# Tools for managing data in The GOA database: Gene Ontology annotation (GOA)
# resource (http://www.ebi.ac.uk/GOA) resource (http://www.ebi.ac.uk/GOA)

# https://academic.oup.com/nar/article/43/D1/D1057/2437623/The-GOA-database-Gene-Ontology-annotation-updates
