"""Data downloaded from NCBI Gene data converted into Python data."""

# Copyright (C) 2014-2018, DV Klopfenstein. All rights reserved

downloaded = "2016_02_22" # 13919 items

# Downloaded as a tsv from NCBI Gene. Then translated to Python:
#
# http://www.ncbi.nlm.nih.gov/gene/?term=genetype+protein+coding%5BProperties%5D+AND+%227227%22%5BTaxonomy+ID%5D+AND+alive%5Bproperty%5D
# 
# or follow link to "Search for Fruit Fly Protein-Coding Genes" found here:
# 
# https://github.com/dvklopfenstein/biocode/blob/master/biodownloads/README.md

import collections as cx

NtData = cx.namedtuple('NtData', 'tax_id Org_name GeneID CurrentID Status Symbol Aliases description other_designations map_location chromosome genomic_nucleotide_accession_version start_position_on_the_genomic_accession end_position_on_the_genomic_accession orientation exon_count OMIM no_hdr0')
GeneID2nt = {
  32768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32768, CurrentID=0, Status='live', Symbol='CG6762', Aliases='Dmel_ Dmel\\CG6762', description='CG6762 gene product from transcript CG6762-RD', other_designations='CG6762-PA|CG6762-PB|CG6762-PC|CG6762-PD', map_location='16E2-16E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17856974, end_position_on_the_genomic_accession=17858610, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32769, CurrentID=0, Status='live', Symbol='Arp8', Aliases='Dmel_CG7846, CG7846, Dmel\\CG7846, arp8, dArp8', description='Actin-related protein 8', other_designations='Arp8-PA|CG7846-PA|actin related protein 8', map_location='16E2-16E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17858379, end_position_on_the_genomic_accession=17860800, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32770, CurrentID=0, Status='live', Symbol='CG6769', Aliases='Dmel_ Dmel\\CG6769', description='CG6769 gene product from transcript CG6769-RB', other_designations='CG6769-PA|CG6769-PB', map_location='16E2-16E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17861002, end_position_on_the_genomic_accession=17862775, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32771, CurrentID=0, Status='live', Symbol='mnb', Aliases='Dmel_CG42273, CG42273, CG7826, CG7835, DYRK1, Dm1, DmDyrk1A, Dmel\\CG42273, Dmel_CG7826, Dmel_CG7835, Dyrk1, Dyrk1A, ME-IV, MNB, Mnb, min', description='minibrain', other_designations='CG42273-PE|CG42273-PF|CG42273-PG|CG42273-PH|CG42273-PI|Dual-specificity tyrosine phosphorylation-regulated kinase 1A|minibrain kinase|mnb-PE|mnb-PF|mnb-PG|mnb-PH|mnb-PI', map_location='1-58.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17862697, end_position_on_the_genomic_accession=17887222, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32773, CurrentID=0, Status='live', Symbol='CG6788', Aliases='Dmel_ AAF48780, Dmel\\CG6788, FBpp0074269', description='CG6788 gene product from transcript CG6788-RA', other_designations='CG6788-PA|Fibrinogen-related protein', map_location='16F1-16F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17888177, end_position_on_the_genomic_accession=17889404, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32774, CurrentID=0, Status='live', Symbol='CG12985', Aliases='Dmel_ Dmel\\CG12985', description='CG12985 gene product from transcript CG12985-RA', other_designations='CG12985-PA|uncharacterized protein', map_location='16F1-16F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17883718, end_position_on_the_genomic_accession=17885238, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32775, CurrentID=0, Status='live', Symbol='GSS', Aliases='Dmel_CG6835, 6835, CG32497, CG33065, CG6835, Dmel\\CG6835, GS', description='Glutathione Synthetase', other_designations='CG6835-PE|CG6835-PF|CG6835-PG|CG6835-PH|CG6835-PI|CG6835-PJ|CG6835-PK|GSS-PE|GSS-PF|GSS-PG|GSS-PH|GSS-PI|GSS-PJ|GSS-PK|glutathione synthase|glutathione synthetase', map_location='16F1-16F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17889547, end_position_on_the_genomic_accession=17898603, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  32776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32776, CurrentID=0, Status='live', Symbol='CG7772', Aliases='Dmel_ Dmel\\CG7772, bs13c01.y1', description='CG7772 gene product from transcript CG7772-RA', other_designations='CG7772-PA', map_location='16F1-16F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17906031, end_position_on_the_genomic_accession=17907261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32777, CurrentID=0, Status='live', Symbol='Vps4', Aliases='Dmel_CG6842, 3B1, BcDNA:GH02678, CG6842, DmBcDNA.GH02678, Dmel\\CG6842, Dvps4B, bcDNA:GH02678, dvps4, vps4, Vps4', description='Vacuolar protein sorting 4', other_designations='CG6842-PA|CG6842-PB|Vps4-PA|Vps4-PB|vacuolar protein sorting 4', map_location='16F1-16F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17907692, end_position_on_the_genomic_accession=17911097, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32778, CurrentID=0, Status='live', Symbol='CG6847', Aliases='Dmel_ Dmel\\CG6847', description='CG6847 gene product from transcript CG6847-RB', other_designations='CG6847-PA|CG6847-PB', map_location='16F3-16F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17911462, end_position_on_the_genomic_accession=17924163, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32780, CurrentID=0, Status='live', Symbol='Sh', Aliases='Dmel_CG12348, BcDNA:GH03046, CG12348, CG17860, CG7640, Dmel\\CG12348, EKO, Eko, F-c, Kv1, SHa, Shw, V-I, minisleep, sh, sha, Sh', description='Shaker', other_designations='CG12348-PA|CG12348-PB|CG12348-PC|CG12348-PD|CG12348-PE|CG12348-PF|CG12348-PH|CG12348-PI|CG12348-PJ|CG12348-PK|CG12348-PL|CG12348-PM|CG12348-PQ|CG12348-PS|CG12348-PT|F-c transcription unit|Sh-PA|Sh-PB|Sh-PC|Sh-PD|Sh-PE|Sh-PF|Sh-PH|Sh-PI|Sh-PJ|Sh-PK|Sh-PL|Sh-PM|Sh-PQ|Sh-PS|Sh-PT|Shaker-downheld|shaker', map_location='1-57.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17924307, end_position_on_the_genomic_accession=18063247, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  32781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32781, CurrentID=0, Status='live', Symbol='CG12672', Aliases='Dmel_ Dmel\\CG12672', description='CG12672 gene product from transcript CG12672-RA', other_designations='CG12672-PA|uncharacterized protein', map_location='16F4-16F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18005897, end_position_on_the_genomic_accession=18007606, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32782, CurrentID=0, Status='live', Symbol='CG6867', Aliases='Dmel_ CT21241, Dmel\\CG6867', description='CG6867 gene product from transcript CG6867-RA', other_designations='CG6867-PA', map_location='16F5-16F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18023870, end_position_on_the_genomic_accession=18028261, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  32784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32784, CurrentID=0, Status='live', Symbol='CG15373', Aliases='Dmel_ Dmel\\CG15373', description='CG15373 gene product from transcript CG15373-RB', other_designations='CG15373-PA|CG15373-PB|uncharacterized protein', map_location='16F5-16F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18044071, end_position_on_the_genomic_accession=18048720, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  32785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32785, CurrentID=0, Status='live', Symbol='Dhc16F', Aliases='Dmel_CG7092, CG7092, DHC, Dhc, Dmel\\CG7092, atoll, cvl-6, dhc', description='Dynein heavy chain at 16F', other_designations='CG7092-PA|Dhc16F-PA|crossveinless-like 6|dynein heavy chain at 16F|dynein-related heavy chain polypeptide', map_location='1-59.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18064483, end_position_on_the_genomic_accession=18078472, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  32786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32786, CurrentID=0, Status='live', Symbol='CG7536', Aliases='Dmel_ Dmel\\CG7536', description='CG7536 gene product from transcript CG7536-RC', other_designations='CG7536-PA|CG7536-PB|CG7536-PC|CG7536-PD', map_location='16F6-16F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18078491, end_position_on_the_genomic_accession=18088532, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32787, CurrentID=0, Status='live', Symbol='Ada3', Aliases='Dmel_CG7098, ADA3, CG7098, DADA3, Dmel\\CG7098, HL-XIV, dADA3, dAda3, dik', description='transcriptional Adaptor 3', other_designations='Ada3-PA|Ada3-PC|CG7098-PA|CG7098-PC|diskette|transcriptional adaptor 3', map_location='16F6-16F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18085761, end_position_on_the_genomic_accession=18088418, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32788, CurrentID=0, Status='live', Symbol='CG7206', Aliases='Dmel_ Dmel\\CG7206', description='CG7206 gene product from transcript CG7206-RB', other_designations='CG7206-PB|CG7206-PC', map_location='16F6-16F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18088725, end_position_on_the_genomic_accession=18090509, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32789, CurrentID=0, Status='live', Symbol='scu', Aliases='Dmel_CG7113, 17beta-HSD10, CG7113, Dmel\\CG7113, HL-X, HLII, Scu, l(1)16Fd, l(1)17Ad', description='scully', other_designations='3-hydroxyacyl-CoA dehydrogenase|CG7113-PA|CG7113-PB|haplo-lethal region II|lethal(1)17Ad|scu-PA|scu-PB', map_location='16F6-16F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18091259, end_position_on_the_genomic_accession=18092853, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32790, CurrentID=0, Status='live', Symbol='RhoGAP16F', Aliases='Dmel_CG7122, CG7122, Dmel\\CG7122', description='Rho GTPase activating protein at 16F', other_designations='CG7122-PB|CG7122-PC|CG7122-PD|CG7122-PE|CG7122-PF|RhoGAP16F-PB|RhoGAP16F-PC|RhoGAP16F-PD|RhoGAP16F-PE|RhoGAP16F-PF', map_location='16F6-16F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18093145, end_position_on_the_genomic_accession=18097413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32791, CurrentID=0, Status='live', Symbol='CG7192', Aliases='Dmel_ Dmel\\CG7192', description='CG7192 gene product from transcript CG7192-RA', other_designations='CG7192-PA|CG7192-PB|CG7192-PD', map_location='16F6-16F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18097558, end_position_on_the_genomic_accession=18101314, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32792, CurrentID=0, Status='live', Symbol='Taf8', Aliases='Dmel_CG7128, CG7128, Dmel\\CG7128, HL-VIII, Prodos, TAF, TAF8, TAF[[II]], TFIID, dmTAF8, l(1)16Fg, pds, pro, prod', description='TBP-associated factor 8', other_designations='CG7128-PA|TBP-associated factor|Taf8-PA|prodos', map_location='16F7-16F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18101970, end_position_on_the_genomic_accession=18103482, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32793, CurrentID=0, Status='live', Symbol='CG7135', Aliases='Dmel_ Dmel\\CG7135', description='CG7135 gene product from transcript CG7135-RA', other_designations='CG7135-PA', map_location='16F7-16F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18103653, end_position_on_the_genomic_accession=18105393, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32794, CurrentID=0, Status='live', Symbol='wupA', Aliases='Dmel_CG7178, CG7178, DROTROPONI, DmTpnI, Dmel\\CG7178, HDP, HL-I, HL08042, HLI, Tn I, TnI, TpnI, WupA, chrX:17947374..17947524, fliH, hdp, hdp-a, hdp2, wup-A', description='wings up A', other_designations='CG7178-PA|CG7178-PB|CG7178-PC|CG7178-PD|CG7178-PE|CG7178-PF|CG7178-PG|CG7178-PH|CG7178-PI|CG7178-PJ|CG7178-PK|CG7178-PL|CG7178-PM|flightless H|flightless-H|haplo-lethal region I|held up|held-up|held-up2|heldup|heldup-a|troponin|troponin I|troponin-I|wings up|wings-up A|wupA-PA|wupA-PB|wupA-PC|wupA-PD|wupA-PE|wupA-PF|wupA-PG|wupA-PH|wupA-PI|wupA-PJ|wupA-PK|wupA-PL|wupA-PM', map_location='1-57.69 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18105568, end_position_on_the_genomic_accession=18116922, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32796, CurrentID=0, Status='live', Symbol='ari-1', Aliases='Dmel_CG5659, AAS65398.1, ARI1_DROME, Ari-1, CG5659, Dmel\\CG5659, HL-V, aria, l(1)16Fj, ari-1', description='ariadne 1', other_designations='Ariadne-1|CG5659-PA|CG5659-PB|CG5659-PC|CG5659-PD|CG5659-PE|CG5659-PF|ari-1-PA|ari-1-PB|ari-1-PC|ari-1-PD|ari-1-PE|ari-1-PF|ariadne', map_location='16F7-16F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18130392, end_position_on_the_genomic_accession=18137784, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32797, CurrentID=0, Status='live', Symbol='Frq1', Aliases='Dmel_CG5744, CG5744, Dmel\\CG5744, Frq, HL-II, frq, frq1', description='Frequenin 1', other_designations='CG5744-PB|CG5744-PC|CG5744-PD|CG5744-PE|Frq1-PB|Frq1-PC|Frq1-PD|Frq1-PE|frequenin|frequenin 1|frequentin|frequinin', map_location='16F7-17A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18155369, end_position_on_the_genomic_accession=18174744, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32798, CurrentID=0, Status='live', Symbol='Andorra', Aliases='Dmel_CG12611, CG12611, Dmel\\CG12611, and, andorra', description='CG12611 gene product from transcript CG12611-RA', other_designations='Andorra-PA|CG12611-PA|andorra', map_location='17A1-17A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18178526, end_position_on_the_genomic_accession=18179472, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32799, CurrentID=0, Status='live', Symbol='Frq2', Aliases='Dmel_CG5907, CG5907, Dmel\\CG5907, Frq, frq, frq2', description='Frequenin 2', other_designations='CG5907-PA|CG5907-PB|CG5907-PC|CG5907-PD|Frq2-PA|Frq2-PB|Frq2-PC|Frq2-PD|frequenin|frequenin 2', map_location='17A1-17A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18177612, end_position_on_the_genomic_accession=18204407, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32800, CurrentID=0, Status='live', Symbol='Hesr', Aliases='Dmel_CG5927, CG5927, Dmel\\CG5927, Her, bHLHc21', description='HES-related', other_designations='CG5927-PA|Hesr-PA', map_location='17A2-17A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18204403, end_position_on_the_genomic_accession=18205144, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32801, CurrentID=0, Status='live', Symbol='Him', Aliases='Dmel_CG15064, CG15064, Dmel\\CG15064, him', description='Holes in muscle', other_designations='CG15064-PA|Him-PA|holes in muscle|holes in muscles|holes-in-muscle', map_location='17A2-17A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18209427, end_position_on_the_genomic_accession=18210741, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32803, CurrentID=0, Status='live', Symbol='CG33639', Aliases='Dmel_ CG16958, CG5936, Dmel\\CG33639, anon-WO0170980.40, anon-WO0170980.41', description='CG33639 gene product from transcript CG33639-RD', other_designations='CG33639-PA|CG33639-PB|CG33639-PC|CG33639-PD|uncharacterized protein', map_location='17A2-17A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18210812, end_position_on_the_genomic_accession=18226076, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  5740834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740834, CurrentID=0, Status='live', Symbol='CG16824', Aliases='Dmel_ Dmel\\CG16824', description='CG16824 gene product from transcript CG16824-RA', other_designations='CG16824-PA|uncharacterized protein', map_location='34B4-34B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13293009, end_position_on_the_genomic_accession=13293415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32805, CurrentID=0, Status='live', Symbol='upd2', Aliases='Dmel_CG5988, CG5988, Dmel\\CG5988, UPD2, Upd-like, Upd2, Upd2/Leptin, cg5988', description='unpaired 2', other_designations='CG5988-PB|Unpaired2|upd homolog|upd2-PB', map_location='17A3-17A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18240654, end_position_on_the_genomic_accession=18243284, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32806, CurrentID=0, Status='live', Symbol='CG15059', Aliases='Dmel_ Dmel\\CG15059', description='CG15059 gene product from transcript CG15059-RC', other_designations='CG15059-PB|CG15059-PC|uncharacterized protein', map_location='17A3-17A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18258551, end_position_on_the_genomic_accession=18259517, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835569, CurrentID=0, Status='live', Symbol='CG45116', Aliases='Dmel_ Dmel\\CG45116', description='CG45116 gene product from transcript CG45116-RA', other_designations='CG45116-PA|uncharacterized protein', map_location='79F6-79F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22699433, end_position_on_the_genomic_accession=22699870, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32813, CurrentID=0, Status='live', Symbol='upd1', Aliases='Dmel_CG5993, CG5993, Dmel\\CG5993, OS, Os, UPD, Upd, Upd1, l(1)YC43, l(1)YM55, od, odsy, os, sis-C, sis-c, sisC, sy, unp, upa, upd', description='unpaired 1', other_designations='CG5993-PA|CG5993-PB|CG5993-PC|Unpaired/Out-stretched|outstretched|outstretched small eye|outstretched-smalleye|sisterless c|small-eye|unpaired|upd homolog|upd1-PA|upd1-PB|upd1-PC', map_location='17A5-17A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18305354, end_position_on_the_genomic_accession=18309291, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32814, CurrentID=0, Status='live', Symbol='CG6023', Aliases='Dmel_ Dmel\\CG6023', description='CG6023 gene product from transcript CG6023-RA', other_designations='CG6023-PA|CG6023-PC', map_location='17A5-17A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18314124, end_position_on_the_genomic_accession=18321193, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32815, CurrentID=0, Status='live', Symbol='CG6123', Aliases='Dmel_ Dmel\\CG6123', description='CG6123 gene product from transcript CG6123-RB', other_designations='CG6123-PA|CG6123-PB', map_location='17A5-17A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18322958, end_position_on_the_genomic_accession=18356191, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32816, CurrentID=0, Status='live', Symbol='CG6106', Aliases='Dmel_ Dmel\\CG6106', description='CG6106 gene product from transcript CG6106-RB', other_designations='CG6106-PA|CG6106-PB', map_location='17A7-17A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18358881, end_position_on_the_genomic_accession=18362527, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32817, CurrentID=0, Status='live', Symbol='CrebB', Aliases='Dmel_CG6103, CG6103, CRE-BP, CREB, CREB-B, CREB-b, CREB2, CREB2a, CREB2b, Creb, Creb B 17A, Creb17A, Creb2, Creb2-b, Creb2.b, Creb2b-17A, CrebB-17B, CrebB17-A, DM9, Dmel\\CG6103, S162, dCREB, dCREB-2, dCREB-2a, dCREB-B, dCREB2, dCREB2-a, dCREB2-b, dCREB2a, dCREB2b, dCbz, dCreb2, dCreb2-b, dCrebB2, dcreb2-b, l(1)17Af, CrebB', description='Cyclic-AMP response element binding protein B', other_designations='CG6103-PF|CG6103-PG|CG6103-PI|CG6103-PJ|CG6103-PM|CG6103-PN|CG6103-PO|CG6103-PP|CrebB-PF|CrebB-PG|CrebB-PI|CrebB-PJ|CrebB-PM|CrebB-PN|CrebB-PO|CrebB-PP|Cyclic-AMP response element binding protein B at 17A|Cyclic-AMP response element binding protein B at 17B|activator|blocker|cAMP response element-binding protein|cAMP response-element binding protein|cAMP-regulated enhancer-binding protein|cAMP-response element binding protein|cAMP-response element-binding protein|cyclic AMP response element binding protein|dCREB2|lethal(1)17Af', map_location='17B-17B', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18366357, end_position_on_the_genomic_accession=18377864, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32818, CurrentID=0, Status='live', Symbol='por', Aliases='Dmel_CG6205, 15175, CG6205, Dmel\\CG6205, Porc, dmPorc, l(1)17Ac, pocc, por', description='porcupine', other_designations='CG6205-PA|CG6205-PB|CG6205-PC|por-PA|por-PB|por-PC', map_location='1-59 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18377521, end_position_on_the_genomic_accession=18379854, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32819, CurrentID=0, Status='live', Symbol='CG6179', Aliases='Dmel_ Dmel\\CG6179, NOSIPL', description='CG6179 gene product from transcript CG6179-RA', other_designations='CG6179-PA', map_location='17A9-17A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18380082, end_position_on_the_genomic_accession=18381479, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32820, CurrentID=0, Status='live', Symbol='betaCOP', Aliases='Dmel_CG6223, BcDNA.GH09317, BcDNA:GH09317, Beta Cop, BetaCop, CG6223, COPI, Dmel\\CG6223, bcop, beta-COP, beta-Cop, beta-CopII, betaCop', description='Coat Protein (coatomer) beta', other_designations='CG6223-PA|COPI beta subunit|beta coatomer protein|beta-coatomer protein|betaCOP-PA|betaCop|coat protein (coatomer) beta', map_location='17A9-17A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18381549, end_position_on_the_genomic_accession=18385242, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32821, CurrentID=0, Status='live', Symbol='Tsf1', Aliases='Dmel_CG6186, 143958_at, CG6186, Dmel\\CG6186, TSF1, Tf, Trf, anon-EST:Posey265, lincRNA.S9601, tsf1', description='Transferrin 1', other_designations='CG6186-PA|CG6186-PC|DTAF-transferrin|Transferrin1|Tsf1-PA|Tsf1-PC|transferrin|transferrin 1|transferrin precursor', map_location='17A9-17A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18387084, end_position_on_the_genomic_accession=18390344, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32822, CurrentID=0, Status='live', Symbol='CG32549', Aliases='Dmel_ 24643014, 24643016, 45555989, 45555998, CG6247, CG6267, CG6278, Dmel\\CG32549', description='CG32549 gene product from transcript CG32549-RD', other_designations='CG32549-PB|CG32549-PD|CG32549-PE|CG32549-PF|CG32549-PG|CG32549-PH|uncharacterized protein', map_location='17A10-17A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18390498, end_position_on_the_genomic_accession=18410144, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32824, CurrentID=0, Status='live', Symbol='CG15056', Aliases='Dmel_ Dmel\\CG15056', description='CG15056 gene product from transcript CG15056-RA', other_designations='CG15056-PA|uncharacterized protein', map_location='17A11-17A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18415415, end_position_on_the_genomic_accession=18416911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32825, CurrentID=0, Status='live', Symbol='CG42323', Aliases='Dmel_ CG15055, CG32550, Dmel\\CG42323, Dmel_CG15055, Dmel_CG32550', description='CG42323 gene product from transcript CG42323-RE', other_designations='CG42323-PC|CG42323-PE|uncharacterized protein', map_location='17B1-17B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18422413, end_position_on_the_genomic_accession=18427001, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32826, CurrentID=0, Status='live', Symbol='CG32548', Aliases='Dmel_ CG15051, CG6306, Dmel\\CG32548, anon-WO0140519.88', description='CG32548 gene product from transcript CG32548-RB', other_designations='CG32548-PB|CG32548-PC|uncharacterized protein', map_location='17B1-17B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18435719, end_position_on_the_genomic_accession=18440581, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32827, CurrentID=0, Status='live', Symbol='CG6290', Aliases='Dmel_ BcDNA:GH02837, Dmel\\CG6290', description='CG6290 gene product from transcript CG6290-RA', other_designations='CG6290-PA', map_location='17B1-17B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18437490, end_position_on_the_genomic_accession=18438233, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32831, CurrentID=0, Status='live', Symbol='Hayan', Aliases='Dmel_CG6361, CG6361, Dmel\\CG6361, SP31, c-SP31', description='CG6361 gene product from transcript CG6361-RB', other_designations='CG6361-PA|CG6361-PB|CG6361-PD|Hayan-PA|Hayan-PB|Hayan-PD|hayan', map_location='17B4-17B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18479983, end_position_on_the_genomic_accession=18483772, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32832, CurrentID=0, Status='live', Symbol='psh', Aliases='Dmel_CG6367, CG6367, Dmel\\CG6367, Psh, SP28, c-SP28', description='persephone', other_designations='CG6367-PA|psh-PA', map_location='17B5-17B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18484487, end_position_on_the_genomic_accession=18486921, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32833, CurrentID=0, Status='live', Symbol='CG15046', Aliases='Dmel_ Dmel\\CG15046', description='CG15046 gene product from transcript CG15046-RA', other_designations='CG15046-PA|uncharacterized protein', map_location='17B5-17B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18487450, end_position_on_the_genomic_accession=18489620, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32834, CurrentID=0, Status='live', Symbol='CG15044', Aliases='Dmel_ BcDNA:RE30049, Dmel\\CG15044', description='CG15044 gene product from transcript CG15044-RA', other_designations='CG15044-PA|uncharacterized protein', map_location='17B5-17B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18491185, end_position_on_the_genomic_accession=18492006, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32835, CurrentID=0, Status='live', Symbol='CG15043', Aliases='Dmel_ BcDNA:RE41194, Dmel\\CG15043', description='CG15043 gene product from transcript CG15043-RB', other_designations='CG15043-PA|CG15043-PB|uncharacterized protein', map_location='17B5-17B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18493193, end_position_on_the_genomic_accession=18494168, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32836, CurrentID=0, Status='live', Symbol='GalNAc-T2', Aliases='Dmel_CG6394, CG6394, Dmel\\CG6394, PGANT7, anon-WO02059370.15, dGalNAc-T2, pgant7', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 2', other_designations='CG6394-PA|CG6394-PB|CG6394-PC|GalNAc-T2-PA|GalNAc-T2-PB|GalNAc-T2-PC|polypeptide GalNAc transferase 7', map_location='17B6-17B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18495067, end_position_on_the_genomic_accession=18499986, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32837, CurrentID=0, Status='live', Symbol='Rad51D', Aliases='Dmel_CG6318, CG6318, Dm CG6318, Dmel\\CG6318, dm', description='CG6318 gene product from transcript CG6318-RA', other_designations='CG6318-PA|Rad51D-PA', map_location='17B6-17B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18500127, end_position_on_the_genomic_accession=18501258, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32838, CurrentID=0, Status='live', Symbol='Wnt5', Aliases='Dmel_CG6407, CG6407, DWnt-3, DWnt-3/5, DWnt-5, DWnt3, DWnt3/5, DWnt5, Dm DWnt3/5, Dm-3, Dmel\\CG6407, Dwnt-3, Dwnt-5, Dwnt5, WNT5, Wnt, Wnt-3, Wnt3, Wnt3/5, anon-WO03040301.228, wnt3, wnt5', description='Wnt oncogene analog 5', other_designations='CG6407-PA|CG6407-PB|Wnt5-PA|Wnt5-PB|wnt-oncogene-analog-3', map_location='17B6-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18501401, end_position_on_the_genomic_accession=18505405, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32839, CurrentID=0, Status='live', Symbol='Ggt-1', Aliases='Dmel_CG6461, CG6461, Dmel\\CG6461, GGT-1', description='gamma-glutamyl transpeptidase', other_designations='CG6461-PA|CG6461-PB|Ggt-1-PA|Ggt-1-PB|gamma-glutamyl transferase', map_location='17C1-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18505883, end_position_on_the_genomic_accession=18512521, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32840, CurrentID=0, Status='live', Symbol='CG6470', Aliases='Dmel_ Dmel\\CG6470, anon-WO0140519.137', description='CG6470 gene product from transcript CG6470-RB', other_designations='CG6470-PA|CG6470-PB', map_location='17C1-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18512829, end_position_on_the_genomic_accession=18514867, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32841, CurrentID=0, Status='live', Symbol='Aats-his', Aliases='Dmel_CG6335, CG6335, Dmel\\CG6335, HRS, Harsl, HisRS, l(1)G0165, l(1)G0358', description='Histidyl-tRNA synthetase', other_designations='Aats-his-PA|Aats-his-PC|Aats-his-PD|Aats-his-PE|CG6335-PA|CG6335-PC|CG6335-PD|CG6335-PE', map_location='17C1-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18515428, end_position_on_the_genomic_accession=18520445, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32843, CurrentID=0, Status='live', Symbol='CG6481', Aliases='Dmel_ Dmel\\CG6481', description='CG6481 gene product from transcript CG6481-RA', other_designations='CG6481-PA', map_location='17C1-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18521074, end_position_on_the_genomic_accession=18522445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32844, CurrentID=0, Status='live', Symbol='CG15042', Aliases='Dmel_ Dmel\\CG15042', description='CG15042 gene product from transcript CG15042-RA', other_designations='CG15042-PA|uncharacterized protein', map_location='17C1-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18526862, end_position_on_the_genomic_accession=18528266, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32845, CurrentID=0, Status='live', Symbol='CG15047', Aliases='Dmel_ Dmel\\CG15047', description='CG15047 gene product from transcript CG15047-RA', other_designations='CG15047-PA|uncharacterized protein', map_location='17C1-17C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18528683, end_position_on_the_genomic_accession=18529593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32846, CurrentID=0, Status='live', Symbol='Bx', Aliases='Dmel_CG44425, Bd, CG10548, CG15048, CG44425, CG6500, DLMO, Dlmo, Dmel\\CG44425, Dmel_CG15048, Dmel_CG6500, LMO, Lmo, Ptd, Rho, beadex/dLMO, dLMO, dLmo, dlmo, dttg, fliH, hdp, hdp-a, hld', description='Beadex', other_designations='Bx-PA|Bx-PB|Bx-PC|Bx-PD|Bx-PF|CG44425-PA|CG44425-PB|CG44425-PC|CG44425-PD|CG44425-PF|LIM-only|beadex|flightless-H|held-up|heldup a|pointedoid|rhombotin', map_location='1-59.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18515428, end_position_on_the_genomic_accession=18572809, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768909, CurrentID=0, Status='live', Symbol='CG15744', Aliases='Dmel_ CT35992, Dmel\\CG15744, Q7KV24_DROME', description='CG15744 gene product from transcript CG15744-RA', other_designations='CG15744-PA|uncharacterized protein', map_location='11E9-11E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13288051, end_position_on_the_genomic_accession=13294433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32848, CurrentID=0, Status='live', Symbol='CG15040', Aliases='Dmel_ Dmel\\CG15040', description='CG15040 gene product from transcript CG15040-RA', other_designations='CG15040-PA|uncharacterized protein', map_location='17C5-17C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18606610, end_position_on_the_genomic_accession=18617589, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32849, CurrentID=0, Status='live', Symbol='wgn', Aliases='Dmel_CG6531, CG6531, Dmel\\CG6531, TNFR, Wgn, vader', description='wengen', other_designations='CG6531-PA|CG6531-PB|TNF-alpha receptor|wgn-PA|wgn-PB', map_location='17C5-17C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18624042, end_position_on_the_genomic_accession=18634908, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32850, CurrentID=0, Status='live', Symbol='Rip11', Aliases='Dmel_CG6606, CG6606, Dmel\\CG6606, dRip11, l(1)G0003, l(1)G0070, l(1)G0092, l(1)G0205, l(1)G0297, l(1)G0434', description='Rab11 interacting protein', other_designations='CG6606-PA|CG6606-PB|CG6606-PC|CG6606-PD|CG6606-PE|Rip11-PA|Rip11-PB|Rip11-PC|Rip11-PD|Rip11-PE|lethal (1) G0003', map_location='17C6-17C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18635642, end_position_on_the_genomic_accession=18647648, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32851, CurrentID=0, Status='live', Symbol='CG6540', Aliases='Dmel_ Dmel\\CG6540, Nup35', description='CG6540 gene product from transcript CG6540-RA', other_designations='CG6540-PA', map_location='17C7-17C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18648182, end_position_on_the_genomic_accession=18649489, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32852, CurrentID=0, Status='live', Symbol='CG6617', Aliases='Dmel_ Dmel\\CG6617, anon-EST:Posey180', description='CG6617 gene product from transcript CG6617-RA', other_designations='CG6617-PA', map_location='17C7-17C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18649396, end_position_on_the_genomic_accession=18650444, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32853, CurrentID=0, Status='live', Symbol='Ing3', Aliases='Dmel_CG6632, CG6632, Dmel\\CG6632, ING3, dIng3', description='Inhibitor of growth family, member 3', other_designations='CG6632-PA|Ing3-PA|inhibitor of growth family, member 3', map_location='17C7-17C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18650718, end_position_on_the_genomic_accession=18653091, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32854, CurrentID=0, Status='live', Symbol='CG6659', Aliases='Dmel_ Dmel\\CG6659', description='CG6659 gene product from transcript CG6659-RA', other_designations='CG6659-PA|CG6659-PB', map_location='17C7-17D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18653392, end_position_on_the_genomic_accession=18657286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32855, CurrentID=0, Status='live', Symbol='fu', Aliases='Dmel_CG6551, CG6551, Dm fu, Dmel\\CG6551, FU, Fu, dFu[mel], l(1)fu, fu', description='fused', other_designations='CG6551-PA|fu-PA', map_location='1-59.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18657755, end_position_on_the_genomic_accession=18661204, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32856, CurrentID=0, Status='live', Symbol='CG6696', Aliases='Dmel_ Dmel\\CG6696', description='CG6696 gene product from transcript CG6696-RA', other_designations='CG6696-PA', map_location='17D1-17D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18660472, end_position_on_the_genomic_accession=18662153, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32857, CurrentID=0, Status='live', Symbol='phm', Aliases='Dmel_CG6578, 306a1, CG6578, CYP306A1, Cyp306a1, Cyp306a1-Dm, Dmel\\CG6578, PHM, exo, ptm', description='phantom', other_designations='CG6578-PA|CG6578-PB|cytochrome P450|exocephalon|phm-PA|phm-PB', map_location='1-64 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18685766, end_position_on_the_genomic_accession=18688433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32858, CurrentID=0, Status='live', Symbol='Cyp18a1', Aliases='Dmel_CG6816, 181H6T, 18a1, CG6816, CT20826, CYP18, CYP18A1-Dm, Cyp18, Cyp18A1, Cyt-P450-17C, CytP450, Dmel\\CG6816, ESTS:181H6T, Eig17-1, P-450, P450, cyp18, cyp18a1', description='Cytochrome P450-18a1', other_designations='CG6816-PA|CG6816-PB|Cyp18a1-PA|Cyp18a1-PB|cytochrome P450 17C|cytochrome P450-18a1', map_location='17D1-17D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18688538, end_position_on_the_genomic_accession=18693042, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32859, CurrentID=0, Status='live', Symbol='Cyp308a1', Aliases='Dmel_CG6585, 308a1, CG6585, Dmel\\CG6585', description='CG6585 gene product from transcript CG6585-RA', other_designations='CG6585-PA|Cyp308a1-PA', map_location='17D1-17D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18709865, end_position_on_the_genomic_accession=18711638, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32860, CurrentID=0, Status='live', Symbol='CCKLR-17D1', Aliases='Dmel_CG42301, CCKLR, CCKLR-D1, CG42301, CG6857, DSK-R2, Dm CG6857, Dmel\\CG42301, Dmel_CG6857, DskR-2, anon-WO0170980.61, anon-WO0170980.62', description='Cholecystokinin-like receptor at 17D1', other_designations='CCK-like receptor at 17D1|CCKLR-17D1-PA|CG42301-PA|cholecystokinin-like receptor', map_location='17D1-17D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18707494, end_position_on_the_genomic_accession=18726572, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32861, CurrentID=0, Status='live', Symbol='CG6873', Aliases='Dmel_ Dmel\\CG6873, adf', description='CG6873 gene product from transcript CG6873-RA', other_designations='Adf/cofilin-like|CG6873-PA|cofilin homologue', map_location='17D3-17D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18739216, end_position_on_the_genomic_accession=18739895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32862, CurrentID=0, Status='live', Symbol='CG12609', Aliases='Dmel_ Dmel\\CG12609', description='CG12609 gene product from transcript CG12609-RB', other_designations='CG12609-PB|uncharacterized protein', map_location='17D3-17D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18746367, end_position_on_the_genomic_accession=18748306, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32864, CurrentID=0, Status='live', Symbol='CCKLR-17D3', Aliases='Dmel_CG32540, CCKLR-D3, CG32540, CG6861, CG6881, CG6894, DSK-R1, Dmel\\CG32540, DskR-1, anon-WO0131005.21, anon-WO0170980.64, anon-WO0170980.65', description='Cholecystokinin-like receptor at 17D3', other_designations='CCK-like receptor at 17D3|CCKLR-17D3-PB|CG32540-PB|DSK-receptor2|sulfakinin receptor', map_location='17D3-17D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18764483, end_position_on_the_genomic_accession=18774027, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32865, CurrentID=0, Status='live', Symbol='CG6891', Aliases='Dmel_ Dmel\\CG6891', description='CG6891 gene product from transcript CG6891-RA', other_designations='CG6891-PA|CG6891-PB|CG6891-PC', map_location='17D4-17D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18775228, end_position_on_the_genomic_accession=18781345, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32866, CurrentID=0, Status='live', Symbol='CG18259', Aliases='Dmel_ Dmel\\CG18259', description='CG18259 gene product from transcript CG18259-RA', other_designations='CG18259-PA|uncharacterized protein', map_location='17D5-17D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18781265, end_position_on_the_genomic_accession=18783494, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32869, CurrentID=0, Status='live', Symbol='CG6961', Aliases='Dmel_ Dmel\\CG6961, cg6961', description='CG6961 gene product from transcript CG6961-RA', other_designations='CG6961-PA', map_location='17D5-17D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18786142, end_position_on_the_genomic_accession=18788210, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32871, CurrentID=0, Status='live', Symbol='Atg101', Aliases='Dmel_CG7053, CG7053, Dmel\\CG7053', description='Autophagy-related 101', other_designations='Atg101-PA|Autophagy-specific gene 101|CG7053-PA', map_location='17D6-17D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18791668, end_position_on_the_genomic_accession=18793517, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32872, CurrentID=0, Status='live', Symbol='bnb', Aliases='Dmel_CG7088, Bnb, CG7088, Dmel\\CG7088, GAP43, Gap43', description='bangles and beads', other_designations='Bangles-n-beads|CG7088-PA|CG7088-PB|CG7088-PC|CG7088-PD|bnb-PA|bnb-PB|bnb-PC|bnb-PD', map_location='17D6-17D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18794246, end_position_on_the_genomic_accession=18797687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32873, CurrentID=0, Status='live', Symbol='CG7058', Aliases='Dmel_ Dmel\\CG7058', description='CG7058 gene product from transcript CG7058-RB', other_designations='CG7058-PB', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18806694, end_position_on_the_genomic_accession=18812002, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  32874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32874, CurrentID=0, Status='live', Symbol='CG42450', Aliases='Dmel_ CG32543, CG7095, Dmel\\CG42450, Dmel_CG32543, Dmel_CG7095', description='CG42450 gene product from transcript CG42450-RB', other_designations='CG42450-PA|CG42450-PB|uncharacterized protein', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18813297, end_position_on_the_genomic_accession=18830406, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32875, CurrentID=0, Status='live', Symbol='CG7101', Aliases='Dmel_ Dmel\\CG7101', description='CG7101 gene product from transcript CG7101-RA', other_designations='CG7101-PA', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18830591, end_position_on_the_genomic_accession=18832123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32876, CurrentID=0, Status='live', Symbol='Pvf1', Aliases='Dmel_CG7103, CG7103, DmVEGF-1, Dmel\\CG7103, EP1624, PVF, PVF1, PVFs, Pvf, VEGF, VEGF2-B, Vegf17E, Vegf17e-a, Vegf17e-b, l(1)G0146, pvf1, vegf1', description='PDGF- and VEGF-related factor 1', other_designations='CG7103-PA|CG7103-PB|CG7103-PC|PDGF (platelet-derived growth factor)- and VEGF (vascular endothelial growth factor)-related factor 1|Pvf1-PA|Pvf1-PB|Pvf1-PC', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18832682, end_position_on_the_genomic_accession=18842901, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32877, CurrentID=0, Status='live', Symbol='CG34422', Aliases='Dmel_ ARID4B, CG7274, CG7282, Dmel\\CG34422, Dmel_CG7274, Dmel_CG7282, Q9VWP3_DROME', description='CG34422 gene product from transcript CG34422-RD', other_designations='CG34422-PC|CG34422-PD|uncharacterized protein', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18843663, end_position_on_the_genomic_accession=18855858, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32878, CurrentID=0, Status='live', Symbol='CG7280', Aliases='Dmel_ Dmel\\CG7280', description='CG7280 gene product from transcript CG7280-RA', other_designations='CG7280-PA', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18846389, end_position_on_the_genomic_accession=18848566, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32880, CurrentID=0, Status='live', Symbol='CG7322', Aliases='Dmel_ Dmel\\CG7322', description='CG7322 gene product from transcript CG7322-RA', other_designations='CG7322-PA|CG7322-PB|CG7322-PC', map_location='17E1-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18855873, end_position_on_the_genomic_accession=18857757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32881, CurrentID=0, Status='live', Symbol='Usp39', Aliases='Dmel_CG7288, CG7288, Dmel\\CG7288, Q9VWP1', description='Ubiquitin specific protease 39', other_designations='CG7288-PA|Usp39-PA|ubiquitin specific protease 39', map_location='17F1-17F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18858203, end_position_on_the_genomic_accession=18860072, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32882, CurrentID=0, Status='live', Symbol='CG7326', Aliases='Dmel_ Dmel\\CG7326', description='CG7326 gene product from transcript CG7326-RB', other_designations='CG7326-PA|CG7326-PB', map_location='17F1-17F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18861197, end_position_on_the_genomic_accession=18863776, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32885, CurrentID=0, Status='live', Symbol='CG7332', Aliases='Dmel_ Dmel\\CG7332', description='CG7332 gene product from transcript CG7332-RB', other_designations='CG7332-PA|CG7332-PB', map_location='17F2-17F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18884513, end_position_on_the_genomic_accession=18888860, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32886, CurrentID=0, Status='live', Symbol='CG7358', Aliases='Dmel_ Dmel\\CG7358', description='CG7358 gene product from transcript CG7358-RB', other_designations='CG7358-PA|CG7358-PB|CG7358-PC', map_location='17F2-17F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18888978, end_position_on_the_genomic_accession=18893636, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32887, CurrentID=0, Status='live', Symbol='SdhBL', Aliases='Dmel_CG7349, CG7349, DHSB, Dmel\\CG7349', description='Succinate dehydrogenase, subunit B (iron-sulfur)-like', other_designations='CG7349-PA|CG7349-PC|SdhBL-PA|SdhBL-PC|ferredoxin|succinate dehydrogenase Iron-sulfur protein|succinate dehydrogenase [ubiquinone] iron-sulfur protein|succinate dehydrogenase, subunit B (iron-sulfur)-like', map_location='17F2-17F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18894446, end_position_on_the_genomic_accession=18896581, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32888, CurrentID=0, Status='live', Symbol='CG7378', Aliases='Dmel_ BcDNA:RH25447, Dmel\\CG7378, MKP-like', description='CG7378 gene product from transcript CG7378-RC', other_designations='CG7378-PA|CG7378-PB|CG7378-PC|MKP-like', map_location='17F2-17F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18900410, end_position_on_the_genomic_accession=18911843, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32889, CurrentID=0, Status='live', Symbol='CG43759', Aliases='Dmel_ CG12527, CG14188, CG32541, CG42506, Dmel\\CG43759, Dmel_CG32541, Dmel_CG42506, EP(X)1644, LD15404', description='CG43759 gene product from transcript CG43759-RC', other_designations='CG43759-PA|CG43759-PB|CG43759-PC|CG43759-PD|uncharacterized protein', map_location='17F3-18A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18930335, end_position_on_the_genomic_accession=19028629, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32891, CurrentID=0, Status='live', Symbol='CG14190', Aliases='Dmel_ Dmel\\CG14190', description='CG14190 gene product from transcript CG14190-RA', other_designations='CG14190-PA|uncharacterized protein', map_location='18A2-18A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19056911, end_position_on_the_genomic_accession=19067792, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32892, CurrentID=0, Status='live', Symbol='CG7406', Aliases='Dmel_ Dmel\\CG7406', description='CG7406 gene product from transcript CG7406-RA', other_designations='CG7406-PA', map_location='18A3-18A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19077720, end_position_on_the_genomic_accession=19078592, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32893, CurrentID=0, Status='live', Symbol='CG14191', Aliases='Dmel_ BcDNA:RE09269, Dmel\\CG14191', description='CG14191 gene product from transcript CG14191-RA', other_designations='CG14191-PA|uncharacterized protein', map_location='18A3-18A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19081665, end_position_on_the_genomic_accession=19082566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32894, CurrentID=0, Status='live', Symbol='CG7423', Aliases='Dmel_ Dmel\\CG7423', description='CG7423 gene product from transcript CG7423-RA', other_designations='CG7423-PA', map_location='18A3-18A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19082609, end_position_on_the_genomic_accession=19083290, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32895, CurrentID=0, Status='live', Symbol='CG15882', Aliases='Dmel_ Dmel\\CG15882', description='CG15882 gene product from transcript CG15882-RA', other_designations='CG15882-PA|uncharacterized protein', map_location='18A3-18A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19123755, end_position_on_the_genomic_accession=19124388, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32896, CurrentID=0, Status='live', Symbol='tgy', Aliases='Dmel_CG7440, CG7440, Dmel\\CG7440, Tgy, dC1GalT5', description='twiggy', other_designations='CG7440-PC|CG7440-PE|tgy-PC|tgy-PE', map_location='18A3-18A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19125003, end_position_on_the_genomic_accession=19127316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32897, CurrentID=0, Status='live', Symbol='Obp18a', Aliases='Dmel_CG15883, 18a, BcDNA:RH25665, CG15883, DmelObp18a, Dmel\\CG15883, obp18a', description='Odorant-binding protein 18a', other_designations='CG15883-PA|Obp18a-PA', map_location='18A3-18A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19135081, end_position_on_the_genomic_accession=19135868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32898, CurrentID=0, Status='live', Symbol='RhoGAP18B', Aliases='Dmel_CG42274, 5-131, CG17015, CG42274, CG7481, CG7502, CG7531, Dmel\\CG42274, Dmel_CG7481, Dmel_CG7502, whir', description='Rho GTPase activating protein at 18B', other_designations='CG42274-PB|CG42274-PC|CG42274-PD|CG42274-PF|CG42274-PH|RhoGAP18B-PB|RhoGAP18B-PC|RhoGAP18B-PD|RhoGAP18B-PF|RhoGAP18B-PH|white rabbit', map_location='18A3-18A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19136107, end_position_on_the_genomic_accession=19167082, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32901, CurrentID=0, Status='live', Symbol='Inx5', Aliases='Dmel_CG7537, CG7537, Dm-inx5, Dmel\\CG7537, FBpp0074467, inx-5, inx5', description='Innexin 5', other_designations='CG7537-PB|Innexin-5|Inx5-PB|innexin 5|inx5', map_location='18A4-18A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19168777, end_position_on_the_genomic_accession=19170883, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32902, CurrentID=0, Status='live', Symbol='CG7556', Aliases='Dmel_ Dmel\\CG7556', description='CG7556 gene product from transcript CG7556-RB', other_designations='CG7556-PA|CG7556-PB', map_location='18A5-18A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19176109, end_position_on_the_genomic_accession=19179830, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32903, CurrentID=0, Status='live', Symbol='CG7453', Aliases='Dmel_ Dmel\\CG7453', description='CG7453 gene product from transcript CG7453-RA', other_designations='CG7453-PA', map_location='18A5-18A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19180173, end_position_on_the_genomic_accession=19182024, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32905, CurrentID=0, Status='live', Symbol='Mec2', Aliases='Dmel_CG7635, CG7635, D Dmel\\CG7635', description='CG7635 gene product from transcript CG7635-RA', other_designations='CG7635-PA|Mec2-PA', map_location='18A6-18A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19188035, end_position_on_the_genomic_accession=19190193, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32906, CurrentID=0, Status='live', Symbol='HP1D3csd', Aliases='Dmel_CG14193, CG14193, Dmel\\CG14193', description='Heterochromatin Protein 1D3 chromoshadow domain', other_designations='CG14193-PA|HP1D3csd-PA|heterochromatin protein 1D3 chromoshadow domain', map_location='18A6-18A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19190214, end_position_on_the_genomic_accession=19191029, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32907, CurrentID=0, Status='live', Symbol='CG7914', Aliases='Dmel_ Dmel\\CG7914', description='CG7914 gene product from transcript CG7914-RA', other_designations='CG7914-PA', map_location='18A7-18A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19192219, end_position_on_the_genomic_accession=19193448, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32908, CurrentID=0, Status='live', Symbol='CG14194', Aliases='Dmel_ Dmel\\CG14194', description='CG14194 gene product from transcript CG14194-RA', other_designations='CG14194-PA|CG14194-PB|uncharacterized protein', map_location='18A7-18A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19193555, end_position_on_the_genomic_accession=19196037, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32909, CurrentID=0, Status='live', Symbol='CG7990', Aliases='Dmel_ Dmel\\CG7990', description='CG7990 gene product from transcript CG7990-RA', other_designations='CG7990-PA|CG7990-PC|CG7990-PD', map_location='18A7-18B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19196058, end_position_on_the_genomic_accession=19201979, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32910, CurrentID=0, Status='live', Symbol='Ulp1', Aliases='Dmel_CG12359, BcDNA:GH02751, CG12359, Dm Dmel\\CG12359, ulp1', description='CG12359 gene product from transcript CG12359-RA', other_designations='CG12359-PA|Ulp1-PA|ubiquitin-like protein-specific protease 1', map_location='18B1-18B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19202103, end_position_on_the_genomic_accession=19208187, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32911, CurrentID=0, Status='live', Symbol='CG14195', Aliases='Dmel_ Dmel\\CG14195', description='CG14195 gene product from transcript CG14195-RB', other_designations='CG14195-PA|CG14195-PB|uncharacterized protein', map_location='18B1-18B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19208694, end_position_on_the_genomic_accession=19212082, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32912, CurrentID=0, Status='live', Symbol='Mur18B', Aliases='Dmel_CG7874, CG7874, Dmel\\CG7874, anon-WO0140519.120', description='Mucin related 18B', other_designations='CG7874-PA|Mur18B-PA|mucin related 18B', map_location='18B1-18B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19213254, end_position_on_the_genomic_accession=19214938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32913, CurrentID=0, Status='live', Symbol='Muc18B', Aliases='Dmel_CG7876, CG7876, Dmel\\CG7876', description='Mucin 18B', other_designations='CG7876-PA|CG7876-PB|Muc18B-PA|Muc18B-PB|mucin 18B', map_location='18B1-18B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19215515, end_position_on_the_genomic_accession=19216660, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740440, CurrentID=0, Status='live', Symbol='CG34190', Aliases='Dmel_ Dmel\\CG34190', description='CG34190 gene product from transcript CG34190-RA', other_designations='CG34190-PA|uncharacterized protein', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16946455, end_position_on_the_genomic_accession=16947554, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32915, CurrentID=0, Status='live', Symbol='CG14196', Aliases='Dmel_ Dmel\\CG14196', description='CG14196 gene product from transcript CG14196-RA', other_designations='CG14196-PA|uncharacterized protein', map_location='18B4-18B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19236059, end_position_on_the_genomic_accession=19241840, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32916, CurrentID=0, Status='live', Symbol='CG7889', Aliases='Dmel_ Dmel\\CG7889', description='CG7889 gene product from transcript CG7889-RA', other_designations='CG7889-PA', map_location='18B4-18B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19243631, end_position_on_the_genomic_accession=19245224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32917, CurrentID=0, Status='live', Symbol='CG7992', Aliases='Dmel_ Dmel\\CG7992', description='CG7992 gene product from transcript CG7992-RA', other_designations='CG7992-PA', map_location='18B4-18B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19245158, end_position_on_the_genomic_accession=19247250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32918, CurrentID=0, Status='live', Symbol='Hs3st-B', Aliases='Dmel_CG7890, CG7890, Dmel\\CG7890, dHS3OSTB, dHs3st-B', description='Heparan sulfate 3-O sulfotransferase-B', other_designations='3-O-sulfotransferase|CG7890-PA|CG7890-PB|Hs3st-B-PA|Hs3st-B-PB|heparan sulfate 3-O sulfotransferase-B', map_location='18B5-18B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19248639, end_position_on_the_genomic_accession=19252045, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32919, CurrentID=0, Status='live', Symbol='rictor', Aliases='Dmel_CG8002, 8002, CG8002, Dmel\\CG8002, Rictor, dRic, dRictor', description='rapamycin-insensitive companion of Tor', other_designations='CG8002-PA|CG8002-PB|CG8002-PC|dRictor|rictor-PA|rictor-PB|rictor-PC', map_location='18B6-18B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19252111, end_position_on_the_genomic_accession=19261240, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32920, CurrentID=0, Status='live', Symbol='Vav', Aliases='Dmel_CG7893, BcDNA:GH01128, CG7893, DmVav, Dmel\\CG7893, DroVav, vav', description='Vav ortholog (H. sapiens)', other_designations='CG7893-PA|CG7893-PB|CG7893-PC|CG7893-PD|Vav ortholog|Vav-PA|Vav-PB|Vav-PC|Vav-PD|vav', map_location='18B6-18B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19261820, end_position_on_the_genomic_accession=19273280, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32921, CurrentID=0, Status='live', Symbol='CG8010', Aliases='Dmel_ Dmel\\CG8010', description='CG8010 gene product from transcript CG8010-RA', other_designations='CG8010-PA', map_location='18B7-18B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19266656, end_position_on_the_genomic_accession=19268896, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32922, CurrentID=0, Status='live', Symbol='CG32536', Aliases='Dmel_ CG8020, Dmel\\CG32536', description='CG32536 gene product from transcript CG32536-RA', other_designations='CG32536-PA|uncharacterized protein', map_location='18B8-18B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19275004, end_position_on_the_genomic_accession=19276898, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32923, CurrentID=0, Status='live', Symbol='CG8028', Aliases='Dmel_ Dmel\\CG8028, MCT', description='CG8028 gene product from transcript CG8028-RC', other_designations='CG8028-PC|CG8028-PD|monocarboxylate transporter', map_location='18B8-18B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19277176, end_position_on_the_genomic_accession=19280349, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32924, CurrentID=0, Status='live', Symbol='CG8034', Aliases='Dmel_ Dmel\\CG8034', description='CG8034 gene product from transcript CG8034-RA', other_designations='CG8034-PA|CG8034-PB', map_location='18B9-18C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19281232, end_position_on_the_genomic_accession=19298162, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32925, CurrentID=0, Status='live', Symbol='CG8051', Aliases='Dmel_ Dmel\\CG8051, EP(X)1550, anon-WO03002137.1', description='CG8051 gene product from transcript CG8051-RA', other_designations='CG8051-PA|CG8051-PB', map_location='18C1-18C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19301330, end_position_on_the_genomic_accession=19307925, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32926, CurrentID=0, Status='live', Symbol='out', Aliases='Dmel_CG8062, BcDNA:LD28120, CG8062, Dmel\\CG8062', description='outsiders', other_designations='CG8062-PA|CG8062-PB|out-PA|out-PB|outsider', map_location='18C2-18C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19310513, end_position_on_the_genomic_accession=19322106, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32928, CurrentID=0, Status='live', Symbol='nAChRalpha7', Aliases='Dmel_CG32538, CG32538, CG8082, CG8109, CT24218, Da7, Dalpha7, Dalpha7 nAChR, Dalpha7-nAChR, Dmel\\CG32538, gfA, gfa, jpl, nAChR, nAChR-Dalpha7, nAcRalpha-18C, nAcRalpha-7, nj507, nj520, nj522, nj54, nj75', description='nicotinic Acetylcholine Receptor alpha7', other_designations='CG32538-PA|CG32538-PD|CG32538-PE|CG32538-PF|Dalpha7|Dalpha7 subunit of the acetylcholine receptor|Dalpha7 subunit of the nicotinic acetylcholine receptor|giant fiber A|giant fibre A|jumpless|nAChRalpha7-PA|nAChRalpha7-PD|nAChRalpha7-PE|nAChRalpha7-PF|nicotinic acetylcholine receptor alpha 18C|nicotinic acetylcholine receptor alpha7|non-jumper-507|non-jumper-520|non-jumper-522|non-jumper-54|non-jumper-75|shaker cognate L', map_location='1-62 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19327075, end_position_on_the_genomic_accession=19348447, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  32930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32930, CurrentID=0, Status='live', Symbol='kek5', Aliases='Dmel_CG12199, CG12199, CT10486, Dmel\\CG12199, Kek5', description='kekkon5', other_designations='CG12199-PA|CG12199-PB|CG12199-PC|CG12199-PD|Kekkon5|kek5-PA|kek5-PB|kek5-PC|kek5-PD|kekkon-like', map_location='18C3-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19353181, end_position_on_the_genomic_accession=19471914, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32932, CurrentID=0, Status='live', Symbol='CG12200', Aliases='Dmel_ Dmel\\CG12200', description='CG12200 gene product from transcript CG12200-RA', other_designations='CG12200-PA|uncharacterized protein', map_location='18C7-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19475063, end_position_on_the_genomic_accession=19476055, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32933, CurrentID=0, Status='live', Symbol='CG32533', Aliases='Dmel_ CG12211, CG14198, Dmel\\CG32533, cg12211', description='CG32533 gene product from transcript CG32533-RA', other_designations='CG32533-PA|CG32533-PB|uncharacterized protein', map_location='18C7-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19474156, end_position_on_the_genomic_accession=19479411, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32934, CurrentID=0, Status='live', Symbol='Nat1', Aliases='Dmel_CG12202, CG12202, Dmel\\CG12202, dNat1, l(1)G0311', description='NAT1 ortholog (S. cerevisiae)', other_designations='CG12202-PA|CG12202-PB|NAT1 ortholog|Nat1-PA|Nat1-PB', map_location='18C7-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19479984, end_position_on_the_genomic_accession=19484643, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32935, CurrentID=0, Status='live', Symbol='pcm', Aliases='Dmel_CG3291, CG3291, DmXrn1, Dmel\\CG3291, Pcm, XRN1, XRN1/pcm, Xrn1, Xrn1p', description='pacman', other_designations='CG3291-PA|CG3291-PB|pcm-PA|pcm-PB', map_location='18C7-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19485759, end_position_on_the_genomic_accession=19492011, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32936, CurrentID=0, Status='live', Symbol='ND-18', Aliases='Dmel_CG12203, CG12203, Dmel\\CG12203, IP11020p, NUYM', description='NADH dehydrogenase (ubiquinone) 18 kDa subunit', other_designations='CG12203-PA|FBpp0074481|NADH ubiquinone oxidoreductase 18 kDa|NADH-ubiquinone oxidoreductase 18 kDa subunit|NADH:ubiquinone oxidoreductase 18 kDa subunit|ND-18-PA', map_location='18C7-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19492421, end_position_on_the_genomic_accession=19493337, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32937, CurrentID=0, Status='live', Symbol='CG12204', Aliases='Dmel_ Dmel\\CG12204', description='CG12204 gene product from transcript CG12204-RA', other_designations='CG12204-PA|uncharacterized protein', map_location='18C7-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19493292, end_position_on_the_genomic_accession=19494401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32938, CurrentID=0, Status='live', Symbol='Pfrx', Aliases='Dmel_CG3400, CG3400, Dmel\\CG3400, PFK2', description='6-phosphofructo-2-kinase', other_designations='CG3400-PA|CG3400-PB|CG3400-PD|CG3400-PE|CG3400-PF|CG3400-PG|CG3400-PH|CG3400-PI|CG3400-PJ|CG3400-PK|CG3400-PL|CG3400-PM|CG3400-PN|Pfrx-PA|Pfrx-PB|Pfrx-PD|Pfrx-PE|Pfrx-PF|Pfrx-PG|Pfrx-PH|Pfrx-PI|Pfrx-PJ|Pfrx-PK|Pfrx-PL|Pfrx-PM|Pfrx-PN|fi-phosphofructo-2-kinase|fructose 2,6 bisphosphatase|phosphofructokinase', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19495505, end_position_on_the_genomic_accession=19509350, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32939, CurrentID=0, Status='live', Symbol='CG14200', Aliases='Dmel_ Dmel\\CG14200', description='CG14200 gene product from transcript CG14200-RA', other_designations='CG14200-PA|CG14200-PB|CG14200-PC|uncharacterized protein', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19495505, end_position_on_the_genomic_accession=19518402, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32940, CurrentID=0, Status='live', Symbol='l(1)G0156', Aliases='Dmel_CG12233, CG12233, Dmel\\CG12233, FBgn0031024, l (1) G0156', description='lethal (1) G0156', other_designations='CG12233-PA|CG12233-PC|CG12233-PD|Lethal(1)G0156|isocitrate dehydrogenase|l(1)G0156-PA|l(1)G0156-PC|l(1)G0156-PD', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19518428, end_position_on_the_genomic_accession=19521810, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32941, CurrentID=0, Status='live', Symbol='CoRest', Aliases='Dmel_CG42687, BcDNA:GH01486, BcDNA:LD11215, CG32534, CG32535, CG33525, CG3878, CG42687, CoREST, CoRESTG, Dmel\\CG42687, Dmel_CG32534, Dmel_CG33525, dCoREST, i169', description='CG42687 gene product from transcript CG42687-RC', other_designations='CG42687-PA|CG42687-PC|CG42687-PD|CG42687-PE|CG42687-PF|CG42687-PG|CG42687-PH|CG42687-PI|CG42687-PJ|CG42687-PK|CoRest-PA|CoRest-PC|CoRest-PD|CoRest-PE|CoRest-PF|CoRest-PG|CoRest-PH|CoRest-PI|CoRest-PJ|CoRest-PK', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19521929, end_position_on_the_genomic_accession=19530896, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  32942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32942, CurrentID=0, Status='live', Symbol='CG12231', Aliases='Dmel_ Dmel\\CG12231', description='CG12231 gene product from transcript CG12231-RA', other_designations='CG12231-PA|uncharacterized protein', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19531032, end_position_on_the_genomic_accession=19531742, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32943, CurrentID=0, Status='live', Symbol='CG32532', Aliases='Dmel_ CG14201, CG14202, CG14203, Dmel\\CG32532', description='CG32532 gene product from transcript CG32532-RB', other_designations='CG32532-PB|CG32532-PC|CG32532-PD|uncharacterized protein', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19531851, end_position_on_the_genomic_accession=19547482, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32946, CurrentID=0, Status='live', Symbol='Grip84', Aliases='Dmel_CG3917, CG3917, DGrip84, Dgrip84, Dgrips84, Dmel\\CG3917, GCP2, grip84, l(1)[[t]]PG36, l(1)[[t]]PG66', description='Gamma-tubulin ring protein 84', other_designations='CG3917-PA|CG3917-PB|CG3917-PC|CG3917-PD|CG3917-PE|CG3917-PF|DGrip84|Grip84-PA|Grip84-PB|Grip84-PC|Grip84-PD|Grip84-PE|Grip84-PF|gamma-tubulin ring protein 84', map_location='18D1-18D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19559431, end_position_on_the_genomic_accession=19565764, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32947, CurrentID=0, Status='live', Symbol='car', Aliases='Dmel_CG12230, CG12230, Car, Dm-Vps33, Dmel\\CG12230, VPS33a, Vps33, Vps33/carnation, Vps33A, Vps33a, Vps33pn, l(1)G0447, car', description='carnation', other_designations='CG12230-PA|CG12230-PB|CG12230-PC|car-PA|car-PB|car-PC|vacuolar protein sorting protein 33', map_location='1-62.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19566019, end_position_on_the_genomic_accession=19569401, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32948, CurrentID=0, Status='live', Symbol='Tao', Aliases='Dmel_CG14217, AAF48973, CG14217, Dmel\\CG14217, MARKK, TAO1-1, dTao-1, tao, tao-1, tao1', description='CG14217 gene product from transcript CG14217-RF', other_designations='CG14217-PA|CG14217-PB|CG14217-PD|CG14217-PE|CG14217-PF|CG14217-PG|CG14217-PH|CG14217-PI|CG14217-PJ|Tao-1|Tao-PA|Tao-PB|Tao-PD|Tao-PE|Tao-PF|Tao-PG|Tao-PH|Tao-PI|Tao-PJ', map_location='18D2-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19569721, end_position_on_the_genomic_accession=19579800, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  32949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32949, CurrentID=0, Status='live', Symbol='CG14218', Aliases='Dmel_ Dmel\\CG14218, SPH211', description='CG14218 gene product from transcript CG14218-RC', other_designations='CG14218-PB|CG14218-PC|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19580350, end_position_on_the_genomic_accession=19581981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32950, CurrentID=0, Status='live', Symbol='CG14204', Aliases='Dmel_ Dmel\\CG14204', description='CG14204 gene product from transcript CG14204-RB', other_designations='CG14204-PA|CG14204-PB|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19583617, end_position_on_the_genomic_accession=19587057, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32951, CurrentID=0, Status='live', Symbol='CG14219', Aliases='Dmel_ Dmel\\CG14219', description='CG14219 gene product from transcript CG14219-RA', other_designations='CG14219-PA|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19587348, end_position_on_the_genomic_accession=19590170, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32952, CurrentID=0, Status='live', Symbol='CG14205', Aliases='Dmel_ Dmel\\CG14205', description='CG14205 gene product from transcript CG14205-RA', other_designations='CG14205-PA|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19592609, end_position_on_the_genomic_accession=19595548, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32953, CurrentID=0, Status='live', Symbol='RpS10b', Aliases='Dmel_CG14206, CG14206, CT33819, Dmel\\CG14206, GM19176, M(1)18C, M(1)36f, M(1)n, M[[n]], Rp S10 b, S10b, i3, i73, l(1)carot13', description='Ribosomal protein S10b', other_designations='CG14206-PB|CG14206-PC|CG14206-PD|CG14206-PE|Minute-n|RpS10b-PB|RpS10b-PC|RpS10b-PD|RpS10b-PE|minute (1) 18C|ribosomal protein S10b', map_location='1-62.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19598031, end_position_on_the_genomic_accession=19599806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32954, CurrentID=0, Status='live', Symbol='CG14220', Aliases='Dmel_ Dmel\\CG14220, SDS3, Sds3, cg14220', description='CG14220 gene product from transcript CG14220-RB', other_designations='CG14220-PA|CG14220-PB|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19599803, end_position_on_the_genomic_accession=19601851, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32955, CurrentID=0, Status='live', Symbol='CG14207', Aliases='Dmel_ Dmel\\CG14207, Hsp20-like protein, HspB8, dHspB8', description='CG14207 gene product from transcript CG14207-RD', other_designations='CG14207-PA|CG14207-PB|CG14207-PC|CG14207-PD|HspB8|small heat shock protein hsp20 family|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19605199, end_position_on_the_genomic_accession=19611413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19835594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835594, CurrentID=0, Status='live', Symbol='CG44815', Aliases='Dmel_ Dmel\\CG44815', description='CG44815 gene product from transcript CG44815-RA', other_designations='CG44815-PA|uncharacterized protein', map_location='8E2-8E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9445815, end_position_on_the_genomic_accession=9446212, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32958, CurrentID=0, Status='live', Symbol='CG14210', Aliases='Dmel_ BcDNA:RE23450, Dmel\\CG14210', description='CG14210 gene product from transcript CG14210-RA', other_designations='CG14210-PA|CG14210-PB|uncharacterized protein', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19616196, end_position_on_the_genomic_accession=19616825, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32961, CurrentID=0, Status='live', Symbol='CG14221', Aliases='Dmel_ Dmel\\CG14221', description='CG14221 gene product from transcript CG14221-RC', other_designations='CG14221-PB|CG14221-PC|uncharacterized protein', map_location='18D7-18D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19618819, end_position_on_the_genomic_accession=19621245, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32962, CurrentID=0, Status='live', Symbol='NAA20', Aliases='Dmel_CG14222, CG14222, DmAAF48987, Dmel\\CG14222, dNAA20', description='N(alpha)-acetyltransferase 20', other_designations='CG14222-PA|CG14222-PB|CG14222-PC|CG14222-PD|NAA20 ortholog|NAA20-PA|NAA20-PB|NAA20-PC|NAA20-PD', map_location='18D7-18D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19621451, end_position_on_the_genomic_accession=19623074, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32963, CurrentID=0, Status='live', Symbol='MKP-4', Aliases='Dmel_CG14211, CG14211, DMKP-4, Dmel\\CG14211, Mkp4', description='MAPK Phosphatase 4', other_designations='CG14211-PB|CG14211-PC|MAPK phosphatase 4|MKP-4-PB|MKP-4-PC|MKP-like', map_location='18D7-18D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19623114, end_position_on_the_genomic_accession=19625729, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32964, CurrentID=0, Status='live', Symbol='CG14212', Aliases='Dmel_ 19527979, CG14122, Dmel\\CG14212', description='CG14212 gene product from transcript CG14212-RA', other_designations='CG14212-PA|uncharacterized protein', map_location='18D8-18D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19625705, end_position_on_the_genomic_accession=19626601, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32965, CurrentID=0, Status='live', Symbol='e(y)3', Aliases='Dmel_CG12238, CG12238, Dmel\\CG12238, SAYP, e(y)[3], l(1)G0084, l(1)G0266, l(1)G0283, l(1)G0381, l(1)G0409, sayp', description='enhancer of yellow 3', other_designations='CG12238-PA|CG12238-PB|CG12238-PC|CG12238-PD|CG12238-PE|e(y)3-PA|e(y)3-PB|e(y)3-PC|e(y)3-PD|e(y)3-PE|lethal (1) G0084|supporter of activation of yellow protein', map_location='1-62.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19626548, end_position_on_the_genomic_accession=19638369, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  32966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32966, CurrentID=0, Status='live', Symbol='Rcd-1', Aliases='Dmel_CG14213, CAF40, CG14213, Caf40, Dmel\\CG14213, Drcd-1', description='Required for cell differentiation 1 ortholog', other_designations='CG14213-PA|CG14213-PB|CG14213-PC|Rcd-1-PA|Rcd-1-PB|Rcd-1-PC|required for cell differentiation 1|required for cell differentiation 1 ortholog', map_location='18D11-18D11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19638767, end_position_on_the_genomic_accession=19640746, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32967, CurrentID=0, Status='live', Symbol='CG12237', Aliases='Dmel_ 16769716, Dmel\\CG12237', description='CG12237 gene product from transcript CG12237-RA', other_designations='CG12237-PA|uncharacterized protein', map_location='18D11-18D11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19641191, end_position_on_the_genomic_accession=19642869, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32968, CurrentID=0, Status='live', Symbol='Sec61gamma', Aliases='Dmel_CG14214, CG14214, Dmel\\CG14214, Sec61', description='Sec61 gamma subunit', other_designations='CG14214-PA|CG14214-PB|CG14214-PC|Sec61gamma-PA|Sec61gamma-PB|Sec61gamma-PC', map_location='18D11-18D12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19643327, end_position_on_the_genomic_accession=19644518, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32969, CurrentID=0, Status='live', Symbol='Arp10', Aliases='Dmel_CG12235, ARP11, Arp11, CG12235, Dmel\\CG12235', description='Actin-related protein 10', other_designations='Arp10-PA|Arp10-PB|Arp11|CG12235-PA|CG12235-PB', map_location='18D12-18D12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19644531, end_position_on_the_genomic_accession=19646176, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32970, CurrentID=0, Status='live', Symbol='Ranbp21', Aliases='Dmel_CG12234, CG12234, Dmel\\CG12234, Exp-5, Exp5, RanBP21, dmExp5', description='CG12234 gene product from transcript CG12234-RB', other_designations='CG12234-PA|CG12234-PB|Exportin-5|Ranbp21-PA|Ranbp21-PB|exportin 5', map_location='18D12-18D13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19646306, end_position_on_the_genomic_accession=19652269, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32971, CurrentID=0, Status='live', Symbol='CG14215', Aliases='Dmel_ Dmel\\CG14215', description='CG14215 gene product from transcript CG14215-RA', other_designations='CG14215-PA|CG14215-PB|uncharacterized protein', map_location='18D13-18D13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19652305, end_position_on_the_genomic_accession=19659407, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32972, CurrentID=0, Status='live', Symbol='CG14223', Aliases='Dmel_ Dmel\\CG14223', description='CG14223 gene product from transcript CG14223-RA', other_designations='CG14223-PA|CG14223-PB|CG14223-PC|uncharacterized protein', map_location='18D13-18D13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19659442, end_position_on_the_genomic_accession=19663971, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32973, CurrentID=0, Status='live', Symbol='Ssu72', Aliases='Dmel_CG14216, BEST:CK01830, CG14216, Dmel\\CG14216, dSsu72', description='Ssu72 CTD phosphatase', other_designations='CG14216-PA|CG14216-PB|CG14216-PC|Ssu72 ortholog|Ssu72-PA|Ssu72-PB|Ssu72-PC', map_location='18D13-18D13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19664561, end_position_on_the_genomic_accession=19666866, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32974, CurrentID=0, Status='live', Symbol='Zw', Aliases='Dmel_CG12529, CG12529, Dmel\\CG12529, G-6-PD, G-6-pdh, G-6PD, G-6pd, G6-PD, G6-pd, G6PD, G6PDH, G6pd, G6pdh, g6pd, zw', description='Zwischenferment', other_designations='CG12529-PA|CG12529-PB|CG12529-PC|Glucose-6-phosphate DH|Zw-PA|Zw-PB|Zw-PC|glucose-6-phosphate 1dehydrogenase|glucose-6-phosphate dehydrogenase|zwischenferment', map_location='1-62.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19667252, end_position_on_the_genomic_accession=19672353, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32975, CurrentID=0, Status='live', Symbol='et', Aliases='Dmel_CG14225, CG14225, Dmel\\CG14225, ET, lat', description='eye transformer', other_designations='CG14225-PB|CG14225-PC|et-PB|et-PC|latran', map_location='18D13-18D13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19672312, end_position_on_the_genomic_accession=19675891, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32976, CurrentID=0, Status='live', Symbol='dome', Aliases='Dmel_CG14226, 142932_at, CG14226, CT33841, DOME, Dmel\\CG14226, Dome, anon-18DEb, l(1)G0217, l(1)G0218, l(1)G0264, l(1)G0282, l(1)G0321, l(1)G0367, l(1)G0405, l(1)G0441, l(1)G0468, mom, vsp', description='domeless', other_designations='CG14226-PA|dome receptor|dome-PA|master of marelle|verstopft', map_location='18D13-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19676061, end_position_on_the_genomic_accession=19683518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32977, CurrentID=0, Status='live', Symbol='Ubqn', Aliases='Dmel_CG14224, BEST:LD26923, CG14224, Chap, DSK2, Dmel\\CG14224, Dsk2, UBQLN homolog, UBQLN1, anon-18DEa, clone 5.42, dUbqln, dUbqn, ubqn', description='Ubiquilin', other_designations='CG14224-PA|CG14224-PB|Ubqn-PA|Ubqn-PB|ubiquilin|ubiquilin 1', map_location='18E1-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19683808, end_position_on_the_genomic_accession=19687658, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32978, CurrentID=0, Status='live', Symbol='CG14227', Aliases='Dmel_ Dmel\\CG14227, SPH158', description='CG14227 gene product from transcript CG14227-RB', other_designations='CG14227-PB|uncharacterized protein', map_location='18E1-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19687791, end_position_on_the_genomic_accession=19688948, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32979, CurrentID=0, Status='live', Symbol='Mer', Aliases='Dmel_CG14228, BG01543, CG14228, D-Mer, Dmel\\CG14228, Dmerlin, EMR2, Emr2, NF2, mer', description='Merlin', other_designations='BG01543|CG14228-PA|CG14228-PB|D merlin|Ezrin-moesin-radixin-2|Mer-PA|Mer-PB|Merlin/Neurofibromin 2|dmerlin|merlin', map_location='18E1-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19689697, end_position_on_the_genomic_accession=19693500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32980, CurrentID=0, Status='live', Symbol='CG14229', Aliases='Dmel_ BcDNA:RE31307, Dmel\\CG14229', description='CG14229 gene product from transcript CG14229-RB', other_designations='CG14229-PA|CG14229-PB|CG14229-PC|uncharacterized protein', map_location='18E1-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19693977, end_position_on_the_genomic_accession=19695250, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32981, CurrentID=0, Status='live', Symbol='Cdc42', Aliases='Dmel_CG12530, CDC42, CG12530Dm, D-CDC42, D-Cdc42, D-cdc42, DCDC42, DCdc42, Dcdc42, Dm Cdc42, DmCDC42, Dmcdc42, Dmel\\CG12530, P40793, cdc42', description='CG12530 gene product from transcript CG12530-RC', other_designations='CG12530-PA|CG12530-PC|CG12530-PD|Cdc42-PA|Cdc42-PC|Cdc42-PD|Cdc42-like', map_location='18E1-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19697091, end_position_on_the_genomic_accession=19699938, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32982, CurrentID=0, Status='live', Symbol='CG14231', Aliases='Dmel_ Dmel\\CG14231', description='CG14231 gene product from transcript CG14231-RB', other_designations='CG14231-PA|CG14231-PB|uncharacterized protein', map_location='18E1-18E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19699820, end_position_on_the_genomic_accession=19701755, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32983, CurrentID=0, Status='live', Symbol='CG14232', Aliases='Dmel_ Dmel\\CG14232, anon-18DEc', description='CG14232 gene product from transcript CG14232-RA', other_designations='CG14232-PA|CG14232-PB|uncharacterized protein', map_location='18E1-18E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19701927, end_position_on_the_genomic_accession=19704259, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32984, CurrentID=0, Status='live', Symbol='CG14230', Aliases='Dmel_ Dmel\\CG14230, Nol8, anon-18DEe, cg14230, clone 9.42', description='CG14230 gene product from transcript CG14230-RA', other_designations='CG14230-PA|uncharacterized protein', map_location='18E1-18E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19704176, end_position_on_the_genomic_accession=19706417, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32986, CurrentID=0, Status='live', Symbol='CG12531', Aliases='Dmel_ Dmel\\CG12531', description='CG12531 gene product from transcript CG12531-RA', other_designations='CG12531-PA|CG12531-PB|uncharacterized protein', map_location='18E3-18E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19718923, end_position_on_the_genomic_accession=19730199, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  32987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32987, CurrentID=0, Status='live', Symbol='AP-1-2beta', Aliases='Dmel_CG12532, AP-1/2beta, AP-1beta1, AP-2beta, AP-beta1, AP1beta1, AP2beta2, BAD1, BAP, Bap, Beta-adaptin, CG12532, Dmel\\CG12532, apl1, apl2, bap, beta1/2, cg12532', description='Adaptor Protein complex 1/2, beta subunit', other_designations='AP-1-2beta-PA|CG12532-PA|adaptor protein complex 1/2, beta subunit|beta adaptin|beta adaptin drosophila 1|beta-adaptin', map_location='18E4-18E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19731004, end_position_on_the_genomic_accession=19734581, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32988, CurrentID=0, Status='live', Symbol='CG14234', Aliases='Dmel_ Dmel\\CG14234', description='CG14234 gene product from transcript CG14234-RA', other_designations='CG14234-PA|uncharacterized protein', map_location='18E5-18E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19735570, end_position_on_the_genomic_accession=19740456, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32989, CurrentID=0, Status='live', Symbol='COX6B', Aliases='Dmel_CG14235, CG14235, COXG, CcoVIb, CoVIb, Dmel\\CG14235, VIb', description='Cytochrome c oxidase subunit 6B', other_designations='CG14235-PA|CG14235-PB|CG14235-PC|CG14235-PD|COX6B-PA|COX6B-PB|COX6B-PC|COX6B-PD|cytochrome c oxidase VIb|cytochrome c oxidase polypeptide VIb|cytochrome c oxidase subunit 6B|cytochrome c oxidase subunit VIb|cytochrome c oxidase subunitVIb|subunit VIb', map_location='18E5-18E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19740537, end_position_on_the_genomic_accession=19744335, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32990, CurrentID=0, Status='live', Symbol='parvin', Aliases='Dmel_CG32528, CG12533, CG32528, Dmel\\CG32528', description='CG32528 gene product from transcript CG32528-RB', other_designations='CG32528-PA|CG32528-PB|parvin-PA|parvin-PB', map_location='18E5-18F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19746041, end_position_on_the_genomic_accession=19748458, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32991, CurrentID=0, Status='live', Symbol='Alr', Aliases='Dmel_CG12534, CG12534, Dmel\\CG12534, alr', description='Augmenter of liver regeneration', other_designations='Alr-PA|Alr-PB|CG12534-PA|CG12534-PB|augmenter of liver regeneration', map_location='18F1-18F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19748314, end_position_on_the_genomic_accession=19750167, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32992, CurrentID=0, Status='live', Symbol='Pmp70', Aliases='Dmel_CG12703, CG12703, Dmel\\CG12703, PMP70, i17', description='CG12703 gene product from transcript CG12703-RD', other_designations='CG12703-PA|CG12703-PB|CG12703-PC|CG12703-PD|Pmp70-PA|Pmp70-PB|Pmp70-PC|Pmp70-PD', map_location='18F1-18F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19750770, end_position_on_the_genomic_accession=19759280, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32993, CurrentID=0, Status='live', Symbol='CG12702', Aliases='Dmel_ Dmel\\CG12702, lincRNA.1023', description='CG12702 gene product from transcript CG12702-RC', other_designations='CG12702-PA|CG12702-PB|CG12702-PC|uncharacterized protein', map_location='18F1-18F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19759173, end_position_on_the_genomic_accession=19764874, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32994, CurrentID=0, Status='live', Symbol='vfl', Aliases='Dmel_CG12701, CG12701, Dmel\\CG12701, EP1344, Vfl, ZLD, Zld, l(1)G0225, l(1)G0353, l(1)G0427, zld', description='vielfaltig', other_designations='CG12701-PA|CG12701-PB|CG12701-PD|CG12701-PF|CG12701-PG|EP1344|ZELDA|lethal (1) G0225|lethal (1) G0353|lethal (1) G0427|vfl-PA|vfl-PB|vfl-PD|vfl-PF|vfl-PG|vielfaltig/Zelda|zinc-finger early drosophila activator', map_location='18F-18F', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19766465, end_position_on_the_genomic_accession=19783548, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32995, CurrentID=0, Status='live', Symbol='SkpD', Aliases='Dmel_CG12700, BcDNA:AT18217, CG12700, Dmel\\CG12700, Skp, dskpD, skpD', description='SKP1-related D', other_designations='CG12700-PA|SkpD-PA|skpD', map_location='18F2-18F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19812729, end_position_on_the_genomic_accession=19813384, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32996, CurrentID=0, Status='live', Symbol='SkpC', Aliases='Dmel_CG11941, CG11941, Dmel\\CG11941, Skp, dskpC, skpC', description='SKP1-related C', other_designations='CG11941-PA|SkpC-PA|skpC', map_location='18F2-18F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19816072, end_position_on_the_genomic_accession=19816700, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32997, CurrentID=0, Status='live', Symbol='SkpE', Aliases='Dmel_CG11942, CG11942, Dmel\\CG11942, Skp, skpE', description='SKP1-related E', other_designations='CG11942-PA|SkpE-PA|skpE', map_location='18F2-18F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19823368, end_position_on_the_genomic_accession=19824080, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32999, CurrentID=0, Status='live', Symbol='amn', Aliases='Dmel_CG11937, Amn, CG11937, Dmel\\CG11937, Drm-amn, EP(X)1442, PACAP, PACAP-38, PACAP-like, PACAP38, PACAP38-like, Pacap38, chpd', description='amnesiac', other_designations='CG11937-PA|amn-PA|cheap date|cheapdate|pituitary adenylyl cyclase-activating polypeptide|pituitary adenylyl cyclase-activating polypeptide-like', map_location='18F4-19A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19886108, end_position_on_the_genomic_accession=19890060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33000, CurrentID=0, Status='live', Symbol='Hers', Aliases='Dmel_CG32529, CG11936, CG15619, CG32529, Dmel\\CG32529, HERS', description='Histone gene-specific Epigenetic Repressor in late S phase', other_designations='CG32529-PA|CG32529-PC|CG32529-PD|CG32529-PE|CG32529-PF|CG32529-PG|CG32529-PH|CG32529-PI|Hers-PA|Hers-PC|Hers-PD|Hers-PE|Hers-PF|Hers-PG|Hers-PH|Hers-PI|histone gene-specific epigenetic repressor in late S phase', map_location='18F4-19A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19866146, end_position_on_the_genomic_accession=19908108, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  33001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33001, CurrentID=0, Status='live', Symbol='CG15618', Aliases='Dmel_ Dmel\\CG15618', description='CG15618 gene product from transcript CG15618-RB', other_designations='CG15618-PB|CG15618-PC|uncharacterized protein', map_location='18F4-18F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19859743, end_position_on_the_genomic_accession=19865524, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33002, CurrentID=0, Status='live', Symbol='Nup205', Aliases='Dmel_CG11943, CG11943, Dmel\\CG11943, Ov4, anon-EST:Posey281, nup205', description='Nucleoporin 205kD', other_designations='CG11943-PA|CG11943-PB|Nup205-PA|Nup205-PB|nucleoporin 205|nucleoporin 205kD', map_location='18F4-18F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19852214, end_position_on_the_genomic_accession=19859379, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33003, CurrentID=0, Status='live', Symbol='pico', Aliases='Dmel_CG11940, BcDNA:LP01106, CG11940, Dmel\\CG11940, Pico, jog', description='CG11940 gene product from transcript CG11940-RB', other_designations='CG11940-PA|CG11940-PB|CG11940-PC|CG11940-PD|john glenn|pico-PA|pico-PB|pico-PC|pico-PD', map_location='18F2-18F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19830215, end_position_on_the_genomic_accession=19852059, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33004, CurrentID=0, Status='live', Symbol='CG12655', Aliases='Dmel_ Dmel\\CG12655', description='CG12655 gene product from transcript CG12655-RA', other_designations='CG12655-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19958393, end_position_on_the_genomic_accession=19959403, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33005, CurrentID=0, Status='live', Symbol='Nep3', Aliases='Dmel_CG9565, CEC, CG9565, DmeNEP3, Dmel\\CG9565, NEP3, dNEP3, nep3', description='Neprilysin 3', other_designations='CG9565-PA|CG9565-PB|CG9565-PC|Nep3-PA|Nep3-PB|Nep3-PC|neprilysin 3', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19961297, end_position_on_the_genomic_accession=19969323, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  33006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33006, CurrentID=0, Status='live', Symbol='CG17003', Aliases='Dmel_ Dmel\\CG17003', description='CG17003 gene product from transcript CG17003-RA', other_designations='CG17003-PA|uncharacterized protein', map_location='19A4-19A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20000968, end_position_on_the_genomic_accession=20002096, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33007, CurrentID=0, Status='live', Symbol='Dop2R', Aliases='Dmel_CG33517, CG17004, CG33517, CG9569, D2R, DD2R, Dmel\\CG33517, anon-WO0170980.130, anon-WO0170980.131, anon-WO0170980.31, anon-WO0170980.32, dD2R', description='Dopamine 2-like receptor', other_designations='CG33517-PB|CG33517-PF|CG33517-PG|CG33517-PH|CG33517-PI|CG33517-PJ|D2-like receptor|D[[2]] receptor|D[[2]]-like receptor|Dop2R-PB|Dop2R-PF|Dop2R-PG|Dop2R-PH|Dop2R-PI|Dop2R-PJ|dopamine 2-like receptor|dopamine D2 receptor', map_location='19A4-19A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20004680, end_position_on_the_genomic_accession=20062706, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  33009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33009, CurrentID=0, Status='live', Symbol='CG9570', Aliases='Dmel_ Dmel\\CG9570', description='CG9570 gene product from transcript CG9570-RA', other_designations='CG9570-PA', map_location='19A6-19A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20051294, end_position_on_the_genomic_accession=20052519, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33010, CurrentID=0, Status='live', Symbol='fd19B', Aliases='Dmel_CG9571, CG9571, Dmcg9571, Dmel\\CG9571', description='forkhead domain 19B', other_designations='CG9571-PA|fd19B-PA', map_location='19B1-19B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20091428, end_position_on_the_genomic_accession=20092408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33011, CurrentID=0, Status='live', Symbol='CG34331', Aliases='Dmel_ Dmel\\CG34331', description='CG34331 gene product from transcript CG34331-RC', other_designations='CG34331-PB|CG34331-PC|uncharacterized protein', map_location='19B1-19B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20122453, end_position_on_the_genomic_accession=20125271, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33012, CurrentID=0, Status='live', Symbol='CG15322', Aliases='Dmel_ Dmel\\CG15322', description='CG15322 gene product from transcript CG15322-RC', other_designations='CG15322-PB|CG15322-PC|uncharacterized protein', map_location='19B3-19B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20133579, end_position_on_the_genomic_accession=20138878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33013, CurrentID=0, Status='live', Symbol='CG9572', Aliases='Dmel_ CT17356, Dmel\\CG9572', description='CG9572 gene product from transcript CG9572-RA', other_designations='CG9572-PA|CG9572-PB|CG9572-PD', map_location='19B3-19B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20148124, end_position_on_the_genomic_accession=20155514, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33014, CurrentID=0, Status='live', Symbol='Rab35', Aliases='Dmel_CG9575, CG9575, D DmRab35, Dmel\\CG9575, Q95RH7, dRab35, rab35', description='CG9575 gene product from transcript CG9575-RD', other_designations='CG9575-PA|CG9575-PB|CG9575-PC|CG9575-PD|Rab35-PA|Rab35-PB|Rab35-PC|Rab35-PD', map_location='19B3-19B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20155766, end_position_on_the_genomic_accession=20159872, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33015, CurrentID=0, Status='live', Symbol='Phf7', Aliases='Dmel_CG9576, CG9576, Dmel\\CG9576', description='PHD finger protein 7 ortholog', other_designations='CG9576-PA|CG9576-PB|CG9576-PC|Phf7-PA|Phf7-PB|Phf7-PC', map_location='19B3-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20160355, end_position_on_the_genomic_accession=20165830, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33016, CurrentID=0, Status='live', Symbol='CG9577', Aliases='Dmel_ Dmel\\CG9577', description='CG9577 gene product from transcript CG9577-RA', other_designations='CG9577-PA', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20166673, end_position_on_the_genomic_accession=20168080, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33017, CurrentID=0, Status='live', Symbol='CG9581', Aliases='Dmel_ Dmel\\CG9581', description='CG9581 gene product from transcript CG9581-RA', other_designations='CG9581-PA', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20168042, end_position_on_the_genomic_accession=20169981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33018, CurrentID=0, Status='live', Symbol='CG9578', Aliases='Dmel_ Dmel\\CG9578', description='CG9578 gene product from transcript CG9578-RA', other_designations='CG9578-PA', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20170222, end_position_on_the_genomic_accession=20171526, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33019, CurrentID=0, Status='live', Symbol='AnxB10', Aliases='Dmel_CG9579, ANXB10, Ann, AnnX, Annx, CG9579, Dmel\\CG9579, annexin, anxX', description='Annexin B10', other_designations='AnxB10-PA|AnxB10-PB|CG9579-PA|CG9579-PB|annexin B10|annexin X', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20171445, end_position_on_the_genomic_accession=20173514, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33022, CurrentID=0, Status='live', Symbol='obst-A', Aliases='Dmel_CG17052, CG17052, Dmel\\CG17052, obst', description='obstructor-A', other_designations='CG17052-PA|CG17052-PB|Obstructor-A|obst-A-PA|obst-A-PB|obstructor', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20239855, end_position_on_the_genomic_accession=20243209, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33023, CurrentID=0, Status='live', Symbol='Peritrophin-A', Aliases='Dmel_CG17058, CG17058, CG31559, Dmel\\CG17058, PeriA, obst-D', description='Peritrophin A', other_designations='CG17058-PA|CG17058-PB|CG17058-PC|Dm-peritrophin-A|Peritrophin-A-PA|Peritrophin-A-PB|Peritrophin-A-PC|peritrophin A', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20244647, end_position_on_the_genomic_accession=20248203, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33024, CurrentID=0, Status='live', Symbol='CG17068', Aliases='Dmel_ Dmel\\CG17068', description='CG17068 gene product from transcript CG17068-RA', other_designations='CG17068-PA|uncharacterized protein', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20248309, end_position_on_the_genomic_accession=20251093, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33025, CurrentID=0, Status='live', Symbol='Rab10', Aliases='Dmel_CG17060, AAF50924, CG17060, DRAB10, D DmRab10, Dmel\\CG17060, O15971, Rab-r10, rab10', description='CG17060 gene product from transcript CG17060-RA', other_designations='CG17060-PA|Rab-protein 10|Rab10-PA', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20251338, end_position_on_the_genomic_accession=20254691, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33026, CurrentID=0, Status='live', Symbol='CG17065', Aliases='Dmel_ Dmel\\CG17065', description='CG17065 gene product from transcript CG17065-RD', other_designations='CG17065-PA|CG17065-PB|CG17065-PD|uncharacterized protein', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20254692, end_position_on_the_genomic_accession=20257199, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33027, CurrentID=0, Status='live', Symbol='Inx7', Aliases='Dmel_CG2977, CG2977, CG32731, D-inx-7, Dm-inx7, Dmel\\CG2977, FBpp0070994, inx-7, inx7, prp7', description='Innexin 7', other_designations='CG2977-PA|Inx7-PA|innexin 7|innexin7|pas related protein 7', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6991612, end_position_on_the_genomic_accession=6993612, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33028, CurrentID=0, Status='live', Symbol='CG1504', Aliases='Dmel_ 1504, CT3673, Dmel\\CG1504', description='CG1504 gene product from transcript CG1504-RE', other_designations='CG1504-PB|CG1504-PC|CG1504-PD|CG1504-PE', map_location='19C3-19C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20302015, end_position_on_the_genomic_accession=20330372, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33029, CurrentID=0, Status='live', Symbol='CG1631', Aliases='Dmel_ Dmel\\CG1631', description='CG1631 gene product from transcript CG1631-RA', other_designations='CG1631-PA', map_location='19C4-19C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20327444, end_position_on_the_genomic_accession=20328680, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33031, CurrentID=0, Status='live', Symbol='CG12679', Aliases='Dmel_ Dmel\\CG12679', description='CG12679 gene product from transcript CG12679-RA', other_designations='CG12679-PA|uncharacterized protein', map_location='19C5-19C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20375764, end_position_on_the_genomic_accession=20376522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33032, CurrentID=0, Status='live', Symbol='jb', Aliases='Dmel_CG15460, CG15460, Dmel\\CG15460', description='jean-baptiste', other_designations='CG15460-PC|jb-PC', map_location='19C5-19C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20390096, end_position_on_the_genomic_accession=20390806, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33033, CurrentID=0, Status='live', Symbol='l(1)G0004', Aliases='Dmel_CG11738, CG11738, Dmel\\CG11738, l(G0004)', description='lethal (1) G0004', other_designations='CG11738-PA|l(1)G0004-PA', map_location='19C5-19C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20391447, end_position_on_the_genomic_accession=20392401, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33034, CurrentID=0, Status='live', Symbol='Syx16', Aliases='Dmel_CG1467, CG1467, Dm Syx16, DmSyx16, Dmel\\CG1467, SYX16, Syx, dSyx 16, dSyx16', description='Syntaxin 16', other_designations='CG1467-PA|Syntaxin16|Syx16-PA|dSyntaxin16|syntaxin|syntaxin 16', map_location='19C6-19C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20392334, end_position_on_the_genomic_accession=20393751, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33035, CurrentID=0, Status='live', Symbol='HERC2', Aliases='Dmel_CG11734, AF189221.1, CG11734, Dmel\\CG11734', description='HECT and RLD domain containing protein 2', other_designations='CG11734-PC|HECT and RLD domain containing protein 2 ortholog|HERC2-PC', map_location='19C6-19C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20394395, end_position_on_the_genomic_accession=20412247, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  33036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33036, CurrentID=0, Status='live', Symbol='CG15459', Aliases='Dmel_ Dmel\\CG15459', description='CG15459 gene product from transcript CG15459-RA', other_designations='CG15459-PA|uncharacterized protein', map_location='19C6-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20412341, end_position_on_the_genomic_accession=20413581, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33037, CurrentID=0, Status='live', Symbol='Obp19a', Aliases='Dmel_CG11748, 19a, CG11748, DmelObp19a, Dmel\\CG11748, OBP19a', description='Odorant-binding protein 19a', other_designations='CG11748-PC|Obp19a-PC', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20415744, end_position_on_the_genomic_accession=20416521, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33038, CurrentID=0, Status='live', Symbol='Obp19b', Aliases='Dmel_CG1670, 19b, CG1670, DmelObp19b, Dmel\\CG1670, anon-WO0140519.23, o19b', description='Odorant-binding protein 19b', other_designations='CG1670-PA|Obp19b-PA', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20416733, end_position_on_the_genomic_accession=20417812, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33039, CurrentID=0, Status='live', Symbol='Obp19c', Aliases='Dmel_CG15457, 19c, CG15457, DmelObp19c, Dmel\\CG15457', description='Odorant-binding protein 19c', other_designations='CG15457-PA|CG15457-PC|Obp19c-PA|Obp19c-PC|odorant binding protein 19c', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20417921, end_position_on_the_genomic_accession=20418964, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33040, CurrentID=0, Status='live', Symbol='Obp19d', Aliases='Dmel_CG1668, 19d, BcDNA:HL07789, BcDNA:RH68082, CG1668, DmelObp19d, Dmel\\CG1668, PB-PRP2, PBP2, PBPRP-2, PBPRP2, Pbprp2, pbprp-2, pbprp2', description='Odorant-binding protein 19d', other_designations='CG1668-PA|CG1668-PB|Obp19d-PA|Obp19d-PB|Pheromone-binding protein-related protein 2', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20419613, end_position_on_the_genomic_accession=20420979, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33043, CurrentID=0, Status='live', Symbol='cactin', Aliases='Dmel_CG1676, CG1676, Cactin, Dmel\\CG1676', description='CG1676 gene product from transcript CG1676-RB', other_designations='CG1676-PB|cactin-PB', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20424069, end_position_on_the_genomic_accession=20426402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33044, CurrentID=0, Status='live', Symbol='CG11710', Aliases='Dmel_ Dmel\\CG11710', description='CG11710 gene product from transcript CG11710-RC', other_designations='CG11710-PA|CG11710-PB|CG11710-PC|uncharacterized protein', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20427411, end_position_on_the_genomic_accession=20429673, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33045, CurrentID=0, Status='live', Symbol='mal', Aliases='Dmel_CG1692, 5, CG1692, Dmel\\CG1692, SRF, bz, ma-l', description='maroon-like', other_designations='CG1692-PA|CG1692-PB|bronzy|mal-PA|mal-PB|maroonlike|section 5', map_location='1-64.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20429675, end_position_on_the_genomic_accession=20433210, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33046, CurrentID=0, Status='live', Symbol='CG1695', Aliases='Dmel_ Dmel\\CG1695', description='CG1695 gene product from transcript CG1695-RC', other_designations='CG1695-PC', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20433950, end_position_on_the_genomic_accession=20445599, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33047, CurrentID=0, Status='live', Symbol='GstT3', Aliases='Dmel_CG1702, CG1702, Dmel\\CG1702, GSTT3, cg1702, gstt3', description='Glutathione S transferase T3', other_designations='CG1702-PA|CG1702-PB|CG1702-PC|GstT3-PA|GstT3-PB|GstT3-PC|glutathione S transferase T3', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20467455, end_position_on_the_genomic_accession=20469796, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33048, CurrentID=0, Status='live', Symbol='RhoGAP19D', Aliases='Dmel_CG1412, 1412, CG1412, Dmel\\CG1412', description='Rho GTPase activating protein at 19D', other_designations='CG1412-PA|CG1412-PB|CG1412-PC|CG1412-PD|CG1412-PE|RhoGAP19D-PA|RhoGAP19D-PB|RhoGAP19D-PC|RhoGAP19D-PD|RhoGAP19D-PE', map_location='19D1-19D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20469955, end_position_on_the_genomic_accession=20510694, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  33049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33049, CurrentID=0, Status='live', Symbol='CG1812', Aliases='Dmel_ Dmel\\CG1812', description='CG1812 gene product from transcript CG1812-RA', other_designations='CG1812-PA', map_location='19D2-19D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20511544, end_position_on_the_genomic_accession=20515828, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33051, CurrentID=0, Status='live', Symbol='RunxB', Aliases='Dmel_CG42267, CG15454, CG15455, CG42267, Dmel\\CG42267, Dmel_CG15454, Dmel_CG15455', description='CG42267 gene product from transcript CG42267-RC', other_designations='CG42267-PC|RunxB-PC', map_location='19D3-19D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20538497, end_position_on_the_genomic_accession=20567097, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33056, CurrentID=0, Status='live', Symbol='Cyp6v1', Aliases='Dmel_CG1829, 6v1, CG1829, Dmel\\CG1829', description='CG1829 gene product from transcript CG1829-RB', other_designations='CG1829-PA|CG1829-PB|CG1829-PC|Cyp6v1-PA|Cyp6v1-PB|Cyp6v1-PC|cytochrome P450', map_location='19E1-19E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20657122, end_position_on_the_genomic_accession=20661425, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33057, CurrentID=0, Status='live', Symbol='CG1835', Aliases='Dmel_ BcDNA:GH25431, Dmel\\CG1835, anon-WO0140519.219', description='CG1835 gene product from transcript CG1835-RC', other_designations='CG1835-PA|CG1835-PC', map_location='19E1-19E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20661917, end_position_on_the_genomic_accession=20669669, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33058, CurrentID=0, Status='live', Symbol='hydra', Aliases='Dmel_CG1338, CG1338, Dmel\\CG1338', description='CG1338 gene product from transcript CG1338-RA', other_designations='CG1338-PA|CG1338-PB|CG1338-PD|CG1338-PF|CG1338-PG|CG1338-PH|CG1338-PI|CG1338-PJ|hydra-PA|hydra-PB|hydra-PD|hydra-PF|hydra-PG|hydra-PH|hydra-PI|hydra-PJ', map_location='19E1-19E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20670832, end_position_on_the_genomic_accession=20677063, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  33059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33059, CurrentID=0, Status='live', Symbol='run', Aliases='Dmel_CG1849, 1, 2, AA33, CG1849, Dmel\\CG1849, LB5, P235, RUN, Rnt, Run, Runt, l(1)19Ea, l(1)AA33, l(1)B2/13.1, l(1)LB9, lLB5, leg', description='runt', other_designations='CG1849-PA|CG1849-PB|CG1849-PC|legless|run-PA|run-PB|run-PC|section 1|section 2', map_location='1-65 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20690670, end_position_on_the_genomic_accession=20697321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33060, CurrentID=0, Status='live', Symbol='CG1324', Aliases='Dmel_ Dmel\\CG1324, anon-WO0140519.192', description='CG1324 gene product from transcript CG1324-RC', other_designations='CG1324-PA|CG1324-PC', map_location='19E2-19E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20734455, end_position_on_the_genomic_accession=20735960, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33061, CurrentID=0, Status='live', Symbol='CG15452', Aliases='Dmel_ Dmel\\CG15452', description='CG15452 gene product from transcript CG15452-RA', other_designations='CG15452-PA|uncharacterized protein', map_location='19E2-19E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20744965, end_position_on_the_genomic_accession=20745965, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33062, CurrentID=0, Status='live', Symbol='shakB', Aliases='Dmel_CG34358, 2, 3, 6, CG12678, CG1321, CG15451, CG32508, CG34358, Dmel\\CG34358, Dmel_CG12678, Dmel_CG32508, E81, FBpp0111481, Pas, R-9-29, R9-29, Shak B, Shak-B, ShakB, W3, inx8, l(1)19Eb, l(1)LB21, l(1)R-10-14, l(1)R-10-3, l(1)R-10-7, l(1)R-9-29, l(1)TH73, l(1)W3, l(1)W3b, lincRNA.1031, nj-156, nj156, pas, pass, shB, shak-B, shkB', description='shaking B', other_designations='CG34358-PA|CG34358-PC|CG34358-PD|CG34358-PE|CG34358-PF|CG34358-PG|CG34358-PH|CG34358-PI|Shaking-B|non-jumper-156|passover|passover-vital protein|section 2|section 3|section 6|shakB-PA|shakB-PC|shakB-PD|shakB-PE|shakB-PF|shakB-PG|shakB-PH|shakB-PI', map_location='19E3-19E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20761071, end_position_on_the_genomic_accession=20927050, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  33064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33064, CurrentID=0, Status='live', Symbol='CG15450', Aliases='Dmel_ Dmel\\CG15450', description='CG15450 gene product from transcript CG15450-RA', other_designations='CG15450-PA|uncharacterized protein', map_location='19E3-19E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20822146, end_position_on_the_genomic_accession=20825641, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33066, CurrentID=0, Status='live', Symbol='CG1314', Aliases='Dmel_ Dmel\\CG1314, anon-WO0140519.251', description='CG1314 gene product from transcript CG1314-RA', other_designations='CG1314-PA', map_location='19E4-19E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20867989, end_position_on_the_genomic_accession=20869397, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33068, CurrentID=0, Status='live', Symbol='I-3', Aliases='Dmel_CG42707, CG11233, CG42707, Dmel\\CG42707, Dmel_CG11233', description='Inhibitor-3', other_designations='CG42707-PB|I-3-PB', map_location='19E4-19E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20881814, end_position_on_the_genomic_accession=20882626, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33071, CurrentID=0, Status='live', Symbol='CG11227', Aliases='Dmel_ Dmel\\CG11227', description='CG11227 gene product from transcript CG11227-RC', other_designations='CG11227-PA|CG11227-PB|CG11227-PC|uncharacterized protein', map_location='19E5-19E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20930972, end_position_on_the_genomic_accession=20933559, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33072, CurrentID=0, Status='live', Symbol='Npc1b', Aliases='Dmel_CG12092, 151, 4, 7, AF2/19, CG12092, Dmel\\CG12092, Dmel_CG11233, NPC1b, Q256, R-9-28, R9-28, dncp1b, l(1)19Ec, l(1)2-3, l(1)R-9-28, l151, l152, lN5, lt2-14a, npc1b', description='Niemann-Pick type C-1b', other_designations='CG12092-PA|NPC1b|Niemann-Pick|Npc1b-PA|lethal (1) 19Ec|section 4|section 7', map_location='19E5-19E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20934869, end_position_on_the_genomic_accession=20939503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33073, CurrentID=0, Status='live', Symbol='Ser6', Aliases='Dmel_CG2071, CG2071, Dmel\\CG2071, SP47', description='Serine protease 6', other_designations='CG2071-PA|Ser6-PA|serine protease 6', map_location='19E5-19E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20962605, end_position_on_the_genomic_accession=20963505, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33074, CurrentID=0, Status='live', Symbol='CG1304', Aliases='Dmel_ Dmel\\CG1304, SP46', description='CG1304 gene product from transcript CG1304-RA', other_designations='CG1304-PA', map_location='19E5-19E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20964527, end_position_on_the_genomic_accession=20965419, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33075, CurrentID=0, Status='live', Symbol='r-cup', Aliases='Dmel_CG10998, CG10998, Dmel\\CG10998', description='ryder cup', other_designations='CG10998-PA|r-cup-PA|ryder-cup', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21023390, end_position_on_the_genomic_accession=21025888, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33076, CurrentID=0, Status='live', Symbol='CG1532', Aliases='Dmel_ Dmel\\CG1532', description='CG1532 gene product from transcript CG1532-RB', other_designations='CG1532-PA|CG1532-PB', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21026125, end_position_on_the_genomic_accession=21028361, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33077, CurrentID=0, Status='live', Symbol='CG1529', Aliases='Dmel_ Dmel\\CG1529', description='CG1529 gene product from transcript CG1529-RD', other_designations='CG1529-PA|CG1529-PC|CG1529-PD', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21028652, end_position_on_the_genomic_accession=21031950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33078, CurrentID=0, Status='live', Symbol='Ntf-2', Aliases='Dmel_CG1740, BcDNA:GM08921, CG1740, DNTF-2, Dmel\\CG1740, Dntf-2, Dntf2, NTF-2, Ntf2, dmNTF2, l(1)G0086, l(1)G0125, l(1)G0337, l(1)G0428, ntf-2, ntf2', description='Nuclear transport factor-2', other_designations='CG1740-PA|CG1740-PB|CG1740-PC|CG1740-PE|Ntf-2-PA|Ntf-2-PB|Ntf-2-PC|Ntf-2-PE|nuclear transport factor-2', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21035469, end_position_on_the_genomic_accession=21041467, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33079, CurrentID=0, Status='live', Symbol='CG15449', Aliases='Dmel_ Dmel\\CG15449', description='CG15449 gene product from transcript CG15449-RA', other_designations='CG15449-PA|uncharacterized protein', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21039514, end_position_on_the_genomic_accession=21040605, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33081, CurrentID=0, Status='live', Symbol='Cbs', Aliases='Dmel_CG1753, CBS, CG1753, DmCBS, Dmel\\CG1753, dCBS', description='Cystathionine beta-synthase', other_designations='CG1753-PA|CG1753-PB|CG1753-PC|Cbs-PA|Cbs-PB|Cbs-PC|cystathionine &bgr-synthase|cystathionine &grb;-synthase|cystathionine beta-synthase', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21045138, end_position_on_the_genomic_accession=21049096, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33082, CurrentID=0, Status='live', Symbol='CG1518', Aliases='Dmel_ Dmel\\CG1518, STT', description='CG1518 gene product from transcript CG1518-RC', other_designations='CG1518-PA|CG1518-PC', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21049885, end_position_on_the_genomic_accession=21053937, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33083, CurrentID=0, Status='live', Symbol='bves', Aliases='Dmel_CG32513, BcDNA:SD21670, CG15448, CG1754, CG32513, Dmel\\CG32513', description='CG32513 gene product from transcript CG32513-RA', other_designations='CG32513-PA|CG32513-PB|CG32513-PC|CG32513-PD|bves-PA|bves-PB|bves-PC|bves-PD', map_location='19E7-19E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21053963, end_position_on_the_genomic_accession=21082391, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33085, CurrentID=0, Status='live', Symbol='CG32512', Aliases='Dmel_ CG15447, CG1739, Dmel\\CG32512', description='CG32512 gene product from transcript CG32512-RD', other_designations='CG32512-PA|CG32512-PB|CG32512-PC|CG32512-PD|uncharacterized protein', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21062277, end_position_on_the_genomic_accession=21067540, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33088, CurrentID=0, Status='live', Symbol='CG15446', Aliases='Dmel_ Dmel\\CG15446', description='CG15446 gene product from transcript CG15446-RD', other_designations='CG15446-PB|CG15446-PC|CG15446-PD|uncharacterized protein', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21095680, end_position_on_the_genomic_accession=21102206, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33090, CurrentID=0, Status='live', Symbol='CG1503', Aliases='Dmel_ Dmel\\CG1503', description='CG1503 gene product from transcript CG1503-RB', other_designations='CG1503-PA|CG1503-PB', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21104695, end_position_on_the_genomic_accession=21106369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33093, CurrentID=0, Status='live', Symbol='unc', Aliases='Dmel_CG1501, 3, 4, 5, CG1501, Dmel\\CG1501, UNC, Unc, W5, l(1)19Ef, l(1)6-3-212, lLV7', description='uncoordinated', other_designations='CG1501-PA|CG1501-PB|section 3|section 4|section 5|unc-PA|unc-PB', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21111414, end_position_on_the_genomic_accession=21116914, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33094, CurrentID=0, Status='live', Symbol='CG15445', Aliases='Dmel_ Dmel\\CG15445, NUB1', description='CG15445 gene product from transcript CG15445-RA', other_designations='CG15445-PA|CG15445-PE|CG15445-PF|CG15445-PG|CG15445-PH|uncharacterized protein', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21117159, end_position_on_the_genomic_accession=21120444, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33096, CurrentID=0, Status='live', Symbol='CG14579', Aliases='Dmel_ Dmel\\CG14579', description='CG14579 gene product from transcript CG14579-RA', other_designations='CG14579-PA|uncharacterized protein', map_location='19F2-19F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21193253, end_position_on_the_genomic_accession=21194521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33097, CurrentID=0, Status='live', Symbol='CG1724', Aliases='Dmel_ BcDNA:AT13347, Dmel\\CG1724', description='CG1724 gene product from transcript CG1724-RA', other_designations='CG1724-PA', map_location='19F2-19F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21202499, end_position_on_the_genomic_accession=21203341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33098, CurrentID=0, Status='live', Symbol='CG32521', Aliases='Dmel_ CG1726, CG1735, Dmel\\CG32521', description='CG32521 gene product from transcript CG32521-RA', other_designations='CG32521-PA|CG32521-PB|CG32521-PC|CG32521-PD|CG32521-PE|CG32521-PF|CG32521-PG|CG32521-PH|CG32521-PI|uncharacterized protein', map_location='19F2-19F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21205310, end_position_on_the_genomic_accession=21281171, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  5740400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740400, CurrentID=0, Status='live', Symbol='CG34169', Aliases='Dmel_ Dmel\\CG34169', description='CG34169 gene product from transcript CG34169-RA', other_designations='CG34169-PA|uncharacterized protein', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17173481, end_position_on_the_genomic_accession=17174105, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33101, CurrentID=0, Status='live', Symbol='CG1722', Aliases='Dmel_ Dmel\\CG1722', description='CG1722 gene product from transcript CG1722-RA', other_designations='CG1722-PA', map_location='19F2-19F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21249976, end_position_on_the_genomic_accession=21251148, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33102, CurrentID=0, Status='live', Symbol='CG1494', Aliases='Dmel_ Dmel\\CG1494', description='CG1494 gene product from transcript CG1494-RC', other_designations='CG1494-PB|CG1494-PC', map_location='19F3-19F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21285301, end_position_on_the_genomic_accession=21292458, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33103, CurrentID=0, Status='live', Symbol='CG1718', Aliases='Dmel_ Dmel\\CG1718', description='CG1718 gene product from transcript CG1718-RB', other_designations='CG1718-PB|CG1718-PC|CG1718-PD', map_location='19F3-19F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21292308, end_position_on_the_genomic_accession=21306665, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33104, CurrentID=0, Status='live', Symbol='CG1801', Aliases='Dmel_ Dmel\\CG1801', description='CG1801 gene product from transcript CG1801-RC', other_designations='CG1801-PC', map_location='19F3-19F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21307586, end_position_on_the_genomic_accession=21317522, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33105, CurrentID=0, Status='live', Symbol='Rpt6', Aliases='Dmel_CG1489, 1489, CG1489, DUG, Dm_Rpt6a, Dmel\\CG1489, Dug, PRS8_DROME, Pros45, S8, Ug, dRpt6, p42C, pros45, sug-1, sug1', description='Regulatory particle triple-A ATPase 6', other_designations='26S proteasome regulatory subunit 8|CG1489-PA|Pros45|Rpt6-PA|p42C|proteasomal subunit 8|regulatory particle triple-A ATPase 6', map_location='19F4-19F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21316052, end_position_on_the_genomic_accession=21317663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33107, CurrentID=0, Status='live', Symbol='Dd', Aliases='Dmel_CG1696, 20151539, BcDNA:LD08201, CG1696, Dmel\\CG1696, anon-WO03040301.167, ddd, l(1)G0269', description='Dullard', other_designations='CG1696-PA|CG1696-PB|CG1696-PC|Dd-PA|Dd-PB|Dd-PC|d-dullard|dullard|lethal (1) G0269', map_location='19F4-19F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21319791, end_position_on_the_genomic_accession=21324009, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33108, CurrentID=0, Status='live', Symbol='CG1486', Aliases='Dmel_ Dmel\\CG1486', description='CG1486 gene product from transcript CG1486-RA', other_designations='CG1486-PA|CG1486-PB', map_location='19F4-19F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21324046, end_position_on_the_genomic_accession=21328150, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33109, CurrentID=0, Status='live', Symbol='tty', Aliases='Dmel_CG1693, CG1693, Dmel\\CG1693, TTYH1, twe', description='tweety', other_designations='CG1693-PA|CG1693-PC|CG1693-PD|CG1693-PE|CG1693-PF|CG1693-PG|tty-PA|tty-PC|tty-PD|tty-PE|tty-PF|tty-PG', map_location='19F4-19F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21328613, end_position_on_the_genomic_accession=21339676, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33110, CurrentID=0, Status='live', Symbol='fliI', Aliases='Dmel_CG1484, 15, BcDNA:LD21753, CG1484, CT3691, DCA3-19, Dmel\\CG1484, EN3, FLI-I, Fli-I, W2, fli, fli I, fli-I, fltI, fltO, l(1)18-80, l(1)19Fd, l(1)ESHS53, l(1)W-2, lDCB2-19, sdby, w-2', description='flightless I', other_designations='CG1484-PA|Flightless-I|fliI-PA|flightless|gluful-2|lgluful-2|section 15|standby', map_location='1-65.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21339741, end_position_on_the_genomic_accession=21345088, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33111, CurrentID=0, Status='live', Symbol='dod', Aliases='Dmel_CG17051, CG17051, Dmel\\CG17051, Pin1', description='dodo', other_designations='CG17051-PA|dod-PA', map_location='19F5-19F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21345358, end_position_on_the_genomic_accession=21346858, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33112, CurrentID=0, Status='live', Symbol='peng', Aliases='Dmel_CG1685, CG1685, Dmel\\CG1685, pen', description='penguin', other_designations='CG1685-PA|peng-PA', map_location='19F5-19F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21346503, end_position_on_the_genomic_accession=21349843, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33114, CurrentID=0, Status='live', Symbol='Sep1', Aliases='Dmel_CG1403, 38960, CG1403, Diff6, Dmel\\CG1403, anon-19Fa, iby, sep1, tu1', description='Septin 1', other_designations='CG1403-PA|Sep1-PA|Septin-1|innocent bystander|septin|septin 1|septin1', map_location='19F5-19F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21358286, end_position_on_the_genomic_accession=21361688, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33117, CurrentID=0, Status='live', Symbol='slgA', Aliases='Dmel_CG1417, CG1417, Dmel\\CG1417, EE85, PRODH, sl, slg, tu3', description='sluggish A', other_designations='CG1417-PA|CG1417-PB|CG1417-PC|CG1417-PD|CG1417-PE|CG1417-PF|CG1417-PG|CG1417-PH|CG1417-PI|CG1417-PJ|CG1417-PK|CG1417-PL|CG1417-PM|CG1417-PN|proline dehydrogenase|slgA-PA|slgA-PB|slgA-PC|slgA-PD|slgA-PE|slgA-PF|slgA-PG|slgA-PH|slgA-PI|slgA-PJ|slgA-PK|slgA-PL|slgA-PM|slgA-PN|sluggish|sluggish-A|sluggishA', map_location='19F6-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21374002, end_position_on_the_genomic_accession=21385159, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  33118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33118, CurrentID=0, Status='live', Symbol='Hlc', Aliases='Dmel_CG1666, 11P1, 14, 7, A112, CG1666, DmRH10, Dmel\\CG1666, HLC, HM-44, hlc, l(1)11P1, l(1)19Ff, l(1)8-1, l(1)A112, l(1)S-19F6, l11P1, lA112', description='Helicase', other_designations='CG1666-PA|CG1666-PB|Hlc-PA|Hlc-PB|RNA helicase gene|helicase|lethal(1)19Ff|section 14|section 7', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21385518, end_position_on_the_genomic_accession=21387946, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33119, CurrentID=0, Status='live', Symbol='mst', Aliases='Dmel_CG1424, CG1424, Dmel\\CG1424, LB-20, LB20, l(1)19Fg, l(1)LB14, l(1)LB20, lLB14', description='misato', other_designations='CG1424-PA|CG1424-PB|lethal(1)19Fg|mst-PA|mst-PB', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21388977, end_position_on_the_genomic_accession=21391533, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33120, CurrentID=0, Status='live', Symbol='lcs', Aliases='Dmel_CG12794, CG12794, Dmel\\CG12794', description='la costa', other_designations='CG12794-PA|lcs-PA|leader sequence, collagen-related repetitive motif and a serine/threonine-rich motif', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21396727, end_position_on_the_genomic_accession=21397316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33121, CurrentID=0, Status='live', Symbol='CG1678', Aliases='Dmel_ BcDNA:GH04760, Dmel\\CG1678', description='CG1678 gene product from transcript CG1678-RB', other_designations='CG1678-PA|CG1678-PB', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21402339, end_position_on_the_genomic_accession=21402772, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33123, CurrentID=0, Status='live', Symbol='CG10918', Aliases='Dmel_ Dmel\\CG10918', description='CG10918 gene product from transcript CG10918-RA', other_designations='CG10918-PA|uncharacterized protein', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21428614, end_position_on_the_genomic_accession=21429256, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33125, CurrentID=0, Status='live', Symbol='DIP-beta', Aliases='Dmel_CG42343, 42343, BP1062, BcDNA:RE23430, CG12564, CG12565, CG14583, CG34019, CG42343, CT34319, Dmel\\CG42343, Dmel_CG12564, Dmel_CG12565, Dmel_CG14583, Dmel_CG34019, pp-CT34319, pp-CT34320, pp-CT34321', description='Dpr-interacting protein beta', other_designations='CG42343-PC|CG42343-PD|CG42343-PE|CG42343-PF|CG42343-PG|DIP-beta-PC|DIP-beta-PD|DIP-beta-PE|DIP-beta-PF|DIP-beta-PG', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21422115, end_position_on_the_genomic_accession=21457738, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33126, CurrentID=0, Status='live', Symbol='Ir20a', Aliases='Dmel_CG14584, CG14584, DmelIR20a, Dmel\\CG14584, IR20a', description='Ionotropic receptor 20a', other_designations='CG14584-PA|Ir20a-PA|ionotropic receptor 20a', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21477254, end_position_on_the_genomic_accession=21478945, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33127, CurrentID=0, Status='live', Symbol='Cyp6t1', Aliases='Dmel_CG1644, 6t1, CG1644, Dmel\\CG1644', description='CG1644 gene product from transcript CG1644-RA', other_designations='CG1644-PA|Cyp6t1-PA|cytochrome P450', map_location='20A1-20A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21517434, end_position_on_the_genomic_accession=21519218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33128, CurrentID=0, Status='live', Symbol='CG14621', Aliases='Dmel_ Dmel\\CG14621, Slc35e1, lincRNA.1043', description='CG14621 gene product from transcript CG14621-RD', other_designations='CG14621-PA|CG14621-PB|CG14621-PC|CG14621-PD|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22444227, end_position_on_the_genomic_accession=22447808, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33129, CurrentID=0, Status='live', Symbol='CG14615', Aliases='Dmel_ Dmel\\CG14615', description='CG14615 gene product from transcript CG14615-RB', other_designations='CG14615-PA|CG14615-PB|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22447183, end_position_on_the_genomic_accession=22449491, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33130, CurrentID=0, Status='live', Symbol='tilB', Aliases='Dmel_CG14620, CG14620, Dmel\\CG14620', description='touch insensitive larva B', other_designations='CG14620-PA|TilB|tilB-PA|touch insensitive larvaeB|touch-insensitive-larva-B', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22449737, end_position_on_the_genomic_accession=22451463, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33131, CurrentID=0, Status='live', Symbol='wap', Aliases='Dmel_CG14614, 15, 16, 19, 20, 21, 22, BcDNA:GM08142, CG14614, Dmel\\CG14614, Q217, Q464, Riq, l(1)20Ab, l(1)A-200, l(1)HM33, l(1)Q217, l(1)Q464, riq', description='wings apart', other_designations='CG14614-PA|CG14614-PB|CG14614-PC|CG14614-PD|riquiqui|section 15|section 16|section 19|section 20|section 21|section 22|wap-PA|wap-PB|wap-PC|wap-PD|wings-apart', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22451445, end_position_on_the_genomic_accession=22455422, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33132, CurrentID=0, Status='live', Symbol='Usp2', Aliases='Dmel_CG14619, CG14619, Dmel\\CG14619, Q9VR54', description='Ubiquitin specific protease 2', other_designations='CG14619-PA|CG14619-PC|CG14619-PD|CG14619-PE|CG14619-PF|CG14619-PG|CG14619-PH|CG14619-PI|CG14619-PK|Usp2-PA|Usp2-PC|Usp2-PD|Usp2-PE|Usp2-PF|Usp2-PG|Usp2-PH|Usp2-PI|Usp2-PK|ubiquitin specific protease 2', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22455803, end_position_on_the_genomic_accession=22470146, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  33133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33133, CurrentID=0, Status='live', Symbol='CG14613', Aliases='Dmel_ ADP/ATP translocase, CT34370, Dmel\\CG14613', description='CG14613 gene product from transcript CG14613-RB', other_designations='CG14613-PA|CG14613-PB|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22469981, end_position_on_the_genomic_accession=22472021, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33134, CurrentID=0, Status='live', Symbol='CG14618', Aliases='Dmel_ Dmel\\CG14618', description='CG14618 gene product from transcript CG14618-RA', other_designations='CG14618-PA|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22473698, end_position_on_the_genomic_accession=22474998, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33135, CurrentID=0, Status='live', Symbol='CG12576', Aliases='Dmel_ Dmel\\CG12576, anon-WO0140519.40', description='CG12576 gene product from transcript CG12576-RF', other_designations='CG12576-PA|CG12576-PB|CG12576-PE|CG12576-PF|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22474851, end_position_on_the_genomic_accession=22482357, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33136, CurrentID=0, Status='live', Symbol='Cp110', Aliases='Dmel_CG14617, CG14617, CG32501, CP110, DCP110, Dmel\\CG14617', description='CG14617 gene product from transcript CG14617-RD', other_designations='CG14617-PA|CG14617-PB|CG14617-PC|CG14617-PD|CG14617-PE|CG14617-PF|CG14617-PG|CG14617-PH|Cp110-PA|Cp110-PB|Cp110-PC|Cp110-PD|Cp110-PE|Cp110-PF|Cp110-PG|Cp110-PH', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22482082, end_position_on_the_genomic_accession=22487458, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33137, CurrentID=0, Status='live', Symbol='l(1)G0196', Aliases='Dmel_CG14616, CG14616, CG32502, Dmel\\CG14616', description='lethal (1) G0196', other_designations='CG14616-PG|CG14616-PH|CG14616-PI|CG14616-PJ|CG14616-PK|CG14616-PL|CG14616-PM|CG14616-PN|CG14616-PO|CG14616-PP|l(1)G0196-PG|l(1)G0196-PH|l(1)G0196-PI|l(1)G0196-PJ|l(1)G0196-PK|l(1)G0196-PL|l(1)G0196-PM|l(1)G0196-PN|l(1)G0196-PO|l(1)G0196-PP', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22487180, end_position_on_the_genomic_accession=22508129, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  33139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33139, CurrentID=0, Status='live', Symbol='S6kII', Aliases='Dmel_CG17596, CG17596, Dmel\\CG17596, RSK, Rsk, S6KII, dRSK, ign, rsk, rsk2', description='Ribosomal protein S6 kinase II', other_designations='CG17596-PA|CG17596-PB|RPS6-protein kinase-II|Ribosomal-S6-kinase|S6 kinase II|S6kII-PA|S6kII-PB|ignorant|p90 ribosomal S6 kinase|ribosomal protein S6 kinase II', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22529964, end_position_on_the_genomic_accession=22536371, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33140, CurrentID=0, Status='live', Symbol='CG17598', Aliases='Dmel_ Dmel\\CG17598, PP2C', description='CG17598 gene product from transcript CG17598-RB', other_designations='CG17598-PA|CG17598-PB|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22534831, end_position_on_the_genomic_accession=22540383, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33141, CurrentID=0, Status='live', Symbol='CG17600', Aliases='Dmel_ CG17602, Dmel\\CG17600', description='CG17600 gene product from transcript CG17600-RA', other_designations='CG17600-PA|CG17600-PB|CG17600-PC|CG17600-PD|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22515197, end_position_on_the_genomic_accession=22542024, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33142, CurrentID=0, Status='live', Symbol='CG17599', Aliases='Dmel_ CLUAP1, Dmel\\CG17599', description='CG17599 gene product from transcript CG17599-RA', other_designations='CG17599-PA|uncharacterized protein', map_location='20C1-20C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22542386, end_position_on_the_genomic_accession=22544322, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33143, CurrentID=0, Status='live', Symbol='CG17601', Aliases='Dmel_ Dmel\\CG17601', description='CG17601 gene product from transcript CG17601-RA', other_designations='CG17601-PA|uncharacterized protein', map_location='20C2-20C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22547097, end_position_on_the_genomic_accession=22548056, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33144, CurrentID=0, Status='live', Symbol='Cda4', Aliases='Dmel_CG32499, BcDNA:RE51076, CG12447, CG13238, CG32499, CG9557, CG9558, DmCDA4, Dmel\\CG32499', description='Chitin deacetylase-like 4', other_designations='CG32499-PA|Cda4-PA|chitin deacetylase-like 4', map_location='20C3-20D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22583634, end_position_on_the_genomic_accession=22632710, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33147, CurrentID=0, Status='live', Symbol='CG12446', Aliases='Dmel_ Dmel\\CG12446', description='CG12446 gene product from transcript CG12446-RA', other_designations='CG12446-PA|uncharacterized protein', map_location='20E2-20E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22809292, end_position_on_the_genomic_accession=22811935, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33148, CurrentID=0, Status='live', Symbol='fog', Aliases='Dmel_CG9559, 114, CG9559, CR41108, CR41109, CR41110, Dmel\\CG9559, Fog, M67, l(1)114, l(1)16-3-190, l(1)20Ba, l114', description='folded gastrulation', other_designations='CG9559-PA|CG9559-PB|CG9559-PC|CG9559-PD|CG9559-PE|CG9559-PF|fog-PA|fog-PB|fog-PC|fog-PD|fog-PE|fog-PF', map_location='1-65 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22824317, end_position_on_the_genomic_accession=22855194, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33155, CurrentID=0, Status='live', Symbol='CG11023', Aliases='Dmel_ Dmel\\CG11023', description='CG11023 gene product from transcript CG11023-RB', other_designations='CG11023-PB|CG11023-PC|CG11023-PD|uncharacterized protein', map_location='21A5-21A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7529, end_position_on_the_genomic_accession=9484, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33156, CurrentID=0, Status='live', Symbol='l(2)gl', Aliases='Dmel_CG2671, CG2671, D-LGL, Dmel\\CG2671, L(2)GL, L(2)gl, LGL, Lgl, MENE (2L)-B, dlgl, gl, l(2), l(2) giant larva, l(2)giant larvae, l-gl, l[[2]]gl, lgl, p127, p127[l(2)gl], p127l(2)gl', description='lethal (2) giant larvae', other_designations='CG2671-PA|CG2671-PB|CG2671-PC|CG2671-PD|CG2671-PE|CG2671-PF|CG2671-PG|CG2671-PH|CG2671-PI|CG2671-PJ|CG2671-PK|complementation group 2.1|l(2) giant larvae|l(2)gl-PA|l(2)gl-PB|l(2)gl-PC|l(2)gl-PD|l(2)gl-PE|l(2)gl-PF|l(2)gl-PG|l(2)gl-PH|l(2)gl-PI|l(2)gl-PJ|l(2)gl-PK|lethal (2) giant larva|lethal giant larva|lethal giant larvae|lethal giant larve|lethal(2) giant larvae|lethal(2)-giant larvae|lethal(2)giant larvae|lethal- giant-larvae|lethal-2-giant larvae|lethal-giant-larvae', map_location='2-0.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9839, end_position_on_the_genomic_accession=21376, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  33157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33157, CurrentID=0, Status='live', Symbol='Ir21a', Aliases='Dmel_CG2657, CG2657, CT8983, DmelIR21a, Dmel\\CG2657, IR21a, ir21a', description='Ionotropic receptor 21a', other_designations='CG2657-PB|Ir21a-PB|ionotropic receptor 21a', map_location='21A5-21B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21823, end_position_on_the_genomic_accession=25155, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33158, CurrentID=0, Status='live', Symbol='Cda5', Aliases='Dmel_CG31973, BcDNA:RH43162, CG2761, CG2776, CG31973, DmCDA5, Dmel\\CG31973', description='Chitin deacetylase-like 5', other_designations='CG31973-PA|CG31973-PB|CG31973-PE|CG31973-PF|CG31973-PG|CG31973-PH|CG31973-PI|CG31973-PJ|Cda5-PA|Cda5-PB|Cda5-PE|Cda5-PF|Cda5-PG|Cda5-PH|Cda5-PI|Cda5-PJ|chitin deacetylase-like 5', map_location='21B1-21B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=25402, end_position_on_the_genomic_accession=65404, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  33161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33161, CurrentID=0, Status='live', Symbol='dbr', Aliases='Dmel_CG11371, BcDNA:LD26519, CG11371, Dbr, Debra, Dmel\\CG11371, EP0456, EP456', description='debra', other_designations='CG11371-PB|CG11371-PC|CG11371-PD|CG11371-PE|CG11371-PF|dbr-PB|dbr-PC|dbr-PD|dbr-PE|dbr-PF', map_location='21B1-21B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=66482, end_position_on_the_genomic_accession=71390, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33162, CurrentID=0, Status='live', Symbol='galectin', Aliases='Dmel_CG11372, CG11372, Dmel\\CG11372, Dmgal', description='CG11372 gene product from transcript CG11372-RD', other_designations='CG11372-PA|CG11372-PB|CG11372-PD|CG11372-PE|CG11372-PF|galectin-PA|galectin-PB|galectin-PD|galectin-PE|galectin-PF', map_location='21B1-21B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=71757, end_position_on_the_genomic_accession=76211, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33163, CurrentID=0, Status='live', Symbol='CG11374', Aliases='Dmel_ Dmel\\CG11374', description='CG11374 gene product from transcript CG11374-RC', other_designations='CG11374-PC|uncharacterized protein', map_location='21B1-21B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=76348, end_position_on_the_genomic_accession=77783, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33165, CurrentID=0, Status='live', Symbol='Zir', Aliases='Dmel_CG11376, CG11376, Dm zir, Dmel\\CG11376, FBgn 31216, anon-EST:Posey54', description='Zizimin-related', other_designations='CG11376-PA|Zir-PA', map_location='21B1-21B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=94739, end_position_on_the_genomic_accession=102086, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33166, CurrentID=0, Status='live', Symbol='CG11377', Aliases='Dmel_ CT31762, Dmel\\CG11377', description='CG11377 gene product from transcript CG11377-RA', other_designations='CG11377-PA|CG11377-PB|uncharacterized protein', map_location='21B2-21B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=102380, end_position_on_the_genomic_accession=104142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33167, CurrentID=0, Status='live', Symbol='Nhe1', Aliases='Dmel_CG12178, CG12178, DmNHE1, Dmel\\CG12178, NHE1, nhe1', description='Na[+]/H[+] hydrogen exchanger 1', other_designations='CG12178-PB|Nhe1-PB', map_location='21B2-21B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=103962, end_position_on_the_genomic_accession=106732, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33169, CurrentID=0, Status='live', Symbol='CG13694', Aliases='Dmel_ Dmel\\CG13694', description='CG13694 gene product from transcript CG13694-RA', other_designations='CG13694-PA|CG13694-PB|uncharacterized protein', map_location='21B2-21B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=115365, end_position_on_the_genomic_accession=116488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33170, CurrentID=0, Status='live', Symbol='CG4822', Aliases='Dmel_ Dmel\\CG4822', description='CG4822 gene product from transcript CG4822-RE', other_designations='CG4822-PA|CG4822-PB|CG4822-PC|CG4822-PD|CG4822-PE|CG4822-PF|CG4822-PG|CG4822-PH|CG4822-PI', map_location='21B2-21B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=116970, end_position_on_the_genomic_accession=121754, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  33171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33171, CurrentID=0, Status='live', Symbol='CG3164', Aliases='Dmel_ BEST:CK02656, CK02656, Dmel\\CG3164', description='CG3164 gene product from transcript CG3164-RA', other_designations='CG3164-PA|CG3164-PB|CG3164-PC|CG3164-PD', map_location='21B2-21B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=122624, end_position_on_the_genomic_accession=130791, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33172, CurrentID=0, Status='live', Symbol='Gs1', Aliases='Dmel_CG2718, CG2718, Dmel\\CG2718, GS1, GSI, GsI, fs(2)P52, fs(2)throng, gs-m, gsI, l(2)21Bc, mt-gs, thg', description='Glutamine synthetase 1', other_designations='CG2718-PB|CG2718-PC|CG2718-PD|Glutamate-ammonia ligase|Gs1-PB|Gs1-PC|Gs1-PD|glutamine synthase 1|glutamine synthetase|glutamine synthetase 1|glutamine synthetase I|glutamine synthetase-1|mitochondrial glutamine synthetase|throng', map_location='21B2-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=132077, end_position_on_the_genomic_accession=134472, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33174, CurrentID=0, Status='live', Symbol='CG31974', Aliases='Dmel_ CG18501, Dmel\\CG31974', description='CG31974 gene product from transcript CG31974-RB', other_designations='CG31974-PB|CG31974-PC|CG31974-PD|uncharacterized protein', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=141077, end_position_on_the_genomic_accession=143091, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33175, CurrentID=0, Status='live', Symbol='CG11454', Aliases='Dmel_ Dmel\\CG11454, cg11454', description='CG11454 gene product from transcript CG11454-RA', other_designations='CG11454-PA|uncharacterized protein', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=143323, end_position_on_the_genomic_accession=144227, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33176, CurrentID=0, Status='live', Symbol='CG42399', Aliases='Dmel_ CG11620, CG4648, Dmel\\CG42399, Dmel_CG11620, Dmel_CG4648', description='CG42399 gene product from transcript CG42399-RC', other_designations='CG42399-PB|CG42399-PC|uncharacterized protein', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=144231, end_position_on_the_genomic_accession=153300, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33178, CurrentID=0, Status='live', Symbol='CG3709', Aliases='Dmel_ Dmel\\CG3709, anon-WO0257455.7', description='CG3709 gene product from transcript CG3709-RA', other_designations='CG3709-PA', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=153539, end_position_on_the_genomic_accession=155305, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33179, CurrentID=0, Status='live', Symbol='ND-15', Aliases='Dmel_CG11455, CG11455, Dmel\\CG11455, NIPM', description='NADH dehydrogenase (ubiquinone) 15 kDa subunit', other_designations='CG11455-PA|CG11455-PB|CG11455-PC|CG11455-PD|CG11455-PF|CG11455-PG|CG11455-PH|CG11455-PI|CG11455-PJ|CG11455-PK|NADH ubiquinone oxidoreductase 15 kDa|NADH-ubiquinone oxidoreductase 15 kDa subunit|NADH:ubiquinone oxidoreductase 15 kDa subunit|ND-15-PA|ND-15-PB|ND-15-PC|ND-15-PD|ND-15-PF|ND-15-PG|ND-15-PH|ND-15-PI|ND-15-PJ|ND-15-PK', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=114726, end_position_on_the_genomic_accession=156030, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33180, CurrentID=0, Status='live', Symbol='CG3436', Aliases='Dmel_ Dmel\\CG3436', description='CG3436 gene product from transcript CG3436-RC', other_designations='CG3436-PA|CG3436-PC', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=156029, end_position_on_the_genomic_accession=157666, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33182, CurrentID=0, Status='live', Symbol='mRpL10', Aliases='Dmel_CG11488, CG11488, Dmel\\CG11488, L10', description='mitochondrial ribosomal protein L10', other_designations='CG11488-PA|mRpL10-PA', map_location='21B4-21B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=203779, end_position_on_the_genomic_accession=204782, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33183, CurrentID=0, Status='live', Symbol='CG11617', Aliases='Dmel_ Dmel\\CG11617, Mkx', description='CG11617 gene product from transcript CG11617-RA', other_designations='CG11617-PA|CG11617-PB|uncharacterized protein', map_location='21B4-21B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=204743, end_position_on_the_genomic_accession=207297, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33184, CurrentID=0, Status='live', Symbol='Tbc1d15-17', Aliases='Dmel_CG11490, CG11490, Dmel\\CG11490', description='TBC1 domain family, member 15/17 ortholog (H. sapiens)', other_designations='CG11490-PA|CG11490-PB|CG11490-PC|TBC1 domain family, member 15/17 ortholog|Tbc1d15-17-PA|Tbc1d15-17-PB|Tbc1d15-17-PC', map_location='21B4-21B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=207179, end_position_on_the_genomic_accession=210725, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33185, CurrentID=0, Status='live', Symbol='kis', Aliases='Dmel_CG3696, 136/31, 2532, 5841, BEST:GM02209, CG18326, CG3660, CG3696, DmelPex20, Dmel\\CG3696, EC2-7, EK2-4, EP(2)0474, EP474, FBtr0078144, GM02209, KIS, KIS-L, Kis, Pex20, Su(Pc)21AB, anon-WO0172774.164, anon-WO0257455.5s, l(2)07812, l(2)k08827, l(2)k11324, l(2)k13631, l(2)k14112, l(2)k16510, l(2)s3527, l(2)s4771, l(2)s4793, kis', description='kismet', other_designations='CG3696-PA|CG3696-PB|CG3696-PC|CG3696-PD|CG3696-PE|CG3696-PF|EP474|kis-PA|kis-PB|kis-PC|kis-PD|kis-PE|kis-PF|lethal (2) k11324|lethal (2) k16510|peroxin 20', map_location='2-0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=210735, end_position_on_the_genomic_accession=250823, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  33187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33187, CurrentID=0, Status='live', Symbol='CG13693', Aliases='Dmel_ Dmel\\CG13693', description='CG13693 gene product from transcript CG13693-RA', other_designations='CG13693-PA|uncharacterized protein', map_location='21B5-21B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=232487, end_position_on_the_genomic_accession=234208, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33190, CurrentID=0, Status='live', Symbol='CG3645', Aliases='Dmel_ Dmel\\CG3645', description='CG3645 gene product from transcript CG3645-RA', other_designations='CG3645-PA|CG3645-PB|CG3645-PC', map_location='21B5-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=252589, end_position_on_the_genomic_accession=271744, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33191, CurrentID=0, Status='live', Symbol='CG17075', Aliases='Dmel_ Dmel\\CG17075', description='CG17075 gene product from transcript CG17075-RB', other_designations='CG17075-PA|CG17075-PB|uncharacterized protein', map_location='21B6-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=259949, end_position_on_the_genomic_accession=267505, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33192, CurrentID=0, Status='live', Symbol='CG3345', Aliases='Dmel_ Dmel\\CG3345, anon-WO0140519.207', description='CG3345 gene product from transcript CG3345-RA', other_designations='CG3345-PA', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=269088, end_position_on_the_genomic_accession=271626, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33193, CurrentID=0, Status='live', Symbol='CG11555', Aliases='Dmel_ Dmel\\CG11555, anon-21Ca', description='CG11555 gene product from transcript CG11555-RB', other_designations='CG11555-PA|CG11555-PB|anon-21Ca|uncharacterized protein', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=273836, end_position_on_the_genomic_accession=274766, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33194, CurrentID=0, Status='live', Symbol='mbm', Aliases='Dmel_CG11604, CG11604, Dmel\\CG11604, mus', description='mushroom body miniature', other_designations='CG11604-PA|MB miniature|mbm-PA|mushroom bodies miniature|mushroom bodies small|mushroom-body-miniature', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=271912, end_position_on_the_genomic_accession=273715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33195, CurrentID=0, Status='live', Symbol='CG17078', Aliases='Dmel_ Dmel\\CG17078, anon-21Cb', description='CG17078 gene product from transcript CG17078-RE', other_designations='CG17078-PB|CG17078-PC|CG17078-PD|CG17078-PE|anon-21Cb|uncharacterized protein', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=274740, end_position_on_the_genomic_accession=278234, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33196, CurrentID=0, Status='live', Symbol='smo', Aliases='Dmel_CG11561, CG11561, Dmel\\CG11561, SMO, Smo, dSmo', description='smoothened', other_designations='CG11561-PA|smo-PA|smooth|smoothene', map_location='2-0.4 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=277582, end_position_on_the_genomic_accession=282167, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33197, CurrentID=0, Status='live', Symbol='CG11601', Aliases='Dmel_ Dmel\\CG11601', description='CG11601 gene product from transcript CG11601-RC', other_designations='CG11601-PA|CG11601-PB|CG11601-PC|uncharacterized protein', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=281670, end_position_on_the_genomic_accession=283269, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33198, CurrentID=0, Status='live', Symbol='CG3625', Aliases='Dmel_ Dmel\\CG3625', description='CG3625 gene product from transcript CG3625-RE', other_designations='CG3625-PA|CG3625-PB|CG3625-PC|CG3625-PE', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=283385, end_position_on_the_genomic_accession=291795, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33199, CurrentID=0, Status='live', Symbol='Amnionless', Aliases='Dmel_CG11592, BEST:CK02467, CG11592, CK02467, Dmel\\CG11592, dAMN', description='Amnionless ortholog', other_designations='Amnionless-PA|Amnionless-PB|CG11592-PA|CG11592-PB|amnionless ortholog', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=287252, end_position_on_the_genomic_accession=289144, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33200, CurrentID=0, Status='live', Symbol='CG11562', Aliases='Dmel_ BcDNA:RE44650, Dmel\\CG11562', description='CG11562 gene product from transcript CG11562-RA', other_designations='CG11562-PA|uncharacterized protein', map_location='21B7-21B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=292419, end_position_on_the_genomic_accession=293222, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33201, CurrentID=0, Status='live', Symbol='U2af38', Aliases='Dmel_CG3582, CG3582, DU2AF38, Dmel\\CG3582, U2AF, U2AF 38, U2AF38, dU2AF38, dU2AF[38], l(2)06751, u2af38', description='U2 small nuclear riboprotein auxiliary factor 38', other_designations='CG3582-PA|CG3582-PB|U2af38-PA|U2af38-PB', map_location='21B7-21B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=292959, end_position_on_the_genomic_accession=294681, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33202, CurrentID=0, Status='live', Symbol='Hop', Aliases='Dmel_CG2720, CG2720, Dmel\\CG2720, EP(2)0418, HOP, Sti1, dHop, dSTI1, dhop, dmSTI1, hop, l(2)k00616', description='Hsp70/Hsp90 organizing protein homolog', other_designations='CG2720-PA|Hop-PA|Hsp70(Hsp90) organizing protein|Hsp70-Hsp90 organizing protein|Hsp70/Hsp90 organizing protein|stress inducible protein 1', map_location='21B8-21B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=295122, end_position_on_the_genomic_accession=297449, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33203, CurrentID=0, Status='live', Symbol='Pi3K21B', Aliases='Dmel_CG2699, CG2699, Dmel\\CG2699, Dp60, PI(3)K, PI3K, PI3K-p60, PI3K21B, PI[[3]]K|p60, Pi3Kp60, dP60, dPI3K, droPIK57, p60', description='CG2699 gene product from transcript CG2699-RE', other_designations='CG2699-PB|CG2699-PC|CG2699-PD|CG2699-PE|PI 3-kinase|PI3 kinase|PI3-kinase|Pi3K21B-PB|Pi3K21B-PC|Pi3K21B-PD|Pi3K21B-PE|dPI 3-kinase', map_location='21B8-21B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=297880, end_position_on_the_genomic_accession=305233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33204, CurrentID=0, Status='live', Symbol='Plc21C', Aliases='Dmel_CG4574, CG4574, Dmel\\CG4574, PLC, PLC-21, PLC-21C, PLC21C, PLC21c, PLC[[Beta]], Plc21, Plcbeta21C, plc-21, plc21, plc21C, plc21c, plcbeta21c', description='Phospholipase C at 21C', other_designations='CG4574-PA|CG4574-PB|CG4574-PC|CG4574-PD|CG4574-PG|CG4574-PH|Plc21C-PA|Plc21C-PB|Plc21C-PC|Plc21C-PD|Plc21C-PG|Plc21C-PH|phospholipase C|phospholipase C 21C|phospholipase C at 21C', map_location='21B8-21C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=305935, end_position_on_the_genomic_accession=355566, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  33205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33205, CurrentID=0, Status='live', Symbol='CG11912', Aliases='Dmel_ Dmel\\CG11912, SP140, anon-WO0140519.112', description='CG11912 gene product from transcript CG11912-RA', other_designations='CG11912-PA|uncharacterized protein', map_location='21B8-21B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=318362, end_position_on_the_genomic_accession=319259, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33206, CurrentID=0, Status='live', Symbol='CG11911', Aliases='Dmel_ Dmel\\CG11911, SP103', description='CG11911 gene product from transcript CG11911-RA', other_designations='CG11911-PA|uncharacterized protein', map_location='21B8-21B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=320279, end_position_on_the_genomic_accession=321248, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33207, CurrentID=0, Status='live', Symbol='Ent1', Aliases='Dmel_CG11907, CG11907, DmENT1, DmEnt1, Dmel\\CG11907, anon-EST:fe2G2, dENT1, ent1', description='Equilibrative nucleoside transporter 1', other_designations='CG11907-PA|CG11907-PB|CG11907-PC|Ent1-PA|Ent1-PB|Ent1-PC|anon-fast-evolving-2G2|equilibrative nucleoside transporter 1', map_location='21C1-21C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=355565, end_position_on_the_genomic_accession=357874, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33208, CurrentID=0, Status='live', Symbol='al', Aliases='Dmel_CG3935, A1, AL, CG3935, Dmel\\CG3935, L13-4, lambda13-4', description='aristaless', other_designations='CG3935-PA|al-PA', map_location='2-0.4 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=378112, end_position_on_the_genomic_accession=387439, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33209, CurrentID=0, Status='live', Symbol='CG4213', Aliases='Dmel_ Dmel\\CG4213', description='CG4213 gene product from transcript CG4213-RA', other_designations='CG4213-PA|CG4213-PB', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=396595, end_position_on_the_genomic_accession=402198, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33210, CurrentID=0, Status='live', Symbol='RpI135', Aliases='Dmel_CG4033, CG4033, DmRP135, Dmel\\CG4033, RP135, RPI135, RpIII135, Rpl135, l(2)k16513', description='RNA polymerase I 135kD subunit', other_designations='CG4033-PA|RNA polymerase I|RpI135-PA', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=404285, end_position_on_the_genomic_accession=407969, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33211, CurrentID=0, Status='live', Symbol='AP-2alpha', Aliases='Dmel_CG4260, AP-2, AP2, Alpha-Adaptin, CG31654, CG4260, D-Ada, D-alphaAda, Dmel\\CG4260, MENE (2L)-A, MENE(2L)-A, ada, alpha, alpha-Ada, alpha-Adaptin, alpha-ada, alpha-adaptin, apl3, cg4260, dAP-2a, l(2)06694, l(2)SH0460, l(2)SH2 0460', description='Adaptor Protein complex 2, alpha subunit', other_designations='AP-2alpha-PA|AP-2alpha-PB|Alpha-adaptin|CG4260-PA|CG4260-PB|adaptor protein complex 2, alpha subunit|alpha adaptin|alphaAdaptin|lethal (2) SH0460', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=408030, end_position_on_the_genomic_accession=414777, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33212, CurrentID=0, Status='live', Symbol='ebi', Aliases='Dmel_CG4063, CG4063, Dmel\\CG4063, E-2f, E-2g, Ebi, FBXW4, Tb11, Tbl1, l(2)k16213', description='CG4063 gene product from transcript CG4063-RA', other_designations='CG4063-PA|beta transducin-like 1 protein|ebi-PA|enhancer on chromosome 2-complementation group 2g', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=415067, end_position_on_the_genomic_accession=418536, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33213, CurrentID=0, Status='live', Symbol='CG13690', Aliases='Dmel_ Dmel\\CG13690', description='CG13690 gene product from transcript CG13690-RA', other_designations='CG13690-PA|uncharacterized protein', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=418434, end_position_on_the_genomic_accession=419652, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33214, CurrentID=0, Status='live', Symbol='RpLP1', Aliases='Dmel_CG4087, CG4087, Dmel\\CG4087, LP1, M(2)21C, M(2)21C1-2, RLA1_DROME, RpA2, RpP2, Rplp1, chr2L:420593..420726, rp21C, rpA2, rpP2, rpa2', description='Ribosomal protein LP1', other_designations='ACIDIC RIBOSOMAL PROTEIN LP1|CG4087-PA|CG4087-PB|Minute(2)21C|Ribosomal-protein-A2|RpLP1-PA|RpLP1-PB|ribosomal protein A2|ribosomal protein LP1|ribosomal protein P2', map_location='2-0.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=419952, end_position_on_the_genomic_accession=420864, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33215, CurrentID=0, Status='live', Symbol='CG11885', Aliases='Dmel_ BcDNA:AT13539, Dmel\\CG11885', description='CG11885 gene product from transcript CG11885-RA', other_designations='CG11885-PA|uncharacterized protein', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=420895, end_position_on_the_genomic_accession=421450, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33216, CurrentID=0, Status='live', Symbol='CG13692', Aliases='Dmel_ Dmel\\CG13692', description='CG13692 gene product from transcript CG13692-RA', other_designations='CG13692-PA|uncharacterized protein', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=421771, end_position_on_the_genomic_accession=422489, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33217, CurrentID=0, Status='live', Symbol='BBS8', Aliases='Dmel_CG13691, BBS2, CG13691, Dmel\\CG13691', description='Bardet-Biedl syndrome 8 ortholog (H. sapiens)', other_designations='BBS8-PA|Bardet-Biedl syndrome 8 ortholog|CG13691-PA', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=422873, end_position_on_the_genomic_accession=425321, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33218, CurrentID=0, Status='live', Symbol='ex', Aliases='Dmel_CG4114, 1270, CG4114, Dmel\\CG4114, Ex, Expanded, brl, l(2)01270, l(2)ey', description='expanded', other_designations='CG4114-PA|CG4114-PB|braille|ex-PA|ex-PB', map_location='2-0.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=431227, end_position_on_the_genomic_accession=448701, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33219, CurrentID=0, Status='live', Symbol='crq', Aliases='Dmel_CG4280, CD36, CG31655, CG4280, CRQ, Cd36, Cqr, Crq, D-CD36, Dmel\\CG4280, anon-WO0140519.8, cqm, cqr', description='croquemort', other_designations='CD36-like|CG4280-PA|CG4280-PB|CG4280-PC|CG4280-PD|CG4280-PE|crq-PA|crq-PB|crq-PC|crq-PD|crq-PE', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=448254, end_position_on_the_genomic_accession=453024, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33220, CurrentID=0, Status='live', Symbol='CG4164', Aliases='Dmel_ Dmel\\CG4164', description='CG4164 gene product from transcript CG4164-RA', other_designations='CG4164-PA', map_location='21C2-21C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=453108, end_position_on_the_genomic_accession=454654, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33221, CurrentID=0, Status='live', Symbol='CG4133', Aliases='Dmel_ Dmel\\CG4133', description='CG4133 gene product from transcript CG4133-RC', other_designations='CG4133-PA|CG4133-PB|CG4133-PC', map_location='21C2-21C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=455531, end_position_on_the_genomic_accession=458745, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33222, CurrentID=0, Status='live', Symbol='CG4297', Aliases='Dmel_ Dmel\\CG4297', description='CG4297 gene product from transcript CG4297-RC', other_designations='CG4297-PC|CG4297-PD', map_location='21C8-21D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=458876, end_position_on_the_genomic_accession=472200, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33223, CurrentID=0, Status='live', Symbol='MED15', Aliases='Dmel_CG4184, Arc105, BcDNA:GH03922, CG4184, Dmel\\CG4184, Med15, Med9/ARC105, dARC105, med15', description='Mediator complex subunit 15', other_designations='CG4184-PA|MED15-PA|mediator complex subunit 15|mediator15', map_location='21D1-21D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=473013, end_position_on_the_genomic_accession=476170, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33224, CurrentID=0, Status='live', Symbol='cbt', Aliases='Dmel_CG4427, CG4427, Cabot, Cabut, Dmel\\CG4427, EP(2)2237, EP2237, dTIEG', description='cabut', other_designations='CG4427-PA|CG4427-PB|CG4427-PC|TGF-beta-inducible early gene|cabot|cbt-PA|cbt-PB|cbt-PC', map_location='21D1-21D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=476446, end_position_on_the_genomic_accession=479688, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33225, CurrentID=0, Status='live', Symbol='ush', Aliases='Dmel_CG2762, CG2762, Dmel\\CG2762, FOG, Ush, ZFPM2, dFOG, l(2)19, u-shaped, ush', description='u-shaped', other_designations='CG2762-PA|CG2762-PB|CG2762-PC|CG2762-PD|CG2762-PE|U-shaped|drosophila friend of GATA|friend of GATA|ush-PA|ush-PB|ush-PC|ush-PD|ush-PE|ushaped', map_location='2-0.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=476220, end_position_on_the_genomic_accession=540560, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33226, CurrentID=0, Status='live', Symbol='lwr', Aliases='Dmel_CG3018, CG3018, DUbc9, DmUbc9, Dmel\\CG3018, Dmubc9, FBgn0010602, Iwr, Lwr, UBC9, Ubc 9, Ubc-9, Ubc9, UbcD9, dUBC9, dUbc9, dip4, hbl, i105, i56, l(2)01519, l(2)02858, l(2)05486, l(2)05487, semi, ubc9', description='lesswright', other_designations='CG3018-PA|CG3018-PB|CG3018-PC|CG3018-PD|hunchback-like|lwr-PA|lwr-PB|lwr-PC|lwr-PD|semushi|ubiquitin conjugating enzyme 9|ubiquitin-conjugating enzyme 9', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=540599, end_position_on_the_genomic_accession=542580, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33227, CurrentID=0, Status='live', Symbol='Spp', Aliases='Dmel_CG11840, CG11840, DmSPP, Dmel\\CG11840, SPP, spp', description='Signal peptide peptidase', other_designations='CG11840-PA|Spp-PA|shanti|signal peptide peptidase|signal peptide protease', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=542786, end_position_on_the_genomic_accession=544653, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33228, CurrentID=0, Status='live', Symbol='nAChRbeta3', Aliases='Dmel_CG11822, CG11822, CT33131, Dbeta3, Dmel\\CG11822, nAChR, nAcR-21Beta, nAcRbeta-21C', description='nicotinic Acetylcholine Receptor beta3', other_designations='CG11822-PA|Dbeta3|nAChRbeta3-PA|nicotinic acetylcholine receptor beta 21C|nicotinic acetylcholine receptor beta3', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=545129, end_position_on_the_genomic_accession=547096, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33229, CurrentID=0, Status='live', Symbol='Ets21C', Aliases='Dmel_CG2914, CG2914, D-Ets-6, D-Ets6, D-ets-6, Dmel\\CG2914, ETS6, Ets, Ets6, ets21C, ets21c', description='Ets at 21C', other_designations='CG2914-PA|CG2914-PB|CG2914-PC|CG2914-PD|Ets21C-PA|Ets21C-PB|Ets21C-PC|Ets21C-PD', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=544925, end_position_on_the_genomic_accession=552923, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33230, CurrentID=0, Status='live', Symbol='rempA', Aliases='Dmel_CG11838, CG11838, CHE-11, Che-11, Dmel\\CG11838, IFT140, IFT40, OSEG, OSEG3, Oseg3, REMPA, oseg3, remp', description='reduced mechanoreceptor potential A', other_designations='CG11838-PB|CG11838-PC|Oseg3|RempA|reduced mechanoreceptor potentials|reduced mechanoreceptor potentials A|rempA-PB|rempA-PC', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=553742, end_position_on_the_genomic_accession=560310, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33231, CurrentID=0, Status='live', Symbol='Tspo', Aliases='Dmel_CG2789, CG2789, Dmel\\CG2789, anon-EST:Posey116, dTSPO, tspo', description='Translocator protein', other_designations='CG2789-PA|Tspo-PA|translocator protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=559612, end_position_on_the_genomic_accession=560523, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33232, CurrentID=0, Status='live', Symbol='CG11835', Aliases='Dmel_ Dmel\\CG11835', description='CG11835 gene product from transcript CG11835-RD', other_designations='CG11835-PA|CG11835-PB|CG11835-PC|CG11835-PD|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=560568, end_position_on_the_genomic_accession=563819, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33233, CurrentID=0, Status='live', Symbol='CG2794', Aliases='Dmel_ Dmel\\CG2794', description='CG2794 gene product from transcript CG2794-RA', other_designations='CG2794-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=564163, end_position_on_the_genomic_accession=566472, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33234, CurrentID=0, Status='live', Symbol='Nle', Aliases='Dmel_CG2863, CG2863, Dmel\\CG2863, anon-WO0026364, l(2)k13714', description='Notchless', other_designations='CG2863-PA|Nle-PA|notchless', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=566366, end_position_on_the_genomic_accession=568121, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33235, CurrentID=0, Status='live', Symbol='CG2807', Aliases='Dmel_ Dmel\\CG2807, SF3B, anon-21Cd', description='CG2807 gene product from transcript CG2807-RB', other_designations='CG2807-PA|CG2807-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=568340, end_position_on_the_genomic_accession=572907, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33236, CurrentID=0, Status='live', Symbol='Ipk2', Aliases='Dmel_CG13688, CG13688, Dmel\\CG13688, dm', description='CG13688 gene product from transcript CG13688-RA', other_designations='CG13688-PA|Ipk2-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=573033, end_position_on_the_genomic_accession=574264, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33237, CurrentID=0, Status='live', Symbol='cold', Aliases='Dmel_CG2813, CG2813, Dmel\\CG2813, cg2813, leaky4', description='coiled', other_designations='CG2813-PA|CG2813-PB|cold-PA|cold-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=574291, end_position_on_the_genomic_accession=575734, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33238, CurrentID=0, Status='live', Symbol='Ptth', Aliases='Dmel_CG13687, CG13687, Dmel\\CG13687, PTTH, phm, ptth', description='Prothoracicotropic hormone', other_designations='CG13687-PE|CG13687-PF|CG13687-PG|Ptth-PE|Ptth-PF|Ptth-PG|prothoracicotopic hormone|prothoracicotropic hormone', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=575711, end_position_on_the_genomic_accession=576896, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33239, CurrentID=0, Status='live', Symbol='Pph13', Aliases='Dmel_CG2819, 60Mun1, CG2819, Dmel\\CG2819, Mu, PPH13, pph13', description='PvuII-PstI homology 13', other_designations='CG2819-PA|Pph13-PA|hazy|munster', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=577486, end_position_on_the_genomic_accession=579549, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33240, CurrentID=0, Status='live', Symbol='Gsc', Aliases='Dmel_CG2851, 60Mun1, 60Mun2, CG2851, D-Gsc, D-gsc, Dmel\\CG2851, M72, PPH25, Pph25, gsc, l(2)05341', description='Goosecoid', other_designations='CG2851-PC|CG2851-PD|CG2851-PE|Gsc-PC|Gsc-PD|Gsc-PE|goosecoid|muenster 72', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=583540, end_position_on_the_genomic_accession=594685, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33241, CurrentID=0, Status='live', Symbol='CG13689', Aliases='Dmel_ Dmel\\CG13689', description='CG13689 gene product from transcript CG13689-RB', other_designations='CG13689-PA|CG13689-PB|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=602813, end_position_on_the_genomic_accession=604430, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33242, CurrentID=0, Status='live', Symbol='lectin-21Cb', Aliases='Dmel_CG13686, AC 004573B, AC004573b, CG13686, Dmel\\CG13686', description='CG13686 gene product from transcript CG13686-RB', other_designations='CG13686-PB|Lectin21Cb|lectin-21Cb-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=620298, end_position_on_the_genomic_accession=621201, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33244, CurrentID=0, Status='live', Symbol='CG2839', Aliases='Dmel_ Dmel\\CG2839', description='CG2839 gene product from transcript CG2839-RA', other_designations='CG2839-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=625652, end_position_on_the_genomic_accession=628200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33245, CurrentID=0, Status='live', Symbol='ds', Aliases='Dmel_CG17941, CG17941, CT39575, Dmel\\CG17941, Ds, EP-822, cad2, dachs, l(2)05142', description='dachsous', other_designations='CG17941-PA|CG17941-PB|cadherin-2|daschous|ds-PA|ds-PB', map_location='2-0.3 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=640013, end_position_on_the_genomic_accession=714983, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  33247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33247, CurrentID=0, Status='live', Symbol='Eaat2', Aliases='Dmel_CG3159, CG3159, Dmel\\CG3159, EAAT, dEAAT2, dEaat2', description='Excitatory amino acid transporter 2', other_designations='CG3159-PA|CG3159-PC|CG3159-PD|Eaat2-PA|Eaat2-PC|Eaat2-PD|excitatory amino acid transporter 2|sodium-dependent excitatory amino acid transporter 2', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=737212, end_position_on_the_genomic_accession=744395, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33248, CurrentID=0, Status='live', Symbol='GABA-B-R3', Aliases='Dmel_CG3022, CG3022, D-GABA[[B]]R3, D-Gaba3, Dmel\\CG3022, GABA B R3, GABA[[B]]-R3, GABA[[B]]R3, GB3, anon-WO0170980.13, anon-WO0170980.14', description='metabotropic GABA-B receptor subtype 3', other_designations='CG3022-PA|CG3022-PC|CG3022-PD|CG3022-PE|CG3022-PF|CG3022-PG|CG3022-PH|GABA-B receptor subtype 3|GABA-B-R3-PA|GABA-B-R3-PC|GABA-B-R3-PD|GABA-B-R3-PE|GABA-B-R3-PF|GABA-B-R3-PG|GABA-B-R3-PH', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=749945, end_position_on_the_genomic_accession=762399, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33249, CurrentID=0, Status='live', Symbol='CG12506', Aliases='Dmel_ Dmel\\CG12506, cg12506', description='CG12506 gene product from transcript CG12506-RA', other_designations='CG12506-PA|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=773547, end_position_on_the_genomic_accession=774017, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33250, CurrentID=0, Status='live', Symbol='CG13947', Aliases='Dmel_ Dmel\\CG13947', description='CG13947 gene product from transcript CG13947-RA', other_designations='CG13947-PA|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=779173, end_position_on_the_genomic_accession=779655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33251, CurrentID=0, Status='live', Symbol='Gr21a', Aliases='Dmel_CG13948, (GR) 21a, 21D.1, CG13948, DmGr21a, Dmel\\CG13948, GR21D.1, GR21D1, GR21a, Gr21D1, gr21a', description='Gustatory receptor 21a', other_designations='CG13948-PA|CG13948-PB|Gr21a-PA|Gr21a-PB|gustatory receptor 21D.1|gustatory receptor 21a', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=780482, end_position_on_the_genomic_accession=782885, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33252, CurrentID=0, Status='live', Symbol='CG3544', Aliases='Dmel_ Dmel\\CG3544', description='CG3544 gene product from transcript CG3544-RB', other_designations='CG3544-PA|CG3544-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=784336, end_position_on_the_genomic_accession=788776, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33253, CurrentID=0, Status='live', Symbol='Pkg21D', Aliases='Dmel_CG3324, CG3324, DG1, Dg1, Dmel\\CG3324, PKG, PKG 21D, PKG-21D, Pkg1, cGMP, dg1, pkg21D', description='Protein kinase, cGMP-dependent at 21D', other_designations='CG3324-PA|Pkg21D-PA|cGMP-dependent protein kinase 21D|protein kinase, cGMP-dependent at 21D', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=786153, end_position_on_the_genomic_accession=790798, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33254, CurrentID=0, Status='live', Symbol='Dbp21E2', Aliases='Dmel_CG3561, CG3561, DmRH9, Dmel\\CG3561, KH1, cg3561', description='DEAD box protein 21E2', other_designations='CG3561-PA|Dbp21E2-PA|KH1', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=811584, end_position_on_the_genomic_accession=813380, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33255, CurrentID=0, Status='live', Symbol='Saf6', Aliases='Dmel_CG3883, CG3883, Dmel\\CG3883, SAF6, saf6', description='SAGA factor-like TAF6', other_designations='CG3883-PA|CG3883-PB|Saf6-PA|Saf6-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=813314, end_position_on_the_genomic_accession=815950, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33256, CurrentID=0, Status='live', Symbol='Pex12', Aliases='Dmel_CG3639, CG3639, DmelPex12, Dmel\\CG3639, PEX12, pex12', description='Peroxin 12', other_designations='CG3639-PA|CG3639-PB|Pex12-PA|Pex12-PB|peroxin 12', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=815993, end_position_on_the_genomic_accession=818096, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33257, CurrentID=0, Status='live', Symbol='CG15880', Aliases='Dmel_ Dmel\\CG15880', description='CG15880 gene product from transcript CG15880-RA', other_designations='CG15880-PA|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=816994, end_position_on_the_genomic_accession=817950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33258, CurrentID=0, Status='live', Symbol='CG3876', Aliases='Dmel_ Dmel\\CG3876', description='CG3876 gene product from transcript CG3876-RA', other_designations='CG3876-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=818079, end_position_on_the_genomic_accession=819508, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33259, CurrentID=0, Status='live', Symbol='Clp', Aliases='Dmel_CG3642, CG3642, CLP, CPSF30, Cpsf30, Dmel\\CG3642, Ssb-c6a, c6a, clp', description='Clipper', other_designations='CG3642-PA|Clp-PA|clipper|single stranded-binding protein c6a', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=819964, end_position_on_the_genomic_accession=821209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33260, CurrentID=0, Status='live', Symbol='CG3662', Aliases='Dmel_ Dmel\\CG3662', description='CG3662 gene product from transcript CG3662-RD', other_designations='CG3662-PA|CG3662-PB|CG3662-PC|CG3662-PD', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=821289, end_position_on_the_genomic_accession=824337, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33261, CurrentID=0, Status='live', Symbol='CG3862', Aliases='Dmel_ Dmel\\CG3862', description='CG3862 gene product from transcript CG3862-RA', other_designations='CG3862-PA|RCC GEF-related protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=824329, end_position_on_the_genomic_accession=825861, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33262, CurrentID=0, Status='live', Symbol='dock', Aliases='Dmel_CG3727, CG3727, DOCK, Dm0447, Dmel\\CG3727, Dock, Nck, dck, doc, l(2)04723', description='dreadlocks', other_designations='CG3727-PA|CG3727-PB|CG3727-PC|CG3727-PD|dock-PA|dock-PB|dock-PC|dock-PD', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=825969, end_position_on_the_genomic_accession=833241, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33263, CurrentID=0, Status='live', Symbol='drongo', Aliases='Dmel_CG3365, CG3365, Dmel\\CG3365', description='CG3365 gene product from transcript CG3365-RH', other_designations='CG3365-PC|CG3365-PF|CG3365-PG|CG3365-PH|CG3365-PI|drongo-PC|drongo-PF|drongo-PG|drongo-PH|drongo-PI', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=833584, end_position_on_the_genomic_accession=851071, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33264, CurrentID=0, Status='live', Symbol='CG4291', Aliases='Dmel_ Dmel\\CG4291', description='CG4291 gene product from transcript CG4291-RB', other_designations='CG4291-PA|CG4291-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=851251, end_position_on_the_genomic_accession=852730, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33265, CurrentID=0, Status='live', Symbol='kraken', Aliases='Dmel_CG3943, CG3943, Dmel\\CG3943, Kraken', description='CG3943 gene product from transcript CG3943-RA', other_designations='CG3943-PA|CG3943-PB|kraken-PA|kraken-PB', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=852768, end_position_on_the_genomic_accession=854539, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33266, CurrentID=0, Status='live', Symbol='CG13949', Aliases='Dmel_ BcDNA:AT06839, Dmel\\CG13949', description='CG13949 gene product from transcript CG13949-RA', other_designations='CG13949-PA|CG13949-PB|CG13949-PC|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=855337, end_position_on_the_genomic_accession=856639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33267, CurrentID=0, Status='live', Symbol='CG13950', Aliases='Dmel_ Dmel\\CG13950', description='CG13950 gene product from transcript CG13950-RA', other_designations='CG13950-PA|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=860309, end_position_on_the_genomic_accession=861806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33268, CurrentID=0, Status='live', Symbol='aru', Aliases='Dmel_CG4276, CG4276, Dmel\\CG4276, EPS8L3', description='arouser', other_designations='CG4276-PA|CG4276-PC|CG4276-PD|EPS8L3|aru-PA|aru-PC|aru-PD', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=861849, end_position_on_the_genomic_accession=868354, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33269, CurrentID=0, Status='live', Symbol='dbe', Aliases='Dmel_CG4258, CG4258, DBE, Dmel\\CG4258, l(2)k05428, l(2)k06708', description='dribble', other_designations='CG4258-PA|dbe-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=868673, end_position_on_the_genomic_accession=869857, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33270, CurrentID=0, Status='live', Symbol='PNUTS', Aliases='Dmel_CG33526, CG31657, CG33526, CG4124, Dmel\\CG33526-RB, PNUTSDm, Pnuts', description='CG33526 gene product from transcript CG33526-RD', other_designations='CG33526-PA|CG33526-PB|CG33526-PD|CG33526-PE|PNUTS-PA|PNUTS-PB|PNUTS-PD|PNUTS-PE', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=870465, end_position_on_the_genomic_accession=877038, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33271, CurrentID=0, Status='live', Symbol='ninaA', Aliases='Dmel_CG3966, CG3966, Dmel\\CG3966, NINAA, NinaA, ninA', description='neither inactivation nor afterpotential A', other_designations='CG3966-PA|either inactivation nor afterpotential A|neither inactivation nor after potential|ninaA-PA', map_location='2-1.4 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=877302, end_position_on_the_genomic_accession=878270, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33272, CurrentID=0, Status='live', Symbol='CG15824', Aliases='Dmel_ Dmel\\CG15824', description='CG15824 gene product from transcript CG15824-RC', other_designations='CG15824-PB|CG15824-PC|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=877172, end_position_on_the_genomic_accession=885141, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  33274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33274, CurrentID=0, Status='live', Symbol='Lsp1beta', Aliases='Dmel_CG4178, CG4178, DmeLSP1b, Dmel\\CG4178, LSP 1, LSP-1, LSP-1 beta, LSP-1beta, LSP1, LSP1beta, Lsp-1b, Lsp-1beta, Lsp1, Lsp1-beta, Lsp1[beta], Lsp1b', description='Larval serum protein 1 beta', other_designations='CG4178-PA|CG4178-PB|Lsp1beta-PA|Lsp1beta-PB|larval serum protein 1|larval serum protein 1 beta|larval serum protein 1[beta]|larval serum protein 1beta', map_location='2-1.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=898500, end_position_on_the_genomic_accession=901316, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33275, CurrentID=0, Status='live', Symbol='GluRIIC', Aliases='Dmel_CG4226, CG4226, CT13021, DGluR-IIC, DGluRIIC, DGluRIII, DmelGluRIIC, Dmel\\CG4226, GLURIIC, GluR, GluR3, GluRC, GluRII-C/III, GluRIII, GlutRIII, glutamate receptor III, GluRIIC', description='Glutamate receptor IIC', other_designations='CG4226-PD|GluRIIC-PD|common type III glutamate receptor subunit|essential glutamate receptor subunit III|glutamate receptor|glutamate receptor IIC|glutamate receptor III|glutamate receptor subunit III', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=914086, end_position_on_the_genomic_accession=917805, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  33276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33276, CurrentID=0, Status='live', Symbol='CG4341', Aliases='Dmel_ BcDNA:GH12144, Dmel\\CG4341', description='CG4341 gene product from transcript CG4341-RA', other_designations='CG4341-PA|CG4341-PC', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=922793, end_position_on_the_genomic_accession=958098, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  33277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33277, CurrentID=0, Status='live', Symbol='IA-2', Aliases='Dmel_CG31795, CG11344, CG31795, CG4355, Dmel\\CG31795, R-PTPX/1A2, anon-WO0153538.69, ia2', description='IA-2 protein tyrosine phosphatase', other_designations='CG31795-PC|CG31795-PD|CG31795-PE|IA-2 ortholog|IA-2-PC|IA-2-PD|IA-2-PE|ia2', map_location='21E3-21E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1008378, end_position_on_the_genomic_accession=1048053, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33278, CurrentID=0, Status='live', Symbol='CG4375', Aliases='Dmel_ BcDNA:AT13819, Dmel\\CG4375', description='CG4375 gene product from transcript CG4375-RA', other_designations='CG4375-PA', map_location='21E3-21E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1026105, end_position_on_the_genomic_accession=1026897, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33279, CurrentID=0, Status='live', Symbol='CG4415', Aliases='Dmel_ Dmel\\CG4415', description='CG4415 gene product from transcript CG4415-RA', other_designations='CG4415-PA|CG4415-PB', map_location='21E3-21E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1037246, end_position_on_the_genomic_accession=1038685, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33281, CurrentID=0, Status='live', Symbol='S', Aliases='Dmel_CG4385, CG4385, Dmel\\CG4385, E(Raf)2B, EC2-2, EK2-2, ES2-4, EY2-4tar/asteroid, fltr, l(2)05671, l(2)07056, l(2)21Eb, l(2)c00080, l(2)k09538, l(2)rK134, star, S', description='Star', other_designations='CG4385-PA|CG4385-PB|S-PA|S-PB|enhancer of GMR-sina 2-4|floater|lethal (2) c00080|star', map_location='2-1.3 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1049674, end_position_on_the_genomic_accession=1077815, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33282, CurrentID=0, Status='live', Symbol='ast', Aliases='Dmel_CG4426, CG4426, Dmel\\CG4426, S[r], Star/asteroid', description='asteroid', other_designations='CG4426-PA|Star-recessive|ast-PA', map_location='21E4-21E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1077948, end_position_on_the_genomic_accession=1080803, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33283, CurrentID=0, Status='live', Symbol='Atg4a', Aliases='Dmel_CG4428, APG4, ATG4, AUT2, Atg4, CG4428, Dmel\\CG4428, atg4', description='Autophagy-related 4a', other_designations='Atg4a-PA|Atg4a-PB|Autophagy-specific gene 4|Autophagy-specific gene 4a|CG4428-PA|CG4428-PB', map_location='21E4-21E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1080911, end_position_on_the_genomic_accession=1085380, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33284, CurrentID=0, Status='live', Symbol='CG4629', Aliases='Dmel_ Dmel\\CG4629, anon-WO0257455.15', description='CG4629 gene product from transcript CG4629-RE', other_designations='CG4629-PE|CG4629-PF', map_location='21E4-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1083091, end_position_on_the_genomic_accession=1102504, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33285, CurrentID=0, Status='live', Symbol='mtRNApol', Aliases='Dmel_CG4644, CG4644, Dmel\\CG4644, POLRMT, mt RNA pol, mtRNA-pol', description='mitochondrial RNA polymerase', other_designations='CG4644-PA|CG4644-PB|mtRNApol-PA|mtRNApol-PB', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1103254, end_position_on_the_genomic_accession=1108373, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33286, CurrentID=0, Status='live', Symbol='CG14339', Aliases='Dmel_ Dmel\\CG14339', description='CG14339 gene product from transcript CG14339-RA', other_designations='CG14339-PA|uncharacterized protein', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1108540, end_position_on_the_genomic_accession=1112694, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33287, CurrentID=0, Status='live', Symbol='CG14340', Aliases='Dmel_ Dmel\\CG14340', description='CG14340 gene product from transcript CG14340-RA', other_designations='CG14340-PA|uncharacterized protein', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1112941, end_position_on_the_genomic_accession=1114138, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33288, CurrentID=0, Status='live', Symbol='Pino', Aliases='Dmel_CG4710, CG4710, Dmel\\CG4710, anon-WO0118547.131, smi21F', description='Pinocchio', other_designations='CG4710-PA|CG4710-PB|CG4710-PC|Pino-PA|Pino-PB|Pino-PC|pinocchio|smell impaired 21 F|smell impaired 21F|smell imparied 21 F', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1114654, end_position_on_the_genomic_accession=1128953, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33289, CurrentID=0, Status='live', Symbol='CG4552', Aliases='Dmel_ Dmel\\CG4552', description='CG4552 gene product from transcript CG4552-RA', other_designations='CG4552-PA', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1129316, end_position_on_the_genomic_accession=1132411, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33290, CurrentID=0, Status='live', Symbol='Iris', Aliases='Dmel_CG4715, CG4715, CG4715-env, Dm-cF, Dmel\\CG4715', description='CG4715 gene product from transcript CG4715-RA', other_designations='CG4715-PA|Iris-PA|iris', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1132332, end_position_on_the_genomic_accession=1134098, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33291, CurrentID=0, Status='live', Symbol='CG4577', Aliases='Dmel_ Dmel\\CG4577', description='CG4577 gene product from transcript CG4577-RB', other_designations='CG4577-PA|CG4577-PB', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1134693, end_position_on_the_genomic_accession=1141156, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33292, CurrentID=0, Status='live', Symbol='MFS3', Aliases='Dmel_CG4726, CG4726, Dmel\\CG4726', description='Major Facilitator Superfamily Transporter 3', other_designations='CG4726-PA|MFS3-PA|major facilitator superfamily transporter 3', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1143201, end_position_on_the_genomic_accession=1147395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33293, CurrentID=0, Status='live', Symbol='CG4749', Aliases='Dmel_ Dmel\\CG4749, anon-WO0118547.81', description='CG4749 gene product from transcript CG4749-RA', other_designations='CG4749-PA', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1147822, end_position_on_the_genomic_accession=1149449, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33294, CurrentID=0, Status='live', Symbol='Tfb4', Aliases='Dmel_CG5041, CG5041, Dmel\\CG5041, Dmp34, TFB4', description='TFB4 ortholog (S. cerevisiae)', other_designations='CG5041-PA|TFB4 ortholog|Tfb4-PA', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1149445, end_position_on_the_genomic_accession=1150677, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33295, CurrentID=0, Status='live', Symbol='Vps29', Aliases='Dmel_CG4764, CG4764, DmVps29, Dmel\\CG4764, VPS29', description='Vacuolar protein sorting 29', other_designations='CG4764-PA|Vps29-PA|vacuolar protein sorting 29', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1150766, end_position_on_the_genomic_accession=1151764, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33298, CurrentID=0, Status='live', Symbol='Tango14', Aliases='Dmel_CG4775, BcDNA:SD10843, CG4775, Dmel\\CG4775, TANGO14, Tango 14, l(2)k00619', description='Transport and Golgi organization 14', other_designations='CG4775-PA|CG4775-PB|Tango14-PA|Tango14-PB|lethal (2) k00619|transport and golgi organization 14', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1158661, end_position_on_the_genomic_accession=1160220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33299, CurrentID=0, Status='live', Symbol='CG5080', Aliases='Dmel_ CT16297, Dmel\\CG5080', description='CG5080 gene product from transcript CG5080-RB', other_designations='CG5080-PA|CG5080-PB|CG5080-PC', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1160290, end_position_on_the_genomic_accession=1163886, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33300, CurrentID=0, Status='live', Symbol='IntS14', Aliases='Dmel_CG4785, CG4785, Dmel\\CG4785, Q9VPY0', description='Integrator 14', other_designations='CG4785-PA|IntS14-PA|integrator 14', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1164443, end_position_on_the_genomic_accession=1166425, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33301, CurrentID=0, Status='live', Symbol='CG14341', Aliases='Dmel_ Dmel\\CG14341', description='CG14341 gene product from transcript CG14341-RB', other_designations='CG14341-PB|uncharacterized protein', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1166506, end_position_on_the_genomic_accession=1167211, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33303, CurrentID=0, Status='live', Symbol='CG5118', Aliases='Dmel_ Dmel\\CG5118, anon-WO0140519.11', description='CG5118 gene product from transcript CG5118-RA', other_designations='CG5118-PA', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1170753, end_position_on_the_genomic_accession=1172805, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33304, CurrentID=0, Status='live', Symbol='CG4887', Aliases='Dmel_ Dmel\\CG4887, cg4887', description='CG4887 gene product from transcript CG4887-RB', other_designations='CG4887-PA|CG4887-PB', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1173009, end_position_on_the_genomic_accession=1177589, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33305, CurrentID=0, Status='live', Symbol='CG4896', Aliases='Dmel_ Dmel\\CG4896, cg4896', description='CG4896 gene product from transcript CG4896-RJ', other_designations='CG4896-PD|CG4896-PG|CG4896-PI|CG4896-PJ|CG4896-PK', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1177690, end_position_on_the_genomic_accession=1182454, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33306, CurrentID=0, Status='live', Symbol='CG5126', Aliases='Dmel_ Dmel\\CG5126', description='CG5126 gene product from transcript CG5126-RA', other_designations='CG5126-PA', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1182297, end_position_on_the_genomic_accession=1184020, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33307, CurrentID=0, Status='live', Symbol='Tgt', Aliases='Dmel_CG4947, CG4947, Dmel\\CG4947, TGT, tgt', description='tRNA-guanine transglycosylase', other_designations='CG4947-PA|Tgt-PA', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1184065, end_position_on_the_genomic_accession=1185698, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33308, CurrentID=0, Status='live', Symbol='CG5001', Aliases='Dmel_ Dmel\\CG5001', description='CG5001 gene product from transcript CG5001-RB', other_designations='CG5001-PA|CG5001-PB|CG5001-PC|CG5001-PD', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1185757, end_position_on_the_genomic_accession=1196807, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33309, CurrentID=0, Status='live', Symbol='CG5139', Aliases='Dmel_ Dmel\\CG5139', description='CG5139 gene product from transcript CG5139-RA', other_designations='CG5139-PA', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1198465, end_position_on_the_genomic_accession=1199266, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33310, CurrentID=0, Status='live', Symbol='CG14342', Aliases='Dmel_ Dmel\\CG14342', description='CG14342 gene product from transcript CG14342-RB', other_designations='CG14342-PB|uncharacterized protein', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1204215, end_position_on_the_genomic_accession=1205250, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33312, CurrentID=0, Status='live', Symbol='CG42329', Aliases='Dmel_ CG14343, CG5156, CT16503, Dmel\\CG42329, Dmel_CG14343, Dmel_CG5156', description='CG42329 gene product from transcript CG42329-RA', other_designations='CG42329-PA|uncharacterized protein', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1219293, end_position_on_the_genomic_accession=1229802, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33313, CurrentID=0, Status='live', Symbol='CG5397', Aliases='Dmel_ CT16527, Dmel\\CG5397', description='CG5397 gene product from transcript CG5397-RB', other_designations='CG5397-PA|CG5397-PB', map_location='21F3-21F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1240036, end_position_on_the_genomic_accession=1243018, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33314, CurrentID=0, Status='live', Symbol='robo3', Aliases='Dmel_CG5423, CG14345, CG5423, CT17158, Dmel\\CG5423, ROBO3, Robo 3, Robo-3, Robo3, dRobo-3', description='roundabout 3', other_designations='CG5423-PA|CG5423-PB|robo3-PA|robo3-PB', map_location='21F3-21F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1256417, end_position_on_the_genomic_accession=1296596, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33317, CurrentID=0, Status='live', Symbol='a5', Aliases='Dmel_CG5430, A5, CG5430, Dmel\\CG5430', description='antennal protein 5', other_designations='CG5430-PA|a5-PA', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1329161, end_position_on_the_genomic_accession=1329993, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33318, CurrentID=0, Status='live', Symbol='CG5440', Aliases='Dmel_ Dmel\\CG5440', description='CG5440 gene product from transcript CG5440-RC', other_designations='CG5440-PB|CG5440-PC', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1334128, end_position_on_the_genomic_accession=1335514, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33319, CurrentID=0, Status='live', Symbol='Cdlc2', Aliases='Dmel_CG5450, CG5450, Cdcl2, Dmel\\CG5450, anon-WO0140519.117, dlc', description='Cytoplasmic dynein light chain 2', other_designations='8kd dynein light chain|CG5450-PA|CG5450-PB|CG5450-PC|Cdlc2-PA|Cdlc2-PB|Cdlc2-PC|cyto dynein light chain 2|cytoplasmic dynein|cytoplasmic dynein light chain 2|dynein', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1344400, end_position_on_the_genomic_accession=1345135, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33320, CurrentID=0, Status='live', Symbol='CG5556', Aliases='Dmel_ Dmel\\CG5556', description='CG5556 gene product from transcript CG5556-RA', other_designations='CG5556-PA|CG5556-PB', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1351728, end_position_on_the_genomic_accession=1352868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33321, CurrentID=0, Status='live', Symbol='CG5561', Aliases='Dmel_ Dmel\\CG5561', description='CG5561 gene product from transcript CG5561-RA', other_designations='CG5561-PA', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1352970, end_position_on_the_genomic_accession=1354109, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33323, CurrentID=0, Status='live', Symbol='CG5565', Aliases='Dmel_ 28557659, Dmel\\CG5565', description='CG5565 gene product from transcript CG5565-RA', other_designations='CG5565-PA', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1354911, end_position_on_the_genomic_accession=1355924, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33324, CurrentID=0, Status='live', Symbol='NLaz', Aliases='Dmel_CG33126, BcDNA:RE67583, CG16933, CG33126, CT33980, DNLaz, Dmel\\CG33126, Nlaz', description='Neural Lazarillo', other_designations='CG33126-PA|CG33126-PB|CG33126-PC|NLaz-PA|NLaz-PB|NLaz-PC|neural lazarillo', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1359837, end_position_on_the_genomic_accession=1361821, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33325, CurrentID=0, Status='live', Symbol='CG14346', Aliases='Dmel_ Dmel\\CG14346', description='CG14346 gene product from transcript CG14346-RC', other_designations='CG14346-PA|CG14346-PC|CG14346-PD|uncharacterized protein', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1367996, end_position_on_the_genomic_accession=1369050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768989, CurrentID=0, Status='live', Symbol='CG33284', Aliases='Dmel_ Dmel\\CG33284, NEST:bs10b10', description='CG33284 gene product from transcript CG33284-RA', other_designations='CG33284-PA|uncharacterized protein', map_location='78B2-78B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21123753, end_position_on_the_genomic_accession=21124884, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33330, CurrentID=0, Status='live', Symbol='CG31661', Aliases='Dmel_ CG10872, Dmel\\CG31661', description='CG31661 gene product from transcript CG31661-RA', other_designations='CG31661-PA|aspartic-type endopeptidase|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1497243, end_position_on_the_genomic_accession=1498583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33331, CurrentID=0, Status='live', Symbol='CG18131', Aliases='Dmel_ BcDNA:AT12853, Dmel\\CG18131', description='CG18131 gene product from transcript CG18131-RG', other_designations='CG18131-PA|CG18131-PB|CG18131-PC|CG18131-PD|CG18131-PG|CG18131-PH|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1498455, end_position_on_the_genomic_accession=1500608, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33332, CurrentID=0, Status='live', Symbol='CG7420', Aliases='Dmel_ Dmel\\CG7420', description='CG7420 gene product from transcript CG7420-RB', other_designations='CG7420-PA|CG7420-PB', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1500784, end_position_on_the_genomic_accession=1502904, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33333, CurrentID=0, Status='live', Symbol='CG18132', Aliases='Dmel_ Dmel\\CG18132', description='CG18132 gene product from transcript CG18132-RA', other_designations='CG18132-PA|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1516795, end_position_on_the_genomic_accession=1517478, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33334, CurrentID=0, Status='live', Symbol='halo', Aliases='Dmel_CG7428, CG7428, Dmel\\CG7428', description='CG7428 gene product from transcript CG7428-RB', other_designations='CG7428-PB|halo-PB', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1517533, end_position_on_the_genomic_accession=1518148, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33335, CurrentID=0, Status='live', Symbol='Or22a', Aliases='Dmel_CG12193, 22A.1, 22a, AN11, CG12193, DOR22A.1, DOR22a, DOR53, DmOr22a, Dmel Or22a, Dmel22a, Dmel\\CG12193, Dor53, OR22a, Or22A.1/b, Or53, dOr22a, dor53, Or22a', description='Odorant receptor 22a', other_designations='CG12193-PA|Or22a-PA|odorant receptor 22a|odorant receptor 53|olfactory receptor 22A.1', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1520491, end_position_on_the_genomic_accession=1522156, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33336, CurrentID=0, Status='live', Symbol='Or22b', Aliases='Dmel_CG4231, 22A.2, 22b, AN12, CG4231, DOR22A.2, DOR22b, DOR67, Dmel\\CG4231, Dor67, OR22b, Or22A.2, Or22a/b, Or67, dOr22b, dor67, or22b', description='Odorant receptor 22b', other_designations='CG4231-PA|Or22b-PA|dmel Or22b|odorant receptor 22b|odorant receptor 67|olfactory receptor 22A.2', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1522697, end_position_on_the_genomic_accession=1524257, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33337, CurrentID=0, Status='live', Symbol='CG10869', Aliases='Dmel_ Dmel\\CG10869', description='CG10869 gene product from transcript CG10869-RA', other_designations='CG10869-PA|uncharacterized protein', map_location='22A3-22A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1584866, end_position_on_the_genomic_accession=1587155, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33338, CurrentID=0, Status='live', Symbol='CG31935', Aliases='Dmel_ CG10870, CG7373, Dmel\\CG31935', description='CG31935 gene product from transcript CG31935-RA', other_designations='CG31935-PA|uncharacterized protein', map_location='22A3-22A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1594822, end_position_on_the_genomic_accession=1611157, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33339, CurrentID=0, Status='live', Symbol='haf', Aliases='Dmel_CG14351, CG14351, CT33985, Dmel\\CG14351, tartan/capricious-like', description='hattifattener', other_designations='CG14351-PB|CG14351-PC|CG14351-PD|CG14351-PE|CG14351-PF|haf-PB|haf-PC|haf-PD|haf-PE|haf-PF', map_location='22A3-22A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1555129, end_position_on_the_genomic_accession=1608404, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33341, CurrentID=0, Status='live', Symbol='CG14352', Aliases='Dmel_ Dmel\\CG14352', description='CG14352 gene product from transcript CG14352-RA', other_designations='CG14352-PA|uncharacterized protein', map_location='22A3-22A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1611410, end_position_on_the_genomic_accession=1612576, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33343, CurrentID=0, Status='live', Symbol='chinmo', Aliases='Dmel_CG31666, CG10871, CG17156, CG17649, CG31666, Chinmo, Dmel\\CG31666, F, l(2)04111', description='Chronologically inappropriate morphogenesis', other_designations='CG31666-PA|CG31666-PD|CG31666-PE|CG31666-PF|CG31666-PG|CG31666-PH|chinmo-PA|chinmo-PD|chinmo-PE|chinmo-PF|chinmo-PG|chinmo-PH|chronologically -inappropriate morphogenesis|chronologically inappropriate morphogenesis|lethal (2) 04111', map_location='22A5-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1651258, end_position_on_the_genomic_accession=1704682, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26067040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067040, CurrentID=0, Status='live', Symbol='CG43844', Aliases='Dmel_ CR43844, Dmel\\CG43844', description='CG43844 gene product from transcript CG43844-RF', other_designations='CG43844-PD|CG43844-PE|CG43844-PF|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21889539, end_position_on_the_genomic_accession=21890171, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33346, CurrentID=0, Status='live', Symbol='cpb', Aliases='Dmel_CG17158, CAPB_DROME, CG17158, CP, CP-beta, CPB, CapZ, Cpb, Dmel\\CG17158, J, bnd', description='capping protein beta', other_designations='CG17158-PA|CG17158-PB|CP beta subunit|Capping-protein-beta|F-actin capping protein|actin capping protein beta|ancpbeta|burned|capping protein|capping proteinbeta|capping proteins beta|cpb-PA|cpb-PB', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1704902, end_position_on_the_genomic_accession=1706962, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33347, CurrentID=0, Status='live', Symbol='CG17660', Aliases='Dmel_ Dmel\\CG17660', description='CG17660 gene product from transcript CG17660-RA', other_designations='CG17660-PA|uncharacterized protein', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1707132, end_position_on_the_genomic_accession=1710421, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33348, CurrentID=0, Status='live', Symbol='mRpL48', Aliases='Dmel_CG17642, CG17642, Dmel\\CG17642, MRP-L48', description='mitochondrial ribosomal protein L48', other_designations='CG17642-PA|CG17642-PB|mRpL48-PA|mRpL48-PB', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1710462, end_position_on_the_genomic_accession=1711465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33349, CurrentID=0, Status='live', Symbol='frtz', Aliases='Dmel_CG17657, CG17657, Dmel\\CG17657, anon-WO0118547.105', description='fritz', other_designations='CG17657-PA|CG17657-PB|CG17657-PD|frtz-PA|frtz-PB|frtz-PD', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1711302, end_position_on_the_genomic_accession=1715851, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33350, CurrentID=0, Status='live', Symbol='Rim2', Aliases='Dmel_CG18317, CG18317, Dmel\\CG18317, drim2', description='Rim2 ortholog (S. cerevisiae)', other_designations='CG18317-PA|CG18317-PC|CG18317-PD|Rim2 ortholog|Rim2-PA|Rim2-PC|Rim2-PD', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1716244, end_position_on_the_genomic_accession=1724388, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33351, CurrentID=0, Status='live', Symbol='Eno', Aliases='Dmel_CG17654, BEST:LD15491, CG17654, CT32526, Dmel\\CG17654, ENOA2, T18, anon-WO0153538.76, eno, Eno', description='Enolase', other_designations='CG17654-PA|CG17654-PB|CG17654-PC|CG17654-PD|CG17654-PE|CG17654-PF|Eno-PA|Eno-PB|Eno-PC|Eno-PD|Eno-PE|Eno-PF|enolase|phosphopyruvate hydratase', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1724768, end_position_on_the_genomic_accession=1729636, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33353, CurrentID=0, Status='live', Symbol='CG17652', Aliases='Dmel_ Dmel\\CG17652', description='CG17652 gene product from transcript CG17652-RB', other_designations='CG17652-PA|CG17652-PB|uncharacterized protein', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1731033, end_position_on_the_genomic_accession=1732356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33354, CurrentID=0, Status='live', Symbol='CG17646', Aliases='Dmel_ ABCG1, BEST:GH05015, CG17647, Dmel\\CG17646, l(2)SH0142, l(2)SH2 0142', description='CG17646 gene product from transcript CG17646-RB', other_designations='ATP-binding cassette transporter G1|CG17646-PA|CG17646-PB|lethal (2) SH0142|uncharacterized protein', map_location='22A4-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1732524, end_position_on_the_genomic_accession=1750612, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  33356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33356, CurrentID=0, Status='live', Symbol='CG17712', Aliases='Dmel_ BcDNA:GH03377, Dmel\\CG17712', description='CG17712 gene product from transcript CG17712-RA', other_designations='CG17712-PA|uncharacterized protein', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1750687, end_position_on_the_genomic_accession=1752122, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33357, CurrentID=0, Status='live', Symbol='CG17648', Aliases='Dmel_ BcDNA:GH02930, Dmel\\CG17648', description='CG17648 gene product from transcript CG17648-RA', other_designations='CG17648-PA|uncharacterized protein', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1752308, end_position_on_the_genomic_accession=1752848, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33358, CurrentID=0, Status='live', Symbol='CG17650', Aliases='Dmel_ Dmel\\CG17650', description='CG17650 gene product from transcript CG17650-RA', other_designations='CG17650-PA|uncharacterized protein', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1758286, end_position_on_the_genomic_accession=1759093, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33359, CurrentID=0, Status='live', Symbol='CG31664', Aliases='Dmel_ CG15622, Dmel\\CG31664', description='CG31664 gene product from transcript CG31664-RC', other_designations='CG31664-PB|CG31664-PC|uncharacterized protein', map_location='22B3-22B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1824172, end_position_on_the_genomic_accession=1826500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33360, CurrentID=0, Status='live', Symbol='c-cup', Aliases='Dmel_CG15623, CG15623, Dmel\\CG15623', description='calcutta cup', other_designations='CG15623-PA|c-cup-PA|calcutta-cup', map_location='22B3-22B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1831627, end_position_on_the_genomic_accession=1832565, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33361, CurrentID=0, Status='live', Symbol='wry', Aliases='Dmel_CG31665, CG12704, CG15621, CG31665, CT35769, Dmel\\CG31665', description='weary', other_designations='CG31665-PB|CG31665-PC|CG31665-PD|fibropellin|wry-PB|wry-PC|wry-PD', map_location='22B3-22B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1836538, end_position_on_the_genomic_accession=1858550, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  33363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33363, CurrentID=0, Status='live', Symbol='CG31663', Aliases='Dmel_ BcDNA:GH27779, CG15359, CG15360, Dmel\\CG31663', description='CG31663 gene product from transcript CG31663-RC', other_designations='CG31663-PA|CG31663-PB|CG31663-PC|uncharacterized protein', map_location='22B4-22B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1858609, end_position_on_the_genomic_accession=1881480, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  33365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33365, CurrentID=0, Status='live', Symbol='CG7295', Aliases='Dmel_ Dmel\\CG7295', description='CG7295 gene product from transcript CG7295-RB', other_designations='CG7295-PA|CG7295-PB', map_location='22B4-22B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1869803, end_position_on_the_genomic_accession=1871032, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33366, CurrentID=0, Status='live', Symbol='CG15358', Aliases='Dmel_ Dmel\\CG15358, mdcds_4396', description='CG15358 gene product from transcript CG15358-RC', other_designations='CG15358-PB|CG15358-PC|uncharacterized protein', map_location='22B4-22B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1882008, end_position_on_the_genomic_accession=1883185, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33367, CurrentID=0, Status='live', Symbol='CG7337', Aliases='Dmel_ Dmel\\CG7337, Q0E8U7_DROME', description='CG7337 gene product from transcript CG7337-RA', other_designations='CG7337-PA|CG7337-PB|CG7337-PC|CG7337-PD|CG7337-PE|CG7337-PF|CG7337-PG|CG7337-PH', map_location='22B4-22B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1884622, end_position_on_the_genomic_accession=1944289, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  33369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33369, CurrentID=0, Status='live', Symbol='Der-1', Aliases='Dmel_CG10908, CG10908, Dmel\\CG10908, l(2)SH1964, l(2)SH2 1964', description='Derlin-1', other_designations='CG10908-PA|CG10908-PB|Der-1-PA|Der-1-PB|Dm Derlin-1|derlin-1|lethal (2) SH1964', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1973957, end_position_on_the_genomic_accession=1975413, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33370, CurrentID=0, Status='live', Symbol='CG15356', Aliases='Dmel_ Dmel\\CG15356, EMSY', description='CG15356 gene product from transcript CG15356-RA', other_designations='CG15356-PA|uncharacterized protein', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1975463, end_position_on_the_genomic_accession=1980755, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33371, CurrentID=0, Status='live', Symbol='CG15362', Aliases='Dmel_ Dmel\\CG15362', description='CG15362 gene product from transcript CG15362-RA', other_designations='CG15362-PA|uncharacterized protein', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1981176, end_position_on_the_genomic_accession=1982099, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33372, CurrentID=0, Status='live', Symbol='CG7289', Aliases='Dmel_ Dmel\\CG7289', description='CG7289 gene product from transcript CG7289-RA', other_designations='CG7289-PA', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1982217, end_position_on_the_genomic_accession=1984514, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33373, CurrentID=0, Status='live', Symbol='Got2', Aliases='Dmel_CG4233, 152801_at, CG4233, Dmel\\CG4233, Got, Got 2, Got-2, got2', description='Glutamate oxaloacetate transaminase 2', other_designations='CG4233-PA|CG4233-PB|GLUTAMATE OXALOACETATE TRANSAMINASE 2|Got2-PA|Got2-PB|glutamate oxaloacetate transaminase 2|mitochondrial aspartate aminotransferase', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1984512, end_position_on_the_genomic_accession=1987758, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33374, CurrentID=0, Status='live', Symbol='Npc2a', Aliases='Dmel_CG7291, BEST:GH08376, BcDNA:GH22047, BcDNA:RH53228, CG7291, CT22505, DmML3, Dmel\\CG7291, NPC2, dnpc2a, npc2a', description='Niemann-Pick type C-2a', other_designations='CG7291-PA|Niemann-Pick type C-2|Npc2a-PA|epididymal secretory protein', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1988290, end_position_on_the_genomic_accession=1988940, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26067045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067045, CurrentID=0, Status='live', Symbol='Mst77Y-13', Aliases='Dmel_CG45767, CG45767, Dmel\\CG45767', description='Male-specific transcript 77Y 13', other_designations='CG45767-PA|Mst77Y-13-PA', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3354105, end_position_on_the_genomic_accession=3354870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33377, CurrentID=0, Status='live', Symbol='CG4238', Aliases='Dmel_ Dmel\\CG4238', description='CG4238 gene product from transcript CG4238-RF', other_designations='CG4238-PD|CG4238-PE|CG4238-PF|CG4238-PG|CG4238-PH|CG4238-PI|CG4238-PJ', map_location='22B8-22C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2009859, end_position_on_the_genomic_accession=2030788, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33379, CurrentID=0, Status='live', Symbol='Su(dx)', Aliases='Dmel_CG4244, AIP4, CG4244, Dmel\\CG4244, ITCH, Itch, Su(Dx), Su(deltex), anon-WO0073329.1, anon-WO0073329.2, anon-WO0073329.4, su(dx)', description='Suppressor of deltex', other_designations='CG4244-PA|CG4244-PB|CG4244-PC|CG4244-PE|CG4244-PF|Su(dx)-PA|Su(dx)-PB|Su(dx)-PC|Su(dx)-PE|Su(dx)-PF|suppressor of deltex', map_location='22C1-22C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2037714, end_position_on_the_genomic_accession=2044372, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33381, CurrentID=0, Status='live', Symbol='Or22c', Aliases='Dmel_CG15377, 22C.1, 22c, CG15377, DOR16, DOR22C.1, Dmel\\CG15377, Or22C.1', description='Odorant receptor 22c', other_designations='CG15377-PA|Or22c-PA|odorant receptor 22c|olfactory receptor 22C.1', map_location='22C1-22C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2054589, end_position_on_the_genomic_accession=2057967, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33384, CurrentID=0, Status='live', Symbol='CG12674', Aliases='Dmel_ Dmel\\CG12674', description='CG12674 gene product from transcript CG12674-RA', other_designations='CG12674-PA|CG12674-PB|uncharacterized protein', map_location='22C2-22C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2079919, end_position_on_the_genomic_accession=2081286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33385, CurrentID=0, Status='live', Symbol='CG4259', Aliases='Dmel_ Dmel\\CG4259, SPH173', description='CG4259 gene product from transcript CG4259-RB', other_designations='CG4259-PA|CG4259-PB', map_location='22C3-22C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2127309, end_position_on_the_genomic_accession=2129237, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33386, CurrentID=0, Status='live', Symbol='GlyP', Aliases='Dmel_CG7254, BcDNA:LD24485, CG7254, DGPH, Dmel\\CG7254, GP, Glp1, Glyp, anon-WO0257455.33, l(2)k07918', description='Glycogen phosphorylase', other_designations='CG7254-PA|CG7254-PB|GlyP-PA|GlyP-PB|glycogen phosphorylase|glycogen phosphorylase b', map_location='22C3-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2130813, end_position_on_the_genomic_accession=2137296, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26067047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067047, CurrentID=0, Status='live', Symbol='Mst77Y-9', Aliases='Dmel_CG45769, CG45769, Dmel\\CG45769', description='Male-specific transcript 77Y 9', other_designations='CG45769-PA|Mst77Y-9-PA', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3336400, end_position_on_the_genomic_accession=3337123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33388, CurrentID=0, Status='live', Symbol='CG11723', Aliases='Dmel_ 11723, Dmel\\CG11723', description='CG11723 gene product from transcript CG11723-RB', other_designations='CG11723-PA|CG11723-PB|uncharacterized protein', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2145297, end_position_on_the_genomic_accession=2147275, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33389, CurrentID=0, Status='live', Symbol='TBCD', Aliases='Dmel_CG7261, BcDNA:LD22320, CG7261, Dmel\\CG7261', description='tubulin folding cofactor D', other_designations='CG7261-PA|TBCD-PA', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2147518, end_position_on_the_genomic_accession=2151664, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33390, CurrentID=0, Status='live', Symbol='AIF', Aliases='Dmel_CG7263, CG7263, DmAIF, Dmel\\CG7263, SD03428p, aif', description='Apoptosis inducing factor', other_designations='AIF-PA|AIF-PB|AIF-PC|CG7263-PA|CG7263-PB|CG7263-PC|apoptosis inducing factor', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2151754, end_position_on_the_genomic_accession=2155389, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33391, CurrentID=0, Status='live', Symbol='CG15382', Aliases='Dmel_ BcDNA:RE26473, Dmel\\CG15382', description='CG15382 gene product from transcript CG15382-RA', other_designations='CG15382-PA|uncharacterized protein', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2155753, end_position_on_the_genomic_accession=2156790, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33392, CurrentID=0, Status='live', Symbol='aop', Aliases='Dmel_CG3166, AOP, Aop, Aop/Yan, CG3166, DROYANET, DROYANETSB, Dmel\\CG3166, SK2-1, YAN, Yan/yan, fus6, pok, yan, yan/pok, aop', description='anterior open', other_designations='CG3166-PA|CG3166-PB|CG3166-PC|CG3166-PD|CG3166-PE|SGP cluster fusion defects 6|anterior-open|aop-PA|aop-PB|aop-PC|aop-PD|aop-PE|pokkuri', map_location='2-12 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2156484, end_position_on_the_genomic_accession=2178754, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33394, CurrentID=0, Status='live', Symbol='CG10874', Aliases='Dmel_ Dmel\\CG10874, FBgn 31395', description='CG10874 gene product from transcript CG10874-RB', other_designations='CG10874-PA|CG10874-PB|CG10874-PC|CG10874-PD|uncharacterized protein', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2189914, end_position_on_the_genomic_accession=2192386, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33395, CurrentID=0, Status='live', Symbol='CG31668', Aliases='Dmel_CG31668, CG10879-PB, Dmel\\CG31668, dSLC5A12', description='CG31668 gene product from transcript CG31668-RC', other_designations='CG31668-PC|uncharacterized protein', map_location='22D1-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2203721, end_position_on_the_genomic_accession=2206166, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33396, CurrentID=0, Status='live', Symbol='CG15385', Aliases='Dmel_ Dmel\\CG15385', description='CG15385 gene product from transcript CG15385-RA', other_designations='CG15385-PA|CG15385-PB|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2209574, end_position_on_the_genomic_accession=2212571, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33397, CurrentID=0, Status='live', Symbol='Uch', Aliases='Dmel_CG4265, CG4265, Dmel\\CG4265, P35122, UBL, UCH, UCH-D, UCH-D19-D, dUch, uch, uch-D, Uch', description='Ubiquitin carboxy-terminal hydrolase', other_designations='CG4265-PA|CG4265-PB|CG4265-PC|Uch-PA|Uch-PB|Uch-PC|ubiquitin C-terminal hydrolase|ubiquitin carboxy-terminal hydrolase|ubiquitin carboxyl terminal hydrolase|ubiquitin carboxyterminal hydrolase', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2212738, end_position_on_the_genomic_accession=2214211, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33398, CurrentID=0, Status='live', Symbol='CG10880', Aliases='Dmel_ Dmel\\CG10880', description='CG10880 gene product from transcript CG10880-RB', other_designations='CG10880-PA|CG10880-PB|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2214972, end_position_on_the_genomic_accession=2216311, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33399, CurrentID=0, Status='live', Symbol='mio', Aliases='Dmel_CG7074, CG15384, CG7074, Dmel\\CG7074, FBtr0077835, Mio, SEA4', description='missing oocyte', other_designations='CG7074-PA|mio-PA', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2217290, end_position_on_the_genomic_accession=2220625, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33401, CurrentID=0, Status='live', Symbol='papi', Aliases='Dmel_CG7082, CG7082, Dmel\\CG7082, PAPI, cg7082', description='CG7082 gene product from transcript CG7082-RA', other_designations='CG7082-PA|CG7082-PB|CG7082-PC|CG7082-PD|papi-PA|papi-PB|papi-PC|papi-PD|partner of PIWIs', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2220748, end_position_on_the_genomic_accession=2225544, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33402, CurrentID=0, Status='live', Symbol='sau', Aliases='Dmel_CG7085, BEST:LD28579, BcDNA:LD10657, CG7085, Dmel\\CG7085, GOLPH3, l(2)s5379, rti', description='sauron', other_designations='CG7085-PA|CG7085-PD|CG7085-PE|lethal (2) s5379|rotini|sau-PA|sau-PD|sau-PE', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2226342, end_position_on_the_genomic_accession=2231099, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33403, CurrentID=0, Status='live', Symbol='CG15387', Aliases='Dmel_ Dmel\\CG15387', description='CG15387 gene product from transcript CG15387-RB', other_designations='CG15387-PB|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2225545, end_position_on_the_genomic_accession=2226294, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33404, CurrentID=0, Status='live', Symbol='Tengl3', Aliases='Dmel_CG31679, CG17006, CG31679, Dmel\\CG31679', description='Testis EndoG-Like 3', other_designations='CG31679-PA|Tengl3-PA|testis EndoG-Like 3', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2237076, end_position_on_the_genomic_accession=2238251, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33405, CurrentID=0, Status='live', Symbol='CG4267', Aliases='Dmel_ CG31686, Dmel\\CG4267, Dmel_CG31686', description='CG4267 gene product from transcript CG4267-RB', other_designations='CG4267-PA|CG4267-PB', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2242378, end_position_on_the_genomic_accession=2245274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33406, CurrentID=0, Status='live', Symbol='Ser12', Aliases='Dmel_CG17240, CG17240, Dmel\\CG17240, SP92', description='Serine protease 12', other_designations='CG17240-PA|Ser12-PA|serine protease 12', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2250431, end_position_on_the_genomic_accession=2251275, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33407, CurrentID=0, Status='live', Symbol='Send1', Aliases='Dmel_CG17012, CG17012, Dmel\\CG17012, SP163', description='Spermathecal endopeptidase 1', other_designations='CG17012-PA|Send1-PA|spermathecal endopeptidase 1', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2254176, end_position_on_the_genomic_accession=2255050, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33408, CurrentID=0, Status='live', Symbol='CG4270', Aliases='Dmel_ BcDNA:AT04879, Dmel\\CG4270', description='CG4270 gene product from transcript CG4270-RA', other_designations='CG4270-PA|CG4270-PB', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2263442, end_position_on_the_genomic_accession=2264241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33409, CurrentID=0, Status='live', Symbol='Sec24CD', Aliases='Dmel_CG10882, BcDNA:LP05220, CG10882, Dmel\\CG10882, FBgn0031408, Sec24, Sten, gho, sec24, sfb3, sten', description='Sec24CD ortholog (H. sapiens)', other_designations='CG10882-PA|CG10882-PB|CG10882-PC|Sec24CD ortholog|Sec24CD-PA|Sec24CD-PB|Sec24CD-PC|ghost|stenosis', map_location='2-68 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2231106, end_position_on_the_genomic_accession=2236844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33410, CurrentID=0, Status='live', Symbol='CG4271', Aliases='Dmel_ Dmel\\CG4271, SPH164', description='CG4271 gene product from transcript CG4271-RA', other_designations='CG4271-PA', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2285163, end_position_on_the_genomic_accession=2285926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33411, CurrentID=0, Status='live', Symbol='CG17237', Aliases='Dmel_ BcDNA:AT15471, DmAAF51272, Dmel\\CG17237', description='CG17237 gene product from transcript CG17237-RA', other_designations='CG17237-PA|uncharacterized protein', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2288503, end_position_on_the_genomic_accession=2289416, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33412, CurrentID=0, Status='live', Symbol='robl22E', Aliases='Dmel_CG10838, CG10838, Dmel\\CG10838, Robl22E', description='CG10838 gene product from transcript CG10838-RB', other_designations='CG10838-PA|CG10838-PB|robl22E-PA|robl22E-PB', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2290224, end_position_on_the_genomic_accession=2290738, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33413, CurrentID=0, Status='live', Symbol='CG16995', Aliases='Dmel_ Dmel\\CG16995', description='CG16995 gene product from transcript CG16995-RA', other_designations='CG16995-PA|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2307948, end_position_on_the_genomic_accession=2308872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33414, CurrentID=0, Status='live', Symbol='CG9967', Aliases='Dmel_ Dmel\\CG9967', description='CG9967 gene product from transcript CG9967-RC', other_designations='CG9967-PA|CG9967-PB|CG9967-PC', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2308761, end_position_on_the_genomic_accession=2357356, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26067052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067052, CurrentID=0, Status='live', Symbol='Myo81F', Aliases='Dmel_CG45784, CG15831, CG40155, CG40204, CG41281, CG41518, CG41527, CG42621, CG42622, CG42623, CG42624, CG45784, Dmel\\CG45784, Dmel_CG15831, Dmel_CG40155, Dmel_CG40204, Dmel_CG41281, Dmel_CG41518, Dmel_CG41527, Dmel_CG42621, Dmel_CG42622, Dmel_CG42623, Dmel_CG42624', description='Myosin 81F', other_designations='CG45784-PB|Myo81F-PB|myosin 81F', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=567076, end_position_on_the_genomic_accession=2532932, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  33418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33418, CurrentID=0, Status='live', Symbol='Rab5', Aliases='Dmel_CG3664, AAF51265, BAA88244, CG3664, D Dm Rab5, DmRab5, Dmel\\CG3664, Drab5, RAB5, RAB5a, Rab, Rab-5, dRab5, drab5, l(2)k08232, rab5', description='CG3664 gene product from transcript CG3664-RC', other_designations='CG3664-PA|CG3664-PB|CG3664-PC|CG3664-PD|CG3664-PE|CG3664-PF|CG3664-PG|CG3664-PH|CG3664-PI|Rab-protein 5|Rab5-PA|Rab5-PB|Rab5-PC|Rab5-PD|Rab5-PE|Rab5-PF|Rab5-PG|Rab5-PH|Rab5-PI', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2359142, end_position_on_the_genomic_accession=2365273, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  33419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33419, CurrentID=0, Status='live', Symbol='Axud1', Aliases='Dmel_CG4272, BcDNA:GH09817, C676, CG4272, D Dmel\\CG4272', description='CG4272 gene product from transcript CG4272-RC', other_designations='Axud1-PA|Axud1-PB|Axud1-PC|Axud1-PD|CG4272-PA|CG4272-PB|CG4272-PC|CG4272-PD', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2365345, end_position_on_the_genomic_accession=2370137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33420, CurrentID=0, Status='live', Symbol='CG3597', Aliases='Dmel_ BVRA, Dmel\\CG3597', description='CG3597 gene product from transcript CG3597-RA', other_designations='CG3597-PA|CG3597-PB', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2370152, end_position_on_the_genomic_accession=2371524, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33421, CurrentID=0, Status='live', Symbol='CG3609', Aliases='Dmel_ CT12127, Dmel\\CG3609', description='CG3609 gene product from transcript CG3609-RA', other_designations='CG3609-PA|CG3609-PB|CG3609-PC', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2372455, end_position_on_the_genomic_accession=2374627, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33422, CurrentID=0, Status='live', Symbol='CG15390', Aliases='Dmel_ BcDNA:RE18748, Dmel\\CG15390', description='CG15390 gene product from transcript CG15390-RA', other_designations='CG15390-PA|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2374713, end_position_on_the_genomic_accession=2375697, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33423, CurrentID=0, Status='live', Symbol='Atxn7', Aliases='Dmel_CG9866, CG9866, Dmel\\CG9866', description='Ataxin 7', other_designations='Ataxin-7|Atxn7-PA|Atxn7-PB|Atxn7-PC|Atxn7-PD|CG9866-PA|CG9866-PB|CG9866-PC|CG9866-PD|ataxin 7', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2375841, end_position_on_the_genomic_accession=2380506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33424, CurrentID=0, Status='live', Symbol='Eogt', Aliases='Dmel_CG9867, CG9867, Dmel\\CG9867, eogt', description='EGF-domain O-GlcNAc transferase', other_designations='CG9867-PA|CG9867-PB|Eogt-PA|Eogt-PB', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2381097, end_position_on_the_genomic_accession=2383073, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33425, CurrentID=0, Status='live', Symbol='CG9870', Aliases='Dmel_ Dmel\\CG9870', description='CG9870 gene product from transcript CG9870-RA', other_designations='CG9870-PA', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2383202, end_position_on_the_genomic_accession=2384747, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33426, CurrentID=0, Status='live', Symbol='CG3557', Aliases='Dmel_ Dmel\\CG3557', description='CG3557 gene product from transcript CG3557-RC', other_designations='CG3557-PA|CG3557-PB|CG3557-PC', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2384761, end_position_on_the_genomic_accession=2385792, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33427, CurrentID=0, Status='live', Symbol='VGlut', Aliases='Dmel_CG9887, CG9887, DV-Glut, DVGLUT, DVGluT, DVGlut, Dmel\\CG9887, DvGluT, DvGlut, MFS11, VGLUT, VGluT, Vglut, dVGLUT, dVGlut, dvGluT, dvGlut, dvglut, vGLUT, vGluT, vGlut, vglut', description='Vesicular glutamate transporter', other_designations='CG9887-PA|CG9887-PB|CG9887-PC|VGlut-PA|VGlut-PB|VGlut-PC|drosophila vesicular glutamate transporter|vesicular glutamate transporter', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2391660, end_position_on_the_genomic_accession=2410662, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  33429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33429, CurrentID=0, Status='live', Symbol='CG18641', Aliases='Dmel_ CG11535, Dmel\\CG18641', description='CG18641 gene product from transcript CG18641-RC', other_designations='CG18641-PC|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2410624, end_position_on_the_genomic_accession=2415322, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33431, CurrentID=0, Status='live', Symbol='CG9886', Aliases='Dmel_ Dmel\\CG9886', description='CG9886 gene product from transcript CG9886-RB', other_designations='CG9886-PA|CG9886-PB', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2420830, end_position_on_the_genomic_accession=2424647, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33432, CurrentID=0, Status='live', Symbol='dpp', Aliases='Dmel_CG9885, BMP, CG9885, DPP, DPP-C, Dm-DPP, Dmel\\CG9885, Dpp, Hin-d, M(2)23AB, M(2)LS1, TGF-b, TGF-beta, TGFbeta, Tg, blk, ho, l(2)10638, l(2)22Fa, l(2)k17036, shv', description='decapentaplegic', other_designations='CG9885-PA|CG9885-PB|CG9885-PC|CG9885-PE|Decapentaplegic/Bone morphogenetic protein|Haplo-insufficient|blink|bone morphogenetic protein|bone morphogenic protein|dpp-PA|dpp-PB|dpp-PC|dpp-PE|heldout|shortvein|tegula', map_location='2-4.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2428372, end_position_on_the_genomic_accession=2459823, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33434, CurrentID=0, Status='live', Symbol='Slh', Aliases='Dmel_CG3539, CG15393, CG3539, Dmel\\CG3539, Dmel_CG15393, SLY1, Sly, Sly1, Sly1p, d-sly, slh', description='SLY-1 homologous', other_designations='CG3539-PC|CG3539-PG|SLY1 homologous|Slh-PC|Slh-PG', map_location='22F3-22F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2488197, end_position_on_the_genomic_accession=2492671, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33435, CurrentID=0, Status='live', Symbol='oaf', Aliases='Dmel_CG9884, CG9884, Dmel\\CG9884, l(2)22Fb, l(2)ND1, tnd', description='out at first', other_designations='CG9884-PA|CG9884-PB|CG9884-PC|CG9884-PD|near dpp complementation group 1|oaf-PA|oaf-PB|oaf-PC|oaf-PD|transcript near decapentaplegic|transcript-near-decapentaplegic', map_location='22F4-22F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2492955, end_position_on_the_genomic_accession=2498846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33436, CurrentID=0, Status='live', Symbol='CG3528', Aliases='Dmel_ Dmel\\CG3528', description='CG3528 gene product from transcript CG3528-RA', other_designations='CG3528-PA', map_location='22F4-22F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2517102, end_position_on_the_genomic_accession=2518037, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33437, CurrentID=0, Status='live', Symbol='CG3515', Aliases='Dmel_ Dmel\\CG3515', description='CG3515 gene product from transcript CG3515-RA', other_designations='CG3515-PA', map_location='22F4-22F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2551145, end_position_on_the_genomic_accession=2552825, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33438, CurrentID=0, Status='live', Symbol='Cyp309a1', Aliases='Dmel_CG9964, 309a1, CG9964, Dmel\\CG9964', description='CG9964 gene product from transcript CG9964-RC', other_designations='CG9964-PB|CG9964-PC|Cyp309a1-PB|Cyp309a1-PC|cytochrome P450', map_location='22F4-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2562483, end_position_on_the_genomic_accession=2564830, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33439, CurrentID=0, Status='live', Symbol='Cyp309a2', Aliases='Dmel_CG18559, 309a2, CG18559, Dmel\\CG18559', description='CG18559 gene product from transcript CG18559-RA', other_designations='CG18559-PA|Cyp309a2-PA|cytochrome P450 309a2', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2564886, end_position_on_the_genomic_accession=2573037, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33440, CurrentID=0, Status='live', Symbol='Rad1', Aliases='Dmel_CG3240, CG3240, DmRad1, Dmel\\CG3240, Dromel_CG3240_FBtr0077747_rad1_mORF[Dm], rad1, Rad1', description='Radiation insensitive 1', other_designations='CG3240-PA|Rad1-PA|radiation insensitive 1', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2573436, end_position_on_the_genomic_accession=2575052, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33441, CurrentID=0, Status='live', Symbol='insv', Aliases='Dmel_CG3227, BcDNA:RE55538, CG3227, Dmel\\CG3227, Insv', description='insensitive', other_designations='CG3227-PA|CG3227-PB|insv-PA|insv-PB', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2575280, end_position_on_the_genomic_accession=2576997, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33442, CurrentID=0, Status='live', Symbol='Elba2', Aliases='Dmel_CG9883, CG9883, Dmel\\CG9883, Elba', description='Early boundary activity 2', other_designations='CG9883-PA|Elba2-PA|early boundary activity 2|early boundary factor', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2577687, end_position_on_the_genomic_accession=2579179, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33443, CurrentID=0, Status='live', Symbol='ND-B17.2', Aliases='Dmel_CG3214, B17.2, BcDNA:GH12382, CG3214, Dmel\\CG3214, N7BM, anon-EST:Posey62', description='NADH dehydrogenase (ubiquinone) B17.2 subunit', other_designations='CG3214-PA|CG3214-PB|CG3214-PC|FBpp0077425|NADH ubiquinone oxidoreductase B17.2|NADH-ubiquinone oxidoreductase subunit B17.2|NADH:ubiquinone oxidoreductase B17.2 subunit|ND-B17.2-PA|ND-B17.2-PB|ND-B17.2-PC', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2579201, end_position_on_the_genomic_accession=2580172, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33444, CurrentID=0, Status='live', Symbol='Arpc5', Aliases='Dmel_CG9881, ARPC5, ArpC5, CG9881, Dmel\\CG9881, P16-ARC, p16, p16-ARC, p16-arc', description='Actin-related protein 2/3 complex, subunit 5', other_designations='Arpc5-PA|Arpc5-PB|CG9881-PA|CG9881-PB|p16-ARC', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2580292, end_position_on_the_genomic_accession=2581423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33445, CurrentID=0, Status='live', Symbol='Drp1', Aliases='Dmel_CG3210, CG3210, DRP, DRP1, Dmel\\CG3210, Dnm1/Drp1, Drp, Drp-1, drp-1, drp1, frat, l(2)22Fc, l(2)22Fd, l(2)ND2, l(2)ND3', description='Dynamin related protein 1', other_designations='CG3210-PA|CG3210-PB|Drp1-PA|Drp1-PB|dynamin related protein 1|dynamin related protein-1|dynamin-2|dynamin-related protein|dynamin-related protein 1|dynaminrelated protein 1|fratboy|lethal(2)22Fc|lethal(2)22Fd|near dpp complementation group 2|near dpp complementation group 3|noodle', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2581281, end_position_on_the_genomic_accession=2585430, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33446, CurrentID=0, Status='live', Symbol='CG15394', Aliases='Dmel_ Dmel\\CG15394, anon-SAGE:Wang-060', description='CG15394 gene product from transcript CG15394-RD', other_designations='CG15394-PC|CG15394-PD|anon-SAGE:Wang-060|uncharacterized protein', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2591869, end_position_on_the_genomic_accession=2592773, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33447, CurrentID=0, Status='live', Symbol='CG15395', Aliases='Dmel_ Dmel\\CG15395', description='CG15395 gene product from transcript CG15395-RB', other_designations='CG15395-PB|uncharacterized protein', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2609994, end_position_on_the_genomic_accession=2611283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33448, CurrentID=0, Status='live', Symbol='CG9962', Aliases='Dmel_ Dmel\\CG9962', description='CG9962 gene product from transcript CG9962-RA', other_designations='CG9962-PA|CKIa-like', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2646577, end_position_on_the_genomic_accession=2647720, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33449, CurrentID=0, Status='live', Symbol='Prosbeta4R1', Aliases='Dmel_CG17301, BcDNA:AT30033, CG17301, Dmel\\CG17301, beta4R1', description='Proteasome beta4 subunit-related 1', other_designations='CG17301-PA|Prosbeta4R1-PA|proteasome beta4 subunit-related 1|proteasome beta4R1 subunit', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2653046, end_position_on_the_genomic_accession=2653926, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33450, CurrentID=0, Status='live', Symbol='Or23a', Aliases='Dmel_CG9880, 23A.1, 23a, AN5, CG9880, DOR23A.1, DOR64, Dmel\\CG9880, Dor64, Or23A.1, Or64, dor64, or23a', description='Odorant receptor 23a', other_designations='CG9880-PA|Or23a-PA|odorant receptor 23a|odorant receptor 64|olfactory receptor 23A.1', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2654049, end_position_on_the_genomic_accession=2655462, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33451, CurrentID=0, Status='live', Symbol='Prosbeta4R2', Aliases='Dmel_CG17302, CG17302, Dmel\\CG17302, anon-WO0140519.122, beta4R2', description='Proteasome beta4 subunit-related 2', other_designations='CG17302-PA|CG17302-PB|Prosbeta4R2-PA|Prosbeta4R2-PB|proteasome beta4 subunit-related 2|proteasome beta4R2 subunit', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2655394, end_position_on_the_genomic_accession=2657102, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33452, CurrentID=0, Status='live', Symbol='CG9879', Aliases='Dmel_ Dmel\\CG9879', description='CG9879 gene product from transcript CG9879-RA', other_designations='CG9879-PA', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2665062, end_position_on_the_genomic_accession=2666265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33453, CurrentID=0, Status='live', Symbol='Gr23a', Aliases='Dmel_CG15396, 23A.1a, 23A1.b, 23A1a, 23A1b, CG15396, Dmel\\CG15396, GR23A.1, GR23a, Gr23A1a, Gr23A1b.a, Gr23a.b, Gr23aA, Gr23aB, Gr23a', description='Gustatory receptor 23a', other_designations='CG15396-PB|CG15396-PC|Gr23a-PB|Gr23a-PC|gustatory receptor 23A.1|gustatory receptor 23a', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2669823, end_position_on_the_genomic_accession=2672186, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33454, CurrentID=0, Status='live', Symbol='CG15398', Aliases='Dmel_ Dmel\\CG15398', description='CG15398 gene product from transcript CG15398-RB', other_designations='CG15398-PB|uncharacterized protein', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2695330, end_position_on_the_genomic_accession=2696437, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33455, CurrentID=0, Status='live', Symbol='CG31690', Aliases='Dmel_ CG11542, CG15397, Dmel\\CG31690', description='CG31690 gene product from transcript CG31690-RE', other_designations='CG31690-PE|uncharacterized protein', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2677734, end_position_on_the_genomic_accession=2719431, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33458, CurrentID=0, Status='live', Symbol='Hrs', Aliases='Dmel_CG2903, CG2903, DmHRS, Dmel\\CG2903, HRS, HRS-2, Vps27, dHrs, hrs, l(2)23AB5, l(2)23Ad, l(2)23Ad(Hrs), vps27', description='Hepatocyte growth factor regulated tyrosine kinase substrate', other_designations='CG2903-PA|CG2903-PB|CG2903-PC|F[[1]]F[[0]] ATPase|HGF-regulated tyrosine kinase substrate|Hrs-PA|Hrs-PB|Hrs-PC|hepatocyte growth factor regulated tyrosine kinase|hepatocyte growth factor regulated tyrosine kinase substrate|hepatocyte growth factor-regulated tyrosine kinase substrate|hepatocyte growth factor-regulated tyrosine substrate|hepatocyte growth-factor-regulated substrate', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2739986, end_position_on_the_genomic_accession=2743377, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33459, CurrentID=0, Status='live', Symbol='CG9961', Aliases='Dmel_ Dmel\\CG9961', description='CG9961 gene product from transcript CG9961-RB', other_designations='CG9961-PA|CG9961-PB', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2743375, end_position_on_the_genomic_accession=2744934, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33460, CurrentID=0, Status='live', Symbol='Cwc25', Aliases='Dmel_CG2843, CG2843, CWC25, Dmel\\CG2843', description='CG2843 gene product from transcript CG2843-RA', other_designations='CG2843-PA|Cwc25-PA', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2745098, end_position_on_the_genomic_accession=2746944, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33461, CurrentID=0, Status='live', Symbol='Pgk', Aliases='Dmel_CG3127, 169C9S, CG3127, Dmel\\CG3127, ESTS:169C9S, PGK, anon-sts18, l(2)23AB3, l(2)23Ac, pgk', description='Phosphoglycerate kinase', other_designations='CG3127-PA|CG3127-PB|CG3127-PC|CG3127-PD|CG3127-PE|Pgk-PA|Pgk-PB|Pgk-PC|Pgk-PD|Pgk-PE|lethal (2) 23AB3|phosphoglycerate kinase|phosphoglycerokinase', map_location='23A3-23B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2746880, end_position_on_the_genomic_accession=2753068, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33462, CurrentID=0, Status='live', Symbol='Bacc', Aliases='Dmel_CG9894, CG9894, Dmel\\CG9894, anon-EST:Liang-2.54, clone 2.54', description='Bacchus', other_designations='Bacc-PA|Bacc-PB|Bacc-PC|Bacc-PD|CG9894-PA|CG9894-PB|CG9894-PC|CG9894-PD|bacchus|baccus', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2752445, end_position_on_the_genomic_accession=2757975, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33465, CurrentID=0, Status='live', Symbol='Trn-SR', Aliases='Dmel_CG2848, CG2848, Dmel\\CG2848, TNPO3, anon-EST:Posey294, dTRN-SR', description='Transportin-Serine/Arginine rich', other_designations='CG2848-PA|CG2848-PB|Transportin-serine/arginine rich|Trn-SR-PA|Trn-SR-PB|transportin', map_location='23A3-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2759924, end_position_on_the_genomic_accession=2763889, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33466, CurrentID=0, Status='live', Symbol='CG3077', Aliases='Dmel_ DmSNX21, Dmel\\CG3077', description='CG3077 gene product from transcript CG3077-RB', other_designations='CG3077-PA|CG3077-PB', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2763416, end_position_on_the_genomic_accession=2765665, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33467, CurrentID=0, Status='live', Symbol='aph-1', Aliases='Dmel_CG2855, Aph-1, Aph1, CG2855, Dm APH-1, Dmel\\CG2855, Pen-1, aph, aph1, dAph-1, l(2)23AB6, l(2)23Ag, pen-1', description='anterior pharynx defective 1', other_designations='CG2855-PA|CG2855-PB|aph-1-PA|aph-1-PB|presenilin enhancer-1|presenilin stabilization factor', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2766233, end_position_on_the_genomic_accession=2767281, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33468, CurrentID=0, Status='live', Symbol='Taf10b', Aliases='Dmel_CG3069, CG3069, Dmel\\CG3069, TAF, TAF10b, TAFII16, TAF[[II]], TAF[[II]]16, TFIID, Taf10, Taf16, TafII16, dTAF10b, dTAF[[II]]16, dmTAF10b, dtafII16', description='TBP-associated factor 10b', other_designations='CG3069-PA|TBP-associated factor|TBP-associated factor 16kD|Taf10b-PA', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2767286, end_position_on_the_genomic_accession=2767845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33469, CurrentID=0, Status='live', Symbol='Taf10', Aliases='Dmel_CG2859, BcDNA:RE73934, CG2859, Dmel\\CG2859, TAF, TAF10, TAF[[II]], TAF[[II]]24, TFIID, Taf24, TafII24, dTAF10, dTAFII24, dTAF[[II]]24, dmTAF10, dmTaf10, dtafII24, taf24', description='TBP-associated factor 10', other_designations='CG2859-PA|TBP-associated factor|TBP-associated factor 24kD|Taf10-PA', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2767890, end_position_on_the_genomic_accession=2768612, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33470, CurrentID=0, Status='live', Symbol='colt', Aliases='Dmel_CG3057, CG3057, Dmel\\CG3057', description='congested-like trachea', other_designations='CG3057-PA|CG3057-PD|collapsed trachea|colt-PA|colt-PD', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2768590, end_position_on_the_genomic_accession=2770069, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33471, CurrentID=0, Status='live', Symbol='CG2862', Aliases='Dmel_ Dmel\\CG2862', description='CG2862 gene product from transcript CG2862-RA', other_designations='CG2862-PA|CG2862-PB|CG2862-PC', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2770135, end_position_on_the_genomic_accession=2770904, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33472, CurrentID=0, Status='live', Symbol='CG15399', Aliases='Dmel_ Dmel\\CG15399', description='CG15399 gene product from transcript CG15399-RA', other_designations='CG15399-PA|uncharacterized protein', map_location='23A5-23A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2770884, end_position_on_the_genomic_accession=2771776, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33473, CurrentID=0, Status='live', Symbol='Syt1', Aliases='Dmel_CG3139, CG3139, D. Syt I, DSYT, DSYT2, DSypt, DSyt, DSyt1, Dmel\\CG3139, Droso1, Dsyt2, STY1, SYT, SYT I, SYT1, Syt, Syt 1, Syt I, Syt-1, SytI, dSYT, dSyt1, dsyt1, dsytI, l(2)23AB1, l(2)23Ba, l(2)k05909, syt, syt 1, syt I, syt-1, syt1, sytI', description='Synaptotagmin 1', other_designations='CG3139-PA|CG3139-PB|CG3139-PC|CG3139-PE|CG3139-PF|CG3139-PG|CG3139-PH|CG3139-PI|Synaptotagmin-1|Synaptotagmin-I|Synaptotagmin1|SynaptotagminI|Syt1-PA|Syt1-PB|Syt1-PC|Syt1-PE|Syt1-PF|Syt1-PG|Syt1-PH|Syt1-PI|synaptoptagmin 1|synaptotagmin|synaptotagmin 1|synaptotagmin I', map_location='23A5-23B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2772327, end_position_on_the_genomic_accession=2799979, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33474, CurrentID=0, Status='live', Symbol='daw', Aliases='Dmel_CG16987, ALP23B, Alp, Alp23B, Alp23b, CG16987, DAW, Daw, Dmel\\CG16987, cg16987', description='dawdle', other_designations='Activin-Like protein|CG16987-PA|CG16987-PB|CG16987-PC|activin like protein|activin like protein at 23B|dActivin2|daw-PA|daw-PB|daw-PC', map_location='23B1-23B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2805261, end_position_on_the_genomic_accession=2812376, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33475, CurrentID=0, Status='live', Symbol='CG2964', Aliases='Dmel_ Dmel\\CG2964', description='CG2964 gene product from transcript CG2964-RA', other_designations='CG2964-PA', map_location='23B2-23B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2812437, end_position_on_the_genomic_accession=2814429, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33476, CurrentID=0, Status='live', Symbol='G6P', Aliases='Dmel_CG15400, CG15400, Dmel\\CG15400, G6pase', description='Glucose-6-Phosphatase', other_designations='CG15400-PB|G6P-PB', map_location='23B2-23B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2814336, end_position_on_the_genomic_accession=2815760, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33477, CurrentID=0, Status='live', Symbol='Duox', Aliases='Dmel_CG3131, CG3131, DUOX, Dmel\\CG3131, dDuox, dduox, duox', description='Dual oxidase', other_designations='CG3131-PB|CG3131-PC|Duox-PB|Duox-PC|NADPH dual oxidase|dual oxidase', map_location='23B2-23B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2815970, end_position_on_the_genomic_accession=2830248, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33478, CurrentID=0, Status='live', Symbol='CG3123', Aliases='Dmel_ Dmel\\CG3123, gi21064005, gi7295901', description='CG3123 gene product from transcript CG3123-RA', other_designations='CG3123-PA', map_location='23B2-23B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2823546, end_position_on_the_genomic_accession=2825506, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33479, CurrentID=0, Status='live', Symbol='CG3119', Aliases='Dmel_ Dmel\\CG3119, dC1GalT7', description='CG3119 gene product from transcript CG3119-RA', other_designations='CG3119-PA|CG3119-PB', map_location='23B3-23B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2834887, end_position_on_the_genomic_accession=2837833, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33480, CurrentID=0, Status='live', Symbol='Cpr23B', Aliases='Dmel_CG2973, CG2973, DmelCpr23B, Dmel\\CG2973', description='Cuticular protein 23B', other_designations='CG2973-PA|Cpr23B-PA|cuticular protein 23B', map_location='23B3-23B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2838411, end_position_on_the_genomic_accession=2839547, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33481, CurrentID=0, Status='live', Symbol='CG2975', Aliases='Dmel_ Dmel\\CG2975, dC1GalT4', description='CG2975 gene product from transcript CG2975-RA', other_designations='CG2975-PA', map_location='23B4-23B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2840124, end_position_on_the_genomic_accession=2842638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33482, CurrentID=0, Status='live', Symbol='CG18558', Aliases='Dmel_ Dmel\\CG18558, dC1GalT3', description='CG18558 gene product from transcript CG18558-RB', other_designations='CG18558-PB|uncharacterized protein', map_location='23B4-23B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2842589, end_position_on_the_genomic_accession=2844513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33483, CurrentID=0, Status='live', Symbol='CG18557', Aliases='Dmel_ Dmel\\CG18557, c-SPH125', description='CG18557 gene product from transcript CG18557-RA', other_designations='CG18557-PA|uncharacterized protein', map_location='23B4-23B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2844601, end_position_on_the_genomic_accession=2845734, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33484, CurrentID=0, Status='live', Symbol='CG3117', Aliases='Dmel_ Dmel\\CG3117, SPH94, c-SPH94', description='CG3117 gene product from transcript CG3117-RB', other_designations='CG3117-PB', map_location='23B4-23B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2845926, end_position_on_the_genomic_accession=2847212, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33485, CurrentID=0, Status='live', Symbol='CG2983', Aliases='Dmel_ Dmel\\CG2983, dC1GalT8', description='CG2983 gene product from transcript CG2983-RB', other_designations='CG2983-PB', map_location='23B5-23B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2847632, end_position_on_the_genomic_accession=2848997, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33486, CurrentID=0, Status='live', Symbol='CG3104', Aliases='Dmel_ Dmel\\CG3104', description='CG3104 gene product from transcript CG3104-RA', other_designations='CG3104-PA|CG3104-PB|CG3104-PC', map_location='23B5-23B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2848989, end_position_on_the_genomic_accession=2855778, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33487, CurrentID=0, Status='live', Symbol='RpS21', Aliases='Dmel_CG2986, CG2986, Dmel\\CG2986, M(2)23B, l(2)03575, l(2)168/14, l(2)k16814, oho23, oho23B, rpS21', description='Ribosomal protein S21', other_designations='CG2986-PA|CG2986-PB|CG2986-PD|CG2986-PE|CG2986-PF|Rp S21|RpS21-PA|RpS21-PB|RpS21-PD|RpS21-PE|RpS21-PF|overgrown hematopoietic organs 23B|overgrown hematopoietic organs at 23B|overgrown hematopoietic organs-23|ribosomal protein S21', map_location='23B5-23B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2856070, end_position_on_the_genomic_accession=2857029, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33488, CurrentID=0, Status='live', Symbol='CG2991', Aliases='Dmel_ CT10107, CT42388, Dmel\\CG2991', description='CG2991 gene product from transcript CG2991-RA', other_designations='CG2991-PA|CG2991-PB|CG2991-PC', map_location='23B6-23B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2857164, end_position_on_the_genomic_accession=2868548, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33489, CurrentID=0, Status='live', Symbol='CG31694', Aliases='Dmel_ CG15401, CG3098, Dmel\\CG31694', description='CG31694 gene product from transcript CG31694-RA', other_designations='CG31694-PA|uncharacterized protein', map_location='23B7-23B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2868627, end_position_on_the_genomic_accession=2874489, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33490, CurrentID=0, Status='live', Symbol='CG8813', Aliases='Dmel_ Dmel\\CG8813, anon-WO0140519.96', description='CG8813 gene product from transcript CG8813-RA', other_designations='CG8813-PA', map_location='23B7-23B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2869652, end_position_on_the_genomic_accession=2870649, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33492, CurrentID=0, Status='live', Symbol='CG8814', Aliases='Dmel_ Dmel\\CG8814', description='CG8814 gene product from transcript CG8814-RA', other_designations='CG8814-PA', map_location='23B8-23C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2874890, end_position_on_the_genomic_accession=2877168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33493, CurrentID=0, Status='live', Symbol='Prx6005', Aliases='Dmel_CG3083, 153181_at, CG3083, DPx-6005, DPx6005, Dmel\\CG3083, GM04269p', description='Peroxiredoxin 6005', other_designations='CG3083-PA|Prx6005-PA|peroxiredoxin 6005', map_location='23C1-23C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2877091, end_position_on_the_genomic_accession=2878232, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33495, CurrentID=0, Status='live', Symbol='NTPase', Aliases='Dmel_CG3059, CG3059, Dmel\\CG3059, NTPDase6, dm', description='CG3059 gene product from transcript CG3059-RD', other_designations='CD39-like NTPase|CG3059-PA|CG3059-PC|CG3059-PD|CG3059-PE|CG3059-PF|NTPase-PA|NTPase-PC|NTPase-PD|NTPase-PE|NTPase-PF', map_location='23C1-23C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2880704, end_position_on_the_genomic_accession=2885850, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33496, CurrentID=0, Status='live', Symbol='lilli', Aliases='Dmel_CG8817, 154454_at, 189Y, AFF4, BEST:GH09955, BcDNA:GM02019, CG8817, Dmel\\CG8817, Lilli, S(ls)4, SS2-1, SY2-1, Su(Raf)2A, Su(Raf)2a, dAFF4, l(2)00632, l(2)23Cc, l(2)K00632, l(2)K07920, l(2)k07920, l(2)k16124', description='lilliputian', other_designations='CG8817-PA|CG8817-PB|CG8817-PC|CG8817-PD|CG8817-PE|CG8817-PF|CG8817-PG|CG8817-PH|liliputian|lilli-PA|lilli-PB|lilli-PC|lilli-PD|lilli-PE|lilli-PF|lilli-PG|lilli-PH|suppressor of GMR-sina 2-1', map_location='23C1-23C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2885929, end_position_on_the_genomic_accession=2954406, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  33498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33498, CurrentID=0, Status='live', Symbol='Rbp9', Aliases='Dmel_CG3151, CG15402, CG3151, Dmel\\CG3151, RBP9, RBP9-2, RRM10, RRM9, Rbp10, Rpb9, fes, fs(2)B, rbp, rbp9, rbp9a, rrm10, rrm9', description='RNA-binding protein 9', other_designations='CG3151-PB|CG3151-PC|CG3151-PE|CG3151-PF|CG3151-PG|CG3151-PH|CG3151-PI|CG3151-PJ|CG3151-PK|RNA binding protein 9|RNA-binding protein 10|Rbp9-PB|Rbp9-PC|Rbp9-PE|Rbp9-PF|Rbp9-PG|Rbp9-PH|Rbp9-PI|Rbp9-PJ|Rbp9-PK|arrest|female sterile|female sterile (2) bridges|female sterile(2)Bridges', map_location='2-5 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2954762, end_position_on_the_genomic_accession=2968434, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33499, CurrentID=0, Status='live', Symbol='Ts', Aliases='Dmel_CG3181, CG3181, Dmel\\CG3181, FBgn0024920, TS, anon-23Ca, c1.5', description='Thymidylate synthase', other_designations='CG3181-PA|CG3181-PB|Ts-PA|Ts-PB|thymidilate synthetase|thymidylate synthase', map_location='23C3-23C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2968293, end_position_on_the_genomic_accession=2969638, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33500, CurrentID=0, Status='live', Symbol='Rrp1', Aliases='Dmel_CG3178, 3178, CG3178, Dmel\\CG3178, RRP1, RRP1_DROME, Rrp-1, rrp', description='Recombination repair protein 1', other_designations='CG3178-PA|CG3178-PB|Rrp1-PA|Rrp1-PB|recombination repair protein 1|strand transferase', map_location='23C3-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2969773, end_position_on_the_genomic_accession=2972508, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33501, CurrentID=0, Status='live', Symbol='gammaTub23C', Aliases='Dmel_CG3157, CG 3157, CG3157, D.m.GAMMA-23C, Dmel\\CG3157, GAMMA23C, GTU88*, GTUB, Tub23, g-Tubulin, gamma tubulin, gamma-Tub, gamma-Tub23C, gamma-tub, gamma-tub 23C, gamma-tub23, gamma-tub23C, gamma-tubulin, gammaTUB, gammaTUB23C, gammaTubD, gamma[zy], gammatub, gammatub23C, l(2)23Ce, gammaTub23C', description='gamma-Tubulin at 23C', other_designations='&ggr-tubulin|23C gamma-tubulin|23C-gamma-tubulin|CG3157-PA|gama-tubulin|gamma tubulin|gamma tubulin 23C|gamma-Tubulin|gamma-tubulin (23C)|gamma-tubulin 23C|gamma-tubulin(23C)|gamma-tubulin-1|gamma-tubulin23C|gammaTub23C-PA|gammatubulin|lethal (2) 23Ce|tubulin', map_location='23C1-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2972891, end_position_on_the_genomic_accession=2974858, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33502, CurrentID=0, Status='live', Symbol='CG9641', Aliases='Dmel_ Dmel\\CG9641', description='CG9641 gene product from transcript CG9641-RB', other_designations='CG9641-PA|CG9641-PB', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2974859, end_position_on_the_genomic_accession=2976862, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33503, CurrentID=0, Status='live', Symbol='CG3165', Aliases='Dmel_ Dmel\\CG3165', description='CG3165 gene product from transcript CG3165-RA', other_designations='CG3165-PA|CG3165-PB', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2977122, end_position_on_the_genomic_accession=2978500, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33504, CurrentID=0, Status='live', Symbol='CG9643', Aliases='Dmel_ Dmel\\CG9643', description='CG9643 gene product from transcript CG9643-RB', other_designations='CG9643-PA|CG9643-PB', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2978411, end_position_on_the_genomic_accession=2979288, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33505, CurrentID=0, Status='live', Symbol='Chd1', Aliases='Dmel_CG3733, 2, CG3733, CHD-1, CHD1, DmCHD1, Dmel\\CG3733, chd1, dCHD1, l(2)23Cd', description='Chromodomain-helicase-DNA-binding protein 1', other_designations='CG3733-PA|CG3733-PB|CG3733-PC|Chd1-PA|Chd1-PB|Chd1-PC|Chromodomain-helicase-DNA-binding protein|chromo-ATPase/helicase-DNA-binding domain|chromo-ATPase/helicase-DNA-binding protein 1|lethal (2) 23Cd', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2979562, end_position_on_the_genomic_accession=2987723, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19835678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835678, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-K', Aliases='Dmel_CG44879, CG44879, Dmel\\CG44879', description='CG44879 gene product from transcript CG44879-RB', other_designations='CG44879-PA|CG44879-PB|pre-mod(mdg4)-K-PA|pre-mod(mdg4)-K-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21355692, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33507, CurrentID=0, Status='live', Symbol='okr', Aliases='Dmel_CG3736, CG3736, DhR54, DmRAD54, DmRad54, Dmel\\CG3736, Dmrad54, Okr, RAD54, RAD54L, Rad54, Ras54, rad54', description='okra', other_designations='CG3736-PA|okr-PA', map_location='2-1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2989594, end_position_on_the_genomic_accession=2992759, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33509, CurrentID=0, Status='live', Symbol='CG17265', Aliases='Dmel_ Dmel\\CG17265', description='CG17265 gene product from transcript CG17265-RA', other_designations='CG17265-PA|uncharacterized protein', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2999286, end_position_on_the_genomic_accession=3016069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33510, CurrentID=0, Status='live', Symbol='CG17224', Aliases='Dmel_ Dmel\\CG17224', description='CG17224 gene product from transcript CG17224-RC', other_designations='CG17224-PA|CG17224-PB|CG17224-PC|CG17224-PD|uncharacterized protein', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3016441, end_position_on_the_genomic_accession=3018308, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33511, CurrentID=0, Status='live', Symbol='CG17264', Aliases='Dmel_ Dmel\\CG17264', description='CG17264 gene product from transcript CG17264-RA', other_designations='CG17264-PA|CG17264-PB|uncharacterized protein', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3018508, end_position_on_the_genomic_accession=3021506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33512, CurrentID=0, Status='live', Symbol='alpha4GT1', Aliases='Dmel_CG17223, CG17223, Dmel\\CG17223, alpha-4GT1', description='CG17223 gene product from transcript CG17223-RA', other_designations='CG17223-PA|alpha1,4-N-acetylgalactosaminyltransferase-1|alpha4GT1-PA', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3021309, end_position_on_the_genomic_accession=3023141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33513, CurrentID=0, Status='live', Symbol='CG3542', Aliases='Dmel_ Dmel\\CG3542, prp40', description='CG3542 gene product from transcript CG3542-RA', other_designations='CG3542-PA', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3023226, end_position_on_the_genomic_accession=3026439, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33514, CurrentID=0, Status='live', Symbol='CG3605', Aliases='Dmel_ Dmel\\CG3605', description='CG3605 gene product from transcript CG3605-RB', other_designations='CG3605-PA|CG3605-PB', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3026252, end_position_on_the_genomic_accession=3029168, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33515, CurrentID=0, Status='live', Symbol='CG17219', Aliases='Dmel_ Dmel\\CG17219', description='CG17219 gene product from transcript CG17219-RA', other_designations='CG17219-PA|uncharacterized protein', map_location='23C4-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3029253, end_position_on_the_genomic_accession=3030616, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33516, CurrentID=0, Status='live', Symbol='GABPI', Aliases='Dmel_CG17257, CG17257, Dmel\\CG17257', description='beta4GalNAcTB pilot', other_designations='CG17257-PA|CG17257-PB|CG17257-PC|CG17257-PD|GABPI-PA|GABPI-PB|GABPI-PC|GABPI-PD', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3030539, end_position_on_the_genomic_accession=3032300, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33517, CurrentID=0, Status='live', Symbol='CG17258', Aliases='Dmel_ Dmel\\CG17258', description='CG17258 gene product from transcript CG17258-RC', other_designations='CG17258-PB|CG17258-PC|CG17258-PD|uncharacterized protein', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3032210, end_position_on_the_genomic_accession=3035348, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33518, CurrentID=0, Status='live', Symbol='CG17259', Aliases='Dmel_ DmSRS1, Dmel\\CG17259, SerRS', description='CG17259 gene product from transcript CG17259-RA', other_designations='CG17259-PA|CG17259-PB|seryl-tRNA synthetase|uncharacterized protein', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3035493, end_position_on_the_genomic_accession=3037403, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33519, CurrentID=0, Status='live', Symbol='CG17260', Aliases='Dmel_ Dmel\\CG17260', description='CG17260 gene product from transcript CG17260-RC', other_designations='CG17260-PA|CG17260-PB|CG17260-PC|uncharacterized protein', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3037230, end_position_on_the_genomic_accession=3038793, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33520, CurrentID=0, Status='live', Symbol='cnir', Aliases='Dmel_CG17262, CG17262, Cnir, Dmel\\CG17262', description='cornichon related', other_designations='CG17262-PA|Cni-related|cnir-PA', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3038826, end_position_on_the_genomic_accession=3039812, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33521, CurrentID=0, Status='live', Symbol='CG17221', Aliases='Dmel_ DM_7295851, Dmel\\CG17221', description='CG17221 gene product from transcript CG17221-RB', other_designations='CG17221-PA|CG17221-PB|CG17221-PC|uncharacterized protein', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3039976, end_position_on_the_genomic_accession=3041890, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33522, CurrentID=0, Status='live', Symbol='CG17261', Aliases='Dmel_ Dmel\\CG17261', description='CG17261 gene product from transcript CG17261-RB', other_designations='CG17261-PA|CG17261-PB|uncharacterized protein', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3042551, end_position_on_the_genomic_accession=3043593, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33524, CurrentID=0, Status='live', Symbol='FASN1', Aliases='Dmel_CG3523, BcDNA:GH07626, BcDNA:gh07626, CG3523, DM_7295848, Dmel\\CG3523, FAS, FASN, FASN[CG3523], FAS[CG3523], Fas, Fatty acid synthase, anon-EST:Posey235, anon-EST:Posey43, anon-WO0138581.1, dFAS, fas', description='Fatty acid synthase 1', other_designations='CG3523-PA|CG3523-PC|FASN1-PA|FASN1-PC|fatty acid synthase|fatty acid synthase 1|fatty acid synthetase', map_location='23C5-23D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3056598, end_position_on_the_genomic_accession=3068177, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33526, CurrentID=0, Status='live', Symbol='toc', Aliases='Dmel_CG9660, BA34, BcDNA:LD27161, CG9660, Dmel\\CG9660, Toc, l(2)01361, l(2)05527, l(2)K08224, l(2)k08224', description='toucan', other_designations='CG9660-PA|CG9660-PB|CG9660-PC|CG9660-PD|CG9660-PE|CG9660-PF|CG9660-PG|CG9660-PH|CG9660-PI|CG9660-PJ|toc-PA|toc-PB|toc-PC|toc-PD|toc-PE|toc-PF|toc-PG|toc-PH|toc-PI|toc-PJ', map_location='23D1-23D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3068345, end_position_on_the_genomic_accession=3144786, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  33527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33527, CurrentID=0, Status='live', Symbol='CG15403', Aliases='Dmel_ Dmel\\CG15403', description='CG15403 gene product from transcript CG15403-RA', other_designations='CG15403-PA|uncharacterized protein', map_location='23D2-23D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3128635, end_position_on_the_genomic_accession=3129489, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33528, CurrentID=0, Status='live', Symbol='ND-B14.5B', Aliases='Dmel_CG12400, BcDNA:RE29623, CG12400, Dmel\\CG12400, N4BM', description='NADH dehydrogenase (ubiquinone) B14.5 B subunit', other_designations='CG12400-PA|CG12400-PB|NADH ubiquinone oxidoreductase B14.5B|NADH-ubiquinone oxidoreductase subunit B14.5b|NADH:ubiquinone oxidoreductase B14.5B subunit|ND-B14.5B-PA|ND-B14.5B-PB', map_location='23D2-23D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3144953, end_position_on_the_genomic_accession=3145848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33529, CurrentID=0, Status='live', Symbol='Mad', Aliases='Dmel_CG12399, 2/23, CG12399, Dmel\\CG12399, E(zen)2, En(vvl), MAD, Mat, P-Mad, P-mad, PMad, Smad, Smad1, apg, c28, dMAD, dMad, l(2)K00237, l(2)k00237, mad, p-Mad, pMAD, pMad, pSmad', description='Mothers against dpp', other_designations='CG12399-PA|CG12399-PB|Mad-PA|Mad-PB|apang|mother against Dpp|mother against decapentaplegic|mothers against decapentaplegi|mothers against decapentaplegic|mothers against dpp|phospho-Smad|phosphoSmad|phosphorylated smad', map_location='2-7.7 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3146056, end_position_on_the_genomic_accession=3159643, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33530, CurrentID=0, Status='live', Symbol='gkt', Aliases='Dmel_CG8825, CG8825, CG8826, Dmel\\CG8825, Gkt, TDP1, Tdp1, anon-23Db, c22', description='glaikit', other_designations='CG8825-PA|Tyrosyl-DNA phosophodiesterase 1|anon-23Db|gkt-PA', map_location='23D4-23D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3159950, end_position_on_the_genomic_accession=3162399, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33532, CurrentID=0, Status='live', Symbol='Hydr2', Aliases='Dmel_CG3488, C24, CG3488, CT11757, DMU29170, Dmel\\CG3488, HYDR2, anon-23D, anon-23Da, c24, cg3488', description='alpha/beta hydrolase2', other_designations='CG3488-PA|CG3488-PB|CG3488-PC|Hydr2-PA|Hydr2-PB|Hydr2-PC', map_location='23D4-23D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3163645, end_position_on_the_genomic_accession=3169462, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33534, CurrentID=0, Status='live', Symbol='CG34393', Aliases='Dmel_ CG15405, CG8829, Dm_2L:14421, Dmel\\CG34393, Dmel_CG15405, Dmel_CG8829', description='CG34393 gene product from transcript CG34393-RA', other_designations='CG34393-PA|CG34393-PB|uncharacterized protein', map_location='23D5-23D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3175112, end_position_on_the_genomic_accession=3200496, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33535, CurrentID=0, Status='live', Symbol='CG44002', Aliases='Dmel_ CG18555, CG34406, CG34406-2, Dmel\\CG44002', description='CG44002 gene product from transcript CG44002-RA', other_designations='CG44002-PA|uncharacterized protein', map_location='23D4-23D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3169820, end_position_on_the_genomic_accession=3171092, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33537, CurrentID=0, Status='live', Symbol='CG15404', Aliases='Dmel_ Dmel\\CG15404', description='CG15404 gene product from transcript CG15404-RB', other_designations='CG15404-PA|CG15404-PB|uncharacterized protein', map_location='23D4-23D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3173972, end_position_on_the_genomic_accession=3174504, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33538, CurrentID=0, Status='live', Symbol='CG3347', Aliases='Dmel_ Dmel\\CG3347', description='CG3347 gene product from transcript CG3347-RB', other_designations='CG3347-PB|CG3347-PC', map_location='23E1-23E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3269437, end_position_on_the_genomic_accession=3271806, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33539, CurrentID=0, Status='live', Symbol='CG3332', Aliases='Dmel_ Dmel\\CG3332', description='CG3332 gene product from transcript CG3332-RB', other_designations='CG3332-PA|CG3332-PB|CG3332-PC|CG3332-PD', map_location='23E1-23E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3275800, end_position_on_the_genomic_accession=3280534, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33540, CurrentID=0, Status='live', Symbol='CG9664', Aliases='Dmel_ Dmel\\CG9664', description='CG9664 gene product from transcript CG9664-RA', other_designations='CG9664-PA|CG9664-PB|CG9664-PC|CG9664-PD', map_location='23E1-23E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3291245, end_position_on_the_genomic_accession=3294706, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33541, CurrentID=0, Status='live', Symbol='CG9663', Aliases='Dmel_ Dmel\\CG9663', description='CG9663 gene product from transcript CG9663-RA', other_designations='CG9663-PA', map_location='23E1-23E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3294926, end_position_on_the_genomic_accession=3303130, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33542, CurrentID=0, Status='live', Symbol='CG15406', Aliases='Dmel_ Dmel\\CG15406', description='CG15406 gene product from transcript CG15406-RA', other_designations='CG15406-PA|uncharacterized protein', map_location='23E3-23E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3309222, end_position_on_the_genomic_accession=3311682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33543, CurrentID=0, Status='live', Symbol='CG3277', Aliases='Dmel_ Dmel\\CG3277', description='CG3277 gene product from transcript CG3277-RC', other_designations='CG3277-PC', map_location='23E3-23E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3304100, end_position_on_the_genomic_accession=3317418, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  33544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33544, CurrentID=0, Status='live', Symbol='CG3326', Aliases='Dmel_ DmCG3326, Dmel\\CG3326', description='CG3326 gene product from transcript CG3326-RB', other_designations='CG3326-PA|CG3326-PB|Dm-fidgetin|fidgetin_DROME', map_location='23E3-23E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3316794, end_position_on_the_genomic_accession=3319872, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33545, CurrentID=0, Status='live', Symbol='CG8837', Aliases='Dmel_ Dmel\\CG8837', description='CG8837 gene product from transcript CG8837-RB', other_designations='CG8837-PA|CG8837-PB', map_location='23E3-23E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3320245, end_position_on_the_genomic_accession=3323141, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33547, CurrentID=0, Status='live', Symbol='CG3285', Aliases='Dmel_ Dmel\\CG3285', description='CG3285 gene product from transcript CG3285-RA', other_designations='CG3285-PA|CG3285-PB', map_location='23E4-23E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3330928, end_position_on_the_genomic_accession=3332876, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33548, CurrentID=0, Status='live', Symbol='CG15408', Aliases='Dmel_ Dmel\\CG15408', description='CG15408 gene product from transcript CG15408-RA', other_designations='CG15408-PA|uncharacterized protein', map_location='23E4-23E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3332814, end_position_on_the_genomic_accession=3334790, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33551, CurrentID=0, Status='live', Symbol='CG8838', Aliases='Dmel_ Dmel\\CG8838, anon-WO0140519.135', description='CG8838 gene product from transcript CG8838-RA', other_designations='CG8838-PA', map_location='23E5-23E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3357646, end_position_on_the_genomic_accession=3358618, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33552, CurrentID=0, Status='live', Symbol='CG34394', Aliases='Dmel_ BcDNA:RE58243, CG15411, CG3304, Dmel\\CG34394, Dmel_CG15411, Dmel_CG3304', description='CG34394 gene product from transcript CG34394-RH', other_designations='CG34394-PA|CG34394-PC|CG34394-PD|CG34394-PE|CG34394-PG|CG34394-PH|uncharacterized protein', map_location='23E5-23E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3358640, end_position_on_the_genomic_accession=3373372, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33553, CurrentID=0, Status='live', Symbol='CG15412', Aliases='Dmel_ Dmel\\CG15412', description='CG15412 gene product from transcript CG15412-RB', other_designations='CG15412-PB|CG15412-PC|uncharacterized protein', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3373567, end_position_on_the_genomic_accession=3375711, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33554, CurrentID=0, Status='live', Symbol='CG9662', Aliases='Dmel_ BcDNA:AT03617, Dmel\\CG9662', description='CG9662 gene product from transcript CG9662-RB', other_designations='CG9662-PA|CG9662-PB', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3375832, end_position_on_the_genomic_accession=3376737, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33555, CurrentID=0, Status='live', Symbol='Ptpa', Aliases='Dmel_CG3289, CG3289, DPTPA, Dmel\\CG3289, PTPA', description='Phosphotyrosyl phosphatase activator', other_designations='CG3289-PA|Ptpa-PA|phosphotyrosyl phosphatase activator', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3377002, end_position_on_the_genomic_accession=3378590, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33556, CurrentID=0, Status='live', Symbol='pgant2', Aliases='Dmel_CG3254, CG3254, Dmel\\CG3254, PGANT2', description='polypeptide GalNAc transferase 2', other_designations='CG3254-PA|CG3254-PB|pgant2-PA|pgant2-PB', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3404460, end_position_on_the_genomic_accession=3426719, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33557, CurrentID=0, Status='live', Symbol='CG8840', Aliases='Dmel_ Dmel\\CG8840', description='CG8840 gene product from transcript CG8840-RA', other_designations='CG8840-PA', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3410560, end_position_on_the_genomic_accession=3411199, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33559, CurrentID=0, Status='live', Symbol='CG2772', Aliases='Dmel_ BcDNA:RE45077, Dmel\\CG2772', description='CG2772 gene product from transcript CG2772-RA', other_designations='CG2772-PA', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3445992, end_position_on_the_genomic_accession=3447589, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33560, CurrentID=0, Status='live', Symbol='Snx1', Aliases='Dmel_CG2774, CG2774, Dmel\\CG2774, Dsnx1, Vps5', description='Sorting nexin 1', other_designations='CG2774-PA|CG2774-PB|Snx1-PA|Snx1-PB|sorting nexin 1', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3449683, end_position_on_the_genomic_accession=3452106, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33561, CurrentID=0, Status='live', Symbol='CG12795', Aliases='Dmel_ Dmel\\CG12795', description='CG12795 gene product from transcript CG12795-RB', other_designations='CG12795-PB|uncharacterized protein', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3452466, end_position_on_the_genomic_accession=3453755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33562, CurrentID=0, Status='live', Symbol='Cog3', Aliases='Dmel_CG3248, CG 3248, CG3248, COG3, Dmel\\CG3248, SEC34, dCOG3', description='CG3248 gene product from transcript CG3248-RA', other_designations='CG3248-PA|Cog3-PA', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3453697, end_position_on_the_genomic_accession=3457064, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33563, CurrentID=0, Status='live', Symbol='Sec5', Aliases='Dmel_CG8843, CG8843, Dmel\\CG8843, SEC5, Sec-5p, dsec5, sec5, Sec5', description='Sec5 ortholog (S. cerevisiae)', other_designations='CG8843-PA|Sec5 ortholog|Sec5-PA|sec5', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3457302, end_position_on_the_genomic_accession=3460367, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33564, CurrentID=0, Status='live', Symbol='CG3246', Aliases='Dmel_ Dmel\\CG3246', description='CG3246 gene product from transcript CG3246-RA', other_designations='CG3246-PA', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3460322, end_position_on_the_genomic_accession=3461952, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33565, CurrentID=0, Status='live', Symbol='msl-2', Aliases='Dmel_CG3241, CG3241, Dmel\\CG3241, MSL, MSL-2, MSL2, Msl 1-3, Msl-2, Msl2, km(2)A, kmA, msl, msl2', description='male-specific lethal 2', other_designations='CG3241-PA|CG3241-PB|killer of males|killer-of-males-A|male specifc lethal|male specific lethal 2|male specific-lethal 2|male-specific-lethal 2|male-specific-lethal-2|msl-2-PA|msl-2-PB', map_location='2-9.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3462218, end_position_on_the_genomic_accession=3466112, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33567, CurrentID=0, Status='live', Symbol='CG3238', Aliases='Dmel_ Dmel\\CG3238', description='CG3238 gene product from transcript CG3238-RB', other_designations='CG3238-PA|CG3238-PB', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3467342, end_position_on_the_genomic_accession=3470004, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33568, CurrentID=0, Status='live', Symbol='CG31776', Aliases='Dmel_ CG8845, CG8845a, Dmel\\CG31776, dppGalNAcT10', description='CG31776 gene product from transcript CG31776-RA', other_designations='CG31776-PA|uncharacterized protein', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3470182, end_position_on_the_genomic_accession=3472364, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33569, CurrentID=0, Status='live', Symbol='Thor', Aliases='Dmel_CG8846, 153432_at, 43-BP, 4E-BP, 4E-BP1, 4EBP, 4e-BP, 4e-bp, 4eBP, 4ebp, BcDNA.HL08053, BcDNA:HL08053, CG8846, CT8705, Dm 4E-BP, Dmel\\CG8846, Phas1/EIF4EBP, anon-WO0172774.158, d4E-BP, d4E-BP1, d4EBP, d4e-bp, d4eBP, dthor, eIF4E-BP, l(2)06270, l(2)k07736, l(2)k13517, pp20, thor', description='CG8846 gene product from transcript CG8846-RA', other_designations='4E binding protein|CG8846-PA|CG8846-PB|Eif4e-binding protein|Thor-PA|Thor-PB|eIF-4E-binding protein|eIF4E binding protein|eif4e-binding protein 4EBP|eukaryotic initiation factor 4E binding protein|eukaryotic initiation factor 4E-binding protein|eukaryotic translation initiation factor 4E binding protein|eukaryotic translation initiation factor 4E-binding protein|insulin-stimulated eIF-4E binding protein|lethal (2) 06270|thor|translation initiation factor 4E binding protein', map_location='23F3-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3478434, end_position_on_the_genomic_accession=3479612, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33570, CurrentID=0, Status='live', Symbol='CG15414', Aliases='Dmel_ Dmel\\CG15414', description='CG15414 gene product from transcript CG15414-RA', other_designations='CG15414-PA|uncharacterized protein', map_location='23F3-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3480755, end_position_on_the_genomic_accession=3491071, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33571, CurrentID=0, Status='live', Symbol='tim', Aliases='Dmel_CG3234, CG3234, Dmel\\CG3234, Ritsu, TIM, Tim, dTIM, dTim, dtim, mel_tim, rit1, tim', description='timeless', other_designations='CG3234-PB|CG3234-PL|CG3234-PM|CG3234-PN|CG3234-PO|CG3234-PP|CG3234-PR|CG3234-PS|TIMELESS|dtimeless|ritsu|tim-PB|tim-PL|tim-PM|tim-PN|tim-PO|tim-PP|tim-PR|tim-PS|timeless1', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3493986, end_position_on_the_genomic_accession=3508119, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  33572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33572, CurrentID=0, Status='live', Symbol='CG31954', Aliases='Dmel_ CG3229, Dmel\\CG31954, SP107', description='CG31954 gene product from transcript CG31954-RA', other_designations='CG31954-PA|uncharacterized protein', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3509058, end_position_on_the_genomic_accession=3510102, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33573, CurrentID=0, Status='live', Symbol='CG17593', Aliases='Dmel_ Dmel\\CG17593', description='CG17593 gene product from transcript CG17593-RB', other_designations='CG17593-PA|CG17593-PB|uncharacterized protein', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3514872, end_position_on_the_genomic_accession=3517264, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33574, CurrentID=0, Status='live', Symbol='CG3213', Aliases='Dmel_ Dmel\\CG3213, bs30c03.y1', description='CG3213 gene product from transcript CG3213-RA', other_designations='CG3213-PA', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3517273, end_position_on_the_genomic_accession=3519651, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740462, CurrentID=0, Status='live', Symbol='CG34398', Aliases='Dmel_ CG15867, CG9487, Dmel\\CG34398, Dmel_CG15867, Dmel_CG9487', description='CG34398 gene product from transcript CG34398-RE', other_designations='CG34398-PA|CG34398-PC|CG34398-PE|CG34398-PF|CG34398-PG|CG34398-PH|CG34398-PI|CG34398-PJ|CG34398-PK|CG34398-PL|uncharacterized protein', map_location='29F3-29F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8898786, end_position_on_the_genomic_accession=8933594, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  33576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33576, CurrentID=0, Status='live', Symbol='Sr-CIV', Aliases='Dmel_CG3212, CG3212, CT10789, Dmel\\CG3212, SR-CIV', description='Scavenger receptor class C, type IV', other_designations='CG3212-PA|Sr-CIV-PA|scavenger receptor class C, type IV', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3522521, end_position_on_the_genomic_accession=3524056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33577, CurrentID=0, Status='live', Symbol='CG8852', Aliases='Dmel_ CT9284, Dmel\\CG8852', description='CG8852 gene product from transcript CG8852-RA', other_designations='CG8852-PA', map_location='23F6-24A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3525131, end_position_on_the_genomic_accession=3527588, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33578, CurrentID=0, Status='live', Symbol='Spindly', Aliases='Dmel_CG15415, CG15415, Dmel\\CG15415', description='CG15415 gene product from transcript CG15415-RB', other_designations='CG15415-PB|Spindly-PB|spindly', map_location='24A1-24A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3527387, end_position_on_the_genomic_accession=3530317, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33579, CurrentID=0, Status='live', Symbol='Che-13', Aliases='Dmel_CG8853, CG8853, CHE-13, Dmel\\CG8853, Hippi, IFT55, IFT57, che-13', description='Che-13 ortholog (C. elegans)', other_designations='CG8853-PA|Che-13 ortholog|Che-13-PA|hippi', map_location='24A1-24A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3530789, end_position_on_the_genomic_accession=3532678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33581, CurrentID=0, Status='live', Symbol='sob', Aliases='Dmel_CG3242, CG3242, Dmel\\CG3242, Sob, org1', description='sister of odd and bowl', other_designations='CG3242-PA|odd skipped-related 1|sob-PA', map_location='24A1-24A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3578013, end_position_on_the_genomic_accession=3580769, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33583, CurrentID=0, Status='live', Symbol='odd', Aliases='Dmel_CG3851, CG3851, Dmel\\CG3851, ODD, Odd, l(2)01863, ods', description='odd skipped', other_designations='CG3851-PA|odd paired|odd-PA|odd-skipped|oddskipped', map_location='2-8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3604224, end_position_on_the_genomic_accession=3606756, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33585, CurrentID=0, Status='live', Symbol='CG15418', Aliases='Dmel_ Dmel\\CG15418', description='CG15418 gene product from transcript CG15418-RA', other_designations='CG15418-PA|uncharacterized protein', map_location='24A2-24A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3621110, end_position_on_the_genomic_accession=3621681, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33589, CurrentID=0, Status='live', Symbol='CG16704', Aliases='Dmel_ BcDNA:RH31195, CT37177, Dmel\\CG16704', description='CG16704 gene product from transcript CG16704-RB', other_designations='CG16704-PA|CG16704-PB|uncharacterized protein', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3692123, end_position_on_the_genomic_accession=3692587, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33590, CurrentID=0, Status='live', Symbol='CG3513', Aliases='Dmel_ Dmel\\CG3513', description='CG3513 gene product from transcript CG3513-RA', other_designations='CG3513-PA', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3692699, end_position_on_the_genomic_accession=3693182, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33591, CurrentID=0, Status='live', Symbol='CG16713', Aliases='Dmel_ CT37195, Dmel\\CG16713', description='CG16713 gene product from transcript CG16713-RA', other_designations='CG16713-PA|uncharacterized protein', map_location='24B2-24B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3695032, end_position_on_the_genomic_accession=3695411, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33592, CurrentID=0, Status='live', Symbol='IM33', Aliases='Dmel_CG16712, BcDNA:RH05411, BcDNA:RH38008, CG16712, CT37183, DIM 33, Dmel\\CG16712', description='Immune induced molecule 33', other_designations='CG16712-PA|CG16712-PB|IM33-PA|IM33-PB|immune induced molecule 33', map_location='24B3-24B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3696218, end_position_on_the_genomic_accession=3696658, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33593, CurrentID=0, Status='live', Symbol='CG3604', Aliases='Dmel_ BcDNA:LP04037, Dmel\\CG3604', description='CG3604 gene product from transcript CG3604-RB', other_designations='CG3604-PA|CG3604-PB', map_location='24B3-24B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3697242, end_position_on_the_genomic_accession=3697872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33594, CurrentID=0, Status='live', Symbol='CG10031', Aliases='Dmel_ Dmel\\CG10031', description='CG10031 gene product from transcript CG10031-RA', other_designations='CG10031-PA|uncharacterized protein', map_location='24B3-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3698451, end_position_on_the_genomic_accession=3699092, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33595, CurrentID=0, Status='live', Symbol='CG2816', Aliases='Dmel_ Dmel\\CG2816', description='CG2816 gene product from transcript CG2816-RB', other_designations='CG2816-PB', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3703659, end_position_on_the_genomic_accession=3704253, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33597, CurrentID=0, Status='live', Symbol='CG2818', Aliases='Dmel_ Dmel\\CG2818', description='CG2818 gene product from transcript CG2818-RA', other_designations='CG2818-PA|CG2818-PB', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3713360, end_position_on_the_genomic_accession=3716810, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33599, CurrentID=0, Status='live', Symbol='Shaw', Aliases='Dmel_CG2822, CG15419, CG2822, Dmel\\CG2822, Kv3, Kv3.12, dShaw, dShaw2, shaw, Shaw', description='Shaker cognate w', other_designations='CG2822-PA|CG2822-PB|CG2822-PC|CG2822-PD|CG2822-PE|Shaw-PA|Shaw-PB|Shaw-PC|Shaw-PD|Shaw-PE|shaker cognate w', map_location='24C1-24C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3718185, end_position_on_the_genomic_accession=3729145, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33600, CurrentID=0, Status='live', Symbol='CG10019', Aliases='Dmel_ BEST:GH25970, Dmel\\CG10019, GH25970', description='CG10019 gene product from transcript CG10019-RC', other_designations='CG10019-PB|CG10019-PC|uncharacterized protein', map_location='24C2-24C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3730466, end_position_on_the_genomic_accession=3754617, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  33601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33601, CurrentID=0, Status='live', Symbol='CG43707', Aliases='Dmel_ CG10020, CG31772, CG31959, Dmel\\CG43707, Dmel_CG31772, Dmel_CG31959', description='CG43707 gene product from transcript CG43707-RF', other_designations='CG43707-PC|CG43707-PD|CG43707-PE|CG43707-PF|CG43707-PG|CG43707-PH|CG43707-PI|uncharacterized protein', map_location='24C3-24C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3757256, end_position_on_the_genomic_accession=3769525, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  33602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33602, CurrentID=0, Status='live', Symbol='bowl', Aliases='Dmel_CG10021, 17-29-5, Bowl, CG10021, Dmel\\CG10021, Su(tor)2-1, l(2)c, org2', description='brother of odd with entrails limited', other_designations='CG10021-PA|CG10021-PB|CG10021-PC|CG10021-PD|CG10021-PF|CG10021-PG|CG10021-PH|bowel|bowl-PA|bowl-PB|bowl-PC|bowl-PD|bowl-PF|bowl-PG|bowl-PH|odd skipped-related 2', map_location='2-16 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3771700, end_position_on_the_genomic_accession=3784121, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33603, CurrentID=0, Status='live', Symbol='CG31958', Aliases='Dmel_ CG10022, Dmel\\CG31958', description='CG31958 gene product from transcript CG31958-RA', other_designations='CG31958-PA|uncharacterized protein', map_location='24C4-24C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3785738, end_position_on_the_genomic_accession=3786343, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33604, CurrentID=0, Status='live', Symbol='bark', Aliases='Dmel_CG3921, CG3921, Dmel\\CG3921, aka', description='bark beetle', other_designations='CG3921-PA|CG3921-PB|CG3921-PC|anakonda|bark-PA|bark-PB|bark-PC', map_location='24C5-24C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3787167, end_position_on_the_genomic_accession=3802726, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  33606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33606, CurrentID=0, Status='live', Symbol='CG3407', Aliases='Dmel_ Dmel\\CG3407, cg3407', description='CG3407 gene product from transcript CG3407-RA', other_designations='CG3407-PA', map_location='24C5-24C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3810801, end_position_on_the_genomic_accession=3813410, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33607, CurrentID=0, Status='live', Symbol='slp1', Aliases='Dmel_CG16738, CG16738, Dmel\\CG16738, Dmslp1, FD6, FoxG, SLP, SLP1, Slp, Slp-1, Slp1, l(2)05965, sko1, slp, slp-1, slp-2', description='sloppy paired 1', other_designations='CG16738-PA|Sloppy-paired|Sloppy-paired1|sloppy|sloppy pair|sloppy paired|sloppy-paired 1|sloppy-paired-1|slp1-PA', map_location='2-8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3825675, end_position_on_the_genomic_accession=3827099, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33608, CurrentID=0, Status='live', Symbol='slp2', Aliases='Dmel_CG2939, CG2939, Dmel\\CG2939, Dmslp2, FD7, FoxG, SLP, Slp, Slp2, slp', description='sloppy paired 2', other_designations='CG2939-PA|Sloppy-paired|sloppy pair|sloppy paired|sloppy-paired 2|slp2-PA', map_location='2-8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3836840, end_position_on_the_genomic_accession=3839185, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33609, CurrentID=0, Status='live', Symbol='CG3964', Aliases='Dmel_ Dmel\\CG3964', description='CG3964 gene product from transcript CG3964-RB', other_designations='CG3964-PA|CG3964-PB|CG3964-PC', map_location='24C8-24C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3862668, end_position_on_the_genomic_accession=3867507, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33610, CurrentID=0, Status='live', Symbol='Cep97', Aliases='Dmel_CG3980, CG3980, CT13231, D Dmel\\CG3980', description='CG3980 gene product from transcript CG3980-RB', other_designations='CG3980-PB|CG3980-PD|Cep97-PB|Cep97-PD', map_location='24C8-24C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3867697, end_position_on_the_genomic_accession=3871425, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33611, CurrentID=0, Status='live', Symbol='capu', Aliases='Dmel_CG3399, CG15420, CG3399, Capu, Dmel\\CG3399, capp', description='cappuccino', other_designations='CG3399-PA|CG3399-PB|CG3399-PD|CG3399-PE|CG3399-PF|CG3399-PG|CG3399-PH|CG3399-PI|CG3399-PJ|cappucino|capu-PA|capu-PB|capu-PD|capu-PE|capu-PF|capu-PG|capu-PH|capu-PI|capu-PJ', map_location='2-8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3872646, end_position_on_the_genomic_accession=3902860, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  33613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33613, CurrentID=0, Status='live', Symbol='fred', Aliases='Dmel_CG31774, CG12675, CG31774, CG3390, CG3393, CT11413, Dmel\\CG31774, Fred', description='friend of echinoid', other_designations='CG31774-PF|CG31774-PG|CG31774-PH|fred-PF|fred-PG|fred-PH', map_location='24C9-24D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3902928, end_position_on_the_genomic_accession=3991713, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33615, CurrentID=0, Status='live', Symbol='CG15422', Aliases='Dmel_ Dmel\\CG15422', description='CG15422 gene product from transcript CG15422-RA', other_designations='CG15422-PA|uncharacterized protein', map_location='24D3-24D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4004984, end_position_on_the_genomic_accession=4005670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33616, CurrentID=0, Status='live', Symbol='CG15423', Aliases='Dmel_ Dmel\\CG15423', description='CG15423 gene product from transcript CG15423-RA', other_designations='CG15423-PA|uncharacterized protein', map_location='24D3-24D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4006041, end_position_on_the_genomic_accession=4006606, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33617, CurrentID=0, Status='live', Symbol='CG43773', Aliases='Dmel_ CG10039, Dmel\\CG43773', description='CG43773 gene product from transcript CG43773-RA', other_designations='CG43773-PA|uncharacterized protein', map_location='24D3-24D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4007690, end_position_on_the_genomic_accession=4007984, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33619, CurrentID=0, Status='live', Symbol='ed', Aliases='Dmel_CG12676, 1X5, CG12676, CG15424, CG16842, CT13476, CT35484, Dmel\\CG12676, E(Elp)24D, E(Elp)[24D], Ed, l(2)k01102', description='echinoid', other_designations='CG12676-PA|CG12676-PB|CG12676-PC|ed-PA|ed-PB|ed-PC', map_location='2-11.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4031377, end_position_on_the_genomic_accession=4115749, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33621, CurrentID=0, Status='live', Symbol='Sr-CI', Aliases='Dmel_CG4099, CG4099, CT9876, D-SR-CI, DSR-CI, Dmel\\CG4099, SR-C1, SR-CI, Sr-C1, Sr-Cl, dSR-C1, dSR-CI, dSRC-I, dScR-I', description='Scavenger receptor class C, type I', other_designations='CG4099-PA|CG4099-PB|Sr-CI-PA|Sr-CI-PB|scavenger receptor|scavenger receptor C, type I|scavenger receptor class C, type 1|scavenger receptor class C, type I', map_location='24D6-24D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4121702, end_position_on_the_genomic_accession=4124153, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33622, CurrentID=0, Status='live', Symbol='CG2955', Aliases='Dmel_ Dmel\\CG2955, anon-WO0140519.175', description='CG2955 gene product from transcript CG2955-RA', other_designations='CG2955-PA', map_location='24D6-24D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4149279, end_position_on_the_genomic_accession=4151203, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33623, CurrentID=0, Status='live', Symbol='Or24a', Aliases='Dmel_CG11767, 24D.1, 24a, CG11767, DOR24D.1, DOR48, Dmel\\CG11767, Or24D.1', description='Odorant receptor 24a', other_designations='CG11767-PB|Or24a-PB|odorant receptor 24a|olfactory receptor 24D.1', map_location='24D7-24D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4164689, end_position_on_the_genomic_accession=4166134, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33626, CurrentID=0, Status='live', Symbol='CG3714', Aliases='Dmel_ AAF51037, Dmel\\CG3714', description='CG3714 gene product from transcript CG3714-RG', other_designations='CG3714-PA|CG3714-PB|CG3714-PC|CG3714-PD|CG3714-PE|CG3714-PF|CG3714-PG|CG3714-PH', map_location='24D8-24D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4189340, end_position_on_the_genomic_accession=4198071, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  33627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33627, CurrentID=0, Status='live', Symbol='ft', Aliases='Dmel_CG3352, 79/18, CG3352, CT11259, Dm Fat, Dmel\\CG3352, Fat, Ft, l(2)24Da, l(2)79/18, l(2)fat, l(2)fd, l(2)ft, l(2)gd-1, l(2)gd-l, l(2)gd2, l(2)k07918', description='fat', other_designations='CG3352-PA|floppy-discs|ft-PA|giant discs 2|giant-discs-1', map_location='2-12.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4198404, end_position_on_the_genomic_accession=4221796, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33628, CurrentID=0, Status='live', Symbol='CG3702', Aliases='Dmel_ Dmel\\CG3702, cg3702', description='CG3702 gene product from transcript CG3702-RA', other_designations='CG3702-PA|CG3702-PB', map_location='24D8-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4224650, end_position_on_the_genomic_accession=4227097, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33629, CurrentID=0, Status='live', Symbol='RpL40', Aliases='Dmel_CG2960, BcDNA:RE10554, CEP52, CG2960, DUb52, Dmel\\CG2960, Dub52, Rpl40, Ub52, Ubi-f, Ubi-f52, Ubiq', description='Ribosomal protein L40', other_designations='CG2960-PA|CG2960-PB|RpL40-PA|RpL40-PB|ribosomal protein L40|ubiquitin|ubiquitin 52-AA extension protein|ubiquitin fusion 52', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4227699, end_position_on_the_genomic_accession=4228561, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33630, CurrentID=0, Status='live', Symbol='CG15425', Aliases='Dmel_ Dmel\\CG15425, anon-SAGE:Wang-056', description='CG15425 gene product from transcript CG15425-RA', other_designations='CG15425-PA|anon-SAGE:Wang-056|uncharacterized protein', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4257572, end_position_on_the_genomic_accession=4258425, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33631, CurrentID=0, Status='live', Symbol='Art2', Aliases='Dmel_CG3675, CG3675, DART2, Dmel\\CG3675', description='Arginine methyltransferase 2', other_designations='Art2-PA|Art2-PB|CG3675-PA|CG3675-PB|arginine methyltransferase 2', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4277696, end_position_on_the_genomic_accession=4281324, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33635, CurrentID=0, Status='live', Symbol='bdl', Aliases='Dmel_CG16857, BcDNA:GH11322, CG16857, CT35491, Dmel\\CG16857, GH11322', description='borderless', other_designations='CG16857-PA|bdl-PA', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4324554, end_position_on_the_genomic_accession=4331830, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33636, CurrentID=0, Status='live', Symbol='Atet', Aliases='Dmel_CG2969, CG2969, Dmel\\CG2969', description='ABC transporter expressed in trachea', other_designations='Atet-PA|Atet-PB|Atet-PC|Atet-PD|CG2969-PA|CG2969-PB|CG2969-PC|CG2969-PD', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4333913, end_position_on_the_genomic_accession=4345776, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33637, CurrentID=0, Status='live', Symbol='CG15429', Aliases='Dmel_ BcDNA:AT23465, Dmel\\CG15429', description='CG15429 gene product from transcript CG15429-RA', other_designations='CG15429-PA|CG15429-PB|uncharacterized protein', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4352975, end_position_on_the_genomic_accession=4354034, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33638, CurrentID=0, Status='live', Symbol='Traf4', Aliases='Dmel_CG3048, CG3048, DTRAF, DTRAF1, Dmel\\CG3048, TRAF, TRAF1, TRAF4, Traf1, dTRAF, dTRAF1, dTraf1, dTraf4, dtraf1, traf-4, traf1, traf4', description='TNF-receptor-associated factor 4', other_designations='CG3048-PA|CG3048-PE|TNF receptor associated factor 1|TNF-receptor-associated factor 1|Traf4-PA|Traf4-PE', map_location='24E1-24E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4361925, end_position_on_the_genomic_accession=4380716, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33639, CurrentID=0, Status='live', Symbol='CG17612', Aliases='Dmel_ Dmel\\CG17612', description='CG17612 gene product from transcript CG17612-RB', other_designations='CG17612-PA|CG17612-PB|uncharacterized protein', map_location='24E4-24E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4380722, end_position_on_the_genomic_accession=4382735, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33640, CurrentID=0, Status='live', Symbol='Vps53', Aliases='Dmel_CG3338, CG3338, Dmel\\CG3338', description='Vacuolar protein sorting 53', other_designations='CG3338-PA|Vps53-PA|vacuolar protein sorting 53', map_location='24E4-24E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4382925, end_position_on_the_genomic_accession=4386062, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33641, CurrentID=0, Status='live', Symbol='Psf2', Aliases='Dmel_CG18013, CG18013, Dmel\\CG18013, P{lacW}l(2)SH0805[SH0805], l(2)SH0805, l(2)SH2 0805', description='CG18013 gene product from transcript CG18013-RC', other_designations='CG18013-PB|CG18013-PC|Psf2-PB|Psf2-PC|lethal (2) SH0805', map_location='24F1-24F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4386182, end_position_on_the_genomic_accession=4387356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33642, CurrentID=0, Status='live', Symbol='Tps1', Aliases='Dmel_CG4104, 19920676, BcDNA:GH08860, CG4104, CG4104 PA, Dmel\\CG4104, TPS, TreS, dtps1, jf5, l(2)24Ea, l(2)jf4, l(2)jf5, l(2)k08903, tps1', description='Trehalose-6-phosphate synthase 1', other_designations='CG4104-PA|Tps1-PA|Trehalos-6-phosphate synthase|lethal(2)24Ea|trehalose synthase|trehalose-6-phosphate synthase|trehalose-phosphate synthase', map_location='24E4-24E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4388327, end_position_on_the_genomic_accession=4392778, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33643, CurrentID=0, Status='live', Symbol='CG3652', Aliases='Dmel_ Dmel\\CG3652', description='CG3652 gene product from transcript CG3652-RA', other_designations='CG3652-PA', map_location='24E5-24E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4392770, end_position_on_the_genomic_accession=4393876, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33644, CurrentID=0, Status='live', Symbol='l(2)k05819', Aliases='Dmel_CG3054, CG3054, Dmel\\CG3054, anon-WO0140519.156', description='lethal (2) k05819', other_designations='CG3054-PA|CG3054-PC|CG3054-PD|l(2)k05819-PA|l(2)k05819-PC|l(2)k05819-PD', map_location='24E5-24F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4394196, end_position_on_the_genomic_accession=4400487, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33645, CurrentID=0, Status='live', Symbol='Dim1', Aliases='Dmel_CG3058, BcDNA:RE13747, CG3058, Dmel\\CG3058, U5-15kD', description='CG3058 gene product from transcript CG3058-RA', other_designations='CG3058-PA|Dim1-PA', map_location='24F1-24F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4403251, end_position_on_the_genomic_accession=4403955, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33646, CurrentID=0, Status='live', Symbol='CG15431', Aliases='Dmel_ Dmel\\CG15431', description='CG15431 gene product from transcript CG15431-RA', other_designations='CG15431-PA|CG15431-PB|uncharacterized protein', map_location='24F1-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4403992, end_position_on_the_genomic_accession=4441282, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33647, CurrentID=0, Status='live', Symbol='CG15432', Aliases='Dmel_ Dmel\\CG15432, cg15432', description='CG15432 gene product from transcript CG15432-RC', other_designations='CG15432-PA|CG15432-PB|CG15432-PC|uncharacterized protein', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4442320, end_position_on_the_genomic_accession=4442985, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33648, CurrentID=0, Status='live', Symbol='morgue', Aliases='Dmel_CG15437, BcDNA:GH02435, CG15437, Dmel\\CG15437, FBXO18, Morgue', description='modifier of rpr and grim, ubiquitously expressed', other_designations='CG15437-PA|morgue-PA', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4442870, end_position_on_the_genomic_accession=4444680, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33649, CurrentID=0, Status='live', Symbol='Elp3', Aliases='Dmel_CG15433, CG15433, D-elp3, Dmel/ELP3, Dmel\\CG15433, Dmel\\ELP3, ELP3, HAT, dmHAG408, elp-3, elp3', description='Elongator complex protein 3', other_designations='CG15433-PA|Elp3-PA|elongator complex protein 3|elongator protein 3', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4444732, end_position_on_the_genomic_accession=4446765, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33650, CurrentID=0, Status='live', Symbol='MFS18', Aliases='Dmel_CG15438, BEST:CK02623, CG15438, CK02623, Dmel\\CG15438', description='Major Facilitator Superfamily Transporter 18', other_designations='CG15438-PA|MFS18-PA|major facilitator superfamily transporter 18', map_location='24F2-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4446670, end_position_on_the_genomic_accession=4448635, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33651, CurrentID=0, Status='live', Symbol='CG15439', Aliases='Dmel_ Dmel\\CG15439', description='CG15439 gene product from transcript CG15439-RA', other_designations='CG15439-PA|uncharacterized protein', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4449295, end_position_on_the_genomic_accession=4453038, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33652, CurrentID=0, Status='live', Symbol='CG15440', Aliases='Dmel_ Dmel\\CG15440, Secp43, cg15440', description='CG15440 gene product from transcript CG15440-RA', other_designations='CG15440-PA|uncharacterized protein', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4453169, end_position_on_the_genomic_accession=4454453, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33653, CurrentID=0, Status='live', Symbol='Gs1l', Aliases='Dmel_CG15441, 19864674, CG15441, Dmel\\CG15441, GS1l, gs1-like', description='GS1-like', other_designations='CG15441-PA|CG15441-PB|GS1-like protein|GS1like|Gs1l-PA|Gs1l-PB', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4455137, end_position_on_the_genomic_accession=4456807, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33654, CurrentID=0, Status='live', Symbol='RpL27A', Aliases='Dmel_CG15442, CG15442, Dmel\\CG15442, M, M(2)24F, M(2)24F-1, M(2)Z, M(2)z, M(2L)z, M[[z]], Mz, Rp L27Aa, RpL27Ab, RpL27a, RpL27a2, anon-EST:Posey257, l(2)24EFh, l(2)24Fb, rpL27a, RpL27A', description='Ribosomal protein L27A', other_designations='CG15442-PA|CG15442-PC|CG15442-PD|RpL27A-PA|RpL27A-PC|RpL27A-PD|minute|minute (2) 24F|ribosomal protein L27A|ribosomal protein L27a', map_location='2-12.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4457186, end_position_on_the_genomic_accession=4458364, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33655, CurrentID=0, Status='live', Symbol='CG15435', Aliases='Dmel_ Dmel\\CG15435', description='CG15435 gene product from transcript CG15435-RB', other_designations='CG15435-PA|CG15435-PB|uncharacterized protein', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4459729, end_position_on_the_genomic_accession=4461746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33656, CurrentID=0, Status='live', Symbol='CG15443', Aliases='Dmel_ Dmel\\CG15443', description='CG15443 gene product from transcript CG15443-RB', other_designations='CG15443-PA|CG15443-PB|uncharacterized protein', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4461657, end_position_on_the_genomic_accession=4463447, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33657, CurrentID=0, Status='live', Symbol='CG15436', Aliases='Dmel_ Dmel\\CG15436', description='CG15436 gene product from transcript CG15436-RA', other_designations='CG15436-PA|uncharacterized protein', map_location='24F3-24F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4463611, end_position_on_the_genomic_accession=4465075, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33658, CurrentID=0, Status='live', Symbol='CG17840', Aliases='Dmel_ Dmel\\CG17840', description='CG17840 gene product from transcript CG17840-RA', other_designations='CG17840-PA|uncharacterized protein', map_location='24F4-24F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4465174, end_position_on_the_genomic_accession=4468022, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33659, CurrentID=0, Status='live', Symbol='ine', Aliases='Dmel_CG15444, CG15444, CG15444d, DMINE, Dmel\\CG15444/rosA, rosA, ine', description='inebriated', other_designations='CG15444-PA|CG15444-PB|CG15444-PC|CG15444-PE|CG15444-PF|ine-PA|ine-PB|ine-PC|ine-PE|ine-PF', map_location='2-12 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4468000, end_position_on_the_genomic_accession=4477022, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  33661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33661, CurrentID=0, Status='live', Symbol='HP6', Aliases='Dmel_CG15636, CG15636, Dmel\\CG15636, UMB, UMBREA, Umbrea, umbrea', description='Heterochromatin protein 6', other_designations='CG15636-PA|HP6-PA|heterochromatin protein 6|umbrea', map_location='25A1-25A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4577719, end_position_on_the_genomic_accession=4578218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33665, CurrentID=0, Status='live', Symbol='CG15635', Aliases='Dmel_ Dmel\\CG15635', description='CG15635 gene product from transcript CG15635-RB', other_designations='CG15635-PB|uncharacterized protein', map_location='25A2-25A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4642972, end_position_on_the_genomic_accession=4646562, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33667, CurrentID=0, Status='live', Symbol='CG3355', Aliases='Dmel_ Dmel\\CG3355, SP21', description='CG3355 gene product from transcript CG3355-RB', other_designations='CG3355-PA|CG3355-PB', map_location='25A2-25A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4651403, end_position_on_the_genomic_accession=4652892, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33668, CurrentID=0, Status='live', Symbol='CG11929', Aliases='Dmel_ Dmel\\CG11929', description='CG11929 gene product from transcript CG11929-RA', other_designations='CG11929-PA|CG11929-PB|uncharacterized protein', map_location='25A3-25A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4682677, end_position_on_the_genomic_accession=4687232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33669, CurrentID=0, Status='live', Symbol='Bsg25A', Aliases='Dmel_CG12205, CG12205, Dmel\\CG12205, Elba, Elba1, anon-EST:Liang-1.70, bsg25A, clone 1.70', description='Blastoderm-specific gene 25A', other_designations='Bsg25A-PA|CG12205-PA|early boundary factor', map_location='25A3-25A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4683052, end_position_on_the_genomic_accession=4684461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33670, CurrentID=0, Status='live', Symbol='Elba3', Aliases='Dmel_CG15634, CG15634, Dmel\\CG15634, Elba', description='Early boundary activity 3', other_designations='CG15634-PA|Elba3-PA|early boundary activity 3|early boundary factor', map_location='25A3-25A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4687511, end_position_on_the_genomic_accession=4688956, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33671, CurrentID=0, Status='live', Symbol='CG3251', Aliases='Dmel_ Dmel\\CG3251', description='CG3251 gene product from transcript CG3251-RA', other_designations='CG3251-PA', map_location='25A3-25A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4688965, end_position_on_the_genomic_accession=4690702, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33672, CurrentID=0, Status='live', Symbol='Taf12L', Aliases='Dmel_CG15632, CG15632, Dmel\\CG15632, Rye, TAF, TAF12L, TAF30alpha-2, TAF[[II]], TFIID, Taf30alpha-2, dTAF12, dTAF12L, dTAF[[II]]30a, dTAF[[II]]30alpha-2, dmTAF12L, dmTAF12b, rye', description='TBP-associated factor 30kD subunit alpha-2', other_designations='CG15632-PA|TBP-associated factor|TBP-associated factor subunit alpha-2|Taf12L-PA|ryan express', map_location='25A3-25A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4700835, end_position_on_the_genomic_accession=4701485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33675, CurrentID=0, Status='live', Symbol='CG15631', Aliases='Dmel_ Dmel\\CG15631', description='CG15631 gene product from transcript CG15631-RB', other_designations='CG15631-PA|CG15631-PB|uncharacterized protein', map_location='25A4-25A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4732413, end_position_on_the_genomic_accession=4734458, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33676, CurrentID=0, Status='live', Symbol='CG15630', Aliases='Dmel_ 15630, CG31970, CT35785, Dmel\\CG15630', description='CG15630 gene product from transcript CG15630-RA', other_designations='CG15630-PA|uncharacterized protein', map_location='25A4-25A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4734592, end_position_on_the_genomic_accession=4794044, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33677, CurrentID=0, Status='live', Symbol='CG3294', Aliases='Dmel_ Dmel\\CG3294, cg3294', description='CG3294 gene product from transcript CG3294-RB', other_designations='CG3294-PA|CG3294-PB', map_location='25A6-25A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4794549, end_position_on_the_genomic_accession=4796581, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33678, CurrentID=0, Status='live', Symbol='Clect27', Aliases='Dmel_CG3244, CG3244, Dmel\\CG3244', description='C-type lectin 27kD', other_designations='CG3244-PA|CG3244-PB|Clect27-PA|Clect27-PB', map_location='25A6-25A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4796600, end_position_on_the_genomic_accession=4801923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33679, CurrentID=0, Status='live', Symbol='CG15629', Aliases='Dmel_ Dmel\\CG15629', description='CG15629 gene product from transcript CG15629-RA', other_designations='CG15629-PA|uncharacterized protein', map_location='25A6-25A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4813009, end_position_on_the_genomic_accession=4815891, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33680, CurrentID=0, Status='live', Symbol='CG3225', Aliases='Dmel_ Dmel\\CG3225, cg3225', description='CG3225 gene product from transcript CG3225-RA', other_designations='CG3225-PA', map_location='25A6-25A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4815613, end_position_on_the_genomic_accession=4818245, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33681, CurrentID=0, Status='live', Symbol='CG15628', Aliases='Dmel_ Dmel\\CG15628, anon-WO02059370.23', description='CG15628 gene product from transcript CG15628-RB', other_designations='CG15628-PA|CG15628-PB|uncharacterized protein', map_location='25A7-25A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4821774, end_position_on_the_genomic_accession=4828945, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33682, CurrentID=0, Status='live', Symbol='Fnta', Aliases='Dmel_CG2976, CG2976, Dmel\\CG2976, FNTA', description='Farnesyl transferase alpha', other_designations='CG2976-PA|CG2976-PB|Fnta-PA|Fnta-PB|farnesyl transferase alpha|farnesyl transferase alpha ortholog', map_location='25A8-25A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4829214, end_position_on_the_genomic_accession=4830693, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33683, CurrentID=0, Status='live', Symbol='Ir25a', Aliases='Dmel_CG15627, CG15627, CT35778, DmelIR25a, Dmel\\CG15627, IR25a', description='Ionotropic receptor 25a', other_designations='CG15627-PB|CG15627-PC|CG15627-PD|CG15627-PE|Ir25a-PB|Ir25a-PC|Ir25a-PD|Ir25a-PE|ionotropic receptor 25a', map_location='25A8-25A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4830846, end_position_on_the_genomic_accession=4835300, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33684, CurrentID=0, Status='live', Symbol='tank', Aliases='Dmel_CG15626, CG15626, Dmel\\CG15626', description='CG15626 gene product from transcript CG15626-RB', other_designations='CG15626-PA|CG15626-PB|tank-PA|tank-PB', map_location='25A8-25A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4835335, end_position_on_the_genomic_accession=4837999, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33685, CurrentID=0, Status='live', Symbol='CG12194', Aliases='Dmel_ Dmel\\CG12194, anon-EST:Posey105', description='CG12194 gene product from transcript CG12194-RA', other_designations='CG12194-PA|CG12194-PB|CG12194-PC|uncharacterized protein', map_location='25A8-25A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4838347, end_position_on_the_genomic_accession=4841465, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33686, CurrentID=0, Status='live', Symbol='mxt', Aliases='Dmel_CG2950, BcDNA:GH06479, CG2950, Dmel\\CG2950, cg2950', description='mextli', other_designations='CG2950-PA|CG2950-PB|CG2950-PC|CG2950-PD|CG2950-PE|CG2950-PF|CG2950-PG|mxt-PA|mxt-PB|mxt-PC|mxt-PD|mxt-PE|mxt-PF|mxt-PG', map_location='25A8-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4841482, end_position_on_the_genomic_accession=4847295, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  33687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33687, CurrentID=0, Status='live', Symbol='CG11927', Aliases='Dmel_ Dmel\\CG11927', description='CG11927 gene product from transcript CG11927-RA', other_designations='CG11927-PA|uncharacterized protein', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4847766, end_position_on_the_genomic_accession=4849499, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33688, CurrentID=0, Status='live', Symbol='mRpS2', Aliases='Dmel_CG2937, CG2937, Dmel\\CG2937, MRP-S2', description='mitochondrial ribosomal protein S2', other_designations='CG2937-PA|mRpS2-PA', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4849439, end_position_on_the_genomic_accession=4850419, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33689, CurrentID=0, Status='live', Symbol='Mon1', Aliases='Dmel_CG11926, CG11926, Dmel\\CG11926, MON1, NP_608868', description='Mon1 ortholog', other_designations='CG11926-PA|Mon1-PA', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4850641, end_position_on_the_genomic_accession=4852640, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33690, CurrentID=0, Status='live', Symbol='pog', Aliases='Dmel_CG31660, CG11923, CG11925, CG31660, Dmel\\CG31660, anon-WO0170980.181, anon-WO0170980.182, pg', description='poor gastrulation', other_designations='CG31660-PB|CG31660-PC|CG31660-PD|CG31660-PE|CG31660-PF|pog-PB|pog-PC|pog-PD|pog-PE|pog-PF', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4852864, end_position_on_the_genomic_accession=4876792, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  33692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33692, CurrentID=0, Status='live', Symbol='Cf2', Aliases='Dmel_CG11924, BcDNA:GM09668, CF2, CF2.5, CG11924, Dmel\\CG11924, cf2', description='Chorion factor 2', other_designations='CG11924-PA|CG11924-PB|CG11924-PC|CG11924-PE|CG11924-PF|Cf2-PA|Cf2-PB|Cf2-PC|Cf2-PE|Cf2-PF|chorion Factor-2|chorion factor 2', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4876891, end_position_on_the_genomic_accession=4883107, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33693, CurrentID=0, Status='live', Symbol='CG3008', Aliases='Dmel_ Dmel\\CG3008', description='CG3008 gene product from transcript CG3008-RA', other_designations='CG3008-PA', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4883298, end_position_on_the_genomic_accession=4885669, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33694, CurrentID=0, Status='live', Symbol='CG15625', Aliases='Dmel_ Dmel\\CG15625', description='CG15625 gene product from transcript CG15625-RA', other_designations='CG15625-PA|uncharacterized protein', map_location='25B1-25B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4889995, end_position_on_the_genomic_accession=4891277, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33695, CurrentID=0, Status='live', Symbol='CG3036', Aliases='Dmel_ 151989_at, CT10168, Dmel\\CG3036', description='CG3036 gene product from transcript CG3036-RA', other_designations='CG3036-PA|CG3036-PB', map_location='25B1-25B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4892112, end_position_on_the_genomic_accession=4903889, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33696, CurrentID=0, Status='live', Symbol='CG2837', Aliases='Dmel_ Dmel\\CG2837', description='CG2837 gene product from transcript CG2837-RC', other_designations='CG2837-PA|CG2837-PB|CG2837-PC|CG2837-PD', map_location='25B2-25B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4903865, end_position_on_the_genomic_accession=4906762, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33700, CurrentID=0, Status='live', Symbol='hoe2', Aliases='Dmel_CG15624, CG15624, Dmel\\CG15624', description='hoepel2', other_designations='CG15624-PA|hoe2-PA', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4933047, end_position_on_the_genomic_accession=4936450, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33701, CurrentID=0, Status='live', Symbol='Sgs1', Aliases='Dmel_CG3047, CG3047, Dmel\\CG3047, Muc25B, Sgs, Sgs-1, sgs-1, sgs1', description='Salivary gland secretion 1', other_designations='CG3047-PA|Sgs1-PA|mucin 25B|salivary gland protein 1|salivary gland secretion 1', map_location='2-13.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4937409, end_position_on_the_genomic_accession=4941423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33702, CurrentID=0, Status='live', Symbol='CG14044', Aliases='Dmel_ Dmel\\CG14044', description='CG14044 gene product from transcript CG14044-RA', other_designations='CG14044-PA|uncharacterized protein', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4942742, end_position_on_the_genomic_accession=4944305, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33703, CurrentID=0, Status='live', Symbol='mRpL24', Aliases='Dmel_CG8849, CG8849, Dmel\\CG8849, L24, RpL24', description='mitochondrial ribosomal protein L24', other_designations='CG8849-PA|mRpL24-PA', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4944256, end_position_on_the_genomic_accession=4945361, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33704, CurrentID=0, Status='live', Symbol='betaggt-I', Aliases='Dmel_CG3469, CG3469, Dmel\\CG3469, GGT1, SR2-2, beta GGT-I, beta-GGT, betaGGT-I, betaGGTI, bro3, ggt-1, ggt1', description='beta subunit of type I geranylgeranyl transferase', other_designations='CG3469-PA|CG3469-PB|betaggt-I-PA|betaggt-I-PB|suppressor of Ras1 2-2|type I geranylgeranyl transferase', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4945579, end_position_on_the_genomic_accession=4948324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33705, CurrentID=0, Status='live', Symbol='jet', Aliases='Dmel_CG8873, CG8873, Dmel\\CG8873, FBXL15, JET', description='jetlag', other_designations='CG8873-PA|CG8873-PB|jet-PA|jet-PB', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4948206, end_position_on_the_genomic_accession=4949784, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33706, CurrentID=0, Status='live', Symbol='Jon25Biii', Aliases='Dmel_CG8871, 25B, CG8871, Dmel\\CG8871, Jon25, Jon25B, SP177', description='Jonah 25Biii', other_designations='CG8871-PA|Jon25Biii-PA|jonah 25B|jonah 25Biii', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4949941, end_position_on_the_genomic_accession=4950767, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33707, CurrentID=0, Status='live', Symbol='Jon25Bii', Aliases='Dmel_CG8869, 25B, CG8869, Dmel\\CG8869, Jon25, Jon25B, SP112', description='Jonah 25Bii', other_designations='CG8869-PA|CG8869-PB|Jon25Bii-PA|Jon25Bii-PB|jonah 25B|jonah 25Bii', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4952243, end_position_on_the_genomic_accession=4953136, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33708, CurrentID=0, Status='live', Symbol='Jon25Bi', Aliases='Dmel_CG8867, 25B, CG8867, Dmel\\CG8867, Jon25, Jon25B, SP137, Ser4', description='Jonah 25Bi', other_designations='CG8867-PB|Jon25Bi-PB|jonah 25B|jonah 25Bi|serine protease 4', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4954279, end_position_on_the_genomic_accession=4955144, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33709, CurrentID=0, Status='live', Symbol='Marcal1', Aliases='Dmel_CG3753, CG3753, Dmel\\CG3753, HARP, Marca1, dmHARP', description='CG3753 gene product from transcript CG3753-RA', other_designations='CG3753-PA|Marcal1-PA|dsmarcal1|hepA-related protein', map_location='25B4-25B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4955509, end_position_on_the_genomic_accession=4958292, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33710, CurrentID=0, Status='live', Symbol='Trip1', Aliases='Dmel_CG8882, CG8882, DmTRIP, Dmel\\CG8882, eIF-3beta, eIF3-S2, eIF3i', description='CG8882 gene product from transcript CG8882-RA', other_designations='CG8882-PA|Trip1-PA', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4966367, end_position_on_the_genomic_accession=4967775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33711, CurrentID=0, Status='live', Symbol='Scox', Aliases='Dmel_CG8885, CG8885, Dmel\\CG8885, SCO, SCO1, SCOX, dSco1, l(2)SH1783, l(2)SH2 1783, scox', description='Synthesis of cytochrome c oxidase', other_designations='CG8885-PA|SCO1 protein homolog|Scox-PA|lethal (2) SH1783|synthesis of cytochrome c oxidase', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4967832, end_position_on_the_genomic_accession=4968935, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33712, CurrentID=0, Status='live', Symbol='CG3756', Aliases='Dmel_ Dmel\\CG3756', description='CG3756 gene product from transcript CG3756-RA', other_designations='CG3756-PA', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4969004, end_position_on_the_genomic_accession=4970216, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33714, CurrentID=0, Status='live', Symbol='CG14043', Aliases='Dmel_ Dmel\\CG14043', description='CG14043 gene product from transcript CG14043-RA', other_designations='CG14043-PA|uncharacterized protein', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4971702, end_position_on_the_genomic_accession=4973687, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33715, CurrentID=0, Status='live', Symbol='mRpL28', Aliases='Dmel_CG3782, CG3782, Dmel\\CG3782, MRP-L28', description='mitochondrial ribosomal protein L28', other_designations='CG3782-PA|CG3782-PB|ecdysteroid regulated 16kd protein|mRpL28-PA|mRpL28-PB', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4973976, end_position_on_the_genomic_accession=4975316, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33716, CurrentID=0, Status='live', Symbol='Gmd', Aliases='Dmel_CG8890, Brazil nut, CG8890, Dm-gmd, DmGMD, Dmel\\CG8890, GMD, gmd, l(2)SH1606, l(2)SH2 1606', description='GDP-mannose 4,6-dehydratase', other_designations='CG8890-PA|GDP-D-mannose 4,6-dehydratase|GDP-mannose 4,6 dehydratase|GDP-mannose 4,6-dehydrase|GDP-mannose dehydratase|Gmd-PA', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4975300, end_position_on_the_genomic_accession=4977023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33717, CurrentID=0, Status='live', Symbol='CG3792', Aliases='Dmel_ Dmel\\CG3792, anon-EST:Posey8', description='CG3792 gene product from transcript CG3792-RA', other_designations='CG3792-PA', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4977312, end_position_on_the_genomic_accession=4978567, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33718, CurrentID=0, Status='live', Symbol='CG8891', Aliases='Dmel_ Dmel\\CG8891', description='CG8891 gene product from transcript CG8891-RA', other_designations='CG8891-PA', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4978484, end_position_on_the_genomic_accession=4979379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33719, CurrentID=0, Status='live', Symbol='CG8892', Aliases='Dmel_ Dmel\\CG8892', description='CG8892 gene product from transcript CG8892-RC', other_designations='CG8892-PB|CG8892-PC|CG8892-PD|CG8892-PE|UBX domain 7 like', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4979513, end_position_on_the_genomic_accession=4981592, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33721, CurrentID=0, Status='live', Symbol='Rtnl1', Aliases='Dmel_CG33113, 142893_at, CG18623, CG33113, CG8895, Dmel\\CG33113, G9, Retl1, Rtnl-1, rtnl1', description='Reticulon-like1', other_designations='CG33113-PA|CG33113-PB|CG33113-PC|CG33113-PD|CG33113-PE|CG33113-PF|CG33113-PG|CG33113-PH|CG33113-PI|CG33113-PJ|CG33113-PL|Reticulon-like 1|Rtnl1-PA|Rtnl1-PB|Rtnl1-PC|Rtnl1-PD|Rtnl1-PE|Rtnl1-PF|Rtnl1-PG|Rtnl1-PH|Rtnl1-PI|Rtnl1-PJ|Rtnl1-PL|reticulon|reticulon-like1', map_location='25B9-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4992808, end_position_on_the_genomic_accession=5009720, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  33725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33725, CurrentID=0, Status='live', Symbol='CG3887', Aliases='Dmel_ Dmel\\CG3887', description='CG3887 gene product from transcript CG3887-RC', other_designations='CG3887-PA|CG3887-PB|CG3887-PC', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5010922, end_position_on_the_genomic_accession=5012127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33726, CurrentID=0, Status='live', Symbol='vkg', Aliases='Dmel_CG16858, 1209, 6072, CG16858, CT25584, Coll IValpha2, DmColA2, Dmel\\CG16858, Vkg, VkgC, alpha(IV)2/vkg, col4a2, coll-IV, coll. IV, collagen-IV, l(2)01209', description='viking', other_designations='CG16858-PA|CG16858-PB|CG16858-PC|collagen IV|collagen IV alpha2|collagenIV a2|type IV collagen|type IV collagen, alpha-chain|type IV collagenS|vkg-PA|vkg-PB|vkg-PC', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5012144, end_position_on_the_genomic_accession=5028191, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33727, CurrentID=0, Status='live', Symbol='Cg25C', Aliases='Dmel_CG4145, CG25C, CG4145, CT12803, Cgc25, Cgc25C, Col IV alpha(1), Coll, Coll IV, DCg1, Dcg, Dcg1, Dmel\\CG4145, alpha(IV)1/Cg25C, alpha1, cg25C, cg25c, col4a1, gp125', description='Collagen type IV', other_designations='CG4145-PA|CG4145-PB|CG4145-PC|Cg25C-PA|Cg25C-PB|Cg25C-PC|collagen|collagen IV|collagen type IV|collagenIV a1|laminin alpha3,5|preprocollagen alpha1(IV)|type IV collagen|type IV collagen, alpha-chain', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5029609, end_position_on_the_genomic_accession=5037279, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  246600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246600, CurrentID=0, Status='live', Symbol='CG30411', Aliases='Dmel_ BcDNA:RE06996, Dmel\\CG30411', description='CG30411 gene product from transcript CG30411-RA', other_designations='CG30411-PA|uncharacterized protein', map_location='59D11-59D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23415939, end_position_on_the_genomic_accession=23417238, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33732, CurrentID=0, Status='live', Symbol='CG9121', Aliases='Dmel_ Dmel\\CG9121', description='CG9121 gene product from transcript CG9121-RD', other_designations='CG9121-PA|CG9121-PB|CG9121-PC|CG9121-PD', map_location='25C1-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5047818, end_position_on_the_genomic_accession=5051062, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33734, CurrentID=0, Status='live', Symbol='CG14040', Aliases='Dmel_ Dmel\\CG14040', description='CG14040 gene product from transcript CG14040-RA', other_designations='CG14040-PA|uncharacterized protein', map_location='25C3-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5052843, end_position_on_the_genomic_accession=5054575, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33735, CurrentID=0, Status='live', Symbol='CG14036', Aliases='Dmel_ Dmel\\CG14036', description='CG14036 gene product from transcript CG14036-RA', other_designations='CG14036-PA|uncharacterized protein', map_location='25C3-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5054579, end_position_on_the_genomic_accession=5055016, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33736, CurrentID=0, Status='live', Symbol='His3.3A', Aliases='Dmel_CG5825, CG5825, Dmel\\CG5825, H3.3, H3.3A, His-3.3A, His3.3, PH3, dH3.3A', description='Histone H3.3A', other_designations='CG5825-PA|CG5825-PC|CG5825-PD|H3L-like histone|His3.3A-PA|His3.3A-PC|His3.3A-PD|histone H3.3|histone H3.3 variant|histone H3.3A|histone variant 3.3|histone variant H3.3', map_location='25C3-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5055058, end_position_on_the_genomic_accession=5056149, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33738, CurrentID=0, Status='live', Symbol='Rpn11', Aliases='Dmel_CG18174, 37B, CG18174, DmS13, Dmel\\CG18174, Q9V3H2, RPN11, RPn11, S13, p37B, rpn11, yip5', description='Regulatory particle non-ATPase 11', other_designations='CG18174-PA|Rpn11-PA|regulatory particle non-ATPase 11|yippee interacting protein 5', map_location='25C3-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5058404, end_position_on_the_genomic_accession=5059790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33739, CurrentID=0, Status='live', Symbol='qtc', Aliases='Dmel_CG14039, CG14039, Dmel\\CG14039', description='quick-to-court', other_designations='CG14039-PA|CG14039-PB|CG14039-PE|CG14039-PF|CG14039-PG|CG14039-PH|CG14039-PI|CG14039-PJ|CG14039-PK|CG14039-PL|CG14039-PM|CG14039-PN|qtc-PA|qtc-PB|qtc-PE|qtc-PF|qtc-PG|qtc-PH|qtc-PI|qtc-PJ|qtc-PK|qtc-PL|qtc-PM|qtc-PN', map_location='25C3-25C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5060815, end_position_on_the_genomic_accession=5070750, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33742, CurrentID=0, Status='live', Symbol='CG5828', Aliases='Dmel_ Dmel\\CG5828', description='CG5828 gene product from transcript CG5828-RB', other_designations='CG5828-PA|CG5828-PB', map_location='25C6-25C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5093976, end_position_on_the_genomic_accession=5095910, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33743, CurrentID=0, Status='live', Symbol='CG4230', Aliases='Dmel_ Dmel\\CG4230', description='CG4230 gene product from transcript CG4230-RA', other_designations='CG4230-PA|CG4230-PB|CG4230-PC', map_location='25C6-25C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5096227, end_position_on_the_genomic_accession=5099294, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33744, CurrentID=0, Status='live', Symbol='ND-13A', Aliases='Dmel_CG8680, CG8680, Dmel\\CG8680, NUMM', description='NADH dehydrogenase (ubiquinone) 13 kDa A subunit', other_designations='CG8680-PA|CG8680-PB|CG8680-PC|NADH ubiquinone oxidoreductase 13 kDa A|NADH-ubiquinone oxidoreductase 13 kDa-A subunit|NADH:ubiquinone oxidoreductase 13 kDa A subunit|ND-13A-PA|ND-13A-PB|ND-13A-PC', map_location='25C6-25C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5099177, end_position_on_the_genomic_accession=5100288, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33748, CurrentID=0, Status='live', Symbol='Cyp28d2', Aliases='Dmel_CG6081, 28d2, CG6081, Dmel\\CG6081', description='CG6081 gene product from transcript CG6081-RA', other_designations='CG6081-PA|Cyp28d2-PA|cytochrome P450', map_location='25C10-25C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5207267, end_position_on_the_genomic_accession=5209345, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33749, CurrentID=0, Status='live', Symbol='Cyp28d1', Aliases='Dmel_CG10833, 28d1, CG10833, Dmel\\CG10833', description='CG10833 gene product from transcript CG10833-RA', other_designations='CG10833-PA|Cyp28d1-PA|cytochrome P450', map_location='25C10-25C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5210460, end_position_on_the_genomic_accession=5212445, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33750, CurrentID=0, Status='live', Symbol='CG7742', Aliases='Dmel_ Dmel\\CG7742', description='CG7742 gene product from transcript CG7742-RD', other_designations='CG7742-PA|CG7742-PB|CG7742-PC|CG7742-PD', map_location='25C10-25D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5212395, end_position_on_the_genomic_accession=5215081, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33751, CurrentID=0, Status='live', Symbol='CG14034', Aliases='Dmel_ Dmel\\CG14034, anon-SAGE:Wang-122', description='CG14034 gene product from transcript CG14034-RC', other_designations='CG14034-PB|CG14034-PC|anon-SAGE:Wang-122|uncharacterized protein', map_location='25D1-25D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5215189, end_position_on_the_genomic_accession=5216542, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33752, CurrentID=0, Status='live', Symbol='TpnC25D', Aliases='Dmel_CG6514, BcDNA:RE53289, CG6514, DmTpnC25D, DmTpnCII, Dmel\\CG6514, TNC25D, TnC, TpnC II, TpnCII, anon-EST:Posey103', description='Troponin C at 25D', other_designations='CG6514-PA|CG6514-PB|CG6514-PC|TpnC25D-PA|TpnC25D-PB|TpnC25D-PC|troponin C at 25D', map_location='25D1-25D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5216701, end_position_on_the_genomic_accession=5219053, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33753, CurrentID=0, Status='live', Symbol='tkv', Aliases='Dmel_CG14026, Atkv, Atr25D, Brk25D, Brk25D1, Brk25D2, CG14026, Dmel\\CG14026, Dtfr, STK-A, TKV, TKV1, Tkv, dtfr, kv, l(2)04415, l(2)25Da, str1, tkva, tkv', description='thickveins', other_designations='Activin-A-receptor-25D|BMP receptor kinase 25D|CG14026-PA|CG14026-PB|CG14026-PC|CG14026-PD|DPP receptor|Thick-vein|Thick-veins|slater|thick veins|thickvein|tkv-PA|tkv-PB|tkv-PC|tkv-PD', map_location='2-16 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5219000, end_position_on_the_genomic_accession=5271384, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33754, CurrentID=0, Status='live', Symbol='Cyp4ac1', Aliases='Dmel_CG14032, 4ac1, CG14032, Dmel\\CG14032, cyp4ac1', description='CG14032 gene product from transcript CG14032-RA', other_designations='CG14032-PA|Cyp4ac1-PA', map_location='25D2-25D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5264001, end_position_on_the_genomic_accession=5265899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33755, CurrentID=0, Status='live', Symbol='Cyp4ac2', Aliases='Dmel_CG17970, 4ac2, CG17970, Dmel\\CG17970', description='CG17970 gene product from transcript CG17970-RB', other_designations='CG17970-PB|Cyp4ac2-PB|cytochrome P450', map_location='25D2-25D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5266392, end_position_on_the_genomic_accession=5268455, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33756, CurrentID=0, Status='live', Symbol='Cyp4ac3', Aliases='Dmel_CG14031, 4ac3, CG14031, Dmel\\CG14031', description='CG14031 gene product from transcript CG14031-RA', other_designations='CG14031-PA|Cyp4ac3-PA|cytochrome P450', map_location='25D2-25D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5268498, end_position_on_the_genomic_accession=5270329, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33757, CurrentID=0, Status='live', Symbol='Bsg25D', Aliases='Dmel_CG14025, CG14025, Dmel\\CG14025, brg25D, bsg25D, bsq25D', description='Blastoderm-specific gene 25D', other_designations='Bsg25D-PB|Bsg25D-PC|Bsg25D-PD|Bsg25D-PE|Bsg25D-PF|Bsg25D-PG|Bsg25D-PH|CG14025-PB|CG14025-PC|CG14025-PD|CG14025-PE|CG14025-PF|CG14025-PG|CG14025-PH', map_location='25D2-25D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5271728, end_position_on_the_genomic_accession=5278548, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  33758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33758, CurrentID=0, Status='live', Symbol='Bub1', Aliases='Dmel_CG14030, BUB1, BUB1-like, BubR1, BubRI, CG14030, CG14030/Bub1, Dmel\\CG14030, bub1', description='Bub1 homologue', other_designations='Bub1-PA|Bub1-PB|Bub1-PC|CG14030-PA|CG14030-PB|CG14030-PC', map_location='25D3-25D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5278865, end_position_on_the_genomic_accession=5283283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33759, CurrentID=0, Status='live', Symbol='vri', Aliases='Dmel_CG14029, CG14029, DM16, Dmel\\CG14029, VRI, Vri, argo, jf23, l(2)25Db, l(2)jf23, mat(2)ea-G, mat(2)earlyRS32, mel_vri', description='vrille', other_designations='CG14029-PA|CG14029-PC|CG14029-PD|CG14029-PE|CG14029-PF|VRILLE|lethal(2)25Db|maternal effect early arrest G|vri-PA|vri-PC|vri-PD|vri-PE|vri-PF', map_location='2-17 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5288944, end_position_on_the_genomic_accession=5311223, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33760, CurrentID=0, Status='live', Symbol='CG14024', Aliases='Dmel_ Dmel\\CG14024, d4ST1', description='CG14024 gene product from transcript CG14024-RB', other_designations='CG14024-PA|CG14024-PB|uncharacterized protein', map_location='25D5-25D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5311311, end_position_on_the_genomic_accession=5314895, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33761, CurrentID=0, Status='live', Symbol='Ncoa6', Aliases='Dmel_CG14023, CG14023, Dmel\\CG14023, NCOA6, dNCOA6', description='Nuclear receptor coactivator 6', other_designations='CG14023-PC|CG14023-PD|CG14023-PF|Ncoa6-PC|Ncoa6-PD|Ncoa6-PF|nuclear receptor coactivator 6', map_location='25D5-25D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5315290, end_position_on_the_genomic_accession=5326683, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33763, CurrentID=0, Status='live', Symbol='CG14022', Aliases='Dmel_ BcDNA:RH62365, Dmel\\CG14022', description='CG14022 gene product from transcript CG14022-RA', other_designations='CG14022-PA|CG14022-PB|uncharacterized protein', map_location='25D6-25D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5327975, end_position_on_the_genomic_accession=5328879, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33764, CurrentID=0, Status='live', Symbol='TotM', Aliases='Dmel_CG14027, CG14027, Dmel\\CG14027, TOTM, Tot M, tot, totM', description='Turandot M', other_designations='CG14027-PA|CG14027-PB|TotM-PA|TotM-PB|turandot|turandot M', map_location='25D6-25D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5329857, end_position_on_the_genomic_accession=5330464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33765, CurrentID=0, Status='live', Symbol='fusl', Aliases='Dmel_CG14021, CG14021, Dmel\\CG14021', description='fuseless', other_designations='CG14021-PA|CG14021-PB|CG14021-PC|CG14021-PD|CG14021-PE|CG14021-PF|fusl-PA|fusl-PB|fusl-PC|fusl-PD|fusl-PE|fusl-PF', map_location='25D6-25D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5330470, end_position_on_the_genomic_accession=5338073, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33766, CurrentID=0, Status='live', Symbol='CG12512', Aliases='Dmel_ Dmel\\CG12512', description='CG12512 gene product from transcript CG12512-RA', other_designations='CG12512-PA|uncharacterized protein', map_location='25D6-25D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5339132, end_position_on_the_genomic_accession=5342234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33768, CurrentID=0, Status='live', Symbol='nompC', Aliases='Dmel_CG11020, CG11020, CG17468, CT30855, DmNompC, Dmel\\CG11020, NOMPC, NompC, ORE-13, TRPN, TRPN1, jf24, l(2)25Dc, l(2)jf24, nomp', description='no mechanoreceptor potential C', other_designations='CG11020-PD|CG11020-PE|CG11020-PF|CG11020-PG|CG11020-PH|No Mechano-receptor potential C|No mechanoreceptor potential C|lethal(2)25Dc|nompC-PD|nompC-PE|nompC-PF|nompC-PG|nompC-PH', map_location='25D6-25D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5346200, end_position_on_the_genomic_accession=5365760, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  33769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33769, CurrentID=0, Status='live', Symbol='H15', Aliases='Dmel_CG6604, CG6604, Dm- Dmel\\CG6604, Nmr, Nmr-1, Tbx20, dm-H15, nmr, nmr1, nmr[H15]', description='CG6604 gene product from transcript CG6604-RA', other_designations='CG6604-PA|CG6604-PB|H15-PA|H15-PB|neuromancer|neuromancer 1|neuromancer 2|neuromancer-1|neuromancer1', map_location='25E1-25E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5404280, end_position_on_the_genomic_accession=5416593, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33770, CurrentID=0, Status='live', Symbol='mid', Aliases='Dmel_CG6634, CG6634, Dmel\\CG6634, H15r, H15r/nmr2, Mid(H15), TBX20, extra, los, nmr2', description='midline', other_designations='CG6634-PA|CG6634-PB|H15-related|extra|lost in space|mid-PA|mid-PB|neuromancer 1|neuromancer 2|neuromancer2', map_location='2-16 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5461639, end_position_on_the_genomic_accession=5467610, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33771, CurrentID=0, Status='live', Symbol='CG14020', Aliases='Dmel_ Dmel\\CG14020', description='CG14020 gene product from transcript CG14020-RA', other_designations='CG14020-PA|CG14020-PB|uncharacterized protein', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5517737, end_position_on_the_genomic_accession=5518949, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33772, CurrentID=0, Status='live', Symbol='CG7382', Aliases='Dmel_ Dmel\\CG7382', description='CG7382 gene product from transcript CG7382-RA', other_designations='CG7382-PA', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5518893, end_position_on_the_genomic_accession=5519896, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33774, CurrentID=0, Status='live', Symbol='Vps52', Aliases='Dmel_CG7371, CG7371, Dmel\\CG7371', description='Vacuolar protein sorting 52', other_designations='CG7371-PA|Vps52-PA|vacuolar protein sorting 52', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5521281, end_position_on_the_genomic_accession=5523898, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33775, CurrentID=0, Status='live', Symbol='CG6907', Aliases='Dmel_ Dmel\\CG6907', description='CG6907 gene product from transcript CG6907-RA', other_designations='CG6907-PA|CG6907-PB', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5524139, end_position_on_the_genomic_accession=5525809, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33776, CurrentID=0, Status='live', Symbol='CG31648', Aliases='Dmel_ CG6922, COXZ, Dmel\\CG31648', description='CG31648 gene product from transcript CG31648-RA', other_designations='CG31648-PA|complex IV assembly protein COX11|cytochrome c oxidase assembly protein COX11|uncharacterized protein', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5526957, end_position_on_the_genomic_accession=5527854, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33777, CurrentID=0, Status='live', Symbol='CG7277', Aliases='Dmel_ Dmel\\CG7277', description='CG7277 gene product from transcript CG7277-RA', other_designations='CG7277-PA', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5530620, end_position_on_the_genomic_accession=5532223, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33778, CurrentID=0, Status='live', Symbol='Cap-D3', Aliases='Dmel_CG31989, BcDNA:RE74832, CAP-D2, CAP-D3, CG14019, CG31989, CG6922, CapD3, Dmel\\CG31989, cap-D3, dCap-D3', description='Chromosome associated protein D3', other_designations='CG31989-PA|CG31989-PB|Cap-D3-PA|Cap-D3-PB|chromosome associated protein D3', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5525905, end_position_on_the_genomic_accession=5535958, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33779, CurrentID=0, Status='live', Symbol='tomb', Aliases='Dmel_CG14016, CG14016, Dmel\\CG14016, Tomb', description='tombola', other_designations='CG14016-PA|tomb-PA', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5535861, end_position_on_the_genomic_accession=5536758, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33780, CurrentID=0, Status='live', Symbol='CG14015', Aliases='Dmel_ Dmel\\CG14015', description='CG14015 gene product from transcript CG14015-RA', other_designations='CG14015-PA|uncharacterized protein', map_location='25E5-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5536850, end_position_on_the_genomic_accession=5539196, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33781, CurrentID=0, Status='live', Symbol='Hel25E', Aliases='Dmel_CG7269, CG7269, Dbp25F, DmRH4, Dmel\\CG7269, HEL, Hel, UAP56, WM6, Wm6, hel, jf26, l(2)25Eb, l(2)gdh-12, l(2)gdh12, l(2)jf26, l(2)k11511, sz, uap56', description='Helicase at 25E', other_designations='CG7269-PA|CG7269-PB|CG7269-PC|CG7269-PD|Hel25E-PA|Hel25E-PB|Hel25E-PC|Hel25E-PD|RNA helicase|helicase|helicase at 25E|lethal(2)25Eb', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5539326, end_position_on_the_genomic_accession=5542314, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33782, CurrentID=0, Status='live', Symbol='Lam', Aliases='Dmel_CG6944, 2459, 74/76, CG6944, D5, DM[[O]], D[[m0]], Dm, Dm(0), Dm0, Dm1, Dm2, DmLamin, DmO, Dm[[0]], Dm[[1]], Dm[[2]], Dm[[mit]], Dm[[o]], Dmel\\CG6944, Dmo, LAM Dm[[0]], Lam(Dm0), LamDm, LamDm0, LamDm[[0]], LamDm[[o]], Lam[[Dm0]], Lamin B, Lamin Dm[[0]], jf27, l(2)04643, l(2)25Ec, l(2)gdh-7, l(2)gdh7, l(2)jf27, lam, lamDm0, lamDm[[0]], lamin, lamin B, lamin Dm0, lamin Dm[[0]], misg, nlam, Lam', description='Lamin', other_designations='B-type lamin|CG6944-PA|CG6944-PB|CG6944-PC|CG6944-PD|Lam-PA|Lam-PB|Lam-PC|Lam-PD|Lamin-B|LaminDm|LaminDm0|alpha-Lamin|intermediate filament-like lamina|lamin|lamin A|lamin B|lamin D|lamin D[[0]]|lamin D[[m0]]|lamin D[[m]]0|lamin D[[o]]|lamin Dm|lamin Dm0|lamin Dm[[0]]|lamin Dmo|lethal(2)25Ec|misguided|misrouted|n-lamin|nuclear lamin', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5542480, end_position_on_the_genomic_accession=5546642, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33783, CurrentID=0, Status='live', Symbol='Oscillin', Aliases='Dmel_CG6957, CG6957, Dmel\\CG6957', description='CG6957 gene product from transcript CG6957-RC', other_designations='CG6957-PA|CG6957-PB|CG6957-PC|CG6957-PD|CG6957-PE|CG6957-PF|CG6957-PG|CG6957-PH|Oscillin-PA|Oscillin-PB|Oscillin-PC|Oscillin-PD|Oscillin-PE|Oscillin-PF|Oscillin-PG|Oscillin-PH|oscillin', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5547086, end_position_on_the_genomic_accession=5549614, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33784, CurrentID=0, Status='live', Symbol='CG14014', Aliases='Dmel_ Dmel\\CG14014', description='CG14014 gene product from transcript CG14014-RB', other_designations='CG14014-PB|uncharacterized protein', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5549709, end_position_on_the_genomic_accession=5550835, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33785, CurrentID=0, Status='live', Symbol='CG18269', Aliases='Dmel_ Dmel\\CG18269', description='CG18269 gene product from transcript CG18269-RA', other_designations='CG18269-PA|uncharacterized protein', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5551838, end_position_on_the_genomic_accession=5552756, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33786, CurrentID=0, Status='live', Symbol='CG14013', Aliases='Dmel_ Dmel\\CG14013', description='CG14013 gene product from transcript CG14013-RA', other_designations='CG14013-PA|uncharacterized protein', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5552843, end_position_on_the_genomic_accession=5553850, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33787, CurrentID=0, Status='live', Symbol='CG14017', Aliases='Dmel_ Dmel\\CG14017', description='CG14017 gene product from transcript CG14017-RA', other_designations='CG14017-PA|uncharacterized protein', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5553897, end_position_on_the_genomic_accession=5555137, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33788, CurrentID=0, Status='live', Symbol='GluRIIA', Aliases='Dmel_CG6992, CG6992, D-GluRIIA, DGluR-II, DGluR-IIA, DGluR2a, DGluRIIA, DGluRIIa, DgluRII, DmelGluRIIA, Dmel\\CG6992, Glu-RII, Glu-RIIA, GluR, GluR-IIA, GluR2, GluR2a, GluRII, GluRII-A, GluRIIa, GluR[[IIa]], GlurIIA, dGluRII, dGluRIIA, dglur-IIA, dglurIIA, gluRIIA, glurIIA, gluriia', description='Glutamate receptor IIA', other_designations='CG6992-PI|GluRIIA-PI|glutamate IIA receptor|glutamate receptor|glutamate receptor IIA|glutamate receptor subunit IIA', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5555072, end_position_on_the_genomic_accession=5558994, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  33789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33789, CurrentID=0, Status='live', Symbol='GluRIIB', Aliases='Dmel_CG7234, CG7234, DGluIIB, DGluR-IIB, DGluRIIB, DmelGluRIIB, Dmel\\CG7234, Glu-RIIB, GluRIIb, dGluRIIB, dglurIIB, glurIIB', description='Glutamate receptor IIB', other_designations='CG7234-PI|GluRIIB-PI|glutamate receptor|glutamate receptor IIB', map_location='25E6-25E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5559342, end_position_on_the_genomic_accession=5563502, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  33790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33790, CurrentID=0, Status='live', Symbol='CG14011', Aliases='Dmel_ Dmel\\CG14011', description='CG14011 gene product from transcript CG14011-RC', other_designations='CG14011-PA|CG14011-PB|CG14011-PC|uncharacterized protein', map_location='25F1-25F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5579580, end_position_on_the_genomic_accession=5581116, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33791, CurrentID=0, Status='live', Symbol='CG7251', Aliases='Dmel_ Dmel\\CG7251, anon-WO0140519.103, bs11h10.y1', description='CG7251 gene product from transcript CG7251-RB', other_designations='CG7251-PA|CG7251-PB', map_location='25F1-25F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5587115, end_position_on_the_genomic_accession=5588414, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33792, CurrentID=0, Status='live', Symbol='CG18266', Aliases='Dmel_ Dmel\\CG18266, anon-WO0140519.54', description='CG18266 gene product from transcript CG18266-RA', other_designations='CG18266-PA|uncharacterized protein', map_location='25F1-25F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5596833, end_position_on_the_genomic_accession=5598636, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33793, CurrentID=0, Status='live', Symbol='DIP-eta', Aliases='Dmel_CG14010, 14010, CG14010, CT33567, Dmel\\CG14010', description='Dpr-interacting protein eta', other_designations='CG14010-PB|CG14010-PD|DIP-eta-PB|DIP-eta-PD', map_location='25F1-25F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5601723, end_position_on_the_genomic_accession=5617946, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  33795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33795, CurrentID=0, Status='live', Symbol='DIP-theta', Aliases='Dmel_CG31646, 31646, CG14008, CG14009, CG31646, CT33566, Dmel\\CG31646', description='Dpr-interacting protein theta', other_designations='CG31646-PA|DIP-theta-PA', map_location='25F1-25F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5626291, end_position_on_the_genomic_accession=5658560, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  33796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33796, CurrentID=0, Status='live', Symbol='Hsp60C', Aliases='Dmel_CG7235, A4V087_DROME, CG7235, Dmel\\CG7235, bs36d10.y1', description='CG7235 gene product from transcript CG7235-RA', other_designations='CG7235-PA|CG7235-PB|CG7235-PC|Hsp60C-PA|Hsp60C-PB|Hsp60C-PC', map_location='25F2-25F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5662849, end_position_on_the_genomic_accession=5670838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33797, CurrentID=0, Status='live', Symbol='CG12511', Aliases='Dmel_ Dmel\\CG12511', description='CG12511 gene product from transcript CG12511-RD', other_designations='CG12511-PB|CG12511-PC|CG12511-PD|uncharacterized protein', map_location='25F2-25F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5678582, end_position_on_the_genomic_accession=5679499, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33798, CurrentID=0, Status='live', Symbol='CG7236', Aliases='Dmel_CG7236, AC018104/Cdk1l, Dmel\\CG7236, KKIALRE', description='CG7236 gene product from transcript CG7236-RE', other_designations='CG7236-PB|CG7236-PC|CG7236-PD|CG7236-PE', map_location='25F3-25F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5706374, end_position_on_the_genomic_accession=5713791, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33799, CurrentID=0, Status='live', Symbol='CG14007', Aliases='Dmel_ Dmel\\CG14007', description='CG14007 gene product from transcript CG14007-RC', other_designations='CG14007-PC|uncharacterized protein', map_location='25F3-25F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5715349, end_position_on_the_genomic_accession=5716680, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33800, CurrentID=0, Status='live', Symbol='CG11149', Aliases='Dmel_ Dmel\\CG11149, dibeta3GnT4', description='CG11149 gene product from transcript CG11149-RD', other_designations='CG11149-PA|CG11149-PB|CG11149-PC|CG11149-PD|CG11149-PE|uncharacterized protein', map_location='25F3-25F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5718121, end_position_on_the_genomic_accession=5720899, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33802, CurrentID=0, Status='live', Symbol='CG14006', Aliases='Dmel_ Dmel\\CG14006', description='CG14006 gene product from transcript CG14006-RB', other_designations='CG14006-PA|CG14006-PB|uncharacterized protein', map_location='25F4-25F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5726415, end_position_on_the_genomic_accession=5728811, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33803, CurrentID=0, Status='live', Symbol='CG11147', Aliases='Dmel_ Dmel\\CG11147', description='CG11147 gene product from transcript CG11147-RA', other_designations='CG11147-PA|CG11147-PB|uncharacterized protein', map_location='25F4-25F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5733675, end_position_on_the_genomic_accession=5741244, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  33804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33804, CurrentID=0, Status='live', Symbol='CG11029', Aliases='Dmel_ Dmel\\CG11029', description='CG11029 gene product from transcript CG11029-RA', other_designations='CG11029-PA|uncharacterized protein', map_location='25F4-25F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5746157, end_position_on_the_genomic_accession=5747961, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33805, CurrentID=0, Status='live', Symbol='CG11030', Aliases='Dmel_ Dmel\\CG11030', description='CG11030 gene product from transcript CG11030-RB', other_designations='CG11030-PA|CG11030-PB|uncharacterized protein', map_location='25F3-25F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5724656, end_position_on_the_genomic_accession=5762420, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33806, CurrentID=0, Status='live', Symbol='obst-E', Aliases='Dmel_CG11142, CG11142, Dmel\\CG11142, Obst-E', description='obstructor-E', other_designations='CG11142-PA|CG11142-PB|obst-E-PA|obst-E-PB', map_location='25F4-25F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5762553, end_position_on_the_genomic_accession=5766142, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33807, CurrentID=0, Status='live', Symbol='CG9171', Aliases='Dmel_ Dmel\\CG9171, dibeta3GnT2', description='CG9171 gene product from transcript CG9171-RD', other_designations='CG9171-PC|CG9171-PD|CG9171-PE|CG9171-PF|CG9171-PG', map_location='25F4-25F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5767002, end_position_on_the_genomic_accession=5800253, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33808, CurrentID=0, Status='live', Symbol='CG14005', Aliases='Dmel_ Dmel\\CG14005', description='CG14005 gene product from transcript CG14005-RB', other_designations='CG14005-PA|CG14005-PB|uncharacterized protein', map_location='25F5-25F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5800517, end_position_on_the_genomic_accession=5801975, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33809, CurrentID=0, Status='live', Symbol='CG7239', Aliases='Dmel_ Dmel\\CG7239, anon-EST:Posey250', description='CG7239 gene product from transcript CG7239-RA', other_designations='CG7239-PA', map_location='25F5-25F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5802220, end_position_on_the_genomic_accession=5804030, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33810, CurrentID=0, Status='live', Symbol='CG11034', Aliases='Dmel_ Dmel\\CG11034', description='CG11034 gene product from transcript CG11034-RC', other_designations='CG11034-PB|CG11034-PC|CG11034-PD|uncharacterized protein', map_location='25F5-25F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5805395, end_position_on_the_genomic_accession=5809063, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33812, CurrentID=0, Status='live', Symbol='TrissinR', Aliases='Dmel_CG34381, CG14002, CG14003, CG14004, CG31645, CG34381, Dmel\\CG34381, Dmel_CG14004, Dmel_CG31645, anon-WO0170980.172, anon-WO0170980.173', description='Trissin receptor', other_designations='CG34381-PB|CG34381-PC|CG34381-PD|CG34381-PE|TrissinR-PB|TrissinR-PC|TrissinR-PD|TrissinR-PE|trissin receptor', map_location='25F5-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5824004, end_position_on_the_genomic_accession=5856767, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  246630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246630, CurrentID=0, Status='live', Symbol='Ir52c', Aliases='Dmel_CG30468, CG12961, CG30468, DmelIR52c, Dmel\\CG30468, IR52c', description='Ionotropic receptor 52c', other_designations='CG30468-PA|Ir52c-PA|ionotropic receptor 52c', map_location='52A6-52A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15504201, end_position_on_the_genomic_accession=15506000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33814, CurrentID=0, Status='live', Symbol='rau', Aliases='Dmel_CG8965, CG8965, Dmel\\CG8965', description='CG8965 gene product from transcript CG8965-RA', other_designations='CG8965-PA|CG8965-PB|rau-PA|rau-PB', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5859769, end_position_on_the_genomic_accession=5879498, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33815, CurrentID=0, Status='live', Symbol='CG9029', Aliases='Dmel_ Dmel\\CG9029', description='CG9029 gene product from transcript CG9029-RB', other_designations='CG9029-PA|CG9029-PB', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5886002, end_position_on_the_genomic_accession=5886609, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33816, CurrentID=0, Status='live', Symbol='Acp26Ab', Aliases='Dmel_CG9024, Acp, Acp26A, Acp26Aab, Acp26a, CG9024, Dmel\\CG9024, Mst26A, Mst26Ab, acp26Ab, msP355b, mst 355b, mst355b', description='Accessory gland protein 26Ab', other_designations='Acp26Ab-PA|Acp26Ab-PB|Acp26Ab-PC|CG9024-PA|CG9024-PB|CG9024-PC|accessory gland protein 26Ab|accessory gland-specific peptide 26Ab|accessory protein 26A', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5892298, end_position_on_the_genomic_accession=5892862, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33817, CurrentID=0, Status='live', Symbol='Acp26Aa', Aliases='Dmel_CG8982, Acp, Acp26Ab, Acp26a, CG8982, Dmel\\CG8982, Mst26A, Mst26Aa, Ovulin, msP355a, mst 355a, mst355a, ovulin, Acp26Aa', description='Accessory gland protein 26Aa', other_designations='Acp26Aa-PA|CG8982-PA|accessory gland peptide 26Aa|accessory gland protein 26Aa|accessory gland-specific peptide 26Aa|accessory protein 26A|ovulin', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5892883, end_position_on_the_genomic_accession=5893896, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33818, CurrentID=0, Status='live', Symbol='CG9021', Aliases='Dmel_ Dmel\\CG9021', description='CG9021 gene product from transcript CG9021-RB', other_designations='CG9021-PA|CG9021-PB', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5903178, end_position_on_the_genomic_accession=5904689, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33819, CurrentID=0, Status='live', Symbol='bchs', Aliases='Dmel_CG14001, Alfy, BcDNA:SD12984, Bchs, Beach1, CG14001, CG9011, DBEACH1, Dmel\\CG14001, EP2299, alfy, bchd', description='blue cheese', other_designations='Beach1|Beached1|CG14001-PA|bchs-PA|beached|beached 1|dbeach1', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5907169, end_position_on_the_genomic_accession=5922776, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  33820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33820, CurrentID=0, Status='live', Symbol='CG14000', Aliases='Dmel_ Dmel\\CG14000', description='CG14000 gene product from transcript CG14000-RA', other_designations='CG14000-PA|uncharacterized protein', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5908927, end_position_on_the_genomic_accession=5910289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33822, CurrentID=0, Status='live', Symbol='CG9016', Aliases='Dmel_ BcDNA:AT19175, Dmel\\CG9016, NEST:bs09a09, NEST:bs20d08, NEST:bs24f02, NEST:bs36a03, bs03c03.y1, bs05g03.y1, bs07e11.y1, bs09c06.y1, bs10d11.y1, bs10f09.y1, bs13h06.y1, bs14b11.y1, bs16c10.y1, bs16d01.y1, bs18b05.y1, bs19d06.y1, bs20d08.y1, bs21a02.y1, bs22a02.y1, bs24f02.y1, bs25d03.y1, bs27b06.y1, bs28b03.y1, bs29e04.y1, bs29f11.y1, bs36f07.y1', description='CG9016 gene product from transcript CG9016-RB', other_designations='CG9016-PA|CG9016-PB', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5922876, end_position_on_the_genomic_accession=5923653, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33823, CurrentID=0, Status='live', Symbol='dsf', Aliases='Dmel_CG9019, CG9019, DSF, Dmel\\CG9019, Dsf, NR2E4', description='dissatisfaction', other_designations='CG9019-PA|CG9019-PB|dsf-PA|dsf-PB', map_location='26A1-26A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5925937, end_position_on_the_genomic_accession=5939261, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33824, CurrentID=0, Status='live', Symbol='Gpdh', Aliases='Dmel_CG9042, CG9042, DROGPDHA, DmG3PDH, Dmel\\CG9042, G3PDH, G3pdh, GAPDH, GPD, GPDA, GPDH, GPDH-1, Gdh, Gpd, alpha-GPD, alpha-GPDH, alpha-GPDH-1, alpha-Gpdh, alphaGPDH, alphaGpd, alphaGpdh, alphaGpdh-1, gpdh, gpdh-1, sn-Gpdh', description='Glycerol 3 phosphate dehydrogenase', other_designations="5' gene|Alpha-Glycerophosphate dehydrogenase|CG9042-PA|CG9042-PB|CG9042-PC|CG9042-PE|CG9042-PF|CG9042-PG|Gpdh-PA|Gpdh-PB|Gpdh-PC|Gpdh-PE|Gpdh-PF|Gpdh-PG|alpha glycerophosphate dehydrogenase|alpha-glycerol phosphate dehydrogenase|alpha-glycerol-3-phosphate dehydrogenase|dehydrogenase|glycerol 3 phosphate dehydrogenase|glycerol-3-phosphate dehydrogenase|glycerophosphate dehydrogenase|sn-glycerol 3-phosphate dehydrogenase|sn-glycerol-3-phosphate|sn-glycerol-3-phosphate dehydrogenase", map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5943643, end_position_on_the_genomic_accession=5948968, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33825, CurrentID=0, Status='live', Symbol='CG9044', Aliases='Dmel_ CT25972, Dmel\\CG9044, mdcds_8484', description='CG9044 gene product from transcript CG9044-RC', other_designations='CG9044-PB|CG9044-PC', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5949450, end_position_on_the_genomic_accession=5955235, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33826, CurrentID=0, Status='live', Symbol='CG13999', Aliases='Dmel_ Dmel\\CG13999', description='CG13999 gene product from transcript CG13999-RA', other_designations='CG13999-PA|uncharacterized protein', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5955312, end_position_on_the_genomic_accession=5956466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33827, CurrentID=0, Status='live', Symbol='Vm26Ab', Aliases='Dmel_CG9046, CG9046, DIM 29, Dm sV23, Dmel\\CG9046, Sv23, TU-4, VM26A.2, Vm26A, fs(2)QJ42, sV23, sV26, sv23', description='Vitelline membrane 26Ab', other_designations='CG9046-PA|CG9046-PB|Vm26Ab-PA|Vm26Ab-PB|vitelline membrane 26Ab', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5957004, end_position_on_the_genomic_accession=5957639, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33828, CurrentID=0, Status='live', Symbol='Vm26Ac', Aliases='Dmel_CG13997, CG13997, Dmel\\CG13997, Fcp26Ac, TU-3, VM26Ac, fc 20, fc20', description='Vitelline membrane 26Aac', other_designations='CG13997-PA|Vm26Ac-PA|follicle cell protein 26Ac|vitelline membrane 26Aac', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5958005, end_position_on_the_genomic_accession=5958631, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33829, CurrentID=0, Status='live', Symbol='Vm26Aa', Aliases='Dmel_CG9048, CG9048, Dmel\\CG9048, TU-2, VM26A.1, VM26Aa, Vm26A, Vm26A.1, Vm26A1, sV17', description='Vitelline membrane 26Aa', other_designations='CG9048-PA|CG9048-PB|Vm26Aa-PA|Vm26Aa-PB|vitelline membrane 26Aa', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5959596, end_position_on_the_genomic_accession=5960340, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33830, CurrentID=0, Status='live', Symbol='psd', Aliases='Dmel_CG9050, BcDNA:GM32356, CG9050, Dmel\\CG9050, Fcp26Aa, TU-1', description='palisade', other_designations='CG9050-PA|CG9050-PB|follicle cell protein 26Aa|psd-PA|psd-PB', map_location='26A4-26A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5963087, end_position_on_the_genomic_accession=5964951, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33831, CurrentID=0, Status='live', Symbol='CG13992', Aliases='Dmel_ Dmel\\CG13992', description='CG13992 gene product from transcript CG13992-RC', other_designations='CG13992-PA|CG13992-PB|CG13992-PC|uncharacterized protein', map_location='26A4-26A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5965391, end_position_on_the_genomic_accession=5967752, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33832, CurrentID=0, Status='live', Symbol='Ucp4C', Aliases='Dmel_CG9064, CG9064, DmUCP4C, Dmel\\CG9064, UCP4C, UCP4c, Ucp4c, anon-WO02079478.1, anon-WO02079478.2, anon-WO0242455.3, anon-WO0242455.4, anon-WO03037362.7, anon-WO03061681.1, anon-WO03061681.2, dmUCP4c', description='CG9064 gene product from transcript CG9064-RA', other_designations='CG9064-PA|Ucp4C-PA', map_location='26A4-26A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5968566, end_position_on_the_genomic_accession=5969761, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33833, CurrentID=0, Status='live', Symbol='Ucp4B', Aliases='Dmel_CG18340, CG18340, DmUCP4B, Dmel\\CG18340, UCP4B, Ucp 4b, Ucp4b, dmUCP4b', description='Uncoupling protein 4B', other_designations='CG18340-PA|CG18340-PB|CG18340-PC|Ucp4B-PA|Ucp4B-PB|Ucp4B-PC|uncoupling protein 4B|uncoupling protein 4b', map_location='26A5-26A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5969859, end_position_on_the_genomic_accession=5971025, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33834, CurrentID=0, Status='live', Symbol='chic', Aliases='Dmel_CG9553, CG9553, Chi, Chic, D88, Dmel\\CG9553, GH4, chi, l(2)27/7, profilin, sand', description='chickadee', other_designations='CG9553-PA|CG9553-PB|CG9553-PC|CG9553-PD|CG9553-PE|CG9553-PF|DProfilin|chic-PA|chic-PB|chic-PC|chic-PD|chic-PE|chic-PF|chicadee|chickaddee|profilin|stranded', map_location='26A5-26B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5972115, end_position_on_the_genomic_accession=5981018, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33835, CurrentID=0, Status='live', Symbol='eIF-4a', Aliases='Dmel_CG9075, 2439, 9075, CG9075, Dm-eIF4A, DmRH2, Dmel\\CG9075, EIF-4A, EIF-4a, EIF4A, EiF4A, Eif-4a, Eif4a, L(2L)162, Su(dpp)YE9, anon-26Aa, anon26A, deIF4A, eIF 4A, eIF-4A-I, eIF4A, eIF4A-I, eIF4AI, eIF4a, eiF4A, eif-4a, eif4A, eif4a, l(2)02439, l(2)162, l(2)26Ac, l(2)gdh-4, l(2)k01501, l(2L)162, pecan, eIF-4a', description='Eukaryotic initiation factor 4a', other_designations='CG9075-PA|CG9075-PB|CG9075-PC|CG9075-PD|CG9075-PE|CG9075-PF|eIF-4a-PA|eIF-4a-PB|eIF-4a-PC|eIF-4a-PD|eIF-4a-PE|eIF-4a-PF|eukaryotic initiation factor 4A|eukaryotic initiation factor 4a|eukaryotic translation initiation factor 4A|translation initiation factor 4A', map_location='26B2-26B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5981764, end_position_on_the_genomic_accession=5985909, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33836, CurrentID=0, Status='live', Symbol='ifc', Aliases='Dmel_CG9078, CG9078, DES-1, Des1, Dmel\\CG9078, Ifc, des, des-1, des1, l(2)26Ab, l(2)gdh-2, ms(2)04875', description='infertile crescent', other_designations='CG9078-PA|CG9078-PB|degenerating spermatocyte|degenerative spermatocyte|ifc-PA|ifc-PB|sphingolipid delta 4 desaturase', map_location='2-20.6 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5986348, end_position_on_the_genomic_accession=5988854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33837, CurrentID=0, Status='live', Symbol='lid', Aliases='Dmel_CG9088, CG9088, Dmel\\CG9088, KDM5, LID, Lid, dJARID1, dKDM5/LID, l(2)10424', description='little imaginal discs', other_designations='CG9088-PA|CG9088-PB|CG9088-PC|CG9088-PD|CG9088-PE|CG9088-PF|lid-PA|lid-PB|lid-PC|lid-PD|lid-PE|lid-PF|little imaginal disks', map_location='26B2-26B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5989400, end_position_on_the_genomic_accession=5999501, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33838, CurrentID=0, Status='live', Symbol='Tsp26A', Aliases='Dmel_CG9093, CG9093, Dm.Tsp26A, Dmel\\CG9093', description='Tetraspanin 26A', other_designations='CG9093-PB|CG9093-PC|CG9093-PD|Tsp26A-PB|Tsp26A-PC|Tsp26A-PD|tetraspanin 26A', map_location='26B2-26B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6000228, end_position_on_the_genomic_accession=6003738, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33839, CurrentID=0, Status='live', Symbol='Gal', Aliases='Dmel_CG9092, CG9092, Dmel\\CG9092, beta-GAL, beta-Gal-1, beta-gal, gal, lacZ-1', description='beta galactosidase', other_designations='CG9092-PA|CG9092-PB|Gal-PA|Gal-PB|beta-galactosidase-1', map_location='26B2-26B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6003574, end_position_on_the_genomic_accession=6006827, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33840, CurrentID=0, Status='live', Symbol='CG9098', Aliases='Dmel_ Dmel\\CG9098', description='CG9098 gene product from transcript CG9098-RC', other_designations='CG9098-PA|CG9098-PB|CG9098-PC', map_location='26B2-26B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6007364, end_position_on_the_genomic_accession=6012338, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33841, CurrentID=0, Status='live', Symbol='H2.0', Aliases='Dmel_CG11607, CG11607, Dmel\\CG11607', description='Homeodomain protein 2.0', other_designations='CG11607-PA|H2.0-PA|homeodomain protein 2.0', map_location='26B2-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6013059, end_position_on_the_genomic_accession=6019402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33842, CurrentID=0, Status='live', Symbol='CG13996', Aliases='Dmel_ Dmel\\CG13996', description='CG13996 gene product from transcript CG13996-RA', other_designations='CG13996-PA|uncharacterized protein', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6020835, end_position_on_the_genomic_accession=6021718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33843, CurrentID=0, Status='live', Symbol='CG9107', Aliases='Dmel_ Dmel\\CG9107, Rrp7', description='CG9107 gene product from transcript CG9107-RA', other_designations='CG9107-PA|CG9107-PB', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6022951, end_position_on_the_genomic_accession=6024539, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33844, CurrentID=0, Status='live', Symbol='CG9109', Aliases='Dmel_ Dmel\\CG9109, dC1GalT9', description='CG9109 gene product from transcript CG9109-RA', other_designations='CG9109-PA|CG9109-PC', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6025870, end_position_on_the_genomic_accession=6034423, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33845, CurrentID=0, Status='live', Symbol='mtm', Aliases='Dmel_CG9115, CG9115, Dmel\\CG9115, MTM1, dMTMH1', description='myotubularin', other_designations='CG9115-PA|CG9115-PB|mtm-PA|mtm-PB', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6034731, end_position_on_the_genomic_accession=6037015, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33846, CurrentID=0, Status='live', Symbol='CG9117', Aliases='Dmel_ Dmel\\CG9117', description='CG9117 gene product from transcript CG9117-RA', other_designations='CG9117-PA', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6036997, end_position_on_the_genomic_accession=6037902, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33847, CurrentID=0, Status='live', Symbol='ade2', Aliases='Dmel_CG9127, 153122_at, Ade2, CG9127, Dmel\\CG9127, Paics, pym', description='adenosine 2', other_designations='CG9127-PA|CG9127-PB|CG9127-PC|ade2-PA|ade2-PB|ade2-PC|adenosine2', map_location='2-17.7 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6041181, end_position_on_the_genomic_accession=6046050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33849, CurrentID=0, Status='live', Symbol='CG12393', Aliases='Dmel_ Dmel\\CG12393', description='CG12393 gene product from transcript CG12393-RB', other_designations='CG12393-PA|CG12393-PB|CG12393-PD|uncharacterized protein', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6048002, end_position_on_the_genomic_accession=6050474, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33850, CurrentID=0, Status='live', Symbol='CG9135', Aliases='Dmel_ Dmel\\CG9135', description='CG9135 gene product from transcript CG9135-RD', other_designations='CG9135-PA|CG9135-PB|CG9135-PC|CG9135-PD', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6050800, end_position_on_the_genomic_accession=6053877, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33851, CurrentID=0, Status='live', Symbol='CG13995', Aliases='Dmel_ Dmel\\CG13995, anon-WO0170980.169, anon-WO0170980.170', description='CG13995 gene product from transcript CG13995-RA', other_designations='CG13995-PA|uncharacterized protein', map_location='26B3-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6054302, end_position_on_the_genomic_accession=6060169, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33852, CurrentID=0, Status='live', Symbol='ND-51', Aliases='Dmel_CG9140, CG9140, Dmel\\CG9140, ND51, NUBM', description='NADH dehydrogenase (ubiquinone) 51 kDa subunit', other_designations='CG9140-PA|CG9140-PB|FBpp0078847|NADH dehydrogenase|NADH ubiquinone oxidoreductase 51 kDa|NADH-UBIQUINONE OXIDOREDUCTASE 51 KDA SUBUNIT|NADH:ubiquinone oxidoreductase 51 kDa subunit|NADH:ubiquinone oxidoreductase, 51kDa subunit|ND-51-PA|ND-51-PB|ubiquinone', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6060842, end_position_on_the_genomic_accession=6062757, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33853, CurrentID=0, Status='live', Symbol='CG13994', Aliases='Dmel_ BcDNA:AT31406, Dmel\\CG13994', description='CG13994 gene product from transcript CG13994-RA', other_designations='CG13994-PA|CG13994-PB|uncharacterized protein', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6062989, end_position_on_the_genomic_accession=6063671, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33854, CurrentID=0, Status='live', Symbol='Fbw5', Aliases='Dmel_CG9144, CG9144, Dmel\\CG9144, FBXW2, FBgn 31773, dFbw5', description='F-box and WD repeat domain containing 5', other_designations='CG9144-PA|Fbw5-PA', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6063562, end_position_on_the_genomic_accession=6065953, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33855, CurrentID=0, Status='live', Symbol='CG9147', Aliases='Dmel_ Dmel\\CG9147', description='CG9147 gene product from transcript CG9147-RD', other_designations='CG9147-PA|CG9147-PB|CG9147-PC|CG9147-PD', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6065997, end_position_on_the_genomic_accession=6067635, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33856, CurrentID=0, Status='live', Symbol='CG9150', Aliases='Dmel_ Dmel\\CG9150', description='CG9150 gene product from transcript CG9150-RB', other_designations='CG9150-PB', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6067668, end_position_on_the_genomic_accession=6068594, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33857, CurrentID=0, Status='live', Symbol='CG13993', Aliases='Dmel_ Dmel\\CG13993, anon-EST:Posey288', description='CG13993 gene product from transcript CG13993-RA', other_designations='CG13993-PA|CG13993-PB|uncharacterized protein', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6069079, end_position_on_the_genomic_accession=6069838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33858, CurrentID=0, Status='live', Symbol='CG9154', Aliases='Dmel_ Dmel\\CG9154', description='CG9154 gene product from transcript CG9154-RB', other_designations='CG9154-PA|CG9154-PB', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6069790, end_position_on_the_genomic_accession=6070686, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33859, CurrentID=0, Status='live', Symbol='Kr-h2', Aliases='Dmel_CG9159, CG9159, Dmel\\CG9159, Kr-H2, Kr-h, anon-26Ba', description='Kruppel homolog 2', other_designations='CG9159-PA|CG9159-PB|CG9159-PC|Kr-h2-PA|Kr-h2-PB|Kr-h2-PC|anon-26Ba|kruppel homolog|kruppel homolog 2', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6070990, end_position_on_the_genomic_accession=6073961, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33861, CurrentID=0, Status='live', Symbol='Kr-h1', Aliases='Dmel_CG45074, 44/11, CG18783, CG45074, CG9167, Dmel\\CG45074, Dmel_CG18783, Kr h, Kr-H1, Kr-h, P381, Pow, l(2)05208, l(2)10642', description='Kruppel homolog 1', other_designations='CG45074-PA|CG45074-PB|Kr-h1-PA|Kr-h1-PB|Krueppel-homolog alpha-isoform|Kruppel-homolog 1|Kruppel-homolog-1|kruppel homolog 1|krupple homolog 1|passion of white', map_location='26B5-26B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6082603, end_position_on_the_genomic_accession=6096498, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33862, CurrentID=0, Status='live', Symbol='CG9175', Aliases='Dmel_ Dmel\\CG9175, Sec12', description='CG9175 gene product from transcript CG9175-RB', other_designations='CG9175-PA|CG9175-PB', map_location='26B5-26B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6096698, end_position_on_the_genomic_accession=6098966, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33863, CurrentID=0, Status='live', Symbol='stai', Aliases='Dmel_CG31641, BEST:LD33989, CG31641, CG5981, D-stat, Dmel\\CG31641, LD33989, SCG10-like, rdt, rdtp', description='stathmin', other_designations='CG31641-PA|CG31641-PB|CG31641-PC|CG31641-PD|CG31641-PE|D-stathmin|redtape|stai-PA|stai-PB|stai-PC|stai-PD|stai-PE', map_location='26B5-26B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6100377, end_position_on_the_genomic_accession=6124653, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33864, CurrentID=0, Status='live', Symbol='Arpc4', Aliases='Dmel_CG5972, ARPC4, Arc-p20, Arp-p20, BcDNA:RE43724, CG5972, Dmel\\CG5972, P-20 ARC, anon-EST:Posey49, arc-p20, p20, p20Arc', description='Actin-related protein 2/3 complex, subunit 4', other_designations='Arc-p20|Arp2/3 complex (p20 subunit)|Arp2/3 subunit p20|Arpc4-PA|CG5972-PA|p20 of Arp2/3 complex|p20 subunit of Arp2/3', map_location='26B7-26B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6125132, end_position_on_the_genomic_accession=6125911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33865, CurrentID=0, Status='live', Symbol='WDR79', Aliases='Dmel_CG9226, CG9226, Dmel\\CG9226, dWDR79, p70', description='WD repeat domain 79 homolog', other_designations='CG9226-PB|CG9226-PC|WDR79-PB|WDR79-PC', map_location='26B7-26B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6125879, end_position_on_the_genomic_accession=6127775, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33866, CurrentID=0, Status='live', Symbol='tectonic', Aliases='Dmel_CG42731, CG42731, CG9227, Dmel\\CG42731, Dmel_CG9227, Tectonic', description='CG42731 gene product from transcript CG42731-RA', other_designations='CG42731-PA|tectonic-PA', map_location='26B7-26B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6127872, end_position_on_the_genomic_accession=6130651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33867, CurrentID=0, Status='live', Symbol='CG9222', Aliases='Dmel_ Dmel\\CG9222', description='CG9222 gene product from transcript CG9222-RA', other_designations='CG9222-PA', map_location='26B8-26B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6131023, end_position_on_the_genomic_accession=6132484, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33868, CurrentID=0, Status='live', Symbol='CG13991', Aliases='Dmel_ Dmel\\CG13991', description='CG13991 gene product from transcript CG13991-RA', other_designations='CG13991-PA|uncharacterized protein', map_location='26B8-26B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6139382, end_position_on_the_genomic_accession=6145010, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33869, CurrentID=0, Status='live', Symbol='CG13989', Aliases='Dmel_ BcDNA:AT09383, Dmel\\CG13989', description='CG13989 gene product from transcript CG13989-RA', other_designations='CG13989-PA|uncharacterized protein', map_location='26B9-26B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6160400, end_position_on_the_genomic_accession=6161309, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33874, CurrentID=0, Status='live', Symbol='AANATL2', Aliases='Dmel_CG9486, CG9486, Dmel\\CG9486', description='Arylalkylamine N-acyltransferase-like 2', other_designations='AANATL2-PA|AANATL2-PB|CG9486-PA|CG9486-PB|arylalkylamine N-acyltransferase-like 2', map_location='26C1-26C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6255859, end_position_on_the_genomic_accession=6256822, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33875, CurrentID=0, Status='live', Symbol='CG13983', Aliases='Dmel_ Dmel\\CG13983', description='CG13983 gene product from transcript CG13983-RA', other_designations='CG13983-PA|uncharacterized protein', map_location='26C1-26C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6263557, end_position_on_the_genomic_accession=6264490, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33879, CurrentID=0, Status='live', Symbol='CG13984', Aliases='Dmel_ Dmel\\CG13984', description='CG13984 gene product from transcript CG13984-RA', other_designations='CG13984-PA|uncharacterized protein', map_location='26C2-26C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6308819, end_position_on_the_genomic_accession=6309475, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33881, CurrentID=0, Status='live', Symbol='PDZ-GEF', Aliases='Dmel_CG9491, CG9491, D-PDZGEF, DGEF, Dmel\\CG9491, Dzy, EC2-8, GEF, Gef26, PDF-GEF Dizzy, dPDZ-GEF, dizzy, dzy, l(2)k13720', description='PDZ domain-containing guanine nucleotide exchange factor', other_designations='CG9491-PA|CG9491-PB|CG9491-PC|CG9491-PD|GEF|Gef26|PDZ-GEF-PA|PDZ-GEF-PB|PDZ-GEF-PC|PDZ-GEF-PD|PDZ-GEF/Dizzy|dizzy', map_location='26C2-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6323767, end_position_on_the_genomic_accession=6332789, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33882, CurrentID=0, Status='live', Symbol='Pez', Aliases='Dmel_CG9493, CG9493, Dmel\\CG9493, PTPD1, d', description='CG9493 gene product from transcript CG9493-RB', other_designations='CG9493-PA|CG9493-PB|Pez-PA|Pez-PB', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6332881, end_position_on_the_genomic_accession=6338671, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33883, CurrentID=0, Status='live', Symbol='Cpr', Aliases='Dmel_CG11567, CCR, CG11567, CPR, DMR, DmCPR, Dmel\\CG11567, NCPR, P450, cpr', description='Cytochrome P450 reductase', other_designations='CG11567-PA|CG11567-PB|CG11567-PC|CG11567-PD|Cpr-PA|Cpr-PB|Cpr-PC|Cpr-PD|NADPH--ferrihemoprotein reductase; NADPH-cytochrome P450 reductase|NADPH-cytochrome P450 oxidoreductase|NADPH-cytochrome P450 reductase|P450 reductase|cytochrome P450 reductase|cytochrome c reductase (NADPH)', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6338836, end_position_on_the_genomic_accession=6346280, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33884, CurrentID=0, Status='live', Symbol='CG9497', Aliases='Dmel_ Dmel\\CG9497', description='CG9497 gene product from transcript CG9497-RB', other_designations='CG9497-PA|CG9497-PB', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6346430, end_position_on_the_genomic_accession=6348427, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33885, CurrentID=0, Status='live', Symbol='CG9498', Aliases='Dmel_ Dmel\\CG9498', description='CG9498 gene product from transcript CG9498-RA', other_designations='CG9498-PA', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6349393, end_position_on_the_genomic_accession=6351034, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33886, CurrentID=0, Status='live', Symbol='ppk7', Aliases='Dmel_CG9499, CG9499, Dmel\\CG9499, PPK7', description='pickpocket 7', other_designations='CG9499-PA|ppk7-PA', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6352324, end_position_on_the_genomic_accession=6354331, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33887, CurrentID=0, Status='live', Symbol='ppk14', Aliases='Dmel_CG9501, CG9501, Dmel\\CG9501, PPK14', description='pickpocket 14', other_designations='CG9501-PB|ppk14-PB', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6354956, end_position_on_the_genomic_accession=6356672, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33888, CurrentID=0, Status='live', Symbol='CG9500', Aliases='Dmel_ AAF52372, Dmel\\CG9500, FBpp0078910', description='CG9500 gene product from transcript CG9500-RA', other_designations='CG9500-PA|Fibrinogen-related protein', map_location='26C3-26C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6357005, end_position_on_the_genomic_accession=6358318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33889, CurrentID=0, Status='live', Symbol='CG9505', Aliases='Dmel_ Dmel\\CG9505', description='CG9505 gene product from transcript CG9505-RA', other_designations='CG9505-PA', map_location='26C4-26C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6367882, end_position_on_the_genomic_accession=6370190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33890, CurrentID=0, Status='live', Symbol='slam', Aliases='Dmel_CG9506, CG9506, Dmel\\CG9506, Slam, wdo', description='slow as molasses', other_designations='CG9506-PB|CG9506-PC|slam-PB|slam-PC|waldo', map_location='26C4-26C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6372388, end_position_on_the_genomic_accession=6377284, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33892, CurrentID=0, Status='live', Symbol='CG9507', Aliases='Dmel_ Dmel\\CG9507', description='CG9507 gene product from transcript CG9507-RA', other_designations='CG9507-PA', map_location='26C4-26C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6381266, end_position_on_the_genomic_accession=6383419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33893, CurrentID=0, Status='live', Symbol='CG42370', Aliases='Dmel_ CG13981, CG33295, CG9508, CG9511, CT26920, Dmel\\CG42370, Dmel_CG9508, Dmel_CG9511', description='CG42370 gene product from transcript CG42370-RA', other_designations='CG42370-PA|Dmel_CG42370|uncharacterized protein', map_location='26C4-26D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6383435, end_position_on_the_genomic_accession=6385512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33894, CurrentID=0, Status='live', Symbol='CG42369', Aliases='Dmel_ CG13981, CG33295, CG9508, CG9511, CT26922, Dmel\\CG42369, Dmel_CG9508, Dmel_CG9511', description='CG42369 gene product from transcript CG42369-RA', other_designations='CG42369-PA|uncharacterized protein', map_location='26D1-26D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6395853, end_position_on_the_genomic_accession=6397249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33895, CurrentID=0, Status='live', Symbol='CG13982', Aliases='Dmel_ Dmel\\CG13982', description='CG13982 gene product from transcript CG13982-RA', other_designations='CG13982-PA|CG13982-PB|uncharacterized protein', map_location='26D1-26D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6412315, end_position_on_the_genomic_accession=6415335, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33896, CurrentID=0, Status='live', Symbol='Tig', Aliases='Dmel_CG11527, CG11527, CT36389, Dmel\\CG11527, TIG, tig', description='Tiggrin', other_designations='CG11527-PA|CG11527-PB|Tig-PA|Tig-PB|tiggrin', map_location='26D1-26D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6415179, end_position_on_the_genomic_accession=6423310, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33897, CurrentID=0, Status='live', Symbol='Fic', Aliases='Dmel_CG9523, CG9523, Dmel\\CG9523, l(2)k11101', description='Fic domain-containing protein', other_designations='CG9523-PA|Fic-PA|lethal (2) k11101', map_location='28D10-28D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6424336, end_position_on_the_genomic_accession=6426371, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33898, CurrentID=0, Status='live', Symbol='CG9527', Aliases='Dmel_ Dmel\\CG9527, LCAD', description='CG9527 gene product from transcript CG9527-RD', other_designations='CG9527-PA|CG9527-PB|CG9527-PC|CG9527-PD|long chain acyl-CoA-dehydrogenase', map_location='26D1-26D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6426724, end_position_on_the_genomic_accession=6448403, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33899, CurrentID=0, Status='live', Symbol='retm', Aliases='Dmel_CG9528, CG9528, Dmel\\CG9528', description='real-time', other_designations='CG9528-PA|CG9528-PB|CG9528-PC|retm-PA|retm-PB|retm-PC', map_location='26D4-26D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6448464, end_position_on_the_genomic_accession=6455929, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33900, CurrentID=0, Status='live', Symbol='frj', Aliases='Dmel_CG9526, CG9526, CT26954, Dmel\\CG9526, Frj, MBOAT', description='farjavit', other_designations='CG9526-PA|CG9526-PB|frj-PA|frj-PB', map_location='26D5-26D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6456052, end_position_on_the_genomic_accession=6459008, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33901, CurrentID=0, Status='live', Symbol='CG16947', Aliases='Dmel_ Dmel\\CG16947, anon-EST:Posey20', description='CG16947 gene product from transcript CG16947-RA', other_designations='CG16947-PA|CG16947-PB|uncharacterized protein', map_location='26D5-26D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6459015, end_position_on_the_genomic_accession=6465669, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33902, CurrentID=0, Status='live', Symbol='CG9531', Aliases='Dmel_ Dmel\\CG9531', description='CG9531 gene product from transcript CG9531-RA', other_designations='CG9531-PA|CG9531-PB', map_location='26D7-26D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6466217, end_position_on_the_genomic_accession=6467698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33903, CurrentID=0, Status='live', Symbol='mmy', Aliases='Dmel_CG9535, BcDNA:LD24639, CG9535, Dmel\\CG9535, LD44264, anon-WO0140519.242, cao, cys, cyst, cystic, l(2)k13717b, mum', description='mummy', other_designations='CG9535-PA|CG9535-PB|CG9535-PD|UDP-GlcNAc pyrophosphorylase|cabrio|cystic|mmy-PA|mmy-PB|mmy-PD', map_location='2-16 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6468538, end_position_on_the_genomic_accession=6473904, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33904, CurrentID=0, Status='live', Symbol='CG9536', Aliases='Dmel_ Dmel\\CG9536', description='CG9536 gene product from transcript CG9536-RB', other_designations='CG9536-PA|CG9536-PB', map_location='26D7-26D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6474438, end_position_on_the_genomic_accession=6476975, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33905, CurrentID=0, Status='live', Symbol='Sec61alpha', Aliases='Dmel_CG9539, 0455, BEST:LD29847, BcDNA:LD29847, CG9539, DSec61alpha, Dmel\\CG9539, E(br)165, LD29847, SEC61-alpha, SEC61ALPHA, Sec61, Sec61-alpha, Sec61a, anon-WO0118547.195, anon-WO0118547.277, dSEC61A, dSec61alpha, endd2, l(2)k04917, sec61, sec61alpha', description='Sec61 alpha subunit', other_designations='CG9539-PA|CG9539-PB|E(br)165|Sec61alpha-PA|Sec61alpha-PB|effector for neural death and degeneration 2', map_location='26D7-26D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6477182, end_position_on_the_genomic_accession=6480575, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33906, CurrentID=0, Status='live', Symbol='Daxx', Aliases='Dmel_CG9537, CG9537, DLP, Dmel\\CG9537', description='Daxx-like protein', other_designations='CG9537-PA|CG9537-PB|Daxx-PA|Daxx-PB', map_location='26D8-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6480929, end_position_on_the_genomic_accession=6487576, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33907, CurrentID=0, Status='live', Symbol='KFase', Aliases='Dmel_CG9542, CG9542, Dmel\\CG9542', description='Kynurenine formamidase', other_designations='CG9542-PA|CG9542-PB|KFase-PA|KFase-PB|kynurenine formamidase', map_location='26D9-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6487342, end_position_on_the_genomic_accession=6488756, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33908, CurrentID=0, Status='live', Symbol='epsilonCOP', Aliases='Dmel_CG9543, BcDNA:LD29885, CG9543, Dmel\\CG9543, bcDNA:LD29885, epsilon-COP, epsilon-Cop', description='Coat Protein (coatomer) epsilon', other_designations='CG9543-PA|COPI epsilon subunit|coat protein (coatomer) epsilon|epsilonCOP-PA', map_location='26D9-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6489090, end_position_on_the_genomic_accession=6490194, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33909, CurrentID=0, Status='live', Symbol='CG9548', Aliases='Dmel_ BcDNA:RH02789, Dmel\\CG9548', description='CG9548 gene product from transcript CG9548-RA', other_designations='CG9548-PA', map_location='26D9-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6490155, end_position_on_the_genomic_accession=6490843, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33910, CurrentID=0, Status='live', Symbol='CG31638', Aliases='Dmel_ CG13764, CG9545, Dmel\\CG31638', description='CG31638 gene product from transcript CG31638-RA', other_designations='CG31638-PA|CG31638-PB|uncharacterized protein', map_location='26D9-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6491061, end_position_on_the_genomic_accession=6496988, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  33911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33911, CurrentID=0, Status='live', Symbol='CG9547', Aliases='Dmel_ 152078_at, Dmel\\CG9547', description='CG9547 gene product from transcript CG9547-RA', other_designations='CG9547-PA', map_location='26D9-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6492974, end_position_on_the_genomic_accession=6494532, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33913, CurrentID=0, Status='live', Symbol='CG9550', Aliases='Dmel_ Dmel\\CG9550, d6ST2', description='CG9550 gene product from transcript CG9550-RA', other_designations='CG9550-PA', map_location='26D9-26D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6497053, end_position_on_the_genomic_accession=6498466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33914, CurrentID=0, Status='live', Symbol='CG31637', Aliases='Dmel_ CG13765, CG9551, Dmel\\CG31637, anon-WO0118547.107, d6ST1', description='CG31637 gene product from transcript CG31637-RA', other_designations='CG31637-PA|CG31637-PB|uncharacterized protein', map_location='26D9-26E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6498647, end_position_on_the_genomic_accession=6526996, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33916, CurrentID=0, Status='live', Symbol='eya', Aliases='Dmel_CG9554, 24582246, BcDNA:LD16029, CG9554, CLI, CLIFT, Dmel\\CG9554, EY2-1, EYA, Eya, cli, cli/eya, dEYA, del1, eay, ey-2', description='eyes absent', other_designations='CG9554-PA|CG9554-PB|CG9554-PC|Eye absent|Eyes-absent|clift|delayed gonad formation 1|eya-PA|eya-PB|eya-PC|eyeless 2', map_location='2-17 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6527447, end_position_on_the_genomic_accession=6546976, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33917, CurrentID=0, Status='live', Symbol='Osm6', Aliases='Dmel_CG9595, CG9595, Dmel\\CG9595, IFT52, NGD5, OSM-6, Osm-6, osm-6', description='Osm-6 ortholog (C. elegans)', other_designations='CG9595-PA|Osm-6 ortholog|Osm6-PA|osm-6', map_location='26E3-26E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6555717, end_position_on_the_genomic_accession=6557502, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33918, CurrentID=0, Status='live', Symbol='COX5B', Aliases='Dmel_CG11015, BcDNA:RH49324, CG11015, COXB, CcO Vb, CoVb, Dmel\\CG11015, Vb, anon-EST:Posey143, anon-WO0118547.187', description='Cytochrome c oxidase subunit 5B', other_designations='CG11015-PA|CG11015-PB|COX5B-PA|COX5B-PB|cytochrome c oxidase Vb|cytochrome c oxidase polypeptide Vb|cytochrome c oxidase subunit 5B|cytochrome c oxidase subunit Vb|subunit Vb', map_location='26E3-26E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6557800, end_position_on_the_genomic_accession=6559308, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33919, CurrentID=0, Status='live', Symbol='COX5BL', Aliases='Dmel_CG11043, CG11043, COXB, Dmel\\CG11043, Vb', description='Cytochrome c oxidase subunit 5B-like', other_designations='CG11043-PA|COX5BL-PA|cytochrome c oxidase polypeptide Vb|cytochrome c oxidase subunit 5B-like', map_location='26E3-26E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6559431, end_position_on_the_genomic_accession=6560133, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33920, CurrentID=0, Status='live', Symbol='CG9596', Aliases='Dmel_ Dmel\\CG9596', description='CG9596 gene product from transcript CG9596-RC', other_designations='CG9596-PB|CG9596-PC|CG9596-PD', map_location='26E3-26E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6559426, end_position_on_the_genomic_accession=6562289, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33921, CurrentID=0, Status='live', Symbol='Ent2', Aliases='Dmel_CG31911, BEST:LD04971, CG11045, CG31911, DmENT2, DmEnt2, Dmel\\CG31911, LD04971, P124, ent2', description='Equilibrative nucleoside transporter 2', other_designations='CG31911-PA|CG31911-PB|Ent2-PA|Ent2-PB|P124|equilibrative nucleoside transporter|equilibrative nucleoside transporter 2', map_location='26E3-26E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6562325, end_position_on_the_genomic_accession=6564709, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33922, CurrentID=0, Status='live', Symbol='CG13766', Aliases='Dmel_ Dmel\\CG13766, FBXO16', description='CG13766 gene product from transcript CG13766-RB', other_designations='CG13766-PA|CG13766-PB|uncharacterized protein', map_location='26E3-26E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6564112, end_position_on_the_genomic_accession=6566181, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33923, CurrentID=0, Status='live', Symbol='CG11319', Aliases='Dmel_ Dmel\\CG11319', description='CG11319 gene product from transcript CG11319-RA', other_designations='CG11319-PA|CG11319-PB|CG11319-PC|uncharacterized protein', map_location='26E3-26F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6568400, end_position_on_the_genomic_accession=6601068, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  33924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33924, CurrentID=0, Status='live', Symbol='CG11050', Aliases='Dmel_ BcDNA:LD22339, Dmel\\CG11050', description='CG11050 gene product from transcript CG11050-RA', other_designations='CG11050-PA|CG11050-PB|CG11050-PC|uncharacterized protein', map_location='26F1-26F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6611011, end_position_on_the_genomic_accession=6617117, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33925, CurrentID=0, Status='live', Symbol='DIP-iota', Aliases='Dmel_CG11320, 11320, CG11320, CT31597, Dmel\\CG11320', description='Dpr-interacting protein iota', other_designations='CG11320-PB|DIP-iota-PB', map_location='26F1-26F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6617134, end_position_on_the_genomic_accession=6621700, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33927, CurrentID=0, Status='live', Symbol='Oatp26F', Aliases='Dmel_CG31634, CG11332, CG13767, CG18784, CG31634, Dmel\\CG31634, oatp 26F, oatp26F', description='Organic anion transporting polypeptide 26F', other_designations='CG31634-PA|Oatp26F-PA|organic anion transporter protein 26f|organic anion transporting polypeptide 26F', map_location='26F3-26F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6625138, end_position_on_the_genomic_accession=6632407, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33928, CurrentID=0, Status='live', Symbol='CG31635', Aliases='Dmel_ BcDNA:GM03875, CG13768, CG13769, CT33252, Dmel\\CG31635', description='CG31635 gene product from transcript CG31635-RG', other_designations='CG31635-PA|CG31635-PB|CG31635-PC|CG31635-PD|CG31635-PE|CG31635-PF|CG31635-PG|CG31635-PH|uncharacterized protein', map_location='26F3-26F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6632798, end_position_on_the_genomic_accession=6649114, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  33930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33930, CurrentID=0, Status='live', Symbol='Tango1', Aliases='Dmel_CG11098, CG11098, Dmel\\CG11098, TANGO1', description='Transport and Golgi organization 1', other_designations='CG11098-PA|CG11098-PC|CG11098-PD|Tango1-PA|Tango1-PC|Tango1-PD|transport and golgi organisation 1|transport and golgi organization 1', map_location='26F3-26F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6649388, end_position_on_the_genomic_accession=6654574, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33931, CurrentID=0, Status='live', Symbol='CG31633', Aliases='Dmel_ CG13770, Dmel\\CG31633, FBXL17', description='CG31633 gene product from transcript CG31633-RA', other_designations='CG31633-PA|uncharacterized protein', map_location='26F5-26F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6654760, end_position_on_the_genomic_accession=6656321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33932, CurrentID=0, Status='live', Symbol='CG11070', Aliases='Dmel_ BcDNA:LD34475, CG 11070, Dmel\\CG11070, E4 A, UBE4A, UFD2a', description='CG11070 gene product from transcript CG11070-RA', other_designations='AABU01002759.1|CG11070-PA|uncharacterized protein', map_location='26F5-26F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6658517, end_position_on_the_genomic_accession=6662033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33933, CurrentID=0, Status='live', Symbol='CG13771', Aliases='Dmel_ Dmel\\CG13771', description='CG13771 gene product from transcript CG13771-RA', other_designations='CG13771-PA|CG13771-PB|uncharacterized protein', map_location='26F6-26F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6662061, end_position_on_the_genomic_accession=6663612, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33934, CurrentID=0, Status='live', Symbol='cup', Aliases='Dmel_CG11181, CG11181, CUP, Cup, DmCup, Dmel\\CG11181, fs(1) fs(2)cup, i239', description='CG11181 gene product from transcript CG11181-RC', other_designations='CG11181-PB|CG11181-PC|Cup|cup-PB|cup-PC', map_location='2-23 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6663664, end_position_on_the_genomic_accession=6674780, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  33935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33935, CurrentID=0, Status='live', Symbol='Galt', Aliases='Dmel_CG9232, CG9232, Dmel\\CG9232, GALT, dGALT', description='Galactose-1-phosphate uridylyltransferase', other_designations='CG9232-PA|CG9232-PB|Galt-PA|Galt-PB|galactose-1-phosphate uridylyltransferase|galactose-1P uridylyltransferase', map_location='26F6-26F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6666264, end_position_on_the_genomic_accession=6668033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33937, CurrentID=0, Status='live', Symbol='cort', Aliases='Dmel_CG11330, CG11330, Cort, Dmel\\CG11330, cor', description='cortex', other_designations='CG11330-PA|CG11330-PB|cdc20-like|cort-PA|cort-PB', map_location='2-49 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6674682, end_position_on_the_genomic_accession=6676635, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  33938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33938, CurrentID=0, Status='live', Symbol='Nse1', Aliases='Dmel_CG11329, CG11329, Dmel\\CG11329, EST2, GM14348, NSE1, anon-WO0118547.82', description='Non-SMC element 1', other_designations='CG11329-PA|CG11329-PB|Nse1-PA|Nse1-PB', map_location='26F7-26F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6676649, end_position_on_the_genomic_accession=6677625, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33939, CurrentID=0, Status='live', Symbol='Nhe3', Aliases='Dmel_CG11328, CG11328, DmNHE3, Dmel\\CG11328, NHE3, nhe3', description='Na[+]/H[+] hydrogen exchanger 3', other_designations='CG11328-PD|CG11328-PE|CG11328-PG|CG11328-PH|CG11328-PI|CG11328-PJ|Nhe3-PD|Nhe3-PE|Nhe3-PG|Nhe3-PH|Nhe3-PI|Nhe3-PJ', map_location='26F7-26F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6677835, end_position_on_the_genomic_accession=6684364, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  33940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33940, CurrentID=0, Status='live', Symbol='CG11327', Aliases='Dmel_ Dmel\\CG11327', description='CG11327 gene product from transcript CG11327-RA', other_designations='CG11327-PA|uncharacterized protein', map_location='26F7-26F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6684842, end_position_on_the_genomic_accession=6686712, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33941, CurrentID=0, Status='live', Symbol='Tsp', Aliases='Dmel_CG11326, CG11326, CT31613, D-TSP, DTSP, Dmel\\CG11326, SP295, tsp', description='Thrombospondin', other_designations='CG11326-PA|CG11326-PC|CG11326-PD|CG11326-PE|CG11326-PF|CG11326-PG|CG11326-PH|Tsp-PA|Tsp-PC|Tsp-PD|Tsp-PE|Tsp-PF|Tsp-PG|Tsp-PH|thrombospondin', map_location='26F7-27A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6686791, end_position_on_the_genomic_accession=6709086, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  33942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33942, CurrentID=0, Status='live', Symbol='AkhR', Aliases='Dmel_CG11325, AKHR, Akhr, BEST:GH19447, CG11325, DAKHR, DAKHR-1, DGRHR, Dmel\\CG11325, GRHR, GRHR/CG11325, GnRHR, akhr', description='Adipokinetic hormone receptor', other_designations='AKH receptor|AkhR-PA|AkhR-PB|AkhR-PC|AkhR-PD|CG11325-PA|CG11325-PB|CG11325-PC|CG11325-PD|Drm-AKH receptor-1|adipokinetic hormone receptor|gonadotrophin-releasing hormone receptor|gonadotropin-releasing hormone receptor|hormone receptor', map_location='27A1-27A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6711286, end_position_on_the_genomic_accession=6716186, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33943, CurrentID=0, Status='live', Symbol='Aatf', Aliases='Dmel_CG11188, AATF, CG11188, Che-1, Dmel\\CG11188, anon-WO0118547.190, nutmeg', description='apoptosis antagonizing transcription factor ortholog', other_designations='Aatf-PA|CG11188-PA|nutmeg', map_location='27A1-27A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6717625, end_position_on_the_genomic_accession=6719389, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33944, CurrentID=0, Status='live', Symbol='homer', Aliases='Dmel_CG11324, BEST:LD03829, CG11324, Dhom, D Dm Homer, Dmel\\CG11324, Dvh, Hom, Homer, LD03829, dhom', description='CG11324 gene product from transcript CG11324-RD', other_designations='CG11324-PA|CG11324-PB|CG11324-PD|CG11324-PE|CG11324-PF|CG11324-PG|d-Homer|drosophila Ves-1 homolog|homer-PA|homer-PB|homer-PD|homer-PE|homer-PF|homer-PG', map_location='27A1-27A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6719394, end_position_on_the_genomic_accession=6723359, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  33946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33946, CurrentID=0, Status='live', Symbol='TTLL3B', Aliases='Dmel_CG11201, CG11201, Dmel\\CG11201, dmTTLL3B', description='Tubulin tyrosine ligase-like 3B', other_designations='CG11201-PB|TTLL3B-PB|tubulin tyrosine ligase-like 3B', map_location='27A2-27A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6732866, end_position_on_the_genomic_accession=6736149, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33947, CurrentID=0, Status='live', Symbol='TTLL3A', Aliases='Dmel_CG11323, CG11323, Dmel\\CG11323, dmTTLL3A', description='Tubulin tyrosine ligase-like 3A', other_designations='CG11323-PA|TTLL3A-PA|tubulin tyrosine ligase-like 3A', map_location='27A2-27A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6736364, end_position_on_the_genomic_accession=6740188, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33948, CurrentID=0, Status='live', Symbol='CG11221', Aliases='Dmel_ Dmel\\CG11221, PKN', description='CG11221 gene product from transcript CG11221-RB', other_designations='CG11221-PA|CG11221-PB|uncharacterized protein', map_location='27A2-27B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6742502, end_position_on_the_genomic_accession=6756133, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33949, CurrentID=0, Status='live', Symbol='CG11322', Aliases='Dmel_ Dmel\\CG11322', description='CG11322 gene product from transcript CG11322-RB', other_designations='CG11322-PB|uncharacterized protein', map_location='27B1-27B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6757864, end_position_on_the_genomic_accession=6759628, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33950, CurrentID=0, Status='live', Symbol='CG11321', Aliases='Dmel_ AAF52435.2, Dmel\\CG11321', description='CG11321 gene product from transcript CG11321-RG', other_designations='CG11321-PC|CG11321-PD|CG11321-PE|CG11321-PF|CG11321-PG|CG11321-PH|CG11321-PI|CG11321-PJ|CG11321-PK|uncharacterized protein', map_location='27B1-27B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6759829, end_position_on_the_genomic_accession=6772755, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  33951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33951, CurrentID=0, Status='live', Symbol='CG17378', Aliases='Dmel_ Dmel\\CG17378', description='CG17378 gene product from transcript CG17378-RA', other_designations='CG17378-PA|CG17378-PB|CG17378-PC|uncharacterized protein', map_location='27B1-27B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6774041, end_position_on_the_genomic_accession=6778444, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  33952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33952, CurrentID=0, Status='live', Symbol='nrv1', Aliases='Dmel_CG9258, 42kDa, CG9258, Dmel\\CG9258, K-ATPase, NKA-alpha, Na, NaK ATPase, Na[+]/K[+] ATPase, Na[+]/K[+]-ATPase, Na[+]/K[+]ATPase, Nrv, Nrv1, nervana', description='nervana 1', other_designations='CG9258-PA|CG9258-PB|Nervana1|nervana|nervous system antigen 1|nirvana|nirvana 1|nrv1-PA|nrv1-PB', map_location='27B1-27B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6778555, end_position_on_the_genomic_accession=6786861, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33953, CurrentID=0, Status='live', Symbol='nrv2', Aliases='Dmel_CG9261, 42kDa, CG9261, Dmel\\CG9261, K-ATPase, NEST:bs13c10, Na, NaK ATPase, Na[+]/K[+] ATPase, Na[+]/K[+]-ATPase, Na[+]/K[+]ATPase, Nrv, Nrv2, Nrv2.2, ecta, l(2)SH0329, l(2)SH2 0329, l(2)k04223, l(2)k13315.2, nrv2', description='nervana 2', other_designations='CG9261-PA|CG9261-PC|CG9261-PD|CG9261-PE|CG9261-PF|CG9261-PG|Na+/K+ ATPase|Na[+]/K[+] ATPase|ectatic|nervana|nervana2|nirvana|nirvana2|nrv2-PA|nrv2-PC|nrv2-PD|nrv2-PE|nrv2-PF|nrv2-PG', map_location='27B1-27B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6789770, end_position_on_the_genomic_accession=6798862, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  33954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33954, CurrentID=0, Status='live', Symbol='CG17377', Aliases='Dmel_ BEST:LP06279, Dmel\\CG17377, anon-WO0140519.32', description='CG17377 gene product from transcript CG17377-RF', other_designations='CG17377-PB|CG17377-PD|CG17377-PE|CG17377-PF|uncharacterized protein', map_location='27B3-27B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6801358, end_position_on_the_genomic_accession=6802840, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33955, CurrentID=0, Status='live', Symbol='CG11236', Aliases='Dmel_ Dmel\\CG11236', description='CG11236 gene product from transcript CG11236-RA', other_designations='CG11236-PA|CG11236-PB|uncharacterized protein', map_location='27B3-27B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6803011, end_position_on_the_genomic_accession=6804372, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33956, CurrentID=0, Status='live', Symbol='CG17375', Aliases='Dmel_ Dmel\\CG17375', description='CG17375 gene product from transcript CG17375-RD', other_designations='CG17375-PB|CG17375-PC|CG17375-PD|uncharacterized protein', map_location='27B3-27B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6804324, end_position_on_the_genomic_accession=6805792, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33957, CurrentID=0, Status='live', Symbol='sens-2', Aliases='Dmel_CG31632, CG11243, CG17373, CG31632, D.m Sens-2, Dmel\\CG31632, Sens-2, sens2', description='senseless-2', other_designations='CG31632-PA|CG31632-PB|sens-2-PA|sens-2-PB', map_location='27B4-27B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6821934, end_position_on_the_genomic_accession=6838995, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33959, CurrentID=0, Status='live', Symbol='Rca1', Aliases='Dmel_CG10800, 152549_at, CG10800, CLD4, Dmel\\CG10800, Emi1, FBXO5, P3300, RCA1, S(rux)2A, l(2)03300, rca1', description='Regulator of cyclin A1', other_designations='CG10800-PA|Rca1-PA|regulator of cyclin A|regulator of cyclin A 1|regulator of cyclin A1|suppressor of rux 2A', map_location='27C1-27C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6852315, end_position_on_the_genomic_accession=6853866, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33960, CurrentID=0, Status='live', Symbol='l(2)k09022', Aliases='Dmel_CG10805, CG10805, Dmel\\CG10805, anon-WO0118547.160', description='lethal (2) k09022', other_designations='CG10805-PA|l(2)k09022-PA', map_location='27C1-27C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6854142, end_position_on_the_genomic_accession=6860849, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33961, CurrentID=0, Status='live', Symbol='Nha1', Aliases='Dmel_CG10806, CG10806, DmNHA1, Dmel\\CG10806, nha1', description='Na[+]/H[+] hydrogen antiporter 1', other_designations='CG10806-PA|CG10806-PB|Nha1-PA|Nha1-PB|serine protease', map_location='27C1-27C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6860905, end_position_on_the_genomic_accession=6870389, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33962, CurrentID=0, Status='live', Symbol='Nlg2', Aliases='Dmel_CG13772, CG13772, DNlg2, Dmel\\CG13772, dnl2, dnlg2, neuroligin, nl2', description='Neuroligin 2', other_designations='CG13772-PA|CG13772-PB|CG13772-PC|Nlg2-PA|Nlg2-PB|Nlg2-PC|neuroligin|neuroligin 2', map_location='27C3-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6874900, end_position_on_the_genomic_accession=6906656, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  33963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33963, CurrentID=0, Status='live', Symbol='CG13773', Aliases='Dmel_ BEST:GH23590, Dmel\\CG13773', description='CG13773 gene product from transcript CG13773-RA', other_designations='CG13773-PA|uncharacterized protein', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6906666, end_position_on_the_genomic_accession=6907844, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33964, CurrentID=0, Status='live', Symbol='Rat1', Aliases='Dmel_CG10354, CG10354, Dmel\\CG10354, RAT1, XRN2, Xrn2, dRAT1, dhp1', description='CG10354 gene product from transcript CG10354-RA', other_designations='CG10354-PA|CG10354-PB|Rat1-PA|Rat1-PB', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6908083, end_position_on_the_genomic_accession=6911423, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  33965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33965, CurrentID=0, Status='live', Symbol='Wee1', Aliases='Dmel_CG4488, CG4488, Dmel\\CG4488, Dwee, Dwee1, WEE1, Wee, Wee 1, dWee1, dwee1, wee, wee1', description='Wee1 kinase', other_designations='CG4488-PA|CG4488-PB|Wee1 ortholog|Wee1-PA|Wee1-PB|wee', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6911090, end_position_on_the_genomic_accession=6914138, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33966, CurrentID=0, Status='live', Symbol='nop5', Aliases='Dmel_CG10206, 2/30, CG10206, DNOP5, Dmel\\CG10206, D FBgn0026196, NOP5, Nop5', description='CG10206 gene product from transcript CG10206-RA', other_designations='CG10206-PA|nop5-PA', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6916753, end_position_on_the_genomic_accession=6918743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33967, CurrentID=0, Status='live', Symbol='x16', Aliases='Dmel_CG10203, 2/30, 9G8, CG10203, Dmel\\CG10203, D Dxl6, RBP1-like, TN166, X16/9G8, XL6, Xl6, anon-WO0118547.78, d9G8, dx16, dxl6, l(2)k00230, xl6', description='CG10203 gene product from transcript CG10203-RB', other_designations='CG10203-PA|CG10203-PB|x16-PA|x16-PB|xl6', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6914301, end_position_on_the_genomic_accession=6920210, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33968, CurrentID=0, Status='live', Symbol='Hrb27C', Aliases='Dmel_CG10377, 10280, CG10377, Dmel\\CG10377, HRB27C, HRP48, Hrb27-C, Hrb27c, Hrb48, Hrp48, Hrp48/Hrb27C, Ov5, RRM7, Rbp7, hnRNP27C, hnRNP48, hrb27, hrb27C, hrb27c, hrp48, l(2)02647, l(2)10280, l(2)2647, linha, p50, rrm7', description='Heterogeneous nuclear ribonucleoprotein at 27C', other_designations='CG10377-PA|CG10377-PB|CG10377-PC|CG10377-PD|CG10377-PE|CG10377-PF|CG10377-PG|CG10377-PH|Hrb27C-PA|Hrb27C-PB|Hrb27C-PC|Hrb27C-PD|Hrb27C-PE|Hrb27C-PF|Hrb27C-PG|Hrb27C-PH|RNA-binding protein 7|heterogeneous nuclear ribonucleoprotein|heterogeneous nuclear ribonucleoprotein at 27C|linha', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6920704, end_position_on_the_genomic_accession=6932057, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33969, CurrentID=0, Status='live', Symbol='CG18304', Aliases='Dmel_ CG13774, Dmel\\CG18304', description='CG18304 gene product from transcript CG18304-RC', other_designations='CG18304-PA|CG18304-PB|CG18304-PC|CG18304-PD|uncharacterized protein', map_location='27C4-27C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6932545, end_position_on_the_genomic_accession=6942814, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  33971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33971, CurrentID=0, Status='live', Symbol='Fgop2', Aliases='Dmel_CG10158, CG10158, Dmel\\CG10158, FGFR1OP2, FGOP2', description='Fibroblast growth factor receptor 1 oncogene partner 2', other_designations='CG10158-PA|CG10158-PB|Fgop2-PA|Fgop2-PB|fibroblast growth factor receptor 1 oncogene partner 2|fibroblast growth factor receptor 1 oncogene partner 2 ortholog', map_location='27C6-27C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6943178, end_position_on_the_genomic_accession=6945289, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33972, CurrentID=0, Status='live', Symbol='ihog', Aliases='Dmel_CG9211, CG9211, CT26314, Dmel\\CG9211, Ihog, iHog', description='interference hedgehog', other_designations='CG9211-PA|ihog-PA', map_location='27C6-27C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6945464, end_position_on_the_genomic_accession=6948808, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33973, CurrentID=0, Status='live', Symbol='Gas41', Aliases='Dmel_CG9207, CG9207, Dmel\\CG9207, GAS41, d', description='CG9207 gene product from transcript CG9207-RA', other_designations='CG9207-PA|Gas41-PA', map_location='27C6-27C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6949310, end_position_on_the_genomic_accession=6950207, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33974, CurrentID=0, Status='live', Symbol='SA', Aliases='Dmel_CG3423, CG3423, DSA, DSA1, DmSA, Dmel\\CG3423, ESTS:92H2T, SCC3, Scc3', description='Stromalin', other_designations='CG3423-PA|SA-PA|cohesin|stromalin', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6950421, end_position_on_the_genomic_accession=6954586, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  33975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33975, CurrentID=0, Status='live', Symbol='CG13775', Aliases='Dmel_ Dmel\\CG13775', description='CG13775 gene product from transcript CG13775-RB', other_designations='CG13775-PA|CG13775-PB|CG13775-PC|uncharacterized protein', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6954357, end_position_on_the_genomic_accession=6956950, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  33976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33976, CurrentID=0, Status='live', Symbol='CG3430', Aliases='Dmel_ Dmel\\CG3430', description='CG3430 gene product from transcript CG3430-RA', other_designations='CG3430-PA', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6957008, end_position_on_the_genomic_accession=6959104, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  33977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33977, CurrentID=0, Status='live', Symbol='Atac1', Aliases='Dmel_CG9200, CG9200, Dmel\\CG9200, dATAC1', description='Ada2a-containing complex component 1', other_designations='ATAC component 1|Atac1-PA|CG9200-PA', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6959099, end_position_on_the_genomic_accession=6960328, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  33978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33978, CurrentID=0, Status='live', Symbol='CG10399', Aliases='Dmel_ 145934_at, Dmel\\CG10399', description='CG10399 gene product from transcript CG10399-RB', other_designations='CG10399-PA|CG10399-PB|uncharacterized protein', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6960377, end_position_on_the_genomic_accession=6961653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33979, CurrentID=0, Status='live', Symbol='sip2', Aliases='Dmel_CG9188, CG9188, Dmel\\CG9188', description='septin interacting protein 2', other_designations='CG9188-PA|CG9188-PB|CG9188-PC|CG9188-PD|CG9188-PE|sip2-PA|sip2-PB|sip2-PC|sip2-PD|sip2-PE', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6961516, end_position_on_the_genomic_accession=6964074, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33981, CurrentID=0, Status='live', Symbol='smt3', Aliases='Dmel_CG4494, CG4494, Dm0342, DmSUMO-1, DmSmt3, Dmel\\CG4494, Dm SMT3, SUMO, Smt3, Sumo, anon-EST:Posey240, dSmt3, dsmt3, l(2)04493, l(2)SH0182, l(2)SH2 0182, sumo', description='CG4494 gene product from transcript CG4494-RA', other_designations='CG4494-PA|CG4494-PB|SUMO|SUMO protein|lethal (2) SH0182|small Ubiquitin-like modifier|smt3-PA|smt3-PB', map_location='27C7-27C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6966780, end_position_on_the_genomic_accession=6967644, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33982, CurrentID=0, Status='live', Symbol='snRNP-U1-70K', Aliases='Dmel_CG8749, 70K, CG8749, Dm70K, Dmel\\CG8749, U1, U1 70K, U1 70K snRNP, U1 70k, U1 snRNP, U1-70K, U1-70k, U170K, U1snRNP-70K, dU1-70K, l(2)02107, snRNP, snRNP27D, snRNP70K', description='small ribonucleoprotein particle U1 subunit 70K', other_designations='CG8749-PA|CG8749-PC|U1 snRNP-specific 70K protein|small nuclear ribonucleoprotein 70K|snRNP-U1-70K-PA|snRNP-U1-70K-PC', map_location='27C7-27D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6968711, end_position_on_the_genomic_accession=6971737, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33983, CurrentID=0, Status='live', Symbol='uif', Aliases='Dmel_CG9138, CG9138, CT26172, Dmel\\CG9138, Poly-EGF, SP1070, Uif, poly-EGF, sp1070', description='uninflatable', other_designations='CG9138-PB|CG9138-PC|CG9138-PD|CG9138-PE|CG9138-PF|CG9138-PG|SP1070|uif-PB|uif-PC|uif-PD|uif-PE|uif-PF|uif-PG', map_location='27D1-27D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6972828, end_position_on_the_genomic_accession=7010359, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  33984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33984, CurrentID=0, Status='live', Symbol='CG43322', Aliases='Dmel_ CG13776, Dmel\\CG43322', description='CG43322 gene product from transcript CG43322-RA', other_designations='CG43322-PA|uncharacterized protein', map_location='27D4-27D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7013396, end_position_on_the_genomic_accession=7014737, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33985, CurrentID=0, Status='live', Symbol='Pcp', Aliases='Dmel_CG3440, CG3440, DMCPGART, DmelPcp, Dmel\\CG3440, Gartgart, Pcp', description='Pupal cuticle protein', other_designations='CG3440-PA|Pcp-PA|gart intron gene|gart intronic gene|pupal cuticle protein', map_location='27D4-27D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7021682, end_position_on_the_genomic_accession=7022470, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33986, CurrentID=0, Status='live', Symbol='ade3', Aliases='Dmel_CG31628, 143062_at, ADE8, Ade3, CG31628, CG8761, D, Dmel\\CG31628, GART, Gart, Gart/ade3, gart, l(2)27Da', description='adenosine 3', other_designations='CG31628-PA|CG31628-PB|CG31628-PC|ade3-PA|ade3-PB|ade3-PC|complementation group D', map_location='27D4-27D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7014623, end_position_on_the_genomic_accession=7023898, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33987, CurrentID=0, Status='live', Symbol='CG3476', Aliases='Dmel_ Dmel\\CG3476', description='CG3476 gene product from transcript CG3476-RA', other_designations='CG3476-PA', map_location='27D5-27D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7027526, end_position_on_the_genomic_accession=7028959, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  33988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33988, CurrentID=0, Status='live', Symbol='Rab30', Aliases='Dmel_CG9100, AAF52477, CG9100, Dm Dmel\\CG9100, Q9VM50, Rab-30, rab30', description='CG9100 gene product from transcript CG9100-RB', other_designations='CG9100-PA|CG9100-PB|CG9100-PC|CG9100-PD|Rab30-PA|Rab30-PB|Rab30-PC|Rab30-PD', map_location='27D5-27D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7030493, end_position_on_the_genomic_accession=7032606, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  33989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33989, CurrentID=0, Status='live', Symbol='Caper', Aliases='Dmel_CG11266, CAPER, CC1.3, CG11266, Dmel\\CG11266, HCC1, cg11266', description='CG11266 gene product from transcript CG11266-RK', other_designations='CG11266-PA|CG11266-PB|CG11266-PH|CG11266-PI|CG11266-PJ|CG11266-PK|CG11266-PL|Caper-PA|Caper-PB|Caper-PH|Caper-PI|Caper-PJ|Caper-PK|Caper-PL|caper', map_location='27D5-27D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7032721, end_position_on_the_genomic_accession=7037878, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  33991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33991, CurrentID=0, Status='live', Symbol='Mnn1', Aliases='Dmel_CG13778, CG13778, CG33266, Dmel\\CG13778, MEN1, Menin, Menin1, menin, mnn1', description='Menin 1', other_designations='CG13778-PA|CG13778-PB|CG13778-PC|Menin1|Mnn1-PA|Mnn1-PB|Mnn1-PC|menin|menin 1', map_location='27D7-27D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7056301, end_position_on_the_genomic_accession=7063586, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  33992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33992, CurrentID=0, Status='live', Symbol='Nuf2', Aliases='Dmel_CG8902, CG8902, Dmel\\CG8902, anon-WO0118547.214, dm l(2)SH2 2276', description='CG8902 gene product from transcript CG8902-RA', other_designations='CG8902-PA|Nuf2-PA', map_location='27D7-27D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7064431, end_position_on_the_genomic_accession=7065837, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  33993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33993, CurrentID=0, Status='live', Symbol='CG11289', Aliases='Dmel_ Dmel\\CG11289', description='CG11289 gene product from transcript CG11289-RB', other_designations='CG11289-PB|uncharacterized protein', map_location='27D7-27E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7067983, end_position_on_the_genomic_accession=7069546, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  33994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33994, CurrentID=0, Status='live', Symbol='Pvf2', Aliases='Dmel_CG13780, CG13780, DmVEGF-2, Dmel\\CG13780, PV2, PVF, PVF2, PVFs, Pvf, VEGF, VEGF-2, VEGF27Cb, Vegf27Cb, pvf2', description='PDGF- and VEGF-related factor 2', other_designations='CG13780-PA|PDGF- and VEGF-receptor related factor 2|Pvf2-PA|VEGF-related factor 2', map_location='27E1-27E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7069536, end_position_on_the_genomic_accession=7084635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  33995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33995, CurrentID=0, Status='live', Symbol='Pvf3', Aliases='Dmel_CG34378, CG13781, CG13782, CG13783, CG31629, CG34378, CG4492, DmVEGF-3, Dmel\\CG34378, Dmel_CG13783, Dmel_CG31629, PVF, PVF3, PVFs, Pvf, VEGF, VEGF27Ca, Vegf27Ca, pvf3', description='PDGF- and VEGF-related factor 3', other_designations='CG34378-PA|CG34378-PD|CG34378-PE|CG34378-PF|Pvf3-PA|Pvf3-PD|Pvf3-PE|Pvf3-PF', map_location='27E2-27E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7098188, end_position_on_the_genomic_accession=7157695, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  33999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33999, CurrentID=0, Status='live', Symbol='CG4495', Aliases='Dmel_ Dmel\\CG4495', description='CG4495 gene product from transcript CG4495-RB', other_designations='CG4495-PA|CG4495-PB', map_location='27E4-27E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7182234, end_position_on_the_genomic_accession=7185529, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  34000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34000, CurrentID=0, Status='live', Symbol='CG4496', Aliases='Dmel_ Dmel\\CG4496', description='CG4496 gene product from transcript CG4496-RA', other_designations='CG4496-PA|CG4496-PB|CG4496-PC', map_location='27E4-27E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7185461, end_position_on_the_genomic_accession=7190010, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34001, CurrentID=0, Status='live', Symbol='CG4497', Aliases='Dmel_ Dmel\\CG4497', description='CG4497 gene product from transcript CG4497-RB', other_designations='CG4497-PA|CG4497-PB', map_location='27E4-27E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7190431, end_position_on_the_genomic_accession=7194588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34002, CurrentID=0, Status='live', Symbol='CG4502', Aliases='Dmel_ Dmel\\CG4502, anon-WO0118547.403', description='CG4502 gene product from transcript CG4502-RA', other_designations='CG4502-PA|CG4502-PB', map_location='27E4-27E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7194908, end_position_on_the_genomic_accession=7204264, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34003, CurrentID=0, Status='live', Symbol='CG13784', Aliases='Dmel_ BcDNA:SD10385, Dmel\\CG13784', description='CG13784 gene product from transcript CG13784-RD', other_designations='CG13784-PB|CG13784-PC|CG13784-PD|CG13784-PE|CG13784-PF|uncharacterized protein', map_location='27E4-27E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7204538, end_position_on_the_genomic_accession=7220033, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34004, CurrentID=0, Status='live', Symbol='ico', Aliases='Dmel_CG4567, B, CG4567, Dmel\\CG4567, G, GFM1, dEF-G1, l(2)27Cb, l(2)27Ea', description='iconoclast', other_designations='CG4567-PA|complementation group B|complementation group G|ico-PA|lethal (2) 27Cb|lethal (2) 27Ea', map_location='27E5-27F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7220369, end_position_on_the_genomic_accession=7223081, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34005, CurrentID=0, Status='live', Symbol='Ndae1', Aliases='Dmel_CG42253, 153385_at, CG31906, CG42253, CG4675, Dmel\\CG42253, Dmel_CG31906, Dmel_CG4675, NDAE1, Ndae-1, anon-WO0257455.13, l(2)k10316, ndae1', description='Na[+]-driven anion exchanger 1', other_designations='CG42253-PA|CG42253-PB|CG42253-PC|CG42253-PD|CG42253-PE|CG42253-PF|CG42253-PG|CG42253-PH|CG42253-PI|CG42253-PJ|Ndae1-PA|Ndae1-PB|Ndae1-PC|Ndae1-PD|Ndae1-PE|Ndae1-PF|Ndae1-PG|Ndae1-PH|Ndae1-PI|Ndae1-PJ', map_location='27E6-27E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7223328, end_position_on_the_genomic_accession=7249343, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  34006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34006, CurrentID=0, Status='live', Symbol='CG13786', Aliases='Dmel_ Dmel\\CG13786', description='CG13786 gene product from transcript CG13786-RB', other_designations='CG13786-PB|uncharacterized protein', map_location='27E7-27E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7250607, end_position_on_the_genomic_accession=7252982, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34007, CurrentID=0, Status='live', Symbol='Wnt4', Aliases='Dmel_CG4698, CG4698, DWnt-4, DWnt4, Dm DWnt4, Dmel\\CG4698, Dwnt4, Wnt, Wnt-4, anon-EST:Liang-2.4, clone 2.4, wnt-4, wnt4', description='Wnt oncogene analog 4', other_designations='CG4698-PA|CG4698-PB|Wnt4-PA|Wnt4-PB', map_location='27E7-27E8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7255419, end_position_on_the_genomic_accession=7277171, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34009, CurrentID=0, Status='live', Symbol='wg', Aliases='Dmel_CG4889, Br, CG4889, DWint-1, DWnt-1, Dint-1, Dm Wg, Dm-1, Dmel\\CG4889, Gla, I, Sp, WG, WNT, Wg, Wnt, Wnt-1, Wnt/Wg, Wnt1, dWnt, fg, int-1, l(2)02657, l(2)rO727, l(2)wg, spdl, wnt, wnt1, wg', description='wingless', other_designations='CG4889-PA|Wnt-1|bristled|complementation group I|flag|glazed|spade|sternopleural|wg-PA', map_location='2-21.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7307159, end_position_on_the_genomic_accession=7316265, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34010, CurrentID=0, Status='live', Symbol='Wnt6', Aliases='Dmel_CG4969, CG4969, D-Wnt-6, DWnt-6, DWnt6, Dmel\\CG4969, wnt6', description='Wnt oncogene analog 6', other_designations='CG4969-PB|CG4969-PC|Wnt6-PB|Wnt6-PC', map_location='27F2-27F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7333714, end_position_on_the_genomic_accession=7352542, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34011, CurrentID=0, Status='live', Symbol='Wnt10', Aliases='Dmel_CG4971, CG4971, D-Wnt-10, DWnt-10, DWnt10, Dm DWnt10, Dmel\\CG4971, wnt10', description='Wnt oncogene analog 10', other_designations='CG4971-PB|Wnt10-PB', map_location='27F2-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7363379, end_position_on_the_genomic_accession=7377528, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34012, CurrentID=0, Status='live', Symbol='ninaC', Aliases='Dmel_CG5125, 2.2, CG 5125, CG5125, CG54125, CT16120, CT42491, DRONINAC, Dm NinaC, Dmel\\CG5125, NINA C, NINAC, Nina C, NinaC, ninac', description='neither inactivation nor afterpotential C', other_designations='CG5125-PA|CG5125-PB|CG5125-PC|CG5125-PD|NINAC myosin III|droninac|either inactivation nor afterpotential C|myosin III|myosin-III|neither inactivation Nor afterpotential C|neither inactivation nor after potential C|neither inactivation nor afterpotential-C|ninaC-PA|ninaC-PB|ninaC-PC|ninaC-PD', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7377702, end_position_on_the_genomic_accession=7384344, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  34013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34013, CurrentID=0, Status='live', Symbol='CG5149', Aliases='Dmel_ Dmel\\CG5149', description='CG5149 gene product from transcript CG5149-RA', other_designations='CG5149-PA', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7384486, end_position_on_the_genomic_accession=7388253, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34014, CurrentID=0, Status='live', Symbol='gudu', Aliases='Dmel_CG5155, CG5155, Dmel\\CG5155, Gudu', description='CG5155 gene product from transcript CG5155-RA', other_designations='CG5155-PA|gudu-PA', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7388508, end_position_on_the_genomic_accession=7390902, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34015, CurrentID=0, Status='live', Symbol='CG5160', Aliases='Dmel_ Dmel\\CG5160', description='CG5160 gene product from transcript CG5160-RA', other_designations='CG5160-PA|CG5160-PB', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7390984, end_position_on_the_genomic_accession=7398750, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34016, CurrentID=0, Status='live', Symbol='CG5171', Aliases='Dmel_ Dmel\\CG5171', description='CG5171 gene product from transcript CG5171-RD', other_designations='CG5171-PA|CG5171-PB|CG5171-PD|CG5171-PE|CG5171-PF', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7401335, end_position_on_the_genomic_accession=7404727, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34017, CurrentID=0, Status='live', Symbol='CG5177', Aliases='Dmel_ 21430192, 28603684, Dmel\\CG5177', description='CG5177 gene product from transcript CG5177-RB', other_designations='CG5177-PA|CG5177-PB|trehalose-phosphatase', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7405227, end_position_on_the_genomic_accession=7408600, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34018, CurrentID=0, Status='live', Symbol='CG5181', Aliases='Dmel_ Dmel\\CG5181', description='CG5181 gene product from transcript CG5181-RA', other_designations='CG5181-PA', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7408533, end_position_on_the_genomic_accession=7409809, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34019, CurrentID=0, Status='live', Symbol='CG15818', Aliases='Dmel_ Dmel\\CG15818', description='CG15818 gene product from transcript CG15818-RA', other_designations='CG15818-PA|uncharacterized protein', map_location='27F3-27F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7410909, end_position_on_the_genomic_accession=7412229, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34021, CurrentID=0, Status='live', Symbol='CG5261', Aliases='Dmel_ DLAT, Dmel\\CG5261, EP(2)0816, EP816, anon-WO0118547.121', description='CG5261 gene product from transcript CG5261-RB', other_designations='CG5261-PA|CG5261-PB|CG5261-PC|CG5261-PD|CG5261-PE|CG5261-PF|Dihydrolipoyllysine-residue acetyltransferase', map_location='27F4-27F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7426866, end_position_on_the_genomic_accession=7430329, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  34022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34022, CurrentID=0, Status='live', Symbol='CG5958', Aliases='Dmel_ Dmel\\CG5958', description='CG5958 gene product from transcript CG5958-RA', other_designations='CG5958-PA', map_location='27F4-27F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7430530, end_position_on_the_genomic_accession=7433516, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34023, CurrentID=0, Status='live', Symbol='CG5973', Aliases='Dmel_ Dmel\\CG5973', description='CG5973 gene product from transcript CG5973-RE', other_designations='CG5973-PA|CG5973-PB|CG5973-PC|CG5973-PD|CG5973-PE|CG5973-PF|CG5973-PG', map_location='27F4-28A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7437492, end_position_on_the_genomic_accession=7445882, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  34024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34024, CurrentID=0, Status='live', Symbol='santa-maria', Aliases='Dmel_CG12789, BEST:CK01577, CG12789, CK01577, Dmel\\CG12789, Santa Maria, santa maria', description='scavenger receptor acting in neural tissue and majority of rhodopsin is absent', other_designations='CG12789-PA|CG12789-PC|SANTA-MARIA|santa maria|santa-maria-PA|santa-maria-PC', map_location='28A1-28A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7445355, end_position_on_the_genomic_accession=7451244, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34027, CurrentID=0, Status='live', Symbol='CG6055', Aliases='Dmel_ Dmel\\CG6055', description='CG6055 gene product from transcript CG6055-RA', other_designations='CG6055-PA|CG6055-PB', map_location='28A1-28A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7466543, end_position_on_the_genomic_accession=7472493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34029, CurrentID=0, Status='live', Symbol='CG6441', Aliases='Dmel_ Dmel\\CG6441, anon-WO0140519.45', description='CG6441 gene product from transcript CG6441-RA', other_designations='CG6441-PA|CG6441-PC', map_location='28A1-28A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7474704, end_position_on_the_genomic_accession=7476952, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34030, CurrentID=0, Status='live', Symbol='Ziz', Aliases='Dmel_CG42533, BcDNA:RH71439, CG15819, CG42533, CG6630, Dm ziz, Dmel\\CG42533, Dmel_CG15819, Dmel_CG6630', description='Zizimin ortholog (H. sapiens)', other_designations='CG42533-PB|CG42533-PC|CG42533-PD|CG42533-PE|CG42533-PG|CG42533-PH|CG42533-PI|CG42533-PJ|CG42533-PK|Dm zizimin|Ziz-PB|Ziz-PC|Ziz-PD|Ziz-PE|Ziz-PG|Ziz-PH|Ziz-PI|Ziz-PJ|Ziz-PK|zizimin ortholog', map_location='28A1-28A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7472513, end_position_on_the_genomic_accession=7496307, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  34031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34031, CurrentID=0, Status='live', Symbol='Obp28a', Aliases='Dmel_CG6641, 28a, CG6641, DmelObp28a, Dmel\\CG6641, PB-PRP5, PBP5, PBPRP-5, PBPRP5, Pbprp5, obp28a, pbprp-5, pbprp5', description='Odorant-binding protein 28a', other_designations='CG6641-PA|Obp28a-PA|Pbprp5|Pheromone-binding protein-related protein 5', map_location='28A4-28A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7496791, end_position_on_the_genomic_accession=7497365, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34032, CurrentID=0, Status='live', Symbol='RapGAP1', Aliases='Dmel_CG44086, CG13789, CG13790, CG13791, CG33529, CG34374, CG44086, CG6682, Dmel\\CG44086, Dmel_CG13790, Dmel_CG13791, Dmel_CG33529, Dmel_CG34374, Q9VM02_DROME, RapGap1, Rapgap1, rapgap1', description='Rap GTPase activating protein 1', other_designations='CG44086-PD|CG44086-PE|CG44086-PF|CG44086-PH|CG44086-PI|RapGAP1-PD|RapGAP1-PE|RapGAP1-PF|RapGAP1-PH|RapGAP1-PI|Rapgap1', map_location='28A4-28B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7497805, end_position_on_the_genomic_accession=7578980, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34035, CurrentID=0, Status='live', Symbol='Spn28B', Aliases='Dmel_CG6717, CG6717, Dmel\\CG6717, Spn7', description='Serpin 28B', other_designations='CG6717-PA|CG6717-PB|Spn28B-PA|Spn28B-PB|serine protease inhibitor 7|serpin 28B', map_location='28B3-28B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7581924, end_position_on_the_genomic_accession=7583333, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34036, CurrentID=0, Status='live', Symbol='Cyp4d21', Aliases='Dmel_CG6730, 4d21, CG6730, Dmel\\CG6730, cyp4d21, sxe1', description='CG6730 gene product from transcript CG6730-RA', other_designations='CG6730-PA|Cyp4d21-PA|cytochrome P450|sex-specific enzyme 1', map_location='28B4-28B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7605600, end_position_on_the_genomic_accession=7607701, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34037, CurrentID=0, Status='live', Symbol='CG6739', Aliases='Dmel_ Dmel\\CG6739', description='CG6739 gene product from transcript CG6739-RB', other_designations='CG6739-PA|CG6739-PB', map_location='28B4-28B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7608389, end_position_on_the_genomic_accession=7616533, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34038, CurrentID=0, Status='live', Symbol='Slob', Aliases='Dmel_CG43756, CG13792, CG43756, CG6772, Dmel\\CG43756, Dmel_CG13792, Dmel_CG6772, slob', description='Slowpoke binding protein', other_designations='CG43756-PA|CG43756-PB|CG43756-PC|CG43756-PD|CG43756-PE|CG43756-PF|CG43756-PG|CG43756-PH|CG43756-PI|CG43756-PJ|CG43756-PK|SLO-binding protein|SLOWPOKE-binding protein|Slob-PA|Slob-PB|Slob-PC|Slob-PD|Slob-PE|Slob-PF|Slob-PG|Slob-PH|Slob-PI|Slob-PJ|Slob-PK|giant slob|slowpoke binding protein|slowpoke channel binding protein', map_location='28B4-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7642005, end_position_on_the_genomic_accession=7675078, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  34041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34041, CurrentID=0, Status='live', Symbol='CG18585', Aliases='Dmel_ Dmel\\CG18585', description='CG18585 gene product from transcript CG18585-RB', other_designations='CG18585-PA|CG18585-PB|uncharacterized protein', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7689311, end_position_on_the_genomic_accession=7691061, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34042, CurrentID=0, Status='live', Symbol='CG7025', Aliases='Dmel_ Dmel\\CG7025', description='CG7025 gene product from transcript CG7025-RA', other_designations='CG7025-PA', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7691109, end_position_on_the_genomic_accession=7692683, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34044, CurrentID=0, Status='live', Symbol='Tep2', Aliases='Dmel_CG7052, 145970_at, 145971_at, CG18589, CG7052, Dmel\\CG7052, T13, TEP II, TEP2, TEP2mel, Tep 2, Tep II, TepII, dTEPII, tepII', description='Thioester-containing protein 2', other_designations='CG7052-PA|CG7052-PB|CG7052-PD|CG7052-PF|CG7052-PG|TEP|Tep2-PA|Tep2-PB|Tep2-PD|Tep2-PF|Tep2-PG|thioester containing protein II|thiolester containing protein II', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7693727, end_position_on_the_genomic_accession=7701601, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34045, CurrentID=0, Status='live', Symbol='Tep3', Aliases='Dmel_CG7068, CG7068, Dmel\\CG7068, TEP3, TEP3mel, TEPIII, TepIII, dTEPIII', description='Thioester-containing protein 3', other_designations='CG7068-PA|TEP|Tep3-PA|thiolester containing protein III', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7702805, end_position_on_the_genomic_accession=7710690, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  34046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34046, CurrentID=0, Status='live', Symbol='Ntl', Aliases='Dmel_CG7075, CG7075, Dmel\\CG7075, Mp, ntl', description='Neurotransmitter transporter-like', other_designations='CG7075-PA|CG7075-PB|Ntl-PA|Ntl-PB|membrane protein|neurotransmitter transporter-like', map_location='28C-28C', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7710880, end_position_on_the_genomic_accession=7713991, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34047, CurrentID=0, Status='live', Symbol='CG13793', Aliases='Dmel_ Dmel\\CG13793', description='CG13793 gene product from transcript CG13793-RB', other_designations='CG13793-PB|uncharacterized protein', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7716974, end_position_on_the_genomic_accession=7719487, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34048, CurrentID=0, Status='live', Symbol='CG13794', Aliases='Dmel_ Dmel\\CG13794', description='CG13794 gene product from transcript CG13794-RA', other_designations='CG13794-PA|uncharacterized protein', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7719847, end_position_on_the_genomic_accession=7722823, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34049, CurrentID=0, Status='live', Symbol='CG13795', Aliases='Dmel_ CT33284, Dmel\\CG13795', description='CG13795 gene product from transcript CG13795-RC', other_designations='CG13795-PB|CG13795-PC|uncharacterized protein', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7723045, end_position_on_the_genomic_accession=7726298, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34051, CurrentID=0, Status='live', Symbol='CG13796', Aliases='Dmel_CG13796-PA, CG18590, CT42497, Dmel\\CG13796', description='CG13796 gene product from transcript CG13796-RD', other_designations='CG13796-PA|CG13796-PB|CG13796-PC|CG13796-PD|uncharacterized protein', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7727157, end_position_on_the_genomic_accession=7732670, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  34052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34052, CurrentID=0, Status='live', Symbol='Acp1', Aliases='Dmel_CG7216, CG7216, CG7216-PA, Dacp-1, Dmel\\CG7216, acp1', description='Adult cuticle protein 1', other_designations='Acp1-PA|CG7216-PA|adult cuticle protein 1', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7740562, end_position_on_the_genomic_accession=7741537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34053, CurrentID=0, Status='live', Symbol='CG7214', Aliases='Dmel_ Dmel\\CG7214', description='CG7214 gene product from transcript CG7214-RA', other_designations='CG7214-PA', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7743682, end_position_on_the_genomic_accession=7744839, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34054, CurrentID=0, Status='live', Symbol='ATPsynGL', Aliases='Dmel_CG7211, ATPN, CG7211, Dmel\\CG7211', description='ATP synthase, subunit G-like', other_designations='ATP synthase g chain|ATPsynGL-PA|CG7211-PA', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7747275, end_position_on_the_genomic_accession=7747874, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34055, CurrentID=0, Status='live', Symbol='CG7203', Aliases='Dmel_ Dmel\\CG7203', description='CG7203 gene product from transcript CG7203-RC', other_designations='CG7203-PB|CG7203-PC', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7752170, end_position_on_the_genomic_accession=7753186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34058, CurrentID=0, Status='live', Symbol='CG7191', Aliases='Dmel_ Dmel\\CG7191', description='CG7191 gene product from transcript CG7191-RA', other_designations='CG7191-PA|CG7191-PB', map_location='28C3-28C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7767628, end_position_on_the_genomic_accession=7775033, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34059, CurrentID=0, Status='live', Symbol='CG7179', Aliases='Dmel_ CT22169, Dmel\\CG7179, Opa, anon-28Cg, tld-like', description='CG7179 gene product from transcript CG7179-RB', other_designations='CG7179-PB', map_location='28C3-28C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7775804, end_position_on_the_genomic_accession=7778029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34060, CurrentID=0, Status='live', Symbol='Uro', Aliases='Dmel_CG7171, CG7171, Dm UO, Dmel\\CG7171, OU, UO, UOX, Uo, anon-WO0140519.210, uro', description='Urate oxidase', other_designations='CG7171-PA|Uro-PA|urate oxidase', map_location='28C3-28C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7780085, end_position_on_the_genomic_accession=7781415, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34061, CurrentID=0, Status='live', Symbol='CG7164', Aliases='Dmel_ Dmel\\CG7164', description='CG7164 gene product from transcript CG7164-RB', other_designations='CG7164-PA|CG7164-PB', map_location='28C3-28C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7781465, end_position_on_the_genomic_accession=7782760, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34062, CurrentID=0, Status='live', Symbol='CG7154', Aliases='Dmel_ Dmel\\CG7154', description='CG7154 gene product from transcript CG7154-RA', other_designations='CG7154-PA', map_location='28C3-28C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7782865, end_position_on_the_genomic_accession=7786061, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34063, CurrentID=0, Status='live', Symbol='CG7149', Aliases='Dmel_ Dmel\\CG7149, cpt, dCCS3', description='CG7149 gene product from transcript CG7149-RB', other_designations='CG7149-PA|CG7149-PB|CG7149-PC|diacylglycerol cholinephosphotransferase', map_location='28C3-28C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7786160, end_position_on_the_genomic_accession=7790557, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34064, CurrentID=0, Status='live', Symbol='LKRSDH', Aliases='Dmel_CG7144, BEST:CK02318, CG7144, CK02318, Dmel\\CG7144, LKR, SDH, anon-WO0149856.1, dLKR', description='Lysine ketoglutarate reductase/saccharopine dehydrogenase', other_designations='CG7144-PA|CG7144-PB|LKRSDH-PA|LKRSDH-PB|drosophila lysine ketoglutarate reductase|drosophila lysine ketoglutarate reductase (dLKR)/saccharopine dehydrogenase (SDH)|lysine ketoglutarate reductase|lysine ketoglutarate reductase/saccharopine dehydrogenase|saccharopine dehydogenase', map_location='28C4-28C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7791175, end_position_on_the_genomic_accession=7795959, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34065, CurrentID=0, Status='live', Symbol='Herp', Aliases='Dmel_CG14536, CG1453, CG14536, Dmel\\CG14536, dHerp, herp', description='Homocysteine-induced endoplasmic reticulum protein', other_designations='CG14536-PA|CG14536-PC|Herp-PA|Herp-PC|homocysteine-induced endoplasmic reticulum protein', map_location='28C4-28C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7796314, end_position_on_the_genomic_accession=7800058, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34066, CurrentID=0, Status='live', Symbol='r2d2', Aliases='Dmel_CG7138, CG7138, Dmel\\CG7138, R2D2, R2d2, cg7138', description='CG7138 gene product from transcript CG7138-RA', other_designations='CG7138-PA|CG7138-PB|CG7138-PC|R2D2|r2d2-PA|r2d2-PB|r2d2-PC', map_location='28C4-28C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7800147, end_position_on_the_genomic_accession=7802098, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34067, CurrentID=0, Status='live', Symbol='cdc14', Aliases='Dmel_CG7134, CDC-14, CDC14, CG7134, Cdc14, Dmel\\CG7134', description='cell division cycle 14', other_designations='CG7134-PA|CG7134-PB|CG7134-PC|CG7134-PD|CG7134-PE|cdc14-PA|cdc14-PB|cdc14-PC|cdc14-PD|cdc14-PE', map_location='28C4-28C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7801668, end_position_on_the_genomic_accession=7810703, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34068, CurrentID=0, Status='live', Symbol='LanB1', Aliases='Dmel_CG7123, BcDNA:LD21794, CG7123, CT22015, Dmel\\CG7123, EP-600, LM-B1/beta, Lam-B1, LamB1, Lambeta, Lan B1, LanB, beta1, lamB1, mspl', description='CG7123 gene product from transcript CG7123-RA', other_designations='CG7123-PA|CG7123-PB|LamininB1|LanB1-PA|LanB1-PB|laminin|laminin B1|laminin B1 subunit|laminin beta|laminin beta-chain|laminin beta1|lamininB|misplaced gonads', map_location='28C4-28D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7811438, end_position_on_the_genomic_accession=7820771, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34069, CurrentID=0, Status='live', Symbol='CG7115', Aliases='Dmel_ BcDNA:LD21794, Dmel\\CG7115, PP2C', description='CG7115 gene product from transcript CG7115-RD', other_designations='CG7115-PA|CG7115-PB|CG7115-PC|CG7115-PD|CG7115-PE', map_location='28D1-28D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7821151, end_position_on_the_genomic_accession=7825350, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34070, CurrentID=0, Status='live', Symbol='Rack1', Aliases='Dmel_CG7111, CG7111, Dmel\\CG7111, RACK1, dRack1, i173, rack1', description='Receptor of activated protein kinase C 1', other_designations='CG7111-PA|CG7111-PB|CG7111-PC|CG7111-PD|Rack1-PA|Rack1-PB|Rack1-PC|Rack1-PD|guanine nucleotide-binding protein beta-subunit like protein|receptor for activated C kinase 1|receptor for activated protein kinase C1|receptor of activated protein kinase C 1', map_location='28D1-28D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7825347, end_position_on_the_genomic_accession=7827315, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34072, CurrentID=0, Status='live', Symbol='CG14537', Aliases='Dmel_ BcDNA:AT11469, Dmel\\CG14537', description='CG14537 gene product from transcript CG14537-RA', other_designations='CG14537-PA|uncharacterized protein', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7840452, end_position_on_the_genomic_accession=7841262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34073, CurrentID=0, Status='live', Symbol='CG14535', Aliases='Dmel_ DmKIF26, Dmel\\CG14535, KIF26, KIF26A/B', description='CG14535 gene product from transcript CG14535-RB', other_designations='CG14535-PA|CG14535-PB|kinesin superfamily protein 26|uncharacterized protein', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7844905, end_position_on_the_genomic_accession=7855475, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34075, CurrentID=0, Status='live', Symbol='TwdlE', Aliases='Dmel_CG14534, CG14534, Dmel\\CG14534', description='TweedleE', other_designations='CG14534-PA|TwdlE-PA', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7857957, end_position_on_the_genomic_accession=7860839, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34077, CurrentID=0, Status='live', Symbol='spz3', Aliases='Dmel_CG7104, CG14533, CG7104, CT21967, CT34263, Dmel\\CG7104, Spz3', description='spatzle 3', other_designations='CG7104-PA|Spz3|spz3-PA', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7873495, end_position_on_the_genomic_accession=7879552, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34079, CurrentID=0, Status='live', Symbol='CG7102', Aliases='Dmel_ CG17973, Dmel\\CG7102', description='CG7102 gene product from transcript CG7102-RB', other_designations='CG7102-PA|CG7102-PB', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7883282, end_position_on_the_genomic_accession=7887513, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34080, CurrentID=0, Status='live', Symbol='SmE', Aliases='Dmel_CG18591, BcDNA:GM19936, CG18591, Dmel\\CG18591, E, E-Sm, dE, snRNPE', description='Small ribonucleoprotein particle protein SmE', other_designations='CG18591-PA|SmE-PA|small ribonucleoprotein particle protein SmE|snRNPE', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7884930, end_position_on_the_genomic_accession=7885450, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34082, CurrentID=0, Status='live', Symbol='Spn28Da', Aliases='Dmel_CG31902, CG12318, CG31902, Dmel\\CG31902', description='Serpin 28Da', other_designations='CG31902-PB|Spn28Da-PB|serpin 28Da', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7887693, end_position_on_the_genomic_accession=7889129, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740878, CurrentID=0, Status='live', Symbol='CG34280', Aliases='Dmel_ Dmel\\CG34280', description='CG34280 gene product from transcript CG34280-RA', other_designations='CG34280-PA|uncharacterized protein', map_location='90A5-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17403272, end_position_on_the_genomic_accession=17403718, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34086, CurrentID=0, Status='live', Symbol='CG7231', Aliases='Dmel_ Dmel\\CG7231', description='CG7231 gene product from transcript CG7231-RE', other_designations='CG7231-PC|CG7231-PD|CG7231-PE|CG7231-PF', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7983946, end_position_on_the_genomic_accession=7986785, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34087, CurrentID=0, Status='live', Symbol='pes', Aliases='Dmel_CG7228, CG7228, Dmel\\CG7228, Pes', description='peste', other_designations='CG7228-PC|CG7228-PD|CG7228-PE|CG7228-PF|CG7228-PG|CG7228-PH|pes-PC|pes-PD|pes-PE|pes-PF|pes-PG|pes-PH', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7986786, end_position_on_the_genomic_accession=7994184, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34088, CurrentID=0, Status='live', Symbol='CG7227', Aliases='Dmel_ Dmel\\CG7227', description='CG7227 gene product from transcript CG7227-RC', other_designations='CG7227-PA|CG7227-PB|CG7227-PC', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7994388, end_position_on_the_genomic_accession=7998151, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34089, CurrentID=0, Status='live', Symbol='Sirup', Aliases='Dmel_CG7224, BcDNA:GH18422, CG7224, Dmel\\CG7224, anon-EST:Posey12, dSdhaf4', description='Starvation-upregulated protein', other_designations='CG7224-PA|CG7224-PC|Sirup-PA|Sirup-PC|starvation-upregulated protein', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7998869, end_position_on_the_genomic_accession=7999676, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34090, CurrentID=0, Status='live', Symbol='Wwox', Aliases='Dmel_CG7221, CG7221, DmWWOX, Dmel\\CG7221, anon-EST:Posey177', description='WW domain containing oxidoreductase', other_designations='CG7221-PA|CG7221-PB|Wwox-PA|Wwox-PB', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7999670, end_position_on_the_genomic_accession=8004286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34091, CurrentID=0, Status='live', Symbol='Spn28Dc', Aliases='Dmel_CG7219, CG7219, Dm-Spn28D, Dmel\\CG7219, Spn28D, spn28D', description='Serpin 28Dc', other_designations='CG7219-PA|Dm-serpin-28D|Serpin-28D|Spn28Dc-PA|serpin 28D|serpin 28Dc', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8004548, end_position_on_the_genomic_accession=8007210, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34092, CurrentID=0, Status='live', Symbol='CG12560', Aliases='Dmel_ Dmel\\CG12560', description='CG12560 gene product from transcript CG12560-RC', other_designations='CG12560-PB|CG12560-PC|uncharacterized protein', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8008302, end_position_on_the_genomic_accession=8009477, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34093, CurrentID=0, Status='live', Symbol='Tg', Aliases='Dmel_CG7356, CG7356, Dmel\\CG7356, TG, dTG', description='Transglutaminase', other_designations='CG7356-PA|CG7356-PB|Tg-PA|Tg-PB|transglutaminase', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8011405, end_position_on_the_genomic_accession=8026898, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34094, CurrentID=0, Status='live', Symbol='CG7367', Aliases='Dmel_ Dmel\\CG7367, TL', description='CG7367 gene product from transcript CG7367-RD', other_designations='CG7367-PB|CG7367-PC|CG7367-PD', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8026914, end_position_on_the_genomic_accession=8039788, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34095, CurrentID=0, Status='live', Symbol='baf', Aliases='Dmel_CG7380, BAF, CG7380, Dmel\\CG7380, anon-WO0172774.178, dBAF', description='barrier to autointegration factor', other_designations='Barrier-to-autointegration factor|CG7380-PA|CG7380-PB|baf-PA|baf-PB', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8029462, end_position_on_the_genomic_accession=8030565, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34096, CurrentID=0, Status='live', Symbol='Cka', Aliases='Dmel_CG7392, 5836, CG7392, CKA, Dcka, Dmel\\CG7392, anon-WO0172774.174, anon-WO0172774.175, anon-WO0172774.177, anon-WO0172774.179, cka, l(2)05836, l(2)s1883', description='Connector of kinase to AP-1', other_designations='CG7392-PA|CG7392-PB|CG7392-PC|CG7392-PD|CG7392-PE|CG7392-PF|CG7392-PG|Cka-PA|Cka-PB|Cka-PC|Cka-PD|Cka-PE|Cka-PF|Cka-PG|WD-40-family-member|connector of kinase to AP-1', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8030721, end_position_on_the_genomic_accession=8040826, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34097, CurrentID=0, Status='live', Symbol='CCDC53', Aliases='Dmel_CG7429, CG7429, Dmel\\CG7429', description='CG7429 gene product from transcript CG7429-RA', other_designations='CCDC53-PA|CG7429-PA', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8041106, end_position_on_the_genomic_accession=8041879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34098, CurrentID=0, Status='live', Symbol='RpL36A', Aliases='Dmel_CG7424, CG7424, Dmel\\CG7424, anon-EST:fe1B2', description='Ribosomal protein L36A', other_designations='CG7424-PA|CG7424-PB|RpL36A-PA|RpL36A-PB|anon-fast-evolving-1B2|ribosomal protein L36A', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8041873, end_position_on_the_genomic_accession=8042909, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34099, CurrentID=0, Status='live', Symbol='CG7466', Aliases='Dmel_ Dmel\\CG7466', description='CG7466 gene product from transcript CG7466-RB', other_designations='CG7466-PB|CG7466-PC', map_location='28D3-28D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8043471, end_position_on_the_genomic_accession=8054326, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34102, CurrentID=0, Status='live', Symbol='Trf', Aliases='Dmel_CG7562, CG7562, Dmel\\CG7562, TRF, TRF11, X70838, dTRF-1, dTRF1, trf, Trf', description='TBP-related factor', other_designations='CG7562-PA|Trf-PA', map_location='28E1-28E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8072895, end_position_on_the_genomic_accession=8073948, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34105, CurrentID=0, Status='live', Symbol='CG8673', Aliases='Dmel_ Dmel\\CG8673, GalT5, dbeta3GnT', description='CG8673 gene product from transcript CG8673-RA', other_designations='CG8673-PA', map_location='28E7-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8124329, end_position_on_the_genomic_accession=8125799, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34106, CurrentID=0, Status='live', Symbol='CG12375', Aliases='Dmel_ Dmel\\CG12375', description='CG12375 gene product from transcript CG12375-RA', other_designations='CG12375-PA|uncharacterized protein', map_location='28E7-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8125668, end_position_on_the_genomic_accession=8126890, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34107, CurrentID=0, Status='live', Symbol='CG8668', Aliases='Dmel_ Dmel\\CG8668, GalT4, dbeta3GnT', description='CG8668 gene product from transcript CG8668-RB', other_designations='CG8668-PA|CG8668-PB|resille', map_location='28E7-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8127001, end_position_on_the_genomic_accession=8133989, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34109, CurrentID=0, Status='live', Symbol='PAPLA1', Aliases='Dmel_CG8552, CG14278, CG8552, Dmel\\CG8552, Sec23-IP', description='Phosphatidic Acid Phospholipase A1', other_designations='CG8552-PA|CG8552-PB|CG8552-PD|CG8552-PE|PAPLA1-PA|PAPLA1-PB|PAPLA1-PD|PAPLA1-PE|Sec23-interacting protein|phosphatidic acid phospholipase A1', map_location='28E7-28E9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8134247, end_position_on_the_genomic_accession=8159638, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34110, CurrentID=0, Status='live', Symbol='Rbsn-5', Aliases='Dmel_CG8506, CG8506, Dmel\\CG8506, EEA1, MENE (2L)-C, MENE(2L)-C, Rbsn, Rbsn5, rbsn, rbsn-5', description='Rabenosyn-5', other_designations='CG8506-PA|Rabenosyn5|Rbsn-5-PA|dRabenosyn|early endosomal antigen 1|rabenosyn', map_location='28E9-28E9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8159832, end_position_on_the_genomic_accession=8161800, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34111, CurrentID=0, Status='live', Symbol='CG8498', Aliases='Dmel_ Dmel\\CG8498', description='CG8498 gene product from transcript CG8498-RB', other_designations='CG8498-PA|CG8498-PB', map_location='28E9-28E9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8161796, end_position_on_the_genomic_accession=8162452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34112, CurrentID=0, Status='live', Symbol='Piezo', Aliases='Dmel_CG44122, CG18103, CG31608, CG44122, CG8486, Dm Dmel\\CG44122, Dmel_CG18103, Dmel_CG8486, Dmpiezo, OrfKD, anon-EST:Liang-1.32, clone 1.32, fos28F, fos500', description='CG44122 gene product from transcript CG44122-RK', other_designations='CG44122-PE|CG44122-PF|CG44122-PG|CG44122-PH|CG44122-PI|CG44122-PJ|CG44122-PK|CG44122-PL|CG44122-PM|CG44122-PO|CG44122-PP|Piezo-PE|Piezo-PF|Piezo-PG|Piezo-PH|Piezo-PI|Piezo-PJ|Piezo-PK|Piezo-PL|Piezo-PM|Piezo-PO|Piezo-PP|dmpiezo|fos-related gene at 28F|piezo', map_location='28F1-28F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8163073, end_position_on_the_genomic_accession=8190137, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  246640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246640, CurrentID=0, Status='live', Symbol='CG30480', Aliases='Dmel_ CG18373, Dmel\\CG30480', description='CG30480 gene product from transcript CG30480-RB', other_designations='CG30480-PA|CG30480-PB|uncharacterized protein', map_location='51B10-51B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14674606, end_position_on_the_genomic_accession=14678136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34114, CurrentID=0, Status='live', Symbol='CG8475', Aliases='Dmel_ Dmel\\CG8475', description='CG8475 gene product from transcript CG8475-RC', other_designations='CG8475-PA|CG8475-PB|CG8475-PC', map_location='28F1-28F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8190808, end_position_on_the_genomic_accession=8196918, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  34115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34115, CurrentID=0, Status='live', Symbol='CG8460', Aliases='Dmel_ Dmel\\CG8460', description='CG8460 gene product from transcript CG8460-RA', other_designations='CG8460-PA', map_location='28F1-28F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8197193, end_position_on_the_genomic_accession=8198813, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34116, CurrentID=0, Status='live', Symbol='PGAP5', Aliases='Dmel_CG8455, CG8455, Dmel\\CG8455', description='Post-GPI attachment to proteins 5 ortholog (H. sapiens)', other_designations='CG8455-PA|CG8455-PB|CG8455-PC|PGAP5-PA|PGAP5-PB|PGAP5-PC|Post-GPI attachment to proteins 5 ortholog', map_location='28F1-28F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8198882, end_position_on_the_genomic_accession=8200652, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34117, CurrentID=0, Status='live', Symbol='SLC5A11', Aliases='Dmel_CG8451, CG8451, CG8451-PA, Dmel\\CG8451', description='Sodium/solute co-transporter-like 5A11', other_designations='CG8451-PA|NA-dependent multivitamin transporter|SLC5A11-PA|cupcake', map_location='28F1-28F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8200465, end_position_on_the_genomic_accession=8205132, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34118, CurrentID=0, Status='live', Symbol='CG8419', Aliases='Dmel_ Dmel\\CG8419', description='CG8419 gene product from transcript CG8419-RA', other_designations='CG8419-PA', map_location='28F1-28F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8206171, end_position_on_the_genomic_accession=8209637, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34119, CurrentID=0, Status='live', Symbol='Su(var)205', Aliases='Dmel_CG8409, CBX5, CG8409, DmHP-1, DmHP1, Dmel\\CG8409, E(var)29, E(var)29A, FBgn0003607, HP-1, HP1, HP1-VS, HP1A, HP1a, HP1alpha, Hp1, Hp1a, SU(VAR)205, Su(Var)2-5, Su(var), Su(var)2-05, Su(var)2-5, Su(var)2-501, Su(var)29A, Su(var)[205], Su-var(2)5, Su[Var]205, Su[var]205, Suvar(2)5, Suvar2-5, dHP-1a, dHP1, hp1, hp1a, su(var)205', description='Suppressor of variegation 205', other_designations='C1A9-nuclear-antigen|CG8409-PA|CG8409-PB|Su(var)2-5|Su(var)205-PA|Su(var)205-PB|heterochromatic protein|heterochromatic protein 1a|heterochromatin protein 1|heterochromatin protein 1A|heterochromatin protein-1|heterochromatin-binding protein|heterochromatin-protein-1|histone protein 1|suppressor of variegated 205|suppressor of variegation 205', map_location='2-31.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8209626, end_position_on_the_genomic_accession=8211122, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34120, CurrentID=0, Status='live', Symbol='Ssb-c31a', Aliases='Dmel_CG8396, C31A, CG8396, Dmel\\CG8396, PC4-like, Ssb-C31a, p11', description='Single stranded-binding protein c31A', other_designations='CG8396-PA|Ssb-c31a-PA|single stranded-binding protein c31A', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8211347, end_position_on_the_genomic_accession=8211938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34121, CurrentID=0, Status='live', Symbol='CG8372', Aliases='Dmel_ Dmel\\CG8372', description='CG8372 gene product from transcript CG8372-RA', other_designations='CG8372-PA|CG8372-PB', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8212039, end_position_on_the_genomic_accession=8212902, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34122, CurrentID=0, Status='live', Symbol='CG8360', Aliases='Dmel_ BcDNA:LP03131, Dmel\\CG8360', description='CG8360 gene product from transcript CG8360-RC', other_designations='CG8360-PA|CG8360-PC', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8212959, end_position_on_the_genomic_accession=8214135, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34123, CurrentID=0, Status='live', Symbol='CG8353', Aliases='Dmel_ Dmel\\CG8353', description='CG8353 gene product from transcript CG8353-RB', other_designations='CG8353-PA|CG8353-PB', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8214045, end_position_on_the_genomic_accession=8215239, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34124, CurrentID=0, Status='live', Symbol='CG8349', Aliases='Dmel_ Dmel\\CG8349', description='CG8349 gene product from transcript CG8349-RA', other_designations='CG8349-PA', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8215915, end_position_on_the_genomic_accession=8216901, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34125, CurrentID=0, Status='live', Symbol='CG8292', Aliases='Dmel_ Dmel\\CG8292', description='CG8292 gene product from transcript CG8292-RA', other_designations='CG8292-PA', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8216851, end_position_on_the_genomic_accession=8218017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34126, CurrentID=0, Status='live', Symbol='Snx6', Aliases='Dmel_CG8282, CG8282, DmSNX6, Dmel\\CG8282, Dsnx6, SNX6', description='Sorting nexin 6', other_designations='CG8282-PA|CG8282-PB|Snx6-PA|Snx6-PB|sorting nexin 6', map_location='28F3-28F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8218071, end_position_on_the_genomic_accession=8220741, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34127, CurrentID=0, Status='live', Symbol='Pvr', Aliases='Dmel_CG8222, 8222, CG8222, CT24332, DmVEGFR, Dmel\\CG8222, PVR, VEGFR, VEGFR-A, VGR1, Vegf, Vegfr, Vegfr-b, Vegfr-c, Vgr1, pvr, stai, vgr1', description='PDGF- and VEGF-receptor related', other_designations='CG8222-PA|CG8222-PB|CG8222-PD|CG8222-PE|CG8222-PG|CG8222-PH|CG8222-PI|CG8222-PJ|CG8222-PK|CG8222-PL|CG8222-PM|CG8222-PN|CG8222-PO|PDGF- and VEGF receptor related|PDGF- and VEGF- receptor related|PDGF- and VEGF-Receptor related|PDGF- and VEGF-related receptor|PDGF/VEGF receptor|PVF receptor|Pvr-PA|Pvr-PB|Pvr-PD|Pvr-PE|Pvr-PG|Pvr-PH|Pvr-PI|Pvr-PJ|Pvr-PK|Pvr-PL|Pvr-PM|Pvr-PN|Pvr-PO|VEGF receptor|platelet-derived growth factor/vascular endothelial growth factor receptor|stasis|vascular endothelial growth factor receptor|vascular endothelial growth factor receptor-1 like', map_location='28F4-28F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8220980, end_position_on_the_genomic_accession=8239878, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  34129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34129, CurrentID=0, Status='live', Symbol='CG14277', Aliases='Dmel_ Dmel\\CG14277', description='CG14277 gene product from transcript CG14277-RB', other_designations='CG14277-PA|CG14277-PB|uncharacterized protein', map_location='28F5-28F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8242486, end_position_on_the_genomic_accession=8243426, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34131, CurrentID=0, Status='live', Symbol='CG8086', Aliases='Dmel_ CG8113, Dmel\\CG8086', description='CG8086 gene product from transcript CG8086-RK', other_designations='CG8086-PA|CG8086-PB|CG8086-PG|CG8086-PH|CG8086-PI|CG8086-PJ|CG8086-PK|CG8086-PL|CG8086-PM|CG8086-PN|CG8086-PO|CG8086-PP', map_location='28F5-29A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8244160, end_position_on_the_genomic_accession=8257502, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  34132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34132, CurrentID=0, Status='live', Symbol='Btk29A', Aliases='Dmel_CG8049, Btk, Btk29a, C-src2, C-src4, CG18355, CG8049, CT2415, CT41718, DSrc28, DTec29, Dm SRC2, Dmel\\CG8049, Dsrc28C, Dsrc29A, EC2-9, S13, SRC 29A, Src2, Src29A, Tec, Tec29, Tec29A, btk, btk29, btk29A, btk29a, c-src/fps, dsrc29A, fic, src-4, src2, src28C, src4, tec29', description='Btk family kinase at 29A', other_designations='Btk29A type 2|Btk29A-PA|Btk29A-PB|Btk29A-PC|Btk29A-PD|Btk29A-PE|Btk29A-PF|Btk29A-PG|Btk29A-PH|CG8049-PA|CG8049-PB|CG8049-PC|CG8049-PD|CG8049-PE|CG8049-PF|CG8049-PG|CG8049-PH|Src oncogene 29A|fickle', map_location='29A1-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8258373, end_position_on_the_genomic_accession=8301079, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  34134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34134, CurrentID=0, Status='live', Symbol='wol', Aliases='Dmel_CG7870, ALG5, CG7870, Dmel\\CG7870, Wol, alg5, dAlg5', description='wollknaeuel', other_designations='CG7870-PA|wol-PA|wollknauel', map_location='29A3-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8301345, end_position_on_the_genomic_accession=8302985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34135, CurrentID=0, Status='live', Symbol='Scgalpha', Aliases='Dmel_CG7851, CG7851, DmSCG-alphaepsilon, Dmel\\CG7851, Scga', description='Sarcoglycan alpha', other_designations='CG7851-PA|CG7851-PB|CG7851-PC|Scgalpha-PA|Scgalpha-PB|Scgalpha-PC|sarcoglycan alpha', map_location='29A3-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8303429, end_position_on_the_genomic_accession=8305376, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34136, CurrentID=0, Status='live', Symbol='CG7840', Aliases='Dmel_ Dmel\\CG7840, cg7840', description='CG7840 gene product from transcript CG7840-RA', other_designations='CG7840-PA', map_location='29A3-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8305586, end_position_on_the_genomic_accession=8306741, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34137, CurrentID=0, Status='live', Symbol='Ostgamma', Aliases='Dmel_CG7830, CG7830, Dmel\\CG7830, MagT1, OST, dMagT1', description='Oligosaccharide transferase gamma subunit', other_designations='CG7830-PA|Ostgamma-PA|magnesium transporter subtype 1|oligosaccharide transferase gamma subunit', map_location='29A3-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8306696, end_position_on_the_genomic_accession=8308192, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34138, CurrentID=0, Status='live', Symbol='CG7818', Aliases='Dmel_ Dmel\\CG7818', description='CG7818 gene product from transcript CG7818-RA', other_designations='CG7818-PA', map_location='29A3-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8308427, end_position_on_the_genomic_accession=8309961, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34139, CurrentID=0, Status='live', Symbol='CG7810', Aliases='Dmel_ Dmel\\CG7810', description='CG7810 gene product from transcript CG7810-RA', other_designations='CG7810-PA', map_location='29A3-29A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8309876, end_position_on_the_genomic_accession=8311195, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34140, CurrentID=0, Status='live', Symbol='CG7806', Aliases='Dmel_ Dme_CG7806, Dmel\\CG7806', description='CG7806 gene product from transcript CG7806-RA', other_designations='CG7806-PA', map_location='29A3-29A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8311399, end_position_on_the_genomic_accession=8317011, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34141, CurrentID=0, Status='live', Symbol='mtsh', Aliases='Dmel_CG7795, CG7795, Dmel\\CG7795', description='mitoshell', other_designations='CG7795-PC|CG7795-PD|mtsh-PC|mtsh-PD', map_location='29A4-29A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8316889, end_position_on_the_genomic_accession=8319908, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34142, CurrentID=0, Status='live', Symbol='CG7787', Aliases='Dmel_ BcDNA:RE45155, Dmel\\CG7787', description='CG7787 gene product from transcript CG7787-RA', other_designations='CG7787-PA', map_location='29A4-29A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8320328, end_position_on_the_genomic_accession=8321285, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34143, CurrentID=0, Status='live', Symbol='CG7781', Aliases='Dmel_ Dmel\\CG7781', description='CG7781 gene product from transcript CG7781-RB', other_designations='CG7781-PA|CG7781-PB', map_location='29A4-29A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8321280, end_position_on_the_genomic_accession=8325409, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34144, CurrentID=0, Status='live', Symbol='CG14275', Aliases='Dmel_ BcDNA:RE73208, CT33900, Dmel\\CG14275', description='CG14275 gene product from transcript CG14275-RA', other_designations='CG14275-PA|CG14275-PB|uncharacterized protein', map_location='29A5-29A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8327456, end_position_on_the_genomic_accession=8332911, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34145, CurrentID=0, Status='live', Symbol='CG14274', Aliases='Dmel_ Dmel\\CG14274', description='CG14274 gene product from transcript CG14274-RA', other_designations='CG14274-PA|uncharacterized protein', map_location='29A5-29B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8338287, end_position_on_the_genomic_accession=8342056, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34146, CurrentID=0, Status='live', Symbol='CG14273', Aliases='Dmel_ Dmel\\CG14273', description='CG14273 gene product from transcript CG14273-RA', other_designations='CG14273-PA|uncharacterized protein', map_location='29B1-29B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8342695, end_position_on_the_genomic_accession=8344378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34147, CurrentID=0, Status='live', Symbol='CG7778', Aliases='Dmel_ Dmel\\CG7778', description='CG7778 gene product from transcript CG7778-RB', other_designations='CG7778-PA|CG7778-PB', map_location='29B1-29B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8346620, end_position_on_the_genomic_accession=8351112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34148, CurrentID=0, Status='live', Symbol='CG7627', Aliases='Dmel_ Dme_CG7627, Dmel\\CG7627', description='CG7627 gene product from transcript CG7627-RB', other_designations='CG7627-PA|CG7627-PB', map_location='29B1-29B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8355678, end_position_on_the_genomic_accession=8362469, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34149, CurrentID=0, Status='live', Symbol='RpS13', Aliases='Dmel_CG13389, CG13389, Dmel\\CG13389, M(2)29B, M(2)32, M(2)32A, RPS13, Rp S13, Rp17, S13, l(2)k09614', description='Ribosomal protein S13', other_designations='CG13389-PA|CG13389-PB|CG13389-PC|Minute(2)32A|RpS13-PA|RpS13-PB|RpS13-PC|ribosomal protein S13|ribosomal protein S17', map_location='29B1-29B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8363477, end_position_on_the_genomic_accession=8364662, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34152, CurrentID=0, Status='live', Symbol='CG17292', Aliases='Dmel_ Dmel\\CG17292', description='CG17292 gene product from transcript CG17292-RD', other_designations='CG17292-PA|CG17292-PB|CG17292-PC|CG17292-PD|uncharacterized protein', map_location='29B2-29B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8370454, end_position_on_the_genomic_accession=8374250, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34153, CurrentID=0, Status='live', Symbol='Wdr82', Aliases='Dmel_CG17293, CG17293, Dmel\\CG17293, WDR82, dWdr82, wdr-82, wdr82', description='WD repeat domain 82', other_designations='CG17293-PA|Wdr82-PA', map_location='29B2-29B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8374461, end_position_on_the_genomic_accession=8375549, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34154, CurrentID=0, Status='live', Symbol='CG13390', Aliases='Dmel_ Dmel\\CG13390', description='CG13390 gene product from transcript CG13390-RA', other_designations='CG13390-PA|uncharacterized protein', map_location='29B3-29B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8375502, end_position_on_the_genomic_accession=8376797, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34155, CurrentID=0, Status='live', Symbol='CG17294', Aliases='Dmel_ 15010468, Dmel\\CG17294, Phos17294', description='CG17294 gene product from transcript CG17294-RA', other_designations='CG17294-PA|uncharacterized protein', map_location='29B3-29B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8377035, end_position_on_the_genomic_accession=8378206, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34156, CurrentID=0, Status='live', Symbol='Aats-ala', Aliases='Dmel_CG13391, ARS, AlaRS, CG13391, Dmel\\CG13391, alaS', description='Alanyl-tRNA synthetase', other_designations='Aats-ala-PA|Aats-ala-PB|CG13391-PA|CG13391-PB', map_location='29B3-29B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8378262, end_position_on_the_genomic_accession=8381764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34157, CurrentID=0, Status='live', Symbol='CG13392', Aliases='Dmel_ Dmel\\CG13392', description='CG13392 gene product from transcript CG13392-RB', other_designations='CG13392-PA|CG13392-PB|uncharacterized protein', map_location='29B3-29B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8381821, end_position_on_the_genomic_accession=8382807, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34158, CurrentID=0, Status='live', Symbol='Rcd4', Aliases='Dmel_CG17295, BcDNA:SD16838, CG17295, Dmel\\CG17295', description='Reduction in Cnn dots 4', other_designations='CG17295-PA|CG17295-PB|Rcd4-PA|Rcd4-PB|reduction in Cnn dots 4', map_location='29B4-29B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8382861, end_position_on_the_genomic_accession=8383657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34159, CurrentID=0, Status='live', Symbol='l(2)k12914', Aliases='Dmel_CG13393, BcDNA:RE23864, CG13393, Dmel\\CG13393, OST', description='lethal (2) k12914', other_designations='CG13393-PA|l(2)k12914-PA', map_location='29C1-29B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8383612, end_position_on_the_genomic_accession=8384137, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34160, CurrentID=0, Status='live', Symbol='CG13384', Aliases='Dmel_ Dmel\\CG13384', description='CG13384 gene product from transcript CG13384-RD', other_designations='CG13384-PB|CG13384-PC|CG13384-PD|CG13384-PE|CG13384-PF|CG13384-PG|CG13384-PH|CG13384-PI|CG13384-PJ|CG13384-PK|CG13384-PL|uncharacterized protein', map_location='29B4-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8384455, end_position_on_the_genomic_accession=8388804, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34161, CurrentID=0, Status='live', Symbol='CG42820', Aliases='Dmel_ CG13394, Dmel\\CG42820, Dmel_CG13394', description='CG42820 gene product from transcript CG42820-RA', other_designations='CG42820-PA|uncharacterized protein', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8389881, end_position_on_the_genomic_accession=8392398, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34162, CurrentID=0, Status='live', Symbol='Acp29AB', Aliases='Dmel_CG17797, AC 004423B, AC004423b, Acp29-II, Acp29B, CG17797, Dmel\\CG17797, Acp29AB', description='Accessory gland protein 29AB', other_designations='Acp29AB-PA|CG17797-PA|accessory gland protein|accessory gland protein 29AB|accessory gland-specific peptide 29AB', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8392456, end_position_on_the_genomic_accession=8393222, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19835795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835795, CurrentID=0, Status='live', Symbol='CG45093', Aliases='Dmel_ Dmel\\CG45093', description='CG45093 gene product from transcript CG45093-RD', other_designations='CG45093-PA|CG45093-PB|CG45093-PC|CG45093-PD|uncharacterized protein', map_location='43E13-43E16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7737019, end_position_on_the_genomic_accession=7782637, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34164, CurrentID=0, Status='live', Symbol='CG13385', Aliases='Dmel_ Dmel\\CG13385', description='CG13385 gene product from transcript CG13385-RA', other_designations='CG13385-PA|uncharacterized protein', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8396649, end_position_on_the_genomic_accession=8397886, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34165, CurrentID=0, Status='live', Symbol='CG13386', Aliases='Dmel_ Dmel\\CG13386', description='CG13386 gene product from transcript CG13386-RA', other_designations='CG13386-PA|uncharacterized protein', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8399021, end_position_on_the_genomic_accession=8400078, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34166, CurrentID=0, Status='live', Symbol='fy', Aliases='Dmel_CG13396, CG13396, Dmel\\CG13396', description='fuzzy', other_designations='CG13396-PA|CG13396-PB|fy-PA|fy-PB', map_location='2-24.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8401476, end_position_on_the_genomic_accession=8403280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34167, CurrentID=0, Status='live', Symbol='emb', Aliases='Dmel_CG13387, CG13387, CRM1, Crm1, DCRM1, Dmel\\CG13387, Emb, XPO-1, XPO1, Xpo1, crm1, dCRM1, l(2)k16715', description='embargoed', other_designations='CG13387-PA|CG13387-PB|Exportin-1|chromosomal region maintenance 1|emb-PA|emb-PB|exportin', map_location='29C1-29C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8403573, end_position_on_the_genomic_accession=8408853, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34169, CurrentID=0, Status='live', Symbol='CG13398', Aliases='Dmel_ Dmel\\CG13398', description='CG13398 gene product from transcript CG13398-RB', other_designations='CG13398-PA|CG13398-PB|uncharacterized protein', map_location='29C3-29C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8412144, end_position_on_the_genomic_accession=8415256, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34170, CurrentID=0, Status='live', Symbol='Akap200', Aliases='Dmel_CG13388, AKAP200, Akap 200, BcDNA:LP02734, CG13388, DAKAP200, Dmel\\CG13388, EP(2)2254, MESR2, akap200, anon-WO0118547.134, anon-WO0118547.213, dpo', description='A kinase anchor protein 200', other_designations='A kinase anchor protein|Akap200-PA|Akap200-PB|Akap200-PC|Akap200-PD|Akap200-PE|Akap200-PF|Akap200-PG|CG13388-PA|CG13388-PB|CG13388-PC|CG13388-PD|CG13388-PE|CG13388-PF|CG13388-PG', map_location='29C3-29C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8415690, end_position_on_the_genomic_accession=8431297, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34171, CurrentID=0, Status='live', Symbol='grk', Aliases='Dmel_CG17610, CG17610, CT32746, Dmel\\CG17610, GRK, Grk, s-Grk', description='gurken', other_designations='CG17610-PA|grk-PA', map_location='2-30 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8431086, end_position_on_the_genomic_accession=8433598, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34172, CurrentID=0, Status='live', Symbol='D12', Aliases='Dmel_CG13400, CG13400, Dmel\\CG13400, Yeats2', description='CG13400 gene product from transcript CG13400-RA', other_designations='CG13400-PA|D12-PA', map_location='29C4-29C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8434058, end_position_on_the_genomic_accession=8437365, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34175, CurrentID=0, Status='live', Symbol='U26', Aliases='Dmel_CG13401, CG13401, Dmel\\CG13401', description='CG13401 gene product from transcript CG13401-RA', other_designations='CG13401-PA|CG13401-PC|CG13401-PD|U26-PA|U26-PC|U26-PD', map_location='29C5-29C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8445861, end_position_on_the_genomic_accession=8449576, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34176, CurrentID=0, Status='live', Symbol='fu12', Aliases='Dmel_CG17608, AGPAT1, AGPAT2, CG17608, Dmel\\CG17608', description='CG17608 gene product from transcript CG17608-RA', other_designations='CG17608-PA|CG17608-PB|CG17608-PC|CG17608-PD|fu12-PA|fu12-PB|fu12-PC|fu12-PD', map_location='29C5-29D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8449691, end_position_on_the_genomic_accession=8463994, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34177, CurrentID=0, Status='live', Symbol='fu2', Aliases='Dmel_CG9233, CG9233, Dmel\\CG9233, FU12', description='CG9233 gene product from transcript CG9233-RB', other_designations='CG9233-PA|CG9233-PB|fu2-PA|fu2-PB', map_location='29D1-29D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8464488, end_position_on_the_genomic_accession=8466694, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34179, CurrentID=0, Status='live', Symbol='d', Aliases='Dmel_CG42840, 29C3-D1, 29CD, AAF52683, CG10595, CG13087, CG31610, CG42840, D, DmMyo20, Dmel\\CG42840, Dmel_CG10595, Dmel_CG13087, Ds, Myo29D, anon-EST:Liang-1.78, clone 1.78, l(2)c00146', description='dachs', other_designations='CG42840-PD|CG42840-PE|d-PD|d-PE|lethal (2) c00146|myosin 29D|unconventional myosin 29D', map_location='2-31.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8466817, end_position_on_the_genomic_accession=8488671, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34180, CurrentID=0, Status='live', Symbol='CG13088', Aliases='Dmel_ Dmel\\CG13088, FBXO15', description='CG13088 gene product from transcript CG13088-RA', other_designations='CG13088-PA|uncharacterized protein', map_location='29D1-29D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8488981, end_position_on_the_genomic_accession=8490546, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34181, CurrentID=0, Status='live', Symbol='Dh31', Aliases='Dmel_CG13094, CG13094, DH 31, DH-II, DH31, DH[[31], DH[[31]], Dh[[31]], Dmel\\CG13094, Drm-DH[[31]], Drome-DH[[31]], anon-EST:Posey114, anon-WO02059370.45, dh31, drome-DH31', description='Diuretic hormone 31', other_designations='CG13094-PA|CG13094-PC|Calcitonin-like diuretic hormone (31 residues)|Dh31-PA|Dh31-PC|calcitonin|diuretic hormone|diuretic hormone 31', map_location='29D1-29D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8491868, end_position_on_the_genomic_accession=8506845, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34182, CurrentID=0, Status='live', Symbol='Bace', Aliases='Dmel_CG13095, CG13095, Dmel\\CG13095, anon-WO0140519.142, dBACE, dBace', description='beta-site APP-cleaving enzyme', other_designations='Bace-PA|Bace-PB|CG13095-PA|CG13095-PB', map_location='29D2-29D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8494893, end_position_on_the_genomic_accession=8496307, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34183, CurrentID=0, Status='live', Symbol='CG13096', Aliases='Dmel_ Dmel\\CG13096', description='CG13096 gene product from transcript CG13096-RB', other_designations='CG13096-PA|CG13096-PB|uncharacterized protein', map_location='29D3-29D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8509069, end_position_on_the_genomic_accession=8511866, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34184, CurrentID=0, Status='live', Symbol='CG13097', Aliases='Dmel_ Dmel\\CG13097', description='CG13097 gene product from transcript CG13097-RA', other_designations='CG13097-PA|uncharacterized protein', map_location='29D3-29D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8512174, end_position_on_the_genomic_accession=8514868, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34185, CurrentID=0, Status='live', Symbol='PIG-U', Aliases='Dmel_CG13089, CG13089, Dmel\\CG13089', description='Phosphatidylinositol glycan anchor biosynthesis, class U ortholog (H. sapiens)', other_designations='CG13089-PA|PIG-U-PA|phosphatidylinositol glycan anchor biosynthesis, class U ortholog', map_location='29D4-29D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8515096, end_position_on_the_genomic_accession=8516755, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34186, CurrentID=0, Status='live', Symbol='mRpL51', Aliases='Dmel_CG13098, CG13098, Dmel\\CG13098, MRP-L51', description='mitochondrial ribosomal protein L51', other_designations='CG13098-PA|mRpL51-PA', map_location='29D4-29D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8516753, end_position_on_the_genomic_accession=8517380, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34187, CurrentID=0, Status='live', Symbol='CG13090', Aliases='Dmel_ Dmel\\CG13090, dMoaD', description='CG13090 gene product from transcript CG13090-RC', other_designations='CG13090-PA|CG13090-PB|CG13090-PC|Uba4 homologue|uncharacterized protein', map_location='29D4-29D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8517411, end_position_on_the_genomic_accession=8519176, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34188, CurrentID=0, Status='live', Symbol='CG13091', Aliases='Dmel_ CT32316, Dmel\\CG13091', description='CG13091 gene product from transcript CG13091-RB', other_designations='CG13091-PA|CG13091-PB|uncharacterized protein', map_location='29D4-29D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8519690, end_position_on_the_genomic_accession=8521809, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34189, CurrentID=0, Status='live', Symbol='Acer', Aliases='Dmel_CG10593, ACE, ACER, CG10593, Dmel\\CG10593, acer, l(2)k07704', description='Angiotensin-converting enzyme-related', other_designations='Acer-PA|Acer-PB|CG10593-PA|CG10593-PB|angiotensin-converting enzyme-related', map_location='29D4-29D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8521933, end_position_on_the_genomic_accession=8525740, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34191, CurrentID=0, Status='live', Symbol='CG17834', Aliases='Dmel_ BcDNA:LP07538, Dmel\\CG17834, anon-290a, anon-29Da', description='CG17834 gene product from transcript CG17834-RA', other_designations='CG17834-PA|CG17834-PB|CG17834-PC|CG17834-PD|CG17834-PE|CG17834-PF|uncharacterized protein', map_location='29D5-29E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8529141, end_position_on_the_genomic_accession=8541439, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34192, CurrentID=0, Status='live', Symbol='Sema-1a', Aliases='Dmel_CG18405, BcDNA:RE36155, CG18405, CT26404, D-Sema 1a, D-Sema I, D-Sema-I, D-SemaI, D-semaI, Dmel\\CG18405, Dsema 1a, Dsema-I, Fas-IV, FasIV, SEMA-1A, Sema 1a, Sema-I, Sema1A, Sema1a, SemaI, d-sema I, d-semaI, l(2)k13702, sema, sema 1a, sema-1a, sema-I, sema1a', description='CG18405 gene product from transcript CG18405-RE', other_designations='CG18405-PA|CG18405-PD|CG18405-PE|CG18405-PF|CG18405-PG|CG18405-PH|CG18405-PI|Fasciclin-IV|Sema-1a-PA|Sema-1a-PD|Sema-1a-PE|Sema-1a-PF|Sema-1a-PG|Sema-1a-PH|Sema-1a-PI|Semaphorin-1|Semaphorin-1A|semaphorin|semaphorin 1|semaphorin 1a|semaphorin I|semaphorin1a|semiphorin', map_location='29E1-29E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8542147, end_position_on_the_genomic_accession=8672041, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  34193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34193, CurrentID=0, Status='live', Symbol='Glt', Aliases='Dmel_CG9280, CG9280, CT26441, Dmel\\CG9280', description='Glutactin', other_designations='CG9280-PA|CG9280-PD|Glt-PA|Glt-PD|glutactin', map_location='29E3-29E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8672698, end_position_on_the_genomic_accession=8676804, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34194, CurrentID=0, Status='live', Symbol='CG9287', Aliases='Dmel_ Dmel\\CG9287', description='CG9287 gene product from transcript CG9287-RA', other_designations='CG9287-PA', map_location='29E3-29E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8676777, end_position_on_the_genomic_accession=8679812, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34195, CurrentID=0, Status='live', Symbol='CG9289', Aliases='Dmel_ CT26467, Dmel\\CG9289', description='CG9289 gene product from transcript CG9289-RA', other_designations='CG9289-PA|glutactin-like', map_location='29E3-29E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8681325, end_position_on_the_genomic_accession=8684376, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34196, CurrentID=0, Status='live', Symbol='PrBP', Aliases='Dmel_CG9296, CG9296, DmPrBP, DmPrBP/delta, Dmel\\CG9296', description='Prenyl-binding protein', other_designations='CG9296-PA|PrBP-PA', map_location='29E3-29E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8684823, end_position_on_the_genomic_accession=8685955, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34197, CurrentID=0, Status='live', Symbol='Trs23', Aliases='Dmel_CG9298, CG9298, Dmel\\CG9298, TRS23p, Trs23', description='TRAPP subunit 23 ortholog (S. cerevisiae)', other_designations='CG9298-PA|TRAPP subunit 23 ortholog|Trs23-PA', map_location='29E3-29E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8686231, end_position_on_the_genomic_accession=8687052, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34198, CurrentID=0, Status='live', Symbol='CG9314', Aliases='Dmel_ CT26521, CatB, Dmel\\CG9314, anon-WO0140519.249', description='CG9314 gene product from transcript CG9314-RC', other_designations='CG9314-PA|CG9314-PC|catalase B|kitty', map_location='29E4-29E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8705572, end_position_on_the_genomic_accession=8707527, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34202, CurrentID=0, Status='live', Symbol='CG12438', Aliases='Dmel_ Dmel\\CG12438', description='CG12438 gene product from transcript CG12438-RA', other_designations='CG12438-PA|uncharacterized protein', map_location='29E6-29E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8752285, end_position_on_the_genomic_accession=8753209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34203, CurrentID=0, Status='live', Symbol='LManIII', Aliases='Dmel_CG9463, CG9463, Dmel\\CG9463', description='Lysosomal alpha-mannosidase III', other_designations='CG9463-PA|CG9463-PB|LManIII-PA|LManIII-PB|lysosomal alpha-mannosidase III', map_location='29F1-29F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8765273, end_position_on_the_genomic_accession=8768710, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34204, CurrentID=0, Status='live', Symbol='LManIV', Aliases='Dmel_CG9465, CG9465, Dmel\\CG9465', description='Lysosomal alpha-mannosidase IV', other_designations='CG9465-PB|LManIV-PB|lysosomal alpha-mannosidase IV', map_location='29F1-29F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8769159, end_position_on_the_genomic_accession=8772344, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34205, CurrentID=0, Status='live', Symbol='LManV', Aliases='Dmel_CG9466, CG9466, Dmel\\CG9466', description='Lysosomal alpha-mannosidase V', other_designations='CG9466-PA|CG9466-PB|LManV-PA|LManV-PB|lysosomal alpha-mannosidase V', map_location='29F1-29F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8772946, end_position_on_the_genomic_accession=8776263, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34206, CurrentID=0, Status='live', Symbol='LManVI', Aliases='Dmel_CG9468, CG9468, Dmel\\CG9468', description='Lysosomal alpha-mannosidase VI', other_designations='CG9468-PA|LManVI-PA|lysosomal alpha-mannosidase VI', map_location='29F1-29F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8777661, end_position_on_the_genomic_accession=8781252, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5740101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740101, CurrentID=0, Status='live', Symbol='CG34459', Aliases='Dmel_ CG6301, Dmel\\CG34459, Dmel_CG6301', description='CG34459 gene product from transcript CG34459-RC', other_designations='CG34459-PA|CG34459-PB|CG34459-PC|uncharacterized protein', map_location='53D11-53D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16824664, end_position_on_the_genomic_accession=16826648, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34208, CurrentID=0, Status='live', Symbol='CG32986', Aliases='Dmel_ CG15866, Dmel\\CG32986', description='CG32986 gene product from transcript CG32986-RA', other_designations='CG32986-PA|uncharacterized protein', map_location='29F3-29F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8862776, end_position_on_the_genomic_accession=8863743, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34209, CurrentID=0, Status='live', Symbol='CG9483', Aliases='Dmel_ Dmel\\CG9483, anon-WO0140519.161', description='CG9483 gene product from transcript CG9483-RA', other_designations='CG9483-PA', map_location='29F3-29F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8866767, end_position_on_the_genomic_accession=8867686, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34211, CurrentID=0, Status='live', Symbol='Tsp29Fa', Aliases='Dmel_CG9494, CG9494, Dm.Tsp29Fa, Dmel\\CG9494', description='Tetraspanin 29Fa', other_designations='CG9494-PA|CG9494-PB|CG9494-PC|Tsp29Fa-PA|Tsp29Fa-PB|Tsp29Fa-PC|tetraspanin 29Fa', map_location='29F4-29F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8933962, end_position_on_the_genomic_accession=8936096, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34212, CurrentID=0, Status='live', Symbol='Tsp29Fb', Aliases='Dmel_CG9496, CG9496, Dm.Tsp29Fb, Dmel\\CG9496, anon-EST:ParkEST130', description='Tetraspanin 29Fb', other_designations='CG9496-PA|CG9496-PB|CG9496-PC|Tsp29Fb-PA|Tsp29Fb-PB|Tsp29Fb-PC|tetraspanin 29Fb', map_location='29F5-29F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8937390, end_position_on_the_genomic_accession=8939743, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34215, CurrentID=0, Status='live', Symbol='C1GalTA', Aliases='Dmel_CG9520, C1GALT1, C1beta3GalT1, CG9520, Dmel\\CG9520, anon-WO0144478.18, dC1GalT1, dC1beta3GalT1', description='Core 1 Galactosyltransferase A', other_designations='C1GalTA-PA|C1GalTA-PB|C1GalTA-PC|CG9520-PA|CG9520-PB|CG9520-PC|core 1 beta1,3-galactosyltransferase|core 1 galactosyltransferase A', map_location='29F5-29F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8943455, end_position_on_the_genomic_accession=8951216, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34217, CurrentID=0, Status='live', Symbol='CG9525', Aliases='Dmel_ Dmel\\CG9525, gi7297470', description='CG9525 gene product from transcript CG9525-RB', other_designations='CG9525-PA|CG9525-PB', map_location='29F5-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8959255, end_position_on_the_genomic_accession=8962354, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34218, CurrentID=0, Status='live', Symbol='CG32984', Aliases='Dmel_ CG17789, Dmel\\CG32984, gi21430910, gi22945998', description='CG32984 gene product from transcript CG32984-RA', other_designations='CG32984-PA|uncharacterized protein', map_location='29F6-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8964982, end_position_on_the_genomic_accession=8967035, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34219, CurrentID=0, Status='live', Symbol='CG18088', Aliases='Dmel_ Dmel\\CG18088, gi21430616, gi7297472', description='CG18088 gene product from transcript CG18088-RC', other_designations='CG18088-PA|CG18088-PB|CG18088-PC|uncharacterized protein', map_location='29F6-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8967535, end_position_on_the_genomic_accession=8970627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34220, CurrentID=0, Status='live', Symbol='CG9541', Aliases='Dmel_ Dmel\\CG9541', description='CG9541 gene product from transcript CG9541-RC', other_designations='CG9541-PC|CG9541-PD', map_location='29F6-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8970520, end_position_on_the_genomic_accession=8975428, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34221, CurrentID=0, Status='live', Symbol='CG13101', Aliases='Dmel_ Dmel\\CG13101', description='CG13101 gene product from transcript CG13101-RB', other_designations='CG13101-PA|CG13101-PB|uncharacterized protein', map_location='29F6-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8977834, end_position_on_the_genomic_accession=8984872, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34222, CurrentID=0, Status='live', Symbol='rost', Aliases='Dmel_CG9552, CG9552, Dmel\\CG9552', description='rolling stone', other_designations='CG9552-PA|CG9552-PB|CG9552-PC|rost-PA|rost-PB|rost-PC', map_location='29F6-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8985936, end_position_on_the_genomic_accession=8989539, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34223, CurrentID=0, Status='live', Symbol='CG9555', Aliases='Dmel_ Dmel\\CG9555', description='CG9555 gene product from transcript CG9555-RC', other_designations='CG9555-PA|CG9555-PC', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8991735, end_position_on_the_genomic_accession=8993632, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34224, CurrentID=0, Status='live', Symbol='CG17906', Aliases='Dmel_ Dmel\\CG17906', description='CG17906 gene product from transcript CG17906-RA', other_designations='CG17906-PA|uncharacterized protein', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8994391, end_position_on_the_genomic_accession=8995617, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34225, CurrentID=0, Status='live', Symbol='alien', Aliases='Dmel_CG9556, Alien, BcDNA:LD10463, CG9556, CH2, CSN2, Csn2, DCH2, Dch2, Dmel\\CG9556', description='CG9556 gene product from transcript CG9556-RD', other_designations='CG9556-PA|CG9556-PB|CG9556-PD|CG9556-PE|COP9 complex homolog subunit 2|alien-PA|alien-PB|alien-PD|alien-PE|d-Alien|drosophila COP9 signalosome homolog 2', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8993818, end_position_on_the_genomic_accession=8997379, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34226, CurrentID=0, Status='live', Symbol='Try29F', Aliases='Dmel_CG9564, CG9564, Dmel\\CG9564, SP106, try29F, try29F.2', description='Trypsin 29F', other_designations='CG9564-PC|CG9564-PD|Try29F-PC|Try29F-PD|trypsin 29F', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8998537, end_position_on_the_genomic_accession=8999816, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34227, CurrentID=0, Status='live', Symbol='CG9568', Aliases='Dmel_ Dmel\\CG9568, FBgn 32087, anon-WO0140519.183', description='CG9568 gene product from transcript CG9568-RA', other_designations='CG9568-PA', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9010583, end_position_on_the_genomic_accession=9011358, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34228, CurrentID=0, Status='live', Symbol='CG13102', Aliases='Dmel_ BcDNA:RE57454, Dmel\\CG13102', description='CG13102 gene product from transcript CG13102-RB', other_designations='CG13102-PA|CG13102-PB|uncharacterized protein', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9012188, end_position_on_the_genomic_accession=9014149, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34229, CurrentID=0, Status='live', Symbol='Rcd-1r', Aliases='Dmel_CG9573, CG9573, Dmel\\CG9573, Drcd-1r, Zeus', description='Rcd-1 related', other_designations='CG9573-PA|Drcd-1 related|Rcd-1r-PA', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9012885, end_position_on_the_genomic_accession=9014068, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34230, CurrentID=0, Status='live', Symbol='CG9582', Aliases='Dmel_ Dmel\\CG9582', description='CG9582 gene product from transcript CG9582-RB', other_designations='CG9582-PA|CG9582-PB', map_location='29F8-29F8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9024320, end_position_on_the_genomic_accession=9031836, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34231, CurrentID=0, Status='live', Symbol='DIP-zeta', Aliases='Dmel_CG31708, 31078, CG13103, CG13104, CG13105, CG31078, CG31708, CT32337, CT32338, Dmel\\CG31708', description='Dpr-interacting protein zeta', other_designations='CG31708-PA|CG31708-PB|CG31708-PC|DIP-zeta-PA|DIP-zeta-PB|DIP-zeta-PC', map_location='29F8-30A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9051631, end_position_on_the_genomic_accession=9071164, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34234, CurrentID=0, Status='live', Symbol='CG12439', Aliases='Dmel_ BcDNA:AT16114, Dmel\\CG12439', description='CG12439 gene product from transcript CG12439-RB', other_designations='CG12439-PA|CG12439-PB|uncharacterized protein', map_location='30A1-30A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9077121, end_position_on_the_genomic_accession=9077988, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34235, CurrentID=0, Status='live', Symbol='Toll-4', Aliases='Dmel_CG18241, CG18241, CT29238, Dmel\\CG18241, TL4, Tl-4, dTLR4, dToll4, toll', description='CG18241 gene product from transcript CG18241-RA', other_designations='CG18241-PA|Toll-4-PA|Toll-like|toll receptor', map_location='30A1-30A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9084107, end_position_on_the_genomic_accession=9089440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34236, CurrentID=0, Status='live', Symbol='Or30a', Aliases='Dmel_CG13106, 30A.1, 30a, CG13106, DOR56, Dmel\\CG13106', description='Odorant receptor 30a', other_designations='CG13106-PA|Or30a-PA|odorant receptor 30a', map_location='30A1-30A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9111814, end_position_on_the_genomic_accession=9113285, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34237, CurrentID=0, Status='live', Symbol='CG32982', Aliases='Dmel_ CG18566, CG9584, CG9585, Dmel\\CG32982', description='CG32982 gene product from transcript CG32982-RF', other_designations='CG32982-PA|CG32982-PB|CG32982-PD|CG32982-PE|CG32982-PF|CG32982-PG|CG32982-PH|uncharacterized protein', map_location='30A2-30A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9125737, end_position_on_the_genomic_accession=9163721, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  34240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34240, CurrentID=0, Status='live', Symbol='CG13108', Aliases='Dmel_ Dmel\\CG13108', description='CG13108 gene product from transcript CG13108-RA', other_designations='CG13108-PA|uncharacterized protein', map_location='30A2-30A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9164235, end_position_on_the_genomic_accession=9165426, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34241, CurrentID=0, Status='live', Symbol='CG9586', Aliases='Dmel_ Dmel\\CG9586', description='CG9586 gene product from transcript CG9586-RB', other_designations='CG9586-PB|CG9586-PC', map_location='30A2-30A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9165421, end_position_on_the_genomic_accession=9166514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34242, CurrentID=0, Status='live', Symbol='tai', Aliases='Dmel_CG13109, 58/2, CG13109, CG18494, DAIB1, DmTaiman, Dmel\\CG13109, EP(2)2630, NEST:bs13g05, TAI, Tai, bs13g05.y1, l(2)01351, l(2)k05802, l(2)k05809', description='taiman', other_designations='CG13109-PC|CG13109-PD|CG13109-PE|CG13109-PF|CG13109-PG|CG13109-PH|drosophila amplified in breast cancer-1|tai-PC|tai-PD|tai-PE|tai-PF|tai-PG|tai-PH', map_location='30A2-30A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9166778, end_position_on_the_genomic_accession=9250211, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  34244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34244, CurrentID=0, Status='live', Symbol='aust', Aliases='Dmel_CG17009, CG17009, Dmel\\CG17009', description='australin', other_designations='CG17009-PA|CG17009-PB|aust-PA|aust-PB', map_location='30A6-30A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9250048, end_position_on_the_genomic_accession=9251042, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34245, CurrentID=0, Status='live', Symbol='borr', Aliases='Dmel_CG4454, BORR, Bor, Borr, CG4454, Dmel\\CG4454, dBor', description='borealin-related', other_designations='Borealin-related|CG4454-PA|CG4454-PB|borealin|borr-PA|borr-PB', map_location='30A6-30A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9251472, end_position_on_the_genomic_accession=9253118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34249, CurrentID=0, Status='live', Symbol='CG17005', Aliases='Dmel_ Dmel\\CG17005', description='CG17005 gene product from transcript CG17005-RE', other_designations='CG17005-PB|CG17005-PC|CG17005-PD|CG17005-PE|uncharacterized protein', map_location='30A8-30A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9327744, end_position_on_the_genomic_accession=9330332, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34250, CurrentID=0, Status='live', Symbol='Cks30A', Aliases='Dmel_CG3738, CDI2, CG3738, Cdi2, Cks, DmCks1, Dmel\\CG3738, U40077, cdi2, cks, cks30A, fs(2)rem, rem', description='Cyclin-dependent kinase subunit 30A', other_designations='CG3738-PA|Cks30A-PA|Cyclin-dependent kinase interactor 2|remnants', map_location='2-32 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9330763, end_position_on_the_genomic_accession=9331981, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34251, CurrentID=0, Status='live', Symbol='Eaat1', Aliases='Dmel_CG3747, AF001784, CG3747, Dglt-1, Dglut-1, Dmel\\CG3747, EAAT, EAAT1, dEAAT, dEAAT1, dEaat1, eaat1', description='Excitatory amino acid transporter 1', other_designations='CG3747-PA|CG3747-PB|CG3747-PC|CG3747-PD|Eaat1-PA|Eaat1-PB|Eaat1-PC|Eaat1-PD|drosophila glutamate transporter-1|excitatory amino acid transporter 1|glutamate transporter|sodium-dependent excitatory amino acid transporter 1|sodium-dependent glutamate transporter|sodium-excitatory-glutamate-cotransporter', map_location='30A8-30A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9333567, end_position_on_the_genomic_accession=9341925, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  34252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34252, CurrentID=0, Status='live', Symbol='CG3748', Aliases='Dmel_ BcDNA:AT30262, Dmel\\CG3748, anon-WO0140519.143', description='CG3748 gene product from transcript CG3748-RC', other_designations='CG3748-PB|CG3748-PC', map_location='30A8-30A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9342832, end_position_on_the_genomic_accession=9344777, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34253, CurrentID=0, Status='live', Symbol='CG13110', Aliases='Dmel_ BcDNA:AT19071, Dmel\\CG13110', description='CG13110 gene product from transcript CG13110-RA', other_designations='CG13110-PA|uncharacterized protein', map_location='30A9-30A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9350470, end_position_on_the_genomic_accession=9350996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34256, CurrentID=0, Status='live', Symbol='Aldh', Aliases='Dmel_CG3752, 146084_at, ACDH, ALDH, ALDHA1, CG 3752, CG3752, Dmel\\CG3752', description='Aldehyde dehydrogenase', other_designations='Aldh-PA|Aldh-PB|CG3752-PA|CG3752-PB|acetaldehyde dehydrogenase|aldehyde dehydrogenase', map_location='30B1-30B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9387389, end_position_on_the_genomic_accession=9391567, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34257, CurrentID=0, Status='live', Symbol='CG4438', Aliases='Dmel_ Dmel\\CG4438', description='CG4438 gene product from transcript CG4438-RB', other_designations='CG4438-PA|CG4438-PB', map_location='30B1-30B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9396999, end_position_on_the_genomic_accession=9397988, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34258, CurrentID=0, Status='live', Symbol='Mco1', Aliases='Dmel_CG3759, CG3759, Dmel\\CG3759, MCO1', description='Multicopper oxidase-1', other_designations='CG3759-PA|CG3759-PB|CG3759-PC|Mco1-PA|Mco1-PB|Mco1-PC|laccase-like|multicopper oxidase-1', map_location='30B2-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9416588, end_position_on_the_genomic_accession=9426342, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34259, CurrentID=0, Status='live', Symbol='Fbp2', Aliases='Dmel_CG3763, CG3763, Dmel\\CG3763, Fbp-2, P6, T19', description='Fat body protein 2', other_designations='CG3763-PA|CG3763-PB|CG3763-PC|Fbp2-PA|Fbp2-PB|Fbp2-PC|Protein-6|fat body protein P6', map_location='30B3-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9426784, end_position_on_the_genomic_accession=9427917, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34260, CurrentID=0, Status='live', Symbol='FucTB', Aliases='Dmel_CG4435, CG4435, CT14430, Dmel\\CG4435', description='CG4435 gene product from transcript CG4435-RD', other_designations='CG4435-PA|CG4435-PC|CG4435-PD|FucTB-PA|FucTB-PC|FucTB-PD|alpha1,3-fucosyltransferase', map_location='30B3-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9428669, end_position_on_the_genomic_accession=9431052, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34262, CurrentID=0, Status='live', Symbol='CG3769', Aliases='Dmel_ Dmel\\CG3769', description='CG3769 gene product from transcript CG3769-RA', other_designations='CG3769-PA', map_location='30B3-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9435399, end_position_on_the_genomic_accession=9436842, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34263, CurrentID=0, Status='live', Symbol='numb', Aliases='Dmel_CG3779, CG3779, Dmel\\CG3779, N7-1, Nb, Numb, d-numb, dNumb, l(2)03235, l(2)3235, l(2)s2201, nbs', description='CG3779 gene product from transcript CG3779-RB', other_designations='CG3779-PA|CG3779-PB|CG3779-PD|d-Numb|numb-PA|numb-PB|numb-PD', map_location='30B3-30B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9437469, end_position_on_the_genomic_accession=9464184, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34264, CurrentID=0, Status='live', Symbol='Gdi', Aliases='Dmel_CG4422, CG4422, Dmel\\CG4422, GDI, GdI, N7-3, dGDI, l(2)N7-3, lincRNA.S746, rab GDI', description='GDP dissociation inhibitor', other_designations='CG4422-PA|CG4422-PB|CG4422-PC|Gdi-PA|Gdi-PB|Gdi-PC|Gdi-related|rab GDP dissociation inhibitor', map_location='30B5-30B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9491176, end_position_on_the_genomic_accession=9494876, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34268, CurrentID=0, Status='live', Symbol='Oatp30B', Aliases='Dmel_CG3811, CG3811, Dmel\\CG3811, oatp 30B, oatp30B', description='Organic anion transporting polypeptide 30B', other_designations='CG3811-PA|CG3811-PB|CG3811-PC|CG3811-PD|Oatp30B-PA|Oatp30B-PB|Oatp30B-PC|Oatp30B-PD|organic anion transporting polypeptide 30B', map_location='30B10-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9521210, end_position_on_the_genomic_accession=9540060, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34269, CurrentID=0, Status='live', Symbol='CG17855', Aliases='Dmel_ Dmel\\CG17855', description='CG17855 gene product from transcript CG17855-RA', other_designations='CG17855-PA|uncharacterized protein', map_location='30B10-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9540523, end_position_on_the_genomic_accession=9542480, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34270, CurrentID=0, Status='live', Symbol='Cpr30B', Aliases='Dmel_CG3818, CG3818, DmelCpr30B, Dmel\\CG3818', description='Cuticular protein 30B', other_designations='CG3818-PA|Cpr30B-PA|cuticular protein 30B', map_location='30B10-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9542878, end_position_on_the_genomic_accession=9543658, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34271, CurrentID=0, Status='live', Symbol='CG13113', Aliases='Dmel_ Dmel\\CG13113', description='CG13113 gene product from transcript CG13113-RA', other_designations='CG13113-PA|uncharacterized protein', map_location='30B10-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9544171, end_position_on_the_genomic_accession=9544875, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34272, CurrentID=0, Status='live', Symbol='CG13114', Aliases='Dmel_ Dmel\\CG13114', description='CG13114 gene product from transcript CG13114-RB', other_designations='CG13114-PB|uncharacterized protein', map_location='30B10-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9545811, end_position_on_the_genomic_accession=9547424, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34274, CurrentID=0, Status='live', Symbol='jp', Aliases='Dmel_CG4405, CG13115, CG4405, Dmel\\CG4405, JP', description='junctophilin', other_designations='CG4405-PA|CG4405-PB|CG4405-PC|CG4405-PD|CG4405-PE|jp-PA|jp-PB|jp-PC|jp-PD|jp-PE', map_location='30B10-30B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9547720, end_position_on_the_genomic_accession=9566750, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  34275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34275, CurrentID=0, Status='live', Symbol='CG3838', Aliases='Dmel_ Dmel\\CG3838, NEST:bs09e08, anon-WO0118547.445, bs09e08.y1', description='CG3838 gene product from transcript CG3838-RA', other_designations='CG3838-PA|CG3838-PB', map_location='30B12-30B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9568890, end_position_on_the_genomic_accession=9573572, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34276, CurrentID=0, Status='live', Symbol='Mtpalpha', Aliases='Dmel_CG4389, BcDNA:GH12558, CG 4389, CG4389, Dmel\\CG4389', description='Mitochondrial trifunctional protein alpha subunit', other_designations='CG4389-PA|CG4389-PB|Mtpalpha-PA|Mtpalpha-PB|mitochondrial trifunctional protein alpha subunit', map_location='30B12-30B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9573563, end_position_on_the_genomic_accession=9576769, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34277, CurrentID=0, Status='live', Symbol='gcm', Aliases='Dmel_CG12245, CG12245, Dmel\\CG12245, GCM, GCM1, GCMa, Gcm, N7-4, glide, glide/gcm, l(2)N7-4, ucc', description='glial cells missing', other_designations='CG12245-PA|CG12245-PB|gcm-PA|gcm-PB|glia cells missing|glial cell deficient|glial cell missing|glial cells-missing|glide|glide glide2|upper class chordotonals', map_location='30B12-30B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9579498, end_position_on_the_genomic_accession=9581745, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34278, CurrentID=0, Status='live', Symbol='CG3841', Aliases='Dmel_ Dmel\\CG3841', description='CG3841 gene product from transcript CG3841-RB', other_designations='CG3841-PA|CG3841-PB', map_location='30B12-30C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9589476, end_position_on_the_genomic_accession=9591783, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34279, CurrentID=0, Status='live', Symbol='CG4382', Aliases='Dmel_ Dmel\\CG4382', description='CG4382 gene product from transcript CG4382-RB', other_designations='CG4382-PB', map_location='30C1-30C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9598880, end_position_on_the_genomic_accession=9600872, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34280, CurrentID=0, Status='live', Symbol='gcm2', Aliases='Dmel_CG3858, CG3858, Dmel\\CG3858, FBgn0019809, Gcm2, dGCM2', description='CG3858 gene product from transcript CG3858-RA', other_designations='CG3858-PA|gcm2-PA|glial cells missing 2|glide glide2|glide2', map_location='30C1-30C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9608562, end_position_on_the_genomic_accession=9612706, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34281, CurrentID=0, Status='live', Symbol='Trx-2', Aliases='Dmel_CG31884, 31884, CG31884, CG3864, DTrx-2, DmTrx-2, DmelTrx-2, Dmel\\CG31884, Trx, anon-WO0118547.188, dmtrx-2', description='thioredoxin-2', other_designations='CG31884-PA|CG31884-PB|Thioredoxin-2|Trx-2-PA|Trx-2-PB|thioredoxin', map_location='30C1-30C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9613166, end_position_on_the_genomic_accession=9616404, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34282, CurrentID=0, Status='live', Symbol='GlcAT-S', Aliases='Dmel_CG3881, CG3881, D DmGlcAT-BSI, Dmel\\CG3881, dGlcAT-BSI', description='CG3881 gene product from transcript CG3881-RA', other_designations='CG3881-PA|CG3881-PB|CG3881-PC|GlcAT-S-PA|GlcAT-S-PB|GlcAT-S-PC', map_location='30C1-30C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9616469, end_position_on_the_genomic_accession=9623109, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34283, CurrentID=0, Status='live', Symbol='Apoltp', Aliases='Dmel_CG15828, ApoLTPI, ApoLTPII, CG15828, CT39214, Dmel\\CG15828, LTP, apoLTP, apoltp', description='Apolipoprotein lipid transfer particle', other_designations='Apoltp-PC|CG15828-PC|apolipoprotein lipid transfer particle|lipid transfer particle', map_location='30C2-30C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9634179, end_position_on_the_genomic_accession=9656026, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  34284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34284, CurrentID=0, Status='live', Symbol='Pka-C1', Aliases='Dmel_CG4379, 6353, C, CG4379, CdkA, Cos, Cos-1, Cos1, DC0, DCO, Dc0, Dco, Dcpk, Dmel\\CG4379, PKA, PKA CI, PKA Cl, PKA-C1, PKAC1, PKAc, PKAcA, Pka, Pka-C, PkaC1, cos1, dPKA, dcO, dco, l(2)01272, l(2)cos[[1]], l(2)s4402, pKA, pkA, pka, pka C1, pka-C1, pka-c1', description='Protein kinase, cAMP-dependent, catalytic subunit 1', other_designations='CG4379-PB|CG4379-PC|CG4379-PD|Costal-1|PKA catalytic subunit|Pka-C1-PB|Pka-C1-PC|Pka-C1-PD|cAMP-dependent protein kinase|cAMP-dependent protein kinase 1|catalytic subunit of PKA|catalytic subunit of cAMP-dependent protein kinase|costal|protein kinase A|protein kinase, cAMP-dependent, catalytic subunit 1', map_location='30C5-30C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9682315, end_position_on_the_genomic_accession=9699297, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34286, CurrentID=0, Status='live', Symbol='pelo', Aliases='Dmel_CG3959, CG3959, Dmel\\CG3959, ms(2)01559', description='pelota', other_designations='CG3959-PA|pelo-PA', map_location='30C5-30C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9700810, end_position_on_the_genomic_accession=9708883, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34287, CurrentID=0, Status='live', Symbol='CG4364', Aliases='Dmel_ Dmel\\CG4364', description='CG4364 gene product from transcript CG4364-RA', other_designations='CG4364-PA', map_location='30C5-30C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9708921, end_position_on_the_genomic_accession=9711271, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34288, CurrentID=0, Status='live', Symbol='CG13116', Aliases='Dmel_ Dmel\\CG13116', description='CG13116 gene product from transcript CG13116-RA', other_designations='CG13116-PA|CG13116-PB|uncharacterized protein', map_location='30C6-30C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9729940, end_position_on_the_genomic_accession=9730945, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34289, CurrentID=0, Status='live', Symbol='CG13117', Aliases='Dmel_ BcDNA:SD27505, Dmel\\CG13117', description='CG13117 gene product from transcript CG13117-RB', other_designations='CG13117-PA|CG13117-PB|uncharacterized protein', map_location='30C6-30C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9736645, end_position_on_the_genomic_accession=9737949, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34291, CurrentID=0, Status='live', Symbol='Cyp4e3', Aliases='Dmel_CG4105, 4e3, CG4105, Cyp4E3, Dmel\\CG4105, P-450, P450, cyp4e3', description='Cytochrome P450-4e3', other_designations='CG4105-PA|Cyp4e3-PA|cytochrome P450|cytochrome P450-4e3', map_location='30C7-30C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9747839, end_position_on_the_genomic_accession=9750071, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34292, CurrentID=0, Status='live', Symbol='zf30C', Aliases='Dmel_CG3998, C282, CG3998, Dmel\\CG3998, Zf30c, Zn30C, anon-EST:Liang-1.52, clone 1.52, l(2)k02506, zf30c', description='zinc finger protein 30C', other_designations='CG3998-PA|zf30C-PA', map_location='30C7-30C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9758342, end_position_on_the_genomic_accession=9762375, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34293, CurrentID=0, Status='live', Symbol='Taf11', Aliases='Dmel_CG4079, CG4079, Dmel\\CG4079, TAF, TAF1, TAF11, TAF30b, TAF30beta, TAF[II]30beta, TAF[[II]], TAF[[II]]30beta, TFIID, TFIID 28 beta, TFIID-p28beta, Taf30beta, Taf[[II]]30beta, d30beta, dTAFII30, dTAFII30 beta, dTAFII30beta, dTAF[[II]]30, dTAF[[II]]30beta', description='TBP-associated factor 11', other_designations='CG4079-PA|TBP-associated factor|TBP-associated factor 30kD subunit beta|Taf11-PA', map_location='30C7-30C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9762424, end_position_on_the_genomic_accession=9763262, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34294, CurrentID=0, Status='live', Symbol='und', Aliases='Dmel_CG4008, CG4008, DMAP2, Dmel\\CG4008, MAP2, p67', description='uninitiated', other_designations='CG4008-PA|CG4008-PB|CG4008-PC|p67|und-PA|und-PB|und-PC', map_location='30C7-30C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9763369, end_position_on_the_genomic_accession=9765338, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34295, CurrentID=0, Status='live', Symbol='ppk18', Aliases='Dmel_CG44152, CG13120, CG13121, CG44152, Dmel\\CG44152, Dmel_CG13120, Dmel_CG13121', description='pickpocket 18', other_designations='CG44152-PA|ppk18-PA', map_location='30C8-30C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9765357, end_position_on_the_genomic_accession=9770996, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34296, CurrentID=0, Status='live', Symbol='CG4017', Aliases='Dmel_ Dmel\\CG4017', description='CG4017 gene product from transcript CG4017-RA', other_designations='CG4017-PA', map_location='30C8-30C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9767058, end_position_on_the_genomic_accession=9768727, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34297, CurrentID=0, Status='live', Symbol='CG17633', Aliases='Dmel_ Dmel\\CG17633', description='CG17633 gene product from transcript CG17633-RA', other_designations='CG17633-PA|uncharacterized protein', map_location='30C8-30C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9769060, end_position_on_the_genomic_accession=9770409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34299, CurrentID=0, Status='live', Symbol='ppk11', Aliases='Dmel_CG34058, CG34058, CG4110, Dmel\\CG34058, PPK11', description='pickpocket 11', other_designations='CG34058-PA|pickpocket11|ppk11-PA', map_location='30C8-30C9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9773108, end_position_on_the_genomic_accession=9775030, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34300, CurrentID=0, Status='live', Symbol='IP3K1', Aliases='Dmel_CG4026, CG4026, D-IP3K1, Dmel\\CG4026, IP3KI, IP[[3]]K-1, UY530, anon-WO0118547.189, dmIP3Kalpha, dmIP[[3]]Kalpha', description='Inositol 1,4,5-triphosphate kinase 1', other_designations='CG4026-PA|CG4026-PB|IP3 kinase 1|IP3K1-PA|IP3K1-PB|inositol 1,4,5-triphosphate kinase 1|inositol 1,4,5-triphosphate kinase1', map_location='30C9-30D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9782384, end_position_on_the_genomic_accession=9788499, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34302, CurrentID=0, Status='live', Symbol='CG4036', Aliases='Dmel_ Dmel\\CG4036', description='CG4036 gene product from transcript CG4036-RB', other_designations='CG4036-PA|CG4036-PB', map_location='30D1-30D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9790882, end_position_on_the_genomic_accession=9791958, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34303, CurrentID=0, Status='live', Symbol='CG13123', Aliases='Dmel_ Dmel\\CG13123', description='CG13123 gene product from transcript CG13123-RB', other_designations='CG13123-PA|CG13123-PB|CG13123-PC|uncharacterized protein', map_location='30D1-30D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9791961, end_position_on_the_genomic_accession=9793107, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34304, CurrentID=0, Status='live', Symbol='nAChRalpha6', Aliases='Dmel_CG4128, BcDNA:GH01410, CG4128, CT13662, Dalpha6, Dmel\\CG4128, alpha6, das, nAChR, nAChR alpha 6, nAChRAlpha-30D, nAChRalpha, nAChRalpha-30D, nAcRalpha-30D, rsn', description='nicotinic Acetylcholine Receptor alpha6', other_designations='CG4128-PA|CG4128-PB|CG4128-PC|CG4128-PD|CG4128-PE|CG4128-PG|CG4128-PH|Dalpha6|alpha6 nicotinic acetylcholine receptor subunit|nAChRalpha6-PA|nAChRalpha6-PB|nAChRalpha6-PC|nAChRalpha6-PD|nAChRalpha6-PE|nAChRalpha6-PG|nAChRalpha6-PH|nicotinic acetylcholine receptor 30D|nicotinic acetylcholine receptor alpha 30D|nicotinic acetylcholine receptor alpha6|resistance to spinosyn|resistant to spinosyns', map_location='30D1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9793317, end_position_on_the_genomic_accession=9886250, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  34306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34306, CurrentID=0, Status='live', Symbol='CG4537', Aliases='Dmel_ Dmel\\CG4537', description='CG4537 gene product from transcript CG4537-RB', other_designations='CG4537-PB', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9890885, end_position_on_the_genomic_accession=9891336, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34307, CurrentID=0, Status='live', Symbol='mtDNA-helicase', Aliases='Dmel_CG5924, CG5924, Dmel\\CG5924', description='mitochondrial DNA helicase', other_designations='CG5924-PA|d-mtDNA helicase|mtDNA helicase|mtDNA-helicase-PA|twinkle', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9891255, end_position_on_the_genomic_accession=9893446, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34309, CurrentID=0, Status='live', Symbol='RpS2', Aliases='Dmel_CG5920, CG5920, Dm01335, Dmel\\CG5920, Rp S2, S2, anon-EST:Posey202, dS2, l(2)03848, l(2)k04603, sop', description='Ribosomal protein S2', other_designations='40S ribosomal protein S2|CG5920-PA|CG5920-PB|CG5920-PC|RpS2-PA|RpS2-PB|RpS2-PC|S2 ribosomal protein|ribosomal protein S2|ribosome RS2|string of pearls', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9896265, end_position_on_the_genomic_accession=9897552, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34310, CurrentID=0, Status='live', Symbol='CG13124', Aliases='Dmel_ Dmel\\CG13124', description='CG13124 gene product from transcript CG13124-RJ', other_designations='CG13124-PA|CG13124-PB|CG13124-PD|CG13124-PE|CG13124-PF|CG13124-PG|CG13124-PH|CG13124-PI|CG13124-PJ|uncharacterized protein', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9903279, end_position_on_the_genomic_accession=9908633, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34311, CurrentID=0, Status='live', Symbol='Etl1', Aliases='Dmel_CG5899, CG5899, Dmel\\CG5899', description='Etl1 homologue', other_designations='CG5899-PA|Etl1-PA', map_location='30E1-30E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9908830, end_position_on_the_genomic_accession=9912608, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34312, CurrentID=0, Status='live', Symbol='Srp54', Aliases='Dmel_CG4602, CG4602, Dmel\\CG4602, SRp54, anon-EST:Liang-2.16, clone 2.16, dSRp54, p75', description='CG4602 gene product from transcript CG4602-RA', other_designations='CG4602-PA|Srp54-PA', map_location='30E4-30E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9912911, end_position_on_the_genomic_accession=9915044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34313, CurrentID=0, Status='live', Symbol='yip2', Aliases='Dmel_CG4600, ACAA, CG4600, Dmel\\CG4600, Yip2', description='yippee interacting protein 2', other_designations='CG4600-PA|acetyl-CoA acyltransferase|yip2-PA', map_location='30E4-30E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9915253, end_position_on_the_genomic_accession=9917220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34314, CurrentID=0, Status='live', Symbol='CG5885', Aliases='Dmel_ AI945207, BEST:CK01296, CK01296, Dmel\\CG5885', description='CG5885 gene product from transcript CG5885-RB', other_designations='CG5885-PA|CG5885-PB', map_location='30E4-30E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9917221, end_position_on_the_genomic_accession=9918330, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34315, CurrentID=0, Status='live', Symbol='CG4598', Aliases='Dmel_ Dmel\\CG4598', description='CG4598 gene product from transcript CG4598-RB', other_designations='CG4598-PA|CG4598-PB', map_location='30E4-30E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9918618, end_position_on_the_genomic_accession=9920120, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34316, CurrentID=0, Status='live', Symbol='CG4594', Aliases='Dmel_ Dmel\\CG4594', description='CG4594 gene product from transcript CG4594-RB', other_designations='CG4594-PA|CG4594-PB', map_location='30E4-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9920160, end_position_on_the_genomic_accession=9921459, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34317, CurrentID=0, Status='live', Symbol='CG4592', Aliases='Dmel_ Dmel\\CG4592', description='CG4592 gene product from transcript CG4592-RB', other_designations='CG4592-PA|CG4592-PB', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9921633, end_position_on_the_genomic_accession=9922860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34318, CurrentID=0, Status='live', Symbol='TbCMF46', Aliases='Dmel_CG13125, CG13125, CT32363, Dmel\\CG13125, mel CG13125', description='CG13125 gene product from transcript CG13125-RA', other_designations='CG13125-PA|TbCMF46-PA', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9922737, end_position_on_the_genomic_accession=9924865, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34319, CurrentID=0, Status='live', Symbol='CG42366', Aliases='Dmel_ CG31711, CG31878, CG4588, Dmel\\CG42366, Dmel_CG31711, Dmel_CG31878, IME2, MAK', description='CG42366 gene product from transcript CG42366-RA', other_designations='CG42366-PA|CG42366-PB|uncharacterized protein', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9925161, end_position_on_the_genomic_accession=9930378, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34320, CurrentID=0, Status='live', Symbol='CG31712', Aliases='Dmel_ CG5879, Dmel\\CG31712', description='CG31712 gene product from transcript CG31712-RA', other_designations='CG31712-PA|CG31712-PB|uncharacterized protein', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9934767, end_position_on_the_genomic_accession=9938134, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34321, CurrentID=0, Status='live', Symbol='CG4619', Aliases='Dmel_ CG31879, Dmel\\CG4619', description='CG4619 gene product from transcript CG4619-RA', other_designations='CG4619-PA|CG4619-PB', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9938863, end_position_on_the_genomic_accession=9943058, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34322, CurrentID=0, Status='live', Symbol='CG5853', Aliases='Dmel_ Dmel\\CG5853', description='CG5853 gene product from transcript CG5853-RC', other_designations='CG5853-PA|CG5853-PB|CG5853-PC', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9940773, end_position_on_the_genomic_accession=9947648, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34323, CurrentID=0, Status='live', Symbol='CG13126', Aliases='Dmel_ Dmel\\CG13126, anon-WO0118547.208', description='CG13126 gene product from transcript CG13126-RB', other_designations='CG13126-PA|CG13126-PB|uncharacterized protein', map_location='30F2-30F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9952072, end_position_on_the_genomic_accession=9954181, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34324, CurrentID=0, Status='live', Symbol='CG4709', Aliases='Dmel_ Dmel\\CG4709', description='CG4709 gene product from transcript CG4709-RB', other_designations='CG4709-PA|CG4709-PB', map_location='30F2-30F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9954325, end_position_on_the_genomic_accession=9956176, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34325, CurrentID=0, Status='live', Symbol='CG4658', Aliases='Dmel_ Dmel\\CG4658', description='CG4658 gene product from transcript CG4658-RD', other_designations='CG4658-PA|CG4658-PB|CG4658-PC|CG4658-PD|CG4658-PE', map_location='30F2-30F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9957132, end_position_on_the_genomic_accession=9960168, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34326, CurrentID=0, Status='live', Symbol='CG5846', Aliases='Dmel_ Dmel\\CG5846', description='CG5846 gene product from transcript CG5846-RA', other_designations='CG5846-PA|CG5846-PB', map_location='30F2-30F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9956175, end_position_on_the_genomic_accession=9957237, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34327, CurrentID=0, Status='live', Symbol='CG5850', Aliases='Dmel_ Dmel\\CG5850', description='CG5850 gene product from transcript CG5850-RD', other_designations='CG5850-PA|CG5850-PC|CG5850-PD|CG5850-PE|CG5850-PF|CG5850-PG', map_location='30F2-30F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9959248, end_position_on_the_genomic_accession=9963949, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34328, CurrentID=0, Status='live', Symbol='Dref', Aliases='Dmel_CG5838, CG5838, DREF, DmDREF, Dmel\\CG5838, Dmel_DREF, dDREF, dref, p80, p80/DREF', description='DNA replication-related element factor', other_designations='CG5838-PA|CG5838-PB|DNA|DNA replication-related element binding factor|DNA replication-related element-binding factor|DNA-replication-related element factor|DRE (DNA replication-related element) binding factor|DRE-binding factor|Dref-PA|Dref-PB|replication-related element-binding factor', map_location='30F2-30F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9964147, end_position_on_the_genomic_accession=9967229, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34329, CurrentID=0, Status='live', Symbol='RpL13', Aliases='Dmel_CG4651, Bbc1, CG4651, Dmbbc1, Dmel\\CG4651, L13, Rp L13, Rpl13, anon-EST:fe1D1, bbc1, rpL13', description='Ribosomal protein L13', other_designations='CG4651-PA|CG4651-PB|CG4651-PC|RpL13-PA|RpL13-PB|RpL13-PC|anon-fast-evolving-1D1|r-protein L13|ribosomal protein L13', map_location='30F3-30F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9967313, end_position_on_the_genomic_accession=9968645, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34330, CurrentID=0, Status='live', Symbol='bib', Aliases='Dmel_CG4722, BIB_DROME, Bib, CG4722, Dmel\\CG4722, EP(2)2278, l(2)neo66, l(2)neo91, m-bib', description='big brain', other_designations='Big brain|CG4722-PA|CG4722-PB|bib-PA|bib-PB', map_location='2-34.7 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9984645, end_position_on_the_genomic_accession=9995538, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34331, CurrentID=0, Status='live', Symbol='SoYb', Aliases='Dmel_CG31755, CG31755, CG5831, Dmel\\CG31755', description='Sister of Yb', other_designations='CG31755-PB|CG31755-PC|SoYb-PB|SoYb-PC|sister of Yb', map_location='30F5-30F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9996706, end_position_on_the_genomic_accession=10002736, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34333, CurrentID=0, Status='live', Symbol='Trp1', Aliases='Dmel_CG4758, CG4758, Dmel\\CG4758, Dtrp1, Tp1, sec62', description='Translocation protein 1', other_designations='CG4758-PA|CG4758-PB|CG4758-PC|CG4758-PD|Trp1-PA|Trp1-PB|Trp1-PC|Trp1-PD|translocation protein 1', map_location='30F5-30F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10009381, end_position_on_the_genomic_accession=10013170, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34334, CurrentID=0, Status='live', Symbol='CG13131', Aliases='Dmel_ Dmel\\CG13131', description='CG13131 gene product from transcript CG13131-RA', other_designations='CG13131-PA|uncharacterized protein', map_location='30F5-30F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10013227, end_position_on_the_genomic_accession=10017463, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34335, CurrentID=0, Status='live', Symbol='CG13127', Aliases='Dmel_ Dmel\\CG13127', description='CG13127 gene product from transcript CG13127-RA', other_designations='CG13127-PA|uncharacterized protein', map_location='30F6-30F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10026676, end_position_on_the_genomic_accession=10029298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34336, CurrentID=0, Status='live', Symbol='obst-B', Aliases='Dmel_CG4778, BEST:CK00178, BcDNA:GH02976, CG4778, CK00178, Dmel\\CG4778', description='obstructor-B', other_designations='CG4778-PA|obst-B-PA', map_location='30F6-31A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10032690, end_position_on_the_genomic_accession=10038954, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34338, CurrentID=0, Status='live', Symbol='Pen', Aliases='Dmel_CG4799, 2.1, CG4799, DPend, Dalpha2, Dimp-alpha2, Dmel\\CG4799, IMPalpha2, Imp-alpha2, Kap alpha2, Kap-alpha2, Kpna2, OHO31, Oho31, Rch1, alpha2, alpha2A-Kap, alphaKap2, anon-WO0140519.258, bs29g06.y1, imp alpha2, imp-alpha2, impalpha2, importin alpha2/pendulin, l(2)144/1, l(2)k06324, l(2)k14401, oho-31, oho31, pen, pen-2', description='Pendulin', other_designations='CG4799-PA|Importin-a2|Importin-alpha|Importin-alpha2|Pen-PA|alpha-importin|i|importin alpha|importin alpha2|lethal (2) k06324|pendulin', map_location='31A1-31A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10056900, end_position_on_the_genomic_accession=10060095, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34339, CurrentID=0, Status='live', Symbol='Spn31A', Aliases='Dmel_CG4804, CG4804, Dmel\\CG4804', description='Serpin 31A', other_designations='CG4804-PA|Spn31A-PA|serpin 31A', map_location='31A2-31A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10060816, end_position_on_the_genomic_accession=10063925, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34341, CurrentID=0, Status='live', Symbol='CG44153', Aliases='Dmel_ BP1029, CG13132, CG13134, CG13136, CG31714, CG34048, CG42843, CG5833, Dmel\\CG44153, Dmel_CG13134, Dmel_CG13136, Dmel_CG31714, Dmel_CG34048, Dmel_CG42843', description='CG44153 gene product from transcript CG44153-RB', other_designations='CG44153-PA|CG44153-PB|uncharacterized protein', map_location='31A3-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10064121, end_position_on_the_genomic_accession=10180256, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  34342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34342, CurrentID=0, Status='live', Symbol='CG13133', Aliases='Dmel_ Dmel\\CG13133', description='CG13133 gene product from transcript CG13133-RA', other_designations='CG13133-PA|uncharacterized protein', map_location='31A2-31A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10091208, end_position_on_the_genomic_accession=10092389, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19835613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835613, CurrentID=0, Status='live', Symbol='CG44286', Aliases='Dmel_ Dmel\\CG44286', description='CG44286 gene product from transcript CG44286-RD', other_designations='CG44286-PA|CG44286-PB|CG44286-PC|CG44286-PD|uncharacterized protein', map_location='45A11-45A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9138486, end_position_on_the_genomic_accession=9141981, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34345, CurrentID=0, Status='live', Symbol='CG13135', Aliases='Dmel_ Dmel\\CG13135', description='CG13135 gene product from transcript CG13135-RC', other_designations='CG13135-PA|CG13135-PC|uncharacterized protein', map_location='31A2-31A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10122129, end_position_on_the_genomic_accession=10122771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34348, CurrentID=0, Status='live', Symbol='CG4839', Aliases='Dmel_ Dmel\\CG4839', description='CG4839 gene product from transcript CG4839-RA', other_designations='CG4839-PA|CG4839-PB|PKG-like', map_location='31A3-31A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10144776, end_position_on_the_genomic_accession=10156407, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34349, CurrentID=0, Status='live', Symbol='CG13137', Aliases='Dmel_ Dmel\\CG13137', description='CG13137 gene product from transcript CG13137-RB', other_designations='CG13137-PB|uncharacterized protein', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10172702, end_position_on_the_genomic_accession=10174090, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34350, CurrentID=0, Status='live', Symbol='Sur', Aliases='Dmel_CG5772, BEST:CK00325, CG5772, CK00325, CYO, Dme_CG5772, Dmel\\CG5772, Dsur, SUR, cyo, dSUR, dSur, dsur, sur', description='Sulfonylurea receptor', other_designations='CG5772-PB|Sur-PB|sulfonyl urea receptor|sulfonylurea receptor|sulphonylurea receptor', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10182557, end_position_on_the_genomic_accession=10199117, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  34351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34351, CurrentID=0, Status='live', Symbol='Ripalpha', Aliases='Dmel_CG18145, CG18145, Dmel\\CG18145, Tes83, dRIPalpha', description='RPA-interacting protein alpha', other_designations='CG18145-PA|Ripalpha-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10199401, end_position_on_the_genomic_accession=10200974, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34352, CurrentID=0, Status='live', Symbol='RpL7', Aliases='Dmel_CG4897, BEST:GH04411, CG4897, Dmel\\CG4897, L7, RPL7, Rp L7, rpL7, rpl7', description='Ribosomal protein L7', other_designations='CG4897-PA|CG4897-PB|RpL7-PA|RpL7-PB|ribosomal protein L7', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10201004, end_position_on_the_genomic_accession=10202321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34353, CurrentID=0, Status='live', Symbol='CG34159', Aliases='Dmel_ CG5739, Dmel\\CG34159', description='CG34159 gene product from transcript CG34159-RA', other_designations='CG34159-PA|uncharacterized protein', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10202520, end_position_on_the_genomic_accession=10204317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34354, CurrentID=0, Status='live', Symbol='CG5734', Aliases='Dmel_ DmSNX17, Dmel\\CG5734', description='CG5734 gene product from transcript CG5734-RA', other_designations='CG5734-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10204757, end_position_on_the_genomic_accession=10206961, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34355, CurrentID=0, Status='live', Symbol='CG5731', Aliases='Dmel_ Dmel\\CG5731', description='CG5731 gene product from transcript CG5731-RA', other_designations='CG5731-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10207085, end_position_on_the_genomic_accession=10209744, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34356, CurrentID=0, Status='live', Symbol='CG5727', Aliases='Dmel_ BcDNA:LD02427, Dmel\\CG5727', description='CG5727 gene product from transcript CG5727-RA', other_designations='CG5727-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10209853, end_position_on_the_genomic_accession=10210681, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34357, CurrentID=0, Status='live', Symbol='CG4901', Aliases='Dmel_ Dmel\\CG4901, cg4901', description='CG4901 gene product from transcript CG4901-RA', other_designations='CG4901-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10211170, end_position_on_the_genomic_accession=10213721, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34358, CurrentID=0, Status='live', Symbol='Npc1a', Aliases='Dmel_CG5722, CG5722, DmNPC, Dmel\\CG5722, NPC1, NPC1a, Npc1, Npc1_Dm, dNPC1, dNPC1a, dmNPC1, dncp1a, dnpc1a, npc1, npc1a', description='Niemann-Pick type C-1a', other_designations='CG5722-PA|CG5722-PB|CG5722-PC|CG5722-PD|CG5722-PE|CG5722-PF|CG5722-PG|Niemann-Pick type C-1|Npc1a-PA|Npc1a-PB|Npc1a-PC|Npc1a-PD|Npc1a-PE|Npc1a-PF|Npc1a-PG', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10213812, end_position_on_the_genomic_accession=10221216, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  34359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34359, CurrentID=0, Status='live', Symbol='Prosalpha6', Aliases='Dmel_CG4904, 143314_at, Alpha-6, CG4904, Dmel\\CG4904, PROS-35, PROS-Dm35, PSMA1, Pros-35, Pros-Dm35, Pros35, alpha-6, alpha6, alpha_dm, anon-SAGE:Wang-116, pros 35, pros35', description='Proteasome alpha6 subunit', other_designations='20S alpha6 subunit|20s proteasome|CG4904-PA|CG4904-PB|Prosalpha6-PA|Prosalpha6-PB|alpha-subunit of 20S proteasome|alpha6|anon-SAGE:Wang-116|proteasome 35|proteasome 35 kD subunit|proteasome 35-kDa subunit|proteasome 35kD subunit|proteasome 35kDa subunit|proteasome alpha6 subunit', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10221351, end_position_on_the_genomic_accession=10222546, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34360, CurrentID=0, Status='live', Symbol='CG4908', Aliases='Dmel_ Dmel\\CG4908', description='CG4908 gene product from transcript CG4908-RB', other_designations='CG4908-PA|CG4908-PB', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10222657, end_position_on_the_genomic_accession=10224447, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34361, CurrentID=0, Status='live', Symbol='CG5708', Aliases='Dmel_ Dmel\\CG5708', description='CG5708 gene product from transcript CG5708-RC', other_designations='CG5708-PA|CG5708-PB|CG5708-PC', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10224457, end_position_on_the_genomic_accession=10226808, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34362, CurrentID=0, Status='live', Symbol='CG5694', Aliases='Dmel_ Dmel\\CG5694', description='CG5694 gene product from transcript CG5694-RA', other_designations='CG5694-PA|CG5694-PB', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10227760, end_position_on_the_genomic_accession=10231692, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34363, CurrentID=0, Status='live', Symbol='eEF1delta', Aliases='Dmel_CG4912, CG4912, Dmel\\CG4912, eEF1d, eEf1delta', description='CG4912 gene product from transcript CG4912-RB', other_designations='CG4912-PA|CG4912-PB|eEF1delta-PA|eEF1delta-PB|eukaryotic translation elongation factor 1delta', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10232507, end_position_on_the_genomic_accession=10234450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34364, CurrentID=0, Status='live', Symbol='me31B', Aliases='Dmel_CG4916, CG4916, DDX6, Dhh1p/Me31b, DmRH3, Dmel\\CG4916, ME31B, Mat31B, Me31B, Me31b, l(2)k06607', description='maternal expression at 31B', other_designations='CG4916-PA|CG4916-PB|me31B-PA|me31B-PB', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10239341, end_position_on_the_genomic_accession=10242172, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34367, CurrentID=0, Status='live', Symbol='Ror', Aliases='Dmel_CG4926, CG4926, D DmHD-2, Dmel\\CG4926, Dror, HD-2, ROR, ROR1', description='CG4926 gene product from transcript CG4926-RA', other_designations='CG4926-PA|Ror-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10251843, end_position_on_the_genomic_accession=10254686, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34368, CurrentID=0, Status='live', Symbol='CG5676', Aliases='Dmel_ BEST:LD31953, Dmel\\CG5676, LD31953', description='CG5676 gene product from transcript CG5676-RA', other_designations='CG5676-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10254610, end_position_on_the_genomic_accession=10255891, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740128, CurrentID=0, Status='live', Symbol='CG34199', Aliases='Dmel_ Dmel\\CG34199', description='CG34199 gene product from transcript CG34199-RA', other_designations='CG34199-PA|uncharacterized protein', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20287596, end_position_on_the_genomic_accession=20288346, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34370, CurrentID=0, Status='live', Symbol='Rsf1', Aliases='Dmel_CG5655, CG5655, Dmel\\CG5655, ROX21, RSF1, Rox21, rox21', description='Repressor splicing factor 1', other_designations='CG5655-PA|CG5655-PB|Rsf1-PA|Rsf1-PB|repressor splicing factor 1', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10261795, end_position_on_the_genomic_accession=10263554, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34371, CurrentID=0, Status='live', Symbol='REPTOR-BP', Aliases='Dmel_CG18619, BcDNA:RH73910, CG18619, DM20, Dmel\\CG18619', description='REPTOR-binding partner', other_designations='CG18619-PA|CG18619-PB|CG18619-PC|CG18619-PD|REPTOR-BP-PA|REPTOR-BP-PB|REPTOR-BP-PC|REPTOR-BP-PD', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10264125, end_position_on_the_genomic_accession=10265229, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34372, CurrentID=0, Status='live', Symbol='Mob3', Aliases='Dmel_CG4946, CG4946, D Dmel\\CG4946, Dmob3, dmob3', description='CG4946 gene product from transcript CG4946-RA', other_designations='CG4946-PA|Mob3-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10265336, end_position_on_the_genomic_accession=10266570, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34373, CurrentID=0, Status='live', Symbol='CG4953', Aliases='Dmel_ Dmel\\CG4953', description='CG4953 gene product from transcript CG4953-RA', other_designations='CG4953-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10266703, end_position_on_the_genomic_accession=10268748, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34374, CurrentID=0, Status='live', Symbol='CG4957', Aliases='Dmel_ Dmel\\CG4957', description='CG4957 gene product from transcript CG4957-RA', other_designations='CG4957-PA', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10268872, end_position_on_the_genomic_accession=10269904, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34376, CurrentID=0, Status='live', Symbol='trk', Aliases='Dmel_CG5619, CG5619, Dmel\\CG5619, fs(2)trunk', description='trunk', other_designations='CG5619-PA|trk-PA', map_location='2-36 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10271443, end_position_on_the_genomic_accession=10272324, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34377, CurrentID=0, Status='live', Symbol='Utx', Aliases='Dmel_CG5640, CG5640, Dmel\\CG5640, UTX, anon-31BCa, dUTX, dUtx, utx', description='Utx histone demethylase', other_designations='CG5640-PA|CG5640-PB|CG5640-PC|CG5640-PD|CG5640-PE|Utx-PA|Utx-PB|Utx-PC|Utx-PD|Utx-PE|anon-31BCa', map_location='31C7-31C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10272325, end_position_on_the_genomic_accession=10278811, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  34378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34378, CurrentID=0, Status='live', Symbol='CG5604', Aliases='Dmel_ Dmel\\CG5604', description='CG5604 gene product from transcript CG5604-RA', other_designations='CG5604-PA', map_location='31C4-31C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10281943, end_position_on_the_genomic_accession=10292365, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34379, CurrentID=0, Status='live', Symbol='Hand', Aliases='Dmel_CG18144, CG18144, Dhand, Dmel\\CG18144, HAND, bHLHa28, dHAND, d hand, hand-C', description='CG18144 gene product from transcript CG18144-RA', other_designations='CG18144-PA|Hand-PA|hand', map_location='31C6-31C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10292679, end_position_on_the_genomic_accession=10295385, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34380, CurrentID=0, Status='live', Symbol='CYLD', Aliases='Dmel_CG5603, CG5603, Cyld, Dmel\\CG5603, Q8IPC5, cyld, dCYLD', description='cylindromatosis ortholog (H. sapiens)', other_designations='CG5603-PB|CG5603-PD|CG5603-PE|CYLD-PB|CYLD-PD|CYLD-PE|cylindromatosis|cylindromatosis ortholog', map_location='31C7-31D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10295859, end_position_on_the_genomic_accession=10299847, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34381, CurrentID=0, Status='live', Symbol='CG13138', Aliases='Dmel_ Dmel\\CG13138', description='CG13138 gene product from transcript CG13138-RC', other_designations='CG13138-PB|CG13138-PC|uncharacterized protein', map_location='31D1-31D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10299977, end_position_on_the_genomic_accession=10302018, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34383, CurrentID=0, Status='live', Symbol='CG5390', Aliases='Dmel_ BEST:GH06341, Dmel\\CG5390, SPH, SPH35, c-SPH35, gh06341', description='CG5390 gene product from transcript CG5390-RA', other_designations='CG5390-PA|CG5390-PB|serine protease homologue protein', map_location='31D1-31D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10304069, end_position_on_the_genomic_accession=10306594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34384, CurrentID=0, Status='live', Symbol='CG4968', Aliases='Dmel_ Dmel\\CG4968, Q9VL00', description='CG4968 gene product from transcript CG4968-RA', other_designations='CG4968-PA', map_location='31D1-31D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10306922, end_position_on_the_genomic_accession=10308045, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34385, CurrentID=0, Status='live', Symbol='Cdk5alpha', Aliases='Dmel_CG5387, CG5387, Cdk5a, D-p35, Dmel\\CG5387, Dp35, P35, dCdk5alpha, p25, p35, p35-21C, p35-31C', description='Cdk5 activator-like protein', other_designations='CG5387-PA|CG5387-PB|Cdk5alpha-PA|Cdk5alpha-PB', map_location='31D1-31D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10308131, end_position_on_the_genomic_accession=10310956, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34386, CurrentID=0, Status='live', Symbol='l(2)SH0834', Aliases='Dmel_CG5385, CG5385, Dmel\\CG5385, l(2)SH2 0834', description='lethal (2) SH0834', other_designations='CG5385-PA|l(2)SH0834-PA', map_location='31D5-31D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10311050, end_position_on_the_genomic_accession=10312099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34387, CurrentID=0, Status='live', Symbol='Usp14', Aliases='Dmel_CG5384, CG5384, DmUsp14, Dmel\\CG5384, Q9VKZ8, Ubp6p', description='Ubiquitin specific protease 14', other_designations='CG5384-PA|CG5384-PB|Usp14-PA|Usp14-PB|ubiquitin specific protease 14', map_location='31D4-31D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10312265, end_position_on_the_genomic_accession=10314396, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34388, CurrentID=0, Status='live', Symbol='CG4972', Aliases='Dmel_ Dmel\\CG4972, Q9VKZ7_DROME', description='CG4972 gene product from transcript CG4972-RA', other_designations='CG4972-PA', map_location='31D5-31D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10314792, end_position_on_the_genomic_accession=10317569, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34389, CurrentID=0, Status='live', Symbol='CG5381', Aliases='Dmel_ Dmel\\CG5381', description='CG5381 gene product from transcript CG5381-RC', other_designations='CG5381-PA|CG5381-PB|CG5381-PC|CG5381-PD', map_location='31D6-31D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10317497, end_position_on_the_genomic_accession=10323102, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34390, CurrentID=0, Status='live', Symbol='CG4995', Aliases='Dmel_ CG18626, Dmel\\CG4995', description='CG4995 gene product from transcript CG4995-RB', other_designations='CG4995-PA|CG4995-PB|CG4995-PC|CG4995-PD', map_location='31D7-31D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10322476, end_position_on_the_genomic_accession=10324990, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34392, CurrentID=0, Status='live', Symbol='RnrL', Aliases='Dmel_CG5371, CG5371, DmRNR1, Dmel\\CG5371, RNR1, RNRL, RnrM1, RntL, l(2)k06709', description='Ribonucleoside diphosphate reductase large subunit', other_designations='CG5371-PA|RnrL-PA|ribonucleoside diphosphate reductase M1 subunit|ribonucleoside diphosphate reductase large subunit|ribonucleoside-diphosphate reductase large subunit', map_location='31D8-31D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10325026, end_position_on_the_genomic_accession=10327943, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34393, CurrentID=0, Status='live', Symbol='CG5375', Aliases='Dmel_ Dmel\\CG5375', description='CG5375 gene product from transcript CG5375-RB', other_designations='CG5375-PB', map_location='31D8-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10328200, end_position_on_the_genomic_accession=10331116, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34394, CurrentID=0, Status='live', Symbol='CG5037', Aliases='Dmel_ COXX, Dmel\\CG5037', description='CG5037 gene product from transcript CG5037-RA', other_designations='CG5037-PA|protoheme IX farnesyltransferase', map_location='31D9-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10331471, end_position_on_the_genomic_accession=10333129, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34395, CurrentID=0, Status='live', Symbol='GATAd', Aliases='Dmel_CG5034, CG5034, Dmel\\CG5034, dGATA-D, d', description='CG5034 gene product from transcript CG5034-RB', other_designations='CG5034-PA|CG5034-PB|CG5034-PC|GATAd-PA|GATAd-PB|GATAd-PC', map_location='31D9-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10333734, end_position_on_the_genomic_accession=10339185, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34396, CurrentID=0, Status='live', Symbol='SamDC', Aliases='Dmel_CG5029, CG5029, Dmel\\CG5029, l(2)31Dc, l(2)31Dd, l(2)31De', description='S-adenosylmethionine decarboxylase', other_designations='CG5029-PA|S-adenosyl decarboxylase|SamDC-PA', map_location='31D9-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10339457, end_position_on_the_genomic_accession=10341503, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34397, CurrentID=0, Status='live', Symbol='Sps2', Aliases='Dmel_CG5025, CG5025, Dmel\\CG5025, Dsps2, SPS2, dSPS2, dsps2, sps2', description='Selenophosphate synthetase 2', other_designations='CG5025-PA|CG5025-PC|Sps2-PA|Sps2-PC|selenophosphate synthetase 2', map_location='31D9-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10342594, end_position_on_the_genomic_accession=10344223, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34398, CurrentID=0, Status='live', Symbol='CG5022', Aliases='Dmel_ Dmel\\CG5022', description='CG5022 gene product from transcript CG5022-RA', other_designations='CG5022-PA', map_location='31D9-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10344442, end_position_on_the_genomic_accession=10347706, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34401, CurrentID=0, Status='live', Symbol='CG5367', Aliases='Dmel_ Dmel\\CG5367', description='CG5367 gene product from transcript CG5367-RA', other_designations='CG5367-PA', map_location='31D10-31D10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10354743, end_position_on_the_genomic_accession=10356597, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34402, CurrentID=0, Status='live', Symbol='CG5045', Aliases='Dmel_ Dmel\\CG5045', description='CG5045 gene product from transcript CG5045-RB', other_designations='CG5045-PA|CG5045-PB', map_location='31D10-31D10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10356795, end_position_on_the_genomic_accession=10358078, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34403, CurrentID=0, Status='live', Symbol='Cand1', Aliases='Dmel_CG5366, BcDNA:GH07774, CAND1, CG5366, Dmel\\CG5366, dmCAND1', description='Cullin-associated and neddylation-dissociated 1', other_designations='CG5366-PA|CG5366-PB|Cand1-PA|Cand1-PB', map_location='31D10-31D10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10358134, end_position_on_the_genomic_accession=10362707, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34404, CurrentID=0, Status='live', Symbol='pim', Aliases='Dmel_CG5052, CG5052, Dmel\\CG5052, PIM, Pim, frag2, l(2)31Df', description='pimples', other_designations='CG5052-PA|CG5052-PB|PIMPLES|fragmented gonads 2|pim-PA|pim-PB|securin', map_location='2-30 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10362989, end_position_on_the_genomic_accession=10364092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34405, CurrentID=0, Status='live', Symbol='lft', Aliases='Dmel_CG13139, CG13139, Dmel\\CG13139', description='lowfat', other_designations='CG13139-PA|CG13139-PB|lft-PA|lft-PB', map_location='31D10-31D10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10364471, end_position_on_the_genomic_accession=10366410, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34406, CurrentID=0, Status='live', Symbol='CG5056', Aliases='Dmel_ Dmel\\CG5056', description='CG5056 gene product from transcript CG5056-RA', other_designations='CG5056-PA', map_location='31D10-31D10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10366692, end_position_on_the_genomic_accession=10367946, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34408, CurrentID=0, Status='live', Symbol='dpr19', Aliases='Dmel_CG13140, BcDNA:LP09705, CG13140, Dmel\\CG13140, Dpr-19, Dpr19', description='defective proboscis extension response 19', other_designations='CG13140-PA|CG13140-PB|CG13140-PC|dpr19-PA|dpr19-PB|dpr19-PC', map_location='31D11-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10376168, end_position_on_the_genomic_accession=10379905, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34409, CurrentID=0, Status='live', Symbol='gny', Aliases='Dmel_CG5091, CG5091, Dalg6, Dmel\\CG5091, Gny[, alg6, dAlg6', description='garnysstan', other_designations='CG5091-PA|CG5091-PB|gny-PA|gny-PB', map_location='31D11-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10380103, end_position_on_the_genomic_accession=10382864, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34410, CurrentID=0, Status='live', Symbol='CG5096', Aliases='Dmel_ CT16355, Dmel\\CG5096', description='CG5096 gene product from transcript CG5096-RA', other_designations='CG5096-PA', map_location='31D11-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10382263, end_position_on_the_genomic_accession=10384723, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34411, CurrentID=0, Status='live', Symbol='Cdk1', Aliases='Dmel_CG5363, 5363, CDC2, CDCDm, CDK1, CDK1/CDC2, CG5363, Cdc2, Cdk-1, Dcdc2, Dm cdc2, DmCdc2, DmCdk1, Dmcdc2, Dmel\\CG5363, cdc, cdc2, cdc2Dm, cdk1, dCdk1, group 4, l(2)31Eh', description='Cyclin-dependent kinase 1', other_designations='CG5363-PA|Cdk1-PA|cdc2|cdc2 kinase|cell division control|cyclin dependent kinase|cyclin-dependent kinase|protein 2', map_location='31D11-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10384739, end_position_on_the_genomic_accession=10386262, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34412, CurrentID=0, Status='live', Symbol='mRpS7', Aliases='Dmel_CG5108, CG5108, Dmel\\CG5108, MRP-S7, RpS7, mRPS7', description='mitochondrial ribosomal protein S7', other_designations='CG5108-PA|CG5108-PB|mRpS7-PA|mRpS7-PB', map_location='31D11-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10386662, end_position_on_the_genomic_accession=10387652, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34413, CurrentID=0, Status='live', Symbol='da', Aliases='Dmel_CG5102, 5102, CG5102, DA, Da, Dmel\\CG5102, bHLHb22, l(2)31Ea, l(2)k08611', description='daughterless', other_designations='CG5102-PA|CG5102-PB|CG5102-PC|da-PA|da-PB|da-PC|daughter-less|lethal (2) k08611', map_location='2-41.3 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10388105, end_position_on_the_genomic_accession=10393228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34414, CurrentID=0, Status='live', Symbol='Mdh1', Aliases='Dmel_CG5362, CG5362, CT17038, Dmel\\CG5362, MDH, MDH-1, Mdh, Mdh-1, Mdh2, MdhD, cMdh, sMdh', description='Malate dehydrogenase 1', other_designations='CG5362-PA|CG5362-PB|Mdh1-PA|Mdh1-PB|cytoplasmic Mdh|cytosolic malate dehydrogenase|malate dehydrogenase|malate dehydrogenase 1|malic DH', map_location='2-37.2 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10393774, end_position_on_the_genomic_accession=10396332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34416, CurrentID=0, Status='live', Symbol='Cnot4', Aliases='Dmel_CG31716, CG31716, CG5244, CG5251, Cont4, Dmel\\CG31716, NOT4, Not4, Q8IPB9_DROME, cg5251, not4', description='Cnot 4 homologue', other_designations='CG31716-PA|CG31716-PB|CG31716-PC|CG31716-PD|CG31716-PE|CG31716-PG|CG31716-PH|CG31716-PI|CG31716-PJ|CG31716-PK|CG31716-PL|Cnot4-PA|Cnot4-PB|Cnot4-PC|Cnot4-PD|Cnot4-PE|Cnot4-PG|Cnot4-PH|Cnot4-PI|Cnot4-PJ|Cnot4-PK|Cnot4-PL|cnot 4 homologue', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10396716, end_position_on_the_genomic_accession=10403206, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  34419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34419, CurrentID=0, Status='live', Symbol='Ip259', Aliases='Dmel_CG5277, CG5277, Dmel\\CG5277, IP259', description='Intronic Protein 259', other_designations='CG5277-PA|CG5277-PB|Ip259-PA|Ip259-PB|intronic protein 259', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10408352, end_position_on_the_genomic_accession=10409610, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34420, CurrentID=0, Status='live', Symbol='RpS27A', Aliases='Dmel_CG5271, CG5271, DUb80, Dmel\\CG5271, Dub80, UB3-D, Ubi-f80, Ubi-m, Ubiq, l(2)04820, l(2)31Ei, mfs(2)31, mfs(2)48, mfs48', description='Ribosomal protein S27A', other_designations='CG5271-PA|RpS27A-PA|Ubiquitin-monomeric|male female sterile 31|ribosomal protein S27A|ubiquitin|ubiquitin fusion 80', map_location='2-41.35 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10408352, end_position_on_the_genomic_accession=10410456, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34421, CurrentID=0, Status='live', Symbol='CG5355', Aliases='Dmel_ Dmel\\CG5355', description='CG5355 gene product from transcript CG5355-RA', other_designations='CG5355-PA', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10410537, end_position_on_the_genomic_accession=10413463, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34422, CurrentID=0, Status='live', Symbol='Klp31E', Aliases='Dmel_CG5300, CG5300, DmCG5300, DmKlp31D, DmKlp31E, Dmel\\CG5300, KIF 21A, KIF21A/B, KLP31E', description='Kinesin-like protein at 31E', other_designations='CG5300-PA|CG5300-PB|CG5300-PC|CG5300-PE|Klp31E-PA|Klp31E-PB|Klp31E-PC|Klp31E-PE', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10414017, end_position_on_the_genomic_accession=10418991, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  34423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34423, CurrentID=0, Status='live', Symbol='RfC3', Aliases='Dmel_CG5313, CG5313, DRFC, DmRFC3, Dmel\\CG5313, Rfc3', description='Replication factor C subunit 3', other_designations='CG5313-PA|RfC3-PA|drosophila replication factor C|replication factor C subunit 3', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10419190, end_position_on_the_genomic_accession=10420612, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34425, CurrentID=0, Status='live', Symbol='CG5168', Aliases='Dmel_ Dmel\\CG5168', description='CG5168 gene product from transcript CG5168-RB', other_designations='CG5168-PA|CG5168-PB', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10423291, end_position_on_the_genomic_accession=10425482, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34426, CurrentID=0, Status='live', Symbol='SmB', Aliases='Dmel_CG5352, B, CG5352, Dmel\\CG5352, SYM10, Sm B1, dSmB, l(2)SH0509, l(2)SH2 0509, snRNP B, SmB', description='Small ribonucleoprotein particle protein SmB', other_designations='CG5352-PB|SmB-PB|lethal (2) SH0509|polypeptide B|small ribonucleoprotein particle protein B|small ribonucleoprotein particle protein SmB', map_location='31E1-31E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10425483, end_position_on_the_genomic_accession=10426627, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34427, CurrentID=0, Status='live', Symbol='KdelR', Aliases='Dmel_CG5183, BEST:CK00230, CG5183, CK00230, Dmel\\CG5183, ERD2, Erd2, dmErd2, l(2)31Em, l(2)k00311', description='KDEL receptor', other_designations='CG5183-PA|CG5183-PB|CG5183-PC|ER lumen protein retaining receptor 2|KDEL-receptor|KdelR-PA|KdelR-PB|KdelR-PC|lethal (2) 31Em', map_location='31E-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10426887, end_position_on_the_genomic_accession=10428632, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34428, CurrentID=0, Status='live', Symbol='CG5188', Aliases='Dmel_ Dmel\\CG5188, MAP1B', description='CG5188 gene product from transcript CG5188-RA', other_designations='CG5188-PA', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10428829, end_position_on_the_genomic_accession=10429988, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34429, CurrentID=0, Status='live', Symbol='Bug22', Aliases='Dmel_CG5343, CG5343, Dmel\\CG5343', description='Basal body up regulated gene 22 ortholog', other_designations='Bug22-PA|CG5343-PA|basal body up regulated gene 22 ortholog', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10429920, end_position_on_the_genomic_accession=10430918, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34430, CurrentID=0, Status='live', Symbol='TfIIB', Aliases='Dmel_CG5193, CG5193, Dmel\\CG5193, TFIIB, Tfiib, TfllB, dTFIIB, dTfIIB, dmTFIIB', description='Transcription factor IIB', other_designations='CG5193-PA|CG5193-PB|TfIIB-PA|TfIIB-PB|transcription factor IIB', map_location='31E2-31E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10431338, end_position_on_the_genomic_accession=10433116, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34431, CurrentID=0, Status='live', Symbol='CG5337', Aliases='Dmel_ Dmel\\CG5337', description='CG5337 gene product from transcript CG5337-RA', other_designations='CG5337-PA', map_location='31E2-31E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10432687, end_position_on_the_genomic_accession=10435656, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34432, CurrentID=0, Status='live', Symbol='holn1', Aliases='Dmel_CG5198, CG5198, Dmel\\CG5198', description='hole-in-one', other_designations='CG5198-PA|holn1-PA', map_location='31E2-31E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10435995, end_position_on_the_genomic_accession=10437316, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34433, CurrentID=0, Status='live', Symbol='CHIP', Aliases='Dmel_CG5203, CG5203, Chip, Dmel\\CG5203, dCHIP', description='CHIP ortholog', other_designations='CG5203-PA|CHIP-PA|NM 058093.3', map_location='31E2-31E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10437548, end_position_on_the_genomic_accession=10439240, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34434, CurrentID=0, Status='live', Symbol='Nse4', Aliases='Dmel_CG13142, CG13142, Dmel\\CG13142', description='Non-SMC element 4', other_designations='CG13142-PB|CG13142-PC|Nse4-PB|Nse4-PC', map_location='31E2-31E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10439440, end_position_on_the_genomic_accession=10441047, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34435, CurrentID=0, Status='live', Symbol='loh', Aliases='Dmel_CG6232, CG6232, CT19510, Dmel\\CG6232, Loh', description='lonely heart', other_designations='CG6232-PA|CG6232-PC|CG6232-PD|loh-PA|loh-PC|loh-PD', map_location='31E3-31E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10442534, end_position_on_the_genomic_accession=10450080, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  34436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34436, CurrentID=0, Status='live', Symbol='LManI', Aliases='Dmel_CG5322, CG5322, Dmel\\CG5322', description='Lysosomal alpha-mannosidase I', other_designations='CG5322-PA|CG5322-PB|LManI-PA|LManI-PB|lysosomal alpha-mannosidase I', map_location='31E4-31E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10452677, end_position_on_the_genomic_accession=10457241, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34437, CurrentID=0, Status='live', Symbol='LManII', Aliases='Dmel_CG6206, BcDNA:GH02419, CG6206, Dmel\\CG6206, LM, LM408, dLM, dLM408', description='Lysosomal alpha-mannosidase II', other_designations='CG6206-PA|CG6206-PB|LManII-PA|LManII-PB|class II lysosomal alpha-mannosidase|lysosomal alpha-mannosidase|lysosomal alpha-mannosidase II|lysosomal mannosidase', map_location='31E5-31E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10457658, end_position_on_the_genomic_accession=10462472, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  34438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34438, CurrentID=0, Status='live', Symbol='RluA-1', Aliases='Dmel_CG31719, CG13143, CG31719, CG7475, Dmel\\CG31719', description='CG31719 gene product from transcript CG31719-RD', other_designations='CG31719-PA|CG31719-PB|CG31719-PC|CG31719-PD|RluA-1-PA|RluA-1-PB|RluA-1-PC|RluA-1-PD|multiple', map_location='31E6-31F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10463340, end_position_on_the_genomic_accession=10474349, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  34440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34440, CurrentID=0, Status='live', Symbol='RluA-2', Aliases='Dmel_CG6187, CG6187, Dmel\\CG6187, yumo', description='CG6187 gene product from transcript CG6187-RA', other_designations='CG6187-PA|CG6187-PC|CG6187-PD|CG6187-PE|RluA-2-PA|RluA-2-PC|RluA-2-PD|RluA-2-PE', map_location='31F1-31F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10477073, end_position_on_the_genomic_accession=10482975, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  34441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34441, CurrentID=0, Status='live', Symbol='Grip75', Aliases='Dmel_CG6176, 75p, CG6176, D Dgrip75, Dgrip75/d75p, Dgrip75s, Dmel\\CG6176, d75p', description='CG6176 gene product from transcript CG6176-RA', other_designations='CG6176-PA|DGrip75|Grip75-PA', map_location='31F1-31F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10483295, end_position_on_the_genomic_accession=10485857, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34442, CurrentID=0, Status='live', Symbol='CG7456', Aliases='Dmel_ Dmel\\CG7456', description='CG7456 gene product from transcript CG7456-RA', other_designations='CG7456-PA', map_location='31F1-31F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10485538, end_position_on_the_genomic_accession=10488622, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34443, CurrentID=0, Status='live', Symbol='CG6144', Aliases='Dmel_ Dmel\\CG6144', description='CG6144 gene product from transcript CG6144-RB', other_designations='CG6144-PB|CG6144-PC', map_location='31F1-31F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10488694, end_position_on_the_genomic_accession=10490461, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34444, CurrentID=0, Status='live', Symbol='CG13144', Aliases='Dmel_ Dmel\\CG13144', description='CG13144 gene product from transcript CG13144-RA', other_designations='CG13144-PA|uncharacterized protein', map_location='31F1-31F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10489900, end_position_on_the_genomic_accession=10491454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34445, CurrentID=0, Status='live', Symbol='Myo31DF', Aliases='Dmel_CG7438, CG7438, Dm IA, DmIA, Dmel\\CG7438, DroMIA, IA, MYOIA, Myo-IA, Myo1A, MyoIA, MyoID, myo31DF, myoID, souther', description='Myosin 31DF', other_designations='CG7438-PA|CG7438-PB|CG7438-PC|Myo31DF-PA|Myo31DF-PB|Myo31DF-PC|Myosin31DF|myosin 1D|myosin 1a|myosin 31DF|myosin I|myosin IA|myosin ID|myosin-IA|myosin1A|souther|unconventional myosin 31DF', map_location='31F1-31F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10491814, end_position_on_the_genomic_accession=10506779, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34446, CurrentID=0, Status='live', Symbol='CG6094', Aliases='Dmel_ Dmel\\CG6094, FBgn 32261', description='CG6094 gene product from transcript CG6094-RB', other_designations='CG6094-PA|CG6094-PB', map_location='31F4-31F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10507081, end_position_on_the_genomic_accession=10508032, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34447, CurrentID=0, Status='live', Symbol='CG7384', Aliases='Dmel_ Dmel\\CG7384', description='CG7384 gene product from transcript CG7384-RA', other_designations='CG7384-PA', map_location='31F4-31F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10507859, end_position_on_the_genomic_accession=10510453, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34449, CurrentID=0, Status='live', Symbol='Lrr47', Aliases='Dmel_CG6098, CG6098, CT19173, Dmel\\CG6098, LRR47, Lrr-47', description='Leucine-rich repeat 47', other_designations='CG6098-PA|CG6098-PB|Lrr47-PA|Lrr47-PB', map_location='31F5-31F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10517335, end_position_on_the_genomic_accession=10519269, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34450, CurrentID=0, Status='live', Symbol='Lip4', Aliases='Dmel_CG6113, CG6113, Dmel\\CG6113, dLip4, dlip4', description='Lipase 4', other_designations='CG6113-PA|CG6113-PB|CG6113-PC|Lip4-PA|Lip4-PB|Lip4-PC|lipase 4|lipase4', map_location='31F5-31F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10527572, end_position_on_the_genomic_accession=10531778, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34451, CurrentID=0, Status='live', Symbol='CG18301', Aliases='Dmel_ Dmel\\CG18301', description='CG18301 gene product from transcript CG18301-RA', other_designations='CG18301-PA|uncharacterized protein', map_location='31F5-31F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10531821, end_position_on_the_genomic_accession=10533647, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34452, CurrentID=0, Status='live', Symbol='CG18302', Aliases='Dmel_ Dmel\\CG18302', description='CG18302 gene product from transcript CG18302-RA', other_designations='CG18302-PA|uncharacterized protein', map_location='31F5-31F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10534322, end_position_on_the_genomic_accession=10536587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34453, CurrentID=0, Status='live', Symbol='Trim9', Aliases='Dmel_CG31721, Asap, CG13145, CG31721, CG6256, Dmel\\CG31721, asap, dTrim9', description='Tripartite motif containing 9', other_designations='CG31721-PA|CG31721-PB|Trim9-PA|Trim9-PB|anomalies in sensory axon patterning|tripartite motif containing 9', map_location='31F5-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10544883, end_position_on_the_genomic_accession=10628054, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34454, CurrentID=0, Status='live', Symbol='CG6138', Aliases='Dmel_ BcDNA:AT25107, Dmel\\CG6138', description='CG6138 gene product from transcript CG6138-RA', other_designations='CG6138-PA', map_location='32A1-32A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10567255, end_position_on_the_genomic_accession=10568012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34455, CurrentID=0, Status='live', Symbol='w-cup', Aliases='Dmel_CG7363, BcDNA:AT31112, CG7363, Dmel\\CG7363', description='world cup', other_designations='CG7363-PA|CG7363-PB|w-cup-PA|w-cup-PB|world-cup', map_location='32A1-32A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10575860, end_position_on_the_genomic_accession=10577792, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34457, CurrentID=0, Status='live', Symbol='CG7329', Aliases='Dmel_ Dmel\\CG7329, ORE-4', description='CG7329 gene product from transcript CG7329-RB', other_designations='CG7329-PA|CG7329-PB', map_location='32A2-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10629352, end_position_on_the_genomic_accession=10638471, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34458, CurrentID=0, Status='live', Symbol='CG31872', Aliases='Dmel_ CG17093, CG17101, Dmel\\CG31872', description='CG31872 gene product from transcript CG31872-RB', other_designations='CG31872-PA|CG31872-PB|uncharacterized protein', map_location='32A2-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10639931, end_position_on_the_genomic_accession=10643539, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740143, CurrentID=0, Status='live', Symbol='CG34207', Aliases='Dmel_ Dmel\\CG34207', description='CG34207 gene product from transcript CG34207-RA', other_designations='CG34207-PA|uncharacterized protein', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22329195, end_position_on_the_genomic_accession=22329720, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34460, CurrentID=0, Status='live', Symbol='CG18284', Aliases='Dmel_ Acp, BEST:GH10507, Dmel\\CG18284', description='CG18284 gene product from transcript CG18284-RC', other_designations='CG18284-PB|CG18284-PC|uncharacterized protein', map_location='32A2-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10643971, end_position_on_the_genomic_accession=10645891, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34461, CurrentID=0, Status='live', Symbol='CG17097', Aliases='Dmel_ Dmel\\CG17097, anon-SAGE:Wang-018', description='CG17097 gene product from transcript CG17097-RB', other_designations='CG17097-PB|anon-SAGE:Wang-018|uncharacterized protein', map_location='32A2-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10646415, end_position_on_the_genomic_accession=10650152, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34462, CurrentID=0, Status='live', Symbol='CG17098', Aliases='Dmel_ Dmel\\CG17098', description='CG17098 gene product from transcript CG17098-RA', other_designations='CG17098-PA|uncharacterized protein', map_location='32A2-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10651613, end_position_on_the_genomic_accession=10654499, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34463, CurrentID=0, Status='live', Symbol='CG31871', Aliases='Dmel_ CG17087, CG17099, Dmel\\CG31871', description='CG31871 gene product from transcript CG31871-RA', other_designations='CG31871-PA|uncharacterized protein', map_location='32A2-32A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10662081, end_position_on_the_genomic_accession=10671487, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34465, CurrentID=0, Status='live', Symbol='CG17104', Aliases='Dmel_ BEST:LD16579, Dmel\\CG17104', description='CG17104 gene product from transcript CG17104-RB', other_designations='CG17104-PA|CG17104-PB|CG17104-PC|uncharacterized protein', map_location='32A3-32A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10677198, end_position_on_the_genomic_accession=10679707, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34466, CurrentID=0, Status='live', Symbol='CG17105', Aliases='Dmel_ Dmel\\CG17105', description='CG17105 gene product from transcript CG17105-RB', other_designations='CG17105-PA|CG17105-PB|uncharacterized protein', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10681030, end_position_on_the_genomic_accession=10681775, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34467, CurrentID=0, Status='live', Symbol='CG17107', Aliases='Dmel_ Dmel\\CG17107', description='CG17107 gene product from transcript CG17107-RB', other_designations='CG17107-PA|CG17107-PB|CG17107-PC|uncharacterized protein', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10681902, end_position_on_the_genomic_accession=10682528, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34468, CurrentID=0, Status='live', Symbol='CG7299', Aliases='Dmel_ BcDNA:RE24635, CT22515, Dmel\\CG7299', description='CG7299 gene product from transcript CG7299-RC', other_designations='CG7299-PA|CG7299-PC', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10683427, end_position_on_the_genomic_accession=10684409, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34469, CurrentID=0, Status='live', Symbol='CG7296', Aliases='Dmel_ BcDNA:RH68528, Dmel\\CG7296', description='CG7296 gene product from transcript CG7296-RB', other_designations='CG7296-PA|CG7296-PB', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10686124, end_position_on_the_genomic_accession=10687066, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34470, CurrentID=0, Status='live', Symbol='CG7294', Aliases='Dmel_ BcDNA:RE28280, Dmel\\CG7294', description='CG7294 gene product from transcript CG7294-RA', other_designations='CG7294-PA', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10688734, end_position_on_the_genomic_accession=10689271, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34471, CurrentID=0, Status='live', Symbol='CG17108', Aliases='Dmel_ BEST:GH23489, CT33647, Dmel\\CG17108, gh23489', description='CG17108 gene product from transcript CG17108-RA', other_designations='CG17108-PA|uncharacterized protein', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10692940, end_position_on_the_genomic_accession=10694103, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34472, CurrentID=0, Status='live', Symbol='CG7300', Aliases='Dmel_ Dmel\\CG7300', description='CG7300 gene product from transcript CG7300-RD', other_designations='CG7300-PB|CG7300-PD', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10695000, end_position_on_the_genomic_accession=10698043, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34474, CurrentID=0, Status='live', Symbol='CG6415', Aliases='Dmel_ 151967_at, Dmel\\CG6415, anon-32Aa', description='CG6415 gene product from transcript CG6415-RA', other_designations='CG6415-PA', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10703621, end_position_on_the_genomic_accession=10704930, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34476, CurrentID=0, Status='live', Symbol='CG6431', Aliases='Dmel_ Dmel\\CG6431, chr2L_10724193_10724543.0', description='CG6431 gene product from transcript CG6431-RC', other_designations='CG6431-PB|CG6431-PC', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10724203, end_position_on_the_genomic_accession=10728740, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34477, CurrentID=0, Status='live', Symbol='CG6443', Aliases='Dmel_ Dmel\\CG6443', description='CG6443 gene product from transcript CG6443-RA', other_designations='CG6443-PA', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10728978, end_position_on_the_genomic_accession=10730302, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34478, CurrentID=0, Status='live', Symbol='CG17118', Aliases='Dmel_ Dmel\\CG17118', description='CG17118 gene product from transcript CG17118-RA', other_designations='CG17118-PA|uncharacterized protein', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10730255, end_position_on_the_genomic_accession=10731356, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34479, CurrentID=0, Status='live', Symbol='EMC3', Aliases='Dmel_CG6750, CG6750, Dmel\\CG6750, dPob', description='ER Membrane protein Complex 3', other_designations='CG6750-PA|EMC3-PA|ER membrane protein complex 3', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10731400, end_position_on_the_genomic_accession=10732423, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34480, CurrentID=0, Status='live', Symbol='Dpy-30L1', Aliases='Dmel_CG6444, CG6444, DPY30, Dmel\\CG6444, Dpy30, dDPY30-L1, dpy-30L1, dpy30', description='Dpy-30-like 1', other_designations='CG6444-PA|Dpy-30L1-PA', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10732653, end_position_on_the_genomic_accession=10733307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34481, CurrentID=0, Status='live', Symbol='Nup107', Aliases='Dmel_CG6743, BcDNA:LD18761, CG6743, Dmel\\CG6743, Nup 107, Nup170, dmNup107, nup107', description='Nucleoporin 107kD', other_designations='CG6743-PA|Nup107-PA|Nup170|nucleoporin 107|nucleoporin 107kD', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10733293, end_position_on_the_genomic_accession=10736071, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34482, CurrentID=0, Status='live', Symbol='Vha16-5', Aliases='Dmel_CG6737, CG6737, Dmel\\CG6737, vha16-5', description='Vacuolar H[+] ATPase 16kD subunit 5', other_designations='CG6737-PA|Vha16-5-PA|vacuolar H[+] ATPase 16kD subunit 5|vacuolar H[+] ATPase subunit 16-5', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10736319, end_position_on_the_genomic_accession=10737043, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34483, CurrentID=0, Status='live', Symbol='CG12299', Aliases='Dmel_ Dmel\\CG12299', description='CG12299 gene product from transcript CG12299-RA', other_designations='CG12299-PA|uncharacterized protein', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10737345, end_position_on_the_genomic_accession=10740762, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34484, CurrentID=0, Status='live', Symbol='CG6729', Aliases='Dmel_ Dmel\\CG6729', description='CG6729 gene product from transcript CG6729-RA', other_designations='CG6729-PA', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10740782, end_position_on_the_genomic_accession=10743935, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34485, CurrentID=0, Status='live', Symbol='CG17124', Aliases='Dmel_ Dmel\\CG17124, anon-WO0118547.153', description='CG17124 gene product from transcript CG17124-RB', other_designations='CG17124-PA|CG17124-PB|CG17124-PC|uncharacterized protein', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10744075, end_position_on_the_genomic_accession=10756325, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34486, CurrentID=0, Status='live', Symbol='CG6495', Aliases='Dmel_ BcDNA:GH10711, Dmel\\CG6495, GH10711', description='CG6495 gene product from transcript CG6495-RA', other_designations='CG6495-PA', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10761116, end_position_on_the_genomic_accession=10764971, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34487, CurrentID=0, Status='live', Symbol='Ubc2', Aliases='Dmel_CG6720, CG6720, Dmel\\CG6720, UbcD2, l(2)k13206, ubcD2', description='Ubiquitin conjugating enzyme 2', other_designations='CG6720-PA|CG6720-PB|CG6720-PC|CG6720-PD|Ubc2-PA|Ubc2-PB|Ubc2-PC|Ubc2-PD|ubiquitin conjugating enzyme 2', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10765010, end_position_on_the_genomic_accession=10768006, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34488, CurrentID=0, Status='live', Symbol='CG6724', Aliases='Dmel_ Dmel\\CG6724', description='CG6724 gene product from transcript CG6724-RB', other_designations='CG6724-PA|CG6724-PB', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10768234, end_position_on_the_genomic_accession=10769952, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34489, CurrentID=0, Status='live', Symbol='CG17127', Aliases='Dmel_ Dmel\\CG17127', description='CG17127 gene product from transcript CG17127-RB', other_designations='CG17127-PA|CG17127-PB|uncharacterized protein', map_location='32A5-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10770868, end_position_on_the_genomic_accession=10771370, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34490, CurrentID=0, Status='live', Symbol='CG31869', Aliases='Dmel_ CG17130, CG18499, Dmel\\CG31869', description='CG31869 gene product from transcript CG31869-RC', other_designations='CG31869-PA|CG31869-PB|CG31869-PC|uncharacterized protein', map_location='32A5-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10772038, end_position_on_the_genomic_accession=10801109, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34493, CurrentID=0, Status='live', Symbol='CG6508', Aliases='Dmel_ Dmel\\CG6508', description='CG6508 gene product from transcript CG6508-RA', other_designations='CG6508-PA', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10819107, end_position_on_the_genomic_accession=10820485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34494, CurrentID=0, Status='live', Symbol='CG17134', Aliases='Dmel_ Dmel\\CG17134', description='CG17134 gene product from transcript CG17134-RA', other_designations='CG17134-PA|uncharacterized protein', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10820705, end_position_on_the_genomic_accession=10821967, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34495, CurrentID=0, Status='live', Symbol='Nos', Aliases='Dmel_CG6713, CG6713, DNOS, DNOS1, Dmel\\CG6713, NOS, dNOS, dNos, drNOSoxy, nNOS, nos', description='Nitric oxide synthase', other_designations='CG6713-PA|CG6713-PF|CG6713-PK|NO synthase|Nos-PA|Nos-PF|Nos-PK|nitric oixide synthase|nitric oxide synthase', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10804274, end_position_on_the_genomic_accession=10837511, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  34496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34496, CurrentID=0, Status='live', Symbol='CG6700', Aliases='Dmel_ Dmel\\CG6700', description='CG6700 gene product from transcript CG6700-RA', other_designations='CG6700-PA|CG6700-PB', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10837732, end_position_on_the_genomic_accession=10842597, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34497, CurrentID=0, Status='live', Symbol='CG17139', Aliases='Dmel_ CG31722, CG31722-A, Dmel\\CG17139', description='CG17139 gene product from transcript CG17139-RA', other_designations='CG17139-PA|uncharacterized protein', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10843014, end_position_on_the_genomic_accession=10845664, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34499, CurrentID=0, Status='live', Symbol='Porin2', Aliases='Dmel_CG17137, BEST:LP07660, CG17137, Dm Dmel\\CG17137, LP07660.3prime, Porin 2, porin 2', description='CG17137 gene product from transcript CG17137-RB', other_designations='CG17137-PA|CG17137-PB|Porin2-PA|Porin2-PB|porin 2|porin2', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10845788, end_position_on_the_genomic_accession=10847044, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34500, CurrentID=0, Status='live', Symbol='porin', Aliases='Dmel_CG6647, CG6647, DVDAC, DmPorin1, DmVDAC, Dmel\\CG6647, POR-1, PORIN, Porin, VDAC, Vdac, anon-WO02059370.56, cg6647, l(2)k05123', description='CG6647 gene product from transcript CG6647-RB', other_designations='CG6647-PA|CG6647-PB|CG6647-PC|CG6647-PD|CG6647-PE|Voltage-dependent anion-selective channel|mitochondrial porin|porin 1|porin-PA|porin-PB|porin-PC|porin-PD|porin-PE|voltage-dependent anion channel', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10847227, end_position_on_the_genomic_accession=10850848, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34502, CurrentID=0, Status='live', Symbol='CG43129', Aliases='Dmel_ CG17085, CG17086, Dmel\\CG43129, Dmel_CG17085, Dmel_CG17086', description='CG43129 gene product from transcript CG43129-RE', other_designations='CG43129-PC|CG43129-PD|CG43129-PE|uncharacterized protein', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10850141, end_position_on_the_genomic_accession=10854604, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34503, CurrentID=0, Status='live', Symbol='Dnz1', Aliases='Dmel_CG6627, CG6627, DNZ1, Dmel\\CG6627', description='DNZDHHC/NEW1 zinc finger protein 11', other_designations='CG6627-PA|DNZDHHC/NEW1 zinc-finger protein 11|Dnz1-PA', map_location='32B1-32B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10854435, end_position_on_the_genomic_accession=10856644, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34504, CurrentID=0, Status='live', Symbol='aurB', Aliases='Dmel_CG6620, ABK, Aur B, AurB, Aurora B, Aurora-B, CG6620, DmAIRK2, DmAurB, DmAuroraB, Dmel\\CG6620, IAL, IAL1, Ial, ial, ial/AurB', description='aurora B', other_designations='AURORA-like|CG6620-PA|DmAuroraB|IplI-aurora-like kinase|aurB-PA|aurora|aurora B kinase|aurora-B|auroraB|dAurora B', map_location='32B2-32B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10856797, end_position_on_the_genomic_accession=10858094, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34505, CurrentID=0, Status='live', Symbol='Stam', Aliases='Dmel_CG6521, CG6521, Dmel\\CG6521, Dstam, STAM1, dstam, stam, Stam', description='Signal transducing adaptor molecule', other_designations='CG6521-PA|Stam-PA|signal transducer adaptor molecule|signal transducing adaptor molecule|signal-transducing adaptor molecule', map_location='32B2-32B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10858536, end_position_on_the_genomic_accession=10862006, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34506, CurrentID=0, Status='live', Symbol='CG12517', Aliases='Dmel_ Dmel\\CG12517', description='CG12517 gene product from transcript CG12517-RA', other_designations='CG12517-PA|uncharacterized protein', map_location='32B3-32B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10870167, end_position_on_the_genomic_accession=10870722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34507, CurrentID=0, Status='live', Symbol='CG14071', Aliases='Dmel_ Dmel\\CG14071', description='CG14071 gene product from transcript CG14071-RB', other_designations='CG14071-PA|CG14071-PB|CG14071-PC|uncharacterized protein', map_location='32B3-32B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10887645, end_position_on_the_genomic_accession=10888927, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34508, CurrentID=0, Status='live', Symbol='CG14070', Aliases='Dmel_ Dmel\\CG14070', description='CG14070 gene product from transcript CG14070-RB', other_designations='CG14070-PA|CG14070-PB|uncharacterized protein', map_location='32B4-32B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10900501, end_position_on_the_genomic_accession=10901785, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34509, CurrentID=0, Status='live', Symbol='CG7309', Aliases='Dmel_ Dmel\\CG7309', description='CG7309 gene product from transcript CG7309-RC', other_designations='CG7309-PC', map_location='32B4-32B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10906270, end_position_on_the_genomic_accession=10908893, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34510, CurrentID=0, Status='live', Symbol='CG14069', Aliases='Dmel_ Dmel\\CG14069', description='CG14069 gene product from transcript CG14069-RB', other_designations='CG14069-PA|CG14069-PB|uncharacterized protein', map_location='32B4-32B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10910051, end_position_on_the_genomic_accession=10910648, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740152, CurrentID=0, Status='live', Symbol='CG34342', Aliases='Dmel_ CG10629, CG32422, Dmel\\CG34342, Dmel_CG10629, Dmel_CG32422', description='CG34342 gene product from transcript CG34342-RE', other_designations='CG34342-PD|CG34342-PE|uncharacterized protein', map_location='64D2-64D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5503528, end_position_on_the_genomic_accession=5510860, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  34514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34514, CurrentID=0, Status='live', Symbol='CG33129', Aliases='Dmel_ CG6087, CG6089, Dmel\\CG33129', description='CG33129 gene product from transcript CG33129-RF', other_designations='CG33129-PA|CG33129-PB|CG33129-PC|CG33129-PE|CG33129-PF|CG33129-PG|CG33129-PH|uncharacterized protein', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10964905, end_position_on_the_genomic_accession=10970275, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34516, CurrentID=0, Status='live', Symbol='YL-1', Aliases='Dmel_CG4621, CG4621, Dmel\\CG4621, d', description='CG4621 gene product from transcript CG4621-RB', other_designations='CG4621-PA|CG4621-PB|YL-1-PA|YL-1-PB', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10970443, end_position_on_the_genomic_accession=10972093, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34517, CurrentID=0, Status='live', Symbol='CG16743', Aliases='Dmel_ BcDNA:RH23514, Dmel\\CG16743', description='CG16743 gene product from transcript CG16743-RB', other_designations='CG16743-PA|CG16743-PB|uncharacterized protein', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10972280, end_position_on_the_genomic_accession=10973408, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34519, CurrentID=0, Status='live', Symbol='SCAR', Aliases='Dmel_CG4636, BEST:SD02991, CG4636, D-SCAR, DWave, DmSCAR, Dmel\\CG4636, Dscar/WAVE, Scar, WAVE, WAVE/SCAR, Wave, dWAVE, l(2)k03107, l(2)k13811, scar, wave', description='CG4636 gene product from transcript CG4636-RD', other_designations='CG4636-PA|CG4636-PB|CG4636-PC|CG4636-PD|CG4636-PE|CG4636-PF|SCAR-PA|SCAR-PB|SCAR-PC|SCAR-PD|SCAR-PE|SCAR-PF|WASP family verprolin-homologous protein|scar|suppressor of cAMP receptors|suppressor of cAR|suppressor of cyclic AMP receptor', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10976772, end_position_on_the_genomic_accession=10982395, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34521, CurrentID=0, Status='live', Symbol='piwi', Aliases='Dmel_CG6122, CG6122, Dmel\\CG6122, PIWI, Piwi', description='CG6122 gene product from transcript CG6122-RA', other_designations='CG6122-PA|CG6122-PB|P-element induced wimpy testis|piwi-PA|piwi-PB', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10982205, end_position_on_the_genomic_accession=10987420, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34524, CurrentID=0, Status='live', Symbol='aub', Aliases='Dmel_CG6137, AUB, Aub, CG6137, Dmel\\CG6137, ms(2)32D, sti, sting', description='aubergine', other_designations='CG6137-PA|CG6137-PC|aub-PA|aub-PC|stellate-interacting gene', map_location='2-39 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10997819, end_position_on_the_genomic_accession=11001476, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  5740154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740154, CurrentID=0, Status='live', Symbol='CG34222', Aliases='Dmel_ Dmel\\CG34222, NEST:bs08e11, NEST:bs11e07, NEST:bs12c06', description='CG34222 gene product from transcript CG34222-RA', other_designations='CG34222-PA|NEST:bs08e11|NEST:bs11e07|NEST:bs12c06|uncharacterized protein', map_location='47A1-47A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10319844, end_position_on_the_genomic_accession=10320500, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34526, CurrentID=0, Status='live', Symbol='RpL9', Aliases='Dmel_CG6141, CG6141, Dmel\\CG6141, L9, M(2)32D, Rp L9, Rpl9, anon-EST:fe3A6, anon-WO0153538.25, anon-WO0153538.26, anon-WO0153538.27, rpL9', description='Ribosomal protein L9', other_designations='CG6141-PA|CG6141-PB|CG6141-PC|RpL9-PA|RpL9-PB|RpL9-PC|anon-fast-evolving-3A6|ribosomal protein L9', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11002758, end_position_on_the_genomic_accession=11004189, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34527, CurrentID=0, Status='live', Symbol='Nup154', Aliases='Dmel_CG4579, BcDNA:LD21772, CG4579, Dmel\\CG4579, Nup, Nup155, Nup32D, l(2)01501, l(2)10432, ms(2)zk, nup154, nup32D, tip, tlp, zk', description='Nucleoporin 154kD', other_designations='CG4579-PA|CG4579-PC|CG4579-PD|Nup154-PA|Nup154-PC|Nup154-PD|nucleoporin 154|nucleoporin 154kD|tulipano|zonder kloten', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11004820, end_position_on_the_genomic_accession=11010921, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  34528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34528, CurrentID=0, Status='live', Symbol='Art8', Aliases='Dmel_CG16840, CG16840, DART7, DART8, Dmel\\CG16840, dART8', description='Arginine methyltransferase 8', other_designations='Art8-PA|CG16840-PA|arginine methyltransferase 8', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11009821, end_position_on_the_genomic_accession=11011086, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34529, CurrentID=0, Status='live', Symbol='dUTPase', Aliases='Dmel_CG4584, BcDNA:LD08534, CG4584, Dmel\\CG4584, LD08534, UTPase, anon-SAGE:Wang-077', description='Deoxyuridine triphosphatase', other_designations='CG4584-PA|CG4584-PB|CG4584-PC|CG4584-PD|anon-SAGE:Wang-077|dUTPase-PA|dUTPase-PB|dUTPase-PC|dUTPase-PD|deoxyuridine triphosphatase', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11011312, end_position_on_the_genomic_accession=11012135, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34530, CurrentID=0, Status='live', Symbol='Samuel', Aliases='Dmel_CG31868, 78BP, CG12291, CG14917, CG14918, CG31868, Dmel\\CG31868, Moses, mdcds_1002, moses', description='SAM-motif ubiquitously expressed punctatedly localized protein', other_designations='CG31868-PA|CG31868-PB|CG31868-PD|CG31868-PE|Samuel-PA|Samuel-PB|Samuel-PD|Samuel-PE|middleman of seventy-eight signaling', map_location='32C1-32D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11013206, end_position_on_the_genomic_accession=11066998, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34531, CurrentID=0, Status='live', Symbol='Acp32CD', Aliases='Dmel_CG4605, CG4605, Dmel\\CG4605', description='Accessory gland protein 32CD', other_designations='Acp32CD-PA|Acp32CD-PB|CG4605-PA|CG4605-PB|accessory gland protein 32CD|accessory gland-specific peptide 32CD', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11015620, end_position_on_the_genomic_accession=11016571, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34532, CurrentID=0, Status='live', Symbol='CG14913', Aliases='Dmel_ Dmel\\CG14913', description='CG14913 gene product from transcript CG14913-RB', other_designations='CG14913-PB|uncharacterized protein', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11016995, end_position_on_the_genomic_accession=11018778, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34536, CurrentID=0, Status='live', Symbol='CG14915', Aliases='Dmel_ Dmel\\CG14915', description='CG14915 gene product from transcript CG14915-RA', other_designations='CG14915-PA|uncharacterized protein', map_location='32D2-32D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11071027, end_position_on_the_genomic_accession=11071532, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34537, CurrentID=0, Status='live', Symbol='AstC', Aliases='Dmel_CG14919, ASC, AST-C, ASTC, Ast-C, Ast2, BEST:GH06087, BcDNA:RH36507, CG14919, CG149199, DAP-C, Dmel\\CG14919, Drm-AST-C, FLT, Flt, M-ASH, ast2, flt, gh06087', description='Allatostatin C', other_designations='AstC-PA|AstC-PB|CG14919-PA|CG14919-PB|Drm-AST C|Drostatin-C|allatostatin 2|allatostatin C|allatostatin Manduca-type|allatostatin-C|flatline', map_location='32D2-32D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11076165, end_position_on_the_genomic_accession=11081399, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34538, CurrentID=0, Status='live', Symbol='AstCC', Aliases='Dmel_CG14920, ASCC, ASTCC, Ast-CC, CG14920, Dmel\\CG14920', description='Allatostatin double C', other_designations='AstCC-PB|AstCC-PC|CG14920-PB|CG14920-PC|allatostatin CC|allatostatin double C', map_location='32D2-32D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11083459, end_position_on_the_genomic_accession=11085229, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34539, CurrentID=0, Status='live', Symbol='CG16854', Aliases='Dmel_ Dmel\\CG16854, gi7297815', description='CG16854 gene product from transcript CG16854-RD', other_designations='CG16854-PB|CG16854-PC|CG16854-PD|uncharacterized protein', map_location='32D2-32D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11086721, end_position_on_the_genomic_accession=11090973, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34540, CurrentID=0, Status='live', Symbol='CG4705', Aliases='Dmel_ Dmel\\CG4705, FBtr0080196, SEA3', description='CG4705 gene product from transcript CG4705-RD', other_designations='CG4705-PB|CG4705-PD', map_location='32D2-32D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11091444, end_position_on_the_genomic_accession=11095096, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34541, CurrentID=0, Status='live', Symbol='Ge-1', Aliases='Dmel_CG6181, CG6181, Dmel\\CG6181, EDC4, GE-1, dGe-1', description='Ge-1 ortholog (H. sapiens)', other_designations='CG6181-PA|CG6181-PB|Ge-1 ortholog|Ge-1-PA|Ge-1-PB', map_location='32D3-32D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11095150, end_position_on_the_genomic_accession=11100888, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34542, CurrentID=0, Status='live', Symbol='Reps', Aliases='Dmel_CG6192, CG6192, Dmel\\CG6192, REPS', description='CG6192 gene product from transcript CG6192-RB', other_designations='CG6192-PA|CG6192-PB|Reps-PA|Reps-PB|reps', map_location='32D3-32D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11100963, end_position_on_the_genomic_accession=11106645, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34543, CurrentID=0, Status='live', Symbol='l(2)gd1', Aliases='Dmel_CG4713, CG4713, Dmel\\CG4713, SH495, gd, l(2)SH0495, l(2)SH2 0495, l(2)gd, l(2)giant disc, l[[2]]gd, lgd', description='lethal (2) giant discs 1', other_designations='CG4713-PA|CG4713-PB|l(2)gd1-PA|l(2)gd1-PB|leathal giant disc|lethal (2) SH0495|lethal (2) giant discs|lethal giant discs', map_location='32D3-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11106943, end_position_on_the_genomic_accession=11110851, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34544, CurrentID=0, Status='live', Symbol='CG6201', Aliases='Dmel_ Dmel\\CG6201', description='CG6201 gene product from transcript CG6201-RA', other_designations='CG6201-PA', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11110557, end_position_on_the_genomic_accession=11113061, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34545, CurrentID=0, Status='live', Symbol='Gr32a', Aliases='Dmel_CG14916, 32D.1, 32D1, CG14916, Dmel\\CG14916, GR32D.1, GR32D1, GR32a, Gr32A, Gr32D1, Gr32d', description='Gustatory receptor 32a', other_designations='CG14916-PA|Gr32a-PA|gustatory receptor 32D.1|gustatory receptor 32a', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11114014, end_position_on_the_genomic_accession=11115823, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34546, CurrentID=0, Status='live', Symbol='CG6230', Aliases='Dmel_ BcDNA:GH06032, Dmel\\CG6230, dCOD1', description='CG6230 gene product from transcript CG6230-RA', other_designations='CG6230-PA', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11115635, end_position_on_the_genomic_accession=11120319, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34547, CurrentID=0, Status='live', Symbol='CG14921', Aliases='Dmel_ Dmel\\CG14921', description='CG14921 gene product from transcript CG14921-RA', other_designations='CG14921-PA|uncharacterized protein', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11120642, end_position_on_the_genomic_accession=11121714, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34548, CurrentID=0, Status='live', Symbol='Csl4', Aliases='Dmel_CG6249, CG6249, CSL4, Dmel\\CG6249, d', description='CG6249 gene product from transcript CG6249-RA', other_designations='CG6249-PA|Csl4-PA', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11122574, end_position_on_the_genomic_accession=11123567, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34549, CurrentID=0, Status='live', Symbol='Nup160', Aliases='Dmel_CG4738, CG4738, Dmel\\CG4738, Nup[mel], l(2)SH2 2055, l(2)SH2055, nup160, Nup160', description='Nucleoporin 160kD', other_designations='CG4738-PA|CG4738-PB|Nup160-PA|Nup160-PB|lethal (2) SH2055|nuclear pore protein 160|nucleoporin 160|nucleoporin 160kD', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11123777, end_position_on_the_genomic_accession=11129335, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  34550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34550, CurrentID=0, Status='live', Symbol='RfC38', Aliases='Dmel_CG6258, BcDNA:LD06837, CG6258, DmRFC5, Dmel\\CG6258, Rfc38, l(2)k13807, l(2)rfc38, n(2)k13807', description='Replication factor C 38kD subunit', other_designations='CG6258-PA|CG6258-PB|DNA replication factor, 38k subunit|RfC38-PA|RfC38-PB|replication factor C 38kD subunit', map_location='32D4-32D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11128932, end_position_on_the_genomic_accession=11130645, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34551, CurrentID=0, Status='live', Symbol='CG4751', Aliases='Dmel_ Dmel\\CG4751, Q9VKJ1', description='CG4751 gene product from transcript CG4751-RA', other_designations='CG4751-PA', map_location='32D4-32D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11130737, end_position_on_the_genomic_accession=11136003, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34552, CurrentID=0, Status='live', Symbol='hgo', Aliases='Dmel_CG4779, CG4779, Dmel\\CG4779', description='homogentisate 1,2-dioxygenase', other_designations='CG4779-PA|hgo-PA', map_location='32D5-32D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11136393, end_position_on_the_genomic_accession=11138565, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34553, CurrentID=0, Status='live', Symbol='Gyc32E', Aliases='Dmel_CG33114, CG33114, CG6275, Cyg1, DMRGC, Dmel\\CG33114, GC, GYC 32E, Gyc-r', description='Guanyl cyclase at 32E', other_designations='CG33114-PB|CG33114-PD|CG33114-PG|Gyc32E-PB|Gyc32E-PD|Gyc32E-PG|guanyl cyclase at 32E|guanylate cyclase receptor-type', map_location='32D5-32D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11138759, end_position_on_the_genomic_accession=11149101, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  34554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34554, CurrentID=0, Status='live', Symbol='CG6287', Aliases='Dmel_ Dmel\\CG6287, GH26', description='CG6287 gene product from transcript CG6287-RA', other_designations='CG6287-PA', map_location='32D5-32D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11151506, end_position_on_the_genomic_accession=11155721, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34557, CurrentID=0, Status='live', Symbol='Ca-beta', Aliases='Dmel_CG42403, CG14922, CG14923, CG33305, CG42403, CG6320, CaBeta, DmCa1beta, Dmel\\CG42403, Dmel_CG33305, Dmel_CG6320, ca-beta, dmcab1', description='Ca2+-channel-protein-beta-subunit', other_designations='CG42403-PE|CG42403-PG|CG42403-PJ|CG42403-PL|CG42403-PM|CG42403-PN|CG42403-PO|CG42403-PP|CG42403-PQ|Ca-beta-PE|Ca-beta-PG|Ca-beta-PJ|Ca-beta-PL|Ca-beta-PM|Ca-beta-PN|Ca-beta-PO|Ca-beta-PP|Ca-beta-PQ|Ca2+-channel-protein beta subunit|Ca[2+]-channel-protein-beta-subunit|calcium channel beta|calcium channel beta subunit', map_location='32D5-32E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11156118, end_position_on_the_genomic_accession=11203544, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  34558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34558, CurrentID=0, Status='live', Symbol='Vm32E', Aliases='Dmel_CG16874, CG16874, Dmel\\CG16874, VM321, VM32E, VMP32E, Vm3a, Vm32Ec, Vm32E', description='Vitelline membrane 32E', other_designations='CG16874-PA|CG16874-PB|Vm32E-PA|Vm32E-PB|vitelline membrane 32E', map_location='32E1-32E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11171167, end_position_on_the_genomic_accession=11171665, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34559, CurrentID=0, Status='live', Symbol='CG4788', Aliases='Dmel_ Dmel\\CG4788', description='CG4788 gene product from transcript CG4788-RA', other_designations='CG4788-PA|CG4788-PB', map_location='32E1-32E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11179032, end_position_on_the_genomic_accession=11180430, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34560, CurrentID=0, Status='live', Symbol='ab', Aliases='Dmel_CG43860, Ab, B2149, CG32830, CG43860, CG4807, Dmel\\CG43860, Dmel_CG32830, Dmel_CG4807, clu, l(2)k02807, ptd', description='abrupt', other_designations='CG43860-PA|CG43860-PB|CG43860-PD|CG43860-PE|CG43860-PF|ab-PA|ab-PB|ab-PD|ab-PE|ab-PF|clueless|parted', map_location='2-44.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11210678, end_position_on_the_genomic_accession=11261080, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  34564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34564, CurrentID=0, Status='live', Symbol='Ppt2', Aliases='Dmel_CG4851, CG4851, Dmel\\CG4851, PPT-2', description='Palmitoyl-protein Thioesterase 2', other_designations='CG4851-PA|Palmitoyl-protein thioesterase 2|Ppt2-PA', map_location='32E2-32E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11280298, end_position_on_the_genomic_accession=11281920, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34565, CurrentID=0, Status='live', Symbol='mre11', Aliases='Dmel_CG16928, 16928, CG16928, Dmel\\CG16928, MRE11, Mre11', description='meiotic recombination 11', other_designations='CG16928-PA|mre11-PA', map_location='32E2-32E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11282090, end_position_on_the_genomic_accession=11284335, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34566, CurrentID=0, Status='live', Symbol='Osi21', Aliases='Dmel_CG14925, CG14925, Dmel\\CG14925', description='Osiris 21', other_designations='CG14925-PA|Osi21-PA|osiris|osiris 21', map_location='32E2-32E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11284327, end_position_on_the_genomic_accession=11285706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34567, CurrentID=0, Status='live', Symbol='CG14926', Aliases='Dmel_ Dmel\\CG14926, bs34f02.y1', description='CG14926 gene product from transcript CG14926-RB', other_designations='CG14926-PA|CG14926-PB|CG14926-PC|uncharacterized protein', map_location='32E4-32E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11342301, end_position_on_the_genomic_accession=11343318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34568, CurrentID=0, Status='live', Symbol='salr', Aliases='Dmel_CG4881, 32F, CG4881, Cf3, Dmel\\CG4881, Sal-r, Salr, Spalt, sal, sal-r, salR, spalt', description='spalt-related', other_designations='CG4881-PA|CG4881-PB|CG4881-PC|Spalt-related|chorion factor 3|sal related|sal-related|salr-PA|salr-PB|salr-PC|spalt|spalt related', map_location='32E4-32F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11358873, end_position_on_the_genomic_accession=11373636, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34569, CurrentID=0, Status='live', Symbol='salm', Aliases='Dmel_CG6464, 3602, B1164, CG6464, Dmel\\CG6464, SAL, SALM, SPALT, Sal, SalM, Salm, Spalt, l(2)03602, sal, salII, salM, slam, spalt', description='spalt major', other_designations='CG6464-PA|CG6464-PB|Spalt-M|sal major|salm-PA|salm-PB|spalt|spalt M|spalt majr|spalt-major', map_location='2-44 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11434311, end_position_on_the_genomic_accession=11445614, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34570, CurrentID=0, Status='live', Symbol='sala', Aliases='Dmel_CG4922, BcDNA:RE40068, CG4922, Dmel\\CG4922, chr2L:11486634..11486740, nsal, sal', description='spalt-adjacent', other_designations='CG4922-PA|CG4922-PC|not-spalt|sal adjacent|sala-PA|sala-PC|spalt|spalt accessory', map_location='2-44 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11484707, end_position_on_the_genomic_accession=11486949, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34571, CurrentID=0, Status='live', Symbol='CG6488', Aliases='Dmel_ Dmel\\CG6488', description='CG6488 gene product from transcript CG6488-RA', other_designations='CG6488-PA', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11496804, end_position_on_the_genomic_accession=11498948, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34572, CurrentID=0, Status='live', Symbol='spz4', Aliases='Dmel_CG14928, CG14928, CT34755, Dmel\\CG14928, Spz4', description='spatzle 4', other_designations='CG14928-PB|spz4-PB', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11501715, end_position_on_the_genomic_accession=11508649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34573, CurrentID=0, Status='live', Symbol='CG6509', Aliases='Dmel_ DLG5, Dlg5, Dmel\\CG6509, LD32687', description='CG6509 gene product from transcript CG6509-RA', other_designations='CG6509-PA|CG6509-PB|CG6509-PC|CG6509-PD', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11508399, end_position_on_the_genomic_accession=11517081, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34576, CurrentID=0, Status='live', Symbol='CG14930', Aliases='Dmel_ Dmel\\CG14930', description='CG14930 gene product from transcript CG14930-RA', other_designations='CG14930-PA|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11520358, end_position_on_the_genomic_accession=11521285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34577, CurrentID=0, Status='live', Symbol='CG31705', Aliases='Dmel_ BcDNA:GH07269, CG6528, Dmel\\CG31705', description='CG31705 gene product from transcript CG31705-RE', other_designations='CG31705-PA|CG31705-PD|CG31705-PE|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11523021, end_position_on_the_genomic_accession=11528205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34578, CurrentID=0, Status='live', Symbol='Mst33A', Aliases='Dmel_CG6541, CG6541, Dmel\\CG6541', description='CG6541 gene product from transcript CG6541-RB', other_designations='CG6541-PA|CG6541-PB|CG6541-PC|Mst33A-PA|Mst33A-PB|Mst33A-PC', map_location='32F3-32F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11585433, end_position_on_the_genomic_accession=11586833, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34581, CurrentID=0, Status='live', Symbol='spag4', Aliases='Dmel_CG6589, CG6589, Dmel\\CG6589, Dspag4, dspag, giac', description='sperm-associated antigen 4', other_designations='CG6589-PB|giacomo|spag4-PB', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11610510, end_position_on_the_genomic_accession=11611772, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34582, CurrentID=0, Status='live', Symbol='kek2', Aliases='Dmel_CG4977, CG4977, CT15962, Dmel\\CG4977, Kek2, NB7', description='kekkon-2', other_designations='CG4977-PA|kek2-PA|kekkon', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11581937, end_position_on_the_genomic_accession=11626883, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34583, CurrentID=0, Status='live', Symbol='CG6614', Aliases='Dmel_ Dmel\\CG6614', description='CG6614 gene product from transcript CG6614-RC', other_designations='CG6614-PA|CG6614-PB|CG6614-PC', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11644619, end_position_on_the_genomic_accession=11648839, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34584, CurrentID=0, Status='live', Symbol='CG12307', Aliases='Dmel_ Dmel\\CG12307', description='CG12307 gene product from transcript CG12307-RA', other_designations='CG12307-PA|uncharacterized protein', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11651331, end_position_on_the_genomic_accession=11652010, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34585, CurrentID=0, Status='live', Symbol='CG4983', Aliases='Dmel_ Dmel\\CG4983, anon-WO0140519.134', description='CG4983 gene product from transcript CG4983-RB', other_designations='CG4983-PA|CG4983-PB', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11652448, end_position_on_the_genomic_accession=11654164, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34586, CurrentID=0, Status='live', Symbol='CG4988', Aliases='Dmel_ Dmel\\CG4988', description='CG4988 gene product from transcript CG4988-RA', other_designations='CG4988-PA', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11694804, end_position_on_the_genomic_accession=11696107, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34587, CurrentID=0, Status='live', Symbol='Vha100-5', Aliases='Dmel_CG12602, CG12602, Dmel\\CG12602, vha100-5', description='Vacuolar H[+] ATPase 100kD subunit 5', other_designations='CG12602-PA|CG12602-PB|Vha100-5-PA|Vha100-5-PB|vacuolar H[+] ATPase 100kD subunit 5|vacuolar H[+] ATPase subunit 100-5', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11738948, end_position_on_the_genomic_accession=11744524, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34588, CurrentID=0, Status='live', Symbol='CG14931', Aliases='Dmel_ Dmel\\CG14931', description='CG14931 gene product from transcript CG14931-RC', other_designations='CG14931-PB|CG14931-PC|uncharacterized protein', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11788334, end_position_on_the_genomic_accession=11789475, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34589, CurrentID=0, Status='live', Symbol='CG14932', Aliases='Dmel_ Dmel\\CG14932', description='CG14932 gene product from transcript CG14932-RB', other_designations='CG14932-PB|CG14932-PC|uncharacterized protein', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11789919, end_position_on_the_genomic_accession=11791011, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34590, CurrentID=0, Status='live', Symbol='Tsp33B', Aliases='Dmel_CG14936, CG14936, Dm.Tsp33B, Dmel\\CG14936', description='Tetraspanin 33B', other_designations='CG14936-PA|Tsp33B-PA|tetraspanin 33B', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11791025, end_position_on_the_genomic_accession=11792365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34591, CurrentID=0, Status='live', Symbol='CG14937', Aliases='Dmel_ Dmel\\CG14937, FBXO17', description='CG14937 gene product from transcript CG14937-RA', other_designations='CG14937-PA|uncharacterized protein', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11792412, end_position_on_the_genomic_accession=11794086, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34592, CurrentID=0, Status='live', Symbol='crol', Aliases='Dmel_CG14938, CG14938, Crol, Dmel\\CG14938, ORE-15, anon-EST:Liang-1.74, anon-EST:Liang-2.48, clone 1.74, clone 2.48 alpha, crol beta, crol gamma, l(2)04418, l(2)s2346, crol', description='crooked legs', other_designations='CG14938-PA|CG14938-PB|CG14938-PC|CG14938-PD|CG14938-PE|CG14938-PF|CG14938-PG|CG14938-PH|CG14938-PI|crol-PA|crol-PB|crol-PC|crol-PD|crol-PE|crol-PF|crol-PG|crol-PH|crol-PI|crooked-legs', map_location='33A1-33A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11794327, end_position_on_the_genomic_accession=11809438, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34593, CurrentID=0, Status='live', Symbol='CycY', Aliases='Dmel_CG14939, CG14939, Dcyclin Y, Dmel\\CG14939, anon-WO0118547.165, i182', description='Cyclin Y', other_designations='CG14939-PA|CG14939-PB|CycY-PA|CycY-PB|cyclin Y|drosophila cyclin Y', map_location='33A2-33A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11810092, end_position_on_the_genomic_accession=11813780, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34594, CurrentID=0, Status='live', Symbol='Pde1c', Aliases='Dmel_CG44007, CG14940, CG14942, CG14943, CG14944, CG31757, CG31758, CG42325, CG44007, Dmel\\CG44007, Dmel_CG14940, Dmel_CG14943, Dmel_CG14944, Dmel_CG31757, Dmel_CG31758, Dmel_CG42325, PDE1C, PDE1c, mdcds_1440', description='CG44007 gene product from transcript CG44007-RG', other_designations='CG44007-PB|CG44007-PC|CG44007-PD|CG44007-PE|CG44007-PG|CG44007-PH|CG44007-PJ|CG44007-PK|CG44007-PL|Pde1c-PB|Pde1c-PC|Pde1c-PD|Pde1c-PE|Pde1c-PG|Pde1c-PH|Pde1c-PJ|Pde1c-PK|Pde1c-PL|phosphodiesterase 1c', map_location='33A8-33B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11814156, end_position_on_the_genomic_accession=11928572, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  34595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34595, CurrentID=0, Status='live', Symbol='esc', Aliases='Dmel_CG14941, CG14941, Dmel\\CG14941, ESC, Esc, L41867, dESC, sc', description='extra sexcombs', other_designations='CG14941-PA|esc-PA|extra Sex combs|extra sex comb', map_location='33A2-33A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11827204, end_position_on_the_genomic_accession=11829405, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34597, CurrentID=0, Status='live', Symbol='Mal-B1', Aliases='Dmel_CG14934, CG14934, Dmel\\CG14934, Drome_B1, mal_B1', description='Maltase B1', other_designations='CG14934-PA|Mal-B1-PA|maltase B1', map_location='33A3-33A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11847226, end_position_on_the_genomic_accession=11849169, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34598, CurrentID=0, Status='live', Symbol='Mal-B2', Aliases='Dmel_CG14935, CG14935, Dmel\\CG14935, Drome_B2, FBgn0032382, mal_B2', description='Maltase B2', other_designations='CG14935-PA|CG14935-PC|CG14935-PD|Mal-B2-PA|Mal-B2-PC|Mal-B2-PD|maltase B2', map_location='33A4-33A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11849958, end_position_on_the_genomic_accession=11852881, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34601, CurrentID=0, Status='live', Symbol='Or33a', Aliases='Dmel_CG16960, 33B.1, 33a, AN3, CG16960, DOR33B.1, DOR73, Dmel\\CG16960, Or33B.1, Or73, dor73', description='Odorant receptor 33a', other_designations='CG16960-PA|Or33a-PA|odorant receptor 33a|olfactory receptor 33B.1', map_location='33B2-33B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11934505, end_position_on_the_genomic_accession=11935716, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34602, CurrentID=0, Status='live', Symbol='Or33b', Aliases='Dmel_CG16961, 33B.2, 33b, AN1, CG16961, DOR33B.2, DOR72, Dmel\\CG16961, Or33B.2, Or72, dor72', description='Odorant receptor 33b', other_designations='CG16961-PA|Or33b-PA|odorant receptor 33b|olfactory receptor 33B.2', map_location='33B2-33B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11936181, end_position_on_the_genomic_accession=11937501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34603, CurrentID=0, Status='live', Symbol='Or33c', Aliases='Dmel_CG5006, 33B.3, 33c, AN2, CG5006, DOR33B.3, DOR71, Dmel\\CG5006, Or33B.3, Or71, dor71', description='Odorant receptor 33c', other_designations='CG5006-PA|Or33c-PA|odorant receptor 33c|olfactory receptor 33B.3', map_location='33B2-33B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11937969, end_position_on_the_genomic_accession=11939202, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34604, CurrentID=0, Status='live', Symbol='Crys', Aliases='Dmel_CG16963, CG16963, Cry, DmelCry, Dmel\\CG16963, c, dcy', description='Crystallin', other_designations='CG16963-PA|CG16963-PB|Crys-PA|Crys-PB|crystallin|cyrstallin|ddosocrystallin|droscrystallin|drosocrystallin', map_location='33B2-33B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11944129, end_position_on_the_genomic_accession=11947131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34605, CurrentID=0, Status='live', Symbol='CG16964', Aliases='Dmel_ BcDNA:GH05159, Dmel\\CG16964', description='CG16964 gene product from transcript CG16964-RB', other_designations='CG16964-PA|CG16964-PB|uncharacterized protein', map_location='33B2-33B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11947941, end_position_on_the_genomic_accession=11948803, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34607, CurrentID=0, Status='live', Symbol='CG16965', Aliases='Dmel_ Dmel\\CG16965', description='CG16965 gene product from transcript CG16965-RC', other_designations='CG16965-PC|uncharacterized protein', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11981591, end_position_on_the_genomic_accession=11984380, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34608, CurrentID=0, Status='live', Symbol='CG6686', Aliases='Dmel_ Dmel\\CG6686, SART1', description='CG6686 gene product from transcript CG6686-RB', other_designations='CG6686-PA|CG6686-PB|CG6686-PC', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11984327, end_position_on_the_genomic_accession=11988039, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34609, CurrentID=0, Status='live', Symbol='zuc', Aliases='Dmel_CG12314, CG12314, Dmel\\CG12314, ZUC, Zuc', description='zucchini', other_designations='CG12314-PA|zuc-PA', map_location='33B5-33F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11988735, end_position_on_the_genomic_accession=11989882, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34610, CurrentID=0, Status='live', Symbol='dgt2', Aliases='Dmel_CG16969, CG16969, Dgt2, Dmel\\CG16969', description='dim gamma-tubulin 2', other_designations='CG16969-PA|CG16969-PB|dgt2-PA|dgt2-PB', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11989936, end_position_on_the_genomic_accession=11990979, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34611, CurrentID=0, Status='live', Symbol='escl', Aliases='Dmel_CG5202, CG5202, CG5202-PA, Dmel\\CG5202, ESC, ESC-LIKE, ESC1, ESCL', description='CG5202 gene product from transcript CG5202-RA', other_designations='CG5202-PA|CG5202-PC|esc-like|escl-PA|escl-PC', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11990927, end_position_on_the_genomic_accession=11992848, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34613, CurrentID=0, Status='live', Symbol='CG12264', Aliases='Dmel_CG12264-PA, Dmel\\CG12264', description='CG12264 gene product from transcript CG12264-RA', other_designations='CG12264-PA|IscS homologue|uncharacterized protein', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11999654, end_position_on_the_genomic_accession=12001244, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34614, CurrentID=0, Status='live', Symbol='CG6746', Aliases='Dmel_ Dmel\\CG6746', description='CG6746 gene product from transcript CG6746-RA', other_designations='CG6746-PA', map_location='33B5-33B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12001234, end_position_on_the_genomic_accession=12002266, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34615, CurrentID=0, Status='live', Symbol='Rh5', Aliases='Dmel_CG5279, CG5279, DMELRH5, Dm Rh5, Dmel\\CG5279, RH5, Rh, rh5', description='Rhodopsin 5', other_designations='CG5279-PA|CG5279-PB|Rh5-PA|Rh5-PB|Rhodopsin5|rhodopsin|rhodopsin 5|rhodopsin-5', map_location='33B8-33B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12008153, end_position_on_the_genomic_accession=12009832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34616, CurrentID=0, Status='live', Symbol='CG6734', Aliases='Dmel_ Dmel\\CG6734, cg6734', description='CG6734 gene product from transcript CG6734-RA', other_designations='CG6734-PA', map_location='33B8-33B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12014304, end_position_on_the_genomic_accession=12020960, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34618, CurrentID=0, Status='live', Symbol='Tom70', Aliases='Dmel_CG6756, CG6756, Dmel\\CG6756, RE66761p, dtom70', description='Translocase of outer membrane 70', other_designations='CG6756-PA|CG6756-PB|CG6756-PC|Tom70-PA|Tom70-PB|Tom70-PC|translocase of outer membrane 70', map_location='33B9-33B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12035984, end_position_on_the_genomic_accession=12038511, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34619, CurrentID=0, Status='live', Symbol='CG6766', Aliases='Dmel_ Dmel\\CG6766, cg6766', description='CG6766 gene product from transcript CG6766-RA', other_designations='CG6766-PA', map_location='33B10-33B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12038631, end_position_on_the_genomic_accession=12040689, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34620, CurrentID=0, Status='live', Symbol='CG6785', Aliases='Dmel_ Dmel\\CG6785, cg6785', description='CG6785 gene product from transcript CG6785-RA', other_designations='CG6785-PA|CG6785-PB|CG6785-PC', map_location='33B10-33B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12041054, end_position_on_the_genomic_accession=12045200, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34621, CurrentID=0, Status='live', Symbol='CG6770', Aliases='Dmel_ CT21019, Dmel\\CG6770, anon-WO02059370.25, p8', description='CG6770 gene product from transcript CG6770-RA', other_designations='CG6770-PA', map_location='33B11-33B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12045343, end_position_on_the_genomic_accession=12046118, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34622, CurrentID=0, Status='live', Symbol='Plzf', Aliases='Dmel_CG6792, CG6792, Dmel\\CG6792, PLZF, cg6792, dPLZF', description='Promyelocytic leukemia zinc finger ortholog', other_designations='CG6792-PA|CG6792-PB|Plzf-PA|Plzf-PB|promyelocytic leukemia zinc finger ortholog', map_location='33B11-33B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12046711, end_position_on_the_genomic_accession=12048918, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34623, CurrentID=0, Status='live', Symbol='CG14945', Aliases='Dmel_ Dmel\\CG14945, cg14945', description='CG14945 gene product from transcript CG14945-RB', other_designations='CG14945-PA|CG14945-PB|uncharacterized protein', map_location='33B11-33B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12049418, end_position_on_the_genomic_accession=12052049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34624, CurrentID=0, Status='live', Symbol='JhI-21', Aliases='Dmel_CG12317, CG12317, Dmel\\CG12317, JHI-21, Jhl-21', description='CG12317 gene product from transcript CG12317-RE', other_designations='CG12317-PA|CG12317-PB|CG12317-PC|CG12317-PD|CG12317-PE|JhI-21-PA|JhI-21-PB|JhI-21-PC|JhI-21-PD|JhI-21-PE', map_location='33B11-33C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12051755, end_position_on_the_genomic_accession=12056042, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  34625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34625, CurrentID=0, Status='live', Symbol='RpL7-like', Aliases='Dmel_CG5317, CG5317, Dmel\\CG5317', description='Ribosomal protein L7-like', other_designations='CG5317-PA|CG5317-PB|CG5317-PC|RpL7-like-PA|RpL7-like-PB|RpL7-like-PC|ribosomal protein L7-like', map_location='33C1-33C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12056326, end_position_on_the_genomic_accession=12058251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34626, CurrentID=0, Status='live', Symbol='Rab3-GAP', Aliases='Dmel_CG7061, BcDNA:LD23336, CG7061, Dmel\\CG7061, rab3-GAP', description='Rab3 GTPase activating protein', other_designations='CG7061-PA|CG7061-PB|Rab3-GAP-PA|Rab3-GAP-PB|rab3-GAP', map_location='33C1-33C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12058304, end_position_on_the_genomic_accession=12063184, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34627, CurrentID=0, Status='live', Symbol='CG14946', Aliases='Dmel_ Dmel\\CG14946', description='CG14946 gene product from transcript CG14946-RD', other_designations='CG14946-PC|CG14946-PD|CG14946-PE|uncharacterized protein', map_location='33C2-33C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12063457, end_position_on_the_genomic_accession=12066775, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34628, CurrentID=0, Status='live', Symbol='CG14947', Aliases='Dmel_ Dmel\\CG14947', description='CG14947 gene product from transcript CG14947-RA', other_designations='CG14947-PA|uncharacterized protein', map_location='33C2-33C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12076100, end_position_on_the_genomic_accession=12077739, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34629, CurrentID=0, Status='live', Symbol='prd', Aliases='Dmel_CG6716, CG6716, Dmel\\CG6716, PRD, Prd, pr', description='paired', other_designations='CG6716-PA|CG6716-PB|prd-PA|prd-PB', map_location='2-45 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12082995, end_position_on_the_genomic_accession=12085827, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34630, CurrentID=0, Status='live', Symbol='Pex19', Aliases='Dmel_CG5325, BEST:GH03076, CG5325, DmelPex19, Dmel\\CG5325, PEX19, Pex-19', description='Peroxin 19', other_designations='CG5325-PA|CG5325-PC|Pex19-PA|Pex19-PC|peroxin 19|peroxisomal farnesylated protein', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12091822, end_position_on_the_genomic_accession=12095250, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34631, CurrentID=0, Status='live', Symbol='CG6712', Aliases='Dmel_ Dmel\\CG6712, ORF2', description='CG6712 gene product from transcript CG6712-RA', other_designations='CG6712-PA', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12095176, end_position_on_the_genomic_accession=12096511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34632, CurrentID=0, Status='live', Symbol='Mt2', Aliases='Dmel_CG10692, CG10692, DNMT, DNMT2, DmMT2, Dmel\\CG10692, Dnmt, Dnmt2, dDNMT, dDNMT2, dDnmt2, dnmt2, nmt2', description='Methyltransferase 2', other_designations='CG10692-PC|CG10692-PD|DNA methyltransferase 2|DNA(5-cytosine) methyltransferase|Mt2-PC|Mt2-PD|cytosine-5 methyltransferase|methyltransferase 2', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12096525, end_position_on_the_genomic_accession=12097687, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34633, CurrentID=0, Status='live', Symbol='Ced-12', Aliases='Dmel_CG5336, CG5336, Ced12, Dced-12, Dmel\\CG5336, ELMO, Elmo, ced-12, d dced-12, elmo', description='CG5336 gene product from transcript CG5336-RA', other_designations='CG5336-PA|Ced-12-PA|engulfment and cell motility', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12097838, end_position_on_the_genomic_accession=12100839, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34634, CurrentID=0, Status='live', Symbol='Patsas', Aliases='Dmel_CG6618, CG6618, Dmel\\CG6618', description='CG6618 gene product from transcript CG6618-RA', other_designations='CG6618-PA|Patsas-PA|patsas', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12100765, end_position_on_the_genomic_accession=12103285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34636, CurrentID=0, Status='live', Symbol='Rab6', Aliases='Dmel_CG6601, 136/6, AAF53168, CG6601, D Dm Rab6, DmRab6, Dmel\\CG6601, Drab6, O18334, dRab6, drab6, l(2)08323, rab6, wrt', description='CG6601 gene product from transcript CG6601-RB', other_designations='CG6601-PA|CG6601-PB|Rab protein 6|Rab-protein 6|Rab6-PA|Rab6-PB|warthog', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12107221, end_position_on_the_genomic_accession=12109116, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34637, CurrentID=0, Status='live', Symbol='Phae1', Aliases='Dmel_CG16996, CG16996, Dmel\\CG16996, SP134', description='Phaedra 1', other_designations='CG16996-PA|CG16996-PB|Phae1-PA|Phae1-PB|phaedra 1', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12109417, end_position_on_the_genomic_accession=12110664, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34638, CurrentID=0, Status='live', Symbol='Phae2', Aliases='Dmel_CG16997, CG16997, Dmel\\CG16997, SP147', description='Phaedra 2', other_designations='CG16997-PB|CG16997-PC|Phae2-PB|Phae2-PC|phaedra 2', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12111110, end_position_on_the_genomic_accession=12112256, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34639, CurrentID=0, Status='live', Symbol='CG17211', Aliases='Dmel_ Dmel\\CG17211', description='CG17211 gene product from transcript CG17211-RA', other_designations='CG17211-PA|uncharacterized protein', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12114360, end_position_on_the_genomic_accession=12129896, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34640, CurrentID=0, Status='live', Symbol='rho-6', Aliases='Dmel_CG17212, CG17212, Dmel\\CG17212, Rho-related [33C1], rho 6, rho6', description='rhomboid-6', other_designations='CG17212-PB|CG17212-PC|Rhomboid-6|rho-6-PB|rho-6-PC', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12130516, end_position_on_the_genomic_accession=12133411, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34641, CurrentID=0, Status='live', Symbol='Gr33a', Aliases='Dmel_CG17213, CG17213, Dmel\\CG17213, GR33a, Gr33C1', description='Gustatory receptor 33a', other_designations='CG17213-PA|Gr33a-PA|gustatory receptor 33C1|gustatory receptor 33a', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12138008, end_position_on_the_genomic_accession=12140392, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34642, CurrentID=0, Status='live', Symbol='CG31760', Aliases='Dmel_ BcDNA:RE52162, CG17214, CG17215, Dmel\\CG31760', description='CG31760 gene product from transcript CG31760-RC', other_designations='CG31760-PB|CG31760-PC|CG31760-PD|uncharacterized protein', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12142780, end_position_on_the_genomic_accession=12170508, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  34644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34644, CurrentID=0, Status='live', Symbol='CG17217', Aliases='Dmel_ Dmel\\CG17217', description='CG17217 gene product from transcript CG17217-RA', other_designations='CG17217-PA|uncharacterized protein', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12170550, end_position_on_the_genomic_accession=12171232, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34645, CurrentID=0, Status='live', Symbol='CG6583', Aliases='Dmel_ Dmel\\CG6583', description='CG6583 gene product from transcript CG6583-RA', other_designations='CG6583-PA', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12171204, end_position_on_the_genomic_accession=12172586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34646, CurrentID=0, Status='live', Symbol='crok', Aliases='Dmel_CG17218, BcDNA:RE65321, CG17218, Dmel\\CG17218', description='crooked', other_designations='CG17218-PA|crok-PA', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12172981, end_position_on_the_genomic_accession=12174091, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34647, CurrentID=0, Status='live', Symbol='atilla', Aliases='Dmel_CG6579, CG6579, Dmel\\CG6579', description='CG6579 gene product from transcript CG6579-RB', other_designations='CG6579-PB|CG6579-PC|atilla-PB|atilla-PC', map_location='33D2-33D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12175451, end_position_on_the_genomic_accession=12177886, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34648, CurrentID=0, Status='live', Symbol='aret', Aliases='Dmel_CG31762, ARET, Aret, BcDNA:GM15173, Bru, Bruno, CG31762, CG6319, Dm Bru, Dmel\\CG31762, NEST:bs01g04, NEST:bs34h11/bruno1, arrest/Bruno, bru, bruno, ms(2)01284, aret', description='arrest', other_designations='CG31762-PA|CG31762-PB|CG31762-PD|CG31762-PE|CG31762-PF|CG31762-PG|CG31762-PH|CG31762-PI|CG31762-PJ|CG31762-PK|aret-PA|aret-PB|aret-PD|aret-PE|aret-PF|aret-PG|aret-PH|aret-PI|aret-PJ|aret-PK|bruno', map_location='33D3-33D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12175004, end_position_on_the_genomic_accession=12313438, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  34650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34650, CurrentID=0, Status='live', Symbol='CG17010', Aliases='Dmel_ Dmel\\CG17010', description='CG17010 gene product from transcript CG17010-RA', other_designations='CG17010-PA|uncharacterized protein', map_location='33E1-33E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12356441, end_position_on_the_genomic_accession=12357906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34651, CurrentID=0, Status='live', Symbol='Pkd2', Aliases='Dmel_CG6504, AMO, Amo, CG6504, Dmel\\CG6504, Dmpkd2, PKD2, TRPP, amo, pkd2', description='Polycystic kidney disease 2 ortholog (H. sapiens)', other_designations='CG6504-PA|Pkd2-PA|almost there|polycystic kidney disease 2 ortholog|polycystic kidney disease gene-2|polycystin-2', map_location='33E3-33E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12398274, end_position_on_the_genomic_accession=12402023, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  34652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34652, CurrentID=0, Status='live', Symbol='vir-1', Aliases='Dmel_CG31764, BEST:GH02216, CG15493, CG31764, CG5453, Dmel\\CG31764, Vir-1, Vir1, vir1', description='virus-induced RNA 1', other_designations='CG31764-PA|CG31764-PB|CG31764-PC|CG31764-PD|CG31764-PE|CG31764-PG|CG31764-PH|vir-1-PA|vir-1-PB|vir-1-PC|vir-1-PD|vir-1-PE|vir-1-PG|vir-1-PH|virus induced RNA-1|virus-induced 1|virus-induced RNA1', map_location='33E3-33E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12403270, end_position_on_the_genomic_accession=12423439, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34654, CurrentID=0, Status='live', Symbol='CG6405', Aliases='Dmel_ Dmel\\CG6405', description='CG6405 gene product from transcript CG6405-RA', other_designations='CG6405-PA', map_location='33E4-33E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12424904, end_position_on_the_genomic_accession=12426440, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34655, CurrentID=0, Status='live', Symbol='CG5446', Aliases='Dmel_ BcDNA:RE36920, Dmel\\CG5446', description='CG5446 gene product from transcript CG5446-RC', other_designations='CG5446-PA|CG5446-PB|CG5446-PC', map_location='33E4-33E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12426631, end_position_on_the_genomic_accession=12427709, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34656, CurrentID=0, Status='live', Symbol='CG6388', Aliases='Dmel_ Dmel\\CG6388', description='CG6388 gene product from transcript CG6388-RA', other_designations='CG6388-PA', map_location='33E4-33E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12428028, end_position_on_the_genomic_accession=12430021, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740176, CurrentID=0, Status='live', Symbol='CG34340', Aliases='Dmel_ CG10017, CG2808, Dmel\\CG34340, Dmel_CG10017, Dmel_CG2808', description='CG34340 gene product from transcript CG34340-RF', other_designations='CG34340-PE|CG34340-PF|uncharacterized protein', map_location='24A5-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3662745, end_position_on_the_genomic_accession=3692114, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34658, CurrentID=0, Status='live', Symbol='Elf', Aliases='Dmel_CG6382, CG6382, DSUP35, Dmel\\CG6382, Dsup35, delf, dsup35, eRF3, elf', description='Ef1alpha-like factor', other_designations='CG6382-PA|CG6382-PB|CG6382-PC|CG6382-PD|Elf-PA|Elf-PB|Elf-PC|Elf-PD|eukaryotic release factor 3', map_location='33E4-33E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12434452, end_position_on_the_genomic_accession=12441279, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34660, CurrentID=0, Status='live', Symbol='Oatp33Ea', Aliases='Dmel_CG5427, CG5427, Dmel\\CG5427, oatp 33Ea', description='Organic anion transporting polypeptide 33Ea', other_designations='CG5427-PA|CG5427-PB|Oatp33Ea-PA|Oatp33Ea-PB|organic anion transporting polypeptide 33Ea', map_location='33E5-33E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12441401, end_position_on_the_genomic_accession=12445924, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34661, CurrentID=0, Status='live', Symbol='CG5421', Aliases='Dmel_ Dmel\\CG5421', description='CG5421 gene product from transcript CG5421-RB', other_designations='CG5421-PA|CG5421-PB', map_location='33E5-33E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12446080, end_position_on_the_genomic_accession=12448132, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34662, CurrentID=0, Status='live', Symbol='Oatp33Eb', Aliases='Dmel_CG6417, CG6417, Dmel\\CG6417, oatp 33Eb', description='Organic anion transporting polypeptide 33Eb', other_designations='CG6417-PA|Oatp33Eb-PA|organic anion transporting polypeptide 33Eb', map_location='33E5-33E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12448484, end_position_on_the_genomic_accession=12452047, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34663, CurrentID=0, Status='live', Symbol='CG5418', Aliases='Dmel_ Dmel\\CG5418', description='CG5418 gene product from transcript CG5418-RA', other_designations='CG5418-PA', map_location='33E5-33E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12451945, end_position_on_the_genomic_accession=12453242, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34665, CurrentID=0, Status='live', Symbol='bun', Aliases='Dmel_CG42281, 1550, Bun, BunA, CG31857, CG42281, CG5461, Dm_2L:60722, Dmel\\CG42281, Dmel_CG31857, Dmel_CG5461, EP(2)0488, TSC-22, Tsc22, bnc, l(2)00255, l(2)02687, l(2)05479, l(2)06475, l(2)k02903, l(2)rI043, shs', description='bunched', other_designations='CG42281-PA|CG42281-PB|CG42281-PC|CG42281-PD|CG42281-PE|CG42281-PF|CG42281-PG|CG42281-PH|CG42281-PO|CG42281-PP|bun-PA|bun-PB|bun-PC|bun-PD|bun-PE|bun-PF|bun-PG|bun-PH|bun-PO|bun-PP|short sighted|short-sighted|shortsighted', map_location='33E5-33E9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12455540, end_position_on_the_genomic_accession=12546611, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  34667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34667, CurrentID=0, Status='live', Symbol='Ref2', Aliases='Dmel_CG17031, CG17031, Dmel\\CG17031, REF2, cg17031, ref2', description='RNA and export factor binding protein 2', other_designations='CG17031-PA|CG17031-PB|CG17031-PC|Ref2-PA|Ref2-PB|Ref2-PC|ref2', map_location='33F1-33F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12610517, end_position_on_the_genomic_accession=12646356, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34669, CurrentID=0, Status='live', Symbol='nub', Aliases='Dmel_CG34395, CG15488, CG15489, CG34395, CG6246, Dmel\\CG34395, Dmel_CG15488, Dmel_CG15489, Dmel_CG6246, NUB, Nb, Nub, Oct-1, Oct1, PDM-1, PDM1, POU33F1, Pdm, Pdm-1, Pdm1, dOct1, dPOU-19, dPOU-19/pdm-1, dPOU19, nb, pdm, pdm-1, pdm1, twn', description='nubbin', other_designations='CG34395-PB|CG34395-PD|CG34395-PE|NUBBIN|POU domain protein 1|nub-PB|nub-PD|nub-PE|twain', map_location='2-47.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12587625, end_position_on_the_genomic_accession=12628135, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34671, CurrentID=0, Status='live', Symbol='CG15485', Aliases='Dmel_ Dmel\\CG15485', description='CG15485 gene product from transcript CG15485-RA', other_designations='CG15485-PA|CG15485-PB|uncharacterized protein', map_location='33F2-33F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12667248, end_position_on_the_genomic_accession=12669714, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34673, CurrentID=0, Status='live', Symbol='pdm2', Aliases='Dmel_CG12287, CG12287, CG15486, CG15487, Dmel\\CG12287, Oct-2, Oct2, Pdm, Pdm-2, Pdm2, dOct2, dPOU-28, dPOU28, dPOU28/pdm-2, dim, dpou28, miti, pdm, pdm-2', description='POU domain protein 2', other_designations='CG12287-PA|CG12287-PB|CG12287-PC|CG12287-PD|Miti-mere|POU domain prot 2|didymous|pdm2-PA|pdm2-PB|pdm2-PC|pdm2-PD', map_location='33F2-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12657165, end_position_on_the_genomic_accession=12686306, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34674, CurrentID=0, Status='live', Symbol='CG5525', Aliases='Dmel_ CCT4, Cct4, Dmel\\CG5525, Q9VK69_DROME, TCP-1, Tcp-1delta, Tcp1-delta', description='CG5525 gene product from transcript CG5525-RA', other_designations='CG5525-PA', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12690336, end_position_on_the_genomic_accession=12692934, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34675, CurrentID=0, Status='live', Symbol='CG6153', Aliases='Dmel_ 6153, Dmel\\CG6153', description='CG6153 gene product from transcript CG6153-RC', other_designations='CG6153-PA|CG6153-PC', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12692993, end_position_on_the_genomic_accession=12693939, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34676, CurrentID=0, Status='live', Symbol='CG5780', Aliases='Dmel_ Dmel\\CG5780, IFT43', description='CG5780 gene product from transcript CG5780-RB', other_designations='CG5780-PB', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12693872, end_position_on_the_genomic_accession=12694988, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34677, CurrentID=0, Status='live', Symbol='PICK1', Aliases='Dmel_CG6167, CG6167, Dmel\\CG6167', description='Protein interacting with C kinase 1', other_designations='CG6167-PA|CG6167-PB|CG6167-PD|CG6167-PE|CG6167-PF|PICK1 ortholog|PICK1-PA|PICK1-PB|PICK1-PD|PICK1-PE|PICK1-PF|protein interacting with C kinase 1', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12695033, end_position_on_the_genomic_accession=12698917, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34678, CurrentID=0, Status='live', Symbol='CG5781', Aliases='Dmel_ Dmel\\CG5781', description='CG5781 gene product from transcript CG5781-RA', other_designations='CG5781-PA', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12696426, end_position_on_the_genomic_accession=12697323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34679, CurrentID=0, Status='live', Symbol='CG17036', Aliases='Dmel_ Dmel\\CG17036', description='CG17036 gene product from transcript CG17036-RA', other_designations='CG17036-PA|CG17036-PB|uncharacterized protein', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12698918, end_position_on_the_genomic_accession=12700746, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34680, CurrentID=0, Status='live', Symbol='CG5776', Aliases='Dmel_ Dmel\\CG5776', description='CG5776 gene product from transcript CG5776-RB', other_designations='CG5776-PA|CG5776-PB', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12700949, end_position_on_the_genomic_accession=12704451, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34681, CurrentID=0, Status='live', Symbol='spict', Aliases='Dmel_CG12292, CG12292, Dmel\\CG12292, Spict, spg6', description='spichthyin', other_designations='CG12292-PA|CG12292-PB|spict-PA|spict-PB', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12704725, end_position_on_the_genomic_accession=12706683, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34682, CurrentID=0, Status='live', Symbol='CG15484', Aliases='Dmel_ Dmel\\CG15484', description='CG15484 gene product from transcript CG15484-RA', other_designations='CG15484-PA|uncharacterized protein', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12706799, end_position_on_the_genomic_accession=12707711, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34683, CurrentID=0, Status='live', Symbol='CG6180', Aliases='Dmel_ Dmel\\CG6180', description='CG6180 gene product from transcript CG6180-RA', other_designations='CG6180-PA', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12707661, end_position_on_the_genomic_accession=12708680, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34684, CurrentID=0, Status='live', Symbol='CG5787', Aliases='Dmel_ Dmel\\CG5787', description='CG5787 gene product from transcript CG5787-RB', other_designations='CG5787-PA|CG5787-PB', map_location='33F3-33F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12708792, end_position_on_the_genomic_accession=12712952, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34685, CurrentID=0, Status='live', Symbol='Pih1D1', Aliases='Dmel_CG5792, CG5792, Dmel\\CG5792', description='CG5792 gene product from transcript CG5792-RA', other_designations='CG5792-PA|CG5792-PD|Pih1D1-PA|Pih1D1-PD', map_location='33F3-33F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12713276, end_position_on_the_genomic_accession=12721843, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34686, CurrentID=0, Status='live', Symbol='MRP', Aliases='Dmel_CG6214, CG6214, Dmel\\CG62141, Mrp, dMRP, dMRP/CG6214, dMRP1, lincRNA.123, MRP', description='Multidrug-Resistance like Protein 1', other_designations='CG6214-PA|CG6214-PB|CG6214-PC|CG6214-PD|CG6214-PE|CG6214-PF|CG6214-PG|CG6214-PH|CG6214-PI|CG6214-PJ|CG6214-PK|CG6214-PL|CG6214-PM|CG6214-PN|CG6214-PO|CG6214-PP|CG6214-PQ|CG6214-PR|MRP-PA|MRP-PB|MRP-PC|MRP-PD|MRP-PE|MRP-PF|MRP-PG|MRP-PH|MRP-PI|MRP-PJ|MRP-PK|MRP-PL|MRP-PM|MRP-PN|MRP-PO|MRP-PP|MRP-PQ|MRP-PR|Multidrug-Resistance like protein 1|Multidrug-resistance protein|multidrug resistance like protein|multidrug resistance-associated protein|multidrug resistance-like protein 1', map_location='33F3-33F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12719045, end_position_on_the_genomic_accession=12766546, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  34687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34687, CurrentID=0, Status='live', Symbol='CG15483', Aliases='Dmel_ Dmel\\CG15483, dibeta3GnT3', description='CG15483 gene product from transcript CG15483-RA', other_designations='CG15483-PA|uncharacterized protein', map_location='33F5-33F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12760346, end_position_on_the_genomic_accession=12761840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34688, CurrentID=0, Status='live', Symbol='kek1', Aliases='Dmel_CG12283, BEST:GM02380, CG12283, CT18186, Dmel\\CG12283, Dmkek1, GM02380, KEK1, Kek, Kek-1, Kek1, NB1, SD01674, kek, kek-1, kekon, l(2)01433, n(2)k07332, uns', description='kekkon-1', other_designations='CG12283-PB|CG12283-PC|Kekkon-1|Kekkon1|Kekon-1|kek1-PB|kek1-PC|kekkon|kekkon 1|kekkonI|unstrung', map_location='34A1-34A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12817000, end_position_on_the_genomic_accession=12822787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34689, CurrentID=0, Status='live', Symbol='ACXC', Aliases='Dmel_CG5983, AC, Acxc, CG5983, D Dmel\\CG5983', description='CG5983 gene product from transcript CG5983-RB', other_designations='ACXC-PA|ACXC-PB|CG5983-PA|CG5983-PB', map_location='34A2-34A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12912530, end_position_on_the_genomic_accession=12916774, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  34693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34693, CurrentID=0, Status='live', Symbol='CG16800', Aliases='Dmel_ Dmel\\CG16800', description='CG16800 gene product from transcript CG16800-RB', other_designations='CG16800-PB|uncharacterized protein', map_location='34A3-34A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12966590, end_position_on_the_genomic_accession=12967810, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34695, CurrentID=0, Status='live', Symbol='Vha68-3', Aliases='Dmel_CG5075, CG5075, Dmel\\CG5075, Vha68, vha67-3, vha68-3', description='Vacuolar H[+] ATPase 68kD subunit 3', other_designations='CG5075-PA|Vha68-3-PA|vacuolar H[+] ATPase 68kD subunit 3|vacuolar H[+] ATPase subunit 68-3', map_location='34A3-34A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12978786, end_position_on_the_genomic_accession=12981293, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34696, CurrentID=0, Status='live', Symbol='CG12404', Aliases='Dmel_ Dmel\\CG12404', description='CG12404 gene product from transcript CG12404-RA', other_designations='CG12404-PA|uncharacterized protein', map_location='34A4-34A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12993757, end_position_on_the_genomic_accession=13000635, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34699, CurrentID=0, Status='live', Symbol='CG9934', Aliases='Dmel_ CG 9934, Dmel\\CG9934, E4 B, UBE4B, UFD2b, dUfd2, ufd2', description='CG9934 gene product from transcript CG9934-RA', other_designations='CG9934-PA|CG9934-PB|NM 135753.2', map_location='34A4-34A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13013276, end_position_on_the_genomic_accession=13019434, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34700, CurrentID=0, Status='live', Symbol='A16', Aliases='Dmel_CG9933, CG9933, Dmel\\CG9933', description='CG9933 gene product from transcript CG9933-RA', other_designations='A16-PA|CG9933-PA', map_location='34A4-34A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13019575, end_position_on_the_genomic_accession=13021312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34701, CurrentID=0, Status='live', Symbol='CG9932', Aliases='Dmel_ Dmel\\CG9932, EP2317, Ef4A, elF-4A', description='CG9932 gene product from transcript CG9932-RE', other_designations='CG9932-PA|CG9932-PC|CG9932-PD|CG9932-PE|elongation factor 4a', map_location='34A5-34A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13021868, end_position_on_the_genomic_accession=13060140, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34702, CurrentID=0, Status='live', Symbol='CG5142', Aliases='Dmel_ Dmel\\CG5142, TTC30', description='CG5142 gene product from transcript CG5142-RD', other_designations='CG5142-PC|CG5142-PD', map_location='34A5-34A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13106807, end_position_on_the_genomic_accession=13109468, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34703, CurrentID=0, Status='live', Symbol='CG5122', Aliases='Dmel_ Dmel\\CG5122', description='CG5122 gene product from transcript CG5122-RA', other_designations='CG5122-PA', map_location='34A5-34A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13130591, end_position_on_the_genomic_accession=13134804, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34704, CurrentID=0, Status='live', Symbol='CG9928', Aliases='Dmel_ BcDNA:RH70154, CT27834, Dmel\\CG9928', description='CG9928 gene product from transcript CG9928-RB', other_designations='CG9928-PA|CG9928-PB', map_location='34A6-34A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13142445, end_position_on_the_genomic_accession=13142903, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34705, CurrentID=0, Status='live', Symbol='Pk34A', Aliases='Dmel_CG5182, CG5182, Dmel\\CG5182, GSK3a', description='CG5182 gene product from transcript CG5182-RA', other_designations='CG5182-PA|Pk34A-PA', map_location='34A7-34A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13158810, end_position_on_the_genomic_accession=13160173, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34706, CurrentID=0, Status='live', Symbol='CG5204', Aliases='Dmel_ Dmel\\CG5204, anon-WO0118547.212', description='CG5204 gene product from transcript CG5204-RA', other_designations='CG5204-PA', map_location='34A7-34A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13160527, end_position_on_the_genomic_accession=13163153, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34707, CurrentID=0, Status='live', Symbol='DnaJ-H', Aliases='Dmel_CG9828, CG9828, DNA J-H, DNAJ-H, Dmel\\CG9828, NEST:bs08d07, dDnaJ-H, dHdj2, dnaJ-H', description='DnaJ homolog', other_designations='CG9828-PA|CG9828-PB|CG9828-PC|DnaJ-H-PA|DnaJ-H-PB|DnaJ-H-PC', map_location='34A7-34A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13163209, end_position_on_the_genomic_accession=13166072, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34708, CurrentID=0, Status='live', Symbol='Sirt1', Aliases='Dmel_CG5216, BEST:LD38188, CG5216, D.mel1, DSir2, Dmel\\CG5216, SIR2, Sir2, Sir2L, dSIR2, dSir2, dmSRT406, dsir2, l(2)05326, sir2', description='Sirtuin 1', other_designations='CG5216-PA|Sir2|Sirt1-PA|silent information regulator 2|sirtuin 1', map_location='34A7-34A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13165564, end_position_on_the_genomic_accession=13169551, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34709, CurrentID=0, Status='live', Symbol='Sfmbt', Aliases='Dmel_CG16975, CG16975, Dmel\\CG16975, SFMBT, dSFMBT, dSfmbt', description='Scm-related gene containing four mbt domains', other_designations='CG16975-PA|CG16975-PB|CG16975-PC|CG16975-PD|Sfmbt-PA|Sfmbt-PB|Sfmbt-PC|Sfmbt-PD', map_location='34A7-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13169649, end_position_on_the_genomic_accession=13176785, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34710, CurrentID=0, Status='live', Symbol='CG5439', Aliases='Dmel_ Dmel\\CG5439', description='CG5439 gene product from transcript CG5439-RA', other_designations='CG5439-PA', map_location='34A8-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13177150, end_position_on_the_genomic_accession=13178978, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34711, CurrentID=0, Status='live', Symbol='CG5287', Aliases='Dmel_ Dmel\\CG5287, dAlg7', description='CG5287 gene product from transcript CG5287-RA', other_designations='CG5287-PA', map_location='34A8-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13179247, end_position_on_the_genomic_accession=13183495, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34712, CurrentID=0, Status='live', Symbol='CG5458', Aliases='Dmel_ Dmel\\CG5458', description='CG5458 gene product from transcript CG5458-RA', other_designations='CG5458-PA', map_location='34A8-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13183450, end_position_on_the_genomic_accession=13184875, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34713, CurrentID=0, Status='live', Symbol='CG16974', Aliases='Dmel_ CT35580, Dmel\\CG16974, Dromel_CG16974_FBtr0080489_mORF', description='CG16974 gene product from transcript CG16974-RB', other_designations='CG16974-PA|CG16974-PB|kekkon-like|uncharacterized protein', map_location='34A8-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13183849, end_position_on_the_genomic_accession=13191979, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34714, CurrentID=0, Status='live', Symbol='Edem2', Aliases='Dmel_CG5682, CG5682, Dmel\\CG5682, EDEM2, edem2', description='ER degradation enhancer, mannosidase alpha-like 2', other_designations='CG5682-PA|CG5682-PB|Edem2-PA|Edem2-PB', map_location='34A8-34A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13192201, end_position_on_the_genomic_accession=13197281, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34715, CurrentID=0, Status='live', Symbol='CG16972', Aliases='Dmel_ Dmel\\CG16972, anon-WO0140519.52', description='CG16972 gene product from transcript CG16972-RA', other_designations='CG16972-PA|uncharacterized protein', map_location='34A9-34A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13197102, end_position_on_the_genomic_accession=13203295, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34716, CurrentID=0, Status='live', Symbol='Pect', Aliases='Dmel_CG5547, CG5547, Dmel\\CG5547, PECT, pect', description='Phosphoethanolamine cytidylyltransferase', other_designations='CG5547-PB|CG5547-PC|CG5547-PD|CG5547-PE|Pect-PB|Pect-PC|Pect-PD|Pect-PE|phosphoethanolamine cytidylyltransferase', map_location='34A9-34A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13203762, end_position_on_the_genomic_accession=13210828, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740629, CurrentID=0, Status='live', Symbol='CG43795', Aliases='Dmel_ CG13552, CG13553, CG13556, CG18678, CG18679, CG34372, CG5357, Dmel\\CG43795, Dmel_CG13552, Dmel_CG13556, Dmel_CG18678, Dmel_CG18679, Dmel_CG34372, Dmel_CG5357', description='CG43795 gene product from transcript CG43795-RA', other_designations='CG43795-PA|uncharacterized protein', map_location='59E2-59E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23444137, end_position_on_the_genomic_accession=23491997, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  34718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34718, CurrentID=0, Status='live', Symbol='kek4', Aliases='Dmel_CG9431, CG9431, CT26742, Dmel\\CG9431', description='kekkon4', other_designations='CG9431-PA|CG9431-PB|CG9431-PC|kek4-PA|kek4-PB|kek4-PC|kekkon-like', map_location='34A9-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13212335, end_position_on_the_genomic_accession=13215350, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34719, CurrentID=0, Status='live', Symbol='CG9426', Aliases='Dmel_ Dmel\\CG9426, Ipp', description='CG9426 gene product from transcript CG9426-RA', other_designations='CG9426-PA', map_location='34A10-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13216290, end_position_on_the_genomic_accession=13219923, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34720, CurrentID=0, Status='live', Symbol='CG5705', Aliases='Dmel_ Dmel\\CG5705', description='CG5705 gene product from transcript CG5705-RA', other_designations='CG5705-PA', map_location='34A10-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13219979, end_position_on_the_genomic_accession=13221398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34721, CurrentID=0, Status='live', Symbol='Ski6', Aliases='Dmel_CG15481, CG15481, Dmel\\CG15481, Rrp41, dRrp41, d', description='CG15481 gene product from transcript CG15481-RB', other_designations='CG15481-PA|CG15481-PB|Ski6-PA|Ski6-PB', map_location='34A10-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13221397, end_position_on_the_genomic_accession=13222429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34722, CurrentID=0, Status='live', Symbol='CG16812', Aliases='Dmel_ Dmel\\CG16812', description='CG16812 gene product from transcript CG16812-RA', other_designations='CG16812-PA|uncharacterized protein', map_location='34A10-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13222587, end_position_on_the_genomic_accession=13224545, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34723, CurrentID=0, Status='live', Symbol='CG15480', Aliases='Dmel_ Dmel\\CG15480', description='CG15480 gene product from transcript CG15480-RA', other_designations='CG15480-PA|uncharacterized protein', map_location='34A10-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13224620, end_position_on_the_genomic_accession=13225275, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34724, CurrentID=0, Status='live', Symbol='CG16813', Aliases='Dmel_ DM26, Dmel\\CG16813', description='CG16813 gene product from transcript CG16813-RB', other_designations='CG16813-PA|CG16813-PB|uncharacterized protein', map_location='34A10-34A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13227529, end_position_on_the_genomic_accession=13228521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34725, CurrentID=0, Status='live', Symbol='CG16815', Aliases='Dmel_ DM27, Dmel\\CG16815', description='CG16815 gene product from transcript CG16815-RB', other_designations='CG16815-PB|uncharacterized protein', map_location='34A11-34A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13229832, end_position_on_the_genomic_accession=13230549, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34726, CurrentID=0, Status='live', Symbol='Prosalpha6T', Aliases='Dmel_CG5648, 5648, Alpha6T, CG5648, Dmel\\CG5648, alpha6T, prosalpha6T', description='Proteasome alpha6 subunit, Testis-specific', other_designations='CG5648-PA|Prosalpha6T-PA|proteasome alpha6 subunit, Testis-specific|proteasome alpha6T subunit', map_location='34A11-34A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13232039, end_position_on_the_genomic_accession=13233058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34727, CurrentID=0, Status='live', Symbol='Mabi', Aliases='Dmel_CG15479, CG15479, DM2, Dmel\\CG15479', description='Mabiki', other_designations='CG15479-PA|Mabi-PA|mabiki', map_location='34A11-34A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13233798, end_position_on_the_genomic_accession=13234660, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34728, CurrentID=0, Status='live', Symbol='CG5867', Aliases='Dmel_ BcDNA:GH05536, CT18390, Dmel\\CG5867', description='CG5867 gene product from transcript CG5867-RA', other_designations='CG5867-PA', map_location='34A11-34A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13236494, end_position_on_the_genomic_accession=13239297, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34729, CurrentID=0, Status='live', Symbol='CG5945', Aliases='Dmel_ CT18564, Dmel\\CG5945, cg5945', description='CG5945 gene product from transcript CG5945-RB', other_designations='CG5945-PA|CG5945-PB', map_location='34A11-34B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13239989, end_position_on_the_genomic_accession=13241773, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34730, CurrentID=0, Status='live', Symbol='CG16820', Aliases='Dmel_ Dmel\\CG16820, cg16820', description='CG16820 gene product from transcript CG16820-RA', other_designations='CG16820-PA|uncharacterized protein', map_location='34B1-34B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13246232, end_position_on_the_genomic_accession=13250189, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34731, CurrentID=0, Status='live', Symbol='CG31728', Aliases='Dmel_ CG16821, Dmel\\CG31728, SP36, c-SP36', description='CG31728 gene product from transcript CG31728-RC', other_designations='CG31728-PC|uncharacterized protein', map_location='34B2-34B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13251509, end_position_on_the_genomic_accession=13257426, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34732, CurrentID=0, Status='live', Symbol='CG44085', Aliases='Dmel_ CG6043, CG6108, Dmel\\CG44085, Dmel_CG6043, Dmel_CG6108', description='CG44085 gene product from transcript CG44085-RF', other_designations='CG44085-PC|CG44085-PD|CG44085-PF|CG44085-PG|CG44085-PH|CG44085-PI|CG44085-PJ|CG44085-PK|CG44085-PL|CG44085-PM|CG44085-PN|CG44085-PO|CG44085-PP|uncharacterized protein', map_location='34B3-34B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13261008, end_position_on_the_genomic_accession=13284043, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  34734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34734, CurrentID=0, Status='live', Symbol='Drep4', Aliases='Dmel_CG9414, CAD, CG9414, DREP-4, Dmel\\CG9414, Drep-4, Rep4, dCAD, drep4', description='DNA fragmentation factor-related protein 4', other_designations='CG9414-PA|Caspase-activated DNase|Drep4-PA|Rep4', map_location='34B4-34B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13283912, end_position_on_the_genomic_accession=13286462, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34735, CurrentID=0, Status='live', Symbol='Uvrag', Aliases='Dmel_CG6116, CG6116, Dmel\\CG6116, UVRAG, Vps38, uvrag', description='UV-resistance associated gene', other_designations='CG6116-PA|CG6116-PB|UV radiation resistance associated|UV radiation resistance associated gene|Uvrag-PA|Uvrag-PB', map_location='34B4-34B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13286777, end_position_on_the_genomic_accession=13289556, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34736, CurrentID=0, Status='live', Symbol='CG31729', Aliases='Dmel_ CG16823, CG6263, DmCG31729, Dmel\\CG31729', description='CG31729 gene product from transcript CG31729-RB', other_designations='CG31729-PA|CG31729-PB|CG31729-PC|CG31729-PD|uncharacterized protein', map_location='34B4-34B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13289997, end_position_on_the_genomic_accession=13301668, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34738, CurrentID=0, Status='live', Symbol='CG43778', Aliases='Dmel_ CG15639, CG16970, Dmel\\CG43778, Dmel_CG16970, FBgn 32504', description='CG43778 gene product from transcript CG43778-RD', other_designations='CG43778-PA|CG43778-PC|CG43778-PD|uncharacterized protein', map_location='34B5-34B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13301814, end_position_on_the_genomic_accession=13331121, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34739, CurrentID=0, Status='live', Symbol='CG16825', Aliases='Dmel_ BcDNA:AT09390, Dmel\\CG16825', description='CG16825 gene product from transcript CG16825-RB', other_designations='CG16825-PA|CG16825-PB|uncharacterized protein', map_location='34B5-34B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13304788, end_position_on_the_genomic_accession=13307807, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34741, CurrentID=0, Status='live', Symbol='CG16826', Aliases='Dmel_ Dmel\\CG16826', description='CG16826 gene product from transcript CG16826-RB', other_designations='CG16826-PA|CG16826-PB|uncharacterized protein', map_location='34B6-34B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13347773, end_position_on_the_genomic_accession=13349020, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34742, CurrentID=0, Status='live', Symbol='CG9395', Aliases='Dmel_ Dmel\\CG9395', description='CG9395 gene product from transcript CG9395-RD', other_designations='CG9395-PC|CG9395-PD', map_location='34B6-34B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13349823, end_position_on_the_genomic_accession=13355073, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34743, CurrentID=0, Status='live', Symbol='CG9377', Aliases='Dmel_ Dmel\\CG9377, SPH166, c-SPH166', description='CG9377 gene product from transcript CG9377-RA', other_designations='CG9377-PA', map_location='34B7-34B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13359901, end_position_on_the_genomic_accession=13361565, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34745, CurrentID=0, Status='live', Symbol='CG6523', Aliases='Dmel_ Dmel\\CG6523, Grx3, Grx4', description='CG6523 gene product from transcript CG6523-RA', other_designations='CG6523-PA', map_location='34B7-34B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13366447, end_position_on_the_genomic_accession=13367410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34747, CurrentID=0, Status='live', Symbol='ND-B22', Aliases='Dmel_CG9306, CG9306, Dmel\\CG9306, NI2M', description='NADH dehydrogenase (ubiquinone) B22 subunit', other_designations='CG9306-PA|CG9306-PB|NADH ubiquinone oxidoreductase B22|NADH-ubiquinone oxidoreductase B22 subunit|NADH:ubiquinone oxidoreductase B22 subunit|ND-B22-PA|ND-B22-PB', map_location='34B8-34B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13370103, end_position_on_the_genomic_accession=13371110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34748, CurrentID=0, Status='live', Symbol='CG9305', Aliases='Dmel_ Bdp1, Dmel\\CG9305', description='CG9305 gene product from transcript CG9305-RB', other_designations='CG9305-PA|CG9305-PB', map_location='34B8-34B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13371427, end_position_on_the_genomic_accession=13373890, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34749, CurrentID=0, Status='live', Symbol='CG6565', Aliases='Dmel_ Dmel\\CG6565', description='CG6565 gene product from transcript CG6565-RB', other_designations='CG6565-PA|CG6565-PB|CG6565-PC', map_location='34B8-34B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13374019, end_position_on_the_genomic_accession=13376404, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34750, CurrentID=0, Status='live', Symbol='CG9302', Aliases='Dmel_ Dmel\\CG9302, anon-WO0118547.80', description='CG9302 gene product from transcript CG9302-RA', other_designations='CG9302-PA', map_location='34B8-34B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13376455, end_position_on_the_genomic_accession=13378538, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34751, CurrentID=0, Status='live', Symbol='loqs', Aliases='Dmel_CG6866, CG6866, Dmel\\CG6866, LOQS, Loqs, R3D1, R3D1-L, R3D1-S, TRBP, cg6866, dRax, loq, r3d1', description='loquacious', other_designations='CG6866-PA|CG6866-PB|CG6866-PC|CG6866-PD|CG6866-PE|CG6866-PF|loqs-PA|loqs-PB|loqs-PC|loqs-PD|loqs-PE|loqs-PF', map_location='34B9-34B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13382652, end_position_on_the_genomic_accession=13385671, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34752, CurrentID=0, Status='live', Symbol='CG9293', Aliases='Dmel_ Dmel\\CG9293', description='CG9293 gene product from transcript CG9293-RB', other_designations='CG9293-PA|CG9293-PB', map_location='34B9-34B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13385848, end_position_on_the_genomic_accession=13387312, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34753, CurrentID=0, Status='live', Symbol='CG7099', Aliases='Dmel_ Dmel\\CG7099, FBgn 32517, dTFIIIC220', description='CG7099 gene product from transcript CG7099-RA', other_designations='CG7099-PA', map_location='34B9-34B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13387571, end_position_on_the_genomic_accession=13393959, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34754, CurrentID=0, Status='live', Symbol='RpL24', Aliases='Dmel_CG9282, CG9282, Dmel\\CG9282, L24, RL24_DROME, Rp L24, Rpl24, anon-EST:Posey242, anon-EST:Posey37', description='Ribosomal protein L24', other_designations='CG9282-PA|CG9282-PB|CG9282-PC|RpL24-PA|RpL24-PB|RpL24-PC|ribosomal protein L24', map_location='34B10-34B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13393687, end_position_on_the_genomic_accession=13395131, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34755, CurrentID=0, Status='live', Symbol='CG16957', Aliases='Dmel_ Dmel\\CG16957', description='CG16957 gene product from transcript CG16957-RA', other_designations='CG16957-PA|CG16957-PB|uncharacterized protein', map_location='34B10-34B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13395360, end_position_on_the_genomic_accession=13396184, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34756, CurrentID=0, Status='live', Symbol='CG10859', Aliases='Dmel_ Dmel\\CG10859', description='CG10859 gene product from transcript CG10859-RA', other_designations='CG10859-PA|CG10859-PB|uncharacterized protein', map_location='34B10-34B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13396392, end_position_on_the_genomic_accession=13398664, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34757, CurrentID=0, Status='live', Symbol='CG7110', Aliases='Dmel_ Dmel\\CG7110', description='CG7110 gene product from transcript CG7110-RB', other_designations='CG7110-PB|CG7110-PC', map_location='34B10-34B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13398991, end_position_on_the_genomic_accession=13411047, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34758, CurrentID=0, Status='live', Symbol='Vm34Ca', Aliases='Dmel_CG9271, CG9271, Dmel\\CG9271, VM26A.1, VM34C, VM34C.1, Vm34C, Vm34Cc', description='Vitelline membrane 34Ca', other_designations='CG9271-PA|Vm34Ca-PA|vitelline membrane 34Ca', map_location='34B11-34B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13411141, end_position_on_the_genomic_accession=13411631, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34759, CurrentID=0, Status='live', Symbol='CG16848', Aliases='Dmel_ Dmel\\CG16848, FBgn0032522', description='CG16848 gene product from transcript CG16848-RA', other_designations='CG16848-PA|uncharacterized protein', map_location='34B11-34B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13412519, end_position_on_the_genomic_accession=13413716, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34760, CurrentID=0, Status='live', Symbol='CG16956', Aliases='Dmel_ Dmel\\CG16956', description='CG16956 gene product from transcript CG16956-RA', other_designations='CG16956-PA|uncharacterized protein', map_location='34B11-34B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13413783, end_position_on_the_genomic_accession=13414575, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34761, CurrentID=0, Status='live', Symbol='Tehao', Aliases='Dmel_CG7121, CG7121, CT22017, Dmel\\CG7121, Tho, Tl-5, Toll 5, Toll-5, dTLR5, dToll5, tehao, toll', description='CG7121 gene product from transcript CG7121-RB', other_designations='CG7121-PA|CG7121-PB|Tehao-PA|Tehao-PB|Toll-5|tehao|toll receptor', map_location='34C1-34C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13435622, end_position_on_the_genomic_accession=13439930, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34762, CurrentID=0, Status='live', Symbol='CG9267', Aliases='Dmel_ Dmel\\CG9267, Q9VJX8', description='CG9267 gene product from transcript CG9267-RA', other_designations='CG9267-PA', map_location='34C1-34C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13444103, end_position_on_the_genomic_accession=13445651, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34763, CurrentID=0, Status='live', Symbol='Hsp60D', Aliases='Dmel_CG16954, CG16954, Dmel\\CG16954', description='CG16954 gene product from transcript CG16954-RA', other_designations='CG16954-PA|CG16954-PB|Hsp60D-PA|Hsp60D-PB', map_location='34C1-34C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13447538, end_position_on_the_genomic_accession=13449497, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34764, CurrentID=0, Status='live', Symbol='CG42810', Aliases='Dmel_ CG16849, Dmel\\CG42810, Dmel_CG16849', description='CG42810 gene product from transcript CG42810-RA', other_designations='CG42810-PA|uncharacterized protein', map_location='34C1-34C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13459681, end_position_on_the_genomic_accession=13460183, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34767, CurrentID=0, Status='live', Symbol='B4', Aliases='Dmel_CG9239/Susi, BG:DS07660.4, CG9239, CG9260, Dmel\\CG9239, Susi, n(2)05337', description='CG9239 gene product from transcript CG9239-RC', other_designations='B4-PA|B4-PB|B4-PC|CG9239-PA|CG9239-PB|CG9239-PC|suppressor of signaling by insulin', map_location='34C3-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13500624, end_position_on_the_genomic_accession=13549328, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34768, CurrentID=0, Status='live', Symbol='CG15638', Aliases='Dmel_ BG:BACR48E02.1, Dmel\\CG15638', description='CG15638 gene product from transcript CG15638-RA', other_designations='CG15638-PA|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13535291, end_position_on_the_genomic_accession=13535778, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34769, CurrentID=0, Status='live', Symbol='CG16852', Aliases='Dmel_ BG:BACR48E02.2, BcDNA:AT28681, Dmel\\CG16852', description='CG16852 gene product from transcript CG16852-RA', other_designations='CG16852-PA|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13540751, end_position_on_the_genomic_accession=13541646, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34770, CurrentID=0, Status='live', Symbol='CG9263', Aliases='Dmel_ Dmel\\CG9263', description='CG9263 gene product from transcript CG9263-RA', other_designations='CG9263-PA', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13541638, end_position_on_the_genomic_accession=13542576, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34771, CurrentID=0, Status='live', Symbol='CG16853', Aliases='Dmel_ BG:BACR48E02.3, Dmel\\CG16853', description='CG16853 gene product from transcript CG16853-RA', other_designations='CG16853-PA|CG16853-PB|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13542268, end_position_on_the_genomic_accession=13543633, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34772, CurrentID=0, Status='live', Symbol='kuz', Aliases='Dmel_CG7147, 11410, 34Da, ADAM10, BG:DS07660.3, CG7147, CT22079, Dmel\\CG7147, GS11410, Kuz, br38, l(2)03782, l(2)34Da, l(2)br38, l(2)c00136, l(2)k01403, l34Da, soy nut', description='kuzbanian', other_designations='CG7147-PA|CG7147-PB|CG7147-PC|CG7147-PF|CG7147-PG|kuz-PA|kuz-PB|kuz-PC|kuz-PF|kuz-PG|kuzbaninan|lethal (2) c00136', map_location='34C4-34C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13550139, end_position_on_the_genomic_accession=13639411, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  34773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34773, CurrentID=0, Status='live', Symbol='CG9254', Aliases='Dmel_ BG:DS07660.1, DS07660.1, Dmel\\CG9254', description='CG9254 gene product from transcript CG9254-RA', other_designations='CG9254-PA', map_location='34C5-34C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13595503, end_position_on_the_genomic_accession=13597186, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34774, CurrentID=0, Status='live', Symbol='ics', Aliases='Dmel_CG9031, BG:BACR48E02.4, CG9031, CT25926, Dmel\\CG9031, RSU-1, RSU1, Rsu-1', description='icarus', other_designations='CG9031-PA|CG9031-PB|CG9031-PC|Ras suppressor-1|ics-PA|ics-PB|ics-PC', map_location='34C6-34C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13639447, end_position_on_the_genomic_accession=13642602, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34775, CurrentID=0, Status='live', Symbol='CG18507', Aliases='Dmel_ BG:DS01368.1, DS01368.1, Dmel\\CG18507, bg:ds01368', description='CG18507 gene product from transcript CG18507-RC', other_designations='CG18507-PA|CG18507-PB|CG18507-PC|CG18507-PD|CG18507-PE|uncharacterized protein', map_location='34C6-34C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13665525, end_position_on_the_genomic_accession=13672967, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  34776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34776, CurrentID=0, Status='live', Symbol='CG7311', Aliases='Dmel_ BG:DS08249.2, Dmel\\CG7311, GPO, GPO-2, Gpo, alpha-GPO', description='CG7311 gene product from transcript CG7311-RD', other_designations='CG7311-PD|CG7311-PE', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13708215, end_position_on_the_genomic_accession=13710583, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34777, CurrentID=0, Status='live', Symbol='CG9014', Aliases='Dmel_ BG:DS08249.3, DS08249.3, Dmel\\CG9014', description='CG9014 gene product from transcript CG9014-RA', other_designations='CG9014-PA|CG9014-PB', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13722161, end_position_on_the_genomic_accession=13723642, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34778, CurrentID=0, Status='live', Symbol='TM9SF4', Aliases='Dmel_CG7364, BG:DS00797.1, CG16861, CG7364, Dmel\\CG7364, n(2)k07245', description='Transmembrane 9 superfamily protein member 4', other_designations='CG7364-PA|CG7364-PB|TM9SF4-PA|TM9SF4-PB|transmembrane 9 superfamily protein member 4', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13746952, end_position_on_the_genomic_accession=13773905, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34779, CurrentID=0, Status='live', Symbol='CG9008', Aliases='Dmel_ BG:DS00797.2, Dmel\\CG9008', description='CG9008 gene product from transcript CG9008-RA', other_designations='CG9008-PA|CG9008-PD|CG9008-PE', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13774340, end_position_on_the_genomic_accession=13780165, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34780, CurrentID=0, Status='live', Symbol='p38b', Aliases='Dmel_CG7393, 186F5S, BG:DS00797.3, CG7393, D-p38, D-p38 MAPK, D-p38b, Dm p38b, Dmel\\CG7393, Dmp38b, Dp38, Dp38b, ESTS:186F5S, MAPK, Mpk34C, anon-sts23, dp38b, p38, p38 MAPK, p38 beta, p38B, p38Kbeta, p38b', description='p38b MAP kinase', other_designations='CG7393-PA|p38 MAP kinase|p38 mitogen-activated protein kinase|p38b-PA|stress-activated p38b MAP kinase', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13780689, end_position_on_the_genomic_accession=13782611, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34781, CurrentID=0, Status='live', Symbol='CG16890', Aliases='Dmel_ BG:DS00797.4, Dmel\\CG16890', description='CG16890 gene product from transcript CG16890-RD', other_designations='CG16890-PA|CG16890-PC|CG16890-PD|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13782231, end_position_on_the_genomic_accession=13784338, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34783, CurrentID=0, Status='live', Symbol='CG31731', Aliases='Dmel_ BG:DS00797.5, CG7491, CG7559, D00797.5, DS00797.5, Dmel\\CG31731', description='CG31731 gene product from transcript CG31731-RB', other_designations='CG31731-PB|CG31731-PC|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13784084, end_position_on_the_genomic_accession=13790562, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  34784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34784, CurrentID=0, Status='live', Symbol='CG16863', Aliases='Dmel_ BG:DS00797.6, DS00797.6, Dmel\\CG16863', description='CG16863 gene product from transcript CG16863-RA', other_designations='CG16863-PA|CG16863-PB|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13790959, end_position_on_the_genomic_accession=13793264, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34785, CurrentID=0, Status='live', Symbol='Sec71', Aliases='Dmel_CG7578, BG:DS00797.7, BIG1, CG7578, Dmel\\CG7578, sec71', description='Sec71 ortholog (S. cerevisiae)', other_designations='CG7578-PA|Sec71 ortholog|Sec71-PA|sec71', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13793521, end_position_on_the_genomic_accession=13799674, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34786, CurrentID=0, Status='live', Symbol='CAH1', Aliases='Dmel_CG7820, BG:DS00941.1, CAH, CG7820, Dmel\\CG7820, Gh7', description='Carbonic anhydrase 1', other_designations='CAH1-PA|CG7820-PA|carbonic anhydrase 1', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13803626, end_position_on_the_genomic_accession=13809994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34787, CurrentID=0, Status='live', Symbol='Adat1', Aliases='Dmel_CG16889, Adat, BG:DS00941.2, CG16889, Dmel\\CG16889, adat, dADAT1', description='Adenosine deaminase, tRNA-specific 1', other_designations='Adat1-PA|Adat1-PB|CG16889-PA|CG16889-PB|adat|adenosine deaminase, tRNA-specific 1', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13808360, end_position_on_the_genomic_accession=13811474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34788, CurrentID=0, Status='live', Symbol='CG16865', Aliases='Dmel_ BG:DS00941.3, Dmel\\CG16865', description='CG16865 gene product from transcript CG16865-RB', other_designations='CG16865-PA|CG16865-PB|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13811522, end_position_on_the_genomic_accession=13813316, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34789, CurrentID=0, Status='live', Symbol='CG16888', Aliases='Dmel_ Dmel\\CG16888', description='CG16888 gene product from transcript CG16888-RA', other_designations='CG16888-PA|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13812700, end_position_on_the_genomic_accession=13813256, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34790, CurrentID=0, Status='live', Symbol='Sos', Aliases='Dmel_CG7793, 2.3, 34Ea, BG:DS00941.4, CG7793, Dmel\\CG7793, E(sev)2A, E(var)189, EC2-5, EK2-8, EY2-3, SOS1, Su(tor)2-2, br24, br25, dSos, dme-SOS, l(2)34Ea, l(2)Sos, l(2)br24, l(2)br25, l(2)k05224, l(2)k06321, sos, Sos', description='Son of sevenless', other_designations='CG7793-PA|Son-of-sevenless|Sos-PA|lethal 24 in the black-reduced region|lethal 25 in the black-reduced region', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13813816, end_position_on_the_genomic_accession=13819824, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34791, CurrentID=0, Status='live', Symbol='b', Aliases='Dmel_CG7811, ADC, B, BG:DS00941.5, CG7811, DGad2, DmGad2, Dmel\\CG7811, GAD[[2]], Gad2, anon-34Db, cDNA 1', description='black', other_designations='Aspartate-alpha-decarboxylase|CG7811-PA|CG7811-PB|CG7811-PC|Oregon-R glutamate decarboxylase|aspartate 1-decarboxylase|b-PA|b-PB|b-PC|glutamic acid decarboxylase 2', map_location='2-48.5 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13821248, end_position_on_the_genomic_accession=13823979, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34792, CurrentID=0, Status='live', Symbol='tam', Aliases='Dmel_CG8987, 34Dc, AAC47658, BG:DS00941.6, CG8987, D-pol gamma-alpha, DNApol-gamma, DNApol-gamma125, Dmel\\CG8987, MtPolA, POLGalpha, Polgamma, Tam, anon-34Dc, br5, cDNA 2, l(2)34Dc, l(2)br14, l(2)br5, l(2)brS, mtDNA pol, mtPolA, pol gamma, pol gamma-alpha, polgamma, polgamma-alpha', description='tamas', other_designations='CG8987-PA|DNA polymerase gamma|DNA polymerase gamma 125kD|alpha subunit|catalytic subunit of the mitochondrial DNA polymerase|catalytic(alpha)subunit|lethal 14 in the black-reduced region|lethal 5 in the black-reduced region|lethal(2)34Dc|mitochondrial gamma polymerase|mtDNA polymerase alpha|tam-PA', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13823980, end_position_on_the_genomic_accession=13827772, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34793, CurrentID=0, Status='live', Symbol='Arpc1', Aliases='Dmel_CG8978, 34Dd, ARC-P41, ARPC1, ARPC1/p41, Arc41, ArcP41, ArpC1, BG:DS00941.7, CG8978, D-sop2, DSop2, Dmel\\CG8978, Dsop2, Sop2, Spo2, arc41, arpc1, br32, l(2)34Dd, l(2)Sop2, l(2)br32, p41, p42, sop, sop2, sop2/arc41', description='Actin-related protein 2/3 complex, subunit 1', other_designations='Arpc1-PA|Arpc1-PB|CG8978-PA|CG8978-PB|lethal(2)34Dd|profilin 2|suppressor of|suppressor of profilin 2', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13827941, end_position_on_the_genomic_accession=13829741, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34794, CurrentID=0, Status='live', Symbol='Orc5', Aliases='Dmel_CG7833, BG:DS00941.8, CG7833, D-orc5, DmORC5, DmOrc5, Dmel\\CG7833, ORC, ORC5, br39, dORC5, l(2)34Df, l(2)br39, orc5', description='Origin recognition complex subunit 5', other_designations='CG7833-PA|Orc5-PA|lethal(2)34Df|origin recognition complex subunit 5', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13829837, end_position_on_the_genomic_accession=13831576, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34796, CurrentID=0, Status='live', Symbol='RpII33', Aliases='Dmel_CG7885, 152117_at, BG:DS00941.10, CG7885, D-rpII33, Dmel\\CG7885, Pol II, Pol II RPII33, PolII, PolIIo, RNA Pol II, RNA pol II, RNA polII, RNAP, RPB3, Rbp3, Rpb3, Rpll33, br17, l(2)34Dg, l(2)br17, l(2)k05605, l34Dg, pol II, polII, rpII33, rpb3', description='RNA polymerase II 33kD subunit', other_designations='CG7885-PA|RNA polymerase II|RpII33-PA|lethal 17 in the black-reduced region|lethal(2)34Dg', map_location='34D1-34D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13833337, end_position_on_the_genomic_accession=13834306, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34799, CurrentID=0, Status='live', Symbol='CG8997', Aliases='Dmel_ BG:DS00941.12, CT25774, DS00941.12, Dmel\\CG8997', description='CG8997 gene product from transcript CG8997-RB', other_designations='CG8997-PA|CG8997-PB', map_location='34D7-34D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13843915, end_position_on_the_genomic_accession=13845325, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34800, CurrentID=0, Status='live', Symbol='CG7916', Aliases='Dmel_ BG:DS00941.13, CT23894, Dmel\\CG7916, LP05416', description='CG7916 gene product from transcript CG7916-RA', other_designations='CG7916-PA', map_location='34D7-34D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13846752, end_position_on_the_genomic_accession=13847799, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34801, CurrentID=0, Status='live', Symbol='CG7953', Aliases='Dmel_ BG:DS00941.14, DS00941.14, Dmel\\CG7953, anon-WO0140519.73', description='CG7953 gene product from transcript CG7953-RB', other_designations='CG7953-PA|CG7953-PB', map_location='34D7-34D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13847925, end_position_on_the_genomic_accession=13849223, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34802, CurrentID=0, Status='live', Symbol='CG7968', Aliases='Dmel_ BG:DS00941.15, DS00941.15, Dmel\\CG7968', description='CG7968 gene product from transcript CG7968-RA', other_designations='CG7968-PA', map_location='34D8-34D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13849400, end_position_on_the_genomic_accession=13850329, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34803, CurrentID=0, Status='live', Symbol='CenG1A', Aliases='Dmel_CG31811, BG:DS08220.1, CG16866, CG16867, CG31811, CG7982, Dmel\\CG31811, cenG, cenG1A, cenG1a, ceng1A', description='Centaurin gamma 1A', other_designations='CG31811-PA|CG31811-PB|CG31811-PC|CenG1A-PA|CenG1A-PB|CenG1A-PC|centaurin gamma|centaurin gamma 1A|centaurin gamma-1A', map_location='34D6-34E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13835564, end_position_on_the_genomic_accession=13898712, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34804, CurrentID=0, Status='live', Symbol='Smg5', Aliases='Dmel_CG8954, BG:DS08220.2, CG8954, Dmel\\CG8954, SMG5, anon-34Ea, dSMG5, smg5', description='CG8954 gene product from transcript CG8954-RB', other_designations='CG8954-PA|CG8954-PB|CG8954-PC|Smg5-PA|Smg5-PB|Smg5-PC', map_location='34E2-34E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13898828, end_position_on_the_genomic_accession=13904661, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34805, CurrentID=0, Status='live', Symbol='Ance', Aliases='Dmel_CG8827, ACE, ANCE, AnCE, BG:DS08220.3, CG8827, Dmel\\CG8827, RACE, Race, ance, anon-EST:fe3D10, br31, l(2)34Eb, l(2)br31, l34Eb, race', description='Angiotensin converting enzyme', other_designations='Ance-PA|Ance-PB|Ance-PC|CG8827-PA|CG8827-PB|CG8827-PC|angiotensin I-converting enzyme|angiotensin converting enzyme|angiotensin-converting enzyme|anon-fast-evolving-3D10|related to angiotensin converting enzyme', map_location='34E2-34E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13905645, end_position_on_the_genomic_accession=13909212, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34806, CurrentID=0, Status='live', Symbol='Ance-2', Aliases='Dmel_CG16869, CG16869, CG31844, Dmel\\CG16869, ance-2', description='CG16869 gene product from transcript CG16869-RA', other_designations='Ance-2-PA|CG16869-PA', map_location='34E2-34E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13909333, end_position_on_the_genomic_accession=13911719, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34807, CurrentID=0, Status='live', Symbol='Acyp', Aliases='Dmel_CG16870, ACPDro, AcPDro, BG:DS00180.1, CG16870, Dmel\\CG16870', description='Acylphosphatase', other_designations='Acyp-PA|Acyp-PB|CG16870-PA|CG16870-PB|acylphosphatase', map_location='34E2-34E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13912083, end_position_on_the_genomic_accession=13912652, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34808, CurrentID=0, Status='live', Symbol='Ance-3', Aliases='Dmel_CG17988, BG:DS00180.5, CG17988, CG31766, CG31843, CT40163, DS00180.5, Dmel\\CG17988, ance-3', description='CG17988 gene product from transcript CG17988-RC', other_designations='Ance-3-PB|Ance-3-PC|CG17988-PB|CG17988-PC', map_location='34E3-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13914725, end_position_on_the_genomic_accession=13956188, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  34809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34809, CurrentID=0, Status='live', Symbol='CG16886', Aliases='Dmel_ BG:DS00180.2, Dmel\\CG16886, vajk-1', description='CG16886 gene product from transcript CG16886-RB', other_designations='CG16886-PA|CG16886-PB|uncharacterized protein', map_location='34E4-34E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13920707, end_position_on_the_genomic_accession=13923553, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34811, CurrentID=0, Status='live', Symbol='CG16885', Aliases='Dmel_ CG18634, Dmel\\CG16885, vajk-2', description='CG16885 gene product from transcript CG16885-RA', other_designations='CG16885-PA|uncharacterized protein', map_location='34E4-34E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13932559, end_position_on_the_genomic_accession=13934692, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34812, CurrentID=0, Status='live', Symbol='CG16884', Aliases='Dmel_ BG:DS00180.3, Dmel\\CG16884, vajk-3', description='CG16884 gene product from transcript CG16884-RA', other_designations='CG16884-PA|uncharacterized protein', map_location='34E4-34E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13944756, end_position_on_the_genomic_accession=13947476, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34813, CurrentID=0, Status='live', Symbol='NimB1', Aliases='Dmel_CG33119, BG:DS00180.12, BcDNA:GH07762, CG33119, CG8855, CT25428, DS00180.12, Dmel\\CG33119, nim B1, nimB1', description='Nimrod B1', other_designations='CG33119-PA|CG33119-PB|NimB1-PA|NimB1-PB|nimrod B1', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13963507, end_position_on_the_genomic_accession=13965088, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34814, CurrentID=0, Status='live', Symbol='NimB4', Aliases='Dmel_CG33115, BG:DS00180.8, CG16882, CG33115, CT37470, DS00180.8, Dmel\\CG33115, nimB4', description='Nimrod B4', other_designations='CG33115-PA|NimB4-PA|nimrod B4|nimrodB4', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13968280, end_position_on_the_genomic_accession=13969976, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34815, CurrentID=0, Status='live', Symbol='NimB5', Aliases='Dmel_CG16873, BG:DS00180.9, CG16873, Dmel\\CG16873, nimB5', description='Nimrod B5', other_designations='CG16873-PA|CG16873-PB|CG16873-PC|NimB5-PA|NimB5-PB|NimB5-PC|nimrod B5', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13970177, end_position_on_the_genomic_accession=13972173, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34816, CurrentID=0, Status='live', Symbol='NimC1', Aliases='Dmel_CG8942, BG:DS00180.10, CG8942, CT25688, Dmel\\CG8942, Nim, NimC, P1, nimC1, nimrod', description='Nimrod C1', other_designations='CG8942-PA|CG8942-PB|CG8942-PC|NimC1-PA|NimC1-PB|NimC1-PC|NimrodC1|nimrod|nimrod C1', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13973158, end_position_on_the_genomic_accession=13976769, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34817, CurrentID=0, Status='live', Symbol='Cyp28a5', Aliases='Dmel_CG8864, 28a5, BEST:GH10483, BG:DS00180.11, CG8864, Cyp28A5, DS00180.11, Dmel\\CG8864, cyp28a5, gh10483', description='CG8864 gene product from transcript CG8864-RA', other_designations='CG8864-PA|Cyp28a5-PA|cytochrome P450', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13977302, end_position_on_the_genomic_accession=13979569, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34818, CurrentID=0, Status='live', Symbol='NimC2', Aliases='Dmel_CG18146, BG:DS00180.14, CG18146, CG18146-PB, CT40902, DS00180.14, Dmel\\CG18146, nimC2', description='Nimrod C2', other_designations='CG18146-PA|CG18146-PB|CG18146-PC|NimC2-PA|NimC2-PB|NimC2-PC|nimrod C2', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13980126, end_position_on_the_genomic_accession=13983269, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34819, CurrentID=0, Status='live', Symbol='rk', Aliases='Dmel_CG8930, BG:DS00180.13, CG8930, CT25644, DLGR-2, DLGR2, Dmel\\CG8930, LGR2/CG8930, rk', description='rickets', other_designations='CG8930-PA|glycoprotein hormone receptor II|glycoprotein-hormone-receptor-II|leucine-rich repeat-containing G protein-coupled receptor 2|rk-PA', map_location='2-48.8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13982743, end_position_on_the_genomic_accession=14001248, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  34822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34822, CurrentID=0, Status='live', Symbol='CG4500', Aliases='Dmel_ 2a9, BG:DS05899.1, Dmel\\CG4500, I, anon-EST:CL2a9', description='CG4500 gene product from transcript CG4500-RA', other_designations='CG4500-PA', map_location='34F1-34F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14027001, end_position_on_the_genomic_accession=14029368, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34823, CurrentID=0, Status='live', Symbol='CG18095', Aliases='Dmel_ BG:DS05899.7, DS05899.7, Dmel\\CG18095', description='CG18095 gene product from transcript CG18095-RB', other_designations='CG18095-PA|CG18095-PB|CG18095-PC|gp150-like|uncharacterized protein', map_location='34F1-34F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14029262, end_position_on_the_genomic_accession=14031539, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34824, CurrentID=0, Status='live', Symbol='NimC4', Aliases='Dmel_CG16876, BG:DS05899.3, CG16876, CT35247, Dmel\\CG16876, nimC4, simu', description='Nimrod C4', other_designations='CG16876-PA|NimC4-PA|nimrod C4|six microns under|six-microns-under', map_location='34F2-34F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14047582, end_position_on_the_genomic_accession=14048863, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38572, CurrentID=0, Status='live', Symbol='CG18418', Aliases='Dmel_ Dmel\\CG18418, anon-WO0140519.190', description='CG18418 gene product from transcript CG18418-RA', other_designations='CG18418-PA|uncharacterized protein', map_location='64B13-64B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4688929, end_position_on_the_genomic_accession=4690070, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34826, CurrentID=0, Status='live', Symbol='nAChRalpha5', Aliases='Dmel_CG32975, BG:DS05899.4, BG:DS05899.5, CG16878, CG32975, CG4498, CT14633, Dalpha5, Dmel\\CG32975, Nic 34E, anon-WO02059370.44, daf, lincRNA.128, lincRNA.S1055, nAChR, nAChRalpha-34E, nAcRalpha-34E, nAcRalpha-34e', description='nicotinic Acetylcholine Receptor alpha5', other_designations='CG32975-PB|CG32975-PD|CG32975-PH|Dalpha5|nAChRalpha5-PB|nAChRalpha5-PD|nAChRalpha5-PH|nicotinic acetylcholine Receptor_34E|nicotinic acetylcholine receptor alpha 34E|nicotinic acetylcholine receptor alpha5', map_location='34F2-34F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14040170, end_position_on_the_genomic_accession=14094401, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  34827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34827, CurrentID=0, Status='live', Symbol='CG16879', Aliases='Dmel_ BG:DS01759.1, DS01759.1, Dmel\\CG16879', description='CG16879 gene product from transcript CG16879-RB', other_designations='CG16879-PA|CG16879-PB|uncharacterized protein', map_location='34F4-34F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14104483, end_position_on_the_genomic_accession=14107461, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34828, CurrentID=0, Status='live', Symbol='CG15293', Aliases='Dmel_ BG:DS01759.2, CT35241, Dmel\\CG15293, HL03474', description='CG15293 gene product from transcript CG15293-RC', other_designations='CG15293-PA|CG15293-PC|uncharacterized protein', map_location='34F4-34F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14108057, end_position_on_the_genomic_accession=14109447, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34829, CurrentID=0, Status='live', Symbol='CG17341', Aliases='Dmel_ BG:DS01523.1, Dmel\\CG17341', description='CG17341 gene product from transcript CG17341-RB', other_designations='CG17341-PB|uncharacterized protein', map_location='35A1-35A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14134491, end_position_on_the_genomic_accession=14160587, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34830, CurrentID=0, Status='live', Symbol='CG43333', Aliases='Dmel_ BG:DS01523.2, CG15291, CG32972, Dmel\\CG43333, Dmel_CG32972', description='CG43333 gene product from transcript CG43333-RA', other_designations='CG43333-PA|uncharacterized protein', map_location='35A1-35A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14163508, end_position_on_the_genomic_accession=14180686, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34831, CurrentID=0, Status='live', Symbol='Dyrk2', Aliases='Dmel_CG4551, BG:DS01523.3, CG4551, DYRK, DYRK2, Dm2, DmDyrk2, Dmel\\CG4551, dDYRK2, smi35A', description='Dual-specificity tyrosine phosphorylation-regulated kinase 2', other_designations='CG4551-PA|CG4551-PB|CG4551-PC|CG4551-PD|CG4551-PE|CG4551-PF|CG4551-PG|Dyrk2-PA|Dyrk2-PB|Dyrk2-PC|Dyrk2-PD|Dyrk2-PE|Dyrk2-PF|Dyrk2-PG|dual specificity tyrosine phosphorylation-regulated kinase|smell impaired 35A', map_location='35A1-35A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14184478, end_position_on_the_genomic_accession=14234126, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34833, CurrentID=0, Status='live', Symbol='ms(2)34Fe', Aliases='Dmel_CG15287, BG:DS01068.1, CG15287, Dmel\\CG15287, br40, l(2)34Fe, ms(2)34Fh, ms(2)br40, ms34Fe', description='male sterile (2) 34Fe', other_designations='CG15287-PA|ms(2)34Fe-PA', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14328429, end_position_on_the_genomic_accession=14333860, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34834, CurrentID=0, Status='live', Symbol='CG15286', Aliases='Dmel_ BG:DS01068.11, Dmel\\CG15286', description='CG15286 gene product from transcript CG15286-RA', other_designations='CG15286-PA|CG15286-PB|uncharacterized protein', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14338240, end_position_on_the_genomic_accession=14340080, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34835, CurrentID=0, Status='live', Symbol='CG33090', Aliases='Dmel_ BG:DS01068.2, CG15287, CG31837, CG7469, Dmel\\CG33090', description='CG33090 gene product from transcript CG33090-RE', other_designations='CG33090-PB|CG33090-PC|CG33090-PD|CG33090-PE|glucosylceramidase|uncharacterized protein', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14334158, end_position_on_the_genomic_accession=14345795, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34836, CurrentID=0, Status='live', Symbol='Send2', Aliases='Dmel_CG18125, BG:DS01068.10, CG18125, DS01068.10, Dmel\\CG18125, SPH190', description='Spermathecal endopeptidase 2', other_designations='CG18125-PA|Send2-PA|spermathecal endopeptidase 2', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14345915, end_position_on_the_genomic_accession=14346688, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34837, CurrentID=0, Status='live', Symbol='mTTF', Aliases='Dmel_CG18124, BG:DS01068.4, CG18124, DS01068.4, DmTTF, Dmel\\CG18124', description='mitochondrial transcription termination factor', other_designations='CG18124-PA|mTTF-PA', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14347239, end_position_on_the_genomic_accession=14348790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34838, CurrentID=0, Status='live', Symbol='l(2)34Fc', Aliases='Dmel_CG7532, A2, BG:DS01068.5, BcDNA:RE60882, CG7532, DS01068.5, Dmel\\CG7532, anon-EST:ParkEST060, br8, l(2)br8, l34Fc', description='lethal (2) 34Fc', other_designations='CG7532-PA|CG7532-PB|l(2)34Fc-PA|l(2)34Fc-PB|lethal 8 in the black-reduced region|lethal group A2', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14349020, end_position_on_the_genomic_accession=14351513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34839, CurrentID=0, Status='live', Symbol='l(2)34Fd', Aliases='Dmel_CG7516, A4, BG:DS01068.6, CG7516, Dmel\\CG7516, br15, l(2)br15', description='lethal (2) 34Fd', other_designations='CG7516-PA|l(2)34Fd-PA|lethal 15 in the black-reduced region|lethal group A4', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14352594, end_position_on_the_genomic_accession=14355085, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34840, CurrentID=0, Status='live', Symbol='Rab14', Aliases='Dmel_CG4212, 57H4T, AAF53390, BG:DS01068.7, CG4212, DRAB14, D DmRab14, Dmel\\CG4212, ESTS:57H4T, O18336, Rab-r14, anon-sts33, rab14', description='CG4212 gene product from transcript CG4212-RA', other_designations='CG4212-PA|CG4212-PB|CG4212-PD|Rab-protein 14|Rab14-PA|Rab14-PB|Rab14-PD', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14355145, end_position_on_the_genomic_accession=14358764, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34842, CurrentID=0, Status='live', Symbol='spel1', Aliases='Dmel_CG4215, BG:DS01068.9, CG4215, Dmel\\CG4215, Msh2, SPEL1, spel, spel-1', description='spellchecker1', other_designations='CG4215-PA|CG4215-PD|CG4215-PE|CG4215-PF|spel1-PA|spel1-PD|spel1-PE|spel1-PF|spellchecker|spellchecker 1', map_location='35A4-35B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14362069, end_position_on_the_genomic_accession=14366840, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34843, CurrentID=0, Status='live', Symbol='ppk', Aliases='Dmel_CG3478, BG:DS06238.1, CG3478, Dmel\\CG3478, PPK, PPK1, dmdNaC1, dmdNaCl, mdNaC1, pkt1, ppk', description='pickpocket', other_designations='CG3478-PA|Pickpocket1|drifter|multidendritic neurons sodium channel 1|pickpocket 1|ppk-PA', map_location='35B1-35B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14378218, end_position_on_the_genomic_accession=14381172, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34844, CurrentID=0, Status='live', Symbol='elB', Aliases='Dmel_CG4220, BG:DS06238.3, CG4220, Dmel\\CG4220, Elb, P10, el, el[B], elb, elbow', description='elbow B', other_designations='CG4220-PC|CG4220-PD|ElbowB|elB-PC|elB-PD|elbow', map_location='2-50.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14390628, end_position_on_the_genomic_accession=14409549, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34845, CurrentID=0, Status='live', Symbol='Pburs', Aliases='Dmel_CG15284, A, Bur beta, Burs beta, Bursbeta, CG15284, Dmel\\CG15284, PBURS, bur &gr;, burs beta, burs-beta, bursbeta, pads-b, pburs, pu, pup', description='Partner of Bursicon', other_designations='CG15284-PA|Pburs-PA|bursicon|bursicon beta|bursicon beta subunit|bursicon beta-subunit|lethal group A|partner of burs|partner of bursicon|pupal', map_location='35B1-35B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14416729, end_position_on_the_genomic_accession=14417331, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34846, CurrentID=0, Status='live', Symbol='Cpr35B', Aliases='Dmel_CG3474, BG:DS06238.4, CG3474, DS06283.4, DmelCpr35B, Dmel\\CG3474', description='Cuticular protein 35B', other_designations='CG3474-PA|Cpr35B-PA|cuticular protein 35B', map_location='35B1-35B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14424315, end_position_on_the_genomic_accession=14425189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34847, CurrentID=0, Status='live', Symbol='noc', Aliases='Dmel_CG4491, BG:DS04641.1, CG4491, Dmel\\CG4491, Noc, br22, el, l(2)35Ba, l(2)35Ba/nocA, l(2)br22, l35BaA, noc[A], noc', description='no ocelli', other_designations='CG4491-PA|CG4491-PB|no-ocelli|noc-PA|noc-PB', map_location='35B2-35B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14490859, end_position_on_the_genomic_accession=14494019, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34848, CurrentID=0, Status='live', Symbol='CG4218', Aliases='Dmel_ BG:DS04641.8, Dmel\\CG4218, anon-SAGE:Wang-118, anon-SAGE:Wang-133', description='CG4218 gene product from transcript CG4218-RC', other_designations='CG4218-PA|CG4218-PB|CG4218-PC|CG4218-PD|anon-SAGE:Wang-118|anon-SAGE:Wang-133', map_location='35B2-35B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14547498, end_position_on_the_genomic_accession=14549109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34849, CurrentID=0, Status='live', Symbol='CG3473', Aliases='Dmel_ BG:DS01486.1, BcDNA:AT16033, Dmel\\CG3473, cg3473', description='CG3473 gene product from transcript CG3473-RB', other_designations='CG3473-PA|CG3473-PB', map_location='35B3-35B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14563015, end_position_on_the_genomic_accession=14563726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34850, CurrentID=0, Status='live', Symbol='osp', Aliases='Dmel_CG3479, BG:DS01486.5, BG:DS01486.7, BG:DS07721.2, CG3479, Dmel\\CG3479, alw, anon-WO02059370.75, l(2)rJ571, v(2)rJ571', description='outspread', other_designations='CG3479-PC|CG3479-PD|CG3479-PE|CG3479-PF|arc-like wing|arclike wing|osp-PC|osp-PD|osp-PE|osp-PF', map_location='2-50.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14599196, end_position_on_the_genomic_accession=14689340, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34853, CurrentID=0, Status='live', Symbol='CG15282', Aliases='Dmel_ BG:DS07721.3, BcDNA:RE15191, DS07721.3, Dmel\\CG15282', description='CG15282 gene product from transcript CG15282-RC', other_designations='CG15282-PA|CG15282-PB|CG15282-PC|uncharacterized protein', map_location='35B4-35B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14712471, end_position_on_the_genomic_accession=14713298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34854, CurrentID=0, Status='live', Symbol='CG42685', Aliases='Dmel_ BG:DS07721.6, CG12636, CG33482, DS07721.6, Dmel\\CG42685, Dmel_CG12636, Dmel_CG33482', description='CG42685 gene product from transcript CG42685-RA', other_designations='CG42685-PA|DS07721.6-like protein|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14724617, end_position_on_the_genomic_accession=14730177, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34856, CurrentID=0, Status='live', Symbol='CG4691', Aliases='Dmel_ BG:DS06874.1, Dmel\\CG4691, anon-WO0140519.47', description='CG4691 gene product from transcript CG4691-RA', other_designations='CG4691-PA', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14760767, end_position_on_the_genomic_accession=14761924, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34857, CurrentID=0, Status='live', Symbol='CSN1a', Aliases='Dmel_CG4697, BG:DS06874.2, CG4697, CH1-1, COP9, CSN1, Csn1A, Csn1a, Dch1, Dch1-1, Dmel\\CG4697, csn1', description='COP9 signalosome subunit 1a', other_designations='CG4697-PA|COP9 complex homolog subunit 1 a|CSN1a-PA|drosophila COP9 signalosome homolog 1', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14762765, end_position_on_the_genomic_accession=14764065, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34858, CurrentID=0, Status='live', Symbol='CG4701', Aliases='Dmel_ BG:DS06874.3, DmDM19DC4Z, Dmel\\CG4701', description='CG4701 gene product from transcript CG4701-RA', other_designations='CG4701-PA', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14765908, end_position_on_the_genomic_accession=14767225, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34859, CurrentID=0, Status='live', Symbol='CG4650', Aliases='Dmel_ Dmel\\CG4650, SPH197', description='CG4650 gene product from transcript CG4650-RC', other_designations='CG4650-PB|CG4650-PC', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14773814, end_position_on_the_genomic_accession=14774981, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34863, CurrentID=0, Status='live', Symbol='CG15279', Aliases='Dmel_ BG:DS03431.1, DS03431.1, DmNAT6, Dmel\\CG15279, anon-WO0149848.9', description='CG15279 gene product from transcript CG15279-RB', other_designations='CG15279-PA|CG15279-PB|CG15279-PD|uncharacterized protein', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14844993, end_position_on_the_genomic_accession=14851749, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34864, CurrentID=0, Status='live', Symbol='CG4480', Aliases='Dmel_ Dmel\\CG4480', description='CG4480 gene product from transcript CG4480-RB', other_designations='CG4480-PB|CG4480-PC', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14870371, end_position_on_the_genomic_accession=14872022, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34865, CurrentID=0, Status='live', Symbol='CG15278', Aliases='Dmel_ Dmel\\CG15278, NEST:bs32h08, bs32h08.y1', description='CG15278 gene product from transcript CG15278-RC', other_designations='CG15278-PA|CG15278-PB|CG15278-PC|uncharacterized protein', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14874376, end_position_on_the_genomic_accession=14876333, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34866, CurrentID=0, Status='live', Symbol='ProtA', Aliases='Dmel_CG4479, BG:DS03431.2, BcDNA:AT31403, CG4479, Dmel\\CG4479, Mst35Ba, dProtA, mst35B, mst35Ba, mst35b, protA', description='Protamine A', other_designations='CG4479-PA|CG4479-PB|Male-specific-transcript-35Ba|ProtA-PA|ProtA-PB|ProtamineA|dProtamineA|protamine|protamine (mst35Ba)|protamine A', map_location='35B6-35B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14879050, end_position_on_the_genomic_accession=14881652, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34867, CurrentID=0, Status='live', Symbol='ProtB', Aliases='Dmel_CG4478, BG:DS03431.3, BcDNA:GH11850, CG4478, Dmel\\CG4478, Mst35Bb, dProtB, mst35b, prot-B, protB', description='Protamine B', other_designations='CG4478-PA|Male-specific-transcript-35Bb|ProtB-PA|ProtamineB|dProtamineB|protamine|protamine (mst35Bb)|protamine B', map_location='35B7-35B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14882101, end_position_on_the_genomic_accession=14883721, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34870, CurrentID=0, Status='live', Symbol='CG3491', Aliases='Dmel_ BG:DS03323.1, Dmel\\CG3491', description='CG3491 gene product from transcript CG3491-RA', other_designations='CG3491-PA', map_location='35B7-35B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14924590, end_position_on_the_genomic_accession=14930565, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34872, CurrentID=0, Status='live', Symbol='mol', Aliases='Dmel_CG4482, 35Bb, B1, BG:DS01219.1, CG15268, CG4482, Dmel\\CG4482, Mol, NIP, br3, l(2)35Bb, l(2)SH1330, l(2)SH2 1330, l(2)br23, l(2)br3, l35Bb, nip', description='moladietz', other_designations='CG4482-PA|CG4482-PC|CG4482-PD|CG4482-PF|CG4482-PG|lethal (2) 35Bb|lethal (2) SH1330|lethal 23 in the black-reduced region|lethal 3 in the black-reduced region|lethal group B1|mol-PA|mol-PC|mol-PD|mol-PF|mol-PG|numb-interacting protein', map_location='35B7-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14975747, end_position_on_the_genomic_accession=14997559, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34873, CurrentID=0, Status='live', Symbol='DCTN5-p25', Aliases='Dmel_CG10846, BG:DS00929.1, C4, CG10846, DS00929.1, Dmel\\CG10846, br2, dyn-p25, l(2)35Bf, l(2)br2, p25', description='Dynactin 5, p25 subunit', other_designations='CG10846-PA|DCTN5-p25-PA|dynactin 5, p25 subunit|dynactin subunit p-25|dynactin, p25 subunit|dynactin-subunit-p25|lethal 2 in the black-reduced region|lethal group C4|lethal(2)35Bf', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15005733, end_position_on_the_genomic_accession=15006607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34874, CurrentID=0, Status='live', Symbol='l(2)35Be', Aliases='Dmel_CG4140, BG:DS00929.2, C3, CG4140, Dmel\\CG4140, br10, l(2)br10', description='lethal (2) 35Be', other_designations='CG4140-PA|l(2)35Be-PA|lethal 10 in the black-reduced region|lethal group C3', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15006583, end_position_on_the_genomic_accession=15007431, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34875, CurrentID=0, Status='live', Symbol='NC2beta', Aliases='Dmel_CG4185, BG:DS00929.3, CG4185, Dmel\\CG4185, Dr1, NC2 beta, Nc2b, dDr1, dNC2', description='Negative Cofactor 2beta', other_designations='CG4185-PA|NC2beta-PA|negative cofactor 2beta', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15007813, end_position_on_the_genomic_accession=15008538, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34876, CurrentID=0, Status='live', Symbol='l(2)35Bc', Aliases='Dmel_CG4103, BG:DS00929.4, C1, CG4103, Dmel\\CG4103, br4, l(2)br, l(2)br11, l(2)br13, l(2)br4, l(2)k08808', description='lethal (2) 35Bc', other_designations='CG4103-PA|CG4103-PB|l(2)35Bc-PA|l(2)35Bc-PB|lethal 11 in the black-reduced region|lethal 13 in the black-reduced region|lethal 4 in the black-reduced region|lethal group C1', map_location='35B8-35B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15008489, end_position_on_the_genomic_accession=15009724, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34877, CurrentID=0, Status='live', Symbol='l(2)35Bd', Aliases='Dmel_CG3688, BG:DS00929.5, C2, CG3688, Dmel\\CG3688, br9, l(2)10408, l(2)br9', description='lethal (2) 35Bd', other_designations='CG3688-PA|l(2)35Bd-PA|lethal 9 in the black-reduced region|lethal group C2', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15009939, end_position_on_the_genomic_accession=15011764, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34878, CurrentID=0, Status='live', Symbol='GABA-B-R1', Aliases='Dmel_CG15274, BG:DS00929.6, CG15274, D-GABA[[B]]R1, D-Gaba1, Dmel\\CG15274, GABA(B)R1, GABA[[B]], GABA[[B]]R, GABA[[B]]R1, GB1, anon-WO0170980.175, anon-WO0170980.176', description='metabotropic GABA-B receptor subtype 1', other_designations='CG15274-PB|CG15274-PC|GABA-B-R1-PB|GABA-B-R1-PC', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15016331, end_position_on_the_genomic_accession=15034704, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  34879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34879, CurrentID=0, Status='live', Symbol='yellow-c', Aliases='Dmel_CG4182, BG:DS00929.8, CG4182, Dmel\\CG4182, anon-EST:ParkEST125, yell-c', description='CG4182 gene product from transcript CG4182-RB', other_designations='CG4182-PA|CG4182-PB|yellow-c-PA|yellow-c-PB', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15035394, end_position_on_the_genomic_accession=15037841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34881, CurrentID=0, Status='live', Symbol='Su(H)', Aliases='Dmel_CG3497, BG:DS00929.10, C, CBF1, CG3497, CSL, D, Dmel\\CG3497, E(H), FTZ-F2, RBJK_DROME, RBP-J kappa, RBP-J[[Kappa]], RBP-Jkappa, SU(H), SUH, Su(h), SuH, Su_H, anon-WO0257455.3, br7, dRBP-JK, dRBP-J[[Kappa]], dRBP-Jk, l(1)br7, l(2)35Bh, l(2)br7, l(2)k07904, oss, su(H)', description='Suppressor of Hairless', other_designations='CG3497-PA|CG3497-PC|Su(H)-PA|Su(H)-PC|lethal 7 in the black-reduced region|lethal group D|suppressor of hairless|suppressor-of-hairless|supressor of hairless|transcription unit C', map_location='2-50.5 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15039488, end_position_on_the_genomic_accession=15043334, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34882, CurrentID=0, Status='live', Symbol='ck', Aliases='Dmel_CG7595, 53D10S, BG:DS00929.11, CG7595, D, D2, DM7a, DMyoVIIa, Dm 35B, DmVIIA, DmVIIa, Dmel\\CG7595, Dro35B, ESTS:53D10S, Mhc35BC, Myo7A, MyoVIIA, anon-35Bb, br27/MyoVIIA, l(1)35Ca, l(2)07130, l(2)35Ca, l(2)br27, l35Ca, myoVIIA, stc, stch, ck', description='crinkled', other_designations='CG7595-PA|CG7595-PB|CG7595-PC|CG7595-PD|ck-PA|ck-PB|ck-PC|ck-PD|lethal 27 in the black-reduced region|lethal group D2|myosin 7a|myosin 7a heavy chain|myosin VII|myosin VIIa|myosin heavy chain at 35BC|myosin-VII|mysoin VIIa|stubby chaetae|transcription unit D', map_location='2-51.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15044963, end_position_on_the_genomic_accession=15057862, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34883, CurrentID=0, Status='live', Symbol='TfIIS', Aliases='Dmel_CG3710, BG:DS00929.12, CG3710, DmS-II, DmSII, Dmel\\CG3710, RnpSII, TFIIS, TFIISA, TFS-II, br52, l(2)35Cf, l(2)35cF, l(2)br52, l35Cf', description='RNA polymerase II elongation factor', other_designations='CG3710-PA|CG3710-PB|TfIIS-PA|TfIIS-PB|lethal(2)35Cf|transcription cleavage factor IIS', map_location='35B10-35B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15058919, end_position_on_the_genomic_accession=15060757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34885, CurrentID=0, Status='live', Symbol='vig', Aliases='Dmel_CG4170, BG:DS00929.13, CG4170, Dmel\\CG4170, EP(2)0812, VIG, Vig', description='vasa intronic gene', other_designations='CG4170-PA|CG4170-PB|CG4170-PC|CG4170-PD|vig-PA|vig-PB|vig-PC|vig-PD', map_location='35C1-35C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15062931, end_position_on_the_genomic_accession=15070316, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34886, CurrentID=0, Status='live', Symbol='CG15270', Aliases='Dmel_ BG:DS04929.1, Dmel\\CG15270', description='CG15270 gene product from transcript CG15270-RC', other_designations='Axs-like|CG15270-PA|CG15270-PB|CG15270-PC|CG15270-PD|CG15270-PE|CG15270-PF|uncharacterized protein', map_location='35C1-35C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15074776, end_position_on_the_genomic_accession=15096915, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  34887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34887, CurrentID=0, Status='live', Symbol='CG15269', Aliases='Dmel_ BG:DS04929.3, DS04929.3, Dmel\\CG15269', description='CG15269 gene product from transcript CG15269-RA', other_designations='CG15269-PA|uncharacterized protein', map_location='35C2-35C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15109405, end_position_on_the_genomic_accession=15113092, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34888, CurrentID=0, Status='live', Symbol='stc', Aliases='Dmel_CG3647, 151662_s_at, BG:DS04929.4, BcDNA.LD22726, BcDNA:LD22726, CG3647, Cf4, Dmel\\CG3647, LD22726, br33, l(2)35Cb, l(2)SH1565, l(2)SH2 1565, l(2)br33, l35Cb', description='shuttle craft', other_designations='CG3647-PA|CG3647-PB|CG3647-PC|CG3647-PD|Chorion-factor-4|shuttlecraft|stc-PA|stc-PB|stc-PC|stc-PD', map_location='35C2-35C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15113480, end_position_on_the_genomic_accession=15118069, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34889, CurrentID=0, Status='live', Symbol='CG4168', Aliases='Dmel_ BG:DS03192.2, Dmel\\CG4168', description='CG4168 gene product from transcript CG4168-RC', other_designations='CG4168-PB|CG4168-PC|chaoptic-like', map_location='35C3-35C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15161882, end_position_on_the_genomic_accession=15200459, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  34890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34890, CurrentID=0, Status='live', Symbol='ZnT35C', Aliases='Dmel_CG3994, BG:DS07295.1, CG3994, Dmel\\CG3994, cg3994, dZnT35C', description='Zinc transporter 35C', other_designations='CG3994-PA|CG3994-PB|ZnT35C-PA|ZnT35C-PB|zinc transporter 35C', map_location='35C4-35C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15228728, end_position_on_the_genomic_accession=15247714, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34891, CurrentID=0, Status='live', Symbol='dao', Aliases='Dmel_CG15267, BG:DS07295.2, CG15267, Dmel\\CG15267, anon-WO0140519.126', description='down and out', other_designations='CG15267-PA|CG15267-PB|CG15267-PC|dao-PA|dao-PB|dao-PC', map_location='35C5-35C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15248879, end_position_on_the_genomic_accession=15254715, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  34892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34892, CurrentID=0, Status='live', Symbol='Pol32', Aliases='Dmel_CG3975, BG:DS07295.3, BcDNA:GH08865, CG3975, Dmel\\CG3975, pol32, rd', description='CG3975 gene product from transcript CG3975-RB', other_designations='CG3975-PB|CG3975-PC|Pol32-PB|Pol32-PC|reduced', map_location='35C5-35C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15254819, end_position_on_the_genomic_accession=15256304, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34893, CurrentID=0, Status='live', Symbol='l(2)35Cc', Aliases='Dmel_CG15266, BG:DS07295.5, CG15266, Dmel\\CG15266, br50, l(2)br50, l35Cc', description='lethal (2) 35Cc', other_designations='CG15266-PA|CG15266-PB|l(2)35Cc-PA|l(2)35Cc-PB', map_location='35C5-35C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15256581, end_position_on_the_genomic_accession=15257642, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34894, CurrentID=0, Status='live', Symbol='yuri', Aliases='Dmel_CG31732, BG:DS05639.1, BG:DS07851.11, CG11860, CG12635, CG31732, Dmel\\CG31732, Yuri', description='yuri gagarin', other_designations='CG31732-PB|CG31732-PC|CG31732-PD|CG31732-PI|CG31732-PN|CG31732-PP|CG31732-PR|CG31732-PS|yuri-PB|yuri-PC|yuri-PD|yuri-PI|yuri-PN|yuri-PP|yuri-PR|yuri-PS', map_location='35C5-35C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15257267, end_position_on_the_genomic_accession=15264692, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  34896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34896, CurrentID=0, Status='live', Symbol='Cul3', Aliases='Dmel_CG42616, BG:DS07851.2, CG11861, CG31829, CG42616, CUL-3, CUL3, Cul-3, Cullin3, DMcul-3, Dmel\\CG42616, Dmel_CG11861, Dmel_CG31829, br34, cul-3, cul3, dCul-3, dCul3, gft, gft/dCul-3, guftagu, l(2)06430, l(2)35Cd, l(2)br34, l35Cd, mds', description='Cullin 3', other_designations='CG42616-PC|CG42616-PD|CG42616-PE|CG42616-PF|Cul3-PC|Cul3-PD|Cul3-PE|Cul3-PF|Cullin3|cullin|cullin 3|cullin-3|guftagu|medusa', map_location='35C5-35C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15265245, end_position_on_the_genomic_accession=15272024, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34897, CurrentID=0, Status='live', Symbol='UK114', Aliases='Dmel_CG15261, BG:DS07851.3, BcDNA:GM01181, CG15261, DUK 114, D Dmel\\CG15261', description='CG15261 gene product from transcript CG15261-RB', other_designations='CG15261-PA|CG15261-PB|UK114-PA|UK114-PB', map_location='35D1-35D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15274181, end_position_on_the_genomic_accession=15275091, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34898, CurrentID=0, Status='live', Symbol='CG15263', Aliases='Dmel_ BG:DS07851.4, Dmel\\CG15263', description='CG15263 gene product from transcript CG15263-RA', other_designations='CG15263-PA|uncharacterized protein', map_location='35D1-35D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15285000, end_position_on_the_genomic_accession=15286047, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34899, CurrentID=0, Status='live', Symbol='CG15260', Aliases='Dmel_ BG:DS07851.8, Dmel\\CG15260', description='CG15260 gene product from transcript CG15260-RB', other_designations='CG15260-PB|uncharacterized protein', map_location='35D1-35D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15291426, end_position_on_the_genomic_accession=15292239, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34901, CurrentID=0, Status='live', Symbol='CG15262', Aliases='Dmel_ BG:DS07851.5, Dmel\\CG15262', description='CG15262 gene product from transcript CG15262-RA', other_designations='CG15262-PA|uncharacterized protein', map_location='35D1-35D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15297808, end_position_on_the_genomic_accession=15299367, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34902, CurrentID=0, Status='live', Symbol='nht', Aliases='Dmel_CG15259, BG:DS07851.6, CG15259, DS07851.6, Dmel\\CG15259, Nht, TAF4, TFIID, dTAF4L, dTAF[[II]]110, dmTAF4b', description='no hitter', other_designations='CG15259-PB|nht-PB|nohitter', map_location='35D1-35D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15317954, end_position_on_the_genomic_accession=15318901, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34903, CurrentID=0, Status='live', Symbol='esg', Aliases='Dmel_CG3758, 4B7, BG:DS07851.7, CG3758, Dmel\\CG3758, Esg, br43, dgl, flg, l(2)07082, l(2)35Ce, l(2)4B7, l(2)br43, l(2)esg, l35Ce, shof, wiz', description='escargot', other_designations='CG3758-PA|D.m.Escargot|Fusion-1|double glazed|esg-PA|esgargot|fleabag|shut off|wizard', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15333866, end_position_on_the_genomic_accession=15336138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34904, CurrentID=0, Status='live', Symbol='CG15258', Aliases='Dmel_ Dmel\\CG15258', description='CG15258 gene product from transcript CG15258-RA', other_designations='CG15258-PA|uncharacterized protein', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15360921, end_position_on_the_genomic_accession=15361627, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34906, CurrentID=0, Status='live', Symbol='wor', Aliases='Dmel_CG4158, BG:DS03023.1, CG4158, Dmel\\CG4158, Wor, br35, l(2)35Da, l(2)br35, l35Da', description='worniu', other_designations='CG4158-PA|D.m.Worniu|lethal(2)35Da|schnecke|warniu|wor-PA', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15423297, end_position_on_the_genomic_accession=15425580, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34907, CurrentID=0, Status='live', Symbol='CG4161', Aliases='Dmel_ BG:DS03023.2, Dmel\\CG4161', description='CG4161 gene product from transcript CG4161-RB', other_designations='CG4161-PA|CG4161-PB', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15433309, end_position_on_the_genomic_accession=15435304, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34908, CurrentID=0, Status='live', Symbol='sna', Aliases='Dmel_CG3956, BG:DS01845.1, CG3956, Dmel\\CG3956, SNA, Sco, Sna, br28, br29, l(2)35Db, l(2)br28, l(2)br29, l(3)br28, l35Db, sn', description='snail', other_designations='CG3956-PA|scutoid|sna-PA', map_location='2-51 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15476593, end_position_on_the_genomic_accession=15478260, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34910, CurrentID=0, Status='live', Symbol='lace', Aliases='Dmel_CG4162, BG:DS01845.3, CG4162, Dmel\\CG4162, Lace, SPT-II, anon-EST:Posey289, br36, l(2)35Dc, l(2)br36, l35Dc', description='CG4162 gene product from transcript CG4162-RA', other_designations='CG4162-PA|CG4162-PB|lace-PA|lace-PB|serine palmitoyltransferase subunit II|sphingosine palmitoyl transferase', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15499127, end_position_on_the_genomic_accession=15505018, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34911, CurrentID=0, Status='live', Symbol='CG15256', Aliases='Dmel_ BG:DS04862.2, Dmel\\CG15256', description='CG15256 gene product from transcript CG15256-RC', other_designations='CG15256-PB|CG15256-PC|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15550372, end_position_on_the_genomic_accession=15552029, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34912, CurrentID=0, Status='live', Symbol='kek3', Aliases='Dmel_CG4192, BG:DS04862.1, CG4192, CT12629, Dmel\\CG4192', description='kekkon-3', other_designations='CG4192-PA|CG4192-PB|Kekkon-3|kek3-PA|kek3-PB|kekkon', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15552723, end_position_on_the_genomic_accession=15583978, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34913, CurrentID=0, Status='live', Symbol='CG15255', Aliases='Dmel_ BG:BACR44L22.1, Dmel\\CG15255', description='CG15255 gene product from transcript CG15255-RB', other_designations='CG15255-PA|CG15255-PB|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15610171, end_position_on_the_genomic_accession=15611719, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34914, CurrentID=0, Status='live', Symbol='Semp1', Aliases='Dmel_CG11864, BACR44L22.8, BG:BACR44L22.8, CG11864, Dmel\\CG11864, semp1', description='Seminal metalloprotease-1', other_designations='CG11864-PA|Semp1-PA|seminal metalloprotease-1', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15612494, end_position_on_the_genomic_accession=15613335, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34915, CurrentID=0, Status='live', Symbol='CG15254', Aliases='Dmel_ BACcr44L22.2, BG:BACR44L22.2, Dmel\\CG15254', description='CG15254 gene product from transcript CG15254-RB', other_designations='CG15254-PA|CG15254-PB|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15613905, end_position_on_the_genomic_accession=15614828, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34916, CurrentID=0, Status='live', Symbol='CG15253', Aliases='Dmel_ BACR44L22.3, BG:BACR44L22.3, Dmel\\CG15253', description='CG15253 gene product from transcript CG15253-RA', other_designations='CG15253-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15615357, end_position_on_the_genomic_accession=15616351, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34917, CurrentID=0, Status='live', Symbol='CG11865', Aliases='Dmel_ BACR44L22.4, BG:BACR44L22.4, Dmel\\CG11865', description='CG11865 gene product from transcript CG11865-RA', other_designations='CG11865-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15616997, end_position_on_the_genomic_accession=15617915, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34918, CurrentID=0, Status='live', Symbol='Or35a', Aliases='Dmel_CG17868, 35a, 36E.1, BACR44L22.5, BG:BACR44L22.5, CG17868, DOR91, Dmel\\CG17868, OR35a', description='Odorant receptor 35a', other_designations='CG17868-PB|Or35a-PB|odorant receptor 35a', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15622068, end_position_on_the_genomic_accession=15623558, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34919, CurrentID=0, Status='live', Symbol='CG7631', Aliases='Dmel_ BG:BACR44L22.6, Dmel\\CG7631', description='CG7631 gene product from transcript CG7631-RA', other_designations='CG7631-PA', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15628678, end_position_on_the_genomic_accession=15629552, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34920, CurrentID=0, Status='live', Symbol='CG18480', Aliases='Dmel_ BG:DS07108.4, BcDNA:LP11031, CT42128, Dmel\\CG18480', description='CG18480 gene product from transcript CG18480-RB', other_designations='CG18480-PA|CG18480-PB|CG18480-PC|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15630002, end_position_on_the_genomic_accession=15634987, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34921, CurrentID=0, Status='live', Symbol='CG4587', Aliases='Dmel_ BG:DS07108.2, DS07108.2, Dmel\\CG4587, mdcds_3066', description='CG4587 gene product from transcript CG4587-RM', other_designations='CG4587-PH|CG4587-PI|CG4587-PJ|CG4587-PK|CG4587-PL|CG4587-PM|CG4587-PN|CG4587-PO', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15644183, end_position_on_the_genomic_accession=15718272, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  34922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34922, CurrentID=0, Status='live', Symbol='CG18477', Aliases='Dmel_ BG:DS07108.1, Dmel\\CG18477, SPH69', description='CG18477 gene product from transcript CG18477-RA', other_designations='CG18477-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15661041, end_position_on_the_genomic_accession=15662753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34923, CurrentID=0, Status='live', Symbol='CG18478', Aliases='Dmel_ BG:DS07108.5, DS07108.5, Dmel\\CG18478, SPH157', description='CG18478 gene product from transcript CG18478-RA', other_designations='CG18478-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15670174, end_position_on_the_genomic_accession=15671223, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34924, CurrentID=0, Status='live', Symbol='CycE', Aliases='Dmel_CG3938, 3938, BG:DS07108.3, CDI7, CG3938, CYCE, Cdi7, Cyc, Cyc EI, CyclE, CyeE, D-CycE, DmCycE, DmcycE, DmcyclinE, Dmel\\CG3938, br37, cdi7, cycE, cyclinE, cycline, dCycE, dm-cycE, fond, l(2)05206, l(2)35Dd, l(2)br37, l(2)k02514, l(2)k02602, l(2)k05007, l35Dd, CycE', description='Cyclin E', other_designations='CG3938-PA|CG3938-PB|CG3938-PC|CG3938-PD|CG3938-PE|CG3938-PF|CycE-PA|CycE-PB|CycE-PC|CycE-PD|CycE-PE|CycE-PF|Cyclin-dependent kinase interactor 7|cyclin|cyclin E|cyclin-E|cyclinE|cycline E|dCyclinE|fondue', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15727581, end_position_on_the_genomic_accession=15748150, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34925, CurrentID=0, Status='live', Symbol='ND-B17', Aliases='Dmel_CG13240, BG:DS09217.1, BcDNA:GM23292, CG13240, Dmel\\CG13240, NB7M, l(2)35Di', description='NADH dehydrogenase (ubiquinone) B17 subunit', other_designations='CG13240-PA|CG13240-PC|NADH ubiquinone oxidoreductase B17|NADH-ubiquinone oxidoreductase B17 subunit|NADH:ubiquinone oxidoreductase B17 subunit|ND-B17-PA|ND-B17-PC|lethal (2) 35Di', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15749319, end_position_on_the_genomic_accession=15751078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34927, CurrentID=0, Status='live', Symbol='Gli', Aliases='Dmel_CG3903, BG:DS09217.3, CG3903, Dmel\\CG3903, gli, l(2)35Dg, l(2)br45, n(2)k09033', description='Gliotactin', other_designations='CG3903-PA|CG3903-PC|CG3903-PD|CG3903-PE|CG3903-PF|Gli-PA|Gli-PC|Gli-PD|Gli-PE|Gli-PF|gliotactin', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15756002, end_position_on_the_genomic_accession=15762758, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34928, CurrentID=0, Status='live', Symbol='CG3793', Aliases='Dmel_ BG:DS09217.4, Dmel\\CG3793', description='CG3793 gene product from transcript CG3793-RB', other_designations='CG3793-PA|CG3793-PB', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15764220, end_position_on_the_genomic_accession=15766267, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34930, CurrentID=0, Status='live', Symbol='Ku80', Aliases='Dmel_CG18801, BG:DS09217.6, CG18801, Dm Dmel\\CG18801, KU80, Ku, ku80', description='CG18801 gene product from transcript CG18801-RA', other_designations='CG18801-PA|Ku80-PA', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15768630, end_position_on_the_genomic_accession=15771159, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34931, CurrentID=0, Status='live', Symbol='CG18109', Aliases='Dmel_ BG:DS02252.3, DS02252.3, Dmel\\CG18109', description='CG18109 gene product from transcript CG18109-RA', other_designations='CG18109-PA|uncharacterized protein', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15784085, end_position_on_the_genomic_accession=15790146, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34932, CurrentID=0, Status='live', Symbol='CG18518', Aliases='Dmel_ BG:DS02252.4, DS02252.4, Dmel\\CG18518', description='CG18518 gene product from transcript CG18518-RA', other_designations='CG18518-PA|uncharacterized protein', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15834569, end_position_on_the_genomic_accession=15835658, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  34933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34933, CurrentID=0, Status='live', Symbol='CG13243', Aliases='Dmel_ BG:DS02252.1, Dmel\\CG13243', description='CG13243 gene product from transcript CG13243-RB', other_designations='CG13243-PA|CG13243-PB|uncharacterized protein', map_location='35D4-35D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15852592, end_position_on_the_genomic_accession=15854110, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34934, CurrentID=0, Status='live', Symbol='Tektin-A', Aliases='Dmel_CG4767, BG:DS02252.2, CG4767, Dmel\\CG4767, anon-WO0140519.235', description='Tektin A', other_designations='CG4767-PA|CG4767-PB|TEKTIN A|Tektin-A-PA|Tektin-A-PB|tektin A', map_location='35D5-35D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15854126, end_position_on_the_genomic_accession=15856450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34935, CurrentID=0, Status='live', Symbol='CG7653', Aliases='Dmel_ BG:DS00365.1, Dmel\\CG7653', description='CG7653 gene product from transcript CG7653-RB', other_designations='CG7653-PB', map_location='35D5-35D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15876938, end_position_on_the_genomic_accession=15880156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34937, CurrentID=0, Status='live', Symbol='Tep1', Aliases='Dmel_CG18096, BG:DS00365.2, CG18096, Dmel\\CG18096, TEP1, TEP1mel, Tep 1, Tep I, Tep-1, TepI, dTEPI, tep1', description='Thioester-containing protein 1', other_designations='CG18096-PA|TEP|Tep1-PA|thioester containing protein I|thiolester containing protein I', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15888640, end_position_on_the_genomic_accession=15893811, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34938, CurrentID=0, Status='live', Symbol='CG31821', Aliases='Dmel_ BG:DS00365.3, CG7647, Dmel\\CG31821', description='CG31821 gene product from transcript CG31821-RB', other_designations='CG31821-PA|CG31821-PB|uncharacterized protein', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15900935, end_position_on_the_genomic_accession=15902410, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34939, CurrentID=0, Status='live', Symbol='beat-Ib', Aliases='Dmel_CG7644, BG:DS00365.4, BcDNA:RE51929, Beat-Ib, CG7644, CT23325, Dmel\\CG7644, beat Ib, beat-B', description='beaten path Ib', other_designations='CG7644-PA|CG7644-PB|beat-Ib-PA|beat-Ib-PB', map_location='35D6-35D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15905134, end_position_on_the_genomic_accession=15962435, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34940, CurrentID=0, Status='live', Symbol='CG13244', Aliases='Dmel_ BcDNA:GH27987, Dmel\\CG13244', description='CG13244 gene product from transcript CG13244-RA', other_designations='CG13244-PA|uncharacterized protein', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15916888, end_position_on_the_genomic_accession=15917607, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34941, CurrentID=0, Status='live', Symbol='CG4793', Aliases='Dmel_ BG:DS07486.3, Dmel\\CG4793, SPH208', description='CG4793 gene product from transcript CG4793-RD', other_designations='CG4793-PD', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15923395, end_position_on_the_genomic_accession=15926287, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34942, CurrentID=0, Status='live', Symbol='CG12448', Aliases='Dmel_ BG:DS07486.4, Dmel\\CG12448', description='CG12448 gene product from transcript CG12448-RC', other_designations='CG12448-PC|uncharacterized protein', map_location='35D7-35D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15941441, end_position_on_the_genomic_accession=15942854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34943, CurrentID=0, Status='live', Symbol='CG18063', Aliases='Dmel_ BG:DS07486.5, DS07486.5, Dmel\\CG18063', description='CG18063 gene product from transcript CG18063-RC', other_designations='CG18063-PC|CG18063-PD|CG18063-PE|uncharacterized protein', map_location='35D7-35D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15957858, end_position_on_the_genomic_accession=15959431, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34944, CurrentID=0, Status='live', Symbol='CG10839', Aliases='Dmel_ BG:DS07486.2, Dmel\\CG10839', description='CG10839 gene product from transcript CG10839-RB', other_designations='CG10839-PA|CG10839-PB|CG10839-PC|uncharacterized protein', map_location='35E1-35E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15979486, end_position_on_the_genomic_accession=15981455, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34945, CurrentID=0, Status='live', Symbol='beat-Ic', Aliases='Dmel_CG4838, BG:DS00913.1, Beat-Ic, CG4838, CT15531, Dmel\\CG4838, beat Ic, beat-C', description='beaten path Ic', other_designations='CG4838-PA|CG4838-PB|CG4838-PC|beat-Ic-PA|beat-Ic-PB|beat-Ic-PC', map_location='35E1-35E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16000291, end_position_on_the_genomic_accession=16041703, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34946, CurrentID=0, Status='live', Symbol='BicC', Aliases='Dmel_CG4824, BG:DS00913.2, BICC, Bic-C, CG4824, Dmel\\CG4824, Dromel_CG4824_FBtr0080804_bicc_mORF, br34, l(2)br34', description='Bicaudal C', other_designations='BicC-PA|BicC-PB|BicC-PD|Bicaudal-C|CG4824-PA|CG4824-PB|CG4824-PD|bicaudal C', map_location='2-51 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16042047, end_position_on_the_genomic_accession=16048757, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  34947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34947, CurrentID=0, Status='live', Symbol='beat-Ia', Aliases='Dmel_CG4846, BG:DS00913.3, Beat Ia, Beat-Ia, CCT, CG4846, CT15567, Cct, Dmel\\CG4846, TRiC, beat, beat Ia, beat1a, l(2)35Ea, l(2)35Eb, l(2)br47, l(2)br48, tric', description='beaten path Ia', other_designations='CG4846-PA|CG4846-PB|beat-Ia-PA|beat-Ia-PB|beaten path|chaperonin containing TCP-1|triclops', map_location='35E2-35E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16049965, end_position_on_the_genomic_accession=16081684, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34948, CurrentID=0, Status='live', Symbol='CG4891', Aliases='Dmel_ BG:DS04095.1, DS04095.1, Dmel\\CG4891', description='CG4891 gene product from transcript CG4891-RA', other_designations='CG4891-PA', map_location='35E2-35E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16094997, end_position_on_the_genomic_accession=16095944, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34949, CurrentID=0, Status='live', Symbol='CG4892', Aliases='Dmel_ BG:DS04095.2, BcDNA:AT24169, DS04095.2, Dmel\\CG4892', description='CG4892 gene product from transcript CG4892-RA', other_designations='CG4892-PA', map_location='35E3-35E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16133508, end_position_on_the_genomic_accession=16134412, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34950, CurrentID=0, Status='live', Symbol='Ca-alpha1D', Aliases='Dmel_CG4894, BG:DS02795.1, CG4894, Ca-_1D, Ca-a1D, Ca-alpha1, Ca-alpha[[1]]D, Ca_alpha1D, Caalpha1D, DmCa1D, DmCa1alpha1, Dmca1D, Dmel\\CG4894, DroCa1, alpha1, alpha1D, ca-alpha1D, ca-alpha1d, cad, l(2)35Fa, l35Fa', description='Ca[2+]-channel protein alpha[[1]] subunit D', other_designations='CG4894-PA|CG4894-PB|CG4894-PC|CG4894-PD|CG4894-PE|CG4894-PF|CG4894-PG|CG4894-PH|CG4894-PI|CG4894-PJ|Ca-alpha1D-PA|Ca-alpha1D-PB|Ca-alpha1D-PC|Ca-alpha1D-PD|Ca-alpha1D-PE|Ca-alpha1D-PF|Ca-alpha1D-PG|Ca-alpha1D-PH|Ca-alpha1D-PI|Ca-alpha1D-PJ|Ca2+-channel protein alpha1 subunit D|Ca[2+]-channel alpha1 subunit|calcium channel alpha-1 subunit|drosophila melanogaster calcium channel alpha[[1]] subunit DHP-sensitive', map_location='35E5-35E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16165831, end_position_on_the_genomic_accession=16190540, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  34951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34951, CurrentID=0, Status='live', Symbol='CG42817', Aliases='Dmel_ BG:DS07473.1, CG12455, DS07473.1A, Dmel\\CG42817, Dmel_CG12455', description='CG42817 gene product from transcript CG42817-RB', other_designations='CG42817-PA|CG42817-PB|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16231475, end_position_on_the_genomic_accession=16241527, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  34952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34952, CurrentID=0, Status='live', Symbol='PRL-1', Aliases='Dmel_CG4993, BG:DS07473.3, BcDNA:RE40268, CG4993, Dmel\\CG4993, PRL, PRL1, PrL-1, dPRL-1', description='PRL-1 phosphatase', other_designations='CG4993-PA|CG4993-PB|CG4993-PC|CG4993-PD|PRL-1-PA|PRL-1-PB|PRL-1-PC|PRL-1-PD', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16245835, end_position_on_the_genomic_accession=16257606, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34953, CurrentID=0, Status='live', Symbol='EndoGI', Aliases='Dmel_CG4930, BG:DS07473.2, CG4930, Dmel\\CG4930, dEndoGI, endoGI', description='Endonuclease G inhibitor', other_designations='CG4930-PA|CG4930-PB|CG4930-PC|EndoGI-PA|EndoGI-PB|EndoGI-PC|endonuclease G inhibitor', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16257688, end_position_on_the_genomic_accession=16259511, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34954, CurrentID=0, Status='live', Symbol='twe', Aliases='Dmel_CG4965, BG:DS02740.1, CG4965, Cdc25, Dmel\\CG4965, TWINE, Twe, cdc25, l(2)35Fh, mat(2)syn-A, mat(2)synHB5, mat(2)syn[HB5], twn', description='twine', other_designations='CG4965-PA|CG4965-PB|CG4965-PC|CG4965-PD|twe-PA|twe-PB|twe-PC|twe-PD', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16259414, end_position_on_the_genomic_accession=16261877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34955, CurrentID=0, Status='live', Symbol='CG4935', Aliases='Dmel_ BG:DS02740.2, Dmel\\CG4935, MET30', description='CG4935 gene product from transcript CG4935-RA', other_designations='CG4935-PA', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16262141, end_position_on_the_genomic_accession=16263483, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  34956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34956, CurrentID=0, Status='live', Symbol='crp', Aliases='Dmel_CG7664, 35F(AT)[[17]], AP-4, AP4, BG:DS02740.3, CG7664, Dmel\\CG7664, MS:35F.AT17, bHLHe63, dAP-4, l(2)00232, l(2)35Fd, l(2)SH1614, l(2)SH2 1614, l(2)k00809, l(2)k03505', description='cropped', other_designations='CG7664-PA|CG7664-PB|crp-PA|crp-PB', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16263556, end_position_on_the_genomic_accession=16287730, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  34957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34957, CurrentID=0, Status='live', Symbol='pkaap', Aliases='Dmel_CG4132, BG:DS02740.4, CG4132, DS02740.4, Dmel\\CG4132', description='CG4132 gene product from transcript CG4132-RA', other_designations='CG4132-PA|pkaap-PA', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16288290, end_position_on_the_genomic_accession=16291195, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34958, CurrentID=0, Status='live', Symbol='CG17329', Aliases='Dmel_ BG:DS02740.5, BcDNA:AT18988, DS02740.5, Dmel\\CG17329', description='CG17329 gene product from transcript CG17329-RA', other_designations='CG17329-PA|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16291284, end_position_on_the_genomic_accession=16291948, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34959, CurrentID=0, Status='live', Symbol='CG13258', Aliases='Dmel_ Dmel\\CG13258', description='CG13258 gene product from transcript CG13258-RB', other_designations='CG13258-PA|CG13258-PB|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16292010, end_position_on_the_genomic_accession=16293025, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34961, CurrentID=0, Status='live', Symbol='heix', Aliases='Dmel_CG5876, 124/1, BG:DS02740.7, CG5876, Dmel\\CG5876, ipa-6d, l(2)35Fc, l(2)k12401', description='heixuedian', other_designations='CG5876-PA|heix-PA|lethal(2)35Fc', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16296051, end_position_on_the_genomic_accession=16299858, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34962, CurrentID=0, Status='live', Symbol='CG17328', Aliases='Dmel_ BG:DS02740.8, DS02740.8, Dmel\\CG17328', description='CG17328 gene product from transcript CG17328-RB', other_designations='CG17328-PA|CG17328-PB|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16299943, end_position_on_the_genomic_accession=16301765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34963, CurrentID=0, Status='live', Symbol='GMF', Aliases='Dmel_CG5869, BG:DS02740.9, BcDNA:SD03793, CG5869, DS02740.9, Dmel\\CG5869, dGMF, gmf', description='Glia maturation factor', other_designations='CG5869-PA|GMF-PA|glia maturation factor|glia-maturation factor', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16301656, end_position_on_the_genomic_accession=16302954, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34964, CurrentID=0, Status='live', Symbol='c(2)M', Aliases='Dmel_CG4249, BG:DS02740.10, C(2)M, CG4249, DS02740.10, Dmel\\CG4249, MEI-910, c(2)m, mei-910', description='crossover suppressor on 2 of Manheim', other_designations='CG4249-PA|alpha-kleisin|c(2)M-PA|crossover suppressor on 2 of manheim', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16303264, end_position_on_the_genomic_accession=16305940, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  34965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34965, CurrentID=0, Status='live', Symbol='CG5861', Aliases='Dmel_ BG:DS02740.11, Dmel\\CG5861, III, anon-35Fa, anon-35Fc', description='CG5861 gene product from transcript CG5861-RA', other_designations='CG5861-PA', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16305801, end_position_on_the_genomic_accession=16307298, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34966, CurrentID=0, Status='live', Symbol='Syx5', Aliases='Dmel_CG4214, BG:DS02740.12, CG4214, DmSyx5, Dmel\\CG4214, II, SED5, SYN5, Sed5, Synx5, SyxA, anon-35Fb, anon-35Fd, dSed5, dSed5p, dSyx5, dsed5, l(2)35Ff, sed5, syn5, syx5, Syx5', description='Syntaxin 5', other_designations='CG4214-PA|CG4214-PB|Syx5-PA|Syx5-PB|dSyntaxin5|lethal(2)35Ff|syntaxin|syntaxin 5|syntaxin5', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16307346, end_position_on_the_genomic_accession=16309274, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34967, CurrentID=0, Status='live', Symbol='cni', Aliases='Dmel_CG5855, BG:DS02740.13, CG5855, Cni, Dmel\\CG5855, IV, anon-35Fe, spnG', description='cornichon', other_designations='CG5855-PA|CG5855-PB|CG5855-PC|cni-PA|cni-PB|cni-PC|spindle-G', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16309292, end_position_on_the_genomic_accession=16310726, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  34968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34968, CurrentID=0, Status='live', Symbol='fzy', Aliases='Dmel_CG4274, BG:DS02740.14, CDC20, CG4274, Cdc20, Cdc20/Fizzy, Cdc20/Fzy, Cdc20[Fzy], Cdc20[fizzy], Dmel\\CG4274, FZY, Fizzy, Fzy, Fzy/Cdc20, cdc20, cdc20/fizzy, dCDC20, dFZY', description='fizzy', other_designations='CG4274-PA|Cdc20|fzy-PA', map_location='2-51 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16310956, end_position_on_the_genomic_accession=16313097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34969, CurrentID=0, Status='live', Symbol='cact', Aliases='Dmel_CG5848, 5848, BG:DS02740.15, CACT, CG5848, Cact, Dmel\\CG5848, Ikappa, NFkappabeta1A, cac, dip6, fs(2)ltoRN48, n(2)k17003', description='cactus', other_designations='CG5848-PA|CG5848-PB|CG5848-PC|CG5848-PD|CG5848-PE|NF-kappa-B inhibitor cactus|cact-PA|cact-PB|cact-PC|cact-PD|cact-PE|female-sterile-(2)-ltoRN48', map_location='2-52 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16313036, end_position_on_the_genomic_accession=16326207, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  34970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34970, CurrentID=0, Status='live', Symbol='CG4278', Aliases='Dmel_ BG:DS02740.16, Dmel\\CG4278, anon-35F/36A, un5', description='CG4278 gene product from transcript CG4278-RA', other_designations='CG4278-PA', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16327069, end_position_on_the_genomic_accession=16328131, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34974, CurrentID=0, Status='live', Symbol='chif', Aliases='Dmel_CG5813, BG:DS09218.1, BG:DS09218.2, CG5813, CG5817, Chif, DS09218.1, Dbf4, Dmel\\CG5813, chf, fs(2)chiffon', description='chiffon', other_designations='CG5813-PA|CG5813-PB|CG5813-PD|chif-PA|chif-PB|chif-PD', map_location='2-53 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16342430, end_position_on_the_genomic_accession=16352628, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34975, CurrentID=0, Status='live', Symbol='CG4455', Aliases='Dmel_ BEST:LD27171, BG:DS09218.3, Dmel\\CG4455, LD27171', description='CG4455 gene product from transcript CG4455-RA', other_designations='CG4455-PA', map_location='35F12-35F12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16352934, end_position_on_the_genomic_accession=16354539, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34976, CurrentID=0, Status='live', Symbol='CaBP1', Aliases='Dmel_CG5809, AAL28897, BG:DS09218.4, CG5809, DmCaBP1, Dmel\\CG5809', description='calcium-binding protein 1', other_designations='CG5809-PA|CG5809-PB|CaBP1-PA|CaBP1-PB', map_location='35F12-35F12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16354886, end_position_on_the_genomic_accession=16356735, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34977, CurrentID=0, Status='live', Symbol='jhamt', Aliases='Dmel_CG17330, BG:DS09218.5, CG17330, DmJHAMT, Dmel\\CG17330, JAHMT, JHAMT, JHamt, Jhamt', description='juvenile hormone acid methyltransferase', other_designations='CG17330-PA|CG17330-PB|jhamt-PA|jhamt-PB|juvenile hormone acid O-methyltransferase|juvenile hormone acid methyl transferase', map_location='35F12-35F12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16365903, end_position_on_the_genomic_accession=16367039, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34978, CurrentID=0, Status='live', Symbol='Idgf1', Aliases='Dmel_CG4472, BG:DS02780.5, CG4472, DmIDGF1, Dmel\\CG4472, IDGF, IDGF-1, IDGF1, idgf1, ldgf1', description='Imaginal disc growth factor 1', other_designations='CG4472-PA|Idgf1-PA|imaginal disc growth factor 1|imaginal disc growth factor1', map_location='36A1-36A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16445278, end_position_on_the_genomic_accession=16446793, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34979, CurrentID=0, Status='live', Symbol='Idgf2', Aliases='Dmel_CG4475, BG:DS02780.4, CG4475, DmIDGF2, Dmel\\CG4475, IDGF, IDGF-2, IDGF2, idgf2', description='Imaginal disc growth factor 2', other_designations='CG4475-PA|CG4475-PB|CG4475-PC|Idgf2-PA|Idgf2-PB|Idgf2-PC|imaginal disc growth factor 2', map_location='36A1-36A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16447330, end_position_on_the_genomic_accession=16450508, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34980, CurrentID=0, Status='live', Symbol='CG5888', Aliases='Dmel_ BG:DS02780.1, CT18477, Dmel\\CG5888', description='CG5888 gene product from transcript CG5888-RA', other_designations='CG5888-PA|CG5888-PB', map_location='35F12-36A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16400175, end_position_on_the_genomic_accession=16448469, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34981, CurrentID=0, Status='live', Symbol='Idgf3', Aliases='Dmel_CG4559, BG:DS02780.2, CG4559, DmIDGF3, Dmel\\CG4559, IDGF, IDGF-3, IDGF3, idgf3, lincRNA.163', description='Imaginal disc growth factor 3', other_designations='CG4559-PA|CG4559-PB|CG4559-PC|CG4559-PD|Idgf3-PA|Idgf3-PB|Idgf3-PC|Idgf3-PD|imaginal disc growth factor 3', map_location='36A1-36A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16450907, end_position_on_the_genomic_accession=16455194, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  34982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34982, CurrentID=0, Status='live', Symbol='dac', Aliases='Dmel_CG4952, BG:DS02780.3, CG4952, DAC, Dac, Dac2-3, Dach, Dm-DAC, Dmel\\CG4952h, l(2)36Ae, l(2)rK364, spn, dac', description='dachshund', other_designations='CG4952-PA|CG4952-PB|CG4952-PC|CG4952-PD|CG4952-PF|CG4952-PG|dac-PA|dac-PB|dac-PC|dac-PD|dac-PF|dac-PG|dachsund|dacshund|daschund|spinosus', map_location='36A1-36A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16466512, end_position_on_the_genomic_accession=16485998, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  34983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34983, CurrentID=0, Status='live', Symbol='CG4580', Aliases='Dmel_ Dmel\\CG4580', description='CG4580 gene product from transcript CG4580-RA', other_designations='CG4580-PA', map_location='36A2-36A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16488148, end_position_on_the_genomic_accession=16490182, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34984, CurrentID=0, Status='live', Symbol='Tpr2', Aliases='Dmel_CG4599, BEST:LD24941, CG4599, Dmel\\CG4599, EU3220, TPR2, Tpr-2, Trp2, dTPR2, dTpr2, dtpr2, dtrpr2, tpr2', description='Tetratricopeptide repeat protein 2', other_designations='CG4599-PA|CG4599-PB|CG4599-PC|CG4599-PD|CG4599-PE|Tpr2-PA|Tpr2-PB|Tpr2-PC|Tpr2-PD|Tpr2-PE|tetratricopeptide repeat protein 2', map_location='36A2-36A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16491540, end_position_on_the_genomic_accession=16507621, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  34985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34985, CurrentID=0, Status='live', Symbol='CG5953', Aliases='Dmel_ Dmel\\CG5953', description='CG5953 gene product from transcript CG5953-RD', other_designations='CG5953-PA|CG5953-PB|CG5953-PC|CG5953-PD', map_location='36A2-36A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16508078, end_position_on_the_genomic_accession=16532877, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  34986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34986, CurrentID=0, Status='live', Symbol='CG5968', Aliases='Dmel_ BcDNA:AT27976, Dmel\\CG5968', description='CG5968 gene product from transcript CG5968-RB', other_designations='CG5968-PA|CG5968-PB', map_location='36A4-36A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16549181, end_position_on_the_genomic_accession=16549959, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34987, CurrentID=0, Status='live', Symbol='CG42389', Aliases='Dmel_ CG13260, CG13261, CG31738, CG4668, CT15043, CT15049, Dmel\\CG42389, Dmel_CG13260, Dmel_CG31738, anon-EST:Liang-1.16, clone 1.16', description='CG42389 gene product from transcript CG42389-RE', other_designations='CG42389-PE|CG42389-PF|CG42389-PG|CG42389-PH|uncharacterized protein', map_location='36A7-36A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16545016, end_position_on_the_genomic_accession=16663218, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34988, CurrentID=0, Status='live', Symbol='CG4631', Aliases='Dmel_ Dmel\\CG4631', description='CG4631 gene product from transcript CG4631-RA', other_designations='CG4631-PA', map_location='36A7-36A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16607370, end_position_on_the_genomic_accession=16610037, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  34991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34991, CurrentID=0, Status='live', Symbol='Trpgamma', Aliases='Dmel_CG5996, CG5996, DmTRPgamma, Dmel\\CG5996, TRPgamma, Trp-gamma, dTRPg, dTRPgamma, trp-gamma, trpgamma', description='Transient receptor potential cation channel gamma', other_designations='CG5996-PA|CG5996-PB|CG5996-PD|Trpgamma-PA|Trpgamma-PB|Trpgamma-PD|transient receptor potential cation channel gamma|trpgamma', map_location='36A9-36A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16663454, end_position_on_the_genomic_accession=16676701, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  34992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34992, CurrentID=0, Status='live', Symbol='her', Aliases='Dmel_CG4694, CG4694, Dmel\\CG4694, FBgn 1185, l(2)mat', description='hermaphrodite', other_designations='CG4694-PA|CG4694-PB|her-PA|her-PB|lethal (2) maternal|lethal(2)maternal', map_location='2-52.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16677185, end_position_on_the_genomic_accession=16679691, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34993, CurrentID=0, Status='live', Symbol='grp', Aliases='Dmel_CG17161, 1C, CG17161, CHK1, CHK1/Grapes, Chk1, Dchk1, DmChk1, Dmel\\CG17161, Grp, Grp/DChk1, Pk36A, Pk?1, chk1, dChk1(chk), grps, l(2)k15102, lemp, grp', description='grapes', other_designations='CG17161-PA|CG17161-PB|CG17161-PC|CG17161-PD|CG17161-PF|Chk1/Grapes|checkpoint kinase 1|grp-PA|grp-PB|grp-PC|grp-PD|grp-PF|last empress|lethal (2) k15102|protein kinase-like 36A', map_location='36A10-36A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16680054, end_position_on_the_genomic_accession=16699926, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  34995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34995, CurrentID=0, Status='live', Symbol='Cyt-c-d', Aliases='Dmel_CG13263, CG13263, CYC, Cyt C, Cyt c, Cyt-c1, CytC, CytC-1, CytC1, Cytc, Cytc-d, DC3, Dmel\\CG13263, bln, bs30f07.y1, cyt c, cyt c-d, cyt-c-d, cyt.c, cyt.c-d, cytochrome c, dc3, ms(2)04445, pDMc02', description='Cytochrome c distal', other_designations='CG13263-PA|CG13263-PB|Cyt-c-d-PA|Cyt-c-d-PB|blanks|cytochrome c|cytochrome c distal|cytochrome c-d|cytochrome-c-d', map_location='36A11-36A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16715891, end_position_on_the_genomic_accession=16719635, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34996, CurrentID=0, Status='live', Symbol='Cyt-c-p', Aliases='Dmel_CG17903, CG17903, CYC, Cyct-c-p, Cyt C, Cyt c, Cyt-c2, CytC, CytC-2, Cytc, Cytc-p, DC4, DMc01, Dmel\\CG17903, cyt c, cyt c-p, cyt-c-p, cyt-c2, cyt.c, cytochrome c, dCyt.C-p, dc4, pDMc01', description='Cytochrome c proximal', other_designations='CG17903-PA|CG17903-PB|Cyt-c-p-PA|Cyt-c-p-PB|Cytochrome-c-proximal|cytochrome c|cytochrome c DC4|cytochrome c proximal|cytochrome-c-p', map_location='36A11-36A12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16719932, end_position_on_the_genomic_accession=16722644, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  34997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34997, CurrentID=0, Status='live', Symbol='VhaSFD', Aliases='Dmel_CG17332, 144191_at, ATP6V1H, CG17332, Dmel\\CG17332, V-ATPase, vhaSFD', description='Vacuolar H[+]-ATPase SFD subunit', other_designations='CG17332-PA|CG17332-PB|CG17332-PD|CG17332-PE|V-ATPase H-subunit|V-ATPase subunit H|V1 H subunit|VhaSFD-PA|VhaSFD-PB|VhaSFD-PD|VhaSFD-PE|vacuolar H+-ATPase SFD subunit|vacuolar H[+]-ATPase SFD subunit', map_location='36A12-36A12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16722883, end_position_on_the_genomic_accession=16727826, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  34998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34998, CurrentID=0, Status='live', Symbol='CG17996', Aliases='Dmel_ BcDNA:GM20557, Dmel\\CG17996', description='CG17996 gene product from transcript CG17996-RA', other_designations='CG17996-PA|uncharacterized protein', map_location='36A12-36A12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16730458, end_position_on_the_genomic_accession=16731039, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  34999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34999, CurrentID=0, Status='live', Symbol='Prosbeta4', Aliases='Dmel_CG17331, 17331, Beta-4, CG17331, Dmel\\CG17331, GC17331, PSMB2, Psmb2, beta-4, beta4', description='Proteasome beta4 subunit', other_designations='CG17331-PA|CG17331-PB|Prosbeta4-PA|Prosbeta4-PB|beta4|proteasome beta4 subunit', map_location='36A12-36A12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16731180, end_position_on_the_genomic_accession=16732133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35000, CurrentID=0, Status='live', Symbol='CG17904', Aliases='Dmel_ Dmel\\CG17904, Nubp1', description='CG17904 gene product from transcript CG17904-RA', other_designations='CG17904-PA|uncharacterized protein', map_location='36A12-36A12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16732550, end_position_on_the_genomic_accession=16733624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35001, CurrentID=0, Status='live', Symbol='glu', Aliases='Dmel_CG11397, BcDNA.LD20207, BcDNA:LD20207, CG11397, DmSMC4, DmSMC4/gluon, Dmel\\CG11397, Gluon, SMC4, Smc4, anon-WO0118547.156, dSMC4, l(2)k08819', description='gluon', other_designations='CG11397-PA|glu-PA', map_location='36A12-36A13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16733617, end_position_on_the_genomic_accession=16738388, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35002, CurrentID=0, Status='live', Symbol='ChLD3', Aliases='Dmel_CG17905, CG17905, DmCDA3, Dm Dmel\\CG17905', description='CG17905 gene product from transcript CG17905-RA', other_designations='CG17905-PA|ChLD3-PA', map_location='36A13-36A14', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16738850, end_position_on_the_genomic_accession=16742610, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35003, CurrentID=0, Status='live', Symbol='LSm7', Aliases='Dmel_CG13277, BcDNA:RH73529, CG13277, Dmel\\CG13277', description='CG13277 gene product from transcript CG13277-RC', other_designations='CG13277-PA|CG13277-PB|CG13277-PC|LSm7-PA|LSm7-PB|LSm7-PC', map_location='36A14-36A14', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16742315, end_position_on_the_genomic_accession=16743100, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35004, CurrentID=0, Status='live', Symbol='BuGZ', Aliases='Dmel_CG17912, BcDNA:LD27810, CG17912, Dm_2L:87597, Dmel\\CG17912, lincRNA.S1359', description='Bub3 interacting GLEBS and Zinc finger domain protein', other_designations='BuGZ-PA|BuGZ-PB|BuGZ-PC|BuGZ-PD|BuGZ-PE|BuGZ-PF|BuGZ-PG|Bub3 interacting GLEBS and zinc finger domain protein|CG17912-PA|CG17912-PB|CG17912-PC|CG17912-PD|CG17912-PE|CG17912-PF|CG17912-PG', map_location='36A14-36A14', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16743405, end_position_on_the_genomic_accession=16753704, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35005, CurrentID=0, Status='live', Symbol='yellow-b', Aliases='Dmel_CG17914, CG17914, Dmel\\CG17914, Yellow-b, yell-b', description='CG17914 gene product from transcript CG17914-RB', other_designations='CG17914-PA|CG17914-PB|Yellow-b|yellow-b-PA|yellow-b-PB', map_location='36A14-36A14', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16754336, end_position_on_the_genomic_accession=16758283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35006, CurrentID=0, Status='live', Symbol='ppk17', Aliases='Dmel_CG13278, CG13278, Dmel\\CG13278', description='pickpocket 17', other_designations='CG13278-PA|CG13278-PB|ppk17-PA|ppk17-PB', map_location='36A14-36A14', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16758184, end_position_on_the_genomic_accession=16760554, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35007, CurrentID=0, Status='live', Symbol='Mhc', Aliases='Dmel_CG17927, Bsh, CG17927, DROMHC, Dm II, DmMHC, Dmel\\CG17927, DroMII, FBgn0002741, Ifm(2)2, Ifm(2)RU1, MHC, MRP36B, Mrp, MyHC, Myo, Nup, Sht, Stp, chr2L:16765388..16765564, chr2L:16781880..16782040, ifm(2)2, ifm(2)RU1, ifm(2)RU2, l(2)36Ae, l(2)M66, l(2)k10423, mMHC, mhc, sMHC, Mhc', description='Myosin heavy chain', other_designations='CG17927-PA|CG17927-PB|CG17927-PC|CG17927-PD|CG17927-PE|CG17927-PF|CG17927-PG|CG17927-PH|CG17927-PI|CG17927-PK|CG17927-PL|CG17927-PM|CG17927-PN|CG17927-PO|CG17927-PP|CG17927-PQ|CG17927-PR|CG17927-PS|CG17927-PT|CG17927-PU|CG17927-PV|Mhc-PA|Mhc-PB|Mhc-PC|Mhc-PD|Mhc-PE|Mhc-PF|Mhc-PG|Mhc-PH|Mhc-PI|Mhc-PK|Mhc-PL|Mhc-PM|Mhc-PN|Mhc-PO|Mhc-PP|Mhc-PQ|Mhc-PR|Mhc-PS|Mhc-PT|Mhc-PU|Mhc-PV|Not-upheld|Shrunken-thorax|beta-myosin heavy chain|chain|heavy chain|indirect flight muscle (2) RU1|indirect flight muscle (2) RU2|muscle myosin|muscle myosin II|muscle myosin II heavy chain|muscle myosin heavy chain|muscle myosin-II|muscle specific myosin II|myo-II heavy chain|myosin|myosin heavy|myosin heavy Chain-PF|myosin heavy chain|myosin heavy-chain|myosin heavy-chain protein|myosin rod protein|samba|sarcomeric myosin heavy-chain|shrunken thorax|stuckup', map_location='36A8-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16766566, end_position_on_the_genomic_accession=16788766, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  35008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35008, CurrentID=0, Status='live', Symbol='Cyt-b5-r', Aliases='Dmel_CG13279, CG13279, Cyt-b, Cyt-b5, Cytb, Cytb5, Cytb5r, Dmel\\CG13279, TU-36B, TU36B', description='Cytochrome b5-related', other_designations='CG13279-PA|CG13279-PB|Cyt-b5-r-PA|Cyt-b5-r-PB|cytochrome b5-related', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16788774, end_position_on_the_genomic_accession=16790721, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35009, CurrentID=0, Status='live', Symbol='CG17928', Aliases='Dmel_ Dmel\\CG17928, anon-WO0118547.16', description='CG17928 gene product from transcript CG17928-RA', other_designations='CG17928-PA|CG17928-PB|uncharacterized protein', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16791568, end_position_on_the_genomic_accession=16793444, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38603, CurrentID=0, Status='live', Symbol='CG4603', Aliases='Dmel_ Dmel\\CG4603, Q9VRJ9, dYOD1', description='CG4603 gene product from transcript CG4603-RC', other_designations='CG4603-PA|CG4603-PB|CG4603-PC', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5133871, end_position_on_the_genomic_accession=5135670, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35013, CurrentID=0, Status='live', Symbol='CG13272', Aliases='Dmel_ CG33312, Dmel\\CG13272, Dmel_CG33312', description='CG13272 gene product from transcript CG13272-RB', other_designations='CG13272-PA|CG13272-PB|uncharacterized protein', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16804884, end_position_on_the_genomic_accession=16816069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35015, CurrentID=0, Status='live', Symbol='CG13280', Aliases='Dmel_ Dmel\\CG13280', description='CG13280 gene product from transcript CG13280-RA', other_designations='CG13280-PA|CG13280-PB|uncharacterized protein', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16802026, end_position_on_the_genomic_accession=16821093, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35016, CurrentID=0, Status='live', Symbol='Cas', Aliases='Dmel_CG13281, BcDNA:LD14270, CAS, CG13281, Dcas, Dmel\\CG13281, cas, dcas, l(2)k03902', description='CAS/CSE1 segregation protein', other_designations='CG13281-PA|Cas-PA|Importin-alpha re-exporter', map_location='36B1-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16821156, end_position_on_the_genomic_accession=16825207, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35019, CurrentID=0, Status='live', Symbol='CG13282', Aliases='Dmel_ Dmel\\CG13282', description='CG13282 gene product from transcript CG13282-RB', other_designations='CG13282-PA|CG13282-PB|uncharacterized protein', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16832065, end_position_on_the_genomic_accession=16834502, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35020, CurrentID=0, Status='live', Symbol='CG13283', Aliases='Dmel_ Dmel\\CG13283', description='CG13283 gene product from transcript CG13283-RA', other_designations='CG13283-PA|uncharacterized protein', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16835136, end_position_on_the_genomic_accession=16837258, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35021, CurrentID=0, Status='live', Symbol='CG13284', Aliases='Dmel_ Dmel\\CG13284', description='CG13284 gene product from transcript CG13284-RE', other_designations='CG13284-PA|CG13284-PB|CG13284-PC|CG13284-PD|CG13284-PE|uncharacterized protein', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16842300, end_position_on_the_genomic_accession=16848716, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35022, CurrentID=0, Status='live', Symbol='CG6012', Aliases='Dmel_ Dmel\\CG6012', description='CG6012 gene product from transcript CG6012-RA', other_designations='CG6012-PA', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16860029, end_position_on_the_genomic_accession=16861489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35026, CurrentID=0, Status='live', Symbol='tweek', Aliases='Dmel_CG42555, CG15133, CG15134, CG42555, CG42706, CG4841, Dmel\\CG42555, Dmel_CG15133, Dmel_CG15134, Dmel_CG4841, Tweek', description='CG42555 gene product from transcript CG42555-RJ', other_designations='CG42555-PF|CG42555-PG|CG42555-PH|CG42555-PI|CG42555-PJ|CG42555-PK|tweek-PF|tweek-PG|tweek-PH|tweek-PI|tweek-PJ|tweek-PK', map_location='36B4-36B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16888490, end_position_on_the_genomic_accession=16917052, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  35027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35027, CurrentID=0, Status='live', Symbol='CG12288', Aliases='Dmel_ Dmel\\CG12288, cg12288', description='CG12288 gene product from transcript CG12288-RA', other_designations='CG12288-PA|uncharacterized protein', map_location='36B4-36B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16900535, end_position_on_the_genomic_accession=16902218, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35029, CurrentID=0, Status='live', Symbol='ApepP', Aliases='Dmel_CG6291, AP-P, Apep[P], CG6291, DAP-P, Daminopep-p, Dmel\\CG6291', description='Aminopeptidase P', other_designations='ApepP-PA|ApepP-PB|CG6291-PA|CG6291-PB|aminopeptidase P', map_location='36B4-36B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16908229, end_position_on_the_genomic_accession=16910418, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35031, CurrentID=0, Status='live', Symbol='beat-IIIb', Aliases='Dmel_CG33179, Beat-IIIb, CG15135, CG33179, CG4855, Dmel\\CG33179, beat IIIb', description='CG33179 gene product from transcript CG33179-RC', other_designations='CG33179-PC|CG33179-PD|beat-IIIb-PC|beat-IIIb-PD', map_location='36B5-36B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16973091, end_position_on_the_genomic_accession=16993984, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35032, CurrentID=0, Status='live', Symbol='CG6304', Aliases='Dmel_ Dmel\\CG6304, anon-WO0140519.248', description='CG6304 gene product from transcript CG6304-RB', other_designations='CG6304-PA|CG6304-PB', map_location='36B6-36B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17011476, end_position_on_the_genomic_accession=17014700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35033, CurrentID=0, Status='live', Symbol='CG15136', Aliases='Dmel_ Dmel\\CG15136', description='CG15136 gene product from transcript CG15136-RB', other_designations='CG15136-PA|CG15136-PB|uncharacterized protein', map_location='36B6-36B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17047161, end_position_on_the_genomic_accession=17048281, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35034, CurrentID=0, Status='live', Symbol='CG12620', Aliases='Dmel_ Dmel\\CG12620', description='CG12620 gene product from transcript CG12620-RB', other_designations='CG12620-PA|CG12620-PB|uncharacterized protein', map_location='36B8-36B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17087641, end_position_on_the_genomic_accession=17089186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35035, CurrentID=0, Status='live', Symbol='beat-IIIa', Aliases='Dmel_CG12621, Beat-IIIa, CG12621, CG42757, Dmel\\CG12621, Dmel_CG42757, beat IIIa', description='CG12621 gene product from transcript CG12621-RE', other_designations='CG12621-PD|CG12621-PE|beat-IIIa-PD|beat-IIIa-PE', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17131024, end_position_on_the_genomic_accession=17173118, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35037, CurrentID=0, Status='live', Symbol='beat-IIIc', Aliases='Dmel_CG15138, BcDNA:RE14414, Beat-IIIC, Beat-IIIc, CG15138, CG6365, Dmel\\CG15138, beat IIIc', description='CG15138 gene product from transcript CG15138-RB', other_designations='Beat-IIIC|CG15138-PA|CG15138-PB|beat-IIIc-PA|beat-IIIc-PB', map_location='36C2-36C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17187637, end_position_on_the_genomic_accession=17260734, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35040, CurrentID=0, Status='live', Symbol='CG6380', Aliases='Dmel_ Dmel\\CG6380', description='CG6380 gene product from transcript CG6380-RB', other_designations='CG6380-PA|CG6380-PB', map_location='36C5-36C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17291075, end_position_on_the_genomic_accession=17292202, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35041, CurrentID=0, Status='live', Symbol='Lrch', Aliases='Dmel_CG6860, CG6860, CT21229, Dmel\\CG6860, dLRCH', description='Leucine-rich-repeats and calponin homology domain protein', other_designations='CG6860-PA|CG6860-PB|CG6860-PC|Lrch-PA|Lrch-PB|Lrch-PC', map_location='36C7-36C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17367511, end_position_on_the_genomic_accession=17384115, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35042, CurrentID=0, Status='live', Symbol='CLIP-190', Aliases='Dmel_CG5020, CG5020, CLIP190, Clip- 190, Clip-190, Clip190, D-CLIP-19, D-CLIP-190, D-CLIP-195, D-CLIP-90, D-Clip-190, Dmel\\CG5020', description='Cytoplasmic linker protein 190', other_designations='CG5020-PA|CG5020-PB|CG5020-PC|CG5020-PI|CG5020-PJ|CG5020-PM|CG5020-PN|CG5020-PO|CG5020-PP|CG5020-PQ|CG5020-PR|CG5020-PS|CG5020-PT|CG5020-PV|CLIP-190-PA|CLIP-190-PB|CLIP-190-PC|CLIP-190-PI|CLIP-190-PJ|CLIP-190-PM|CLIP-190-PN|CLIP-190-PO|CLIP-190-PP|CLIP-190-PQ|CLIP-190-PR|CLIP-190-PS|CLIP-190-PT|CLIP-190-PV|cytoplasmic linker protein 190|cytoplasmic linker protein-190', map_location='36C7-36C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17384700, end_position_on_the_genomic_accession=17409698, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  35043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35043, CurrentID=0, Status='live', Symbol='Rpb11', Aliases='Dmel_CG6840, BcDNA:GM15177, CG6840, Dmel\\CG6840, RPB11', description='CG6840 gene product from transcript CG6840-RA', other_designations='CG6840-PA|Rpb11-PA', map_location='36C7-36C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17409698, end_position_on_the_genomic_accession=17410406, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35044, CurrentID=0, Status='live', Symbol='CG15141', Aliases='Dmel_ BEST:LD24839, Dmel\\CG15141', description='CG15141 gene product from transcript CG15141-RA', other_designations='CG15141-PA|uncharacterized protein', map_location='36C7-36C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17410641, end_position_on_the_genomic_accession=17413499, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35045, CurrentID=0, Status='live', Symbol='Dif', Aliases='Dmel_CG6794, 6794, CG6794, DIF, Dmel\\CG6794, dif', description='Dorsal-related immunity factor', other_designations='CG6794-PA|CG6794-PB|CG6794-PC|CG6794-PD|Dif-PA|Dif-PB|Dif-PC|Dif-PD|Dorsal-like immunity factor|Dorsal-related transcription factor|dorsal related immunity factor|dorsal-related immunity factor', map_location='36C7-36C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17413248, end_position_on_the_genomic_accession=17433161, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35046, CurrentID=0, Status='live', Symbol='CG5043', Aliases='Dmel_ Dmel\\CG5043', description='CG5043 gene product from transcript CG5043-RA', other_designations='CG5043-PA', map_location='36C8-36C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17433862, end_position_on_the_genomic_accession=17435800, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35047, CurrentID=0, Status='live', Symbol='dl', Aliases='Dmel_CG6667, CG6667, DL, Dl, Dmel\\CG6667, Dor, Dorsal, GSd447, NFkappaB, REL, anon-EST:GressD7, d, dL, fs(2)k10816, mat(2)dorsal', description='dorsal', other_designations='CG6667-PA|CG6667-PB|CG6667-PC|CG6667-PD|CG6667-PE|CG6667-PF|NFkappabeta|dl-PA|dl-PB|dl-PC|dl-PD|dl-PE|dl-PF|embryonic polarity protein dorsal', map_location='2-52.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17436830, end_position_on_the_genomic_accession=17450364, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  35048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35048, CurrentID=0, Status='live', Symbol='CG5050', Aliases='Dmel_ Dmel\\CG5050', description='CG5050 gene product from transcript CG5050-RA', other_designations='CG5050-PA|CG5050-PB', map_location='36C9-36C9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17452924, end_position_on_the_genomic_accession=17458775, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35049, CurrentID=0, Status='live', Symbol='SPH93', Aliases='Dmel_CG6639, CG6639, Dmel\\CG6639, c-', description='CG6639 gene product from transcript CG6639-RA', other_designations='CG6639-PA|SPH93-PA', map_location='36C9-36C9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17454263, end_position_on_the_genomic_accession=17455923, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35050, CurrentID=0, Status='live', Symbol='CG18563', Aliases='Dmel_ Dmel\\CG18563, SPH156', description='CG18563 gene product from transcript CG18563-RB', other_designations='CG18563-PB|uncharacterized protein', map_location='36C9-36C9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17456670, end_position_on_the_genomic_accession=17458044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35051, CurrentID=0, Status='live', Symbol='BicD', Aliases='Dmel_CG6605, BIC-D, BICD, Bic-D, Bic[D], CG6605, Dmel\\CG6605, almond, anon-EST:fe2A11', description='Bicaudal D', other_designations='BicD-PA|BicD-PB|BicD-PC|BicD-PD|Bicaudal-D|BicaudalD|Bicuadal-D|CG6605-PA|CG6605-PB|CG6605-PC|CG6605-PD|anon-fast-evolving-2A11|bicaudal D', map_location='2-52.91 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17459605, end_position_on_the_genomic_accession=17473215, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35052, CurrentID=0, Status='live', Symbol='Sgt', Aliases='Dmel_CG5094, CG5094, Dmel\\CG5094, dsgt', description='small glutamine-rich tetratricopeptide containing protein', other_designations='CG5094-PA|CG5094-PB|Sgt-PA|Sgt-PB', map_location='36C9-36C9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17471603, end_position_on_the_genomic_accession=17474773, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35053, CurrentID=0, Status='live', Symbol='Aac11', Aliases='Dmel_CG6582, AAC-11, Api5, BcDNA.LD09852, BcDNA:LD09852, CG6582, Dmel\\CG6582, l(2)k06710', description='CG6582 gene product from transcript CG6582-RA', other_designations='Aac11-PA|CG6582-PA|antiapoptosis clone 11|antiapoptosis clone-11|apoptosis inhibitor-5', map_location='36C9-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17474912, end_position_on_the_genomic_accession=17477888, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35054, CurrentID=0, Status='live', Symbol='fws', Aliases='Dmel_CG6549, CG6549, COG5, Cog5, Dmel\\CG6549, FWS, Fws, anon-EST:Posey162, anon-EST:Posey186, dCOG5', description='four way stop', other_designations='CG6549-PA|CG6549-PC|CG6549-PD|CG6549-PH|CG6549-PI|Four-way stop|drosophila Cog5|four Way stop|fws-PA|fws-PC|fws-PD|fws-PH|fws-PI', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17478442, end_position_on_the_genomic_accession=17481029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35055, CurrentID=0, Status='live', Symbol='CG5110', Aliases='Dmel_ BcDNA:RE09173, Dmel\\CG5110, MP1, dMP1', description='CG5110 gene product from transcript CG5110-RB', other_designations='CG5110-PA|CG5110-PB', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17481231, end_position_on_the_genomic_accession=17482037, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35056, CurrentID=0, Status='live', Symbol='GCS2beta', Aliases='Dmel_CG6453, CG6453, Dmel\\CG6453, Q9VJD1_DROME', description='Glucosidase 2 beta subunit', other_designations='CG6453-PA|CG6453-PB|GCS2beta-PA|GCS2beta-PB|glucosidase 2 beta subunit', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17481988, end_position_on_the_genomic_accession=17484307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35057, CurrentID=0, Status='live', Symbol='CG5131', Aliases='Dmel_ Dmel\\CG5131', description='CG5131 gene product from transcript CG5131-RA', other_designations='CG5131-PA', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17484501, end_position_on_the_genomic_accession=17485624, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35058, CurrentID=0, Status='live', Symbol='qua', Aliases='Dmel_CG6433, CG6433, Dmel\\CG6433, Qua, anon-WO0140519.71, fs(2)A12', description='quail', other_designations='CG6433-PA|CG6433-PB|CG6433-PC|female sterile(2)A12|qua-PA|qua-PB|qua-PC|villin', map_location='2-53 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17486806, end_position_on_the_genomic_accession=17497208, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35059, CurrentID=0, Status='live', Symbol='CG15142', Aliases='Dmel_ Dmel\\CG15142', description='CG15142 gene product from transcript CG15142-RA', other_designations='CG15142-PA|uncharacterized protein', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17497994, end_position_on_the_genomic_accession=17498821, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35060, CurrentID=0, Status='live', Symbol='CG6412', Aliases='Dmel_ Dmel\\CG6412, mtEF-Ts', description='CG6412 gene product from transcript CG6412-RA', other_designations='CG6412-PA', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17499268, end_position_on_the_genomic_accession=17501106, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35061, CurrentID=0, Status='live', Symbol='Dhc36C', Aliases='Dmel_CG5526, CG5526, DHC, Dhc, Dhc36c, Dmel\\CG5526, dhc', description='Dynein heavy chain at 36C', other_designations='CG5526-PB|Dhc36C-PB|dynein heavy chain at 36C|dynein-related heavy chain polypeptide', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17501401, end_position_on_the_genomic_accession=17525515, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  35062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35062, CurrentID=0, Status='live', Symbol='CG15143', Aliases='Dmel_ Dmel\\CG15143', description='CG15143 gene product from transcript CG15143-RA', other_designations='CG15143-PA|uncharacterized protein', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17520892, end_position_on_the_genomic_accession=17524549, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35063, CurrentID=0, Status='live', Symbol='CG15144', Aliases='Dmel_ Dmel\\CG15144', description='CG15144 gene product from transcript CG15144-RB', other_designations='CG15144-PB|uncharacterized protein', map_location='36C10-36C11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17525585, end_position_on_the_genomic_accession=17529122, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35064, CurrentID=0, Status='live', Symbol='CG15145', Aliases='Dmel_ Dmel\\CG15145', description='CG15145 gene product from transcript CG15145-RA', other_designations='CG15145-PA|uncharacterized protein', map_location='36C11-36C11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17529311, end_position_on_the_genomic_accession=17532234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35065, CurrentID=0, Status='live', Symbol='CG7094', Aliases='Dmel_ Dmel\\CG7094', description='CG7094 gene product from transcript CG7094-RA', other_designations='CG7094-PA|CKIa-like', map_location='36D1-36D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17554354, end_position_on_the_genomic_accession=17555765, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35066, CurrentID=0, Status='live', Symbol='Oli', Aliases='Dmel_CG5545, CG5545, Dmel\\CG5545, Dolig, bHLHb23, oli, Oli', description='Olig family', other_designations='CG5545-PA|CG5545-PB|CG5545-PC|CG5545-PD|Oli-PA|Oli-PB|Oli-PC|Oli-PD|olig family', map_location='36D1-36D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17589184, end_position_on_the_genomic_accession=17591267, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35067, CurrentID=0, Status='live', Symbol='CG6870', Aliases='Dmel_ Dmel\\CG6870', description='CG6870 gene product from transcript CG6870-RB', other_designations='CG6870-PA|CG6870-PB|cytochrome B5', map_location='36D1-36D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17591315, end_position_on_the_genomic_accession=17592299, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35068, CurrentID=0, Status='live', Symbol='Sytalpha', Aliases='Dmel_CG5559, CG5559, Dmel\\CG5559, SYT-alpha, SYT3, Syt alpha, Syt3, dsytalpha, sytalpha', description='Synaptotagmin alpha', other_designations='CG5559-PA|CG5559-PB|CG5559-PC|CG5559-PD|Syt3|Sytalpha-PA|Sytalpha-PB|Sytalpha-PC|Sytalpha-PD|synaptotagmin alpha|sytalpha', map_location='36D-36D', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17592260, end_position_on_the_genomic_accession=17604387, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35069, CurrentID=0, Status='live', Symbol='CG15147', Aliases='Dmel_ Dmel\\CG15147', description='CG15147 gene product from transcript CG15147-RA', other_designations='CG15147-PA|uncharacterized protein', map_location='36D1-36D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17625746, end_position_on_the_genomic_accession=17626381, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35070, CurrentID=0, Status='live', Symbol='CadN', Aliases='Dmel_CG7100, B, CG7100, CT21941, Cad-N, D-cad, DN, DN-CAD, DN-CAd, DN-Cad, DN-cad, DNCad, DNcad, Dmel\\CG7100, N-Cad, N-Cadh, N-cad, N-cadherin, NCAD, NCad, Ncad, S(DmcycE[JP])2.12, anon-EST:CL32, cad1, cadN, l(2)36Da, l(2)Bld, n-cad, ncad', description='Cadherin-N', other_designations='CG7100-PA|CG7100-PB|CG7100-PC|CG7100-PD|CG7100-PE|CG7100-PF|CG7100-PG|CG7100-PH|CG7100-PI|CG7100-PJ|CG7100-PK|CG7100-PL|CadN-PA|CadN-PB|CadN-PC|CadN-PD|CadN-PE|CadN-PF|CadN-PG|CadN-PH|CadN-PI|CadN-PJ|CadN-PK|CadN-PL|CadherinN|DN cadherin|DN-cadherin|DN-cadherin1|NCadherin|drosophila neuronal cadherin|lethal(2)36Da|n-cadherin', map_location='2-53.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17647053, end_position_on_the_genomic_accession=17778340, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  35071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35071, CurrentID=0, Status='live', Symbol='CadN2', Aliases='Dmel_CG42829, CG42829, CG7527, CT23007, DN-Cad2, DN-cad2, DN-like, DNcad2, DmNCad2, Dmel\\CG42829, Dmel_CG7527, N-Cad2, NCad2, Ncad2, cadN2', description='Cadherin-N2', other_designations='CG42829-PC|CG42829-PD|CadN2-PC|CadN2-PD|DN-cadh-like|DN-cadherin|N-cadherin2|cadherin N2r', map_location='36D3-36E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17786971, end_position_on_the_genomic_accession=17875247, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  35072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35072, CurrentID=0, Status='live', Symbol='CG5674', Aliases='Dmel_ Dmel\\CG5674', description='CG5674 gene product from transcript CG5674-RB', other_designations='CG5674-PA|CG5674-PB|CG5674-PD|CG5674-PE', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17963379, end_position_on_the_genomic_accession=17974868, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35073, CurrentID=0, Status='live', Symbol='btv', Aliases='Dmel_CG15148, CG15148, DHC, DHC36D, Dhc36D, Dmel\\CG15148', description='beethoven', other_designations='CG15148-PD|DyneinHC|btv-PD|dynein heavy chain 36D', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17962625, end_position_on_the_genomic_accession=17982472, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  35074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35074, CurrentID=0, Status='live', Symbol='CG5681', Aliases='Dmel_ Dmel\\CG5681', description='CG5681 gene product from transcript CG5681-RA', other_designations='CG5681-PA', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17986399, end_position_on_the_genomic_accession=17988619, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35075, CurrentID=0, Status='live', Symbol='CG42635', Aliases='Dmel_ CG15149, Dmel\\CG42635, Dmel_CG15149', description='CG42635 gene product from transcript CG42635-RA', other_designations='CG42635-PA|CG42635-PB|uncharacterized protein', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18007935, end_position_on_the_genomic_accession=18008962, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35076, CurrentID=0, Status='live', Symbol='elfless', Aliases='Dmel_CG15150, BcDNA:AT24563, CG15150, Dmel\\CG15150', description='CG15150 gene product from transcript CG15150-RB', other_designations='CG15150-PB|CG15150-PC|elfless-PB|elfless-PC', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18009643, end_position_on_the_genomic_accession=18011475, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35077, CurrentID=0, Status='live', Symbol='rdo', Aliases='Dmel_CG15151, CG15151, CT35050, Dmel\\CG15151, PFE, Rdo', description='reduced ocelli', other_designations='CG15151-PA|CG15151-PB|pray For elves|rdo-PA|rdo-PB', map_location='2-53 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18012380, end_position_on_the_genomic_accession=18070246, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  35078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35078, CurrentID=0, Status='live', Symbol='Arr1', Aliases='Dmel_CG5711, 3H10, ARR1, Arr, ArrA, CG5711, DA, DAI, DS-Ag, DmArr1, Dmarrestin, Dmel\\CG5711, Dro.-Arr.I, FBgn0000120, PRII, arr1', description='Arrestin 1', other_designations='Arr1-PA|Arr1-PB|Arr1-PC|Arrestin1|CG5711-PA|CG5711-PB|CG5711-PC|Phosrestin-II|arrestin|arrestin 1|arrestin A|drosophila arrestin', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18078269, end_position_on_the_genomic_accession=18081307, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35079, CurrentID=0, Status='live', Symbol='CG31741', Aliases='Dmel_ CG5750, Dmel\\CG31741', description='CG31741 gene product from transcript CG31741-RA', other_designations='CG31741-PA|uncharacterized protein', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18084256, end_position_on_the_genomic_accession=18086714, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35080, CurrentID=0, Status='live', Symbol='CG15153', Aliases='Dmel_ Dmel\\CG15153', description='CG15153 gene product from transcript CG15153-RA', other_designations='CG15153-PA|uncharacterized protein', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18090737, end_position_on_the_genomic_accession=18092256, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35081, CurrentID=0, Status='live', Symbol='CG5755', Aliases='Dmel_ Dmel\\CG5755, anon-WO0140519.163', description='CG5755 gene product from transcript CG5755-RA', other_designations='CG5755-PA', map_location='36E4-36E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18096568, end_position_on_the_genomic_accession=18098120, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35082, CurrentID=0, Status='live', Symbol='CG15152', Aliases='Dmel_ Dmel\\CG15152', description='CG15152 gene product from transcript CG15152-RA', other_designations='CG15152-PA|CG15152-PB|uncharacterized protein', map_location='36E5-36E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18114547, end_position_on_the_genomic_accession=18115220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35083, CurrentID=0, Status='live', Symbol='CG5758', Aliases='Dmel_ Dmel\\CG5758', description='CG5758 gene product from transcript CG5758-RE', other_designations='CG5758-PA|CG5758-PB|CG5758-PC|CG5758-PD|CG5758-PE|CG5758-PF', map_location='36E5-36E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18119995, end_position_on_the_genomic_accession=18129862, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35084, CurrentID=0, Status='live', Symbol='kel', Aliases='Dmel_CG7210, CG7210, Dm.Ring_canal/KELC_DROME, Dmel\\CG7210, Kel, Kelch', description='kelch', other_designations='CG7210-PA|CG7210-PB|kel-PA|kel-PB', map_location='36E5-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18129884, end_position_on_the_genomic_accession=18138083, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35085, CurrentID=0, Status='live', Symbol='Socs36E', Aliases='Dmel_CG15154, CG15154, Dmel\\CG15154, SOCS, SOCS36E, SocS36E, Socs, anon-EST:Liang-2.2, clone 2.2, dSOCS, dmSocs36E, socs36E, socs36e, socs60E1', description='Suppressor of cytokine signaling at 36E', other_designations='CG15154-PA|CG15154-PC|CG15154-PD|Socs36E-PA|Socs36E-PC|Socs36E-PD|suppressor of cytokine signaling|suppressor of cytokine signaling 36E|suppressor of cytokine signaling at 36E', map_location='36E6-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18138675, end_position_on_the_genomic_accession=18152417, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35086, CurrentID=0, Status='live', Symbol='CG17681', Aliases='Dmel_ Dmel\\CG17681', description='CG17681 gene product from transcript CG17681-RA', other_designations='CG17681-PA|uncharacterized protein', map_location='36E6-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18153123, end_position_on_the_genomic_accession=18154516, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35087, CurrentID=0, Status='live', Symbol='CG15155', Aliases='Dmel_ CT35055, Dmel\\CG15155', description='CG15155 gene product from transcript CG15155-RA', other_designations='CG15155-PA|CG15155-PB|uncharacterized protein', map_location='36E6-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18154851, end_position_on_the_genomic_accession=18155942, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35088, CurrentID=0, Status='live', Symbol='CG5783', Aliases='Dmel_ Dmel\\CG5783', description='CG5783 gene product from transcript CG5783-RA', other_designations='CG5783-PA', map_location='36E6-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18156043, end_position_on_the_genomic_accession=18157162, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740248, CurrentID=0, Status='live', Symbol='robls54B', Aliases='Dmel_CG34192, CG34192, Dmel\\CG34192, GENE 3, HDC06929, gene 3', description='roadblock similar 54B', other_designations='CG34192-PA|CG34192-PB|robls54B-PA|robls54B-PB', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17455743, end_position_on_the_genomic_accession=17456441, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35090, CurrentID=0, Status='live', Symbol='CG42750', Aliases='Dmel_ CG15157, CG15158, CG31746, CG31786, CG42400, CG7167, CG7198, Dmel\\CG42750, Dmel_CG15157, Dmel_CG31746, Dmel_CG31786, Dmel_CG42400', description='CG42750 gene product from transcript CG42750-RC', other_designations='CG42750-PB|CG42750-PC|uncharacterized protein', map_location='36F1-36F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18159811, end_position_on_the_genomic_accession=18275770, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  35091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35091, CurrentID=0, Status='live', Symbol='Ptp36E', Aliases='Dmel_CG7180, CG7180, DPTP36E, Dmel\\CG7180, R-PTPk', description='Protein tyrosine phosphatase 36E', other_designations='CG7180-PA|CG7180-PB|CG7180-PC|Ptp36E-PA|Ptp36E-PB|Ptp36E-PC|protein tyrosine phosphatase 36E', map_location='36E6-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18180884, end_position_on_the_genomic_accession=18185007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35095, CurrentID=0, Status='live', Symbol='CG5790', Aliases='Dmel_ Dmel\\CG5790', description='CG5790 gene product from transcript CG5790-RB', other_designations='CDC 7-related|CG5790-PA|CG5790-PB', map_location='36F2-36F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18294840, end_position_on_the_genomic_accession=18297347, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35096, CurrentID=0, Status='live', Symbol='Acp36DE', Aliases='Dmel_CG7157, Acp, Acp36D, CG7157, CT22109, Dmel\\CG7157, anon-WO0140519.55', description='Accessory gland protein 36DE', other_designations='36DE accessory gland protein|Acp36DE-PA|Acp36DE-PB|CG7157-PA|CG7157-PB|accessory gland peptide 36DE|accessory gland protein|accessory gland protein 36DE', map_location='36F3-36F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18356047, end_position_on_the_genomic_accession=18359026, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35097, CurrentID=0, Status='live', Symbol='Fas3', Aliases='Dmel_CG5803, CG5803, Dmel\\CG5803, FAS III, FAS3, FASIII, Fas, Fas III, Fas-3, Fas-III, FasIII, FasIIIface, FascIII, fas III, fas-III, fas3, fasIII, fascIII', description='Fasciclin 3', other_designations='CG5803-PA|CG5803-PB|CG5803-PC|CG5803-PD|CG5803-PE|CG5803-PF|CG5803-PG|Fas3-PA|Fas3-PB|Fas3-PC|Fas3-PD|Fas3-PE|Fas3-PF|Fas3-PG|Fasciclin-3|Fasciclin-III|Fasciclin3|FasciclinIII|FsciclinIII|fascicilin III|fasciclin|fasciclin 3|fasciclin III|fascilin III|fasiciclin III|fasicilin III|fasiclin III', map_location='36F2-36F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18320102, end_position_on_the_genomic_accession=18393441, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  35098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35098, CurrentID=0, Status='live', Symbol='RpS26', Aliases='Dmel_CG10305, CG10305, Dmel\\CG10305, M(2)36F, M(2)S13, M(2)m, M[[s]], Rps26, anon-EST:fe2A3, l(2)04553, l(2)36Fa, rpS26, rpS31', description='Ribosomal protein S26', other_designations='CG10305-PA|CG10305-PB|CG10305-PC|CG10305-PD|CG10305-PE|RpS26-PA|RpS26-PB|RpS26-PC|RpS26-PD|RpS26-PE|anon-fast-evolving-2A3|minute (2) 36F|ribosomal protein S26|ribosomal protein S31', map_location='36F2-36F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18441335, end_position_on_the_genomic_accession=18443277, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35099, CurrentID=0, Status='live', Symbol='ncm', Aliases='Dmel_CG12750, CG12750, Dmel\\CG12750, E-2a, l(2)SH0931, l(2)SH2 0931', description='nucampholin', other_designations='CG12750-PA|lethal (2) SH0931|ncm-PA', map_location='36F5-36F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18444507, end_position_on_the_genomic_accession=18448957, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35100, CurrentID=0, Status='live', Symbol='bsf', Aliases='Dmel_CG10302, BSF, CG10302, DmLrpprc1, Dmel\\CG10302', description='bicoid stability factor', other_designations='CG10302-PA|bicoid stabilization factor|bsf-PA', map_location='36F5-36F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18449517, end_position_on_the_genomic_accession=18454587, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35101, CurrentID=0, Status='live', Symbol='Ntf-2r', Aliases='Dmel_CG10174, CG10174, Dmel\\CG10174, Dntf-2, Dntf-2R, Dntf-2r', description='Nuclear transport factor-2-related', other_designations='CG10174-PA|Ntf-2r-PA|drosophila nuclear transport factor-2-related|nuclear transport factor-2-related', map_location='36F5-36F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18454649, end_position_on_the_genomic_accession=18455419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35102, CurrentID=0, Status='live', Symbol='CG10283', Aliases='Dmel_ Dmel\\CG10283', description='CG10283 gene product from transcript CG10283-RD', other_designations='CG10283-PA|CG10283-PB|CG10283-PC|CG10283-PD|uncharacterized protein', map_location='36F5-36F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18472964, end_position_on_the_genomic_accession=18484513, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35103, CurrentID=0, Status='live', Symbol='CG10176', Aliases='Dmel_ Dmel\\CG10176', description='CG10176 gene product from transcript CG10176-RC', other_designations='CG10176-PB|CG10176-PC|uncharacterized protein', map_location='36F5-36F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18485417, end_position_on_the_genomic_accession=18488021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35104, CurrentID=0, Status='live', Symbol='kon', Aliases='Dmel_CG10275, CG10275, CT28869, Dmel\\CG10275, Kon, perd', description='kon-tiki', other_designations='CG10275-PB|CG10275-PC|CG10275-PD|NG2-like|kon-PB|kon-PC|kon-PD|perdido', map_location='36F5-36F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18487747, end_position_on_the_genomic_accession=18503978, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  35105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35105, CurrentID=0, Status='live', Symbol='CG10178', Aliases='Dmel_ Dmel\\CG10178', description='CG10178 gene product from transcript CG10178-RB', other_designations='CG10178-PA|CG10178-PB|uncharacterized protein', map_location='36F6-36F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18509560, end_position_on_the_genomic_accession=18513512, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35106, CurrentID=0, Status='live', Symbol='CG10211', Aliases='Dmel_ Dmel\\CG10211', description='CG10211 gene product from transcript CG10211-RB', other_designations='CG10211-PA|CG10211-PB|uncharacterized protein', map_location='36F6-36F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18517487, end_position_on_the_genomic_accession=18531967, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35107, CurrentID=0, Status='live', Symbol='Pde11', Aliases='Dmel_CG34341, CG10231, CG15159, CG34341, Dmel\\CG34341, Dmel_CG10231, Dmel_CG15159, PDE11', description='Phosphodiesterase 11', other_designations='CG34341-PB|CG34341-PC|CG34341-PD|CG34341-PE|CG34341-PF|CG34341-PG|Pde11-PB|Pde11-PC|Pde11-PD|Pde11-PE|Pde11-PF|Pde11-PG|phosphodiesterase 11', map_location='36F6-36F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18532025, end_position_on_the_genomic_accession=18590342, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  35109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35109, CurrentID=0, Status='live', Symbol='CG15160', Aliases='Dmel_ Dmel\\CG15160', description='CG15160 gene product from transcript CG15160-RA', other_designations='CG15160-PA|CG15160-PB|CG15160-PC|uncharacterized protein', map_location='36F6-36F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18590891, end_position_on_the_genomic_accession=18595429, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35110, CurrentID=0, Status='live', Symbol='amos', Aliases='Dmel_CG10393, Amos, CG10393, Dmel\\CG10393, Roi, Tft, asolo, bHLHa10, rolo', description='absent MD neurons and olfactory sensilla', other_designations='CG10393-PA|Rough-eye|absent MD neurons|amos-PA|rough eye|tufted', map_location='2-53.6 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18596136, end_position_on_the_genomic_accession=18596921, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35111, CurrentID=0, Status='live', Symbol='CG10413', Aliases='Dmel_ Dmel\\CG10413', description='CG10413 gene product from transcript CG10413-RB', other_designations='CG10413-PA|CG10413-PB|uncharacterized protein', map_location='36F7-36F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18599755, end_position_on_the_genomic_accession=18605249, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35112, CurrentID=0, Status='live', Symbol='CG10333', Aliases='Dmel_ DmPrp28, DmRH19, Dmel\\CG10333, cg10333', description='CG10333 gene product from transcript CG10333-RA', other_designations='CG10333-PA|uncharacterized protein', map_location='36F7-36F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18606832, end_position_on_the_genomic_accession=18609550, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740252, CurrentID=0, Status='live', Symbol='Not1', Aliases='Dmel_CG34407, CG1874, CG1884, CG34407, Dmel\\CG34407, Dmel_CG1874, Dmel_CG1884, NOT, NOT1, anon-EST:Liang-1.29, anon-WO0118547.221, clone 1.29, not1', description='CG34407 gene product from transcript CG34407-RG', other_designations='CG34407-PC|CG34407-PD|CG34407-PE|CG34407-PF|CG34407-PG|CG34407-PH|Not1-PC|Not1-PD|Not1-PE|Not1-PF|Not1-PG|Not1-PH', map_location='45F4-45F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9566146, end_position_on_the_genomic_accession=9579007, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  35114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35114, CurrentID=0, Status='live', Symbol='CG15161', Aliases='Dmel_ Dmel\\CG15161, IFT46', description='CG15161 gene product from transcript CG15161-RA', other_designations='CG15161-PA|uncharacterized protein', map_location='36F7-36F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18614339, end_position_on_the_genomic_accession=18616106, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35115, CurrentID=0, Status='live', Symbol='Cyp310a1', Aliases='Dmel_CG10391, 310a1, CG10391, Cyp310, Cyp310a, Dmel\\CG10391, H, anon-EST:CL1d7, cyp310a1', description='CG10391 gene product from transcript CG10391-RA', other_designations='CG10391-PA|Cyp310a1-PA|cytochrome P450', map_location='36F8-36F8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18651058, end_position_on_the_genomic_accession=18653010, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35116, CurrentID=0, Status='live', Symbol='MESR3', Aliases='Dmel_CG15162, CG15162, Dmel\\CG15162', description='Misexpression suppressor of ras 3', other_designations='CG15162-PA|CG15162-PB|CG15162-PC|CG15162-PD|MESR3-PA|MESR3-PB|MESR3-PC|MESR3-PD|mis-expression suppressor of ras 3|misexpression suppressor of ras 3', map_location='36F7-36F9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18617256, end_position_on_the_genomic_accession=18661776, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35117, CurrentID=0, Status='live', Symbol='CG43338', Aliases='Dmel_ CG15165, CG31790, CG42627, Dmel\\CG43338, Dmel_CG31790', description='CG43338 gene product from transcript CG43338-RB', other_designations='CG43338-PA|CG43338-PB|uncharacterized protein', map_location='36F10-36F10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18666892, end_position_on_the_genomic_accession=18671246, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35118, CurrentID=0, Status='live', Symbol='CG31751', Aliases='Dmel_ CG15163, CG15164, Dmel\\CG31751, anon-ct, ct', description='CG31751 gene product from transcript CG31751-RA', other_designations='CG31751-PA|CG31751-PB|CG31751-PC|constitutive transcript|uncharacterized protein', map_location='36F10-36F11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18677221, end_position_on_the_genomic_accession=18683715, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35119, CurrentID=0, Status='live', Symbol='tos', Aliases='Dmel_CG10387, CG10387, DmTosca, Dmel\\CG10387, Tosca', description='tosca', other_designations='CG10387-PA|tos-PA', map_location='36F11-36F11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18679631, end_position_on_the_genomic_accession=18682075, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35121, CurrentID=0, Status='live', Symbol='msl-1', Aliases='Dmel_CG10385, CG10385, Dmel\\CG10385, MSL, MSL-1, MSL1, MSL3, Msl 1-3, Msl-1, Msl1, i94, km(2)B, kmB, mls-1, msl, msl1', description='male-specific lethal 1', other_designations='CG10385-PA|CG10385-PB|killer-of-males-B|male specifc lethal|male specific lethal 1|msl-1-PA|msl-1-PB', map_location='2-20 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18683720, end_position_on_the_genomic_accession=18688800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35122, CurrentID=0, Status='live', Symbol='CG10336', Aliases='Dmel_ Dmel\\CG10336', description='CG10336 gene product from transcript CG10336-RC', other_designations='CG10336-PA|CG10336-PC|uncharacterized protein', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18689528, end_position_on_the_genomic_accession=18690906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35123, CurrentID=0, Status='live', Symbol='CG10383', Aliases='Dmel_ Dmel\\CG10383', description='CG10383 gene product from transcript CG10383-RB', other_designations='CG10383-PA|CG10383-PB|uncharacterized protein', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18690904, end_position_on_the_genomic_accession=18695602, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35124, CurrentID=0, Status='live', Symbol='CG10338', Aliases='Dmel_ Dmel\\CG10338', description='CG10338 gene product from transcript CG10338-RB', other_designations='CG10338-PA|CG10338-PB|uncharacterized protein', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18695664, end_position_on_the_genomic_accession=18698571, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35125, CurrentID=0, Status='live', Symbol='CG10341', Aliases='Dmel_ Dmel\\CG10341', description='CG10341 gene product from transcript CG10341-RA', other_designations='CG10341-PA|CG10341-PC|uncharacterized protein', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18698714, end_position_on_the_genomic_accession=18700645, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35126, CurrentID=0, Status='live', Symbol='CG10376', Aliases='Dmel_ Dmel\\CG10376, PP2C', description='CG10376 gene product from transcript CG10376-RA', other_designations='CG10376-PA|uncharacterized protein', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18700606, end_position_on_the_genomic_accession=18703105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35127, CurrentID=0, Status='live', Symbol='CG10343', Aliases='Dmel_ Dmel\\CG10343, anon-WO0118547.123', description='CG10343 gene product from transcript CG10343-RA', other_designations='CG10343-PA|uncharacterized protein', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18703473, end_position_on_the_genomic_accession=18704454, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35128, CurrentID=0, Status='live', Symbol='Jwa', Aliases='Dmel_CG10373, CG10373, Dmel\\CG10373, GH05842p, djwa', description='Jwa ortholog', other_designations='CG10373-PB|Jwa-PB', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18704570, end_position_on_the_genomic_accession=18705703, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35129, CurrentID=0, Status='live', Symbol='Faf2', Aliases='Dmel_CG10372, CG10372, Dmel\\CG10372, FFAF, Faf, GT#49', description='Fas-associated factor', other_designations='CG10372-PA|CG10372-PB|Faf2-PA|Faf2-PB|Fly Fas-associated factor|UBX domain 8 like', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18706010, end_position_on_the_genomic_accession=18708688, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35130, CurrentID=0, Status='live', Symbol='Grip71', Aliases='Dmel_CG10346, CG10346, Dgp71WD, Dgrip71, Dmel\\CG10346wd', description='CG10346 gene product from transcript CG10346-RA', other_designations='CG10346-PA|Grip71-PA', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18708846, end_position_on_the_genomic_accession=18711297, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35131, CurrentID=0, Status='live', Symbol='Irk3', Aliases='Dmel_CG10369, CG10369, Dmel\\CG10369, dKirIII, irk3', description='Inwardly rectifying potassium channel 3', other_designations='CG10369-PA|CG10369-PC|CG10369-PD|Irk3-PA|Irk3-PC|Irk3-PD|inwardly rectifying potassium channel 3', map_location='37A1-37A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18711071, end_position_on_the_genomic_accession=18713748, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35132, CurrentID=0, Status='live', Symbol='CG10348', Aliases='Dmel_ Dmel\\CG10348, lincRNA.186', description='CG10348 gene product from transcript CG10348-RC', other_designations='CG10348-PB|CG10348-PC|uncharacterized protein', map_location='37A1-37A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18714453, end_position_on_the_genomic_accession=18732524, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35134, CurrentID=0, Status='live', Symbol='CG15167', Aliases='Dmel_ Dmel\\CG15167', description='CG15167 gene product from transcript CG15167-RA', other_designations='CG15167-PA|uncharacterized protein', map_location='37A2-37A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18740358, end_position_on_the_genomic_accession=18740807, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35135, CurrentID=0, Status='live', Symbol='ham', Aliases='Dmel_CG31753, CG10568, CG15906, CG15907, CG31753, Dmel\\CG31753, HAM_DROME, Ham', description='hamlet', other_designations='CG31753-PB|CG31753-PC|CG31753-PD|CG31753-PE|CG31753-PF|ham-PB|ham-PC|ham-PD|ham-PE|ham-PF', map_location='37A2-37A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18762874, end_position_on_the_genomic_accession=18792906, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35137, CurrentID=0, Status='live', Symbol='CG17324', Aliases='Dmel_ BEST:GH06505, Dmel\\CG17324, GH06505', description='CG17324 gene product from transcript CG17324-RB', other_designations='CG17324-PB|uncharacterized protein', map_location='37A6-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18819344, end_position_on_the_genomic_accession=18822573, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35138, CurrentID=0, Status='live', Symbol='CG17323', Aliases='Dmel_ CT35071, Dmel\\CG17323', description='CG17323 gene product from transcript CG17323-RD', other_designations='CG17323-PA|CG17323-PC|CG17323-PD|CG17323-PE|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18823548, end_position_on_the_genomic_accession=18826716, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35139, CurrentID=0, Status='live', Symbol='CG17322', Aliases='Dmel_ BEST:LD25345, Dmel\\CG17322, LD25345', description='CG17322 gene product from transcript CG17322-RB', other_designations='CG17322-PA|CG17322-PB|CG17322-PC|CG17322-PD|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18826770, end_position_on_the_genomic_accession=18829059, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35140, CurrentID=0, Status='live', Symbol='CG17597', Aliases='Dmel_ Dmel\\CG17597', description='CG17597 gene product from transcript CG17597-RB', other_designations='CG17597-PA|CG17597-PB|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18829012, end_position_on_the_genomic_accession=18831153, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35142, CurrentID=0, Status='live', Symbol='CG10600', Aliases='Dmel_ Dmel\\CG10600', description='CG10600 gene product from transcript CG10600-RC', other_designations='CG10600-PA|CG10600-PB|CG10600-PC|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18833710, end_position_on_the_genomic_accession=18839679, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35143, CurrentID=0, Status='live', Symbol='CG31752', Aliases='Dmel_ CG10599, Dmel\\CG31752', description='CG31752 gene product from transcript CG31752-RB', other_designations='CG31752-PB|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18839628, end_position_on_the_genomic_accession=18841310, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35144, CurrentID=0, Status='live', Symbol='CG17321', Aliases='Dmel_ Dmel\\CG17321', description='CG17321 gene product from transcript CG17321-RB', other_designations='CG17321-PA|CG17321-PB|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18845184, end_position_on_the_genomic_accession=18855110, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35145, CurrentID=0, Status='live', Symbol='mRpL13', Aliases='Dmel_CG10603, CG10602, CG10603, Dmel\\CG10603, L13, anon-EST:fe3C4', description='mitochondrial ribosomal protein L13', other_designations='CG10603-PA|anon-fast-evolving-3C4|mRpL13-PA', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18858028, end_position_on_the_genomic_accession=18859248, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35146, CurrentID=0, Status='live', Symbol='CG10602', Aliases='Dmel_ Dmel\\CG10602, mRpL13, pepCG10602', description='CG10602 gene product from transcript CG10602-RF', other_designations='CG10602-PC|CG10602-PD|CG10602-PF|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18855126, end_position_on_the_genomic_accession=18859248, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35147, CurrentID=0, Status='live', Symbol='tup', Aliases='Dmel_CG10619, CG10619, Dmel\\CG10619, Isl, Islet, Tup, isl, islet, l(2)37Aa, l(2)E41', description='tailup', other_designations='CG10619-PA|CG10619-PB|CG10619-PC|dIslet|islet|islet 1|lethal(2)37Aa|tail-up|tailup/islet|tup-PA|tup-PB|tup-PC', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18859507, end_position_on_the_genomic_accession=18881261, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35149, CurrentID=0, Status='live', Symbol='ssp3', Aliases='Dmel_CG18397, CG10666, CG18397, Dmel\\CG18397, Ssp3', description='short spindle 3', other_designations='CG18397-PA|CG18397-PB|ssp3-PA|ssp3-PB', map_location='37B5-37B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18916347, end_position_on_the_genomic_accession=18943960, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35150, CurrentID=0, Status='live', Symbol='CG10428', Aliases='Dmel_ Dmel\\CG10428', description='CG10428 gene product from transcript CG10428-RA', other_designations='CG10428-PA|uncharacterized protein', map_location='37B6-37B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18948884, end_position_on_the_genomic_accession=18950973, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35151, CurrentID=0, Status='live', Symbol='Nedd8', Aliases='Dmel_CG10679, CG10679, Dmel\\CG10679, NEDD8, dNEDD8, nedd8', description='CG10679 gene product from transcript CG10679-RB', other_designations='CG10679-PA|CG10679-PB|Nedd8-PA|Nedd8-PB|neural precursor cell-expressed developmentally downregulated', map_location='37B7-37B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18951006, end_position_on_the_genomic_accession=18951837, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35152, CurrentID=0, Status='live', Symbol='CG10621', Aliases='Dmel_ BEST:LD26123, Dmel\\CG10621, anon-WO0140519.13', description='CG10621 gene product from transcript CG10621-RA', other_designations='CG10621-PA|CG10621-PB|CG10621-PC|selenocysteine methyltransferase|uncharacterized protein', map_location='37B7-37B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18952206, end_position_on_the_genomic_accession=18954431, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35153, CurrentID=0, Status='live', Symbol='CG10623', Aliases='Dmel_ Dmel\\CG10623', description='CG10623 gene product from transcript CG10623-RC', other_designations='CG10623-PA|CG10623-PB|CG10623-PC|selenocysteine methyltransferase|uncharacterized protein', map_location='37B7-37B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18955481, end_position_on_the_genomic_accession=18957311, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35154, CurrentID=0, Status='live', Symbol='Nak', Aliases='Dmel_CG10637, CG10637, Dmel\\CG10637, nak', description='Numb-associated kinase', other_designations='CG10637-PA|CG10637-PC|CG10637-PD|CG10637-PE|Nak-PA|Nak-PC|Nak-PD|Nak-PE', map_location='37B7-37B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18957732, end_position_on_the_genomic_accession=18968744, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35155, CurrentID=0, Status='live', Symbol='Tango6', Aliases='Dmel_CG18398, CG18398, Dmel\\CG18398, TANGO6', description='Transport and Golgi organization 6', other_designations='CG18398-PA|Tango6-PA|transport and golgi organization 6', map_location='37B7-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18969028, end_position_on_the_genomic_accession=18972004, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35156, CurrentID=0, Status='live', Symbol='CG10639', Aliases='Dmel_ Dmel\\CG10639', description='CG10639 gene product from transcript CG10639-RA', other_designations='CG10639-PA|CG10639-PB|CG10639-PC|uncharacterized protein', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18971993, end_position_on_the_genomic_accession=18974448, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35157, CurrentID=0, Status='live', Symbol='CG10431', Aliases='Dmel_ Dmel\\CG10431, dmCG10431', description='CG10431 gene product from transcript CG10431-RA', other_designations='CG10431-PA|CG10431-PB|uncharacterized protein', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18974710, end_position_on_the_genomic_accession=18978187, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35158, CurrentID=0, Status='live', Symbol='Swip-1', Aliases='Dmel_CG10641, CG10641, DSwip-1, Dmel\\CG10641, EFhd2, efhd, efhd2', description='Swiprosin-1', other_designations='CG10641-PA|CG10641-PB|EF-hand domain containing 2|Swip-1-PA|Swip-1-PB', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18977519, end_position_on_the_genomic_accession=18987280, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35159, CurrentID=0, Status='live', Symbol='CG15168', Aliases='Dmel_ BcDNA:RE09053, Dmel\\CG15168', description='CG15168 gene product from transcript CG15168-RA', other_designations='CG15168-PA|uncharacterized protein', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18987707, end_position_on_the_genomic_accession=18988331, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35160, CurrentID=0, Status='live', Symbol='CG15170', Aliases='Dmel_ Dmel\\CG15170', description='CG15170 gene product from transcript CG15170-RB', other_designations='CG15170-PA|CG15170-PB|uncharacterized protein', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18988268, end_position_on_the_genomic_accession=18990233, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35161, CurrentID=0, Status='live', Symbol='CG15169', Aliases='Dmel_ Dmel\\CG15169', description='CG15169 gene product from transcript CG15169-RA', other_designations='CG15169-PA|uncharacterized protein', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18991020, end_position_on_the_genomic_accession=18992409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35162, CurrentID=0, Status='live', Symbol='CG10650', Aliases='Dmel_ BEST:GH09876, Dmel\\CG10650', description='CG10650 gene product from transcript CG10650-RB', other_designations='CG10650-PA|CG10650-PB|uncharacterized protein', map_location='37B8-37B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18993360, end_position_on_the_genomic_accession=18995934, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35163, CurrentID=0, Status='live', Symbol='CG31792', Aliases='Dmel_ CG17338, Dmel\\CG31792', description='CG31792 gene product from transcript CG31792-RC', other_designations='CG31792-PB|CG31792-PC|uncharacterized protein', map_location='37B9-37B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18997448, end_position_on_the_genomic_accession=19002713, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  35166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35166, CurrentID=0, Status='live', Symbol='robl37BC', Aliases='Dmel_CG15171, CG15171, Dmel\\CG15171, Robl37BC', description='CG15171 gene product from transcript CG15171-RA', other_designations='CG15171-PA|robl37BC-PA', map_location='37B9-37B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19010866, end_position_on_the_genomic_accession=19011532, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35167, CurrentID=0, Status='live', Symbol='CG15172', Aliases='Dmel_ Dmel\\CG15172', description='CG15172 gene product from transcript CG15172-RB', other_designations='CG15172-PA|CG15172-PB|CG15172-PC|CG15172-PD|uncharacterized protein', map_location='37B9-37B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19022417, end_position_on_the_genomic_accession=19023646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35168, CurrentID=0, Status='live', Symbol='Sidpn', Aliases='Dmel_CG10446, CG10446, Dmel\\CG10446, Side, Sidestep, bHLHe51, side', description='similar to Deadpan', other_designations='CG10446-PA|Sidpn-PA|similar to deadpan', map_location='37B9-37B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19027710, end_position_on_the_genomic_accession=19030780, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35169, CurrentID=0, Status='live', Symbol='hook', Aliases='Dmel_CG10653, CG10653, Dmel\\CG10653, Hook, dHk, hk', description='CG10653 gene product from transcript CG10653-RA', other_designations='CG10653-PA|hook-PA', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19030844, end_position_on_the_genomic_accession=19033498, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35170, CurrentID=0, Status='live', Symbol='mib2', Aliases='Dmel_CG17492, Be, CG17492, D-mibl, Dmel\\CG17492, Mib-2, i115, l(2)37Be, mib1/2, suelto', description='mind bomb 2', other_designations='CG17492-PA|CG17492-PB|CG17492-PC|lethal (2) 37Be|mib2-PA|mib2-PB|mib2-PC|mindbomb2', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19034444, end_position_on_the_genomic_accession=19041541, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35172, CurrentID=0, Status='live', Symbol='Ttc19', Aliases='Dmel_CG15173, CG15173, Dmel\\CG15173, TTC19, cg15173', description='Tetratricopeptide repeat domain 19', other_designations='CG15173-PA|Ttc19-PA|tetratricopeptide repeat domain 19', map_location='37B11-37B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19043535, end_position_on_the_genomic_accession=19045148, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35173, CurrentID=0, Status='live', Symbol='Acn', Aliases='Dmel_CG10473, Ba, CG10437, CG10473, Dmel\\CG10473, dAcn, dacn, hkl, l(2)37Ba', description='Acinus', other_designations='Acn-PA|Acn-PB|Acn-PC|CG10473-PA|CG10473-PB|CG10473-PC|acinus|drosophlia homolog of acinus|hook-like|lethal (2) 37Ba', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19045355, end_position_on_the_genomic_accession=19048817, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35174, CurrentID=0, Status='live', Symbol='CG10470', Aliases='Dmel_ B2, Dmel\\CG10470, TU37B2, TUB2, TuB2, anon-37B-2', description='CG10470 gene product from transcript CG10470-RA', other_designations='CG10470-PA|transcription unit 37B2|uncharacterized protein', map_location='37B12-37B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19049076, end_position_on_the_genomic_accession=19049884, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35176, CurrentID=0, Status='live', Symbol='Rpn3', Aliases='Dmel_CG42641, CG10484, CG42641, DXA2_DROME, Dmel\\CG42641, Dmel_CG10484, Dox, Dox-A2, DoxA2, RPN3, S3, l(2)37Bf, l(2)37Bf[+], p58', description='Regulatory particle non-ATPase 3', other_designations='CG42641-PA|CG42641-PB|Rpn3-PA|Rpn3-PB|diphenol oxidase A2|diphenol oxidase-A2|regulatory particle non-ATPase 3', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19051790, end_position_on_the_genomic_accession=19053681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35177, CurrentID=0, Status='live', Symbol='CG10492', Aliases='Dmel_ B1, Dmel\\CG10492, TU37B1, TuB1, anon-37B-1', description='CG10492 gene product from transcript CG10492-RA', other_designations='CG10492-PA|CG10492-PB|transcription unit 37B1|uncharacterized protein', map_location='37B12-37B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19054078, end_position_on_the_genomic_accession=19058821, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35178, CurrentID=0, Status='live', Symbol='Phlpp', Aliases='Dmel_CG10493, CG10493, CT29458, Dmel\\CG10493, PP2C, dPHLPP', description='PH domain leucine-rich repeat protein phosphatase', other_designations='CG10493-PA|CG10493-PB|Phlpp-PA|Phlpp-PB', map_location='37B12-37B13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19059129, end_position_on_the_genomic_accession=19063132, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35179, CurrentID=0, Status='live', Symbol='CG10495', Aliases='Dmel_ Dmel\\CG10495, anon-WO0118547.200', description='CG10495 gene product from transcript CG10495-RA', other_designations='CG10495-PA|uncharacterized protein', map_location='37B13-37B13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19063295, end_position_on_the_genomic_accession=19064726, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35180, CurrentID=0, Status='live', Symbol='CG17343', Aliases='Dmel_ Dmel\\CG17343', description='CG17343 gene product from transcript CG17343-RA', other_designations='CG17343-PA|uncharacterized protein', map_location='37B13-37B13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19064929, end_position_on_the_genomic_accession=19069673, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35181, CurrentID=0, Status='live', Symbol='CG10702', Aliases='Dmel_ Dmel\\CG10702', description='CG10702 gene product from transcript CG10702-RD', other_designations='CG10702-PA|CG10702-PB|CG10702-PC|CG10702-PD|uncharacterized protein', map_location='37B13-37B13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19064832, end_position_on_the_genomic_accession=19069851, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35182, CurrentID=0, Status='live', Symbol='CG17572', Aliases='Dmel_ Dmel\\CG17572, SPH64, c-SPH64', description='CG17572 gene product from transcript CG17572-RA', other_designations='CG17572-PA|uncharacterized protein', map_location='37B13-37B13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19071131, end_position_on_the_genomic_accession=19074149, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35183, CurrentID=0, Status='live', Symbol='CG10700', Aliases='Dmel_ Dmel\\CG10700', description='CG10700 gene product from transcript CG10700-RA', other_designations='CG10700-PA|uncharacterized protein', map_location='37B13-37B13', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19077276, end_position_on_the_genomic_accession=19079028, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35184, CurrentID=0, Status='live', Symbol='Lim3', Aliases='Dmel_CG10699, CG10699, CG15174, Dmel\\CG10699, Lim-3A, l(2)37Bd, l(2)Bd, lim3', description='CG10699 gene product from transcript CG10699-RF', other_designations='CG10699-PA|CG10699-PB|CG10699-PC|CG10699-PD|CG10699-PE|CG10699-PF|CG10699-PG|Lim3-PA|Lim3-PB|Lim3-PC|Lim3-PD|Lim3-PE|Lim3-PF|Lim3-PG|lethal(2)37Bd', map_location='37B13-37C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19079256, end_position_on_the_genomic_accession=19108427, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  5740264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740264, CurrentID=0, Status='live', Symbol='CG34175', Aliases='Dmel_ Dmel\\CG34175', description='CG34175 gene product from transcript CG34175-RA', other_designations='CG34175-PA|uncharacterized protein', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3398023, end_position_on_the_genomic_accession=3398684, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35188, CurrentID=0, Status='live', Symbol='amd', Aliases='Dmel_CG10501, AMD-r, CG10501, CG17345, Dmel\\CG10501, X04695, fAMD, l(2)37Bk, l(2)amd, l(2)amd alpha-methyldopa hypersensitive, l(2)amd[H]', description='alpha methyl dopa-resistant', other_designations='CG10501-PA|CG10501-PB|DHPAA synthase|alpha methyl dopa|alpha-methotrexate deficient|alpha-methyl dopa resistant|amd-PA|amd-PB|dopa decarboxylase|transcription unit No. 1', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19110154, end_position_on_the_genomic_accession=19114110, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35189, CurrentID=0, Status='live', Symbol='CG10561', Aliases='Dmel_ C2, Cs, Dmel\\CG10561, ORF1, ORF2, ORF3, TU37C1, TU37C2, TuC1, anon-37Cb, anon-37Cc, anon-37Cd, anon-37Cs, cs, l(2)37Cs', description='CG10561 gene product from transcript CG10561-RA', other_designations='CG10561-PA|lethal(2)37Cs|transcription unit 37C2|transcription unit No. 2|uncharacterized protein', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19114401, end_position_on_the_genomic_accession=19116570, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35190, CurrentID=0, Status='live', Symbol='Ddc', Aliases='Dmel_CG10697, 5-HT, AADC, CG10697, DDCDm, Dmel\\CG10697, ddc, fDDC, l(2)37Bl, l(2)37Ch, l(2)k02104, Ddc', description='Dopa decarboxylase', other_designations='3,4-Dihydroxyphenylalanine decarboxylase|CG10697-PB|CG10697-PC|CG10697-PD|Ddc-PB|Ddc-PC|Ddc-PD|aromatic L-amino acid decarboxylase|dopa decarboxylase|dopa-decarboxylase|dopadecarboxylase|dopamine decarboxylase|transcription unit No. 3', map_location='37C1-37C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19116483, end_position_on_the_genomic_accession=19120306, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35192, CurrentID=0, Status='live', Symbol='l(2)37Cb', Aliases='Dmel_CG10689, CG10689, Cb, Dmel\\CG10689, cg10689, l(2)E104', description='lethal (2) 37Cb', other_designations='CG10689-PA|l(2)37Cb-PA', map_location='37C5-37C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19123998, end_position_on_the_genomic_accession=19126976, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35193, CurrentID=0, Status='live', Symbol='l(2)37Cd', Aliases='Dmel_CG10563, CG10563, Cd, Dmel\\CG10563', description='lethal (2) 37Cd', other_designations='CG10563-PA|l(2)37Cd-PA', map_location='37C5-37C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19127260, end_position_on_the_genomic_accession=19129326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35194, CurrentID=0, Status='live', Symbol='Aats-asn', Aliases='Dmel_CG10687, AsnRS, BcDNA:GH06451, CG10687, Ca, Dmel\\CG10687, NRS, l(2)37Ca, l(2)37Ca)E11, l(2)E11', description='Asparaginyl-tRNA synthetase', other_designations='Aats-asn-PA|Aats-asn-PB|CG10687-PA|CG10687-PB|lethal (2) 37Ca', map_location='37C5-37C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19129469, end_position_on_the_genomic_accession=19131954, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35195, CurrentID=0, Status='live', Symbol='DCTN6-p27', Aliases='Dmel_CG17347, CG17347, Ce, Dmel\\CG17347, l(2)37Ce, l(2)37Ce)E62, l(2)E62', description='Dynactin 6, p27 subunit', other_designations='CG17347-PA|DCTN6-p27-PA|P27 DYNACTIN|dynactin 6, p27 subunit|dynactin, p27 subunit|lethal (2) 37Ce', map_location='37C5-37C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19132323, end_position_on_the_genomic_accession=19133097, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35196, CurrentID=0, Status='live', Symbol='l(2)37Cg', Aliases='Dmel_CG10685, CG10685, Cg, Dmel\\CG10685', description='lethal (2) 37Cg', other_designations='CG10685-PA|CG10685-PC|CG10685-PD|l(2)37Cg-PA|l(2)37Cg-PC|l(2)37Cg-PD', map_location='37C5-37C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19133023, end_position_on_the_genomic_accession=19135233, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35197, CurrentID=0, Status='live', Symbol='brat', Aliases='Dmel_CG10719, BRAT, Brat, CG10719, Cf, Dmel\\CG10719, E60, anon-37CDa, fs(2)ltoPM43, l(2)37Cf, l(2)E60, l(2)brat, lincRNA.S1565', description='brain tumor', other_designations='CG10719-PA|CG10719-PB|CG10719-PC|CG10719-PE|CG10719-PF|brain tumour|brat-PA|brat-PB|brat-PC|brat-PE|brat-PF', map_location='37C1-37C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19133816, end_position_on_the_genomic_accession=19179758, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35198, CurrentID=0, Status='live', Symbol='CG17568', Aliases='Dmel_ Dmel\\CG17568', description='CG17568 gene product from transcript CG17568-RA', other_designations='CG17568-PA|uncharacterized protein', map_location='37C6-37C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19179749, end_position_on_the_genomic_accession=19181702, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35199, CurrentID=0, Status='live', Symbol='gammaTub37C', Aliases='Dmel_CG17566, CG17566, D.m.GAMMA-37C, Dmel\\CG17566, GTU88*, PL10, Rb1011, TW1, fs(2)TW1, fs(2)TWI, fsTW1, g-Tubulin, gamma Tub37CD, gamma tubulin, gamma-Tub, gamma-Tub37C, gamma-tub37C, gamma-tub37CD, gamma-tubulin, gammaTUB, gammaTUB37C, gammaTub, gammaTub2D, gammaTub37c, gamma[mat], gammatub, gammaTub37C', description='gamma-Tubulin at 37C', other_designations='37CD gamma-tubulin|CG17566-PA|CG17566-PB|female sterile(2)Ted wright 1|gama-tubulin|gamma Tubulin2|gamma tubulin|gamma tubulin at 37C|gamma-Tubulin|gamma-tubulin (37C)|gamma-tubulin-2|gamma-tubulin37C|gammaTub37C-PA|gammaTub37C-PB|gammatubulin|maternal gamma-tubulin|tubulin', map_location='37C7-37C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19183863, end_position_on_the_genomic_accession=19185726, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35200, CurrentID=0, Status='live', Symbol='pigeon', Aliases='Dmel_CG10739, CG10739, Dmel\\CG10739, lio, pio', description='CG10739 gene product from transcript CG10739-RB', other_designations='CG10739-PA|CG10739-PB|linotte|lionette|pigeon-PA|pigeon-PB', map_location='37C7-37C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19186043, end_position_on_the_genomic_accession=19189938, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35203, CurrentID=0, Status='live', Symbol='CG31797', Aliases='Dmel_ CG15175, CG15176, CT35089, Dmel\\CG31797', description='CG31797 gene product from transcript CG31797-RA', other_designations='CG31797-PA|uncharacterized protein', map_location='37D1-37D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19259432, end_position_on_the_genomic_accession=19262240, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35205, CurrentID=0, Status='live', Symbol='CG17564', Aliases='Dmel_ Dmel\\CG17564', description='CG17564 gene product from transcript CG17564-RB', other_designations='CG17564-PB|uncharacterized protein', map_location='37D2-37D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19298345, end_position_on_the_genomic_accession=19299933, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35206, CurrentID=0, Status='live', Symbol='CG10750', Aliases='Dmel_ Dmel\\CG10750', description='CG10750 gene product from transcript CG10750-RB', other_designations='CG10750-PA|CG10750-PB|uncharacterized protein', map_location='37D2-37D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19300255, end_position_on_the_genomic_accession=19301600, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35207, CurrentID=0, Status='live', Symbol='dnt', Aliases='Dmel_CG17559, CG17559, DNT, Dmel\\CG17559', description='doughnut on 2', other_designations='CG17559-PA|CG17559-PB|dnt-PA|dnt-PB|doughnut', map_location='37D2-37D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19338669, end_position_on_the_genomic_accession=19363227, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35208, CurrentID=0, Status='live', Symbol='CG13086', Aliases='Dmel_ Dmel\\CG13086', description='CG13086 gene product from transcript CG13086-RB', other_designations='CG13086-PA|CG13086-PB|CG13086-PC|uncharacterized protein', map_location='37D2-37D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19364375, end_position_on_the_genomic_accession=19365470, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35209, CurrentID=0, Status='live', Symbol='CG17349', Aliases='Dmel_ Dmel\\CG17349, anon-WO0140519.204', description='CG17349 gene product from transcript CG17349-RA', other_designations='CG17349-PA|uncharacterized protein', map_location='37D3-37D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19370445, end_position_on_the_genomic_accession=19372711, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35210, CurrentID=0, Status='live', Symbol='CG17350', Aliases='Dmel_ Dmel\\CG17350', description='CG17350 gene product from transcript CG17350-RA', other_designations='CG17350-PA|uncharacterized protein', map_location='37D3-37D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19380930, end_position_on_the_genomic_accession=19383093, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35211, CurrentID=0, Status='live', Symbol='fon', Aliases='Dmel_CG15825, CG15825, Dmel\\CG15825, Fon', description='fondue', other_designations='CG15825-PA|CG15825-PB|CG15825-PC|CG15825-PD|fon-PA|fon-PB|fon-PC|fon-PD|gpdhdue', map_location='37D3-37D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19383138, end_position_on_the_genomic_accession=19386205, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35212, CurrentID=0, Status='live', Symbol='CG17549', Aliases='Dmel_ Dmel\\CG17549', description='CG17549 gene product from transcript CG17549-RB', other_designations='CG17549-PA|CG17549-PB|CG17549-PC|CG17549-PD|CG17549-PE|uncharacterized protein', map_location='37D4-37D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19387050, end_position_on_the_genomic_accession=19390075, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35213, CurrentID=0, Status='live', Symbol='CG17544', Aliases='Dmel_ Dmel\\CG17544', description='CG17544 gene product from transcript CG17544-RA', other_designations='CG17544-PA|CG17544-PB|CG17544-PC|CG17544-PD|uncharacterized protein', map_location='37D4-37D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19390370, end_position_on_the_genomic_accession=19398493, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35215, CurrentID=0, Status='live', Symbol='Pax', Aliases='Dmel_CG31794, CG18061, CG18576, CG31794, CT40481, CT42454, DPaxillin, DPxn, DPxn37, Dmel\\CG31794, Dpax, DpaxA, PDLP, dPax, pax', description='Paxillin', other_designations='CG31794-PA|CG31794-PB|CG31794-PC|CG31794-PD|CG31794-PF|CG31794-PG|CG31794-PH|CG31794-PJ|CG31794-PK|D-Paxillin|Pax-PA|Pax-PB|Pax-PC|Pax-PD|Pax-PF|Pax-PG|Pax-PH|Pax-PJ|Pax-PK|Pax-derived LIM only protein|Paxillin-derived LIM-only protein|paxilin|paxillin', map_location='37D5-37D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19400753, end_position_on_the_genomic_accession=19426596, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  35216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35216, CurrentID=0, Status='live', Symbol='Lectin-galC1', Aliases='Dmel_CG9976, AC007082a, CG9976, DL1, Dgal-1, Dmel\\CG9976, Lecti-galC1, lectin-galC1', description='Galactose-specific C-type lectin', other_designations='CG9976-PA|CG9976-PB|Lectin-galC1-PA|Lectin-galC1-PB|drosophila lectin 1|galactin|galactose specific C-type lectin|lectin|lectin-galC1', map_location='37D6-37D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19417447, end_position_on_the_genomic_accession=19418274, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35218, CurrentID=0, Status='live', Symbol='CG16771', Aliases='Dmel_ Dmel\\CG16771', description='CG16771 gene product from transcript CG16771-RE', other_designations='CG16771-PA|CG16771-PB|CG16771-PC|CG16771-PD|CG16771-PE|uncharacterized protein', map_location='37D7-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19424050, end_position_on_the_genomic_accession=19429224, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35219, CurrentID=0, Status='live', Symbol='CG13085', Aliases='Dmel_ Dmel\\CG13085, FBXO3', description='CG13085 gene product from transcript CG13085-RB', other_designations='CG13085-PA|CG13085-PB|uncharacterized protein', map_location='37D7-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19424050, end_position_on_the_genomic_accession=19430270, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35220, CurrentID=0, Status='live', Symbol='CG9987', Aliases='Dmel_ Dmel\\CG9987', description='CG9987 gene product from transcript CG9987-RA', other_designations='CG9987-PA', map_location='37E1-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19430643, end_position_on_the_genomic_accession=19432444, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35221, CurrentID=0, Status='live', Symbol='Rab9', Aliases='Dmel_CG9994, 146494_at, CG9994, Dm Dmel\\CG9994, rab9', description='CG9994 gene product from transcript CG9994-RA', other_designations='CG9994-PA|Rab9-PA', map_location='37E1-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19432574, end_position_on_the_genomic_accession=19435841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35222, CurrentID=0, Status='live', Symbol='CG10237', Aliases='Dmel_ Dmel\\CG10237', description='CG10237 gene product from transcript CG10237-RD', other_designations='CG10237-PA|CG10237-PB|CG10237-PC|CG10237-PD|CG10237-PE|uncharacterized protein', map_location='37E1-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19435850, end_position_on_the_genomic_accession=19442727, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35223, CurrentID=0, Status='live', Symbol='RanGAP', Aliases='Dmel_CG9999, CG9999, CT28175, Dmel\\CG9999, RGP1_DROME, Ran, RanGap, RanGap1, SD, Scl, Sd, Sd-RanGAP, Sd-RanGap, dRanGAP, ranGAP, ranGap, rangap', description='Ran GTPase activating protein', other_designations='CG9999-PA|CG9999-PB|CG9999-PC|RanGAP-PA|RanGAP-PB|RanGAP-PC|segregation distorter', map_location='2-54 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19441959, end_position_on_the_genomic_accession=19447317, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35225, CurrentID=0, Status='live', Symbol='Top2', Aliases='Dmel_CG10223, CG10223, CHRAC, Dmel\\CG10223, TOP2, TOPO II, TOPOII, TopII, Topo II, TopoII, dCHRAC, dTopoII, top2, topo II, topoII', description='Topoisomerase 2', other_designations='CG10223-PA|CG10223-PB|DNA topoisomerase II|DNA topoisomeriase II|Top 2|Top2-PA|Top2-PB|Topoisomerase2|topoisomerase 2|topoisomerase II|type II DNA topoisomerase|type II topoisomerase', map_location='37E1-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19447365, end_position_on_the_genomic_accession=19453490, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35226, CurrentID=0, Status='live', Symbol='CG10026', Aliases='Dmel_ Dmel\\CG10026', description='CG10026 gene product from transcript CG10026-RA', other_designations='CG10026-PA|CG10026-PB|CG10026-PC|CG10026-PD|uncharacterized protein', map_location='37E1-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19454256, end_position_on_the_genomic_accession=19457723, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35227, CurrentID=0, Status='live', Symbol='tj', Aliases='Dmel_CG10034, CG10034, DM13, Dmaf, Dmel\\CG10034, PL3, TJ, Tj, fs(2)eo-B, fs(2)eo2, fs(2)eoPL3, maf-L, maf1', description='traffic jam', other_designations='CG10034-PB|CG10034-PC|Traffic-Jam|female sterile(2)eo2|tj-PB|tj-PC', map_location='37E3-37E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19464267, end_position_on_the_genomic_accession=19467758, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35228, CurrentID=0, Status='live', Symbol='CG10195', Aliases='Dmel_ Dmel\\CG10195', description='CG10195 gene product from transcript CG10195-RC', other_designations='CG10195-PB|CG10195-PC|uncharacterized protein', map_location='37E3-37E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19482232, end_position_on_the_genomic_accession=19484705, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35229, CurrentID=0, Status='live', Symbol='CG13084', Aliases='Dmel_ Dmel\\CG13084', description='CG13084 gene product from transcript CG13084-RA', other_designations='CG13084-PA|uncharacterized protein', map_location='37E3-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19484930, end_position_on_the_genomic_accession=19486532, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35230, CurrentID=0, Status='live', Symbol='CG13083', Aliases='Dmel_ Dmel\\CG13083', description='CG13083 gene product from transcript CG13083-RB', other_designations='CG13083-PB|uncharacterized protein', map_location='37E4-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19487490, end_position_on_the_genomic_accession=19488691, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35231, CurrentID=0, Status='live', Symbol='CG10194', Aliases='Dmel_ Dmel\\CG10194', description='CG10194 gene product from transcript CG10194-RA', other_designations='CG10194-PA|uncharacterized protein', map_location='37E4-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19488682, end_position_on_the_genomic_accession=19490104, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35232, CurrentID=0, Status='live', Symbol='CG18094', Aliases='Dmel_ CG10048, Dmel\\CG18094', description='CG18094 gene product from transcript CG18094-RB', other_designations='CG18094-PA|CG18094-PB|uncharacterized protein', map_location='37E4-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19490262, end_position_on_the_genomic_accession=19491679, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35234, CurrentID=0, Status='live', Symbol='CG10189', Aliases='Dmel_ Dmel\\CG10189, i180, i199', description='CG10189 gene product from transcript CG10189-RB', other_designations='CG10189-PA|CG10189-PB|uncharacterized protein', map_location='37E4-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19491647, end_position_on_the_genomic_accession=19493253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35235, CurrentID=0, Status='live', Symbol='swm', Aliases='Dmel_CG10084, CG10084, Dmel\\CG10084, E1, EC2-10, S(Rux)2B, S(rux)2B, Swm, l(2)37Dh, l(2)E1', description='second mitotic wave missing', other_designations='CG10084-PA|CG10084-PB|CG10084-PC|CG10084-PD|lethal (2) 37Dh|swm-PA|swm-PB|swm-PC|swm-PD', map_location='37E4-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19493251, end_position_on_the_genomic_accession=19497978, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35236, CurrentID=0, Status='live', Symbol='Top3alpha', Aliases='Dmel_CG10123, CG10123, DmTop3alpha, Dmel\\CG10123, Top3a, Topo IIIalpha, dTopoIIIa, top3alpha, topo IIIalpha', description='Topoisomerase 3alpha', other_designations='CG10123-PA|Top 3alpha|Top3alpha-PA|Topoisomerase3alpha|topoisomerase 3-alpha|topoisomerase 3alpha|topoisomerase IIIa|topoisomerase IIIalpha', map_location='37E4-37E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19498168, end_position_on_the_genomic_accession=19502268, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35237, CurrentID=0, Status='live', Symbol='CG10188', Aliases='Dmel_ Dmel\\CG10188', description='CG10188 gene product from transcript CG10188-RB', other_designations='CG10188-PA|CG10188-PB|CG10188-PC|uncharacterized protein', map_location='37E4-37E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19502335, end_position_on_the_genomic_accession=19508764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35238, CurrentID=0, Status='live', Symbol='CG10186', Aliases='Dmel_ CT28647, Dmel\\CG10186, anon-EST:Posey189', description='CG10186 gene product from transcript CG10186-RA', other_designations='CG10186-PA|CG10186-PC|CG10186-PD|CG10186-PG|CG10186-PH|uncharacterized protein', map_location='37E5-37E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19508953, end_position_on_the_genomic_accession=19517204, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  35239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35239, CurrentID=0, Status='live', Symbol='CG10132', Aliases='Dmel_ Dmel\\CG10132', description='CG10132 gene product from transcript CG10132-RA', other_designations='CG10132-PA|CG10132-PB|uncharacterized protein', map_location='37E5-37E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19518824, end_position_on_the_genomic_accession=19525453, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  35240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35240, CurrentID=0, Status='live', Symbol='CG10166', Aliases='Dmel_ Dmel\\CG10166, Dol-P-ManTase, dDPM', description='CG10166 gene product from transcript CG10166-RA', other_designations='CG10166-PA|Dolichol-phosphate mannosyltransferase|uncharacterized protein', map_location='37E5-37E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19525434, end_position_on_the_genomic_accession=19526368, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35241, CurrentID=0, Status='live', Symbol='CG10137', Aliases='Dmel_ Dmel\\CG10137', description='CG10137 gene product from transcript CG10137-RA', other_designations='CG10137-PA|uncharacterized protein', map_location='37E5-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19526686, end_position_on_the_genomic_accession=19532985, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35242, CurrentID=0, Status='live', Symbol='CG10165', Aliases='Dmel_ Dmel\\CG10165', description='CG10165 gene product from transcript CG10165-RC', other_designations='CG10165-PA|CG10165-PB|CG10165-PC|CG10165-PD|CG10165-PE|uncharacterized protein', map_location='37E5-37E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19526910, end_position_on_the_genomic_accession=19530305, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35243, CurrentID=0, Status='live', Symbol='CG31697', Aliases='Dmel_ CG10355, Dmel\\CG31697', description='CG31697 gene product from transcript CG31697-RA', other_designations='CG31697-PA|uncharacterized protein', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19535448, end_position_on_the_genomic_accession=19536104, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35244, CurrentID=0, Status='live', Symbol='CG13082', Aliases='Dmel_ Dmel\\CG13082', description='CG13082 gene product from transcript CG13082-RA', other_designations='CG13082-PA|uncharacterized protein', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19538425, end_position_on_the_genomic_accession=19539740, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35245, CurrentID=0, Status='live', Symbol='CG13081', Aliases='Dmel_ Dmel\\CG13081', description='CG13081 gene product from transcript CG13081-RB', other_designations='CG13081-PA|CG13081-PB|uncharacterized protein', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19540607, end_position_on_the_genomic_accession=19541540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35246, CurrentID=0, Status='live', Symbol='ref(2)P', Aliases='Dmel_CG10360, CG10360, Dmel\\CG10360, Ref(2)P, Ref(2)p, Ref2P, SQSTM1, p62, p63, refn, ref(2)Po2, ref(2)p, ref2p, ref(2)P', description='refractory to sigma P', other_designations='CG10360-PA|CG10360-PB|ref(2)P-PA|ref(2)P-PB|refractaire|refractaire(2)Paris|sequestosome|sequestosome-1', map_location='2-54.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19542468, end_position_on_the_genomic_accession=19545548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35247, CurrentID=0, Status='live', Symbol='CG10337', Aliases='Dmel_ Dmel\\CG10337', description='CG10337 gene product from transcript CG10337-RA', other_designations='CG10337-PA|uncharacterized protein', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19545376, end_position_on_the_genomic_accession=19546959, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35248, CurrentID=0, Status='live', Symbol='Tep4', Aliases='Dmel_CG10363, CG10363, CT29102, Dmel\\CG10363, TEP IV, TEP4, TEP4mel, TEPIV, Tep IV, TepIV, dTEPIV, tep4, tepIV', description='Thioester-containing protein 4', other_designations='CG10363-PA|CG10363-PB|CG10363-PC|CG10363-PD|TEP|Tep4-PA|Tep4-PB|Tep4-PC|Tep4-PD|thioester containing protein IV|thiolester containing protein IV', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19548507, end_position_on_the_genomic_accession=19556451, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35250, CurrentID=0, Status='live', Symbol='Tep5', Aliases='Dmel_CG13079, CG13079, CG13080, Dmel\\CG13079, TEP5mel, TepV, dTEPV', description='Thioester-containing protein 5', other_designations='CG13079-PB|TEP|Tep5-PB|thiolester containing protein V', map_location='37F1-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19558047, end_position_on_the_genomic_accession=19563049, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35251, CurrentID=0, Status='live', Symbol='CG13078', Aliases='Dmel_ Dmel\\CG13078', description='CG13078 gene product from transcript CG13078-RA', other_designations='CG13078-PA|uncharacterized protein', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19563930, end_position_on_the_genomic_accession=19565158, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35252, CurrentID=0, Status='live', Symbol='CG13077', Aliases='Dmel_ Dmel\\CG13077', description='CG13077 gene product from transcript CG13077-RA', other_designations='CG13077-PA|CG13077-PB|uncharacterized protein', map_location='37F1-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19565863, end_position_on_the_genomic_accession=19567359, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35253, CurrentID=0, Status='live', Symbol='spi', Aliases='Dmel_CG10334, CG10334, CT29014, Dmel\\CG10334, EP(2)2378, Spi, Spitz, anon-WO0118547.158, l(2)01068, l(2)37Fa, l(2)s3547, s-spi, sSpi', description='spitz', other_designations='CG10334-PA|CG10334-PB|CG10334-PE|CG10334-PF|CG10334-PG|epidermal growth factor protein|spi-PA|spi-PB|spi-PE|spi-PF|spi-PG', map_location='2-54 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19567900, end_position_on_the_genomic_accession=19577315, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35254, CurrentID=0, Status='live', Symbol='msb1l', Aliases='Dmel_CG10364, CG10364, Dmel\\CG10364, Msb1l, msh1l', description='CG10364 gene product from transcript CG10364-RA', other_designations='CG10364-PA|msb1l-PA', map_location='37F2-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19577701, end_position_on_the_genomic_accession=19579131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35255, CurrentID=0, Status='live', Symbol='CG10268', Aliases='Dmel_ BcDNA:SD12705, Dmel\\CG10268, MevPK', description='CG10268 gene product from transcript CG10268-RA', other_designations='CG10268-PA|uncharacterized protein', map_location='37F2-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19579093, end_position_on_the_genomic_accession=19580351, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35256, CurrentID=0, Status='live', Symbol='Hakai', Aliases='Dmel_CG10263, CG10263, Dmel\\CG10263, hakai', description='CG10263 gene product from transcript CG10263-RE', other_designations='CG10263-PA|CG10263-PC|CG10263-PE|CG10263-PF|Hakai-PA|Hakai-PC|Hakai-PE|Hakai-PF|hakai', map_location='37F2-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19580612, end_position_on_the_genomic_accession=19582474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35257, CurrentID=0, Status='live', Symbol='CG10262', Aliases='Dmel_ DmPCNA2, Dmel\\CG10262, PCNA2', description='CG10262 gene product from transcript CG10262-RA', other_designations='CG10262-PA|uncharacterized protein', map_location='37F2-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19582823, end_position_on_the_genomic_accession=19583990, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35258, CurrentID=0, Status='live', Symbol='CG10366', Aliases='Dmel_ Dmel\\CG10366', description='CG10366 gene product from transcript CG10366-RA', other_designations='CG10366-PA|uncharacterized protein', map_location='37F2-37F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19584235, end_position_on_the_genomic_accession=19586440, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35259, CurrentID=0, Status='live', Symbol='Lar', Aliases='Dmel_CG10443, CG10443, DLAR, DLar, Dlar, Dmel\\CG10443, LAR, PTP38A, Ptp38A, dlar, lar', description='Leukocyte-antigen-related-like', other_designations='CG10443-PA|CG10443-PB|CG10443-PD|CG10443-PF|CG10443-PG|CG10443-PH|CG10443-PI|LAR receptor tyrosine phosphatase|Lar-PA|Lar-PB|Lar-PD|Lar-PF|Lar-PG|Lar-PH|Lar-PI|bypass|leukocyte- antigen-related-like|leukocyte-antigen-related-like', map_location='37F2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19586623, end_position_on_the_genomic_accession=19732069, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  35260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35260, CurrentID=0, Status='live', Symbol='CG10462', Aliases='Dmel_ BcDNA:SD07008, Dmel\\CG10462', description='CG10462 gene product from transcript CG10462-RD', other_designations='CG10462-PA|CG10462-PB|CG10462-PC|CG10462-PD|CG10462-PE|CG10462-PF|uncharacterized protein', map_location='38A2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19732080, end_position_on_the_genomic_accession=19735980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35261, CurrentID=0, Status='live', Symbol='Nf-YB', Aliases='Dmel_CG10447, BcDNA:RH50436, CG10447, Dmel\\CG10447, NF-YB, dNF-YB', description='Nuclear factor Y-box B', other_designations='CG10447-PA|CG10447-PB|Nf-YB-PA|Nf-YB-PB|nuclear factor Y-box B', map_location='38A2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19736044, end_position_on_the_genomic_accession=19737088, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35262, CurrentID=0, Status='live', Symbol='CG10631', Aliases='Dmel_ Dmel\\CG10631', description='CG10631 gene product from transcript CG10631-RA', other_designations='CG10631-PA|uncharacterized protein', map_location='38A2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19742817, end_position_on_the_genomic_accession=19756904, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35263, CurrentID=0, Status='live', Symbol='CG10628', Aliases='Dmel_ Dmel\\CG10628', description='CG10628 gene product from transcript CG10628-RA', other_designations='CG10628-PA|uncharacterized protein', map_location='38A2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19757245, end_position_on_the_genomic_accession=19758615, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35264, CurrentID=0, Status='live', Symbol='CG10463', Aliases='Dmel_ Dmel\\CG10463', description='CG10463 gene product from transcript CG10463-RA', other_designations='CG10463-PA|uncharacterized protein', map_location='38A2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19758879, end_position_on_the_genomic_accession=19760986, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35265, CurrentID=0, Status='live', Symbol='fbp', Aliases='Dmel_CG31692, CG10611, CG31692, Dmel\\CG31692, FbPase', description='fructose-1,6-bisphosphatase', other_designations='CG31692-PA|CG31692-PC|CG31692-PD|CG31692-PF|fbp-PA|fbp-PC|fbp-PD|fbp-PF', map_location='38A2-38A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19761587, end_position_on_the_genomic_accession=19764136, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35266, CurrentID=0, Status='live', Symbol='bsh', Aliases='Dmel_CG10604, BSH, Bsh, CG10604, Dmel\\CG10604, E86, Hox11-310, lambdaE86', description='brain-specific homeobox', other_designations='Brain-specific homeobox|CG10604-PB|CG10604-PC|brain specific homeobox|bsh-PB|bsh-PC', map_location='38A3-38A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19765776, end_position_on_the_genomic_accession=19769067, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35267, CurrentID=0, Status='live', Symbol='CdGAPr', Aliases='Dmel_CG10538, 10538, CG10538, Dmel\\CG10538, FBgn 32821, GAP, cdGAPr, d-CdGAPr', description='Cd GTPase activating protein-related', other_designations='CG10538-PA|CG10538-PB|CG10538-PC|CG10538-PD|CdGAPr-PA|CdGAPr-PB|CdGAPr-PC|CdGAPr-PD', map_location='38A3-38A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19774874, end_position_on_the_genomic_accession=19791882, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35268, CurrentID=0, Status='live', Symbol='CG10466', Aliases='Dmel_ Dmel\\CG10466, anon-WO0140519.255, cg10466', description='CG10466 gene product from transcript CG10466-RB', other_designations='CG10466-PA|CG10466-PB|uncharacterized protein', map_location='38A3-38A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19792104, end_position_on_the_genomic_accession=19792782, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35270, CurrentID=0, Status='live', Symbol='CG13962', Aliases='Dmel_ Dmel\\CG13962', description='CG13962 gene product from transcript CG13962-RB', other_designations='CG13962-PA|CG13962-PB|uncharacterized protein', map_location='38A4-38A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19813704, end_position_on_the_genomic_accession=19814532, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35273, CurrentID=0, Status='live', Symbol='CG10481', Aliases='Dmel_ Dmel\\CG10481', description='CG10481 gene product from transcript CG10481-RB', other_designations='CG10481-PB|uncharacterized protein', map_location='38A6-38A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19890201, end_position_on_the_genomic_accession=19892550, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35275, CurrentID=0, Status='live', Symbol='Victoria', Aliases='Dmel_CG33117, CG13963, CG33117, Dmel\\CG33117, TotE', description='CG33117 gene product from transcript CG33117-RA', other_designations='CG33117-PA|Victoria-PA|turandot E|victoria', map_location='38A7-38A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19912938, end_position_on_the_genomic_accession=19913516, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35277, CurrentID=0, Status='live', Symbol='sick', Aliases='Dmel_CG43720, CG10662, CG13958, CG13959, CG13960, CG13961, CG13964, CG33316, CG34343, CG42589, CG43720, Dm_2L:104545, Dmel\\CG43720, Dmel_CG10662, Dmel_CG13958, Dmel_CG13964, Dmel_CG33316, Dmel_CG34343, Dmel_CG42589', description='sickie', other_designations='CG43720-PB|CG43720-PD|CG43720-PE|CG43720-PF|CG43720-PG|CG43720-PH|CG43720-PI|CG43720-PJ|CG43720-PK|CG43720-PL|CG43720-PM|CG43720-PN|sick-PB|sick-PD|sick-PE|sick-PF|sick-PG|sick-PH|sick-PI|sick-PJ|sick-PK|sick-PL|sick-PM|sick-PN', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19796365, end_position_on_the_genomic_accession=19958424, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  35279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35279, CurrentID=0, Status='live', Symbol='COX4', Aliases='Dmel_CG10664, CG10664, COX, COX IV, COXIV, CX41, CX42, CcO IV, CoIV, CoxIV, Dmel\\CG10664, IV, cox4, dCOX-IV', description='Cytochrome c oxidase subunit 4', other_designations='CG10664-PA|CG10664-PB|CG10664-PC|CG10664-PD|COX4-PA|COX4-PB|COX4-PC|COX4-PD|cytochrome c oxidase|cytochrome c oxidase IV|cytochrome c oxidase polypeptide IV|cytochrome c oxidase subunit 4|cytochrome c oxidase subunit IV|subunit IV', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19958713, end_position_on_the_genomic_accession=19960065, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35280, CurrentID=0, Status='live', Symbol='CG13965', Aliases='Dmel_ Dmel\\CG13965', description='CG13965 gene product from transcript CG13965-RA', other_designations='CG13965-PA|uncharacterized protein', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19961758, end_position_on_the_genomic_accession=19962352, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35281, CurrentID=0, Status='live', Symbol='CG16772', Aliases='Dmel_ Dmel\\CG16772, anon-WO0138581.5', description='CG16772 gene product from transcript CG16772-RA', other_designations='CG16772-PA|CG16772-PB|uncharacterized protein', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19962679, end_position_on_the_genomic_accession=19963844, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35282, CurrentID=0, Status='live', Symbol='CG10680', Aliases='Dmel_ Dmel\\CG10680', description='CG10680 gene product from transcript CG10680-RB', other_designations='CG10680-PA|CG10680-PB|uncharacterized protein', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19964136, end_position_on_the_genomic_accession=19965702, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35284, CurrentID=0, Status='live', Symbol='CG43861', Aliases='Dmel_ CG12508, CG13966, Dmel\\CG43861, Dmel_CG12508, Dmel_CG13966', description='CG43861 gene product from transcript CG43861-RD', other_designations='CG43861-PA|CG43861-PB|CG43861-PC|CG43861-PD|uncharacterized protein', map_location='38A8-38B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19968684, end_position_on_the_genomic_accession=20017282, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35285, CurrentID=0, Status='live', Symbol='CG10659', Aliases='Dmel_ Dmel\\CG10659', description='CG10659 gene product from transcript CG10659-RA', other_designations='CG10659-PA|uncharacterized protein', map_location='38B1-38B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20024975, end_position_on_the_genomic_accession=20025811, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35286, CurrentID=0, Status='live', Symbol='sNPF', Aliases='Dmel_CG13968, 38B.2, CG13968, Dmel\\CG13968, Drm-sNPF, Drm-sNPF-1, Drm-sNPF-2, Drm-sNPF-3, Drm-sNPF-4, Drm-sNPF-AP1, LRLRFamide, SNPF, lrlrfa, s-NPF-1, sNPF-1[4-11], sNPF-2, sNPF-2[12-19], sNPF-2s, sNPF-3, sNPF-4, sNPF1, sNPF3, short NPF, snpf, sNPF', description='short neuropeptide F precursor', other_designations='CG13968-PB|LRLRFamides|neuropeptide F (short)|precursor of short neuropeptide F|sNPF precursor|sNPF-PB|short NPF|short neuropeptide F|short neuropeptide F-1|short neuropeptide F1|small neuropeptide F|small neuropeptide F-1|small neuropeptide-F', map_location='38B1-38B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20027915, end_position_on_the_genomic_accession=20056373, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35287, CurrentID=0, Status='live', Symbol='barr', Aliases='Dmel_CG10726, 38B.3, Barr, Barren, CAP-H, CG10726, CG14684, Cap-H, CapH, Dmel\\CG10726, dCap-H', description='barren', other_designations='CG10726-PA|CG10726-PB|CG10726-PC|barr-PA|barr-PB|barr-PC', map_location='38B1-38B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20058163, end_position_on_the_genomic_accession=20061370, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35288, CurrentID=0, Status='live', Symbol='lok', Aliases='Dmel_CG10895, 10895, 38B.4, CG10895, CHK-2, CHK2, Chk2, DmCHk2, DmChk2, Dmchk2, Dmel\\CG10895, Dmnk, Dmnk-L, Dmnk-S, Dmnk/DChk2, LOKI/CHK2, MNK, chk, chk2, chk2/lok, mnk', description='loki', other_designations='CG10895-PA|CG10895-PB|CG10895-PC|CG10895-PD|checkpoint kinase 2|drosophila maternal nuclear kinase|localized ovarian kinase|lok-PA|lok-PB|lok-PC|lok-PD|maternal nuclear kinase', map_location='38B2-38B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20061422, end_position_on_the_genomic_accession=20064381, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35289, CurrentID=0, Status='live', Symbol='vls', Aliases='Dmel_CG10728, 38B.5, CG107278, CG10728, Dmel\\CG10728, val', description='valois', other_designations='CG10728-PA|vls-PA', map_location='38B2-38B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20064261, end_position_on_the_genomic_accession=20065679, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35291, CurrentID=0, Status='live', Symbol='CG10730', Aliases='Dmel_ 38B.7, Dmel\\CG10730, gi21430366, gi7298655', description='CG10730 gene product from transcript CG10730-RC', other_designations='CG10730-PC|uncharacterized protein', map_location='38B2-38B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20070081, end_position_on_the_genomic_accession=20072195, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35292, CurrentID=0, Status='live', Symbol='pr', Aliases='Dmel_CG16784, 38B.9, CG16784, Dmel\\CG16784, anon-38Bb', description='purple', other_designations='CG16784-PC|pr-PC', map_location='2-54.5 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20073719, end_position_on_the_genomic_accession=20075458, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35293, CurrentID=0, Status='live', Symbol='neb', Aliases='Dmel_CG10718, 38B.10, 38B.12, CG10718, Dm0332, DmKlp38B, DmNeb, Dmel\\CG10718, KIF 14, KIF14, KLP 38B, KLP-38B, KLP38B, Klp38, Klp38B, Mot, Neb, klp38B, l(2)03552, l(2)k00802, l(2)k07614b, sl(2)ry, sl(2)ry3, tio, neb', description='nebbish', other_designations='CG10718-PA|Kinesin-like protein at 38B|mothra|neb-PA|tiovivo', map_location='2-54.5 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20075757, end_position_on_the_genomic_accession=20089698, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35295, CurrentID=0, Status='live', Symbol='CG10747', Aliases='Dmel_ 38B.13, Dmel\\CG10747', description='CG10747 gene product from transcript CG10747-RB', other_designations='CG10747-PA|CG10747-PB|uncharacterized protein', map_location='38B5-38B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20089636, end_position_on_the_genomic_accession=20091348, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35296, CurrentID=0, Status='live', Symbol='CG10721', Aliases='Dmel_ 38B.14, Dmel\\CG10721', description='CG10721 gene product from transcript CG10721-RB', other_designations='CG10721-PA|CG10721-PB|uncharacterized protein', map_location='38B6-38B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20091435, end_position_on_the_genomic_accession=20093152, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35297, CurrentID=0, Status='live', Symbol='Taf13', Aliases='Dmel_CG10756, 38B.15, CG10756, Dmel\\CG10756, TAF, TAF13, TAF18, TAF[[II]], TAF[[II]]18, TFIID, Taf18, dTAF[[II]]18', description='TBP-associated factor 13', other_designations='CG10756-PA|TBP-associated factor|TBP-associated factor 18kD|Taf13-PA', map_location='38B6-38B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20093088, end_position_on_the_genomic_accession=20093849, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35298, CurrentID=0, Status='live', Symbol='nesd', Aliases='Dmel_CG10722, 38B.16, CG10722, Dmel\\CG10722', description='nessun dorma', other_designations='CG10722-PA|CG10722-PB|nesd-PA|nesd-PB', map_location='38B6-38B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20093972, end_position_on_the_genomic_accession=20096253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35299, CurrentID=0, Status='live', Symbol='mRpS18B', Aliases='Dmel_CG10757, 38B.17, CG10757, Dmel\\CG10757, MRP-S18-(2), mRpS18b', description='mitochondrial ribosomal protein S18B', other_designations='CG10757-PA|CG10757-PB|mRpS18B-PA|mRpS18B-PB|mitochondrial ribosomal protein S18b', map_location='38B6-38B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20096329, end_position_on_the_genomic_accession=20097253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35300, CurrentID=0, Status='live', Symbol='Kua', Aliases='Dmel_CG10723, 38B.18, CG10723, Dmel\\CG10723', description='CG10723 gene product from transcript CG10723-RD', other_designations='CG10723-PA|CG10723-PB|CG10723-PC|CG10723-PD|Kua-PA|Kua-PB|Kua-PC|Kua-PD', map_location='38C1-38C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20097581, end_position_on_the_genomic_accession=20103375, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35303, CurrentID=0, Status='live', Symbol='CG10651', Aliases='Dmel_ 38C.2, Dmel\\CG10651', description='CG10651 gene product from transcript CG10651-RB', other_designations='CG10651-PA|CG10651-PB|uncharacterized protein', map_location='38C2-38C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20133755, end_position_on_the_genomic_accession=20134923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35305, CurrentID=0, Status='live', Symbol='La', Aliases='Dmel_CG10922, 38C.38, BEST:LD24519, CG10922, D-La, DLa/SS-B, Dmel\\CG10922, dmLa', description='La autoantigen-like', other_designations='CG10922-PA|La-PA|ribonucleoprotein La', map_location='38C5-38C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20348393, end_position_on_the_genomic_accession=20350012, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35306, CurrentID=0, Status='live', Symbol='RtGEF', Aliases='Dmel_CG10043, 38C.40, ARHGEF7, CG10043, Dmel\\CG10043, DrhoGEF, DrtGEF, Pix, Rho/RacGEF, RhoGEF, beta-PIX, betaPIX, dPix, dpix, pix, rhoGEF, rtGEF', description='Rho-type guanine nucleotide exchange factor', other_designations='CG10043-PA|CG10043-PB|CG10043-PD|CG10043-PE|CG10043-PF|CG10043-PG|CG10043-PH|PAK-interacting exchange factor|RtGEF-PA|RtGEF-PB|RtGEF-PD|RtGEF-PE|RtGEF-PF|RtGEF-PG|RtGEF-PH|drosophila rho-type guanine exchange factor|rho-type guanine exchange factor', map_location='38C5-38C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20350434, end_position_on_the_genomic_accession=20365254, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  35307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35307, CurrentID=0, Status='live', Symbol='Ugt37a1', Aliases='Dmel_CG11012, 38C.41, CG11012, DmUgt37a1, DmeUgt37a1, Dmel\\CG11012', description='UDP-glycosyltransferase 37a1', other_designations='CG11012-PA|Ugt37a1-PA', map_location='38C5-38C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20372409, end_position_on_the_genomic_accession=20374083, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35308, CurrentID=0, Status='live', Symbol='CG16798', Aliases='Dmel_ 38C.41, Dmel\\CG16798', description='CG16798 gene product from transcript CG16798-RA', other_designations='CG16798-PA|uncharacterized protein', map_location='38C5-38C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20376981, end_position_on_the_genomic_accession=20382356, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35309, CurrentID=0, Status='live', Symbol='CG10947', Aliases='Dmel_ 38C.42, Dmel\\CG10947', description='CG10947 gene product from transcript CG10947-RD', other_designations='CG10947-PB|CG10947-PC|CG10947-PD|uncharacterized protein', map_location='38C6-38C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20385293, end_position_on_the_genomic_accession=20414800, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35310, CurrentID=0, Status='live', Symbol='CG10949', Aliases='Dmel_ 38C.46, Dmel\\CG10949, FBgn0032858', description='CG10949 gene product from transcript CG10949-RB', other_designations='CG10949-PA|CG10949-PB|uncharacterized protein', map_location='38C6-38C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20415173, end_position_on_the_genomic_accession=20416918, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35311, CurrentID=0, Status='live', Symbol='Arpc2', Aliases='Dmel_CG10954, 38C.47, ARC-P34, ARC-p34, ARPC2, ARPC2/p34, Arc-p34, ArcP34, CG10954, Dmel\\CG10954, anon-WO0118547.154, p34', description='Actin-related protein 2/3 complex, subunit 2', other_designations='Arc-p34|Arpc2-PA|Arpc2-PB|CG10954-PA|CG10954-PB', map_location='38C6-38C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20417057, end_position_on_the_genomic_accession=20423027, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35312, CurrentID=0, Status='live', Symbol='CG15130', Aliases='Dmel_ 38C.48, Dmel\\CG15130', description='CG15130 gene product from transcript CG15130-RB', other_designations='CG15130-PB|CG15130-PC|uncharacterized protein', map_location='38C6-38C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20423106, end_position_on_the_genomic_accession=20424660, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35314, CurrentID=0, Status='live', Symbol='CG31688', Aliases='Dmel_CG31688, 38C.49, 38C.52, CG11016, CG11017, CG11019, CG18597-PA, Dmel\\CG31688, Dmel_CG11019', description='CG31688 gene product from transcript CG31688-RD', other_designations='CG31688-PB|CG31688-PD|CG31688-PE|CG31688-PF|CG31688-PG|uncharacterized protein', map_location='38C6-38C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20426199, end_position_on_the_genomic_accession=20447905, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35315, CurrentID=0, Status='live', Symbol='Cdc23', Aliases='Dmel_CG2508, 38C.54, APC8, Apc8, CDC23, CG2508, CG2508-PA, Dmel\\CG2508, cdc23', description='Cell division cycle 23 ortholog', other_designations='CG2508-PA|CG2508-PB|Cdc23-PA|Cdc23-PB|cdc23|cell division cycle 23 ortholog', map_location='38C7-38C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20459342, end_position_on_the_genomic_accession=20461729, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35316, CurrentID=0, Status='live', Symbol='CG2493', Aliases='Dmel_ 38C.55, Dmel\\CG2493', description='CG2493 gene product from transcript CG2493-RA', other_designations='CG2493-PA|CG2493-PB', map_location='38C8-38C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20461585, end_position_on_the_genomic_accession=20463444, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35319, CurrentID=0, Status='live', Symbol='CG15475', Aliases='Dmel_ 38D.19, Dmel\\CG15475', description='CG15475 gene product from transcript CG15475-RA', other_designations='CG15475-PA|uncharacterized protein', map_location='38D1-38D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20614522, end_position_on_the_genomic_accession=20615559, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35320, CurrentID=0, Status='live', Symbol='CG17472', Aliases='Dmel_ 38D.23, BcDNA:GH20534, Dmel\\CG17472', description='CG17472 gene product from transcript CG17472-RA', other_designations='CG17472-PA|uncharacterized protein', map_location='38D1-38D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20634515, end_position_on_the_genomic_accession=20635167, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35321, CurrentID=0, Status='live', Symbol='CG17470', Aliases='Dmel_ 38D.25, Dmel\\CG17470, anon-WO0140519.68, anon-WO0153538.51', description='CG17470 gene product from transcript CG17470-RA', other_designations='CG17470-PA|uncharacterized protein', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20642306, end_position_on_the_genomic_accession=20643244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35322, CurrentID=0, Status='live', Symbol='phr6-4', Aliases='Dmel_CG2488, 38D.25, 64PHR, CG2488, D6-4, Dm 64, Dmel\\CG2488, phr(6-4)', description='(6-4)-photolyase', other_designations='(6-4) photolyase|(6-4)photolyase|6-4 photolyase|CG2488-PA|CG2488-PB|CG2488-PC|CG2488-PD|photolyase|phr6-4-PA|phr6-4-PB|phr6-4-PC|phr6-4-PD', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20643351, end_position_on_the_genomic_accession=20645791, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35323, CurrentID=0, Status='live', Symbol='CG2608', Aliases='Dmel_ 38D.26, Dmel\\CG2608', description='CG2608 gene product from transcript CG2608-RA', other_designations='CG2608-PA', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20645822, end_position_on_the_genomic_accession=20647381, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35324, CurrentID=0, Status='live', Symbol='CG2611', Aliases='Dmel_ 38D.27, Dmel\\CG2611', description='CG2611 gene product from transcript CG2611-RA', other_designations='CG2611-PA', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20650189, end_position_on_the_genomic_accession=20650898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35325, CurrentID=0, Status='live', Symbol='bru', Aliases='Dmel_CG2478, 38D.28, CG2478, Dmel\\CG2478', description='brunelleschi', other_designations='CG2478-PA|bru-PA', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20650755, end_position_on_the_genomic_accession=20656178, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35326, CurrentID=0, Status='live', Symbol='CG2614', Aliases='Dmel_ 38D.29, Dmel\\CG2614', description='CG2614 gene product from transcript CG2614-RA', other_designations='CG2614-PA|CG2614-PC', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20656426, end_position_on_the_genomic_accession=20659218, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35327, CurrentID=0, Status='live', Symbol='CG31678', Aliases='Dmel_ 38D.30, CG17465, CG17466, Dmel\\CG31678', description='CG31678 gene product from transcript CG31678-RB', other_designations='CG31678-PB|CG31678-PC|CG31678-PE|uncharacterized protein', map_location='38D2-38D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20659132, end_position_on_the_genomic_accession=20675941, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35329, CurrentID=0, Status='live', Symbol='IKKepsilon', Aliases='Dmel_CG2615, 38D.31, APTX7, CG2615, DIK2, Dik2, DmIKKepsilon, DmIKKepsilon/dIK2, Dmel\\CG2615, Dmik2, IK2, IKK, Ik2, Ik2/DmIKK, dIKK, dik2, ik2, l(2)38Ea', description='I-kappaB kinase epsilon', other_designations='CG2615-PB|CG2615-PC|IKK-related kinase|IKKepsilon-PB|IKKepsilon-PC|IkappaB kinase-like 2|apoptoxin 7|ik2|lethal (2) 38Ea', map_location='38D4-38E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20676045, end_position_on_the_genomic_accession=20678979, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35330, CurrentID=0, Status='live', Symbol='Cen', Aliases='Dmel_CG1962, 38E.1, CG1962, Dmel\\CG1962, cen', description='Centrocortin', other_designations='CG1962-PA|CG1962-PB|Cen-PA|Cen-PB|centrocortin', map_location='38D5-38D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20678921, end_position_on_the_genomic_accession=20681846, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35331, CurrentID=0, Status='live', Symbol='CG2617', Aliases='Dmel_ 38E.2, Dmel\\CG2617', description='CG2617 gene product from transcript CG2617-RA', other_designations='CG2617-PA|CG2617-PB', map_location='38D5-38D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20682368, end_position_on_the_genomic_accession=20683629, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35332, CurrentID=0, Status='live', Symbol='Hr38', Aliases='Dmel_CG1864, 38E.3, 38E.7, CG1864, DHR38, Dhr38, Dmel\\CG1864, HR38, NR4A4, dHR38, hr38, l(2)02306', description='Hormone receptor-like in 38', other_designations='CG1864-PB|CG1864-PD|Hr38-PB|Hr38-PD|hormone receptor 38|hormone receptor-like in 38', map_location='38D5-38E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20683462, end_position_on_the_genomic_accession=20714588, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35333, CurrentID=0, Status='live', Symbol='CG9316', Aliases='Dmel_ 38E.9, Dmel\\CG9316, FBXL10', description='CG9316 gene product from transcript CG9316-RA', other_designations='CG9316-PA', map_location='38E4-38E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20724209, end_position_on_the_genomic_accession=20726551, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35334, CurrentID=0, Status='live', Symbol='CG9317', Aliases='Dmel_ 38E.10, Dmel\\CG9317', description='CG9317 gene product from transcript CG9317-RA', other_designations='CG9317-PA|CG9317-PC', map_location='38E4-38E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20727151, end_position_on_the_genomic_accession=20730282, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35335, CurrentID=0, Status='live', Symbol='TM9SF2', Aliases='Dmel_CG9318, 38E.10, CG9318, Dmel\\CG9318', description='Transmembrane 9 superfamily protein member 2', other_designations='CG9318-PA|CG9318-PB|TM9SF2-PA|TM9SF2-PB|transmembrane 9 superfamily protein member 2', map_location='38E4-38E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20730416, end_position_on_the_genomic_accession=20734309, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35336, CurrentID=0, Status='live', Symbol='Fs(2)Ket', Aliases='Dmel_CG2637, 38E.11, CG2637, Dmel\\CG2637, Fs(2)Hon, Fs(2)Sz6, Fs(2)Sz7, Fs(2)ket, Hon, IMPbeta1, Impbeta, Ket, Ketel, Kpnb1, MRE26, dKap-beta, fs(2)Ket, fs(2)lo-B, fs(2)lto2, fs(2)ltoHD43, imp-beta, impbeta, ketel, kettle', description='Female sterile (2) Ketel', other_designations='CG2637-PA|CG2637-PB|CG2637-PC|CG2637-PD|CG2637-PE|DImportin-beta|Fs(2)Ket-PA|Fs(2)Ket-PB|Fs(2)Ket-PC|Fs(2)Ket-PD|Fs(2)Ket-PE|Importin-beta|female sterile (2) ketel|female sterile(2)lto2|fs(2) ketel|hont|importin beta|importin beta1|ketel|mRNA-like ncRNA in embryogenesis 26', map_location='38E5-38E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20734609, end_position_on_the_genomic_accession=20750941, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  35337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35337, CurrentID=0, Status='live', Symbol='CG9319', Aliases='Dmel_ 38E.13, Dmel\\CG9319', description='CG9319 gene product from transcript CG9319-RB', other_designations='CG9319-PA|CG9319-PB', map_location='38E6-38E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20750934, end_position_on_the_genomic_accession=20752341, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35338, CurrentID=0, Status='live', Symbol='Ns4', Aliases='Dmel_CG9320, 38E.14, CG9320, Dmel\\CG9320, NS4, dmNS4, ns4', description='Nucleostemin ortholog (H. sapiens) 4', other_designations='CG9320-PA|Ns4-PA|nucleostemin 4|nucleostemin ortholog 4', map_location='38E6-38E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20752476, end_position_on_the_genomic_accession=20754655, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35339, CurrentID=0, Status='live', Symbol='CG9323', Aliases='Dmel_ 38E.15, Dmel\\CG9323, cg9323', description='CG9323 gene product from transcript CG9323-RB', other_designations='CG9323-PA|CG9323-PB', map_location='38E7-38E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20754579, end_position_on_the_genomic_accession=20757985, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35340, CurrentID=0, Status='live', Symbol='dia', Aliases='Dmel_CG1768, 38E.16, CG1768, DIA, Dia, Dias, Dmel\\CG1768, l(2)k07135, ms(2)04138', description='diaphanous', other_designations='CG1768-PA|CG1768-PB|CG1768-PC|CG1768-PD|CG1768-PE|dia-PA|dia-PB|dia-PC|dia-PD|dia-PE', map_location='38E7-38E8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20758144, end_position_on_the_genomic_accession=20768053, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35341, CurrentID=0, Status='live', Symbol='cad', Aliases='Dmel_CG1759, 38E.19, CAD, CG1759, Cad, Dmel\\CG1759, S67, anon-WO2004063362.83, cd', description='caudal', other_designations='CG1759-PA|CG1759-PB|CG1759-PC|cad-PA|cad-PB|cad-PC', map_location='38E9-38E10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20770731, end_position_on_the_genomic_accession=20783135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35342, CurrentID=0, Status='live', Symbol='Pomp', Aliases='Dmel_CG9324, 38E.20, CG9324, Dmel\\CG9324, POMP, UMP1, pomp', description='CG9324 gene product from transcript CG9324-RA', other_designations='20S maturase|CG9324-PA|Pomp-PA|maturase|pomp|proteosome maturation protein', map_location='38E10-38E10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20787656, end_position_on_the_genomic_accession=20789098, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35343, CurrentID=0, Status='live', Symbol='vari', Aliases='Dmel_CG9326, 38E.21, CG9326, Dmel\\CG9326, Vari, l(2)03953, l(2)03953b, l(2)38EFa, szar', description='varicose', other_designations="CG9326-PB|CG9326-PC|CG9326-PD|CG9326-PE|CG9326-PF|lethal (2) 38EFa|senz'aria|senz-aria|vari-PB|vari-PC|vari-PD|vari-PE|vari-PF", map_location='39D3-40B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20788662, end_position_on_the_genomic_accession=20797042, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35344, CurrentID=0, Status='live', Symbol='CG9328', Aliases='Dmel_ 38E.22, BEST:LD01639, BcDNA:SD01241, Dmel\\CG9328, LD01639', description='CG9328 gene product from transcript CG9328-RB', other_designations='CG9328-PA|CG9328-PB', map_location='38E10-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20797715, end_position_on_the_genomic_accession=20814331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35346, CurrentID=0, Status='live', Symbol='CheB38c', Aliases='Dmel_CG14405, CG14405, Dmel\\CG14405', description='Chemosensory protein B 38c', other_designations='CG14405-PA|CheB38c-PA|chemosensory protein B 38c', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20820089, end_position_on_the_genomic_accession=20820867, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35347, CurrentID=0, Status='live', Symbol='CG9331', Aliases='Dmel_ Dmel\\CG9331', description='CG9331 gene product from transcript CG9331-RK', other_designations='CG9331-PA|CG9331-PB|CG9331-PC|CG9331-PD|CG9331-PF|CG9331-PG|CG9331-PH|CG9331-PI|CG9331-PJ|CG9331-PK|CG9331-PL|CG9331-PM', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20822828, end_position_on_the_genomic_accession=20827744, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35348, CurrentID=0, Status='live', Symbol='CG31673', Aliases='Dmel_ BcDNA:LD14730, CCG31673, CG9332, Dmel\\CG31673', description='CG31673 gene product from transcript CG31673-RB', other_designations='CG31673-PA|CG31673-PB|uncharacterized protein', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20829609, end_position_on_the_genomic_accession=20831161, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35349, CurrentID=0, Status='live', Symbol='Oseg5', Aliases='Dmel_CG9333, CG9333, CHE-2, Che-2, Dmel\\CG9333, IFT80, OSEG, OSEG5, oseg5', description='Outer segment 5', other_designations='CG9333-PB|Oseg5-PB|outer segment 5', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20832849, end_position_on_the_genomic_accession=20835840, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35351, CurrentID=0, Status='live', Symbol='CG31676', Aliases='Dmel_ BcDNA:RH34376, CG14403, CG14404, Dmel\\CG31676', description='CG31676 gene product from transcript CG31676-RB', other_designations='CG31676-PA|CG31676-PB|uncharacterized protein', map_location='38F2-38F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20840750, end_position_on_the_genomic_accession=20849182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35352, CurrentID=0, Status='live', Symbol='CG14402', Aliases='Dmel_ BcDNA:AT08210, Dmel\\CG14402', description='CG14402 gene product from transcript CG14402-RA', other_designations='CG14402-PA|CG14402-PB|uncharacterized protein', map_location='38F2-38F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20849300, end_position_on_the_genomic_accession=20850987, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35353, CurrentID=0, Status='live', Symbol='twit', Aliases='Dmel_CG9335, BcDNA:LD40063, CG9335, Dmel\\CG9335', description='target of wit', other_designations='CG9335-PA|twit-PA', map_location='38F2-38F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20852578, end_position_on_the_genomic_accession=20857113, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35354, CurrentID=0, Status='live', Symbol='CG14400', Aliases='Dmel_ Dmel\\CG14400', description='CG14400 gene product from transcript CG14400-RA', other_designations='CG14400-PA|uncharacterized protein', map_location='38F3-38F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20856987, end_position_on_the_genomic_accession=20857891, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35355, CurrentID=0, Status='live', Symbol='CG9336', Aliases='Dmel_ BcDNA:RE71975, Dmel\\CG9336', description='CG9336 gene product from transcript CG9336-RB', other_designations='CG9336-PA|CG9336-PB', map_location='38F3-38F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20858695, end_position_on_the_genomic_accession=20861014, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35357, CurrentID=0, Status='live', Symbol='CG9338', Aliases='Dmel_ BcDNA:GH07967, Dmel\\CG9338', description='CG9338 gene product from transcript CG9338-RB', other_designations='CG9338-PA|CG9338-PB', map_location='38F3-38F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20865215, end_position_on_the_genomic_accession=20867709, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35358, CurrentID=0, Status='live', Symbol='CG14401', Aliases='Dmel_ Dmel\\CG14401', description='CG14401 gene product from transcript CG14401-RA', other_designations='CG14401-PA|CG14401-PB|CG14401-PC|uncharacterized protein', map_location='38F3-38F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20869854, end_position_on_the_genomic_accession=20871565, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35359, CurrentID=0, Status='live', Symbol='sky', Aliases='Dmel_CG9339, CG14399, CG9339, CG9340, Dmel\\CG9339, anon-WO0118547.455', description='skywalker', other_designations='CG9339-PA|CG9339-PB|CG9339-PC|CG9339-PD|CG9339-PE|CG9339-PF|CG9339-PG|CG9339-PH|CG9339-PI|CG9339-PJ|CG9339-PK|CG9339-PL|sky-PA|sky-PB|sky-PC|sky-PD|sky-PE|sky-PF|sky-PG|sky-PH|sky-PI|sky-PJ|sky-PK|sky-PL', map_location='38F3-38F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20871335, end_position_on_the_genomic_accession=20918931, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  35362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35362, CurrentID=0, Status='live', Symbol='Mtp', Aliases='Dmel_CG9342, 2L3637, 2L4501, CG9342, CT3751, Dmel\\CG9342, MTP, dMTP, mtp', description='Microsomal triacylglycerol transfer protein', other_designations='CG9342-PA|Mtp-PA|microsomal triacylglycerol transfer protein|microsomal triglyceride transfer protein', map_location='38F5-38F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20919094, end_position_on_the_genomic_accession=20922876, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35364, CurrentID=0, Status='live', Symbol='RPA2', Aliases='Dmel_CG9273, 153194_at, CG9273, Dmel\\CG9273, RpA2, dRPA2', description='Replication protein A2', other_designations='CG9273-PA|CG9273-PB|RPA2-PA|RPA2-PB|replication protein A, subunit 2|replication protein A2', map_location='38F6-38F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20924975, end_position_on_the_genomic_accession=20926147, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35365, CurrentID=0, Status='live', Symbol='CG9272', Aliases='Dmel_ 9272, Dmel\\CG9272', description='CG9272 gene product from transcript CG9272-RA', other_designations='CG9272-PA', map_location='38F6-38F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20926319, end_position_on_the_genomic_accession=20927860, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35366, CurrentID=0, Status='live', Symbol='CG9270', Aliases='Dmel_ Dme_CG9270, Dmel\\CG9270', description='CG9270 gene product from transcript CG9270-RC', other_designations='CG9270-PC', map_location='38F6-38F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20929456, end_position_on_the_genomic_accession=20934113, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35368, CurrentID=0, Status='live', Symbol='Itgbn', Aliases='Dmel_CG1762, Beta[[nu]], CG1762, CT5192, Dmel\\CG1762, Fly betanu, Itgbetanu, beta-nu, betaInt, betaInt nu, betaInt-nu, betaInt[nu], betaV-integrin, beta[[v]], beta[nu]-integrin, betagr;-nu, betanu, betav', description='Integrin betanu subunit', other_designations='CG1762-PA|CG1762-PB|Itgbn-PA|Itgbn-PB|beta[nu] integrin|beta[v]integrin|betanu integrin|betav integrin|integrin beta neu|integrin beta v|integrin betanu subunit|integrin bn|integrinbetanu', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21053033, end_position_on_the_genomic_accession=21058044, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35369, CurrentID=0, Status='live', Symbol='CG9265', Aliases='Dmel_ Dmel\\CG9265', description='CG9265 gene product from transcript CG9265-RB', other_designations='CG9265-PA|CG9265-PB|CG9265-PC|CG9265-PD', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21057706, end_position_on_the_genomic_accession=21071502, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35370, CurrentID=0, Status='live', Symbol='tadr', Aliases='Dmel_CG9264, CG9264, Dmel\\CG9264', description='torn and diminished rhabdomeres', other_designations='CG9264-PC|CG9264-PD|tadr-PC|tadr-PD', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21075405, end_position_on_the_genomic_accession=21081053, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35372, CurrentID=0, Status='live', Symbol='CG9259', Aliases='Dmel_ Dmel\\CG9259', description='CG9259 gene product from transcript CG9259-RC', other_designations='CG9259-PA|CG9259-PB|CG9259-PC', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21089375, end_position_on_the_genomic_accession=21090999, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35374, CurrentID=0, Status='live', Symbol='CG12050', Aliases='Dmel_ Dmel\\CG12050', description='CG12050 gene product from transcript CG12050-RB', other_designations='CG12050-PB|uncharacterized protein', map_location='39A1-39A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21094966, end_position_on_the_genomic_accession=21102429, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35375, CurrentID=0, Status='live', Symbol='CG9257', Aliases='Dmel_ Dmel\\CG9257', description='CG9257 gene product from transcript CG9257-RA', other_designations='CG9257-PA', map_location='39A2-39A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21102743, end_position_on_the_genomic_accession=21104321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35376, CurrentID=0, Status='live', Symbol='Nhe2', Aliases='Dmel_CG9256, CG9255, CG9256, DmNHE2, Dmel\\CG9256, Dnhe2, NHE2, nhe2', description='Na[+]/H[+] hydrogen exchanger 2', other_designations='CG9256-PA|CG9256-PB|CG9256-PC|CG9256-PD|CG9256-PE|CG9256-PF|CG9256-PG|CG9256-PH|CG9256-PI|CG9256-PJ|CG9256-PK|CG9256-PL|Nhe2-PA|Nhe2-PB|Nhe2-PC|Nhe2-PD|Nhe2-PE|Nhe2-PF|Nhe2-PG|Nhe2-PH|Nhe2-PI|Nhe2-PJ|Nhe2-PK|Nhe2-PL', map_location='39A2-39A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21104568, end_position_on_the_genomic_accession=21134255, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  35378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35378, CurrentID=0, Status='live', Symbol='Dap160', Aliases='Dmel_CG1099, CG1099, DAP-160, DAP160/intersectin, Dmel\\CG1099, dap160, noncoding_4110, p160, Dap160', description='Dynamin associated protein 160', other_designations='CG1099-PA|CG1099-PB|CG1099-PC|CG1099-PD|CG1099-PG|CG1099-PH|Dap160-PA|Dap160-PB|Dap160-PC|Dap160-PD|Dap160-PG|Dap160-PH|Dynamin-associated protein 160 kDa|dynamin associated protein 160|intersectin', map_location='39A4-39A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21134439, end_position_on_the_genomic_accession=21142347, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  35379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35379, CurrentID=0, Status='live', Symbol='CG9253', Aliases='Dmel_ DmRH24, Dmel\\CG9253, anon-EST:GressD10, cg9253', description='CG9253 gene product from transcript CG9253-RA', other_designations='CG9253-PA', map_location='39A4-39A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21143229, end_position_on_the_genomic_accession=21145146, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35380, CurrentID=0, Status='live', Symbol='del', Aliases='Dmel_CG9252, CG9252, Del, Dmel\\CG9252', description='deadlock', other_designations='CG9252-PA|CG9252-PB|CG9252-PC|del-PA|del-PB|del-PC', map_location='39A4-39A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21145137, end_position_on_the_genomic_accession=21148546, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35381, CurrentID=0, Status='live', Symbol='E2f2', Aliases='Dmel_CG1071, CG1071, Dmel\\CG1071, E2F, E2F2, E2f, dE2F-2, dE2F2, dE2f2, de2f2, e2f2', description='E2F transcription factor 2', other_designations='CG1071-PA|CG1071-PB|E2f2-PA|E2f2-PB', map_location='39A5-39A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21154293, end_position_on_the_genomic_accession=21156749, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35382, CurrentID=0, Status='live', Symbol='Mpp6', Aliases='Dmel_CG9250, BcDNA:RE45276, CG9250, Dmel\\CG9250, MPP6, Mpps, dMPP6, l(2)k16403, mpp6', description='M-phase phosphoprotein 6', other_designations='CG9250-PA|CG9250-PB|Mpp6-PA|Mpp6-PB', map_location='39A6-39A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21156288, end_position_on_the_genomic_accession=21157685, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35383, CurrentID=0, Status='live', Symbol='CG9249', Aliases='Dmel_ Dmel\\CG9249', description='CG9249 gene product from transcript CG9249-RA', other_designations='CG9249-PA|CG9249-PB', map_location='39A6-39A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21157433, end_position_on_the_genomic_accession=21158611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35384, CurrentID=0, Status='live', Symbol='CG9248', Aliases='Dmel_ Dmel\\CG9248', description='CG9248 gene product from transcript CG9248-RB', other_designations='CG9248-PA|CG9248-PB|CG9248-PC', map_location='39A6-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21158847, end_position_on_the_genomic_accession=21160903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35385, CurrentID=0, Status='live', Symbol='Nbr', Aliases='Dmel_CG9247, CG9247, Dmel\\CG9247, nbr', description='Nibbler', other_designations='CG9247-PA|Nbr-PA|nibbler', map_location='39A6-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21160636, end_position_on_the_genomic_accession=21163083, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35386, CurrentID=0, Status='live', Symbol='CG9246', Aliases='Dmel_ Dmel\\CG9246', description='CG9246 gene product from transcript CG9246-RA', other_designations='CG9246-PA', map_location='39A7-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21163641, end_position_on_the_genomic_accession=21166090, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35387, CurrentID=0, Status='live', Symbol='CG43345', Aliases='Dmel_ CG9243, Dmel\\CG43345, Dmel_CG9243', description='CG43345 gene product from transcript CG43345-RC', other_designations='CG43345-PB|CG43345-PC|uncharacterized protein', map_location='39A7-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21166046, end_position_on_the_genomic_accession=21168581, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35390, CurrentID=0, Status='live', Symbol='Mcm10', Aliases='Dmel_CG9241, CG9241, DmMCM10, Dmel\\CG9241, Scim19, dMcm10, mcm10', description='Minichromosome maintenance 10', other_designations='CG9241-PA|Mcm10-PA|minichromosome maintenance 10|sensitized chromosome inheritance modifier 19', map_location='39B1-39B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21179933, end_position_on_the_genomic_accession=21182812, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35393, CurrentID=0, Status='live', Symbol='CG31624', Aliases='Dmel_ CG17401, Dmel\\CG31624, Tes88', description='CG31624 gene product from transcript CG31624-RC', other_designations='CG31624-PA|CG31624-PB|CG31624-PC|uncharacterized protein', map_location='39B1-39B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21199045, end_position_on_the_genomic_accession=21200375, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35394, CurrentID=0, Status='live', Symbol='clumsy', Aliases='Dmel_CG8681, CG8681, CT9037, Clumsy, Dmel\\CG8681, Dmel GluR39B', description='CG8681 gene product from transcript CG8681-RB', other_designations='CG8681-PA|CG8681-PB|clumsy-PA|clumsy-PB', map_location='39B2-39B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21206493, end_position_on_the_genomic_accession=21211818, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  35395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35395, CurrentID=0, Status='live', Symbol='CG8679', Aliases='Dmel_ Dmel\\CG8679, dLEM3', description='CG8679 gene product from transcript CG8679-RA', other_designations='CG8679-PA', map_location='39B3-39B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21216004, end_position_on_the_genomic_accession=21218544, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35396, CurrentID=0, Status='live', Symbol='Atg18b', Aliases='Dmel_CG8678, ATG18.2, Atg18, Atg18-like, CG8678, Dmel\\CG8678, l(2)k09410', description='Autophagy-related 18b', other_designations='Atg18b-PA|Atg18b-PB|Autophagy-specific gene 18b|CG8678-PA|CG8678-PB', map_location='39B3-39B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21212322, end_position_on_the_genomic_accession=21220572, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35397, CurrentID=0, Status='live', Symbol='CG8677', Aliases='Dmel_ BEST:LD14959, BEST:LD19244, Dmel\\CG8677, LD14959, dRsf-1', description='CG8677 gene product from transcript CG8677-RC', other_designations='CG8677-PB|CG8677-PC|HBXAP ortholog|remodeling and spacing factor', map_location='39B3-39B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21221312, end_position_on_the_genomic_accession=21232059, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35398, CurrentID=0, Status='live', Symbol='Hr39', Aliases='Dmel_CG8676, CG8676, DHR30, DHR39, DHR39/Ftz-F1beta, Dmel\\CG8676, EP(2)2490, FTZ-F1B, FTZ-F1beta, HR39, NR5B1, betaFTZ-E1, dFTZ-F1beta, fs(2)04443, fs(2)neo8, ftz-F1beta, ftz-f1beta, i70', description='Hormone receptor-like in 39', other_designations='CG8676-PA|CG8676-PB|CG8676-PC|CG8676-PD|Hr39-PA|Hr39-PB|Hr39-PC|Hr39-PD|NR5A-class nuclear hormone receptor|fs(2)04443|fs(2)neo8|hormone receptor like in 39|hormone receptor-like in 39', map_location='39C-60', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21237250, end_position_on_the_genomic_accession=21259675, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35399, CurrentID=0, Status='live', Symbol='l(2)k14505', Aliases='Dmel_CG8674, CG8674, Dmel\\CG8674', description='lethal (2) k14505', other_designations='CG8674-PA|CG8674-PB|l(2)k14505-PA|l(2)k14505-PB', map_location='39C1-39C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21261168, end_position_on_the_genomic_accession=21262411, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35400, CurrentID=0, Status='live', Symbol='CG8671', Aliases='Dmel_ Dmel\\CG8671, anon-WO0118547.182', description='CG8671 gene product from transcript CG8671-RG', other_designations='CG8671-PA|CG8671-PB|CG8671-PC|CG8671-PD|CG8671-PE|CG8671-PF|CG8671-PG', map_location='39C1-39C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21262956, end_position_on_the_genomic_accession=21282790, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  35402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35402, CurrentID=0, Status='live', Symbol='Mondo', Aliases='Dmel_CG18362, 154826_at, CG18362, ChREBP, Dmel\\CG18362, Mio, anon-39Da, bHLHd12, dmondo, mio', description='CG18362 gene product from transcript CG18362-RL', other_designations='CG18362-PA|CG18362-PB|CG18362-PE|CG18362-PF|CG18362-PG|CG18362-PH|CG18362-PI|CG18362-PJ|CG18362-PK|CG18362-PL|CG18362-PM|Mlx interactor|Mondo-PA|Mondo-PB|Mondo-PE|Mondo-PF|Mondo-PG|Mondo-PH|Mondo-PI|Mondo-PJ|Mondo-PK|Mondo-PL|Mondo-PM|WBSCR14-related|mondo', map_location='39C1-39C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21286078, end_position_on_the_genomic_accession=21310050, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  35404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35404, CurrentID=0, Status='live', Symbol='dimm', Aliases='Dmel_CG8667, CG8667, DIMM, Dimm, Dmel\\CG8667, Mistr, bHLHa16, c929, dim', description='dimmed', other_designations='CG8667-PA|CG8667-PB|dimm-PA|dimm-PB|mist 1-related', map_location='39D1-39D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21331756, end_position_on_the_genomic_accession=21342155, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35405, CurrentID=0, Status='live', Symbol='Tsp39D', Aliases='Dmel_CG8666, CG8666, Dm.Tsp39D, Dmel\\CG8666', description='Tetraspanin 39D', other_designations='CG8666-PB|Tsp39D-PB|tetraspanin 39D', map_location='39D1-39D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21344607, end_position_on_the_genomic_accession=21358271, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35407, CurrentID=0, Status='live', Symbol='CG8665', Aliases='Dmel_ Dmel\\CG8665', description='CG8665 gene product from transcript CG8665-RA', other_designations='CG8665-PA', map_location='39D2-39D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21371227, end_position_on_the_genomic_accession=21377995, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35408, CurrentID=0, Status='live', Symbol='nrv3', Aliases='Dmel_CG8663, CG8663, Dmel\\CG8663, JYbeta1, Nrv3, nervana3', description='nervana 3', other_designations='CG8663-PA|CG8663-PC|CG8663-PD|CG8663-PE|Na,K-ATPase beta-subunit|nrv3-PA|nrv3-PC|nrv3-PD|nrv3-PE', map_location='39D2-39D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21380000, end_position_on_the_genomic_accession=21393319, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35409, CurrentID=0, Status='live', Symbol='CG17571', Aliases='Dmel_ 38C.22, Dmel\\CG17571, SP83, anon-38C.22', description='CG17571 gene product from transcript CG17571-RB', other_designations='CG17571-PA|CG17571-PB|anon-38C.22|uncharacterized protein', map_location='38C3-38C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20255988, end_position_on_the_genomic_accession=20257192, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35410, CurrentID=0, Status='live', Symbol='CG17570', Aliases='Dmel_ 38C.24, Dmel\\CG17570, anon-38C.24', description='CG17570 gene product from transcript CG17570-RA', other_designations='CG17570-PA|anon-38C.24|uncharacterized protein', map_location='38C3-38C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20263419, end_position_on_the_genomic_accession=20264424, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35411, CurrentID=0, Status='live', Symbol='Lamp1', Aliases='Dmel_CG3305, CG3305, Dmel\\CG3305, LAMP, LAMP1, Lamp, Lamp-1, dLAMP, dLamp, lamp1', description='CG3305 gene product from transcript CG3305-RA', other_designations='CG3305-PA|Lamp1-PA|lysosomal-associated membrane protein 1|lysosome-associated membrane protein', map_location='39E2-39E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21573389, end_position_on_the_genomic_accession=21577103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35414, CurrentID=0, Status='live', Symbol='nompB', Aliases='Dmel_CG12548, 146590_s_at, CG12548, CG14465, CG15190, CG4137, Dmel\\CG12548, IFT88, NOMPB, OSM-5, Tg737, nomp, osm-5', description='no mechanoreceptor potential B', other_designations='CG12548-PC|CG12548-PD|NompB|nompB-PC|nompB-PD', map_location='39E2-39E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21578292, end_position_on_the_genomic_accession=21589423, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35417, CurrentID=0, Status='live', Symbol='CG2201', Aliases='Dmel_ Dmel\\CG2201', description='CG2201 gene product from transcript CG2201-RF', other_designations='CG2201-PA|CG2201-PB|CG2201-PC|CG2201-PE|CG2201-PF|CG2201-PG', map_location='39E3-39E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21614232, end_position_on_the_genomic_accession=21624080, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35418, CurrentID=0, Status='live', Symbol='Df31', Aliases='Dmel_CG2207, BEST:LD04967, CG2207, DF 31, DF31, Dmel\\CG2207, LD04967, anon-EST:Liang-1.46, anon-EST:Posey126, anon-EST:fe1A4, anon-WO0153538.53, anon1A4, clone 1.46, l(2)k05815', description='Decondensation factor 31', other_designations='CG2207-PA|CG2207-PB|CG2207-PF|Df31-PA|Df31-PB|Df31-PF|anon-fast-evolving-1A4|chromatin decondensation factor 31|decondensation factor 31', map_location='39E3-39E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21626309, end_position_on_the_genomic_accession=21629932, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35419, CurrentID=0, Status='live', Symbol='Ac3', Aliases='Dmel_CG1506, AC 3, AC3, CG1506, DAC3, DAC39E, Dmel\\CG1506', description='CG1506 gene product from transcript CG1506-RA', other_designations='Ac3-PA|Ac3-PC|CG1506-PA|CG1506-PC|adenylate cyclase 3', map_location='39E3-39E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21632587, end_position_on_the_genomic_accession=21647964, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  35420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35420, CurrentID=0, Status='live', Symbol='Cul2', Aliases='Dmel_CG1512, CG1512, CUL2, Cul-2, Cullin2, DMcul-2, Dmel\\CG1512, cul-2, cul2, dCul2, l(2)02074', description='Cullin 2', other_designations='CG1512-PA|CG1512-PB|Cul2-PA|Cul2-PB|Cullin-2|cul-2|cullin 2', map_location='39E3-39E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21658735, end_position_on_the_genomic_accession=21663698, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35421, CurrentID=0, Status='live', Symbol='CG2225', Aliases='Dmel_ Dmel\\CG2225', description='CG2225 gene product from transcript CG2225-RH', other_designations='CG2225-PA|CG2225-PB|CG2225-PC|CG2225-PE|CG2225-PF|CG2225-PG|CG2225-PH', map_location='39E6-39E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21663806, end_position_on_the_genomic_accession=21677530, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  35422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35422, CurrentID=0, Status='live', Symbol='EF2', Aliases='Dmel_CG2238, CG2238, Dmel\\CG2238, EF-2, EF-2bb, Ef2, Ef2b, anon-EST:Liang-1.44, chr2L:21668915..21669164, clone 1.44, eEF2, eF2, ef2b, EF2', description='Elongation factor 2', other_designations='CG2238-PA|CG2238-PC|CG2238-PD|EF2-PA|EF2-PC|EF2-PD|elongation factor 2|elongation factor 2b|eukaryotic translation elongation factor 2b', map_location='39E7-39E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21677913, end_position_on_the_genomic_accession=21682164, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35424, CurrentID=0, Status='live', Symbol='nolo', Aliases='Dmel_CG31619, CG13236, CG2122, CG2131, CG31619, CT6940, Dmel\\CG31619', description='no long nerve cord', other_designations='CG31619-PA|CG31619-PB|CG31619-PC|CG31619-PD|CG31619-PE|CG31619-PF|CG31619-PG|nolo-PA|nolo-PB|nolo-PC|nolo-PD|nolo-PE|nolo-PF|nolo-PG', map_location='39F1-39F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21684106, end_position_on_the_genomic_accession=21729051, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  35425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35425, CurrentID=0, Status='live', Symbol='step', Aliases='Dmel_CG11628, CG11628, CG11633, CYH1, Dmel\\CG11628, GPH, GRP1, GRP1/cytohesin 1, Grp1, Step, cytohesin/GRP1, l(2)SH0323, l(2)SH2 0323, l(2)k08110k, step', description='steppke', other_designations='CG11628-PA|CG11628-PC|CG11628-PD|CG11628-PE|general receptor for phosphoinositides 1|general receptor for phosphoinositides-1|lethal (2) SH0323|lethal (2) k08110|step-PA|step-PC|step-PD|step-PE', map_location='40A1-39F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21740041, end_position_on_the_genomic_accession=21757466, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35426, CurrentID=0, Status='live', Symbol='CG1416', Aliases='Dmel_ Dmel\\CG1416', description='CG1416 gene product from transcript CG1416-RA', other_designations='CG1416-PA|CG1416-PB|CG1416-PC', map_location='40A2-40A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21758010, end_position_on_the_genomic_accession=21761000, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35427, CurrentID=0, Status='live', Symbol='CG31612', Aliases='Dmel_ CG11631, CG11632, Dmel\\CG31612', description='CG31612 gene product from transcript CG31612-RA', other_designations='CG31612-PA|CG31612-PB|uncharacterized protein', map_location='40A2-40A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21762314, end_position_on_the_genomic_accession=21795447, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35429, CurrentID=0, Status='live', Symbol='CG11630', Aliases='Dmel_ Dmel\\CG11630, gi7302136', description='CG11630 gene product from transcript CG11630-RA', other_designations='CG11630-PA|uncharacterized protein', map_location='40A3-40A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21782046, end_position_on_the_genomic_accession=21784065, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35430, CurrentID=0, Status='live', Symbol='tsh', Aliases='Dmel_CG1374, CG1374, Dmel\\CG1374, T shirt, Tsh, ae, ae[l], l(2)04319, l(2)B4-2-12', description='teashirt', other_designations='CG1374-PA|T-shirt|aeroplane|tsh-PA', map_location='2-55.8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21828593, end_position_on_the_genomic_accession=21837011, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35431, CurrentID=0, Status='live', Symbol='CG11629', Aliases='Dmel_ Dmel\\CG11629', description='CG11629 gene product from transcript CG11629-RA', other_designations='CG11629-PA|uncharacterized protein', map_location='40B3-40B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21896666, end_position_on_the_genomic_accession=21898289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35432, CurrentID=0, Status='live', Symbol='CG1421', Aliases='Dmel_ BcDNA:LP11549, Dmel\\CG1421', description='CG1421 gene product from transcript CG1421-RA', other_designations='CG1421-PA', map_location='40B4-40B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21902993, end_position_on_the_genomic_accession=21903739, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35433, CurrentID=0, Status='live', Symbol='CG1428', Aliases='Dmel_ Dmel\\CG1428, gi19527787, gi20129693', description='CG1428 gene product from transcript CG1428-RA', other_designations='CG1428-PA', map_location='40B4-40B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21906289, end_position_on_the_genomic_accession=21908319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35434, CurrentID=0, Status='live', Symbol='CG11634', Aliases='Dmel_ BcDNA:AT09142, Dmel\\CG11634', description='CG11634 gene product from transcript CG11634-RB', other_designations='CG11634-PA|CG11634-PB|uncharacterized protein', map_location='40B4-40B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21916300, end_position_on_the_genomic_accession=21920324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35435, CurrentID=0, Status='live', Symbol='CG2528', Aliases='Dmel_ Dmel\\CG2528, lincRNA.224', description='CG2528 gene product from transcript CG2528-RG', other_designations='CG2528-PE|CG2528-PG', map_location='40D1-40D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21974749, end_position_on_the_genomic_accession=21992933, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35437, CurrentID=0, Status='live', Symbol='ttm3', Aliases='Dmel_CG6691, CG6691, Dmel\\CG6691', description='tiny tim 3', other_designations='CG6691-PA|ttm3-PA', map_location='40E2-40E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22079632, end_position_on_the_genomic_accession=22080738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35438, CurrentID=0, Status='live', Symbol='CG10834', Aliases='Dmel_ Dmel\\CG10834', description='CG10834 gene product from transcript CG10834-RA', other_designations='CG10834-PA|uncharacterized protein', map_location='40E3-40E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22106862, end_position_on_the_genomic_accession=22107304, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35439, CurrentID=0, Status='live', Symbol='CG6675', Aliases='Dmel_ Dmel\\CG6675', description='CG6675 gene product from transcript CG6675-RB', other_designations='CG6675-PB|CG6675-PC', map_location='40E3-40E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22108964, end_position_on_the_genomic_accession=22110449, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35440, CurrentID=0, Status='live', Symbol='CG3651', Aliases='Dmel_ Dmel\\CG3651', description='CG3651 gene product from transcript CG3651-RB', other_designations='CG3651-PB|CG3651-PC|CG3651-PD', map_location='40E4-40E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22121338, end_position_on_the_genomic_accession=22126614, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35442, CurrentID=0, Status='live', Symbol='CG42748', Aliases='Dmel_ CG15216, CG15217, CG6448, Dmel\\CG42748, Dmel_CG15216, Dmel_CG6448, anon-WO0140519.28', description='CG42748 gene product from transcript CG42748-RJ', other_designations='CG42748-PB|CG42748-PC|CG42748-PD|CG42748-PE|CG42748-PF|CG42748-PG|CG42748-PH|CG42748-PJ|CG42748-PK|CG42748-PL|CG42748-PM|uncharacterized protein', map_location='40F1-40F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22131221, end_position_on_the_genomic_accession=22166221, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  35445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35445, CurrentID=0, Status='live', Symbol='Clamp', Aliases='Dmel_CG1832, CG1832, CLAMP, Dmel\\CG1832', description='Chromatin-linked adaptor for MSL proteins', other_designations='CG1832-PA|CG1832-PB|Clamp-PA|Clamp-PB|chromatin-linked adaptor for MSL proteins', map_location='40F2-40F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22165720, end_position_on_the_genomic_accession=22169143, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35447, CurrentID=0, Status='live', Symbol='CG3635', Aliases='Dmel_ Dmel\\CG3635', description='CG3635 gene product from transcript CG3635-RB', other_designations='CG3635-PB|CG3635-PC', map_location='40F2-40F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22175415, end_position_on_the_genomic_accession=22183991, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35449, CurrentID=0, Status='live', Symbol='Ir40a', Aliases='Dmel_CG42352, CG42352, CG5922, CG5929, DmelIR40a, Dmel\\CG42352, Dmel_CG5922, Dmel_CG5929, IR40a', description='Ionotropic receptor 40a', other_designations='CG42352-PE|CG42352-PF|CG42352-PG|Ir40a-PE|Ir40a-PF|Ir40a-PG|ionotropic receptor 40a', map_location='40F4-40F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22189526, end_position_on_the_genomic_accession=22214376, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35452, CurrentID=0, Status='live', Symbol='CG3262', Aliases='Dmel_ Dmel\\CG3262', description='CG3262 gene product from transcript CG3262-RE', other_designations='CG3262-PD|CG3262-PE|CG3262-PF', map_location='40F6-40F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22239916, end_position_on_the_genomic_accession=22246988, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35453, CurrentID=0, Status='live', Symbol='RpL21', Aliases='Dmel_CG12775, BcDNA:RH06526, CG12775, Dmel\\CG12775, RPL21, Rp L21, anon-EST:Posey18, anon-EST:Posey30', description='Ribosomal protein L21', other_designations='CG12775-PA|CG12775-PB|CG12775-PC|RpL21-PA|RpL21-PB|RpL21-PC|ribosomal protein L21', map_location='40F6-40F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22247111, end_position_on_the_genomic_accession=22247875, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35454, CurrentID=0, Status='live', Symbol='Tif-IA', Aliases='Dmel_CG3278, CG16938, CG3278, CG5951, Dmel\\CG3278, TIF-1A, TIF-IA, TIf-IA, Tif-1A, TifIA, dTIF-IA, tif1a', description='CG3278 gene product from transcript CG3278-RA', other_designations='CG3278-PA|Tif-IA-PA', map_location='40F6-40F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22249876, end_position_on_the_genomic_accession=22258260, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38677, CurrentID=0, Status='live', Symbol='PVRAP', Aliases='Dmel_CG32406, CG13292, CG32406, CG6062, Dmel\\CG32406', description='PVR adaptor protein', other_designations='CG32406-PA|PVRAP-PA', map_location='65A2-65A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5997208, end_position_on_the_genomic_accession=6026317, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740634, CurrentID=0, Status='live', Symbol='CG34181', Aliases='Dmel_ Dmel\\CG34181', description='CG34181 gene product from transcript CG34181-RA', other_designations='CG34181-PA|uncharacterized protein', map_location='30B1-30B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9367640, end_position_on_the_genomic_accession=9368434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35473, CurrentID=0, Status='live', Symbol='TpnC41C', Aliases='Dmel_CG2981, CG2981, DmTnC1, DmTpnC41C, DmTpnCIIIa, Dmel\\CG2981, TNC41C, TPNC, TnC, TnC1, TnC41C, TpnC IIIa, TpnC41, TpnCIIIa', description='Troponin C at 41C', other_designations='CG2981-PA|CG2981-PB|TpnC41C-PA|TpnC41C-PB|TroponinC at 41C|troponin C|troponin C at 41C|troponin-C', map_location='41D1-41D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5065237, end_position_on_the_genomic_accession=5070210, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35475, CurrentID=0, Status='live', Symbol='CG3107', Aliases='Dmel_ Dmel\\CG3107', description='CG3107 gene product from transcript CG3107-RB', other_designations='CG3107-PA|CG3107-PB|CG3107-PC|CG3107-PD', map_location='41D3-41D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5090581, end_position_on_the_genomic_accession=5095090, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35478, CurrentID=0, Status='live', Symbol='gus', Aliases='Dmel_CG2944, A11, CG2944, CG2946, Dmel\\CG2944, GUS, Gus, dmSSB4, i25', description='gustavus', other_designations='CG2944-PG|CG2944-PH|CG2944-PI|CG2944-PJ|CG2944-PK|CG2944-PL|CG2944-PM|gus-PG|gus-PH|gus-PI|gus-PJ|gus-PK|gus-PL|gus-PM', map_location='41D3-41D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5116990, end_position_on_the_genomic_accession=5128321, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5740313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740313, CurrentID=0, Status='live', Symbol='CG34423', Aliases='Dmel_ CG11079, Dmel\\CG34423, Dmel_CG11079, anon-WO0118547.139', description='CG34423 gene product from transcript CG34423-RB', other_designations='CG34423-PA|CG34423-PB|uncharacterized protein', map_location='59D8-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23358612, end_position_on_the_genomic_accession=23359204, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35480, CurrentID=0, Status='live', Symbol='Atf6', Aliases='Dmel_CG3136, ATF6, CG3136, DM15, Dmel\\CG3136, atf6', description='CG3136 gene product from transcript CG3136-RA', other_designations='Atf6-PA|Atf6-PB|Atf6-PC|CG3136-PA|CG3136-PB|CG3136-PC', map_location='41D4-41E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5144733, end_position_on_the_genomic_accession=5154756, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35483, CurrentID=0, Status='live', Symbol='Nipped-A', Aliases='Dmel_CG33554, 41Ah, CG10549, CG2905, CG33554, Dmel\\CG33554, GroupV, Nip-A, NipA, Nipped A, Nipped[A], TRA1, TRRAP, Tra-1, Tra1, dTRRAP, dTra1, dmTRRAP, l(2)41Ah, l(2)C123, l(2)EMS34-12, tra1', description='CG33554 gene product from transcript CG33554-RE', other_designations='CG33554-PE|CG33554-PF|Nipped-A-PE|Nipped-A-PF|NippedA|lethal(2)41Ah|nipped', map_location='2-55.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5177965, end_position_on_the_genomic_accession=5251022, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  35485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35485, CurrentID=0, Status='live', Symbol='d4', Aliases='Dmel_CG2682, CG2682, D4, D Dmel\\CG2682, dd4', description='CG2682 gene product from transcript CG2682-RB', other_designations='CG2682-PA|CG2682-PB|CG2682-PC|CG2682-PD|d4-PA|d4-PB|d4-PC|d4-PD', map_location='41E3-41E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5259004, end_position_on_the_genomic_accession=5295965, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35486, CurrentID=0, Status='live', Symbol='sxc', Aliases='Dmel_CG10392, BcDNA:GH04245, CG10392, Dmel\\CG10392, OGT, Ogt, P1201, SXC, Sxc, dO-GnT, l(2)02637, l(2)NC130, ogt/Ogt, sxc', description='super sex combs', other_designations='CG10392-PA|CG10392-PB|CG10392-PC|O-glycosyltransferase|O-linked N-acetylglucosamine transferase|lethal (2) NC130|sxc-PA|sxc-PB|sxc-PC', map_location='2-55.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5307288, end_position_on_the_genomic_accession=5329555, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35488, CurrentID=0, Status='live', Symbol='CG10465', Aliases='Dmel_ Dmel\\CG10465', description='CG10465 gene product from transcript CG10465-RA', other_designations='CG10465-PA|uncharacterized protein', map_location='41E5-41E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5339706, end_position_on_the_genomic_accession=5340995, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35490, CurrentID=0, Status='live', Symbol='CG10395', Aliases='Dmel_ Dmel\\CG10395, mdcds_12730', description='CG10395 gene product from transcript CG10395-RE', other_designations='CG10395-PD|CG10395-PE|uncharacterized protein', map_location='41E5-41E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5350477, end_position_on_the_genomic_accession=5351972, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35491, CurrentID=0, Status='live', Symbol='COX4L', Aliases='Dmel_CG10396, BcDNA:AT07685, CG10396, COX IV, COXIV, CX41, CX42, Dmel\\CG10396, IV, cox4', description='Cytochrome c oxidase subunit 4-like', other_designations='CG10396-PA|CG10396-PB|COX4L-PA|COX4L-PB|cytochrome c oxidase subunit 4-like|cytochrome c oxidase subunit IV', map_location='41E5-41E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5353994, end_position_on_the_genomic_accession=5354723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35492, CurrentID=0, Status='live', Symbol='CG10417', Aliases='Dmel_ Dmel\\CG10417, PP2Cg', description='CG10417 gene product from transcript CG10417-RB', other_designations='CG10417-PA|CG10417-PB|uncharacterized protein', map_location='41E5-41E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5360331, end_position_on_the_genomic_accession=5363104, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35494, CurrentID=0, Status='live', Symbol='laccase2', Aliases='Dmel_CG42345, CG10398, CG10408, CG30437, CG32838, CG42345, Dmel\\CG42345, Dmel_CG30437, Dmel_CG32838, MCO2', description='laccase 2', other_designations='CG42345-PA|CG42345-PD|CG42345-PE|CG42345-PF|CG42345-PG|laccase2-PA|laccase2-PD|laccase2-PE|laccase2-PF|laccase2-PG', map_location='41F1-41F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5413744, end_position_on_the_genomic_accession=5456095, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  35496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35496, CurrentID=0, Status='live', Symbol='CG30440', Aliases='Dmel_ CG10412, CG10416, CG12183, Dmel\\CG30440', description='CG30440 gene product from transcript CG30440-RA', other_designations='CG30440-PA|uncharacterized protein', map_location='41F2-41F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5468319, end_position_on_the_genomic_accession=5495861, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38684, CurrentID=0, Status='live', Symbol='Jon65Aii', Aliases='Dmel_CG6580, CG6580, Dmel\\CG6580, Jon65A, SP153', description='Jonah 65Aii', other_designations='CG6580-PA|Jon65Aii-PA|jonah 65A|jonah 65Aii', map_location='65A3-65A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6052321, end_position_on_the_genomic_accession=6053215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35498, CurrentID=0, Status='live', Symbol='TpnC4', Aliases='Dmel_CG12408, BcDNA:RH03361, CG12408, DmTnC4, DmTpnC41F, DmTpnCIIIb, Dmel\\CG12408, TNC41F6, TnC, TnC4, TpnC IIIb1F, TpnCIIIb, tnC4, TpnC4', description='Troponin C isoform 4', other_designations='CG12408-PA|CG12408-PB|TpnC4-PA|TpnC4-PB|troponin C|troponin C 41F|troponin C isoform 4|troponin C4|troponin-C', map_location='41F2-41F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5503437, end_position_on_the_genomic_accession=5507866, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35499, CurrentID=0, Status='live', Symbol='CG11665', Aliases='Dmel_ Dmel\\CG11665', description='CG11665 gene product from transcript CG11665-RB', other_designations='CG11665-PA|CG11665-PB|uncharacterized protein', map_location='41F6-41F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5640915, end_position_on_the_genomic_accession=5652815, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35500, CurrentID=0, Status='live', Symbol='CG1344', Aliases='Dmel_ BcDNA:LD22679, Dmel\\CG1344', description='CG1344 gene product from transcript CG1344-RB', other_designations='CG1344-PA|CG1344-PB', map_location='41F6-41F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5653676, end_position_on_the_genomic_accession=5656778, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35501, CurrentID=0, Status='live', Symbol='Not3', Aliases='Dmel_CG8426, CG8426, CNOT3, Dmel\\CG8426, NOT3, NOT3.5, NOT3/5, l(2)NC136, not3, not3/5', description='CG8426 gene product from transcript CG8426-RB', other_designations='CG8426-PA|CG8426-PB|Not3-PA|Not3-PB|lethal (2) NC136|lethal(2)NC136', map_location='41F6-41F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5656406, end_position_on_the_genomic_accession=5663222, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35503, CurrentID=0, Status='live', Symbol='CG8245', Aliases='Dmel_ Dmel\\CG8245', description='CG8245 gene product from transcript CG8245-RA', other_designations='CG8245-PA', map_location='41F6-41F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5668383, end_position_on_the_genomic_accession=5669986, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35504, CurrentID=0, Status='live', Symbol='kune', Aliases='Dmel_CG1298, CG1298, Dmel\\CG1298, Kune', description='kune-kune', other_designations='CG1298-PA|kune-PA', map_location='41F6-41F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5671741, end_position_on_the_genomic_accession=5672800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35505, CurrentID=0, Status='live', Symbol='scaf', Aliases='Dmel_CG11066, CG11066, CG15900, Dmel\\CG11066, SPH142, scarf', description='scarface', other_designations='CG11066-PA|CG11066-PB|scaf-PA|scaf-PB', map_location='41F6-41F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5676750, end_position_on_the_genomic_accession=5691723, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35507, CurrentID=0, Status='live', Symbol='CG17337', Aliases='Dmel_ Dm52, Dmel\\CG17337, Y, anon-41Fb', description='CG17337 gene product from transcript CG17337-RA', other_designations='CG17337-PA|CG17337-PC|anon-41Fb|transcription unit Y|uncharacterized protein', map_location='41F9-41F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5697872, end_position_on_the_genomic_accession=5700212, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35509, CurrentID=0, Status='live', Symbol='ap', Aliases='Dmel_CG8376, AP, Ap, Apt, CG8376, Dmel\\CG8376, LIM, S-2a, Xa, blt, neuro, trw', description='apterous', other_designations='APTEROUS|CG8376-PA|CG8376-PB|CG8376-PC|CG8376-PE|ap-PA|ap-PB|ap-PC|ap-PE|apteroys|neuromuscular|no-wings|tornwing|xasta', map_location='41F8-41F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5706202, end_position_on_the_genomic_accession=5727525, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35511, CurrentID=0, Status='live', Symbol='CG7791', Aliases='Dmel_ Dmel\\CG7791', description='CG7791 gene product from transcript CG7791-RA', other_designations='CG7791-PA', map_location='41F9-41F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5758065, end_position_on_the_genomic_accession=5760407, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35512, CurrentID=0, Status='live', Symbol='Gp210', Aliases='Dmel_CG7897, CG14467, CG7897, Dm gp210, Dmel\\CG7897, GP210, Gp188, gp188, gp210, gp210[[D]], lyadi', description='Gp210 ortholog (H. sapiens)', other_designations='CG7897-PA|CG7897-PB|Gp188|Gp210 ortholog|Gp210-PA|Gp210-PB', map_location='41F9-41F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5760862, end_position_on_the_genomic_accession=5774324, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  35513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35513, CurrentID=0, Status='live', Symbol='ZnT41F', Aliases='Dmel_CG11163, BEST:CK02137, CG11163, CK02137, Dmel\\CG11163, cg11163, dZnT41F', description='Zinc transporter 41F', other_designations='CG11163-PA|CG11163-PB|CG11163-PC|CG11163-PD|ZnT41F-PA|ZnT41F-PB|ZnT41F-PC|ZnT41F-PD|zinc transporter 41F', map_location='41F10-41F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5774776, end_position_on_the_genomic_accession=5782858, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35514, CurrentID=0, Status='live', Symbol='Or42a', Aliases='Dmel_CG17250, 41E.2, 42a, CG17250, DOR117, Dmel\\CG17250, OR42a, Or42A', description='Odorant receptor 42a', other_designations='CG17250-PA|Or42a-PA|odorant receptor 42a', map_location='41F11-41F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5791441, end_position_on_the_genomic_accession=5792967, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35515, CurrentID=0, Status='live', Symbol='Tsp42A', Aliases='Dmel_CG14468, CG14468, Dm.Tsp42A, Dmel\\CG14468', description='Tetraspanin 42A', other_designations='CG14468-PA|CG14468-PB|Tsp42A-PA|Tsp42A-PB|tetraspanin 42A', map_location='41F11-41F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5795413, end_position_on_the_genomic_accession=5796710, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35516, CurrentID=0, Status='live', Symbol='Or42b', Aliases='Dmel_CG12754, 41E.1, 42b, CG12754, DOR118, Dmel\\CG12754, OR42b', description='Odorant receptor 42b', other_designations='CG12754-PA|Or42b-PA|odorant receptor 42b', map_location='41F11-41F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5797195, end_position_on_the_genomic_accession=5798512, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35519, CurrentID=0, Status='live', Symbol='CG43366', Aliases='Dmel_ CG12551, CG14470, Dmel\\CG43366, Dmel_CG12551, Dmel_CG14470', description='CG43366 gene product from transcript CG43366-RB', other_designations='CG43366-PB|uncharacterized protein', map_location='42A3-42A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5893895, end_position_on_the_genomic_accession=5931363, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35520, CurrentID=0, Status='live', Symbol='CG7882', Aliases='Dmel_ Dmel\\CG7882', description='CG7882 gene product from transcript CG7882-RC', other_designations='CG7882-PA|CG7882-PC', map_location='42A4-42A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5957621, end_position_on_the_genomic_accession=5960183, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38688, CurrentID=0, Status='live', Symbol='CG10472', Aliases='Dmel_ Dmel\\CG10472, SP78, anon-WO0140519.101', description='CG10472 gene product from transcript CG10472-RA', other_designations='CG10472-PA|uncharacterized protein', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6061372, end_position_on_the_genomic_accession=6062910, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35522, CurrentID=0, Status='live', Symbol='BubR1', Aliases='Dmel_CG7838, 31/13, BUB1, BUBR1, BcDNA:LD23835, Bub1, CG 7838, CG7838, DmBUB1, Dmbub1, Dmel\\CG7838, bub1, bubR1, bubr1, dBUB1, l(2)k03113, l(2)k06109, mcl(2)Z1525', description='Bub1-related kinase', other_designations='Bub1|BubR1-PA|CG7838-PA|bub related one|mcl(2)Z1525', map_location='42A5-42A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5969099, end_position_on_the_genomic_accession=5974385, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35523, CurrentID=0, Status='live', Symbol='mle', Aliases='Dmel_CG11680, CG11680, Dmel\\CG11680, MLE, Mle, mak, mll, mls, nap, nap[ts]', description='maleless', other_designations='CG11680-PA|CG11680-PC|male lethal|male-killer|male-lethal|maleless-no-action potential|mle-PA|mle-PC|no action potential|no-action-potential', map_location='2-55.2 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5974857, end_position_on_the_genomic_accession=5980778, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35524, CurrentID=0, Status='live', Symbol='Src42A', Aliases='Dmel_CG44128, CG14471, CG44128, CG7873, DSrc42A, Dm SRC41, DmHD-29, Dmel\\CG44128, Dmel_CG14471, Dmel_CG7873, Dsrc41, Dsrc42A, Dtk5, HD-29, SK2-4, SRC 42A, SRC42A, Src, Src41, Src42, Src42a, Su(D-raf)1, Su(Raf)1, Su(phl)1, Su1, Tk5, dSrc, dtk-5, dtk5, l(2)k10108, src, src42, src42A, src42a', description='Src oncogene at 42A', other_designations='CG44128-PA|CG44128-PB|Src42A-PA|Src42A-PB|src oncogene at 42A|suppressor of pole hole|suppressor of raf-like-oncogene|tyrosine kinase 5', map_location='42A6-42A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5981844, end_position_on_the_genomic_accession=6012795, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  35526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35526, CurrentID=0, Status='live', Symbol='Act42A', Aliases='Dmel_CG12051, 42A, A, ACT2_DROME, AFFX-Dros-ACTIN_M_r_at, Act, Act42, Act42a, Actin, Actin/BAP47, BAP47, C, CG12051, Dmel\\CG12051, act 42A, act42A, actin', description='Actin 42A', other_designations='42A actin|Act42A-PA|Actin-42A|CG12051-PA|F-actin|actin|actin 42A|actin42a', map_location='42A7-42A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6013910, end_position_on_the_genomic_accession=6015689, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35527, CurrentID=0, Status='live', Symbol='Pngl', Aliases='Dmel_CG7865, CG7865, Dmel\\CG7865, PNGase, png1', description='PNGase-like', other_designations='CG7865-PA|CG7865-PC|PNGase|Pngl-PA|Pngl-PC', map_location='42A7-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6017739, end_position_on_the_genomic_accession=6020695, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35528, CurrentID=0, Status='live', Symbol='Strica', Aliases='Dmel_CG7863, CG7863, DREAM, Dmel\\CG7863, Dream, Dream/Strica, STRICA, STRICA/DREAM/Dream, dream, strica, Strica', description='Ser/Thr-rich caspase', other_designations='CG7863-PA|CG7863-PB|Ser/Thr-RIch-CAspase|Strica-PA|Strica-PB|dream|strica', map_location='42A8-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6020311, end_position_on_the_genomic_accession=6023144, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35529, CurrentID=0, Status='live', Symbol='SCAP', Aliases='Dmel_CG33131, CG14592, CG33131, CG8356, Dmel\\CG33131, Scap, Scap_Dm, dSCAP, dScap, dscap', description='SREBP cleavage activating protein', other_designations='CG33131-PA|SCAP-PA', map_location='42A8-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6026273, end_position_on_the_genomic_accession=6036388, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35531, CurrentID=0, Status='live', Symbol='CG14591', Aliases='Dmel_ Dmel\\CG14591', description='CG14591 gene product from transcript CG14591-RD', other_designations='CG14591-PA|CG14591-PB|CG14591-PC|CG14591-PD|uncharacterized protein', map_location='42A8-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6036557, end_position_on_the_genomic_accession=6040647, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35532, CurrentID=0, Status='live', Symbol='Tbce', Aliases='Dmel_CG7861, CG7861, Dmel\\CG7861, TBCE, tbce', description='Tubulin-specific chaperone E', other_designations='CG7861-PB|CG7861-PC|Tbce-PB|Tbce-PC|tubulin-specific chaperone E', map_location='42A8-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6042532, end_position_on_the_genomic_accession=6044964, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35533, CurrentID=0, Status='live', Symbol='CG7856', Aliases='Dmel_ Dmel\\CG7856', description='CG7856 gene product from transcript CG7856-RA', other_designations='CG7856-PA', map_location='42A8-42A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6057357, end_position_on_the_genomic_accession=6059558, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35535, CurrentID=0, Status='live', Symbol='CCHa2-R', Aliases='Dmel_CG14593, CCHa2r, CG14593, CG14594, Dmel\\CG14593, anon-WO0131005.23', description='CCHamide-2 receptor', other_designations='CCHa2-R-PA|CCHa2-R-PB|CG14593-PA|CG14593-PB', map_location='42A8-42A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6050949, end_position_on_the_genomic_accession=6068680, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35536, CurrentID=0, Status='live', Symbol='CG7845', Aliases='Dmel_ CG745, Dmel\\CG7845, WDR74', description='CG7845 gene product from transcript CG7845-RB', other_designations='CG7845-PA|CG7845-PB', map_location='42A9-42A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6074471, end_position_on_the_genomic_accession=6076367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35537, CurrentID=0, Status='live', Symbol='CG7849', Aliases='Dmel_ Dmel\\CG7849', description='CG7849 gene product from transcript CG7849-RB', other_designations='CG7849-PB', map_location='42A9-42A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6076910, end_position_on_the_genomic_accession=6078088, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35538, CurrentID=0, Status='live', Symbol='CG14590', Aliases='Dmel_ Dmel\\CG14590', description='CG14590 gene product from transcript CG14590-RA', other_designations='CG14590-PA|uncharacterized protein', map_location='42A9-42A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6078641, end_position_on_the_genomic_accession=6080572, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35539, CurrentID=0, Status='live', Symbol='Ars2', Aliases='Dmel_CG7843, CG7843, Dmel\\CG7843, anon-WO0118547.166', description='CG7843 gene product from transcript CG7843-RE', other_designations='Ars2-PA|Ars2-PC|Ars2-PD|Ars2-PE|Ars2-PF|Ars2-PG|CG7843-PA|CG7843-PC|CG7843-PD|CG7843-PE|CG7843-PF|CG7843-PG|arsenic resistance protein 2', map_location='42A9-42A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6080828, end_position_on_the_genomic_accession=6085625, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35540, CurrentID=0, Status='live', Symbol='EcR', Aliases='Dmel_CG1765, CG1765, CG8347, DEcR, Dhr23, DmEcR, Dmel\\CG1765, ECR, ECR-C-A, EcR-B, EcR-B1, EcR1b, EcRA, EcRB, EcRB-1, EcRB1, EcRC, EcdR, Ecr, Ecr1B, NR1H1, anon-WO0229075.1, dECR, dmEcR, ecr, lie, ms(2)06410, ms(2)42A, snt, EcR', description='Ecdysone receptor', other_designations='CG1765-PA|CG1765-PB|CG1765-PC|CG1765-PD|CG1765-PE|CG1765-PG|EcR-PA|EcR-PB|EcR-PC|EcR-PD|EcR-PE|EcR-PG|ecdyson receptor 1b|ecdysone receptor|ecdysone receptor complex|ecdysteroid receptor|ecdysterone receptor|long island expressway|male sterile(2)42A|snaggletooth', map_location='42A9-42A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6087873, end_position_on_the_genomic_accession=6169087, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35541, CurrentID=0, Status='live', Symbol='CG14589', Aliases='Dmel_ Dmel\\CG14589', description='CG14589 gene product from transcript CG14589-RA', other_designations='CG14589-PA|CG14589-PB|CG14589-PC|uncharacterized protein', map_location='42A11-42A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6137303, end_position_on_the_genomic_accession=6138411, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35543, CurrentID=0, Status='live', Symbol='Cyp6w1', Aliases='Dmel_CG8345, 6w1, AntP450, CG8345, CYP6W1, Cyp6W1, Dmel\\CG8345', description='CG8345 gene product from transcript CG8345-RB', other_designations='CG8345-PA|CG8345-PB|Cyp6w1-PA|Cyp6w1-PB|cytochrome P450', map_location='42A12-42A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6173684, end_position_on_the_genomic_accession=6176195, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35544, CurrentID=0, Status='live', Symbol='CG8343', Aliases='Dmel_ BEST:GH10831, Dmel\\CG8343, GH10831, GH10831-Inc', description='CG8343 gene product from transcript CG8343-RA', other_designations='CG8343-PA|Incilarin-like', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6179726, end_position_on_the_genomic_accession=6180356, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35545, CurrentID=0, Status='live', Symbol='CG11211', Aliases='Dmel_ BcDNA:RH26557, Dmel\\CG11211', description='CG11211 gene product from transcript CG11211-RA', other_designations='CG11211-PA|uncharacterized protein', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6180926, end_position_on_the_genomic_accession=6181626, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35546, CurrentID=0, Status='live', Symbol='Ptr', Aliases='Dmel_CG11212, CG11212, Dmel\\CG11212, Dmptr_Dm, ptc-related-gene, ptr, Ptr', description='Patched-related', other_designations='CG11212-PA|CG11212-PB|CG11212-PC|CG11212-PD|CG11212-PE|Ptc related|Ptc-related|Ptc-related Disp-like|Ptr-PA|Ptr-PB|Ptr-PC|Ptr-PD|Ptr-PE|patched-related|ptc-related gene', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6181748, end_position_on_the_genomic_accession=6196320, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35547, CurrentID=0, Status='live', Symbol='CG8335', Aliases='Dmel_ Dmel\\CG8335, eIF3-S5, eIF3f', description='CG8335 gene product from transcript CG8335-RA', other_designations='CG8335-PA', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6200853, end_position_on_the_genomic_accession=6201956, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35549, CurrentID=0, Status='live', Symbol='CG30431', Aliases='Dmel_ CG8324, CG8341, Dmel\\CG30431', description='CG30431 gene product from transcript CG30431-RA', other_designations='CG30431-PA|uncharacterized protein', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6202725, end_position_on_the_genomic_accession=6205224, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35550, CurrentID=0, Status='live', Symbol='CG17994', Aliases='Dmel_ Dmel\\CG17994', description='CG17994 gene product from transcript CG17994-RA', other_designations='CG17994-PA|uncharacterized protein', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6207219, end_position_on_the_genomic_accession=6209468, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38693, CurrentID=0, Status='live', Symbol='CG6610', Aliases='Dmel_ BcDNA:RE60135, Dmel\\CG6610', description='CG6610 gene product from transcript CG6610-RA', other_designations='CG6610-PA|CG6610-PB|CG6610-PC', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6075536, end_position_on_the_genomic_accession=6076258, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35552, CurrentID=0, Status='live', Symbol='bin3', Aliases='Dmel_CG8276, Bin3, CG8276, Dmel\\CG8276, l(2)SH1014, l(2)SH2 1014', description='bicoid-interacting protein 3', other_designations='Bicoid-interacting protein 3|CG8276-PA|CG8276-PB|CG8276-PC|CG8276-PD|bicoid interacting protein 3|bin3-PA|bin3-PB|bin3-PC|bin3-PD|lethal (2) SH1014', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6214568, end_position_on_the_genomic_accession=6240602, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35553, CurrentID=0, Status='live', Symbol='tomboy40', Aliases='Dmel_CG8330, CG8330, Dmel\\CG8330, Tomboy40', description='CG8330 gene product from transcript CG8330-RA', other_designations='CG8330-PA|dtomboy40|tomboy40-PA', map_location='42A14-42A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6225634, end_position_on_the_genomic_accession=6226884, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35554, CurrentID=0, Status='live', Symbol='Pld', Aliases='Dmel_CG12110, CG12110, Dmel\\CG12110, PLD, dPLD, dPld, mitoPLD, pld, pld1', description='Phospholipase D', other_designations='CG12110-PA|CG12110-PB|CG12110-PC|CG12110-PD|CG12110-PF|CG12110-PG|Pld-PA|Pld-PB|Pld-PC|Pld-PD|Pld-PF|Pld-PG|mitochondrial phospholipase|phospholipase D', map_location='42A14-42A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6242004, end_position_on_the_genomic_accession=6255266, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  35555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35555, CurrentID=0, Status='live', Symbol='jing', Aliases='Dmel_CG9397, 1.28, AEBP2, CG9397, CG9403, CT9093, Dmel\\CG9397, Dmel_CG9403, Jing, l(2)01094, l(2)22F3', description='CG9397 gene product from transcript CG9397-RI', other_designations='1.28|CG9397-PD|CG9397-PH|CG9397-PI|CG9397-PJ|deformed|jing-PD|jing-PH|jing-PI|jing-PJ|rhumba', map_location='42B1-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6502006, end_position_on_the_genomic_accession=6621792, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35556, CurrentID=0, Status='live', Symbol='CG15233', Aliases='Dmel_ BEST:LD11861, Dmel\\CG15233', description='CG15233 gene product from transcript CG15233-RB', other_designations='CG15233-PB|uncharacterized protein', map_location='42B2-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6600044, end_position_on_the_genomic_accession=6600978, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35559, CurrentID=0, Status='live', Symbol='Trap1', Aliases='Dmel_CG3152, CG3152, Dmel\\CG3152, TRAP1, TRAP1dr, dtrap-1, trap1', description='CG3152 gene product from transcript CG3152-RA', other_designations='CG3152-PA|CG3152-PB|Hsp90-related protein|TNF receptor-associated protein 1|Trap1-PA|Trap1-PB', map_location='42B2-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6633647, end_position_on_the_genomic_accession=6636262, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35560, CurrentID=0, Status='live', Symbol='Bap170', Aliases='Dmel_CG3274, BAP170, BcDNA:GH12174, CG3274, Dmel\\CG3274, bap170, hfl', description='Brahma associated protein 170kD', other_designations='Bap170-PA|CG3274-PA|brahma associated protein 170kD|half-life', map_location='42B2-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6636512, end_position_on_the_genomic_accession=6642358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35561, CurrentID=0, Status='live', Symbol='Fmo-2', Aliases='Dmel_CG3174, CG3174, DmFMO-2, Dmel\\CG3174, fmo2', description='Flavin-containing monooxygenase 2', other_designations='CG3174-PA|Fmo-2-PA', map_location='42B2-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6642418, end_position_on_the_genomic_accession=6645098, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35563, CurrentID=0, Status='live', Symbol='geminin', Aliases='Dmel_CG3183, CG3183, Dm Geminin, Dm  Dmel\\CG3183, Gem, Geminin, anon-WO0257455.1, gem, l(2)k14019, v(2)k09107', description='CG3183 gene product from transcript CG3183-RA', other_designations='CG3183-PA|CG3183-PB|geminin-PA|geminin-PB', map_location='42C2-42C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6655463, end_position_on_the_genomic_accession=6656594, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35564, CurrentID=0, Status='live', Symbol='sced', Aliases='Dmel_CG3273, CG3273, CT9143, Dmel\\CG3273, Sced', description='scrambled', other_designations='CG3273-PA|CG3273-PB|sced-PA|sced-PB', map_location='42C2-42C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6657326, end_position_on_the_genomic_accession=6660624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35565, CurrentID=0, Status='live', Symbol='Dpit47', Aliases='Dmel_CG3189, CG3189, Dmel\\CG3189, cn20', description='DNA polymerase interacting tpr containing protein of 47kD', other_designations='CG3189-PA|CG3189-PB|Dpit47-PA|Dpit47-PB|cn20', map_location='42C3-42C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6660645, end_position_on_the_genomic_accession=6662118, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35567, CurrentID=0, Status='live', Symbol='CG15908', Aliases='Dmel_ Dmel\\CG15908', description='CG15908 gene product from transcript CG15908-RA', other_designations='CG15908-PA|uncharacterized protein', map_location='42C3-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6668150, end_position_on_the_genomic_accession=6668738, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35568, CurrentID=0, Status='live', Symbol='CG9410', Aliases='Dmel_ Dmel\\CG9410', description='CG9410 gene product from transcript CG9410-RA', other_designations='CG9410-PA|CG9410-PB|CG9410-PE', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6669024, end_position_on_the_genomic_accession=6670076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35569, CurrentID=0, Status='live', Symbol='Hsepi', Aliases='Dmel_CG3194, CG3194, Dmel\\CG3194', description='Heparan sulfate C5-epimerase', other_designations='C5-epimerase|CG3194-PA|CG3194-PB|Hsepi-PA|Hsepi-PB|heparan sulfate C5-epimerase|heparan sulfate glucuronyl C5-epimerase', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6670154, end_position_on_the_genomic_accession=6673372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35570, CurrentID=0, Status='live', Symbol='PGAP3', Aliases='Dmel_CG3271, CG3271, Dmel\\CG3271', description='Post-GPI attachment to proteins 3 ortholog (H. sapiens)', other_designations='CG3271-PA|CG3271-PB|PGAP3-PA|PGAP3-PB|Post-GPI attachment to proteins 3 ortholog', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6673604, end_position_on_the_genomic_accession=6675061, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35571, CurrentID=0, Status='live', Symbol='CG17266', Aliases='Dmel_ Dmel\\CG17266', description='CG17266 gene product from transcript CG17266-RA', other_designations='CG17266-PA|CG17266-PB|uncharacterized protein', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6675028, end_position_on_the_genomic_accession=6676001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35572, CurrentID=0, Status='live', Symbol='CG15909', Aliases='Dmel_ Dmel\\CG15909', description='CG15909 gene product from transcript CG15909-RA', other_designations='CG15909-PA|uncharacterized protein', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6683186, end_position_on_the_genomic_accession=6684710, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35573, CurrentID=0, Status='live', Symbol='Tdc1', Aliases='Dmel_CG30445, CG30445, CG3686, Dmel\\CG30445, dTdc1, tdc, tdc1', description='Tyrosine decarboxylase 1', other_designations='CG30445-PA|Tdc1-PA|tdc1|tyramine decarboxylase 1|tyrosine decarboxylase|tyrosine decarboxylase 1', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6677001, end_position_on_the_genomic_accession=6680391, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35574, CurrentID=0, Status='live', Symbol='CG9422', Aliases='Dmel_ BcDNA:RE35072, Dmel\\CG9422', description='CG9422 gene product from transcript CG9422-RC', other_designations='CG9422-PA|CG9422-PC|CG9422-PD', map_location='42C4-42C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6675822, end_position_on_the_genomic_accession=6693162, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35575, CurrentID=0, Status='live', Symbol='CG3270', Aliases='Dmel_ Dmel\\CG3270', description='CG3270 gene product from transcript CG3270-RA', other_designations='CG3270-PA', map_location='42C5-42C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6693430, end_position_on_the_genomic_accession=6695187, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35576, CurrentID=0, Status='live', Symbol='Mob4', Aliases='Dmel_CG3403, CG3403, D Dmel\\CG3403, Dmob4, MOB4, dmob4', description='CG3403 gene product from transcript CG3403-RA', other_designations='CG3403-PA|CG3403-PB|Mob4-PA|Mob4-PB|Mps one binder kinase activator-like 4', map_location='42C5-42C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6695114, end_position_on_the_genomic_accession=6696661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35577, CurrentID=0, Status='live', Symbol='Rab2', Aliases='Dmel_CG3269, CG3269, DRAB2, D DmRab2, Dmel\\CG3269, Drab2, O18333, rab2', description='CG3269 gene product from transcript CG3269-RA', other_designations='CG3269-PA|CG3269-PB|Rab-protein 2|Rab2-PA|Rab2-PB', map_location='42C5-42C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6696739, end_position_on_the_genomic_accession=6699469, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35578, CurrentID=0, Status='live', Symbol='CG3409', Aliases='Dmel_ Dmel\\CG3409', description='CG3409 gene product from transcript CG3409-RB', other_designations='CG3409-PA|CG3409-PB|CG3409-PC', map_location='42C5-42C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6699695, end_position_on_the_genomic_accession=6715892, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35579, CurrentID=0, Status='live', Symbol='Zip42C.1', Aliases='Dmel_CG9428, CG9428, Dmel\\CG9428, ZIP1, Zip1, cg9428, dZIP1, dZip1, dZip42C.1', description='Zinc/iron regulated transporter-related protein 42C.1', other_designations='CG9428-PA|Zinc/iron regulated transporter-related protein 1|Zinc/iron transporter-related protein 1|Zip42C.1-PA', map_location='42C6-42C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6716342, end_position_on_the_genomic_accession=6717707, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35580, CurrentID=0, Status='live', Symbol='Zip42C.2', Aliases='Dmel_CG9430, CG9430, Dmel\\CG9430, cg9430, dZip2, dZip42C.2', description='Zinc/iron regulated transporter-related protein 42C.2', other_designations='CG9430-PB|Zip42C.2-PB', map_location='42C6-42C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6718196, end_position_on_the_genomic_accession=6719533, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35583, CurrentID=0, Status='live', Symbol='phtf', Aliases='Dmel_CG3268, BcDNA.GH08636, BcDNA:GH08636, CG3268, Dmel\\CG3268, d-', description='CG3268 gene product from transcript CG3268-RA', other_designations='CG3268-PA|phtf-PA|putative homeodomain transcription factor', map_location='42C7-42C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6741899, end_position_on_the_genomic_accession=6746141, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35584, CurrentID=0, Status='live', Symbol='Eb1', Aliases='Dmel_CG3265, BcDNA.LD08743, BcDNA:LD08743, CG3265, Dm EB1, DmEB1, Dmel\\CG3265, EB-1, EB1, S(DmcycE[JP])2.5, anon-EST:Liang-2.33, anon-EST:Liang-2.52, clone 2.33, clone 2.52, dEB-1, dEB1, d eb1, l(2)04524, l(2)4524', description='CG3265 gene product from transcript CG3265-RF', other_designations='+TIP End-Binding 1|CG3265-PA|CG3265-PB|CG3265-PD|CG3265-PE|CG3265-PF|CG3265-PG|Eb1-PA|Eb1-PB|Eb1-PD|Eb1-PE|Eb1-PF|Eb1-PG|MT-plus-end-binding protein|microtubule plus-tip protein EB1', map_location='42C8-42C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6749227, end_position_on_the_genomic_accession=6756300, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  35585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35585, CurrentID=0, Status='live', Symbol='CG3420', Aliases='Dmel_ BcDNA:RE40412, Dmel\\CG3420', description='CG3420 gene product from transcript CG3420-RA', other_designations='CG3420-PA|CG3420-PB', map_location='42C8-42C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6756296, end_position_on_the_genomic_accession=6757209, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35586, CurrentID=0, Status='live', Symbol='CG9436', Aliases='Dmel_ Dmel\\CG9436', description='CG9436 gene product from transcript CG9436-RA', other_designations='CG9436-PA', map_location='42C8-42C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6757601, end_position_on_the_genomic_accession=6758903, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35587, CurrentID=0, Status='live', Symbol='Cyp6a2', Aliases='Dmel_CG9438, 143127_at, 6a2, 91-R, 91R, CG9438, CYP6A2, CYP6a2, Cyp6A2, Cyp6a, Cyt-P450-B1, Dmel\\CG9438, P-450, P-450-B, P450, P450-B, P450-B1, cyp6a2', description='Cytochrome P450-6a2', other_designations='CG9438-PA|Cyp6a2-PA|cytochrome P450|cytochrome P450 6a2|cytochrome P450 B1|cytochrome P450-6a2|cytochrome P450-B1', map_location='42D1-42D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6779749, end_position_on_the_genomic_accession=6781491, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35588, CurrentID=0, Status='live', Symbol='Epac', Aliases='Dmel_CG34392, CG15235, CG3427, CG34392, D-EPAC, D-Epac, Dm_2R:10367, Dmel\\CG34392, Dmel_CG15235, Dmel_CG3427, epac', description='Exchange protein directly activated by cAMP ortholog (H. sapiens)', other_designations='CG34392-PD|CG34392-PE|CG34392-PF|CG34392-PG|Epac-PD|Epac-PE|Epac-PF|Epac-PG|epac guanine nucleotide exchange factor|exchange protein directly activated by cAMP ortholog', map_location='42D1-42D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6759715, end_position_on_the_genomic_accession=6798941, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  35590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35590, CurrentID=0, Status='live', Symbol='SdhB', Aliases='Dmel_CG3283, CG3283, CII-SDHB, DHSB, Dmel\\CG3283, Ip, SDH, SDH-B, SDH-IP, SDH-Ip, SDHb, sdhB', description='Succinate dehydrogenase, subunit B (iron-sulfur)', other_designations='CG3283-PA|COMPLEX II: iron-sulfur subunit|SdhB-PA|iron-sulfur protein subunit of succinate dehydrogenase|succinate dehydrogenase|succinate dehydrogenase B|succinate dehydrogenase Iron-sulfur|succinate dehydrogenase Iron-sulfur protein|succinate dehydrogenase [ubiquinone] iron-sulfur protein|succinate dehydrogenase iron protein|succinate dehydrogenase subunit b|succinate dehydrogenase, subunit B (iron-sulfur)', map_location='42D3-42D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6807626, end_position_on_the_genomic_accession=6809172, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35591, CurrentID=0, Status='live', Symbol='CG15237', Aliases='Dmel_ Dmel\\CG15237', description='CG15237 gene product from transcript CG15237-RA', other_designations='CG15237-PA|uncharacterized protein', map_location='42D4-42D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6809044, end_position_on_the_genomic_accession=6809766, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35592, CurrentID=0, Status='live', Symbol='ubl', Aliases='Dmel_CG3450, CG3450, Dmel\\CG3450, dubl, l(2)k03203, mut182', description='ubiquitin like', other_designations='CG3450-PA|drosophila ubiquitin like|lethal (2) k03203|ubl-PA', map_location='42D4-42D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6810011, end_position_on_the_genomic_accession=6810530, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35594, CurrentID=0, Status='live', Symbol='koi', Aliases='Dmel_CG44154, BEST:LD11166, CG18584, CG3287, CG44154, Dmel\\CG44154, Dmel_CG18584, Dmel_CG3287, n(2)09967', description='klaroid', other_designations='CG44154-PA|CG44154-PB|CG44154-PC|CG44154-PD|CG44154-PE|CG44154-PF|koi-PA|koi-PB|koi-PC|koi-PD|koi-PE|koi-PF', map_location='42D4-42D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6811019, end_position_on_the_genomic_accession=6818199, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  35595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35595, CurrentID=0, Status='live', Symbol='CG15236', Aliases='Dmel_ Dmel\\CG15236', description='CG15236 gene product from transcript CG15236-RA', other_designations='CG15236-PA|CG15236-PB|uncharacterized protein', map_location='42D4-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6828356, end_position_on_the_genomic_accession=6838360, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35596, CurrentID=0, Status='live', Symbol='coro', Aliases='Dmel_CG9446, CG9446, Dmel\\CG9446, l(2)k08011', description='CG9446 gene product from transcript CG9446-RA', other_designations='CG9446-PA|CG9446-PB|CG9446-PC|CG9446-PD|CG9446-PE|CG9446-PF|CG9446-PG|coro-PA|coro-PB|coro-PC|coro-PD|coro-PE|coro-PF|coro-PG|coronin|lethal (2) k08011', map_location='42D4-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6862001, end_position_on_the_genomic_accession=6872718, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35597, CurrentID=0, Status='live', Symbol='CG9447', Aliases='Dmel_ Dmel\\CG9447', description='CG9447 gene product from transcript CG9447-RC', other_designations='CG9447-PC', map_location='42D6-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6872996, end_position_on_the_genomic_accession=6875819, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35599, CurrentID=0, Status='live', Symbol='Spn42Db', Aliases='Dmel_CG9454, CG9454, Dmel\\CG9454', description='Serpin 42Db', other_designations='CG9454-PC|CG9454-PD|Spn42Db-PC|Spn42Db-PD|serpin 42Db', map_location='42D6-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6879285, end_position_on_the_genomic_accession=6880770, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35600, CurrentID=0, Status='live', Symbol='Spn42Dc', Aliases='Dmel_CG9455, CG9455, Dmel\\CG9455', description='Serpin 42Dc', other_designations='CG9455-PA|CG9455-PB|Spn42Dc-PA|Spn42Dc-PB|serpin 42Dc', map_location='42D6-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6880995, end_position_on_the_genomic_accession=6883407, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35602, CurrentID=0, Status='live', Symbol='Spn42De', Aliases='Dmel_CG9460, CG9460, Dmel\\CG9460', description='Serpin 42De', other_designations='CG9460-PA|CG9460-PB|Spn42De-PA|Spn42De-PB|serpin 42De', map_location='42D6-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6885552, end_position_on_the_genomic_accession=6888262, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35604, CurrentID=0, Status='live', Symbol='CG3358', Aliases='Dmel_ Dmel\\CG3358', description='CG3358 gene product from transcript CG3358-RF', other_designations='CG3358-PE|CG3358-PF|CG3358-PG', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6923366, end_position_on_the_genomic_accession=6924929, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740334, CurrentID=0, Status='live', Symbol='CG34300', Aliases='Dmel_ Dmel\\CG34300', description='CG34300 gene product from transcript CG34300-RA', other_designations='CG34300-PA|CG34300-PB|uncharacterized protein', map_location='99E5-99E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30305808, end_position_on_the_genomic_accession=30306640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35608, CurrentID=0, Status='live', Symbol='Cyp6u1', Aliases='Dmel_CG3567, 6u1, CG3567, Dmel\\CG3567', description='CG3567 gene product from transcript CG3567-RC', other_designations='CG3567-PA|CG3567-PB|CG3567-PC|Cyp6u1-PA|Cyp6u1-PB|Cyp6u1-PC|cytochrome P450', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6971996, end_position_on_the_genomic_accession=6975026, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35609, CurrentID=0, Status='live', Symbol='vimar', Aliases='Dmel_CG3572, CG3572, Dmel\\CG3572, Vimar, anon-EST:Posey213, l(2)k16722', description='visceral mesodermal armadillo-repeats', other_designations='CG3572-PB|CG3572-PC|vimar-PB|vimar-PC|visceral mesoderm armadillo repeats|visceral mesodermal armadillo repeats', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6975285, end_position_on_the_genomic_accession=6985712, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35610, CurrentID=0, Status='live', Symbol='CG17002', Aliases='Dmel_ Dmel\\CG17002', description='CG17002 gene product from transcript CG17002-RB', other_designations='CG17002-PB|CG17002-PC|CG17002-PD|uncharacterized protein', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6986043, end_position_on_the_genomic_accession=6988922, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35611, CurrentID=0, Status='live', Symbol='CG30159', Aliases='Dmel_ CG3364, Dmel\\CG30159', description='CG30159 gene product from transcript CG30159-RC', other_designations='CG30159-PA|CG30159-PB|CG30159-PC|uncharacterized protein', map_location='42E4-42E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6992681, end_position_on_the_genomic_accession=6994306, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35612, CurrentID=0, Status='live', Symbol='Tsp42Ec', Aliases='Dmel_CG12847, CG12847, Dm.Tsp42Ec, Dmel\\CG12847', description='Tetraspanin 42Ec', other_designations='CG12847-PA|Tsp42Ec-PA|tetraspanin 42E|tetraspanin 42Ec', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7006442, end_position_on_the_genomic_accession=7008622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35613, CurrentID=0, Status='live', Symbol='Tsp42Ed', Aliases='Dmel_CG12846, BcDNA:GH07074, CG12846, Dm.Tsp42Ed, Dmel\\CG12846, GH07074', description='Tetraspanin 42Ed', other_designations='CG12846-PA|CG12846-PB|Tsp42Ed-PA|Tsp42Ed-PB|tetraspanin 42E|tetraspanin 42Ed', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7009498, end_position_on_the_genomic_accession=7011739, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35614, CurrentID=0, Status='live', Symbol='Tsp42Ee', Aliases='Dmel_CG10106, BcDNA:LD16071, CG10106, Dm.Tsp42Ee, Dmel\\CG10106, anon-WO0118547.151', description='Tetraspanin 42Ee', other_designations='CG10106-PA|CG10106-PB|Tsp42Ee-PA|Tsp42Ee-PB|tetraspanin|tetraspanin 42E|tetraspanin 42Ee', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7012254, end_position_on_the_genomic_accession=7018362, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35615, CurrentID=0, Status='live', Symbol='Tsp42Ef', Aliases='Dmel_CG12845, CG12845, Dm.Tsp42Ef, Dmel\\CG12845', description='Tetraspanin 42Ef', other_designations='CG12845-PA|Tsp42Ef-PA|tetraspanin 42E|tetraspanin 42Ef|tetraspanin42Ef', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7025010, end_position_on_the_genomic_accession=7028082, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35616, CurrentID=0, Status='live', Symbol='Tsp42Eg', Aliases='Dmel_CG12142, BcDNA:GM06962, CG12142, Dm.Tsp42Eg, Dmel\\CG12142', description='Tetraspanin 42Eg', other_designations='CG12142-PA|Tsp42Eg-PA|tetraspanin 42E|tetraspanin 42Eg', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7028911, end_position_on_the_genomic_accession=7033233, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35617, CurrentID=0, Status='live', Symbol='Tsp42Eh', Aliases='Dmel_CG12844, CG12844, Dm.Tsp42Eh, Dmel\\CG12844', description='Tetraspanin 42Eh', other_designations='CG12844-PB|Tsp42Eh-PB|tetraspanin 42E|tetraspanin 42Eh', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7033562, end_position_on_the_genomic_accession=7034920, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35618, CurrentID=0, Status='live', Symbol='Tsp42Ei', Aliases='Dmel_CG12843, CG12843, Dm.Tsp42Ei, Dmel\\CG12843, Tsp42EI', description='Tetraspanin 42Ei', other_designations='CG12843-PA|CG12843-PB|Tsp42Ei-PA|Tsp42Ei-PB|tetraspanin 42E|tetraspanin 42Ei', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7035125, end_position_on_the_genomic_accession=7036949, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35620, CurrentID=0, Status='live', Symbol='Tsp42Ej', Aliases='Dmel_CG12143, CG12143, Dm.Tsp42Ej, Dmel\\CG12143, Sun, sun, tsp42ej', description='Tetraspanin 42Ej', other_designations='CG12143-PA|Tsp42Ej-PA|sunglass|sunglasses|tetraspanin 42E|tetraspanin 42Ej', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7040136, end_position_on_the_genomic_accession=7042762, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35621, CurrentID=0, Status='live', Symbol='Tsp42Ek', Aliases='Dmel_CG12841, CG12841, Dm.Tsp42Ek, Dmel\\CG12841', description='Tetraspanin 42Ek', other_designations='CG12841-PA|CG12841-PB|Tsp42Ek-PA|Tsp42Ek-PB|tetraspanin 42E|tetraspanin 42Ek', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7042656, end_position_on_the_genomic_accession=7044710, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35622, CurrentID=0, Status='live', Symbol='Tsp42El', Aliases='Dmel_CG12840, CG12840, Dm.Tsp42El, Dmel\\CG12840, Tsp42E1, anon-EST:Posey244', description='Tetraspanin 42El', other_designations='CG12840-PA|Tsp42El-PA|tetraspanin 42E|tetraspanin 42El|tetraspanin at 42El', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7045769, end_position_on_the_genomic_accession=7048538, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35623, CurrentID=0, Status='live', Symbol='lbm', Aliases='Dmel_CG2374, CG2374, Dm.Tsp42Em, Dmel\\CG2374, Ibm, LBM, Lbm, lbl', description='late bloomer', other_designations='CG2374-PA|CG2374-PB|latebloomer|lbm-PA|lbm-PB|tetraspanin 42E', map_location='42E6-42E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7049238, end_position_on_the_genomic_accession=7051964, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35624, CurrentID=0, Status='live', Symbol='Tsp42En', Aliases='Dmel_CG12839, BcDNA:RH02160, CG12839, Dm.Tsp42En, Dmel\\CG12839', description='Tetraspanin 42En', other_designations='CG12839-PA|CG12839-PB|Tsp42En-PA|Tsp42En-PB|tetraspanin 42E|tetraspanin 42En', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7052747, end_position_on_the_genomic_accession=7054399, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35625, CurrentID=0, Status='live', Symbol='Tsp42Eo', Aliases='Dmel_CG12838, BcDNA:RH51444, CG12838, Dm.Tsp42Eo, Dmel\\CG12838', description='Tetraspanin 42Eo', other_designations='CG12838-PA|Tsp42Eo-PA|tetraspanin 42E|tetraspanin 42Eo', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7054717, end_position_on_the_genomic_accession=7056576, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35626, CurrentID=0, Status='live', Symbol='Tsp42Ep', Aliases='Dmel_CG4471, CG4471, Dm.Tsp42Ep, Dmel\\CG4471', description='Tetraspanin 42Ep', other_designations='CG4471-PA|CG4471-PB|Tsp42Ep-PA|Tsp42Ep-PB|tetraspanin 42E|tetraspanin 42Ep', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7054405, end_position_on_the_genomic_accession=7057548, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35627, CurrentID=0, Status='live', Symbol='pgant3', Aliases='Dmel_CG4445, BcDNA:GH09147, CG4445, Dmel\\CG4445, anon-WO0118547.119', description='polypeptide GalNAc transferase 3', other_designations='CG4445-PA|pgant3-PA', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7057654, end_position_on_the_genomic_accession=7061732, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35628, CurrentID=0, Status='live', Symbol='Tsp42Eq', Aliases='Dmel_CG12832, CG12832, Dm.Tsp42Eq, Dmel\\CG12832', description='Tetraspanin 42Eq', other_designations='CG12832-PA|Tsp42Eq-PA|tetraspanin 42E|tetraspanin 42Eq', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7062108, end_position_on_the_genomic_accession=7063569, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35629, CurrentID=0, Status='live', Symbol='Tsp42Er', Aliases='Dmel_CG12837, CG12837, Dm.Tsp42Er, Dmel\\CG12837', description='Tetraspanin 42Er', other_designations='CG12837-PA|Tsp42Er-PA|tetraspanin 42E|tetraspanin 42Er', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7064399, end_position_on_the_genomic_accession=7066113, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35630, CurrentID=0, Status='live', Symbol='CG12836', Aliases='Dmel_ Dmel\\CG12836, mdcds_19139', description='CG12836 gene product from transcript CG12836-RC', other_designations='CG12836-PC|uncharacterized protein', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7068572, end_position_on_the_genomic_accession=7069896, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35631, CurrentID=0, Status='live', Symbol='CG12831', Aliases='Dmel_ Dmel\\CG12831', description='CG12831 gene product from transcript CG12831-RC', other_designations='CG12831-PB|CG12831-PC|uncharacterized protein', map_location='42E7-42E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7070495, end_position_on_the_genomic_accession=7074152, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35634, CurrentID=0, Status='live', Symbol='Cyp9b1', Aliases='Dmel_CG4485, 9b1, CG4485, CYP9B1, Dmel\\CG4485, P-450, P450', description='Cytochrome P450-9b1', other_designations='CG4485-PA|Cyp9b1-PA|cytochrome P450-9b1', map_location='42F2-42F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7127838, end_position_on_the_genomic_accession=7129808, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35635, CurrentID=0, Status='live', Symbol='Cyp9b2', Aliases='Dmel_CG4486, 9b2, BEST:GH08116, CG4486, Dmel\\CG4486, P-450, P-450 Cyp9b2, P450, cyp9b2', description='Cytochrome P450-9b2', other_designations='CG4486-PA|Cyp9b2-PA|cytochrome P450|cytochrome P450-9b2|fly plexin a', map_location='42F2-42F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7130277, end_position_on_the_genomic_accession=7132139, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35639, CurrentID=0, Status='live', Symbol='Spn43Ad', Aliases='Dmel_CG1859, CG1859, Dmel\\CG1859', description='Serpin 43Ad', other_designations='CG1859-PA|Spn43Ad-PA|serpin 43Ad', map_location='42F3-42F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7153738, end_position_on_the_genomic_accession=7155424, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35641, CurrentID=0, Status='live', Symbol='CG11060', Aliases='Dmel_ Dmel\\CG11060', description='CG11060 gene product from transcript CG11060-RA', other_designations='CG11060-PA|uncharacterized protein', map_location='42F3-42F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7175886, end_position_on_the_genomic_accession=7178290, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35642, CurrentID=0, Status='live', Symbol='CG33140', Aliases='Dmel_ BcDNA:AT21895, CG11074, Dmel\\CG33140', description='CG33140 gene product from transcript CG33140-RA', other_designations='CG33140-PA|uncharacterized protein', map_location='42F3-43A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7178390, end_position_on_the_genomic_accession=7180730, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35644, CurrentID=0, Status='live', Symbol='Or43a', Aliases='Dmel_CG1854, 43B.1, 43B1, 43a, AN14, CG1854, DOR43B.1, DOR43a, DOR87, Dmel\\CG1854, Dor87, OR43a, Or, Or43A, Or43B.1, Or87, dor87, or43a', description='Odorant receptor 43a', other_designations='CG1854-PA|Or43a-PA|odorant receptor 43a|odorant receptor 87|olfactory receptor 43B.1', map_location='43A1-43A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7232840, end_position_on_the_genomic_accession=7234913, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35646, CurrentID=0, Status='live', Symbol='Gadd45', Aliases='Dmel_CG11086, BcDNA:RH70774, CG11086, D-GADD45, Dmel\\CG11086, GADD45, GG1086', description='CG11086 gene product from transcript CG11086-RA', other_designations='CG11086-PA|Gadd45-PA|growth arrest and DNA damage-inducible gene 45', map_location='43A2-43A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7249161, end_position_on_the_genomic_accession=7250655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35647, CurrentID=0, Status='live', Symbol='CG1850', Aliases='Dmel_ Dmel\\CG1850', description='CG1850 gene product from transcript CG1850-RA', other_designations='CG1850-PA', map_location='43A2-43A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7258867, end_position_on_the_genomic_accession=7261269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35648, CurrentID=0, Status='live', Symbol='Br140', Aliases='Dmel_CG1845, CG1845, Dmel\\CG1845', description='CG1845 gene product from transcript CG1845-RB', other_designations='Br140-PA|Br140-PB|CG1845-PA|CG1845-PB', map_location='43A2-43A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7272275, end_position_on_the_genomic_accession=7278101, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35649, CurrentID=0, Status='live', Symbol='Incenp', Aliases='Dmel_CG12165, CG12165, DmINCENP, Dmel\\CG12165, INCENP, anon-WO0118547.171, dIncenp, incenp, mat(2)ea-C, mat(2)earlyQA26', description='Inner centromere protein', other_designations='CG12165-PA|CG12165-PB|CG12165-PC|Incenp-PA|Incenp-PB|Incenp-PC|inner centromere protein|maternal effect early arrest C', map_location='2-61 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7278467, end_position_on_the_genomic_accession=7281795, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35651, CurrentID=0, Status='live', Symbol='CG12164', Aliases='Dmel_ Dmel\\CG12164', description='CG12164 gene product from transcript CG12164-RB', other_designations='CG12164-PA|CG12164-PB|uncharacterized protein', map_location='43A4-43A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7311395, end_position_on_the_genomic_accession=7313495, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35652, CurrentID=0, Status='live', Symbol='Dscam1', Aliases='Dmel_CG17800, 43Bc, CG17800, CT39257, DSCAM, DScam, DmDscam, Dm_2R:13579, Dm_2R:13612, Dmel\\CG17800, DsCam, Dscam, Dscam-hv, Dsm, FBgn0033159, Neu1, dScam, dscam, l(2)05518, l(2)43Bc, p270', description='Down syndrome cell adhesion molecule 1', other_designations="CG17800-PA|CG17800-PAA|CG17800-PAB|CG17800-PAC|CG17800-PAD|CG17800-PAE|CG17800-PAF|CG17800-PAG|CG17800-PAH|CG17800-PAI|CG17800-PAJ|CG17800-PAK|CG17800-PAL|CG17800-PAM|CG17800-PAN|CG17800-PAO|CG17800-PAP|CG17800-PAQ|CG17800-PAR|CG17800-PAS|CG17800-PAT|CG17800-PAU|CG17800-PAV|CG17800-PAW|CG17800-PAX|CG17800-PAY|CG17800-PAZ|CG17800-PBA|CG17800-PBB|CG17800-PBC|CG17800-PBD|CG17800-PBE|CG17800-PBF|CG17800-PBG|CG17800-PBH|CG17800-PBI|CG17800-PBJ|CG17800-PBK|CG17800-PBL|CG17800-PBM|CG17800-PBN|CG17800-PBO|CG17800-PBP|CG17800-PBQ|CG17800-PBR|CG17800-PBS|CG17800-PBT|CG17800-PBU|CG17800-PBV|CG17800-PBW|CG17800-PBX|CG17800-PBY|CG17800-PBZ|CG17800-PC|CG17800-PCA|CG17800-PCB|CG17800-PCC|CG17800-PCD|CG17800-PD|CG17800-PE|CG17800-PF|CG17800-PG|CG17800-PH|CG17800-PI|CG17800-PJ|CG17800-PL|CG17800-PM|CG17800-PO|CG17800-PR|CG17800-PT|CG17800-PV|CG17800-PW|CG17800-PX|CG17800-PY|CG17800-PZ|Down's syndrome cell adhesion molecule|Dscam1-PA|Dscam1-PAA|Dscam1-PAB|Dscam1-PAC|Dscam1-PAD|Dscam1-PAE|Dscam1-PAF|Dscam1-PAG|Dscam1-PAH|Dscam1-PAI|Dscam1-PAJ|Dscam1-PAK|Dscam1-PAL|Dscam1-PAM|Dscam1-PAN|Dscam1-PAO|Dscam1-PAP|Dscam1-PAQ|Dscam1-PAR|Dscam1-PAS|Dscam1-PAT|Dscam1-PAU|Dscam1-PAV|Dscam1-PAW|Dscam1-PAX|Dscam1-PAY|Dscam1-PAZ|Dscam1-PBA|Dscam1-PBB|Dscam1-PBC|Dscam1-PBD|Dscam1-PBE|Dscam1-PBF|Dscam1-PBG|Dscam1-PBH|Dscam1-PBI|Dscam1-PBJ|Dscam1-PBK|Dscam1-PBL|Dscam1-PBM|Dscam1-PBN|Dscam1-PBO|Dscam1-PBP|Dscam1-PBQ|Dscam1-PBR|Dscam1-PBS|Dscam1-PBT|Dscam1-PBU|Dscam1-PBV|Dscam1-PBW|Dscam1-PBX|Dscam1-PBY|Dscam1-PBZ|Dscam1-PC|Dscam1-PCA|Dscam1-PCB|Dscam1-PCC|Dscam1-PCD|Dscam1-PD|Dscam1-PE|Dscam1-PF|Dscam1-PG|Dscam1-PH|Dscam1-PI|Dscam1-PJ|Dscam1-PL|Dscam1-PM|Dscam1-PO|Dscam1-PR|Dscam1-PT|Dscam1-PV|Dscam1-PW|Dscam1-PX|Dscam1-PY|Dscam1-PZ|down syndrome cell adhesion molecule|down syndrome cell adhesion molecule 1|down syndrome cell-adhesion molecule|drosophila down syndrome cell adhesion molecule|drosophila down syndrome cell adhesion molecule 1|lethal(2)43Bc", map_location='43A4-43B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7317924, end_position_on_the_genomic_accession=7381899, orientation='minus', exon_count=115, OMIM='', no_hdr0=''),
  35653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35653, CurrentID=0, Status='live', Symbol='cos', Aliases='Dmel_CG1708, CG1708, COS2, COS[2], Cos, Cos-2, Cos2, Cos[2], DmCos2, Dmel\\CG1708, KIF26A/B, Kif7-2, cos2, l(2)161/28, l(2)cos[[2]], l(2)k16101, l(2)k16128, cos', description='costa', other_designations='CG1708-PA|CG1708-PB|Costal-2|Costal2|DmCostal2|cos-PA|cos-PB|costal 2|enhancer of epaulette', map_location='2-57 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7382176, end_position_on_the_genomic_accession=7387115, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35654, CurrentID=0, Status='live', Symbol='Dhx15', Aliases='Dmel_CG11107, CG11107, DHX15, Dmel\\CG11107, anon-EST:Liang-1.18, cg11107, clone 1.18', description='DHX15 ortholog', other_designations='CG11107-PA|CG11107-PB|Dhx15-PA|Dhx15-PB', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7387187, end_position_on_the_genomic_accession=7391023, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35655, CurrentID=0, Status='live', Symbol='Gr43a', Aliases='Dmel_CG1712, 43C.1, 43C1, CG1712, DmGr43a, Dmel\\CG1712, GR43C.1, GR43a, Gr43C1, gr43a', description='Gustatory receptor 43a', other_designations='CG1712-PA|CG1712-PB|CG1712-PC|Gr43a-PA|Gr43a-PB|Gr43a-PC|gustatory receptor 43C.1|gustatory receptor 43a', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7390869, end_position_on_the_genomic_accession=7393161, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35656, CurrentID=0, Status='live', Symbol='Glo1', Aliases='Dmel_CG1707, CG1707, Dmel\\CG1707, glo1', description='Glyoxalase 1', other_designations='CG1707-PA|Glo1-PA|glyoxalase 1', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7393546, end_position_on_the_genomic_accession=7395932, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35658, CurrentID=0, Status='live', Symbol='CG11112', Aliases='Dmel_ CT8227, Dmel\\CG11112', description='CG11112 gene product from transcript CG11112-RC', other_designations='CG11112-PC|uncharacterized protein', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7400743, end_position_on_the_genomic_accession=7401482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35659, CurrentID=0, Status='live', Symbol='CG11113', Aliases='Dmel_ Dmel\\CG11113', description='CG11113 gene product from transcript CG11113-RB', other_designations='CG11113-PB|uncharacterized protein', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7401859, end_position_on_the_genomic_accession=7402364, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35660, CurrentID=0, Status='live', Symbol='Eaf', Aliases='Dmel_CG11166, CG11166, CG1706, Dmel\\CG11166, EP(2)2475, LD35207, dEaf', description='ELL-associated factor', other_designations='CG11166-PA|CG11166-PB|CG11166-PC|CG11166-PD|Eaf-PA|Eaf-PB|Eaf-PC|Eaf-PD|dELL-associated factor', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7396850, end_position_on_the_genomic_accession=7407819, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35661, CurrentID=0, Status='live', Symbol='CG1701', Aliases='Dmel_ Dmel\\CG1701', description='CG1701 gene product from transcript CG1701-RC', other_designations='CG1701-PB|CG1701-PC', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7408371, end_position_on_the_genomic_accession=7409385, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35662, CurrentID=0, Status='live', Symbol='so', Aliases='Dmel_CG11121, CG11121, Dmel\\CG11121, Drl, SO, So, ami, mda, med, oldmda, so', description='sine oculis', other_designations='CG11121-PA|absent minded|droplet|medusa|optic lobe defective|sine oculus|sinus oculis|so-PA', map_location='2-57.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7419350, end_position_on_the_genomic_accession=7434872, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35663, CurrentID=0, Status='live', Symbol='CG11145', Aliases='Dmel_ BcDNA:AT20622, Dmel\\CG11145', description='CG11145 gene product from transcript CG11145-RA', other_designations='CG11145-PA|uncharacterized protein', map_location='43B3-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7427905, end_position_on_the_genomic_accession=7428593, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35664, CurrentID=0, Status='live', Symbol='CG11123', Aliases='Dmel_ Dmel\\CG11123', description='CG11123 gene product from transcript CG11123-RA', other_designations='CG11123-PA|uncharacterized protein', map_location='43C1-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7440706, end_position_on_the_genomic_accession=7442885, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35665, CurrentID=0, Status='live', Symbol='sPLA2', Aliases='Dmel_CG11124, CG11124, CG14507, Dmel\\CG11124, sPLA[[2]], secretory phospholipase A[[2]]', description='secretory Phospholipase A2', other_designations='CG11124-PA|CG11124-PB|CG11124-PC|sPLA2-PA|sPLA2-PB|sPLA2-PC|secreted phospholipase A2|secretory phospholipase A2', map_location='43C1-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7442864, end_position_on_the_genomic_accession=7443827, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35667, CurrentID=0, Status='live', Symbol='kappaB-Ras', Aliases='Dmel_CG1669, B7YZS7, BcDNA:RH38596, CG1669, Dmel\\CG1669, Q9V4L4-1, d', description='CG1669 gene product from transcript CG1669-RB', other_designations='CG1669-PA|CG1669-PB|CG1669-PC|kappaB-Ras-PA|kappaB-Ras-PB|kappaB-Ras-PC', map_location='43C1-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7443824, end_position_on_the_genomic_accession=7445792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35669, CurrentID=0, Status='live', Symbol='CG11125', Aliases='Dmel_ Dmel\\CG11125', description='CG11125 gene product from transcript CG11125-RC', other_designations='CG11125-PA|CG11125-PC|uncharacterized protein', map_location='43C1-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7447473, end_position_on_the_genomic_accession=7449995, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35670, CurrentID=0, Status='live', Symbol='fa2h', Aliases='Dmel_CG30502, CG1642, CG1645, CG30502, Dmel\\CG30502, FA2H', description='fatty acid 2-hydroylase', other_designations='CG30502-PB|fa2h-PB|fatty acid hydroxlase|sphingolipid-specific fatty acid hydroxylase', map_location='43C1-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7446189, end_position_on_the_genomic_accession=7451553, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35671, CurrentID=0, Status='live', Symbol='Inos', Aliases='Dmel_CG11143, CG11143, Dmel\\CG11143, INOS, bs36h12.y1', description='CG11143 gene product from transcript CG11143-RA', other_designations='CG11143-PA|Inos-PA|inos|myo-inositol-1-phosphate-synthase', map_location='43C3-43C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7454922, end_position_on_the_genomic_accession=7458874, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35672, CurrentID=0, Status='live', Symbol='CG11141', Aliases='Dmel_ Dmel\\CG11141', description='CG11141 gene product from transcript CG11141-RC', other_designations='CG11141-PA|CG11141-PB|CG11141-PC|uncharacterized protein', map_location='43C3-43C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7459685, end_position_on_the_genomic_accession=7463395, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35673, CurrentID=0, Status='live', Symbol='CG11127', Aliases='Dmel_ Dmel\\CG11127', description='CG11127 gene product from transcript CG11127-RA', other_designations='CG11127-PA|uncharacterized protein', map_location='43C3-43C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7463504, end_position_on_the_genomic_accession=7464987, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35674, CurrentID=0, Status='live', Symbol='p47', Aliases='Dmel_CG11139, CG11139, Dmel\\CG11139, anon-WO0118547.216, d', description='CG11139 gene product from transcript CG11139-RA', other_designations='CG11139-PA|p47 a variant|p47-PA', map_location='43C3-43C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7465018, end_position_on_the_genomic_accession=7466759, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35677, CurrentID=0, Status='live', Symbol='Coop', Aliases='Dmel_CG1621, CG1621, Dmel\\CG1621, coop', description='Corepressor of Pangolin', other_designations='CG1621-PA|CG1621-PB|Coop-PA|Coop-PB|co-repressor of pangolin|corepressor of Pan|corepressor of pangolin', map_location='43C5-43C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7490836, end_position_on_the_genomic_accession=7492589, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35678, CurrentID=0, Status='live', Symbol='CG1620', Aliases='Dmel_ 146745_at, Dmel\\CG1620', description='CG1620 gene product from transcript CG1620-RA', other_designations='CG1620-PA|CG1620-PB|CG1620-PC', map_location='43C5-43C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7493806, end_position_on_the_genomic_accession=7496466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35679, CurrentID=0, Status='live', Symbol='dpa', Aliases='Dmel_CG1616, 43Ca, CG1616, DmMCM4, DmMcm4, DmeMCM4, Dmel\\CG1616, Dpa, MCM4, MCM4_DROME, Mcm4, l(2)43Ca, mcm4', description='disc proliferation abnormal', other_designations='CG1616-PA|CG1616-PB|dpa-PA|dpa-PB|drosophila proliferation abnormal', map_location='43C6-43C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7496702, end_position_on_the_genomic_accession=7500145, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35680, CurrentID=0, Status='live', Symbol='didum', Aliases='Dmel_CG2146, 43CD, CG2146, Didum, DmV, Dmel\\CG2146, M5, MYOV, Myo5, MyoV, MyoV43CD, NEST:bs14c05, NMC7, myoV, soy', description='dilute class unconventional myosin', other_designations='CG2146-PA|CG2146-PB|CG2146-PC|MyosinV|didum-PA|didum-PB|didum-PC|myosin V|myosin-V|shorty', map_location='43C7-43D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7500147, end_position_on_the_genomic_accession=7508625, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  35681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35681, CurrentID=0, Status='live', Symbol='CG12736', Aliases='Dmel_ Dmel\\CG12736', description='CG12736 gene product from transcript CG12736-RA', other_designations='CG12736-PA|uncharacterized protein', map_location='43D1-43D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7508527, end_position_on_the_genomic_accession=7510357, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35682, CurrentID=0, Status='live', Symbol='az2', Aliases='Dmel_CG1605, CG1605, Dmel\\CG1605, zf43C', description='CG1605 gene product from transcript CG1605-RA', other_designations='CG1605-PA|az2-PA|zinc finger protein 43C', map_location='43D1-43D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7510622, end_position_on_the_genomic_accession=7512749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35683, CurrentID=0, Status='live', Symbol='CG1603', Aliases='Dmel_ Dmel\\CG1603', description='CG1603 gene product from transcript CG1603-RB', other_designations='CG1603-PA|CG1603-PB', map_location='43D1-43D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7513846, end_position_on_the_genomic_accession=7516390, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35684, CurrentID=0, Status='live', Symbol='CG1602', Aliases='Dmel_ Dmel\\CG1602', description='CG1602 gene product from transcript CG1602-RA', other_designations='CG1602-PA', map_location='43D1-43D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7516923, end_position_on_the_genomic_accession=7518936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35685, CurrentID=0, Status='live', Symbol='CG2144', Aliases='Dmel_ Dmel\\CG2144', description='CG2144 gene product from transcript CG2144-RA', other_designations='CG2144-PA', map_location='43D1-43D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7519280, end_position_on_the_genomic_accession=7522541, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35686, CurrentID=0, Status='live', Symbol='Orc1', Aliases='Dmel_CG10667, CG10667, DmORC1, DmOrc1, DmOrc1p, Dmel\\CG10667, ORC, ORC1, dORC1, orc1, rDmORC', description='Origin recognition complex subunit 1', other_designations='CG10667-PA|Orc1-PA|origin recognition complex 1|origin recognition complex subunit 1', map_location='43D1-43D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7522437, end_position_on_the_genomic_accession=7525716, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35687, CurrentID=0, Status='live', Symbol='Drat', Aliases='Dmel_CG1600, CG1600, Dmel\\CG1600', description='Death resistor Adh domain containing target', other_designations='CG1600-PA|CG1600-PB|CG1600-PC|CG1600-PD|Drat-PA|Drat-PB|Drat-PC|Drat-PD|death resistor Adh domain containing target', map_location='43D2-43D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7526118, end_position_on_the_genomic_accession=7532981, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35688, CurrentID=0, Status='live', Symbol='Cyt-b5', Aliases='Dmel_CG2140, 1857, BcDNA:GH02503, BcDNA:RE66521, CG2140, CYT-B5, CYTB5, Dmel\\CG2140, dpld, l(2)01857, l(2)k00107', description='Cytochrome b5', other_designations='CG2140-PA|CG2140-PB|Cyt-b5-PA|Cyt-b5-PB|cytochrome b5|dappled|lethal (2) k00107', map_location='43C5-43E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7533506, end_position_on_the_genomic_accession=7535936, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35689, CurrentID=0, Status='live', Symbol='CG2137', Aliases='Dmel_ 146747_at, Dmel\\CG2137, GPO, GPO-3, Gpo, alpha-GPO, gpo-3', description='CG2137 gene product from transcript CG2137-RA', other_designations='CG2137-PA', map_location='43D3-43D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7535919, end_position_on_the_genomic_accession=7538297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35690, CurrentID=0, Status='live', Symbol='CG1598', Aliases='Dmel_ BEST:GH10483, BcDNA:GM18141, Dmel\\CG1598, gh10483', description='CG1598 gene product from transcript CG1598-RA', other_designations='CG1598-PA', map_location='43D3-43D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7538232, end_position_on_the_genomic_accession=7539696, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35691, CurrentID=0, Status='live', Symbol='Corin', Aliases='Dmel_CG2105, CG2105, Dmel\\CG2105, SP75', description='CG2105 gene product from transcript CG2105-RC', other_designations='CG2105-PB|CG2105-PC|Corin-PB|Corin-PC|DCorin|corin', map_location='43D3-43D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7540111, end_position_on_the_genomic_accession=7564435, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  35692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35692, CurrentID=0, Status='live', Symbol='CG30499', Aliases='Dmel_ CG1364, Dmel\\CG30499', description='CG30499 gene product from transcript CG30499-RA', other_designations='CG30499-PA|CG30499-PB|uncharacterized protein', map_location='43D7-43D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7564434, end_position_on_the_genomic_accession=7565751, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35693, CurrentID=0, Status='live', Symbol='Vps13', Aliases='Dmel_CG2093, BEST:LD02743, CG2093, Dmel\\CG2093', description='Vacuolar protein sorting 13', other_designations='CG2093-PA|CG2093-PB|Vps13-PA|Vps13-PB|vacuolar protein sorting 13', map_location='43D7-43E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7566809, end_position_on_the_genomic_accession=7580675, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  35694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35694, CurrentID=0, Status='live', Symbol='blow', Aliases='Dmel_CG1363, Blow, CG1363, Dmel\\CG1363, l(2)43Eb', description='blown fuse', other_designations='CG1363-PA|CG1363-PB|CG1363-PD|blow-PA|blow-PB|blow-PD|blown-fuse|blownfuse', map_location='43E1-43E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7580488, end_position_on_the_genomic_accession=7587079, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5740349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740349, CurrentID=0, Status='live', Symbol='CG34357', Aliases='Dmel_ CG12584, CG12585, CG14652, CG41467, CG9783, Dmel\\CG34357, Dmel_CG12585, Dmel_CG14652, Dmel_CG41467, Dmel_CG9783, GYC', description='CG34357 gene product from transcript CG34357-RD', other_designations='CG34357-PB|CG34357-PC|CG34357-PD|uncharacterized protein', map_location='82C1-82C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4491943, end_position_on_the_genomic_accession=4558067, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  35696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35696, CurrentID=0, Status='live', Symbol='scra', Aliases='Dmel_CG2092, ABP8, Abp8, Ani, Anil, Anillin, CG2092, Dmel\\CG2092, Scra, ani, anil, anillin, l(2)03427, l(2)43Ec, l(2)k08255', description='scraps', other_designations='Actin-binding protein 8|CG2092-PA|CG2092-PB|CG2092-PC|anillin|annilin|lethal (2) k08255|scra-PA|scra-PB|scra-PC', map_location='43E4-43E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7590119, end_position_on_the_genomic_accession=7595242, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  35697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35697, CurrentID=0, Status='live', Symbol='CG1358', Aliases='Dmel_ Dmel\\CG1358', description='CG1358 gene product from transcript CG1358-RD', other_designations='CG1358-PA|CG1358-PB|CG1358-PC|CG1358-PD|FLVCR haem transporter', map_location='43E3-43E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7595265, end_position_on_the_genomic_accession=7620711, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  35698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35698, CurrentID=0, Status='live', Symbol='CG43340', Aliases='Dmel_ BcDNA:GH02712, BcDNA:HL07907, CG17983, CG17984, CG2080, CG2088, CG30492, CG30494, CT6764, Dmel\\CG43340, Dmel_CG30492, Dmel_CG30494', description='CG43340 gene product from transcript CG43340-RH', other_designations='CG43340-PA|CG43340-PC|CG43340-PD|CG43340-PE|CG43340-PF|CG43340-PG|CG43340-PH|CG43340-PJ|CG43340-PK|CG43340-PL|CG43340-PM|CG43340-PN|CG43340-PO|CG43340-PP|CG43340-PQ|uncharacterized protein', map_location='43E6-43E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7623242, end_position_on_the_genomic_accession=7651787, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  35701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35701, CurrentID=0, Status='live', Symbol='Rpt1', Aliases='Dmel_CG1341, 1341, CG1341, Dm_Rpt1, Dmel\\CG1341, Dmp48B, Q7KMQ0_DROME, S7, anon-WO0118547.90, dRpt1, l(2)05643, l(2)43Ed, p48B', description='Regulatory particle triple-A ATPase 1', other_designations='19S complex subunit Rpt1|AAA ATPase|CG1341-PA|Rpt1-PA|lethal(2)43Ed|regulatory particle triple-A ATPase 1', map_location='43E7-43E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7652440, end_position_on_the_genomic_accession=7653954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35702, CurrentID=0, Status='live', Symbol='CG17985', Aliases='Dmel_ Dmel\\CG17985', description='CG17985 gene product from transcript CG17985-RA', other_designations='CG17985-PA|uncharacterized protein', map_location='43E7-43E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7651625, end_position_on_the_genomic_accession=7655302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35704, CurrentID=0, Status='live', Symbol='CG30491', Aliases='Dmel_ CG17986, CG2074, Dmel\\CG30491', description='CG30491 gene product from transcript CG30491-RA', other_designations='CG30491-PA|CG30491-PB|uncharacterized protein', map_location='43E7-43E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7655072, end_position_on_the_genomic_accession=7656770, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35705, CurrentID=0, Status='live', Symbol='CG1339', Aliases='Dmel_ Dmel\\CG1339, Gr43b, mdcds_19638', description='CG1339 gene product from transcript CG1339-RB', other_designations='CG1339-PB|Gr43b', map_location='43E8-43E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7659104, end_position_on_the_genomic_accession=7660588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35706, CurrentID=0, Status='live', Symbol='CG2070', Aliases='Dmel_ Dmel\\CG2070', description='CG2070 gene product from transcript CG2070-RB', other_designations='CG2070-PA|CG2070-PB', map_location='43E8-43E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7661855, end_position_on_the_genomic_accession=7663408, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35707, CurrentID=0, Status='live', Symbol='CG2065', Aliases='Dmel_ Dmel\\CG2065', description='CG2065 gene product from transcript CG2065-RB', other_designations='CG2065-PA|CG2065-PB', map_location='43E8-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7663786, end_position_on_the_genomic_accession=7665269, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35708, CurrentID=0, Status='live', Symbol='CG2064', Aliases='Dmel_ BcDNA:SD07613, Dmel\\CG2064', description='CG2064 gene product from transcript CG2064-RA', other_designations='CG2064-PA', map_location='43E9-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7665828, end_position_on_the_genomic_accession=7667343, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35709, CurrentID=0, Status='live', Symbol='DCTN4-p62', Aliases='Dmel_CG12042, CG12042, Dmel\\CG12042, p62', description='Dynactin 4, p62 subunit', other_designations='CG12042-PA|DCTN4-p62-PA|P62 DYNACTIN|dynactin 4, p62 subunit|p62 subunit of dynactin', map_location='43E9-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7667248, end_position_on_the_genomic_accession=7669029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35710, CurrentID=0, Status='live', Symbol='CG12826', Aliases='Dmel_ Dmel\\CG12826', description='CG12826 gene product from transcript CG12826-RA', other_designations='CG12826-PA|uncharacterized protein', map_location='43E9-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7670327, end_position_on_the_genomic_accession=7671103, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35711, CurrentID=0, Status='live', Symbol='mRpL52', Aliases='Dmel_CG1577, BcDNA:RH01479, CG1577, Dmel\\CG1577, MRP-L52, rpl-52, rpl52', description='mitochondrial ribosomal protein L52', other_designations='CG1577-PA|mRpL52-PA', map_location='43E9-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7673721, end_position_on_the_genomic_accession=7674400, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35712, CurrentID=0, Status='live', Symbol='CG12107', Aliases='Dmel_ Dmel\\CG12107', description='CG12107 gene product from transcript CG12107-RA', other_designations='CG12107-PA|uncharacterized protein', map_location='43E9-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7674450, end_position_on_the_genomic_accession=7675362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35713, CurrentID=0, Status='live', Symbol='U2A', Aliases="Dmel_CG1406, BcDNA:RE56869, CG1406, Dmel\\CG1406', U2A-prime, dU2A', l(2)43Ee, u2a", description='CG1406 gene product from transcript CG1406-RA', other_designations='CG1406-PA|U2A-PA|lethal(2)43Ee', map_location='43E9-43E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7675362, end_position_on_the_genomic_accession=7676362, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35715, CurrentID=0, Status='live', Symbol='LRR', Aliases='Dmel_CG1399, CG1399, CG1576, Dmel\\CG1399, Dmel\\LRR', description='Leucine-rich repeat', other_designations='CG1399-PA|CG1399-PB|CG1399-PC|CG1399-PD|CG1399-PF|CG1399-PG|LRR-PA|LRR-PB|LRR-PC|LRR-PD|LRR-PF|LRR-PG|dcarmil|leucine-rich repeat', map_location='43E9-43E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7676540, end_position_on_the_genomic_accession=7698902, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  35717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35717, CurrentID=0, Status='live', Symbol='tor', Aliases='Dmel_CG1389, CG1389, Dmel\\CG1389, TOR, Tor, Torso, splc', description='torso', other_designations='CG1389-PA|CG1389-PD|spliced|tor-PA|tor-PD', map_location='43E11-43E12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7705183, end_position_on_the_genomic_accession=7709184, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  35718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35718, CurrentID=0, Status='live', Symbol='CG1941', Aliases='Dmel_ Dmel\\CG1941', description='CG1941 gene product from transcript CG1941-RC', other_designations='CG1941-PA|CG1941-PB|CG1941-PC|CG1941-PD', map_location='43E12-43E12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7709873, end_position_on_the_genomic_accession=7711981, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35719, CurrentID=0, Status='live', Symbol='CG1942', Aliases='Dmel_ DGAT2, Dmel\\CG1942', description='CG1942 gene product from transcript CG1942-RA', other_designations='CG1942-PA', map_location='43E12-43E12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7712578, end_position_on_the_genomic_accession=7714182, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35720, CurrentID=0, Status='live', Symbol='CG1946', Aliases='Dmel_ Dmel\\CG1946', description='CG1946 gene product from transcript CG1946-RA', other_designations='CG1946-PA', map_location='43E12-43E12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7714923, end_position_on_the_genomic_accession=7716781, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35724, CurrentID=0, Status='live', Symbol='cn', Aliases='Dmel_CG1555, CG1555, Dmel\\CG1555, Dmel\\cn, KMO', description='cinnabar', other_designations='CG1555-PB|cn-PB', map_location='2-57.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7782797, end_position_on_the_genomic_accession=7785067, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35726, CurrentID=0, Status='live', Symbol='CG12825', Aliases='Dmel_ Dmel\\CG12825, dmCG12825', description='CG12825 gene product from transcript CG12825-RA', other_designations='CG12825-PA|uncharacterized protein', map_location='43E16-43E16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7789206, end_position_on_the_genomic_accession=7789988, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35727, CurrentID=0, Status='live', Symbol='CG12824', Aliases='Dmel_ Dmel\\CG12824', description='CG12824 gene product from transcript CG12824-RA', other_designations='CG12824-PA|CG12824-PB|uncharacterized protein', map_location='43E16-43E16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7790368, end_position_on_the_genomic_accession=7791170, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35728, CurrentID=0, Status='live', Symbol='Gapdh1', Aliases='Dmel_CG12055, BEST:GH12586, CG12055, Dmel\\CG12055, GA3PDH, GADPH, GAP, GAPDH, GAPDH I, GAPDH-1, GAPDH1, GAPDHI, Gapd, Gapdh, Gapdh-1, Gapdh43E, gapdh, gapdh-1, gh12586', description='Glyceraldehyde 3 phosphate dehydrogenase 1', other_designations='CG12055-PA|CG12055-PB|G3-P dehydrogenase|GAPDH1|Gapdh1-PA|Gapdh1-PB|Glyceraldehyde-3-phosphate dehydrogenase1|glyceraldehyde 3 phosphate dehydrogenase|glyceraldehyde 3 phosphate dehydrogenase 1|glyceraldehyde 3 phosphate dehydrogenase1|glyceraldehyde-3-phosphate dehydrogenase', map_location='43E17-43E17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7791901, end_position_on_the_genomic_accession=7793380, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35730, CurrentID=0, Status='live', Symbol='Nop17l', Aliases='Dmel_CG1553, CG1553, Dmel\\CG1553', description='Nop17 like', other_designations='CG1553-PA|CG1553-PB|CG1553-PC|NOP17 like|Nop17l-PA|Nop17l-PB|Nop17l-PC', map_location='43E17-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7800785, end_position_on_the_genomic_accession=7806385, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35731, CurrentID=0, Status='live', Symbol='sax', Aliases='Dmel_CG1891, Bkr43E, Brk43E, CG1891, Dmel\\CG1891, SAX, STK-B, Sa, Sax', description='saxophone', other_designations='CG1891-PA|CG1891-PB|CG1891-PC|sax-PA|sax-PB|sax-PC', map_location='2-57 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7806484, end_position_on_the_genomic_accession=7810067, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35732, CurrentID=0, Status='live', Symbol='CG1550', Aliases='Dmel_ Dmel\\CG1550', description='CG1550 gene product from transcript CG1550-RA', other_designations='CG1550-PA', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7810068, end_position_on_the_genomic_accession=7812028, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35733, CurrentID=0, Status='live', Symbol='CG1882', Aliases='Dmel_ Dmel\\CG1882', description='CG1882 gene product from transcript CG1882-RG', other_designations='CG1882-PA|CG1882-PB|CG1882-PC|CG1882-PD|CG1882-PE|CG1882-PF|CG1882-PG|abhydrolase 1', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7812504, end_position_on_the_genomic_accession=7815486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35735, CurrentID=0, Status='live', Symbol='phr', Aliases='Dmel_CG11205, C2PHR, CG11205, CG11205-PA, Dm CPD, Dmel\\CG11205, Phl, anon-OV1, anonOV1', description='photorepair', other_designations='CG11205-PA|CG11205-PB|CPD photolyase|DNA photolyase|anonymous ovarian cDNA 1|photolyase|phr-PA|phr-PB', map_location='2-56.8 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7825289, end_position_on_the_genomic_accession=7827319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35737, CurrentID=0, Status='live', Symbol='CG12822', Aliases='Dmel_CG12822-PA, Dmel\\CG12822', description='CG12822 gene product from transcript CG12822-RA', other_designations='CG12822-PA|CG12822-PB|uncharacterized protein', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7832171, end_position_on_the_genomic_accession=7833841, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35738, CurrentID=0, Status='live', Symbol='Dgk', Aliases='Dmel_CG34361, CG12820, CG1535, CG18654, CG34361, DGKb, Dmel\\CG34361, Dmel_CG12820, Dmel_CG18654, dDGK, dgk', description='Diacyl glycerol kinase', other_designations='CG34361-PE|CG34361-PF|CG34361-PG|CG34361-PH|CG34361-PI|CG34361-PJ|CG34361-PK|Dgk-PE|Dgk-PF|Dgk-PG|Dgk-PH|Dgk-PI|Dgk-PJ|Dgk-PK|diacyl glycerol kinase|diacylglycerol kinase', map_location='43E18-43F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7834812, end_position_on_the_genomic_accession=7885982, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  35741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35741, CurrentID=0, Status='live', Symbol='CG12159', Aliases='Dmel_ Dmel\\CG12159', description='CG12159 gene product from transcript CG12159-RB', other_designations='CG12159-PA|CG12159-PB|uncharacterized protein', map_location='43F1-43F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7915027, end_position_on_the_genomic_accession=7916478, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35742, CurrentID=0, Status='live', Symbol='Cul1', Aliases='Dmel_CG1877, CG1877, CUL1, Cul-1[Ex], Cullin-1, Cullin1, Dmel\\CG1877, Lin-19, Lin19, Q9V312, cul-1, cul1, dCUL1, dCul-1, dCul1, dmCUL1, l(2)02074, l(2)k01207, lin 19, lin-19/cul-1, lin19, lin19-like, Cul1', description='Cullin 1', other_designations='CG1877-PA|CG1877-PB|CG1877-PC|CG1877-PD|Cul1-PA|Cul1-PB|Cul1-PC|Cul1-PD|cullin|cullin 1|cullin-1|cullin1|dCullin|dCullin1|lin-19-like|lin19-like', map_location='43F1-43F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7916856, end_position_on_the_genomic_accession=7921579, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35743, CurrentID=0, Status='live', Symbol='Or43b', Aliases='Dmel_CG17853, 25A.1/43, 43b, AN7, CG17853, DOR25A.1, DOR81, Dmel\\CG17853, OR43b, Or25A.1', description='Odorant receptor 43b', other_designations='CG17853-PA|Or43b-PA|odorant receptor 43b|olfactory receptor 25A.1', map_location='43F2-43F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7920835, end_position_on_the_genomic_accession=7922604, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35744, CurrentID=0, Status='live', Symbol='Kdm4A', Aliases='Dmel_CG15835, CG15835, Dmel\\CG15835, KDM4A, dJMJD2(1), dKDM4A, dkdm4a', description='Lysine (K)-specific demethylase 4A', other_designations='CG15835-PA|CG15835-PB|Kdm4A-PA|Kdm4A-PB|histone demethylase 4A|lysine (K)-specific demethylase 4A', map_location='43F2-43F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7922775, end_position_on_the_genomic_accession=7924988, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35745, CurrentID=0, Status='live', Symbol='MFS12', Aliases='Dmel_CG8791, CG8791, Dmel\\CG8791, anon-WO0140519.99', description='Major Facilitator Superfamily Transporter 12', other_designations='CG8791-PA|CG8791-PB|CG8791-PC|MFS12-PA|MFS12-PB|MFS12-PC|major facilitator superfamily transporter 12', map_location='43F2-43F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7926653, end_position_on_the_genomic_accession=7929922, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35746, CurrentID=0, Status='live', Symbol='rnh1', Aliases='Dmel_CG8729, 74/9, CG8729, Dmel\\CG8729, RNase H1, RnH1, l(2)43F1-5, l(2)43Fa, l(2)k07409, l(2)k07624, rnhl', description='ribonuclease H1', other_designations='CG8729-PA|CG8729-PB|RNaseH1|rnh1-PA|rnh1-PB', map_location='43F2-43F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7931054, end_position_on_the_genomic_accession=7932380, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35747, CurrentID=0, Status='live', Symbol='drosha', Aliases='Dmel_CG8730, CG8730, DROSHA, Dmel\\CG8730, Drosha', description='CG8730 gene product from transcript CG8730-RA', other_designations='CG8730-PA|drosha-PA', map_location='43F2-43F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7932429, end_position_on_the_genomic_accession=7936840, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35748, CurrentID=0, Status='live', Symbol='CG8728', Aliases='Dmel_ Dmel\\CG8728', description='CG8728 gene product from transcript CG8728-RA', other_designations='CG8728-PA', map_location='43F3-43F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7936866, end_position_on_the_genomic_accession=7939114, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35749, CurrentID=0, Status='live', Symbol='CG14764', Aliases='Dmel_ CG2923, Dmel\\CG14764', description='CG14764 gene product from transcript CG14764-RA', other_designations='CG14764-PA|CG14764-PB|CG14764-PC|uncharacterized protein', map_location='43F4-43F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7942043, end_position_on_the_genomic_accession=7955882, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35751, CurrentID=0, Status='live', Symbol='azot', Aliases='Dmel_CG11165, BcDNA:AT10229, BcDNA:AT23738, CG11165, Dmel\\CG11165', description='ahuizotl', other_designations='CG11165-PA|azot-PA', map_location='43F6-43F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7949839, end_position_on_the_genomic_accession=7950465, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35753, CurrentID=0, Status='live', Symbol='CG2906', Aliases='Dmel_ Dmel\\CG2906', description='CG2906 gene product from transcript CG2906-RE', other_designations='CG2906-PD|CG2906-PE', map_location='43F7-43F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7956150, end_position_on_the_genomic_accession=7958254, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35754, CurrentID=0, Status='live', Symbol='CG2915', Aliases='Dmel_ Dmel\\CG2915', description='CG2915 gene product from transcript CG2915-RA', other_designations='CG2915-PA|CG2915-PB', map_location='43F7-43F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7958207, end_position_on_the_genomic_accession=7960015, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38727, CurrentID=0, Status='live', Symbol='Best2', Aliases='Dmel_CG10173, CG10173, Dmel\\CG10173, dBest2, dbest2', description='Bestrophin 2', other_designations='Best2-PA|CG10173-PA|bestrophin 2', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6245933, end_position_on_the_genomic_accession=6256665, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  35756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35756, CurrentID=0, Status='live', Symbol='nito', Aliases='Dmel_CG2910, BcDNA:GH11110, CG2910, Dm44A/spenito, DmSSp, Dmel\\CG2910, Nito, SSLP, cg2910, dOTT', description='spenito', other_designations='CG2910-PA|CG2910-PB|CG2910-PC|One twenty Two|Spen-like protein|nito-PA|nito-PB|nito-PC', map_location='43F8-43F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7962463, end_position_on_the_genomic_accession=7969805, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35757, CurrentID=0, Status='live', Symbol='CG14763', Aliases='Dmel_ Dmel\\CG14763', description='CG14763 gene product from transcript CG14763-RC', other_designations='CG14763-PA|CG14763-PB|CG14763-PC|uncharacterized protein', map_location='43F9-43F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7966029, end_position_on_the_genomic_accession=7968024, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35758, CurrentID=0, Status='live', Symbol='CG8726', Aliases='Dmel_ Dmel\\CG8726', description='CG8726 gene product from transcript CG8726-RB', other_designations='CG8726-PA|CG8726-PB|CG8726-PC', map_location='43F9-43F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7968129, end_position_on_the_genomic_accession=7972003, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35759, CurrentID=0, Status='live', Symbol='CSN4', Aliases='Dmel_CG8725, CG8725, CH4, Cops4, Csn4, DCH4, Dch4, Dmel\\CG8725, csn4, l(2)k08018', description='COP9 signalosome subunit 4', other_designations='CG8725-PA|CG8725-PB|COP9 complex homolog subunit 4|CSN4-PA|CSN4-PB|drosophila COP9 signalosome homolog 4', map_location='43F9-44A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7972175, end_position_on_the_genomic_accession=7974099, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35761, CurrentID=0, Status='live', Symbol='ACC', Aliases='Dmel_CG11198, A1Z784_DROME, ACoT, CG11198, CG8723, DmACC, Dmel\\CG11198, FBgn0043811, dACC', description='Acetyl-CoA carboxylase', other_designations='ACC-PA|ACC-PB|ACC-PC|ACC-PE|ACC-PF|ACC-PG|CG11198-PA|CG11198-PB|CG11198-PC|CG11198-PE|CG11198-PF|CG11198-PG|aceto-acetyl-CoA-thiolase|acetyl CoA carboxylase|acetyl coenzyme A carboxylase|acetyl-CoA carboxylase|acetyl-coenzyme A carboxylase', map_location='44A1-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7974115, end_position_on_the_genomic_accession=7991862, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  35762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35762, CurrentID=0, Status='live', Symbol='Nup44A', Aliases='Dmel_CG8722, CG8722, Dmel\\CG8722, FBgn 33247/Seh1, Seh1, nup44A, seh1, Nup44A', description='Nucleoporin at 44A', other_designations='CG8722-PA|CG8722-PB|CG8722-PC|Nup44A-PA|Nup44A-PB|Nup44A-PC|nucleoporin 44A|nucleoporin at 44A|nucleoporin-44A', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7988462, end_position_on_the_genomic_accession=7991283, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35763, CurrentID=0, Status='live', Symbol='Dic3', Aliases='Dmel_CG11196, CG11196p, Dmel\\CG11196, Dic3', description='Dicarboxylate carrier 3', other_designations='CG11196-PA|CG11196-PB|Dic3-PA|Dic3-PB|dicarboxylate carrier 3', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7993798, end_position_on_the_genomic_accession=7994968, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35764, CurrentID=0, Status='live', Symbol='Hey', Aliases='Dmel_CG11194, CG11194, Dmel\\CG11194, Hesr-1, bHLHe44, d-hey, dHey, hesr-1, hey', description='Hairy/E(spl)-related with YRPW motif', other_designations='CG11194-PA|Hairy/E(spl)|Hairy/E(spl) related with YRPW motif|Hairy/E(spl)-related|Hey-PA|hairy and enhancer of split related protein|hairy and enhancer of split related-1', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7995438, end_position_on_the_genomic_accession=7998278, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35765, CurrentID=0, Status='live', Symbol='CG11191', Aliases='Dmel_ Dmel\\CG11191', description='CG11191 gene product from transcript CG11191-RB', other_designations='CG11191-PA|CG11191-PB|uncharacterized protein', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8000490, end_position_on_the_genomic_accession=8002120, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35766, CurrentID=0, Status='live', Symbol='Odc1', Aliases='Dmel_CG8721, CG8721, Dmel\\CG8721, Odc-1, dodc1', description='Ornithine decarboxylase 1', other_designations='CG8721-PA|Odc1-PA|ornithine decarboxylase|ornithine decarboxylase 1|ornithine decarboxylase-1', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8004824, end_position_on_the_genomic_accession=8006672, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35767, CurrentID=0, Status='live', Symbol='Odc2', Aliases='Dmel_CG8719, CG8719, Dmel\\CG8719, dodc2', description='Ornithine decarboxylase 2', other_designations='CG8719-PA|Odc2-PA|ornithine decarboxylase 2', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8007246, end_position_on_the_genomic_accession=8009062, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35768, CurrentID=0, Status='live', Symbol='CG14762', Aliases='Dmel_ CT34565, Dmel\\CG14762', description='CG14762 gene product from transcript CG14762-RA', other_designations='CG14762-PA|CG14762-PB|gp150-like|uncharacterized protein', map_location='44A2-44A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8009119, end_position_on_the_genomic_accession=8013755, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35770, CurrentID=0, Status='live', Symbol='CG12769', Aliases='Dmel_ Dmel\\CG12769', description='CG12769 gene product from transcript CG12769-RB', other_designations='CG12769-PA|CG12769-PB|CG12769-PC|uncharacterized protein', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8052080, end_position_on_the_genomic_accession=8065554, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35771, CurrentID=0, Status='live', Symbol='lig', Aliases='Dmel_CG8715, CG8715, Dmel\\CG8715', description='lingerer', other_designations='CG8715-PA|CG8715-PB|CG8715-PC|CG8715-PD|CG8715-PE|CG8715-PF|CG8715-PG|CG8715-PH|CG8715-PI|CG8715-PJ|CG8715-PK|CG8715-PL|lig-PA|lig-PB|lig-PC|lig-PD|lig-PE|lig-PF|lig-PG|lig-PH|lig-PI|lig-PJ|lig-PK|lig-PL|transcription unit A', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8067977, end_position_on_the_genomic_accession=8078815, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  35773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35773, CurrentID=0, Status='live', Symbol='slv', Aliases='Dmel_CG8717, CG8717, Dmel\\CG8717', description='saliva', other_designations='CG8717-PA|CG8717-PB|CG8717-PC|slv-PA|slv-PB|slv-PC', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8080559, end_position_on_the_genomic_accession=8083357, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35774, CurrentID=0, Status='live', Symbol='sut1', Aliases='Dmel_CG8714, CG8714, Dmel\\CG8714, Sut1', description='sugar transporter 1', other_designations='CG8714-PA|sut1-PA', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8082976, end_position_on_the_genomic_accession=8087151, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35777, CurrentID=0, Status='live', Symbol='CG8713', Aliases='Dmel_ Dmel\\CG8713, anon-WO0109301.39, anon-WO0109301.40, anon-WO0109301.41, anon-WO0109301.42, anon-WO0109301.43, anon-WO0109301.44, anon-WO0109301.45, anon-WO0109301.46, anon-WO0109301.5', description='CG8713 gene product from transcript CG8713-RA', other_designations='CG8713-PA', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8091696, end_position_on_the_genomic_accession=8096948, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35778, CurrentID=0, Status='live', Symbol='CG8712', Aliases='Dmel_ Dmel\\CG8712', description='CG8712 gene product from transcript CG8712-RB', other_designations='CG8712-PA|CG8712-PB', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8096792, end_position_on_the_genomic_accession=8098484, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35779, CurrentID=0, Status='live', Symbol='CG11210', Aliases='Dmel_ CT31310, Dmel\\CG11210', description='CG11210 gene product from transcript CG11210-RA', other_designations='CG11210-PA|uncharacterized protein', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8098684, end_position_on_the_genomic_accession=8102306, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35780, CurrentID=0, Status='live', Symbol='Cul4', Aliases='Dmel_CG8711, CG8711, CUL4, Cul-4, Cullin4, DMcul-4, Dmel\\CG8711, cul-4, cul4, dCul4', description='Cullin 4', other_designations='CG8711-PA|CG8711-PB|Cul4-PA|Cul4-PB|Cullin-4|Cullin4|cul-4|cullin 4', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8102260, end_position_on_the_genomic_accession=8106318, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  35781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35781, CurrentID=0, Status='live', Symbol='udd', Aliases='Dmel_CG18316, CG18316, Dmel\\CG18316, Udd', description='under-developed', other_designations='CG18316-PA|CG18316-PB|Under-developed|udd-PA|udd-PB', map_location='44A4-44B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8106385, end_position_on_the_genomic_accession=8113949, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35783, CurrentID=0, Status='live', Symbol='Asap', Aliases='Dmel_CG303721, CG14758, CG2226, CG30372, Dmel\\CG30372, dAsap, Asap', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain', other_designations='ArfGAP with SH3 domain, ankyrin repeat and PH domain 1|Asap-PB|Asap-PC|Asap-PD|CG30372-PB|CG30372-PC|CG30372-PD', map_location='44B2-44B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8114286, end_position_on_the_genomic_accession=8123001, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  35784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35784, CurrentID=0, Status='live', Symbol='Nup50', Aliases='Dmel_CG2158, CG2158, Dmel\\CG2158, Nup, nup50', description='Nucleoporin 50kD', other_designations='CG2158-PA|Nup50-PA|nuclear pore protein 50|nucleoporin 50|nucleoporin 50kD', map_location='44B2-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8123036, end_position_on_the_genomic_accession=8125058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35785, CurrentID=0, Status='live', Symbol='coil', Aliases='Dmel_CG8710, CG8710, Coilin, Dmel\\CG8710', description='coilin', other_designations='CG8710-PC|CG8710-PD|CG8710-PE|coil-PC|coil-PD|coil-PE|dcoilin', map_location='44B3-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8124981, end_position_on_the_genomic_accession=8127242, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35786, CurrentID=0, Status='live', Symbol='Socs44A', Aliases='Dmel_CG2160, CG2160, Dmel\\CG2160, SOCS, SOCS44A, Socs44E, Socs44a, dmSocs44A, socs42A, socs44A', description='Suppressor of Cytokine Signaling at 44A', other_designations='CG2160-PA|Socs44A-PA|suppressor of cytokine signaling at 44A|suppressor of cytokine signaling at 44E', map_location='44B3-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8127575, end_position_on_the_genomic_accession=8129476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35787, CurrentID=0, Status='live', Symbol='Pbp49', Aliases='Dmel_CG42515, CG11508, CG42515, DMPBP49, DmPBP, DmPBP49, DmSNAP50, DmSNAPc, Dmel\\CG42515, Dmel_CG11508, Dromel_CG11508_FBtr0088817_mORF', description='PSEA-binding protein 49kD', other_designations='CG42515-PA|CG42515-PB|CG42515-PD|PSEA-binding protein|PSEA-binding protein of 49 kDa|Pbp49-PA|Pbp49-PB|Pbp49-PD|drosophila melanogaster PSE-binding protein 49-kDa subunit|proximal sequence Element-Binding protein 49|snRNA activator protein 50', map_location='44B3-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8129354, end_position_on_the_genomic_accession=8131323, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35788, CurrentID=0, Status='live', Symbol='Pabp2', Aliases='Dmel_CG2163, CG2163, Dmel\\CG2163, PABP2, PABPN1, ROX2, Rox2, dpabII, pAbp2, pabp2, rox2', description='CG2163 gene product from transcript CG2163-RA', other_designations='CG2163-PA|CG2163-PB|Pabp2-PA|Pabp2-PB|nuclear poly(A)-binding protein|poly-A binding protein 2|polyA-binding protein II', map_location='44B3-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8131431, end_position_on_the_genomic_accession=8134374, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35789, CurrentID=0, Status='live', Symbol='Obp44a', Aliases='Dmel_CG2297, 44a, CG2297, DmelObp44a, Dmel\\CG2297, anon-EST:ParkEST107, bs11e05.y1, obp44a', description='Odorant-binding protein 44a', other_designations='CG2297-PA|CG2297-PB|Obp44a-PA|Obp44a-PB|odorant-binding protein 44a', map_location='44B3-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8134170, end_position_on_the_genomic_accession=8135315, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35790, CurrentID=0, Status='live', Symbol='Lpin', Aliases='Dmel_CG8709, CG8709, DLpin, DmLpin, Dmel\\CG8709, EP2431, LIPIN, anon-WO0118547.133', description='Lipin', other_designations='CG8709-PA|CG8709-PB|CG8709-PC|CG8709-PD|CG8709-PE|CG8709-PG|CG8709-PH|CG8709-PJ|CG8709-PK|CG8709-PL|CG8709-PM|DmLipin|EP2431|Lpin-PA|Lpin-PB|Lpin-PC|Lpin-PD|Lpin-PE|Lpin-PG|Lpin-PH|Lpin-PJ|Lpin-PK|Lpin-PL|Lpin-PM|dLipin|lipin', map_location='44B4-44B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8136986, end_position_on_the_genomic_accession=8156654, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  5740365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740365, CurrentID=0, Status='live', Symbol='CG34267', Aliases='Dmel_ Dmel\\CG34267', description='CG34267 gene product from transcript CG34267-RA', other_designations='CG34267-PA|uncharacterized protein', map_location='61C3-61C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=500014, end_position_on_the_genomic_accession=500451, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35792, CurrentID=0, Status='live', Symbol='CG8708', Aliases='Dmel_ Dmel\\CG8708, anon-WO0140519.114, anon-WO0144478.10, dC1GalT2', description='CG8708 gene product from transcript CG8708-RC', other_designations='CG8708-PB|CG8708-PC|CG8708-PD', map_location='44B5-44B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8159406, end_position_on_the_genomic_accession=8161565, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35793, CurrentID=0, Status='live', Symbol='RagC-D', Aliases='Dmel_CG8707, CG8707, Dmel\\CG8707, Gtr2, RagC, dRagC', description='Ras-related GTP binding C/D ortholog (H. sapiens)', other_designations='CG8707-PA|RagC-D-PA|Ras-related GTP binding C/D ortholog', map_location='44B7-44B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8161679, end_position_on_the_genomic_accession=8163500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35795, CurrentID=0, Status='live', Symbol='Gasz', Aliases='Dmel_CG2183, BEST:LD05795, CG2183, Dmel\\CG2183, GASZ, gasz', description='GASZ ortholog', other_designations='CG2183-PA|Gasz-PA', map_location='44B7-44B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8167948, end_position_on_the_genomic_accession=8170089, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35796, CurrentID=0, Status='live', Symbol='Mlh1', Aliases='Dmel_CG11482, CG11482, Dmel\\CG11482, dmlh-1, dmlh1', description='CG11482 gene product from transcript CG11482-RB', other_designations='CG11482-PA|CG11482-PB|Mlh1-PA|Mlh1-PB', map_location='44B8-44B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8170274, end_position_on_the_genomic_accession=8172401, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35797, CurrentID=0, Status='live', Symbol='CG14757', Aliases='Dmel_ 146793_at, Dmel\\CG14757', description='CG14757 gene product from transcript CG14757-RA', other_designations='CG14757-PA|uncharacterized protein', map_location='44B8-44B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8172364, end_position_on_the_genomic_accession=8173613, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35798, CurrentID=0, Status='live', Symbol='CG14756', Aliases='Dmel_ Dmel\\CG14756', description='CG14756 gene product from transcript CG14756-RA', other_designations='CG14756-PA|uncharacterized protein', map_location='44C1-44C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8204888, end_position_on_the_genomic_accession=8205706, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35799, CurrentID=0, Status='live', Symbol='LRP1', Aliases='Dmel_CG33087, CG30368, CG33087, CG8706, CT9297, Dm CG33087, Dmel\\CG33087, LDLR, lrp1', description='LDL receptor protein 1', other_designations='CG33087-PF|CG33087-PG|LDL receptor protein|LDL-receptor-related protein|LDL-receptor-related protein 1|LDLR-like|LRP1-PF|LRP1-PG', map_location='44C1-44C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8174165, end_position_on_the_genomic_accession=8226438, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  35800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35800, CurrentID=0, Status='live', Symbol='dpn', Aliases='Dmel_CG8704, 44C, CG8704, DPN, Dmel\\CG8704, Dpn, anon-EST:fe1B12, bHLHe50', description='deadpan', other_designations='CG8704-PA|anon-fast-evolving-1B12|dpn-PA', map_location='44C2-44C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8228971, end_position_on_the_genomic_accession=8232330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35801, CurrentID=0, Status='live', Symbol='pnut', Aliases='Dmel_CG8705, CG8705, Dmel\\CG8705, E(sina)44C, E(sina)8, PNUT, Pnut, anon-EST:Liang-2.18, clone 2.18, l(2)02502', description='peanut', other_designations='CG8705-PA|CG8705-PB|enhancer of seven in absentia 8|pnut-PA|pnut-PB|septin|septin 3|septin[Pnut]', map_location='44C2-44C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8237271, end_position_on_the_genomic_accession=8240297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35802, CurrentID=0, Status='live', Symbol='CG14760', Aliases='Dmel_ Dmel\\CG14760, SP120', description='CG14760 gene product from transcript CG14760-RA', other_designations='CG14760-PA|uncharacterized protein', map_location='44C2-44C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8254223, end_position_on_the_genomic_accession=8256931, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35803, CurrentID=0, Status='live', Symbol='CG14759', Aliases='Dmel_ Dmel\\CG14759', description='CG14759 gene product from transcript CG14759-RA', other_designations='CG14759-PA|CG14759-PB|uncharacterized protein', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8259841, end_position_on_the_genomic_accession=8260483, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35804, CurrentID=0, Status='live', Symbol='CG2291', Aliases='Dmel_ BcDNA:AT16234, Dmel\\CG2291', description='CG2291 gene product from transcript CG2291-RB', other_designations='CG2291-PA|CG2291-PB', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8260607, end_position_on_the_genomic_accession=8261463, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35805, CurrentID=0, Status='live', Symbol='CG12126', Aliases='Dmel_ BcDNA:AT16918, Dmel\\CG12126', description='CG12126 gene product from transcript CG12126-RA', other_designations='CG12126-PA|uncharacterized protein', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8262674, end_position_on_the_genomic_accession=8263278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35806, CurrentID=0, Status='live', Symbol='CG30371', Aliases='Dmel_ CG2280, Dmel\\CG30371, SP80', description='CG30371 gene product from transcript CG30371-RA', other_designations='CG30371-PA|uncharacterized protein', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8267643, end_position_on_the_genomic_accession=8269334, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35808, CurrentID=0, Status='live', Symbol='CG11635', Aliases='Dmel_ Dmel\\CG11635', description='CG11635 gene product from transcript CG11635-RA', other_designations='CG11635-PA|uncharacterized protein', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8358538, end_position_on_the_genomic_accession=8359472, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35809, CurrentID=0, Status='live', Symbol='boly', Aliases='Dmel_CG30362, CG2149, CG30362, Dmel\\CG30362, NEST:bs34d07', description='borrelly', other_designations='CG30362-PA|boly-PA', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8365550, end_position_on_the_genomic_accession=8366688, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35810, CurrentID=0, Status='live', Symbol='CG18449', Aliases='Dmel_ Dmel\\CG18449', description='CG18449 gene product from transcript CG18449-RA', other_designations='CG18449-PA|uncharacterized protein', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8367163, end_position_on_the_genomic_accession=8367927, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35811, CurrentID=0, Status='live', Symbol='CG2127', Aliases='Dmel_ Dmel\\CG2127, NEST:bs19a07, anon-WO0140519.25, bs19a07.y1', description='CG2127 gene product from transcript CG2127-RB', other_designations='CG2127-PA|CG2127-PB', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8368344, end_position_on_the_genomic_accession=8370038, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35812, CurrentID=0, Status='live', Symbol='CG8701', Aliases='Dmel_ Dmel\\CG8701, anon-WO0140519.78', description='CG8701 gene product from transcript CG8701-RA', other_designations='CG8701-PA', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8371073, end_position_on_the_genomic_accession=8372142, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35813, CurrentID=0, Status='live', Symbol='pdm3', Aliases='Dmel_CG42698, BcDNA:AT16994, CG11641, CG14755, CG42698, Dmel\\CG42698, Dmel_CG11641, Dmel_CG14755, Pdm3', description='pou domain motif 3', other_designations='CG42698-PA|CG42698-PC|CG42698-PE|CG42698-PF|POU domain motif 3|pdm3-PA|pdm3-PC|pdm3-PE|pdm3-PF', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8327490, end_position_on_the_genomic_accession=8396280, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  35814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35814, CurrentID=0, Status='live', Symbol='CG2121', Aliases='Dmel_ CG8698, Dmel\\CG2121', description='CG2121 gene product from transcript CG2121-RB', other_designations='CG2121-PA|CG2121-PB', map_location='44C4-44C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8396397, end_position_on_the_genomic_accession=8411767, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35816, CurrentID=0, Status='live', Symbol='CSN7', Aliases='Dmel_CG2038, CG2038, CH7b, Csn7, Dch7, Dmel\\CG2038, csn7, CSN7', description='COP9 signalosome subunit 7', other_designations='CG2038-PA|COP9 complex homolog subunit 7|CSN7-PA|drosophila COP9 signalosome homolog 7', map_location='44C6-44C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8421033, end_position_on_the_genomic_accession=8428577, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35817, CurrentID=0, Status='live', Symbol='Lcp1', Aliases='Dmel_CG11650, 1, CG11650, CP1, DMLCP1, DmelLcp1, Dmel\\CG11650, I, LCP-1, LCP1, L[[3]]CP1, lcp1', description='Larval cuticle protein 1', other_designations='CG11650-PA|CG11650-PB|Lcp1-PA|Lcp1-PB|cuticle p. I|gene I|larval cuticle protein 1', map_location='44C6-44C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8430115, end_position_on_the_genomic_accession=8431243, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35818, CurrentID=0, Status='live', Symbol='Lcp2', Aliases='Dmel_CG8697, 2, CG8697, CP2, DmelLcp2, Dmel\\CG8697, II, LCP-2, LCP2, L[[3]]CP2, lcp2', description='Larval cuticle protein 2', other_designations='CG8697-PA|CG8697-PB|Lcp2-PA|Lcp2-PB|cuticle p. II|gene II|larval cuticle protein 2', map_location='44C6-44C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8433797, end_position_on_the_genomic_accession=8434785, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35819, CurrentID=0, Status='live', Symbol='Lcp3', Aliases='Dmel_CG2043, 3, CG2043, CP3, DMLCP3, DmelLcp3, Dmel\\CG2043, III, LCP-3, LCP3, L[[3]]CP3, LcpIII, lcp3', description='Larval cuticle protein 3', other_designations='CG2043-PA|CG2043-PB|Lcp3-PA|Lcp3-PB|cuticle p. III|gene III|larval cuticle protein 3', map_location='44C6-44C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8435310, end_position_on_the_genomic_accession=8436095, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35820, CurrentID=0, Status='live', Symbol='Lcp4', Aliases='Dmel_CG2044, 4, CG2044, DmelLcp4, Dmel\\CG2044, IV, LCP-4, LCP4, L[[3]]CP4, lcp4', description='Larval cuticle protein 4', other_designations='CG2044-PA|CG2044-PB|Lcp4-PA|Lcp4-PB|cuticle p. IV|gene IV|larval cuticle protein 4', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8437647, end_position_on_the_genomic_accession=8438495, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35821, CurrentID=0, Status='live', Symbol='Cyp4ad1', Aliases='Dmel_CG2110, 4ad1, BEST:LD02646, CG2110, Dmel\\CG2110', description='CG2110 gene product from transcript CG2110-RA', other_designations='CG2110-PA|CG2110-PB|CYP4-related|Cyp4ad1-PA|Cyp4ad1-PB|cytochrome P450', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8439395, end_position_on_the_genomic_accession=8442806, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35822, CurrentID=0, Status='live', Symbol='Cyp4e2', Aliases='Dmel_CG2060, 4e2, CG2060, CYP4E2, CyP4e2, Cyp4E2, Cyt-P450-4E2, Dmel\\CG2060, P-450, P450', description='Cytochrome P450-4e2', other_designations='CG2060-PA|CG2060-PB|Cyp4e2-PA|Cyp4e2-PB|cytochrome P450|cytochrome P450-4e2', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8444560, end_position_on_the_genomic_accession=8447164, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35824, CurrentID=0, Status='live', Symbol='Mal-A1', Aliases='Dmel_CG8696, CG8696, Dmel\\CG8696, Drome_A1, H, LvpH, V, lvpH, mal_A1', description='Maltase A1', other_designations='CG8696-PA|Mal-A1-PA|gene V|hdl cuticle gene cluster|larval visceral protein H|maltase|maltase A1|maltase H precursor', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8449664, end_position_on_the_genomic_accession=8451904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35825, CurrentID=0, Status='live', Symbol='Mal-A2', Aliases='Dmel_CG8694, CG8694, D, Dmel\\CG8694, Drome_A2, LvpD, lvpD, mal_A2', description='Maltase A2', other_designations='CG8694-PA|Mal-A2-PA|hdl cuticle gene cluster|larval visceral protein D|maltase|maltase A2', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8453627, end_position_on_the_genomic_accession=8455782, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35826, CurrentID=0, Status='live', Symbol='Mal-A3', Aliases='Dmel_CG8695, CG8695, Dmel\\CG8695, Drome_A3, L, LvpL, V, lvpL, mal_A3', description='Maltase A3', other_designations='CG8695-PA|Mal-A3-PA|gene V|hdl cuticle gene cluster|larval visceral protein L|maltase|maltase A3', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8455909, end_position_on_the_genomic_accession=8457892, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35827, CurrentID=0, Status='live', Symbol='Mal-A4', Aliases='Dmel_CG8693, CG8693, Dmel\\CG8693, Drome_A4, mal_A4', description='Maltase A4', other_designations='CG8693-PA|Mal-A4-PA|maltase A4', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8458815, end_position_on_the_genomic_accession=8460884, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35828, CurrentID=0, Status='live', Symbol='Mal-A5', Aliases='Dmel_CG30359, CG30359, CG8689, Dmel\\CG30359, Drome_A5, mal_A5', description='Maltase A5', other_designations='CG30359-PA|CG30359-PB|CG30359-PC|Mal-A5-PA|Mal-A5-PB|Mal-A5-PC|maltase A5', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8462753, end_position_on_the_genomic_accession=8465092, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35829, CurrentID=0, Status='live', Symbol='Mal-A7', Aliases='Dmel_CG11669, CG11669, CT36683, Dmel\\CG11669, Drome_A7, mal_A7', description='Maltase A7', other_designations='CG11669-PA|CG11669-PB|Mal-A7-PA|Mal-A7-PB|maltase A7', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8468646, end_position_on_the_genomic_accession=8471663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35830, CurrentID=0, Status='live', Symbol='Mal-A8', Aliases='Dmel_CG8690, CG8690, CT6532, Dmel\\CG8690, Drome_A8, mal_A8', description='Maltase A8', other_designations='CG8690-PA|Mal-A8-PA|maltase A8', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8471954, end_position_on_the_genomic_accession=8473961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35832, CurrentID=0, Status='live', Symbol='mtt', Aliases='Dmel_CG30361, CG18447, CG30361, CG8692, DmGluBR, DmGluRB, DmXR, Dmel\\CG30361, Glu-RB, GluRB, anon-WO0170980.4, anon-WO0170980.5, mXr', description='mangetout', other_designations='CG30361-PC|CG30361-PD|CG30361-PE|CG30361-PF|CG30361-PG|DmX receptor|mXr|mtt-PC|mtt-PD|mtt-PE|mtt-PF|mtt-PG', map_location='44D1-44D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8473956, end_position_on_the_genomic_accession=8532294, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  35834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35834, CurrentID=0, Status='live', Symbol='CG12780', Aliases='Dmel_ Dmel\\CG12780, GNBP-like', description='CG12780 gene product from transcript CG12780-RA', other_designations='CG12780-PA|uncharacterized protein', map_location='44D2-44D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8551859, end_position_on_the_genomic_accession=8552290, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35835, CurrentID=0, Status='live', Symbol='Cyp6a14', Aliases='Dmel_CG8687, 6a14, CG8687, Dmel\\CG8687, Dmel_CR8687', description='CG8687 gene product from transcript CG8687-RE', other_designations='CG8687-PC|CG8687-PE|Cyp6a14-PC|Cyp6a14-PE|cytochrome P450', map_location='44D3-44D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8564187, end_position_on_the_genomic_accession=8567202, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35837, CurrentID=0, Status='live', Symbol='Cyp6a13', Aliases='Dmel_CG2397, 6a13, CG2397, Dmel\\CG2397, cyp6a13', description='CG2397 gene product from transcript CG2397-RA', other_designations='CG2397-PA|Cyp6a13-PA|cytochrome P450', map_location='44D3-44D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8570499, end_position_on_the_genomic_accession=8572147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35838, CurrentID=0, Status='live', Symbol='CG42326', Aliases='Dmel_ CG14747, CG14748, Dmel\\CG42326, Dmel_CG14747, Dmel_CG14748', description='CG42326 gene product from transcript CG42326-RE', other_designations='CG42326-PD|CG42326-PE|uncharacterized protein', map_location='44D3-44D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8573368, end_position_on_the_genomic_accession=8584779, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  35840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35840, CurrentID=0, Status='live', Symbol='CG14752', Aliases='Dmel_ Dmel\\CG14752', description='CG14752 gene product from transcript CG14752-RA', other_designations='CG14752-PA|uncharacterized protein', map_location='44D4-44D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8590497, end_position_on_the_genomic_accession=8591948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35841, CurrentID=0, Status='live', Symbol='CG8736', Aliases='Dmel_ BcDNA:RH08259, Dmel\\CG8736', description='CG8736 gene product from transcript CG8736-RB', other_designations='CG8736-PB', map_location='44D4-44D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8594031, end_position_on_the_genomic_accession=8595001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35842, CurrentID=0, Status='live', Symbol='CG8735', Aliases='Dmel_ Dmel\\CG8735', description='CG8735 gene product from transcript CG8735-RB', other_designations='CG8735-PA|CG8735-PB|lunapark', map_location='44D4-44D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8596972, end_position_on_the_genomic_accession=8599962, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35843, CurrentID=0, Status='live', Symbol='rgr', Aliases='Dmel_CG8643, CG14751, CG8643, Dmel\\CG8643, l(2)k02605', description='regular', other_designations='CG8643-PA|CG8643-PC|lethal (2) k02605|rgr-PA|rgr-PC', map_location='44D4-31F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8600415, end_position_on_the_genomic_accession=8612295, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35845, CurrentID=0, Status='live', Symbol='CG8642', Aliases='Dmel_ AAF59068, Dmel\\CG8642, FBpp0087808', description='CG8642 gene product from transcript CG8642-RB', other_designations='CG8642-PB|Fibrinogen-related protein', map_location='44D4-44D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8611379, end_position_on_the_genomic_accession=8614155, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35846, CurrentID=0, Status='live', Symbol='Cirl', Aliases='Dmel_CG8639, BcDNA:GH07331, CG8639, CIRL, Dmel\\CG8639, anon-WO0170980.7, anon-WO0170980.8, dcirl', description='Calcium-independent receptor for alpha-latrotoxin', other_designations='CG8639-PB|CG8639-PE|CG8639-PF|CG8639-PG|CG8639-PH|CG8639-PI|CG8639-PJ|CG8639-PK|Cirl-PB|Cirl-PE|Cirl-PF|Cirl-PG|Cirl-PH|Cirl-PI|Cirl-PJ|Cirl-PK', map_location='44D4-44D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8614722, end_position_on_the_genomic_accession=8625203, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  35847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35847, CurrentID=0, Status='live', Symbol='Vps25', Aliases='Dmel_CG14750, CG14750, Dmel\\CG14750, ESCRT-I, ESCRT-II, VPS25, dVps25, l(2)44Db, l(2)k08904, vps25', description='Vacuolar protein sorting 25', other_designations='CG14750-PA|Vps25-PA|lethal (2) 44Db|vacuolar protein sorting 25|vesicle protein sorting 25', map_location='44D5-44D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8625451, end_position_on_the_genomic_accession=8626180, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35848, CurrentID=0, Status='live', Symbol='beta3GalTII', Aliases='Dmel_CG8734, CG8734, Dmel\\CG8734, GalTII, dGAGbeta3GalTII, dbeta3GalTII', description='Galactosyltransferase II', other_designations='CG8734-PA|beta3GalTII-PA|galactosyltransferase II', map_location='44D5-44D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8626181, end_position_on_the_genomic_accession=8627611, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35849, CurrentID=0, Status='live', Symbol='CG14749', Aliases='Dmel_ Dmel\\CG14749, GLE1, Gle1', description='CG14749 gene product from transcript CG14749-RA', other_designations='CG14749-PA|uncharacterized protein', map_location='44D5-44D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8627900, end_position_on_the_genomic_accession=8630186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35850, CurrentID=0, Status='live', Symbol='CG8635', Aliases='Dmel_ Dmel\\CG8635', description='CG8635 gene product from transcript CG8635-RA', other_designations='CG8635-PA|CG8635-PB', map_location='44D5-44D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8630207, end_position_on_the_genomic_accession=8631962, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35851, CurrentID=0, Status='live', Symbol='ptc', Aliases='Dmel_CG2411, BcDNA:RH36596, CG2411, Conf, Dmel\\CG2411, PTC, Ptc, Ptc_Dm, Ptch, com4, dPtc, l(2)k02507, path, rubr, tuf, ptc', description='patched', other_designations='CG2411-PA|compaction defects 4|confused|patch|ptc-PA|rubberneck|tufted', map_location='44D5-44E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8649649, end_position_on_the_genomic_accession=8665223, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35853, CurrentID=0, Status='live', Symbol='Jon44E', Aliases='Dmel_CG8579, CG8579, Dmel\\CG8579, Jon43E, SP100', description='Jonah 44E', other_designations='CG8579-PA|CG8579-PB|Jon44E-PA|Jon44E-PB|jonah 43E|jonah 44E', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8691448, end_position_on_the_genomic_accession=8692844, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35854, CurrentID=0, Status='live', Symbol='CG8586', Aliases='Dmel_ Dmel\\CG8586, SPH58, c-SPH58', description='CG8586 gene product from transcript CG8586-RC', other_designations='CG8586-PA|CG8586-PB|CG8586-PC', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8692796, end_position_on_the_genomic_accession=8696245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35855, CurrentID=0, Status='live', Symbol='CG8738', Aliases='Dmel_ Dmel\\CG8738, SPH121, c-SPH121', description='CG8738 gene product from transcript CG8738-RA', other_designations='CG8738-PA', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8696428, end_position_on_the_genomic_accession=8698583, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35856, CurrentID=0, Status='live', Symbol='CG8584', Aliases='Dmel_ 19527701, BEST:LP01468, Dmel\\CG8584', description='CG8584 gene product from transcript CG8584-RA', other_designations='CG8584-PA', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8700692, end_position_on_the_genomic_accession=8701794, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35857, CurrentID=0, Status='live', Symbol='CG12376', Aliases='Dmel_ Dmel\\CG12376, anon-WO0140519.77', description='CG12376 gene product from transcript CG12376-RA', other_designations='CG12376-PA|CG12376-PB|uncharacterized protein', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8701835, end_position_on_the_genomic_accession=8704502, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35858, CurrentID=0, Status='live', Symbol='CG14744', Aliases='Dmel_ Dmel\\CG14744', description='CG14744 gene product from transcript CG14744-RD', other_designations='CG14744-PC|CG14744-PD|CG14744-PE|uncharacterized protein', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8706388, end_position_on_the_genomic_accession=8709486, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35859, CurrentID=0, Status='live', Symbol='PGRP-SC1a', Aliases='Dmel_CG14746, CG14746, Dmel\\CG14746, PGRP SC, PGRP-SC, PGRP-SC1, PGRP-SC1A, picky', description='CG14746 gene product from transcript CG14746-RA', other_designations='CG14746-PA|PGRP-SC1a-PA|Peptidoglycan-recognition protein-SC1a precursor', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8709733, end_position_on_the_genomic_accession=8710320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35860, CurrentID=0, Status='live', Symbol='CG14743', Aliases='Dmel_ Dmel\\CG14743', description='CG14743 gene product from transcript CG14743-RB', other_designations='CG14743-PB|uncharacterized protein', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8710398, end_position_on_the_genomic_accession=8712913, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35861, CurrentID=0, Status='live', Symbol='PGRP-SC1b', Aliases='Dmel_CG8577, CG8577, CT8705, Dmel\\CG8577, PGRP SC, PGRP-SC, PGRP-SC1, PGRP-SC1B, SC1B', description='CG8577 gene product from transcript CG8577-RB', other_designations='CG8577-PA|CG8577-PB|PGRP-SC1b-PA|PGRP-SC1b-PB|Peptidoglycan-recognition protein-SC1b precursor', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8713446, end_position_on_the_genomic_accession=8715094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35862, CurrentID=0, Status='live', Symbol='PGRP-SC2', Aliases='Dmel_CG14745, CG14745, CG14745 PGRP, Dmel\\CG14745, PCRP-SC2, PGRP-SC', description='CG14745 gene product from transcript CG14745-RA', other_designations='CG14745-PA|PGRP-SC2-PA|Peptidoglycan-recognition protein-SC2 precursor|peptidoglycan recognition protein SC2', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8716950, end_position_on_the_genomic_accession=8717695, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740377, CurrentID=0, Status='live', Symbol='CG34228', Aliases='Dmel_ Dmel\\CG34228, HDC05827', description='CG34228 gene product from transcript CG34228-RA', other_designations='CG34228-PA|uncharacterized protein', map_location='48B4-48B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11651857, end_position_on_the_genomic_accession=11652339, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35864, CurrentID=0, Status='live', Symbol='gcl', Aliases='Dmel_CG8411, CG8411, Dmel\\CG8411, GCL, Gcl-1, gcl', description='germ cell-less', other_designations='CG8411-PA|CG8411-PB|gcl-PA|gcl-PB|germ cell less|germ-cell-less', map_location='44E3-44E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8724258, end_position_on_the_genomic_accession=8728999, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35865, CurrentID=0, Status='live', Symbol='stmA', Aliases='Dmel_CG8739, CG8739, Dmel\\CG8739, cmp44E, rbo, stambhA, stm-A', description='stambha A', other_designations='CG8739-PA|CG8739-PB|CG8739-PC|CG8739-PD|conserved membrane protein at 44E|rolling blackout|stambh A|stambhA|stmA-PA|stmA-PB|stmA-PC|stmA-PD', map_location='2-56.8 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8729502, end_position_on_the_genomic_accession=8734304, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35866, CurrentID=0, Status='live', Symbol='CG8740', Aliases='Dmel_ BcDNA:GH05582, Dmel\\CG8740', description='CG8740 gene product from transcript CG8740-RD', other_designations='CG8740-PA|CG8740-PB|CG8740-PC|CG8740-PD|CG8740-PE', map_location='44E3-44E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8735810, end_position_on_the_genomic_accession=8759590, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  35867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35867, CurrentID=0, Status='live', Symbol='CG8746', Aliases='Dmel_ Dmel\\CG8746', description='CG8746 gene product from transcript CG8746-RA', other_designations='CG8746-PA', map_location='44F1-44F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8792633, end_position_on_the_genomic_accession=8793773, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740378, CurrentID=0, Status='live', Symbol='CG34328', Aliases='Dmel_ Dmel\\CG34328', description='CG34328 gene product from transcript CG34328-RA', other_designations='CG34328-PA|uncharacterized protein', map_location='17A10-17A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18399569, end_position_on_the_genomic_accession=18400065, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35875, CurrentID=0, Status='live', Symbol='CG8272', Aliases='Dmel_ CT8327, Dmel\\CG8272, FBXL9', description='CG8272 gene product from transcript CG8272-RA', other_designations='CG8272-PA|CG8272-PB', map_location='44F2-44F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8888404, end_position_on_the_genomic_accession=8891401, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35877, CurrentID=0, Status='live', Symbol='Sec31', Aliases='Dmel_CG8266, CG8266, Dmel\\CG8266, SEC31, dSec31p, sec31', description='Sec31 ortholog (S. cerevisiae)', other_designations='CG8266-PA|CG8266-PB|CG8266-PC|CG8266-PD|Sec31 ortholog|Sec31-PA|Sec31-PB|Sec31-PC|Sec31-PD|sec31', map_location='44F3-44F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8893613, end_position_on_the_genomic_accession=8898798, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  35878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35878, CurrentID=0, Status='live', Symbol='ana2', Aliases='Dmel_CG8262, Ana-2, Ana2, BcDNA:LD22033, CG8262, Dmel\\CG8262', description='anastral spindle 2', other_designations='CG8262-PA|CG8262-PB|ana2-PA|ana2-PB|anastral spindle phenotype 2', map_location='44F3-44F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8898729, end_position_on_the_genomic_accession=8900211, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35879, CurrentID=0, Status='live', Symbol='CG13751', Aliases='Dmel_ BcDNA:RH09719, Dmel\\CG13751', description='CG13751 gene product from transcript CG13751-RA', other_designations='CG13751-PA|uncharacterized protein', map_location='44F3-44F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8900241, end_position_on_the_genomic_accession=8900617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35880, CurrentID=0, Status='live', Symbol='MrgBP', Aliases='Dmel_CG13746, CG13746, Dmel\\CG13746, MRGBP, dMrgB, dMrgBP', description='MRG/MORF4L binding protein', other_designations='CG13746-PA|MrgBP-PA', map_location='44F3-44F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8900566, end_position_on_the_genomic_accession=8901375, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35881, CurrentID=0, Status='live', Symbol='Ggamma1', Aliases='Dmel_CG8261, CG8261, D-G[[gamma]]1, Dmel\\CG8261, G[[gamma1]], G[[gamma]], Gbetagamma, Gg1, Ggamma, Ggamma-1, bro4, dg1alpha, l(2)k08017', description='G protein gamma 1', other_designations='CG8261-PA|CG8261-PB|CG8261-PC|G protein Gg1|G protein gamma1|G protein subunit gamma1|G-protein gamma subunit 1|Ggamma1-PA|Ggamma1-PB|Ggamma1-PC', map_location='44F3-44F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8901404, end_position_on_the_genomic_accession=8904791, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35882, CurrentID=0, Status='live', Symbol='CG8258', Aliases='Dmel_ CCT8, Dmel\\CG8258, Q7K3J0_DROME, TCP-1theta, Tcp1-theta', description='CG8258 gene product from transcript CG8258-RA', other_designations='CG8258-PA', map_location='44F5-44F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8904844, end_position_on_the_genomic_accession=8906990, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35884, CurrentID=0, Status='live', Symbol='CG8252', Aliases='Dmel_ Dmel\\CG8252', description='CG8252 gene product from transcript CG8252-RA', other_designations='CG8252-PA', map_location='44F6-44F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8909432, end_position_on_the_genomic_accession=8910349, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  35885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35885, CurrentID=0, Status='live', Symbol='CG30349', Aliases='Dmel_ CG13750, CG2396, Dmel\\CG30349', description='CG30349 gene product from transcript CG30349-RA', other_designations='CG30349-PA|uncharacterized protein', map_location='44F5-44F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8907518, end_position_on_the_genomic_accession=8910885, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35886, CurrentID=0, Status='live', Symbol='Pgi', Aliases='Dmel_CG8251, CG8251, Dmel\\CG8251, GPI, Gpi, PGI, pgi', description='Phosphoglucose isomerase', other_designations='CG8251-PA|CG8251-PB|CG8251-PC|Pgi-PA|Pgi-PB|Pgi-PC|glucose phosphate isomerase|phosphoglucoisomerase|phosphoglucose isomerase', map_location='44F6-44F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8910900, end_position_on_the_genomic_accession=8914069, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35888, CurrentID=0, Status='live', Symbol='CG8248', Aliases='Dmel_ Dmel\\CG8248', description='CG8248 gene product from transcript CG8248-RB', other_designations='CG8248-PA|CG8248-PB', map_location='44F7-44F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8918940, end_position_on_the_genomic_accession=8920368, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35889, CurrentID=0, Status='live', Symbol='Spt', Aliases='Dmel_CG8247, CG8247, Dmel\\CG8247, spt', description='Spitting Image', other_designations='CG8247-PB|Spt-PB|spitting image', map_location='44F7-44F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8920270, end_position_on_the_genomic_accession=8922755, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35890, CurrentID=0, Status='live', Symbol='CG8243', Aliases='Dmel_ Dmel\\CG8243, gap40', description='CG8243 gene product from transcript CG8243-RA', other_designations='CG8243-PA', map_location='44F7-44F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8922785, end_position_on_the_genomic_accession=8925346, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35891, CurrentID=0, Status='live', Symbol='CG8237', Aliases='Dmel_ Dmel\\CG8237', description='CG8237 gene product from transcript CG8237-RA', other_designations='CG8237-PA', map_location='44F8-44F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8925671, end_position_on_the_genomic_accession=8926890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35892, CurrentID=0, Status='live', Symbol='CG8235', Aliases='Dmel_ Dmel\\CG8235', description='CG8235 gene product from transcript CG8235-RA', other_designations='CG8235-PA', map_location='44F8-44F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8927014, end_position_on_the_genomic_accession=8928105, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35893, CurrentID=0, Status='live', Symbol='PAN2', Aliases='Dmel_CG8232, A1Z7K9, CG8232, Dmel\\CG8232, Pan2, pan2', description='PAN2 ortholog (S. cerevisiae)', other_designations='CG8232-PA|CG8232-PB|PAN2 ortholog|PAN2-PA|PAN2-PB', map_location='44F8-44F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8928169, end_position_on_the_genomic_accession=8933103, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  35894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35894, CurrentID=0, Status='live', Symbol='CG13749', Aliases='Dmel_ Dmel\\CG13749', description='CG13749 gene product from transcript CG13749-RB', other_designations='CG13749-PB|uncharacterized protein', map_location='44F9-44F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8933423, end_position_on_the_genomic_accession=8934445, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35895, CurrentID=0, Status='live', Symbol='FANCI', Aliases='Dmel_CG13745, CG13745, Dmel\\CG13745, fanci', description='Fanconi anemia complementation group I homologue', other_designations='CG13745-PA|FANCI-PA|fanconi anemia complementation group I homologue', map_location='44F9-44F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8934576, end_position_on_the_genomic_accession=8939789, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  35896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35896, CurrentID=0, Status='live', Symbol='CG13748', Aliases='Dmel_ Dmel\\CG13748', description='CG13748 gene product from transcript CG13748-RA', other_designations='CG13748-PA|uncharacterized protein', map_location='44F9-44F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8940415, end_position_on_the_genomic_accession=8941136, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35897, CurrentID=0, Status='live', Symbol='CG8230', Aliases='Dmel_ BcDNA:GH02536, Dmel\\CG8230', description='CG8230 gene product from transcript CG8230-RA', other_designations='CG8230-PA', map_location='44F9-44F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8944298, end_position_on_the_genomic_accession=8947344, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35898, CurrentID=0, Status='live', Symbol='CG8229', Aliases='Dmel_ Dmel\\CG8229, Uspl1l', description='CG8229 gene product from transcript CG8229-RE', other_designations='CG8229-PD|CG8229-PE', map_location='44F11-44F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8948113, end_position_on_the_genomic_accession=8951086, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35899, CurrentID=0, Status='live', Symbol='Tom7', Aliases='Dmel_CG8226, BcDNA:AT23266, CG8226, Dmel\\CG8226, dtom7', description='Translocase of outer membrane 7', other_designations='CG8226-PA|CG8226-PB|Tom7-PA|Tom7-PB|translocase of outer membrane 7', map_location='44F11-44F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8951595, end_position_on_the_genomic_accession=8952242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35900, CurrentID=0, Status='live', Symbol='babo', Aliases='Dmel_CG8224, ATR-1, ATR-I, Atf1, Atr-1, Atr-I, Atr45A, AtrI, BABO, Bab, Babo, CG8224, Dmel\\CG8224, Gprk-3, Gprk3, STK-E, atr-I, l(2)44Fd, l(2)k16912', description='baboon', other_designations='CG8224-PA|CG8224-PB|CG8224-PC|G protein-coupled receptor kinase 3|activin A receptor 45A|activin type I receptor|babo-PA|babo-PB|babo-PC|baboa', map_location='44F11-44F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8952339, end_position_on_the_genomic_accession=8962796, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  35901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35901, CurrentID=0, Status='live', Symbol='CG8216', Aliases='Dmel_ Dmel\\CG8216', description='CG8216 gene product from transcript CG8216-RA', other_designations='Apis-ITG-like|CG8216-PA', map_location='44F12-44F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8963164, end_position_on_the_genomic_accession=8968184, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35902, CurrentID=0, Status='live', Symbol='CG8213', Aliases='Dmel_ Dmel\\CG8213, SP44, c-SP44', description='CG8213 gene product from transcript CG8213-RB', other_designations='CG8213-PB|CG8213-PC|CG8213-PD', map_location='44F12-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8969346, end_position_on_the_genomic_accession=8985900, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35904, CurrentID=0, Status='live', Symbol='Np', Aliases='Dmel_CG34350, CG11824, CG34350, CG8181, Dmel\\CG34350, Dmel_CG11824, Dmel_CG8181, SP59, np', description='Notopleural', other_designations='CG34350-PA|CG34350-PB|Np-PA|Np-PB|notopleural', map_location='44F10-45A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8992945, end_position_on_the_genomic_accession=9006587, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35905, CurrentID=0, Status='live', Symbol='CG8172', Aliases='Dmel_ Dmel\\CG8172, SP32, c-SP32', description='CG8172 gene product from transcript CG8172-RD', other_designations='CG8172-PD|CG8172-PE|CG8172-PF|CG8172-PG', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9007234, end_position_on_the_genomic_accession=9013526, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35906, CurrentID=0, Status='live', Symbol='CG13744', Aliases='Dmel_ Dmel\\CG13744, SP67, c-SP67', description='CG13744 gene product from transcript CG13744-RA', other_designations='CG13744-PA|uncharacterized protein', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9014368, end_position_on_the_genomic_accession=9017195, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35907, CurrentID=0, Status='live', Symbol='CG13747', Aliases='Dmel_ Dmel\\CG13747, dmCG13747', description='CG13747 gene product from transcript CG13747-RA', other_designations='CG13747-PA|uncharacterized protein', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9017815, end_position_on_the_genomic_accession=9019032, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35908, CurrentID=0, Status='live', Symbol='CG8170', Aliases='Dmel_ Dmel\\CG8170, SP54', description='CG8170 gene product from transcript CG8170-RA', other_designations='CG8170-PA|CG8170-PB', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9019032, end_position_on_the_genomic_accession=9031045, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35909, CurrentID=0, Status='live', Symbol='Ance-4', Aliases='Dmel_CG8196, CG8196, Dmel\\CG8196, ance-4', description='CG8196 gene product from transcript CG8196-RA', other_designations='Ance-4-PA|Ance-4-PB|CG8196-PA|CG8196-PB', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9039728, end_position_on_the_genomic_accession=9042142, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35910, CurrentID=0, Status='live', Symbol='PPO2', Aliases='Dmel_CG8193, CG8193, DmePPO2, Dmel\\CG8193, PO45, ProPO45, proPO45, proPOA(3)', description='Prophenoloxidase 2', other_designations='CG8193-PA|PPO2-PA|phenoloxidase A3|phenoloxidase subunit A3|prophenoloxidase 2|prophenoloxidase 45', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9042261, end_position_on_the_genomic_accession=9044709, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35911, CurrentID=0, Status='live', Symbol='CG13743', Aliases='Dmel_ Dmel\\CG13743', description='CG13743 gene product from transcript CG13743-RC', other_designations='CG13743-PA|CG13743-PB|CG13743-PC|uncharacterized protein', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9044850, end_position_on_the_genomic_accession=9060608, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  35912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35912, CurrentID=0, Status='live', Symbol='CG8197', Aliases='Dmel_ Dmel\\CG8197', description='CG8197 gene product from transcript CG8197-RB', other_designations='CG8197-PA|CG8197-PB', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9060461, end_position_on_the_genomic_accession=9061924, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35913, CurrentID=0, Status='live', Symbol='ana', Aliases='Dmel_CG8084, CG8084, Dmel\\CG8084, med', description='anachronism', other_designations='CG8084-PA|CG8084-PB|CG8084-PC|ana-PA|ana-PB|ana-PC|medullaless', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9062295, end_position_on_the_genomic_accession=9077554, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35914, CurrentID=0, Status='live', Symbol='CG8083', Aliases='Dmel_ BEST:CK00459, CK00459, DmCnt2, Dmel\\CG8083', description='CG8083 gene product from transcript CG8083-RA', other_designations='CG8083-PA|CG8083-PB|CG8083-PC', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9078721, end_position_on_the_genomic_accession=9082374, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35915, CurrentID=0, Status='live', Symbol='CNT1', Aliases='Dmel_CG11778, CG11778, DmCnt1, Dmel\\CG11778', description='Concentrative Nucleoside Transporter 1', other_designations='CG11778-PB|CNT1-PB|concentrative nucleoside transporter 1', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9082395, end_position_on_the_genomic_accession=9085717, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  35916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35916, CurrentID=0, Status='live', Symbol='CG13742', Aliases='Dmel_ BcDNA:LD16446, Dmel\\CG13742', description='CG13742 gene product from transcript CG13742-RB', other_designations='CG13742-PB|CG13742-PC|CG13742-PD|CG13742-PE|uncharacterized protein', map_location='45A1-45A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9084598, end_position_on_the_genomic_accession=9087808, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35917, CurrentID=0, Status='live', Symbol='Dbp45A', Aliases='Dmel_CG12759, CG12759, DBP45A, Dbp45, DmRH18, Dmel\\CG12759, dbp45A', description='DEAD box protein 45A', other_designations='ATP-dependent RNA helicase|CG12759-PA|Dbp45A-PA', map_location='45A1-45A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9087834, end_position_on_the_genomic_accession=9089802, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35918, CurrentID=0, Status='live', Symbol='CG8080', Aliases='Dmel_ CG30346, Dmel\\CG8080', description='CG8080 gene product from transcript CG8080-RC', other_designations='CG8080-PA|CG8080-PB|CG8080-PC|CG8080-PD|CG8080-PE', map_location='45A2-45A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9089774, end_position_on_the_genomic_accession=9092893, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35919, CurrentID=0, Status='live', Symbol='CG13741', Aliases='Dmel_ CG30346, Dmel\\CG13741', description='CG13741 gene product from transcript CG13741-RA', other_designations='CG13741-PA|uncharacterized protein', map_location='45A2-45A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9092506, end_position_on_the_genomic_accession=9094658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35920, CurrentID=0, Status='live', Symbol='CG8078', Aliases='Dmel_ Dmel\\CG8078, anon-WO0118547.76', description='CG8078 gene product from transcript CG8078-RA', other_designations='CG8078-PA', map_location='45A3-45A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9096931, end_position_on_the_genomic_accession=9098525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35921, CurrentID=0, Status='live', Symbol='CG8777', Aliases='Dmel_ Dmel\\CG8777, SKI3, ski3', description='CG8777 gene product from transcript CG8777-RA', other_designations='CG8777-PA', map_location='45A3-45A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9098539, end_position_on_the_genomic_accession=9103153, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  35922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35922, CurrentID=0, Status='live', Symbol='Vang', Aliases='Dmel_CG8075, CG8075, Dmel\\CG8075, Stb, Stbm/Stbm, stb, stbm, stbm/vang, strabismus/Van Gogh, vang, vang/stbm, vangogh/strabismus, Vang', description='Van Gogh', other_designations='CG8075-PA|CG8075-PB|Van Gogh/Strabismus|Van gogh|VanGogh|Vang-PA|Vang-PB|Vang-gogh|sltrabismus|strabismus|van gogh|van-gogh|vang gogh', map_location='45A4-45A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9103238, end_position_on_the_genomic_accession=9106796, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35923, CurrentID=0, Status='live', Symbol='Pmm45A', Aliases='Dmel_CG8073, CG8073, Dmel\\CG8073', description='Phosphomannomutase 45A', other_designations='CG8073-PA|Pmm45A-PA|phosphomannomutase 45A', map_location='45A6-45A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9107151, end_position_on_the_genomic_accession=9109718, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35924, CurrentID=0, Status='live', Symbol='tsu', Aliases='Dmel_CG8781, CG8781, Dmel\\CG8781, TSU, Tsu, Y14, Y14/Tsunagi, cg8781, y14', description='tsunagi', other_designations='CG8781-PA|Y14|tsu-PA', map_location='45A7-45A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9110799, end_position_on_the_genomic_accession=9111578, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35925, CurrentID=0, Status='live', Symbol='Mys45A', Aliases='Dmel_CG8070, CG8070, Dmel\\CG8070, Mys-45A, Mys45a, anon-WO0118547.170', description='Mystery 45A', other_designations='CG8070-PA|Mys45A-PA|mystery|mystery 45A', map_location='45A7-45A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9111505, end_position_on_the_genomic_accession=9113951, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35926, CurrentID=0, Status='live', Symbol='Phax', Aliases='Dmel_CG8069, BEST:GH22533, CG8069, Dmel\\CG8069, dPHAX', description='Phosphorylated adaptor for RNA export', other_designations='CG8069-PA|CG8069-PB|PHosphorylated adaptor for RNA export|Phax-PA|Phax-PB|phosphorylated adaptor for RNA export', map_location='45A8-45A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9114274, end_position_on_the_genomic_accession=9116030, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35927, CurrentID=0, Status='live', Symbol='Su(var)2-10', Aliases='Dmel_CG8068, CG8068, CLOT2057, DPIAS, DmPias, Dmel\\CG8068, Dpias, PIAS, SU(VAR)2-10, Su(Var)2-10, Su(var)-10, Su-var(2)10, Suvar(2)10, ZIMP, ZimpA, ZimpB, dPIAS, dpias, i184, l(2)03697, pias, zimp', description='Suppressor of variegation 2-10', other_designations='CG8068-PA|CG8068-PB|CG8068-PC|CG8068-PD|CG8068-PE|CG8068-PF|CG8068-PG|CG8068-PH|CG8068-PI|CG8068-PJ|CG8068-PK|CG8068-PL|CG8068-PM|CG8068-PN|Su(var)2-10-PA|Su(var)2-10-PB|Su(var)2-10-PC|Su(var)2-10-PD|Su(var)2-10-PE|Su(var)2-10-PF|Su(var)2-10-PG|Su(var)2-10-PH|Su(var)2-10-PI|Su(var)2-10-PJ|Su(var)2-10-PK|Su(var)2-10-PL|Su(var)2-10-PM|Su(var)2-10-PN|protein inhibitor of activated STAT|suppressor of variegation 2-10|zinc finger-containing Miz1 PIAS3-like', map_location='2-57.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9116122, end_position_on_the_genomic_accession=9122000, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  35928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35928, CurrentID=0, Status='live', Symbol='GstE13', Aliases='Dmel_CG11784, BcDNA:LD22317, CG11784, Dmel\\CG11784, GSTE13, gste13', description='Glutathione S transferase E13', other_designations='CG11784-PA|CG11784-PB|GstE13-PA|GstE13-PB|glutathione S transferase E13', map_location='45A9-45A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9122125, end_position_on_the_genomic_accession=9123309, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35929, CurrentID=0, Status='live', Symbol='CG18659', Aliases='Dmel_ BcDNA:GH08789, CG30347, CG30348, CG8059, Dmel\\CG18659, anon-WO0118547.140, anon-WO0118547.145', description='CG18659 gene product from transcript CG18659-RH', other_designations='CG18659-PA|CG18659-PC|CG18659-PD|CG18659-PE|CG18659-PF|CG18659-PH|CG18659-PI|uncharacterized protein', map_location='45A9-45A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9123304, end_position_on_the_genomic_accession=9132030, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  35930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35930, CurrentID=0, Status='live', Symbol='Hydr1', Aliases='Dmel_CG8058, CG8058, Dm.HYDR1, Dmel\\CG8058', description='alpha/beta hydrolase 1', other_designations='CG8058-PA|CG8058-PB|CG8058-PC|CG8058-PD|CG8058-PE|Hydr1-PA|Hydr1-PB|Hydr1-PC|Hydr1-PD|Hydr1-PE', map_location='45A11-45A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9132480, end_position_on_the_genomic_accession=9135453, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35931, CurrentID=0, Status='live', Symbol='alc', Aliases='Dmel_CG8057, 8057, AMPKbeta, CG8057, Dmel\\CG8057, FBgn0033383, betaAMPK, l(2)45Ad', description='alicorn', other_designations='CG8057-PA|CG8057-PB|alc-PA|alc-PB|beta homologue|beta subunit of the SNF1/AMP-activated protein kinase|lethal (2) 45Ad', map_location='45A11-45B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9135375, end_position_on_the_genomic_accession=9137933, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35932, CurrentID=0, Status='live', Symbol='CG8788', Aliases='Dmel_ BEST:HL05962, Dmel\\CG8788, n(2)k04512', description='CG8788 gene product from transcript CG8788-RA', other_designations='CG8788-PA|CG8788-PC', map_location='45A11-45A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9138486, end_position_on_the_genomic_accession=9141981, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35933, CurrentID=0, Status='live', Symbol='shrb', Aliases='Dmel_CG8055, CG8055, DmSNF2, Dmel\\CG8055, Dvps32, ESCRT-III, Shrb, Snf7, Snf7/Vps32, Vps32, anon-WO0118547.180, vps32, vps32/snf7', description='shrub', other_designations='CG8055-PA|Vps32|shrb-PA', map_location='45A12-45E', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9142146, end_position_on_the_genomic_accession=9143914, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35934, CurrentID=0, Status='live', Symbol='Prp38', Aliases='Dmel_CG30342, CG30342, CG8054, Dmel\\CG30342, dprp38', description='pre-mRNA processing factor 38', other_designations='CG30342-PA|Prp38-PA', map_location='45A12-45A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9143956, end_position_on_the_genomic_accession=9145151, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35935, CurrentID=0, Status='live', Symbol='CG8008', Aliases='Dmel_ Dmel\\CG8008', description='CG8008 gene product from transcript CG8008-RB', other_designations='CG8008-PA|CG8008-PB', map_location='45B1-45B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9157907, end_position_on_the_genomic_accession=9161436, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35936, CurrentID=0, Status='live', Symbol='CG8046', Aliases='Dmel_ Dmel\\CG8046', description='CG8046 gene product from transcript CG8046-RB', other_designations='CG8046-PA|CG8046-PB', map_location='45B1-45B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9162037, end_position_on_the_genomic_accession=9164967, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35937, CurrentID=0, Status='live', Symbol='Rad51C', Aliases='Dmel_CG2412, CG2412, Dm CG2412, Dmel\\CG2412, dm', description='CG2412 gene product from transcript CG2412-RB', other_designations='CG2412-PB|Rad51C-PB', map_location='45B1-45B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9165632, end_position_on_the_genomic_accession=9167670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35938, CurrentID=0, Status='live', Symbol='CG42382', Aliases='Dmel_ CG8016, Dmel\\CG42382, Dmel_CG8016', description='CG42382 gene product from transcript CG42382-RA', other_designations='CG42382-PA|uncharacterized protein', map_location='45B1-45B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9166566, end_position_on_the_genomic_accession=9167508, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35939, CurrentID=0, Status='live', Symbol='Rme-8', Aliases='Dmel_CG8014, CG8014, Dmel\\CG8014, E(shi)2-1, l(2)45Ba, rme-8', description='Receptor mediated endocytosis 8', other_designations='CG8014-PA|CG8014-PB|Rme-8-PA|Rme-8-PB|receptor mediated endocytosis 8', map_location='45B1-45B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9167763, end_position_on_the_genomic_accession=9177261, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  35940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35940, CurrentID=0, Status='live', Symbol='Rab32', Aliases='Dmel_CG8024, 24652026, 24652028, A1Z7S1, CG8024, DRABR1, DmRab32, Dmel\\CG8024, DrabRP1, RAB-RP1, Rab-RP1, Rab-RP1/CG8024, Rab-r1/RP1, Rab7L1, RabRP1, anon-WO0118547.106, beck, ltd, rab32', description='CG8024 gene product from transcript CG8024-RD', other_designations='CG8024-PB|CG8024-PC|CG8024-PD|CG8024-PE|CG8024-PH|CG8024-PI|CG8024-PJ|Rab-related protein 1|Rab32-PB|Rab32-PC|Rab32-PD|Rab32-PE|Rab32-PH|Rab32-PI|Rab32-PJ|lightoid', map_location='45B2-45B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9177864, end_position_on_the_genomic_accession=9186342, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35941, CurrentID=0, Status='live', Symbol='CG8026', Aliases='Dmel_ Dmel\\CG8026, Dromel_CG8026_FBtr0088611_mORF, anon-WO0242455.53, anon-WO0242455.6, anon-WO0242455.7', description='CG8026 gene product from transcript CG8026-RD', other_designations='CG8026-PB|CG8026-PC|CG8026-PD', map_location='45B3-45B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9186670, end_position_on_the_genomic_accession=9192096, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  35942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35942, CurrentID=0, Status='live', Symbol='unpg', Aliases='Dmel_CG1650, CG1650, Dmel\\CG1650, Gbx, anon-45C, unp, upg', description='unplugged', other_designations='CG1650-PA|unpg-PA', map_location='45B3-45B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9195205, end_position_on_the_genomic_accession=9199870, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35943, CurrentID=0, Status='live', Symbol='CG8027', Aliases='Dmel_ Cg8027, Dmel\\CG8027', description='CG8027 gene product from transcript CG8027-RA', other_designations='CG8027-PA', map_location='45B3-45B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9204410, end_position_on_the_genomic_accession=9207124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35944, CurrentID=0, Status='live', Symbol='VhaAC45', Aliases='Dmel_CG8029, CG8029, Dmel\\CG8029, anon-WO0118547.117, l(2)SH1634, l(2)SH2 1634, vhaAC45', description='Vacuolar H[+] ATPase AC45 accessory subunit', other_designations='CG8029-PA|CG8029-PB|VhaAC45-PA|VhaAC45-PB|lethal (2) SH1634|vacuolar H[+] ATPase AC45 accessory subunit|vacuolar H[+] ATPase accessory protein AC45', map_location='45B4-45B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9207561, end_position_on_the_genomic_accession=9209774, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35946, CurrentID=0, Status='live', Symbol='Cyp4p2', Aliases='Dmel_CG1944, 4p2, CG1944, CYP4P2, Dmel\\CG1944', description='CG1944 gene product from transcript CG1944-RA', other_designations='CG1944-PA|Cyp4p2-PA|cytochrome P450', map_location='45B7-45B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9237512, end_position_on_the_genomic_accession=9239831, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  5740541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740541, CurrentID=0, Status='live', Symbol='CG34180', Aliases='Dmel_ Dmel\\CG34180', description='CG34180 gene product from transcript CG34180-RB', other_designations='CG34180-PB|uncharacterized protein', map_location='26B4-26B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6060510, end_position_on_the_genomic_accession=6060938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35948, CurrentID=0, Status='live', Symbol='Cyp4p3', Aliases='Dmel_CG10843, 4p3, CG10843, Dmel\\CG10843', description='CG10843 gene product from transcript CG10843-RA', other_designations='CG10843-PA|Cyp4p3-PA|cytochrome P450', map_location='45B7-45B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9242320, end_position_on_the_genomic_accession=9244410, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35949, CurrentID=0, Status='live', Symbol='hig', Aliases='Dmel_CG2040, CG2040, CT6282, Dmel\\CG2040, HIG', description='hikaru genki', other_designations='CG2040-PA|CG2040-PB|CG2040-PC|CG2040-PD|hig-PA|hig-PB|hig-PC|hig-PD', map_location='45B7-45B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9227785, end_position_on_the_genomic_accession=9247249, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35950, CurrentID=0, Status='live', Symbol='Pkn', Aliases='Dmel_CG2049, 3-2, CG2049, CG2055, CT6637, CT6660, DPKN, Dmel\\CG2049, Dpkn, PKC-related, PKN, PRK2, Pk45C, Pk?3, dPKN, l(2)06736, l(2)45Ca, pkn, v(2)rG232', description='Protein kinase N', other_designations='CG2049-PB|CG2049-PC|CG2049-PD|CG2049-PE|CG2049-PF|CG2049-PG|CG2049-PH|CG2049-PJ|CG2049-PK|CG2049-PL|CG2049-PM|CG2049-PN|CG2049-PO|CG2049-PP|CG2049-PQ|CG2049-PS|Pkn-PB|Pkn-PC|Pkn-PD|Pkn-PE|Pkn-PF|Pkn-PG|Pkn-PH|Pkn-PJ|Pkn-PK|Pkn-PL|Pkn-PM|Pkn-PN|Pkn-PO|Pkn-PP|Pkn-PQ|Pkn-PS|protein kinase N|protein kinase c related|protein kinase related to protein kinase N|protein kinase-like 45C', map_location='45C1-45C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9252310, end_position_on_the_genomic_accession=9285097, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  35952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35952, CurrentID=0, Status='live', Symbol='CG2063', Aliases='Dmel_ Dmel\\CG2063', description='CG2063 gene product from transcript CG2063-RA', other_designations='CG2063-PA', map_location='45C1-45C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9285310, end_position_on_the_genomic_accession=9286634, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35953, CurrentID=0, Status='live', Symbol='CG1968', Aliases='Dmel_ Dmel\\CG1968', description='CG1968 gene product from transcript CG1968-RB', other_designations='CG1968-PA|CG1968-PB', map_location='45C1-45C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9286827, end_position_on_the_genomic_accession=9289026, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35954, CurrentID=0, Status='live', Symbol='Mad1', Aliases='Dmel_CG2072, CG2072, Dmel\\CG2072, TXBP181-like, mad1', description='CG2072 gene product from transcript CG2072-RA', other_designations='CG2072-PA|Mad1-PA|TXBP181-like', map_location='45C4-45C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9288957, end_position_on_the_genomic_accession=9291565, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35955, CurrentID=0, Status='live', Symbol='Drep2', Aliases='Dmel_CG1975, CG1975, DREP-2, Dmel\\CG1975, Drep-2, Rep2, anon-WO0118547.162, anon-WO0118547.168, drep2, rep2', description='DNA fragmentation factor-related protein 2', other_designations='CG1975-PA|CG1975-PB|CG1975-PC|CG1975-PD|Drep2-PA|Drep2-PB|Drep2-PC|Drep2-PD|Rep2', map_location='45C4-45C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9291856, end_position_on_the_genomic_accession=9302821, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35956, CurrentID=0, Status='live', Symbol='Myd88', Aliases='Dmel_CG2078, CG2078, DMMYD88, DmMyD88, DmMyd88, Dmel\\CG2078, EP(2)2535, Kra, LD20892, MYD88, MyD88F, dMYd88, dMyD88, dMyd88, kra, myd88', description='CG2078 gene product from transcript CG2078-RB', other_designations='CG2078-PA|CG2078-PB|CG2078-PC|Myd88-PA|Myd88-PB|Myd88-PC|Myd88F|krapfen', map_location='45C5-45C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9302823, end_position_on_the_genomic_accession=9308722, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35957, CurrentID=0, Status='live', Symbol='CG13739', Aliases='Dmel_ BcDNA:GH21254, Dmel\\CG13739', description='CG13739 gene product from transcript CG13739-RA', other_designations='CG13739-PA|uncharacterized protein', map_location='45C6-45D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9309296, end_position_on_the_genomic_accession=9351655, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35958, CurrentID=0, Status='live', Symbol='Or45a', Aliases='Dmel_CG1978, 45C.1, 45a, CG1978, DOR58, Dmel\\CG1978', description='Odorant receptor 45a', other_designations='CG1978-PA|Or45a-PA|odorant receptor 45a', map_location='45D1-45D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9345349, end_position_on_the_genomic_accession=9348863, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35959, CurrentID=0, Status='live', Symbol='CG13954', Aliases='Dmel_ Dmel\\CG13954', description='CG13954 gene product from transcript CG13954-RC', other_designations='CG13954-PC|CG13954-PD|uncharacterized protein', map_location='45D1-45D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9309296, end_position_on_the_genomic_accession=9389467, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35962, CurrentID=0, Status='live', Symbol='CG8800', Aliases='Dmel_ Dmel\\CG8800', description='CG8800 gene product from transcript CG8800-RA', other_designations='CG8800-PA', map_location='45D1-45D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9395330, end_position_on_the_genomic_accession=9396206, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35963, CurrentID=0, Status='live', Symbol='Non1', Aliases='Dmel_CG8801, BcDNA:LD23830, CG30341, CG8801, Dmel\\CG8801, LD23830', description='Novel nucleolar protein 1', other_designations='CG8801-PA|CG8801-PB|Non1-PA|Non1-PB|novel nucleolar protein 1', map_location='45D2-45D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9396265, end_position_on_the_genomic_accession=9398734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740394, CurrentID=0, Status='live', Symbol='CG34164', Aliases='Dmel_ Dmel\\CG34164', description='CG34164 gene product from transcript CG34164-RB', other_designations='CG34164-PB|CG34164-PC|uncharacterized protein', map_location='33C1-33C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12056348, end_position_on_the_genomic_accession=12058251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35966, CurrentID=0, Status='live', Symbol='wun', Aliases='Dmel_CG8804, CG8804, Dmel\\CG8804, LPP-like, Wun, Wunen, l(2)k16806', description='wunen', other_designations='CG8804-PA|CG8804-PB|CG8804-PC|wun-PA|wun-PB|wun-PC', map_location='45D3-45D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9400531, end_position_on_the_genomic_accession=9413463, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  35968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35968, CurrentID=0, Status='live', Symbol='CG13955', Aliases='Dmel_ Dmel\\CG13955', description='CG13955 gene product from transcript CG13955-RA', other_designations='CG13955-PA|uncharacterized protein', map_location='45D5-45D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9421972, end_position_on_the_genomic_accession=9423688, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35969, CurrentID=0, Status='live', Symbol='prel', Aliases='Dmel_CG8806, BcDNA:LD36516, CG8806, Dmel\\CG8806', description='preli-like', other_designations='CG8806-PB|CG8806-PC|Preli-like|prel-PB|prel-PC|preli|protein of relevant evolutionary and lymphoid interest-like', map_location='45D5-45D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9424087, end_position_on_the_genomic_accession=9426107, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35970, CurrentID=0, Status='live', Symbol='Pdk', Aliases='Dmel_CG8808, BcDNA:LD09837, CG8808, DmPDK, DmPdk, Dmel\\CG8808, PDK, pdk', description='Pyruvate dehydrogenase kinase', other_designations='CG8808-PA|CG8808-PB|CG8808-PC|CG8808-PD|Pdk-PA|Pdk-PB|Pdk-PC|Pdk-PD|pyruvate dehydrogenase kinase', map_location='45D5-45D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9426439, end_position_on_the_genomic_accession=9434638, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35971, CurrentID=0, Status='live', Symbol='ced-6', Aliases='Dmel_CG11804, CG11804, Ced-6, Dmel\\CG11804, ced6, dCed-6, d drCed-6, drced-6', description='CG11804 gene product from transcript CG11804-RD', other_designations='CG11804-PA|CG11804-PB|CG11804-PC|CG11804-PD|CG11804-PF|ced-6-PA|ced-6-PB|ced-6-PC|ced-6-PD|ced-6-PF', map_location='45D7-45D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9434669, end_position_on_the_genomic_accession=9444937, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  35974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35974, CurrentID=0, Status='live', Symbol='Camta', Aliases='Dmel_CG42332, CG13952, CG13953, CG42332, CG8809, DmCAMTA, Dmel\\CG42332, Dmel_CG13952, Dmel_CG13953, Dmel_CG8809, camta, dCAMTA', description='Calmodulin-binding transcription activator', other_designations='CG42332-PB|CG42332-PD|CG42332-PE|CG42332-PF|CG42332-PG|CG42332-PH|CG42332-PI|Calmodulin-binding transcription factor|Camta-PB|Camta-PD|Camta-PE|Camta-PF|Camta-PG|Camta-PH|Camta-PI|calmodulin-binding transcription activator', map_location='45E1-45E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9446046, end_position_on_the_genomic_accession=9480076, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  35975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35975, CurrentID=0, Status='live', Symbol='Wnt2', Aliases='Dmel_CG1916, CG1916, D-wnt-2, DWnt-2, DWnt2, Dm DWnt2, Dm-2, Dmel\\CG1916, Dwnt-2, Wnt, Wnt-2, dWnt2, wnt2', description='Wnt oncogene analog 2', other_designations='CG1916-PA|CG1916-PB|Wnt2-PA|Wnt2-PB', map_location='45E1-45E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9494166, end_position_on_the_genomic_accession=9503188, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  35977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35977, CurrentID=0, Status='live', Symbol='brp', Aliases='Dmel_CG42344, BRP, Brp, CG12932, CG12933, CG1931, CG30336, CG30337, CG34146, CG42344, Cast, Dmel\\CG42344, Dmel_CG12932, Dmel_CG34146, MAbnc82, Mabnc82, NC82, Nc82, mAb nc82, mAbnc82, nc-82, nc82', description='bruchpilot', other_designations='CG42344-PD|CG42344-PE|CG42344-PG|CG42344-PH|CG42344-PI|CG42344-PJ|CG42344-PK|CG42344-PL|CG42344-PM|Nc82|brp-PD|brp-PE|brp-PG|brp-PH|brp-PI|brp-PJ|brp-PK|brp-PL|brp-PM|burchpilot|monoclonal antibody nc82', map_location='45E1-45F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9503765, end_position_on_the_genomic_accession=9541565, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  35979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35979, CurrentID=0, Status='live', Symbol='CG1888', Aliases='Dmel_ Dmel\\CG1888', description='CG1888 gene product from transcript CG1888-RC', other_designations='CG1888-PA|CG1888-PB|CG1888-PC|CG1888-PD', map_location='45F1-45F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9541947, end_position_on_the_genomic_accession=9548135, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35980, CurrentID=0, Status='live', Symbol='Or45b', Aliases='Dmel_CG12931, 45F.1, 45b, CG12931, DOR107, Dmel\\CG12931', description='Odorant receptor 45b', other_designations='CG12931-PA|Or45b-PA|odorant receptor 45b', map_location='45F1-45F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9559598, end_position_on_the_genomic_accession=9561068, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  35981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35981, CurrentID=0, Status='live', Symbol='CG1809', Aliases='Dmel_ Dmel\\CG1809', description='CG1809 gene product from transcript CG1809-RA', other_designations='CG1809-PA', map_location='45F1-45F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9561638, end_position_on_the_genomic_accession=9563871, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35984, CurrentID=0, Status='live', Symbol='CG1814', Aliases='Dmel_ 45550393, 45551045, Dmel\\CG1814', description='CG1814 gene product from transcript CG1814-RC', other_designations='CG1814-PA|CG1814-PB|CG1814-PC', map_location='45F4-45F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9577427, end_position_on_the_genomic_accession=9581847, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  35985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35985, CurrentID=0, Status='live', Symbol='CG1868', Aliases='Dmel_ Dmel\\CG1868', description='CG1868 gene product from transcript CG1868-RB', other_designations='CG1868-PB|CG1868-PC', map_location='45F4-45F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9581761, end_position_on_the_genomic_accession=9584545, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  35986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35986, CurrentID=0, Status='live', Symbol='Updo', Aliases='Dmel_CG1818, CG1818, Dmel\\CG1818', description='CG1818 gene product from transcript CG1818-RA', other_designations='CG1818-PA|Updo-PA|updo', map_location='45F5-45F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9584643, end_position_on_the_genomic_accession=9586126, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  35987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35987, CurrentID=0, Status='live', Symbol='CG12929', Aliases='Dmel_ Dmel\\CG12929', description='CG12929 gene product from transcript CG12929-RA', other_designations='CG12929-PA|CG12929-PB|uncharacterized protein', map_location='45F5-45F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9586110, end_position_on_the_genomic_accession=9586714, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  35988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35988, CurrentID=0, Status='live', Symbol='RpL31', Aliases='Dmel_CG1821, BcDNA:RE59131, CG1821, Dmel\\CG1821, M(2)45F, Rp L31, anon-EST:Posey68', description='Ribosomal protein L31', other_designations='CG1821-PA|CG1821-PB|CG1821-PC|RpL31-PA|RpL31-PB|RpL31-PC|ribosomal protein L31', map_location='45F5-45F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9587193, end_position_on_the_genomic_accession=9587989, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35989, CurrentID=0, Status='live', Symbol='CG1827', Aliases='Dmel_ Dmel\\CG1827, cg1827', description='CG1827 gene product from transcript CG1827-RD', other_designations='CG1827-PD', map_location='45F5-45F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9589050, end_position_on_the_genomic_accession=9590443, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35991, CurrentID=0, Status='live', Symbol='Map60', Aliases='Dmel_CG1825, CG1825, CP-60, CP60, Cp60, DMAP 60, DMAP-60, DMAP60, Dmel\\CG1825, MAP60', description='Microtubule-associated protein 60', other_designations='CG1825-PA|Map60-PA|centrosome associated protein 60', map_location='45F5-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9590867, end_position_on_the_genomic_accession=9592744, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35993, CurrentID=0, Status='live', Symbol='CG1902', Aliases='Dmel_ Dmel\\CG1902', description='CG1902 gene product from transcript CG1902-RA', other_designations='CG1902-PA|CG1902-PC', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9594045, end_position_on_the_genomic_accession=9605624, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35995, CurrentID=0, Status='live', Symbol='GstT1', Aliases='Dmel_CG30000, BcDNA:RE63575, CG12930, CG12930B, CG30000, Dmel\\CG30000, GSTT1, gstt1', description='Glutathione S transferase T1', other_designations='CG30000-PA|GstT1-PA|glutathione S transferase T1', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9605988, end_position_on_the_genomic_accession=9607184, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  35996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35996, CurrentID=0, Status='live', Symbol='CG12926', Aliases='Dmel_ Dmel\\CG12926', description='CG12926 gene product from transcript CG12926-RA', other_designations='CG12926-PA|uncharacterized protein', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9608128, end_position_on_the_genomic_accession=9610290, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35997, CurrentID=0, Status='live', Symbol='Mmp2', Aliases='Dmel_CG1794, 2-MMP, CG1794, Dm2-MMP, Dmel\\CG1794, MMP2, anon-WO0118547.84, dm-2MMP, dmmp2, l(2)02353, mmp2', description='Matrix metalloproteinase 2', other_designations='CG1794-PB|CG1794-PC|CG1794-PD|Mmp2-PB|Mmp2-PC|Mmp2-PD|matrix metalloprotease 2|matrix metalloproteinase|matrix metalloproteinase 2|matrix metalloprotinase 2', map_location='45F6-46A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9611139, end_position_on_the_genomic_accession=9683858, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  35998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35998, CurrentID=0, Status='live', Symbol='Uba1', Aliases='Dmel_CG1782, 2/31, CG1782, DmUba1, Dmel\\CG1782, E1, FBtr0088499, UBA-1, Uba 1, l(2)03405, l(2)05642, l(2)s3484, uba-1, uba1', description='Ubiquitin activating enzyme 1', other_designations='CG1782-PA|CG1782-PC|E1|E1 ligase|Uba1-PA|Uba1-PC|ubiquitin activating enzyme|ubiquitin activating enzyme 1|ubiquitin-activating enzyme (E1)', map_location='46A1-46A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9688289, end_position_on_the_genomic_accession=9693328, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  35999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35999, CurrentID=0, Status='live', Symbol='CG1773', Aliases='Dmel_ Dmel\\CG1773, SP74', description='CG1773 gene product from transcript CG1773-RA', other_designations='CG1773-PA', map_location='46A1-46A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9706267, end_position_on_the_genomic_accession=9707580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36000, CurrentID=0, Status='live', Symbol='CG10459', Aliases='Dmel_ Dmel\\CG10459', description='CG10459 gene product from transcript CG10459-RA', other_designations='CG10459-PA|uncharacterized protein', map_location='46A1-46A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9707536, end_position_on_the_genomic_accession=9708722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36001, CurrentID=0, Status='live', Symbol='dap', Aliases='Dmel_CG1772, CDI4, CDKN2B, CG1772, CIB1, Cdi4, DAP, Dac, Dap, Decapo, Dmel\\CG1772, E(Sev-CycE)2B, P15, cdi4, dacapo/cyclin-dependent kinase interactor 4, l(2)04454, p21, p21[dacapo], p27, p27[Dap], p27cip/kip', description='dacapo', other_designations='CG1772-PA|Cyclin-dependent kinase interactor 4|Dacap0|Dacapo/p21|chakra|dacpao|dap-PA|decapo|p21/Dacapo', map_location='46B1-46B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9712321, end_position_on_the_genomic_accession=9716256, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36002, CurrentID=0, Status='live', Symbol='sqa', Aliases='Dmel_CG42347, CG1776, CG30004, CG42347, Dmel\\CG42347, Dmel_CG1776, Dmel_CG30004, FBgn0259678', description='spaghetti-squash activator', other_designations='CG42347-PB|CG42347-PC|CG42347-PD|MLCK-like|sqa-PB|sqa-PC|sqa-PD', map_location='46B2-46B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9716348, end_position_on_the_genomic_accession=9748121, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36003, CurrentID=0, Status='live', Symbol='trpl', Aliases='Dmel_CG18345, CG1694, CG18345, DmTRPL, Dmel\\CG18345, TRP-Like, TRPL, TrpL, Trpl, dTRPL, dTrpL, trp-l, trpL', description='transient receptor potential-like', other_designations='CG18345-PA|CG18345-PB|CG18345-PD|TRP-like|transient receptor Potential-Like|trpl-PA|trpl-PB|trpl-PD', map_location='46B2-46B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9753697, end_position_on_the_genomic_accession=9765518, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36004, CurrentID=0, Status='live', Symbol='CG1690', Aliases='Dmel_ Dmel\\CG1690', description='CG1690 gene product from transcript CG1690-RB', other_designations='CG1690-PA|CG1690-PB', map_location='46B2-46B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9756293, end_position_on_the_genomic_accession=9758068, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36005, CurrentID=0, Status='live', Symbol='CG1688', Aliases='Dmel_ BcDNA:GH04802, Dmel\\CG1688, anon-WO0109301.12, anon-WO0109301.47, anon-WO0109301.48, anon-WO0109301.49, anon-WO0109301.50', description='CG1688 gene product from transcript CG1688-RA', other_designations='CG1688-PA', map_location='46B2-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9766364, end_position_on_the_genomic_accession=9806875, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36006, CurrentID=0, Status='live', Symbol='CG1698', Aliases='Dmel_ DmNAT5, Dmel\\CG1698', description='CG1698 gene product from transcript CG1698-RA', other_designations='CG1698-PA|CG1698-PB', map_location='46B3-46B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9793326, end_position_on_the_genomic_accession=9798006, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36009, CurrentID=0, Status='live', Symbol='CG1648', Aliases='Dmel_ Dmel\\CG1648', description='CG1648 gene product from transcript CG1648-RC', other_designations='CG1648-PA|CG1648-PB|CG1648-PC', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9810974, end_position_on_the_genomic_accession=9820053, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36010, CurrentID=0, Status='live', Symbol='dila', Aliases='Dmel_CG1625, CG1625, CG1626, Dmel\\CG1625', description='dilatory', other_designations='CG1625-PC|CG1625-PD|CG1625-PE|dila-PC|dila-PD|dila-PE', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9823082, end_position_on_the_genomic_accession=9827206, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36011, CurrentID=0, Status='live', Symbol='hebe', Aliases='Dmel_CG1623, CG1623, Dmel\\CG1623', description='CG1623 gene product from transcript CG1623-RD', other_designations='CG1623-PB|CG1623-PD|CG1623-PE|CG1623-PF|hebe-PB|hebe-PD|hebe-PE|hebe-PF', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9828104, end_position_on_the_genomic_accession=9836642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36012, CurrentID=0, Status='live', Symbol='CG1663', Aliases='Dmel_ Dmel\\CG1663', description='CG1663 gene product from transcript CG1663-RA', other_designations='CG1663-PA', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9836985, end_position_on_the_genomic_accession=9838481, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36013, CurrentID=0, Status='live', Symbol='Lsm11', Aliases='Dmel_CG12924, CG12924, Dmel\\CG12924, LSM11, d', description='CG12924 gene product from transcript CG12924-RA', other_designations='CG12924-PA|Lsm11-PA', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9838434, end_position_on_the_genomic_accession=9839500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36014, CurrentID=0, Status='live', Symbol='CG1665', Aliases='Dmel_ Dmel\\CG1665, IP19920p', description='CG1665 gene product from transcript CG1665-RA', other_designations='CG1665-PA', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9839776, end_position_on_the_genomic_accession=9841163, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36015, CurrentID=0, Status='live', Symbol='Vamp7', Aliases='Dmel_CG1599, CG1599, Dmel\\CG1599, TI-VAMP, VAMP7', description='Vesicle-associated membrane protein 7', other_designations='CG1599-PA|CG1599-PC|Vamp7-PA|Vamp7-PC', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9841080, end_position_on_the_genomic_accession=9842502, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36016, CurrentID=0, Status='live', Symbol='CG1667', Aliases='Dmel_ Dmel\\CG1667', description='CG1667 gene product from transcript CG1667-RC', other_designations='CG1667-PB|CG1667-PC', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9842686, end_position_on_the_genomic_accession=9845512, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36017, CurrentID=0, Status='live', Symbol='Orc6', Aliases='Dmel_CG1584, CG1584, DmORC 6, DmORC6, Dmel\\CG1584, ORC, ORC6, orc6, rDmORC', description='Origin recognition complex subunit 6', other_designations='CG1584-PA|Orc6-PA|origin recognition complex subunit 6', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9845458, end_position_on_the_genomic_accession=9846485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36018, CurrentID=0, Status='live', Symbol='Prosalpha7', Aliases='Dmel_CG1519, 1519, Alpha-7, Alpha7, CG1519, CPalpha7, Dmel\\CG1519, PROSalpha7, PSA3_DROME, alpha-7, alpha7, alpha7_dm', description='Proteasome alpha7 subunit', other_designations='20S proteasome subunit alpha 7|20S proteasome subunit alpha-7|20s proteasome|CG1519-PA|Prosalpha7-PA|alpha-subunit of 20S proteasome|alpha7|proteasome alpha7 subunit|proteasome subunit alpha type-3', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9846636, end_position_on_the_genomic_accession=9847836, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36019, CurrentID=0, Status='live', Symbol='CG1671', Aliases='Dmel_ Dmel\\CG1671', description='CG1671 gene product from transcript CG1671-RA', other_designations='CG1671-PA', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9848061, end_position_on_the_genomic_accession=9850642, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36020, CurrentID=0, Status='live', Symbol='PCB', Aliases='Dmel_CG1516, BcDNA, BcDNA:GH06348, CG1516, Dmel\\CG1516, PC, Q0E9E2_DROME', description='Pyruvate carboxylase', other_designations='CG1516-PA|CG1516-PB|CG1516-PD|CG1516-PE|CG1516-PG|CG1516-PI|CG1516-PJ|CG1516-PK|CG1516-PL|CG1516-PM|GH06348|PCB-PA|PCB-PB|PCB-PD|PCB-PE|PCB-PG|PCB-PI|PCB-PJ|PCB-PK|PCB-PL|PCB-PM|pyruvate carboxykinase|pyruvate carboxylase', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9850703, end_position_on_the_genomic_accession=9861543, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36022, CurrentID=0, Status='live', Symbol='Ntmt', Aliases='Dmel_CG1675, CG1675, Dmel\\CG1675, dNTMT', description='N-terminal methyltransferase', other_designations='CG1675-PA|Ntmt-PA', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9865632, end_position_on_the_genomic_accession=9866684, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36023, CurrentID=0, Status='live', Symbol='CG18446', Aliases='Dmel_ Dmel\\CG18446', description='CG18446 gene product from transcript CG18446-RA', other_designations='CG18446-PA|uncharacterized protein', map_location='46C1-46C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9871470, end_position_on_the_genomic_accession=9873646, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36024, CurrentID=0, Status='live', Symbol='CG12744', Aliases='Dmel_ Dmel\\CG12744', description='CG12744 gene product from transcript CG12744-RA', other_designations='CG12744-PA|CG12744-PB|uncharacterized protein', map_location='46C1-46C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9875281, end_position_on_the_genomic_accession=9876242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36025, CurrentID=0, Status='live', Symbol='Sec24AB', Aliases='Dmel_CG1472, CG1472, Dmel\\CG1472, Hau, SEC24, Sec24, sec24', description='Sec24AB ortholog (H. sapiens)', other_designations='CG1472-PA|CG1472-PB|CG1472-PC|Sec24AB ortholog|Sec24AB-PA|Sec24AB-PB|Sec24AB-PC|sec24', map_location='46C1-46C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9876232, end_position_on_the_genomic_accession=9882357, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36026, CurrentID=0, Status='live', Symbol='CG12923', Aliases='Dmel_ Dmel\\CG12923', description='CG12923 gene product from transcript CG12923-RB', other_designations='CG12923-PA|CG12923-PB|uncharacterized protein', map_location='46C2-46C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9887648, end_position_on_the_genomic_accession=9889485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36027, CurrentID=0, Status='live', Symbol='CG30007', Aliases='Dmel_ CG12138, Dmel\\CG30007', description='CG30007 gene product from transcript CG30007-RB', other_designations='CG30007-PB|CG30007-PC|CG30007-PD|CG30007-PE|uncharacterized protein', map_location='46C2-46C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9891955, end_position_on_the_genomic_accession=9898513, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36028, CurrentID=0, Status='live', Symbol='CG1513', Aliases='Dmel_ Dmel\\CG1513', description='CG1513 gene product from transcript CG1513-RA', other_designations='CG1513-PA|CG1513-PB|CG1513-PC', map_location='46C2-46C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9882572, end_position_on_the_genomic_accession=9891689, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  36029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36029, CurrentID=0, Status='live', Symbol='CG1441', Aliases='Dmel_ Dmel\\CG1441', description='CG1441 gene product from transcript CG1441-RB', other_designations='CG1441-PA|CG1441-PB', map_location='46C4-46C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9899018, end_position_on_the_genomic_accession=9903287, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36030, CurrentID=0, Status='live', Symbol='FMRFa', Aliases='Dmel_CG2346, CG2346, DPKQDFMRF-amide, DPKQDFMRFamide, Dmel\\CG2346, Drm-FMRF-1, Drm-FMRF-2, Drm-FMRF-3, Drm-FMRF-4, Drm-FMRF-5, Drm-FMRFa1, Drm-FMRFa2, Drm-FMRFa3, Drm-FMRFa4, FMRF-1, FMRFa-2, FMRFa-3, FMRFa-4, FMRFa-5, FMRFa-6, FMRFa-7, FMRFa-8, FMRFa-like, Fmrf, dFMRF, dFMRFa, dFMRFa-1, dFMRFa-2, dFMRFa-3, dFMRFa-4, dFMRFa-5, dFMRFa-6, dFMRFa-7, dFMRFa-8, dFmrf, dFmrfa, fmrf, FMRFa', description='FMRFamide', other_designations='CG2346-PA|DPKQDFMRFamide|Drm-FMRFamide-1|Drm-FMRFamide-2|Drm-FMRFamide-3|Drm-FMRFamide-4|Drm-FMRFamide-5|FMRF amide homologue precursor|FMRF-amide|FMRFa-PA|FMRFamide prohormone|FMRFamide-related|FRMFamide|SDNFMRFamide|TPAEDFMRFamide|amino acid amidated neuropeptide|dFMRFamide', map_location='46C4-46C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9906294, end_position_on_the_genomic_accession=9910457, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36031, CurrentID=0, Status='live', Symbol='Etf-QO', Aliases='Dmel_CG12140, CG 12140, CG12140, Dmel\\CG12140, ETF:QO, ETFD, ETFDH', description='Electron transfer flavoprotein-ubiquinone oxidoreductase', other_designations='CG12140-PA|CG12140-PB|Etf-QO-PA|Etf-QO-PB|electron transfer flavoprotein-ubiquinone oxidoreductase|electron transfer flavoprotein:ubiquinone oxidoreductase', map_location='46C4-46C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9910893, end_position_on_the_genomic_accession=9913488, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36032, CurrentID=0, Status='live', Symbol='Mef2', Aliases='Dmel_CG1429, 22.21, BEST:SD04091, C, CG1429, D-MEF2, D-Mef2, D-mef2, DMEF-2, DMEF2, DMedf2, DMef-2, DMef2, Dmef, Dmef-2, Dmef2, Dmel\\CG1429, MEF-2, MEF2, Mef, Mef-2, SD04091, dMEF2, dMEf2, dMef2, dmef2, l(2)46CFr, mb247, mef, mef-2, mef2, mef2c', description='Myocyte enhancer factor 2', other_designations='CG1429-PA|CG1429-PB|CG1429-PC|CG1429-PD|CG1429-PF|CG1429-PG|CG1429-PH|DMEF2|Drosophila-myocyte enhancer factor 2|Mef2-PA|Mef2-PB|Mef2-PC|Mef2-PD|Mef2-PF|Mef2-PG|Mef2-PH|Myocyte-enhancer factor 2|Myocyte-enhancer factor-2|complementation group C|drosophila myocyte-specific enhancer factor 2|group C|myocyte enhancer factor 2|myocyte enhancer factor-2|myocyte enhancing factor 2|myocyte-specific enhancer factor 2', map_location='46C4-46C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9913496, end_position_on_the_genomic_accession=9958858, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  36033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36033, CurrentID=0, Status='live', Symbol='Pal1', Aliases='Dmel_CG12130, CG12130, DPAL1, Dmel\\CG12130, PAL1, anon-WO0149856.3, dPAL1', description='Peptidyl-alpha-hydroxyglycine-alpha-amidating lyase 1', other_designations='CG12130-PA|CG12130-PC|CG12130-PD|Pal1-PA|Pal1-PC|Pal1-PD', map_location='46C7-46C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9959549, end_position_on_the_genomic_accession=9963853, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36034, CurrentID=0, Status='live', Symbol='CG15863', Aliases='Dmel_ Dmel\\CG15863', description='CG15863 gene product from transcript CG15863-RA', other_designations='CG15863-PA|uncharacterized protein', map_location='46C7-46C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9959119, end_position_on_the_genomic_accession=9959732, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36035, CurrentID=0, Status='live', Symbol='CG1418', Aliases='Dmel_ Dmel\\CG1418', description='CG1418 gene product from transcript CG1418-RA', other_designations='CG1418-PA', map_location='46C7-46C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9963835, end_position_on_the_genomic_accession=9965067, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36036, CurrentID=0, Status='live', Symbol='CG12133', Aliases='Dmel_ Dmel\\CG12133, SP15', description='CG12133 gene product from transcript CG12133-RB', other_designations='CG12133-PB|uncharacterized protein', map_location='46C8-46C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9967231, end_position_on_the_genomic_accession=9968388, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36037, CurrentID=0, Status='live', Symbol='Adam', Aliases='Dmel_CG12131, 5.3, CG12131, Dmel\\CG12131, E, Q7K550_DROME, eIF3-S1, eIF3j, l(2)46CFd, l(2)46Cf, l(2)k13906', description='CG12131 gene product from transcript CG12131-RA', other_designations='Adam-PA|Adam-PB|CG12131-PA|CG12131-PB|adam|complementation group E', map_location='46C8-46C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9968929, end_position_on_the_genomic_accession=9970897, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740544, CurrentID=0, Status='live', Symbol='futsch', Aliases='Dmel_CG34387, 22C10, 22C10/FUTSCH, 22C20, 22c10, CG14772, CG3064, CG34387, Dmel\\CG34387, Dmel_CG14772, Dmel_CG3064, EG:49E4.1, FUTSCH, Futch, Futsch, MAP-IB, MAP1B, MAb22C10, Mab 22C10, Mab22C10, Mab22c10, Map-1B, Map1B, Olk, lincRNA.936, mAb 22C10, mAb-22C10, mAb22C10, mAb22c10, map1b, olk, ssC10', description='CG34387 gene product from transcript CG34387-RF', other_designations='CG34387-PC|CG34387-PE|CG34387-PF|futch|futsch-PC|futsch-PE|futsch-PF|omb-like', map_location='2A3-2A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1409069, end_position_on_the_genomic_accession=1453686, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36039, CurrentID=0, Status='live', Symbol='eve', Aliases='Dmel_CG2328, 10.5, 10.9, 14.10, 20.35, CG2328, Dmel\\CG2328, E(eve), EVE, Eve, F, V, VI, dm-eve2, even, l(2)46CFg, l(2)46CFh, l(2)46CFj, l(2)46CFp, l(2)46Ce, l(2)46Cg, eve', description='even skipped', other_designations='CG2328-PA|complementation group F|eve-PA|even-skiped|even-skipped|evenskip|evenskipped|group V|group VI|lethal(2)46Ce', map_location='46C10-46C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9979319, end_position_on_the_genomic_accession=9980795, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36040, CurrentID=0, Status='live', Symbol='TER94', Aliases='Dmel_CG2331, 22.26, 7.12, Apple, BcDNA.GM02885, BcDNA:GM02885, CDC48, CG2331, Dm Dmel\\CG2331, I, Ter94, Tera, VCP, VCP ATPase, VCP/p97, Vcp, anon-WO2004063362.65, anon-WO2004063362.67, cdc48, dTERA, dVCP, dmTERA, dvcp, l(2)03775, l(2)46CFf, l(2)46CFs, l(2)46Ch, p97, pTER94, ter94, vcp', description='CG2331 gene product from transcript CG2331-RA', other_designations='CG2331-PA|CG2331-PC|CG2331-PD|CG2331-PE|TER94-PA|TER94-PC|TER94-PD|TER94-PE|Ter94|Valosin-Containing protein|complementation group I|p97 ATPase|transitional endoplasmic reticulum 94|transitional endoplasmic reticulum ATPase|valosin containing protein', map_location='46D-46D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9989159, end_position_on_the_genomic_accession=9993559, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36041, CurrentID=0, Status='live', Symbol='Pka-R2', Aliases='Dmel_CG15862, BcDNA:GM01761, CG15862, Cos, Cos-1, Cos1, Dmel\\CG15862, Epa, PKA, PKA RII, PKA-R2, PKA-RII, PKa-R2, Pka-RII, RII, RII[[DR]], cos1, pkA, pka-RII', description='Protein kinase, cAMP-dependent, regulatory subunit type 2', other_designations='CG15862-PA|CG15862-PB|CG15862-PD|CG15862-PE|Costal-1|Pka-R2-PA|Pka-R2-PB|Pka-R2-PD|Pka-R2-PE|cAMP-dependent protein kinase R2|cAMP-dependent protein kinase type II|costal|epaulette|protein kinase A|protein kinase, cAMP-dependent, regulatory subunit type 2', map_location='46D1-46D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9993890, end_position_on_the_genomic_accession=10026042, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36042, CurrentID=0, Status='live', Symbol='CG12128', Aliases='Dmel_ Dmel\\CG12128', description='CG12128 gene product from transcript CG12128-RB', other_designations='CG12128-PA|CG12128-PB|uncharacterized protein', map_location='46D4-46D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10026839, end_position_on_the_genomic_accession=10030668, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36043, CurrentID=0, Status='live', Symbol='CG1407', Aliases='Dmel_ Dmel\\CG1407', description='CG1407 gene product from transcript CG1407-RB', other_designations='CG1407-PA|CG1407-PB|CG1407-PC|CG1407-PD|CG1407-PE|CG1407-PF|CG1407-PG', map_location='46D5-46D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10028443, end_position_on_the_genomic_accession=10032893, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36044, CurrentID=0, Status='live', Symbol='CG12129', Aliases='Dmel_ Dmel\\CG12129, cg12129', description='CG12129 gene product from transcript CG12129-RA', other_designations='CG12129-PA|uncharacterized protein', map_location='46D7-46D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10033304, end_position_on_the_genomic_accession=10034698, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36045, CurrentID=0, Status='live', Symbol='oys', Aliases='Dmel_CG18445, CG18445, Dmel\\CG18445, Oys', description='oysgedart', other_designations='CG18445-PA|CG18445-PB|CG18445-PD|CG18445-PE|O-acyl transferase|oys-PA|oys-PB|oys-PD|oys-PE', map_location='46D7-46D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10034595, end_position_on_the_genomic_accession=10048619, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36046, CurrentID=0, Status='live', Symbol='sel', Aliases='Dmel_CG12918, CG12918, Dmel\\CG12918, Sle', description='seele', other_designations='CG12918-PA|sel-PA', map_location='46D9-46D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10050862, end_position_on_the_genomic_accession=10052167, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36047, CurrentID=0, Status='live', Symbol='Def', Aliases='Dmel_CG1385, 143607_at, CG1385, DEF, DIM 26, Dmel\\CG1385, def', description='Defensin', other_designations='CG1385-PA|Def-PA|defensin|defensin precursor', map_location='46D9-46D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10054178, end_position_on_the_genomic_accession=10054576, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36048, CurrentID=0, Status='live', Symbol='magu', Aliases='Dmel_CG2264, CG2264, CT7504, Dmel\\CG2264, PENT, pent', description='CG2264 gene product from transcript CG2264-RA', other_designations='CG2264-PA|CG2264-PB|CG2264-PD|CG2264-PF|magu-PA|magu-PB|magu-PD|magu-PF|pentagone', map_location='46D9-46E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10052541, end_position_on_the_genomic_accession=10063712, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  36049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36049, CurrentID=0, Status='live', Symbol='CG2292', Aliases='Dmel_ Dmel\\CG2292', description='CG2292 gene product from transcript CG2292-RA', other_designations='CG2292-PA', map_location='46E1-46E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10063385, end_position_on_the_genomic_accession=10066514, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36050, CurrentID=0, Status='live', Symbol='mRpL42', Aliases='Dmel_CG12921, BcDNA:RE72116, CG12921, Dmel\\CG12921, MRP-S32, MRPL31, mRpS32', description='mitochondrial ribosomal protein L42', other_designations='CG12921-PA|CG12921-PB|mRpL42-PA|mRpL42-PB|mitochondrial ribosomal protein S32', map_location='46E1-46E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10066650, end_position_on_the_genomic_accession=10067242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36051, CurrentID=0, Status='live', Symbol='Cdc2rk', Aliases='Dmel_CG1362, CG1362, Dcdrk, Dmel\\CG1362, cdc2rk', description='Cdc2-related kinase', other_designations='CG1362-PA|Cdc2rk-PA|cdc2-related-kinase', map_location='46E1-46E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10067217, end_position_on_the_genomic_accession=10068702, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36052, CurrentID=0, Status='live', Symbol='CG12920', Aliases='Dmel_ Dmel\\CG12920', description='CG12920 gene product from transcript CG12920-RA', other_designations='CG12920-PA|uncharacterized protein', map_location='46E1-46E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10069106, end_position_on_the_genomic_accession=10071067, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36053, CurrentID=0, Status='live', Symbol='CG1371', Aliases='Dmel_ Dmel\\CG1371', description='CG1371 gene product from transcript CG1371-RA', other_designations='CG1371-PA', map_location='46E1-46E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10070981, end_position_on_the_genomic_accession=10075645, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36054, CurrentID=0, Status='live', Symbol='egr', Aliases='Dmel_CG12919, BcDNA:RH51659, CG12919, Dmel\\CG12919, Ect1, Egr, Regg1, TNF, TNF-alpha, darth, dt1, regg1', description='eiger', other_designations='CG12919-PA|CG12919-PB|TNF-alpha|drosophila aspartate rich TNF homolog|egr-PA|egr-PB|tumor necrosis factor family member DT1|wengen', map_location='46E1-46E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10078366, end_position_on_the_genomic_accession=10084210, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36055, CurrentID=0, Status='live', Symbol='sut4', Aliases='Dmel_CG1380, CG1380, Dmel\\CG1380', description='sugar transporter 4', other_designations='CG1380-PA|sut4-PA', map_location='46E3-46E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10085442, end_position_on_the_genomic_accession=10087127, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36056, CurrentID=0, Status='live', Symbol='CG2269', Aliases='Dmel_ Dmel\\CG2269, anon-WO0118547.141', description='CG2269 gene product from transcript CG2269-RD', other_designations='CG2269-PA|CG2269-PB|CG2269-PC|CG2269-PD', map_location='46E4-46E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10085169, end_position_on_the_genomic_accession=10096416, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36057, CurrentID=0, Status='live', Symbol='Jra', Aliases='Dmel_CG2275, AP-1, AP1, CG2275, D-Jun, D-jun, D-jun/Jra, DJUN, DJun, DM14, Djun, Dmel\\CG2275, JRA, JUN, Jun, V, c-Jun, cJun, d-JRA, d-Jun, d-jun, dAP-1, dJRA, dJUN, dJra, dJun, djun, dm-Jun, jra, jun, l(2)46Ef, l(2)IA109, l(2R)IA109', description='Jun-related antigen', other_designations='CG2275-PA|CG2275-PB|CG2275-PC|Jra-PA|Jra-PB|Jra-PC|Jun oncogene|Jun related antigen|activator protein 1|complementation group V|lethal(2)1A109|lethal(2)IA109', map_location='2-60 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10096480, end_position_on_the_genomic_accession=10098556, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36058, CurrentID=0, Status='live', Symbol='RpLP0-like', Aliases='Dmel_CG1381, CG1381, Dmel\\CG1381, FBgn0033485', description='Ribosomal protein LP0-like', other_designations='CG1381-PA|RpLP0-like-PA|ribosomal protein LP0-like', map_location='46E5-46E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10098518, end_position_on_the_genomic_accession=10099590, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36059, CurrentID=0, Status='live', Symbol='14-3-3zeta', Aliases='Dmel_CG17870, 14-3-3, 14-3-3 zeta, 14-3-3-Leo, 14-3-3-zeta, 14-3-3ZETA, 14-3-3leo, 2G1, 4-3-3 zeta, 5.11, 549, BEST:GH05075, CG17870, D14-3-3, D Dmel\\CG17870, K, LEO, Leo, PAR-5, PAR5, Par-5, d14-3-3zeta, l(2)07103, l(2)46CFe, l(2)46Ee, leo, par-5', description='CG17870 gene product from transcript CG17870-RE', other_designations='14-3-3zeta-PA|14-3-3zeta-PB|14-3-3zeta-PC|14-3-3zeta-PD|14-3-3zeta-PE|14-3-3zeta-PF|14-3-3zeta-PG|14-3-3zeta-PH|14-3-3zeta-PI|14-3-3zeta-PJ|14-3-3zeta-PK|14-3-3zeta-PL|CG17870-PA|CG17870-PB|CG17870-PC|CG17870-PD|CG17870-PE|CG17870-PF|CG17870-PG|CG17870-PH|CG17870-PI|CG17870-PJ|CG17870-PK|CG17870-PL|D14 3 3 protein|Leonardo-13-3-3|complementation group K|leonardo|leonardo 14-3-3', map_location='46E6-46E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10099677, end_position_on_the_genomic_accession=10109687, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36060, CurrentID=0, Status='live', Symbol='Pfk', Aliases='Dmel_CG4001, BcDNA.GH12192, BcDNA:GH12192, CG4001, Dmel\\CG4001, PFK, PFK1, l(2)06339, pfk', description='Phosphofructokinase', other_designations='6-phosphofructokinase|CG4001-PA|CG4001-PB|CG4001-PC|CG4001-PD|CG4001-PE|Pfk-PA|Pfk-PB|Pfk-PC|Pfk-PD|Pfk-PE|phosphofructokinase|phosphofructokinase 1', map_location='2-85.2 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10109740, end_position_on_the_genomic_accession=10117457, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36061, CurrentID=0, Status='live', Symbol='dmpd', Aliases='Dmel_CG11866, CG11866, Dmel\\CG11866, FBXO13', description='dampened', other_designations='CG11866-PA|CG11866-PB|dmpd-PA|dmpd-PB', map_location='46F1-46F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10117694, end_position_on_the_genomic_accession=10121451, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36064, CurrentID=0, Status='live', Symbol='gem', Aliases='Dmel_CG30011, BcDNA:HL07889, CG11867, CG30011, CG3459, Dmel\\CG30011, anon-AE003830.1, dCP2', description='gemini', other_designations='CG30011-PA|CG30011-PC|CG30011-PF|CG30011-PG|gem-PA|gem-PC|gem-PF|gem-PG|persephone', map_location='46F1-46F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10124293, end_position_on_the_genomic_accession=10137472, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  36065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36065, CurrentID=0, Status='live', Symbol='CG12917', Aliases='Dmel_ Dmel\\CG12917', description='CG12917 gene product from transcript CG12917-RB', other_designations='CG12917-PB|uncharacterized protein', map_location='46F2-46F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10137656, end_position_on_the_genomic_accession=10139267, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740411, CurrentID=0, Status='live', Symbol='CG41562', Aliases='Dmel_ Dmel\\CG41562', description='CG41562 gene product from transcript CG41562-RA', other_designations='CG41562-PA|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22251626, end_position_on_the_genomic_accession=22252340, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36068, CurrentID=0, Status='live', Symbol='CG18011', Aliases='Dmel_ CG17847, Dmel\\CG18011', description='CG18011 gene product from transcript CG18011-RA', other_designations='CG18011-PA|uncharacterized protein', map_location='46F2-46F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10142338, end_position_on_the_genomic_accession=10146206, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36071, CurrentID=0, Status='live', Symbol='KCNQ', Aliases='Dmel_CG33135, CG12215, CG12915, CG33135, DKCNQ, Dmel\\CG33135, KQT, KQT 1, dKCNQ, kcnq1', description='KCNQ potassium channel', other_designations='CG33135-PC|CG33135-PD|CG33135-PE|CG33135-PF|CG33135-PG|CG33135-PH|CG33135-PI|CG33135-PJ|CG33135-PK|KCNQ-PC|KCNQ-PD|KCNQ-PE|KCNQ-PF|KCNQ-PG|KCNQ-PH|KCNQ-PI|KCNQ-PJ|KCNQ-PK|KCNQ-type K[+] channel', map_location='46F3-46F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10145897, end_position_on_the_genomic_accession=10192394, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  36072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36072, CurrentID=0, Status='live', Symbol='mlt', Aliases='Dmel_CG12214, CG12214, Dmel\\CG12214, TBCE-like, TBCEL, Tbcel, ms(2)3, ms(2)4210, ms(2)ry-3, ms(2)ry3', description='mulet', other_designations='CG12214-PA|CG12214-PB|CG12214-PC|mlt-PA|mlt-PB|mlt-PC|tubulin-specific chaperone E-like protein', map_location='46F3-46F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10155916, end_position_on_the_genomic_accession=10158931, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36073, CurrentID=0, Status='live', Symbol='Hr46', Aliases='Dmel_CG33183, 12.6, 2.4, CG11823, CG12208, CG33183, D, DHR3, Dhr3, DmRORalpha, Dmel\\CG33183, HR3, Hr3, NR1F4, dHR3, hr46, l(2)46CFi, l(2)k09242', description='Hormone receptor-like in 46', other_designations='CG33183-PA|CG33183-PC|CG33183-PD|CG33183-PE|CG33183-PF|CG33183-PG|CG33183-PH|Hr46-PA|Hr46-PC|Hr46-PD|Hr46-PE|Hr46-PF|Hr46-PG|Hr46-PH|complementation group D|hormone receptor 3|hormone receptor-like 46|hormone receptor-like in 46', map_location='46F5-46F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10203995, end_position_on_the_genomic_accession=10237348, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  36075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36075, CurrentID=0, Status='live', Symbol='CG12912', Aliases='Dmel_ Dmel\\CG12912', description='CG12912 gene product from transcript CG12912-RB', other_designations='CG12912-PA|CG12912-PB|uncharacterized protein', map_location='46F7-46F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10205454, end_position_on_the_genomic_accession=10234607, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36076, CurrentID=0, Status='live', Symbol='Hdc', Aliases='Dmel_CG3454, CG3454, Dmel\\CG3454, JK910, anon-WO0140519.259, hdc', description='Histidine decarboxylase', other_designations='CG3454-PA|CG3454-PB|CG3454-PC|CG3454-PD|Hdc-PA|Hdc-PB|Hdc-PC|Hdc-PD|histidine decarboxylase|histidine-decarboxylase', map_location='2-58.4 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10237988, end_position_on_the_genomic_accession=10246077, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36077, CurrentID=0, Status='live', Symbol='CG12209', Aliases='Dmel_ BcDNA:AT13439, Dmel\\CG12209', description='CG12209 gene product from transcript CG12209-RA', other_designations='CG12209-PA|uncharacterized protein', map_location='46F7-46F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10246071, end_position_on_the_genomic_accession=10247560, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36078, CurrentID=0, Status='live', Symbol='CG12914', Aliases='Dmel_ Dmel\\CG12914', description='CG12914 gene product from transcript CG12914-RA', other_designations='CG12914-PA|CG12914-PB|CG12914-PC|uncharacterized protein', map_location='46F7-46F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10247814, end_position_on_the_genomic_accession=10249404, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36079, CurrentID=0, Status='live', Symbol='CG12913', Aliases='Dmel_ Dmel\\CG12913, dCSGalNAcT1', description='CG12913 gene product from transcript CG12913-RB', other_designations='CG12913-PA|CG12913-PB|uncharacterized protein', map_location='46F8-46F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10249220, end_position_on_the_genomic_accession=10255495, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36080, CurrentID=0, Status='live', Symbol='Syb', Aliases='Dmel_CG12210, 21.15, CG12210, Dmel\\CG12210, END1-2, J, SYB, VAMP, VAMP2, Vamp, c-Syb, dVAMP, l(2)46CFq, l(2)46El, syb, synaptobrevin', description='Synaptobrevin', other_designations='CG12210-PA|CG12210-PB|CG12210-PC|CG12210-PD|CG12210-PE|Syb-PA|Syb-PB|Syb-PC|Syb-PD|Syb-PE|VAMP3|Vesicle-associated membrane protein|complementation group J|synaptobrevin', map_location='46F8-46F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10249435, end_position_on_the_genomic_accession=10252417, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36081, CurrentID=0, Status='live', Symbol='CG12911', Aliases='Dmel_ Dmel\\CG12911', description='CG12911 gene product from transcript CG12911-RB', other_designations='CG12911-PA|CG12911-PB|CG12911-PC|CG12911-PD|CG12911-PE|CG12911-PF|uncharacterized protein', map_location='46F9-46F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10256686, end_position_on_the_genomic_accession=10261276, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36082, CurrentID=0, Status='live', Symbol='CG12910', Aliases='Dmel_ Dmel\\CG12910', description='CG12910 gene product from transcript CG12910-RB', other_designations='CG12910-PA|CG12910-PB|CG12910-PC|uncharacterized protein', map_location='46F9-46F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10265057, end_position_on_the_genomic_accession=10269205, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36084, CurrentID=0, Status='live', Symbol='CAP', Aliases='Dmel_CG18408/Vinexin, CG18408, CG18409, CG3451, DCAP, Dmel\\CG18408, Rexin, cap, rexin', description='CG18408 gene product from transcript CG18408-RI', other_designations='CAP-PA|CAP-PB|CAP-PC|CAP-PD|CAP-PE|CAP-PF|CAP-PG|CAP-PH|CAP-PI|CAP-PJ|CAP-PK|CAP-PL|CAP-PN|CAP-PO|CAP-PP|CAP-PQ|CAP-PR|CAP-PS|CAP-PU|CAP-PV|CAP-PW|CAP-PX|CAP-PY|CG18408-PA|CG18408-PB|CG18408-PC|CG18408-PD|CG18408-PE|CG18408-PF|CG18408-PG|CG18408-PH|CG18408-PI|CG18408-PJ|CG18408-PK|CG18408-PL|CG18408-PN|CG18408-PO|CG18408-PP|CG18408-PQ|CG18408-PR|CG18408-PS|CG18408-PU|CG18408-PV|CG18408-PW|CG18408-PX|CG18408-PY|Rexin/CAP|cap', map_location='46F9-47A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10269016, end_position_on_the_genomic_accession=10307024, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  38782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38782, CurrentID=0, Status='live', Symbol='CG15829', Aliases='Dmel_ Dmel\\CG15829', description='CG15829 gene product from transcript CG15829-RB', other_designations='CG15829-PA|CG15829-PB|uncharacterized protein', map_location='65E4-65E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7128903, end_position_on_the_genomic_accession=7129673, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36086, CurrentID=0, Status='live', Symbol='JhI-1', Aliases='Dmel_CG3298, CG3298, Dmel\\CG3298, JHI-1, RNZ, dRNase Z, dRNaseZ, dtRNaseZ, tRNAse Z, tRNase Z', description='Juvenile hormone-inducible protein 1', other_designations='CG3298-PB|JhI-1-PB|juvenile hormone-inducible protein 1|tRNase Z', map_location='47A1-47A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10307030, end_position_on_the_genomic_accession=10309794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36087, CurrentID=0, Status='live', Symbol='CG12909', Aliases='Dmel_ Dmel\\CG12909', description='CG12909 gene product from transcript CG12909-RA', other_designations='CG12909-PA|uncharacterized protein', map_location='47A1-47A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10310113, end_position_on_the_genomic_accession=10311143, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36088, CurrentID=0, Status='live', Symbol='Obp46a', Aliases='Dmel_CG12905, CG12905, DmelObp46a, Dmel\\CG12905', description='Odorant-binding protein 46a', other_designations='CG12905-PA|Obp46a-PA', map_location='47A1-47A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10311172, end_position_on_the_genomic_accession=10312023, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36089, CurrentID=0, Status='live', Symbol='Ndg', Aliases='Dmel_CG12908, BEST:CK00552, CG12908, CT32053, Dmel\\CG12908, ndg', description='Nidogen/entactin', other_designations='CG12908-PA|CG12908-PC|Ndg-PA|Ndg-PC|Nidogen-like|nidogen|nidogen/entactin', map_location='47A1-47A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10313891, end_position_on_the_genomic_accession=10322104, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36091, CurrentID=0, Status='live', Symbol='CG12907', Aliases='Dmel_ BcDNA:AT31822, Dmel\\CG12907, anon-SAGE:Wang-109, anon-WO0140519.79', description='CG12907 gene product from transcript CG12907-RB', other_designations='CG12907-PA|CG12907-PB|anon-SAGE:Wang-109|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10353371, end_position_on_the_genomic_accession=10354384, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36092, CurrentID=0, Status='live', Symbol='CG12902', Aliases='Dmel_ BcDNA:AT11401, Dmel\\CG12902', description='CG12902 gene product from transcript CG12902-RB', other_designations='CG12902-PA|CG12902-PB|CG12902-PC|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10354724, end_position_on_the_genomic_accession=10355831, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36093, CurrentID=0, Status='live', Symbol='CG42733', Aliases='Dmel_ CG12901, Dmel\\CG42733, Dmel_CG12901, Tes85-2', description='CG42733 gene product from transcript CG42733-RD', other_designations='CG42733-PB|CG42733-PC|CG42733-PD|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10356171, end_position_on_the_genomic_accession=10356800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36094, CurrentID=0, Status='live', Symbol='Gr47a', Aliases='Dmel_CG12906, 47A.1, 47A1, CG12906, Dmel\\CG12906, GR47A.1, Gr47A, Gr47A1', description='Gustatory receptor 47a', other_designations='CG12906-PB|Gr47a-PB|gustatory receptor 47A.1|gustatory receptor 47a', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10362185, end_position_on_the_genomic_accession=10363584, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36095, CurrentID=0, Status='live', Symbol='Ir47a', Aliases='Dmel_CG12900, CG12900, CG12900/CG34363, DmelIR47a, Dmel\\CG12900, IR47a', description='Ionotropic receptor 47a', other_designations='CG12900-PA|Ir47a-PA|ionotropic receptor 47a', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10377128, end_position_on_the_genomic_accession=10380109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36096, CurrentID=0, Status='live', Symbol='CG12898', Aliases='Dmel_ Dmel\\CG12898', description='CG12898 gene product from transcript CG12898-RA', other_designations='CG12898-PA|uncharacterized protein', map_location='47A6-47A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10406975, end_position_on_the_genomic_accession=10407445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740416, CurrentID=0, Status='live', Symbol='CG34245', Aliases='Dmel_ Dmel\\CG34245', description='CG34245 gene product from transcript CG34245-RD', other_designations='CG34245-PD|uncharacterized protein', map_location='71A2-71A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14893734, end_position_on_the_genomic_accession=14894416, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36098, CurrentID=0, Status='live', Symbol='Prx2540-2', Aliases='Dmel_CG11765, CG11765, DPx-2540-2, DPx2540, Dmel\\CG11765, Prx2540', description='Peroxiredoxin 2540-2', other_designations='CG11765-PA|Prx2540-2-PA|peroxiredoxin|peroxiredoxin 2540|peroxiredoxin 2540-2', map_location='47A7-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10418718, end_position_on_the_genomic_accession=10419761, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36099, CurrentID=0, Status='live', Symbol='CG11825', Aliases='Dmel_ BcDNA:RE25483, Dmel\\CG11825, dHIG1', description='CG11825 gene product from transcript CG11825-RC', other_designations='CG11825-PA|CG11825-PC|CG11825-PD|uncharacterized protein', map_location='47A7-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10420414, end_position_on_the_genomic_accession=10423650, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36101, CurrentID=0, Status='live', Symbol='CG12896', Aliases='Dmel_ BEST:LD11278, DPx-2540-1 variant, Dmel\\CG12896, RE05501p, RH54193', description='CG12896 gene product from transcript CG12896-RA', other_designations='AOP2-related|CG12896-PA|uncharacterized protein', map_location='47A7-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10426880, end_position_on_the_genomic_accession=10427915, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36102, CurrentID=0, Status='live', Symbol='RanBPM', Aliases='Dmel_CG42236, CG11763, CG42236, Dmel\\CG42236, Dmel_CG11763, l(2)k05201', description='Ran-binding protein M', other_designations='CG42236-PA|CG42236-PC|CG42236-PD|CG42236-PE|CG42236-PF|CG42236-PG|Ran binding protein M|Ran-Binding protein in the microtubule organizing center|Ran-binding proteins 9/10 homologue|RanBPM-PA|RanBPM-PC|RanBPM-PD|RanBPM-PE|RanBPM-PF|RanBPM-PG|lethal (2) k05201', map_location='47A7-47A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10429708, end_position_on_the_genomic_accession=10436581, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  36103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36103, CurrentID=0, Status='live', Symbol='CG12895', Aliases='Dmel_ BcDNA:RE29446, Dmel\\CG12895', description='CG12895 gene product from transcript CG12895-RB', other_designations='CG12895-PA|CG12895-PB|uncharacterized protein', map_location='47A7-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10436598, end_position_on_the_genomic_accession=10437382, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36104, CurrentID=0, Status='live', Symbol='Galphao', Aliases='Dmel_CG2204, Bkh, CG2204, DG[[o]], DGalpha0, DGalpha[[o]], DROGPAD, Dmel\\CG2204, G(o)alpha47A, G-alpha-47A, G-oalpha, G-oalpha47A, G-oalpha47a, G[[o]], G[[o]]alpha, G[[oalpha]], Galpha47A, Galpha[[047A]], Galpha[[0]], Galpha[[o]], Galpha[[o]]47A47A, Go, Go-alpha, Goa, Goalpha, Goalpha47a, alpha(o), bkh, dgo, dgoalpha, drogpad, goalpha, Galphao', description='G protein alpha o subunit', other_designations='CG2204-PA|CG2204-PB|CG2204-PC|CG2204-PD|CG2204-PE|CG2204-PF|CG2204-PG|CG2204-PH|CG2204-PI|CG2204-PJ|G protein oalpha 47A|Galphao-PA|Galphao-PB|Galphao-PC|Galphao-PD|Galphao-PE|Galphao-PF|Galphao-PG|Galphao-PH|Galphao-PI|Galphao-PJ|brokenheart|brokenhearted|drogpad', map_location='47A7-47A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10437985, end_position_on_the_genomic_accession=10465701, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36105, CurrentID=0, Status='live', Symbol='Cyp49a1', Aliases='Dmel_CG18377, 49a1, CG12894, CG18377, Cyp49A1p, Dmel\\CG18377', description='CG18377 gene product from transcript CG18377-RA', other_designations='CG18377-PA|CG18377-PC|CG18377-PD|CG18377-PE|Cyp49a1-PA|Cyp49a1-PC|Cyp49a1-PD|Cyp49a1-PE|cytochrome P450', map_location='47A7-47A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10449350, end_position_on_the_genomic_accession=10458892, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  36107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36107, CurrentID=0, Status='live', Symbol='Caf1-105', Aliases='Dmel_CG12892, CAF p105, CAF1-105, CAF1-p105, CAF1-p75, CAF1-p75/p105, CAF105/75, CAF1p105, CG12892, Caf1-10, Dmel\\CG12892, caf1-105, dCAF-1-p105, dCAF1-p105, dCAF1-p75, p105, p105/p75, p75, p75/p105', description='Chromatin assembly factor 1, p105 subunit', other_designations='CG12892-PA|Caf1-105-PA|chromatin assembly factor 1, p105 subunit', map_location='47A9-47A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10466001, end_position_on_the_genomic_accession=10468540, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36108, CurrentID=0, Status='live', Symbol='CG11777', Aliases='Dmel_ BcDNA:RE61564, Dmel\\CG11777, cg11777', description='CG11777 gene product from transcript CG11777-RA', other_designations='CG11777-PA|uncharacterized protein', map_location='47A9-47A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10468616, end_position_on_the_genomic_accession=10469643, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36109, CurrentID=0, Status='live', Symbol='whd', Aliases='Dmel_CG12891, BEST:GH27794, CG12891, CPT1, CPTI, Cpt1, Dmel\\CG12891, gh27794', description='withered', other_designations='CG12891-PA|CG12891-PB|CG12891-PC|CG12891-PD|carnitin-palmitoyl transferase-I|carnitine palmitoyltransferase 1|carnitine-O-palmitoyltransferase|mitochondrial carnitine palamitoyltransferase I|mitochondrial carnitine palmitoyltransferase I|whd-PA|whd-PB|whd-PC|whd-PD', map_location='2-61 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10469473, end_position_on_the_genomic_accession=10478567, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36110, CurrentID=0, Status='live', Symbol='trsn', Aliases='Dmel_CG11761, CG11761, Dmel\\CG11761, TB-RBP', description='translin', other_designations='CG11761-PA|testis-brain RNA binding protein|trsn-PA', map_location='47A11-47A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10479011, end_position_on_the_genomic_accession=10480111, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36111, CurrentID=0, Status='live', Symbol='CG17765', Aliases='Dmel_ Dmel\\CG17765', description='CG17765 gene product from transcript CG17765-RA', other_designations='CG17765-PA|uncharacterized protein', map_location='47A11-47A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10480163, end_position_on_the_genomic_accession=10481886, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36118, CurrentID=0, Status='live', Symbol='psq', Aliases='Dmel_CG2368, BTB-V, BtbV, CG2368, Dmel\\CG2368, PSQ, Piq, Psq, eyeful, pip, psk-1, rum, zep, psq', description='pipsqueak', other_designations='BTB-protein-V|CG2368-PA|CG2368-PB|CG2368-PC|CG2368-PD|CG2368-PE|CG2368-PF|CG2368-PG|CG2368-PH|CG2368-PJ|CG2368-PK|CG2368-PM|CG2368-PN|eyeful|peepsqueak|pipsqeak|psq-PA|psq-PB|psq-PC|psq-PD|psq-PE|psq-PF|psq-PG|psq-PH|psq-PJ|psq-PK|psq-PM|psq-PN|rumpf|zeppelin|zepplin', map_location='47A13-47B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10557888, end_position_on_the_genomic_accession=10617280, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19893529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893529, CurrentID=0, Status='live', Symbol='ND2', Aliases='', description='NADH dehydrogenase subunit 2', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=240, end_position_on_the_genomic_accession=1263, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  36122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36122, CurrentID=0, Status='live', Symbol='Git', Aliases='Dmel_CG16728, CG16728, CT37209, Dmel\\CG16728, arfgap2, dGit, dgit, git', description='G protein-coupled receptor kinase interacting ArfGAP', other_designations='CG16728-PA|G-protein-coupled receptor kinase-interacting protein|Git-PA', map_location='47B1-47B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10641818, end_position_on_the_genomic_accession=10645427, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36123, CurrentID=0, Status='live', Symbol='Elp2', Aliases='Dmel_CG11887, CG1187, CG11887, D-elp2, Dmel\\CG11887, StIP, Stip, elp-2', description='Elongator complex protein 2', other_designations='CG11887-PA|Elp2-PA|StIP|elongator complex protein 2', map_location='47B1-47B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10645735, end_position_on_the_genomic_accession=10648634, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36124, CurrentID=0, Status='live', Symbol='CG12934', Aliases='Dmel_ Dmel\\CG12934', description='CG12934 gene product from transcript CG12934-RB', other_designations='CG12934-PA|CG12934-PB|uncharacterized protein', map_location='47B2-47B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10648888, end_position_on_the_genomic_accession=10649830, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36125, CurrentID=0, Status='live', Symbol='stan', Aliases='Dmel_CG11895, CG11895, CT20776, Dmel\\CG11895, Flam, Fmi, STAN, Stan, fmi, fmi/stan', description='starry night', other_designations='CG11895-PA|CG11895-PB|CG11895-PC|CG11895-PD|CG11895-PE|CG11895-PF|CG11895-PG|flamingo|serpentine cadherin|stan-PA|stan-PB|stan-PC|stan-PD|stan-PE|stan-PF|stan-PG', map_location='2-62 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10673345, end_position_on_the_genomic_accession=10721138, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36126, CurrentID=0, Status='live', Symbol='CG12309', Aliases='Dmel_ Dmel\\CG12309', description='CG12309 gene product from transcript CG12309-RA', other_designations='CG12309-PA|uncharacterized protein', map_location='47B7-47B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10725270, end_position_on_the_genomic_accession=10727272, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19893535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893535, CurrentID=0, Status='live', Symbol='COX2', Aliases='', description='cytochrome c oxidase subunit II', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=3083, end_position_on_the_genomic_accession=3767, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  36128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36128, CurrentID=0, Status='live', Symbol='CG12338', Aliases='Dmel_ Dmel\\CG12338', description='CG12338 gene product from transcript CG12338-RB', other_designations='CG12338-PA|CG12338-PB|uncharacterized protein', map_location='47B7-47B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10729578, end_position_on_the_genomic_accession=10731528, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36129, CurrentID=0, Status='live', Symbol='CG7220', Aliases='Dmel_ BEST:GH28095, BcDNA:GH28095, Dmel\\CG7220', description='CG7220 gene product from transcript CG7220-RA', other_designations='CG7220-PA|CG7220-PB|CG7220-PC|CG7220-PD|CG7220-PE|CG7220-PF', map_location='47B7-47B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10731335, end_position_on_the_genomic_accession=10738370, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36130, CurrentID=0, Status='live', Symbol='Mat1', Aliases='Dmel_CG7614, AT-1, CG7614, Dmel\\CG7614, MAT1', description='Menage a trois 1 ortholog (H. sapiens)', other_designations='CG7614-PA|Mat1-PA|menage a trois 1 ortholog', map_location='47B9-47B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10738617, end_position_on_the_genomic_accession=10739829, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19893539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893539, CurrentID=0, Status='live', Symbol='ATP6', Aliases='', description='ATP synthase F0 subunit 6', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=4062, end_position_on_the_genomic_accession=4736, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  19893540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893540, CurrentID=0, Status='live', Symbol='COX3', Aliases='', description='cytochrome c oxidase subunit III', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=4736, end_position_on_the_genomic_accession=5524, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  36133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36133, CurrentID=0, Status='live', Symbol='wde', Aliases='Dmel_CG12340, BcDNA:LD26050, CG12340, Dmel\\CG12340, LD26050, Wde, i10, i119', description='windei', other_designations='CG12340-PA|CG12340-PB|wde-PA|wde-PB', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10800420, end_position_on_the_genomic_accession=10808938, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19893542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893542, CurrentID=0, Status='live', Symbol='ND3', Aliases='', description='NADH dehydrogenase subunit 3', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=5608, end_position_on_the_genomic_accession=5961, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  36135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36135, CurrentID=0, Status='live', Symbol='CG7637', Aliases='Dmel_ Dmel\\CG7637', description='CG7637 gene product from transcript CG7637-RA', other_designations='CG7637-PA', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10810556, end_position_on_the_genomic_accession=10810936, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36136, CurrentID=0, Status='live', Symbol='mms4', Aliases='Dmel_CG12936, CG12936, Dmel\\CG12936, EME1', description='Methyl methanesulfonate sensitivity 4 ortholog (S. cerevisiae)', other_designations='CG12936-PA|methyl methanesulfonate sensitivity 4 ortholog|mms4-PA', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10811078, end_position_on_the_genomic_accession=10812215, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36137, CurrentID=0, Status='live', Symbol='CG12341', Aliases='Dmel_ BcDNA:SD17342, Dmel\\CG12341', description='CG12341 gene product from transcript CG12341-RA', other_designations='CG12341-PA|uncharacterized protein', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10812157, end_position_on_the_genomic_accession=10813310, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36138, CurrentID=0, Status='live', Symbol='CG7222', Aliases='Dmel_ Dmel\\CG7222', description='CG7222 gene product from transcript CG7222-RB', other_designations='CG7222-PA|CG7222-PB', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10813405, end_position_on_the_genomic_accession=10815781, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36139, CurrentID=0, Status='live', Symbol='dgo', Aliases='Dmel_CG12342, CG12342, Dgo, Dmel\\CG12342, EP(2)2619, EP2619', description='diego', other_designations='CG12342-PA|CG12342-PB|EP2619|dgo-PA|dgo-PB', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10816037, end_position_on_the_genomic_accession=10819884, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19893549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893549, CurrentID=0, Status='live', Symbol='ND5', Aliases='', description='NADH dehydrogenase subunit 5', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=6409, end_position_on_the_genomic_accession=8125, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  36142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36142, CurrentID=0, Status='live', Symbol='RpS15Ab', Aliases='Dmel_CG12324, CG12324, CT22283, Dmel\\CG12324, RpS15A', description='Ribosomal protein S15Ab', other_designations='CG12324-PA|RpS15Ab-PA|ribosomal protein S15Ab', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10821676, end_position_on_the_genomic_accession=10822399, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19893551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893551, CurrentID=0, Status='live', Symbol='ND4', Aliases='', description='NADH dehydrogenase subunit 4', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=8207, end_position_on_the_genomic_accession=9545, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  36144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36144, CurrentID=0, Status='live', Symbol='CG12325', Aliases='Dmel_ Dmel\\CG12325', description='CG12325 gene product from transcript CG12325-RB', other_designations='CG12325-PA|CG12325-PB|uncharacterized protein', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10823405, end_position_on_the_genomic_accession=10827644, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36145, CurrentID=0, Status='live', Symbol='CG12344', Aliases='Dmel_ CT23391, Dmel\\CG12344', description='CG12344 gene product from transcript CG12344-RD', other_designations='CG12344-PA|CG12344-PB|CG12344-PC|CG12344-PD|glycine receptor alpha2-like|uncharacterized protein', map_location='47C1-47C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10827430, end_position_on_the_genomic_accession=10835709, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36146, CurrentID=0, Status='live', Symbol='CG7686', Aliases='Dmel_ BcDNA:LD21529, Dmel\\CG7686', description='CG7686 gene product from transcript CG7686-RB', other_designations='CG7686-PA|CG7686-PB', map_location='47C3-47C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10836570, end_position_on_the_genomic_accession=10838727, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36147, CurrentID=0, Status='live', Symbol='CG30015', Aliases='Dmel_ CG7699, CG7701, Dmel\\CG30015, anon-EST:Posey2', description='CG30015 gene product from transcript CG30015-RA', other_designations='CG30015-PA|CG30015-PB|CG30015-PC|CG30015-PD|CG30015-PE|CG30015-PF|CG30015-PG|uncharacterized protein', map_location='47C3-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10839102, end_position_on_the_genomic_accession=10874771, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19893556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893556, CurrentID=0, Status='live', Symbol='CYTB', Aliases='', description='cytochrome b', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=10499, end_position_on_the_genomic_accession=11635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  19893558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893558, CurrentID=0, Status='live', Symbol='ND1', Aliases='', description='NADH dehydrogenase subunit 1', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=11721, end_position_on_the_genomic_accession=12659, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  36154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36154, CurrentID=0, Status='live', Symbol='CG18004', Aliases='Dmel_ Dmel\\CG18004', description='CG18004 gene product from transcript CG18004-RB', other_designations='CG18004-PA|CG18004-PB|CG18004-PD|uncharacterized protein', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10885019, end_position_on_the_genomic_accession=10886375, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36156, CurrentID=0, Status='live', Symbol='CG30020', Aliases='Dmel_ BEST:LD40262, CG12941, CG18353, Dmel\\CG30020, p190', description='CG30020 gene product from transcript CG30020-RA', other_designations='CG30020-PA|CG30020-PC|uncharacterized protein', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10886810, end_position_on_the_genomic_accession=10893191, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  36157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36157, CurrentID=0, Status='live', Symbol='cag', Aliases='Dmel_CG12346, CG12346, Dmel\\CG12346', description='CG12346 gene product from transcript CG12346-RA', other_designations='CG12346-PA|CG12346-PB|cag-PA|cag-PB', map_location='47C5-47C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10893034, end_position_on_the_genomic_accession=10894414, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36158, CurrentID=0, Status='live', Symbol='CG12942', Aliases='Dmel_ Dmel\\CG12942', description='CG12942 gene product from transcript CG12942-RA', other_designations='CG12942-PA|uncharacterized protein', map_location='47C6-47C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10894459, end_position_on_the_genomic_accession=10897222, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36159, CurrentID=0, Status='live', Symbol='ND-B14', Aliases='Dmel_CG7712, BcDNA:RE25411, CG7712, Dmel\\CG7712, NB4M', description='NADH dehydrogenase (ubiquinone) B14 subunit', other_designations='CG7712-PA|CG7712-PB|FBpp0087333|NADH ubiquinone oxidoreductase B14|NADH-ubiquinone oxidoreductase B14 subunit|NADH:ubiquinone oxidoreductase B14 subunit|ND-B14-PA|ND-B14-PB', map_location='47C6-47C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10897144, end_position_on_the_genomic_accession=10897885, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36160, CurrentID=0, Status='live', Symbol='Rpb5', Aliases='Dmel_CG11979, CG11979, DmRPB5, Dmel\\CG11979, RPB5', description='CG11979 gene product from transcript CG11979-RB', other_designations='CG11979-PA|CG11979-PB|Rpb5-PA|Rpb5-PB', map_location='47C6-47C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10898031, end_position_on_the_genomic_accession=10899060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36161, CurrentID=0, Status='live', Symbol='polyph', Aliases='Dmel_CG12943, CG12943, Dmel\\CG12943', description='polyphemus', other_designations='CG12943-PA|CG12943-PC|polyph-PA|polyph-PC', map_location='47C6-47C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10903570, end_position_on_the_genomic_accession=10905563, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36162, CurrentID=0, Status='live', Symbol='Obp47a', Aliases='Dmel_CG12944, 47a, CG12944, DmelObp47a, Dmel\\CG12944', description='Odorant-binding protein 47a', other_designations='CG12944-PA|CG12944-PB|Obp47a-PA|Obp47a-PB', map_location='47C6-47C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10909923, end_position_on_the_genomic_accession=10910485, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36163, CurrentID=0, Status='live', Symbol='Spn47C', Aliases='Dmel_CG7722, CG7722, Dmel\\CG7722, spn47C', description='Serpin 47C', other_designations='CG7722-PA|CG7722-PB|CG7722-PC|Spn47C-PA|Spn47C-PB|Spn47C-PC|serpin 47C', map_location='47C7-47C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10938425, end_position_on_the_genomic_accession=10941738, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36167, CurrentID=0, Status='live', Symbol='BBS4', Aliases='Dmel_CG13232, CG13232, Dmel\\CG13232', description='Bardet-Biedl syndrome 4 ortholog (H. sapiens)', other_designations='BBS4-PA|BBS4-PB|Bardet-Biedl syndrome 4 ortholog|CG13232-PA|CG13232-PB', map_location='47D5-47D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11135923, end_position_on_the_genomic_accession=11139315, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36168, CurrentID=0, Status='live', Symbol='CG13229', Aliases='Dmel_ Dm CG13229, Dmel\\CG13229, anon-WO0170980.142, anon-WO0170980.143', description='CG13229 gene product from transcript CG13229-RC', other_designations='CG13229-PA|CG13229-PB|CG13229-PC|uncharacterized protein', map_location='47D5-47D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11139304, end_position_on_the_genomic_accession=11172206, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36169, CurrentID=0, Status='live', Symbol='CG13231', Aliases='Dmel_ Dmel\\CG13231', description='CG13231 gene product from transcript CG13231-RA', other_designations='CG13231-PA|uncharacterized protein', map_location='47D5-47D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11146289, end_position_on_the_genomic_accession=11147128, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36170, CurrentID=0, Status='live', Symbol='CG12391', Aliases='Dmel_ Dmel\\CG12391, anon-WO0140519.201', description='CG12391 gene product from transcript CG12391-RA', other_designations='CG12391-PA|uncharacterized protein', map_location='47D5-47D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11147602, end_position_on_the_genomic_accession=11150068, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36171, CurrentID=0, Status='live', Symbol='shn', Aliases='Dmel_CG7734, BEST:SD06302, CG7734, Dmel\\CG7734, EP(2)2359, EP2359, SD06302, SHN, Shn, l(2)04738, quo', description='schnurri', other_designations='CG7734-PA|CG7734-PB|CG7734-PC|CG7734-PD|CG7734-PE|CG7734-PF|quo vadis|shn-PA|shn-PB|shn-PC|shn-PD|shn-PE|shn-PF|umnitza', map_location='2-62 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11159086, end_position_on_the_genomic_accession=11217394, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  3771724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771724, CurrentID=0, Status='live', Symbol='CG33680', Aliases='Dmel_ Dmel\\CG33680', description='CG33680 gene product from transcript CG33680-RA', other_designations='CG33680-PA|uncharacterized protein', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25249007, end_position_on_the_genomic_accession=25250316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36173, CurrentID=0, Status='live', Symbol='Syx7', Aliases='Dmel_CG5081, Aval, Avl, CG5081, Dm Syx7, DmSyx7, Dmel\\CG5081, FCP-C, SYX7, Syx, avl, braided, dSyn7, syx7, syx7/avl', description='Syntaxin 7', other_designations='CG5081-PA|CG5081-PB|Syntaxin7|Syx7-PA|Syx7-PB|avalanche|braided|dSyntaxin-7|synatxin7/avalanche|syntaxin|syntaxin 7', map_location='78D2-78D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21445834, end_position_on_the_genomic_accession=21448128, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36174, CurrentID=0, Status='live', Symbol='CG7737', Aliases='Dmel_ Dmel\\CG7737', description='CG7737 gene product from transcript CG7737-RA', other_designations='CG7737-PA|CG7737-PB', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11222871, end_position_on_the_genomic_accession=11226746, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36176, CurrentID=0, Status='live', Symbol='metro', Aliases='Dmel_CG30021, CG13219, CG18338, CG30021, Dmel\\CG30021, FBgn0050021, Metro, skf, skiff', description='menage a trois', other_designations='CG30021-PA|CG30021-PB|CG30021-PC|CG30021-PD|metro-PA|metro-PB|metro-PC|metro-PD|skiff', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11226458, end_position_on_the_genomic_accession=11232516, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3771729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771729, CurrentID=0, Status='live', Symbol='His3:CG33842', Aliases='Dmel_CG33842, CG33842, Dmel\\CG33842', description='CG33842 gene product from transcript CG33842-RA', other_designations='CG33842-PA|His3:CG33842-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21499920, end_position_on_the_genomic_accession=21500448, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771730, CurrentID=0, Status='live', Symbol='CG33666', Aliases='Dmel_ Dmel\\CG33666', description='CG33666 gene product from transcript CG33666-RB', other_designations='CG33666-PB|uncharacterized protein', map_location='6A2-6A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6476243, end_position_on_the_genomic_accession=6476967, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36179, CurrentID=0, Status='live', Symbol='CG13227', Aliases='Dmel_ Dmel\\CG13227, cg13227', description='CG13227 gene product from transcript CG13227-RA', other_designations='CG13227-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11235252, end_position_on_the_genomic_accession=11235755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771732, CurrentID=0, Status='live', Symbol='CG30461', Aliases='Dmel_ BcDNA:AT13932, CG33142, CR30461, Dmel\\CG30461', description='CG30461 gene product from transcript CG30461-RK', other_designations='CG30461-PI|CG30461-PJ|CG30461-PK|uncharacterized protein', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16995991, end_position_on_the_genomic_accession=16998594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740430, CurrentID=0, Status='live', Symbol='CG34299', Aliases='Dmel_ Dmel\\CG34299', description='CG34299 gene product from transcript CG34299-RA', other_designations='CG34299-PA|CG34299-PB|uncharacterized protein', map_location='99E3-99E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30249583, end_position_on_the_genomic_accession=30250661, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3771734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771734, CurrentID=0, Status='live', Symbol='His1:CG33807', Aliases='Dmel_CG33807, CG33807, Dmel\\CG33807', description='CG33807 gene product from transcript CG33807-RA', other_designations='CG33807-PA|His1:CG33807-PA|histone H1', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21431963, end_position_on_the_genomic_accession=21432985, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36183, CurrentID=0, Status='live', Symbol='Listericin', Aliases='Dmel_CG9080, 146991_at, BcDNA:RH62530, CG9080, Dmel\\CG9080', description='CG9080 gene product from transcript CG9080-RA', other_designations='CG9080-PA|Listericin-PA|listericin', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11241745, end_position_on_the_genomic_accession=11242194, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771736, CurrentID=0, Status='live', Symbol='His1:CG33861', Aliases='Dmel_CG33861, CG33861, Dmel\\CG33861', description='CG33861 gene product from transcript CG33861-RA', other_designations='CG33861-PA|His1:CG33861-PA|histone H1', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21531302, end_position_on_the_genomic_accession=21532324, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771738, CurrentID=0, Status='live', Symbol='CG9510', Aliases='Dmel_ BEST:GH06087, CG33085, Dmel\\CG9510, gh06087', description='CG9510 gene product from transcript CG9510-RD', other_designations='CG9510-PB|CG9510-PC|CG9510-PD', map_location='29F5-29F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8939725, end_position_on_the_genomic_accession=8942105, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36187, CurrentID=0, Status='live', Symbol='Or47a', Aliases='Dmel_CG13225, 47E.1, 47a, AN10, CG13225, DOR24, DOR47E.1, Dmel\\CG13225, OR47a, Or24, Or47E.1, dor24, or47a', description='Odorant receptor 47a', other_designations='CG13225-PA|Or47a-PA|odorant receptor 47a|olfactory receptor 47E.1', map_location='47E1-47E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11250368, end_position_on_the_genomic_accession=11251913, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36188, CurrentID=0, Status='live', Symbol='Cpr47Ea', Aliases='Dmel_CG9079, CG9079, DmelCpr47Ea, Dmel\\CG9079', description='Cuticular protein 47Ea', other_designations='CG9079-PA|Cpr47Ea-PA|cuticular protein 47Ea', map_location='47E3-47E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11251751, end_position_on_the_genomic_accession=11253398, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36189, CurrentID=0, Status='live', Symbol='Cpr47Eb', Aliases='Dmel_CG13224, CG13224, DmelCpr47Eb, Dmel\\CG13224', description='Cuticular protein 47Eb', other_designations='CG13224-PA|Cpr47Eb-PA|cuticular protein 47Eb', map_location='47E3-47E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11255167, end_position_on_the_genomic_accession=11255983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36190, CurrentID=0, Status='live', Symbol='CG13223', Aliases='Dmel_ Dmel\\CG13223', description='CG13223 gene product from transcript CG13223-RA', other_designations='CG13223-PA|uncharacterized protein', map_location='47E3-47E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11256507, end_position_on_the_genomic_accession=11258767, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36191, CurrentID=0, Status='live', Symbol='Cpr47Ec', Aliases='Dmel_CG9077, CG9077, Cpr47eg, DmelCpr47Ec, Dmel\\CG9077', description='Cuticular protein 47Ec', other_designations='CG9077-PA|Cpr47Ec-PA|cuticular protein 47Ec', map_location='47E3-47E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11259106, end_position_on_the_genomic_accession=11259976, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36192, CurrentID=0, Status='live', Symbol='Cpr47Ed', Aliases='Dmel_CG9076, CG9076, DmelCpr47Ed, Dmel\\CG9076', description='Cuticular protein 47Ed', other_designations='CG9076-PA|Cpr47Ed-PA|cuticular protein 47Ed', map_location='47E4-47E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11261079, end_position_on_the_genomic_accession=11261806, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36193, CurrentID=0, Status='live', Symbol='Cpr47Ee', Aliases='Dmel_CG13222, CG13222, DmelCpr47Ee, Dmel\\CG13222', description='Cuticular protein 47Ee', other_designations='CG13222-PA|Cpr47Ee-PA|cuticular protein 47Ee', map_location='47E4-47E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11265126, end_position_on_the_genomic_accession=11266856, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36194, CurrentID=0, Status='live', Symbol='Cpr47Ef', Aliases='Dmel_CG13214, CG13214, DmelCpr47Ef, Dmel\\CG13214', description='Cuticular protein 47Ef', other_designations='CG13214-PC|CG13214-PD|Cpr47Ef-PC|Cpr47Ef-PD|cuticular protein 47Ef', map_location='47E4-47E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11266915, end_position_on_the_genomic_accession=11272575, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  36195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36195, CurrentID=0, Status='live', Symbol='TpnC47D', Aliases='Dmel_CG9073, BcDNA:RE34610, CG9073, DmTnC2, DmTpnC47D, DmTpnCIb, Dmel\\CG9073, TNC, TNC47D, TnC, TnC47D, TpnC Ib, TpnCIb', description='Troponin C at 47D', other_designations='CG9073-PA|CG9073-PB|TpnC47D-PA|TpnC47D-PB|troponin C|troponin C at 47D', map_location='47E5-47E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11273845, end_position_on_the_genomic_accession=11275286, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36196, CurrentID=0, Status='live', Symbol='Cpr47Eg', Aliases='Dmel_CG9070, BcDNA:RE09259, CG9070, Cpr47eg, Dmel\\CG9070, Lcp2a', description='Cuticular protein 47Eg', other_designations='CG9070-PA|Cpr47Eg-PA|Lcp2a|cuticular protein 47Eg|larval cuticle protein 2a', map_location='47E5-47E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11277957, end_position_on_the_genomic_accession=11278445, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36197, CurrentID=0, Status='live', Symbol='CG9067', Aliases='Dmel_ Dmel\\CG9067, TRAPPC2L, Tca17', description='CG9067 gene product from transcript CG9067-RA', other_designations='CG9067-PA', map_location='47E5-47E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11278893, end_position_on_the_genomic_accession=11279683, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36199, CurrentID=0, Status='live', Symbol='CG9062', Aliases='Dmel_ Dmel\\CG9062', description='CG9062 gene product from transcript CG9062-RC', other_designations='CG9062-PB|CG9062-PC', map_location='47E5-47E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11280484, end_position_on_the_genomic_accession=11283971, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36200, CurrentID=0, Status='live', Symbol='CG13220', Aliases='Dmel_ Dmel\\CG13220', description='CG13220 gene product from transcript CG13220-RB', other_designations='CG13220-PA|CG13220-PB|uncharacterized protein', map_location='47E6-47E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11284283, end_position_on_the_genomic_accession=11285098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36201, CurrentID=0, Status='live', Symbol='Fbl6', Aliases='Dmel_CG13213, CG13213, Dmel\\CG13213, FBXL6, fbl6', description='F-box and leucine-rich repeat protein 6', other_designations='CG13213-PA|CG13213-PB|CG13213-PC|Fbl6-PA|Fbl6-PB|Fbl6-PC', map_location='47E6-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11285148, end_position_on_the_genomic_accession=11289772, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36202, CurrentID=0, Status='live', Symbol='CG18335', Aliases='Dmel_ Dmel\\CG18335', description='CG18335 gene product from transcript CG18335-RA', other_designations='CG18335-PA|uncharacterized protein', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11289688, end_position_on_the_genomic_accession=11291096, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36203, CurrentID=0, Status='live', Symbol='dare', Aliases='Dmel_CG12390, AR, CG12390, Dare, Dmel\\CG12390, W8', description='defective in the avoidance of repellents', other_designations='ADXRDTASE|CG12390-PA|adrenodoxin reductase|dare-PA', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11290994, end_position_on_the_genomic_accession=11293174, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3771756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771756, CurrentID=0, Status='live', Symbol='CG10097', Aliases='Dmel_ CG32919, CT28427, Dmel\\CG10097, anon-WO0140519.58', description='CG10097 gene product from transcript CG10097-RD', other_designations='CG10097-PB|CG10097-PC|CG10097-PD|uncharacterized protein', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12390166, end_position_on_the_genomic_accession=12395980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36205, CurrentID=0, Status='live', Symbol='CG42336', Aliases='Dmel_ BcDNA:AT13707, CG13210, CG13211, Dmel\\CG42336, Dmel_CG13210, Dmel_CG13211', description='CG42336 gene product from transcript CG42336-RB', other_designations='CG42336-PB|CG42336-PC|CG42336-PD|CG42336-PE|CG42336-PF|uncharacterized protein', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11294536, end_position_on_the_genomic_accession=11297520, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36207, CurrentID=0, Status='live', Symbol='Obp47b', Aliases='Dmel_CG13208, CG13208, DmelObp47b, Dmel\\CG13208', description='Odorant-binding protein 47b', other_designations='CG13208-PA|Obp47b-PA', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11302247, end_position_on_the_genomic_accession=11302987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36208, CurrentID=0, Status='live', Symbol='CG7741', Aliases='Dmel_ Dmel\\CG7741', description='CG7741 gene product from transcript CG7741-RA', other_designations='CG7741-PA|CG7741-PB|CG7741-PC', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11301916, end_position_on_the_genomic_accession=11305607, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36209, CurrentID=0, Status='live', Symbol='Fpps', Aliases='Dmel_CG12389, CG12389, DmFPPS, DmFPPSase, DmFPS, Dmel\\CG12389, FPPS, FPS, fpps, fps, l(2)k06103, v(2)k03514', description='Farnesyl pyrophosphate synthase', other_designations='CG12389-PA|Fpps-PA|farnesyl diphosphase synthase|farnesyl diphosphate synthase|farnesyl pyrophosphate synthase', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11305511, end_position_on_the_genomic_accession=11307528, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36210, CurrentID=0, Status='live', Symbol='CG7745', Aliases='Dmel_ Dmel\\CG7745', description='CG7745 gene product from transcript CG7745-RA', other_designations='CG7745-PA', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11307879, end_position_on_the_genomic_accession=11309829, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36212, CurrentID=0, Status='live', Symbol='Or47b', Aliases='Dmel_CG13206, 47E.2, 47b, CG13206, DOR25, DOR47E.2, Dmel\\CG13206, OR 47b, OR47b, Or47E.2, or47b', description='Odorant receptor 47b', other_designations='CG13206-PA|Or47b-PA|odorant receptor 47b|olfactory receptor 47E.2', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11319710, end_position_on_the_genomic_accession=11321312, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36213, CurrentID=0, Status='live', Symbol='sha', Aliases='Dmel_CG13209, CG13209, Dmel\\CG13209, koj, lincRNA.286', description='shavenoid', other_designations='CG13209-PA|CG13209-PB|CG13209-PC|kojak|sha-PA|sha-PB|sha-PC|shavenoid/kojak', map_location='2-62 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11324314, end_position_on_the_genomic_accession=11336463, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  36214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36214, CurrentID=0, Status='live', Symbol='lambdaTry', Aliases='Dmel_CG12350, CG12350, Dmel\\CG12350, SP76', description='CG12350 gene product from transcript CG12350-RA', other_designations='CG12350-PA|lambdaTry-PA|lambdaTrypsin', map_location='47F3-47F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11336840, end_position_on_the_genomic_accession=11337791, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36215, CurrentID=0, Status='live', Symbol='kappaTry', Aliases='Dmel_CG12388, CG12388, Dmel\\CG12388, SPH109', description='CG12388 gene product from transcript CG12388-RC', other_designations='CG12388-PC|kappaTry-PC', map_location='47F3-47F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11337797, end_position_on_the_genomic_accession=11338733, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36216, CurrentID=0, Status='live', Symbol='zetaTry', Aliases='Dmel_CG12387, CG12387, DmZetaTry, Dmel\\CG12387, SP110', description='zetaTrypsin', other_designations='CG12387-PA|zetaTry-PA', map_location='47F3-47F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11339186, end_position_on_the_genomic_accession=11340132, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36217, CurrentID=0, Status='live', Symbol='etaTry', Aliases='Dmel_CG12386, 169C4S, CG12386, DmEtaTry, Dmel\\CG12386, ESTS:169C4S, SP57, anon-sts17', description='etaTrypsin', other_designations='CG12386-PA|etaTry-PA', map_location='47F3-47F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11340513, end_position_on_the_genomic_accession=11341423, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36218, CurrentID=0, Status='live', Symbol='thetaTry', Aliases='Dmel_CG12385, CG12385, DmThetaTry, Dmel\\CG12385, SP139, Try', description='thetaTrypsin', other_designations='CG12385-PA|thetaTry-PA', map_location='47F3-47F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11341714, end_position_on_the_genomic_accession=11342760, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771771, CurrentID=0, Status='live', Symbol='His3:CG33815', Aliases='Dmel_CG33815, CG33815, Dmel\\CG33815', description='CG33815 gene product from transcript CG33815-RA', other_designations='CG33815-PA|His3:CG33815-PA|histone H3', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21454388, end_position_on_the_genomic_accession=21454916, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36221, CurrentID=0, Status='live', Symbol='gammaTry', Aliases='Dmel_CG30028, BcDNA:RE40325, CG30028, DmGammaTry, Dmel\\CG30028, SPH180, Try, endo-I, gamma, gamma-Try', description='gammaTrypsin', other_designations='CG30028-PA|endoproteinase I|gammaTry-PA|gummaTrypsin', map_location='47F4-47F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11349061, end_position_on_the_genomic_accession=11349870, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771774, CurrentID=0, Status='live', Symbol='His2A:CG33847', Aliases='Dmel_CG33847, CG33847, Dmel\\CG33847', description='CG33847 gene product from transcript CG33847-RA', other_designations='CG33847-PA|His2A:CG33847-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21508566, end_position_on_the_genomic_accession=21509058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36223, CurrentID=0, Status='live', Symbol='iotaTry', Aliases='Dmel_CG7754, 143775_at, CG7754, Dmel\\CG7754, SP129', description='iotaTrypsin', other_designations='CG7754-PA|iotaTry-PA|trypsin iota', map_location='47F5-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11356720, end_position_on_the_genomic_accession=11357537, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36224, CurrentID=0, Status='live', Symbol='CG12384', Aliases='Dmel_ BcDNA:GH16986, BcDNA:RH17411, Dmel\\CG12384', description='CG12384 gene product from transcript CG12384-RA', other_designations='CG12384-PA|uncharacterized protein', map_location='47F5-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11357902, end_position_on_the_genomic_accession=11359176, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3771779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771779, CurrentID=0, Status='live', Symbol='CG18003', Aliases='Dmel_ CG30019, Dmel\\CG18003', description='CG18003 gene product from transcript CG18003-RB', other_designations='CG18003-PB|uncharacterized protein', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10883547, end_position_on_the_genomic_accession=10884936, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36229, CurrentID=0, Status='live', Symbol='CG13203', Aliases='Dmel_ Dmel\\CG13203, anon-EST:Posey175', description='CG13203 gene product from transcript CG13203-RC', other_designations='CG13203-PC|CG13203-PD|uncharacterized protein', map_location='47F7-47F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11373607, end_position_on_the_genomic_accession=11377109, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36230, CurrentID=0, Status='live', Symbol='Tsp47F', Aliases='Dmel_CG9033, CG9033, Dm.Tsp47F, Dmel\\CG9033', description='Tetraspanin 47F', other_designations='CG9033-PB|Tsp47F-PB|tetraspanin 47F', map_location='47F7-47F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11377193, end_position_on_the_genomic_accession=11379177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771783, CurrentID=0, Status='live', Symbol='His2A:CG33823', Aliases='Dmel_CG33823, CG33823, Dmel\\CG33823', description='CG33823 gene product from transcript CG33823-RA', other_designations='CG33823-PA|His2A:CG33823-PA|histone H2A', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21468098, end_position_on_the_genomic_accession=21468590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36232, CurrentID=0, Status='live', Symbol='Sod3', Aliases='Dmel_CG9027, CG13200, CG9027, CT25938, Dmel/CG9027, Dmel\\CG9027, SOD3, dSod3', description='Superoxide dismutase 3', other_designations='CG9027-PA|CG9027-PB|CG9027-PD|CG9027-PE|CG9027-PF|Sod3-PA|Sod3-PB|Sod3-PD|Sod3-PE|Sod3-PF|superoxide dismutase 3', map_location='47F7-47F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11380347, end_position_on_the_genomic_accession=11384114, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36233, CurrentID=0, Status='live', Symbol='CG30022', Aliases='Dmel_ BcDNA:RE56416, CG9026, Dmel\\CG30022', description='CG30022 gene product from transcript CG30022-RB', other_designations='CG30022-PB|uncharacterized protein', map_location='47F8-47F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11387160, end_position_on_the_genomic_accession=11388779, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36234, CurrentID=0, Status='live', Symbol='CG7759', Aliases='Dmel_ Dmel\\CG7759', description='CG7759 gene product from transcript CG7759-RA', other_designations='CG7759-PA', map_location='47F10-47F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11405597, end_position_on_the_genomic_accession=11408079, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36235, CurrentID=0, Status='live', Symbol='CG7763', Aliases='Dmel_ BEST:CK02422, CK02422A, CK02422a, Dmel\\CG7763', description='CG7763 gene product from transcript CG7763-RB', other_designations='CG7763-PA|CG7763-PB', map_location='47F11-47F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11414510, end_position_on_the_genomic_accession=11415462, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36236, CurrentID=0, Status='live', Symbol='Drip', Aliases='Dmel_CG9023, CG9023, DRIP, Dmel\\CG9023, I, anon-48Aa, drip', description='CG9023 gene product from transcript CG9023-RG', other_designations='CG9023-PA|CG9023-PB|CG9023-PC|CG9023-PD|CG9023-PE|CG9023-PF|CG9023-PG|Drip-PA|Drip-PB|Drip-PC|Drip-PD|Drip-PE|Drip-PF|Drip-PG|aquaporin|drip|drosophila integral protein|transcript group I', map_location='47F11-47F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11411203, end_position_on_the_genomic_accession=11429645, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36237, CurrentID=0, Status='live', Symbol='CG7777', Aliases='Dmel_ Aqp7777, Dmel\\CG7777', description='CG7777 gene product from transcript CG7777-RB', other_designations='CG7777-PA|CG7777-PB|CG7777-PC|CG7777-PD', map_location='47F12-47F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11436243, end_position_on_the_genomic_accession=11442216, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36238, CurrentID=0, Status='live', Symbol='E(Pc)', Aliases='Dmel_CG7776, CG7776, Dmel\\CG7776, E(PC), GH05739, GH14582, III, anon-48Ac, e(Pc), e(pc), l(2)28-28-12', description='Enhancer of Polycomb', other_designations='CG7776-PA|CG7776-PB|CG7776-PC|CG7776-PD|CG7776-PE|E(Pc)-PA|E(Pc)-PB|E(Pc)-PC|E(Pc)-PD|E(Pc)-PE|enhancer of polycomb|transcript group III', map_location='2-61.9 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11452043, end_position_on_the_genomic_accession=11465270, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36239, CurrentID=0, Status='live', Symbol='inv', Aliases='Dmel_CG17835, CG17835, Dmel\\CG17835, Engrailed/Invected, IV, Inv, er, in', description='invected', other_designations='CG17835-PB|CG17835-PD|CG17835-PE|engrailed-invected|engrailed-related|inv-PB|inv-PD|inv-PE|transcript group IV', map_location='47F15-47F17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11474465, end_position_on_the_genomic_accession=11509695, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36240, CurrentID=0, Status='live', Symbol='en', Aliases='Dmel_CG9015, 153867_at, Apa, CG9015, Dmel\\CG9015, EN, En, EnR, Eng, Engr, Engrailed, Engrailed/Invected, Es, V1, spa2, spt, en', description='engrailed', other_designations='CG9015-PA|CG9015-PB|ENGRAILED|apigmented abdomen|en-PA|en-PB|erased|invected|sparse SGPs 2|spermatheca|transcript group V', map_location='2-62.0 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11524006, end_position_on_the_genomic_accession=11528210, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36241, CurrentID=0, Status='live', Symbol='tou', Aliases='Dmel_CG10897, CG10897, Dm Tou, Dmel\\CG10897, EP(2)0622, EP0622, EP622, TIP5, VI, anon-48Ad, gene VI', description='toutatis', other_designations='CG10897-PA|CG10897-PB|CG10897-PE|CG10897-PF|CG10897-PG|tou-PA|tou-PB|tou-PE|tou-PF|tou-PG|transcript group VI', map_location='48A1-48A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11577485, end_position_on_the_genomic_accession=11616045, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  36242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36242, CurrentID=0, Status='live', Symbol='Egm', Aliases='Dmel_CG9006, Acyl CoA DH, CG9006, Dmel\\CG9006, egm, l(2)k14708', description='Enigma', other_designations='CG9006-PA|Egm-PA|enigma|l(2)k14708|lethal (2) k14708', map_location='48A3-48A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11616869, end_position_on_the_genomic_accession=11619023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36243, CurrentID=0, Status='live', Symbol='CG9005', Aliases='Dmel_ BcDNA:GH05710, Dmel\\CG9005', description='CG9005 gene product from transcript CG9005-RB', other_designations='CG9005-PA|CG9005-PB|CG9005-PC', map_location='48A3-48B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11619479, end_position_on_the_genomic_accession=11640941, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36244, CurrentID=0, Status='live', Symbol='CG9003', Aliases='Dmel_ Dmel\\CG9003, FBXL2', description='CG9003 gene product from transcript CG9003-RG', other_designations='CG9003-PA|CG9003-PB|CG9003-PC|CG9003-PD|CG9003-PE|CG9003-PF|CG9003-PG|Fbxl5 iron sensing', map_location='48B2-48B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11641477, end_position_on_the_genomic_accession=11651616, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  36245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36245, CurrentID=0, Status='live', Symbol='CG13198', Aliases='Dmel_ Dmel\\CG13198', description='CG13198 gene product from transcript CG13198-RA', other_designations='CG13198-PA|uncharacterized protein', map_location='48B5-48B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11654898, end_position_on_the_genomic_accession=11655762, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36246, CurrentID=0, Status='live', Symbol='Roc2', Aliases='Dmel_CG8998, BcDNA:RE61847, CG8998, Dmel\\CG8998, Roc, dRoc2, roc2', description='Regulator of cullins 2', other_designations='CG8998-PA|CG8998-PB|Roc2-PA|Roc2-PB|regulator of cullins 2', map_location='48B4-48C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11652756, end_position_on_the_genomic_accession=11679354, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740441, CurrentID=0, Status='live', Symbol='CG34161', Aliases='Dmel_ Dmel\\CG34161', description='CG34161 gene product from transcript CG34161-RA', other_designations='CG34161-PA|CG34161-PB|CG34161-PC|uncharacterized protein', map_location='32A1-32A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10580957, end_position_on_the_genomic_accession=10581702, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36248, CurrentID=0, Status='live', Symbol='Tret1-1', Aliases='Dmel_CG30035, CG30035, CG7797, CG7801, DmTret1-1, Dmel\\CG30035, Tre T1', description='Trehalose transporter 1-1', other_designations='CG30035-PA|CG30035-PB|Tret1-1-PA|Tret1-1-PB|trehalose transporter 1|trehalose transporter 1-1', map_location='48B6-48B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11665768, end_position_on_the_genomic_accession=11674782, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36249, CurrentID=0, Status='live', Symbol='Tret1-2', Aliases='Dmel_CG8234, CG8234, DmTret1-2, Dmel\\CG8234', description='Trehalose transporter 1-2', other_designations='CG8234-PA|Tret1-2-PA|trehalose transporter 1-2', map_location='48B8-48C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11675935, end_position_on_the_genomic_accession=11678014, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36250, CurrentID=0, Status='live', Symbol='CG13196', Aliases='Dmel_ Dmel\\CG13196', description='CG13196 gene product from transcript CG13196-RA', other_designations='CG13196-PA|uncharacterized protein', map_location='48C1-48C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11684165, end_position_on_the_genomic_accession=11686434, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36251, CurrentID=0, Status='live', Symbol='Buffy', Aliases='Dmel_CG8238, BUFFY, Bcl-2-48AE/Borg2, Buffy/DBorg-2, Buffy/dBorg-2, CG8238, Dbx, Dmel\\CG8238, Drob2, buffy, dBORG-2/BUFFY, dBorg-2, dBorg2, dbuffy', description='CG8238 gene product from transcript CG8238-RA', other_designations='48A-E Bcl-2-like|Buffy-PA|CG8238-PA|buffy', map_location='48C1-48C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11679492, end_position_on_the_genomic_accession=11688557, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36252, CurrentID=0, Status='live', Symbol='wal', Aliases='Dmel_CG8996, BcDNA.LD07532, BcDNA:GH09945, BcDNA:LD07532, CG8996, Dmel\\CG8996, EP2253, ETF, ETFA, Etf, P447, l(2)02516', description='walrus', other_designations='CG8996-PA|CG8996-PB|electron transfer flavoprotein|electron transfer flavoprotein alpha subunit|electron transfer flavoprotein alpha-subunit|wal-PA|wal-PB', map_location='2-59 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11688569, end_position_on_the_genomic_accession=11690863, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36253, CurrentID=0, Status='live', Symbol='CG13197', Aliases='Dmel_ BcDNA:RE27552, Dmel\\CG13197', description='CG13197 gene product from transcript CG13197-RB', other_designations='CG13197-PA|CG13197-PB|uncharacterized protein', map_location='48C2-48C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11691285, end_position_on_the_genomic_accession=11697785, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3771806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771806, CurrentID=0, Status='live', Symbol='mei-217', Aliases='Dmel_CG33935, CG33067, CG33935, Dmel\\CG33935', description='CG33935 gene product from transcript CG33935-RC', other_designations='CG33935-PC|mei-217-PC', map_location='15E5-15E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17136578, end_position_on_the_genomic_accession=17142990, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36255, CurrentID=0, Status='live', Symbol='pyr', Aliases='Dmel_CG13194, CG13194, DFGF8-2, DmPyr, Dmel\\CG13194, FGF8-like2, Pyr, fgf8-like', description='pyramus', other_designations='CG13194-PB|pyr-PB', map_location='48C2-48C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11712263, end_position_on_the_genomic_accession=11726345, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36256, CurrentID=0, Status='live', Symbol='CG13193', Aliases='Dmel_ Dmel\\CG13193', description='CG13193 gene product from transcript CG13193-RA', other_designations='CG13193-PA|uncharacterized protein', map_location='48C2-48C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11768453, end_position_on_the_genomic_accession=11769561, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36257, CurrentID=0, Status='live', Symbol='Ir48c', Aliases='Dmel_CG12444, CG12444, DmelIR48c, Dmel\\CG12444, IR48c, TANGO3, Tango3', description='Ionotropic receptor 48c', other_designations='CG12444-PB|Ir48c-PB|ionotropic receptor 48c|transport and golgi organization 3', map_location='48C3-48C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11787808, end_position_on_the_genomic_accession=11789493, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36258, CurrentID=0, Status='live', Symbol='ths', Aliases='Dmel_CG12443, CG12443, DFGF8-1, DmThs, Dmel\\CG12443, FGF8-like1, Neu4, Ths, fgf8-like', description='thisbe', other_designations='CG12443-PA|CG12443-PB|ths-PA|ths-PB', map_location='48C3-48C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11756817, end_position_on_the_genomic_accession=11814424, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36259, CurrentID=0, Status='live', Symbol='CG13192', Aliases='Dmel_ Dmel\\CG13192', description='CG13192 gene product from transcript CG13192-RA', other_designations='CG13192-PA|uncharacterized protein', map_location='48C4-48C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11825345, end_position_on_the_genomic_accession=11827097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36260, CurrentID=0, Status='live', Symbol='Sobp', Aliases='Dmel_CG8991, CG8991, Dmel\\CG8991, Sbp', description='Sine oculis-binding protein', other_designations='CG8991-PA|CG8991-PB|Sobp-PA|Sobp-PB|sine oculis-binding protein', map_location='48C4-48C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11827170, end_position_on_the_genomic_accession=11839825, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36262, CurrentID=0, Status='live', Symbol='S2P', Aliases='Dmel_CG8988, CG8988, Dmel\\CG8988, anon-WO0118547.346, dS2P, ds2p', description='site-2 protease', other_designations='CG8988-PA|S2P-PA', map_location='48C4-48C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11842103, end_position_on_the_genomic_accession=11844240, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36263, CurrentID=0, Status='live', Symbol='Sln', Aliases='Dmel_CG8271, CG8271, Dmel\\CG8271', description='Silnoon', other_designations='CG8271-PA|Sln-PA|silnoon', map_location='48C4-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11845324, end_position_on_the_genomic_accession=11852324, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36264, CurrentID=0, Status='live', Symbol='Mtor', Aliases='Dmel_CG8274, Bx34, CG8274, Dmel\\CG8274, MTOR, TPR, Tpr, l(2)k03905', description='Megator', other_designations='CG8274-PA|CG8274-PB|Mtor-PA|Mtor-PB|lethal (2) k03905|megator', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11852617, end_position_on_the_genomic_accession=11861202, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36265, CurrentID=0, Status='live', Symbol='TwdlBeta', Aliases='Dmel_CG8986, CG8986, CT25838, Dmel\\CG8986', description='CG8986 gene product from transcript CG8986-RA', other_designations='CG8986-PA|TwdlBeta-PA', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11862537, end_position_on_the_genomic_accession=11863252, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36266, CurrentID=0, Status='live', Symbol='Damm', Aliases='Dmel_CG18188, CG18188, DAMM/Daydream, Daydream, Dmel\\CG18188, damm, Damm', description='Death associated molecule related to Mch2 caspase', other_designations='CG18188-PA|CG18188-PB|Damm-PA|Damm-PB|daydream|death associated molecule related to Mch2|death associated molecule related to Mch2 caspase', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11864112, end_position_on_the_genomic_accession=11866383, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36268, CurrentID=0, Status='live', Symbol='CG13185', Aliases='Dmel_ CG18329, Dmel\\CG13185', description='CG13185 gene product from transcript CG13185-RD', other_designations='CG13185-PD|uncharacterized protein', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11867876, end_position_on_the_genomic_accession=11885585, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  36269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36269, CurrentID=0, Status='live', Symbol='cuff', Aliases='Dmel_CG13190, CG13190, Dmel\\CG13190', description='cutoff', other_designations='CG13190-PA|cuff-PA', map_location='2-61 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11886067, end_position_on_the_genomic_accession=11887603, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36270, CurrentID=0, Status='live', Symbol='ERp60', Aliases='Dmel_CG8983, CG8983, D- Dmel\\CG8983, Erp60, Ov6, anon-EST:fe3D9', description='CG8983 gene product from transcript CG8983-RA', other_designations='CG8983-PA|ERp60-PA|anon-fast-evolving-3D9|protein disulfide isomerase homolog', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11887568, end_position_on_the_genomic_accession=11890979, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36271, CurrentID=0, Status='live', Symbol='Ef1alpha48D', Aliases='Dmel_CG8280, 1275, CG8280, Dm0084, DmEF1, Dmel\\CG8280, EF, EF-1, EF-1[a], EF-1a, EF-1alpha, EF-1alpha F1, EF-1alphaF, EF-1alphaF1, EF-1alphaF[[1]], EF1-alpha, EF11_DROME, EF1alpha, EF1alpha 48D, EF1alpha48D, Ef1, Ef1a10, Ef1a48D, Ef1alpha1, Ef1alpha48E, Efalpha48D, F1, F1 EF-1alpha, anon-EST:Liang-1.17, anon-EST:Liang-2.46, chr2R:7409490..7409690, chr2R:7410009..7410129, clone 1.17, clone 2.46, eEF1alpha, ef-1alpha, l(2)01275, l(2)1275, l(2)k07303', description='Elongation factor 1alpha48D', other_designations='CG8280-PA|CG8280-PC|CG8280-PD|CG8280-PE|Ef1alpha48D-PA|Ef1alpha48D-PC|Ef1alpha48D-PD|Ef1alpha48D-PE|elongation factor 1 alpha|elongation factor 1 alpha 48D|elongation factor 1-alpha F1|elongation factor 1alpha48D|translational elongation factor1', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11891135, end_position_on_the_genomic_accession=11895762, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36272, CurrentID=0, Status='live', Symbol='MCPH1', Aliases='Dmel_CG42572, CG13184, CG30038, CG42572, CG8981, Dmcph1, Dmel\\CG42572, Dmel_CG30038, Dmel_CG8981, awol, mcph1', description='Microcephalin', other_designations='CG42572-PA|CG42572-PB|CG42572-PC|CG42572-PD|MCPH1-PA|MCPH1-PB|MCPH1-PC|MCPH1-PD|absent without leave|microcephalin', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11895997, end_position_on_the_genomic_accession=11903918, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36273, CurrentID=0, Status='live', Symbol='Zip48C', Aliases='Dmel_CG13189, CG13189, Dmel\\CG13189, cg13189, dZip48C', description='Zinc/iron regulated transporter-related protein 48C', other_designations='CG13189-PA|Zip48C-PA', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11899517, end_position_on_the_genomic_accession=11901394, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36275, CurrentID=0, Status='live', Symbol='reb', Aliases='Dmel_CG13183, CG13183, Dmel\\CG13183', description='rebuf', other_designations='CG13183-PA|reb-PA', map_location='48C5-48D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11904323, end_position_on_the_genomic_accession=11912779, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36276, CurrentID=0, Status='live', Symbol='exp', Aliases='Dmel_CG13188, CG13188, Dmel\\CG13188', description='expansion', other_designations='CG13188-PA|CG13188-PB|exp-PA|exp-PB', map_location='48D3-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11939557, end_position_on_the_genomic_accession=11967435, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36277, CurrentID=0, Status='live', Symbol='PI31', Aliases='Dmel_CG8979, 8979, CG8979, Dm Dmel\\CG8979, cg8979, dmPI31', description='CG8979 gene product from transcript CG8979-RD', other_designations='CG8979-PA|CG8979-PB|CG8979-PC|CG8979-PD|CG8979-PF|PI31-PA|PI31-PB|PI31-PC|PI31-PD|PI31-PF', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11967343, end_position_on_the_genomic_accession=11970115, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36278, CurrentID=0, Status='live', Symbol='ADD1', Aliases='Dmel_CG8290, BEST:LD29214, CG8290, Dmel\\CG8290, LD29214, dADD1', description='ADD domain-containing protein 1', other_designations='ADD1-PA|ADD1-PB|ADD1-PE|CG8290-PA|CG8290-PB|CG8290-PE', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11970686, end_position_on_the_genomic_accession=11977721, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36279, CurrentID=0, Status='live', Symbol='GalT1', Aliases='Dmel_CG33145, CG33145, CG8976, Dmel\\CG33145, dbeta3GnT', description='beta1,3-galactosyltransferase 1', other_designations='CG33145-PA|CG33145-PB|GalT1-PA|GalT1-PB', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11980577, end_position_on_the_genomic_accession=11982956, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36280, CurrentID=0, Status='live', Symbol='RnrS', Aliases='Dmel_CG8975, CG8975, DmRNR2, Dmel\\CG8975, RNR2, RNRS, RNrS, Rnr2, Rnrs, anon-WO0118547.148, dRNR2, rnr2', description='Ribonucleoside diphosphate reductase small subunit', other_designations='CG8975-PA|RnrS-PA|ribonucleoside diphosphate reductase small subunit|ribonucleoside-diphosphate reductase small subunit|ribonucleotide reductase|ribonucleotide reductase small subunit', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11983145, end_position_on_the_genomic_accession=11985393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36281, CurrentID=0, Status='live', Symbol='rho-7', Aliases='Dmel_CG8972, CG8972, Dmel\\CG8972, PARL, R7, Rho-7, Rhomboid-7, anon-WO0138581.3', description='rhomboid-7', other_designations='CG8972-PA|CG8972-PB|Rhomboid-7|rho-7-PA|rho-7-PB', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11986513, end_position_on_the_genomic_accession=11988787, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36282, CurrentID=0, Status='live', Symbol='CG8298', Aliases='Dmel_ Dmel\\CG8298', description='CG8298 gene product from transcript CG8298-RC', other_designations='CG8298-PA|CG8298-PC', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11989131, end_position_on_the_genomic_accession=11992350, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36283, CurrentID=0, Status='live', Symbol='otk', Aliases='Dmel_CG8967, CG8967, CT25769, DTrk, Dmel\\CG8967, Dtkr, Dtrk, IGGtyk, OFT/TRK, OTK, Otk, Tk48D, Trk48D, anon-WO2004063362.81, dTrk', description='off-track', other_designations='CG8967-PA|CG8967-PB|Off-track|off track|otk-PA|otk-PB|tyrosine kinase at 48D', map_location='48D6-48D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12001473, end_position_on_the_genomic_accession=12019846, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36284, CurrentID=0, Status='live', Symbol='otk2', Aliases='Dmel_CG8964, CG8964, CT25760, Dmel\\CG8964, l(2)SH1639, l(2)SH2 1639', description='off-track2', other_designations='CG8964-PA|lethal (2) SH1639|otk2-PA', map_location='48D7-48D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12023146, end_position_on_the_genomic_accession=12025272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36285, CurrentID=0, Status='live', Symbol='Mppe', Aliases='Dmel_CG8889, 11471, CG8889, Dmel\\CG8889, GS11471, dMPPE, dmppe', description='Metallophosphoesterase', other_designations='CG8889-PA|CG8889-PB|CG8889-PD|CG8889-PE|Mppe-PA|Mppe-PB|Mppe-PD|Mppe-PE|metallophosphoesterase', map_location='48D7-48D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12025614, end_position_on_the_genomic_accession=12027707, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36286, CurrentID=0, Status='live', Symbol='pds5', Aliases='Dmel_CG17509, CG17509, Dmel\\CG17509, PDS5, Pds5, anon-WO0118547.183, dPds5, l(2)k13312', description='CG17509 gene product from transcript CG17509-RB', other_designations='CG17509-PA|CG17509-PB|cohiba|lethal (2) k13312|pds5-PA|pds5-PB', map_location='48D7-48D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12027985, end_position_on_the_genomic_accession=12033725, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36287, CurrentID=0, Status='live', Symbol='CG8321', Aliases='Dmel_ Dmel\\CG8321', description='CG8321 gene product from transcript CG8321-RA', other_designations='CG8321-PA|CG8321-PB', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12033946, end_position_on_the_genomic_accession=12036260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36288, CurrentID=0, Status='live', Symbol='128up', Aliases='Dmel_CG8340, CG8340, Dmel\\CG8340, GTP-bp, Gtp-bp, X71866', description='upstream of RpIII128', other_designations='128up-PA|CG8340-PA', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12037298, end_position_on_the_genomic_accession=12038851, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36289, CurrentID=0, Status='live', Symbol='RpIII128', Aliases='Dmel_CG8344, CG8344, DmRP128, Dmel\\CG8344, RP128', description='RNA polymerase III 128kD subunit', other_designations='CG8344-PA|RNA polymerase III|RNA polymerase IIIb|RpIII128-PA', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12039041, end_position_on_the_genomic_accession=12042623, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36290, CurrentID=0, Status='live', Symbol='Drep1', Aliases='Dmel_CG8357, CG8357, DREP-1, DREP-1/dICAD, Dmel\\CG8357, Drep-1, ICAD, ICAD/DREP-1, Rep1, dICAD, drep1', description='DNA fragmentation factor-related protein 1', other_designations='CG8357-PA|CG8357-PB|DFF45-related protein-1|Drep1-PA|Drep1-PB|Rep1|inhibitor of CAD', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12044612, end_position_on_the_genomic_accession=12050172, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36292, CurrentID=0, Status='live', Symbol='Drep3', Aliases='Dmel_CG8364, BcDNA:AT08574, CG13187, CG8364, DREP-3, Dmel\\CG8364, Drep-3, Rep3, drep-3, drep3', description='DNA fragmentation factor-related protein 3', other_designations='CG8364-PA|CG8364-PB|Drep3-PA|Drep3-PB|Rep3', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12055089, end_position_on_the_genomic_accession=12062866, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36293, CurrentID=0, Status='live', Symbol='CG8888', Aliases='Dmel_ Dmel\\CG8888', description='CG8888 gene product from transcript CG8888-RA', other_designations='CG8888-PA', map_location='48D8-48E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12064565, end_position_on_the_genomic_accession=12079543, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36294, CurrentID=0, Status='live', Symbol='CG13186', Aliases='Dmel_ AAF58585, BcDNA:AT11581, Dmel\\CG13186', description='CG13186 gene product from transcript CG13186-RA', other_designations='CG13186-PA|uncharacterized protein', map_location='48E1-48E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12081276, end_position_on_the_genomic_accession=12082057, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36295, CurrentID=0, Status='live', Symbol='jeb', Aliases='Dmel_CG30040, CG13180, CG13182, CG30040, Dmel\\CG30040, Jeb, l(2)SH0422, l(2)k05644', description='jelly belly', other_designations='CG30040-PA|CG30040-PB|jeb-PA|jeb-PB|jelly-belly|jellybelly|lethal (2) SH0422', map_location='48E1-48E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12085194, end_position_on_the_genomic_accession=12119340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36297, CurrentID=0, Status='live', Symbol='CG18343', Aliases='Dmel_ Dmel\\CG18343', description='CG18343 gene product from transcript CG18343-RB', other_designations='CG18343-PA|CG18343-PB|uncharacterized protein', map_location='48E2-48E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12138695, end_position_on_the_genomic_accession=12139324, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36298, CurrentID=0, Status='live', Symbol='SkpB', Aliases='Dmel_CG8881, CG8881, Dmel\\CG8881, Skp, dskpB, skpB', description='SKP1-related B', other_designations='CG8881-PA|SkpB-PA|skpB', map_location='48E2-48E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12139295, end_position_on_the_genomic_accession=12140229, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3771851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771851, CurrentID=0, Status='live', Symbol='CG33927', Aliases='Dmel_ Dmel\\CG33927', description='CG33927 gene product from transcript CG33927-RA', other_designations='CG33927-PA|uncharacterized protein', map_location='82D1-82D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4689951, end_position_on_the_genomic_accession=4690608, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36300, CurrentID=0, Status='live', Symbol='OSCP1', Aliases='Dmel_CG13178, CG13178, Dmel\\CG13178, dOSCP1', description='Organic solute carrier partner 1', other_designations='CG13178-PA|OSCP1-PA|organic solute carrier partner 1', map_location='48E2-48E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12143191, end_position_on_the_genomic_accession=12144552, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36301, CurrentID=0, Status='live', Symbol='Hen1', Aliases='Dmel_CG12367, CG12367, Dm Dmel\\CG12367, FBgn 33686, PIMET, Pimet, Pimet/DmHen1, hen-1, hen1', description='CG12367 gene product from transcript CG12367-RA', other_designations='CG12367-PA|Hen1-PA|piRNA methyltransferase', map_location='48E2-48E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12145710, end_position_on_the_genomic_accession=12152752, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36302, CurrentID=0, Status='live', Symbol='CG8878', Aliases='Dmel_ BcDNA:LD23371, Dmel\\CG8878, VRK', description='CG8878 gene product from transcript CG8878-RA', other_designations='CG8878-PA|CG8878-PB|CG8878-PC', map_location='48E2-48E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12146024, end_position_on_the_genomic_accession=12151057, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36303, CurrentID=0, Status='live', Symbol='CG8407', Aliases='Dmel_ Dmel\\CG8407', description='CG8407 gene product from transcript CG8407-RB', other_designations='CG8407-PA|CG8407-PB', map_location='48E3-48E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12153071, end_position_on_the_genomic_accession=12154263, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36304, CurrentID=0, Status='live', Symbol='Prp8', Aliases='Dmel_CG8877, A1Z8U0, CG8877, Dmel\\CG8877, E-2e, prp8', description='pre-mRNA processing factor 8', other_designations='CG8877-PA|Prp8-PA|U5 small nuclear ribonucleoprotein-specific factor|prp8', map_location='48E3-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12154258, end_position_on_the_genomic_accession=12162998, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  36306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36306, CurrentID=0, Status='live', Symbol='SmD3', Aliases='Dmel_CG8427, CG8427, Dmel\\CG8427, SYM10, Sm D3, dD3, guf, guf2, l(2)131/7, l(2)154/1, l(2)90/37, smD3, smd3, snRNP Sm D3', description='Small ribonucleoprotein particle protein SmD3', other_designations='CG8427-PA|SmD3-PA|gut feeling|gutfeeling|small ribonucleoprotein Sm D3|small ribonucleoprotein particle protein SmD3', map_location='48E4-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12168731, end_position_on_the_genomic_accession=12170003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36307, CurrentID=0, Status='live', Symbol='Oda', Aliases='Dmel_CG16747, AZ, CG16747, D-AZ, Dmel\\CG16747, ODA, guf, l(2)02833, l(2)k09037, oda', description='Ornithine decarboxylase antizyme', other_designations='CG16747-PA|CG16747-PB|CG16747-PC|Oda-PA|Oda-PB|Oda-PC|antizyme|gut feeling|gutfeeling|ornithine deacarboxylase antizyme|ornithine decarboxylase antizyme|orthinine decarboyxlase antizyme', map_location='48E4-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12166295, end_position_on_the_genomic_accession=12174264, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36308, CurrentID=0, Status='live', Symbol='Cct5', Aliases='Dmel_CG8439, BcDNA.GM12270, BcDNA:GM12270, CCT5, CG8439, Dmel\\CG8439, TCP-1epsilon, Tcp1-epsilon, cct5, l(2)06444', description='T-complex Chaperonin 5', other_designations='CG8439-PA|CG8439-PB|Cct5-PA|Cct5-PB|T-complex chaperonin 5', map_location='48E4-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12174802, end_position_on_the_genomic_accession=12177516, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36309, CurrentID=0, Status='live', Symbol='EndoG', Aliases='Dmel_CG8862, CG8862, Dmel\\CG8862, dEndoG, endoG', description='Endonuclease G', other_designations='CG8862-PA|EndoG-PA|endonuclease G', map_location='48E4-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12177430, end_position_on_the_genomic_accession=12178553, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36310, CurrentID=0, Status='live', Symbol='CG8860', Aliases='Dmel_ BcDNA:GM14157, Dmel\\CG8860', description='CG8860 gene product from transcript CG8860-RA', other_designations='CG8860-PA', map_location='48E6-48E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12179483, end_position_on_the_genomic_accession=12179902, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36311, CurrentID=0, Status='live', Symbol='wash', Aliases='Dmel_CG13176, CG13176, Dmel\\CG13176, Wash, p63, p65', description='washout', other_designations='CG13176-PA|CG13176-PB|wash-PA|wash-PB', map_location='48E6-48E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12180493, end_position_on_the_genomic_accession=12182373, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771864, CurrentID=0, Status='live', Symbol='CG33669', Aliases='Dmel_ Dmel\\CG33669, NEST:bs03a05, bs03a05.y1', description='CG33669 gene product from transcript CG33669-RC', other_designations='CG33669-PC|uncharacterized protein', map_location='5F6-5F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6399455, end_position_on_the_genomic_accession=6400100, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36314, CurrentID=0, Status='live', Symbol='SmF', Aliases='Dmel_CG16792, BcDNA:GM23968, CG16792, Deb, Deb-B, DebB, Dmel\\CG16792, F-Sm, dF, deb-b, snRNP-F', description='Small ribonucleoprotein particle protein SmF', other_designations='CG16792-PA|SmF-PA|developmental embryonic B|small ribonucleoprotein particle protein SmF', map_location='48E6-48E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12183844, end_position_on_the_genomic_accession=12184310, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36316, CurrentID=0, Status='live', Symbol='Cyp6g1', Aliases='Dmel_CG8453, 6g1, CG8453, CYP6G1, CYP6g1, Cyp6-like, Cyp6G1, Cyp6gl, DDT-R, Dmel- Dmel\\CG8453, RDDT, RI, RI[DDT], RI[II], Rst(2)DDT, anon-WO03025223.16, anon-WO03025223.17, cyp6g1', description='CG8453 gene product from transcript CG8453-RB', other_designations='CG8453-PA|CG8453-PB|CYP6-like|Cyp6g1-PA|Cyp6g1-PB|cytochrome P450|resistance to DDT|resistance to Insecticide-DDT|resistance to insecticide on chromosome 2', map_location='48E7-48E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12185667, end_position_on_the_genomic_accession=12188431, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36317, CurrentID=0, Status='live', Symbol='Cyp6g2', Aliases='Dmel_CG8859, 6g2, CG8859, Cyp6G2, Dmel\\CG8859', description='CG8859 gene product from transcript CG8859-RA', other_designations='CG8859-PA|Cyp6g2-PA|cytochrome P450', map_location='48E7-48E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12188338, end_position_on_the_genomic_accession=12190377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36318, CurrentID=0, Status='live', Symbol='Cyp6t3', Aliases='Dmel_CG8457, 6t3, CG8457, Cyp6T3, Dmel\\CG8457', description='CG8457 gene product from transcript CG8457-RA', other_designations='CG8457-PA|Cyp6t3-PA', map_location='48E7-48E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12190682, end_position_on_the_genomic_accession=12192436, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36319, CurrentID=0, Status='live', Symbol='CG8858', Aliases='Dmel_ Dmel\\CG8858', description='CG8858 gene product from transcript CG8858-RA', other_designations='CG8858-PA', map_location='48E8-48E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12192374, end_position_on_the_genomic_accession=12198854, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3771872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771872, CurrentID=0, Status='live', Symbol='Hsp67Bb', Aliases='Dmel_CG4456, CG32041, CG4456, Dmel\\CG4456, EP(3)3247, Hsp-G2, Hspg2, anon-WO0118547.50, anon-WO0140519.138, gene 2, gene-2, gene2, gene4, hsp67B, hsp67Bb, small hsp locus 67B', description='Heat shock gene 67Bb', other_designations='CG4456-PA|CG4456-PB|CG4456-PC|CG4456-PD|Hsp67Bb-PA|Hsp67Bb-PB|Hsp67Bb-PC|Hsp67Bb-PD|gene 2|heat shock gene 2|heat shock gene 67Bb|heatshock protein 67Bb', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9372722, end_position_on_the_genomic_accession=9374964, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36321, CurrentID=0, Status='live', Symbol='RpS11', Aliases='Dmel_CG8857, CG8857, Dmel\\CG8857, M(2)48E, RPS11, S11, anon-EST:fe1D2, anon-MMS23, clone 23', description='Ribosomal protein S11', other_designations='CG8857-PB|CG8857-PC|RpS11-PB|RpS11-PC|anon-fast-evolving-1D2|ribosomal protein S11', map_location='48E8-48E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12199024, end_position_on_the_genomic_accession=12200815, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3771875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771875, CurrentID=0, Status='live', Symbol='CG33920', Aliases='Dmel_ Dmel\\CG33920', description='CG33920 gene product from transcript CG33920-RA', other_designations='CG33920-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31044391, end_position_on_the_genomic_accession=31045105, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36324, CurrentID=0, Status='live', Symbol='CG8854', Aliases='Dmel_ Dmel\\CG8854', description='CG8854 gene product from transcript CG8854-RC', other_designations='CG8854-PB|CG8854-PC', map_location='48E10-48E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12224461, end_position_on_the_genomic_accession=12229310, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771877, CurrentID=0, Status='live', Symbol='Adh', Aliases='Dmel_CG3481, ADH3, BG:DS01486.8, CG32954, CG3481, DM-ADH, DmADH, Dmel\\CG3481, Dreg-1, Reg-1, T16, adh, dADH, Adh', description='Alcohol dehydrogenase', other_designations='Adh-PC|Adh-PE|Adh-PF|Adh-PH|Adh-PI|CG3481-PC|CG3481-PE|CG3481-PF|CG3481-PH|CG3481-PI|alcohol dehydrogenase|alcohol dehyrdrogenase|alcoholdehydrogenase|drosophila rhythmically expressed gene 1', map_location='2-50.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14615552, end_position_on_the_genomic_accession=14618902, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36326, CurrentID=0, Status='live', Symbol='CG43316', Aliases='Dmel_ CG13169, Dmel\\CG43316, Dmel_CG13169', description='CG43316 gene product from transcript CG43316-RA', other_designations='CG43316-PA|uncharacterized protein', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12242546, end_position_on_the_genomic_accession=12243273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771879, CurrentID=0, Status='live', Symbol='His1:CG33810', Aliases='Dmel_CG33810, CG33810, Dmel\\CG33810', description='CG33810 gene product from transcript CG33810-RA', other_designations='CG33810-PA|His1:CG33810-PA|histone H1', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21437011, end_position_on_the_genomic_accession=21438033, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36328, CurrentID=0, Status='live', Symbol='Vha36-2', Aliases='Dmel_CG13167, CG13167, Dmel\\CG13167, vha36-2', description='Vacuolar H[+] ATPase 36kD subunit 2', other_designations='CG13167-PA|Vha36-2-PA|vacuolar H[+] ATPase 36kD subunit 2|vacuolar H[+] ATPase subunit 36-2', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12257282, end_position_on_the_genomic_accession=12258652, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36329, CurrentID=0, Status='live', Symbol='Cam', Aliases='Dmel_CG8472, 3909, CAM, CG8472, CaM, Cal, Cal49A1, D-CaM, DmCaM, DmCalm, Dmel\\CG8472, anon-EST:Posey59, cam, dCaM, l(2)03909, l(2)3909, lincRNA.S2634, Cam', description='Calmodulin', other_designations='CG8472-PA|CG8472-PB|CG8472-PC|CG8472-PD|CG8472-PE|Cam-PA|Cam-PB|Cam-PC|Cam-PD|Cam-PE|calmodulin', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12259398, end_position_on_the_genomic_accession=12278809, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36330, CurrentID=0, Status='live', Symbol='CG42700', Aliases='Dmel_ CG13165, CG34233, Dmel\\CG42700, Dmel_CG13165, Dmel_CG34233, ORE-16', description='CG42700 gene product from transcript CG42700-RJ', other_designations='CG42700-PC|CG42700-PD|CG42700-PE|CG42700-PF|CG42700-PG|CG42700-PH|CG42700-PI|CG42700-PJ|uncharacterized protein', map_location='48F6-48F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12279771, end_position_on_the_genomic_accession=12300795, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  36331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36331, CurrentID=0, Status='live', Symbol='CG8850', Aliases='Dmel_ DmNAT4, Dmel\\CG8850', description='CG8850 gene product from transcript CG8850-RB', other_designations='CG8850-PA|CG8850-PB', map_location='48F6-48F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12298747, end_position_on_the_genomic_accession=12302240, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36332, CurrentID=0, Status='live', Symbol='CG13164', Aliases='Dmel_ Dmel\\CG13164, SIP2, Sip2', description='CG13164 gene product from transcript CG13164-RE', other_designations='CG13164-PA|CG13164-PC|CG13164-PE|CG13164-PG|syntaxin Interact. Prot. 2|syntaxin interacting protein 2|uncharacterized protein', map_location='48F6-48F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12302544, end_position_on_the_genomic_accession=12306269, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36333, CurrentID=0, Status='live', Symbol='CG17739', Aliases='Dmel_ CT39321, Dmel\\CG17739, anon-WO0153538.35', description='CG17739 gene product from transcript CG17739-RA', other_designations='CG17739-PA|Fat-spondin|uncharacterized protein', map_location='48F6-48F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12306528, end_position_on_the_genomic_accession=12310412, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36334, CurrentID=0, Status='live', Symbol='CG30046', Aliases='Dmel_ CG17740, CT39323, Dmel\\CG30046', description='CG30046 gene product from transcript CG30046-RC', other_designations='CG30046-PB|CG30046-PC|uncharacterized protein', map_location='48F7-48F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12314117, end_position_on_the_genomic_accession=12317573, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36335, CurrentID=0, Status='live', Symbol='CG13163', Aliases='Dmel_ Dmel\\CG13163', description='CG13163 gene product from transcript CG13163-RB', other_designations='CG13163-PA|CG13163-PB|uncharacterized protein', map_location='48F7-48F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12321238, end_position_on_the_genomic_accession=12322797, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36336, CurrentID=0, Status='live', Symbol='CG8841', Aliases='Dmel_ Dmc1, Dmel\\CG8841', description='CG8841 gene product from transcript CG8841-RC', other_designations='CG8841-PA|CG8841-PB|CG8841-PC', map_location='48F8-48F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12322507, end_position_on_the_genomic_accession=12326258, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36337, CurrentID=0, Status='live', Symbol='garz', Aliases='Dmel_CG8487, ARF1GEF, ArfGEF, CG 8487, CG8487, Dmel\\CG8487, GBF1, GEF, Garz, ang', description='gartenzwerg', other_designations='CG8487-PA|CG8487-PB|CG8487-PC|angina|angle wing|garz-PA|garz-PB|garz-PC|sec7-family GEF', map_location='2-10.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12326360, end_position_on_the_genomic_accession=12333860, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3771890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771890, CurrentID=0, Status='live', Symbol='eys', Aliases='Dmel_CG33955, 21A6, CG15388, CG15389, CG33955, CG7245, CT35437, Dmel\\CG33955, EYS, Eyes Shut, Eys, MAb21A6, SPAM, Spam, mAb 21A6, mAb21A6, spacemaker, spam', description='eyes shut', other_designations='CG33955-PE|CG33955-PF|agrin-like|eyeshut|eys-PE|eys-PF|fibropellin|spacemaker|spam', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2311680, end_position_on_the_genomic_accession=2358944, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36339, CurrentID=0, Status='live', Symbol='Den1', Aliases='Dmel_CG8493, CG8493, DEN1, Dmel\\CG8493', description='Deneddylase 1', other_designations='CG8493-PA|CG8493-PC|Den1-PA|Den1-PC|deneddylase 1', map_location='48F10-48F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12337797, end_position_on_the_genomic_accession=12339544, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36340, CurrentID=0, Status='live', Symbol='CG8839', Aliases='Dmel_ Dmel\\CG8839', description='CG8839 gene product from transcript CG8839-RA', other_designations='CG8839-PA|CG8839-PC|CG8839-PD|CG8839-PE', map_location='48F10-48F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12339610, end_position_on_the_genomic_accession=12344617, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36341, CurrentID=0, Status='live', Symbol='ana3', Aliases='Dmel_CG13162, Ana3, CG13162, Dmel\\CG13162, l(2)SH0558', description='anastral spindle 3', other_designations='CG13162-PA|ana3-PA|anastral spindle phenotype 3|lethal (2) SH0558', map_location='48F11-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12344983, end_position_on_the_genomic_accession=12351325, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36342, CurrentID=0, Status='live', Symbol='CG33012', Aliases='Dmel_ CG13161, CG30041, Dmel\\CG33012', description='CG33012 gene product from transcript CG33012-RC', other_designations='CG33012-PC|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12367353, end_position_on_the_genomic_accession=12370886, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  36343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36343, CurrentID=0, Status='live', Symbol='CG13160', Aliases='Dmel_ Dmel\\CG13160', description='CG13160 gene product from transcript CG13160-RB', other_designations='CG13160-PA|CG13160-PB|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12371594, end_position_on_the_genomic_accession=12375081, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36344, CurrentID=0, Status='live', Symbol='CG13159', Aliases='Dmel_ BcDNA:RE25177, Dmel\\CG13159', description='CG13159 gene product from transcript CG13159-RA', other_designations='CG13159-PA|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12375286, end_position_on_the_genomic_accession=12375678, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36345, CurrentID=0, Status='live', Symbol='Cpr49Ab', Aliases='Dmel_CG30042, CG30042, CG8496, DmelCpr49Ab, Dmel\\CG30042', description='Cuticular protein 49Ab', other_designations='CG30042-PA|Cpr49Ab-PA|cuticular protein 49Ab', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12378440, end_position_on_the_genomic_accession=12379396, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36346, CurrentID=0, Status='live', Symbol='CG13155', Aliases='Dmel_ Dmel\\CG13155', description='CG13155 gene product from transcript CG13155-RA', other_designations='CG13155-PA|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12379614, end_position_on_the_genomic_accession=12380426, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36347, CurrentID=0, Status='live', Symbol='CG8501', Aliases='Dmel_ Dmel\\CG8501', description='CG8501 gene product from transcript CG8501-RA', other_designations='CG8501-PA|CG8501-PB', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12383045, end_position_on_the_genomic_accession=12384520, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36348, CurrentID=0, Status='live', Symbol='Cpr49Ac', Aliases='Dmel_CG8502, CG8502, Dmel\\CG8502', description='Cuticular protein 49Ac', other_designations='CG8502-PA|CG8502-PC|CG8502-PD|CG8502-PE|CG8502-PF|CG8502-PG|Cpr49Ac-PA|Cpr49Ac-PC|Cpr49Ac-PD|Cpr49Ac-PE|Cpr49Ac-PF|Cpr49Ac-PG|cuticular protein 49Ac', map_location='49A2-49A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12386666, end_position_on_the_genomic_accession=12390057, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36349, CurrentID=0, Status='live', Symbol='Cpr49Ad', Aliases='Dmel_CG8836, CG8836, Dmel\\CG8836', description='Cuticular protein 49Ad', other_designations='CG8836-PA|Cpr49Ad-PA|cuticular protein 49Ad', map_location='49A2-49A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12390896, end_position_on_the_genomic_accession=12391699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36350, CurrentID=0, Status='live', Symbol='Or49a', Aliases='Dmel_CG13158, 49A.1, 49a, CG13158, DOR10, Dmel\\CG13158, OR49a', description='Odorant receptor 49a', other_designations='CG13158-PA|Or49a-PA|odorant receptor 49a', map_location='49A2-49A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12393483, end_position_on_the_genomic_accession=12394842, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36351, CurrentID=0, Status='live', Symbol='Cpr49Ae', Aliases='Dmel_CG8505, CG8505, DmelCpr49Ae, Dmel\\CG8505', description='Cuticular protein 49Ae', other_designations='CG8505-PA|Cpr49Ae-PA|cuticular protein 49Ae', map_location='49A2-49A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12402216, end_position_on_the_genomic_accession=12405108, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36352, CurrentID=0, Status='live', Symbol='Cpr49Af', Aliases='Dmel_CG8510, CG8510, DmelCpr49Af, Dmel\\CG8510', description='Cuticular protein 49Af', other_designations='CG8510-PA|CG8510-PB|Cpr49Af-PA|Cpr49Af-PB|cuticular protein 49Af', map_location='49A2-49A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12405932, end_position_on_the_genomic_accession=12407045, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36353, CurrentID=0, Status='live', Symbol='Cpr49Ag', Aliases='Dmel_CG8511, CG8511, DmelCpr49Ag, Dmel\\CG8511', description='Cuticular protein 49Ag', other_designations='CG8511-PA|Cpr49Ag-PA|cuticular protein 49Ag', map_location='49A3-49A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12407793, end_position_on_the_genomic_accession=12408523, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36354, CurrentID=0, Status='live', Symbol='Cpr49Ah', Aliases='Dmel_CG8515, CG8515, DmelCpr49Ah, Dmel\\CG8515', description='Cuticular protein 49Ah', other_designations='CG8515-PA|Cpr49Ah-PA|cuticular protein 49Ah', map_location='49A3-49A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12413962, end_position_on_the_genomic_accession=12415170, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36355, CurrentID=0, Status='live', Symbol='CG13157', Aliases='Dmel_ Dmel\\CG13157', description='CG13157 gene product from transcript CG13157-RA', other_designations='CG13157-PA|uncharacterized protein', map_location='49A3-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12416044, end_position_on_the_genomic_accession=12418231, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36356, CurrentID=0, Status='live', Symbol='CG8834', Aliases='Dmel_ Dmel\\CG8834, mdcds_23070', description='CG8834 gene product from transcript CG8834-RA', other_designations='CG8834-PA', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12420475, end_position_on_the_genomic_accession=12422535, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36357, CurrentID=0, Status='live', Symbol='CG8520', Aliases='Dmel_ Dmel\\CG8520, FBgn 33734', description='CG8520 gene product from transcript CG8520-RA', other_designations='CG8520-PA|CG8520-PB', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12422931, end_position_on_the_genomic_accession=12425459, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36358, CurrentID=0, Status='live', Symbol='CG8525', Aliases='Dmel_ Dmel\\CG8525', description='CG8525 gene product from transcript CG8525-RA', other_designations='CG8525-PA|CG8525-PD', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12425483, end_position_on_the_genomic_accession=12426975, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3771911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771911, CurrentID=0, Status='live', Symbol='His1:CG33834', Aliases='Dmel_CG33834, CG33834, Dmel\\CG33834', description='CG33834 gene product from transcript CG33834-RA', other_designations='CG33834-PA|His1:CG33834-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21486496, end_position_on_the_genomic_accession=21487518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36360, CurrentID=0, Status='live', Symbol='Nup54', Aliases='Dmel_CG8831, CG8831, Dmel\\CG8831, Nup53, nup54', description='Nucleoporin 54kD', other_designations='CG8831-PA|Nup54-PA|nucleoporin 54|nucleoporin 54kD', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12428449, end_position_on_the_genomic_accession=12431099, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36361, CurrentID=0, Status='live', Symbol='DUBAI', Aliases='Dmel_CG8830, CG8830, Dmel\\CG8830, Q7JQI1, Usp38', description='Deubiquitinating apoptotic inhibitor', other_designations='CG8830-PA|CG8830-PB|DUBAI-PA|DUBAI-PB|DeUBiquitinating-Apoptotic-Inhibitor|deubiquitinating apoptotic inhibitor', map_location='49A4-49A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12431447, end_position_on_the_genomic_accession=12436577, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36362, CurrentID=0, Status='live', Symbol='Dyb', Aliases='Dmel_CG8529, CG8529, DmDYB, Dmel\\CG8529', description='Dystrobrevin', other_designations='CG8529-PA|CG8529-PC|CG8529-PD|CG8529-PF|CG8529-PG|CG8529-PH|CG8529-PI|Dyb-PA|Dyb-PC|Dyb-PD|Dyb-PF|Dyb-PG|Dyb-PH|Dyb-PI|Dystrobrevin-like|dystrobrevin', map_location='49A6-49A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12437345, end_position_on_the_genomic_accession=12452916, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36363, CurrentID=0, Status='live', Symbol='Lac', Aliases='Dmel_CG12369, CG12369, CT24921, Dmel\\CG12369, bulb, l(2)k11012, l(2)k11012b, lac', description='Lachesin', other_designations='CG12369-PA|Lac-PA|bulbous|lachesin|lachesin bulbous', map_location='49A6-49A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12453190, end_position_on_the_genomic_accession=12465630, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36364, CurrentID=0, Status='live', Symbol='dgt5', Aliases='Dmel_CG8828, CG8828, Dgt5, Dmel\\CG8828, msd2', description='dim gamma-tubulin 5', other_designations='CG8828-PA|CG8828-PB|dgt5-PA|dgt5-PB|mitotic spindle density 2', map_location='49A7-49A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12465678, end_position_on_the_genomic_accession=12471775, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36365, CurrentID=0, Status='live', Symbol='CG8545', Aliases='Dmel_ Dmel\\CG8545, Nol1Nt', description='CG8545 gene product from transcript CG8545-RA', other_designations='CG8545-PA', map_location='49A7-49A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12468651, end_position_on_the_genomic_accession=12471722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36366, CurrentID=0, Status='live', Symbol='CG8550', Aliases='Dmel_ Dmel\\CG8550', description='CG8550 gene product from transcript CG8550-RA', other_designations='CG8550-PA', map_location='49A7-49A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12471915, end_position_on_the_genomic_accession=12474021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36368, CurrentID=0, Status='live', Symbol='Dh44-R2', Aliases='Dmel_CG12370, CG12370, CG1315, CG13156, DH-R, DH44-R2, Dmel\\CG12370, anon-WO0170980.103, anon-WO0170980.104', description='Diuretic hormone 44 receptor 2', other_designations='CG12370-PA|CG12370-PB|CRF-like diuretic hormonereceptor 2|Dh44-R2-PA|Dh44-R2-PB|diuretic hormone 44 receptor 2', map_location='49A7-49A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12474424, end_position_on_the_genomic_accession=12487986, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  36371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36371, CurrentID=0, Status='live', Symbol='s-cup', Aliases='Dmel_CG30044, CG13152, CG13153, CG30044, Dmel\\CG30044', description='stanley-cup', other_designations='CG30044-PB|s-cup-PB', map_location='49A9-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12495672, end_position_on_the_genomic_accession=12501459, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36372, CurrentID=0, Status='live', Symbol='vis', Aliases='Dmel_CG8821, CG8821, Dmel\\CG8821, TGIFa, achi/vis, dTGIFa, zaa', description='vismay', other_designations='CG8821-PA|CG8821-PB|achi/vis|vis-PA|vis-PB|zuker-aly-like a', map_location='49A10-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12507896, end_position_on_the_genomic_accession=12510615, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36373, CurrentID=0, Status='live', Symbol='achi', Aliases='Dmel_CG8819, Achi, CG8819, Dmel\\CG8819, TGIF, ach/vis, dTGIF, dTGIFa, dTGIFb, vis, zaa, achi', description='achintya', other_designations='CG8819-PA|CG8819-PC|CG8819-PD|achi-PA|achi-PC|achi-PD|zuker-aly-like a', map_location='49A10-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12511534, end_position_on_the_genomic_accession=12515062, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36374, CurrentID=0, Status='live', Symbol='CG13151', Aliases='Dmel_ Dmel\\CG13151', description='CG13151 gene product from transcript CG13151-RA', other_designations='CG13151-PA|uncharacterized protein', map_location='49A10-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12515375, end_position_on_the_genomic_accession=12517284, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36375, CurrentID=0, Status='live', Symbol='CG8818', Aliases='Dmel_ Dmel\\CG8818', description='CG8818 gene product from transcript CG8818-RA', other_designations='CG8818-PA', map_location='49A10-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12517264, end_position_on_the_genomic_accession=12518912, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36376, CurrentID=0, Status='live', Symbol='CG8569', Aliases='Dmel_ Dmel\\CG8569', description='CG8569 gene product from transcript CG8569-RA', other_designations='CG8569-PA', map_location='49A10-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12519166, end_position_on_the_genomic_accession=12521624, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36377, CurrentID=0, Status='live', Symbol='fra', Aliases='Dmel_CG8581, CG8581, CT24981, DCC, Dmel\\CG8581, Fra', description='frazzled', other_designations='CG8581-PA|CG8581-PB|CG8581-PC|DCC/Frazzled|Netrin-receptor|deleted in colorectal cancer|deleted in colorectal carcinoma|fra-PA|fra-PB|fra-PC', map_location='49A10-49B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12529053, end_position_on_the_genomic_accession=12564184, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36378, CurrentID=0, Status='live', Symbol='Cyp301a1', Aliases='Dmel_CG8587, 301a1, CG8587, Cyp301A1, Dmel\\CG8587', description='CG8587 gene product from transcript CG8587-RA', other_designations='CG8587-PA|Cyp301a1-PA|cytochrome P450', map_location='49B4-49B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12566185, end_position_on_the_genomic_accession=12569061, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36379, CurrentID=0, Status='live', Symbol='Ak6', Aliases='Dmel_CG8816, CG8816, DAK6, Dmel\\CG8816', description='Adenylate kinase 6', other_designations='Ak6-PA|CG8816-PA|adenylate kinase 6', map_location='49B5-49B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12569014, end_position_on_the_genomic_accession=12569847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36380, CurrentID=0, Status='live', Symbol='stil', Aliases='Dmel_CG8592, 152072_at, CG8592, Dmel\\CG8592, Stil, Y10276', description='stand still', other_designations='CG8592-PA|stil-PA', map_location='2-63 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12570367, end_position_on_the_genomic_accession=12571883, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36381, CurrentID=0, Status='live', Symbol='ClC-b', Aliases='Dmel_CG8594, CG8594, Clcn7, DmClC-b, DmClCb, Dmel\\CG8594', description='Chloride channel-b', other_designations='CG8594-PA|ClC-b-PA|chloride channel-b', map_location='49B5-49B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12572068, end_position_on_the_genomic_accession=12575299, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36382, CurrentID=0, Status='live', Symbol='Sin3A', Aliases='Dmel_CG8815, 159A11T, 27/3, 97/15, CG8815, Dmel\\CG8815, ES2-3, ESTS:159A11T, EY2-9, SIN3, SIN3A, Sin 3A, Sin3, Sin3a, dSin3, d l(2)08269, l(2)k02703, l(2)k05415, l(2)k09715, ld, sin3, sin3A, sin3a, sni3', description='CG8815 gene product from transcript CG8815-RC', other_designations='CG8815-PA|CG8815-PB|CG8815-PC|CG8815-PD|CG8815-PE|CG8815-PF|CG8815-PG|CG8815-PH|CG8815-PI|CG8815-PJ|Sin3A-PA|Sin3A-PB|Sin3A-PC|Sin3A-PD|Sin3A-PE|Sin3A-PF|Sin3A-PG|Sin3A-PH|Sin3A-PI|Sin3A-PJ|enhancer of GMR-sina 2-3', map_location='49B5-49B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12575240, end_position_on_the_genomic_accession=12592917, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  36383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36383, CurrentID=0, Status='live', Symbol='Amph', Aliases='Dmel_CG8604, Amp, CG8604, DAMP, Damp, Dmel\\CG8604, amph, dAmp, dAmph, damph', description='Amphiphysin', other_designations='Amph-PA|Amph-PB|Amph-PC|CG8604-PA|CG8604-PB|CG8604-PC|amphiphysin|dAmphiphysin', map_location='49B7-49B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12593604, end_position_on_the_genomic_accession=12612068, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36384, CurrentID=0, Status='live', Symbol='Galphaq', Aliases='Dmel_CG17759, CG17759, DGalpha[[q]], DGalphaq, DGq, DGqalpha, Dgq, DmGalpha[[q]], Dmel\\CG17759, Dromel_CG17759_galpha49b_mORF, G alpha 49B, G?q, G[[a]]q, G[[alphaq]], G[[q]], G[[q]]alpha, G[[qalpha]], G[q], Galpha, Galpha49B, Galpha49b, Galpha[[q]], Gaq, Gq, Gq/11, Gqa-3, Gqalpha, Gqalpha[[e]], Gqalphae, Gqalpha{e}, dG[[alphaq]], dG[[q]]alpha, dGaalpha, dGq, dGqa-49b, dGqalpha, dGqalpha-3, dGqalpha-49b, dgq, dgqalpha, galpha49B, sast', description='G protein alpha q subunit', other_designations='CG17759-PA|CG17759-PC|CG17759-PD|CG17759-PE|CG17759-PG|CG17759-PJ|CG17759-PK|G protein 49B|G protein alpha49B|G-protein alpha subunit|G-protein alpha49B|G-protein-alpha[[q]]|Galphaq-PA|Galphaq-PC|Galphaq-PD|Galphaq-PE|Galphaq-PG|Galphaq-PJ|Galphaq-PK|small amplify and slow termination', map_location='49B8-49B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12612740, end_position_on_the_genomic_accession=12623197, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36385, CurrentID=0, Status='live', Symbol='CG17760', Aliases='Dmel_ BcDNA:SD21019, Dmel\\CG17760', description='CG17760 gene product from transcript CG17760-RB', other_designations='CG17760-PB|uncharacterized protein', map_location='49B9-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12625808, end_position_on_the_genomic_accession=12628619, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36386, CurrentID=0, Status='live', Symbol='muskelin', Aliases='Dmel_CG8811, CG8811, DmMK, Dmel\\CG8811', description='CG8811 gene product from transcript CG8811-RC', other_designations='CG8811-PA|CG8811-PB|CG8811-PC|muskelin-PA|muskelin-PB|muskelin-PC', map_location='49B10-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12628697, end_position_on_the_genomic_accession=12632378, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36387, CurrentID=0, Status='live', Symbol='spt4', Aliases='Dmel_CG12372, 142904_at, CG12372, DSIF, Dmel\\CG12372, D SPT4, Spt4, dDSIF, dspt4, l(2)k05316', description='CG12372 gene product from transcript CG12372-RA', other_designations='CG12372-PA|DSIF|spt4-PA', map_location='49B10-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12633589, end_position_on_the_genomic_accession=12634172, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3771941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771941, CurrentID=0, Status='live', Symbol='His4:CG33875', Aliases='Dmel_CG33875, CG33875, Dmel\\CG33875', description='CG33875 gene product from transcript CG33875-RA', other_designations='CG33875-PA|His4:CG33875-PA|histone H4', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21519556, end_position_on_the_genomic_accession=21520083, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36390, CurrentID=0, Status='live', Symbol='Iswi', Aliases='Dmel_CG8625, ACF, CG8625, CHRAC, Dmel\\CG8625, ISW, ISWI, NURF, NURF 140, NURF-140, Nurf-140, SWI, anon-EP1279744.124, dCHRAC, dISWI, dNURF, iswi, p140, p140/ISWI', description='Imitation SWI', other_designations='CG8625-PA|CG8625-PB|CG8625-PC|ISWI|ISWI ATPase|Iswi-PA|Iswi-PB|Iswi-PC|imitation SW|imitation SWI|imitation switch|imitation-SWI|nucleosome remodeling factor|nucleosome remodeling factor - 140kD', map_location='49B10-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12636671, end_position_on_the_genomic_accession=12640628, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36391, CurrentID=0, Status='live', Symbol='CG8785', Aliases='Dmel_ Dmel\\CG8785', description='CG8785 gene product from transcript CG8785-RA', other_designations='CG8785-PA|CG8785-PB', map_location='49B10-49B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12640690, end_position_on_the_genomic_accession=12648224, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36392, CurrentID=0, Status='live', Symbol='CG8778', Aliases='Dmel_ Dmel\\CG8778', description='CG8778 gene product from transcript CG8778-RA', other_designations='CG8778-PA|Enoyl-CoA hydratase', map_location='49B11-49B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12651301, end_position_on_the_genomic_accession=12652487, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36393, CurrentID=0, Status='live', Symbol='ZnT49B', Aliases='Dmel_CG8632, CG8632, Dmel\\CG8632, cg8632, dZnT49B', description='Zinc transporter 49B', other_designations='CG8632-PA|CG8632-PB|ZnT49B-PA|ZnT49B-PB|zinc transporter 49B', map_location='49B11-49B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12652805, end_position_on_the_genomic_accession=12655743, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36394, CurrentID=0, Status='live', Symbol='CG8646', Aliases='Dmel_ Dmel\\CG8646', description='CG8646 gene product from transcript CG8646-RB', other_designations='CG8646-PB', map_location='49B11-49B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12655954, end_position_on_the_genomic_accession=12659341, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36395, CurrentID=0, Status='live', Symbol='nemy', Aliases='Dmel_CG8776, CG13147, CG8776, CytB561, Dmel\\CG8776, Dmel_CG8772, Nemy, P153, anon-WO0140519.146', description='no extended memory', other_designations='CG8776-PA|CG8776-PB|CG8776-PC|CG8776-PE|CG8776-PG|CG8776-PK|CG8776-PL|CG8776-PM|DcytB-like|cytochrome b561|nemy-PA|nemy-PB|nemy-PC|nemy-PE|nemy-PG|nemy-PK|nemy-PL|nemy-PM|no-extended memory', map_location='49B12-49C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12659497, end_position_on_the_genomic_accession=12670219, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36396, CurrentID=0, Status='live', Symbol='GLS', Aliases='Dmel_CG42708, CG42708, CG8772, CG8872, Dmel\\CG42708, Dmel_CG8772', description='Glutaminase', other_designations='CG42708-PA|CG42708-PB|CG42708-PC|CG42708-PD|CG42708-PE|CG42708-PF|CG42708-PH|CG42708-PI|CG42708-PJ|CG42708-PK|GLS-PA|GLS-PB|GLS-PC|GLS-PD|GLS-PE|GLS-PF|GLS-PH|GLS-PI|GLS-PJ|GLS-PK|glutaminase', map_location='49B12-49B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12670483, end_position_on_the_genomic_accession=12679593, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36397, CurrentID=0, Status='live', Symbol='CG8771', Aliases='Dmel_ Dmel\\CG8771, Nup188', description='CG8771 gene product from transcript CG8771-RB', other_designations='CG8771-PB|CG8771-PC', map_location='49B12-49B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12679967, end_position_on_the_genomic_accession=12686179, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36398, CurrentID=0, Status='live', Symbol='CG13148', Aliases='Dmel_ Dmel\\CG13148', description='CG13148 gene product from transcript CG13148-RC', other_designations='CG13148-PA|CG13148-PB|CG13148-PC|CG13148-PD|uncharacterized protein', map_location='49B12-49B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12686249, end_position_on_the_genomic_accession=12688786, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36399, CurrentID=0, Status='live', Symbol='Obp49a', Aliases='Dmel_CG30052, BcDNA:GH25962, CG30052, CG8769, DmelObp49a, Dmel\\CG30052, Obp49A, obp49a', description='Odorant-binding protein 49a', other_designations='CG30052-PA|Obp49a-PA', map_location='49B12-49B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12690059, end_position_on_the_genomic_accession=12690955, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36400, CurrentID=0, Status='live', Symbol='CG8768', Aliases='Dmel_ Dmel\\CG8768', description='CG8768 gene product from transcript CG8768-RA', other_designations='CG8768-PA', map_location='49C1-49C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12691308, end_position_on_the_genomic_accession=12692690, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36401, CurrentID=0, Status='live', Symbol='wuc', Aliases='Dmel_CG12442, 147114_at, BcDNA:AT27450, CG12442, Dmel\\CG12442', description='Wake-up-call', other_designations='CG12442-PA|CG12442-PB|wake-up-call|wuc-PA|wuc-PB', map_location='49C1-49C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12700081, end_position_on_the_genomic_accession=12700851, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36403, CurrentID=0, Status='live', Symbol='CG42663', Aliases='Dmel_ BcDNA:SD15892, CG12488, CG13486, CG18356, Dmel\\CG42663, Dmel_CG12488, Dmel_CG18356', description='CG42663 gene product from transcript CG42663-RB', other_designations='CG42663-PB|CG42663-PC|CG42663-PD|uncharacterized protein', map_location='49C2-49C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12705172, end_position_on_the_genomic_accession=12751545, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36404, CurrentID=0, Status='live', Symbol='Mos', Aliases='Dmel_CG8767, CG8767, D-mos, DMOS, Dmel\\CG8767, MOS, dmos, mos', description='Mos oncogene', other_designations='CG8767-PA|Mos-PA|mos', map_location='49C2-49C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12751483, end_position_on_the_genomic_accession=12752936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36405, CurrentID=0, Status='live', Symbol='Taz', Aliases='Dmel_CG8766, CG8766, Dmel\\CG8766, TAZ, dTAZ, tafazzin', description='Tafazzin', other_designations='CG8766-PA|CG8766-PB|CG8766-PC|CG8766-PD|Taz-PA|Taz-PB|Taz-PC|Taz-PD|tafazzin|tafazzins', map_location='49C2-49C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12753026, end_position_on_the_genomic_accession=12756414, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36406, CurrentID=0, Status='live', Symbol='mRpL18', Aliases='Dmel_CG12373, CG12373, Dmel\\CG12373, HL, MRP-L18, anon-49Ca', description='mitochondrial ribosomal protein L18', other_designations='CG12373-PA|mRpL18-PA', map_location='49C2-49C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12756434, end_position_on_the_genomic_accession=12757173, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771959, CurrentID=0, Status='live', Symbol='His3:CG33806', Aliases='Dmel_CG33806, CG33806, Dmel\\CG33806', description='CG33806 gene product from transcript CG33806-RA', other_designations='CG33806-PA|His3:CG33806-PA|histone H3', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21430255, end_position_on_the_genomic_accession=21430783, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36408, CurrentID=0, Status='live', Symbol='Dgkepsilon', Aliases='Dmel_CG8657, CG8657, DGKe, DGKepsilon, Dmel\\CG8657, dgkepsilon', description='Diacyl glycerol kinase epsilon', other_designations='CG8657-PA|CG8657-PB|Dgkepsilon-PA|Dgkepsilon-PB|diacyl glycerol kinase epsilon|diacylglycerol kinase epsilon', map_location='49C2-49C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12757546, end_position_on_the_genomic_accession=12759724, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36409, CurrentID=0, Status='live', Symbol='Nacalpha', Aliases='Dmel_CG8759, CG8759, Dmel\\CG8759, NAC, NACalpha, aic, alpha-NAC, alphaNAC, l(2)04329', description='Nascent polypeptide associated complex protein alpha subunit', other_designations='CG8759-PA|CG8759-PB|CG8759-PC|CG8759-PD|NAC protein alpha subunit|Nacalpha-PA|Nacalpha-PB|Nacalpha-PC|Nacalpha-PD|alpha NAC|lethal (2) 04329|nascent polypeptide associated complex protein alpha subunit', map_location='49C2-49C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12759693, end_position_on_the_genomic_accession=12760999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36410, CurrentID=0, Status='live', Symbol='CG12374', Aliases='Dmel_ Dmel\\CG12374, anon-WO0140519.63', description='CG12374 gene product from transcript CG12374-RA', other_designations='CG12374-PA|uncharacterized protein', map_location='49C3-49C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12762113, end_position_on_the_genomic_accession=12763825, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36411, CurrentID=0, Status='live', Symbol='sca', Aliases='Dmel_CG17579, AAA28880, AAF58455, CG17579, Dmel\\CG17579, FBpp0086968, FBpp0086969, Sca, Sca1b, anon-EST:Liang-38, clone 38, mAb sca1', description='scabrous', other_designations='CG17579-PA|CG17579-PB|Fibrinogen-related protein|sca-PA|sca-PB', map_location='2-66.7 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12780541, end_position_on_the_genomic_accession=12802010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36412, CurrentID=0, Status='live', Symbol='Cyp9h1', Aliases='Dmel_CG17577, 9h1, CG17577, Dmel\\CG17577', description='CG17577 gene product from transcript CG17577-RA', other_designations='CG17577-PA|Cyp9h1-PA|cytochrome P450', map_location='49D4-49D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12840813, end_position_on_the_genomic_accession=12842558, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36413, CurrentID=0, Status='live', Symbol='Or49b', Aliases='Dmel_CG17584, 49D.1, 49b, AN13, CG17584, DOR105, Dmel\\CG17584', description='Odorant receptor 49b', other_designations='CG17584-PA|Or49b-PA|odorant receptor 49b', map_location='49D4-49D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12842665, end_position_on_the_genomic_accession=12844704, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36414, CurrentID=0, Status='live', Symbol='CG17575', Aliases='Dmel_ Dmel\\CG17575, anon-SAGE:Wang-113', description='CG17575 gene product from transcript CG17575-RA', other_designations='CG17575-PA|CG17575-PB|anon-SAGE:Wang-113|uncharacterized protein', map_location='49D4-49D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12845544, end_position_on_the_genomic_accession=12846596, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36415, CurrentID=0, Status='live', Symbol='CG17574', Aliases='Dmel_ BcDNA:RH03863, Dmel\\CG17574', description='CG17574 gene product from transcript CG17574-RA', other_designations='CG17574-PA|CG17574-PE|CG17574-PF|CG17574-PG|CG17574-PH|CG17574-PI|CG17574-PJ|CG17574-PK|CG17574-PL|uncharacterized protein', map_location='49D4-49D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12849358, end_position_on_the_genomic_accession=12869423, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  3771968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771968, CurrentID=0, Status='live', Symbol='Msp300', Aliases='Dmel_CG42768, BBF-1, Bbbf1, BcDNA:AT10293, BcDNA:GH02446, BcDNA:GH18470, BcDNA:LP10936, CG14035, CG18251, CG18252, CG31649, CG31916, CG33549, CG33715, CG42768, Dmel\\CG42768, Dmel_CG14035, Dmel_CG33715, FBgn0010070, MSP-300, MSP300, Msp-300, Spec25CD, jf22, l(2)25Ch, l(2)jf22, msp-300, msp300', description='Muscle-specific protein 300 kDa', other_designations='Box B-binding factor 1|CG42768-PB|CG42768-PD|CG42768-PE|CG42768-PF|CG42768-PG|CG42768-PH|CG42768-PI|CG42768-PJ|CG42768-PK|CG42768-PL|CG42768-PM|Msp-300|Msp300-PB|Msp300-PD|Msp300-PE|Msp300-PF|Msp300-PG|Msp300-PH|Msp300-PI|Msp300-PJ|Msp300-PK|Msp300-PL|Msp300-PM|Muscle-specific protein 300|Muscle-specific protein-300|lethal(2)25Ch|nesprin|repetitive ORF', map_location='25C6-25C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5100877, end_position_on_the_genomic_accession=5207002, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  36417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36417, CurrentID=0, Status='live', Symbol='CG3790', Aliases='Dmel_ CT, Dmel\\CG3790', description='CG3790 gene product from transcript CG3790-RA', other_designations='CG3790-PA|carnitine transporter', map_location='49D6-49D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12872787, end_position_on_the_genomic_accession=12875012, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36418, CurrentID=0, Status='live', Symbol='CG3814', Aliases='Dmel_ BEST:CK02213, CK02213, Dmel\\CG3814', description='CG3814 gene product from transcript CG3814-RB', other_designations='CG3814-PA|CG3814-PB|CG3814-PC|CG3814-PD', map_location='49D6-49D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12875164, end_position_on_the_genomic_accession=12876966, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36419, CurrentID=0, Status='live', Symbol='Nmda1', Aliases='Dmel_CG3798, CG3798, DRONMDA, Dmel\\CG3798, NMDARA1, cDNA2, dNMDARA1, dTmbim3', description='N-methyl-D-aspartate receptor-associated protein', other_designations='CG3798-PA|CG3798-PB|CG3798-PC|CG3798-PD|CG3798-PE|CG3798-PF|CG3798-PG|CG3798-PH|CG3798-PI|NMDARA1-like|Nmda1-PA|Nmda1-PB|Nmda1-PC|Nmda1-PD|Nmda1-PE|Nmda1-PF|Nmda1-PG|Nmda1-PH|Nmda1-PI', map_location='49D6-49D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12877094, end_position_on_the_genomic_accession=12880230, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36420, CurrentID=0, Status='live', Symbol='Aats-asp', Aliases='Dmel_CG3821, Aats-Asp, AspRS, CG3821, DRS, Dmel\\CG3821, aats-asp, anon-EST:Liang-2.49, cDNA1, clone 2.49, l(2)49Db, l(2)k04508, l(2)v27, l(2)vr1, poney, vr1', description='Aspartyl-tRNA synthetase', other_designations='Aats-asp-PA|Aats-asp-PB|Aats-asp-PC|CG3821-PA|CG3821-PB|CG3821-PC|lethal(2)49Db', map_location='49D7-49D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12877004, end_position_on_the_genomic_accession=12883168, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36421, CurrentID=0, Status='live', Symbol='vg', Aliases='Dmel_CG3830, CG3830, Dmel\\CG3830, VG, Vg21, vg', description='vestigial', other_designations='CG3830-PA|vestgial|vestigal|vg-PA', map_location='2-67.0 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12884632, end_position_on_the_genomic_accession=12899385, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36423, CurrentID=0, Status='live', Symbol='CG13319', Aliases='Dmel_ BcDNA:GH19585, Dmel\\CG13319', description='CG13319 gene product from transcript CG13319-RA', other_designations='CG13319-PA|CG13319-PB|uncharacterized protein', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12923904, end_position_on_the_genomic_accession=12924664, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36424, CurrentID=0, Status='live', Symbol='sug', Aliases='Dmel_CG3850, CG3850, Dmel\\CG3850', description='sugarbabe', other_designations='CG3850-PA|CG3850-PB|sug-PA|sug-PB', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12924567, end_position_on_the_genomic_accession=12930644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36425, CurrentID=0, Status='live', Symbol='CG17019', Aliases='Dmel_ Dmel\\CG17019', description='CG17019 gene product from transcript CG17019-RA', other_designations='CG17019-PA|CG17019-PB|uncharacterized protein', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12930912, end_position_on_the_genomic_accession=12934045, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36426, CurrentID=0, Status='live', Symbol='CG13322', Aliases='Dmel_ Dmel\\CG13322, anon-WO0118547.137', description='CG13322 gene product from transcript CG13322-RA', other_designations='CG13322-PA|CG13322-PB|CG13322-PC|uncharacterized protein', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12934234, end_position_on_the_genomic_accession=12936535, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36427, CurrentID=0, Status='live', Symbol='Sans', Aliases='Dmel_CG13320, CG13320, DSANS, Dmel\\CG13320', description='Sans ortholog', other_designations='CG13320-PA|CG13320-PB|Sans-PA|Sans-PB|sans ortholog', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12936891, end_position_on_the_genomic_accession=12940412, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36428, CurrentID=0, Status='live', Symbol='Mdr49', Aliases='Dmel_CG3879, CG3879, Dm5, Dmel\\CG3879, MDR49, MDR5, Mes5, mdr49, mrd49', description='Multi drug resistance 49', other_designations='CG3879-PA|CG3879-PB|Mdr49-PA|Mdr49-PB|multi drug resistance 49', map_location='49E1-49E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12940323, end_position_on_the_genomic_accession=12946347, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36429, CurrentID=0, Status='live', Symbol='CG44251', Aliases='Dmel_ CG3884, Dmel\\CG44251', description='CG44251 gene product from transcript CG44251-RA', other_designations='CG44251-PA|CG44251-PB|uncharacterized protein', map_location='49E4-49E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12948348, end_position_on_the_genomic_accession=12952293, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36430, CurrentID=0, Status='live', Symbol='CG13321', Aliases='Dmel_ Dmel\\CG13321', description='CG13321 gene product from transcript CG13321-RB', other_designations='CG13321-PA|CG13321-PB|uncharacterized protein', map_location='49E4-49E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12953112, end_position_on_the_genomic_accession=12955712, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36431, CurrentID=0, Status='live', Symbol='Psc', Aliases='Dmel_CG3886, 49Ea, CG3886, D-Bmi, Dmel\\CG3886, PSC1, Su(z)2-C, dPSC, l(2)49Ea, l(2)Psc, l(2)vr14, psc, vr14, Psc', description='Posterior sex combs', other_designations='CG3886-PA|CG3886-PB|CG3886-PC|Psc-PA|Psc-PB|Psc-PC|posterior sex comb|posterior sex combs|posterior-sex-combs', map_location='2-67 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12965997, end_position_on_the_genomic_accession=12980994, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36432, CurrentID=0, Status='live', Symbol='Su(z)2', Aliases='Dmel_CG3905, Arp, CG3905, Dmel\\CG3905, SU(Z)2, SUZ2, Su(Z)2D, Su(z)3, Su(z)[[2]], Suz2, l(2)k06344, su(z)2, su(z)3, Su(z)2', description='Suppressor of zeste 2', other_designations='CG3905-PA|CG3905-PB|CG3905-PC|Su(z)2-PA|Su(z)2-PB|Su(z)2-PC|aristapedioid|aristopedoid|suppressor 2 of zeste|suppressor of zeste (2)|suppressor of zeste (3)|suppressor of zeste 2|suppressor of zeste 3|suppressor of zeste two|supressor of zeste 2', map_location='2-67 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12996501, end_position_on_the_genomic_accession=13007911, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36433, CurrentID=0, Status='live', Symbol='CG13323', Aliases='Dmel_ Dmel\\CG13323, FBgn0033788', description='CG13323 gene product from transcript CG13323-RB', other_designations='CG13323-PA|CG13323-PB|uncharacterized protein', map_location='49F1-49F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13044695, end_position_on_the_genomic_accession=13047256, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36434, CurrentID=0, Status='live', Symbol='CG13324', Aliases='Dmel_ Dmel\\CG13324', description='CG13324 gene product from transcript CG13324-RA', other_designations='CG13324-PA|uncharacterized protein', map_location='49F1-49F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13049146, end_position_on_the_genomic_accession=13049656, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36436, CurrentID=0, Status='live', Symbol='Drl-2', Aliases='Dmel_CG3915, CG12463, CG3915, DNT-like, DRL-2, Dmel\\CG3915, dnl, drl-2', description='Derailed 2', other_designations='CG3915-PB|CG3915-PC|Drl-2-PB|Drl-2-PC|derailed 2|derailed-2|doughnut-like', map_location='49F1-49F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13055246, end_position_on_the_genomic_accession=13079539, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36437, CurrentID=0, Status='live', Symbol='CG13325', Aliases='Dmel_ Dmel\\CG13325', description='CG13325 gene product from transcript CG13325-RB', other_designations='CG13325-PA|CG13325-PB|CG13325-PC|CG13325-PD|uncharacterized protein', map_location='49F2-49F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13110769, end_position_on_the_genomic_accession=13126014, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36438, CurrentID=0, Status='live', Symbol='CG3955', Aliases='Dmel_ Dmel\\CG3955', description='CG3955 gene product from transcript CG3955-RA', other_designations='CG3955-PA', map_location='49F2-49F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13127915, end_position_on_the_genomic_accession=13129052, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771991, CurrentID=0, Status='live', Symbol='His1:CG33855', Aliases='Dmel_CG33855, CG33855, Dmel\\CG33855', description='CG33855 gene product from transcript CG33855-RA', other_designations='CG33855-PA|His1:CG33855-PA|histone H1', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21521616, end_position_on_the_genomic_accession=21522638, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36440, CurrentID=0, Status='live', Symbol='Cap-G', Aliases='Dmel_CG34438, CAP-G, CG13327, CG17054, CG30057, CG34438, CapG, DCAP-G, DmCAP-G, Dmel\\CG34438, Dmel_CG17054, cap-G, cap-g, dCAP-G, dCap-G, dcap-g, l(2)49Ff, l(2)vr9, vr-9, vr9', description='Chromosome associated protein G', other_designations='CG34438-PF|CG34438-PG|Cap-G-PF|Cap-G-PG|chromosome associated protein G|lethal (2) 49Ff', map_location='49E7-49F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13133142, end_position_on_the_genomic_accession=13142820, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36442, CurrentID=0, Status='live', Symbol='Ack-like', Aliases='Dmel_CG43741, CG33007, CG3969, CG43741, DPR2, DRODPR2, DmHD-11, Dmel\\CG43741, Dmel_CG33007, Dmel_CG3969, HD-11, PR2, PR2/ACK, Rp2, dACK', description='Activated Cdc42 kinase-like', other_designations='Ack-like-PA|Ack-like-PC|Ack-like-PD|Ack-like-PE|CG43741-PA|CG43741-PC|CG43741-PD|CG43741-PE|Fak-like tyrosine kinase|activated Cdc42 kinase-like|tyrosine kinase', map_location='49F3-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13142834, end_position_on_the_genomic_accession=13152677, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36444, CurrentID=0, Status='live', Symbol='TppII', Aliases='Dmel_CG3991, CG3991, Dmel\\CG3991, TPP II, TPP2, TPPII, anon-WO0118547.127, dTPP II, dTPPII, l(2)k14301', description='tripeptidyl-peptidase II', other_designations='CG3991-PA|CG3991-PD|CG3991-PE|CG3991-PF|TppII-PA|TppII-PD|TppII-PE|TppII-PF|Tripeptidyl-peptidase II|tripeptidyl peptidase II|tripeptidylpeptidase II', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13152885, end_position_on_the_genomic_accession=13159141, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  36445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36445, CurrentID=0, Status='live', Symbol='Nrk', Aliases='Dmel_CG4007, CG4007, DNRK, DmHD-434, Dmel\\CG4007, Dnrk, HD-434, ROR2, anon-WO2004063362.79, l(2)k14301, nrk, nrtk_dros', description='Neurospecific receptor kinase', other_designations='CG4007-PA|Nrk-PA|drosophila neurospecific receptor kinase|neuronal receptor kinase|neurospecific receptor kinase|neurospecific receptor tyrosine kinase', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13160063, end_position_on_the_genomic_accession=13163484, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36446, CurrentID=0, Status='live', Symbol='CG15870', Aliases='Dmel_ Dmel\\CG15870', description='CG15870 gene product from transcript CG15870-RA', other_designations='CG15870-PA|uncharacterized protein', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13163363, end_position_on_the_genomic_accession=13163948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36447, CurrentID=0, Status='live', Symbol='GLaz', Aliases='Dmel_CG4604, CG4604, DGLaz, Dmel\\CG4604', description='Glial Lazarillo', other_designations='CG4604-PA|CG4604-PB|CG4604-PC|CG4604-PD|CG4604-PE|CG4604-PF|GLaz-PA|GLaz-PB|GLaz-PC|GLaz-PD|GLaz-PE|GLaz-PF|glial lazarillo', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13163959, end_position_on_the_genomic_accession=13168032, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36448, CurrentID=0, Status='live', Symbol='Spt-I', Aliases='Dmel_CG4016, CG4016, Dmel\\CG4016, SPT-I, Spt1, l(2)SH1626, l(2)SH2 1626, lcb1', description='Serine palmitoyltransferase subunit I', other_designations='CG4016-PA|CG4016-PB|CG4016-PC|Spt-I-PA|Spt-I-PB|Spt-I-PC|lethal (2) SH1626|serine palmitoyltransferase subunit I', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13168498, end_position_on_the_genomic_accession=13170744, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36449, CurrentID=0, Status='live', Symbol='CG33137', Aliases='Dmel_ CG4023, Dmel\\CG33137', description='CG33137 gene product from transcript CG33137-RB', other_designations='CG33137-PB|uncharacterized protein', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13175395, end_position_on_the_genomic_accession=13176075, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740475, CurrentID=0, Status='live', Symbol='CG41378', Aliases='Dmel_ Dmel\\CG41378', description='CG41378 gene product from transcript CG41378-RE', other_designations='CG41378-PB|CG41378-PC|CG41378-PD|CG41378-PE|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=2770380, end_position_on_the_genomic_accession=2846708, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3772004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772004, CurrentID=0, Status='live', Symbol='His1:CG33843', Aliases='Dmel_CG33843, CG33843, Dmel\\CG33843', description='CG33843 gene product from transcript CG33843-RA', other_designations='CG33843-PA|His1:CG33843-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21501633, end_position_on_the_genomic_accession=21502655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36454, CurrentID=0, Status='live', Symbol='lat', Aliases='Dmel_CG4088, CG4088, DmORC3, DmOrc3, Dmel\\CG4088, LAT, ORC, ORC3, Orc3, l(2)49Fd, l(2)vr6, orc3, vr-6, vr6', description='latheo', other_designations='CG4088-PA|CG4088-PB|lat-PA|lat-PB|origin recognition complex subunit 3', map_location='49F7-49F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13202728, end_position_on_the_genomic_accession=13206547, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36455, CurrentID=0, Status='live', Symbol='FLASH', Aliases='Dmel_CG4616, CG4616, Dmel\\CG4616, dFLASH', description='FLASH ortholog', other_designations='CG4616-PA|CG4616-PB|FLASH-PA|FLASH-PB', map_location='49F8-49F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13206745, end_position_on_the_genomic_accession=13210469, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36456, CurrentID=0, Status='live', Symbol='AQP', Aliases='Dmel_CG12251-dr, CG12251, Dmel\\CG12251, AQP', description='aquaporin', other_designations='AQP-PA|AQP-PB|CG12251-PA|CG12251-PB', map_location='49F9-49F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13212405, end_position_on_the_genomic_accession=13214496, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36457, CurrentID=0, Status='live', Symbol='CG4627', Aliases='Dmel_ Dmel\\CG4627', description='CG4627 gene product from transcript CG4627-RA', other_designations='CG4627-PA', map_location='49F10-49F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13214706, end_position_on_the_genomic_accession=13215760, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36458, CurrentID=0, Status='live', Symbol='CG4630', Aliases='Dmel_ Dmel\\CG4630', description='CG4630 gene product from transcript CG4630-RB', other_designations='CG4630-PA|CG4630-PB', map_location='49F10-49F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13215659, end_position_on_the_genomic_accession=13219247, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36459, CurrentID=0, Status='live', Symbol='CG4646', Aliases='Dmel_ BcDNA:SD14289, Dmel\\CG4646', description='CG4646 gene product from transcript CG4646-RB', other_designations='CG4646-PA|CG4646-PB', map_location='49F10-49F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13219396, end_position_on_the_genomic_accession=13220251, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36460, CurrentID=0, Status='live', Symbol='Fsn', Aliases='Dmel_CG4643, BEST:GH20696, CG4643, DFsn, Dmel\\CG4643, FBXO2, Wals, dFsn, fsn', description='F-box synaptic protein', other_designations='CG4643-PA|CG4643-PB|Fsn-PA|Fsn-PB|wallenstein', map_location='49F10-49F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13220939, end_position_on_the_genomic_accession=13222459, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36461, CurrentID=0, Status='live', Symbol='Dp', Aliases='Dmel_CG4654, 49Fk, CG4654, DDP1, DP, DP1, DmDP, Dmel\\CG46541, TFPD2, dDP, dDP1, dDp, dp, l(2)49Fk, l(2)vr10, vr10, Dp', description='DP transcription factor', other_designations='CG4654-PA|CG4654-PB|CG4654-PC|Dp-PA|Dp-PB|Dp-PC|dimerization partner|lethal(2)49Fk', map_location='49F10-49F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13222433, end_position_on_the_genomic_accession=13228202, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36462, CurrentID=0, Status='live', Symbol='Pex13', Aliases='Dmel_CG4663, CG4663, DmelPex13, Dmel\\CG4663, PEX13, pex13', description='Peroxin 13', other_designations='CG4663-PA|Pex13-PA|peroxin 13', map_location='49F11-49F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13228276, end_position_on_the_genomic_accession=13230234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772015, CurrentID=0, Status='live', Symbol='CG43164', Aliases='Dmel_ CG18431, CG33006, CG4844, CG4844a, CG4844b, Dmel\\CG43164, Dmel_CG18431, Dmel_CG4844, anon-EST:fe2D7, anon2D7', description='CG43164 gene product from transcript CG43164-RC', other_designations='CG43164-PA|CG43164-PB|CG43164-PC|CG43164-PD|CG43164-PE|anon-fast-evolving-2D7|uncharacterized protein', map_location='54B18-54C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17502354, end_position_on_the_genomic_accession=17508645, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36464, CurrentID=0, Status='live', Symbol='CG4670', Aliases='Dmel_ DmQSOX1, Dmel\\CG4670, QSOX1, lincRNA.300', description='CG4670 gene product from transcript CG4670-RB', other_designations='CG4670-PA|CG4670-PB', map_location='49F11-49F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13231758, end_position_on_the_genomic_accession=13235296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36465, CurrentID=0, Status='live', Symbol='CG4676', Aliases='Dmel_ Dmel\\CG4676', description='CG4676 gene product from transcript CG4676-RA', other_designations='CG4676-PA', map_location='49F11-49F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13235980, end_position_on_the_genomic_accession=13237055, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36466, CurrentID=0, Status='live', Symbol='CG4679', Aliases='Dmel_ Dmel\\CG4679, RE17985p', description='CG4679 gene product from transcript CG4679-RA', other_designations='CG4679-PA', map_location='49F11-49F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13237256, end_position_on_the_genomic_accession=13239468, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36467, CurrentID=0, Status='live', Symbol='GstE14', Aliases='Dmel_CG4688, CG4688, Dmel\\CG4688, GSTE14, gste14, nobo', description='Glutathione S transferase E14', other_designations='CG4688-PA|GstE14-PA|glutathione S transferase E14|noppera-bo', map_location='49F12-49F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13239729, end_position_on_the_genomic_accession=13241013, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36468, CurrentID=0, Status='live', Symbol='Mp20', Aliases='Dmel_CG4696, CG4696, DMMP20, Dmel\\CG4696, Dmmp20, MP20, Tpn, mp20', description='Muscle protein 20', other_designations='CG4696-PA|CG4696-PC|Mp20-PA|Mp20-PC|muscle protein 20|muscle protein 20 kd|myophilin', map_location='49F13-49F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13246670, end_position_on_the_genomic_accession=13248880, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36469, CurrentID=0, Status='live', Symbol='CG4712', Aliases='Dmel_ Dmel\\CG4712', description='CG4712 gene product from transcript CG4712-RA', other_designations='CG4712-PA|CG4712-PB', map_location='49F13-49F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13248894, end_position_on_the_genomic_accession=13251724, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36470, CurrentID=0, Status='live', Symbol='CG4714', Aliases='Dmel_ Dmel\\CG4714, anon-WO0140519.159', description='CG4714 gene product from transcript CG4714-RA', other_designations='CG4714-PA', map_location='49F13-49F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13251852, end_position_on_the_genomic_accession=13254337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36471, CurrentID=0, Status='live', Symbol='CG4716', Aliases='Dmel_ BEST:GH18008, Dmel\\CG4716, GH16, gh18008', description='CG4716 gene product from transcript CG4716-RC', other_designations='CG4716-PA|CG4716-PB|CG4716-PC', map_location='49F13-49F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13256587, end_position_on_the_genomic_accession=13257446, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36472, CurrentID=0, Status='live', Symbol='CG10799', Aliases='Dmel_ CT30256, Dmel\\CG10799', description='CG10799 gene product from transcript CG10799-RA', other_designations='CG10799-PA|uncharacterized protein', map_location='49F13-49F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13257886, end_position_on_the_genomic_accession=13258477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36473, CurrentID=0, Status='live', Symbol='CG42319', Aliases='Dmel_ CG17041, CG17042, Dmel\\CG42319, Dmel_CG17041, Dmel_CG17042', description='CG42319 gene product from transcript CG42319-RE', other_designations='CG42319-PA|CG42319-PB|CG42319-PC|CG42319-PD|CG42319-PE|CG42319-PF|uncharacterized protein', map_location='49F14-49F15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13260101, end_position_on_the_genomic_accession=13266336, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36475, CurrentID=0, Status='live', Symbol='Dh31-R', Aliases='Dmel_CG32843, CG17043, CG17415, CG32843, DH31, DH31-R, DH31-R1, DH31R1, DmeCG32843, Dmel\\CG32843, anon-WO0170980.136, anon-WO0170980.137, dh31-R1, Dh31-R', description='Diuretic hormone 31 Receptor', other_designations='CG32843-PA|CG32843-PB|CG32843-PC|Calcitonin-like diuretic hormone receptor|DH31 receptor|DH31-receptor|DH[[31]] receptor|Dh31-R-PA|Dh31-R-PB|Dh31-R-PC|diuretic hormone 31 receptor|diuretic hormone 31 receptor 1', map_location='49F15-50A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13267798, end_position_on_the_genomic_accession=13295927, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36477, CurrentID=0, Status='live', Symbol='CG4734', Aliases='Dmel_ Dmel\\CG4734', description='CG4734 gene product from transcript CG4734-RA', other_designations='CG4734-PA', map_location='50A1-50A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13297202, end_position_on_the_genomic_accession=13298467, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36478, CurrentID=0, Status='live', Symbol='CG17047', Aliases='Dmel_ Dmel\\CG17047', description='CG17047 gene product from transcript CG17047-RB', other_designations='CG17047-PA|CG17047-PB|uncharacterized protein', map_location='50A1-50A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13298995, end_position_on_the_genomic_accession=13301547, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36479, CurrentID=0, Status='live', Symbol='CG17048', Aliases='Dmel_ Dmel\\CG17048', description='CG17048 gene product from transcript CG17048-RB', other_designations='CG17048-PA|CG17048-PB|uncharacterized protein', map_location='50A3-50A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13353638, end_position_on_the_genomic_accession=13354499, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772032, CurrentID=0, Status='live', Symbol='His3:CG33839', Aliases='Dmel_CG33839, CG33839, Dmel\\CG33839', description='CG33839 gene product from transcript CG33839-RA', other_designations='CG33839-PA|His3:CG33839-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21494875, end_position_on_the_genomic_accession=21495403, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36481, CurrentID=0, Status='live', Symbol='CG10814', Aliases='Dmel_ Dmel\\CG10814', description='CG10814 gene product from transcript CG10814-RA', other_designations='CG10814-PA|uncharacterized protein', map_location='50A3-50A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13375975, end_position_on_the_genomic_accession=13377731, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36484, CurrentID=0, Status='live', Symbol='AttC', Aliases='Dmel_CG4740, Att, Attacin C, C, CG4740, DIM 16, DIM-16, Dmel\\CG4740, MPAC, att, att C, attC', description='Attacin-C', other_designations='AttC-PB|Attacin-C precursor|AttacinC|CG4740-PB|attacin|attacin C|attackin|drosophila immune induced molecule 16', map_location='50A3-50A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13393705, end_position_on_the_genomic_accession=13394667, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36485, CurrentID=0, Status='live', Symbol='CG4744', Aliases='Dmel_ Dmel\\CG4744', description='CG4744 gene product from transcript CG4744-RA', other_designations='CG4744-PA', map_location='50A3-50A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13394615, end_position_on_the_genomic_accession=13399368, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36486, CurrentID=0, Status='live', Symbol='IMPPP', Aliases='Dmel_CG18279, CG18279, DIM 10, DIM 12, DIM 13, DIM 20, DIM 24, DIM 5, DIM 6, DIM 8, DIM-10, DIM-12, DIM-13, DIM-24, DIM10, DIM12, DIM13, Dmel\\CG18279, IM10, IM20, IM5, IM6, IM8, im10', description='Immune induced molecule prepropeptide', other_designations='CG18279-PA|CG18279-PC|DIM 24 family precursor|IM10|IMPPP-PA|IMPPP-PC|Immune-induced peptides precursor|immune induced molecule 10|immune induced molecule 20|immune induced molecule 5|immune induced molecule 6|immune induced molecule 8|immune induced molecule prepropeptide|immune-induced molecule 10', map_location='50A5-50A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13407011, end_position_on_the_genomic_accession=13408358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36487, CurrentID=0, Status='live', Symbol='CG18278', Aliases='Dmel_ Dmel\\CG18278', description='CG18278 gene product from transcript CG18278-RA', other_designations='CG18278-PA|uncharacterized protein', map_location='50A5-50A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13408042, end_position_on_the_genomic_accession=13409887, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36488, CurrentID=0, Status='live', Symbol='CG42321', Aliases='Dmel_ CG17034, CG30061, DmCG42321, Dmel\\CG42321, Dmel_CG17034, Dmel_CG30061, anon-WO0153538.54', description='CG42321 gene product from transcript CG42321-RM', other_designations='CG42321-PE|CG42321-PG|CG42321-PH|CG42321-PI|CG42321-PJ|CG42321-PK|CG42321-PL|CG42321-PM|CG42321-PN|CG42321-PO|CG42321-PP|CG42321-PQ|CG42321-PR|CG42321-PS|uncharacterized protein', map_location='50A8-50A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13414978, end_position_on_the_genomic_accession=13439051, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  36489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36489, CurrentID=0, Status='live', Symbol='Ser8', Aliases='Dmel_CG4812, CG4812, Dmel\\CG4812, SP#3, SP3, SP90', description='CG4812 gene product from transcript CG4812-RA', other_designations='CG4812-PA|Ser8-PA|serine protease 3', map_location='50A8-50A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13434207, end_position_on_the_genomic_accession=13435045, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36491, CurrentID=0, Status='live', Symbol='cnn', Aliases='Dmel_CG4832, CG 4832, CG18370, CG4832, CNN, Cnn, Cnn1, Dmel\\CG4832, cen, mat(2)syn-C, mat(2)synHK21, mfs, obv', description='centrosomin', other_designations='CG4832-PA|CG4832-PB|CG4832-PD|CG4832-PE|CG4832-PG|CG4832-PH|CG4832-PI|CG4832-PJ|CG4832-PK|CG4832-PL|CG4832-PM|CG4832-PN|centrosmin|cnn-PA|cnn-PB|cnn-PD|cnn-PE|cnn-PG|cnn-PH|cnn-PI|cnn-PJ|cnn-PK|cnn-PL|cnn-PM|cnn-PN|maternal effect syncytial blastoderm arrest C|oblivious', map_location='2-65 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13439314, end_position_on_the_genomic_accession=13450475, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  36492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36492, CurrentID=0, Status='live', Symbol='cbs', Aliases='Dmel_CG4840, CG4840, Cbs, Dmel\\CG4840, dGolgin-97, dem', description="centrosomin's beautiful sister", other_designations="CG4840-PA|CG4840-PB|CG4840-PC|Centrosomin's beautiful sister|cbs-PA|cbs-PB|cbs-PC|dGolgin-97|dementin|golgin 97", map_location='50A9-50A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13450586, end_position_on_the_genomic_accession=13453393, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36494, CurrentID=0, Status='live', Symbol='cbc', Aliases='Dmel_CG5970, CG5970, Dmel\\CG5970, cg5970', description='crowded by cid', other_designations='CG5970-PA|cbc-PA', map_location='50A11-50A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13483512, end_position_on_the_genomic_accession=13485391, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36495, CurrentID=0, Status='live', Symbol='cid', Aliases='Dmel_CG13329, BcDNA:RE21270, CENH3, CENP-A, CENP-A/CID, CENP-A/Cid, CENPA, CG13329, CID, CID/CENP-A, CenH3, CenH3[CID], CenH3[Cid], Cenp-A, CenpA, CenpA/CID, Cid, Cid[Mel], Dmel\\CG13329, cenH3, cenpA', description='centromere identifier', other_designations='CG13329-PA|centromere identifier protein|centromere protein A|centromeric histone H3|cid-PA', map_location='50A11-50A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13485617, end_position_on_the_genomic_accession=13486631, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36496, CurrentID=0, Status='live', Symbol='bbc', Aliases='Dmel_CG6016, Bbc, CG6016, Dmel\\CG6016, cept, dCCS2', description='bb in a boxcar', other_designations='CDP-ethanolamine phosphotransferase|CG6016-PA|CG6016-PB|CG6016-PC|CG6016-PE|bbc-PA|bbc-PB|bbc-PC|bbc-PE', map_location='50A11-50A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13486634, end_position_on_the_genomic_accession=13494862, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36497, CurrentID=0, Status='live', Symbol='drk', Aliases='Dmel_CG6033, 24/1, CG6033, DRK, Dmel\\CG6033, Drk, Drk/Grb2, E(sev)2B, GRB2, Grb-2, Grb/drk, Grb2, Grb2/drk, P1112, Su(sev)R1, crkl, l(2)10626, l(2)k13809', description='downstream of receptor kinase', other_designations='CG6033-PA|CG6033-PB|CG6033-PC|CG6033-PD|CG6033-PF|downstream of receptor kinases|downstream receptor kinase|drk-PA|drk-PB|drk-PC|drk-PD|drk-PF|enhancer of sevenless-2B', map_location='2-70 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13495222, end_position_on_the_genomic_accession=13502841, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36498, CurrentID=0, Status='live', Symbol='mars', Aliases='Dmel_CG17064, CG17064, D-HURP, Dmel\\CG17064, HURP, Mars, dhrp/gkap, gkap', description='CG17064 gene product from transcript CG17064-RD', other_designations='CG17064-PA|CG17064-PB|CG17064-PC|CG17064-PD|mars-PA|mars-PB|mars-PC|mars-PD', map_location='50A13-50A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13500621, end_position_on_the_genomic_accession=13506011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36499, CurrentID=0, Status='live', Symbol='mip120', Aliases='Dmel_CG6061, CG6061, Dmel\\CG6061, MIP120, Mip120, l(3)L4569, p120', description='Myb-interacting protein 120', other_designations='CG6061-PA|CG6061-PB|mip120-PA|mip120-PB', map_location='50A14-50A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13506430, end_position_on_the_genomic_accession=13513112, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36501, CurrentID=0, Status='live', Symbol='CG13330', Aliases='Dmel_ Dmel\\CG13330', description='CG13330 gene product from transcript CG13330-RA', other_designations='CG13330-PA|uncharacterized protein', map_location='50A14-50A15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13515259, end_position_on_the_genomic_accession=13517452, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772054, CurrentID=0, Status='live', Symbol='CG33775', Aliases='Dmel_ BP1034, Dmel\\CG33775', description='CG33775 gene product from transcript CG33775-RA', other_designations='CG33775-PA|uncharacterized protein', map_location='49B3-49B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12564224, end_position_on_the_genomic_accession=12564901, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36503, CurrentID=0, Status='live', Symbol='CG13331', Aliases='Dmel_ Dmel\\CG13331', description='CG13331 gene product from transcript CG13331-RA', other_designations='CG13331-PA|uncharacterized protein', map_location='50A15-50A15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13521769, end_position_on_the_genomic_accession=13523375, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772056, CurrentID=0, Status='live', Symbol='His1:CG33801', Aliases='Dmel_CG33801, CG33801, Dmel\\CG33801', description='CG33801 gene product from transcript CG33801-RA', other_designations='CG33801-PA|His1:CG33801-PA|histone H1', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21416953, end_position_on_the_genomic_accession=21417975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36506, CurrentID=0, Status='live', Symbol='CG6145', Aliases='Dmel_ CT19307, Dmel\\CG6145', description='CG6145 gene product from transcript CG6145-RE', other_designations='CG6145-PA|CG6145-PB|CG6145-PC|CG6145-PE|CG6145-PF', map_location='50B1-50B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13537371, end_position_on_the_genomic_accession=13547577, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36507, CurrentID=0, Status='live', Symbol='CG33156', Aliases='Dmel_ CG30063, CG6152, Dmel\\CG33156', description='CG33156 gene product from transcript CG33156-RF', other_designations='CG33156-PA|CG33156-PB|CG33156-PC|CG33156-PE|CG33156-PF|uncharacterized protein', map_location='50B1-50B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13548244, end_position_on_the_genomic_accession=13553240, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36508, CurrentID=0, Status='live', Symbol='Roe1', Aliases='Dmel_CG6155, BcDNA:GM03203, CG6155, DROE1, Dmel\\CG6155, Droe1, cg6155, droe1', description='CG6155 gene product from transcript CG6155-RA', other_designations='CG6155-PA|Roe1-PA', map_location='50B1-50B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13554077, end_position_on_the_genomic_accession=13555606, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36509, CurrentID=0, Status='live', Symbol='link', Aliases='Dmel_CG13333, CG13333, Dmel\\CG13333', description='CG13333 gene product from transcript CG13333-RA', other_designations='CG13333-PA|link-PA', map_location='50B1-50B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13555949, end_position_on_the_genomic_accession=13557342, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36510, CurrentID=0, Status='live', Symbol='CG13334', Aliases='Dmel_ Dmel\\CG13334, NEST:bs27e02', description='CG13334 gene product from transcript CG13334-RB', other_designations='CG13334-PA|CG13334-PB|uncharacterized protein', map_location='50B1-50B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13558545, end_position_on_the_genomic_accession=13560768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36511, CurrentID=0, Status='live', Symbol='CG42807', Aliases='Dmel_ CG13335, CG6183, Dmel\\CG42807, Dmel_CG13335', description='CG42807 gene product from transcript CG42807-RB', other_designations='CG42807-PA|CG42807-PB|uncharacterized protein', map_location='50B2-50B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13570595, end_position_on_the_genomic_accession=13572759, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772064, CurrentID=0, Status='live', Symbol='DNApol-gamma35', Aliases='Dmel_CG33650, BG:DS00941.9, CG33084, CG33650, CG8969, D-pol gamma-beta, DNA pol-gam.35, DNApol-gamma, DNApol-gamma.35, Dmel\\CG33650, MtPolB, PolG, Polgamma, br16, l(2)34De, l(2)br16, pol gamma-beta, pol gamma35, polgamma, polgamma-beta', description='DNA polymerase gamma 35kD', other_designations='CG33650-PA|DNA polymerase -gamma|DNA polymerase gamma|DNApol-gamma35-PA|accessory subunit of mitochondrial DNA polymerase|accessory(beta)subunit|beta subunit|lethal 16 in the black-reduced region|lethal(2)34De|mitochondrial DNA polymerase accessory subunit|mitochondrial gamma polymerase', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13831482, end_position_on_the_genomic_accession=13833105, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36513, CurrentID=0, Status='live', Symbol='CG6191', Aliases='Dmel_ BcDNA:GH06193, Dmel\\CG6191, Dromel_CG6191_FBtr0087678_mORF', description='CG6191 gene product from transcript CG6191-RB', other_designations='CG6191-PA|CG6191-PB|CG6191-PC|cables', map_location='50B3-50B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13580313, end_position_on_the_genomic_accession=13593339, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36514, CurrentID=0, Status='live', Symbol='fand', Aliases='Dmel_CG6197, CG6197, Dmel\\CG6197, Prp5, XAB2, Xab2', description='fandango', other_designations='CG6197-PA|complementation group 7|fand-PA', map_location='50B4-50B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13593644, end_position_on_the_genomic_accession=13596557, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36515, CurrentID=0, Status='live', Symbol='S-Lap5', Aliases='Dmel_CG18369, CG18369, Dmel\\CG18369, S-LAP 5', description='Sperm-Leucylaminopeptidase 5', other_designations='CG18369-PA|S-Lap5-PA', map_location='50B4-50B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13596677, end_position_on_the_genomic_accession=13598750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36516, CurrentID=0, Status='live', Symbol='CG12464', Aliases='Dmel_ Dmel\\CG12464', description='CG12464 gene product from transcript CG12464-RA', other_designations='CG12464-PA|uncharacterized protein', map_location='50B5-50B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13608673, end_position_on_the_genomic_accession=13609178, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772069, CurrentID=0, Status='live', Symbol='mus201', Aliases='Dmel_CG10890, 55A11T, CG10890, CG32956, DmXPG, Dmel\\CG10890, ESTS:55A11T, XPG, XPG[DM], XPG[Dm], mus(2)201, mus-201, rad202, xpg/mus201', description='mutagen-sensitive 201', other_designations='CG10890-PC|CG10890-PD|mus201-PC|mus201-PD', map_location='29C4-29C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8438169, end_position_on_the_genomic_accession=8442342, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36518, CurrentID=0, Status='live', Symbol='CG13337', Aliases='Dmel_ Dmel\\CG13337', description='CG13337 gene product from transcript CG13337-RA', other_designations='CG13337-PA|CG13337-PB|uncharacterized protein', map_location='50C1-50C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13665114, end_position_on_the_genomic_accession=13667756, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36519, CurrentID=0, Status='live', Symbol='CG18368', Aliases='Dmel_ Dmel\\CG18368', description='CG18368 gene product from transcript CG18368-RA', other_designations='CG18368-PA|uncharacterized protein', map_location='50C1-50C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13667736, end_position_on_the_genomic_accession=13671594, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36520, CurrentID=0, Status='live', Symbol='CG6220', Aliases='Dmel_ Dmel\\CG6220', description='CG6220 gene product from transcript CG6220-RA', other_designations='CG6220-PA', map_location='50C2-50C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13691707, end_position_on_the_genomic_accession=13693369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36521, CurrentID=0, Status='live', Symbol='fas', Aliases='Dmel_CG17716, CG17716, CT39263, Dmel\\CG17716, l(2)05488, l(2)5488', description='faint sausage', other_designations='CG17716-PA|CG17716-PB|fas-PA|fas-PB', map_location='2-68 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13623044, end_position_on_the_genomic_accession=13749367, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  36522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36522, CurrentID=0, Status='live', Symbol='CG6280', Aliases='Dmel_ Dmel\\CG6280', description='CG6280 gene product from transcript CG6280-RA', other_designations='CG6280-PA', map_location='50C3-50C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13753964, end_position_on_the_genomic_accession=13758971, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3772075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772075, CurrentID=0, Status='live', Symbol='His1:CG33831', Aliases='Dmel_CG33831, CG33831, Dmel\\CG33831', description='CG33831 gene product from transcript CG33831-RA', other_designations='CG33831-PA|His1:CG33831-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21481345, end_position_on_the_genomic_accession=21482367, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36524, CurrentID=0, Status='live', Symbol='S-Lap7', Aliases='Dmel_CG13340, CG13340, Dmel\\CG13340, S-LAP 7', description='Sperm-Leucylaminopeptidase 7', other_designations='CG13340-PA|CG13340-PB|S-Lap7-PA|S-Lap7-PB|Sperm-LeucylAminoPeptidase 7', map_location='50C5-50C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13778124, end_position_on_the_genomic_accession=13780245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36525, CurrentID=0, Status='live', Symbol='Cpr50Cb', Aliases='Dmel_CG6305, CG6305, DmelCpr50Cb, Dmel\\CG6305', description='Cuticular protein 50Cb', other_designations='CG6305-PA|Cpr50Cb-PA|cuticular protein 50Cb', map_location='50C5-50C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13794777, end_position_on_the_genomic_accession=13805231, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36526, CurrentID=0, Status='live', Symbol='stj', Aliases='Dmel_CG12295, CG12295, Dmel\\CG12295, alpha[[2]]delta, alpha[[2]]delta-3, dalpha[[2]]delta-3, l(2)k10814', description='straightjacket', other_designations='CG12295-PB|CG12295-PC|CG12295-PD|lethal (2) k10814|stj-PB|stj-PC|stj-PD', map_location='50C5-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13805794, end_position_on_the_genomic_accession=13819571, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  36527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36527, CurrentID=0, Status='live', Symbol='fl(2)d', Aliases='Dmel_CG6315, CG6315, Dmel\\CG6315, v(2)k16105', description='female lethal d', other_designations='CG6315-PA|CG6315-PB|CG6315-PC|CG6315-PD|female-lethal(2)d|fl(2)d-PA|fl(2)d-PB|fl(2)d-PC|fl(2)d-PD', map_location='2-70.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13819624, end_position_on_the_genomic_accession=13824553, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3772080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772080, CurrentID=0, Status='live', Symbol='CG12788', Aliases='Dmel_ CG33066, Dmel\\CG12788, Dromel_CG12788_CG17767_FBtr0074_tim9b_mORF, PSTK', description='CG12788 gene product from transcript CG12788-RC', other_designations='CG12788-PA|CG12788-PC|CG12788-PE|uncharacterized protein', map_location='18D3-18D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19616663, end_position_on_the_genomic_accession=19618537, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772081, CurrentID=0, Status='live', Symbol='His2B:CG33870', Aliases='Dmel_CG33870, CG33870, Dmel\\CG33870', description='CG33870 gene product from transcript CG33870-RA', other_designations='CG33870-PA|His2B:CG33870-PA|histone H2B', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21532461, end_position_on_the_genomic_accession=21533055, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772082, CurrentID=0, Status='live', Symbol='kay', Aliases='Dmel_CG33956, AP-1, AP1, CG15507, CG15509, CG33956, D-Fos, D-fos, D-fos/kay, DFOS, DFos, DFra, DM3, DM30, Dfos, Dmel\\CG33956, FOS, Fos, Fra, KAY, Kay, c-Fos, c-fos, d-fos, dAP-1, dFRA, dFos, dFra, dfos, dlhD, fbz, fos, fra', description='kayak', other_designations='CG33956-PA|CG33956-PB|CG33956-PD|CG33956-PF|CG33956-PG|Dfos/kayak|Fos|Fos-related antigen|activator protein 1|dorsal holes D|kay-PA|kay-PB|kay-PD|kay-PF|kay-PG', map_location='3-99 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29765995, end_position_on_the_genomic_accession=29794113, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36531, CurrentID=0, Status='live', Symbol='CG6347', Aliases='Dmel_ Dmel\\CG6347', description='CG6347 gene product from transcript CG6347-RA', other_designations='CG6347-PA', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13836419, end_position_on_the_genomic_accession=13839363, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36532, CurrentID=0, Status='live', Symbol='CG6357', Aliases='Dmel_ Dmel\\CG6357', description='CG6357 gene product from transcript CG6357-RA', other_designations='CG6357-PA', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13840053, end_position_on_the_genomic_accession=13841743, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36533, CurrentID=0, Status='live', Symbol='Syngr', Aliases='Dmel_CG10808, CG10808, Dmel\\CG10808, Gyr, gyr, sng-1, synaptogyrin', description='Synaptogyrin', other_designations='CG10808-PA|Syngr-PA|synaptogyrin', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13841761, end_position_on_the_genomic_accession=13846255, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36534, CurrentID=0, Status='live', Symbol='CG30484', Aliases='Dmel_ CG13342, Dmel\\CG30484', description='CG30484 gene product from transcript CG30484-RB', other_designations='CG30484-PB|uncharacterized protein', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13846393, end_position_on_the_genomic_accession=13849832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740489, CurrentID=0, Status='live', Symbol='CG34430', Aliases='Dmel_ CG14766, Dmel\\CG34430, Dmel_CG14766', description='CG34430 gene product from transcript CG34430-RA', other_designations='CG34430-PA|uncharacterized protein', map_location='43F5-43F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7943442, end_position_on_the_genomic_accession=7944475, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36536, CurrentID=0, Status='live', Symbol='CG6543', Aliases='Dmel_ CG18645, Dmel\\CG6543, Echs1', description='CG6543 gene product from transcript CG6543-RA', other_designations='CG6543-PA|CG6543-PB', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13852947, end_position_on_the_genomic_accession=13854331, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36537, CurrentID=0, Status='live', Symbol='CG6553', Aliases='Dmel_ Dmel\\CG6553', description='CG6553 gene product from transcript CG6553-RA', other_designations='CG6553-PA', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13854578, end_position_on_the_genomic_accession=13855778, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36538, CurrentID=0, Status='live', Symbol='tum', Aliases='Dmel_CG13345, 13345, AcGAP, CG13345, DRacGAP, Dmel\\CG13345, RacGAP, RacGAP50C, RacGAP50c, RacGap, RacGap50C, RhoGAP-50C14, TUM, acGAP, eon, i249, racGAP50, racGAP50C, racGAP50c', description='tumbleweed', other_designations='CG13345-PA|RacGAP50C|enhancer of naked|tum-PA', map_location='49C-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13856150, end_position_on_the_genomic_accession=13858689, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36539, CurrentID=0, Status='live', Symbol='Uba3', Aliases='Dmel_CG13343, CG13343, Dmel\\CG13343, dUba3, l(2)SH2 2028, l(2)SH2028', description='Ubiquitin activating enzyme 3', other_designations='CG13343-PA|Uba3-PA|human Uba3 homologue|lethal (2) SH2028|ubiquitin activating enzyme 3', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13858952, end_position_on_the_genomic_accession=13860432, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36540, CurrentID=0, Status='live', Symbol='CG16935', Aliases='Dmel_ DM_7303260, Dmel\\CG16935', description='CG16935 gene product from transcript CG16935-RA', other_designations='CG16935-PA|CG16935-PB|uncharacterized protein', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13860349, end_position_on_the_genomic_accession=13862002, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36541, CurrentID=0, Status='live', Symbol='CG13344', Aliases='Dmel_ Dmel\\CG13344, anon-EST:GreesD1, anon-EST:GressD1', description='CG13344 gene product from transcript CG13344-RA', other_designations='CG13344-PA|CG13344-PC|uncharacterized protein', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13861997, end_position_on_the_genomic_accession=13864156, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36542, CurrentID=0, Status='live', Symbol='shot', Aliases='Dmel_CG18076, 42/4, CG18076, CG18637, Dmel\\CG18076, Grv, Shot, Su(Mir)1, grv, kak, kop, l(2)CA4, l(2)k03010, l(2)k04204, l(2)k05434, l(2)k05821, l(2)k10821, l(2)k15606, shortstop/kakapo', description='short stop', other_designations='CG18076-PA|CG18076-PAA|CG18076-PAB|CG18076-PAC|CG18076-PAD|CG18076-PB|CG18076-PC|CG18076-PE|CG18076-PG|CG18076-PH|CG18076-PI|CG18076-PJ|CG18076-PK|CG18076-PL|CG18076-PM|CG18076-PN|CG18076-PO|CG18076-PP|CG18076-PQ|CG18076-PX|CG18076-PY|CG18076-PZ|groovin|kakapo|kakapoo|kopupu|short-stop|shortshop|shortstop|shot-PA|shot-PAA|shot-PAB|shot-PAC|shot-PAD|shot-PB|shot-PC|shot-PE|shot-PG|shot-PH|shot-PI|shot-PJ|shot-PK|shot-PL|shot-PM|shot-PN|shot-PO|shot-PP|shot-PQ|shot-PX|shot-PY|shot-PZ|spectraplakin', map_location='50C6-50C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13864237, end_position_on_the_genomic_accession=13942110, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  36543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36543, CurrentID=0, Status='live', Symbol='DJ-1alpha', Aliases='Dmel_CG6646, CG6646, DJ-1, DJ-1A, DJ-1a, Dmel\\CG6646', description='CG6646 gene product from transcript CG6646-RA', other_designations='CG6646-PA|DJ-1A|DJ-1alpha-PA', map_location='50C9-50C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13942534, end_position_on_the_genomic_accession=13943377, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36544, CurrentID=0, Status='live', Symbol='AGO1', Aliases='Dmel_CG6671, AGO-1, Ago, Ago-1, Ago1, Agp1, CG6671, Dm Ago1, DmAgo1, Dmel\\CG6671, MRE20, ago1, ago1-1, anon-WO0257455.29, dAGO1, dAgo1, l(2)04845, l(2)4845, l(2)k00208, l(2)k08121', description='Argonaute-1', other_designations='AGO1-PA|AGO1-PB|AGO1-PC|AGO1-PD|Argonaute1|CG6671-PA|CG6671-PB|CG6671-PC|CG6671-PD|argonaute|argonaute 1|argonaute protein 1|argonaute-1|mRNA-like ncRNA in embryogenesis 20', map_location='50C9-50C17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13943387, end_position_on_the_genomic_accession=13958089, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36545, CurrentID=0, Status='live', Symbol='Rpn13', Aliases='Dmel_CG13349, CG13349, Dmel\\CG13349, GC13349, p42E', description='Regulatory particle non-ATPase 13', other_designations='CG13349-PA|CG13349-PB|CG13349-PD|CG13349-PE|CG13349-PF|CG13349-PG|Rpn13-PA|Rpn13-PB|Rpn13-PD|Rpn13-PE|Rpn13-PF|Rpn13-PG|regulatory particle non-ATPase 13', map_location='50C17-50C18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13958545, end_position_on_the_genomic_accession=13960404, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36546, CurrentID=0, Status='live', Symbol='Cp1', Aliases='Dmel_CG6692, CATL_DROME, CG6692, CL, CP1, CT20780, CTSL, CatL, Cath L, CysP-1, DCP1, Dcp-1, Dcp1, Dmel\\CG6692, cp1, fs(2)50Ca', description='Cysteine proteinase-1', other_designations='CG6692-PA|CG6692-PC|CG6692-PD|CG6692-PE|Cp1-PA|Cp1-PC|Cp1-PD|Cp1-PE|L-like cysteine protease|cathepsin L|cysteine proteinase 1|cysteine proteinase-1|drosophila cysteine proteinase-1', map_location='50C18-50C20', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13960675, end_position_on_the_genomic_accession=13967981, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36547, CurrentID=0, Status='live', Symbol='Aats-phe-m', Aliases='Dmel_CG13348, Aats-phe, CG13348, Dmel\\CG13348, FRS, PheRS, Pts, mtPheRS', description='mitochondrial phenylalanyl-tRNA synthetase', other_designations='Aats-phe-m-PA|CG13348-PA|Phenylalanyl-tRNA synthetase', map_location='50C20-50C20', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13968308, end_position_on_the_genomic_accession=13969902, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772100, CurrentID=0, Status='live', Symbol='Pmi', Aliases='Dmel_CG33718, CG32912, CG33718, CG5523, Dmel\\CG33718, PMI', description='CG33718 gene product from transcript CG33718-RC', other_designations='CG33718-PC|CG33718-PD|CG33718-PE|Pmi-PC|Pmi-PD|Pmi-PE', map_location='64E7-64E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5780051, end_position_on_the_genomic_accession=5784685, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3772101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772101, CurrentID=0, Status='live', Symbol='Ext2', Aliases='Dmel_CG8433, BcDNA:GH02288, CG33038, CG8433, DEXT2, D Dmel\\CG8433, Sotv, dExt2, sotv', description='CG8433 gene product from transcript CG8433-RB', other_designations='CG8433-PB|Ext2-PB|sister of tout-velu|sister-of-tout-velu', map_location='52E11-52E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16145904, end_position_on_the_genomic_accession=16148719, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36550, CurrentID=0, Status='live', Symbol='CG6701', Aliases='Dmel_ Dmel\\CG6701, EP(2)2054, EP2054', description='CG6701 gene product from transcript CG6701-RD', other_designations='CG6701-PB|CG6701-PC|CG6701-PD|CG6701-PE', map_location='50C22-50C23', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13972984, end_position_on_the_genomic_accession=13980165, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36551, CurrentID=0, Status='live', Symbol='Ctf4', Aliases='Dmel_CG13350, CG13350, Dmel\\CG13350, anon-WO0140519.180', description='CG13350 gene product from transcript CG13350-RA', other_designations='CG13350-PA|Ctf4-PA', map_location='50C23-50C23', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13980362, end_position_on_the_genomic_accession=13983317, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36552, CurrentID=0, Status='live', Symbol='CG8067', Aliases='Dmel_ Dmel\\CG8067', description='CG8067 gene product from transcript CG8067-RB', other_designations='CG8067-PA|CG8067-PB', map_location='50C23-50C23', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13983395, end_position_on_the_genomic_accession=13984893, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772105, CurrentID=0, Status='live', Symbol='CG9130', Aliases='Dmel_ BEST:GH19274, BEST:GH19464, CG32332, Dmel\\CG9130, anon-WO0140519.145, anon-WO0140519.53', description='CG9130 gene product from transcript CG9130-RB', other_designations='CG9130-PA|CG9130-PB', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1252012, end_position_on_the_genomic_accession=1255986, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36554, CurrentID=0, Status='live', Symbol='RN-tre', Aliases='Dmel_CG8085, CG13351, CG8085, DRN-tre, Dmel\\CG8085', description='Related to the N terminus of tre oncogene', other_designations='CG8085-PA|CG8085-PB|CG8085-PC|CG8085-PE|CG8085-PF|RN-tre-PA|RN-tre-PB|RN-tre-PC|RN-tre-PE|RN-tre-PF|TRE oncoprotein related|related to the N terminus of tre oncogene|tre oncogene-related protein', map_location='50C23-50C23', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13984815, end_position_on_the_genomic_accession=13990707, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36555, CurrentID=0, Status='live', Symbol='mam', Aliases='Dmel_CG8118, CG8118, Dm0064, Dmel\\CG8118, MAM, Mam, N-2G, SR2-4, dMam, l(2)03505, l(2)04615(CAG)[[7]], mam', description='mastermind', other_designations='CG8118-PA|CG8118-PB|CG8118-PC|CG8118-PD|mam-PA|mam-PB|mam-PC|mam-PD|master mind', map_location='2-70.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13991203, end_position_on_the_genomic_accession=14060356, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3772108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772108, CurrentID=0, Status='live', Symbol='CG33798', Aliases='Dmel_ BP1039, Dmel\\CG33798', description='CG33798 gene product from transcript CG33798-RA', other_designations='CG33798-PA|uncharacterized protein', map_location='49E7-49E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12993897, end_position_on_the_genomic_accession=12994622, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772109, CurrentID=0, Status='live', Symbol='Meltrin', Aliases='Dmel_CG7649, CG31314, CG31385, CG7649, CT23341, D Dmel\\CG7649, FBgn0051314, Neu3, meltrin', description='CG7649 gene product from transcript CG7649-RC', other_designations='CG7649-PC|CG7649-PD|DMeltrin|Meltrin-PC|Meltrin-PD|Meltrin-like|Neu3|Neuroectoderm-expressed 3|dMeltrin-alpha|meltrin', map_location='88C10-88D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14664907, end_position_on_the_genomic_accession=14697292, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36559, CurrentID=0, Status='live', Symbol='CG42288', Aliases='Dmel_ CG13353, Dmel\\CG42288, Dmel_CG13353', description='CG42288 gene product from transcript CG42288-RA', other_designations='CG42288-PA|CG42288-PB|uncharacterized protein', map_location='50D5-50D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14091031, end_position_on_the_genomic_accession=14091814, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36560, CurrentID=0, Status='live', Symbol='Rcd1', Aliases='Dmel_CG8233, CG8233, Dmel\\CG8233, NSL3, Nsl3, cg8233', description='Reduction in Cnn dots 1', other_designations='CG8233-PA|CG8233-PB|CG8233-PC|CG8233-PD|CG8233-PE|CG8233-PF|CG8233-PG|CG8233-PH|Rcd1-PA|Rcd1-PB|Rcd1-PC|Rcd1-PD|Rcd1-PE|Rcd1-PF|Rcd1-PG|Rcd1-PH|non-specific lethal 3|reduction in Cnn dots 1', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14148263, end_position_on_the_genomic_accession=14153123, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36561, CurrentID=0, Status='live', Symbol='pea', Aliases='Dmel_CG8241, CG8241, Dmel\\CG8241, Prp22, cg8241, prp22', description='peanuts', other_designations='CG8241-PA|pea-PA', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14153153, end_position_on_the_genomic_accession=14157106, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36562, CurrentID=0, Status='live', Symbol='CG13016', Aliases='Dmel_ Dmel\\CG13016', description='CG13016 gene product from transcript CG13016-RA', other_designations='CG13016-PA|uncharacterized protein', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14157762, end_position_on_the_genomic_accession=14159434, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36563, CurrentID=0, Status='live', Symbol='CG8257', Aliases='Dmel_ Dmel\\CG8257', description='CG8257 gene product from transcript CG8257-RB', other_designations='CG8257-PA|CG8257-PB', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14159096, end_position_on_the_genomic_accession=14161262, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36564, CurrentID=0, Status='live', Symbol='O-fut1', Aliases='Dmel_CG12366, 4R6, AAF58290.1, CG12366, Dmel\\CG12366, O-FUT1, O-fut, OFUT1, OFut1, Ofut-1, Ofut1, l(2)SH2 2260, l(2)SH2260, ntc, nti', description='O-fucosyltransferase 1', other_designations='CG12366-PA|CG12366-PB|O-fucosyl transferase|O-fucosyltransferase|O-fut1-PA|O-fut1-PB|neurotic', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14161430, end_position_on_the_genomic_accession=14163054, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36565, CurrentID=0, Status='live', Symbol='Tango7', Aliases='Dmel_CG8309, CG8309, Dmel\\CG8309, TANGO7, Tango 7, Tng7, anon-EST:Posey67, anon-EST:Posey76, eIF3M, eIF3m, tango7', description='Transport and Golgi organization 7', other_designations='CG8309-PA|CG8309-PB|Tango7-PA|Tango7-PB|transport and golgi organization 7', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14163066, end_position_on_the_genomic_accession=14164825, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36566, CurrentID=0, Status='live', Symbol='CG8323', Aliases='Dmel_ Dmel\\CG8323', description='CG8323 gene product from transcript CG8323-RA', other_designations='CG8323-PA', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14165051, end_position_on_the_genomic_accession=14166991, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36567, CurrentID=0, Status='live', Symbol='CG18327', Aliases='Dmel_ Dmel\\CG18327', description='CG18327 gene product from transcript CG18327-RA', other_designations='CG18327-PA|CG18327-PB|uncharacterized protein', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14166982, end_position_on_the_genomic_accession=14168233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36568, CurrentID=0, Status='live', Symbol='CG18324', Aliases='Dmel_ Dmel\\CG18324', description='CG18324 gene product from transcript CG18324-RA', other_designations='CG18324-PA|CG18324-PB|uncharacterized protein', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14168786, end_position_on_the_genomic_accession=14170233, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36569, CurrentID=0, Status='live', Symbol='CG8331', Aliases='Dmel_ CT24597, Dmel\\CG8331', description='CG8331 gene product from transcript CG8331-RD', other_designations='CG8331-PA|CG8331-PB|CG8331-PC|CG8331-PD', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14170462, end_position_on_the_genomic_accession=14172998, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36570, CurrentID=0, Status='live', Symbol='mRpS16', Aliases='Dmel_CG8338, 147189_at, BcDNA:SD15809, CG8338, Dmel\\CG8338, MRP-S16', description='mitochondrial ribosomal protein S16', other_designations='CG8338-PA|mRpS16-PA', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14172986, end_position_on_the_genomic_accession=14173596, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36571, CurrentID=0, Status='live', Symbol='cg', Aliases='Dmel_CG8367, CG8367, Dmel\\CG8367, l(2)07659, l(2)k11504, mig', description='combgap', other_designations='CG8367-PA|CG8367-PB|CG8367-PC|CG8367-PD|CG8367-PF|CG8367-PI|CG8367-PJ|CG8367-PK|CG8367-PL|CG8367-PM|CG8367-PN|cg-PA|cg-PB|cg-PC|cg-PD|cg-PF|cg-PI|cg-PJ|cg-PK|cg-PL|cg-PM|cg-PN|comb gap|migraine', map_location='2-71.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14173693, end_position_on_the_genomic_accession=14182838, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  36573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36573, CurrentID=0, Status='live', Symbol='CG30069', Aliases='Dmel_ CG13017, CG8381, Dmel\\CG30069', description='CG30069 gene product from transcript CG30069-RA', other_designations='CG30069-PA|uncharacterized protein', map_location='50E1-50E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14185204, end_position_on_the_genomic_accession=14198368, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36574, CurrentID=0, Status='live', Symbol='VGAT', Aliases='Dmel_CG8394, CG8394, CG8394.2, Dmel\\CG8394, VIAAT, vGAT', description='Vesicular GABA Transporter', other_designations='CG8394-PA|VGAT-PA|vesicular GABA transporter', map_location='50E4-50E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14198478, end_position_on_the_genomic_accession=14201249, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36575, CurrentID=0, Status='live', Symbol='Sox15', Aliases='Dmel_CG8404, CG13945, CG8404, DSox-15, Dmel\\CG8404, Dsox-15, SOX15, SOX50E, SOXF, Sox F, Sox50E, SoxE, SoxF', description='Sox box protein 15', other_designations='CG8404-PA|Sox15-PA', map_location='50E4-50E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14201390, end_position_on_the_genomic_accession=14212413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36576, CurrentID=0, Status='live', Symbol='RpS23', Aliases='Dmel_CG8415, CG8415, CT24703, Dmel\\CG8415, RPS23, Rp S23, anon-EST:Posey158, anon-EST:Posey63', description='Ribosomal protein S23', other_designations='CG8415-PA|CG8415-PB|RpS23-PA|RpS23-PB|ribosomal protein S23', map_location='50E4-50E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14215096, end_position_on_the_genomic_accession=14216627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36577, CurrentID=0, Status='live', Symbol='CG8468', Aliases='Dmel_ Dmel\\CG8468', description='CG8468 gene product from transcript CG8468-RB', other_designations='CG8468-PA|CG8468-PB|CG8468-PC', map_location='50E4-50E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14219898, end_position_on_the_genomic_accession=14230620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36578, CurrentID=0, Status='live', Symbol='Opa1', Aliases='Dmel_CG8479, CG8479, Dmel\\CG8479, OPA1, OPA1-like-like, dOPA1, dOpa1, l(2)s3475, opa1, opa1-like, Opa1', description='Optic atrophy 1 ortholog (H. sapiens)', other_designations='CG8479-PA|CG8479-PB|CG8479-PC|CG8479-PD|CG8479-PE|Opa1-PA|Opa1-PB|Opa1-PC|Opa1-PD|Opa1-PE|Opa1-like|dOpa1|drosophila optic atrophy 1-like|lethal (2) s3475|opa1|optic atrophy 1|optic atrophy 1 ortholog|optic atrophy 1-like', map_location='50E4-50E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14230611, end_position_on_the_genomic_accession=14235449, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  36579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36579, CurrentID=0, Status='live', Symbol='CG8485', Aliases='Dmel_ 152964_at, Dmel\\CG8485, SNRK', description='CG8485 gene product from transcript CG8485-RA', other_designations='CG8485-PA|CG8485-PB|CG8485-PC', map_location='50E4-50E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14235787, end_position_on_the_genomic_accession=14239703, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36580, CurrentID=0, Status='live', Symbol='Usp20-33', Aliases='Dmel_CG8494, CG8494, Dmel\\CG8494, Q7JW61, USP20, Usp20, Usp33', description='Ubiquitin specific protease 20/33', other_designations='CG8494-PA|Usp20-33-PA|ubiquitin specific protease 20/33', map_location='50E4-50E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14239932, end_position_on_the_genomic_accession=14243655, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3346149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346149, CurrentID=0, Status='live', Symbol='upd3', Aliases='Dmel_CG33542, CG15062, CG15062/CG5963, CG33542, CG5963, Dmel\\CG33542, UPD3, Upd-3, Upd-like, Upd3, cg15062, cg5963, od, odsy, os, sy, upd-3', description='unpaired 3', other_designations='CG33542-PB|Unpaired3|outstretched|outstretched-smalleye|small-eye|unpaired|unpaired 3beta|unpaired-3|unpaired-like protein 3|upd homolog|upd3-PB', map_location='17A4-17A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18277233, end_position_on_the_genomic_accession=18284596, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36582, CurrentID=0, Status='live', Symbol='Mdr50', Aliases='Dmel_CG8523, CG8523, Dm2, Dm3, Dmel\\CG8523, MDR50, mdr50', description='Multi drug resistance 50', other_designations='CG8523-PA|Mdr50-PA|multi drug resistance 50|multi-drug resistance 50', map_location='2-50 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14246966, end_position_on_the_genomic_accession=14252597, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36583, CurrentID=0, Status='live', Symbol='Hsc70-5', Aliases='Dmel_CG8542, CG8542, Dmel\\CG8542, HSC-70, HSC5, HSC70, HSC70-5, Hsc-5, Hsc5, Hsc70, Hsp-c5, dhsc70, hsc-70-5, hsc5, hsc70, hsc70-5, l(2)k04907', description='Heat shock protein cognate 5', other_designations='CG8542-PA|CG8542-PB|Hsc70-5-PA|Hsc70-5-PB|Hsp cognate 5|heat shock 70 kDa protein cognate 5|heat shock protein cognate 5|mortalin', map_location='50E6-50E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14252598, end_position_on_the_genomic_accession=14256192, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36584, CurrentID=0, Status='live', Symbol='CG8531', Aliases='Dmel_ Dmel\\CG8531', description='CG8531 gene product from transcript CG8531-RA', other_designations='CG8531-PA', map_location='50E6-50E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14256493, end_position_on_the_genomic_accession=14258787, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36585, CurrentID=0, Status='live', Symbol='beta4GalNAcTA', Aliases='Dmel_CG8536, BcDNA:GH13356, CG8536, DmGalT, Dmel\\CG8536, anon-WO0118547.128, b4GalNAcTA, beta4Gal-NAcTA, beta4GalNAcT1, beta4GalNacTA, beta4GalTA, dbeta4GalTA', description='CG8536 gene product from transcript CG8536-RA', other_designations='CG8536-PA|beta1,4-N-acetylgalactosaminyltransferase|beta4GalNAcTA-PA', map_location='50E6-50E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14259030, end_position_on_the_genomic_accession=14261150, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36586, CurrentID=0, Status='live', Symbol='CG8547', Aliases='Dmel_ Dmel\\CG8547', description='CG8547 gene product from transcript CG8547-RE', other_designations='CG8547-PC|CG8547-PD|CG8547-PE', map_location='50E6-50E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14261206, end_position_on_the_genomic_accession=14267919, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36587, CurrentID=0, Status='live', Symbol='SelD', Aliases='Dmel_CG8553, 142926_at, CG8553, Dm SelD, Dm-SelD, Dmel\\CG8553, SPS1, SeS, anon-WO0172774.145, anon-WO0172774.146, dSPS1, dsps1, l(2)SH1599, l(2)SH2 1599, l(2)k12303, ptf1, ptuf, ptuf1, selD, selD/sps1', description='Selenide,water dikinase', other_designations='CG8553-PA|CG8553-PB|SelD-PA|SelD-PB|lethal (2) SH1599|patufet|selD-like|selenide,water dikinase|selenophosphate synthetase|selenophosphate synthetase 1', map_location='50E8-50E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14268772, end_position_on_the_genomic_accession=14270658, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36588, CurrentID=0, Status='live', Symbol='conv', Aliases='Dmel_CG8561, ALS, Als, CG8561, CT24963, Dmel\\CG8561, dALS, l(2)k06507b, tartan/capricious-like', description='convoluted', other_designations='CG8561-PA|CG8561-PB|acid labile substrate|acid labile subunit|conv-PA|conv-PB', map_location='50E6-50E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14270729, end_position_on_the_genomic_accession=14275495, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36589, CurrentID=0, Status='live', Symbol='Ih', Aliases='Dmel_CG8585, BcDNA:GH25778, CG8585, CR30068, DMIH, DmIh, Dmel\\CG8585, Dmel_CR30068, I[[h]]', description='I[[h]] channel', other_designations='CG8585-PB|CG8585-PC|CG8585-PE|CG8585-PF|CG8585-PG|CG8585-PH|CG8585-PI|CG8585-PJ|CG8585-PK|CG8585-PL|CG8585-PM|Ih channel|Ih-PB|Ih-PC|Ih-PE|Ih-PF|Ih-PG|Ih-PH|Ih-PI|Ih-PJ|Ih-PK|Ih-PL|Ih-PM|Ih-channel|drosophila melanogaster I[[h]]-channel|putative Ih-channel|putative-voltage-and-cyclic-nucleotide-gated-ion-channel', map_location='50E6-50F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14276130, end_position_on_the_genomic_accession=14300095, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  36590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36590, CurrentID=0, Status='live', Symbol='tej', Aliases='Dmel_CG8589, CG8589, Dmel\\CG8589, Tej, anon-EST:fe1D11', description='tejas', other_designations='CG8589-PA|anon-fast-evolving-1D11|tej-PA', map_location='50F1-50F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14300103, end_position_on_the_genomic_accession=14302353, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36592, CurrentID=0, Status='live', Symbol='Shroom', Aliases='Dmel_CG34379, CG13942, CG13944, CG34379, CG8603, Dmel\\CG34379, Dmel_CG13942, Dmel_CG8603, Shrm, dShrm, dm', description='CG34379 gene product from transcript CG34379-RH', other_designations='CG34379-PD|CG34379-PF|CG34379-PG|CG34379-PH|Shroom-PD|Shroom-PF|Shroom-PG|Shroom-PH|shroom', map_location='50F4-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14308142, end_position_on_the_genomic_accession=14351802, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  3772145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772145, CurrentID=0, Status='live', Symbol='CheA86a', Aliases='Dmel_CG33698, CG33698, Dmel\\CG33698', description='Chemosensory protein A 86a', other_designations='CG33698-PA|CheA86a-PA|chemosensory protein A 86a', map_location='86D4-86D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11144346, end_position_on_the_genomic_accession=11145235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3346162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346162, CurrentID=0, Status='live', Symbol='Gfrl', Aliases='Dmel_CG43224, BcDNA:GH26447, CG17203, CG17204, CG17205, CG17208, CG33518, CG34118, CG34119, CG42401, CG42625, CG42801, CG43224, CG4342, DmGfrl, Dmel\\CG43224, Dmel_CG17208, Dmel_CG33518, Dmel_CG34118, Dmel_CG34119, Dmel_CG42401, Dmel_CG42625, Dmel_CG42801, Dmel_CG4342, mun', description='Glial cell line-derived neurotrophic family receptor-like', other_designations='CG43224-PE|CG43224-PF|CG43224-PG|CG43224-PH|CG43224-PI|CG43224-PJ|CG43224-PK|CG43224-PL|CG43224-PM|CG43224-PN|GDNF Receptor-like|Gfrl-PE|Gfrl-PF|Gfrl-PG|Gfrl-PH|Gfrl-PI|Gfrl-PJ|Gfrl-PK|Gfrl-PL|Gfrl-PM|Gfrl-PN|glial cell line-derived neurotrophic family receptor-like|munin', map_location='92E8-92E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20378458, end_position_on_the_genomic_accession=20484588, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  36595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36595, CurrentID=0, Status='live', Symbol='Arc1', Aliases='Dmel_CG12505, Arc, CG12505, Dmel\\CG12505, anon-WO0153538.31, anon-WO0153538.40, anon-WO0153538.41, anon-WO0153538.43, anon-WO0153538.45, dARC1, dArc1', description='Activity-regulated cytoskeleton associated protein 1', other_designations='Arc1-PA|CG12505-PA|dArc1', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14357927, end_position_on_the_genomic_accession=14360297, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772148, CurrentID=0, Status='live', Symbol='His2A:CG33850', Aliases='Dmel_CG33850, CG33850, Dmel\\CG33850', description='CG33850 gene product from transcript CG33850-RA', other_designations='CG33850-PA|His2A:CG33850-PA|histone H2A', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21513610, end_position_on_the_genomic_accession=21514102, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36597, CurrentID=0, Status='live', Symbol='Arc2', Aliases='Dmel_CG13941, BcDNA:SD11637, CG13941, Dmel\\CG13941, anon-WO0118547.456, dARC2, d', description='CG13941 gene product from transcript CG13941-RA', other_designations='Arc2-PA|CG13941-PA', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14361924, end_position_on_the_genomic_accession=14362825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36598, CurrentID=0, Status='live', Symbol='Tfb1', Aliases='Dmel_CG8151, CG8151, Dmel\\CG8151, Dmp62, TFB1, TfBl, l(2)06949, l(2)06950, p62', description='TFB1 ortholog (S. cerevisiae)', other_designations='CG8151-PA|CG8151-PB|CG8151-PC|CG8151-PD|TFB1 ortholog|TfBl|Tfb1-PA|Tfb1-PB|Tfb1-PC|Tfb1-PD', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14363048, end_position_on_the_genomic_accession=14365473, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3346167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346167, CurrentID=0, Status='live', Symbol='btsz', Aliases='Dmel_CG44012, BcDNA:GH06647, Btsz, CG14858, CG17304, CG18280, CG31306, CG33555, CG44012, CG7343, Dmel\\CG44012, Dmel_CG17304, Dmel_CG33555, Granulophilin, l(3)10418', description='bitesize', other_designations='CG44012-PB|CG44012-PC|CG44012-PD|CG44012-PE|CG44012-PF|CG44012-PG|CG44012-PH|CG44012-PI|CG44012-PJ|bite size|btsz-PB|btsz-PC|btsz-PD|btsz-PE|btsz-PF|btsz-PG|btsz-PH|btsz-PI|btsz-PJ|granulophilin', map_location='88D5-88D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14804092, end_position_on_the_genomic_accession=14875488, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  36600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36600, CurrentID=0, Status='live', Symbol='Obp50e', Aliases='Dmel_CG13939, CG13939, DmelObp50e, Dmel\\CG13939', description='Odorant-binding protein 50e', other_designations='CG13939-PA|Obp50e-PA', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14374462, end_position_on_the_genomic_accession=14375385, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36601, CurrentID=0, Status='live', Symbol='Dh44-R1', Aliases='Dmel_CG8422, CG8422, DH-R, DH44-R1, DH[[44]]-R1, DmeCG8422, Dmel\\CG8422, anon-WO0170980.100, anon-WO0170980.101', description='Diuretic hormone 44 receptor 1', other_designations='CG8422-PA|CRF-like diuretic hormone receptor 1|DH44 receptor|DH44-receptor|DH[[44]] receptor|Dh44-R1-PA|diuretic hormone 44 receptor 1', map_location='50F8-50F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14378097, end_position_on_the_genomic_accession=14385359, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  3772154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772154, CurrentID=0, Status='live', Symbol='Acp33A', Aliases='Dmel_CG6555, Acp33a, CG15841, CG32952, CG32952-A, CG6555, Dmel\\CG6555', description='Accessory gland-specific peptide 33A', other_designations='Acp33A-PA|CG6555-PA|accessory gland protein 33A|accessory gland-specific peptide 33A', map_location='33A-33A', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11591339, end_position_on_the_genomic_accession=11591678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36603, CurrentID=0, Status='live', Symbol='CG17385', Aliases='Dmel_ Dmel\\CG17385', description='CG17385 gene product from transcript CG17385-RB', other_designations='CG17385-PA|CG17385-PB|uncharacterized protein', map_location='51A1-51A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14409295, end_position_on_the_genomic_accession=14411572, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36604, CurrentID=0, Status='live', Symbol='Sin1', Aliases='Dmel_CG10105, CG10105, Dmel\\CG10105, dSin1, sin1', description='SAPK-interacting protein 1', other_designations='CG10105-PA|Sin1-PA|Stress-activated protein kinase [SAPK]-interacting protein|stress-activated map kinase-interacting protein 1', map_location='51A2-51A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14411825, end_position_on_the_genomic_accession=14413851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36605, CurrentID=0, Status='live', Symbol='CG17386', Aliases='Dmel_ CT33484, Dmel\\CG17386', description='CG17386 gene product from transcript CG17386-RB', other_designations='CG17386-PA|CG17386-PB|uncharacterized protein', map_location='51A2-51A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14420324, end_position_on_the_genomic_accession=14424764, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36606, CurrentID=0, Status='live', Symbol='phyl', Aliases='Dmel_CG10108, CG10108, D(DmcycE[JP])2.2, Dmel\\CG10108, PHYL, Phyl, SR2-3, Su(Raf)2C, phy1', description='phyllopod', other_designations='CG10108-PA|phyl-PA', map_location='51A2-51A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14427336, end_position_on_the_genomic_accession=14432337, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772159, CurrentID=0, Status='live', Symbol='CG33934', Aliases='Dmel_ CG32921, CG32921-ORFB, CG4961, Dmel\\CG33934, anon-WO0140519.217', description='CG33934 gene product from transcript CG33934-RE', other_designations='CG33934-PC|CG33934-PD|CG33934-PE|uncharacterized protein', map_location='92C3-92C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20042757, end_position_on_the_genomic_accession=20048716, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  3346176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346176, CurrentID=0, Status='live', Symbol='U3-55K', Aliases='Dmel_CG33505, CG10220, CG33505, Dmel\\CG33505', description='U3 small nuclear riboprotein factor 55K', other_designations='CG33505-PA|U3-55K-PA', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14858845, end_position_on_the_genomic_accession=14860758, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36609, CurrentID=0, Status='live', Symbol='Oaz', Aliases='Dmel_CG42702, CG17388, CG17389, CG17390, CG42702, Dm_2R:42259, Dmel\\CG42702, Dmel_CG17388, Dmel_CG17389, Dmel_CG17390, dmOAZ, oaz', description='O/E-associated zinc finger protein', other_designations='CG42702-PB|CG42702-PC|Oaz-PB|Oaz-PC', map_location='51A4-51A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14435899, end_position_on_the_genomic_accession=14465881, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3772163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772163, CurrentID=0, Status='live', Symbol='His3:CG33809', Aliases='Dmel_CG33809, CG33809, Dmel\\CG33809', description='CG33809 gene product from transcript CG33809-RA', other_designations='CG33809-PA|His3:CG33809-PA|histone H3', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21435299, end_position_on_the_genomic_accession=21435827, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36612, CurrentID=0, Status='live', Symbol='Asx', Aliases='Dmel_CG8787, ASX, CG8787, Dmel\\CG8787, FBgn0261823, xen', description='Additional sex combs', other_designations='Asx-PA|Asx-PB|Asx-PC|Asx-PD|CG8787-PA|CG8787-PB|CG8787-PC|CG8787-PD|additional Sex combs|additional sex comb|additional sex combs|additionnal sex comb|xenicid', map_location='2-72 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14503958, end_position_on_the_genomic_accession=14512493, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  36613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36613, CurrentID=0, Status='live', Symbol='Cpr51A', Aliases='Dmel_CG10112, CG10112, Dmel\\CG10112', description='Cuticular protein 51A', other_designations='CG10112-PA|Cpr51A-PA|cuticular protein 51A', map_location='51A6-51A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14512559, end_position_on_the_genomic_accession=14515211, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36614, CurrentID=0, Status='live', Symbol='ttv', Aliases='Dmel_CG10117, CG10117, DEXT1, Dmel\\CG10117, Ext1, P1.15, Ttv, dExt1, l(2)00681, l(2)05282, l(2)SH0631, l(2)SH2 0631, l(2)k00115, l(2)k03617', description='tout-velu', other_designations='CG10117-PA|CG10117-PB|CG10117-PC|lethal (2) SH0631|tout velu|tout-velo|toutvelu|ttv-PA|ttv-PB|ttv-PC', map_location='51B1-51B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14526267, end_position_on_the_genomic_accession=14587917, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  36615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36615, CurrentID=0, Status='live', Symbol='LamC', Aliases='Dmel_CG10119, CG10119, DmC, Dmel\\CG10119, G-IF, LAMC, lamC, pG-IF', description='Lamin C', other_designations='CG10119-PA|CG10119-PB|G intermediate filament|LamC-PA|LamC-PB|LaminC|lamin A|lamin C|lamin DmC', map_location='51B1-51B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14570810, end_position_on_the_genomic_accession=14575771, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36616, CurrentID=0, Status='live', Symbol='CG12869', Aliases='Dmel_ Dmel\\CG12869', description='CG12869 gene product from transcript CG12869-RA', other_designations='CG12869-PA|CG12869-PB|CG12869-PC|uncharacterized protein', map_location='51B4-51B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14591203, end_position_on_the_genomic_accession=14595617, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36617, CurrentID=0, Status='live', Symbol='RpI1', Aliases='Dmel_CG10122, 153129_at, CG10122, DmRPA1, Dmel\\CG10122, RPA1', description='RNA polymerase I subunit', other_designations='CG10122-PA|RNA polymerase I|RpI1-PA', map_location='51B4-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14595611, end_position_on_the_genomic_accession=14601216, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36619, CurrentID=0, Status='live', Symbol='tra2', Aliases='Dmel_CG10128, CG10128, Dmel\\CG10128, TRA 2, TRA-2, TRA2, Tra-2, Tra2, tra-2', description='transformer 2', other_designations='CG10128-PA|CG10128-PB|CG10128-PC|CG10128-PD|CG10128-PE|CG10128-PF|CG10128-PG|Transformer-2|tra2-PA|tra2-PB|tra2-PC|tra2-PD|tra2-PE|tra2-PF|tra2-PG|transformer2', map_location='2-70 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14602004, end_position_on_the_genomic_accession=14604337, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3346188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346188, CurrentID=0, Status='live', Symbol='Vkor', Aliases='Dmel_CG33544, BP1071, CG33544, Dmel\\CG33544, HDC06808, VKORC1', description='Vitamin-K epoxide reductase', other_designations='CG33544-PA|Vkor-PA', map_location='53D8-53D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16778365, end_position_on_the_genomic_accession=16778994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36621, CurrentID=0, Status='live', Symbol='Su(var)2-HP2', Aliases='Dmel_CG12864, CG12864, Dmel\\CG12864, HP2, Hp2, SU(VAR)2-HP2', description='CG12864 gene product from transcript CG12864-RB', other_designations='CG12864-PA|CG12864-PB|Su(var)2-HP2-PA|Su(var)2-HP2-PB|heterochromatic protein 2|heterochromatin protein 2', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14607154, end_position_on_the_genomic_accession=14618118, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36623, CurrentID=0, Status='live', Symbol='CG12863', Aliases='Dmel_ Dmel\\CG12863', description='CG12863 gene product from transcript CG12863-RA', other_designations='CG12863-PA|uncharacterized protein', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14619798, end_position_on_the_genomic_accession=14621320, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3346192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346192, CurrentID=0, Status='live', Symbol='Vmat', Aliases='Dmel_CG33528, CG33528, CG6119, CG6139, DVMAT, DVMAT-A, DVMAT-B, Dmel\\CG33528, VMAT, dVMAT, l(2)SH0459, l(2)SH2 0459, vmat', description='Vesicular monoamine transporter', other_designations='CG33528-PC|CG33528-PD|CG33528-PE|CG33528-PF|Vmat-PC|Vmat-PD|Vmat-PE|Vmat-PF|drosophila vesicular monoamine transporter|lethal (2) SH0459|vesicle transporter|vesicular monoamine transporter', map_location='50B2-50B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13513132, end_position_on_the_genomic_accession=13534251, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36625, CurrentID=0, Status='live', Symbol='CG12862', Aliases='Dmel_ Dmel\\CG12862', description='CG12862 gene product from transcript CG12862-RA', other_designations='CG12862-PA|uncharacterized protein', map_location='51B7-51B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14640214, end_position_on_the_genomic_accession=14640626, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36626, CurrentID=0, Status='live', Symbol='CG10139', Aliases='Dmel_ Dmel\\CG10139', description='CG10139 gene product from transcript CG10139-RB', other_designations='CG10139-PA|CG10139-PB|uncharacterized protein', map_location='51B7-51B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14645270, end_position_on_the_genomic_accession=14646344, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36627, CurrentID=0, Status='live', Symbol='Adgf-E', Aliases='Dmel_CG10143, ADGF-E, CG10143, Dmel\\CG10143, adgfe', description='Adenosine deaminase-related growth factor E', other_designations='Adgf-E-PA|CG10143-PA|adenosine deaminase-related growth factor E', map_location='51B7-51B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14656798, end_position_on_the_genomic_accession=14658755, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772180, CurrentID=0, Status='live', Symbol='Sxl', Aliases='Dmel_CG43770, CG14425, CG14425/CG18350, CG18350, CG33070, CG43770, DmSxl, Dmel\\CG43770, Dmel_CG14425, Dmel_CG18350, Fl, Mex156, ORF2, SXL, Su(da), Sx1, SxL-1 Sxl-2, SxlPe, SxlPm, dSXL, fs(1)K1274, fs(1)M106, l(1)6Fa, melSxl, sxl, Sxl', description='Sex lethal', other_designations='CG43770-PA|CG43770-PAA|CG43770-PAB|CG43770-PAC|CG43770-PB|CG43770-PC|CG43770-PD|CG43770-PE|CG43770-PF|CG43770-PG|CG43770-PH|CG43770-PJ|CG43770-PK|CG43770-PL|CG43770-PM|CG43770-PN|CG43770-PO|CG43770-PP|CG43770-PQ|CG43770-PR|CG43770-PT|CG43770-PW|CG43770-PX|CG43770-PY|CG43770-PZ|Sex-kethal|Sxl-PA|Sxl-PAA|Sxl-PAB|Sxl-PAC|Sxl-PB|Sxl-PC|Sxl-PD|Sxl-PE|Sxl-PF|Sxl-PG|Sxl-PH|Sxl-PJ|Sxl-PK|Sxl-PL|Sxl-PM|Sxl-PN|Sxl-PO|Sxl-PP|Sxl-PQ|Sxl-PR|Sxl-PT|Sxl-PW|Sxl-PX|Sxl-PY|Sxl-PZ|female lethal|sex-lethal|sexlethal', map_location='1-19.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7074550, end_position_on_the_genomic_accession=7098053, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  3772181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772181, CurrentID=0, Status='live', Symbol='CG33784', Aliases='Dmel_ Dmel\\CG33784, HDC12790', description='CG33784 gene product from transcript CG33784-RA', other_designations='CG33784-PA|uncharacterized protein', map_location='85F9-85F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10067905, end_position_on_the_genomic_accession=10068572, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36630, CurrentID=0, Status='live', Symbol='CG12866', Aliases='Dmel_ Dmel\\CG12866', description='CG12866 gene product from transcript CG12866-RE', other_designations='CG12866-PB|CG12866-PC|CG12866-PD|CG12866-PE|uncharacterized protein', map_location='51B9-51B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14668391, end_position_on_the_genomic_accession=14671675, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3772183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772183, CurrentID=0, Status='live', Symbol='CG10096', Aliases='Dmel_ CG32919, Dmel\\CG10096, anon-WO0140519.58', description='CG10096 gene product from transcript CG10096-RC', other_designations='CG10096-PA|CG10096-PB|CG10096-PC|CG10096-PD|uncharacterized protein', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12390166, end_position_on_the_genomic_accession=12395980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36632, CurrentID=0, Status='live', Symbol='mspo', Aliases='Dmel_CG10145, CG10145, CT28539, Dmel\\CG10145, Mspo, m-spo', description='M-spondin', other_designations='CG10145-PA|CG10145-PB|m-spondin|mspo-PA|mspo-PB', map_location='51B9-51C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14671396, end_position_on_the_genomic_accession=14713622, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3346201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346201, CurrentID=0, Status='live', Symbol='Unc-89', Aliases='Dmel_CG33519, BEST:HL01080, CG18019, CG18020, CG18021, CG30171, CG30175, CG33519, CG3901, CT12925, CT40322, CT40326, Dmel\\CG33519, UNC-89, obs, unc-89', description='CG33519 gene product from transcript CG33519-RH', other_designations='CG33519-PC|CG33519-PD|CG33519-PE|CG33519-PF|CG33519-PG|CG33519-PH|PROJECTIN-like|Unc-89-PC|Unc-89-PD|Unc-89-PE|Unc-89-PF|Unc-89-PG|Unc-89-PH|obscurin', map_location='60B1-60B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23995543, end_position_on_the_genomic_accession=24014254, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  3346202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346202, CurrentID=0, Status='live', Symbol='SIFa', Aliases='Dmel_CG33527, CG33527, Dmel\\CG33527, Drm-IFa, Drm-SIFa, IFAMIDE, IFa, Neb-LFamide, SIFR', description='SIFamide', other_designations='AYRKPPFNGSLFamide|CG33527-PB|IFamide|Neb-LFamide|SIFa-PB|neuropeptide AYRKPPFNGSIFamide|neuropeptide IFamide', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24577566, end_position_on_the_genomic_accession=24578143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36635, CurrentID=0, Status='live', Symbol='Dro', Aliases='Dmel_CG10816, 143609_at, BcDNA:RH31634, CG10816, DIM 11, DIM 15, DIM 9, DIM-15, DROC, Dmel\\CG10816, Drc1, Dros, Drs, IM15, drc, dro, dros, drosocin, Dro', description='Drosocin', other_designations='CG10816-PA|CG10816-PB|Dro-PA|Dro-PB|drosocin|drosocin precursor|drosophila immune induced molecule 11|drosophila immune induced molecule 15|drosophila immune induced molecule 9', map_location='51C1-51C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14745961, end_position_on_the_genomic_accession=14746714, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36636, CurrentID=0, Status='live', Symbol='AttA', Aliases='Dmel_CG10146, ATT, Att, Att A, Att-A, BcDNA:LP05763, CG10146, Dmel\\CG10146, anon-WO0140519.5, att, att A, attA', description='Attacin-A', other_designations='AttA-PA|Attacin-A precursor|AttacinA|CG10146-PA|attacin|attacin A|attackin', map_location='51C1-51C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14747362, end_position_on_the_genomic_accession=14748215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36637, CurrentID=0, Status='live', Symbol='AttB', Aliases='Dmel_CG18372, Att, Att B1, AttB2, AttBn, Attb, CG18372, Dmel\\CG18372, att, att B1, attB, AttB', description='Attacin-B', other_designations='AttB-PA|AttB-PB|Attacin-B precursor|AttacinB|CG18372-PA|CG18372-PB|attacin|attacin B|attacin B1|attackin', map_location='51C1-51C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14749223, end_position_on_the_genomic_accession=14750165, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36638, CurrentID=0, Status='live', Symbol='Rpn6', Aliases='Dmel_CG10149, BcDNA.LD18931, BcDNA:LD18931, CG10149, CG5289, Dmel\\CG10149, Pros44.5, S9, dPSMD11, l(2)k00103, p42B, rpn6', description='Regulatory particle non-ATPase 6', other_designations='CG10149-PA|CG10149-PB|CG10149-PC|Rpn6-PA|Rpn6-PB|Rpn6-PC|proteasome p44.5 subunit|regulatory particle non-ATPase 6', map_location='51C1-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14750166, end_position_on_the_genomic_accession=14752910, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3772191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772191, CurrentID=0, Status='live', Symbol='His3:CG33854', Aliases='Dmel_CG33854, CG33854, Dmel\\CG33854', description='CG33854 gene product from transcript CG33854-RA', other_designations='CG33854-PA|His3:CG33854-PA|histone H3', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21520109, end_position_on_the_genomic_accession=21520637, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36640, CurrentID=0, Status='live', Symbol='ND-B15', Aliases='Dmel_CG12859, CG 12859, CG12859, Dmel\\CG12859, NB5M, anon-EST:fe3B12', description='NADH dehydrogenase (ubiquinone) B15 subunit', other_designations='CG12859-PA|NADH ubiquinone oxidoreductase B15|NADH-ubiquinone oxidoreductase B15 subunit|NADH:ubiquinone oxidoreductase B15 subunit|ND-B15-PA|anon-fast-evolving-3B12', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14756516, end_position_on_the_genomic_accession=14757072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3346209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346209, CurrentID=0, Status='live', Symbol='Ddr', Aliases='Dmel_CG33531, CG11573, CG13984, CG13985, CG33531, CG9488, CG9490, CT33540, DDR, Dmel\\CG33531, Dmel_CG13984', description='Discoidin domain receptor', other_designations='CG33531-PD|CG33531-PF|CG33531-PG|Ddr-PD|Ddr-PF|Ddr-PG|discoidin domain receptor', map_location='26C1-26C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6253123, end_position_on_the_genomic_accession=6323466, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  36642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36642, CurrentID=0, Status='live', Symbol='CG12857', Aliases='Dmel_ Dmel\\CG12857', description='CG12857 gene product from transcript CG12857-RB', other_designations='CG12857-PA|CG12857-PB|uncharacterized protein', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14757907, end_position_on_the_genomic_accession=14759602, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36643, CurrentID=0, Status='live', Symbol='Spred', Aliases='Dmel_CG10155, AE33, CG10155, Dm AE33, Dmel\\CG10155, dSPRED, spred', description='Sprouty-related protein with EVH-1 domain', other_designations='CG10155-PA|CG10155-PB|CG10155-PC|Spred-PA|Spred-PB|Spred-PC', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14759646, end_position_on_the_genomic_accession=14769423, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3772196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772196, CurrentID=0, Status='live', Symbol='CG18170', Aliases='Dmel_ CG32316, Dmel\\CG18170', description='CG18170 gene product from transcript CG18170-RA', other_designations='CG18170-PA|CG18170-PB|uncharacterized protein', map_location='62A7-62A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1602954, end_position_on_the_genomic_accession=1617355, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36645, CurrentID=0, Status='live', Symbol='BEAF-32', Aliases='Dmel_CG10159, BEAFA, BEAF-32B, BEAF32, BEAF32A, BEAF32B, BEAF32b, Beaf-32, CG10159, Dmel\\CG10159, beaf, bs04c01.y1, BEAF-32', description='Boundary element-associated factor of 32kD', other_designations='BEAF-32-PA|BEAF-32-PB|BEAF-32-PC|CG10159-PA|CG10159-PB|CG10159-PC|boundary element associated factor|boundary element associated factor 32|boundary element associated factor of 32|boundary element-associated factor|boundary element-associated factor of 32kD', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14770469, end_position_on_the_genomic_accession=14772630, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772198, CurrentID=0, Status='live', Symbol='His3:CG33818', Aliases='Dmel_CG33818, CG33818, Dmel\\CG33818', description='CG33818 gene product from transcript CG33818-RA', other_designations='CG33818-PA|His3:CG33818-PA|histone H3', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21459447, end_position_on_the_genomic_accession=21459975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36648, CurrentID=0, Status='live', Symbol='hui', Aliases='Dmel_CG10200, CG10200, Dmel\\CG10200', description='hase und igel', other_designations='CG10200-PA|CG10200-PB|hare and hedgehog|hui-PA|hui-PB', map_location='51C3-51C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14808419, end_position_on_the_genomic_accession=14809931, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36649, CurrentID=0, Status='live', Symbol='CG10202', Aliases='Dmel_ Dmel\\CG10202, anon-WO0140519.148', description='CG10202 gene product from transcript CG10202-RA', other_designations='CG10202-PA|uncharacterized protein', map_location='51C4-51C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14817857, end_position_on_the_genomic_accession=14819987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36650, CurrentID=0, Status='live', Symbol='CG10205', Aliases='Dmel_ CT28709, Dmel\\CG10205', description='CG10205 gene product from transcript CG10205-RA', other_designations='CG10205-PA|CG10205-PB|uncharacterized protein', map_location='51C4-51C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14832832, end_position_on_the_genomic_accession=14833951, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36651, CurrentID=0, Status='live', Symbol='NaPi-T', Aliases='Dmel_CG10207, CG10207, Dmel\\CG10207, Dropo, MFS2, Na/Pi-cotransporter, NapiT', description='Na[+]-dependent inorganic phosphate cotransporter', other_designations='CG10207-PA|CG10207-PB|NaPi-T-PA|NaPi-T-PB', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14842720, end_position_on_the_genomic_accession=14845874, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36652, CurrentID=0, Status='live', Symbol='CG10209', Aliases='Dmel_ Dmel\\CG10209', description='CG10209 gene product from transcript CG10209-RA', other_designations='CG10209-PA|CG10209-PB|uncharacterized protein', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14846064, end_position_on_the_genomic_accession=14848366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36653, CurrentID=0, Status='live', Symbol='SMC2', Aliases='Dmel_CG10212, AAF58197, CG10212, DmSMC2, Dmel\\CG10212, Smc2, dSMC2, smc-2', description='Structural maintenance of chromosomes 2', other_designations='CG10212-PA|SMC2-PA|structural maintenance of chromosomes 2', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14848621, end_position_on_the_genomic_accession=14852653, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36654, CurrentID=0, Status='live', Symbol='Ercc1', Aliases='Dmel_CG10215, CG10215, Dmel\\CG10215, ERCC1', description='CG10215 gene product from transcript CG10215-RA', other_designations='CG10215-PA|Ercc1-PA', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14852941, end_position_on_the_genomic_accession=14853948, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36655, CurrentID=0, Status='live', Symbol='Ciao1', Aliases='Dmel_CG12797, CG12797, CiaO1, Dmel\\CG12797', description='CG12797 gene product from transcript CG12797-RA', other_designations='CG12797-PA|Ciao1-PA', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14853863, end_position_on_the_genomic_accession=14855048, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36656, CurrentID=0, Status='live', Symbol='HPS1', Aliases='Dmel_CG12855, CG12855, Dmel\\CG12855, HPS, NP_610997, dHPS1', description='Hermansky-Pudlak Syndrome 1 ortholog (H. sapiens)', other_designations='CG12855-PA|HPS|HPS1-PA|Hermansky-Pudlak syndrome 1 ortholog', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14855429, end_position_on_the_genomic_accession=14857502, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3346225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346225, CurrentID=0, Status='live', Symbol='CG33509', Aliases='Dmel_ CG14398, Dmel\\CG33509', description='CG33509 gene product from transcript CG33509-RA', other_designations='CG33509-PA|uncharacterized protein', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21084605, end_position_on_the_genomic_accession=21086506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3346226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346226, CurrentID=0, Status='live', Symbol='ppk13', Aliases='Dmel_CG33508, CG14398, CG33508, Dmel\\CG33508, PPK13', description='pickpocket 13', other_designations='CG33508-PA|ppk13-PA', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21086790, end_position_on_the_genomic_accession=21088751, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3346227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346227, CurrentID=0, Status='live', Symbol='dpr2', Aliases='Dmel_CG33507, BP1042, CG14067, CG14068, CG33507, CT33638, Dmel\\CG33507, Dpr-2, Dpr2', description='defective proboscis extension response 2', other_designations='BP1042|CG33507-PF|CG33507-PG|dpr2-PF|dpr2-PG', map_location='32B4-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10917015, end_position_on_the_genomic_accession=10964153, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36661, CurrentID=0, Status='live', Symbol='Cyp6a23', Aliases='Dmel_CG10242, 6a23, CG10242, Cypa23, Dmel\\CG10242, cyp6a23', description='CG10242 gene product from transcript CG10242-RA', other_designations='CG10242-PA|Cyp6a23-PA|cytochrome P450', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14875829, end_position_on_the_genomic_accession=14877654, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36662, CurrentID=0, Status='live', Symbol='Cyp6a19', Aliases='Dmel_CG10243, 6a19, CG10243, Dmel\\CG10243', description='CG10243 gene product from transcript CG10243-RB', other_designations='CG10243-PA|CG10243-PB|Cyp6a19-PA|Cyp6a19-PB', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14877786, end_position_on_the_genomic_accession=14879752, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36663, CurrentID=0, Status='live', Symbol='Cyp6a9', Aliases='Dmel_CG10246, 6a9, CG10246, CYP6A9, Cyt-P450-rAF4, Dmel\\CG10246, P-450, P450', description='Cytochrome P450-6a9', other_designations='CG10246-PB|Cyp6a9-PB|cytochrome P450 related AF4|cytochrome P450-6a9', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14879356, end_position_on_the_genomic_accession=14881372, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36664, CurrentID=0, Status='live', Symbol='Cyp6a20', Aliases='Dmel_CG10245, 6a20, CG10245, Dmel\\CG10245, p450', description='CG10245 gene product from transcript CG10245-RB', other_designations='CG10245-PB|Cyp6a20-PB', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14882118, end_position_on_the_genomic_accession=14884046, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36665, CurrentID=0, Status='live', Symbol='Cyp6a21', Aliases='Dmel_CG10247, 6a21, CG10247, Dmel\\CG10247, cyp6a21', description='CG10247 gene product from transcript CG10247-RA', other_designations='CG10247-PA|Cyp6a21-PA|cytochrome P450|cytochrome P450 6a21', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14885264, end_position_on_the_genomic_accession=14886947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36666, CurrentID=0, Status='live', Symbol='Cyp6a8', Aliases='Dmel_CG10248, 6a8, CG10248, CYP6A8, Cyt-P450-rAF5, Dmel\\CG10248, P-450, P450', description='Cytochrome P450-6a8', other_designations='CG10248-PA|Cyp6a8-PA|cytochrome P450|cytochrome P450 related AF5|cytochrome P450-6a8', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14887171, end_position_on_the_genomic_accession=14889010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3346235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346235, CurrentID=0, Status='live', Symbol='scaf6', Aliases='Dmel_CG33522, BP1040, BP1047, CG14058, CG32168, CG33522, CG6615, CG6626, Dmel\\CG33522', description='CG33522 gene product from transcript CG33522-RE', other_designations='BP1047|CG33522-PA|CG33522-PC|CG33522-PE|scaf6-PA|scaf6-PC|scaf6-PE', map_location='73E5-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17036859, end_position_on_the_genomic_accession=17042574, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  36668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36668, CurrentID=0, Status='live', Symbol='Kank', Aliases='Dmel_CG10249, BEST:GH19041, BcDNA:GH03482, CG10249, Dmel\\CG10249, kank', description='CG10249 gene product from transcript CG10249-RG', other_designations='CG10249-PA|CG10249-PB|CG10249-PC|CG10249-PD|CG10249-PE|CG10249-PF|CG10249-PG|CG10249-PI|CG10249-PJ|Kank-PA|Kank-PB|Kank-PC|Kank-PD|Kank-PE|Kank-PF|Kank-PG|Kank-PI|Kank-PJ|kank', map_location='51D2-51D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14901475, end_position_on_the_genomic_accession=14928425, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  3772221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772221, CurrentID=0, Status='live', Symbol='sea', Aliases='Dmel_CG6782, CG31305, CG6782, DmCIC, Dmel\\CG6782, SLC25A1, Sea, anon-WO0140519.12, l(3)EP3364', description='scheggia', other_designations='CG6782-PA|CG6782-PC|CG6782-PD|CG6782-PE|sea-PA|sea-PC|sea-PD|sea-PE', map_location='86E10-86E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11559896, end_position_on_the_genomic_accession=11567054, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36670, CurrentID=0, Status='live', Symbol='Lap1', Aliases='Dmel_CG10255, CG10255, CT28815, Dmel\\CG10255, dLAP1, dlap1', description='CG10255 gene product from transcript CG10255-RB', other_designations='CG10255-PA|CG10255-PB|Lap1-PA|Lap1-PB', map_location='51D4-51D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14935660, end_position_on_the_genomic_accession=14940722, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3346239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346239, CurrentID=0, Status='live', Symbol='CG33514', Aliases='Dmel_ CG15017, CG15018, Dmel\\CG33514', description='CG33514 gene product from transcript CG33514-RA', other_designations='CG33514-PA|uncharacterized protein', map_location='64B2-64B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4365434, end_position_on_the_genomic_accession=4366671, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36673, CurrentID=0, Status='live', Symbol='ckn', Aliases='Dmel_CG12424, BcDNA:GM06913, CG12424, Dmel\\CG12424', description='caskin', other_designations='CG12424-PB|CG12424-PC|CG12424-PD|CG12424-PE|CG12424-PF|ckn-PB|ckn-PC|ckn-PD|ckn-PE|ckn-PF', map_location='51D6-51D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14963292, end_position_on_the_genomic_accession=14980795, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36674, CurrentID=0, Status='live', Symbol='pcs', Aliases='Dmel_CG7761, CG7761, Dmel\\CG7761, prt', description='parcas', other_designations='CG7761-PB|pcs-PB|poirot', map_location='51D6-51D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14981263, end_position_on_the_genomic_accession=14991336, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36675, CurrentID=0, Status='live', Symbol='CG7639', Aliases='Dmel_ Dmel\\CG7639', description='CG7639 gene product from transcript CG7639-RB', other_designations='CG7639-PB', map_location='51D7-51D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14987660, end_position_on_the_genomic_accession=14992230, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36676, CurrentID=0, Status='live', Symbol='CG10265', Aliases='Dmel_ CG30474, Dmel\\CG10265', description='CG10265 gene product from transcript CG10265-RC', other_designations='CG10265-PA|CG10265-PC|uncharacterized protein', map_location='51D7-51D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14992332, end_position_on_the_genomic_accession=14993835, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36678, CurrentID=0, Status='live', Symbol='chn', Aliases='Dmel_CG11798, CG11798, Chn, Dmel\\CG11798, EPIL#6, EPIL6, anon-EST:Liang-1.13, clone 1.13, l(2)02064, l(2)k04218, taco', description='charlatan', other_designations='CG11798-PA|CG11798-PB|CG11798-PC|CG11798-PD|CG11798-PE|CG11798-PF|CG11798-PG|CG11798-PH|CG11798-PI|chn-PA|chn-PB|chn-PC|chn-PD|chn-PE|chn-PF|chn-PG|chn-PH|chn-PI', map_location='51E1-51E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15115236, end_position_on_the_genomic_accession=15144783, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38881, CurrentID=0, Status='live', Symbol='CG8111', Aliases='Dmel_ Dmel\\CG8111', description='CG8111 gene product from transcript CG8111-RA', other_designations='CG8111-PA', map_location='66A19-66A19', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7913664, end_position_on_the_genomic_accession=7915120, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772232, CurrentID=0, Status='live', Symbol='fabp', Aliases='Dmel_CG6783, BcDNA:RH06221, BcDNA:RH46282, BcDNA:RH49003, BcDNA:SD12036, CG31305, CG31305-PI, CG6783, CT21061, Dmel\\CG6783, anon-EST:Posey234, anon-WO0140519.12, dFabp', description='fatty acid bindin protein', other_designations='CG6783-PA|CG6783-PB|CG6783-PC|fabp-PA|fabp-PB|fabp-PC|fatty-acid binding protein', map_location='86E10-86E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11559896, end_position_on_the_genomic_accession=11567054, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36681, CurrentID=0, Status='live', Symbol='igl', Aliases='Dmel_CG18285, CG18285, Dmel\\CG18285, Igl', description='igloo', other_designations='CG18285-PA|CG18285-PB|CG18285-PC|igl-PA|igl-PB|igl-PC|invertebrate GAP-43-like', map_location='51E3-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15155437, end_position_on_the_genomic_accession=15190292, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36682, CurrentID=0, Status='live', Symbol='GPHR', Aliases='Dmel_CG8090, CG8090, Dmel\\CG8090, dGPHR', description='Golgi pH Regulator', other_designations='CG8090-PA|GPHR-PA|golgi pH regulator', map_location='51E5-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15199065, end_position_on_the_genomic_accession=15201200, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36683, CurrentID=0, Status='live', Symbol='CG11807', Aliases='Dmel_ CT14564, Dmel\\CG11807', description='CG11807 gene product from transcript CG11807-RB', other_designations='CG11807-PA|CG11807-PB|CG11807-PC|uncharacterized protein', map_location='51E5-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15201253, end_position_on_the_genomic_accession=15204466, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36685, CurrentID=0, Status='live', Symbol='row', Aliases='Dmel_CG8092, CG8092, Dmel\\CG8092, Row, anon-WO0118547.194', description='relative of woc', other_designations='CG8092-PA|CG8092-PB|CG8092-PC|Relative-of-WOC|row-PA|row-PB|row-PC', map_location='51E5-51E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15206527, end_position_on_the_genomic_accession=15213634, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36686, CurrentID=0, Status='live', Symbol='CG8093', Aliases='Dmel_ CT22063, Dmel\\CG8093', description='CG8093 gene product from transcript CG8093-RB', other_designations='CG8093-PA|CG8093-PB|triglyceride lipase', map_location='51E7-51E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15214060, end_position_on_the_genomic_accession=15215686, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246413, CurrentID=0, Status='live', Symbol='Cpr49Aa', Aliases='Dmel_CG30045, CG30045, CG8496, DmelCpr49Aa, Dmel\\CG30045', description='Cuticular protein 49Aa', other_designations='CG30045-PB|Cpr49Aa-PB|cuticular protein 49Aa', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12376626, end_position_on_the_genomic_accession=12377378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36688, CurrentID=0, Status='live', Symbol='mRpL41', Aliases='Dmel_CG12954, BcDNA:RE19416, CG12954, Dmel\\CG12954, YmL27', description='mitochondrial ribosomal protein L41', other_designations='CG12954-PA|mRpL41-PA', map_location='51E7-51E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15217024, end_position_on_the_genomic_accession=15217617, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36689, CurrentID=0, Status='live', Symbol='CG8079', Aliases='Dmel_ Dmel\\CG8079', description='CG8079 gene product from transcript CG8079-RC', other_designations='CG8079-PA|CG8079-PB|CG8079-PC|CG8079-PD|CG8079-PE', map_location='51E7-51E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15218379, end_position_on_the_genomic_accession=15224795, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  5740515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740515, CurrentID=0, Status='live', Symbol='CG34178', Aliases='Dmel_ CG12677, Dmel\\CG34178, Dmel_CG12677', description='CG34178 gene product from transcript CG34178-RB', other_designations='CG34178-PA|CG34178-PB|CG34178-PC|uncharacterized protein', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4433908, end_position_on_the_genomic_accession=4435057, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36692, CurrentID=0, Status='live', Symbol='scb', Aliases='Dmel_CG8095, CG8095, CT21280, CT36929, Dm0620, Dmel\\CG8095, E(Sev-CycE)E93, E(sev-cycE)[e93], ItgaPS3, PS3, Vol, aPS3, alpha-PS3, alphaInt3, alphaPS3, alpha[[PS3]], beta[[PS3], l(2)01288, vol', description='scab', other_designations='CG8095-PA|CG8095-PB|aVolado|alpha Position-Specific antigen 3|alphaPS3 integrin|alphaPS3 subunit|beta[[PS3]]integrin|complementation group 2.11|integrin alphaPS3 subunit|scb-PA|scb-PB|volado', map_location='2-73 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15248784, end_position_on_the_genomic_accession=15258498, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36693, CurrentID=0, Status='live', Symbol='ItgaPS4', Aliases='Dmel_CG16827, CG16827, CT37393, Dmel\\CG16827, ItgalphaPS4, alpha-PS4, alpha-ps4, alphaPS4', description='Integrin alphaPS4 subunit', other_designations='CG16827-PA|ItgaPS4-PA|alphaPS4|alphaPS4 integrin|integrin alphaPS4 subunit', map_location='51E11-51E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15258761, end_position_on_the_genomic_accession=15262873, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36694, CurrentID=0, Status='live', Symbol='Trpm', Aliases='Dmel_CG44240, BcDNA:GH04950, CG16805, CG30078, CG30079, CG34123, CG34187, CG44240, DmTRPM, Dmel\\CG44240, Dmel_CG34123, TRPM, TrpM, dTRPM, mdcds_11815, trpm', description='Transient receptor potential cation channel, subfamily M ortholog', other_designations='CG44240-PA|CG44240-PB|CG44240-PD|CG44240-PE|CG44240-PF|CG44240-PG|CG44240-PH|CG44240-PI|CG44240-PJ|Trpm-PA|Trpm-PB|Trpm-PD|Trpm-PE|Trpm-PF|Trpm-PG|Trpm-PH|Trpm-PI|Trpm-PJ|transient receptor potential cation channel, subfamily M ortholog|transient receptor potential channel M', map_location='51E11-51F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15262808, end_position_on_the_genomic_accession=15309368, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  3772247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772247, CurrentID=0, Status='live', Symbol='CG33671', Aliases='Dmel_ CG16804, CG16804b, CG33009, CG33009-ORFB, CG33009ORFB, CG8810, Dmel\\CG33671, Dromel_CG33672_FBtr0091651_mORF, MevK', description='CG33671 gene product from transcript CG33671-RA', other_designations='CG33671-PA|CG33671-PB|uncharacterized protein', map_location='49B10-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12634166, end_position_on_the_genomic_accession=12636141, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36696, CurrentID=0, Status='live', Symbol='CG8152', Aliases='Dmel_ Dmel\\CG8152', description='CG8152 gene product from transcript CG8152-RA', other_designations='CG8152-PA', map_location='51F4-51F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15309931, end_position_on_the_genomic_accession=15311134, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36697, CurrentID=0, Status='live', Symbol='Xpc', Aliases='Dmel_CG8153, CG8153, DhR4, DhXPC, DmXPC, Dmel\\CG8153, Mus 210, Mus210, XPC[DM]c, mus(2)201, mus(2)201G1, mus(2)201[GI], mus201[G1], mus210, mus212, Xpc', description='Xeroderma pigmentosum, complementation group C', other_designations='CG8153-PA|CG8153-PC|CG8153-PD|Xpc-PA|Xpc-PC|Xpc-PD|mus210|mutagen-sensitive 210|xeroderma pigmentosum group C complementing factor|xeroderma pigmentosum group C-like|xeroderma pigmentosum, complementation group C', map_location='2-69.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15311510, end_position_on_the_genomic_accession=15318239, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36698, CurrentID=0, Status='live', Symbol='CG8155', Aliases='Dmel_ Dmel\\CG8155', description='CG8155 gene product from transcript CG8155-RA', other_designations='CG8155-PA', map_location='51F5-51F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15317817, end_position_on_the_genomic_accession=15322171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36699, CurrentID=0, Status='live', Symbol='Arf51F', Aliases='Dmel_CG8156, AFR51F, ARF3, ARF6, ARFIII, Arf51f, Arf6, CG8156, Dm Arf51F, Dmel\\CG8156, EP(2)2612, LD12493, P40946, arf51F, arf51f, arf6, d-Arf6, dARFIII, dArf6', description='ADP ribosylation factor at 51F', other_designations='ADP ribosylation factor 51F|ADP ribosylation factor III|Arf51F-PA|Arf51F-PB|Arf51F-PC|Arf51F-PD|Arf51F-PE|CG8156-PA|CG8156-PB|CG8156-PC|CG8156-PD|CG8156-PE', map_location='51F6-51F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15322501, end_position_on_the_genomic_accession=15325490, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36700, CurrentID=0, Status='live', Symbol='CG8157', Aliases='Dmel_ BcDNA:RE07473, Dmel\\CG8157', description='CG8157 gene product from transcript CG8157-RB', other_designations='CG8157-PA|CG8157-PB', map_location='51F6-51F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15325503, end_position_on_the_genomic_accession=15326386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36701, CurrentID=0, Status='live', Symbol='CG8160', Aliases='Dmel_ Dmel\\CG8160', description='CG8160 gene product from transcript CG8160-RA', other_designations='CG8160-PA', map_location='51F6-51F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15327143, end_position_on_the_genomic_accession=15328271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36702, CurrentID=0, Status='live', Symbol='Hr51', Aliases='Dmel_CG16801, CG16801, CG16801/NR2E3, DHR51, Dmel\\CG16801, HR51, NR2E3, Nr2e3, PNR, UNF, fdx, unf', description='Hormone receptor 51', other_designations='CG16801-PC|Hr51-PC|drosophila hormone receptor 51|failure to deploy or expand|hormone receptor 51|unfulfilled', map_location='51F7-51F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15331325, end_position_on_the_genomic_accession=15339384, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36703, CurrentID=0, Status='live', Symbol='unc-5', Aliases='Dmel_CG8166, AAF7419, CG8166, CG8168, CT20824, Dmel\\CG8166, Dunc5, UNC-5, UNC5, UNC5D, Unc-5, Unc5, Unc5d, dUnc5, unc5', description='CG8166 gene product from transcript CG8166-RB', other_designations='CG8166-PA|CG8166-PB|UNC5 protein|Uncoordinated5|unc-5-PA|unc-5-PB|unc-5-like|uncoordinated 5', map_location='51F9-51F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15340472, end_position_on_the_genomic_accession=15368300, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3772256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772256, CurrentID=0, Status='live', Symbol='CG33770', Aliases='Dmel_ CG14458, Dmel\\CG33770', description='CG33770 gene product from transcript CG33770-RB', other_designations='CG33770-PB|uncharacterized protein', map_location='79F3-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22648672, end_position_on_the_genomic_accession=22649287, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36705, CurrentID=0, Status='live', Symbol='Pms2', Aliases='Dmel_CG8169, CG8169, Dmel\\CG8169, Mlh2, dmlh-2, dmlh2, dpms2, pms2', description='CG8169 gene product from transcript CG8169-RA', other_designations='CG8169-PA|Pms2-PA', map_location='51F11-51F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15368074, end_position_on_the_genomic_accession=15371520, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36706, CurrentID=0, Status='live', Symbol='SRPK', Aliases='Dmel_CG8174, BcDNA:SD03158, CG8174, Dmel\\CG8174, dSRPK, dSRPK1, srpk', description='SR Protein Kinase', other_designations='CG8174-PA|CG8174-PB|CG8174-PC|CG8174-PD|SR protein kinase|SRPK-PA|SRPK-PB|SRPK-PC|SRPK-PD|cuaba', map_location='51F11-51F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15380724, end_position_on_the_genomic_accession=15387851, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36708, CurrentID=0, Status='live', Symbol='Mtk', Aliases='Dmel_CG8175, BcDNA:RH07954, CG8175, DIM 17, Dmel\\CG8175, METCH, Metc, Metch, metch, metk, mtk', description='Metchnikowin', other_designations='CG8175-PA|Mtk-PA|drosophila immune induced molecule 17|metchikowin|metchinkowin|metchnikovin|metchnikowin|metchnikowin A|metchnikowin B|metchnikowin precursor|metchnikowin-1|metchnikowin-2|metchnikowine', map_location='52A1-52A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15408846, end_position_on_the_genomic_accession=15409113, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246417, CurrentID=0, Status='live', Symbol='CG30050', Aliases='Dmel_ Dmel\\CG30050', description='CG30050 gene product from transcript CG30050-RB', other_designations='CG30050-PB|uncharacterized protein', map_location='49A3-49A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12408954, end_position_on_the_genomic_accession=12409735, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772264, CurrentID=0, Status='live', Symbol='His2B:CG33874', Aliases='Dmel_CG33874, CG33874, Dmel\\CG33874', description='CG33874 gene product from transcript CG33874-RA', other_designations='CG33874-PA|His2B:CG33874-PA|histone H2B', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21522775, end_position_on_the_genomic_accession=21523369, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36713, CurrentID=0, Status='live', Symbol='CG8180', Aliases='Dmel_ Dmel\\CG8180', description='CG8180 gene product from transcript CG8180-RC', other_designations='CG8180-PA|CG8180-PB|CG8180-PC', map_location='52A4-52A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15472175, end_position_on_the_genomic_accession=15482834, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36714, CurrentID=0, Status='live', Symbol='CG12964', Aliases='Dmel_ Dmel\\CG12964', description='CG12964 gene product from transcript CG12964-RB', other_designations='CG12964-PB|CG12964-PC|uncharacterized protein', map_location='52A4-52A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15485953, end_position_on_the_genomic_accession=15497231, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36715, CurrentID=0, Status='live', Symbol='Ir52a', Aliases='Dmel_CG12960, CG12960, DmelIR52a, Dmel\\CG12960, IR52a', description='Ionotropic receptor 52a', other_designations='CG12960-PB|Ir52a-PB|ionotropic receptor 52a', map_location='52A6-52A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15499632, end_position_on_the_genomic_accession=15501431, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36716, CurrentID=0, Status='live', Symbol='Ir52d', Aliases='Dmel_CG30464, CG12961, CG30464, DmelIR52d, Dmel\\CG30464, IR52d', description='Ionotropic receptor 52d', other_designations='CG30464-PB|Ir52d-PB|ionotropic receptor 52d', map_location='52A6-52A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15506582, end_position_on_the_genomic_accession=15508366, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36717, CurrentID=0, Status='live', Symbol='GalNAc-T1', Aliases='Dmel_CG8182, CG8182, Dmel\\CG8182, pgant1', description='CG8182 gene product from transcript CG8182-RA', other_designations='CG8182-PA|CG8182-PB|CG8182-PC|GalNAc-T1-PA|GalNAc-T1-PB|GalNAc-T1-PC|N-acetylgalactosaminyltransferase|polypeptide GalNAc transferase 1', map_location='52A6-52A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15511710, end_position_on_the_genomic_accession=15515597, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36718, CurrentID=0, Status='live', Symbol='Khc-73', Aliases='Dmel_CG8183, CG8183, DmKIN73, DmKin73, DmKlp73, Dmel\\CG8183, KHC-73, KIF 13A, KIF13A/B, KLP73, Khc73, Kin73, Kinesin-73, anon-EST:fe3C9, khc-73', description='Kinesin heavy chain 73', other_designations='CG8183-PA|CG8183-PC|CG8183-PD|CG8183-PE|CG8183-PF|Khc-73-PA|Khc-73-PC|Khc-73-PD|Khc-73-PE|Khc-73-PF|Kinesin-3|Kinesin-73|anon-fast-evolving-3C9|kinesin heavy chain 73|kinesin heavy-chain 73', map_location='52A6-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15515778, end_position_on_the_genomic_accession=15532259, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  3772271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772271, CurrentID=0, Status='live', Symbol='His2B:CG33886', Aliases='Dmel_CG33886, CG33886, Dmel\\CG33886', description='CG33886 gene product from transcript CG33886-RA', other_designations='CG33886-PA|His2B:CG33886-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21492700, end_position_on_the_genomic_accession=21493294, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36720, CurrentID=0, Status='live', Symbol='CG8187', Aliases='Dmel_ Dmel\\CG8187', description='CG8187 gene product from transcript CG8187-RA', other_designations='CG8187-PA', map_location='52A10-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15539554, end_position_on_the_genomic_accession=15541363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36722, CurrentID=0, Status='live', Symbol='eIF2B-gamma', Aliases='Dmel_CG8190, CG8190, Dmel\\CG8190, eIF2Bgamma', description='CG8190 gene product from transcript CG8190-RA', other_designations='CG8190-PA|eIF2B-gamma-PA', map_location='52A10-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15542661, end_position_on_the_genomic_accession=15544409, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36723, CurrentID=0, Status='live', Symbol='CG8192', Aliases='Dmel_ Dmel\\CG8192', description='CG8192 gene product from transcript CG8192-RA', other_designations='CG8192-PA', map_location='52A10-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15544316, end_position_on_the_genomic_accession=15549984, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772276, CurrentID=0, Status='live', Symbol='His2B:CG33882', Aliases='Dmel_CG33882, CG33882, Dmel\\CG33882', description='CG33882 gene product from transcript CG33882-RA', other_designations='CG33882-PA|His2B:CG33882-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21502798, end_position_on_the_genomic_accession=21503392, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36725, CurrentID=0, Status='live', Symbol='CG8195', Aliases='Dmel_ Dmel\\CG8195', description='CG8195 gene product from transcript CG8195-RA', other_designations='CG8195-PA', map_location='52A12-52A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15562697, end_position_on_the_genomic_accession=15565801, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36726, CurrentID=0, Status='live', Symbol='Flo1', Aliases='Dmel_CG8200, CG8200, CK02126, Dmel\\CG8200, FLO1, FLO[Dm], FLO[Dm]-1, Flo, Flo-1, Flo[Dm], flo, flo1, flotillin, reggie-2', description='Flotillin 1', other_designations='CG8200-PA|CG8200-PB|Flo1-PA|Flo1-PB|Flotillin-1|Flotillin1|flotillin|flotillin 1|reggie2', map_location='52A13-52A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15566271, end_position_on_the_genomic_accession=15569404, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36727, CurrentID=0, Status='live', Symbol='Cdk5', Aliases='Dmel_CG8203, CDK5, CG8203, DmCdk5, Dmel\\CG8203, cdk5', description='Cyclin-dependent kinase 5', other_designations='CG8203-PA|Cdk5-PA', map_location='52A13-52A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15569528, end_position_on_the_genomic_accession=15571417, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36728, CurrentID=0, Status='live', Symbol='CG8204', Aliases='Dmel_ BcDNA:AT23078, Dmel\\CG8204', description='CG8204 gene product from transcript CG8204-RA', other_designations='CG8204-PA|CG8204-PB', map_location='52A13-52A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15573144, end_position_on_the_genomic_accession=15573991, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36730, CurrentID=0, Status='live', Symbol='CG8207', Aliases='Dmel_ Dmel\\CG8207', description='CG8207 gene product from transcript CG8207-RA', other_designations='CG8207-PA', map_location='52B5-52B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15677218, end_position_on_the_genomic_accession=15678956, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36731, CurrentID=0, Status='live', Symbol='Vha14-1', Aliases='Dmel_CG8210, ATP6V1F, ATPase, CG8210, Dmel\\CG8210, VAF1_DROME, Vha14, l(2)SH1248, l(2)SH2 1248, vha14, vha14-1', description='Vacuolar H[+] ATPase 14kD subunit 1', other_designations='CG8210-PA|F subunit|V-ATPase 14kD subunit|V-ATPase proton pump|Vha14-1-PA|lethal (2) SH1248|vacuolar H[+] ATPase 14kD subunit|vacuolar H[+] ATPase 14kD subunit 1|vacuolar H[+] ATPase subunit 14-1', map_location='52B5-52B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15679426, end_position_on_the_genomic_accession=15680084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36732, CurrentID=0, Status='live', Symbol='Ranbp11', Aliases='Dmel_CG33139, CG33139, CG8212, Dmel\\CG33139, SP49, imp-beta11, impbeta11', description='CG33139 gene product from transcript CG33139-RA', other_designations='CG33139-PA|Ranbp11-PA|importin-beta11|ran-binding protein 11', map_location='52B5-52B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15682719, end_position_on_the_genomic_accession=15687415, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36733, CurrentID=0, Status='live', Symbol='Cep89', Aliases='Dmel_CG8214, CG8214, Dmel\\CG8214', description='Centrosomal protein 89kDa homolog', other_designations='CG8214-PA|Cep89-PA|centrosomal protein 89kDa homolog', map_location='52B5-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15687388, end_position_on_the_genomic_accession=15688985, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740752, CurrentID=0, Status='live', Symbol='CG34183', Aliases='Dmel_ Dmel\\CG34183', description='CG34183 gene product from transcript CG34183-RA', other_designations='CG34183-PA|uncharacterized protein', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9890061, end_position_on_the_genomic_accession=9890772, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36736, CurrentID=0, Status='live', Symbol='CG30089', Aliases='Dmel_ CG12305, CG12966, CG12967, Dmel\\CG30089', description='CG30089 gene product from transcript CG30089-RA', other_designations='CG30089-PA|CG30089-PB|uncharacterized protein', map_location='52C2-52C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15727381, end_position_on_the_genomic_accession=15762044, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3772289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772289, CurrentID=0, Status='live', Symbol='CG33764', Aliases='Dmel_ Dmel\\CG33764', description='CG33764 gene product from transcript CG33764-RA', other_designations='CG33764-PA|uncharacterized protein', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4266789, end_position_on_the_genomic_accession=4267503, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36740, CurrentID=0, Status='live', Symbol='Zasp52', Aliases='Dmel_CG30084, CG12968, CG12969, CG30084, CG8242, Dmel\\CG30084, Z(210), ZASP52, Zasp, Zbp210, dzasp, tun, tungus, zasp, zasp52', description='Z band alternatively spliced PDZ-motif protein 52', other_designations='CG30084-PC|CG30084-PE|CG30084-PF|CG30084-PG|CG30084-PI|CG30084-PK|CG30084-PL|CG30084-PM|CG30084-PN|CG30084-PO|CG30084-PP|CG30084-PQ|CG30084-PR|CG30084-PS|CG30084-PT|CG30084-PU|CG30084-PV|CG30084-PW|CG30084-PX|CG30084-PY|CG30084-PZ|Z band alternatively spliced PDZ-motif protein|Z-band-protein-210|Zasp52-PC|Zasp52-PE|Zasp52-PF|Zasp52-PG|Zasp52-PI|Zasp52-PK|Zasp52-PL|Zasp52-PM|Zasp52-PN|Zasp52-PO|Zasp52-PP|Zasp52-PQ|Zasp52-PR|Zasp52-PS|Zasp52-PT|Zasp52-PU|Zasp52-PV|Zasp52-PW|Zasp52-PX|Zasp52-PY|Zasp52-PZ|zetalin', map_location='52C4-52C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15762565, end_position_on_the_genomic_accession=15817004, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  36741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36741, CurrentID=0, Status='live', Symbol='Poxn', Aliases='Dmel_CG8246, CG8246, Dmel\\CG8246, P4, Pox-n, neuro, pox-n, poxn', description='Pox neuro', other_designations='CG8246-PA|CG8246-PB|Poxn-PA|Poxn-PB|paired box neural|paired box-neuro|pox neuro|pox-neural|pox-neuro', map_location='52C7-52C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15826693, end_position_on_the_genomic_accession=15835055, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36742, CurrentID=0, Status='live', Symbol='CG8249', Aliases='Dmel_ Dmel\\CG8249', description='CG8249 gene product from transcript CG8249-RC', other_designations='CG8249-PA|CG8249-PB|CG8249-PC', map_location='52C8-52C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15842760, end_position_on_the_genomic_accession=15847362, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36743, CurrentID=0, Status='live', Symbol='tun', Aliases='Dmel_CG8253, BcDNA:LD38553, CG8253, Dmel\\CG8253, Ntaq1', description='tungus', other_designations='CG8253-PA|CG8253-PB|CG8253-PC|CG8253-PD|tun-PA|tun-PB|tun-PC|tun-PD', map_location='52C8-52C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15794000, end_position_on_the_genomic_accession=15855645, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36744, CurrentID=0, Status='live', Symbol='CG12970', Aliases='Dmel_ Dmel\\CG12970', description='CG12970 gene product from transcript CG12970-RC', other_designations='CG12970-PC|uncharacterized protein', map_location='52C8-52C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15855961, end_position_on_the_genomic_accession=15857522, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36746, CurrentID=0, Status='live', Symbol='sli', Aliases='Dmel_CG43758, CG33464, CG43758, CG8355, CT21700, CT37068, Dmel\\CG43758, Dmel_CG33464, Dmel_CG8355, Sli, Slit, dSlit, fus3, l(2)05248, wdh', description='slit', other_designations='CG43758-PA|CG43758-PB|CG43758-PC|CG43758-PD|CG43758-PE|CG43758-PF|SGP cluster fusion defects 3|lethal (2) 05248|sli-PA|sli-PB|sli-PC|sli-PD|sli-PE|sli-PF', map_location='2-77 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15869495, end_position_on_the_genomic_accession=15922118, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  36747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36747, CurrentID=0, Status='live', Symbol='bdg', Aliases='Dmel_CG8291, CG8291, Dmel\\CG8291', description='bedraggled', other_designations='CG8291-PA|CG8291-PB|CG8291-PC|CG8291-PD|bdg-PA|bdg-PB|bdg-PC|bdg-PD', map_location='52D2-52D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15923040, end_position_on_the_genomic_accession=15931935, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36748, CurrentID=0, Status='live', Symbol='Diap2', Aliases='Dmel_CG8293, CG8293, D-IAP2, D-iap2, DIAP, DIAP-2, DIAP2, DIAPII, DIHA, Diap, Diha, Dmel\\CG8293, IAP, IAP2, Iap2, Ilp, dIAP, dIAP2, dILP, diap2', description='Death-associated inhibitor of apoptosis 2', other_designations='CG8293-PA|CG8293-PB|Diap2-PA|Diap2-PB|IAP-like protein|drosophila inhibitor of apoptosis 2|drosophila melanogaster inhibitor of apoptosis 2|inhibitor of apoptosis 2|inhibitor of apoptosis protein 2', map_location='52D2-52D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15932170, end_position_on_the_genomic_accession=15934825, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36749, CurrentID=0, Status='live', Symbol='CG8297', Aliases='Dmel_ Dmel\\CG8297', description='CG8297 gene product from transcript CG8297-RA', other_designations='CG8297-PA', map_location='52D2-52D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15934825, end_position_on_the_genomic_accession=15935938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36750, CurrentID=0, Status='live', Symbol='Mlf', Aliases='Dmel_CG8295, BcDNA:SD18148, CG8295, Dmel\\CG8295, MLF, dMLF, dmlf, mlf', description='Myelodysplasia/myeloid leukemia factor', other_designations='CG8295-PA|CG8295-PB|CG8295-PC|CG8295-PD|CG8295-PE|CG8295-PF|CG8295-PG|CG8295-PH|Mlf-PA|Mlf-PB|Mlf-PC|Mlf-PD|Mlf-PE|Mlf-PF|Mlf-PG|Mlf-PH|drosophila myeloid leukemia factor|myeloid leukemia factor', map_location='52D2-52D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15935893, end_position_on_the_genomic_accession=15940073, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  36751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36751, CurrentID=0, Status='live', Symbol='CG8299', Aliases='Dmel_ Dmel\\CG8299, SP65', description='CG8299 gene product from transcript CG8299-RA', other_designations='CG8299-PA', map_location='52D3-52D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15940610, end_position_on_the_genomic_accession=15941483, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36752, CurrentID=0, Status='live', Symbol='Cyp4aa1', Aliases='Dmel_CG8302, 4aa1, CG8302, Cyp4AA1, Dmel\\CG8302', description='CG8302 gene product from transcript CG8302-RB', other_designations='CG8302-PB|Cyp4aa1-PB|cytochrome P450', map_location='52D3-52D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15942864, end_position_on_the_genomic_accession=15947266, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36753, CurrentID=0, Status='live', Symbol='Strn-Mlck', Aliases='Dmel_CG44162, 143736_at, A(225), BEST:GH22543, BcDNA:RH74685, CG15699, CG18255, CG44162, CG8304, CT41348, Dmel\\CG44162, Dmel_CG15699, Dmel_CG18255, MLCK, MLCK-1, MLCK-2, MLCK-3, Mlc-k, Mlck, Strn-MLCK, Strn-mlck, Strn/Mlck, mlck, strn-mlck', description='Stretchin-Mlck', other_designations='225kD protein|A-band-protein-225|CG44162-PB|CG44162-PG|CG44162-PJ|CG44162-PM|CG44162-PN|CG44162-PR|CG44162-PS|CG44162-PT|CG44162-PU|CG44162-PV|MLC-kinase|Stretchin-MLCK|Strn-Mlck-PB|Strn-Mlck-PG|Strn-Mlck-PJ|Strn-Mlck-PM|Strn-Mlck-PN|Strn-Mlck-PR|Strn-Mlck-PS|Strn-Mlck-PT|Strn-Mlck-PU|Strn-Mlck-PV|myosin light chain kinase|stretchin', map_location='52D9-52D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15946933, end_position_on_the_genomic_accession=15989498, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  36754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36754, CurrentID=0, Status='live', Symbol='CG8366', Aliases='Dmel_ Dmel\\CG8366', description='CG8366 gene product from transcript CG8366-RA', other_designations='CG8366-PA', map_location='52D6-52D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15955040, end_position_on_the_genomic_accession=15956636, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36757, CurrentID=0, Status='live', Symbol='CG8314', Aliases='Dmel_ Dmel\\CG8314, anon-WO0118547.93', description='CG8314 gene product from transcript CG8314-RA', other_designations='CG8314-PA', map_location='52D9-52D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15989683, end_position_on_the_genomic_accession=15991286, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36758, CurrentID=0, Status='live', Symbol='Pex11', Aliases='Dmel_CG8315, CG8315, DmelPex11, Dmel\\CG8315, PEX11A, PEX11B', description='Peroxin 11', other_designations='CG8315-PA|Pex11-PA|peroxin 11', map_location='52D9-52D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15991367, end_position_on_the_genomic_accession=15992375, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36759, CurrentID=0, Status='live', Symbol='CG8320', Aliases='Dmel_ Dmel\\CG8320', description='CG8320 gene product from transcript CG8320-RA', other_designations='CG8320-PA', map_location='52D9-52D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15992408, end_position_on_the_genomic_accession=15993707, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36760, CurrentID=0, Status='live', Symbol='ATPCL', Aliases='Dmel_CG8322, ACL, BcDNA:LD21334, CG8322, DmATPCL, Dmel\\CG8322, anon-WO0140519.179, dATPCL, l(2)01466, l(2)k09217, n(2)k09217', description='ATP citrate lyase', other_designations='ATP-citrate lyase|ATPCL-PD|ATPCL-PE|ATPCL-PF|ATPCL-PG|CG8322-PD|CG8322-PE|CG8322-PF|CG8322-PG', map_location='52D10-52D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15993531, end_position_on_the_genomic_accession=16001290, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36761, CurrentID=0, Status='live', Symbol='CG8370', Aliases='Dmel_ Dmel\\CG8370', description='CG8370 gene product from transcript CG8370-RA', other_designations='CG8370-PA|CG8370-PB', map_location='52D11-52D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16001181, end_position_on_the_genomic_accession=16006901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36762, CurrentID=0, Status='live', Symbol='CG8386', Aliases='Dmel_ Dmel\\CG8386', description='CG8386 gene product from transcript CG8386-RA', other_designations='CG8386-PA', map_location='52D12-52D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16006893, end_position_on_the_genomic_accession=16007667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36763, CurrentID=0, Status='live', Symbol='CG8388', Aliases='Dmel_ Dmel\\CG8388', description='CG8388 gene product from transcript CG8388-RA', other_designations='CG8388-PA', map_location='52D12-52D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16009401, end_position_on_the_genomic_accession=16011438, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36764, CurrentID=0, Status='live', Symbol='CG8389', Aliases='Dmel_ Dmel\\CG8389', description='CG8389 gene product from transcript CG8389-RB', other_designations='CG8389-PA|CG8389-PB', map_location='52D12-52D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16008339, end_position_on_the_genomic_accession=16013815, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3772317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772317, CurrentID=0, Status='live', Symbol='His4:CG33889', Aliases='Dmel_CG33889, CG33889, Dmel\\CG33889', description='CG33889 gene product from transcript CG33889-RA', other_designations='CG33889-PA|His4:CG33889-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21484085, end_position_on_the_genomic_accession=21484497, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36766, CurrentID=0, Status='live', Symbol='Rrp42', Aliases='Dmel_CG8395, CG8395, Dmel\\CG8395, anon-EST:Posey140, anon-EST:Posey187, d', description='CG8395 gene product from transcript CG8395-RA', other_designations='CG8395-PA|Rrp42-PA', map_location='52D14-52D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16014908, end_position_on_the_genomic_accession=16016050, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36767, CurrentID=0, Status='live', Symbol='CG8397', Aliases='Dmel_ Dmel\\CG8397', description='CG8397 gene product from transcript CG8397-RA', other_designations='CG8397-PA', map_location='52D14-52D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16016042, end_position_on_the_genomic_accession=16017154, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36768, CurrentID=0, Status='live', Symbol='CG8399', Aliases='Dmel_ Dmel\\CG8399, SDR2, anon-EST:ParkEST009, sdr2', description='CG8399 gene product from transcript CG8399-RB', other_designations='CG8399-PA|CG8399-PB|stromal cell-derived receptor 2', map_location='52D14-52D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16017737, end_position_on_the_genomic_accession=16024627, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36769, CurrentID=0, Status='live', Symbol='casp', Aliases='Dmel_CG8400, CASP, CG8400, Dmel\\CG8400', description='caspar', other_designations='CG8400-PA|CG8400-PB|CG8400-PC|casp-PA|casp-PB|casp-PC|casper', map_location='52D14-52D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16024628, end_position_on_the_genomic_accession=16028161, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  36770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36770, CurrentID=0, Status='live', Symbol='CG8401', Aliases='Dmel_ Dmel\\CG8401', description='CG8401 gene product from transcript CG8401-RB', other_designations='CG8401-PB', map_location='52D15-52D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16028603, end_position_on_the_genomic_accession=16030091, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36771, CurrentID=0, Status='live', Symbol='SP2353', Aliases='Dmel_CG8403, BcDNA:GH12319, CG8403, CT18665, Dmel\\CG8403', description='CG8403 gene product from transcript CG8403-RB', other_designations='CG8403-PA|CG8403-PB|SP2353-PA|SP2353-PB|agrin-like', map_location='52D15-52E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16031930, end_position_on_the_genomic_accession=16047243, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  36772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36772, CurrentID=0, Status='live', Symbol='CG8405', Aliases='Dmel_ Dmel\\CG8405', description='CG8405 gene product from transcript CG8405-RB', other_designations='CG8405-PA|CG8405-PB|membralin', map_location='52E1-52E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16047648, end_position_on_the_genomic_accession=16060835, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36773, CurrentID=0, Status='live', Symbol='Dg', Aliases='Dmel_CG18250, CG18250, CT41273, DG, DmDG, Dmel\\CG18250, atu, dg, dgn, dys', description='Dystroglycan', other_designations='CG18250-PA|CG18250-PB|CG18250-PC|CG18250-PD|Dg-PA|Dg-PB|Dg-PC|Dg-PD|alphaDystroglycan|atsugari|dystroglican|dystroglycan|dystroglycan-like', map_location='52E2-52E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16081176, end_position_on_the_genomic_accession=16098564, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  36774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36774, CurrentID=0, Status='live', Symbol='CG8414', Aliases='Dmel_ Dmel\\CG8414', description='CG8414 gene product from transcript CG8414-RA', other_designations='CG8414-PA', map_location='52E4-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16099306, end_position_on_the_genomic_accession=16102608, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36775, CurrentID=0, Status='live', Symbol='Rho1', Aliases='Dmel_CG8416, 19549712, 8416, AAF01186, CG8416, D- DRho, DRho1, DRhoA, Dm Rho1, DmRHO-A, Dmel\\CG8416, Drho1, DrhoA, E3.10/J3.8, P48148, RHO, RHO1, Rho, Rho A, Rho GTPase, Rho kinase, RhoA, RhoN19, dRho1, dRhoA, dnRho, l(2)52Fa, l(2)k02107b, n(2)k07236, rho, rho1, rhoA', description='CG8416 gene product from transcript CG8416-RB', other_designations='CG8416-PA|CG8416-PB|CG8416-PC|CG8416-PD|CG8416-PF|CG8416-PG|Rho GTPase|Rho1-PA|Rho1-PB|Rho1-PC|Rho1-PD|Rho1-PF|Rho1-PG|RhoA|RhoA kinase|rho kinase|rho1|rho1 kinase', map_location='52E5-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16102687, end_position_on_the_genomic_accession=16107229, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36776, CurrentID=0, Status='live', Symbol='Ric', Aliases='Dmel_CG8418, 17137290, CG8418, D-Ric, Dmel\\CG8418, Q7JMZ0, RIC, Roc', description='Ras-related protein interacting with calmodulin', other_designations='CG8418-PA|CG8418-PB|Ras operated by calmodulin|Ras which interacts with calmodulin|Ras-related which interacted with calmodulin|Ric-PA|Ric-PB', map_location='52E5-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16107440, end_position_on_the_genomic_accession=16110017, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772329, CurrentID=0, Status='live', Symbol='Chrac-14', Aliases='Dmel_CG13399, 13399, 18K protein, 55A11T, CG13399, CG32956, CHRAC, CHRAC-14, CHRAC14, Chrac14, Dmel\\CG13399, ESTS:55A11T, chrac-14, mary', description='Chromatin accessibility complex 14kD protein', other_designations='CG13399-PB|Chrac-14-PB|chromatin accessibility complex 14kD protein', map_location='29C4-29C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8437405, end_position_on_the_genomic_accession=8437996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36778, CurrentID=0, Status='live', Symbol='Asph', Aliases='Dmel_CG8421, BAH/humbug, CG18658, CG8421, Dmel\\CG8421', description='Aspartyl beta-hydroxylase', other_designations='Asph-PA|Asph-PB|Asph-PG|Asph-PH|Asph-PI|Asph-PJ|Asph-PK|Asph-PL|CG8421-PA|CG8421-PB|CG8421-PG|CG8421-PH|CG8421-PI|CG8421-PJ|CG8421-PK|CG8421-PL|aspartyl beta-hydroxylase', map_location='52E5-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16110331, end_position_on_the_genomic_accession=16116996, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  36779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36779, CurrentID=0, Status='live', Symbol='Jhedup', Aliases='Dmel_CG8424, CG8424, Dmel\\CG8424, JHE, JHEdup', description='Juvenile hormone esterase duplication', other_designations='CG8424-PA|Jhedup-PA|juvenile hormone esterase|juvenile hormone esterase duplication', map_location='52E5-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16117109, end_position_on_the_genomic_accession=16119149, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36780, CurrentID=0, Status='live', Symbol='Jhe', Aliases='Dmel_CG8425, CG8425, DmJhe, Dmel\\CG8425, JH, JHE, jhe', description='Juvenile hormone esterase', other_designations='CG8425-PA|CG8425-PB|CG8425-PC|JH esterase|JH-esterase|Jhe-PA|Jhe-PB|Jhe-PC|juvenile hormone esterase', map_location='52E5-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16119793, end_position_on_the_genomic_accession=16122773, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36782, CurrentID=0, Status='live', Symbol='Got1', Aliases='Dmel_CG8430, BEST:GH07823, CG8430, Dmel\\CG8430, FBgn 1124, GOT-1, Got-1, Got2', description='Glutamate oxaloacetate transaminase 1', other_designations='CG8430-PA|CG8430-PB|Got1-PA|Got1-PB|glutamate oxaloacetate transaminase 1|glutamate oxaloacetic transaminase', map_location='2-75 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16138352, end_position_on_the_genomic_accession=16141928, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36784, CurrentID=0, Status='live', Symbol='Aats-cys', Aliases='Dmel_CG8431, BcDNA:LD21177, CG8431, CRS, CysRS, Dmel\\CG8431', description='Cysteinyl-tRNA synthetase', other_designations='Aats-cys-PA|CG8431-PA', map_location='52E11-52E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16143147, end_position_on_the_genomic_accession=16145906, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3772338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772338, CurrentID=0, Status='live', Symbol='CheA46a', Aliases='Dmel_CG33800, CG33800, Dmel\\CG33800', description='Chemosensory protein A 46a', other_designations='CG33800-PA|CheA46a-PA|chemosensory protein A 46a', map_location='46B2-46B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9748800, end_position_on_the_genomic_accession=9749389, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36787, CurrentID=0, Status='live', Symbol='CG10734', Aliases='Dmel_ Dmel\\CG10734', description='CG10734 gene product from transcript CG10734-RA', other_designations='CG10734-PA|uncharacterized protein', map_location='52F1-52F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16157267, end_position_on_the_genomic_accession=16158794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36788, CurrentID=0, Status='live', Symbol='lbk', Aliases='Dmel_CG8434, CG8434, Dmel\\CG8434, Lbk', description='lambik', other_designations='CG8434-PB|kekkon-like|lbk-PB', map_location='52F2-52F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16150123, end_position_on_the_genomic_accession=16169519, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36789, CurrentID=0, Status='live', Symbol='CG8435', Aliases='Dmel_ Dmel\\CG8435', description='CG8435 gene product from transcript CG8435-RA', other_designations='CG8435-PA', map_location='52F3-52F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16169600, end_position_on_the_genomic_accession=16170861, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36790, CurrentID=0, Status='live', Symbol='Ptp52F', Aliases='Dmel_CG18243, CG18243, CT18044, DPTP52F, Dmel\\CG18243, PTP52F, ptp52F', description='Protein tyrosine phosphatase 52F', other_designations='CG18243-PB|Ptp52F-PB|protein tyrosine phosphatase 52F', map_location='52F3-52F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16171112, end_position_on_the_genomic_accession=16178307, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  36791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36791, CurrentID=0, Status='live', Symbol='Lis-1', Aliases='Dmel_CG8440, CG8440, D-Lis1, DLIS-1, DLIS1, DLis-1, DLis1, Dlis-1, Dlis1, DmLIS1, Dmel\\CG8440, LIS-1, Lis 1, Lis1, dLis1, l(2)k13209, l(2R)W8, lis-1, lis1, n(2)k11702', description='Lissencephaly-1', other_designations='CG8440-PA|CG8440-PB|CG8440-PF|CG8440-PG|Lis-1-PA|Lis-1-PB|Lis-1-PF|Lis-1-PG|Paf-AHbeta|lissencephaly-1|platelet-activating factor acetylhydrolase Ib beta-subunit homolog', map_location='52F4-52F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16178826, end_position_on_the_genomic_accession=16185103, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  3772344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772344, CurrentID=0, Status='live', Symbol='CG33785', Aliases='Dmel_ CG10438, CG10438a, CG30144, CG33005, CG33005-ORFA, CG33005ORFA, Dmel\\CG33785, Dromel_CG33786_FBtr0091786_uORF', description='CG33785 gene product from transcript CG33785-RA', other_designations='CG33785-PA|uncharacterized protein', map_location='57A8-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20642466, end_position_on_the_genomic_accession=20643785, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36793, CurrentID=0, Status='live', Symbol='clu', Aliases='Dmel_CG8443, CG8443, Clu, Dmel\\CG8443, anon-WO0118547.143', description='clueless', other_designations='CG8443-PA|CG8443-PB|clu-PA|clu-PB', map_location='52F5-52F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16186584, end_position_on_the_genomic_accession=16194277, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36794, CurrentID=0, Status='live', Symbol='calypso', Aliases='Dmel_CG8445, BAP1, CG8445, Dmel\\CG8445, Q7K5N4', description='CG8445 gene product from transcript CG8445-RB', other_designations='CG8445-PA|CG8445-PB|calypso-PA|calypso-PB|ubiquitin thiolesterase', map_location='52F7-52F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16194062, end_position_on_the_genomic_accession=16196333, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36795, CurrentID=0, Status='live', Symbol='CG44243', Aliases='Dmel_ CG8446, Dmel\\CG44243', description='CG44243 gene product from transcript CG44243-RA', other_designations='CG44243-PA|CG44243-PB|uncharacterized protein', map_location='52F7-52F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16197042, end_position_on_the_genomic_accession=16199555, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36797, CurrentID=0, Status='live', Symbol='mrj', Aliases='Dmel_CG8448, CG8448, Dmel\\CG8448, MRJ, dMRJ, d', description='CG8448 gene product from transcript CG8448-RD', other_designations='CG8448-PA|CG8448-PB|CG8448-PC|CG8448-PD|CG8448-PE|CG8448-PF|CG8448-PG|CG8448-PH|mrj-PA|mrj-PB|mrj-PC|mrj-PD|mrj-PE|mrj-PF|mrj-PG|mrj-PH', map_location='52F8-52F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16203329, end_position_on_the_genomic_accession=16221793, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  36798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36798, CurrentID=0, Status='live', Symbol='CG7798', Aliases='Dmel_ Dmel\\CG7798', description='CG7798 gene product from transcript CG7798-RA', other_designations='CG7798-PA', map_location='52F11-52F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16222112, end_position_on_the_genomic_accession=16222627, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772351, CurrentID=0, Status='live', Symbol='His2A:CG33817', Aliases='Dmel_CG33817, CG33817, Dmel\\CG33817', description='CG33817 gene product from transcript CG33817-RA', other_designations='CG33817-PA|His2A:CG33817-PA|histone H2A', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21457995, end_position_on_the_genomic_accession=21458487, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36800, CurrentID=0, Status='live', Symbol='Tsf3', Aliases='Dmel_CG3666, CG3666, Dmel\\CG3666, tsf3', description='Transferrin 3', other_designations='CG3666-PA|Tsf3-PA|transferrin|transferrin 3', map_location='52F11-52F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16226501, end_position_on_the_genomic_accession=16229869, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36801, CurrentID=0, Status='live', Symbol='CG15701', Aliases='Dmel_ Dmel\\CG15701', description='CG15701 gene product from transcript CG15701-RA', other_designations='CG15701-PA|uncharacterized protein', map_location='52F11-53A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16230198, end_position_on_the_genomic_accession=16234087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36802, CurrentID=0, Status='live', Symbol='CG7786', Aliases='Dmel_ DM33, DM7, Dmel\\CG7786', description='CG7786 gene product from transcript CG7786-RA', other_designations='CG7786-PA', map_location='53A1-53A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16234358, end_position_on_the_genomic_accession=16234936, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36803, CurrentID=0, Status='live', Symbol='CG3687', Aliases='Dmel_ BcDNA:AT22273, Dmel\\CG3687', description='CG3687 gene product from transcript CG3687-RA', other_designations='CG3687-PA', map_location='53A1-53A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16236715, end_position_on_the_genomic_accession=16237479, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36804, CurrentID=0, Status='live', Symbol='krimp', Aliases='Dmel_CG15707, CG15707, Dmel\\CG15707, KRIMP, Krimp, mtc', description='krimper', other_designations='CG15707-PA|krimp-PA|montecristo', map_location='53A1-53A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16238103, end_position_on_the_genomic_accession=16240722, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36805, CurrentID=0, Status='live', Symbol='CG15708', Aliases='Dmel_ Dmel\\CG15708', description='CG15708 gene product from transcript CG15708-RA', other_designations='CG15708-PA|uncharacterized protein', map_location='53A1-53A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16240936, end_position_on_the_genomic_accession=16241934, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36806, CurrentID=0, Status='live', Symbol='CngA', Aliases='Dmel_CG42701, CG15709, CG42701, CG7779, CNG, Cng, DmCNGC, Dmel\\CG42701, Dmel_CG15709, Dmel_CG7779, cng, dCNCalpha, dmCNGA', description='Cyclic nucleotide-gated ion channel subunit A', other_designations='CG42701-PA|CG42701-PB|CNG channel alpha subunit|CNG-channel|CngA-PA|CngA-PB|Cyclic-nucleotide-gated ion channel protein|cyclic nucleotide-gated ion channel subunit A', map_location='53A2-53A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16243145, end_position_on_the_genomic_accession=16259805, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  3772359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772359, CurrentID=0, Status='live', Symbol='CG33704', Aliases='Dmel_ Dmel\\CG33704, NEST:bs05b09, NEST:bs14b07, NEST:bs27f10, NEST:bs29a11, NEST:bs30h04, bs30h04.y1', description='CG33704 gene product from transcript CG33704-RA', other_designations='CG33704-PA|uncharacterized protein', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20950102, end_position_on_the_genomic_accession=20951094, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36808, CurrentID=0, Status='live', Symbol='fidipidine', Aliases='Dmel_CG7773, CG7773, Dmel\\CG7773', description='CG7773 gene product from transcript CG7773-RA', other_designations='CG7773-PA|fidipidine-PA', map_location='53A2-53A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16262284, end_position_on_the_genomic_accession=16264619, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36809, CurrentID=0, Status='live', Symbol='csul', Aliases='Dmel_CG3730, CG3730, Csul, DART5, Dart5, Dmel\\CG3730, JBP1, PRMT5, dPRMT5, dart5, dprmt5', description='capsuleen', other_designations='CG3730-PA|CG3730-PB|capsuleenn|csul-PA|csul-PB|drosophila arginine methyltransferase 5', map_location='53A3-53A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16264652, end_position_on_the_genomic_accession=16266959, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36810, CurrentID=0, Status='live', Symbol='Khc', Aliases='Dmel_CG7765, CG7765, DK, DKH, DmK, DmKHC, Dmel\\CG7765, Dmkin, KHC, KIF 5A, KIF5B, KIN, Kin, Kin-1, Kinesin-1, khc, kin, kinesin, kinesin-1, l(2)W12, l(2)k13219, l(2)k13314, l(2R)W12, pgs', description='Kinesin heavy chain', other_designations='CG7765-PA|Khc-PA|Kin-1 heavy chain|conventional kinesin|heavy chain of kinesin|kinesin|kinesin 1|kinesin I|kinesin alpha subunit|kinesin alpha-chain|kinesin heavy chain|kinesin heavy chain I|kinesin-1|kinesin-1 heavy chain|partagas', map_location='53A3-53A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16266960, end_position_on_the_genomic_accession=16271971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36811, CurrentID=0, Status='live', Symbol='CG33017', Aliases='Dmel_ CG15702, CG15703, Dmel\\CG33017', description='CG33017 gene product from transcript CG33017-RD', other_designations='CG33017-PB|CG33017-PC|CG33017-PD|uncharacterized protein', map_location='53A4-53A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16272114, end_position_on_the_genomic_accession=16277655, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36813, CurrentID=0, Status='live', Symbol='cato', Aliases='Dmel_CG7760, CG7760, Cato, DATH-B, Dmel\\CG7760, bHLHa12', description='cousin of atonal', other_designations='CG7760-PA|cato-PA|drosophila atonal homologue', map_location='53A4-53A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16277685, end_position_on_the_genomic_accession=16278396, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36814, CurrentID=0, Status='live', Symbol='CG15704', Aliases='Dmel_ BcDNA:RH57845, Dmel\\CG15704', description='CG15704 gene product from transcript CG15704-RA', other_designations='CG15704-PA|uncharacterized protein', map_location='53A4-53A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16280015, end_position_on_the_genomic_accession=16280470, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772367, CurrentID=0, Status='live', Symbol='mRpL53', Aliases='Dmel_CG30481, BcDNA:GM05817, CG30481, CG33184, Dmel\\CG30481', description='mitochondrial ribosomal protein L53', other_designations='CG30481-PA|mRpL53-PA', map_location='50C16-50C16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13954195, end_position_on_the_genomic_accession=13955664, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36816, CurrentID=0, Status='live', Symbol='CG7755', Aliases='Dmel_ Dmel\\CG7755', description='CG7755 gene product from transcript CG7755-RA', other_designations='CG7755-PA', map_location='53A5-53A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16281630, end_position_on_the_genomic_accession=16282946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36817, CurrentID=0, Status='live', Symbol='CG9068', Aliases='Dmel_ Dmel\\CG9068', description='CG9068 gene product from transcript CG9068-RA', other_designations='CG9068-PA', map_location='53A5-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16283435, end_position_on_the_genomic_accession=16287450, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3772370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772370, CurrentID=0, Status='live', Symbol='His3:CG33821', Aliases='Dmel_CG33821, CG33821, Dmel\\CG33821', description='CG33821 gene product from transcript CG33821-RA', other_designations='CG33821-PA|His3:CG33821-PA|histone H3', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21464507, end_position_on_the_genomic_accession=21465035, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36819, CurrentID=0, Status='live', Symbol='JhI-26', Aliases='Dmel_CG3767, CG3767, Dmel\\CG3767', description='Juvenile hormone-inducible protein 26', other_designations='CG3767-PA|JhI-26-PA|juvenile hormone-inducible protein 26', map_location='53B1-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16289320, end_position_on_the_genomic_accession=16291333, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36820, CurrentID=0, Status='live', Symbol='CG7747', Aliases='Dmel_ CYC4, Dmel\\CG7747', description='CG7747 gene product from transcript CG7747-RA', other_designations='CG7747-PA|NM 137269.2', map_location='53B1-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16292264, end_position_on_the_genomic_accession=16294130, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36821, CurrentID=0, Status='live', Symbol='Atg9', Aliases='Dmel_CG3615, ATG9, CG3615, Dmel\\CG3615, atg9', description='Autophagy-related 9', other_designations='Atg9-PA|Atg9-PB|Autophagy-specific gene 9|CG3615-PA|CG3615-PB', map_location='53B1-53B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16294400, end_position_on_the_genomic_accession=16297232, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772374, CurrentID=0, Status='live', Symbol='His3:CG33824', Aliases='Dmel_CG33824, CG33824, Dmel\\CG33824', description='CG33824 gene product from transcript CG33824-RA', other_designations='CG33824-PA|His3:CG33824-PA|histone H3', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21469550, end_position_on_the_genomic_accession=21470078, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772375, CurrentID=0, Status='live', Symbol='CG32270', Aliases='Dmel_ CG11530, CG33161, Dmel\\CG32270, SP124', description='CG32270 gene product from transcript CG32270-RA', other_designations='CG32270-PA|uncharacterized protein', map_location='63B12-63B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3206949, end_position_on_the_genomic_accession=3208985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36824, CurrentID=0, Status='live', Symbol='CG8060', Aliases='Dmel_ Dmel\\CG8060', description='CG8060 gene product from transcript CG8060-RA', other_designations='CG8060-PA', map_location='53B4-53B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16316313, end_position_on_the_genomic_accession=16320732, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36825, CurrentID=0, Status='live', Symbol='CG4282', Aliases='Dmel_ Dmel\\CG4282', description='CG4282 gene product from transcript CG4282-RA', other_designations='CG4282-PA', map_location='53B5-53B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16322186, end_position_on_the_genomic_accession=16324801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36826, CurrentID=0, Status='live', Symbol='Vha44', Aliases='Dmel_CG8048, 6072, ATP6V1C, CG8048, Dmel\\CG8048, l(2)6072, l(2)SH1339, l(2)SH2 1339, vha44', description='Vacuolar H[+] ATPase 44kD subunit', other_designations='C subunit|CG8048-PA|CG8048-PB|CG8048-PC|CG8048-PE|CG8048-PF|V-ATPase C subunit|Vha44-PA|Vha44-PB|Vha44-PC|Vha44-PE|Vha44-PF|lethal (2) SH1339|vacuolar ATPase C-subunit|vacuolar H+ ATPase 44kD C subunit|vacuolar H[+] ATPase 44 kDa subunit|vacuolar H[+] ATPase 44kD C subunit|vacuolar H[+] ATPase 44kD subunit', map_location='53B5-53C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16324813, end_position_on_the_genomic_accession=16336637, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  3772380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772380, CurrentID=0, Status='live', Symbol='CG33766', Aliases='Dmel_ CG14458, Dmel\\CG33766', description='CG33766 gene product from transcript CG33766-RA', other_designations='CG33766-PA|uncharacterized protein', map_location='79F3-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22645627, end_position_on_the_genomic_accession=22646236, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36829, CurrentID=0, Status='live', Symbol='CG7997', Aliases='Dmel_ CT24026, Dmel\\CG7997', description='CG7997 gene product from transcript CG7997-RB', other_designations='CG7997-PA|CG7997-PB', map_location='53C1-53C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16341251, end_position_on_the_genomic_accession=16344182, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  3772382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772382, CurrentID=0, Status='live', Symbol='Plp', Aliases='Dmel_CG33957, AKAP450, CG13459, CG18648, CG33957, CG6735, CP309, Cp309, D-PLP, DPLP, Dmel\\CG33957, Dplp, PACT, PLP, cp309, d-plp, dPLP, dplp, l(3)s2172, pPlp, plp', description='Pericentrin-like protein', other_designations='CG33957-PC|CG33957-PD|CG33957-PF|CG33957-PG|CG33957-PH|CG33957-PJ|CG33957-PK|CG33957-PL|CG33957-PM|CG33957-PN|CG33957-PO|CG33957-PP|Plp-PC|Plp-PD|Plp-PF|Plp-PG|Plp-PH|Plp-PJ|Plp-PK|Plp-PL|Plp-PM|Plp-PN|Plp-PO|Plp-PP|cp309|drosophila pericentrin like protein|lethal (3) s2172|pericentrin|pericentrin-like|pericentrin-like protein|pericentrin-like-protein|pericentrin/AKAP450 centrosomal targeting', map_location='71B2-71B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15066081, end_position_on_the_genomic_accession=15094724, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  36831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36831, CurrentID=0, Status='live', Symbol='wcd', Aliases='Dmel_CG7989, CG7989, Dmel\\CG7989, Wcd, l(2)k07824, l(2)kO7824', description='wicked', other_designations='CG7989-PA|lethal (2) k07824|wcd-PA', map_location='53C1-53C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16345838, end_position_on_the_genomic_accession=16347500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36832, CurrentID=0, Status='live', Symbol='CG15710', Aliases='Dmel_ Dmel\\CG15710', description='CG15710 gene product from transcript CG15710-RA', other_designations='CG15710-PA|uncharacterized protein', map_location='53C1-53C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16347614, end_position_on_the_genomic_accession=16348527, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36833, CurrentID=0, Status='live', Symbol='CG6262', Aliases='Dmel_ Dmel\\CG6262', description='CG6262 gene product from transcript CG6262-RB', other_designations='CG6262-PA|CG6262-PB|CG6262-PC', map_location='53C2-53C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16359730, end_position_on_the_genomic_accession=16363524, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36834, CurrentID=0, Status='live', Symbol='CG15711', Aliases='Dmel_ Dmel\\CG15711', description='CG15711 gene product from transcript CG15711-RB', other_designations='CG15711-PA|CG15711-PB|uncharacterized protein', map_location='53C2-53C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16363551, end_position_on_the_genomic_accession=16365895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36838, CurrentID=0, Status='live', Symbol='CG4398', Aliases='Dmel_ Dmel\\CG4398', description='CG4398 gene product from transcript CG4398-RA', other_designations='CG4398-PA', map_location='53C4-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16432244, end_position_on_the_genomic_accession=16433183, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36839, CurrentID=0, Status='live', Symbol='CG7848', Aliases='Dmel_ Dmel\\CG7848', description='CG7848 gene product from transcript CG7848-RA', other_designations='CG7848-PA', map_location='53C4-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16433852, end_position_on_the_genomic_accession=16435954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36840, CurrentID=0, Status='live', Symbol='CG4409', Aliases='Dmel_ Dmel\\CG4409', description='CG4409 gene product from transcript CG4409-RB', other_designations='CG4409-PB', map_location='53C4-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16443211, end_position_on_the_genomic_accession=16444845, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36841, CurrentID=0, Status='live', Symbol='CG15925', Aliases='Dmel_ Dmel\\CG15925', description='CG15925 gene product from transcript CG15925-RA', other_designations='CG15925-PA|uncharacterized protein', map_location='53C4-53C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16445986, end_position_on_the_genomic_accession=16447512, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36843, CurrentID=0, Status='live', Symbol='CG15712', Aliases='Dmel_ Dmel\\CG15712', description='CG15712 gene product from transcript CG15712-RA', other_designations='CG15712-PA|uncharacterized protein', map_location='53C6-53C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16484841, end_position_on_the_genomic_accession=16485552, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772396, CurrentID=0, Status='live', Symbol='seq', Aliases='Dmel_CG32904, 2R5, 49Fc, CG17724b, CG32904, CG4037, CG4055, Dmel\\CG32904, Seq, l(2)49Fc, l(2)vr5, vr5', description='sequoia', other_designations='CG32904-PA|CG32904-PB|lethal(2)49Fc|seq-PA|seq-PB', map_location='49F4-49F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13176553, end_position_on_the_genomic_accession=13190086, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36845, CurrentID=0, Status='live', Symbol='CG7813', Aliases='Dmel_ Dmel\\CG7813', description='CG7813 gene product from transcript CG7813-RA', other_designations='CG7813-PA', map_location='53C6-53C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16496889, end_position_on_the_genomic_accession=16499429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36846, CurrentID=0, Status='live', Symbol='Sema-2a', Aliases='Dmel_CG4700, 112/40, BcDNA:LD09561, CG4700, CT14922, D-Sema II, D-Sema-II, D-Sema2a, D-SemaII, D-semaII, Dmel\\CG4700, Dsema-II, Sema2a, SemaII, dSema-II, l(2)03021, sema 2a, sema II, sema-2a, sema-II, sema2a, semaII', description='CG4700 gene product from transcript CG4700-RC', other_designations='CG4700-PA|CG4700-PB|CG4700-PC|CG4700-PE|D-Semaphorin II|Sema-2a-PA|Sema-2a-PB|Sema-2a-PC|Sema-2a-PE|semaphorin|semaphorin 2a|semaphorin II|semiphorin', map_location='53C6-53C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16500178, end_position_on_the_genomic_accession=16533459, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  36847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36847, CurrentID=0, Status='live', Symbol='loopin-1', Aliases='Dmel_CG4750, CG4750, Dmel\\CG4750, S-LAP 6, anon-WO0140519.81, kl3 lp-1, kl3lp1, loopin', description='CG4750 gene product from transcript CG4750-RC', other_designations='CG4750-PA|CG4750-PC|Sperm-LeucylAminoPeptidase 6|kl-3 loop protein-1|loopin-1-PA|loopin-1-PC', map_location='53C9-53C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16534943, end_position_on_the_genomic_accession=16537086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36848, CurrentID=0, Status='live', Symbol='Syn2', Aliases='Dmel_CG4905, CG4905, DmSYN-2, Dmel\\CG4905', description='Syntrophin-like 2', other_designations='CG4905-PA|CG4905-PB|CG4905-PC|CG4905-PE|Syn2-PA|Syn2-PB|Syn2-PC|Syn2-PE|Syntrophin-2|syntrophin 2', map_location='53C7-53C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16552258, end_position_on_the_genomic_accession=16558486, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36849, CurrentID=0, Status='live', Symbol='DAT', Aliases='Dmel_CG8380, CG8380, DmDAT, Dmel\\CG8380, dDAT, fmn, fumin', description='Dopamine transporter', other_designations='CG8380-PA|CG8380-PB|DAT-PA|DAT-PB|dopamine transporter|fumin', map_location='53C7-53C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16558567, end_position_on_the_genomic_accession=16565238, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36850, CurrentID=0, Status='live', Symbol='CG4945', Aliases='Dmel_ CT15864, Dmel\\CG4945, PKN', description='CG4945 gene product from transcript CG4945-RA', other_designations='CG4945-PA|CG4945-PB|CG4945-PC|CG4945-PD', map_location='53C8-53C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16566023, end_position_on_the_genomic_accession=16570120, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36851, CurrentID=0, Status='live', Symbol='RpS15', Aliases='Dmel_CG8332, CG8332, Dmel\\CG8332, S15, anon-EST:Posey137, anon-EST:Posey185, s15', description='Ribosomal protein S15', other_designations='CG8332-PA|CG8332-PB|RpS15-PA|RpS15-PB|ribosomal protein S15', map_location='53C8-53C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16570112, end_position_on_the_genomic_accession=16571230, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36852, CurrentID=0, Status='live', Symbol='CG4927', Aliases='Dmel_ Dmel\\CG4927, SP42', description='CG4927 gene product from transcript CG4927-RC', other_designations='CG4927-PC', map_location='53C8-53C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16572082, end_position_on_the_genomic_accession=16574154, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36853, CurrentID=0, Status='live', Symbol='Lst', Aliases='Dmel_CG8317, CG8317, Dmel\\CG8317, lst', description='Limostatin', other_designations='CG8317-PA|CG8317-PB|CG8317-PC|Lst-PA|Lst-PB|Lst-PC|limostatin', map_location='53C8-53C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16574579, end_position_on_the_genomic_accession=16576586, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36854, CurrentID=0, Status='live', Symbol='Cdk4', Aliases='Dmel_CG5072, 8-6, CDK4, CDK4/6, CG5072/6, DmCdk4, Dmel\\CG5072, Pk53C, Pk?7, cdk, cdk4, cdk4/6, dCdk4, l(2)05428, l(2)0671, l(2)k06503, l(2)s4639, l(2)sh0671, Cdk4', description='Cyclin-dependent kinase 4', other_designations='CG5072-PA|CG5072-PB|CG5072-PC|CG5072-PD|CG5072-PE|Cdk4-PA|Cdk4-PB|Cdk4-PC|Cdk4-PD|Cdk4-PE|Cyclin-dependent kinase 4/6|cyclin-dependent kinase 4|protein kinase-like 53C', map_location='53C9-53C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16579470, end_position_on_the_genomic_accession=16585448, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36855, CurrentID=0, Status='live', Symbol='RpLP2', Aliases='Dmel_CG4918, BcDNA:SD22208, CG4918, Dmel\\CG4918, LP2, RLA2_DROME, RpA1, RpP1, RplP2, T1, Tp1, bs34h02.y1, chr2R:12101099..12101299, rpA1', description='Ribosomal protein LP2', other_designations='ACIDIC RIBOSOMAL PROTEIN LP2|CG4918-PA|RpLP2-PA|Temporal-protein-1|ribosomal protein A1|ribosomal protein LP2|ribosomal protein P1|rpA1|temporal protein 1', map_location='53C9-53C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16586129, end_position_on_the_genomic_accession=16586732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36856, CurrentID=0, Status='live', Symbol='CG8311', Aliases='Dmel_ Dmel\\CG8311', description='CG8311 gene product from transcript CG8311-RA', other_designations='CG8311-PA|dolichol kinase', map_location='53C9-53C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16587045, end_position_on_the_genomic_accession=16589184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36857, CurrentID=0, Status='live', Symbol='CG8306', Aliases='Dmel_ Dmel\\CG8306', description='CG8306 gene product from transcript CG8306-RA', other_designations='CG8306-PA', map_location='53C9-53C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16589518, end_position_on_the_genomic_accession=16594646, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36858, CurrentID=0, Status='live', Symbol='CG8303', Aliases='Dmel_ Dmel\\CG8303', description='CG8303 gene product from transcript CG8303-RD', other_designations='CG8303-PC|CG8303-PD', map_location='53C9-53C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16595028, end_position_on_the_genomic_accession=16604373, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36859, CurrentID=0, Status='live', Symbol='CG5089', Aliases='Dmel_ Dmel\\CG5089, anon-WO0140519.170', description='CG5089 gene product from transcript CG5089-RA', other_designations='CG5089-PA', map_location='53C9-53C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16598748, end_position_on_the_genomic_accession=16600411, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36860, CurrentID=0, Status='live', Symbol='CG5065', Aliases='Dmel_ Dmel\\CG5065', description='CG5065 gene product from transcript CG5065-RA', other_designations='CG5065-PA|CG5065-PB|CG5065-PC|CG5065-PD|CG5065-PE', map_location='53C11-53C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16607654, end_position_on_the_genomic_accession=16624894, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  36862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36862, CurrentID=0, Status='live', Symbol='gprs', Aliases='Dmel_CG18471, CG18471, CG5967, Dmel\\CG18471, Gprs', description='CG18471 gene product from transcript CG18471-RE', other_designations='CG18471-PA|CG18471-PB|CG18471-PC|CG18471-PD|CG18471-PE|CG18471-PF|gprs-PA|gprs-PB|gprs-PC|gprs-PD|gprs-PE|gprs-PF', map_location='53C13-53C14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16641243, end_position_on_the_genomic_accession=16651125, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36863, CurrentID=0, Status='live', Symbol='Amyrel', Aliases='Dmel_CG8221, AmyRel, CG8221, Dmel\\CG8221, amyrel', description='CG8221 gene product from transcript CG8221-RA', other_designations='Amyrel-PA|CG8221-PA|amyrel', map_location='53C14-53C14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16651095, end_position_on_the_genomic_accession=16652699, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36864, CurrentID=0, Status='live', Symbol='Nach', Aliases='Dmel_CG8178, CG8178, Dmel\\CG8178, PPK4, ppk4', description='CG8178 gene product from transcript CG8178-RA', other_designations='CG8178-PA|Nach-PA|nach|pickpocket 4', map_location='53C14-53C14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16653205, end_position_on_the_genomic_accession=16655046, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36865, CurrentID=0, Status='live', Symbol='Picot', Aliases='Dmel_CG8098, CG8098, Dmel\\CG8098, MFS8, picot', description='CG8098 gene product from transcript CG8098-RA', other_designations='CG8098-PA|CG8098-PB|CG8098-PC|CG8098-PD|Picot-PA|Picot-PB|Picot-PC|Picot-PD|picot|putative inorganic phosphate cotransporter', map_location='53C15-53D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16655323, end_position_on_the_genomic_accession=16667675, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  3772419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772419, CurrentID=0, Status='live', Symbol='CG18065', Aliases='Dmel_ CG33008, Dmel\\CG18065', description='CG18065 gene product from transcript CG18065-RA', other_designations='CG18065-PA|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20550334, end_position_on_the_genomic_accession=20552090, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36868, CurrentID=0, Status='live', Symbol='CG5210', Aliases='Dmel_ Chit, Chit13, Cht13, DS47, DmDS47, Dmel\\CG5210, chit, ds47', description='CG5210 gene product from transcript CG5210-RB', other_designations='CG5210-PA|CG5210-PB|chitinase-like', map_location='53D2-53D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16691022, end_position_on_the_genomic_accession=16693751, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772421, CurrentID=0, Status='live', Symbol='His3:CG33845', Aliases='Dmel_CG33845, CG33845, Dmel\\CG33845', description='CG33845 gene product from transcript CG33845-RA', other_designations='CG33845-PA|His3:CG33845-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21504973, end_position_on_the_genomic_accession=21505501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36871, CurrentID=0, Status='live', Symbol='CG15617', Aliases='Dmel_ CT35755, Dmel\\CG15617', description='CG15617 gene product from transcript CG15617-RA', other_designations='CG15617-PA|uncharacterized protein', map_location='53D5-53D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16745223, end_position_on_the_genomic_accession=16746579, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36872, CurrentID=0, Status='live', Symbol='Acp53C14a', Aliases='Dmel_CG8626, Acp53, Acp53Ca, Apc53C14a, CG8626, Dmel\\CG8626', description='Accessory gland protein 53C14a', other_designations='Acp53C14a-PA|Acp53C14a-PB|CG8626-PA|CG8626-PB|accessory gland protein 53C14a', map_location='53D5-53D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16747834, end_position_on_the_genomic_accession=16748590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36873, CurrentID=0, Status='live', Symbol='Acp53C14b', Aliases='Dmel_CG15616, Acp53, Acp53Cb, CG15616, Dmel\\CG15616', description='Accessory gland protein 53C14b', other_designations='Acp53C14b-PA|Acp53C14b-PB|CG15616-PA|CG15616-PB|accessory gland protein 53C14b', map_location='53D6-53D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16748815, end_position_on_the_genomic_accession=16749460, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36874, CurrentID=0, Status='live', Symbol='Acp53Ea', Aliases='Dmel_CG8622, ACP53EA, Acp53E, BcDNA:AT09943, CG8622, Dmel\\CG8622', description='Accessory gland protein 53Ea', other_designations='Acp53Ea-PA|CG8622-PA|accessory gland protein 53Ea|accessory gland-specific peptide 53Ea', map_location='53D6-53D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16749803, end_position_on_the_genomic_accession=16750335, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772427, CurrentID=0, Status='live', Symbol='primo-2', Aliases='Dmel_CG33747, BEO2, CG31311, CG33747, CG9599, Dmel\\CG33747, LMW-PTP, Primo-2, primo', description='CG33747 gene product from transcript CG33747-RC', other_designations='CG33747-PC|primo-2-PC', map_location='87F12-87F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13710017, end_position_on_the_genomic_accession=13711835, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36876, CurrentID=0, Status='live', Symbol='unc-104', Aliases='Dmel_CG8566, CG8566, DUNC104, DUnc-104, DmCG8566, DmDUnc104, DmKlp53D, Dmel\\CG8566, Dunc104, KIF 1B, KIF1A, KIF1A/1B, Klp53D, Unc-104, Unc104, bris, imac, unc104', description='unc-104 ortholog (C. elegans)', other_designations='CG8566-PB|CG8566-PC|CG8566-PD|CG8566-PE|CG8566-PF|CG8566-PG|CG8566-PH|CG8566-PI|D uncoordinated protein 104|Kinesin-3|bristly|immaculate connections|kinesin 3|unc-104 ortholog|unc-104-PB|unc-104-PC|unc-104-PD|unc-104-PE|unc-104-PF|unc-104-PG|unc-104-PH|unc-104-PI', map_location='53D6-53D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16751503, end_position_on_the_genomic_accession=16772671, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  36877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36877, CurrentID=0, Status='live', Symbol='CG5348', Aliases='Dmel_ Dmel\\CG5348', description='CG5348 gene product from transcript CG5348-RC', other_designations='CG5348-PB|CG5348-PC', map_location='53D7-53D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16766804, end_position_on_the_genomic_accession=16769238, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36878, CurrentID=0, Status='live', Symbol='Sod2', Aliases='Dmel_CG8905, CG8905, Dmel\\CG8905, Mito SOD, Mn SOD, Mn-SOD, Mn-SOD2, MnSOD, MnSODII, Mn[2+]SOD, SOD, SOD-2, SOD2, Sod-2, dSOD2, dSod2, mitSOD2, mnSOD, sod2', description='Superoxide dismutase 2 (Mn)', other_designations='CG8905-PA|CG8905-PB|Mn superoxide dismutase|Mn-superoxide dismutase|MnSOD|Sod2-PA|Sod2-PB|manganese superoxide dismutase|manganese-SOD|mitochodnrial SOD|mitochondrial Mn-superoxide dismutase 2|mitochondrial MnSOD|mitochondrial superoxide dismutase 2|superoxide dismutase|superoxide dismutase 2|superoxide dismutase 2 (Mn)|superoxide dismutase-2|superoxide dismutatase 2', map_location='53D8-53D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16772785, end_position_on_the_genomic_accession=16773915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36879, CurrentID=0, Status='live', Symbol='Arp53D', Aliases='Dmel_CG5409, ARP53D, Actr53D, CG5409, Dmel\\CG5409, arp5, arp53D', description='Actin-related protein 53D', other_designations='Arp53D-PB|CG5409-PB', map_location='53D8-53D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16774246, end_position_on_the_genomic_accession=16775676, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36880, CurrentID=0, Status='live', Symbol='resilin', Aliases='Dmel_CG15920, CG15920, Dmel\\CG15920', description='CG15920 gene product from transcript CG15920-RB', other_designations='CG15920-PA|CG15920-PB|resilin-PA|resilin-PB', map_location='53D8-53D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16779905, end_position_on_the_genomic_accession=16783171, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36881, CurrentID=0, Status='live', Symbol='CG5522', Aliases='Dmel_ Dmel\\CG5522, RalGPS, RasGPS', description='CG5522 gene product from transcript CG5522-RE', other_designations='CG5522-PA|CG5522-PB|CG5522-PC|CG5522-PD|CG5522-PE|CG5522-PF', map_location='53D8-53D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16783671, end_position_on_the_genomic_accession=16792472, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36882, CurrentID=0, Status='live', Symbol='CG15615', Aliases='Dmel_ Dmel\\CG15615', description='CG15615 gene product from transcript CG15615-RB', other_designations='CG15615-PB|uncharacterized protein', map_location='53D9-53D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16793119, end_position_on_the_genomic_accession=16796407, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36883, CurrentID=0, Status='live', Symbol='CG5550', Aliases='Dmel_ AAF57948, AAL48972, Dmel\\CG5550, FBpp0086184', description='CG5550 gene product from transcript CG5550-RB', other_designations='CG5550-PA|CG5550-PB|Fibrinogen-like|Fibrinogen-related protein', map_location='53D10-53D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16816589, end_position_on_the_genomic_accession=16817577, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772436, CurrentID=0, Status='live', Symbol='CG33773', Aliases='Dmel_ BP1061, Dmel\\CG33773', description='CG33773 gene product from transcript CG33773-RA', other_designations='CG33773-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31035346, end_position_on_the_genomic_accession=31036124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36886, CurrentID=0, Status='live', Symbol='IntS8', Aliases='Dmel_CG5859, CG5859, Dmel\\CG5859, EG:EG0003.5, Int8', description='Integrator 8', other_designations='CG5859-PA|IntS8-PA|integrator 8', map_location='53D13-53D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16850437, end_position_on_the_genomic_accession=16854104, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36887, CurrentID=0, Status='live', Symbol='Fen1', Aliases='Dmel_CG8648, CG8648, DmFEN-1, Dmel\\CG8648, EG:EG0003.3, FEN1, fen1', description='Flap endonuclease 1', other_designations='CG8648-PA|Fen1-PA|flap endonuclease 1|flap endonuclease-1', map_location='53D13-53D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16854025, end_position_on_the_genomic_accession=16855464, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36889, CurrentID=0, Status='live', Symbol='Psi', Aliases='Dmel_CG8912, CG8912, Dmel\\CG8912, EG:EG0003.2, PSI', description='P-element somatic inhibitor', other_designations='CG8912-PA|CG8912-PB|CG8912-PC|P element somatic inhibitor|P-element somatic inhibitor protein|Psi-PA|Psi-PB|Psi-PC', map_location='53D14-53D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16861321, end_position_on_the_genomic_accession=16867714, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36891, CurrentID=0, Status='live', Symbol='CG6426', Aliases='Dmel_ BcDNA:RH01665, Dmel\\CG6426, l(2)k04810, n(2)k04810', description='CG6426 gene product from transcript CG6426-RA', other_designations='CG6426-PA', map_location='53D14-53D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16869402, end_position_on_the_genomic_accession=16871851, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36893, CurrentID=0, Status='live', Symbol='CG6429', Aliases='Dmel_ BcDNA:RH62928, Dmel\\CG6429', description='CG6429 gene product from transcript CG6429-RA', other_designations='CG6429-PA', map_location='53D15-53D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16873015, end_position_on_the_genomic_accession=16873755, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36894, CurrentID=0, Status='live', Symbol='CG6435', Aliases='Dmel_ Dmel\\CG6435', description='CG6435 gene product from transcript CG6435-RA', other_designations='CG6435-PA', map_location='53D15-53D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16873772, end_position_on_the_genomic_accession=16874977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36895, CurrentID=0, Status='live', Symbol='CG6472', Aliases='Dmel_ Dmel\\CG6472', description='CG6472 gene product from transcript CG6472-RB', other_designations='CG6472-PB', map_location='53D15-53E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16880713, end_position_on_the_genomic_accession=16884695, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3772448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772448, CurrentID=0, Status='live', Symbol='His2A:CG33835', Aliases='Dmel_CG33835, CG33835, Dmel\\CG33835', description='CG33835 gene product from transcript CG33835-RA', other_designations='CG33835-PA|His2A:CG33835-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21488378, end_position_on_the_genomic_accession=21488870, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36897, CurrentID=0, Status='live', Symbol='inaC', Aliases='Dmel_CG6518, CG6518, Dmel\\CG6518, Dpkc2, INAC, InaC, PK-C, PKC, PKC 53E, PKC2, PKCi, Pkc2, dPKC53E, dPKC53E(ey), ePKC, eye PKC, eye-PKC, eyePKC, pkc-2', description='inactivation no afterpotential C', other_designations='CG6518-PA|PKC-eye|Protein-kinase-2|eye-PKC|eye-protein-kinase C|eye-specific PKC|inaC-PA|protein kinase C|protein kinase-c2|protein-kinase-C', map_location='2-82 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16898535, end_position_on_the_genomic_accession=16901575, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3772451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772451, CurrentID=0, Status='live', Symbol='CG33723', Aliases='Dmel_ Dmel\\CG33723', description='CG33723 gene product from transcript CG33723-RA', other_designations='CG33723-PA|uncharacterized protein', map_location='30B4-30B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9440011, end_position_on_the_genomic_accession=9440740, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36900, CurrentID=0, Status='live', Symbol='Vha16-4', Aliases='Dmel_CG9013, CG9013, Dmel\\CG9013, vha16-4', description='Vacuolar H[+] ATPase 16kD subunit 4', other_designations='CG9013-PA|Vha16-4-PA|vacuolar H[+] ATPase 16kD subunit 4|vacuolar H[+] ATPase subunit 16-4', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16947658, end_position_on_the_genomic_accession=16948224, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772453, CurrentID=0, Status='live', Symbol='CG9133', Aliases='Dmel_ AT07490, BEST:GH19274, BEST:GH19464, CG32332, Dmel\\CG9133, anon-WO0140519.145, anon-WO0140519.53, bs29f09.y1', description='CG9133 gene product from transcript CG9133-RD', other_designations='CG9133-PB|CG9133-PC|CG9133-PD', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1252012, end_position_on_the_genomic_accession=1255986, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36903, CurrentID=0, Status='live', Symbol='CG6665', Aliases='Dmel_ Dmel\\CG6665', description='CG6665 gene product from transcript CG6665-RC', other_designations='CG6665-PC', map_location='53E2-53E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16957465, end_position_on_the_genomic_accession=16958927, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772456, CurrentID=0, Status='live', Symbol='Indy-2', Aliases='Dmel_CG33933, CG32921, CG33933, CG4961, Dmel\\CG33933, anon-WO0140519.217', description="I'm not dead yet 2", other_designations='CG33933-PA|CG33933-PB|Indy-2-PA|Indy-2-PB', map_location='92C3-92C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20042757, end_position_on_the_genomic_accession=20048716, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  36905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36905, CurrentID=0, Status='live', Symbol='Cbp53E', Aliases='Dmel_CG6702, CALB2, CALB2/calretinin, CG6702, Calb, Cbp33, DCBP, Dcbp, Dcp, Dmel\\CG6702, cbn, cbp53E', description='Calbindin 53E', other_designations='CG6702-PA|CG6702-PB|Cbp53E-PA|Cbp53E-PB|calbindin|calbindin 32|calbindin 53E|calbindin-32|calretinin', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16973927, end_position_on_the_genomic_accession=16996057, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  36907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36907, CurrentID=0, Status='live', Symbol='ste24b', Aliases='Dmel_CG9001, CG9001, Dmel\\CG9001, Dmel\\Ste24, Dmel\\Ste24b', description='ste24b prenyl protease type I', other_designations='CG9001-PA|prenyl protease type I|ste24b-PA', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16998610, end_position_on_the_genomic_accession=17000126, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36908, CurrentID=0, Status='live', Symbol='ste24a', Aliases='Dmel_CG9000, CG9000, Dmel\\CG9000, Dmel\\Ste24, Dmel\\Ste24a', description='prenyl protease type I', other_designations='CG9000-PA|ste24a-PA', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17000244, end_position_on_the_genomic_accession=17001941, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36909, CurrentID=0, Status='live', Symbol='CG6796', Aliases='Dmel_ Dmel\\CG6796', description='CG6796 gene product from transcript CG6796-RA', other_designations='CG6796-PA', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17001994, end_position_on_the_genomic_accession=17003599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36911, CurrentID=0, Status='live', Symbol='CG6805', Aliases='Dmel_ BEST:CK01299, CK01299, Dmel\\CG6805, IPP', description='CG6805 gene product from transcript CG6805-RA', other_designations='CG6805-PA|CG6805-PB', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17005451, end_position_on_the_genomic_accession=17007501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36912, CurrentID=0, Status='live', Symbol='Ehbp1', Aliases='Dmel_CG15609, 2R11, CG15609, Dmel\\CG15609, EHBP1, dEHBP-1, dEHBP1', description='Eps15 homology domain containing protein-binding protein 1', other_designations='CG15609-PB|CG15609-PF|CG15609-PG|CG15609-PI|Ehbp1-PB|Ehbp1-PF|Ehbp1-PG|Ehbp1-PI', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17007674, end_position_on_the_genomic_accession=17015894, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  36913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36913, CurrentID=0, Status='live', Symbol='CG8963', Aliases='Dmel_ Dmel\\CG8963', description='CG8963 gene product from transcript CG8963-RB', other_designations='CG8963-PA|CG8963-PB|CG8963-PC|CG8963-PE', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17016226, end_position_on_the_genomic_accession=17019522, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36914, CurrentID=0, Status='live', Symbol='Dark', Aliases='Dmel_CG6829, APAF-1, APAF1, ARK, Apaf, Apaf-1, Apaf1, Ark, CG6829, D-Apaf-1, DARK, Dapaf-1, Dapaf-1/HAC-1/Apaf-I, Dark/Dapaf-1/HAC1, Dark/Hac-1/dApaf-1, Dark/Hac-1/dApaf1, Dmel\\CG6829, Hac-1, Hac-1/Dark, Hac1, T1, anon-53Fa, apaf-1, apaf1, arc, ark, dAPAF-1, dApaf-1, dApaf-1/DARK/HAC-1, dApaf1, dArk, dapaf, dapaf-1, dapaf-1L, dapaf-1S, dark, dark/dapaf-1/hac-1, dark/hac-1/dapaf-1, hac-1, hac1, l(2)SH0173, Dark', description='Death-associated APAF1-related killer', other_designations='Apaf-1 related|Apaf-1 related killer|Apaf-1-related killer|Apaf-1-related-killer|CG6829-PA|CG6829-PB|Dark-PA|Dark-PB|apaf1-related killer|dark|drosophila Apaf-1-related killer|lethal (2) SH0173|transcription unit 1', map_location='54A2-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17020027, end_position_on_the_genomic_accession=17026703, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36915, CurrentID=0, Status='live', Symbol='RhoGEF2', Aliases='Dmel_CG9635, CG9635, DRhoGEF, DRhoGEF2, Dmel\\CG9635, DrhoGEF2, Gef2, RG2, Rho-GEF2, RhoGEF, RhoGF2, RhoGef2, Su(Rho1)2B, T2, dRhoGEF, dRhoGEF2, l(2)04291, rhoGEF2, rhogef2, shar', description='Rho guanine nucleotide exchange factor 2', other_designations='CG9635-PD|CG9635-PE|CG9635-PF|CG9635-PG|CG9635-PH|RhoGEF2-PD|RhoGEF2-PE|RhoGEF2-PF|RhoGEF2-PG|RhoGEF2-PH|akkordion|guanine nucleotide exchange factor 2|shar pei|transcription unit 2', map_location='53E4-53F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17027190, end_position_on_the_genomic_accession=17044653, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  36916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36916, CurrentID=0, Status='live', Symbol='CG9640', Aliases='Dmel_ Dmel\\CG9640', description='CG9640 gene product from transcript CG9640-RB', other_designations='CG9640-PA|CG9640-PB', map_location='53F1-53F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17044743, end_position_on_the_genomic_accession=17046768, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36917, CurrentID=0, Status='live', Symbol='CG9642', Aliases='Dmel_ Dmel\\CG9642', description='CG9642 gene product from transcript CG9642-RA', other_designations='CG9642-PA', map_location='53F1-53F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17046839, end_position_on_the_genomic_accession=17048794, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36918, CurrentID=0, Status='live', Symbol='CG9646', Aliases='Dmel_ Dmel\\CG9646', description='CG9646 gene product from transcript CG9646-RA', other_designations='CG9646-PA|CG9646-PB', map_location='53F2-53F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17048875, end_position_on_the_genomic_accession=17053855, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36919, CurrentID=0, Status='live', Symbol='fat-spondin', Aliases='Dmel_CG6953, CG6953, CT21501, Dmel\\CG6953, Fat-spondin', description='CG6953 gene product from transcript CG6953-RA', other_designations='CG6953-PA|fat-spondin-PA', map_location='53F2-53F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17054002, end_position_on_the_genomic_accession=17059299, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36920, CurrentID=0, Status='live', Symbol='tef', Aliases='Dmel_CG8961, CG8961, Dmel\\CG8961, l(2)k15914', description='teflon', other_designations='CG8961-PA|lethal (2) k15914|tef-PA', map_location='2-80.0 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17054546, end_position_on_the_genomic_accession=17056835, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36921, CurrentID=0, Status='live', Symbol='CG8950', Aliases='Dmel_ Dmel\\CG8950', description='CG8950 gene product from transcript CG8950-RB', other_designations='CG8950-PA|CG8950-PB', map_location='53F3-53F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17059331, end_position_on_the_genomic_accession=17062634, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36922, CurrentID=0, Status='live', Symbol='CG6967', Aliases='Dmel_ CT21565, Dmel\\CG6967', description='CG6967 gene product from transcript CG6967-RD', other_designations='CG6967-PC|CG6967-PD', map_location='53F4-53F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17062775, end_position_on_the_genomic_accession=17066546, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36924, CurrentID=0, Status='live', Symbol='CG30460', Aliases='Dmel_ CG15607, CG15608, Dmel\\CG30460, anon-WO0140519.237', description='CG30460 gene product from transcript CG30460-RE', other_designations='CG30460-PA|CG30460-PB|CG30460-PC|CG30460-PD|CG30460-PE|CG30460-PF|CG30460-PG|CG30460-PH|CG30460-PI|uncharacterized protein', map_location='53F4-53F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17065971, end_position_on_the_genomic_accession=17087633, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38922, CurrentID=0, Status='live', Symbol='CG13678', Aliases='Dmel_ BcDNA:RE24439, Dmel\\CG13678', description='CG13678 gene product from transcript CG13678-RA', other_designations='CG13678-PA|uncharacterized protein', map_location='66C1-66C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8222141, end_position_on_the_genomic_accession=8222953, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36926, CurrentID=0, Status='live', Symbol='CG6984', Aliases='Dmel_ Dmel\\CG6984', description='CG6984 gene product from transcript CG6984-RA', other_designations='CG6984-PA', map_location='53F7-53F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17091984, end_position_on_the_genomic_accession=17093070, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36927, CurrentID=0, Status='live', Symbol='GstS1', Aliases='Dmel_CG8938, CG8938, DmGST-2, DmGST2, DmGSTS1, DmGSTS1-1, DmGst-2, DmGstS1-1, Dmel\\CG8938, Est2, GST, GST S1, GST-2, GST2, GSTII, GSTS1, Gst2-1, S1, gstD2, gsts1, l(2)04227, l(2)k08805, lincRNA.330, n(2)06253, GstS1', description='Glutathione S transferase S1', other_designations='CG8938-PA|CG8938-PB|CG8938-PC|CG8938-PD|Glutathione-S-Transferase 2|GstS1-PA|GstS1-PB|GstS1-PC|GstS1-PD|glutathione S transferase 2|glutathione S transferase S1|glutathione S-transferase|glutathione S-transferase 2|glutathione S-transferase S1|glutathione S-transferase-2|glutathione S-transferase-S1|glutathione transferase-related protein|glutathione-S-transferase|glutathione-S-transferase (GST) S1', map_location='53F7-53F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17093246, end_position_on_the_genomic_accession=17099501, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36929, CurrentID=0, Status='live', Symbol='CG30456', Aliases='Dmel_ CG15612, CG15613, Dmel\\CG30456', description='CG30456 gene product from transcript CG30456-RB', other_designations='CG30456-PB|CG30456-PC|CG30456-PD|uncharacterized protein', map_location='53F8-53F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17100241, end_position_on_the_genomic_accession=17107542, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36930, CurrentID=0, Status='live', Symbol='CG15611', Aliases='Dmel_ Dmel\\CG15611', description='CG15611 gene product from transcript CG15611-RA', other_designations='CG15611-PA|CG15611-PB|uncharacterized protein', map_location='53F10-53F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17108844, end_position_on_the_genomic_accession=17114360, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36931, CurrentID=0, Status='live', Symbol='Spn53F', Aliases='Dmel_CG10956, CG10956, Dmel\\CG10956', description='Serpin 53F', other_designations='CG10956-PB|Spn53F-PB|serpin 53F', map_location='53F12-53F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17122898, end_position_on_the_genomic_accession=17124237, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36932, CurrentID=0, Status='live', Symbol='Amy-d', Aliases='Dmel_CG17876, AMY, Amy, Amy(d)istal, Amy1, Amy3, Amy6, AmyB, AmyD, CG17876, Dmel\\CG17876, alpha-AMY, alphaAmy, amy, amy-d, amyB(3), Amy-d', description='Amylase distal', other_designations='Amy-d-PA|Amylase-B|CG17876-PA|alpha-amylase|alpha-amylase distal|amylase|amylase distal', map_location='53F12-53F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17124764, end_position_on_the_genomic_accession=17126369, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36933, CurrentID=0, Status='live', Symbol='CG15605', Aliases='Dmel_ BcDNA:AT23338, Dmel\\CG15605', description='CG15605 gene product from transcript CG15605-RA', other_designations='CG15605-PA|uncharacterized protein', map_location='53F12-53F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17126410, end_position_on_the_genomic_accession=17127023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36934, CurrentID=0, Status='live', Symbol='Cda9', Aliases='Dmel_CG15918, CG15918, DmCDA9, Dmel\\CG15918', description='Chitin deacetylase-like 9', other_designations='CG15918-PA|CG15918-PB|Cda9-PA|Cda9-PB|chitin deacetylase-like 9', map_location='53F13-53F13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17128480, end_position_on_the_genomic_accession=17130334, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36935, CurrentID=0, Status='live', Symbol='CG11400', Aliases='Dmel_ Dmel\\CG11400', description='CG11400 gene product from transcript CG11400-RB', other_designations='CG11400-PA|CG11400-PB|uncharacterized protein', map_location='54A1-54A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17134129, end_position_on_the_genomic_accession=17135345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36936, CurrentID=0, Status='live', Symbol='Gbp', Aliases='Dmel_CG15917, CG15917, Dmel\\CG15917, GBP', description='Growth-blocking peptide', other_designations='CG15917-PA|Gbp-PA|growth-blocking peptide-like factor', map_location='54A1-54A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17136451, end_position_on_the_genomic_accession=17136903, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36937, CurrentID=0, Status='live', Symbol='CG11395', Aliases='Dmel_ Dmel\\CG11395', description='CG11395 gene product from transcript CG11395-RB', other_designations='CG11395-PA|CG11395-PB|uncharacterized protein', map_location='54A1-54A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17137240, end_position_on_the_genomic_accession=17139595, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36938, CurrentID=0, Status='live', Symbol='CG17290', Aliases='Dmel_ BcDNA:RE15976, Dmel\\CG17290', description='CG17290 gene product from transcript CG17290-RC', other_designations='CG17290-PA|CG17290-PC|uncharacterized protein', map_location='54A2-54A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17154589, end_position_on_the_genomic_accession=17157125, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36939, CurrentID=0, Status='live', Symbol='CG17287', Aliases='Dmel_ Dmel\\CG17287', description='CG17287 gene product from transcript CG17287-RA', other_designations='CG17287-PA|uncharacterized protein', map_location='54A2-54A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17156534, end_position_on_the_genomic_accession=17158316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36940, CurrentID=0, Status='live', Symbol='CG30457', Aliases='Dmel_CG30457, CG17288-PA, Dmel\\CG30457', description='CG30457 gene product from transcript CG30457-RA', other_designations='CG30457-PA|uncharacterized protein', map_location='54A2-54A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17160122, end_position_on_the_genomic_accession=17160829, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36941, CurrentID=0, Status='live', Symbol='CG10953', Aliases='Dmel_CG10953-PA, Dmel\\CG10953', description='CG10953 gene product from transcript CG10953-RB', other_designations='CG10953-PA|CG10953-PB|uncharacterized protein', map_location='54A2-54A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17165084, end_position_on_the_genomic_accession=17166643, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36942, CurrentID=0, Status='live', Symbol='CG10950', Aliases='Dmel_ Dmel\\CG10950', description='CG10950 gene product from transcript CG10950-RA', other_designations='CG10950-PA|uncharacterized protein', map_location='54A2-54A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17195203, end_position_on_the_genomic_accession=17197159, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36943, CurrentID=0, Status='live', Symbol='CG18469', Aliases='Dmel_ BcDNA:AT30190, Dmel\\CG18469', description='CG18469 gene product from transcript CG18469-RA', other_designations='CG18469-PA|uncharacterized protein', map_location='54B1-54B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17245769, end_position_on_the_genomic_accession=17246277, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772496, CurrentID=0, Status='live', Symbol='His2B:CG33880', Aliases='Dmel_CG33880, CG33880, Dmel\\CG33880', description='CG33880 gene product from transcript CG33880-RA', other_designations='CG33880-PA|His2B:CG33880-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21507843, end_position_on_the_genomic_accession=21508437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36945, CurrentID=0, Status='live', Symbol='mbl', Aliases='Dmel_CG33197, BG01127, CG10941, CG14477, CG33197, Dmel\\CG33197, Mbl, chaste, chst, cht, l(2)01038, l(2)k01212, l(2)k05507, mb1, mbn, mm, n(2)k07110', description='CG33197 gene product from transcript CG33197-RP', other_designations='BG01127|CG33197-PA|CG33197-PB|CG33197-PD|CG33197-PH|CG33197-PJ|CG33197-PK|CG33197-PM|CG33197-PN|CG33197-PO|CG33197-PP|CG33197-PR|CG33197-PS|CG33197-PT|MUSCLEBLIND|chaste|mbl-PA|mbl-PB|mbl-PD|mbl-PH|mbl-PJ|mbl-PK|mbl-PM|mbl-PN|mbl-PO|mbl-PP|mbl-PR|mbl-PS|mbl-PT|mindmelt', map_location='54B1-54B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17216549, end_position_on_the_genomic_accession=17379376, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  3772499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772499, CurrentID=0, Status='live', Symbol='CG33769', Aliases='Dmel_ CG14458, Dmel\\CG33769', description='CG33769 gene product from transcript CG33769-RA', other_designations='CG33769-PA|uncharacterized protein', map_location='79F3-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22647927, end_position_on_the_genomic_accession=22648519, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36948, CurrentID=0, Status='live', Symbol='CG6568', Aliases='Dmel_ Dmel\\CG6568', description='CG6568 gene product from transcript CG6568-RA', other_designations='CG6568-PA|CG6568-PB', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17405304, end_position_on_the_genomic_accession=17407427, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36950, CurrentID=0, Status='live', Symbol='CG30101', Aliases='Dmel_ CG15901, CG6564, Dmel\\CG30101, FBgn0050101, vajk-4', description='CG30101 gene product from transcript CG30101-RA', other_designations='CG30101-PA|uncharacterized protein', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17408258, end_position_on_the_genomic_accession=17410204, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36951, CurrentID=0, Status='live', Symbol='Prosalpha5', Aliases='Dmel_CG10938, 154644_at, Alpha-5, CG10938, Dmel\\CG10938, PRCZ, PSA5_DROME, PSMA5, ProsMA5, alpha-5, alpha5, alpha5_dm, prosMA5', description='Proteasome alpha5 subunit', other_designations='20s proteasome|CG10938-PA|CG10938-PB|Prosalpha5-PA|Prosalpha5-PB|alpha-subunit of 20S proteasome|alpha5|proteasome alpha subunit|proteasome alpha5 subunit', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17412767, end_position_on_the_genomic_accession=17413769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36952, CurrentID=0, Status='live', Symbol='cnk', Aliases='Dmel_CG6556, CG6556, CNK, Cnk, Dmel\\CG6556, EC2-3, EK2-3, anon-WO0140519.129, anon-WO0257455.27, l(2)k16314', description='connector enhancer of ksr', other_designations='CG6556-PA|CG6556-PB|cnk-PA|cnk-PB|connector enhancer of KSR|enhancer of KDN', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17413970, end_position_on_the_genomic_accession=17420007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36953, CurrentID=0, Status='live', Symbol='l(2)k01209', Aliases='Dmel_CG4798, CG4798, Dmel\\CG4798, l(2)K01209, l(2)k08901', description='lethal (2) k01209', other_designations='CG4798-PA|CG4798-PB|CG4798-PC|CG4798-PD|l(2)k01209-PA|l(2)k01209-PB|l(2)k01209-PC|l(2)k01209-PD', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17420407, end_position_on_the_genomic_accession=17425679, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36954, CurrentID=0, Status='live', Symbol='CG6550', Aliases='Dmel_ Dmel\\CG6550', description='CG6550 gene product from transcript CG6550-RA', other_designations='CG6550-PA', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17425670, end_position_on_the_genomic_accession=17427511, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36955, CurrentID=0, Status='live', Symbol='Mtap', Aliases='Dmel_CG4802, CG4802, Dmel\\CG4802, MTAP', description='Methylthioadenosine phosphorylase ortholog', other_designations='CG4802-PA|CG4802-PB|CG4802-PC|Mtap-PA|Mtap-PB|Mtap-PC|methylthioadenosine phosphorylase ortholog', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17427795, end_position_on_the_genomic_accession=17429315, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36956, CurrentID=0, Status='live', Symbol='Bap55', Aliases='Dmel_CG6546, BAF53, BAP47, BAP55, Baf53, CG6546, Dmel\\CG6546, bap55', description='Brahma associated protein 55kD', other_designations='BRM-associated protein 55|Bap55-PA|CG6546-PA|brahma associated protein 55 kD|brahma associated protein 55kD|brahma associated protein of 55kD', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17429353, end_position_on_the_genomic_accession=17430963, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36957, CurrentID=0, Status='live', Symbol='Lhr', Aliases='Dmel_CG18468, CG18468, Dmel\\CG18468, HP3, LHR, LHR[[mel]], mel-Lhr', description='Lethal hybrid rescue', other_designations='CG18468-PA|Lhr-PA|lethal hybrid rescue', map_location='54B7-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17431032, end_position_on_the_genomic_accession=17432219, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36958, CurrentID=0, Status='live', Symbol='CG18467', Aliases='Dmel_ Dmel\\CG18467', description='CG18467 gene product from transcript CG18467-RA', other_designations='CG18467-PA|uncharacterized protein', map_location='54B14-54B15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17440297, end_position_on_the_genomic_accession=17442483, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36959, CurrentID=0, Status='live', Symbol='EDTP', Aliases='Dmel_CG6542, BcDNA:LD23181, CG6542, D-EDTP, Dmel\\CG6542, anon-WO0118547.124', description='Egg-derived tyrosine phosphatase', other_designations='CG6542-PA|CG6542-PC|EDTP-PA|EDTP-PC', map_location='54B7-54B15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17432173, end_position_on_the_genomic_accession=17444938, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36960, CurrentID=0, Status='live', Symbol='mthl4', Aliases='Dmel_CG6536, CG6536, Dmel\\CG6536, anon-WO0170980.55, anon-WO0170980.56', description='methuselah-like 4', other_designations='CG6536-PB|Mth-like 4|mthl4-PB', map_location='54B15-54B15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17445185, end_position_on_the_genomic_accession=17447483, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36961, CurrentID=0, Status='live', Symbol='mthl3', Aliases='Dmel_CG6530, 104/8, BEST:GM02553, CG6530, Dmel\\CG6530, GENE 1, Mth-3, anon-WO0170980.52, anon-WO0170980.53, gene 1, l(2)k10408, mthl-3', description='methuselah-like 3', other_designations='CG6530-PA|CG6530-PB|CG6530-PC|Mth-like 3|mthl3-PA|mthl3-PB|mthl3-PC', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17448533, end_position_on_the_genomic_accession=17452598, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  36962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36962, CurrentID=0, Status='live', Symbol='CG10764', Aliases='Dmel_ Dmel\\CG10764, FBgn0034221, GENE 2, SP77, anon-54Ba, gene 2', description='CG10764 gene product from transcript CG10764-RA', other_designations='CG10764-PA|uncharacterized protein', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17453399, end_position_on_the_genomic_accession=17455729, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36963, CurrentID=0, Status='live', Symbol='robl', Aliases='Dmel_CG10751, CG10751, DLC, Dmel\\CG10751, LC7, l(2)k10408k, robl', description='roadblock', other_designations='CG10751-PA|dynein light chain|dynein light chain LC7|dynein, light chain, roadblock-type 2|robl-PA', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17456472, end_position_on_the_genomic_accession=17457129, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36964, CurrentID=0, Status='live', Symbol='CG14478', Aliases='Dmel_ BEST:LD13441, Dmel\\CG14478, GENE 4, gene 4', description='CG14478 gene product from transcript CG14478-RD', other_designations='CG14478-PA|CG14478-PB|CG14478-PC|CG14478-PD|uncharacterized protein', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17457663, end_position_on_the_genomic_accession=17465795, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36965, CurrentID=0, Status='live', Symbol='Tes', Aliases='Dmel_CG6522, CG6522, Dmel\\CG6522, dTES', description='Testin ortholog', other_designations='CG6522-PA|Tes-PA|testin ortholog', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17465879, end_position_on_the_genomic_accession=17469177, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  36966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36966, CurrentID=0, Status='live', Symbol='qkr54B', Aliases='Dmel_CG4816, CG4816, Dmel\\CG4816, QKR54B, SAM, SAM50, Sam50, p62GAP, qrk54B', description='quaking related 54B', other_designations='CG4816-PA|CG4816-PB|CG4816-PD|CG4816-PE|qkr54B-PA|qkr54B-PB|qkr54B-PD|qkr54B-PE', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17469624, end_position_on_the_genomic_accession=17475129, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  36967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36967, CurrentID=0, Status='live', Symbol='insb', Aliases='Dmel_CG6520, CG6520, Dmel\\CG6520', description='insensible', other_designations='CG6520-PA|CG6520-PB|insb-PA|insb-PB', map_location='54B16-54B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17475214, end_position_on_the_genomic_accession=17476272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36968, CurrentID=0, Status='live', Symbol='veil', Aliases='Dmel_CG4827, CG4827, Dmel\\CG4827, NT5E-1', description='CG4827 gene product from transcript CG4827-RB', other_designations="5' nucleotidase precursor|5'-nucleotidase-precursor|CG4827-PA|CG4827-PB|veil-PA|veil-PB", map_location='54B16-54B17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17478158, end_position_on_the_genomic_accession=17482118, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36969, CurrentID=0, Status='live', Symbol='CG30103', Aliases='Dmel_ BcDNA:AT08275, CG4837, Dmel\\CG30103', description='CG30103 gene product from transcript CG30103-RA', other_designations='CG30103-PA|uncharacterized protein', map_location='54B17-54B17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17493695, end_position_on_the_genomic_accession=17495998, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3772524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772524, CurrentID=0, Status='live', Symbol='CG33658', Aliases='Dmel_ BP1056, Dmel\\CG33658', description='CG33658 gene product from transcript CG33658-RA', other_designations='CG33658-PA|uncharacterized protein', map_location='96A25-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24840509, end_position_on_the_genomic_accession=24841163, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36973, CurrentID=0, Status='live', Symbol='CG4847', Aliases='Dmel_ BEST:GH19547, Dmel\\CG4847', description='CG4847 gene product from transcript CG4847-RA', other_designations='CG4847-PA|CG4847-PC|CG4847-PD|CG4847-PE|cathepsinK', map_location='54C1-54C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17511513, end_position_on_the_genomic_accession=17513855, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36974, CurrentID=0, Status='live', Symbol='CG4853', Aliases='Dmel_ Dmel\\CG4853', description='CG4853 gene product from transcript CG4853-RB', other_designations='CG4853-PA|CG4853-PB|CG4853-PC', map_location='54C1-54C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17514217, end_position_on_the_genomic_accession=17517445, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36975, CurrentID=0, Status='live', Symbol='APC10', Aliases='Dmel_CG11419, Apc10, CG11419, Dmel\\CG11419', description='Anaphase Promoting Complex subunit 10', other_designations='APC10-PA|CG11419-PA|anaphase promoting complex subunit 10', map_location='54C1-54C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17517261, end_position_on_the_genomic_accession=17518186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36976, CurrentID=0, Status='live', Symbol='CG4866', Aliases='Dmel_ Dmel\\CG4866, mRpS4', description='CG4866 gene product from transcript CG4866-RA', other_designations='CG4866-PA', map_location='54C1-54C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17518223, end_position_on_the_genomic_accession=17519027, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36978, CurrentID=0, Status='live', Symbol='Patronin', Aliases='Dmel_CG33130, BcDNA:LD17191, CG18459, CG18460, CG18462, CG30102, CG33130, CG6516, Dmel\\CG33130, Ssp4, l(2)k07433, ssp4', description='CG33130 gene product from transcript CG33130-RN', other_designations='CG33130-PA|CG33130-PB|CG33130-PC|CG33130-PD|CG33130-PE|CG33130-PF|CG33130-PI|CG33130-PJ|CG33130-PK|CG33130-PL|CG33130-PM|CG33130-PN|CG33130-PO|Patronin-PA|Patronin-PB|Patronin-PC|Patronin-PD|Patronin-PE|Patronin-PF|Patronin-PI|Patronin-PJ|Patronin-PK|Patronin-PL|Patronin-PM|Patronin-PN|Patronin-PO|lethal (2) k07433|patronin|short spindle 4', map_location='54C1-54C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17519024, end_position_on_the_genomic_accession=17535866, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  36981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36981, CurrentID=0, Status='live', Symbol='eIF3-S9', Aliases='Dmel_CG4878, CG4878, Dmel\\CG4878, cg4878, eIF3 S9, eIF3-s9, eIF3b', description='CG4878 gene product from transcript CG4878-RB', other_designations='CG4878-PA|CG4878-PB|eIF3-S9-PA|eIF3-S9-PB', map_location='54C3-54C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17536213, end_position_on_the_genomic_accession=17539142, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  36982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36982, CurrentID=0, Status='live', Symbol='CG43370', Aliases='Dmel_ CG18432, CG18631, Dmel\\CG43370, Dmel_CG18432, Dmel_CG18631', description='CG43370 gene product from transcript CG43370-RD', other_designations='CG43370-PD|uncharacterized protein', map_location='54C3-54C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17539322, end_position_on_the_genomic_accession=17544353, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  36984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36984, CurrentID=0, Status='live', Symbol='cyp33', Aliases='Dmel_CG4886, CG4886, CYP33, Cyp33, Dcyp33, Dmel\\CG4886', description='cyclophilin-33', other_designations='CG4886-PA|cyp33-PA', map_location='54C3-54C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17545222, end_position_on_the_genomic_accession=17546400, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36985, CurrentID=0, Status='live', Symbol='RpL18A', Aliases='Dmel_CG6510, CG6510, Dmel\\CG6510, L18, L18A, L18a, RPL18a, RpL18a, anon-EST:Posey75', description='Ribosomal protein L18A', other_designations='CG6510-PA|RpL18A-PA|ribosomal protein L18A', map_location='54C3-54C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17546406, end_position_on_the_genomic_accession=17547462, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36986, CurrentID=0, Status='live', Symbol='MESR4', Aliases='Dmel_CG4903, CG4903, Dmel\\CG4903, EP(2)0988, EP0386, EP0988, EP988, MSR4, Mesr4', description='Misexpression suppressor of ras 4', other_designations='CG4903-PA|CG4903-PB|CG4903-PC|CG4903-PD|MESR4-PA|MESR4-PB|MESR4-PC|MESR4-PD|misexpression of Ras 4|misexpression suppressor of Ras4|misexpression suppressor of ras 4', map_location='54C3-54C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17547769, end_position_on_the_genomic_accession=17556289, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772540, CurrentID=0, Status='live', Symbol='CG33635', Aliases='Dmel_ BP1075, Dmel\\CG33635, HDC00027', description='CG33635 gene product from transcript CG33635-RA', other_designations='CG33635-PA|uncharacterized protein', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=157836, end_position_on_the_genomic_accession=158664, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36989, CurrentID=0, Status='live', Symbol='Ns2', Aliases='Dmel_CG6501, CG6501, Dmel\\CG6501, NS2, Ngp, Ngp1, dmNS2, ns2', description='Nucleostemin ortholog (H. sapiens) 2', other_designations='CG6501-PA|Ngp|Ns2-PA|nucleostemin 2|nucleostemin ortholog 2', map_location='54C8-54C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17565779, end_position_on_the_genomic_accession=17568456, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36990, CurrentID=0, Status='live', Symbol='POSH', Aliases='Dmel_CG4909, 155116_at, BEST:GM09451, CG4909, DPOSH, Dmel\\CG4909, GM09451, dPOSH, l(2)k15815, posh', description='Plenty of SH3s', other_designations='CG4909-PA|POSH-PA|Plenty-of-SH3s|plenty of SH3 domains|plenty of SH3s', map_location='54C9-54C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17568732, end_position_on_the_genomic_accession=17572322, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38933, CurrentID=0, Status='live', Symbol='CG7194', Aliases='Dmel_ BEST:CK00048, BEST:CK01174, CK01174, Dmel\\CG7194, ck00048', description='CG7194 gene product from transcript CG7194-RA', other_designations='CG7194-PA', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8302364, end_position_on_the_genomic_accession=8304059, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36992, CurrentID=0, Status='live', Symbol='Rab4', Aliases='Dmel_CG4921, CG 4921, CG4921, DRAB4, Dm Dmel\\CG4921, Q7KY04, RAB4c, rab4', description='CG4921 gene product from transcript CG4921-RA', other_designations='CG4921-PA|CG4921-PB|Rab-protein 4|Rab4-PA|Rab4-PB', map_location='54C9-54C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17573462, end_position_on_the_genomic_accession=17574979, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36993, CurrentID=0, Status='live', Symbol='Dcr-2', Aliases='Dmel_CG6493, CG6493, DCR-2, DCR2, DICER, Dcr, Dcr2, Dicer, Dicer2, Dmel\\CG6493, cg6493, dcr, dcr-2, dcr2, dic2, dicer, dicer2', description='Dicer-2', other_designations='CG6493-PA|CG6493-PB|Dcr-2-PA|Dcr-2-PB|Dicer2|dicer|dicer 2|dicer-2', map_location='54C10-54C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17574979, end_position_on_the_genomic_accession=17581525, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  36994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36994, CurrentID=0, Status='live', Symbol='CG6484', Aliases='Dmel_ Dmel\\CG6484', description='CG6484 gene product from transcript CG6484-RA', other_designations='CG6484-PA', map_location='54C10-54C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17582007, end_position_on_the_genomic_accession=17584012, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36995, CurrentID=0, Status='live', Symbol='CG14483', Aliases='Dmel_ Dmel\\CG14483', description='CG14483 gene product from transcript CG14483-RC', other_designations='CG14483-PA|CG14483-PB|CG14483-PC|uncharacterized protein', map_location='54C11-54C11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17585195, end_position_on_the_genomic_accession=17586294, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36996, CurrentID=0, Status='live', Symbol='RhoGAP54D', Aliases='Dmel_CG6477, CG6477, Dmel\\CG6477', description='Rho GTPase activating protein at 54D', other_designations='CG6477-PA|CG6477-PC|RhoGAP54D-PA|RhoGAP54D-PC', map_location='54C11-54C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17586425, end_position_on_the_genomic_accession=17590884, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36997, CurrentID=0, Status='live', Symbol='icln', Aliases='Dmel_CG4924, CG4924, Dmel\\CG4924, pICln', description='CG4924 gene product from transcript CG4924-RA', other_designations='CG4924-PA|icln-PA', map_location='54C12-54C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17590964, end_position_on_the_genomic_accession=17591881, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36998, CurrentID=0, Status='live', Symbol='ND-51L1', Aliases='Dmel_CG11423, CG11423, Dmel\\CG11423, NUBM', description='NADH dehydrogenase (ubiquinone) 51 kDa subunit-like 1', other_designations='CG11423-PA|NADH-UBIQUINONE OXIDOREDUCTASE 51 KDA SUBUNIT|NADH:ubiquinone oxidoreductase 51 kDa subunit|ND-51L1-PA', map_location='54C12-54C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17595267, end_position_on_the_genomic_accession=17597409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36999, CurrentID=0, Status='live', Symbol='Smurf', Aliases='Dmel_CG4943, CG4943, D-smurf, DSmurf, Dmel\\CG4943, Dsmurf, LACK, Lack, SMURF ubiquitin ligase, d-smurf, dSmurf, dSmurf1, lack, smurf, Smurf', description='SMAD specific E3 ubiquitin protein ligase', other_designations='CG4943-PA|CG4943-PB|Smurf-PA|Smurf-PB|Smurf1|lethal with a checkpoint kinase', map_location='54C12-54D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17592791, end_position_on_the_genomic_accession=17604461, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3772552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772552, CurrentID=0, Status='live', Symbol='His3:CG33833', Aliases='Dmel_CG33833, CG33833, Dmel\\CG33833', description='CG33833 gene product from transcript CG33833-RA', other_designations='CG33833-PA|His3:CG33833-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21484681, end_position_on_the_genomic_accession=21485209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37002, CurrentID=0, Status='live', Symbol='Oxp', Aliases='Dmel_CG18186, CG18186, Dmel\\CG18186, OXP', description='Oxpecker', other_designations='CG18186-PA|Oxp-PA|oxpecker', map_location='54D2-54D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17633797, end_position_on_the_genomic_accession=17634167, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37004, CurrentID=0, Status='live', Symbol='CCHa1-R', Aliases='Dmel_CG30106, CCHa1r, CCHamide1r, CG14484, CG18192, CG30106, Dmel\\CG30106', description='CCHamide-1 receptor', other_designations='ASB-receptor|CCHa1-R-PA|CG30106-PA', map_location='54D3-54D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17646848, end_position_on_the_genomic_accession=17665068, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37005, CurrentID=0, Status='live', Symbol='eIF3-S8', Aliases='Dmel_CG4954, CG4954, Dmel\\CG4954, eIF3-s8, eIF3c', description='CG4954 gene product from transcript CG4954-RB', other_designations='CG4954-PA|CG4954-PB|eIF3-S8-PA|eIF3-S8-PB', map_location='54D4-54D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17665678, end_position_on_the_genomic_accession=17669275, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37006, CurrentID=0, Status='live', Symbol='P32', Aliases='Dmel_CG6459, CG 6459, CG6459, Dmel\\CG6459, TAP, dp32, p32', description='CG6459 gene product from transcript CG6459-RA', other_designations='CG6459-PA|P32-PA', map_location='54D4-54D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17671309, end_position_on_the_genomic_accession=17672640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772559, CurrentID=0, Status='live', Symbol='Skeletor', Aliases='Dmel_CG43161, 2Ab3, CG14681, CG14682, CG31261, CG32922, CG33676, CG43161, Dmel\\CG43161, Dmel_CG14681, Dmel_CG33676, skeletor', description='CG43161 gene product from transcript CG43161-RD', other_designations='CG43161-PB|CG43161-PC|CG43161-PD|CG43161-PE|Skeletor-PB|Skeletor-PC|Skeletor-PD|Skeletor-PE|skeletor', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10686943, end_position_on_the_genomic_accession=10696270, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3772560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772560, CurrentID=0, Status='live', Symbol='SLIRP1', Aliases='Dmel_CG33714, CG1704, CG33071, CG33071-ORFA, CG33071ORFA, CG33714, DmSLIRP1, Dmel\\CG33714, Dromel_CG33713_FBtr0091704_uORF, cg1704', description='SRA stem-loop interacting RNA binding protein 1', other_designations='CG33714-PA|CG33714-PB|SLIRP1-PA|SLIRP1-PB', map_location='19F4-19F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21317936, end_position_on_the_genomic_accession=21319296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37009, CurrentID=0, Status='live', Symbol='HPS4', Aliases='Dmel_CG4966, CG4966, Dmel\\CG4966, NP_725707, dHPS4', description='Hermansky-Pudlak Syndrome 4 ortholog (H. sapiens)', other_designations='CG4966-PA|CG4966-PB|HPS4-PA|HPS4-PB|Hermansky-Pudlak syndrome 4 ortholog', map_location='54D5-54D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17683574, end_position_on_the_genomic_accession=17690822, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37010, CurrentID=0, Status='live', Symbol='swi2', Aliases='Dmel_CG14485, CG14485, Dmel\\CG14485', description='CG14485 gene product from transcript CG14485-RA', other_designations='CG14485-PA|swi2-PA', map_location='54D5-54D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17690275, end_position_on_the_genomic_accession=17695353, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37011, CurrentID=0, Status='live', Symbol='rdgBbeta', Aliases='Dmel_CG17818, BcDNA.LD14189, BcDNA:LD14189, CG17818, Dmel\\CG17818, D RdgB-beta, RdgBbeta', description='CG17818 gene product from transcript CG17818-RA', other_designations='CG17818-PA|rdgBbeta-PA', map_location='54D6-54D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17695776, end_position_on_the_genomic_accession=17698362, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37012, CurrentID=0, Status='live', Symbol='CG6424', Aliases='Dmel_ BcDNA:GH05095, Dmel\\CG6424, GH05095, anon-WO0118547.96, anon-WO0153538.32, anon-WO0153538.33, anon-WO0153538.34', description='CG6424 gene product from transcript CG6424-RA', other_designations='CG6424-PA|CG6424-PB|CG6424-PC|CG6424-PD', map_location='54E1-54E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17703335, end_position_on_the_genomic_accession=17719346, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3772565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772565, CurrentID=0, Status='live', Symbol='His2A:CG33844', Aliases='Dmel_CG33844, CG33844, Dmel\\CG33844', description='CG33844 gene product from transcript CG33844-RA', other_designations='CG33844-PA|His2A:CG33844-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21503521, end_position_on_the_genomic_accession=21504013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37014, CurrentID=0, Status='live', Symbol='CG10933', Aliases='Dmel_ Dmel\\CG10933', description='CG10933 gene product from transcript CG10933-RA', other_designations='CG10933-PA|uncharacterized protein', map_location='54E2-54E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17723309, end_position_on_the_genomic_accession=17725056, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37015, CurrentID=0, Status='live', Symbol='Snx16', Aliases='Dmel_CG6410, CG6410, DmSNX16, Dmel\\CG6410', description='Sorting nexin 16', other_designations='CG6410-PA|Snx16-PA|sorting nexin 16', map_location='54E2-54E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17725030, end_position_on_the_genomic_accession=17726847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37016, CurrentID=0, Status='live', Symbol='CG4975', Aliases='Dmel_ Dmel\\CG4975', description='CG4975 gene product from transcript CG4975-RE', other_designations='CG4975-PC|CG4975-PE', map_location='54E2-54E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17726914, end_position_on_the_genomic_accession=17732987, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3772569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772569, CurrentID=0, Status='live', Symbol='CG32269', Aliases='Dmel_ CG11530, CG33161, Dmel\\CG32269, SP124', description='CG32269 gene product from transcript CG32269-RA', other_designations='CG32269-PA|CG32269-PB|uncharacterized protein', map_location='63B12-63B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3206949, end_position_on_the_genomic_accession=3208985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37018, CurrentID=0, Status='live', Symbol='CG34195', Aliases='Dmel_ CG4989, Dmel\\CG34195', description='CG34195 gene product from transcript CG34195-RA', other_designations='CG34195-PA|uncharacterized protein', map_location='54E4-54E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17733238, end_position_on_the_genomic_accession=17734483, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37019, CurrentID=0, Status='live', Symbol='CG6406', Aliases='Dmel_ Dmel\\CG6406', description='CG6406 gene product from transcript CG6406-RA', other_designations='CG6406-PA|CG6406-PB', map_location='54E4-54E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17734419, end_position_on_the_genomic_accession=17737269, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37020, CurrentID=0, Status='live', Symbol='CG6401', Aliases='Dmel_ Dm*AAF57802/5-442, Dmel\\CG6401', description='CG6401 gene product from transcript CG6401-RB', other_designations='CG6401-PA|CG6401-PB', map_location='54E4-54E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17737622, end_position_on_the_genomic_accession=17739685, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37021, CurrentID=0, Status='live', Symbol='CG4996', Aliases='Dmel_ Dmel\\CG4996', description='CG4996 gene product from transcript CG4996-RA', other_designations='CG4996-PA', map_location='54E5-54E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17739901, end_position_on_the_genomic_accession=17743969, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  3772574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772574, CurrentID=0, Status='live', Symbol='Fuca', Aliases='Dmel_CG6128, CG33047, CG6128, DmFuca, Dmel\\CG6128_Dm, Fuca', description='alpha-L-fucosidase', other_designations='CG6128-PB|CG6128-PC|Fuca-PB|Fuca-PC', map_location='68C12-68C12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11518798, end_position_on_the_genomic_accession=11522713, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3772575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772575, CurrentID=0, Status='live', Symbol='His2B:CG33884', Aliases='Dmel_CG33884, CG33884, Dmel\\CG33884', description='CG33884 gene product from transcript CG33884-RA', other_designations='CG33884-PA|His2B:CG33884-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21497745, end_position_on_the_genomic_accession=21498339, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772576, CurrentID=0, Status='live', Symbol='Hsp22', Aliases='Dmel_CG4460, 22, CG32041, CG4460, DmHsp22, Dmel\\CG4460, EP(3)3247, HSP22, anon-WO0118547.50, anon-WO0140519.138, hsp22, small hsp locus 67B', description='Heat shock protein 22', other_designations='CG4460-PA|CG4460-PB|Hsp22-PA|Hsp22-PB|heat shock protein 22|heat shock protein hsp22|heat-shock protein 22|protein 22', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9372931, end_position_on_the_genomic_accession=9374964, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37025, CurrentID=0, Status='live', Symbol='CG5002', Aliases='Dmel_ Dmel\\CG5002', description='CG5002 gene product from transcript CG5002-RA', other_designations='CG5002-PA', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17750293, end_position_on_the_genomic_accession=17753214, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37026, CurrentID=0, Status='live', Symbol='Sardh', Aliases='Dmel_CG6385, CG6385, Dmel\\CG6385, sardh', description='Sarcosine dehydrogenase', other_designations='CG6385-PA|Sardh-PA|sarcosine dehydrogenase', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17753555, end_position_on_the_genomic_accession=17756422, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37027, CurrentID=0, Status='live', Symbol='HLH54F', Aliases='Dmel_CG5005, CG5005, Dmel\\CG5005, HLH102C-Dm, bHLH54F, bHLHa9', description='CG5005 gene product from transcript CG5005-RA', other_designations='CG5005-PA|HLH54F-PA', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17757405, end_position_on_the_genomic_accession=17760517, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37028, CurrentID=0, Status='live', Symbol='CG5009', Aliases='Dmel_ BcDNA:GH07485, Dmel\\CG5009, GH07485', description='CG5009 gene product from transcript CG5009-RA', other_designations='CG5009-PA', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17760940, end_position_on_the_genomic_accession=17764541, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37029, CurrentID=0, Status='live', Symbol='OstDelta', Aliases='Dmel_CG6370, CG6370, Dmel\\CG6370, OST', description='Oligosaccharide transferase Delta subunit', other_designations='CG6370-PA|CG6370-PB|OstDelta-PA|OstDelta-PB|oligosaccharide transferase delta subunit', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17764574, end_position_on_the_genomic_accession=17767070, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37030, CurrentID=0, Status='live', Symbol='CG14488', Aliases='Dmel_ Dmel\\CG14488', description='CG14488 gene product from transcript CG14488-RA', other_designations='CG14488-PA|uncharacterized protein', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17767165, end_position_on_the_genomic_accession=17767786, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772583, CurrentID=0, Status='live', Symbol='gdl', Aliases='Dmel_CG33756, CG33088, CG33756, CG7268, Dmel\\CG33756', description='gonadal', other_designations='CG33756-PA|CG33756-PB|CG33756-PC|gdl-PA|gdl-PB|gdl-PC', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15509445, end_position_on_the_genomic_accession=15511047, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37032, CurrentID=0, Status='live', Symbol='CG6362', Aliases='Dmel_ Dmel\\CG6362, EG:52C10.6', description='CG6362 gene product from transcript CG6362-RA', other_designations='CG6362-PA', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17770405, end_position_on_the_genomic_accession=17772552, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37033, CurrentID=0, Status='live', Symbol='fab1', Aliases='Dmel_CG6355, CG6355, Dmel\\CG6355, EG:52C10.5, Fab1, PIKfyve, PtdIns3P, fab-1, l(2)SJB', description='CG6355 gene product from transcript CG6355-RA', other_designations='CG6355-PA|CG6355-PB|PtdIns(3)P 5-kinase|fab1-PA|fab1-PB', map_location='54E9-54E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17773964, end_position_on_the_genomic_accession=17781930, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37034, CurrentID=0, Status='live', Symbol='aft', Aliases='Dmel_CG5032, CG5032, Dmel\\CG5032, EG:52C10.7', description='adrift', other_designations='CG5032-PA|aft-PA', map_location='54E9-54E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17786135, end_position_on_the_genomic_accession=17788556, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37035, CurrentID=0, Status='live', Symbol='Ubc10', Aliases='Dmel_CG5788, CG5788, Dmel\\CG5788, EG:52C10.4, UbcD10, ubc, ubcD10', description='Ubiquitin conjugating enzyme 10', other_designations='CG5788-PA|Ubc10-PA|ubiquitin conjugating enzyme 10', map_location='54E9-54E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17788412, end_position_on_the_genomic_accession=17789410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37036, CurrentID=0, Status='live', Symbol='CG5033', Aliases='Dmel_ Dmel\\CG5033, EG:52C10.1, dWDR36', description='CG5033 gene product from transcript CG5033-RA', other_designations='CG5033-PA', map_location='54E9-54E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17789679, end_position_on_the_genomic_accession=17792448, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37037, CurrentID=0, Status='live', Symbol='Dhit', Aliases='Dmel_CG5036, CG5036, Dmel\\CG5036, EG:52C10.2, dhit', description='Double hit', other_designations='CG5036-PB|CG5036-PC|Dhit-PB|Dhit-PC|double hit', map_location='54E9-54E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17792870, end_position_on_the_genomic_accession=17800677, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37038, CurrentID=0, Status='live', Symbol='grh', Aliases='Dmel_CG42311, BcDNA:LD38807, CG30111, CG42311, CG5058, DREB, Dmel\\CG42311, Dmel_CG30111, Dmel_CG5058, EG:191D12.1, ELF1, Elf-1, Elf-1/NTF-1, Elf1, GRH, Gh, Grh, NTF, NTF-1, NTF-1/Elf-1, NTF1, Ntf, Ntf1, elf1, elf1(CAG)[[7]], hsk, l(2)06850, l(2)IM45, l(2)s2140, ntf-1', description='grainy head', other_designations='CG42311-PH|CG42311-PI|CG42311-PJ|CG42311-PK|CG42311-PL|CG42311-PN|CG42311-PO|CG42311-PP|decapentaplegic repression element binding protein|element I binding activity|grainyhead|grh-PH|grh-PI|grh-PJ|grh-PK|grh-PL|grh-PN|grh-PO|grh-PP|head skeleton|neurogenic element binding transcription factor|neuronal transcription factor 1', map_location='2-86 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17801132, end_position_on_the_genomic_accession=17842820, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  37039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37039, CurrentID=0, Status='live', Symbol='olf186-M', Aliases='Dmel_CG14489, CG14489, Dmel\\CG14489', description='CG14489 gene product from transcript CG14489-RA', other_designations='CG14489-PA|olf186-M-PA', map_location='54F1-54F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17846699, end_position_on_the_genomic_accession=17848444, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37040, CurrentID=0, Status='live', Symbol='olf186-F', Aliases='Dmel_CG11430, BcDNA:LD03241, CG11430, CRACM1, Dmel\\CG11430, Olf186-F, Orai, SPH206, dOrai, dOrai1, dorai, l(2)k11505', description='CG11430 gene product from transcript CG11430-RB', other_designations='CG11430-PA|CG11430-PB|CG11430-PE|CG11430-PF|CG11430-PG|CRAC modulator 1|lethal (2) k11505|olf186-F-PA|olf186-F-PB|olf186-F-PE|olf186-F-PF|olf186-F-PG|orai', map_location='54F1-54F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17843442, end_position_on_the_genomic_accession=17860582, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37041, CurrentID=0, Status='live', Symbol='CG14490', Aliases='Dmel_ Dmel\\CG14490', description='CG14490 gene product from transcript CG14490-RA', other_designations='CG14490-PA|uncharacterized protein', map_location='54F3-54F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17860760, end_position_on_the_genomic_accession=17861591, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37042, CurrentID=0, Status='live', Symbol='thr', Aliases='Dmel_CG5785, CG5785, Dmel\\CG5785, THR, Thr, ana, anch, frag1', description='three rows', other_designations='CG5785-PB|CG5785-PC|anarchist|fragmented gonads 1|separase|thr-PB|thr-PC|three row', map_location='2-86 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17861273, end_position_on_the_genomic_accession=17865963, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37043, CurrentID=0, Status='live', Symbol='Mapmodulin', Aliases='Dmel_CG5784, CG5784, CT18148, CT42180, Dmel\\CG5784, anon-EST:fe3A2', description='CG5784 gene product from transcript CG5784-RB', other_designations='CG5784-PA|CG5784-PB|CG5784-PC|CG5784-PD|Mapmodulin-PA|Mapmodulin-PB|Mapmodulin-PC|Mapmodulin-PD|anon-fast-evolving-3A2|mapmodulin', map_location='54F4-54F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17866178, end_position_on_the_genomic_accession=17872015, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37044, CurrentID=0, Status='live', Symbol='PPO1', Aliases='Dmel_CG42639, A1, A[[1]], Bc, CG42639, CG5779, DmePPOA1, Dmel\\CG42639, Dmel_CG5779, Dox-A1, DoxA1, Mox, PO, PO A1, PO54, PPO, Phox, ProPO, ProPO A1, ProPO54, ProPo, Propo, bc, i12, pro-A, pro-PO, pro-POA[[1]], proPO, proPO A[[1]], proPO-A1, proPO54', description='Prophenoloxidase 1', other_designations='CG42639-PA|PPO1-PA|black cell|black cells|diphenol oxidase A1 subunit|monophenol oxidase|monophenoloxidase|phenoloxidase|pro-phenol oxidase A1|prophenol oxidase A1|prophenoloxidase|prophenoloxidase 1|prophenoloxidase A1|prophenoloxidase A[[1]]', map_location='2-80.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17887220, end_position_on_the_genomic_accession=17889973, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  5740574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740574, CurrentID=0, Status='live', Symbol='CG34221', Aliases='Dmel_ Dmel\\CG34221', description='CG34221 gene product from transcript CG34221-RC', other_designations='CG34221-PB|CG34221-PC|uncharacterized protein', map_location='46F5-46F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10197136, end_position_on_the_genomic_accession=10199711, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37046, CurrentID=0, Status='live', Symbol='CG14491', Aliases='Dmel_ Dmel\\CG14491', description='CG14491 gene product from transcript CG14491-RA', other_designations='CG14491-PA|uncharacterized protein', map_location='55A1-55A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17895078, end_position_on_the_genomic_accession=17895578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37047, CurrentID=0, Status='live', Symbol='Elk', Aliases='Dmel_CG5076, CG5076, Dmel\\CG5076, dmELK, elk', description='Eag-like K[+] channel', other_designations='CG5076-PA|CG5076-PB|Elk-PA|Elk-PB|eag-like K+ channel|eag-like K[+] channel', map_location='54F4-55A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17873939, end_position_on_the_genomic_accession=17926695, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  3772600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772600, CurrentID=0, Status='live', Symbol='CG33725', Aliases='Dmel_ Dmel\\CG33725', description='CG33725 gene product from transcript CG33725-RC', other_designations='CG33725-PA|CG33725-PC|uncharacterized protein', map_location='69C2-69C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12482609, end_position_on_the_genomic_accession=12486446, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772602, CurrentID=0, Status='live', Symbol='CG33757', Aliases='Dmel_ Dmel\\CG33757', description='CG33757 gene product from transcript CG33757-RA', other_designations='CG33757-PA|uncharacterized protein', map_location='45D8-45D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9454297, end_position_on_the_genomic_accession=9454948, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772603, CurrentID=0, Status='live', Symbol='CG15892', Aliases='Dmel_ CG32918, Dmel\\CG15892', description='CG15892 gene product from transcript CG15892-RA', other_designations='CG15892-PA|uncharacterized protein', map_location='5E6-5E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6293106, end_position_on_the_genomic_accession=6294343, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37052, CurrentID=0, Status='live', Symbol='CG10910', Aliases='Dmel_ Dmel\\CG10910', description='CG10910 gene product from transcript CG10910-RB', other_designations='CG10910-PB|uncharacterized protein', map_location='55B1-55B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18020268, end_position_on_the_genomic_accession=18022406, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37053, CurrentID=0, Status='live', Symbol='CG5773', Aliases='Dmel_ BcDNA:RH46075, Dmel\\CG5773', description='CG5773 gene product from transcript CG5773-RC', other_designations='CG5773-PA|CG5773-PB|CG5773-PC', map_location='55B1-55B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18042296, end_position_on_the_genomic_accession=18043744, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37054, CurrentID=0, Status='live', Symbol='CG5770', Aliases='Dmel_ BcDNA:LP06072, Dmel\\CG5770', description='CG5770 gene product from transcript CG5770-RA', other_designations='CG5770-PA', map_location='55B1-55B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18044046, end_position_on_the_genomic_accession=18045073, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772607, CurrentID=0, Status='live', Symbol='His3:CG33827', Aliases='Dmel_CG33827, CG33827, Dmel\\CG33827', description='CG33827 gene product from transcript CG33827-RA', other_designations='CG33827-PA|His3:CG33827-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21474593, end_position_on_the_genomic_accession=21475121, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37056, CurrentID=0, Status='live', Symbol='CG14495', Aliases='Dmel_ Dmel\\CG14495', description='CG14495 gene product from transcript CG14495-RA', other_designations='CG14495-PA|CG14495-PB|uncharacterized protein', map_location='55B2-55B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18048804, end_position_on_the_genomic_accession=18049922, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37057, CurrentID=0, Status='live', Symbol='Muc55B', Aliases='Dmel_CG5765, CG5765, Dmel\\CG5765', description='Mucin 55B', other_designations='CG5765-PA|CG5765-PB|Muc55B-PA|Muc55B-PB|mucin 55B', map_location='55B2-55B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18050019, end_position_on_the_genomic_accession=18052107, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37058, CurrentID=0, Status='live', Symbol='CG10911', Aliases='Dmel_ Dmel\\CG10911, anon-WO0140519.211', description='CG10911 gene product from transcript CG10911-RA', other_designations='CG10911-PA|uncharacterized protein', map_location='55B2-55B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18052550, end_position_on_the_genomic_accession=18054066, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37059, CurrentID=0, Status='live', Symbol='CG10912', Aliases='Dmel_ Dmel\\CG10912', description='CG10912 gene product from transcript CG10912-RA', other_designations='CG10912-PA|uncharacterized protein', map_location='55B2-55B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18056064, end_position_on_the_genomic_accession=18057065, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37062, CurrentID=0, Status='live', Symbol='CG5757', Aliases='Dmel_ Dmel\\CG5757', description='CG5757 gene product from transcript CG5757-RB', other_designations='CG5757-PA|CG5757-PB', map_location='55B4-55B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18093723, end_position_on_the_genomic_accession=18094900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37063, CurrentID=0, Status='live', Symbol='CG5098', Aliases='Dmel_ Dmel\\CG5098', description='CG5098 gene product from transcript CG5098-RA', other_designations='CG5098-PA|CG5098-PB|CG5098-PC|CG5098-PD', map_location='55B4-55B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18094940, end_position_on_the_genomic_accession=18101031, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37064, CurrentID=0, Status='live', Symbol='CG5756', Aliases='Dmel_ Dmel\\CG5756', description='CG5756 gene product from transcript CG5756-RB', other_designations='CG5756-PB|CG5756-PC', map_location='55B5-55B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18101061, end_position_on_the_genomic_accession=18108389, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37065, CurrentID=0, Status='live', Symbol='stau', Aliases='Dmel_CG5753, CG5753, Dmel\\CG5753, Dmstau, STAU, Stau, Stauf, dStau', description='staufen', other_designations='CG5753-PA|CG5753-PB|CG5753-PC|stau-PA|stau-PB|stau-PC|stauffen', map_location='2-83.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18117896, end_position_on_the_genomic_accession=18125779, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37066, CurrentID=0, Status='live', Symbol='Sema-5c', Aliases='Dmel_CG5661, BEST:CK02567, BG01245, CG5661, CK02567, CT17880, Dmel\\CG5661, Sema 5c, Sema-5C, sema 5c, sema-5c', description='Semaphorin-5c', other_designations='CG5661-PA|CG5661-PB|Sema-5c-PA|Sema-5c-PB|semaphorin|semaphorin 5C|semiphorin', map_location='68F2-68F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12067511, end_position_on_the_genomic_accession=12081830, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  3772619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772619, CurrentID=0, Status='live', Symbol='His3:CG33848', Aliases='Dmel_CG33848, CG33848, Dmel\\CG33848', description='CG33848 gene product from transcript CG33848-RA', other_designations='CG33848-PA|His3:CG33848-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21510018, end_position_on_the_genomic_accession=21510546, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37068, CurrentID=0, Status='live', Symbol='Hsf', Aliases='Dmel_CG5748, CG5748, D-HSF, Dm-Hsf, DmHSF, DmHSFa, Dmel\\CG5748, HSF, HSF1, HSFa, dHSF, dHSFa, hsf, l(2)03091', description='Heat shock factor', other_designations='CG5748-PA|CG5748-PB|CG5748-PC|CG5748-PD|Heat-Shock factor|Hsf-PA|Hsf-PB|Hsf-PC|Hsf-PD|heat shock factor|heat shock transcription factor|heat shock transcription factor 1|heat-shock transcription factor', map_location='55B7-55B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18130303, end_position_on_the_genomic_accession=18134634, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37069, CurrentID=0, Status='live', Symbol='Pcl', Aliases='Dmel_CG5109, CG5109, Dmel\\CG5109, PCL, l(2)s1859, pcl', description='Polycomblike', other_designations='CG5109-PA|CG5109-PB|Pcl-PA|Pcl-PB|Polycomb-like|polycomblike', map_location='2-84 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18135426, end_position_on_the_genomic_accession=18139990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37070, CurrentID=0, Status='live', Symbol='pAbp', Aliases='Dmel_CG5119, BcDNA:LD24412, CG16803, CG5119, Dmel\\CG5119, PABP, PABP55B, PABPC1, PAbp, Pabp, anon-EST:Posey247, anon-EST:Posey249, anon-EST:fe2E5, dPABP, duo, l(2)SH1908, l(2)SH2 1908, l(2)k10109, pAB, pABP, pabp', description='CG5119 gene product from transcript CG5119-RC', other_designations='CG5119-PA|CG5119-PB|CG5119-PC|Poly(A)-binding protein|Poly(A)binding protein|PolyA-binding protein|anon-fast-evolving-2E5|doppio fuso|lethal (2) SH1908|pAbp-PA|pAbp-PB|pAbp-PC|poly (A)-binding protein|poly(A) binding protein|poly-A binding protein|polyadenylate-binding protein', map_location='55B8-55B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18140078, end_position_on_the_genomic_accession=18147191, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37071, CurrentID=0, Status='live', Symbol='CG17680', Aliases='Dmel_ BcDNA:RE55001, Dmel\\CG17680', description='CG17680 gene product from transcript CG17680-RA', other_designations='CG17680-PA|uncharacterized protein', map_location='55B9-55B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18147297, end_position_on_the_genomic_accession=18147997, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37072, CurrentID=0, Status='live', Symbol='CG5742', Aliases='Dmel_ Dmel\\CG5742', description='CG5742 gene product from transcript CG5742-RB', other_designations='CG5742-PA|CG5742-PB|CG5742-PC', map_location='55B9-55B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18148096, end_position_on_the_genomic_accession=18151410, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37073, CurrentID=0, Status='live', Symbol='adp', Aliases='Dmel_CG5124, CG5124, Dmel\\CG5124, anon-WO0196371.1, anon-WO0196371.27, anon-WO0196371.3', description='adipose', other_designations='CG5124-PA|CG5124-PB|adp-PA|adp-PB', map_location='2-83.4 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18151704, end_position_on_the_genomic_accession=18154060, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37075, CurrentID=0, Status='live', Symbol='CG10914', Aliases='Dmel_ Dmel\\CG10914', description='CG10914 gene product from transcript CG10914-RA', other_designations='CG10914-PA|uncharacterized protein', map_location='55B9-55B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18158257, end_position_on_the_genomic_accession=18160310, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37076, CurrentID=0, Status='live', Symbol='CG10915', Aliases='Dmel_ Dmel\\CG10915', description='CG10915 gene product from transcript CG10915-RB', other_designations='CG10915-PA|CG10915-PB|uncharacterized protein', map_location='55B9-55B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18160720, end_position_on_the_genomic_accession=18164746, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37078, CurrentID=0, Status='live', Symbol='Nup75', Aliases='Dmel_CG5733, CG5733, Dmel\\CG5733, Nup 75, nup75', description='Nucleoporin 75kD', other_designations='CG5733-PA|Nup75-PA|nucleoporin 75|nucleoporin 75kD', map_location='55B10-55B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18164694, end_position_on_the_genomic_accession=18167140, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37079, CurrentID=0, Status='live', Symbol='MED9', Aliases='Dmel_CG42517, BcDNA:LD07740, BcDNA:RH42446, CG30112, CG30113, CG42517, CG5134, Dmel\\CG42517, Dmel_CG5134, med9', description='Mediator complex subunit 9', other_designations='CG42517-PC|MED9-PC|mediator complex subunit 9', map_location='55B11-55B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18167255, end_position_on_the_genomic_accession=18168380, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37080, CurrentID=0, Status='live', Symbol='Dgp-1', Aliases='Dmel_CG5729, CG5729, Dmel\\CG5729', description='CG5729 gene product from transcript CG5729-RB', other_designations='CG5729-PA|CG5729-PB|Dgp-1-PA|Dgp-1-PB', map_location='55B11-55B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18168413, end_position_on_the_genomic_accession=18172607, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37081, CurrentID=0, Status='live', Symbol='CG10916', Aliases='Dmel_ Dmel\\CG10916', description='CG10916 gene product from transcript CG10916-RB', other_designations='CG10916-PA|CG10916-PB|uncharacterized protein', map_location='55B11-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18173669, end_position_on_the_genomic_accession=18175725, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37082, CurrentID=0, Status='live', Symbol='CG5726', Aliases='Dmel_ Dmel\\CG5726', description='CG5726 gene product from transcript CG5726-RA', other_designations='CG5726-PA', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18175255, end_position_on_the_genomic_accession=18180197, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37083, CurrentID=0, Status='live', Symbol='nopo', Aliases='Dmel_CG5140, CG5140, Dmel\\CG5140, NOPO, Nopo, TRIP', description='no poles', other_designations='CG5140-PA|CG5140-PB|nopo-PA|nopo-PB', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18180496, end_position_on_the_genomic_accession=18182519, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37084, CurrentID=0, Status='live', Symbol='CG5721', Aliases='Dmel_ Dmel\\CG5721', description='CG5721 gene product from transcript CG5721-RA', other_designations='CG5721-PA', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18182435, end_position_on_the_genomic_accession=18184090, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772638, CurrentID=0, Status='live', Symbol='Gr64d', Aliases='Dmel_CG14987, CG14987, CG33157, Dmel64d, Dmel\\CG14987, GR64d, Gr64', description='Gustatory receptor 64d', other_designations='CG14987-PA|CG14987-PC|Gr64d-PA|Gr64d-PC|gustatory receptor 64d', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4032062, end_position_on_the_genomic_accession=4035703, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  37087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37087, CurrentID=0, Status='live', Symbol='CG14500', Aliases='Dmel_ Dmel\\CG14500', description='CG14500 gene product from transcript CG14500-RA', other_designations='CG14500-PA|uncharacterized protein', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18188978, end_position_on_the_genomic_accession=18189654, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37088, CurrentID=0, Status='live', Symbol='CG33958', Aliases='Dmel_ CG14501, CG5719, Dmel\\CG33958, GYC', description='CG33958 gene product from transcript CG33958-RA', other_designations='CG33958-PA|uncharacterized protein', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18184382, end_position_on_the_genomic_accession=18192851, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  37089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37089, CurrentID=0, Status='live', Symbol='fj', Aliases='Dmel_CG10917, CG10917, Dmel\\CG10917, Fj', description='four-jointed', other_designations='CG10917-PA|fj-PA|four jointed', map_location='2-81.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18232761, end_position_on_the_genomic_accession=18236311, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37090, CurrentID=0, Status='live', Symbol='Ote', Aliases='Dmel_CG5581, CG5581, Dmel\\CG5581, hal, ote, otefin', description='Otefin', other_designations='CG5581-PA|Ote-PA|halted|haulted|otefin', map_location='55A1-55A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18275607, end_position_on_the_genomic_accession=18277177, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37092, CurrentID=0, Status='live', Symbol='CG14502', Aliases='Dmel_ Dmel\\CG14502', description='CG14502 gene product from transcript CG14502-RC', other_designations='CG14502-PA|CG14502-PB|CG14502-PC|CG14502-PD|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18288696, end_position_on_the_genomic_accession=18294712, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37093, CurrentID=0, Status='live', Symbol='CG18536', Aliases='Dmel_ Dmel\\CG18536', description='CG18536 gene product from transcript CG18536-RA', other_designations='CG18536-PA|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18294759, end_position_on_the_genomic_accession=18295658, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772646, CurrentID=0, Status='live', Symbol='mei-218', Aliases='Dmel_CG8923, CG33067, CG8923, Dmel\\CG8923', description='meiotic 218', other_designations='CG8923-PB|CG8923-PC|CG8923-PD|CG8923-PE|mei-218-PB|mei-218-PC|mei-218-PD|mei-218-PE|transcript c', map_location='15E5-15E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17135985, end_position_on_the_genomic_accession=17142990, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37095, CurrentID=0, Status='live', Symbol='CG18538', Aliases='Dmel_ Dmel\\CG18538', description='CG18538 gene product from transcript CG18538-RA', other_designations='CG18538-PA|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18297070, end_position_on_the_genomic_accession=18297751, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37096, CurrentID=0, Status='live', Symbol='CG18539', Aliases='Dmel_ Dmel\\CG18539', description='CG18539 gene product from transcript CG18539-RC', other_designations='CG18539-PB|CG18539-PC|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18297795, end_position_on_the_genomic_accession=18298903, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37097, CurrentID=0, Status='live', Symbol='CG18540', Aliases='Dmel_ Dmel\\CG18540', description='CG18540 gene product from transcript CG18540-RB', other_designations='CG18540-PB|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18299317, end_position_on_the_genomic_accession=18300072, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37098, CurrentID=0, Status='live', Symbol='CG14505', Aliases='Dmel_ Dmel\\CG14505', description='CG14505 gene product from transcript CG14505-RB', other_designations='CG14505-PB|CG14505-PC|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18302945, end_position_on_the_genomic_accession=18303977, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37099, CurrentID=0, Status='live', Symbol='IM23', Aliases='Dmel_CG15066, BcDNA:RH67033, CG15066, DIM 23, DIM-23, Dmel\\CG15066', description='Immune induced molecule 23', other_designations='CG15066-PA|IM23-PA|Immune-induced peptide 23 precursor|immune induced molecule 23', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18382704, end_position_on_the_genomic_accession=18383235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37100, CurrentID=0, Status='live', Symbol='IM1', Aliases='Dmel_CG18108, BcDNA:RH08745, CG18108, DIM 1, DIM-1, Dmel\\CG18108, IM2-like', description='Immune induced molecule 1', other_designations='CG18108-PA|IM1-PA|Immune-induced peptide 1 precursor|immune induced molecule 1|immune-induced molecule 1', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18383949, end_position_on_the_genomic_accession=18384378, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37101, CurrentID=0, Status='live', Symbol='CG18107', Aliases='Dmel_ BcDNA:RH09147, Dmel\\CG18107', description='CG18107 gene product from transcript CG18107-RA', other_designations='CG18107-PA|uncharacterized protein', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18384567, end_position_on_the_genomic_accession=18384911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37102, CurrentID=0, Status='live', Symbol='CG15067', Aliases='Dmel_ Dmel\\CG15067, anon-EST:Posey85', description='CG15067 gene product from transcript CG15067-RA', other_designations='CG15067-PA|CG15067-PB|uncharacterized protein', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18384824, end_position_on_the_genomic_accession=18385589, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772655, CurrentID=0, Status='live', Symbol='CG33155', Aliases='Dmel_ BcDNA:GM05817, CG33184, CR33155, Dmel\\CG33155', description='CG33155 gene product from transcript CG33155-RC', other_designations='CG33155-PA|CG33155-PC|CG33155-PD|uncharacterized protein', map_location='50C16-50C16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13954195, end_position_on_the_genomic_accession=13955664, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37104, CurrentID=0, Status='live', Symbol='Idgf5', Aliases='Dmel_CG5154, CG5154, Cht14, DmIDGF5, Dmel\\CG5154, IDGF, IDGF-5, IDGF5', description='Imaginal disc growth factor 5', other_designations='CG5154-PA|Idgf5-PA|imaginal disc growth factor 5', map_location='55C4-55C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18393666, end_position_on_the_genomic_accession=18395282, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37105, CurrentID=0, Status='live', Symbol='GstE10', Aliases='Dmel_CG17522, CG17522, DmGSTE10, Dmel\\CG17522, GSTE10, gste10', description='Glutathione S transferase E10', other_designations='CG17522-PA|CG17522-PB|GstE10-PA|GstE10-PB|glutathione S transferase E10', map_location='55C6-55C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18396638, end_position_on_the_genomic_accession=18398519, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37106, CurrentID=0, Status='live', Symbol='GstE1', Aliases='Dmel_CG5164, CG5164, CT16545, DmGST-3, DmGSTE1, DmGSTE1-1, Dmel\\CG5164, GST-3, GST-E1, GST3, GSTE1, Gst-3, Gst-theta 55F, Gst3, Gst55F, gst-3, gstE1, gste1', description='Glutathione S transferase E1', other_designations='CG5164-PA|GstE1-PA|glutathione S transferase E1|glutathione S-transferase|glutathione-S-transferase|glutathione-S-transferase E1', map_location='55C6-55C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18398393, end_position_on_the_genomic_accession=18399223, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37107, CurrentID=0, Status='live', Symbol='GstE2', Aliases='Dmel_CG17523, CG17523, DmGSTE2, Dmel\\CG17523, GSTE2, gste2', description='Glutathione S transferase E2', other_designations='CG17523-PA|GstE2-PA|glutathione S transferase E2', map_location='55C6-55C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18399486, end_position_on_the_genomic_accession=18400375, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37108, CurrentID=0, Status='live', Symbol='GstE3', Aliases='Dmel_CG17524, BEST:GH09055, CG17524, DmGSTE3, Dmel\\CG17524, E3, GST-E3, GSTE3, gste3', description='Glutathione S transferase E3', other_designations='CG17524-PA|GstE3-PA|glutathione S transferase E3', map_location='55C7-55C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18400713, end_position_on_the_genomic_accession=18401467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37109, CurrentID=0, Status='live', Symbol='GstE4', Aliases='Dmel_CG17525, CG17525, DmGSTE4, Dmel\\CG17525, GSTE4, gste4', description='Glutathione S transferase E4', other_designations='CG17525-PA|GstE4-PA|glutathione S transferase E4', map_location='55C7-55C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18403248, end_position_on_the_genomic_accession=18404086, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37110, CurrentID=0, Status='live', Symbol='GstE5', Aliases='Dmel_CG17527, CG17527, DmGSTE5, Dmel\\CG17527, GSTE5, gstE5, gste5', description='Glutathione S transferase E5', other_designations='CG17527-PA|GstE5-PA|glutathione S transferase E5', map_location='55C7-55C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18404605, end_position_on_the_genomic_accession=18405338, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37111, CurrentID=0, Status='live', Symbol='GstE6', Aliases='Dmel_CG17530, CG17530, CT38747, DmGSTE6, Dmel\\CG17530, E6, GSTE6, gstE6, gste6', description='Glutathione S transferase E6', other_designations='CG17530-PA|GstE6-PA|glutathione S transferase E6|glutathione transferase E6', map_location='55C7-55C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18406011, end_position_on_the_genomic_accession=18406761, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37112, CurrentID=0, Status='live', Symbol='GstE7', Aliases='Dmel_CG17531, CG17531, DmGSTE7, Dmel\\CG17531, E7, GST-E7, GSTE7, gstE7, gste7', description='Glutathione S transferase E7', other_designations='CG17531-PA|GstE7-PA|glutathione S transferase E7|glutathione S-transferase|glutathione transferase E7', map_location='55C7-55C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18406924, end_position_on_the_genomic_accession=18407688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37113, CurrentID=0, Status='live', Symbol='GstE8', Aliases='Dmel_CG17533, BcDNA:GM08326, CG17533, DmGSTE8, Dmel\\CG17533, GST3-2, GSTE8, Gst3-2, gste8', description='Glutathione S transferase E8', other_designations='CG17533-PA|CG17533-PB|GstE8-PA|GstE8-PB|glutathione S transferase E8', map_location='55C7-55C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18407834, end_position_on_the_genomic_accession=18408955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772666, CurrentID=0, Status='live', Symbol='His4:CG33907', Aliases='Dmel_CG33907, CG33907, Dmel\\CG33907', description='CG33907 gene product from transcript CG33907-RA', other_designations='CG33907-PA|His4:CG33907-PA|histone H4', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21429655, end_position_on_the_genomic_accession=21430066, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37116, CurrentID=0, Status='live', Symbol='Dp1', Aliases='Dmel_CG5170, BcDNA:LD21383, BcDNA:LD21677, CG5170, DDP1, Dmel\\CG5170, anon-WO0118547.516, anon-WO0172774.168, ddp1', description='Dodeca-satellite-binding protein 1', other_designations='CG5170-PA|CG5170-PB|CG5170-PD|CG5170-PE|CG5170-PF|CG5170-PG|CG5170-PH|Dp1-PA|Dp1-PB|Dp1-PD|Dp1-PE|Dp1-PF|Dp1-PG|Dp1-PH|dodeca satellite binding protein 1|dodeca-satellite protein 1|dodeca-satellite-binding protein 1|drosophila dodeca-satellite-binding protein 1|vigilin', map_location='55C8-55C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18412027, end_position_on_the_genomic_accession=18421124, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  37117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37117, CurrentID=0, Status='live', Symbol='CG5174', Aliases='Dmel_ BcDNA:RE18604, BcDNA:RH27395, Dmel\\CG5174', description='CG5174 gene product from transcript CG5174-RM', other_designations='CG5174-PA|CG5174-PB|CG5174-PG|CG5174-PH|CG5174-PJ|CG5174-PL|CG5174-PM|CG5174-PN|CG5174-PO|CG5174-PP', map_location='55C8-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18421413, end_position_on_the_genomic_accession=18426221, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3772670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772670, CurrentID=0, Status='live', Symbol='bbx', Aliases='Dmel_CG1414, CG1414, CG32527, Dmel\\CG1414, anon-19Fb, tu2', description='bobby sox', other_designations='CG1414-PB|CG1414-PC|CG1414-PD|CG1414-PE|bbx-PB|bbx-PC|bbx-PD|bbx-PE', map_location='19F5-19F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21363028, end_position_on_the_genomic_accession=21373145, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37119, CurrentID=0, Status='live', Symbol='CG30118', Aliases='Dmel_ CG16859, CG5544, Dmel\\CG30118', description='CG30118 gene product from transcript CG30118-RC', other_designations='CG30118-PA|CG30118-PB|CG30118-PC|uncharacterized protein', map_location='55C9-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18430315, end_position_on_the_genomic_accession=18440475, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  37121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37121, CurrentID=0, Status='live', Symbol='slim', Aliases='Dmel_CG5186, CG5186, Dmel\\CG5186, l(2)s1866', description='scruin like at the midline', other_designations='CG5186-PA|CG5186-PC|lethal (2) s1866|slim-PA|slim-PC', map_location='55C9-56F12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18440965, end_position_on_the_genomic_accession=18445526, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37122, CurrentID=0, Status='live', Symbol='CG5189', Aliases='Dmel_ Dmel\\CG5189, dp14, p14', description='CG5189 gene product from transcript CG5189-RB', other_designations='CG5189-PA|CG5189-PB', map_location='55C9-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18446548, end_position_on_the_genomic_accession=18449420, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37123, CurrentID=0, Status='live', Symbol='Prp19', Aliases='Dmel_CG5519, BcDNA:LD02793, CG5519, Dmel\\CG5519, EC2-11, GPB, Gbp, PRP19, l(2)07838, prp19', description='Pre-RNA processing factor 19', other_designations='CG5519-PA|GTP-binding regulatory protein beta chain|NM 079059.2|Prp19-PA', map_location='55D1-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18447361, end_position_on_the_genomic_accession=18451041, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772676, CurrentID=0, Status='live', Symbol='ste24c', Aliases='Dmel_CG9002, CG30462, CG33142, CG9002, Dmel\\CG9002, Dmel\\Ste24, Dmel\\Ste24c', description='ste24c prenyl protease type I', other_designations='CG9002-PA|CG9002-PC|CG9002-PD|prenyl protease type I|ste24c-PA|ste24c-PC|ste24c-PD', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16995991, end_position_on_the_genomic_accession=16998594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3772677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772677, CurrentID=0, Status='live', Symbol='Snapin', Aliases='Dmel_CG9958, CG32951, CG9958, Dmel\\CG9958, Dromel_CG9960_CG9958_FBtr007773_snapin_uORF, NP_722835, SNAPIN, snapin', description='Snapin ortholog (H. sapiens)', other_designations='CG9958-PA|Snapin-PA|snapin|snapin ortholog', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2758427, end_position_on_the_genomic_accession=2759677, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37126, CurrentID=0, Status='live', Symbol='CG30116', Aliases='Dmel_ BcDNA:GH04922, CG17671, CG5512, Dmel\\CG30116', description='CG30116 gene product from transcript CG30116-RC', other_designations='CG30116-PA|CG30116-PB|CG30116-PC|CG30116-PD|uncharacterized protein', map_location='55C10-55D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18456602, end_position_on_the_genomic_accession=18498026, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  37128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37128, CurrentID=0, Status='live', Symbol='GstE11', Aliases='Dmel_CG5224, CG5224, Dmel\\CG5224, GSTE11, cg5224, gste11', description='Glutathione S transferase E11', other_designations='CG5224-PA|CG5224-PB|GstE11-PA|GstE11-PB|glutathione S transferase E11', map_location='55D1-55D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18503608, end_position_on_the_genomic_accession=18505025, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37129, CurrentID=0, Status='live', Symbol='CG43066', Aliases='Dmel_ CG42510, CG5226, Dmel\\CG43066, Dmel_CG42510, Dmel_CG5226, anon-WO0149848.1', description='CG43066 gene product from transcript CG43066-RC', other_designations='CG43066-PA|CG43066-PC|uncharacterized protein', map_location='55D1-55D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18514574, end_position_on_the_genomic_accession=18526439, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37130, CurrentID=0, Status='live', Symbol='CG10924', Aliases='Dmel_ Dmel\\CG10924, PEPCK', description='CG10924 gene product from transcript CG10924-RA', other_designations='CG10924-PA|CG10924-PB|pepck isogene|predicted phosphoenolpyruvate carboxykinase|uncharacterized protein', map_location='55D1-55D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18527430, end_position_on_the_genomic_accession=18535749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37131, CurrentID=0, Status='live', Symbol='Pepck', Aliases='Dmel_CG17725, 143299_at, CG10924, CG17725, Dmel\\CG17725, Dromel_CG17725_FBtr0086701_pepck_mORF, PEPCK, dPEPCK, pepck', description='Phosphoenolpyruvate carboxykinase', other_designations='CG17725-PA|Pepck-PA|Phosphoenolpyruvate-carboxykinase|phosphoenol pyruvate carboxykinase|phosphoenolpyruvate carboxy kinase|phosphoenolpyruvate carboxykinase', map_location='55D3-55D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18536767, end_position_on_the_genomic_accession=18539416, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37132, CurrentID=0, Status='live', Symbol='Rgk2', Aliases='Dmel_CG34390, CG15069, CG30119, CG34390, Dmel\\CG34390, Dmel_CG15069, Dmel_CG30119, RGK2', description='Rad, Gem/Kir family member 2', other_designations='CG34390-PA|CG34390-PC|Rgk2-PA|Rgk2-PC', map_location='55D3-55D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18539585, end_position_on_the_genomic_accession=18548969, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37134, CurrentID=0, Status='live', Symbol='GEFmeso', Aliases='Dmel_CG30115, CG18430, CG30115, CG5503, Dmel\\CG30115', description='Guanine nucleotide exchange factor in mesoderm', other_designations='CG30115-PE|CG30115-PF|CG30115-PG|GEFmeso-PE|GEFmeso-PF|GEFmeso-PG|guanine nucleotide exchange factor in mesoderm', map_location='55D3-55E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18554119, end_position_on_the_genomic_accession=18611777, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  37135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37135, CurrentID=0, Status='live', Symbol='CG10927', Aliases='Dmel_ Dmel\\CG10927', description='CG10927 gene product from transcript CG10927-RA', other_designations='CG10927-PA|uncharacterized protein', map_location='55E1-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18613090, end_position_on_the_genomic_accession=18614255, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37136, CurrentID=0, Status='live', Symbol='mRpS28', Aliases='Dmel_CG5497, BcDNA:RH13070, CG5497, Dmel\\CG5497, MRP-S35, mrpS28', description='mitochondrial ribosomal protein S28', other_designations='CG5497-PA|mRpS28-PA', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18614210, end_position_on_the_genomic_accession=18615018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37137, CurrentID=0, Status='live', Symbol='CG5323', Aliases='Dmel_ Dmel\\CG5323', description='CG5323 gene product from transcript CG5323-RA', other_designations='CG5323-PA', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18615060, end_position_on_the_genomic_accession=18615999, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37138, CurrentID=0, Status='live', Symbol='CG5327', Aliases='Dmel_ Dmel\\CG5327, NEST:bs33e07', description='CG5327 gene product from transcript CG5327-RA', other_designations='CG5327-PA', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18615988, end_position_on_the_genomic_accession=18618317, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37139, CurrentID=0, Status='live', Symbol='CG5493', Aliases='Dmel_ Dmel\\CG5493', description='CG5493 gene product from transcript CG5493-RB', other_designations='CG5493-PA|CG5493-PB|CG5493-PC|CG5493-PD', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18616692, end_position_on_the_genomic_accession=18619328, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37140, CurrentID=0, Status='live', Symbol='CG5335', Aliases='Dmel_ Dmel\\CG5335', description='CG5335 gene product from transcript CG5335-RA', other_designations='CG5335-PA', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18621541, end_position_on_the_genomic_accession=18622986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37141, CurrentID=0, Status='live', Symbol='Atg7', Aliases='Dmel_CG5489, ATG7, CG5489, Dmel\\CG5489, atg-7, atg7', description='Autophagy-related 7', other_designations='Atg7-PA|Atg7-PB|Autophagy-specific gene 7|CG5489-PA|CG5489-PB|autophagy-related gene 7|autophagy-related protein 7', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18620466, end_position_on_the_genomic_accession=18625044, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37142, CurrentID=0, Status='live', Symbol='Sec6', Aliases='Dmel_CG5341, CG5341, Dmel\\CG5341, Dsec6p, dsec6, sec 6, sec6, Sec6', description='Sec6 ortholog (S. cerevisiae)', other_designations='CG5341-PA|Sec6 ortholog|Sec6-PA|sec6', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18625249, end_position_on_the_genomic_accession=18627718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37143, CurrentID=0, Status='live', Symbol='Eip55E', Aliases='Dmel_CG5345, CG5345, Dmel\\CG5345, EIP40, Eip40, Eip55BD, LD22255', description='CG5345 gene product from transcript CG5345-RB', other_designations='CG5345-PA|CG5345-PB|Ecdysone-induced polypeptide 40kD|Ecdysone-induced protein 40kD|Eip55E-PA|Eip55E-PB|ecdysone induced protein 55E', map_location='55E2-55E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18627918, end_position_on_the_genomic_accession=18629395, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37144, CurrentID=0, Status='live', Symbol='CG5482', Aliases='Dmel_ Dmel\\CG5482', description='CG5482 gene product from transcript CG5482-RC', other_designations='CG5482-PA|CG5482-PB|CG5482-PC|CG5482-PD', map_location='55E2-55E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18629594, end_position_on_the_genomic_accession=18632045, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37146, CurrentID=0, Status='live', Symbol='CG30122', Aliases='Dmel_ CG18610, CG5477, Dmel\\CG30122, LD11002, SAF-A', description='CG30122 gene product from transcript CG30122-RB', other_designations='CG30122-PB|CG30122-PC|CG30122-PD|uncharacterized protein', map_location='55E3-55E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18632328, end_position_on_the_genomic_accession=18639270, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  37147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37147, CurrentID=0, Status='live', Symbol='SP2637', Aliases='Dmel_CG5473, BEST:GH24992, CG5473, Dmel\\CG5473, dm-ntan1', description='CG5473 gene product from transcript CG5473-RC', other_designations='CG5473-PA|CG5473-PB|CG5473-PC|CG5473-PD|SP2637-PA|SP2637-PB|SP2637-PC|SP2637-PD', map_location='55E3-55E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18640862, end_position_on_the_genomic_accession=18658222, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37148, CurrentID=0, Status='live', Symbol='Gint3', Aliases='Dmel_CG5469, CG5469, Dmel\\CG5469, Dromel_CG5469_FBtr0086654_gint3_mORF, UBX 6 like', description='GDI interacting protein 3', other_designations='CG5469-PB|CG5469-PD|CG5469-PE|Gint3-PB|Gint3-PD|Gint3-PE', map_location='55E5-55E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18658418, end_position_on_the_genomic_accession=18661548, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37149, CurrentID=0, Status='live', Symbol='edl', Aliases='Dmel_CG15085, CG15085, Dmel\\CG15085, EDL, Edl, MAE, Mae, anon-WO02077276.2, anon-WO02077276.3/mae, l(2)k06602, mae, edl', description='ETS-domain lacking', other_designations='CG15085-PA|Ets domain lacking|edl-PA|modulator of the activity of Ets', map_location='55E6-55E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18667527, end_position_on_the_genomic_accession=18673536, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772702, CurrentID=0, Status='live', Symbol='His1:CG33846', Aliases='Dmel_CG33846, CG33846, Dmel\\CG33846', description='CG33846 gene product from transcript CG33846-RA', other_designations='CG33846-PA|His1:CG33846-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21506684, end_position_on_the_genomic_accession=21507706, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37152, CurrentID=0, Status='live', Symbol='Sik3', Aliases='Dmel_CG42856, CG15072, CG18604, CG42856, Dmel\\CG42856, Dmel_CG15072, Dmel_CG18604, EMK/KIAA0999, SIK, SIK3', description='Salt-inducible kinase 3', other_designations='CG42856-PA|CG42856-PB|CG42856-PC|CG42856-PD|CG42856-PE|CG42856-PF|Sik3-PA|Sik3-PB|Sik3-PC|Sik3-PD|Sik3-PE|Sik3-PF', map_location='55E8-55E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18688132, end_position_on_the_genomic_accession=18705805, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  37153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37153, CurrentID=0, Status='live', Symbol='CG15071', Aliases='Dmel_ Dmel\\CG15071', description='CG15071 gene product from transcript CG15071-RA', other_designations='CG15071-PA|uncharacterized protein', map_location='55E8-55E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18696585, end_position_on_the_genomic_accession=18697345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37155, CurrentID=0, Status='live', Symbol='CG15073', Aliases='Dmel_ Dmel\\CG15073', description='CG15073 gene product from transcript CG15073-RA', other_designations='CG15073-PA|CG15073-PB|uncharacterized protein', map_location='55E9-55E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18705977, end_position_on_the_genomic_accession=18708226, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37156, CurrentID=0, Status='live', Symbol='Vps51', Aliases='Dmel_CG15087, CG15087, Dmel\\CG15087', description='Vacuolar protein sorting 51', other_designations='CG15087-PA|Vps51-PA|vacuolar protein sorting 51', map_location='55E9-55E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18708152, end_position_on_the_genomic_accession=18710865, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37157, CurrentID=0, Status='live', Symbol='List', Aliases='Dmel_CG15088, CG15088, DmNAT2, Dmel\\CG15088', description='Lithium-inducible SLC6 transporter', other_designations='CG15088-PC|CG15088-PD|List-PC|List-PD', map_location='55E10-55E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18727817, end_position_on_the_genomic_accession=18730281, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37158, CurrentID=0, Status='live', Symbol='CG18609', Aliases='Dmel_ CT42569, Dmel\\CG18609', description='CG18609 gene product from transcript CG18609-RA', other_designations='CG18609-PA|uncharacterized protein', map_location='55E11-55E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18742517, end_position_on_the_genomic_accession=18743626, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37159, CurrentID=0, Status='live', Symbol='CG17821', Aliases='Dmel_ Dmel\\CG17821', description='CG17821 gene product from transcript CG17821-RA', other_designations='CG17821-PA|uncharacterized protein', map_location='55E11-55E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18743806, end_position_on_the_genomic_accession=18744996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37161, CurrentID=0, Status='live', Symbol='Hs3st-A', Aliases='Dmel_CG33147, CG15074, CG15075, CG15076, CG30123, CG30124, CG33147, Dmel\\CG33147, dHS3OSTA, lincRNA.S3111', description='Heparan sulfate 3-O sulfotransferase-A', other_designations='CG33147-PA|CG33147-PB|Hs3st-A-PA|Hs3st-A-PB|heparan sulfate 3-O sulfotransferase-A', map_location='55E9-55F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18718114, end_position_on_the_genomic_accession=18754675, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3772715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772715, CurrentID=0, Status='live', Symbol='His1:CG33825', Aliases='Dmel_CG33825, CG33825, Dmel\\CG33825', description='CG33825 gene product from transcript CG33825-RA', other_designations='CG33825-PA|His1:CG33825-PA|histone H1', map_location='39D4-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21471259, end_position_on_the_genomic_accession=21472281, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37164, CurrentID=0, Status='live', Symbol='Jabba', Aliases='Dmel_CG42351, 141242_at, BP1008, CG1509, CG15092, CG34020, CG42351, Dmel\\CG42351, Dmel_CG15092, Dmel_CG34020, anon-WO0118547.167', description='CG42351 gene product from transcript CG42351-RI', other_designations='CG42351-PB|CG42351-PC|CG42351-PD|CG42351-PE|CG42351-PF|CG42351-PG|CG42351-PH|CG42351-PI|Jabba-PB|Jabba-PC|Jabba-PD|Jabba-PE|Jabba-PF|Jabba-PG|Jabba-PH|Jabba-PI|jabba', map_location='55F1-55F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18757074, end_position_on_the_genomic_accession=18761082, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  37165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37165, CurrentID=0, Status='live', Symbol='Mctp', Aliases='Dmel_CG15078, CG15078, CG33146, CG33148, DmMCTP, Dmel\\CG15078', description='Multiple C2 domain and transmembrane region protein', other_designations='CG15078-PA|CG15078-PB|CG15078-PD|CG15078-PE|CG15078-PF|CG15078-PG|Mctp-PA|Mctp-PB|Mctp-PD|Mctp-PE|Mctp-PF|Mctp-PG|multiple C2 domain and transmembrane region protein', map_location='55F1-55F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18761758, end_position_on_the_genomic_accession=18774824, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  37166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37166, CurrentID=0, Status='live', Symbol='CG15093', Aliases='Dmel_ CG 15093, CT34968, Dmel\\CG15093', description='CG15093 gene product from transcript CG15093-RB', other_designations='3-hydroxyisobutyrate dehydrogenase|CG15093-PA|CG15093-PB|CG15093-PE|uncharacterized protein', map_location='55F2-55F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18775669, end_position_on_the_genomic_accession=18776978, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772719, CurrentID=0, Status='live', Symbol='CG33964', Aliases='Dmel_ CG13174, CG33154, CG8447, CT32415, Dmel\\CG33964', description='CG33964 gene product from transcript CG33964-RA', other_designations='CG33964-PA|uncharacterized protein', map_location='48E6-48E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12182639, end_position_on_the_genomic_accession=12185506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37168, CurrentID=0, Status='live', Symbol='MFS15', Aliases='Dmel_CG15094, CG15094, Dmel\\CG15094', description='Major Facilitator Superfamily Transporter 15', other_designations='CG15094-PA|CG15094-PB|MFS15-PA|MFS15-PB|major facilitator superfamily transporter 15', map_location='55F3-55F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18794618, end_position_on_the_genomic_accession=18796678, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37170, CurrentID=0, Status='live', Symbol='CG15096', Aliases='Dmel_ CT42567, Dmel\\CG15096', description='CG15096 gene product from transcript CG15096-RA', other_designations='CG15096-PA|CG15096-PB|uncharacterized protein', map_location='55F5-55F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18808356, end_position_on_the_genomic_accession=18815019, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37172, CurrentID=0, Status='live', Symbol='CG15097', Aliases='Dmel_ Dm.CG15097, Dmel\\CG15097', description='CG15097 gene product from transcript CG15097-RC', other_designations='CG15097-PA|CG15097-PC|CG15097-PD|CG15097-PE|uncharacterized protein', map_location='55F6-55F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18822188, end_position_on_the_genomic_accession=18829699, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37173, CurrentID=0, Status='live', Symbol='CG15082', Aliases='Dmel_ Dmel\\CG15082', description='CG15082 gene product from transcript CG15082-RB', other_designations='CG15082-PA|CG15082-PB|uncharacterized protein', map_location='55F6-55F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18830331, end_position_on_the_genomic_accession=18833241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37174, CurrentID=0, Status='live', Symbol='CG15098', Aliases='Dmel_ Dmel\\CG15098, dmCG15098', description='CG15098 gene product from transcript CG15098-RA', other_designations='CG15098-PA|uncharacterized protein', map_location='55F6-55F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18833372, end_position_on_the_genomic_accession=18834775, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37175, CurrentID=0, Status='live', Symbol='CG15083', Aliases='Dmel_ BcDNA:RH17222, Dmel\\CG15083', description='CG15083 gene product from transcript CG15083-RA', other_designations='CG15083-PA|uncharacterized protein', map_location='55F6-55F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18835263, end_position_on_the_genomic_accession=18836842, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37176, CurrentID=0, Status='live', Symbol='CG15099', Aliases='Dmel_ Dmel\\CG15099', description='CG15099 gene product from transcript CG15099-RB', other_designations='CG15099-PA|CG15099-PB|uncharacterized protein', map_location='55F6-55F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18836821, end_position_on_the_genomic_accession=18846619, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  37177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37177, CurrentID=0, Status='live', Symbol='MetRS', Aliases='Dmel_CG15100, CG15100, Dmel\\CG15100, MRS, anon-EST:Liang-2.15, clone 2.15, dMetRS, p120', description='Methionyl-tRNA synthetase', other_designations='CG15100-PA|MetRS-PA', map_location='55F7-55F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18846858, end_position_on_the_genomic_accession=18850599, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37178, CurrentID=0, Status='live', Symbol='CG15084', Aliases='Dmel_ Dmel\\CG15084', description='CG15084 gene product from transcript CG15084-RA', other_designations='CG15084-PA|uncharacterized protein', map_location='55F7-55F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18850844, end_position_on_the_genomic_accession=18851963, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37179, CurrentID=0, Status='live', Symbol='CG18190', Aliases='Dmel_ Dmel\\CG18190', description='CG18190 gene product from transcript CG18190-RB', other_designations='CG18190-PB|uncharacterized protein', map_location='55F7-55F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18851907, end_position_on_the_genomic_accession=18853043, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37181, CurrentID=0, Status='live', Symbol='Jheh2', Aliases='Dmel_CG15102, CG15102, CT34977, DmEH, Dmel\\CG15102, jheh2', description='Juvenile hormone epoxide hydrolase 2', other_designations='CG15102-PA|CG15102-PD|Jheh2-PA|Jheh2-PD|juvenile hormone epoxide hydrolase 2|juvenile hormone esterase', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18855767, end_position_on_the_genomic_accession=18858245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37182, CurrentID=0, Status='live', Symbol='Jheh3', Aliases='Dmel_CG15106, CG15106, Dmel\\CG15106', description='Juvenile hormone epoxide hydrolase 3', other_designations='CG15106-PA|Jheh3-PA|juvenile hormone epoxide hydrolase 3', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18860375, end_position_on_the_genomic_accession=18862078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37183, CurrentID=0, Status='live', Symbol='Dpt', Aliases='Dmel_CG12763, 143443_at, CG12763, DIM 27, DIPT, DPT, Dep, Dip, Dipt, DiptA, Dit, Dmel\\CG12763A, dip, dipt, dpt, Dpt', description='Diptericin', other_designations='CG12763-PA|Diptericin-A|DiptericinA|Dpt-PA|dipt|diptericin|diptericin A|diptericin precursor', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18865765, end_position_on_the_genomic_accession=18866260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37184, CurrentID=0, Status='live', Symbol='DptB', Aliases='Dmel_CG10794, 147473_at, BcDNA:RH29451, CG10794, DiptB, Diptericin, Dmel\\CG10794, Dpt, dpt B, dptB, dptlp', description='Diptericin B', other_designations='CG10794-PA|Diptericin-B|DiptericinB|DptB-PA|diptercin B|diptericin|diptericin B|diptericin-like protein', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18867391, end_position_on_the_genomic_accession=18867895, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37185, CurrentID=0, Status='live', Symbol='sano', Aliases='Dmel_CG12758, CG12758, Dmel\\CG12758', description='serrano', other_designations='CG12758-PA|CG12758-PB|CG12758-PC|CG12758-PD|CG12758-PE|CG12758-PF|sano-PA|sano-PB|sano-PC|sano-PD|sano-PE|sano-PF', map_location='55F8-56A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18878391, end_position_on_the_genomic_accession=18959500, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37187, CurrentID=0, Status='live', Symbol='prod', Aliases='Dmel_CG18608, CG18608, Dmel\\CG18608, PROD, Prod, l(2)88/10, l(2)k08810', description='proliferation disrupter', other_designations='CG18608-PA|prod-PA', map_location='56A2-56A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18969404, end_position_on_the_genomic_accession=18971133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37188, CurrentID=0, Status='live', Symbol='Topors', Aliases='Dmel_CG15104, CG15104, Dmel\\CG15104, TOPORS, dTopors, dtopors, nbl', description='Topoisomerase I-interacting protein', other_designations='CG15104-PA|CG15104-PB|Topors-PA|Topors-PB|nuclear blebber|topoisomerase I-interacting protein', map_location='56A2-56A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18972612, end_position_on_the_genomic_accession=18977262, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37189, CurrentID=0, Status='live', Symbol='CG18605', Aliases='Dmel_ Dmel\\CG18605', description='CG18605 gene product from transcript CG18605-RB', other_designations='CG18605-PB|uncharacterized protein', map_location='56A2-56A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18977514, end_position_on_the_genomic_accession=18979279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37190, CurrentID=0, Status='live', Symbol='tn', Aliases='Dmel_CG15105, Abba, CG15105, Dmel\\CG15105, abba, l(2)tn', description='thin', other_designations='CG15105-PA|CG15105-PB|CG15105-PC|CG15105-PD|CG15105-PE|CG15105-PF|another B-box affiliate|l(2)thin|tn-PA|tn-PB|tn-PC|tn-PD|tn-PE|tn-PF', map_location='2-85.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18987344, end_position_on_the_genomic_accession=19008944, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  37191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37191, CurrentID=0, Status='live', Symbol='5-HT1B', Aliases='Dmel_CG15113, 5-HT-dro2BDro, 5-HT[1B]Dro, 5-HT[[1BDro]], 5-HT[[1B]], 5-HT[[1B]]DRO, 5-HT[[1B]]Dro, 5-HT[[dro2B]], 5HT-R2B, 5HT-dro2B, 5HT-dro[[2B]], 5HT1B, 5htdro2b, CG15113, CG30126, DRO2B, Dm5HTdro2B, Dmel\\CG15113, Dmel_CG30126, d5-HT1B, 5-HT1B', description='5-hydroxytryptamine (serotonin) receptor 1B', other_designations='5-HT1B-PA|5-HT1B-PD|5-HT1B-PE|CG15113-PA|CG15113-PD|CG15113-PE|drosophila serotonin receptor 1B|serotonin receptor 1B|serotonin receptor 2B', map_location='56B1-56B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19022325, end_position_on_the_genomic_accession=19032642, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37192, CurrentID=0, Status='live', Symbol='CG15115', Aliases='Dmel_ Dmel\\CG15115', description='CG15115 gene product from transcript CG15115-RA', other_designations='CG15115-PA|uncharacterized protein', map_location='56B1-56B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19035492, end_position_on_the_genomic_accession=19036387, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37194, CurrentID=0, Status='live', Symbol='CG15116', Aliases='Dmel_ Dmel\\CG15116, GPx-like', description='CG15116 gene product from transcript CG15116-RA', other_designations='CG15116-PA|uncharacterized protein', map_location='56B2-56B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19055231, end_position_on_the_genomic_accession=19055924, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37195, CurrentID=0, Status='live', Symbol='CG15109', Aliases='Dmel_ Dmel\\CG15109', description='CG15109 gene product from transcript CG15109-RC', other_designations='CG15109-PC|CG15109-PD|CG15109-PE|uncharacterized protein', map_location='56B2-56B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19059421, end_position_on_the_genomic_accession=19061065, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37196, CurrentID=0, Status='live', Symbol='5-HT1A', Aliases='Dmel_CG16720, 5-HT-dro2ADro, 5-HT[1A]Dro, 5-HT[[1ADro]], 5-HT[[1A]], 5-HT[[1A]]DRO, 5HT-R2A, 5HT-dro2A, 5HT-drp[[2A]], 5HT1A, 5HT[[1ADro]], 5htdro2a, CG16720, DRO2A, Dm5HTdro2A, Dmel\\CG16720, d5-HT1A, 5-HT1A', description='5-hydroxytryptamine (serotonin) receptor 1A', other_designations='5-HT1A-PB|CG16720-PB|serotonin receptor 1A|serotonin receptor 2A|serotonin-receptor-2A', map_location='56B2-56B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19067155, end_position_on_the_genomic_accession=19121958, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37197, CurrentID=0, Status='live', Symbol='CG15117', Aliases='Dmel_ Dmel\\CG15117', description='CG15117 gene product from transcript CG15117-RA', other_designations='CG15117-PA|CG15117-PB|CG15117-PC|CG15117-PD|CG15117-PE|CG15117-PF|uncharacterized protein', map_location='56B5-56B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19122231, end_position_on_the_genomic_accession=19129300, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  37198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37198, CurrentID=0, Status='live', Symbol='botv', Aliases='Dmel_CG15110, BcDNA:LD21192, Botv, CG15110, DEXT3, DEXTL3, DExt3, Dextl3, Dmel\\CG15110, Extl3, dExt3', description='brother of tout-velu', other_designations='CG15110-PA|botv-PA|brother of tout velu|brother of tout-velo|brother-of-tout-velu|drosophila Ext-like 3', map_location='56B5-56B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19129589, end_position_on_the_genomic_accession=19133706, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37199, CurrentID=0, Status='live', Symbol='CG15118', Aliases='Dmel_ Dmel\\CG15118, anon-EST:fe1H12', description='CG15118 gene product from transcript CG15118-RA', other_designations='CG15118-PA|CG15118-PB|CG15118-PD|CG15118-PE|anon-fast-evolving-1H12|uncharacterized protein', map_location='56B5-56B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19133727, end_position_on_the_genomic_accession=19137169, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37200, CurrentID=0, Status='live', Symbol='CG15111', Aliases='Dmel_ Dmel\\CG15111', description='CG15111 gene product from transcript CG15111-RB', other_designations='CG15111-PA|CG15111-PB|uncharacterized protein', map_location='56B5-56B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19137372, end_position_on_the_genomic_accession=19141034, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37201, CurrentID=0, Status='live', Symbol='ena', Aliases='Dmel_CG15112, CG15112, Dmel\\CG15112, ENA/VASP, Ena, Enb, VASP, Vasp, enb, l(2)02029', description='enabled', other_designations='CG15112-PA|CG15112-PB|CG15112-PC|CG15112-PD|CG15112-PE|CG15112-PF|CG15112-PG|VASP|ena-PA|ena-PB|ena-PC|ena-PD|ena-PE|ena-PF|ena-PG', map_location='2-87 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19141632, end_position_on_the_genomic_accession=19163705, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37202, CurrentID=0, Status='live', Symbol='CG10737', Aliases='Dmel_ 10737, Dmel\\CG10737', description='CG10737 gene product from transcript CG10737-RS', other_designations='CG10737-PA|CG10737-PC|CG10737-PD|CG10737-PE|CG10737-PN|CG10737-PO|CG10737-PP|CG10737-PR|CG10737-PS|CG10737-PT|CG10737-PU|CG10737-PV|CG10737-PW|CG10737-PX|gene 1|uncharacterized protein', map_location='56C1-56C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19164143, end_position_on_the_genomic_accession=19178223, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  37203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37203, CurrentID=0, Status='live', Symbol='hppy', Aliases='Dmel_CG7097, CG7097, CG7097/hppy, Dmel\\CG7097, EP-704, GLK/KHS1, MAP4K3, anon-WO0153538.68, l(2)SH1261, l(2)SH2 1261', description='happyhour', other_designations='CG7097-PA|CG7097-PB|CG7097-PC|CG7097-PD|CG7097-PE|gene 2|hppy-PA|hppy-PB|hppy-PC|hppy-PD|hppy-PE|lethal (2) SH1261', map_location='56C1-56C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19178738, end_position_on_the_genomic_accession=19227193, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  37204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37204, CurrentID=0, Status='live', Symbol='CG7137', Aliases='Dmel_ Dmel\\CG7137', description='CG7137 gene product from transcript CG7137-RA', other_designations='CG7137-PA|gene 3', map_location='56C4-56C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19227344, end_position_on_the_genomic_accession=19228613, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37205, CurrentID=0, Status='live', Symbol='cora', Aliases='Dmel_CG11949, CG11949, COR, Cor, Cora, Coracle, D4.1, Dmel\\CG11949, cor, l(2)k08713, mdcds_14526', description='coracle', other_designations='CG11949-PA|CG11949-PB|CG11949-PC|CG11949-PE|CG11949-PF|CG11949-PG|D4.1-Coracle|cora-PA|cora-PB|cora-PC|cora-PE|cora-PF|cora-PG', map_location='2-86.9 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19229063, end_position_on_the_genomic_accession=19243751, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  37206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37206, CurrentID=0, Status='live', Symbol='wbl', Aliases='Dmel_CG7225, CG7225, Dmel\\CG7225, Wbl, Wind, Windb, wbt, wind', description='windbeutel', other_designations='CG7225-PA|CG7225-PB|wbl-PA|wbl-PB', map_location='2-86 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19249395, end_position_on_the_genomic_accession=19250917, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37207, CurrentID=0, Status='live', Symbol='tbrd-2', Aliases='Dmel_CG7229, 7229, CG7229, Dmel\\CG7229, tBRD-2', description='testis-specifically expressed bromodomain containing protein-2', other_designations='CG7229-PA|gene 4|tbrd-2-PA', map_location='56C6-56C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19259909, end_position_on_the_genomic_accession=19262294, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37209, CurrentID=0, Status='live', Symbol='CG11906', Aliases='Dmel_ 11906, Dmel\\CG11906', description='CG11906 gene product from transcript CG11906-RA', other_designations='CG11906-PA|gene 6|uncharacterized protein', map_location='56C6-56C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19281418, end_position_on_the_genomic_accession=19283961, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37210, CurrentID=0, Status='live', Symbol='CG10476', Aliases='Dmel_ Dmel\\CG10476', description='CG10476 gene product from transcript CG10476-RA', other_designations='CG10476-PA|uncharacterized protein', map_location='56C6-56C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19284551, end_position_on_the_genomic_accession=19285219, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37211, CurrentID=0, Status='live', Symbol='CG10474', Aliases='Dmel_ Dmel\\CG10474', description='CG10474 gene product from transcript CG10474-RA', other_designations='CG10474-PA|uncharacterized protein', map_location='56C7-56C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19285395, end_position_on_the_genomic_accession=19286730, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37212, CurrentID=0, Status='live', Symbol='CG18606', Aliases='Dmel_ Dmel\\CG18606', description='CG18606 gene product from transcript CG18606-RB', other_designations='CG18606-PB|uncharacterized protein', map_location='56C8-56C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19287983, end_position_on_the_genomic_accession=19288995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37213, CurrentID=0, Status='live', Symbol='CG18607', Aliases='Dmel_ Dmel\\CG18607', description='CG18607 gene product from transcript CG18607-RA', other_designations='CG18607-PA|uncharacterized protein', map_location='56C8-56C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19289109, end_position_on_the_genomic_accession=19289852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37214, CurrentID=0, Status='live', Symbol='FK506-bp2', Aliases='Dmel_CG11001, 11001, 143729_at, CG11001, Dmel\\CG11001, DrFKBP12, FKBP, FKBP12, Fk506-bp2, dFKBP12, fkbp', description='FK506-binding protein 2', other_designations='CG11001-PA|FK506-binding protein|FK506-bp2-PA', map_location='56C8-56C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19290109, end_position_on_the_genomic_accession=19290962, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37215, CurrentID=0, Status='live', Symbol='mip40', Aliases='Dmel_CG15119, CG15119, Dmel\\CG15119, Mip40, p40', description='Myb-interacting protein 40', other_designations='CG15119-PA|mip40-PA', map_location='56C8-56C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19291105, end_position_on_the_genomic_accession=19292196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37216, CurrentID=0, Status='live', Symbol='Tab2', Aliases='Dmel_CG7417, 7417, CG7417, Dmel\\CG7417, ORF1, TAB, TAB2, cg7417, dTAB2, dTab2, galere, tab-2, tab2', description='TAK1-associated binding protein 2', other_designations='CG7417-PA|CG7417-PB|CG7417-PC|TAK1-associated binding Protein2|TGF-beta-activated kinase 1/MAP3K7 binding protein 2|Tab2-PA|Tab2-PB|Tab2-PC|galere', map_location='56C8-56C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19292529, end_position_on_the_genomic_accession=19304508, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37217, CurrentID=0, Status='live', Symbol='CG7461', Aliases='Dmel_ Dmel\\CG7461, cg7461', description='CG7461 gene product from transcript CG7461-RC', other_designations='CG7461-PB|CG7461-PC|drosophila very-long chain ACAD', map_location='56C9-56C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19305172, end_position_on_the_genomic_accession=19307445, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37218, CurrentID=0, Status='live', Symbol='EndoB', Aliases='Dmel_CG9834, CG9834, D-EndoB, D-endoB, Dmel\\CG9834, endoB', description='Endophilin B', other_designations='CG9834-PA|CG9834-PB|CG9834-PC|CG9834-PD|EndoB-PA|EndoB-PB|EndoB-PC|EndoB-PD|endophilin|endophilin B', map_location='56C10-56C10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19307447, end_position_on_the_genomic_accession=19310665, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  5740603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740603, CurrentID=0, Status='live', Symbol='CG34442', Aliases='Dmel_ CG30070, Dmel\\CG34442, Dmel_CG30070', description='CG34442 gene product from transcript CG34442-RA', other_designations='CG34442-PA|uncharacterized protein', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14366798, end_position_on_the_genomic_accession=14367874, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37220, CurrentID=0, Status='live', Symbol='CG9975', Aliases='Dmel_ Dmel\\CG9975', description='CG9975 gene product from transcript CG9975-RA', other_designations='CG9975-PA|CG9975-PB', map_location='56D1-56D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19341809, end_position_on_the_genomic_accession=19345754, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37221, CurrentID=0, Status='live', Symbol='CG11961', Aliases='Dmel_CG11961-PA, Dmel\\CG11961', description='CG11961 gene product from transcript CG11961-RB', other_designations='CG11961-PA|CG11961-PB|CG11961-PC|uncharacterized protein', map_location='56D1-56D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19357532, end_position_on_the_genomic_accession=19363124, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37222, CurrentID=0, Status='live', Symbol='CG10051', Aliases='Dmel_ Dmel\\CG10051', description='CG10051 gene product from transcript CG10051-RA', other_designations='CG10051-PA|uncharacterized protein', map_location='56D2-56D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19363641, end_position_on_the_genomic_accession=19367294, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37223, CurrentID=0, Status='live', Symbol='CG9416', Aliases='Dmel_CG9416-PA, Dmel\\CG9416', description='CG9416 gene product from transcript CG9416-RB', other_designations='CG9416-PA|CG9416-PB', map_location='56D2-56D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19367248, end_position_on_the_genomic_accession=19373702, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  37224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37224, CurrentID=0, Status='live', Symbol='CG10062', Aliases='Dmel_ Dmel\\CG10062, anon-WO0140519.226', description='CG10062 gene product from transcript CG10062-RC', other_designations='CG10062-PA|CG10062-PC|CG10062-PD|uncharacterized protein', map_location='56D2-56D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19375313, end_position_on_the_genomic_accession=19380174, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37225, CurrentID=0, Status='live', Symbol='CG10073', Aliases='Dmel_ Dmel\\CG10073', description='CG10073 gene product from transcript CG10073-RC', other_designations='CG10073-PA|CG10073-PC|uncharacterized protein', map_location='56D2-56D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19380543, end_position_on_the_genomic_accession=19384965, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37226, CurrentID=0, Status='live', Symbol='CG10081', Aliases='Dmel_ Dmel\\CG10081', description='CG10081 gene product from transcript CG10081-RA', other_designations='CG10081-PA|CG10081-PB|uncharacterized protein', map_location='56D2-56D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19385923, end_position_on_the_genomic_accession=19389318, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37227, CurrentID=0, Status='live', Symbol='CG11257', Aliases='Dmel_ Dmel\\CG11257', description='CG11257 gene product from transcript CG11257-RA', other_designations='CG11257-PA|uncharacterized protein', map_location='56D3-56D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19390128, end_position_on_the_genomic_accession=19391900, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37228, CurrentID=0, Status='live', Symbol='SdhA', Aliases='Dmel_CG17246, CG17246, DHSA, Dmel\\CG17246, SDH, Scs-fp, Sdh, Sdh A, Sdha, scs-fp, sdh', description='Succinate dehydrogenase, subunit A (flavoprotein)', other_designations='CG17246-PA|CG17246-PB|CG17246-PC|COMPLEX II: flavoprotein subunit|SUCCINYL COENZIME A SYNTHETASE FLAVOPROTEIN SUBUNIT|SdhA-PA|SdhA-PB|SdhA-PC|succinate dehydrogenase|succinate dehydrogenase A|succinate dehydrogenase [ubiquinone] flavoprotein subunit|succinate dehydrogenase flavoprotein|succinate dehydrogenase flavoprotein subunit|succinate dehydrogenase, subunit A (flavoprotein)|succinyl CoA synthetase flavoprotein subunit of complex II|succinyl coenzyme A synthetase flavoprotein subunit', map_location='56D3-56D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19391863, end_position_on_the_genomic_accession=19395971, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37229, CurrentID=0, Status='live', Symbol='cer', Aliases='Dmel_CG10460, CER, CG10460, D/CTLA-2, Dmel\\CG10460', description='crammer', other_designations='CG10460-PA|cer-PA|drosophila CTLA-2like protein', map_location='56D3-56D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19396215, end_position_on_the_genomic_accession=19396772, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37230, CurrentID=0, Status='live', Symbol='hts', Aliases='Dmel_CG43443, 1B1, ADD-87, Add, Add-hts, Adducin, CG34197, CG43443, CG9325, Dmel\\CG43443, Dmel_CG34197, Dmel_CG9325, EST D, HTS, HTS-R1, HTS-RC, HTS_DROME, Hts, Hts-RC, HtsF, HtsRC, Ovhts, Ovhts-RC, add, add-like, adducin, anon-EST:Posey9RC, l(2)00634, l(2)01103, l(2)k06121, l(2)k14523, hts', description='hu li tai shao', other_designations='Adducin-like 56D|Adducin-like-56D|CG43443-PA|CG43443-PC|CG43443-PD|CG43443-PH|CG43443-PI|CG43443-PJ|CG43443-PK|CG43443-PL|CG43443-PM|CG43443-PN|CG43443-PO|CG43443-PP|CG43443-PQ|CG43443-PR|CG43443-PS|Hu Li Tai shao|Hu li-tai shao|Hu-li tai shao|Hu-li tai-shao|Hu-li-tai-shao|adducin|adducin-like|hts-PA|hts-PC|hts-PD|hts-PH|hts-PI|hts-PJ|hts-PK|hts-PL|hts-PM|hts-PN|hts-PO|hts-PP|hts-PQ|hts-PR|hts-PS|hu-li tai shao protein|ovarian hu-li tai shao-ring canal protein', map_location='56D3-56D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19397032, end_position_on_the_genomic_accession=19424903, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  5740605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740605, CurrentID=0, Status='live', Symbol='CG34279', Aliases='Dmel_ Dmel\\CG34279', description='CG34279 gene product from transcript CG34279-RA', other_designations='CG34279-PA|CG34279-PC|uncharacterized protein', map_location='90A5-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17402362, end_position_on_the_genomic_accession=17403288, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37232, CurrentID=0, Status='live', Symbol='CalpA', Aliases='Dmel_CG7563, CG18152, CG7563, Dmel\\CG7563, EST C, calpA', description='Calpain-A', other_designations='CG7563-PA|CG7563-PB|CG7563-PC|CG7563-PD|CalpA-PA|CalpA-PB|CalpA-PC|CalpA-PD|calpain|calpain A|calpain lacking calmodulin-like domain|calpainA', map_location='56D5-56D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19425113, end_position_on_the_genomic_accession=19430564, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  37233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37233, CurrentID=0, Status='live', Symbol='Fak', Aliases='Dmel_CG10023, CG10023, CT28129, DFAK, DFak, DFak56, Dfak, Dfak56, DmFAK, Dmel\\CG10023, FAK, FAK65D56, Fak56D, dFak, fak, fak56, ptk2, Fak', description='Focal adhesion kinase', other_designations='CG10023-PA|CG10023-PB|CG10023-PC|CG10023-PD|CG10023-PF|CG10023-PG|Fak-PA|Fak-PB|Fak-PC|Fak-PD|Fak-PF|Fak-PG|focal adhesion kinase|focal adhesion kinase homolog|protein tyrosine kinase 2', map_location='56D5-56D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19430659, end_position_on_the_genomic_accession=19437259, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  37235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37235, CurrentID=0, Status='live', Symbol='RpL11', Aliases='Dmel_CG7726, BcDNA:LD17235, CG7726, DL11, Dmel\\CG7726, EST A, L11, RPL11, Rp L11, chr2R:14957112..14957299, l(2)k16914', description='Ribosomal protein L11', other_designations='60S ribosomal protein L11|CG7726-PA|CG7726-PB|RpL11-PA|RpL11-PB|ribosomal protein DL11|ribosomal protein L11', map_location='56D7-56D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19442342, end_position_on_the_genomic_accession=19443825, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37236, CurrentID=0, Status='live', Symbol='Arl6', Aliases='Dmel_CG7735, A1ZBK9, ARL6, BBS3, CG7735, Dmel\\CG7735', description='ADP ribosylation factor-like 6', other_designations='Arl6-PA|Arl6-PB|CG7735-PA|CG7735-PB', map_location='56D7-56D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19444179, end_position_on_the_genomic_accession=19445058, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37237, CurrentID=0, Status='live', Symbol='EloC', Aliases='Dmel_CG9291, BcDNA:RH71704, CG9291, Dmel\\CG9291, Elongin-C, ElonginC, dElongin C, l(2)SH1299, l(2)SH1520, l(2)SH2 1520', description='Elongin C', other_designations='CG9291-PA|CG9291-PB|EloC-PA|EloC-PB|ElonginC|d-elongin C|elongin C|lethal (2) SH1520', map_location='56D7-56D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19445085, end_position_on_the_genomic_accession=19446286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37238, CurrentID=0, Status='live', Symbol='betaTub56D', Aliases='Dmel_CG9277, 1t, A1ZBL0_DROME, B1t, BETA 56D, CG9277, DTB2, Dmbeta1, Dmel\\CG9277, T, TBB1_DROME, Tub, Tubulin, beta, beta-Tub, beta-Tub56D, beta-tub, beta-tubulin56D, beta1, beta1-Tubulin, beta1-tub, beta1Tub, beta1t, beta1tub, beta56D, betaTub, betaTub1, beta[[1]] tubulin, beta[[1]]-tubulin, betatub(56D), betatub56D', description='beta-Tubulin at 56D', other_designations='CG9277-PA|CG9277-PB|b-tubulin56D|beta tubulin|beta-1 tubulin|beta-1-tubulin|beta-Tubulin|beta-Tubulin 56D|beta-Tubulin at 56 D|beta-tubulin-56D|beta1 tubulin|beta1-Tubulin|betaTub56D-PA|betaTub56D-PB|betatubulin|tubulin|tubulin beta-1 chain|tubulin beta1-chain', map_location='56D7-56D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19447319, end_position_on_the_genomic_accession=19451600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37239, CurrentID=0, Status='live', Symbol='CG7744', Aliases='Dmel_ Dmel\\CG7744', description='CG7744 gene product from transcript CG7744-RA', other_designations='CG7744-PA', map_location='56D8-56D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19452504, end_position_on_the_genomic_accession=19455527, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37244, CurrentID=0, Status='live', Symbol='TBCB', Aliases='Dmel_CG11242, CG11242, Dmel\\CG11242, dTBCB', description='tubulin-binding cofactor B', other_designations='CG11242-PA|TBCB-PA', map_location='56D11-56D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19485494, end_position_on_the_genomic_accession=19486837, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37245, CurrentID=0, Status='live', Symbol='Oseg6', Aliases='Dmel_CG11237, CG11237, Dmel\\CG11237, OSEG, OSEG6, WDR19, Wdr19, oseg6', description='Outer segment 6', other_designations='CG11237-PA|Oseg6-PA|outer segment 6', map_location='56D12-56D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19487046, end_position_on_the_genomic_accession=19491846, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37246, CurrentID=0, Status='live', Symbol='Rep', Aliases='Dmel_CG8432, CG8432, Dmel\\CG8432, REP', description='Rab escort protein', other_designations='CG8432-PA|Rab escort protein homolog|Rep-PA', map_location='56D13-56D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19492087, end_position_on_the_genomic_accession=19494125, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37247, CurrentID=0, Status='live', Symbol='hpo', Aliases='Dmel_CG11228, CG11228, Dmel\\CG11228, Hippo, Hpo, MST, MST2, dMST, hipo', description='hippo', other_designations='CG11228-PA|CG11228-PB|MST|hippo kinase|hpo-PA|hpo-PB', map_location='56D13-56D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19493996, end_position_on_the_genomic_accession=19496856, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37248, CurrentID=0, Status='live', Symbol='CG15120', Aliases='Dmel_ Dmel\\CG15120', description='CG15120 gene product from transcript CG15120-RA', other_designations='CG15120-PA|uncharacterized protein', map_location='56D13-56D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19497752, end_position_on_the_genomic_accession=19500372, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37249, CurrentID=0, Status='live', Symbol='CG11007', Aliases='Dmel_ Dmel\\CG11007', description='CG11007 gene product from transcript CG11007-RB', other_designations='CG11007-PA|CG11007-PB|uncharacterized protein', map_location='56D14-56D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19503068, end_position_on_the_genomic_accession=19504594, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37250, CurrentID=0, Status='live', Symbol='Ir56b', Aliases='Dmel_CG15121, CG15121, DmelIR56b, Dmel\\CG15121, IR56b', description='Ionotropic receptor 56b', other_designations='CG15121-PA|Ir56b-PA|ionotropic receptor 56b', map_location='56D15-56D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19508167, end_position_on_the_genomic_accession=19509653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37251, CurrentID=0, Status='live', Symbol='Ir56c', Aliases='Dmel_CG15122, CG15122, DmelIR56c, Dmel\\CG15122, IR56c', description='Ionotropic receptor 56c', other_designations='CG15122-PA|Ir56c-PA|ionotropic receptor 56c', map_location='56D15-56D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19509768, end_position_on_the_genomic_accession=19511471, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37252, CurrentID=0, Status='live', Symbol='Ir56d', Aliases='Dmel_CG15904, CG15904, DmelIR56d, Dmel\\CG15904, IR56d', description='Ionotropic receptor 56d', other_designations='CG15904-PA|Ir56d-PA|ionotropic receptor 56d', map_location='56D15-56D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19511673, end_position_on_the_genomic_accession=19513571, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37253, CurrentID=0, Status='live', Symbol='CG16716', Aliases='Dmel_ Dmel\\CG16716, anon-WO0140519.27', description='CG16716 gene product from transcript CG16716-RC', other_designations='CG16716-PC|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19593790, end_position_on_the_genomic_accession=19596866, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37254, CurrentID=0, Status='live', Symbol='sm', Aliases='Dmel_CG9218, BcDNA:GH10856, CG9218, Dmel\\CG9218, l(2)05338o, sm', description='smooth', other_designations='CG9218-PA|CG9218-PAA|CG9218-PAB|CG9218-PAC|CG9218-PB|CG9218-PE|CG9218-PG|CG9218-PH|CG9218-PI|CG9218-PJ|CG9218-PK|CG9218-PL|CG9218-PM|CG9218-PO|CG9218-PP|CG9218-PT|CG9218-PU|CG9218-PV|CG9218-PW|CG9218-PY|CG9218-PZ|sm-PA|sm-PAA|sm-PAB|sm-PAC|sm-PB|sm-PE|sm-PG|sm-PH|sm-PI|sm-PJ|sm-PK|sm-PL|sm-PM|sm-PO|sm-PP|sm-PT|sm-PU|sm-PV|sm-PW|sm-PY|sm-PZ', map_location='2-91.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19517864, end_position_on_the_genomic_accession=19631508, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  37255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37255, CurrentID=0, Status='live', Symbol='CG18367', Aliases='Dmel_ CG15123, Dmel\\CG18367', description='CG18367 gene product from transcript CG18367-RA', other_designations='CG18367-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19636660, end_position_on_the_genomic_accession=19637400, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37256, CurrentID=0, Status='live', Symbol='CG15124', Aliases='Dmel_ Dmel\\CG15124', description='CG15124 gene product from transcript CG15124-RA', other_designations='CG15124-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19642947, end_position_on_the_genomic_accession=19643698, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37257, CurrentID=0, Status='live', Symbol='CG15905', Aliases='Dmel_ Dmel\\CG15905', description='CG15905 gene product from transcript CG15905-RC', other_designations='CG15905-PB|CG15905-PC|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19646483, end_position_on_the_genomic_accession=19647645, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772186, CurrentID=0, Status='live', Symbol='CG33695', Aliases='Dmel_ CG16917, CG32831, CG32955, CG32955-ORFB, CG4831, Dmel\\CG33695, l(2)k07716', description='CG33695 gene product from transcript CG33695-RC', other_designations='CG33695-PC|CG33695-PD|CG33695-PE|CG33695-PF|uncharacterized protein', map_location='32E2-32E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11278021, end_position_on_the_genomic_accession=11279982, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37259, CurrentID=0, Status='live', Symbol='CG11018', Aliases='Dmel_ Dmel\\CG11018, anon-WO0140519.121', description='CG11018 gene product from transcript CG11018-RA', other_designations='CG11018-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19662436, end_position_on_the_genomic_accession=19663738, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37261, CurrentID=0, Status='live', Symbol='isopeptidase-T-3', Aliases='Dmel_CG11025, CG11025, Dmel\\CG11025, ISOT-3, IsoT-3', description='CG11025 gene product from transcript CG11025-RC', other_designations='CG11025-PA|CG11025-PB|CG11025-PC|isopeptidase T-3|isopeptidase-T-3-PA|isopeptidase-T-3-PB|isopeptidase-T-3-PC', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19670618, end_position_on_the_genomic_accession=19673477, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37263, CurrentID=0, Status='live', Symbol='CG15128', Aliases='Dmel_ Dmel\\CG15128', description='CG15128 gene product from transcript CG15128-RA', other_designations='CG15128-PA|uncharacterized protein', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19685614, end_position_on_the_genomic_accession=19688078, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37264, CurrentID=0, Status='live', Symbol='Obp56a', Aliases='Dmel_CG11797, 56a, BcDNA:RE46170, CG11797, DmelObp56a, Dmel\\CG11797, obp56a', description='Odorant-binding protein 56a', other_designations='CG11797-PA|Obp56a-PA', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19697723, end_position_on_the_genomic_accession=19698508, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37265, CurrentID=0, Status='live', Symbol='Obp56b', Aliases='Dmel_CG30129, 56b, CG15129, CG30129, DmelObp56b, Dmel\\CG30129, OBP56b', description='Odorant-binding protein 56b', other_designations='CG30129-PA|Obp56b-PA', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19698714, end_position_on_the_genomic_accession=19699346, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37266, CurrentID=0, Status='live', Symbol='Obp56d', Aliases='Dmel_CG11218, 56d, BEST:GH09027, BcDNA:RE22207, CG11218, CT31326, DmelObp56d, Dmel\\CG11218, OBP56d, obp56d', description='Odorant-binding protein 56d', other_designations='CG11218-PA|CG11218-PB|Obp56d-PA|Obp56d-PB', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19702924, end_position_on_the_genomic_accession=19703915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37267, CurrentID=0, Status='live', Symbol='Obp56e', Aliases='Dmel_CG8462, 56e, CG8462, DmelObp56e, Dmel\\CG8462, obp56e', description='Odorant-binding protein 56e', other_designations='CG8462-PA|CG8462-PB|Obp56e-PA|Obp56e-PB|odorant binding protein 56e', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19712345, end_position_on_the_genomic_accession=19712932, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37268, CurrentID=0, Status='live', Symbol='CG8517', Aliases='Dmel_ 8517, Dmel\\CG8517, mtr', description='CG8517 gene product from transcript CG8517-RA', other_designations='CG8517-PA|thioredoxin', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19714076, end_position_on_the_genomic_accession=19714868, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37269, CurrentID=0, Status='live', Symbol='Or56a', Aliases='Dmel_CG12501, 56E.1, 56a, CG12501, DOR59, Dmel\\CG12501', description='Odorant receptor 56a', other_designations='CG12501-PA|Or56a-PA|odorant receptor 56a', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19769461, end_position_on_the_genomic_accession=19771233, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37270, CurrentID=0, Status='live', Symbol='Obp56g', Aliases='Dmel_CG13873, 56g, CG13873, DmelObp56g, Dmel\\CG13873, OBP56g', description='Odorant-binding protein 56g', other_designations='CG13873-PA|CG13873-PB|Obp56g-PA|Obp56g-PB', map_location='56E3-56E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19783484, end_position_on_the_genomic_accession=19784043, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37271, CurrentID=0, Status='live', Symbol='Obp56h', Aliases='Dmel_CG13874, 56h, CG13874, DmelObp56h, Dmel\\CG13874, OBP56h', description='Odorant-binding protein 56h', other_designations='CG13874-PA|CG13874-PB|Obp56h-PA|Obp56h-PB', map_location='56E4-56E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19815554, end_position_on_the_genomic_accession=19816204, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37272, CurrentID=0, Status='live', Symbol='Toll-7', Aliases='Dmel_CG8595, CG8595, CT24947, Dmel\\CG8595, TOLL 7, Tl-7, dTLR7, dToll7, toll', description='CG8595 gene product from transcript CG8595-RA', other_designations='CG8595-PA|Toll-7-PA|Toll-like|toll receptor', map_location='56E4-56E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19826578, end_position_on_the_genomic_accession=19832973, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37273, CurrentID=0, Status='live', Symbol='CG13872', Aliases='Dmel_ Dmel\\CG13872', description='CG13872 gene product from transcript CG13872-RA', other_designations='CG13872-PA|uncharacterized protein', map_location='56E4-56E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19878836, end_position_on_the_genomic_accession=19880920, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37274, CurrentID=0, Status='live', Symbol='CG10822', Aliases='Dmel_ Dmel\\CG10822', description='CG10822 gene product from transcript CG10822-RA', other_designations='CG10822-PA|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19951345, end_position_on_the_genomic_accession=19951950, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37275, CurrentID=0, Status='live', Symbol='CG8654', Aliases='Dmel_ Dmel\\CG8654', description='CG8654 gene product from transcript CG8654-RC', other_designations='CG8654-PA|CG8654-PB|CG8654-PC|CG8654-PD', map_location='56F4-56F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19979242, end_position_on_the_genomic_accession=19984895, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37276, CurrentID=0, Status='live', Symbol='CG16898', Aliases='Dmel_ Dmel\\CG16898', description='CG16898 gene product from transcript CG16898-RA', other_designations='CG16898-PA|uncharacterized protein', map_location='56F4-56F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20003620, end_position_on_the_genomic_accession=20005003, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37277, CurrentID=0, Status='live', Symbol='18w', Aliases='Dmel_CG8896, 18-w, 18-wheeler, 18W, CG8896, CT25100, Dmel\\CG8896, Toll-2, l(2)00053, tlr, toll', description='18 wheeler', other_designations='18-wheeler|18Wheeler|18w-PA|CG8896-PA|Toll-2|toll like receptor|toll receptor|wheeler', map_location='56F8-56F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20111511, end_position_on_the_genomic_accession=20116932, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37278, CurrentID=0, Status='live', Symbol='CG11041', Aliases='Dmel_ Dmel\\CG11041', description='CG11041 gene product from transcript CG11041-RB', other_designations='CG11041-PB|uncharacterized protein', map_location='56F8-56F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20146118, end_position_on_the_genomic_accession=20146824, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37280, CurrentID=0, Status='live', Symbol='CG16894', Aliases='Dmel_ Dmel\\CG16894', description='CG16894 gene product from transcript CG16894-RA', other_designations='CG16894-PA|uncharacterized protein', map_location='56F9-56F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20211585, end_position_on_the_genomic_accession=20212619, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37281, CurrentID=0, Status='live', Symbol='CG11044', Aliases='Dmel_ Dmel\\CG11044, FBXO10', description='CG11044 gene product from transcript CG11044-RA', other_designations='CG11044-PA|uncharacterized protein', map_location='56F10-56F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20231829, end_position_on_the_genomic_accession=20234927, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37282, CurrentID=0, Status='live', Symbol='CG11099', Aliases='Dmel_ CT31033, Dmel\\CG11099', description='CG11099 gene product from transcript CG11099-RA', other_designations='CG11099-PA|CG11099-PB|uncharacterized protein', map_location='56F10-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20236235, end_position_on_the_genomic_accession=20239200, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37283, CurrentID=0, Status='live', Symbol='CG13869', Aliases='Dmel_ Dmel\\CG13869', description='CG13869 gene product from transcript CG13869-RA', other_designations='CG13869-PA|uncharacterized protein', map_location='56F10-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20238495, end_position_on_the_genomic_accession=20239429, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37284, CurrentID=0, Status='live', Symbol='Efhc1.2', Aliases='Dmel_CG11048, CG11048, Dmel\\CG11048', description='EFHC1 homologue 2', other_designations='CG11048-PA|Defhc1.2|EFHC1 homolog|Efhc1.2-PA', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20239936, end_position_on_the_genomic_accession=20242873, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37285, CurrentID=0, Status='live', Symbol='CG11208', Aliases='Dmel_ Dmel\\CG11208, anon-WO0118547.196', description='CG11208 gene product from transcript CG11208-RA', other_designations='CG11208-PA|uncharacterized protein', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20242752, end_position_on_the_genomic_accession=20245388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37286, CurrentID=0, Status='live', Symbol='ppk6', Aliases='Dmel_CG11209, CG11209, Dmel\\CG11209, PPK6', description='pickpocket 6', other_designations='CG11209-PA|ppk6-PA', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20245760, end_position_on_the_genomic_accession=20247624, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37287, CurrentID=0, Status='live', Symbol='CG9864', Aliases='Dmel_ CT27832, Dmel\\CG9864', description='CG9864 gene product from transcript CG9864-RB', other_designations='CG9864-PA|CG9864-PB', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20248130, end_position_on_the_genomic_accession=20254670, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37288, CurrentID=0, Status='live', Symbol='Ate1', Aliases='Dmel_CG9204, ATE1, Ate, CG9204, Dm- Dmel\\CG9204, ate1, l(2)k10809', description='CG9204 gene product from transcript CG9204-RA', other_designations='Ate1-PA|Ate1-PB|Ate1-PC|Ate1-PD|CG9204-PA|CG9204-PB|CG9204-PC|CG9204-PD', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20254966, end_position_on_the_genomic_accession=20257390, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37289, CurrentID=0, Status='live', Symbol='Hsl', Aliases='Dmel_CG11055, A, CG11055, Dmel\\CG11055, anon-56Ea, dHSL', description='Hormone-sensitive lipase ortholog', other_designations='CG11055-PA|CG11055-PB|CG11055-PC|Hormone-sensitive lipase|Hsl-PA|Hsl-PB|Hsl-PC|hormone sensitive lipase|transcription unit A', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20257650, end_position_on_the_genomic_accession=20261726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37290, CurrentID=0, Status='live', Symbol='PCNA', Aliases='Dmel_CG9193, 53/13, CG9193, DmPCNA, Dmel\\CG9193, MUS209, Mus209/mus209, Pcna, dPCNA, l(2)02448, l(2)56F, l(2)56Fa, l(2)s1534, mus 209, mus209, pcna, PCNA', description='Proliferating cell nuclear antigen', other_designations='CG9193-PA|CG9193-PB|PCNA-PA|PCNA-PB|Proliferating-cell-nuclear-antigen|mutagen-sensitive 209|proliferating cell nuclear antigen|proliferating-cell nuclear antigen', map_location='2-92.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20261754, end_position_on_the_genomic_accession=20262883, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37292, CurrentID=0, Status='live', Symbol='RpS18', Aliases='Dmel_CG8900, B, BcDNA:RH43343, CG8900, Dmel\\CG8900, M(2)56F, RS18_DROME, Rps18, S18, anon-56Fa, anon-EST:Posey27, anon-EST:fe2D3, rpS18', description='Ribosomal protein S18', other_designations='CG8900-PA|CG8900-PB|RpS18-PA|RpS18-PB|anon-fast-evolving-2D3|ribosomal protein S18|transcription unit B', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20263979, end_position_on_the_genomic_accession=20264848, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37293, CurrentID=0, Status='live', Symbol='CG8908', Aliases='Dmel_ Dmel\\CG8908', description='CG8908 gene product from transcript CG8908-RE', other_designations='CG8908-PE', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20264920, end_position_on_the_genomic_accession=20270994, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37294, CurrentID=0, Status='live', Symbol='CG10444', Aliases='Dmel_ Dmel\\CG10444, dSLC5A3', description='CG10444 gene product from transcript CG10444-RA', other_designations='CG10444-PA|uncharacterized protein', map_location='56F11-56F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20271653, end_position_on_the_genomic_accession=20274499, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37295, CurrentID=0, Status='live', Symbol='CG11788', Aliases='Dmel_ Dmel\\CG11788', description='CG11788 gene product from transcript CG11788-RA', other_designations='CG11788-PA|uncharacterized protein', map_location='56F11-56F14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20274597, end_position_on_the_genomic_accession=20276142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37296, CurrentID=0, Status='live', Symbol='CG9143', Aliases='Dmel_ Dmel\\CG9143, cg9143', description='CG9143 gene product from transcript CG9143-RA', other_designations='CG9143-PA', map_location='56F14-56F15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20275973, end_position_on_the_genomic_accession=20278915, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37297, CurrentID=0, Status='live', Symbol='CG9090', Aliases='Dmel_ Dmel\\CG9090', description='CG9090 gene product from transcript CG9090-RA', other_designations='CG9090-PA|CG9090-PB|CG9090-PC', map_location='56F15-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20279066, end_position_on_the_genomic_accession=20281127, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37298, CurrentID=0, Status='live', Symbol='CG16868', Aliases='Dmel_ Dmel\\CG16868', description='CG16868 gene product from transcript CG16868-RA', other_designations='CG16868-PA|uncharacterized protein', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20281629, end_position_on_the_genomic_accession=20287094, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37299, CurrentID=0, Status='live', Symbol='Cpr56F', Aliases='Dmel_CG9036, CG9036, DmelCpr56F, Dmel\\CG9036', description='Cuticular protein 56F', other_designations='CG9036-PA|Cpr56F-PA|cuticular protein 56F', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20288371, end_position_on_the_genomic_accession=20291945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37300, CurrentID=0, Status='live', Symbol='CkIIbeta2', Aliases="Dmel_CG8914, CG8914, CK2, CK2beta', CKIIb2, CKIIbeta', CKIIbeta2, DmCK2beta', DmCKIIbeta', Dmel\\CG8914, dCK2beta'", description='Casein kinase II beta2 subunit', other_designations='CG8914-PA|CkIIbeta2-PA|casein kinase II beta2 subunit', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20289545, end_position_on_the_genomic_accession=20290550, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37301, CurrentID=0, Status='live', Symbol='CG11200', Aliases='Dmel_CG11200, Dmel\\CG11200', description='Carbonyl reductase', other_designations='CG11200-PA|CG11200-PB|CG11200-PC|carbonyl reductase', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20304517, end_position_on_the_genomic_accession=20308671, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37302, CurrentID=0, Status='live', Symbol='CG13868', Aliases='Dmel_ Dmel\\CG13868, anon-WO0153538.60', description='CG13868 gene product from transcript CG13868-RB', other_designations='CG13868-PA|CG13868-PB|uncharacterized protein', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20308777, end_position_on_the_genomic_accession=20316917, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37303, CurrentID=0, Status='live', Symbol='CG13871', Aliases='Dmel_ Dmel\\CG13871', description='CG13871 gene product from transcript CG13871-RA', other_designations='CG13871-PA|uncharacterized protein', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20317323, end_position_on_the_genomic_accession=20319039, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37304, CurrentID=0, Status='live', Symbol='CG8920', Aliases='Dmel_ BcDNA:LD21403, Dmel\\CG8920', description='CG8920 gene product from transcript CG8920-RB', other_designations='CG8920-PA|CG8920-PB|CG8920-PD|CG8920-PF', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20319180, end_position_on_the_genomic_accession=20325973, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37305, CurrentID=0, Status='live', Symbol='MED8', Aliases='Dmel_CG13867, ARC32, Arc32, CG13867, CG30140, Dmel\\CG13867, Med8/ARC32, dARC32', description='Mediator complex subunit 8', other_designations='CG13867-PA|MED8-PA|mediator complex subunit 8', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20325844, end_position_on_the_genomic_accession=20326806, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37306, CurrentID=0, Status='live', Symbol='CG8929', Aliases='Dmel_ Dmel\\CG8929, anon-WO0118547.91', description='CG8929 gene product from transcript CG8929-RA', other_designations='CG8929-PA|CG8929-PB|CG8929-PC|CG8929-PD', map_location='56F16-56F16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20327270, end_position_on_the_genomic_accession=20330721, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37307, CurrentID=0, Status='live', Symbol='CG16739', Aliases='Dmel_ Dmel\\CG16739', description='CG16739 gene product from transcript CG16739-RA', other_designations='CG16739-PA|uncharacterized protein', map_location='56F17-56F17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20341993, end_position_on_the_genomic_accession=20342749, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37308, CurrentID=0, Status='live', Symbol='CG13870', Aliases='Dmel_ Dmel\\CG13870', description='CG13870 gene product from transcript CG13870-RB', other_designations='CG13870-PA|CG13870-PB|uncharacterized protein', map_location='56F17-56F17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20343845, end_position_on_the_genomic_accession=20346018, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37309, CurrentID=0, Status='live', Symbol='CG11192', Aliases='Dmel_ Dmel\\CG11192, SP82', description='CG11192 gene product from transcript CG11192-RB', other_designations='CG11192-PB|uncharacterized protein', map_location='56F17-56F17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20354108, end_position_on_the_genomic_accession=20355067, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37310, CurrentID=0, Status='live', Symbol='CG12484', Aliases='Dmel_ 12484, CG30143, CG33041, CT32777, Dmel\\CG12484, Dmel_CG30143, Dmel_CG33041', description='CG12484 gene product from transcript CG12484-RB', other_designations='CG12484-PA|CG12484-PB|CG12484-PD|CG12484-PE|uncharacterized protein', map_location='57A3-57A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20424338, end_position_on_the_genomic_accession=20478951, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  37311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37311, CurrentID=0, Status='live', Symbol='Obp57c', Aliases='Dmel_CG13421, 57c, BcDNA:RH68952, CG13421, DmelObp57c, Dmel\\CG13421, Obp57a, obp57c', description='Odorant-binding protein 57c', other_designations='CG13421-PA|Obp57c-PA', map_location='57A4-57A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20503584, end_position_on_the_genomic_accession=20504202, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37312, CurrentID=0, Status='live', Symbol='CG13426', Aliases='Dmel_ Dmel\\CG13426', description='CG13426 gene product from transcript CG13426-RA', other_designations='CG13426-PA|uncharacterized protein', map_location='57A4-57A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20512697, end_position_on_the_genomic_accession=20513169, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37313, CurrentID=0, Status='live', Symbol='GNBP-like3', Aliases='Dmel_CG13422, CG13422, Dmel\\CG13422, GNBP-like, GNBPlike 3', description='GNBP-like 3', other_designations='CG13422-PA|GNBP-like|GNBP-like3-PA|Gram-negative binding protein-like protein', map_location='57A4-57A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20526311, end_position_on_the_genomic_accession=20526825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37314, CurrentID=0, Status='live', Symbol='CG18067', Aliases='Dmel_ CT40493, Dmel\\CG18067', description='CG18067 gene product from transcript CG18067-RB', other_designations='CG18067-PA|CG18067-PB|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20534245, end_position_on_the_genomic_accession=20535835, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37315, CurrentID=0, Status='live', Symbol='CG13423', Aliases='Dmel_ Dmel\\CG13423', description='CG13423 gene product from transcript CG13423-RA', other_designations='CG13423-PA|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20536578, end_position_on_the_genomic_accession=20538264, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37316, CurrentID=0, Status='live', Symbol='CG13427', Aliases='Dmel_ Dmel\\CG13427', description='CG13427 gene product from transcript CG13427-RA', other_designations='CG13427-PA|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20538877, end_position_on_the_genomic_accession=20539316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37317, CurrentID=0, Status='live', Symbol='CG13428', Aliases='Dmel_ Dmel\\CG13428', description='CG13428 gene product from transcript CG13428-RB', other_designations='CG13428-PA|CG13428-PB|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20539512, end_position_on_the_genomic_accession=20540448, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37318, CurrentID=0, Status='live', Symbol='CG30148', Aliases='Dmel_ BcDNA:AT18611, CG13429, Dmel\\CG30148', description='CG30148 gene product from transcript CG30148-RB', other_designations='CG30148-PA|CG30148-PB|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20546595, end_position_on_the_genomic_accession=20547658, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37319, CurrentID=0, Status='live', Symbol='Cpr57A', Aliases='Dmel_CG18066, CG18066, Dmel\\CG18066', description='Cuticular protein 57A', other_designations='CG18066-PA|CG18066-PB|Cpr57A-PA|Cpr57A-PB|cuticular protein 57A', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20548368, end_position_on_the_genomic_accession=20550117, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37322, CurrentID=0, Status='live', Symbol='lms', Aliases='Dmel_CG13424, CG13424, Dmel\\CG13424', description='lateral muscles scarcer', other_designations='CG13424-PB|CG13424-PC|lms-PB|lms-PC', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20556877, end_position_on_the_genomic_accession=20559953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37323, CurrentID=0, Status='live', Symbol='Mgat1', Aliases='Dmel_CG13431, CG13431, Dmel\\CG13431, GlcNAc-TI, GlcNAcT1, MGAT1, dMGAT1', description='Mannosyl (alpha-1,3-)-glycoprotein beta-1,2-N-acetylglucosaminyltransferase 1', other_designations='CG13431-PA|Mgat1-PA|N-acetylglucosaminyltransferase I|UDP-GlcNAc:a-3-D-mannoside-beta-1,2-N-acetylglucosaminyltransferase I|mannosyl (alpha-1,3-)-glycoprotein beta-1,2-N-acetylglucosaminyltransferase 1', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20559156, end_position_on_the_genomic_accession=20562331, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37325, CurrentID=0, Status='live', Symbol='Nnf1a', Aliases='Dmel_CG13434, CG13434, Dmel\\CG13434, dmNnf1R-2', description='CG13434 gene product from transcript CG13434-RA', other_designations='CG13434-PA|Nnf1-related protein 2|Nnf1a-PA', map_location='57A6-57A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20586222, end_position_on_the_genomic_accession=20587238, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37328, CurrentID=0, Status='live', Symbol='Hil', Aliases='Dmel_CG30147, CG10439, CG13435, CG30147, D-hil, Dmel\\CG30147', description='Hillarin', other_designations='CG30147-PA|CG30147-PB|D-hillarin|Hil-PA|Hil-PB|hillarin', map_location='57A7-57A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20606881, end_position_on_the_genomic_accession=20618217, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37329, CurrentID=0, Status='live', Symbol='CG9945', Aliases='Dmel_ Dmel\\CG9945', description='CG9945 gene product from transcript CG9945-RA', other_designations='CG9945-PA|CG9945-PB', map_location='57A7-57A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20618503, end_position_on_the_genomic_accession=20621081, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37330, CurrentID=0, Status='live', Symbol='CG11180', Aliases='Dmel_ Dmel\\CG11180', description='CG11180 gene product from transcript CG11180-RB', other_designations='CG11180-PA|CG11180-PB|uncharacterized protein', map_location='57A7-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20620914, end_position_on_the_genomic_accession=20623429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37331, CurrentID=0, Status='live', Symbol='FAM21', Aliases='Dmel_CG16742, CG16742, Dmel\\CG16742', description='CG16742 gene product from transcript CG16742-RA', other_designations='CG16742-PA|FAM21-PA', map_location='57A8-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20623547, end_position_on_the_genomic_accession=20628315, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37332, CurrentID=0, Status='live', Symbol='Rcd6', Aliases='Dmel_CG11175, CG11175, Dmel\\CG11175', description='Reduction in Cnn dots 6', other_designations='CG11175-PA|CG11175-PB|Rcd6-PA|Rcd6-PB|reduction in Cnn dots 6', map_location='57A8-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20628186, end_position_on_the_genomic_accession=20633506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740622, CurrentID=0, Status='live', Symbol='CG34426', Aliases='Dmel_ CG13307, Dmel\\CG34426, Dmel_CG13307', description='CG34426 gene product from transcript CG34426-RA', other_designations='CG34426-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8802463, end_position_on_the_genomic_accession=8803599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37334, CurrentID=0, Status='live', Symbol='CG13436', Aliases='Dmel_ Dmel\\CG13436', description='CG13436 gene product from transcript CG13436-RA', other_designations='CG13436-PA|uncharacterized protein', map_location='57A8-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20636392, end_position_on_the_genomic_accession=20638423, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37335, CurrentID=0, Status='live', Symbol='rig', Aliases='Dmel_CG30149, CG11171, CG13440, CG30149, Dmel\\CG30149, Gem5, Rig, anon-SAGE:Wang-132, l(2)05056', description='rigor mortis', other_designations='CG30149-PB|Gemin5|anon-SAGE:Wang-132|rig-PB', map_location='57A8-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20633876, end_position_on_the_genomic_accession=20640681, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37336, CurrentID=0, Status='live', Symbol='maf-S', Aliases='Dmel_CG9954, CG9954, DM19, DM31, Dmel\\CG9954, Maf, Maf-S, MafS, anon-EST:Posey173, maf2', description='CG9954 gene product from transcript CG9954-RB', other_designations='CG9954-PA|CG9954-PB|maf-S-PA|maf-S-PB', map_location='57A8-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20640976, end_position_on_the_genomic_accession=20641668, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37337, CurrentID=0, Status='live', Symbol='CG11110', Aliases='Dmel_ Dmel\\CG11110', description='CG11110 gene product from transcript CG11110-RA', other_designations='CG11110-PA|serine-type peptidase|uncharacterized protein', map_location='57A8-57A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20641742, end_position_on_the_genomic_accession=20642558, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37339, CurrentID=0, Status='live', Symbol='DMAP1', Aliases='Dmel_CG11132, CG11132, DMA, Dmel\\CG11132, Q7K3D8_DROME, dDMAP1', description='DNA methyltransferase 1 associated protein 1', other_designations='CG11132-PA|DMAP1-PA', map_location='57A9-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20644036, end_position_on_the_genomic_accession=20645865, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37340, CurrentID=0, Status='live', Symbol='CG16799', Aliases='Dmel_ Dmel\\CG16799', description='CG16799 gene product from transcript CG16799-RB', other_designations='CG16799-PA|CG16799-PB|uncharacterized protein', map_location='57A9-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20652557, end_position_on_the_genomic_accession=20655096, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37341, CurrentID=0, Status='live', Symbol='CG11159', Aliases='Dmel_ Dmel\\CG11159', description='CG11159 gene product from transcript CG11159-RA', other_designations='CG11159-PA|uncharacterized protein', map_location='57A9-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20655756, end_position_on_the_genomic_accession=20656609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37342, CurrentID=0, Status='live', Symbol='Lrt', Aliases='Dmel_CG11136, CG11136, CT31131, Dmel\\CG11136, lrt, tartan/capricious-like', description='Leucine-rich tendon-specific protein', other_designations='CG11136-PB|CG11136-PC|Lrt-PB|Lrt-PC', map_location='57A9-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20646360, end_position_on_the_genomic_accession=20661070, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37343, CurrentID=0, Status='live', Symbol='CG13437', Aliases='Dmel_ Dmel\\CG13437', description='CG13437 gene product from transcript CG13437-RA', other_designations='CG13437-PA|uncharacterized protein', map_location='57A9-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20661585, end_position_on_the_genomic_accession=20662595, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37344, CurrentID=0, Status='live', Symbol='Fem-1', Aliases='Dmel_CG9025, CG9025, Dmel\\CG9025, dFEM-1', description='CG9025 gene product from transcript CG9025-RA', other_designations='CG9025-PA|CG9025-PB|Fem-1-PA|Fem-1-PB', map_location='57A9-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20662474, end_position_on_the_genomic_accession=20667306, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37345, CurrentID=0, Status='live', Symbol='exu', Aliases='Dmel_CG8994, CG8994, Dmel\\CG8994, Exu, bs30h09.y11, exu', description='exuperantia', other_designations='CG8994-PA|CG8994-PB|CG8994-PC|CG8994-PD|exu-PA|exu-PB|exu-PC|exu-PD', map_location='2-93 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20667427, end_position_on_the_genomic_accession=20671083, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37347, CurrentID=0, Status='live', Symbol='Gr57a', Aliases='Dmel_CG13441, 57B.1, CG13441, Dmel\\CG13441, GR57B.1, GR57a, Gr57B1', description='Gustatory receptor 57a', other_designations='CG13441-PA|Gr57a-PA|gustatory receptor 57B.1|gustatory receptor 57a', map_location='57B1-57B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20685605, end_position_on_the_genomic_accession=20686910, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37348, CurrentID=0, Status='live', Symbol='CG13438', Aliases='Dmel_ Dmel\\CG13438', description='CG13438 gene product from transcript CG13438-RA', other_designations='CG13438-PA|uncharacterized protein', map_location='57B1-57B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20692829, end_position_on_the_genomic_accession=20693739, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37349, CurrentID=0, Status='live', Symbol='CG13442', Aliases='Dmel_ Dmel\\CG13442', description='CG13442 gene product from transcript CG13442-RA', other_designations='CG13442-PA|uncharacterized protein', map_location='57B1-57B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20727829, end_position_on_the_genomic_accession=20730610, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37351, CurrentID=0, Status='live', Symbol='CG13443', Aliases='Dmel_ Dmel\\CG13443', description='CG13443 gene product from transcript CG13443-RA', other_designations='CG13443-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20773687, end_position_on_the_genomic_accession=20775364, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37354, CurrentID=0, Status='live', Symbol='CG15225', Aliases='Dmel_ Dmel\\CG15225', description='CG15225 gene product from transcript CG15225-RB', other_designations='CG15225-PB|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20792982, end_position_on_the_genomic_accession=20793795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37355, CurrentID=0, Status='live', Symbol='insc', Aliases='Dmel_CG11312, 25/17, CG11312, Dmel\\CG11312, INSC, Ins, Insc, fam, l(2)05475, l(2)k12405, nem', description='inscuteable', other_designations='CG11312-PA|fata morgana|insc-PA|inscrutable|inscutable|not enough muscles', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20821262, end_position_on_the_genomic_accession=20836028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37356, CurrentID=0, Status='live', Symbol='sktl', Aliases='Dmel_CG9985, 25/17, CG9985, Dmel\\CG9985, PIP5-kinase, PIP5K, PIP5K 57B6, Sktl, fam, l(2)05475, l(2)k12405, ski', description='skittles', other_designations='CG9985-PA|CG9985-PB|CG9985-PC|PIP5-kinase|fata morgana|phosphatidylinositol 4-phosphate 5-kinase|sktl-PA|sktl-PB|sktl-PC|type I phosphatidylinositol 4-phosphate 5-kinase', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20827282, end_position_on_the_genomic_accession=20836028, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37357, CurrentID=0, Status='live', Symbol='CG17999', Aliases='Dmel_ Dmel\\CG17999', description='CG17999 gene product from transcript CG17999-RA', other_designations='CG17999-PA|uncharacterized protein', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20847980, end_position_on_the_genomic_accession=20849939, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37358, CurrentID=0, Status='live', Symbol='CG9993', Aliases='Dmel_ Dmel\\CG9993', description='CG9993 gene product from transcript CG9993-RA', other_designations='CG9993-PA', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20850146, end_position_on_the_genomic_accession=20852125, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37359, CurrentID=0, Status='live', Symbol='CG15227', Aliases='Dmel_ Dmel\\CG15227', description='CG15227 gene product from transcript CG15227-RB', other_designations='CG15227-PB|uncharacterized protein', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20863781, end_position_on_the_genomic_accession=20867076, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37360, CurrentID=0, Status='live', Symbol='Ipk1', Aliases='Dmel_CG30295, CG15228, CG15229, CG15230, CG30295, Dmel\\CG30295, dm', description='CG30295 gene product from transcript CG30295-RB', other_designations='CG30295-PA|CG30295-PB|CG30295-PC|CG30295-PD|Ipk1-PA|Ipk1-PB|Ipk1-PC|Ipk1-PD', map_location='57B3-57B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20870791, end_position_on_the_genomic_accession=20879145, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  37363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37363, CurrentID=0, Status='live', Symbol='Cib2', Aliases='Dmel_CG9236, CG9236, Dmel\\CG9236, cib2', description='CIB2 ortholog', other_designations='CG9236-PB|Cib2-PB', map_location='57B4-57B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20879930, end_position_on_the_genomic_accession=20882663, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37364, CurrentID=0, Status='live', Symbol='otp', Aliases='Dmel_CG10036, BcDNA:RE58095, CG10036, CG2965, Dmel\\CG10036, Otp, W26, bk24, w26', description='orthopedia', other_designations='CG10036-PC|CG10036-PE|CG10036-PF|CG10036-PG|CG10036-PH|CG10036-PI|CG10036-PJ|CG10036-PK|otp-PC|otp-PE|otp-PF|otp-PG|otp-PH|otp-PI|otp-PJ|otp-PK', map_location='57B4-57B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20883625, end_position_on_the_genomic_accession=20903414, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37366, CurrentID=0, Status='live', Symbol='CG9235', Aliases='Dmel_ Dmel\\CG9235', description='CG9235 gene product from transcript CG9235-RA', other_designations='CG9235-PA', map_location='57B4-57B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20909704, end_position_on_the_genomic_accession=20910801, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37367, CurrentID=0, Status='live', Symbol='Rx', Aliases='Dmel_CG10052, CG10052, DRx, Dmel\\CG10052, Drx, E97, PPH17, W60, bk50, drx, wom', description='Retinal Homeobox', other_designations='CG10052-PB|Rx-PB|drosophila Rx|retinal homeobox|wombat', map_location='57B4-57B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20916753, end_position_on_the_genomic_accession=20937030, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37368, CurrentID=0, Status='live', Symbol='Act57B', Aliases='Dmel_CG10067, Act57, Act57A, Act57b, Actin, BEST:RE38067, C, CG10067, CG30294, Dmel\\CG10067, M, T10, act 57A, act57A, act57B, act57b, actin, anon-EST:CL2c12, chr2R:16452771..16452966', description='Actin 57B', other_designations='Act57B-PA|CG10067-PA|F-actin|actin|actin 57B|actin57B', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20944021, end_position_on_the_genomic_accession=20946431, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37371, CurrentID=0, Status='live', Symbol='CG15649', Aliases='Dmel_ BcDNA:GH08880, Dmel\\CG15649', description='CG15649 gene product from transcript CG15649-RB', other_designations='CG15649-PA|CG15649-PB|uncharacterized protein', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20972916, end_position_on_the_genomic_accession=20974017, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37372, CurrentID=0, Status='live', Symbol='CG9344', Aliases='Dmel_ BcDNA:RE43665, Dmel\\CG9344', description='CG9344 gene product from transcript CG9344-RA', other_designations='CG9344-PA', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20977056, end_position_on_the_genomic_accession=20977600, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37373, CurrentID=0, Status='live', Symbol='CG15650', Aliases='Dmel_ Dmel\\CG15650', description='CG15650 gene product from transcript CG15650-RB', other_designations='CG15650-PA|CG15650-PB|uncharacterized protein', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20977595, end_position_on_the_genomic_accession=20978818, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37374, CurrentID=0, Status='live', Symbol='CG9313', Aliases='Dmel_ Dmel\\CG9313', description='CG9313 gene product from transcript CG9313-RC', other_designations='CG9313-PA|CG9313-PB|CG9313-PC', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20978901, end_position_on_the_genomic_accession=20984043, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37375, CurrentID=0, Status='live', Symbol='CG15651', Aliases='Dmel_ Dmel\\CG15651, STT', description='CG15651 gene product from transcript CG15651-RA', other_designations='CG15651-PA|uncharacterized protein', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20983838, end_position_on_the_genomic_accession=20986500, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37376, CurrentID=0, Status='live', Symbol='CG3216', Aliases='Dmel_ Dmel\\CG3216, GYC', description='CG3216 gene product from transcript CG3216-RD', other_designations='CG3216-PC|CG3216-PD', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20986833, end_position_on_the_genomic_accession=20991171, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  37377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37377, CurrentID=0, Status='live', Symbol='dgt3', Aliases='Dmel_CG3221, CG3221, Dgt3, Dmel\\CG3221, dim', description='dim gamma-tubulin 3', other_designations='CG3221-PB|dgt3-PB', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20992276, end_position_on_the_genomic_accession=20994381, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37378, CurrentID=0, Status='live', Symbol='Prosalpha3', Aliases='Dmel_CG9327, Alpha-3, CG9327, Dmel\\CG9327, PROS-29, PSB3_DROME, Pros20S-alpha3, Pros29, alpha-3, alpha3, alpha3_dm, pros29', description='Proteasome alpha3 subunit', other_designations='20s proteasome|CG9327-PA|Prosalpha3-PA|alpha-subunit of 20S proteasome|alpha3|proteasome 20S alpha3 subunit|proteasome 29kD subunit|proteasome alpha3 subunit|proteasome subunit alpha-type 4', map_location='57B5-57B', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20994900, end_position_on_the_genomic_accession=20996118, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37379, CurrentID=0, Status='live', Symbol='CG10543', Aliases='Dmel_ Dmel\\CG10543', description='CG10543 gene product from transcript CG10543-RF', other_designations='CG10543-PA|CG10543-PD|CG10543-PE|CG10543-PF|uncharacterized protein', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20996220, end_position_on_the_genomic_accession=21007740, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37381, CurrentID=0, Status='live', Symbol='CG9346', Aliases='Dmel_ Dmel\\CG9346, cg9346', description='CG9346 gene product from transcript CG9346-RA', other_designations='CG9346-PA', map_location='57B9-57B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21011970, end_position_on_the_genomic_accession=21015628, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37382, CurrentID=0, Status='live', Symbol='CG3295', Aliases='Dmel_ Dmel\\CG3295', description='CG3295 gene product from transcript CG3295-RA', other_designations='CG3295-PA', map_location='57B9-57B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21015594, end_position_on_the_genomic_accession=21017446, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37384, CurrentID=0, Status='live', Symbol='RIC-3', Aliases='Dmel_CG30296, CG15652, CG30296, CG9349, Dm Dmel\\CG30296', description='CG30296 gene product from transcript CG30296-RE', other_designations='CG30296-PC|CG30296-PD|CG30296-PE|CG30296-PG|CG30296-PH|CG30296-PI|CG30296-PJ|CG30296-PK|CG30296-PL|CG30296-PM|CG30296-PN|CG30296-PO|CG30296-PP|CG30296-PQ|CG30296-PR|CG30296-PS|CG30296-PT|CG30296-PU|CG30296-PV|RIC-3-PC|RIC-3-PD|RIC-3-PE|RIC-3-PG|RIC-3-PH|RIC-3-PI|RIC-3-PJ|RIC-3-PK|RIC-3-PL|RIC-3-PM|RIC-3-PN|RIC-3-PO|RIC-3-PP|RIC-3-PQ|RIC-3-PR|RIC-3-PS|RIC-3-PT|RIC-3-PU|RIC-3-PV', map_location='57B9-57B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21017784, end_position_on_the_genomic_accession=21024918, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  37385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37385, CurrentID=0, Status='live', Symbol='ND-B14.7', Aliases='Dmel_CG9350, BcDNA:RH51452, CG9350, Dmel\\CG9350', description='NADH dehydrogenase (ubiquinone) B14.7 subunit', other_designations='CG9350-PA|NADH ubiquinone oxidoreductase B14.7|ND-B14.7-PA', map_location='57B12-57B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21025078, end_position_on_the_genomic_accession=21026086, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37386, CurrentID=0, Status='live', Symbol='shg', Aliases='Dmel_CG3722, CADH, CG3722, CT12481, Cad, CadE, Cadh, D E-cad, D-CAD2, D-cad, DCAD, DCAD2, DCad, DCad2, DE, DE Cad, DE Cadh, DE-CAD, DE-CAD2, DE-Cad, DE-Cad2, DE-Cadh, DE-Cadherin, DE-cad, DE-cadh, DE-cadherin, DEC, DECAD, DECad, DECadh, DE[cyto], DEcad, De-Cad, Dmel\\CG3722, E-CAD, E-Ca, E-Cad, E-Cadherin, E-cad, E-cadherin, ECAD, ECad, ECad2, ECadh, Ecad, N-cad, Shg, cad, cadh, com5, dEcad, de-cad, e-cad, gp150, l(2)10469, l(2)k03401, l(2)k10220/DE-Cad, shg', description='shotgun', other_designations='(DE)-cadherin|CG3722-PA|Cadherin-E|D-Ecadherin|D-cadherin|DCAD2|DE cadherin|DE-cad|DE-cadherin/Shotgun|DEcad|DEcadherin|De-cadherin|ECadherin|Shotgun/E-cadherin|cadherin|compaction defects 5|drosophila E-cadherin|drosophila epithelial cadherin|drosophila melanogaster epithelial (DE)-cadherin|e-cadherin|epithelial cadherin|eta-catenin|shg-PA', map_location='2-92 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21049654, end_position_on_the_genomic_accession=21057197, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37387, CurrentID=0, Status='live', Symbol='cpa', Aliases='Dmel_CG10540, CG10540, Cpa, Dmel\\CG10540, miro, scrd', description='capping protein alpha', other_designations='CG10540-PA|capping alpha|capping protein|capping proteinalpha|cpa-PA', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21057787, end_position_on_the_genomic_accession=21059883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37388, CurrentID=0, Status='live', Symbol='CG15653', Aliases='Dmel_ Dmel\\CG15653', description='CG15653 gene product from transcript CG15653-RA', other_designations='CG15653-PA|uncharacterized protein', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21059780, end_position_on_the_genomic_accession=21060547, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37389, CurrentID=0, Status='live', Symbol='mRpL54', Aliases='Dmel_CG9353, BcDNA:SD09147, CG9353, Dmel\\CG9353, MRP-L54', description='mitochondrial ribosomal protein L54', other_designations='CG9353-PA|mRpL54-PA', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21060789, end_position_on_the_genomic_accession=21061587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37390, CurrentID=0, Status='live', Symbol='Cht8', Aliases='Dmel_CG9357, CG9357, Dmel\\CG9357', description='CG9357 gene product from transcript CG9357-RA', other_designations='CG9357-PA|Cht8-PA', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21061646, end_position_on_the_genomic_accession=21063212, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37392, CurrentID=0, Status='live', Symbol='Cht9', Aliases='Dmel_CG10531, CG10531, DmCHT9, Dmel\\CG10531', description='CG10531 gene product from transcript CG10531-RA', other_designations='CG10531-PA|Cht9-PA', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21067992, end_position_on_the_genomic_accession=21069286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37393, CurrentID=0, Status='live', Symbol='CG10527', Aliases='Dmel_ Dmel\\CG10527, FAMeT', description='CG10527 gene product from transcript CG10527-RA', other_designations='CG10527-PA|farnesoic acid O-methyltransferase|uncharacterized protein', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21070368, end_position_on_the_genomic_accession=21072789, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37395, CurrentID=0, Status='live', Symbol='Rbpn-5', Aliases='Dmel_CG4030, CG4030, Dmel\\CG4030', description='Rabaptin-5', other_designations='CG4030-PA|Rabaptin5|Rbpn-5-PA', map_location='57B19-57B19', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21088386, end_position_on_the_genomic_accession=21091004, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37397, CurrentID=0, Status='live', Symbol='CG4038', Aliases='Dmel_CG4038-PA, Dmel\\CG4038, GCR 101, GCR101, GCR101 SnRNP, Pen101, anon-Pen101, dGAR1', description='CG4038 gene product from transcript CG4038-RA', other_designations='CG4038-PA', map_location='57B20-57B20', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21095958, end_position_on_the_genomic_accession=21097305, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37398, CurrentID=0, Status='live', Symbol='CG34396', Aliases='Dmel_ CG15654, CG15655, Dmel\\CG34396, Dmel_CG15654, Dmel_CG15655, anon-WO0109301.1, anon-WO0109301.16, anon-WO0109301.17', description='CG34396 gene product from transcript CG34396-RC', other_designations='CG34396-PC|CG34396-PD|uncharacterized protein', map_location='57B20-57B20', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21092222, end_position_on_the_genomic_accession=21101453, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39001, CurrentID=0, Status='live', Symbol='Cp16', Aliases='Dmel_CG6533, CG6533, CP16, Dmel\\CG6533, Dms16, Ov31, S16, X16715, cp16, melcp16, s16, s16-1', description='Chorion protein 16', other_designations='CG6533-PA|Cp16-PA|chorion prot 16|chorion protein 16', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8731944, end_position_on_the_genomic_accession=8732599, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37400, CurrentID=0, Status='live', Symbol='ktub', Aliases='Dmel_CG9398, 87D3S, CG9398, D-tulp, Dmel\\CG9398, Dtulp, ESTS:87D3S, King-tubby, TULP, Tulp, king-tubby, tulp', description='king tubby', other_designations='CG9398-PB|CG9398-PC|Tubby-like|king-tubby|ktub-PB|ktub-PC|tubby-like protein', map_location='57B20-57C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21101928, end_position_on_the_genomic_accession=21123704, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37401, CurrentID=0, Status='live', Symbol='CG4050', Aliases='Dmel_ Dmel\\CG4050, anon-DM192, anon-j', description='CG4050 gene product from transcript CG4050-RA', other_designations='CG4050-PA|CG4050-PB', map_location='57C2-57C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21123735, end_position_on_the_genomic_accession=21128510, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37402, CurrentID=0, Status='live', Symbol='mago', Aliases='Dmel_CG9401, CG9401, Dm MGN, DmMago, Dmel\\CG9401, MAGOH, MGN, Mago, Magoh, l(2)57Ca, mgn', description='mago nashi', other_designations='CG9401-PA|mago nishi|mago-PA|mago-nashi', map_location='57C2-57C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21128577, end_position_on_the_genomic_accession=21129737, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37403, CurrentID=0, Status='live', Symbol='Magi', Aliases='Dmel_CG30388, CG15656, CG30388, CG4117, D-MAGI, Dmel\\CG30388, Ima, ima, magi', description='CG30388 gene product from transcript CG30388-RA', other_designations='CG30388-PA|CG30388-PB|CG30388-PC|Magi-PA|Magi-PB|Magi-PC|magi', map_location='57C2-57C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21129463, end_position_on_the_genomic_accession=21142261, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37405, CurrentID=0, Status='live', Symbol='CG9406', Aliases='Dmel_ Dmel\\CG9406', description='CG9406 gene product from transcript CG9406-RA', other_designations='CG9406-PA', map_location='57C3-57C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21142368, end_position_on_the_genomic_accession=21143261, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37407, CurrentID=0, Status='live', Symbol='Hmg-2', Aliases='Dmel_CG9418, BEST:LD07122, CG9418, Dmel\\CG9418, LD07122', description='High mobility group protein 2', other_designations='CG9418-PA|Hmg-2-PA|high mobility group protein 1/2|high mobility group protein 2', map_location='57C4-57C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21146203, end_position_on_the_genomic_accession=21147608, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37408, CurrentID=0, Status='live', Symbol='CG15657', Aliases='Dmel_ BcDNA:AT24036, Dmel\\CG15657', description='CG15657 gene product from transcript CG15657-RA', other_designations='CG15657-PA|CG15657-PB|uncharacterized protein', map_location='57C4-57C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21147478, end_position_on_the_genomic_accession=21148153, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37410, CurrentID=0, Status='live', Symbol='CG30389', Aliases='Dmel_ BcDNA:AT02070, CG10518, CG4126, Dmel\\CG30389', description='CG30389 gene product from transcript CG30389-RD', other_designations='CG30389-PA|CG30389-PC|CG30389-PD|CG30389-PE|CG30389-PF|uncharacterized protein', map_location='57C4-57C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21148265, end_position_on_the_genomic_accession=21160545, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  37411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37411, CurrentID=0, Status='live', Symbol='CG4266', Aliases='Dmel_ Dmel\\CG4266, cg4266', description='CG4266 gene product from transcript CG4266-RD', other_designations='CG4266-PA|CG4266-PB|CG4266-PC|CG4266-PD', map_location='57C6-57C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21161713, end_position_on_the_genomic_accession=21168659, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  37412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37412, CurrentID=0, Status='live', Symbol='hng1', Aliases='Dmel_CG9437, CG9437, Dmel\\CG9437', description='hinge1', other_designations='CG9437-PA|hng1-PA', map_location='57C7-57C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21168930, end_position_on_the_genomic_accession=21170010, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37413, CurrentID=0, Status='live', Symbol='LSm1', Aliases='Dmel_CG4279, CG4279, Dmel\\CG4279, Lsm1', description='CG4279 gene product from transcript CG4279-RA', other_designations='CG4279-PA|LSm1-PA', map_location='57C7-57C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21169969, end_position_on_the_genomic_accession=21170672, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37414, CurrentID=0, Status='live', Symbol='Xpd', Aliases='Dmel_CG9433, CG9433, DhR3, DhXPD, DmXPD, Dmel\\CG9433, ERCC2, XPD, XPD/ERCC2, l(2)SH2 2137, l(2)SH2137, xpd', description='Xeroderma pigmentosum D ortholog (H. sapiens)', other_designations='CG9433-PA|Xpd-PA|lethal (2) SH2137|xeroderma pigmentosum D|xeroderma pigmentosum D ortholog', map_location='57C7-57C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21171001, end_position_on_the_genomic_accession=21174504, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37415, CurrentID=0, Status='live', Symbol='Pu', Aliases='Dmel_CG9441, CG9441, DGTPCH, Dmel\\CG9441, GCH1, GTPC, GTPCH, Gr, l(2)57Cf, p[Gr], upi', description='Punch', other_designations='CG9441-PA|CG9441-PB|CG9441-PC|GTP cyclohydrolase|GTP cyclohydrolase I|GTP cyclohydrolase I isoform c|Pu-PA|Pu-PB|Pu-PC|grape|punch|unpigmented', map_location='2-97 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21175315, end_position_on_the_genomic_accession=21182608, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37416, CurrentID=0, Status='live', Symbol='CG4286', Aliases='Dmel_ Dmel\\CG4286', description='CG4286 gene product from transcript CG4286-RA', other_designations='CG4286-PA', map_location='57C7-57C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21177754, end_position_on_the_genomic_accession=21178636, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37417, CurrentID=0, Status='live', Symbol='tud', Aliases='Dmel_CG9450, CG9450, Dmel\\CG9450, TUD, Tud, tu', description='tudor', other_designations='CG9450-PA|CG9450-PB|TUDOR|tud-PA|tud-PB', map_location='57C8-57C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21183322, end_position_on_the_genomic_accession=21192913, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37418, CurrentID=0, Status='live', Symbol='Lapsyn', Aliases='Dmel_CG15658, CG15658, CT35842, Dmel\\CG15658', description='Leucine-rich repeat activity-regulated protein at synapses', other_designations='CG15658-PA|Lapsyn-PA', map_location='57C9-57D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21193571, end_position_on_the_genomic_accession=21199431, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37419, CurrentID=0, Status='live', Symbol='Glycogenin', Aliases='Dmel_CG44244, CG44244, CG9480, Dmel\\CG44244, Dmel_CG9480, NEST:bs16a02', description='CG44244 gene product from transcript CG44244-RE', other_designations='CG44244-PB|CG44244-PD|CG44244-PE|CG44244-PG|CG44244-PH|CG44244-PI|Glycogenin-PB|Glycogenin-PD|Glycogenin-PE|Glycogenin-PG|Glycogenin-PH|Glycogenin-PI|glycogenin', map_location='57D1-57D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21200483, end_position_on_the_genomic_accession=21210821, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37420, CurrentID=0, Status='live', Symbol='CG4302', Aliases='Dmel_ BEST:GH09393, Dmel\\CG4302, GH09393', description='CG4302 gene product from transcript CG4302-RA', other_designations='CG4302-PA', map_location='57D1-57D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21212880, end_position_on_the_genomic_accession=21214972, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37421, CurrentID=0, Status='live', Symbol='CG15661', Aliases='Dmel_ Dmel\\CG15661', description='CG15661 gene product from transcript CG15661-RA', other_designations='CG15661-PA|CG15661-PB|uncharacterized protein', map_location='57D2-57D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21215435, end_position_on_the_genomic_accession=21217779, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37422, CurrentID=0, Status='live', Symbol='ASPP', Aliases='Dmel_CG18375, CG18375, Dmel\\CG18375, dASPP', description='Ankyrin-repeat, SH3-domain, and Proline-rich-region containing Protein', other_designations='ASPP-PA|ASPP-PB|ASPP-PC|Ankyrin-repeat, SH3-domain and Proline-rich-region containing protein|Ankyrin-repeat, SH3-domain, and Proline-rich-region containing protein|CG18375-PA|CG18375-PB|CG18375-PC', map_location='57D2-57D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21217856, end_position_on_the_genomic_accession=21251150, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37423, CurrentID=0, Status='live', Symbol='CG30391', Aliases='Dmel_ CG15662, Dmel\\CG30391', description='CG30391 gene product from transcript CG30391-RA', other_designations='CG30391-PA|uncharacterized protein', map_location='57D4-57D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21264805, end_position_on_the_genomic_accession=21266342, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37427, CurrentID=0, Status='live', Symbol='MFS16', Aliases='Dmel_CG10069, CG10069, Dmel\\CG10069', description='Major Facilitator Superfamily Transporter 16', other_designations='CG10069-PA|CG10069-PB|CG10069-PC|CG10069-PD|MFS16-PA|MFS16-PB|MFS16-PC|MFS16-PD|major facilitator superfamily transporter 16', map_location='57D7-57D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21282298, end_position_on_the_genomic_accession=21285679, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37428, CurrentID=0, Status='live', Symbol='CG10505', Aliases='Dmel_ Dme_CG10505, Dmel\\CG10505', description='CG10505 gene product from transcript CG10505-RA', other_designations='ABC transporter|CG10505-PA|uncharacterized protein', map_location='57D7-57D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21285248, end_position_on_the_genomic_accession=21290343, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37429, CurrentID=0, Status='live', Symbol='Sgf29', Aliases='Dmel_CG30390, CG10509, CG30390, Dmel\\CG30390, dSgf29', description='SAGA-associated factor 29kDa', other_designations='CG30390-PA|CG30390-PB|SAGA-associated factor 29 ortholog|Sgf29-PA|Sgf29-PB', map_location='57D8-57D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21292239, end_position_on_the_genomic_accession=21293646, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37430, CurrentID=0, Status='live', Symbol='RpL29', Aliases='Dmel_CG10071, BcDNA:RH58777, CG10071, DL43, Dmel\\CG10071, L29, RpL43, anon-EST:Posey204, anon-EST:Posey73', description='Ribosomal protein L29', other_designations='60S ribosomal component|60S ribosomal protein L43|CG10071-PA|CG10071-PB|CG10071-PD|RpL29-PA|RpL29-PB|RpL29-PD|ribosomal protein L29', map_location='57D8-57D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21293768, end_position_on_the_genomic_accession=21294404, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37431, CurrentID=0, Status='live', Symbol='CG9752', Aliases='Dmel_ Dmel\\CG9752', description='CG9752 gene product from transcript CG9752-RA', other_designations='CG9752-PA', map_location='57D8-57D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21294402, end_position_on_the_genomic_accession=21295588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37433, CurrentID=0, Status='live', Symbol='CG42672', Aliases='Dmel_ CG10074, CG15665, CG30386, CG30387, Dmel\\CG42672, Dmel_CG30386, Dmel_CG30387', description='CG42672 gene product from transcript CG42672-RP', other_designations='CG42672-PD|CG42672-PF|CG42672-PG|CG42672-PI|CG42672-PJ|CG42672-PK|CG42672-PL|CG42672-PM|CG42672-PO|CG42672-PP|uncharacterized protein', map_location='57D8-57D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21295652, end_position_on_the_genomic_accession=21310411, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  37434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37434, CurrentID=0, Status='live', Symbol='CG9754', Aliases='Dmel_ Dmel\\CG9754', description='CG9754 gene product from transcript CG9754-RA', other_designations='CG9754-PA', map_location='57D11-57D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21310368, end_position_on_the_genomic_accession=21312354, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37435, CurrentID=0, Status='live', Symbol='CG9485', Aliases='Dmel_ BEST:GH22856, Dmel\\CG9485, FBgn0034618', description='CG9485 gene product from transcript CG9485-RF', other_designations='CG9485-PA|CG9485-PB|CG9485-PC|CG9485-PD|CG9485-PE|CG9485-PF', map_location='57D11-57D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21312399, end_position_on_the_genomic_accession=21319569, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37437, CurrentID=0, Status='live', Symbol='CG30394', Aliases='Dmel_ CG10500, CG15670, Dmel\\CG30394, anon-WO0172774.187, anon-WO0172774.188', description='CG30394 gene product from transcript CG30394-RB', other_designations='CG30394-PA|CG30394-PB|uncharacterized protein', map_location='57D11-57D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21320040, end_position_on_the_genomic_accession=21323320, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37439, CurrentID=0, Status='live', Symbol='CG15666', Aliases='Dmel_ BBS9, BcDNA:RH67628, Dmel\\CG15666', description='CG15666 gene product from transcript CG15666-RD', other_designations='CG15666-PD|uncharacterized protein', map_location='57D12-57D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21341972, end_position_on_the_genomic_accession=21345608, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37440, CurrentID=0, Status='live', Symbol='CG9822', Aliases='Dmel_ Agr-like, Dmel\\CG9822', description='CG9822 gene product from transcript CG9822-RA', other_designations='CG9822-PA', map_location='57D12-57D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21346875, end_position_on_the_genomic_accession=21347883, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37441, CurrentID=0, Status='live', Symbol='CG17974', Aliases='Dmel_ Dmel\\CG17974', description='CG17974 gene product from transcript CG17974-RA', other_designations='CG17974-PA|uncharacterized protein', map_location='57D12-57D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21348945, end_position_on_the_genomic_accession=21350578, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37443, CurrentID=0, Status='live', Symbol='CG10795', Aliases='Dmel_ Dmel\\CG10795, LD27358', description='CG10795 gene product from transcript CG10795-RA', other_designations='CG10795-PA|uncharacterized protein', map_location='57E1-57E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21383227, end_position_on_the_genomic_accession=21384391, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37444, CurrentID=0, Status='live', Symbol='EfSec', Aliases='Dmel_CG9841, CG9841, Dmel\\CG9841, EFSec, EFsec, Efsec, Eftu, SelB/eEFsec, dEFsec, deEFsec, eEFsec', description='CG9841 gene product from transcript CG9841-RA', other_designations='CG9841-PA|EfSec-PA', map_location='57E1-57E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21384552, end_position_on_the_genomic_accession=21386221, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37445, CurrentID=0, Status='live', Symbol='Acox57D-p', Aliases='Dmel_CG9707, Acox57Dp, CG9707, D-acox, Dmel\\CG9707, acox', description='acyl-Coenzyme A oxidase at 57D proximal', other_designations='Acox57D-p-PA|Acox57D-p-PB|CG9707-PA|CG9707-PB|acyl co-enzyme A oxidase', map_location='57E1-57E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21386309, end_position_on_the_genomic_accession=21389276, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37446, CurrentID=0, Status='live', Symbol='Acox57D-d', Aliases='Dmel_CG9709, Acox57Dd, CG9709, D-acox, Dmel\\CG9709, acox', description='acyl-Coenzyme A oxidase at 57D distal', other_designations='Acox57D-d-PA|Acyl-coenzyme A oxidase|CG9709-PA|acyl co-enzyme A oxidase', map_location='57E1-57E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21389405, end_position_on_the_genomic_accession=21392439, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37447, CurrentID=0, Status='live', Symbol='Sdc', Aliases='Dmel_CG10497, CG10497, CT27296, Dmel\\CG10497, Dsyn, Syd, Syn, dSdc, l(2)10608, sdc, syd, syn', description='Syndecan', other_designations='CG10497-PA|CG10497-PB|CG10497-PD|CG10497-PG|CG10497-PH|CG10497-PI|CG10497-PJ|CG10497-PK|Sdc-PA|Sdc-PB|Sdc-PD|Sdc-PG|Sdc-PH|Sdc-PI|Sdc-PJ|Sdc-PK|dSyndecan|syndecan', map_location='57E1-57E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21392402, end_position_on_the_genomic_accession=21481176, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37449, CurrentID=0, Status='live', Symbol='Fkbp14', Aliases='Dmel_CG9847, BcDNA.GM07659, BcDNA:GH08925, BcDNA:GM07659, CG9847, Dmel\\CG9847, FKB13, FKBP13, FKBP14, Fkbp13, GM07659, fkbp13, l(2)00734', description='FK506-binding protein 14 ortholog', other_designations='CG9847-PA|CG9847-PC|CG9847-PD|CG9847-PE|CG9847-PF|Fkbp13|Fkbp14-PA|Fkbp14-PC|Fkbp14-PD|Fkbp14-PE|Fkbp14-PF', map_location='57E6-57E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21487167, end_position_on_the_genomic_accession=21497929, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37450, CurrentID=0, Status='live', Symbol='TAF1C-like', Aliases='Dmel_CG10496, CG10496, Dmel\\CG10496', description='TATA box-binding protein-associated factor RNA polymerase I subunit C-like', other_designations='CG10496-PA|TAF1C-like-PA', map_location='57E6-57E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21498903, end_position_on_the_genomic_accession=21501732, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37451, CurrentID=0, Status='live', Symbol='MESK2', Aliases='Dmel_CG15669, BcDNA:GH12032, BcDNA:LD27504, BcDNA:RH10174, CG15668, CG15669, Dmel\\CG15669, Mesk2, l(2)01467, mesk2', description='Misexpression suppressor of KSR 2', other_designations='CG15669-PA|CG15669-PB|CG15669-PC|CG15669-PD|CG15669-PE|CG15669-PF|CG15669-PI|CG15669-PJ|CG15669-PK|MESK2-PA|MESK2-PB|MESK2-PC|MESK2-PD|MESK2-PE|MESK2-PF|MESK2-PI|MESK2-PJ|MESK2-PK|misexpression suppressor of KSR 2', map_location='57E8-57E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21502097, end_position_on_the_genomic_accession=21517241, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37453, CurrentID=0, Status='live', Symbol='CG10494', Aliases='Dmel_ Dmel\\CG10494', description='CG10494 gene product from transcript CG10494-RB', other_designations='CG10494-PA|CG10494-PB|uncharacterized protein', map_location='57E9-57E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21516774, end_position_on_the_genomic_accession=21519125, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37454, CurrentID=0, Status='live', Symbol='CG30283', Aliases='Dmel_ CG10450, Dmel\\CG30283, SP37, SPH37', description='CG30283 gene product from transcript CG30283-RB', other_designations='CG30283-PB|uncharacterized protein', map_location='57E11-57E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21539584, end_position_on_the_genomic_accession=21540573, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37455, CurrentID=0, Status='live', Symbol='Egfr', Aliases='Dmel_CG10079, C-erb, CG10079, D-EGFR, D-Egf, DEGFR, DER, DER flb, DER/EGFR, DER/top, DER/torpedo, DER1, DEgfr, Degfr, Der, DmHD-33, Dmel\\CG10079, EC2-4, EFG-R, EGF-R, EGFR, EGFr, EGfr, EK2-6, Egf, Egf-r, EgfR, El, Elp, Elp-1, Elp-B1, Elp-B1RB1, Flb, HD-33, TOP, Tor, Torpedo/DER, Torpedo/Egfr, c-erbB, d-egf-r, dEGFR, dEGFR1, dEGFr, dEgfr, degfr, der, egfr, flb, l(2)05351, l(2)09261, l(2)57DEFa, l(2)57EFa, l(2)57Ea, mor1, top, top/DER, top/flb, torpedo/Egfr, torpedo/egfr', description='Epidermal growth factor receptor', other_designations='CG10079-PA|CG10079-PB|DER-Ellipse|EGF-receptor|EGFR|Egf receptor|Egfr-PA|Egfr-PB|drosophila epidermal growth factor receptor homologue|ellipse|ellipse torpedo|epidermal growth factor receptor|faint little ball|morphological defects 1|torpedo|torpedo/DER', map_location='57E9-57F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21522420, end_position_on_the_genomic_accession=21559977, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37456, CurrentID=0, Status='live', Symbol='twz', Aliases='Dmel_CG10440, CG10440, Dmel\\CG10440, Twz', description='tiwaz', other_designations='CG10440-PA|CG10440-PB|twz-PA|twz-PB', map_location='57F1-57F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21560246, end_position_on_the_genomic_accession=21576035, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37458, CurrentID=0, Status='live', Symbol='CG10433', Aliases='Dmel_ Dmel\\CG10433', description='CG10433 gene product from transcript CG10433-RC', other_designations='CG10433-PA|CG10433-PB|CG10433-PC|uncharacterized protein', map_location='57F3-57F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21605676, end_position_on_the_genomic_accession=21608125, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37459, CurrentID=0, Status='live', Symbol='CG15673', Aliases='Dmel_ Dmel\\CG15673', description='CG15673 gene product from transcript CG15673-RC', other_designations='CG15673-PC|uncharacterized protein', map_location='57F3-57F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21608393, end_position_on_the_genomic_accession=21613071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37461, CurrentID=0, Status='live', Symbol='PTP-ER', Aliases='Dmel_CG9856, CG9856, Dmel\\CG9856, ER2-5, ptp-er', description='Protein tyrosine phosphatase-ERK/Enhancer of Ras1', other_designations='CG9856-PA|PTP-ER-PA|protein tyrosine phosphatase|protein tyrosine phosphatase-ERK/Enhancer of Ras1', map_location='57F4-57F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21619321, end_position_on_the_genomic_accession=21624446, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37462, CurrentID=0, Status='live', Symbol='mahj', Aliases='Dmel_CG10080, BcDNA:SD08722, CG10080, Dmel\\CG10080, VprBP, anon-WO0172774.151', description='mahjong', other_designations='CG10080-PA|CG10080-PB|mahj-PA|mahj-PB', map_location='57F4-57F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21625030, end_position_on_the_genomic_accession=21631422, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37463, CurrentID=0, Status='live', Symbol='CG15674', Aliases='Dmel_ Dmel\\CG15674', description='CG15674 gene product from transcript CG15674-RA', other_designations='CG15674-PA|CG15674-PB|uncharacterized protein', map_location='57F5-57F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21632007, end_position_on_the_genomic_accession=21634362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37464, CurrentID=0, Status='live', Symbol='CG10321', Aliases='Dmel_ Dmel\\CG10321', description='CG10321 gene product from transcript CG10321-RA', other_designations='CG10321-PA|CG10321-PB|uncharacterized protein', map_location='57F5-57F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21634679, end_position_on_the_genomic_accession=21638399, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37465, CurrentID=0, Status='live', Symbol='CG10082', Aliases='Dmel_ Dmel\\CG10082, anon-WO0172774.149, dmIP6K', description='CG10082 gene product from transcript CG10082-RA', other_designations='CG10082-PA|CG10082-PB|uncharacterized protein', map_location='57F6-57F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21646177, end_position_on_the_genomic_accession=21657935, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37466, CurrentID=0, Status='live', Symbol='ND-B12', Aliases='Dmel_CG10320, 10320, CG10320, Dmel\\CG10320, NB2M, anon-EST:Posey47, anon-EST:Posey84', description='NADH dehydrogenase (ubiquinone) B12 subunit', other_designations='CG10320-PA|CG10320-PB|CG10320-PC|CG10320-PD|CG10320-PE|NADH ubiquinone oxidoreductase B12|NADH-ubiquinone oxidoreductase B12 subunit|NADH:ubiquinone oxidoreductase B12 subunit|ND-B12-PA|ND-B12-PB|ND-B12-PC|ND-B12-PD|ND-B12-PE', map_location='57F6-57F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21658078, end_position_on_the_genomic_accession=21658777, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37467, CurrentID=0, Status='live', Symbol='Rae1', Aliases='Dmel_CG9862, CG9862, Dmel\\CG9862, dmrae1, rae1', description='CG9862 gene product from transcript CG9862-RA', other_designations='CG9862-PA|Rae1-PA', map_location='2-96.2 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21658770, end_position_on_the_genomic_accession=21660007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37468, CurrentID=0, Status='live', Symbol='pirk', Aliases='Dmel_CG15678, BcDNA:SD16010, CG15678, Dmel\\CG15678, PIRK, Pims, Pirk, Rudra, pims, rudra', description='poor Imd response upon knock-in', other_designations='CG15678-PA|CG15678-PB|PGRP-LC-interacting inhibitor of Imd signaling|pirk-PA|pirk-PB', map_location='57F6-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21660967, end_position_on_the_genomic_accession=21662995, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37469, CurrentID=0, Status='live', Symbol='CG42365', Aliases='Dmel_ CG15675, Dmel\\CG42365, Dmel_CG15675', description='CG42365 gene product from transcript CG42365-RA', other_designations='CG42365-PA|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21666687, end_position_on_the_genomic_accession=21667594, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37470, CurrentID=0, Status='live', Symbol='CG9865', Aliases='Dmel_ Dmel\\CG9865, Dromel_CG9865_FBtr0071680_cx9c_morf', description='CG9865 gene product from transcript CG9865-RA', other_designations='CG9865-PA|CG9865-PB', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21667541, end_position_on_the_genomic_accession=21669916, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37471, CurrentID=0, Status='live', Symbol='NC2alpha', Aliases='Dmel_CG10318, CG10318, Dmel\\CG10318, Drap1, NC2, Nc2alpha, dDrap, dDrap1, dDrap1/dNC2a, dNC2', description='Negative Cofactor 2alpha', other_designations='CG10318-PA|NC2alpha-PA|negative cofactor 2alpha', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21670152, end_position_on_the_genomic_accession=21671593, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37472, CurrentID=0, Status='live', Symbol='CG15676', Aliases='Dmel_ Dmel\\CG15676', description='CG15676 gene product from transcript CG15676-RB', other_designations='CG15676-PA|CG15676-PB|uncharacterized protein', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21671700, end_position_on_the_genomic_accession=21672845, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37473, CurrentID=0, Status='live', Symbol='stum', Aliases='Dmel_CG30263, CG15677, CG15679, CG30263, CT35864, Dmel\\CG30263, mdcds_15969', description='stumble', other_designations='CG30263-PB|CG30263-PD|CG30263-PE|stum-PB|stum-PD|stum-PE', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21674901, end_position_on_the_genomic_accession=21685000, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37475, CurrentID=0, Status='live', Symbol='CG10306', Aliases='Dmel_ Dmel\\CG10306, eIF3k', description='CG10306 gene product from transcript CG10306-RA', other_designations='CG10306-PA|uncharacterized protein', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21685283, end_position_on_the_genomic_accession=21686201, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37476, CurrentID=0, Status='live', Symbol='Tbp', Aliases='Dmel_CG9874, CG9874, Dmel\\CG9874, RBP, TBP, TBP38, TFIID, TFIIDtau, TfIID, dTBP, dTFIID, dTbp, dtbp', description='TATA binding protein', other_designations='42kDa polypeptide|CG9874-PA|TATA box binding protein|TATA box-binding protein|TATA-binding protein|TATA-box binding protein|Tbp-PA', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21686192, end_position_on_the_genomic_accession=21687642, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37477, CurrentID=0, Status='live', Symbol='CG10307', Aliases='Dmel_ CT28947, Dmel\\CG10307', description='CG10307 gene product from transcript CG10307-RA', other_designations='CG10307-PA|CG10307-PB|uncharacterized protein', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21688773, end_position_on_the_genomic_accession=21690253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37478, CurrentID=0, Status='live', Symbol='Tim10', Aliases='Dmel_CG9878, CG9878, Dmel\\CG9878, Dromel_CG9878_FBtr0071676_tim10_mORF, bs31d12.y1, tim10', description='Translocase of inner membrane 10', other_designations='CG9878-PA|CG9878-PB|Tim10-PA|Tim10-PB|translocase of inner membrane 10', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21690215, end_position_on_the_genomic_accession=21691139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37479, CurrentID=0, Status='live', Symbol='CngB', Aliases='Dmel_CG17922, CG17922, Cncbeta, Dmel\\CG17922, dCNCbeta, dmCNGB', description='Cyclic nucleotide-gated ion channel subunit B', other_designations='CG17922-PA|CG17922-PB|CngB-PA|CngB-PB|Cyclic-nucleotide-gated ion channel protein beta-subunit|cyclic nucleotide-gated ion channel subunit B', map_location='57F9-57F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21692691, end_position_on_the_genomic_accession=21697152, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37480, CurrentID=0, Status='live', Symbol='HmgZ', Aliases='Dmel_CG17921, BcDNA:RE28596, CG17921, Dmel\\CG17921, HMG Z, HMG-Z, HMG1, anon-EST:Liang-1.58, clone 1.58, dHMG-Z, hmg-Z, hmgz', description='HMG protein Z', other_designations='CG17921-PA|CG17921-PB|CG17921-PC|CG17921-PD|HmgZ-PA|HmgZ-PB|HmgZ-PC|HmgZ-PD', map_location='57F9-57F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21697488, end_position_on_the_genomic_accession=21707861, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37481, CurrentID=0, Status='live', Symbol='HmgD', Aliases='Dmel_CG17950, CG17950, Dm-HMG-D100, Dmel\\CG17950, HMG D, HMG-D, HMG-D100, HMG1, HMGD, Hmg-D, anon-EST:Liang-1.9, clone 1.9, dHMG-D, hmg-D, hmgd', description='High mobility group protein D', other_designations='CG17950-PA|CG17950-PB|CG17950-PC|CG17950-PD|HmgD-PA|HmgD-PB|HmgD-PC|HmgD-PD|high mobility group protein D', map_location='57F9-57F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21701458, end_position_on_the_genomic_accession=21716727, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37482, CurrentID=0, Status='live', Symbol='LBR', Aliases='Dmel_CG17952, CG17952, Dmel\\CG17952, dLBR', description='Lamin B receptor', other_designations='CG17952-PA|CG17952-PB|CG17952-PC|LBR-PA|LBR-PB|LBR-PC|lamin B receptor|lamin-B receptor', map_location='57F10-57F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21720495, end_position_on_the_genomic_accession=21724055, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37483, CurrentID=0, Status='live', Symbol='Grx-1', Aliases='Dmel_CG7975, CG7975, Dmel\\CG7975', description='CG7975 gene product from transcript CG7975-RA', other_designations='CG7975-PA|Grx-1-PA', map_location='57F11-57F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21757345, end_position_on_the_genomic_accession=21757913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37484, CurrentID=0, Status='live', Symbol='CG4021', Aliases='Dmel_ Dmel\\CG4021, cg4021', description='CG4021 gene product from transcript CG4021-RA', other_designations='CG4021-PA', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21773295, end_position_on_the_genomic_accession=21774876, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37485, CurrentID=0, Status='live', Symbol='lox2', Aliases='Dmel_CG4402, CG4402, Dmel\\CG4402, Dmloxl-2, dmlox-2', description='lysyl oxidase-like 2', other_designations='CG4402-PA|lox2-PA', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21789623, end_position_on_the_genomic_accession=21791580, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37486, CurrentID=0, Status='live', Symbol='CG4386', Aliases='Dmel_ Dmel\\CG4386, SP17', description='CG4386 gene product from transcript CG4386-RA', other_designations='CG4386-PA', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21799330, end_position_on_the_genomic_accession=21800904, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37487, CurrentID=0, Status='live', Symbol='CG13492', Aliases='Dmel_ Dmel\\CG13492', description='CG13492 gene product from transcript CG13492-RD', other_designations='CG13492-PD|uncharacterized protein', map_location='58A2-58A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21820662, end_position_on_the_genomic_accession=21831409, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37488, CurrentID=0, Status='live', Symbol='CG4363', Aliases='Dmel_ BcDNA:LP06375, CT14230, Dmel\\CG4363', description='CG4363 gene product from transcript CG4363-RA', other_designations='CG4363-PA', map_location='58A2-58A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21834230, end_position_on_the_genomic_accession=21835047, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37489, CurrentID=0, Status='live', Symbol='CG4377', Aliases='Dmel_ CT14296, Dmel\\CG4377', description='CG4377 gene product from transcript CG4377-RA', other_designations='CG4377-PA', map_location='58A2-58A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21835448, end_position_on_the_genomic_accession=21836357, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37490, CurrentID=0, Status='live', Symbol='CG4372', Aliases='Dmel_ Dmel\\CG4372', description='CG4372 gene product from transcript CG4372-RA', other_designations='CG4372-PA', map_location='58A2-58A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21837605, end_position_on_the_genomic_accession=21838798, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37491, CurrentID=0, Status='live', Symbol='CG9294', Aliases='Dmel_ Dmel\\CG9294, SP20', description='CG9294 gene product from transcript CG9294-RB', other_designations='CG9294-PB', map_location='58A3-58A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21860825, end_position_on_the_genomic_accession=21861994, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37492, CurrentID=0, Status='live', Symbol='comr', Aliases='Dmel_CG13493, CG13493, Comr, Dmel\\CG13493, Z2-1340, zab', description='cookie monster', other_designations='CG13493-PA|comr-PA|zuker-aly-like b', map_location='58A3-58A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21864870, end_position_on_the_genomic_accession=21866859, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37493, CurrentID=0, Status='live', Symbol='PpN58A', Aliases='Dmel_CG3245, CG3245, DmPpN-58A, Dmel\\CG3245, PPN, PPN 58A, PpN-58A', description='Protein phosphatase N at 58A', other_designations='CG3245-PA|PpN58A-PA|protein phosphatase N at 58A', map_location='58A3-58A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21881352, end_position_on_the_genomic_accession=21882533, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37496, CurrentID=0, Status='live', Symbol='CG13488', Aliases='Dmel_ Dmel\\CG13488', description='CG13488 gene product from transcript CG13488-RC', other_designations='CG13488-PA|CG13488-PB|CG13488-PC|uncharacterized protein', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21951592, end_position_on_the_genomic_accession=21952562, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37497, CurrentID=0, Status='live', Symbol='CG13494', Aliases='Dmel_ BcDNA:AT17606, Dmel\\CG13494', description='CG13494 gene product from transcript CG13494-RA', other_designations='CG13494-PA|uncharacterized protein', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21954242, end_position_on_the_genomic_accession=21954845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37498, CurrentID=0, Status='live', Symbol='ppk9', Aliases='Dmel_CG34369, CG13489, CG13490, CG34369, Dmel\\CG34369, Dmel_CG13489, Dmel_CG13490', description='pickpocket 9', other_designations='CG34369-PA|CG34369-PB|ppk9-PA|ppk9-PB', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21964173, end_position_on_the_genomic_accession=21970664, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  37500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37500, CurrentID=0, Status='live', Symbol='CG9304', Aliases='Dmel_ Dmel\\CG9304', description='CG9304 gene product from transcript CG9304-RB', other_designations='CG9304-PA|CG9304-PB', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21973122, end_position_on_the_genomic_accession=21976386, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37501, CurrentID=0, Status='live', Symbol='Gr58c', Aliases='Dmel_CG13491, 58A.3, 58A3, CG13491, Dmel\\CG13491, GR58A.3, GR58c, Gr58A3', description='Gustatory receptor 58c', other_designations='CG13491-PA|Gr58c-PA|gustatory receptor 58A.3|gustatory receptor 58c', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21976662, end_position_on_the_genomic_accession=21977957, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37502, CurrentID=0, Status='live', Symbol='Gr58b', Aliases='Dmel_CG13495, 58A.2, 58A2, CG13495, Dmel\\CG13495, GR58A.2, Gr58A2', description='Gustatory receptor 58b', other_designations='CG13495-PA|Gr58b-PA|gustatory receptor 58A.2|gustatory receptor 58b', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21978215, end_position_on_the_genomic_accession=21979677, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37507, CurrentID=0, Status='live', Symbol='CG9308', Aliases='Dmel_ Dmel\\CG9308, p24', description='CG9308 gene product from transcript CG9308-RA', other_designations='CG9308-PA', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22034844, end_position_on_the_genomic_accession=22035666, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740593, CurrentID=0, Status='live', Symbol='ND-MWFE', Aliases='Dmel_CG34439, BcDNA:RH49308, CG13328, CG17054, CG30057, CG34439, Dmel\\CG34439, l(2)c00093', description='NADH dehydrogenase (ubiquinone) MWFE subunit', other_designations='CG34439-PA|CG34439-PB|NADH ubiquinone oxidoreductase MWFE|ND-MWFE-PA|ND-MWFE-PB|lethal (2) c00093', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13158666, end_position_on_the_genomic_accession=13160190, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37510, CurrentID=0, Status='live', Symbol='CG13501', Aliases='Dmel_ Dmel\\CG13501', description='CG13501 gene product from transcript CG13501-RB', other_designations='CG13501-PA|CG13501-PB|uncharacterized protein', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22058904, end_position_on_the_genomic_accession=22061517, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37511, CurrentID=0, Status='live', Symbol='Snp', Aliases='Dmel_CG44248, CG11478, CG30327, CG42257, CG44248, CG6393, Dmel\\CG44248, Dmel_CG30327, Dmel_CG42257, Dmel_CG6393', description='Snipper', other_designations='CG44248-PD|CG44248-PE|CG44248-PF|CG44248-PG|Snp-PD|Snp-PE|Snp-PF|Snp-PG|snipper', map_location='58B1-58B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22060950, end_position_on_the_genomic_accession=22069757, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37513, CurrentID=0, Status='live', Symbol='CG11475', Aliases='Dmel_ Dmel\\CG11475', description='CG11475 gene product from transcript CG11475-RA', other_designations='CG11475-PA|uncharacterized protein', map_location='58B2-58B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22069725, end_position_on_the_genomic_accession=22071286, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37514, CurrentID=0, Status='live', Symbol='CG11474', Aliases='Dmel_ Dmel\\CG11474', description='CG11474 gene product from transcript CG11474-RC', other_designations='CG11474-PA|CG11474-PC|uncharacterized protein', map_location='58B2-58B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22071234, end_position_on_the_genomic_accession=22073443, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37515, CurrentID=0, Status='live', Symbol='CG2921', Aliases='Dmel_ Dmel\\CG2921', description='CG2921 gene product from transcript CG2921-RA', other_designations='CG2921-PA', map_location='58B3-58B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22073538, end_position_on_the_genomic_accession=22075308, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37516, CurrentID=0, Status='live', Symbol='GlcT-1', Aliases='Dmel_CG6437, CG6437, DGlcT-1, Dmel\\CG6437, GlcT, GlcT1, dGlcCerT', description='Ceramide glucosyltransferase', other_designations='CG6437-PA|GlcT-1-PA|ceramide glucosyltransferase|glucosylceramide synthase|glucosyltransferase', map_location='58B3-58B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22075988, end_position_on_the_genomic_accession=22079381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37517, CurrentID=0, Status='live', Symbol='Synj', Aliases='Dmel_CG6562, CG6562, Dmel\\CG6562, IPP, synj', description='Synaptojanin', other_designations='CG6562-PA|CG6562-PB|CG6562-PC|Synj-PA|Synj-PB|Synj-PC|synaptojanin', map_location='58B4-58B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22080176, end_position_on_the_genomic_accession=22085767, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37518, CurrentID=0, Status='live', Symbol='CG13502', Aliases='Dmel_ Dmel\\CG13502', description='CG13502 gene product from transcript CG13502-RD', other_designations='CG13502-PA|CG13502-PD|uncharacterized protein', map_location='58B4-58B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22085782, end_position_on_the_genomic_accession=22088727, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37519, CurrentID=0, Status='live', Symbol='CG11073', Aliases='Dmel_ Dmel\\CG11073', description='CG11073 gene product from transcript CG11073-RD', other_designations='CG11073-PA|CG11073-PB|CG11073-PC|CG11073-PD|CG11073-PE|uncharacterized protein', map_location='58B5-58B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22091489, end_position_on_the_genomic_accession=22099534, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37520, CurrentID=0, Status='live', Symbol='CG6613', Aliases='Dmel_ Dmel\\CG6613', description='CG6613 gene product from transcript CG6613-RA', other_designations='CG6613-PA', map_location='58B7-58B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22103177, end_position_on_the_genomic_accession=22105963, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37521, CurrentID=0, Status='live', Symbol='Vrp1', Aliases='Dmel_CG13503, CG13503, D-WIP, Dm_Vrp1, Dmel\\CG13503, WIP, dWIP, sltr, solas, vrp1, wip', description='Verprolin 1', other_designations='CG13503-PA|CG13503-PB|CG13503-PC|CG13503-PD|CG13503-PE|CG13503-PF|CG13503-PG|CG13503-PH|CG13503-PI|CG13503-PJ|CG13503-PK|CG13503-PL|D-WASp-interacting protein|Verprolin1|Vrp1-PA|Vrp1-PB|Vrp1-PC|Vrp1-PD|Vrp1-PE|Vrp1-PF|Vrp1-PG|Vrp1-PH|Vrp1-PI|Vrp1-PJ|Vrp1-PK|Vrp1-PL|WASP-interacting protein|WASp interacting protein|drosophila WIP|solitary|veroprolin|verprolin|verprolin 1', map_location='58B8-58B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22106107, end_position_on_the_genomic_accession=22122562, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  37522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37522, CurrentID=0, Status='live', Symbol='CG30280', Aliases='Dmel_ AAF46801, CG6676, Dmel\\CG30280, FBpp0071673', description='CG30280 gene product from transcript CG30280-RB', other_designations='CG30280-PB|CG30280-PC|Fibrinogen-related protein|uncharacterized protein', map_location='58B8-58B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22113589, end_position_on_the_genomic_accession=22114841, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37523, CurrentID=0, Status='live', Symbol='mei-S332', Aliases='Dmel_CG5303, BEST:LD13774, CG5303, Dmel\\CG5303, MEI-S322, MEI-S332, Mei-S332, Mei-S332/Shugoshin, MeiS332, mei-S322, mei-s332, meiS322, meiS332', description='meiotic from via Salaria 332', other_designations='CG5303-PA|CG5303-PB|mei-S332-PA|mei-S332-PB|meiotic from salaria 332|meiotic from via salaria 332|shugoshin', map_location='58B9-58B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22122101, end_position_on_the_genomic_accession=22124566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37524, CurrentID=0, Status='live', Symbol='GM130', Aliases='Dmel_CG11061, CG11061, Dmel\\CG11061, GM-130, dGM130, dgm130', description='Golgi matrix protein 130 kD ortholog', other_designations='CG11061-PA|GM130-PA|golgi matrix protein 130 kD ortholog', map_location='58B10-58B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22124567, end_position_on_the_genomic_accession=22127502, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37526, CurrentID=0, Status='live', Symbol='CG34205', Aliases='Dmel_ CG13504, Dmel\\CG34205', description='CG34205 gene product from transcript CG34205-RA', other_designations='CG34205-PA|uncharacterized protein', map_location='58C1-58C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22130112, end_position_on_the_genomic_accession=22131299, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37527, CurrentID=0, Status='live', Symbol='CG11269', Aliases='Dmel_ Dmel\\CG11269', description='CG11269 gene product from transcript CG11269-RA', other_designations='CG11269-PA|uncharacterized protein', map_location='58C1-58C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22138895, end_position_on_the_genomic_accession=22139295, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37528, CurrentID=0, Status='live', Symbol='Fmr1', Aliases='Dmel_CG6203, AT24755, BcDNA:GM08679, CG6203, Dmel\\CG6203, EP(3)3517, FMR, FMR1, FMRP, FMRp, FXR, Fmrp, cg6203, dFMR, dFMR1, dFMRP, dFXR, dFXR1, dFXRP, d dFmrp, dfmr, dfmr1, dfxr, dfxr1, dmfr1, fmr, fmr1', description='CG6203 gene product from transcript CG6203-RC', other_designations='CG6203-PA|CG6203-PB|CG6203-PC|CG6203-PD|CG6203-PE|CG6203-PF|CG6203-PG|CG6203-PH|CG6203-PI|CG6203-PJ|CG6203-PK|Fmr1-PA|Fmr1-PB|Fmr1-PC|Fmr1-PD|Fmr1-PE|Fmr1-PF|Fmr1-PG|Fmr1-PH|Fmr1-PI|Fmr1-PJ|Fmr1-PK|Fragile-X mental retardation protein|Fragile-X-related|dFMR|dFmr1|drosophila fragile X mental retardation protein|fragile X|fragile X mental retardation|fragile X mental retardation 1|fragile X mental retardation gene|fragile X mental retardation protein|fragile X mental retardation protein 1|fragile X protein|fragile X related protein|fragile X-related|fragile x related|fragile-X', map_location='85F10-85F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10101420, end_position_on_the_genomic_accession=10110121, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  37531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37531, CurrentID=0, Status='live', Symbol='CG3045', Aliases='Dmel_ Dmel\\CG3045', description='CG3045 gene product from transcript CG3045-RA', other_designations='CG3045-PA', map_location='58C5-58C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22172803, end_position_on_the_genomic_accession=22174691, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37532, CurrentID=0, Status='live', Symbol='CG6758', Aliases='Dmel_ Dmel\\CG6758, FBXO6', description='CG6758 gene product from transcript CG6758-RA', other_designations='CG6758-PA', map_location='58C5-58C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22174922, end_position_on_the_genomic_accession=22177428, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37533, CurrentID=0, Status='live', Symbol='CG11170', Aliases='Dmel_ CG30279, Dmel\\CG11170', description='CG11170 gene product from transcript CG11170-RC', other_designations='CG11170-PB|CG11170-PC|CG11170-PD|uncharacterized protein', map_location='58C5-58C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22177419, end_position_on_the_genomic_accession=22182879, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37534, CurrentID=0, Status='live', Symbol='CG11275', Aliases='Dmel_ Dmel\\CG11275, i188', description='CG11275 gene product from transcript CG11275-RA', other_designations='CG11275-PA|uncharacterized protein', map_location='58C5-58C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22179695, end_position_on_the_genomic_accession=22182642, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37535, CurrentID=0, Status='live', Symbol='MED16', Aliases='Dmel_CG5465, CG5465, Dmel\\CG5465, Med16, Sin4/TRAP95, Trap95, dTRAP95, med16', description='Mediator complex subunit 16', other_designations='CG5465-PA|MED16-PA|mediator complex subunit 16', map_location='58C5-58C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22183138, end_position_on_the_genomic_accession=22185892, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37536, CurrentID=0, Status='live', Symbol='Vps35', Aliases='Dmel_CG5625, CG5625, DVps35, DmVps35, Dmel\\CG5625, Dvps35, VPS35, dvps35, vps35', description='Vacuolar protein sorting 35', other_designations='CG5625-PA|CG5625-PB|Vps35-PA|Vps35-PB|vacuolar protein sorting 35', map_location='58C7-58C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22185972, end_position_on_the_genomic_accession=22189886, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37537, CurrentID=0, Status='live', Symbol='Swim', Aliases='Dmel_CG3074, CG3074, Dmel\\CG3074, dmw, swim', description='Secreted Wg-interacting molecule', other_designations='CG3074-PA|CG3074-PB|Swim-PA|Swim-PB|dead man walking|secreted Wg-interacting molecule', map_location='58C7-58D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22189817, end_position_on_the_genomic_accession=22202671, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37538, CurrentID=0, Status='live', Symbol='CG3292', Aliases='Dmel_ Dmel\\CG3292', description='CG3292 gene product from transcript CG3292-RA', other_designations='CG3292-PA', map_location='58D1-58D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22203694, end_position_on_the_genomic_accession=22205593, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37539, CurrentID=0, Status='live', Symbol='Alp2', Aliases='Dmel_CG3290, Aph-2, CG3290, Dmel\\CG3290', description='Alkaline phosphatase 2', other_designations='Alp2-PA|CG3290-PA|alkaline phosphatase 2', map_location='58D1-58D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22206005, end_position_on_the_genomic_accession=22207987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37540, CurrentID=0, Status='live', Symbol='CG3264', Aliases='Dmel_ Dmel\\CG3264', description='CG3264 gene product from transcript CG3264-RB', other_designations='CG3264-PA|CG3264-PB', map_location='58D1-58D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22208464, end_position_on_the_genomic_accession=22210765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37541, CurrentID=0, Status='live', Symbol='CG11291', Aliases='Dmel_ 22026920, Dmel\\CG11291, Phos11291', description='CG11291 gene product from transcript CG11291-RA', other_designations='CG11291-PA|uncharacterized protein', map_location='58D1-58D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22211371, end_position_on_the_genomic_accession=22212704, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37542, CurrentID=0, Status='live', Symbol='ari-2', Aliases='Dmel_CG5709, AAF46823.1, CG5709, Dmel\\CG5709, ari2, l(2)07768', description='ariadne 2', other_designations='CG5709-PA|ari-2-PA|ariadne-2', map_location='58D1-58D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22212835, end_position_on_the_genomic_accession=22217264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37543, CurrentID=0, Status='live', Symbol='Oatp58Da', Aliases='Dmel_CG30277, CG30277, CG3387, Dmel\\CG30277, oatp 58Da', description='Organic anion transporting polypeptide 58Da', other_designations='CG30277-PA|Oatp58Da-PA|organic anion transporting polypeptide 58Da', map_location='58D2-58D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22218456, end_position_on_the_genomic_accession=22220905, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37544, CurrentID=0, Status='live', Symbol='Oatp58Db', Aliases='Dmel_CG3382, 58Db, CG3382, Dmel\\CG3382, oatp 58Db', description='Organic anion transporting polypeptide 58Db', other_designations='CG3382-PB|Oatp58Db-PB|organic anion transporting polypeptide 58Db', map_location='58D2-58D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22221180, end_position_on_the_genomic_accession=22223499, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37545, CurrentID=0, Status='live', Symbol='Oatp58Dc', Aliases='Dmel_CG3380, CG3380, Dmel\\CG3380, OATP, OATP58Dc, oatp 58Dc', description='Organic anion transporting polypeptide 58Dc', other_designations='CG3380-PA|CG3380-PB|CG3380-PC|Oatp58Dc-PA|Oatp58Dc-PB|Oatp58Dc-PC|organic anion transporting polypeptide|organic anion transporting polypeptide 58Dc', map_location='58D2-58D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22223681, end_position_on_the_genomic_accession=22228574, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37546, CurrentID=0, Status='live', Symbol='dve', Aliases='Dmel_CG5799, CG5799, DVE, Dmel\\CG5799, Dve, Dve-s, SATB1, dep, l(2)01738', description='defective proventriculus', other_designations='CG5799-PA|CG5799-PB|compass|dve-PA|dve-PB', map_location='58D2-58D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22243963, end_position_on_the_genomic_accession=22286469, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37547, CurrentID=0, Status='live', Symbol='CG5819', Aliases='Dmel_ 5819, CT18212, Dmel\\CG5819, tartan/capricious-like', description='CG5819 gene product from transcript CG5819-RC', other_designations='CG5819-PA|CG5819-PB|CG5819-PC', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22293400, end_position_on_the_genomic_accession=22296955, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37548, CurrentID=0, Status='live', Symbol='wdp', Aliases='Dmel_CG3413, CG30504, CG3413, CT11415, CT39092, Dmel\\CG3413', description='windpipe', other_designations='CG3413-PA|CG3413-PB|CG3413-PC|CG3413-PD|CG3413-PE|connectin-like|wdp-PA|wdp-PB|wdp-PC|wdp-PD|wdp-PE', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22298214, end_position_on_the_genomic_accession=22311937, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37549, CurrentID=0, Status='live', Symbol='Gp150', Aliases='Dmel_CG5820, CG5820, CT18251, Dmel\\CG5820, GP150, anon-EST:fe2A1, gp150, l(2)52A, p150', description='CG5820 gene product from transcript CG5820-RC', other_designations='CG5820-PA|CG5820-PB|CG5820-PC|CG5820-PD|CG5820-PE|CG5820-PF|Gp150-PA|Gp150-PB|Gp150-PC|Gp150-PD|Gp150-PE|Gp150-PF|anon-fast-evolving-2A1|glycoprotein 150', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22317662, end_position_on_the_genomic_accession=22329416, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37551, CurrentID=0, Status='live', Symbol='T3dh', Aliases='Dmel_CG3425, CG3425, Dmel\\CG3425', description='Type III alcohol dehydrogenase', other_designations='CG3425-PA|T3dh-PA|type III alcohol dehydrogenase', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22329726, end_position_on_the_genomic_accession=22332072, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37552, CurrentID=0, Status='live', Symbol='Liprin-gamma', Aliases='Dmel_CG11206, CG11206, Dmel\\CG11206, Liprin CG11206, liprin gamma', description='CG11206 gene product from transcript CG11206-RA', other_designations='CG11206-PA|CG11206-PB|CG11206-PC|CG11206-PD|CG11206-PE|Liprin-gamma-PA|Liprin-gamma-PB|Liprin-gamma-PC|Liprin-gamma-PD|Liprin-gamma-PE|liprin gamma|liprin-gamma', map_location='58D3-58D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22333500, end_position_on_the_genomic_accession=22373109, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  37553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37553, CurrentID=0, Status='live', Symbol='CG11298', Aliases='Dmel_ BcDNA:AT10824, Dmel\\CG11298', description='CG11298 gene product from transcript CG11298-RA', other_designations='CG11298-PA|uncharacterized protein', map_location='58D4-58D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22348307, end_position_on_the_genomic_accession=22348875, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37554, CurrentID=0, Status='live', Symbol='Rtf1', Aliases='Dmel_CG10955, CG10955, Dmel\\CG10955, d', description='CG10955 gene product from transcript CG10955-RB', other_designations='CG10955-PA|CG10955-PB|Rtf1-PA|Rtf1-PB', map_location='58D4-58D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22375090, end_position_on_the_genomic_accession=22378484, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37555, CurrentID=0, Status='live', Symbol='wrapper', Aliases='Dmel_CG10382, CG10382, CT19059, Dmel\\CG10382, Wr, Wrapper', description='CG10382 gene product from transcript CG10382-RA', other_designations='CG10382-PA|wrapper-PA', map_location='58D4-58D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22379637, end_position_on_the_genomic_accession=22382675, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37556, CurrentID=0, Status='live', Symbol='CG13506', Aliases='Dmel_ 13506, CT32874, Dmel\\CG13506', description='CG13506 gene product from transcript CG13506-RB', other_designations='CG13506-PA|CG13506-PB|uncharacterized protein', map_location='58D4-58D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22383134, end_position_on_the_genomic_accession=22394015, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37557, CurrentID=0, Status='live', Symbol='babos', Aliases='Dmel_CG3624, CG3624, CT12183, Dmel\\CG3624', description='CG3624 gene product from transcript CG3624-RB', other_designations='CG3624-PA|CG3624-PB|babos-PA|babos-PB|cell adhesion protein', map_location='58D4-58D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22393876, end_position_on_the_genomic_accession=22402059, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37558, CurrentID=0, Status='live', Symbol='CG6044', Aliases='Dmel_ BcDNA:RH44966, Dmel\\CG6044', description='CG6044 gene product from transcript CG6044-RG', other_designations='CG6044-PA|CG6044-PG', map_location='58D4-58D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22404177, end_position_on_the_genomic_accession=22408540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37559, CurrentID=0, Status='live', Symbol='qkr58E-3', Aliases='Dmel_CG3584, CG3584, Dmel\\CG3584, KEP1, QKR58E-3, kep1, qkr58e-3, qrk58E-3', description='quaking related 58E-3', other_designations='CG3584-PB|CG3584-PC|CG3584-PD|KH domain encompassing protein 1|qkr58E-3-PB|qkr58E-3-PC|qkr58E-3-PD', map_location='58D7-58D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22409208, end_position_on_the_genomic_accession=22411545, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37560, CurrentID=0, Status='live', Symbol='Mes4', Aliases='Dmel_CG11301, 11301, BcDNA:GM21739, CG11301, Dmel\\CG11301, MES4, NF-Yc', description='Mesoderm-expressed 4', other_designations='CG11301-PA|CG11301-PB|Mes4-PA|Mes4-PB|NF-YC-like', map_location='58D8-58D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22411789, end_position_on_the_genomic_accession=22412582, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37561, CurrentID=0, Status='live', Symbol='qkr58E-1', Aliases='Dmel_CG3613, CG3613, Dmel\\CG3613, QKR58E-1, qkr-58-E, qrk58E-1', description='quaking related 58E-1', other_designations='CG3613-PA|CG3613-PC|CG3613-PD|KH-domain|qkr58E-1-PA|qkr58E-1-PC|qkr58E-1-PD', map_location='58D8-58D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22412556, end_position_on_the_genomic_accession=22415518, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37562, CurrentID=0, Status='live', Symbol='qkr58E-2', Aliases='Dmel_CG5821, CG5821, Dmel\\CG5821, qkr58e-2, qrk58E-2', description='quaking related 58E-2', other_designations='CG5821-PA|CG5821-PB|CG5821-PC|qkr58E-2-PA|qkr58E-2-PB|qkr58E-2-PC', map_location='58D8-58E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22415560, end_position_on_the_genomic_accession=22419001, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37563, CurrentID=0, Status='live', Symbol='mRpS29', Aliases='Dmel_CG3633, CG3633, DAP3, Dmel\\CG3633, MRP-S29, anon-WO0118547.102', description='mitochondrial ribosomal protein S29', other_designations='CG3633-PA|mRpS29-PA', map_location='58E1-58E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22418937, end_position_on_the_genomic_accession=22420376, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37564, CurrentID=0, Status='live', Symbol='rad50', Aliases='Dmel_CG6339, 6339, CG6339, Dmel\\CG6339, Dm RAD50, Rad50', description='CG6339 gene product from transcript CG6339-RE', other_designations='CG6339-PD|CG6339-PE|rad50-PD|rad50-PE', map_location='58E1-58E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22421013, end_position_on_the_genomic_accession=22425482, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37565, CurrentID=0, Status='live', Symbol='CG10344', Aliases='Dmel_ Dmel\\CG10344', description='CG10344 gene product from transcript CG10344-RA', other_designations='CG10344-PA|CG10344-PB|uncharacterized protein', map_location='58E1-58E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22425733, end_position_on_the_genomic_accession=22426977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37566, CurrentID=0, Status='live', Symbol='ppk12', Aliases='Dmel_CG10972, CG10972, Dmel\\CG10972, PPK12', description='pickpocket 12', other_designations='CG10972-PA|ppk12-PA', map_location='58E1-58E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22427479, end_position_on_the_genomic_accession=22429584, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39029, CurrentID=0, Status='live', Symbol='CG5644', Aliases='Dmel_ Dmel\\CG5644, anon-EST:Posey104', description='CG5644 gene product from transcript CG5644-RA', other_designations='CG5644-PA', map_location='66E6-66E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8977395, end_position_on_the_genomic_accession=8979880, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246560, CurrentID=0, Status='live', Symbol='UQCR-11L', Aliases='Dmel_CG30354, CG30354, Dmel\\CG30354, NEST:bs01g03, UCRH', description='Ubiquinol-cytochrome c reductase 11 kDa subunit-like', other_designations='CG30354-PA|CG30354-PB|UQCR-11L-PA|UQCR-11L-PB|Ubiquinol-cytochrome C reductase complex 11 kDa protein|Ubiquinol-cytochrome c reductase 11 kDa protein', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8677219, end_position_on_the_genomic_accession=8677661, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37570, CurrentID=0, Status='live', Symbol='CG4554', Aliases='Dmel_ Dmel\\CG4554, GM01751p', description='CG4554 gene product from transcript CG4554-RA', other_designations='CG4554-PA', map_location='58E3-58E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22483303, end_position_on_the_genomic_accession=22492033, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37571, CurrentID=0, Status='live', Symbol='CG4610', Aliases='Dmel_ Dmel\\CG4610', description='CG4610 gene product from transcript CG4610-RA', other_designations='CG4610-PA', map_location='58E3-58E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22492197, end_position_on_the_genomic_accession=22494295, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37572, CurrentID=0, Status='live', Symbol='CG6018', Aliases='Dmel_ DmalphaE1-O, Dmel\\CG6018, alphaE1-O', description='CG6018 gene product from transcript CG6018-RA', other_designations='CG6018-PA|orphan alpha-esterase', map_location='58E8-58E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22550094, end_position_on_the_genomic_accession=22552113, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37573, CurrentID=0, Status='live', Symbol='px', Aliases='Dmel_CG4444, CG4444, Dmel\\CG4444, l(2)k08134, l(2)k08316', description='plexus', other_designations='CG4444-PA|CG4444-PB|CG4444-PD|CG4444-PE|px-PA|px-PB|px-PD|px-PE', map_location='2-100.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22494297, end_position_on_the_genomic_accession=22562065, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  37574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37574, CurrentID=0, Status='live', Symbol='CG11362', Aliases='Dmel_ Dmel\\CG11362', description='CG11362 gene product from transcript CG11362-RA', other_designations='CG11362-PA|uncharacterized protein', map_location='58E8-58E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22562307, end_position_on_the_genomic_accession=22563131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37575, CurrentID=0, Status='live', Symbol='dnr1', Aliases='Dmel_CG12489, CG12489, DNR1, Dmel\\CG12489, Dnr1', description='defense repressor 1', other_designations='CG12489-PA|CG12489-PB|defense repressor I|dnr1-PA|dnr1-PB', map_location='58E9-58F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22563152, end_position_on_the_genomic_accession=22592964, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37576, CurrentID=0, Status='live', Symbol='CG3927', Aliases='Dmel_ BcDNA:AT29209, Dmel\\CG3927, cg3927', description='CG3927 gene product from transcript CG3927-RA', other_designations='CG3927-PA', map_location='58F1-58F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22594677, end_position_on_the_genomic_accession=22596348, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37577, CurrentID=0, Status='live', Symbol='nsr', Aliases='Dmel_CG3875, CG3875, Dmel\\CG3875, cg3875', description='novel spermatogenesis regulator', other_designations='CG3875-PA|CG3875-PB|nsr-PA|nsr-PB', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22650556, end_position_on_the_genomic_accession=22653595, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37578, CurrentID=0, Status='live', Symbol='CG4269', Aliases='Dmel_ Dmel\\CG4269, cg4269', description='CG4269 gene product from transcript CG4269-RB', other_designations='CG4269-PA|CG4269-PB|CG4269-PC', map_location='58F1-58F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22597245, end_position_on_the_genomic_accession=22600695, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37579, CurrentID=0, Status='live', Symbol='CG4294', Aliases='Dmel_ Dmel\\CG4294, FBgn0034742', description='CG4294 gene product from transcript CG4294-RB', other_designations='CG4294-PB|CG4294-PC', map_location='58F1-58F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22600405, end_position_on_the_genomic_accession=22605667, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37580, CurrentID=0, Status='live', Symbol='RpS16', Aliases='Dmel_CG4046, CG4046, Dmel\\CG4046, Rp S16, S16, bs32d05.y1', description='Ribosomal protein S16', other_designations='CG4046-PA|CG4046-PB|RpS16-PA|RpS16-PB|ribosomal protein S16', map_location='58F1-58F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22605745, end_position_on_the_genomic_accession=22607995, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37581, CurrentID=0, Status='live', Symbol='Vps20', Aliases='Dmel_CG4071, CG4071, DmVPS20, Dmel\\CG4071, ESCRT-III, anon-EST:Posey40, vps20', description='Vacuolar protein sorting 20', other_designations='CG4071-PA|CG4071-PC|Vps20-PA|Vps20-PC|vacuolar protein sorting 20', map_location='58F1-58F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22608310, end_position_on_the_genomic_accession=22609419, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37582, CurrentID=0, Status='live', Symbol='CG4329', Aliases='Dmel_ Dmel\\CG4329, anon-WO0140519.244', description='CG4329 gene product from transcript CG4329-RA', other_designations='CG4329-PA', map_location='58F2-58F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22609334, end_position_on_the_genomic_accession=22613376, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37583, CurrentID=0, Status='live', Symbol='CG33143', Aliases='Dmel_ 33143, CG13509, CG4232, Dmel\\CG33143, mdcds_16487', description='CG33143 gene product from transcript CG33143-RC', other_designations='CG33143-PC|CG33143-PD|CG33143-PE|CG33143-PF|uncharacterized protein', map_location='58F2-58F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22614328, end_position_on_the_genomic_accession=22634084, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37585, CurrentID=0, Status='live', Symbol='CG17807', Aliases='Dmel_ Dmel\\CG17807', description='CG17807 gene product from transcript CG17807-RA', other_designations='CG17807-PA|uncharacterized protein', map_location='58F2-58F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22634019, end_position_on_the_genomic_accession=22635942, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37586, CurrentID=0, Status='live', Symbol='Cdk9', Aliases='Dmel_CG5179, CDK9, CG5179, Dmel\\CG5179, P-TEF, P-TEFb, PTefb, cdk9', description='Cyclin-dependent kinase 9', other_designations='CG5179-PA|Cdk9-PA|Cyclin-dependent kinase9 2.7|p43 PITALRE|positive transcription elongation factor b|positive transcription elongation factor b, small subunit', map_location='58F2-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22636063, end_position_on_the_genomic_accession=22637727, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37587, CurrentID=0, Status='live', Symbol='bonsai', Aliases='Dmel_CG4207, CG4207, DmMRPS15, Dmel\\CG4207, MRP-S15, mRpS15', description='CG4207 gene product from transcript CG4207-RA', other_designations='CG4207-PA|bonsai-PA', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22637770, end_position_on_the_genomic_accession=22638868, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37588, CurrentID=0, Status='live', Symbol='CG3732', Aliases='Dmel_ Dmel\\CG3732', description='CG3732 gene product from transcript CG3732-RA', other_designations='CG3732-PA', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22638807, end_position_on_the_genomic_accession=22640201, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37589, CurrentID=0, Status='live', Symbol='RpS24', Aliases='Dmel_CG3751, BcDNA:RE59324, CG3751, Dmel\\CG3751, Rp S24, S24, l(2)SH2 2053, l(2)SH2053', description='Ribosomal protein S24', other_designations='40S ribosomal protein S24|CG3751-PA|RpS24-PA|lethal (2) SH2053|ribosomal protein S24', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22640526, end_position_on_the_genomic_accession=22641419, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37590, CurrentID=0, Status='live', Symbol='Ugt58Fa', Aliases='Dmel_CG4414, AC 004377A, CG4414, Dmel\\CG4414', description='CG4414 gene product from transcript CG4414-RB', other_designations='CG4414-PA|CG4414-PB|Ugt58Fa-PA|Ugt58Fa-PB|lectin-58Fg', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22641786, end_position_on_the_genomic_accession=22643917, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37591, CurrentID=0, Status='live', Symbol='CG2852', Aliases='Dmel_ Dmel\\CG2852', description='CG2852 gene product from transcript CG2852-RD', other_designations='CG2852-PA|CG2852-PD', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22644646, end_position_on_the_genomic_accession=22646344, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37592, CurrentID=0, Status='live', Symbol='CG30195', Aliases='Dmel_ CG13513, Dmel\\CG30195', description='CG30195 gene product from transcript CG30195-RA', other_designations='CG30195-PA|uncharacterized protein', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22646616, end_position_on_the_genomic_accession=22647245, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37593, CurrentID=0, Status='live', Symbol='CG3746', Aliases='Dmel_ Dmel\\CG3746', description='CG3746 gene product from transcript CG3746-RB', other_designations='CG3746-PA|CG3746-PB', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22648995, end_position_on_the_genomic_accession=22649750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37594, CurrentID=0, Status='live', Symbol='Cyp6d2', Aliases='Dmel_CG4373, 6d2, CG4373, Dmel\\CG4373', description='CG4373 gene product from transcript CG4373-RA', other_designations='CG4373-PA|Cyp6d2-PA|cytochrome P450', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22652573, end_position_on_the_genomic_accession=22654530, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37595, CurrentID=0, Status='live', Symbol='CG30196', Aliases='Dmel_ CG13514, Dmel\\CG30196', description='CG30196 gene product from transcript CG30196-RA', other_designations='CG30196-PA|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22654805, end_position_on_the_genomic_accession=22655392, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37596, CurrentID=0, Status='live', Symbol='CG13510', Aliases='Dmel_ Dmel\\CG13510', description='CG13510 gene product from transcript CG13510-RC', other_designations='CG13510-PA|CG13510-PB|CG13510-PC|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22657901, end_position_on_the_genomic_accession=22661389, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37597, CurrentID=0, Status='live', Symbol='CG13511', Aliases='Dmel_ Dmel\\CG13511', description='CG13511 gene product from transcript CG13511-RB', other_designations='CG13511-PA|CG13511-PB|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22657901, end_position_on_the_genomic_accession=22661389, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37598, CurrentID=0, Status='live', Symbol='CG42565', Aliases='Dmel_ CG13512, Dmel\\CG42565, Dmel_CG13512', description='CG42565 gene product from transcript CG42565-RB', other_designations='CG42565-PA|CG42565-PB|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22657901, end_position_on_the_genomic_accession=22661389, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37599, CurrentID=0, Status='live', Symbol='CG4250', Aliases='Dmel_ BcDNA:RH52355, CT13934, Dmel\\CG4250', description='CG4250 gene product from transcript CG4250-RB', other_designations='CG4250-PA|CG4250-PB', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22663692, end_position_on_the_genomic_accession=22664474, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37600, CurrentID=0, Status='live', Symbol='CG30269', Aliases='Dmel_ CG13515, Dmel\\CG30269', description='CG30269 gene product from transcript CG30269-RB', other_designations='CG30269-PA|CG30269-PB|uncharacterized protein', map_location='58F4-58F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22667155, end_position_on_the_genomic_accession=22668848, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37601, CurrentID=0, Status='live', Symbol='RYBP', Aliases='Dmel_CG12190, CG12190, Dmel\\CG12190, dRYBP', description='Ring and YY1 Binding Protein', other_designations='CG12190-PA|CG12190-PB|RYBP-PA|RYBP-PB|ring and YY1 binding protein', map_location='58F7-58F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22670587, end_position_on_the_genomic_accession=22673212, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37602, CurrentID=0, Status='live', Symbol='Ppa', Aliases='Dmel_CG9952, CG9952, Dmel\\CG9952, FBXL14, I-55, PPA, ppa', description='Partner of paired', other_designations='CG9952-PA|CG9952-PB|CG9952-PC|Ppa-PA|Ppa-PB|Ppa-PC|partner of paired', map_location='59A1-59A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22684955, end_position_on_the_genomic_accession=22688987, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37603, CurrentID=0, Status='live', Symbol='robo1', Aliases='Dmel_CG13521, BEST:GM07780, CG13521, CT32892, D-Robo1, Dmel\\CG13521, ROBO, Robo, Robo-1, Robo1, dRobo, dRobo1, roboS, robo1', description='roundabout 1', other_designations='CG13521-PA|CG13521-PC|robo|robo1-PA|robo1-PC|roundabout', map_location='59A2-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22692524, end_position_on_the_genomic_accession=22700970, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  37605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37605, CurrentID=0, Status='live', Symbol='Obp59a', Aliases='Dmel_CG13517, 59a, CG13517, DmelObp59a, Dmel\\CG13517, Obp58a', description='Odorant-binding protein 59a', other_designations='CG13517-PB|Obp59a-PB', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22702695, end_position_on_the_genomic_accession=22703786, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37606, CurrentID=0, Status='live', Symbol='CG30259', Aliases='Dmel_ CG13522, CG13523, Dmel\\CG30259', description='CG30259 gene product from transcript CG30259-RA', other_designations='CG30259-PA|uncharacterized protein', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22701345, end_position_on_the_genomic_accession=22704447, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37607, CurrentID=0, Status='live', Symbol='Obp58b', Aliases='Dmel_CG13518, C13518, CG13518, DmelObp58b, Dmel\\CG13518', description='Odorant-binding protein 58b', other_designations='CG13518-PA|Obp58b-PA|odorant binding protein 58b', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22704704, end_position_on_the_genomic_accession=22705639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37608, CurrentID=0, Status='live', Symbol='Obp58c', Aliases='Dmel_CG13524, AAF46892, CG13524, DmelObp58c, Dmel\\CG13524', description='Odorant-binding protein 58c', other_designations='CG13524-PA|Obp58c-PA', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22705872, end_position_on_the_genomic_accession=22706678, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39036, CurrentID=0, Status='live', Symbol='Doc3', Aliases='Dmel_CG5093, CG5093, Dm-DOC3, Dmel\\CG5093, Doc, doc3', description='Dorsocross3', other_designations='CG5093-PA|Doc3-PA|dorsocross', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9005196, end_position_on_the_genomic_accession=9007249, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37610, CurrentID=0, Status='live', Symbol='CG30268', Aliases='Dmel_ CG13520, Dmel\\CG30268', description='CG30268 gene product from transcript CG30268-RB', other_designations='CG30268-PB|uncharacterized protein', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22712701, end_position_on_the_genomic_accession=22717247, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  33575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=33575, CurrentID=0, Status='live', Symbol='CG8851', Aliases='Dmel_ Dmel\\CG8851', description='CG8851 gene product from transcript CG8851-RC', other_designations='CG8851-PA|CG8851-PC', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3519677, end_position_on_the_genomic_accession=3522555, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37613, CurrentID=0, Status='live', Symbol='CG13526', Aliases='Dmel_ BcDNA:AT13807, Dmel\\CG13526', description='CG13526 gene product from transcript CG13526-RA', other_designations='CG13526-PA|uncharacterized protein', map_location='59A4-59A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22755219, end_position_on_the_genomic_accession=22755896, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37614, CurrentID=0, Status='live', Symbol='CG42260', Aliases='Dmel_ CG13530, CG3536, Dmel\\CG42260, Dmel_CG13530, Dmel_CG3536, dmCNG4, lincRNA.378', description='CG42260 gene product from transcript CG42260-RB', other_designations='CG42260-PA|CG42260-PB|CG42260-PD|uncharacterized protein', map_location='59B1-59B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22740999, end_position_on_the_genomic_accession=22798289, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  37615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37615, CurrentID=0, Status='live', Symbol='CG13527', Aliases='Dmel_ Dmel\\CG13527, SPH182', description='CG13527 gene product from transcript CG13527-RB', other_designations='CG13527-PA|CG13527-PB|uncharacterized protein', map_location='59A4-59A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22763706, end_position_on_the_genomic_accession=22765422, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37617, CurrentID=0, Status='live', Symbol='blw', Aliases='Dmel_CG3612, A2(2nd)27, ATP syn-alpha, ATP5a, ATPA, ATPA_DROME, ATPsyn-alpha, ATPsynalpha, Blw, C-V-alpha, CG3612, CValpha, CxV, Dmel\\CG3612, l(2)01410, l(2)03972, l(2)59ABa, l(2)k00212, ms(2)01504, ms(2)Pry58F', description='bellwether', other_designations='ATP SYNTHASE ALPHA|ATP synthase|ATP synthase alpha chain|ATP synthase alpha subunit of complex V|ATP synthase subunit alpha|ATP synthase subunits alpha|ATP synthase, alpha subunit|ATP synthase, subunit alpha|ATP synthetase alpha chain|CG3612-PA|COMPLEX V: alpha chain|F0/F1 ATP synthase alpha|alpha subunit of ATP synthase|alpha subunit of the mitochondrial F[[1]]F[[0]] ATPase|alpha-ATP synthase|alpha-subunit of ATP synthase|bellweather|blw-PA|colibri|complex V alpha|complex valpha|complex valpha subunit|mitochondrial ATP synthase alpha subunit|mitochondrial ATP synthase subunit B|mitochondrial ATP synthase subunit alpha of complex V|mitochondrial ATPase complex V-subunit alpha', map_location='59B1-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22799099, end_position_on_the_genomic_accession=22802180, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37618, CurrentID=0, Status='live', Symbol='CycB', Aliases='Dmel_CG3510, 2g, 3510, CG3510, CYC-B, CYCB, Cyc B, Cycb, DmcycB, Dmel\\CG3510, chr2R:18312808..18312907, cyc B, cyc-B, cycB, cyclin B, cyclinB, mAbF2F4', description='Cyclin B', other_designations='CG3510-PA|CG3510-PB|CG3510-PC|CG3510-PE|CycB-PA|CycB-PB|CycB-PC|CycB-PE|Cyclin-B|cyclin B|cyclin B1|cyclinB|dCyclinB', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22803521, end_position_on_the_genomic_accession=22806948, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37619, CurrentID=0, Status='live', Symbol='stl', Aliases='Dmel_CG3622, BcDNA:GM08694, CG3622, Dmel\\CG3622, fs(2)A16', description='stall', other_designations='CG3622-PC|CG3622-PE|female sterile(2)A16|stl-PC|stl-PE', map_location='2-102 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22807860, end_position_on_the_genomic_accession=22815993, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  37621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37621, CurrentID=0, Status='live', Symbol='CG42284', Aliases='Dmel_ CG13528, CG13529, CG30264, CG30266, Dmel\\CG42284, Dmel_CG30264, Dmel_CG30266', description='CG42284 gene product from transcript CG42284-RD', other_designations='CG42284-PD|CG42284-PE|uncharacterized protein', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22821430, end_position_on_the_genomic_accession=22828903, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37622, CurrentID=0, Status='live', Symbol='CG30265', Aliases='Dmel_ CG11682, Dmel\\CG30265', description='CG30265 gene product from transcript CG30265-RA', other_designations='CG30265-PA|CG30265-PB|uncharacterized protein', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22832456, end_position_on_the_genomic_accession=22834580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37623, CurrentID=0, Status='live', Symbol='CG12490', Aliases='Dmel_ Dmel\\CG12490', description='CG12490 gene product from transcript CG12490-RB', other_designations='CG12490-PB|uncharacterized protein', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22834647, end_position_on_the_genomic_accession=22836701, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37624, CurrentID=0, Status='live', Symbol='CG9825', Aliases='Dmel_ Dmel\\CG9825', description='CG9825 gene product from transcript CG9825-RA', other_designations='CG9825-PA', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22837361, end_position_on_the_genomic_accession=22839291, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37625, CurrentID=0, Status='live', Symbol='CG9826', Aliases='Dmel_ Dmel\\CG9826', description='CG9826 gene product from transcript CG9826-RC', other_designations='CG9826-PA|CG9826-PC', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22840739, end_position_on_the_genomic_accession=22842977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37626, CurrentID=0, Status='live', Symbol='CG3649', Aliases='Dmel_ Dmel\\CG3649', description='CG3649 gene product from transcript CG3649-RB', other_designations='CG3649-PB', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22845977, end_position_on_the_genomic_accession=22847911, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37627, CurrentID=0, Status='live', Symbol='CG13531', Aliases='Dmel_ Dmel\\CG13531', description='CG13531 gene product from transcript CG13531-RC', other_designations='CG13531-PB|CG13531-PC|uncharacterized protein', map_location='59B2-59B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22847832, end_position_on_the_genomic_accession=22853827, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37628, CurrentID=0, Status='live', Symbol='RpL23', Aliases='Dmel_CG3661, CG3661, Dmel\\CG3661, L17, L23, RL17, RPL17A, RpL17A, Rpl17A, TA01212p, rpl23', description='Ribosomal protein L23', other_designations='CG3661-PA|RpL23-PA|ribosomal protein L17A|ribosomal protein L23', map_location='59B3-59B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22854383, end_position_on_the_genomic_accession=22855694, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37629, CurrentID=0, Status='live', Symbol='inaD', Aliases='Dmel_CG3504, CG3504, Dmel\\CG3504, INAD, Ina-D, InaD', description='inactivation no afterpotential D', other_designations='CG3504-PA|CG3504-PB|INAD|inaD-PA|inaD-PB|inactivation no after-potential D|inactivation, no after-depolarization|inactivation-no after potential D|inactivation-no-afterpotential D|inactivation-no-afterpotential-D', map_location='2-101 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22855688, end_position_on_the_genomic_accession=22858794, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37631, CurrentID=0, Status='live', Symbol='fd59A', Aliases='Dmel_CG3668, CG3668, Dmel\\CG3668, Dmfd3, FD3, FD59A, Fd59a, fd3, fd59a, foxd', description='forkhead domain 59A', other_designations='CG3668-PA|fd59A-PA', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22868159, end_position_on_the_genomic_accession=22870706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37632, CurrentID=0, Status='live', Symbol='CG13532', Aliases='Dmel_ CT32907, Dmel\\CG13532', description='CG13532 gene product from transcript CG13532-RA', other_designations='CG13532-PA|uncharacterized protein', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22871190, end_position_on_the_genomic_accession=22874427, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37633, CurrentID=0, Status='live', Symbol='PIP5K59B', Aliases='Dmel_CG3682, CG17281, CG3682, Dmel\\CG3682, PI4P5K, PIP5K 59B, dPIP5K', description='Phosphatidylinositol 4-phosphate 5-kinase at 59B', other_designations='CG3682-PA|CG3682-PE|CG3682-PF|CG3682-PH|CG3682-PI|PIP5K59B-PA|PIP5K59B-PE|PIP5K59B-PF|PIP5K59B-PH|PIP5K59B-PI|phosphatidylinositol 4-phosphate 5-kinase at 59B', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22874769, end_position_on_the_genomic_accession=22881436, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  37635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37635, CurrentID=0, Status='live', Symbol='EMC8-9', Aliases='Dmel_CG3501, BcDNA:SD01117, CG3501, Dmel\\CG3501, EMC8/9', description='ER Membrane protein Complex 8-9', other_designations='CG3501-PA|EMC8-9-PA|ER membrane protein complex 8-9', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22881322, end_position_on_the_genomic_accession=22882690, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37636, CurrentID=0, Status='live', Symbol='CG3499', Aliases='Dmel_ DmCG3499, Dmel\\CG3499', description='CG3499 gene product from transcript CG3499-RD', other_designations='CG3499-PB|CG3499-PC|CG3499-PD', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22882747, end_position_on_the_genomic_accession=22886247, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37637, CurrentID=0, Status='live', Symbol='asrij', Aliases='Dmel_CG13533, Asrij, CG13533, Dm Dmel\\CG13533, FBgn 34793, arj', description='CG13533 gene product from transcript CG13533-RA', other_designations='CG13533-PA|CG13533-PB|asrij-PA|asrij-PB', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22886398, end_position_on_the_genomic_accession=22887578, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37638, CurrentID=0, Status='live', Symbol='Gmer', Aliases='Dmel_CG3495, CG3495, DmGMER, Dmel\\CG3495, GER, gmer, l(2)SH1931', description='GDP-4-keto-6-deoxy-D-mannose 3,5-epimerase/4-reductase', other_designations='CG3495-PA|GDP-keto-6-deoxymannose 3,5-epimerase/4-reductase|Gmer-PA|lethal (2) SH1931', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22887574, end_position_on_the_genomic_accession=22888857, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37639, CurrentID=0, Status='live', Symbol='MED23', Aliases='Dmel_CG3695, CG3695, Dmel\\CG3695, Gal11/SUR2, Med23, Trap150, Trap150beta, dSUR2, dTRAP150beta', description='Mediator complex subunit 23', other_designations='CG3695-PA|MED23-PA|mediator complex subunit 23', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22888928, end_position_on_the_genomic_accession=22894425, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  37640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37640, CurrentID=0, Status='live', Symbol='CG3700', Aliases='Dmel_ Dmel\\CG3700, SP48, c-SP48', description='CG3700 gene product from transcript CG3700-RA', other_designations='CG3700-PA', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22894435, end_position_on_the_genomic_accession=22895745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37641, CurrentID=0, Status='live', Symbol='nahoda', Aliases='Dmel_CG12781, CG12781, CT36793, Dmel\\CG12781', description='CG12781 gene product from transcript CG12781-RC', other_designations='CG12781-PA|CG12781-PB|CG12781-PC|nahoda-PA|nahoda-PB|nahoda-PC', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22896174, end_position_on_the_genomic_accession=22916719, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37643, CurrentID=0, Status='live', Symbol='Reep1', Aliases='Dmel_CG42678, BcDNA:LD42159, BcDNA:RE29641, CG30190, CG30193, CG42678, CG9848, D-reep1, Dmel\\CG42678, Dmel_CG30190, Dmel_CG30193, anon-WO0118547.113, reep1', description='REEP1 homolog', other_designations='CG42678-PD|CG42678-PE|CG42678-PG|CG42678-PH|CG42678-PI|CG42678-PJ|CG42678-PK|CG42678-PL|CG42678-PO|CG42678-PP|CG42678-PQ|CG42678-PR|CG42678-PS|Reep1-PD|Reep1-PE|Reep1-PG|Reep1-PH|Reep1-PI|Reep1-PJ|Reep1-PK|Reep1-PL|Reep1-PO|Reep1-PP|Reep1-PQ|Reep1-PR|Reep1-PS', map_location='59B6-59B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22925107, end_position_on_the_genomic_accession=22932044, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37644, CurrentID=0, Status='live', Symbol='CG3788', Aliases='Dmel_ Dmel\\CG3788', description='CG3788 gene product from transcript CG3788-RB', other_designations='CG3788-PA|CG3788-PB', map_location='59B6-59B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22932035, end_position_on_the_genomic_accession=22934077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37646, CurrentID=0, Status='live', Symbol='CG3800', Aliases='Dmel_ BEST:GM06241, CG11716, Dmel\\CG3800, anon-EST:fe1B4, anon-EST:fe2B6', description='CG3800 gene product from transcript CG3800-RA', other_designations='CG3800-PA|CG3800-PB|anon-fast-evolving-1B4|anon-fast-evolving-2B6', map_location='59B6-59B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22934569, end_position_on_the_genomic_accession=22937077, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37647, CurrentID=0, Status='live', Symbol='CG9849', Aliases='Dmel_ Dmel\\CG9849', description='CG9849 gene product from transcript CG9849-RA', other_designations='CG9849-PA|CG9849-PB', map_location='59B6-59B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22937279, end_position_on_the_genomic_accession=22939025, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37648, CurrentID=0, Status='live', Symbol='CG3831', Aliases='Dmel_ Dmel\\CG3831', description='CG3831 gene product from transcript CG3831-RC', other_designations='CG3831-PA|CG3831-PC', map_location='59B6-59B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22938987, end_position_on_the_genomic_accession=22941169, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37650, CurrentID=0, Status='live', Symbol='CG32833', Aliases='Dmel_ BcDNA:GH11992, CG30331, CG9898, Dmel\\CG32833, SPH196', description='CG32833 gene product from transcript CG32833-RA', other_designations='CG32833-PA|uncharacterized protein', map_location='59C1-59C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22984273, end_position_on_the_genomic_accession=22985334, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37651, CurrentID=0, Status='live', Symbol='CG9897', Aliases='Dmel_ Dmel\\CG9897, SPH195', description='CG9897 gene product from transcript CG9897-RA', other_designations='CG9897-PA|CG9897-PB', map_location='59C1-59C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22985420, end_position_on_the_genomic_accession=22986339, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37652, CurrentID=0, Status='live', Symbol='CG9896', Aliases='Dmel_ Dmel\\CG9896', description='CG9896 gene product from transcript CG9896-RA', other_designations='CG9896-PA', map_location='59C1-59C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22993669, end_position_on_the_genomic_accession=22998884, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37654, CurrentID=0, Status='live', Symbol='CG42741', Aliases='Dmel_ CG34209, CG9895, Dmel\\CG42741, Dmel_CG34209, Dmel_CG9895', description='CG42741 gene product from transcript CG42741-RA', other_designations='CG42741-PA|uncharacterized protein', map_location='59C2-59C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23011078, end_position_on_the_genomic_accession=23042943, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37655, CurrentID=0, Status='live', Symbol='twi', Aliases='Dmel_CG2956, CG2956, DTwist, Dmel\\CG2956, EC9, TWI, Tw, Twi, bHLHb17, dip5', description='twist', other_designations='CG2956-PA|CG2956-PB|CG2956-PC|twi-PA|twi-PB|twi-PC', map_location='2-100 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22985374, end_position_on_the_genomic_accession=23048325, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39044, CurrentID=0, Status='live', Symbol='CG4942', Aliases='Dmel_ Dmel\\CG4942', description='CG4942 gene product from transcript CG4942-RA', other_designations='CG4942-PA', map_location='66F4-66F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9073105, end_position_on_the_genomic_accession=9074467, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37659, CurrentID=0, Status='live', Symbol='CG9890', Aliases='Dmel_ Dmel\\CG9890, anon-EST:Posey169, ld46465', description='CG9890 gene product from transcript CG9890-RA', other_designations='CG9890-PA', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23062747, end_position_on_the_genomic_accession=23064260, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37661, CurrentID=0, Status='live', Symbol='l(2)k09913', Aliases='Dmel_CG3082, 1(2)k09913, CG3082, Dmel\\CG3082', description='lethal (2) k09913', other_designations='CG3082-PA|CG3082-PB|CG3082-PC|CG3082-PD|CG3082-PE|CG3082-PF|CG3082-PG|CG3082-PH|CG3082-PI|l(2)k09913-PA|l(2)k09913-PB|l(2)k09913-PC|l(2)k09913-PD|l(2)k09913-PE|l(2)k09913-PF|l(2)k09913-PG|l(2)k09913-PH|l(2)k09913-PI', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23067196, end_position_on_the_genomic_accession=23071875, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37662, CurrentID=0, Status='live', Symbol='Fib', Aliases='Dmel_CG9888, CG9888, Dmel\\CG9888ri, GCR-6, GCR6, Pen59C5, fib, pen59C5, Fib', description='Fibrillarin', other_designations='CG9888-PA|Fib-PA|fibrillarin', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23072909, end_position_on_the_genomic_accession=23074725, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37663, CurrentID=0, Status='live', Symbol='CG3085', Aliases='Dmel_ Dmel\\CG3085, anon-WO0140519.123', description='CG3085 gene product from transcript CG3085-RB', other_designations='CG3085-PA|CG3085-PB', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23074823, end_position_on_the_genomic_accession=23076857, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37664, CurrentID=0, Status='live', Symbol='Art7', Aliases='Dmel_CG9882, CG9882, DART7, DART8, Dmel\\CG9882, dart7', description='Arginine methyltransferase 7', other_designations='Art7-PB|CG9882-PB|arginine methyltransferase 7', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23076688, end_position_on_the_genomic_accession=23079038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246576, CurrentID=0, Status='live', Symbol='CG30376', Aliases='Dmel_ BcDNA:AT30260, Dmel\\CG30376', description='CG30376 gene product from transcript CG30376-RA', other_designations='CG30376-PA|uncharacterized protein', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8263486, end_position_on_the_genomic_accession=8264426, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37666, CurrentID=0, Status='live', Symbol='CG9877', Aliases='Dmel_ BcDNA:RE26296, Dmel\\CG9877', description='CG9877 gene product from transcript CG9877-RA', other_designations='CG9877-PA', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23085881, end_position_on_the_genomic_accession=23086268, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37667, CurrentID=0, Status='live', Symbol='CG13538', Aliases='Dmel_ Dmel\\CG13538', description='CG13538 gene product from transcript CG13538-RA', other_designations='CG13538-PA|uncharacterized protein', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23087177, end_position_on_the_genomic_accession=23088165, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37668, CurrentID=0, Status='live', Symbol='CG9876', Aliases='Dmel_ Dmel\\CG9876', description='CG9876 gene product from transcript CG9876-RA', other_designations='CG9876-PA', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23090100, end_position_on_the_genomic_accession=23094266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37669, CurrentID=0, Status='live', Symbol='RpL37b', Aliases='Dmel_CG9873, CG9873, Dmel\\CG9873, Rpl37b', description='Ribosomal protein L37b', other_designations='CG9873-PA|RpL37b-PA|ribosomal protein L37b', map_location='59C4-59C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23100057, end_position_on_the_genomic_accession=23100478, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37671, CurrentID=0, Status='live', Symbol='Klp59C', Aliases='Dmel_CG3219, CG3219, DmCG3219, DmKLP59C, DmKlp59C, Dmel\\CG3219, KIF 2B, KIF2A/B/C, KLP59C', description='Kinesin-like protein at 59C', other_designations='CG3219-PA|Klp59C-PA|kinesin-13', map_location='59C4-59C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23123923, end_position_on_the_genomic_accession=23126042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37672, CurrentID=0, Status='live', Symbol='CG3215', Aliases='Dmel_ Dmel\\CG3215, gpdh-2', description='CG3215 gene product from transcript CG3215-RA', other_designations='CG3215-PA', map_location='59C4-59C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23128604, end_position_on_the_genomic_accession=23130189, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37673, CurrentID=0, Status='live', Symbol='CG13544', Aliases='Dmel_ Dmel\\CG13544', description='CG13544 gene product from transcript CG13544-RA', other_designations='CG13544-PA|uncharacterized protein', map_location='59D1-59D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23139080, end_position_on_the_genomic_accession=23140577, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37674, CurrentID=0, Status='live', Symbol='Klp59D', Aliases='Dmel_CG12192, CG12192, DmKlp59D, Dmel\\CG12192, KIF2A/B/C, KLP59D', description='Kinesin-like protein at 59D', other_designations='CG12192-PA|Klp59D-PA', map_location='59D1-59D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23141999, end_position_on_the_genomic_accession=23144497, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37675, CurrentID=0, Status='live', Symbol='CG13545', Aliases='Dmel_ Dmel\\CG13545', description='CG13545 gene product from transcript CG13545-RB', other_designations='CG13545-PB|uncharacterized protein', map_location='59D1-59D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23157577, end_position_on_the_genomic_accession=23158185, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37676, CurrentID=0, Status='live', Symbol='CG9899', Aliases='Dmel_ Dmel\\CG9899', description='CG9899 gene product from transcript CG9899-RA', other_designations='CG9899-PA', map_location='59D1-59D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23153854, end_position_on_the_genomic_accession=23162077, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37679, CurrentID=0, Status='live', Symbol='CG34371', Aliases='Dmel_ 34371, CG13546, CG13547, CG13548, CG30188, CT32921, CT32922, Dmel\\CG34371, Dmel_CG13548, Dmel_CG30188', description='CG34371 gene product from transcript CG34371-RA', other_designations='CG34371-PA|CG34371-PB|uncharacterized protein', map_location='59D1-59D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23163012, end_position_on_the_genomic_accession=23203973, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  37680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37680, CurrentID=0, Status='live', Symbol='CG13539', Aliases='Dmel_ Dmel\\CG13539', description='CG13539 gene product from transcript CG13539-RB', other_designations='CG13539-PA|CG13539-PB|uncharacterized protein', map_location='59D2-59D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23183977, end_position_on_the_genomic_accession=23185288, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37681, CurrentID=0, Status='live', Symbol='LS2', Aliases='Dmel_CG3162, CG3162, Dmel\\CG3162, cg3162', description='Large Subunit 2', other_designations='CG3162-PA|LS2-PA|large subunit 2', map_location='59D2-59D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23210130, end_position_on_the_genomic_accession=23211539, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37682, CurrentID=0, Status='live', Symbol='CG3092', Aliases='Dmel_ Dmel\\CG3092, anon-WO0140519.90', description='CG3092 gene product from transcript CG3092-RA', other_designations='CG3092-PA', map_location='59D3-59D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23232626, end_position_on_the_genomic_accession=23233927, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37683, CurrentID=0, Status='live', Symbol='yip3', Aliases='Dmel_CG13549, BcDNA:AT20887, CG13549, Dmel\\CG13549, NEST:bs18b12, bs18b12.y1', description='yippee interacting protein 3', other_designations='CG13549-PA|CG13549-PB|yip3-PA|yip3-PB', map_location='59D3-59D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23244044, end_position_on_the_genomic_accession=23245399, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37684, CurrentID=0, Status='live', Symbol='RpL22-like', Aliases='Dmel_CG9871, CG9871, Dmel\\CG9871, rpL22-like', description='Ribosomal protein L22-like', other_designations='CG9871-PA|RpL22-like-PA|ribosomal protein L22-like', map_location='59D3-59D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23246610, end_position_on_the_genomic_accession=23247893, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37685, CurrentID=0, Status='live', Symbol='CG12782', Aliases='Dmel_ Dmel\\CG12782', description='CG12782 gene product from transcript CG12782-RA', other_designations='CG12782-PA|uncharacterized protein', map_location='59D4-59D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23264269, end_position_on_the_genomic_accession=23265543, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37686, CurrentID=0, Status='live', Symbol='CG13540', Aliases='Dmel_ Dmel\\CG13540', description='CG13540 gene product from transcript CG13540-RB', other_designations='CG13540-PB|uncharacterized protein', map_location='59D4-59D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23270477, end_position_on_the_genomic_accession=23271059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37687, CurrentID=0, Status='live', Symbol='ord', Aliases='Dmel_CG3134, CG3134, Dmel\\CG3134, ORD, Ord', description='orientation disruptor', other_designations='CG3134-PA|CG3134-PB|ord-PA|ord-PB|orientation disrupter', map_location='2-103.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23271257, end_position_on_the_genomic_accession=23275536, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37688, CurrentID=0, Status='live', Symbol='CG3124', Aliases='Dmel_ Dmel\\CG3124, anon-WO0140519.230', description='CG3124 gene product from transcript CG3124-RA', other_designations='CG3124-PA|CG3124-PB', map_location='59D4-59D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23282743, end_position_on_the_genomic_accession=23284113, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37689, CurrentID=0, Status='live', Symbol='CG13541', Aliases='Dmel_ Dmel\\CG13541', description='CG13541 gene product from transcript CG13541-RB', other_designations='CG13541-PA|CG13541-PB|uncharacterized protein', map_location='59D4-59D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23284227, end_position_on_the_genomic_accession=23285821, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37690, CurrentID=0, Status='live', Symbol='Prosbeta5R1', Aliases='Dmel_CG9868, CG9868, Dmel\\CG9868, Prosbeta5R, beta5R1', description='Proteasome beta5 subunit-related 1', other_designations='CG9868-PA|Prosbeta5R1-PA|proteasome beta5 subunit-related 1|proteasome beta5R subunit', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23308654, end_position_on_the_genomic_accession=23309816, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37691, CurrentID=0, Status='live', Symbol='CG30412', Aliases='Dmel_ CG9803, Dmel\\CG30412, anon-WO0140519.195', description='CG30412 gene product from transcript CG30412-RA', other_designations='CG30412-PA|CG30412-PB|CG30412-PC|uncharacterized protein', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23318189, end_position_on_the_genomic_accession=23319348, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37692, CurrentID=0, Status='live', Symbol='CG9861', Aliases='Dmel_ Dmel\\CG9861', description='CG9861 gene product from transcript CG9861-RA', other_designations='CG9861-PA', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23323659, end_position_on_the_genomic_accession=23326745, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37694, CurrentID=0, Status='live', Symbol='CG9863', Aliases='Dmel_ BcDNA:AT26616, Dmel\\CG9863', description='CG9863 gene product from transcript CG9863-RB', other_designations='CG9863-PB', map_location='59D7-59D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23346635, end_position_on_the_genomic_accession=23347768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37695, CurrentID=0, Status='live', Symbol='CG3502', Aliases='Dmel_ BEST:LD34564, Dmel\\CG3502', description='CG3502 gene product from transcript CG3502-RB', other_designations='CG3502-PB', map_location='59D7-59D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23343405, end_position_on_the_genomic_accession=23347641, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37696, CurrentID=0, Status='live', Symbol='CG30409', Aliases='Dmel_ CG9872, Dmel\\CG30409', description='CG30409 gene product from transcript CG30409-RA', other_designations='CG30409-PA|CG30409-PB|uncharacterized protein', map_location='59D8-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23354367, end_position_on_the_genomic_accession=23355321, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37697, CurrentID=0, Status='live', Symbol='CG3500', Aliases='Dmel_ BcDNA:AT03803, Dmel\\CG3500', description='CG3500 gene product from transcript CG3500-RA', other_designations='CG3500-PA|CG3500-PB', map_location='59D8-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23356372, end_position_on_the_genomic_accession=23357565, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37698, CurrentID=0, Status='live', Symbol='CG9875', Aliases='Dmel_ BcDNA:GH23338, Dmel\\CG9875', description='CG9875 gene product from transcript CG9875-RB', other_designations='CG9875-PA|CG9875-PB', map_location='59D8-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23357660, end_position_on_the_genomic_accession=23358508, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37699, CurrentID=0, Status='live', Symbol='CG34424', Aliases='Dmel_ CG11079, Dmel\\CG34424, anon-WO0118547.139', description='CG34424 gene product from transcript CG34424-RA', other_designations='CG34424-PA|CG34424-PB|uncharacterized protein', map_location='59D8-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23359215, end_position_on_the_genomic_accession=23360266, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37701, CurrentID=0, Status='live', Symbol='Ice1', Aliases='Dmel_CG13550, CG13550, Dmel\\CG13550, ICE1, dICE1', description='Interacts with the C terminus of ELL 1', other_designations='CG13550-PA|Ice1-PA|interacts with the C terminus of ELL 1', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23366866, end_position_on_the_genomic_accession=23371459, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37702, CurrentID=0, Status='live', Symbol='Golgin245', Aliases='Dmel_CG3493, CG3493, Dmel\\CG3493, dGolgin-245, dGolgin245', description='Golgin-245 ortholog (H. sapiens)', other_designations='CG3493-PA|Golgin-245 ortholog|Golgin245-PA', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23371360, end_position_on_the_genomic_accession=23377045, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37703, CurrentID=0, Status='live', Symbol='yellow-d', Aliases='Dmel_CG9889, BcDNA:GH20107, CG9889, Dmel\\CG9889, d, yell-d', description='CG9889 gene product from transcript CG9889-RA', other_designations='CG9889-PA|yellow-d-PA', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23378573, end_position_on_the_genomic_accession=23380438, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37704, CurrentID=0, Status='live', Symbol='yellow-d2', Aliases='Dmel_CG9891, CG9891, CT27880, Dmel\\CG9891, d2, yel-d2', description='CG9891 gene product from transcript CG9891-RA', other_designations='CG9891-PA|yellow-d2-PA', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23381197, end_position_on_the_genomic_accession=23382591, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37705, CurrentID=0, Status='live', Symbol='CG30414', Aliases='Dmel_ CG3487, Dmel\\CG30414, FBgn0050414, SP62', description='CG30414 gene product from transcript CG30414-RB', other_designations='CG30414-PB|CG30414-PC|uncharacterized protein', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23384200, end_position_on_the_genomic_accession=23385468, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37706, CurrentID=0, Status='live', Symbol='eIF2B-delta', Aliases='Dmel_CG10315, CG10315, Dmel\\CG10315, eIF2Bdelta', description='CG10315 gene product from transcript CG10315-RA', other_designations='CG10315-PA|CG10315-PB|CG10315-PC|CG10315-PD|eIF2B-delta-PA|eIF2B-delta-PB|eIF2B-delta-PC|eIF2B-delta-PD', map_location='59D9-59D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23386521, end_position_on_the_genomic_accession=23389052, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37707, CurrentID=0, Status='live', Symbol='CG3530', Aliases='Dmel_ BcDNA, BcDNA:GH04637, CG18093, Dmel\\CG3530', description='CG3530 gene product from transcript CG3530-RB', other_designations='CG3530-PA|CG3530-PB|CG3530-PC|MYOTUBULARIN-like', map_location='59D10-59D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23389009, end_position_on_the_genomic_accession=23396680, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  37708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37708, CurrentID=0, Status='live', Symbol='CG3520', Aliases='Dmel_ Dmel\\CG3520', description='CG3520 gene product from transcript CG3520-RA', other_designations='CG3520-PA', map_location='59D10-59D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23396770, end_position_on_the_genomic_accession=23402309, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37709, CurrentID=0, Status='live', Symbol='CG9812', Aliases='Dmel_ BcDNA:RH38235, Dmel\\CG9812', description='CG9812 gene product from transcript CG9812-RB', other_designations='CG9812-PB|CG9812-PD|CG9812-PE', map_location='59D11-59D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23412320, end_position_on_the_genomic_accession=23419684, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37710, CurrentID=0, Status='live', Symbol='CG9815', Aliases='Dmel_ Dmel\\CG9815', description='CG9815 gene product from transcript CG9815-RD', other_designations='CG9815-PA|CG9815-PB|CG9815-PC|CG9815-PD', map_location='59D11-59D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23421521, end_position_on_the_genomic_accession=23433354, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37711, CurrentID=0, Status='live', Symbol='Or59a', Aliases='Dmel_CG9820, 59D.1, 59a, AN6, CG9820, DOR46, DOR59D.1, Dmel\\CG9820, Or46, Or59D.1, dor46', description='Odorant receptor 59a', other_designations='CG9820-PA|Or59a-PA|odorant receptor 59a|olfactory receptor 59D.1', map_location='59E1-59E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23439868, end_position_on_the_genomic_accession=23441065, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37713, CurrentID=0, Status='live', Symbol='SkpF', Aliases='Dmel_CG12227, CG12227, Dmel\\CG12227, Skp, skpF', description='SKP1-related F', other_designations='CG12227-PA|SkpF-PA|skpF', map_location='59E1-59E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23458065, end_position_on_the_genomic_accession=23458759, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37715, CurrentID=0, Status='live', Symbol='Or59b', Aliases='Dmel_CG3569, 59E.1, 59b, CG3569, DOR119, DmOR59B, Dmel\\CG3569', description='Odorant receptor 59b', other_designations='CG3569-PA|Or59b-PA|odorant receptor 59b', map_location='59E1-59E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23470165, end_position_on_the_genomic_accession=23471775, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37716, CurrentID=0, Status='live', Symbol='Or59c', Aliases='Dmel_CG17226, 59E.2, 59c, CG17226, DOR120, Dmel\\CG17226', description='Odorant receptor 59c', other_designations='CG17226-PA|Or59c-PA|odorant receptor 59c', map_location='59E1-59E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23472300, end_position_on_the_genomic_accession=23473753, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37717, CurrentID=0, Status='live', Symbol='CG13557', Aliases='Dmel_ Dmel\\CG13557', description='CG13557 gene product from transcript CG13557-RA', other_designations='CG13557-PA|uncharacterized protein', map_location='59E1-59E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23475698, end_position_on_the_genomic_accession=23476418, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37719, CurrentID=0, Status='live', Symbol='CG13558', Aliases='Dmel_ Dmel\\CG13558', description='CG13558 gene product from transcript CG13558-RA', other_designations='CG13558-PA|uncharacterized protein', map_location='59E2-59E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23499651, end_position_on_the_genomic_accession=23500702, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37720, CurrentID=0, Status='live', Symbol='CG13559', Aliases='Dmel_ Dmel\\CG13559', description='CG13559 gene product from transcript CG13559-RA', other_designations='CG13559-PA|uncharacterized protein', map_location='59E2-59E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23502946, end_position_on_the_genomic_accession=23503815, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37721, CurrentID=0, Status='live', Symbol='CG3906', Aliases='Dmel_ Dmel\\CG3906', description='CG3906 gene product from transcript CG3906-RA', other_designations='CG3906-PA', map_location='59E2-59E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23511286, end_position_on_the_genomic_accession=23512225, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37723, CurrentID=0, Status='live', Symbol='mi', Aliases='Dmel_CG5360, CG5360, Dmel\\CG5360, l(2)SH0818, l(2)SH2 0818', description='minus', other_designations='CG5360-PA|lethal (2) SH0818|mi-PA', map_location='2-104.7 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23523856, end_position_on_the_genomic_accession=23528260, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37724, CurrentID=0, Status='live', Symbol='bw', Aliases='Dmel_CG17632, CG17632, Dmel\\CG17632, Pm, Su(w[co2]), Su(w[coJ]), ptm', description='brown', other_designations='CG17632-PA|CG17632-PD|brow|bw-PA|bw-PD|plum|pteridine-modifier|suppressor of white-coral J|suppressor of white-coral-2', map_location='2-104.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23527805, end_position_on_the_genomic_accession=23538499, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37725, CurrentID=0, Status='live', Symbol='Gr59e', Aliases='Dmel_CG33151, 59E.2, 59E2, CG15872, CG33151, Dmel\\CG33151, GR59E.2, GR59e, Gr59E, Gr59E2', description='Gustatory receptor 59e', other_designations='CG33151-PA|Gr59e-PA|gustatory receptor 59E.2|gustatory receptor 59e', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23541054, end_position_on_the_genomic_accession=23542379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37726, CurrentID=0, Status='live', Symbol='Gr59f', Aliases='Dmel_CG33150, 59E.1, CG33150, CG5365, Dmel\\CG33150, GR59E.1, Gr59E1', description='Gustatory receptor 59f', other_designations='CG33150-PA|Gr59f-PA|gustatory receptor 59E.1|gustatory receptor 59f', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23542510, end_position_on_the_genomic_accession=23543885, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37727, CurrentID=0, Status='live', Symbol='wmd', Aliases='Dmel_CG3957, CG3957, CG3957/wmd, Dmel\\CG3957, WMD, pter', description='wing morphogenesis defect', other_designations='CG3957-PA|CG3957-PB|pterodactyl|wmd-PA|wmd-PB', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23544823, end_position_on_the_genomic_accession=23548151, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37728, CurrentID=0, Status='live', Symbol='levy', Aliases='Dmel_CG17280, BcDNA:RH03295, CG17280, COX6A, COXD, COXE, Dmel\\CG17280, VIa', description='CG17280 gene product from transcript CG17280-RA', other_designations='CG17280-PA|CG17280-PB|cytochrome c oxidase VIa|cytochrome c oxidase polypeptide VIA|cytochrome c oxidase subunit 6A|levy-PA|levy-PB', map_location='2-103 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23548239, end_position_on_the_genomic_accession=23548982, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37729, CurrentID=0, Status='live', Symbol='Dcp-1', Aliases='Dmel_CG5370, CG5370, DCP-1, DCP1, Dcp1, Dmel\\CG5370, ccp1, dcp-1, dcp1, l(2)01862, l(2)02132', description='Death caspase-1', other_designations='CG5370-PA|Dcp-1-PA|caspase|caspase 1|caspase-1|death caspase-1', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23551313, end_position_on_the_genomic_accession=23553547, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37730, CurrentID=0, Status='live', Symbol='pita', Aliases='Dmel_CG3941, CG3941, Dmel\\CG3941, PITA, Pita, spdk', description='CG3941 gene product from transcript CG3941-RA', other_designations='CG3941-PA|CG3941-PB|pita-PA|pita-PB|spotted dick', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23548967, end_position_on_the_genomic_accession=23554712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37731, CurrentID=0, Status='live', Symbol='Rrp4', Aliases='Dmel_CG3931, CG3931, Dmel\\CG3931, RRP4, d rrp4', description='CG3931 gene product from transcript CG3931-RA', other_designations='CG3931-PA|CG3931-PB|Rrp4-PA|Rrp4-PB', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23555018, end_position_on_the_genomic_accession=23556217, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37732, CurrentID=0, Status='live', Symbol='ItgaPS5', Aliases='Dmel_CG5372, CG5372, CT17068, Dmel\\CG5372, ItgalphaPS5, alpha-PS5, alphaPS5', description='Integrin alphaPS5 subunit', other_designations='CG5372-PA|ItgaPS5-PA|alphaPS5|alphaPS5 integrin|integrin alphaPS5 subunit', map_location='59E4-59E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23556672, end_position_on_the_genomic_accession=23560263, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37733, CurrentID=0, Status='live', Symbol='Pi3K59F', Aliases='Dmel_CG5373, CG5373, DmVps34, Dmel\\CG5373, PI(3)K, PI3K, PI3K-59F, PI3K_59F, PI[[3]]K, Pi3K_59F, VPS34, Vps34, dPI3K, dVps34, dVps34/PI3K59F, vps34', description='Phosphotidylinositol 3 kinase 59F', other_designations='CG5373-PA|PI 3-kinase|PI3-kinase|Pi3K59F-PA|dPI 3-kinase|phosphotidylinositol 3 kinase 59F', map_location='59E4-59F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23560492, end_position_on_the_genomic_accession=23564045, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37734, CurrentID=0, Status='live', Symbol='apt', Aliases='Dmel_CG5393, 3041, Apt, CG5393, Dmel\\CG5393, TDF, TDF/APT, l(2)03041, l(2)06369, l(2)09049, l(2)59Ea, l(2)k11531, tdf', description='apontic', other_designations='CG5393-PA|CG5393-PB|CG5393-PC|CG5393-PD|CG5393-PE|CG5393-PF|apt-PA|apt-PB|apt-PC|apt-PD|apt-PE|apt-PF|trachea defective|tracheae defective|tracheal defective', map_location='59F1-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23564891, end_position_on_the_genomic_accession=23599526, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37735, CurrentID=0, Status='live', Symbol='CG10332', Aliases='Dmel_CG10332, BcDNA:AT19986-ORFA, Dmel\\CG10332, anon-WO0153538.56, anon-WO0153538.57, anon-WO0153538.58', description='CG10332 gene product from transcript CG10332-RA', other_designations='CG10332-PA|uncharacterized protein', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23600914, end_position_on_the_genomic_accession=23601773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37736, CurrentID=0, Status='live', Symbol='CG5398', Aliases='Dmel_ Aqp5398, Dmel\\CG5398, anon-WO0140519.151', description='CG5398 gene product from transcript CG5398-RA', other_designations='CG5398-PA', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23602585, end_position_on_the_genomic_accession=23603641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37737, CurrentID=0, Status='live', Symbol='CG17664', Aliases='Dmel_ Aqp17664, Dmel\\CG17664', description='CG17664 gene product from transcript CG17664-RC', other_designations='CG17664-PA|CG17664-PC|uncharacterized protein', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23603722, end_position_on_the_genomic_accession=23616754, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37738, CurrentID=0, Status='live', Symbol='CG17662', Aliases='Dmel_ Aqp17662, Dmel\\CG17662', description='CG17662 gene product from transcript CG17662-RB', other_designations='CG17662-PB|uncharacterized protein', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23605689, end_position_on_the_genomic_accession=23606654, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37739, CurrentID=0, Status='live', Symbol='CG4019', Aliases='Dmel_ Aqp4019, Dmel\\CG4019', description='CG4019 gene product from transcript CG4019-RA', other_designations='CG4019-PA|CG4019-PB|CG4019-PC|CG4019-PD|CG4019-PE|CG4019-PF', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23607592, end_position_on_the_genomic_accession=23614710, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37741, CurrentID=0, Status='live', Symbol='Pde8', Aliases='Dmel_CG45019, CG42241, CG45019, CG5411, Dmel\\CG45019, Dmel_CG42241, Dmel_CG5411, PDE8, pde8', description='Phosphodiesterase 8', other_designations='CG45019-PA|CG45019-PB|CG45019-PC|CG45019-PD|CG45019-PE|CG45019-PI|CG45019-PK|CG45019-PN|CG45019-PO|Pde8-PA|Pde8-PB|Pde8-PC|Pde8-PD|Pde8-PE|Pde8-PI|Pde8-PK|Pde8-PN|Pde8-PO|phosphodiesterase 8', map_location='59F5-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23657567, end_position_on_the_genomic_accession=23680844, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  37742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37742, CurrentID=0, Status='live', Symbol='St1', Aliases='Dmel_CG5428, CG5428, Dmel\\CG5428, dmST1, st1', description='Sulfotransferase 1', other_designations='CG5428-PB|St1-PB|sulfotransferase 1', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23681623, end_position_on_the_genomic_accession=23682952, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37743, CurrentID=0, Status='live', Symbol='St3', Aliases='Dmel_CG44167, CG11293, CG44167, CG5431, Dmel\\CG44167, Dmel_CG11293, Dmel_CG5431, dmST3, st3', description='Sulfotransferase 3', other_designations='CG44167-PA|CG44167-PB|CG44167-PC|St3-PA|St3-PB|St3-PC|sulfotransferase 3', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23683377, end_position_on_the_genomic_accession=23687980, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37744, CurrentID=0, Status='live', Symbol='l(2)efl', Aliases='Dmel_CG4533, CG4533, Cryab, Dmel\\CG4533, L(2)efl, el(2)13-1, l(2)13-1, l(2)59Fa, l(2)Efl', description='lethal (2) essential for life', other_designations='CG4533-PA|CG4533-PC|l(2)efl-PA|l(2)efl-PC|lethal(2)essential for life', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23684651, end_position_on_the_genomic_accession=23685543, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37746, CurrentID=0, Status='live', Symbol='Pal2', Aliases='Dmel_CG5472, CG5472, Dmel\\CG5472, PAL, PAL2, Pal, anon-WO0149856.5, dPAL2', description='Peptidyl-alpha-hydroxyglycine-alpha-amidating lyase 2', other_designations='CG5472-PA|CG5472-PD|CG5472-PE|Pal2-PA|Pal2-PD|Pal2-PE|Peptidyl-alpha-hydroxyglycine-alpha-amidating lyase', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23688077, end_position_on_the_genomic_accession=23690838, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37748, CurrentID=0, Status='live', Symbol='angel', Aliases='Dmel_CG12273, CG12273, Dmel\\CG12273', description='CG12273 gene product from transcript CG12273-RA', other_designations='CG12273-PA|angel-PA', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23694003, end_position_on_the_genomic_accession=23695232, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37749, CurrentID=0, Status='live', Symbol='CG30183', Aliases='Dmel_ CG4093, CG4530, Dmel\\CG30183', description='CG30183 gene product from transcript CG30183-RC', other_designations='CG30183-PA|CG30183-PC|uncharacterized protein', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23690247, end_position_on_the_genomic_accession=23697824, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37750, CurrentID=0, Status='live', Symbol='mRpL43', Aliases='Dmel_CG5479, BcDNA:RE35686, CG5479, CI-B8, Dmel\\CG5479, mRpL-CI-B8', description='mitochondrial ribosomal protein L43', other_designations='CG5479-PA|mRpL43-PA|mitochondrial ribosomal protein L-CI-B8', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23698105, end_position_on_the_genomic_accession=23699014, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37751, CurrentID=0, Status='live', Symbol='sigmar', Aliases='Dmel_CG4091, CG4091, Dmel\\CG4091', description='salivary glands marred', other_designations='CG4091-PA|CG4091-PB|CG4091-PC|sigmar-PA|sigmar-PB|sigmar-PC', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23698949, end_position_on_the_genomic_accession=23702406, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37754, CurrentID=0, Status='live', Symbol='l(2)not', Aliases='Dmel_CG4084, 1(2) not, CG4084, Dmel\\CG4084, cDNA 10F, dAlg3, ntid', description='lethal (2) neighbor of tid', other_designations='CG4084-PA|l(2)not-PA|lethal(2) neighbour of 1(2)tid', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23705439, end_position_on_the_genomic_accession=23709792, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37755, CurrentID=0, Status='live', Symbol='Sesn', Aliases='Dmel_CG11299, BEST:GH20113, BcDNA:GH20113, CG11299, Dmel\\CG11299, dSesn', description='Sestrin', other_designations='CG11299-PA|CG11299-PB|CG11299-PC|CG11299-PD|Sesn-PA|Sesn-PB|Sesn-PC|Sesn-PD|sestrin', map_location='59F6-59F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23713900, end_position_on_the_genomic_accession=23734846, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  37756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37756, CurrentID=0, Status='live', Symbol='CG18128', Aliases='Dmel_ Dmel\\CG18128', description='CG18128 gene product from transcript CG18128-RA', other_designations='CG18128-PA|uncharacterized protein', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23718798, end_position_on_the_genomic_accession=23719936, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37757, CurrentID=0, Status='live', Symbol='egl', Aliases='Dmel_CG4051, CG4051, Dmel\\CG4051, EGL, Egl, egal', description='egalitarian', other_designations='CG4051-PB|CG4051-PC|egl-PB|egl-PC', map_location='2-105 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23734904, end_position_on_the_genomic_accession=23746829, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37758, CurrentID=0, Status='live', Symbol='CG13560', Aliases='Dmel_ Dmel\\CG13560', description='CG13560 gene product from transcript CG13560-RB', other_designations='CG13560-PA|CG13560-PB|uncharacterized protein', map_location='59F7-59F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23747560, end_position_on_the_genomic_accession=23748477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37759, CurrentID=0, Status='live', Symbol='CG12491', Aliases='Dmel_ Dmel\\CG12491', description='CG12491 gene product from transcript CG12491-RA', other_designations='CG12491-PA|uncharacterized protein', map_location='59F7-59F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23750646, end_position_on_the_genomic_accession=23751225, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37760, CurrentID=0, Status='live', Symbol='CG11300', Aliases='Dmel_ BcDNA:LP08443, Dmel\\CG11300', description='CG11300 gene product from transcript CG11300-RA', other_designations='CG11300-PA|uncharacterized protein', map_location='59F7-59F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23751915, end_position_on_the_genomic_accession=23752495, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37761, CurrentID=0, Status='live', Symbol='CG5532', Aliases='Dmel_ Dmel\\CG5532', description='CG5532 gene product from transcript CG5532-RB', other_designations='CG5532-PA|CG5532-PB', map_location='59F7-59F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23752754, end_position_on_the_genomic_accession=23753515, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37762, CurrentID=0, Status='live', Symbol='CG9850', Aliases='Dmel_ CG11169, Dimp, Dmel\\CG9850, dimp', description='CG9850 gene product from transcript CG9850-RC', other_designations='CG9850-PA|CG9850-PC|CG9850-PD|CG9850-PE|CG9850-PF|Dlg-interacting metalloprotease', map_location='59F7-60A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23753477, end_position_on_the_genomic_accession=23798395, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  37763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37763, CurrentID=0, Status='live', Symbol='CG15800', Aliases='Dmel_ Dmel\\CG15800, SkpG', description='CG15800 gene product from transcript CG15800-RA', other_designations='CG15800-PA|uncharacterized protein', map_location='60A1-60A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23790614, end_position_on_the_genomic_accession=23791227, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37765, CurrentID=0, Status='live', Symbol='CG13561', Aliases='Dmel_ Dmel\\CG13561', description='CG13561 gene product from transcript CG13561-RB', other_designations='CG13561-PB|uncharacterized protein', map_location='60A1-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23804315, end_position_on_the_genomic_accession=23806249, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37766, CurrentID=0, Status='live', Symbol='CG5539', Aliases='Dmel_ Dmel\\CG5539', description='CG5539 gene product from transcript CG5539-RB', other_designations='CG5539-PA|CG5539-PB', map_location='60A1-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23804811, end_position_on_the_genomic_accession=23806144, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37767, CurrentID=0, Status='live', Symbol='Ssrp', Aliases='Dmel_CG4817, CG4817, Cf5, Dmel\\CG4817, DssRP, SSRP, SSRP1, dSSRP, dSSRP1, dSsrp, ssrp', description='Structure specific recognition protein', other_designations='CG4817-PA|Ssrp-PA|chorion factor|structure specific recognition protein|structure-specific recognition protein', map_location='60A2-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23806361, end_position_on_the_genomic_accession=23809004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37768, CurrentID=0, Status='live', Symbol='CG5543', Aliases='Dmel_ Dmel\\CG5543, anon-WO0118547.181', description='CG5543 gene product from transcript CG5543-RA', other_designations='CG5543-PA', map_location='60A2-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23809236, end_position_on_the_genomic_accession=23811398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37769, CurrentID=0, Status='live', Symbol='Nxt1', Aliases='Dmel_CG12752, CG12752, DmNXT1, Dmel\\CG12752, Nxt, dNxt, dmNXT, dmNXT1, dmP15, nxt1, p15', description='NTF2-related export protein 1', other_designations='CG12752-PA|CG12752-PB|Nxt1-PA|Nxt1-PB', map_location='60A2-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23811321, end_position_on_the_genomic_accession=23811974, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37770, CurrentID=0, Status='live', Symbol='CG4797', Aliases='Dmel_ Dmel\\CG4797', description='CG4797 gene product from transcript CG4797-RB', other_designations='CG4797-PB|sugar transporter-like', map_location='60A2-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23812185, end_position_on_the_genomic_accession=23818327, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37771, CurrentID=0, Status='live', Symbol='CG4763', Aliases='Dmel_ Dmel\\CG4763', description='CG4763 gene product from transcript CG4763-RA', other_designations='CG4763-PA', map_location='60A2-60A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23823357, end_position_on_the_genomic_accession=23824868, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37772, CurrentID=0, Status='live', Symbol='CG5549', Aliases='Dmel_ Dmel\\CG5549', description='CG5549 gene product from transcript CG5549-RB', other_designations='CG5549-PB', map_location='60A2-60A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23820354, end_position_on_the_genomic_accession=23839864, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37774, CurrentID=0, Status='live', Symbol='Snap29', Aliases='Dmel_CG11173, CG11173, Dmel\\CG11173, MENE (2R)-E, MENE(2R)-E, SNAP-29, SNAP29, Ubisnap, ubisnap, usnp', description='Synaptosomal-associated protein 29kDa', other_designations='CG11173-PA|Snap29-PA|ubisnap', map_location='60A3-60A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23843483, end_position_on_the_genomic_accession=23844763, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37775, CurrentID=0, Status='live', Symbol='CG5554', Aliases='Dmel_ Dmel\\CG5554', description='CG5554 gene product from transcript CG5554-RB', other_designations='CG5554-PA|CG5554-PB', map_location='60A3-60A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23845001, end_position_on_the_genomic_accession=23847024, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37776, CurrentID=0, Status='live', Symbol='eIF6', Aliases='Dmel_CG17611, CG17611, D Dmel\\CG17611, anon-WO0172774.153, anon-WO0172774.155, eif6, elF6, l(2)k13214', description='CG17611 gene product from transcript CG17611-RA', other_designations='CG17611-PA|eIF6-PA', map_location='60A3-60A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23849246, end_position_on_the_genomic_accession=23850301, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37778, CurrentID=0, Status='live', Symbol='gbb', Aliases='Dmel_CG5562, 60A, BMP, CG5562, Dm-GBB, Dmel\\CG5562, GBB, Gbb, Gbb-60A, SixtyA, TGF-b, TGFbeta-60A, Tgfb-60, Tgfbeta-60A-60A, gcn, l(2)60A-J, tgfb-60A, vgr/60A, gbb', description='glass bottom boat', other_designations='CG5562-PA|CG5562-PB|bone morphogenetic protein|bone morphogenic protein|gbb-PA|gbb-PB|glass bottom boat-60A|glass-bottom boat|glass-bottomed boat|transforming growth factor beta at 60A', map_location='60A3-60A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23851562, end_position_on_the_genomic_accession=23853247, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37780, CurrentID=0, Status='live', Symbol='Pym', Aliases='Dmel_CG30176, BEST:HL05757, CG10330, CG30176, Dmel\\CG30176, PYM, UD3, Wibg, pym, wibg', description='Partner of Y14 and Mago', other_designations='CG30176-PA|Pym-PA|partner of Y14 and mago|undefined 3|within bgcn', map_location='60A4-60A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23855954, end_position_on_the_genomic_accession=23856936, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37781, CurrentID=0, Status='live', Symbol='TBPH', Aliases='Dmel_CG10327, CG10327, Dmel\\CG10327, TDP-43, TDP43, dTDP, dTDP-43, dmTDP43, dtar, tbph', description='TAR DNA-binding protein-43 homolog', other_designations='CG10327-PA|CG10327-PB|CG10327-PC|CG10327-PD|CG10327-PE|CG10327-PF|TAR DNA binding protein|TBPH-PA|TBPH-PB|TBPH-PC|TBPH-PD|TBPH-PE|TBPH-PF|Tar DNA binding protein homolog', map_location='60A4-60A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23859067, end_position_on_the_genomic_accession=23862593, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37782, CurrentID=0, Status='live', Symbol='CG4585', Aliases='Dmel_ CK00282, Dmel\\CG4585, UD2, anon-60Ab, dCCS4', description='CG4585 gene product from transcript CG4585-RA', other_designations='CG4585-PA|CG4585-PB|CG4585-PC|CG4585-PD|undefined 2', map_location='60A5-60A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23862962, end_position_on_the_genomic_accession=23865398, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37783, CurrentID=0, Status='live', Symbol='CG5569', Aliases='Dmel_ BcDNA:SD18780, Dmel\\CG5569', description='CG5569 gene product from transcript CG5569-RA', other_designations='CG5569-PA', map_location='60A5-60A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23865414, end_position_on_the_genomic_accession=23866654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37784, CurrentID=0, Status='live', Symbol='Thiolase', Aliases='Dmel_CG4581, ACoT, CG4581, Dmel\\CG4581, Mtpbeta, l(2)k09828, thiolase', description='CG4581 gene product from transcript CG4581-RA', other_designations='CG4581-PA|Thiolase-PA|thiolase', map_location='60A5-60A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23866648, end_position_on_the_genomic_accession=23868603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37785, CurrentID=0, Status='live', Symbol='ken', Aliases='Dmel_CG5575, 2970, 3907, 5029, 8253, BCL-6/KEN, CG5575, Dmel\\CG5575, KEN, Ken, P420, l(2)00628, l(2)02970, l(2)05029, ok, okina', description='ken and barbie', other_designations='CG5575-PA|Ken & barbie|ken-PA|ken-and-barbie', map_location='60A6-60A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23870273, end_position_on_the_genomic_accession=23877442, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37786, CurrentID=0, Status='live', Symbol='TM4SF', Aliases='Dmel_CG11303, CG11303, Dm.Tsp60A, Dmel\\CG11303', description='Transmembrane 4 superfamily', other_designations='CG11303-PA|CG11303-PB|CG11303-PC|TM4SF-PA|TM4SF-PB|TM4SF-PC|transmembrane 4 superfamily', map_location='60A7-60A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23878197, end_position_on_the_genomic_accession=23880860, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37787, CurrentID=0, Status='live', Symbol='CG4882', Aliases='Dmel_ Dmel\\CG4882, GM14313, LD24471, UD1, anon-60Aa', description='CG4882 gene product from transcript CG4882-RB', other_designations='CG4882-PA|CG4882-PB|undefined 1', map_location='60A7-60A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23880817, end_position_on_the_genomic_accession=23882927, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37788, CurrentID=0, Status='live', Symbol='PHDP', Aliases='Dmel_CG11182, CG11182, Dmel\\CG11182, anon-60Ac', description='Putative homeodomain protein', other_designations='CG11182-PA|PHDP-PA|putative homeodomain protein', map_location='60A7-60A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23883136, end_position_on_the_genomic_accession=23884440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37789, CurrentID=0, Status='live', Symbol='CG5597', Aliases='Dmel_ 5597, CT17698, Dmel\\CG5597', description='CG5597 gene product from transcript CG5597-RA', other_designations='CG5597-PA', map_location='60A8-60A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23886973, end_position_on_the_genomic_accession=23888777, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37790, CurrentID=0, Status='live', Symbol='DCP1', Aliases='Dmel_CG11183, CG11183, Dcp-1, Dcp1, Dcp1p, Dmel\\CG11183, anon-WO0118547.163, dDCP1, dDcp1', description='Decapping protein 1', other_designations='CG11183-PA|CG11183-PB|DCP1-PA|DCP1-PB|decapping protein 1', map_location='60A8-60A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23888764, end_position_on_the_genomic_accession=23890276, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37791, CurrentID=0, Status='live', Symbol='DNA-ligI', Aliases='Dmel_CG5602, CG5602, DmDNA Lig I, Dmel\\CG5602, LIG1', description='DNA ligase I', other_designations='CG5602-PA|DNA-ligI-PA', map_location='60A8-60A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23890512, end_position_on_the_genomic_accession=23893177, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37792, CurrentID=0, Status='live', Symbol='Upf3', Aliases='Dmel_CG11184, CG11184, Dmel\\CG11184, UPF3, upf3', description='CG11184 gene product from transcript CG11184-RB', other_designations='CG11184-PB|CG11184-PD|Upf3-PB|Upf3-PD', map_location='60A8-60A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23893066, end_position_on_the_genomic_accession=23894987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37793, CurrentID=0, Status='live', Symbol='CG17658', Aliases='Dmel_ Dmel\\CG17658', description='CG17658 gene product from transcript CG17658-RA', other_designations='CG17658-PA|uncharacterized protein', map_location='60A9-60A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23895063, end_position_on_the_genomic_accession=23896997, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37794, CurrentID=0, Status='live', Symbol='CG5339', Aliases='Dmel_ Dmel\\CG5339, drev', description='CG5339 gene product from transcript CG5339-RA', other_designations='CG5339-PA', map_location='60A9-60A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23896896, end_position_on_the_genomic_accession=23898246, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37795, CurrentID=0, Status='live', Symbol='Lpt', Aliases='Dmel_CG5591, CG5591, Cmi, Dmel\\CG5591, LPT, cg5591, cmi', description='Lost PHDs of trr', other_designations='CG5591-PA|CG5591-PB|Lpt-PA|Lpt-PB|cara mitad|lost PHD finger of Trr|lost PHDs of Trr|lost PHDs of trr', map_location='60A9-60A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23898561, end_position_on_the_genomic_accession=23903661, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37797, CurrentID=0, Status='live', Symbol='CG13562', Aliases='Dmel_ Dmel\\CG13562', description='CG13562 gene product from transcript CG13562-RA', other_designations='CG13562-PA|CG13562-PC|uncharacterized protein', map_location='60A9-60A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23903750, end_position_on_the_genomic_accession=23905269, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37798, CurrentID=0, Status='live', Symbol='Nap1', Aliases='Dmel_CG5330, CG5330, Dmel\\CG5330, NAP-1, NAP1, Nap-1, dNAP-1, dNAP1, dNap, dNap1, nap1, p56/dCAF-4', description='Nucleosome assembly protein 1', other_designations='CG5330-PA|CG5330-PB|CG5330-PC|Nap1-PA|Nap1-PB|Nap1-PC|nucleosome assembly protein 1', map_location='60A9-60A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23905175, end_position_on_the_genomic_accession=23907294, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37800, CurrentID=0, Status='live', Symbol='kcc', Aliases='Dmel_CG5594, BEST:CK01510, CG2768, CG5594, CK01510, Dmel\\CG5594, E(sda)J, KCC, Line J, l(2)60A-C, mdcds_17605', description='kazachoc', other_designations='CG5594-PA|CG5594-PB|CG5594-PC|CG5594-PD|CG5594-PE|CG5594-PF|CG5594-PG|E(sda)J|kcc-PA|kcc-PB|kcc-PC|kcc-PD|kcc-PE|kcc-PF|kcc-PG|lethal (2) 60A-C', map_location='60A9-60A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23907654, end_position_on_the_genomic_accession=23925066, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  37801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37801, CurrentID=0, Status='live', Symbol='CG2812', Aliases='Dmel_ Dmel\\CG2812', description='CG2812 gene product from transcript CG2812-RA', other_designations='CG2812-PA', map_location='60A11-60A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23925236, end_position_on_the_genomic_accession=23926911, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37803, CurrentID=0, Status='live', Symbol='CG3735', Aliases='Dmel_ Dmel\\CG3735', description='CG3735 gene product from transcript CG3735-RB', other_designations='CG3735-PA|CG3735-PB', map_location='60A12-60A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23937077, end_position_on_the_genomic_accession=23939754, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37804, CurrentID=0, Status='live', Symbol='Taldo', Aliases='Dmel_CG2827, 154176_at, CG2827, CT9666, Dmel\\CG2827, TA, Tal, tal', description='Transaldolase', other_designations='CG2827-PA|Tal|Taldo-PA|transaldolase', map_location='60A12-60A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23940410, end_position_on_the_genomic_accession=23942114, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37805, CurrentID=0, Status='live', Symbol='Galphas', Aliases='Dmel_CG2835, CG2835, DGs, DGsalpha, Dmel\\CG2835, G-5alpha, G-alpha-60A, G-salpha 60A, G-salpha60A, GSalpha, G[[S]]alpha, G[[alphas]], G[[s]], G[[s]]-alpha-60A, G[[s]]alpha, Gaalpha, Galpha60A, GalphaS, Galpha[[s]], Galpha[[s]]60A, Gs, Gs-alpha, Gs-alpha60A, Gsalpha, Gsalpha60A, dG[[alphas]], dgs, dgsalpha', description='G protein alpha s subunit', other_designations='CG2835-PA|CG2835-PB|CG2835-PC|CG2835-PD|CG2835-PE|G protein salpha 60A|G-alpha s|Galphas-PA|Galphas-PB|Galphas-PC|Galphas-PD|Galphas-PE|stimulatory alpha subunit of G protein', map_location='60A12-60A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23942538, end_position_on_the_genomic_accession=23946936, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37806, CurrentID=0, Status='live', Symbol='Orcokinin', Aliases='Dmel_CG13565, BcDNA:RH09340, CG13565, Dmel\\CG13565', description='CG13565 gene product from transcript CG13565-RA', other_designations='CG13565-PA|CG13565-PB|CG13565-PC|Orcokinin-PA|Orcokinin-PB|Orcokinin-PC|orcokinin', map_location='60A13-60A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23946959, end_position_on_the_genomic_accession=23949076, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37807, CurrentID=0, Status='live', Symbol='CG2970', Aliases='Dmel_ Dmel\\CG2970, SD03319p', description='CG2970 gene product from transcript CG2970-RA', other_designations='CG2970-PA|CG2970-PB', map_location='60A13-60A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23949201, end_position_on_the_genomic_accession=23952459, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37808, CurrentID=0, Status='live', Symbol='fzr2', Aliases='Dmel_CG16783, CG16783, Dmel\\CG16783, FZR2', description='fizzy-related 2', other_designations='CG16783-PA|fzr2-PA', map_location='60A13-60A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23950647, end_position_on_the_genomic_accession=23952096, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37809, CurrentID=0, Status='live', Symbol='CG3803', Aliases='Dmel_ Dmel\\CG3803', description='CG3803 gene product from transcript CG3803-RA', other_designations='CG3803-PA', map_location='60A13-60A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23952513, end_position_on_the_genomic_accession=23954048, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37810, CurrentID=0, Status='live', Symbol='thoc5', Aliases='Dmel_CG2980, CG2980, Dmel\\CG2980, THOC5, garm', description='CG2980 gene product from transcript CG2980-RA', other_designations='CG2980-PA|garmcho|thoc5-PA', map_location='60A13-60A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23954091, end_position_on_the_genomic_accession=23956406, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37811, CurrentID=0, Status='live', Symbol='CG16787', Aliases='Dmel_ Dmel\\CG16787', description='CG16787 gene product from transcript CG16787-RA', other_designations='CG16787-PA|uncharacterized protein', map_location='60A13-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23956335, end_position_on_the_genomic_accession=23957405, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740702, CurrentID=0, Status='live', Symbol='CG34433', Aliases='Dmel_ CG15541, Dmel\\CG34433, Dmel_CG15541', description='CG34433 gene product from transcript CG34433-RA', other_designations='CG34433-PA|CG34433-PB|uncharacterized protein', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30649961, end_position_on_the_genomic_accession=30650858, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37814, CurrentID=0, Status='live', Symbol='Fmo-1', Aliases='Dmel_CG3006, CG3006, DmFMO-1, Dmel\\CG3006, FMO-1', description='Flavin-containing monooxygenase 1', other_designations='CG3006-PA|Fmo-1-PA', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23962956, end_position_on_the_genomic_accession=23964641, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37815, CurrentID=0, Status='live', Symbol='Alas', Aliases='Dmel_CG3017, ALAS, CG3017, Dmel\\CG3017, alas', description='Aminolevulinate synthase', other_designations='Alas-PA|Alas/50-Aminolevulinate synthase|CG3017-PA|aminolevulinate synthase|delta-Aminolevulinate synthase', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23964772, end_position_on_the_genomic_accession=23966757, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37817, CurrentID=0, Status='live', Symbol='CG10904', Aliases='Dmel_ Dmel\\CG10904', description='CG10904 gene product from transcript CG10904-RB', other_designations='CG10904-PA|CG10904-PB|uncharacterized protein', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23968941, end_position_on_the_genomic_accession=23970073, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37818, CurrentID=0, Status='live', Symbol='CG3065', Aliases='Dmel_ Dmel\\CG3065', description='CG3065 gene product from transcript CG3065-RA', other_designations='CG3065-PA|CG3065-PB|CG3065-PC', map_location='60A14-60A15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23970305, end_position_on_the_genomic_accession=23972233, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37820, CurrentID=0, Status='live', Symbol='Gadd34', Aliases='Dmel_CG3825, CG3825, Dm-GADD34, Dmel\\CG3825, GADD34, dGADD34, dPPP1R15, gadd34', description='CG3825 gene product from transcript CG3825-RA', other_designations='CG3825-PA|Gadd34-PA', map_location='60A15-60A16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23975599, end_position_on_the_genomic_accession=23977255, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37822, CurrentID=0, Status='live', Symbol='Sox14', Aliases='Dmel_CG3090, CG17263, CG3090, DSox-14, DSox14, DSox60B, Dmel\\CG3090, Dsox-14, SOX14, SOX60B, SOXC, sox14', description='Sox box protein 14', other_designations='CG3090-PA|CG3090-PB|CG3090-PC|Sox14-PA|Sox14-PB|Sox14-PC|sox box protein 14', map_location='60A16-60A16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23978801, end_position_on_the_genomic_accession=23985177, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37823, CurrentID=0, Status='live', Symbol='Phm', Aliases='Dmel_CG3832, CG3832, Dmel\\CG3832, PHM, dPHM', description='Peptidylglycine-alpha-hydroxylating monooxygenase', other_designations='CG3832-PA|CG3832-PB|Phm-PA|Phm-PB|peptidyl gycine alpha hydroxylating monooxygenase|peptidylglycine alpha-hydroxylating monooxygenase|peptidylglycine-alpha-hydroxylating monooxygenase', map_location='60A16-60A16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23985175, end_position_on_the_genomic_accession=23988455, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37824, CurrentID=0, Status='live', Symbol='Pask', Aliases='Dmel_CG3105, CG3105, Dmel\\CG3105, pask', description='PAS kinase', other_designations='CG3105-PA|Pask-PA', map_location='60A16-60B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23988738, end_position_on_the_genomic_accession=23992049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37825, CurrentID=0, Status='live', Symbol='CG3860', Aliases='Dmel_ Dmel\\CG3860', description='CG3860 gene product from transcript CG3860-RA', other_designations='CG3860-PA|CG3860-PB', map_location='60B1-60B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23992138, end_position_on_the_genomic_accession=23995055, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  34513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34513, CurrentID=0, Status='live', Symbol='CG14072', Aliases='Dmel_ Dmel\\CG14072', description='CG14072 gene product from transcript CG14072-RB', other_designations='CG14072-PA|CG14072-PB|CG14072-PC|uncharacterized protein', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10936662, end_position_on_the_genomic_accession=10937931, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772205, CurrentID=0, Status='live', Symbol='CG33772', Aliases='Dmel_ BP1031, Dmel\\CG33772', description='CG33772 gene product from transcript CG33772-RC', other_designations='CG33772-PC|uncharacterized protein', map_location='79F3-79F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22651855, end_position_on_the_genomic_accession=22652525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37830, CurrentID=0, Status='live', Symbol='CG4324', Aliases='Dmel_ Dmel\\CG4324', description='CG4324 gene product from transcript CG4324-RC', other_designations='CG4324-PA|CG4324-PC', map_location='60B2-60B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24014917, end_position_on_the_genomic_accession=24017198, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37831, CurrentID=0, Status='live', Symbol='CG3121', Aliases='Dmel_ Dmel\\CG3121, anon-WO0140519.109', description='CG3121 gene product from transcript CG3121-RA', other_designations='CG3121-PA', map_location='60B3-60B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24017421, end_position_on_the_genomic_accession=24019896, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37832, CurrentID=0, Status='live', Symbol='CG3907', Aliases='Dmel_ Dmel\\CG3907', description='CG3907 gene product from transcript CG3907-RB', other_designations='CG3907-PA|CG3907-PB', map_location='60B3-60B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24019898, end_position_on_the_genomic_accession=24022298, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37833, CurrentID=0, Status='live', Symbol='CG11388', Aliases='Dmel_ Dmel\\CG11388, anon-EST:Posey287, dibeta3GnT6', description='CG11388 gene product from transcript CG11388-RA', other_designations='CG11388-PA|uncharacterized protein', map_location='60B4-60B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24022343, end_position_on_the_genomic_accession=24023981, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37834, CurrentID=0, Status='live', Symbol='Adk2', Aliases='Dmel_CG3140, ADK-2-44B, Adk2-60B, BcDNA:SD09634, CG3140, Dak2, Dmel\\CG3140, anon-Dak2, l(2)04201, l(2)04209, l(2)k04201, l(2)k04209, l(2)k16120, Adk2', description='Adenylate kinase 2', other_designations='Adk2-44B|Adk2-60B|Adk2-PA|CG3140-PA|adenylate kinase 2|adenylate kinase-2', map_location='60B4-44B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24024207, end_position_on_the_genomic_accession=24025877, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37836, CurrentID=0, Status='live', Symbol='CG3163', Aliases='Dmel_ Dmel\\CG3163', description='CG3163 gene product from transcript CG3163-RA', other_designations='CG3163-PA', map_location='60B4-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24029317, end_position_on_the_genomic_accession=24030555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37837, CurrentID=0, Status='live', Symbol='Chi', Aliases='Dmel_CG3924, CG3924, CHIP, Dmel\\CG3924, LDB, Ldb, chip, dLDB/Chip, dLdb, l(2)04405, l(2)k04405', description='Chip', other_designations='CG3924-PA|CG3924-PB|Chi-PA|Chi-PB|chip', map_location='2-106.8 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24031538, end_position_on_the_genomic_accession=24033943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37838, CurrentID=0, Status='live', Symbol='MAN1', Aliases='Dmel_CG3167, BcDNA:LD24895, CG3167, Dmel\\CG3167, d dMan1', description='CG3167 gene product from transcript CG3167-RC', other_designations='CG3167-PA|CG3167-PB|CG3167-PC|MAN1-PA|MAN1-PB|MAN1-PC', map_location='60B5-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24033974, end_position_on_the_genomic_accession=24036707, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37839, CurrentID=0, Status='live', Symbol='Not11', Aliases='Dmel_CG13567, C2orf29, CG13567, Dmel\\CG13567, NOT11', description='CG13567 gene product from transcript CG13567-RC', other_designations='CG13567-PA|CG13567-PB|CG13567-PC|Not11-PA|Not11-PB|Not11-PC', map_location='60B5-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24036720, end_position_on_the_genomic_accession=24037915, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37840, CurrentID=0, Status='live', Symbol='IntS1', Aliases='Dmel_CG3173, CG3173, Dmel\\CG3173, Int1', description='Integrator 1', other_designations='CG3173-PA|IntS1-PA|integrator 1', map_location='60B5-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24037943, end_position_on_the_genomic_accession=24044458, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37841, CurrentID=0, Status='live', Symbol='tsr', Aliases='Dmel_CG4254, 12/8, ADF, ADF/cofilin, Adf, CG4254, Cadf, D-cof, D61, Dmel\\CG4254, T8, Tsr, cofilin, cofilin/ADF, cofilin/actin depolymerizing factor homolog, l(2)03101, l(2)k03101, l(2)k05633, l(2)k13409, ntf', description='twinstar', other_designations='ADFcofilin|CG4254-PA|Cofilin/actin depolymerizing factor-like|DCofilin|actin-depolymerizing factor|cofilin|no terminal filament|tsr-PA|twin star', map_location='60B5-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24044453, end_position_on_the_genomic_accession=24046805, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37842, CurrentID=0, Status='live', Symbol='gammaSnap1', Aliases='Dmel_CG3988, CG3988, Dmel\\CG3988, gSnap, gamma SNAP, gamma-SNAP, gamma-Snap, gammaSNAP-2, gammaSnap', description='gamma Soluble NSF attachment protein 1', other_designations='CG3988-PA|gamma SNAP|gamma soluble NSF attachment protein 1|gamma-soluble NSF attachment protein|gammaSnap|gammaSnap1-PA', map_location='60B5-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24047145, end_position_on_the_genomic_accession=24048903, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37843, CurrentID=0, Status='live', Symbol='sei', Aliases='Dmel_CG3182, CG3182, Derg, Dmel\\CG3182, erg', description='seizure', other_designations='CG3182-PA|CG3182-PB|CG3182-PC|Ether-a-go-go-related|Seizure/HERG|eag related gene|sei-PA|sei-PB|sei-PC|seizure potassium channel', map_location='2-106 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24047411, end_position_on_the_genomic_accession=24053035, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  37844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37844, CurrentID=0, Status='live', Symbol='ppk29', Aliases='Dmel_CG13568, CG13568, Dmel\\CG13568, nope', description='pickpocket 29', other_designations='CG13568-PD|CG13568-PE|CG13568-PF|CG13568-PG|CG13568-PH|not passionate|ppk29-PD|ppk29-PE|ppk29-PF|ppk29-PG|ppk29-PH', map_location='60B6-60B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24052917, end_position_on_the_genomic_accession=24057137, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  37845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37845, CurrentID=0, Status='live', Symbol='CG13563', Aliases='Dmel_ Dmel\\CG13563', description='CG13563 gene product from transcript CG13563-RB', other_designations='CG13563-PB|uncharacterized protein', map_location='60B6-60B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24055306, end_position_on_the_genomic_accession=24057599, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37846, CurrentID=0, Status='live', Symbol='eIF-5A', Aliases='Dmel_CG3186, 1296, CG3186, Dmel\\CG3186, anon-EST:fe2H10, chr2R:19566878..19566998, eIF5A, eIF5a, l(2)01296', description='CG3186 gene product from transcript CG3186-RB', other_designations='CG3186-PA|CG3186-PB|anon-fast-evolving-2H10|eIF-5A-PA|eIF-5A-PB|eukaryotic initiation factor 5A|eukaryotic translation factor 5A', map_location='60B7-60B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24057798, end_position_on_the_genomic_accession=24059483, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37849, CurrentID=0, Status='live', Symbol='RpL39', Aliases='Dmel_CG3997, BcDNA:RH27094, CG3997, Dmel\\CG3997, L39, RPl46, Rp46, RpL46, anon-EST:Posey16, rp46', description='Ribosomal protein L39', other_designations='CG3997-PA|RpL39-PA|ribosomal protein L39|ribosomal protein L46', map_location='60B7-60B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24061853, end_position_on_the_genomic_accession=24062431, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37850, CurrentID=0, Status='live', Symbol='Rap2l', Aliases='Dmel_CG3204, CG3204, Dm Rap21, Dmel\\CG3204, O96692, Rap2, Rap21, Rap2L', description='Ras-associated protein 2-like', other_designations='CG3204-PA|CG3204-PC|CG3204-PD|Rap2l-PA|Rap2l-PC|Rap2l-PD', map_location='60B7-60B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24063051, end_position_on_the_genomic_accession=24065405, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37851, CurrentID=0, Status='live', Symbol='yki', Aliases='Dmel_CG4005, CG4005, Dmel\\CG4005, YAP, YKI, Yki, yap', description='yorkie', other_designations='CG4005-PD|CG4005-PF|CG4005-PG|yki-PD|yki-PF|yki-PG', map_location='60B7-60B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24065975, end_position_on_the_genomic_accession=24068485, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37852, CurrentID=0, Status='live', Symbol='CG3209', Aliases='Dmel_ Dmel\\CG3209, GPAT4', description='CG3209 gene product from transcript CG3209-RE', other_designations='CG3209-PB|CG3209-PC|CG3209-PD|CG3209-PE', map_location='60B8-60B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24068757, end_position_on_the_genomic_accession=24073552, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37853, CurrentID=0, Status='live', Symbol='Mlp60A', Aliases='Dmel_CG42309, BcDNA:RE14544, BcDNA:RH74176, CG30174, CG30179, CG3220, CG33149, CG42309, DLim-2, DMLP1, DMlp1, DmLIM-2, Dmel\\CG42309, Dmel_CG30174, Dmel_CG30179, Dmel_CG33149, FBgn0259209, MLIM 60A, MLP, MLP1, Mlp1, Mlp60, dMlp60A', description='Muscle LIM protein at 60A', other_designations='CG42309-PB|CG42309-PC|CG42309-PE|CG42309-PF|Mlp60A-PB|Mlp60A-PC|Mlp60A-PE|Mlp60A-PF|Muscle-specific LIM protein at 60A|muscle LIM protein 60A|muscle LIM protein at 60A', map_location='60B8-60B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24074602, end_position_on_the_genomic_accession=24078223, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37854, CurrentID=0, Status='live', Symbol='CG10339', Aliases='Dmel_ Dmel\\CG10339', description='CG10339 gene product from transcript CG10339-RA', other_designations='CG10339-PA|uncharacterized protein', map_location='60B8-60B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24078234, end_position_on_the_genomic_accession=24080527, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37855, CurrentID=0, Status='live', Symbol='CG13564', Aliases='Dmel_ Dmel\\CG13564, Minos1c', description='CG13564 gene product from transcript CG13564-RA', other_designations='CG13564-PA|mitochondrial INner membrane organizing system 1|uncharacterized protein', map_location='60B8-60B8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24080925, end_position_on_the_genomic_accession=24081424, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37856, CurrentID=0, Status='live', Symbol='CG16786', Aliases='Dmel_ Dmel\\CG16786', description='CG16786 gene product from transcript CG16786-RC', other_designations='CG16786-PA|CG16786-PB|CG16786-PC|uncharacterized protein', map_location='60B8-60B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24081782, end_position_on_the_genomic_accession=24086287, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37857, CurrentID=0, Status='live', Symbol='snama', Aliases='Dmel_CG3231, BCDNA:LD21643, BcDNA:LD21643, CG3231, Dmel\\CG3231, FBgn0086129, SNAMA, Snama, l(2)rQ313, mnm', description='something that sticks like glue', other_designations='CG3231-PA|CG3231-PB|lethal (2) rQ313|mini-me|retinoblastoma binding protein 6 family|snama-PA|snama-PB', map_location='60B9-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24086604, end_position_on_the_genomic_accession=24092014, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37858, CurrentID=0, Status='live', Symbol='gek', Aliases='Dmel_CG4012, CG4012, Dmel\\CG4012, Gek, MRCKalpha, MdaPk, Mrck, anon-WO0200864.1, dMRCK, l(2)09373', description='genghis khan', other_designations='CG4012-PA|chingis khan|gek-PA|genghis kahn|myotonic dystrophy kinase related', map_location='60B9-60B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24092079, end_position_on_the_genomic_accession=24098636, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  37859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37859, CurrentID=0, Status='live', Symbol='enok', Aliases='Dmel_CG11290, CG11290, Cg11290, Dmel\\CG11290, Enok, KAT6, S1, anon-60Ba, anon-WO0200864.1, anon-WO0200864.2, dmHAG406, rot', description='enoki mushroom', other_designations='CG11290-PA|CG11290-PB|enok-PA|enok-PB|rotkehlchen', map_location='60B10-60B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24099210, end_position_on_the_genomic_accession=24107032, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37860, CurrentID=0, Status='live', Symbol='CG4049', Aliases='Dmel_ Arip4, BEST:LD04705, Dmel\\CG4049', description='CG4049 gene product from transcript CG4049-RA', other_designations='CG4049-PA|R-gene', map_location='60B10-60B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24107087, end_position_on_the_genomic_accession=24114292, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37861, CurrentID=0, Status='live', Symbol='CG3253', Aliases='Dmel_ BEST:GH04269, Dmel\\CG3253, dibeta3GnT1', description='CG3253 gene product from transcript CG3253-RB', other_designations='CG3253-PA|CG3253-PB|L-gene', map_location='60B10-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24114398, end_position_on_the_genomic_accession=24116359, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37862, CurrentID=0, Status='live', Symbol='CG3257', Aliases='Dmel_ Dmel\\CG3257, anon-EST:Posey117', description='CG3257 gene product from transcript CG3257-RA', other_designations='CG3257-PA', map_location='60B11-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24116792, end_position_on_the_genomic_accession=24123263, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  50001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50001, CurrentID=0, Status='live', Symbol='CG15864', Aliases='Dmel_ CG7553, Dmel\\CG15864', description='CG15864 gene product from transcript CG15864-RB', other_designations='CG15864-PB|uncharacterized protein', map_location='84F6-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8248438, end_position_on_the_genomic_accession=8250324, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37864, CurrentID=0, Status='live', Symbol='tamo', Aliases='Dmel_CG4057, 60B, CG4057, Dmel\\CG4057', description='CG4057 gene product from transcript CG4057-RA', other_designations='CG4057-PA|CG4057-PB|CG4057-PC|tamo-PA|tamo-PB|tamo-PC', map_location='60B11-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24126638, end_position_on_the_genomic_accession=24133551, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37866, CurrentID=0, Status='live', Symbol='CG4065', Aliases='Dmel_ Dmel\\CG4065, anon-EST:Nelson1', description='CG4065 gene product from transcript CG4065-RB', other_designations='CG4065-PA|CG4065-PB', map_location='60B12-60B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24135180, end_position_on_the_genomic_accession=24138050, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37867, CurrentID=0, Status='live', Symbol='Dat', Aliases='Dmel_CG3318, AA-NAT1, AANAT, AANATA, AANATB, Aanat1, CG3318, DAT1, Dmel\\CG3318, NAT1, aaNAT, aaNAT1, aaNat, Dat', description='Dopamine N acetyltransferase', other_designations='CG3318-PA|CG3318-PB|Dat-PA|Dat-PB|arylalkylamine N-acetyltransferase|arylalkylamine N-acetyltransferase-1|dopamine N acetyltransferase|dopamine N-acetyltransferase|dopamine acetyltransferase', map_location='2-107 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24139563, end_position_on_the_genomic_accession=24157762, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37869, CurrentID=0, Status='live', Symbol='DnaJ-60', Aliases='Dmel_CG42567, CG12240, CG42567, DmJ60, Dmel\\CG42567, Dmel_CG12240, DnaJ60, anon-WO0172774.147', description='DnaJ-like-60', other_designations='CG42567-PA|CG42567-PC|DnaJ-60-PA|DnaJ-60-PC', map_location='60C1-60C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24157668, end_position_on_the_genomic_accession=24159312, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37871, CurrentID=0, Status='live', Symbol='CG3328', Aliases='Dmel_ Dmel\\CG3328', description='CG3328 gene product from transcript CG3328-RB', other_designations='CG3328-PA|CG3328-PB|CG3328-PC', map_location='60C1-60C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24162489, end_position_on_the_genomic_accession=24173660, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  37872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37872, CurrentID=0, Status='live', Symbol='mRpS17', Aliases='Dmel_CG4326, CG4326, Dmel\\CG4326, MRP-S17', description='mitochondrial ribosomal protein S17', other_designations='CG4326-PA|mRpS17-PA', map_location='60C1-60C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24173665, end_position_on_the_genomic_accession=24174441, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37873, CurrentID=0, Status='live', Symbol='Nop60B', Aliases='Dmel_CG3333, CG333, CG3333, Dmel\\CG3333, FBgn0023184, MFL/Nop60B, Nop60, Nop60b, Uhg6, dUhg6, l(2)k05318, mfl, minifly, nop60B', description='Nucleolar protein at 60B', other_designations='CG3333-PA|CG3333-PB|CG3333-PC|CG3333-PF|CG3333-PG|Nop60B-PA|Nop60B-PB|Nop60B-PC|Nop60B-PF|Nop60B-PG|Uhg6|dUhg-like 6|minifly|nucleolar protein at 60B', map_location='60C1-60C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24174877, end_position_on_the_genomic_accession=24186343, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37874, CurrentID=0, Status='live', Symbol='ocm', Aliases='Dmel_CG3363, CG3363, Dmel\\CG3363, Mga, Mod(GMroX1)A', description='over compensating males', other_designations='166 complementation group|CG3363-PA|CG3363-PB|CG3363-PC|complementation group A|ocm-PA|ocm-PB|ocm-PC', map_location='60C2-60C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24186703, end_position_on_the_genomic_accession=24195290, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37875, CurrentID=0, Status='live', Symbol='cN-IIIB', Aliases='Dmel_CG3362, 16769374, CG3362, Dmel\\CG3362, cNIII-like', description="cytosolic 5'-nucleotidase IIIB", other_designations='CG3362-PA|cN-IIIB-PA|cytosolic nucleotidase III-like', map_location='60C2-60C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24195498, end_position_on_the_genomic_accession=24197027, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37876, CurrentID=0, Status='live', Symbol='CG3356', Aliases='Dmel_ Dmel\\CG3356', description='CG3356 gene product from transcript CG3356-RA', other_designations='CG3356-PA', map_location='60C2-60C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24197824, end_position_on_the_genomic_accession=24202187, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37877, CurrentID=0, Status='live', Symbol='CG11406', Aliases='Dmel_ Dmel\\CG11406', description='CG11406 gene product from transcript CG11406-RB', other_designations='CG11406-PA|CG11406-PB|CG11406-PC|CG11406-PD|uncharacterized protein', map_location='60C2-60C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24202091, end_position_on_the_genomic_accession=24208203, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37879, CurrentID=0, Status='live', Symbol='CG30419', Aliases='Dmel_ CG13572, CG13573, Dmel\\CG30419', description='CG30419 gene product from transcript CG30419-RA', other_designations='CG30419-PA|uncharacterized protein', map_location='60C2-60C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24212491, end_position_on_the_genomic_accession=24229795, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37881, CurrentID=0, Status='live', Symbol='Ir60a', Aliases='Dmel_CG13576, CG13576, DmelIR60a, Dmel\\CG13576, IR60a', description='Ionotropic receptor 60a', other_designations='CG13576-PA|Ir60a-PA|ionotropic receptor 60a', map_location='60C4-60C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24249055, end_position_on_the_genomic_accession=24251303, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37882, CurrentID=0, Status='live', Symbol='nord', Aliases='Dmel_CG30418, CG13574, CG30418, CG3372, Dmel\\CG30418, E3272', description='CG30418 gene product from transcript CG30418-RA', other_designations='CG30418-PA|nord-PA', map_location='60C3-60C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24243255, end_position_on_the_genomic_accession=24255094, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37883, CurrentID=0, Status='live', Symbol='CG13575', Aliases='Dmel_ Dmel\\CG13575, anon-WO0170980.148, anon-WO0170980.149', description='CG13575 gene product from transcript CG13575-RA', other_designations='CG13575-PA|uncharacterized protein', map_location='60C4-60C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24255327, end_position_on_the_genomic_accession=24257414, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37884, CurrentID=0, Status='live', Symbol='CG3376', Aliases='Dmel_ Dmel\\CG3376, aSMase', description='CG3376 gene product from transcript CG3376-RB', other_designations='CG3376-PA|CG3376-PB|CG3376-PC|CG3376-PD', map_location='60C4-60C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24258013, end_position_on_the_genomic_accession=24265213, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37885, CurrentID=0, Status='live', Symbol='CG13577', Aliases='Dmel_ Dmel\\CG13577', description='CG13577 gene product from transcript CG13577-RB', other_designations='CG13577-PB|CG13577-PC|uncharacterized protein', map_location='60C4-60C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24267377, end_position_on_the_genomic_accession=24269179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37886, CurrentID=0, Status='live', Symbol='nvy', Aliases='Dmel_CG3385, CG3385, Dmel\\CG3385, Nvy', description='nervy', other_designations='CG3385-PA|CG3385-PB|CG3385-PC|CG3385-PD|nvy-PA|nvy-PB|nvy-PC|nvy-PD', map_location='60C5-60C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24275734, end_position_on_the_genomic_accession=24291084, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37887, CurrentID=0, Status='live', Symbol='CG3394', Aliases='Dmel_ Dmel\\CG3394', description='CG3394 gene product from transcript CG3394-RA', other_designations='CG3394-PA|CG3394-PB', map_location='60C6-60C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24294097, end_position_on_the_genomic_accession=24298111, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37888, CurrentID=0, Status='live', Symbol='betaTub60D', Aliases='Dmel_CG3401, 143391_i_at, 3t, B3t, BETA 60D, CG3401, D.m.BETA-60D, DTB3, Dmbeta3, Dmel\\CG3401, T, Tub, Tub60D, beta-Tub60D, beta-Tub6D, beta-tub, beta3, beta3 TU, beta3-Tub, beta3-tubulin, beta3TUB, beta3Tub, beta3t, beta3tub, beta60C, betaTub, betaTub3, betaTub60C, beta[[3]] tubulin, beta[[3]]-Tub, beta[[3]]-tubulin, betatub60D, p50, p50/tubulin, p53, p53/tubulin', description='beta-Tubulin at 60D', other_designations='60C beta tubulin|B3-tubulin|CG3401-PA|CG3401-PB|b-3 tubulin|b3 tubulin|beta 3 tubulin|beta tubulin|beta-3 tubulin|beta-3-tubulin|beta-3tubulin|beta-Tubulin|beta-Tubulin at 60A|beta-tubulin 60D|beta-tubulin at 60D|beta-tubulin60D|beta3 tubulin|beta3-tubulin|beta3tubulin|betaTub60D-PA|betaTub60D-PB|tubulin|tubulin b-3', map_location='60C6-60C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24305881, end_position_on_the_genomic_accession=24313099, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37889, CurrentID=0, Status='live', Symbol='slbo', Aliases='Dmel_CG4354, C/EBP, CG4354, DC/EBP, DM8, Dm-c/EBP, DmC/EBP, Dmel\\CG4354, SLBO, Slbo, fs(2)7, fs(2)8, fs(2)ry7, fs(2)ry8, slobo', description='slow border cells', other_designations='CCAAT/enhancer binding protein|CG4354-PA|CG4354-PB|slbo-PA|slbo-PB', map_location='60C6-60C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24331522, end_position_on_the_genomic_accession=24334307, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37890, CurrentID=0, Status='live', Symbol='bs', Aliases='Dmel_CG3411, BS, Bs, CG3411, D-SRF, DSRF, DSrf, DSrf/bs, DmSRF, Dmel\\CG3411, Group IIc, Mal, SRF, Serf, Srf, ba/DSRF, bs/Dsrf, dSRF, dsrf, l(2)03267, pruned/DSRF, px, srf, bs', description='blistered', other_designations='CG3411-PA|CG3411-PB|CG3411-PC|SRF|balloon|blistered/DSRF|blistered/Drosophila serum response factor|bs-PA|bs-PB|bs-PC|drosophila serum response factor|pruned|serum response factor', map_location='2-107.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24342861, end_position_on_the_genomic_accession=24376473, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740715, CurrentID=0, Status='live', Symbol='CG34191', Aliases='Dmel_ Dmel\\CG34191', description='CG34191 gene product from transcript CG34191-RA', other_designations='CG34191-PA|CG34191-PB|uncharacterized protein', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16951837, end_position_on_the_genomic_accession=16952581, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37892, CurrentID=0, Status='live', Symbol='mAChR-A', Aliases='Dmel_CG4356, Acr60C, AcrC, CG4356, DM1, Dm-mAChR-A, Dm1, Dmel\\CG4356, mAChR, mAChR A, mAChR-1, mAcR, mAcR-60C, mAcR60C', description='muscarinic Acetylcholine Receptor, A-type', other_designations='A-type mAChR|CG4356-PA|CG4356-PB|mAChR-A-PA|mAChR-A-PB|muscarinic acetylcholine Receptor, A-type|muscarinic acetylcholine receptor|muscarinic acetylcholine receptor 60C|muscarinic receptor', map_location='60C7-60C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24378630, end_position_on_the_genomic_accession=24390294, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37893, CurrentID=0, Status='live', Symbol='Slik', Aliases='Dmel_CG4527, CG4527, DPlkk1, Dmel\\CG4527, PLKK1, Plkk/Slik, Plkk1, SLIK, SLK1, Slk, dPlkk, dPlkk/Slik, slik, slk, Slik', description='Sterile20-like kinase', other_designations='CG4527-PA|CG4527-PB|CG4527-PD|CG4527-PE|CG4527-PF|CG4527-PG|Slik-PA|Slik-PB|Slik-PD|Slik-PE|Slik-PF|Slik-PG|Ste20 kinase|Sterile-20|dPlkk/Slik threonine kinase|polo kinase kinase 1', map_location='60C7-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24392129, end_position_on_the_genomic_accession=24402680, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  37894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37894, CurrentID=0, Status='live', Symbol='Rpn8', Aliases='Dmel_CG3416, CG3416, Dmel\\CG3416, Mov34, P26270, PRSC_DROME, RPN8, S12, l(2)k08003, p39B', description='Regulatory particle non-ATPase 8', other_designations='CG3416-PA|CG3416-PB|CG3416-PC|Mov34|Rpn8-PA|Rpn8-PB|Rpn8-PC|regulatory particle non-ATPase 8', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24400028, end_position_on_the_genomic_accession=24404527, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37895, CurrentID=0, Status='live', Symbol='SerT', Aliases='Dmel_CG4545, CG4545, DmSERT, DmSerT, Dmel\\CG4545, SERT, Sert, dSERT, dSERT1, dSerT, dSert, dsert, serT', description='Serotonin transporter', other_designations='CG4545-PA|CG4545-PB|SerT-PA|SerT-PB|cocaine-sensitive serotonin transporter|drosophila serotonin transporter|serotonin transporter', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24404595, end_position_on_the_genomic_accession=24411074, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37897, CurrentID=0, Status='live', Symbol='CG42383', Aliases='Dmel_ CG4556, Dmel\\CG42383, Dmel_CG4556, eyc, p47, p47 NSFL1', description='CG42383 gene product from transcript CG42383-RA', other_designations='CG42383-PA|eyes closed|uncharacterized protein', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24423926, end_position_on_the_genomic_accession=24425102, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37898, CurrentID=0, Status='live', Symbol='CG15873', Aliases='Dmel_ Dmel\\CG15873, SP184, SPH184', description='CG15873 gene product from transcript CG15873-RA', other_designations='CG15873-PA|uncharacterized protein', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24429578, end_position_on_the_genomic_accession=24430699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740606, CurrentID=0, Status='live', Symbol='CG34429', Aliases='Dmel_ CG14116, Dmel\\CG34429, Dmel_CG14116', description='CG34429 gene product from transcript CG34429-RA', other_designations='CG34429-PA|uncharacterized protein', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13063046, end_position_on_the_genomic_accession=13063929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37900, CurrentID=0, Status='live', Symbol='CG3483', Aliases='Dmel_ Dmel\\CG3483, anon-WO0140519.111', description='CG3483 gene product from transcript CG3483-RA', other_designations='CG3483-PA', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24445500, end_position_on_the_genomic_accession=24446839, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37901, CurrentID=0, Status='live', Symbol='CG4563', Aliases='Dmel_ Dmel\\CG4563', description='CG4563 gene product from transcript CG4563-RA', other_designations='CG4563-PA', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24451084, end_position_on_the_genomic_accession=24453040, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37902, CurrentID=0, Status='live', Symbol='CG3492', Aliases='Dmel_ Dmel\\CG3492, anon-WO0140519.253', description='CG3492 gene product from transcript CG3492-RA', other_designations='CG3492-PA', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24472712, end_position_on_the_genomic_accession=24474226, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37903, CurrentID=0, Status='live', Symbol='CG3494', Aliases='Dmel_ CT11751, Dmel\\CG3494', description='CG3494 gene product from transcript CG3494-RA', other_designations='CG3494-PA|CG3494-PB', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24474299, end_position_on_the_genomic_accession=24475261, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37904, CurrentID=0, Status='live', Symbol='CG16837', Aliases='Dmel_ Dmel\\CG16837', description='CG16837 gene product from transcript CG16837-RA', other_designations='CG16837-PA|uncharacterized protein', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24475386, end_position_on_the_genomic_accession=24476032, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37905, CurrentID=0, Status='live', Symbol='CG13579', Aliases='Dmel_ Dmel\\CG13579, anon-WO0170980.151, anon-WO0170980.152', description='CG13579 gene product from transcript CG13579-RA', other_designations='CG13579-PA|CG13579-PB|uncharacterized protein', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24455292, end_position_on_the_genomic_accession=24490656, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37906, CurrentID=0, Status='live', Symbol='CG13589', Aliases='Dmel_ Dmel\\CG13589', description='CG13589 gene product from transcript CG13589-RA', other_designations='CG13589-PA|uncharacterized protein', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24492214, end_position_on_the_genomic_accession=24492855, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37907, CurrentID=0, Status='live', Symbol='CG13590', Aliases='Dmel_ Dmel\\CG13590', description='CG13590 gene product from transcript CG13590-RA', other_designations='CG13590-PA|uncharacterized protein', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24493256, end_position_on_the_genomic_accession=24493948, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37909, CurrentID=0, Status='live', Symbol='Ssl', Aliases='Dmel_CG13591, CG13591, CK2betates, CKIIb1, DmCK2betates, Dmel\\CG13591, NEST:bs30a09, SSL, Su(Ste)l, betaCK2tes', description='Suppressor of Stellate-like', other_designations='CG13591-PA|Ssl-PA|Su(Ste)-like|regulatory subunit of protein kinase CK2|suppressor of Stellate-like', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24499243, end_position_on_the_genomic_accession=24500267, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37910, CurrentID=0, Status='live', Symbol='Prosalpha4T2', Aliases='Dmel_CG4569, CG4569, Dmel\\CG4569, Pros28.1B, a4-t2, alpha4T2, alpha4t2, alpha4t2_dm, dalpha7, pros28.1B', description='Proteasome alpha4 subunit, Testis-specific 2', other_designations='20s proteasome|CG4569-PA|CG4569-PB|Pros28.1B|Prosalpha4T2-PA|Prosalpha4T2-PB|proteasome 28kD subunit 1B|proteasome alpha4 subunit, Testis-specific 2|testes-specific proteasome subunit', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24500780, end_position_on_the_genomic_accession=24502277, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37911, CurrentID=0, Status='live', Symbol='CG13581', Aliases='Dmel_ BcDNA:AT03240, Dmel\\CG13581', description='CG13581 gene product from transcript CG13581-RA', other_designations='CG13581-PA|uncharacterized protein', map_location='60D3-60D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24506870, end_position_on_the_genomic_accession=24507706, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37912, CurrentID=0, Status='live', Symbol='Letm1', Aliases='Dmel_CG4589, B4D, CG4589, DmLETM1, Dmel\\CG4589, anon-60Da, anon60Da, dLetm1', description='LETM1 ortholog', other_designations='CG4589-PA|CG4589-PB|CG4589-PC|Letm1-PA|Letm1-PB|Letm1-PC', map_location='60D3-60D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24509508, end_position_on_the_genomic_accession=24515320, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37914, CurrentID=0, Status='live', Symbol='CG4612', Aliases='Dmel_ Dmel\\CG4612, cg4612', description='CG4612 gene product from transcript CG4612-RC', other_designations='CG4612-PA|CG4612-PC', map_location='60D3-60D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24521874, end_position_on_the_genomic_accession=24526089, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37916, CurrentID=0, Status='live', Symbol='Brca2', Aliases='Dmel_CG30169, 30169, BRCA2, BcDNA:SD25109, CG13583, CG13584, CG30169, Dmbrca2, Dmel\\CG30169, brca2, dmbrca2', description='Breast cancer 2, early onset homolog', other_designations='Brca2-PA|CG30169-PA|breast cancer 2 early onset homolog|breast cancer 2, early onset homolog', map_location='60D4-60D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24526172, end_position_on_the_genomic_accession=24529581, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37917, CurrentID=0, Status='live', Symbol='Ir60e', Aliases='Dmel_CG13592, CG13592, DmelIR60e, Dmel\\CG13592, IR60e', description='Ionotropic receptor 60e', other_designations='CG13592-PD|Ir60e-PD|ionotropic receptor 60e', map_location='60D4-60D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24529554, end_position_on_the_genomic_accession=24531272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37918, CurrentID=0, Status='live', Symbol='CG13585', Aliases='Dmel_ BcDNA:AT28250, Dmel\\CG13585', description='CG13585 gene product from transcript CG13585-RA', other_designations='CG13585-PA|CG13585-PB|CG13585-PC|CG13585-PD|uncharacterized protein', map_location='60D4-60D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24531594, end_position_on_the_genomic_accession=24533391, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37919, CurrentID=0, Status='live', Symbol='CG4622', Aliases='Dmel_ Dmel\\CG4622', description='CG4622 gene product from transcript CG4622-RC', other_designations='CG4622-PA|CG4622-PC', map_location='60D4-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24533374, end_position_on_the_genomic_accession=24556300, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37920, CurrentID=0, Status='live', Symbol='CG11413', Aliases='Dmel_ BcDNA:RH33060, CT31865, Dmel\\CG11413', description='CG11413 gene product from transcript CG11413-RA', other_designations='CG11413-PA|uncharacterized protein', map_location='60D4-60D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24536190, end_position_on_the_genomic_accession=24536791, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37921, CurrentID=0, Status='live', Symbol='ITP', Aliases='Dmel_CG13586, BcDNA:SD05282, CG13586, CHH-like, Dmel\\CG13586, Drm-ITP, DrmITP, DrmITPL1, DrmITPL2, Itp, chgh, itp', description='Ion transport peptide', other_designations='CG13586-PC|CG13586-PD|CG13586-PE|CG13586-PF|CG13586-PG|ITP-PC|ITP-PD|ITP-PE|ITP-PF|ITP-PG|ion transport peptide|ion transport polypeptide', map_location='60D4-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24537421, end_position_on_the_genomic_accession=24558961, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37922, CurrentID=0, Status='live', Symbol='Nurf-38', Aliases='Dmel_CG4634, 38, CG4634, Dmel\\CG4634, N38, NURF, NURF 38, NURF-38, NURF38, Nurf 38, Nurf38, dNURF, l(2)k16102, p38, p38/NURF-38', description='Nucleosome remodeling factor - 38kD', other_designations='CG4634-PA|CG4634-PB|CG4634-PC|Nurf-38-PA|Nurf-38-PB|Nurf-38-PC|inorganic pyrophosphatase|nucleosome remodeling factor|nucleosome remodeling factor - 38kD|nucleosome remodeling factor-38 kDa|pyrophosphatase', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24558967, end_position_on_the_genomic_accession=24561087, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37923, CurrentID=0, Status='live', Symbol='CG11414', Aliases='Dmel_ Dmel\\CG11414', description='CG11414 gene product from transcript CG11414-RA', other_designations='CG11414-PA|uncharacterized protein', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24561607, end_position_on_the_genomic_accession=24565058, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37924, CurrentID=0, Status='live', Symbol='uri', Aliases='Dmel_CG11416, CG11416, Dmel\\CG11416, dURI, ori', description='unconventional prefoldin RPB5 interactor', other_designations='CG11416-PA|origami|uri-PA', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24565554, end_position_on_the_genomic_accession=24568082, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37925, CurrentID=0, Status='live', Symbol='Fcp1', Aliases='Dmel_CG12252, 21483550, CG12252, DmFcp1, Dmel\\CG12252, Dmfcp1, fcp1', description='TFIIF-interacting CTD phosphatase', other_designations='CG12252-PA|CG12252-PB|Fcp1-PA|Fcp1-PB|RNApolymerase II carboxyl-terminal domain phosphatase', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24567987, end_position_on_the_genomic_accession=24571821, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37926, CurrentID=0, Status='live', Symbol='CG3511', Aliases='Dmel_ Dmel\\CG3511', description='CG3511 gene product from transcript CG3511-RA', other_designations='CG3511-PA|CG3511-PB', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24571627, end_position_on_the_genomic_accession=24574503, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  37927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37927, CurrentID=0, Status='live', Symbol='Start1', Aliases='Dmel_CG3522, CG3522, Dmel\\CG3522, StART1, Start, start1', description='CG3522 gene product from transcript CG3522-RB', other_designations='CG3522-PA|CG3522-PB|CG3522-PC|CG3522-PD|Start1-PA|Start1-PB|Start1-PC|Start1-PD', map_location='60D5-60D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24574559, end_position_on_the_genomic_accession=24577676, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  37928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37928, CurrentID=0, Status='live', Symbol='CG4681', Aliases='Dmel_ Dmel\\CG4681, anon-WO0140519.29', description='CG4681 gene product from transcript CG4681-RA', other_designations='CG4681-PA', map_location='60D5-60D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24578247, end_position_on_the_genomic_accession=24579827, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37929, CurrentID=0, Status='live', Symbol='pio', Aliases='Dmel_CG3541, CG30167, CG3541, Dmel\\CG3541, Pio', description='piopio', other_designations='CG3541-PB|CG3541-PC|CG3541-PD|CG3541-PE|pio-PB|pio-PC|pio-PD|pio-PE', map_location='60D7-60D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24581245, end_position_on_the_genomic_accession=24597862, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  37930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37930, CurrentID=0, Status='live', Symbol='CG13587', Aliases='Dmel_ Dmel\\CG13587', description='CG13587 gene product from transcript CG13587-RA', other_designations='CG13587-PA|uncharacterized protein', map_location='60D8-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24597995, end_position_on_the_genomic_accession=24599091, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37931, CurrentID=0, Status='live', Symbol='ATPsynF', Aliases='Dmel_CG4692, ATPK, CG4692, Dmel\\CG4692, anon-EST:Posey45', description='ATP synthase, subunit F', other_designations='ATP synthase|ATP synthase f chain|ATPsynF-PA|ATPsynF-PB|CG4692-PA|CG4692-PB|F0/F1 ATP synthase F', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24599088, end_position_on_the_genomic_accession=24599821, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37932, CurrentID=0, Status='live', Symbol='CG3548', Aliases='Dmel_ Dmel\\CG3548', description='CG3548 gene product from transcript CG3548-RA', other_designations='CG3548-PA', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24599947, end_position_on_the_genomic_accession=24602367, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37933, CurrentID=0, Status='live', Symbol='CG3565', Aliases='Dmel_ Dmel\\CG3565, anon-WO0140519.176', description='CG3565 gene product from transcript CG3565-RA', other_designations='CG3565-PA', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24602472, end_position_on_the_genomic_accession=24603488, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37934, CurrentID=0, Status='live', Symbol='CG3570', Aliases='Dmel_ Dmel\\CG3570', description='CG3570 gene product from transcript CG3570-RA', other_designations='CG3570-PA', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24603572, end_position_on_the_genomic_accession=24604670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37935, CurrentID=0, Status='live', Symbol='CG4707', Aliases='Dmel_ Cg4707, Dmel\\CG4707', description='CG4707 gene product from transcript CG4707-RA', other_designations='CG4707-PA', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24604595, end_position_on_the_genomic_accession=24607023, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37936, CurrentID=0, Status='live', Symbol='CG42361', Aliases='Dmel_ CG3579, Dmel\\CG42361, Dmel_CG13588, Dmel_CG3579', description='CG42361 gene product from transcript CG42361-RA', other_designations='CG42361-PA|uncharacterized protein', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24607673, end_position_on_the_genomic_accession=24608788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37937, CurrentID=0, Status='live', Symbol='CG42360', Aliases='Dmel_ CG13588, CG3579, Dmel\\CG42360, Dmel_CG13588, Dmel_CG3579', description='CG42360 gene product from transcript CG42360-RA', other_designations='CG42360-PA|uncharacterized protein', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24607267, end_position_on_the_genomic_accession=24610847, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  37938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37938, CurrentID=0, Status='live', Symbol='Adck', Aliases='Dmel_CG3608, CG3608, Dmel\\CG3608, cg3608', description='aarF domain containing kinase', other_designations='Adck-PA|Adck-PB|CG3608-PA|CG3608-PB', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24611703, end_position_on_the_genomic_accession=24613906, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37939, CurrentID=0, Status='live', Symbol='CG4741', Aliases='Dmel_ Dmel\\CG4741', description='CG4741 gene product from transcript CG4741-RB', other_designations='CG4741-PA|CG4741-PB|CG4741-PC', map_location='60D9-60D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24610801, end_position_on_the_genomic_accession=24615158, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37940, CurrentID=0, Status='live', Symbol='CG13594', Aliases='Dmel_ Dmel\\CG13594', description='CG13594 gene product from transcript CG13594-RB', other_designations='CG13594-PA|CG13594-PB|uncharacterized protein', map_location='60D9-60D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24615423, end_position_on_the_genomic_accession=24638588, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37941, CurrentID=0, Status='live', Symbol='Cyp9c1', Aliases='Dmel_CG3616, 9c1, CG3616, CYP9C1, Dmel\\CG3616, P-450, P450', description='Cytochrome P450-9c1', other_designations='CG3616-PA|Cyp9c1-PA|cytochrome P450-9c1', map_location='60D10-60D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24631063, end_position_on_the_genomic_accession=24632923, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37942, CurrentID=0, Status='live', Symbol='CG3640', Aliases='Dmel_ Dmel\\CG3640', description='CG3640 gene product from transcript CG3640-RA', other_designations='CG3640-PA', map_location='60D10-60D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24639688, end_position_on_the_genomic_accession=24640756, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37943, CurrentID=0, Status='live', Symbol='CG4781', Aliases='Dmel_ 4781, CT15357, Dmel\\CG4781', description='CG4781 gene product from transcript CG4781-RA', other_designations='CG4781-PA', map_location='60D10-60D10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24640683, end_position_on_the_genomic_accession=24642239, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37944, CurrentID=0, Status='live', Symbol='CG3663', Aliases='Dmel_ Dmel\\CG3663', description='CG3663 gene product from transcript CG3663-RA', other_designations='CG3663-PA', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24661503, end_position_on_the_genomic_accession=24662409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37945, CurrentID=0, Status='live', Symbol='CG30161', Aliases='Dmel_ BcDNA:AT15604, CG4856, Dmel\\CG30161', description='CG30161 gene product from transcript CG30161-RA', other_designations='CG30161-PA|uncharacterized protein', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24663532, end_position_on_the_genomic_accession=24664377, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37946, CurrentID=0, Status='live', Symbol='ND-19', Aliases='Dmel_CG3683, CG3683, Dmel\\CG3683, NDUFA8, NUPM, RH23043p', description='NADH dehydrogenase (ubiquinone) 19 kDa subunit', other_designations='CG3683-PA|CG3683-PB|CG3683-PC|CG3683-PD|FBpp0072231|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 8|NADH ubiquinone oxidoreductase 19 kDa|NADH-UBIQUINONE OXIDOREDUCTASE 19 KDA SUBUNIT|NADH:ubiquinone oxidoreductase 19 kDa subunit|ND-19-PA|ND-19-PB|ND-19-PC|ND-19-PD', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24664398, end_position_on_the_genomic_accession=24665439, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37947, CurrentID=0, Status='live', Symbol='Pof', Aliases='Dmel_CG3691, CG3691, Dmel\\CG3691, POF, cg3691, pof', description='Painting of fourth', other_designations='CG3691-PA|CG3691-PB|CG3691-PC|Pof-PA|Pof-PB|Pof-PC|painting of fourth|painting of fourth chromosome|painting on the fourth|painting-of-fourth', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24669397, end_position_on_the_genomic_accession=24671577, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37948, CurrentID=0, Status='live', Symbol='CG4806', Aliases='Dmel_ Dmel\\CG4806, cg4806', description='CG4806 gene product from transcript CG4806-RA', other_designations='CG4806-PA', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24666158, end_position_on_the_genomic_accession=24668343, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37949, CurrentID=0, Status='live', Symbol='Mmp1', Aliases='Dmel_CG4859, CG4859, Dm1-MMP, Dmel\\CG4859, MMP-1, MMP1, MMp1, Mmp 1, dMMP1, dm1-MMP, dmmp1, l(2)k04809, mmp1', description='Matrix metalloproteinase 1', other_designations='CG4859-PC|CG4859-PD|CG4859-PE|CG4859-PF|CG4859-PG|CG4859-PH|CG4859-PI|CG4859-PJ|CG4859-PK|CG4859-PL|Mmp1-PC|Mmp1-PD|Mmp1-PE|Mmp1-PF|Mmp1-PG|Mmp1-PH|Mmp1-PI|Mmp1-PJ|Mmp1-PK|Mmp1-PL|matrix metalloprotease|matrix metalloprotease 1|matrix metalloproteinase 1|matrix metalloproteinase-1|matrix metalloproteinase1|matrix metalloprotinase 1|metalloproteinase', map_location='60D13-60D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24671294, end_position_on_the_genomic_accession=24688184, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37950, CurrentID=0, Status='live', Symbol='ST6Gal', Aliases='Dmel_CG4871, CG4871, D.SiaT, D.SialT, DSiaT, DmPST, DmSialT, Dmel\\CG4871, SialT, dST6Gal I, dSiaT', description='Sialyltransferase', other_designations='CG4871-PA|CG4871-PB|ST6Gal-PA|ST6Gal-PB|alpha-2,6-sialyltransferase|alpha2,6-sialyltransferase|b-galactoside-a-2,6-sialyltransferase|sialyltransferase|sialyltransferase-like', map_location='60D14-60D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24691422, end_position_on_the_genomic_accession=24694386, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37953, CurrentID=0, Status='live', Symbol='Mid1', Aliases='Dmel_CG33988, CG13595, CG13596, CG18510, CG33988, Dmel\\CG33988, cg13596', description='Mid1 ortholog (S. cerevisiae)', other_designations='CG33988-PA|CG33988-PB|Mid1 ortholog|Mid1-PA|Mid1-PB', map_location='60D14-60D16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24695378, end_position_on_the_genomic_accession=24736657, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37954, CurrentID=0, Status='live', Symbol='Usp15-31', Aliases='Dmel_CG30421, CG30421, CG3872, CG9189, Dmel\\CG30421, Q9W117, USP43, Usp15, Usp31', description='Ubiquitin specific protease 15/31', other_designations='CG30421-PA|CG30421-PB|CG30421-PC|Usp15-31-PA|Usp15-31-PB|Usp15-31-PC|ubiquitin specific protease 15/31', map_location='60D16-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24743749, end_position_on_the_genomic_accession=24756976, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37956, CurrentID=0, Status='live', Symbol='spz6', Aliases='Dmel_CG9196, BcDNA:LP10725, CG9196, CT26280, Dmel\\CG9196, Spz6', description='spatzle 6', other_designations='CG9196-PA|spz6-PA', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24757378, end_position_on_the_genomic_accession=24760466, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37957, CurrentID=0, Status='live', Symbol='CG3880', Aliases='Dmel_ Dmel\\CG3880', description='CG3880 gene product from transcript CG3880-RA', other_designations='CG3880-PA', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24760664, end_position_on_the_genomic_accession=24762605, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37959, CurrentID=0, Status='live', Symbol='CG3894', Aliases='Dmel_ CG9197, Dmel\\CG3894', description='CG3894 gene product from transcript CG3894-RC', other_designations='CG3894-PC|dmNeural-like', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24763167, end_position_on_the_genomic_accession=24766750, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37960, CurrentID=0, Status='live', Symbol='GstE12', Aliases='Dmel_CG16936, BcDNA:GH04753, CG16936, Dmel\\CG16936, GST, GSTE12, gste12', description='Glutathione S transferase E12', other_designations='CG16936-PA|CG16936-PB|CG16936-PC|CG16936-PD|GstE12-PA|GstE12-PB|GstE12-PC|GstE12-PD|glutathione S transferase E12|glutathione S-transferase|glutathione transferase', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24766718, end_position_on_the_genomic_accession=24768172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37961, CurrentID=0, Status='live', Symbol='Eps-15', Aliases='Dmel_CG16932, CG16932, Dmel\\CG16932, EH1, EPS-15, EPS15, Eps15, anon-WO0118547.85, eps-15, eps15, i31', description='Epidermal growth factor receptor pathway substrate clone 15', other_designations='CG16932-PA|CG16932-PB|CG16932-PC|CG16932-PD|Eps-15-PA|Eps-15-PB|Eps-15-PC|Eps-15-PD|Eps15 homologue 1|epidermal growth factor receptor pathway substrate clone 15', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24768314, end_position_on_the_genomic_accession=24775265, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  37962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37962, CurrentID=0, Status='live', Symbol='egg', Aliases='Dmel_CG12196, AAF47268 Dm, AAM70794, BcDNA:AT13877, CG12196, CG30422, CG30426, Dm-setdb1, DmSetdb1, Dmel\\CG12196, Dmel_CG30422, Dmel_CG30426, Dmsetdb1, EGG, Egg, Eggless, SETDB1, SetDB1, Setdb1, dEset, dSETDB1, dSetdb1, dsetdb1', description='eggless', other_designations='CG12196-PA|Histone-lysine N-methyltransferase|dSETDB1|egg-PA', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24775534, end_position_on_the_genomic_accession=24779901, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  37963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37963, CurrentID=0, Status='live', Symbol='Lcp9', Aliases='Dmel_CG16914, CG16914, DmelLcp9, Dmel\\CG16914, LCP9, lcp9', description='Larval cuticle protein 9', other_designations='CG16914-PA|Lcp9-PA|larval cuticle protein 9', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24779979, end_position_on_the_genomic_accession=24780428, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37964, CurrentID=0, Status='live', Symbol='CG3594', Aliases='Dmel_ Dmel\\CG3594, Eap, cg3594', description='CG3594 gene product from transcript CG3594-RA', other_designations='CG3594-PA|Exu-associated protein', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24780917, end_position_on_the_genomic_accession=24781779, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37965, CurrentID=0, Status='live', Symbol='Aats-tyr-m', Aliases='Dmel_CG16912, Aatm, CG16912, Dmel\\CG16912, anon-EST:fe1E12', description='mitochondrial tyrosyl-tRNA synthetase', other_designations='Aats-tyr-m-PA|CG16912-PA|anon-fast-evolving-1E12', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24781726, end_position_on_the_genomic_accession=24783500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37966, CurrentID=0, Status='live', Symbol='CG3589', Aliases='Dmel_ Dmel\\CG3589', description='CG3589 gene product from transcript CG3589-RA', other_designations='CG3589-PA', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24783555, end_position_on_the_genomic_accession=24785542, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37967, CurrentID=0, Status='live', Symbol='key', Aliases='Dmel_CG16910, CG16910, DmIKK-gamma, DmIKKgamma, Dmel\\CG16910, Dmikkgamma, IKK, IKK-gamma, IKKG, IKK[[gamma]], IKKg, IKKgamma, KEY, Kenny, Key, dIKK, dIKK-gamma, dmIKKgamma', description='kenny', other_designations='CG16910-PA|CG16910-PB|kenney|key-PA|key-PB', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24785514, end_position_on_the_genomic_accession=24787411, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37968, CurrentID=0, Status='live', Symbol='Reg-5', Aliases='Dmel_CG2928, CG2928, DREG-5, Dmel\\CG2928, Dreg, Dreg-5, Dreg5, dreg-5, reg-5', description='Rhythmically expressed gene 5', other_designations='CG2928-PA|CG2928-PB|Reg-5-PA|Reg-5-PB|drosophila rhythmically expressed gene 5|rhythmically expressed gene 5', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24788008, end_position_on_the_genomic_accession=24793410, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318783, CurrentID=0, Status='live', Symbol='CG31526', Aliases='Dmel_ Dmel\\CG31526', description='CG31526 gene product from transcript CG31526-RA', other_designations='CG31526-PA|CG31526-PC|uncharacterized protein', map_location='82C1-82C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4533248, end_position_on_the_genomic_accession=4533940, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37970, CurrentID=0, Status='live', Symbol='Orc4', Aliases='Dmel_CG2917, CG2917, DmORC4, Dmel\\CG2917, ORC, ORC4, dmOrc4, orc4, rDmORC', description='Origin recognition complex subunit 4', other_designations='CG2917-PA|Orc4-PA|origin recognition complex subunit 4', map_location='60E1-60E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24795740, end_position_on_the_genomic_accession=24797414, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37971, CurrentID=0, Status='live', Symbol='CG12849', Aliases='Dmel_ Dmel\\CG12849', description='CG12849 gene product from transcript CG12849-RA', other_designations='CG12849-PA|uncharacterized protein', map_location='60E2-60E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24797925, end_position_on_the_genomic_accession=24798613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37972, CurrentID=0, Status='live', Symbol='CG3611', Aliases='Dmel_ BcDNA:AT12780, Dmel\\CG3611', description='CG3611 gene product from transcript CG3611-RC', other_designations='CG3611-PA|CG3611-PB|CG3611-PC', map_location='60E2-60E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24798798, end_position_on_the_genomic_accession=24800776, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37973, CurrentID=0, Status='live', Symbol='Dll', Aliases='Dmel_CG3629, 2.7, Art, Ba, BcDNA:LP01770, CG3629, DLL, Dmel\\CG3629, E(Arp), En(Arp), dl, dll, l(2)01092, l(2)387', description='Distal-less', other_designations='CG3629-PA|CG3629-PB|CG3629-PC|Distall-less|Dll-PA|Dll-PB|Dll-PC|brista|distal-less|distalless|enhancer of Arp', map_location='2-107.8 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24814822, end_position_on_the_genomic_accession=24835667, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37974, CurrentID=0, Status='live', Symbol='CG3650', Aliases='Dmel_ Dmel\\CG3650, SP159', description='CG3650 gene product from transcript CG3650-RA', other_designations='CG3650-PA', map_location='60E3-60E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24848994, end_position_on_the_genomic_accession=24850200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37977, CurrentID=0, Status='live', Symbol='CG16896', Aliases='Dmel_ Dmel\\CG16896', description='CG16896 gene product from transcript CG16896-RA', other_designations='CG16896-PA|uncharacterized protein', map_location='60E4-60E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24871407, end_position_on_the_genomic_accession=24874912, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37978, CurrentID=0, Status='live', Symbol='Atf-2', Aliases='Dmel_CG44246, ATF-2, ATF2, Atf2, CG12850, CG30420, CG44246, DM5, Dmel\\CG44246, Dmel_CG30420, atf-2, dATF-2, dATF2', description='Activating transcription factor-2', other_designations='Atf-2-PA|CG44246-PA|activating transcription factor-2', map_location='60E4-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24870108, end_position_on_the_genomic_accession=24871509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37979, CurrentID=0, Status='live', Symbol='NKAIN', Aliases='Dmel_CG34413, CG32836, CG34413, CG9047, Dmel\\CG34413, Dmel_CG32836, Dmel_CG9047, anon-WO0118547.94, dNKAIN', description='Na,K-ATPase Interacting', other_designations='CG34413-PF|CG34413-PG|CG34413-PH|CG34413-PI|CG34413-PJ|CG34413-PK|CG34413-PL|NKAIN-PF|NKAIN-PG|NKAIN-PH|NKAIN-PI|NKAIN-PJ|NKAIN-PK|NKAIN-PL|Na,K-ATPase interacting', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24881570, end_position_on_the_genomic_accession=24888615, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37980, CurrentID=0, Status='live', Symbol='Ance-5', Aliases='Dmel_CG10142, CG10142, Dmel\\CG10142, ance-5', description='CG10142 gene product from transcript CG10142-RA', other_designations='Ance-5-PA|CG10142-PA', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24888628, end_position_on_the_genomic_accession=24891023, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37981, CurrentID=0, Status='live', Symbol='NaCP60E', Aliases='Dmel_CG34405, CG18506, CG34405, CG9071, CT24831, DIC60, DSC, DSC1, DSCI, Dmel\\CG34405, Dmel_CG18506, Dmel_CG9071, Na-CP, dsc, dsc1, smi60E', description='Na channel protein 60E', other_designations='CG34405-PG|CG34405-PH|CG34405-PI|CG34405-PJ|CG34405-PK|CG34405-PL|CG34405-PM|CG34405-PN|CG34405-PO|CG34405-PP|Na-channel-protein|NaCP60E-PG|NaCP60E-PH|NaCP60E-PI|NaCP60E-PJ|NaCP60E-PK|NaCP60E-PL|NaCP60E-PM|NaCP60E-PN|NaCP60E-PO|NaCP60E-PP|put. sodium channel gene|smell impaired 60E|sodium channel 1', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24891614, end_position_on_the_genomic_accession=24921677, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  37982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37982, CurrentID=0, Status='live', Symbol='CG9083', Aliases='Dmel_ Dmel\\CG9083', description='CG9083 gene product from transcript CG9083-RB', other_designations='CG9083-PB', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24898273, end_position_on_the_genomic_accession=24899505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37983, CurrentID=0, Status='live', Symbol='Tpc2', Aliases='Dmel_CG2857, CG2857, DmTpc2, Dmel\\CG2857', description='Thiamine pyrophosphate carrier protein 2', other_designations='CG2857-PB|Tpc2-PB|thiamine pyrophosphate carrier protein|thiamine pyrophosphate carrier protein 2', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24897902, end_position_on_the_genomic_accession=24900734, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37985, CurrentID=0, Status='live', Symbol='pain', Aliases='Dmel_CG15860, CG15860, CT31987, Dmel\\CG15860, EP2251, Pain', description='painless', other_designations='CG15860-PA|CG15860-PB|CG15860-PC|pain-PA|pain-PB|pain-PC', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24921505, end_position_on_the_genomic_accession=24928275, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37986, CurrentID=0, Status='live', Symbol='CG30427', Aliases='Dmel_ BEST:LD36843, CG18680, CG2858, Dmel\\CG30427', description='CG30427 gene product from transcript CG30427-RC', other_designations='CG30427-PA|CG30427-PB|CG30427-PC|CG30427-PD|CG30427-PE|CG30427-PF|CG30427-PG|uncharacterized protein', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24928747, end_position_on_the_genomic_accession=24937477, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  37987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37987, CurrentID=0, Status='live', Symbol='CG3760', Aliases='Dmel_ Dmel\\CG3760, anon-EST:fe2C9, anon-fe2C9, anon2C9', description='CG3760 gene product from transcript CG3760-RC', other_designations='CG3760-PA|CG3760-PB|CG3760-PC|CG3760-PD|anon-fast-evolving-2C9', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24938181, end_position_on_the_genomic_accession=24939769, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  37988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37988, CurrentID=0, Status='live', Symbol='CG2811', Aliases='Dmel_ BcDNA:RE67445, Dmel\\CG2811', description='CG2811 gene product from transcript CG2811-RB', other_designations='CG2811-PA|CG2811-PB|CG2811-PC', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24939706, end_position_on_the_genomic_accession=24940483, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37989, CurrentID=0, Status='live', Symbol='Tina-1', Aliases='Dmel_CG2803, BEST:GM14784, CG2803, Dmel\\CG2803', description='Troponin C-akin-1', other_designations='CG2803-PA|CG2803-PB|Tina-1-PA|Tina-1-PB|troponin C-akin-1', map_location='60E5-60E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24940625, end_position_on_the_genomic_accession=24942413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37990, CurrentID=0, Status='live', Symbol='CG15861', Aliases='Dmel_ Dmel\\CG15861', description='CG15861 gene product from transcript CG15861-RB', other_designations='CG15861-PA|CG15861-PB|uncharacterized protein', map_location='60E8-60E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24943392, end_position_on_the_genomic_accession=24945201, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37991, CurrentID=0, Status='live', Symbol='CG3770', Aliases='Dmel_ CT12588, Dmel\\CG3770', description='CG3770 gene product from transcript CG3770-RA', other_designations='CG3770-PA|CG3770-PB', map_location='60E8-60E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24945139, end_position_on_the_genomic_accession=24946636, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37992, CurrentID=0, Status='live', Symbol='CG2790', Aliases='Dmel_ BcDNA:GH03108, Dmel\\CG2790', description='CG2790 gene product from transcript CG2790-RA', other_designations='CG2790-PA', map_location='60E8-60E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24946561, end_position_on_the_genomic_accession=24948378, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37993, CurrentID=0, Status='live', Symbol='CG12851', Aliases='Dmel_ Dmel\\CG12851', description='CG12851 gene product from transcript CG12851-RD', other_designations='CG12851-PB|CG12851-PC|CG12851-PD|uncharacterized protein', map_location='60E8-60E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24948645, end_position_on_the_genomic_accession=24955677, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  37994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37994, CurrentID=0, Status='live', Symbol='CG2765', Aliases='Dmel_ Dmel\\CG2765', description='CG2765 gene product from transcript CG2765-RF', other_designations='CG2765-PC|CG2765-PD|CG2765-PE|CG2765-PF', map_location='60E10-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24959841, end_position_on_the_genomic_accession=24965076, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  37995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37995, CurrentID=0, Status='live', Symbol='RpL19', Aliases='Dmel_CG2746, CG2746, Dmel\\CG2746, E(Dl)KP135, L19, M(2)60E, M(2)a, M(2)c, M60E, Rp L19, Rpl19, anon-EST:Posey33, l(2)k03704, rpL19, rpl19', description='Ribosomal protein L19', other_designations='CG2746-PA|CG2746-PB|Minute(2)60E|RpL19-PA|RpL19-PB|enhancer of delta KP135|lethal (2) k03704|ribosomal protein L19', map_location='2-108 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24967877, end_position_on_the_genomic_accession=24969547, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37996, CurrentID=0, Status='live', Symbol='CG3776', Aliases='Dmel_ DmP29, Dmel\\CG3776, EP(2)0835, Jhebp29, P29', description='CG3776 gene product from transcript CG3776-RB', other_designations='CG3776-PA|CG3776-PB|CG3776-PC|drosophila P29|juvenile hormone esterase binding protein 29', map_location='60E11-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24969678, end_position_on_the_genomic_accession=24970855, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  37997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37997, CurrentID=0, Status='live', Symbol='Phk-3', Aliases='Dmel_CG9358, BcDNA:RE09339, CG9358, DmelCSP3, Dmel\\CG9358, PHK-3', description='Pherokine 3', other_designations='CG9358-PA|CG9358-PB|CG9358-PC|Phk-3-PA|Phk-3-PB|Phk-3-PC|pherokine 3', map_location='60E11-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24970843, end_position_on_the_genomic_accession=24973207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37998, CurrentID=0, Status='live', Symbol='CG2736', Aliases='Dmel_ Dmel\\CG2736, cg2736', description='CG2736 gene product from transcript CG2736-RA', other_designations='CG2736-PA', map_location='60E11-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24973431, end_position_on_the_genomic_accession=24975341, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  37999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37999, CurrentID=0, Status='live', Symbol='emp', Aliases='Dmel_CG2727, CG2727, Dmel\\CG2727', description='epithelial membrane protein', other_designations='CG2727-PA|CG2727-PB|CG2727-PC|CG2727-PD|CG2727-PE|emp-PA|emp-PB|emp-PC|emp-PD|emp-PE', map_location='60E11-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24976456, end_position_on_the_genomic_accession=24986070, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38000, CurrentID=0, Status='live', Symbol='CG3829', Aliases='Dmel_ Dmel\\CG3829', description='CG3829 gene product from transcript CG3829-RA', other_designations='CG3829-PA|CG3829-PB', map_location='60E11-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24986044, end_position_on_the_genomic_accession=24990489, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38001, CurrentID=0, Status='live', Symbol='zip', Aliases='Dmel_CG15792, CG15792, DROMHC, Dm nmII, Dmel\\CG15792, DmnmII, Dronm-MII, E(br), MHC, Mhc-c, Mhc-c[1], Myo, Myo II, Myo-II, Myo-II HC, Myo2, MyoII, NMM, NMMII, ZIP, Zip, anon-WO0140519.37, l(2)02957, l(2)17F1, myo II, myoII, nmMHC, nmy-2/MyoII, zpr, zip', description='zipper', other_designations='CG15792-PA|CG15792-PB|CG15792-PC|CG15792-PD|CG15792-PE|CG15792-PF|CG15792-PG|CG15792-PH|CG15792-PI|Myo-II heavy chain|Myosin-2|Myosin-II|cytoplasmic myosin|cytoplasmic myosin II|cytoplasmic myosin heavy chain|cytoplasmic myosin-II|enhancer of broad|lethal (2) 17F1|myosin|myosin 2|myosin HC|myosin II|myosin II heavy chain|myosin heavy chain|myosin heavy chain-cytoplasmic|myosin-heavy-chain-cytoplasmic|myosins II|non muscle myosin|non-muscle myosin|non-muscle myosin II|non-muscle myosin II heavy chain|non-muscle myosin heavy-chain|nonmuscle myosin II|nonmuscle myosin II heavy chain|nonmuscle myosin II heavychain|nonmuscle myosin heavy chain|nonmuscle myosin-II|nonmuscle myosin-II heavy chain|zip-PA|zip-PB|zip-PC|zip-PD|zip-PE|zip-PF|zip-PG|zip-PH|zip-PI|zipper heavy chain', map_location='60E11-60E12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24990570, end_position_on_the_genomic_accession=25011965, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  38002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38002, CurrentID=0, Status='live', Symbol='uzip', Aliases='Dmel_CG3533, CG3533, Dmel\\CG3533, nzip, zip', description='unzipped', other_designations='CG3533-PA|CG3533-PB|not-zipper|uzip-PA|uzip-PB|zipper', map_location='2-107.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25012062, end_position_on_the_genomic_accession=25031737, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38003, CurrentID=0, Status='live', Symbol='Nplp1', Aliases='Dmel_CG3441, APK, CG3441, Dmel\\CG3441, IPNa, IPNamide, MTY-amide, MTYa, MTYamide, NPLP-1, NPLP1, VQQ, nplp1', description='Neuropeptide-like precursor 1', other_designations='APK peptide|CG3441-PA|CG3441-PB|CG3441-PC|IPNamide|IPNamide peptide|MTYamide peptide|Nplp1-PA|Nplp1-PB|Nplp1-PC|neuropeptide like precursor 1|neuropeptide like precursor protein 1|neuropeptide-like precursor 1', map_location='60F1-60F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25032377, end_position_on_the_genomic_accession=25040377, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38004, CurrentID=0, Status='live', Symbol='gsb-n', Aliases='Dmel_CG2692, BSH-4, BSH4, CG2692, Dmel\\CG2692, Gsb-p, GsbN, Gsbn, gsbeuro, gsb-p, gsbN, gsbn, gsbp, gsb-n', description='gooseberry-neuro', other_designations='CG2692-PA|gooseberry|gooseberry neuro|gooseberry proximal|gsb proximal|gsb-n-PA|gsb-neuro|gsb-proximal', map_location='2-107.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25040383, end_position_on_the_genomic_accession=25052331, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38005, CurrentID=0, Status='live', Symbol='gsb', Aliases='Dmel_CG3388, BSH-9, BSH9, BSH9c2, CG3388, Dmel\\CG3388, GSB-D, Gsb, Gsb-d-2, gsb-d, gsb-distal, gsbBSH9, gsbd, l(2)01155, l(2)gsb, gsb', description='gooseberry', other_designations='CG3388-PA|Gooseberry-distal|gooseberry distal|gsb-PA', map_location='2-107.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25061964, end_position_on_the_genomic_accession=25064643, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38006, CurrentID=0, Status='live', Symbol='gol', Aliases='Dmel_CG2679, CG2679, Dmel\\CG2679, Gol, anon-G1, g1, meso', description='goliath', other_designations='CG2679-PB|CG2679-PC|CG2679-PD|CG2679-PE|gol-PB|gol-PC|gol-PD|gol-PE', map_location='60F3-60F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25072627, end_position_on_the_genomic_accession=25084347, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38007, CurrentID=0, Status='live', Symbol='lov', Aliases='Dmel_CG16778, 091y, 91Y, BTB-III, BtbIII, CG16778, CG2672, Dmel\\CG16778, Lov, TKR, Tkr, anon-G2, dTKR, g2, tkr', description='jim lovell', other_designations='91Y|BTB-protein-III|CG16778-PB|CG16778-PC|CG16778-PD|CG16778-PE|CG16778-PF|Jim lovell|lov-PB|lov-PC|lov-PD|lov-PE|lov-PF|tyrosine kinase related|tyrosine kinase-related|tyrosine kinase-related protein', map_location='2-107.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25085780, end_position_on_the_genomic_accession=25139627, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38009, CurrentID=0, Status='live', Symbol='Ebp', Aliases='Dmel_CG2668, BcDNA:GH06048, CG2668, Dmel\\CG2668, P38, PEB, PEB-me, Peb, anon-WO0140519.140, peb', description='Ejaculatory bulb protein', other_designations='CG2668-PA|CG2668-PB|Ebp-PA|Ebp-PB|ejaculatory bulb protein|protein ejaculatory bulb', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25175126, end_position_on_the_genomic_accession=25176715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38011, CurrentID=0, Status='live', Symbol='CG9380', Aliases='Dmel_ Dmel\\CG9380', description='CG9380 gene product from transcript CG9380-RG', other_designations='CG9380-PD|CG9380-PE|CG9380-PF|CG9380-PG', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25183628, end_position_on_the_genomic_accession=25255318, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  38012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38012, CurrentID=0, Status='live', Symbol='Kr', Aliases='Dmel_CG3340, CG3340, Dm-Kr, Dmel\\CG3340, If, KR, kr', description='Kruppel', other_designations='CG3340-PA|CG3340-PB|Kr-PA|Kr-PB|irregular facets|krpple|krueppel|kruppel|kuppel|lruppel', map_location='2-107.6 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25226611, end_position_on_the_genomic_accession=25231394, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38013, CurrentID=0, Status='live', Symbol='mthl8', Aliases='Dmel_CG32475, CG11869, CG13406, CG32475, Dmel\\CG32475, Mth-8, anon-WO0170980.145, anon-WO0170980.146, mthl-8', description='methuselah-like 8', other_designations='CG32475-PA|CG32475-PB|Methuselah-like 8|mth-like 8|mthl8-PA|mthl8-PB', map_location='61A5-61A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22428, end_position_on_the_genomic_accession=26308, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38015, CurrentID=0, Status='live', Symbol='Lsp1gamma', Aliases='Dmel_CG6821, CG6821, DmeLSP1g, Dmel\\CG6821, LSP 1, LSP-1 gamma, LSP-1gamma, LSP1, Lsp-1gamma, Lsp1, Lsp1-gamma, Lsp1[gamma], Lsp1gama', description='Larval serum protein 1 gamma', other_designations='CG6821-PA|Lsp1gamma-PA|larval serum protein 1 gamma|larval serum protein 1[gamma]|larval serum protein 1gamma', map_location='3-1.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=66430, end_position_on_the_genomic_accession=68995, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38016, CurrentID=0, Status='live', Symbol='CG13405', Aliases='Dmel_ BcDNA:AT11049, Dmel\\CG13405', description='CG13405 gene product from transcript CG13405-RB', other_designations='CG13405-PA|CG13405-PB|uncharacterized protein', map_location='61A6-61A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=73945, end_position_on_the_genomic_accession=75332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38017, CurrentID=0, Status='live', Symbol='Pdk1', Aliases='Dmel_CG1210, 1210, CG1201, CG1210, DSTPK61, Dmel\\CG1210, Dstpk61, PDK, PDK-1, PDK1, PDK1/Pk61C, PK61C, Pk61C, Pk61C/PDK1, Pk61C|PDK1, Pk61c/PDK1, dPDK, dPDK-1, dPDK1, dSTPK61, pdk1, pk61c', description='Phosphoinositide-dependent kinase 1', other_designations='CG1210-PA|CG1210-PB|CG1210-PC|CG1210-PD|CG1210-PE|CG1210-PF|CG1210-PG|CG1210-PH|CG1210-PJ|Pdk1-PA|Pdk1-PB|Pdk1-PC|Pdk1-PD|Pdk1-PE|Pdk1-PF|Pdk1-PG|Pdk1-PH|Pdk1-PJ|phosphoinositide dependent kinase 1|phosphoinositide-dependent kinase 1|pk61c|protein kinase 61C|serine/threonine protein kinase', map_location='61B1-61B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=129450, end_position_on_the_genomic_accession=145895, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38019, CurrentID=0, Status='live', Symbol='CG6845', Aliases='Dmel_ BEST:LD22345, Dmel\\CG6845', description='CG6845 gene product from transcript CG6845-RE', other_designations='CG6845-PC|CG6845-PD|CG6845-PE', map_location='61B1-61B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=146433, end_position_on_the_genomic_accession=149056, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38020, CurrentID=0, Status='live', Symbol='Dic61B', Aliases='Dmel_CG7051, CG7051, Dmel\\CG7051, ms, ms(3)21', description='Dynein intermediate chain at 61B', other_designations='CG7051-PA|CG7051-PB|Dic61B-PA|Dic61B-PB|dynein intermediate chain at 61B|male sterile (3) 21', map_location='61B1-61B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=149679, end_position_on_the_genomic_accession=152589, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38021, CurrentID=0, Status='live', Symbol='p130CAS', Aliases='Dmel_CG1212, CG1212, CT1293, DCas, Dcas, Dmel\\CG1212, p130Cas', description='CG1212 gene product from transcript CG1212-RD', other_designations='CG1212-PB|CG1212-PD|CG1212-PE|CG1212-PF|CG1212-PG|p130CAS-PB|p130CAS-PD|p130CAS-PE|p130CAS-PF|p130CAS-PG', map_location='61B1-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=152884, end_position_on_the_genomic_accession=174517, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38022, CurrentID=0, Status='live', Symbol='CG7049', Aliases='Dmel_ Dmel\\CG7049', description='CG7049 gene product from transcript CG7049-RA', other_designations='CG7049-PA', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=175112, end_position_on_the_genomic_accession=176228, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38023, CurrentID=0, Status='live', Symbol='Vdup1', Aliases='Dmel_CG7047, CG7047, Dmel\\CG7047, VDUP1, dvdup1, vdup1', description='Vitamin D[[3]] up-regulated protein 1', other_designations='CG7047-PA|CG7047-PB|CG7047-PC|CG7047-PD|CG7047-PE|Vdup1-PA|Vdup1-PB|Vdup1-PC|Vdup1-PD|Vdup1-PE|vitamin D3 up-regulated protein 1|vitamin D[[3]] up-regulated protein 1|vitamin-D[[3]] upregulated protein 1', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=176306, end_position_on_the_genomic_accession=178963, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38024, CurrentID=0, Status='live', Symbol='CG13875', Aliases='Dmel_ Dmel\\CG13875, RE30146P', description='CG13875 gene product from transcript CG13875-RB', other_designations='CG13875-PA|CG13875-PB|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=179923, end_position_on_the_genomic_accession=182535, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38026, CurrentID=0, Status='live', Symbol='Mtch', Aliases='Dmel_CG6851, CG6851, Dmel\\CG6851', description='Mitochondrial carrier homolog 1', other_designations='CG6851-PA|CG6851-PB|CG6851-PC|CG6851-PE|CG6851-PF|CG6851-PG|Mtch-PA|Mtch-PB|Mtch-PC|Mtch-PE|Mtch-PF|Mtch-PG|mitochondrial carrier homolog 1', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=184626, end_position_on_the_genomic_accession=186310, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38027, CurrentID=0, Status='live', Symbol='rno', Aliases='Dmel_CG7036, CG7036, Dmel\\CG7036', description='rhinoceros', other_designations='CG7036-PB|CG7036-PC|CG7036-PD|rno-PB|rno-PC|rno-PD', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=187207, end_position_on_the_genomic_accession=199762, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38028, CurrentID=0, Status='live', Symbol='mri', Aliases='Dmel_CG1216, CG1216, Dmel\\CG1216, Mri', description='mrityu', other_designations='CG1216-PA|CG1216-PB|CG1216-PE|mri-PA|mri-PB|mri-PE', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=200174, end_position_on_the_genomic_accession=202152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38029, CurrentID=0, Status='live', Symbol='NitFhit', Aliases='Dmel_CG7067, CG7067, Dmel\\CG7067', description='Nitrilase and fragile histidine triad fusion protein', other_designations='CG7067-PA|NitFhit-PA|NitFlit|nitrilase and fragile histidine triad fusion protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=203730, end_position_on_the_genomic_accession=205309, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38031, CurrentID=0, Status='live', Symbol='CG13876', Aliases='Dmel_ Dmel\\CG13876', description='CG13876 gene product from transcript CG13876-RA', other_designations='CG13876-PA|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=206134, end_position_on_the_genomic_accession=207527, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38032, CurrentID=0, Status='live', Symbol='CG7028', Aliases='Dmel_ BcDNA:GH04978, Dmel\\CG7028, FBgn0027587, PRP4, prp4', description='CG7028 gene product from transcript CG7028-RE', other_designations='CG7028-PA|CG7028-PB|CG7028-PC|CG7028-PD|CG7028-PE', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=207184, end_position_on_the_genomic_accession=211457, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38033, CurrentID=0, Status='live', Symbol='thoc7', Aliases='Dmel_CG17143, CG17143, Dmel\\CG17143, THOC7', description='CG17143 gene product from transcript CG17143-RB', other_designations='CG17143-PA|CG17143-PB|thoc7-PA|thoc7-PB', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=211636, end_position_on_the_genomic_accession=212687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38034, CurrentID=0, Status='live', Symbol='CG16940', Aliases='Dmel_ Dmel\\CG16940', description='CG16940 gene product from transcript CG16940-RA', other_designations='CG16940-PA|CG16940-PC|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=212709, end_position_on_the_genomic_accession=216479, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38036, CurrentID=0, Status='live', Symbol='CG13877', Aliases='Dmel_ Dmel\\CG13877', description='CG13877 gene product from transcript CG13877-RB', other_designations='CG13877-PB|CG13877-PC|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=218068, end_position_on_the_genomic_accession=218663, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38037, CurrentID=0, Status='live', Symbol='pyx', Aliases='Dmel_CG17142, CG17142, CT33412, Dmel\\CG17142, Pyx, dTRPA2, trpA2', description='pyrexia', other_designations='CG17142-PA|CG17142-PB|pyx-PA|pyx-PB', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=217932, end_position_on_the_genomic_accession=223069, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38044, CurrentID=0, Status='live', Symbol='wac', Aliases='Dmel_CG13879, CG13879, Dmel\\CG13879, Wac, was', description='wee Augmin', other_designations='CG13879-PB|CG13879-PC|CG13879-PD|augmin|wac-PB|wac-PC|wac-PD|wee augmin', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=258176, end_position_on_the_genomic_accession=259097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38045, CurrentID=0, Status='live', Symbol='Tudor-SN', Aliases='Dmel_CG7008, CG7008, Dmel\\CG7008, SND1, TSN, TSN-1, TUDOR-SN, dP100, p100, tudor-SN', description='Tudor staphylococcal nuclease', other_designations='CG7008-PA|CG7008-PB|Tudor-SN-PA|Tudor-SN-PB|tudor staphylococcal nuclease', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=259043, end_position_on_the_genomic_accession=264380, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38046, CurrentID=0, Status='live', Symbol='mRpL17', Aliases='Dmel_CG13880, BcDNA:RE43278, CG13880, Dmel\\CG13880, L17, mRPL17, mrpL17', description='mitochondrial ribosomal protein L17', other_designations='CG13880-PA|CG13880-PB|mRpL17-PA|mRpL17-PB|mitochondrial ribosomal protein large subunit 17', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=264779, end_position_on_the_genomic_accession=265932, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38047, CurrentID=0, Status='live', Symbol='miple1', Aliases='Dmel_CG1221, CG1221, Dmel\\CG1221, Miple, anon-EST:Posey203, miple', description='midkine and pleiotrophin 1', other_designations='CG1221-PA|CG1221-PB|CG1221-PC|miple|miple1-PA|miple1-PB|miple1-PC', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=267299, end_position_on_the_genomic_accession=270336, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38049, CurrentID=0, Status='live', Symbol='ttm2', Aliases='Dmel_CG12313, 20130405, CG12313, Dmel\\CG12313', description='tiny tim 2', other_designations='CG12313-PA|ttm2-PA', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=275155, end_position_on_the_genomic_accession=276667, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38050, CurrentID=0, Status='live', Symbol='RhoGEF3', Aliases='Dmel_CG43976, CG1225, CG13881, CG13882, CG13883, CG42377, CG42378, CG43976, Dmel\\CG43976, Dmel_CG1225, Dmel_CG13881, Dmel_CG13882, Dmel_CG13883, Dmel_CG42377, Dmel_CG42378, DrhoGEF3, RhoGEF, rhoGEF3', description='Rho guanine nucleotide exchange factor 3', other_designations='CG43976-PA|CG43976-PB|CG43976-PC|CG43976-PE|CG43976-PF|CG43976-PG|CG43976-PH|CG43976-PJ|CG43976-PK|CG43976-PL|CG43976-PM|RhoGEF3-PA|RhoGEF3-PB|RhoGEF3-PC|RhoGEF3-PE|RhoGEF3-PF|RhoGEF3-PG|RhoGEF3-PH|RhoGEF3-PJ|RhoGEF3-PK|RhoGEF3-PL|RhoGEF3-PM', map_location='61B3-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=277437, end_position_on_the_genomic_accession=305292, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  38055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38055, CurrentID=0, Status='live', Symbol='Atac3', Aliases='Dmel_CG32343, CG32343, CG6994, Dmel\\CG32343, cg6994, dATAC3', description='Ada2a-containing complex component 3', other_designations='ATAC complex component 3|Atac3-PB|Atac3-PC|Atac3-PD|Atac3-PE|CG32343-PB|CG32343-PC|CG32343-PD|CG32343-PE', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=324985, end_position_on_the_genomic_accession=327992, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38056, CurrentID=0, Status='live', Symbol='mthl9', Aliases='Dmel_CG17084, CG17084, Dmel\\CG17084, Mth-9, anon-WO0170980.163, anon-WO0170980.164', description='methuselah-like 9', other_designations='CG17084-PA|mth-like 9|mthl9-PA', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=329822, end_position_on_the_genomic_accession=332309, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38057, CurrentID=0, Status='live', Symbol='mthl10', Aliases='Dmel_CG17061, CG17061, Dmel\\CG17061, anon-WO0170980.166, anon-WO0170980.167, mth2, mthl-10', description='methuselah-like 10', other_designations='CG17061-PA|CG17061-PB|CG17061-PC|CG17061-PD|CG17061-PE|CG17061-PF|CG17061-PG|Mth-like 10|mthl10-PA|mthl10-PB|mthl10-PC|mthl10-PD|mthl10-PE|mthl10-PF|mthl10-PG', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=332801, end_position_on_the_genomic_accession=342915, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38058, CurrentID=0, Status='live', Symbol='mth', Aliases='Dmel_CG6936, CG6936, Dmel\\CG6936, MTH, Mth', description='methuselah', other_designations='CG6936-PA|CG6936-PB|CG6936-PE|methusalah|mth-PA|mth-PB|mth-PE', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=343378, end_position_on_the_genomic_accession=346957, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38059, CurrentID=0, Status='live', Symbol='Ptpmeg', Aliases='Dmel_CG1228, CG1228, Dmel\\CG1228, MEG1, Meg, PTP-meg, anon-WO0118547.211, dMEG1, dPtpmeg, ptmpeg, ptpmeg, scc', description='Protein tyrosine phosphatase Meg', other_designations='CG1228-PG|CG1228-PH|CG1228-PI|CG1228-PJ|CG1228-PK|Meg|Ptpmeg-PG|Ptpmeg-PH|Ptpmeg-PI|Ptpmeg-PJ|Ptpmeg-PK|protein tyrosine phosphatase Meg|split central complex', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=328097, end_position_on_the_genomic_accession=356050, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38060, CurrentID=0, Status='live', Symbol='CG1231', Aliases='Dmel_ Dmel\\CG1231', description='CG1231 gene product from transcript CG1231-RA', other_designations='CG1231-PA', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=356422, end_position_on_the_genomic_accession=357664, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38062, CurrentID=0, Status='live', Symbol='Cdc5', Aliases='Dmel_CG6905, CG6905Dm, Dmel\\CG6905, cdc5, cdc5-like, cef1, Cdc5', description='Cell division cycle 5 ortholog', other_designations='CG6905-PA|CG6905-PB|Cdc5-PA|Cdc5-PB|cdc5|cdc5-like|cell division cycle 5 ortholog', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=357851, end_position_on_the_genomic_accession=361383, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38063, CurrentID=0, Status='live', Symbol='CG1233', Aliases='Dmel_ Dmel\\CG1233', description='CG1233 gene product from transcript CG1233-RA', other_designations='CG1233-PA|CG1233-PB', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=361472, end_position_on_the_genomic_accession=365681, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38064, CurrentID=0, Status='live', Symbol='CG13884', Aliases='Dmel_ BcDNA:AT18596, Dmel\\CG13884', description='CG13884 gene product from transcript CG13884-RA', other_designations='CG13884-PA|uncharacterized protein', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=365664, end_position_on_the_genomic_accession=366437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38065, CurrentID=0, Status='live', Symbol='trh', Aliases='Dmel_CG42865, 3.1, BP1081, CG13885, CG42865, CG6883, DMU42699, Dmel\\CG42865, Dmel_CG13885, Dmel_CG6883, HDC07747, TRH, Trh, bHLHe13, l(3)10512', description='trachealess', other_designations='CG42865-PA|CG42865-PB|CG42865-PC|CG42865-PD|CG42865-PE|CG42865-PF|CG42865-PG|trh-PA|trh-PB|trh-PC|trh-PD|trh-PE|trh-PF|trh-PG', map_location='3-1 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=366524, end_position_on_the_genomic_accession=392633, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  38066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38066, CurrentID=0, Status='live', Symbol='CG13891', Aliases='Dmel_ Dmel\\CG13891', description='CG13891 gene product from transcript CG13891-RA', other_designations='CG13891-PA|uncharacterized protein', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=433146, end_position_on_the_genomic_accession=433906, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38067, CurrentID=0, Status='live', Symbol='klar', Aliases='Dmel_CG17046, CG17046, Dmel\\CG17046, Klar, kls, marb', description='klarsicht', other_designations='CG17046-PA|CG17046-PB|CG17046-PC|CG17046-PD|CG17046-PE|CG17046-PF|CG17046-PG|CG17046-PH|CG17046-PI|klar-PA|klar-PB|klar-PC|klar-PD|klar-PE|klar-PF|klar-PG|klar-PH|klar-PI|klarischt|marbles', map_location='3-0.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=434083, end_position_on_the_genomic_accession=540614, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  38068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38068, CurrentID=0, Status='live', Symbol='CG17180', Aliases='Dmel_ Dmel\\CG17180', description='CG17180 gene product from transcript CG17180-RA', other_designations='CG17180-PA|CG17180-PB|uncharacterized protein', map_location='61C3-61C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=540851, end_position_on_the_genomic_accession=542147, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38069, CurrentID=0, Status='live', Symbol='Cypl', Aliases='Dmel_CG13892, CG13892, Dmel\\CG13892, cycl, cypl', description='Cyclophilin-like', other_designations='CG13892-PA|Cypl-PA', map_location='61C3-61C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=542122, end_position_on_the_genomic_accession=542931, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38070, CurrentID=0, Status='live', Symbol='Hipk', Aliases='Dmel_CG17090, CG17090, Dmel\\CG17090, HIPK, HIPK2, dhipk2, hipk', description='Homeodomain interacting protein kinase', other_designations='CG17090-PA|CG17090-PB|CG17090-PC|Hipk-PA|Hipk-PB|Hipk-PC|Homeodomain-interacting protein kinase|Homeodomain-interacting protein kinase 2|homeodomain interactin protein kinase|homeodomain interacting protein kinase', map_location='61C3-61C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=543060, end_position_on_the_genomic_accession=581099, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38071, CurrentID=0, Status='live', Symbol='Ppm1', Aliases='Dmel_CG12169, CG12169, Dmel\\CG12169, PP2C, dPPM1', description='CG12169 gene product from transcript CG12169-RA', other_designations='CG12169-PA|Ppm1-PA', map_location='61C5-61C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=581337, end_position_on_the_genomic_accession=582810, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38072, CurrentID=0, Status='live', Symbol='Kah', Aliases='Dmel_CG17181, CG17181, Dmel\\CG17181, kah', description='Kahuli', other_designations='CG17181-PA|Kah-PA|kahuli|scratch 3|scratch-like 1', map_location='61C5-61C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=582824, end_position_on_the_genomic_accession=589274, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38073, CurrentID=0, Status='live', Symbol='MED14', Aliases='Dmel_CG12031, CG12031, Dmel\\CG12031, Med14, Trap170, dTRAP170, p160', description='Mediator complex subunit 14', other_designations='CG12031-PA|CG12031-PC|MED14-PA|MED14-PC|mediator complex subunit 14', map_location='61C6-61C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=593246, end_position_on_the_genomic_accession=599239, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38074, CurrentID=0, Status='live', Symbol='CG13893', Aliases='Dmel_ Dmel\\CG13893', description='CG13893 gene product from transcript CG13893-RA', other_designations='CG13893-PA|uncharacterized protein', map_location='61C6-61C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=599094, end_position_on_the_genomic_accession=604008, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38075, CurrentID=0, Status='live', Symbol='Reg-2', Aliases='Dmel_CG3200, 21483246, CG3200, Dmel\\CG3200, Dreg-2', description='Rhythmically expressed gene 2', other_designations='CG3200-PA|Reg-2-PA|drosophila rhythmically expressed gene 2|rhythmically expressed gene 2', map_location='61C7-61C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=605165, end_position_on_the_genomic_accession=606569, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38076, CurrentID=0, Status='live', Symbol='Gale', Aliases='Dmel_CG12030, CG12030, Dmel\\CG12030, GALE, dGALE', description="UDP-galactose 4'-epimerase", other_designations="CG12030-PA|CG12030-PB|Gale-PA|Gale-PB|UDP-gal 4' epimerase", map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=648555, end_position_on_the_genomic_accession=651311, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38077, CurrentID=0, Status='live', Symbol='CG3402', Aliases='Dmel_ BcDNA:LD06317, Dmel\\CG3402', description='CG3402 gene product from transcript CG3402-RA', other_designations='CG3402-PA', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=651542, end_position_on_the_genomic_accession=652881, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38078, CurrentID=0, Status='live', Symbol='MED30', Aliases='Dmel_CG17183, CG17183, Dmel\\CG17183, Med11/TRAP25, Med20, Med30, Trap25, dMED20, dTRAP25, med30', description='Mediator complex subunit 30', other_designations='CG17183-PA|MED30-PA|mediator complex subunit 30', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=652987, end_position_on_the_genomic_accession=654089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38079, CurrentID=0, Status='live', Symbol='Rev1', Aliases='Dmel_CG12189, AAF47401, CG12189, DmREV1, Dm Dmel\\CG12189, REV1, dREV1, dRev1, drev1, rev1', description='CG12189 gene product from transcript CG12189-RA', other_designations='CG12189-PA|Rev1-PA|UV-revertible gene 1', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=654363, end_position_on_the_genomic_accession=657971, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38080, CurrentID=0, Status='live', Symbol='CG17129', Aliases='Dmel_ Dmel\\CG17129', description='CG17129 gene product from transcript CG17129-RC', other_designations='CG17129-PA|CG17129-PB|CG17129-PC|uncharacterized protein', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=677879, end_position_on_the_genomic_accession=680326, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38081, CurrentID=0, Status='live', Symbol='CG3386', Aliases='Dmel_ Dmel\\CG3386', description='CG3386 gene product from transcript CG3386-RD', other_designations='CG3386-PA|CG3386-PB|CG3386-PC|CG3386-PD', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=680436, end_position_on_the_genomic_accession=682470, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38082, CurrentID=0, Status='live', Symbol='ebd1', Aliases='Dmel_CG3371, CG3371, Dmel\\CG3371, EP3592, Ebd1', description='earthbound 1', other_designations='CG3371-PA|earthbound|ebd1-PA', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=682885, end_position_on_the_genomic_accession=685549, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38083, CurrentID=0, Status='live', Symbol='CG3344', Aliases='Dmel_ Dmel\\CG3344', description='CG3344 gene product from transcript CG3344-RA', other_designations='CG3344-PA', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=685942, end_position_on_the_genomic_accession=687958, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38084, CurrentID=0, Status='live', Symbol='RabX6', Aliases='Dmel_CG12015, CG12015, Dm Dmel\\CG12015, rabX6, rabx6', description='CG12015 gene product from transcript CG12015-RA', other_designations='CG12015-PA|CG12015-PB|RabX6-PA|RabX6-PB', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=690517, end_position_on_the_genomic_accession=691951, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38085, CurrentID=0, Status='live', Symbol='Vti1a', Aliases='Dmel_CG3279, CG3279, Dmel\\CG3279, VTI1, Vti1', description='Vesicle transport through interaction with t-SNAREs 1a', other_designations='CG3279-PA|VTI1 ortholog|Vti1|Vti1a-PA|vesicle transport through interaction with t-SNAREs 1a', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=691486, end_position_on_the_genomic_accession=692919, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38086, CurrentID=0, Status='live', Symbol='CG13894', Aliases='Dmel_ Dmel\\CG13894, dmCG13894', description='CG13894 gene product from transcript CG13894-RA', other_designations='CG13894-PA|uncharacterized protein', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=693083, end_position_on_the_genomic_accession=699144, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38087, CurrentID=0, Status='live', Symbol='CG13895', Aliases='Dmel_ Dmel\\CG13895', description='CG13895 gene product from transcript CG13895-RA', other_designations='CG13895-PA|CG13895-PB|uncharacterized protein', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=705948, end_position_on_the_genomic_accession=708205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38088, CurrentID=0, Status='live', Symbol='CkIIalpha-i3', Aliases='Dmel_CG3217, CG3217, CKIIalpha-i3, Dmel\\CG3217', description='CKII-alpha subunit interactor-3', other_designations='CG3217-PA|CG3217-PB|CG3217-PC|CkIIalpha-i3-PA|CkIIalpha-i3-PB|CkIIalpha-i3-PC', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=708601, end_position_on_the_genomic_accession=710607, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38089, CurrentID=0, Status='live', Symbol='CG13896', Aliases='Dmel_ Dmel\\CG13896', description='CG13896 gene product from transcript CG13896-RB', other_designations='CG13896-PA|CG13896-PB|CG13896-PC|uncharacterized protein', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=710820, end_position_on_the_genomic_accession=711850, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38090, CurrentID=0, Status='live', Symbol='hng3', Aliases='Dmel_CG13897, CG13897, Dmel\\CG13897, anon-WO0118547.557', description='hinge3', other_designations='CG13897-PA|CG13897-PB|Hinge3|hng3-PA|hng3-PB', map_location='61C9-61C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=737680, end_position_on_the_genomic_accession=755489, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38091, CurrentID=0, Status='live', Symbol='emc', Aliases='Dmel_CG1007, 0094/26, 0203/10, 0587/01, 0977/09, Ach, CG1007, Dm0688, Dmel\\CG1007, Emc, bHLHb28, gov, l(3)04322, l(3)05592, l(3)61Da, l(3)j4E11, ms(3)61CD', description='extra macrochaetae', other_designations='CG1007-PA|achetous|emc-PA|extra macrochaete|extra macrocheatae|extramachrochaetae|extramacrochaetae|extramacrochaete|extramacrocheate|golovan|male sterile(3)61CD', map_location='3-0.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=749400, end_position_on_the_genomic_accession=753492, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38092, CurrentID=0, Status='live', Symbol='CG13898', Aliases='Dmel_ Dmel\\CG13898', description='CG13898 gene product from transcript CG13898-RA', other_designations='CG13898-PA|uncharacterized protein', map_location='61D1-61D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=803427, end_position_on_the_genomic_accession=804051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38093, CurrentID=0, Status='live', Symbol='CG13900', Aliases='Dmel_ Dmel\\CG13900, SF3B3', description='CG13900 gene product from transcript CG13900-RA', other_designations='CG13900-PA|CG13900-PC|putative damaged DNA-binding activity|splicing factor 3b subunit 3|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=826720, end_position_on_the_genomic_accession=833385, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  38094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38094, CurrentID=0, Status='live', Symbol='CG42553', Aliases='Dmel_ CG13886, CT33419, Dmel\\CG42553, Dmel_CG13886', description='CG42553 gene product from transcript CG42553-RA', other_designations='CG42553-PA|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=834093, end_position_on_the_genomic_accession=835515, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38095, CurrentID=0, Status='live', Symbol='CG13901', Aliases='Dmel_ Dmel\\CG13901', description='CG13901 gene product from transcript CG13901-RB', other_designations='CG13901-PA|CG13901-PB|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=836667, end_position_on_the_genomic_accession=837821, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38096, CurrentID=0, Status='live', Symbol='CG13887', Aliases='Dmel_ Dmel\\CG13887', description='CG13887 gene product from transcript CG13887-RE', other_designations='CG13887-PB|CG13887-PC|CG13887-PD|CG13887-PE|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=837861, end_position_on_the_genomic_accession=839931, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38097, CurrentID=0, Status='live', Symbol='CG13902', Aliases='Dmel_ Dmel\\CG13902', description='CG13902 gene product from transcript CG13902-RC', other_designations='CG13902-PB|CG13902-PC|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=839980, end_position_on_the_genomic_accession=841360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38098, CurrentID=0, Status='live', Symbol='Gr61a', Aliases='Dmel_CG13888, CG13888, Dmel61a, Dmel\\CG13888, GR61a, Gr61D1', description='Gustatory receptor 61a', other_designations='CG13888-PA|CG13888-PB|CG13888-PC|Gr61a-PA|Gr61a-PB|Gr61a-PC|gustatory receptor 61D1|gustatory receptor 61a', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=841451, end_position_on_the_genomic_accession=843701, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38099, CurrentID=0, Status='live', Symbol='cep290', Aliases='Dmel_CG13889, CG13889, Dmel\\CG13889, Q9W0M1_DROME', description='CG13889 gene product from transcript CG13889-RB', other_designations='CG13889-PA|CG13889-PB|cep290-PA|cep290-PB', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=843926, end_position_on_the_genomic_accession=851080, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38100, CurrentID=0, Status='live', Symbol='CG13890', Aliases='Dmel_ Dmel\\CG13890', description='CG13890 gene product from transcript CG13890-RA', other_designations='CG13890-PA|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=851346, end_position_on_the_genomic_accession=852673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38101, CurrentID=0, Status='live', Symbol='dpr20', Aliases='Dmel_CG12191, CG12191, CT9818, Dmel\\CG12191, Dpr-20, Dpr20', description='defective proboscis extension response 20', other_designations='CG12191-PA|dpr20-PA', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=852720, end_position_on_the_genomic_accession=857248, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38102, CurrentID=0, Status='live', Symbol='CG12502', Aliases='Dmel_ Dmel\\CG12502', description='CG12502 gene product from transcript CG12502-RB', other_designations='CG12502-PB|CG12502-PC|CG12502-PD|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=865396, end_position_on_the_genomic_accession=871701, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38103, CurrentID=0, Status='live', Symbol='Usp10', Aliases='Dmel_CG32479, CG12503, CG13903, CG32479, Dmel\\CG32479, Q9W0L7', description='Ubiquitin specific protease 10', other_designations='CG32479-PA|CG32479-PB|CG32479-PC|Usp10-PA|Usp10-PB|Usp10-PC|ubiquitin specific protease 10', map_location='61D2-61D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=871899, end_position_on_the_genomic_accession=895313, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  38104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38104, CurrentID=0, Status='live', Symbol='CG13907', Aliases='Dmel_ Dmel\\CG13907', description='CG13907 gene product from transcript CG13907-RA', other_designations='CG13907-PA|CG13907-PB|uncharacterized protein', map_location='61D2-61D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=877291, end_position_on_the_genomic_accession=882013, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38106, CurrentID=0, Status='live', Symbol='CG34056', Aliases='Dmel_ CG13904, CG13904-1, Dmel\\CG34056, dC1GalT6', description='CG34056 gene product from transcript CG34056-RA', other_designations='CG34056-PA|uncharacterized protein', map_location='61D4-61D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=895951, end_position_on_the_genomic_accession=897364, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740751, CurrentID=0, Status='live', Symbol='Cpr65Ax1', Aliases='Dmel_CG34270, CG18774, CG34270, Dmel\\CG34270', description='Cuticular protein 65Ax1', other_designations='CG34270-PA|Cpr65Ax1-PA|cuticular protein 65Ax1', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6147531, end_position_on_the_genomic_accession=6147987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38109, CurrentID=0, Status='live', Symbol='Glut1', Aliases='Dmel_CG43946, 0074/12, CG1086, CG13908, CG43946, DmGLUT1, DmGluT1, Dmel\\CG43946, Dmel_CG1086, Dmel_CG13908, Dmglut1, GLU, Glt1, Glu T1, Glut-1, glut1, l(3)61Ea, l(3)L1H, l(3)S007412', description='Glucose transporter 1', other_designations='CG43946-PAA|CG43946-PB|CG43946-PC|CG43946-PD|CG43946-PE|CG43946-PJ|CG43946-PL|CG43946-PM|CG43946-PO|CG43946-PP|CG43946-PQ|CG43946-PR|CG43946-PS|CG43946-PT|CG43946-PU|CG43946-PV|CG43946-PW|CG43946-PX|CG43946-PY|CG43946-PZ|Glut1-PAA|Glut1-PB|Glut1-PC|Glut1-PD|Glut1-PE|Glut1-PJ|Glut1-PL|Glut1-PM|Glut1-PO|Glut1-PP|Glut1-PQ|Glut1-PR|Glut1-PS|Glut1-PT|Glut1-PU|Glut1-PV|Glut1-PW|Glut1-PX|Glut1-PY|Glut1-PZ|glucose transporter|glucose transporter 1|lethal (3) S007412', map_location='61D4-61E', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=914074, end_position_on_the_genomic_accession=993788, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  38113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38113, CurrentID=0, Status='live', Symbol='CG9205', Aliases='Dmel_ Dmel\\CG9205', description='CG9205 gene product from transcript CG9205-RA', other_designations='CG9205-PA|CG9205-PB', map_location='61E2-61E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1034697, end_position_on_the_genomic_accession=1036298, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38116, CurrentID=0, Status='live', Symbol='bab1', Aliases='Dmel_CG9097, BAB, BAB-1, BAB1, Bab1, CG13910, CG9097, Dmel\\CG9097, anon-WO0118547.639, bab, bab-I, bric-a-brac', description='bric a brac 1', other_designations='Bric-a-brac|CG9097-PB|CG9097-PC|CG9097-PD|CG9097-PE|bab1-PB|bab1-PC|bab1-PD|bab1-PE|bric a brac|bric a brac I|bric a brac1|bric-a-brac 1', map_location='61E2-61F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1036369, end_position_on_the_genomic_accession=1101089, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38120, CurrentID=0, Status='live', Symbol='CG13912', Aliases='Dmel_ Dmel\\CG13912', description='CG13912 gene product from transcript CG13912-RA', other_designations='CG13912-PA|uncharacterized protein', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1189588, end_position_on_the_genomic_accession=1190472, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38121, CurrentID=0, Status='live', Symbol='Trh', Aliases='Dmel_CG9122, CG9122, CT9937, DTPHu, DTRH, Dmel\\CG9122, TRH, TRHn, dTrh', description='Tryptophan hydroxylase', other_designations='CG9122-PA|Trh-PA|tryptophan hydroxylase', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1190384, end_position_on_the_genomic_accession=1193173, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38122, CurrentID=0, Status='live', Symbol='LysX', Aliases='Dmel_CG9120, BcDNA:LP09906, CG9120, Dmel\\CG9120, lysX', description='Lysozyme X', other_designations='CG9120-PA|LysX-PA|lysozyme X', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1194739, end_position_on_the_genomic_accession=1195302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38124, CurrentID=0, Status='live', Symbol='CG9119', Aliases='Dmel_ CT9987, Dmel\\CG9119', description='CG9119 gene product from transcript CG9119-RA', other_designations='CG9119-PA', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1203316, end_position_on_the_genomic_accession=1204795, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38125, CurrentID=0, Status='live', Symbol='LysB', Aliases='Dmel_CG1179, CG1179, Dmel\\CG1179, Lys B, lysB', description='Lysozyme B', other_designations='CG1179-PA|CG1179-PB|LysB-PA|LysB-PB|lysozyme B|lyzozyme B', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1207371, end_position_on_the_genomic_accession=1207981, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38127, CurrentID=0, Status='live', Symbol='LysD', Aliases='Dmel_CG9118, CG9118, Dmel\\CG9118, lysD', description='Lysozyme D', other_designations='CG9118-PA|LysD-PA|lysozyme|lysozyme D', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1210717, end_position_on_the_genomic_accession=1211196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38128, CurrentID=0, Status='live', Symbol='LysE', Aliases='Dmel_CG1180, BcDNA:LP07339, CG1180, Dmel\\CG1180, Lys E, lysE', description='Lysozyme E', other_designations='CG1180-PA|LysE-PA|lysozyme E', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1212827, end_position_on_the_genomic_accession=1213406, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38129, CurrentID=0, Status='live', Symbol='LysP', Aliases='Dmel_CG9116, CG9116, Dmel\\CG9116, lysP', description='Lysozyme P', other_designations='CG9116-PA|LysP-PA|lysozyme|lysozyme P', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1218082, end_position_on_the_genomic_accession=1218696, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38130, CurrentID=0, Status='live', Symbol='LysS', Aliases='Dmel_CG1165, BcDNA:LP06719, CG1165, Dmel\\CG1165, lysS', description='Lysozyme S', other_designations='CG1165-PA|LysS-PA|lysozyme S', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1227747, end_position_on_the_genomic_accession=1228238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38131, CurrentID=0, Status='live', Symbol='mwh', Aliases='Dmel_CG43772, BP1018, CG13913, CG34022, CG43772, Dmel\\CG43772, Dmel_CG13913, Dmel_CG34022', description='multiple wing hairs', other_designations='CG43772-PA|CG43772-PB|CG43772-PC|multiple wing hair|multiple-wing-hairs|mwh-PA|mwh-PB|mwh-PC', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1200786, end_position_on_the_genomic_accession=1232700, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38133, CurrentID=0, Status='live', Symbol='CG9194', Aliases='Dmel_ C9194, Dmel\\CG9194, anon-WO0109301.19, anon-WO0109301.20, anon-WO0109301.21, anon-WO0109301.22, anon-WO0109301.23, anon-WO0109301.24, anon-WO0109301.25, anon-WO0109301.26, anon-WO0109301.27, anon-WO0109301.28, anon-WO0109301.29, anon-WO0109301.3, anon-WO0109301.31, anon-WO0109301.32, anon-WO0109301.33, anon-WO0109301.34, anon-WO0109301.36, anon-WO0109301.37, anon-WO0109301.38', description='CG9194 gene product from transcript CG9194-RA', other_designations='CG9194-PA', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1235726, end_position_on_the_genomic_accession=1242025, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38134, CurrentID=0, Status='live', Symbol='CG9192', Aliases='Dmel_ Dmel\\CG9192', description='CG9192 gene product from transcript CG9192-RA', other_designations='CG9192-PA', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1237133, end_position_on_the_genomic_accession=1238319, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38135, CurrentID=0, Status='live', Symbol='Klp61F', Aliases='Dmel_CG9191, 6345, CG9191, CG9191-PA, DmKLP61F, DmKRP130, DmKlp61F, Dmel\\CG9191, KIF 8, KIF11, KLP2, KLP61, KLP61F, KLP61f, KRP130, KRP[[(130)]], KRP[[130]], Kin-5, Klp 61F, Klp2/KRP130, Klp61f, Krp130, klp61F, klp61f, l(3)07012, l(3)6D283, l(3)Klp61F, urc, urcn, Klp61F', description='Kinesin-like protein at 61F', other_designations='CG9191-PA|Kinesin-related protein-130kD|Klp61F-PA|kinesin-5|kinesin-5 motor|kinesin-5s|kinesin-like protein|urchin', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1241568, end_position_on_the_genomic_accession=1245910, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38136, CurrentID=0, Status='live', Symbol='Psf1', Aliases='Dmel_CG9187, CG9187, CG9187-PA, Dmel\\CG9187, psf1', description='CG9187 gene product from transcript CG9187-RA', other_designations='CG9187-PA|Psf1-PA', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1246986, end_position_on_the_genomic_accession=1247677, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38137, CurrentID=0, Status='live', Symbol='Sac1', Aliases='Dmel_CG9128, CG9128, Dmel\\CG9128, dsac1, l(3)2107, sac1', description='CG9128 gene product from transcript CG9128-RB', other_designations='CG9128-PA|CG9128-PB|Sac1-PA|Sac1-PB|suppressor of actin-1', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1247817, end_position_on_the_genomic_accession=1250451, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38138, CurrentID=0, Status='live', Symbol='CG9129', Aliases='Dmel_ Dmel\\CG9129', description='CG9129 gene product from transcript CG9129-RA', other_designations='CG9129-PA', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1250696, end_position_on_the_genomic_accession=1251778, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38141, CurrentID=0, Status='live', Symbol='CG9134', Aliases='Dmel_ Dmel\\CG9134', description='CG9134 gene product from transcript CG9134-RB', other_designations='CG9134-PA|CG9134-PB|CG9134-PC', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1256449, end_position_on_the_genomic_accession=1274498, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38142, CurrentID=0, Status='live', Symbol='CG32333', Aliases='Dmel_ CG9137, CG9146, Dmel\\CG32333', description='CG32333 gene product from transcript CG32333-RC', other_designations='CG32333-PA|CG32333-PB|CG32333-PC|uncharacterized protein', map_location='61F4-61F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1275956, end_position_on_the_genomic_accession=1295699, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  5740757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740757, CurrentID=0, Status='live', Symbol='CG34316', Aliases='Dmel_ Dm_3R:60797, Dmel\\CG34316, JHBP', description='CG34316 gene product from transcript CG34316-RA', other_designations='CG34316-PA|uncharacterized protein', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15230585, end_position_on_the_genomic_accession=15231686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38144, CurrentID=0, Status='live', Symbol='Rabex-5', Aliases='Dmel_CG9139, CG9139, Dmel\\CG9139, RabGEF1, Rabex5', description='Rabaptin-5-associated exchange factor for Rab5 ortholog (H. sapiens)', other_designations='CG9139-PA|Rabaptin-5-associated exchange factor for Rab5 ortholog|Rabex-5-PA', map_location='61F5-61F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1295913, end_position_on_the_genomic_accession=1298497, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38145, CurrentID=0, Status='live', Symbol='scf', Aliases='Dmel_CG9148, BEST:CK02682, CG9148, CK02682, DCB-45, Dmel\\CG9148, SCF, anon-EST:Posey14', description='supercoiling factor', other_designations='CG9148-PA|CG9148-PB|DNA supercoiling factor|scf-PA|scf-PB', map_location='61F5-61F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1298672, end_position_on_the_genomic_accession=1300389, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38146, CurrentID=0, Status='live', Symbol='Rac1', Aliases='Dmel_CG2248, 2248, CG2248, D-Rac, D-Rac 1, D- DRac, DRac1, DRacA, Dm Rac1, DmRAC, DmRAC1, Dmel\\CG2248, Dmrac1, Drac, Drac1, Drac1a, DracA, GTPase Rac1, P40792, RAC1, Rac, Rac GTPase, Rac-1, Rac3, RacA, dRac, dRac1, rac, rac1', description='CG2248 gene product from transcript CG2248-RB', other_designations='CG2248-PA|CG2248-PB|Rac|Rac GTPase|Rac1-PA|Rac1-PB', map_location='61F5-61F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1300879, end_position_on_the_genomic_accession=1302683, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38147, CurrentID=0, Status='live', Symbol='CG9149', Aliases='Dmel_ ACAT, Dmel\\CG9149', description='CG9149 gene product from transcript CG9149-RA', other_designations='CG9149-PA', map_location='61F5-61F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1302944, end_position_on_the_genomic_accession=1304717, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38148, CurrentID=0, Status='live', Symbol='CG2277', Aliases='Dmel_ 28575006, Dmel\\CG2277', description='CG2277 gene product from transcript CG2277-RB', other_designations='CG2277-PA|CG2277-PB|CG2277-PC', map_location='61F5-61F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1304850, end_position_on_the_genomic_accession=1307418, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38149, CurrentID=0, Status='live', Symbol='CG2469', Aliases='Dmel_ Dmel\\CG2469, cg2469', description='CG2469 gene product from transcript CG2469-RB', other_designations='CG2469-PA|CG2469-PB', map_location='61F5-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1307224, end_position_on_the_genomic_accession=1311805, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38150, CurrentID=0, Status='live', Symbol='CG9186', Aliases='Dmel_ Dmel\\CG9186', description='CG9186 gene product from transcript CG9186-RB', other_designations='CG9186-PA|CG9186-PB', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1311719, end_position_on_the_genomic_accession=1313377, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38151, CurrentID=0, Status='live', Symbol='CG9153', Aliases='Dmel_ Dmel\\CG9153, FBgn0035207, NM_167868.1', description='CG9153 gene product from transcript CG9153-RA', other_designations='CG9153-PA|CG9153-PB|CG9153-PC', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1313696, end_position_on_the_genomic_accession=1318330, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38152, CurrentID=0, Status='live', Symbol='CG9184', Aliases='Dmel_ Dmel\\CG9184', description='CG9184 gene product from transcript CG9184-RB', other_designations='CG9184-PA|CG9184-PB', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1321792, end_position_on_the_genomic_accession=1322806, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38153, CurrentID=0, Status='live', Symbol='Myo61F', Aliases='Dmel_CG9155, CG9155, Dm IB, DmIB, Dmel\\CG9155, DroMIB, IB, MIB, MYO61F, MYOIB, Myo-IB, Myo1B, MyoIC, myo61F, myo61f', description='Myosin 61F', other_designations='CG9155-PA|CG9155-PB|CG9155-PC|CG9155-PD|Myo61F-PA|Myo61F-PB|Myo61F-PC|Myo61F-PD|Myosin-IB|myosin 1b|myosin 61F|myosin I|myosin IB|myosin61F', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1318480, end_position_on_the_genomic_accession=1329617, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38154, CurrentID=0, Status='live', Symbol='ND-ACP', Aliases='Dmel_CG9160, ACPM, CG9160, Dmel\\CG9160, ND-AcC, l(3)1303, l(3)j4A6, mtACP, mtACP1, mtacp1', description='NADH dehydrogenase (ubiquinone) acyl carrier protein', other_designations='CG9160-PA|CG9160-PB|CG9160-PC|NADH ubiquinone oxidoreductase ACP|NADH ubiquinone oxidoreductase acyl carrier protein|NADH-ubiquinone oxidoreductase acyl carrier protein|NADH-ubiquinone oxidoreductase acyl-carrier subunit|NADH:ubiquinone oxidoreductase acyl carrier|ND-ACP-PA|ND-ACP-PB|ND-ACP-PC|acyl carrier protein|mitochondrial acyl carrier protein|mitochondrial acyl carrier protein 1|mitochondrial acyl carrier protein of electron transport complex I', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1329949, end_position_on_the_genomic_accession=1331631, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38155, CurrentID=0, Status='live', Symbol='msd1', Aliases='Dmel_CG13914, CG13914, Dmel\\CG13914', description='mitotic spindle density 1', other_designations='CG13914-PA|msd1-PA', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1331744, end_position_on_the_genomic_accession=1332326, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38156, CurrentID=0, Status='live', Symbol='msd5', Aliases='Dmel_CG2213, CG2213, Dgt7, Dmel\\CG2213', description='mitotic spindle density 5', other_designations='CG2213-PA|msd5-PA', map_location='61F6-61F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1332307, end_position_on_the_genomic_accession=1333146, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38157, CurrentID=0, Status='live', Symbol='CG2211', Aliases='Dmel_ Dmel\\CG2211, anon-WO0118547.230', description='CG2211 gene product from transcript CG2211-RA', other_designations='CG2211-PA|CG2211-PB', map_location='61F6-61F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1333239, end_position_on_the_genomic_accession=1335359, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38159, CurrentID=0, Status='live', Symbol='CG2199', Aliases='Dmel_ Dmel\\CG2199', description='CG2199 gene product from transcript CG2199-RB', other_designations='CG2199-PA|CG2199-PB|CG2199-PC', map_location='61F7-61F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1339460, end_position_on_the_genomic_accession=1342403, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38160, CurrentID=0, Status='live', Symbol='Ptp61F', Aliases='Dmel_CG9181, BEST:LP01280, CG9178, CG9181, CPtp62A, DCPTP62A, DPTP61F, Dmel\\CG9181, PTP1B, PTP61F, Ptp-61F, R-PTP 61F, anon-WO0118547.297, anon-WO0140519.98, dPTP61F, dptp61F, ptp61F', description='Protein tyrosine phosphatase 61F', other_designations='CG9181-PA|CG9181-PB|CG9181-PC|CG9181-PD|CG9181-PE|Ptp61F-PA|Ptp61F-PB|Ptp61F-PC|Ptp61F-PD|Ptp61F-PE|cytoplasmic protein tyrosine phosphatase 62A|protein tyrosine phosphatase|protein tyrosine phosphatase 61F|protein-tyrosine-phosphatase', map_location='61F7-62A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1342493, end_position_on_the_genomic_accession=1475257, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38161, CurrentID=0, Status='live', Symbol='312', Aliases='Dmel_CG9166, BEST:CK01660, CG9166, Dmel\\CG9166', description='CG9166 gene product from transcript CG9166-RA', other_designations='312-PA|312-PB|CG9166-PA|CG9166-PB', map_location='61F7-61F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1351116, end_position_on_the_genomic_accession=1352853, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38163, CurrentID=0, Status='live', Symbol='CG9168', Aliases='Dmel_ Dmel\\CG9168, gi23092768', description='CG9168 gene product from transcript CG9168-RA', other_designations='CG9168-PA', map_location='62A1-62A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1411088, end_position_on_the_genomic_accession=1413089, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38164, CurrentID=0, Status='live', Symbol='FucTD', Aliases='Dmel_CG9169, CG9169, Dmel\\CG9169', description='CG9169 gene product from transcript CG9169-RA', other_designations='CG9169-PA|FucTD-PA', map_location='62A1-62A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1416231, end_position_on_the_genomic_accession=1417933, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38165, CurrentID=0, Status='live', Symbol='CG9173', Aliases='Dmel_ Dmel\\CG9173', description='CG9173 gene product from transcript CG9173-RA', other_designations='CG9173-PA', map_location='62A1-62A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1421199, end_position_on_the_genomic_accession=1423471, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38167, CurrentID=0, Status='live', Symbol='SA-2', Aliases='Dmel_CG13916, CG1, CG13916, Dmel\\CG13916, SNM, anon-WO0109301.35, pf-1, snm', description='Stromalin-2', other_designations='CG13916-PC|SA-2-PC|complementation group 1|pairing failure-1|stromalin in meiosis', map_location='62A1-62A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1434336, end_position_on_the_genomic_accession=1437872, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38168, CurrentID=0, Status='live', Symbol='rho', Aliases='Dmel_CG1004, CG1004, DMRHO, DMRHOa, DMRHOb, DRORHO, Dmel\\CG1004, Dmel\\rho, Dmrho, RHO, RHOb, Rho, Rho-1, Rho1, Ve, iks-1, rho1, rhom, rhomboid/veinlet, ve, rho', description='rhomboid', other_designations='CG1004-PA|CG1004-PB|rho-PA|rho-PB|rhomboid 1|rhomboid-1|rhomboid/veinlet|rhomboid1|veinlet', map_location='3-0.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1463811, end_position_on_the_genomic_accession=1468944, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38169, CurrentID=0, Status='live', Symbol='stet', Aliases='Dmel_CG33166, B-Rho, Brho, CG12083, CG12504, CG33166, Dmel\\CG33166, Rho-2, Rho2, brho, rho-2, rho2', description='stem cell tumor', other_designations='CG33166-PA|CG33166-PB|brother of rhomboid|rhomboid-2|stet-PA|stet-PB', map_location='62A2-62A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1475438, end_position_on_the_genomic_accession=1489547, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38170, CurrentID=0, Status='live', Symbol='robl62A', Aliases='Dmel_CG1014, CG1014, Dmel\\CG1014, rob162A', description='CG1014 gene product from transcript CG1014-RA', other_designations='CG1014-PA|CG1014-PB|robl62A-PA|robl62A-PB', map_location='62A2-62A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1477619, end_position_on_the_genomic_accession=1478369, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38173, CurrentID=0, Status='live', Symbol='pUf68', Aliases='Dmel_CG12085, CG12085, DPUF68, Dmel\\CG12085, Fir, HFP, Hfp, PUF60, PUF68, Puf60, Puf68, anon-WO0118547.408, dFIR, hfp, pUF68, pUbsf, puf68', description='poly U binding factor 68kD', other_designations='CG12085-PA|CG12085-PB|CG12085-PC|CG12085-PD|CG12085-PE|CG12085-PF|CG12085-PG|CG12085-PH|CG12085-PI|CG12085-PJ|half pint|half-pint|pUf68-PA|pUf68-PB|pUf68-PC|pUf68-PD|pUf68-PE|pUf68-PF|pUf68-PG|pUf68-PH|pUf68-PI|pUf68-PJ|polyU binding splicing factor', map_location='62A2-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1496853, end_position_on_the_genomic_accession=1502590, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38174, CurrentID=0, Status='live', Symbol='cue', Aliases='Dmel_CG12086, CG12086, CT5570, Dmel\\CG12086, ms(3)02509', description='cueball', other_designations='CG12086-PA|cue-PA', map_location='62A3-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1503226, end_position_on_the_genomic_accession=1517271, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38175, CurrentID=0, Status='live', Symbol='Psa', Aliases='Dmel_CG1009, CG1009, Dmel\\CG1009, PSA, dPSA, dPsa, l(3)06226', description='Puromycin sensitive aminopeptidase', other_designations='CG1009-PA|CG1009-PB|CG1009-PC|CG1009-PD|CG1009-PE|CG1009-PF|CG1009-PH|CG1009-PI|Psa-PA|Psa-PB|Psa-PC|Psa-PD|Psa-PE|Psa-PF|Psa-PH|Psa-PI|lethal (3) 06226|puromycin sensitive aminopeptidase|puromycin-sensitive aminopeptidase|puromycin-sensitive-aminopeptidase', map_location='62A3-62A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1502843, end_position_on_the_genomic_accession=1523380, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38176, CurrentID=0, Status='live', Symbol='Iml1', Aliases='Dmel_CG12090, CG12090, Dmel\\CG12090, SEA1, iml1', description='Increased minichromosome loss 1', other_designations='CG12090-PA|CG12090-PB|CG12090-PC|CG12090-PD|CG12090-PE|Iml1-PA|Iml1-PB|Iml1-PC|Iml1-PD|Iml1-PE|increased minichromosome loss 1', map_location='62A3-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1523414, end_position_on_the_genomic_accession=1534034, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  38177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38177, CurrentID=0, Status='live', Symbol='CG12091', Aliases='Dmel_ Dmel\\CG12091', description='CG12091 gene product from transcript CG12091-RB', other_designations='CG12091-PA|CG12091-PB|uncharacterized protein', map_location='62A3-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1534929, end_position_on_the_genomic_accession=1536835, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38178, CurrentID=0, Status='live', Symbol='CG7852', Aliases='Dmel_ Dmel\\CG7852', description='CG7852 gene product from transcript CG7852-RC', other_designations='CG7852-PA|CG7852-PB|CG7852-PC', map_location='62A3-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1537137, end_position_on_the_genomic_accession=1544823, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  38180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38180, CurrentID=0, Status='live', Symbol='Cct2', Aliases='Dmel_CG18330, CCT2, CG18330, CT1187, Dmel\\CG18330, cct2', description='CTP:phosphocholine cytidylyltransferase 2', other_designations='CG18330-PA|Cct2-PA', map_location='62A5-62A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1553493, end_position_on_the_genomic_accession=1555315, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38181, CurrentID=0, Status='live', Symbol='CG12099', Aliases='Dmel_ Dmel\\CG12099', description='CG12099 gene product from transcript CG12099-RA', other_designations='CG12099-PA|CG12099-PB|uncharacterized protein', map_location='62A5-62A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1555363, end_position_on_the_genomic_accession=1558484, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38182, CurrentID=0, Status='live', Symbol='Pex10', Aliases='Dmel_CG7864, CG7864, DmelPex10, Dmel\\CG7864, PEX10, pex10', description='Peroxin 10', other_designations='CG7864-PA|CG7864-PB|Pex10-PA|Pex10-PB|peroxin 10', map_location='62A5-62A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1558228, end_position_on_the_genomic_accession=1559863, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38183, CurrentID=0, Status='live', Symbol='CG12003', Aliases='Dmel_ Dmel\\CG12003', description='CG12003 gene product from transcript CG12003-RA', other_designations='CG12003-PA|uncharacterized protein', map_location='62A5-62A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1565392, end_position_on_the_genomic_accession=1567802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38184, CurrentID=0, Status='live', Symbol='CG7879', Aliases='Dmel_ Dmel\\CG7879, cg7879', description='CG7879 gene product from transcript CG7879-RB', other_designations='CG7879-PA|CG7879-PB', map_location='62A5-62A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1559813, end_position_on_the_genomic_accession=1568608, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38185, CurrentID=0, Status='live', Symbol='CG12004', Aliases='Dmel_ Dmel\\CG12004, anon-EST:fe3A8', description='CG12004 gene product from transcript CG12004-RD', other_designations='CG12004-PA|CG12004-PB|CG12004-PC|CG12004-PD|anon-fast-evolving-3A8|uncharacterized protein', map_location='62A6-62A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1569181, end_position_on_the_genomic_accession=1576330, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38186, CurrentID=0, Status='live', Symbol='CG13917', Aliases='Dmel_ Dmel\\CG13917', description='CG13917 gene product from transcript CG13917-RC', other_designations='CG13917-PA|CG13917-PB|CG13917-PC|CG13917-PD|uncharacterized protein', map_location='62A7-62A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1576699, end_position_on_the_genomic_accession=1600845, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  50055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50055, CurrentID=0, Status='live', Symbol='CG5791', Aliases='Dmel_ BcDNA:RH16331, Dmel\\CG5791', description='CG5791 gene product from transcript CG5791-RB', other_designations='CG5791-PA|CG5791-PB|CG5791-PC', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22087075, end_position_on_the_genomic_accession=22088501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38190, CurrentID=0, Status='live', Symbol='CG12105', Aliases='Dmel_ Dmel\\CG12105', description='CG12105 gene product from transcript CG12105-RA', other_designations='CG12105-PA|CG12105-PB|uncharacterized protein', map_location='62A8-62A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1610696, end_position_on_the_genomic_accession=1615801, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38192, CurrentID=0, Status='live', Symbol='CG13926', Aliases='Dmel_ Dmel\\CG13926', description='CG13926 gene product from transcript CG13926-RA', other_designations='CG13926-PA|uncharacterized protein', map_location='62A8-62A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1617348, end_position_on_the_genomic_accession=1618158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38193, CurrentID=0, Status='live', Symbol='ABCB7', Aliases='Dmel_CG7955, CG7955, Dmel\\CG7955, d', description='CG7955 gene product from transcript CG7955-RA', other_designations='ABCB7-PA|ABCB7-PB|CG7955-PA|CG7955-PB', map_location='62A8-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1618414, end_position_on_the_genomic_accession=1623413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38194, CurrentID=0, Status='live', Symbol='GC', Aliases='Dmel_CG13927, CG13927, DgammaC, Dmel\\CG13927, dGC, dgammaC, gammaC', description='gamma-glutamyl carboxylase', other_designations='CG13927-PA|CG13927-PB|GC-PA|GC-PB|gamma-glutamyl-carboxylase', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1623179, end_position_on_the_genomic_accession=1629301, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38195, CurrentID=0, Status='live', Symbol='CG13928', Aliases='Dmel_ Dmel\\CG13928', description='CG13928 gene product from transcript CG13928-RA', other_designations='CG13928-PA|uncharacterized protein', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1625791, end_position_on_the_genomic_accession=1628377, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38196, CurrentID=0, Status='live', Symbol='nSyb', Aliases='Dmel_CG17248, CG17248, Dmel\\CG17248, Dn-syb, N-SYB, N-Syb, N-syb, NSYB, Nsyb, SYB, Syb, Syb62A, n-Syb, n-syb, nSYB, nsyb, syb', description='neuronal Synaptobrevin', other_designations='CG17248-PA|CG17248-PE|CG17248-PF|CG17248-PG|CG17248-PH|CG17248-PI|CG17248-PJ|CG17248-PK|N-synaptobrevin|Synaptobrevin-62A|VAMP3|nSyb-PA|nSyb-PE|nSyb-PF|nSyb-PG|nSyb-PH|nSyb-PI|nSyb-PJ|nSyb-PK|neural-synaptobrevin|neuronal SYNAPTOBREVIN|neuronal synaptobrevin|neuronal synaptogrevin|neuronal-Synaptobrevin|synaptobrevin', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1630348, end_position_on_the_genomic_accession=1637752, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38197, CurrentID=0, Status='live', Symbol='metl', Aliases='Dmel_CG13929, CG13929, Dmel\\CG13929, anon-EST:Posey109', description='methyltransferase-like', other_designations='CG13929-PA|CG13929-PB|metl-PA|metl-PB', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1637942, end_position_on_the_genomic_accession=1639235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38198, CurrentID=0, Status='live', Symbol='Bgb', Aliases='Dmel_CG7959, CG7959, Dmel\\CG7959, bgb, en(lz)D/9, rpbeta2', description='Big brother', other_designations='Bgb-PA|Bgb-PB|Big-brother|CG7959-PA|CG7959-PB|runt domain partner beta2', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1639711, end_position_on_the_genomic_accession=1640952, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38199, CurrentID=0, Status='live', Symbol='alphaCOP', Aliases='Dmel_CG7961, CG7961, Dmel\\CG7961, alpha-COP, alphaCop', description='Coat Protein (coatomer) alpha', other_designations='CG7961-PA|CG7961-PB|COPI alpha subunit|alpha-coatomer protein|alphaCOP-PA|alphaCOP-PB|coat protein (coatomer) alpha|coatomer I', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1641066, end_position_on_the_genomic_accession=1645199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38200, CurrentID=0, Status='live', Symbol='CG13919', Aliases='Dmel_ Dmel\\CG13919', description='CG13919 gene product from transcript CG13919-RA', other_designations='CG13919-PA|uncharacterized protein', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1645467, end_position_on_the_genomic_accession=1645983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38201, CurrentID=0, Status='live', Symbol='CG17249', Aliases='Dmel_ Dmel\\CG17249', description='CG17249 gene product from transcript CG17249-RA', other_designations='CG17249-PA|uncharacterized protein', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1645931, end_position_on_the_genomic_accession=1647464, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38202, CurrentID=0, Status='live', Symbol='Bro', Aliases='Dmel_CG7960, CG7960, Dmel\\CG7960, bro, rpbeta1', description='Brother', other_designations='Bro-PA|CG7960-PA|beta for runt and OTHERs|brother|runt domain partner beta1', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1647869, end_position_on_the_genomic_accession=1648612, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38203, CurrentID=0, Status='live', Symbol='CG13920', Aliases='Dmel_ BEST:CK00246, CK00246, Dmel\\CG13920', description='CG13920 gene product from transcript CG13920-RC', other_designations='CG13920-PA|CG13920-PC|uncharacterized protein', map_location='62A9-62A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1648944, end_position_on_the_genomic_accession=1651501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38204, CurrentID=0, Status='live', Symbol='CG7967', Aliases='Dmel_ Dmel\\CG7967', description='CG7967 gene product from transcript CG7967-RA', other_designations='CG7967-PA', map_location='62A9-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1651432, end_position_on_the_genomic_accession=1652900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38205, CurrentID=0, Status='live', Symbol='CG7970', Aliases='Dmel_ Dmel\\CG7970, anon-WO0118547.268', description='CG7970 gene product from transcript CG7970-RC', other_designations='CG7970-PA|CG7970-PB|CG7970-PC', map_location='62A10-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1652750, end_position_on_the_genomic_accession=1654238, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38206, CurrentID=0, Status='live', Symbol='CG7971', Aliases='Dmel_ BcDNA:GH04236, Dmel\\CG7971, anon-EST:fe1A7', description='CG7971 gene product from transcript CG7971-RC', other_designations='CG7971-PA|CG7971-PB|CG7971-PC|CG7971-PF|CG7971-PG|CG7971-PH|anon-fast-evolving-1A7', map_location='62A10-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1654247, end_position_on_the_genomic_accession=1665064, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  38207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38207, CurrentID=0, Status='live', Symbol='CG7974', Aliases='Dmel_ Dmel\\CG7974', description='CG7974 gene product from transcript CG7974-RB', other_designations='CG7974-PA|CG7974-PB', map_location='62A10-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1665138, end_position_on_the_genomic_accession=1666594, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38208, CurrentID=0, Status='live', Symbol='RpL23A', Aliases='Dmel_CG7977, 153215_at, CG7977, Dmel\\CG7977, L23A, L23a, PBP-3, RPL23a, Rp L23A, RpL23a, RpL25, anon-EST:Liang-1.4, anon-EST:fe2D1, clone 1.4, rpL25, rpl23a', description='Ribosomal protein L23A', other_designations='CG7977-PA|PARP-binding protein-3|RpL23A-PA|anon-fast-evolving-2D1|ribosomal protein L23A|ribosomal protein L25', map_location='62A10-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1667192, end_position_on_the_genomic_accession=1668654, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38209, CurrentID=0, Status='live', Symbol='RabX5', Aliases='Dmel_CG7980, CG7980, Dm Dmel\\CG7980, Q9W0A7, rabX5', description='CG7980 gene product from transcript CG7980-RB', other_designations='CG7980-PB|RabX5-PB', map_location='62A10-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1668901, end_position_on_the_genomic_accession=1670108, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38210, CurrentID=0, Status='live', Symbol='CG13930', Aliases='Dmel_ Dmel\\CG13930', description='CG13930 gene product from transcript CG13930-RA', other_designations='CG13930-PA|uncharacterized protein', map_location='62A10-62A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1670407, end_position_on_the_genomic_accession=1672727, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38211, CurrentID=0, Status='live', Symbol='CG12011', Aliases='Dmel_ Dmel\\CG12011, anon-EST:Liang-1.72, clone 1.72', description='CG12011 gene product from transcript CG12011-RA', other_designations='CG12011-PA|uncharacterized protein', map_location='62A11-62A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1694617, end_position_on_the_genomic_accession=1696412, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38212, CurrentID=0, Status='live', Symbol='CG13931', Aliases='Dmel_ Dmel\\CG13931', description='CG13931 gene product from transcript CG13931-RB', other_designations='CG13931-PA|CG13931-PB|uncharacterized protein', map_location='62A12-62A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1700956, end_position_on_the_genomic_accession=1702407, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38214, CurrentID=0, Status='live', Symbol='CG7991', Aliases='Dmel_ Dmel\\CG7991', description='CG7991 gene product from transcript CG7991-RB', other_designations='CG7991-PB|CG7991-PC', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1674709, end_position_on_the_genomic_accession=1715541, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5740769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740769, CurrentID=0, Status='live', Symbol='CG44005', Aliases='Dmel_ CG34252, Dmel\\CG44005', description='CG44005 gene product from transcript CG44005-RB', other_designations='CG44005-PA|CG44005-PB|uncharacterized protein', map_location='75B6-75B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18070300, end_position_on_the_genomic_accession=18072123, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38216, CurrentID=0, Status='live', Symbol='CG18171', Aliases='Dmel_ Dmel\\CG18171', description='CG18171 gene product from transcript CG18171-RA', other_designations='CG18171-PA|uncharacterized protein', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1720616, end_position_on_the_genomic_accession=1723153, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38217, CurrentID=0, Status='live', Symbol='CG12035', Aliases='Dmel_ Dmel\\CG12035', description='CG12035 gene product from transcript CG12035-RA', other_designations='CG12035-PA|uncharacterized protein', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1724390, end_position_on_the_genomic_accession=1725647, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38218, CurrentID=0, Status='live', Symbol='drpr', Aliases='Dmel_CG2086, BcDNA:GH03529, CG18172, CG2086, CT41022, CT6730, Dmel\\CG2086, Drpr', description='draper', other_designations='CG2086-PA|CG2086-PB|CG2086-PC|CG2086-PD|CG2086-PE|CG2086-PF|drpr-PA|drpr-PB|drpr-PC|drpr-PD|drpr-PE|drpr-PF', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1715595, end_position_on_the_genomic_accession=1731107, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38219, CurrentID=0, Status='live', Symbol='Oseg4', Aliases='Dmel_CG2069, CG2069, Dmel\\CG2069, OSEG, OSEG4, WDR35, Wdr35, cytC40, oseg4', description='Outer segment 4', other_designations='CG2069-PA|CG2069-PB|Oseg4-PA|Oseg4-PB|outer segment 4', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1730993, end_position_on_the_genomic_accession=1735767, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38220, CurrentID=0, Status='live', Symbol='CG18173', Aliases='Dmel_ Dmel\\CG18173', description='CG18173 gene product from transcript CG18173-RB', other_designations='CG18173-PA|CG18173-PB|uncharacterized protein', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1735853, end_position_on_the_genomic_accession=1738106, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38221, CurrentID=0, Status='live', Symbol='Gk', Aliases='Dmel_CG7995, CG7995, Dmel\\CG7995, dGK', description='Glycerol kinase', other_designations='CG7995-PA|CG7995-PB|CG7995-PC|CG7995-PD|CG7995-PE|CG7995-PF|Gk-PA|Gk-PB|Gk-PC|Gk-PD|Gk-PE|Gk-PF|glycerol kinase', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1738407, end_position_on_the_genomic_accession=1745421, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38222, CurrentID=0, Status='live', Symbol='CG13921', Aliases='Dmel_ Dmel\\CG13921', description='CG13921 gene product from transcript CG13921-RF', other_designations='CG13921-PB|CG13921-PC|CG13921-PD|CG13921-PE|CG13921-PF|uncharacterized protein', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1746907, end_position_on_the_genomic_accession=1752390, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38223, CurrentID=0, Status='live', Symbol='Cht2', Aliases='Dmel_CG2054, CG2054, CHT2, DmCHT2, Dmel\\CG2054, cht2', description='Chitinase 2', other_designations='CG2054-PA|CG2054-PB|CG2054-PC|Cht2-PA|Cht2-PB|Cht2-PC|chitinase 2', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1752328, end_position_on_the_genomic_accession=1755934, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38224, CurrentID=0, Status='live', Symbol='CG8001', Aliases='Dmel_ Dmel\\CG8001', description='CG8001 gene product from transcript CG8001-RA', other_designations='CG8001-PA|CG8001-PB|CG8001-PC|CG8001-PD', map_location='62B1-62B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1756159, end_position_on_the_genomic_accession=1759208, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38226, CurrentID=0, Status='live', Symbol='Dhc62B', Aliases='Dmel_CG15804, CG15804, CG18613, DHC, Dhc, Dmel\\CG15804, dhc', description='Dynein heavy chain at 62B', other_designations='CG15804-PC|Dhc62B-PC|dynein heavy chain 62B|dynein heavy chain at 62B|dynein-related heavy chain polypeptide', map_location='62B1-62B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1759172, end_position_on_the_genomic_accession=1771758, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  5740771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740771, CurrentID=0, Status='live', Symbol='CG34235', Aliases='Dmel_ Dmel\\CG34235', description='CG34235 gene product from transcript CG34235-RB', other_designations='CG34235-PB|uncharacterized protein', map_location='49F9-49F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13209809, end_position_on_the_genomic_accession=13212237, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38228, CurrentID=0, Status='live', Symbol='CG12018', Aliases='Dmel_ 12018, BcDNA:HL03874, Dmel\\CG12018', description='CG12018 gene product from transcript CG12018-RA', other_designations='CG12018-PA|DNA polymerase delta small subunit|DNApol- delta small subunit|uncharacterized protein', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1775633, end_position_on_the_genomic_accession=1777019, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38229, CurrentID=0, Status='live', Symbol='CG2021', Aliases='Dmel_ Dmel\\CG2021', description='CG2021 gene product from transcript CG2021-RA', other_designations='CG2021-PA', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1777010, end_position_on_the_genomic_accession=1777411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38230, CurrentID=0, Status='live', Symbol='mRpL46', Aliases='Dmel_CG13922, CG13922, Dmel\\CG13922, MRP-L46, MRpL46', description='mitochondrial ribosomal protein L46', other_designations='CG13922-PA|mRpL46-PA', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1777439, end_position_on_the_genomic_accession=1778347, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38231, CurrentID=0, Status='live', Symbol='alpha-Spec', Aliases='Dmel_CG1977, 3A9, CG1977, Dmel\\CG1977, FBtr0072789, FCP-B, Spec, Spectrin, alfa-Spec, alpha, alpha-SPEC, alpha-Sp, alpha-Spc, alpha-Spet, alpha-Spectrin, alpha-spec, alpha-spectrin, alphaSp, alphaSpec, alphaSpectrin, anon-EST:fe2E2, dre3, klo, l(3)04276, l(3)62Bd, l(3)alpha-Spec, l(3)dre3, spectrin, alpha-Spec', description='alpha Spectrin', other_designations='CG1977-PA|CG1977-PB|CG1977-PC|CG1977-PD|aSpc 3A9|alpha spectrin|alpha-SPECTRIN|alpha-Spec-PA|alpha-Spec-PB|alpha-Spec-PC|alpha-Spec-PD|alpha-spectrin II|alphaSpectrin|anon-fast-evolving-2E2|klotzchen|spectrin', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1778485, end_position_on_the_genomic_accession=1795328, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38232, CurrentID=0, Status='live', Symbol='Cdc37', Aliases='Dmel_CG12019, CG12019, CPDR-1, CPDR-2, DROCPDR, Dmel\\CG12019, E(sev)3B, cdc37', description='CG12019 gene product from transcript CG12019-RA', other_designations='CG12019-PA|Cdc37-PA|cdc37 protein', map_location='3-5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1795392, end_position_on_the_genomic_accession=1797307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38233, CurrentID=0, Status='live', Symbol='CG12020', Aliases='Dmel_ Dmel\\CG12020', description='CG12020 gene product from transcript CG12020-RA', other_designations='CG12020-PA|uncharacterized protein', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1797305, end_position_on_the_genomic_accession=1798610, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38235, CurrentID=0, Status='live', Symbol='JTBR', Aliases='Dmel_CG1935, BcDNA:RH07980, CG1935, Dmel\\CG1935, anon-EST:fe2G4, dJTB-like, d', description='CG1935 gene product from transcript CG1935-RB', other_designations='CG1935-PA|CG1935-PB|CG1935-PC|JTBR-PA|JTBR-PB|JTBR-PC|anon-fast-evolving-2G4|jumping translocation breakpoint', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1802008, end_position_on_the_genomic_accession=1802856, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38238, CurrentID=0, Status='live', Symbol='GV1', Aliases='Dmel_CG12023, BcDNA:RH49123, CG12023, DIP1, Dmel\\CG12023', description='CG12023 gene product from transcript CG12023-RA', other_designations='CG12023-PA|CG12023-PB|GV1-PA|GV1-PB', map_location='62B4-62B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1810588, end_position_on_the_genomic_accession=1822903, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38239, CurrentID=0, Status='live', Symbol='Cpr62Ba', Aliases='Dmel_CG13934, CG13934, DmelCpr62Ba, Dmel\\CG13934', description='Cuticular protein 62Ba', other_designations='CG13934-PB|CG13934-PC|Cpr62Ba-PB|Cpr62Ba-PC|cuticular protein 62Ba', map_location='62B6-62B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1824129, end_position_on_the_genomic_accession=1831783, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38240, CurrentID=0, Status='live', Symbol='Cpr62Bb', Aliases='Dmel_CG13935, CG13935, DmelCpr62Bb, Dmel\\CG13935', description='Cuticular protein 62Bb', other_designations='CG13935-PA|CG13935-PB|CG13935-PC|Cpr62Bb-PA|Cpr62Bb-PB|Cpr62Bb-PC|cuticular protein 62Bb', map_location='62B6-62B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1832636, end_position_on_the_genomic_accession=1835164, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38241, CurrentID=0, Status='live', Symbol='Cpr62Bc', Aliases='Dmel_CG1919, CG1919, DmelCpr62Bc, Dmel\\CG1919', description='Cuticular protein 62Bc', other_designations='CG1919-PA|CG1919-PB|Cpr62Bc-PA|Cpr62Bc-PB|cuticular protein 62Bc', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1840281, end_position_on_the_genomic_accession=1842084, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38242, CurrentID=0, Status='live', Symbol='CNMa', Aliases='Dmel_CG13936, CG13936, Dmel\\CG13936', description='CNMamide', other_designations='CG13936-PB|CG13936-PC|CG13936-PD|CNMa-PB|CNMa-PC|CNMa-PD', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1850915, end_position_on_the_genomic_accession=1855591, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38243, CurrentID=0, Status='live', Symbol='CG12024', Aliases='Dmel_ Dmel\\CG12024', description='CG12024 gene product from transcript CG12024-RA', other_designations='CG12024-PA|CG12024-PB|CG12024-PC|uncharacterized protein', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1857165, end_position_on_the_genomic_accession=1859047, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38244, CurrentID=0, Status='live', Symbol='Rap1', Aliases='Dmel_CG1956, BEST:GH18528, C-ras3, CG1956, D-Rap1, DRap, DRap1, Dmel\\CG1956, Dras3, Dras62B, E(faf), EC3-7, FCG-C, P08645, R, RAP1, Rap, Ras3, RasIII, l(3)62Bf, l(3)R, rap1, ras 3, ras3', description='Rap1 GTPase', other_designations='CG1956-PA|CG1956-PB|CG1956-PC|Rap1-PA|Rap1-PB|Rap1-PC|enhancer of faf|roughened', map_location='3-1.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1859115, end_position_on_the_genomic_accession=1862239, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740774, CurrentID=0, Status='live', Symbol='CG34214', Aliases='Dmel_ Dm_2R:92173, Dmel\\CG34214', description='CG34214 gene product from transcript CG34214-RA', other_designations='CG34214-PA|uncharacterized protein', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24665427, end_position_on_the_genomic_accession=24666022, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38246, CurrentID=0, Status='live', Symbol='CG12025', Aliases='Dmel_ Dmel\\CG12025', description='CG12025 gene product from transcript CG12025-RC', other_designations='CG12025-PA|CG12025-PB|CG12025-PC|uncharacterized protein', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1865430, end_position_on_the_genomic_accession=1869491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38247, CurrentID=0, Status='live', Symbol='CG13924', Aliases='Dmel_ Dmel\\CG13924', description='CG13924 gene product from transcript CG13924-RC', other_designations='CG13924-PB|CG13924-PC|uncharacterized protein', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1870019, end_position_on_the_genomic_accession=1871564, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38248, CurrentID=0, Status='live', Symbol='dre4', Aliases='Dmel_CG1828, CG1828, DRE4/dSPT16, Dmel\\CG1828, Dre4, SPT16, Spt16, dSPT16, dSpt16, dspt16, l(3)62Bg, l(3) spt16', description='CG1828 gene product from transcript CG1828-RB', other_designations='CG1828-PA|CG1828-PB|dre4-PA|dre4-PB|lethal(3)62Bg|supressor of Ty element 16', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1871574, end_position_on_the_genomic_accession=1876312, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38249, CurrentID=0, Status='live', Symbol='CG13937', Aliases='Dmel_ Dmel\\CG13937, d4ST2', description='CG13937 gene product from transcript CG13937-RI', other_designations='CG13937-PD|CG13937-PE|CG13937-PF|CG13937-PG|CG13937-PH|CG13937-PI|uncharacterized protein', map_location='62B7-62B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1876551, end_position_on_the_genomic_accession=1883041, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38251, CurrentID=0, Status='live', Symbol='Tmhs', Aliases='Dmel_CG12026, CG12026, Dmel\\CG12026, TMHS, l(3)62Be, l(3)neo7, neo7', description='Tetraspan membrane protein in hair cell stereocilia ortholog', other_designations='CG12026-PA|CG12026-PB|Tmhs-PA|Tmhs-PB|lethal (3) 62Be|tetraspan membrane protein in hair cell stereocilia|tetraspan membrane protein in hair cell stereocilia ortholog', map_location='3-1.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1885657, end_position_on_the_genomic_accession=1891462, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38252, CurrentID=0, Status='live', Symbol='dsb', Aliases='Dmel_CG1887, CG1887, Dmel\\CG1887', description='debris buster', other_designations='CG1887-PD|CG1887-PF|CG1887-PG|dsb-PD|dsb-PF|dsb-PG', map_location='62B9-62B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1893408, end_position_on_the_genomic_accession=1910197, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38254, CurrentID=0, Status='live', Symbol='Dbx', Aliases='Dmel_CG42234, CG12361, CG13925, CG42234, Dmel\\CG42234, Dmel_CG12361, Dmel_CG13925, d dbx', description='CG42234 gene product from transcript CG42234-RC', other_designations='CG42234-PC|Dbx-PC', map_location='62B10-62B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1921307, end_position_on_the_genomic_accession=1933602, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38255, CurrentID=0, Status='live', Symbol='CG5687', Aliases='Dmel_ Dmel\\CG5687, anon-EST:Posey42, dSLC5A2', description='CG5687 gene product from transcript CG5687-RB', other_designations='CG5687-PA|CG5687-PB', map_location='62B11-62B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1933633, end_position_on_the_genomic_accession=1939103, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38256, CurrentID=0, Status='live', Symbol='Mfap1', Aliases='Dmel_CG1017, CG1017, Dmel\\CG1017, dmfap1', description='Microfibril-associated protein 1', other_designations='CG1017-PA|Mfap1-PA', map_location='62B11-62B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1941068, end_position_on_the_genomic_accession=1942872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38257, CurrentID=0, Status='live', Symbol='mu2', Aliases='Dmel_CG1960, CG1960, Dmel\\CG1960, MDC1, MU2, cDNA D, mu-2, mutator[2]', description='mutator 2', other_designations='CG1960-PA|CG1960-PB|CG1960-PC|mu2-PA|mu2-PB|mu2-PC', map_location='62B11-62B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1942734, end_position_on_the_genomic_accession=1947475, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38258, CurrentID=0, Status='live', Symbol='Cnb', Aliases='Dmel_CG5690, CG5690, Dmel\\CG5690, anon-62BCa, cDNA C, cnb', description='Centrobin', other_designations='CG5690-PA|CG5690-PB|Cnb-PA|Cnb-PB|centrobin', map_location='62B11-62B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1947565, end_position_on_the_genomic_accession=1950262, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38259, CurrentID=0, Status='live', Symbol='mv', Aliases='Dmel_CG42863, CG11814, CG42863, CG5691, Dmel\\CG42863, Dmel_CG11814, Dmel_CG5691', description='mauve', other_designations='CG42863-PB|mv-PB', map_location='62B11-62B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1950306, end_position_on_the_genomic_accession=1964178, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  38261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38261, CurrentID=0, Status='live', Symbol='SCOT', Aliases='Dmel_CG1140, BcDNA:RH54340, CG1140, CT1116, DmSCOT, Dmel\\CG1140, cDNA B', description='Succinyl-CoA:3-ketoacid CoA transferase', other_designations='3-oxoacid CoA-transferase|3-oxoacid-CoA transferase|3-oxoacid-CoA-transferase|CG1140-PA|CG1140-PB|CG1140-PC|SCOT-PA|SCOT-PB|SCOT-PC|citric acid synthetase', map_location='62B11-62B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1963143, end_position_on_the_genomic_accession=1966805, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38262, CurrentID=0, Status='live', Symbol='CG1927', Aliases='Dmel_ BcDNA:GH11112, Dmel\\CG1927, anon-62BCb, anon-WO0118547.340, cDNA A', description='CG1927 gene product from transcript CG1927-RA', other_designations='CG1927-PA|CG1927-PB', map_location='62B12-62B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1966725, end_position_on_the_genomic_accession=1970079, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38263, CurrentID=0, Status='live', Symbol='CG11815', Aliases='Dmel_ Dmel\\CG11815', description='CG11815 gene product from transcript CG11815-RA', other_designations='CG11815-PA|uncharacterized protein', map_location='62B12-62B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1970191, end_position_on_the_genomic_accession=1971200, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38264, CurrentID=0, Status='live', Symbol='CG1139', Aliases='Dmel_ Dmel\\CG1139', description='CG1139 gene product from transcript CG1139-RA', other_designations='CG1139-PA', map_location='62B12-62B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1971638, end_position_on_the_genomic_accession=1974285, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38273, CurrentID=0, Status='live', Symbol='CG15822', Aliases='Dmel_ Dmel\\CG15822', description='CG15822 gene product from transcript CG15822-RE', other_designations='CG15822-PD|CG15822-PE|CG15822-PF|CG15822-PG|uncharacterized protein', map_location='62D1-62D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2153201, end_position_on_the_genomic_accession=2163863, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38274, CurrentID=0, Status='live', Symbol='CG15879', Aliases='Dmel_ Dmel\\CG15879', description='CG15879 gene product from transcript CG15879-RA', other_designations='CG15879-PA|uncharacterized protein', map_location='62D1-62D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2167550, end_position_on_the_genomic_accession=2169108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38277, CurrentID=0, Status='live', Symbol='CG15820', Aliases='Dmel_ BcDNA:SD08776, Dmel\\CG15820', description='CG15820 gene product from transcript CG15820-RA', other_designations='CG15820-PA|uncharacterized protein', map_location='62D1-62D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2172995, end_position_on_the_genomic_accession=2174332, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38278, CurrentID=0, Status='live', Symbol='CG13810', Aliases='Dmel_ Dmel\\CG13810', description='CG13810 gene product from transcript CG13810-RB', other_designations='CG13810-PB|uncharacterized protein', map_location='62D2-62D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2185930, end_position_on_the_genomic_accession=2187676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38280, CurrentID=0, Status='live', Symbol='CG8960', Aliases='Dmel_ Dmel\\CG8960', description='CG8960 gene product from transcript CG8960-RA', other_designations='CG8960-PA|CG8960-PB', map_location='62D2-62D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2197013, end_position_on_the_genomic_accession=2198089, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38281, CurrentID=0, Status='live', Symbol='CG15878', Aliases='Dmel_ Dmel\\CG15878', description='CG15878 gene product from transcript CG15878-RA', other_designations='CG15878-PA|uncharacterized protein', map_location='62D2-62D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2210873, end_position_on_the_genomic_accession=2212503, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38282, CurrentID=0, Status='live', Symbol='Oseg2', Aliases='Dmel_CG13809, CG13809, Dmel\\CG13809, IFT172, OSEG, OSEG2, OSM-1, oseg2, osm-1', description='Outer segment 2', other_designations='CG13809-PB|Oseg2-PB|osm-1|outer segment 2|selective LIM binding factor', map_location='62D3-62D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2229646, end_position_on_the_genomic_accession=2235253, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38283, CurrentID=0, Status='live', Symbol='CG9018', Aliases='Dmel_ Dmel\\CG9018, FBgn0035318', description='CG9018 gene product from transcript CG9018-RA', other_designations='CG9018-PA|CG9018-PB', map_location='62D3-62D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2235284, end_position_on_the_genomic_accession=2237076, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38284, CurrentID=0, Status='live', Symbol='ACXD', Aliases='Dmel_CG5712, AC, CG5712, D Dmel\\CG5712', description='CG5712 gene product from transcript CG5712-RB', other_designations='ACXD-PA|ACXD-PB|CG5712-PA|CG5712-PB', map_location='62D3-62D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2237608, end_position_on_the_genomic_accession=2243719, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  38285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38285, CurrentID=0, Status='live', Symbol='CG32301', Aliases='Dmel_ AC, CG8970, Dmel\\CG32301', description='CG32301 gene product from transcript CG32301-RB', other_designations='CG32301-PB|CG32301-PC|CG32301-PD|uncharacterized protein', map_location='62D3-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2244478, end_position_on_the_genomic_accession=2248614, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38286, CurrentID=0, Status='live', Symbol='CG1275', Aliases='Dmel_ Cyt-b561-1, Dmel\\CG1275, anon-WO0118547.260', description='CG1275 gene product from transcript CG1275-RD', other_designations='CG1275-PA|CG1275-PB|CG1275-PC|CG1275-PD|DcytB-like|cytochrome [[b561-1]]', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2252744, end_position_on_the_genomic_accession=2257423, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38287, CurrentID=0, Status='live', Symbol='CG2034', Aliases='Dmel_ Dmel\\CG2034, anon-i1', description='CG2034 gene product from transcript CG2034-RB', other_designations='CG2034-PA|CG2034-PB', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2257456, end_position_on_the_genomic_accession=2258922, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38288, CurrentID=0, Status='live', Symbol='oxt', Aliases='Dmel_CG32300, CG17771, CG17772, CG32300, Dmel\\CG32300, OXT, Oxt, anon-i2, dOXT, dXylT', description='peptide O-xylosyltransferase', other_designations='CG32300-PB|Peptide-O-xylosyl-transferase|Peptide-o-xylosyltransferase|drosophila peptide O-xylosyltransferase|oxt-PB', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2258839, end_position_on_the_genomic_accession=2262664, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38290, CurrentID=0, Status='live', Symbol='CG13807', Aliases='Dmel_ Dmel\\CG13807', description='CG13807 gene product from transcript CG13807-RA', other_designations='CG13807-PA|uncharacterized protein', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2262919, end_position_on_the_genomic_accession=2263615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38291, CurrentID=0, Status='live', Symbol='ecd', Aliases='Dmel_CG5714, CG5714, Dmel\\CG5714, Ed-1, ecd1, ecdl, l(3)62Db, l(3)ecd, l(3)ecd-1, l(3)ecd1, ecd', description='ecdysoneless', other_designations='CG5714-PA|ecd-PA|ecdysone|echinoid', map_location='3-1.3 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2263581, end_position_on_the_genomic_accession=2265857, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38292, CurrentID=0, Status='live', Symbol='CG13806', Aliases='Dmel_ Dmel\\CG13806', description='CG13806 gene product from transcript CG13806-RA', other_designations='CG13806-PA|uncharacterized protein', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2266394, end_position_on_the_genomic_accession=2267640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38293, CurrentID=0, Status='live', Symbol='CG32302', Aliases='Dmel_ CG13805, Dmel\\CG32302', description='CG32302 gene product from transcript CG32302-RA', other_designations='CG32302-PA|uncharacterized protein', map_location='62D4-62D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2268825, end_position_on_the_genomic_accession=2269903, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38294, CurrentID=0, Status='live', Symbol='yellow-g', Aliases='Dmel_CG5717, CG5717, Dmel\\CG5717, g, yell-g', description='CG5717 gene product from transcript CG5717-RA', other_designations='CG5717-PA|yellow-g-PA', map_location='62D5-62D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2269881, end_position_on_the_genomic_accession=2272769, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38295, CurrentID=0, Status='live', Symbol='yellow-g2', Aliases='Dmel_CG13804, CG13804, Dmel\\CG13804, g2, yel-g2, yg2', description='CG13804 gene product from transcript CG13804-RA', other_designations='CG13804-PA|yellow-g2-PA', map_location='62D5-62D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2276550, end_position_on_the_genomic_accession=2278038, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38296, CurrentID=0, Status='live', Symbol='MsR2', Aliases='Dmel_CG43745, CG13802, CG13803, CG43745, DMS-R2, DMSR-2, Dm CG13803, Dmel\\CG43745, Dmel_CG13802, Dmel_CG13803, DmsR, DmsR-2, anon-WO0170980.160, anon-WO0170980.161', description='Myosuppressin receptor 2', other_designations='CG43745-PA|CG43745-PB|CG43745-PC|DMS-receptor2|MsR2-PA|MsR2-PB|MsR2-PC|dromyosuppressin receptor 2|myosuppressin receptor 2', map_location='62D5-62D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2282883, end_position_on_the_genomic_accession=2300944, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38298, CurrentID=0, Status='live', Symbol='MsR1', Aliases='Dmel_CG8985, CG8985, DMS-R1, DMSR-1, Dm CG8985, Dmel\\CG8985, DmsR, DmsR-1, anon-WO0170980.106, anon-WO0170980.107', description='Myosuppressin receptor 1', other_designations='CG8985-PA|CG8985-PB|DMS-receptor1|MsR1-PA|MsR1-PB|dromyosuppressin receptor 1|myosuppressin receptor 1', map_location='62D6-62D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2322693, end_position_on_the_genomic_accession=2345799, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38299, CurrentID=0, Status='live', Symbol='CG13801', Aliases='Dmel_ Dmel\\CG13801', description='CG13801 gene product from transcript CG13801-RA', other_designations='CG13801-PA|uncharacterized protein', map_location='62D6-62D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2360557, end_position_on_the_genomic_accession=2362686, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38300, CurrentID=0, Status='live', Symbol='CG1317', Aliases='Dmel_ Dmel\\CG1317', description='CG1317 gene product from transcript CG1317-RD', other_designations='CG1317-PB|CG1317-PC|CG1317-PD', map_location='62D7-62D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2367843, end_position_on_the_genomic_accession=2371939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38301, CurrentID=0, Status='live', Symbol='CG8993', Aliases='Dmel_ 8993, Dmel\\CG8993', description='CG8993 gene product from transcript CG8993-RA', other_designations='CG8993-PA|thioredoxin', map_location='62D7-62D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2372199, end_position_on_the_genomic_accession=2372876, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38302, CurrentID=0, Status='live', Symbol='mRpL23', Aliases='Dmel_CG1320, BcDNA:RE49852, CG1320, Dmel\\CG1320, L23', description='mitochondrial ribosomal protein L23', other_designations='CG1320-PA|CG1320-PB|CG1320-PC|CG1320-PD|mRpL23-PA|mRpL23-PB|mRpL23-PC|mRpL23-PD', map_location='62D7-62D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2372844, end_position_on_the_genomic_accession=2373704, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38303, CurrentID=0, Status='live', Symbol='CG9004', Aliases='Dmel_ Dmel\\CG9004', description='CG9004 gene product from transcript CG9004-RA', other_designations='CG9004-PA', map_location='62D7-62D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2373728, end_position_on_the_genomic_accession=2376488, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38304, CurrentID=0, Status='live', Symbol='CG15877', Aliases='Dmel_ Dmel\\CG15877', description='CG15877 gene product from transcript CG15877-RA', other_designations='CG15877-PA|uncharacterized protein', map_location='62D7-62D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2376484, end_position_on_the_genomic_accession=2377556, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740784, CurrentID=0, Status='live', Symbol='CG11741', Aliases='Dmel_ Dmel\\CG11741', description='CG11741 gene product from transcript CG11741-RA', other_designations='CG11741-PA|CG11741-PB|uncharacterized protein', map_location='85A3-85A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8528423, end_position_on_the_genomic_accession=8528877, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38306, CurrentID=0, Status='live', Symbol='CG45186', Aliases='Dmel_ CG1141, CG12790, CG13797, CG13798, CG13799, CG13800, CG16763, CG16764, CG32306, CG33231, CG33232, CG42669, Dmel\\CG45186, Dmel_CG13800, Dmel_CG32306, Dmel_CG33231, Dmel_CG33232, Dmel_CG42669, c62E-11', description='CG45186 gene product from transcript CG45186-RT', other_designations='CG45186-PA|CG45186-PAA|CG45186-PAB|CG45186-PAC|CG45186-PAD|CG45186-PAE|CG45186-PAF|CG45186-PAG|CG45186-PAH|CG45186-PC|CG45186-PE|CG45186-PF|CG45186-PG|CG45186-PH|CG45186-PJ|CG45186-PK|CG45186-PL|CG45186-PM|CG45186-PN|CG45186-PO|CG45186-PP|CG45186-PQ|CG45186-PR|CG45186-PS|CG45186-PT|CG45186-PU|CG45186-PV|CG45186-PW|CG45186-PX|CG45186-PY|CG45186-PZ|uncharacterized protein|villin-like', map_location='62E2-62E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2377655, end_position_on_the_genomic_accession=2499226, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  38310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38310, CurrentID=0, Status='live', Symbol='CG16762', Aliases='Dmel_ Dmel\\CG16762', description='CG16762 gene product from transcript CG16762-RA', other_designations='CG16762-PA|uncharacterized protein', map_location='62E3-62E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2474667, end_position_on_the_genomic_accession=2475614, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38311, CurrentID=0, Status='live', Symbol='Cyp4d20', Aliases='Dmel_CG16761, 4d20, CG16761, Dmel\\CG16761, RH59895', description='CG16761 gene product from transcript CG16761-RA', other_designations='CG16761-PA|Cyp4d20-PA', map_location='62E3-62E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2476259, end_position_on_the_genomic_accession=2478448, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38313, CurrentID=0, Status='live', Symbol='CG1146', Aliases='Dmel_ Dmel\\CG1146', description='CG1146 gene product from transcript CG1146-RB', other_designations='CG1146-PA|CG1146-PB|CG1146-PC|CG1146-PD', map_location='62E3-62E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2488804, end_position_on_the_genomic_accession=2492613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38315, CurrentID=0, Status='live', Symbol='CG16758', Aliases='Dmel_ CG16760, Dmel\\CG16758, c62E-15', description='CG16758 gene product from transcript CG16758-RD', other_designations='CG16758-PB|CG16758-PD|CG16758-PE|CG16758-PF|CG16758-PG|CG16758-PH|uncharacterized protein', map_location='62E4-62E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2499372, end_position_on_the_genomic_accession=2503606, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38321, CurrentID=0, Status='live', Symbol='dos', Aliases='Dmel_CG1044, 141511_at, CG1044, DOS, Dmel\\CG1044, Dos, E(csw)3A, E(csw[CS])3A, EC3-4, Gab-1, Su(sev)3A, Su(sev[S11])3A', description='daughter of sevenless', other_designations='CG1044-PA|CG1044-PB|dos-PA|dos-PB|enhancer of corkscrew-CS|suppressor of sevenless-S11', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2589414, end_position_on_the_genomic_accession=2595590, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38322, CurrentID=0, Status='live', Symbol='CG16984', Aliases='Dmel_ BcDNA:AT01695, Dmel\\CG16984', description='CG16984 gene product from transcript CG16984-RA', other_designations='CG16984-PA|uncharacterized protein', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2595781, end_position_on_the_genomic_accession=2596998, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38323, CurrentID=0, Status='live', Symbol='CG16985', Aliases='Dmel_ BcDNA:RE30174, Dmel\\CG16985', description='CG16985 gene product from transcript CG16985-RA', other_designations='CG16985-PA|uncharacterized protein', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2599297, end_position_on_the_genomic_accession=2600242, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38324, CurrentID=0, Status='live', Symbol='CG12182', Aliases='Dmel_ BcDNA:GH02340, Dmel\\CG12182', description='CG12182 gene product from transcript CG12182-RC', other_designations='CG12182-PA|CG12182-PB|CG12182-PC|uncharacterized protein', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2595842, end_position_on_the_genomic_accession=2599579, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38325, CurrentID=0, Status='live', Symbol='CG16986', Aliases='Dmel_ Dmel\\CG16986', description='CG16986 gene product from transcript CG16986-RA', other_designations='CG16986-PA|CG16986-PB|uncharacterized protein', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2600331, end_position_on_the_genomic_accession=2601536, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38326, CurrentID=0, Status='live', Symbol='Pxn', Aliases='Dmel_CG12002, CG12002, CT1239, DmPxd-a, Dmel\\CG12002, Dpxn, pxn', description='Peroxidasin', other_designations='CG12002-PA|CG12002-PB|CG12002-PC|CG12002-PD|CG12002-PE|Pxn-PA|Pxn-PB|Pxn-PC|Pxn-PD|Pxn-PE|peroxidasin|peroxydasin|protein X', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2601455, end_position_on_the_genomic_accession=2630064, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38327, CurrentID=0, Status='live', Symbol='MEP-1', Aliases='Dmel_CG1244, CG1244, Dmel\\CG1244, MEP1, Q0E8J0_DROME, anon-EST:Liang-1.37, anon-WO0118547.584, clone 1.37, d', description='CG1244 gene product from transcript CG1244-RG', other_designations='CG1244-PA|CG1244-PB|CG1244-PC|CG1244-PD|CG1244-PE|CG1244-PF|CG1244-PG|MEP-1-PA|MEP-1-PB|MEP-1-PC|MEP-1-PD|MEP-1-PE|MEP-1-PF|MEP-1-PG', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2631715, end_position_on_the_genomic_accession=2641984, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  38328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38328, CurrentID=0, Status='live', Symbol='CG14949', Aliases='Dmel_ Dmel\\CG14949', description='CG14949 gene product from transcript CG14949-RA', other_designations='CG14949-PA|uncharacterized protein', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2642184, end_position_on_the_genomic_accession=2643424, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740788, CurrentID=0, Status='live', Symbol='CG34382', Aliases='Dmel_ CG14165, CG14166, Dmel\\CG34382, Dmel_CG14165, Dmel_CG14166', description='CG34382 gene product from transcript CG34382-RC', other_designations='CG34382-PC|CG34382-PD|CG34382-PE|CG34382-PF|uncharacterized protein', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9878985, end_position_on_the_genomic_accession=9884493, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38330, CurrentID=0, Status='live', Symbol='CG1246', Aliases='Dmel_ Dmel\\CG1246', description='CG1246 gene product from transcript CG1246-RC', other_designations='CG1246-PB|CG1246-PC|CG1246-PD|CG1246-PE', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2648048, end_position_on_the_genomic_accession=2650205, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38331, CurrentID=0, Status='live', Symbol='Aats-pro', Aliases='Dmel_CG12186, CG12186, Dmel\\CG12186, PRS, ProRS', description='Prolyl-tRNA synthetase', other_designations='Aats-pro-PA|Aats-pro-PB|CG12186-PA|CG12186-PB', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2650161, end_position_on_the_genomic_accession=2651850, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38332, CurrentID=0, Status='live', Symbol='CG42304', Aliases='Dmel_ CG1248, CG32297, Dmel\\CG42304, Dmel_CG32297', description='CG42304 gene product from transcript CG42304-RA', other_designations='CG42304-PA|CG42304-PB|uncharacterized protein', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2652088, end_position_on_the_genomic_accession=2655455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38337, CurrentID=0, Status='live', Symbol='Fife', Aliases='Dmel_CG43955, CG12187, CG14950, CG16976, CG43375, CG43955, Dmel\\CG43955, Dmel_CG12187, Dmel_CG14950, Dmel_CG16976, Dmel_CG43375, anon-EST:Posey171, fife', description='CG43955 gene product from transcript CG43955-RD', other_designations='CG43955-PB|CG43955-PD|CG43955-PF|CG43955-PG|Fife-PB|Fife-PD|Fife-PF|Fife-PG|fife', map_location='62E9-62F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2655415, end_position_on_the_genomic_accession=2711690, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  38338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38338, CurrentID=0, Status='live', Symbol='Mrtf', Aliases='Dmel_CG32296, CG12188, CG14951, CG32296, DMRTF, Dmel\\CG32296, MAL-D, MRTF, Mal-D, dMRTF, dMrtf, mal, mal-d', description='Myocardin-related transcription factor', other_designations='CG32296-PD|CG32296-PG|CG32296-PH|Mrtf-PD|Mrtf-PG|Mrtf-PH', map_location='62F2-62F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2717615, end_position_on_the_genomic_accession=2766536, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38340, CurrentID=0, Status='live', Symbol='Acp62F', Aliases='Dmel_CG1262, Acp62, Acp62f, CG1262, CT2528, Dmel\\CG1262', description='Accessory gland protein 62F', other_designations='Acp62F-PA|CG1262-PA|accessory gland peptide 62F|accessory gland protein 62F', map_location='62F2-62F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2739750, end_position_on_the_genomic_accession=2740228, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38341, CurrentID=0, Status='live', Symbol='Non2', Aliases='Dmel_CG1240, CG1240, Dmel\\CG1240', description='Novel nucleolar protein 2', other_designations='CG1240-PA|Non2-PA|novel nucleolar protein 2', map_location='62F3-62F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2766999, end_position_on_the_genomic_accession=2768392, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38342, CurrentID=0, Status='live', Symbol='CG9977', Aliases='Dmel_ Dmel\\CG9977', description='CG9977 gene product from transcript CG9977-RA', other_designations='CG9977-PA', map_location='62F3-62F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2768466, end_position_on_the_genomic_accession=2772325, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38343, CurrentID=0, Status='live', Symbol='CG12093', Aliases='Dmel_ Dmel\\CG12093', description='CG12093 gene product from transcript CG12093-RA', other_designations='CG12093-PA|uncharacterized protein', map_location='62F3-62F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2772560, end_position_on_the_genomic_accession=2774007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38344, CurrentID=0, Status='live', Symbol='Atg2', Aliases='Dmel_CG1241, ATG2, CG1241, Dmel\\CG1241, EP3697, atg2', description='Autophagy-related 2', other_designations='Atg2-PA|Atg2-PB|Autophagy-specific gene 2|CG1241-PA|CG1241-PB', map_location='62F3-62F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2774338, end_position_on_the_genomic_accession=2780677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38345, CurrentID=0, Status='live', Symbol='mRpS35', Aliases='Dmel_CG2101, CG2101, Dmel\\CG2101, MRP-S28', description='mitochondrial ribosomal protein S35', other_designations='CG2101-PA|mRpS35-PA', map_location='62F4-62F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2780899, end_position_on_the_genomic_accession=2782105, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38346, CurrentID=0, Status='live', Symbol='pgant6', Aliases='Dmel_CG2103, CG2103, Dmel\\CG2103, Pgant6', description='polypeptide GalNAc transferase 6', other_designations='CG2103-PA|CG2103-PB|CG2103-PC|pgant6-PA|pgant6-PB|pgant6-PC', map_location='62F4-62F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2782255, end_position_on_the_genomic_accession=2786010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38347, CurrentID=0, Status='live', Symbol='Tet', Aliases='Dmel_CG43444, CG2083, CG43444, CG9973, Dmel\\CG43444, Dmel_CG2083, Dmel_CG9973, Idax, dTet', description='Ten-Eleven Translocation (TET) family protein', other_designations='CG43444-PA|CG43444-PB|CG43444-PC|CG43444-PD|CG43444-PE|CG43444-PF|TET family protein|Ten-Eleven translocation (TET) family protein|Tet-PA|Tet-PB|Tet-PC|Tet-PD|Tet-PE|Tet-PF', map_location='62F6-63A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2786207, end_position_on_the_genomic_accession=2879158, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38350, CurrentID=0, Status='live', Symbol='spz5', Aliases='Dmel_CG9972, CG9972, CT28115, DNT2, Dmel\\CG9972, NT2, Spz5', description='spatzle 5', other_designations='CG9972-PA|spz5-PA', map_location='63A1-63A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2883641, end_position_on_the_genomic_accession=2892166, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38351, CurrentID=0, Status='live', Symbol='CG9970', Aliases='Dmel_ Dmel\\CG9970', description='CG9970 gene product from transcript CG9970-RA', other_designations='CG9970-PA', map_location='63A1-63A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2915024, end_position_on_the_genomic_accession=2916654, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38352, CurrentID=0, Status='live', Symbol='Shab', Aliases='Dmel_CG43128, CG1066, CG43128, CG9965, Dmel\\CG43128, Dmel_CG1066, Dmel_CG9965, Kv2, dShab, shab, shb', description='Shaker cognate b', other_designations='CG43128-PA|CG43128-PB|CG43128-PC|CG43128-PH|CG43128-PI|CG43128-PJ|CG43128-PK|CG43128-PL|CG43128-PM|CG43128-PN|Shab-PA|Shab-PB|Shab-PC|Shab-PH|Shab-PI|Shab-PJ|Shab-PK|Shab-PL|Shab-PM|Shab-PN|shaker|shaker cognate b', map_location='3-14 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2895017, end_position_on_the_genomic_accession=2959579, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  38354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38354, CurrentID=0, Status='live', Symbol='Or63a', Aliases='Dmel_CG9969, 63B.1, 63a, CG9969, DOR28, Dmel\\CG9969', description='Odorant receptor 63a', other_designations='CG9969-PA|CG9969-PB|Or63a-PA|Or63a-PB|odorant receptor 63a', map_location='63A2-63A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2961534, end_position_on_the_genomic_accession=2963788, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38355, CurrentID=0, Status='live', Symbol='CG2107', Aliases='Dmel_ Dmel\\CG2107, SD01848p', description='CG2107 gene product from transcript CG2107-RA', other_designations='CG2107-PA', map_location='63A2-63A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2981022, end_position_on_the_genomic_accession=2983551, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38356, CurrentID=0, Status='live', Symbol='CG2113', Aliases='Dmel_ BcDNA:AT18037, Dmel\\CG2113', description='CG2113 gene product from transcript CG2113-RA', other_designations='CG2113-PA|CG2113-PB', map_location='63A2-63A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2995172, end_position_on_the_genomic_accession=2996155, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38357, CurrentID=0, Status='live', Symbol='FMRFaR', Aliases='Dmel_CG2114, CG2114, Dm FMRFa-R, Dmel\\CG2114, DrmFMRFa-R, FMRFa-R, FR', description='FMRFamide Receptor', other_designations='CG2114-PA|CG2114-PB|CG2114-PC|FMRFa-receptor|FMRFaR-PA|FMRFaR-PB|FMRFaR-PC|FMRFamide receptor|dFMRFa receptor|fmrf receptor|larval-opioid-receptor', map_location='63A3-63A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3001716, end_position_on_the_genomic_accession=3011860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38358, CurrentID=0, Status='live', Symbol='CG32487', Aliases='Dmel_ 24656330, CG2077, Dmel\\CG32487, Phos32487', description='CG32487 gene product from transcript CG32487-RA', other_designations='CG32487-PA|uncharacterized protein', map_location='63A3-63A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3016150, end_position_on_the_genomic_accession=3017256, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740793, CurrentID=0, Status='live', Symbol='CG34312', Aliases='Dmel_ CG30058, Dmel\\CG34312, Dmel_CG30058', description='CG34312 gene product from transcript CG34312-RA', other_designations='CG34312-PA|CG34312-PB|CG34312-PC|uncharacterized protein', map_location='49F9-49F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13209809, end_position_on_the_genomic_accession=13212237, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38360, CurrentID=0, Status='live', Symbol='CG2162', Aliases='Dmel_ Dmel\\CG2162, anon-EST:Posey170', description='CG2162 gene product from transcript CG2162-RA', other_designations='CG2162-PA|CG2162-PE|CG2162-PF', map_location='63A3-63A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3035196, end_position_on_the_genomic_accession=3040469, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38362, CurrentID=0, Status='live', Symbol='scramb2', Aliases='Dmel_CG1893, CG1893, Dmel\\CG1893, Scramb2', description='scramblase 2', other_designations='CG1893-PA|scramb2-PA|scramblase', map_location='63A3-63A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3044113, end_position_on_the_genomic_accession=3046812, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38363, CurrentID=0, Status='live', Symbol='CG32485', Aliases='Dmel_ CG2159, Dmel\\CG32485', description='CG32485 gene product from transcript CG32485-RA', other_designations='CG32485-PA|uncharacterized protein', map_location='63A5-63A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3050323, end_position_on_the_genomic_accession=3051765, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38364, CurrentID=0, Status='live', Symbol='CG1271', Aliases='Dmel_ Dmel\\CG1271', description='CG1271 gene product from transcript CG1271-RC', other_designations='CG1271-PA|CG1271-PB|CG1271-PC|CG1271-PD', map_location='63A5-63A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3052514, end_position_on_the_genomic_accession=3054906, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38365, CurrentID=0, Status='live', Symbol='CG16753', Aliases='Dmel_ Dmel\\CG16753', description='CG16753 gene product from transcript CG16753-RA', other_designations='CG16753-PA|uncharacterized protein', map_location='63A5-63A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3055079, end_position_on_the_genomic_accession=3055842, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38367, CurrentID=0, Status='live', Symbol='CG32486', Aliases='Dmel_ CG14954, CG16745, Dmel\\CG32486, anon-WO0118547.267, l(2)SH3 095, l(3)SH095', description='CG32486 gene product from transcript CG32486-RD', other_designations='CG32486-PD|lethal (3) SH095|uncharacterized protein', map_location='63A5-63A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3055908, end_position_on_the_genomic_accession=3070839, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38369, CurrentID=0, Status='live', Symbol='CG11486', Aliases='Dmel_ Dmel\\CG11486, PAN3', description='CG11486 gene product from transcript CG11486-RM', other_designations='CG11486-PA|CG11486-PB|CG11486-PC|CG11486-PD|CG11486-PF|CG11486-PG|CG11486-PH|CG11486-PJ|CG11486-PK|CG11486-PL|CG11486-PM|CG11486-PN|CG11486-PO|CG11486-PP|CG11486-PQ|CG11486-PR|polyA-nuclease related|uncharacterized protein', map_location='63A6-63B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3071133, end_position_on_the_genomic_accession=3090406, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  38370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38370, CurrentID=0, Status='live', Symbol='Cht7', Aliases='Dmel_CG1869, CG1869, DmCHT7, Dmel\\CG1869', description='CG1869 gene product from transcript CG1869-RA', other_designations='CG1869-PA|Cht7-PA', map_location='63B1-63B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3090985, end_position_on_the_genomic_accession=3106958, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38371, CurrentID=0, Status='live', Symbol='CG45067', Aliases='Dmel_ CG14955, Dmel\\CG45067', description='CG45067 gene product from transcript CG45067-RA', other_designations='CG45067-PA|uncharacterized protein', map_location='63B1-63B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3111299, end_position_on_the_genomic_accession=3114248, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38372, CurrentID=0, Status='live', Symbol='prominin-like', Aliases='Dmel_CG7740, CG7740, Dmel\\CG7740, bs29e02.y1', description='CG7740 gene product from transcript CG7740-RD', other_designations='CG7740-PA|CG7740-PB|CG7740-PC|CG7740-PD|CG7740-PE|CG7740-PF|CG7740-PG|Prominin-like|prominin-like protein|prominin-like-PA|prominin-like-PB|prominin-like-PC|prominin-like-PD|prominin-like-PE|prominin-like-PF|prominin-like-PG', map_location='63B2-63B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3114450, end_position_on_the_genomic_accession=3127724, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  38373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38373, CurrentID=0, Status='live', Symbol='CG11537', Aliases='Dmel_ Dmel\\CG11537', description='CG11537 gene product from transcript CG11537-RB', other_designations='CG11537-PA|CG11537-PB|CG11537-PC|CG11537-PD|CG11537-PE|CG11537-PF|CG11537-PG|uncharacterized protein', map_location='63B3-63B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3128034, end_position_on_the_genomic_accession=3143457, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38374, CurrentID=0, Status='live', Symbol='CG1291', Aliases='Dmel_ Alg2, Dmel\\CG1291, dAlg2', description='CG1291 gene product from transcript CG1291-RA', other_designations='CG1291-PA', map_location='63B5-63B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3143808, end_position_on_the_genomic_accession=3145284, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38375, CurrentID=0, Status='live', Symbol='Usp5', Aliases='Dmel_CG12082, CG12082, DmUsp5, Dmel\\CG12082, Q9VZU7, USP5, leon, usp5', description='Ubiquitin specific protease 5', other_designations='CG12082-PA|CG12082-PB|Usp5-PA|Usp5-PB|isopeptidase T|ubiquitin specific protease 5', map_location='63B6-63B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3145405, end_position_on_the_genomic_accession=3148358, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38376, CurrentID=0, Status='live', Symbol='BtbVII', Aliases='Dmel_CG43365, BTB-VII, BTBVII, BcDNA:GH07729, CG11494, CG14956, CG43365, Dmel\\CG43365, Dmel_CG11494, Dmel_CG14956', description='BTB-protein-VII', other_designations='BTB protein VII|BtbVII-PA|BtbVII-PB|BtbVII-PD|BtbVII-PE|BtbVII-PF|BtbVII-PG|BtbVII-PH|BtbVII-PI|CG43365-PA|CG43365-PB|CG43365-PD|CG43365-PE|CG43365-PF|CG43365-PG|CG43365-PH|CG43365-PI', map_location='63B7-63B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3148646, end_position_on_the_genomic_accession=3158408, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38378, CurrentID=0, Status='live', Symbol='ND-30', Aliases='Dmel_CG12079, C-I-NDUFS3, CG 12079, CG12079, Dmel\\CG12079, NDUFS3, NUGM', description='NADH dehydrogenase (ubiquinone) 30 kDa subunit', other_designations='CG12079-PA|NADH dehydrogenase (ubiquinone) Fe-S protein 3|NADH ubiquinone oxidoreductase 30 kDa|NADH-UBIQUINONE OXIDOREDUCTASE 30 KDA SUBUNIT|NADH:ubiquinone oxidoreductase 30 kDa subunit|ND-30-PA|nuclear-encoded subunit NDUFS3 of complex I', map_location='63B7-63B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3158485, end_position_on_the_genomic_accession=3159578, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38379, CurrentID=0, Status='live', Symbol='CG15812', Aliases='Dmel_ Dmel\\CG15812, Pfk, pfk', description='CG15812 gene product from transcript CG15812-RA', other_designations='CG15812-PA|CG15812-PB|piefke|uncharacterized protein', map_location='63B7-63B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3159991, end_position_on_the_genomic_accession=3162019, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38381, CurrentID=0, Status='live', Symbol='Asciz', Aliases='Dmel_CG14962, CG14962, Dmel\\CG14962, dASCIZ', description='ASCIZ homolog', other_designations='Asciz-PA|Asciz-PB|CG14962-PA|CG14962-PB', map_location='63B8-63B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3163786, end_position_on_the_genomic_accession=3166250, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38382, CurrentID=0, Status='live', Symbol='CG32278', Aliases='Dmel_ CG12032, Dmel\\CG32278', description='CG32278 gene product from transcript CG32278-RA', other_designations='CG32278-PA|uncharacterized protein', map_location='63B8-63B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3169286, end_position_on_the_genomic_accession=3171767, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38383, CurrentID=0, Status='live', Symbol='CG14963', Aliases='Dmel_ Dmel\\CG14963', description='CG14963 gene product from transcript CG14963-RA', other_designations='CG14963-PA|uncharacterized protein', map_location='63B8-63B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3171614, end_position_on_the_genomic_accession=3172545, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38384, CurrentID=0, Status='live', Symbol='CG14964', Aliases='Dmel_ CT34810, Dmel\\CG14964', description='CG14964 gene product from transcript CG14964-RB', other_designations='CG14964-PB|uncharacterized protein', map_location='63B9-63B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3172817, end_position_on_the_genomic_accession=3178580, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38385, CurrentID=0, Status='live', Symbol='Girdin', Aliases='Dmel_CG12734, CG12734, Dmel\\CG12734', description='CG12734 gene product from transcript CG12734-RA', other_designations='CG12734-PA|CG12734-PB|CG12734-PC|CG12734-PD|CG12734-PE|Girdin-PA|Girdin-PB|Girdin-PC|Girdin-PD|Girdin-PE|dGirdin|girdin', map_location='63B9-63B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3178930, end_position_on_the_genomic_accession=3185287, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38386, CurrentID=0, Status='live', Symbol='CG14957', Aliases='Dmel_ Dmel\\CG14957', description='CG14957 gene product from transcript CG14957-RA', other_designations='CG14957-PA|uncharacterized protein', map_location='63B10-63B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3187563, end_position_on_the_genomic_accession=3188106, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38387, CurrentID=0, Status='live', Symbol='CG14958', Aliases='Dmel_ Dmel\\CG14958', description='CG14958 gene product from transcript CG14958-RA', other_designations='CG14958-PA|uncharacterized protein', map_location='63B10-63B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3189978, end_position_on_the_genomic_accession=3190623, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38388, CurrentID=0, Status='live', Symbol='CG14965', Aliases='Dmel_ 63BC-T3, Dmel\\CG14965, ORF1, anon-63BC-T3, dmLD47616p', description='CG14965 gene product from transcript CG14965-RA', other_designations='CG14965-PA|uncharacterized protein', map_location='63B10-63B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3190622, end_position_on_the_genomic_accession=3192491, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38389, CurrentID=0, Status='live', Symbol='Hsp83', Aliases='Dmel_CG1242, 143198_at, 83, 83K HSP, CG1242, DMHSP82, DmHsp83, Dmel\\CG1242, E(sev)3A, E(sina)2, HSP82, HSP83, HSP90, Hsp82, Hsp90, ORF1, Su(Raf)3A, anon-EST:Liang-2.53, anon-WO0068693, anon-WO0140519.209, clone 2.53, en(lz)3C/4C, hsp82, hsp83, hsp84, hsp90, l(3)j5C2, ms(3)08445, stc', description='Heat shock protein 83', other_designations='CG1242-PA|CG1242-PB|Hsp83-PA|Hsp83-PB|enhancer of seven in absentia 2|enhancer of sevenless 3A|heat shock protein 83|heat-shock protein 90 kDa|scratch', map_location='3-5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3192969, end_position_on_the_genomic_accession=3197059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38390, CurrentID=0, Status='live', Symbol='CG14966', Aliases='Dmel_ Dmel\\CG14966', description='CG14966 gene product from transcript CG14966-RA', other_designations='CG14966-PA|uncharacterized protein', map_location='63B11-63B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3197045, end_position_on_the_genomic_accession=3197631, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38391, CurrentID=0, Status='live', Symbol='gry', Aliases='Dmel_CG17569, CG17569, Dmel\\CG17569', description='gryzun', other_designations='CG17569-PB|gry-PB', map_location='63B11-63B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3199090, end_position_on_the_genomic_accession=3211348, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38392, CurrentID=0, Status='live', Symbol='CG32271', Aliases='Dmel_ CG11532, Dmel\\CG32271, SP176', description='CG32271 gene product from transcript CG32271-RA', other_designations='CG32271-PA|uncharacterized protein', map_location='63B11-63B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3202238, end_position_on_the_genomic_accession=3203162, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38393, CurrentID=0, Status='live', Symbol='CG32277', Aliases='Dmel_ CG11531, Dmel\\CG32277, SP183', description='CG32277 gene product from transcript CG32277-RA', other_designations='CG32277-PA|uncharacterized protein', map_location='63B12-63B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3204600, end_position_on_the_genomic_accession=3205449, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38395, CurrentID=0, Status='live', Symbol='CG14967', Aliases='Dmel_ Dmel\\CG14967', description='CG14967 gene product from transcript CG14967-RB', other_designations='CG14967-PB|CG14967-PC|uncharacterized protein', map_location='63B13-63B14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3210992, end_position_on_the_genomic_accession=3218920, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38396, CurrentID=0, Status='live', Symbol='CG12034', Aliases='Dmel_ Dmel\\CG12034, nSMase', description='CG12034 gene product from transcript CG12034-RA', other_designations='CG12034-PA|uncharacterized protein', map_location='63B14-63B14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3219156, end_position_on_the_genomic_accession=3220925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38397, CurrentID=0, Status='live', Symbol='RpL28', Aliases='Dmel_CG12740, BcDNA:GH04183, CG12740, Dmel\\CG12740, GH04183p, L28, anon-WO0118547.321, chr3L:3205394..3205484', description='Ribosomal protein L28', other_designations='60 S ribosomal protein L28|CG12740-PA|CG12740-PB|CG12740-PD|CG12740-PE|CG12740-PF|CG12740-PG|RpL28-PA|RpL28-PB|RpL28-PD|RpL28-PE|RpL28-PF|RpL28-PG|putativeribosomal L28e protein|ribosomal protein L28', map_location='63B14-63B14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3220911, end_position_on_the_genomic_accession=3222960, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38398, CurrentID=0, Status='live', Symbol='CG17737', Aliases='Dmel_ BcDNA:RE14985, Dmel\\CG17737, anon-EST:Posey272', description='CG17737 gene product from transcript CG17737-RA', other_designations='CG17737-PA|uncharacterized protein', map_location='63C1-63C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3223341, end_position_on_the_genomic_accession=3224650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38399, CurrentID=0, Status='live', Symbol='CG11505', Aliases='Dmel_ Dmel\\CG11505', description='CG11505 gene product from transcript CG11505-RD', other_designations='CG11505-PA|CG11505-PB|CG11505-PD|uncharacterized protein', map_location='63C1-63C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3224838, end_position_on_the_genomic_accession=3236604, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38400, CurrentID=0, Status='live', Symbol='CG17746', Aliases='Dmel_ Cg17746, Dmel\\CG17746, PP2C', description='CG17746 gene product from transcript CG17746-RB', other_designations='CG17746-PA|CG17746-PB|protein phosphatase 2C|uncharacterized protein', map_location='63C1-63C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3236805, end_position_on_the_genomic_accession=3239578, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38401, CurrentID=0, Status='live', Symbol='CG12078', Aliases='Dmel_ 24656659, Dmel\\CG12078', description='CG12078 gene product from transcript CG12078-RB', other_designations='CG12078-PA|CG12078-PB|uncharacterized protein', map_location='63C1-63C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3243274, end_position_on_the_genomic_accession=3244294, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38402, CurrentID=0, Status='live', Symbol='ckd', Aliases='Dmel_CG14959, BcDNA:RH52423, CG14959, CKD, Dmel\\CG14959, EP(2)3139, EP(3)0714, EP3139, EP714', description='cracked', other_designations='CG14959-PC|CG14959-PD|CG14959-PE|ckd-PC|ckd-PD|ckd-PE', map_location='63C2-63C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3257878, end_position_on_the_genomic_accession=3268184, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38403, CurrentID=0, Status='live', Symbol='CG14960', Aliases='Dmel_ Dmel\\CG14960', description='CG14960 gene product from transcript CG14960-RA', other_designations='CG14960-PA|uncharacterized protein', map_location='63C4-63C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3271036, end_position_on_the_genomic_accession=3276255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38404, CurrentID=0, Status='live', Symbol='CG12017', Aliases='Dmel_ Dmel\\CG12017', description='CG12017 gene product from transcript CG12017-RB', other_designations='CG12017-PA|CG12017-PB|CG12017-PC|CG12017-PD|uncharacterized protein', map_location='63C5-63C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3280239, end_position_on_the_genomic_accession=3290896, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38405, CurrentID=0, Status='live', Symbol='CG12009', Aliases='Dmel_ Dmel\\CG12009', description='CG12009 gene product from transcript CG12009-RB', other_designations='CG12009-PA|CG12009-PB|uncharacterized protein', map_location='63C5-63C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3291459, end_position_on_the_genomic_accession=3299000, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38406, CurrentID=0, Status='live', Symbol='CG14968', Aliases='Dmel_ Dmel\\CG14968, anon-WO0107627.1, anon-WO0107627.2, anon-WO0107627.3', description='CG14968 gene product from transcript CG14968-RB', other_designations='CG14968-PA|CG14968-PB|CG14968-PC|CG14968-PD|CG14968-PE|uncharacterized protein', map_location='63C5-63C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3299336, end_position_on_the_genomic_accession=3303112, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38407, CurrentID=0, Status='live', Symbol='ZnT63C', Aliases='Dmel_CG17723, CG17723, Dmel\\CG17723, ZnT1, cg17723, dZnT1, dZnT63C', description='Zinc transporter 63C', other_designations='CG17723-PA|CG17723-PB|CG17723-PC|CG17723-PD|CG17723-PE|CG17723-PF|CG17723-PG|ZnT63C-PA|ZnT63C-PB|ZnT63C-PC|ZnT63C-PD|ZnT63C-PE|ZnT63C-PF|ZnT63C-PG|zinc transporter 63C', map_location='63C6-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3303708, end_position_on_the_genomic_accession=3314009, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38408, CurrentID=0, Status='live', Symbol='Drsl2', Aliases='Dmel_CG32279, CG10813, CG32279, CT30310, Dmel\\CG32279, Dro-D, Dro2, Drs-lD, dmy2, dro-2, dro2', description='Drosomycin-like 2', other_designations='CG32279-PA|Drosomycin-like D|Drsl2-PA|drosomycin|drosomycin 2|drosomycin-2|drosomycin-D', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3314349, end_position_on_the_genomic_accession=3314681, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38409, CurrentID=0, Status='live', Symbol='Drsl5', Aliases='Dmel_CG10812, BcDNA:GH09576, CG10812, Dmel\\CG10812, Dro-G, Dro5, Drs-lG, dmy5, dro5', description='Drosomycin-like 5', other_designations='CG10812-PA|Drosomycin-like G|Drsl5-PA|dro5|drosmycin-5|drosomycin|drosomycin 5|drosomycin B|drosomycin-5|drosomycin-G', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3316781, end_position_on_the_genomic_accession=3317144, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38410, CurrentID=0, Status='live', Symbol='PIG-C', Aliases='Dmel_CG12077, CG12077, Dmel\\CG12077', description='Phosphatidylinositol glycan anchor biosynthesis, class C ortholog (H. sapiens)', other_designations='CG12077-PB|PIG-C-PB|phosphatidylinositol glycan anchor biosynthesis, class C ortholog', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3318028, end_position_on_the_genomic_accession=3319232, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38411, CurrentID=0, Status='live', Symbol='CG12016', Aliases='Dmel_ Dmel\\CG12016, Dromel_CG12016_FBtr0073063_mORF', description='CG12016 gene product from transcript CG12016-RA', other_designations='CG12016-PA|CG12016-PB|uncharacterized protein', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3319555, end_position_on_the_genomic_accession=3322489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38412, CurrentID=0, Status='live', Symbol='Strip', Aliases='Dmel_CG11526, CG11526, Dmel\\CG11526, strip', description='Striatin interacting protein', other_designations='CG11526-PA|CG11526-PB|CG11526-PC|Strip-PA|Strip-PB|Strip-PC|striatin interacting protein|striatin interacting protein ortholog', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3322025, end_position_on_the_genomic_accession=3328871, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38413, CurrentID=0, Status='live', Symbol='PHGPx', Aliases='Dmel_CG12013, CG12013, DmGPx, Dm Dmel\\CG12013, GPx, GTPx, GTPx-1, Gtpx-1, PGTPx, anon-WO0257455.17, phGpx', description='CG12013 gene product from transcript CG12013-RC', other_designations='CG12013-PA|CG12013-PC|CG12013-PD|PHGPx-PA|PHGPx-PC|PHGPx-PD|glutathione peroxidase|glutathione peroxidase I', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3326265, end_position_on_the_genomic_accession=3328626, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38414, CurrentID=0, Status='live', Symbol='CG14961', Aliases='Dmel_ Dmel\\CG14961', description='CG14961 gene product from transcript CG14961-RA', other_designations='CG14961-PA|uncharacterized protein', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3330063, end_position_on_the_genomic_accession=3333181, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38415, CurrentID=0, Status='live', Symbol='CG14969', Aliases='Dmel_ Dmel\\CG14969', description='CG14969 gene product from transcript CG14969-RA', other_designations='CG14969-PA|uncharacterized protein', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3333119, end_position_on_the_genomic_accession=3334978, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38416, CurrentID=0, Status='live', Symbol='Drsl6', Aliases='Dmel_CG32268, CG10815, CG11520, CG32268, Dmel\\CG32268, Dro-I, Dro6, Drs-lI, dro6', description='Drosomycin-like 6', other_designations='CG32268-PA|Drosomycin-6|Drosomycin-like I|Drsl6-PA|drosomycin|drosomycin 6|drosomycin-I', map_location='63D1-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3336032, end_position_on_the_genomic_accession=3336418, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38418, CurrentID=0, Status='live', Symbol='kst', Aliases='Dmel_CG12008, CG12008, Dmel\\CG12008, Kst, Spec-beta[[H]], beta-Spectrin, betaH, betaH-Spec, betaHS, beta[[H]], beta[[H]]-SPEC, beta[[H]]-Sp, beta[[H]]-Spec, beta[[H]]-Spectrin, beta[[H]]spectrin, beta[[Heavy]]spectrin, beta[[heavy]]-Spec, beta[[heavy]]-Spectrin, beta[[heavy]]spectrin, betah-Spec, l(3)01318', description='karst', other_designations='CG12008-PA|CG12008-PB|CG12008-PC|CG12008-PE|CG12008-PF|CG12008-PG|CG12008-PH|Spectrin-beta[[H]]|b[Heavy]-spectrin|beta-H spectrin|beta-Heavy spectrin|beta-heavy-spectrin|betaH-Spectrin|betaHeavy--spectrin|betaHeavy-Spectrin|beta[[(HEAVY)]]-spectrin|beta[[H]] spectrin|beta[[H]]-spectrin|beta[[Heavy]]-spectrin|kst-PA|kst-PB|kst-PC|kst-PE|kst-PF|kst-PG|kst-PH|spectrin|spectrin beta-heavy chain', map_location='63D2-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3336860, end_position_on_the_genomic_accession=3368133, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  38419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38419, CurrentID=0, Status='live', Symbol='Drs', Aliases='Dmel_CG10810, BcDNA:LP03851, CG10810, Crp, DIM 19, DIM 21, DRO, DROM, DROS, DRS, Dmel\\CG10810, Drm, Drom, Dros, Droso, IM21, dmy1, drm, drom, drs, drsm', description='Drosomycin', other_designations='CG10810-PA|Cysteine-rich-protein|Drosomycin-B|Drs-PA|cysteine-rich peptide|drosomcycin|drosomyci|drosomycin|drosomycin precursor|drosomycine|drosomyocin|drosomysin|drosophila immune induced molecule 19|immune induced molecule 21', map_location='63D2-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3369556, end_position_on_the_genomic_accession=3369942, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38420, CurrentID=0, Status='live', Symbol='YT521-B', Aliases='Dmel_CG12076, BcDNA:GH01918, CG12076, Dmel\\CG12076, YT521, d', description='CG12076 gene product from transcript CG12076-RA', other_designations='CG12076-PA|CG12076-PB|YT521-B-PA|YT521-B-PB', map_location='63D2-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3370451, end_position_on_the_genomic_accession=3374170, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38421, CurrentID=0, Status='live', Symbol='CG12010', Aliases='Dmel_ Dmel\\CG12010', description='CG12010 gene product from transcript CG12010-RB', other_designations='CG12010-PB|uncharacterized protein', map_location='63D2-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3374282, end_position_on_the_genomic_accession=3376649, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38422, CurrentID=0, Status='live', Symbol='CG12012', Aliases='Dmel_ Dmel\\CG12012', description='CG12012 gene product from transcript CG12012-RC', other_designations='CG12012-PA|CG12012-PB|CG12012-PC|uncharacterized protein', map_location='63D2-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3376757, end_position_on_the_genomic_accession=3377992, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38423, CurrentID=0, Status='live', Symbol='CG12014', Aliases='Dmel_ Dmel\\CG12014, anon-EST:Posey282', description='CG12014 gene product from transcript CG12014-RA', other_designations='CG12014-PA|uncharacterized protein', map_location='63D2-63D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3378268, end_position_on_the_genomic_accession=3380083, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38424, CurrentID=0, Status='live', Symbol='sty', Aliases='Dmel_CG1921, CG1921, CS3-3, Dmel\\CG1921, STY, Spry, Sry, Sty, dSPRY, spry', description='sprouty', other_designations='CG1921-PB|CG1921-PC|calcineurin suppressor, chromosome 3, group 3|dSprouty|sty-PB|sty-PC', map_location='63D2-63D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3401153, end_position_on_the_genomic_accession=3424935, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38427, CurrentID=0, Status='live', Symbol='armi', Aliases='Dmel_CG11513, APC16, ARMI, Armi, CG11513, Dmel\\CG11513', description='armitage', other_designations='CG11513-PB|CG11513-PC|CG11513-PD|armi-PB|armi-PC|armi-PD', map_location='63E1-63E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3461317, end_position_on_the_genomic_accession=3466303, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38428, CurrentID=0, Status='live', Symbol='CycJ', Aliases='Dmel_CG10308, CDI5, CG10308, Cdi5, Dmel\\CG10308, cdi5, cycJ, cyclinJ', description='Cyclin J', other_designations='CG10308-PA|CG10308-PB|CycJ-PA|CycJ-PB|Cyclin-dependent kinase interactor 5|cyclin J', map_location='63E1-63E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3466428, end_position_on_the_genomic_accession=3468098, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38429, CurrentID=0, Status='live', Symbol='CG14971', Aliases='Dmel_ Dmel\\CG14971', description='CG14971 gene product from transcript CG14971-RA', other_designations='CG14971-PA|uncharacterized protein', map_location='63E1-63E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3468023, end_position_on_the_genomic_accession=3471309, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38430, CurrentID=0, Status='live', Symbol='CG42324', Aliases='Dmel_ CG14972, CG14973, Dmel\\CG42324, Dmel_CG14972, Dmel_CG14973', description='CG42324 gene product from transcript CG42324-RK', other_designations='CG42324-PF|CG42324-PG|CG42324-PH|CG42324-PI|CG42324-PJ|CG42324-PK|uncharacterized protein', map_location='63E2-63E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3472447, end_position_on_the_genomic_accession=3510895, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740805, CurrentID=0, Status='live', Symbol='CG34208', Aliases='Dmel_ Dmel\\CG34208', description='CG34208 gene product from transcript CG34208-RA', other_designations='CG34208-PA|uncharacterized protein', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22333126, end_position_on_the_genomic_accession=22333505, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38432, CurrentID=0, Status='live', Symbol='ImpE2', Aliases='Dmel_CG1934, CG1934, Dmel\\CG1934, IMP-E2, Imp-E2', description='Ecdysone-inducible gene E2', other_designations='CG1934-PA|ImpE2-PA', map_location='63E2-63E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3510974, end_position_on_the_genomic_accession=3512681, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38433, CurrentID=0, Status='live', Symbol='Eip63E', Aliases='Dmel_CG10579, CG10579, Dmel\\CG10579, EIP63E, L63, PFTAIRE, cdc2-63E', description='Ecdysone-induced protein 63E', other_designations='CG10579-PA|CG10579-PB|CG10579-PC|CG10579-PD|CG10579-PE|CG10579-PF|CG10579-PG|CG10579-PH|CG10579-PI|CG10579-PJ|CG10579-PK|CG10579-PL|CG10579-PM|CG10579-PN|Eip63E-PA|Eip63E-PB|Eip63E-PC|Eip63E-PD|Eip63E-PE|Eip63E-PF|Eip63E-PG|Eip63E-PH|Eip63E-PI|Eip63E-PJ|Eip63E-PK|Eip63E-PL|Eip63E-PM|Eip63E-PN|PFTAIRE', map_location='63E3-63E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3515751, end_position_on_the_genomic_accession=3610521, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  38434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38434, CurrentID=0, Status='live', Symbol='CG10359', Aliases='Dmel_ AAF47782, Dmel\\CG10359, FBpp0072995', description='CG10359 gene product from transcript CG10359-RE', other_designations='CG10359-PE|CG10359-PG|Fibrinogen-related protein|Fibronogen-like|uncharacterized protein', map_location='63E5-63E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3612192, end_position_on_the_genomic_accession=3615244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38435, CurrentID=0, Status='live', Symbol='CG10357', Aliases='Dmel_ Dmel\\CG10357', description='CG10357 gene product from transcript CG10357-RC', other_designations='CG10357-PC|uncharacterized protein', map_location='63E5-63E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3615382, end_position_on_the_genomic_accession=3616682, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38436, CurrentID=0, Status='live', Symbol='dar1', Aliases='Dmel_CG12029, CG12029, Dar1, Dmel\\CG12029, l(3)63Ed, l(3)SH5', description='dendritic arbor reduction 1', other_designations='CG12029-PB|dar1-PB|lethal (3) 63Ed', map_location='63E5-63F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3626352, end_position_on_the_genomic_accession=3653442, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38437, CurrentID=0, Status='live', Symbol='CG10862', Aliases='Dmel_ Dmel\\CG10862', description='CG10862 gene product from transcript CG10862-RA', other_designations='CG10862-PA|uncharacterized protein', map_location='63E6-63E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3653405, end_position_on_the_genomic_accession=3654771, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38438, CurrentID=0, Status='live', Symbol='CG32264', Aliases='Dmel_ CG10857, CG10860, CG12604, CG14978, Dmel\\CG32264', description='CG32264 gene product from transcript CG32264-RK', other_designations='CG32264-PB|CG32264-PD|CG32264-PE|CG32264-PG|CG32264-PI|CG32264-PJ|CG32264-PK|CG32264-PL|CG32264-PM|uncharacterized protein', map_location='63E8-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3714826, end_position_on_the_genomic_accession=3806459, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  38440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38440, CurrentID=0, Status='live', Symbol='ppk27', Aliases='Dmel_CG10858, CG10858, Dmel\\CG10858', description='pickpocket 27', other_designations='CG10858-PB|ppk27-PB', map_location='63E9-63E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3753693, end_position_on_the_genomic_accession=3755201, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38443, CurrentID=0, Status='live', Symbol='ntc', Aliases='Dmel_CG10855, CG10855, Dmel\\CG10855, FBXO7', description='nutcracker', other_designations='CG10855-PA|ntc-PA', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3803832, end_position_on_the_genomic_accession=3806459, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38444, CurrentID=0, Status='live', Symbol='IntS10', Aliases='Dmel_CG1120, CG1120, Dmel\\CG1120, Int10', description='Integrator 10', other_designations='CG1120-PA|IntS10-PA|integrator 10', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3806710, end_position_on_the_genomic_accession=3809090, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38445, CurrentID=0, Status='live', Symbol='CG32262', Aliases='Dmel_ CG10854, Dmel\\CG32262', description='CG32262 gene product from transcript CG32262-RA', other_designations='CG32262-PA|uncharacterized protein', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3809146, end_position_on_the_genomic_accession=3810196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38446, CurrentID=0, Status='live', Symbol='CG12006', Aliases='Dmel_ Dmel\\CG12006', description='CG12006 gene product from transcript CG12006-RA', other_designations='CG12006-PA|CG12006-PB|uncharacterized protein', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3811263, end_position_on_the_genomic_accession=3813958, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38447, CurrentID=0, Status='live', Symbol='Scsalpha', Aliases='Dmel_CG1065, CG1065, Dmel\\CG1065, Scs-alpha, scsalpha', description='Succinyl coenzyme A synthetase alpha subunit', other_designations='CG1065-PA|CG1065-PB|Scsalpha-PA|Scsalpha-PB|succinyl coenzyme A synthatase, alpha subunit|succinyl coenzyme A synthetase alpha subunit|succinyl-CoA synthetase alpha-subunit', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3814211, end_position_on_the_genomic_accession=3818383, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38448, CurrentID=0, Status='live', Symbol='Acp63F', Aliases='Dmel_CG10852, Acp63/64A, Acp63f, CG10852, Dmel\\CG10852, acp63F, Acp63F', description='Accessory gland protein 63F', other_designations='Acp63F-PA|Acp63F-PB|CG10852-PA|CG10852-PB|accessory gland peptide 63F|accessory gland protein 63F', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3815839, end_position_on_the_genomic_accession=3816354, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38451, CurrentID=0, Status='live', Symbol='Awh', Aliases='Dmel_CG1072, CG1072, Dmel\\CG1072, awh, l(3)63Ea, l(3)63Ea/Awh, l(3)64Ea, l(3)SH9, l63Ea', description='Arrowhead', other_designations='Awh-PA|Awh-PB|Awh-PC|Awh-PD|CG1072-PA|CG1072-PB|CG1072-PC|CG1072-PD|arrowhead|lethal(3)63Ea', map_location='63F4-63F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3846206, end_position_on_the_genomic_accession=3861692, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38452, CurrentID=0, Status='live', Symbol='Fie', Aliases='Dmel_CG1079, BEST:HL04053, CG1079, Dmel\\CG1079, HL04053', description='Fire exit', other_designations='CG1079-PA|CG1079-PB|Fie-PA|Fie-PB|fire exit', map_location='63F5-63F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3892574, end_position_on_the_genomic_accession=3894833, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38453, CurrentID=0, Status='live', Symbol='Gr63a', Aliases='Dmel_CG14979, CG14979, DmGr63a, Dmel\\CG14979, GR63a, Gr63F1, gr63a', description='Gustatory receptor 63a', other_designations='CG14979-PB|Gr63a-PB|gustatory receptor 63F1|gustatory receptor 63a', map_location='63F5-63F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3894999, end_position_on_the_genomic_accession=3896603, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38454, CurrentID=0, Status='live', Symbol='CG14977', Aliases='Dmel_ BcDNA:RH34338, Dmel\\CG14977, dC7orf59', description='CG14977 gene product from transcript CG14977-RA', other_designations='CG14977-PA|CG14977-PB|uncharacterized protein', map_location='63F5-63F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3896900, end_position_on_the_genomic_accession=3897498, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38455, CurrentID=0, Status='live', Symbol='Ccz1', Aliases='Dmel_CG14980, C7orf28B, CG14980, Dccz1, Dmel\\CG14980, NP_647835', description='Ccz1 ortholog', other_designations='CG14980-PB|Ccz1-PB', map_location='63F5-63F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3897444, end_position_on_the_genomic_accession=3899156, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38456, CurrentID=0, Status='live', Symbol='Ubi-p63E', Aliases='Dmel_CG11624, CG11624, DmUb, DmUbi-p63E, Dmel\\CG11624, UB, Ub, Ubi-P63e, Ubi-p, Ubi63E, Ubi63p, Ubip, Ubip63E, Ubiq, Ubp, magn, ubi, ubi-p63E, ubi63E', description='Ubiquitin-63E', other_designations='CG11624-PA|CG11624-PB|CG11624-PC|CG11624-PD|Ubi-p63E-PA|Ubi-p63E-PB|Ubi-p63E-PC|Ubi-p63E-PD|Ubiquitin-63-E|Ubiquitin-polymeric|magellan|polyubiquitin|ubiquitin|ubiquitin 63E', map_location='63F5-63F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3899259, end_position_on_the_genomic_accession=3903184, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38457, CurrentID=0, Status='live', Symbol='Sc2', Aliases='Dmel_CG10849, CG10849, Dmel\\CG10849, d l(3)05634, l(3)63Eb, l(3)SH4', description='CG10849 gene product from transcript CG10849-RA', other_designations='CG10849-PA|Sc2-PA|Ubiquitin-63E|lethal(3)63Eb', map_location='63F5-63F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3903492, end_position_on_the_genomic_accession=3905035, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38459, CurrentID=0, Status='live', Symbol='mge', Aliases='Dmel_CG14981, BcDNA:LP07226, CG14981, Dmel\\CG14981, l(3)63Fc, l(3)SH3', description='maggie', other_designations='CG14981-PA|CG14981-PB|CG14981-PC|mge-PA|mge-PB|mge-PC', map_location='63F6-63F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3907852, end_position_on_the_genomic_accession=3909931, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38461, CurrentID=0, Status='live', Symbol='CG10866', Aliases='Dmel_ Dmel\\CG10866', description='CG10866 gene product from transcript CG10866-RB', other_designations='CG10866-PA|CG10866-PB|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3936060, end_position_on_the_genomic_accession=3937992, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38462, CurrentID=0, Status='live', Symbol='CG12766', Aliases='Dmel_ Dmel\\CG12766', description='CG12766 gene product from transcript CG12766-RA', other_designations='CG12766-PA|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3938624, end_position_on_the_genomic_accession=3940125, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38463, CurrentID=0, Status='live', Symbol='CG10863', Aliases='Dmel_ BcDNA:GH10614, Dmel\\CG10863, dm1/AKR2E3', description='CG10863 gene product from transcript CG10863-RA', other_designations='CG10863-PA|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3941008, end_position_on_the_genomic_accession=3942465, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38464, CurrentID=0, Status='live', Symbol='CG14982', Aliases='Dmel_ Dmel\\CG14982', description='CG14982 gene product from transcript CG14982-RC', other_designations='CG14982-PB|CG14982-PC|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3942710, end_position_on_the_genomic_accession=3947278, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38465, CurrentID=0, Status='live', Symbol='CG10853', Aliases='Dmel_ BcDNA:LP09837, Dmel\\CG10853', description='CG10853 gene product from transcript CG10853-RA', other_designations='CG10853-PA|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3947891, end_position_on_the_genomic_accession=3948687, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38466, CurrentID=0, Status='live', Symbol='CG14983', Aliases='Dmel_ Dmel\\CG14983', description='CG14983 gene product from transcript CG14983-RA', other_designations='CG14983-PA|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3948699, end_position_on_the_genomic_accession=3949305, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740811, CurrentID=0, Status='live', Symbol='CG34231', Aliases='Dmel_ Dmel\\CG34231', description='CG34231 gene product from transcript CG34231-RC', other_designations='CG34231-PC|uncharacterized protein', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11986008, end_position_on_the_genomic_accession=11987653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38468, CurrentID=0, Status='live', Symbol='CG12605', Aliases='Dmel_ Dmel\\CG12605, anon-WO0140519.93', description='CG12605 gene product from transcript CG12605-RD', other_designations='CG12605-PA|CG12605-PB|CG12605-PC|CG12605-PD|CG12605-PE|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3954251, end_position_on_the_genomic_accession=3964771, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38469, CurrentID=0, Status='live', Symbol='scrt', Aliases='Dmel_CG1130, BcDNA:RH71479, CG1130, Dmel\\CG1130, Scrt, anon-EST:Liang-1.31, anon-EST:Liang-1.67, anon-WO02059370.57, clone 1.31, clone 1.67, lincRNA.S4061', description='scratch', other_designations='CG1130-PA|CG1130-PB|scrt-PA|scrt-PB', map_location='64A2-64A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3983896, end_position_on_the_genomic_accession=3994003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38470, CurrentID=0, Status='live', Symbol='CG14985', Aliases='Dmel_ Dmel\\CG14985', description='CG14985 gene product from transcript CG14985-RB', other_designations='CG14985-PB|CG14985-PC|uncharacterized protein', map_location='64A3-64A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4003474, end_position_on_the_genomic_accession=4006626, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38471, CurrentID=0, Status='live', Symbol='fd64A', Aliases='Dmel_CG1132, CG1132, Dmel\\CG1132, Dmfd2, FD2, FoxL1, fd2', description='forkhead domain 64A', other_designations='CG1132-PA|CG1132-PB|fd64A-PA|fd64A-PB', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4018878, end_position_on_the_genomic_accession=4021279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38472, CurrentID=0, Status='live', Symbol='Mul1', Aliases='Dmel_CG1134, CG1134, Dmel\\CG1134, MUL1', description='Mitochondrial E3 ubiquitin protein ligase 1 homologue', other_designations='CG1134-PA|CG1134-PB|Mul1-PA|Mul1-PB|mitochondrial E3 ubiquitin protein ligase 1 homologue', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4021457, end_position_on_the_genomic_accession=4023501, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38473, CurrentID=0, Status='live', Symbol='CG11594', Aliases='Dmel_ Dmel\\CG11594', description='CG11594 gene product from transcript CG11594-RA', other_designations='CG11594-PA|CG11594-PC|CG11594-PD|CG11594-PE|uncharacterized protein', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4022700, end_position_on_the_genomic_accession=4026011, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5740108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740108, CurrentID=0, Status='live', Symbol='CG34188', Aliases='Dmel_ Dmel\\CG34188', description='CG34188 gene product from transcript CG34188-RA', other_designations='CG34188-PA|uncharacterized protein', map_location='52A1-52A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15427091, end_position_on_the_genomic_accession=15428023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38477, CurrentID=0, Status='live', Symbol='CG11593', Aliases='Dmel_ Dmel\\CG11593, anon-EST:Posey64', description='CG11593 gene product from transcript CG11593-RB', other_designations='CG11593-PB|CG11593-PC|uncharacterized protein', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4037523, end_position_on_the_genomic_accession=4040482, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38478, CurrentID=0, Status='live', Symbol='Rcd5', Aliases='Dmel_CG1135, CG1135, Dmel\\CG1135, MCRS1, MCRS2, dMCRS1, dMCRS2, l(3)r0166, l(3)rG166, p78', description='Reduction in Cnn dots 5', other_designations='CG1135-PA|Rcd5-PA|lethal (3) rG166|reduction in Cnn dots 5', map_location='64A4-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4040529, end_position_on_the_genomic_accession=4042597, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38479, CurrentID=0, Status='live', Symbol='CG1136', Aliases='Dmel_ Dmel\\CG1136, novel', description='CG1136 gene product from transcript CG1136-RA', other_designations='CG1136-PA', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4046596, end_position_on_the_genomic_accession=4059660, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38480, CurrentID=0, Status='live', Symbol='Dpy-30L2', Aliases='Dmel_CG11591, CG11591, Dmel\\CG11591, dpy-30L2', description='Dpy-30-like 2', other_designations='CG11591-PA|CG11591-PB|Dpy-30L2-PA|Dpy-30L2-PB', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4046988, end_position_on_the_genomic_accession=4047735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740113, CurrentID=0, Status='live', Symbol='Pp1-Y1', Aliases='Dmel_CG41534, CG41534, DmPp1-Y1, Dmel\\CG41534, Pp1Y1', description='Protein phosphatase 1, Y-linked 1', other_designations='CG41534-PC|Pp1-Y1-PC|protein phosphatase 1, Y-linked 1', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=9833, end_position_on_the_genomic_accession=10786, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740115, CurrentID=0, Status='live', Symbol='CG34250', Aliases='Dmel_ Dmel\\CG34250', description='CG34250 gene product from transcript CG34250-RB', other_designations='CG34250-PA|CG34250-PB|uncharacterized protein', map_location='74C3-74C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17476222, end_position_on_the_genomic_accession=17477103, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38484, CurrentID=0, Status='live', Symbol='Gad1', Aliases='Dmel_CG14994, CG14994, DGAD1, DGad1, Dmel\\CG14994, GAD, GAD-1, GAD1, GAD[[1]], Gad, Glb, dGad1, dgad1, gad, gad1, l(3)64An, l(3)SH10', description='Glutamic acid decarboxylase 1', other_designations='CG14994-PA|CG14994-PB|CG14994-PC|CG14994-PD|GABA synthetase|Gad1-PA|Gad1-PB|Gad1-PC|Gad1-PD|glutamatic acid decarboxylase|glutamic acid decarboxylase|glutamic acid decarboxylase 1|glutamic acid decarboxylase-1|vesicular glutamic acid decarboxylase', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4074077, end_position_on_the_genomic_accession=4088744, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38485, CurrentID=0, Status='live', Symbol='CG14989', Aliases='Dmel_ BcDNA:GH06302, Dmel\\CG14989', description='CG14989 gene product from transcript CG14989-RE', other_designations='CG14989-PC|CG14989-PD|CG14989-PE|CG14989-PF|uncharacterized protein', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4087581, end_position_on_the_genomic_accession=4092706, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38486, CurrentID=0, Status='live', Symbol='CG14990', Aliases='Dmel_ Dmel\\CG14990, SPH97', description='CG14990 gene product from transcript CG14990-RA', other_designations='CG14990-PA|uncharacterized protein', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4093207, end_position_on_the_genomic_accession=4094344, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38487, CurrentID=0, Status='live', Symbol='CG14995', Aliases='Dmel_ CT34848, Dmel\\CG14995, NEST:bs15f01, NEST:bs17f06, anon-WO0140519.252, bs17f06.y1', description='CG14995 gene product from transcript CG14995-RA', other_designations='CG14995-PA|CG14995-PC|CG14995-PD|CG14995-PF|CG14995-PG|CG14995-PH|CG14995-PI|gandalf|uncharacterized protein', map_location='64A5-64A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4094343, end_position_on_the_genomic_accession=4103256, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38488, CurrentID=0, Status='live', Symbol='Fit1', Aliases='Dmel_CG14991, CG14991, Dmel\\CG14991, anon-WO0118547.421, fit 1', description='Fermitin 1', other_designations='CG14991-PA|CG14991-PB|Fit1-PA|Fit1-PB|fermitin 1', map_location='64A6-64A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4103564, end_position_on_the_genomic_accession=4107116, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38489, CurrentID=0, Status='live', Symbol='Ack', Aliases='Dmel_CG14992, ACK, ACK2, BcDNA:GH10777, CG14992, DACK, DAck, Dmel\\CG14992, p145', description='Activated Cdc42 kinase', other_designations='Ack-PA|CG14992-PA|activated Cdc42 kinase', map_location='64A6-64A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4108186, end_position_on_the_genomic_accession=4114157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38490, CurrentID=0, Status='live', Symbol='Chd64', Aliases='Dmel_CG14996, CG-14996, CG14996, CT34849, Dmel\\CG14996, anon-EST:Liang-1.80, clone 1.80', description='CG14996 gene product from transcript CG14996-RB', other_designations='CG14996-PB|CG14996-PC|Chd64-PB|Chd64-PC', map_location='64A6-64A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4114289, end_position_on_the_genomic_accession=4122388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38491, CurrentID=0, Status='live', Symbol='ens', Aliases='Dmel_CG14998, CG14998, Dm-308, Dmel\\CG14998c, swoosh, ens', description='ensconsin', other_designations='CG14998-PA|CG14998-PB|CG14998-PC|CG14998-PD|CG14998-PE|CG14998-PF|CG14998-PG|CG14998-PH|CG14998-PI|CG14998-PJ|CG14998-PK|CG14998-PL|ens-PA|ens-PB|ens-PC|ens-PD|ens-PE|ens-PF|ens-PG|ens-PH|ens-PI|ens-PJ|ens-PK|ens-PL|swoosh', map_location='64A7-64A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4122907, end_position_on_the_genomic_accession=4134550, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  38492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38492, CurrentID=0, Status='live', Symbol='RfC4', Aliases='Dmel_CG14999, 154910_at, BEST:GH24992, CG14999, DmRFC4, Dmel\\CG14999, RFC400, Rfc4, Rfc40, l(3)64Ai, l(3)A1, l(3)Rfc4, l(3)SG13, l(3)SH17, l(3)dsl-1, l(3)dsl1, l(3)e20, l(3)rfc40, rfc, rfc4, rfc40, rfc[40], RfC4', description='Replication factor C subunit 4', other_designations='CG14999-PA|Replication-factor-C 40kD subunit|RfC4-PA|lethal (3) SG13|replication factor C subunit 4', map_location='64A7-64A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4134860, end_position_on_the_genomic_accession=4136038, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38493, CurrentID=0, Status='live', Symbol='Rop', Aliases='Dmel_CG15811, CG15811, Dmel\\CG15811, ROP, SEC1, cs1, l(3)64Ah, l(3)SH16, ras2, rop, rop(Sec1)', description='Ras opposite', other_designations='CG15811-PA|CG15811-PB|Rop-PA|Rop-PB|ras opposite', map_location='64A7-64A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4136359, end_position_on_the_genomic_accession=4138996, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38494, CurrentID=0, Status='live', Symbol='Ras64B', Aliases='Dmel_CG1167, BcDNA:RE36103, C-ras2, CG1167, D-ras-2, D-ras2, DRas2, Dm-ras-64B, Dmel\\CG1167, Dmras64B, Dras2, Dras64B, RAS, RAS2, RAS64B, Ras, Ras2, c-ras2, ras, ras-2, ras2', description='Ras oncogene at 64B', other_designations='CG1167-PA|Ras64B-PA|ras oncogene|small GTP binding protein', map_location='64A7-64A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4139076, end_position_on_the_genomic_accession=4141398, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38495, CurrentID=0, Status='live', Symbol='Akh', Aliases='Dmel_CG1171, AKH, AKH1, CG1171, Dmel\\CG1171, Drm-AKH, Hrth, akh, dAkh', description='Adipokinetic hormone', other_designations='Akh-PA|CG1171-PA|Hypertrehaloseaemic-hormone|adipo- kinetic hormone|adipokinetic hormone|adipokinetic hormone-like', map_location='64A7-64A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4141467, end_position_on_the_genomic_accession=4142070, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38496, CurrentID=0, Status='live', Symbol='CG1299', Aliases='Dmel_ Dmel\\CG1299, SP18, c-SP18', description='CG1299 gene product from transcript CG1299-RA', other_designations='CG1299-PA|CG1299-PB', map_location='64A8-64A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4148696, end_position_on_the_genomic_accession=4152473, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740129, CurrentID=0, Status='live', Symbol='CG34206', Aliases='Dmel_ Dmel\\CG34206', description='CG34206 gene product from transcript CG34206-RA', other_designations='CG34206-PA|uncharacterized protein', map_location='58C4-58C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22163738, end_position_on_the_genomic_accession=22164594, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740130, CurrentID=0, Status='live', Symbol='CG34454', Aliases='Dmel_ CG34018, Dmel\\CG34454, Dmel_CG34018', description='CG34454 gene product from transcript CG34454-RA', other_designations='CG34454-PA|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=231681, end_position_on_the_genomic_accession=232350, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38499, CurrentID=0, Status='live', Symbol='mas', Aliases='Dmel_CG15002, CG15002, Dmel\\CG15002, SPH79, c-SPH79', description='masquerade', other_designations='CG15002-PB|CG15002-PC|mas-PB|mas-PC', map_location='64A8-64A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4161666, end_position_on_the_genomic_accession=4167464, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38500, CurrentID=0, Status='live', Symbol='Ero1L', Aliases='Dmel_CG1333, CG1333, Dmel\\CG1333, ERO1-L, Ero1l, kiga, l(3)64Ag, l(3)SH22', description='Endoplasmic reticulum oxidoreductin-1-like', other_designations='CG1333-PB|Ero1L-PB|endoplasmic reticulum oxidoreductin-1-like|endoplasmic reticulum oxidoreductin-1L|lethal (3) 64Ag', map_location='64A9-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4170404, end_position_on_the_genomic_accession=4173789, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38501, CurrentID=0, Status='live', Symbol='Teh4', Aliases='Dmel_CG15003, CG15003, Dmel\\CG15003, TEH4, anon-WO0172774.111', description='tipE homolog 4', other_designations='CG15003-PB|Teh4-PB', map_location='64A9-64A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4174859, end_position_on_the_genomic_accession=4178270, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38502, CurrentID=0, Status='live', Symbol='Teh2', Aliases='Dmel_CG15004, CG15004, Dmel\\CG15004, TEH2', description='tipE homolog 2', other_designations='CG15004-PA|CG15004-PB|Teh2-PA|Teh2-PB', map_location='64A9-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4178980, end_position_on_the_genomic_accession=4188023, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38503, CurrentID=0, Status='live', Symbol='VhaM9.7-c', Aliases='Dmel_CG11589, CG11589, Dmel\\CG11589, VhaM9.7-1, vhaM9.2-1, vhaM9.7-1, vhaM9.7-3', description='Vacuolar H[+] ATPase M9.7 subunit c', other_designations='CG11589-PA|VhaM9.7-1|VhaM9.7-c-PA|vacuolar H[+] ATPase M9.7 subunit c|vacuolar H[+] ATPase subunit M9.7-c', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4186018, end_position_on_the_genomic_accession=4186865, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38504, CurrentID=0, Status='live', Symbol='tipE', Aliases='Dmel_CG1232, CG1232, Dmel\\CG1232, TipE, tip-E', description='temperature-induced paralytic E', other_designations='CG1232-PA|CG1232-PB|temperature-induced paralysis locus E|temperature-induced-paralytic-E|tipE-PA|tipE-PB', map_location='3-13.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4188499, end_position_on_the_genomic_accession=4193788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38505, CurrentID=0, Status='live', Symbol='CG43367', Aliases='Dmel_ CG1295, CG1308, CG1332, Dmel\\CG43367, Dmel_CG1295, Dmel_CG1308, Dmel_CG1332', description='CG43367 gene product from transcript CG43367-RE', other_designations='CG43367-PC|CG43367-PD|CG43367-PE|CG43367-PF|uncharacterized protein', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4193993, end_position_on_the_genomic_accession=4224003, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  38506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38506, CurrentID=0, Status='live', Symbol='CG15005', Aliases='Dmel_ Dmel\\CG15005', description='CG15005 gene product from transcript CG15005-RA', other_designations='CG15005-PA|uncharacterized protein', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4200825, end_position_on_the_genomic_accession=4204339, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38508, CurrentID=0, Status='live', Symbol='Cpr64Aa', Aliases='Dmel_CG15006, CG15006, DmelCpr64Aa, Dmel\\CG15006', description='Cuticular protein 64Aa', other_designations='CG15006-PA|Cpr64Aa-PA|cuticular protein 64Aa', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4207838, end_position_on_the_genomic_accession=4209178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38509, CurrentID=0, Status='live', Symbol='Cpr64Ab', Aliases='Dmel_CG15007, CG15007, DmelCpr64Ab, Dmel\\CG15007', description='Cuticular protein 64Ab', other_designations='CG15007-PA|Cpr64Ab-PA|cuticular protein 64Ab', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4210880, end_position_on_the_genomic_accession=4211540, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740142, CurrentID=0, Status='live', Symbol='Gpa2', Aliases='Dmel_CG17878, CG17878, Dmel\\CG17878, GPA2, gpa2', description='Glycoprotein hormone alpha 2 ortholog (H. sapiens)', other_designations='CG17878-PB|Gpa2-PB|glycoprotein hormone alpha 2|glycoprotein hormone alpha 2 ortholog', map_location='', chromosome='Unknown', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  38511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38511, CurrentID=0, Status='live', Symbol='Cpr64Ad', Aliases='Dmel_CG1259, CG1259, DmelCpr64Ad, Dmel\\CG1259', description='Cuticular protein 64Ad', other_designations='CG1259-PB|Cpr64Ad-PB|cuticular protein 64Ad', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4215379, end_position_on_the_genomic_accession=4216414, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38513, CurrentID=0, Status='live', Symbol='ImpL2', Aliases='Dmel_CG15009, CG15009, CT34862, Dmel\\CG15009, GH28, IMP-L2, IMPL2, Imp-12, Imp-L2, ImpL-2', description='Ecdysone-inducible gene L2', other_designations='CG15009-PA|CG15009-PB|CG15009-PC|CG15009-PD|ImpL2-PA|ImpL2-PB|ImpL2-PC|ImpL2-PD|ecdysone-inducible gene L2|imaginal morphogenesis protein-late 2|neural and ectodermal development factor', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4224006, end_position_on_the_genomic_accession=4236147, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38514, CurrentID=0, Status='live', Symbol='CG14997', Aliases='Dmel_ BcDNA:GH04863, Dmel\\CG14997', description='CG14997 gene product from transcript CG14997-RB', other_designations='CG14997-PA|CG14997-PB|uncharacterized protein', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4240173, end_position_on_the_genomic_accession=4242603, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38515, CurrentID=0, Status='live', Symbol='pav', Aliases='Dmel_CG1258, 0934/13, 0973/08, 2A12, 2A12MEL, CG1258, DmKLP1/pav, DmMKLP1, DmPAV, DmPav, Dmel\\CG1258, KIF 23, KIF23, Mklp1, PAV, PAV-KLP, PAV-MKLP, PAV/KLP, Pav, Pav KLP, Pav-KLP, Pav-Klp, Pav-klp, PavKLP, anon-EST:fe2A12, anon2A12, group D-KLP, pav-Klp, pav', description='pavarotti', other_designations='CG1258-PA|Kinesin-6|Pavarotti-KLP|anon-fast-evolving-2A12|mitotic kinesin-like protein 1|pav-PA|pavarotti KLP|pavarotti kinesin-like protein|pavarroti', map_location='64A10-64A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4242734, end_position_on_the_genomic_accession=4246233, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38516, CurrentID=0, Status='live', Symbol='ago', Aliases='Dmel_CG15010, Ago, CG15010, DmFbw7, Dmel\\CG15010, FBXW3, FBXW7, SCF[Ago], anon-WO0118547.345, cdc4', description='archipelago', other_designations='CG15010-PA|CG15010-PB|CG15010-PC|ago-PA|ago-PB|ago-PC', map_location='64A11-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4246343, end_position_on_the_genomic_accession=4254306, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38517, CurrentID=0, Status='live', Symbol='CG1265', Aliases='Dmel_ BcDNA:RE09466, Dmel\\CG1265', description='CG1265 gene product from transcript CG1265-RB', other_designations='CG1265-PB', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4254776, end_position_on_the_genomic_accession=4256053, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38518, CurrentID=0, Status='live', Symbol='CG15011', Aliases='Dmel_ Dmel\\CG15011, cg15011', description='CG15011 gene product from transcript CG15011-RA', other_designations='CG15011-PA|uncharacterized protein', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4255991, end_position_on_the_genomic_accession=4259152, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38519, CurrentID=0, Status='live', Symbol='CG1309', Aliases='Dmel_ Dmel\\CG1309', description='CG1309 gene product from transcript CG1309-RA', other_designations='CG1309-PA', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4259430, end_position_on_the_genomic_accession=4261643, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38520, CurrentID=0, Status='live', Symbol='CG11586', Aliases='Dmel_ Dmel\\CG11586', description='CG11586 gene product from transcript CG11586-RA', other_designations='CG11586-PA|CG11586-PB|uncharacterized protein', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4261297, end_position_on_the_genomic_accession=4262470, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38521, CurrentID=0, Status='live', Symbol='VhaM9.7-a', Aliases='Dmel_CG1268, CG1268, Dmel\\CG1268, vhaM9.7-1', description='Vacuolar H[+] ATPase M9.7 subunit a', other_designations='CG1268-PB|CG1268-PC|VhaM9.7-a-PB|VhaM9.7-a-PC|vacuolar H[+] ATPase M9.7 subunit a|vacuolar H[+] ATPase subunit M9.7-a', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4262705, end_position_on_the_genomic_accession=4263464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38522, CurrentID=0, Status='live', Symbol='CG1273', Aliases='Dmel_ 1273, Dmel\\CG1273', description='CG1273 gene product from transcript CG1273-RB', other_designations='CG1273-PB', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4263719, end_position_on_the_genomic_accession=4277778, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38523, CurrentID=0, Status='live', Symbol='CG1311', Aliases='Dmel_ Dmel\\CG1311', description='CG1311 gene product from transcript CG1311-RA', other_designations='CG1311-PA|CG1311-PB', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4277962, end_position_on_the_genomic_accession=4281585, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38525, CurrentID=0, Status='live', Symbol='srw', Aliases='Dmel_CG11582, CG11582, Dmel\\CG11582', description='shrew', other_designations='CG11582-PB|srw-PB', map_location='64B12-64C12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4282956, end_position_on_the_genomic_accession=4283478, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38526, CurrentID=0, Status='live', Symbol='CG1316', Aliases='Dmel_ Dmel\\CG1316, cg1316', description='CG1316 gene product from transcript CG1316-RA', other_designations='CG1316-PA', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4283851, end_position_on_the_genomic_accession=4286865, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38527, CurrentID=0, Status='live', Symbol='TfIIEbeta', Aliases='Dmel_CG1276, CG1276, Dmel\\CG1276, E[[S]], TFIIE, TFIIEbeta, TFIIbeta, TfIIE, dTFIIE-S, dTFIIES', description='Transcription factor IIEbeta', other_designations='CG1276-PA|CG1276-PB|TFIIE small subunit|TfIIEbeta-PA|TfIIEbeta-PB|transcription factor IIEbeta', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4287038, end_position_on_the_genomic_accession=4288349, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38528, CurrentID=0, Status='live', Symbol='Hexo1', Aliases='Dmel_CG1318, CG1318, DmHex1, Dmel\\CG1318, HEX 1, HEX1, HEXO1', description='Hexosaminidase 1', other_designations='CG1318-PA|CG1318-PD|CG1318-PE|CG1318-PF|Hexo1-PA|Hexo1-PD|Hexo1-PE|Hexo1-PF|hexosaminidase 1', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4288641, end_position_on_the_genomic_accession=4291211, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38529, CurrentID=0, Status='live', Symbol='CG15012', Aliases='Dmel_ Dmel\\CG15012', description='CG15012 gene product from transcript CG15012-RC', other_designations='CG15012-PC|uncharacterized protein', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4291208, end_position_on_the_genomic_accession=4292246, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38530, CurrentID=0, Status='live', Symbol='Fdx2', Aliases='Dmel_CG1319, BcDNA:RE08888, BcDNA:RH09920, CG1319, Dmel\\CG1319, Fdxh2', description='Ferredoxin 2', other_designations='CG1319-PB|Fdx2-PB|ferredoxin|ferredoxin 2', map_location='64B1-64B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4292482, end_position_on_the_genomic_accession=4293432, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38531, CurrentID=0, Status='live', Symbol='dyl', Aliases='Dmel_CG15013, CG15013, Dmel\\CG15013', description='dusky-like', other_designations='CG15013-PA|CG15013-PB|CG15013-PC|Dusky-like|dyl-PA|dyl-PB|dyl-PC', map_location='64B1-64B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4294004, end_position_on_the_genomic_accession=4307718, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38532, CurrentID=0, Status='live', Symbol='PMP34', Aliases='Dmel_CG32250, BcDNA:RE36975, CG1326, CG32250, Dmel\\CG32250, dPMP34', description='Peroxisomal Membrane Protein 34', other_designations='CG32250-PA|CG32250-PB|PMP34-PA|PMP34-PB|peroxisomal membrane protein 34', map_location='64B1-64B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4318755, end_position_on_the_genomic_accession=4321109, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38533, CurrentID=0, Status='live', Symbol='CG15014', Aliases='Dmel_ Dmel\\CG15014', description='CG15014 gene product from transcript CG15014-RA', other_designations='CG15014-PA|uncharacterized protein', map_location='64B1-64B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4321105, end_position_on_the_genomic_accession=4322259, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38534, CurrentID=0, Status='live', Symbol='Cip4', Aliases='Dmel_CG15015, CG11341, CG15015, DCIP4, Dmel\\CG15015, cip4, dCIP4, dcip4', description='Cdc42-interacting protein 4', other_designations='CG15015-PA|CG15015-PB|CG15015-PC|CG15015-PD|CG15015-PE|CG15015-PF|CG15015-PG|CG15015-PH|CG15015-PI|CG15015-PJ|CG15015-PK|Cdc42-Interacting protein 4|Cip4-PA|Cip4-PB|Cip4-PC|Cip4-PD|Cip4-PE|Cip4-PF|Cip4-PG|Cip4-PH|Cip4-PI|Cip4-PJ|Cip4-PK', map_location='64B1-64B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4322491, end_position_on_the_genomic_accession=4364102, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  38535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38535, CurrentID=0, Status='live', Symbol='mRpS6', Aliases='Dmel_CG15016, BcDNA:RH10862, CG15016, Dmel\\CG15016, MRP-S6, Mrps6', description='mitochondrial ribosomal protein S6', other_designations='CG15016-PA|CG15016-PB|mRpS6-PA|mRpS6-PB', map_location='64B2-64B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4364550, end_position_on_the_genomic_accession=4365243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740168, CurrentID=0, Status='live', Symbol='CG34456', Aliases='Dmel_ CG4446, Dmel\\CG34456, Dmel_CG4446', description='CG34456 gene product from transcript CG34456-RA', other_designations='CG34456-PA|uncharacterized protein', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9362108, end_position_on_the_genomic_accession=9362734, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38538, CurrentID=0, Status='live', Symbol='CG11342', Aliases='Dmel_ Dmel\\CG11342', description='CG11342 gene product from transcript CG11342-RA', other_designations='CG11342-PA|uncharacterized protein', map_location='64B2-64B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4366494, end_position_on_the_genomic_accession=4367458, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38539, CurrentID=0, Status='live', Symbol='DopEcR', Aliases='Dmel_CG18314, CG18314, DmDopEcR, Dmel\\CG18314, anon-WO0170980.139, anon-WO0170980.140, dop', description='Dopamine/Ecdysteroid receptor', other_designations='CG18314-PA|CG18314-PB|CG18314-PC|DopEcR-PA|DopEcR-PB|DopEcR-PC|dopamine/ecdysteroid receptor', map_location='64B2-64B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4367649, end_position_on_the_genomic_accession=4380387, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38540, CurrentID=0, Status='live', Symbol='slow', Aliases='Dmel_CG7447, CG7447, CT22877, Dmel\\CG7447, Slow, blr, slomo', description='slowdown', other_designations='CG7447-PA|CG7447-PB|N-like|blade runner|slow-PA|slow-PB|slowdowm|slowmotion', map_location='64B3-64B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4384442, end_position_on_the_genomic_accession=4403181, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38541, CurrentID=0, Status='live', Symbol='Syx17', Aliases='Dmel_CG7452, CG7452, Dm Syx17, DmSyx17, Dmel\\CG7452, S17, Syx', description='Syntaxin 17', other_designations='CG7452-PA|CG7452-PB|Syx17-PA|Syx17-PB|syntaxin|syntaxin 17', map_location='64B4-64B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4403487, end_position_on_the_genomic_accession=4405168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38542, CurrentID=0, Status='live', Symbol='CG15019', Aliases='Dmel_ Dmel\\CG15019', description='CG15019 gene product from transcript CG15019-RB', other_designations='CG15019-PA|CG15019-PB|uncharacterized protein', map_location='64B4-64B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4405127, end_position_on_the_genomic_accession=4406059, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38543, CurrentID=0, Status='live', Symbol='DOR', Aliases='Dmel_CG11347, CG11347, Dmel\\CG11347, d', description='CG11347 gene product from transcript CG11347-RE', other_designations='CG11347-PA|CG11347-PB|CG11347-PC|CG11347-PD|CG11347-PE|CG11347-PF|CG11347-PG|CG11347-PH|DOR-PA|DOR-PB|DOR-PC|DOR-PD|DOR-PE|DOR-PF|DOR-PG|DOR-PH', map_location='64B4-64B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4406943, end_position_on_the_genomic_accession=4420648, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38544, CurrentID=0, Status='live', Symbol='CG15020', Aliases='Dmel_ Dmel\\CG15020', description='CG15020 gene product from transcript CG15020-RA', other_designations='CG15020-PA|uncharacterized protein', map_location='64B5-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4421662, end_position_on_the_genomic_accession=4426621, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38545, CurrentID=0, Status='live', Symbol='nAChRbeta1', Aliases='Dmel_CG11348, AChR, AChR64B, ARD, AchR64B, Acr64B, AcrD, Ard, CG11348, CG11348-PA, CG12606, Dbeta1, Dmel\\CG11348, ard, beta 64B, nAChR, nAcR64B, nAcRbeta-64B', description='nicotinic Acetylcholine Receptor beta1', other_designations='AChR protein of drosophila|CG11348-PA|CG11348-PB|nAChRbeta1-PA|nAChRbeta1-PB|nicotinic acetylcholine receptor beta 64B|nicotinic acetylcholine receptor beta1|nicotinic acetylcholine receptorb-64B', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4429349, end_position_on_the_genomic_accession=4435993, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38546, CurrentID=0, Status='live', Symbol='CG15021', Aliases='Dmel_ CT34875, Dmel\\CG15021', description='CG15021 gene product from transcript CG15021-RA', other_designations='CG15021-PA|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4436721, end_position_on_the_genomic_accession=4438311, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740179, CurrentID=0, Status='live', Symbol='CG34437', Aliases='Dmel_ CG16918, Dmel\\CG34437, Dmel_CG16918, SPH201', description='CG34437 gene product from transcript CG34437-RA', other_designations='CG34437-PA|uncharacterized protein', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28283719, end_position_on_the_genomic_accession=28284725, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38548, CurrentID=0, Status='live', Symbol='CG11345', Aliases='Dmel_ Dmel\\CG11345', description='CG11345 gene product from transcript CG11345-RA', other_designations='CG11345-PA|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4457177, end_position_on_the_genomic_accession=4458163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38549, CurrentID=0, Status='live', Symbol='CG15022', Aliases='Dmel_ BcDNA:RE24507, Dmel\\CG15022', description='CG15022 gene product from transcript CG15022-RA', other_designations='CG15022-PA|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4459193, end_position_on_the_genomic_accession=4460317, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38550, CurrentID=0, Status='live', Symbol='CG15023', Aliases='Dmel_ Dmel\\CG15023', description='CG15023 gene product from transcript CG15023-RD', other_designations='CG15023-PB|CG15023-PC|CG15023-PD|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4460622, end_position_on_the_genomic_accession=4461504, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740825, CurrentID=0, Status='live', Symbol='CG34305', Aliases='Dmel_ Dmel\\CG34305', description='CG34305 gene product from transcript CG34305-RA', other_designations='CG34305-PA|uncharacterized protein', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4242785, end_position_on_the_genomic_accession=4243146, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38552, CurrentID=0, Status='live', Symbol='CG11349', Aliases='Dmel_ Dmel\\CG11349', description='CG11349 gene product from transcript CG11349-RB', other_designations='CG11349-PB|uncharacterized protein', map_location='64B7-64B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4473068, end_position_on_the_genomic_accession=4475095, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740185, CurrentID=0, Status='live', Symbol='UQCR-11', Aliases='Dmel_CG41623, CG41623, Dmel\\CG41623, UCRH, Ucrh', description='Ubiquinol-cytochrome c reductase 11 kDa subunit', other_designations='CG41623-PB|CG41623-PC|CG41623-PD|CG41623-PE|CG41623-PF|UQCR-11-PB|UQCR-11-PC|UQCR-11-PD|UQCR-11-PE|UQCR-11-PF|Ubiquinol-cytochrome C reductase complex 11 kDa protein|Ubiquinol-cytochrome c oxidoreductase 11 kDa|Ubiquinol-cytochrome c reductase 11 kDa protein|ucrh', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=25120709, end_position_on_the_genomic_accession=25126171, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38554, CurrentID=0, Status='live', Symbol='CG11350', Aliases='Dmel_ CG1135, Dmel\\CG11350', description='CG11350 gene product from transcript CG11350-RC', other_designations='CG11350-PC|uncharacterized protein', map_location='64B8-64B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4482809, end_position_on_the_genomic_accession=4484370, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38555, CurrentID=0, Status='live', Symbol='CG13722', Aliases='Dmel_ Dmel\\CG13722', description='CG13722 gene product from transcript CG13722-RB', other_designations='CG13722-PB|uncharacterized protein', map_location='64B8-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4485159, end_position_on_the_genomic_accession=4490963, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38556, CurrentID=0, Status='live', Symbol='CG13721', Aliases='Dmel_ Dmel\\CG13721', description='CG13721 gene product from transcript CG13721-RA', other_designations='CG13721-PA|uncharacterized protein', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4496883, end_position_on_the_genomic_accession=4497777, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38557, CurrentID=0, Status='live', Symbol='CG13720', Aliases='Dmel_ Dmel\\CG13720', description='CG13720 gene product from transcript CG13720-RB', other_designations='CG13720-PB|uncharacterized protein', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4498856, end_position_on_the_genomic_accession=4500500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38558, CurrentID=0, Status='live', Symbol='NT1', Aliases='Dmel_CG42576, CG18318, CG32242, CG32244, CG42576, CT30097, DNT1, Dmel\\CG42576, Dmel_CG32242, Dmel_CG32244, Spz2, dNT1, spz2', description='Neurotrophin 1', other_designations='CG42576-PB|CG42576-PC|CG42576-PD|CG42576-PE|CG42576-PF|CG42576-PG|CG42576-PH|NT1-PB|NT1-PC|NT1-PD|NT1-PE|NT1-PF|NT1-PG|NT1-PH|drosophila Neurotrophin1|neurotrophin 1|spatzle 2|spz2', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4501652, end_position_on_the_genomic_accession=4508433, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38559, CurrentID=0, Status='live', Symbol='Tie', Aliases='Dmel_CG7525, CG7525, DPR1, Dmel\\CG7525, RTK, tie', description='Tie-like receptor tyrosine kinase', other_designations='CG7525-PA|Tie-PA|tyrosine kinase|tyrosine kinase with Ig and epidermal growth factor homology domain', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4510698, end_position_on_the_genomic_accession=4532145, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38560, CurrentID=0, Status='live', Symbol='CG11353', Aliases='Dmel_ Dmel\\CG11353, anon-WO0172774.162', description='CG11353 gene product from transcript CG11353-RB', other_designations='CG11353-PA|CG11353-PB|uncharacterized protein', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4533170, end_position_on_the_genomic_accession=4539270, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38561, CurrentID=0, Status='live', Symbol='CG11357', Aliases='Dmel_ BcDNA:GM04393, Dmel\\CG11357, GalT2, dbeta3GnT', description='CG11357 gene product from transcript CG11357-RD', other_designations='CG11357-PA|CG11357-PB|CG11357-PC|CG11357-PD|uncharacterized protein', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4542051, end_position_on_the_genomic_accession=4553457, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38562, CurrentID=0, Status='live', Symbol='CG42540', Aliases='Dmel_ CG13718, CG13719, CG32245, CG7468, Dmel\\CG42540, Dmel_CG13719, Dmel_CG32245', description='CG42540 gene product from transcript CG42540-RB', other_designations='CG42540-PB|CG42540-PC|CG42540-PD|CG42540-PF|CG42540-PG|CG42540-PH|CG42540-PI|CG42540-PJ|CG42540-PK|uncharacterized protein', map_location='64B11-64B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4556087, end_position_on_the_genomic_accession=4592295, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  5740827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740827, CurrentID=0, Status='live', Symbol='CG40635', Aliases='Dmel_ Dmel\\CG40635', description='CG40635 gene product from transcript CG40635-RB', other_designations='CG40635-PB|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  5740196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740196, CurrentID=0, Status='live', Symbol='CG34292', Aliases='Dmel_ Dmel\\CG34292', description='CG34292 gene product from transcript CG34292-RA', other_designations='CG34292-PA|uncharacterized protein', map_location='97E1-97E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27062216, end_position_on_the_genomic_accession=27063059, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38565, CurrentID=0, Status='live', Symbol='HDAC1', Aliases='Dmel_CG7471, CG7471, DHDAC1, DRpd3, DmHDAC1, Dmel\\CG7471, E(var)3-64BC, HDAC, HDAC-1, Hdac, Hdac1, RPD3, Rpd3, Rpd3/HDAC, Su(var)3-26, Su(var)326, Su(var)328, dHDAC-1, dHDAC1, dRPD3, dRpd3, dmHDA401, drpd3, hdac1, l(3)04556, l(3)64Cc, rpd3, rpd[3]', description='Histone deacetylase 1', other_designations='CG7471-PA|HDAC1-PA|Rpd3|Rpd3 histone deacetylase|histone Deacetylase-1|histone deacetylase|histone deacetylase 1|histone deacetylase 1 (HDAC1)|reduced potassium dependency 3|suppressor of variegation 326', map_location='64B12-64B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4626734, end_position_on_the_genomic_accession=4629467, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38567, CurrentID=0, Status='live', Symbol='CG13716', Aliases='Dmel_ Dmel\\CG13716, HL1', description='CG13716 gene product from transcript CG13716-RA', other_designations='CG13716-PA|uncharacterized protein', map_location='64B12-64B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4634436, end_position_on_the_genomic_accession=4634789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38571, CurrentID=0, Status='live', Symbol='CG7514', Aliases='Dmel_ Dmel\\CG7514', description='CG7514 gene product from transcript CG7514-RA', other_designations='CG7514-PA', map_location='64B13-64B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4687587, end_position_on_the_genomic_accession=4688774, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740204, CurrentID=0, Status='live', Symbol='clos', Aliases='Dmel_CG42600, CG12927, CG12928, CG1863, CG33479, CG34364, CG42600, Dmel\\CG42600, Dmel_CG12927, Dmel_CG12928, Dmel_CG1863, Dmel_CG33479, Dmel_CG34364', description='closca', other_designations='CG42600-PD|clos-PD', map_location='45F5-45F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9587953, end_position_on_the_genomic_accession=9599732, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  38573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38573, CurrentID=0, Status='live', Symbol='CG15876', Aliases='Dmel_ Dmel\\CG15876, HL3', description='CG15876 gene product from transcript CG15876-RA', other_designations='CG15876-PA|uncharacterized protein', map_location='64B13-64B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4694205, end_position_on_the_genomic_accession=4694845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38574, CurrentID=0, Status='live', Symbol='CG13712', Aliases='Dmel_ Dmel\\CG13712, HL4', description='CG13712 gene product from transcript CG13712-RA', other_designations='CG13712-PA|uncharacterized protein', map_location='64B14-64B14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4709209, end_position_on_the_genomic_accession=4709890, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38575, CurrentID=0, Status='live', Symbol='CG12493', Aliases='Dmel_ Dmel\\CG12493, cg12493', description='CG12493 gene product from transcript CG12493-RB', other_designations='CG12493-PA|CG12493-PB|uncharacterized protein', map_location='64B16-64B16', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4752393, end_position_on_the_genomic_accession=4753738, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38576, CurrentID=0, Status='live', Symbol='CG13711', Aliases='Dmel_ Dmel\\CG13711, HL5', description='CG13711 gene product from transcript CG13711-RA', other_designations='CG13711-PA|uncharacterized protein', map_location='64B16-64B16', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4757883, end_position_on_the_genomic_accession=4758708, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38578, CurrentID=0, Status='live', Symbol='RhoGEF64C', Aliases='Dmel_CG32239, CG13709, CG13710, CG32239, Dmel\\CG32239, EP(3)3035, EP3035, GEF, GEF(64C), GEF64C, Gef64C, RhoGEF, RhoGef64C', description='Rho guanine nucleotide exchange factor at 64C', other_designations='CG32239-PA|CG32239-PB|Rho guanine nucleotide exchange factor|RhoGEF64C-PA|RhoGEF64C-PB|guanine nucleotide exchange factor|guanine nucleotide exchange factor 64C|guanine nucleotide exchange factor GEF64C', map_location='64B13-64B17', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4692783, end_position_on_the_genomic_accession=4796253, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38579, CurrentID=0, Status='live', Symbol='CG7509', Aliases='Dmel_ CT1509, Dmel\\CG7509, tartan/capricious-like', description='CG7509 gene product from transcript CG7509-RA', other_designations='CG7509-PA', map_location='64B17-64B17', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4803077, end_position_on_the_genomic_accession=4805274, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38580, CurrentID=0, Status='live', Symbol='Dhc64C', Aliases='Dmel_CG7507, CD, CG7507, Cdhc, DHC, DHC64C, Dhc, Dhc46C, Dhc64, Dhc64c, Dmel\\CG7507, Dynein, FCP-A, Fs(3)Lab, Fs(3)Laborc, Fs(3)Sz18, HMW MAP, Lab, Su(Gl)77, cDhc, cDhc64C, dhc, dhc64C, dhc64c, l(3)64Ca', description='Dynein heavy chain 64C', other_designations='CG7507-PA|CG7507-PC|CG7507-PD|CG7507-PE|CG7507-PF|CG7507-PG|CG7507-PH|CG7507-PI|Dhc64C-PA|Dhc64C-PC|Dhc64C-PD|Dhc64C-PE|Dhc64C-PF|Dhc64C-PG|Dhc64C-PH|Dhc64C-PI|cytoplasmic dynein|cytoplasmic dynein heavy chain|cytoplasmic dynenin|dynein|dynein HC|dynein heavy chain|dynein heavy chain 64C|dynein heavy chain 64c|dynein heavy chain at 64C|dynein heavy-chain|female sterile(3)Laborc|laborc|suppressor of glued 77', map_location='64B17-64C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4807368, end_position_on_the_genomic_accession=4825687, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  38581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38581, CurrentID=0, Status='live', Symbol='Aats-leu', Aliases='Dmel_CG7479, CG7479, Dmel\\CG7479, LRS, LeuRS', description='Leucyl-tRNA synthetase', other_designations='Aats-leu-PA|CG7479-PA', map_location='64C1-64C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4825853, end_position_on_the_genomic_accession=4829038, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38582, CurrentID=0, Status='live', Symbol='CG13708', Aliases='Dmel_ CT33170, Dmel\\CG13708', description='CG13708 gene product from transcript CG13708-RB', other_designations='CG13708-PA|CG13708-PB|uncharacterized protein', map_location='64C1-64C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4829092, end_position_on_the_genomic_accession=4833591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38583, CurrentID=0, Status='live', Symbol='CG13707', Aliases='Dmel_ Dmel\\CG13707', description='CG13707 gene product from transcript CG13707-RB', other_designations='CG13707-PB|CG13707-PC|CG13707-PD|uncharacterized protein', map_location='64C1-64C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4837791, end_position_on_the_genomic_accession=4840108, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38584, CurrentID=0, Status='live', Symbol='CG32237', Aliases='Dmel_ BcDNA:RE17910, CG13703, CG13706, Dmel\\CG32237', description='CG32237 gene product from transcript CG32237-RA', other_designations='CG32237-PA|uncharacterized protein', map_location='64C1-64C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4849438, end_position_on_the_genomic_accession=4852461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38586, CurrentID=0, Status='live', Symbol='CG17150', Aliases='Dmel_ Dmel\\CG17150', description='CG17150 gene product from transcript CG17150-RD', other_designations='CG17150-PD|CG17150-PE|uncharacterized protein', map_location='64C1-64C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4861269, end_position_on_the_genomic_accession=4895749, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  38587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38587, CurrentID=0, Status='live', Symbol='CG13705', Aliases='Dmel_CG13705-PA, Dmel\\CG13705', description='CG13705 gene product from transcript CG13705-RA', other_designations='CG13705-PA|uncharacterized protein', map_location='64C2-64C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4896864, end_position_on_the_genomic_accession=4898470, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38588, CurrentID=0, Status='live', Symbol='CG13704', Aliases='Dmel_ BcDNA:RE40159, Dmel\\CG13704', description='CG13704 gene product from transcript CG13704-RA', other_designations='CG13704-PA|uncharacterized protein', map_location='64C2-64C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4898900, end_position_on_the_genomic_accession=4900869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38589, CurrentID=0, Status='live', Symbol='Rh50', Aliases='Dmel_CG7499, BcDNA:GH03016, CG7499, Dmel\\CG7499, Rh-like', description='CG7499 gene product from transcript CG7499-RB', other_designations='CG7499-PA|CG7499-PB|CG7499-PC|Rh50-PA|Rh50-PB|Rh50-PC', map_location='64C2-64C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4910368, end_position_on_the_genomic_accession=4920198, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38590, CurrentID=0, Status='live', Symbol='Con', Aliases='Dmel_CG7503, CG7503, CON, CT1840, Dmel\\CG7503, Ubx-t35, anon-64C, con, conn', description='Connectin', other_designations='CG7503-PA|CG7503-PB|Con-PA|Con-PB|connectin|transcript-35', map_location='64C2-64C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4936758, end_position_on_the_genomic_accession=5079033, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38591, CurrentID=0, Status='live', Symbol='CG17030', Aliases='Dmel_ Dmel\\CG17030', description='CG17030 gene product from transcript CG17030-RA', other_designations='CG17030-PA|uncharacterized protein', map_location='64C4-64C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4980003, end_position_on_the_genomic_accession=4980769, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740224, CurrentID=0, Status='live', Symbol='CG34317', Aliases='Dmel_ CG7950, Dmel\\CG34317, Dmel_CG7950, Dromel_CG7950_FBtr0085563_CHCHD4inmorf_uORF', description='CG34317 gene product from transcript CG34317-RA', other_designations='CG34317-PA|uncharacterized protein', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30056859, end_position_on_the_genomic_accession=30058131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38593, CurrentID=0, Status='live', Symbol='Srp54k', Aliases='Dmel_CG4659, BcDNA.GM09489, BcDNA:GM09489, CG4659, Dmel\\CG4659, Srp54, l(3)01418', description='Signal recognition particle protein 54k', other_designations='CG4659-PA|Srp54k-PA|signal recognition particle protein 54k', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5152678, end_position_on_the_genomic_accession=5154777, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38594, CurrentID=0, Status='live', Symbol='Gen', Aliases='Dmel_CG10670, BcDNA:GM10765, CG10670, DmGEN, Dmel\\CG10670, mus324', description='XPG-like endonuclease', other_designations='CG10670-PA|Gen-PA|mutagen-sensitive 324', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5149849, end_position_on_the_genomic_accession=5152376, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38595, CurrentID=0, Status='live', Symbol='Aats-ala-m', Aliases='Dmel_CG4633, AlaS, CG4633, Dm mt AlaRS, Dmel\\CG4633', description='mitochondrial alanyl-tRNA synthetase', other_designations='Aats-ala-m-PA|CG4633-PA', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5146098, end_position_on_the_genomic_accession=5149852, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38596, CurrentID=0, Status='live', Symbol='CG10671', Aliases='Dmel_ Dmel\\CG10671, FIT2, i136', description='CG10671 gene product from transcript CG10671-RA', other_designations='CG10671-PA|CG10671-PB|uncharacterized protein', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5144082, end_position_on_the_genomic_accession=5146063, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38597, CurrentID=0, Status='live', Symbol='Gdap1', Aliases='Dmel_CG4623, CG4623, Dmel\\CG4623, GDAP1', description='GDAP1 ortholog', other_designations='CG4623-PA|CG4623-PB|CG4623-PC|Gdap1-PA|Gdap1-PB|Gdap1-PC', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5141953, end_position_on_the_genomic_accession=5144317, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38598, CurrentID=0, Status='live', Symbol='CG10672', Aliases='Dmel_ DHRS4, Dmel\\CG10672', description='CG10672 gene product from transcript CG10672-RA', other_designations='CG10672-PA|uncharacterized protein', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5140440, end_position_on_the_genomic_accession=5141646, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38599, CurrentID=0, Status='live', Symbol='CHMP2B', Aliases='Dmel_CG4618, CG4618, Dmel\\CG4618, cg4618', description='Charged multivesicular body protein 2b', other_designations='CG4618-PA|CHMP2B-PA|charged multivesicular body protein 2b', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5139619, end_position_on_the_genomic_accession=5140473, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38600, CurrentID=0, Status='live', Symbol='Prpk', Aliases='Dmel_CG10673, CG10673, Dmel\\CG10673', description='p53-related protein kinase ortholog', other_designations='CG10673-PA|Prpk-PA', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5138652, end_position_on_the_genomic_accession=5139395, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38601, CurrentID=0, Status='live', Symbol='CG4611', Aliases='Dmel_ Dmel\\CG4611', description='CG4611 gene product from transcript CG4611-RA', other_designations='CG4611-PA', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5136376, end_position_on_the_genomic_accession=5138679, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38602, CurrentID=0, Status='live', Symbol='CG10674', Aliases='Dmel_ Dmel\\CG10674', description='CG10674 gene product from transcript CG10674-RA', other_designations='CG10674-PA|uncharacterized protein', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5135748, end_position_on_the_genomic_accession=5136292, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740235, CurrentID=0, Status='live', Symbol='RhoGAP1A', Aliases='Dmel_CG40494, BCR, CG17617, CG17960, CG40453, CG40494, Dmel\\CG40494, Dmel_CG00000, EG:23E12.2, EG:23E12.5, RhoGAP, rhoGAP1A', description='Rho GTPase activating protein at 1A', other_designations='CG40494-PA|CG40494-PB|CG40494-PC|CG40494-PD|CG40494-PE|CG40494-PF|RhoGAP1A-PA|RhoGAP1A-PB|RhoGAP1A-PC|RhoGAP1A-PD|RhoGAP1A-PE|RhoGAP1A-PF', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=127449, end_position_on_the_genomic_accession=140340, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  38604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38604, CurrentID=0, Status='live', Symbol='CG4597', Aliases='Dmel_ Dmel\\CG4597', description='CG4597 gene product from transcript CG4597-RA', other_designations='CG4597-PA', map_location='64C6-64C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5129674, end_position_on_the_genomic_accession=5130240, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38605, CurrentID=0, Status='live', Symbol='shep', Aliases='Dmel_CG32423, BcDNA:LD40028, BcDNA:RH63980, CG10647, CG10649, CG10668, CG32423, Dmel\\CG32423, Shep, alan, alan-shepard, anon-EST:Posey83, anon-WO0118547.198, anon-WO0172774.41, cg10668', description='alan shepard', other_designations='CG32423-PA|CG32423-PB|CG32423-PD|CG32423-PE|CG32423-PF|CG32423-PG|CG32423-PH|CG32423-PI|shep-PA|shep-PB|shep-PD|shep-PE|shep-PF|shep-PG|shep-PH|shep-PI', map_location='64C8-64C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5155821, end_position_on_the_genomic_accession=5277944, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  5740238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740238, CurrentID=0, Status='live', Symbol='CG34238', Aliases='Dmel_ Dmel\\CG34238', description='CG34238 gene product from transcript CG34238-RA', other_designations='CG34238-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10626499, end_position_on_the_genomic_accession=10627439, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38608, CurrentID=0, Status='live', Symbol='CG4669', Aliases='Dmel_ Dmel\\CG4669, anon-WO0140519.194', description='CG4669 gene product from transcript CG4669-RA', other_designations='CG4669-PA', map_location='64C9-64C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5234483, end_position_on_the_genomic_accession=5236631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38610, CurrentID=0, Status='live', Symbol='lama', Aliases='Dmel_CG10645, CG10645, Dmel\\CG10645', description='lamina ancestor', other_designations='CG10645-PA|CG10645-PB|CG10645-PC|CG10645-PD|lama-PA|lama-PB|lama-PC|lama-PD', map_location='64C12-64C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5343619, end_position_on_the_genomic_accession=5355923, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38611, CurrentID=0, Status='live', Symbol='Klp64D', Aliases='Dmel_CG10642, CG10642, DmKlp64D, Dmel\\CG10642, KIF 3A, KIF3A, KLP4, KLP64D, KLP64Ddm, Kif3A, Klp 64D, Klp4, klp64D', description='Kinesin-like protein at 64D', other_designations='CG10642-PA|Kinesin-2|Klp64D-PA', map_location='64C13-64C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5356864, end_position_on_the_genomic_accession=5359392, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38612, CurrentID=0, Status='live', Symbol='Cyt-c1', Aliases='Dmel_CG4769, CG4769, CY1, Dmel\\CG4769, anon-EST:Posey141', description='Cytochrome c1', other_designations='CG4769-PA|CG4769-PB|Cyt-c1-PA|Cyt-c1-PB|Ubiquinol-cytochrome c oxidoreductase cytochrome C1|Ubiquinol-cytochrome c reductase cytochrome C1, heme protein|cytochrome c1|cytochrome c1, heme protein, mitochondrial', map_location='64C13-64C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5360211, end_position_on_the_genomic_accession=5362798, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38613, CurrentID=0, Status='live', Symbol='Uev1A', Aliases='Dmel_CG10640, CG10640, DmUEV, Dmel\\CG10640, UEV, UEV1A, UEV1a, Uev1a, dUEV1A, dUEV1a, dUev1A, uev1a', description='Ubiquitin-conjugating enzyme variant 1A', other_designations='CG10640-PA|CG10640-PB|CG10640-PC|Uev1A-PA|Uev1A-PB|Uev1A-PC', map_location='64C13-64C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5362884, end_position_on_the_genomic_accession=5366225, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38614, CurrentID=0, Status='live', Symbol='Membrin', Aliases='Dmel_CG4780, BcDNA:SD22572, CG4780, Dmel\\CG4780, Memb, SNARE/membrin, membrin', description='CG4780 gene product from transcript CG4780-RA', other_designations='CG4780-PA|Membrin-PA|fmembrin|membrin|moon cheese', map_location='64C13-64C14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5366455, end_position_on_the_genomic_accession=5367381, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38615, CurrentID=0, Status='live', Symbol='CG10635', Aliases='Dmel_ BcDNA:SD11994, Dmel\\CG10635', description='CG10635 gene product from transcript CG10635-RA', other_designations='CG10635-PA|uncharacterized protein', map_location='64C13-64C14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5367341, end_position_on_the_genomic_accession=5368008, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38616, CurrentID=0, Status='live', Symbol='Ir64a', Aliases='Dmel_CG10633, CG10633, CT29782, DmelIR64a, Dmel\\CG10633, IR64a, Ir64, ir64a', description='Ionotropic receptor 64a', other_designations='CG10633-PA|Ir64a-PA|ionotropic receptor 64a', map_location='64C15-64C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5378786, end_position_on_the_genomic_accession=5382821, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  5740836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740836, CurrentID=0, Status='live', Symbol='CG34273', Aliases='Dmel_ Dmel\\CG34273', description='CG34273 gene product from transcript CG34273-RA', other_designations='CG34273-PA|uncharacterized protein', map_location='88C4-88C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14517042, end_position_on_the_genomic_accession=14517487, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38619, CurrentID=0, Status='live', Symbol='CG4835', Aliases='Dmel_ Dmel\\CG4835', description='CG4835 gene product from transcript CG4835-RB', other_designations='CG4835-PB', map_location='64D2-64D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5479790, end_position_on_the_genomic_accession=5483663, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38620, CurrentID=0, Status='live', Symbol='blanks', Aliases='Dmel_CG10630, CG10630, Dmel\\CG10630, lump', description='CG10630 gene product from transcript CG10630-RA', other_designations='CG10630-PA|blanks-PA', map_location='64D2-64D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5497523, end_position_on_the_genomic_accession=5498958, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38622, CurrentID=0, Status='live', Symbol='Lkr', Aliases='Dmel_CG10626, CG10626, DLKR, Dmel\\CG10626, LK-R, LKR, LkR, anon-WO0131005.17, anon-WO0170980.118, anon-WO0170980.119, dLKR, lkr', description='Leucokinin receptor', other_designations='CG10626-PA|Lkr-PA|drosokinin receptor|kinin receptor|leucokinin receptor', map_location='64D2-64D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5512421, end_position_on_the_genomic_accession=5529341, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38623, CurrentID=0, Status='live', Symbol='CG13285', Aliases='Dmel_ Dmel\\CG13285, FBgn0035611', description='CG13285 gene product from transcript CG13285-RB', other_designations='CG13285-PA|CG13285-PB|uncharacterized protein', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5533968, end_position_on_the_genomic_accession=5535759, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38624, CurrentID=0, Status='live', Symbol='CG10625', Aliases='Dmel_ Dmel\\CG10625, anon-WO0118547.236', description='CG10625 gene product from transcript CG10625-RJ', other_designations='CG10625-PA|CG10625-PB|CG10625-PC|CG10625-PD|CG10625-PE|CG10625-PG|CG10625-PH|CG10625-PI|CG10625-PJ|CG10625-PK|uncharacterized protein', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5538411, end_position_on_the_genomic_accession=5551159, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  38628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38628, CurrentID=0, Status='live', Symbol='bc10', Aliases='Dmel_CG4867, CG4867, Dmel\\CG4867, d', description='CG4867 gene product from transcript CG4867-RA', other_designations='CG4867-PA|bc10-PA', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5565231, end_position_on_the_genomic_accession=5566765, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38629, CurrentID=0, Status='live', Symbol='Msr-110', Aliases='Dmel_CG10596, CG10596, Dmel\\CG10596', description='CG10596 gene product from transcript CG10596-RD', other_designations='CG10596-PA|CG10596-PB|CG10596-PC|CG10596-PD|Msr-110-PA|Msr-110-PB|Msr-110-PC|Msr-110-PD|engrailed nuclear homeoprotein-regulated gene', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5566919, end_position_on_the_genomic_accession=5577348, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38630, CurrentID=0, Status='live', Symbol='l(3)psg2', Aliases='Dmel_CG5146, CG5146, Dmel\\CG5146, anon-EST:Liang-1.47, clone 1.47, psg2', description='lethal (3) persistent salivary gland 2', other_designations='CG5146-PA|l(3)psg2-PA', map_location='3-25 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5583669, end_position_on_the_genomic_accession=5590390, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38631, CurrentID=0, Status='live', Symbol='spo', Aliases='Dmel_CG10594, 307a1, CG10594, CYP307A1, CYP307A1-Dm, Cyp307a1, Dmel\\CG10594, Dmspo, Spo', description='spook', other_designations='CG10594-PA|CG10594-PC|spo-PA|spo-PC', map_location='3-19 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5590083, end_position_on_the_genomic_accession=5592571, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38632, CurrentID=0, Status='live', Symbol='CG10592', Aliases='Dmel_ Dmel\\CG10592', description='CG10592 gene product from transcript CG10592-RA', other_designations='CG10592-PA|uncharacterized protein', map_location='64D5-64D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5592741, end_position_on_the_genomic_accession=5595272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38633, CurrentID=0, Status='live', Symbol='CG5150', Aliases='Dmel_ Dmel\\CG5150', description='CG5150 gene product from transcript CG5150-RB', other_designations='CG5150-PA|CG5150-PB', map_location='64D5-64D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5595542, end_position_on_the_genomic_accession=5597422, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38634, CurrentID=0, Status='live', Symbol='CG10591', Aliases='Dmel_ Dmel\\CG10591', description='CG10591 gene product from transcript CG10591-RA', other_designations='CG10591-PA|uncharacterized protein', map_location='64D5-64D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5598557, end_position_on_the_genomic_accession=5599692, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38635, CurrentID=0, Status='live', Symbol='TM9SF3', Aliases='Dmel_CG10590, CG10590, Dmel\\CG10590', description='Transmembrane 9 superfamily protein member 3', other_designations='CG10590-PA|TM9SF3-PA|transmembrane 9 superfamily protein member 3', map_location='64D5-64D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5601096, end_position_on_the_genomic_accession=5603979, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38636, CurrentID=0, Status='live', Symbol='mthl2', Aliases='Dmel_CG17795, CG17795, Dmel\\CG17795, Mth-2, anon-WO0170980.28, anon-WO0170980.29', description='methuselah-like 2', other_designations='CG17795-PB|mth-like 2|mthl2-PB', map_location='64D6-64D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5604790, end_position_on_the_genomic_accession=5607190, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38637, CurrentID=0, Status='live', Symbol='Eaf6', Aliases='Dmel_CG12756, CG12756, Dmel\\CG12756, dEaf6', description='Esa1-associated factor 6', other_designations='CG12756-PA|Eaf6-PA', map_location='64D6-64D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5607162, end_position_on_the_genomic_accession=5608300, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38638, CurrentID=0, Status='live', Symbol='Blimp-1', Aliases='Dmel_CG5249, CG5249, Dmel\\CG5249, blimp-1, d dBlimp1', description='CG5249 gene product from transcript CG5249-RF', other_designations='Blimp-1-PA|Blimp-1-PB|Blimp-1-PC|Blimp-1-PD|Blimp-1-PE|Blimp-1-PF|CG5249-PA|CG5249-PB|CG5249-PC|CG5249-PD|CG5249-PE|CG5249-PF|dBlimp1', map_location='64E1-64E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5630983, end_position_on_the_genomic_accession=5652094, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38639, CurrentID=0, Status='live', Symbol='lin-28', Aliases='Dmel_CG17334, BcDNA:RE05342, CG17334, Dmel\\CG17334, EP(3)0915, lin28', description='CG17334 gene product from transcript CG17334-RA', other_designations='CG17334-PA|lin-28-PA', map_location='64E1-64E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5654681, end_position_on_the_genomic_accession=5657384, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38640, CurrentID=0, Status='live', Symbol='Sse', Aliases='Dmel_CG10583, CG10583, Dmel\\CG10583, SSE, l(3)13m-281, l(3)SG14, sse', description='Separase', other_designations='CG10583-PA|Sse-PA|lethal(3)SG14|separase', map_location='3-26.6 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5659700, end_position_on_the_genomic_accession=5662247, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39208, CurrentID=0, Status='live', Symbol='CG14153', Aliases='Dmel_ Dmel\\CG14153', description='CG14153 gene product from transcript CG14153-RA', other_designations='CG14153-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10643528, end_position_on_the_genomic_accession=10644615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38643, CurrentID=0, Status='live', Symbol='DnaJ-1', Aliases='Dmel_CG10578, CG10578, DNAJ-1, DNAJ1, DROJ1, DmHsp40, Dmel\\CG10578, DnaJ1 64EF, DroJ1, EU3500, HSP40, HSP40/HDJ1, Hdj1, Hsp-40, Hsp40, anon-WO0140519.166, anon-WO0172774.135, dHDJ1, dHDJ1/HSP40, dHdj1, dhdJ1, dhdj-1, dhdj1, dnaJ-1, droj1, hsp40', description='DnaJ-like-1', other_designations='CG10578-PA|CG10578-PB|DnaJ-1-PA|DnaJ-1-PB|Hsp40|dnaJ like1', map_location='64E5-64E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5750029, end_position_on_the_genomic_accession=5752189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38644, CurrentID=0, Status='live', Symbol='Usp47', Aliases='Dmel_CG5486, BcDNA:LD12764, CG5486, D-UBP-64E, D-Ubp-64E, Dmel\\CG5486, E(var)1, E(var)3-64E, E-var(3)64E, Q24574, UBP, Ubp64, Ubp64E, dUSP47, dUbp64, ubp64, ubp64E', description='Ubiquitin specific protease 47', other_designations='CG5486-PA|CG5486-PB|CG5486-PD|Ubiquitin-specific protease 64E|Usp47-PA|Usp47-PB|Usp47-PD|ubiquitin specific protease 47', map_location='64E5-64E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5752973, end_position_on_the_genomic_accession=5761886, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38645, CurrentID=0, Status='live', Symbol='CG10576', Aliases='Dmel_ BEST:CK00496, Dmel\\CG10576', description='CG10576 gene product from transcript CG10576-RD', other_designations='CG10576-PA|CG10576-PC|CG10576-PD|proliferation-associated protein|uncharacterized protein', map_location='64E6-64E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5761963, end_position_on_the_genomic_accession=5765702, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38646, CurrentID=0, Status='live', Symbol='Txl', Aliases='Dmel_CG5495, 5495, CG5495, Dmel\\CG5495, anon-WO0118547.248, txl', description='Thioredoxin-like', other_designations='CG5495-PA|Txl-PA|thioredoxin', map_location='64E6-64E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5765846, end_position_on_the_genomic_accession=5767084, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38647, CurrentID=0, Status='live', Symbol='Ppat-Dpck', Aliases='Dmel_CG10575, CG10575, DPCK/PPAT, Dmel\\CG10575, PPAT-DPCK, dPPAT-DPCK', description='Bifunctional Phosphopantetheine adenylyltransferase - Dephospho-CoA kinase', other_designations="4'-phosphopantetheine adenylyltransferase - dephospho-CoA kinase|CG10575-PA|Ppat-Dpck-PA|bifunctional phosphopantetheine adenylyltransferase - Dephospho-CoA kinase", map_location='64E7-64E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5767017, end_position_on_the_genomic_accession=5768794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38648, CurrentID=0, Status='live', Symbol='scny', Aliases='Dmel_CG5505, 2331, CG5505, Dmel\\CG5505, Q9VRP5, Scny, USP36, Usp36, anon-WO0118547.247, dUSP36, dUsp36, et, l(3)02331, mule', description='scrawny', other_designations="CG5505-PB|CG5505-PC|CG5505-PD|CG5505-PG|CG5505-PH|CG5505-PI|CG5505-PJ|emperor's thumb|lethal (3) 02331|mule|scny-PB|scny-PC|scny-PD|scny-PG|scny-PH|scny-PI|scny-PJ|ubiquitin C-terminal hydrolase 36|ubiquitin specific protease 36|ubiquitin-specific protease 36", map_location='64E8-64E12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5769002, end_position_on_the_genomic_accession=5776271, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38649, CurrentID=0, Status='live', Symbol='Myt1', Aliases='Dmel_CG32417, CG10569, CG32417, Dmel\\CG32417, Dmyt1, MYT 1, MYT1, d myt1', description='CG32417 gene product from transcript CG32417-RA', other_designations='CG32417-PA|Myt1-PA', map_location='64E7-64E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5777476, end_position_on_the_genomic_accession=5779591, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740282, CurrentID=0, Status='live', Symbol='CG34315', Aliases='Dmel_ Dmel\\CG34315', description='CG34315 gene product from transcript CG34315-RB', other_designations='CG34315-PB|uncharacterized protein', map_location='49F7-49F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13202728, end_position_on_the_genomic_accession=13206547, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38651, CurrentID=0, Status='live', Symbol='Cralbp', Aliases='Dmel_CG10546, CG10546, CRALBP, Dmel\\CG10546', description='Cellular retinaldehyde binding protein', other_designations='CG10546-PA|Cralbp-PA|cellular retinaldehyde binding protein', map_location='64E8-64E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5783215, end_position_on_the_genomic_accession=5786718, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38652, CurrentID=0, Status='live', Symbol='Bre1', Aliases='Dmel_CG10542, BRE1, Bre-1, CG10542, Dmel\\CG10542, EC3-8, bre-1, bre1, dBRE1, d l(3)01640', description='CG10542 gene product from transcript CG10542-RB', other_designations='Bre1-PA|Bre1-PB|CG10542-PA|CG10542-PB|lethal (3) 01640', map_location='64E8-64E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5789831, end_position_on_the_genomic_accession=5794192, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38653, CurrentID=0, Status='live', Symbol='Tektin-C', Aliases='Dmel_CG10541, CG10541, Dmel\\CG10541', description='Tektin C', other_designations='CG10541-PA|CG10541-PC|CG10541-PD|TEKTIN C|Tektin-C-PA|Tektin-C-PC|Tektin-C-PD|tektin C', map_location='64E8-64E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5794405, end_position_on_the_genomic_accession=5796543, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38654, CurrentID=0, Status='live', Symbol='S6k', Aliases='Dmel_CG10539, 10539, 7-10, 7084, CG10539, DS6K, Dmel\\CG10539, Dp70S6k, Dp70[s6k], Dp70s6k, Pk64F, Pk?6, S6K, S6K1|S6K, dS6K, dS6k, dp70[S6k], dp70s6k, dps6k, ds6k, fs(3)07084, l(3)07084, p-S6K, p70 S6K, p70/S6K, p70S6K, p70[S6 kinase], p70[S6K], p70[S6k], p70[S6kinase], p70s6K, s6k, s6k11, S6k', description='Ribosomal protein S6 kinase', other_designations='CG10539-PA|CG10539-PB|RPS6-p70-protein kinase|S6 kinase|S6-kinase|S6K kinase|S6k-PA|S6k-PB|S6kinase|dS6 kinase|p70 S6 kinase|p70 S6 kinase 1|p70 ribosomal S6 kinase|p70S6 kinase|p70S6K kinase|protein kinase-like 64F|ribosomal S6 protein kinase|ribosomal protein S6 kinase', map_location='64F-64E11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5797549, end_position_on_the_genomic_accession=5810931, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38655, CurrentID=0, Status='live', Symbol='CG5537', Aliases='Dmel_ Dmel\\CG5537, anon-EST:Posey275, uprt', description='CG5537 gene product from transcript CG5537-RC', other_designations='CG5537-PA|CG5537-PB|CG5537-PC', map_location='64E11-64E11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5810973, end_position_on_the_genomic_accession=5812228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38656, CurrentID=0, Status='live', Symbol='mad2', Aliases='Dmel_CG17498, CG17498, Dmel\\CG17498, MAD2, Mad2', description='CG17498 gene product from transcript CG17498-RA', other_designations='CG17498-PA|mad2-PA', map_location='64E11-64E12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5812135, end_position_on_the_genomic_accession=5813093, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38657, CurrentID=0, Status='live', Symbol='vn', Aliases='Dmel_CG10491, CG10491, CT29452, Dm vn, Dmel\\CG10491, P1749, Vein, Vn, ddd, l(3)10567, l(3)10568, l(3)L6A, l(3)ddd, l(3)rF264, l(3)vn10567, wvn', description='vein', other_designations='CG10491-PA|CG10491-PC|Dm vein|defective dorsal discs|vn-PA|vn-PC', map_location='3-16.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5813338, end_position_on_the_genomic_accession=5845664, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38658, CurrentID=0, Status='live', Symbol='CG5568', Aliases='Dmel_ Dmel\\CG5568', description='CG5568 gene product from transcript CG5568-RA', other_designations='CG5568-PA', map_location='64F3-64F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5871501, end_position_on_the_genomic_accession=5873518, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38659, CurrentID=0, Status='live', Symbol='CG18586', Aliases='Dmel_ Dmel\\CG18586', description='CG18586 gene product from transcript CG18586-RC', other_designations='CG18586-PB|CG18586-PC|uncharacterized protein', map_location='64F3-64F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5873311, end_position_on_the_genomic_accession=5875466, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38660, CurrentID=0, Status='live', Symbol='CG13287', Aliases='Dmel_ Dmel\\CG13287', description='CG13287 gene product from transcript CG13287-RA', other_designations='CG13287-PA|CG13287-PB|uncharacterized protein', map_location='64F3-64F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5882304, end_position_on_the_genomic_accession=5884090, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38661, CurrentID=0, Status='live', Symbol='DNApol-epsilon58', Aliases='Dmel_CG10489, 10489, CG10489, DNApol-epsilon, Dmel\\CG10489, POLE2, Pole2, dpolepsilonp58', description='DNA polymerase epsilon 58kD subunit', other_designations='CG10489-PA|DNA polymerase epsilon B subunit|DNApol-epsilon58-PA|Pole2', map_location='64F4-64F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5895691, end_position_on_the_genomic_accession=5897435, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740294, CurrentID=0, Status='live', Symbol='CG41520', Aliases='Dmel_ CG17443, CG40067, CG40109, Dmel\\CG41520', description='CG41520 gene product from transcript CG41520-RD', other_designations='CG41520-PA|CG41520-PB|CG41520-PC|CG41520-PD|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1649061, end_position_on_the_genomic_accession=1778937, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38663, CurrentID=0, Status='live', Symbol='QC', Aliases='Dmel_CG32412, CG10487, CG32412, Dmel\\CG32412, DromeQC', description='Glutaminyl cyclase', other_designations='CG32412-PA|QC-PA|glutaminyl cyclase', map_location='64F4-64F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5900835, end_position_on_the_genomic_accession=5902456, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38664, CurrentID=0, Status='live', Symbol='CG10486', Aliases='Dmel_ Dmel\\CG10486', description='CG10486 gene product from transcript CG10486-RB', other_designations='CG10486-PB|CG10486-PC|uncharacterized protein', map_location='64F5-64F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5904211, end_position_on_the_genomic_accession=5906699, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38665, CurrentID=0, Status='live', Symbol='CG13288', Aliases='Dmel_ Dmel\\CG13288, dmCG13288', description='CG13288 gene product from transcript CG13288-RB', other_designations='CG13288-PB|CG13288-PC|uncharacterized protein', map_location='64F5-64F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5907381, end_position_on_the_genomic_accession=5917540, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38666, CurrentID=0, Status='live', Symbol='CG10483', Aliases='Dmel_ Dmel\\CG10483', description='CG10483 gene product from transcript CG10483-RA', other_designations='CG10483-PA|uncharacterized protein', map_location='64F5-64F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5917905, end_position_on_the_genomic_accession=5921316, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  5740299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740299, CurrentID=0, Status='live', Symbol='Sws1', Aliases='Dmel_CG34314, CG34314, Dmel\\CG34314, Dromel_CG3240_FBtr0077747_rad1_uORF, dmSws1', description='SWIM domain containing Srs2 interacting protein 1', other_designations='CG34314-PA|Sws1-PA', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2573436, end_position_on_the_genomic_accession=2575052, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38668, CurrentID=0, Status='live', Symbol='Vap-33B', Aliases='Dmel_CG33523, BcDNA:GH09028, CG10727, CG10977, CG32410, CG33523, Dmel\\CG33523', description='VAMP-associated protein of 33kDa ortholog B', other_designations='CG33523-PA|CG33523-PB|CG33523-PC|CG33523-PD|CG33523-PE|CG33523-PF|MSP protein|Vap-33B-PA|Vap-33B-PB|Vap-33B-PC|Vap-33B-PD|Vap-33B-PE|Vap-33B-PF', map_location='64F5-65A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5923862, end_position_on_the_genomic_accession=5930120, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38669, CurrentID=0, Status='live', Symbol='Rcc1', Aliases='Dmel_CG10480, BJ1, Bj1, CG10480, CG18640 (Bj1), Dmel\\CG10480, RCC1, RanGEF, dRCC1, inxs, sdt', description='Regulator of chromosome condensation 1 ortholog (H. sapiens)', other_designations='Bj1 protein|CG10480-PA|CG10480-PB|CG10480-PC|Rcc1-PA|Rcc1-PB|Rcc1-PC|in excess|regulator of chromosome condensation 1|regulator of chromosome condensation 1 ortholog|side-out', map_location='65A1-65A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5930262, end_position_on_the_genomic_accession=5932773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740303, CurrentID=0, Status='live', Symbol='CG34386', Aliases='Dmel_ CG14496, CG14497, Dmel\\CG34386, Dmel_CG14496, Dmel_CG14497', description='CG34386 gene product from transcript CG34386-RD', other_designations='CG34386-PC|CG34386-PD|uncharacterized protein', map_location='55B2-55B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18062415, end_position_on_the_genomic_accession=18093628, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38674, CurrentID=0, Status='live', Symbol='CG10479', Aliases='Dmel_ Dmel\\CG10479', description='CG10479 gene product from transcript CG10479-RA', other_designations='CG10479-PA|CG10479-PB|uncharacterized protein', map_location='65A1-65A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5967545, end_position_on_the_genomic_accession=5976007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38675, CurrentID=0, Status='live', Symbol='alphaKap4', Aliases='Dmel_CG10478, CG10478, CG11040, Dmel\\CG10478, g7295403', description='alpha Karyopherin-4', other_designations='CG10478-PA|alphaKap4-PA|importin alpha4', map_location='65A2-65A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6009446, end_position_on_the_genomic_accession=6011021, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740846, CurrentID=0, Status='live', Symbol='JYalpha', Aliases='Dmel_CG45760, CG17923, CG40625, CG45760, Dmel\\CG45760, Dmel_CG17923, Dmel_CG40625, JY-alpha, JYAlpha', description='CG45760 gene product from transcript CG45760-RB', other_designations='CG45760-PB|JYalpha-PB', map_location='101F1-101F1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=879, end_position_on_the_genomic_accession=5039, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740310, CurrentID=0, Status='live', Symbol='Nox', Aliases='Dmel_CG34399, CG15924, CG34399, CG3896, Dmel\\CG34399, Dmel_CG15924, Dmel_CG3896, NOX-11, dNox, dnox, Nox', description='NADPH oxidase', other_designations='CG34399-PB|CG34399-PC|CG34399-PD|CG34399-PE|Nox-PB|Nox-PC|Nox-PD|Nox-PE', map_location='53B3-53B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16307269, end_position_on_the_genomic_accession=16316485, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  38679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38679, CurrentID=0, Status='live', Symbol='CG10477', Aliases='Dmel_ Dmel\\CG10477, SP51, anon-WO0140519.232', description='CG10477 gene product from transcript CG10477-RA', other_designations='CG10477-PA|CG10477-PB|NF-YB-like|uncharacterized protein', map_location='65A3-65A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6033199, end_position_on_the_genomic_accession=6034611, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38680, CurrentID=0, Status='live', Symbol='yip7', Aliases='Dmel_CG6457, CG6457, Dmel\\CG6457, MAC, SP116, SPH116, mac', description='yippee interacting protein 7', other_designations='CG6457-PA|machete|yip7-PA', map_location='65A3-65A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6035216, end_position_on_the_genomic_accession=6036175, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38681, CurrentID=0, Status='live', Symbol='CG6462', Aliases='Dmel_ Dmel\\CG6462, SP127', description='CG6462 gene product from transcript CG6462-RA', other_designations='CG6462-PA', map_location='65A3-65A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6037111, end_position_on_the_genomic_accession=6038188, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38682, CurrentID=0, Status='live', Symbol='Jon65Aiv', Aliases='Dmel_CG6467, BEST:GH23808, CG6467, Dmel\\CG6467, Jon65, Jon65A, SP117, jon65aiv', description='Jonah 65Aiv', other_designations='BEST:GH23808|CG6467-PA|Jon65Aiv-PA|Jonah65A-IV|jonah 65A|jonah 65Aiv', map_location='65A3-65A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6039155, end_position_on_the_genomic_accession=6040135, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38683, CurrentID=0, Status='live', Symbol='Jon65Aiii', Aliases='Dmel_CG6483, CG6483, CG6484, Dmel\\CG6483, Jon65, Jon65A, SP98, anon-EST:GressD16, anon-WO0140519.59, jon65Aiii', description='Jonah 65Aiii', other_designations='CG6483-PA|Jon65Aiii-PA|jonah 65|jonah 65A|jonah 65Aiii', map_location='65A3-65A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6050757, end_position_on_the_genomic_accession=6051690, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740316, CurrentID=0, Status='live', Symbol='CG34460', Aliases='Dmel_ CG6301, Dmel\\CG34460, Dmel_CG6301, bs26g12, bs26g12.y1', description='CG34460 gene product from transcript CG34460-RA', other_designations='CG34460-PA|uncharacterized protein', map_location='53D11-53D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16826872, end_position_on_the_genomic_accession=16827755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38685, CurrentID=0, Status='live', Symbol='Jon65Ai', Aliases='Dmel_CG10475, CG10475, CT29408, Dmel\\CG10475, Jon65, Jon65A, SP145', description='Jonah 65Ai', other_designations='CG10475-PA|CG10475-PB|Jon65Ai-PA|Jon65Ai-PB|jonah 65A|jonah 65Ai', map_location='65A4-65A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6053493, end_position_on_the_genomic_accession=6054443, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740318, CurrentID=0, Status='live', Symbol='CG43729', Aliases='Dmel_ CG12958, CG12959, CG14163, CG18625, CG30470, CG34318, CG34318/CG34365/CG8179, CG34365, CG43223, CG8179, Dm_2R:47664, Dm_2R:47672, Dmel\\CG43729, Dmel_CG12958, Dmel_CG12959, Dmel_CG18625, Dmel_CG30470, Dmel_CG34318, Dmel_CG34365, Dmel_CG43223, Dmel_CG8179, mdcds_11908', description='CG43729 gene product from transcript CG43729-RH', other_designations='CG43729-PA|CG43729-PC|CG43729-PE|CG43729-PH|CG43729-PI|CG43729-PJ|CG43729-PM|CG43729-PN|CG43729-PO|CG43729-PP|CG43729-PQ|CG43729-PR|CG43729-PS|CG43729-PT|CG43729-PU|CG43729-PV|uncharacterized protein', map_location='51F13-52A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15401096, end_position_on_the_genomic_accession=15471884, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  38687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38687, CurrentID=0, Status='live', Symbol='CG6592', Aliases='Dmel_ Dmel\\CG6592, SP105, lincRNA.S4213', description='CG6592 gene product from transcript CG6592-RA', other_designations='CG6592-PA', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6059658, end_position_on_the_genomic_accession=6061378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740320, CurrentID=0, Status='live', Symbol='uex', Aliases="Dmel_CG42595, 41Ad, CG40084, CG42595, CR41594, Dmel\\CG42595, Dmel_CG40084, Dmel_CR41594, EMS45-73, GroupIII, IR3, l(2)41Ad, l(2)EMS34-7, l(2)IR3, l(2R)A', l(2R)EMS45-73, l(2Rh)IR3", description='unextended', other_designations='CG42595-PE|CG42595-PF|CG42595-PG|CG42595-PH|lethal (2) IR3|uex-PE|uex-PF|uex-PG|uex-PH', map_location='2-55.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=3900285, end_position_on_the_genomic_accession=3949425, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38689, CurrentID=0, Status='live', Symbol='CG42269', Aliases='Dmel_ CG6596, CG6600, Dmel\\CG42269, Dmel_CG6596, Dmel_CG6600', description='CG42269 gene product from transcript CG42269-RE', other_designations='CG42269-PE|uncharacterized protein', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6066296, end_position_on_the_genomic_accession=6071545, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740323, CurrentID=0, Status='live', Symbol='CG34380', Aliases='Dmel_ CG13986, CG13987, CG13988, CG15813, CG31640, Dmel\\CG34380, Dmel_CG13988, Dmel_CG31640, anon-WO0073329.6', description='CG34380 gene product from transcript CG34380-RD', other_designations='CG34380-PC|CG34380-PD|CG34380-PE|uncharacterized protein', map_location='26B11-26B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6179133, end_position_on_the_genomic_accession=6211114, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38692, CurrentID=0, Status='live', Symbol='CG13295', Aliases='Dmel_ Dmel\\CG13295', description='CG13295 gene product from transcript CG13295-RA', other_designations='CG13295-PA|uncharacterized protein', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6073490, end_position_on_the_genomic_accession=6075245, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740325, CurrentID=0, Status='live', Symbol='CG42732', Aliases='Dmel_ CG12897, CG12899, CG12900/CG34363, CG12901, CG12904, CG34363, Dm_2R:24301, Dm_2R:24438, Dmel\\CG42732, Dmel_CG12897, Dmel_CG12899, Dmel_CG12904, Dmel_CG34363, Slack', description='CG42732 gene product from transcript CG42732-RN', other_designations='CG42732-PD|CG42732-PE|CG42732-PF|CG42732-PG|CG42732-PH|CG42732-PI|CG42732-PL|CG42732-PN|CG42732-PO|uncharacterized protein', map_location='47A1-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10322195, end_position_on_the_genomic_accession=10416996, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  38694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38694, CurrentID=0, Status='live', Symbol='ssp6', Aliases='Dmel_CG6619, CG6619, Dmel\\CG6619, Ssp6', description='short spindle 6', other_designations='CG6619-PA|CG6619-PB|ssp6-PA|ssp6-PB', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6076722, end_position_on_the_genomic_accession=6082491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38695, CurrentID=0, Status='live', Symbol='CG13293', Aliases='Dmel_ Dmel\\CG13293', description='CG13293 gene product from transcript CG13293-RC', other_designations='CG13293-PB|CG13293-PC|CG13293-PD|CG13293-PE|uncharacterized protein', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6082883, end_position_on_the_genomic_accession=6092188, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38696, CurrentID=0, Status='live', Symbol='CG10469', Aliases='Dmel_ Dmel\\CG10469, SP123', description='CG10469 gene product from transcript CG10469-RA', other_designations='CG10469-PA|uncharacterized protein', map_location='65A5-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6092188, end_position_on_the_genomic_accession=6093313, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38697, CurrentID=0, Status='live', Symbol='CG10467', Aliases='Dmel_ Dmel\\CG10467', description='CG10467 gene product from transcript CG10467-RA', other_designations='CG10467-PA|uncharacterized protein', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6093204, end_position_on_the_genomic_accession=6094747, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38698, CurrentID=0, Status='live', Symbol='loj', Aliases='Dmel_CG10733, CG10733, Dmel\\CG10733, LOJ, l(3)04026, p24', description='logjam', other_designations='CG10733-PD|CG10733-PE|CG10733-PF|loj-PD|loj-PE|loj-PF', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6094987, end_position_on_the_genomic_accession=6096517, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5740331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740331, CurrentID=0, Status='live', Symbol='CG34291', Aliases='Dmel_ Dmel\\CG34291', description='CG34291 gene product from transcript CG34291-RA', other_designations='CG34291-PA|uncharacterized protein', map_location='97B2-97B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26425290, end_position_on_the_genomic_accession=26425781, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38700, CurrentID=0, Status='live', Symbol='Ets65A', Aliases='Dmel_CG7018, CG13294, CG6882, CG7018, D-Ets-3, D-ets-3, Dmel\\CG7018, ETS3, Ets, Ets3, ets65A', description='Ets at 65A', other_designations='CG7018-PA|CG7018-PB|CG7018-PC|CG7018-PD|CG7018-PE|Ets65A-PA|Ets65A-PB|Ets65A-PC|Ets65A-PD|Ets65A-PE', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6097617, end_position_on_the_genomic_accession=6123913, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38701, CurrentID=0, Status='live', Symbol='l(3)mbn', Aliases='Dmel_CG12755, CG12755, Dmel\\CG12755, MBN, l(3)malignant blood neoplasm1-1, l(3)mbn1, l(3)mbn', description='lethal (3) malignant blood neoplasm', other_designations='CG12755-PB|CG12755-PD|CG12755-PE|l(3)mbn-PB|l(3)mbn-PD|l(3)mbn-PE|lethal(3)malignant blood neoplasm-1', map_location='3-13.3 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6123934, end_position_on_the_genomic_accession=6128807, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38702, CurrentID=0, Status='live', Symbol='Lcp65Ag2', Aliases='Dmel_CG10534, BcDNA:RE01644, CG10534, DCP8-beta, Dmel\\CG10534, LCP8, L[[3]]CP8, Lcp-g2, Lcp8, chr3L:6106987..6107153', description='CG10534 gene product from transcript CG10534-RA', other_designations='CG10534-PA|Lcp65Ag2-PA', map_location='3-11 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6132990, end_position_on_the_genomic_accession=6133593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38703, CurrentID=0, Status='live', Symbol='Lcp65Ag1', Aliases='Dmel_CG10530, CG10530, DCP8-alpha, Dmel\\CG10530, LCP8, L[[3]]CP8, Lcp-g1, Lcp8', description='CG10530 gene product from transcript CG10530-RA', other_designations='CG10530-PA|Lcp65Ag1-PA', map_location='3-11 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6134697, end_position_on_the_genomic_accession=6135274, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38706, CurrentID=0, Status='live', Symbol='Cpr65Aw', Aliases='Dmel_CG32404, C32404, CG10461, CG32404, DmelCpr65Aw, Dmel\\CG32404', description='Cuticular protein 65Aw', other_designations='CG32404-PA|CG32404-PB|Cpr65Aw-PA|Cpr65Aw-PB|cuticular protein 65Aw', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6141903, end_position_on_the_genomic_accession=6142413, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38707, CurrentID=0, Status='live', Symbol='Lcp65Ad', Aliases='Dmel_CG6955, BcDNA:LP05231, CG6955, DCP1, DMCLP65d, Dmel Dmel\\CG6955, Lcp-d', description='CG6955 gene product from transcript CG6955-RB', other_designations='CG6955-PA|CG6955-PB|Lcp65Ad-PA|Lcp65Ad-PB', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6143395, end_position_on_the_genomic_accession=6144101, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38708, CurrentID=0, Status='live', Symbol='Lcp65Ac', Aliases='Dmel_CG6956, CG6956, DCP2, DMLCP65c, Dmel\\CG6956, LCP65Ac, Lcp-c', description='CG6956 gene product from transcript CG6956-RA', other_designations='CG6956-PA|Lcp65Ac-PA', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6144819, end_position_on_the_genomic_accession=6145393, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38709, CurrentID=0, Status='live', Symbol='Lcp65Aa', Aliases='Dmel_CG7287, CG7287, DCP4, DMLCP65a, DmelLCp65Aa, Dmel\\CG7287, Lcp-a', description='CG7287 gene product from transcript CG7287-RA', other_designations='CG7287-PA|Lcp65Aa-PA', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6151304, end_position_on_the_genomic_accession=6151803, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38710, CurrentID=0, Status='live', Symbol='Acp65Aa', Aliases='Dmel_CG10297, ACP65A, Acp65A, BcDNA:RH61147, CG10297, DCP5, DMACP65A, DmelAcp65Aa, Dmel\\CG10297', description='Adult cuticle protein 65Aa', other_designations='Acp65Aa-PA|CG10297-PA|adult cuticle protein 65Aa', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6152050, end_position_on_the_genomic_accession=6152745, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38711, CurrentID=0, Status='live', Symbol='CG13297', Aliases='Dmel_ BcDNA:RE64858, Dmel\\CG13297', description='CG13297 gene product from transcript CG13297-RA', other_designations='CG13297-PA|uncharacterized protein', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6156592, end_position_on_the_genomic_accession=6157377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38712, CurrentID=0, Status='live', Symbol='Cpr65Az', Aliases='Dmel_CG12330, CG12330, DmelCpr65Az, Dmel\\CG12330', description='Cuticular protein 65Az', other_designations='CG12330-PA|Cpr65Az-PA|cuticular protein 65Az', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6160814, end_position_on_the_genomic_accession=6161589, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38713, CurrentID=0, Status='live', Symbol='CG13296', Aliases='Dmel_ Dmel\\CG13296', description='CG13296 gene product from transcript CG13296-RA', other_designations='CG13296-PA|uncharacterized protein', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6163127, end_position_on_the_genomic_accession=6164682, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38714, CurrentID=0, Status='live', Symbol='CG10289', Aliases='Dmel_ Dmel\\CG10289, PpV[[R]]', description='CG10289 gene product from transcript CG10289-RC', other_designations='CG10289-PA|CG10289-PB|CG10289-PC|CG10289-PD|uncharacterized protein', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6164916, end_position_on_the_genomic_accession=6173880, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38715, CurrentID=0, Status='live', Symbol='CG7376', Aliases='Dmel_ Dmel\\CG7376, SHPRH', description='CG7376 gene product from transcript CG7376-RB', other_designations='CG7376-PA|CG7376-PB', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6174248, end_position_on_the_genomic_accession=6178941, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38716, CurrentID=0, Status='live', Symbol='CG10274', Aliases='Dmel_ Dmel\\CG10274', description='CG10274 gene product from transcript CG10274-RB', other_designations='CG10274-PB|uncharacterized protein', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6178677, end_position_on_the_genomic_accession=6181829, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38717, CurrentID=0, Status='live', Symbol='D19B', Aliases='Dmel_CG10270, CG10270, Dmel\\CG10270, d19b', description='CG10270 gene product from transcript CG10270-RA', other_designations='CG10270-PA|D19B-PA', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6182176, end_position_on_the_genomic_accession=6184920, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38718, CurrentID=0, Status='live', Symbol='D19A', Aliases='Dmel_CG10269, CG10269, Dmel\\CG10269, d19a', description='CG10269 gene product from transcript CG10269-RA', other_designations='CG10269-PA|D19A-PA|zinc finger AT-hook protein', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6185830, end_position_on_the_genomic_accession=6189123, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39221, CurrentID=0, Status='live', Symbol='CG7949', Aliases='Dmel_ Dmel\\CG7949', description='CG7949 gene product from transcript CG7949-RA', other_designations='CG7949-PA', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10882890, end_position_on_the_genomic_accession=10883573, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38720, CurrentID=0, Status='live', Symbol='CG13298', Aliases='Dmel_ Dmel\\CG13298, cg13298', description='CG13298 gene product from transcript CG13298-RA', other_designations='CG13298-PA|uncharacterized protein', map_location='65A8-65A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6192124, end_position_on_the_genomic_accession=6193179, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38721, CurrentID=0, Status='live', Symbol='Trn', Aliases='Dmel_CG7398, CG7398, Dmel\\CG7398, IMPbeta2, TRN, Tm1, dTRN, Trn', description='Transportin', other_designations='CG7398-PA|CG7398-PB|CG7398-PC|CG7398-PD|Trn-PA|Trn-PB|Trn-PC|Trn-PD|transportin', map_location='65A8-65A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6193480, end_position_on_the_genomic_accession=6200385, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38722, CurrentID=0, Status='live', Symbol='CG8219', Aliases='Dmel_ 148274_at, Dmel\\CG8219, EP(3)3072', description='CG8219 gene product from transcript CG8219-RA', other_designations='CG8219-PA', map_location='65A8-65A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6200473, end_position_on_the_genomic_accession=6203776, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38723, CurrentID=0, Status='live', Symbol='LanA', Aliases='Dmel_CG10236, 5]], CG10236, CT28769, Dmel\\CG10236, FBgn0002526, FBtr0077014, LM-A/alpha1, Lam, Lam-A, LamA, Lan, alpha3/5, alpha[[3, hdln, lamA, laminin alpha3/5, lanA', description='Laminin A', other_designations='CG10236-PA|LanA-PA|alpha3,5 laminin|headline|laminin|laminin A|laminin A chain|laminin A subunit|laminin A5|laminin alpha|laminin alpha-chain|laminin alpha3,5|laminin alpha[[3,5]]|lamininA', map_location='65A8-65A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6203831, end_position_on_the_genomic_accession=6218022, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  38724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38724, CurrentID=0, Status='live', Symbol='CG13299', Aliases='Dmel_ Dmel\\CG13299', description='CG13299 gene product from transcript CG13299-RA', other_designations='CG13299-PA|uncharacterized protein', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6231805, end_position_on_the_genomic_accession=6232361, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38725, CurrentID=0, Status='live', Symbol='CG10226', Aliases='Dmel_ AMT, Dmel\\CG10226', description='CG10226 gene product from transcript CG10226-RA', other_designations='ABC multidrug transporter|CG10226-PA|CG10226-PB|uncharacterized protein', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6234323, end_position_on_the_genomic_accession=6240068, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38726, CurrentID=0, Status='live', Symbol='Mdr65', Aliases='Dmel_CG10181, CG10181, Dmel\\CG10181, MDR65A, P-gp, mdr65, Mdr65', description='Multi drug resistance 65', other_designations='CG10181-PA|Mdr65-PA|P-glycoprotein|P-glycoprotein65-Dm|multi drug resistance 65|multiple drug resistance 65', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6240500, end_position_on_the_genomic_accession=6245746, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  5740359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740359, CurrentID=0, Status='live', Symbol='spok', Aliases='Dmel_CG41624, 307a2p, CG40123, CG41624, CYP307A2, Cyp307a2, Cyp307a2Psi, Cyp307a2p, Dmel\\CG41624, Dmspok, spo2', description='spookier', other_designations='CG41624-PB|Cyp307a2Psi|cytochrome P450|spok-PB|spook 2', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=3322810, end_position_on_the_genomic_accession=3354486, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38728, CurrentID=0, Status='live', Symbol='CG10163', Aliases='Dmel_ Dmel\\CG10163', description='CG10163 gene product from transcript CG10163-RA', other_designations='CG10163-PA|uncharacterized protein', map_location='65A10-65A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6257423, end_position_on_the_genomic_accession=6258760, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740362, CurrentID=0, Status='live', Symbol='CG34384', Aliases='Dmel_ CG14462, CG2667, CG31187, CG9851, DGKd, Dmel\\CG34384, Dmel_CG14462, Dmel_CG31187, dgkd', description='CG34384 gene product from transcript CG34384-RC', other_designations='CG34384-PA|CG34384-PB|CG34384-PC|CG34384-PD|CG34384-PE|CG34384-PF|uncharacterized protein', map_location='84E1-84E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7814361, end_position_on_the_genomic_accession=7855392, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  38732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38732, CurrentID=0, Status='live', Symbol='CG42747', Aliases='Dmel_ CG14910, CG14911, Dmel\\CG42747, Dmel_CG14910, Dmel_CG14911', description='CG42747 gene product from transcript CG42747-RC', other_designations='CG42747-PC|uncharacterized protein', map_location='65B3-65B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6400590, end_position_on_the_genomic_accession=6485964, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38733, CurrentID=0, Status='live', Symbol='CG10147', Aliases='Dmel_ Dmel\\CG10147', description='CG10147 gene product from transcript CG10147-RB', other_designations='CG10147-PA|CG10147-PB|uncharacterized protein', map_location='65B3-65B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6486167, end_position_on_the_genomic_accession=6488424, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38734, CurrentID=0, Status='live', Symbol='CG8270', Aliases='Dmel_ Dmel\\CG8270', description='CG8270 gene product from transcript CG8270-RB', other_designations='CG8270-PB', map_location='65B3-65B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6488681, end_position_on_the_genomic_accession=6491210, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38735, CurrentID=0, Status='live', Symbol='CG10144', Aliases='Dmel_ Dmel\\CG10144, Vps8, anon-WO0118547.365', description='CG10144 gene product from transcript CG10144-RA', other_designations='CG10144-PA|uncharacterized protein', map_location='65B3-65B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6491230, end_position_on_the_genomic_accession=6495297, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38736, CurrentID=0, Status='live', Symbol='sfl', Aliases='Dmel_CG8339, CG8339, CT24567, Dmel\\CG8339, NDST, Sfl, dw, l(3)03844, lincRNA.S4253', description='sulfateless', other_designations='CG8339-PA|CG8339-PB|diet-wingless|sfl-PA|sfl-PB', map_location='65B3-65B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6495680, end_position_on_the_genomic_accession=6550104, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38738, CurrentID=0, Status='live', Symbol='ndl', Aliases='Dmel_CG10129, CG10129, Dmel\\CG10129, Ndl, SP89, mat(3)2, mel(3)5, nud', description='nudel', other_designations='CG10129-PA|ndl-PA', map_location='3-17 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6593938, end_position_on_the_genomic_accession=6603366, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  38740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38740, CurrentID=0, Status='live', Symbol='CG8368', Aliases='Dmel_ Dmel\\CG8368', description='CG8368 gene product from transcript CG8368-RB', other_designations='CG8368-PA|CG8368-PB|RNase H-like', map_location='65B5-65C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6603684, end_position_on_the_genomic_accession=6608739, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38741, CurrentID=0, Status='live', Symbol='CG8398', Aliases='Dmel_ Dmel\\CG8398, anon-EST:Posey31', description='CG8398 gene product from transcript CG8398-RA', other_designations='CG8398-PA|CG8398-PC|CG8398-PD|CG8398-PE', map_location='65C1-65C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6609325, end_position_on_the_genomic_accession=6619113, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38742, CurrentID=0, Status='live', Symbol='GluRIA', Aliases='Dmel_CG8442, CG8442, DGluR-I, DGluR-IA, DGluRI, DmelGlu-R1, Dmel\\CG8442, Dro GluRIA, Glu-R1, Glu-RI, GluR1, GluRI, dGluRI, dglur-IA', description='Glutamate receptor IA', other_designations='CG8442-PA|GluRIA-PA|glutamate receptor|glutamate receptor I|glutamate receptor IA|glutamate receptor kainate-selective subunit|kainate receptor', map_location='65C1-65C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6624032, end_position_on_the_genomic_accession=6635059, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38743, CurrentID=0, Status='live', Symbol='eIF4E-4', Aliases='Dmel_CG10124, BEST:GH01027, CG10124, CT28485, Dmel\\CG10124, NEST:bs04e06, eIF4E', description='CG10124 gene product from transcript CG10124-RA', other_designations='CG10124-PA|eIF4E-4-PA', map_location='65C1-65C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6664224, end_position_on_the_genomic_accession=6665641, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38744, CurrentID=0, Status='live', Symbol='SP1173', Aliases='Dmel_CG10121, CG10121, Dmel\\CG10121', description='CG10121 gene product from transcript CG10121-RC', other_designations='CG10121-PA|CG10121-PB|CG10121-PC|CG10121-PD|SP1173-PA|SP1173-PB|SP1173-PC|SP1173-PD', map_location='65C1-65C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6689327, end_position_on_the_genomic_accession=6703521, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38745, CurrentID=0, Status='live', Symbol='CG8519', Aliases='Dmel_ Dmel\\CG8519', description='CG8519 gene product from transcript CG8519-RA', other_designations='CG8519-PA|CG8519-PB', map_location='65C3-65C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6704860, end_position_on_the_genomic_accession=6709000, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38746, CurrentID=0, Status='live', Symbol='ple', Aliases='Dmel_CG10118, CG10118, DH65B, DTH, Dmel\\CG10118, Pale, Ple, TH, Th, VII, dTH1, dTH65B, pale/ple, th', description='pale', other_designations='CG10118-PA|CG10118-PB|CG10118-PC|Tyrosine-3-hydroxylase|XBB70 group VII|drosophila tyrosine hydroxylase|ple-PA|ple-PB|ple-PC|tyrosine hydrodroxylase|tyrosine hydroxylase|tyrosine-hydroxylase|tyrosone hydroxylase', map_location='3-18 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6713356, end_position_on_the_genomic_accession=6719525, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38747, CurrentID=0, Status='live', Symbol='dikar', Aliases='Dmel_CG42799, CG10115, CG32393, CG32394, CG42799, Dmel\\CG42799, Dmel_CG32393, Dmel_CG32394', description='CG42799 gene product from transcript CG42799-RG', other_designations='CG42799-PB|CG42799-PC|CG42799-PD|CG42799-PE|CG42799-PF|CG42799-PG|dikar-PB|dikar-PC|dikar-PD|dikar-PE|dikar-PF|dikar-PG', map_location='65C3-65C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6728800, end_position_on_the_genomic_accession=6743164, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38748, CurrentID=0, Status='live', Symbol='velo', Aliases='Dmel_CG10107, CG10107, Dmel\\CG10107', description='veloren', other_designations='CG10107-PA|CG10107-PB|CG10107-PC|CG10107-PD|velo-PA|velo-PB|velo-PC|velo-PD|verloren', map_location='65C3-65C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6744426, end_position_on_the_genomic_accession=6752752, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38749, CurrentID=0, Status='live', Symbol='CG8549', Aliases='Dmel_ Dmel\\CG8549', description='CG8549 gene product from transcript CG8549-RA', other_designations='CG8549-PA|CG8549-PB', map_location='65C3-65C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6752913, end_position_on_the_genomic_accession=6754029, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38750, CurrentID=0, Status='live', Symbol='CG10103', Aliases='Dmel_ Dmel\\CG10103', description='CG10103 gene product from transcript CG10103-RD', other_designations='CG10103-PA|CG10103-PB|CG10103-PC|CG10103-PD|CG10103-PE|CG10103-PF|uncharacterized protein', map_location='65C3-65C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6754017, end_position_on_the_genomic_accession=6756497, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38751, CurrentID=0, Status='live', Symbol='CG32392', Aliases='Dmel_ CG10099, CG14912, Dmel\\CG32392', description='CG32392 gene product from transcript CG32392-RA', other_designations='CG32392-PA|CG32392-PB|uncharacterized protein', map_location='65C4-65C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6756629, end_position_on_the_genomic_accession=6763246, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38752, CurrentID=0, Status='live', Symbol='vvl', Aliases='Dmel_CG10037, CG10037, Cf1-a, Cf1a, Cf1a/drifter, Cfla, DFR, Dfr, Dfr/Vvl, Dmel\\CG10037, Drf, VVL, Vv1, Vvl, dfr, dfr-vvl, dfr/vvl, drf, prd3, sep, ut/dfr, vvl/dfr/ut, vvl', description='ventral veins lacking', other_designations='CG10037-PA|CG10037-PC|CG10037-PD|CG10037-PE|CG10037-PF|Drifter/Ventral veinless|PRD gene 3|drifter|paired-like 3|separated|u-turn|ventral veinless|vvl-PA|vvl-PC|vvl-PD|vvl-PE|vvl-PF', map_location='65C5-65C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6790158, end_position_on_the_genomic_accession=6794941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38753, CurrentID=0, Status='live', Symbol='Prat2', Aliases='Dmel_CG10078, CG10078, DmPrat2, Dmel\\CG10078', description='Phosphoribosylamidotransferase 2', other_designations='CG10078-PA|CG10078-PB|Phosphoribosylamidotransferase-2|Prat2-PA|Prat2-PB|amidophosphoribosyltransferase|phosphoribosyl amidotransferase 2|phosphoribosylamidotransferase 2', map_location='65D2-65D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6911999, end_position_on_the_genomic_accession=6914876, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38754, CurrentID=0, Status='live', Symbol='CG14820', Aliases='Dmel_ Dmel\\CG14820', description='CG14820 gene product from transcript CG14820-RA', other_designations='CG14820-PA|uncharacterized protein', map_location='65D3-65D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6932926, end_position_on_the_genomic_accession=6934824, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38755, CurrentID=0, Status='live', Symbol='tow', Aliases='Dmel_CG14821, CG14821, Dmel\\CG14821, Tow', description='target of wingless', other_designations='CG14821-PC|CG14821-PD|CG14821-PE|CG14821-PF|CG14821-PG|tow-PC|tow-PD|tow-PE|tow-PF|tow-PG', map_location='65D3-65D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6942856, end_position_on_the_genomic_accession=6949313, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38756, CurrentID=0, Status='live', Symbol='CG10077', Aliases='Dmel_ BcDNA:HL01868, BcDNA:HL07910, DmRH5, Dmel\\CG10077, cg10077, p90', description='CG10077 gene product from transcript CG10077-RA', other_designations='CG10077-PA|CG10077-PC|CG10077-PD|uncharacterized protein', map_location='65D3-65D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6949491, end_position_on_the_genomic_accession=6955904, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38757, CurrentID=0, Status='live', Symbol='CG9948', Aliases='Dmel_ Dmel\\CG9948', description='CG9948 gene product from transcript CG9948-RB', other_designations='CG9948-PA|CG9948-PB|CG9948-PC', map_location='65D4-65D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6955991, end_position_on_the_genomic_accession=6957440, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38758, CurrentID=0, Status='live', Symbol='CG10075', Aliases='Dmel_ Dmel\\CG10075', description='CG10075 gene product from transcript CG10075-RA', other_designations='CG10075-PA|uncharacterized protein', map_location='65D4-65D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6956946, end_position_on_the_genomic_accession=6958108, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740391, CurrentID=0, Status='live', Symbol='CG34176', Aliases='Dmel_ Dmel\\CG34176', description='CG34176 gene product from transcript CG34176-RA', other_designations='CG34176-PA|uncharacterized protein', map_location='24D3-24D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4004080, end_position_on_the_genomic_accession=4004480, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38760, CurrentID=0, Status='live', Symbol='sgl', Aliases='Dmel_CG10072, 8310, BG01214, CG10072, Dmel\\CG10072, Sgl, UDP-GlcDH, UDPGDH, XIII, kiwi, l(3)05007, l(3)08310, l(3)10503, l(3)65CEd, l(3)SG9, ska, sl', description='sugarless', other_designations='CG10072-PA|UDP-glucose 6-dehydrogenase|UDP-glucose dehydrogenase|distal W5.4 group XIII|kiwi|lethal(3)SG9|sgl-PA|suppenkasper', map_location='3-22.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6959554, end_position_on_the_genomic_accession=6964758, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38761, CurrentID=0, Status='live', Symbol='CG10064', Aliases='Dmel_ Dmel\\CG10064', description='CG10064 gene product from transcript CG10064-RA', other_designations='CG10064-PA|uncharacterized protein', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6965260, end_position_on_the_genomic_accession=6967897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38762, CurrentID=0, Status='live', Symbol='Mis12', Aliases='Dmel_CG18156, BcDNA:RE19545, CG18156, Dmel\\CG18156, Mis-12, dm', description='CG18156 gene product from transcript CG18156-RA', other_designations='CG18156-PA|Mis12-PA', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6968035, end_position_on_the_genomic_accession=6969048, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38763, CurrentID=0, Status='live', Symbol='CG9953', Aliases='Dmel_ Dmel\\CG9953', description='CG9953 gene product from transcript CG9953-RA', other_designations='CG9953-PA', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6969658, end_position_on_the_genomic_accession=6971666, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38764, CurrentID=0, Status='live', Symbol='CG10063', Aliases='Dmel_ BcDNA:AT30027, Dmel\\CG10063', description='CG10063 gene product from transcript CG10063-RA', other_designations='CG10063-PA|uncharacterized protein', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6971579, end_position_on_the_genomic_accession=6972467, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38765, CurrentID=0, Status='live', Symbol='Galphai', Aliases='Dmel_CG10060, CG10060, DG[[alpha1]], DGalpha1, Dmel\\CG10060, G-alpha-65A, G-ia65A, G-ialpha65A, G-ialpha65a, G-oalpha65A, G[[alphai]], G[[i]], G[[i]]alpha, Gai, Galpha i, Galpha65A, Galpha[[i65A]], Galpha[[i]], Galpha[[i]]65Aota65A, Gi, Gi-alpha, Gi/0, Gialpha, i, Galphai', description='G protein alpha i subunit', other_designations='CG10060-PA|G protein I a-subunit 65A|G protein alpha o subunit 65A|G protein alpha subunit|G protein alpha-subunit|G protein alphai subunit 65A|G protein alphao subunit 65A|Galphai-PA', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6972750, end_position_on_the_genomic_accession=6980044, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38766, CurrentID=0, Status='live', Symbol='bin', Aliases='Dmel_CG18647, BIN, Bin, CG18647, DmFoxF, Dmbin, Dmel\\CG18647, FOXF, FoxF, XIVI, l(3)65CEe, bin', description='biniou', other_designations='CG18647-PA|bin-PA|distal W5.4 group XIV', map_location='65D6-65D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6990622, end_position_on_the_genomic_accession=6994447, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38768, CurrentID=0, Status='live', Symbol='CG17744', Aliases='Dmel_ Dmel\\CG17744', description='CG17744 gene product from transcript CG17744-RA', other_designations='CG17744-PA|uncharacterized protein', map_location='65E1-65E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7025801, end_position_on_the_genomic_accession=7026794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38769, CurrentID=0, Status='live', Symbol='Mp', Aliases='Dmel_CG42543, CG14822, CG33171, CG42543, CG8645, CG8647, COL18A1, CT14872, Dmel\\CG42543, Dmel_CG33171, Dmel_CG8647, Dmp, dmp, mp', description='Multiplexin', other_designations='CG42543-PB|CG42543-PE|CG42543-PF|CG42543-PG|CG42543-PH|CG42543-PI|CG42543-PJ|CG42543-PL|CG42543-PM|CG42543-PN|CG42543-PO|CG42543-PP|CG42543-PR|CG42543-PS|CG42543-PT|CG42543-PU|Collagen-like|Mp-PB|Mp-PE|Mp-PF|Mp-PG|Mp-PH|Mp-PI|Mp-PJ|Mp-PL|Mp-PM|Mp-PN|Mp-PO|Mp-PP|Mp-PR|Mp-PS|Mp-PT|Mp-PU|multiplexin', map_location='65E1-65E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6998843, end_position_on_the_genomic_accession=7053499, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  5740402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740402, CurrentID=0, Status='live', Symbol='CG34345', Aliases='Dmel_ BP1012, CG11053, CG34009, Dmel\\CG34345, Dmel_CG11053, Dmel_CG34009', description='CG34345 gene product from transcript CG34345-RA', other_designations='CG34345-PA|uncharacterized protein', map_location='26F3-26F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6623112, end_position_on_the_genomic_accession=6625359, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38771, CurrentID=0, Status='live', Symbol='CG8641', Aliases='Dmel_ Dmel\\CG8641, Q9VS10', description='CG8641 gene product from transcript CG8641-RB', other_designations='CG8641-PA|CG8641-PB', map_location='65E2-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7058930, end_position_on_the_genomic_accession=7077462, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38772, CurrentID=0, Status='live', Symbol='CG14823', Aliases='Dmel_ Dmel\\CG14823', description='CG14823 gene product from transcript CG14823-RA', other_designations='CG14823-PA|CG14823-PB|CG14823-PD|uncharacterized protein', map_location='65E3-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7066314, end_position_on_the_genomic_accession=7069463, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38773, CurrentID=0, Status='live', Symbol='Cpr65Ea', Aliases='Dmel_CG8640, CG8640, DmelCpr65Ea, Dmel\\CG8640', description='Cuticular protein 65Ea', other_designations='CG8640-PA|Cpr65Ea-PA|cuticular protein 65Ea', map_location='65E3-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7086003, end_position_on_the_genomic_accession=7086608, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38774, CurrentID=0, Status='live', Symbol='Cpr65Eb', Aliases='Dmel_CG8638, BcDNA:RE28679, CG8638, DmelCpr65Eb, Dmel\\CG8638', description='Cuticular protein 65Eb', other_designations='CG8638-PA|Cpr65Eb-PA|cuticular protein 65Eb', map_location='65E3-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7087713, end_position_on_the_genomic_accession=7088787, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38775, CurrentID=0, Status='live', Symbol='Cpr65Ec', Aliases='Dmel_CG8634, CG8634, DmelCpr65Ec, Dmel\\CG8634', description='Cuticular protein 65Ec', other_designations='CG8634-PA|Cpr65Ec-PA|cuticular protein 65Ec', map_location='65E3-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7093204, end_position_on_the_genomic_accession=7093841, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740408, CurrentID=0, Status='live', Symbol='CG17707', Aliases='Dmel_ Dmel\\CG17707, EG:23E12.3', description='CG17707 gene product from transcript CG17707-RC', other_designations='CG17707-PB|CG17707-PC|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=142731, end_position_on_the_genomic_accession=148426, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38779, CurrentID=0, Status='live', Symbol='msl-3', Aliases='Dmel_CG8631, CG8631, DmMsl-3, Dmel\\CG8631, MSL, MSL-3, MSL3, MSL3_DROME, Msl 1-3, Msl3, mle(3)132, mle3, mls3, msl, msl3', description='male-specific lethal 3', other_designations='CG8631-PA|gene 2|male sex lethal 3|male sex lethal-3|male specifc lethal|male specific lethal 3|msl-3-PA', map_location='3-25.8 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7123213, end_position_on_the_genomic_accession=7125533, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38780, CurrentID=0, Status='live', Symbol='BBS1', Aliases='Dmel_CG14825, CG14825, Dmel\\CG14825', description='Bardet-Biedl syndrome 1 ortholog (H. sapiens)', other_designations='BBS1-PA|Bardet-Biedl syndrome 1 ortholog|CG14825-PA', map_location='65E4-65E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7125905, end_position_on_the_genomic_accession=7128118, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38781, CurrentID=0, Status='live', Symbol='CG8629', Aliases='Dmel_ BcDNA:RE33457, Dmel\\CG8629', description='CG8629 gene product from transcript CG8629-RA', other_designations='CG8629-PA', map_location='65E4-65E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7127913, end_position_on_the_genomic_accession=7128264, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740414, CurrentID=0, Status='live', Symbol='Snoo', Aliases='Dmel_CG34421, CG34421, CG7093, CG7233, Dmel\\CG34421, Dmel_CG7093, Dmel_CG7233, DskiL, Sno, SnoA, SnoI, SnoN, SnoN2, dSki, dSno, dSnoN, dsno, dsnoN, l(2)SH1402, l(2)SH2 1402, snoN', description='Sno oncogene', other_designations='CG34421-PB|CG34421-PC|CG34421-PD|CG34421-PG|Sno-oncogene|Snoo-PB|Snoo-PC|Snoo-PD|Snoo-PG|lethal (2) SH1402|snoN', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7891187, end_position_on_the_genomic_accession=7984196, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38783, CurrentID=0, Status='live', Symbol='CG8628', Aliases='Dmel_ Dmel\\CG8628, anon-SAGE:Wang-073', description='CG8628 gene product from transcript CG8628-RA', other_designations='CG8628-PA|CG8628-PB|anon-SAGE:Wang-073', map_location='65E4-65E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7130874, end_position_on_the_genomic_accession=7131505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38784, CurrentID=0, Status='live', Symbol='Dbi', Aliases='Dmel_CG8627, ACBP, BcDNA:RH39533, CG8627, DBI, Dmel\\CG8627, anon-WO0172774.131, dbi', description='Diazepam-binding inhibitor', other_designations='CG8627-PA|CG8627-PB|Dbi-PA|Dbi-PB|acyl-CoA binding protein|diazepam binding inhibitor|diazepam-binding protein|endozepine', map_location='65E4-65E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7131863, end_position_on_the_genomic_accession=7133463, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38785, CurrentID=0, Status='live', Symbol='melt', Aliases='Dmel_CG8624, 1441/14, CG8624, Dmel\\CG8624, Melt, Q9VS24_DROME, anon-WO03014152.10, dMelt', description='melted', other_designations='CG8624-PB|CG8624-PC|CG8624-PD|melt-PB|melt-PC|melt-PD', map_location='65E4-65E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7134516, end_position_on_the_genomic_accession=7143074, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38786, CurrentID=0, Status='live', Symbol='corn', Aliases='Dmel_CG42278, CG32386, CG32390, CG42278, CG8621, Dmel\\CG42278, Dmel_CG32386, Dmel_CG32390', description='cornetto', other_designations='CG42278-PB|CG42278-PC|CG42278-PD|CG42278-PE|corn-PB|corn-PC|corn-PD|corn-PE', map_location='65E5-65E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7142991, end_position_on_the_genomic_accession=7157073, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38788, CurrentID=0, Status='live', Symbol='Dscam2', Aliases='Dmel_CG42256, BcDNA:AT03867, CG14828, CG17742, CG32387, CG32389, CG42256, CG8618, CG8619, CT14580, CT14584, CT34644, Dmel\\CG42256, Dmel_CG32387, Dmel_CG32389, Dscam-like', description='Down syndrome cell adhesion molecule 2', other_designations='CG42256-PE|CG42256-PF|CG42256-PG|CG42256-PH|CG42256-PI|CG42256-PJ|CG42256-PK|CG42256-PL|CG42256-PM|CG42256-PN|CG42256-PO|CG42256-PP|CG42256-PQ|CG42256-PR|CG42256-PS|Dscam2-PE|Dscam2-PF|Dscam2-PG|Dscam2-PH|Dscam2-PI|Dscam2-PJ|Dscam2-PK|Dscam2-PL|Dscam2-PM|Dscam2-PN|Dscam2-PO|Dscam2-PP|Dscam2-PQ|Dscam2-PR|Dscam2-PS|down syndrome cell adhesion molecule 2', map_location='65E6-65E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7159199, end_position_on_the_genomic_accession=7189931, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  38791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38791, CurrentID=0, Status='live', Symbol='CG14826', Aliases='Dmel_ Dmel\\CG14826', description='CG14826 gene product from transcript CG14826-RA', other_designations='CG14826-PA|CG14826-PC|uncharacterized protein', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7234512, end_position_on_the_genomic_accession=7235965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38792, CurrentID=0, Status='live', Symbol='CG14829', Aliases='Dmel_ Dmel\\CG14829', description='CG14829 gene product from transcript CG14829-RB', other_designations='CG14829-PB|uncharacterized protein', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7236800, end_position_on_the_genomic_accession=7238122, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38793, CurrentID=0, Status='live', Symbol='BHD', Aliases='Dmel_CG8616, CG8616, CT25610, DBHD, Dmel\\CG8616, dFLCN', description='Birt-Hogg-Dube homolog', other_designations='BHD-PA|Birt-Hogg-Dube|Birt-Hogg-Dube syndrome|CG8616-PA|GABA A receptor beta3-like|folliculin homolog', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7238542, end_position_on_the_genomic_accession=7240245, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38794, CurrentID=0, Status='live', Symbol='RpL18', Aliases='Dmel_CG8615, CG8615, Dmel\\CG8615, Rp L18e, RpL18', description='Ribosomal protein L18', other_designations='CG8615-PA|CG8615-PB|CG8615-PC|RpL18-PA|RpL18-PB|RpL18-PC|ribosomal protein L18', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7240203, end_position_on_the_genomic_accession=7241351, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38795, CurrentID=0, Status='live', Symbol='Neos', Aliases='Dmel_CG8614, BcDNA.HL05936, BcDNA:HL05936, CG8614, Dmel\\CG8614, l(3)L2249', description='Neosin', other_designations='CG8614-PA|CG8614-PB|Neos-PA|Neos-PB|neosin', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7241628, end_position_on_the_genomic_accession=7243139, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38796, CurrentID=0, Status='live', Symbol='mRpL50', Aliases='Dmel_CG8612, ANON-65Eb, BcDNA:GM06185, BcDNA:RE06140, CG8612, Dmel\\CG8612, MRP-L50, anon-65Eb', description='mitochondrial ribosomal protein L50', other_designations='CG8612-PA|mRpL50-PA', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7243080, end_position_on_the_genomic_accession=7243933, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38797, CurrentID=0, Status='live', Symbol='mei-P22', Aliases='Dmel_CG14827, CG14827, Dmel\\CG14827, MEI-P22, Mei-P22', description='meiotic P22', other_designations='CG14827-PA|mei-P22-PA', map_location='3-21.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7244388, end_position_on_the_genomic_accession=7245383, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38798, CurrentID=0, Status='live', Symbol='Cdc27', Aliases='Dmel_CG8610, APC3, Apc3, CDC27, CDC27Dm, CG8610, Dmcdc27, Dmel\\CG8610, cdc27, l(3)L7123, mks', description='Cell division cycle 27 ortholog', other_designations='CG8610-PA|CG8610-PB|Cdc27-PA|Cdc27-PB|cell division cycle 27 ortholog|makos', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7245481, end_position_on_the_genomic_accession=7249340, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38799, CurrentID=0, Status='live', Symbol='MED4', Aliases='Dmel_CG8609, CG8609, Dmel\\CG8609, Med4, Med4/TRAP36, Trap36, dTRAP36, dp34, med4, p34', description='Mediator complex subunit 4', other_designations='CG8609-PA|MED4-PA|mediator|mediator complex subunit 4', map_location='65E9-65E9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7249529, end_position_on_the_genomic_accession=7250527, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38800, CurrentID=0, Status='live', Symbol='CG14830', Aliases='Dmel_ Dmel\\CG14830', description='CG14830 gene product from transcript CG14830-RC', other_designations='CG14830-PA|CG14830-PB|CG14830-PC|uncharacterized protein', map_location='65E9-65E10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7250718, end_position_on_the_genomic_accession=7263920, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38801, CurrentID=0, Status='live', Symbol='unc-13-4A', Aliases='Dmel_CG32381, CG14831, CG14832, CG14833, CG32381, CG8608, Dmel\\CG32381, d', description='CG32381 gene product from transcript CG32381-RC', other_designations='CG32381-PB|CG32381-PC|unc-13-4A-PB|unc-13-4A-PC', map_location='65E10-65F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7266192, end_position_on_the_genomic_accession=7312896, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  39235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39235, CurrentID=0, Status='live', Symbol='CG32069', Aliases='Dmel_ CG6316, Dmel\\CG32069', description='CG32069 gene product from transcript CG32069-RA', other_designations='CG32069-PA|uncharacterized protein', map_location='68A3-68A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11062131, end_position_on_the_genomic_accession=11062614, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740436, CurrentID=0, Status='live', Symbol='CG34298', Aliases='Dmel_ Dmel\\CG34298', description='CG34298 gene product from transcript CG34298-RA', other_designations='CG34298-PA|uncharacterized protein', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29999795, end_position_on_the_genomic_accession=30000403, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38805, CurrentID=0, Status='live', Symbol='CG8607', Aliases='Dmel_ Dmel\\CG8607', description='CG8607 gene product from transcript CG8607-RA', other_designations='CG8607-PA', map_location='65F4-65F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7320068, end_position_on_the_genomic_accession=7321211, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38806, CurrentID=0, Status='live', Symbol='RhoGEF4', Aliases='Dmel_CG8606, CG8606, DRhoGEF4, Dmel\\CG8606, RhoGEF', description='Rho guanine nucleotide exchange factor 4', other_designations='CG8606-PA|CG8606-PB|RhoGEF4-PA|RhoGEF4-PB', map_location='65F4-65F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7321099, end_position_on_the_genomic_accession=7323411, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38807, CurrentID=0, Status='live', Symbol='Rint1', Aliases='Dmel_CG8605, CG8605, Dmel\\CG8605, Rint-1, Tip20, rint1', description='RINT1 ortholog', other_designations='CG8605-PA|Rint1-PA', map_location='65F4-65F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7323619, end_position_on_the_genomic_accession=7326255, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38808, CurrentID=0, Status='live', Symbol='CG8602', Aliases='Dmel_ Dmel\\CG8602, anon-WO0118547.349, anon-WO0118547.382', description='CG8602 gene product from transcript CG8602-RA', other_designations='CG8602-PA|CG8602-PB', map_location='65F4-65F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7326273, end_position_on_the_genomic_accession=7328842, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38809, CurrentID=0, Status='live', Symbol='mus312', Aliases='Dmel_CG8601, CG8601, Dmel\\CG8601, MUS312, mus(3)312', description='mutagen-sensitive 312', other_designations='CG8601-PA|CG8601-PD|CG8601-PE|mus312-PA|mus312-PD|mus312-PE', map_location='3-18 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7329132, end_position_on_the_genomic_accession=7333471, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740442, CurrentID=0, Status='live', Symbol='CG34356', Aliases='Dmel_ CG12524, CG14163, Dmel\\CG34356, Dmel_CG12524, Dmel_CG14163', description='CG34356 gene product from transcript CG34356-RF', other_designations='CG34356-PD|CG34356-PE|CG34356-PF|uncharacterized protein', map_location='67D2-67D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9908120, end_position_on_the_genomic_accession=9967298, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  38811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38811, CurrentID=0, Status='live', Symbol='lark', Aliases='Dmel_CG8597, CG8597, Dmel\\CG8597, LARK, Lark, dLARK', description='CG8597 gene product from transcript CG8597-RD', other_designations='CG8597-PA|CG8597-PB|CG8597-PC|CG8597-PD|CG8597-PE|lark-PA|lark-PB|lark-PC|lark-PD|lark-PE', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7339266, end_position_on_the_genomic_accession=7341893, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38812, CurrentID=0, Status='live', Symbol='eco', Aliases='Dmel_CG8598, CG8598, Deco, Dmel\\CG8598, ECO, Eco1, deco', description='Establishment of cohesion', other_designations='CG8598-PA|CG8598-PB|drosophila eco1|eco-PA|eco-PB|establishment of cohesion', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7335442, end_position_on_the_genomic_accession=7339340, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38813, CurrentID=0, Status='live', Symbol='CG8596', Aliases='Dmel_ Dmel\\CG8596, cln7', description='CG8596 gene product from transcript CG8596-RA', other_designations='CG8596-PA|CG8596-PB|CG8596-PC', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7341884, end_position_on_the_genomic_accession=7345183, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  5740446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740446, CurrentID=0, Status='live', Symbol='Rgk3', Aliases='Dmel_CG34397, CG15660, CG15663, CG15664, CG33949, CG34397, Dmel\\CG34397, Dmel_CG15660, Dmel_CG33949, RGK3, mdcds_15673, mdcds_15683', description='Rad, Gem/Kir family member 3', other_designations='CG34397-PC|CG34397-PE|CG34397-PF|CG34397-PG|Rgk3-PC|Rgk3-PE|Rgk3-PF|Rgk3-PG', map_location='57D5-57D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21255165, end_position_on_the_genomic_accession=21280995, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38815, CurrentID=0, Status='live', Symbol='Srp19', Aliases='Dmel_CG4457, CG4457, Dmel\\CG4457, SRP19, srp19', description='Signal recognition particle protein 19', other_designations='CG4457-PA|Srp19-PA|signal recognition particle 19 kD protein|signal recognition particle protein 19', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7347310, end_position_on_the_genomic_accession=7348034, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38816, CurrentID=0, Status='live', Symbol='qm', Aliases='Dmel_CG8593, CG8593, Dmel\\CG8593, GGPPS, GGPPS/qm, bro2', description='quemao', other_designations='CG8593-PA|CG8593-PB|CG8593-PC|geranylgeranyl pyrophosphate synthase|qm-PA|qm-PB|qm-PC|que mao', map_location='3-23.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7348313, end_position_on_the_genomic_accession=7353622, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38817, CurrentID=0, Status='live', Symbol='CTCF', Aliases='Dmel_CG8591, CG8591, Dmel\\CG8591, d', description='CG8591 gene product from transcript CG8591-RA', other_designations='CCCTC-binding factor|CG8591-PA|CTCF-PA', map_location='65F6-65F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7353578, end_position_on_the_genomic_accession=7356713, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38818, CurrentID=0, Status='live', Symbol='pst', Aliases='Dmel_CG8588, CG8588, Dm-65, Dmel\\CG8588', description='pastrel', other_designations='CG8588-PE|CG8588-PF|CG8588-PG|pst-PE|pst-PF|pst-PG', map_location='65F6-65F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7357275, end_position_on_the_genomic_accession=7360263, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38819, CurrentID=0, Status='live', Symbol='Sec63', Aliases='Dmel_CG8583, CG8583, Dmel\\CG8583, anon-WO0118547.333, peg, sec63', description='Sec63 ortholog (S. cerevisiae)', other_designations='CG8583-PA|Sec63 ortholog|Sec63-PA|sec63', map_location='65F7-65F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7360561, end_position_on_the_genomic_accession=7364366, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38820, CurrentID=0, Status='live', Symbol='Sh3beta', Aliases='Dmel_CG8582, CG8582, Dmel\\CG8582, LD12912, SH3beta', description='CG8582 gene product from transcript CG8582-RD', other_designations='CG8582-PA|CG8582-PB|CG8582-PC|CG8582-PD|CG8582-PE|Sh3beta-PA|Sh3beta-PB|Sh3beta-PC|Sh3beta-PD|Sh3beta-PE', map_location='65F7-65F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7364584, end_position_on_the_genomic_accession=7369260, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38821, CurrentID=0, Status='live', Symbol='akirin', Aliases='Dmel_CG8580, Akirin, CG8580, DmAkirin, Dmel\\CG8580, EP(3)0906, anon-WO0118547.397, bhr', description='CG8580 gene product from transcript CG8580-RD', other_designations='CG8580-PA|CG8580-PB|CG8580-PC|CG8580-PD|CG8580-PE|CG8580-PF|akirin-PA|akirin-PB|akirin-PC|akirin-PD|akirin-PE|akirin-PF|bhringi', map_location='65F7-65F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7369843, end_position_on_the_genomic_accession=7373711, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38823, CurrentID=0, Status='live', Symbol='SMSr', Aliases='Dmel_CG32380, CG32380, CG8572, CG8576, Css3beta, Dmel\\CG32380, d dmSMSr', description='CG32380 gene product from transcript CG32380-RA', other_designations='CG32380-PA|CPE synthase|SMSr-PA', map_location='65F7-65F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7374050, end_position_on_the_genomic_accession=7381474, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38824, CurrentID=0, Status='live', Symbol='smid', Aliases='Dmel_CG8571, CG8571, Dmel\\CG8571, Smid, l(3)j6B8, l(3)j8A6, l(3)s2898', description='smallminded', other_designations='CG8571-PA|CG8571-PC|CG8571-PG|CG8571-PH|CG8571-PI|small-minded|smid-PA|smid-PC|smid-PG|smid-PH|smid-PI', map_location='65F9-65F10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7381614, end_position_on_the_genomic_accession=7385637, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38825, CurrentID=0, Status='live', Symbol='CG8564', Aliases='Dmel_ Dmel\\CG8564', description='CG8564 gene product from transcript CG8564-RC', other_designations='CG8564-PA|CG8564-PC', map_location='65F10-65F10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7385648, end_position_on_the_genomic_accession=7387914, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38826, CurrentID=0, Status='live', Symbol='CG8563', Aliases='Dmel_ Dmel\\CG8563', description='CG8563 gene product from transcript CG8563-RB', other_designations='CG8563-PA|CG8563-PB', map_location='65F11-65F11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7390742, end_position_on_the_genomic_accession=7396013, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38828, CurrentID=0, Status='live', Symbol='CG8562', Aliases='Dmel_ CT15463, Dmel\\CG8562', description='CG8562 gene product from transcript CG8562-RA', other_designations='CG8562-PA', map_location='65F11-65F11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7398285, end_position_on_the_genomic_accession=7399912, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38829, CurrentID=0, Status='live', Symbol='CG18417', Aliases='Dmel_ Dmel\\CG18417, carboxypeptidase b1', description='CG18417 gene product from transcript CG18417-RA', other_designations='CG18417-PA|uncharacterized protein', map_location='65F11-65F11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7400070, end_position_on_the_genomic_accession=7401671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38830, CurrentID=0, Status='live', Symbol='CG8560', Aliases='Dmel_ CT14378, Dmel\\CG8560', description='CG8560 gene product from transcript CG8560-RA', other_designations='CG8560-PA|CG8560-PB', map_location='65F11-65F11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7402332, end_position_on_the_genomic_accession=7404181, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38831, CurrentID=0, Status='live', Symbol='Rac2', Aliases='Dmel_CG8556, CG8556, D-Rac 2, D- DRac2, DRacB, DmRAC2, Dmel\\CG8556, Drac1b, Drac2, P48554, Rac, Rac GTPase, Rac-2, Rac1, Rac1b, RacB, rac2', description='CG8556 gene product from transcript CG8556-RB', other_designations='CG8556-PA|CG8556-PB|Rac|Rac GTPase|Rac2-PA|Rac2-PB', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7432762, end_position_on_the_genomic_accession=7435190, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38832, CurrentID=0, Status='live', Symbol='CG14835', Aliases='Dmel_ Dmel\\CG14835', description='CG14835 gene product from transcript CG14835-RA', other_designations='CG14835-PA|uncharacterized protein', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7436808, end_position_on_the_genomic_accession=7437466, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38833, CurrentID=0, Status='live', Symbol='sphinx2', Aliases='Dmel_CG32382, CG32382, CG8555, Dmel\\CG32382, SPH194, Sphinx2', description='CG32382 gene product from transcript CG32382-RC', other_designations='CG32382-PB|CG32382-PC|Sphinx2|sphinx2-PB|sphinx2-PC', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7439554, end_position_on_the_genomic_accession=7440733, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246634, CurrentID=0, Status='live', Symbol='CG30472', Aliases='Dmel_ Dmel\\CG30472', description='CG30472 gene product from transcript CG30472-RB', other_designations='CG30472-PB|uncharacterized protein', map_location='52A1-52A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15423185, end_position_on_the_genomic_accession=15424139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38835, CurrentID=0, Status='live', Symbol='ppk26', Aliases='Dmel_CG8546, CG8546, Dmel\\CG8546, bba', description='pickpocket 26', other_designations='CG8546-PC|balboa|ppk26-PC', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7456149, end_position_on_the_genomic_accession=7458608, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740468, CurrentID=0, Status='live', Symbol='CG34213', Aliases='Dmel_ Dmel\\CG34213', description='CG34213 gene product from transcript CG34213-RA', other_designations='CG34213-PA|uncharacterized protein', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23967124, end_position_on_the_genomic_accession=23967745, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38837, CurrentID=0, Status='live', Symbol='CG8543', Aliases='Dmel_ Dmel\\CG8543', description='CG8543 gene product from transcript CG8543-RA', other_designations='CG8543-PA', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7469502, end_position_on_the_genomic_accession=7470302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38838, CurrentID=0, Status='live', Symbol='CG8541', Aliases='Dmel_ Dmel\\CG8541', description='CG8541 gene product from transcript CG8541-RA', other_designations='CG8541-PA', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7471559, end_position_on_the_genomic_accession=7472514, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38839, CurrentID=0, Status='live', Symbol='mthl6', Aliases='Dmel_CG16992, CG16992, Dmel\\CG16992, anon-WO0170980.127, anon-WO0170980.128', description='methuselah-like 6', other_designations='CG16992-PA|mth-like 6|mthl6-PA', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7485967, end_position_on_the_genomic_accession=7487918, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740472, CurrentID=0, Status='live', Symbol='Fili', Aliases='Dmel_CG34368, CG13487, CG34368, CG4054, CT13464, CT32855, Dmel\\CG34368, Dmel_CG13487, Dmel_CG4045, Dmel_CG4054', description='Fish-lips', other_designations='CG34368-PC|CG34368-PD|Fili-PC|Fili-PD', map_location='58A3-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21872030, end_position_on_the_genomic_accession=21946259, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38841, CurrentID=0, Status='live', Symbol='Cyp4d8', Aliases='Dmel_CG4321, 4d8, CG4321, Dmel\\CG4321, P-450, P450', description='Cytochrome P450-4d8', other_designations='CG4321-PA|Cyp4d8-PA|cytochrome P450|cytochrome P450-4d8', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7493557, end_position_on_the_genomic_accession=7496179, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38842, CurrentID=0, Status='live', Symbol='CG8539', Aliases='Dmel_ Dmel\\CG8539', description='CG8539 gene product from transcript CG8539-RB', other_designations='CG8539-PA|CG8539-PB', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7497211, end_position_on_the_genomic_accession=7498831, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38843, CurrentID=0, Status='live', Symbol='CG7548', Aliases='Dmel_ Dmel\\CG7548', description='CG7548 gene product from transcript CG7548-RA', other_designations='CG7548-PA', map_location='66A3-66A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7505540, end_position_on_the_genomic_accession=7506217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38844, CurrentID=0, Status='live', Symbol='CG7546', Aliases='Dmel_ Dmel\\CG7546, anon-EST:Liang-2.45, clone 2.45', description='CG7546 gene product from transcript CG7546-RH', other_designations='CG7546-PA|CG7546-PB|CG7546-PC|CG7546-PD|CG7546-PE|CG7546-PF|CG7546-PG|CG7546-PH', map_location='66A3-66A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7510986, end_position_on_the_genomic_accession=7517316, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  38845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38845, CurrentID=0, Status='live', Symbol='ltl', Aliases='Dmel_CG32372, CG14836, CG32372, CG7545, CT15760, Dmel\\CG32372, F, anon-EST:CL1a11, l(3)tr, lincRNA.S4366', description='larval translucida', other_designations='CG32372-PA|CG32372-PB|lethal (3) translucida|lethal translucida|ltl-PA|ltl-PB', map_location='66A3-66A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7518997, end_position_on_the_genomic_accession=7528013, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38846, CurrentID=0, Status='live', Symbol='lqf', Aliases='Dmel_CG8532, 0110/27, CG8532, Dmel\\CG8532, E(faf)bE25, Epsin, LqF, Lqf, anon-WO0118547.239, dEpsin, epsin, l(3)S011027', description='liquid facets', other_designations='CG8532-PA|CG8532-PB|CG8532-PC|CG8532-PD|CG8532-PE|CG8532-PF|CG8532-PG|CG8532-PH|CG8532-PI|CG8532-PJ|CG8532-PK|CG8532-PL|Epsin1/Liquid facets|drosophila epsin|enhancer of faf bE25|epsin|epsin 1|lqf-PA|lqf-PB|lqf-PC|lqf-PD|lqf-PE|lqf-PF|lqf-PG|lqf-PH|lqf-PI|lqf-PJ|lqf-PK|lqf-PL', map_location='66A4-66A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7528780, end_position_on_the_genomic_accession=7540012, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  38847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38847, CurrentID=0, Status='live', Symbol='CG16998', Aliases='Dmel_ Dmel\\CG16998, SP104', description='CG16998 gene product from transcript CG16998-RA', other_designations='CG16998-PA|uncharacterized protein', map_location='66A5-66A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7539060, end_position_on_the_genomic_accession=7539836, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38848, CurrentID=0, Status='live', Symbol='CG32374', Aliases='Dmel_ CG8528, Dmel\\CG32374, SP148', description='CG32374 gene product from transcript CG32374-RA', other_designations='CG32374-PA|uncharacterized protein', map_location='66A5-66A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7540269, end_position_on_the_genomic_accession=7541168, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740481, CurrentID=0, Status='live', Symbol='CG34198', Aliases='Dmel_ Dmel\\CG34198', description='CG34198 gene product from transcript CG34198-RA', other_designations='CG34198-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19682446, end_position_on_the_genomic_accession=19682977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38850, CurrentID=0, Status='live', Symbol='frac', Aliases='Dmel_CG7526, CG7526, CT16044, Dmel\\CG7526', description='faulty attraction', other_designations='CG7526-PC|CG7526-PD|CG7526-PE|CG7526-PF|CG7526-PG|frac-PC|frac-PD|frac-PE|frac-PF|frac-PG', map_location='66A5-66A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7562610, end_position_on_the_genomic_accession=7569499, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38851, CurrentID=0, Status='live', Symbol='CG14838', Aliases='Dmel_ Dmel\\CG14838', description='CG14838 gene product from transcript CG14838-RA', other_designations='CG14838-PA|uncharacterized protein', map_location='66A6-66A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7583454, end_position_on_the_genomic_accession=7587248, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38852, CurrentID=0, Status='live', Symbol='CG7716', Aliases='Dmel_ Dmel\\CG7716', description='CG7716 gene product from transcript CG7716-RA', other_designations='CG7716-PA', map_location='66A6-66A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7588608, end_position_on_the_genomic_accession=7590998, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740488, CurrentID=0, Status='live', Symbol='CG34263', Aliases='Dmel_ Dmel\\CG34263', description='CG34263 gene product from transcript CG34263-RA', other_designations='CG34263-PA|uncharacterized protein', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=265595, end_position_on_the_genomic_accession=265908, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38857, CurrentID=0, Status='live', Symbol='CG7506', Aliases='Dmel_ Dmel\\CG7506', description='CG7506 gene product from transcript CG7506-RA', other_designations='CG7506-PA', map_location='66A8-66A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7640798, end_position_on_the_genomic_accession=7641648, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38858, CurrentID=0, Status='live', Symbol='PGRP-SD', Aliases='Dmel_CG7496, CG7496, Dm  Dmel\\CG7496, pgrp-sd', description='CG7496 gene product from transcript CG7496-RA', other_designations='CG7496-PA|PGRP-SD-PA|Peptidoglycan-recognition protein-SD precursor', map_location='66A8-66A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7651180, end_position_on_the_genomic_accession=7651900, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38859, CurrentID=0, Status='live', Symbol='CG7492', Aliases='Dmel_ DPLG4, Dmel\\CG7492, anon-EST:fe1A11', description='CG7492 gene product from transcript CG7492-RA', other_designations='CG7492-PA|PIF-like gene 4|anon-fast-evolving-1A11', map_location='66A8-66A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7651866, end_position_on_the_genomic_accession=7654752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740493, CurrentID=0, Status='live', Symbol='CG34258', Aliases='Dmel_ Dmel\\CG34258', description='CG34258 gene product from transcript CG34258-RA', other_designations='CG34258-PA|uncharacterized protein', map_location='76D6-76D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19951235, end_position_on_the_genomic_accession=19951968, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38863, CurrentID=0, Status='live', Symbol='Ank2', Aliases='Dmel_CG42734, CG18025, CG18227, CG18768, CG32373, CG32375, CG32377, CG32377/CG34416, CG34121, CG34158, CG34416, CG42734, CG7462, CG7493, DAnk2, Dank2, Dmel\\CG42734, Dmel_CG32375, Dmel_CG32377, Dmel_CG34121, Dmel_CG34158, Dmel_CG34416, Dmel_CG7462, ESTS:149B10S, SP2523, ank2, ank2-L, dank2, sp2523', description='Ankyrin 2', other_designations='Ank2-PAA|Ank2-PE|Ank2-PF|Ank2-PG|Ank2-PJ|Ank2-PK|Ank2-PL|Ank2-PM|Ank2-PN|Ank2-PP|Ank2-PQ|Ank2-PR|Ank2-PS|Ank2-PT|Ank2-PU|Ank2-PV|Ank2-PW|Ank2-PX|Ank2-PY|Ank2-PZ|CG42734-PAA|CG42734-PE|CG42734-PF|CG42734-PG|CG42734-PJ|CG42734-PK|CG42734-PL|CG42734-PM|CG42734-PN|CG42734-PP|CG42734-PQ|CG42734-PR|CG42734-PS|CG42734-PT|CG42734-PU|CG42734-PV|CG42734-PW|CG42734-PX|CG42734-PY|CG42734-PZ|Dank2|SP2523|ankyrin 2|ankyrin2|ankyrin2-XL', map_location='66A10-66A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7655389, end_position_on_the_genomic_accession=7718395, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  38864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38864, CurrentID=0, Status='live', Symbol='CG12262', Aliases='Dmel_ ACAD, CG 12262, Dmel\\CG12262, MCAD', description='CG12262 gene product from transcript CG12262-RA', other_designations='CG12262-PA|acyl-CoA dehdrogenase|medium chain acyl-CoA dehydrogenase|uncharacterized protein', map_location='66A10-66A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7719350, end_position_on_the_genomic_accession=7721472, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740497, CurrentID=0, Status='live', Symbol='inaF-C', Aliases='Dmel_CG34321, CG34321, Dmel\\CG34321, INAF, InaF, inaF', description='CG34321 gene product from transcript CG34321-RA', other_designations='CG34321-PA|inaF-C-PA', map_location='10D8-10E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11726390, end_position_on_the_genomic_accession=11731765, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38866, CurrentID=0, Status='live', Symbol='CG8492', Aliases='Dmel_ Dmel\\CG8492', description='CG8492 gene product from transcript CG8492-RD', other_designations='CG8492-PD', map_location='66A11-66A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7727804, end_position_on_the_genomic_accession=7732191, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38868, CurrentID=0, Status='live', Symbol='Snmp2', Aliases='Dmel_CG7422, 148358_at, CG16832, CG7422, Dmel\\CG7422, SNMP2', description='Sensory neuron membrane protein 2', other_designations='CG7422-PB|CG7422-PC|CG7422-PD|Snmp2-PB|Snmp2-PC|Snmp2-PD|sensory neuron membrane protein 2', map_location='66A11-66A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7733250, end_position_on_the_genomic_accession=7743041, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  5740501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740501, CurrentID=0, Status='live', Symbol='CG34253', Aliases='Dmel_ CG32194, Dmel\\CG34253', description='CG34253 gene product from transcript CG34253-RA', other_designations='CG34253-PA|uncharacterized protein', map_location='75B7-75B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18086674, end_position_on_the_genomic_accession=18087320, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38870, CurrentID=0, Status='live', Symbol='CG7409', Aliases='Dmel_ CT16169, Dmel\\CG7409', description='CG7409 gene product from transcript CG7409-RA', other_designations='CG7409-PA', map_location='66A12-66A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7754409, end_position_on_the_genomic_accession=7755168, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38871, CurrentID=0, Status='live', Symbol='Hn', Aliases='Dmel_CG7399, CG7399, DTPH, Dmel\\CG7399, PAH, Pah, TPH, TRH, Tph, Trh, bu, pah', description='Henna', other_designations='CG7399-PA|CG7399-PB|CG7399-PC|CG7399-PD|Hn-PA|Hn-PB|Hn-PC|Hn-PD|Phenylalanine-hydroxylase|brunette|henna|phenylalanine hydroxylase|phenylalanine-tryptophan hydroxylase|tryptophan hydrolase', map_location='3-23.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7760453, end_position_on_the_genomic_accession=7763166, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38872, CurrentID=0, Status='live', Symbol='Clk', Aliases='Dmel_CG7391, CG7391, CLK, CLOCK, Dmel\\CG7391, Jerk, Jrk, PAS1, bHLHe10, clk, clock, dCLK, dCLK/JRK, dCLOCK, dClck, dClk, dClock, dclk, jrk', description='Clock', other_designations='CG7391-PA|CG7391-PD|CG7391-PF|CG7391-PG|CG7391-PH|CLOCK|Clk-PA|Clk-PD|Clk-PF|Clk-PG|Clk-PH|clock|dClock', map_location='66A12-66A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7763233, end_position_on_the_genomic_accession=7775603, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  5740505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740505, CurrentID=0, Status='live', Symbol='Lgr4', Aliases='Dmel_CG34411, CG32637, CG34411, CG4187, DLGR4, Dmel\\CG34411, Dmel_CG32637, Dmel_CG4187, anon-WO0170980.19, anon-WO0170980.20, dLGR4', description='Leucine-rich repeat-containing G protein-coupled receptor 4', other_designations='CG34411-PB|CG34411-PC|Lgr4-PB|Lgr4-PC', map_location='11E2-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13161714, end_position_on_the_genomic_accession=13189164, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  5740507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740507, CurrentID=0, Status='live', Symbol='CG34259', Aliases='Dmel_ Dmel\\CG34259', description='CG34259 gene product from transcript CG34259-RB', other_designations='CG34259-PB|uncharacterized protein', map_location='62C2-62C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2026991, end_position_on_the_genomic_accession=2028326, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38876, CurrentID=0, Status='live', Symbol='CG32365', Aliases='Dmel_ CG13682, CG13683, CG8294, Dmel\\CG32365', description='CG32365 gene product from transcript CG32365-RB', other_designations='CG32365-PA|CG32365-PB|uncharacterized protein', map_location='66A17-66A18', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7869042, end_position_on_the_genomic_accession=7894814, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38878, CurrentID=0, Status='live', Symbol='eIF4E-5', Aliases='Dmel_CG8277, CG8277, CT24507, Dmel\\CG8277, eIF4E', description='CG8277 gene product from transcript CG8277-RA', other_designations='CG8277-PA|eIF4E-5-PA', map_location='66A18-66A18', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7894880, end_position_on_the_genomic_accession=7896009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38879, CurrentID=0, Status='live', Symbol='pbl', Aliases='Dmel_CG8114, 0083/20, 0293/09, 0542/03, CG8114, Dmel\\CG8114, ECT2, Ect-2, GEF, PBL, Pbl, anon-WO0172774.22, anon-WO0172774.25, l(3)09645', description='pebble', other_designations='CG8114-PA|CG8114-PB|CG8114-PC|CG8114-PD|CG8114-PE|CG8114-PF|pbl-PA|pbl-PB|pbl-PC|pbl-PD|pbl-PE|pbl-PF', map_location='3-26 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7896109, end_position_on_the_genomic_accession=7911954, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  38880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38880, CurrentID=0, Status='live', Symbol='CG8281', Aliases='Dmel_ Dmel\\CG8281', description='CG8281 gene product from transcript CG8281-RA', other_designations='CG8281-PA', map_location='66A19-66A19', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7912463, end_position_on_the_genomic_accession=7913763, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740513, CurrentID=0, Status='live', Symbol='CG44837', Aliases='Dmel_ CG14126, CG14129, CG34420, CG5917, Dmel\\CG44837, Dmel_CG14126, Dmel_CG14129, Dmel_CG34420, Dmel_CG5917', description='CG44837 gene product from transcript CG44837-RB', other_designations='CG44837-PB|CG44837-PC|CG44837-PD|rugose|uncharacterized protein', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11850645, end_position_on_the_genomic_accession=11938872, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38883, CurrentID=0, Status='live', Symbol='Srp9', Aliases='Dmel_CG8268, CG8268, Dmel\\CG8268', description='Signal recognition particle protein 9', other_designations='CG8268-PA|Srp9-PA|signal recognition particle protein 9', map_location='66A20-66A20', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7936373, end_position_on_the_genomic_accession=7937258, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38884, CurrentID=0, Status='live', Symbol='exex', Aliases='Dmel_CG8254, CG8254, DHB9, Dmel\\CG8254, HB9, Hb9, dHb9, dhb9, hb9', description='extra-extra', other_designations='CG8254-PA|exex-PA', map_location='66A21-66A21', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7948186, end_position_on_the_genomic_accession=7951563, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38885, CurrentID=0, Status='live', Symbol='RNaseX25', Aliases='Dmel_CG8194, CG8194, DmRNase-66B, DmRNase-X25, Dmel\\CG8194', description='Ribonuclease X25', other_designations='CG8194-PA|RNaseX25-PA|acid RNase|ribonuclease X25', map_location='66A21-66A21', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7970588, end_position_on_the_genomic_accession=7972449, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38886, CurrentID=0, Status='live', Symbol='CG8042', Aliases='Dmel_ BcDNA:GH10229, Dmel\\CG8042, anon-WO0140519.245, anon-WO0172774.24', description='CG8042 gene product from transcript CG8042-RA', other_designations='CG8042-PA|UBX domain 2 like', map_location='66A21-66A21', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7972449, end_position_on_the_genomic_accession=7974991, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38887, CurrentID=0, Status='live', Symbol='HP4', Aliases='Dmel_CG8044, BcDNA:GM01628, CG8044, Dmel\\CG8044, Hip, hip', description='Heterochromatin protein 4', other_designations='CG8044-PA|CG8044-PB|HP1-interacting protein|HP4-PA|HP4-PB|heterochromatin protein 4', map_location='66A21-66A21', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7975438, end_position_on_the_genomic_accession=7976154, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38888, CurrentID=0, Status='live', Symbol='CG8209', Aliases='Dmel_ Dmel\\CG8209, anon-WO0172774.138', description='CG8209 gene product from transcript CG8209-RA', other_designations='CG8209-PA|UBA/UBX 33.3 kDa protein like', map_location='66A21-66A21', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7976406, end_position_on_the_genomic_accession=7977849, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38889, CurrentID=0, Status='live', Symbol='CG8038', Aliases='Dmel_ BcDNA:LD45688, Dmel\\CG8038, anon-WO0172774.38', description='CG8038 gene product from transcript CG8038-RA', other_designations='CG8038-PA', map_location='66A21-66A22', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7977805, end_position_on_the_genomic_accession=7978650, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38890, CurrentID=0, Status='live', Symbol='nmo', Aliases='Dmel_CG7892, CG7892, Dmel\\CG7892, NEMO, NEMO/NLK, NLK, Nemo, Nlk, Nmo, ORE-6, adk, adk1, anon-EST:Gibbs3, anon-EST:Liang-2.43, anon-WO0118547.332, anon-WO0140519.256, anon-WO0172774.138, clone 2.43', description='nemo', other_designations='CG7892-PA|CG7892-PB|CG7892-PC|CG7892-PE|CG7892-PF|CG7892-PG|CG7892-PH|NEMO|Nemo-like kinase|adirondack|dNEMO|nmo-PA|nmo-PB|nmo-PC|nmo-PE|nmo-PF|nmo-PG|nmo-PH', map_location='66A22-66B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7979046, end_position_on_the_genomic_accession=8054147, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  38891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38891, CurrentID=0, Status='live', Symbol='bip1', Aliases='Dmel_CG7574, Bip1, CG7574, Dmel\\CG7574, anon-WO0118547.246, bip-I', description='CG7574 gene product from transcript CG7574-RA', other_designations='Bab interating protein 1|CG7574-PA|CG7574-PB|bip1-PA|bip1-PB', map_location='66B3-66B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8054333, end_position_on_the_genomic_accession=8058713, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38893, CurrentID=0, Status='live', Symbol='CG7565', Aliases='Dmel_ CT23145, Dmel\\CG7565', description='CG7565 gene product from transcript CG7565-RA', other_designations='CG7565-PA', map_location='66B5-66B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8087635, end_position_on_the_genomic_accession=8091786, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38895, CurrentID=0, Status='live', Symbol='Ect4', Aliases='Dmel_CG43119, CG13680, CG13681, CG34373, CG43119, CG7905, CG7915, Dmel\\CG43119, Dmel_CG13680, Dmel_CG13681, Dmel_CG34373, Dmel_CG7915/SARM, anon-WO0118547.272, dsarm, ect4, Ect4', description='Ectoderm-expressed 4', other_designations='CG43119-PD|CG43119-PE|CG43119-PF|CG43119-PG|CG43119-PH|CG43119-PI|CG43119-PJ|CG43119-PK|Ect4-PD|Ect4-PE|Ect4-PF|Ect4-PG|Ect4-PH|Ect4-PI|Ect4-PJ|Ect4-PK|drosophila sterile alpha and armadillo motif', map_location='66B5-66B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8063874, end_position_on_the_genomic_accession=8108897, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  5740528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740528, CurrentID=0, Status='live', Symbol='CG34354', Aliases='Dmel_ CG12425, CG12852, CG4787, Dmel\\CG34354, Dmel_CG12425, Dmel_CG12852, Dmel_CG4787, cg4787', description='CG34354 gene product from transcript CG34354-RE', other_designations='CG34354-PD|CG34354-PE|CG34354-PF|CG34354-PG|uncharacterized protein', map_location='98C2-98C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28146880, end_position_on_the_genomic_accession=28240526, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38897, CurrentID=0, Status='live', Symbol='CG7927', Aliases='Dmel_ BcDNA:GH11023, Dmel\\CG7927, anon-WO0172774.138', description='CG7927 gene product from transcript CG7927-RC', other_designations='CG7927-PA|CG7927-PB|CG7927-PC', map_location='66B6-66B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8110368, end_position_on_the_genomic_accession=8113406, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38898, CurrentID=0, Status='live', Symbol='Arp3', Aliases='Dmel_CG7558, ARP3, AcTr66B, Actr66B, Arp66B, CG7558, Dmel\\CG7558, anon-WO0172774.138, arp3, arp66B, arp66b, schwachling', description='Actin-related protein 3', other_designations='Actin-related protein|Actin-related protein 66B|Arp3-PA|Arp3-PB|CG7558-PA|CG7558-PB|actin-related protein-3|schwachling', map_location='66B6-66B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8113407, end_position_on_the_genomic_accession=8115902, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  5740883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740883, CurrentID=0, Status='live', Symbol='CG34248', Aliases='Dmel_ Dmel\\CG34248', description='CG34248 gene product from transcript CG34248-RB', other_designations='CG34248-PA|CG34248-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16270369, end_position_on_the_genomic_accession=16270902, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38900, CurrentID=0, Status='live', Symbol='ldbr', Aliases='Dmel_CG7942, CG7942, Dmel\\CG7942, Lbdr, Ldbr', description='lariat debranching enzyme', other_designations='CG7942-PB|lariat-debranching enzyme|ldbr-PB', map_location='66B6-66B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8121827, end_position_on_the_genomic_accession=8123615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38901, CurrentID=0, Status='live', Symbol='CG7550', Aliases='Dmel_ Dmel\\CG7550', description='CG7550 gene product from transcript CG7550-RA', other_designations='CG7550-PA', map_location='66B7-66B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8123612, end_position_on_the_genomic_accession=8124531, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740534, CurrentID=0, Status='live', Symbol='CG34194', Aliases='Dmel_ Dmel\\CG34194', description='CG34194 gene product from transcript CG34194-RB', other_designations='CG34194-PA|CG34194-PB|uncharacterized protein', map_location='54C3-54C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17544190, end_position_on_the_genomic_accession=17545045, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38904, CurrentID=0, Status='live', Symbol='CG7504', Aliases='Dmel_ Dmel\\CG7504', description='CG7504 gene product from transcript CG7504-RC', other_designations='CG7504-PB|CG7504-PC', map_location='66B7-66B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8131429, end_position_on_the_genomic_accession=8137498, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38905, CurrentID=0, Status='live', Symbol='mus301', Aliases='Dmel_CG7972, CG7972, Dmel\\CG7972, Mus301, SpnC, anon-WO0118547.354, cg7972, mus(3)301, spn-C, spnC', description='mutagen-sensitive 301', other_designations='CG7972-PA|mus301-PA|spindel|spindle C|spindle-C', map_location='3-23 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8137808, end_position_on_the_genomic_accession=8142014, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38906, CurrentID=0, Status='live', Symbol='CG13676', Aliases='Dmel_ Dmel\\CG13676', description='CG13676 gene product from transcript CG13676-RB', other_designations='CG13676-PA|CG13676-PB|uncharacterized protein', map_location='66B8-66B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8141868, end_position_on_the_genomic_accession=8151813, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38907, CurrentID=0, Status='live', Symbol='CG13675', Aliases='Dmel_ Dmel\\CG13675', description='CG13675 gene product from transcript CG13675-RD', other_designations='CG13675-PC|CG13675-PD|uncharacterized protein', map_location='66B10-66B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8160315, end_position_on_the_genomic_accession=8165518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38909, CurrentID=0, Status='live', Symbol='Nmt', Aliases='Dmel_CG7436, 153290_at, CG7436, Dmel\\CG74361, dNMT, l(3)j1C7, l(3)j1c7, nmt, Nmt', description='N-myristoyl transferase', other_designations='CG7436-PA|Nmt-PA', map_location='66B10-66B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8176135, end_position_on_the_genomic_accession=8178429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38910, CurrentID=0, Status='live', Symbol='mthl7', Aliases='Dmel_CG7476, CG7476, Dmel\\CG7476, anon-WO0170980.85, anon-WO0170980.86', description='methuselah-like 7', other_designations='CG7476-PA|CG7476-PB|Mth-like 7|mthl7-PA|mthl7-PB', map_location='66B11-66B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8178675, end_position_on_the_genomic_accession=8180514, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38911, CurrentID=0, Status='live', Symbol='CG7979', Aliases='Dmel_ Dmel\\CG7979', description='CG7979 gene product from transcript CG7979-RA', other_designations='CG7979-PA', map_location='66B11-66B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8180923, end_position_on_the_genomic_accession=8183369, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38912, CurrentID=0, Status='live', Symbol='ERR', Aliases='Dmel_CG7404, CG7404, Dmel\\CG7404, NR3B4, dERR', description='estrogen-related receptor', other_designations='CG7404-PA|CG7404-PB|ERR-PA|ERR-PB|Estrogen-Related receptor|estrogen related receptor', map_location='66B11-66B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8183361, end_position_on_the_genomic_accession=8186671, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38913, CurrentID=0, Status='live', Symbol='Atg18a', Aliases='Dmel_CG7986, ATG18, ATG18.1, Atg18, CG7986, Dmel\\CG7986, atg18', description='Autophagy-related 18a', other_designations='Atg18a-PA|Atg18a-PB|Atg18a-PC|Atg18a-PD|Atg18a-PE|Atg18a-PF|Atg18a-PG|Autophagy-specific gene 18|Autophagy-specific gene 18a|CG7986-PA|CG7986-PB|CG7986-PC|CG7986-PD|CG7986-PE|CG7986-PF|CG7986-PG', map_location='66B11-66B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8186857, end_position_on_the_genomic_accession=8191118, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38914, CurrentID=0, Status='live', Symbol='MED24', Aliases='Dmel_CG7999, CG7999, Dmel\\CG7999, Med24, Trap100, dTRAP100, l(3)psg5, med24, psg5', description='Mediator complex subunit 24', other_designations='CG7999-PA|MED24-PA|mediator|mediator complex subunit 24', map_location='3-34 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8191270, end_position_on_the_genomic_accession=8196573, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740547, CurrentID=0, Status='live', Symbol='CG34461', Aliases='Dmel_ CG7072, Cpr66Ca, DmelCpr66Ca, Dmel\\CG34461, Dmel_CG7072, anon-SAGE:Wang-027', description='CG34461 gene product from transcript CG34461-RB', other_designations='CG34461-PA|CG34461-PB|anon-SAGE:Wang-027|uncharacterized protein', map_location='66C7-66C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8325859, end_position_on_the_genomic_accession=8328307, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38916, CurrentID=0, Status='live', Symbol='UbcE2M', Aliases='Dmel_CG7375, CG7375, Dmel\\CG7375, Ubc12, UbcD12, anon-WO0118547.309, dUbc12', description='Ubiquitin conjugating enzyme E2M', other_designations='CG7375-PA|CG7375-PB|CG7375-PC|UbcE2M-PA|UbcE2M-PB|UbcE2M-PC|ubiquitin conjugating enzyme 12|ubiquitin conjugating enzyme E2M', map_location='66B12-66B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8195847, end_position_on_the_genomic_accession=8197987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38917, CurrentID=0, Status='live', Symbol='CG8005', Aliases='Dmel_ Dmel\\CG8005', description='CG8005 gene product from transcript CG8005-RA', other_designations='CG8005-PA|CG8005-PB', map_location='66B12-66B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8198207, end_position_on_the_genomic_accession=8199491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38918, CurrentID=0, Status='live', Symbol='CG7366', Aliases='Dmel_ Dmel\\CG7366, anon-WO0140519.185', description='CG7366 gene product from transcript CG7366-RA', other_designations='CG7366-PA', map_location='66B12-66B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8201078, end_position_on_the_genomic_accession=8202417, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38919, CurrentID=0, Status='live', Symbol='CG13679', Aliases='Dmel_ BcDNA:RE17113, Dmel\\CG13679', description='CG13679 gene product from transcript CG13679-RB', other_designations='CG13679-PB|uncharacterized protein', map_location='66B12-66B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8202792, end_position_on_the_genomic_accession=8203345, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38920, CurrentID=0, Status='live', Symbol='CG8006', Aliases='Dmel_ Dmel\\CG8006', description='CG8006 gene product from transcript CG8006-RA', other_designations='CG8006-PA', map_location='66C1-66C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8211890, end_position_on_the_genomic_accession=8213622, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38921, CurrentID=0, Status='live', Symbol='CG13674', Aliases='Dmel_ Dmel\\CG13674', description='CG13674 gene product from transcript CG13674-RA', other_designations='CG13674-PA|CG13674-PB|uncharacterized protein', map_location='66C1-66C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8218232, end_position_on_the_genomic_accession=8218983, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740554, CurrentID=0, Status='live', Symbol='CG34448', Aliases='Dmel_ CG31684, CG9966, Dmel\\CG34448, Dmel_CG31684', description='CG34448 gene product from transcript CG34448-RA', other_designations='CG34448-PA|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2416792, end_position_on_the_genomic_accession=2417965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38923, CurrentID=0, Status='live', Symbol='eIF4E-3', Aliases='Dmel_CG8023, CG8023, Dmel\\CG8023, anon-WO0140519.227, eIF4E, elF4E-3, l(3)0139, l(3)L0139', description='eukaryotic translation initiation factor 4E-3', other_designations='CG8023-PA|eIF4E-3-PA|lethal (3) L0139', map_location='66C1-66C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8229224, end_position_on_the_genomic_accession=8230568, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38924, CurrentID=0, Status='live', Symbol='CG7213', Aliases='Dmel_ Dmel\\CG7213', description='CG7213 gene product from transcript CG7213-RA', other_designations='CG7213-PA', map_location='66C1-66C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8231814, end_position_on_the_genomic_accession=8240601, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740558, CurrentID=0, Status='live', Symbol='CG34230', Aliases='Dmel_ Dmel\\CG34230', description='CG34230 gene product from transcript CG34230-RA', other_designations='CG34230-PA|uncharacterized protein', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11861310, end_position_on_the_genomic_accession=11862015, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38928, CurrentID=0, Status='live', Symbol='cert', Aliases='Dmel_CG7207, BcDNA:GH07688, CERT, CG7207, Dcert, Dmel\\CG7207, dcert', description='ceramide transfer protein', other_designations='CG7207-PA|CG7207-PB|CG7207-PC|cert-PA|cert-PB|cert-PC', map_location='66C4-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8282528, end_position_on_the_genomic_accession=8289870, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  38929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38929, CurrentID=0, Status='live', Symbol='CG7201', Aliases='Dmel_ Dmel\\CG7201', description='CG7201 gene product from transcript CG7201-RA', other_designations='CG7201-PA|CG7201-PB', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8290286, end_position_on_the_genomic_accession=8295327, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38930, CurrentID=0, Status='live', Symbol='Rab19', Aliases='Dmel_CG7062, AAF50452, CG7062, DRABR3, DRabRP3, DmRab19, Dmel\\CG7062, O18339, Rab-RP3, Rab-r3/RP3, Rab19s, rab19', description='CG7062 gene product from transcript CG7062-RB', other_designations='CG7062-PA|CG7062-PB|Rab19-PA|Rab19-PB|rab-related protein 3', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8297018, end_position_on_the_genomic_accession=8298506, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38931, CurrentID=0, Status='live', Symbol='Arl5', Aliases='Dmel_CG7197, CG7197, Dmel\\CG7197, Q9VSG8', description='ADP-ribosylation factor-like 5', other_designations='Arl5-PA|CG7197-PA', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8298501, end_position_on_the_genomic_accession=8299839, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38932, CurrentID=0, Status='live', Symbol='CG13671', Aliases='Dmel_ Dmel\\CG13671', description='CG13671 gene product from transcript CG13671-RA', other_designations='CG13671-PA|uncharacterized protein', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8300064, end_position_on_the_genomic_accession=8302341, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740565, CurrentID=0, Status='live', Symbol='CG34370', Aliases='Dmel_ BP1037, CG13497, CG13498, CG13499, CG30405, CG30408, CG33227, CG34004, CT32866, Dmel\\CG34370, Dmel_CG13499, Dmel_CG33227, Dmel_CG34004, tld-like', description='CG34370 gene product from transcript CG34370-RG', other_designations='CG34370-PG|CG34370-PH|uncharacterized protein', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21987068, end_position_on_the_genomic_accession=22039559, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  38934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38934, CurrentID=0, Status='live', Symbol='Sbp2', Aliases='Dmel_CG7066, CG7066, Dmel\\CG7066, SBP2', description='SECIS-binding protein 2', other_designations='CG7066-PA|SECIS binding protein 2|Sbp2-PA', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8304157, end_position_on_the_genomic_accession=8305325, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38935, CurrentID=0, Status='live', Symbol='Gr66a', Aliases='Dmel_CG7189, CG7189, Dmel\\CG7189, GR66A, GR66C1, GR66a, Gr66, Gr66A, Gr66C1', description='Gustatory receptor 66a', other_designations='CG7189-PB|Gr66a-PB|gustatory receptor 66C1|gustatory receptor 66a', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8305336, end_position_on_the_genomic_accession=8306980, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38936, CurrentID=0, Status='live', Symbol='BI-1', Aliases='Dmel_CG7188, CG7188, Dmel\\CG7188, d-BI-1, dBI-1, dTmbim6', description='Bax Inhibitor-1', other_designations='BI-1-PA|BI-1-PB|Bax inhibitor-1|CG7188-PA|CG7188-PB', map_location='66C5-66C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8307145, end_position_on_the_genomic_accession=8308941, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38937, CurrentID=0, Status='live', Symbol='CG7185', Aliases='Dmel_ CF I[[m]]68, CFlm-68k, Dmel\\CG7185, cg7185', description='CG7185 gene product from transcript CG7185-RB', other_designations='CG7185-PA|CG7185-PB', map_location='66C5-66C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8310493, end_position_on_the_genomic_accession=8315502, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38938, CurrentID=0, Status='live', Symbol='CG13670', Aliases='Dmel_ Dmel\\CG13670', description='CG13670 gene product from transcript CG13670-RA', other_designations='CG13670-PA|uncharacterized protein', map_location='66C7-66C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8320763, end_position_on_the_genomic_accession=8322280, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38940, CurrentID=0, Status='live', Symbol='Cpr66Cb', Aliases='Dmel_CG7076, CG7076, DmelCpr66Cb, Dmel\\CG7076', description='Cuticular protein 66Cb', other_designations='CG7076-PA|Cpr66Cb-PA|cuticular protein 66Cb', map_location='66C7-66C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8336672, end_position_on_the_genomic_accession=8338595, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38941, CurrentID=0, Status='live', Symbol='Pex2', Aliases='Dmel_CG7081, CG7081, DmelPex2, Dmel\\CG7081, PEX2, pex2', description='Peroxin 2', other_designations='CG7081-PA|Pex2-PA|peroxin 2', map_location='66C7-66C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8339051, end_position_on_the_genomic_accession=8340276, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38942, CurrentID=0, Status='live', Symbol='DNApol-alpha50', Aliases='Dmel_CG7108, CG7108, DNA polalpha, DNApol-alpha, DNApolalpha, DNApolalpha50, Dmel\\CG7108, Dmpol alpha, Dpri50, POLA, PRI1, Pol, Polalpha, Polalpha 50kDa, anon-WO0140519.257, dnapol alpha 50, pol alpha, polalpha, polalpha50', description='DNA polymerase alpha 50kD', other_designations='CG7108-PA|DNA polymerase 50K|DNA polymerase alpha|DNA polymerase alpha 46,000 gamma subunit|DNA polymerase alpha 50 kda primase subunit|DNA polymerase alpha 50,000 gamma subunit|DNA-polymerase alpha|DNApol-alpha50-PA|p50 subunit of DNA polymerase alpha', map_location='66C7-66C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8340314, end_position_on_the_genomic_accession=8344404, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38943, CurrentID=0, Status='live', Symbol='CG7083', Aliases='Dmel_ Dmel\\CG7083', description='CG7083 gene product from transcript CG7083-RA', other_designations='CG7083-PA', map_location='66C7-66C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8340785, end_position_on_the_genomic_accession=8342876, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38944, CurrentID=0, Status='live', Symbol='CG7182', Aliases='Dmel_ Dmel\\CG7182', description='CG7182 gene product from transcript CG7182-RB', other_designations='CG7182-PA|CG7182-PB', map_location='66C8-66C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8344421, end_position_on_the_genomic_accession=8346382, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38945, CurrentID=0, Status='live', Symbol='GAPcenA', Aliases='Dmel_CG7112, CG7112, Dmel\\CG7112, FBgn 35879, FBgn0035879, GapcenA', description='GTPase activating protein and centrosome-associated ortholog (H. sapiens)', other_designations='CG7112-PA|CG7112-PB|GAPcenA-PA|GAPcenA-PB|GTPase activating protein and centrosome-associated ortholog', map_location='66C8-66C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8346561, end_position_on_the_genomic_accession=8352279, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  38946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38946, CurrentID=0, Status='live', Symbol='CG17352', Aliases='Dmel_ CT33100, Dmel\\CG17352', description='CG17352 gene product from transcript CG17352-RA', other_designations='CG17352-PA|CG17352-PB|CG17352-PC|CG17352-PE|uncharacterized protein', map_location='66C8-66C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8352355, end_position_on_the_genomic_accession=8356296, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  5740580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740580, CurrentID=0, Status='live', Symbol='CG34443', Aliases='Dmel_ CG30070, Dmel\\CG34443, Dmel_CG30070', description='CG34443 gene product from transcript CG34443-RA', other_designations='CG34443-PA|uncharacterized protein', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14367979, end_position_on_the_genomic_accession=14369067, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5740583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740583, CurrentID=0, Status='live', Symbol='CG13051', Aliases='Dmel_ Dmel\\CG13051', description='CG13051 gene product from transcript CG13051-RA', other_designations='CG13051-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16264713, end_position_on_the_genomic_accession=16265100, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38952, CurrentID=0, Status='live', Symbol='Jon66Ci', Aliases='Dmel_CG7118, 66C, CG7118, Dmel\\CG7118, Jon66C, SP171', description='Jonah 66Ci', other_designations='CG7118-PA|Jon66Ci-PA|jonah 66C|jonah 66Ci', map_location='66C10-66C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8391540, end_position_on_the_genomic_accession=8392377, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38953, CurrentID=0, Status='live', Symbol='Jon66Cii', Aliases='Dmel_CG7170, 66C, CCG7170, CG7170, Dmel\\CG7170, Jon66C, SP165', description='Jonah 66Cii', other_designations='CG7170-PA|Jon66Cii-PA|chymotrypsin-type peptidase|jonah 66C|jonah 66Cii', map_location='66C10-66C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8392820, end_position_on_the_genomic_accession=8393649, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38954, CurrentID=0, Status='live', Symbol='CG7120', Aliases='Dmel_ BcDNA:AT15424, Dmel\\CG7120, gi7295100', description='CG7120 gene product from transcript CG7120-RG', other_designations='CG7120-PD|CG7120-PE|CG7120-PF|CG7120-PG', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8396000, end_position_on_the_genomic_accession=8402295, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38955, CurrentID=0, Status='live', Symbol='mkg-p', Aliases='Dmel_CG7163, CG7163, Dmel\\CG7163, mkg, mkg-r, mkgr', description='monkey king protein', other_designations='CG7163-PA|CG7163-PB|CG7163-PD|mkg-p-PA|mkg-p-PB|mkg-p-PD|monkey-king', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8402168, end_position_on_the_genomic_accession=8405277, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38956, CurrentID=0, Status='live', Symbol='CG13667', Aliases='Dmel_ Dmel\\CG13667', description='CG13667 gene product from transcript CG13667-RB', other_designations='CG13667-PA|CG13667-PB|uncharacterized protein', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8405454, end_position_on_the_genomic_accession=8407354, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38957, CurrentID=0, Status='live', Symbol='Oseg1', Aliases='Dmel_CG7161, CG7161, DAF-10, Dmel\\CG7161, IFT122, OSEG, OSEG1, OSM-4, Wdr10, mecC, oseg1', description='Outer segment 1', other_designations='CG7161-PA|CG7161-PB|Oseg1-PA|Oseg1-PB|mechanosensory C|outer segment 1', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8407268, end_position_on_the_genomic_accession=8412825, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  5740590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740590, CurrentID=0, Status='live', Symbol='CG34353', Aliases='Dmel_ CG12274, CG12427, CG12873, CG12874, CG12875, CG5566, CG5572, Dmel\\CG34353, Dmel_CG12274, Dmel_CG12427, Dmel_CG12873, Dmel_CG12874, Dmel_CG12875, Dmel_CG5566, Dmel_CG5572', description='CG34353 gene product from transcript CG34353-RB', other_designations='CG34353-PA|CG34353-PB|CG34353-PC|uncharacterized protein', map_location='98B3-98B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27767966, end_position_on_the_genomic_accession=27905843, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  38959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38959, CurrentID=0, Status='live', Symbol='Exo70', Aliases='Dmel_CG7127, CG7127, Dmel\\CG7127p, exo70, Exo70', description='Exo70 ortholog (S. cerevisiae)', other_designations='CG7127-PA|Exo70 ortholog|Exo70-PA|exo70', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8411851, end_position_on_the_genomic_accession=8415910, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  5740592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740592, CurrentID=0, Status='live', Symbol='ATbp', Aliases='Dmel_CG17172, ATBP, CG17172, CG40145, Dmel\\CG17172', description='(A+T)-stretch binding protein', other_designations='ATbp-PB|CG17172-PB', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23106003, end_position_on_the_genomic_accession=23108331, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38961, CurrentID=0, Status='live', Symbol='Cbl', Aliases='Dmel_CG7037, CBL, CG7037, CG7043, D-Cbl, D-Cbl L, D-cbl, D-cblL, D-cblS, DCbl, Dcbl, Dmel\\CG7037, Dv-Cbl, Dv-cbl, F165, anon-WO0118547.68, cbl, d-cbl, dCbl, v-Cbl', description='Cbl proto-oncogene ortholog', other_designations='CG7037-PA|CG7037-PB|Cbl-PA|Cbl-PB|D-cblLONG|D-cblSHORT|cable', map_location='3-24.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8424954, end_position_on_the_genomic_accession=8432246, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38963, CurrentID=0, Status='live', Symbol='Unr', Aliases='Dmel_CG7015, BcDNA:LD13080, CG7015, CR32028, Dmel\\CG7015, Dmel_CR32028, MRE30, UNR, dUNR, unr', description='Upstream of N-ras', other_designations='CG7015-PA|CG7015-PB|CG7015-PC|CG7015-PD|CG7015-PE|Unr-PA|Unr-PB|Unr-PC|Unr-PD|Unr-PE|upstream of N-ras', map_location='66C13-66C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8432322, end_position_on_the_genomic_accession=8442381, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  38964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38964, CurrentID=0, Status='live', Symbol='CG6983', Aliases='Dmel_ Dmel\\CG6983', description='CG6983 gene product from transcript CG6983-RB', other_designations='CG6983-PA|CG6983-PB', map_location='66C13-66C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8443042, end_position_on_the_genomic_accession=8449377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740597, CurrentID=0, Status='live', Symbol='RYa', Aliases='Dmel_CG40733, CG40733, Dmel\\CG40733, dRYamide', description='RYamide', other_designations='CG40733-PC|RYa-PC', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=3647346, end_position_on_the_genomic_accession=3647765, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38966, CurrentID=0, Status='live', Symbol='CG43163', Aliases='Dmel_ CG6902, CG6915, Dmel\\CG43163, Dmel_CG6902, Dmel_CG6915', description='CG43163 gene product from transcript CG43163-RC', other_designations='CG43163-PC|CG43163-PD|uncharacterized protein', map_location='66D1-66D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8472807, end_position_on_the_genomic_accession=8507146, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  5740599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740599, CurrentID=0, Status='live', Symbol='pgc', Aliases='Dmel_CG32885, CG11296, CG32885, CR32885, Dmel\\CG32885, Pgc, Pgc1-1, pgr, pgc', description='polar granule component', other_designations='58D gene|CG32885-PA|CG32885-PC|CG32885-PD|pgc-PA|pgc-PC|pgc-PD', map_location='58D3-58D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22330200, end_position_on_the_genomic_accession=22332813, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38968, CurrentID=0, Status='live', Symbol='ZC3H3', Aliases='Dmel_CG6694, CG6694, Dmel\\CG6694, d', description='CG6694 gene product from transcript CG6694-RA', other_designations='CG6694-PA|ZC3H3-PA', map_location='66D4-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8507469, end_position_on_the_genomic_accession=8509763, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38969, CurrentID=0, Status='live', Symbol='CG6745', Aliases='Dmel_ CT20937, Dmel\\CG6745', description='CG6745 gene product from transcript CG6745-RA', other_designations='CG6745-PA', map_location='66D5-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8510003, end_position_on_the_genomic_accession=8513007, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38970, CurrentID=0, Status='live', Symbol='CG6683', Aliases='Dmel_ Dmel\\CG6683', description='CG6683 gene product from transcript CG6683-RA', other_designations='CG6683-PA', map_location='66D5-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8513874, end_position_on_the_genomic_accession=8514921, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38971, CurrentID=0, Status='live', Symbol='CG6765', Aliases='Dmel_ Dmel\\CG6765', description='CG6765 gene product from transcript CG6765-RB', other_designations='CG6765-PA|CG6765-PB|CG6765-PC', map_location='66D5-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8512767, end_position_on_the_genomic_accession=8518738, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  38972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38972, CurrentID=0, Status='live', Symbol='GstO3', Aliases='Dmel_CG6776, CG6776, Dmel\\CG6776, GSTO3, T14', description='Glutathione S transferase O3', other_designations='CG6776-PA|GstO3-PA|glutathione S transferase O3|glutathione transferase', map_location='66D5-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8518885, end_position_on_the_genomic_accession=8520295, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38973, CurrentID=0, Status='live', Symbol='se', Aliases='Dmel_CG6781, CG6781, Dmel\\CG6781, GSTO1, GSTO4, cg6781, dmgsto1', description='sepia', other_designations='CG6781-PA|se-PA', map_location='3-26.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8520552, end_position_on_the_genomic_accession=8521489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38974, CurrentID=0, Status='live', Symbol='GstO2', Aliases='Dmel_CG6673, CG6673, CG6673-A, CG6673-B, CG6673A, CG6673B, DmGSTO1, Dmel\\CG6673, GSTO2, anon-EST:Posey29, gsto2', description='Glutathione S transferase O2', other_designations='CG6673-PA|CG6673-PB|GstO2-PA|GstO2-PB|glutathione S transferase O2', map_location='66D5-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8521429, end_position_on_the_genomic_accession=8523501, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38975, CurrentID=0, Status='live', Symbol='GstO1', Aliases='Dmel_CG6662, CG6662, Dmel\\CG6662, GSTO1', description='Glutathione S transferase O1', other_designations='CG6662-PA|GstO1-PA|glutathione S transferase O1', map_location='66D5-66D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8523742, end_position_on_the_genomic_accession=8524866, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38976, CurrentID=0, Status='live', Symbol='foi', Aliases='Dmel_CG6817, CG6817, Dmel\\CG6817, FOI, cg6817, dFOI, kas, l(3)j8E8, l(3)neo13', description='fear-of-intimacy', other_designations='CG6817-PA|CG6817-PB|CG6817-PC|CG6817-PD|fear of intimacy|foi-PA|foi-PB|foi-PC|foi-PD|kastchen', map_location='66D5-66D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8525244, end_position_on_the_genomic_accession=8534503, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740609, CurrentID=0, Status='live', Symbol='CG34290', Aliases='Dmel_ Dmel\\CG34290', description='CG34290 gene product from transcript CG34290-RB', other_designations='CG34290-PB|uncharacterized protein', map_location='95F9-95F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24288553, end_position_on_the_genomic_accession=24289818, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  38978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38978, CurrentID=0, Status='live', Symbol='rhea', Aliases='Dmel_CG6831, CG6831, CT21159, Dmel\\CG6831, Rhea, Talin, Tln, talin, tendrils', description='CG6831 gene product from transcript CG6831-RB', other_designations='CG6831-PB|CG6831-PC|CG6831-PD|CG6831-PE|CG6831-PF|CG6831-PG|CG6831-PH|CG6831-PI|rhea-PB|rhea-PC|rhea-PD|rhea-PE|rhea-PF|rhea-PG|rhea-PH|rhea-PI|talin|tendrils', map_location='66D6-66D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8545001, end_position_on_the_genomic_accession=8574409, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  38979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38979, CurrentID=0, Status='live', Symbol='CG6638', Aliases='Dmel_ Dmel\\CG6638', description='CG6638 gene product from transcript CG6638-RA', other_designations='CG6638-PA', map_location='66D7-66D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8574407, end_position_on_the_genomic_accession=8576078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740612, CurrentID=0, Status='live', Symbol='CG34451', Aliases='Dmel_ CG33999, Dmel\\CG34451, Dmel_CG33999', description='CG34451 gene product from transcript CG34451-RC', other_designations='CG34451-PA|CG34451-PB|CG34451-PC|uncharacterized protein', map_location='71F1-71F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15688517, end_position_on_the_genomic_accession=15692047, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38981, CurrentID=0, Status='live', Symbol='CG43078', Aliases='Dmel_ CG17926, CG32352, CG32353, CG32355, CG6239, Dmel\\CG43078, Dmel_CG32352, Dmel_CG32353, Dmel_CG32355, lincRNA.473', description='CG43078 gene product from transcript CG43078-RN', other_designations='CG43078-PA|CG43078-PB|CG43078-PC|CG43078-PD|CG43078-PE|CG43078-PG|CG43078-PH|CG43078-PI|CG43078-PJ|CG43078-PK|CG43078-PL|CG43078-PM|CG43078-PN|uncharacterized protein', map_location='66D7-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8576808, end_position_on_the_genomic_accession=8599617, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38982, CurrentID=0, Status='live', Symbol='Nelf-E', Aliases='Dmel_CG5994, CG5994, Dmel\\CG5994, NELF, NELF-E, Nelf-e, NelfE, anon-66Da, cg5994, dNelf-E', description='Negative elongation factor E', other_designations='CG5994-PA|Nelf-E-PA|anonymous-66Da|negative elongation factor|negative elongation factor E', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8599442, end_position_on_the_genomic_accession=8600879, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38983, CurrentID=0, Status='live', Symbol='RpL14', Aliases='Dmel_CG6253, CG6253, Dmel\\CG6253, L14, M(3), M(3)66, M(3)66D, M66D, RPL14, Rp L14, anon-EST:Posey193U, anon-EST:Posey6, rpl14', description='Ribosomal protein L14', other_designations='60 S ribosomal protein L14|CG6253-PA|Minute(3)66D|RpL14-PA|minute (3)66D|ribosomal protein L14', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8601006, end_position_on_the_genomic_accession=8602392, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38984, CurrentID=0, Status='live', Symbol='CG5989', Aliases='Dmel_ Dmel\\CG5989, anon-66Db', description='CG5989 gene product from transcript CG5989-RC', other_designations='CG5989-PA|CG5989-PB|CG5989-PC|SAM methionine methyltransferase|anonymous-66Db', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8602411, end_position_on_the_genomic_accession=8604149, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38985, CurrentID=0, Status='live', Symbol='CG6282', Aliases='Dmel_ Dmel\\CG6282', description='CG6282 gene product from transcript CG6282-RC', other_designations='CG6282-PA|CG6282-PB|CG6282-PC|CG6282-PD', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8604657, end_position_on_the_genomic_accession=8617074, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  38986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38986, CurrentID=0, Status='live', Symbol='S-Lap1', Aliases='Dmel_CG6372, CG6372, Dmel\\CG6372, S-LAP 1, anon-WO0140519.89, bs33g03.y1, bs34e01.y1', description='Sperm-Leucylaminopeptidase 1', other_designations='CG6372-PA|S-Lap1-PA', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8617235, end_position_on_the_genomic_accession=8619474, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38987, CurrentID=0, Status='live', Symbol='GAPsec', Aliases='Dmel_CG5978, BEST:LD24460, CG5978, Dmel\\CG5978, dGAPsec', description='GTPase activating protein, SECIS-dependent read-through', other_designations='CG5978-PA|GAPsec-PA', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8622039, end_position_on_the_genomic_accession=8623901, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  38988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38988, CurrentID=0, Status='live', Symbol='Zasp66', Aliases='Dmel_CG6416, CG6416, Dmel\\CG6416, ZASP66, Zaps66, anon-WO0153538.30', description='Z band alternatively spliced PDZ-motif protein 66', other_designations='CG6416-PA|CG6416-PB|CG6416-PE|CG6416-PF|CG6416-PG|CG6416-PH|CG6416-PI|CG6416-PJ|CG6416-PK|CG6416-PM|CG6416-PN|CG6416-PO|Zasp66-PA|Zasp66-PB|Zasp66-PE|Zasp66-PF|Zasp66-PG|Zasp66-PH|Zasp66-PI|Zasp66-PJ|Zasp66-PK|Zasp66-PM|Zasp66-PN|Zasp66-PO', map_location='66D9-66D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8624371, end_position_on_the_genomic_accession=8638813, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38989, CurrentID=0, Status='live', Symbol='Cdc6', Aliases='Dmel_CG5971, CDC6, CG5971, Dm Dmel\\CG5971, cdc6', description='CG5971 gene product from transcript CG5971-RA', other_designations='CG5971-PA|Cdc6-PA', map_location='66D9-66D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8631509, end_position_on_the_genomic_accession=8634069, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38990, CurrentID=0, Status='live', Symbol='Cpr66D', Aliases='Dmel_CG32029, CG32029, CG6458, CG6469, DmelCpr66D, Dmel\\CG32029', description='Cuticular protein 66D', other_designations='CG32029-PA|Cpr66D-PA|cuticular protein 66D', map_location='66D9-66D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8640439, end_position_on_the_genomic_accession=8643584, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740624, CurrentID=0, Status='live', Symbol='Hsc20', Aliases='Dmel_CG34246, CG34246, Dmel\\CG34246, l(3)72Do', description='Heat shock protein cognate 20', other_designations='CG34246-PA|Hsc20-PA|heat shock protein cognate 20|lethal (3) 72Do', map_location='72D10-72D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16233127, end_position_on_the_genomic_accession=16234165, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38993, CurrentID=0, Status='live', Symbol='Arr2', Aliases='Dmel_CG5962, 49, AH, ARR2, Arr, Arr-2, ArrB, ArrB-ml, CG5962, DAII, Dmel\\CG5962, Dro.-Arr.II, PRI, arr, arr2', description='Arrestin 2', other_designations='49kD arrestin homolog|Arr2-PA|Arrestin-2|Beta-Arrestin|CG5962-PA|arrestin|arrestin 2|arrestin B|arrestin-like protein|arrestin2|phosrestin I|phosrestin-1', map_location='66D9-66D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8647261, end_position_on_the_genomic_accession=8648936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38994, CurrentID=0, Status='live', Symbol='Pex7', Aliases='Dmel_CG6486, CG6486, DmelPex7, Dmel\\CG6486, PEX7', description='Peroxin 7', other_designations='CG6486-PA|CG6486-PB|Pex7-PA|Pex7-PB|peroxin 7', map_location='66D9-66D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8647145, end_position_on_the_genomic_accession=8651011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  38995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38995, CurrentID=0, Status='live', Symbol='h', Aliases='Dmel_CG6494, 8247, CG6494, Dmel\\CG6494, H, HRY, Hairy, bHLHb39, bHLHc14, brr, dDr1ry, l(3)08247, l(3)rM384, h', description='hairy', other_designations='CG6494-PA|CG6494-PB|barrel|h-PA|h-PB', map_location='3-26.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8675759, end_position_on_the_genomic_accession=8679253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38997, CurrentID=0, Status='live', Symbol='CG6511', Aliases='Dmel_ Dmel\\CG6511', description='CG6511 gene product from transcript CG6511-RB', other_designations='CG6511-PA|CG6511-PB', map_location='66D11-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8721324, end_position_on_the_genomic_accession=8726108, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38998, CurrentID=0, Status='live', Symbol='Cp18', Aliases='Dmel_CG6517, ACE3, CG6517, CP18, Dmel\\CG6517, Dms18, E5, S-18, S18, cp18, s-18, s18, s18-1', description='Chorion protein 18', other_designations='CG6517-PA|Cp18-PA|Shell-18|chorion protein 18', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8726791, end_position_on_the_genomic_accession=8727613, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38999, CurrentID=0, Status='live', Symbol='Cp15', Aliases='Dmel_CG6519, C.s15, CG6519, Cs15, Dmel\\CG6519, Dms15, E6, S15, cp15, melcp15, s-15, s15, s15-1', description='Chorion protein 15', other_designations='CG6519-PA|Cp15-PA|Shell-15|chorion prot 15|chorion protein 15|s15 chorion gene', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8728481, end_position_on_the_genomic_accession=8729066, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39000, CurrentID=0, Status='live', Symbol='Cp19', Aliases='Dmel_CG6524, C.s19, CG6524, CP19, Dmel\\CG6524, Dms19, S19, cp19, s19, s19-1', description='Chorion protein 19', other_designations='CG6524-PA|Cp19-PA|Shell-19|chorion protein 19', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8729865, end_position_on_the_genomic_accession=8730604, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740633, CurrentID=0, Status='live', Symbol='nvd', Aliases='Dmel_CG40050, CG40050, Dmel\\CG40050-Dm, nvd', description='neverland', other_designations='CG40050-PA|lethal 6|nvd-PA', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24324866, end_position_on_the_genomic_accession=24401655, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39002, CurrentID=0, Status='live', Symbol='Prm', Aliases='Dmel_CG5939, 0106/05, CG5939, Dmel\\CG5939, FBgn0003149, PM, PM/mPM, Para, l(3)10631, l(3)S010605, mPM, pm, prm', description='Paramyosin', other_designations='CG5939-PA|CG5939-PB|CG5939-PC|CG5939-PD|CG5939-PE|Prm-PA|Prm-PB|Prm-PC|Prm-PD|Prm-PE|lethal (3) S010605|miniparamyosin|paramyosin|paramyosin/miniparamyosin', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8732858, end_position_on_the_genomic_accession=8745310, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  39003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39003, CurrentID=0, Status='live', Symbol='CG6576', Aliases='Dmel_ Dmel\\CG6576', description='CG6576 gene product from transcript CG6576-RA', other_designations='CG6576-PA', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8755989, end_position_on_the_genomic_accession=8757167, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39004, CurrentID=0, Status='live', Symbol='Fhos', Aliases='Dmel_CG42610, CG32025, CG32030, CG42610, CG5775, CG5797, Dmel\\CG42610, Dmel_CG32025, Dmel_CG32030, dFHOS, desert, jaz, knit, l(3)01629', description='Formin homology 2 domain containing ortholog', other_designations='CG42610-PA|CG42610-PB|CG42610-PD|CG42610-PE|CG42610-PF|CG42610-PG|CG42610-PH|CG42610-PI|CG42610-PJ|Fhos-PA|Fhos-PB|Fhos-PD|Fhos-PE|Fhos-PF|Fhos-PG|Fhos-PH|Fhos-PI|Fhos-PJ|desert|formin homology 2 domain containing ortholog|jazigo|knittrig|lethal (3) 01629', map_location='66D12-66E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8753619, end_position_on_the_genomic_accession=8798719, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  39005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39005, CurrentID=0, Status='live', Symbol='CG5804', Aliases='Dmel_ BcDNA:RE05521, Dmel\\CG5804', description='CG5804 gene product from transcript CG5804-RA', other_designations='CG5804-PA', map_location='66D14-66D14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8788144, end_position_on_the_genomic_accession=8788545, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39007, CurrentID=0, Status='live', Symbol='CG13310', Aliases='Dmel_ Dmel\\CG13310', description='CG13310 gene product from transcript CG13310-RA', other_designations='CG13310-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8798730, end_position_on_the_genomic_accession=8800532, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39008, CurrentID=0, Status='live', Symbol='CG13311', Aliases='Dmel_ BcDNA:RE29808, Dmel\\CG13311', description='CG13311 gene product from transcript CG13311-RA', other_designations='CG13311-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8801540, end_position_on_the_genomic_accession=8802164, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39010, CurrentID=0, Status='live', Symbol='CG13312', Aliases='Dmel_ Dmel\\CG13312', description='CG13312 gene product from transcript CG13312-RA', other_designations='CG13312-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8805811, end_position_on_the_genomic_accession=8807213, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39011, CurrentID=0, Status='live', Symbol='CG13308', Aliases='Dmel_ Dmel\\CG13308', description='CG13308 gene product from transcript CG13308-RB', other_designations='CG13308-PB|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8808544, end_position_on_the_genomic_accession=8809307, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39012, CurrentID=0, Status='live', Symbol='CG13309', Aliases='Dmel_ CT32604, Dmel\\CG13309', description='CG13309 gene product from transcript CG13309-RA', other_designations='CG13309-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8810395, end_position_on_the_genomic_accession=8811265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39013, CurrentID=0, Status='live', Symbol='dally', Aliases='Dmel_CG4974, CG4974, CT15952, Dally, Dmel\\CG4974, anon-EST:Liang-1.63, clone 1.63, gem, l(3)06464, l(3)SG18', description='division abnormally delayed', other_designations='CG4974-PA|CG4974-PB|dally-PA|dally-PB|genitalia missing|lethal (3) SG18', map_location='3-31.9 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8827505, end_position_on_the_genomic_accession=8891192, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39014, CurrentID=0, Status='live', Symbol='Mcm7', Aliases='Dmel_CG4978, CG4978, DmMCM7, DmeMCM7, Dmel\\CG4978, MCM7, McM7, Mcp-PCR1, PCR1, anon-EST:Liang-1.66, clone 1.66, dMcm7, mcm7', description='Minichromosome maintenance 7', other_designations='CG4978-PA|Mcm7-PA|minichromosome maintenance 7|minichromosome maintenance-PCR1', map_location='66E3-66E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8891596, end_position_on_the_genomic_accession=8894310, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39015, CurrentID=0, Status='live', Symbol='TrpA1', Aliases='Dmel_CG5751, ANKTM1, Anktm1, CG5751, CG5761, CT18073, DmTRPA1, Dmel\\CG5751, TRPA, TRPA1, Trp1A, TrpA, Trpa1, dANKTM1, dTRPA, dTRPA1, dTrpA, dTrpA1, dtrpA1, trpA1', description='Transient receptor potential cation channel A1 ortholog', other_designations='CG5751-PG|CG5751-PH|CG5751-PI|CG5751-PJ|CG5751-PK|TrpA1-PG|TrpA1-PH|TrpA1-PI|TrpA1-PJ|TrpA1-PK|transient receptor potential A 1|transient receptor potential A1|transient receptor potential cation channel|transient receptor potential cation channel A1 ortholog|transient-receptor-potential A1', map_location='66E3-66E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8894404, end_position_on_the_genomic_accession=8904790, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  39016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39016, CurrentID=0, Status='live', Symbol='mfr', Aliases='Dmel_CG5747, CG5747, Dmel\\CG5747, Otoferlin', description='misfire', other_designations='CG5747-PB|CG5747-PC|CG5747-PD|CG5747-PE|CG5747-PF|CG5747-PI|CG5747-PJ|mfr-PB|mfr-PC|mfr-PD|mfr-PE|mfr-PF|mfr-PI|mfr-PJ|otoferlin', map_location='66E3-66E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8905918, end_position_on_the_genomic_accession=8912671, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  39017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39017, CurrentID=0, Status='live', Symbol='Tsp66E', Aliases='Dmel_CG4999, CG4999, Dm.Tsp66E, Dmel\\CG4999', description='Tetraspanin 66E', other_designations='CG4999-PA|CG4999-PB|CG4999-PC|Tsp66E-PA|Tsp66E-PB|Tsp66E-PC|tetraspanin 66E', map_location='66E3-66E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8912499, end_position_on_the_genomic_accession=8932065, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39018, CurrentID=0, Status='live', Symbol='orb2', Aliases='Dmel_CG43782, CG43113, CG43782, CG5735, Dmel\\CG43782, Dmel_CG43113, Dmel_CG5735, Dmel_CG5741, Orb2, anon-WO0140519.222, cg5735', description='CG43782 gene product from transcript CG43782-RH', other_designations='CG43782-PA|CG43782-PB|CG43782-PC|CG43782-PD|CG43782-PH|orb2-PA|orb2-PB|orb2-PC|orb2-PD|orb2-PH', map_location='66E4-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8937100, end_position_on_the_genomic_accession=8954463, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39020, CurrentID=0, Status='live', Symbol='GNBP3', Aliases='Dmel_CG5008, CG5008, DGNBP-3, Dmel\\CG5008, GNBP, GNBP-3, Gnbp3, Hades, gnbp3', description='Gram-negative bacteria binding protein 3', other_designations='CG5008-PA|GNBP3-PA|Gram-negative binding protein 3', map_location='66E5-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8955021, end_position_on_the_genomic_accession=8956551, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39022, CurrentID=0, Status='live', Symbol='CG13313', Aliases='Dmel_ Dmel\\CG13313', description='CG13313 gene product from transcript CG13313-RA', other_designations='CG13313-PA|uncharacterized protein', map_location='66E5-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8958857, end_position_on_the_genomic_accession=8960921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39023, CurrentID=0, Status='live', Symbol='CG5660', Aliases='Dmel_ Dmel\\CG5660', description='CG5660 gene product from transcript CG5660-RB', other_designations='CG5660-PA|CG5660-PB|mitochondrial valyl-tRNA synthetase', map_location='66E5-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8961274, end_position_on_the_genomic_accession=8964769, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39024, CurrentID=0, Status='live', Symbol='CG5653', Aliases='Dmel_ Dmel\\CG5653', description='CG5653 gene product from transcript CG5653-RA', other_designations='CG5653-PA', map_location='66E5-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8964912, end_position_on_the_genomic_accession=8966374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39025, CurrentID=0, Status='live', Symbol='CG5021', Aliases='Dmel_ Dmel\\CG5021', description='CG5021 gene product from transcript CG5021-RC', other_designations='CG5021-PA|CG5021-PB|CG5021-PC|CG5021-PD', map_location='66E5-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8966829, end_position_on_the_genomic_accession=8968171, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39026, CurrentID=0, Status='live', Symbol='CG5026', Aliases='Dmel_ # CG5026, Dmel\\CG5026', description='CG5026 gene product from transcript CG5026-RB', other_designations='CG5026-PA|CG5026-PB|CG5026-PC', map_location='66E5-66E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8968528, end_position_on_the_genomic_accession=8971752, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39027, CurrentID=0, Status='live', Symbol='pix', Aliases='Dmel_CG5651, ABCE1, BEST:GH18088, CG5651, Dmel\\CG5651, E3-c', description='pixie', other_designations='CG5651-PA|CG5651-PB|pix-PA|pix-PB', map_location='66E6-66E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8972056, end_position_on_the_genomic_accession=8974765, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39028, CurrentID=0, Status='live', Symbol='Srp68', Aliases='Dmel_CG5064, CG5064, Dmel\\CG5064', description='Signal recognition particle protein 68', other_designations='CG5064-PA|Srp68-PA|signal recognition particle protein 68', map_location='66E6-66E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8975046, end_position_on_the_genomic_accession=8977401, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740661, CurrentID=0, Status='live', Symbol='CG34457', Aliases='Dmel_ CG5197, Dmel\\CG34457, Dmel_CG5197, SP118', description='CG34457 gene product from transcript CG34457-RA', other_designations='CG34457-PA|CG34457-PB|uncharacterized protein', map_location='53D1-53D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16670326, end_position_on_the_genomic_accession=16671836, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39030, CurrentID=0, Status='live', Symbol='CG13314', Aliases='Dmel_ Dmel\\CG13314', description='CG13314 gene product from transcript CG13314-RB', other_designations='CG13314-PA|CG13314-PB|uncharacterized protein', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8981795, end_position_on_the_genomic_accession=8983212, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39031, CurrentID=0, Status='live', Symbol='Galk', Aliases='Dmel_CG5288, CG5288, CT16885, Dmel\\CG5288, GALK, cg5288, dGALK', description='Galactokinase', other_designations='CG5288-PA|CG5288-PB|CG5288-PC|CG5288-PD|Galk-PA|Galk-PB|Galk-PC|Galk-PD|galactokinase', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8982212, end_position_on_the_genomic_accession=8986091, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  5740664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740664, CurrentID=0, Status='live', Symbol='CG40228', Aliases='Dmel_ BcDNA:RE67573, Dmel\\CG40228', description='CG40228 gene product from transcript CG40228-RC', other_designations='CG40228-PC|CG40228-PD|CG40228-PE|CG40228-PF|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=27082091, end_position_on_the_genomic_accession=27083906, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39033, CurrentID=0, Status='live', Symbol='CG5280', Aliases='Dmel_ Dmel\\CG5280', description='CG5280 gene product from transcript CG5280-RA', other_designations='CG5280-PA', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8989069, end_position_on_the_genomic_accession=8990228, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39034, CurrentID=0, Status='live', Symbol='smg', Aliases='Dmel_CG5263, CG5263, D69, Dmel\\CG5263, SMG, Smg, anon-D69, anonD69', description='smaug', other_designations='CG5263-PA|CG5263-PB|CG5263-PC|CG5263-PD|CG5263-PE|SMAUG|dSmaug|smau|smg-PA|smg-PB|smg-PC|smg-PD|smg-PE', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8990317, end_position_on_the_genomic_accession=8997989, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39035, CurrentID=0, Status='live', Symbol='CG5087', Aliases='Dmel_ Dmel\\CG5087, NM_140022.1', description='CG5087 gene product from transcript CG5087-RA', other_designations='CG5087-PA|CG5087-PB', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8998184, end_position_on_the_genomic_accession=9002566, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740668, CurrentID=0, Status='live', Symbol='CG34347', Aliases='Dmel_ CG11339, CG15566, Dmel\\CG34347, Dmel_CG11339, Dmel_CG15566, EST:LP08211, LP8211', description='CG34347 gene product from transcript CG34347-RB', other_designations='CG34347-PB|CG34347-PE|uncharacterized protein', map_location='100C1-100C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31311157, end_position_on_the_genomic_accession=31371748, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  39037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39037, CurrentID=0, Status='live', Symbol='CG5194', Aliases='Dmel_ Dmel\\CG5194, anon-WO0140519.239', description='CG5194 gene product from transcript CG5194-RB', other_designations='CG5194-PA|CG5194-PB', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9008624, end_position_on_the_genomic_accession=9009838, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39038, CurrentID=0, Status='live', Symbol='Doc2', Aliases='Dmel_CG5187, CG5187, Dm-DOC2, Dmel\\CG5187, Doc, doc2', description='Dorsocross2', other_designations='CG5187-PA|Doc2-PA|dorsocross', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9012686, end_position_on_the_genomic_accession=9019246, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39039, CurrentID=0, Status='live', Symbol='Doc1', Aliases='Dmel_CG5133, CG5133, Dm-DOC1, Dmel\\CG5133, Doc, Ect2, TBX6, Tb66F2, doc, doc1', description='Dorsocross1', other_designations='CG5133-PA|CG5133-PB|Doc1-PA|Doc1-PB|dorsocross', map_location='66F2-66F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9041375, end_position_on_the_genomic_accession=9045042, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740672, CurrentID=0, Status='live', Symbol='CG34351', Aliases='Dmel_ CG11928, CG15633, Dmel\\CG34351, Dmel_CG11928, Dmel_CG15633, ORE-3', description='CG34351 gene product from transcript CG34351-RC', other_designations='CG34351-PC|CG34351-PD|uncharacterized protein', map_location='25A3-25A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4694721, end_position_on_the_genomic_accession=4722465, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39041, CurrentID=0, Status='live', Symbol='Argk', Aliases='Dmel_CG32031, AK, AK-1, ARK, ArgK, CG32031, CG4929, CG5173, Dmel\\CG32031, argK, argk', description='Arginine kinase', other_designations='Argk-PA|Argk-PB|Argk-PC|Argk-PD|Argk-PE|Argk-PG|Argk-PH|CG32031-PA|CG32031-PB|CG32031-PC|CG32031-PD|CG32031-PE|CG32031-PG|CG32031-PH|arginine kinase|insect intestinal mucin iim22', map_location='3-25.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9048781, end_position_on_the_genomic_accession=9066027, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39043, CurrentID=0, Status='live', Symbol='CG4911', Aliases='Dmel_ Dmel\\CG4911, FBXO4', description='CG4911 gene product from transcript CG4911-RC', other_designations='CG4911-PA|CG4911-PB|CG4911-PC', map_location='66F4-66F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9067195, end_position_on_the_genomic_accession=9072731, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5740676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740676, CurrentID=0, Status='live', Symbol='inaF-A', Aliases='Dmel_CG34322, CG34322, Dm_X:58729, Dmel\\CG34322, INAF, InaF, inaF', description='CG34322 gene product from transcript CG34322-RA', other_designations='CG34322-PA|inaF-A-PA', map_location='10D8-10E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11726390, end_position_on_the_genomic_accession=11733895, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39048, CurrentID=0, Status='live', Symbol='Tequila', Aliases='Dmel_CG4821, BEST:CK00198, BEST:GH18038, CG18153, CG18403, CG4821, CG4948, CK00198, Dmel\\CG4821, SP72, Teq, anon-EST:ParkEST161, graal, teq', description='CG4821 gene product from transcript CG4821-RG', other_designations='CG4821-PC|CG4821-PD|CG4821-PF|CG4821-PG|Tequila-PC|Tequila-PD|Tequila-PF|Tequila-PG|tequila', map_location='66F4-66F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9074643, end_position_on_the_genomic_accession=9092131, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  39049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39049, CurrentID=0, Status='live', Symbol='bol', Aliases='Dmel_CG4760, BOL, BcDNA:AT13332, Bol, Boule, CG4727, CG4760, Dmel\\CG4760, lincRNA.485, ms(3)07272', description='boule', other_designations='CG4760-PA|CG4760-PB|CG4760-PC|CG4760-PD|CG4760-PE|CG4760-PF|CG4760-PG|bol-PA|bol-PB|bol-PC|bol-PD|bol-PE|bol-PF|bol-PG', map_location='66F5-67A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9097629, end_position_on_the_genomic_accession=9134405, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  39050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39050, CurrentID=0, Status='live', Symbol='Dhpr', Aliases='Dmel_CG4665, CG4665, DHPR, Dmel\\CG4665', description='Dihydropteridine reductase', other_designations='CG4665-PA|CG4665-PB|CG4665-PC|Dhpr-PA|Dhpr-PB|Dhpr-PC|dihydropteridine reductase', map_location='67A1-67A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9134560, end_position_on_the_genomic_accession=9135916, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39051, CurrentID=0, Status='live', Symbol='Use1', Aliases='Dmel_CG14181, CG14181, Dmel\\CG14181, FBgn 35965', description='USE1 ortholog (S. cerevisiae)', other_designations='CG14181-PA|USE1 ortholog|Use1-PA|unconventional SNARE in the ER 1 ortholog', map_location='67A1-67A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9135709, end_position_on_the_genomic_accession=9137181, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39052, CurrentID=0, Status='live', Symbol='nwk', Aliases='Dmel_CG43479, CG43479, CG4641, CG4684, Dmel\\CG43479, Dmel_CG4641, Dmel_CG4684, NWK, Nwk', description='nervous wreck', other_designations='CG43479-PB|CG43479-PC|CG43479-PD|CG43479-PF|CG43479-PG|CG43479-PH|CG43479-PI|CG43479-PJ|CG43479-PK|CG43479-PL|nwk-PB|nwk-PC|nwk-PD|nwk-PF|nwk-PG|nwk-PH|nwk-PI|nwk-PJ|nwk-PK|nwk-PL', map_location='67A1-67A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9137860, end_position_on_the_genomic_accession=9145693, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  5740685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740685, CurrentID=0, Status='live', Symbol='CG34165', Aliases='Dmel_ Dmel\\CG34165', description='CG34165 gene product from transcript CG34165-RB', other_designations='CG34165-PA|CG34165-PB|uncharacterized protein', map_location='35B4-35B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14713750, end_position_on_the_genomic_accession=14714526, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39054, CurrentID=0, Status='live', Symbol='Rdl', Aliases='Dmel_CG10537, CG10537, CT29555, DmRDL, DmRdl, Dmel\\CG10537, GABA, GABA[[A]], GABA[[A]] receptor, GABA[[A]]R, GABAr, LCCH1, RDL, Rd1, gaba, rdl', description='Resistant to dieldrin', other_designations='CG10537-PA|CG10537-PC|CG10537-PD|CG10537-PE|CG10537-PF|CG10537-PG|GABA A beta-subunit|GABA receptor|GABA[[A]] receptor|Rdl-PA|Rdl-PC|Rdl-PD|Rdl-PE|Rdl-PF|Rdl-PG|resistance to dieldrin|resistant to dieldrin', map_location='3-24.7 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9145838, end_position_on_the_genomic_accession=9182149, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  39055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39055, CurrentID=0, Status='live', Symbol='Slc45-1', Aliases='Dmel_CG4484, CG4484, Dmel\\CG4484, scrt', description='Slc45 ortholog 1', other_designations='CG4484-PA|CG4484-PB|CG4484-PC|CG4484-PD|Slc45-1-PA|Slc45-1-PB|Slc45-1-PC|Slc45-1-PD|sucrose transporter', map_location='67A3-67A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9203065, end_position_on_the_genomic_accession=9208853, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39056, CurrentID=0, Status='live', Symbol='CG4476', Aliases='Dmel_ DmNAT3, Dmel\\CG4476, anon-WO0149848.7', description='CG4476 gene product from transcript CG4476-RB', other_designations='CG4476-PB', map_location='67A3-67A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9211880, end_position_on_the_genomic_accession=9215256, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39057, CurrentID=0, Status='live', Symbol='CG4483', Aliases='Dmel_ Dmel\\CG4483', description='CG4483 gene product from transcript CG4483-RA', other_designations='CG4483-PA', map_location='67A4-67A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9235150, end_position_on_the_genomic_accession=9236648, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39058, CurrentID=0, Status='live', Symbol='CG4477', Aliases='Dmel_ Dmel\\CG4477', description='CG4477 gene product from transcript CG4477-RB', other_designations='CG4477-PB', map_location='67A4-67A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9250225, end_position_on_the_genomic_accession=9251338, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740691, CurrentID=0, Status='live', Symbol='l(2)41Ab', Aliases='Dmel_CG41265, 1(2R)B, 41Ab, CG40214, CG40334, CG41265, Dmel\\CG41265, GroupI, l(2)EMS45-10, l(2R)B', description='lethal (2) 41Ab', other_designations='CG41265-PA|CG41265-PB|l(2)41Ab-PA|l(2)41Ab-PB', map_location='2-55.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1349392, end_position_on_the_genomic_accession=1443935, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  5740693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740693, CurrentID=0, Status='live', Symbol='CG34453', Aliases='Dmel_ BP1026, CG34018, Dmel\\CG34453, Dmel_CG34018', description='CG34453 gene product from transcript CG34453-RB', other_designations='CG34453-PB|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=232480, end_position_on_the_genomic_accession=233351, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39062, CurrentID=0, Status='live', Symbol='PGRP-LA', Aliases='Dmel_CG32042, CG18614, CG32042, CG4361, CG4384, Dmel\\CG32042, PGRPLA, pgrp-la', description='Peptidoglycan recognition protein LA', other_designations='CG32042-PC|CG32042-PD|CG32042-PE|CG32042-PF|CG32042-PG|PGRP-LA-PC|PGRP-LA-PD|PGRP-LA-PE|PGRP-LA-PF|PGRP-LA-PG|peptidoglycan recognition protein LA', map_location='67B1-67B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9334332, end_position_on_the_genomic_accession=9338336, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39063, CurrentID=0, Status='live', Symbol='PGRP-LC', Aliases='Dmel_CG4432, (PGRP)-LC, CG4432, Dm PGRP-LC, Dmel\\CG4432a, PGRP-LCx, PGRPLC, Pgrp-LC, ird7, pgrp-lc, tot, PGRP-LC', description='Peptidoglycan recognition protein LC', other_designations='CG4432-PA|CG4432-PB|CG4432-PC|CG4432-PD|CG4432-PE|CG4432-PG|CG4432-PH|CG4432-PI|PGRP-LC-PA|PGRP-LC-PB|PGRP-LC-PC|PGRP-LC-PD|PGRP-LC-PE|PGRP-LC-PG|PGRP-LC-PH|PGRP-LC-PI|Peptidoglycan-recognition protein-LC|immune response deficient 7|peptidoglycan recognition protein LC|peptidoglycan recognition protein-LC|totem', map_location='67B1-67B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9338810, end_position_on_the_genomic_accession=9348336, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  39064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39064, CurrentID=0, Status='live', Symbol='PGRP-LF', Aliases='Dmel_CG4437, CG4437, Dmel\\CG4437, PGRP-like, PGRPLF, anon-WO0118547.581, pgrp-lf', description='Peptidoglycan recognition protein LF', other_designations='CG4437-PA|PGRP-LF-PA|PRGP-like|Peptidoglycan-recognition protein-LF|peptidoglycan recognition protein|peptidoglycan recognition protein LF', map_location='67B1-67B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9349609, end_position_on_the_genomic_accession=9351489, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39065, CurrentID=0, Status='live', Symbol='UGP', Aliases='Dmel_CG4347, CG4347, Dmel\\CG4347, UDPGPPase', description='CG4347 gene product from transcript CG4347-RE', other_designations='CG4347-PA|CG4347-PC|CG4347-PD|CG4347-PE|UDP-glucose pyrophosphorylase|UGP-PA|UGP-PC|UGP-PD|UGP-PE|UTP-G-1-P uridyiyltransferase|uridine diphosphate-glucose pyrophosphorylase', map_location='67B1-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9350867, end_position_on_the_genomic_accession=9358935, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39066, CurrentID=0, Status='live', Symbol='Pdxk', Aliases='Dmel_CG34455, CG34455, CG4446, Dmel\\CG34455, Dmel_CG4446, dPdxk', description='Pyridoxal kinase', other_designations='CG34455-PC|Pdxk-PC|pyridoxal kinase', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9360316, end_position_on_the_genomic_accession=9362042, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39067, CurrentID=0, Status='live', Symbol='CG4447', Aliases='Dmel_ Dmel\\CG4447', description='CG4447 gene product from transcript CG4447-RA', other_designations='CG4447-PA', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9364350, end_position_on_the_genomic_accession=9365240, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39068, CurrentID=0, Status='live', Symbol='Klp67A', Aliases='Dmel_CG10923, 0427/05, CG10923, DmKlp67A, Dmel\\CG10923, KIF 18A, KIF18A/B, KLP3, KLP67A, KRP67A, Kin-8, Klp3, Klp67, anon-WO0140519.224, anon-WO0172774.120, klp67A, l(3)S042705', description='Kinesin-like protein at 67A', other_designations='CG10923-PA|CG10923-PB|CG10923-PC|Kinesin-8|Klp67A-PA|Klp67A-PB|Klp67A-PC|kinesin 8|kinesin like protein 67a|lethal (3) S042705', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9362636, end_position_on_the_genomic_accession=9367046, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39069, CurrentID=0, Status='live', Symbol='CG4452', Aliases='Dmel_ Dmel\\CG4452, anon-WO0172774.118, cg4452', description='CG4452 gene product from transcript CG4452-RE', other_designations='CG4452-PA|CG4452-PE', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9367377, end_position_on_the_genomic_accession=9370300, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39070, CurrentID=0, Status='live', Symbol='Fdx1', Aliases='Dmel_CG4205, BcDNA:RH67819, CG4205, Dmel\\CG4205, Fdxh', description='Ferredoxin 1', other_designations='CG4205-PA|CG4205-PB|Fdx1-PA|Fdx1-PB|ferredoxin|ferredoxin 1', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9370195, end_position_on_the_genomic_accession=9371238, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39071, CurrentID=0, Status='live', Symbol='Hsp67Bc', Aliases='Dmel_CG4190, CG4190, Dm-HSP67Bc, Dmel\\CG4190, HSP67, HSP67Bc, HSPB8, Hsp-G3, Hsp67BC, Hspg3, gene 3, gene-3, gene3, gene5, hsp67B, small hsp locus 67B', description='Heat shock gene 67Bc', other_designations='CG4190-PA|Hsp67Bc-PA|gene 3|heat shock gene 67Bc', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9371511, end_position_on_the_genomic_accession=9372497, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740704, CurrentID=0, Status='live', Symbol='CG34212', Aliases='Dmel_ Dmel\\CG34212', description='CG34212 gene product from transcript CG34212-RB', other_designations='CG34212-PA|CG34212-PB|uncharacterized protein', map_location='42A8-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6049874, end_position_on_the_genomic_accession=6050286, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39074, CurrentID=0, Status='live', Symbol='CG4461', Aliases='Dmel_ 148460_at, 4461, BEST:GH12586, DmCG4461, Dmel\\CG4461, Hsp20, gh12586', description='CG4461 gene product from transcript CG4461-RA', other_designations='CG4461-PA', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9375430, end_position_on_the_genomic_accession=9376293, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39075, CurrentID=0, Status='live', Symbol='Hsp26', Aliases='Dmel_CG4183, 26, 26K, CG4183, DmHsp26, Dmel\\CG4183, HSP26, hsp 26, hsp26, small hsp locus 67B', description='Heat shock protein 26', other_designations='CG4183-PA|CG4183-PB|Hsp 26|Hsp26-PA|Hsp26-PB|heat shock protein 26|heat shock protein hsp26|protein 26', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9376418, end_position_on_the_genomic_accession=9377427, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39076, CurrentID=0, Status='live', Symbol='Hsp67Ba', Aliases='Dmel_CG4167, CG4167, DROHSP, DROHSP1, DmHsp67, Dmel\\CG4167, Hsp-G1, gene 1, gene1, hsp67Ba, small hsp locus 67B', description='Heat shock gene 67Ba', other_designations='CG4167-PA|Hsp67Ba-PA|gene 1|heat shock gene 67Ba', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9377802, end_position_on_the_genomic_accession=9379534, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39077, CurrentID=0, Status='live', Symbol='Hsp23', Aliases='Dmel_CG4463, 23, CG4463, DmHSP23, DmHsp23, Dmel\\CG4463, HSP23, HSP23/IAP, Hsp-23, hsp 22K, hsp 23, hsp 23K, hsp23, hsp23/22, small hsp locus 67B', description='Heat shock protein 23', other_designations='CG4463-PA|CG4463-PB|Hsp 23|Hsp23-PA|Hsp23-PB|heat shock protein 23|heat-shock protein 23|protein 23', map_location='67B3-67B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9381882, end_position_on_the_genomic_accession=9382765, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39078, CurrentID=0, Status='live', Symbol='Hsp27', Aliases='Dmel_CG4466, 27, 27K, 28, CG4466, Dhsp27, DmHSP27, DmHsp27, Dmel\\CG4466, HSP27, HSP28, Hsp 27, Hsp28, anon-WO0140519.69, hsp 27, hsp-27, hsp26, hsp27, hsp27/26, hsp28, small hsp locus 67B', description='Heat shock protein 27', other_designations='CG4466-PA|CG4466-PB|Hsp 27|Hsp27-PA|Hsp27-PB|heat shock protein 237|heat shock protein 27|heat shock protein hsp27|heatshock protein 27|protein 27', map_location='67B3-67B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9384063, end_position_on_the_genomic_accession=9385694, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39079, CurrentID=0, Status='live', Symbol='CG4080', Aliases='Dmel_ 4080, Dmel\\CG4080', description='CG4080 gene product from transcript CG4080-RB', other_designations='CG4080-PA|CG4080-PB', map_location='67B3-67B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9386827, end_position_on_the_genomic_accession=9399227, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39081, CurrentID=0, Status='live', Symbol='Cpr67B', Aliases='Dmel_CG3672, CG3672, Dmel\\CG3672', description='Cuticular protein 67B', other_designations='CG3672-PA|Cpr67B-PA|cuticular protein 67B', map_location='67B4-67B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9404797, end_position_on_the_genomic_accession=9406174, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39082, CurrentID=0, Status='live', Symbol='CG4022', Aliases='Dmel_ Dmel\\CG4022', description='CG4022 gene product from transcript CG4022-RC', other_designations='CG4022-PA|CG4022-PB|CG4022-PC', map_location='67B4-67B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9406179, end_position_on_the_genomic_accession=9410832, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39083, CurrentID=0, Status='live', Symbol='CG3689', Aliases='Dmel_ CF I[[m]]25, Dmel\\CG3689', description='CG3689 gene product from transcript CG3689-RC', other_designations='CG3689-PB|CG3689-PC', map_location='67B5-67B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9411038, end_position_on_the_genomic_accession=9413835, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39084, CurrentID=0, Status='live', Symbol='CG3982', Aliases='Dmel_ Dmel\\CG3982, bs12d10.y1', description='CG3982 gene product from transcript CG3982-RA', other_designations='CG3982-PA', map_location='67B5-67B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9415755, end_position_on_the_genomic_accession=9417388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39085, CurrentID=0, Status='live', Symbol='aay', Aliases='Dmel_CG3705, 0423/14, 24661601, CG3705, Dmel\\CG3705, anon-WO0172774.117', description='astray', other_designations='3-phosphoserine phosphatase|CG3705-PA|aay-PA', map_location='67B5-67B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9423253, end_position_on_the_genomic_accession=9424781, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39086, CurrentID=0, Status='live', Symbol='CG3967', Aliases='Dmel_ Dmel\\CG3967', description='CG3967 gene product from transcript CG3967-RC', other_designations='CG3967-PA|CG3967-PB|CG3967-PC|CG3967-PD|CG3967-PE|CG3967-PF', map_location='67B5-67B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9413596, end_position_on_the_genomic_accession=9426758, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39088, CurrentID=0, Status='live', Symbol='RpS17', Aliases='Dmel_CG3922, BcDNA:RE44119, CG3922, D9, Dmel\\CG3922, M, M(3)67, M(3)67C, M(3)RpS17, M(3)S33, M(3)i, M(3)i(55), M(3)i[55], M(3)q, M(3L)i, M(3L)i[55], Minute, RPS17, RS17_DROME, Rp S17, Rps17, anon-EST:Posey48, l(3)67BDo, l(3)dtOA4, rpS17', description='Ribosomal protein S17', other_designations='CG3922-PB|Minute(3)67C|RpS17-PB|ribosomal protein S17', map_location='3-28.9 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9428795, end_position_on_the_genomic_accession=9429821, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39089, CurrentID=0, Status='live', Symbol='MTF-1', Aliases='Dmel_CG3743, CG3743, Dmel\\CG3743, Mtf-1, anon-WO0118547.411, dMTF-1, mtf-1', description='Metal response element-binding Transcription Factor-1', other_designations='CG3743-PA|CG3743-PB|CG3743-PC|CG3743-PD|CG3743-PE|CG3743-PF|CG3743-PG|MTF-1-PA|MTF-1-PB|MTF-1-PC|MTF-1-PD|MTF-1-PE|MTF-1-PF|MTF-1-PG|Metal-responsive transcription factor 1|metal response element binding transcription factor|metal response element binding transcription factor 1|metal response element-binding transcription Factor-1|metal response element-binding transcription factor-1|metal responsive transcription factor|metal responsive transcription factor 1|metal-regulatory transcription factor 1|metal-responsive transcription factor-1', map_location='67B5-67B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9430312, end_position_on_the_genomic_accession=9439490, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39090, CurrentID=0, Status='live', Symbol='Bet3', Aliases='Dmel_CG3911, BET3p, CG3911, Dmel\\CG3911, dBet3, Bet3', description='BET3 ortholog (S. cerevisiae)', other_designations='BET3 ortholog|Bet3-PA|CG3911-PA', map_location='67B6-67B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9438482, end_position_on_the_genomic_accession=9442267, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39091, CurrentID=0, Status='live', Symbol='Nf-YA', Aliases='Dmel_CG3891, CG3891, Dmel\\CG3891, NF-YA, dNF-YA', description='Nuclear factor Y-box A', other_designations='CG3891-PA|Nf-YA-PA|nuclear factor Y-box A', map_location='67B6-67B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9440056, end_position_on_the_genomic_accession=9442267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39092, CurrentID=0, Status='live', Symbol='CG33926', Aliases='Dmel_ Dmel\\CG33926', description='CG33926 gene product from transcript CG33926-RB', other_designations='CG33926-PA|CG33926-PB|uncharacterized protein', map_location='67B6-67B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9443232, end_position_on_the_genomic_accession=9446388, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39093, CurrentID=0, Status='live', Symbol='CG3529', Aliases='Dmel_ DmTOM1, Dmel\\CG3529', description='CG3529 gene product from transcript CG3529-RB', other_designations='CG3529-PB|CG3529-PC|CG3529-PD', map_location='67B7-67B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9446864, end_position_on_the_genomic_accession=9450785, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39094, CurrentID=0, Status='live', Symbol='CG3448', Aliases='Dmel_ Dmel\\CG3448', description='CG3448 gene product from transcript CG3448-RB', other_designations='CG3448-PB', map_location='67B7-67B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9450545, end_position_on_the_genomic_accession=9451572, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39095, CurrentID=0, Status='live', Symbol='phol', Aliases='Dmel_CG3445, CG3445, Dmel\\CG3445, PHOL, Phol', description='pleiohomeotic like', other_designations='CG3445-PA|CG3445-PB|Pleiohomeotic-like|pho-like|phol-PA|phol-PB', map_location='67B7-67B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9452803, end_position_on_the_genomic_accession=9456281, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39096, CurrentID=0, Status='live', Symbol='CG3437', Aliases='Dmel_ Dmel\\CG3437', description='CG3437 gene product from transcript CG3437-RB', other_designations='CG3437-PB', map_location='67B8-67B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9456342, end_position_on_the_genomic_accession=9457669, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39097, CurrentID=0, Status='live', Symbol='CG3552', Aliases='Dmel_ Dmel\\CG3552', description='CG3552 gene product from transcript CG3552-RD', other_designations='CG3552-PA|CG3552-PD', map_location='67B8-67B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9456318, end_position_on_the_genomic_accession=9459370, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39098, CurrentID=0, Status='live', Symbol='CG3434', Aliases='Dmel_ Dmel\\CG3434', description='CG3434 gene product from transcript CG3434-RA', other_designations='CG3434-PA', map_location='67B8-67B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9459369, end_position_on_the_genomic_accession=9460799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19836003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836003, CurrentID=0, Status='live', Symbol='CG45069', Aliases='Dmel_ CG3419, Dmel\\CG45069', description='CG45069 gene product from transcript CG45069-RB', other_designations='CG45069-PA|CG45069-PB|uncharacterized protein', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24411870, end_position_on_the_genomic_accession=24415295, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39102, CurrentID=0, Status='live', Symbol='Uch-L5', Aliases='Dmel_CG3431, BcDNA.LD24440, BcDNA:LD24440, CG3431, Dmel\\CG3431, Q9XZ61, UCH, UCH 37, Uch-L3, Uch37, dUCH37, l(3)j2B8, p37A, uch-L3', description='Ubiquitin carboxy-terminal hydrolase L5 ortholog', other_designations='CG3431-PA|CG3431-PB|Uch-L5-PA|Uch-L5-PB|ubiquitin C-terminal hydrolase|ubiquitin C-terminal hydrolase related|ubiquitin carboxy-terminal hydrolase L5 ortholog', map_location='67B9-67B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9475493, end_position_on_the_genomic_accession=9476965, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39103, CurrentID=0, Status='live', Symbol='Jarid2', Aliases='Dmel_CG3654, CG3654, Dmel\\CG3654, JARID2, dJARID2, dJmj, djmj', description='Jumonji, AT rich interactive domain 2', other_designations='CG3654-PD|Jarid2-PD|jumonji', map_location='67B9-67B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9477038, end_position_on_the_genomic_accession=9486530, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39104, CurrentID=0, Status='live', Symbol='pall', Aliases='Dmel_CG3428, CG3428, Dmel\\CG3428, FBXO1, Pall', description='pallbearer', other_designations='CG3428-PA|pall-PA', map_location='67B10-67B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9486448, end_position_on_the_genomic_accession=9487790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39105, CurrentID=0, Status='live', Symbol='CG32036', Aliases='Dmel_ CG3426, Dmel\\CG32036', description='CG32036 gene product from transcript CG32036-RC', other_designations='CG32036-PC|uncharacterized protein', map_location='67B10-67B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9488050, end_position_on_the_genomic_accession=9489233, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39106, CurrentID=0, Status='live', Symbol='path', Aliases='Dmel_CG3424, CG3424, Dmel\\CG3424, Path, anon-WO0118547.271, anon-WO0118547.370, pth', description='pathetic', other_designations='CG3424-PA|CG3424-PD|CG3424-PE|path-PA|path-PD|path-PE', map_location='67B10-67B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9494555, end_position_on_the_genomic_accession=9505328, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39107, CurrentID=0, Status='live', Symbol='CG3408', Aliases='Dmel_ CT11439, Dmel\\CG3408', description='CG3408 gene product from transcript CG3408-RB', other_designations='CG3408-PA|CG3408-PB', map_location='67B10-67B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9506818, end_position_on_the_genomic_accession=9509583, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39108, CurrentID=0, Status='live', Symbol='RpS9', Aliases='Dmel_CG3395, CG3395, Dmel\\CG3395, Rp S9, RpL11, S9, chr3L:9483841..9483937, chr3L:9484246..9484474', description='Ribosomal protein S9', other_designations='CG3395-PA|CG3395-PB|CG3395-PD|CG3395-PE|RpS9-PA|RpS9-PB|RpS9-PD|RpS9-PE|ribosomal protein S9', map_location='67B11-67B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9509793, end_position_on_the_genomic_accession=9511531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39109, CurrentID=0, Status='live', Symbol='Or67a', Aliases='Dmel_CG12526, 67B.1, 67a, CG12526, DOR84, DmOR67a, Dmel\\CG12526', description='Odorant receptor 67a', other_designations='CG12526-PA|Or67a-PA|odorant receptor 67a', map_location='67B11-67B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9529067, end_position_on_the_genomic_accession=9531386, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39110, CurrentID=0, Status='live', Symbol='Ir67a', Aliases='Dmel_CG12525, CG12525, DmelIR67a, Dmel\\CG12525, IR67a', description='Ionotropic receptor 67a', other_designations='CG12525-PB|Ir67a-PB|ionotropic receptor 67a', map_location='67B13-67B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9542245, end_position_on_the_genomic_accession=9543999, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740743, CurrentID=0, Status='live', Symbol='CG34436', Aliases='Dmel_ CG16918, Dmel\\CG34436, Dmel_CG16918, SPH201', description='CG34436 gene product from transcript CG34436-RA', other_designations='CG34436-PA|uncharacterized protein', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28282580, end_position_on_the_genomic_accession=28283548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740745, CurrentID=0, Status='live', Symbol='CG34444', Aliases='Dmel_ CG30070, Dmel\\CG34444, Dmel_CG30070', description='CG34444 gene product from transcript CG34444-RA', other_designations='CG34444-PA|uncharacterized protein', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14369217, end_position_on_the_genomic_accession=14370326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740746, CurrentID=0, Status='live', Symbol='CG34172', Aliases='Dmel_ Dmel\\CG34172', description='CG34172 gene product from transcript CG34172-RC', other_designations='CG34172-PA|CG34172-PB|CG34172-PC|CG34172-PD|uncharacterized protein', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2192525, end_position_on_the_genomic_accession=2197912, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39115, CurrentID=0, Status='live', Symbol='CG3088', Aliases='Dmel_ Dmel\\CG3088, SPH188', description='CG3088 gene product from transcript CG3088-RA', other_designations='CG3088-PA', map_location='67C2-67C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9600490, end_position_on_the_genomic_accession=9601325, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39116, CurrentID=0, Status='live', Symbol='CG3306', Aliases='Dmel_ Dmel\\CG3306', description='CG3306 gene product from transcript CG3306-RA', other_designations='CG3306-PA', map_location='67C2-67C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9601523, end_position_on_the_genomic_accession=9602548, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39118, CurrentID=0, Status='live', Symbol='LanB2', Aliases='Dmel_CG3322, CG3322, CT11145, DROLAMB2A, Dmel\\CG3322, LAMB2A, LM-B2/gamma, LMG1, Lam-B2, anon-EST:fe2H3, gamma1, knod, l(3)67BDb, lamB2, lanB2', description='Laminin B2', other_designations='CG3322-PA|CG3322-PB|LamininB2|LanB2-PA|LanB2-PB|anon-fast-evolving-2H3|knodel|laminin|laminin B2|laminin B2 subunit|laminin gamma|laminin gamma-chain|laminin gamma1|laminin gamma2|lamininB|lethal (3) 67BDb', map_location='67C2-67D13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9614914, end_position_on_the_genomic_accession=9623980, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39119, CurrentID=0, Status='live', Symbol='CG3335', Aliases='Dmel_ Dm-RBD-1, Dmel\\CG3335, Q9VT19, cg3335', description='CG3335 gene product from transcript CG3335-RA', other_designations='CG3335-PA', map_location='67C2-67C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9624237, end_position_on_the_genomic_accession=9627321, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39120, CurrentID=0, Status='live', Symbol='Or67b', Aliases='Dmel_CG14176, 67B.2, 67b, CG14176, Dmel\\CG14176', description='Odorant receptor 67b', other_designations='CG14176-PB|Or67b-PB|odorant receptor 67b', map_location='67C2-67C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9627316, end_position_on_the_genomic_accession=9629193, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39121, CurrentID=0, Status='live', Symbol='CG8336', Aliases='Dmel_ Dmel\\CG8336', description='CG8336 gene product from transcript CG8336-RB', other_designations='CG8336-PA|CG8336-PB|CG8336-PC|CG8336-PD', map_location='67C2-67C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9628676, end_position_on_the_genomic_accession=9631795, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39122, CurrentID=0, Status='live', Symbol='fry', Aliases='Dmel_CG32045, CG14171, CG32045, CG6774, CG6780, Dmel\\CG32045, Fry, RE-B, SS3-7, l(3)02240, yeti', description='furry', other_designations='CG32045-PB|CG32045-PC|CG32045-PE|CG32045-PF|CG32045-PG|CG32045-PH|CG32045-PI|CG32045-PJ|fry-PB|fry-PC|fry-PE|fry-PF|fry-PG|fry-PH|fry-PI|fry-PJ|suppressor of GMR-sina 3-7', map_location='67C3-67C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9631834, end_position_on_the_genomic_accession=9679103, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  39123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39123, CurrentID=0, Status='live', Symbol='CG8329', Aliases='Dmel_ Dmel\\CG8329, SP170', description='CG8329 gene product from transcript CG8329-RA', other_designations='CG8329-PA', map_location='67C3-67C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9638112, end_position_on_the_genomic_accession=9639021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39124, CurrentID=0, Status='live', Symbol='CG18179', Aliases='Dmel_ BEST:LP02275, Dmel\\CG18179, SP152', description='CG18179 gene product from transcript CG18179-RA', other_designations='CG18179-PA|uncharacterized protein', map_location='67C3-67C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9639271, end_position_on_the_genomic_accession=9640140, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39125, CurrentID=0, Status='live', Symbol='CG18180', Aliases='Dmel_ Dmel\\CG18180, SP151', description='CG18180 gene product from transcript CG18180-RA', other_designations='CG18180-PA|uncharacterized protein', map_location='67C3-67C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9641013, end_position_on_the_genomic_accession=9641913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39127, CurrentID=0, Status='live', Symbol='CNMaR', Aliases='Dmel_CG33696, CG16726, CG32047, CG33696, Dm CG33696, Dmel\\CG33696', description='CNMamide Receptor', other_designations='CG33696-PA|CNMaR-PA|CNMamide receptor', map_location='67C3-67C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9644109, end_position_on_the_genomic_accession=9659433, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39129, CurrentID=0, Status='live', Symbol='CG16717', Aliases='Dmel_ Dmel\\CG16717', description='CG16717 gene product from transcript CG16717-RA', other_designations='CG16717-PA|uncharacterized protein', map_location='67C4-67C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9679201, end_position_on_the_genomic_accession=9681060, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39130, CurrentID=0, Status='live', Symbol='alphaTub67C', Aliases='Dmel_CG8308, 4t, ALPHA 67C, CG8308, D.m.ALPHA-67C, DTA2, Dmel\\CG8308, Fs(3)Kav, Fs(3)Kavar, Fs(3)Sz14, Fs(3)Sz22, Fs(3)Tom, Fs(3)Tomaj, Kavar, T, Tom, Tub, aTub67C, alpha-Tub, alpha-Tub67C, alpha-tub, alpha-tub67C, alpha-tubulin, alpha-tubulin[4], alpha4, alpha4-t, alpha4-tubulin, alpha4Tub67C, alpha4t, alpha67C, alphaTUB, alphaTub, alphaTub 67C, alphaTub4, alphaTub[67C], alphatub, alphatub67C, gamma-Tub, mum-tub67C, tub, tub67c', description='alpha-Tubulin at 67C', other_designations='CG8308-PA|alpha tubulin|alpha tubulin 67C|alpha-Tubulin|alpha4-tubulin|alphaT|alphaTub67C-PA|alphaTubulin 67C|alphaTubulin67C|alphatubulin|female sterile (3) kavar|kavar|tomaj|tubulin|tubulin at 67C|tubulin67C', map_location='67C4-67C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9681268, end_position_on_the_genomic_accession=9683370, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39131, CurrentID=0, Status='live', Symbol='CG16719', Aliases='Dmel_ Dmel\\CG16719', description='CG16719 gene product from transcript CG16719-RA', other_designations='CG16719-PA|uncharacterized protein', map_location='67C4-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9685171, end_position_on_the_genomic_accession=9686250, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39132, CurrentID=0, Status='live', Symbol='CG6767', Aliases='Dmel_ Dmel\\CG6767', description='CG6767 gene product from transcript CG6767-RA', other_designations='CG6767-PA|CG6767-PB|CG6767-PC|CG6767-PD|CG6767-PF|CG6767-PG|CG6767-PH|CG6767-PI|CG6767-PJ', map_location='67C4-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9683579, end_position_on_the_genomic_accession=9696700, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39133, CurrentID=0, Status='live', Symbol='Ubc4', Aliases='Dmel_CG8284, CG8284, Dmel\\CG8284, E2, UbcD4', description='Ubiquitin conjugating enzyme 4', other_designations='CG8284-PA|CG8284-PB|CG8284-PC|Ubc4-PA|Ubc4-PB|Ubc4-PC|ubiquitin conjugating enzyme 4', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9698336, end_position_on_the_genomic_accession=9700481, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39134, CurrentID=0, Status='live', Symbol='CG6761', Aliases='Dmel_ Dmel\\CG6761', description='CG6761 gene product from transcript CG6761-RA', other_designations='CG6761-PA|CG6761-PB', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9700517, end_position_on_the_genomic_accession=9703595, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39135, CurrentID=0, Status='live', Symbol='CG16711', Aliases='Dmel_ Dmel\\CG16711', description='CG16711 gene product from transcript CG16711-RB', other_designations='CG16711-PB|CG16711-PC|uncharacterized protein', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9704331, end_position_on_the_genomic_accession=9710329, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39136, CurrentID=0, Status='live', Symbol='SH3PX1', Aliases='Dmel_CG6757, CG6757, DSH3PX, DSH3PX1, DmSNX18, Dmel\\CG6757, dSH3PX1, p63', description='SH3PX1 ortholog (H. sapiens)', other_designations='CG6757-PA|SH3PX1 ortholog|SH3PX1-PA', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9710422, end_position_on_the_genomic_accession=9713827, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39137, CurrentID=0, Status='live', Symbol='vsg', Aliases='Dmel_CG16707, BEST:GH15083, CG16707, Dmel\\CG16707, FBgn0045823, GH15083, Vsg, anon-WO0118547.164', description='visgun', other_designations='CG16707-PA|CG16707-PB|CG16707-PC|CG16707-PD|CG16707-PE|endolyn|vsg-PA|vsg-PB|vsg-PC|vsg-PD|vsg-PE', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9714324, end_position_on_the_genomic_accession=9717716, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39138, CurrentID=0, Status='live', Symbol='CG18178', Aliases='Dmel_ Dmel\\CG18178', description='CG18178 gene product from transcript CG18178-RA', other_designations='CG18178-PA|CG18178-PB|uncharacterized protein', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9717786, end_position_on_the_genomic_accession=9719144, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39139, CurrentID=0, Status='live', Symbol='CG14174', Aliases='Dmel_ Dmel\\CG14174', description='CG14174 gene product from transcript CG14174-RA', other_designations='CG14174-PA|uncharacterized protein', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9719328, end_position_on_the_genomic_accession=9720464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39141, CurrentID=0, Status='live', Symbol='defl', Aliases='Dmel_CG18176, CG18176, Dmel\\CG18176, Int7, IntS7', description='deflated', other_designations='CG18176-PA|defl-PA|integrator 7', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9723423, end_position_on_the_genomic_accession=9726706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39142, CurrentID=0, Status='live', Symbol='Naa60', Aliases='Dmel_CG18177, CG18177, DmAAF50213, Dmel\\CG18177, Hat4, NAA60, NatF, dNAA60, dNaa60, dNatF', description='N(alpha)-acetyltransferase 60', other_designations='CG18177-PA|CG18177-PB|CG18177-PC|Naa60-PA|Naa60-PB|Naa60-PC', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9726895, end_position_on_the_genomic_accession=9729650, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39143, CurrentID=0, Status='live', Symbol='ATPsynB', Aliases='Dmel_CG8189, ATP syn-b, ATPF, ATPsyn-b, CG8189, Dmel\\CG8189', description='ATP synthase, subunit B', other_designations='ATP synthase B chain|ATP synthase B chain, mt precursor|ATP synthase, b|ATPsynB-PA|CG8189-PA|F0/F1 ATP synthase B|Fo-ATP synthase subunit b', map_location='67C5-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9729999, end_position_on_the_genomic_accession=9731224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39144, CurrentID=0, Status='live', Symbol='CG6749', Aliases='Dmel_ Dmel\\CG6749', description='CG6749 gene product from transcript CG6749-RA', other_designations='CG6749-PA', map_location='67C5-67C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9731238, end_position_on_the_genomic_accession=9733807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39145, CurrentID=0, Status='live', Symbol='CG42268', Aliases='Dmel_ CG14170, CG32044, CG32046, CG6740, Dmel\\CG42268, Dmel_CG32044, Dmel_CG32046', description='CG42268 gene product from transcript CG42268-RI', other_designations='CG42268-PD|CG42268-PF|CG42268-PH|CG42268-PI|CG42268-PJ|CG42268-PK|CG42268-PL|CG42268-PM|uncharacterized protein', map_location='67C6-67C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9734012, end_position_on_the_genomic_accession=9760395, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  39147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39147, CurrentID=0, Status='live', Symbol='CG8177', Aliases='Dmel_ BcDNA:RE59468, DAE, Dmel\\CG8177, anon-EST:Posey259', description='CG8177 gene product from transcript CG8177-RM', other_designations='CG8177-PA|CG8177-PB|CG8177-PC|CG8177-PD|CG8177-PE|CG8177-PF|CG8177-PG|CG8177-PH|CG8177-PI|CG8177-PJ|CG8177-PK|CG8177-PL|CG8177-PM', map_location='67C7-67C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9763544, end_position_on_the_genomic_accession=9786466, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  39148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39148, CurrentID=0, Status='live', Symbol='Zasp67', Aliases='Dmel_CG14168, CG14168, Dmel\\CG14168', description='CG14168 gene product from transcript CG14168-RE', other_designations='CG14168-PD|CG14168-PE|CG14168-PG|CG14168-PH|Zasp67-PD|Zasp67-PE|Zasp67-PG|Zasp67-PH', map_location='67C8-67C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9787933, end_position_on_the_genomic_accession=9797687, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  39149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39149, CurrentID=0, Status='live', Symbol='Ilp1', Aliases='Dmel_CG14173, CG14173, DILP, DILP-1, DILP1, Dilp, Dilp 1, Dilp-1, Dilp1, Dmel\\CG14173, Ilp, Ilp-1, dILP1, dilp, dilp-1, dilp1, ilp1', description='Insulin-like peptide 1', other_designations='CG14173-PA|Ilp1-PA|drosophila insulin-like peptide 1|insulin-like peptide', map_location='67C8-67C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9798420, end_position_on_the_genomic_accession=9799050, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39150, CurrentID=0, Status='live', Symbol='Ilp2', Aliases='Dmel_CG8167, CG8167, DILP, DILP-2, DILP2, Dilp, Dilp 2, Dilp-2, Dilp2, Dmel\\CG8167, ILP2, IRP, Ilp, Ilp-2, dILP-2, dILP2, dIlp2, dilp, dilp-2, dilp2, ilp2', description='Insulin-like peptide 2', other_designations='CG8167-PA|Ilp2-PA|Insulin- related peptide 2|Insulin-related peptide|drosophila insulin like peptide 2|drosophila insulin-like peptide 2|insulin-like 2|insulin-like peptide|insulin-like peptide-2', map_location='67C8-67C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9799698, end_position_on_the_genomic_accession=9800440, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39151, CurrentID=0, Status='live', Symbol='Ilp3', Aliases='Dmel_CG14167, CG14167, DILP, DILP-3, DILP3, Dilp, Dilp 3, Dilp-3, Dilp3, Dmel\\CG14167, Ilp, Ilp-3, dILP-3, dILP3, dIlp3, dilp, dilp-3, dilp3, ilp3', description='Insulin-like peptide 3', other_designations='CG14167-PA|Ilp3-PA|drosophila insulin-like peptide 3|insulin-like peptide', map_location='67C8-67C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9801298, end_position_on_the_genomic_accession=9802120, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39152, CurrentID=0, Status='live', Symbol='Ilp4', Aliases='Dmel_CG6736, A, BcDNA:RE27042, CG6736, DILP, DILP-4, DILP4, Dilp, Dilp 4, Dilp4, Dirlp, Dmel\\CG6736, Ilp, Ilp-4, Mes3, anon-EST:CLB3, dILP4, dilp, dilp-4, dilp4, ilp4, mes3', description='Insulin-like peptide 4', other_designations='CG6736-PA|Ilp4-PA|drosophila insulin-like peptide 4|drosophila insulin/relaxin-like peptide|insulin-like peptide|insulin-like peptide-4', map_location='67C8-67C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9803399, end_position_on_the_genomic_accession=9804263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39153, CurrentID=0, Status='live', Symbol='CG43897', Aliases='Dmel_ BcDNA:RE55923, BcDNA:RH19933, CG14172, CG32049, CG32050, CG32051, CG33205, CG8154, CG8163, Dmel\\CG43897, Dmel_CG32050, Dmel_CG33205', description='CG43897 gene product from transcript CG43897-RL', other_designations='CG43897-PA|CG43897-PB|CG43897-PC|CG43897-PE|CG43897-PF|CG43897-PG|CG43897-PH|CG43897-PI|CG43897-PJ|CG43897-PK|CG43897-PL|CG43897-PM|CG43897-PN|CG43897-PO|CG43897-PP|CG43897-PQ|CG43897-PR|uncharacterized protein', map_location='67C9-67C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9813847, end_position_on_the_genomic_accession=9834345, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  39156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39156, CurrentID=0, Status='live', Symbol='I-2', Aliases='Dmel_CG10574, CG10574, DI-2, Dm I-2, Dmel\\CG10574Dm, I-2PP1, SET, I-2', description='Inhibitor-2', other_designations='CG10574-PA|CG10574-PB|I-2-PA|I-2-PB|inhibitor-2|phosphatase inhibitor-2', map_location='67C10-67C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9834367, end_position_on_the_genomic_accession=9836805, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39157, CurrentID=0, Status='live', Symbol='Cdk8', Aliases='Dmel_CG10572, CDK8, CG10572, DmCdk8, Dmel\\CG10572, cdk8, dCdk8, dTRAP56, p58', description='Cyclin-dependent kinase 8', other_designations='CG10572-PA|Cdk8-PA', map_location='67C10-67C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9837080, end_position_on_the_genomic_accession=9838670, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39158, CurrentID=0, Status='live', Symbol='RasGAP1', Aliases='Dmel_CG6721, CG6721, Dmel\\CG6721, GAP, GAP1, Gap, Gap 1, Gap1, Ras-GAP, RasGAP, gap, gap1, mip, rI533, rasGAP, sxt', description='Ras GTPase activating protein 1', other_designations='CG6721-PA|CG6721-PB|CG6721-PC|GTPase activating protein 1|GTPase activating protein1|GTPase-activating protein 1|RasGAP1-PA|RasGAP1-PB|RasGAP1-PC|more inner photoreceptors|seven extra|sextra', map_location='67C10-67C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9839093, end_position_on_the_genomic_accession=9855285, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39159, CurrentID=0, Status='live', Symbol='CG10809', Aliases='Dmel_ Dmel\\CG10809', description='CG10809 gene product from transcript CG10809-RB', other_designations='CG10809-PA|CG10809-PB|uncharacterized protein', map_location='67C11-67C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9857383, end_position_on_the_genomic_accession=9859889, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39160, CurrentID=0, Status='live', Symbol='iPLA2-VIA', Aliases='Dmel_CG6718, CG6718, Dmel\\CG6718', description='calcium-independent phospholipase A2 VIA', other_designations='CG6718-PA|CG6718-PB|CG6718-PC|CG6718-PD|iPLA2-VIA-PA|iPLA2-VIA-PB|iPLA2-VIA-PC|iPLA2-VIA-PD|phospholipase A[[2]]', map_location='67C11-67C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9860217, end_position_on_the_genomic_accession=9864349, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39161, CurrentID=0, Status='live', Symbol='CG8108', Aliases='Dmel_ BcDNA:GH07921, Dmel\\CG8108', description='CG8108 gene product from transcript CG8108-RA', other_designations='CG8108-PA|CG8108-PB', map_location='67C11-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9864639, end_position_on_the_genomic_accession=9870674, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39164, CurrentID=0, Status='live', Symbol='Taf2', Aliases='Dmel_CG6711, CG6711, Dmel\\CG6711, TAF, TAF150, TAF2, TAFII150, TAF[II]150, TAF[[II]], TAF[[II]]150, TFIID, Taf150, TafII150, Taf[[II]]150, dTAF II 150, dTAF150, dTAFII150, dTAF[[II]]150', description='TBP-associated factor 2', other_designations='CG6711-PA|CG6711-PB|TBP-associated factor|TBP-associated factor 150kD|Taf2-PA|Taf2-PB', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9884411, end_position_on_the_genomic_accession=9888683, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39165, CurrentID=0, Status='live', Symbol='CalpB', Aliases='Dmel_CG8107, CALPB, CG8107, Dmel\\CG8107, calpB', description='Calpain-B', other_designations='CG8107-PA|CG8107-PB|CalpB-PA|CalpB-PB|calpain B|calpainB', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9888762, end_position_on_the_genomic_accession=9893319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39166, CurrentID=0, Status='live', Symbol='CG6709', Aliases='Dmel_ Dmel\\CG6709', description='CG6709 gene product from transcript CG6709-RB', other_designations='CG6709-PA|CG6709-PB', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9893188, end_position_on_the_genomic_accession=9894743, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39167, CurrentID=0, Status='live', Symbol='CG14164', Aliases='Dmel_ Dmel\\CG14164', description='CG14164 gene product from transcript CG14164-RB', other_designations='CG14164-PA|CG14164-PB|uncharacterized protein', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9893188, end_position_on_the_genomic_accession=9894743, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39168, CurrentID=0, Status='live', Symbol='CG6707', Aliases='Dmel_ Dmel\\CG6707', description='CG6707 gene product from transcript CG6707-RA', other_designations='CG6707-PA|CG6707-PB|CG6707-PC', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9894807, end_position_on_the_genomic_accession=9898575, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39169, CurrentID=0, Status='live', Symbol='nudE', Aliases='Dmel_CG8104, CG8104, Dmel\\CG8104, D NudE', description='CG8104 gene product from transcript CG8104-RE', other_designations='CG8104-PA|CG8104-PB|CG8104-PC|CG8104-PD|CG8104-PE|CG8104-PF|nudE-PA|nudE-PB|nudE-PC|nudE-PD|nudE-PE|nudE-PF', map_location='67D1-67D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9898703, end_position_on_the_genomic_accession=9903031, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  5740802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740802, CurrentID=0, Status='live', Symbol='CG34297', Aliases='Dmel_ Dmel\\CG34297', description='CG34297 gene product from transcript CG34297-RA', other_designations='CG34297-PA|uncharacterized protein', map_location='84A5-84A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6747068, end_position_on_the_genomic_accession=6748007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39172, CurrentID=0, Status='live', Symbol='CG6685', Aliases='Dmel_ Dmel\\CG6685', description='CG6685 gene product from transcript CG6685-RB', other_designations='CG6685-PA|CG6685-PB', map_location='67D2-67D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9966603, end_position_on_the_genomic_accession=9968317, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39173, CurrentID=0, Status='live', Symbol='Dronc', Aliases='Dmel_CG8091, CG8090, CG8091, DRONC, Dmel\\CG8091/Casp9, Nc, Nc\\Dronc, anon-EST:GreesD6, anon-EST:GressD6, anon-WO0118547.313, caspase-9, dronc, Dronc', description='Death regulator Nedd2-like caspase', other_designations='CG8091-PA|Caspase-9-like|Dronc-PA|Nedd2-like caspase|caspase-9|death regulator Nedd2-like caspase|dronc|drosophila Caspase-9 homologue|drosophila Nedd2-like caspase|nedd 2-like caspase', map_location='67D2-67D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9968479, end_position_on_the_genomic_accession=9971002, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39174, CurrentID=0, Status='live', Symbol='CG6674', Aliases='Dmel_ Dmel\\CG6674', description='CG6674 gene product from transcript CG6674-RB', other_designations='CG6674-PA|CG6674-PB', map_location='67D2-67D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9971151, end_position_on_the_genomic_accession=9972054, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39175, CurrentID=0, Status='live', Symbol='vnc', Aliases='Dmel_CG11989, ARD1, Ard1, CG11989, DmAAF50178, Dmel\\CG11989, Dromel_CG11989_FBtr0076348_ard1_mORF, NAA10, ard1, dArd1, l(3)67BDk', description='variable nurse cells', other_designations='Ard1|CG11989-PA|CG11989-PB|CG11989-PC|CG11989-PD|CG11989-PE|arrest defective 1|arrest defective1|lethal (3) 67BDk|variable nurse cell|vnc-PA|vnc-PB|vnc-PC|vnc-PD|vnc-PE', map_location='67B1-67D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9972312, end_position_on_the_genomic_accession=9974023, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39177, CurrentID=0, Status='live', Symbol='CG14160', Aliases='Dmel_ Dmel\\CG14160', description='CG14160 gene product from transcript CG14160-RA', other_designations='CG14160-PA|uncharacterized protein', map_location='67D8-67D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10086747, end_position_on_the_genomic_accession=10089400, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39178, CurrentID=0, Status='live', Symbol='CG32053', Aliases='Dmel_ CG6645, Dmel\\CG32053', description='CG32053 gene product from transcript CG32053-RA', other_designations='CG32053-PA|uncharacterized protein', map_location='67D8-67D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10090170, end_position_on_the_genomic_accession=10092346, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39179, CurrentID=0, Status='live', Symbol='CG42825', Aliases='Dmel_ CG6640, Dmel\\CG42825, Dmel_CG6640, GH19118, anon-WO0140519.30', description='CG42825 gene product from transcript CG42825-RA', other_designations='CG42825-PA|uncharacterized protein', map_location='67D8-67D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10095754, end_position_on_the_genomic_accession=10098115, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39180, CurrentID=0, Status='live', Symbol='dpr10', Aliases='Dmel_CG32057, BcDNA:RE37920, CG14158, CG14159, CG32057, CT33762, Dmel\\CG32057, Dpr-10, Dpr10', description='defective proboscis extension response 10', other_designations='CG32057-PA|CG32057-PB|CG32057-PC|CG32057-PD|dpr10-PA|dpr10-PB|dpr10-PC|dpr10-PD', map_location='67D9-67D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10147454, end_position_on_the_genomic_accession=10186108, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  5740813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740813, CurrentID=0, Status='live', Symbol='CG34447', Aliases='Dmel_ CG31684, CG9966, Dmel\\CG34447, Dmel_CG31684', description='CG34447 gene product from transcript CG34447-RA', other_designations='CG34447-PA|CG34447-PB|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2418422, end_position_on_the_genomic_accession=2420013, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39183, CurrentID=0, Status='live', Symbol='CG6628', Aliases='Dmel_ Dmel\\CG6628', description='CG6628 gene product from transcript CG6628-RA', other_designations='CG6628-PA', map_location='67D9-67D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10165809, end_position_on_the_genomic_accession=10167023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740816, CurrentID=0, Status='live', Symbol='DIP-delta', Aliases='Dmel_CG34391, 34391, CG15214, CG34391, CG4814, CT15465, CT35147, Dmel\\CG34391, Dmel_CG15214, Dmel_CG4814', description='Dpr-interacting protein delta', other_designations='CG34391-PB|CG34391-PD|CG34391-PE|DIP-delta-PB|DIP-delta-PD|DIP-delta-PE', map_location='64D1-64D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5424109, end_position_on_the_genomic_accession=5442399, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  5740817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740817, CurrentID=0, Status='live', Symbol='CG34283', Aliases='Dmel_ Dmel\\CG34283', description='CG34283 gene product from transcript CG34283-RA', other_designations='CG34283-PA|uncharacterized protein', map_location='91C3-91C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18721967, end_position_on_the_genomic_accession=18722590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39186, CurrentID=0, Status='live', Symbol='CG8065', Aliases='Dmel_ Dmel\\CG8065', description='CG8065 gene product from transcript CG8065-RC', other_designations='CG8065-PB|CG8065-PC', map_location='67D10-67D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10215665, end_position_on_the_genomic_accession=10216980, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317827, CurrentID=0, Status='live', Symbol='CG32023', Aliases='Dmel_ Dmel\\CG32023', description='CG32023 gene product from transcript CG32023-RA', other_designations='CG32023-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8803602, end_position_on_the_genomic_accession=8804182, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39188, CurrentID=0, Status='live', Symbol='CG32055', Aliases='Dmel_ CG6590, CT20492, Dmel\\CG32055', description='CG32055 gene product from transcript CG32055-RA', other_designations='CG32055-PA|uncharacterized protein', map_location='67D11-67D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10218504, end_position_on_the_genomic_accession=10220164, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39189, CurrentID=0, Status='live', Symbol='Or67c', Aliases='Dmel_CG14156, 67D.1, 67c, CG14156, DOR77, DmOR67C, Dmel\\CG14156', description='Odorant receptor 67c', other_designations='CG14156-PA|Or67c-PA|odorant receptor 67c', map_location='67D11-67D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10243320, end_position_on_the_genomic_accession=10244706, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39191, CurrentID=0, Status='live', Symbol='Or67d', Aliases='Dmel_CG14157, 67d, CG14157, Dmel\\CG14157, OR67d', description='Odorant receptor 67d', other_designations='CG14157-PA|Or67d-PA|odorant receptor 67d', map_location='67D12-67D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10273204, end_position_on_the_genomic_accession=10274624, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39193, CurrentID=0, Status='live', Symbol='CG12362', Aliases='Dmel_ AAN11912.1, Dmel\\CG12362', description='CG12362 gene product from transcript CG12362-RB', other_designations='CG12362-PA|CG12362-PB|uncharacterized protein', map_location='67E2-67E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10412818, end_position_on_the_genomic_accession=10414774, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39194, CurrentID=0, Status='live', Symbol='Ir67b', Aliases='Dmel_CG12303, CG12303, DmelIR67b, Dmel\\CG12303, IR67b', description='Ionotropic receptor 67b', other_designations='CG12303-PA|Ir67b-PA|ionotropic receptor 67b', map_location='67E3-67E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10429689, end_position_on_the_genomic_accession=10431983, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39195, CurrentID=0, Status='live', Symbol='Gem3', Aliases='Dmel_CG6539, BcDNA.LD05563, BcDNA:LD05563, CG6539, DHH1, Dhh1, DmRH22, Dmel\\CG6539, dGem3, l(3)rL562', description='Gemin3', other_designations='CG6539-PA|CG6539-PB|DEAD/DEAH RNA helicase 1|Gem3-PA|Gem3-PB|dGemin3|gemin3', map_location='67E3-67E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10459081, end_position_on_the_genomic_accession=10463101, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39196, CurrentID=0, Status='live', Symbol='CG32061', Aliases='Dmel_ CG8040, Dmel\\CG32061', description='CG32061 gene product from transcript CG32061-RA', other_designations='CG32061-PA|uncharacterized protein', map_location='67E3-67E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10463816, end_position_on_the_genomic_accession=10464635, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39197, CurrentID=0, Status='live', Symbol='CG6527', Aliases='Dmel_ BcDNA:AT13283, Dmel\\CG6527', description='CG6527 gene product from transcript CG6527-RB', other_designations='CG6527-PA|CG6527-PB', map_location='67E4-67E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10557023, end_position_on_the_genomic_accession=10559082, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39198, CurrentID=0, Status='live', Symbol='A2bp1', Aliases='Dmel_CG32062, A2BP1, BcDNA:GM06073, CG14155, CG18441, CG32062, Dmel\\CG32062, cg18441, dA2BP1', description='Ataxin-2 binding protein 1', other_designations='A2bp1-PE|A2bp1-PF|A2bp1-PH|A2bp1-PI|A2bp1-PJ|A2bp1-PK|A2bp1-PL|A2bp1-PM|CG32062-PE|CG32062-PF|CG32062-PH|CG32062-PI|CG32062-PJ|CG32062-PK|CG32062-PL|CG32062-PM', map_location='67E4-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10481412, end_position_on_the_genomic_accession=10594012, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  5740831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740831, CurrentID=0, Status='live', Symbol='CG34167', Aliases='Dmel_ Dmel\\CG34167', description='CG34167 gene product from transcript CG34167-RA', other_designations='CG34167-PA|uncharacterized protein', map_location='35E1-35E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16002981, end_position_on_the_genomic_accession=16003641, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39201, CurrentID=0, Status='live', Symbol='CG14151', Aliases='Dmel_ Dmel\\CG14151', description='CG14151 gene product from transcript CG14151-RA', other_designations='CG14151-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10627680, end_position_on_the_genomic_accession=10628484, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39202, CurrentID=0, Status='live', Symbol='hay', Aliases='Dmel_CG8019, CG8019, DhR25, DhXPB, DmXPB, Dmel\\CG8019, ERCC3, Hay, TFIIH, XPB, Xpb, Xpb/hay, hwr, i50, ms(3)nc2, nc2', description='haywire', other_designations='CG8019-PA|CG8019-PB|hay-PA|hay-PB', map_location='3-34.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10631079, end_position_on_the_genomic_accession=10633969, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39203, CurrentID=0, Status='live', Symbol='E(z)', Aliases='Dmel_CG6502, CG6502, Dmel\\CG6502, E(Z)1, EZ, EZH1/2, E[z], Ez, Su(z)301, l(3)1902, l(3)67Fa, l(3)B12, l(3)SG17, l(3)ds12, pco, pco/E(z), E(z)', description='Enhancer of zeste', other_designations='CG6502-PA|CG6502-PB|CG6502-PC|E(z)-PA|E(z)-PB|E(z)-PC|enhancer of zest|enhancer of zeste|enhancer-of-zeste|polycombeotic', map_location='3-34.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10634575, end_position_on_the_genomic_accession=10638130, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39204, CurrentID=0, Status='live', Symbol='CG8009', Aliases='Dmel_ BcDNA:LD14173, Dmel\\CG8009', description='CG8009 gene product from transcript CG8009-RB', other_designations='CG8009-PA|CG8009-PB', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10638146, end_position_on_the_genomic_accession=10639678, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39205, CurrentID=0, Status='live', Symbol='CG18628', Aliases='Dmel_ BcDNA:GH08251, Dmel\\CG18628', description='CG18628 gene product from transcript CG18628-RA', other_designations='CG18628-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10639740, end_position_on_the_genomic_accession=10640136, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39206, CurrentID=0, Status='live', Symbol='CG32066', Aliases='Dmel_ CG6487, CG6491, Dmel\\CG32066', description='CG32066 gene product from transcript CG32066-RB', other_designations='CG32066-PB|CG32066-PC|uncharacterized protein', map_location='67E5-67E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10640286, end_position_on_the_genomic_accession=10661013, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39207, CurrentID=0, Status='live', Symbol='CG14154', Aliases='Dmel_ Dmel\\CG14154', description='CG14154 gene product from transcript CG14154-RA', other_designations='CG14154-PA|CG14154-PB|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10641889, end_position_on_the_genomic_accession=10643009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740840, CurrentID=0, Status='live', Symbol='Shawl', Aliases='Dmel_CG34366, CG13111, CG34366, CG4450, Dmel\\CG34366, Dmel_CG13111, Dmel_CG4450, Kv3, Kv3.2, shawl', description='Shaw-like', other_designations='CG34366-PC|CG34366-PD|Shawl-PC|Shawl-PD', map_location='30B1-30B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9370309, end_position_on_the_genomic_accession=9415156, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  39210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39210, CurrentID=0, Status='live', Symbol='CG8003', Aliases='Dmel_ Dmel\\CG8003', description='CG8003 gene product from transcript CG8003-RB', other_designations='CG8003-PA|CG8003-PB', map_location='67E6-67E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10661475, end_position_on_the_genomic_accession=10663157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39212, CurrentID=0, Status='live', Symbol='simj', Aliases='Dmel_CG32067, CG32067, CG7983, CG8000, Dmel\\CG32067, SIMJ, anon-WO0118547.361, dp66, l(3)01814, l(3)j4A5, l(3)rN672, p66, p66/68-like', description='simjang', other_designations='CG32067-PA|CG32067-PB|CG32067-PC|CG32067-PD|CG32067-PE|simj-PA|simj-PB|simj-PC|simj-PD|simj-PE', map_location='67E6-67E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10664739, end_position_on_the_genomic_accession=10690734, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39213, CurrentID=0, Status='live', Symbol='CG11811', Aliases='Dmel_ Dmel\\CG11811', description='CG11811 gene product from transcript CG11811-RA', other_designations='CG11811-PA|uncharacterized protein', map_location='67E6-67E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10691024, end_position_on_the_genomic_accession=10692598, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39214, CurrentID=0, Status='live', Symbol='ND-13B', Aliases='Dmel_CG6463, BcDNA:RH11203, CG6463, Dmel\\CG6463, NUFM', description='NADH dehydrogenase (ubiquinone) 13 kDa B subunit', other_designations='CG6463-PA|NADH ubiquinone oxidoreductase 13 kDa B|NADH-ubiquinone oxidoreductase 13 kDa-B subunit|NADH:ubiquinone oxidoreductase 13 kDa B subunit|ND-13B-PA', map_location='67E7-67E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10693184, end_position_on_the_genomic_accession=10694108, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39215, CurrentID=0, Status='live', Symbol='NijA', Aliases='Dmel_CG6449, CG6449, Dmel\\CG6449, ninA', description='Ninjurin A', other_designations='CG6449-PA|CG6449-PB|CG6449-PC|CG6449-PD|NijA-PA|NijA-PB|NijA-PC|NijA-PD|ninjurin A', map_location='67E7-67E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10694360, end_position_on_the_genomic_accession=10696877, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740848, CurrentID=0, Status='live', Symbol='CG34229', Aliases='Dmel_ CG13191, Dmel\\CG34229, Sdh6', description='CG34229 gene product from transcript CG34229-RA', other_designations='CG34229-PA|uncharacterized protein', map_location='48C4-48C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11841608, end_position_on_the_genomic_accession=11842164, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39217, CurrentID=0, Status='live', Symbol='tna', Aliases='Dmel_CG7958, 0583/02, CG7958, Dmel\\CG7958, anon-WO0118547.590, l(3)L6731, l(3)S058302, l(3)rI075', description='tonalli', other_designations='CG7958-PA|CG7958-PB|CG7958-PD|tna-PA|tna-PB|tna-PD|tonalli A', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10857889, end_position_on_the_genomic_accession=10876496, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39218, CurrentID=0, Status='live', Symbol='CG6418', Aliases='Dmel_ DmRH27, Dmel\\CG6418, cg6418', description='CG6418 gene product from transcript CG6418-RB', other_designations='CG6418-PB', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10876315, end_position_on_the_genomic_accession=10879037, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39219, CurrentID=0, Status='live', Symbol='Blos4', Aliases='Dmel_CG14149, CG14149, Dm Blos4, Dmel\\CG14149, NP_648414, blos4', description='Biogenesis of lysosome-related organelles complex 1, subunit 4', other_designations='BLOC-1 subunit 4|Blos4-PA|Blos4-PB|CG14149-PA|CG14149-PB|biogenesis of lysosome-related organelles complex 1, subunit 4|cappuccino', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10879085, end_position_on_the_genomic_accession=10880188, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39220, CurrentID=0, Status='live', Symbol='CG6409', Aliases='Dmel_ Dmel\\CG6409', description='CG6409 gene product from transcript CG6409-RB', other_designations='CG6409-PA|CG6409-PB', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10880699, end_position_on_the_genomic_accession=10882386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740853, CurrentID=0, Status='live', Symbol='lmgA', Aliases='Dmel_CG34440, APC11, Apc11, CG18042, CG34440, Dmel\\CG34440, Dmel_CG18042, Dromel_CG18042_FBtr0079695_lmg_uORF, Img, LMG, ORF1, l(2)03424, lmg', description='lemming A', other_designations='CG34440-PA|CG34440-PB|CG34440-PC|CG34440-PD|lemming|lmgA-PA|lmgA-PB|lmgA-PC|lmgA-PD', map_location='29D4-29D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8524999, end_position_on_the_genomic_accession=8528618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39222, CurrentID=0, Status='live', Symbol='CG6404', Aliases='Dmel_ BcDNA:GH02220, Dmel\\CG6404, OXA1', description='CG6404 gene product from transcript CG6404-RA', other_designations='CG6404-PA|CG6404-PB|biogenesis protein OXA1|inner membrane protein OXA1L', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10884261, end_position_on_the_genomic_accession=10886471, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39223, CurrentID=0, Status='live', Symbol='Cpr67Fa1', Aliases='Dmel_CG7941, BcDNA:RE57116, CG7941, DmelCpr67Fa1, Dmel\\CG7941', description='Cuticular protein 67Fa1', other_designations='CG7941-PA|Cpr67Fa1-PA|cuticular protein 67Fa1', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10889990, end_position_on_the_genomic_accession=10890700, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39224, CurrentID=0, Status='live', Symbol='Cpr67Fa2', Aliases='Dmel_CG18349, BcDNA:LP01241, CG18349, DmelCpr67Fa2, Dmel\\CG18349', description='Cuticular protein 67Fa2', other_designations='CG18349-PA|Cpr67Fa2-PA|cuticular protein 67Fa2', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10891714, end_position_on_the_genomic_accession=10892516, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39225, CurrentID=0, Status='live', Symbol='Cpr67Fb', Aliases='Dmel_CG18348, CG18348, DmelCpr67Fb, Dmel\\CG18348', description='Cuticular protein 67Fb', other_designations='CG18348-PA|Cpr67Fb-PA|cuticular protein 67Fb', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10894173, end_position_on_the_genomic_accession=10894693, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39226, CurrentID=0, Status='live', Symbol='Aps', Aliases='Dmel_CG6391, CG6391, Dmel\\CG6391, anon-WO0118547.266', description='CG6391 gene product from transcript CG6391-RC', other_designations='Aps-PA|Aps-PB|Aps-PC|CG6391-PA|CG6391-PB|CG6391-PC', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10894776, end_position_on_the_genomic_accession=10899123, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39227, CurrentID=0, Status='live', Symbol='CG14147', Aliases='Dmel_ BcDNA:RE16005, Dmel\\CG14147', description='CG14147 gene product from transcript CG14147-RA', other_designations='CG14147-PA|uncharacterized protein', map_location='68A1-68A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10972648, end_position_on_the_genomic_accession=10973476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39228, CurrentID=0, Status='live', Symbol='klu', Aliases='Dmel_CG12296, CG12296, Dmel\\CG12296, Klu, P09036, l(3)09036, l(3)10052', description='klumpfuss', other_designations='CG12296-PA|CG12296-PB|CG12296-PC|CG12296-PD|klu-PA|klu-PB|klu-PC|klu-PD|klumphuss', map_location='68A1-68A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10981171, end_position_on_the_genomic_accession=11009188, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740861, CurrentID=0, Status='live', Symbol='CG34234', Aliases='Dmel_ Dmel\\CG34234', description='CG34234 gene product from transcript CG34234-RA', other_designations='CG34234-PA|uncharacterized protein', map_location='49A9-49A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12486719, end_position_on_the_genomic_accession=12487378, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39230, CurrentID=0, Status='live', Symbol='CG32079', Aliases='Dmel_ CG11806, Dmel\\CG32079', description='CG32079 gene product from transcript CG32079-RB', other_designations='CG32079-PB|uncharacterized protein', map_location='68A1-68A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11025561, end_position_on_the_genomic_accession=11027332, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39231, CurrentID=0, Status='live', Symbol='CG43693', Aliases='Dmel_ CG34239, CG6327, Dmel\\CG43693, Dmel_CG34239, Dmel_CG6327', description='CG43693 gene product from transcript CG43693-RD', other_designations='CG43693-PA|CG43693-PB|CG43693-PC|CG43693-PD|CG43693-PE|CG43693-PF|CG43693-PG|uncharacterized protein', map_location='68A2-68A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11036358, end_position_on_the_genomic_accession=11051371, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  39232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39232, CurrentID=0, Status='live', Symbol='CG7888', Aliases='Dmel_ Dmel\\CG7888', description='CG7888 gene product from transcript CG7888-RC', other_designations='CG7888-PA|CG7888-PB|CG7888-PC', map_location='68A3-68A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11054035, end_position_on_the_genomic_accession=11059659, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39233, CurrentID=0, Status='live', Symbol='CG6321', Aliases='Dmel_ Dmel\\CG6321', description='CG6321 gene product from transcript CG6321-RA', other_designations='CG6321-PA', map_location='68A3-68A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11059437, end_position_on_the_genomic_accession=11061224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39234, CurrentID=0, Status='live', Symbol='Blos2', Aliases='Dmel_CG14145, BLOS2, CG14145, DmBlos2, Dmel\\CG14145, NP_648427, blos2', description='Biogenesis of lysosome-related organelles complex 1, subunit 2', other_designations='BLOC-1 subunit 2|Blos2-PA|CG14145-PA|biogenesis of lysosome-related organelles complex 1, subunit 2', map_location='68A3-68A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11061571, end_position_on_the_genomic_accession=11062215, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740867, CurrentID=0, Status='live', Symbol='CG34428', Aliases='Dmel_ CG14116, Dmel\\CG34428, Dmel_CG14116', description='CG34428 gene product from transcript CG34428-RA', other_designations='CG34428-PA|uncharacterized protein', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13061879, end_position_on_the_genomic_accession=13062687, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740868, CurrentID=0, Status='live', Symbol='CG34458', Aliases='Dmel_ CG5197, Dmel\\CG34458, Dmel_CG5197, SP118', description='CG34458 gene product from transcript CG34458-RA', other_designations='CG34458-PA|uncharacterized protein', map_location='53D1-53D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16672033, end_position_on_the_genomic_accession=16672889, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39237, CurrentID=0, Status='live', Symbol='CG6310', Aliases='Dmel_ BcDNA:AT03632, Dmel\\CG6310', description='CG6310 gene product from transcript CG6310-RA', other_designations='CG6310-PA', map_location='68A4-68A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11069043, end_position_on_the_genomic_accession=11069864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39238, CurrentID=0, Status='live', Symbol='Mocs1', Aliases='Dmel_CG33048, CG33048, CG7858, Dmel\\CG33048, MOCS1A, MOCS1B, MOCSIA, Mocs1A-B, Mocs1B, MocsI, Mosc1A+Mosc1B, lxd, mocsI, Mocs1', description='Molybdenum cofactor synthesis 1 ortholog', other_designations='CG33048-PA|CG33048-PC|Mocs1-PA|Mocs1-PC|low xanthine dehydrogenase|molybdenum cofactor synthesis 1 ortholog', map_location='3-34.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11070526, end_position_on_the_genomic_accession=11073348, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39240, CurrentID=0, Status='live', Symbol='CG7839', Aliases='Dmel_ Dmel\\CG7839, anon-EST:fe2H6', description='CG7839 gene product from transcript CG7839-RA', other_designations='CG7839-PA|anon-fast-evolving-2H6', map_location='68A4-68A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11074769, end_position_on_the_genomic_accession=11078736, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39241, CurrentID=0, Status='live', Symbol='JIL-1', Aliases='Dmel_CG6297, 2Ab17, CG6297, Dmel\\CG6297, IL-1, JIL, JIL1, Jil-1, Jil1, MSK, MSK1, Su(var)3-1, Su(var)3-103, Su-var(3)1, Su-var(3)103, anon-WO0118547.379, mAb2A', description='JIL-1 kinase', other_designations='CG6297-PA|CG6297-PB|CG6297-PC|JIL-1 histone H3S10 kinase|JIL-1 histone kinase|JIL-1-PA|JIL-1-PB|JIL-1-PC|suppressor of variegation 3-1', map_location='3-31.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11078752, end_position_on_the_genomic_accession=11093193, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39242, CurrentID=0, Status='live', Symbol='CG6279', Aliases='Dmel_ Dmel\\CG6279', description='CG6279 gene product from transcript CG6279-RA', other_designations='CG6279-PA|CG6279-PB', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11093371, end_position_on_the_genomic_accession=11096245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39243, CurrentID=0, Status='live', Symbol='CG6272', Aliases='Dmel_ BcDNA:GH10915, CT19628, DM10, Dmel\\CG6272', description='CG6272 gene product from transcript CG6272-RB', other_designations='CG6272-PA|CG6272-PB', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11097446, end_position_on_the_genomic_accession=11098186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39244, CurrentID=0, Status='live', Symbol='APP-BP1', Aliases='Dmel_CG7828, APPBP1, CG7828, Dmel\\CG7828, anon-WO0140519.225, dAPP-BP1, dAPPBP1', description='beta-Amyloid precursor protein binding protein 1', other_designations='APP-BP1-PA|APP-BP1-PB|CG7828-PA|CG7828-PB', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11098412, end_position_on_the_genomic_accession=11100578, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39245, CurrentID=0, Status='live', Symbol='Elo68beta', Aliases='Dmel_CG11801, CG11801, Dmel\\CG11801, elo68beta', description='Elongase 68beta', other_designations='CG11801-PB|Elo68beta-PB|elongase 68beta', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11100675, end_position_on_the_genomic_accession=11101698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39246, CurrentID=0, Status='live', Symbol='CG32071', Aliases='Dmel_ CG6261, Dmel\\CG32071', description='CG32071 gene product from transcript CG32071-RA', other_designations='CG32071-PA|uncharacterized protein', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11103421, end_position_on_the_genomic_accession=11103873, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740879, CurrentID=0, Status='live', Symbol='CG34367', Aliases='Dmel_ CG13141, CG5369, Dmel\\CG34367, Dmel_CG13141, Dmel_CG5369', description='CG34367 gene product from transcript CG34367-RC', other_designations='CG34367-PB|CG34367-PC|CG34367-PE|uncharacterized protein', map_location='31D10-31D10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10347851, end_position_on_the_genomic_accession=10353905, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39248, CurrentID=0, Status='live', Symbol='CG12522', Aliases='Dmel_ Dmel\\CG12522', description='CG12522 gene product from transcript CG12522-RA', other_designations='CG12522-PA|uncharacterized protein', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11108235, end_position_on_the_genomic_accession=11108732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39250, CurrentID=0, Status='live', Symbol='CG7638', Aliases='Dmel_ Dmel\\CG7638', description='CG7638 gene product from transcript CG7638-RA', other_designations='CG7638-PA', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11109690, end_position_on_the_genomic_accession=11112117, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39251, CurrentID=0, Status='live', Symbol='Sod', Aliases="Dmel_CG11793, CG11793, Cu, Cu-Zn SOD, Cu/Zn SOD, Cu/Zn sod, Cu/Zn superoxide dismutase, Cu/ZnSOD, CuSOD, CuZn SOD, CuZn-SOD, CuZn-SOD1, CuZnSOD, Cu[2+]/Zn[2+]SOD, Dmel\\CG11793, G, Mn SOD, SOD, SOD-1, SOD1, SOD2, SODC_DROME-1, Sod1, To, To-1, Zn SOD, Zn Sod, Zn-SOD, ZnSod, cSOD, cSod, dSOD1, l(3)108, l(3)68Af', l(3)G, sod, sod1, Sod", description='Superoxide dismutase', other_designations='CG11793-PA|CG11793-PD|Cu, Zn superoxide dismutase|Cu-Zn superoxide dismutase|Cu/Zn superoxide dismutase|Cu/Zn-Superoxide dismutase|CuZn superoxide dismutase|CuZn-superoxide dismutase|CuZn-superoxide dismutase (SOD)1|CuZnSOD|Cu[2+] Zn[2+] superoxide dismutase|Cu[2+]Zn[2+] superoxide dismutase|Mn superoxide dismutase|Sod-PA|Sod-PD|complementation group G|copper and zinc SOD|copper-zinc superoxide|copper-zinc superoxide dismutase|cytoplasmic Cu/ZnSOD|dismutase|heart morphogenesis; GO:0003007 | IMP|paracrine signaling; GO:0038001 | IMP|reactive oxygen species metabolic process; GO:0072593 | IMP|super oxide dismutase|superoxidase dismutase|superoxide dismutase|superoxide dismutase 1|superoxide dismutatase|superoxide-dismutase|superoxido dismutase|tetrazolium oxidase|tetrazolium oxidase-1', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11112281, end_position_on_the_genomic_accession=11113740, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39252, CurrentID=0, Status='live', Symbol='FoxK', Aliases='Dmel_CG11799, CG11799, DmLD16137, DmMNF, Dmel\\CG11799, MNF, Mnf, dFoxK, fd68A, fork head transcription factor long isoform, fork head transcription factor short isoform', description='Forkhead box K', other_designations='CG11799-PG|CG11799-PH|CG11799-PI|CG11799-PJ|CG11799-PK|CG11799-PN|CG11799-PO|CG11799-PP|FoxK-PG|FoxK-PH|FoxK-PI|FoxK-PJ|FoxK-PK|FoxK-PN|FoxK-PO|FoxK-PP|Mnf|fork head box K|forkhead box K|forkhead domain 68A|myocyte nuclear factor', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11114169, end_position_on_the_genomic_accession=11120783, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  39253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39253, CurrentID=0, Status='live', Symbol='mRpL2', Aliases='Dmel_CG7636, CG7636, Dmel\\CG7636, L2', description='mitochondrial ribosomal protein L2', other_designations='CG7636-PA|mRpL2-PA', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11120847, end_position_on_the_genomic_accession=11122041, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39254, CurrentID=0, Status='live', Symbol='Ufd1-like', Aliases='Dmel_CG6233, CG6233, Dmel\\CG6233, UFD1, Udf1, UfD1-like, Ufd-1, Ufd1, Ufd1 like, Ufd1L, anon-WO0118547.405, dUFD1L, dUfd1, ufd1, ufd1-like', description='Ubiquitin fusion-degradation 1-like', other_designations='CG6233-PA|UfD1-like|Ufd1-like-PA|ubiquitin fusion-degradation|ubiquitin fusion-degradation 1-like', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11121966, end_position_on_the_genomic_accession=11123216, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39255, CurrentID=0, Status='live', Symbol='NaPi-III', Aliases='Dmel_CG42575, CG42575, CG7628, CR32078, Dmel\\CG42575, Dmel_CG7628, Dmel_CR32078, Dromel_CG7628_FBtr0076213_mORF, dPit', description='Na[+]-dependent inorganic phosphate cotransporter type III', other_designations='CG42575-PA|CG42575-PB|CR32078|NaPi-III-PA|NaPi-III-PB', map_location='68A7-68A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11123348, end_position_on_the_genomic_accession=11134663, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39256, CurrentID=0, Status='live', Symbol='CG14143', Aliases='Dmel_ Dmel\\CG14143, chr3L_11128040_11128584.0', description='CG14143 gene product from transcript CG14143-RA', other_designations='CG14143-PA|uncharacterized protein', map_location='68A8-68A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11134987, end_position_on_the_genomic_accession=11135480, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39257, CurrentID=0, Status='live', Symbol='CG6216', Aliases='Dmel_ Dmel\\CG6216', description='CG6216 gene product from transcript CG6216-RB', other_designations='CG6216-PB', map_location='68A8-68A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11149578, end_position_on_the_genomic_accession=11150785, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39258, CurrentID=0, Status='live', Symbol='CG32082', Aliases='Dmel_ CG7624, Dmel\\CG32082, anon-WO0118547.199', description='CG32082 gene product from transcript CG32082-RA', other_designations='CG32082-PA|CG32082-PB|CG32082-PD|CG32082-PE|uncharacterized protein', map_location='68A8-68A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11136049, end_position_on_the_genomic_accession=11163984, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  39259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39259, CurrentID=0, Status='live', Symbol='wls', Aliases='Dmel_CG6210, CG6210, Dmel\\CG6210, Evi, Evi/Wls, Srt, Wls, Wls/Evi, evi, srt/evi, wls', description='wntless', other_designations='CG6210-PA|CG6210-PB|evenness interrupted|sprinter|wls-PA|wls-PB', map_location='68A9-68A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11166620, end_position_on_the_genomic_accession=11169625, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39260, CurrentID=0, Status='live', Symbol='CG7616', Aliases='Dmel_ Dmel\\CG7616, FBgn 36142', description='CG7616 gene product from transcript CG7616-RA', other_designations='CG7616-PA|CG7616-PB', map_location='68A9-68A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11169817, end_position_on_the_genomic_accession=11172411, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39261, CurrentID=0, Status='live', Symbol='CG14142', Aliases='Dmel_ Dmel\\CG14142', description='CG14142 gene product from transcript CG14142-RC', other_designations='CG14142-PB|CG14142-PC|CG14142-PD|uncharacterized protein', map_location='68A9-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11173514, end_position_on_the_genomic_accession=11176892, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39262, CurrentID=0, Status='live', Symbol='GlcAT-P', Aliases='Dmel_CG6207, CG6207, D DmGlcAT-BSII, Dmel\\CG6207, dGlcAT-BSII', description='CG6207 gene product from transcript CG6207-RE', other_designations='CG6207-PA|CG6207-PB|CG6207-PC|CG6207-PD|CG6207-PE|GlcAT-P-PA|GlcAT-P-PB|GlcAT-P-PC|GlcAT-P-PD|GlcAT-P-PE', map_location='68B1-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11176954, end_position_on_the_genomic_accession=11197536, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39263, CurrentID=0, Status='live', Symbol='CG7607', Aliases='Dmel_ BcDNA:RH68451, CT23219, Dmel\\CG7607', description='CG7607 gene product from transcript CG7607-RA', other_designations='CG7607-PA', map_location='68B1-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11199620, end_position_on_the_genomic_accession=11200587, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39264, CurrentID=0, Status='live', Symbol='CG14141', Aliases='Dmel_ BcDNA:RH33338, CT33742, Dmel\\CG14141', description='CG14141 gene product from transcript CG14141-RA', other_designations='CG14141-PA|uncharacterized protein', map_location='68B1-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11201351, end_position_on_the_genomic_accession=11202158, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39265, CurrentID=0, Status='live', Symbol='Plod', Aliases='Dmel_CG6199, CG6199, Dmel\\CG6199, LH3, RE-H, anon-EST:ParkEST325, anon-WO0118547.416, dPlod', description='procollagen lysyl hydroxylase', other_designations='CG6199-PA|CG6199-PB|Plod-PA|Plod-PB|drosophila plod|lysyl hydroxylase', map_location='68B1-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11198166, end_position_on_the_genomic_accession=11206016, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  39266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39266, CurrentID=0, Status='live', Symbol='Ube3a', Aliases='Dmel_CG6190, AS, As, CG6190, Dmel\\CG6190, Dube3a, NM_140195.2, UBE3A, dUBE3A, das, dube3A, dube3a, ube3a', description='Ubiquitin protein ligase E3A', other_designations='CG6190-PA|Ube3a-PA|angelman syndrome|drosophila angelman syndrome|ubiquitin protein ligase E3A', map_location='68B1-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11206422, end_position_on_the_genomic_accession=11210583, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39267, CurrentID=0, Status='live', Symbol='CG7600', Aliases='Dmel_ Dmel\\CG7600', description='CG7600 gene product from transcript CG7600-RA', other_designations='CG7600-PA', map_location='68B1-68B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11210786, end_position_on_the_genomic_accession=11214097, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39268, CurrentID=0, Status='live', Symbol='CG42671', Aliases='Dmel_ CG18490, CG34240, Dmel\\CG42671, Dmel_CG18490, Dmel_CG34240', description='CG42671 gene product from transcript CG42671-RG', other_designations='CG42671-PB|CG42671-PC|CG42671-PD|CG42671-PE|CG42671-PF|CG42671-PG|CG42671-PH|CG42671-PI|CG42671-PJ|CG42671-PK|CG42671-PL|uncharacterized protein', map_location='68B3-68B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11214505, end_position_on_the_genomic_accession=11230745, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39269, CurrentID=0, Status='live', Symbol='Ir68a', Aliases='Dmel_CG6185, CG6185, CT19410, DmelIR68a, Dmel\\CG6185, IR68a', description='Ionotropic receptor 68a', other_designations='CG6185-PA|CG6185-PB|Ir68a-PA|Ir68a-PB|ionotropic receptor 68a', map_location='68B3-68B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11232972, end_position_on_the_genomic_accession=11236218, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39270, CurrentID=0, Status='live', Symbol='scyl', Aliases='Dmel_CG7590, 7590, CG7590, Dmel\\CG7590, anon-WO0118547.215, scy', description='scylla', other_designations='CG7590-PA|CG7590-PC|scyl-PA|scyl-PC', map_location='68B4-68C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11252307, end_position_on_the_genomic_accession=11256172, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39271, CurrentID=0, Status='live', Symbol='CG6175', Aliases='Dmel_ Dmel\\CG6175, anon-WO0257455.23', description='CG6175 gene product from transcript CG6175-RC', other_designations='CG6175-PB|CG6175-PC|CG6175-PD', map_location='68C1-68C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11287936, end_position_on_the_genomic_accession=11295919, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39272, CurrentID=0, Status='live', Symbol='CG7573', Aliases='Dmel_ Dmel\\CG7573', description='CG7573 gene product from transcript CG7573-RB', other_designations='CG7573-PB', map_location='68C2-68C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11311747, end_position_on_the_genomic_accession=11313397, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39273, CurrentID=0, Status='live', Symbol='CG6168', Aliases='Dmel_ Dmel\\CG6168', description='CG6168 gene product from transcript CG6168-RC', other_designations='CG6168-PB|CG6168-PC', map_location='68C2-68C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11319642, end_position_on_the_genomic_accession=11321719, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39274, CurrentID=0, Status='live', Symbol='CG6163', Aliases='Dmel_ Dmel\\CG6163', description='CG6163 gene product from transcript CG6163-RA', other_designations='CG6163-PA|CG6163-PB', map_location='68C4-68C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11361986, end_position_on_the_genomic_accession=11364885, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39275, CurrentID=0, Status='live', Symbol='CG11726', Aliases='Dmel_ Dmel\\CG11726, cg11726', description='CG11726 gene product from transcript CG11726-RA', other_designations='CG11726-PA|uncharacterized protein', map_location='68C6-68C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11394690, end_position_on_the_genomic_accession=11395673, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39276, CurrentID=0, Status='live', Symbol='CG7560', Aliases='Dmel_ BEST:GH18008, Dmel\\CG7560, gh18008', description='CG7560 gene product from transcript CG7560-RA', other_designations='CG7560-PA', map_location='68C7-68C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11425158, end_position_on_the_genomic_accession=11426395, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39277, CurrentID=0, Status='live', Symbol='CG6149', Aliases='Dmel_ Dmel\\CG6149', description='CG6149 gene product from transcript CG6149-RB', other_designations='CG6149-PB', map_location='68C8-68C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11434695, end_position_on_the_genomic_accession=11435476, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39278, CurrentID=0, Status='live', Symbol='CG7557', Aliases='Dmel_ Dmel\\CG7557, anon-WO0140519.44', description='CG7557 gene product from transcript CG7557-RA', other_designations='CG7557-PA|CG7557-PB', map_location='68C9-68C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11445161, end_position_on_the_genomic_accession=11446579, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39279, CurrentID=0, Status='live', Symbol='CG12289', Aliases='Dmel_ BcDNA:AT04143, Dmel\\CG12289', description='CG12289 gene product from transcript CG12289-RA', other_designations='CG12289-PA|uncharacterized protein', map_location='68C9-68C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11457076, end_position_on_the_genomic_accession=11458470, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39280, CurrentID=0, Status='live', Symbol='CG7551', Aliases='Dmel_ BcDNA:AT07524, Dmel\\CG7551', description='CG7551 gene product from transcript CG7551-RB', other_designations='CG7551-PB', map_location='68C9-68C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11458490, end_position_on_the_genomic_accession=11459625, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39281, CurrentID=0, Status='live', Symbol='CG6140', Aliases='Dmel_ Dmel\\CG6140', description='CG6140 gene product from transcript CG6140-RB', other_designations='CG6140-PA|CG6140-PB', map_location='68C9-68C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11463570, end_position_on_the_genomic_accession=11465864, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39282, CurrentID=0, Status='live', Symbol='Vha16-2', Aliases='Dmel_CG32089, CG32089, CG7547, Dmel\\CG32089, vha16-2', description='Vacuolar H[+] ATPase 16kD subunit 2', other_designations='CG32089-PA|Vha16-2-PA|vacuolar H[+] ATPase 16kD subunit 2|vacuolar H[+] ATPase subunit 16-2', map_location='68C9-68C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11474958, end_position_on_the_genomic_accession=11475552, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39284, CurrentID=0, Status='live', Symbol='chrb', Aliases='Dmel_CG7533, BcDNA:GH09771, CG7533, Dmel\\CG7533, EP(3)1035, anon-EST:Liang-2.51, char, clone 2.51', description='charybde', other_designations='CG7533-PC|CG7533-PE|charybdis|chrb-PC|chrb-PE', map_location='68C10-68C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11487550, end_position_on_the_genomic_accession=11499055, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39285, CurrentID=0, Status='live', Symbol='Sgs8', Aliases='Dmel_CG6132, 68C GII, CG6132, Dmel\\CG6132, Sgs, Sgs-8, sgs-8, sgs8', description='Salivary gland secretion 8', other_designations='CG6132-PA|Sgs8-PA|gene II|group II|salivary gland secretion 8', map_location='3-35.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11508967, end_position_on_the_genomic_accession=11509394, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39288, CurrentID=0, Status='live', Symbol='Sgs3', Aliases='Dmel_CG11720, 68C GII, CG11720, DMSGS378, Dmel\\CG11720, Glue, Muc68Cb, SGS-3, SGS378, Sgs, Sgs-3, sgs-3, sgs3', description='Salivary gland secretion 3', other_designations='CG11720-PA|Sgs3-PA|gene IV|group IV|mucin 68Cb|salivary gland glue protein 3|salivary gland secretion 3', map_location='3-35.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11512221, end_position_on_the_genomic_accession=11513402, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39289, CurrentID=0, Status='live', Symbol='CG7512', Aliases='Dmel_ Dmel\\CG7512', description='CG7512 gene product from transcript CG7512-RB', other_designations='CG7512-PB', map_location='68C11-68C12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11513705, end_position_on_the_genomic_accession=11518904, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39293, CurrentID=0, Status='live', Symbol='Mob2', Aliases='Dmel_CG11711, BEST:GH06341, BcDNA:GH13236, CG11711, CG11712, D Dmel\\CG11711, Dmob1, Dmob2, Mob1, MobI, dmob2, gh06341, trezor', description='CG11711 gene product from transcript CG11711-RA', other_designations='CG11711-PA|CG11711-PB|CG11711-PC|CG11711-PD|CG11711-PE|Mob2-PA|Mob2-PB|Mob2-PC|Mob2-PD|Mob2-PE', map_location='68C12-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11522757, end_position_on_the_genomic_accession=11562233, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39294, CurrentID=0, Status='live', Symbol='CG7394', Aliases='Dmel_ BcDNA:RE52392, D.M.UKP, Dmel\\CG7394, TIM14', description='CG7394 gene product from transcript CG7394-RD', other_designations='CG7394-PA|CG7394-PB|CG7394-PC|CG7394-PD', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11561996, end_position_on_the_genomic_accession=11564312, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39297, CurrentID=0, Status='live', Symbol='rt', Aliases='Dmel_CG6097, CG6097, DPOMT1, Dmel\\CG6097, POMT1, Rot_abd, Rt, dPOMT1, dPomt1, fs(3)neo1, pomt, roa', description='rotated abdomen', other_designations='CG6097-PA|drosophila protein O-mannosyltransferase 1|rt-PA', map_location='3-37 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11584122, end_position_on_the_genomic_accession=11587770, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39300, CurrentID=0, Status='live', Symbol='CG14137', Aliases='Dmel_ Dmel\\CG14137', description='CG14137 gene product from transcript CG14137-RA', other_designations='CG14137-PA|uncharacterized protein', map_location='68C14-68C14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11599636, end_position_on_the_genomic_accession=11599947, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39301, CurrentID=0, Status='live', Symbol='CG7368', Aliases='Dmel_ Dmel\\CG7368', description='CG7368 gene product from transcript CG7368-RB', other_designations='CG7368-PB|CG7368-PC', map_location='68C14-68C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11597202, end_position_on_the_genomic_accession=11605374, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39302, CurrentID=0, Status='live', Symbol='CG6091', Aliases='Dmel_ Dmel\\CG6091, Q9VTK7, ped', description='CG6091 gene product from transcript CG6091-RF', other_designations='CG6091-PD|CG6091-PE|CG6091-PF|CG6091-PG', map_location='68C15-68C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11605436, end_position_on_the_genomic_accession=11609465, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39304, CurrentID=0, Status='live', Symbol='CG6084', Aliases='Dmel_ 34F4T, Dmel\\CG6084, ESTS:34F4T', description='CG6084 gene product from transcript CG6084-RC', other_designations='Aldo/keto reductase|CG6084-PA|CG6084-PB|CG6084-PC|CG6084-PD|aldehyde reductase', map_location='68C15-68D', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11620535, end_position_on_the_genomic_accession=11624381, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39305, CurrentID=0, Status='live', Symbol='CG6083', Aliases='Dmel_ Dmel\\CG6083', description='CG6083 gene product from transcript CG6083-RA', other_designations='CG6083-PA', map_location='68C15-68C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11624528, end_position_on_the_genomic_accession=11626125, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39306, CurrentID=0, Status='live', Symbol='PCID2', Aliases='Dmel_CG7351, CG7351, Dmel\\CG7351, dmPCID2', description='PCI domain-containing protein 2', other_designations='CG7351-PA|PCID2-PA|drosophila melanogaster PCID2', map_location='68C15-68C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11626441, end_position_on_the_genomic_accession=11627715, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39307, CurrentID=0, Status='live', Symbol='CG32091', Aliases='Dmel_ CG14136, CG7346, Dmel\\CG32091', description='CG32091 gene product from transcript CG32091-RB', other_designations='CG32091-PB|uncharacterized protein', map_location='68D1-68D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11631392, end_position_on_the_genomic_accession=11642902, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39308, CurrentID=0, Status='live', Symbol='CG6071', Aliases='Dmel_ Dmel\\CG6071', description='CG6071 gene product from transcript CG6071-RA', other_designations='CG6071-PA', map_location='68D1-68D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11642906, end_position_on_the_genomic_accession=11646033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39309, CurrentID=0, Status='live', Symbol='RIOK1', Aliases='Dmel_CG11660, CG11660, Dmel\\CG11660, Rio1, dRIOK1', description='RIO kinase 1', other_designations='CG11660-PA|CG11660-PB|RIOK1-PA|RIOK1-PB', map_location='68D1-68D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11646234, end_position_on_the_genomic_accession=11648474, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39310, CurrentID=0, Status='live', Symbol='CG7339', Aliases='Dmel_ Dmel\\CG7339, lincRNA.S4825', description='CG7339 gene product from transcript CG7339-RB', other_designations='CG7339-PA|CG7339-PB', map_location='68D1-68D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11648756, end_position_on_the_genomic_accession=11650499, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39311, CurrentID=0, Status='live', Symbol='CG32085', Aliases='Dmel_ BcDNA:RH06780, CG14134, CG6060, CT33735, Dmel\\CG32085, FBXL16', description='CG32085 gene product from transcript CG32085-RA', other_designations='CG32085-PA|uncharacterized protein', map_location='68D1-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11650517, end_position_on_the_genomic_accession=11680344, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39313, CurrentID=0, Status='live', Symbol='TfIIEalpha', Aliases='Dmel_CG10415, CG10415, Dmel\\CG10415, TFIIE, TFIIE-L, TFIIEalpha, TfIIE, TfIIEa, dTFIIE-L, dTFIIEL', description='Transcription factor IIEalpha', other_designations='CG10415-PA|TfIIEalpha-PA|transcription factor IIEalpha', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11682038, end_position_on_the_genomic_accession=11683805, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39314, CurrentID=0, Status='live', Symbol='Sugb', Aliases='Dmel_CG7334, CG7334, Dmel\\CG7334, Sug, sug', description='Sugar baby', other_designations='CG7334-PC|CG7334-PE|CG7334-PF|Sug|Sugar-baby|Sugb-PC|Sugb-PE|Sugb-PF|sugar baby', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11683980, end_position_on_the_genomic_accession=11690155, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39315, CurrentID=0, Status='live', Symbol='Pallidin', Aliases='Dmel_CG14133, CG14133, Dmel\\CG14133, NP_648494, pallidin', description='Pallidin ortholog (H. sapiens)', other_designations='CG14133-PB|Pallidin-PB|pallidin|pallidin ortholog', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11689631, end_position_on_the_genomic_accession=11691399, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39316, CurrentID=0, Status='live', Symbol='CG14135', Aliases='Dmel_ Dmel\\CG14135', description='CG14135 gene product from transcript CG14135-RC', other_designations='CG14135-PB|CG14135-PC|uncharacterized protein', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11691571, end_position_on_the_genomic_accession=11692757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39317, CurrentID=0, Status='live', Symbol='CG11652', Aliases='Dmel_ Dmel\\CG11652, Dph1', description='CG11652 gene product from transcript CG11652-RA', other_designations='CG11652-PA|CG11652-PB|CG11652-PC|diphthamide biosynthesis-like 1|uncharacterized protein', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11693879, end_position_on_the_genomic_accession=11696963, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39318, CurrentID=0, Status='live', Symbol='CG6053', Aliases='Dmel_ Dmel\\CG6053', description='CG6053 gene product from transcript CG6053-RC', other_designations='CG6053-PB|CG6053-PC', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11697126, end_position_on_the_genomic_accession=11699263, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39319, CurrentID=0, Status='live', Symbol='CG11658', Aliases='Dmel_ Dmel\\CG11658, FBXO12, anon-WO0118547.413', description='CG11658 gene product from transcript CG11658-RC', other_designations='CG11658-PA|CG11658-PB|CG11658-PC|CG11658-PD|CG11658-PE|uncharacterized protein', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11699496, end_position_on_the_genomic_accession=11708553, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39320, CurrentID=0, Status='live', Symbol='Ccdc56', Aliases='Dmel_CG42630, CCDC56, CG42630, CG7319, Dmel\\CG42630, Dmel_CG7319, ccdc56', description='Coiled-coil domain containing 56', other_designations='CG42630-PA|Ccdc56-PA|mitochondrial transcription factor B1', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11708979, end_position_on_the_genomic_accession=11710570, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39321, CurrentID=0, Status='live', Symbol='crim', Aliases='Dmel_CG6038, CG6038, Crim, Dmel\\CG6038', description='crimpled', other_designations='CG6038-PA|crim-PA', map_location='68D2-68D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11710507, end_position_on_the_genomic_accession=11711386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39322, CurrentID=0, Status='live', Symbol='Bmcp', Aliases='Dmel_CG7314, CG7314, DmBMCP, DmUCP5, Dmel\\CG7314, UCP5, Ucp5, bmcp, dmUCP5, ucp5', description='CG7314 gene product from transcript CG7314-RB', other_designations='Bmcp-PA|Bmcp-PB|CG7314-PA|CG7314-PB|bmcp', map_location='68D3-68D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11720934, end_position_on_the_genomic_accession=11728118, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39323, CurrentID=0, Status='live', Symbol='CG32087', Aliases='Dmel_ CG7308, Dmel\\CG32087', description='CG32087 gene product from transcript CG32087-RA', other_designations='CG32087-PA|uncharacterized protein', map_location='68D3-68D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11732141, end_position_on_the_genomic_accession=11733610, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39324, CurrentID=0, Status='live', Symbol='Gr68a', Aliases='Dmel_CG7303, 68D.1, CG7303, Dmel\\CG7303, GR68D.1, GR68a, Gr68D1', description='Gustatory receptor 68a', other_designations='CG7303-PA|Gr68a-PA|gustatory receptor 68D.1|gustatory receptor 68a', map_location='68D4-68D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11751305, end_position_on_the_genomic_accession=11752474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39325, CurrentID=0, Status='live', Symbol='CG6024', Aliases='Dmel_ CT18912, Dmel\\CG6024', description='CG6024 gene product from transcript CG6024-RF', other_designations='CG6024-PC|CG6024-PF', map_location='68D3-68D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11733589, end_position_on_the_genomic_accession=11766182, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39326, CurrentID=0, Status='live', Symbol='Muc68D', Aliases='Dmel_CG6004, CG6004, Dmel\\CG6004, anon-WO0153538.47, anon-WO0153538.48', description='Mucin 68D', other_designations='CG6004-PB|Muc68D-PB|mucin 68D', map_location='68D4-68D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11767299, end_position_on_the_genomic_accession=11772157, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39327, CurrentID=0, Status='live', Symbol='Tim13', Aliases='Dmel_CG11611, CG11611, Dmel\\CG11611', description='CG11611 gene product from transcript CG11611-RA', other_designations='CG11611-PA|Tim13-PA', map_location='68D4-68D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11774060, end_position_on_the_genomic_accession=11774501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39328, CurrentID=0, Status='live', Symbol='CG14131', Aliases='Dmel_ Dmel\\CG14131, Pi3K68D', description='CG14131 gene product from transcript CG14131-RA', other_designations='CG14131-PA|CG14131-PB|uncharacterized protein', map_location='68D4-68D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11777073, end_position_on_the_genomic_accession=11780485, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39329, CurrentID=0, Status='live', Symbol='Pi3K68D', Aliases='Dmel_CG11621, 11621, BcDNA:LD15217, CG11621, Cpk, Dmel\\CG11621, PI(3)K, PI3K, PI3K 68D, PI3K 68_D, PI3K-68D, PI3K-68D/E, PI3K68D, PI3K_68D, PI[[3]]K, Pi3K, cpk, dPI3K, dPIK, pi3k68d', description='Phosphotidylinositol 3 kinase 68D', other_designations='CG11621-PA|CG11621-PB|CG11621-PC|CG11621-PD|CG11621-PE|PI 3-kinase|PI3-kinase|Pi3K68D-PA|Pi3K68D-PB|Pi3K68D-PC|Pi3K68D-PD|Pi3K68D-PE|dPI 3-kinase|phosphotidylinositol 3 kinase 68D', map_location='68D4-68D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11774650, end_position_on_the_genomic_accession=11790013, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  39330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39330, CurrentID=0, Status='live', Symbol='CG5964', Aliases='Dmel_ Dmel\\CG5964', description='CG5964 gene product from transcript CG5964-RA', other_designations='CG5964-PA', map_location='68D6-68D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11790212, end_position_on_the_genomic_accession=11793283, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39331, CurrentID=0, Status='live', Symbol='CG10907', Aliases='Dmel_ Dmel\\CG10907', description='CG10907 gene product from transcript CG10907-RA', other_designations='CG10907-PA|uncharacterized protein', map_location='68D6-68D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11793514, end_position_on_the_genomic_accession=11795418, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39332, CurrentID=0, Status='live', Symbol='Klp68D', Aliases='Dmel_CG7293, CG7293, DmKlp68D, Dmel\\CG7293, KIF 3B, KIF3B, KLP-5, KLP5, KLP68D, KLP68Ddm, KLP[[64D/68D]], Klp5, klp68d', description='Kinesin-like protein at 68D', other_designations='CG7293-PA|CG7293-PB|Kinesin-2|Klp68D-PA|Klp68D-PB|kinesin II', map_location='68D6-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11795614, end_position_on_the_genomic_accession=11798875, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39333, CurrentID=0, Status='live', Symbol='CG10361', Aliases='Dmel_ Dmel\\CG10361', description='CG10361 gene product from transcript CG10361-RA', other_designations='CG10361-PA|uncharacterized protein', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11798948, end_position_on_the_genomic_accession=11800861, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39334, CurrentID=0, Status='live', Symbol='CG42255', Aliases='Dmel_ CG32092, CG32094, CG5956, CT18701, Dmel\\CG42255, Dmel_CG32092, Dmel_CG32094', description='CG42255 gene product from transcript CG42255-RB', other_designations='CG42255-PB|uncharacterized protein', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11801951, end_position_on_the_genomic_accession=11814755, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  39335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39335, CurrentID=0, Status='live', Symbol='CG14130', Aliases='Dmel_ Dmel\\CG14130', description='CG14130 gene product from transcript CG14130-RA', other_designations='CG14130-PA|uncharacterized protein', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11815444, end_position_on_the_genomic_accession=11816710, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39336, CurrentID=0, Status='live', Symbol='CG5946', Aliases='Dmel_ Dmel\\CG5946', description='CG5946 gene product from transcript CG5946-RB', other_designations='CG5946-PA|CG5946-PB|CG5946-PC|CG5946-PD|CG5946-PE|CG5946-PF|CG5946-PG', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11816695, end_position_on_the_genomic_accession=11822275, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39337, CurrentID=0, Status='live', Symbol='CG11597', Aliases='Dmel_ Dmel\\CG11597, PP4', description='CG11597 gene product from transcript CG11597-RC', other_designations='CG11597-PA|CG11597-PB|CG11597-PC|PP4-like|uncharacterized protein', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11820338, end_position_on_the_genomic_accession=11822275, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39338, CurrentID=0, Status='live', Symbol='RpL10Ab', Aliases='Dmel_CG7283, BcDNA:RH06366, CG7283, Dmel\\CG7283, L10Ab, RpL10A, RpL10Ae, anon-EST:Posey188', description='Ribosomal protein L10Ab', other_designations='CG7283-PA|CG7283-PF|RpL10Ab-PA|RpL10Ab-PF|ribosomal protein L10Ab', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11822851, end_position_on_the_genomic_accession=11824211, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39339, CurrentID=0, Status='live', Symbol='CG7264', Aliases='Dmel_ Dmel\\CG7264', description='CG7264 gene product from transcript CG7264-RA', other_designations='CG7264-PA|CG7264-PB', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11829030, end_position_on_the_genomic_accession=11830844, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39340, CurrentID=0, Status='live', Symbol='CycA', Aliases='Dmel_CG5940, 0033/02, 0040/24, 0046/23, 0104/07, 0245/34, 0521/06, 0545/13, 1329/16, 1449/05, CG5940, CLD3, CYCA, Cyc A, DmcycA, Dmel\\CG5940, S(rux)3B, anon-WO0140519.128, cycA, dCycA, hari, l(3)03946, l(3)183, l(3)68Ea, l(3)j3C8, l(3)neo114, l(3)rsg11, rsg11', description='Cyclin A', other_designations='CG5940-PA|CG5940-PC|CycA-PA|CycA-PC|cyclin A|cyclinA|rose-gespleten region interval 11|suppressor of rux 3B', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11826719, end_position_on_the_genomic_accession=11833578, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39343, CurrentID=0, Status='live', Symbol='CG5906', Aliases='Dmel_ BcDNA:AT16953, Dmel\\CG5906', description='CG5906 gene product from transcript CG5906-RA', other_designations='CG5906-PA', map_location='68E2-68E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11870457, end_position_on_the_genomic_accession=11871305, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39344, CurrentID=0, Status='live', Symbol='Sprn', Aliases='Dmel_CG14128, CG14128, Dmel\\CG14128, sprn', description='Spermitin', other_designations='CG14128-PB|CG14128-PC|CG14128-PE|CG14128-PF|Sprn-PB|Sprn-PC|Sprn-PE|Sprn-PF|spermitin', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11879115, end_position_on_the_genomic_accession=11881182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39345, CurrentID=0, Status='live', Symbol='CCDC151', Aliases='Dmel_CG14127, CG14127, Dmel\\CG14127', description='Coiled-coil domain containing protein 151', other_designations='CCDC151-PA|CG14127-PA|Ccdc151', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11881384, end_position_on_the_genomic_accession=11883546, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39346, CurrentID=0, Status='live', Symbol='CG5897', Aliases='Dmel_ Dmel\\CG5897', description='CG5897 gene product from transcript CG5897-RB', other_designations='CG5897-PB', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11887259, end_position_on_the_genomic_accession=11888981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39347, CurrentID=0, Status='live', Symbol='CG11588', Aliases='Dmel_ Dmel\\CG11588', description='CG11588 gene product from transcript CG11588-RB', other_designations='CG11588-PA|CG11588-PB|uncharacterized protein', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11921513, end_position_on_the_genomic_accession=11922599, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39348, CurrentID=0, Status='live', Symbol='SdhAL', Aliases='Dmel_CG5718, CG5718, DHSA, Dmel\\CG5718', description='Succinate dehydrogenase, subunit A (flavoprotein)-like', other_designations='CG5718-PA|SdhAL-PA|succinate dehydrogenase [ubiquinone] flavoprotein subunit|succinate dehydrogenase flavoprotein subunit|succinate dehydrogenase, subunit A (flavoprotein)-like', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11922610, end_position_on_the_genomic_accession=11924757, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39349, CurrentID=0, Status='live', Symbol='byn', Aliases='Dmel_CG7260, Byn, CG7260, D-TRA, D-Trg, DTrg, Dm-BYN, Dmel\\CG7260, Tra, Trg, apro/apro, dm-Trg, trg, byn', description='brachyenteron', other_designations='CG7260-PB|CG7260-PC|T related antigen|T-related gene|aproctous|brachyury|byn-PB|byn-PC', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11927693, end_position_on_the_genomic_accession=11935919, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39351, CurrentID=0, Status='live', Symbol='Rpt4R', Aliases='Dmel_CG7257, 7257, CG7257, DmCG7257, Dm_Rpt4c, Dmel\\CG7257', description='Regulatory particle triple-A ATPase 4-related', other_designations='CG7257-PA|CG7257-PB|Rpt4R-PA|Rpt4R-PB|regulatory particle triple-A ATPase 4-related', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11941912, end_position_on_the_genomic_accession=11943485, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39352, CurrentID=0, Status='live', Symbol='CG5883', Aliases='Dmel_ Dmel\\CG5883', description='CG5883 gene product from transcript CG5883-RA', other_designations='CG5883-PA', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11943640, end_position_on_the_genomic_accession=11944852, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39353, CurrentID=0, Status='live', Symbol='CG7252', Aliases='Dmel_ BcDNA:LP01340, Dmel\\CG7252', description='CG7252 gene product from transcript CG7252-RA', other_designations='CG7252-PA', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11947675, end_position_on_the_genomic_accession=11949248, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39354, CurrentID=0, Status='live', Symbol='CG17826', Aliases='Dmel_ Dmel\\CG17826', description='CG17826 gene product from transcript CG17826-RA', other_designations='CG17826-PA|uncharacterized protein', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11949839, end_position_on_the_genomic_accession=11952225, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39355, CurrentID=0, Status='live', Symbol='obst-G', Aliases='Dmel_CG9781, CG9781, Dmel\\CG9781', description='obstructor-G', other_designations='CG9781-PA|obst-G-PA', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11952301, end_position_on_the_genomic_accession=11953317, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39356, CurrentID=0, Status='live', Symbol='CG7248', Aliases='Dmel_ Dmel\\CG7248', description='CG7248 gene product from transcript CG7248-RA', other_designations='CG7248-PA', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11954083, end_position_on_the_genomic_accession=11956780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39357, CurrentID=0, Status='live', Symbol='CG11570', Aliases='Dmel_ Dmel\\CG11570, mlncRNA68E3', description='CG11570 gene product from transcript CG11570-RB', other_designations='CG11570-PA|CG11570-PB|uncharacterized protein', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11956708, end_position_on_the_genomic_accession=11957765, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39359, CurrentID=0, Status='live', Symbol='CG14125', Aliases='Dmel_ Dmel\\CG14125', description='CG14125 gene product from transcript CG14125-RB', other_designations='CG14125-PA|CG14125-PB|uncharacterized protein', map_location='68E4-68E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11974905, end_position_on_the_genomic_accession=11975810, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39360, CurrentID=0, Status='live', Symbol='CG43896', Aliases='Dmel_ CG17824, CG6947, Dmel\\CG43896, Dmel_CG17824, Dmel_CG6947', description='CG43896 gene product from transcript CG43896-RD', other_designations='CG43896-PB|CG43896-PC|CG43896-PD|uncharacterized protein', map_location='68E4-68E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11976111, end_position_on_the_genomic_accession=11983764, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39362, CurrentID=0, Status='live', Symbol='CG6938', Aliases='Dmel_ Dmel\\CG6938', description='CG6938 gene product from transcript CG6938-RB', other_designations='CG6938-PB|CG6938-PC', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11984906, end_position_on_the_genomic_accession=11990505, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  39363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39363, CurrentID=0, Status='live', Symbol='CG6931', Aliases='Dmel_ Dmel\\CG6931', description='CG6931 gene product from transcript CG6931-RA', other_designations='CG6931-PA', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11990736, end_position_on_the_genomic_accession=11992000, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39364, CurrentID=0, Status='live', Symbol='viaf', Aliases='Dmel_CG18593, CG18593, Dmel\\CG18593, VIAF, anon-WO0134798.17, anon-WO0134798.181, viaf', description='viral IAP-associated factor', other_designations='CG18593-PA|viaf-PA', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11991853, end_position_on_the_genomic_accession=11993179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39365, CurrentID=0, Status='live', Symbol='Grip163', Aliases='Dmel_CG5688, CG5688, D Dgrip163, Dgrips163, Dmel\\CG5688, GCP6, grip163', description='CG5688 gene product from transcript CG5688-RA', other_designations='CG5688-PA|DGrip163|Grip163-PA', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11993274, end_position_on_the_genomic_accession=11998138, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39366, CurrentID=0, Status='live', Symbol='Pop2', Aliases='Dmel_CG5684, CAF1, CG5684, Caf1, Dmel\\CG5684, POP2, pop2', description='CG5684 gene product from transcript CG5684-RA', other_designations='CG5684-PA|CG5684-PB|CG5684-PC|CG5684-PD|CG5684-PE|Pop2-PA|Pop2-PB|Pop2-PC|Pop2-PD|Pop2-PE', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11998398, end_position_on_the_genomic_accession=12002915, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39367, CurrentID=0, Status='live', Symbol='CG6928', Aliases='Dmel_ Dmel\\CG6928', description='CG6928 gene product from transcript CG6928-RD', other_designations='CG6928-PA|CG6928-PB|CG6928-PC|CG6928-PD', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12003155, end_position_on_the_genomic_accession=12008051, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39368, CurrentID=0, Status='live', Symbol='rols', Aliases='Dmel_CG32096, ANTS, Ants, CG12277, CG17155, CG32096, CG5679, Dmel\\CG32096, Rol/Ants, Rols, Rols6, Rols7, anon-WO0257455.19, ants, ants/rols, l(3)08232, l(3)68Fd, l(3)j2A6\\ants, rols', description='rolling pebbles', other_designations='CG32096-PA|CG32096-PB|CG32096-PC|CG32096-PD|CG32096-PE|CG32096-PF|CG32096-PG|CG32096-PH|CG32096-PI|antisocial|rolling pebbles 7|rols-PA|rols-PB|rols-PC|rols-PD|rols-PE|rols-PF|rols-PG|rols-PH|rols-PI', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12008225, end_position_on_the_genomic_accession=12064859, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39369, CurrentID=0, Status='live', Symbol='CG6793', Aliases='Dmel_ Dmel\\CG6793', description='CG6793 gene product from transcript CG6793-RA', other_designations='CG6793-PA', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12042387, end_position_on_the_genomic_accession=12044217, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740655, CurrentID=0, Status='live', Symbol='CG34431', Aliases='Dmel_ CG14766, Dmel\\CG34431, Dmel_CG14766', description='CG34431 gene product from transcript CG34431-RA', other_designations='CG34431-PA|uncharacterized protein', map_location='43F5-43F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7944526, end_position_on_the_genomic_accession=7947076, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36227, CurrentID=0, Status='live', Symbol='CG13204', Aliases='Dmel_ 13204, Dmel\\CG13204', description='CG13204 gene product from transcript CG13204-RA', other_designations='CG13204-PA|CG13204-PB|uncharacterized protein', map_location='47F6-47F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11366710, end_position_on_the_genomic_accession=11370287, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39373, CurrentID=0, Status='live', Symbol='CG17154', Aliases='Dmel_ Dmel\\CG17154', description='CG17154 gene product from transcript CG17154-RA', other_designations='CG17154-PA|uncharacterized protein', map_location='68F2-68F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12088626, end_position_on_the_genomic_accession=12089225, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39374, CurrentID=0, Status='live', Symbol='l(3)j2D3', Aliases='Dmel_CG6801, CG6801, Dmel\\CG6801, L(3)JD3, l(3)68Fb, l(3)jD3', description='lethal (3) j2D3', other_designations='CG6801-PA|l(3)j2D3-PA', map_location='68F3-68F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12116101, end_position_on_the_genomic_accession=12118055, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39375, CurrentID=0, Status='live', Symbol='ssp', Aliases='Dmel_CG17153, CG17153, Dmel\\CG17153, Ssp', description='sunspot', other_designations='CG17153-PA|ssp-PA', map_location='68F4-68F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12118257, end_position_on_the_genomic_accession=12119789, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39376, CurrentID=0, Status='live', Symbol='CG11560', Aliases='Dmel_ Dmel\\CG11560', description='CG11560 gene product from transcript CG11560-RB', other_designations='CG11560-PA|CG11560-PB|uncharacterized protein', map_location='68F4-68F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12120019, end_position_on_the_genomic_accession=12121185, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39377, CurrentID=0, Status='live', Symbol='yps', Aliases='Dmel_CG5654, CG5654, Dmel\\CG5654, YPS, Yps, p45, p57', description='ypsilon schachtel', other_designations='CG5654-PA|CG5654-PB|yps-PA|yps-PB|ypsilon schactel', map_location='68F4-68F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12121274, end_position_on_the_genomic_accession=12125150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39378, CurrentID=0, Status='live', Symbol='Ir68b', Aliases='Dmel_CG17152, CG17152, CT38092, DmelIR68b, Dmel\\CG17152, IR68b', description='Ionotropic receptor 68b', other_designations='CG17152-PA|Ir68b-PA|ionotropic receptor 68b', map_location='68F4-68F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12125487, end_position_on_the_genomic_accession=12127394, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39381, CurrentID=0, Status='live', Symbol='CG11534', Aliases='Dmel_ BEST:GH12489, Dmel\\CG11534', description='CG11534 gene product from transcript CG11534-RA', other_designations='CG11534-PA|uncharacterized protein', map_location='68F5-68F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12132146, end_position_on_the_genomic_accession=12133834, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39382, CurrentID=0, Status='live', Symbol='CG5645', Aliases='Dmel_ Dmel\\CG5645', description='CG5645 gene product from transcript CG5645-RA', other_designations='CG5645-PA', map_location='68F5-68F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12133771, end_position_on_the_genomic_accession=12136706, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39383, CurrentID=0, Status='live', Symbol='Atg12', Aliases='Dmel_CG10861, ATG12, CG10861, Dmel\\CG10861, atg12', description='Autophagy-related 12', other_designations='Atg12-PB|Autophagy-specific gene 12|CG10861-PB|autophagy related protein 12', map_location='68F5-68F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12136967, end_position_on_the_genomic_accession=12137559, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39385, CurrentID=0, Status='live', Symbol='RhoGAP68F', Aliases='Dmel_CG6811, CG 6811, CG6811, Dmel\\CG6811, RhoGAP', description='Rho GTPase activating protein at 68F', other_designations='CG6811-PA|CG6811-PC|CG6811-PD|CG6811-PE|RhoGAP68F-PA|RhoGAP68F-PC|RhoGAP68F-PD|RhoGAP68F-PE', map_location='68F5-68F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12138868, end_position_on_the_genomic_accession=12143380, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39386, CurrentID=0, Status='live', Symbol='CG5642', Aliases='Dmel_ Dmel\\CG5642, eIF3l', description='CG5642 gene product from transcript CG5642-RA', other_designations='CG5642-PA', map_location='68F6-68F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12143446, end_position_on_the_genomic_accession=12146097, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39387, CurrentID=0, Status='live', Symbol='Nrx-IV', Aliases='Dmel_CG6827, CG682, CG6827, CT21123, Dmel\\CG6827, Dnrx, NRX, NRX IV, Nrx, Nrx IV, NrxIV, Nx IV, l(3)68Ff, l(3)rsg23, nrx, nrx IV, nrx-IV, nrxIV, rsg23', description='Neurexin IV', other_designations='CG6827-PA|CG6827-PB|Neurexin-4|Neurexin-IV|Nrx-IV-PA|Nrx-IV-PB|neurexin|neurexin 4|neurexin IV|neurexinIV|rose-gespleten region interval 23', map_location='68F6-68F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12146519, end_position_on_the_genomic_accession=12156990, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  39388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39388, CurrentID=0, Status='live', Symbol='CG9760', Aliases='Dmel_ Dmel\\CG9760', description='CG9760 gene product from transcript CG9760-RC', other_designations='CG9760-PC|CG9760-PD', map_location='68F7-68F8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12157052, end_position_on_the_genomic_accession=12163735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39389, CurrentID=0, Status='live', Symbol='Rh7', Aliases='Dmel_CG5638, CG5638, DMELRH7, Dmel\\CG5638, Rh, anon-WO0170980.34, anon-WO0170980.35, rh7', description='Rhodopsin 7', other_designations='CG5638-PA|Rh7-PA|rhodopsin|rhodopsin 7', map_location='68F8-69A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12164205, end_position_on_the_genomic_accession=12175482, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39390, CurrentID=0, Status='live', Symbol='CAH2', Aliases='Dmel_CG6906, CG6906, CT21171, Dmel\\CG6906', description='Carbonic anhydrase 2', other_designations='CAH2-PA|CAH2-PB|CG6906-PA|CG6906-PB|carbonic anhydrase 2', map_location='69A1-69A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12177601, end_position_on_the_genomic_accession=12183347, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39391, CurrentID=0, Status='live', Symbol='CG6910', Aliases='Dmel_ Dmel\\CG6910', description='CG6910 gene product from transcript CG6910-RC', other_designations='CG6910-PB|CG6910-PC', map_location='69A1-69A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12184984, end_position_on_the_genomic_accession=12187860, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39392, CurrentID=0, Status='live', Symbol='Est-6', Aliases='Dmel_CG6917, CG6917, Dmel\\CG6917, EST-5, EST-6, EST6, Est, Est 6, Est-5, Est-5B, Est-D, Est6, est-6, est6', description='Esterase 6', other_designations='CG6917-PA|CG6917-PB|Est-6-PA|Est-6-PB|Esterase-6|Esterase6|Stress-inducible humoral factor turandot A|carboxyl ester hydrolase|esterase 6', map_location='3-35.9 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12188818, end_position_on_the_genomic_accession=12190705, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39393, CurrentID=0, Status='live', Symbol='Est-P', Aliases='Dmel_CG17148, CG17148, Dmel\\CG17148, EST-5, EST7, ESTP, Est-5, Est-7, Est7, EstP, PsiEst-6, est-P, psiEst-6, yEst-6, yEst6', description='Esterase P', other_designations='CG17148-PA|Est-P-PA|esterase P', map_location='3-35.9 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12190733, end_position_on_the_genomic_accession=12192528, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39394, CurrentID=0, Status='live', Symbol='thoc6', Aliases='Dmel_CG5632, CG5632, Dmel\\CG5632, THOC6, Thoc6', description='CG5632 gene product from transcript CG5632-RC', other_designations='CG5632-PA|CG5632-PB|CG5632-PC|thoc6-PA|thoc6-PB|thoc6-PC', map_location='69A1-69A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12190696, end_position_on_the_genomic_accession=12193688, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39395, CurrentID=0, Status='live', Symbol='CG11529', Aliases='Dmel_ 148640_at, Dmel\\CG11529, SP84', description='CG11529 gene product from transcript CG11529-RA', other_designations='CG11529-PA|uncharacterized protein', map_location='69A2-69A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12197791, end_position_on_the_genomic_accession=12199030, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39396, CurrentID=0, Status='live', Symbol='Adk1', Aliases='Dmel_CG17146, ADK-1, AK1, CG17146, DAK1, Dak1, Dmel\\CG17146, adk1, bs34e10.y1', description='Adenylate kinase 1', other_designations='Adk1-PA|Adk1-PB|CG17146-PA|CG17146-PB|adenylate kinase 1|adenylate kinase-1', map_location='69A2-69A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12200322, end_position_on_the_genomic_accession=12204082, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39397, CurrentID=0, Status='live', Symbol='CG5626', Aliases='Dmel_ Dmel\\CG5626', description='CG5626 gene product from transcript CG5626-RA', other_designations='CG5626-PA', map_location='69A2-69A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12204123, end_position_on_the_genomic_accession=12206022, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39399, CurrentID=0, Status='live', Symbol='app', Aliases='Dmel_CG42318, CG17144, CG42318, CG5620, CR32097, Dmel\\CG42318, Dmel_CG17144, Dmel_CG5620', description='approximated', other_designations='CG42318-PH|CG42318-PI|CG42318-PJ|CG42318-PK|CG42318-PL|CG42318-PM|CG42318-PO|CG42318-PR|CG42318-PS|CG42318-PT|app-PH|app-PI|app-PJ|app-PK|app-PL|app-PM|app-PO|app-PR|app-PS|app-PT', map_location='3-37.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12206238, end_position_on_the_genomic_accession=12266841, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  39400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39400, CurrentID=0, Status='live', Symbol='CG32100', Aliases='Dmel_ CG10429, CG18638, Dmel\\CG32100', description='CG32100 gene product from transcript CG32100-RA', other_designations='CG32100-PA|uncharacterized protein', map_location='69A4-69A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12267528, end_position_on_the_genomic_accession=12269803, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740656, CurrentID=0, Status='live', Symbol='CG34278', Aliases='Dmel_ Dmel\\CG34278', description='CG34278 gene product from transcript CG34278-RA', other_designations='CG34278-PA|uncharacterized protein', map_location='89E13-89E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17120465, end_position_on_the_genomic_accession=17120808, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39402, CurrentID=0, Status='live', Symbol='Pbgs', Aliases='Dmel_CG10335, CG10335, DmPBGS, Dmel\\CG10335, delta-ALA-D', description='Porphobilinogen synthase', other_designations='CG10335-PA|CG10335-PB|Pbgs-PA|Pbgs-PB|delta-aminolevulinate dehydratase|porphobilinogen synthase', map_location='69A4-69A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12269452, end_position_on_the_genomic_accession=12271345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39403, CurrentID=0, Status='live', Symbol='CG4300', Aliases='Dmel_ Dmel\\CG4300, anon-WO0118547.387', description='CG4300 gene product from transcript CG4300-RB', other_designations='CG4300-PA|CG4300-PB|CG4300-PC|CG4300-PD', map_location='69A4-69A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12271656, end_position_on_the_genomic_accession=12277251, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39404, CurrentID=0, Status='live', Symbol='CG10426', Aliases='Dmel_ Dmel\\CG10426, INPPE5, IPP', description='CG10426 gene product from transcript CG10426-RA', other_designations='CG10426-PA|CG10426-PB|pharbin-like|uncharacterized protein', map_location='69A5-69A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12277655, end_position_on_the_genomic_accession=12281476, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39405, CurrentID=0, Status='live', Symbol='CG4328', Aliases='Dmel_ Dmel\\CG4328', description='CG4328 gene product from transcript CG4328-RB', other_designations='CG4328-PB', map_location='69A5-69A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12290956, end_position_on_the_genomic_accession=12296123, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39406, CurrentID=0, Status='live', Symbol='CG32105', Aliases='Dmel_ CG10432, CG17615, Dmel\\CG32105', description='CG32105 gene product from transcript CG32105-RB', other_designations='CG32105-PB|CG32105-PC|uncharacterized protein', map_location='69B1-69B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12329232, end_position_on_the_genomic_accession=12335901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39408, CurrentID=0, Status='live', Symbol='CG10418', Aliases='Dmel_ BcDNA:GM14851, Dmel\\CG10418', description='CG10418 gene product from transcript CG10418-RA', other_designations='CG10418-PA|uncharacterized protein', map_location='69B2-69B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12336676, end_position_on_the_genomic_accession=12337236, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39409, CurrentID=0, Status='live', Symbol='CrzR', Aliases='Dmel_CG10698, CG10698, Crz-R, DCR, DGRHRII, Dmel\\CG10698, GRH-RII, GRHRII, GRHRII_Dro, anon-WO0170980.196, anon-WO0170980.197, crzR', description='Corazonin receptor', other_designations='CG10698-PA|CG10698-PB|Crz receptor|CrzR-PA|CrzR-PB|Drm-AKH receptor-2|GRHRII|corazonin receptor|drosophila corazonin receptor', map_location='69B2-69B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12339116, end_position_on_the_genomic_accession=12351544, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39410, CurrentID=0, Status='live', Symbol='Ncc69', Aliases='Dmel_CG4357, BEST:CK01027, CG43567, CG4357, CK01027, Dmel\\CG4357, NCC69, NKCC, mdcds_25302', description='sodium chloride cotransporter 69', other_designations='CG4357-PA|CG4357-PC|CG4357-PD|CG4357-PE|CG4357-PF|CG4357-PG|Ncc69-PA|Ncc69-PC|Ncc69-PD|Ncc69-PE|Ncc69-PF|Ncc69-PG', map_location='69B3-69B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12368646, end_position_on_the_genomic_accession=12391444, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  39411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39411, CurrentID=0, Status='live', Symbol='Obp69a', Aliases='Dmel_CG10436, 69a, CG10436, DmelObp69a, Dmel\\CG10436, PB-PRP1, PBP1, PBPRP-1, Pbprp1, pbprp-1, pbprp1', description='Odorant-binding protein 69a', other_designations='CG10436-PA|Obp69a-PA|Pbprp1|Pheromone-binding protein-related protein 1', map_location='69B4-69B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12396171, end_position_on_the_genomic_accession=12396957, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39413, CurrentID=0, Status='live', Symbol='CG32104', Aliases='Dmel_ CG14123, CG14124, Dmel\\CG32104, anon-69Ag', description='CG32104 gene product from transcript CG32104-RB', other_designations='CG32104-PB|uncharacterized protein', map_location='69B4-69B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12393804, end_position_on_the_genomic_accession=12399322, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  39414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39414, CurrentID=0, Status='live', Symbol='Smyd4', Aliases='Dmel_CG14122, CG14122, Dmel\\CG14122, dSmyd4', description='SET and MYND domain protein 4', other_designations='CG14122-PA|Smyd4-PA|dSmyd4', map_location='69B4-69B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12399539, end_position_on_the_genomic_accession=12401935, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39415, CurrentID=0, Status='live', Symbol='CG32103', Aliases='Dmel_ CG18022, CG4392, Dmel\\CG32103', description='CG32103 gene product from transcript CG32103-RE', other_designations='CG32103-PB|CG32103-PC|CG32103-PD|CG32103-PE|uncharacterized protein', map_location='69B4-69B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12402295, end_position_on_the_genomic_accession=12414842, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39417, CurrentID=0, Status='live', Symbol='ArfGAP1', Aliases='Dmel_CG4237, ARF1-GAP, ARF1GAP, CG4237, Dmel\\CG4237, GAP, GAP69C, Gap69C, Gap69c, arf1GAP, dArf1-GAP, gap69, i51', description='ADP-ribosylation factor GTPase activating protein 1', other_designations='ArfGAP1-PA|ArfGAP1-PB|CG4237-PA|CG4237-PB|GTPase-activating protein 69C', map_location='69B5-69B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12415026, end_position_on_the_genomic_accession=12417580, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39418, CurrentID=0, Status='live', Symbol='toe', Aliases='Dmel_CG10704, CG10704, Dmel\\CG10704, Toe, anon-WO0172774.89', description='twin of eyg', other_designations='CG10704-PA|CG10704-PB|toe-PA|toe-PB|twin of eyegone', map_location='69C1-69C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12422510, end_position_on_the_genomic_accession=12432668, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39419, CurrentID=0, Status='live', Symbol='eyg', Aliases='Dmel_CG10488, BK27, Bk27, CG10488, Dmel\\CG10488, EYG, Eyg, lun', description='eyegone', other_designations='CG10488-PA|CG10488-PB|CG10488-PC|eye gone|eyeone|eyg-PA|eyg-PB|eyg-PC|eygone|lune', map_location='69C2-69C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12462537, end_position_on_the_genomic_accession=12468021, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39420, CurrentID=0, Status='live', Symbol='CG10616', Aliases='Dmel_ Dmel\\CG10616', description='CG10616 gene product from transcript CG10616-RC', other_designations='CG10616-PA|CG10616-PB|CG10616-PC|uncharacterized protein', map_location='69C2-69C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12471728, end_position_on_the_genomic_accession=12474428, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39421, CurrentID=0, Status='live', Symbol='CG10663', Aliases='Dmel_ CT29868, Dmel\\CG10663, SP30', description='CG10663 gene product from transcript CG10663-RD', other_designations='CG10663-PB|CG10663-PD|uncharacterized protein', map_location='69C2-69C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12474418, end_position_on_the_genomic_accession=12485035, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39422, CurrentID=0, Status='live', Symbol='CG10660', Aliases='Dmel_ Dmel\\CG10660', description='CG10660 gene product from transcript CG10660-RD', other_designations='CG10660-PA|CG10660-PB|CG10660-PC|CG10660-PD|uncharacterized protein', map_location='69C2-69C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12486668, end_position_on_the_genomic_accession=12490676, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39423, CurrentID=0, Status='live', Symbol='CG10657', Aliases='Dmel_ BEST:LD32330, Dmel\\CG10657', description='CG10657 gene product from transcript CG10657-RA', other_designations='CG10657-PA|retinaldehyde-binding related protein|uncharacterized protein', map_location='69C3-69C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12491830, end_position_on_the_genomic_accession=12497216, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39424, CurrentID=0, Status='live', Symbol='CG10638', Aliases='Dmel_ AKR, CG32101, Dmel\\CG10638, anon-WO0172774.89', description='CG10638 gene product from transcript CG10638-RD', other_designations='CG10638-PA|CG10638-PB|CG10638-PD|aldo-keto reductase|uncharacterized protein', map_location='69C3-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12498106, end_position_on_the_genomic_accession=12501311, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39425, CurrentID=0, Status='live', Symbol='CG10681', Aliases='Dmel_ Dmel\\CG10681', description='CG10681 gene product from transcript CG10681-RB', other_designations='CG10681-PA|CG10681-PB|uncharacterized protein', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12501513, end_position_on_the_genomic_accession=12502647, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39426, CurrentID=0, Status='live', Symbol='CG10646', Aliases='Dmel_ Dmel\\CG10646, anon-WO0172774.52', description='CG10646 gene product from transcript CG10646-RA', other_designations='CG10646-PA|uncharacterized protein', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12502467, end_position_on_the_genomic_accession=12503894, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39428, CurrentID=0, Status='live', Symbol='CG10654', Aliases='Dmel_ Dmel\\CG10654', description='CG10654 gene product from transcript CG10654-RB', other_designations='CG10654-PA|CG10654-PB|uncharacterized protein', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12505374, end_position_on_the_genomic_accession=12507840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39429, CurrentID=0, Status='live', Symbol='sti', Aliases='Dmel_CG10522, CG10522, CIT-K, Citron K, DCK, Dmel\\CG10522, STI, anon-WO0118547.388, cit, dck, l(3)7m-62, l(3)7m62, l(3)SG21/Citron, sti', description='sticky', other_designations='CG10522-PA|CG10522-PB|citron|citron kinase|drosophila citron kinase|sti-PA|sti-PB', map_location='3-36.8 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12507998, end_position_on_the_genomic_accession=12514815, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39430, CurrentID=0, Status='live', Symbol='tral', Aliases='Dmel_CG10686, BEST:GH08269, CG10686, DmLSm15, Dmel\\CG10686, LSm15, Lsm15, Scd6p, TRAL, Tral, anon-EST:Posey139, anon-WO0172774.89, anon-WO02059370.24, p67, p70', description='trailer hitch', other_designations='CG10686-PA|CG10686-PB|CG10686-PC|CG10686-PD|CG10686-PE|CG10686-PF|CG10686-PG|CG10686-PH|CG10686-PI|Trailer-Hitch|tral-PA|tral-PB|tral-PC|tral-PD|tral-PE|tral-PF|tral-PG|tral-PH|tral-PI', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12515059, end_position_on_the_genomic_accession=12520954, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39431, CurrentID=0, Status='live', Symbol='Gcn5', Aliases='Dmel_CG4107, Ada4, CG4107, Dmel\\CG4107, GCN5/PCAF, Gcn5/Pcaf, KAT, KAT2, P/CAF, PCAF, Pcaf, anon-WO0172774.89, dGCN5, dGCN5 HAT, dGcn5, dGcn5i, dKAT2, dPCAF, dmGCN5, dmHAG401, gcn5, p/CAF, Gcn5', description='Gcn5 ortholog', other_designations='CG4107-PA|Gcn5-PA|pcaf', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12521129, end_position_on_the_genomic_accession=12524014, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39432, CurrentID=0, Status='live', Symbol='ver', Aliases='Dmel_CG14121, 1479/10, CG14121, Dmel\\CG14121, Ver, anon-WO0172774.86, anon-WO0172774.87, anon-WO0172774.89, l(3)S147910', description='verrocchio', other_designations='CG14121-PA|CG14121-PB|lethal (3) S147910|ver-PA|ver-PB', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12523987, end_position_on_the_genomic_accession=12525150, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39433, CurrentID=0, Status='live', Symbol='eIF-2beta', Aliases='Dmel_CG4153, CG4153, D-eIF2beta, Dmel\\CG4153, Eif2beta, IF2B, eIF-2, eIF2beta, elF2beta', description='Eukaryotic initiation factor 2beta', other_designations='CG4153-PA|eIF-2beta-PA|eukaryotic initiation factor 2beta', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12525093, end_position_on_the_genomic_accession=12526731, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39434, CurrentID=0, Status='live', Symbol='nst', Aliases='Dmel_CG10627, CG10627, Dmel\\CG10627', description='nesthocker', other_designations='CG10627-PA|nst-PA', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12527215, end_position_on_the_genomic_accession=12530171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39435, CurrentID=0, Status='live', Symbol='Tsf2', Aliases='Dmel_CG10620, CG10620, Dmel\\CG10620, MTf, tsf2', description='Transferrin 2', other_designations='CG10620-PA|Tsf2-PA|melanotransferrin|transferrin 2', map_location='69C4-69C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12530436, end_position_on_the_genomic_accession=12533724, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39436, CurrentID=0, Status='live', Symbol='CG10688', Aliases='Dmel_ 33589256, Dmel\\CG10688', description='CG10688 gene product from transcript CG10688-RA', other_designations='CG10688-PA|uncharacterized protein', map_location='69C6-69C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12534590, end_position_on_the_genomic_accession=12535513, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39437, CurrentID=0, Status='live', Symbol='CG4069', Aliases='Dmel_ Dmel\\CG4069, anon-WO0172774.58', description='CG4069 gene product from transcript CG4069-RA', other_designations='CG4069-PA', map_location='69C6-69C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12535426, end_position_on_the_genomic_accession=12537315, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39438, CurrentID=0, Status='live', Symbol='sowah', Aliases='Dmel_CG10632, BcDNA:GM09162, CG10632, Dmel\\CG10632', description='sosondowah', other_designations='CG10632-PG|CG10632-PH|CG10632-PI|CG10632-PJ|sowah-PG|sowah-PH|sowah-PI|sowah-PJ', map_location='69C6-69C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12537513, end_position_on_the_genomic_accession=12572542, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  39439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39439, CurrentID=0, Status='live', Symbol='ara', Aliases='Dmel_CG10571, Ara, CG10571, Dmel\\CG10571, IRO-C, Iro, Iro-C, IroC, iro, iro-C', description='araucan', other_designations='CG10571-PA|CG10571-PB|ara-PA|ara-PB|aracucan|iroquis complex|iroquois|iroquois gene complex|iroquois-complex', map_location='69C8-69C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12580672, end_position_on_the_genomic_accession=12596018, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39440, CurrentID=0, Status='live', Symbol='caup', Aliases='Dmel_CG10605, CG10605, Caup, Dmel\\CG10605, IRO-C, Iro, Iro-C, IroC, iro, iro-C', description='caupolican', other_designations='CG10605-PA|caup-PA|cauplican|iroquis complex|iroquois|iroquois gene complex|iroquois-complex', map_location='69D1-69D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12609631, end_position_on_the_genomic_accession=12621798, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39441, CurrentID=0, Status='live', Symbol='mirr', Aliases='Dmel_CG10601, CG10601, DH1, De1, De3, Dmel\\CG10601, Group F, IRO-C, Iro, Iro-C, IroC, MIRR, Mirr, Mrr, Sai, cre, crep, group F, iro, iro-C, l(3)69Ca, l(3)69Da, l(3)6D1, l(3)A5-3-42, mir, mrr', description='mirror', other_designations='CG10601-PA|CG10601-PB|CG10601-PC|caupalican|crepuscule|iroquis complex|iroquois|mirr-PA|mirr-PB|mirr-PC|sail', map_location='69D3-69D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12693639, end_position_on_the_genomic_accession=12710249, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39443, CurrentID=0, Status='live', Symbol='Ptp69D', Aliases='Dmel_CG10975, 69D, CG10975, CT30751, DPTP, DPTP69D, Dmel\\CG10975, Dptp69D, PTP69D, PTPase, Ptp, Ptp69d, R-PTP 69D, dptp69D, dptp69d, ptp69D, ptp69d', description='Protein tyrosine phosphatase 69D', other_designations='CG10975-PA|CG10975-PB|Ptp69D-PA|Ptp69D-PB|protein tyrosine phosphatase 69D|protein tyrosine phosphatase PTP69D', map_location='69D6-69E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12734945, end_position_on_the_genomic_accession=12741568, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  39444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39444, CurrentID=0, Status='live', Symbol='CG32109', Aliases='Dmel_ CG10982, Dmel\\CG32109', description='CG32109 gene product from transcript CG32109-RA', other_designations='CG32109-PA|uncharacterized protein', map_location='69E1-69E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12745344, end_position_on_the_genomic_accession=12746770, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39445, CurrentID=0, Status='live', Symbol='Klc', Aliases='Dmel_CG5433, CG5433, DKLC, Dmel\\CG5433, KLC, kinesin, klc', description='Kinesin light chain', other_designations='CG5433-PA|CG5433-PB|CG5433-PC|Klc-PA|Klc-PB|Klc-PC|kinesin|kinesin 1 light chain|kinesin I|kinesin light chain|kinesin light chain subunit of kinesin-1|kinesin-1|kinesin-1 light chain', map_location='69E1-69E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12746616, end_position_on_the_genomic_accession=12750056, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39446, CurrentID=0, Status='live', Symbol='CG10984', Aliases='Dmel_ Dmel\\CG10984', description='CG10984 gene product from transcript CG10984-RC', other_designations='CG10984-PA|CG10984-PB|CG10984-PC|CG10984-PD|uncharacterized protein', map_location='69E1-69E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12750119, end_position_on_the_genomic_accession=12753402, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39447, CurrentID=0, Status='live', Symbol='CG10973', Aliases='Dmel_ Dmel\\CG10973', description='CG10973 gene product from transcript CG10973-RB', other_designations='CG10973-PB|uncharacterized protein', map_location='69E1-69E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12753239, end_position_on_the_genomic_accession=12754774, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39448, CurrentID=0, Status='live', Symbol='CG32113', Aliases='Dmel_ CG11002, CG11003, Dmel\\CG32113', description='CG32113 gene product from transcript CG32113-RB', other_designations='CG32113-PB|uncharacterized protein', map_location='69E1-69E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12755049, end_position_on_the_genomic_accession=12768605, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  39450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39450, CurrentID=0, Status='live', Symbol='Hip1', Aliases='Dmel_CG10971, CG10971, Dmel\\CG10971R, Hip1', description='Huntingtin interacting protein 1', other_designations='CG10971-PA|CG10971-PB|Hip1-PA|Hip1-PB|huntingtin interacting protein 1', map_location='69E2-69E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12768432, end_position_on_the_genomic_accession=12773573, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39451, CurrentID=0, Status='live', Symbol='CG10969', Aliases='Dmel_ Dmel\\CG10969', description='CG10969 gene product from transcript CG10969-RA', other_designations='CG10969-PA|uncharacterized protein', map_location='69E2-69E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12778750, end_position_on_the_genomic_accession=12781097, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39452, CurrentID=0, Status='live', Symbol='CG17666', Aliases='Dmel_ Dmel\\CG17666, bs31g10.y1', description='CG17666 gene product from transcript CG17666-RA', other_designations='CG17666-PA|CG17666-PB|uncharacterized protein', map_location='69E2-69E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12781325, end_position_on_the_genomic_accession=12784377, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39453, CurrentID=0, Status='live', Symbol='CG42709', Aliases='Dmel_ CG17667, Dmel\\CG42709, Dmel_CG17667, EP(3)3622, EP3622', description='CG42709 gene product from transcript CG42709-RA', other_designations='CG42709-PA|CG42709-PB|CG42709-PC|CG42709-PD|uncharacterized protein', map_location='69E2-69E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12790192, end_position_on_the_genomic_accession=12796084, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39454, CurrentID=0, Status='live', Symbol='Atg1', Aliases='Dmel_CG10967, ATG1, CG10967, DK-4, DmATG1, Dmel\\CG10967, ULK1, UNC 51-like, Unc-51, Unc51, anon-WO0118547.287, atg1, dATG1, dAtg1, l(3)00305, unc-51, unc51', description='Autophagy-related 1', other_designations='Atg1-PA|Atg1-PB|Autophagy-specific gene 1|CG10967-PA|CG10967-PB|autophagy-related gene 1|autophagy-specific gene1|pegarn|uncoordinated movement-51', map_location='69E2-69E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12784429, end_position_on_the_genomic_accession=12805116, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  39455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39455, CurrentID=0, Status='live', Symbol='Sap130', Aliases='Dmel_CG11006, BcDNA:LD21213, CG11006, Dmel\\CG11006, SAP130, l(3)06924', description='Sin3A-associated protein 130', other_designations='CG11006-PA|CG11006-PB|CG11006-PC|CG11006-PE|CG11006-PF|CG11006-PG|Sap130-PA|Sap130-PB|Sap130-PC|Sap130-PE|Sap130-PF|Sap130-PG|Sin3A-Associated protein|lethal (3) 06924', map_location='69E3-69E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12805902, end_position_on_the_genomic_accession=12813036, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39456, CurrentID=0, Status='live', Symbol='CG10754', Aliases='Dmel_ Dmel\\CG10754', description='CG10754 gene product from transcript CG10754-RA', other_designations='CG10754-PA|uncharacterized protein', map_location='69E4-69E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12813028, end_position_on_the_genomic_accession=12814019, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39457, CurrentID=0, Status='live', Symbol='CG42588', Aliases='Dmel_ BP1016, BcDNA:SD20456, CG11008, CG32108, CG34017, Dmel\\CG42588, Dmel_CG11008, Dmel_CG32108, Dmel_CG34017', description='CG42588 gene product from transcript CG42588-RA', other_designations='CG42588-PA|CG42588-PB|uncharacterized protein', map_location='69E5-69E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12814254, end_position_on_the_genomic_accession=12824650, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  39458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39458, CurrentID=0, Status='live', Symbol='CG10960', Aliases='Dmel_ CT30701, Dmel\\CG10960', description='CG10960 gene product from transcript CG10960-RC', other_designations='CG10960-PA|CG10960-PB|CG10960-PC|uncharacterized protein', map_location='69E5-69E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12824880, end_position_on_the_genomic_accession=12843106, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39459, CurrentID=0, Status='live', Symbol='CG10948', Aliases='Dmel_ CT30663, Dmel\\CG10948, cg10948', description='CG10948 gene product from transcript CG10948-RB', other_designations='CG10948-PB|CG10948-PC|CG10948-PD|uncharacterized protein', map_location='69E6-69E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12844254, end_position_on_the_genomic_accession=12850117, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39460, CurrentID=0, Status='live', Symbol='Wbp2', Aliases='Dmel_CG11009, CG11009, Dmel\\CG11009, wbp2', description='Wbp2 ortholog', other_designations='CG11009-PB|CG11009-PC|CG11009-PD|CG11009-PE|CG11009-PF|CG11009-PG|Wbp2-PB|Wbp2-PC|Wbp2-PD|Wbp2-PE|Wbp2-PF|Wbp2-PG', map_location='69E6-69E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12850165, end_position_on_the_genomic_accession=12853169, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39461, CurrentID=0, Status='live', Symbol='Ent3', Aliases='Dmel_CG11010, CG11010, DmENT3, DmEnt3, Dmel\\CG11010, ent3', description='Equilibrative nucleoside transporter 3', other_designations='CG11010-PA|CG11010-PB|CG11010-PC|Ent3-PA|Ent3-PB|Ent3-PC|equilibrative nucleoside transporter 3', map_location='69E6-69E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12853646, end_position_on_the_genomic_accession=12857428, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39462, CurrentID=0, Status='live', Symbol='CG10943', Aliases='Dmel_ Dmel\\CG10943', description='CG10943 gene product from transcript CG10943-RA', other_designations='CG10943-PA|uncharacterized protein', map_location='69E6-69E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12861717, end_position_on_the_genomic_accession=12862920, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39463, CurrentID=0, Status='live', Symbol='CG14120', Aliases='Dmel_ Dmel\\CG14120', description='CG14120 gene product from transcript CG14120-RA', other_designations='CG14120-PA|uncharacterized protein', map_location='69E8-69E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12885039, end_position_on_the_genomic_accession=12889244, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39464, CurrentID=0, Status='live', Symbol='CG43894', Aliases='Dmel_ BP1086, CG14119, CG34032, Dmel\\CG43894, Dmel_CG14119, Dmel_CG34032, HDC09916', description='CG43894 gene product from transcript CG43894-RB', other_designations='CG43894-PB|CG43894-PC|CG43894-PD|CG43894-PE|uncharacterized protein', map_location='69E8-69E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12889770, end_position_on_the_genomic_accession=12892162, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39465, CurrentID=0, Status='live', Symbol='CG14118', Aliases='Dmel_ Dmel\\CG14118', description='CG14118 gene product from transcript CG14118-RA', other_designations='CG14118-PA|uncharacterized protein', map_location='69E8-69F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12892658, end_position_on_the_genomic_accession=12894758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39466, CurrentID=0, Status='live', Symbol='CG12520', Aliases='Dmel_ Dmel\\CG12520, FBgn0036324', description='CG12520 gene product from transcript CG12520-RB', other_designations='CG12520-PB|CG12520-PC|uncharacterized protein', map_location='69F1-69F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12901247, end_position_on_the_genomic_accession=12902737, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39467, CurrentID=0, Status='live', Symbol='CG10752', Aliases='Dmel_ Dmel\\CG10752, anon-WO0140519.236', description='CG10752 gene product from transcript CG10752-RA', other_designations='CG10752-PA|uncharacterized protein', map_location='69F2-69F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12952686, end_position_on_the_genomic_accession=12954647, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39469, CurrentID=0, Status='live', Symbol='CG10748', Aliases='Dmel_ Dmel\\CG10748', description='CG10748 gene product from transcript CG10748-RA', other_designations='CG10748-PA|uncharacterized protein', map_location='69F2-69F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12971214, end_position_on_the_genomic_accession=12972359, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39470, CurrentID=0, Status='live', Symbol='CG10749', Aliases='Dmel_ Dmel\\CG10749', description='CG10749 gene product from transcript CG10749-RA', other_designations='CG10749-PA|uncharacterized protein', map_location='69F2-69F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12972474, end_position_on_the_genomic_accession=12973650, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39471, CurrentID=0, Status='live', Symbol='CG11262', Aliases='Dmel_ Dmel\\CG11262', description='CG11262 gene product from transcript CG11262-RA', other_designations='CG11262-PA|uncharacterized protein', map_location='69F4-69F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12995135, end_position_on_the_genomic_accession=12997418, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39472, CurrentID=0, Status='live', Symbol='CG11263', Aliases='Dmel_ Dmel\\CG11263', description='CG11263 gene product from transcript CG11263-RA', other_designations='CG11263-PA|CG11263-PB|uncharacterized protein', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13006817, end_position_on_the_genomic_accession=13007881, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39473, CurrentID=0, Status='live', Symbol='CG14117', Aliases='Dmel_ BcDNA:RE44251, Dmel\\CG14117', description='CG14117 gene product from transcript CG14117-RE', other_designations='CG14117-PD|CG14117-PE|uncharacterized protein', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13008285, end_position_on_the_genomic_accession=13009450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39474, CurrentID=0, Status='live', Symbol='CG11261', Aliases='Dmel_ CUL6, Cullin6, Dmel\\CG11261', description='CG11261 gene product from transcript CG11261-RA', other_designations='CG11261-PA|uncharacterized protein', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13009399, end_position_on_the_genomic_accession=13011685, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39475, CurrentID=0, Status='live', Symbol='MICAL-like', Aliases='Dmel_CG11259, AAK93415, CG11259, Dmel\\CG11259', description='CG11259 gene product from transcript CG11259-RA', other_designations='CG11259-PA|MICAL-like-PA', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13011976, end_position_on_the_genomic_accession=13015332, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39476, CurrentID=0, Status='live', Symbol='CG11267', Aliases='Dmel_ Dmel\\CG11267, Hsp10, anon-WO0118547.175', description='CG11267 gene product from transcript CG11267-RA', other_designations='CG11267-PA|uncharacterized protein', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13016050, end_position_on_the_genomic_accession=13016903, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39477, CurrentID=0, Status='live', Symbol='mRpL20', Aliases='Dmel_CG11258, CG11258, Dmel\\CG11258, L20, anon-WO0118547.265', description='mitochondrial ribosomal protein L20', other_designations='CG11258-PA|mRpL20-PA', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13017149, end_position_on_the_genomic_accession=13017867, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39478, CurrentID=0, Status='live', Symbol='ste14', Aliases='Dmel_CG11268, CG11268, Dmel\\CG11268, Dmel\\Ste14, Ste14', description='isoprenylcysteine carboxylmethyltransferase', other_designations='CG11268-PA|CG11268-PB|ste14-PA|ste14-PB', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13018266, end_position_on_the_genomic_accession=13019687, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39479, CurrentID=0, Status='live', Symbol='AdenoK', Aliases='Dmel_CG11255, CG11255, DmAdk2, Dmel\\CG11255', description='Adenosine Kinase', other_designations='AdenoK-PA|AdenoK-PB|AdenoK-PC|CG11255-PA|CG11255-PB|CG11255-PC|adenosine kinase', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13019715, end_position_on_the_genomic_accession=13022688, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39480, CurrentID=0, Status='live', Symbol='RpS12', Aliases='Dmel_CG11271, BcDNA:RH29995, CG11271, Dmel\\CG11271, RS12E, S12, l(3)s2783, RpS12', description='Ribosomal protein S12', other_designations='40S ribosomal component|40S ribosomal protein S12|CG11271-PB|RpS12-PB|putative ribosomal protein S12|ribosomal protein S12|ribosomal protein gene 12', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13023352, end_position_on_the_genomic_accession=13024762, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39481, CurrentID=0, Status='live', Symbol='Zmynd10', Aliases='Dmel_CG11253, CG11253, Dmel\\CG11253', description='ZMYND10 homolog', other_designations='CG11253-PA|Zmynd10-PA', map_location='69F5-69F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13024763, end_position_on_the_genomic_accession=13026437, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39483, CurrentID=0, Status='live', Symbol='SRm160', Aliases='Dmel_CG11274, CG11274, Dmel\\CG11274, SRM160', description='CG11274 gene product from transcript CG11274-RA', other_designations='CG11274-PA|SRm160-PA', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13036821, end_position_on_the_genomic_accession=13040976, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39484, CurrentID=0, Status='live', Symbol='RpS4', Aliases='Dmel_CG11276, CG11276, Dmel\\CG11276, RPS4, Rp S4, SX4, anon-EST:Posey270, rpS4', description='Ribosomal protein S4', other_designations='CG11276-PA|CG11276-PB|CG11276-PC|RpS4-PA|RpS4-PB|RpS4-PC|ribosomal protein S4', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13041756, end_position_on_the_genomic_accession=13044234, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39485, CurrentID=0, Status='live', Symbol='Syx13', Aliases='Dmel_CG11278, CG11278, Dm Syx13, DmSyx13, Dmel\\CG11278, SYN13, Syx, Syx20, ff16, ff16 (69D1-70A2), l(3)01470, syx13', description='Syntaxin 13', other_designations='CG11278-PA|CG11278-PB|Syntaxin13|Syx13-PA|Syx13-PB|dSyntaxin13|syntaxin|syntaxin 13', map_location='3-38.63 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13044434, end_position_on_the_genomic_accession=13046894, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39486, CurrentID=0, Status='live', Symbol='CG11279', Aliases='Dmel_ BcDNA:GM21080, Dmel\\CG11279', description='CG11279 gene product from transcript CG11279-RC', other_designations='CG11279-PA|CG11279-PB|CG11279-PC|uncharacterized protein', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13046977, end_position_on_the_genomic_accession=13048919, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39487, CurrentID=0, Status='live', Symbol='CG14115', Aliases='Dmel_ Dmel\\CG14115', description='CG14115 gene product from transcript CG14115-RA', other_designations='CG14115-PA|uncharacterized protein', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13060522, end_position_on_the_genomic_accession=13061394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39489, CurrentID=0, Status='live', Symbol='CG17300', Aliases='Dmel_ Dmel\\CG17300', description='CG17300 gene product from transcript CG17300-RA', other_designations='CG17300-PA|uncharacterized protein', map_location='70A1-70A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13086387, end_position_on_the_genomic_accession=13087289, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39490, CurrentID=0, Status='live', Symbol='CG11251', Aliases='Dmel_ Dmel\\CG11251', description='CG11251 gene product from transcript CG11251-RA', other_designations='CG11251-PA|uncharacterized protein', map_location='70A1-70A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13088581, end_position_on_the_genomic_accession=13089969, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39491, CurrentID=0, Status='live', Symbol='trn', Aliases='Dmel_CG11280, 0024/05, 0641/17, CG11280, CT31487, Dmel\\CG11280, anon-EST:fe2B1', description='tartan', other_designations='CG11280-PA|CG11280-PB|TARTAN|anon-fast-evolving-2B1|trn-PA|trn-PB', map_location='70A1-70A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13114271, end_position_on_the_genomic_accession=13118077, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39492, CurrentID=0, Status='live', Symbol='snky', Aliases='Dmel_CG11281, CG11281, Dmel\\CG11281, ms(3)4, ms(3)ry4, ms(3)sneaky, ms(3)snky', description='sneaky', other_designations='CG11281-PB|snky-PB', map_location='3-40 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13174478, end_position_on_the_genomic_accession=13176781, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39493, CurrentID=0, Status='live', Symbol='caps', Aliases='Dmel_CG11282, BcDNA:LD07388, CAPS, CG11282, CT31495, Caps, Dmel\\CG11282, l(3)02937, l(3)rM632', description='capricious', other_designations='CG11282-PA|CG11282-PB|CG11282-PD|CG11282-PE|caps-PA|caps-PB|caps-PD|caps-PE', map_location='70A3-70A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13228692, end_position_on_the_genomic_accession=13279642, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39494, CurrentID=0, Status='live', Symbol='SP', Aliases='Dmel_CG17673, ACP, ACP70A, Acp-70A, Acp70, Acp70A, Acp70a, Apc70A, CG17673, Dmel\\CG17673, DrmSP, PapB, Spg70', description='Sex Peptide', other_designations='CG17673-PA|DrmSP|Paragonial-peptide-B|SP-PA|Sex peptide|Sex-Peptide|accessory gland peptide 70A|accessory gland protein 70A|accessory protein 70A|sex peptide gene in 70|sexpeptide', map_location='70A4-70A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13301635, end_position_on_the_genomic_accession=13301922, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39495, CurrentID=0, Status='live', Symbol='CG17687', Aliases='Dmel_ Dmel\\CG17687', description='CG17687 gene product from transcript CG17687-RA', other_designations='CG17687-PA|uncharacterized protein', map_location='70A5-70A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13349165, end_position_on_the_genomic_accession=13354940, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  39496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39496, CurrentID=0, Status='live', Symbol='SNCF', Aliases='Dmel_CG14112, BcDNA:RE54375, CG14112, Dmel\\CG14112, HL6, SCNF', description='SoxNeuro Co-Factor', other_designations='CG14112-PA|SNCF-PA', map_location='70A7-70A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13378583, end_position_on_the_genomic_accession=13379190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39497, CurrentID=0, Status='live', Symbol='CG14111', Aliases='Dmel_ Dmel\\CG14111', description='CG14111 gene product from transcript CG14111-RB', other_designations='CG14111-PB|uncharacterized protein', map_location='70A7-70A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13377957, end_position_on_the_genomic_accession=13380380, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39498, CurrentID=0, Status='live', Symbol='CG14107', Aliases='Dmel_ Dmel\\CG14107', description='CG14107 gene product from transcript CG14107-RA', other_designations='CG14107-PA|uncharacterized protein', map_location='70A7-70A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13380359, end_position_on_the_genomic_accession=13381346, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39499, CurrentID=0, Status='live', Symbol='ImpL1', Aliases='Dmel_CG10717, CG10717, Dmel\\CG10717, IMP-L1, Imp-L1', description='Ecdysone-inducible gene L1', other_designations='CG10717-PA|CG10717-PB|CG10717-PC|ImpL1-PA|ImpL1-PB|ImpL1-PC', map_location='70A7-70A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13382765, end_position_on_the_genomic_accession=13383999, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39500, CurrentID=0, Status='live', Symbol='CG14110', Aliases='Dmel_ Dmel\\CG14110', description='CG14110 gene product from transcript CG14110-RA', other_designations='CG14110-PA|uncharacterized protein', map_location='70A7-70A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13384789, end_position_on_the_genomic_accession=13385925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39501, CurrentID=0, Status='live', Symbol='CG10171', Aliases='Dmel_ Dmel\\CG10171', description='CG10171 gene product from transcript CG10171-RB', other_designations='CG10171-PA|CG10171-PB|uncharacterized protein', map_location='70A7-70A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13385946, end_position_on_the_genomic_accession=13387907, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39502, CurrentID=0, Status='live', Symbol='Poc1', Aliases='Dmel_CG10191, CG10191, Dmel\\CG10191, Pix, poc1', description='Proteome of centrioles 1', other_designations='CG10191-PA|CG10191-PB|Poc1-PA|Poc1-PB|proteome of centrioles 1', map_location='70A8-70A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13388299, end_position_on_the_genomic_accession=13390006, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39504, CurrentID=0, Status='live', Symbol='CG10222', Aliases='Dmel_ Dmel\\CG10222', description='CG10222 gene product from transcript CG10222-RB', other_designations='CG10222-PA|CG10222-PB|uncharacterized protein', map_location='70A8-70A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13411122, end_position_on_the_genomic_accession=13412453, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39505, CurrentID=0, Status='live', Symbol='flr', Aliases='Dmel_CG10724, AIP1, Aip1, CG10724, Dmel\\CG10724, aip1', description='flare', other_designations='CG10724-PA|CG10724-PB|actin interacting protein 1|actin-interacting protein 1|flare hair|flr-PA|flr-PB', map_location='3-38.8 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13412843, end_position_on_the_genomic_accession=13416453, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39506, CurrentID=0, Status='live', Symbol='CG14106', Aliases='Dmel_ Dmel\\CG14106', description='CG14106 gene product from transcript CG14106-RA', other_designations='CG14106-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13424571, end_position_on_the_genomic_accession=13425671, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39507, CurrentID=0, Status='live', Symbol='CG14105', Aliases='Dmel_ Dmel\\CG14105', description='CG14105 gene product from transcript CG14105-RA', other_designations='CG14105-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13425891, end_position_on_the_genomic_accession=13426673, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39508, CurrentID=0, Status='live', Symbol='CG10713', Aliases='Dmel_ CT28585, Dmel\\CG10713', description='CG10713 gene product from transcript CG10713-RA', other_designations='CG10713-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13428839, end_position_on_the_genomic_accession=13435229, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39509, CurrentID=0, Status='live', Symbol='CG10154', Aliases='Dmel_ Dmel\\CG10154', description='CG10154 gene product from transcript CG10154-RA', other_designations='CG10154-PA|CG10154-PB|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13435680, end_position_on_the_genomic_accession=13437136, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39510, CurrentID=0, Status='live', Symbol='CG10725', Aliases='Dmel_ Dmel\\CG10725', description='CG10725 gene product from transcript CG10725-RB', other_designations='CG10725-PB|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13437954, end_position_on_the_genomic_accession=13439009, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39511, CurrentID=0, Status='live', Symbol='CG10140', Aliases='Dmel_ Dmel\\CG10140', description='CG10140 gene product from transcript CG10140-RA', other_designations='CG10140-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13439317, end_position_on_the_genomic_accession=13440470, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39512, CurrentID=0, Status='live', Symbol='CG14109', Aliases='Dmel_ Dmel\\CG14109', description='CG14109 gene product from transcript CG14109-RB', other_designations='CG14109-PA|CG14109-PB|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13440910, end_position_on_the_genomic_accession=13442420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39513, CurrentID=0, Status='live', Symbol='cmb', Aliases='Dmel_CG10732, CG10732, Dmel\\CG10732, NEST:bs28e06', description='combover', other_designations='CG10732-PA|CG10732-PB|CG10732-PC|CG10732-PD|cmb-PA|cmb-PB|cmb-PC|cmb-PD', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13442594, end_position_on_the_genomic_accession=13449245, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39514, CurrentID=0, Status='live', Symbol='CG10133', Aliases='Dmel_ Dmel\\CG10133', description='CG10133 gene product from transcript CG10133-RA', other_designations='CG10133-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13449187, end_position_on_the_genomic_accession=13450283, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39515, CurrentID=0, Status='live', Symbol='CG10116', Aliases='Dmel_ Dmel\\CG10116', description='CG10116 gene product from transcript CG10116-RA', other_designations='CG10116-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13456208, end_position_on_the_genomic_accession=13457238, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39516, CurrentID=0, Status='live', Symbol='CG10738', Aliases='Dmel_ Dmel\\CG10738, GYC', description='CG10738 gene product from transcript CG10738-RD', other_designations='CG10738-PC|CG10738-PD|CG10738-PE|CG10738-PF|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13450770, end_position_on_the_genomic_accession=13465539, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  39517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39517, CurrentID=0, Status='live', Symbol='CG10089', Aliases='Dmel_ Dmel\\CG10089', description='CG10089 gene product from transcript CG10089-RE', other_designations='CG10089-PA|CG10089-PB|CG10089-PC|CG10089-PD|CG10089-PE|CG10089-PF|MKP-like|uncharacterized protein', map_location='70B2-70B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13465552, end_position_on_the_genomic_accession=13474787, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39518, CurrentID=0, Status='live', Symbol='stv', Aliases='Dmel_CG32130, CG10745, CG14108, CG32130, Dmel\\CG32130, GH04349, Stv, cg32130, l(3)00543', description='starvin', other_designations='CG32130-PA|CG32130-PB|CG32130-PC|CG32130-PD|CG32130-PE|CG32130-PF|CG32130-PG|Dm-Starvin|lethal (3) 00543|starving|stv-PA|stv-PB|stv-PC|stv-PD|stv-PE|stv-PF|stv-PG', map_location='70B2-70B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13477541, end_position_on_the_genomic_accession=13483515, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39520, CurrentID=0, Status='live', Symbol='Abp1', Aliases='Dmel_CG10083, CG10083, Dmel\\CG10083, abp1', description='Actin binding protein 1', other_designations='Abp1-PA|CG10083-PA|Drebrin-like|actin binding protein 1', map_location='70B2-70B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13483533, end_position_on_the_genomic_accession=13486280, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39521, CurrentID=0, Status='live', Symbol='Tgi', Aliases='Dmel_CG10741, CG10741, Dmel\\CG10741, NEST:bs23g10, SdBP, tgi', description='Tondu-domain-containing Growth Inhibitor', other_designations='CG10741-PA|CG10741-PB|Sd-Binding-Protein|Tgi-PA|Tgi-PB|Tondu-domain-containing growth inhibitor', map_location='70B2-70B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13486797, end_position_on_the_genomic_accession=13502492, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39522, CurrentID=0, Status='live', Symbol='Spt20', Aliases='Dmel_CG17689, CG17689, Dmel\\CG17689, lincRNA.S5070', description='CG17689 gene product from transcript CG17689-RA', other_designations='CG17689-PA|CG17689-PC|CG17689-PD|Spt20-PA|Spt20-PC|Spt20-PD', map_location='70B3-70B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13502707, end_position_on_the_genomic_accession=13511541, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39523, CurrentID=0, Status='live', Symbol='Vps36', Aliases='Dmel_CG10711, CG10711, Dmel\\CG10711, l(3)L5212, vps36', description='Vacuolar protein sorting 36', other_designations='CG10711-PA|Vps36-PA|lethal (3) L5212|vacuolar protein sorting 36', map_location='70B7-70B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13511646, end_position_on_the_genomic_accession=13513112, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39524, CurrentID=0, Status='live', Symbol='Liprin-beta', Aliases='Dmel_CG10743, CG10743, Dmel\\CG10743', description='CG10743 gene product from transcript CG10743-RB', other_designations='CG10743-PA|CG10743-PB|CG10743-PC|Liprin-beta-PA|Liprin-beta-PB|Liprin-beta-PC', map_location='70B4-70B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13513254, end_position_on_the_genomic_accession=13516276, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39525, CurrentID=0, Status='live', Symbol='CG10710', Aliases='Dmel_ Dmel\\CG10710', description='CG10710 gene product from transcript CG10710-RA', other_designations='CG10710-PA|uncharacterized protein', map_location='70B4-70B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13516273, end_position_on_the_genomic_accession=13519605, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39527, CurrentID=0, Status='live', Symbol='bru-3', Aliases='Dmel_CG43744, Bru-3, CG10046, CG12478, CG34243, CG43744, Dm Bru-3, Dmel\\CG43744, Dmel_CG12478, Dmel_CG34243, cg10046, cg12478, l(3)05871', description='bruno-3', other_designations='CG43744-PA|CG43744-PB|CG43744-PC|CG43744-PD|CG43744-PE|CG43744-PF|CG43744-PG|CG43744-PH|CG43744-PI|CG43744-PJ|CG43744-PK|CG43744-PL|CG43744-PM|CG43744-PN|CG43744-PO|bru-3-PA|bru-3-PB|bru-3-PC|bru-3-PD|bru-3-PE|bru-3-PF|bru-3-PG|bru-3-PH|bru-3-PI|bru-3-PJ|bru-3-PK|bru-3-PL|bru-3-PM|bru-3-PN|bru-3-PO|bruno3|lethal (3) 05871', map_location='70C1-70C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13521907, end_position_on_the_genomic_accession=13803400, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  39528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39528, CurrentID=0, Status='live', Symbol='CG8757', Aliases='Dmel_ Dmel\\CG8757, FOHSDR', description='CG8757 gene product from transcript CG8757-RB', other_designations='CG8757-PB', map_location='70C4-70C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13840197, end_position_on_the_genomic_accession=13842240, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39529, CurrentID=0, Status='live', Symbol='Hml', Aliases='Dmel_CG7002, CG7002, CT21553, Dmel\\CG7002, d-hml, hml', description='Hemolectin', other_designations='CG7002-PA|CG7002-PB|Hml-PA|Hml-PB|hemolectin', map_location='70C4-70C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13846054, end_position_on_the_genomic_accession=13860001, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  39530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39530, CurrentID=0, Status='live', Symbol='CG8745', Aliases='Dmel_ Dmel\\CG8745', description='CG8745 gene product from transcript CG8745-RA', other_designations='CG8745-PA', map_location='70C5-70C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13860027, end_position_on_the_genomic_accession=13866905, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39531, CurrentID=0, Status='live', Symbol='CG13737', Aliases='Dmel_ Dmel\\CG13737', description='CG13737 gene product from transcript CG13737-RA', other_designations='CG13737-PA|CG13737-PB|uncharacterized protein', map_location='70C5-70C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13878994, end_position_on_the_genomic_accession=13883056, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39533, CurrentID=0, Status='live', Symbol='dysc', Aliases='Dmel_CG43749, CG17368, CG32131, CG32132, CG34400, CG43749, CG8760, Dm_3L:69335, Dmel\\CG43749, Dmel_CG17368, Dmel_CG32131, Dmel_CG32132, Dmel_CG34400, Dmel_CG8760, anon-AE003536.1', description='dyschronic', other_designations='CG43749-PC|CG43749-PE|CG43749-PF|CG43749-PG|CG43749-PH|CG43749-PI|CG43749-PJ|dysc-PC|dysc-PE|dysc-PF|dysc-PG|dysc-PH|dysc-PI|dysc-PJ', map_location='70C5-70C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13867414, end_position_on_the_genomic_accession=13903236, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  39535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39535, CurrentID=0, Status='live', Symbol='CG8833', Aliases='Dmel_ Dmel\\CG8833, GPTC1', description='CG8833 gene product from transcript CG8833-RA', other_designations='CG8833-PA', map_location='70C5-70C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13925882, end_position_on_the_genomic_accession=13929000, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39536, CurrentID=0, Status='live', Symbol='DCTN1-p150', Aliases='Dmel_CG9206, CG9206, DYNA_DROME, Dmel\\CG9206, FCP-D, Gl, Glu, Glued, P150, P150[Glued], p150, p150/Glued, p150/glued, p150[Glued], p150[glued], t1', description='Dynactin 1, p150 subunit', other_designations='CG9206-PA|DCTN1-p150-PA|dynactin|dynactin 1, p150 subunit|dynactin P150|glued|p150-Glued|p150[glued]|p150glued', map_location='3-41.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13929387, end_position_on_the_genomic_accession=13934656, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39537, CurrentID=0, Status='live', Symbol='CG32137', Aliases='Dmel_ CG17365, CG17366, Dmel\\CG32137', description='CG32137 gene product from transcript CG32137-RB', other_designations='CG32137-PA|CG32137-PB|uncharacterized protein', map_location='70C6-70C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13939250, end_position_on_the_genomic_accession=13962647, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39539, CurrentID=0, Status='live', Symbol='Meics', Aliases='Dmel_CG8474, CG8474, Dmel\\CG8474', description='Meiotic central spindle', other_designations='CG8474-PA|CG8474-PB|Meics-PA|Meics-PB|meiotic central spindle|zinc-finger-motif-protein', map_location='70C7-70C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13962658, end_position_on_the_genomic_accession=13964702, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39540, CurrentID=0, Status='live', Symbol='ssp2', Aliases='Dmel_CG9028, CG9028, Dmel\\CG9028, Ssp2', description='short spindle 2', other_designations='CG9028-PA|CG9028-PC|sentin|short spindle|ssp2-PA|ssp2-PC', map_location='70C7-70C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13965203, end_position_on_the_genomic_accession=13973067, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39541, CurrentID=0, Status='live', Symbol='CG13738', Aliases='Dmel_ BcDNA:RE61257, Dmel\\CG13738', description='CG13738 gene product from transcript CG13738-RA', other_designations='CG13738-PA|uncharacterized protein', map_location='70C9-70C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13973123, end_position_on_the_genomic_accession=13973477, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39542, CurrentID=0, Status='live', Symbol='Hsc70-1', Aliases='Dmel_CG8937, 143191_at, BcDNA:RH49358, CG8937, DmHsp70, Dmel\\CG8937, HSC-70, HSC1, HSC70, HSC70-1, HSP7A_DROME, Hsc1, Hsc70, Hsp-c1, dhsc70, dhsc70-1, hsc70', description='Heat shock protein cognate 1', other_designations='CG8937-PA|CG8937-PC|CG8937-PD|Hsc70-1-PA|Hsc70-1-PC|Hsc70-1-PD|heat shock protein cognate 1', map_location='70C9-70C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13974061, end_position_on_the_genomic_accession=13977761, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39543, CurrentID=0, Status='live', Symbol='CG17364', Aliases='Dmel_ Dmel\\CG17364', description='CG17364 gene product from transcript CG17364-RF', other_designations='CG17364-PC|CG17364-PD|CG17364-PE|CG17364-PF|uncharacterized protein', map_location='70C9-70C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13977905, end_position_on_the_genomic_accession=13993400, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39545, CurrentID=0, Status='live', Symbol='CG17362', Aliases='Dmel_ Dmel\\CG17362', description='CG17362 gene product from transcript CG17362-RB', other_designations='CG17362-PB|uncharacterized protein', map_location='70C9-70C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13986620, end_position_on_the_genomic_accession=13987184, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39546, CurrentID=0, Status='live', Symbol='CG9040', Aliases='Dmel_ BcDNA:LP10528, Dmel\\CG9040', description='CG9040 gene product from transcript CG9040-RA', other_designations='CG9040-PA', map_location='70C9-70C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13987878, end_position_on_the_genomic_accession=13988626, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39547, CurrentID=0, Status='live', Symbol='26-29-p', Aliases='Dmel_CG8947, 26/29kD-proteinase, CG8947, Dmel\\CG8947, anon-SAGE:Wang-123, l(3)s3635', description='26-29kD-proteinase', other_designations='26,29kDa proteinase|26-29 kD-proteinase|26-29-p-PA|26-29kDa proteinase|26-29kd prot|CG8947-PA|anon-SAGE:Wang-123|cathepsin K', map_location='70C10-70C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13994231, end_position_on_the_genomic_accession=13996571, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39548, CurrentID=0, Status='live', Symbol='CG17361', Aliases='Dmel_ Dmel\\CG17361', description='CG17361 gene product from transcript CG17361-RA', other_designations='CG17361-PA|uncharacterized protein', map_location='70C10-70C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13997191, end_position_on_the_genomic_accession=13998055, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39549, CurrentID=0, Status='live', Symbol='CG17359', Aliases='Dmel_ Dmel\\CG17359', description='CG17359 gene product from transcript CG17359-RA', other_designations='CG17359-PA|uncharacterized protein', map_location='70C11-70C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13997900, end_position_on_the_genomic_accession=13999251, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39550, CurrentID=0, Status='live', Symbol='Nprl3', Aliases='Dmel_CG8783, CG8783, Dmel\\CG8783, NPR3, nprl3', description='Nitrogen permease regulator-like 3', other_designations='CG8783-PA|CG8783-PB|CG8783-PC|Nprl3-PA|Nprl3-PB|Nprl3-PC|nitrogen permease regulator-like 3', map_location='70C11-70C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13999380, end_position_on_the_genomic_accession=14001880, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39551, CurrentID=0, Status='live', Symbol='upSET', Aliases='Dmel_CG9007, CG9007, Dmel\\CG9007', description='CG9007 gene product from transcript CG9007-RB', other_designations='CG9007-PA|CG9007-PB|upSET-PA|upSET-PB', map_location='70C11-70C12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14002127, end_position_on_the_genomic_accession=14020148, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39552, CurrentID=0, Status='live', Symbol='Ptip', Aliases='Dmel_CG32133, CG32133, CG6532, CG8797, Dmel\\CG32133, PTIP, ptip', description='PAX transcription activation domain interacting protein', other_designations='CG32133-PA|Ptip-PA', map_location='70C12-70C14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14020315, end_position_on_the_genomic_accession=14029121, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39554, CurrentID=0, Status='live', Symbol='endos', Aliases='Dmel_CG6513, 0670/06, CG6513, Dmel\\CG6513, End, dendos, l(3)S067006', description='endosulfine', other_designations='CG6513-PA|CG6513-PB|CG6513-PC|CG6513-PD|CG6513-PE|CG6513-PF|alpha-endosulfine|dendos|endos-PA|endos-PB|endos-PC|endos-PD|endos-PE|endos-PF', map_location='70C14-70C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14029579, end_position_on_the_genomic_accession=14031706, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39555, CurrentID=0, Status='live', Symbol='CG6650', Aliases='Dmel_ 0670/06, Dmel\\CG6650, anon-WO0172774.121, anon-WO0172774.122, anon-WO0172774.123, l(3)067006, l(3)S067006', description='CG6650 gene product from transcript CG6650-RA', other_designations='CG6650-PA|CG6650-PB|CG6650-PC', map_location='70C15-70C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14031450, end_position_on_the_genomic_accession=14034909, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39556, CurrentID=0, Status='live', Symbol='CG6661', Aliases='Dmel_ CR33944, Dmel\\CG6661, PUT2', description='CG6661 gene product from transcript CG6661-RA', other_designations='CG6661-PA', map_location='70C15-70C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14034776, end_position_on_the_genomic_accession=14036751, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39557, CurrentID=0, Status='live', Symbol='Hsc70Cb', Aliases='Dmel_CG6603, 0041/12, 0318/20, 1342/17, 1348/02, 1485/13, BEST:LD32772, CG6603, Dmel\\CG6603, HSC-70, HSC70, HSC70cb, HSP110, HSP110-1, Hsc70, LD32772, anon-WO0118547.302, dHsp110, hsc70Cb, hsc70cb, l(3)00082, l(3)07621, l(3)134217, l(3)148513, l(3)70Ca, l(3)70Cb, l(3)S134217, l(3)S134802, l(3)S148513', description='CG6603 gene product from transcript CG6603-RI', other_designations='CG6603-PA|CG6603-PB|CG6603-PC|CG6603-PE|CG6603-PF|CG6603-PG|CG6603-PH|CG6603-PI|Hsc70Cb-PA|Hsc70Cb-PB|Hsc70Cb-PC|Hsc70Cb-PE|Hsc70Cb-PF|Hsc70Cb-PG|Hsc70Cb-PH|Hsc70Cb-PI', map_location='70C15-70C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14037019, end_position_on_the_genomic_accession=14043686, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39558, CurrentID=0, Status='live', Symbol='blue', Aliases='Dmel_CG6451, CG6451, Dmel\\CG6451, bls', description='bluestreak', other_designations='CG6451-PA|blue-PA', map_location='70C15-70D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14043694, end_position_on_the_genomic_accession=14049892, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39559, CurrentID=0, Status='live', Symbol='CG6833', Aliases='Dmel_ Dmel\\CG6833', description='CG6833 gene product from transcript CG6833-RA', other_designations='CG6833-PA|CG6833-PB', map_location='70D1-70D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14051206, end_position_on_the_genomic_accession=14052469, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39560, CurrentID=0, Status='live', Symbol='CG13484', Aliases='Dmel_ Dmel\\CG13484', description='CG13484 gene product from transcript CG13484-RA', other_designations='CG13484-PA|uncharacterized protein', map_location='70D1-70D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14052428, end_position_on_the_genomic_accession=14053079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39561, CurrentID=0, Status='live', Symbol='CG32138', Aliases='Dmel_ CG13485, CG43986, CG6807, Dmel\\CG32138, Dmel_CG43986, Q9VUC5_DROME, Q9VUC6_DROME, anon-WO0172774.98', description='CG32138 gene product from transcript CG32138-RC', other_designations='CG32138-PA|CG32138-PB|CG32138-PC|CG32138-PD|uncharacterized protein', map_location='70D1-70D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14053215, end_position_on_the_genomic_accession=14071583, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  39564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39564, CurrentID=0, Status='live', Symbol='btl', Aliases='Dmel_CG32134, 0844/01, BTL/FGFR2, Btl, CG32134, CG6714, CT20816, D-FGFR, DFGF-R1, DFR2, Dfr-2, DmHD-311, Dmel\\CG32134, Dtk2, FGFR, FGFR1, Fgf-r, HD-311, Tk2, dFGFR, dev, dtk2, fgf-r, l(3)00208, lambdatop', description='breathless', other_designations='CG32134-PA|CG32134-PB|FGF receptor|FGFR|Fibroblast-growth-factor-receptor|btl-PA|btl-PB|devenir|fibrobast growth factor receptor|fibroblast growth factor receptor|fibroblast growth factor receptor 1|no terminal cell clones-L', map_location='3-40.9 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14075772, end_position_on_the_genomic_accession=14083229, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39565, CurrentID=0, Status='live', Symbol='CG8100', Aliases='Dmel_ Dmel\\CG8100, lincRNA.S5087', description='CG8100 gene product from transcript CG8100-RC', other_designations='CG8100-PA|CG8100-PB|CG8100-PC', map_location='70D2-70D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14083615, end_position_on_the_genomic_accession=14086421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39566, CurrentID=0, Status='live', Symbol='Fbp1', Aliases='Dmel_CG17285, CG17285, DmeFBP1, Dmel\\CG17285, FBP-1, FBP1, FbP1, Fbp-1, LP07910p, P1, fbp-1, fbp1', description='Fat body protein 1', other_designations='CG17285-PA|CG17285-PB|CG17285-PD|Fbp1-PA|Fbp1-PB|Fbp1-PD|Protein-1|fat body protein 1|fat body protein-1|fat body protein1', map_location='70D2-70D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14094589, end_position_on_the_genomic_accession=14098053, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39567, CurrentID=0, Status='live', Symbol='Sox21a', Aliases='Dmel_CG7345, CG7345, Dmel\\CG7345, SOXB2.3, Sox B2-3, Sox21, SoxB2.3', description='CG7345 gene product from transcript CG7345-RA', other_designations='CG7345-PA|CG7345-PB|Sox21a-PA|Sox21a-PB', map_location='70D2-70D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14103481, end_position_on_the_genomic_accession=14108118, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39569, CurrentID=0, Status='live', Symbol='Sox21b', Aliases='Dmel_CG32139, CG13483, CG32139, CG6419, Dmel\\CG32139, SOXB2.2, Sox B2-2, SoxB2.2', description='CG32139 gene product from transcript CG32139-RA', other_designations='CG32139-PA|CG32139-PB|CG32139-PC|Sox21b-PA|Sox21b-PB|Sox21b-PC|sox-like', map_location='70D2-70D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14112405, end_position_on_the_genomic_accession=14131324, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39570, CurrentID=0, Status='live', Symbol='D', Aliases='Dmel_CG5893, CG5893M10, DM23, DM33, DM36, DM63, Dmel\\CG5893, SOX70, SOX70D, SOXB2.1, SOXDP, Sox B2-1, Sox70, Sox70D, Sry-rDM10, Sry-rDM23, Sry-rDM33, Sry-rDM36, Sry-rDM63, fish, l(3)LG9, loD, D', description='Dichaete', other_designations='CG5893-PA|CG5893-PB|CG5893-PC|CG5893-PD|D-PA|D-PB|D-PC|D-PD|Fish-Hook|Sox box protein 70D|Sry-related DM10|Sry-related DM23|Sry-related DM33|Sry-related DM36|Sry-related DM63|dichaete|fish hook|lethal of dichaete', map_location='3-40.7 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14175610, end_position_on_the_genomic_accession=14178620, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39571, CurrentID=0, Status='live', Symbol='nan', Aliases='Dmel_CG5842, CG5842, CT18317, DmNan, Dmel\\CG5842, NAN, Nan, OCR, TRPV', description='nanchung', other_designations='CG5842-PA|CG5842-PB|NANCHUNG|nan-PA|nan-PB|transient receptor potential vanilloid', map_location='70D3-70D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14186249, end_position_on_the_genomic_accession=14189648, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39572, CurrentID=0, Status='live', Symbol='nuf', Aliases='Dmel_CG33991, CG32140, CG33991, CG7867, CIP-D2, Cy3-23, D2, Dmel\\CG33991, EP3077, FIP3, NUF, Nuf, Nuf1, Nuf2p, anon-D2', description='nuclear fallout', other_designations='CG33991-PA|CG33991-PB|CG33991-PC|CG33991-PD|CG33991-PE|CG33991-PF|CG33991-PG|CG33991-PH|CG33991-PJ|CG33991-PK|CG33991-PL|CG33991-PM|corkscrew interacting Protein-D2|nuclear-fallout|nuf-PA|nuf-PB|nuf-PC|nuf-PD|nuf-PE|nuf-PF|nuf-PG|nuf-PH|nuf-PJ|nuf-PK|nuf-PL|nuf-PM', map_location='70D3-70D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14190870, end_position_on_the_genomic_accession=14233250, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  39573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39573, CurrentID=0, Status='live', Symbol='CG7768', Aliases='Dmel_ Dmel\\CG7768', description='CG7768 gene product from transcript CG7768-RA', other_designations='CG7768-PA|CG7768-PB', map_location='70D4-70D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14218204, end_position_on_the_genomic_accession=14218974, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39574, CurrentID=0, Status='live', Symbol='CG7924', Aliases='Dmel_ Dmel\\CG7924', description='CG7924 gene product from transcript CG7924-RA', other_designations='CG7924-PA', map_location='70D4-70D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14233469, end_position_on_the_genomic_accession=14234609, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39575, CurrentID=0, Status='live', Symbol='CG7906', Aliases='Dmel_ Dmel\\CG7906', description='CG7906 gene product from transcript CG7906-RB', other_designations='CG7906-PA|CG7906-PB', map_location='70D4-70D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14236635, end_position_on_the_genomic_accession=14238272, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39577, CurrentID=0, Status='live', Symbol='CG13482', Aliases='Dmel_ Dmel\\CG13482', description='CG13482 gene product from transcript CG13482-RA', other_designations='CG13482-PA|uncharacterized protein', map_location='70D5-70D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14329346, end_position_on_the_genomic_accession=14329885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39579, CurrentID=0, Status='live', Symbol='CG13481', Aliases='Dmel_ Dmel\\CG13481', description='CG13481 gene product from transcript CG13481-RC', other_designations='CG13481-PB|CG13481-PC|uncharacterized protein', map_location='70D5-70D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14368925, end_position_on_the_genomic_accession=14369707, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39580, CurrentID=0, Status='live', Symbol='CG3868', Aliases='Dmel_ Dmel\\CG3868', description='CG3868 gene product from transcript CG3868-RA', other_designations='CG3868-PA', map_location='70D6-70D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14400769, end_position_on_the_genomic_accession=14402725, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39581, CurrentID=0, Status='live', Symbol='stwl', Aliases='Dmel_CG3836, 84, CG3836, Dmel\\CG3836, Stwl, fs(3)02024, snw', description='stonewall', other_designations='CG3836-PA|CG3836-PB|CG3836-PC|alpha transcript|stwl-PA|stwl-PB|stwl-PC', map_location='70D6-70D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14402930, end_position_on_the_genomic_accession=14409939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39582, CurrentID=0, Status='live', Symbol='CG3919', Aliases='Dmel_ Dmel\\CG3919', description='CG3919 gene product from transcript CG3919-RC', other_designations='CG3919-PB|CG3919-PC', map_location='70D7-70D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14410626, end_position_on_the_genomic_accession=14412732, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39583, CurrentID=0, Status='live', Symbol='bbg', Aliases='Dmel_CG42230, BBG, C96, CG42230, CG9587, CG9598, Dmel\\CG42230, Dmel_CG9587, Dmel_CG9598', description='big bang', other_designations='C96|CG42230-PC|CG42230-PE|CG42230-PF|CG42230-PG|CG42230-PH|CG42230-PI|CG42230-PJ|CG42230-PK|bbg-PC|bbg-PE|bbg-PF|bbg-PG|bbg-PH|bbg-PI|bbg-PJ|bbg-PK', map_location='70E1-70D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14412828, end_position_on_the_genomic_accession=14536276, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  39585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39585, CurrentID=0, Status='live', Symbol='CG9592', Aliases='Dmel_ Dmel\\CG9592', description='CG9592 gene product from transcript CG9592-RB', other_designations='CG9592-PB', map_location='70E1-70E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14517825, end_position_on_the_genomic_accession=14518475, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39586, CurrentID=0, Status='live', Symbol='CG4613', Aliases='Dmel_ Dmel\\CG4613, SP40', description='CG4613 gene product from transcript CG4613-RC', other_designations='CG4613-PB|CG4613-PC', map_location='70E1-70E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14519760, end_position_on_the_genomic_accession=14521690, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39587, CurrentID=0, Status='live', Symbol='Mpcp', Aliases='Dmel_CG4994, CG4994, Dmel\\CG4994, MPCP, Mcpc, PpsM, l(3)00564', description='Mitochondrial phosphate carrier protein', other_designations='CG4994-PA|CG4994-PB|CG4994-PC|Mpcp-PA|Mpcp-PB|Mpcp-PC|mitochondrial phophate carrier protein|mitochondrial phosphate carrier protein|mitochondrial proton/phosphate symporter', map_location='70E1-70E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14537651, end_position_on_the_genomic_accession=14540441, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39588, CurrentID=0, Status='live', Symbol='Gbs-70E', Aliases='Dmel_CG9238, CG9238, Dmel\\CG9238, anon-WO0118547.400', description='Glycogen binding subunit 70E', other_designations='CG9238-PA|CG9238-PB|CG9238-PC|Gbs-70E-PA|Gbs-70E-PB|Gbs-70E-PC|glycogen binding subunit 70E', map_location='70E1-70E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14540937, end_position_on_the_genomic_accession=14550975, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39592, CurrentID=0, Status='live', Symbol='shd', Aliases='Dmel_CG13478, 314a1, CG13478, CYP314A1, Cyp314a1, Dmel\\CG13478, cyp314a1', description='shade', other_designations='CG13478-PB|CG13478-PC|CG13478-PD|cytochrome P450|ecdysone 20-monooxygenase|shd-PB|shd-PC|shd-PD', map_location='3-41 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14614458, end_position_on_the_genomic_accession=14622205, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39593, CurrentID=0, Status='live', Symbol='CG9628', Aliases='Dmel_ Dmel\\CG9628', description='CG9628 gene product from transcript CG9628-RA', other_designations='CG9628-PA|CG9628-PB|CG9628-PC', map_location='70E4-70E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14622796, end_position_on_the_genomic_accession=14628174, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39594, CurrentID=0, Status='live', Symbol='RecQ5', Aliases='Dmel_CG4879, CG4879, DQ1, Dm RECQ5/QE, DmRECQ5, Dmel\\CG4879, Dq1, Q1, RECQ5, RECQ5/QE, RECQE, anon-EST:Posey293, dRecQ5, dmRecQ5b', description='RecQ5 helicase', other_designations='CG4879-PA|CG4879-PB|CG4879-PC|RecQ5-PA|RecQ5-PB|RecQ5-PC|drosophila homolog of RecQ|homolog of RecQ', map_location='70E5-70E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14628160, end_position_on_the_genomic_accession=14632077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39596, CurrentID=0, Status='live', Symbol='dlp', Aliases='Dmel_CG32146, CG17703, CG32146, CG5031, CT16138, D-gpcB, DLP, Dally-like, Dlp, Dly, Dmel\\CG32146, dly', description='dally-like', other_designations='CG32146-PA|CG32146-PC|D-glypican B|Dally-like|Dally-like protein|dallylike|dlp-PA|dlp-PC', map_location='70E5-70E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14632601, end_position_on_the_genomic_accession=14671609, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39597, CurrentID=0, Status='live', Symbol='CG4914', Aliases='Dmel_ Dmel\\CG4914, SP39', description='CG4914 gene product from transcript CG4914-RA', other_designations='CG4914-PA', map_location='70E7-70E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14674538, end_position_on_the_genomic_accession=14678147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39598, CurrentID=0, Status='live', Symbol='CG5048', Aliases='Dmel_ Dmel\\CG5048, anon-WO0140519.238', description='CG5048 gene product from transcript CG5048-RA', other_designations='CG5048-PA', map_location='70F1-70F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14684768, end_position_on_the_genomic_accession=14685842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39599, CurrentID=0, Status='live', Symbol='CG13477', Aliases='Dmel_ Dmel\\CG13477', description='CG13477 gene product from transcript CG13477-RA', other_designations='CG13477-PA|uncharacterized protein', map_location='70F1-70F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14691347, end_position_on_the_genomic_accession=14691962, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39600, CurrentID=0, Status='live', Symbol='CG13474', Aliases='Dmel_ BcDNA:AT29351, Dmel\\CG13474', description='CG13474 gene product from transcript CG13474-RA', other_designations='CG13474-PA|uncharacterized protein', map_location='70F1-70F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14696551, end_position_on_the_genomic_accession=14697105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39601, CurrentID=0, Status='live', Symbol='CG17177', Aliases='Dmel_ Dmel\\CG17177', description='CG17177 gene product from transcript CG17177-RA', other_designations='CG17177-PA|CG17177-PB|uncharacterized protein', map_location='70F1-70F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14697313, end_position_on_the_genomic_accession=14701678, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39602, CurrentID=0, Status='live', Symbol='CG13476', Aliases='Dmel_ Dmel\\CG13476', description='CG13476 gene product from transcript CG13476-RB', other_designations='CG13476-PB|uncharacterized protein', map_location='70F1-70F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14705464, end_position_on_the_genomic_accession=14706157, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39603, CurrentID=0, Status='live', Symbol='CG13473', Aliases='Dmel_ Dmel\\CG13473', description='CG13473 gene product from transcript CG13473-RA', other_designations='CG13473-PA|uncharacterized protein', map_location='70F3-70F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14728760, end_position_on_the_genomic_accession=14729343, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39604, CurrentID=0, Status='live', Symbol='CG13471', Aliases='Dmel_ Dmel\\CG13471', description='CG13471 gene product from transcript CG13471-RA', other_designations='CG13471-PA|uncharacterized protein', map_location='70F3-70F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14732589, end_position_on_the_genomic_accession=14733517, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462750, CurrentID=0, Status='live', Symbol='CG44008', Aliases='Dmel_ CG31758, CG42325, Dmel\\CG44008', description='CG44008 gene product from transcript CG44008-RA', other_designations='CG44008-PA|CG44008-PB|uncharacterized protein', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11841502, end_position_on_the_genomic_accession=11842445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39608, CurrentID=0, Status='live', Symbol='CG9384', Aliases='Dmel_ Dmel\\CG9384, dMGAT4-1', description='CG9384 gene product from transcript CG9384-RB', other_designations='CG9384-PA|CG9384-PB', map_location='70F4-70F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14762028, end_position_on_the_genomic_accession=14767703, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39609, CurrentID=0, Status='live', Symbol='CG17173', Aliases='Dmel_ Dmel\\CG17173, dMGAT4-2', description='CG17173 gene product from transcript CG17173-RA', other_designations='CG17173-PA|uncharacterized protein', map_location='70F4-70F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14767703, end_position_on_the_genomic_accession=14769283, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39610, CurrentID=0, Status='live', Symbol='mop', Aliases='Dmel_CG9311, CG9311, Dmel\\CG9311, sfv', description='myopic', other_designations='CG9311-PA|mop-PA|slaughterhouse-five', map_location='70F4-70F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14769341, end_position_on_the_genomic_accession=14776129, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39611, CurrentID=0, Status='live', Symbol='bmm', Aliases='Dmel_CG5295, ATGL, Atgl, Bmm, CG5295, Dmel\\CG5295', description='brummer', other_designations='CG5295-PA|CG5295-PB|adipocyte triglyceride lipase|bmm-PA|bmm-PB|brummer lipase', map_location='70F5-70F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14776496, end_position_on_the_genomic_accession=14786412, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39612, CurrentID=0, Status='live', Symbol='Tdrd3', Aliases='Dmel_CG13472, CG13472, Dmel\\CG13472, dTDRD3', description='tudor domain containing 3 ortholog (H. sapiens)', other_designations='CG13472-PA|Tdrd3-PA|tudor domain containing 3 ortholog', map_location='70F5-70F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14786791, end_position_on_the_genomic_accession=14789665, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39613, CurrentID=0, Status='live', Symbol='CG9425', Aliases='Dmel_ Dmel\\CG9425', description='CG9425 gene product from transcript CG9425-RB', other_designations='CG9425-PB|CG9425-PC|CG9425-PD', map_location='70F6-70F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14789935, end_position_on_the_genomic_accession=14801089, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39615, CurrentID=0, Status='live', Symbol='gnu', Aliases='Dmel_CG5272, CG5272, Dmel\\CG5272, Gnu', description='giant nuclei', other_designations='CG5272-PA|gnu-PA', map_location='3-42.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14802380, end_position_on_the_genomic_accession=14803360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39616, CurrentID=0, Status='live', Symbol='CG17839', Aliases='Dmel_ 17839, CT39634, Dmel\\CG17839', description='CG17839 gene product from transcript CG17839-RB', other_designations='CG17839-PB|CG17839-PC|uncharacterized protein', map_location='70F7-71A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14805136, end_position_on_the_genomic_accession=14882224, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  39617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39617, CurrentID=0, Status='live', Symbol='DCX-EMAP', Aliases='Dmel_CG42247, CG13466, CG13467, CG42247, Dmel\\CG42247, Dmel_CG13466, Dmel_CG13467, anon-71Aa, dmEMAP', description='Doublecortin-domain-containing echinoderm-microtubule-associated protein ortholog', other_designations='CG42247-PC|CG42247-PD|CG42247-PF|CG42247-PG|DCX-EMAP-PC|DCX-EMAP-PD|DCX-EMAP-PF|DCX-EMAP-PG|doublecortin|doublecortin-domain-containing echinoderm-microtubule-associated protein', map_location='71A2-71A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14903805, end_position_on_the_genomic_accession=14942907, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  39619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39619, CurrentID=0, Status='live', Symbol='Tom', Aliases='Dmel_CG5185, Bbu, CG5185, DmelCG5185, Dmel\\CG5185, anon-EST:fe1F6, barbu, bbu/tom, tom', description='Twin of m4', other_designations='CG5185-PA|Tom-PA|anon-fast-evolving-1F6|barbu|twin of m4', map_location='71A3-71A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14969344, end_position_on_the_genomic_accession=14970306, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39620, CurrentID=0, Status='live', Symbol='Brd', Aliases='Dmel_CG3096-C, CG3096, DmelCG3096, Dmel\\CG3096, brd, Brd', description='Bearded', other_designations='Brd-PA|CG3096-PA|bearded', map_location='3-42 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14972667, end_position_on_the_genomic_accession=14973198, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39621, CurrentID=0, Status='live', Symbol='Ocho', Aliases='Dmel_CG3396, CG3396, CG5138, DmelCG3396, Dmel\\CG3396, ocho', description='CG3396 gene product from transcript CG3396-RB', other_designations='CG3396-PB|CG3396-PC|Ocho-PB|Ocho-PC|ocho', map_location='71A4-71A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14974848, end_position_on_the_genomic_accession=14976040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39623, CurrentID=0, Status='live', Symbol='CG3349', Aliases='Dmel_ Dmel\\CG3349', description='CG3349 gene product from transcript CG3349-RB', other_designations='CG3349-PB', map_location='71A4-71A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14979620, end_position_on_the_genomic_accession=14982432, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39624, CurrentID=0, Status='live', Symbol='CG5114', Aliases='Dmel_ Dmel\\CG5114', description='CG5114 gene product from transcript CG5114-RB', other_designations='CG5114-PA|CG5114-PB', map_location='71A4-71A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14982835, end_position_on_the_genomic_accession=14984486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39625, CurrentID=0, Status='live', Symbol='mnd', Aliases='Dmel_CG3297, CG3297, Dmel\\CG3297, SG29, l(3)B5, l(3)SG29, l(3)ds-5, l(3)ds5, l(3)o52, md', description='minidiscs', other_designations='CG3297-PA|CG3297-PC|CG3297-PD|minidisks|mnd-PA|mnd-PC|mnd-PD', map_location='3-47.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14985100, end_position_on_the_genomic_accession=14991667, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39626, CurrentID=0, Status='live', Symbol='Zip71B', Aliases='Dmel_CG10006, CG10006, Dmel\\CG10006, cg10006, dZip71B', description='Zinc/iron regulated transporter-related protein 71B', other_designations='CG10006-PB|CG10006-PC|Zip71B-PB|Zip71B-PC', map_location='71B1-71B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14992289, end_position_on_the_genomic_accession=14997983, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39627, CurrentID=0, Status='live', Symbol='mRpL39', Aliases='Dmel_CG17166, CG17166, Dmel\\CG17166, MRP-L5, mRpL5', description='mitochondrial ribosomal protein L39', other_designations='CG17166-PA|mRpL39-PA|mitochondrial ribosomal protein L5', map_location='71B1-71B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14997981, end_position_on_the_genomic_accession=14999666, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39628, CurrentID=0, Status='live', Symbol='Prosbeta2', Aliases='Dmel_CG3329, 3329, Beta-2, CG3329, DTS-7, DTS7, Dmel\\CG3329, Dts7, PROSbeta2, Prosbeta, Prosbeta 2, anon-EST:fe1E6, anon-EST:fe1G6, b2, beta 2, beta-2, beta2, beta2_dm, l(3)DTS7, prosbeta, prosbeta2', description='Proteasome beta2 subunit', other_designations='20s proteasome|CG3329-PA|Prosbeta2-PA|anon-fast-evolving-1E6|anon-fast-evolving-1G6|b2 subunit|beta2|beta2 proteosome subunit|beta2 subunit|proteasome beta2 subunit', map_location='71B1-71B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14999951, end_position_on_the_genomic_accession=15001319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39629, CurrentID=0, Status='live', Symbol='Reck', Aliases='Dmel_CG5392, CG5392, Dmel\\CG5392, RECK', description='Reversion-inducing-cysteine-rich protein with kazal motifs', other_designations='CG5392-PA|CG5392-PB|Reck-PA|Reck-PB|reversion-inducing-cysteine-rich protein with kazal motifs', map_location='71B1-71B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15001238, end_position_on_the_genomic_accession=15014062, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39630, CurrentID=0, Status='live', Symbol='Sytbeta', Aliases='Dmel_CG42333, CG13462, CG3020, CG42333, Dmel\\CG42333, Dmel_CG13462, Dmel_CG3020, SYT-beta, Syt V, Syt beta, Syt13, SytV, dsytbeta, syt13, sytbeta', description='Synaptotagmin beta', other_designations='CG42333-PC|CG42333-PD|Sytbeta-PC|Sytbeta-PD|synaptotagmin V|synaptotagmin beta|sytbeta', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15014695, end_position_on_the_genomic_accession=15064977, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  3772265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772265, CurrentID=0, Status='live', Symbol='His2B:CG33868', Aliases='Dmel_CG33868, CG33868, Dmel\\CG33868', description='CG33868 gene product from transcript CG33868-RA', other_designations='CG33868-PA|His2B:CG33868-PA|histone H2B', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21537505, end_position_on_the_genomic_accession=21538099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39633, CurrentID=0, Status='live', Symbol='CG18581', Aliases='Dmel_ Dmel\\CG18581', description='CG18581 gene product from transcript CG18581-RA', other_designations='CG18581-PA|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15049550, end_position_on_the_genomic_accession=15050209, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39634, CurrentID=0, Status='live', Symbol='CG12310', Aliases='Dmel_ BcDNA:RE02926, Dmel\\CG12310, chr3L:15015907..15016009', description='CG12310 gene product from transcript CG12310-RA', other_designations='CG12310-PA|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15050263, end_position_on_the_genomic_accession=15050806, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39635, CurrentID=0, Status='live', Symbol='CG13461', Aliases='Dmel_ Dmel\\CG13461', description='CG13461 gene product from transcript CG13461-RA', other_designations='CG13461-PA|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15051931, end_position_on_the_genomic_accession=15052490, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39636, CurrentID=0, Status='live', Symbol='CG18649', Aliases='Dmel_ BcDNA:RE73239, Dmel\\CG18649', description='CG18649 gene product from transcript CG18649-RB', other_designations='CG18649-PA|CG18649-PB|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15053568, end_position_on_the_genomic_accession=15054135, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39637, CurrentID=0, Status='live', Symbol='CG13463', Aliases='Dmel_ BcDNA:RE10167, Dmel\\CG13463', description='CG13463 gene product from transcript CG13463-RA', other_designations='CG13463-PA|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15054777, end_position_on_the_genomic_accession=15055565, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39638, CurrentID=0, Status='live', Symbol='CG13460', Aliases='Dmel_ Dmel\\CG13460', description='CG13460 gene product from transcript CG13460-RB', other_designations='CG13460-PB|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15061029, end_position_on_the_genomic_accession=15061705, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39641, CurrentID=0, Status='live', Symbol='Or71a', Aliases='Dmel_CG17871, 71B.1, 71a, CG17871, DOR14, Dmel\\CG17871, OR71a, Or71c, or71a', description='Odorant receptor 71a', other_designations='CG17871-PB|CG17871-PD|Or71a-PB|Or71a-PD|odorant receptor 71a', map_location='71B3-71B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15076907, end_position_on_the_genomic_accession=15078447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39643, CurrentID=0, Status='live', Symbol='sstn', Aliases='Dmel_CG6945, CG6945, Dmel\\CG6945', description='stepping stone', other_designations='CG6945-PA|sstn-PA', map_location='71B4-71B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15090823, end_position_on_the_genomic_accession=15094719, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39644, CurrentID=0, Status='live', Symbol='Aats-gly', Aliases='Dmel_CG6778, Aat-gly, CG6778, Dmel\\CG6778, GARS, GRS, Gars, GlyRS, dGars, gars, team', description='Glycyl-tRNA synthetase', other_designations='Aats-gly-PA|Aats-gly-PB|Aats-gly-PC|CG6778-PA|CG6778-PB|CG6778-PC|glycyl-tRNA synthetase|terminal arbors missing', map_location='71B4-71B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15095142, end_position_on_the_genomic_accession=15097686, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39645, CurrentID=0, Status='live', Symbol='CTPsyn', Aliases='Dmel_CG45070, CG45070, CG6854, CTPS, Dmel\\CG45070, Dmel_CG6854', description='CTP synthase', other_designations='CG45070-PA|CG45070-PB|CG45070-PC|CTP synthetase|CTP-synthase|CTPsyn-PA|CTPsyn-PB|CTPsyn-PC|ctp synthase|cytidine synthase', map_location='71B4-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15097914, end_position_on_the_genomic_accession=15113003, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39646, CurrentID=0, Status='live', Symbol='CG13458', Aliases='Dmel_ Dmel\\CG13458', description='CG13458 gene product from transcript CG13458-RA', other_designations='CG13458-PA|uncharacterized protein', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15113557, end_position_on_the_genomic_accession=15116373, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39647, CurrentID=0, Status='live', Symbol='Cep135', Aliases='Dmel_CG17081, Bld-10, Bld10, Bld10p, CEP135, CG17081, D DCep135/Bld10, DmBLD10, Dmel\\CG17081, bld10, dbld10', description='CG17081 gene product from transcript CG17081-RB', other_designations='CG17081-PA|CG17081-PB|CG17081-PC|Cep135-PA|Cep135-PB|Cep135-PC', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15116737, end_position_on_the_genomic_accession=15128029, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  39648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39648, CurrentID=0, Status='live', Symbol='CG16959', Aliases='Dmel_ CT21630, Dmel\\CG16959', description='CG16959 gene product from transcript CG16959-RE', other_designations='CG16959-PA|CG16959-PB|CG16959-PC|CG16959-PD|CG16959-PE|uncharacterized protein', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15119878, end_position_on_the_genomic_accession=15125321, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39649, CurrentID=0, Status='live', Symbol='CG13457', Aliases='Dmel_ BcDNA:AT15860, BcDNA:AT16542, Dmel\\CG13457', description='CG13457 gene product from transcript CG13457-RA', other_designations='CG13457-PA|CG13457-PB|uncharacterized protein', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15127940, end_position_on_the_genomic_accession=15131378, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39650, CurrentID=0, Status='live', Symbol='CG12316', Aliases='Dmel_ Dmel\\CG12316', description='CG12316 gene product from transcript CG12316-RA', other_designations='CG12316-PA|CG12316-PB|uncharacterized protein', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15131508, end_position_on_the_genomic_accession=15136934, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39651, CurrentID=0, Status='live', Symbol='Pdi', Aliases='Dmel_CG6988, CG6988, Dmel\\CG6988, Dpdi, EP(3)1022, PDI, dpdi, pdi', description='Protein disulfide isomerase', other_designations='CG6988-PA|CG6988-PE|Pdi-PA|Pdi-PE|disulfide isomerase|proline diisomerease|protein disulfide isomerase|protein disulphide isomerase|protein-disulfide isomerase', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15137738, end_position_on_the_genomic_accession=15142840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39652, CurrentID=0, Status='live', Symbol='Pex3', Aliases='Dmel_CG6859, CG6859, DmelPex3, Dmel\\CG6859, PEX3, pex3', description='Peroxin 3', other_designations='CG6859-PA|CG6859-PB|Pex3-PA|Pex3-PB|peroxin 3|peroxisomal multifunctional enzyme type 2', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15144355, end_position_on_the_genomic_accession=15146379, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39653, CurrentID=0, Status='live', Symbol='FucTA', Aliases='Dmel_CG6869, 3-FucT, CG6869, Dm alpha1, Dmel\\CG6869, anon-EST:Posey285, fucTA', description='CG6869 gene product from transcript CG6869-RA', other_designations='CG6869-PA|CG6869-PB|FucTA-PA|FucTA-PB|alpha1,3-fucosyltransferase|core alpha1,3-fucosyltransferase|core fucosyltransferase', map_location='71B5-71B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15147137, end_position_on_the_genomic_accession=15154395, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39654, CurrentID=0, Status='live', Symbol='Msh6', Aliases='Dmel_CG7003, CG7003, Dmel\\CG7003', description='CG7003 gene product from transcript CG7003-RA', other_designations='CG7003-PA|Msh6-PA', map_location='71B6-71B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15154233, end_position_on_the_genomic_accession=15158562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39655, CurrentID=0, Status='live', Symbol='Prp31', Aliases='Dmel_CG6876, CG6876, Dmel\\CG6876, FBgn0036487, anon-EST:Liang-2.38, clone 2.38', description='CG6876 gene product from transcript CG6876-RA', other_designations='CG6876-PA|Prp31-PA', map_location='71B6-71B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15159100, end_position_on_the_genomic_accession=15161138, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39656, CurrentID=0, Status='live', Symbol='CG6878', Aliases='Dmel_ Dmel\\CG6878, Romo1', description='CG6878 gene product from transcript CG6878-RA', other_designations='CG6878-PA', map_location='71B6-71B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15161817, end_position_on_the_genomic_accession=15162470, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39657, CurrentID=0, Status='live', Symbol='CG7011', Aliases='Dmel_ Dmel\\CG7011', description='CG7011 gene product from transcript CG7011-RA', other_designations='CG7011-PA', map_location='71B6-71B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15160876, end_position_on_the_genomic_accession=15163628, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39658, CurrentID=0, Status='live', Symbol='CG6888', Aliases='Dmel_ 6888, BcDNA:AT16346, Dmel\\CG6888', description='CG6888 gene product from transcript CG6888-RA', other_designations='CG6888-PA', map_location='71B8-71B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15207844, end_position_on_the_genomic_accession=15208586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39659, CurrentID=0, Status='live', Symbol='Tak1', Aliases='Dmel_CG18492, CG1388, CG18492, D-TAK, D-tak, D10, DTAK1, DTak, DTak1, DmTak1, Dmel\\CG18492, TAK, TAK-1, TAK1, Tak, Tak 1, dTAK, dTAK-1, dTAK1, dTak, dTak1, dtak1, tak, tak1', description='TGF-beta activated kinase 1', other_designations='CG18492-PA|CG18492-PB|TGF beta-activated kinase 1|TGF-beta activated-kinase 1|TGFbeta-activated kinase 1|Tak1-PA|Tak1-PB|transforming growth factor beta-activated kinase|transforming growth factor-beta activated kinase 1|transforming growth factor-beta-activated kinase 1', map_location='19D2-19D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20515908, end_position_on_the_genomic_accession=20524928, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39660, CurrentID=0, Status='live', Symbol='Best4', Aliases='Dmel_CG7259, CG7259, CT22395, Dmel\\CG7259, dBest4, dbest4', description='Bestrophin 4', other_designations='Best4-PB|CG7259-PB|bestrophin 4', map_location='71C2-71C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15298958, end_position_on_the_genomic_accession=15300942, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39661, CurrentID=0, Status='live', Symbol='Best3', Aliases='Dmel_CG12327, CG12327, Dmel\\CG12327, dBest3, dbest3', description='Bestrophin 3', other_designations='Best3-PA|CG12327-PA|bestrophin 3', map_location='71C2-71C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15300915, end_position_on_the_genomic_accession=15302785, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39662, CurrentID=0, Status='live', Symbol='CG7255', Aliases='Dmel_ Dmel\\CG7255', description='CG7255 gene product from transcript CG7255-RE', other_designations='CG7255-PE|CG7255-PF|CG7255-PG|CG7255-PH|CG7255-PI', map_location='71C2-71C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15304255, end_position_on_the_genomic_accession=15311195, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  39663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39663, CurrentID=0, Status='live', Symbol='Toll-6', Aliases='Dmel_CG7250, 18w-like, CG7250, CT22359, Dmel\\CG7250, TOLL 6, Tl-6, Toll 6, dTLR6, dToll6, toll, toll-6', description='CG7250 gene product from transcript CG7250-RC', other_designations='CG7250-PA|CG7250-PB|CG7250-PC|Toll-6-PA|Toll-6-PB|Toll-6-PC|toll receptor', map_location='71C2-71C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15336692, end_position_on_the_genomic_accession=15344315, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39665, CurrentID=0, Status='live', Symbol='CG7804', Aliases='Dmel_ Dmel\\CG7804, cg7804', description='CG7804 gene product from transcript CG7804-RA', other_designations='CG7804-PA|CG7804-PB', map_location='71D2-71D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15472197, end_position_on_the_genomic_accession=15474110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39666, CurrentID=0, Status='live', Symbol='Ran-like', Aliases='Dmel_CG7815, BcDNA:GH25818, CG7815, Dmel\\CG7815, anon-WO0140519.150, ran-like', description='CG7815 gene product from transcript CG7815-RA', other_designations='CG7815-PA|Ran-like-PA', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15484396, end_position_on_the_genomic_accession=15485402, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39667, CurrentID=0, Status='live', Symbol='CG13455', Aliases='Dmel_ Dmel\\CG13455', description='CG13455 gene product from transcript CG13455-RA', other_designations='CG13455-PA|uncharacterized protein', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15496574, end_position_on_the_genomic_accession=15498055, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39668, CurrentID=0, Status='live', Symbol='CG7276', Aliases='Dmel_ Dmel\\CG7276', description='CG7276 gene product from transcript CG7276-RB', other_designations='CG7276-PB', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15498202, end_position_on_the_genomic_accession=15498821, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39669, CurrentID=0, Status='live', Symbol='CG7275', Aliases='Dmel_ Dmel\\CG7275', description='CG7275 gene product from transcript CG7275-RA', other_designations='CG7275-PA', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15498959, end_position_on_the_genomic_accession=15500561, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39670, CurrentID=0, Status='live', Symbol='CG7272', Aliases='Dmel_ Dmel\\CG7272', description='CG7272 gene product from transcript CG7272-RA', other_designations='CG7272-PA', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15500938, end_position_on_the_genomic_accession=15503769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39671, CurrentID=0, Status='live', Symbol='CG7857', Aliases='Dmel_ Dmel\\CG7857, Q9VUN9, anon-Becker1', description='CG7857 gene product from transcript CG7857-RA', other_designations='CG7857-PA', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15503600, end_position_on_the_genomic_accession=15504894, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39672, CurrentID=0, Status='live', Symbol='CG7841', Aliases='Dmel_ Dmel\\CG7841', description='CG7841 gene product from transcript CG7841-RB', other_designations='CG7841-PA|CG7841-PB', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15505188, end_position_on_the_genomic_accession=15508252, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39673, CurrentID=0, Status='live', Symbol='Z600', Aliases='Dmel_CG17962, 3.2, CG17962, DMZ60MEX, Dmel\\CG17962, Frs, frs, z600', description='CG17962 gene product from transcript CG17962-RA', other_designations='CG17962-PA|Z600-PA|fruhstart|frustart', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15509061, end_position_on_the_genomic_accession=15509513, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39675, CurrentID=0, Status='live', Symbol='Eip71CD', Aliases='Dmel_CG7266, CG7266, DROZ60MEX, Dmel\\CG7266, Eip28, Eip28 /29, Eip28/29, Eip71C, MsrA, dmsrA, eip28/29', description='Ecdysone-induced protein 28/29kD', other_designations='CG7266-PA|CG7266-PB|CG7266-PC|CG7266-PE|CG7266-PF|CG7266-PG|Eip28/29|Eip71CD-PA|Eip71CD-PB|Eip71CD-PC|Eip71CD-PE|Eip71CD-PF|Eip71CD-PG|Methionine-S-sulfoxide reductase|methionine reductase A|methionine sulfoxide reductase A', map_location='71D3-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15511052, end_position_on_the_genomic_accession=15513202, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39676, CurrentID=0, Status='live', Symbol='CG13454', Aliases='Dmel_ Dmel\\CG13454', description='CG13454 gene product from transcript CG13454-RA', other_designations='CG13454-PA|uncharacterized protein', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15516533, end_position_on_the_genomic_accession=15517056, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39677, CurrentID=0, Status='live', Symbol='mex1', Aliases='Dmel_CG7936, CG7936, Dmel\\CG7936', description='midgut expression 1', other_designations='CG7936-PA|CG7936-PB|CG7936-PC|mex1-PA|mex1-PB|mex1-PC', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15517424, end_position_on_the_genomic_accession=15519592, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39678, CurrentID=0, Status='live', Symbol='yellow-k', Aliases='Dmel_CG7463, CG7463, Dmel\\CG7463, yell-k', description='CG7463 gene product from transcript CG7463-RA', other_designations='CG7463-PA|yellow-k-PA', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15520620, end_position_on_the_genomic_accession=15521653, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39679, CurrentID=0, Status='live', Symbol='CG7945', Aliases='Dmel_ CG17014, Dmel\\CG7945', description='CG7945 gene product from transcript CG7945-RA', other_designations='CG7945-PA|CG7945-PB', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15521711, end_position_on_the_genomic_accession=15524137, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39681, CurrentID=0, Status='live', Symbol='obst-H', Aliases='Dmel_CG33983, CG17014, CG33983, CG7973, CG7973(short), Dmel\\CG33983, FBgn0036507', description='obstructor-H', other_designations='CG33983-PA|obst-H-PA', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15528265, end_position_on_the_genomic_accession=15529243, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39682, CurrentID=0, Status='live', Symbol='CrebA', Aliases='Dmel_CG7450, BBF-2, BBF2_DROME, BOX B Binding Factor-2, Bbbf2, BcDNA:SD05937, CG7450, CRE-BP, CREB, CREB-A, CREB-a, CREBA, Creb, Creb A, Creb-A, DM4, Dmel\\CG7450, creb, crebA, dCREB-A, dCREBA, dCreb-A, dCrebA, l(3)03576', description='Cyclic-AMP response element binding protein A', other_designations='Box B-binding factor 2|CG7450-PA|CG7450-PB|CrebA-PA|CrebA-PB|cAMP response element binding protein|cAMP response element binding protein A|cyclic AMP response element binding protein|dCREB-A', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15535395, end_position_on_the_genomic_accession=15553214, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39683, CurrentID=0, Status='live', Symbol='AGO2', Aliases='Dmel_CG7439, AG02, AGO 2, Ago-2, Ago2, CG13452, CG7439, Dm Ago2, DmAgo2, Dmel\\CG7439, ago, ago-2, ago2, dAGO2, dAgo2, eIF2C2', description='Argonaute 2', other_designations='AGO2-PB|AGO2-PC|AGO2-PE|ARGONAUTE2|CG7439-PB|CG7439-PC|CG7439-PE|argonaute|argonaute 2|argonaute-2', map_location='71C3-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15554113, end_position_on_the_genomic_accession=15561042, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39684, CurrentID=0, Status='live', Symbol='CG7739', Aliases='Dmel_ Dmel\\CG7739', description='CG7739 gene product from transcript CG7739-RA', other_designations='CG7739-PA', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15560726, end_position_on_the_genomic_accession=15563065, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39685, CurrentID=0, Status='live', Symbol='CG7427', Aliases='Dmel_ Dmel\\CG7427', description='CG7427 gene product from transcript CG7427-RA', other_designations='CG7427-PA|CG7427-PB|CG7427-PC|CG7427-PD|CG7427-PE', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15563427, end_position_on_the_genomic_accession=15565559, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39686, CurrentID=0, Status='live', Symbol='dop', Aliases='Dmel_CG6498, CG6498, Dmel\\CG6498, MAST 205, MAST205', description='drop out', other_designations='CG6498-PA|CG6498-PB|dop-PA|dop-PB', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15566013, end_position_on_the_genomic_accession=15576313, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39687, CurrentID=0, Status='live', Symbol='CG16979', Aliases='Dmel_ Dmel\\CG16979', description='CG16979 gene product from transcript CG16979-RA', other_designations='CG16979-PA|uncharacterized protein', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15576570, end_position_on_the_genomic_accession=15578674, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39688, CurrentID=0, Status='live', Symbol='mrn', Aliases='Dmel_CG7764, 154538_at, CG7764, Dmel\\CG7764, Dmp52, TFB2, Tfb2, anon-WO0118547.648, l(3)E35, ms(3)nc16, nc16, p52', description='marionette', other_designations='CG7764-PA|Tfb2|marionnette|mrn-PA', map_location='3-42.6 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15578769, end_position_on_the_genomic_accession=15580537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39689, CurrentID=0, Status='live', Symbol='CG12301', Aliases='Dmel_ Dmel\\CG12301', description='CG12301 gene product from transcript CG12301-RA', other_designations='CG12301-PA|uncharacterized protein', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15580800, end_position_on_the_genomic_accession=15583373, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39690, CurrentID=0, Status='live', Symbol='CG12304', Aliases='Dmel_ Dmel\\CG12304, p38', description='CG12304 gene product from transcript CG12304-RA', other_designations='CG12304-PA|CG12304-PB|uncharacterized protein', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15583638, end_position_on_the_genomic_accession=15585347, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39691, CurrentID=0, Status='live', Symbol='CG7656', Aliases='Dmel_ Dmel\\CG7656', description='CG7656 gene product from transcript CG7656-RF', other_designations='CG7656-PA|CG7656-PD|CG7656-PE|CG7656-PF', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15585431, end_position_on_the_genomic_accession=15588175, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39693, CurrentID=0, Status='live', Symbol='RhoGAP71E', Aliases='Dmel_CG32149, BcDNA:LD07615, CG16980, CG32149, CG7396, Dmel\\CG32149, RhoGAP-71E1, RhoGap71E, l(3)j6B9', description='Rho GTPase activating protein at 71E', other_designations='CG32149-PA|CG32149-PB|CG32149-PC|CG32149-PD|CG32149-PE|CG32149-PF|CG32149-PG|RhoGAP71E-PA|RhoGAP71E-PB|RhoGAP71E-PC|RhoGAP71E-PD|RhoGAP71E-PE|RhoGAP71E-PF|RhoGAP71E-PG', map_location='71E1-71E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15588860, end_position_on_the_genomic_accession=15609125, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39694, CurrentID=0, Status='live', Symbol='CG7650', Aliases='Dmel_ BcDNA:AT27057, Dmel\\CG7650', description='CG7650 gene product from transcript CG7650-RA', other_designations='CG7650-PA', map_location='71E2-71E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15608707, end_position_on_the_genomic_accession=15610386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39695, CurrentID=0, Status='live', Symbol='CG13449', Aliases='Dmel_ Dmel\\CG13449', description='CG13449 gene product from transcript CG13449-RA', other_designations='CG13449-PA|CG13449-PB|uncharacterized protein', map_location='71E2-71E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15610631, end_position_on_the_genomic_accession=15612828, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39696, CurrentID=0, Status='live', Symbol='comm3', Aliases='Dmel_CG42334, CG13446, CG13447, CG13448, CG33209, CG42334, Dmel\\CG42334, Dmel_CG13446, Dmel_CG33209', description='CG42334 gene product from transcript CG42334-RC', other_designations='CG42334-PC|CG42334-PD|CG42334-PE|comm3-PC|comm3-PD|comm3-PE', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15613184, end_position_on_the_genomic_accession=15647762, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39697, CurrentID=0, Status='live', Symbol='CG7372', Aliases='Dmel_ Dmel\\CG7372', description='CG7372 gene product from transcript CG7372-RA', other_designations='CG7372-PA', map_location='71E3-71E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15619901, end_position_on_the_genomic_accession=15624130, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39700, CurrentID=0, Status='live', Symbol='Eig71Ea', Aliases='Dmel_CG16931, CG16931, Dmel\\CG16931, I, L71, L71-1, gene I, gene-I', description='Ecdysone-induced gene 71Ea', other_designations='71E-Gene-I|CG16931-PA|Eig71Ea-PA|gene I', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15649126, end_position_on_the_genomic_accession=15649661, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39701, CurrentID=0, Status='live', Symbol='Eig71Eb', Aliases='Dmel_CG7355, CG7355, Dmel\\CG7355, II, L71, L71-2, gene II, gene-II', description='Ecdysone-induced gene 71Eb', other_designations='71E-Gene-II|CG7355-PA|CG7355-PB|Eig71Eb-PA|Eig71Eb-PB|gene II', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15649936, end_position_on_the_genomic_accession=15650541, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39703, CurrentID=0, Status='live', Symbol='Eig71Ec', Aliases='Dmel_CG7608, CG7608, Dmel\\CG7608, Eig 71Ec, III, L71, L71-3, LP11175p, gene III, gene-III', description='Ecdysone-induced gene 71Ec', other_designations='71E-Gene-III|CG7608-PA|Eig71Ec-PA|gene III', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15651479, end_position_on_the_genomic_accession=15652254, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39704, CurrentID=0, Status='live', Symbol='Eig71Ed', Aliases='Dmel_CG7350, BcDNA:LP04160, CG7350, Dmel\\CG7350, Eig 71Ed, IV, L71, L71-4, gene IV, gene-IV', description='Ecdysone-induced gene 71Ed', other_designations='71E-Gene-IV|CG7350-PA|Eig71Ed-PA|gene IV', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15652537, end_position_on_the_genomic_accession=15653096, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39706, CurrentID=0, Status='live', Symbol='Eig71Ee', Aliases='Dmel_CG7604, CG7604, Dmel\\CG7604, I71-7, L71, L71-7, VII, eig71Ee, gene VII, gene-VII, gp150', description='Ecdysone-induced gene 71Ee', other_designations='71E-Gene-VII|CG7604-PA|Eig71Ee-PA|gene VII', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15654472, end_position_on_the_genomic_accession=15656010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39707, CurrentID=0, Status='live', Symbol='Eig71Ef', Aliases='Dmel_CG7599, CG7599, Dmel\\CG7599, EIG71-EF, L71, L71-5, V, gene V, gene-V', description='Ecdysone-induced gene 71Ef', other_designations='71E-Gene-V|CG7599-PA|Eig71Ef-PA|gene V', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15656655, end_position_on_the_genomic_accession=15657390, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39708, CurrentID=0, Status='live', Symbol='Eig71Eg', Aliases='Dmel_CG7336, CG7336, Dmel\\CG7336, L71, L71-6, VI, eig71Eg, gene VI, gene-VI', description='Ecdysone-induced gene 71Eg', other_designations='71E-Gene-VI|CG7336-PA|Eig71Eg-PA|gene VI', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15657628, end_position_on_the_genomic_accession=15658149, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39709, CurrentID=0, Status='live', Symbol='Eig71Eh', Aliases='Dmel_CG7594, CG7594, Dmel\\CG7594, EIG71-H, L71, L71-8', description='CG7594 gene product from transcript CG7594-RB', other_designations='CG7594-PB|Eig71Eh-PB', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15658936, end_position_on_the_genomic_accession=15659441, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39710, CurrentID=0, Status='live', Symbol='Eig71Ei', Aliases='Dmel_CG7327, CG7327, Dmel\\CG7327, EIG71-EI, L71, L71-9', description='CG7327 gene product from transcript CG7327-RB', other_designations='CG7327-PA|CG7327-PB|Eig71Ei-PA|Eig71Ei-PB', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15659707, end_position_on_the_genomic_accession=15660351, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39711, CurrentID=0, Status='live', Symbol='Eig71Ej', Aliases='Dmel_CG7588, CG7588, Dmel\\CG7588, L71, L71-10', description='CG7588 gene product from transcript CG7588-RA', other_designations='CG7588-PA|Eig71Ej-PA', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15660549, end_position_on_the_genomic_accession=15661015, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39712, CurrentID=0, Status='live', Symbol='Eig71Ek', Aliases='Dmel_CG7325, CG7325, Dmel\\CG7325, EIG71-K, L71, L71-11', description='CG7325 gene product from transcript CG7325-RA', other_designations='CG7325-PA|Eig71Ek-PA', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15661248, end_position_on_the_genomic_accession=15661809, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39713, CurrentID=0, Status='live', Symbol='CG7304', Aliases='Dmel_ Dmel\\CG7304, dppGalNAcT11', description='CG7304 gene product from transcript CG7304-RB', other_designations='CG7304-PB', map_location='71F1-71F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15679720, end_position_on_the_genomic_accession=15681470, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39714, CurrentID=0, Status='live', Symbol='CG7579', Aliases='Dmel_ Dmel\\CG7579, dppGalNAcT12', description='CG7579 gene product from transcript CG7579-RA', other_designations='CG7579-PA', map_location='71F1-71F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15681633, end_position_on_the_genomic_accession=15683445, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39715, CurrentID=0, Status='live', Symbol='pgant8', Aliases='Dmel_CG7297, CG7297, Dmel\\CG7297', description='polypeptide GalNAc transferase 8', other_designations='CG7297-PA|pgant8-PA', map_location='71F1-71F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15683872, end_position_on_the_genomic_accession=15686598, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39716, CurrentID=0, Status='live', Symbol='comm2', Aliases='Dmel_CG7554, CG7554, Dmel\\CG7554', description='CG7554 gene product from transcript CG7554-RA', other_designations='CG7554-PA|comm2-PA', map_location='71F1-71F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15697594, end_position_on_the_genomic_accession=15699850, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39717, CurrentID=0, Status='live', Symbol='comm', Aliases='Dmel_CG17943, CG17943, Comm, Dmel\\CG17943', description='commissureless', other_designations='CG17943-PA|CG17943-PB|comm-PA|comm-PB|commisureless', map_location='71F2-71F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15721511, end_position_on_the_genomic_accession=15728430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39719, CurrentID=0, Status='live', Symbol='CG13445', Aliases='Dmel_ Dmel\\CG13445', description='CG13445 gene product from transcript CG13445-RA', other_designations='CG13445-PA|uncharacterized protein', map_location='72A1-72A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15809485, end_position_on_the_genomic_accession=15809977, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39720, CurrentID=0, Status='live', Symbol='fwe', Aliases='Dmel_CG6151, CG6151, Dmel\\CG6151, anon-WO0118547.242, dFwe', description='flower', other_designations='CG6151-PA|CG6151-PB|CG6151-PC|fwe-PA|fwe-PB|fwe-PC', map_location='72A1-72A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15810081, end_position_on_the_genomic_accession=15816410, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39721, CurrentID=0, Status='live', Symbol='CkIIalpha-i1', Aliases='Dmel_CG6215, CG6215, CKIIalpha-I1, CKIIalpha-iI, DmA35, Dmel\\CG6215, ZFP47, dCKIIalpha-i1', description='CKII-alpha subunit interactor-1', other_designations='CG6215-PA|CkIIalpha-i1-PA', map_location='72A1-72A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15816979, end_position_on_the_genomic_accession=15818621, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39722, CurrentID=0, Status='live', Symbol='DCP2', Aliases='Dmel_CG6169, BEST:LD07107, CG6169, DCp2p, Dcp2, Dmel\\CG6169, LD07107, dDCP2, dDcp, dDcp2, dcp2', description='Decapping protein 2', other_designations='CG6169-PA|CG6169-PB|CG6169-PD|CG6169-PE|DCP2-PA|DCP2-PB|DCP2-PD|DCP2-PE|decapping protein 2', map_location='72A1-72A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15818734, end_position_on_the_genomic_accession=15826423, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19835885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835885, CurrentID=0, Status='live', Symbol='CG44257', Aliases='Dmel_ Dmel\\CG44257', description='CG44257 gene product from transcript CG44257-RA', other_designations='CG44257-PA|uncharacterized protein', map_location='90A1-90A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17265278, end_position_on_the_genomic_accession=17265821, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39724, CurrentID=0, Status='live', Symbol='CG12713', Aliases='Dmel_ Dmel\\CG12713', description='CG12713 gene product from transcript CG12713-RA', other_designations='CG12713-PA|uncharacterized protein', map_location='72A2-72A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15832464, end_position_on_the_genomic_accession=15833947, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39725, CurrentID=0, Status='live', Symbol='CG18081', Aliases='Dmel_ BcDNA:RE63284, Dmel\\CG18081', description='CG18081 gene product from transcript CG18081-RB', other_designations='CG18081-PA|CG18081-PB|CG18081-PC|uncharacterized protein', map_location='72A2-72A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15833795, end_position_on_the_genomic_accession=15835388, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39726, CurrentID=0, Status='live', Symbol='CG15715', Aliases='Dmel_ BcDNA:RE26825, Dmel\\CG15715', description='CG15715 gene product from transcript CG15715-RB', other_designations='CG15715-PA|CG15715-PB|uncharacterized protein', map_location='72A2-72A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15835693, end_position_on_the_genomic_accession=15837482, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39727, CurrentID=0, Status='live', Symbol='CG32150', Aliases='Dmel_ CG12712, CG15714, Dmel\\CG32150', description='CG32150 gene product from transcript CG32150-RA', other_designations='CG32150-PA|CG32150-PC|CG32150-PD|uncharacterized protein', map_location='72A2-72A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15837773, end_position_on_the_genomic_accession=15844062, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39730, CurrentID=0, Status='live', Symbol='pHCl', Aliases='Dmel_CG44099, BP1024, BcDNA:GH21757, CG12486, CG13444, CG33989, CG44099, CG6112, CT19189, Dmel\\CG44099, Dmel_CG12486, Dmel_CG33989, pHCL-A', description='CG44099 gene product from transcript CG44099-RL', other_designations='CG44099-PC|CG44099-PD|CG44099-PE|CG44099-PF|CG44099-PG|CG44099-PH|CG44099-PI|CG44099-PJ|CG44099-PK|CG44099-PL|CG44099-PM|CG44099-PN|CG44099-PO|CG44099-PP|CG44099-PQ|CG44099-PR|glutamate-gated chloride channel-like|pHCl-PC|pHCl-PD|pHCl-PE|pHCl-PF|pHCl-PG|pHCl-PH|pHCl-PI|pHCl-PJ|pHCl-PK|pHCl-PL|pHCl-PM|pHCl-PN|pHCl-PO|pHCl-PP|pHCl-PQ|pHCl-PR', map_location='72A2-72A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15848578, end_position_on_the_genomic_accession=15899029, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  39732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39732, CurrentID=0, Status='live', Symbol='sff', Aliases='Dmel_CG6114, CG6114, Dmel\\CG6114', description='sugar-free frosting', other_designations='CG6114-PA|CG6114-PB|CG6114-PC|SNF1-related|sff-PA|sff-PB|sff-PC', map_location='72A4-72B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15900156, end_position_on_the_genomic_accession=15924861, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  39733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39733, CurrentID=0, Status='live', Symbol='Pka-C3', Aliases='Dmel_CG6117, 5-23, CG6117, DC2, Dmel\\CG6117, Dpck, PKA, PKA C3, PKA-C-3, PKA-C3, Pk72A, Pk?5, pka-C3', description='Protein kinase, cAMP-dependent, catalytic subunit 3', other_designations='CG6117-PA|CG6117-PB|Pka-C3-PA|Pka-C3-PB|Protein-kinase-like-72A|cAMP-activated protein kinase A|cAMP-dependent protein kinase|cAMP-dependent protein kinase 3|protein kinase, cAMP-dependent, catalytic subunit 3|swiss cheese', map_location='72B1-72B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15925805, end_position_on_the_genomic_accession=15954797, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39734, CurrentID=0, Status='live', Symbol='GXIVsPLA2', Aliases='Dmel_CG17035, CG17035, Dmel\\CG17035, GXIVsPLA', description='CG17035 gene product from transcript CG17035-RD', other_designations='CG17035-PA|CG17035-PC|CG17035-PD|GXIVsPLA2-PA|GXIVsPLA2-PC|GXIVsPLA2-PD', map_location='72B2-72B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15955096, end_position_on_the_genomic_accession=15956465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39735, CurrentID=0, Status='live', Symbol='elgi', Aliases='Dmel_CG17033, CG17033, Dmel\\CG17033', description='early girl', other_designations='CG17033-PA|CG17033-PB|elgi-PA|elgi-PB', map_location='72B2-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15956550, end_position_on_the_genomic_accession=15958433, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39736, CurrentID=0, Status='live', Symbol='CG17032', Aliases='Dmel_ Dmel\\CG17032', description='CG17032 gene product from transcript CG17032-RA', other_designations='CG17032-PA|CG17032-PB|uncharacterized protein', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15958843, end_position_on_the_genomic_accession=15960609, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39737, CurrentID=0, Status='live', Symbol='l(3)72Ab', Aliases='Dmel_CG5931, ASCC3L1, Brr2, CG5931, Dmel\\CG5931, anon-EST:Liang-1.21, brr2, cg5931, clone 1.21', description='lethal (3) 72Ab', other_designations='CG5931-PA|l(3)72Ab-PA', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15960836, end_position_on_the_genomic_accession=15967872, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39738, CurrentID=0, Status='live', Symbol='CG10516', Aliases='Dmel_ Dmel\\CG10516', description='CG10516 gene product from transcript CG10516-RA', other_designations='CG10516-PA|uncharacterized protein', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15968204, end_position_on_the_genomic_accession=15969902, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39739, CurrentID=0, Status='live', Symbol='CG17026', Aliases='Dmel_ Dmel\\CG17026, cg17026', description='CG17026 gene product from transcript CG17026-RA', other_designations='CG17026-PA|uncharacterized protein', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15975368, end_position_on_the_genomic_accession=15976428, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39740, CurrentID=0, Status='live', Symbol='CG17029', Aliases='Dmel_ Dmel\\CG17029, IMP', description='CG17029 gene product from transcript CG17029-RB', other_designations='CG17029-PA|CG17029-PB|uncharacterized protein', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15976870, end_position_on_the_genomic_accession=15978435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39741, CurrentID=0, Status='live', Symbol='CG17028', Aliases='Dmel_ Dmel\\CG17028, IMP', description='CG17028 gene product from transcript CG17028-RA', other_designations='CG17028-PA|CG17028-PB|uncharacterized protein', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15978524, end_position_on_the_genomic_accession=15980004, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39742, CurrentID=0, Status='live', Symbol='CG17027', Aliases='Dmel_ 152392_at, Dmel\\CG17027, IMP', description='CG17027 gene product from transcript CG17027-RB', other_designations='CG17027-PA|CG17027-PB|uncharacterized protein', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15980100, end_position_on_the_genomic_accession=15981888, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39744, CurrentID=0, Status='live', Symbol='brm', Aliases='Dmel_CG5942, BRD, BRM, Brm, Brm @ Kto, CG18438, CG5942, Dmel\\CG5942, E(E2F)3A, dBRM, dSNF2, l(3)72Aa', description='brahma', other_designations='CG5942-PA|CG5942-PB|CG5942-PC|CG5942-PD|CG5942-PE|CG5942-PF|brm-PA|brm-PB|brm-PC|brm-PD|brm-PE|brm-PF|complementation group 3.5', map_location='3-43.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15970082, end_position_on_the_genomic_accession=15982869, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39745, CurrentID=0, Status='live', Symbol='Arl1', Aliases='Dmel_CG6025, 72CDd, ARL1, ARL1_DROME, Arf, Arf72A, BcDNA:GM20805, CG 6025, CG6025, Dmel\\CG6025, P25160, arf72A, arf[72A], arl, arl1, dARL1p, dArl1, l(3)72Ae, l(3)72CDd', description='ADP ribosylation factor-like 1', other_designations='Arf-like 1|Arf72A|Arl1-PA|CG6025-PA|arflike at 72A', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15983135, end_position_on_the_genomic_accession=15984362, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39746, CurrentID=0, Status='live', Symbol='DNApol-delta', Aliases='Dmel_CG5949, 5949, CAA61369, CG5949, DNA Pol Delta, DNA pol delta, Dm Poldelta, Dmel\\CG5949, Poldelta, dnapol delta, l(3)72Ac', description='DNA-polymerase-delta', other_designations='CG5949-PA|DNA polymerase delta|DNA polymerase-delta|DNApol-delta-PA|lethal (3) 72Ac', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15984322, end_position_on_the_genomic_accession=15987815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39747, CurrentID=0, Status='live', Symbol='Hip14', Aliases='Dmel_CG6017, 3L1, CG 6017, CG6017, Dmel\\CG6017, dHIP14, hip14, l(3)72Ad', description='Huntingtin-interacting protein 14', other_designations='CG6017-PA|CG6017-PB|CG6017-PC|Hip14-PA|Hip14-PB|Hip14-PC|ankyrin domain|htt-interacting protein 14|lethal (3) 72Ad', map_location='72C1-72C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15988236, end_position_on_the_genomic_accession=15992464, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39748, CurrentID=0, Status='live', Symbol='CG5830', Aliases='Dmel_ 24664944, Dmel\\CG5830', description='CG5830 gene product from transcript CG5830-RA', other_designations='CG5830-PA|CG5830-PB', map_location='72C1-72C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15992669, end_position_on_the_genomic_accession=16004736, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39749, CurrentID=0, Status='live', Symbol='mRpS31', Aliases='Dmel_CG5904, CG5904, Dmel\\CG5904, MRP-S31, RpS31', description='mitochondrial ribosomal protein S31', other_designations='CG5904-PA|mRpS31-PA', map_location='72C2-72C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16005193, end_position_on_the_genomic_accession=16006599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39750, CurrentID=0, Status='live', Symbol='mib1', Aliases='Dmel_CG5841, CG5841, D-mib, D-mib1, Dmel\\CG5841, Dmib, MIB1, Mib, Mib1, dmib1, l(3)72CDa, l(3)72Db, mib/2, mind-bomb, mib1', description='mind bomb 1', other_designations='CG5841-PA|CG5841-PB|Mindbomb1|lethal (3) 72CDa|lethal (3) 72Db|mib1-PA|mib1-PB|mind bomb|mind bomb-1|mind bomb1|mind-bomb|mindbomb|mindbomb 1', map_location='72C2-72C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16006552, end_position_on_the_genomic_accession=16011812, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39751, CurrentID=0, Status='live', Symbol='Notum', Aliases='Dmel_CG13076, CG13076, Dmel\\CG13076, Not, Nt, Wf, anon-WO0118547.482, anon-WO03062410.10, anon-WO03062410.8, l(3)72CDf, l(3)72Da, notum, ntm, wf', description='CG13076 gene product from transcript CG13076-RA', other_designations='CG13076-PA|Notum-PA|lethal(3)72CDf|lethal(3)72Da|notum|wingful|wingful/notum', map_location='72C3-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16023480, end_position_on_the_genomic_accession=16032752, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39752, CurrentID=0, Status='live', Symbol='CG5895', Aliases='Dmel_ Dmel\\CG5895', description='CG5895 gene product from transcript CG5895-RA', other_designations='CG5895-PA|CG5895-PC', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16035485, end_position_on_the_genomic_accession=16037227, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39753, CurrentID=0, Status='live', Symbol='Diap1', Aliases='Dmel_CG12284, 0736/01, 1065/03, CG12284, D-IAP1, D-iap1, DIAP, DIAP I, DIAP-1, DIAP1, DIAPI, Diap, Diap-1, Diap-1/Th, Dmel\\CG12284, E(rpr)3-1, IAP, IAP1, Iap1, S048915, TH, Thread, dIAP, dIAP-1, dIAP1, dIap1, diap, diap 1, diap-1, diap1, diap1/thread, l(3)72CDe, l(3)72Dc, l(3)S030605, l(3)S030605b, l(3)S106503, l(3)j5C8, th, thread', description='Death-associated inhibitor of apoptosis 1', other_designations='CG12284-PA|CG12284-PB|CG12284-PC|CG12284-PD|CG12284-PE|CG12284-PF|DIAP1|Diap1-PA|Diap1-PB|Diap1-PC|Diap1-PD|Diap1-PE|Diap1-PF|degradation of inhibitors of apoptosis|drosophila IAP1|drosophila inhibitor of apoptosis|drosophila inhibitor of apoptosis 1|drosophila inhibitor of apoptosis Protein-1|drosophila inhibitor of apoptosis protein|drosophila inhibitor of apoptosis protein 1|drosophila inhibitor-of-apoptosis protein 1|drosophila thread/inhibitor of apoptosis protein-1|inhibitor of apoptosis|inhibitor of apoptosis 1|inhibitor of apoptosis protein 1|inhibitor of apoptosis-1|inhibitor of apoptotic protein 1|inhibitors of apoptosis protein|lethal (3) 72CDe|thread', map_location='3-43.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16038410, end_position_on_the_genomic_accession=16051034, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14462775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462775, CurrentID=0, Status='live', Symbol='CG43923', Aliases='Dmel_ Dmel\\CG43923', description='CG43923 gene product from transcript CG43923-RA', other_designations='CG43923-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15675582, end_position_on_the_genomic_accession=15676102, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39756, CurrentID=0, Status='live', Symbol='CG13075', Aliases='Dmel_ Dmel\\CG13075', description='CG13075 gene product from transcript CG13075-RB', other_designations='CG13075-PB|uncharacterized protein', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16083993, end_position_on_the_genomic_accession=16085264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39757, CurrentID=0, Status='live', Symbol='Tasp1', Aliases='Dmel_CG5241, CG5241, Dmel\\CG5241, Taspase1, anon-WO0118547.172, dTasp, l(3)72Dl', description='Taspase 1', other_designations='CG5241-PA|CG5241-PB|Tasp1-PA|Tasp1-PB|Taspase1|drosophila threonine Aspartase1|lethal (3) 72Dl|taspase 1', map_location='72D5-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16085221, end_position_on_the_genomic_accession=16087525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39758, CurrentID=0, Status='live', Symbol='CG5235', Aliases='Dmel_ Dmel\\CG5235', description='CG5235 gene product from transcript CG5235-RA', other_designations='CG5235-PA|CG5235-PB', map_location='72D4-72D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16087627, end_position_on_the_genomic_accession=16090392, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39759, CurrentID=0, Status='live', Symbol='ClC-c', Aliases='Dmel_CG5284, CG5284, Clc-c, DmClC-c, DmClCc, Dmel\\CG5284, anon-WO0138359.7', description='Chloride channel-c', other_designations='CG5284-PB|CG5284-PC|CG5284-PD|ClC-c-PB|ClC-c-PC|ClC-c-PD|chloride channel-c', map_location='72D5-72D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16090454, end_position_on_the_genomic_accession=16096151, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39760, CurrentID=0, Status='live', Symbol='CG13074', Aliases='Dmel_ Dmel\\CG13074', description='CG13074 gene product from transcript CG13074-RA', other_designations='CG13074-PA|uncharacterized protein', map_location='72D5-72D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16095768, end_position_on_the_genomic_accession=16097493, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39761, CurrentID=0, Status='live', Symbol='ATPsynbetaL', Aliases='Dmel_CG5389, ATPB, CG5389, Dmel\\CG5389, ms(3)72Dt', description='ATP synthase, beta subunit-like', other_designations='ATP synthase beta chain|ATPsynbetaL-PA|CG5389-PA|male sterile (3) 72Dt', map_location='72D5-72D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16097834, end_position_on_the_genomic_accession=16100088, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39762, CurrentID=0, Status='live', Symbol='CG5414', Aliases='Dmel_ Dmel\\CG5414', description='CG5414 gene product from transcript CG5414-RB', other_designations='CG5414-PB', map_location='72D6-72D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16100280, end_position_on_the_genomic_accession=16103482, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39763, CurrentID=0, Status='live', Symbol='IntS9', Aliases='Dmel_CG5222, CG5222, Dmel\\CG5222, Int9, Ints9', description='Integrator 9', other_designations='CG5222-PA|IntS9-PA|integrator 9', map_location='72D6-72D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16103432, end_position_on_the_genomic_accession=16105627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39764, CurrentID=0, Status='live', Symbol='Zn72D', Aliases='Dmel_CG5215, CG5215, Dmel\\CG5215, ZN72D, l(3)72Dk, zn72D', description='Zinc-finger protein at 72D', other_designations='CG5215-PA|CG5215-PB|CG5215-PC|CG5215-PD|CG5215-PE|CG5215-PF|Zn72D-PA|Zn72D-PB|Zn72D-PC|Zn72D-PD|Zn72D-PE|Zn72D-PF|lethal (3) 72Dk', map_location='72D6-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16106594, end_position_on_the_genomic_accession=16112626, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39765, CurrentID=0, Status='live', Symbol='Taf4', Aliases='Dmel_CG5444, CG5444, Dmel\\CG5444, SR3-3, SY3-2 TAF110, TAF, TAF110, TAF4, TAFII110, TAF[II]110, TAF[[II]], TAF[[II]]110, TFIID, TFIID 110, Taf110, Taf[[II]]110, d110, dTAF4, dTAFII110, dTAF[[II]]110, dmTAF4, l(3)72Dj, p110, taf110', description='TBP-associated factor 4', other_designations='CG5444-PA|CG5444-PB|CG5444-PC|CG5444-PD|CG5444-PE|CG5444-PF|CG5444-PG|CG5444-PH|CG5444-PI|TATA-associated factor 110|TBP associated factor 110|TBP-associated factor|TBP-associated factor 110|TBP-associated factor 110kD|Taf4-PA|Taf4-PB|Taf4-PC|Taf4-PD|Taf4-PE|Taf4-PF|Taf4-PG|Taf4-PH|Taf4-PI|lethal(3)72Dj', map_location='72D7-72D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16113212, end_position_on_the_genomic_accession=16121651, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39766, CurrentID=0, Status='live', Symbol='Strumpellin', Aliases='Dmel_CG12272, CG12272, Dmel\\CG12272, p118', description='CG12272 gene product from transcript CG12272-RB', other_designations='CG12272-PA|CG12272-PB|CG12272-PC|Strumpellin-PA|Strumpellin-PB|Strumpellin-PC|strumpellin', map_location='72D8-72D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16122002, end_position_on_the_genomic_accession=16126052, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39768, CurrentID=0, Status='live', Symbol='SsRbeta', Aliases='Dmel_CG5474, BcDNA.LD10457, BcDNA:LD10457, CG5474, Dmel\\CG5474, anon-EST:Liang-1.1, clone 1.1, l(3)s1939', description='Signal sequence receptor beta', other_designations='CG5474-PA|SsRbeta-PA|signal sequence receptor beta', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16129617, end_position_on_the_genomic_accession=16130983, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39769, CurrentID=0, Status='live', Symbol='Trs20', Aliases='Dmel_CG5161, BcDNA:RE62842, CG5161, Dmel\\CG5161, E(smoDN)B-left, Group B-leftp, l(3)72Dh, Trs20', description='TRAPP subunit 20 ortholog (S. cerevisiae)', other_designations='CG5161-PA|TRAPP subunit 20 ortholog|Trs20-PA|enhancer of smoDN group B-left|lethal (3) 72Dh', map_location='72D9-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16131001, end_position_on_the_genomic_accession=16131632, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39770, CurrentID=0, Status='live', Symbol='elg1', Aliases='Dmel_CG16838, CG16838, Dmel\\CG16838', description='elg1 homolog', other_designations='CG16838-PC|CG16838-PD|CG16838-PE|elg1-PC|elg1-PD|elg1-PE', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16132028, end_position_on_the_genomic_accession=16136691, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39771, CurrentID=0, Status='live', Symbol='CG5157', Aliases='Dmel_ Dmel\\CG5157', description='CG5157 gene product from transcript CG5157-RA', other_designations='CG5157-PA', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16136487, end_position_on_the_genomic_accession=16137879, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39772, CurrentID=0, Status='live', Symbol='CG5151', Aliases='Dmel_ Dmel\\CG5151', description='CG5151 gene product from transcript CG5151-RE', other_designations='CG5151-PC|CG5151-PD|CG5151-PE|CG5151-PF', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16137802, end_position_on_the_genomic_accession=16164701, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39773, CurrentID=0, Status='live', Symbol='CG13073', Aliases='Dmel_ Dmel\\CG13073', description='CG13073 gene product from transcript CG13073-RB', other_designations='CG13073-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16213051, end_position_on_the_genomic_accession=16217303, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39774, CurrentID=0, Status='live', Symbol='l(3)72Dn', Aliases='Dmel_CG5018, CG5018, Dmel\\CG5018, l(3)j5A4', description='lethal (3) 72Dn', other_designations='CG5018-PA|CG5018-PB|l(3)72Dn-PA|l(3)72Dn-PB', map_location='72D10-72D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16226126, end_position_on_the_genomic_accession=16228737, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39775, CurrentID=0, Status='live', Symbol='CG5027', Aliases='Dmel_ Dmel\\CG5027, LD24073p', description='CG5027 gene product from transcript CG5027-RA', other_designations='CG5027-PA', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16229100, end_position_on_the_genomic_accession=16231139, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39776, CurrentID=0, Status='live', Symbol='PDCD-5', Aliases='Dmel_CG13072, BcDNA:RE17056, CG13072, DMPDCD-5, Dmel\\CG13072, PDCD5', description='Programmed Cell Death 5', other_designations='CG13072-PA|PDCD-5-PA|programmed cell death 5|programmed cell death gene-5', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16231081, end_position_on_the_genomic_accession=16232038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39777, CurrentID=0, Status='live', Symbol='MED10', Aliases='Dmel_CG5057, BcDNA:SD24044, CG5057, Dmel\\CG5057, Med10, Nut2, dMED10, dTRAP15, l(3)72Ds, med10', description='Mediator complex subunit 10', other_designations='CG5057-PB|CG5057-PC|MED10-PB|MED10-PC|mediator|mediator complex subunit 10', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16232084, end_position_on_the_genomic_accession=16232777, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39778, CurrentID=0, Status='live', Symbol='l(3)72Dr', Aliases='Dmel_CG32154, CG32154, CG5082, Dmel\\CG32154', description='lethal (3) 72Dr', other_designations='CG32154-PA|l(3)72Dr-PA', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16237723, end_position_on_the_genomic_accession=16239462, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39779, CurrentID=0, Status='live', Symbol='CG13055', Aliases='Dmel_ Dmel\\CG13055', description='CG13055 gene product from transcript CG13055-RB', other_designations='CG13055-PA|CG13055-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16249560, end_position_on_the_genomic_accession=16252355, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39780, CurrentID=0, Status='live', Symbol='CG13054', Aliases='Dmel_ Dmel\\CG13054', description='CG13054 gene product from transcript CG13054-RA', other_designations='CG13054-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16255852, end_position_on_the_genomic_accession=16257191, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39781, CurrentID=0, Status='live', Symbol='CG13071', Aliases='Dmel_ Dmel\\CG13071', description='CG13071 gene product from transcript CG13071-RB', other_designations='CG13071-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16262740, end_position_on_the_genomic_accession=16263247, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39782, CurrentID=0, Status='live', Symbol='CG13070', Aliases='Dmel_ Dmel\\CG13070', description='CG13070 gene product from transcript CG13070-RB', other_designations='CG13070-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16263343, end_position_on_the_genomic_accession=16264151, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39783, CurrentID=0, Status='live', Symbol='CG4950', Aliases='Dmel_ CT15890, Dmel\\CG4950', description='CG4950 gene product from transcript CG4950-RC', other_designations='CG4950-PB|CG4950-PC', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16265797, end_position_on_the_genomic_accession=16268156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39784, CurrentID=0, Status='live', Symbol='CG13068', Aliases='Dmel_ Dmel\\CG13068', description='CG13068 gene product from transcript CG13068-RA', other_designations='CG13068-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16269392, end_position_on_the_genomic_accession=16269911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39785, CurrentID=0, Status='live', Symbol='CG13067', Aliases='Dmel_ Dmel\\CG13067, anon-WO0153538.39, anon-WO0153538.46', description='CG13067 gene product from transcript CG13067-RA', other_designations='CG13067-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16271282, end_position_on_the_genomic_accession=16271918, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39786, CurrentID=0, Status='live', Symbol='CG13065', Aliases='Dmel_ BcDNA:RH56745, Dmel\\CG13065', description='CG13065 gene product from transcript CG13065-RA', other_designations='CG13065-PA|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16277435, end_position_on_the_genomic_accession=16278180, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39787, CurrentID=0, Status='live', Symbol='CG13050', Aliases='Dmel_ Dmel\\CG13050', description='CG13050 gene product from transcript CG13050-RA', other_designations='CG13050-PA|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16278458, end_position_on_the_genomic_accession=16279271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39788, CurrentID=0, Status='live', Symbol='CG13049', Aliases='Dmel_ BcDNA:RE26528, Dmel\\CG13049', description='CG13049 gene product from transcript CG13049-RA', other_designations='CG13049-PA|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16281132, end_position_on_the_genomic_accession=16281930, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39789, CurrentID=0, Status='live', Symbol='CG13048', Aliases='Dmel_ Dmel\\CG13048', description='CG13048 gene product from transcript CG13048-RB', other_designations='CG13048-PA|CG13048-PB|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16282809, end_position_on_the_genomic_accession=16283921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39790, CurrentID=0, Status='live', Symbol='CG13047', Aliases='Dmel_ Dmel\\CG13047', description='CG13047 gene product from transcript CG13047-RB', other_designations='CG13047-PB|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16284680, end_position_on_the_genomic_accession=16285396, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39791, CurrentID=0, Status='live', Symbol='CG13046', Aliases='Dmel_ Dmel\\CG13046', description='CG13046 gene product from transcript CG13046-RB', other_designations='CG13046-PB|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16287580, end_position_on_the_genomic_accession=16288422, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39792, CurrentID=0, Status='live', Symbol='CG13045', Aliases='Dmel_ Dmel\\CG13045', description='CG13045 gene product from transcript CG13045-RB', other_designations='CG13045-PB|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16289712, end_position_on_the_genomic_accession=16290950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39793, CurrentID=0, Status='live', Symbol='CG4962', Aliases='Dmel_ CT15908, Dmel\\CG4962', description='CG4962 gene product from transcript CG4962-RA', other_designations='CG4962-PA', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16294711, end_position_on_the_genomic_accession=16295547, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39794, CurrentID=0, Status='live', Symbol='CG4982', Aliases='Dmel_ Dmel\\CG4982', description='CG4982 gene product from transcript CG4982-RA', other_designations='CG4982-PA', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16301453, end_position_on_the_genomic_accession=16302027, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39795, CurrentID=0, Status='live', Symbol='CG13044', Aliases='Dmel_ Dmel\\CG13044', description='CG13044 gene product from transcript CG13044-RA', other_designations='CG13044-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16302490, end_position_on_the_genomic_accession=16303178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39796, CurrentID=0, Status='live', Symbol='CG13043', Aliases='Dmel_ BcDNA:RH10371, CT32262, Dmel\\CG13043', description='CG13043 gene product from transcript CG13043-RA', other_designations='CG13043-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16305225, end_position_on_the_genomic_accession=16305870, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39797, CurrentID=0, Status='live', Symbol='CG13063', Aliases='Dmel_ Dmel\\CG13063', description='CG13063 gene product from transcript CG13063-RA', other_designations='CG13063-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16306600, end_position_on_the_genomic_accession=16307205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39798, CurrentID=0, Status='live', Symbol='CG13042', Aliases='Dmel_ Dmel\\CG13042', description='CG13042 gene product from transcript CG13042-RA', other_designations='CG13042-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16307846, end_position_on_the_genomic_accession=16308480, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39799, CurrentID=0, Status='live', Symbol='CG13062', Aliases='Dmel_ Dmel\\CG13062', description='CG13062 gene product from transcript CG13062-RB', other_designations='CG13062-PB|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16314285, end_position_on_the_genomic_accession=16315149, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39801, CurrentID=0, Status='live', Symbol='CG13041', Aliases='Dmel_ BcDNA:RE63063, Dmel\\CG13041', description='CG13041 gene product from transcript CG13041-RA', other_designations='CG13041-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16318954, end_position_on_the_genomic_accession=16319526, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39802, CurrentID=0, Status='live', Symbol='CG13060', Aliases='Dmel_ BcDNA:RE40431, Dmel\\CG13060', description='CG13060 gene product from transcript CG13060-RA', other_designations='CG13060-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16319856, end_position_on_the_genomic_accession=16320454, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39803, CurrentID=0, Status='live', Symbol='CG13059', Aliases='Dmel_ Dmel\\CG13059', description='CG13059 gene product from transcript CG13059-RA', other_designations='CG13059-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16323672, end_position_on_the_genomic_accession=16324308, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  170876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=170876, CurrentID=0, Status='live', Symbol='Pif2', Aliases='Dmel_CG31483, BcDNA:AT26658, CG31483, Dmel\\CG31483, NEST:bs18a02, NEST:bs19e11, NEST:bs20e01, NEST:bs25a03, NEST:bs35g11, PIF-2, bs20e01.y1', description='PFTAIRE-interacting factor 2', other_designations='CG31483-PA|Pif2-PA', map_location='83F1-83F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6549036, end_position_on_the_genomic_accession=6549983, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39805, CurrentID=0, Status='live', Symbol='CG13039', Aliases='Dmel_ BcDNA:AT27312, Dmel\\CG13039', description='CG13039 gene product from transcript CG13039-RA', other_designations='CG13039-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16326057, end_position_on_the_genomic_accession=16326762, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  170878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=170878, CurrentID=0, Status='live', Symbol='Obp83g', Aliases='Dmel_CG31558, 83g, BcDNA:RE27306, CG15583, CG31558, DmelObp83g, Dmel\\CG31558, Obp83f, obp83g', description='Odorant-binding protein 83g', other_designations='CG31558-PA|Obp83g-PA', map_location='83D4-83D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6111671, end_position_on_the_genomic_accession=6112302, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39807, CurrentID=0, Status='live', Symbol='CG33060', Aliases='Dmel_ CG13057, Dmel\\CG33060', description='CG33060 gene product from transcript CG33060-RB', other_designations='CG33060-PB|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16331040, end_position_on_the_genomic_accession=16331767, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39808, CurrentID=0, Status='live', Symbol='CG4998', Aliases='Dmel_ Dmel\\CG4998, SPH66', description='CG4998 gene product from transcript CG4998-RB', other_designations='CG4998-PA|CG4998-PB', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16339003, end_position_on_the_genomic_accession=16345241, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39809, CurrentID=0, Status='live', Symbol='mRpS34', Aliases='Dmel_CG42648, CG13037, CG42648, Dmel\\CG42648, Dmel_CG13037, MRP-S34', description='mitochondrial ribosomal protein S34', other_designations='CG42648-PA|mRpS34-PA', map_location='72E1-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16348074, end_position_on_the_genomic_accession=16348782, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  170882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=170882, CurrentID=0, Status='live', Symbol='Obp56c', Aliases='Dmel_CG30128, 56c, CG15129, CG30128, DmelObp56c, Dmel\\CG30128, OBP56c', description='Odorant-binding protein 56c', other_designations='CG30128-PB|CG30128-PC|Obp56c-PB|Obp56c-PC', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19699995, end_position_on_the_genomic_accession=19701124, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39811, CurrentID=0, Status='live', Symbol='CG4933', Aliases='Dmel_ Dmel\\CG4933', description='CG4933 gene product from transcript CG4933-RA', other_designations='CG4933-PA', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16351991, end_position_on_the_genomic_accession=16353286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39812, CurrentID=0, Status='live', Symbol='Pdh', Aliases='Dmel_CG4899, CG4899, Dmel\\CG4899, pcdr, pdh', description='Photoreceptor dehydrogenase', other_designations='CG4899-PA|CG4899-PB|CG4899-PC|CG4899-PD|Pdh-PA|Pdh-PB|Pdh-PC|Pdh-PD|photoreceptor dehydrogenase|pigment cell dehydrogenase reductase', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16353233, end_position_on_the_genomic_accession=16354297, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39813, CurrentID=0, Status='live', Symbol='CG45057', Aliases='Dmel_ CG34249, CG45027, CG4893, Dmel\\CG45057, Dmel_CG34249, Dmel_CG4893, HDC10482, mdcds_27436', description='CG45057 gene product from transcript CG45057-RA', other_designations='CG45057-PA|CG45057-PB|CG45057-PC|CG45057-PD|uncharacterized protein', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16354746, end_position_on_the_genomic_accession=16359428, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39814, CurrentID=0, Status='live', Symbol='Cpr72Ea', Aliases='Dmel_CG4818, CG4818, Dmel\\CG4818', description='Cuticular protein 72Ea', other_designations='CG4818-PA|Cpr72Ea-PA|cuticular protein 72Ea', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16362603, end_position_on_the_genomic_accession=16363750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39815, CurrentID=0, Status='live', Symbol='Cpr72Eb', Aliases='Dmel_CG12255, CG12255, Dmel\\CG12255', description='Cuticular protein 72Eb', other_designations='CG12255-PA|Cpr72Eb-PA|cuticular protein 72Eb', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16364899, end_position_on_the_genomic_accession=16365707, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39816, CurrentID=0, Status='live', Symbol='Cpr72Ec', Aliases='Dmel_CG4784, CG4784, CT15377, Dmel\\CG4784', description='Cuticular protein 72Ec', other_designations='CG4784-PA|Cpr72Ec-PA|cuticular protein 72Ec', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16365894, end_position_on_the_genomic_accession=16367340, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39817, CurrentID=0, Status='live', Symbol='CG4842', Aliases='Dmel_ Dmel\\CG4842', description='CG4842 gene product from transcript CG4842-RA', other_designations='CG4842-PA', map_location='72E2-72E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16368389, end_position_on_the_genomic_accession=16369518, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39818, CurrentID=0, Status='live', Symbol='roq', Aliases='Dmel_CG16807, CG16807, Dmel\\CG16807, NEST:bs21b12', description='roquin', other_designations='CG16807-PA|CG16807-PB|CG16807-PC|roq-PA|roq-PB|roq-PC', map_location='72E3-72E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16370938, end_position_on_the_genomic_accession=16378566, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39819, CurrentID=0, Status='live', Symbol='CG4753', Aliases='Dmel_ AGPAT4, Dmel\\CG4753', description='CG4753 gene product from transcript CG4753-RB', other_designations='CG4753-PA|CG4753-PB|CG4753-PC', map_location='72E4-72E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16378974, end_position_on_the_genomic_accession=16381570, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39820, CurrentID=0, Status='live', Symbol='CG4729', Aliases='Dmel_ AGPAT3, Dmel\\CG4729', description='CG4729 gene product from transcript CG4729-RA', other_designations='CG4729-PA|CG4729-PB|CG4729-PC|CG4729-PD|CG4729-PE', map_location='72E4-72E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16381772, end_position_on_the_genomic_accession=16385957, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39821, CurrentID=0, Status='live', Symbol='RAF2', Aliases='Dmel_CG4877, CG4689, CG4877, Dmel\\CG4877, dRAF2', description='RING-associated factor 2', other_designations='CG4877-PA|CG4877-PB|CG4877-PC|RAF2-PA|RAF2-PB|RAF2-PC|dRING-associated factor 2', map_location='72E4-72E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16386327, end_position_on_the_genomic_accession=16390816, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39824, CurrentID=0, Status='live', Symbol='Gagr', Aliases='Dmel_CG4680, CG4680, Dmel\\CG4680, Grp', description='Gag-related', other_designations='CG4680-PA|CG4680-PB|Gag related protein|Gagr-PA|Gagr-PB', map_location='72E4-72E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16394586, end_position_on_the_genomic_accession=16396880, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39825, CurrentID=0, Status='live', Symbol='CG44836', Aliases='Dmel_ CG13035, CG13036, Dmel\\CG44836, Dmel_CG13035, Dmel_CG13036', description='CG44836 gene product from transcript CG44836-RF', other_designations='CG44836-PE|CG44836-PF|rugose|uncharacterized protein', map_location='72E5-72E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16390952, end_position_on_the_genomic_accession=16400275, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39826, CurrentID=0, Status='live', Symbol='fax', Aliases='Dmel_CG4609, CG4609, Dmel\\CG4609, Fax, anon-EST:Liang-1.34, anon-WO0172774.62, anon-WO0172774.63, anon-WO0172774.65, anon-WO0172774.66, anon-WO0172774.67, anon-WO0172774.68, anon-WO0172774.69, clone 1.34, fac', description='failed axon connections', other_designations='CG4609-PA|CG4609-PB|CG4609-PC|CG4609-PD|CG4609-PE|CG4609-PF|failed axonal connections|fax-PA|fax-PB|fax-PC|fax-PD|fax-PE|fax-PF', map_location='72E5-72F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16402796, end_position_on_the_genomic_accession=16411828, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39827, CurrentID=0, Status='live', Symbol='TMS1', Aliases='Dmel_CG4672, CG4672, Dmel\\CG4672d', description='CG4672 gene product from transcript CG4672-RA', other_designations='CG4672-PA|CG4672-PB|CG4672-PC|CG4672-PD|CG4672-PE|CG4672-PF|TMS1-PA|TMS1-PB|TMS1-PC|TMS1-PD|TMS1-PE|TMS1-PF|target of methylation-induced silencing 1', map_location='72F1-72F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16412722, end_position_on_the_genomic_accession=16415386, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39828, CurrentID=0, Status='live', Symbol='CG4573', Aliases='Dmel_ Dmel\\CG4573', description='CG4573 gene product from transcript CG4573-RA', other_designations='CG4573-PA', map_location='72F1-72F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16415292, end_position_on_the_genomic_accession=16417177, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39829, CurrentID=0, Status='live', Symbol='Aats-tyr', Aliases='Dmel_CG4561, CG4561, Dmel\\CG4561, TyrRS, YARS, YRS, anon-EST:Posey261, dTyrRS, dYARS', description='Tyrosyl-tRNA synthetase', other_designations='Aats-tyr-PA|CG4561-PA', map_location='72F1-72F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16417349, end_position_on_the_genomic_accession=16419302, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39831, CurrentID=0, Status='live', Symbol='dsx-c73A', Aliases='Dmel_CG32159, CG13034, CG32159, CG4541, DSC73, Dmel\\CG32159, dsc73, dsx-c73, l(3)j10E8', description='doublesex cognate 73A', other_designations='CG32159-PB|drosophila secreted cuticle protein 73|dsx-c73A-PB|lethal (3) j10E8', map_location='73A3-73A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16422795, end_position_on_the_genomic_accession=16451958, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39833, CurrentID=0, Status='live', Symbol='aos', Aliases='Dmel_CG4531, Aos, Argos, CG4531, DmAos, Dmel\\CG4531, arg, gil, l(3)05845, l(3)05959, l(3)j5E11, l(3)rJ472, l(3)rJ812, l(3)rQ526, rlt, sty', description='argos', other_designations='CG4531-PA|aos-PA|giant lens|roulette|strawberry', map_location='73A2-73A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16470386, end_position_on_the_genomic_accession=16483650, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39834, CurrentID=0, Status='live', Symbol='CG33158', Aliases='Dmel_ CG10162, CG4638, Dmel\\CG33158', description='CG33158 gene product from transcript CG33158-RB', other_designations='CG33158-PB|uncharacterized protein', map_location='72F1-73A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16419597, end_position_on_the_genomic_accession=16496560, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39835, CurrentID=0, Status='live', Symbol='CG43373', Aliases='Dmel_ AC, Ago2, CG12436, CG32158, CG42513, Dmel\\CG43373, Dmel_CG12436, Dmel_CG32158, Dmel_CG42513', description='CG43373 gene product from transcript CG43373-RB', other_designations='CG43373-PB|CG43373-PC|uncharacterized protein', map_location='73A4-73A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16491838, end_position_on_the_genomic_accession=16543370, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  39836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39836, CurrentID=0, Status='live', Symbol='st', Aliases='Dmel_CG4314, CG4314, Dmel\\CG4314', description='scarlet', other_designations='CG4314-PA|st-PA', map_location='3-44.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16497651, end_position_on_the_genomic_accession=16500460, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39839, CurrentID=0, Status='live', Symbol='CG13033', Aliases='Dmel_ BcDNA:RE50345, Dmel\\CG13033', description='CG13033 gene product from transcript CG13033-RA', other_designations='CG13033-PA|uncharacterized protein', map_location='73A5-73A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16544567, end_position_on_the_genomic_accession=16545628, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39840, CurrentID=0, Status='live', Symbol='CG4229', Aliases='Dmel_ Dmel\\CG4229', description='CG4229 gene product from transcript CG4229-RA', other_designations='CG4229-PA', map_location='73A7-73A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16556823, end_position_on_the_genomic_accession=16557958, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39841, CurrentID=0, Status='live', Symbol='Mipp1', Aliases='Dmel_CG4123, CG4123, CK2082, Dmel\\CG4123, mipp1', description='Multiple inositol polyphosphate phosphatase 1', other_designations='CG4123-PA|CG4123-PB|CG4123-PC|Mipp1-PA|Mipp1-PB|Mipp1-PC|multiple inositol polyphosphate phosphatase 1', map_location='73A7-73A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16563448, end_position_on_the_genomic_accession=16573334, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39842, CurrentID=0, Status='live', Symbol='mbf1', Aliases='Dmel_CG4143, BcDNA:GH10148, BcDNA:GM01267, CG4143, Dmel\\CG4143, MBF-1, MBF1, MRE4, Mbf-1, Mbf1', description='multiprotein bridging factor 1', other_designations='CG4143-PA|CG4143-PB|CG4143-PC|CG4143-PD|CG4143-PE|mRNA-like ncRNA in embryogenesis 4|mbf1-PA|mbf1-PB|mbf1-PC|mbf1-PD|mbf1-PE', map_location='73A9-73A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16574436, end_position_on_the_genomic_accession=16576871, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39843, CurrentID=0, Status='live', Symbol='nxf2', Aliases='Dmel_CG4118, CG4118, Dmel\\CG4118, NXF2', description='nuclear RNA export factor 2', other_designations='CG4118-PA|CG4118-PB|CG4118-PC|nxf2-PA|nxf2-PB|nxf2-PC', map_location='73A9-73A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16576859, end_position_on_the_genomic_accession=16580306, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39844, CurrentID=0, Status='live', Symbol='Smn', Aliases='Dmel_CG16725, CG16725, DmSMN, Dmel\\CG16725, SMN, dSMN, l(3)73Ao, smn', description='survival motor neuron', other_designations='CG16725-PA|CG16725-PB|Smn-PA|Smn-PB|survival of motor neurons', map_location='73A9-73A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16580398, end_position_on_the_genomic_accession=16581547, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39845, CurrentID=0, Status='live', Symbol='Rpn12', Aliases='Dmel_CG4157, CG4157, Dmel\\CG4157, S14, dRpn12, p30', description='Regulatory particle non-ATPase 12', other_designations='CG4157-PA|Rpn12-PA|regulatory particle non-ATPase 12', map_location='73A10-73A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16583028, end_position_on_the_genomic_accession=16584022, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39846, CurrentID=0, Status='live', Symbol='UQCR-C2', Aliases='Dmel_CG4169, BEST:GH10978, CG 4169, CG4169, Dmel\\CG4169, UCR2, anon-EST:Posey245', description='Ubiquinol-cytochrome c reductase core protein 2', other_designations='BEST:GH10978|CG4169-PA|CG4169-PB|UQCR-C2-PA|UQCR-C2-PB|Ubiquinol-cytochrome C reductase complex core protein 2|Ubiquinol-cytochrome c oxidoreductase core protein 2|ubiquinol-cytochrome c reductase', map_location='73A10-73A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16584124, end_position_on_the_genomic_accession=16586111, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39847, CurrentID=0, Status='live', Symbol='Syx8', Aliases='Dmel_CG4109, CG4109, Dm Syx8, DmSyx8, Dmel\\CG4109, SYX8, Syx', description='Syntaxin 8', other_designations='CG4109-PA|Syx8-PA|dSyntaxin8|syntaxin|syntaxin 8', map_location='73A10-73A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16586108, end_position_on_the_genomic_accession=16587057, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39849, CurrentID=0, Status='live', Symbol='tra', Aliases='Dmel_CG16724, CG16724, Dmel\\CG16724, Dmtra, TRA, Tra', description='transformer', other_designations='CG16724-PA|CG16724-PB|tra-PA|tra-PB', map_location='3-45 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16590059, end_position_on_the_genomic_accession=16591050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39850, CurrentID=0, Status='live', Symbol='spd-2', Aliases='Dmel_CG17286, BcDNA:LD24702, CEP192, CG17286, CG17286-PA, Cep192, D-SPD2, D-Spd2, DSPD-2, DSpd-2, DSpd2, Dmel\\CG17286, Dspd-2, Dspd2, SPD-2, SPD2, Spd-2, Spd2, anon-73Ba, dSpd2, spd2', description='spindle defective 2', other_designations='CG17286-PA|spd-2-PA', map_location='73A10-73A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16591207, end_position_on_the_genomic_accession=16595014, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39854, CurrentID=0, Status='live', Symbol='CG4098', Aliases='Dmel_CG4098-PA, Dmel\\CG4098', description='CG4098 gene product from transcript CG4098-RA', other_designations='CG4098-PA', map_location='73B1-73B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16608138, end_position_on_the_genomic_accession=16609347, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39855, CurrentID=0, Status='live', Symbol='Prosbeta6', Aliases='Dmel_CG4097, Beta-6, CG4097, DTS-5, DTS5, Dmel\\CG4097, PSB1_DROME, PSMB1, Pros26, Pros26-DTS, b6, beta-6, beta6, beta6_dm, l(3)73Ai, l(3)DTS-5, l(3)DTS5, pros26, pros26B, pros26beta6', description='Proteasome beta6 subunit', other_designations='20S proteasome|CG4097-PA|Prosbeta6-PA|beta6|beta6 proteosome subunit|proteasome 26kD subunit|proteasome beta6 subunit|proteasome subunit 20S', map_location='3-44.3 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16609400, end_position_on_the_genomic_accession=16610418, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39856, CurrentID=0, Status='live', Symbol='Su(P)', Aliases='Dmel_CG4086, CG4086, Dmel\\CG4086', description='Suppressor of ref(2)P sterility', other_designations='CG4086-PA|Su(P)-PA|suppressor of ref(2)P sterility', map_location='3-44 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16610662, end_position_on_the_genomic_accession=16612199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39857, CurrentID=0, Status='live', Symbol='CG4101', Aliases='Dmel_ Dmel\\CG4101, anon-73B1, anon73B1', description='CG4101 gene product from transcript CG4101-RA', other_designations='CG4101-PA', map_location='73B1-73B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16612486, end_position_on_the_genomic_accession=16612950, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39858, CurrentID=0, Status='live', Symbol='Mo25', Aliases='Dmel_CG4083, CG4083, Dmel\\CG4083, Dmo25, d l(3)00274, mo25', description='CG4083 gene product from transcript CG4083-RA', other_designations='CG4083-PA|CG4083-PB|Mo25-PA|Mo25-PB', map_location='73B1-73B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16613218, end_position_on_the_genomic_accession=16615472, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39860, CurrentID=0, Status='live', Symbol='Baldspot', Aliases='Dmel_CG3971, CG 3971, CG3971, Dmel\\CG3971, NEST:bs18f10, NEST:bs18g10, blp, l(3)01895, l(3)02281, l(3)04106, l(3)neo21, l(3)neo22, noa', description='CG3971 gene product from transcript CG3971-RA', other_designations='Baldspot-PA|Baldspot-PB|CG3971-PA|CG3971-PB|baldspot|neighbor of abl', map_location='73B4-73B', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16648442, end_position_on_the_genomic_accession=16661615, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39861, CurrentID=0, Status='live', Symbol='Galphaf', Aliases='Dmel_CG12232, CG12232, Dmel\\CG12232, G[[f]], Galpha73B, Galpha[[f]]73B, Gf, Gfalpha', description='G protein alpha f subunit', other_designations='CG12232-PA|G protein alpha 73B|GTP-binding protein alpha-subunit|Galphaf-PA|Gf alpha', map_location='73B5-73B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16662557, end_position_on_the_genomic_accession=16667934, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39862, CurrentID=0, Status='live', Symbol='zetaCOP', Aliases='Dmel_CG3948, CG3948, CT13124, Dmel\\CG3948, copz1, zeta-Cop, zetaCop', description='Coat Protein (coatomer) zeta', other_designations='CG3948-PA|CG3948-PB|CG3948-PC|COPI zeta subunit|coat protein (coatomer) zeta|zetaCOP-PA|zetaCOP-PB|zetaCOP-PC|zetacop', map_location='73B5-73B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16671664, end_position_on_the_genomic_accession=16673555, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39863, CurrentID=0, Status='live', Symbol='CG13032', Aliases='Dmel_ Dmel\\CG13032', description='CG13032 gene product from transcript CG13032-RA', other_designations='CG13032-PA|uncharacterized protein', map_location='73B6-73B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16673964, end_position_on_the_genomic_accession=16676264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39864, CurrentID=0, Status='live', Symbol='Lasp', Aliases='Dmel_CG3849, CG3849, Dmel\\CG3849, anon-73Bd, cDNA B, lasp', description='CG3849 gene product from transcript CG3849-RB', other_designations='CG3849-PA|CG3849-PB|CG3849-PC|CG3849-PD|Lasp-PA|Lasp-PB|Lasp-PC|Lasp-PD|lasp', map_location='73B6-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16676406, end_position_on_the_genomic_accession=16711677, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39865, CurrentID=0, Status='live', Symbol='CG9692', Aliases='Dmel_ Dmel\\CG9692', description='CG9692 gene product from transcript CG9692-RB', other_designations='CG9692-PB|CG9692-PC', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16704973, end_position_on_the_genomic_accession=16706428, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39866, CurrentID=0, Status='live', Symbol='Dab', Aliases='Dmel_CG9695, CG9695, Dmel\\CG9695, cDNA C, dab', description='Disabled', other_designations='CG9695-PA|CG9695-PC|CG9695-PD|CG9695-PE|CG9695-PF|CG9695-PG|CG9695-PH|CG9695-PI|Dab-PA|Dab-PC|Dab-PD|Dab-PE|Dab-PF|Dab-PG|Dab-PH|Dab-PI|disabled', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16712126, end_position_on_the_genomic_accession=16724575, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39867, CurrentID=0, Status='live', Symbol='CG13031', Aliases='Dmel_ Dmel\\CG13031', description='CG13031 gene product from transcript CG13031-RA', other_designations='CG13031-PA|uncharacterized protein', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16724861, end_position_on_the_genomic_accession=16725375, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39868, CurrentID=0, Status='live', Symbol='CG13026', Aliases='Dmel_ Dmel\\CG13026', description='CG13026 gene product from transcript CG13026-RC', other_designations='CG13026-PB|CG13026-PC|uncharacterized protein', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16724678, end_position_on_the_genomic_accession=16726169, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39869, CurrentID=0, Status='live', Symbol='PGRP-SB2', Aliases='Dmel_CG9697, CG9697, Dmel\\CG9697, PGRP-SB', description='CG9697 gene product from transcript CG9697-RA', other_designations='CG9697-PA|CG9697-PB|PGRP-SB2-PA|PGRP-SB2-PB', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16726541, end_position_on_the_genomic_accession=16727288, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39870, CurrentID=0, Status='live', Symbol='PGRP-SB1', Aliases='Dmel_CG9681, 142657_at, CG9681, Dmel\\CG9681, PGRP SB1, PGRP-SB', description='CG9681 gene product from transcript CG9681-RA', other_designations='CG9681-PA|PGRP-SB1-PA|Peptidoglycan-recognition protein-SB1 precursor|peptidoglycan receptor protein SB1|peptidoglycan recognition|peptidoglycan recognition protein SB1|protein-like', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16727299, end_position_on_the_genomic_accession=16727989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39871, CurrentID=0, Status='live', Symbol='Dbp73D', Aliases='Dmel_CG9680, CG9680, DBP73D, DmRH7, Dmel\\CG9680', description='Dead box protein 73D', other_designations='CG9680-PA|Dbp73D-PA|dead box protein 73D', map_location='73C1-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16728496, end_position_on_the_genomic_accession=16730818, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39872, CurrentID=0, Status='live', Symbol='CG9701', Aliases='Dmel_ Dmel\\CG9701', description='CG9701 gene product from transcript CG9701-RA', other_designations='CG9701-PA', map_location='73C2-73C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16733546, end_position_on_the_genomic_accession=16739364, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39873, CurrentID=0, Status='live', Symbol='Nrt', Aliases='Dmel_CG9704, BP106, BcDNA:LD22004, CG9704, DAB, Dab, Dmel\\CG9704, MabBP106, NRT, NRT BP106, NT, Neur BP106, dab, l(3)73Be, neu, nrt, Nrt', description='Neurotactin', other_designations='CG9704-PA|CG9704-PB|CG9704-PC|Nrt-PA|Nrt-PB|Nrt-PC|disabled|neurotactin', map_location='73C3-73C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16766072, end_position_on_the_genomic_accession=16777465, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39874, CurrentID=0, Status='live', Symbol='CG13025', Aliases='Dmel_ Dmel\\CG13025', description='CG13025 gene product from transcript CG13025-RA', other_designations='CG13025-PA|CG13025-PB|uncharacterized protein', map_location='73C4-73C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16776531, end_position_on_the_genomic_accession=16779736, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39875, CurrentID=0, Status='live', Symbol='CG9705', Aliases='Dmel_ Dmel\\CG9705', description='CG9705 gene product from transcript CG9705-RA', other_designations='CG9705-PA|CG9705-PB|CG9705-PC', map_location='73C4-73C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16780100, end_position_on_the_genomic_accession=16782419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39876, CurrentID=0, Status='live', Symbol='CG9706', Aliases='Dmel_ Dmel\\CG9706', description='CG9706 gene product from transcript CG9706-RA', other_designations='CG9706-PA|CG9706-PB', map_location='73C4-73C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16783010, end_position_on_the_genomic_accession=16785246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39877, CurrentID=0, Status='live', Symbol='Int6', Aliases='Dmel_CG9677, BcDNA.GM01233, BcDNA:GM01233, CG9677, DINT6, Dmel\\CG9677, eIF-3p48/INT-6, eIF3-S6, eIF3-p48/INT6, eIF3e, l(3)10547, l(3)j9E8', description='Int6 homologue', other_designations='CG9677-PA|Int6-PA', map_location='73C4-73C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16785242, end_position_on_the_genomic_accession=16786941, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39878, CurrentID=0, Status='live', Symbol='CG9674', Aliases='Dmel_ Dmel\\CG9674', description='CG9674 gene product from transcript CG9674-RD', other_designations='CG9674-PA|CG9674-PD|CG9674-PE|CG9674-PF|CG9674-PG|CG9674-PH|CG9674-PI|glutamate synthase', map_location='73C4-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16787573, end_position_on_the_genomic_accession=16807731, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  39879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39879, CurrentID=0, Status='live', Symbol='nudC', Aliases='Dmel_CG9710, CG9710, Dmel\\CG9710, D NUDC, NudC, Q9VVA6', description='CG9710 gene product from transcript CG9710-RB', other_designations='CG9710-PA|CG9710-PB|nudC-PA|nudC-PB', map_location='73D1-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16804994, end_position_on_the_genomic_accession=16819532, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39880, CurrentID=0, Status='live', Symbol='CG13024', Aliases='Dmel_ BcDNA:RH61817, Dmel\\CG13024', description='CG13024 gene product from transcript CG13024-RC', other_designations='CG13024-PB|CG13024-PC|uncharacterized protein', map_location='73D1-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16809658, end_position_on_the_genomic_accession=16829074, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39881, CurrentID=0, Status='live', Symbol='TSG101', Aliases='Dmel_CG9712, CG9712, DmTSG101, Dmel\\CG9712, ESCRT-I, Ept, Tsg1-1, Tsg101, burs, dTSg101, dTsg101, dVps23, ept, tsg101, vps23', description='Tumor susceptibility gene 101', other_designations='CG9712-PA|TSG101-PA|erupted|tumor suppressor protein 101|tumor susceptibility Gene-101|tumor susceptibility gene 101|tumor susceptibility gene 101 ortholog', map_location='73D1-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16842294, end_position_on_the_genomic_accession=16844585, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39882, CurrentID=0, Status='live', Symbol='CG9669', Aliases='Dmel_ Dmel\\CG9669', description='CG9669 gene product from transcript CG9669-RB', other_designations='CG9669-PA|CG9669-PB', map_location='73D1-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16844568, end_position_on_the_genomic_accession=16844991, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39883, CurrentID=0, Status='live', Symbol='CG9715', Aliases='Dmel_ Dmel\\CG9715', description='CG9715 gene product from transcript CG9715-RA', other_designations='CG9715-PA', map_location='73D1-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16845386, end_position_on_the_genomic_accession=16852350, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39884, CurrentID=0, Status='live', Symbol='sina', Aliases='Dmel_CG9949, CG9949, Dmel\\CG9949, SINA, Sina, anon-EST:fe2H7, sia', description='seven in absentia', other_designations='CG9949-PA|CG9949-PB|CG9949-PC|anon-fast-evolving-2H7|seven in absenti|seven-in-absentia|sina-PA|sina-PB|sina-PC', map_location='3-45.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16853751, end_position_on_the_genomic_accession=16858712, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39885, CurrentID=0, Status='live', Symbol='sinah', Aliases='Dmel_CG13030, CG13030, Dmel\\CG13030, SinaH, anon-WO0140519.228, sina-h', description='sina homologue', other_designations='CG13030-PC|Sina-Homologue|sinah-PC', map_location='73D3-73D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16859080, end_position_on_the_genomic_accession=16860529, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39886, CurrentID=0, Status='live', Symbol='CG13029', Aliases='Dmel_ BcDNA:AT31918, Dmel\\CG13029, anon-WO0118547.38', description='CG13029 gene product from transcript CG13029-RC', other_designations='CG13029-PC|CG13029-PD|uncharacterized protein', map_location='73D3-73D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16860622, end_position_on_the_genomic_accession=16861838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39887, CurrentID=0, Status='live', Symbol='Rh4', Aliases='Dmel_CG9668, CG9668, DMELRH4, Dm Rh4, Dmel\\CG9668, FBgn0003250, RH4, Rh, rh4', description='Rhodopsin 4', other_designations='CG9668-PA|Rh4-PA|Rhodopsin4|rhodopsin|rhodopsin 4', map_location='73D2-73D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16852527, end_position_on_the_genomic_accession=16862517, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39888, CurrentID=0, Status='live', Symbol='CG9951', Aliases='Dmel_ Dmel\\CG9951', description='CG9951 gene product from transcript CG9951-RA', other_designations='CG9951-PA', map_location='73D3-73D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16863164, end_position_on_the_genomic_accession=16864989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39889, CurrentID=0, Status='live', Symbol='Lmpt', Aliases='Dmel_CG42679, BcDNA:RE37250, CG11914, CG11915, CG11916, CG32171, CG33256, CG42679, CG9959, Dmel\\CG42679, Dmel_CG11915, Dmel_CG32171, Dmel_CG33256, lmpt', description='Limpet', other_designations='CG42679-PA|CG42679-PB|CG42679-PC|CG42679-PD|CG42679-PE|CG42679-PF|CG42679-PG|CG42679-PH|CG42679-PI|CG42679-PJ|CG42679-PK|CG42679-PL|CG42679-PM|CG42679-PN|CG42679-PO|Lmpt-PA|Lmpt-PB|Lmpt-PC|Lmpt-PD|Lmpt-PE|Lmpt-PF|Lmpt-PG|Lmpt-PH|Lmpt-PI|Lmpt-PJ|Lmpt-PK|Lmpt-PL|Lmpt-PM|Lmpt-PN|Lmpt-PO|limpet', map_location='73D5-73D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16865295, end_position_on_the_genomic_accession=16922487, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  39894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39894, CurrentID=0, Status='live', Symbol='CG13023', Aliases='Dmel_ Dmel\\CG13023', description='CG13023 gene product from transcript CG13023-RB', other_designations='CG13023-PA|CG13023-PB|uncharacterized protein', map_location='73D6-73D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16923914, end_position_on_the_genomic_accession=16927434, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39895, CurrentID=0, Status='live', Symbol='CG11905', Aliases='Dmel_ Dmel\\CG11905', description='CG11905 gene product from transcript CG11905-RB', other_designations='CG11905-PA|CG11905-PB|CG11905-PC|CG11905-PD|CG11905-PE|CG11905-PF|CG11905-PG|CG11905-PH|CG11905-PI|CG11905-PL|uncharacterized protein', map_location='73D6-73D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16928381, end_position_on_the_genomic_accession=16935536, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39896, CurrentID=0, Status='live', Symbol='CG13022', Aliases='Dmel_ Dmel\\CG13022', description='CG13022 gene product from transcript CG13022-RB', other_designations='CG13022-PB|uncharacterized protein', map_location='73D7-73D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16938330, end_position_on_the_genomic_accession=16938952, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39897, CurrentID=0, Status='live', Symbol='Cpr73D', Aliases='Dmel_CG9665, CG9665, Dmel\\CG9665', description='Cuticular protein 73D', other_designations='CG9665-PB|CG9665-PC|Cpr73D-PB|Cpr73D-PC|cuticular protein 73D', map_location='73D7-73E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16939100, end_position_on_the_genomic_accession=16957689, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39899, CurrentID=0, Status='live', Symbol='Nc73EF', Aliases='Dmel_CG11661, 141810_at, CG11661, Dmel\\CG11661, E1, FBgn0036682, Nc73eF, anon-WO0118547.329', description='Neural conserved at 73EF', other_designations='CG11661-PA|CG11661-PB|CG11661-PC|CG11661-PF|CG11661-PG|CG11661-PH|CG11661-PI|CG11661-PJ|Nc73EF-PA|Nc73EF-PB|Nc73EF-PC|Nc73EF-PF|Nc73EF-PG|Nc73EF-PH|Nc73EF-PI|Nc73EF-PJ|alpha-ketoglutarate dehydrogenase E1 subunit|neural conserved at 73EF|neural conserved at 73KF', map_location='73E1-73E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16954822, end_position_on_the_genomic_accession=16968724, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  39900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39900, CurrentID=0, Status='live', Symbol='Exn', Aliases='Dmel_CG42665, BcDNA:GH03693, CG14060, CG3799, CG42665, Dmel\\CG42665, Dmel_CG14060, Dmel_CG3799, exn', description='Ephexin', other_designations='CG42665-PA|CG42665-PB|CG42665-PC|CG42665-PD|CG42665-PE|Exn-PA|Exn-PB|Exn-PC|Exn-PD|Exn-PE|ephexin', map_location='73E1-73E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16970013, end_position_on_the_genomic_accession=16994388, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  39902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39902, CurrentID=0, Status='live', Symbol='CG3764', Aliases='Dmel_ Dmel\\CG3764', description='CG3764 gene product from transcript CG3764-RB', other_designations='CG3764-PA|CG3764-PB|CG3764-PC', map_location='73E2-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16994717, end_position_on_the_genomic_accession=17012565, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39903, CurrentID=0, Status='live', Symbol='CG6664', Aliases='Dmel_ Dmel\\CG6664', description='CG6664 gene product from transcript CG6664-RC', other_designations='CG6664-PA|CG6664-PB|CG6664-PC', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17012886, end_position_on_the_genomic_accession=17016180, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39904, CurrentID=0, Status='live', Symbol='CG7728', Aliases='Dmel_ Dmel\\CG7728', description='CG7728 gene product from transcript CG7728-RA', other_designations='CG7728-PA', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17016536, end_position_on_the_genomic_accession=17020270, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39905, CurrentID=0, Status='live', Symbol='CG6652', Aliases='Dmel_ Dmel\\CG6652', description='CG6652 gene product from transcript CG6652-RA', other_designations='CG6652-PA', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17020348, end_position_on_the_genomic_accession=17022856, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39906, CurrentID=0, Status='live', Symbol='a10', Aliases='Dmel_CG6642, A10, CG6642, DmelCSP4, Dmel\\CG6642, OS-D, OS-D/A10, Os-D, Phk-1', description='antennal protein 10', other_designations='CG6642-PA|Pherokine1|a10-PA|pherokine 1', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17023123, end_position_on_the_genomic_accession=17023741, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39907, CurrentID=0, Status='live', Symbol='Fit2', Aliases='Dmel_CG7729, CG7729, Dmel\\CG7729', description='Fermitin 2', other_designations='CG7729-PA|Fit2-PA|fermitin 2', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17024328, end_position_on_the_genomic_accession=17027117, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39908, CurrentID=0, Status='live', Symbol='CG7730', Aliases='Dmel_ Dmel\\CG7730', description='CG7730 gene product from transcript CG7730-RC', other_designations='CG7730-PC', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17027515, end_position_on_the_genomic_accession=17029814, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39909, CurrentID=0, Status='live', Symbol='Ilp8', Aliases='Dmel_CG14059, CG14059, DILP8, Dmel\\CG14059, dILP8, dilp8, dlip8', description='Insulin-like peptide 8', other_designations='CG14059-PB|Ilp8-PB|drosophila insulin-like peptide 8', map_location='73E4-73E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17029812, end_position_on_the_genomic_accession=17032080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39910, CurrentID=0, Status='live', Symbol='beg', Aliases='Dmel_CG7842, CG7842, Dmel\\CG7842', description='bad egg', other_designations='CG7842-PA|beg-PA', map_location='73E4-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17035668, end_position_on_the_genomic_accession=17036865, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39914, CurrentID=0, Status='live', Symbol='Papst2', Aliases='Dmel_CG7853, CG7853, Dmel\\CG7853, PAPST2, dPAPST2', description='PAPS transporter 2', other_designations='CG7853-PA|Papst2-PA', map_location='73E5-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17042797, end_position_on_the_genomic_accession=17044465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39915, CurrentID=0, Status='live', Symbol='CG14057', Aliases='Dmel_ Dmel\\CG14057', description='CG14057 gene product from transcript CG14057-RA', other_designations='CG14057-PA|CG14057-PB|uncharacterized protein', map_location='73E5-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17044401, end_position_on_the_genomic_accession=17045038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39916, CurrentID=0, Status='live', Symbol='TpnC73F', Aliases='Dmel_CG7930, CG7930, DmTnC3, DmTpnC73F, DmTpnCIa, Dmel\\CG7930, TNC, TNC73F, TnC, TnC73F, TpnC Ia, TpnCIa, tpnC73F', description='Troponin C at 73F', other_designations='CG7930-PA|CG7930-PC|TpnC73F-PA|TpnC73F-PC|TroponinC73F|troponin C|troponin C at 73F|troponin C73F', map_location='73E5-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17046063, end_position_on_the_genomic_accession=17048390, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39917, CurrentID=0, Status='live', Symbol='rogdi', Aliases='Dmel_CG7725, CG7725, Dmel\\CG7725', description='CG7725 gene product from transcript CG7725-RB', other_designations='CG7725-PA|CG7725-PB|rogdi-PA|rogdi-PB', map_location='73E5-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17049512, end_position_on_the_genomic_accession=17055454, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  39918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39918, CurrentID=0, Status='live', Symbol='CG7724', Aliases='Dmel_ Dmel\\CG7724', description='CG7724 gene product from transcript CG7724-RA', other_designations='CG7724-PA', map_location='73E5-73E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17055613, end_position_on_the_genomic_accession=17058533, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39919, CurrentID=0, Status='live', Symbol='Rbp6', Aliases='Dmel_CG32169, BcDNA:RE25373, CG14055, CG14056, CG32169, CG6573, Dmel\\CG32169, RRM6, rbp6, rrm6', description='RNA-binding protein 6', other_designations='CG32169-PA|CG32169-PB|CG32169-PC|CG32169-PD|CG32169-PE|CG32169-PF|CG32169-PG|Rbp6-PA|Rbp6-PB|Rbp6-PC|Rbp6-PD|Rbp6-PE|Rbp6-PF|Rbp6-PG', map_location='73E5-74A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17058774, end_position_on_the_genomic_accession=17238212, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  39922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39922, CurrentID=0, Status='live', Symbol='CG6512', Aliases='Dmel_ AFG3L2, DmCG6512, Dmel\\CG6512', description='CG6512 gene product from transcript CG6512-RB', other_designations='CG6512-PA|CG6512-PB', map_location='74A1-74A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17238613, end_position_on_the_genomic_accession=17242128, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39923, CurrentID=0, Status='live', Symbol='CG7707', Aliases='Dmel_ Dmel\\CG7707, FBXO8', description='CG7707 gene product from transcript CG7707-RA', other_designations='CG7707-PA', map_location='74A2-74A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17243436, end_position_on_the_genomic_accession=17244746, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39924, CurrentID=0, Status='live', Symbol='CG6497', Aliases='Dmel_ Dmel\\CG6497', description='CG6497 gene product from transcript CG6497-RB', other_designations='CG6497-PA|CG6497-PB', map_location='74A2-74A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17252377, end_position_on_the_genomic_accession=17253937, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39925, CurrentID=0, Status='live', Symbol='CG13723', Aliases='Dmel_ Dmel\\CG13723', description='CG13723 gene product from transcript CG13723-RA', other_designations='CG13723-PA|uncharacterized protein', map_location='74A2-74A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17264416, end_position_on_the_genomic_accession=17264986, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39926, CurrentID=0, Status='live', Symbol='ND-24L', Aliases='Dmel_CG6485, CG6485, Dmel\\CG6485, NUHM', description='NADH dehydrogenase (ubiquinone) 24 kDa subunit-like', other_designations='CG6485-PA|NADH-ubiquinone oxidoreductase 24 kDa subunit|NADH:ubiquinone oxidoreductase 24 kDa subunit|ND-24L-PA', map_location='74A4-74A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17319969, end_position_on_the_genomic_accession=17320811, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39927, CurrentID=0, Status='live', Symbol='CG13724', Aliases='Dmel_ Dmel\\CG13724', description='CG13724 gene product from transcript CG13724-RA', other_designations='CG13724-PA|uncharacterized protein', map_location='74A4-74A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17322841, end_position_on_the_genomic_accession=17324493, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39928, CurrentID=0, Status='live', Symbol='CG13725', Aliases='Dmel_ Dmel\\CG13725', description='CG13725 gene product from transcript CG13725-RA', other_designations='CG13725-PA|uncharacterized protein', map_location='74A4-74A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17324618, end_position_on_the_genomic_accession=17325518, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39929, CurrentID=0, Status='live', Symbol='Or74a', Aliases='Dmel_CG13726, 74A.1, 74a, CG13726, DOR26, Dmel\\CG13726', description='Odorant receptor 74a', other_designations='CG13726-PA|Or74a-PA|odorant receptor 74a', map_location='74A4-74A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17327733, end_position_on_the_genomic_accession=17329169, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39930, CurrentID=0, Status='live', Symbol='CG6479', Aliases='Dmel_ Dmel\\CG6479', description='CG6479 gene product from transcript CG6479-RC', other_designations='CG6479-PA|CG6479-PB|CG6479-PC', map_location='74A5-74A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17337172, end_position_on_the_genomic_accession=17339262, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39931, CurrentID=0, Status='live', Symbol='CG13727', Aliases='Dmel_ Dmel\\CG13727', description='CG13727 gene product from transcript CG13727-RA', other_designations='CG13727-PA|uncharacterized protein', map_location='74A5-74A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17340279, end_position_on_the_genomic_accession=17341316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39932, CurrentID=0, Status='live', Symbol='brv2', Aliases='Dmel_CG16793, CG16793, CT33194, Dmel\\CG16793', description='brivido-2', other_designations='CG16793-PA|CG16793-PB|brv2-PA|brv2-PB', map_location='74A5-74A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17342050, end_position_on_the_genomic_accession=17344696, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39933, CurrentID=0, Status='live', Symbol='Mip', Aliases='Dmel_CG6456, ASB, AST-B, Ast-B, AstB, AstB-1, CG6456, DAP-B, DmMIP1, DmMIP2, DmMIP3, DmMIP4, DmMIP5, Dmel\\CG6456, Drm-MIP, Drm-MIP-1, Drm-MIP-2, Drm-MIP-3, Drm-MIP-4, Drm-MIP-5, MIP, MIP-1, MIP-2, MIP-3, MIP-4, MIP-5, mip', description='Myoinhibiting peptide precursor', other_designations='CG6456-PA|Drostatin-B1|Drostatin-B3|Drostatin-B4|Drostatin-B5|Mip-PA|Mip-like peptide|allatostatin (B-type)|allatostatin B|allatostatin B-1|allatostatin cricket-type|allatostatin-B|allatostatin-B1|allostatin B|drostatin-B2|mioinhibiting peptide precursor|myoinhibiting peptide|myoinhibiting peptide precursor|myoinhibitory peptide', map_location='74A5-74A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17344947, end_position_on_the_genomic_accession=17349666, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39934, CurrentID=0, Status='live', Symbol='CG7692', Aliases='Dmel_ Dmel\\CG7692', description='CG7692 gene product from transcript CG7692-RD', other_designations='CG7692-PA|CG7692-PB|CG7692-PC|CG7692-PD', map_location='74A5-74A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17351822, end_position_on_the_genomic_accession=17358056, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39935, CurrentID=0, Status='live', Symbol='tap', Aliases='Dmel_CG7659, CG7659, Dmel\\CG7659, Tap, bHLHc9, bps, l(3)01658', description='target of Poxn', other_designations='CG7659-PA|biparous|tap-PA|target of Pox-n|target of poxn', map_location='74A5-74A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17359682, end_position_on_the_genomic_accession=17361811, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39936, CurrentID=0, Status='live', Symbol='Cad74A', Aliases='Dmel_CG6445, CG6445, CT20086, Cad74B, DCad74B, Dmel\\CG6445, cad74A', description='Cadherin 74A', other_designations='CG6445-PA|CG6445-PB|CG6445-PC|Cad74A-PA|Cad74A-PB|Cad74A-PC|cadherin 74A', map_location='74A5-74A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17367753, end_position_on_the_genomic_accession=17379530, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39937, CurrentID=0, Status='live', Symbol='CG13728', Aliases='Dmel_ Dmel\\CG13728', description='CG13728 gene product from transcript CG13728-RA', other_designations='CG13728-PA|uncharacterized protein', map_location='74A6-74A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17380026, end_position_on_the_genomic_accession=17382728, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39938, CurrentID=0, Status='live', Symbol='CG13731', Aliases='Dmel_ Dmel\\CG13731', description='CG13731 gene product from transcript CG13731-RB', other_designations='CG13731-PB|uncharacterized protein', map_location='74A6-74B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17383636, end_position_on_the_genomic_accession=17388956, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39939, CurrentID=0, Status='live', Symbol='blot', Aliases='Dmel_CG3897, CG3897, DmBLOT, Dmel\\CG3897, l(3)1658', description='bloated tubules', other_designations='CG3897-PA|CG3897-PB|CG3897-PC|CG3897-PD|CG3897-PE|blated tubules|blot-PA|blot-PB|blot-PC|blot-PD|blot-PE', map_location='74B1-74B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17396031, end_position_on_the_genomic_accession=17421514, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  39940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39940, CurrentID=0, Status='live', Symbol='Sec3', Aliases='Dmel_CG3885, CG3885, Dmel\\CG3885, sec3', description='Sec3 ortholog (S. cerevisiae)', other_designations='CG3885-PA|Sec3 ortholog|Sec3-PA|sec3', map_location='74B2-74B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17421787, end_position_on_the_genomic_accession=17425217, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39941, CurrentID=0, Status='live', Symbol='CG33051', Aliases='Dmel_ CG3882, Dmel\\CG33051', description='CG33051 gene product from transcript CG33051-RA', other_designations='CG33051-PA|uncharacterized protein', map_location='74B4-74B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17426286, end_position_on_the_genomic_accession=17427130, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39943, CurrentID=0, Status='live', Symbol='frc', Aliases='Dmel_CG3874, CG3874, Dmel\\CG3874, FRC, Frc, UDP-GlcNAc/UDP-GlcUA, UST74C, anon-WO0172774.5, anon-WO0172774.7, dFRC, l(3)00073, l(3)02619, l(3)10098, l(3)L6750, l(3)S070006, l(3)j2E3, ust74C', description='fringe connection', other_designations='CG3874-PA|CG3874-PB|Fringe-connection|UDP-sugar transporter in 74C|frc-PA|frc-PB', map_location='74B5-74C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17428936, end_position_on_the_genomic_accession=17433190, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39944, CurrentID=0, Status='live', Symbol='CG32176', Aliases='Dmel_ CG13729, CG13730, Dmel\\CG32176', description='CG32176 gene product from transcript CG32176-RA', other_designations='CG32176-PA|uncharacterized protein', map_location='74C1-74C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17433868, end_position_on_the_genomic_accession=17437971, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39945, CurrentID=0, Status='live', Symbol='CG12229', Aliases='Dmel_ Dmel\\CG12229', description='CG12229 gene product from transcript CG12229-RA', other_designations='CG12229-PA|pyruvate kinase|uncharacterized protein', map_location='74C1-74C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17434472, end_position_on_the_genomic_accession=17436431, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39947, CurrentID=0, Status='live', Symbol='CG18265', Aliases='Dmel_ Dmel\\CG18265, ORE-10', description='CG18265 gene product from transcript CG18265-RB', other_designations='CG18265-PA|CG18265-PB|CG18265-PC|uncharacterized protein', map_location='74C2-74C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17439401, end_position_on_the_genomic_accession=17457461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39948, CurrentID=0, Status='live', Symbol='QIL1', Aliases='Dmel_CG7603, CG7603, Dmel\\CG7603', description='CG7603 gene product from transcript CG7603-RA', other_designations='CG7603-PA|QIL1-PA', map_location='74C3-74C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17465945, end_position_on_the_genomic_accession=17466768, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39949, CurrentID=0, Status='live', Symbol='CG7589', Aliases='Dmel_ CT23187, Dmel\\CG7589', description='CG7589 gene product from transcript CG7589-RB', other_designations='CG7589-PA|CG7589-PB|ligand-gated chloride channel', map_location='74C3-74C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17470127, end_position_on_the_genomic_accession=17474517, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39950, CurrentID=0, Status='live', Symbol='UQCR-Q', Aliases='Dmel_CG7580, BcDNA:AT13736, CG7580, Dmel\\CG7580, UCRQ', description='Ubiquinol-cytochrome c reductase ubiquinone-binding protein', other_designations='CG7580-PA|CG7580-PB|CG7580-PC|CG7580-PD|UQCR-Q-PA|UQCR-Q-PB|UQCR-Q-PC|UQCR-Q-PD|Ubiquinol-cytochrome C reductase complex ubiquinone-binding protein QP-C|Ubiquinol-cytochrome c oxidoreductase Ubiquinone-binding protein QP-C|Ubiquinol-cytochrome c reductase Ubiquinone-binding protein QP-C|ubiquinone binding protein', map_location='74C3-74C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17474761, end_position_on_the_genomic_accession=17476124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39951, CurrentID=0, Status='live', Symbol='CG13733', Aliases='Dmel_ Dmel\\CG13733', description='CG13733 gene product from transcript CG13733-RA', other_designations='CG13733-PA|uncharacterized protein', map_location='74C3-74C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17477795, end_position_on_the_genomic_accession=17478926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39952, CurrentID=0, Status='live', Symbol='qjt', Aliases='Dmel_CG13732, CG13732, Dmel\\CG13732, Quijote', description='quijote', other_designations='CG13732-PA|CG13732-PB|qjt-PA|qjt-PB', map_location='74C4-74C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17479096, end_position_on_the_genomic_accession=17480141, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39953, CurrentID=0, Status='live', Symbol='CG6333', Aliases='Dmel_ Dmel\\CG6333', description='CG6333 gene product from transcript CG6333-RB', other_designations='CG6333-PA|CG6333-PB', map_location='74C4-74C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17480129, end_position_on_the_genomic_accession=17486230, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39954, CurrentID=0, Status='live', Symbol='Oatp74D', Aliases='Dmel_CG7571, CG7571, Dmel\\CG7571, oatp 74D', description='Organic anion transporting polypeptide 74D', other_designations='CG7571-PA|CG7571-PC|CG7571-PD|CG7571-PE|Oatp74D-PA|Oatp74D-PC|Oatp74D-PD|Oatp74D-PE|organic anion transporting polypeptide 74D', map_location='74D1-74D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17487118, end_position_on_the_genomic_accession=17511693, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  39955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39955, CurrentID=0, Status='live', Symbol='U4-U6-60K', Aliases='Dmel_CG6322, CG6322, Dmel\\CG6322, Prp4, U4/U6-60K', description='U4-U6 small nuclear riboprotein factor 60K', other_designations='CG6322-PA|U4-U6-60K-PA', map_location='74D2-74D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17511855, end_position_on_the_genomic_accession=17513912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39956, CurrentID=0, Status='live', Symbol='CG7564', Aliases='Dmel_ Alsin2, Dmel\\CG7564', description='CG7564 gene product from transcript CG7564-RD', other_designations='CG7564-PA|CG7564-PB|CG7564-PC|CG7564-PD', map_location='74D2-74D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17514298, end_position_on_the_genomic_accession=17524740, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39957, CurrentID=0, Status='live', Symbol='Edc3', Aliases='Dmel_CG6311, CG6311, DmEDC3, Dmel\\CG6311, EDC3, LSM16, LSm16', description='Enhancer of decapping 3', other_designations='CG6311-PB|Edc3-PB|enhancer of decapping 3|enhancer of decapping-3', map_location='74D2-74D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17525333, end_position_on_the_genomic_accession=17529861, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39958, CurrentID=0, Status='live', Symbol='Nedd4', Aliases='Dmel_CG42279, ACN62428.1, CG32184, CG42279, CG7555, D Dmel\\CG42279, Dmel_CG32184, Dmel_CG7555, NEDD4, dNedd4, dNedd4-1, nedd4', description='CG42279 gene product from transcript CG42279-RJ', other_designations='CG42279-PE|CG42279-PF|CG42279-PG|CG42279-PH|CG42279-PI|CG42279-PJ|CG42279-PK|CG42279-PL|CG42279-PM|DNedd4|Nedd4-PE|Nedd4-PF|Nedd4-PG|Nedd4-PH|Nedd4-PI|Nedd4-PJ|Nedd4-PK|Nedd4-PL|Nedd4-PM', map_location='74D3-74D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17530081, end_position_on_the_genomic_accession=17546660, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  39959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39959, CurrentID=0, Status='live', Symbol='Jon74E', Aliases='Dmel_CG6298, CG6298, CT19724, Dmel\\CG6298, SP85', description='Jonah 74E', other_designations='CG6298-PA|CG6298-PB|Jon74E-PA|Jon74E-PB|jonah 74E', map_location='74D3-74D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17546739, end_position_on_the_genomic_accession=17549127, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39960, CurrentID=0, Status='live', Symbol='CG7542', Aliases='Dmel_ Dmel\\CG7542, SP71', description='CG7542 gene product from transcript CG7542-RB', other_designations='CG7542-PA|CG7542-PB', map_location='74D4-74D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17549826, end_position_on_the_genomic_accession=17550980, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39961, CurrentID=0, Status='live', Symbol='CycT', Aliases='Dmel_CG6292, CG6292, Dmcyclin T, Dmel\\CG6292, ORE-14, P-TEFb, anon-74EFc, dT, p124', description='Cyclin T', other_designations='CG6292-PA|CG6292-PB|CG6292-PC|CG6292-PD|CycT-PA|CycT-PB|CycT-PC|CycT-PD|CyclinT|cyclin T|cyclin T1|cyclin T2|dCyclin T|dmcyclin T', map_location='74D4-74D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17551004, end_position_on_the_genomic_accession=17557788, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39962, CurrentID=0, Status='live', Symbol='Eip74EF', Aliases='Dmel_CG32180, 3.3, 74EF, CG32180, CG6273, CG6285, Dmel\\CG32180, E-74, E74, E74-B, E74A, E74A/B, E74B, ECIP, Eip74, Eip74 A, Eip74A, Eip74E, e74, e74B, eip74EF, eip74ef, l(3)24, l(3)neo24, mE74', description='Ecdysone-induced protein 74EF', other_designations='CG32180-PA|CG32180-PB|CG32180-PC|CG32180-PD|CG32180-PE|Ecdysone-induced prot 74EF|Ecdysone-induced-protein 74EF|Eip74EF-PA|Eip74EF-PB|Eip74EF-PC|Eip74EF-PD|Eip74EF-PE|ecdysone induced protein 74EF|ecdysone inducible protein|ecdysone-induced protein 74EF', map_location='74D4-74E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17558672, end_position_on_the_genomic_accession=17619325, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  39964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39964, CurrentID=0, Status='live', Symbol='Vps60', Aliases='Dmel_CG6259, CG6259, Chmp5, Dmel\\CG6259, VPS60, anon-74EFb, bs36h07.y1', description='Vacuolar protein sorting 60', other_designations='CG6259-PA|CG6259-PB|Vps60-PA|Vps60-PB|vacuolar protein sorting 60', map_location='74E2-74E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17620123, end_position_on_the_genomic_accession=17621625, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  39965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39965, CurrentID=0, Status='live', Symbol='CG7510', Aliases='Dmel_ Dmel\\CG7510, anon-74EFa', description='CG7510 gene product from transcript CG7510-RB', other_designations='CG7510-PA|CG7510-PB', map_location='74E2-74E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17621824, end_position_on_the_genomic_accession=17630971, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  39966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39966, CurrentID=0, Status='live', Symbol='CG7497', Aliases='Dmel_ BcDNA:GH27361, CHED-related, Dmel\\CG7497, anon-WO0170980.88, anon-WO0170980.89', description='CG7497 gene product from transcript CG7497-RC', other_designations='CG7497-PB|CG7497-PC', map_location='74E3-74E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17633449, end_position_on_the_genomic_accession=17638001, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  39969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39969, CurrentID=0, Status='live', Symbol='CG7484', Aliases='Dmel_ BcDNA:SD16138, Dmel\\CG7484, Prise 8, Sep15', description='CG7484 gene product from transcript CG7484-RB', other_designations='CG7484-PB', map_location='74E4-74E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17654617, end_position_on_the_genomic_accession=17655645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39970, CurrentID=0, Status='live', Symbol='Crtc', Aliases='Dmel_CG6064, CG6064, CRTC, Dmel\\CG6064, ORE-12, TORC, TORC1, cg6064, dTORC', description='CREB-regulated transcription coactivator', other_designations='CG6064-PA|CG6064-PB|CG6064-PC|Crtc-PA|Crtc-PB|Crtc-PC|TORC', map_location='74E4-74E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17659818, end_position_on_the_genomic_accession=17669234, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39971, CurrentID=0, Status='live', Symbol='CG6052', Aliases='Dmel_ Dmel\\CG6052', description='CG6052 gene product from transcript CG6052-RA', other_designations='CG6052-PA', map_location='74F3-74F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17723718, end_position_on_the_genomic_accession=17729244, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39972, CurrentID=0, Status='live', Symbol='Ccn', Aliases='Dmel_CG32183, CG13735, CG13736, CG32183, Dmel\\CG32183, dCCN', description='CG32183 gene product from transcript CG32183-RB', other_designations='CG32183-PB|Ccn-PB', map_location='74F1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17690015, end_position_on_the_genomic_accession=17748839, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  39973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39973, CurrentID=0, Status='live', Symbol='CG7460', Aliases='Dmel_ Dmel\\CG7460', description='CG7460 gene product from transcript CG7460-RB', other_designations='CG7460-PB|CG7460-PC', map_location='74F4-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17740540, end_position_on_the_genomic_accession=17742540, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39974, CurrentID=0, Status='live', Symbol='CG6034', Aliases='Dmel_ Dmel\\CG6034', description='CG6034 gene product from transcript CG6034-RA', other_designations='CG6034-PA', map_location='75A1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17745173, end_position_on_the_genomic_accession=17746860, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39975, CurrentID=0, Status='live', Symbol='Adgf-B', Aliases='Dmel_CG5998, ADGF-B, AdgfB, CG5998, Dmel\\CG5998, adgfb', description='Adenosine deaminase-related growth factor B', other_designations='Adgf-B-PA|CG5998-PA|adenosine deaminase-related growth factor B', map_location='75A1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17748833, end_position_on_the_genomic_accession=17750588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39976, CurrentID=0, Status='live', Symbol='Adgf-A', Aliases='Dmel_CG5992, ADGDF-A, ADGF-A, Adgf-a, AdgfA, CG5992, Dmel\\CG5992, adgf-a, adgfa', description='Adenosine deaminase-related growth factor A', other_designations='Adgf-A-PA|Adgf-A-PB|CG5992-PA|CG5992-PB|adenosine deaminase|adenosine deaminase-related growth factor A', map_location='75A1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17751731, end_position_on_the_genomic_accession=17764158, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39977, CurrentID=0, Status='live', Symbol='CG42816', Aliases='Dmel_ CG32186, CG5587, CG5944, Dmel\\CG42816, Dmel_CG32186', description='CG42816 gene product from transcript CG42816-RB', other_designations='CG42816-PB|uncharacterized protein', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17792088, end_position_on_the_genomic_accession=17797594, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  39978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39978, CurrentID=0, Status='live', Symbol='NUCB1', Aliases='Dmel_CG32190, CG32190, CG5587, Dmel\\CG32190, NucB1, nucb1', description='CG32190 gene product from transcript CG32190-RA', other_designations='CG32190-PA|NUCB1-PA', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17797738, end_position_on_the_genomic_accession=17802279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39979, CurrentID=0, Status='live', Symbol='CG5589', Aliases='Dmel_ DmRH17, Dmel\\CG5589, cg5589', description='CG5589 gene product from transcript CG5589-RA', other_designations='CG5589-PA', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17802490, end_position_on_the_genomic_accession=17804715, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39980, CurrentID=0, Status='live', Symbol='CG32187', Aliases='Dmel_ CG5584, DPLG2, Dmel\\CG32187', description='CG32187 gene product from transcript CG32187-RA', other_designations='CG32187-PA|PIF-like gene 2|uncharacterized protein', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17804916, end_position_on_the_genomic_accession=17806365, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39981, CurrentID=0, Status='live', Symbol='cln3', Aliases='Dmel_CG5582, CG5582, CLN3, Dmel\\CG5582', description='CG5582 gene product from transcript CG5582-RA', other_designations='CG5582-PA|CG5582-PB|ceroid-lipofuscinosis, neuronal 3|cln3-PA|cln3-PB', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17809231, end_position_on_the_genomic_accession=17814119, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39982, CurrentID=0, Status='live', Symbol='Ir75a', Aliases='Dmel_CG14585, CG14585, DmelIR75a, Dmel\\CG14585, IR75a, IR75abc', description='Ionotropic receptor 75a', other_designations='CG14585-PB|Ir75a-PB|ionotropic receptor 75a', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17815570, end_position_on_the_genomic_accession=17817921, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39983, CurrentID=0, Status='live', Symbol='Ir75c', Aliases='Dmel_CG42642, CG14586, CG14586-A, CG42642, DmelIR75c, Dmel\\CG42642, Dmel_CG14586, IR75abc, IR75c', description='Ionotropic receptor 75c', other_designations='CG42642-PA|Ir75c-PA|ionotropic receptor 75c', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17820493, end_position_on_the_genomic_accession=17822834, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39984, CurrentID=0, Status='live', Symbol='Grd', Aliases='Dmel_CG7446, CG7446, CT22917, DmGABA(GRD), Dmel\\CG7446, GRD', description='Glycine receptor', other_designations='CG7446-PA|GABA and glycine-like receptor of drosophila|GABA-and-glycine-receptor-like-subunit-of-Drosophila|GABA/Glycine-gated chloride channel-like|Grd-PA|glycine receptor', map_location='75A3-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17823333, end_position_on_the_genomic_accession=17839767, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39985, CurrentID=0, Status='live', Symbol='CG5577', Aliases='Dmel_ 20151379, Dmel\\CG5577', description='CG5577 gene product from transcript CG5577-RB', other_designations='CG5577-PA|CG5577-PB', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17838650, end_position_on_the_genomic_accession=17841427, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39986, CurrentID=0, Status='live', Symbol='CG5567', Aliases='Dmel_ 24666141, Dmel\\CG5567, Phos5567', description='CG5567 gene product from transcript CG5567-RA', other_designations='CG5567-PA', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17841540, end_position_on_the_genomic_accession=17843544, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39987, CurrentID=0, Status='live', Symbol='MED19', Aliases='Dmel_CG5546, CG5546, Dmel\\CG5546, Med19', description='Mediator complex subunit 19', other_designations='CG5546-PA|CG5546-PB|MED19-PA|MED19-PB|mediator complex subunit 19', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17843725, end_position_on_the_genomic_accession=17846501, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  39988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39988, CurrentID=0, Status='live', Symbol='CG7430', Aliases='Dmel_ Dmel\\CG7430, E3', description='CG7430 gene product from transcript CG7430-RA', other_designations='CG7430-PA|CG7430-PB|alpha-ketoglutarate dehydrogenase E3 subunit|pyruvate dehydrogenase E3 subunit', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17846916, end_position_on_the_genomic_accession=17849601, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39989, CurrentID=0, Status='live', Symbol='CG7441', Aliases='Dmel_ BEST:CK01295, Dmel\\CG7441, GH31', description='CG7441 gene product from transcript CG7441-RA', other_designations='CG7441-PA|CG7441-PB', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17849727, end_position_on_the_genomic_accession=17852453, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39990, CurrentID=0, Status='live', Symbol='CG5535', Aliases='Dmel_ Dmel\\CG5535, anon-WO0118547.111', description='CG5535 gene product from transcript CG5535-RC', other_designations='CG5535-PB|CG5535-PC', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17852624, end_position_on_the_genomic_accession=17856794, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39991, CurrentID=0, Status='live', Symbol='CG7408', Aliases='Dmel_ Dmel\\CG7408', description='CG7408 gene product from transcript CG7408-RB', other_designations='CG7408-PB|CG7408-PC|CG7408-PD', map_location='75A7-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17879096, end_position_on_the_genomic_accession=17883653, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  39992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39992, CurrentID=0, Status='live', Symbol='CG5506', Aliases='Dmel_ BcDNA:LP12049, Dmel\\CG5506', description='CG5506 gene product from transcript CG5506-RA', other_designations='CG5506-PA', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17883649, end_position_on_the_genomic_accession=17884376, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  39993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39993, CurrentID=0, Status='live', Symbol='CG16775', Aliases='Dmel_ Dmel\\CG16775', description='CG16775 gene product from transcript CG16775-RB', other_designations='CG16775-PB|uncharacterized protein', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17885948, end_position_on_the_genomic_accession=17886645, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39994, CurrentID=0, Status='live', Symbol='CG7402', Aliases='Dmel_ Dmel\\CG7402', description='CG7402 gene product from transcript CG7402-RA', other_designations='CG7402-PA', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17887705, end_position_on_the_genomic_accession=17891404, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39995, CurrentID=0, Status='live', Symbol='Tsp74F', Aliases='Dmel_CG5492, CG5492, Dm.Tsp74F, Dmel\\CG5492', description='Tetraspanin 74F', other_designations='CG5492-PA|CG5492-PB|Tsp74F-PA|Tsp74F-PB|tetraspanin 74F', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17891820, end_position_on_the_genomic_accession=17895862, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  39996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39996, CurrentID=0, Status='live', Symbol='Prestin', Aliases='Dmel_CG5485, CG5485, Dmel\\CG5485, Slc26a5, d dpres', description='CG5485 gene product from transcript CG5485-RA', other_designations='CG5485-PA|Prestin-PA|prestin', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17898889, end_position_on_the_genomic_accession=17907812, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  39997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39997, CurrentID=0, Status='live', Symbol='CG14353', Aliases='Dmel_ Dmel\\CG14353', description='CG14353 gene product from transcript CG14353-RA', other_designations='CG14353-PA|uncharacterized protein', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17908143, end_position_on_the_genomic_accession=17909683, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39998, CurrentID=0, Status='live', Symbol='CG5290', Aliases='Dmel_ Dmel\\CG5290', description='CG5290 gene product from transcript CG5290-RA', other_designations='CG5290-PA', map_location='75A8-75A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17909483, end_position_on_the_genomic_accession=17912565, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  39999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39999, CurrentID=0, Status='live', Symbol='Eip75B', Aliases='Dmel_CG8127, 57B, 75B, BcDNA:GM02640, CG8127, DmE75A, DmE75B, Dmel\\CG8127, E75, E75-C, E75A, E75B, E75C, EP1121b, Eip75, Eip75 A, Eip75A, Mgn00274, NR1D3, Rev-Erb, anon-WO0172774.31, anon-WO0172774.32, dE75, eip75B, l(3)07041, l(3)j11A6, l(3)j12E8, l(3)j3A6, l(3)j5E1, l(3)neo25', description='Ecdysone-induced protein 75B', other_designations='CG8127-PA|CG8127-PB|CG8127-PC|CG8127-PD|CG8127-PE|CG8127-PF|CG8127-PG|E75 nuclear receptor|Eip75B-PA|Eip75B-PB|Eip75B-PC|Eip75B-PD|Eip75B-PE|Eip75B-PF|Eip75B-PG|ecdysone inducible protein 75B|ecdysone-induced protein 75', map_location='75A10-75B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17950953, end_position_on_the_genomic_accession=18064696, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40000, CurrentID=0, Status='live', Symbol='CG13698', Aliases='Dmel_ Dmel\\CG13698', description='CG13698 gene product from transcript CG13698-RB', other_designations='CG13698-PB|CG13698-PC|uncharacterized protein', map_location='75B7-75B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18088417, end_position_on_the_genomic_accession=18106149, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40001, CurrentID=0, Status='live', Symbol='mRpS26', Aliases='Dmel_CG7354, CG7354, Dmel\\CG7354, MRP-S26', description='mitochondrial ribosomal protein S26', other_designations='CG7354-PA|mRpS26-PA', map_location='75B9-75B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18106579, end_position_on_the_genomic_accession=18107499, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40002, CurrentID=0, Status='live', Symbol='CG5147', Aliases='Dmel_ Dmel\\CG5147', description='CG5147 gene product from transcript CG5147-RA', other_designations='CG5147-PA', map_location='75B9-75B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18107848, end_position_on_the_genomic_accession=18109574, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40003, CurrentID=0, Status='live', Symbol='mus304', Aliases='Dmel_CG7347, ATRIP, CG7347, Dmel\\CG7347, Mus304, anon-WO0140519.82, atrip, mus(3)304, mus-304', description='mutagen-sensitive 304', other_designations='ATR-interacting protein|CG7347-PA|mus304-PA|mutagen sensitive 304', map_location='3-46 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18109814, end_position_on_the_genomic_accession=18112847, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40004, CurrentID=0, Status='live', Symbol='CG7341', Aliases='Dmel_ Dmel\\CG7341', description='CG7341 gene product from transcript CG7341-RB', other_designations='CG7341-PA|CG7341-PB', map_location='75B10-75B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18112959, end_position_on_the_genomic_accession=18121003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40005, CurrentID=0, Status='live', Symbol='Cyp312a1', Aliases='Dmel_CG5137, 312a1, CG5137, Dmel\\CG5137, P450 IV', description='CG5137 gene product from transcript CG5137-RB', other_designations='CG5137-PA|CG5137-PB|Cyp312a1-PA|Cyp312a1-PB|cytochrome P450', map_location='75B11-75B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18128940, end_position_on_the_genomic_accession=18132689, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40007, CurrentID=0, Status='live', Symbol='CG7330', Aliases='Dmel_ BcDNA:RE22290, Dmel\\CG7330', description='CG7330 gene product from transcript CG7330-RB', other_designations='CG7330-PA|CG7330-PB', map_location='75B13-75B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18148923, end_position_on_the_genomic_accession=18153807, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40008, CurrentID=0, Status='live', Symbol='CG13699', Aliases='Dmel_ Dmel\\CG13699', description='CG13699 gene product from transcript CG13699-RA', other_designations='CG13699-PA|uncharacterized protein', map_location='75B13-75C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18154897, end_position_on_the_genomic_accession=18162522, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40009, CurrentID=0, Status='live', Symbol='hid', Aliases='Dmel_CG5123, CG5123, Dmel\\CG5123, HID, Hid, Hid1, W1, his, l(3)05014, hid', description='head involution defective', other_designations='CG5123-PA|CG5123-PB|head involuted|head involuted defective|head involution defect|head involution defective 1|head-involution defective|head-involution-defective|hid-PA|hid-PB|wrinkled|wrinkler', map_location='75C1-75C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18167745, end_position_on_the_genomic_accession=18185641, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40010, CurrentID=0, Status='live', Symbol='CG7320', Aliases='Dmel_ Dmel\\CG7320', description='CG7320 gene product from transcript CG7320-RB', other_designations='CG7320-PB', map_location='75C2-75C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18201494, end_position_on_the_genomic_accession=18203637, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40011, CurrentID=0, Status='live', Symbol='CheA75a', Aliases='Dmel_CG7313, BcDNA:HL02449, CG7313, Dmel\\CG7313', description='Chemosensory protein A 75a', other_designations='CG7313-PA|CheA75a-PA|chemosensory protein A 75a', map_location='75C2-75C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18252088, end_position_on_the_genomic_accession=18252846, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40012, CurrentID=0, Status='live', Symbol='CG5103', Aliases='Dmel_ Dmel\\CG5103', description='CG5103 gene product from transcript CG5103-RA', other_designations='CG5103-PA', map_location='75C3-75C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18292125, end_position_on_the_genomic_accession=18294262, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40013, CurrentID=0, Status='live', Symbol='CG13700', Aliases='Dmel_ Dmel\\CG13700', description='CG13700 gene product from transcript CG13700-RD', other_designations='CG13700-PB|CG13700-PC|CG13700-PD|CG13700-PE|uncharacterized protein', map_location='75C4-75C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18298742, end_position_on_the_genomic_accession=18301260, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40014, CurrentID=0, Status='live', Symbol='grim', Aliases='Dmel_CG4345, BcDNA:RE28551, CG4345, Dmel\\CG4345, GRIM, Grim, gri, grm', description='CG4345 gene product from transcript CG4345-RA', other_designations='CG4345-PA|grim-PA', map_location='75C4-75C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18302719, end_position_on_the_genomic_accession=18304417, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40015, CurrentID=0, Status='live', Symbol='rpr', Aliases='Dmel_CG4319, CG4319, Dmel\\CG4319, RPR, Reaper, Rpr, anon-WO0162936.19, rp', description='reaper', other_designations='CG4319-PA|reaper L|rpr-PA', map_location='75C6-75C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18397535, end_position_on_the_genomic_accession=18398435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40016, CurrentID=0, Status='live', Symbol='skl', Aliases='Dmel_CG13701, BcDNA:RE14076, CG13701, Dmel\\CG13701, Skl, meph, veto', description='sickle', other_designations='CG13701-PA|mephisto|skl-PA', map_location='75C6-75C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18438561, end_position_on_the_genomic_accession=18439942, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40017, CurrentID=0, Status='live', Symbol='CG4306', Aliases='Dmel_ Dmel\\CG4306', description='CG4306 gene product from transcript CG4306-RA', other_designations='CG4306-PA', map_location='75C7-75C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18461957, end_position_on_the_genomic_accession=18463163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40019, CurrentID=0, Status='live', Symbol='AstC-R2', Aliases='Dmel_CG13702, ASTC-R2, AlCR-2, AlCR2, CG13702, Dmel\\CG13702, Drostar2, anon-WO0170980.154, anon-WO0170980.155, lincRNA.610', description='Allatostatin C receptor 2', other_designations='ASC-receptor|AstC-R2-PB|AstC-R2-PD|AstC-R2-PE|AstC-R2-PF|CG13702-PB|CG13702-PD|CG13702-PE|CG13702-PF|allatostatin C receptor 2|allatostatin C receptor2|drosophila allatostatin C receptor 2|somatostatin receptor', map_location='75D1-75D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18485137, end_position_on_the_genomic_accession=18502557, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40020, CurrentID=0, Status='live', Symbol='AstC-R1', Aliases='Dmel_CG7285, ASTC-R1, CG7285, Dmel\\CG7285, Drostar1, anon-WO0131005.19, anon-WO0170980.79, anon-WO0170980.80, star1', description='Allatostatin C receptor 1', other_designations='ASC-receptor|ASTC receptor 1|AstC-R1-PA|CG7285-PA|allatostatin C receptor 1|drosophila allatostatin C receptor 1', map_location='75D2-75D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18560258, end_position_on_the_genomic_accession=18576326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40021, CurrentID=0, Status='live', Symbol='CG7271', Aliases='Dmel_ Dmel\\CG7271', description='CG7271 gene product from transcript CG7271-RA', other_designations='CG7271-PA', map_location='75D2-75D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18588759, end_position_on_the_genomic_accession=18590165, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40023, CurrentID=0, Status='live', Symbol='CG4174', Aliases='Dmel_ Dmel\\CG4174', description='CG4174 gene product from transcript CG4174-RE', other_designations='CG4174-PD|CG4174-PE|CG4174-PF', map_location='75D3-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18600150, end_position_on_the_genomic_accession=18617943, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  40024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40024, CurrentID=0, Status='live', Symbol='CG18234', Aliases='Dmel_ Dmel\\CG18234, anon-SAGE:Wang-055', description='CG18234 gene product from transcript CG18234-RC', other_designations='CG18234-PC|anon-SAGE:Wang-055|uncharacterized protein', map_location='75D3-75D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18603425, end_position_on_the_genomic_accession=18605415, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40025, CurrentID=0, Status='live', Symbol='CG18233', Aliases='Dmel_ Dmel\\CG18233', description='CG18233 gene product from transcript CG18233-RB', other_designations='CG18233-PB|uncharacterized protein', map_location='75D3-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18605472, end_position_on_the_genomic_accession=18607349, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40026, CurrentID=0, Status='live', Symbol='CG18231', Aliases='Dmel_ Dmel\\CG18231', description='CG18231 gene product from transcript CG18231-RB', other_designations='CG18231-PB|uncharacterized protein', map_location='75D4-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18607462, end_position_on_the_genomic_accession=18609512, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40028, CurrentID=0, Status='live', Symbol='CG32199', Aliases='Dmel_ CG13382, Dmel\\CG32199', description='CG32199 gene product from transcript CG32199-RB', other_designations='CG32199-PB|uncharacterized protein', map_location='75D4-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18611900, end_position_on_the_genomic_accession=18614004, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40029, CurrentID=0, Status='live', Symbol='CG13380', Aliases='Dmel_ Dmel\\CG13380', description='CG13380 gene product from transcript CG13380-RB', other_designations='CG13380-PB|CG13380-PC|CG13380-PD|uncharacterized protein', map_location='75D3-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18600150, end_position_on_the_genomic_accession=18617943, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  40030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40030, CurrentID=0, Status='live', Symbol='not', Aliases='Dmel_CG4166, CG4166, Dmel\\CG4166, NOT, Nonstop, Not, Q9VVR1, UBP8, USP22, Upb8, anon-WO0257455.9, i155, l(3)02069, l(3)j9A6, nonstop', description='non-stop', other_designations='CG4166-PB|CG4166-PC|Non-stop|nonstop|not-PB|not-PC', map_location='75D4-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18618305, end_position_on_the_genomic_accession=18622129, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40031, CurrentID=0, Status='live', Symbol='bora', Aliases='Dmel_CG6897, CG6897, Dmel\\CG6897', description='aurora borealis', other_designations='CG6897-PA|bora-PA|borealis', map_location='75D4-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18622692, end_position_on_the_genomic_accession=18624751, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40032, CurrentID=0, Status='live', Symbol='MYPT-75D', Aliases='Dmel_CG6896, CG6896, Dmel\\CG6896, MYPT75D', description='CG6896 gene product from transcript CG6896-RB', other_designations='CG6896-PA|CG6896-PB|CG6896-PC|CG6896-PD|MYPT-75D-PA|MYPT-75D-PB|MYPT-75D-PC|MYPT-75D-PD', map_location='75D4-75D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18625959, end_position_on_the_genomic_accession=18667177, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  40033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40033, CurrentID=0, Status='live', Symbol='GNBP2', Aliases='Dmel_CG4144, CG4144, DGNBP-2, Dmel\\CG4144, GNBP, gnbp2', description='Gram-negative bacteria binding protein 2', other_designations='CG4144-PA|CG4144-PB|CG4144-PD|GNBP2-PA|GNBP2-PB|GNBP2-PD', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18673474, end_position_on_the_genomic_accession=18675413, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40034, CurrentID=0, Status='live', Symbol='GNBP1', Aliases='Dmel_CG6895, CG6895, DGNBP-1, DGNBP1, Dmel\\CG6895, GNBP, GNBP 1, GNBP-1', description='Gram-negative bacteria binding protein 1', other_designations='CG6895-PA|GNBP1-PA|glucan-binding protein 1|gram-negative binding protein 1|osiris', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18675800, end_position_on_the_genomic_accession=18677876, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40035, CurrentID=0, Status='live', Symbol='Sgf11', Aliases='Dmel_CG13379, CG13379, Dmel\\CG13379, dSgf11', description='SAGA associated factor 11kDa', other_designations='CG13379-PB|Sgf11-PB', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18681037, end_position_on_the_genomic_accession=18681836, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40036, CurrentID=0, Status='live', Symbol='Chmp1', Aliases='Dmel_CG4108, CG4108, CHMP1, Dmel\\CG4108, VPS46', description='Charged multivesicular body protein 1', other_designations='CG4108-PA|CG4108-PB|Chmp1-PA|Chmp1-PB|charged multivesicular body protein 1', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18682895, end_position_on_the_genomic_accession=18684061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40037, CurrentID=0, Status='live', Symbol='Cyp12c1', Aliases='Dmel_CG4120, 12c1, CG4120, Dmel\\CG4120, cyp12c1', description='CG4120 gene product from transcript CG4120-RB', other_designations='CG4120-PA|CG4120-PB|Cyp12c1-PA|Cyp12c1-PB', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18684259, end_position_on_the_genomic_accession=18686319, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40038, CurrentID=0, Status='live', Symbol='CG6893', Aliases='Dmel_ Dmel\\CG6893', description='CG6893 gene product from transcript CG6893-RB', other_designations='CG6893-PB', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18694187, end_position_on_the_genomic_accession=18695648, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40039, CurrentID=0, Status='live', Symbol='Dic4', Aliases='Dmel_CG18363, CG18363, DmDic4p, Dmel\\CG18363', description='Dicarboxylate carrier 4', other_designations='CG18363-PA|CG18363-PB|CG18363-PC|Dic4-PA|Dic4-PB|Dic4-PC|dicarboxylate carrier 4', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18696402, end_position_on_the_genomic_accession=18698412, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40040, CurrentID=0, Status='live', Symbol='CG12477', Aliases='Dmel_ Dmel\\CG12477', description='CG12477 gene product from transcript CG12477-RA', other_designations='CG12477-PA|uncharacterized protein', map_location='75D7-75D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18721893, end_position_on_the_genomic_accession=18723272, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40041, CurrentID=0, Status='live', Symbol='CG6885', Aliases='Dmel_ Dmel\\CG6885, anon-WO0118547.731', description='CG6885 gene product from transcript CG6885-RA', other_designations='CG6885-PA', map_location='75D8-75D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18744488, end_position_on_the_genomic_accession=18745699, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40042, CurrentID=0, Status='live', Symbol='MED11', Aliases='Dmel_CG6884, BcDNA:GH01072, CG6884, Dmel\\CG6884, Med11, Med21, anon-WO0153538.71, dMED21, med11', description='Mediator complex subunit 11', other_designations='CG6884-PA|MED11-PA|RNA poll II transcription mediator|mediator complex subunit 11', map_location='75D8-75D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18745856, end_position_on_the_genomic_accession=18746623, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40043, CurrentID=0, Status='live', Symbol='Nufip', Aliases='Dmel_CG4076, CG4076, Dmel\\CG4076, NUFIP', description='CG4076 gene product from transcript CG4076-RA', other_designations='CG4076-PA|Nufip-PA|nufip', map_location='75D8-75D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18746611, end_position_on_the_genomic_accession=18748367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40044, CurrentID=0, Status='live', Symbol='Atg3', Aliases='Dmel_CG6877, AUT1, Aut-1, Aut1, CG6877, Dmel\\CG6877, Draut1, atg3', description='Autophagy-related 3', other_designations='Atg3-PA|Aut1|Autophagy-specific gene 3|CG6877-PA', map_location='75D8-75D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18748687, end_position_on_the_genomic_accession=18750483, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40045, CurrentID=0, Status='live', Symbol='ftz-f1', Aliases='Dmel_CG4059, BFTZ-F1, CG13378, CG4059, DEP1, DmFTZ-F1, Dmel\\CG4059, FIP1, FTZ-F1, FTZ-F1A, FTZ-F1alpha, FTZ-Fl, FTZF1, FTZf1alpha, Ftz-F1, Ftz-f1, NR5A3, alphaFTZ-F1, alphaFTZF1, alphaftz-f1, bFTZ-F1, beta FTZ-F1, beta Ftz-f1, beta-FTZ-F1, beta-Ftz-F1, betaFTZ-F1, betaFTZF1, betaFtz-F1, betaFtz-f1, betaftz-F1, betaftz-f1, betaftzf1, dFTZ-F1alpha, dFtz-F1alpha, dmFTZ-f1, ftz-F1alpha, ftzf1, l(3)03649, ftz-f1', description='ftz transcription factor 1', other_designations='CG4059-PA|CG4059-PB|CG4059-PC|CG4059-PD|Ftz interacting protein 1|beta-Ftz transcription factor 1|ftz factor 1|ftz-f1-PA|ftz-f1-PB|ftz-f1-PC|ftz-f1-PD|ftz-transcription factor 1|fushi tarazu transcription factor 1', map_location='75D8-75E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18750607, end_position_on_the_genomic_accession=18801309, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40046, CurrentID=0, Status='live', Symbol='CG14073', Aliases='Dmel_ Dmel\\CG14073, Nirvana', description='CG14073 gene product from transcript CG14073-RC', other_designations='CG14073-PB|CG14073-PC|nirvana|uncharacterized protein', map_location='75E1-75E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18802265, end_position_on_the_genomic_accession=18820675, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40047, CurrentID=0, Status='live', Symbol='HipHop', Aliases='Dmel_CG6874, BcDNA:RE10558, CG6874, Dmel\\CG6874, HIPHOP, hiphop, l(3)neo26', description='HP1-HOAP-interacting protein', other_designations='CG6874-PA|CG6874-PB|HP1 and HOAP-interacting protein|HipHop-PA|HipHop-PB|l(3)neo26|lethal (3) neo26', map_location='75E1-75E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18821171, end_position_on_the_genomic_accession=18822353, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40048, CurrentID=0, Status='live', Symbol='Cat', Aliases='Dmel_CG6871, CAT, CATA, CG6871, CT21282A, DMCATHPO, DROCATHPO, Dmel\\CG6871, U00145, bs36h11.y1, cat, Cat', description='Catalase', other_designations='CATHPO|CG6871-PA|Cat-PA|catalase|cytosolic catalase', map_location='3-47.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18822604, end_position_on_the_genomic_accession=18828188, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40049, CurrentID=0, Status='live', Symbol='Indy', Aliases='Dmel_CG3979, BEST:LP01220, CG3979, Dmel\\CG3979, INDY, anon-EST:fe3A7, anon-WO0172774.70, drIndy, indy', description="I'm not dead yet", other_designations='CG3979-PA|CG3979-PB|CG3979-PC|CG3979-PD|I am not dead yet|Indy-PA|Indy-PB|Indy-PC|Indy-PD|anon-fast-evolving-3A7', map_location='75E1-75E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18828632, end_position_on_the_genomic_accession=18846267, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  40050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40050, CurrentID=0, Status='live', Symbol='CG6865', Aliases='Dmel_ Dmel\\CG6865, SP63', description='CG6865 gene product from transcript CG6865-RB', other_designations='CG6865-PB', map_location='75E2-75E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18863647, end_position_on_the_genomic_accession=18865514, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40051, CurrentID=0, Status='live', Symbol='CG14074', Aliases='Dmel_ Dmel\\CG14074', description='CG14074 gene product from transcript CG14074-RB', other_designations='CG14074-PB|uncharacterized protein', map_location='75E2-75E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18866225, end_position_on_the_genomic_accession=18867431, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40052, CurrentID=0, Status='live', Symbol='Dysb', Aliases='Dmel_CG6856, CG6856, Ddysb, Dmel\\CG6856, NP_649064, dysb', description='Dysbindin ortholog (H. sapiens)', other_designations='CG6856-PA|CG6856-PB|Dysb-PA|Dysb-PB|dysbindin|dysbindin ortholog', map_location='75E2-75E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18867771, end_position_on_the_genomic_accession=18869219, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40053, CurrentID=0, Status='live', Symbol='CG6852', Aliases='Dmel_ BcDNA:GH24739, Dmel\\CG6852', description='CG6852 gene product from transcript CG6852-RC', other_designations='CG6852-PA|CG6852-PC', map_location='75E2-75E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18869230, end_position_on_the_genomic_accession=18870057, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40054, CurrentID=0, Status='live', Symbol='Rad9', Aliases='Dmel_CG3945, CG3945, Dm Dmel\\CG3945, rad9', description='CG3945 gene product from transcript CG3945-RA', other_designations='CG3945-PA|CG3945-PB|Rad9-PA|Rad9-PB', map_location='75E2-75E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18870045, end_position_on_the_genomic_accession=18872553, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40055, CurrentID=0, Status='live', Symbol='Ugt', Aliases='Dmel_CG6850, CG6850, DUGT, Dmel\\CG6850, GT, UGGT, UGT, anon-WO0172774.73, anon-WO0172774.75, ugt', description='UDP-glucose-glycoprotein glucosyltransferase', other_designations='CG6850-PA|UDP-glucose:glycoprotein glucosyltransferase|Ugt-PA', map_location='75E3-75E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18872747, end_position_on_the_genomic_accession=18878245, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40056, CurrentID=0, Status='live', Symbol='CG3961', Aliases='Dmel_ Dmel\\CG3961', description='CG3961 gene product from transcript CG3961-RC', other_designations='CG3961-PA|CG3961-PC|CG3961-PD|CG3961-PE', map_location='75E3-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18878250, end_position_on_the_genomic_accession=18890483, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  40057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40057, CurrentID=0, Status='live', Symbol='NijB', Aliases='Dmel_CG11637, CG11637, Dmel\\CG11637', description='Ninjurin B', other_designations='CG11637-PA|NijB-PA|ninjurin B', map_location='75E4-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18892383, end_position_on_the_genomic_accession=18893278, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40059, CurrentID=0, Status='live', Symbol='CG3902', Aliases='Dmel_ CG 3902, CG18168, CT12987, Dmel\\CG3902, GH07925p', description='CG3902 gene product from transcript CG3902-RA', other_designations='Acyl-coenzyme A dehydrogenase homolog|CG3902-PA', map_location='75E4-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18893120, end_position_on_the_genomic_accession=18898293, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40060, CurrentID=0, Status='live', Symbol='RpL26', Aliases='Dmel_CG6846, CG6846, Dmel\\CG6846, L26, RPL26, Rp L26, lincRNA.S5499', description='Ribosomal protein L26', other_designations='CG6846-PA|CG6846-PB|CG6846-PC|RpL26-PA|RpL26-PB|RpL26-PC|ribosomal protein L26', map_location='75E4-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18898875, end_position_on_the_genomic_accession=18900408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40061, CurrentID=0, Status='live', Symbol='arx', Aliases='Dmel_CG3893, BcDNA:LD18018, CG3893, DmGTSF1, Dmel\\CG3893, Gtsf1', description='asterix', other_designations='CG3893-PA|arx-PA|gametocyte-specific factor 1', map_location='75E4-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18900595, end_position_on_the_genomic_accession=18901554, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40062, CurrentID=0, Status='live', Symbol='CG6843', Aliases='Dmel_ Dmel\\CG6843', description='CG6843 gene product from transcript CG6843-RA', other_designations='CG6843-PA', map_location='75E4-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18901594, end_position_on_the_genomic_accession=18902952, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40063, CurrentID=0, Status='live', Symbol='CSN1b', Aliases='Dmel_CG3889, CG3889, CH1-2, CSN1, CSN1B, Csn1B, Csn1b, DCH1-2, Dch1, Dch1-2, Dmel\\CG3889, csn1, dCSN1', description='COP9 signalosome subunit 1b', other_designations='CG3889-PA|COP9 complex homolog subunit 1 b|COP9 complex homolog subunit 1-2|CSN1b-PA|drosophila COP9 signalosome homolog 1', map_location='75E4-75E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18903057, end_position_on_the_genomic_accession=18904930, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40064, CurrentID=0, Status='live', Symbol='CG6841', Aliases='Dmel_ BcDNA:LD43276, Dmel\\CG6841, Prp6', description='CG6841 gene product from transcript CG6841-RB', other_designations='CG6841-PA|CG6841-PB', map_location='75E4-75E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18905372, end_position_on_the_genomic_accession=18908974, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40065, CurrentID=0, Status='live', Symbol='Ir75d', Aliases='Dmel_CG14076, CG14076, CT33663, DmelIR75d, Dmel\\CG14076, IR75d, lr75d', description='Ionotropic receptor 75d', other_designations='CG14076-PC|Ir75d-PC|ionotropic receptor 75d', map_location='75E5-75E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18908260, end_position_on_the_genomic_accession=18911182, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40066, CurrentID=0, Status='live', Symbol='CG14077', Aliases='Dmel_ Dmel\\CG14077, VIa', description='CG14077 gene product from transcript CG14077-RB', other_designations='CG14077-PB|CG14077-PC|uncharacterized protein', map_location='75E5-75E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18908260, end_position_on_the_genomic_accession=18914150, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40067, CurrentID=0, Status='live', Symbol='CG6839', Aliases='Dmel_ Dmel\\CG6839', description='CG6839 gene product from transcript CG6839-RA', other_designations='CG6839-PA', map_location='75E5-75E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18917581, end_position_on_the_genomic_accession=18919259, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40068, CurrentID=0, Status='live', Symbol='CG18223', Aliases='Dmel_ BcDNA:AT03311, Dmel\\CG18223, SPH200', description='CG18223 gene product from transcript CG18223-RA', other_designations='CG18223-PA|CG18223-PC|uncharacterized protein', map_location='75E6-75E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18919224, end_position_on_the_genomic_accession=18920319, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40069, CurrentID=0, Status='live', Symbol='CG3819', Aliases='Dmel_ CT12789, Dmel\\CG3819', description='CG3819 gene product from transcript CG3819-RB', other_designations='CG3819-PA|CG3819-PB', map_location='75E6-75E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18920430, end_position_on_the_genomic_accession=18922108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40070, CurrentID=0, Status='live', Symbol='CG6836', Aliases='Dmel_ Dmel\\CG6836', description='CG6836 gene product from transcript CG6836-RA', other_designations='CG6836-PA', map_location='75F1-75F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18950629, end_position_on_the_genomic_accession=18955406, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40071, CurrentID=0, Status='live', Symbol='CG14075', Aliases='Dmel_ Dmel\\CG14075', description='CG14075 gene product from transcript CG14075-RA', other_designations='CG14075-PA|uncharacterized protein', map_location='75F2-75F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18984961, end_position_on_the_genomic_accession=18985634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40072, CurrentID=0, Status='live', Symbol='CG11619', Aliases='Dmel_ Dmel\\CG11619', description='CG11619 gene product from transcript CG11619-RA', other_designations='CG11619-PA|uncharacterized protein', map_location='75F2-75F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18985905, end_position_on_the_genomic_accession=18988323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40073, CurrentID=0, Status='live', Symbol='CG18135', Aliases='Dmel_ Dmel\\CG18135', description='CG18135 gene product from transcript CG18135-RB', other_designations='CG18135-PA|CG18135-PB|CG18135-PC|CG18135-PD|CG18135-PE|uncharacterized protein', map_location='75F2-75F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18990026, end_position_on_the_genomic_accession=18998283, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40074, CurrentID=0, Status='live', Symbol='CG3808', Aliases='Dmel_ Dmel\\CG3808, cg3808', description='CG3808 gene product from transcript CG3808-RA', other_designations='CG3808-PA', map_location='75F2-75F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18997875, end_position_on_the_genomic_accession=19000608, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40075, CurrentID=0, Status='live', Symbol='CG18136', Aliases='Dmel_ Dmel\\CG18136', description='CG18136 gene product from transcript CG18136-RA', other_designations='CG18136-PA|uncharacterized protein', map_location='75F2-75F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19000985, end_position_on_the_genomic_accession=19003252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40078, CurrentID=0, Status='live', Symbol='Acp76A', Aliases='Dmel_CG3801, Acp76, CG3801, Dmel\\CG3801, Spn76A', description='Accessory gland protein 76A', other_designations='Acp76A-PA|Acp76A-PB|CG3801-PA|CG3801-PB|accessory gland protein 76A|accessory gland-specific peptide 76A', map_location='75F5-75F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19050757, end_position_on_the_genomic_accession=19052026, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40079, CurrentID=0, Status='live', Symbol='CG3797', Aliases='Dmel_ Dmel\\CG3797', description='CG3797 gene product from transcript CG3797-RB', other_designations='CG3797-PA|CG3797-PB', map_location='75F6-75F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19059400, end_position_on_the_genomic_accession=19062816, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40080, CurrentID=0, Status='live', Symbol='CG6812', Aliases='Dmel_ Dmel\\CG6812', description='CG6812 gene product from transcript CG6812-RA', other_designations='CG6812-PA', map_location='75F6-75F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19063165, end_position_on_the_genomic_accession=19066151, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40081, CurrentID=0, Status='live', Symbol='Mkp3', Aliases='Dmel_CG14080, CG14080, CG14081, DMKP-3, Dmel\\CG14080, Dmkp3, MKP3, dMKP3, mkp-3, mkp3', description='Mitogen-activated protein kinase phosphatase 3', other_designations='CG14080-PA|CG14080-PB|CG14080-PC|CG14080-PD|ERK-inhibitory phosphatase|MAP kinase phosphatase 3|MAPK phosphatase 3|MAPK phosphatase-3|MKP-3-like|MKP-like|Mkp3-PA|Mkp3-PB|Mkp3-PC|Mkp3-PD|drosophila MKP-3', map_location='75F6-75F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19066155, end_position_on_the_genomic_accession=19086426, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40083, CurrentID=0, Status='live', Symbol='MESR6', Aliases='Dmel_CG6818, CG6818, Dmel\\CG6818', description='Misexpression suppressor of ras 6', other_designations='CG6818-PA|CG6818-PB|MESR6-PA|MESR6-PB|misexpression suppressor of Ras 6|misexpression suppressor of ras 6', map_location='75F7-75F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19092670, end_position_on_the_genomic_accession=19094760, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40084, CurrentID=0, Status='live', Symbol='CG11577', Aliases='Dmel_ Dmel\\CG11577, anon-WO0118547.323', description='CG11577 gene product from transcript CG11577-RA', other_designations='CG11577-PA|uncharacterized protein', map_location='75F7-75F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19094890, end_position_on_the_genomic_accession=19096217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40085, CurrentID=0, Status='live', Symbol='CG10424', Aliases='Dmel_ Dmel\\CG10424', description='CG10424 gene product from transcript CG10424-RA', other_designations='CG10424-PA|uncharacterized protein', map_location='75F7-75F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19096399, end_position_on_the_genomic_accession=19097709, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40086, CurrentID=0, Status='live', Symbol='CG14079', Aliases='Dmel_ Dmel\\CG14079', description='CG14079 gene product from transcript CG14079-RA', other_designations='CG14079-PA|uncharacterized protein', map_location='75F7-75F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19097999, end_position_on_the_genomic_accession=19099064, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40087, CurrentID=0, Status='live', Symbol='Gem2', Aliases='Dmel_CG10419, CG10419, Dmel\\CG10419, dGem2', description='Gemin2', other_designations='CG10419-PA|CG10419-PB|Gem2-PA|Gem2-PB|dGemin2', map_location='75F7-75F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19099123, end_position_on_the_genomic_accession=19100363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40088, CurrentID=0, Status='live', Symbol='CG14082', Aliases='Dmel_ Dmel\\CG14082', description='CG14082 gene product from transcript CG14082-RE', other_designations='CG14082-PC|CG14082-PD|CG14082-PE|uncharacterized protein', map_location='75F7-75F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19105381, end_position_on_the_genomic_accession=19138975, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40090, CurrentID=0, Status='live', Symbol='fz2', Aliases='Dmel_CG9739, CG9739, D-Fz2, D-fz2, DFZ2, DFz-2, DFz2, Dfrz2, Dfz2, Dfz[[2]], Dm Fz2, Dmel\\CG9739, FZ2, Fz2, dFz2, dfz(2), dfz2, dfz2r, frz2, fz-2', description='frizzled 2', other_designations='CG9739-PA|CG9739-PB|CG9739-PC|CG9739-PD|CG9739-PE|CG9739-PF|CG9739-PG|D-frizzled2|DFrizzled2|Dfrizzled[[2]]|Frizzled2|drosophila frizzled 2|frizzled-2|fz2-PA|fz2-PB|fz2-PC|fz2-PD|fz2-PE|fz2-PF|fz2-PG', map_location='75F9-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19140975, end_position_on_the_genomic_accession=19235373, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40091, CurrentID=0, Status='live', Symbol='mRpL21', Aliases='Dmel_CG9730, CG9730, Dmel\\CG9730, MRP-L21', description='mitochondrial ribosomal protein L21', other_designations='CG9730-PA|CG9730-PB|CG9730-PC|mRpL21-PA|mRpL21-PB|mRpL21-PC', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19253290, end_position_on_the_genomic_accession=19255061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40092, CurrentID=0, Status='live', Symbol='rept', Aliases='Dmel_CG9750, CG9750, Dmel\\CG9750, Q9V3K3_DROME, REP, REPT, Rept, Reptin, Tip48, Tip48/Reptin, anon-WO0142479.1, dRep, dReptin, drep, l(3)06945, psg10, rep', description='reptin', other_designations='CG9750-PA|CG9750-PB|CG9750-PC|dReptin|rept-PA|rept-PB|rept-PC', map_location='3-55 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19255340, end_position_on_the_genomic_accession=19257111, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40093, CurrentID=0, Status='live', Symbol='nes', Aliases='Dmel_CG9655, CG9655, Dmel\\CG9655, Nes', description='nessy', other_designations='CG9655-PA|CG9655-PB|CG9655-PC|nes-PA|nes-PB|nes-PC', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19257282, end_position_on_the_genomic_accession=19260679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40094, CurrentID=0, Status='live', Symbol='Bet1', Aliases='Dmel_CG14084, BET1, BcDNA:RH40423, CG14084, Dm Bet1, Dmel\\CG14084, G14084, bs16e01.y1', description='BET1 ortholog (S. cerevisiae)', other_designations='BET1 ortholog|Bet1-PA|Bet1-PB|CG14084-PA|CG14084-PB', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19261394, end_position_on_the_genomic_accession=19262437, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40095, CurrentID=0, Status='live', Symbol='Max', Aliases='Dmel_CG9648, CG9648, Dmel\\CG9648, MAX, bHLHd9, d dmax, max', description='CG9648 gene product from transcript CG9648-RD', other_designations='CG9648-PA|CG9648-PC|CG9648-PD|Max-PA|Max-PC|Max-PD', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19262232, end_position_on_the_genomic_accession=19264184, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40096, CurrentID=0, Status='live', Symbol='CG9666', Aliases='Dmel_ Dmel\\CG9666, Dromel_CG9666_FBtr0074991_mORF', description='CG9666 gene product from transcript CG9666-RD', other_designations='CG9666-PA|CG9666-PD', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19263766, end_position_on_the_genomic_accession=19266838, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40097, CurrentID=0, Status='live', Symbol='CG9629', Aliases='Dmel_ Dmel\\CG9629', description='CG9629 gene product from transcript CG9629-RA', other_designations='CG9629-PA', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19266473, end_position_on_the_genomic_accession=19268913, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40098, CurrentID=0, Status='live', Symbol='CG14088', Aliases='Dmel_ Dmel\\CG14088, SPH199', description='CG14088 gene product from transcript CG14088-RB', other_designations='CG14088-PB|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19269772, end_position_on_the_genomic_accession=19270642, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40099, CurrentID=0, Status='live', Symbol='CG14085', Aliases='Dmel_ Dmel\\CG14085', description='CG14085 gene product from transcript CG14085-RA', other_designations='CG14085-PA|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19269351, end_position_on_the_genomic_accession=19274168, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40100, CurrentID=0, Status='live', Symbol='CG14086', Aliases='Dmel_ Dmel\\CG14086', description='CG14086 gene product from transcript CG14086-RD', other_designations='CG14086-PD|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19275791, end_position_on_the_genomic_accession=19279919, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40101, CurrentID=0, Status='live', Symbol='CG14089', Aliases='Dmel_ Dmel\\CG14089', description='CG14089 gene product from transcript CG14089-RA', other_designations='CG14089-PA|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19280120, end_position_on_the_genomic_accession=19280800, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40102, CurrentID=0, Status='live', Symbol='Gbs-76A', Aliases='Dmel_CG9619, 149039_at, CG9619, Dmel\\CG9619, GBS-76A', description='Glycogen binding subunit 76A', other_designations='CG9619-PA|CG9619-PB|CG9619-PC|CG9619-PD|Gbs-76A-PA|Gbs-76A-PB|Gbs-76A-PC|Gbs-76A-PD|glycogen binding subunit 76A', map_location='76A3-76A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19281364, end_position_on_the_genomic_accession=19292766, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40104, CurrentID=0, Status='live', Symbol='pip', Aliases='Dmel_CG9614, CG9614, Dmel\\CG9614, Pipe, l(3)76BDpe-ST2, pip', description='pipe', other_designations='CG9614-PA|CG9614-PC|CG9614-PD|CG9614-PF|CG9614-PG|CG9614-PH|CG9614-PI|CG9614-PJ|CG9614-PK|CG9614-PL|CG9614-PM|CG9614-PN|pip-PA|pip-PC|pip-PD|pip-PF|pip-PG|pip-PH|pip-PI|pip-PJ|pip-PK|pip-PL|pip-PM|pip-PN', map_location='3-47 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19298623, end_position_on_the_genomic_accession=19337609, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  40105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40105, CurrentID=0, Status='live', Symbol='CG32206', Aliases='Dmel_ CG14090, CG14091, CG14092, CG9600, Dmel\\CG32206', description='CG32206 gene product from transcript CG32206-RE', other_designations='CG32206-PB|CG32206-PE|uncharacterized protein', map_location='76A6-76B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19340321, end_position_on_the_genomic_accession=19412912, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  40109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40109, CurrentID=0, Status='live', Symbol='ms(3)76Ba', Aliases='Dmel_CG14087, CG14087, Dmel\\CG14087', description='male sterile (3) 76Ba', other_designations='CG14087-PC|ms(3)76Ba-PC', map_location='76B1-76B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19387580, end_position_on_the_genomic_accession=19390033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40111, CurrentID=0, Status='live', Symbol='Chd3', Aliases='Dmel_CG9594, CG9594, CHD-3, CHD3, DmCHD3, Dmel\\CG9594, chd3, dCHD3', description='CG9594 gene product from transcript CG9594-RA', other_designations='CG9594-PA|Chd3-PA', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19447173, end_position_on_the_genomic_accession=19450238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40112, CurrentID=0, Status='live', Symbol='CG14095', Aliases='Dmel_ Dmel\\CG14095', description='CG14095 gene product from transcript CG14095-RA', other_designations='CG14095-PA|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19464049, end_position_on_the_genomic_accession=19464759, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40113, CurrentID=0, Status='live', Symbol='CG14096', Aliases='Dmel_ Dmel\\CG14096', description='CG14096 gene product from transcript CG14096-RA', other_designations='CG14096-PA|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19466567, end_position_on_the_genomic_accession=19467079, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40114, CurrentID=0, Status='live', Symbol='CG12519', Aliases='Dmel_ Dmel\\CG12519', description='CG12519 gene product from transcript CG12519-RB', other_designations='CG12519-PA|CG12519-PB|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19472491, end_position_on_the_genomic_accession=19473145, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40115, CurrentID=0, Status='live', Symbol='CG18294', Aliases='Dmel_ BcDNA:RE51966, Dmel\\CG18294', description='CG18294 gene product from transcript CG18294-RA', other_designations='CG18294-PA|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19475017, end_position_on_the_genomic_accession=19475578, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40116, CurrentID=0, Status='live', Symbol='brv1', Aliases='Dmel_CG9472, CG9472, CT26822, Dmel\\CG9472', description='brivido-1', other_designations='CG9472-PA|brv1-PA', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19483826, end_position_on_the_genomic_accession=19486498, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40117, CurrentID=0, Status='live', Symbol='CG9449', Aliases='Dmel_ Dmel\\CG9449', description='CG9449 gene product from transcript CG9449-RH', other_designations='CG9449-PE|CG9449-PF|CG9449-PG|CG9449-PH|CG9449-PJ|CG9449-PK', map_location='76B5-76B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19492053, end_position_on_the_genomic_accession=19497723, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40118, CurrentID=0, Status='live', Symbol='CG9451', Aliases='Dmel_ Dmel\\CG9451', description='CG9451 gene product from transcript CG9451-RB', other_designations='CG9451-PA|CG9451-PB', map_location='76B5-76B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19498197, end_position_on_the_genomic_accession=19500798, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40119, CurrentID=0, Status='live', Symbol='CG9452', Aliases='Dmel_ Dmel\\CG9452', description='CG9452 gene product from transcript CG9452-RA', other_designations='CG9452-PA', map_location='76B5-76B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19501010, end_position_on_the_genomic_accession=19503870, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40120, CurrentID=0, Status='live', Symbol='Cpr76Ba', Aliases='Dmel_CG9283, CG9283, DmelCpr76Ba, Dmel\\CG9283', description='Cuticular protein 76Ba', other_designations='CG9283-PA|Cpr76Ba-PA|cuticular protein 76Ba', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19517552, end_position_on_the_genomic_accession=19518332, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40121, CurrentID=0, Status='live', Symbol='Cpr76Bb', Aliases='Dmel_CG9290, CG9290, DmelCpr76Bb, Dmel\\CG9290', description='Cuticular protein 76Bb', other_designations='CG9290-PA|Cpr76Bb-PA|cuticular protein 76Bb', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19519649, end_position_on_the_genomic_accession=19520419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40122, CurrentID=0, Status='live', Symbol='Cpr76Bc', Aliases='Dmel_CG9295, CG9295, DmelCpr76Bc, Dmel\\CG9295', description='Cuticular protein 76Bc', other_designations='CG9295-PC|CG9295-PD|Cpr76Bc-PC|Cpr76Bc-PD|cuticular protein 76Bc', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19520945, end_position_on_the_genomic_accession=19527272, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40123, CurrentID=0, Status='live', Symbol='Cpr76Bd', Aliases='Dmel_CG9299, CG9299, DmelCpr76Bd, Dmel\\CG9299', description='Cuticular protein 76Bd', other_designations='CG9299-PB|CG9299-PC|CG9299-PD|Cpr76Bd-PB|Cpr76Bd-PC|Cpr76Bd-PD|cuticular protein 76Bd', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19532743, end_position_on_the_genomic_accession=19538289, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40124, CurrentID=0, Status='live', Symbol='CG9279', Aliases='Dmel_ Dmel\\CG9279', description='CG9279 gene product from transcript CG9279-RB', other_designations='CG9279-PB|CG9279-PC', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19538425, end_position_on_the_genomic_accession=19544172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40126, CurrentID=0, Status='live', Symbol='Rcd7', Aliases='Dmel_CG14098, CG14098, Dmel\\CG14098', description='Reduction in Cnn dots 7', other_designations='CG14098-PA|CG14098-PB|Rcd7-PA|Rcd7-PB|reduction in Cnn dots 7', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19545110, end_position_on_the_genomic_accession=19547293, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40127, CurrentID=0, Status='live', Symbol='Ltn1', Aliases='Dmel_CG32210, CG32210, CG9268, CG9274, Dmel\\CG32210, l(3)76BDr', description='Listerin E3 ubiquitin protein ligase 1', other_designations='CG32210-PA|Ltn1-PA|l(3)76BDr|listerin E3 ubiquitin protein ligase 1', map_location='76B6-76B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19544345, end_position_on_the_genomic_accession=19552290, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  40128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40128, CurrentID=0, Status='live', Symbol='CG9300', Aliases='Dmel_ Dmel\\CG9300', description='CG9300 gene product from transcript CG9300-RA', other_designations='CG9300-PA', map_location='76B7-76B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19552420, end_position_on_the_genomic_accession=19554767, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40129, CurrentID=0, Status='live', Symbol='Shal', Aliases='Dmel_CG9262, CG9262, DNK[[v]]4, Dmel\\CG9262, Kv4, Sh-l, Shl, dShal, shal', description='Shaker cognate l', other_designations='CG9262-PA|CG9262-PB|CG9262-PC|Shaker-like|Shal-PA|Shal-PB|Shal-PC|shaker cognate 1|shaker cognate L|shaker cognate l', map_location='76B7-76B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19566851, end_position_on_the_genomic_accession=19584568, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40130, CurrentID=0, Status='live', Symbol='CG9231', Aliases='Dmel_ Dmel\\CG9231', description='CG9231 gene product from transcript CG9231-RB', other_designations='CG9231-PA|CG9231-PB', map_location='76B8-76B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19585093, end_position_on_the_genomic_accession=19586003, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40131, CurrentID=0, Status='live', Symbol='CG9330', Aliases='Dmel_ Dmel\\CG9330', description='CG9330 gene product from transcript CG9330-RA', other_designations='CG9330-PA', map_location='76B8-76B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19586253, end_position_on_the_genomic_accession=19589111, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40132, CurrentID=0, Status='live', Symbol='CG14100', Aliases='Dmel_ Dmel\\CG14100', description='CG14100 gene product from transcript CG14100-RA', other_designations='CG14100-PA|uncharacterized protein', map_location='76B8-76B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19589072, end_position_on_the_genomic_accession=19590524, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40133, CurrentID=0, Status='live', Symbol='ash1', Aliases='Dmel_CG8887, ASH1, Ash, Ash 1, Ash1, CG8887, Dmel\\CG8887, ash, ash-1, dash, l(3)76BDb, l(3)SG29a', description='absent, small, or homeotic discs 1', other_designations='Absent, small or homeotic discs 1|Absent, small, or homeotic disc 1|CG8887-PB|CG8887-PC|absent small or homeotic 1|ash1-PB|ash1-PC|discs absent, small, or homeotic|discs absent, small, or homeotic-1', map_location='3-46.6 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19590773, end_position_on_the_genomic_accession=19598755, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40134, CurrentID=0, Status='live', Symbol='Taf6', Aliases='Dmel_CG32211, CG32211, CG9348, Dmel\\CG32211, SR3-4B, TAF, TAF6, TAF60, TAF60/62, TAFII60, TAFII60(62), TAFII62, TAF[[II]], TAF[[II]]60, TAF[[II]]60/62, TAF[[II]]62, TFIID, TFIID 62, Taf-60, Taf62, Taf[[II]]60, d60, dTAF6, dTAF62, dTAFII62, dTAF[[II]]262, dTAF[[II]]60, dTAF[[II]]62, dmTAF6, dme-TAFII60, l(3)76BDf, p62, taf6, taf60, Taf6', description='TBP-associated factor 6', other_designations='CG32211-PA|CG32211-PB|TBP associated factor 60|TBP-Associated factor 6|TBP-associated factor|TBP-associated factor 60|TBP-associated factor 60kD|Taf6-PA|Taf6-PB|suppressor of Ras85D 3-4B', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19599088, end_position_on_the_genomic_accession=19601382, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40135, CurrentID=0, Status='live', Symbol='CG9368', Aliases='Dmel_ BcDNA:RH51268, Dmel\\CG9368', description='CG9368 gene product from transcript CG9368-RD', other_designations='CG9368-PA|CG9368-PB|CG9368-PD', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19601633, end_position_on_the_genomic_accession=19604120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40136, CurrentID=0, Status='live', Symbol='lush', Aliases='Dmel_CG8807, 76a, 76c, 76c(LUSH), CG8807, DmelObp76a, Dmel\\CG8807, LUSH, Lush, Obp76a, obp76a', description='CG8807 gene product from transcript CG8807-RA', other_designations='CG8807-PA|CG8807-PB|lush-PA|lush-PB', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19604398, end_position_on_the_genomic_accession=19606090, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40137, CurrentID=0, Status='live', Symbol='CG9372', Aliases='Dmel_ Dmel\\CG9372, SP34, c-SP34', description='CG9372 gene product from transcript CG9372-RA', other_designations='CG9372-PA', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19606208, end_position_on_the_genomic_accession=19608368, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40138, CurrentID=0, Status='live', Symbol='Lon', Aliases='Dmel_CG8798, CG8798, Dmel\\CG8798, anon-WO0118547.350', description='Lon protease', other_designations='CG8798-PA|CG8798-PC|Lon-PA|Lon-PC', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19607857, end_position_on_the_genomic_accession=19612199, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40139, CurrentID=0, Status='live', Symbol='CG9376', Aliases='Dmel_ Dmel\\CG9376', description='CG9376 gene product from transcript CG9376-RA', other_designations='CG9376-PA', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19612455, end_position_on_the_genomic_accession=19613489, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40140, CurrentID=0, Status='live', Symbol='l(3)76BDm', Aliases='Dmel_CG8793, CG8793, Dmel\\CG8793, l(3)L3809', description='lethal (3) 76BDm', other_designations='CG8793-PA|CG8793-PB|l(3)76BDm-PA|l(3)76BDm-PB|lethal (3) L3809', map_location='76B9-76B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19613427, end_position_on_the_genomic_accession=19618521, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40141, CurrentID=0, Status='live', Symbol='asf1', Aliases='Dmel_CG9383, ASF1, Asf1, CG9383, Dmel\\CG9383, dASF/SF2, dASF1, dAsf1, l(3)76BDq', description='anti-silencing factor 1', other_designations='CG9383-PA|CG9383-PB|anti- silencing function 1|anti-silencing factor|asf1-PA|asf1-PB', map_location='3-46.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19618894, end_position_on_the_genomic_accession=19620398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40142, CurrentID=0, Status='live', Symbol='ms(3)76Cc', Aliases='Dmel_CG9392, CG9392, Dmel\\CG9392', description='male sterile (3) 76Cc', other_designations='CG9392-PA|ms(3)76Cc-PA', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19620502, end_position_on_the_genomic_accession=19624309, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40143, CurrentID=0, Status='live', Symbol='wnd', Aliases='Dmel_CG8789, CG8789, DLK, DLK/ZPK, Dmel\\CG8789, MAPKKK, Wnd, Wnd/DLK', description='wallenda', other_designations='CG8789-PA|CG8789-PB|CG8789-PC|CG8789-PD|dual leucine zipper-bearing kinase|wnd-PA|wnd-PB|wnd-PC|wnd-PD', map_location='76B9-76B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19624406, end_position_on_the_genomic_accession=19635783, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40144, CurrentID=0, Status='live', Symbol='CG8786', Aliases='Dmel_ BcDNA:GM01527, DRNF146, Dmel\\CG8786, RNF146', description='CG8786 gene product from transcript CG8786-RG', other_designations='CG8786-PB|CG8786-PC|CG8786-PF|CG8786-PG', map_location='76B10-76B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19637207, end_position_on_the_genomic_accession=19642816, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40145, CurrentID=0, Status='live', Symbol='Oat', Aliases='Dmel_CG8782, 151832_at, CG8782, Dmel\\CG8782', description='Ornithine aminotransferase precursor', other_designations='CG8782-PA|Oat-PA|ornithine aminotransferase|ornithine aminotransferase precursor', map_location='76B11-76B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19647280, end_position_on_the_genomic_accession=19649851, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40146, CurrentID=0, Status='live', Symbol='tey', Aliases='Dmel_CG8780, CG8780, Dmel\\CG8780', description='teyrha-meyrha', other_designations='CG8780-PA|CG8780-PB|CG8780-PC|CG8780-PD|tey-PA|tey-PB|tey-PC|tey-PD|teyrha meyrha|teyrha-meyhra', map_location='76B11-76C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19650506, end_position_on_the_genomic_accession=19664104, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40147, CurrentID=0, Status='live', Symbol='CG8765', Aliases='Dmel_ Dmel\\CG8765', description='CG8765 gene product from transcript CG8765-RB', other_designations='CG8765-PA|CG8765-PB|CG8765-PC', map_location='76C1-76C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19683941, end_position_on_the_genomic_accession=19687353, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40148, CurrentID=0, Status='live', Symbol='Gbeta76C', Aliases='Dmel_CG8770, CG8770, DGqbeta, Dmel\\CG8770, G&bgr76C, G[[beta76C]], G[[beta]], G[[beta]]e, G[q], Galpha76C, Gbe, Gbeta, Gbeta76c, Gbeta[[e]], Gbetae, Gbeta{e}, Gqbeta, eGbeta, eye Gbeta, gbe', description='G protein beta-subunit 76C', other_designations='CG8770-PA|G[[beta]] eye|Gbeta76C-PA|guanine nucleotide regulatory protein beta subunit', map_location='76C1-76C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19687637, end_position_on_the_genomic_accession=19689992, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40149, CurrentID=0, Status='live', Symbol='verm', Aliases='Dmel_CG8756, CG8756, DmCDA2, Dmel\\CG8756, LCBP1, N63, Verm, hlm, l(3)76BDx', description='vermiform', other_designations='CG8756-PG|CG8756-PH|CG8756-PI|LDLa domain containing chitin binding protein 1|helmsman|verm-PG|verm-PH|verm-PI', map_location='76C2-76C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19690033, end_position_on_the_genomic_accession=19695260, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40150, CurrentID=0, Status='live', Symbol='serp', Aliases='Dmel_CG32209, CG32209, CG8747, CG8748, DmCDA1, Dmel\\CG32209, Serp, l(3)76BDk', description='serpentine', other_designations='CG32209-PB|CG32209-PC|l(3)76BDk|serp-PB|serp-PC', map_location='76C2-76C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19700977, end_position_on_the_genomic_accession=19708015, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40152, CurrentID=0, Status='live', Symbol='Trpml', Aliases='Dmel_CG8743, CG8743, CT25240, DmTRPML, Dmel\\CG8743, TRPML, l(3)76BDi, trpml', description='Transient receptor potential cation channel, mucolipin ortholog', other_designations='CG8743-PA|TRP mucolipin|Trpml-PA|l(3)76BDi|transient receptor potential cation channel, mucolipin ortholog|transient receptor potential mucolipin', map_location='76C2-76C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19714418, end_position_on_the_genomic_accession=19718328, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40153, CurrentID=0, Status='live', Symbol='CG42637', Aliases='Dmel_ Dmel\\CG42637, Dmel_CG8742', description='CG42637 gene product from transcript CG42637-RB', other_designations='CG42637-PA|CG42637-PB|CG42637-PC|uncharacterized protein', map_location='76C2-76C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19719527, end_position_on_the_genomic_accession=19793569, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  40154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40154, CurrentID=0, Status='live', Symbol='ms(3)76Ca', Aliases='Dmel_CG14101, CG14101, Dmel\\CG14101, anon-WO0140519.34', description='male sterile (3) 76Ca', other_designations='CG14101-PA|ms(3)76Ca-PA', map_location='76C3-76C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19760409, end_position_on_the_genomic_accession=19761817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40155, CurrentID=0, Status='live', Symbol='SREBP', Aliases='Dmel_CG8522, CG8522, Dmel\\CG8522, HLH106, HlH106, SREBF1, Srebp, dSREBP, dSrebp, dsrebp, l(3)76BDw, srebp', description='Sterol regulatory element binding protein', other_designations='CG8522-PA|CG8522-PB|CG8522-PC|CG8522-PD|SREBP-PA|SREBP-PB|SREBP-PC|SREBP-PD|drosophila sterol regulatory element binding protein|helix loop helix protein 106|lethal (3) 76BDw|sterol regulatory element binding protein|sterol regulatory element-binding protein|sterol-responsive element-binding protein|transcription factor sterol regulatory element binding protein', map_location='76C5-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19793935, end_position_on_the_genomic_accession=19798979, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40156, CurrentID=0, Status='live', Symbol='CG14102', Aliases='Dmel_ Dmel\\CG14102', description='CG14102 gene product from transcript CG14102-RA', other_designations='CG14102-PA|uncharacterized protein', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19798975, end_position_on_the_genomic_accession=19801291, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40157, CurrentID=0, Status='live', Symbol='Ir76a', Aliases='Dmel_CG42584, CG34257, CG42584, CG8533, CT24909, DmelIR76a, Dmel\\CG42584, Dmel_CG34257, Dmel_CG8533, IR76a, ir76a', description='Ionotropic receptor 76a', other_designations='CG42584-PA|CG42584-PB|CG42584-PD|CG42584-PE|Ir76a-PA|Ir76a-PB|Ir76a-PD|Ir76a-PE|ionotropic receptor 76a', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19799278, end_position_on_the_genomic_accession=19803427, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40158, CurrentID=0, Status='live', Symbol='CG14103', Aliases='Dmel_ Dmel\\CG14103', description='CG14103 gene product from transcript CG14103-RB', other_designations='CG14103-PA|CG14103-PB|uncharacterized protein', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19803738, end_position_on_the_genomic_accession=19805080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40159, CurrentID=0, Status='live', Symbol='CG15881', Aliases='Dmel_ BcDNA:HL01254, Ccdc58, Dmel\\CG15881', description='CG15881 gene product from transcript CG15881-RB', other_designations='CG15881-PA|CG15881-PB|CG15881-PC|uncharacterized protein', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19805181, end_position_on_the_genomic_accession=19806932, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40160, CurrentID=0, Status='live', Symbol='Kap-alpha1', Aliases='Dmel_CG8548, CG8548, Dalpha1, Dmalpha1, Dmel\\CG8548, IMPalpha1, Imp-alpha1, Kap alpha1, Kapalpha1S1, Kpna1, alphaKap1, anon-WO0118547.276, imp alpha1, imp-alpha1, impalpha1', description='karyopherin alpha1', other_designations='CG8548-PA|Kap-alpha1-PA|importin alpha1|importin-alpha1|karyopherin alpha 1|karyopherin-alpha1', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19807232, end_position_on_the_genomic_accession=19810398, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40161, CurrentID=0, Status='live', Symbol='Cyp305a1', Aliases='Dmel_CG8733, 305a1, AAF49108.1, CG8733, CYP305A1-Dm, Cyp305A1, Dmel\\CG8733', description='CG8733 gene product from transcript CG8733-RA', other_designations='CG8733-PA|Cyp305a1-PA|cytochrome P450|cytochrome P450 305a1', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19810403, end_position_on_the_genomic_accession=19813339, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40162, CurrentID=0, Status='live', Symbol='cyc', Aliases='Dmel_CG8727, BMAL1, Bmal1, CG8727, CYC, CYCLE, Cyc, Cycle, Dmel\\CG8727, MOP3, Mop3, bHLHe7, bMAL1, bmal1, dBMAL, dBMAL1, dBma1, dBmal, dCYC, dbmal, dbmal1', description='cycle', other_designations='CG8727-PA|brain and muscle ARNT-like 1|cyc-PA', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19813770, end_position_on_the_genomic_accession=19815938, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40163, CurrentID=0, Status='live', Symbol='Fibp', Aliases='Dmel_CG8660, CG8660, Dmel\\CG8660, DrFIBP, F1bp', description='CG8660 gene product from transcript CG8660-RF', other_designations='CG8660-PB|CG8660-PC|CG8660-PE|CG8660-PF|CG8660-PG|Fibp-PB|Fibp-PC|Fibp-PE|Fibp-PF|Fibp-PG|fibp', map_location='76D1-76D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19816085, end_position_on_the_genomic_accession=19817967, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40164, CurrentID=0, Status='live', Symbol='Deaf1', Aliases='Dmel_CG8567, CG8567, DEAF-1, DEAF1, Deaf-1, Dmel\\CG8567, deaf1, l(3)76BDv', description='Deformed epidermal autoregulatory factor-1', other_designations='CG8567-PA|CG8567-PB|CG8567-PC|CG8567-PD|CG8567-PE|Deaf1-PA|Deaf1-PB|Deaf1-PC|Deaf1-PD|Deaf1-PE|deaf-1|deformed epidermal autoregulatory factor-1', map_location='76D1-76D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19818174, end_position_on_the_genomic_accession=19830686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40165, CurrentID=0, Status='live', Symbol='trc', Aliases='Dmel_CG8637, 0669/17, CG8637, Dmel\\CG8637, NDR, Ndr, RE-D, Trc, dmTRC, l(3)76BDa, l(3)S066917', description='tricornered', other_designations='CG8637-PA|CG8637-PB|nuclear Dbf2-related protein kinase|trc-PA|trc-PB|tricorner', map_location='3-46 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19830864, end_position_on_the_genomic_accession=19834304, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40167, CurrentID=0, Status='live', Symbol='Papss', Aliases='Dmel_CG8363, CG8363, DmPAPSS, Dmel\\CG8363, PAPS, PAPSS, Paps, l(3)76BDn, paps, papss', description='PAPS synthetase', other_designations='CG8363-PA|CG8363-PB|CG8363-PC|CG8363-PD|CG8363-PE|CG8363-PF|CG8363-PG|CG8363-PH|PAPs synthase|Papss-PA|Papss-PB|Papss-PC|Papss-PD|Papss-PE|Papss-PF|Papss-PG|Papss-PH|paps synthetase', map_location='76D1-76D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19845293, end_position_on_the_genomic_accession=19860925, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40168, CurrentID=0, Status='live', Symbol='Rab8', Aliases='Dmel_CG8287, AAF49101, CG8287, DRAB8, D Dm Rab8, DmRab8, Dmel\\CG8287, O18338, Rab-r8, dm-Rab8, l(3)76BDu, rab8', description='CG8287 gene product from transcript CG8287-RA', other_designations='CG8287-PA|CG8287-PB|CG8287-PC|Rab-protein 8|Rab8-PA|Rab8-PB|Rab8-PC|l(3)76BDu', map_location='76D2-76D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19867488, end_position_on_the_genomic_accession=19871078, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40169, CurrentID=0, Status='live', Symbol='Usp32', Aliases='Dmel_CG8334, CG8334, Dmel\\CG8334, Q9VW49, USP15', description='Ubiquitin specific protease 32', other_designations='CG8334-PB|CG8334-PC|CG8334-PD|CG8334-PE|Usp32-PB|Usp32-PC|Usp32-PD|Usp32-PE|ubiquitin specific protease 32', map_location='76D3-76D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19871139, end_position_on_the_genomic_accession=19881783, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  40170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40170, CurrentID=0, Status='live', Symbol='Mi-2', Aliases='Dmel_CG8103, 0006/06, 0854/01, CG8103, Chd4, Dm- Dmel\\CG8103, M-i2, MI-2, MI2, Mi 2, Mi2, Pha, dMI-2, dMi, dMi-2, dMi2, hip76, l(3)01058, l(3)76BDe, l(3)A154.3M3, l(3)L1243, l(3)S000606, l(3)S085401, l(3)j3D4, pha', description='CG8103 gene product from transcript CG8103-RA', other_designations='CG8103-PA|CG8103-PB|CG8103-PC|CG8103-PD|Mi-2 ortholog|Mi-2-PA|Mi-2-PB|Mi-2-PC|Mi-2-PD|PHD-containing ATPase|dMi-2|hb-interacting protein 76|lethal (3) L1243', map_location='76D3-76D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19882051, end_position_on_the_genomic_accession=19910215, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40171, CurrentID=0, Status='live', Symbol='Su(Tpl)', Aliases='Dmel_CG32217, 0854/01, CG32217, CG8037, CR32216, Dmel\\CG32217, ELL, Ell, SR3-4A, SS3-4, SY3-1/, dELL, l (3)76BD, l(3)76BDs, l(3)S000606, l(3)S085401', description='CG32217 gene product from transcript CG32217-RB', other_designations='CG32217-PA|CG32217-PB|CG32217-PC|Su(Tpl)-PA|Su(Tpl)-PB|Su(Tpl)-PC|eleven-nineteen lysine-rich leukemia|elongation factor RNA polymerase II|suppressor of GMR-sina 3-4|suppressor of triplolethal', map_location='3-46.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19889120, end_position_on_the_genomic_accession=19909410, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40172, CurrentID=0, Status='live', Symbol='Prp3', Aliases='Dmel_CG7757, CG7757, Dmel\\CG7757, PRPF3, dPrp3', description='Precursor RNA processing 3', other_designations='CG7757-PA|CG7757-PB|Prp3-PA|Prp3-PB|precursor RNA processing 3', map_location='76D4-76D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19910300, end_position_on_the_genomic_accession=19912504, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40173, CurrentID=0, Status='live', Symbol='Mtr3', Aliases='Dmel_CG8025, CG8025, Dmel\\CG8025, d mtr3', description='CG8025 gene product from transcript CG8025-RA', other_designations='CG8025-PA|Mtr3-PA', map_location='76D4-76D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19912405, end_position_on_the_genomic_accession=19913715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40174, CurrentID=0, Status='live', Symbol='Rpn1', Aliases='Dmel_CG7762, BcDNA:LD21723, CG7762, Dmel\\CG7762, RPN1, S2, p97', description='Regulatory particle non-ATPase 1', other_designations='CG7762-PA|Rpn1-PA|regulatory particle non-ATPase 1', map_location='76D4-76D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19915051, end_position_on_the_genomic_accession=19918160, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40176, CurrentID=0, Status='live', Symbol='CG7770', Aliases='Dmel_ BcDNA:GH28557, Dmel\\CG7770', description='CG7770 gene product from transcript CG7770-RB', other_designations='CG7770-PA|CG7770-PB', map_location='76D4-76D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19923671, end_position_on_the_genomic_accession=19924688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40177, CurrentID=0, Status='live', Symbol='Grasp65', Aliases='Dmel_CG7809, CG7809, Dmel\\CG7809, GRASP, GRASP65, dGRASP, d dgrasp', description='CG7809 gene product from transcript CG7809-RA', other_designations='CG7809-PA|Grasp65-PA', map_location='76D5-76D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19924943, end_position_on_the_genomic_accession=19927175, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40178, CurrentID=0, Status='live', Symbol='CG8004', Aliases='Dmel_ Dmel\\CG8004', description='CG8004 gene product from transcript CG8004-RA', other_designations='CG8004-PA', map_location='76D5-76D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19927194, end_position_on_the_genomic_accession=19928508, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40179, CurrentID=0, Status='live', Symbol='RhoGDI', Aliases='Dmel_CG7823, CG7823, Dmel\\CG7823', description='CG7823 gene product from transcript CG7823-RA', other_designations='CG7823-PA|CG7823-PB|RhoGDI-PA|RhoGDI-PB', map_location='76D5-76D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19928743, end_position_on_the_genomic_accession=19932930, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40180, CurrentID=0, Status='live', Symbol='Ac76E', Aliases='Dmel_CG7978, AC 76E, AC76E, CG7978, DAC2, DAC76E, Dmel\\CG7978, ac76e', description='Adenylyl cyclase 76E', other_designations='Ac76E-PA|Ac76E-PB|AdenylateCyclase E|CG7978-PA|CG7978-PB|adenylyl cyclase|adenylyl cyclase 76E', map_location='76D6-76D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19939287, end_position_on_the_genomic_accession=19977960, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  40181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40181, CurrentID=0, Status='live', Symbol='CG14182', Aliases='Dmel_ Dmel\\CG14182', description='CG14182 gene product from transcript CG14182-RD', other_designations='CG14182-PA|CG14182-PB|CG14182-PC|CG14182-PD|uncharacterized protein', map_location='76D7-76D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19978713, end_position_on_the_genomic_accession=19979812, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40182, CurrentID=0, Status='live', Symbol='CG17732', Aliases='Dmel_ Dmel\\CG17732', description='CG17732 gene product from transcript CG17732-RB', other_designations='CG17732-PB|uncharacterized protein', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19981685, end_position_on_the_genomic_accession=19984191, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40183, CurrentID=0, Status='live', Symbol='hale', Aliases='Dmel_CG7570, CG7570, Dmel\\CG7570', description='hale-bopp', other_designations='CG7570-PA|hale-PA', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19984426, end_position_on_the_genomic_accession=19985868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40184, CurrentID=0, Status='live', Symbol='schuy', Aliases='Dmel_CG17736, CG17736, Dmel\\CG17736', description='schumacher-levy', other_designations='CG17736-PA|schuy-PA', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19986093, end_position_on_the_genomic_accession=19988072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40185, CurrentID=0, Status='live', Symbol='fln', Aliases='Dmel_CG7445, CG7445, Dmel\\CG7445, Fln, Ftn, mp27', description='flightin', other_designations='CG7445-PA|CG7445-PB|fln-PA|fln-PB', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19989551, end_position_on_the_genomic_accession=19991220, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40186, CurrentID=0, Status='live', Symbol='Nca', Aliases='Dmel_CG7641, CG7641, Dmel\\CG7641lc, nca, Nca', description='Neurocalcin', other_designations='CG7641-PA|Nca-PA|neurocalcin', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19991887, end_position_on_the_genomic_accession=19997568, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40187, CurrentID=0, Status='live', Symbol='CG7646', Aliases='Dmel_ BcDNA:RH69723, Dmel\\CG7646, anon-EST:Posey160', description='CG7646 gene product from transcript CG7646-RC', other_designations='CG7646-PA|CG7646-PB|CG7646-PC', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19991887, end_position_on_the_genomic_accession=19996269, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40188, CurrentID=0, Status='live', Symbol='CG7433', Aliases='Dmel_ 149085_at, Dmel\\CG7433', description='CG7433 gene product from transcript CG7433-RC', other_designations='CG7433-PA|CG7433-PB|CG7433-PC|CG7433-PD', map_location='76D8-76E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19997714, end_position_on_the_genomic_accession=20000686, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40189, CurrentID=0, Status='live', Symbol='Tom20', Aliases='Dmel_CG7654, CG7654, Dmel\\CG7654, TOM 20, TOM20, anon-WO0118547.225, dtom20', description='Translocase of outer membrane 20', other_designations='CG7654-PA|CG7654-PB|Tom20-PA|Tom20-PB|translocase of outer membrane 20', map_location='76E1-76E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20000644, end_position_on_the_genomic_accession=20002446, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40190, CurrentID=0, Status='live', Symbol='CG7668', Aliases='Dmel_ AAF49079, AAM11109, AAN11645, BcDNA:GM07843, BcDNA:RE40293, Dmel\\CG7668, FBpp0074673, FBpp0074674', description='CG7668 gene product from transcript CG7668-RC', other_designations='CG7668-PA|CG7668-PC|Fibrinogen-related protein', map_location='76E1-76E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20002736, end_position_on_the_genomic_accession=20004834, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40191, CurrentID=0, Status='live', Symbol='kug', Aliases='Dmel_CG7749, CG7749, CT23445, Cad76E, DCad76E, Dmel\\CG7749, Ftl, RE-A, fat2, ftlel, kug', description='kugelei', other_designations='CG7749-PC|CG7749-PD|CG7749-PE|fat-like|fat2|kug-PC|kug-PD|kug-PE', map_location='3-47 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20005301, end_position_on_the_genomic_accession=20024607, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  40192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40192, CurrentID=0, Status='live', Symbol='CG14183', Aliases='Dmel_ Dmel\\CG14183', description='CG14183 gene product from transcript CG14183-RA', other_designations='CG14183-PA|uncharacterized protein', map_location='76E1-76E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20024687, end_position_on_the_genomic_accession=20027835, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40193, CurrentID=0, Status='live', Symbol='CG14184', Aliases='Dmel_ Dmel\\CG14184, anon-WO0118547.633, dp18, p18', description='CG14184 gene product from transcript CG14184-RA', other_designations='CG14184-PA|CG14184-PB|uncharacterized protein', map_location='76E1-76E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20027819, end_position_on_the_genomic_accession=20029020, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40195, CurrentID=0, Status='live', Symbol='sNPF-R', Aliases='Dmel_CG7395, CG18639, CG7395, CG7395/CG18639, CT22771, Dmel\\CG7395, Drm-NPFR76F, Drm-sNPF-R, GPCR60, NPFR76F, anon-WO0131005.1, sNPFR, sNPFR1, snpfr', description='short neuropeptide F receptor', other_designations='CG7395-PA|CG7395-PB|neuropeptide F-like receptor 76F|neuropeptide-Y-receptor|sNPF receptor|sNPF-R-PA|sNPF-R-PB', map_location='76E2-76E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20058511, end_position_on_the_genomic_accession=20089007, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40196, CurrentID=0, Status='live', Symbol='CG14186', Aliases='Dmel_ Dmel\\CG14186', description='CG14186 gene product from transcript CG14186-RA', other_designations='CG14186-PA|uncharacterized protein', map_location='76F1-76F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20098084, end_position_on_the_genomic_accession=20103827, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40197, CurrentID=0, Status='live', Symbol='CG14185', Aliases='Dmel_ CT33794, Dmel\\CG14185', description='CG14185 gene product from transcript CG14185-RA', other_designations='CG14185-PA|uncharacterized protein', map_location='76F1-76F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20106094, end_position_on_the_genomic_accession=20107557, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40198, CurrentID=0, Status='live', Symbol='Ir76b', Aliases='Dmel_CG7385, CG7385, CT22733, DmelIR76b, Dmel\\CG7385, IR76b, ir76b', description='Ionotropic receptor 76b', other_designations='CG7385-PA|Ir76b-PA|ionotropic receptor 76b', map_location='76F1-76F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20108484, end_position_on_the_genomic_accession=20111534, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40199, CurrentID=0, Status='live', Symbol='CG14187', Aliases='Dmel_ Dmel\\CG14187', description='CG14187 gene product from transcript CG14187-RB', other_designations='CG14187-PA|CG14187-PB|uncharacterized protein', map_location='76F1-76F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20112265, end_position_on_the_genomic_accession=20112847, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40200, CurrentID=0, Status='live', Symbol='CG7365', Aliases='Dmel_ Dmel\\CG7365', description='CG7365 gene product from transcript CG7365-RA', other_designations='CG7365-PA|CG7365-PB', map_location='76F2-76F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20124298, end_position_on_the_genomic_accession=20127381, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40201, CurrentID=0, Status='live', Symbol='gig', Aliases='Dmel_CG6975, 6975, C1, CG6975, Dmel\\CG6975, Gigas, ME 109, TSC2, Tsc2, Tsc2/gig, dTSC2, dTsc2|Tsc2, l(3)109, tsc2, gig', description='gigas', other_designations='CG6975-PA|CG6975-PB|gig-PA|gig-PB|lethal(3)109|tuberous sclerosis complex 2', map_location='3-46 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20129181, end_position_on_the_genomic_accession=20139010, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  40202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40202, CurrentID=0, Status='live', Symbol='obst-J', Aliases='Dmel_CG7348, CG7348, Dmel\\CG7348, anon-EST:ParkEST360', description='obstructor-J', other_designations='CG7348-PA|obst-J-PA', map_location='76F2-76F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20130451, end_position_on_the_genomic_accession=20131743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40203, CurrentID=0, Status='live', Symbol='CG7335', Aliases='Dmel_ BcDNA:AT09463, Dmel\\CG7335', description='CG7335 gene product from transcript CG7335-RA', other_designations='CG7335-PA', map_location='76F2-76F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20132032, end_position_on_the_genomic_accession=20133268, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40204, CurrentID=0, Status='live', Symbol='CG7328', Aliases='Dmel_ Dmel\\CG7328', description='CG7328 gene product from transcript CG7328-RA', other_designations='CG7328-PA', map_location='76F2-76F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20139210, end_position_on_the_genomic_accession=20140607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40205, CurrentID=0, Status='live', Symbol='CG42674', Aliases='Dmel_ CG13811, CG13812, CG32223, CG42348, CG7323, Dmel\\CG42674, Dmel_CG13811, Dmel_CG32223, Dmel_CG42348, Dmel_CG7323', description='CG42674 gene product from transcript CG42674-RF', other_designations='CG42674-PA|CG42674-PB|CG42674-PC|CG42674-PD|CG42674-PE|CG42674-PF|CG42674-PG|uncharacterized protein', map_location='77A1-77A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20149614, end_position_on_the_genomic_accession=20198317, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  40207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40207, CurrentID=0, Status='live', Symbol='Ssk', Aliases='Dmel_CG6981, CG6981, CT21613, Dmel\\CG6981, ssk', description='Snakeskin', other_designations='CG6981-PA|CG6981-PB|Ssk-PA|Ssk-PB|snakeskin', map_location='77A1-77A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20188522, end_position_on_the_genomic_accession=20189779, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40209, CurrentID=0, Status='live', Symbol='obst-F', Aliases='Dmel_CG7306, CG7306, Dmel\\CG7306', description='obstructor-F', other_designations='CG7306-PA|obst-F-PA', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20200443, end_position_on_the_genomic_accession=20201638, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40210, CurrentID=0, Status='live', Symbol='CG7298', Aliases='Dmel_ Dmel\\CG7298, anon-EST:ParkEST398', description='CG7298 gene product from transcript CG7298-RA', other_designations='CG7298-PA', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20202070, end_position_on_the_genomic_accession=20203687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40211, CurrentID=0, Status='live', Symbol='CG7290', Aliases='Dmel_ Dmel\\CG7290', description='CG7290 gene product from transcript CG7290-RB', other_designations='CG7290-PA|CG7290-PB', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20204309, end_position_on_the_genomic_accession=20206046, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40212, CurrentID=0, Status='live', Symbol='CG6996', Aliases='Dmel_ Dmel\\CG6996', description='CG6996 gene product from transcript CG6996-RC', other_designations='CG6996-PB|CG6996-PC', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20207534, end_position_on_the_genomic_accession=20209600, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40213, CurrentID=0, Status='live', Symbol='CG7017', Aliases='Dmel_ Dmel\\CG7017', description='CG7017 gene product from transcript CG7017-RB', other_designations='CG7017-PA|CG7017-PB', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20210292, end_position_on_the_genomic_accession=20212122, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40214, CurrentID=0, Status='live', Symbol='CG6933', Aliases='Dmel_ Dmel\\CG6933', description='CG6933 gene product from transcript CG6933-RA', other_designations='CG6933-PA|CG6933-PB|CG6933-PC|CG6933-PD|CG6933-PE', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20218500, end_position_on_the_genomic_accession=20220420, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40215, CurrentID=0, Status='live', Symbol='CG17145', Aliases='Dmel_ Dmel\\CG17145', description='CG17145 gene product from transcript CG17145-RB', other_designations='CG17145-PA|CG17145-PB|uncharacterized protein', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20221814, end_position_on_the_genomic_accession=20223197, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40216, CurrentID=0, Status='live', Symbol='CG17147', Aliases='Dmel_ Dmel\\CG17147', description='CG17147 gene product from transcript CG17147-RB', other_designations='CG17147-PB|uncharacterized protein', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20223932, end_position_on_the_genomic_accession=20225324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40217, CurrentID=0, Status='live', Symbol='Su(var)3-3', Aliases='Dmel_CG17149, CG17149, Dmel/HDM, Dmel/LSD1, Dmel\\CG17149, Dmel\\LSD1, Hdm, Lsd1, SU(VAR)3-3, Su-var(3)3, Suvar(3)3, dLSD1, dLsd1, lsd1', description='Suppressor of variegation 3-3', other_designations='CG17149-PA|CG17149-PB|CG17149-PC|Su(var)3-3-PA|Su(var)3-3-PB|Su(var)3-3-PC|flavoprotein|histone demethylase|histone lysine-specific demethylase 1|suppressor of variegation 3-3', map_location='3-46.6 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20226381, end_position_on_the_genomic_accession=20230250, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40218, CurrentID=0, Status='live', Symbol='CG13813', Aliases='Dmel_ Dmel\\CG13813', description='CG13813 gene product from transcript CG13813-RA', other_designations='CG13813-PA|uncharacterized protein', map_location='77A3-77A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20231669, end_position_on_the_genomic_accession=20233255, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40220, CurrentID=0, Status='live', Symbol='CG17233', Aliases='Dmel_ CG6918, Dmel\\CG17233', description='CG17233 gene product from transcript CG17233-RG', other_designations='CG17233-PA|CG17233-PC|CG17233-PD|CG17233-PE|CG17233-PF|CG17233-PG|CG17233-PH|CG17233-PI|uncharacterized protein', map_location='77A3-77A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20233173, end_position_on_the_genomic_accession=20241903, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40221, CurrentID=0, Status='live', Symbol='Clc', Aliases='Dmel_CG6948, CG6948, CLC, Dmel\\CG6948, clc, clc1', description='Clathrin light chain', other_designations='CG6948-PA|CG6948-PB|CG6948-PC|CG6948-PD|Clc-PA|Clc-PB|Clc-PC|Clc-PD|clathrin|clathrin light chain|clathrin light-chain|clathrin-light-chain', map_location='77A3-77A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20241053, end_position_on_the_genomic_accession=20243735, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40222, CurrentID=0, Status='live', Symbol='CG6951', Aliases='Dmel_ Dmel\\CG6951, anon-WO0118547.579', description='CG6951 gene product from transcript CG6951-RB', other_designations='CG6951-PA|CG6951-PB', map_location='77A3-77A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20241267, end_position_on_the_genomic_accession=20245312, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40224, CurrentID=0, Status='live', Symbol='rdgC', Aliases='Dmel_CG44746, CG13814, CG44746, CG6571, Dmel\\CG44746, Dmel_CG13814, Dmel_CG6571, DmrdgC-77B, RDGC, RDGC/PP5, RdgC, RdgC 77B', description='retinal degeneration C', other_designations='CG44746-PB|CG44746-PF|CG44746-PG|CG44746-PH|rdgC-PB|rdgC-PF|rdgC-PG|rdgC-PH', map_location='77B1-77A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20248080, end_position_on_the_genomic_accession=20265554, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  40225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40225, CurrentID=0, Status='live', Symbol='gogo', Aliases='Dmel_CG32227, CG13815, CG13817, CG32227, CT33316, Dmel\\CG32227, Gogo', description='golden goal', other_designations='CG32227-PA|gogo-PA', map_location='77B1-77B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20268859, end_position_on_the_genomic_accession=20295133, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40226, CurrentID=0, Status='live', Symbol='CG17122', Aliases='Dmel_ Dmel\\CG17122', description='CG17122 gene product from transcript CG17122-RA', other_designations='CG17122-PA|uncharacterized protein', map_location='77B1-77B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20273389, end_position_on_the_genomic_accession=20275335, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40228, CurrentID=0, Status='live', Symbol='FRG1', Aliases='Dmel_CG6480, CG6480, Dmel\\CG6480, dFRG1', description='FSHD region gene 1 ortholog (H. sapiens)', other_designations='CG6480-PA|FRG1-PA|FSHD region gene 1 ortholog', map_location='77B1-77B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20295802, end_position_on_the_genomic_accession=20296971, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40229, CurrentID=0, Status='live', Symbol='Pex23', Aliases='Dmel_CG32226, CG18565, CG32226, CG6468, DmelPex23, Dmel\\CG32226', description='Peroxin 23', other_designations='CG32226-PA|Pex23-PA|peroxin 23', map_location='77B1-77B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20297129, end_position_on_the_genomic_accession=20302141, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40231, CurrentID=0, Status='live', Symbol='CG6597', Aliases='Dmel_ Dmel\\CG6597', description='CG6597 gene product from transcript CG6597-RE', other_designations='CG6597-PA|CG6597-PB|CG6597-PD|CG6597-PE', map_location='77B2-77B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20299410, end_position_on_the_genomic_accession=20309511, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40232, CurrentID=0, Status='live', Symbol='polo', Aliases='Dmel_CG12306, 0256/04, 1324/08, CG12306, Dmel\\CG12306, PLK1, POLO, POLO/PLK1, Polo, Polo kinase, anon-WO0172774.3, l(3)01673, l(3)77Aa, l(3)S025604, l(3)S132408', description='CG12306 gene product from transcript CG12306-RA', other_designations='CG12306-PA|CG12306-PB|polo kinase|polo protein kinase|polo-PA|polo-PB', map_location='3-46 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20309654, end_position_on_the_genomic_accession=20313466, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40233, CurrentID=0, Status='live', Symbol='alphaSnap', Aliases='Dmel_CG6625, CG6625, Dmel\\CG6625, SNAP, Snap, aSnap, alpha-SNAP, dSNAP, dalpha-SNAP, dalphaSNAP, dsnap, l(3)77ABa, l(3)77ABc, snap', description='alpha Soluble NSF attachment protein', other_designations='CG6625-PA|alpha soluble NSF attachment protein|alphaSnap-PA|lethal (3) 77ABa|lethal (3) 77ABc|soluble N-ethylmaleimide-sensitive factor attachment protein|soluble NSF attachment protein|soluble NSF-attachment protein', map_location='77B3-77B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20313632, end_position_on_the_genomic_accession=20316577, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40234, CurrentID=0, Status='live', Symbol='Spn77Ba', Aliases='Dmel_CG6680, CG6680, Dmel\\CG6680, Q0E8C8_DROME, l(3)77ABi, spn77Ba', description='Serpin 77Ba', other_designations='CG6680-PA|CG6680-PB|CG6680-PC|Spn77Ba-PA|Spn77Ba-PB|Spn77Ba-PC|lethal (3) 77ABi|serpin 77Ba', map_location='77B3-77B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20317190, end_position_on_the_genomic_accession=20320116, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40235, CurrentID=0, Status='live', Symbol='Spn77Bb', Aliases='Dmel_CG6663, CG6266, CG6663, Dmel\\CG6663, Spn77', description='Serpin 77Bb', other_designations='CG6663-PA|Spn77Bb-PA|serpin 77Bb', map_location='77B4-77B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20320952, end_position_on_the_genomic_accession=20322358, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40236, CurrentID=0, Status='live', Symbol='Spn77Bc', Aliases='Dmel_CG6289, CG6289, Dmel\\CG6289, Spn77', description='Serpin 77Bc', other_designations='CG6289-PA|CG6289-PB|Spn77Bc-PA|Spn77Bc-PB|serpin 77Bc', map_location='77B4-77B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20342634, end_position_on_the_genomic_accession=20344319, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40239, CurrentID=0, Status='live', Symbol='Rbbp5', Aliases='Dmel_CG5585, CG5585, Dmel\\CG5585, RBBP5, dRbbp5, rbbp5', description='Retinoblastoma binding protein 5', other_designations='CG5585-PA|Rbbp5-PA|retinoblastoma binding protein 5', map_location='77B4-77B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20345736, end_position_on_the_genomic_accession=20347942, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40240, CurrentID=0, Status='live', Symbol='eRF1', Aliases='Dmel_CG5605, CG5605, Dm0184, Dmel\\CG5605, Group J, group J, l(3)00103, l(3)05637, l(3)77ABb, l(3)neo28', description='eukaryotic release factor 1', other_designations='CG5605-PA|CG5605-PB|CG5605-PC|CG5605-PE|CG5605-PF|CG5605-PG|CG5605-PH|CG5605-PI|eRF1-PA|eRF1-PB|eRF1-PC|eRF1-PE|eRF1-PF|eRF1-PG|eRF1-PH|eRF1-PI', map_location='77B4-77B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20348355, end_position_on_the_genomic_accession=20353873, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40241, CurrentID=0, Status='live', Symbol='CG5618', Aliases='Dmel_ Dmel\\CG5618', description='CG5618 gene product from transcript CG5618-RD', other_designations='CG5618-PA|CG5618-PD|CG5618-PE', map_location='77B5-77B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20353449, end_position_on_the_genomic_accession=20356632, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40242, CurrentID=0, Status='live', Symbol='kin17', Aliases='Dmel_CG5649, CG5649, Dmel\\CG5649', description='CG5649 gene product from transcript CG5649-RA', other_designations='CG5649-PA|kin17-PA', map_location='77B5-77B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20356807, end_position_on_the_genomic_accession=20358331, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40243, CurrentID=0, Status='live', Symbol='CG5665', Aliases='Dmel_ Dmel\\CG5665, cg5665', description='CG5665 gene product from transcript CG5665-RC', other_designations='CG5665-PB|CG5665-PC', map_location='77B5-77B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20358512, end_position_on_the_genomic_accession=20359820, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40245, CurrentID=0, Status='live', Symbol='Toll-9', Aliases='Dmel_CG5528, CG5528, CT17508, Dm  Dmel\\CG5528, TOLL 9, Tak/Toll-like, Tl-5, Tl-9, Toll 9, dTLR9, dToll9, toll', description='CG5528 gene product from transcript CG5528-RA', other_designations='CG5528-PA|CG5528-PB|CG5528-PC|Toll-9-PA|Toll-9-PB|Toll-9-PC|toll receptor', map_location='77B6-77B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20361843, end_position_on_the_genomic_accession=20366724, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40246, CurrentID=0, Status='live', Symbol='in', Aliases='Dmel_CG16993, B, CG16993, Dmel\\CG16993', description='inturned', other_designations='CG16993-PA|in-PA|intwined|transcript B', map_location='77B6-77B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20367198, end_position_on_the_genomic_accession=20370259, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40247, CurrentID=0, Status='live', Symbol='CG13247', Aliases='Dmel_ Dmel\\CG13247', description='CG13247 gene product from transcript CG13247-RA', other_designations='CG13247-PA|uncharacterized protein', map_location='77B6-77B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20370313, end_position_on_the_genomic_accession=20371746, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40248, CurrentID=0, Status='live', Symbol='Ide', Aliases='Dmel_CG5517, CG5517, Dmel\\CG5517, IDE, dIDE, dIde, dide, dlde, ide', description='Insulin degrading metalloproteinase', other_designations='CG5517-PB|Ide-PB|insulin degrading enzyme|insulin degrading metalloproteinase|insulin-degrading enzyme', map_location='77B6-77B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20371672, end_position_on_the_genomic_accession=20375896, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  40249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40249, CurrentID=0, Status='live', Symbol='RhoBTB', Aliases='Dmel_CG5701, 28574635, CG5701, Dmel\\CG5701, RhoGTPase, RhoX', description='Rho-related BTB domain containing', other_designations='CG5701-PA|CG5701-PB|CG5701-PC|RhoBTB-PA|RhoBTB-PB|RhoBTB-PC', map_location='77B6-77B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20376242, end_position_on_the_genomic_accession=20382398, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  40250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40250, CurrentID=0, Status='live', Symbol='CG5498', Aliases='Dmel_ BcDNA:GH08385, Dmel\\CG5498', description='CG5498 gene product from transcript CG5498-RB', other_designations='CG5498-PB|transcript 2', map_location='77B9-77B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20382405, end_position_on_the_genomic_accession=20384109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40252, CurrentID=0, Status='live', Symbol='CG33969', Aliases='Dmel_ CG32429, CG5571, Dmel\\CG33969', description='CG33969 gene product from transcript CG33969-RC', other_designations='CG33969-PA|CG33969-PC|uncharacterized protein', map_location='77C1-77C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20391685, end_position_on_the_genomic_accession=20393553, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40254, CurrentID=0, Status='live', Symbol='CG13248', Aliases='Dmel_ CAT-4, Dmel\\CG13248', description='CG13248 gene product from transcript CG13248-RA', other_designations='CG13248-PA|uncharacterized protein', map_location='77C1-77C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20407103, end_position_on_the_genomic_accession=20410566, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40255, CurrentID=0, Status='live', Symbol='zye', Aliases='Dmel_CG5847, CG5847, Dmel\\CG5847', description='CG5847 gene product from transcript CG5847-RA', other_designations='CG5847-PA|zye-PA', map_location='77C2-77C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20413604, end_position_on_the_genomic_accession=20424309, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40256, CurrentID=0, Status='live', Symbol='CG5282', Aliases='Dmel_ Dmel\\CG5282', description='CG5282 gene product from transcript CG5282-RA', other_designations='CG5282-PA', map_location='77C2-77C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20424424, end_position_on_the_genomic_accession=20426071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40257, CurrentID=0, Status='live', Symbol='CG5274', Aliases='Dmel_ Dmel\\CG5274', description='CG5274 gene product from transcript CG5274-RB', other_designations='CG5274-PA|CG5274-PB', map_location='77C2-77C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20426199, end_position_on_the_genomic_accession=20427666, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40258, CurrentID=0, Status='live', Symbol='CG5262', Aliases='Dmel_ Dmel\\CG5262, anon-WO0118547.347', description='CG5262 gene product from transcript CG5262-RB', other_designations='CG5262-PA|CG5262-PB', map_location='77C2-77C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20427962, end_position_on_the_genomic_accession=20430475, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40259, CurrentID=0, Status='live', Symbol='Las', Aliases='Dmel_CG5231, CG5231, Dmel\\CG5231', description='Lipoic acid synthase', other_designations='CG5231-PC|Las-PC|lipoic acid synthase', map_location='77C2-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20430605, end_position_on_the_genomic_accession=20432361, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40260, CurrentID=0, Status='live', Symbol='Psn', Aliases='Dmel_CG18803, CG18803, CG5868, DPS, DPsn, DmPS, Dmel\\CG18803, Dps, DrosPS, PS, PS1, PSN, dPsn, l(3)77CDb, pres, psn', description='Presenilin', other_designations='CG18803-PA|CG18803-PB|CG18803-PC|CG18803-PD|CG18803-PE|Psn-PA|Psn-PB|Psn-PC|Psn-PD|Psn-PE|dPresenilin|pre-senelin|presenilin', map_location='77C3-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20432468, end_position_on_the_genomic_accession=20435156, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40261, CurrentID=0, Status='live', Symbol='mRpL15', Aliases='Dmel_CG5219, CG5219, Dmel\\CG5219, L15, l(3)77CDc', description='mitochondrial ribosomal protein L15', other_designations='50S ribosomal protein L15|CG5219-PA|mRpL15-PA', map_location='77C3-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20435110, end_position_on_the_genomic_accession=20436209, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40262, CurrentID=0, Status='live', Symbol='CG5872', Aliases='Dmel_ Dmel\\CG5872', description='CG5872 gene product from transcript CG5872-RB', other_designations='CG5872-PA|CG5872-PB', map_location='77C3-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20436250, end_position_on_the_genomic_accession=20438370, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40263, CurrentID=0, Status='live', Symbol='CG11796', Aliases='Dmel_ Dmel\\CG11796', description='CG11796 gene product from transcript CG11796-RA', other_designations='CG11796-PA|CG11796-PB|uncharacterized protein', map_location='77C3-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20438166, end_position_on_the_genomic_accession=20439872, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40264, CurrentID=0, Status='live', Symbol='CG5910', Aliases='Dmel_ CK02563, Dmel\\CG5910', description='CG5910 gene product from transcript CG5910-RA', other_designations='CG5910-PA|CG5910-PB', map_location='77C3-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20442419, end_position_on_the_genomic_accession=20450580, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40265, CurrentID=0, Status='live', Symbol='CG5199', Aliases='Dmel_ CUT8, Dmel\\CG5199', description='CG5199 gene product from transcript CG5199-RB', other_designations='CG5199-PA|CG5199-PB', map_location='77C3-77C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20450532, end_position_on_the_genomic_accession=20452864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40266, CurrentID=0, Status='live', Symbol='atk', Aliases='Dmel_CG5195, CG5195, CT16609, Dmel\\CG5195, tartan/capricious-like', description='artichoke', other_designations='CG5195-PA|CG5195-PB|CG5195-PC|atk-PA|atk-PB|atk-PC', map_location='77C3-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20454690, end_position_on_the_genomic_accession=20464862, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40267, CurrentID=0, Status='live', Symbol='mag', Aliases='Dmel_CG5932, CG5932, Dmel\\CG5932, Mag, lipA', description='magro', other_designations='CG5932-PA|mag-PA', map_location='77C4-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20460683, end_position_on_the_genomic_accession=20462274, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40268, CurrentID=0, Status='live', Symbol='CG5955', Aliases='Dmel_ Dmel\\CG5955', description='CG5955 gene product from transcript CG5955-RA', other_designations='CG5955-PA', map_location='77C4-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20466944, end_position_on_the_genomic_accession=20469592, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40269, CurrentID=0, Status='live', Symbol='CG5969', Aliases='Dmel_ BcDNA:GM09417, BcDNA:RE53796, CG32426, Dmel\\CG5969', description='CG5969 gene product from transcript CG5969-RA', other_designations='CG5969-PA', map_location='77C4-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20473783, end_position_on_the_genomic_accession=20474791, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40270, CurrentID=0, Status='live', Symbol='isoQC', Aliases='Dmel_CG5976, CG32426, CG5976, Dmel\\CG5976, isoDromeQC', description='iso Glutaminyl cyclase', other_designations='CG5976-PA|CG5976-PB|iso glutaminyl cyclase|isoQC-PA|isoQC-PB', map_location='77C4-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20474466, end_position_on_the_genomic_accession=20477003, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40271, CurrentID=0, Status='live', Symbol='ZnT77C', Aliases='Dmel_CG5130, CG5130, Dmel\\CG5130, anon-WO0118547.435, cg5130, dZnT77C', description='Zinc transporter 77C', other_designations='CG5130-PA|CG5130-PB|ZnT77C-PA|ZnT77C-PB|zinc transporter 77C', map_location='77C4-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20484293, end_position_on_the_genomic_accession=20494921, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40272, CurrentID=0, Status='live', Symbol='ND-39', Aliases='Dmel_CG6020, CG6020, Dmel\\CG6020, GM13757p, NDUFA9, NUEM', description='NADH dehydrogenase (ubiquinone) 39 kDa subunit', other_designations='CG6020-PA|CG6020-PB|FBpp0077859|NADH dehydrogenase|NADH ubiquinone oxidoreductase 39 kDa|NADH-ubiquinone oxidoreductase 39 kDa subunit|NADH:ubiquinone oxidoreductase 39 kDa subunit|ND-39-PA|ND-39-PB', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20492082, end_position_on_the_genomic_accession=20497084, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40273, CurrentID=0, Status='live', Symbol='CG32425', Aliases='Dmel_ CG15869, CG5100, Dmel\\CG32425', description='CG32425 gene product from transcript CG32425-RE', other_designations='CG32425-PA|CG32425-PB|CG32425-PC|CG32425-PE|CG32425-PF|uncharacterized protein', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20497323, end_position_on_the_genomic_accession=20514107, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40274, CurrentID=0, Status='live', Symbol='CG18281', Aliases='Dmel_ Dmel\\CG18281', description='CG18281 gene product from transcript CG18281-RA', other_designations='CG18281-PA|uncharacterized protein', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20499836, end_position_on_the_genomic_accession=20501500, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40275, CurrentID=0, Status='live', Symbol='CG17637', Aliases='Dmel_ Dmel\\CG17637', description='CG17637 gene product from transcript CG17637-RA', other_designations='CG17637-PA|uncharacterized protein', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20504186, end_position_on_the_genomic_accession=20505700, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40276, CurrentID=0, Status='live', Symbol='CG5078', Aliases='Dmel_ Dmel\\CG5078', description='CG5078 gene product from transcript CG5078-RA', other_designations='CG5078-PA', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20506032, end_position_on_the_genomic_accession=20507375, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40278, CurrentID=0, Status='live', Symbol='CG5059', Aliases='Dmel_ BNIP3, Dmel\\CG5059', description='CG5059 gene product from transcript CG5059-RD', other_designations='CG5059-PA|CG5059-PB|CG5059-PC|CG5059-PD|CG5059-PE', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20514598, end_position_on_the_genomic_accession=20519393, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40279, CurrentID=0, Status='live', Symbol='mTerf3', Aliases='Dmel_CG5047, CG5047, D-MTERF3, DmMTERF3, DmMterf3, Dmel\\CG5047, MTERF3, Mterf3', description='CG5047 gene product from transcript CG5047-RA', other_designations='CG5047-PA|mTerf3-PA|mitochondrial transcription termination factor', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20519590, end_position_on_the_genomic_accession=20520772, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40280, CurrentID=0, Status='live', Symbol='CG5104', Aliases='Dmel_ BcDNA:RH26504, Dmel\\CG5104', description='CG5104 gene product from transcript CG5104-RB', other_designations='CG5104-PB', map_location='77C6-77C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20521009, end_position_on_the_genomic_accession=20522434, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40281, CurrentID=0, Status='live', Symbol='CG4825', Aliases='Dmel_ Dmel\\CG4825, ptdss1', description='CG4825 gene product from transcript CG4825-RB', other_designations='CDP-diacylglycerol-serine O-phosphatidyltransferase|CG4825-PA|CG4825-PB|phosphatidylserine synthase', map_location='77C6-77C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20522440, end_position_on_the_genomic_accession=20526174, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40282, CurrentID=0, Status='live', Symbol='CG4858', Aliases='Dmel_ Dmel\\CG4858, Nubp2', description='CG4858 gene product from transcript CG4858-RA', other_designations='CG4858-PA', map_location='77C7-77C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20526576, end_position_on_the_genomic_accession=20527910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40283, CurrentID=0, Status='live', Symbol='Rcd2', Aliases='Dmel_CG4786, CG4786, CT15369, Dmel\\CG4786', description='Reduction in Cnn dots 2', other_designations='CG4786-PA|Rcd2-PA|reduction in Cnn dots 2', map_location='77C7-77D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20527972, end_position_on_the_genomic_accession=20540417, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40284, CurrentID=0, Status='live', Symbol='CG13250', Aliases='Dmel_ Dmel\\CG13250', description='CG13250 gene product from transcript CG13250-RA', other_designations='CG13250-PA|CG13250-PB|uncharacterized protein', map_location='77C7-77D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20533709, end_position_on_the_genomic_accession=20535335, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40285, CurrentID=0, Status='live', Symbol='knrl', Aliases='Dmel_CG4761, CG4761, Dmel\\CG4761, Knrl, NR0A2, knl', description='knirps-like', other_designations='CG4761-PA|CG4761-PB|Knirps-related|knirps related|knrl-PA|knrl-PB', map_location='77D4-77E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20597141, end_position_on_the_genomic_accession=20620659, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40286, CurrentID=0, Status='live', Symbol='CG13251', Aliases='Dmel_ Dmel\\CG13251', description='CG13251 gene product from transcript CG13251-RA', other_designations='CG13251-PA|uncharacterized protein', map_location='77E1-77E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20637199, end_position_on_the_genomic_accession=20640910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40287, CurrentID=0, Status='live', Symbol='kni', Aliases='Dmel_CG4717, CG4717, Dmel\\CG4717, KNI, KNIRPS, Kn, Kni, NR0A1, kn, knr, ri', description='knirps', other_designations='CG4717-PA|CG4717-PB|kni-PA|kni-PB|radius incompletus', map_location='77E3-77E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20692330, end_position_on_the_genomic_accession=20695378, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40288, CurrentID=0, Status='live', Symbol='cmpy', Aliases='Dmel_CG13253, CG13253, Dmel\\CG13253', description='crimpy', other_designations='CG13253-PB|CG13253-PC|CG13253-PD|cmpy-PB|cmpy-PC|cmpy-PD', map_location='77E3-77E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20712838, end_position_on_the_genomic_accession=20719200, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40289, CurrentID=0, Status='live', Symbol='CG13252', Aliases='Dmel_ Dmel\\CG13252', description='CG13252 gene product from transcript CG13252-RA', other_designations='CG13252-PA|uncharacterized protein', map_location='77E4-77E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20730591, end_position_on_the_genomic_accession=20744508, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40290, CurrentID=0, Status='live', Symbol='CG4074', Aliases='Dmel_ DSCR3, DmDSCR3, Dmel\\CG4074', description='CG4074 gene product from transcript CG4074-RB', other_designations='CG4074-PA|CG4074-PB', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20768207, end_position_on_the_genomic_accession=20769482, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40291, CurrentID=0, Status='live', Symbol='CG4042', Aliases='Dmel_ Dmel\\CG4042', description='CG4042 gene product from transcript CG4042-RB', other_designations='CG4042-PA|CG4042-PB', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20770143, end_position_on_the_genomic_accession=20771507, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40292, CurrentID=0, Status='live', Symbol='Pitslre', Aliases='Dmel_CG4268, CG4268, Dmel\\CG4268, Gat, Gta, PITSLRE, PitsIre, cdk11, group 3, pitslre', description='CG4268 gene product from transcript CG4268-RA', other_designations='CG4268-PA|CG4268-PC|CG4268-PE|CG4268-PF|CG4268-PG|Galactosyltransferase-associated kinase-like|Gat-like|PITSLRE|PITSLRE kinase|Pitslre-PA|Pitslre-PC|Pitslre-PE|Pitslre-PF|Pitslre-PG|pitslre', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20771528, end_position_on_the_genomic_accession=20776931, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40293, CurrentID=0, Status='live', Symbol='Pex16', Aliases='Dmel_CG3947, CG3947, DmelPex16, Dmel\\CG3947, PEX16, pex16', description='Peroxin 16', other_designations='CG3947-PA|Pex16-PA|peroxin 16', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20776766, end_position_on_the_genomic_accession=20778264, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40294, CurrentID=0, Status='live', Symbol='Pex14', Aliases='Dmel_CG4289, CG4289, DmelPex14, Dmel\\CG4289, PEX14', description='Peroxin 14', other_designations='CG4289-PA|Pex14-PA|peroxin 14', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20778799, end_position_on_the_genomic_accession=20780378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40295, CurrentID=0, Status='live', Symbol='CG11399', Aliases='Dmel_ Dmel\\CG11399', description='CG11399 gene product from transcript CG11399-RB', other_designations='CG11399-PB|uncharacterized protein', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20780350, end_position_on_the_genomic_accession=20786197, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40296, CurrentID=0, Status='live', Symbol='CG11396', Aliases='Dmel_ Dmel\\CG11396', description='CG11396 gene product from transcript CG11396-RA', other_designations='CG11396-PA|uncharacterized protein', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20786340, end_position_on_the_genomic_accession=20788137, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40297, CurrentID=0, Status='live', Symbol='Six4', Aliases='Dmel_CG3871, CG3871, D- D-six4, DSix4, Dmel\\CG3871, Dsix4, six4, six4/5', description='CG3871 gene product from transcript CG3871-RA', other_designations='CG3871-PA|CG3871-PB|Six4-PA|Six4-PB|myotonix', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20788886, end_position_on_the_genomic_accession=20792777, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40298, CurrentID=0, Status='live', Symbol='CG3698', Aliases='Dmel_ Dmel\\CG3698', description='CG3698 gene product from transcript CG3698-RA', other_designations='CG3698-PA', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20797241, end_position_on_the_genomic_accession=20798375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40299, CurrentID=0, Status='live', Symbol='CG4365', Aliases='Dmel_ Dmel\\CG4365', description='CG4365 gene product from transcript CG4365-RB', other_designations='CG4365-PA|CG4365-PB|CG4365-PC', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20798645, end_position_on_the_genomic_accession=20800081, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40300, CurrentID=0, Status='live', Symbol='Spc105R', Aliases='Dmel_CG11451, CG11451, DmSpc105, Dmel\\CG11451, Spc105, dmKNL-1, dmSpc105R, spc105R', description='Spc105-related', other_designations='CG11451-PA|CG11451-PB|Spc105R-PA|Spc105R-PB', map_location='77E6-77E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20800410, end_position_on_the_genomic_accession=20806853, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40301, CurrentID=0, Status='live', Symbol='CG3634', Aliases='Dmel_ Dmel\\CG3634', description='CG3634 gene product from transcript CG3634-RA', other_designations='CG3634-PA|CG3634-PB|CG3634-PC', map_location='77E8-77E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20806868, end_position_on_the_genomic_accession=20809272, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40302, CurrentID=0, Status='live', Symbol='HIPP1', Aliases='Dmel_CG3680, CG3680, Dmel\\CG3680', description='HP1 and insulator partner protein 1', other_designations='CG3680-PA|CG3680-PB|CG3680-PC|CG3680-PD|HIPP1-PA|HIPP1-PB|HIPP1-PC|HIPP1-PD|HP1 and insulator partner protein-1', map_location='77E8-77E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20809669, end_position_on_the_genomic_accession=20813332, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40303, CurrentID=0, Status='live', Symbol='CG3618', Aliases='Dmel_ Dmel\\CG3618, GYCb100B', description='CG3618 gene product from transcript CG3618-RB', other_designations='CG3618-PA|CG3618-PB', map_location='77E8-77E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20813353, end_position_on_the_genomic_accession=20816075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40304, CurrentID=0, Status='live', Symbol='Mst77F', Aliases='Dmel_CG3354, 18c, CG3354, Dmel\\CG3354, R1-18c, ms(3)nc3, mst77F, nc3', description='Male-specific transcript 77F', other_designations='CG3354-PA|Mst77F-PA|male sterile (3) nc3', map_location='77E8-77F', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20816901, end_position_on_the_genomic_accession=20818027, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40305, CurrentID=0, Status='live', Symbol='Pka-R1', Aliases='Dmel_CG42341, 18304, CG12452, CG18677, CG3263, CG42341, CdkR, DRI, Dmel\\CG42341, Dmel_CG12452, Dmel_CG3263, PKA, PKA-R1, PKA-RI, PKAR, PKAi, PKAr, PKa-R1, Pka-RI, PkaR, RI, dPKA-RI, dr[[1]], pKA, pkA, pka-R1', description='Protein kinase, cAMP-dependent, regulatory subunit type 1', other_designations='CG42341-PAA|CG42341-PAB|CG42341-PAC|CG42341-PK|CG42341-PL|CG42341-PM|CG42341-PN|CG42341-PO|CG42341-PP|CG42341-PQ|CG42341-PR|CG42341-PS|CG42341-PT|CG42341-PU|CG42341-PV|CG42341-PW|CG42341-PX|CG42341-PZ|Pka-R1-PAA|Pka-R1-PAB|Pka-R1-PAC|Pka-R1-PK|Pka-R1-PL|Pka-R1-PM|Pka-R1-PN|Pka-R1-PO|Pka-R1-PP|Pka-R1-PQ|Pka-R1-PR|Pka-R1-PS|Pka-R1-PT|Pka-R1-PU|Pka-R1-PV|Pka-R1-PW|Pka-R1-PX|Pka-R1-PZ|cAMP-dependent protein kinase|cAMP-dependent protein kinase R1|protein kinase A|protein kinase, cAMP-dependent, regulatory subunit type 1|regulatory subunit of PKA|regulatory subunit of protein kinase A|type I regulatory subunit of cAMP-dependent protein kinase', map_location='77E8-77F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20816118, end_position_on_the_genomic_accession=20836048, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40307, CurrentID=0, Status='live', Symbol='CG3288', Aliases='Dmel_ Dmel\\CG3288', description='CG3288 gene product from transcript CG3288-RB', other_designations='CG3288-PA|CG3288-PB', map_location='77F1-77F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20836182, end_position_on_the_genomic_accession=20837153, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40308, CurrentID=0, Status='live', Symbol='CSN3', Aliases='Dmel_CG18332, CG18332, CH3, Csn3, DCH3, Dch3, Dmel\\CG18332, FBgn0027055, dCSN3', description='COP9 signalosome subunit 3', other_designations='CG18332-PA|COP9 complex homolog subunit 3|CSN3-PA|drosophila COP9 signalosome homolog 3', map_location='77F1-77F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20839096, end_position_on_the_genomic_accession=20841576, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40309, CurrentID=0, Status='live', Symbol='CG11456', Aliases='Dmel_ Dmel\\CG11456', description='CG11456 gene product from transcript CG11456-RB', other_designations='CG11456-PA|CG11456-PB|CG11456-PC|uncharacterized protein', map_location='77F1-77F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20841987, end_position_on_the_genomic_accession=20844340, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40310, CurrentID=0, Status='live', Symbol='CG32432', Aliases='Dmel_ CG13256, CG13257, CG3116, Dmel\\CG32432, FBgn0037032', description='CG32432 gene product from transcript CG32432-RB', other_designations='CG32432-PA|CG32432-PB|uncharacterized protein', map_location='77F1-77F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20844332, end_position_on_the_genomic_accession=20885188, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  40313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40313, CurrentID=0, Status='live', Symbol='CG10589', Aliases='Dmel_ Dmel\\CG10589, anon-WO0140519.139', description='CG10589 gene product from transcript CG10589-RA', other_designations='CG10589-PA|uncharacterized protein', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20927826, end_position_on_the_genomic_accession=20929190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40314, CurrentID=0, Status='live', Symbol='fng', Aliases='Dmel_CG10580, CG10580, D-Fng, D-fng, Dfng, Dmel\\CG10580, Fng, Frg, fg, frg, l(3)rG554', description='fringe', other_designations='CG10580-PA|fng-PA', map_location='3-47 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20925891, end_position_on_the_genomic_accession=20949512, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40315, CurrentID=0, Status='live', Symbol='Sems', Aliases='Dmel_CG10586, CG10586, Dmel\\CG10586, SP168, sems', description='Seminase', other_designations='CG10586-PA|Sems-PA|seminase', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20955755, end_position_on_the_genomic_accession=20956750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40316, CurrentID=0, Status='live', Symbol='CG10588', Aliases='Dmel_ Dmel\\CG10588', description='CG10588 gene product from transcript CG10588-RB', other_designations='CG10588-PA|CG10588-PB|uncharacterized protein', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20956934, end_position_on_the_genomic_accession=20960660, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40317, CurrentID=0, Status='live', Symbol='CG11037', Aliases='Dmel_ Dmel\\CG11037, SP191', description='CG11037 gene product from transcript CG11037-RA', other_designations='CG11037-PA|uncharacterized protein', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20960398, end_position_on_the_genomic_accession=20961386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40318, CurrentID=0, Status='live', Symbol='CG10587', Aliases='Dmel_ Dmel\\CG10587, SP167', description='CG10587 gene product from transcript CG10587-RA', other_designations='CG10587-PA|CG10587-PC|CG10587-PD|uncharacterized protein', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20961537, end_position_on_the_genomic_accession=20962555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40319, CurrentID=0, Status='live', Symbol='CG12983', Aliases='Dmel_ Dmel\\CG12983, NEST:bs29a07, bs29a07.y1', description='CG12983 gene product from transcript CG12983-RE', other_designations='CG12983-PB|CG12983-PC|CG12983-PD|CG12983-PE|CG12983-PF|uncharacterized protein', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20962655, end_position_on_the_genomic_accession=20966762, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40321, CurrentID=0, Status='live', Symbol='CG12984', Aliases='Dmel_ Dmel\\CG12984', description='CG12984 gene product from transcript CG12984-RA', other_designations='CG12984-PA|CG12984-PB|uncharacterized protein', map_location='78A2-78A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20991673, end_position_on_the_genomic_accession=20992930, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772288, CurrentID=0, Status='live', Symbol='CG33774', Aliases='Dmel_ BcDNA:GM20929, Dm_2R:21419, Dmel\\CG33774', description='CG33774 gene product from transcript CG33774-RA', other_designations='CG33774-PA|uncharacterized protein', map_location='45D3-45D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9399753, end_position_on_the_genomic_accession=9400329, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40323, CurrentID=0, Status='live', Symbol='CG10585', Aliases='Dmel_ Dmel\\CG10585', description='CG10585 gene product from transcript CG10585-RA', other_designations='CG10585-PA|CG10585-PB|uncharacterized protein', map_location='78A2-78A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21027915, end_position_on_the_genomic_accession=21030141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40324, CurrentID=0, Status='live', Symbol='CG10584', Aliases='Dmel_ Dmel\\CG10584, anon-WO0118547.713', description='CG10584 gene product from transcript CG10584-RA', other_designations='CG10584-PA|uncharacterized protein', map_location='78A2-78A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21030280, end_position_on_the_genomic_accession=21031360, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40325, CurrentID=0, Status='live', Symbol='Sin', Aliases='Dmel_CG10582, CG10582, Dmel\\CG10582, sin', description='Sex-lethal interactor', other_designations='CG10582-PA|CG10582-PC|Sin-PA|Sin-PC', map_location='78A5-78A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21031825, end_position_on_the_genomic_accession=21033561, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40326, CurrentID=0, Status='live', Symbol='CG10581', Aliases='Dmel_ Dmel\\CG10581', description='CG10581 gene product from transcript CG10581-RA', other_designations='CG10581-PA|uncharacterized protein', map_location='78A5-78A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21033554, end_position_on_the_genomic_accession=21034203, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40327, CurrentID=0, Status='live', Symbol='siz', Aliases='Dmel_CG32434, CG10577, CG12980, CG12981, CG32434, Dmel\\CG32434, Loner, loner', description='schizo', other_designations='CG32434-PA|CG32434-PB|CG32434-PC|CG32434-PD|dBRAG|loner|siz-PA|siz-PB|siz-PC|siz-PD', map_location='78A5-78B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21034535, end_position_on_the_genomic_accession=21075444, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40330, CurrentID=0, Status='live', Symbol='ko', Aliases='Dmel_CG10573, CG10573, Dmel\\CG10573', description='knockout', other_designations='CG10573-PA|CG10573-PB|ko-PA|ko-PB', map_location='78B1-78B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21075581, end_position_on_the_genomic_accession=21131159, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40331, CurrentID=0, Status='live', Symbol='ICA69', Aliases='Dmel_CG10566, CG10566, Dm ric-19, Dmel\\CG10566, dCG10566-PA', description='CG10566 gene product from transcript CG10566-RA', other_designations='CG10566-PA|CG10566-PB|ICA69-PA|ICA69-PB', map_location='78B2-78B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21131446, end_position_on_the_genomic_accession=21133310, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40332, CurrentID=0, Status='live', Symbol='CG10565', Aliases='Dmel_ Dmel\\CG10565', description='CG10565 gene product from transcript CG10565-RB', other_designations='CG10565-PA|CG10565-PB|uncharacterized protein', map_location='78B2-78B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21133250, end_position_on_the_genomic_accession=21136353, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40333, CurrentID=0, Status='live', Symbol='Ac78C', Aliases='Dmel_CG10564, AC 78C, AC78C, CG10554, CG10564, DAC78C, DAC78C-s, Dmel\\CG10564', description='Adenylyl cyclase 78C', other_designations='Ac78C-PB|Ac78C-PC|Ac78C-PD|Ac78C-PE|Ac78C-PF|CG10564-PB|CG10564-PC|CG10564-PD|CG10564-PE|CG10564-PF|adenylyl cyclase|adenylyl cyclase 78C', map_location='78B3-78C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21136447, end_position_on_the_genomic_accession=21171930, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  40335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40335, CurrentID=0, Status='live', Symbol='CG42337', Aliases='Dmel_ CG12976, CG12977, Dmel\\CG42337, Dmel_CG12976, Dmel_CG12977', description='CG42337 gene product from transcript CG42337-RC', other_designations='CG42337-PB|CG42337-PC|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21183313, end_position_on_the_genomic_accession=21193363, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40336, CurrentID=0, Status='live', Symbol='park', Aliases='Dmel_CG10523, AAN12155.1, CG10523, Dmel\\CG10523, Dpark, PK, Park, SD01679, dpk, pk', description='parkin', other_designations='CG10523-PB|CG10523-PC|D-parkin|dparkin|park-PB|park-PC', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21194754, end_position_on_the_genomic_accession=21196853, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40339, CurrentID=0, Status='live', Symbol='CG10512', Aliases='Dmel_ Dmel\\CG10512', description='CG10512 gene product from transcript CG10512-RA', other_designations='CG10512-PA|CG10512-PB|CG10512-PC|CG10512-PD|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21202920, end_position_on_the_genomic_accession=21208092, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40340, CurrentID=0, Status='live', Symbol='CG33054', Aliases='Dmel_ CG10167, CG10517, Dmel\\CG33054', description='CG33054 gene product from transcript CG33054-RA', other_designations='CG33054-PA|CG33054-PB|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21197411, end_position_on_the_genomic_accession=21202882, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40341, CurrentID=0, Status='live', Symbol='CG10510', Aliases='Dmel_ Dmel\\CG10510', description='CG10510 gene product from transcript CG10510-RA', other_designations='CG10510-PA|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21208148, end_position_on_the_genomic_accession=21210084, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40342, CurrentID=0, Status='live', Symbol='CG10508', Aliases='Dmel_ Dmel\\CG10508', description='CG10508 gene product from transcript CG10508-RK', other_designations='CG10508-PI|CG10508-PJ|CG10508-PK|CG10508-PL|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21210215, end_position_on_the_genomic_accession=21216051, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40343, CurrentID=0, Status='live', Symbol='CG12975', Aliases='Dmel_ BcDNA:RE57810, Dmel\\CG12975', description='CG12975 gene product from transcript CG12975-RA', other_designations='CG12975-PA|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21215995, end_position_on_the_genomic_accession=21216839, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40345, CurrentID=0, Status='live', Symbol='Eip78C', Aliases='Dmel_CG18023, CG18023, DR-78, Dmel\\CG18023, E78, E78A, E78B, E78D, EIP78C, Eip 78C, Eip78, Eip78D, NR1E1, anon-WO0172774.42, anon-WO0172774.44, dE78', description='Ecdysone-induced protein 78C', other_designations='CG18023-PA|CG18023-PB|CG18023-PD|CG18023-PE|Eip78C-PA|Eip78C-PB|Eip78C-PD|Eip78C-PE|ecdysone induced protein 78C|nuclear hormone receptor', map_location='78C2-78C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21226609, end_position_on_the_genomic_accession=21266503, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40346, CurrentID=0, Status='live', Symbol='CG9391', Aliases='Dmel_ 142335_at, Dmel\\CG9391, IMP', description='CG9391 gene product from transcript CG9391-RC', other_designations='CG9391-PA|CG9391-PC', map_location='78C3-78C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21266680, end_position_on_the_genomic_accession=21268335, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40347, CurrentID=0, Status='live', Symbol='CG9389', Aliases='Dmel_ 142336_at, Dmel\\CG9389, IMP', description='CG9389 gene product from transcript CG9389-RA', other_designations='CG9389-PA', map_location='78C3-78C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21268392, end_position_on_the_genomic_accession=21270580, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40348, CurrentID=0, Status='live', Symbol='AcCoAS', Aliases='Dmel_CG9390, ACS, AcCoAs, BEST:GH28401, CG9390, Dmel\\CG9390, dACS', description='Acetyl Coenzyme A synthase', other_designations='Ac CoA synthase|AcCoAS-PA|AcCoAS-PB|AcCoAS-PD|Acetyl-CoA synthase|Acetyl-CoA sythase|AcetylCoA synthase|CG9390-PA|CG9390-PB|CG9390-PD|acetate-coenzyme A ligase|acetyl CoA synthase|acetyl CoA synthetase|acetyl coenzyme A synthase|acetyl-CoA synthetase|acetyl-coenzyme A synthase', map_location='78C3-78C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21270722, end_position_on_the_genomic_accession=21280243, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40349, CurrentID=0, Status='live', Symbol='ppl', Aliases='Dmel_CG7758, CG7758, Dmel\\CG7758, l(3)00217, l(3)78Cb', description='pumpless', other_designations='CG7758-PA|ppl-PA', map_location='78C4-78C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21280491, end_position_on_the_genomic_accession=21281506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40350, CurrentID=0, Status='live', Symbol='CG12974', Aliases='Dmel_ Dmel\\CG12974', description='CG12974 gene product from transcript CG12974-RD', other_designations='CG12974-PA|CG12974-PB|CG12974-PC|CG12974-PD|uncharacterized protein', map_location='78C4-78C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21281752, end_position_on_the_genomic_accession=21285962, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40351, CurrentID=0, Status='live', Symbol='pzg', Aliases='Dmel_CG7752, CG7752, Dmel\\CG7752, Pzg, Pzg/Z4, Q9VP57_DROME, Z4, p160, z4', description='putzig', other_designations='CG7752-PA|CG7752-PB|Z4|pzg-PA|pzg-PB', map_location='78C5-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21286099, end_position_on_the_genomic_accession=21290310, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40352, CurrentID=0, Status='live', Symbol='Cpr78Ca', Aliases='Dmel_CG11310, CG11310, DmelCpr78Ca, Dmel\\CG11310', description='Cuticular protein 78Ca', other_designations='CG11310-PA|Cpr78Ca-PA|cuticular protein 78Ca', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21291860, end_position_on_the_genomic_accession=21292800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40353, CurrentID=0, Status='live', Symbol='Cpr78Cb', Aliases='Dmel_CG7663, BcDNA:RE38359, CG7663, DmelCpr78Cb, Dmel\\CG7663, anon-EST:ParkEST240', description='Cuticular protein 78Cb', other_designations='CG7663-PA|CG7663-PB|Cpr78Cb-PA|Cpr78Cb-PB|cuticular protein 78Cb', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21294492, end_position_on_the_genomic_accession=21296698, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40354, CurrentID=0, Status='live', Symbol='Edg78E', Aliases='Dmel_CG7673, CG7673, DMLCP78, DmelEdg78E, Dmel\\CG7673, EDG-78, EDG78, EDG78E, Edg78, Edg78e', description='Ecdysone-dependent gene 78E', other_designations='CG7673-PA|CG7673-PB|Edg78E-PA|Edg78E-PB', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21296763, end_position_on_the_genomic_accession=21297803, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40355, CurrentID=0, Status='live', Symbol='Cpr78Cc', Aliases='Dmel_CG7658, BcDNA:RE32113, CG7658, DmelCpr78Cc, Dmel\\CG7658, anon-EST:ParkEST167', description='Cuticular protein 78Cc', other_designations='CG7658-PA|Cpr78Cc-PA|cuticular protein 78Cc', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21300580, end_position_on_the_genomic_accession=21301171, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40356, CurrentID=0, Status='live', Symbol='CG11309', Aliases='Dmel_ Dmel\\CG11309', description='CG11309 gene product from transcript CG11309-RF', other_designations='CG11309-PA|CG11309-PB|CG11309-PC|CG11309-PD|CG11309-PE|CG11309-PF|uncharacterized protein', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21301785, end_position_on_the_genomic_accession=21304658, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40357, CurrentID=0, Status='live', Symbol='CG7632', Aliases='Dmel_ Dmel\\CG7632', description='CG7632 gene product from transcript CG7632-RA', other_designations='CG7632-PA|CG7632-PB', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21304697, end_position_on_the_genomic_accession=21306197, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40358, CurrentID=0, Status='live', Symbol='Pc', Aliases='Dmel_CG32443, CG32443, CG7618, DmPc, Dmel\\CG32443, PC-G, PcG, dPC, Pc', description='Polycomb', other_designations='CG32443-PA|Pc-PA|polycomb', map_location='78C6-78C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21306138, end_position_on_the_genomic_accession=21318020, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40359, CurrentID=0, Status='live', Symbol='Rab26', Aliases='Dmel_CG34410, AAF51708, CG32439, CG34410, CG7605, Dm Dmel\\CG34410, Dmel_CG32439, Dmel_CG7605, Q8IPT6, Q9VP48, rab26', description='CG34410 gene product from transcript CG34410-RG', other_designations='CG34410-PB|CG34410-PC|CG34410-PD|CG34410-PE|CG34410-PF|CG34410-PG|Rab26-PB|Rab26-PC|Rab26-PD|Rab26-PE|Rab26-PF|Rab26-PG', map_location='78C7-78C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21318774, end_position_on_the_genomic_accession=21335027, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40360, CurrentID=0, Status='live', Symbol='Tsr1', Aliases='Dmel_CG7338, CG7338, Dmel\\CG7338', description='Tsr1 ribosome assembly factor', other_designations='CG7338-PA|Tsr1-PA|ribosome assembly factor Tsr1', map_location='78C7-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21335215, end_position_on_the_genomic_accession=21338232, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40361, CurrentID=0, Status='live', Symbol='CG7324', Aliases='Dmel_ Dmel\\CG7324', description='CG7324 gene product from transcript CG7324-RB', other_designations='CG7324-PA|CG7324-PB', map_location='78C7-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21337681, end_position_on_the_genomic_accession=21343716, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  40362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40362, CurrentID=0, Status='live', Symbol='CG32436', Aliases='Dmel_ CG12973, CG6968, Dmel\\CG32436, HD35012', description='CG32436 gene product from transcript CG32436-RA', other_designations='CG32436-PA|uncharacterized protein', map_location='78C8-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21344033, end_position_on_the_genomic_accession=21355049, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40363, CurrentID=0, Status='live', Symbol='ebd2', Aliases='Dmel_CG12972, CG12972, Dmel\\CG12972', description='earthbound 2', other_designations='CG12972-PA|CG12972-PB|ebd2-PA|ebd2-PB', map_location='78C8-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21344768, end_position_on_the_genomic_accession=21349288, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40365, CurrentID=0, Status='live', Symbol='CG12971', Aliases='Dmel_ Dmel\\CG12971', description='CG12971 gene product from transcript CG12971-RC', other_designations='CG12971-PB|CG12971-PC|uncharacterized protein', map_location='78C8-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21355100, end_position_on_the_genomic_accession=21356495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40366, CurrentID=0, Status='live', Symbol='CG32437', Aliases='Dmel_ CG5051, Dmel\\CG32437', description='CG32437 gene product from transcript CG32437-RA', other_designations='CG32437-PA|uncharacterized protein', map_location='78C8-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21357683, end_position_on_the_genomic_accession=21359867, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40367, CurrentID=0, Status='live', Symbol='sa', Aliases='Dmel_CG11308, CG11308, Dmel\\CG11308, Sa, TAF-like, TAF8L, dTAF8L, gfb, ms(3)VO45, ms(3)VO5, ms(3)sa, tTAF', description='spermatocyte arrest', other_designations='CG11308-PA|golfballs|sa-PA|spermatocytes arrest|testis-specific TBP associated factor', map_location='3-44 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21405425, end_position_on_the_genomic_accession=21406738, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40368, CurrentID=0, Status='live', Symbol='rgn', Aliases='Dmel_CG6014, BcDNA:GH11973, CG6014, Dmel\\CG6014, l(3)00534', description='regeneration', other_designations='CG6014-PA|CG6014-PB|CG6014-PC|CG6014-PD|lethal (3) 00534|rgn-PA|rgn-PB|rgn-PC|rgn-PD', map_location='78D1-78D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21382501, end_position_on_the_genomic_accession=21424416, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40369, CurrentID=0, Status='live', Symbol='barc', Aliases='Dmel_CG6049, CG6049, Dmel\\CG6049, Tat-SF1, cg6049, cus2', description='barricade', other_designations='CG6049-PA|CG6049-PC|barc-PA|barc-PC', map_location='78D1-78D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21420388, end_position_on_the_genomic_accession=21433134, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40370, CurrentID=0, Status='live', Symbol='Aef1', Aliases='Dmel_CG5683, AEF-1, CG5683, Dmel\\CG5683', description='Adult enhancer factor 1', other_designations='Adh enhancer factor 1|Aef1-PA|Aef1-PB|Aef1-PC|Aef1-PD|CG5683-PA|CG5683-PB|CG5683-PC|CG5683-PD|adult enhancer factor 1', map_location='78D2-78D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21433422, end_position_on_the_genomic_accession=21438639, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40371, CurrentID=0, Status='live', Symbol='CG5664', Aliases='Dmel_ Dmel\\CG5664', description='CG5664 gene product from transcript CG5664-RA', other_designations='CG5664-PA|CG5664-PB', map_location='78D2-78D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21438628, end_position_on_the_genomic_accession=21441363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40372, CurrentID=0, Status='live', Symbol='Alp1', Aliases='Dmel_CG5656, APH, Aph, Aph-1, CG5656, Dmel\\CG5656, l-Aph, p-Aph', description='Alkaline phosphatase 1', other_designations='Alp1-PA|Alp1-PB|CG5656-PA|CG5656-PB|alkaline phosphatase|alkaline phosphatase 1|alkaline phosphatase-1|larval alkaline phosphatase|pupal alkaline phosphatase', map_location='3-47.3 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21442227, end_position_on_the_genomic_accession=21444819, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40373, CurrentID=0, Status='live', Symbol='Syx6', Aliases='Dmel_CG7736, CG7736, CT23545, Dm Syx6, DmSyx6, Dmel\\CG7736, SYN6, SYX6/10, Syx', description='Syntaxin 6', other_designations='CG7736-PB|CG7736-PD|CG7736-PE|Syx6-PB|Syx6-PD|Syx6-PE|syntaxin|syntaxin 6', map_location='47D7-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11219539, end_position_on_the_genomic_accession=11222480, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40374, CurrentID=0, Status='live', Symbol='croc', Aliases='Dmel_CG5069, CG5069, Dmel\\CG5069, Dmfd1, FD1, FoxC, fd1, fd78E', description='crocodile', other_designations='CG5069-PA|FD1 transcription unit|croc-PA|forkhead domain 78E', map_location='78D3-78D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21474133, end_position_on_the_genomic_accession=21476479, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40375, CurrentID=0, Status='live', Symbol='Neu2', Aliases='Dmel_CG7204, CG7204, Dmel\\CG7204', description='Neuroectoderm-expressed 2', other_designations='CG7204-PA|CG7204-PB|Neu2-PA|Neu2-PB', map_location='78D3-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21484891, end_position_on_the_genomic_accession=21490586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40376, CurrentID=0, Status='live', Symbol='CG7202', Aliases='Dmel_ Dmel\\CG7202', description='CG7202 gene product from transcript CG7202-RA', other_designations='CG7202-PA', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21491399, end_position_on_the_genomic_accession=21492773, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40377, CurrentID=0, Status='live', Symbol='CG7519', Aliases='Dmel_ Dmel\\CG7519', description='CG7519 gene product from transcript CG7519-RA', other_designations='CG7519-PA|CG7519-PB', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21493083, end_position_on_the_genomic_accession=21493907, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40378, CurrentID=0, Status='live', Symbol='Hr78', Aliases='Dmel_CG7199, CG7199, DHR78, Dhr78, Dmel\\CG7199, HR78D, NR2D1, XR78E/F, Hr78', description='Hormone-receptor-like in 78', other_designations='CG7199-PA|CG7199-PB|CG7199-PC|CG7199-PD|CG7199-PE|CG7199-PF|Hr78-PA|Hr78-PB|Hr78-PC|Hr78-PD|Hr78-PE|Hr78-PF|X receptor at 78E/F', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21494019, end_position_on_the_genomic_accession=21497494, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40381, CurrentID=0, Status='live', Symbol='Est-Q', Aliases='Dmel_CG7529, CG7529, Dmel\\CG7529', description='Esterase Q', other_designations='CG7529-PA|Est-Q-PA|esterase Q', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21501421, end_position_on_the_genomic_accession=21503959, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40382, CurrentID=0, Status='live', Symbol='Glg1', Aliases='Dmel_CG33214, 120 kDa, 120-kDa, 120KD, 120kDa, CG33214, CG7190, CG7193, CG7195, DG13, Dmel\\CG33214, GLG1, d120, d120kD, d120kd, dGLG1, gp120, p120, p120-Golgi, p120kD', description='Golgi complex-localized glycoprotein 1', other_designations='CG33214-PA|CG33214-PB|Glg1-PA|Glg1-PB|golgi 120-kDa protein|golgi complex-localized glycoprotein 1|golgi complex-localized glycoprotein-1|p120', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21497549, end_position_on_the_genomic_accession=21507538, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40383, CurrentID=0, Status='live', Symbol='M6', Aliases='Dmel_CG7540, CG7540, Dmel\\CG7540, da', description='CG7540 gene product from transcript CG7540-RF', other_designations='CG7540-PB|CG7540-PC|CG7540-PD|CG7540-PE|CG7540-PF|CG7540-PG|M6-PB|M6-PC|M6-PD|M6-PE|M6-PF|M6-PG', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21507852, end_position_on_the_genomic_accession=21512749, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40384, CurrentID=0, Status='live', Symbol='SAK', Aliases='Dmel_CG7186, CG7186, DmSAK, Dmel\\CG7186, PLK4, Plk4/PLK, SAK/PLK4, Sak, plk4, sak, SAK', description='Sak kinase', other_designations='CG7186-PA|Polo-like kinase 4|SAK-PA', map_location='78D4-78D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21512141, end_position_on_the_genomic_accession=21515887, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40385, CurrentID=0, Status='live', Symbol='Cdk12', Aliases='Dmel_CG7597, AC017581, CG7597, Dmel\\CG7597, anon-WO0140519.165, cg7597, dCDK12, gi24668141', description='Cyclin-dependent kinase 12', other_designations='CG7597-PA|CG7597-PB|CG7597-PC|Cdk12-PA|Cdk12-PB|Cdk12-PC', map_location='78D4-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21516306, end_position_on_the_genomic_accession=21522684, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40386, CurrentID=0, Status='live', Symbol='CG7611', Aliases='Dmel_ Dmel\\CG7611, WDR26', description='CG7611 gene product from transcript CG7611-RI', other_designations='CG7611-PA|CG7611-PB|CG7611-PC|CG7611-PD|CG7611-PE|CG7611-PG|CG7611-PH|CG7611-PI|CG7611-PJ', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21523186, end_position_on_the_genomic_accession=21528163, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40388, CurrentID=0, Status='live', Symbol='Rpn10', Aliases='Dmel_CG7619, 7619, CG7619, DmS5a, Dmel\\CG7619, Pros54/p54, S5A, S5a, dRpn10, mu-54, p54, pros54, Rpn10', description='Regulatory particle non-ATPase 10', other_designations='54 kda mu particle subunit of 26S protease|CG7619-PA|CG7619-PB|Rpn10-PA|Rpn10-PB|proteasome 54kD subunit|regulatory particle non-ATPase 10|subunit S5a', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21537000, end_position_on_the_genomic_accession=21538582, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40389, CurrentID=0, Status='live', Symbol='VhaM9.7-b', Aliases='Dmel_CG7625, ATP6V0M9.7-2, CG7625, Dmel\\CG7625, Vha M9.7-2, VhaM9.7-2, anon-WO0118547.551, vhaM9.2-2, vhaM9.7-2, vhaM9.7-b', description='Vacuolar H[+] ATPase M9.7 subunit b', other_designations='CG7625-PA|CG7625-PB|VhaM9.7-2|VhaM9.7-b-PA|VhaM9.7-b-PB|vacuolar H[+] ATPase M9.7 subunit b|vacuolar H[+] ATPase subunit M9.7-b', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21539862, end_position_on_the_genomic_accession=21541215, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40390, CurrentID=0, Status='live', Symbol='COX8', Aliases='Dmel_CG7181, 0554/18, BcDNA:GH15688, CG7181, CcO VIIc, CoVIII, Dmel\\CG7181, VIII, anon-WO0118547.331, l(3)S055418', description='Cytochrome c oxidase subunit 8', other_designations='CG7181-PA|CG7181-PB|COX8-PA|COX8-PB|cytochrome c oxidase subunit 8|cytochrome c oxidase subunit VIII|lethal (3) S055418|subunit VIII', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21538859, end_position_on_the_genomic_accession=21539447, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40391, CurrentID=0, Status='live', Symbol='Wnk', Aliases='Dmel_CG7177, CG7177, DWNK, Dmel\\CG7177, WNK, cg7177, dWNK, wnk', description='Wnk kinase', other_designations='CG7177-PL|CG7177-PM|CG7177-PN|CG7177-PO|WNK homolog|WNK kinase|Wnk-PL|Wnk-PM|Wnk-PN|Wnk-PO|with-no-lysine kinase', map_location='78D5-78D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21543850, end_position_on_the_genomic_accession=21554928, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40392, CurrentID=0, Status='live', Symbol='CG7173', Aliases='Dmel_ Dmel\\CG7173', description='CG7173 gene product from transcript CG7173-RA', other_designations='CG7173-PA', map_location='78D6-78D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21555373, end_position_on_the_genomic_accession=21558226, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40393, CurrentID=0, Status='live', Symbol='CapaR', Aliases='Dmel_CG14575, CAP2B-R, CG14575, Dm CG14575, Dmel\\CG14575, capaR', description='Capability receptor', other_designations='CAPA receptor|CG14575-PB|CG14575-PC|CapaR-PB|CapaR-PC|capa peptide receptor|capability receptor', map_location='78D6-78D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21558974, end_position_on_the_genomic_accession=21562181, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40394, CurrentID=0, Status='live', Symbol='CG7634', Aliases='Dmel_ Dmel\\CG7634', description='CG7634 gene product from transcript CG7634-RA', other_designations='CG7634-PA', map_location='78D6-78D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21565426, end_position_on_the_genomic_accession=21567785, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40395, CurrentID=0, Status='live', Symbol='CRIF', Aliases='Dmel_CG7172, CG7172, Dmel\\CG7172, anon-WO0118547.316, dCRIF', description='CR6-interacting factor', other_designations='CG7172-PA|CRIF-PA', map_location='78D6-78D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21568059, end_position_on_the_genomic_accession=21568898, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40396, CurrentID=0, Status='live', Symbol='CG32441', Aliases='Dmel_ CG7783, Dmel\\CG32441', description='CG32441 gene product from transcript CG32441-RD', other_designations='CG32441-PC|CG32441-PD|uncharacterized protein', map_location='78D7-78D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21574131, end_position_on_the_genomic_accession=21575874, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40398, CurrentID=0, Status='live', Symbol='TfAP-2', Aliases='Dmel_CG7807, AP-2, AP-2alpha, CG7807, DAP-2, Dmel\\CG7807, dAP-2, dAP-2alpha, dTfap2, l(3)1215', description='Transcription factor AP-2', other_designations='AP-2|CG7807-PA|CG7807-PB|CG7807-PC|CG7807-PD|CG7807-PE|CG7807-PF|TfAP-2-PA|TfAP-2-PB|TfAP-2-PC|TfAP-2-PD|TfAP-2-PE|TfAP-2-PF|lethal (3) 1215|stummelbein|transcription factor AP-2', map_location='3-50.6 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21586828, end_position_on_the_genomic_accession=21613349, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40399, CurrentID=0, Status='live', Symbol='S1P', Aliases='Dmel_CG7169, CG7169, Dmel\\CG7169, d', description='CG7169 gene product from transcript CG7169-RA', other_designations='CG7169-PA|S1P-PA', map_location='78E1-78E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21613797, end_position_on_the_genomic_accession=21617376, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40400, CurrentID=0, Status='live', Symbol='CG11307', Aliases='Dmel_ Dmel\\CG11307', description='CG11307 gene product from transcript CG11307-RC', other_designations='CG11307-PC|CG11307-PD|uncharacterized protein', map_location='78E1-78E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21617623, end_position_on_the_genomic_accession=21619873, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40401, CurrentID=0, Status='live', Symbol='CG7166', Aliases='Dmel_ CT22143, Dmel\\CG7166', description='CG7166 gene product from transcript CG7166-RB', other_designations='CG7166-PB|CG7166-PC', map_location='78E1-78E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21620000, end_position_on_the_genomic_accession=21625647, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40402, CurrentID=0, Status='live', Symbol='CG11306', Aliases='Dmel_ Alg11, Dmel\\CG11306, dAlg11', description='CG11306 gene product from transcript CG11306-RB', other_designations='CG11306-PA|CG11306-PB|putative alpha-1,2-mannosyltransferase activity|uncharacterized protein', map_location='78E1-78E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21625806, end_position_on_the_genomic_accession=21628571, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40403, CurrentID=0, Status='live', Symbol='MED1', Aliases='Dmel_CG7162, CG7162, Dmel\\CG7162, FBgn0037109, Med1, Med1/SOP3/TRAP220, Trap220, dTRAP220', description='Mediator complex subunit 1', other_designations='CG7162-PA|CG7162-PB|CG7162-PC|MED1-PA|MED1-PB|MED1-PC|mediator|mediator complex subunit 1', map_location='78E1-78E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21628325, end_position_on_the_genomic_accession=21634446, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40404, CurrentID=0, Status='live', Symbol='ORMDL', Aliases='Dmel_CG14577, CG14577, Dmel\\CG14577', description='CG14577 gene product from transcript CG14577-RA', other_designations='CG14577-PA|ORMDL-PA', map_location='78E1-78E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21635305, end_position_on_the_genomic_accession=21636251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40406, CurrentID=0, Status='live', Symbol='CG32444', Aliases='Dmel_ BcDNA:GH08902, CG12562, CG14576, CT3430, Dmel\\CG32444', description='CG32444 gene product from transcript CG32444-RA', other_designations='CG32444-PA|aldose 1-epimerase|uncharacterized protein', map_location='78E1-78E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21636947, end_position_on_the_genomic_accession=21639060, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40408, CurrentID=0, Status='live', Symbol='Cpr78E', Aliases='Dmel_CG7160, BcDNA:LP06027, CG7160, DmelCpr78E, Dmel\\CG7160', description='Cuticular protein 78E', other_designations='CG7160-PA|Cpr78E-PA|cuticular protein 78E', map_location='78E2-78E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21666452, end_position_on_the_genomic_accession=21666980, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40409, CurrentID=0, Status='live', Symbol='CG11249', Aliases='Dmel_ Dmel\\CG11249', description='CG11249 gene product from transcript CG11249-RA', other_designations='CG11249-PA|uncharacterized protein', map_location='78E3-78E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21672139, end_position_on_the_genomic_accession=21674518, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40410, CurrentID=0, Status='live', Symbol='Als2', Aliases='Dmel_CG7158, CG7158, Dmel\\CG7158, dALS2', description='Amyotrophic lateral sclerosis 2 ortholog', other_designations='Als2-PA|CG7158-PA|amyotrophic lateral sclerosis 2 ortholog', map_location='78E3-78E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21674394, end_position_on_the_genomic_accession=21679755, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318031, CurrentID=0, Status='live', Symbol='CG32442', Aliases='Dmel_ BcDNA:RE62711, BcDNA:RH17480, Dmel\\CG32442', description='CG32442 gene product from transcript CG32442-RA', other_designations='CG32442-PA|CG32442-PB|CG32442-PC|uncharacterized protein', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21528218, end_position_on_the_genomic_accession=21529333, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40412, CurrentID=0, Status='live', Symbol='CG32447', Aliases='Dmel_ CG14571, CG7155, Dmel\\CG32447, anon-WO0170980.76, anon-WO0170980.77', description='CG32447 gene product from transcript CG32447-RD', other_designations='CG32447-PB|CG32447-PC|CG32447-PD|uncharacterized protein', map_location='78E4-78E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21690612, end_position_on_the_genomic_accession=21706365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40414, CurrentID=0, Status='live', Symbol='CG11247', Aliases='Dmel_ Dmel\\CG11247', description='CG11247 gene product from transcript CG11247-RA', other_designations='CG11247-PA|CG11247-PB|CG11247-PC|CG11247-PD|uncharacterized protein', map_location='78E5-78E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21724326, end_position_on_the_genomic_accession=21728806, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40415, CurrentID=0, Status='live', Symbol='Rpb8', Aliases='Dmel_CG11246, CG11246, Dmel\\CG11246, RPB8', description='CG11246 gene product from transcript CG11246-RA', other_designations='CG11246-PA|Rpb8-PA', map_location='78E5-78E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21728932, end_position_on_the_genomic_accession=21729694, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40416, CurrentID=0, Status='live', Symbol='CG14570', Aliases='Dmel_ Dmel\\CG14570', description='CG14570 gene product from transcript CG14570-RA', other_designations='CG14570-PA|uncharacterized protein', map_location='78E5-78E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21729675, end_position_on_the_genomic_accession=21730454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40417, CurrentID=0, Status='live', Symbol='CG14569', Aliases='Dmel_ Dmel\\CG14569', description='CG14569 gene product from transcript CG14569-RA', other_designations='CG14569-PA|uncharacterized protein', map_location='78E5-78E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21730725, end_position_on_the_genomic_accession=21731550, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40418, CurrentID=0, Status='live', Symbol='CG14568', Aliases='Dmel_ BcDNA:RE21944, Dmel\\CG14568', description='CG14568 gene product from transcript CG14568-RA', other_designations='CG14568-PA|uncharacterized protein', map_location='78E5-78E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21732806, end_position_on_the_genomic_accession=21733487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40419, CurrentID=0, Status='live', Symbol='CG14573', Aliases='Dmel_ Dmel\\CG14573', description='CG14573 gene product from transcript CG14573-RA', other_designations='CG14573-PA|uncharacterized protein', map_location='78E5-78E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21734099, end_position_on_the_genomic_accession=21734788, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40420, CurrentID=0, Status='live', Symbol='CG14567', Aliases='Dmel_ Dmel\\CG14567', description='CG14567 gene product from transcript CG14567-RA', other_designations='CG14567-PA|uncharacterized protein', map_location='78E6-78E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21737706, end_position_on_the_genomic_accession=21738430, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40421, CurrentID=0, Status='live', Symbol='CG14566', Aliases='Dmel_ Dmel\\CG14566', description='CG14566 gene product from transcript CG14566-RB', other_designations='CG14566-PA|CG14566-PB|uncharacterized protein', map_location='78E6-78E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21739479, end_position_on_the_genomic_accession=21740309, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40422, CurrentID=0, Status='live', Symbol='CG14572', Aliases='Dmel_ Dmel\\CG14572', description='CG14572 gene product from transcript CG14572-RA', other_designations='CG14572-PA|uncharacterized protein', map_location='78E6-78E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21741534, end_position_on_the_genomic_accession=21742441, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40423, CurrentID=0, Status='live', Symbol='CG14565', Aliases='Dmel_ Dmel\\CG14565', description='CG14565 gene product from transcript CG14565-RA', other_designations='CG14565-PA|uncharacterized protein', map_location='78E6-78E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21743248, end_position_on_the_genomic_accession=21744269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40424, CurrentID=0, Status='live', Symbol='Syn1', Aliases='Dmel_CG7152, CG7151, CG7152, DmSYN-1, Dmel\\CG7152', description='Syntrophin-like 1', other_designations='CG7152-PB|CG7152-PC|CG7152-PD|CG7152-PE|Syn1-PB|Syn1-PC|Syn1-PD|Syn1-PE|syntrophin', map_location='78E6-78F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21745359, end_position_on_the_genomic_accession=21770534, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40425, CurrentID=0, Status='live', Symbol='CG14564', Aliases='Dmel_ Dmel\\CG14564', description='CG14564 gene product from transcript CG14564-RA', other_designations='CG14564-PA|uncharacterized protein', map_location='78F1-78F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21757157, end_position_on_the_genomic_accession=21758046, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40427, CurrentID=0, Status='live', Symbol='CG7370', Aliases='Dmel_ BcDNA:GM07065, Dmel\\CG7370', description='CG7370 gene product from transcript CG7370-RA', other_designations='CG7370-PA', map_location='78F1-78F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21769815, end_position_on_the_genomic_accession=21771321, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40428, CurrentID=0, Status='live', Symbol='eg', Aliases='Dmel_CG7383, CG7383, Dmel\\CG7383, EGON, Eg, Egl, Egon, NR0A3, T1, eagl, egon, spy, eg', description='eagle', other_designations='CG7383-PA|CG7383-PB|CG7383-PC|T1 transcription unit|eg-PA|eg-PB|eg-PC|embryonic gonad|spready', map_location='3-47.3 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21808703, end_position_on_the_genomic_accession=21817600, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40429, CurrentID=0, Status='live', Symbol='CycH', Aliases='Dmel_CG7405, 142767_at, CCNH, CG7405, Dmel\\CG7405', description='Cyclin H', other_designations='CG7405-PA|CycH-PA|CyclinH|cyclin H|cyclin-H', map_location='78F3-78F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21821997, end_position_on_the_genomic_accession=21823309, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40430, CurrentID=0, Status='live', Symbol='CG7407', Aliases='Dmel_ BcDNA:LD29166, Dmel\\CG7407', description='CG7407 gene product from transcript CG7407-RA', other_designations='CG7407-PA', map_location='78F3-78F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21823461, end_position_on_the_genomic_accession=21824571, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40431, CurrentID=0, Status='live', Symbol='CG7414', Aliases='Dmel_ Dmel\\CG7414, anon-WO0118547.243', description='CG7414 gene product from transcript CG7414-RC', other_designations='CG7414-PA|CG7414-PC|CG7414-PD|CG7414-PE', map_location='78F3-78F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21824740, end_position_on_the_genomic_accession=21827856, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40432, CurrentID=0, Status='live', Symbol='CG7148', Aliases='Dmel_ BEST:GH10766, Dmel\\CG7148, FBXL12', description='CG7148 gene product from transcript CG7148-RA', other_designations='CG7148-PA', map_location='78F4-78F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21830544, end_position_on_the_genomic_accession=21832253, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40433, CurrentID=0, Status='live', Symbol='Nopp140', Aliases='Dmel_CG7421, CG7421, Dm Dmel\\CG7421, Nopp- 140, nopp140', description='CG7421 gene product from transcript CG7421-RE', other_designations='CG7421-PA|CG7421-PB|CG7421-PE|CG7421-PF|Nopp140-PA|Nopp140-PB|Nopp140-PE|Nopp140-PF', map_location='78F4-78F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21832774, end_position_on_the_genomic_accession=21836685, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40434, CurrentID=0, Status='live', Symbol='P5CDh1', Aliases='Dmel_CG7145, CG7145, DmP5CDh1, Dmel\\CG7145, P5CDh', description='delta-1-Pyrroline-5-carboxylate dehydrogenase 1', other_designations='CG7145-PA|CG7145-PB|CG7145-PD|CG7145-PE|CG7145-PF|P5CDh1-PA|P5CDh1-PB|P5CDh1-PD|P5CDh1-PE|P5CDh1-PF|delta-1-pyrroline-5-carboxylate dehydrogenase', map_location='78F4-78F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21836665, end_position_on_the_genomic_accession=21841356, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40435, CurrentID=0, Status='live', Symbol='CG14563', Aliases='Dmel_ Dmel\\CG14563', description='CG14563 gene product from transcript CG14563-RA', other_designations='CG14563-PA|uncharacterized protein', map_location='78F4-78F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21841808, end_position_on_the_genomic_accession=21842386, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40436, CurrentID=0, Status='live', Symbol='mub', Aliases='Dmel_CG7437, CG7437, Dmel\\CG7437, MRE17, MUB, Mub, anon-WO0172774.33, anon-WO0172774.35, anon-WO0172774.36, anon-WO0172774.37, l(3)04093, l(3)j6B10, l(3)j8B3', description='mushroom-body expressed', other_designations='CG7437-PA|CG7437-PC|CG7437-PD|CG7437-PE|CG7437-PF|CG7437-PG|CG7437-PH|CG7437-PI|CG7437-PK|CG7437-PL|CG7437-PM|CG7437-PN|Mushroom-body expressed|mRNA-like ncRNA in embryogenesis 17|mub-PA|mub-PC|mub-PD|mub-PE|mub-PF|mub-PG|mub-PH|mub-PI|mub-PK|mub-PL|mub-PM|mub-PN|mushroom bodies|mushroom body expressed|poly (rC) binding protein', map_location='78F4-79A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21844788, end_position_on_the_genomic_accession=21932702, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  40437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40437, CurrentID=0, Status='live', Symbol='CG7442', Aliases='Dmel_ Dmel\\CG7442', description='CG7442 gene product from transcript CG7442-RA', other_designations='CG7442-PA', map_location='79A4-79A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21934704, end_position_on_the_genomic_accession=21938636, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40438, CurrentID=0, Status='live', Symbol='DNApol-eta', Aliases='Dmel_CG7143, AAF51794, CG7143, DNA pol eta, DNA pol-eta, DNApoleta, Dmel\\CG7143, Dmpol eta, POLH, dPoleta, drad30A, pol eta, pol h', description='CG7143 gene product from transcript CG7143-RA', other_designations='CG7143-PA|DNA polymerase eta|DNApol-eta-PA|polymerase eta', map_location='79A4-79A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21938634, end_position_on_the_genomic_accession=21941812, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40439, CurrentID=0, Status='live', Symbol='CG14562', Aliases='Dmel_ Dmel\\CG14562, HPS3, NP_649372, anon-WO0140519.241', description='CG14562 gene product from transcript CG14562-RB', other_designations='CG14562-PA|CG14562-PB|uncharacterized protein', map_location='79A4-79A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21942221, end_position_on_the_genomic_accession=21949866, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40441, CurrentID=0, Status='live', Symbol='CG7458', Aliases='Dmel_ Dmel\\CG7458', description='CG7458 gene product from transcript CG7458-RA', other_designations='CG7458-PA', map_location='79A4-79A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21955110, end_position_on_the_genomic_accession=21958041, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40442, CurrentID=0, Status='live', Symbol='CS-2', Aliases='Dmel_CG7464, CG7464, ChSA, DmCS-2, DmeChSA, Dmel\\CG7464', description='Chitin synthase 2', other_designations='CG7464-PC|CG7464-PD|CS-2-PC|CS-2-PD|chitin Syn2|chitin synthase 2|chitin synthase A|chitin-synthase', map_location='79A5-79A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21962401, end_position_on_the_genomic_accession=21969964, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40443, CurrentID=0, Status='live', Symbol='CG7470', Aliases='Dmel_ Dmel\\CG7470', description='CG7470 gene product from transcript CG7470-RB', other_designations='CG7470-PA|CG7470-PB|glutamate 5-kinase', map_location='79A5-79A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21972002, end_position_on_the_genomic_accession=21984606, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  40444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40444, CurrentID=0, Status='live', Symbol='Act79B', Aliases='Dmel_CG7478, 143060_f_at, Actin, CG7478, D, Dmel\\CG7478, act 79B, act79B, actin, mus79B', description='Actin 79B', other_designations='79B actin|Act79B-PA|Act79B-PB|CG7478-PA|CG7478-PB|F-actin|actin|actin 79B|actin-79B', map_location='79A6-79A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21985210, end_position_on_the_genomic_accession=21987190, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40445, CurrentID=0, Status='live', Symbol='CG7140', Aliases='Dmel_ Dmel\\CG7140', description='CG7140 gene product from transcript CG7140-RB', other_designations='CG7140-PB', map_location='79B1-79B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22026230, end_position_on_the_genomic_accession=22028325, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40447, CurrentID=0, Status='live', Symbol='CG14561', Aliases='Dmel_ Dmel\\CG14561', description='CG14561 gene product from transcript CG14561-RB', other_designations='CG14561-PA|CG14561-PB|uncharacterized protein', map_location='79B2-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22065907, end_position_on_the_genomic_accession=22067216, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40448, CurrentID=0, Status='live', Symbol='CG7139', Aliases='Dmel_ BcDNA:LD21293, Dmel\\CG7139, anon-WO0118547.415, mdcds_30855', description='CG7139 gene product from transcript CG7139-RA', other_designations='CG7139-PA', map_location='79B2-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22067270, end_position_on_the_genomic_accession=22071848, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40449, CurrentID=0, Status='live', Symbol='CG7133', Aliases='Dmel_ Dmel\\CG7133', description='CG7133 gene product from transcript CG7133-RA', other_designations='CG7133-PA', map_location='79B2-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22073068, end_position_on_the_genomic_accession=22074523, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40450, CurrentID=0, Status='live', Symbol='CG7130', Aliases='Dmel_ Dmel\\CG7130', description='CG7130 gene product from transcript CG7130-RA', other_designations='CG7130-PA', map_location='79B2-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22075056, end_position_on_the_genomic_accession=22075756, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40451, CurrentID=0, Status='live', Symbol='RpLP0', Aliases='Dmel_CG7490, AP3, Ap3, Ape, CG7490, Dmel\\CG7490, LP0, P0, P35, PO, RLA0, RPL P0, RpP0, RpPO, Rplp0, l(3)0154, l(3)01544, p0', description='Ribosomal protein LP0', other_designations='60S ribosomal protein P0|ACIDIC RIBOSOMAL PROTEIN LP0|CG7490-PA|CG7490-PB|RpLP0-PA|RpLP0-PB|apurinic endonuclease 3|ribosomal protein LP0|ribosomal protein P0', map_location='79B2-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22075867, end_position_on_the_genomic_accession=22077487, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40453, CurrentID=0, Status='live', Symbol='olf413', Aliases='Dmel_CG12673, CG12673, CG14461, CG7495, CR33185, Dmel\\CG12673, Olf413', description='CG12673 gene product from transcript CG12673-RC', other_designations='CG12673-PB|CG12673-PC|olf413-PB|olf413-PC', map_location='79C2-79D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22136090, end_position_on_the_genomic_accession=22247702, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  40456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40456, CurrentID=0, Status='live', Symbol='Rich', Aliases='Dmel_CG9063, 3L6, BcDNA:GH03694, BcDNAGH03694, CG9063, Dmel\\CG9063, rich', description='RIC1 homolog', other_designations='CG9063-PA|CG9063-PB|Rich-PA|Rich-PB|ric1 homolog|ric1 homologue', map_location='79D2-79D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22257749, end_position_on_the_genomic_accession=22262921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40457, CurrentID=0, Status='live', Symbol='Ddx1', Aliases='Dmel_CG9054, CG9054, DDX1, Dbp79E, DmDDX1, DmRH12, Dmel\\CG9054, anon-79E', description='Dead-box-1', other_designations='CG9054-PA|Ddx1-PA', map_location='79D3-79D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22263369, end_position_on_the_genomic_accession=22266017, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40458, CurrentID=0, Status='live', Symbol='CG11523', Aliases='Dmel_ Dmel\\CG11523', description='CG11523 gene product from transcript CG11523-RA', other_designations='CG11523-PA|uncharacterized protein', map_location='79D3-79D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22265806, end_position_on_the_genomic_accession=22266679, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40459, CurrentID=0, Status='live', Symbol='Csp', Aliases='Dmel_CG6395, CG6395, CSP, CSP32, Cys, D-CSP, D.M.CSP, DCSP-1, DNAJC5, DSP/DNAJC5, Dcsp, Dmel\\CG6395, MAB ab49, ab49, anon-WO02059370.50, csp, csp29, csp32, l(3)03988', description='Cysteine string protein', other_designations='CG6395-PA|CG6395-PB|CG6395-PC|CG6395-PD|CG6395-PE|CG6395-PF|Csp-PA|Csp-PB|Csp-PC|Csp-PD|Csp-PE|Csp-PF|cystein string protein|cysteine string protein|cysteine string protein 3|cysteine-string protein', map_location='79D3-79D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22267382, end_position_on_the_genomic_accession=22273745, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40461, CurrentID=0, Status='live', Symbol='Srpk79D', Aliases='Dmel_CG11489, BEST:CK01209, CG11489, CG9085, CK01209, Dmel\\CG11489, SRPK, SRPK1, atc, srpk79D', description='Serine-arginine protein kinase at 79D', other_designations='CG11489-PB|CG11489-PE|CG11489-PF|CG11489-PH|CG11489-PI|CG11489-PJ|CG11489-PK|CG11489-PL|CG11489-PM|CG11489-PN|CG11489-PO|Srpk79D-PB|Srpk79D-PE|Srpk79D-PF|Srpk79D-PH|Srpk79D-PI|Srpk79D-PJ|Srpk79D-PK|Srpk79D-PL|Srpk79D-PM|Srpk79D-PN|Srpk79D-PO|air traffic controller|serine-arginine protein kinase at 79D', map_location='79D4-79D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22273985, end_position_on_the_genomic_accession=22284335, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40462, CurrentID=0, Status='live', Symbol='Hem', Aliases='Dmel_CG5837, CG5837, Dmel\\CG5837, HEM1, Kette, Kette/Nap1, Nap, Nap1, dhem, dhem-2, kette, kte, l(3)03335, Hem', description='HEM-protein', other_designations='CG5837-PA|Hem-PA|hemkette|kette', map_location='79D4-79D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22284354, end_position_on_the_genomic_accession=22288195, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40464, CurrentID=0, Status='live', Symbol='Ten-m', Aliases='Dmel_CG5723, 2017, 5301, BcDNA:AT25108, CG11452, CG5723, CT16449, Dmel\\CG5723, Odzc, Ten79E, Ten[m], l(3)00844, l(3)05301, l(3)05309, l(3)rJ307, l(3)rL201, l(3)rP126, odd(Oz), odz, odz/ten-m, ten(m), ten-m, ten-m/odz, ten[m], Ten-m', description='Tenascin major', other_designations='CG5723-PB|CG5723-PD|CG5723-PE|Ten-m-PB|Ten-m-PD|Ten-m-PE|odd Oz|odd-Oz|tenascin major|tenascin-like molecule major|tenascin-major|teneurin', map_location='79D4-79E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22293044, end_position_on_the_genomic_accession=22407863, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  40467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40467, CurrentID=0, Status='live', Symbol='CG11449', Aliases='Dmel_ Dmel\\CG11449', description='CG11449 gene product from transcript CG11449-RA', other_designations='CG11449-PA|uncharacterized protein', map_location='79E4-79E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22454373, end_position_on_the_genomic_accession=22455989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40468, CurrentID=0, Status='live', Symbol='laza', Aliases='Dmel_CG11440, CG11440, Dmel\\CG11440, LPP, LPP-like, Laza, WUN-like', description='lazaro', other_designations='CG11440-PA|laza-PA', map_location='79E4-79E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22458816, end_position_on_the_genomic_accession=22461131, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40469, CurrentID=0, Status='live', Symbol='CG11438', Aliases='Dmel_ Css1alpha, Dmel\\CG11438, LPP-like, PAP2, WUN-like', description='CG11438 gene product from transcript CG11438-RB', other_designations='CG11438-PA|CG11438-PB|uncharacterized protein', map_location='79E4-79E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22465619, end_position_on_the_genomic_accession=22468076, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40470, CurrentID=0, Status='live', Symbol='CG11437', Aliases='Dmel_ Dmel\\CG11437, LPP-like, WUN-like', description='CG11437 gene product from transcript CG11437-RA', other_designations='CG11437-PA|uncharacterized protein', map_location='79E4-79E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22468862, end_position_on_the_genomic_accession=22469853, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40471, CurrentID=0, Status='live', Symbol='CG11426', Aliases='Dmel_ Css1beta, Dmel\\CG11426, LPP, LPP-like, PAP2, WUN-like', description='CG11426 gene product from transcript CG11426-RA', other_designations='CG11426-PA|lipid phosphate phosphohydrolase|uncharacterized protein', map_location='79E4-79E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22470556, end_position_on_the_genomic_accession=22473509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40472, CurrentID=0, Status='live', Symbol='CG11425', Aliases='Dmel_ Dmel\\CG11425, LPP-like, WUN-like', description='CG11425 gene product from transcript CG11425-RA', other_designations='CG11425-PA|uncharacterized protein', map_location='79E4-79E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22474795, end_position_on_the_genomic_accession=22475792, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40474, CurrentID=0, Status='live', Symbol='CG11404', Aliases='Dmel_ BcDNA:GH11730, Dmel\\CG11404', description='CG11404 gene product from transcript CG11404-RA', other_designations='CG11404-PA|CG11404-PB|uncharacterized protein', map_location='79E8-79E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22548437, end_position_on_the_genomic_accession=22549342, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40475, CurrentID=0, Status='live', Symbol='Trxr-2', Aliases='Dmel_CG11401, CG11401, Dm-TrxR, DmTrx, DmTrxR-2, Dmel\\CG11401, TRR, TrxR, Trxr2, trxr-2', description='thioredoxin reductase 2', other_designations='CG11401-PA|Trxr-2-PA|thioredoxin reductase|thioredoxin-reductase-2', map_location='79E8-79E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22550226, end_position_on_the_genomic_accession=22552019, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40476, CurrentID=0, Status='live', Symbol='CG14459', Aliases='Dmel_ Dmel\\CG14459', description='CG14459 gene product from transcript CG14459-RD', other_designations='CG14459-PC|CG14459-PD|CG14459-PE|uncharacterized protein', map_location='79F1-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22573114, end_position_on_the_genomic_accession=22637916, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40477, CurrentID=0, Status='live', Symbol='ND-B14.5AL', Aliases='Dmel_CG6914, BcDNA:AT13913, CG6914, Dmel\\CG6914, N4AM', description='NADH dehydrogenase (ubiquinone) B14.5 A subunit-like', other_designations='CG6914-PA|NADH-ubiquinone oxidoreductase subunit B14.5a|NADH:ubiquinone oxidoreductase B14.5A subunit|ND-B14.5AL-PA', map_location='79F2-79F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22606369, end_position_on_the_genomic_accession=22607126, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40479, CurrentID=0, Status='live', Symbol='CG14457', Aliases='Dmel_ BcDNA:LP08817, CG18618, Dmel\\CG14457', description='CG14457 gene product from transcript CG14457-RA', other_designations='CG14457-PA|CG14457-PB|CG14457-PC|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22669956, end_position_on_the_genomic_accession=22680034, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40480, CurrentID=0, Status='live', Symbol='CG14455', Aliases='Dmel_ Dmel\\CG14455', description='CG14455 gene product from transcript CG14455-RB', other_designations='CG14455-PB|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22686360, end_position_on_the_genomic_accession=22687087, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40481, CurrentID=0, Status='live', Symbol='CG14456', Aliases='Dmel_ Dmel\\CG14456', description='CG14456 gene product from transcript CG14456-RB', other_designations='CG14456-PA|CG14456-PB|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22687222, end_position_on_the_genomic_accession=22688288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40482, CurrentID=0, Status='live', Symbol='CG14454', Aliases='Dmel_ BcDNA:RE01153, CT34135, Dmel\\CG14454', description='CG14454 gene product from transcript CG14454-RA', other_designations='CG14454-PA|CG14454-PB|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22691674, end_position_on_the_genomic_accession=22692528, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40483, CurrentID=0, Status='live', Symbol='CG12546', Aliases='Dmel_ BcDNA:RE45259, Dmel\\CG12546', description='CG12546 gene product from transcript CG12546-RA', other_designations='CG12546-PA|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22693324, end_position_on_the_genomic_accession=22693812, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40484, CurrentID=0, Status='live', Symbol='CG14453', Aliases='Dmel_ BcDNA:RE13473, Dmel\\CG14453', description='CG14453 gene product from transcript CG14453-RA', other_designations='CG14453-PA|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22694083, end_position_on_the_genomic_accession=22694648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40485, CurrentID=0, Status='live', Symbol='CG14452', Aliases='Dmel_ BcDNA:RE59202, Dmel\\CG14452', description='CG14452 gene product from transcript CG14452-RA', other_designations='CG14452-PA|uncharacterized protein', map_location='79F5-79F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22695953, end_position_on_the_genomic_accession=22696435, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40486, CurrentID=0, Status='live', Symbol='CG11370', Aliases='Dmel_ Dmel\\CG11370', description='CG11370 gene product from transcript CG11370-RA', other_designations='CG11370-PA|CG11370-PB|uncharacterized protein', map_location='79F6-79F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22712155, end_position_on_the_genomic_accession=22713660, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40487, CurrentID=0, Status='live', Symbol='ArfGAP3', Aliases='Dmel_CG6838, CG6838, Dmel\\CG6838, arfgap1, dArfGAP3', description='ADP-ribosylation factor GTPase activating protein 3', other_designations='ArfGAP3-PA|ArfGAP3-PB|ArfGAP3-PC|ArfGAP3-PE|ArfGAP3-PF|ArfGAP3-PG|ArfGAP3-PH|CG6838-PA|CG6838-PB|CG6838-PC|CG6838-PE|CG6838-PF|CG6838-PG|CG6838-PH', map_location='79F6-79F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22714313, end_position_on_the_genomic_accession=22717577, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40488, CurrentID=0, Status='live', Symbol='CG14451', Aliases='Dmel_ Dmel\\CG14451', description='CG14451 gene product from transcript CG14451-RA', other_designations='CG14451-PA|uncharacterized protein', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22721230, end_position_on_the_genomic_accession=22722091, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40489, CurrentID=0, Status='live', Symbol='mael', Aliases='Dmel_CG11254, CG11254, Dmel\\CG11254, MAEL, Mael, mae1', description='maelstrom', other_designations='CG11254-PA|CG11254-PB|CG11254-PC|CG11254-PD|mael-PA|mael-PB|mael-PC|mael-PD', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22721203, end_position_on_the_genomic_accession=22726234, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40490, CurrentID=0, Status='live', Symbol='CG14450', Aliases='Dmel_ Dmel\\CG14450', description='CG14450 gene product from transcript CG14450-RA', other_designations='CG14450-PA|CG14450-PB|uncharacterized protein', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22726234, end_position_on_the_genomic_accession=22727789, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40491, CurrentID=0, Status='live', Symbol='CG11367', Aliases='Dmel_ Dmel\\CG11367', description='CG11367 gene product from transcript CG11367-RA', other_designations='CG11367-PA|uncharacterized protein', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22728111, end_position_on_the_genomic_accession=22733446, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40492, CurrentID=0, Status='live', Symbol='CG11241', Aliases='Dmel_ Dmel\\CG11241', description='CG11241 gene product from transcript CG11241-RC', other_designations='CG11241-PA|CG11241-PB|CG11241-PC|uncharacterized protein', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22732013, end_position_on_the_genomic_accession=22738420, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40495, CurrentID=0, Status='live', Symbol='SPoCk', Aliases='Dmel_CG32451, CG14449, CG32451, CG7651, Dmel\\CG32451, SPoCK, anon-WO0118547.249', description='Secretory Pathway Calcium atpase', other_designations='CG32451-PA|CG32451-PB|CG32451-PC|CG32451-PD|CG32451-PE|CG32451-PF|CG32451-PG|SPoCk-PA|SPoCk-PB|SPoCk-PC|SPoCk-PD|SPoCk-PE|SPoCk-PF|SPoCk-PG|secretory pathway calcium atpase', map_location='80A1-80A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22749444, end_position_on_the_genomic_accession=22787631, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  40497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40497, CurrentID=0, Status='live', Symbol='CG14448', Aliases='Dmel_ Dmel\\CG14448', description='CG14448 gene product from transcript CG14448-RA', other_designations='CG14448-PA|uncharacterized protein', map_location='80A2-80A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22788439, end_position_on_the_genomic_accession=22789247, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40501, CurrentID=0, Status='live', Symbol='CG11226', Aliases='Dmel_ BcDNA:AT16729, Dmel\\CG11226', description='CG11226 gene product from transcript CG11226-RA', other_designations='CG11226-PA|uncharacterized protein', map_location='80A4-80A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22837068, end_position_on_the_genomic_accession=22838170, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40502, CurrentID=0, Status='live', Symbol='CG32459', Aliases='Dmel_ CG8220, Dmel\\CG32459, anon-WO0140519.31', description='CG32459 gene product from transcript CG32459-RA', other_designations='CG32459-PA|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22854583, end_position_on_the_genomic_accession=22856059, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40503, CurrentID=0, Status='live', Symbol='CG13239', Aliases='Dmel_ Dmel\\CG13239', description='CG13239 gene product from transcript CG13239-RA', other_designations='CG13239-PA|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22860787, end_position_on_the_genomic_accession=22861490, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40504, CurrentID=0, Status='live', Symbol='CG33169', Aliases='Dmel_ BcDNA:RH56317, CG8244, Dmel\\CG33169, EP3524', description='CG33169 gene product from transcript CG33169-RA', other_designations='CG33169-PA|CG33169-PC|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22863864, end_position_on_the_genomic_accession=22864496, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40505, CurrentID=0, Status='live', Symbol='CG11137', Aliases='Dmel_ BcDNA:LD31024, CT31141, Dmel\\CG11137', description='CG11137 gene product from transcript CG11137-RB', other_designations='CG11137-PA|CG11137-PB|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22867175, end_position_on_the_genomic_accession=22868243, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40506, CurrentID=0, Status='live', Symbol='Arf79F', Aliases='Dmel_CG8385, ARF-1, ARF1, ARF1_DROME, Arf, Arf1, Arf79f, CG8385, Darf1, Dm Arf79F, Dmel\\CG8385, P61209, arf1, d-Arf1, dARF79F, dARFI, dArf1', description='ADP ribosylation factor at 79F', other_designations='ADP ribosylation factor 79F|ADP-ribosylation factor 1|Arf79F-PA|Arf79F-PB|Arf79F-PC|Arf79F-PE|Arf79F-PF|Arf79F-PH|Arf79F-PI|Arf79F-PJ|CG8385-PA|CG8385-PB|CG8385-PC|CG8385-PE|CG8385-PF|CG8385-PH|CG8385-PI|CG8385-PJ', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22868380, end_position_on_the_genomic_accession=22872231, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40507, CurrentID=0, Status='live', Symbol='CG11109', Aliases='Dmel_ CT31087, Dmel\\CG11109', description='CG11109 gene product from transcript CG11109-RB', other_designations='CG11109-PA|CG11109-PB|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22872200, end_position_on_the_genomic_accession=22873702, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40508, CurrentID=0, Status='live', Symbol='Chro', Aliases='Dmel_CG10712, CG10712, CHRIZ, CHRIZ/CHRO, CHRO, Chiz, Chr, Chriz(Chriz), Dmel\\CG10712, anon-EST:fe2E3, Chro', description='Chromator', other_designations='CG10712-PA|CG10712-PB|CG10712-PC|CG10712-PD|CG10712-PE|Chro-PA|Chro-PB|Chro-PC|Chro-PD|Chro-PE|anon-fast-evolving-2E3|chromator|chromo domain protein interacting with Z4|chromobox protein interacting with Z4', map_location='80B1-80B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22873702, end_position_on_the_genomic_accession=22878984, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40509, CurrentID=0, Status='live', Symbol='Ssl1', Aliases='Dmel_CG11115, CG11115, Dmel\\CG11115, SSL1, p44, ssl1', description='Suppressor of Stem-Loop mutation ortholog (S. cerevisiae)', other_designations='CG11115-PA|CG11115-PB|Ssl1-PA|Ssl1-PB|suppressor of Stem-Loop mutation ortholog', map_location='80B2-80B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22879089, end_position_on_the_genomic_accession=22880963, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40510, CurrentID=0, Status='live', Symbol='slif', Aliases='Dmel_CG11128, CG11128, Dmel\\CG11128, SlifA, slif', description='slimfast', other_designations='CG11128-PA|CG11128-PB|CG11128-PC|slif-PA|slif-PB|slif-PC', map_location='80B2-80B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22881045, end_position_on_the_genomic_accession=22893821, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40511, CurrentID=0, Status='live', Symbol='CG11131', Aliases='Dmel_ Dmel\\CG11131', description='CG11131 gene product from transcript CG11131-RB', other_designations='CG11131-PA|CG11131-PB|uncharacterized protein', map_location='80B2-80B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22898744, end_position_on_the_genomic_accession=22900262, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40512, CurrentID=0, Status='live', Symbol='BoYb', Aliases='Dmel_CG11133, CG11133, Dmel\\CG11133', description='Brother of Yb', other_designations='BoYb-PA|CG11133-PA|brother of Yb', map_location='80B2-80B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22901912, end_position_on_the_genomic_accession=22905470, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40513, CurrentID=0, Status='live', Symbol='CG12768', Aliases='Dmel_ Dmel\\CG12768', description='CG12768 gene product from transcript CG12768-RD', other_designations='CG12768-PA|CG12768-PB|CG12768-PC|CG12768-PD|uncharacterized protein', map_location='80B2-80B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22910491, end_position_on_the_genomic_accession=22923532, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40514, CurrentID=0, Status='live', Symbol='Mes2', Aliases='Dmel_CG11100, CG11100, Dmel\\CG11100, adfR, mes2', description='Mesoderm-expressed 2', other_designations='CG11100-PB|CG11100-PC|Mes2-PB|Mes2-PC', map_location='80C1-80C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22945657, end_position_on_the_genomic_accession=22952248, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40515, CurrentID=0, Status='live', Symbol='nrm', Aliases='Dmel_CG43079, CG32458, CG41452, CG43079, CG8779, CT16082, Dmel\\CG43079, Dmel_CG32458, Dmel_CG41452, Dmel_CG8779', description='neuromusculin', other_designations='CG43079-PA|CG43079-PB|CG43079-PE|CG43079-PF|CG43079-PH|CG43079-PI|nrm-PA|nrm-PB|nrm-PE|nrm-PF|nrm-PH|nrm-PI', map_location='80C3-80C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22998372, end_position_on_the_genomic_accession=23040300, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  40517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40517, CurrentID=0, Status='live', Symbol='alpha-Cat', Aliases='Dmel_CG17947, CG17947, CT39986, Dalpha-Cat, Dalpha-cat, DalphaC, Dalphacat, Dmcatp, Dmel\\CG17947, LD30423p, a-Cat, a-cat, alpha-CAT, alpha-cat, alpha-catenin, alpha-ctn, alphaCAT, alphaCat, alphacat', description='alpha Catenin', other_designations='CG17947-PA|CG17947-PB|D-alpha-catenin|Dalpha-Catenin|Dalpha-cadherin|a-catenin|alpha catenin|alpha-Cat-PA|alpha-Cat-PB|alpha-Catenin|alphaCatenin|dalphacatenin', map_location='80F1-80F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23318646, end_position_on_the_genomic_accession=23338100, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  40521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40521, CurrentID=0, Status='live', Symbol='nAChRalpha4', Aliases='Dmel_CG12414, CG12414, CG17552, Dalpha4, Dmel\\CG12414, anon-WO0138359.1, nAChR, nAcRalpha-4, nAcRalpha-80, nAcRalpha-80B, nAcRaplha-80B, rye', description='nicotinic Acetylcholine Receptor alpha4', other_designations='CG12414-PC|CG12414-PD|CG12414-PE|CG12414-PF|CG12414-PG|CG12414-PH|Dalpha4|nAChRalpha4-PC|nAChRalpha4-PD|nAChRalpha4-PE|nAChRalpha4-PF|nAChRalpha4-PG|nAChRalpha4-PH|nicotinic acetylcholine receptor alpha 4|nicotinic acetylcholine receptor alpha 80B|nicotinic acetylcholine receptor alpha4|redeye', map_location='80D5-80E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23206669, end_position_on_the_genomic_accession=23287448, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  40522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40522, CurrentID=0, Status='live', Symbol='CG12581', Aliases='Dmel_ Dmel\\CG12581', description='CG12581 gene product from transcript CG12581-RB', other_designations='CG12581-PA|CG12581-PB|uncharacterized protein', map_location='81F5-81F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4174632, end_position_on_the_genomic_accession=4186751, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40524, CurrentID=0, Status='live', Symbol='beta-Man', Aliases='Dmel_CG12582, CG12582, CG12583, Dmel\\CG12582', description='beta-Mannosidase', other_designations='CG12582-PA|CG12582-PD|CG12582-PF|CG12582-PG|CG12582-PH|beta-Man-PA|beta-Man-PD|beta-Man-PF|beta-Man-PG|beta-Man-PH', map_location='81F6-81F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4197206, end_position_on_the_genomic_accession=4204572, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40526, CurrentID=0, Status='live', Symbol='CG14636', Aliases='Dmel_ Dmel\\CG14636', description='CG14636 gene product from transcript CG14636-RA', other_designations='CG14636-PA|CG14636-PB|uncharacterized protein', map_location='81F6-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4204493, end_position_on_the_genomic_accession=4215314, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40527, CurrentID=0, Status='live', Symbol='aux', Aliases='Dmel_CG1107, Aux, Auxillin/DNAJC6, CG1107, DNAJC6, Dmel\\CG1107, GAK, anon-WO0118547.393illin, dAux, aux', description='auxilin', other_designations='CG1107-PA|CG1107-PB|CG1107-PC|CG1107-PD|aux-PA|aux-PB|aux-PC|aux-PD|auxillin', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4211783, end_position_on_the_genomic_accession=4227518, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  40528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40528, CurrentID=0, Status='live', Symbol='DhpD', Aliases='Dmel_CG18143, CG18143, Dmel\\CG18143', description='Dihydropterin deaminase', other_designations='CG18143-PA|DhpD-PA|dihydropterin deaminase', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4227387, end_position_on_the_genomic_accession=4229238, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40529, CurrentID=0, Status='live', Symbol='CG14641', Aliases='Dmel_ Dmel\\CG14641, cg14641', description='CG14641 gene product from transcript CG14641-RA', other_designations='CG14641-PA|uncharacterized protein', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4230779, end_position_on_the_genomic_accession=4232342, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40530, CurrentID=0, Status='live', Symbol='abs', Aliases='Dmel_CG14637, ABS, CG14637, DmRH23, Dmel\\CG14637, anon-WO0118547.315, l(3)00620, l(3)04505, l(3)06862, l(3)06863', description='abstrakt', other_designations='CG14637-PA|abs-PA|abstract', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4233047, end_position_on_the_genomic_accession=4235041, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40532, CurrentID=0, Status='live', Symbol='CG14642', Aliases='Dmel_ Dmel\\CG14642, SP29', description='CG14642 gene product from transcript CG14642-RA', other_designations='CG14642-PA|CG14642-PB|CG14642-PC|uncharacterized protein', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4241045, end_position_on_the_genomic_accession=4242717, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40534, CurrentID=0, Status='live', Symbol='TwdlF', Aliases='Dmel_CG14639, CG14639, Dmel\\CG14639', description='TweedleF', other_designations='CG14639-PA|TwdlF-PA', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4247020, end_position_on_the_genomic_accession=4248313, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40535, CurrentID=0, Status='live', Symbol='TwdlG', Aliases='Dmel_CG14643, CG14643, Dmel\\CG14643', description='TweedleG', other_designations='CG14643-PA|CG14643-PB|CG14643-PC|TwdlG-PA|TwdlG-PB|TwdlG-PC|tweedle G', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4248723, end_position_on_the_genomic_accession=4251341, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40536, CurrentID=0, Status='live', Symbol='CG14644', Aliases='Dmel_ Dmel\\CG14644, NEST:bs06h05', description='CG14644 gene product from transcript CG14644-RA', other_designations='CG14644-PA|CG14644-PB|NEST:bs06h05|uncharacterized protein', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4251894, end_position_on_the_genomic_accession=4253043, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40537, CurrentID=0, Status='live', Symbol='TwdlV', Aliases='Dmel_CG14640, CG14640, Dmel\\CG14640', description='TweedleV', other_designations='CG14640-PA|TwdlV-PA|tweedle V', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4254708, end_position_on_the_genomic_accession=4255833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40538, CurrentID=0, Status='live', Symbol='CG1092', Aliases='Dmel_ Dmel\\CG1092', description='CG1092 gene product from transcript CG1092-RC', other_designations='CG1092-PC|CG1092-PD', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4266770, end_position_on_the_genomic_accession=4268316, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40539, CurrentID=0, Status='live', Symbol='CG9766', Aliases='Dmel_ Dmel\\CG9766', description='CG9766 gene product from transcript CG9766-RC', other_designations='CG9766-PC', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4269221, end_position_on_the_genomic_accession=4277928, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40540, CurrentID=0, Status='live', Symbol='CG9780', Aliases='Dmel_ Dmel\\CG9780', description='CG9780 gene product from transcript CG9780-RA', other_designations='CG9780-PA|neprilysin', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4292274, end_position_on_the_genomic_accession=4294840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40541, CurrentID=0, Status='live', Symbol='MP1', Aliases='Dmel_CG1102, BEST:GH02921, CG1102, Dmel\\CG1102, SP25, proPO-AE', description='Melanization Protease 1', other_designations='CG1102-PA|CG1102-PC|CG1102-PE|MP1-PA|MP1-PC|MP1-PE|melanization protease 1', map_location='82A3-82A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4307908, end_position_on_the_genomic_accession=4309835, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40542, CurrentID=0, Status='live', Symbol='Vps24', Aliases='Dmel_CG9779, CG9779, DmVPS24, Dmel\\CG9779, ESCRT, dvps24, vps24', description='Vacuolar protein sorting 24', other_designations='CG9779-PA|CG9779-PB|Vps24-PA|Vps24-PB|charged multivesicular body protein 3|vacuolar protein sorting 24', map_location='82A3-82A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4307435, end_position_on_the_genomic_accession=4310947, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40543, CurrentID=0, Status='live', Symbol='CG9791', Aliases='Dmel_ Dmel\\CG9791', description='CG9791 gene product from transcript CG9791-RD', other_designations='CG9791-PC|CG9791-PD', map_location='82A3-82A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4311555, end_position_on_the_genomic_accession=4315398, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40544, CurrentID=0, Status='live', Symbol='Syt14', Aliases='Dmel_CG9778, CG9778, Dmel\\CG9778, Syt 14, Syt16, dsyt14, syt14, syt16', description='Synaptotagmin 14', other_designations='CG9778-PA|CG9778-PC|CG9778-PD|Syt14-PA|Syt14-PC|Syt14-PD|synaptotagmin 14|syt14', map_location='82A3-82A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4314455, end_position_on_the_genomic_accession=4319069, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40545, CurrentID=0, Status='live', Symbol='CG9795', Aliases='Dmel_ Dmel\\CG9795', description='CG9795 gene product from transcript CG9795-RD', other_designations='CG9795-PA|CG9795-PD|CG9795-PE|CG9795-PF', map_location='82A3-82A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4319683, end_position_on_the_genomic_accession=4326379, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  40546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40546, CurrentID=0, Status='live', Symbol='CG9776', Aliases='Dmel_ BcDNA:LD19168, Dmel\\CG9776', description='CG9776 gene product from transcript CG9776-RA', other_designations='CG9776-PA|CG9776-PB', map_location='82A3-82A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4327814, end_position_on_the_genomic_accession=4334071, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40547, CurrentID=0, Status='live', Symbol='CG1103', Aliases='Dmel_ Dmel\\CG1103, anon-WO0118547.152', description='CG1103 gene product from transcript CG1103-RA', other_designations='CG1103-PA|CG1103-PB', map_location='82A4-82A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4337755, end_position_on_the_genomic_accession=4339914, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40548, CurrentID=0, Status='live', Symbol='Skp2', Aliases='Dmel_CG9772, 153298_at, CG9772, Dmel\\CG9772, FBXL1, SKP2, d', description='CG9772 gene product from transcript CG9772-RA', other_designations='CG9772-PA|CG9772-PE|Skp2-PA|Skp2-PE', map_location='82A4-82A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4334879, end_position_on_the_genomic_accession=4337674, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40549, CurrentID=0, Status='live', Symbol='hkb', Aliases='Dmel_CG9768, 5953, CG9768, Dmel\\CG9768, HKB, Hkb, anon-82B3, gurt', description='huckebein', other_designations='CG9768-PA|gurtelchen|hkb-PA|huckbein|hucklebein', map_location='82A4-82A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4345015, end_position_on_the_genomic_accession=4346646, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40551, CurrentID=0, Status='live', Symbol='CG1090', Aliases='Dmel_ Dmel\\CG1090', description='CG1090 gene product from transcript CG1090-RI', other_designations='CG1090-PE|CG1090-PG|CG1090-PH|CG1090-PI', map_location='82A4-82A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4359883, end_position_on_the_genomic_accession=4366855, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40552, CurrentID=0, Status='live', Symbol='CG11739', Aliases='Dmel_ Dmel\\CG11739, anon-EST:Posey55', description='CG11739 gene product from transcript CG11739-RA', other_designations='CG11739-PA|CG11739-PB|CG11739-PC|CG11739-PD|uncharacterized protein', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4378600, end_position_on_the_genomic_accession=4381210, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40553, CurrentID=0, Status='live', Symbol='Cont', Aliases='Dmel_CG1084, CG1084, CONT, CT1481, D-Axonin, DCONT, Dmel\\CG1084, cont', description='Contactin', other_designations='CG1084-PA|Cont-PA|contactin', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4381306, end_position_on_the_genomic_accession=4387018, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40554, CurrentID=0, Status='live', Symbol='tub', Aliases='Dmel_CG10520, CG10520, Dmel\\CG10520, Tube', description='tube', other_designations='CG10520-PC|tub-PC', map_location='3-47 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4387742, end_position_on_the_genomic_accession=4389813, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40555, CurrentID=0, Status='live', Symbol='CG14646', Aliases='Dmel_ Dmel\\CG14646, anon-WO0118547.373', description='CG14646 gene product from transcript CG14646-RA', other_designations='CG14646-PA|CG14646-PB|uncharacterized protein', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4390366, end_position_on_the_genomic_accession=4392111, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40556, CurrentID=0, Status='live', Symbol='CG9855', Aliases='Dmel_ Dmel\\CG9855', description='CG9855 gene product from transcript CG9855-RC', other_designations='CG9855-PA|CG9855-PC', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4395979, end_position_on_the_genomic_accession=4397875, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40557, CurrentID=0, Status='live', Symbol='CG9853', Aliases='Dmel_ BEST:LD21829, CG41469, Dmel\\CG9853, Dmel_CG41469', description='CG9853 gene product from transcript CG9853-RC', other_designations='CG9853-PA|CG9853-PB|CG9853-PC', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4398381, end_position_on_the_genomic_accession=4402027, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40558, CurrentID=0, Status='live', Symbol='CG14647', Aliases='Dmel_ Dmel\\CG14647', description='CG14647 gene product from transcript CG14647-RB', other_designations='CG14647-PB|uncharacterized protein', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4400510, end_position_on_the_genomic_accession=4402017, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40559, CurrentID=0, Status='live', Symbol='lost', Aliases='Dmel_CG14648, CG14648, Dmel\\CG14648, Growl, anon-WO0118547.126, growl, nev', description='CG14648 gene product from transcript CG14648-RA', other_designations='CG14648-PA|growl|lost boys|lost-PA|neverland', map_location='82A6-82A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4403076, end_position_on_the_genomic_accession=4407182, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40560, CurrentID=0, Status='live', Symbol='CG32944', Aliases='Dmel_ CG10532, CG9818, Dmel\\CG32944', description='CG32944 gene product from transcript CG32944-RE', other_designations='CG32944-PE|CG32944-PF|CG32944-PG|uncharacterized protein', map_location='82A6-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4407227, end_position_on_the_genomic_accession=4419578, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  40562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40562, CurrentID=0, Status='live', Symbol='srl', Aliases='Dmel_CG9809, CG 9809, CG9809, Dmel\\CG9809, PGC-1, PGC-1-alpha, Spargel, Srl, cg9809, dPCG1alpha, dPGC, dPGC-1, dPGC1', description='spargel', other_designations='CG9809-PB|CG9809-PD|CG9809-PE|peroxisome proliferator-activated receptor gamma coactivator 1|srl-PB|srl-PD|srl-PE', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4422483, end_position_on_the_genomic_accession=4429335, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40563, CurrentID=0, Status='live', Symbol='eIF3-S10', Aliases='Dmel_CG9805, CG9805, Dmel\\CG9805, EIF3-S10, MMAP140, anon-WO0118547.261, eIF3-s10, eIF3a', description='CG9805 gene product from transcript CG9805-RD', other_designations='CG9805-PA|CG9805-PD|eIF3-S10-PA|eIF3-S10-PD', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4429798, end_position_on_the_genomic_accession=4433895, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40564, CurrentID=0, Status='live', Symbol='CG1074', Aliases='Dmel_ Dmel\\CG1074, anon-WO0118547.308', description='CG1074 gene product from transcript CG1074-RA', other_designations='CG1074-PA', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4433965, end_position_on_the_genomic_accession=4436175, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40565, CurrentID=0, Status='live', Symbol='CG9804', Aliases='Dmel_ Dmel\\CG9804', description='CG9804 gene product from transcript CG9804-RA', other_designations='CG9804-PA|CG9804-PB', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4436222, end_position_on_the_genomic_accession=4437329, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40566, CurrentID=0, Status='live', Symbol='CG14650', Aliases='Dmel_ Dmel\\CG14650', description='CG14650 gene product from transcript CG14650-RA', other_designations='CG14650-PA|uncharacterized protein', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4437387, end_position_on_the_genomic_accession=4441328, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40567, CurrentID=0, Status='live', Symbol='CG31522', Aliases='Dmel_ BcDNA:GH22993, BcDNA:LD10305, CG9798, Dmel\\CG31522, anon-WO0118547.381', description='CG31522 gene product from transcript CG31522-RF', other_designations='CG31522-PA|CG31522-PB|CG31522-PD|CG31522-PF|CG31522-PG|CG31522-PH|CG31522-PI|uncharacterized protein', map_location='82B1-82B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4441415, end_position_on_the_genomic_accession=4462883, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40568, CurrentID=0, Status='live', Symbol='CG14651', Aliases='Dmel_ Dmel\\CG14651', description='CG14651 gene product from transcript CG14651-RB', other_designations='CG14651-PB|uncharacterized protein', map_location='82B3-82B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4480812, end_position_on_the_genomic_accession=4484215, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40569, CurrentID=0, Status='live', Symbol='Fip1', Aliases='Dmel_CG1078, CG1078, Dmel\\CG1078, FIP1, anon-EST:Liang-1.35, clone 1.35, d', description='CG1078 gene product from transcript CG1078-RA', other_designations='CG1078-PA|Fip1-PA', map_location='82B3-82B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4485008, end_position_on_the_genomic_accession=4487677, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40574, CurrentID=0, Status='live', Symbol='Mur82C', Aliases='Dmel_CG12586, CG12586, Dmel\\CG12586', description='Mucin related 82C', other_designations='CG12586-PB|Mur82C-PB|mucin related 82C', map_location='82C2-82C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4541356, end_position_on_the_genomic_accession=4543215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40575, CurrentID=0, Status='live', Symbol='5-HT2A', Aliases='Dmel_CG1056, 5-HT2, 5-HT2Dro, 5-HT[2]Dro, 5-HT[[2Dro]], 5-HT[[2]], 5-HT[[2]]Dro, 5-ht2Dro, 5HT-R82A, 5HT-dro[[1]], 5HT2, 5HT2-Dro, 5HT2Dro, 5HT[[2Dro]], CG1056, Dm5HT2, Dmel\\CG1056, d5-HT2, l(3)82CDb', description='5-hydroxytryptamine (serotonin) receptor 2A', other_designations='5-HT2A-PA|5-HT2A-PB|5-HT2A-PD|5-HT2A-PF|5-HT2A-PH|5-HT[[2]]Dro receptor|CG1056-PA|CG1056-PB|CG1056-PD|CG1056-PF|CG1056-PH|lethal (3) 82CDb|serotonin receptor 2', map_location='82C3-82D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4612870, end_position_on_the_genomic_accession=4633304, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40576, CurrentID=0, Status='live', Symbol='CG9775', Aliases='Dmel_ Dmel\\CG9775', description='CG9775 gene product from transcript CG9775-RE', other_designations='CG9775-PA|CG9775-PC|CG9775-PD|CG9775-PE', map_location='82C5-82C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4638010, end_position_on_the_genomic_accession=4641838, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40577, CurrentID=0, Status='live', Symbol='MED31', Aliases='Dmel_CG1057, BcDNA:GM28912, CG1057, Dmel\\CG1057, Med31, Trap18, dSOH1, dTRAP18, med31, p36', description='Mediator complex subunit 31', other_designations='CG1057-PA|CG1057-PB|CG1057-PC|CG1057-PD|CG1057-PE|MED31-PA|MED31-PB|MED31-PC|MED31-PD|MED31-PE|mediator|mediator complex subunit 31', map_location='82C5-82C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4641963, end_position_on_the_genomic_accession=4643292, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40578, CurrentID=0, Status='live', Symbol='slx1', Aliases='Dmel_CG18271, CG1827, CG18271, Dmel\\CG18271, SLX1', description='nuclease slx1', other_designations='CG18271-PB|slx1-PB', map_location='82C5-82C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4641963, end_position_on_the_genomic_accession=4644636, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40579, CurrentID=0, Status='live', Symbol='Dlip2', Aliases='Dmel_CG9771, CG9771, DIP2, Dip2, Dmel\\CG9771, dip2, dmDIP2, sip4', description='Dorsal interacting protein 2', other_designations='CG9771-PA|Dlip2-PA|dorsal interacting protein 2|septin interacting protein 4', map_location='82C5-82C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4644649, end_position_on_the_genomic_accession=4645655, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40580, CurrentID=0, Status='live', Symbol='rpk', Aliases='Dmel_CG1058, CG1058, Dmel\\CG1058, GNaC1, RPK, dGNaC1, ppk2', description='ripped pocket', other_designations='CG1058-PA|CG1058-PB|gonad-specific amiloride-sensitive sodium channel 1|rpk-PA|rpk-PB', map_location='82C5-82C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4645981, end_position_on_the_genomic_accession=4648884, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40581, CurrentID=0, Status='live', Symbol='Karybeta3', Aliases='Dmel_CG1059, CG1059, Dmel\\CG1059, Karbeta3, Kary-beta3, l(3)82CDd, l(3)j3A4, l(3)j7E8', description='Karyopherin beta 3', other_designations='CG1059-PA|CG1059-PB|Karybeta3-PA|Karybeta3-PB|karyopherin beta|karyopherin beta 3|karyopherin-beta3|lethal (3) 82CDd', map_location='82C5-82C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4649176, end_position_on_the_genomic_accession=4654624, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40582, CurrentID=0, Status='live', Symbol='spartin', Aliases='Dmel_CG12001, CG12001, Dmel\\CG12001', description='CG12001 gene product from transcript CG12001-RA', other_designations='CG12001-PA|CG12001-PB|spartin-PA|spartin-PB', map_location='82C5-82D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4654800, end_position_on_the_genomic_accession=4657985, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40583, CurrentID=0, Status='live', Symbol='CG43427', Aliases='Dmel_ CG1054, CG1055, CG14653, CG18253, CG31531, CG31534, Dmel\\CG43427, Dmel_CG31531, Dmel_CG31534, anon-WO0140519.95', description='CG43427 gene product from transcript CG43427-RP', other_designations='CG43427-PE|CG43427-PF|CG43427-PG|CG43427-PH|CG43427-PI|CG43427-PJ|CG43427-PM|CG43427-PN|CG43427-PO|CG43427-PP|uncharacterized protein', map_location='82D2-82D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4659579, end_position_on_the_genomic_accession=4712193, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  40587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40587, CurrentID=0, Status='live', Symbol='CG9769', Aliases='Dmel_ Dmel\\CG9769, Q9VN50, eIF3-S5, eIF3f, eif3f', description='CG9769 gene product from transcript CG9769-RA', other_designations='CG9769-PA', map_location='82D2-82D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4712882, end_position_on_the_genomic_accession=4714182, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40588, CurrentID=0, Status='live', Symbol='Nep2', Aliases='Dmel_CG9761, BcDNA:GH07643, CG9761, DmNEP2, Dmel\\CG9761, NEP2, dNEP2, nep2', description='Neprilysin 2', other_designations='CG9761-PA|CG9761-PB|CG9761-PC|Nep2-PA|Nep2-PB|Nep2-PC|neprilysin 2|neprilysin-2', map_location='82D2-82D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4719458, end_position_on_the_genomic_accession=4731878, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40589, CurrentID=0, Status='live', Symbol='tacc', Aliases='Dmel_CG9765, CG9765, D-TACC, D-tacc, DTACC, Dmel\\CG9765, TACC, TACCdm, acc, d-tacc, dTACC, dtacc', description='transforming acidic coiled-coil protein', other_designations='CG9765-PA|CG9765-PC|CG9765-PD|CG9765-PE|CG9765-PF|CG9765-PG|CG9765-PH|CG9765-PI|CG9765-PJ|CG9765-PK|CG9765-PL|CG9765-PM|CG9765-PN|D. melanogaster transforming acidic coiled coil|drosophila transforming acidic Coiled-Coil protein|drosophila transforming acidic coiled-coil|tacc-PA|tacc-PC|tacc-PD|tacc-PE|tacc-PF|tacc-PG|tacc-PH|tacc-PI|tacc-PJ|tacc-PK|tacc-PL|tacc-PM|tacc-PN|transforming acidic coiled-coil', map_location='82D2-82D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4733654, end_position_on_the_genomic_accession=4749056, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  40590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40590, CurrentID=0, Status='live', Symbol='CG44098', Aliases='Dmel_ CG12588, CG14654, CG31530, CG9767, Dmel\\CG44098, Dmel_CG14654, Dmel_CG31530', description='CG44098 gene product from transcript CG44098-RB', other_designations='CG44098-PB|CG44098-PC|CG44098-PD|CG44098-PE|uncharacterized protein', map_location='82D4-82D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4750409, end_position_on_the_genomic_accession=4773197, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40593, CurrentID=0, Status='live', Symbol='atms', Aliases='Dmel_CG2503, Atms, CG2503, Dmel\\CG2503, PAF, PAF1, Paf1, anon-WO0172774.113, anon-WO0172774.114, anon-WO0172774.116, l(3)01456, l(3)02466, l(3)rK509', description='antimeros', other_designations='CG2503-PA|CG2503-PB|atms-PA|atms-PB', map_location='82D3-82D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4750035, end_position_on_the_genomic_accession=4777490, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40594, CurrentID=0, Status='live', Symbol='CG14655', Aliases='Dmel_ Dmel\\CG14655', description='CG14655 gene product from transcript CG14655-RA', other_designations='CG14655-PA|uncharacterized protein', map_location='82D5-82D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4777497, end_position_on_the_genomic_accession=4780302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40595, CurrentID=0, Status='live', Symbol='CG17387', Aliases='Dmel_ Dmel\\CG17387', description='CG17387 gene product from transcript CG17387-RA', other_designations='CG17387-PA|uncharacterized protein', map_location='82D5-82D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4781120, end_position_on_the_genomic_accession=4784722, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40596, CurrentID=0, Status='live', Symbol='ctrip', Aliases='Dmel_CG42574, CG14656, CG17735, CG42574, Dmel\\CG42574, Dmel_CG14656, Dmel_CG17735', description='circadian trip', other_designations='CG42574-PA|CG42574-PB|CG42574-PC|CG42574-PE|CG42574-PF|CG42574-PH|CG42574-PI|CG42574-PJ|CG42574-PK|CG42574-PL|CG42574-PM|ctrip-PA|ctrip-PB|ctrip-PC|ctrip-PE|ctrip-PF|ctrip-PH|ctrip-PI|ctrip-PJ|ctrip-PK|ctrip-PL|ctrip-PM', map_location='82D5-82D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4786284, end_position_on_the_genomic_accession=4812096, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  40598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40598, CurrentID=0, Status='live', Symbol='Hus1-like', Aliases='Dmel_CG2525, CG2525, DmHus1, Dmel\\CG2525, Hus-1, Hus1, hus1', description='CG2525 gene product from transcript CG2525-RA', other_designations='CG2525-PA|CG2525-PB|Hus1-like-PA|Hus1-like-PB', map_location='82D6-82D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4808336, end_position_on_the_genomic_accession=4812096, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40599, CurrentID=0, Status='live', Symbol='CG1129', Aliases='Dmel_ Dmel\\CG1129', description='CG1129 gene product from transcript CG1129-RE', other_designations='CG1129-PA|CG1129-PB|CG1129-PD|CG1129-PE', map_location='82D6-82D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4811444, end_position_on_the_genomic_accession=4816025, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40600, CurrentID=0, Status='live', Symbol='CG1126', Aliases='Dmel_ BBS5, Dmel\\CG1126', description='CG1126 gene product from transcript CG1126-RA', other_designations='CG1126-PA', map_location='82D6-82D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4816318, end_position_on_the_genomic_accession=4817776, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40602, CurrentID=0, Status='live', Symbol='CG14657', Aliases='Dmel_ Dmel\\CG14657', description='CG14657 gene product from transcript CG14657-RC', other_designations='CG14657-PB|CG14657-PC|CG14657-PD|uncharacterized protein', map_location='82D6-82D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4817980, end_position_on_the_genomic_accession=4821432, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40603, CurrentID=0, Status='live', Symbol='CG14658', Aliases='Dmel_ Dmel\\CG14658', description='CG14658 gene product from transcript CG14658-RA', other_designations='CG14658-PA|uncharacterized protein', map_location='82D7-82D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4829911, end_position_on_the_genomic_accession=4830513, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40604, CurrentID=0, Status='live', Symbol='CG14659', Aliases='Dmel_ Dmel\\CG14659', description='CG14659 gene product from transcript CG14659-RA', other_designations='CG14659-PA|uncharacterized protein', map_location='82D7-82D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4830849, end_position_on_the_genomic_accession=4831297, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40605, CurrentID=0, Status='live', Symbol='opa', Aliases='Dmel_CG1133, CG1133, Dmel\\CG1133, Opa', description='odd paired', other_designations='CG1133-PA|GLI-Kr zinc finger pair-rule protein|Odd-paired|opa-PA', map_location='3-48 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4852813, end_position_on_the_genomic_accession=4869979, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40606, CurrentID=0, Status='live', Symbol='laf', Aliases='Dmel_CG14660, C6, CG14660, Dmel\\CG14660', description='labial associated factor', other_designations='CG14660-PA|CG14660-PB|clone C6|laf-PA|laf-PB', map_location='82E1-82E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4876005, end_position_on_the_genomic_accession=4878533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40607, CurrentID=0, Status='live', Symbol='Gnf1', Aliases='Dmel_CG1119, CG1119, DmRFC1, Dmel\\CG1119, GTF-1, Gnf-1, Gtf1, ONF-CAACAA, RF-C 140, RFC, RFC1, RfC, RfC140, dRF-Cp140, dRFC140, gnf1, gtf1, rfc1', description='Germ line transcription factor 1', other_designations='CG1119-PA|CG1119-PB|Gnf1-PA|Gnf1-PB|Replication-factor-C 140kD subunit|germ line transcription factor 1|germline transcription factor|replication factor C 140kD subunit', map_location='82E2-82E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4906622, end_position_on_the_genomic_accession=4910081, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40608, CurrentID=0, Status='live', Symbol='Cdep', Aliases='Dmel_CG44193, BcDNA:AT26639, BcDNA:RH50880, CDEP, CG1283, CG2008, CG31536, CG34306, CG44193, Dmel\\CG44193, Dmel_CG31536, Dmel_CG34306, i23', description='Chondrocyte-derived ezrin-like domain containing protein ortholog (H. sapiens)', other_designations='CG44193-PC|CG44193-PE|CG44193-PF|CG44193-PG|Cdep-PC|Cdep-PE|Cdep-PF|Cdep-PG|Chondrocyte-derived ezrin-like domain containing protein ortholog', map_location='82E2-82E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4910552, end_position_on_the_genomic_accession=4945576, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  40610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40610, CurrentID=0, Status='live', Symbol='Ubc6', Aliases='Dmel_CG2013, BcDNA:RE56673, CG2013, Dhr6, Dmel\\CG2013, RAD6, Rad6, Rad6a, UbcD6, dRAD6, dRad6', description='Ubiquitin conjugating enzyme 6', other_designations='CG2013-PA|CG2013-PC|Ubc6-PA|Ubc6-PC|ubiquitin conjugating enzyme|ubiquitin conjugating enzyme 6', map_location='82E3-82E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4948755, end_position_on_the_genomic_accession=4952682, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40611, CurrentID=0, Status='live', Symbol='CG14661', Aliases='Dmel_ CT34439, Dmel\\CG14661', description='CG14661 gene product from transcript CG14661-RB', other_designations='CG14661-PA|CG14661-PB|uncharacterized protein', map_location='82E4-82E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4953506, end_position_on_the_genomic_accession=4955252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40612, CurrentID=0, Status='live', Symbol='CG2016', Aliases='Dmel_ Dmel\\CG2016, anon-SAGE:Wang-083', description='CG2016 gene product from transcript CG2016-RB', other_designations='CG2016-PB|CG2016-PC|CG2016-PD|CG2016-PE|anon-SAGE:Wang-083', map_location='82E4-82E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4956999, end_position_on_the_genomic_accession=4961351, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40613, CurrentID=0, Status='live', Symbol='CG1124', Aliases='Dmel_ CT1789, Dmel\\CG1124', description='CG1124 gene product from transcript CG1124-RA', other_designations='CG1124-PA', map_location='82E4-82E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4965445, end_position_on_the_genomic_accession=4968485, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40614, CurrentID=0, Status='live', Symbol='CG14662', Aliases='Dmel_ CT34440, Dmel\\CG14662', description='CG14662 gene product from transcript CG14662-RA', other_designations='CG14662-PA|uncharacterized protein', map_location='82E4-82E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4981999, end_position_on_the_genomic_accession=4984232, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40615, CurrentID=0, Status='live', Symbol='CG2022', Aliases='Dmel_ Dmel\\CG2022', description='CG2022 gene product from transcript CG2022-RB', other_designations='CG2022-PB', map_location='82E4-82E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4986500, end_position_on_the_genomic_accession=5011031, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40616, CurrentID=0, Status='live', Symbol='corto', Aliases='Dmel_CG2530, 7128, CG2530, CP-1, Ccf, Corto, Dmel\\CG2530, anon-WO02059370.59, ccf, l(3)07128, l(3)neo31, l(3)neo32', description='CG2530 gene product from transcript CG2530-RC', other_designations='CG2530-PA|CG2530-PB|CG2530-PC|CG2530-PD|CG2530-PE|centrosomal and chromosomal factor|corto-PA|corto-PB|corto-PC|corto-PD|corto-PE', map_location='3-47.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5077779, end_position_on_the_genomic_accession=5087072, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40617, CurrentID=0, Status='live', Symbol='CG12007', Aliases='Dmel_ Dmel\\CG12007, PTAR3, RabGGTalpha', description='CG12007 gene product from transcript CG12007-RA', other_designations='CG12007-PA|uncharacterized protein', map_location='82F1-82F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5128087, end_position_on_the_genomic_accession=5129939, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40618, CurrentID=0, Status='live', Symbol='CG12590', Aliases='Dmel_ Dmel\\CG12590', description='CG12590 gene product from transcript CG12590-RA', other_designations='CG12590-PA|uncharacterized protein', map_location='82F1-82F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5159612, end_position_on_the_genomic_accession=5160901, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40619, CurrentID=0, Status='live', Symbol='dpr16', Aliases='Dmel_CG12591, CG12591, CT34444, Dmel\\CG12591, Dpr-16, Dpr16, mdcds_51595', description='defective proboscis extension response 16', other_designations='CG12591-PB|CG12591-PC|dpr16-PB|dpr16-PC', map_location='82F1-82F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5150908, end_position_on_the_genomic_accession=5170121, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40620, CurrentID=0, Status='live', Symbol='cno', Aliases='Dmel_CG42312, AF-6, CG2534, CG31537, CG42312, Canoe, Cno, Dmel\\CG42312, Dmel_CG2534, Dmel_CG31537, EC3-10, anon-WO0172774.47?, dlhA, lip, mis, cno', description='canoe', other_designations='CG42312-PC|CG42312-PD|CG42312-PE|CG42312-PF|CG42312-PG|CG42312-PH|CG42312-PI|CG42312-PJ|afadin|cno-PC|cno-PD|cno-PE|cno-PF|cno-PG|cno-PH|cno-PI|cno-PJ|dorsal holes A|lips|misty', map_location='3-49 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5170004, end_position_on_the_genomic_accession=5217456, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  40621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40621, CurrentID=0, Status='live', Symbol='Prosbeta2R2', Aliases='Dmel_CG12161, CG12161, Dmel\\CG12161, beta2R2', description='Proteasome beta2 subunit-related 2', other_designations='CG12161-PA|Prosbeta2R2-PA|proteasome beta2 subunit-related 2|proteasome beta2R2 subunit', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5217751, end_position_on_the_genomic_accession=5219528, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40622, CurrentID=0, Status='live', Symbol='CG1116', Aliases='Dmel_ Dmel\\CG1116, Rgp1, anon-WO0118547.337', description='CG1116 gene product from transcript CG1116-RB', other_designations='CG1116-PA|CG1116-PB|CG1116-PD', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5219657, end_position_on_the_genomic_accession=5221548, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40623, CurrentID=0, Status='live', Symbol='CG2604', Aliases='Dmel_ CT40966, Dmel\\CG2604', description='CG2604 gene product from transcript CG2604-RC', other_designations='CG2604-PA|CG2604-PB|CG2604-PC|saccheropin dehydrogenase', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5221632, end_position_on_the_genomic_accession=5223365, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40624, CurrentID=0, Status='live', Symbol='Vps37B', Aliases='Dmel_CG1115, CG1115, DmVPS37, Dmel\\CG1115, Vps37, vps37B', description='Vacuolar protein sorting 37B', other_designations='CG1115-PA|CG1115-PB|CG1115-PC|Vps37B-PA|Vps37B-PB|Vps37B-PC|vacuolar protein sorting 37B', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5226158, end_position_on_the_genomic_accession=5227526, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40626, CurrentID=0, Status='live', Symbol='Mms19', Aliases='Dmel_CG12005, CG12005, Dmel\\CG12005, MMS19, NEST:bs07f08, dMMS19, mms19', description='CG12005 gene product from transcript CG12005-RB', other_designations='CG12005-PB|Mms19-PB', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5232543, end_position_on_the_genomic_accession=5236288, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40627, CurrentID=0, Status='live', Symbol='rtp', Aliases='Dmel_CG10233, 9C4, CG10233, Dmel\\CG10233, anon-83ABa, retin, uta', description='retinophilin', other_designations='CG10233-PA|rtp-PA|undertaker', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5236092, end_position_on_the_genomic_accession=5237315, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40628, CurrentID=0, Status='live', Symbol='CG12163', Aliases='Dmel_ Dmel\\CG12163, anon-WO0118547.305', description='CG12163 gene product from transcript CG12163-RB', other_designations='CG12163-PA|CG12163-PB|CG12163-PC|uncharacterized protein', map_location='82F6-82F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5237651, end_position_on_the_genomic_accession=5251745, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40629, CurrentID=0, Status='live', Symbol='CG1113', Aliases='Dmel_ Dmel\\CG1113', description='CG1113 gene product from transcript CG1113-RB', other_designations='CG1113-PA|CG1113-PB', map_location='82F7-82F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5253153, end_position_on_the_genomic_accession=5255403, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40630, CurrentID=0, Status='live', Symbol='CG12173', Aliases='Dmel_ 24644163, Dmel\\CG12173', description='CG12173 gene product from transcript CG12173-RA', other_designations='CG12173-PA|uncharacterized protein', map_location='82F7-82F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5255444, end_position_on_the_genomic_accession=5256681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40632, CurrentID=0, Status='live', Symbol='Tim17a2', Aliases='Dmel_CG14666, BcDNA:RH03914, CG14666, Dmel\\CG14666', description='CG14666 gene product from transcript CG14666-RA', other_designations='CG14666-PA|Tim17a2-PA', map_location='82F7-82F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5258336, end_position_on_the_genomic_accession=5262692, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40633, CurrentID=0, Status='live', Symbol='Hph', Aliases='Dmel_CG44015, 0303/04, 31543, CG1114, CG14665, CG2676, CG31543, CG44015, CR31541, Dm.egl-9, Dmel\\CG44015, Dmel_CG31543, Dmel_CR31541, PHD, anon-EST:Posey69, dHPH, dPHD, dPHD2, dmHPH, fga, hph, l(3)02255, l(3)82Fe, l(3)S030304', description='HIF prolyl hydroxylase', other_designations='CG44015-PA|CG44015-PB|CG44015-PC|CG44015-PD|HIF prolyl 4-hydroxylase|HIF-1a prolyl hyroxylase|Hph-PA|Hph-PB|Hph-PC|Hph-PD|fatiga|lethal (3) 82Fe|lethal (3) S030304|lopped|prolyl-4-hydroxylase', map_location='82F7-82F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5257037, end_position_on_the_genomic_accession=5269507, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40635, CurrentID=0, Status='live', Symbol='Tim17b1', Aliases='Dmel_CG1158, BcDNA:AT16284, CG1158, Dmel\\CG1158', description='CG1158 gene product from transcript CG1158-RA', other_designations='CG1158-PA|Tim17b1-PA', map_location='82F8-82F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5300145, end_position_on_the_genomic_accession=5301186, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40637, CurrentID=0, Status='live', Symbol='CG11999', Aliases='Dmel_ BEST:LD29996, CG-11999, Dmel\\CG11999', description='CG11999 gene product from transcript CG11999-RA', other_designations='CG11999-PA|uncharacterized protein', map_location='82F10-82F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5353717, end_position_on_the_genomic_accession=5354747, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40638, CurrentID=0, Status='live', Symbol='CG1161', Aliases='Dmel_ Dmel\\CG1161', description='CG1161 gene product from transcript CG1161-RC', other_designations='CG1161-PA|CG1161-PC', map_location='82F10-82F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5355092, end_position_on_the_genomic_accession=5356748, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40639, CurrentID=0, Status='live', Symbol='Prosbeta7', Aliases='Dmel_CG12000, 12000, 154275_at, Beta-7, CG12000, Dmel\\CG12000, GC12000, Prosbeta4, Prot. beta4, anon-WO0118547.326, beta-4, beta-7, beta7', description='Proteasome beta7 subunit', other_designations='CG12000-PA|CG12000-PB|Prosbeta4|Prosbeta7-PA|Prosbeta7-PB|beta-type 4 subunit|beta7|proteasome beta4 subunit|proteasome beta7 subunit', map_location='82F11-82F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5357012, end_position_on_the_genomic_accession=5358364, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40640, CurrentID=0, Status='live', Symbol='Cerk', Aliases='Dmel_CG16708, CG16708, CerK, DCERK, Dcerk, Dmel\\CG16708, dcerk', description='Ceramide kinase', other_designations='CG16708-PA|CG16708-PB|Cerk-PA|Cerk-PB|ceramide kinase', map_location='82F11-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5358873, end_position_on_the_genomic_accession=5368306, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40642, CurrentID=0, Status='live', Symbol='hd', Aliases='Dmel_CG2669, CG2669, Dmel\\CG2669, fs(3)272, fs(3)272-9, l(3)82Ff, l(3)82Fk', description='humpty dumpty', other_designations='CG2669-PA|CG2669-PB|female sterile(3)272-9|hd-PA|hd-PB|lethal (3) 82Ff|lethal (3) 82Fk', map_location='3-47.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5372592, end_position_on_the_genomic_accession=5374962, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40643, CurrentID=0, Status='live', Symbol='CG14667', Aliases='Dmel_ Dmel\\CG14667', description='CG14667 gene product from transcript CG14667-RB', other_designations='CG14667-PB|uncharacterized protein', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5375022, end_position_on_the_genomic_accession=5376460, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40644, CurrentID=0, Status='live', Symbol='7B2', Aliases='Dmel_CG1168, CG1168, Dmel\\CG1168, FBgn0040073, d', description='CG1168 gene product from transcript CG1168-RB', other_designations='7B2-PA|7B2-PB|CG1168-PA|CG1168-PB', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5376780, end_position_on_the_genomic_accession=5377975, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40646, CurrentID=0, Status='live', Symbol='CG14668', Aliases='Dmel_ Dmel\\CG14668', description='CG14668 gene product from transcript CG14668-RA', other_designations='CG14668-PA|uncharacterized protein', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5386717, end_position_on_the_genomic_accession=5387664, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40647, CurrentID=0, Status='live', Symbol='CG1172', Aliases='Dmel_ CC1172, CR43473, Dmel\\CG1172, Dmel_CR43473', description='CG1172 gene product from transcript CG1172-RC', other_designations='CG1172-PA|CG1172-PB|CG1172-PC', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5393922, end_position_on_the_genomic_accession=5399510, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40648, CurrentID=0, Status='live', Symbol='Or83a', Aliases='Dmel_CG10612, 83A.1, 83a, CG10612, DOR44, Dmel\\CG10612', description='Odorant receptor 83a', other_designations='CG10612-PA|Or83a-PA|odorant receptor 83a', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5397643, end_position_on_the_genomic_accession=5399196, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40649, CurrentID=0, Status='live', Symbol='CG2663', Aliases='Dmel_ Dmel\\CG2663', description='CG2663 gene product from transcript CG2663-RA', other_designations='CG2663-PA|CG2663-PC|alpha tocopherol transfer protein', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5400594, end_position_on_the_genomic_accession=5404135, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40650, CurrentID=0, Status='live', Symbol='Orco', Aliases='Dmel_CG10609, 83A.2, 83b, A45, CG10609, DOR45, DOR83b, DmORCO, DmOrco, DmelOrco, Dmel\\CG10609, OR49, OR83B, OR83b, ORCO, Or83b, or83b, orco, vainsA', description='Odorant receptor co-receptor', other_designations='CG10609-PA|CG10609-PB|Or83b|Orco-PA|Orco-PB|odorant receptor 83b|odorant receptor co-receptor|odorant receptor coreceptor|olfactory co-receptor|olfactory coreceptor|olfactory receptor co-receptor', map_location='83A2-83A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5407183, end_position_on_the_genomic_accession=5412406, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40651, CurrentID=0, Status='live', Symbol='CG12147', Aliases='Dmel_ CKIalpha like, Dmel\\CG12147', description='CG12147 gene product from transcript CG12147-RA', other_designations='CG12147-PA|CKIa-like|uncharacterized protein', map_location='83A2-83A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5425369, end_position_on_the_genomic_accession=5427257, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40652, CurrentID=0, Status='live', Symbol='CG14669', Aliases='Dmel_ Dmel\\CG14669, Q9VNB7', description='CG14669 gene product from transcript CG14669-RB', other_designations='CG14669-PA|CG14669-PB|CG14669-PC|CG14669-PD|uncharacterized protein', map_location='83A2-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5414527, end_position_on_the_genomic_accession=5458394, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40653, CurrentID=0, Status='live', Symbol='PEK', Aliases='Dmel_CG2087, CG2087, DPERK, DmPEK, Dmel\\CG2087, DpERK, EIF2-like, PERK, Pek, Perk, dPERK, pERK, pek, perk', description='pancreatic eIF-2alpha kinase', other_designations='CG2087-PA|CG2087-PB|CG2087-PC|PEK-PA|PEK-PB|PEK-PC|PERK|pancreatic eIF2alpha kinase', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5458563, end_position_on_the_genomic_accession=5464753, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40654, CurrentID=0, Status='live', Symbol='RpL35A', Aliases='Dmel_CG2099, CG2099, Dmel\\CG2099, L35a, Rp L35A, Rpl35A, anon-EST:Posey39', description='Ribosomal protein L35A', other_designations='CG2099-PA|RpL35A-PA|ribosomal protein L35A', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5465630, end_position_on_the_genomic_accession=5467101, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40655, CurrentID=0, Status='live', Symbol='POLDIP2', Aliases='Dmel_CG12162, CG12162, Dmel\\CG12162', description='Polymerase (DNA-directed), delta interacting protein 2', other_designations='CG12162-PA|CG12162-PB|POLDIP2-PA|POLDIP2-PB|polymerase (DNA-directed), delta interacting protein 2', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5467778, end_position_on_the_genomic_accession=5469566, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40656, CurrentID=0, Status='live', Symbol='mRpL44', Aliases='Dmel_CG2109, CG2109, Dmel\\CG2109, MRP-L44', description='mitochondrial ribosomal protein L44', other_designations='CG2109-PA|mRpL44-PA', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5469463, end_position_on_the_genomic_accession=5470636, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40657, CurrentID=0, Status='live', Symbol='Snr1', Aliases='Dmel_CG1064, BAP45, CG1064, Dmel\\CG1064, SNR1, Snr/BAP45, l(3)01319, snr, snr-1, snr1, Snr1', description='Snf5-related 1', other_designations='CG1064-PA|Snr1-PA', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5470926, end_position_on_the_genomic_accession=5472754, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40659, CurrentID=0, Status='live', Symbol='Hcs', Aliases='Dmel_CG14670, CG14670, Dmel\\CG14670, HCS, HSC, NEST:bs27h05', description='Holocarboxylase synthetase', other_designations='CG14670-PA|CG14670-PB|CG14670-PC|CG14670-PD|Hcs-PA|Hcs-PB|Hcs-PC|Hcs-PD|holocarboxylase synthetase', map_location='83A5-83A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5474899, end_position_on_the_genomic_accession=5485393, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  40660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40660, CurrentID=0, Status='live', Symbol='ksr', Aliases='Dmel_CG2899, CG2899, Dmel\\CG2899, EC3-1, EK3-1, KSR, KSR-1, Ksr, Ksr-1, SR3-1, dKsr, l(3)j5E2', description='kinase suppressor of ras', other_designations='CG2899-PA|CG2899-PB|CG2899-PC|CG2899-PD|kinase suppressor of Ras|kinase suppressor of Ras85D|ksr-PA|ksr-PB|ksr-PC|ksr-PD|suppressor of Ras1 3-1', map_location='83A5-83A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5478390, end_position_on_the_genomic_accession=5483906, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40661, CurrentID=0, Status='live', Symbol='CG31550', Aliases='Dmel_ CG12167, CG12168, Dmel\\CG31550', description='CG31550 gene product from transcript CG31550-RE', other_designations='CG31550-PB|CG31550-PC|CG31550-PD|CG31550-PE|uncharacterized protein', map_location='83A5-83A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5485543, end_position_on_the_genomic_accession=5489489, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40663, CurrentID=0, Status='live', Symbol='CG31547', Aliases='Dmel_ CG10544, CG2509, Dmel\\CG31547', description='CG31547 gene product from transcript CG31547-RB', other_designations='CG31547-PA|CG31547-PB|uncharacterized protein', map_location='83A5-83A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5489505, end_position_on_the_genomic_accession=5502702, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  40664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40664, CurrentID=0, Status='live', Symbol='Itp-r83A', Aliases='Dmel_CG1063, CG1063, DIP, DmInsP[[3R]], DmInsP[[3]]R, Dmel\\CG1063, IP, IP3R, IP[[3]] receptor, IP[[3]]R, ITP-r83A, ITPR1, Ins3PR, InsP[[3]], InsP[[3]]R, Insp3R, Ip3R, Itp-r, Itp-r-83A, Itp-r42B, Itp-r[1], Itpr, anon-WO02059370.49, dIP[[3]]R, dip, itpr, itpr-83, l(3)05616, l(3)itpr, l(3)j5B4', description='Inositol 1,4,5,-tris-phosphate receptor', other_designations='1, 4, 5-inositol trisphosphate receptor|CG1063-PA|CG1063-PB|CG1063-PC|IP3 receptor|IP[[3]] receptor|Inositol(1,4,5)-triphosphate receptor 42B|Ins(1,4,5)P[[3]] receptor|InsP[[3]] receptor|Itp-r83A-PA|Itp-r83A-PB|Itp-r83A-PC|dm_IP3 receptor|inositol 1,4,5,-tris-phosphate receptor|inositol 1,4,5-triphosphate receptor|inositol 1,4,5-trisphosphate (InsP3) receptor|inositol 1,4,5-trisphosphate receptor|inositol triphosphate receptor 1|inositol trisphosphate receptor|inositol-1,3,5,-trisphosphate receptor|inositol-1,4,5-trisphosphate receptor|ip3-receptor|type 1 InsP{3} receptor', map_location='83A7-83B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5517117, end_position_on_the_genomic_accession=5539411, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  40665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40665, CurrentID=0, Status='live', Symbol='Nmdar1', Aliases='Dmel_CG2902, CG2902, DMNMDAR1, DNMDAR-I, DNmdar, DmelNmdar1, Dmel\\CG2902, DrNR1, NMDA, NMDA R1, NMDAR1, NMdar1, NR1, NRI, Nmdar, dNR1, dnmdar-I, dsNR1, nmr', description='NMDA receptor 1', other_designations='CG2902-PA|NMDA receptor|NMDA receptor I|NMDAR|Nmdar1-PA|dNR1', map_location='83A6-83A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5508842, end_position_on_the_genomic_accession=5523328, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  40666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40666, CurrentID=0, Status='live', Symbol='CG2519', Aliases='Dmel_ Dmel\\CG2519', description='CG2519 gene product from transcript CG2519-RB', other_designations='CG2519-PA|CG2519-PB|CG2519-PC', map_location='83B1-83B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5539510, end_position_on_the_genomic_accession=5547775, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40667, CurrentID=0, Status='live', Symbol='mia', Aliases='Dmel_CG10390, CG10390, Dmel\\CG10390, Mia, TAF, TAF60-2, TAF6L, TAF[[II]], TAF[[II]]60-2, TFIID, Taf60-2, Taf6L, dTAF6L, dTAF[[II]]60, dTAF[[II]]60-2, dmTAF6b, ms(3)570', description='meiosis I arrest', other_designations='CG10390-PA|CG10390-PB|CG10390-PC|TAF6-like|TBP-associated factor 6-like|TBP-associated factor 60kD-2|mia-PA|mia-PB|mia-PC', map_location='83B1-83B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5548115, end_position_on_the_genomic_accession=5551168, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40668, CurrentID=0, Status='live', Symbol='Pcmt', Aliases='Dmel_CG2152, CG2152, Dmel\\CG2152, PIMT, Pimt, dPIMT, pcmt', description='Protein-L-isoaspartate (D-aspartate) O-methyltransferase', other_designations='CG2152-PA|CG2152-PC|Pcmt-PA|Pcmt-PC|protein D-aspartyl, L-isoaspartyl O-methyltransferase|protein isoaspartyl methyltransferase', map_location='101F-83B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5552823, end_position_on_the_genomic_accession=5554316, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40669, CurrentID=0, Status='live', Symbol='Snm1', Aliases='Dmel_CG10018, CG10018, Dmel\\CG10018, SNM1, mus322', description='CG10018 gene product from transcript CG10018-RA', other_designations='CG10018-PA|Snm1-PA|mutagen-sensitive 322', map_location='83B1-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5555084, end_position_on_the_genomic_accession=5557786, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40670, CurrentID=0, Status='live', Symbol='Pi4KIIalpha', Aliases='Dmel_CG2929, CG2929, Dmel\\CG2929, PI4KII, PI4KIIalpha, PI4KIIbeta, Pi4KIIa', description='CG2929 gene product from transcript CG2929-RC', other_designations='CG2929-PA|CG2929-PB|CG2929-PC|Pi4KIIalpha-PA|Pi4KIIalpha-PB|Pi4KIIalpha-PC', map_location='83B1-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5557663, end_position_on_the_genomic_accession=5563652, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40671, CurrentID=0, Status='live', Symbol='CG14671', Aliases='Dmel_ Dmel\\CG14671', description='CG14671 gene product from transcript CG14671-RB', other_designations='CG14671-PA|CG14671-PB|uncharacterized protein', map_location='83B2-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5563977, end_position_on_the_genomic_accession=5564815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40672, CurrentID=0, Status='live', Symbol='CG12746', Aliases='Dmel_ Dmel\\CG12746, WUN-like', description='CG12746 gene product from transcript CG12746-RD', other_designations='CG12746-PB|CG12746-PD|CG12746-PE|PPAPDC1|uncharacterized protein', map_location='83B2-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5564865, end_position_on_the_genomic_accession=5567435, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40673, CurrentID=0, Status='live', Symbol='CG2931', Aliases='Dmel_ Dmel\\CG2931, cg2931', description='CG2931 gene product from transcript CG2931-RA', other_designations='CG2931-PA', map_location='83B2-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5567502, end_position_on_the_genomic_accession=5568615, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40675, CurrentID=0, Status='live', Symbol='CRMP', Aliases='Dmel_CG1411, BcDNA:HL02693, CG1411, DHP, Dmel\\CG1411, EP(3)3238, EP3238, Pyd2, crmp, l(3)83Ba, pyd2', description='Collapsin Response Mediator Protein', other_designations='CG1411-PA|CG1411-PB|CG1411-PE|CRMP-PA|CRMP-PB|CRMP-PE|collapsin response mediating protein|collapsin response mediator protein|dihydropyrimidinase|dihydropyrimidine amidohydrolase', map_location='83B2-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5571108, end_position_on_the_genomic_accession=5579189, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  40676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40676, CurrentID=0, Status='live', Symbol='CG2926', Aliases='Dmel_ Dmel\\CG2926', description='CG2926 gene product from transcript CG2926-RA', other_designations='CG2926-PA|CG2926-PB', map_location='83B2-83B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5579301, end_position_on_the_genomic_accession=5588380, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40677, CurrentID=0, Status='live', Symbol='rev7', Aliases='Dmel_CG2948, CG2948, DmREV7, Dmel\\CG2948, Rev7, mad2B', description='CG2948 gene product from transcript CG2948-RA', other_designations='CG2948-PA|rev7-PA', map_location='83B4-83B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5588873, end_position_on_the_genomic_accession=5589703, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40678, CurrentID=0, Status='live', Symbol='noi', Aliases='Dmel_CG2925, CG2925, Dmel\\CG2925, Noi, SF3A3, SF3a[60], U2, l(3)j3E7, ms3225', description='noisette', other_designations='CG2925-PA|noi-PA', map_location='83B4-83B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5589601, end_position_on_the_genomic_accession=5591459, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40679, CurrentID=0, Status='live', Symbol='Vha26', Aliases='Dmel_CG1088, ATP6V1E1, BcDNA.GH03683, BcDNA:GH03683, CG1088, Dmel\\CG1088, VhaE, anon-83Bb, chr3R:1419698..1419836, vATPase, vha26', description='Vacuolar H[+]-ATPase 26kD subunit', other_designations='CG1088-PA|CG1088-PB|CG1088-PC|E subunit|Vha26-PA|Vha26-PB|Vha26-PC|vacuolar ATP synthase subunit E|vacuolar H+ATPase subunit E|vacuolar H[+]-ATPase 26 kD E subunit|vacuolar H[+]-ATPase 26kD E subunit|vacuolar H[+]-ATPase 26kD subunit', map_location='83B4-83B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5591763, end_position_on_the_genomic_accession=5595356, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40680, CurrentID=0, Status='live', Symbol='kra', Aliases='Dmel_CG2922, CG2922, CT7240, Decp, Dmel\\CG2922, ECP, Exba, Kra, anon-WO0172774.49, chr3R:1423015..1423173, eIF, eIF-5C, eIF-5c, eIF5, eIF5c, eIf-5c, ecp, eiF-5C, elF-5C, elF5C, elf-5C, exba, l(3)03022, l(3)j9B6', description='krasavietz', other_designations='CG2922-PA|CG2922-PB|CG2922-PC|CG2922-PD|CG2922-PE|CG2922-PF|CG2922-PG|drosophila eIF5C-containing protein|eukaryotic initiation factor 5C|extra bases|extra bses|extra-bases|kra-PA|kra-PB|kra-PC|kra-PD|kra-PE|kra-PF|kra-PG|lethal (3) j9B6', map_location='83B4-83B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5595364, end_position_on_the_genomic_accession=5601136, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40681, CurrentID=0, Status='live', Symbol='CG1427', Aliases='Dmel_ Dmel\\CG1427, SecS', description='CG1427 gene product from transcript CG1427-RC', other_designations='CG1427-PC|CG1427-PD|CG1427-PE', map_location='83B4-83B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5601433, end_position_on_the_genomic_accession=5603185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40682, CurrentID=0, Status='live', Symbol='asl', Aliases='Dmel_CG2919, ASL, Asl, CG2919, Dmel\\CG2919, mecD, sdu', description='asterless', other_designations='CG2919-PA|asl-PA|mechanosensory D|mechanosensory mutant D|star dust', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5603236, end_position_on_the_genomic_accession=5606543, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40683, CurrentID=0, Status='live', Symbol='Rga', Aliases='Dmel_CG2161, CG2138, CG2161, Dmel\\CG2161, NOT2, Not2, RGA, l(3)03834, not2', description='Regena', other_designations='CG2161-PA|CG2161-PB|CG2161-PC|Rga-PA|Rga-PB|Rga-PC|regena', map_location='83B5-83B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5607275, end_position_on_the_genomic_accession=5613112, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40684, CurrentID=0, Status='live', Symbol='Atu', Aliases='Dmel_CG1433, CG1433, Dmel\\CG1433, l(3)s1938', description='Another transcription unit', other_designations='Atu-PA|CG1433-PA|another transcription unit', map_location='83B6-83B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5613358, end_position_on_the_genomic_accession=5615939, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40685, CurrentID=0, Status='live', Symbol='Spec2', Aliases='Dmel_CG14672, BcDNA:LD24607, CG14672, Dmel\\CG14672, SPEC, SPEC2, dSPEC, dSPEC2', description='CG14672 gene product from transcript CG14672-RA', other_designations='CG14672-PA|Spec2-PA|small protein effector of Cdc42 2', map_location='83B6-83B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5616644, end_position_on_the_genomic_accession=5624125, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40686, CurrentID=0, Status='live', Symbol='CG2911', Aliases='Dmel_ Dmel\\CG2911', description='CG2911 gene product from transcript CG2911-RB', other_designations='CG2911-PA|CG2911-PB', map_location='83B6-83B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5622119, end_position_on_the_genomic_accession=5624125, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40687, CurrentID=0, Status='live', Symbol='RpL13A', Aliases='Dmel_CG1475, CG1475, Dmel\\CG1475, L13A, L13a2, M(3)83B, Rp L13A, anon-EST:Posey125, anon-EST:fe1A5, bs25f04.y1, cg1475, chr3R:1451210..1451407', description='Ribosomal protein L13A', other_designations='60S ribosomal protein L13A|CG1475-PB|CG1475-PC|RpL13A-PB|RpL13A-PC|anon-fast-evolving-1A5|putative ribosomal L13 protein|ribosomal protein L13A|ribosomal protein L13a', map_location='83B6-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5624474, end_position_on_the_genomic_accession=5625947, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40689, CurrentID=0, Status='live', Symbol='CG31549', Aliases='Dmel_ CG11005, Dmel\\CG31549', description='CG31549 gene product from transcript CG31549-RA', other_designations='CG31549-PA|CG31549-PB|uncharacterized protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5632192, end_position_on_the_genomic_accession=5633321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40690, CurrentID=0, Status='live', Symbol='CG12171', Aliases='Dmel_ Dmel\\CG12171, anon-WO0118547.83', description='CG12171 gene product from transcript CG12171-RA', other_designations='CG12171-PA|uncharacterized protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5633627, end_position_on_the_genomic_accession=5634640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40691, CurrentID=0, Status='live', Symbol='CG31546', Aliases='Dmel_ CG2907, Dmel\\CG31546, anon-WO0140519.189', description='CG31546 gene product from transcript CG31546-RA', other_designations='CG31546-PA|uncharacterized protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5634626, end_position_on_the_genomic_accession=5635624, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40692, CurrentID=0, Status='live', Symbol='CG12170', Aliases='Dmel_ Dmel\\CG12170, anon-WO0118547.367', description='CG12170 gene product from transcript CG12170-RA', other_designations='CG12170-PA|uncharacterized protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5637234, end_position_on_the_genomic_accession=5638657, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40693, CurrentID=0, Status='live', Symbol='MTA1-like', Aliases='Dmel_CG2244, CG2244, Dmel\\CG2244, MTA, MTA-like, MTA1, dMTA, dMTA-1, dMTA1-like, p120', description='Metastasis associated 1-like', other_designations='CG2244-PA|CG2244-PB|CG2244-PC|CG2244-PD|MTA-like|MTA1-like-PA|MTA1-like-PB|MTA1-like-PC|MTA1-like-PD|metastasis associated 1-like|metastasis-associated-1-like-protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5638655, end_position_on_the_genomic_accession=5648738, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40694, CurrentID=0, Status='live', Symbol='Sec23', Aliases='Dmel_CG1250, CG1250, COPII, Dmel\\CG1250, SEC23A, Sec23p, dSec23, dSec23p, dar2, dsec23, dsec23p, hau, sec23, Sec23', description='Sec23 ortholog (S. cerevisiae)', other_designations='CG1250-PA|CG1250-PB|CG1250-PD|CG1250-PE|CG1250-PF|CG1250-PG|CG1250-PH|Sec23 ortholog|Sec23-PA|Sec23-PB|Sec23-PD|Sec23-PE|Sec23-PF|Sec23-PG|Sec23-PH|coat protein II|dendritic arbor reduction 2|haunted|sec23', map_location='3-48.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5649065, end_position_on_the_genomic_accession=5652629, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40695, CurrentID=0, Status='live', Symbol='elm', Aliases='Dmel_CG2185, CG2185, Dmel\\CG2185', description='ethanol sensitive with low memory', other_designations='CG2185-PA|CG2185-PB|elm-PA|elm-PB', map_location='83B8-83B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5652640, end_position_on_the_genomic_accession=5654861, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40696, CurrentID=0, Status='live', Symbol='MED27', Aliases='Dmel_CG1245, CG1245, CRSP34, Dmel\\CG1245, Med27, Trap37, dCRSP34, dCrsp34, dTRAP37, med27', description='Mediator complex subunit 27', other_designations='CG1245-PA|MED27-PA|mediator complex subunit 27', map_location='83B8-83B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5655139, end_position_on_the_genomic_accession=5656336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40697, CurrentID=0, Status='live', Symbol='CG2182', Aliases='Dmel_ Dmel\\CG2182', description='CG2182 gene product from transcript CG2182-RA', other_designations='CG2182-PA', map_location='83B8-83B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5656236, end_position_on_the_genomic_accession=5659912, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40698, CurrentID=0, Status='live', Symbol='CG1109', Aliases='Dmel_ BcDNA:LD24014, Dmel\\CG1109, WDR33', description='CG1109 gene product from transcript CG1109-RA', other_designations='CG1109-PA|CG1109-PB', map_location='83B8-83B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5660635, end_position_on_the_genomic_accession=5665169, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40699, CurrentID=0, Status='live', Symbol='Xe7', Aliases='Dmel_CG2179, CG2179, Dmel\\CG2179, l(3)02248', description='CG2179 gene product from transcript CG2179-RD', other_designations='CG2179-PA|CG2179-PD|CG2179-PE|Xe7-PA|Xe7-PD|Xe7-PE', map_location='83B8-83B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5665139, end_position_on_the_genomic_accession=5669334, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40700, CurrentID=0, Status='live', Symbol='Atg17', Aliases='Dmel_CG1347, CG1347, Dmel\\CG1347, FIP200, Fip200, RB1CC1, fip200', description='Autophagy-related 17', other_designations='Atg17-PB|Atg17-PC|Autophagy-specific gene 17|CG1347-PB|CG1347-PC', map_location='83B8-83B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5669713, end_position_on_the_genomic_accession=5679930, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40701, CurrentID=0, Status='live', Symbol='Rab23', Aliases='Dmel_CG2108, CG2108, Dm Dmel\\CG2108, Q9VNG6, rab23', description='CG2108 gene product from transcript CG2108-RA', other_designations='CG2108-PA|Rab23-PA', map_location='83B9-83B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5680054, end_position_on_the_genomic_accession=5685434, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40702, CurrentID=0, Status='live', Symbol='CG2104', Aliases='Dmel_ Dmel\\CG2104', description='CG2104 gene product from transcript CG2104-RB', other_designations='CG2104-PA|CG2104-PB', map_location='83C1-83C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5690891, end_position_on_the_genomic_accession=5693286, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40704, CurrentID=0, Status='live', Symbol='plx', Aliases='Dmel_CG42612, CG1093, CG14674, CG34122, CG42612, Dmel\\CG42612, Dmel_CG1093, Dmel_CG14674, Dmel_CG34122', description='pollux', other_designations='CG42612-PA|CG42612-PC|CG42612-PD|CG42612-PE|CG42612-PF|plx-PA|plx-PC|plx-PD|plx-PE|plx-PF', map_location='83B9-83C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5685855, end_position_on_the_genomic_accession=5712837, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  40706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40706, CurrentID=0, Status='live', Symbol='CG1239', Aliases='Dmel_ Dmel\\CG1239, cg1239', description='CG1239 gene product from transcript CG1239-RB', other_designations='CG1239-PA|CG1239-PB|CG1239-PC', map_location='83C1-83C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5733513, end_position_on_the_genomic_accession=5734702, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40707, CurrentID=0, Status='live', Symbol='CG2100', Aliases='Dmel_ Dmel\\CG2100', description='CG2100 gene product from transcript CG2100-RB', other_designations='CG2100-PA|CG2100-PB', map_location='83C1-83C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5734698, end_position_on_the_genomic_accession=5736355, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40708, CurrentID=0, Status='live', Symbol='CG1236', Aliases='Dmel_ Dmel\\CG1236', description='CG1236 gene product from transcript CG1236-RA', other_designations='CG1236-PA', map_location='83C1-83C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5736748, end_position_on_the_genomic_accession=5738013, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40709, CurrentID=0, Status='live', Symbol='Sym', Aliases='Dmel_CG2097, CG2097, Dmel\\CG2097p, dSymp, Sym', description='Symplekin', other_designations='CG2097-PA|Sym-PA|symplekin', map_location='83C1-83C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5738178, end_position_on_the_genomic_accession=5742476, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40710, CurrentID=0, Status='live', Symbol='Madm', Aliases='Dmel_CG1098, BcDNA:LD28657, CG1098, Dmel\\CG1098, MADM, anon-WO0118547.406, dMADM, lincRNA.S6031', description='MLF1-adaptor molecule', other_designations='CG1098-PA|CG1098-PB|Madm-PA|Madm-PB|Mif1 adaptor molecule', map_location='83C2-83C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5742510, end_position_on_the_genomic_accession=5747138, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40711, CurrentID=0, Status='live', Symbol='CG2091', Aliases='Dmel_ Dmel\\CG2091', description='CG2091 gene product from transcript CG2091-RA', other_designations='CG2091-PA', map_location='83C2-83C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5753318, end_position_on_the_genomic_accession=5754774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40712, CurrentID=0, Status='live', Symbol='Sec8', Aliases='Dmel_CG2095, CG2095, Dmel\\CG2095p, dsec8, fun, sec8, Sec8', description='Sec8 ortholog (S. cerevisiae)', other_designations='CG2095-PA|Sec8 ortholog|Sec8-PA|funnel cakes|sec8', map_location='3-48.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5754932, end_position_on_the_genomic_accession=5758997, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40713, CurrentID=0, Status='live', Symbol='CG2082', Aliases='Dmel_ BcDNA:GH02439, Dmel\\CG2082', description='CG2082 gene product from transcript CG2082-RO', other_designations='CG2082-PA|CG2082-PB|CG2082-PC|CG2082-PD|CG2082-PG|CG2082-PH|CG2082-PJ|CG2082-PK|CG2082-PL|CG2082-PM|CG2082-PN|CG2082-PO', map_location='83C3-83C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5758680, end_position_on_the_genomic_accession=5781076, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  40714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40714, CurrentID=0, Status='live', Symbol='jagn', Aliases='Dmel_CG10978, CG10978, Dmel\\CG10978', description='jagunal', other_designations='CG10978-PA|CG10978-PB|jagn-PA|jagn-PB', map_location='83C3-83C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5779315, end_position_on_the_genomic_accession=5781308, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40715, CurrentID=0, Status='live', Symbol='kat-60L1', Aliases='Dmel_CG1193, CG1193, DmCG1193, Dmel\\CG1193', description='katanin p60-like 1', other_designations='CG1193-PA|CG1193-PB|CG1193-PC|CG1193-PD|CG1193-PE|CG1193-PF|CG1193-PG|CG1193-PH|Katanin-p60L1|kat-60L1-PA|kat-60L1-PB|kat-60L1-PC|kat-60L1-PD|kat-60L1-PE|kat-60L1-PF|kat-60L1-PG|kat-60L1-PH', map_location='83C3-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5781762, end_position_on_the_genomic_accession=5787465, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40716, CurrentID=0, Status='live', Symbol='CG2051', Aliases='Dmel_ Dmel\\CG2051, dmHAG409', description='CG2051 gene product from transcript CG2051-RB', other_designations='CG2051-PA|CG2051-PB', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5787337, end_position_on_the_genomic_accession=5789384, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40717, CurrentID=0, Status='live', Symbol='Rpn5', Aliases='Dmel_CG1100, CG1100, Dmel\\CG1100, p55, rpn5', description='Regulatory particle non-ATPase 5', other_designations='CG1100-PA|Rpn5-PA|regulatory particle non-ATPase 5', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5789493, end_position_on_the_genomic_accession=5791569, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40718, CurrentID=0, Status='live', Symbol='CG1218', Aliases='Dmel_ Dmel\\CG1218', description='CG1218 gene product from transcript CG1218-RA', other_designations='CG1218-PA', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5791879, end_position_on_the_genomic_accession=5793383, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40719, CurrentID=0, Status='live', Symbol='CG2046', Aliases='Dmel_ Dmel\\CG2046', description='CG2046 gene product from transcript CG2046-RA', other_designations='CG2046-PA', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5793308, end_position_on_the_genomic_accession=5794317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40720, CurrentID=0, Status='live', Symbol='CG10979', Aliases='Dmel_ Dmel\\CG10979', description='CG10979 gene product from transcript CG10979-RB', other_designations='CG10979-PA|CG10979-PB|uncharacterized protein', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5794705, end_position_on_the_genomic_accession=5799994, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40721, CurrentID=0, Status='live', Symbol='CG11000', Aliases='Dmel_ BcDNA:HL04389, CG31555, Dmel\\CG11000', description='CG11000 gene product from transcript CG11000-RC', other_designations='CG11000-PB|CG11000-PC|uncharacterized protein', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5802328, end_position_on_the_genomic_accession=5813258, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740700, CurrentID=0, Status='live', Symbol='natalisin', Aliases='Dmel_CG34388, CG14842, CG14843, CG34388, Dmel\\CG34388, Dmel_CG14842, Dmel_CG14843, NTL, ntl', description='CG34388 gene product from transcript CG34388-RC', other_designations='CG34388-PC|CG34388-PD|natalisin-PC|natalisin-PD', map_location='88C1-88C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14418593, end_position_on_the_genomic_accession=14421865, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40723, CurrentID=0, Status='live', Symbol='Hpr1', Aliases='Dmel_CG2031, CG2031, Dmel\\CG2031, HPR1', description='CG2031 gene product from transcript CG2031-RA', other_designations='CG2031-PA|Hpr1-PA', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5818794, end_position_on_the_genomic_accession=5821440, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40724, CurrentID=0, Status='live', Symbol='CG2023', Aliases='Dmel_ Cg2023, Dmel\\CG2023, Sec20, lincRNA.683', description='CG2023 gene product from transcript CG2023-RA', other_designations='CG2023-PA|CG2023-PB', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5822026, end_position_on_the_genomic_accession=5824532, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40725, CurrentID=0, Status='live', Symbol='dgrn', Aliases='Dmel_CG10981, CG10981, Dmel\\CG10981', description='degringolade', other_designations='CG10981-PA|CG10981-PB|CG10981-PD|dgrn-PA|dgrn-PB|dgrn-PD', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5824615, end_position_on_the_genomic_accession=5826576, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40726, CurrentID=0, Status='live', Symbol='glob3', Aliases='Dmel_CG14675, CG14675, DmeGb3, Dmel\\CG14675', description='globin 3', other_designations='CG14675-PA|glob3-PA', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5830062, end_position_on_the_genomic_accession=5831785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40727, CurrentID=0, Status='live', Symbol='CG1208', Aliases='Dmel_ Dmel\\CG1208', description='CG1208 gene product from transcript CG1208-RB', other_designations='CG1208-PB|CG1208-PC', map_location='83C4-83C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5831639, end_position_on_the_genomic_accession=5838782, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40728, CurrentID=0, Status='live', Symbol='CG1213', Aliases='Dmel_ CT30783, Dmel\\CG1213', description='CG1213 gene product from transcript CG1213-RB', other_designations='CG1213-PA|CG1213-PB|CG1213-PC|CG1213-PD', map_location='83C5-83C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5841170, end_position_on_the_genomic_accession=5846568, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40729, CurrentID=0, Status='live', Symbol='CG14676', Aliases='Dmel_ Dmel\\CG14676', description='CG14676 gene product from transcript CG14676-RA', other_designations='CG14676-PA|uncharacterized protein', map_location='83C5-83C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5847220, end_position_on_the_genomic_accession=5848177, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40732, CurrentID=0, Status='live', Symbol='ECSIT', Aliases='Dmel_CG10610, CG10610, Dm Dmel\\CG10610, dECSIT', description='CG10610 gene product from transcript CG10610-RA', other_designations='CG10610-PA|ECSIT-PA', map_location='83C5-83C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5856585, end_position_on_the_genomic_accession=5857914, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40733, CurrentID=0, Status='live', Symbol='CG2017', Aliases='Dmel_ CG 2017, Dmel\\CG2017', description='CG2017 gene product from transcript CG2017-RD', other_designations='CG2017-PA|CG2017-PB|CG2017-PC|CG2017-PD|CG2017-PE|GP-1 related', map_location='83C5-83C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5858083, end_position_on_the_genomic_accession=5862405, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40737, CurrentID=0, Status='live', Symbol='Obp83a', Aliases='Dmel_CG11421, 83a, CG11421, DmelObp83a, Dmel\\CG11421, OS-F, OS-F/PBP3, Obp33a, Os-F, PB-PRP3, PBP-H2, PBPRP-3, PBPRP3, Pbprp3, obp83a, pbprp-3, pbprp3, pbprp3(OS-F)', description='Odorant-binding protein 83a', other_designations='CG11421-PA|CG11421-PB|CG11421-PC|Obp83a-PA|Obp83a-PB|Obp83a-PC|Olfactory-specific-F|Pbprp3|Pheromone-binding protein homolog 2|Pheromone-binding protein-related protein 3', map_location='83D1-83D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5972155, end_position_on_the_genomic_accession=5975245, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40738, CurrentID=0, Status='live', Symbol='Obp83b', Aliases='Dmel_CG11422, 83b, CG11422, DmelObp83b, Dmel\\CG11422, OS-E, Os-E, PBP-H1, obp83b', description='Odorant-binding protein 83b', other_designations='CG11422-PA|Obp83b-PA|Olfactory-specific E|Os-E|Pheromone-binding protein homolog 1', map_location='83D1-83D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5976177, end_position_on_the_genomic_accession=5976817, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40739, CurrentID=0, Status='live', Symbol='Rm62', Aliases='Dmel_CG10279, 4136, CG10279, DmRH8, Dmel\\CG10279, Dmp68, Lip, RM62, dmP68, l(3)01086, l(3)j3D2, l(3)j3D5, l(3)rG338, l(3)s5196, p68, rm62', description='CG10279 gene product from transcript CG10279-RK', other_designations='CG10279-PA|CG10279-PB|CG10279-PC|CG10279-PD|CG10279-PE|CG10279-PF|CG10279-PH|CG10279-PI|CG10279-PJ|CG10279-PK|CG10279-PL|CG10279-PM|Lighten-up|Rm62-PA|Rm62-PB|Rm62-PC|Rm62-PD|Rm62-PE|Rm62-PF|Rm62-PH|Rm62-PI|Rm62-PJ|Rm62-PK|Rm62-PL|Rm62-PM|helicase p62|lighten up', map_location='83D1-83D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6000426, end_position_on_the_genomic_accession=6008579, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40740, CurrentID=0, Status='live', Symbol='CG10280', Aliases='Dmel_ Dmel\\CG10280', description='CG10280 gene product from transcript CG10280-RA', other_designations='CG10280-PA|CG10280-PB|uncharacterized protein', map_location='83D2-83D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6010573, end_position_on_the_genomic_accession=6011932, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40741, CurrentID=0, Status='live', Symbol='CG11459', Aliases='Dmel_ Dmel\\CG11459', description='CG11459 gene product from transcript CG11459-RB', other_designations='CG11459-PA|CG11459-PB|uncharacterized protein', map_location='83D2-83D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6025826, end_position_on_the_genomic_accession=6027238, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40743, CurrentID=0, Status='live', Symbol='CG15580', Aliases='Dmel_ Dmel\\CG15580', description='CG15580 gene product from transcript CG15580-RA', other_designations='CG15580-PA|uncharacterized protein', map_location='83D2-83D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6054712, end_position_on_the_genomic_accession=6057536, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40744, CurrentID=0, Status='live', Symbol='Or83c', Aliases='Dmel_CG15581, 83D.1, 83c, CG15581, DOR116, Dmel\\CG15581', description='Odorant receptor 83c', other_designations='CG15581-PA|Or83c-PA|odorant receptor 83c', map_location='83D4-83D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6076804, end_position_on_the_genomic_accession=6078283, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40745, CurrentID=0, Status='live', Symbol='Gasp', Aliases='Dmel_CG10287, CG10287, Dmel\\CG10287, anon-EST:ParkEST180, gammaBert27, gasp, obst-C', description='CG10287 gene product from transcript CG10287-RB', other_designations='CG10287-PA|CG10287-PB|Gasp-PA|Gasp-PB|gasp', map_location='83D4-83D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6088799, end_position_on_the_genomic_accession=6100887, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40746, CurrentID=0, Status='live', Symbol='Obp83cd', Aliases='Dmel_CG15582, 83c, 83c-1, 83c-2, 83d, CG15582, DmelObp83cd, Dmel\\CG15582, Obp83c, Obp83d', description='Odorant-binding protein 83cd', other_designations='CG15582-PA|Obp83cd-PA|Odorant-binding protein 83c|Odorant-binding protein 83d|odorant receptor 83cd', map_location='83D4-83D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6108134, end_position_on_the_genomic_accession=6109733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40747, CurrentID=0, Status='live', Symbol='Obp83ef', Aliases='Dmel_CG31557, 83f, 83f-1, 83f-2, CG15583, CG31557, DmelObp83ef, Dmel\\CG31557, OBP83e, Obp83e, Obp83f', description='Odorant-binding protein 83ef', other_designations='CG31557-PA|Obp83ef-PA|Odorant-binding protein 83f|pseudogene 83e', map_location='83D4-83D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6109642, end_position_on_the_genomic_accession=6111034, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40748, CurrentID=0, Status='live', Symbol='Vha14-2', Aliases='Dmel_CG1076, CG1076, Dmel\\CG1076, vha14-2', description='Vacuolar H[+] ATPase 14kD subunit 2', other_designations='CG1076-PA|CG1076-PB|CG1076-PC|Vha14-2-PA|Vha14-2-PB|Vha14-2-PC|vacuolar H[+] ATPase 14kD subunit 2|vacuolar H[+] ATPase subunit 14-2', map_location='83D4-83D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6122592, end_position_on_the_genomic_accession=6124405, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40749, CurrentID=0, Status='live', Symbol='CG31559', Aliases='Dmel_ 31559, CG11461, CG15584, Dmel\\CG31559', description='CG31559 gene product from transcript CG31559-RA', other_designations='CG31559-PA|uncharacterized protein', map_location='83D5-83D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6150116, end_position_on_the_genomic_accession=6157235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40751, CurrentID=0, Status='live', Symbol='CG1077', Aliases='Dmel_ Dmel\\CG1077, anon-f', description='CG1077 gene product from transcript CG1077-RA', other_designations='CG1077-PA', map_location='83D5-83D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6167483, end_position_on_the_genomic_accession=6170309, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40752, CurrentID=0, Status='live', Symbol='Osi1', Aliases='Dmel_CG15585, CG15585, Dmel\\CG15585, osi1', description='Osiris 1', other_designations='CG15585-PA|Osi1-PA|osiris|osiris 1', map_location='83D5-83D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6173549, end_position_on_the_genomic_accession=6174971, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40754, CurrentID=0, Status='live', Symbol='NPFR', Aliases='Dmel_CG1147, CG1147, DmNPFR1, Dmel\\CG1147, NPF-R-1, NPFR1, NPFr, RE18294p, anon-WO0131005.3, anon-WO0131005.5, dNPFR1, npfr1, NPFR', description='Neuropeptide F receptor', other_designations='CG1147-PA|CG1147-PB|CG1147-PC|CG1147-PD|NPF-receptor|NPFR-PA|NPFR-PB|NPFR-PC|NPFR-PD|dNPF receptor|neuropeptide F receptor|neuropeptide F receptor 1|neuropeptide F receptor-1', map_location='83D5-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6187315, end_position_on_the_genomic_accession=6195403, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40755, CurrentID=0, Status='live', Symbol='Osi24', Aliases='Dmel_CG15589, CG15589, Dmel\\CG15589', description='Osiris 24', other_designations='CG15589-PA|Osi24-PA|osiris 24', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6198811, end_position_on_the_genomic_accession=6207678, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40756, CurrentID=0, Status='live', Symbol='Osi2', Aliases='Dmel_CG1148, CG1148, Dmel\\CG1148', description='Osiris 2', other_designations='CG1148-PA|CG1148-PB|CG1148-PC|Osi2-PA|Osi2-PB|Osi2-PC|osiris|osiris 2', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6209373, end_position_on_the_genomic_accession=6213493, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40757, CurrentID=0, Status='live', Symbol='Osi3', Aliases='Dmel_CG1150, CG1150, Dmel\\CG1150, osi3', description='Osiris 3', other_designations='CG1150-PA|CG1150-PB|Osi3-PA|Osi3-PB|osiris|osiris 3', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6215832, end_position_on_the_genomic_accession=6218199, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40758, CurrentID=0, Status='live', Symbol='Osi4', Aliases='Dmel_CG10303, CG10303, Dmel\\CG10303', description='Osiris 4', other_designations='CG10303-PA|CG10303-PB|Osi4-PA|Osi4-PB|osiris|osiris 4', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6218318, end_position_on_the_genomic_accession=6221781, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40759, CurrentID=0, Status='live', Symbol='Osi5', Aliases='Dmel_CG15590, CG15590, Dmel\\CG15590', description='Osiris 5', other_designations='CG15590-PA|CG15590-PB|Osi5-PA|Osi5-PB|osiris|osiris 5', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6227512, end_position_on_the_genomic_accession=6228828, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40760, CurrentID=0, Status='live', Symbol='Osi6', Aliases='Dmel_CG1151, BcDNA:LD21503, CG1151, Dmel\\CG1151', description='Osiris 6', other_designations='CG1151-PA|CG1151-PB|Osi6-PA|Osi6-PB|osiris|osiris 6', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6234708, end_position_on_the_genomic_accession=6237181, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40761, CurrentID=0, Status='live', Symbol='Osi7', Aliases='Dmel_CG1153, CG1153, Dmel\\CG1153, osi7', description='Osiris 7', other_designations='CG1153-PA|Osi7-PA|osiris|osiris 7', map_location='83E1-83E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6248994, end_position_on_the_genomic_accession=6251683, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40762, CurrentID=0, Status='live', Symbol='Osi8', Aliases='Dmel_CG15591, CG15591, Dmel\\CG15591', description='Osiris 8', other_designations='CG15591-PA|Osi8-PA|osiris|osiris 8', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6255175, end_position_on_the_genomic_accession=6256602, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40763, CurrentID=0, Status='live', Symbol='Osi9', Aliases='Dmel_CG15592, CG15592, Dmel\\CG15592', description='Osiris 9', other_designations='CG15592-PA|Osi9-PA|osiris|osiris 9', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6260211, end_position_on_the_genomic_accession=6262105, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40764, CurrentID=0, Status='live', Symbol='Osi10', Aliases='Dmel_CG15593, CG15593, Dmel\\CG15593', description='Osiris 10', other_designations='CG15593-PA|Osi10-PA|osiris|osiris 10', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6263976, end_position_on_the_genomic_accession=6267395, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40765, CurrentID=0, Status='live', Symbol='Osi11', Aliases='Dmel_CG15596, CG15596, Dmel\\CG15596, osi11', description='Osiris 11', other_designations='CG15596-PA|Osi11-PA|osiris|osiris 11', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6267591, end_position_on_the_genomic_accession=6268757, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40766, CurrentID=0, Status='live', Symbol='Osi12', Aliases='Dmel_CG1154, CG1154, Dmel\\CG1154, osi12', description='Osiris 12', other_designations='CG1154-PA|Osi12-PA|osiris|osiris 12', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6278531, end_position_on_the_genomic_accession=6282495, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40767, CurrentID=0, Status='live', Symbol='CG15597', Aliases='Dmel_ Dmel\\CG15597', description='CG15597 gene product from transcript CG15597-RB', other_designations='CG15597-PB|uncharacterized protein', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6284462, end_position_on_the_genomic_accession=6285200, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40768, CurrentID=0, Status='live', Symbol='CG15594', Aliases='Dmel_ Dmel\\CG15594', description='CG15594 gene product from transcript CG15594-RB', other_designations='CG15594-PB|uncharacterized protein', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6287733, end_position_on_the_genomic_accession=6289824, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40769, CurrentID=0, Status='live', Symbol='Osi13', Aliases='Dmel_CG15595, CG15595, Dmel\\CG15595', description='Osiris 13', other_designations='CG15595-PA|Osi13-PA|osiris|osiris 13', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6291737, end_position_on_the_genomic_accession=6293489, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40770, CurrentID=0, Status='live', Symbol='Osi14', Aliases='Dmel_CG1155, BEST:LD24139, CG1155, Dmel\\CG1155, SP558, Sp558, osi14', description='Osiris 14', other_designations='CG1155-PA|Osi14-PA|osiris|osiris 14', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6299128, end_position_on_the_genomic_accession=6300738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40771, CurrentID=0, Status='live', Symbol='Osi15', Aliases='Dmel_CG1157, CG1157, Dmel\\CG1157, osi15', description='Osiris 15', other_designations='CG1157-PA|Osi15-PA|osiris|osiris 15', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6301804, end_position_on_the_genomic_accession=6303653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40772, CurrentID=0, Status='live', Symbol='CG31560', Aliases='Dmel_ CG10314, Dmel\\CG31560', description='CG31560 gene product from transcript CG31560-RA', other_designations='CG31560-PA|uncharacterized protein', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6309976, end_position_on_the_genomic_accession=6310998, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40774, CurrentID=0, Status='live', Symbol='Osi17', Aliases='Dmel_CG15598, CG15598, Dmel\\CG15598, mdcds_43703', description='Osiris 17', other_designations='CG15598-PA|CG15598-PB|CG15598-PC|CG15598-PD|Osi17-PA|Osi17-PB|Osi17-PC|Osi17-PD|osiris|osiris 17', map_location='83E2-83E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6311772, end_position_on_the_genomic_accession=6327308, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40775, CurrentID=0, Status='live', Symbol='Osi18', Aliases='Dmel_CG1169, CG1169, Dmel\\CG1169', description='Osiris 18', other_designations='CG1169-PA|Osi18-PA|osiris|osiris 18', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6330248, end_position_on_the_genomic_accession=6331809, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40776, CurrentID=0, Status='live', Symbol='Osi19', Aliases='Dmel_CG15189, CG15189, Dmel\\CG15189', description='Osiris 19', other_designations='CG15189-PA|CG15189-PB|Osi19-PA|Osi19-PB|osiris|osiris 19', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6335362, end_position_on_the_genomic_accession=6337196, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40777, CurrentID=0, Status='live', Symbol='Osi20', Aliases='Dmel_CG15188, CG15188, Dmel\\CG15188', description='Osiris 20', other_designations='CG15188-PA|Osi20-PA|osiris|osiris 20', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6340095, end_position_on_the_genomic_accession=6341424, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40778, CurrentID=0, Status='live', Symbol='CG17917', Aliases='Dmel_ Dmel\\CG17917', description='CG17917 gene product from transcript CG17917-RA', other_designations='CG17917-PA|uncharacterized protein', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6341968, end_position_on_the_genomic_accession=6342873, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40779, CurrentID=0, Status='live', Symbol='CG10298', Aliases='Dmel_ BcDNA:GH28351, Dmel\\CG10298', description='CG10298 gene product from transcript CG10298-RA', other_designations='CG10298-PA|uncharacterized protein', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6343002, end_position_on_the_genomic_accession=6343799, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40780, CurrentID=0, Status='live', Symbol='CG17919', Aliases='Dmel_ Dmel\\CG17919', description='CG17919 gene product from transcript CG17919-RA', other_designations='CG17919-PA|uncharacterized protein', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6344029, end_position_on_the_genomic_accession=6345087, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40781, CurrentID=0, Status='live', Symbol='SmD2', Aliases='Dmel_CG1249, BcDNA:LD03002, CG1249, D2-Sm, Dmel\\CG1249, SMD2_DROME, dD2, snRNP D2, snRNP2', description='Small ribonucleoprotein particle protein SmD2', other_designations='CG1249-PA|SmD2-PA|small ribonucleoprotein particle protein D2|small ribonucleoprotein particle protein SmD2|snRNP2', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6345000, end_position_on_the_genomic_accession=6345911, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40782, CurrentID=0, Status='live', Symbol='CG18048', Aliases='Dmel_ Dmel\\CG18048', description='CG18048 gene product from transcript CG18048-RA', other_designations='CG18048-PA|uncharacterized protein', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6346111, end_position_on_the_genomic_accession=6347519, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40783, CurrentID=0, Status='live', Symbol='Hr83', Aliases='Dmel_CG10296, CG10296, DHR83, Dmel\\CG10296, NR2E5', description='Hormone receptor 83', other_designations='CG10296-PA|Hr83-PA|drosophila hormone receptor 83|hormone receptor 83', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6349243, end_position_on_the_genomic_accession=6350512, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40785, CurrentID=0, Status='live', Symbol='CG42724', Aliases='Dmel_ CG15187, CG31367, CG33097, Dmel\\CG42724, Dmel_CG31367, Dmel_CG33097', description='CG42724 gene product from transcript CG42724-RC', other_designations='CG42724-PC|CG42724-PE|CG42724-PG|CG42724-PI|CG42724-PK|CG42724-PL|CG42724-PM|uncharacterized protein', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6361672, end_position_on_the_genomic_accession=6368288, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  40786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40786, CurrentID=0, Status='live', Symbol='CG10286', Aliases='Dmel_ Dmel\\CG10286', description='CG10286 gene product from transcript CG10286-RA', other_designations='CG10286-PA|uncharacterized protein', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6368965, end_position_on_the_genomic_accession=6371115, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40787, CurrentID=0, Status='live', Symbol='CRAT', Aliases='Dmel_CG1041, CG1041, Dmel\\CG1041', description='Carnitine O-Acetyl-Transferase', other_designations='CG1041-PA|CG1041-PB|CG1041-PD|CG1041-PE|CRAT-PA|CRAT-PB|CRAT-PD|CRAT-PE|carnitine O-Acetyl-Transferase|carnitine O-acetyl-transferase', map_location='83E4-83E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6371026, end_position_on_the_genomic_accession=6376044, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40788, CurrentID=0, Status='live', Symbol='CG42564', Aliases='Dmel_ CG10284, Dmel\\CG42564, Dmel_CG10284', description='CG42564 gene product from transcript CG42564-RA', other_designations='CG42564-PA|uncharacterized protein', map_location='83E4-83E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6374475, end_position_on_the_genomic_accession=6377348, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40789, CurrentID=0, Status='live', Symbol='CG1024', Aliases='Dmel_ BcDNA:LD21969, Dmel\\CG1024', description='CG1024 gene product from transcript CG1024-RE', other_designations='CG1024-PB|CG1024-PD|CG1024-PE', map_location='83E5-83E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6377349, end_position_on_the_genomic_accession=6380328, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40790, CurrentID=0, Status='live', Symbol='TfIIFalpha', Aliases='Dmel_CG10281, 154355_at, CG10281, DmTFIIF, Dmel\\CG10281, F5a, TFIIF, TFIIF-L, TFIIFa, TFIIFalpha, TfIIF', description='Transcription factor IIFalpha', other_designations='CG10281-PA|TFIIF large subunit|TfIIFalpha-PA|factor 5|factor 5a|large subunit of transcription factor 5|transcription factor IIFalpha', map_location='83E5-83E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6380467, end_position_on_the_genomic_accession=6382604, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40791, CurrentID=0, Status='live', Symbol='godzilla', Aliases='Dmel_CG10277, CG10277, Dmel\\CG10277[CG10277], godzilla', description='godzilla E3 ubiquitin protein ligase', other_designations='CG10277-PA|CG10277-PB|CG10277-PC|CG10277-PD|godzilla-PA|godzilla-PB|godzilla-PC|godzilla-PD', map_location='83E5-83E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6382914, end_position_on_the_genomic_accession=6388304, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40792, CurrentID=0, Status='live', Symbol='Dmtn', Aliases='Dmel_CG1021, CG1021, Dmel\\CG1021, anon-WO0118547.375, dmtn', description='Dementin', other_designations='CG1021-PD|CG1021-PE|CG1021-PF|CG1021-PG|CG1021-PH|Dmtn-PD|Dmtn-PE|Dmtn-PF|Dmtn-PG|Dmtn-PH|dementin', map_location='83E5-83E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6389162, end_position_on_the_genomic_accession=6405895, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  40793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40793, CurrentID=0, Status='live', Symbol='gpp', Aliases='Dmel_CG42803, 0732/14, BcDNA:SD05230, CG10272, CG33324, CG42803, DOT1L, Dmel\\CG42803, Dmel_CG10272, Dmel_CG33324, Gpp, anon-WO0118547.319, dDOT1L, dDot1, grp, l(3)03342, l(3)S073214b, l(3)sA3484', description='grappa', other_designations='CG42803-PB|CG42803-PC|CG42803-PD|CG42803-PE|CG42803-PF|gpp-PB|gpp-PC|gpp-PD|gpp-PE|gpp-PF|lethal (3) S073214b', map_location='83E6-83E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6406855, end_position_on_the_genomic_accession=6449274, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40794, CurrentID=0, Status='live', Symbol='CG9727', Aliases='Dmel_ Dmel\\CG9727', description='CG9727 gene product from transcript CG9727-RC', other_designations='CG9727-PA|CG9727-PB|CG9727-PC', map_location='83E7-83E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6449944, end_position_on_the_genomic_accession=6456676, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40795, CurrentID=0, Status='live', Symbol='Zif', Aliases='Dmel_CG10267, CG10267, Dmel\\CG10267, anon-WO0118547.391', description='Zinc-finger protein', other_designations='CG10267-PA|CG10267-PB|CG10267-PC|Zif-PA|Zif-PB|Zif-PC', map_location='83E8-83E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6457088, end_position_on_the_genomic_accession=6458822, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40796, CurrentID=0, Status='live', Symbol='Neurochondrin', Aliases='Dmel_CG2330, CG2330, Dmel\\CG2330', description='CG2330 gene product from transcript CG2330-RA', other_designations='CG2330-PA|Neurochondrin-PA|dNeurochondrin|neurochondrin', map_location='83E8-83E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6468437, end_position_on_the_genomic_accession=6471016, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40797, CurrentID=0, Status='live', Symbol='CG15186', Aliases='Dmel_ Dmel\\CG15186', description='CG15186 gene product from transcript CG15186-RI', other_designations='CG15186-PA|CG15186-PB|CG15186-PC|CG15186-PE|CG15186-PF|CG15186-PG|CG15186-PH|CG15186-PI|CG15186-PJ|CG15186-PK|CG15186-PL|CG15186-PM|uncharacterized protein', map_location='83E8-83F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6478311, end_position_on_the_genomic_accession=6495025, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  40798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40798, CurrentID=0, Status='live', Symbol='CG15185', Aliases='Dmel_ Dmel\\CG15185', description='CG15185 gene product from transcript CG15185-RA', other_designations='CG15185-PA|uncharacterized protein', map_location='83F1-83F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6527508, end_position_on_the_genomic_accession=6528674, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40800, CurrentID=0, Status='live', Symbol='dpr11', Aliases='Dmel_CG33202, BcDNA:GH22307, CG15181, CG15182, CG15183, CG31309, CG33202, CT35098, Dmel\\CG33202, Dpr-11, Dpr11', description='defective proboscis extension response 11', other_designations='CG33202-PB|CG33202-PC|dpr11-PB|dpr11-PC', map_location='83F1-83F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6529169, end_position_on_the_genomic_accession=6593330, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40803, CurrentID=0, Status='live', Symbol='CG1137', Aliases='Dmel_ Dmel\\CG1137, anon-WO0118547.86', description='CG1137 gene product from transcript CG1137-RA', other_designations='CG1137-PA', map_location='83F1-83F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6558618, end_position_on_the_genomic_accession=6560351, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40804, CurrentID=0, Status='live', Symbol='CG2336', Aliases='Dmel_ Dmel\\CG2336', description='CG2336 gene product from transcript CG2336-RA', other_designations='CG2336-PA', map_location='83F3-83F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6580784, end_position_on_the_genomic_accession=6582271, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40805, CurrentID=0, Status='live', Symbol='CG1138', Aliases='Dmel_ Dmel\\CG1138', description='CG1138 gene product from transcript CG1138-RA', other_designations='CG1138-PA', map_location='83F4-83F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6596625, end_position_on_the_genomic_accession=6598022, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40806, CurrentID=0, Status='live', Symbol='CG31482', Aliases='Dmel_ CG2337, Dmel\\CG31482', description='CG31482 gene product from transcript CG31482-RA', other_designations='CG31482-PA|uncharacterized protein', map_location='83F4-83F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6601242, end_position_on_the_genomic_accession=6602201, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40807, CurrentID=0, Status='live', Symbol='glob2', Aliases='Dmel_CG15180, CG15180, DmeGb2, Dmel\\CG15180, anon-SAGE:Wang-121', description='globin 2', other_designations='CG15180-PB|CG15180-PC|CG15180-PD|anon-SAGE:Wang-121|glob2-PB|glob2-PC|glob2-PD', map_location='83F4-83F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6605997, end_position_on_the_genomic_accession=6608914, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40809, CurrentID=0, Status='live', Symbol='sowi', Aliases='Dmel_CG15178, CG15178, Dmel\\CG15178', description='solwind', other_designations='CG15178-PA|sowi-PA', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6626930, end_position_on_the_genomic_accession=6627900, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40810, CurrentID=0, Status='live', Symbol='CG15177', Aliases='Dmel_ Dmel\\CG15177', description='CG15177 gene product from transcript CG15177-RB', other_designations='CG15177-PA|CG15177-PB|uncharacterized protein', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6628204, end_position_on_the_genomic_accession=6629321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40811, CurrentID=0, Status='live', Symbol='sunz', Aliases='Dmel_CG15179, BcDNA:AT31060, CG15179, Dmel\\CG15179', description='sungrazer', other_designations='CG15179-PA|sunz-PA', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6630151, end_position_on_the_genomic_accession=6631184, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40812, CurrentID=0, Status='live', Symbol='CG1315', Aliases='Dmel_ BG:DS00004.14, Dmel\\CG1315', description='CG1315 gene product from transcript CG1315-RB', other_designations='CG1315-PA|CG1315-PB|argininosuccinate synthase|argininosuccinate synthase-like', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6642262, end_position_on_the_genomic_accession=6644434, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40813, CurrentID=0, Status='live', Symbol='Taf1', Aliases='Dmel_CG17603, BG:DS00004.13, CG17603, Dmel\\CG17603, EfW1, SR3-5, TAF, TAF1, TAF200, TAF230, TAF250, TAF250/230, TAFII-250, TAFII250, TAF[II]250, TAF[[II]], TAF[[II]]230, TAF[[II]]250, TAF[[II]]250/230, TFIID, TFIID TAF250p, Taf200, Taf230, Taf250, Taf[[II]]250, cel, cell, d230, dTAF230, dTAF250, dTAFII250, dTAF[[II]]230, dTAF[[II]]250, dmTAF1, dmTAF[[II]]230, l(3)84Ab, p230, Taf1', description='TBP-associated factor 1', other_designations='CG17603-PA|CG17603-PB|CG17603-PC|CG17603-PD|CG17603-PE|CG17603-PF|TATA-associated factor 250|TBP-associated factor|TBP-associated factor 250|TBP-associated factor 250kD|TFIID 230 kda subunit|Taf1-PA|Taf1-PB|Taf1-PC|Taf1-PD|Taf1-PE|Taf1-PF|cell lethal|suppressor of Ras85D 3-5', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6646889, end_position_on_the_genomic_accession=6656048, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  40814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40814, CurrentID=0, Status='live', Symbol='CG1307', Aliases='Dmel_ BG:DS00004.12, Dmel\\CG1307', description='CG1307 gene product from transcript CG1307-RE', other_designations='CG1307-PC|CG1307-PE|hypothetical 23.1kd-like protein', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6656030, end_position_on_the_genomic_accession=6657442, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40815, CurrentID=0, Status='live', Symbol='twr', Aliases='Dmel_CG2358, AF160889, BG:DS00004.11, BcDNA:GM04682, CG2358, Dmel\\CG2358, EfW5, Ov15, Spase18-21, Spase18/21, l(3)05614, l(3)84Aa', description='twisted bristles roughened eye', other_designations='CG2358-PA|SPC 21-kDa-like|spase 18/21-subunit|twr-PA', map_location='84A1-84A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6657882, end_position_on_the_genomic_accession=6659442, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40816, CurrentID=0, Status='live', Symbol='agt', Aliases='Dmel_CG1303, Agt, BG:DS00004.10, CG1303, DmAGT, Dmel\\CG1303', description='O-6-alkylguanine-DNA alkyltransferase', other_designations='CG1303-PA|CG1303-PB|agt-PA|agt-PB', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6658806, end_position_on_the_genomic_accession=6660100, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40817, CurrentID=0, Status='live', Symbol='lab', Aliases='Dmel_CG1264, BG:DS00004.9, CG1264, Dm lab, DmLab, Dmel\\CG1264, Dmlab, EfR9, F121, F24, F90, F90-2, LAB, Lab, l(3)01241, l(3)84Ac, lb', description='labial', other_designations='CG1264-PA|CG1264-PB|lab-PA|lab-PB', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6661427, end_position_on_the_genomic_accession=6678590, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40818, CurrentID=0, Status='live', Symbol='Edg84A', Aliases='Dmel_CG2345, BG:DS00004.8, CG2345, DM-EDG84, DmelEDG84A, Dmel\\CG2345, EDG-84, EDG-84A, EDG84, EDG84A, Edg84, cc1', description='Ecdysone-dependent gene 84A', other_designations='CG2345-PA|Edg84A-PA|cuticle cluster 1|cuticle-like cluster|ecdysone-dependent gene-84A-1', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6684445, end_position_on_the_genomic_accession=6685157, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40819, CurrentID=0, Status='live', Symbol='Ccp84Ag', Aliases='Dmel_CG2342, BG:DS00004.7, BcDNA:RE04513, CG2342, Ccp, Dmel Dmel\\CG2342, cc2', description='CG2342 gene product from transcript CG2342-RA', other_designations='CG2342-PA|Ccp84Ag-PA|cuticle cluster 2', map_location='84A2-84A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6686728, end_position_on_the_genomic_accession=6688039, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40820, CurrentID=0, Status='live', Symbol='Ccp84Af', Aliases='Dmel_CG1331, BG:DS00004.6, CG1331, Ccp, Dmel Dmel\\CG1331, cc3', description='CG1331 gene product from transcript CG1331-RA', other_designations='CG1331-PA|Ccp84Af-PA|cuticle cluster 3', map_location='84A3-84A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6689602, end_position_on_the_genomic_accession=6690239, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40821, CurrentID=0, Status='live', Symbol='Ccp84Ae', Aliases='Dmel_CG1330, BG:DS00004.5, CG1330, Ccp, Dmel Dmel\\CG1330, cc4', description='CG1330 gene product from transcript CG1330-RA', other_designations='CG1330-PA|Ccp84Ae-PA|cuticle cluster 4', map_location='84A3-84A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6690887, end_position_on_the_genomic_accession=6691758, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40822, CurrentID=0, Status='live', Symbol='Ccp84Ad', Aliases='Dmel_CG2341, BG:DS00004.4, CG2341, Ccp, Dmel Dmel\\CG2341, cc5', description='CG2341 gene product from transcript CG2341-RA', other_designations='CG2341-PA|Ccp84Ad-PA|cuticle cluster 5', map_location='84A3-84A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6692732, end_position_on_the_genomic_accession=6693527, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40823, CurrentID=0, Status='live', Symbol='Ccp84Ac', Aliases='Dmel_CG1327, BG:DS00004.3, CG1327, Ccp, Dmel Dmel\\CG1327, cc6', description='CG1327 gene product from transcript CG1327-RA', other_designations='CG1327-PA|Ccp84Ac-PA|cuticle cluster 6|larval cuticle', map_location='84A3-84A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6695148, end_position_on_the_genomic_accession=6696271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40824, CurrentID=0, Status='live', Symbol='Ccp84Ab', Aliases='Dmel_CG1252, BG:DS00004.2, CG1252, Ccp, Dmel Dmel\\CG1252, cc7', description='CG1252 gene product from transcript CG1252-RA', other_designations='CG1252-PA|Ccp84Ab-PA|cuticle cluster 7', map_location='84A4-84A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6701968, end_position_on_the_genomic_accession=6702877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40825, CurrentID=0, Status='live', Symbol='Ccp84Aa', Aliases='Dmel_CG2360, BG:DS00004.1, CG2360, Ccp, Dmel Dmel\\CG2360, cc8', description='CG2360 gene product from transcript CG2360-RA', other_designations='CG2360-PA|Ccp84Aa-PA|cuticle cluster 8', map_location='84A4-84A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6704084, end_position_on_the_genomic_accession=6704903, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40826, CurrentID=0, Status='live', Symbol='pb', Aliases='Dmel_CG31481, BG:DS01719.1, CG17880, CG31481, Dmel\\CG31481, Dmpb, PB, Pb, l(3)04498, pd, pr', description='proboscipedia', other_designations='CG31481-PA|CG31481-PB|CG31481-PC|CG31481-PD|pb-PA|pb-PB|pb-PC|pb-PD|proboscopedia|proposcipedia', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6707800, end_position_on_the_genomic_accession=6742063, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40827, CurrentID=0, Status='live', Symbol='zen2', Aliases='Dmel_CG1048, BG:DS00276.10, CG1048, Dmel\\CG1048, z2, zpr', description='zerknullt-related', other_designations='CG1048-PA|CG1048-PB|Zerknullt-related|zen-pattern-related|zen2-PA|zen2-PB|zerknullt 2', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6743135, end_position_on_the_genomic_accession=6744136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40828, CurrentID=0, Status='live', Symbol='zen', Aliases='Dmel_CG1046, BG:DS00276.9, CG1046, Dmel\\CG1046, ZEN, ZEN VRE, Zen, dm-zen, l(3)84Ad, z11, zen', description='zerknullt', other_designations='CG1046-PA|zen-PA|zerknuellt|zerknullt 1|zerknult', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6752864, end_position_on_the_genomic_accession=6754194, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40830, CurrentID=0, Status='live', Symbol='bcd', Aliases='Dmel_CG1034, BCD, BG:DS00276.7, Bcd, CG1034, Dmel\\CG1034, bic, mum, prd4', description='bicoid', other_designations='CG1034-PA|CG1034-PD|CG1034-PE|CG1034-PF|CG1034-PG|PRD gene 4|bcd-PA|bcd-PD|bcd-PE|bcd-PF|bcd-PG|multimorph', map_location='3-48 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6755842, end_position_on_the_genomic_accession=6759466, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40831, CurrentID=0, Status='live', Symbol='Ama', Aliases='Dmel_CG2198, AMA, BG:DS00276.6, CG2198, CT7244, Dmel\\CG2198, M109, ama', description='Amalgam', other_designations='Ama-PA|Ama-PB|Ama-PC|CG2198-PA|CG2198-PB|CG2198-PC|amalgam', map_location='84A5-84A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6762593, end_position_on_the_genomic_accession=6765261, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40832, CurrentID=0, Status='live', Symbol='Dfd', Aliases='Dmel_CG2189, BG:DS00276.5, CG2189, DFD, DmDfd, Dmel\\CG2189, EbR11, dfd, l(3)84Ae', description='Deformed', other_designations='CG2189-PA|Dfd-PA|deformed', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6791836, end_position_on_the_genomic_accession=6802428, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40833, CurrentID=0, Status='live', Symbol='Scr', Aliases='Dmel_CG1030, BG:DS07876.2, CG1030, DmScr, Dmel\\CG1030, Msc, SCR, l(3)84Af, scr', description='Sex combs reduced', other_designations='CG1030-PA|CG1030-PB|Scr-PA|Scr-PB|multi sex combs|multiple sex comb|sex combs reduced|sex-combs reduced', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6823120, end_position_on_the_genomic_accession=6849981, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40834, CurrentID=0, Status='live', Symbol='ftz', Aliases='Dmel_CG2047, BG:DS07876.1, CG2047, Dm-Ftz, Dm-ftz, DmFtz, Dmel\\CG2047, Dmftz, FTZ, Ftz, Ual, l(3)84Ag', description='fushi tarazu', other_designations='CG2047-PA|Fushi-Tarazu|Ultra-abdominal-like|ftz-PA|fushitarazu', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6864324, end_position_on_the_genomic_accession=6866244, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40835, CurrentID=0, Status='live', Symbol='Antp', Aliases='Dmel_CG1028, 3.4, ANT-C, ANT-P, ANTC, ANTP, Ant, AntP, AntP1 P1, Antp P2, Antp1, Aus, BG:DS07700.1, CG1028, DMANTPE1, DRO15DC96Z, DmAntp, Dmel\\CG1028, Hu, Ns, Scx, antp, l(3)84Ba, Antp', description='Antennapedia', other_designations='Antp-PD|Antp-PE|Antp-PF|Antp-PG|Antp-PH|Antp-PI|Antp-PJ|Antp-PK|Antp-PL|Antp-PM|Antp-PN|Aus|CG1028-PD|CG1028-PE|CG1028-PF|CG1028-PG|CG1028-PH|CG1028-PI|CG1028-PJ|CG1028-PK|CG1028-PL|CG1028-PM|CG1028-PN|Dm-Antennapedia|Sex comb extra|Sex combs extra|antenapedia|antennapaedia|antennapedia|antennapedia P1|antennapedia P2|atennapedia|humeral|nasobemia', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6896253, end_position_on_the_genomic_accession=6999228, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40836, CurrentID=0, Status='live', Symbol='Sodh-1', Aliases='Dmel_CG1982, BG:DS00464.2, CG1982, DM_7298873, Dmel\\CG1982, Sdh-1, Sodh', description='Sorbitol dehydrogenase 1', other_designations='CG1982-PA|CG1982-PB|Sodh-1-PA|Sodh-1-PB|sorbitol dehydrogenase 1|sorbitol dehydrogenase like', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7052335, end_position_on_the_genomic_accession=7054541, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40837, CurrentID=0, Status='live', Symbol='CG1979', Aliases='Dmel_ BG:DS00464.1, BcDNA:GH19145, Dmel\\CG1979', description='CG1979 gene product from transcript CG1979-RA', other_designations='CG1979-PA', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7054586, end_position_on_the_genomic_accession=7058671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40838, CurrentID=0, Status='live', Symbol='dj', Aliases='Dmel_CG1980, CG1980, DJ, Dj, Dmel\\CG1980, bs02h03.y1', description='don juan', other_designations='CG1980-PA|CG1980-PB|CG1980-PC|Don juan|dj-PA|dj-PB|dj-PC|don-juan', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7059315, end_position_on_the_genomic_accession=7060524, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40839, CurrentID=0, Status='live', Symbol='djl', Aliases='Dmel_CG1984, CG1984, Dmel\\CG1984, Don Juan-like, dj like, dj-like', description='don juan like', other_designations='CG1984-PA|Don Juan-like|djl-PA', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7060767, end_position_on_the_genomic_accession=7061919, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40840, CurrentID=0, Status='live', Symbol='CG1988', Aliases='Dmel_ Dmel\\CG1988', description='CG1988 gene product from transcript CG1988-RB', other_designations='CG1988-PB|CG1988-PC|CG1988-PD', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7062175, end_position_on_the_genomic_accession=7064370, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40841, CurrentID=0, Status='live', Symbol='CG1105', Aliases='Dmel_ Dmel\\CG1105', description='CG1105 gene product from transcript CG1105-RA', other_designations='CG1105-PA', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7064186, end_position_on_the_genomic_accession=7067136, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40842, CurrentID=0, Status='live', Symbol='CG1965', Aliases='Dmel_ Dmel\\CG1965', description='CG1965 gene product from transcript CG1965-RC', other_designations='CG1965-PA|CG1965-PB|CG1965-PC', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7067401, end_position_on_the_genomic_accession=7071624, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  40843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40843, CurrentID=0, Status='live', Symbol='CG1104', Aliases='Dmel_ Dmel\\CG1104', description='CG1104 gene product from transcript CG1104-RC', other_designations='CG1104-PA|CG1104-PC', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7071624, end_position_on_the_genomic_accession=7074452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40844, CurrentID=0, Status='live', Symbol='CG1943', Aliases='Dmel_ Dmel\\CG1943', description='CG1943 gene product from transcript CG1943-RD', other_designations='CG1943-PA|CG1943-PB|CG1943-PC|CG1943-PD', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7074859, end_position_on_the_genomic_accession=7077981, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40846, CurrentID=0, Status='live', Symbol='Dpck', Aliases='Dmel_CG1939, CG1939, DPCK, Dmel\\CG1939, dDPCK', description='Dephospho-CoA kinase', other_designations='CG1939-PA|Dpck-PA', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7082182, end_position_on_the_genomic_accession=7083251, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40847, CurrentID=0, Status='live', Symbol='Tailor', Aliases='Dmel_CG1091, CG1091, Dmel\\CG1091, anon-WO0118547.222', description='CG1091 gene product from transcript CG1091-RB', other_designations='CG1091-PA|CG1091-PB|CG1091-PC|Tailor-PA|Tailor-PB|Tailor-PC|tailor', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7083225, end_position_on_the_genomic_accession=7086083, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40848, CurrentID=0, Status='live', Symbol='alphaTub84B', Aliases='Dmel_CG1913, 1t, ALPHA 84C, BEST:LD32507, CG1913, DM1alpha, DTA1, Dmel\\CG1913, T, TBA1_DROME, Talpha1, Tub, Tuba84B, Tubalpha1, a-tub84B, a1t, aTub84B, alpha-Tub, alpha-Tub84B, alpha-tub, alpha-tub84B, alpha-tubulin, alpha-tubulin 84B, alpha1, alpha1t, alpha1tub, alpha1tub84B, alpha84B, alphaT, alphaTUB, alphaTUB84B, alphaTub, alphaTub1, alphaTub84, alphat-1, alphatub, alphatub84B, anon-EST:Liang-1.59, anon-EST:Liang-2.30, chr3R:2914276..2914416, clone 1.59, clone 2.30, l(3)84Bd, l(3)g3, ms(3)nc33, nc33, tub, tubA84B, tubalpha1, tubulin', description='alpha-Tubulin at 84B', other_designations='CG1913-PA|Tubulin-alpha1|alpha tubulin|alpha tubulin 84B|alpha-Tubulin|alpha-Tubulin 84B|alpha-tubulin 1|alpha-tubulin84B|alpha1 tubulin|alpha1-tubulin|alphaT|alphaTub84B-PA|alphaTubulin 84B|alphaTubulin84B|alpha[[1]] tubulin|alpha[[1]]-tubulin|alphatubulin|tubulin|tubulin 1 alpha-chain|tubulin 1alpha|tubulin alpha1|tubulin alpha1 promoter|tubulin(alpha1)|tubulin1alpha|tubulin84B', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7086599, end_position_on_the_genomic_accession=7088839, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40849, CurrentID=0, Status='live', Symbol='Mlp84B', Aliases='Dmel_CG1019, CG1019, DLim-3, DmLIM-3, Dmel\\CG1019, MLP, MLP84B, Mlmu84B, Mlp, Mlp84b, dMlp84B, mlp, mlp84B', description='Muscle LIM protein at 84B', other_designations='CG1019-PA|CG1019-PB|CG1019-PC|LIM protein 84B|Mlp84B-PA|Mlp84B-PB|Mlp84B-PC|Muscle-specific LIM protein at 84B|muscle LIM protein|muscle LIM protein at 84B', map_location='84C1-84C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7112850, end_position_on_the_genomic_accession=7117732, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40850, CurrentID=0, Status='live', Symbol='Alh', Aliases='Dmel_CG1070, AF10, CG1070, Dalf, Dmel\\CG1070, anon-WO0118547.226, dAF10, l(3)84Be, l(3)g2, l(3)g5, l(3)j13A6, l(3)j8C8, l(3)j8c8, l(3)r13, l(3j8C8, stk', description='Alhambra', other_designations='Alh-PA|Alh-PB|Alh-PD|Alh-PF|Alh-PH|Alh-PI|Alh-PJ|Alh-PK|Alh-PL|Alh-PM|Alh-PN|Alh-PO|Alh-PP|CG1070-PA|CG1070-PB|CG1070-PD|CG1070-PF|CG1070-PH|CG1070-PI|CG1070-PJ|CG1070-PK|CG1070-PL|CG1070-PM|CG1070-PN|CG1070-PO|CG1070-PP|alhambra|sticking', map_location='84B6-84C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7094596, end_position_on_the_genomic_accession=7124185, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  40851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40851, CurrentID=0, Status='live', Symbol='CG10098', Aliases='Dmel_ Dmel\\CG10098', description='CG10098 gene product from transcript CG10098-RA', other_designations='CG10098-PA|uncharacterized protein', map_location='84C3-84C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7124548, end_position_on_the_genomic_accession=7126503, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40852, CurrentID=0, Status='live', Symbol='CG10068', Aliases='Dmel_ Dmel\\CG10068', description='CG10068 gene product from transcript CG10068-RA', other_designations='CG10068-PA|uncharacterized protein', map_location='84C4-84C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7126271, end_position_on_the_genomic_accession=7128027, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40853, CurrentID=0, Status='live', Symbol='CG31248', Aliases='Dmel_ CG2640, Dmel\\CG31248', description='CG31248 gene product from transcript CG31248-RA', other_designations='CG31248-PA|CG31248-PB|uncharacterized protein', map_location='84C4-84C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7129320, end_position_on_the_genomic_accession=7130908, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40856, CurrentID=0, Status='live', Symbol='CG14610', Aliases='Dmel_ Dmel\\CG14610', description='CG14610 gene product from transcript CG14610-RB', other_designations='CG14610-PB|uncharacterized protein', map_location='84C6-84C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7143729, end_position_on_the_genomic_accession=7144295, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40857, CurrentID=0, Status='live', Symbol='CG2656', Aliases='Dmel_ Dmel\\CG2656, NEST:bs33f03', description='CG2656 gene product from transcript CG2656-RA', other_designations='CG2656-PA', map_location='84C6-84C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7144724, end_position_on_the_genomic_accession=7151270, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40858, CurrentID=0, Status='live', Symbol='gfzf', Aliases='Dmel_CG33546, CG10065, CG31329, CG31492, CG33546, Dmel\\CG33546, GFZF, Su(Kpn), cg10065, dGFZF, l(3)84Cc', description='GST-containing FLYWCH zinc-finger protein', other_designations='CG33546-PB|CG33546-PD|CG33546-PE|gfzf-PB|gfzf-PD|gfzf-PE|lethal (3) 84Cc|suppressor of killer of prune', map_location='84C6-84D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7145880, end_position_on_the_genomic_accession=7150968, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40859, CurrentID=0, Status='live', Symbol='Sas-4', Aliases='Dmel_CG10061, CG10061, CPAP, D-Sas4, D-sas4, DSAS-4, DSas-4, DSas4, Dmel\\CG10061, Dsas-4, Dsas4, SAS-4, SAS4, Sas4, d-SAS-4, dsas-4, l(3)S2214, l(3)s2214, sas-4, sas4', description='Spindle assembly abnormal 4 ortholog (C. elegans)', other_designations='CG10061-PA|DSas-4|Sas-4-PA|drosophila Sas-4|lethal (3) s2214|spindle assembly abnormal 4 ortholog', map_location='84C6-84C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7151367, end_position_on_the_genomic_accession=7154219, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40860, CurrentID=0, Status='live', Symbol='MAGE', Aliases='Dmel_CG10059, CG10059, D Dmel\\CG10059, Mage, Nse3, sst', description='CG10059 gene product from transcript CG10059-RA', other_designations='CG10059-PA|MAGE-PA|necdin|sleepless in seattle', map_location='84C7-84C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7154238, end_position_on_the_genomic_accession=7155180, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40861, CurrentID=0, Status='live', Symbol='sas', Aliases='Dmel_CG2507, CG2507, Dmel\\CG2507, SAS, Sas, l(3)84Cd, rfd', description='stranded at second', other_designations='CG2507-PA|CG2507-PB|CG2507-PC|CG2507-PD|responsible for dominant phenotype|sas-PA|sas-PB|sas-PC|sas-PD|stranded-at-second', map_location='84C8-84D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7162808, end_position_on_the_genomic_accession=7183702, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  40862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40862, CurrentID=0, Status='live', Symbol='CG10055', Aliases='Dmel_ Dmel\\CG10055, RE52393', description='CG10055 gene product from transcript CG10055-RA', other_designations='CG10055-PA|uncharacterized protein', map_location='84D1-84D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7184101, end_position_on_the_genomic_accession=7186673, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40863, CurrentID=0, Status='live', Symbol='lap', Aliases='Dmel_CG2520, AP180, BcDNA:GM10787, CG2520, Dmel\\CG2520, EP(3)3060, EP3060, LAP, Lap, dAP-3, l(3)neo34/AP180, sig, lap', description='like-AP180', other_designations='CG2520-PA|CG2520-PC|CG2520-PD|CG2520-PE|CG2520-PF|CG2520-PG|CG2520-PH|CG2520-PI|CG2520-PJ|EP3060|Like-AP180|complementation group 12|lap-PA|lap-PC|lap-PD|lap-PE|lap-PF|lap-PG|lap-PH|lap-PI|lap-PJ|like AP180|saisoog', map_location='84D1-84D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7186994, end_position_on_the_genomic_accession=7200941, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  40864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40864, CurrentID=0, Status='live', Symbol='CG14609', Aliases='Dmel_ Dmel\\CG14609', description='CG14609 gene product from transcript CG14609-RB', other_designations='CG14609-PA|CG14609-PB|uncharacterized protein', map_location='84D2-84D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7200970, end_position_on_the_genomic_accession=7203450, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  40865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40865, CurrentID=0, Status='live', Symbol='CG1288', Aliases='Dmel_ BEST:LP10010, BcDNA:AT26093, Dmel\\CG1288, anon-SAGE:Wang-119', description='CG1288 gene product from transcript CG1288-RA', other_designations='CG1288-PA|CG1288-PB|anon-SAGE:Wang-119', map_location='84D2-84D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7206652, end_position_on_the_genomic_accession=7208431, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40866, CurrentID=0, Status='live', Symbol='CG14606', Aliases='Dmel_ Dmel\\CG14606', description='CG14606 gene product from transcript CG14606-RB', other_designations='CG14606-PA|CG14606-PB|uncharacterized protein', map_location='84D2-84D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7214339, end_position_on_the_genomic_accession=7217083, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40867, CurrentID=0, Status='live', Symbol='CG14605', Aliases='Dmel_ BcDNA:AT16877, Dmel\\CG14605', description='CG14605 gene product from transcript CG14605-RE', other_designations='CG14605-PB|CG14605-PD|CG14605-PE|uncharacterized protein', map_location='84D2-84D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7217679, end_position_on_the_genomic_accession=7219705, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40868, CurrentID=0, Status='live', Symbol='CG14608', Aliases='Dmel_ Dmel\\CG14608', description='CG14608 gene product from transcript CG14608-RC', other_designations='CG14608-PC|uncharacterized protein', map_location='84D2-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7224414, end_position_on_the_genomic_accession=7241138, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40869, CurrentID=0, Status='live', Symbol='CG14607', Aliases='Dmel_ Dmel\\CG14607', description='CG14607 gene product from transcript CG14607-RB', other_designations='CG14607-PB|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7241760, end_position_on_the_genomic_accession=7244006, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40870, CurrentID=0, Status='live', Symbol='CG1234', Aliases='Dmel_ Dmel\\CG1234', description='CG1234 gene product from transcript CG1234-RA', other_designations='CG1234-PA', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7244232, end_position_on_the_genomic_accession=7246969, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40871, CurrentID=0, Status='live', Symbol='CG10053', Aliases='Dmel_ BEST:LD41878, BEST:LD45537, Dmel\\CG10053', description='CG10053 gene product from transcript CG10053-RB', other_designations='CG10053-PA|CG10053-PB|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7247090, end_position_on_the_genomic_accession=7248140, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40872, CurrentID=0, Status='live', Symbol='CG1227', Aliases='Dmel_ Dmel\\CG1227, MPSK/PSK, Psk', description='CG1227 gene product from transcript CG1227-RA', other_designations='CG1227-PA|CG1227-PB', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7248151, end_position_on_the_genomic_accession=7249738, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40873, CurrentID=0, Status='live', Symbol='CG10050', Aliases='Dmel_ Dmel\\CG10050', description='CG10050 gene product from transcript CG10050-RA', other_designations='CG10050-PA|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7249795, end_position_on_the_genomic_accession=7251750, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40874, CurrentID=0, Status='live', Symbol='Obp84a', Aliases='Dmel_CG1176, 84a, CG1176, DmelObp84a, Dmel\\CG1176, PBP4, PBPRP-4, PBPRP4, Pbprp4, pbprp-4, pbprp4', description='Odorant-binding protein 84a', other_designations='CG1176-PA|CG1176-PC|Obp84a-PA|Obp84a-PC|Pbprp4|Pheromone-binding protein-related protein 4', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7250725, end_position_on_the_genomic_accession=7251957, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40875, CurrentID=0, Status='live', Symbol='Gld', Aliases='Dmel_CG1152, CG1152, Dmel\\CG1152, GDH, GLD, Go, gld, glucose dehydrogenase, l(3)84Cg', description='Glucose dehydrogenase', other_designations='CG1152-PA|Gld-PA|glucose dehydrogenase|glucose oxidase', map_location='3-48 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7252063, end_position_on_the_genomic_accession=7265703, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40876, CurrentID=0, Status='live', Symbol='Syt4', Aliases='Dmel_CG10047, CG10047, D.Syt IV, Dmel\\CG10047, SYT4, Syt IV, SytIV, dsyt4, syt 4, syt4, sytIV', description='Synaptotagmin 4', other_designations='CG10047-PA|Syt4-PA|synaptotagmin 4|synaptotagmin IV', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7266065, end_position_on_the_genomic_accession=7270809, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40877, CurrentID=0, Status='live', Symbol='CG10032', Aliases='Dmel_ Dmel\\CG10032', description='CG10032 gene product from transcript CG10032-RA', other_designations='CG10032-PA|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7271592, end_position_on_the_genomic_accession=7272509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40879, CurrentID=0, Status='live', Symbol='rn', Aliases='Dmel_CG42277, 5.3, CG10040, CG14600, CG14601, CG14603, CG32466, CG32467, CG42277, Dipr, Dmel\\CG42277, Dmel_CG32466, Dmel_CG32467, P5.3, Rn, roe, tod, tot', description='rotund', other_designations='CG42277-PC|CG42277-PE|CG42277-PF|CG42277-PG|Distal-into-proximal|T-one deficient|rn-PC|rn-PE|rn-PF|rn-PG|roughened eye|t-one transformed', map_location='3-47.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7272617, end_position_on_the_genomic_accession=7309629, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  40883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40883, CurrentID=0, Status='live', Symbol='CG10029', Aliases='Dmel_ Dmel\\CG10029', description='CG10029 gene product from transcript CG10029-RA', other_designations='CG10029-PA|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7315724, end_position_on_the_genomic_accession=7317207, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40884, CurrentID=0, Status='live', Symbol='nxf4', Aliases='Dmel_CG31501, BcDNA:AT07692, CG14604, CG31501, Dmel\\CG31501', description='nuclear export factor 4', other_designations='CG31501-PA|nxf4-PA', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7346730, end_position_on_the_genomic_accession=7347836, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40885, CurrentID=0, Status='live', Symbol='CG17944', Aliases='Dmel_ Dmel\\CG17944', description='CG17944 gene product from transcript CG17944-RB', other_designations='CG17944-PB|uncharacterized protein', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7361368, end_position_on_the_genomic_accession=7362123, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40886, CurrentID=0, Status='live', Symbol='Mst84Dd', Aliases='Dmel_CG17935, CG17935, Dmel\\CG17935, Mst, Mst ORF, Mst84, Mst87D, Mstprox, mst(3)gl-4', description='Male-specific RNA 84Dd', other_designations='CG17935-PA|Mst84Dd-PA', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7363307, end_position_on_the_genomic_accession=7363773, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40887, CurrentID=0, Status='live', Symbol='Mst84Dc', Aliases='Dmel_CG17945, CG17945, Dmel\\CG17945, Mst, Mst ORF, Mst84, Mst84D, Mst87D, Mstprox, mst(3)gl-3', description='Male-specific RNA 84Dc', other_designations='CG17945-PA|CG17945-PB|Mst84Dc-PA|Mst84Dc-PB', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7364254, end_position_on_the_genomic_accession=7364673, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40888, CurrentID=0, Status='live', Symbol='Mst84Db', Aliases='Dmel_CG17934, BcDNA:AT08471, CG17934, Dmel\\CG17934, Mst, Mst ORF, Mst84, Mst87D, Mstprox, NEST:bs22c06, bs22c06.y1, mst(3)gl-2', description='Male-specific RNA 84Db', other_designations='CG17934-PA|Mst84Db-PA', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7365160, end_position_on_the_genomic_accession=7365627, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40889, CurrentID=0, Status='live', Symbol='Mst84Da', Aliases='Dmel_CG17946, CG17946, Dmel\\CG17946, Mst, Mst ORF, Mst84, Mst87D, Mstprox, mst(3)gl-1', description='Male-specific RNA 84Da', other_designations='CG17946-PA|Mst84Da-PA', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7365850, end_position_on_the_genomic_accession=7366262, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40890, CurrentID=0, Status='live', Symbol='MstProx', Aliases='Dmel_CG1149, CG1149, Dm Dmel\\CG1149, Tl-3, Toll-3, dTLR3, dToll3, mstprox, toll', description='CG1149 gene product from transcript CG1149-RA', other_designations='CG1149-PA|MstProx-PA|Toll-3', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7365939, end_position_on_the_genomic_accession=7369305, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40891, CurrentID=0, Status='live', Symbol='Ir84a', Aliases='Dmel_CG10101, CG10101, CT28433, DmelIR84a, Dmel\\CG10101, IR84a', description='Ionotropic receptor 84a', other_designations='CG10101-PB|Ir84a-PB|ionotropic receptor 84a', map_location='84D6-84D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7406869, end_position_on_the_genomic_accession=7409384, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40892, CurrentID=0, Status='live', Symbol='wa-cup', Aliases='Dmel_CG10113, CG10113, Dmel\\CG10113', description='walker cup', other_designations='CG10113-PB|wa-cup-PB|walker-cup', map_location='84D6-84D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7420765, end_position_on_the_genomic_accession=7423057, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40893, CurrentID=0, Status='live', Symbol='CG14598', Aliases='Dmel_ Dmel\\CG14598', description='CG14598 gene product from transcript CG14598-RA', other_designations='CG14598-PA|uncharacterized protein', map_location='84D6-84D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7439760, end_position_on_the_genomic_accession=7441976, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40894, CurrentID=0, Status='live', Symbol='CG1142', Aliases='Dmel_ Dmel\\CG1142', description='CG1142 gene product from transcript CG1142-RA', other_designations='CG1142-PA', map_location='84D8-84D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7472458, end_position_on_the_genomic_accession=7473781, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40896, CurrentID=0, Status='live', Symbol='alpha-Est10', Aliases='Dmel_CG1131, CG1131, DmalphaE10, Dmel\\CG1131, Esterase-10, K, aE10, alpha-Ests, alphaE10, anon-WO0172774.144', description='alpha-Esterase-10', other_designations='CG1131-PB|CG1131-PC|CG1131-PD|Esterase-10|alpha-Est10-PB|alpha-Est10-PC|alpha-Est10-PD|fragment K', map_location='84D8-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7486583, end_position_on_the_genomic_accession=7493845, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40897, CurrentID=0, Status='live', Symbol='alpha-Est9', Aliases='Dmel_CG1128, CG1128, Dm_3R:18691, DmalphaE9, Dmel\\CG1128, Est9, J, aE9, alphaE9', description='alpha-Esterase-9', other_designations='CG1128-PC|CG1128-PD|alpha-Est9-PC|alpha-Est9-PD|fragment J', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7493921, end_position_on_the_genomic_accession=7505098, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40898, CurrentID=0, Status='live', Symbol='CG1287', Aliases='Dmel_ BEST:GH03396, Dmel\\CG1287', description='CG1287 gene product from transcript CG1287-RA', other_designations='CG1287-PA', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7500951, end_position_on_the_genomic_accession=7502376, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40899, CurrentID=0, Status='live', Symbol='alpha-Est8', Aliases='Dmel_CG1121, CG1121, DmalphaE8, Dmel\\CG1121, Est8, I, aE8, alphaE8', description='alpha-Esterase-8', other_designations='CG1121-PA|alpha-Est8-PA|fragment I', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7509343, end_position_on_the_genomic_accession=7515670, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40900, CurrentID=0, Status='live', Symbol='Ubc84D', Aliases='Dmel_CG12799, CG12799, Dmel\\CG12799, UbcD84, ubc', description='Ubiquitin conjugating enzyme 84D', other_designations='CG12799-PA|Ubc84D-PA|ubiquitin conjugating enzyme|ubiquitin conjugating enzyme 84D', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7518820, end_position_on_the_genomic_accession=7519456, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40901, CurrentID=0, Status='live', Symbol='alpha-Est7', Aliases='Dmel_CG1112, Ali, CG1112, DmalphaE7, Dmel\\CG1112, EST23, Est-23, Est23, H, Mce, aE7, ali, alphaE7, alphaest-7', description='alpha-Esterase-7', other_designations='CG1112-PA|aliesterase|aliesteraseless|alpha-Est7-PA|carboxylesterase 23|esterase 23|fragment H|malathion carboxylesterase', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7519563, end_position_on_the_genomic_accession=7523753, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40902, CurrentID=0, Status='live', Symbol='alpha-Est6', Aliases='Dmel_CG1108, CG1108, DmalphaE6, Dmel\\CG1108, E, Esterase-6, F, G, aE6, alphaE6', description='alpha-Esterase-6', other_designations='CG1108-PA|Esterase-6|alpha-Est6-PA|fragment E|fragment F|fragment G', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7524567, end_position_on_the_genomic_accession=7528777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40903, CurrentID=0, Status='live', Symbol='Rtnl2', Aliases='Dmel_CG1279, CG1279, Dmel\\CG1279, Tml84D, anon-WO0170980.134, rtnl2, trop', description='CG1279 gene product from transcript CG1279-RC', other_designations='CG1279-PB|CG1279-PC|Rtnl2-PB|Rtnl2-PC|Tropomyosin-like-84D|tropomyosin-like', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7528864, end_position_on_the_genomic_accession=7530241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40904, CurrentID=0, Status='live', Symbol='alphaTub84D', Aliases='Dmel_CG2512, 3t, ALPHA 84D, CG2512, D.m.ALPHA-84D, DTA4, Dmel\\CG2512, T, Tub, aTub84D, alpha-Tub, alpha-Tub84D, alpha-tub, alpha-tub84D, alpha-tubulin, alpha3, alpha3t, alpha84D, alphaTUB, alphaTUB84D, alphaTub, alphaTub3, alphaTub84', description='alpha-Tubulin at 84D', other_designations='CG2512-PA|Tubulin-alpha3|alpha tubulin|alpha-Tubulin|alphaT|alphaTub84D-PA|alphaTubulin 84D|alphatubulin|tubulin|tubulin alpha-3 chain|tubulin alpha3|tubulin at 84D', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7530753, end_position_on_the_genomic_accession=7532732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40905, CurrentID=0, Status='live', Symbol='alpha-Est5', Aliases='Dmel_CG1089, CG1089, D, DmalphaE5, Dmel\\CG1089, EST-C, EST9, ESTC, Est-9, Est-C, Est-c, aE5, alphaE5', description='alpha-Esterase-5', other_designations='CG1089-PA|CG1089-PB|alpha-Est5-PA|alpha-Est5-PB|esterase C|esterase-C|fragment D', map_location='3-47.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7533496, end_position_on_the_genomic_accession=7536390, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40906, CurrentID=0, Status='live', Symbol='alpha-Est4', Aliases='Dmel_CG1082, C, CG1082, DmalphaE4, Dmel\\CG1082, FBgn0015572, aE4, alphaE4', description='alpha-Esterase-4', other_designations='CG1082-PA|alpha-Est4-PA|fragment C', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7536687, end_position_on_the_genomic_accession=7538615, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40907, CurrentID=0, Status='live', Symbol='alpha-Est3', Aliases='Dmel_CG1257, B, CG1257, DmalphaE3, Dmel\\CG1257, aE3, alpha-Ests, alphaE3', description='alpha-Esterase-3', other_designations='CG1257-PA|CG1257-PB|CG1257-PC|CG1257-PD|alpha-Est3-PA|alpha-Est3-PB|alpha-Est3-PC|alpha-Est3-PD|fragment B', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7538972, end_position_on_the_genomic_accession=7541631, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40908, CurrentID=0, Status='live', Symbol='alpha-Est2', Aliases='Dmel_CG2505, Alpha-Est2, B, CG2505, DmalphaE2, Dmel\\CG2505, aE2, alpha-EST2, alphaE2', description='alpha-Esterase-2', other_designations='CG2505-PA|CG2505-PB|alpha-Est2-PA|alpha-Est2-PB|alpha-esterase 2|fragment B', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7541810, end_position_on_the_genomic_accession=7545230, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40909, CurrentID=0, Status='live', Symbol='alpha-Est1', Aliases='Dmel_CG1031, A, Alpha-Est1, CG1031, DmalphaE1, Dmel\\CG1031, Est1, Q9VIC3_DROME, aE1, alpha est1, alphaE1', description='alpha-Esterase-1', other_designations='CG1031-PA|alpha-Est1-PA|fragment A', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7545733, end_position_on_the_genomic_accession=7549124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40912, CurrentID=0, Status='live', Symbol='Nlg3', Aliases='Dmel_CG34127, CG1062, CG14597, CG18291, CG32465, CG34127, DNLG3, DNlg3, Dmel\\CG34127, dnlg3', description='Neuroligin 3', other_designations='CG34127-PB|CG34127-PC|Nlg3-PB|Nlg3-PC|drosokinin receptor|neuroligin 3|neuroligin3', map_location='84D10-84D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7570281, end_position_on_the_genomic_accession=7639434, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40913, CurrentID=0, Status='live', Symbol='Nlg1', Aliases='Dmel_CG31146, CG12148, CG14596, CG31146, DNlg1, Dmel\\CG31146, anon-WO0118547.256, dnlg1, nlg1', description='Neuroligin 1', other_designations='CG31146-PD|CG31146-PE|Nlg1-PD|Nlg1-PE|neuroligin|neuroligin 1', map_location='84D11-84D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7658247, end_position_on_the_genomic_accession=7694874, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  40915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40915, CurrentID=0, Status='live', Symbol='CG2616', Aliases='Dmel_ Dmel\\CG2616', description='CG2616 gene product from transcript CG2616-RB', other_designations='CG2616-PA|CG2616-PB', map_location='84D13-84D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7722417, end_position_on_the_genomic_accession=7724715, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40916, CurrentID=0, Status='live', Symbol='pyd3', Aliases='Dmel_CG3027, CG3027, Dmel\\CG3027, betaAS, betaASase', description='CG3027 gene product from transcript CG3027-RA', other_designations='CG3027-PA|beta-Alanine synthase|pyd3-PA', map_location='84D14-84D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7747913, end_position_on_the_genomic_accession=7750079, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40917, CurrentID=0, Status='live', Symbol='CG10919', Aliases='Dmel_ Dmel\\CG10919', description='CG10919 gene product from transcript CG10919-RA', other_designations='CG10919-PA|uncharacterized protein', map_location='84D14-84D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7758122, end_position_on_the_genomic_accession=7759677, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40918, CurrentID=0, Status='live', Symbol='Sp7', Aliases='Dmel_CG3066, CG3066, Dmel\\CG3066, MP2, PAE, PAE1, SP7, Sp 7, anon-Ryu, proPO-AE, sp7', description='Serine protease 7', other_designations='CG3066-PA|CG3066-PE|CG3066-PF|Sp7-PA|Sp7-PE|Sp7-PF|melanization protease 2|prophenoloxidase-activating enzyme|serine protease 7', map_location='84D14-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7759786, end_position_on_the_genomic_accession=7763947, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40919, CurrentID=0, Status='live', Symbol='CG11286', Aliases='Dmel_ Dmel\\CG11286', description='CG11286 gene product from transcript CG11286-RA', other_designations='CG11286-PA|CG11286-PB|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7770860, end_position_on_the_genomic_accession=7771756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40920, CurrentID=0, Status='live', Symbol='CG10086', Aliases='Dmel_ Dmel\\CG10086', description='CG10086 gene product from transcript CG10086-RA', other_designations='CG10086-PA|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7790633, end_position_on_the_genomic_accession=7793483, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40921, CurrentID=0, Status='live', Symbol='CG2641', Aliases='Dmel_ Dmel\\CG2641', description='CG2641 gene product from transcript CG2641-RA', other_designations='CG2641-PA', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7794069, end_position_on_the_genomic_accession=7796599, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40922, CurrentID=0, Status='live', Symbol='CG3014', Aliases='Dmel_ Dmel\\CG3014', description='CG3014 gene product from transcript CG3014-RC', other_designations='CG3014-PA|CG3014-PC', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7799264, end_position_on_the_genomic_accession=7802793, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40923, CurrentID=0, Status='live', Symbol='CG18268', Aliases='Dmel_ Dmel\\CG18268', description='CG18268 gene product from transcript CG18268-RA', other_designations='CG18268-PA|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7802831, end_position_on_the_genomic_accession=7804432, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40924, CurrentID=0, Status='live', Symbol='CG2993', Aliases='Dmel_ Dmel\\CG2993', description='CG2993 gene product from transcript CG2993-RB', other_designations='CG2993-PA|CG2993-PB|CG2993-PC', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7804665, end_position_on_the_genomic_accession=7810014, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40925, CurrentID=0, Status='live', Symbol='sgll', Aliases='Dmel_CG31472, CG2649, CG31472, Dmel\\CG31472, dPNPO', description='sugarlethal', other_designations="CG31472-PA|CG31472-PB|pyridoxine 5'-phosphate oxidase|sgll-PA|sgll-PB", map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7810532, end_position_on_the_genomic_accession=7811912, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40928, CurrentID=0, Status='live', Symbol='CG17816', Aliases='Dmel_ Dmel\\CG17816', description='CG17816 gene product from transcript CG17816-RE', other_designations='CG17816-PD|CG17816-PE|CG17816-PF|CG17816-PG|CG17816-PH|CG17816-PI|CG17816-PJ|CG17816-PK|uncharacterized protein', map_location='84E3-84E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7855582, end_position_on_the_genomic_accession=7886733, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  40929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40929, CurrentID=0, Status='live', Symbol='CG10092', Aliases='Dmel_ Dmel\\CG10092, atl', description='CG10092 gene product from transcript CG10092-RB', other_designations='CG10092-PA|CG10092-PB|arginyl tRNA ligase|uncharacterized protein', map_location='84E4-84E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7887045, end_position_on_the_genomic_accession=7890150, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40931, CurrentID=0, Status='live', Symbol='wtrw', Aliases='Dmel_CG31284, 153412_at, CG10409, CG17817, CG31284, CT29234, Dmel\\CG31284, Wtrw, dTRPA3, trpA3', description='water witch', other_designations='CG31284-PA|CG31284-PB|CG31284-PC|CG31284-PD|waterwitch|wtrw-PA|wtrw-PB|wtrw-PC|wtrw-PD', map_location='84E4-84E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7891331, end_position_on_the_genomic_accession=7901147, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40932, CurrentID=0, Status='live', Symbol='mRpS9', Aliases='Dmel_CG2957, CG2957, Dmel\\CG2957, MRP-S9', description='mitochondrial ribosomal protein S9', other_designations='CG2957-PA|mRpS9-PA', map_location='84E4-84E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7889277, end_position_on_the_genomic_accession=7902667, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40933, CurrentID=0, Status='live', Symbol='EMC1', Aliases='Dmel_CG2943, CG2943, Dmel\\CG2943, NEST:bs09f06', description='ER Membrane protein Complex 1', other_designations='CG2943-PA|CG2943-PB|CG2943-PC|CG2943-PD|EMC1-PA|EMC1-PB|EMC1-PC|EMC1-PD|ER membrane protein complex 1', map_location='84E4-84E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7903030, end_position_on_the_genomic_accession=7907363, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  40934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40934, CurrentID=0, Status='live', Symbol='Taf7', Aliases='Dmel_CG2670, CG2670, Dmel\\CG2670, E, TAF, TAF55, TAF7, TAF[[II]], TAF[[II]]55, TFIID, Taf55, anon-84Ec, dTAF[[II]]55', description='TBP-associated factor 7', other_designations='CG2670-PA|CG2670-PB|TBP-associated factor|TBP-associated factor 55kD|Taf7-PA|Taf7-PB', map_location='84E5-84E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7907333, end_position_on_the_genomic_accession=7909721, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40935, CurrentID=0, Status='live', Symbol='Prat', Aliases='Dmel_CG2867, CG2867, DmPrat, Dmel\\CG2867', description='Phosphoribosylamidotransferase', other_designations='CG2867-PA|CG2867-PB|Prat-PA|Prat-PB|phosphoribosylamidotransferase|transcript D', map_location='84E5-84E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7909598, end_position_on_the_genomic_accession=7911779, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40936, CurrentID=0, Status='live', Symbol='CG2846', Aliases='Dmel_ C, Dmel\\CG2846, anon-84Ea', description='CG2846 gene product from transcript CG2846-RA', other_designations='CG2846-PA|CG2846-PB|transcript C', map_location='84E5-84E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7912017, end_position_on_the_genomic_accession=7913290, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40937, CurrentID=0, Status='live', Symbol='CG2678', Aliases='Dmel_ B, Dmel\\CG2678, anon-84Eb', description='CG2678 gene product from transcript CG2678-RB', other_designations='CG2678-PA|CG2678-PB|transcript B', map_location='84E5-84E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7913851, end_position_on_the_genomic_accession=7916550, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40938, CurrentID=0, Status='live', Symbol='CG10445', Aliases='Dmel_ Dmel\\CG10445', description='CG10445 gene product from transcript CG10445-RD', other_designations='CG10445-PC|CG10445-PD|CG10445-PE|uncharacterized protein', map_location='84E5-84E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7916599, end_position_on_the_genomic_accession=7920299, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40940, CurrentID=0, Status='live', Symbol='dsx', Aliases='Dmel_CG11094, CG11094, DSX, Dmdsx, Dmel\\CG11094, Dsx, HrF, dsxM, ix-62c, dsx', description='doublesex', other_designations='CG11094-PA|CG11094-PB|CG11094-PC|CG11094-PD|CG11094-PE|CG11094-PF|double sex|dsx-PA|dsx-PB|dsx-PC|dsx-PD|dsx-PE|dsx-PF|hermaphrodite|intersex-62c', map_location='3-48.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7924323, end_position_on_the_genomic_accession=7967408, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  40941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40941, CurrentID=0, Status='live', Symbol='CD98hc', Aliases='Dmel_CG2791, CG2791, Dmel\\CG2791, RE39378p, cd98hc', description='CD98 heavy chain', other_designations='CD98hc-PA|CD98hc-PB|CD98hc-PD|CD98hc-PF|CG2791-PA|CG2791-PB|CG2791-PD|CG2791-PF', map_location='84E6-84E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7974409, end_position_on_the_genomic_accession=7978428, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  40942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40942, CurrentID=0, Status='live', Symbol='Os-C', Aliases='Dmel_CG3250, CG3250, Dmel\\CG3250, OS-C, Os-c', description='CG3250 gene product from transcript CG3250-RB', other_designations='CG3250-PB|CG3250-PC|CG3250-PD|Os-C-PB|Os-C-PC|Os-C-PD', map_location='84E6-84E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7978959, end_position_on_the_genomic_accession=7979614, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40943, CurrentID=0, Status='live', Symbol='CG2781', Aliases='Dmel_ Dmel\\CG2781', description='CG2781 gene product from transcript CG2781-RA', other_designations='CG2781-PA', map_location='84E6-84E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7981817, end_position_on_the_genomic_accession=7995852, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  40944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40944, CurrentID=0, Status='live', Symbol='CG14463', Aliases='Dmel_ Dmel\\CG14463', description='CG14463 gene product from transcript CG14463-RB', other_designations='CG14463-PA|CG14463-PB|uncharacterized protein', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7999832, end_position_on_the_genomic_accession=8001064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40945, CurrentID=0, Status='live', Symbol='CG2698', Aliases='Dmel_ Dmel\\CG2698', description='CG2698 gene product from transcript CG2698-RD', other_designations='CG2698-PA|CG2698-PC|CG2698-PD|CG2698-PE|CG2698-PG|CG2698-PH|CG2698-PI', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8001562, end_position_on_the_genomic_accession=8006033, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  40946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40946, CurrentID=0, Status='live', Symbol='CG2767', Aliases='Dmel_ Dmel\\CG2767', description='CG2767 gene product from transcript CG2767-RA', other_designations='Aldo/keto reductase|CG2767-PA', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8005924, end_position_on_the_genomic_accession=8008110, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  40947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40947, CurrentID=0, Status='live', Symbol='CG3223', Aliases='Dmel_ Dmel\\CG3223', description='CG3223 gene product from transcript CG3223-RA', other_designations='CG3223-PA', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8009400, end_position_on_the_genomic_accession=8011018, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40948, CurrentID=0, Status='live', Symbol='CG10435', Aliases='Dmel_ Dmel\\CG10435', description='CG10435 gene product from transcript CG10435-RA', other_designations='CG10435-PA|uncharacterized protein', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8010907, end_position_on_the_genomic_accession=8013201, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40949, CurrentID=0, Status='live', Symbol='Pbp95', Aliases='Dmel_CG2702, BEST:LD28511, CG2702, DmPBP, DmPBP95, DmSNAP190, DmSNAPc, Dmel\\CG2702, anon-WO0118547.359', description='PSEA-binding protein 95kD', other_designations='CG2702-PA|PSEA-binding protein|Pbp95-PA|drosophila melanogaster PSE-binding protein 95-kDa subunit', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8013791, end_position_on_the_genomic_accession=8016610, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40950, CurrentID=0, Status='live', Symbol='CG2747', Aliases='Dmel_ Dmel\\CG2747', description='CG2747 gene product from transcript CG2747-RD', other_designations='CG2747-PA|CG2747-PB|CG2747-PC|CG2747-PD|CG2747-PE|CG2747-PF', map_location='84E8-84E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8016210, end_position_on_the_genomic_accession=8030790, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  8673961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673961, CurrentID=0, Status='live', Symbol='CG42580', Aliases='Dmel_ Dmel\\CG42580, HDC19854', description='CG42580 gene product from transcript CG42580-RA', other_designations='CG42580-PA|uncharacterized protein', map_location='19D3-19D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20574282, end_position_on_the_genomic_accession=20574693, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40952, CurrentID=0, Status='live', Symbol='CG10903', Aliases='Dmel_ Dmel\\CG10903', description='CG10903 gene product from transcript CG10903-RB', other_designations='CG10903-PA|CG10903-PB|uncharacterized protein', map_location='84E9-84E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8030411, end_position_on_the_genomic_accession=8032145, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40953, CurrentID=0, Status='live', Symbol='CG11035', Aliases='Dmel_ Dmel\\CG11035, anon-WO0118547.426', description='CG11035 gene product from transcript CG11035-RB', other_designations='CG11035-PA|CG11035-PB|uncharacterized protein', map_location='84E9-84E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8031130, end_position_on_the_genomic_accession=8033009, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40955, CurrentID=0, Status='live', Symbol='mAChR-B', Aliases='Dmel_CG7918, CG7918, Dm-mAChR-B, Dmel\\CG7918, anon-WO0170980.91, anon-WO0170980.92, mAChR B, mAcR-CG7918', description='muscarinic Acetylcholine Receptor, B-type', other_designations='B-type mAChR|CG7918-PA|CG7918-PB|mAChR-B-PA|mAChR-B-PB|muscarinic acetylcholine Receptor, B-type', map_location='84E10-84E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8059284, end_position_on_the_genomic_accession=8087049, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40956, CurrentID=0, Status='live', Symbol='CG7910', Aliases='Dmel_ Dmel\\CG7910', description='CG7910 gene product from transcript CG7910-RA', other_designations='CG7910-PA|CG7910-PB', map_location='84E10-84E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8087252, end_position_on_the_genomic_accession=8089526, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  40957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40957, CurrentID=0, Status='live', Symbol='CG7900', Aliases='Dmel_ Dmel\\CG7900', description='CG7900 gene product from transcript CG7900-RB', other_designations='CG7900-PB', map_location='84E11-84E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8094124, end_position_on_the_genomic_accession=8105083, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40958, CurrentID=0, Status='live', Symbol='puc', Aliases='Dmel_CG7850, 0238/03, 1351/08, CG7850, Dmel\\CG7850, JNK-DSP, PUC, PUC/MKP, Puc, hrt, l(3)84Eh, l(3)A251.1, l(3)j4E1, vco', description='puckered', other_designations='CG7850-PA|CG7850-PB|Jun N-terminal kinase dual-specificity phosphatase|hearty|lethal (3) j4E1|puc-PA|puc-PB|velcro', map_location='3-47.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8105332, end_position_on_the_genomic_accession=8122294, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  40959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40959, CurrentID=0, Status='live', Symbol='CG7878', Aliases='Dmel_ DmRH26, Dmel\\CG7878, cg7878', description='CG7878 gene product from transcript CG7878-RA', other_designations='CG7878-PA', map_location='84E13-84E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8122725, end_position_on_the_genomic_accession=8125273, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40960, CurrentID=0, Status='live', Symbol='CG9667', Aliases='Dmel_ Dmel\\CG9667, ISY1', description='CG9667 gene product from transcript CG9667-RA', other_designations='CG9667-PA|CG9667-PB', map_location='84E13-84E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8125290, end_position_on_the_genomic_accession=8126357, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40961, CurrentID=0, Status='live', Symbol='Gie', Aliases='Dmel_CG7891, Arl8, BcDNA:LD29185, CG7891, Dmel\\CG7891, Q9VHV5', description='GTPase indispensable for equal segregation of chromosomes', other_designations='CG7891-PA|CG7891-PB|CG7891-PC|Gie-PA|Gie-PB|Gie-PC|novel GTPase indispensable for equal segregation of chromosomes', map_location='84E13-84E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8126605, end_position_on_the_genomic_accession=8129835, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40962, CurrentID=0, Status='live', Symbol='grn', Aliases='Dmel_CG9656, CG9656, CG9656-PA, Dmel\\CG9656, GATAc, GRAIN, Gata-c, Gata2, Grn, dGATA-2, dGATAc, dGatac, gra, l(3)84Fa', description='grain', other_designations='CG9656-PA|CG9656-PB|CG9656-PC|GATA-factor-c|grn-PA|grn-PB|grn-PC|lethal(3)84Fa', map_location='3-47 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8147368, end_position_on_the_genomic_accession=8181444, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  40963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40963, CurrentID=0, Status='live', Symbol='CG7800', Aliases='Dmel_ CT23295, Dmel\\CG7800', description='CG7800 gene product from transcript CG7800-RA', other_designations='CG7800-PA', map_location='84F4-84F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8217957, end_position_on_the_genomic_accession=8219734, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40964, CurrentID=0, Status='live', Symbol='CG18249', Aliases='Dmel_ CT41336, Dmel\\CG18249', description='CG18249 gene product from transcript CG18249-RB', other_designations='CG18249-PB|CG18249-PC|uncharacterized protein', map_location='84F4-84F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8220146, end_position_on_the_genomic_accession=8222188, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40965, CurrentID=0, Status='live', Symbol='DNApol-iota', Aliases='Dmel_CG7602, AAF54198, CG7602, DNA Pol II-iota, DNA-pol iota, DNApol-i, Dm-Poliota, Dmel\\CG7602, Poliota, drad30B', description='CG7602 gene product from transcript CG7602-RB', other_designations='CG7602-PA|CG7602-PB|CG7602-PC|DNApol-iota-PA|DNApol-iota-PB|DNApol-iota-PC', map_location='84F4-84F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8233507, end_position_on_the_genomic_accession=8237389, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40966, CurrentID=0, Status='live', Symbol='Ada2b', Aliases='Dmel_CG9638, ADA2b, Ada2B, Ada2S, BcDNA:LD24527, CG9638, Dmel\\CG9638, ada2b, dADA2b, dAda2B, dAda2S, dAda2b', description='transcriptional Adaptor 2b', other_designations='Ada2b-PA|Ada2b-PB|Ada2b-PC|CG9638-PA|CG9638-PB|CG9638-PC|transcriptional adapter 2S|transcriptional adaptor 2b', map_location='84F5-84F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8236632, end_position_on_the_genomic_accession=8240515, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40967, CurrentID=0, Status='live', Symbol='CG9636', Aliases='Dmel_ Dmel\\CG9636', description='CG9636 gene product from transcript CG9636-RC', other_designations='CG9636-PC|CG9636-PD', map_location='84F5-84F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8240747, end_position_on_the_genomic_accession=8243792, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40970, CurrentID=0, Status='live', Symbol='mtg', Aliases='Dmel_CG7549, 0849/10, CG7549, Dmel\\CG7549, Mtg, l(3)S084910, l(3)S084910b', description='mind the gap', other_designations='CG7549-PE|CG7549-PF|CG7549-PG|CG7549-PH|Mind-the-Gap|lethal (3) S084910b|mtg-PE|mtg-PF|mtg-PG|mtg-PH', map_location='84F11-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8250497, end_position_on_the_genomic_accession=8260718, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  40972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40972, CurrentID=0, Status='live', Symbol='RpA-70', Aliases='Dmel_CG9633, CG9633, D-RPA70, D-SSB, DRP-A, DmRPA, Dmel\\CG9633, RP-A, RPA, RPA 70, RPA 70 kDa, RPA1, RPA70, RpA1, RpA70, Rpa-70, Ssb-70, dRP-A, dmrpa1, i164', description='Replication Protein A 70', other_designations='CG9633-PA|RpA-70-PA|drosophila replication protein A|replication protein A 70|single stranded-binding protein 70', map_location='84F6-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8262120, end_position_on_the_genomic_accession=8264451, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  40973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40973, CurrentID=0, Status='live', Symbol='Mcm2', Aliases='Dmel_CG7538, CG7538, DmMCM2, DmMcm2, DmeMCM2, Dmel\\CG7538, MCM2, MCM2-7, MCM2_DROME, McM2, PCR3, dMCM2, dMcm2, l(3)rL074', description='Minichromosome maintenance 2', other_designations='CG7538-PA|Mcm2-PA|minichromosome maintenance 2', map_location='84F6-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8264970, end_position_on_the_genomic_accession=8267924, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40974, CurrentID=0, Status='live', Symbol='CG9630', Aliases='Dmel_ DmRH16, Dmel\\CG9630, cg9630', description='CG9630 gene product from transcript CG9630-RA', other_designations='CG9630-PA', map_location='84F6-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8267943, end_position_on_the_genomic_accession=8269896, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40975, CurrentID=0, Status='live', Symbol='ato', Aliases='Dmel_CG7508, ATO, At, Ato, CG7508, Dmel\\CG75081, ato[Dm], bHLHa10, ato', description='atonal', other_designations='CG7508-PA|ato-PA|atona', map_location='84F6-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8278052, end_position_on_the_genomic_accession=8279274, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40976, CurrentID=0, Status='live', Symbol='CG11671', Aliases='Dmel_ BcDNA:GM23845, Dmel\\CG11671', description='CG11671 gene product from transcript CG11671-RA', other_designations='CG11671-PA|uncharacterized protein', map_location='84F8-84F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8287876, end_position_on_the_genomic_accession=8288613, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40977, CurrentID=0, Status='live', Symbol='CG11672', Aliases='Dmel_ BcDNA:RE26525, Dmel\\CG11672', description='CG11672 gene product from transcript CG11672-RB', other_designations='CG11672-PA|CG11672-PB|uncharacterized protein', map_location='84F9-84F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8289395, end_position_on_the_genomic_accession=8290578, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40978, CurrentID=0, Status='live', Symbol='CG31463', Aliases='Dmel_ CG11673, Dmel\\CG31463', description='CG31463 gene product from transcript CG31463-RA', other_designations='CG31463-PA|uncharacterized protein', map_location='84F9-84F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8292335, end_position_on_the_genomic_accession=8293609, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40979, CurrentID=0, Status='live', Symbol='CG9626', Aliases='Dmel_ Dmel\\CG9626, Q9VHT6_DROME', description='CG9626 gene product from transcript CG9626-RC', other_designations='CG9626-PA|CG9626-PB|CG9626-PC|CG9626-PD', map_location='84F9-84F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8293843, end_position_on_the_genomic_accession=8300978, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  40980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40980, CurrentID=0, Status='live', Symbol='mRpL1', Aliases='Dmel_CG7494, CG7494, Dmel\\CG7494, L1', description='mitochondrial ribosomal protein L1', other_designations='CG7494-PA|mRpL1-PA', map_location='84F9-84F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8301440, end_position_on_the_genomic_accession=8302770, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40981, CurrentID=0, Status='live', Symbol='nac', Aliases='Dmel_CG9620, CG9620, Dmel\\CG9620, Gfr, Nac, gfr', description='neuronally altered carbohydrate', other_designations='CG9620-PA|Gfr|golgi GDP-fucose transporter|nac-PA|neuronally-altered-carbohydrate', map_location='84F10-84F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8302686, end_position_on_the_genomic_accession=8304023, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40982, CurrentID=0, Status='live', Symbol='Sgt1', Aliases='Dmel_CG9617, CG9617, DmSgt1, Dmel\\CG9617, Q9VHT3, l(3)s2383, sgt1', description='suppressor-of-G2-allele-of-skp1', other_designations='CG9617-PA|Sgt1-PA|lethal (3) s2383', map_location='84F10-84F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8304190, end_position_on_the_genomic_accession=8305274, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40983, CurrentID=0, Status='live', Symbol='tex', Aliases='Dmel_CG9615, CG9615, Dmel\\CG9615, TEX1', description='CG9615 gene product from transcript CG9615-RA', other_designations='CG9615-PA|tex-PA', map_location='84F10-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8307473, end_position_on_the_genomic_accession=8308645, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40984, CurrentID=0, Status='live', Symbol='CG11693', Aliases='Dmel_ Dmel\\CG11693', description='CG11693 gene product from transcript CG11693-RB', other_designations='CG11693-PA|CG11693-PB|uncharacterized protein', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8308881, end_position_on_the_genomic_accession=8310778, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40985, CurrentID=0, Status='live', Symbol='CG11694', Aliases='Dmel_ Dmel\\CG11694', description='CG11694 gene product from transcript CG11694-RA', other_designations='CG11694-PA|uncharacterized protein', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8311203, end_position_on_the_genomic_accession=8312137, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40986, CurrentID=0, Status='live', Symbol='CG11698', Aliases='Dmel_ Dmel\\CG11698', description='CG11698 gene product from transcript CG11698-RA', other_designations='CG11698-PA|uncharacterized protein', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8312188, end_position_on_the_genomic_accession=8313097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40987, CurrentID=0, Status='live', Symbol='eIF4AIII', Aliases='Dmel_CG7483, 7483, CG7483, DmRH28, Dmel\\CG7483, EIF4AIII, anon-WO0118547.364, eIF4A, eIF4A-III, eIF4A3, l(3)84Fi', description='CG7483 gene product from transcript CG7483-RA', other_designations='CG7483-PA|eIF4AIII-PA|eukaryotic translation initiation factor 4AIII|lethal(3)84Fi', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8313208, end_position_on_the_genomic_accession=8314864, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40988, CurrentID=0, Status='live', Symbol='Coq2', Aliases='Dmel_CG9613, CG9613, COQ2, Dmel\\CG9613, coq2, sbo', description='Coenzyme Q biosynthesis protein 2', other_designations='CG9613-PA|Coq2-PA|coenzyme Q biosynthesis protein 2|small boy', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8314784, end_position_on_the_genomic_accession=8316543, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  40989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40989, CurrentID=0, Status='live', Symbol='Ctr1B', Aliases='Dmel_CG7459, CG7459, Dmel\\CG7459, ctr1B', description='Copper transporter 1B', other_designations='CG7459-PA|Ctr1B-PA|copper transporter 1B', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8317612, end_position_on_the_genomic_accession=8319205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40990, CurrentID=0, Status='live', Symbol='Poxm', Aliases='Dmel_CG9610, CG9610, Dmel\\CG9610, P29, Pox-m, meso, poxm', description='Pox meso', other_designations='CG9610-PC|Pox-meso|Poxm-PC|paired box mesodermal|paired box-meso', map_location='84F11-84F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8319025, end_position_on_the_genomic_accession=8333792, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  40991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40991, CurrentID=0, Status='live', Symbol='Or85a', Aliases='Dmel_CG7454, 85A.4, 85a, CG7454, DOR31, DmOR85A, Dmel\\CG7454, or85a', description='Odorant receptor 85a', other_designations='CG7454-PA|Or85a-PA|odorant receptor 85a', map_location='84F11-84F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8321747, end_position_on_the_genomic_accession=8323054, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40992, CurrentID=0, Status='live', Symbol='CG7443', Aliases='Dmel_ Dmel\\CG7443', description='CG7443 gene product from transcript CG7443-RA', other_designations='CG7443-PA|CG7443-PB', map_location='84F12-84F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8324316, end_position_on_the_genomic_accession=8325036, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40993, CurrentID=0, Status='live', Symbol='Arl2', Aliases='Dmel_CG7435, ARL2, Arf, Arf84D, Arf84F, Arl84F, CG 7435, CG7435, Dm Arf84F, Dmel\\CG7435, Q06849, arl2', description='ADP ribosylation factor-like 2', other_designations='ADP ribosylation factor 84F|Arl2-PA|CG7435-PA|arflike-2', map_location='84F13-84F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8341100, end_position_on_the_genomic_accession=8341872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40994, CurrentID=0, Status='live', Symbol='CG9601', Aliases='Dmel_ 28573184, Dmel\\CG9601, PNKP', description='CG9601 gene product from transcript CG9601-RA', other_designations='CG9601-PA', map_location='84F13-84F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8341661, end_position_on_the_genomic_accession=8343516, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40995, CurrentID=0, Status='live', Symbol='COX7AL', Aliases='Dmel_CG18193, CG18193, COXJ, COXK, Dmel\\CG18193, VIIa', description='Cytochrome c oxidase subunit 7A-like', other_designations='CG18193-PA|COX7AL-PA|cytochrome c oxidase polypeptide VIIa|cytochrome c oxidase subunit 7A-like', map_location='84F13-84F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8343669, end_position_on_the_genomic_accession=8344310, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  40996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40996, CurrentID=0, Status='live', Symbol='DppIII', Aliases='Dmel_CG7415, CG7415, DPP, DPP III, Dmel\\CG7415', description='Dipeptidyl aminopeptidase III', other_designations='CG7415-PB|CG7415-PC|Dipeptidyl-peptidase III|DppIII-PB|DppIII-PC|dipeptidyl aminopeptidase|dipeptidyl aminopeptidase III', map_location='84F13-84F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8344400, end_position_on_the_genomic_accession=8347465, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  40997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40997, CurrentID=0, Status='live', Symbol='CG7352', Aliases='Dmel_ Dmel\\CG7352', description='CG7352 gene product from transcript CG7352-RC', other_designations='CG7352-PA|CG7352-PC', map_location='84F13-84F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8347579, end_position_on_the_genomic_accession=8349296, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40998, CurrentID=0, Status='live', Symbol='Atg13', Aliases='Dmel_CG7331, ATG13, BcDNA:HL01508, CG7331, Dmel\\CG7331, atg13', description='Autophagy-related 13', other_designations='Atg13-PA|Autophagy-specific gene 13|CG7331-PA', map_location='84F13-84F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8349815, end_position_on_the_genomic_accession=8352817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  40999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40999, CurrentID=0, Status='live', Symbol='stck', Aliases='Dmel_CG7954, CG7954, D-PINCH, Dmel\\CG7954, PINCH, Pinch, dPINCH, dpin, pin85A, pinch, stk', description='steamer duck', other_designations='CG7954-PA|CG7954-PB|CG7954-PC|CG7954-PD|PINCH|d-pinch|stck-PA|stck-PB|stck-PC|stck-PD|steamerduck', map_location='84F15-84F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8356978, end_position_on_the_genomic_accession=8359964, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41000, CurrentID=0, Status='live', Symbol='CG9684', Aliases='Dmel_ Dmel\\CG9684', description='CG9684 gene product from transcript CG9684-RA', other_designations='CG9684-PA', map_location='84F15-84F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8359958, end_position_on_the_genomic_accession=8362554, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41001, CurrentID=0, Status='live', Symbol='CG7963', Aliases='Dmel_ Dmel\\CG7963', description='CG7963 gene product from transcript CG7963-RB', other_designations='CG7963-PB', map_location='84F15-84F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8362826, end_position_on_the_genomic_accession=8363894, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41006, CurrentID=0, Status='live', Symbol='Obp85a', Aliases='Dmel_CG11732, CG11732, DmelObp85a, Dmel\\CG11732', description='Odorant-binding protein 85a', other_designations='CG11732-PA|Obp85a-PA', map_location='85A1-85A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8419822, end_position_on_the_genomic_accession=8420597, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41007, CurrentID=0, Status='live', Symbol='Or85b', Aliases='Dmel_CG11735, 85A.1, 85b, CG11735, DOR115, DmOR85b, Dmel\\CG11735', description='Odorant receptor 85b', other_designations='CG11735-PA|Or85b-PA|odorant receptor 85b', map_location='85A2-85A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8510077, end_position_on_the_genomic_accession=8511358, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41008, CurrentID=0, Status='live', Symbol='Or85c', Aliases='Dmel_CG17911, 85A.3, 85c, CG17911, DOR78, Dmel\\CG17911', description='Odorant receptor 85c', other_designations='CG17911-PA|Or85c-PA|odorant receptor 85c', map_location='85A2-85A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8511734, end_position_on_the_genomic_accession=8513022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41009, CurrentID=0, Status='live', Symbol='CG11737', Aliases='Dmel_ Dmel\\CG11737', description='CG11737 gene product from transcript CG11737-RA', other_designations='CG11737-PA|CG11737-PB|CG11737-PC|uncharacterized protein', map_location='85A2-85A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8517868, end_position_on_the_genomic_accession=8521378, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41010, CurrentID=0, Status='live', Symbol='CG31259', Aliases='Dmel_ CG11740, Dmel\\CG31259', description='CG31259 gene product from transcript CG31259-RB', other_designations='CG31259-PA|CG31259-PB|uncharacterized protein', map_location='85A2-85A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8523944, end_position_on_the_genomic_accession=8526667, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41011, CurrentID=0, Status='live', Symbol='Or85d', Aliases='Dmel_CG11742, 85.2, 85A.2, 85d, CG11742, DOR114, Dmel\\CG11742', description='Odorant receptor 85d', other_designations='CG11742-PA|Or85d-PA|odorant receptor 85d', map_location='85A3-85A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8552608, end_position_on_the_genomic_accession=8553968, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41013, CurrentID=0, Status='live', Symbol='pch2', Aliases='Dmel_CG31453, CG11743, CG11744, CG31453, Dmel\\CG31453', description='CG31453 gene product from transcript CG31453-RB', other_designations='CG31453-PA|CG31453-PB|pch2-PA|pch2-PB', map_location='85A3-85A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8559540, end_position_on_the_genomic_accession=8561764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41014, CurrentID=0, Status='live', Symbol='Cenp-C', Aliases='Dmel_CG31258, CENP-C, CG11745, CG11746, CG31258, CLD1, CenpC, Dmel\\CG31258, L1, cenp-C, l(2)SH3 157, l(3)85Aa, l(3)SH157', description='Centromeric protein-C', other_designations='CG31258-PA|CG31258-PB|Cenp-C-PA|Cenp-C-PB|centromeric protein C|centromeric protein-C|lethal (3) 85Aa|lethal (3) SH157', map_location='85A3-85A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8564105, end_position_on_the_genomic_accession=8569248, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41017, CurrentID=0, Status='live', Symbol='5-HT2B', Aliases='Dmel_CG42796, CG42796, CG7994, CG8007, Dmel\\CG42796, Dmel_CG7994, Dmel_CG8007, anon-WO0170980.94, anon-WO0170980.95, anon-WO0170980.97, anon-WO0170980.98', description='5-hydroxytryptamine (serotonin) receptor 2B', other_designations='5-HT2B-PD|5-HT2B-PE|5-HT2B-PF|CG42796-PD|CG42796-PE|CG42796-PF', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8578205, end_position_on_the_genomic_accession=8629996, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41018, CurrentID=0, Status='live', Symbol='Or85e', Aliases='Dmel_CG9700, 85B.1, 85e, CG9700, CR9700, DOR104, DOR85e, Dmel\\CG9700, Dor104, OR55, OR85e, Or104, dor104', description='Odorant receptor 85e', other_designations='CG9700-PB|Or85e-PB|odorant receptor 104|odorant receptor 85e', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8634320, end_position_on_the_genomic_accession=8635878, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41019, CurrentID=0, Status='live', Symbol='Cyp313b1', Aliases='Dmel_CG9716, 313b1, CG9716, Dmel\\CG9716', description='CG9716 gene product from transcript CG9716-RB', other_designations='CG9716-PA|CG9716-PB|Cyp313b1-PA|Cyp313b1-PB', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8643070, end_position_on_the_genomic_accession=8645650, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41020, CurrentID=0, Status='live', Symbol='Mkk4', Aliases='Dmel_CG9738, 142758_at, CG9738, CT27508, D-MKK4, DMKK4, Dmel\\CG9738, JNKK2, MAPKK4, MKK4, Mpk4, SEK1/MKK4, dMKK4, dMkk4, hep, mkk4', description='MAP kinase kinase 4', other_designations='CG9738-PA|CG9738-PB|CG9738-PC|JNK kinase 2|MAPK kinase 4|Map kinase kinase 4|Mkk4-PA|Mkk4-PB|Mkk4-PC|drosophila MAP kinase kinase 4', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8646172, end_position_on_the_genomic_accession=8650796, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41021, CurrentID=0, Status='live', Symbol='SLIRP2', Aliases='Dmel_CG8021, CG8021, DmSLIRP2, Dmel\\CG8021, cg8021', description='SRA stem-loop interacting RNA binding protein 2', other_designations='CG8021-PA|SLIRP2-PA', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8651084, end_position_on_the_genomic_accession=8651493, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41022, CurrentID=0, Status='live', Symbol='Dhod', Aliases='Dmel_CG9741, CG9741, DHOD, DHODH, Dmel\\CG9741, dhod, l(3)s3512', description='Dihydroorotate dehydrogenase', other_designations='CG9741-PA|CG9741-PD|Dhod-PA|Dhod-PD|dihydroorotate dehydrogenase', map_location='3-48.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8651391, end_position_on_the_genomic_accession=8652916, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318133, CurrentID=0, Status='live', Symbol='CG32638', Aliases='Dmel_ Dmel\\CG32638', description='CG32638 gene product from transcript CG32638-RB', other_designations='CG32638-PB|uncharacterized protein', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13122876, end_position_on_the_genomic_accession=13128794, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41025, CurrentID=0, Status='live', Symbol='p', Aliases='Dmel_CG9770, CG9770, Dmel\\CG9770, HPS5, NP_649810, dHPS5, snb', description='pink', other_designations='CG9770-PA|CG9770-PB|CG9770-PC|drosophila Hermansky-Pudlak syndrome 5 subunit|p-PA|p-PB|p-PC|pink-peach|sunburst', map_location='3-48.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8660903, end_position_on_the_genomic_accession=8670684, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41026, CurrentID=0, Status='live', Symbol='CG8032', Aliases='Dmel_ Dmel\\CG8032', description='CG8032 gene product from transcript CG8032-RB', other_designations='CG8032-PA|CG8032-PB', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8664115, end_position_on_the_genomic_accession=8667772, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41027, CurrentID=0, Status='live', Symbol='CG8036', Aliases='Dmel_ Dmel\\CG8036, anon-WO0118547.344', description='CG8036 gene product from transcript CG8036-RB', other_designations='CG8036-PB|CG8036-PC|CG8036-PD|transketolase', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8667868, end_position_on_the_genomic_accession=8673702, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41028, CurrentID=0, Status='live', Symbol='mRpL19', Aliases='Dmel_CG8039, CG8039, Dmel\\CG8039, L19', description='mitochondrial ribosomal protein L19', other_designations='CG8039-PA|mRpL19-PA', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8674109, end_position_on_the_genomic_accession=8675196, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41029, CurrentID=0, Status='live', Symbol='CG9773', Aliases='Dmel_ Dmel\\CG9773', description='CG9773 gene product from transcript CG9773-RA', other_designations='CG9773-PA', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8675283, end_position_on_the_genomic_accession=8676613, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41030, CurrentID=0, Status='live', Symbol='CG8043', Aliases='Dmel_ Dmel\\CG8043, Iba57', description='CG8043 gene product from transcript CG8043-RA', other_designations='CG8043-PA', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8676915, end_position_on_the_genomic_accession=8678233, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41031, CurrentID=0, Status='live', Symbol='CG11755', Aliases='Dmel_ Dmel\\CG11755', description='CG11755 gene product from transcript CG11755-RB', other_designations='CG11755-PA|CG11755-PB|uncharacterized protein', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8676915, end_position_on_the_genomic_accession=8678955, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41032, CurrentID=0, Status='live', Symbol='hb', Aliases='Dmel_CG9786, CG9786, Dmel\\CG9786, HB, Hb, Hunchback, R-pbx, Rg-bx, Rg-pbxHLH, l(3)85Ah, hb', description='hunchback', other_designations='CG9786-PA|CG9786-PB|Regulator-of-postbithorax|hb-PA|hb-PB', map_location='3-48 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8690980, end_position_on_the_genomic_accession=8697822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41033, CurrentID=0, Status='live', Symbol='CG8112', Aliases='Dmel_ Dmel\\CG8112', description='CG8112 gene product from transcript CG8112-RD', other_designations='CG8112-PB|CG8112-PC|CG8112-PD', map_location='85A8-85A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8712933, end_position_on_the_genomic_accession=8715263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41034, CurrentID=0, Status='live', Symbol='Cks85A', Aliases='Dmel_CG9790, BcDNA:GM01344, CG9790, Cks, Dmel\\CG9790, cks85A', description='Cyclin-dependent kinase subunit 85A', other_designations='CG9790-PA|CG9790-PB|CG9790-PC|Cks85A-PA|Cks85A-PB|Cks85A-PC', map_location='85A8-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8712906, end_position_on_the_genomic_accession=8718456, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41035, CurrentID=0, Status='live', Symbol='CG8116', Aliases='Dmel_ Dmel\\CG8116, EP(3)3246, GH04517', description='CG8116 gene product from transcript CG8116-RB', other_designations='CG8116-PA|CG8116-PB|CG8116-PC', map_location='85A8-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8716243, end_position_on_the_genomic_accession=8721320, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41036, CurrentID=0, Status='live', Symbol='CG11760', Aliases='Dmel_ BcDNA:LD16427, Dmel\\CG11760', description='CG11760 gene product from transcript CG11760-RB', other_designations='CG11760-PB|uncharacterized protein', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8719198, end_position_on_the_genomic_accession=8721320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41037, CurrentID=0, Status='live', Symbol='CG8136', Aliases='Dmel_ Dmel\\CG8136, anon-WO0140519.254, anon-WO0140519.56', description='CG8136 gene product from transcript CG8136-RB', other_designations='CG8136-PA|CG8136-PB|CG8136-PC|CG8136-PD', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8721371, end_position_on_the_genomic_accession=8723051, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41038, CurrentID=0, Status='live', Symbol='nom', Aliases='Dmel_CG8145, CG8145, Dmel\\CG8145', description='numerous disordered muscles', other_designations='CG8145-PA|CG8145-PB|CG8145-PC|nom-PA|nom-PB|nom-PC', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8723174, end_position_on_the_genomic_accession=8725357, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41039, CurrentID=0, Status='live', Symbol='CG11762', Aliases='Dmel_ Dmel\\CG11762, i36', description='CG11762 gene product from transcript CG11762-RA', other_designations='CG11762-PA|uncharacterized protein', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8725479, end_position_on_the_genomic_accession=8726730, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41040, CurrentID=0, Status='live', Symbol='CG8159', Aliases='Dmel_ Dmel\\CG8159', description='CG8159 gene product from transcript CG8159-RB', other_designations='CG8159-PB', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8726854, end_position_on_the_genomic_accession=8728370, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41041, CurrentID=0, Status='live', Symbol='ranshi', Aliases='Dmel_CG9793, CG9793, Dmel\\CG9793', description='CG9793 gene product from transcript CG9793-RA', other_designations='CG9793-PA|ranshi-PA', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8728367, end_position_on_the_genomic_accession=8729693, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41042, CurrentID=0, Status='live', Symbol='M1BP', Aliases='Dmel_CG9797, CG9797, Dmel\\CG9797', description='Motif 1 Binding Protein', other_designations='CG9797-PA|M1BP-PA|motif 1 binding protein', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8729790, end_position_on_the_genomic_accession=8731698, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41043, CurrentID=0, Status='live', Symbol='CG8202', Aliases='Dmel_ Dmel\\CG8202', description='CG8202 gene product from transcript CG8202-RA', other_designations='CG8202-PA', map_location='85A9-85A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8732091, end_position_on_the_genomic_accession=8735837, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  41044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41044, CurrentID=0, Status='live', Symbol='CG9801', Aliases='Dmel_ Dmel\\CG9801', description='CG9801 gene product from transcript CG9801-RC', other_designations='CG9801-PA|CG9801-PB|CG9801-PC', map_location='85A9-85A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8735879, end_position_on_the_genomic_accession=8747675, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41045, CurrentID=0, Status='live', Symbol='CG8223', Aliases='Dmel_ Dmel\\CG8223', description='CG8223 gene product from transcript CG8223-RB', other_designations='CG8223-PA|CG8223-PB', map_location='85A10-85A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8747798, end_position_on_the_genomic_accession=8754104, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41046, CurrentID=0, Status='live', Symbol='CG11768', Aliases='Dmel_ Dmel\\CG11768', description='CG11768 gene product from transcript CG11768-RC', other_designations='CG11768-PA|CG11768-PB|CG11768-PC|uncharacterized protein', map_location='85A10-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8749206, end_position_on_the_genomic_accession=8768200, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41047, CurrentID=0, Status='live', Symbol='CG8236', Aliases='Dmel_ Dmel\\CG8236', description='CG8236 gene product from transcript CG8236-RA', other_designations='CG8236-PA', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8769394, end_position_on_the_genomic_accession=8771150, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41048, CurrentID=0, Status='live', Symbol='CG13318', Aliases='Dmel_ Dmel\\CG13318, SPH101, c-SPH101', description='CG13318 gene product from transcript CG13318-RA', other_designations='CG13318-PA|uncharacterized protein', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8771246, end_position_on_the_genomic_accession=8772691, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41052, CurrentID=0, Status='live', Symbol='Ir85a', Aliases='Dmel_CG11775, CG11775, CT32618, DmelIR85a, Dmel\\CG11775, IR85a', description='Ionotropic receptor 85a', other_designations='CG11775-PA|Ir85a-PA|ionotropic receptor 85a', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8794684, end_position_on_the_genomic_accession=8796966, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41054, CurrentID=0, Status='live', Symbol='Tcp-1eta', Aliases='Dmel_CG8351, CG8351, Cct7, Dmel\\CG8351, Q9VHL2_DROME, Tcp-1-eta1, tcp-1eta, tcp1eta', description='CG8351 gene product from transcript CG8351-RA', other_designations='CG8351-PA|TCP1-zeta homologue|Tcp-1eta-PA', map_location='85B1-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8801564, end_position_on_the_genomic_accession=8806234, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41055, CurrentID=0, Status='live', Symbol='CG9839', Aliases='Dmel_ Dmel\\CG9839', description='CG9839 gene product from transcript CG9839-RB', other_designations='CG9839-PA|CG9839-PB', map_location='85B1-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8803471, end_position_on_the_genomic_accession=8805802, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41056, CurrentID=0, Status='live', Symbol='hng2', Aliases='Dmel_CG8359, CG8359, Dmel\\CG8359', description='hinge2', other_designations='CG8359-PA|hng2-PA', map_location='85B2-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8806322, end_position_on_the_genomic_accession=8807581, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41057, CurrentID=0, Status='live', Symbol='CG9837', Aliases='Dmel_ Dmel\\CG9837', description='CG9837 gene product from transcript CG9837-RF', other_designations='CG9837-PD|CG9837-PE|CG9837-PF', map_location='85B2-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8807679, end_position_on_the_genomic_accession=8813579, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41058, CurrentID=0, Status='live', Symbol='CG9821', Aliases='Dmel_ CG31442, Dmel\\CG9821, Ov29, anon-WO0118547.399, anon-WO0118547.424', description='CG9821 gene product from transcript CG9821-RC', other_designations='CG9821-PA|CG9821-PB|CG9821-PC|CG9821-PD', map_location='85B2-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8813490, end_position_on_the_genomic_accession=8820682, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41059, CurrentID=0, Status='live', Symbol='IscU', Aliases='Dmel_CG9836, CG9836, Dmel\\CG9836, dIscU, iscu', description='IscU homolog', other_designations='CG9836-PA|IscU homologue|IscU-PA', map_location='85B2-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8822849, end_position_on_the_genomic_accession=8824177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41060, CurrentID=0, Status='live', Symbol='CG8379', Aliases='Dmel_ Dmel\\CG8379, anon-WO0118547.250', description='CG8379 gene product from transcript CG8379-RB', other_designations='CG8379-PA|CG8379-PB', map_location='85B2-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8824411, end_position_on_the_genomic_accession=8827279, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41062, CurrentID=0, Status='live', Symbol='pyd', Aliases='Dmel_CG43140, CG11782, CG11962, CG12409, CG31349, CG43140, CG9729, CG9731, CG9763, CT36877, DZO-1, Dmel\\CG43140, Dmel_CG31349, Dmel_CG9731, Pch, Pyd, Pyd/ZO-1, Tamou, ZO-1, ZO1, dzo-1(Z01), tam, xvt, pyd', description='polychaetoid', other_designations='CG43140-PA|CG43140-PB|CG43140-PC|CG43140-PF|CG43140-PG|CG43140-PI|CG43140-PJ|CG43140-PK|CG43140-PL|CG43140-PM|CG43140-PO|extra verticals|failed fusions|pyd-PA|pyd-PB|pyd-PC|pyd-PF|pyd-PG|pyd-PI|pyd-PJ|pyd-PK|pyd-PL|pyd-PM|pyd-PO|tamou', map_location='3-48.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8827068, end_position_on_the_genomic_accession=8931896, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  318140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318140, CurrentID=0, Status='live', Symbol='CG32651', Aliases='Dmel_ Dmel\\CG32651', description='CG32651 gene product from transcript CG32651-RA', other_designations='CG32651-PA|uncharacterized protein', map_location='11A9-11A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12404616, end_position_on_the_genomic_accession=12408477, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41066, CurrentID=0, Status='live', Symbol='osk', Aliases='Dmel_CG10901, CG10901, Dm-osk, Dmel\\CG10901, OSK, Osk, Oskar, norka', description='oskar', other_designations='CG10901-PA|CG10901-PC|CG10901-PD|dmoskar|osk-PA|osk-PC|osk-PD', map_location='3-48.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8935126, end_position_on_the_genomic_accession=8938395, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41067, CurrentID=0, Status='live', Symbol='skap', Aliases='Dmel_CG11963, CG11963, CGI 1963, Dmel\\CG11963, SkAP, Skap', description='skpA associated protein', other_designations='CG11963-PA|CG11963-PB|CG11963-PC|CG11963-PD|CG11963-PG|CG11963-PH|CG11963-PI|CG11963-PJ|SkpA associated protein|skap-PA|skap-PB|skap-PC|skap-PD|skap-PG|skap-PH|skap-PI|skap-PJ', map_location='85B7-85B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8938724, end_position_on_the_genomic_accession=8943027, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  41068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41068, CurrentID=0, Status='live', Symbol='CG11964', Aliases='Dmel_ Dmel\\CG11964', description='CG11964 gene product from transcript CG11964-RA', other_designations='CG11964-PA|uncharacterized protein', map_location='85B7-85B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8942954, end_position_on_the_genomic_accession=8945020, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41069, CurrentID=0, Status='live', Symbol='CG11966', Aliases='Dmel_ Dmel\\CG11966', description='CG11966 gene product from transcript CG11966-RA', other_designations='CG11966-PA|CG11966-PB|uncharacterized protein', map_location='85B8-85B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8950246, end_position_on_the_genomic_accession=8963037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41070, CurrentID=0, Status='live', Symbol='CAHbeta', Aliases='Dmel_CG11967, CG11967, DmBCA, Dmel\\CG11967, beta-CA', description='Carbonic anhydrase beta', other_designations='CAHbeta-PA|CAHbeta-PB|CG11967-PA|CG11967-PB|GH10821P|beta-carbonic anhydrase|carbonic anhydrase beta', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8980860, end_position_on_the_genomic_accession=8982793, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41071, CurrentID=0, Status='live', Symbol='RagA-B', Aliases='Dmel_CG11968, CG11968, Dmel\\CG11968, Gtr1, Rag, RagA, dRagA, dRagB', description='Ras-related GTP binding A/B ortholog (H. sapiens)', other_designations='CG11968-PA|RagA-B-PA|Ras-related GTP binding A/B ortholog', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8982732, end_position_on_the_genomic_accession=8984315, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41072, CurrentID=0, Status='live', Symbol='CG11970', Aliases='Dmel_ BcDNA:LD23876, Dmel\\CG11970', description='CG11970 gene product from transcript CG11970-RB', other_designations='CG11970-PA|CG11970-PB|uncharacterized protein', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8984718, end_position_on_the_genomic_accession=8990649, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41073, CurrentID=0, Status='live', Symbol='E(var)3-9', Aliases='Dmel_CG11971, 1E9, Anon1E9, CG11971, Dmel\\CG11971, E-var(3)9, anon-EST:fe1E9, anon1E9', description='Enhancer of variegation 3-9', other_designations='CG11971-PA|E(var)3-9-PA|anon-fast-evolving-1E9|enhancer of variegation 3-9', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8990645, end_position_on_the_genomic_accession=8992949, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41074, CurrentID=0, Status='live', Symbol='CG11975', Aliases='Dmel_ Dmel\\CG11975', description='CG11975 gene product from transcript CG11975-RB', other_designations='CG11975-PA|CG11975-PB|PROPPIN-3|uncharacterized protein', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8993253, end_position_on_the_genomic_accession=8995212, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41075, CurrentID=0, Status='live', Symbol='CG31100', Aliases='Dmel_ CG11976, CG11978, Dmel\\CG31100', description='CG31100 gene product from transcript CG31100-RB', other_designations='CG31100-PA|CG31100-PB|uncharacterized protein', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8995229, end_position_on_the_genomic_accession=9003650, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41076, CurrentID=0, Status='live', Symbol='CG11977', Aliases='Dmel_ Dmel\\CG11977', description='CG11977 gene product from transcript CG11977-RB', other_designations='CG11977-PB|uncharacterized protein', map_location='85C1-85C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8997697, end_position_on_the_genomic_accession=8998678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41078, CurrentID=0, Status='live', Symbol='CG11980', Aliases='Dmel_ Dmel\\CG11980, anon-WO0118547.327', description='CG11980 gene product from transcript CG11980-RC', other_designations='CG11980-PA|CG11980-PB|CG11980-PC|uncharacterized protein', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9004017, end_position_on_the_genomic_accession=9005331, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41079, CurrentID=0, Status='live', Symbol='Prosbeta3', Aliases='Dmel_CG11981, 11981, Beta-3, CG11981, Dmel\\CG11981, Pros beta3, beta-3, beta3', description='Proteasome beta3 subunit', other_designations='20S proteasome beta3 subunit|20s proteasome|CG11981-PA|Prosbeta3-PA|beta3|proteasome beta3 subunit', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9005313, end_position_on_the_genomic_accession=9006226, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41080, CurrentID=0, Status='live', Symbol='CG11982', Aliases='Dmel_ Dmel\\CG11982', description='CG11982 gene product from transcript CG11982-RA', other_designations='CG11982-PA|uncharacterized protein', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9006456, end_position_on_the_genomic_accession=9008030, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41081, CurrentID=0, Status='live', Symbol='CG11983', Aliases='Dmel_ Dmel\\CG11983', description='CG11983 gene product from transcript CG11983-RA', other_designations='CG11983-PA|CG11983-PB|uncharacterized protein', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9008321, end_position_on_the_genomic_accession=9010274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41082, CurrentID=0, Status='live', Symbol='CG11984', Aliases='Dmel_ Dmel\\CG11984', description='CG11984 gene product from transcript CG11984-RD', other_designations='CG11984-PD|CG11984-PE|CG11984-PF|CG11984-PG|CG11984-PH|CG11984-PI|uncharacterized protein', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9010282, end_position_on_the_genomic_accession=9014909, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41083, CurrentID=0, Status='live', Symbol='CG11986', Aliases='Dmel_ Dmel\\CG11986', description='CG11986 gene product from transcript CG11986-RA', other_designations='CG11986-PA|uncharacterized protein', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9015446, end_position_on_the_genomic_accession=9016822, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41084, CurrentID=0, Status='live', Symbol='tgo', Aliases='Dmel_CG11987, ARNT, Arnt, CG11987, DARNT, Dmel\\CG11987, HIF-1, HIF-1beta, HIF1b, Hif1, TGO, Tango, Tgo, arnt, bHLHe4, dARNT, dHIF-1beta, dTgo, darnt, jay, prd7', description='tango', other_designations='CG11987-PA|CG11987-PB|PRD gene 7|jaywalker|paired-like 7|tgo-PA|tgo-PB', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9016773, end_position_on_the_genomic_accession=9020022, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41085, CurrentID=0, Status='live', Symbol='neur', Aliases='Dmel_CG11988, 0328/12, 0971/15, 1293/09, 1315/05, BEST:LD21494, CG11988, Dmel\\CG11988, Neur, l(3)j6B12, l(3)j9B8, l(3)j9B9, l(3)j9C8, l(3)neo37, neu', description='neuralized', other_designations='CG11988-PA|CG11988-PB|CG11988-PC|CG11988-PD|CG11988-PE|neur-PA|neur-PB|neur-PC|neur-PD|neur-PE|neuralised', map_location='3-48.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9020348, end_position_on_the_genomic_accession=9039471, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41086, CurrentID=0, Status='live', Symbol='hyx', Aliases='Dmel_CG11990, CG11990, Cdc73, Dmel\\CG11990, Hyx/CDC73, hyx', description='hyrax', other_designations='CG11990-PA|CG11990-PB|hyx-PA|hyx-PB|parafibromin', map_location='85C3-85C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9038760, end_position_on_the_genomic_accession=9042349, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41087, CurrentID=0, Status='live', Symbol='Rel', Aliases='Dmel_CG11992, CG11992, Dmel\\CG11992, NF-kappaB, NFkappaB, REL, RELI-p110, Rel/NF-kappaB, RelA, ird, ird4, l(3)neo36, rel, relish, Rel', description='Relish', other_designations='CG11992-PA|CG11992-PB|CG11992-PC|CG11992-PD|Rel-PA|Rel-PB|Rel-PC|Rel-PD|immune response deficient 4|nuclear factor NF-kappa-B p110 subunit|positive regulation of transcription from RNA polymerase II promoter; GO:0045944 | IMP|relish', map_location='85C3-85C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9044183, end_position_on_the_genomic_accession=9047977, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41089, CurrentID=0, Status='live', Symbol='Mst85C', Aliases='Dmel_CG11993, CG11993, Dmel\\CG11993', description='CG11993 gene product from transcript CG11993-RB', other_designations='CG11993-PA|CG11993-PB|Mst85C-PA|Mst85C-PB', map_location='85C3-85C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9048645, end_position_on_the_genomic_accession=9052231, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41090, CurrentID=0, Status='live', Symbol='Kdm2', Aliases='Dmel_CG11033, CG11033, Dmel\\CG11033, FBXL19, JHDM1, KDM2, dKDM2, dKdm2, dRAF1, dkdm2', description='Lysine (K)-specific demethylase 2', other_designations='CG11033-PB|CG11033-PC|CG11033-PD|CG11033-PE|JmjC domain-containing histone demethylase 1|Kdm2-PB|Kdm2-PC|Kdm2-PD|Kdm2-PE|dRING-associated factor 1|lysine (K)-specific demethylase 2', map_location='85C3-85C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9052495, end_position_on_the_genomic_accession=9063245, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41091, CurrentID=0, Status='live', Symbol='beag', Aliases='Dmel_CG18005, CG18005, Dmel\\CG18005', description='CG18005 gene product from transcript CG18005-RA', other_designations='CG18005-PA|beag-PA', map_location='85C4-85C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9063132, end_position_on_the_genomic_accession=9065078, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41092, CurrentID=0, Status='live', Symbol='Ada', Aliases='Dmel_CG11994, ADA, CG11994, DmADA, Dmel\\CG11994, DrosADA, ada, dADA', description='Adenosine deaminase', other_designations='Ada-PA|CG11994-PA|adenosine deaminase', map_location='85C4-85C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9065183, end_position_on_the_genomic_accession=9066303, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41093, CurrentID=0, Status='live', Symbol='CG11997', Aliases='Dmel_ Dmel\\CG11997', description='CG11997 gene product from transcript CG11997-RA', other_designations='CG11997-PA|uncharacterized protein', map_location='85C12-85C12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9177517, end_position_on_the_genomic_accession=9179098, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41094, CurrentID=0, Status='live', Symbol='pum', Aliases='Dmel_CG9755, CG9755, CG9763, Dmel\\CG9755, EP(3)0883, PUM, Pum, anon-WO0172774.19, bem, fs(3)02003, l(3)01688, ova, ovt, pkl', description='pumilio', other_designations='CG9755-PA|CG9755-PB|CG9755-PC|CG9755-PD|CG9755-PE|CG9755-PF|CG9755-PG|CG9755-PH|bemused|milord|ovarette|pum-PA|pum-PB|pum-PC|pum-PD|pum-PE|pum-PF|pum-PG|pum-PH|pumillio|pumuckel', map_location='3-48.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9066343, end_position_on_the_genomic_accession=9237682, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  41095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41095, CurrentID=0, Status='live', Symbol='D1', Aliases='Dmel_CG9745, CG9745, Dmel\\CG9745, EP(3)0473, HMGI, RE69409, dD1, dHMGx', description='D1 chromosomal protein', other_designations='CG9745-PA|CG9745-PC|CG9745-PD|D1-PA|D1-PC|D1-PD|chromosomal protein D1', map_location='85D1-85D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9238083, end_position_on_the_genomic_accession=9240654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41096, CurrentID=0, Status='live', Symbol='Vps15', Aliases='Dmel_CG9746, CG9746, Dmel\\CG9746, Ird1, dVps15, dVps15/p150, ird, ird1, vps15', description='Vacuolar protein sorting 15', other_designations='CG9746-PA|CG9746-PB|CG9746-PC|Vps15-PA|Vps15-PB|Vps15-PC|immune response deficient 1|immune response-deficient 1|vacuolar protein sorting 15', map_location='85D1-85D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9240757, end_position_on_the_genomic_accession=9245869, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41097, CurrentID=0, Status='live', Symbol='CG8420', Aliases='Dmel_ Dmel\\CG8420, anon-WO0140519.182', description='CG8420 gene product from transcript CG8420-RB', other_designations='CG8420-PA|CG8420-PB', map_location='85D2-85D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9246649, end_position_on_the_genomic_accession=9251431, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41098, CurrentID=0, Status='live', Symbol='St2', Aliases='Dmel_CG16733, CG16733, Dmel\\CG16733, dmST2, st2', description='Sulfotransferase 2', other_designations='CG16733-PA|CG16733-PB|St2-PA|St2-PB|sulfotransferase 2', map_location='85D4-85D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9251843, end_position_on_the_genomic_accession=9253514, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41100, CurrentID=0, Status='live', Symbol='CG16734', Aliases='Dmel_ Dmel\\CG16734', description='CG16734 gene product from transcript CG16734-RD', other_designations='CG16734-PD|uncharacterized protein', map_location='85D4-85D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9255052, end_position_on_the_genomic_accession=9256731, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41101, CurrentID=0, Status='live', Symbol='CG16736', Aliases='Dmel_ Dmel\\CG16736', description='CG16736 gene product from transcript CG16736-RA', other_designations='CG16736-PA|uncharacterized protein', map_location='85D4-85D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9256689, end_position_on_the_genomic_accession=9257716, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41102, CurrentID=0, Status='live', Symbol='Ibf2', Aliases='Dmel_CG9740, CG9740, Dmel\\CG9740', description='Insulator binding factor 2', other_designations='CG9740-PA|Ibf2-PA|insulator binding factor 2', map_location='85D4-85D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9257727, end_position_on_the_genomic_accession=9258501, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41103, CurrentID=0, Status='live', Symbol='Ibf1', Aliases='Dmel_CG8436, CG8436, Dmel\\CG8436', description='Insulator binding factor 1', other_designations='CG8436-PA|CG8436-PB|Ibf1-PA|Ibf1-PB|insulator binding factor 1', map_location='85D6-85D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9258964, end_position_on_the_genomic_accession=9260400, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41104, CurrentID=0, Status='live', Symbol='VhaM8.9', Aliases='Dmel_CG8444, ATP6AP2, CG8444, Dmel\\CG8444, VhaM8-9, VhaPRR, dPRR, vhaM8.9', description='Vacuolar H[+] ATPase M8.9 accessory subunit', other_designations='CG8444-PA|VhaM8.9-PA|vacuolar H[+] ATPase M8.9 accessory subunit', map_location='85D6-85D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9260638, end_position_on_the_genomic_accession=9262590, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41105, CurrentID=0, Status='live', Symbol='sage', Aliases='Dmel_CG12952, CG12952, Dmel\\CG12952, Sage, bHLHc7', description='salivary gland-expressed bHLH', other_designations='CG12952-PA|CG12952-PC|sage-PA|sage-PC', map_location='85D7-85D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9262654, end_position_on_the_genomic_accession=9264142, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38467, CurrentID=0, Status='live', Symbol='CG14984', Aliases='Dmel_ Dmel\\CG14984', description='CG14984 gene product from transcript CG14984-RC', other_designations='CG14984-PB|CG14984-PC|uncharacterized protein', map_location='64A1-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3951193, end_position_on_the_genomic_accession=3952550, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41107, CurrentID=0, Status='live', Symbol='Vps16A', Aliases='Dmel_CG8454, CG8454, DmVps16, Dmel\\CG8454, V16, Vps16, dVps16A, vps16A', description='Vacuolar protein sorting 16A', other_designations='CG8454-PA|Vps16|Vps16A-PA|vacuolar protein sorting 16A', map_location='85D7-85D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9267019, end_position_on_the_genomic_accession=9270224, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41108, CurrentID=0, Status='live', Symbol='HP1e', Aliases='Dmel_CG8120, CG8120, Dmel\\CG8120, HP1E', description='Heterochromatin Protein 1e', other_designations='CG8120-PA|HP1e-PA|heterochromatin protein 1e', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9321770, end_position_on_the_genomic_accession=9322361, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41109, CurrentID=0, Status='live', Symbol='CG8861', Aliases='Dmel_ Dmel\\CG8861', description='CG8861 gene product from transcript CG8861-RC', other_designations='CG8861-PB|CG8861-PC', map_location='85D8-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9293642, end_position_on_the_genomic_accession=9328511, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41110, CurrentID=0, Status='live', Symbol='CG12951', Aliases='Dmel_ Dmel\\CG12951, SP87', description='CG12951 gene product from transcript CG12951-RA', other_designations='CG12951-PA|uncharacterized protein', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9329223, end_position_on_the_genomic_accession=9330331, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41111, CurrentID=0, Status='live', Symbol='CG16749', Aliases='Dmel_ Dmel\\CG16749, SP91', description='CG16749 gene product from transcript CG16749-RA', other_designations='CG16749-PA|uncharacterized protein', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9330885, end_position_on_the_genomic_accession=9331885, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41112, CurrentID=0, Status='live', Symbol='CG8866', Aliases='Dmel_ Dmel\\CG8866, UNC 51-like, dULK3', description='CG8866 gene product from transcript CG8866-RA', other_designations='CG8866-PA|CG8866-PB', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9332064, end_position_on_the_genomic_accession=9334714, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41113, CurrentID=0, Status='live', Symbol='CG8121', Aliases='Dmel_ Dmel\\CG8121, dmCG8121', description='CG8121 gene product from transcript CG8121-RA', other_designations='CG8121-PA|CG8121-PB|CG8121-PC', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9335170, end_position_on_the_genomic_accession=9337459, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41114, CurrentID=0, Status='live', Symbol='CG45050', Aliases='Dmel_ BcDNA:RE70963, CG16750, CG16751, CG33187, CG33188, CG3393, CG33936, CG33937, CG43674, Dmel\\CG45050, Dmel_CG33937, Dmel_CG43674, anon-85Da, anon-EST:fe1B11, anon-EST:fe2G11, drn', description='CG45050 gene product from transcript CG45050-RF', other_designations='CG45050-PB|CG45050-PC|CG45050-PD|CG45050-PE|CG45050-PF|CG45050-PG|CG45050-PH|CG45050-PI|anon-fast-evolving-1B11|anon-fast-evolving-2G11|uncharacterized protein', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9340216, end_position_on_the_genomic_accession=9362994, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41116, CurrentID=0, Status='live', Symbol='CG18473', Aliases='Dmel_ CG8125, Dmel\\CG18473', description='CG18473 gene product from transcript CG18473-RA', other_designations='CG18473-PA|uncharacterized protein', map_location='85D11-85D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9363305, end_position_on_the_genomic_accession=9364838, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41117, CurrentID=0, Status='live', Symbol='CG8129', Aliases='Dmel_ CT24308, Dmel\\CG8129', description='CG8129 gene product from transcript CG8129-RB', other_designations='CG8129-PA|CG8129-PB|serine racemase', map_location='85D12-85D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9365398, end_position_on_the_genomic_accession=9367763, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41118, CurrentID=0, Status='live', Symbol='FER', Aliases='Dmel_CG8874, CG8874, DFER, DFer, Dfer, Dfps, Dm FPS, DmHD-179, Dmel\\CG8874, FPS, Fer, Fps85D, HD-179, dFer, dfer, dfps, dfps 85D, dfps85D, fer, fps, fps85D', description='FER ortholog (H. sapiens)', other_designations='CG8874-PA|CG8874-PB|CG8874-PC|CG8874-PD|CG8874-PE|CG8874-PF|FER ortholog|FER-PA|FER-PB|FER-PC|FER-PD|FER-PE|FER-PF|Fps oncogene analog', map_location='85D13-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9367603, end_position_on_the_genomic_accession=9394546, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  41119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41119, CurrentID=0, Status='live', Symbol='Or85f', Aliases='Dmel_CG16755, 85D.1, 85f, CG16755, DOR32, DmOr85f, Dmel\\CG16755', description='Odorant receptor 85f', other_designations='CG16755-PA|Or85f-PA|odorant receptor 85f', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9398190, end_position_on_the_genomic_accession=9399551, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41120, CurrentID=0, Status='live', Symbol='RpL34b', Aliases='Dmel_CG9354, CG9354, Dmel\\CG9354, L34b, Rp L34b, RpL34, anon-EST:fe1C10, anon-EST:fe1E8', description='Ribosomal protein L34b', other_designations='CG9354-PA|CG9354-PB|CG9354-PC|CG9354-PD|RpL34b-PA|RpL34b-PB|RpL34b-PC|RpL34b-PD|anon-fast-evolving-1C10|anon-fast-evolving-1E8|ribosomal protein L34b', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9400237, end_position_on_the_genomic_accession=9401226, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41121, CurrentID=0, Status='live', Symbol='CG8132', Aliases='Dmel_ Dmel\\CG8132', description='CG8132 gene product from transcript CG8132-RA', other_designations='CG8132-PA', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9401495, end_position_on_the_genomic_accession=9402618, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41122, CurrentID=0, Status='live', Symbol='CG9356', Aliases='Dmel_ Dmel\\CG9356', description='CG9356 gene product from transcript CG9356-RA', other_designations='CG9356-PA|CG9356-PB', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9402564, end_position_on_the_genomic_accession=9404269, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41123, CurrentID=0, Status='live', Symbol='CG8135', Aliases='Dmel_ Dmel\\CG8135', description='CG8135 gene product from transcript CG8135-RA', other_designations='CG8135-PA', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9404385, end_position_on_the_genomic_accession=9407857, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41124, CurrentID=0, Status='live', Symbol='betaTub85D', Aliases='Dmel_CG9359, 2t, B2t, BETA 85D, BETA2, CG9359, D.m.BETA-85D, DTB4, Dmbeta2, Dmel\\CG9359, Tub, beta(2)Tu, beta(2)Tub, beta-Tub85D, beta-tub, beta-tub85D, beta2, beta2-tub, beta2t, beta2tub, beta85D, beta;Tub85D, betaTub2, beta[[2]]-tubulin, betatub85D, ms(3)KK[D]', description='beta-Tubulin at 85D', other_designations='CG9359-PA|b2 tubulin|beta 2 tubulin|beta tubulin|beta(2)Tubulin|beta-Tubulin|beta-Tubulin 85D|beta-tubulin85D|beta2 tubulin|beta2-tubulin|beta2tubulin|betaTub85D-PA|betaTubulin85D|testis-specific beta-tubulin|tubulin', map_location='3-48.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9408383, end_position_on_the_genomic_accession=9410282, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41125, CurrentID=0, Status='live', Symbol='Task7', Aliases='Dmel_CG9361, CG9361, Dmel\\CG9361, K2P-TASK7, dTASK-7', description='TWIK-related acid-sensitive K[+] channel 7', other_designations='CG9361-PA|Task7-PA', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9410329, end_position_on_the_genomic_accession=9412293, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41126, CurrentID=0, Status='live', Symbol='alpha-Man-IIa', Aliases='Dmel_CG18802, CG18474, CG18802, CG8139, DM-GII.1, Dmel\\CG18802, GMII, GmII, MAN-2, Man, Man-II, alpha-Man-II, dGMII, dGMIIb', description='alpha-Mannosidase class II a', other_designations='CG18802-PA|CG18802-PB|alpha mannosidase II|alpha-Man-IIa-PA|alpha-Man-IIa-PB|alpha-mannosidase 2|alpha-mannosidase II|alpha-mannosidase-II|chain A, golgi Alpha-Mannosidase Ii|golgi Man II|golgi alpha-mannosidase|golgi alpha-mannosidase II|golgi mannosidase II|mannosidase II', map_location='85D15-85D16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9412548, end_position_on_the_genomic_accession=9417264, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  8673991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673991, CurrentID=0, Status='live', Symbol='CG42581', Aliases='Dmel_ Dmel\\CG42581', description='CG42581 gene product from transcript CG42581-RA', other_designations='CG42581-PA|uncharacterized protein', map_location='19B1-19B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20096162, end_position_on_the_genomic_accession=20096474, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41132, CurrentID=0, Status='live', Symbol='GstZ1', Aliases='Dmel_CG9362, CG9362, Dmel\\CG9362, GSTZ1, gstz1', description='Glutathione S transferase Z1', other_designations='CG9362-PA|GstZ1-PA|glutathione S transferase Z1', map_location='85D17-85D17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9455517, end_position_on_the_genomic_accession=9456546, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41133, CurrentID=0, Status='live', Symbol='GstZ2', Aliases='Dmel_CG9363, CG9363, Dmel\\CG9363, GSTZ2, cg9363, gstz2', description='Glutathione S transferase Z2', other_designations='CG9363-PA|CG9363-PB|CG9363-PC|GstZ2-PA|GstZ2-PB|GstZ2-PC|glutathione S transferase Z2', map_location='85D17-85D17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9456815, end_position_on_the_genomic_accession=9460652, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41134, CurrentID=0, Status='live', Symbol='CG16779', Aliases='Dmel_ Dmel\\CG16779', description='CG16779 gene product from transcript CG16779-RD', other_designations='CG16779-PA|CG16779-PB|CG16779-PC|CG16779-PD|CG16779-PE|uncharacterized protein', map_location='85D17-85D18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9461084, end_position_on_the_genomic_accession=9493677, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  41135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41135, CurrentID=0, Status='live', Symbol='CG8147', Aliases='Dmel_ 154821_at, BEST:LD38109, Dmel\\CG8147', description='CG8147 gene product from transcript CG8147-RA', other_designations='CG8147-PA', map_location='85D18-85D18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9494801, end_position_on_the_genomic_accession=9496723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41136, CurrentID=0, Status='live', Symbol='RhoL', Aliases='Dmel_CG9366, 9366, CG9366, D-RhoL, DRhoL, DmRHO-L, Dmel\\CG9366, Drac3, G, Mes1, Rac3, RhoG/Rac3, Rhol, anon-EST:CL1c2, rhoL, rhol, rol, RhoL', description='Rho-like', other_designations='CG9366-PA|CG9366-PB|Rho-Like|RhoL-PA|RhoL-PB', map_location='85D18-85D18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9497101, end_position_on_the_genomic_accession=9502581, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41137, CurrentID=0, Status='live', Symbol='CG8149', Aliases='Dmel_ Dmel\\CG8149', description='CG8149 gene product from transcript CG8149-RA', other_designations='CG8149-PA', map_location='85D18-85D18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9504656, end_position_on_the_genomic_accession=9505865, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41138, CurrentID=0, Status='live', Symbol='rump', Aliases='Dmel_CG9373, CG9373, Dmel\\CG9373, HnRNP-S5, Hrp59, S5, cg9373, hnRNP M, hrp59, p70, p75', description='rumpelstiltskin', other_designations='CG9373-PA|CG9373-PB|Hrp59|heterogeneous nuclear ribonucleoprotein M|heterogeneous nuclear ribonucleoprotein S5|p75|rump-PA|rump-PB', map_location='85D18-85D19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9505853, end_position_on_the_genomic_accession=9510324, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41139, CurrentID=0, Status='live', Symbol='Rlb1', Aliases='Dmel_CG8161, CG8161, Dmel\\CG8161, Ov30, Rlb', description='CG8161 gene product from transcript CG8161-RA', other_designations='CG8161-PA|Rlb1-PA', map_location='85D19-85D19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9513327, end_position_on_the_genomic_accession=9514827, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41140, CurrentID=0, Status='live', Symbol='Ras85D', Aliases='Dmel_CG9375, C-ras1, CG9375, D-Ras, D-Ras1, D-ras-1, D-ras1, DRAS1, DRas, DRas85D/Ras, Dm Ras1, Dmel\\CG9375, Dmras85D, Dras, Dras1, Dras85D, E(faf), E(sev)3C, EC3-3, EK3-4, P08646, RAS, RAS1, RAS85D, RTK, Ras, Ras 85D, Ras-1, Ras1, Ras1/RAs85D, Ras185D, Ras85Nt, RasI, RasV12, Ras[V12], S35097, Su(tor)3-2, dRAS1, dRas, dRas1, dRas85D, dras1, fs(3)05703, l(3)06677, l(3)s1747, p21[Ras1], ras, ras 1, ras1, ras85B, ras85D, Ras85D', description='Ras oncogene at 85D', other_designations='CG9375-PA|Ras GTPase|Ras oncogene at 85B|Ras1|Ras85D-PA|enhancer of faf', map_location='3-50 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9510561, end_position_on_the_genomic_accession=9513067, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41141, CurrentID=0, Status='live', Symbol='mRpL47', Aliases='Dmel_CG9378, BcDNA:AT29239, CG9378, Dmel\\CG9378, MRP-L47, Rlc1', description='mitochondrial ribosomal protein L47', other_designations='CG9378-PB|Rlc1|mRpL47-PB', map_location='85D19-85D19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9514822, end_position_on_the_genomic_accession=9515908, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41143, CurrentID=0, Status='live', Symbol='JHDM2', Aliases='Dmel_CG8165, CG31123, CG8165, Dmel\\CG8165', description='JmjC domain-containing histone demethylase 2', other_designations='CG8165-PA|CG8165-PB|JHDM2-PA|JHDM2-PB', map_location='85D19-85D19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9516007, end_position_on_the_genomic_accession=9520938, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41144, CurrentID=0, Status='live', Symbol='by', Aliases='Dmel_CG9379, 3R-B, BcDNA:RE65113, CG9379, CT26639, Dmel\\CG9379', description='blistery', other_designations='3R-B|CG9379-PA|CG9379-PC|CG9379-PD|by-PA|by-PC|by-PD|tensin', map_location='3-48.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9528238, end_position_on_the_genomic_accession=9533011, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41145, CurrentID=0, Status='live', Symbol='mura', Aliases='Dmel_CG9381, CG9381, CG9382, Dmel\\CG9381, Mura', description='murashka', other_designations='CG9381-PA|CG9381-PB|CG9381-PC|mura-PA|mura-PB|mura-PC', map_location='85D22-85D24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9535625, end_position_on_the_genomic_accession=9551806, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41147, CurrentID=0, Status='live', Symbol='RnpS1', Aliases='Dmel_CG16788, CG16788, DEK, Dmel\\CG16788, RNPS1, cg16788, rnps1', description='RNA-binding protein S1', other_designations='CG16788-PA|RRM-domain protein|RnpS1-PA', map_location='85D24-85D24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9552034, end_position_on_the_genomic_accession=9553747, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41148, CurrentID=0, Status='live', Symbol='CG9386', Aliases='Dmel_ Dmel\\CG9386', description='CG9386 gene product from transcript CG9386-RA', other_designations='CG9386-PA', map_location='85D24-85D24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9553696, end_position_on_the_genomic_accession=9555408, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41149, CurrentID=0, Status='live', Symbol='CG8199', Aliases='Dmel_ Dmel\\CG8199', description='CG8199 gene product from transcript CG8199-RA', other_designations='CG8199-PA', map_location='85D24-85D24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9555785, end_position_on_the_genomic_accession=9557776, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41150, CurrentID=0, Status='live', Symbol='AP-1mu', Aliases='Dmel_CG93881, AP-47, AP1mu1, AP47, APmu1, CG9388, Dmel\\CG9388, Dmu1, ap-47, apm1, cg9388, mu1, AP-1mu', description='Adaptor Protein complex 1, mu subunit', other_designations='AP-1mu-PA|AP-47|CG9388-PA|adaptor protein complex 1, mu subunit', map_location='85D24-85D24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9557712, end_position_on_the_genomic_accession=9559525, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41151, CurrentID=0, Status='live', Symbol='MBD-like', Aliases='Dmel_CG8208, CG8208, Dm Mbd2/3, DmMbd2/3, Dmel\\CG8208, MBD2, MBD2-3, MBD2/3, MBD3, dMBD-like, dMBD2/3, dMbD2/3', description='Methyl-CpG binding domain protein-like', other_designations='CG8208-PA|CG8208-PB|CG8208-PC|MBD-like-PA|MBD-like-PB|MBD-like-PC|methyl-CpG-binding-domain-like-protein', map_location='85D24-85D24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9559760, end_position_on_the_genomic_accession=9561533, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41152, CurrentID=0, Status='live', Symbol='CG9393', Aliases='Dmel_ Dmel\\CG9393', description='CG9393 gene product from transcript CG9393-RA', other_designations='CG9393-PA|CG9393-PB', map_location='85D24-85D25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9561583, end_position_on_the_genomic_accession=9563673, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41153, CurrentID=0, Status='live', Symbol='Vps45', Aliases='Dmel_CG8228, CG8228, DmVps45, Dmel\\CG8228, MENE (3R)-C, MENE(3R)-C, VPS45p, dVps45, Vps45', description='Vacuolar protein sorting 45', other_designations='CG8228-PA|Vps45-PA|vacuolar protein sorting 45', map_location='85D25-85E', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9563924, end_position_on_the_genomic_accession=9566092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41154, CurrentID=0, Status='live', Symbol='CG16789', Aliases='Dmel_ Dmel\\CG16789', description='CG16789 gene product from transcript CG16789-RA', other_designations='CG16789-PA|CG16789-PB|uncharacterized protein', map_location='85D25-85D25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9566195, end_position_on_the_genomic_accession=9569189, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41155, CurrentID=0, Status='live', Symbol='CG16790', Aliases='Dmel_ Dmel\\CG16790', description='CG16790 gene product from transcript CG16790-RB', other_designations='CG16790-PA|CG16790-PB|uncharacterized protein', map_location='85D25-85D25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9569074, end_position_on_the_genomic_accession=9570214, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41156, CurrentID=0, Status='live', Symbol='CG9396', Aliases='Dmel_ Dmel\\CG9396', description='CG9396 gene product from transcript CG9396-RA', other_designations='CG9396-PA', map_location='85D25-85D25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9570414, end_position_on_the_genomic_accession=9571623, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41157, CurrentID=0, Status='live', Symbol='CG9399', Aliases='Dmel_ Dmel\\CG9399', description='CG9399 gene product from transcript CG9399-RE', other_designations='CG9399-PA|CG9399-PB|CG9399-PD|CG9399-PE', map_location='85D25-85D25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9571911, end_position_on_the_genomic_accession=9574211, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41158, CurrentID=0, Status='live', Symbol='Kap-alpha3', Aliases='Dmel_CG9423, 0335/13, CG9423, Dalpha3, Dmel\\CG9423, IMPalpha3, Imp alpha-3, Imp-alpha3, KAP-ALPHA3, Kap alpha3, Kap-?3, Kapalpha3, Kpna3, alphaKap3, dKap-alpha3, dimpalpha3, imp alpha3, imp-a3, imp-alpha3, impalpha3, kpn-alpha3, l(3)S033513', description='karyopherin alpha3', other_designations='CG9423-PA|CG9423-PB|CG9423-PC|CG9423-PD|Importin-alpha3|Kap-alpha3-PA|Kap-alpha3-PB|Kap-alpha3-PC|Kap-alpha3-PD|dImportin-alpha3|importin a3|importin alpha|importin alpha-3|importin alpha3|karyopherin a3|karyopherin alpha4|karyopherin-alpha3|karyopherin-alpha4', map_location='85D25-85D25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9574484, end_position_on_the_genomic_accession=9577733, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41159, CurrentID=0, Status='live', Symbol='CG8273', Aliases='Dmel_ Dmel\\CG8273, cg8273', description='CG8273 gene product from transcript CG8273-RA', other_designations='CG8273-PA|CG8273-PB', map_location='85D25-85D26', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9578042, end_position_on_the_genomic_accession=9581225, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41160, CurrentID=0, Status='live', Symbol='CG8301', Aliases='Dmel_ Dmel\\CG8301', description='CG8301 gene product from transcript CG8301-RA', other_designations='CG8301-PA', map_location='85D26-85D27', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9581306, end_position_on_the_genomic_accession=9586648, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41161, CurrentID=0, Status='live', Symbol='P58IPK', Aliases='Dmel_CG8286, BEST:SD05682, CG8286, Dmel\\CG8286, SD05682.5prime, dIPK', description='CG8286 gene product from transcript CG8286-RA', other_designations='CG8286-PA|P58IPK-PA|inhibitor of protein kinases', map_location='85D27-85D27', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9586942, end_position_on_the_genomic_accession=9589581, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41162, CurrentID=0, Status='live', Symbol='bocks', Aliases='Dmel_CG9424, Bocks, Bocksbeutel, CG9424, Dmel\\CG9424, emerin', description='bocksbeutel', other_designations='CG9424-PA|CG9424-PB|bocks-PA|bocks-PB', map_location='85D27-85D27', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9589468, end_position_on_the_genomic_accession=9591002, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41163, CurrentID=0, Status='live', Symbol='CG8312', Aliases='Dmel_ Dmel\\CG8312', description='CG8312 gene product from transcript CG8312-RA', other_designations='CG8312-PA|CG8312-PB|CG8312-PC|CG8312-PD', map_location='85D27-85E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9591309, end_position_on_the_genomic_accession=9625073, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41164, CurrentID=0, Status='live', Symbol='CG9427', Aliases='Dmel_ Dmel\\CG9427', description='CG9427 gene product from transcript CG9427-RB', other_designations='CG9427-PA|CG9427-PB', map_location='85E1-85E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9625157, end_position_on_the_genomic_accession=9626930, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41165, CurrentID=0, Status='live', Symbol='CG8319', Aliases='Dmel_ Dmel\\CG8319', description='CG8319 gene product from transcript CG8319-RA', other_designations='CG8319-PA', map_location='85E1-85E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9627321, end_position_on_the_genomic_accession=9628676, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41166, CurrentID=0, Status='live', Symbol='Calr', Aliases='Dmel_CG9429, CG9429, Clr, Crc, Dmel\\CG9429, anon-EST:Liang-2.11, anon-EST:Liang-2.17, anon-EST:fe3D7, chr3R:5454619..5454778, clone 2.11, clone 2.17, crc, eth, potp', description='Calreticulin', other_designations='CG9429-PA|CG9429-PB|Calr-PA|Calr-PB|anon-fast-evolving-3D7|calreticulin|pot pourri|potpourri', map_location='85E1-85E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9628691, end_position_on_the_genomic_accession=9630851, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41167, CurrentID=0, Status='live', Symbol='SpdS', Aliases='Dmel_CG8327, CG8327, Dmel\\CG8327, Spds', description='Spermidine Synthase', other_designations='CG8327-PA|CG8327-PB|CG8327-PC|SpdS-PA|SpdS-PB|SpdS-PC|spermidine synthase', map_location='85E1-85E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9631641, end_position_on_the_genomic_accession=9633673, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41168, CurrentID=0, Status='live', Symbol='Scm', Aliases='Dmel_CG9495, CG9495, Dmel\\CG9495, SCM, Su(z)302, l(3)85Ef, scm', description='Sex comb on midleg', other_designations='CG9495-PA|CG9495-PB|Scm-PA|Scm-PB|Sex combs on midleg|Sex combs on midlegs|sex comb on midleg', map_location='3-48.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9633359, end_position_on_the_genomic_accession=9637611, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41169, CurrentID=0, Status='live', Symbol='Fst', Aliases='Dmel_CG9434, CG9434, Dmel\\CG9434, frost', description='Frost', other_designations='CG9434-PA|Fst-PA|frost', map_location='85E2-85E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9644978, end_position_on_the_genomic_accession=9646154, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41170, CurrentID=0, Status='live', Symbol='Dh44', Aliases='Dmel_CG8348, CG8348, CRF, DH, DH 44, DH44, DH[[44]], Dh, Dmel\\CG8348, Drm-DH, Drome-DH[[44]], drome-DH44', description='Diuretic hormone 44', other_designations='CG8348-PA|CG8348-PC|CRF-like diuretic hormone|CRF-like peptide|Dh44-PA|Dh44-PC|diuretic hormone|diuretic hormone 44', map_location='85E2-85E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9638182, end_position_on_the_genomic_accession=9650706, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41171, CurrentID=0, Status='live', Symbol='CG9492', Aliases='Dmel_ Dmel\\CG9492', description='CG9492 gene product from transcript CG9492-RF', other_designations='CG9492-PD|CG9492-PE|CG9492-PF|CG9492-PG', map_location='85E2-85E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9651179, end_position_on_the_genomic_accession=9674643, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  41172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41172, CurrentID=0, Status='live', Symbol='CG8358', Aliases='Dmel_ Dmel\\CG8358', description='CG8358 gene product from transcript CG8358-RA', other_designations='CG8358-PA', map_location='85E3-85E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9675523, end_position_on_the_genomic_accession=9677778, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41173, CurrentID=0, Status='live', Symbol='CG16817', Aliases='Dmel_ Dmel\\CG16817, YC17', description='CG16817 gene product from transcript CG16817-RA', other_designations='CG16817-PA|CG16817-PB|homolog to co-chaperone p23|uncharacterized protein', map_location='85E3-85E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9677927, end_position_on_the_genomic_accession=9679658, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41174, CurrentID=0, Status='live', Symbol='nmdyn-D7', Aliases='Dmel_CG8362, CG8362, Dmel\\CG8362', description='CG8362 gene product from transcript CG8362-RA', other_designations='CG8362-PA|nmdyn-D7-PA', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9679939, end_position_on_the_genomic_accession=9681502, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41175, CurrentID=0, Status='live', Symbol='CG9444', Aliases='Dmel_ Dmel\\CG9444', description='CG9444 gene product from transcript CG9444-RB', other_designations='CG9444-PA|CG9444-PB', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9681415, end_position_on_the_genomic_accession=9684264, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41176, CurrentID=0, Status='live', Symbol='CG18542', Aliases='Dmel_ Dmel\\CG18542', description='CG18542 gene product from transcript CG18542-RA', other_designations='CG18542-PA|uncharacterized protein', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9685760, end_position_on_the_genomic_accession=9687141, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41177, CurrentID=0, Status='live', Symbol='eca', Aliases='Dmel_CG33104, CG33104, CG9443, Dmel\\CG33104, ECA, p24', description='eclair', other_designations='CG33104-PA|eca-PA', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9684469, end_position_on_the_genomic_accession=9685476, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41178, CurrentID=0, Status='live', Symbol='Unc-115b', Aliases='Dmel_CG31332, CG31332, CG9489, Dmel\\CG31332, anon-WO0140519.9, unc-115', description='CG31332 gene product from transcript CG31332-RD', other_designations='CG31332-PA|CG31332-PB|CG31332-PC|CG31332-PD|CG31332-PE|CG31332-PF|Unc-115b-PA|Unc-115b-PB|Unc-115b-PC|Unc-115b-PD|Unc-115b-PE|Unc-115b-PF|unc-115', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9687060, end_position_on_the_genomic_accession=9692775, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  41179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41179, CurrentID=0, Status='live', Symbol='trbd', Aliases='Dmel_CG9448, CG9448, Dmel\\CG9448, Q9VH90', description='trabid', other_designations='CG9448-PA|trbd-PA', map_location='85E5-85E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9705996, end_position_on_the_genomic_accession=9711037, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41180, CurrentID=0, Status='live', Symbol='dmt', Aliases='Dmel_CG8374, 0481/03, CG8374, Dmel\\CG8374, Dmt, anon-WO0118547.317, aul, s048103', description='dalmatian', other_designations='CG8374-PA|anduril|dmt-PA', map_location='85E5-85E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9711142, end_position_on_the_genomic_accession=9714675, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41181, CurrentID=0, Status='live', Symbol='hyd', Aliases='Dmel_CG9484, CG9484, Dmel\\CG9484, HYD, Hyd, NP_524296.2, UBE 3A, c43, l(3)85Ea, l(3)C43, l(3)C43-3, l(3)SG33, l(3)c43, l[[3]]c43', description='hyperplastic discs', other_designations='CG9484-PA|CG9484-PB|hyd-PA|hyd-PB', map_location='3-49.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9714623, end_position_on_the_genomic_accession=9725615, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  41182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41182, CurrentID=0, Status='live', Symbol='FoxP', Aliases='Dmel_CG43067, CG16899, CG32937, CG43067, DmFoxP, Dmcg16899, Dmel\\CG43067, Dmel_CG16899, Dmel_CG32937, dFoxP, fd85E', description='Forkhead box P', other_designations='CG43067-PC|CG43067-PD|CG43067-PE|FoxP-PC|FoxP-PD|FoxP-PE|forkhead box P|forkhead domain 85E', map_location='85E5-85E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9726020, end_position_on_the_genomic_accession=9730872, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41183, CurrentID=0, Status='live', Symbol='alphaTub85E', Aliases='Dmel_CG9476, 2t, ALPHA 85E, CG9476, DTA3, Dmel\\CG9476, T, Tub, a2t, aTub85E, alpha-Tub, alpha-Tub85E, alpha-tub, alpha-tubulin, alpha2, alpha2t, alpha85E, alpha85E-tub, alphaTub, alphaTub2, alphaTub84E', description='alpha-Tubulin at 85E', other_designations='CG9476-PA|alpha tubulin|alpha-Tubulin|alpha-Tubulin 85E|alpha2-tubulin|alpha2tubulin|alpha85E-tubulin|alphaT|alphaTub85E-PA|alphatubulin|tubulin', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9731221, end_position_on_the_genomic_accession=9733250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41184, CurrentID=0, Status='live', Symbol='CG12950', Aliases='Dmel_ 12950, CT32101, Dmel\\CG12950', description='CG12950 gene product from transcript CG12950-RB', other_designations='CG12950-PA|CG12950-PB|uncharacterized protein', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9734012, end_position_on_the_genomic_accession=9739979, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  41185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41185, CurrentID=0, Status='live', Symbol='Pnn', Aliases='Dmel_CG8383, CG8383, Dmel\\CG8383, FBgn0037737, dPnn', description='Pinin', other_designations='CG8383-PA|CG8383-PB|Pnn-PA|Pnn-PB|pinin', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9740899, end_position_on_the_genomic_accession=9742844, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41187, CurrentID=0, Status='live', Symbol='CG12948', Aliases='Dmel_ Dmel\\CG12948', description='CG12948 gene product from transcript CG12948-RB', other_designations='CG12948-PA|CG12948-PB|uncharacterized protein', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9746167, end_position_on_the_genomic_accession=9747744, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41189, CurrentID=0, Status='live', Symbol='CG16908', Aliases='Dmel_ Dmel\\CG16908', description='CG16908 gene product from transcript CG16908-RA', other_designations='CG16908-PA|uncharacterized protein', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9751088, end_position_on_the_genomic_accession=9754686, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41190, CurrentID=0, Status='live', Symbol='Rpt3R', Aliases='Dmel_CG9475, 9475, CG9475, DmCG9475, Dm_Rpt3b, Dmel\\CG9475, S6', description='Regulatory particle triple-A ATPase 3-related', other_designations='CG9475-PC|CG9475-PD|Rpt3R-PC|Rpt3R-PD|regulatory particle triple-A ATPase 3-related', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9754735, end_position_on_the_genomic_accession=9756510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41191, CurrentID=0, Status='live', Symbol='CG8412', Aliases='Dmel_ Dmel\\CG8412, dAlg12', description='CG8412 gene product from transcript CG8412-RA', other_designations='CG8412-PA', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9756491, end_position_on_the_genomic_accession=9759313, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41192, CurrentID=0, Status='live', Symbol='CG8417', Aliases='Dmel_ Dmel\\CG8417', description='CG8417 gene product from transcript CG8417-RA', other_designations='CG8417-PA', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9759622, end_position_on_the_genomic_accession=9761208, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41193, CurrentID=0, Status='live', Symbol='Snap24', Aliases='Dmel_CG9474, CG9474, Dmel\\CG9474, SNAP-24, SNAP-25b, SNAP24, Snap25b', description='Synaptosomal-associated protein 24kDa', other_designations='CG9474-PA|Snap24-PA|Snap25b|synapse protein 24', map_location='85E6-85E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9761305, end_position_on_the_genomic_accession=9763150, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41194, CurrentID=0, Status='live', Symbol='CG8478', Aliases='Dmel_ Dmel\\CG8478', description='CG8478 gene product from transcript CG8478-RB', other_designations='CG8478-PB|CG8478-PC|CG8478-PD|CG8478-PE', map_location='85E7-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9763650, end_position_on_the_genomic_accession=9766135, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41195, CurrentID=0, Status='live', Symbol='CG8481', Aliases='Dmel_ Dmel\\CG8481', description='CG8481 gene product from transcript CG8481-RB', other_designations='CG8481-PA|CG8481-PB', map_location='85E8-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9767834, end_position_on_the_genomic_accession=9770027, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41197, CurrentID=0, Status='live', Symbol='CG9471', Aliases='Dmel_ BVR, Dmel\\CG9471', description='CG9471 gene product from transcript CG9471-RA', other_designations='CG9471-PA|CG9471-PB|biliverdin IX-beta reductase|biliverdin reductase', map_location='85E8-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9770532, end_position_on_the_genomic_accession=9771542, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41198, CurrentID=0, Status='live', Symbol='Whamy', Aliases='Dmel_CG12946, CG12946, Dmel\\CG12946', description='CG12946 gene product from transcript CG12946-RC', other_designations='CG12946-PA|CG12946-PB|CG12946-PC|CG12946-PD|CG12946-PE|Whamy-PA|Whamy-PB|Whamy-PC|Whamy-PD|Whamy-PE|whamy', map_location='85E8-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9771795, end_position_on_the_genomic_accession=9774419, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41199, CurrentID=0, Status='live', Symbol='topi', Aliases='Dmel_CG8484, CG8484, Dmel\\CG8484, Topi', description='matotopetli', other_designations='CG8484-PA|topi-PA', map_location='85E8-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9775121, end_position_on_the_genomic_accession=9777943, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41200, CurrentID=0, Status='live', Symbol='RpS29', Aliases='Dmel_CG8495, BcDNA:AT13329, BcDNA:AT28563, BcDNA:RH06643, CG8495, Dmel\\CG8495, M(3)85E, M(3)LS5, anon-EST:fe3B4, anon-WO0118547.360, chr3R:5604479..5604622', description='Ribosomal protein S29', other_designations='CG8495-PA|CG8495-PB|RpS29-PA|RpS29-PB|anon-fast-evolving-3B4|minute (3) 85E|ribosomal protein S29', map_location='85E11-85E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9778042, end_position_on_the_genomic_accession=9779528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41201, CurrentID=0, Status='live', Symbol='CG12947', Aliases='Dmel_ Dmel\\CG12947', description='CG12947 gene product from transcript CG12947-RC', other_designations='CG12947-PC|uncharacterized protein', map_location='85E8-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9780064, end_position_on_the_genomic_accession=9783054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41202, CurrentID=0, Status='live', Symbol='MtnA', Aliases='Dmel_CG9470, BcDNA:GH18460, CG9470, Dmel\\CG9470, DrosMtn, MTN, Mtn, anon-WO0153538.42, anon-WO0153538.44, anon-WO0153538.49, mtnA', description='Metallothionein A', other_designations='CG9470-PA|CG9470-PB|MtnA-PA|MtnA-PB|metallothionein|metallothionein A|metallothionein N', map_location='85E8-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9783407, end_position_on_the_genomic_accession=9784370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41203, CurrentID=0, Status='live', Symbol='CG8500', Aliases='Dmel_ Dmel\\CG8500, Q9VH66', description='CG8500 gene product from transcript CG8500-RA', other_designations='CG8500-PA|CG8500-PB', map_location='85E9-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9784653, end_position_on_the_genomic_accession=9789323, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41204, CurrentID=0, Status='live', Symbol='CG12945', Aliases='Dmel_ Dmel\\CG12945', description='CG12945 gene product from transcript CG12945-RD', other_designations='CG12945-PB|CG12945-PC|CG12945-PD|uncharacterized protein', map_location='85E9-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9787816, end_position_on_the_genomic_accession=9791221, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41205, CurrentID=0, Status='live', Symbol='CG8507', Aliases='Dmel_ Dmel\\CG8507', description='CG8507 gene product from transcript CG8507-RA', other_designations='CG8507-PA', map_location='85E9-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9791483, end_position_on_the_genomic_accession=9793098, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41206, CurrentID=0, Status='live', Symbol='CG8516', Aliases='Dmel_ Dmel\\CG8516, FBgn0037757', description='CG8516 gene product from transcript CG8516-RA', other_designations='CG8516-PA', map_location='85E9-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9793295, end_position_on_the_genomic_accession=9797754, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41207, CurrentID=0, Status='live', Symbol='CG9467', Aliases='Dmel_ Dmel\\CG9467', description='CG9467 gene product from transcript CG9467-RA', other_designations='CG9467-PA', map_location='85E9-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9797769, end_position_on_the_genomic_accession=9801026, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41208, CurrentID=0, Status='live', Symbol='CG8526', Aliases='Dmel_ Dmel\\CG8526', description='CG8526 gene product from transcript CG8526-RA', other_designations='CG8526-PA|CG8526-PB', map_location='85E9-85E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9801280, end_position_on_the_genomic_accession=9803461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41209, CurrentID=0, Status='live', Symbol='FBXO11', Aliases='Dmel_CG9461, CG9461, Dmel\\CG9461, FBX011, dFBX011', description='F-box protein 11', other_designations='CG9461-PA|CG9461-PB|FBX011 ortholog|FBXO11-PA|FBXO11-PB', map_location='85E9-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9803486, end_position_on_the_genomic_accession=9809368, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41210, CurrentID=0, Status='live', Symbol='CG8534', Aliases='Dmel_ Dmel\\CG8534', description='CG8534 gene product from transcript CG8534-RA', other_designations='CG8534-PA', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9812747, end_position_on_the_genomic_accession=9813939, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41211, CurrentID=0, Status='live', Symbol='eloF', Aliases='Dmel_CG16905, CG16905, Dmel\\CG16905', description='elongase F', other_designations='CG16905-PA|eloF-PA|elongase', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9814161, end_position_on_the_genomic_accession=9815158, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41212, CurrentID=0, Status='live', Symbol='CG16904', Aliases='Dmel_ Dmel\\CG16904', description='CG16904 gene product from transcript CG16904-RA', other_designations='CG16904-PA|uncharacterized protein', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9815928, end_position_on_the_genomic_accession=9816979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318891, CurrentID=0, Status='live', Symbol='CG33124', Aliases='Dmel_CG33124, CG10879-PB, Dmel\\CG33124, dSLC5A13', description='CG33124 gene product from transcript CG33124-RD', other_designations='CG33124-PD|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2206525, end_position_on_the_genomic_accession=2209102, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41214, CurrentID=0, Status='live', Symbol='CG9458', Aliases='Dmel_ CT26802, Dmel\\CG9458', description='CG9458 gene product from transcript CG9458-RA', other_designations='CG9458-PA', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9818914, end_position_on_the_genomic_accession=9819877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41215, CurrentID=0, Status='live', Symbol='Teh1', Aliases='Dmel_CG12806, CG12806, Dmel\\CG12806, TEH1', description='tipE homolog 1', other_designations='CG12806-PA|CG12806-PB|Teh1-PA|Teh1-PB', map_location='85E10-85E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9826081, end_position_on_the_genomic_accession=9868120, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41217, CurrentID=0, Status='live', Symbol='Glut4EF', Aliases='Dmel_CG34360, BcDNA:GH25505, BcDNA:GM06804, BcDNA:RE66512, BcDNA:RH63124, CG11676, CG12802, CG12804, CG12805, CG32469, CG33975, CG34360, CG43757, D- Dmel\\CG34360, Dmel_CG12802, Dmel_CG33975, stretch', description='CG34360 gene product from transcript CG34360-RH', other_designations='BcDNA:GH25505|BcDNA:GM06804|CG34360-PA|CG34360-PB|CG34360-PC|CG34360-PD|CG34360-PE|CG34360-PF|CG34360-PG|CG34360-PH|Glut4EF-PA|Glut4EF-PB|Glut4EF-PC|Glut4EF-PD|Glut4EF-PE|Glut4EF-PF|Glut4EF-PG|Glut4EF-PH|glucose transporter 4 enhancer factor|stretched out', map_location='85E14-85F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9874681, end_position_on_the_genomic_accession=9995364, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41219, CurrentID=0, Status='live', Symbol='Art4', Aliases='Dmel_CG5358, CARM1, CG5358, DART4, Dmel\\CG5358, dArt 4', description='Arginine methyltransferase 4', other_designations='Art4-PA|Art4-PB|CARMER|CG5358-PA|CG5358-PB|arginie methyl transferase 4|arginine methyl-transferase 4|arginine methyltransferase 4|coactivator Arg methyltransferase for Ecr/Usp', map_location='85F4-85F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9982816, end_position_on_the_genomic_accession=9986060, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41221, CurrentID=0, Status='live', Symbol='Spn85F', Aliases='Dmel_CG12807, CG12807, Dmel\\CG12807', description='Serpin 85F', other_designations='CG12807-PA|Spn85F-PA|serpin 85F', map_location='85F5-85F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9997163, end_position_on_the_genomic_accession=10000314, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41222, CurrentID=0, Status='live', Symbol='CG5359', Aliases='Dmel_ BcDNA:RE37471, Dmel\\CG5359', description='CG5359 gene product from transcript CG5359-RC', other_designations='CG5359-PA|CG5359-PB|CG5359-PC', map_location='85F5-85F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9999771, end_position_on_the_genomic_accession=10001276, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41225, CurrentID=0, Status='live', Symbol='Mical', Aliases='Dmel_CG33208, CG11685, CG11687, CG18484, CG18667, CG18668, CG33186, CG33190, CG33208, Dmel\\CG33208, MICAL, mdcds_49226, mical, milch', description='Molecule interacting with CasL', other_designations='CG33208-PA|CG33208-PB|CG33208-PC|CG33208-PD|CG33208-PE|CG33208-PF|CG33208-PG|CG33208-PI|CG33208-PK|CG33208-PL|CG33208-PN|CG33208-PO|CG33208-PP|CG33208-PQ|Mical-PA|Mical-PB|Mical-PC|Mical-PD|Mical-PE|Mical-PF|Mical-PG|Mical-PI|Mical-PK|Mical-PL|Mical-PN|Mical-PO|Mical-PP|Mical-PQ|mical|milchstrasse|molecule interacting with CasL|molecule that interacts with CasL', map_location='85F5-85F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10001554, end_position_on_the_genomic_accession=10042895, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  41226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41226, CurrentID=0, Status='live', Symbol='CG3909', Aliases='Dmel_ BcDNA:LD21537, CG3039, Dmel\\CG3909, Ski8', description='CG3909 gene product from transcript CG3909-RA', other_designations='CG3909-PA', map_location='85F7-85F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10043270, end_position_on_the_genomic_accession=10044606, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41227, CurrentID=0, Status='live', Symbol='CG11722', Aliases='Dmel_ BcDNA:AT14909, Dmel\\CG11722', description='CG11722 gene product from transcript CG11722-RA', other_designations='CG11722-PA|uncharacterized protein', map_location='85F7-85F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10044596, end_position_on_the_genomic_accession=10045542, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41228, CurrentID=0, Status='live', Symbol='mtTFB2', Aliases='Dmel_CG3910, CG3910, Dmel\\CG3910, TFB2, TFB2M, d-TF2BM, d-mtTFB2', description='mitochondrial transcription factor B2', other_designations='CG3910-PA|mtTFB2-PA', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10045574, end_position_on_the_genomic_accession=10047260, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41229, CurrentID=0, Status='live', Symbol='CG12811', Aliases='Dmel_ Dmel\\CG12811', description='CG12811 gene product from transcript CG12811-RA', other_designations='CG12811-PA|uncharacterized protein', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10047252, end_position_on_the_genomic_accession=10048797, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41230, CurrentID=0, Status='live', Symbol='CG3925', Aliases='Dmel_ Dmel\\CG3925', description='CG3925 gene product from transcript CG3925-RA', other_designations='CG3925-PA', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10049199, end_position_on_the_genomic_accession=10051527, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41231, CurrentID=0, Status='live', Symbol='Fancl', Aliases='Dmel_CG12812, CG12812, Dmel\\CG12812, FAL, FANCL', description='CG12812 gene product from transcript CG12812-RB', other_designations='CG12812-PA|CG12812-PB|Fancl-PA|Fancl-PB|fancl', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10051508, end_position_on_the_genomic_accession=10052869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41232, CurrentID=0, Status='live', Symbol='Npc2d', Aliases='Dmel_CG12813, CG12813, Dmel\\CG12813, npc2d', description='Niemann-Pick type C-2d', other_designations='CG12813-PA|Npc2d-PA', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10055356, end_position_on_the_genomic_accession=10056083, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41233, CurrentID=0, Status='live', Symbol='Npc2c', Aliases='Dmel_CG3934, CG3934, DmML5A, Dmel\\CG3934, npc2c', description='Niemann-Pick type C-2c', other_designations='CG3934-PA|Npc2c-PA', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10060252, end_position_on_the_genomic_accession=10061136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41235, CurrentID=0, Status='live', Symbol='nerfin-2', Aliases='Dmel_CG12809, CG12809, Dmel\\CG12809, Nerfin', description='nervous fingers 2', other_designations='CG12809-PA|CG12809-PB|nerfin-2-PA|nerfin-2-PB|nervous fingers', map_location='85F9-85F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10071117, end_position_on_the_genomic_accession=10073761, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41236, CurrentID=0, Status='live', Symbol='CG5361', Aliases='Dmel_ Dmel\\CG5361', description='CG5361 gene product from transcript CG5361-RB', other_designations='CG5361-PB', map_location='85F9-85F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10073821, end_position_on_the_genomic_accession=10075373, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41238, CurrentID=0, Status='live', Symbol='CG3940', Aliases='Dmel_ Dmel\\CG3940', description='CG3940 gene product from transcript CG3940-RB', other_designations='CG3940-PA|CG3940-PB|CG3940-PC', map_location='85F12-85F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10112237, end_position_on_the_genomic_accession=10120081, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41239, CurrentID=0, Status='live', Symbol='gammaSnap2', Aliases='Dmel_CG6208, CG6208, Dmel\\CG6208, SNAP-31, gammaSNAP, gammaSnap-2', description='gamma Soluble NSF attachment protein 2', other_designations='CG6208-PA|gamma soluble NSF attachment protein 2|gammaSnap-2|gammaSnap2-PA', map_location='85F13-85F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10120399, end_position_on_the_genomic_accession=10121400, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41242, CurrentID=0, Status='live', Symbol='TAF1B', Aliases='Dmel_CG6241, CG6241, Dmel\\CG6241', description='TATA box-binding protein-associated factor RNA polymerase I subunit B', other_designations='CG6241-PA|CG6241-PB|CG6241-PC|TAF1B-PA|TAF1B-PB|TAF1B-PC', map_location='85F14-85F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10140103, end_position_on_the_genomic_accession=10144184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41244, CurrentID=0, Status='live', Symbol='CG6254', Aliases='Dmel_ Dmel\\CG6254', description='CG6254 gene product from transcript CG6254-RA', other_designations='CG6254-PA', map_location='85F15-85F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10159144, end_position_on_the_genomic_accession=10161331, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41246, CurrentID=0, Status='live', Symbol='CG12814', Aliases='Dmel_ CG12815, Dmel\\CG12814', description='CG12814 gene product from transcript CG12814-RC', other_designations='CG12814-PB|CG12814-PC|CG12814-PD|CG12814-PE|uncharacterized protein', map_location='85F16-85F16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10171156, end_position_on_the_genomic_accession=10189603, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41247, CurrentID=0, Status='live', Symbol='Syn', Aliases='Dmel_CG3985, CG3985, Dmel\\CG3985, Dsyn-1, SYN, SYN1, SYN2, syn', description='Synapsin', other_designations='CG3985-PA|CG3985-PD|CG3985-PE|CG3985-PF|Syn-PA|Syn-PD|Syn-PE|Syn-PF|Synapsin1|synapsin|synapsin I', map_location='85F16-86A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10190689, end_position_on_the_genomic_accession=10222442, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  41248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41248, CurrentID=0, Status='live', Symbol='Timp', Aliases='Dmel_CG6281, CG6281, DTIMP, Dmel\\CG6281, Dtimp, TIMP, dTIMP, timp', description='Tissue inhibitor of metalloproteases', other_designations='CG6281-PB|Timp-PB|tissue inhibitor of metalloproteases|tissue inhibitor of metalloproteinases', map_location='86A1-86A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10203875, end_position_on_the_genomic_accession=10213007, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41249, CurrentID=0, Status='live', Symbol='CG12420', Aliases='Dmel_ Dmel\\CG12420', description='CG12420 gene product from transcript CG12420-RA', other_designations='CG12420-PA|uncharacterized protein', map_location='86A1-86A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10222698, end_position_on_the_genomic_accession=10223745, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41250, CurrentID=0, Status='live', Symbol='CG12817', Aliases='Dmel_ BcDNA:RE53107, Dmel\\CG12817', description='CG12817 gene product from transcript CG12817-RB', other_designations='CG12817-PB|uncharacterized protein', map_location='86A1-86A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10223440, end_position_on_the_genomic_accession=10224725, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41252, CurrentID=0, Status='live', Symbol='CG42795', Aliases='Dmel_ CG12816, CG34108, CG3996, Dmel\\CG42795, Dmel_CG34108, Dmel_CG3996', description='CG42795 gene product from transcript CG42795-RF', other_designations='CG42795-PA|CG42795-PB|CG42795-PC|CG42795-PE|CG42795-PF|CG42795-PG|uncharacterized protein', map_location='86A2-86A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10225690, end_position_on_the_genomic_accession=10251277, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  41253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41253, CurrentID=0, Status='live', Symbol='CG3999', Aliases='Dmel_ Dmel\\CG3999', description='CG3999 gene product from transcript CG3999-RA', other_designations='CG3999-PA|CG3999-PB', map_location='86A2-86A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10251518, end_position_on_the_genomic_accession=10255268, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41254, CurrentID=0, Status='live', Symbol='Sirt6', Aliases='Dmel_CG6284, CG6284, D.mel4, Dmel\\CG6284, dSIRT6, dmSRT408', description='Sirtuin 6', other_designations='CG6284-PB|Sirt6-PB|sirtuin 6', map_location='86A3-86A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10255349, end_position_on_the_genomic_accession=10261849, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41256, CurrentID=0, Status='live', Symbol='CG43143', Aliases='Dmel_ CG11870, CG11871, CG11872, Dmel\\CG43143, Dmel_CG11870, Dmel_CG11872, EMK/KIAA0537, NUAK, Omphk1', description='CG43143 gene product from transcript CG43143-RI', other_designations='CG43143-PA|CG43143-PB|CG43143-PC|CG43143-PD|CG43143-PE|CG43143-PF|CG43143-PG|CG43143-PH|CG43143-PI|uncharacterized protein', map_location='86A7-86A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10264306, end_position_on_the_genomic_accession=10308467, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  41259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41259, CurrentID=0, Status='live', Symbol='CG6293', Aliases='Dmel_ Dmel\\CG6293', description='CG6293 gene product from transcript CG6293-RB', other_designations='CG6293-PA|CG6293-PB', map_location='86A7-86A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10307806, end_position_on_the_genomic_accession=10311613, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41260, CurrentID=0, Status='live', Symbol='Bruce', Aliases='Dmel_CG6303, BRUCE, CG6303, Dmel\\CG6303, bruce, dBRUCE, dBruce, dbruce, mod86', description='BIR repeat containing ubiquitin-conjugating enzyme', other_designations='Bruce-PA|Bruce-PB|Bruce-PC|Bruce-PD|Bruce-PE|CG6303-PA|CG6303-PB|CG6303-PC|CG6303-PD|CG6303-PE|dbruce|mod86', map_location='86A7-86A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10312039, end_position_on_the_genomic_accession=10331353, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  41261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41261, CurrentID=0, Status='live', Symbol='CG12818', Aliases='Dmel_ Dmel\\CG12818', description='CG12818 gene product from transcript CG12818-RA', other_designations='CG12818-PA|uncharacterized protein', map_location='86A8-86A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10331581, end_position_on_the_genomic_accession=10332695, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41262, CurrentID=0, Status='live', Symbol='sle', Aliases='Dmel_CG12819, CG12819, CG12819-PA, Dmel\\CG12819, FBtr0082231, SLE', description='slender lobes', other_designations='CG12819-PA|CG12819-PB|sle-PA|sle-PB|slender lobe', map_location='86A8-86B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10333115, end_position_on_the_genomic_accession=10338265, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41263, CurrentID=0, Status='live', Symbol='CG12592', Aliases='Dmel_CG12592-PA, Dmel\\CG12592', description='CG12592 gene product from transcript CG12592-RA', other_designations='CG12592-PA|uncharacterized protein', map_location='86B1-86B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10338342, end_position_on_the_genomic_accession=10340872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41264, CurrentID=0, Status='live', Symbol='CG18545', Aliases='Dmel_CG18545-PA, Dmel\\CG18545', description='CG18545 gene product from transcript CG18545-RA', other_designations='CG18545-PA|uncharacterized protein', map_location='86B1-86B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10340949, end_position_on_the_genomic_accession=10341668, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41265, CurrentID=0, Status='live', Symbol='jumu', Aliases='Dmel_CG4029, 5295, CG4029, Dmel\\CG4029, Dmjumu, Dom, Dwhn, E(var)631, Jumeau, Jumu, Scim31, Whn, jumeaux/Dom, l(3)06142, l(3)06439, jumu', description='jumeau', other_designations='CG4029-PA|domina|jumeaux|jumu-PA|winged-helix nude|winged-helix nude-like', map_location='86B1-86B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10350182, end_position_on_the_genomic_accession=10361612, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41266, CurrentID=0, Status='live', Symbol='Rfx', Aliases='Dmel_CG6312, CG6312, Dmel\\CG6312, RFX, dRFX, d rfx', description='CG6312 gene product from transcript CG6312-RC', other_designations='CG6312-PC|CG6312-PF|CG6312-PG|CG6312-PH|CG6312-PI|CG6312-PJ|Rfx-PC|Rfx-PF|Rfx-PG|Rfx-PH|Rfx-PI|Rfx-PJ|regulatory factor X', map_location='86B1-86B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10361760, end_position_on_the_genomic_accession=10378068, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41268, CurrentID=0, Status='live', Symbol='CG6325', Aliases='Dmel_ Dmel\\CG6325', description='CG6325 gene product from transcript CG6325-RA', other_designations='CG6325-PA', map_location='86B4-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10402967, end_position_on_the_genomic_accession=10407285, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41269, CurrentID=0, Status='live', Symbol='Irp-1B', Aliases='Dmel_CG6342, 144037_at, C-Acon, CG6342, Dmel\\CG6342, IRF, IRP, IRP-1, IRP-1B, IRP1, cACON, cAcon, dIRP, irp-1B, irp1B', description='Iron regulatory protein 1B', other_designations='CG6342-PA|Irp-1B-PA|c-aconitase|cytoplasmic aconitase|cytosolic aconitase|iron regulatory factor|iron regulatory protein|iron regulatory protein 1B', map_location='86B4-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10408618, end_position_on_the_genomic_accession=10412869, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  41270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41270, CurrentID=0, Status='live', Symbol='Rrp46', Aliases='Dmel_CG4043, CG4043, Dmel\\CG4043, d rrp46', description='CG4043 gene product from transcript CG4043-RA', other_designations='CG4043-PA|Rrp46-PA', map_location='86B4-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10413301, end_position_on_the_genomic_accession=10414242, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41271, CurrentID=0, Status='live', Symbol='CG6345', Aliases='Dmel_ 149631_at, Dmel\\CG6345', description='CG6345 gene product from transcript CG6345-RA', other_designations='CG6345-PA', map_location='86B4-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10414218, end_position_on_the_genomic_accession=10416046, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41272, CurrentID=0, Status='live', Symbol='Cyp12e1', Aliases='Dmel_CG14680, 12e1, CG14680, Dmel\\CG14680', description='CG14680 gene product from transcript CG14680-RC', other_designations='CG14680-PC|Cyp12e1-PC', map_location='86B4-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10434789, end_position_on_the_genomic_accession=10436704, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41273, CurrentID=0, Status='live', Symbol='hth', Aliases='Dmel_CG17117, 1323/07, 1422/04, CG17117, Dm-HTH, Dmel\\CG17117, HTH, Hth, Meis1, P53, anon-EST:Liang-2.13, clone 2.13, dtl1, hth2, l(3)05745, l(3)86Ca, hth', description='homothorax', other_designations='CG17117-PA|CG17117-PC|CG17117-PE|CG17117-PF|CG17117-PI|dorsotonals|gene II|hth-PA|hth-PC|hth-PE|hth-PF|hth-PI', map_location='86C1-86C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10507561, end_position_on_the_genomic_accession=10639568, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  41274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41274, CurrentID=0, Status='live', Symbol='CG6465', Aliases='Dmel_ Dmel\\CG6465', description='CG6465 gene product from transcript CG6465-RA', other_designations='CG6465-PA', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10675560, end_position_on_the_genomic_accession=10677251, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41275, CurrentID=0, Status='live', Symbol='CG14688', Aliases='Dmel_ Dmel\\CG14688', description='CG14688 gene product from transcript CG14688-RB', other_designations='CG14688-PB|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10677737, end_position_on_the_genomic_accession=10680925, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41278, CurrentID=0, Status='live', Symbol='CG14683', Aliases='Dmel_ Dmel\\CG14683', description='CG14683 gene product from transcript CG14683-RA', other_designations='CG14683-PA|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10696647, end_position_on_the_genomic_accession=10698976, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41279, CurrentID=0, Status='live', Symbol='pug', Aliases='Dmel_CG4067, CG4067, Dmel\\CG4067, MTHFD, Pug, bs14h02.y1D, pug', description='pugilist', other_designations='CG4067-PA|CG4067-PB|CG4067-PC|CG4067-PD|CG4067-PE|pug-PA|pug-PB|pug-PC|pug-PD|pug-PE', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10696534, end_position_on_the_genomic_accession=10703316, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41280, CurrentID=0, Status='live', Symbol='CG31391', Aliases='Dmel_ BcDNA:AT04406, CG6471, Dmel\\CG31391', description='CG31391 gene product from transcript CG31391-RA', other_designations='CG31391-PA|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10703372, end_position_on_the_genomic_accession=10705075, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41281, CurrentID=0, Status='live', Symbol='CG14684', Aliases='Dmel_ Dmel\\CG14684', description='CG14684 gene product from transcript CG14684-RA', other_designations='CG14684-PA|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10706974, end_position_on_the_genomic_accession=10707460, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41282, CurrentID=0, Status='live', Symbol='CG31278', Aliases='Dmel_ CG6482, Dmel\\CG31278, PDF', description='CG31278 gene product from transcript CG31278-RA', other_designations='CG31278-PA|CG31278-PB|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10707398, end_position_on_the_genomic_accession=10708340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41283, CurrentID=0, Status='live', Symbol='CG14689', Aliases='Dmel_ Dmel\\CG14689', description='CG14689 gene product from transcript CG14689-RA', other_designations='CG14689-PA|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10709588, end_position_on_the_genomic_accession=10711248, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41284, CurrentID=0, Status='live', Symbol='CG4073', Aliases='Dmel_ Dmel\\CG4073', description='CG4073 gene product from transcript CG4073-RA', other_designations='CG4073-PA', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10714798, end_position_on_the_genomic_accession=10715901, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41285, CurrentID=0, Status='live', Symbol='tomboy20', Aliases='Dmel_CG14690, CG14690, Dmel\\CG14690', description='CG14690 gene product from transcript CG14690-RA', other_designations='CG14690-PA|dtomboy20|tomboy20-PA', map_location='86C5-86C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10735277, end_position_on_the_genomic_accession=10735957, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41286, CurrentID=0, Status='live', Symbol='TkR86C', Aliases='Dmel_CG6515, CG6515, DmNTLR, Dmel\\CG6515, NKD, NKD12, NKr, Nkd, TAKR86C, TakR86C, Takr86C', description='Tachykinin-like receptor at 86C', other_designations='CG6515-PA|CG6515-PB|TkR86C-PA|TkR86C-PB|natalisin receptor|neurokinin receptor from drosophila|neuropeptide receptor|tachykinin receptor', map_location='86C5-86C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10737592, end_position_on_the_genomic_accession=10751228, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41287, CurrentID=0, Status='live', Symbol='CG14691', Aliases='Dmel_ Dmel\\CG14691', description='CG14691 gene product from transcript CG14691-RA', other_designations='CG14691-PA|CG14691-PB|CG14691-PC|uncharacterized protein', map_location='86C5-86C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10753331, end_position_on_the_genomic_accession=10755634, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41288, CurrentID=0, Status='live', Symbol='MED7', Aliases='Dmel_CG31390, CG31390, CG6529, Dmel\\CG31390, Med7, dMED7, dTRAP34', description='Mediator complex subunit 7', other_designations='CG31390-PA|MED7-PA|mediator|mediator complex subunit 7', map_location='86C5-86C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10760071, end_position_on_the_genomic_accession=10761032, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41289, CurrentID=0, Status='live', Symbol='Cap-H2', Aliases='Dmel_CG14685, CAP-H2, CG14685, CapH2, Dmel\\CG14685, cap-H2, dCap-H2', description='Chromosome associated protein H2', other_designations='CG14685-PD|CG14685-PE|CG14685-PF|CG14685-PG|Cap-H2-PD|Cap-H2-PE|Cap-H2-PF|Cap-H2-PG|chromosome associated protein H2', map_location='3-49 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10761310, end_position_on_the_genomic_accession=10766756, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41290, CurrentID=0, Status='live', Symbol='TfIIFbeta', Aliases='Dmel_CG6538, CG6538, Dmel\\CG6538, F[[S]], RAP30, T2FB_DROME, TFIIF, TFIIFbeta, TfIIF, dTFIIF30, l(3)j3C1', description='Transcription factor TFIIFbeta', other_designations='CG6538-PA|TFIIF small subunit|TfIIFbeta-PA|factor 5|transcription factor TFIIFbeta', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10766622, end_position_on_the_genomic_accession=10767804, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41291, CurrentID=0, Status='live', Symbol='fau', Aliases='Dmel_CG45077, BcDNA:HL07933, CG45077, CG6544, Dmel\\CG45077, Dmel_CG6544, Fau', description='CG45077 gene product from transcript CG45077-RA', other_designations='CG45077-PA|fau-PA', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10773197, end_position_on_the_genomic_accession=10777374, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41292, CurrentID=0, Status='live', Symbol='Desi', Aliases='Dmel_CG14686, CG14686, Dmel\\CG14686', description='Desiccate', other_designations='CG14686-PA|Desi-PA|desiccate', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10779569, end_position_on_the_genomic_accession=10780825, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41293, CurrentID=0, Status='live', Symbol='mRpL37', Aliases='Dmel_CG42632, CG42632, CG6547, Dmel\\CG42632, Dmel_CG6547, MRP-L2, MRpL37, mRpL2a', description='mitochondrial ribosomal protein L37', other_designations='CG42632-PB|mRpL37-PB|mitochondrial ribosomal protein L2a', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10782563, end_position_on_the_genomic_accession=10784092, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41294, CurrentID=0, Status='live', Symbol='Rbp1', Aliases='Dmel_CG17136, CG17136, Dmel\\CG17136, RBP1, RRM1, RRM11, Rbp-11, Rbp1D, rbp1, rrm1, rrm11, Rbp1', description='RNA-binding protein 1', other_designations='CG17136-PA|CG17136-PD|RNA-binding protein 11|Rbp1-PA|Rbp1-PD', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10784383, end_position_on_the_genomic_accession=10786465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41295, CurrentID=0, Status='live', Symbol='Art1', Aliases='Dmel_CG6554, CG6554, DART1, Dart, Dmel\\CG6554, dart1', description='Arginine methyltransferase 1', other_designations='Arg methyltransferase 1|Art1-PA|CG6554-PA|arginine methyltransferase 1', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10786481, end_position_on_the_genomic_accession=10788226, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41296, CurrentID=0, Status='live', Symbol='Mcm5', Aliases='Dmel_CG4082, CG4082, DmCDC46, DmCDC465, DmMCM5, DmMcm5, DmeMCM5, Dmel\\CG4082, MCM5, McM5, PCR4, mcm5', description='Minichromosome maintenance 5', other_designations='CG4082-PA|Mcm5-PA|mini-chromosome maintenance 5|minichromosome maintenance 5', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10788524, end_position_on_the_genomic_accession=10791150, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41297, CurrentID=0, Status='live', Symbol='CG14687', Aliases='Dmel_ BcDNA:RE74890, Dmel\\CG14687', description='CG14687 gene product from transcript CG14687-RA', other_designations='CG14687-PA|uncharacterized protein', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10792904, end_position_on_the_genomic_accession=10793643, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41298, CurrentID=0, Status='live', Symbol='CG14692', Aliases='Dmel_ Dmel\\CG14692', description='CG14692 gene product from transcript CG14692-RE', other_designations='CG14692-PE|uncharacterized protein', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10794058, end_position_on_the_genomic_accession=10803955, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  41299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41299, CurrentID=0, Status='live', Symbol='CG14693', Aliases='Dmel_ Dmel\\CG14693', description='CG14693 gene product from transcript CG14693-RC', other_designations='CG14693-PB|CG14693-PC|CG14693-PD|CG14693-PE|CG14693-PF|uncharacterized protein', map_location='86C6-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10811581, end_position_on_the_genomic_accession=10815047, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  41300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41300, CurrentID=0, Status='live', Symbol='CG4089', Aliases='Dmel_ Dmel\\CG4089', description='CG4089 gene product from transcript CG4089-RA', other_designations='CG4089-PA', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10815288, end_position_on_the_genomic_accession=10817490, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41302, CurrentID=0, Status='live', Symbol='Cad86C', Aliases='Dmel_CG42601, BcDNA:LP07955, CG31393, CG31425, CG33112, CG42601, CG4509, CG4655, CT14654, CT15041, DCad86C, Dmel\\CG42601, Dmel_CG4509, Dmel_CG4655', description='Cadherin 86C', other_designations='CG42601-PC|CG42601-PD|CG42601-PE|CG42601-PF|CG42601-PG|Cad86C-PC|Cad86C-PD|Cad86C-PE|Cad86C-PF|Cad86C-PG|cadherin 86C|cadherin Cad86C', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10827079, end_position_on_the_genomic_accession=10849030, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  41303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41303, CurrentID=0, Status='live', Symbol='CG4565', Aliases='Dmel_ Dmel\\CG4565', description='CG4565 gene product from transcript CG4565-RB', other_designations='CG4565-PB', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10852724, end_position_on_the_genomic_accession=10853644, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41304, CurrentID=0, Status='live', Symbol='CG6567', Aliases='Dmel_ Dmel\\CG6567', description='CG6567 gene product from transcript CG6567-RB', other_designations='CG6567-PB', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10853374, end_position_on_the_genomic_accession=10854451, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41305, CurrentID=0, Status='live', Symbol='CG4511', Aliases='Dmel_ Dmel\\CG4511', description='CG4511 gene product from transcript CG4511-RA', other_designations='CG4511-PA|CG4511-PB', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10854501, end_position_on_the_genomic_accession=10856412, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41306, CurrentID=0, Status='live', Symbol='CG4570', Aliases='Dmel_ Dmel\\CG4570', description='CG4570 gene product from transcript CG4570-RA', other_designations='CG4570-PA', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10856807, end_position_on_the_genomic_accession=10858066, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41307, CurrentID=0, Status='live', Symbol='CG14694', Aliases='Dmel_ Dmel\\CG14694', description='CG14694 gene product from transcript CG14694-RA', other_designations='CG14694-PA|uncharacterized protein', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10858202, end_position_on_the_genomic_accession=10859813, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41308, CurrentID=0, Status='live', Symbol='CG6574', Aliases='Dmel_ Dmel\\CG6574', description='CG6574 gene product from transcript CG6574-RA', other_designations='CG6574-PA', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10860531, end_position_on_the_genomic_accession=10862988, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41309, CurrentID=0, Status='live', Symbol='SelR', Aliases='Dmel_CG6584, BcDNA:RE36040, BcDNA:RH61230, CG6584, Dmel\\CG6584, EP3340, MsrB, dMsrB, selR', description='Selenoprotein R', other_designations='CG6584-PA|CG6584-PB|CG6584-PC|CG6584-PE|CG6584-PG|CG6584-PH|CG6584-PI|CG6584-PJ|EP3340|SelR-PA|SelR-PB|SelR-PC|SelR-PE|SelR-PG|SelR-PH|SelR-PI|SelR-PJ|methionine-R-sulfoxide reductase|selenoprotein R', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10863355, end_position_on_the_genomic_accession=10868642, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41310, CurrentID=0, Status='live', Symbol='Tsp86D', Aliases='Dmel_CG4591, CG4591, Dm.Tsp86D, Dmel\\CG4591', description='Tetraspanin 86D', other_designations='CG4591-PA|CG4591-PB|CG4591-PC|Tsp86D-PA|Tsp86D-PB|Tsp86D-PC|tetraspanin 86D', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10868708, end_position_on_the_genomic_accession=10870555, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41311, CurrentID=0, Status='live', Symbol='Fdh', Aliases='Dmel_CG6598, ADHX, ADHX_DROME, CG6598, Dmel\\CG6598, GSNOR, ODH, Odh, fdh, gfd, gfd/odh, odh', description='Formaldehyde dehydrogenase', other_designations='CG6598-PA|Fdh-PA|formaldehyde dehydrogenase|glutathione-dependent formaldehyde dehydrogenase|octanol dehydrogenase', map_location='3-49.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10870906, end_position_on_the_genomic_accession=10872381, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41312, CurrentID=0, Status='live', Symbol='CG4596', Aliases='Dmel_ Dmel\\CG4596', description='CG4596 gene product from transcript CG4596-RA', other_designations='CG4596-PA', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10873038, end_position_on_the_genomic_accession=10875582, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41313, CurrentID=0, Status='live', Symbol='Sodh-2', Aliases='Dmel_CG4649, CG4649, DM_7299382, Dmel\\CG4649, SODH-2, Sdh, Sdh-2', description='Sorbitol dehydrogenase-2', other_designations='CG4649-PA|Sodh-2-PA|sorbitol dehydrogenase 2|sorbitol dehydrogenase-2', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10876384, end_position_on_the_genomic_accession=10878257, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41314, CurrentID=0, Status='live', Symbol='CG14695', Aliases='Dmel_ Dmel\\CG14695', description='CG14695 gene product from transcript CG14695-RA', other_designations='CG14695-PA|uncharacterized protein', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10878204, end_position_on_the_genomic_accession=10879319, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41316, CurrentID=0, Status='live', Symbol='Tpc1', Aliases='Dmel_CG6608, CG6608, DmTpc1, Dmel\\CG6608', description='Thiamine pyrophosphate carrier protein 1', other_designations='CG6608-PA|CG6608-PB|Tpc1-PA|Tpc1-PB|thiamine pyrophosphate carrier protein|thiamine pyrophosphate carrier protein 1', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10879572, end_position_on_the_genomic_accession=10882188, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41317, CurrentID=0, Status='live', Symbol='CG14696', Aliases='Dmel_ Dmel\\CG14696', description='CG14696 gene product from transcript CG14696-RA', other_designations='CG14696-PA|uncharacterized protein', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10882433, end_position_on_the_genomic_accession=10885001, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41318, CurrentID=0, Status='live', Symbol='Adk3', Aliases='Dmel_CG6612, AK3, CG6612, DAK3, Dak3, Dmel\\CG6612, bs12h03.y1', description='Adenylate kinase 3', other_designations='ADENYLATE KINASE - 3|Adk3-PA|Adk3-PB|Adk3-PC|Adk3-PD|CG6612-PA|CG6612-PB|CG6612-PC|CG6612-PD|adenylate kinase 3|adenylate kinase-3', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10887345, end_position_on_the_genomic_accession=10889588, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41319, CurrentID=0, Status='live', Symbol='CG6621', Aliases='Dmel_ Dmel\\CG6621', description='CG6621 gene product from transcript CG6621-RB', other_designations='CG6621-PA|CG6621-PB', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10889768, end_position_on_the_genomic_accession=10893531, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41320, CurrentID=0, Status='live', Symbol='Leash', Aliases='Dmel_CG4674, CG32471, CG4674, Dmel\\CG4674, leash', description='CG4674 gene product from transcript CG4674-RA', other_designations='CG4674-PA|Leash-PA|leash', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10885540, end_position_on_the_genomic_accession=10896418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41321, CurrentID=0, Status='live', Symbol='Tengl4', Aliases='Dmel_CG4683, CG4683, Dmel\\CG4683', description='Testis EndoG-Like 4', other_designations='CG4683-PA|Tengl4-PA|testis EndoG-Like 4', map_location='86C12-86C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10950284, end_position_on_the_genomic_accession=10951797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41324, CurrentID=0, Status='live', Symbol='CG6629', Aliases='Dmel_ Dmel\\CG6629', description='CG6629 gene product from transcript CG6629-RA', other_designations='CG6629-PA', map_location='86D4-86D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11114671, end_position_on_the_genomic_accession=11115706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41326, CurrentID=0, Status='live', Symbol='CG4706', Aliases='Dmel_ Dmel\\CG4706', description='CG4706 gene product from transcript CG4706-RA', other_designations='CG4706-PA', map_location='86D4-86D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11137177, end_position_on_the_genomic_accession=11139742, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41330, CurrentID=0, Status='live', Symbol='CG4757', Aliases='Dmel_ BcDNA:GH05741, Dmel\\CG4757, GH05702, GH05741', description='CG4757 gene product from transcript CG4757-RB', other_designations='CG4757-PA|CG4757-PB', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11161207, end_position_on_the_genomic_accession=11163498, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246605, CurrentID=0, Status='live', Symbol='CG30423', Aliases='Dmel_ BcDNA:RH04491, Dmel\\CG30423', description='CG30423 gene product from transcript CG30423-RE', other_designations='CG30423-PE|uncharacterized protein', map_location='60E4-60E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24874976, end_position_on_the_genomic_accession=24875865, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41333, CurrentID=0, Status='live', Symbol='Ugt35b', Aliases='Dmel_CG6649, AC 006491B, CG6649, DmeUgt35b, Dmel\\CG6649, UDPGT2, UGT35b, Ugt-35b, Ugt35B, Ugt86Db, ugt35b', description='UDP-glycosyltransferase 35b', other_designations='CG6649-PA|UDP-glycosyltransferase|Ugt35b-PA', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11168503, end_position_on_the_genomic_accession=11170246, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41334, CurrentID=0, Status='live', Symbol='Ugt35a', Aliases='Dmel_CG6644, AC 006491F, CG6644, DmeUgt35a, Dmel\\CG6644, UDPGT1, Ugt86Df, ugt35a', description='UDP-glycosyltransferase 35a', other_designations='CG6644-PA|Ugt35a-PA', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11170817, end_position_on_the_genomic_accession=11172664, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41337, CurrentID=0, Status='live', Symbol='CG18577', Aliases='Dmel_ Dmel\\CG18577', description='CG18577 gene product from transcript CG18577-RA', other_designations='CG18577-PA|uncharacterized protein', map_location='86D7-86D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11196092, end_position_on_the_genomic_accession=11197285, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41339, CurrentID=0, Status='live', Symbol='cu', Aliases='Dmel_CG31299, CG31299, CG4782, CG4796, Dmel\\CG31299, no, no(cu), noc, noct, nocturnin', description='curled', other_designations='CG31299-PC|CG31299-PD|CG31299-PE|CG31299-PF|CG31299-PG|CG31299-PH|cu-PC|cu-PD|cu-PE|cu-PF|cu-PG|cu-PH|dnocturnin|nocturin|nocturnin', map_location='3-50.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11197592, end_position_on_the_genomic_accession=11208633, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41340, CurrentID=0, Status='live', Symbol='SdhC', Aliases='Dmel_CG6666, BcDNA:RH04289, C560, CG6666, Dmel\\CG6666, RH04289p, SDH, sdhC', description='Succinate dehydrogenase, subunit C', other_designations='CG6666-PA|CG6666-PB|SdhC-PA|SdhC-PB|succinate dehydrogenase|succinate dehydrogenase C|succinate dehydrogenase cytochrome B560|succinate dehydrogenase cytochrome B560 subunit|succinate dehydrogenase, subunit C', map_location='86D7-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11208887, end_position_on_the_genomic_accession=11209849, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41341, CurrentID=0, Status='live', Symbol='Tctp', Aliases='Dmel_CG4800, CG4800, Dmel\\CG4800, TCTP, dTCTP', description='Translationally controlled tumor protein ortholog (H. sapiens)', other_designations='CG4800-PA|CG4800-PB|Tctp-PA|Tctp-PB|translationally controlled tumor protein|translationally controlled tumor protein ortholog', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11210247, end_position_on_the_genomic_accession=11211898, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41342, CurrentID=0, Status='live', Symbol='ZnT86D', Aliases='Dmel_CG6672, CG6672, Dmel\\CG6672, cg6672, dZnT7, dZnT86D', description='Zinc transporter 86D', other_designations='CG6672-PA|ZnT86D-PA|zinc transporter 86D', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11212308, end_position_on_the_genomic_accession=11214539, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41343, CurrentID=0, Status='live', Symbol='RpS25', Aliases='Dmel_CG6684, CG6684, D25, Dmel\\CG6684, M(3)86D, M(3)S31, M(3)S32, M(3)s[32], Rp S25, RpS31, S25, anon-EST:Posey260', description='Ribosomal protein S25', other_designations='CG6684-PA|CG6684-PB|RpS25-PA|RpS25-PB|minute (3) 86D|minute (3) S32|ribosomal protein S25', map_location='3-50.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11215028, end_position_on_the_genomic_accession=11216287, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41344, CurrentID=0, Status='live', Symbol='CG4820', Aliases='Dmel_ Dmel\\CG4820', description='CG4820 gene product from transcript CG4820-RB', other_designations='CG4820-PB', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11216535, end_position_on_the_genomic_accession=11217898, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41345, CurrentID=0, Status='live', Symbol='CG6689', Aliases='Dmel_ Dmel\\CG6689, dmCG6689', description='CG6689 gene product from transcript CG6689-RA', other_designations='CG6689-PA', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11217898, end_position_on_the_genomic_accession=11220064, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41346, CurrentID=0, Status='live', Symbol='CG6693', Aliases='Dmel_ Dmel\\CG6693', description='CG6693 gene product from transcript CG6693-RA', other_designations='CG6693-PA|CG6693-PB', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11220268, end_position_on_the_genomic_accession=11221639, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41347, CurrentID=0, Status='live', Symbol='RpL3', Aliases='Dmel_CG4863, CG4863, Dmel\\CG4863, L3, Rpl3, anon-EST:GressD4, chr3R:7047878..7048035, rpL3', description='Ribosomal protein L3', other_designations='CG4863-PA|CG4863-PD|CG4863-PG|CG4863-PH|RpL3-PA|RpL3-PD|RpL3-PG|RpL3-PH|ribosomal protein L3', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11221894, end_position_on_the_genomic_accession=11225173, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41348, CurrentID=0, Status='live', Symbol='scpr-C', Aliases='Dmel_CG5106, BEST:CK01680, CG5106, Dmel\\CG5106, SCPR-C, Scpr-C, cg5106, scp-C', description='SCP-containing protein C', other_designations='CG5106-PA|CG5106-PB|SCP-Related-C|scpr-C-PA|scpr-C-PB', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11225520, end_position_on_the_genomic_accession=11226681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41349, CurrentID=0, Status='live', Symbol='CG17726', Aliases='Dmel_ Dmel\\CG17726', description='CG17726 gene product from transcript CG17726-RA', other_designations='CG17726-PA|uncharacterized protein', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11226762, end_position_on_the_genomic_accession=11228227, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41350, CurrentID=0, Status='live', Symbol='GCC88', Aliases='Dmel_CG10703, CG10703, Dmel\\CG10703, dGCC88', description='GCC88 ortholog (H. sapiens)', other_designations='CG10703-PA|GCC88 ortholog|GCC88-PA', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11228158, end_position_on_the_genomic_accession=11230600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41351, CurrentID=0, Status='live', Symbol='CG17187', Aliases='Dmel_ BcDNA:RE47242, Dmel\\CG17187', description='CG17187 gene product from transcript CG17187-RA', other_designations='CG17187-PA|uncharacterized protein', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11230617, end_position_on_the_genomic_accession=11231879, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41352, CurrentID=0, Status='live', Symbol='CG14701', Aliases='Dmel_ Dmel\\CG14701', description='CG14701 gene product from transcript CG14701-RA', other_designations='CG14701-PA|uncharacterized protein', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11231912, end_position_on_the_genomic_accession=11232539, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41353, CurrentID=0, Status='live', Symbol='Arfip', Aliases='Dmel_CG17184, CG17184, Dmel\\CG17184, arfip, dCG17184-PA', description='Arfaptin', other_designations='Arfip-PA|CG17184-PA|arfaptin|arfaptin-2|partner of Rac1', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11232429, end_position_on_the_genomic_accession=11235013, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41354, CurrentID=0, Status='live', Symbol='CG17721', Aliases='Dmel_ BcDNA:LP09251, Dmel\\CG17721', description='CG17721 gene product from transcript CG17721-RA', other_designations='CG17721-PA|CG17721-PB|CG17721-PC|uncharacterized protein', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11235140, end_position_on_the_genomic_accession=11236077, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41355, CurrentID=0, Status='live', Symbol='CG31388', Aliases='Dmel_ BcDNA:GH10523, CG17722, CG5135, Dmel\\CG31388', description='CG31388 gene product from transcript CG31388-RB', other_designations='CG31388-PA|CG31388-PB|uncharacterized protein', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11237815, end_position_on_the_genomic_accession=11240114, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41357, CurrentID=0, Status='live', Symbol='scpr-B', Aliases='Dmel_CG17210, CG17210, Dmel\\CG17210, SCPR-B, anon-WO0140519.115, cg17210, scp-B', description='SCP-containing protein B', other_designations='CG17210-PA|SCP-Related-B|scpr-B-PA', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11240041, end_position_on_the_genomic_accession=11240991, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41358, CurrentID=0, Status='live', Symbol='scpr-A', Aliases='Dmel_CG5207, CG5207, Dmel\\CG5207, SCPR-A, bs03f04.y1, cg5207, scp-A', description='SCP-containing protein A', other_designations='CG5207-PA|SCP-Related-A|scpr-A-PA', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11241183, end_position_on_the_genomic_accession=11242120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41359, CurrentID=0, Status='live', Symbol='CG17734', Aliases='Dmel_ BcDNA:RH54117, Dmel\\CG17734', description='CG17734 gene product from transcript CG17734-RA', other_designations='CG17734-PA|CG17734-PB|CG17734-PC|uncharacterized protein', map_location='86D8-86D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11242208, end_position_on_the_genomic_accession=11243999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41360, CurrentID=0, Status='live', Symbol='CG5214', Aliases='Dmel_ Dmel\\CG5214, alpha-KGDHC', description='CG5214 gene product from transcript CG5214-RA', other_designations='CG5214-PA', map_location='86D9-86D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11244302, end_position_on_the_genomic_accession=11246862, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41361, CurrentID=0, Status='live', Symbol='KP78b', Aliases='Dmel_CG17216, CG17216, Dmel\\CG17216, KP78B, Kp78B, Kp78b', description='CG17216 gene product from transcript CG17216-RB', other_designations='CG17216-PA|CG17216-PB|KP78b-PA|KP78b-PB', map_location='86E2-86E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11346294, end_position_on_the_genomic_accession=11353131, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41362, CurrentID=0, Status='live', Symbol='KP78a', Aliases='Dmel_CG6715, CG6715, Dmel\\CG6715, KP78A, Kp78A', description='CG6715 gene product from transcript CG6715-RB', other_designations='CG6715-PB|KP78a-PB', map_location='86E2-86E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11346294, end_position_on_the_genomic_accession=11353131, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41363, CurrentID=0, Status='live', Symbol='pros', Aliases='Dmel_CG17228, 0244/09, 0320/10, 0441/16, 0451/09, 0563/18, 0585/13, 0664/07, 0671/02, 0763/13, 0989/01, 1135/07, 1135/09, 1167/13, 1316/02, 671/2, BcDNA:HL08040, CG17228, DMPROSPER, DROPROSA, Dmel\\CG17228, PROS, PROS-1, PROS-2, Pro, Pros, Prosp, Voila, anon-WO0140519.15, l(3)10419, l(3)j12C8, l(3)j6E2, l(3)rH013, l(3)rI160, l(3)rJ806, l(3)rK137, l(3)rK204, l(3)rL433, l(3)rO534, pro, voila', description='prospero', other_designations='CG17228-PH|CG17228-PI|CG17228-PJ|CG17228-PK|CG17228-PL|CG17228-PM|a la voile et a la vapeur|pros-PH|pros-PI|pros-PJ|pros-PK|pros-PL|pros-PM', map_location='86E2-86E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11328480, end_position_on_the_genomic_accession=11407627, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41364, CurrentID=0, Status='live', Symbol='mRpL40', Aliases='Dmel_CG5242, CG5242, Dmel\\CG5242, MRPL22(24), mRpL22-24', description='mitochondrial ribosomal protein L40', other_designations='CG5242-PA|mRpL40-PA|mitochondrial ribosomal protein L22/24', map_location='86E4-86E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11407918, end_position_on_the_genomic_accession=11408756, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41365, CurrentID=0, Status='live', Symbol='mgr', Aliases='Dmel_CG6719, BcDNA:LD34406, CG6719, Dmel\\CG6719, PAC10', description='merry-go-round', other_designations='CG6719-PA|mgr-PA', map_location='3-51.3 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11408792, end_position_on_the_genomic_accession=11409875, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41366, CurrentID=0, Status='live', Symbol='Blm', Aliases='Dmel_CG6920, BLM, CG6920, DmBLM, DmBlm, DmKu70, Dmblm, Dmel\\CG6920, Ku70, MUS309, Mus309, RecQ, blm, dBLM, dmblm, ku70, mus 309, mus309', description='Bloom syndrome helicase ortholog', other_designations="Blm-PA|Bloom's helicase|CG6920-PA|RecQ-like|bloom|bloom helicase|bloom syndrome helicase|bloom syndrome helicase ortholog|drosophila Blm|genome stabilizer 1|gesta1|mutagen-sensitive 309", map_location='86E17-86E17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11736810, end_position_on_the_genomic_accession=11741783, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41367, CurrentID=0, Status='live', Symbol='Ranbp9', Aliases='Dmel_CG5252, CG5252, Dmel\\CG5252', description='CG5252 gene product from transcript CG5252-RA', other_designations='CG5252-PA|CG5252-PD|Ranbp9-PA|Ranbp9-PD', map_location='86E4-86E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11412522, end_position_on_the_genomic_accession=11416142, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41368, CurrentID=0, Status='live', Symbol='CG6723', Aliases='Dmel_ Dmel\\CG6723, dSLC5A8', description='CG6723 gene product from transcript CG6723-RA', other_designations='CG6723-PA', map_location='86E4-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11416218, end_position_on_the_genomic_accession=11418668, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41369, CurrentID=0, Status='live', Symbol='ninaG', Aliases='Dmel_CG6728, CG6728, Dmel\\CG6728, NinaG', description='CG6728 gene product from transcript CG6728-RB', other_designations='CG6728-PA|CG6728-PB|ninaG-PA|ninaG-PB', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11419376, end_position_on_the_genomic_accession=11421693, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41370, CurrentID=0, Status='live', Symbol='CG5270', Aliases='Dmel_ Dmel\\CG5270', description='CG5270 gene product from transcript CG5270-RB', other_designations='CG5270-PB', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11422092, end_position_on_the_genomic_accession=11429969, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41371, CurrentID=0, Status='live', Symbol='CG18643', Aliases='Dmel_ Dmel\\CG18643', description='CG18643 gene product from transcript CG18643-RC', other_designations='CG18643-PB|CG18643-PC|uncharacterized protein', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11430061, end_position_on_the_genomic_accession=11430762, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41372, CurrentID=0, Status='live', Symbol='RpL24-like', Aliases='Dmel_CG6764, CG6764, Dmel\\CG6764', description='Ribosomal protein L24-like', other_designations='CG6764-PA|RpL24-like-PA|ribosomal protein L24-like', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11430740, end_position_on_the_genomic_accession=11431582, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41373, CurrentID=0, Status='live', Symbol='CG5276', Aliases='Dmel_ Dmel\\CG5276', description='CG5276 gene product from transcript CG5276-RA', other_designations='CG5276-PA|CG5276-PB', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11431738, end_position_on_the_genomic_accession=11433543, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41374, CurrentID=0, Status='live', Symbol='CG6744', Aliases='Dmel_ Dmel\\CG6744', description='CG6744 gene product from transcript CG6744-RA', other_designations='CG6744-PA|CG6744-PB', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11433488, end_position_on_the_genomic_accession=11435703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41375, CurrentID=0, Status='live', Symbol='CG5281', Aliases='Dmel_ Dmel\\CG5281', description='CG5281 gene product from transcript CG5281-RA', other_designations='CG5281-PA', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11435850, end_position_on_the_genomic_accession=11437523, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41376, CurrentID=0, Status='live', Symbol='CG17230', Aliases='Dmel_ BcDNA:GH24095, CG5286, Dmel\\CG17230', description='CG17230 gene product from transcript CG17230-RA', other_designations='CG17230-PA|CG17230-PB|CG17230-PC|uncharacterized protein', map_location='86E5-86E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11437785, end_position_on_the_genomic_accession=11467372, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41377, CurrentID=0, Status='live', Symbol='sals', Aliases='Dmel_CG31374, CG14703, CG31374, CG6771, Dmel\\CG31374', description='sarcomere length short', other_designations='CG31374-PA|CG31374-PB|CG31374-PC|CG31374-PD|CG31374-PE|CG31374-PF|CG31374-PG|CG31374-PH|sals-PA|sals-PB|sals-PC|sals-PD|sals-PE|sals-PF|sals-PG|sals-PH', map_location='86E5-86E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11437547, end_position_on_the_genomic_accession=11451791, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41379, CurrentID=0, Status='live', Symbol='PGRP-LB', Aliases='Dmel_CG14704, CG14704, Dmel\\CG14704, pgrp-LB', description='Peptidoglycan recognition protein LB', other_designations='CG14704-PA|CG14704-PB|CG14704-PC|CG14704-PD|CG14704-PE|CG14704-PF|PGRP-LB-PA|PGRP-LB-PB|PGRP-LB-PC|PGRP-LB-PD|PGRP-LB-PE|PGRP-LB-PF|Peptidoglycan-recognition protein-LB precursor|peptidoglycan recognition protein LB', map_location='86E6-86E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11452849, end_position_on_the_genomic_accession=11460552, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41381, CurrentID=0, Status='live', Symbol='dpr5', Aliases='Dmel_CG5308, CG5308, CT16895, Dmel\\CG5308, Dpr-5, Dpr5', description='defective proboscis extension response 5', other_designations='CG5308-PA|CG5308-PB|dpr5-PA|dpr5-PB', map_location='86E7-86E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11499984, end_position_on_the_genomic_accession=11505951, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  8674033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674033, CurrentID=0, Status='live', Symbol='CG42456', Aliases='Dmel_ Dmel\\CG42456, Dromel_CG12016_FBtr0073063_uORF', description='CG42456 gene product from transcript CG42456-RA', other_designations='CG42456-PA|uncharacterized protein', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3319555, end_position_on_the_genomic_accession=3322489, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41384, CurrentID=0, Status='live', Symbol='CG10898', Aliases='Dmel_ Dmel\\CG10898, anon-WO0118547.280', description='CG10898 gene product from transcript CG10898-RA', other_designations='CG10898-PA|uncharacterized protein', map_location='86E10-86E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11557776, end_position_on_the_genomic_accession=11559877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41387, CurrentID=0, Status='live', Symbol='Mrp4', Aliases='Dmel_CG14709, BEST:CK01227, BEST:CK02472, CG14709, CK01227, Dme_CG14709, Dmel\\CG14709, EP3221, dMRP4', description='Multidrug resistance protein 4 ortholog', other_designations='CG14709-PB|EP3221|Mrp4-PB|multidrug resistance protein 4 ortholog', map_location='86E11-86E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11569250, end_position_on_the_genomic_accession=11575937, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41388, CurrentID=0, Status='live', Symbol='CG6790', Aliases='Dmel_ Dmel\\CG6790, NEST:bs22e02, anon-WO0140519.186', description='CG6790 gene product from transcript CG6790-RA', other_designations='CG6790-PA', map_location='86E11-86E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11575901, end_position_on_the_genomic_accession=11578745, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41389, CurrentID=0, Status='live', Symbol='CG5342', Aliases='Dmel_ Dmel\\CG5342', description='CG5342 gene product from transcript CG5342-RB', other_designations='CG5342-PB', map_location='86E11-86E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11578826, end_position_on_the_genomic_accession=11581476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41390, CurrentID=0, Status='live', Symbol='wkd', Aliases='Dmel_CG5344, CG5344, Dmel\\CG5344, Wkd, cg5344', description='whacked', other_designations='CG5344-PA|CG5344-PB|wkd-PA|wkd-PB', map_location='86E11-86E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11581576, end_position_on_the_genomic_accession=11585305, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41391, CurrentID=0, Status='live', Symbol='CG6791', Aliases='Dmel_ Dmel\\CG6791', description='CG6791 gene product from transcript CG6791-RA', other_designations='CG6791-PA|CG6791-PB', map_location='86E11-86E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11585354, end_position_on_the_genomic_accession=11590146, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41392, CurrentID=0, Status='live', Symbol='Jupiter', Aliases='Dmel_CG31363, BEST:LP05168, BcDNA:GH06373, BcDNA:LD21358, BcDNA:LD43690, BcDNA:RH53211, CG17238, CG18762, CG18763, CG31363, Dmel\\CG31363, anon-EST:Liang-1.76, anon-WO0140519.173, clone 1.76, jupiter', description='CG31363 gene product from transcript CG31363-RI', other_designations='CG31363-PA|CG31363-PB|CG31363-PC|CG31363-PD|CG31363-PE|CG31363-PH|CG31363-PI|Jupiter-PA|Jupiter-PB|Jupiter-PC|Jupiter-PD|Jupiter-PE|Jupiter-PH|Jupiter-PI|jupiter', map_location='86E11-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11590386, end_position_on_the_genomic_accession=11619906, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41393, CurrentID=0, Status='live', Symbol='CG14710', Aliases='Dmel_ Dmel\\CG14710', description='CG14710 gene product from transcript CG14710-RB', other_designations='CG14710-PB|uncharacterized protein', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11620090, end_position_on_the_genomic_accession=11621728, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41394, CurrentID=0, Status='live', Symbol='CG6808', Aliases='Dmel_ CG31372, Dmel\\CG6808', description='CG6808 gene product from transcript CG6808-RA', other_designations='CG6808-PA|CG6808-PB', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11621466, end_position_on_the_genomic_accession=11625644, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41395, CurrentID=0, Status='live', Symbol='CG14711', Aliases='Dmel_ Dmel\\CG14711', description='CG14711 gene product from transcript CG14711-RA', other_designations='CG14711-PA|uncharacterized protein', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11623339, end_position_on_the_genomic_accession=11624948, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41396, CurrentID=0, Status='live', Symbol='CG6813', Aliases='Dmel_ CG31372, Dmel\\CG6813', description='CG6813 gene product from transcript CG6813-RD', other_designations='CG6813-PB|CG6813-PC|CG6813-PD', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11625051, end_position_on_the_genomic_accession=11626638, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41397, CurrentID=0, Status='live', Symbol='CG14712', Aliases='Dmel_ Dmel\\CG14712', description='CG14712 gene product from transcript CG14712-RA', other_designations='CG14712-PA|uncharacterized protein', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11628759, end_position_on_the_genomic_accession=11633206, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41398, CurrentID=0, Status='live', Symbol='Csk', Aliases='Dmel_CG42317, CG17309, CG32852, CG42317, CSK, DCsk, Dmel\\CG42317, Dmel_CG17309, Dmel_CG32852, csk, d-Csk, dCSK, dCsk, dcsk, flik, l(3)S017909, l(3)j1D8', description='C-terminal Src kinase', other_designations='C-Src kinase|C-terminal-Src kinase|CG42317-PF|CG42317-PG|CG42317-PH|CG42317-PI|CG42317-PJ|CG42317-PK|CG42317-PL|Csk-PF|Csk-PG|Csk-PH|Csk-PI|Csk-PJ|Csk-PK|Csk-PL|drosophila C-terminal Src kinase|flik|lethal (3) S017909|lethal (3) j1D8', map_location='86E13-86E', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11633278, end_position_on_the_genomic_accession=11650706, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  41399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41399, CurrentID=0, Status='live', Symbol='Elp1', Aliases='Dmel_CG10535, CG10535, D-elp1, Dmel\\CG10535, ELP1, elp1', description='Elongator complex protein 1', other_designations='CG10535-PA|CG10535-PB|CG10535-PC|Elp1-PA|Elp1-PB|Elp1-PC|elongator complex protein 1', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11650971, end_position_on_the_genomic_accession=11655691, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41401, CurrentID=0, Status='live', Symbol='CG42327', Aliases='Dmel_ CG14713, CG14714, CG18479, Dmel\\CG42327, Dmel_CG14713, Dmel_CG14714, anon-WO0118547.710', description='CG42327 gene product from transcript CG42327-RE', other_designations='CG42327-PE|uncharacterized protein', map_location='86E15-86E15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11656149, end_position_on_the_genomic_accession=11673974, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41403, CurrentID=0, Status='live', Symbol='CG14715', Aliases='Dmel_ BcDNA:GM09283, BcDNA:RH50927, Dmel\\CG14715', description='CG14715 gene product from transcript CG14715-RA', other_designations='CG14715-PA|uncharacterized protein', map_location='86E15-86E15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11674221, end_position_on_the_genomic_accession=11674923, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41404, CurrentID=0, Status='live', Symbol='CG18476', Aliases='Dmel_ CG17822, Dmel\\CG18476', description='CG18476 gene product from transcript CG18476-RA', other_designations='CG18476-PA|uncharacterized protein', map_location='86E15-86E15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11674920, end_position_on_the_genomic_accession=11678116, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41406, CurrentID=0, Status='live', Symbol='Taf12', Aliases='Dmel_CG17358, CG17358, Dmel\\CG17358, T2DA_DROME, TAF, TAF12, TAF30a, TAF30alpha, TAF30alpha/28, TAF30alpha/28/22, TAFII30alpha/22, TAF[II]30alpha, TAF[[II]], TAF[[II]]230, TAF[[II]]30alpha, TAF[[II]]30alpha/28 TAF[[II]]22, TAF[[II]]30alpha/28/22, TFIID, TFIID 22, TFIID 28, TFIID-p28, Taf22, Taf28, Taf30alpha, Taf32, Taf[[II]]30alpha, d28/22, dTAF12, dTAF22, dTAF28, dTAFII30, dTAFII30 alpha, dTAF[[II]]30, dTAF[[II]]30alpha, dTAF[[II]]30alpha/22, dmTAF12, p22, p28', description='TBP-associated factor 12', other_designations='26kDa polypeptide|CG17358-PA|CG17358-PB|CG17358-PD|CG17358-PE|TBP-associated factor|TBP-associated factor 30kD subunit alpha|Taf12-PA|Taf12-PB|Taf12-PD|Taf12-PE', map_location='86E15-86E15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11678386, end_position_on_the_genomic_accession=11679690, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41407, CurrentID=0, Status='live', Symbol='Ho', Aliases='Dmel_CG14716, CG14716, CT34507, DmHO, Dmel\\CG14716, HO, dHO', description='Heme oxygenase', other_designations='CG14716-PA|Ho-PA|haem oxygenase|heme oxygenase', map_location='86E15-86E15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11679743, end_position_on_the_genomic_accession=11680976, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41408, CurrentID=0, Status='live', Symbol='CG6830', Aliases='Dmel_ Dmel\\CG6830', description='CG6830 gene product from transcript CG6830-RA', other_designations='CG6830-PA', map_location='86E15-86E15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11680972, end_position_on_the_genomic_accession=11684342, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41409, CurrentID=0, Status='live', Symbol='CG6834', Aliases='Dmel_ Dmel\\CG6834', description='CG6834 gene product from transcript CG6834-RA', other_designations='CG6834-PA', map_location='86E15-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11684672, end_position_on_the_genomic_accession=11688133, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41410, CurrentID=0, Status='live', Symbol='CG6908', Aliases='Dmel_ Dmel\\CG6908', description='CG6908 gene product from transcript CG6908-RA', other_designations='CG6908-PA|CG6908-PB', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11688381, end_position_on_the_genomic_accession=11690051, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41411, CurrentID=0, Status='live', Symbol='Fer3', Aliases='Dmel_CG6913, BcDNA:LD04689, CG6913, Dmel\\CG6913, Dnato3, bHLHa32, nato3', description='48 related 3', other_designations='CG6913-PA|Fer3-PA|N-twist|nephew of atonal 3|ntwist', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11693410, end_position_on_the_genomic_accession=11694311, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41412, CurrentID=0, Status='live', Symbol='CG14717', Aliases='Dmel_ Dmel\\CG14717, anon-SAGE:Wang-126', description='CG14717 gene product from transcript CG14717-RB', other_designations='CG14717-PA|CG14717-PB|anon-SAGE:Wang-126|uncharacterized protein', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11696509, end_position_on_the_genomic_accession=11698327, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41413, CurrentID=0, Status='live', Symbol='CG14718', Aliases='Dmel_ Dmel\\CG14718, cg14718', description='CG14718 gene product from transcript CG14718-RA', other_designations='CG14718-PA|uncharacterized protein', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11700614, end_position_on_the_genomic_accession=11702720, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41414, CurrentID=0, Status='live', Symbol='I-t', Aliases='Dmel_CG14719, CG14719, Dmel\\CG14719, pp1d6', description='inhibitor-t', other_designations='CG14719-PA|I-t-PA', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11704030, end_position_on_the_genomic_accession=11704916, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41415, CurrentID=0, Status='live', Symbol='CG14720', Aliases='Dmel_ Dmel\\CG14720', description='CG14720 gene product from transcript CG14720-RA', other_designations='CG14720-PA|uncharacterized protein', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11711718, end_position_on_the_genomic_accession=11712371, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41416, CurrentID=0, Status='live', Symbol='CG12594', Aliases='Dmel_ Dmel\\CG12594', description='CG12594 gene product from transcript CG12594-RB', other_designations='CG12594-PA|CG12594-PB|uncharacterized protein', map_location='86E16-86E17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11721734, end_position_on_the_genomic_accession=11732479, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41417, CurrentID=0, Status='live', Symbol='CG14721', Aliases='Dmel_ Dmel\\CG14721', description='CG14721 gene product from transcript CG14721-RA', other_designations='CG14721-PA|CG14721-PB|uncharacterized protein', map_location='86E17-86E17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11735447, end_position_on_the_genomic_accession=11736877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41419, CurrentID=0, Status='live', Symbol='CG14722', Aliases='Dmel_ Dmel\\CG14722', description='CG14722 gene product from transcript CG14722-RA', other_designations='CG14722-PA|uncharacterized protein', map_location='86E17-86E17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11741890, end_position_on_the_genomic_accession=11744027, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41420, CurrentID=0, Status='live', Symbol='CG6923', Aliases='Dmel_ Dmel\\CG6923', description='CG6923 gene product from transcript CG6923-RA', other_designations='CG6923-PA|CG6923-PB|CG6923-PC|CG6923-PD', map_location='86E17-86E17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11744229, end_position_on_the_genomic_accession=11750029, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41423, CurrentID=0, Status='live', Symbol='l(3)neo38', Aliases='Dmel_CG6930, BcDNA:SD16005, CG14725, CG31364, CG6930, Dmel\\CG6930, anon-EST:fe2F4, anon2F4', description='lethal (3) neo38', other_designations='CG6930-PA|CG6930-PB|CG6930-PC|CG6930-PD|CG6930-PF|CG6930-PG|CG6930-PI|CG6930-PJ|CG6930-PK|CG6930-PL|anon-fast-evolving-2F4|l(3)neo38-PA|l(3)neo38-PB|l(3)neo38-PC|l(3)neo38-PD|l(3)neo38-PF|l(3)neo38-PG|l(3)neo38-PI|l(3)neo38-PJ|l(3)neo38-PK|l(3)neo38-PL', map_location='86E18-86E18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11765038, end_position_on_the_genomic_accession=11779128, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41425, CurrentID=0, Status='live', Symbol='CG17360', Aliases='Dmel_ BcDNA:RH29885, Dmel\\CG17360', description='CG17360 gene product from transcript CG17360-RA', other_designations='CG17360-PA|CG17360-PB|uncharacterized protein', map_location='86E18-86F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11781131, end_position_on_the_genomic_accession=11787014, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  41426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41426, CurrentID=0, Status='live', Symbol='HisCl1', Aliases='Dmel_CG14723, CG14723, CT34515, DMHisClalpha2, Dm-HCLB, DmHisCl1, Dmel\\CG14723, HA-CI II, HisCl-alpha2, hclB, hisCl1', description='Histamine-gated chloride channel subunit 1', other_designations='CG14723-PA|CG14723-PB|CG14723-PC|CG14723-PD|HisCl1-PA|HisCl1-PB|HisCl1-PC|HisCl1-PD|glycine receptor alpha3 subunit-like|histamine-gated chloride channel subunit 1', map_location='86F7-86F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11787176, end_position_on_the_genomic_accession=11789393, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41427, CurrentID=0, Status='live', Symbol='Sbf', Aliases='Dmel_CG6939, CG6939, DSbf1, Dmel\\CG6939, SBF, SBF11, dSbf1, sbf, Sbf', description='SET domain binding factor', other_designations='CG6939-PA|CG6939-PB|Sbf-PA|Sbf-PB', map_location='86F8-86F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11789293, end_position_on_the_genomic_accession=11796948, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41428, CurrentID=0, Status='live', Symbol='ClC-a', Aliases='Dmel_CG31116, CG14726, CG31116, CG6942, CLC-2, Clc-a, DmClC-2, DmClC-a, DmClCa, Dmel\\CG31116, FBgn0037951', description='Chloride channel-a', other_designations='CG31116-PA|CG31116-PC|CG31116-PD|CG31116-PF|CG31116-PG|CG31116-PH|CG31116-PI|ClC-a-PA|ClC-a-PC|ClC-a-PD|ClC-a-PF|ClC-a-PG|ClC-a-PH|ClC-a-PI|chloride channel-a', map_location='86F8-86F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11797244, end_position_on_the_genomic_accession=11813207, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  41431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41431, CurrentID=0, Status='live', Symbol='glo', Aliases='Dmel_CG6946, 110/46, CG6946, Dmel\\CG6946, Glo, anon-WO0118547.328, cg6946, l(3)S011046, l(3)S11046, p67, p67-K', description='glorund', other_designations='CG6946-PA|CG6946-PB|CG6946-PC|glo-PA|glo-PB|glo-PC|p67-K', map_location='86F8-86F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11815206, end_position_on_the_genomic_accession=11819339, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41432, CurrentID=0, Status='live', Symbol='COX5A', Aliases='Dmel_CG14724, CG14724, COX, COXA, COXA_DROME, CoVa, Dmel\\CG14724, Va, cova, tend', description='Cytochrome c oxidase subunit 5A', other_designations='CG14724-PA|CG14724-PB|CG14724-PC|CG14724-PD|COX5A-PA|COX5A-PB|COX5A-PC|COX5A-PD|cytochrome c oxidase Va|cytochrome c oxidase Va, mt precursor|cytochrome c oxidase polypeptide Va|cytochrome c oxidase subunit 5A|cytochrome c oxidase subunit Va|cytochrome c oxidase subunit Va preprotein|cytochrome oxidase Va|cytochrome oxidase subunit Va|cytochrome-c oxidase subunit Va|subunit Va|subunit Va of complex IV|tenured', map_location='86F9-86F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11819359, end_position_on_the_genomic_accession=11820299, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41433, CurrentID=0, Status='live', Symbol='CG6950', Aliases='Dmel_ Dmel\\CG6950', description='CG6950 gene product from transcript CG6950-RC', other_designations='CG6950-PA|CG6950-PB|CG6950-PC|CG6950-PD', map_location='86F9-86F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11820381, end_position_on_the_genomic_accession=11822493, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41434, CurrentID=0, Status='live', Symbol='CG6959', Aliases='Dmel_ 6959, CT21563, Dmel\\CG6959', description='CG6959 gene product from transcript CG6959-RA', other_designations='CG6959-PA|CG6959-PC|connectin-like', map_location='86F9-86F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11822702, end_position_on_the_genomic_accession=11869460, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41436, CurrentID=0, Status='live', Symbol='CG6962', Aliases='Dmel_ Dmel\\CG6962', description='CG6962 gene product from transcript CG6962-RA', other_designations='CG6962-PA', map_location='87A1-87A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11879702, end_position_on_the_genomic_accession=11882534, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41437, CurrentID=0, Status='live', Symbol='CG31368', Aliases='Dmel_ CG14729, CG14730, Dmel\\CG31368', description='CG31368 gene product from transcript CG31368-RD', other_designations='CG31368-PC|CG31368-PD|uncharacterized protein', map_location='87A1-87A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11882790, end_position_on_the_genomic_accession=11892382, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41438, CurrentID=0, Status='live', Symbol='mthl5', Aliases='Dmel_CG6965, CG6965, Dmel\\CG6965, Mth-5, anon-WO0170980.70, anon-WO0170980.71', description='methuselah-like 5', other_designations='CG6965-PA|CG6965-PB|Mth-like 5|mthl5-PA|mthl5-PB', map_location='87A1-87A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11883938, end_position_on_the_genomic_accession=11887168, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41440, CurrentID=0, Status='live', Symbol='CG6971', Aliases='Dmel_ DNALI1, Dmel\\CG6971, anon-WO0140519.67', description='CG6971 gene product from transcript CG6971-RA', other_designations='CG6971-PA', map_location='87A1-87A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11892306, end_position_on_the_genomic_accession=11893297, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41441, CurrentID=0, Status='live', Symbol='Cad87A', Aliases='Dmel_CG6977, CG6977, CT21611, Cad87, DCad87A, Dmel\\CG6977', description='Cadherin 87A', other_designations='CG6977-PA|Cad87A-PA|cadherin 87A', map_location='87A1-87A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11893512, end_position_on_the_genomic_accession=11949129, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  41442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41442, CurrentID=0, Status='live', Symbol='CG14731', Aliases='Dmel_ Dmel\\CG14731, gi21483202, gi7299531', description='CG14731 gene product from transcript CG14731-RA', other_designations='CG14731-PA|uncharacterized protein', map_location='87A2-87A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11931262, end_position_on_the_genomic_accession=11933640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41443, CurrentID=0, Status='live', Symbol='CG31211', Aliases='Dmel_ CG14732, CG18546, Dmel\\CG31211, anon-WO03070958.3, cDNA', description='CG31211 gene product from transcript CG31211-RB', other_designations='CG31211-PA|CG31211-PB|CG31211-PC|uncharacterized protein', map_location='87A2-87A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11949680, end_position_on_the_genomic_accession=11954038, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41445, CurrentID=0, Status='live', Symbol='CG3281', Aliases='Dmel_ 87A7-9/1, Dmel\\CG3281, anon-87Aa, anon-WO0140519.243', description='CG3281 gene product from transcript CG3281-RA', other_designations='CG3281-PA|anon-87Aa', map_location='87A7-87A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11960883, end_position_on_the_genomic_accession=11962805, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41446, CurrentID=0, Status='live', Symbol='aurA', Aliases='Dmel_CG3068, 87A7-9/2, Aur-A, AurA, CG3068, Dmel\\CG3068, aur, aur-A, ck10, ck[10], l(3)87Ac, l(3)ck10', description='aurora A', other_designations='A-type aurora kinase|Aurora-A kinase|AuroraA|CG3068-PA|aurA-PA|aurora|aurora A kinase|aurora kinase|aurora-A|early-A|fruhe2', map_location='3-53 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11963124, end_position_on_the_genomic_accession=11964826, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41447, CurrentID=0, Status='live', Symbol='CG12213', Aliases='Dmel_ 87A7-9/3, Dmel\\CG12213, anon-87Ab', description='CG12213 gene product from transcript CG12213-RD', other_designations='CG12213-PB|CG12213-PC|CG12213-PD|CG12213-PE|anon-87Ab|uncharacterized protein', map_location='87A7-87A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11964674, end_position_on_the_genomic_accession=11967691, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41448, CurrentID=0, Status='live', Symbol='GC1', Aliases='Dmel_CG18347, 87A7-9/4, CG18347, DmGC1, Dmel\\CG18347, anon-87Ac', description='Glutamate Carrier 1', other_designations='CG18347-PA|CG18347-PB|GC1-PA|GC1-PB|anon-87Ac|glutamate carrier 1', map_location='87A7-87A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11967708, end_position_on_the_genomic_accession=11971686, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41449, CurrentID=0, Status='live', Symbol='GC2', Aliases='Dmel_CG12201, CG12201, DmGC2, Dmel\\CG12201', description='Glutamate Carrier 2', other_designations='CG12201-PB|CG12201-PD|GC2-PB|GC2-PD|glutamate carrier 2', map_location='87A4-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11971694, end_position_on_the_genomic_accession=11972896, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41450, CurrentID=0, Status='live', Symbol='Tango9', Aliases='Dmel_CG10007, 87A7-9/5, CG10007, Dmel\\CG10007, TANGO9, anon-87Ad', description='Transport and Golgi organization 9', other_designations='CG10007-PA|CG10007-PB|Tango9-PA|Tango9-PB|anon-87Ad|transport and golgi organization 9', map_location='87A7-87A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11971007, end_position_on_the_genomic_accession=11974910, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41451, CurrentID=0, Status='live', Symbol='CG10005', Aliases='Dmel_ BcDNA:RE59626, Dmel\\CG10005', description='CG10005 gene product from transcript CG10005-RA', other_designations='CG10005-PA|CG10005-PB|uncharacterized protein', map_location='87A4-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11975156, end_position_on_the_genomic_accession=11977197, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41452, CurrentID=0, Status='live', Symbol='CG18547', Aliases='Dmel_ Dmel\\CG18547', description='CG18547 gene product from transcript CG18547-RA', other_designations='CG18547-PA|uncharacterized protein', map_location='87A4-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11978574, end_position_on_the_genomic_accession=11980390, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41453, CurrentID=0, Status='live', Symbol='CG12224', Aliases='Dmel_ Dmel\\CG12224', description='CG12224 gene product from transcript CG12224-RC', other_designations='CG12224-PB|CG12224-PC|uncharacterized protein', map_location='87A4-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11980704, end_position_on_the_genomic_accession=11982622, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41454, CurrentID=0, Status='live', Symbol='CG3397', Aliases='Dmel_ Dmel\\CG3397', description='CG3397 gene product from transcript CG3397-RA', other_designations='CG3397-PA', map_location='87A4-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11983098, end_position_on_the_genomic_accession=11984541, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41455, CurrentID=0, Status='live', Symbol='mfas', Aliases='Dmel_CG3359, 87A7-9/7, BEST:GH06752, CG3359, Dmel\\CG3359, Mfas, anon-87Af, anon-EST:Liang-97, anon-WO0140519.174, clone 97, gh06752', description='midline fasciclin', other_designations='CG3359-PB|CG3359-PC|CG3359-PD|CG3359-PE|CG3359-PF|CG3359-PG|CG3359-PH|CG3359-PI|CG3359-PJ|CG3359-PK|CG3359-PL|CG3359-PM|CG3359-PN|CG3359-PO|CG3359-PP|CG3359-PQ|anon-87Af|mfas-PB|mfas-PC|mfas-PD|mfas-PE|mfas-PF|mfas-PG|mfas-PH|mfas-PI|mfas-PJ|mfas-PK|mfas-PL|mfas-PM|mfas-PN|mfas-PO|mfas-PP|mfas-PQ', map_location='87A7-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11984862, end_position_on_the_genomic_accession=11993315, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41456, CurrentID=0, Status='live', Symbol='Tk', Aliases='Dmel_CG14734, CG14734, DTK, DTK-1, DTK-2, DTK-3, DTK-4, DTK-5, DTK-6, DTK1, DmTK6, Dmel\\CG14734, Drm-TK, Drm-TK-1, Drm-TK-2, Drm-TK-3, Drm-TK-4, Drm-TK-5, Drm-TK-6, Dtk, TAP-5, TK, TK-1, Tyk, dtk', description='Tachykinin', other_designations='CG14734-PA|CG14734-PB|Tachykinin-related|Tk-PA|Tk-PB|neurokinin|preprotachykinin|tachykinin|tachykinin-1|tachykinin-3|tachykinin-6|tachykinin-like peptide', map_location='87A4-87A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11994175, end_position_on_the_genomic_accession=12005473, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41457, CurrentID=0, Status='live', Symbol='Ect3', Aliases='Dmel_CG3132, 87A7-9/8, CG3132, Dmel\\CG3132, anon-87Ag', description='Ectoderm-expressed 3', other_designations='CG3132-PA|Ect3-PA|anon-87Ag|beta-galactosidase', map_location='87A7-87A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11999278, end_position_on_the_genomic_accession=12003516, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41458, CurrentID=0, Status='live', Symbol='KLHL18', Aliases='Dmel_CG3571, CG3571, Dmel\\CG3571, anon-WO0118547.179, d', description='CG3571 gene product from transcript CG3571-RC', other_designations='CG3571-PA|CG3571-PB|CG3571-PC|KLHL18-PA|KLHL18-PB|KLHL18-PC', map_location='87A4-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12007476, end_position_on_the_genomic_accession=12010434, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41459, CurrentID=0, Status='live', Symbol='GCC185', Aliases='Dmel_CG3532, CG3532, Dmel\\CG3532, dGCC185', description='GCC185 ortholog (H. sapiens)', other_designations='CG3532-PB|GCC185 ortholog|GCC185-PB', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12010772, end_position_on_the_genomic_accession=12015059, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41460, CurrentID=0, Status='live', Symbol='CG3313', Aliases='Dmel_ Dmel\\CG3313', description='CG3313 gene product from transcript CG3313-RB', other_designations='CG3313-PA|CG3313-PB', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12014943, end_position_on_the_genomic_accession=12017031, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41461, CurrentID=0, Status='live', Symbol='Spt3', Aliases='Dmel_CG3169, CG3169, Dmel\\CG3169, SPT3, TFIID, d', description='CG3169 gene product from transcript CG3169-RA', other_designations='CG3169-PA|Spt3-PA', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12017156, end_position_on_the_genomic_accession=12018550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41462, CurrentID=0, Status='live', Symbol='CG31358', Aliases='Dmel_ CG17424, CG18141, CG18142, Dmel\\CG31358', description='CG31358 gene product from transcript CG31358-RB', other_designations='CG31358-PA|CG31358-PB|uncharacterized protein', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12027736, end_position_on_the_genomic_accession=12031757, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41465, CurrentID=0, Status='live', Symbol='ssp5', Aliases='Dmel_CG14735, CG14735, Dmel\\CG14735, Ssp5, Tes59, anon-WO0140519.127', description='short spindle 5', other_designations='CG14735-PA|ssp5-PA', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12033142, end_position_on_the_genomic_accession=12035525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41466, CurrentID=0, Status='live', Symbol='CG14736', Aliases='Dmel_ Dmel\\CG14736', description='CG14736 gene product from transcript CG14736-RE', other_designations='CG14736-PD|CG14736-PE|CG14736-PF|CG14736-PG|uncharacterized protein', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12040528, end_position_on_the_genomic_accession=12043453, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41467, CurrentID=0, Status='live', Symbol='CG14739', Aliases='Dmel_ Dmel\\CG14739', description='CG14739 gene product from transcript CG14739-RA', other_designations='CG14739-PA|uncharacterized protein', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12048253, end_position_on_the_genomic_accession=12049213, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41468, CurrentID=0, Status='live', Symbol='CG14740', Aliases='Dmel_ CS, Dmel\\CG14740, anon-WO0140519.178', description='CG14740 gene product from transcript CG14740-RA', other_designations='CG14740-PA|uncharacterized protein', map_location='87A6-87A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12061070, end_position_on_the_genomic_accession=12062944, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41469, CurrentID=0, Status='live', Symbol='ATP8B', Aliases='Dmel_CG14741, CG14741, DmCG14741, Dmel\\CG14741, dATP8B', description='ATPase 8B', other_designations='ATP8B-PB|ATP8B-PC|ATP8B-PD|ATP8B-PF|ATP8B-PG|ATP8B-PH|ATP8B-PI|CG14741-PB|CG14741-PC|CG14741-PD|CG14741-PF|CG14741-PG|CG14741-PH|CG14741-PI', map_location='87A7-87A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12068375, end_position_on_the_genomic_accession=12090612, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  41470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41470, CurrentID=0, Status='live', Symbol='dpr17', Aliases='Dmel_CG31361, CG14737, CG14738, CG31361, CT34531, Dmel\\CG31361, Dpr-17, Dpr17', description='defective proboscis extension response 17', other_designations='CG31361-PA|CG31361-PB|CG31361-PC|CG31361-PD|CG31361-PE|dpr17-PA|dpr17-PB|dpr17-PC|dpr17-PD|dpr17-PE', map_location='87A8-87A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12094876, end_position_on_the_genomic_accession=12111420, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41472, CurrentID=0, Status='live', Symbol='CG4702', Aliases='Dmel_ 4702, Dmel\\CG4702', description='CG4702 gene product from transcript CG4702-RC', other_designations='CG4702-PA|CG4702-PB|CG4702-PC|CG4702-PD', map_location='87A9-87A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12121583, end_position_on_the_genomic_accession=12128201, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41473, CurrentID=0, Status='live', Symbol='dpr15', Aliases='Dmel_CG10095, CG10095, CT28423, Dmel\\CG10095, Dpr-15, Dpr15', description='defective proboscis extension response 15', other_designations='CG10095-PA|CG10095-PB|dpr15-PA|dpr15-PB', map_location='87A10-87B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12138833, end_position_on_the_genomic_accession=12158595, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41474, CurrentID=0, Status='live', Symbol='Cpn', Aliases='Dmel_CG4795, 77H5, Ag 72H5, Ag72H5, CG4795, Dmel\\CG4795, Mab77H5, cap, cpn', description='Calphotin', other_designations='CG4795-PA|CG4795-PB|Cpn-PA|Cpn-PB|calcium-binding protein|calphotin|monoclonal antibody 77H5', map_location='87A9-87B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12162887, end_position_on_the_genomic_accession=12166186, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41475, CurrentID=0, Status='live', Symbol='CG4810', Aliases='Dmel_ 149735_at, Dmel\\CG4810, eIF3-S7', description='CG4810 gene product from transcript CG4810-RA', other_designations='CG4810-PA', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12192703, end_position_on_the_genomic_accession=12195011, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41476, CurrentID=0, Status='live', Symbol='CG3809', Aliases='Dmel_ DmAdk1, Dmel\\CG3809, anon-WO0140519.193', description='CG3809 gene product from transcript CG3809-RA', other_designations='CG3809-PA', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12197941, end_position_on_the_genomic_accession=12199473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41477, CurrentID=0, Status='live', Symbol='CG4830', Aliases='Dmel_ Dmel\\CG4830', description='CG4830 gene product from transcript CG4830-RA', other_designations='CG4830-PA', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12208159, end_position_on_the_genomic_accession=12210360, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41479, CurrentID=0, Status='live', Symbol='CG4848', Aliases='Dmel_ Dmel\\CG4848', description='CG4848 gene product from transcript CG4848-RC', other_designations='CG4848-PA|CG4848-PB|CG4848-PC', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12213792, end_position_on_the_genomic_accession=12219686, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41480, CurrentID=0, Status='live', Symbol='CG4860', Aliases='Dmel_ Dmel\\CG4860', description='CG4860 gene product from transcript CG4860-RA', other_designations='CG4860-PA', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12219881, end_position_on_the_genomic_accession=12221464, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41481, CurrentID=0, Status='live', Symbol='CG10014', Aliases='Dmel_ Dmel\\CG10014', description='CG10014 gene product from transcript CG10014-RA', other_designations='CG10014-PA|uncharacterized protein', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12222387, end_position_on_the_genomic_accession=12224048, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41482, CurrentID=0, Status='live', Symbol='CG17404', Aliases='Dmel_ Dmel\\CG17404, SP149', description='CG17404 gene product from transcript CG17404-RB', other_designations='CG17404-PB|uncharacterized protein', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12223991, end_position_on_the_genomic_accession=12225019, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41483, CurrentID=0, Status='live', Symbol='Su(var)3-9', Aliases='Dmel_CG43664, CG43664, CG6476, Dmel\\CG43664, Dmel_CG6476, E(var)3-2, SU(VAR)3-9, SUV39, SUV39H1, Su(VAR)3-9, Su(Var)3-9, Su(var), Su(var) 3-902, Su(var)306, Su(var)310, Su(var)328, Su-var(3)9, Su-var(3)902, SuVar3-9, Suv 3-9, Suvar(3)9, Suvar3-9, Suvar39, caa56376 Suvar39 Dm, dSU(VAR)3-9, dSu(var)3-9, ptn, su(var)3-9, Su(var)3-9', description='Suppressor of variegation 3-9', other_designations='CG43664-PA|Su(var)3-9-PA|enhancer of variegation 3-2|histone methyltransferase|pitkin|suppressor of variegation 3-9|suppressor of variegation 306|suppressor of variegation 310|suppressor of variegation 328', map_location='3-56.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15259864, end_position_on_the_genomic_accession=15263458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41484, CurrentID=0, Status='live', Symbol='CG3916', Aliases='Dmel_ Dmel\\CG3916, SP154', description='CG3916 gene product from transcript CG3916-RA', other_designations='CG3916-PA', map_location='87B3-87B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12229413, end_position_on_the_genomic_accession=12230425, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674050, CurrentID=0, Status='live', Symbol='CG42504', Aliases='Dmel_ Dmel\\CG42504', description='CG42504 gene product from transcript CG42504-RA', other_designations='CG42504-PA|uncharacterized protein', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12028070, end_position_on_the_genomic_accession=12031757, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41486, CurrentID=0, Status='live', Symbol='Cyp313a5', Aliases='Dmel_CG15807, 313a5, CG15807, CYP313a5, Dmel\\CG15807', description='CG15807 gene product from transcript CG15807-RA', other_designations='CG15807-PA|Cyp313a5-PA', map_location='87B3-87B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12232794, end_position_on_the_genomic_accession=12234684, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41487, CurrentID=0, Status='live', Symbol='Cyp313a2', Aliases='Dmel_CG10094, 313a2, CG10094, CYP313A2, Cyp313A2, Dmel\\CG10094', description='CG10094 gene product from transcript CG10094-RB', other_designations='CG10094-PB|Cyp313a2-PB', map_location='87B3-87B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12234980, end_position_on_the_genomic_accession=12236904, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41488, CurrentID=0, Status='live', Symbol='Cyp313a3', Aliases='Dmel_CG10093, 313a3, CG10093, Dmel\\CG10093', description='CG10093 gene product from transcript CG10093-RA', other_designations='CG10093-PA|Cyp313a3-PA', map_location='87B3-87B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12237138, end_position_on_the_genomic_accession=12239266, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41489, CurrentID=0, Status='live', Symbol='CG3942', Aliases='Dmel_ Dmel\\CG3942', description='CG3942 gene product from transcript CG3942-RA', other_designations='CG3942-PA', map_location='87B4-87B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12254187, end_position_on_the_genomic_accession=12256592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41490, CurrentID=0, Status='live', Symbol='CG17738', Aliases='Dmel_ Dmel\\CG17738', description='CG17738 gene product from transcript CG17738-RA', other_designations='CG17738-PA|uncharacterized protein', map_location='87B4-87B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12270890, end_position_on_the_genomic_accession=12271300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41491, CurrentID=0, Status='live', Symbol='svp', Aliases='Dmel_CG11502, 0005/12, 0042/01, 0106/18, 0114/20, 0351/01, 0598/01, 0861/07, BcDNA:GH08189, CG11502, CG18158, Dmel\\CG11502, NR2F3, SVP, SVP-1, SVP1, Svp, ck16, don, l(3)07842, l(3)87Bd, l(3)ck16, l(3)j2E2, l(3)rA028, l(3)rL069', description='seven up', other_designations='CG11502-PA|CG11502-PB|CG11502-PC|CG11502-PE|SEVEN-UP|complementation group 7|down-and-out|seven Up|sevenup|svp-PA|svp-PB|svp-PC|svp-PE', map_location='87B4-87B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12258737, end_position_on_the_genomic_accession=12302785, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41493, CurrentID=0, Status='live', Symbol='CG4066', Aliases='Dmel_ Dmel\\CG4066', description='CG4066 gene product from transcript CG4066-RA', other_designations='CG4066-PA', map_location='87B6-87B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12311353, end_position_on_the_genomic_accession=12313119, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41494, CurrentID=0, Status='live', Symbol='CG10013', Aliases='Dmel_ Dmel\\CG10013', description='CG10013 gene product from transcript CG10013-RA', other_designations='CG10013-PA|uncharacterized protein', map_location='87B6-87B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12323521, end_position_on_the_genomic_accession=12325210, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41495, CurrentID=0, Status='live', Symbol='CG10038', Aliases='Dmel_ Dmel\\CG10038', description='CG10038 gene product from transcript CG10038-RB', other_designations='CG10038-PA|CG10038-PB|uncharacterized protein', map_location='87B7-87B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12348987, end_position_on_the_genomic_accession=12350644, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41496, CurrentID=0, Status='live', Symbol='CG10041', Aliases='Dmel_ Dmel\\CG10041, SP178', description='CG10041 gene product from transcript CG10041-RA', other_designations='CG10041-PA|uncharacterized protein', map_location='87B7-87B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12350815, end_position_on_the_genomic_accession=12352079, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41498, CurrentID=0, Status='live', Symbol='MBD-R2', Aliases='Dmel_CG10042, CG10042, CG17393, Dm Tam3, Dmel\\CG10042, MBD, MBDR2, Mbdr2, Mbrd2, TAM3, d mbd-R2', description='CG10042 gene product from transcript CG10042-RB', other_designations='CG10042-PB|MBD-R2-PB', map_location='87B7-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12352143, end_position_on_the_genomic_accession=12356265, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41499, CurrentID=0, Status='live', Symbol='CG4115', Aliases='Dmel_ Dmel\\CG4115', description='CG4115 gene product from transcript CG4115-RA', other_designations='CG4115-PA', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12360008, end_position_on_the_genomic_accession=12363255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41500, CurrentID=0, Status='live', Symbol='Tim17a1', Aliases='Dmel_CG10090, CG10090, Dmel\\CG10090', description='CG10090 gene product from transcript CG10090-RA', other_designations='CG10090-PA|CG10090-PB|Tim17a1-PA|Tim17a1-PB', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12363667, end_position_on_the_genomic_accession=12364680, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41502, CurrentID=0, Status='live', Symbol='GstD9', Aliases='Dmel_CG10091, CG10091, DmGSTD9, DmGstd9, Dmel\\CG10091, GST-D9, GSTD9, gstD9, gstd9', description='Glutathione S transferase D9', other_designations='CG10091-PA|CG10091-PB|GstD9-PA|GstD9-PB|glutathione S transferase D9', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12366163, end_position_on_the_genomic_accession=12367564, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41503, CurrentID=0, Status='live', Symbol='GstD1', Aliases='Dmel_CG10045, CG10045, D1, DmGST-1, DmGST1, DmGSTD1, DmGSTD1-1, DmGst1, Dmel\\CG10045, G-S-T, GST, GST-1, GST-D1, GST1, GST1-1, GSTD1, GSTase, GSt1-1, Gst, Gst-D1, Gst1, Gst1-1, Gst1.1, GstD, T5, dmGSTD1, gst-D, gst-D1-1, gstD, gstD1, gstd1', description='Glutathione S transferase D1', other_designations='CG10045-PA|CG10045-PB|Glutathione-S-transferase 1|GstD1-PA|GstD1-PB|glutathione S transferase D1|glutathione S-transferase 1|glutathione S-transferase D1|glutathione S-transferase-D1', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12367817, end_position_on_the_genomic_accession=12369268, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41511, CurrentID=0, Status='live', Symbol='CG10035', Aliases='Dmel_ Dmel\\CG10035, anon-EST:Liang-20, clone 20', description='CG10035 gene product from transcript CG10035-RB', other_designations='CG10035-PA|CG10035-PB|uncharacterized protein', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12380869, end_position_on_the_genomic_accession=12382691, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41512, CurrentID=0, Status='live', Symbol='GstD11', Aliases='Dmel_CG17639, BcDNA:RE72392, CG17639, DmGSTD11, Dmel\\CG17639, GSTD11, cg17639, gstd11', description='Glutathione S transferase D11', other_designations='CG17639-PA|CG17639-PB|GstD11-PA|GstD11-PB|glutathione S transferase D11', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12383336, end_position_on_the_genomic_accession=12385624, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41513, CurrentID=0, Status='live', Symbol='CG34402', Aliases='Dmel_ CG17756, CG17793, CG31218, CG4940, CT39356, Dmel\\CG34402, Dmel_CG17793, Dmel_CG31218, Dmel_CG4940, SOL-1, sol-1', description='CG34402 gene product from transcript CG34402-RC', other_designations='CG34402-PC|CG34402-PD|CG34402-PE|suppressor of lurcher movement defect-1|uncharacterized protein', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12386249, end_position_on_the_genomic_accession=12402902, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318215, CurrentID=0, Status='live', Symbol='CG32797', Aliases='Dmel_ Dmel\\CG32797', description='CG32797 gene product from transcript CG32797-RA', other_designations='CG32797-PA|uncharacterized protein', map_location='3A2-3A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2424925, end_position_on_the_genomic_accession=2425855, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41518, CurrentID=0, Status='live', Symbol='lig3', Aliases='Dmel_CG17227, CG17227, Dmel\\CG17227, LIG3, Lig3, lig 3', description='DNA ligase III', other_designations='CG17227-PA|lig3-PA|ligase III', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12399011, end_position_on_the_genomic_accession=12401839, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41520, CurrentID=0, Status='live', Symbol='Cyp9f2', Aliases='Dmel_CG11466, 9f2, CG11466, Cyp 9f2, Cyp9F2, Dmel\\CG11466, anon-EST:Posey150', description='CG11466 gene product from transcript CG11466-RA', other_designations='CG11466-PA|Cyp9f2-PA|cytochrome P450', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12403340, end_position_on_the_genomic_accession=12406028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41521, CurrentID=0, Status='live', Symbol='CG5167', Aliases='Dmel_ Dmel\\CG5167', description='CG5167 gene product from transcript CG5167-RA', other_designations='CG5167-PA|saccheropin dehydrogenase', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12406532, end_position_on_the_genomic_accession=12408107, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41522, CurrentID=0, Status='live', Symbol='CG5196', Aliases='Dmel_ Dmel\\CG5196', description='CG5196 gene product from transcript CG5196-RA', other_designations='CG5196-PA|CG5196-PB', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12408316, end_position_on_the_genomic_accession=12411768, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41524, CurrentID=0, Status='live', Symbol='pps', Aliases='Dmel_CG6525, CG6525, Dmel\\CG6525, FBgn0082831, SPP, Spp, spp', description='protein partner of snf', other_designations='CG6525-PA|CG6525-PB|SNFs protein partner|pps-PA|pps-PB|protein partner of sans-fille', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12413828, end_position_on_the_genomic_accession=12421251, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41525, CurrentID=0, Status='live', Symbol='Scgbeta', Aliases='Dmel_CG5657, CG5657, DmSCG-beta, Dmel\\CG5657, dscgbeta, scg-beta', description='Sarcoglycan beta', other_designations='CG5657-PA|Scgbeta-PA|beta-sarcoglycan|beta-sarcolgycan|sarcoglycan beta', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12421236, end_position_on_the_genomic_accession=12422429, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41526, CurrentID=0, Status='live', Symbol='CG17202', Aliases='Dmel_ BcDNA:RH21090, Dmel\\CG17202', description='CG17202 gene product from transcript CG17202-RA', other_designations='CG17202-PA|uncharacterized protein', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12422768, end_position_on_the_genomic_accession=12423606, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41528, CurrentID=0, Status='live', Symbol='NANS', Aliases='Dmel_CG5232, CG5232, DmSAS, Dmel\\CG5232, Neu5Ac, SAS, Sas, sas', description='N-acetylneuraminic acid synthase', other_designations='CG5232-PA|N-acetylneuraminic acid phosphate synthase|NANS-PA|Neu5Ac synthase|sialic acid phosphate synthase|sialic acid synthase', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12426578, end_position_on_the_genomic_accession=12427877, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41529, CurrentID=0, Status='live', Symbol='CG5641', Aliases='Dmel_ Dmel\\CG5641', description='CG5641 gene product from transcript CG5641-RA', other_designations='CG5641-PA|interleukin enhancer binding factor', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12427795, end_position_on_the_genomic_accession=12429282, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41530, CurrentID=0, Status='live', Symbol='CG5245', Aliases='Dmel_ Dmel\\CG5245, anon-WO0118547.611', description='CG5245 gene product from transcript CG5245-RA', other_designations='CG5245-PA', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12429560, end_position_on_the_genomic_accession=12431235, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41531, CurrentID=0, Status='live', Symbol='Prosalpha2', Aliases='Dmel_CG5266, 153303_at, Alpha-2, CG5266, Dmel\\CG5266, PRC3, PROS Dm25, PROS-Dm25g, PSA2_DROME, Pros25, Prot. 25 Kda, Su(DTS), Su(DTS)-1, alpha-2, alpha2, alpha2_dm, dalpha2', description='Proteasome alpha2 subunit', other_designations='20s proteasome|CG5266-PA|Pros25|Prosalpha2-PA|Su(DTS)-1|alpha-subunit of 20S proteasome|alpha2|alpha2 subunit|proteasome 25kD subunit|proteasome alpha2 subunit|proteasome subunit alpha-type 2|proteosome', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12431321, end_position_on_the_genomic_accession=12432556, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41532, CurrentID=0, Status='live', Symbol='Aos1', Aliases='Dmel_CG12276, AOS1, Aos, CG12276, DmAos1, DmSAE1, DmSae1, Dmaos1, Dmel\\CG12276, SAE1, Sae1', description='Activator of SUMO 1', other_designations='Aos1-PA|CG12276-PA|Smt3 activating enzyme 1|activator of SUMO 1|ubiquitin activating enzyme|ubiquitin activating enzyme 1', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12432533, end_position_on_the_genomic_accession=12433822, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41533, CurrentID=0, Status='live', Symbol='CG5844', Aliases='Dmel_ Dmel\\CG5844', description='CG5844 gene product from transcript CG5844-RA', other_designations='CG5844-PA|Enoyl-CoA hydratase', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12434098, end_position_on_the_genomic_accession=12435643, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41536, CurrentID=0, Status='live', Symbol='Desat2', Aliases='Dmel_CG5925, 7-11HD, CG5925, Dmel\\CG5925, Fad, desat, desat-2, desat2, desat2/Fad, ds2', description='Desaturase 2', other_designations='7,11-heptacosadiene modifier|CG5925-PA|Desat2-PA|Desaturase2|desat2|desaturase|desaturase 2|desaturase-2|fatty acid desaturase', map_location='3-51.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12436248, end_position_on_the_genomic_accession=12437857, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41537, CurrentID=0, Status='live', Symbol='CG5538', Aliases='Dmel_ Dmel\\CG5538, anon-WO0140519.84, bs11a11.y1', description='CG5538 gene product from transcript CG5538-RA', other_designations='CG5538-PA', map_location='87B11-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12447653, end_position_on_the_genomic_accession=12449046, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41538, CurrentID=0, Status='live', Symbol='CG18549', Aliases='Dmel_ Dmel\\CG18549', description='CG18549 gene product from transcript CG18549-RA', other_designations='CG18549-PA|uncharacterized protein', map_location='87B11-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12449296, end_position_on_the_genomic_accession=12456355, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41539, CurrentID=0, Status='live', Symbol='CG5509', Aliases='Dmel_ Dmel\\CG5509', description='CG5509 gene product from transcript CG5509-RA', other_designations='CG5509-PA', map_location='87B11-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12452464, end_position_on_the_genomic_accession=12453196, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41540, CurrentID=0, Status='live', Symbol='trus', Aliases='Dmel_CG5333, CG5333, Dmel\\CG5333', description='toys are us', other_designations='CG5333-PA|trus-PA', map_location='87B11-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12456573, end_position_on_the_genomic_accession=12458444, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41541, CurrentID=0, Status='live', Symbol='CG5961', Aliases='Dmel_ Dmel\\CG5961, FBXO9', description='CG5961 gene product from transcript CG5961-RB', other_designations='CG5961-PA|CG5961-PB|CG5961-PC', map_location='87B11-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12458421, end_position_on_the_genomic_accession=12460294, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41542, CurrentID=0, Status='live', Symbol='CG12267', Aliases='Dmel_ Dmel\\CG12267', description='CG12267 gene product from transcript CG12267-RB', other_designations='CG12267-PA|CG12267-PB|uncharacterized protein', map_location='87B11-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12460414, end_position_on_the_genomic_accession=12462418, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41543, CurrentID=0, Status='live', Symbol='CG5608', Aliases='Dmel_ Dmel\\CG5608', description='CG5608 gene product from transcript CG5608-RA', other_designations='CG5608-PA', map_location='87B11-87B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12462562, end_position_on_the_genomic_accession=12465298, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  8674060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674060, CurrentID=0, Status='live', Symbol='CG42587', Aliases='Dmel_ Dmel\\CG42587', description='CG42587 gene product from transcript CG42587-RA', other_designations='CG42587-PA|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14744845, end_position_on_the_genomic_accession=14745210, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41547, CurrentID=0, Status='live', Symbol='Hug', Aliases='Dmel_CG6371, CG6371, Dmel\\CG6371, Drm-MT2, Drm-PK-2, HUG, HUGgamma-gamma, Hugin-PK-2, PK-2, hug, hug gamma, hug-1, hug-2, hug-PK, hug-gamma, hugg, huggamma, hugin, pyrokinin/PBAN-like, Hug', description='Hugin', other_designations='CG6371-PA|Hug-PA|Hugin-gamma|Hugin-pyrokinin|SVPFKPRLamide|hugin|hugin gamma|pyrokinin|pyrokinin-2', map_location='87B15-87B15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12526188, end_position_on_the_genomic_accession=12529076, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41549, CurrentID=0, Status='live', Symbol='Octbeta2R', Aliases='Dmel_CG33976, CG18553, CG33976, CG6989, DmCG6989, DmOctbeta2R, Dm_3R:45121, Dmel\\CG33976, OctB2R, OctBeta2, Octbeta2r, anon-WO0170980.184, anon-WO0170980.185, octbeta2R', description='Octopamine beta2 receptor', other_designations='CG33976-PA|CG33976-PC|CG33976-PD|CG33976-PE|CG33976-PF|CG33976-PG|Octbeta2R-PA|Octbeta2R-PC|Octbeta2R-PD|Octbeta2R-PE|Octbeta2R-PF|Octbeta2R-PG|octopamine beta receptor|octopamine beta2 receptor', map_location='87C1-87C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12550625, end_position_on_the_genomic_accession=12595725, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41550, CurrentID=0, Status='live', Symbol='Vha55', Aliases='Dmel_CG17369, 1208/13, ATP6V1B1, CG17369, Dmel\\CG17369, SzA, V-ATPase, VAT-2, l(3)87Ca, l(3)SzA, l(3)j2E9, vha55', description='Vacuolar H[+]-ATPase 55kD subunit', other_designations='B subunit|B subunit of V-ATPase|CG17369-PA|CG17369-PB|CG17369-PC|V-ATPase B-subunit|V-ATPase subunit B|V1 B subunit|Vha55-PA|Vha55-PB|Vha55-PC|vacuolar ATP synthase subunit B|vacuolar ATPase B subunit|vacuolar H+-ATPase 55kD B subunit|vacuolar H+ATPase subunit B|vacuolar H[+]-ATPase 55-kDa B subunit|vacuolar H[+]-ATPase 55kD B subunit|vacuolar H[+]-ATPase 55kD subunit', map_location='3-51.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12623762, end_position_on_the_genomic_accession=12627829, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41551, CurrentID=0, Status='live', Symbol='Snx3', Aliases='Dmel_CG6359, CG6359, DmSNX3, Dmel\\CG6359, Dsnx3, lincRNA.S6645, snx3', description='Sorting nexin 3', other_designations='CG6359-PA|CG6359-PB|Snx3-PA|Snx3-PB|sorting nexin 3', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12628031, end_position_on_the_genomic_accession=12630495, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41552, CurrentID=0, Status='live', Symbol='Not10', Aliases='Dmel_CG18616, CG18616, Dmel\\CG18616, NOT10, d', description='CG18616 gene product from transcript CG18616-RB', other_designations='CG18616-PA|CG18616-PB|Not10-PA|Not10-PB', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12630647, end_position_on_the_genomic_accession=12633188, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41554, CurrentID=0, Status='live', Symbol='CG11598', Aliases='Dmel_ Dmel\\CG11598', description='CG11598 gene product from transcript CG11598-RB', other_designations='CG11598-PB|uncharacterized protein', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12634917, end_position_on_the_genomic_accession=12636283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41555, CurrentID=0, Status='live', Symbol='CG11600', Aliases='Dmel_ Dmel\\CG11600', description='CG11600 gene product from transcript CG11600-RB', other_designations='CG11600-PB|uncharacterized protein', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12636540, end_position_on_the_genomic_accession=12637976, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41556, CurrentID=0, Status='live', Symbol='CG11608', Aliases='Dmel_ Dmel\\CG11608', description='CG11608 gene product from transcript CG11608-RB', other_designations='CG11608-PB|uncharacterized protein', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12638481, end_position_on_the_genomic_accession=12640033, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41557, CurrentID=0, Status='live', Symbol='CG6753', Aliases='Dmel_ Dmel\\CG6753', description='CG6753 gene product from transcript CG6753-RB', other_designations='CG6753-PB', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12640163, end_position_on_the_genomic_accession=12642482, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41558, CurrentID=0, Status='live', Symbol='Dtg', Aliases='Dmel_CG6234, CG6234, Dmel\\CG6234', description='Dpp target gene', other_designations='CG6234-PA|Dtg-PA', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12648058, end_position_on_the_genomic_accession=12652910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41559, CurrentID=0, Status='live', Symbol='CG6225', Aliases='Dmel_ Dmel\\CG6225', description='CG6225 gene product from transcript CG6225-RB', other_designations='CG6225-PA|CG6225-PB', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12654577, end_position_on_the_genomic_accession=12660860, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41560, CurrentID=0, Status='live', Symbol='CG14395', Aliases='Dmel_ Dmel\\CG14395', description='CG14395 gene product from transcript CG14395-RD', other_designations='CG14395-PB|CG14395-PD|CG14395-PE|CG14395-PF|CG14395-PG|CG14395-PH|CG14395-PI|CG14395-PJ|uncharacterized protein', map_location='87C4-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12662831, end_position_on_the_genomic_accession=12673959, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41561, CurrentID=0, Status='live', Symbol='Gnmt', Aliases='Dmel_CG6188, CG6188, Dmel\\CG6188, gnmt', description='Glycine N-methyltransferase', other_designations='CG6188-PA|Gnmt-PA|glycine N-methyltransferase', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12674608, end_position_on_the_genomic_accession=12676084, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41562, CurrentID=0, Status='live', Symbol='mbo', Aliases='Dmel_CG6819, 1465/07, 1465/7, CG6819, DNup88, Dmel\\CG6819, Dnup 88, Dnup88, Mbo, Nup88, anon-WO0118547.263, dNup88, l(3)05043, l(3)S146507, nup88', description='members only', other_designations='CG6819-PA|mbo-PA', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12676487, end_position_on_the_genomic_accession=12679052, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41563, CurrentID=0, Status='live', Symbol='Cyp313a4', Aliases='Dmel_CG6802, 313a4, CG6802, Cyp313A4, Dmel\\CG6802', description='CG6802 gene product from transcript CG6802-RA', other_designations='CG6802-PA|Cyp313a4-PA', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12679279, end_position_on_the_genomic_accession=12681375, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41565, CurrentID=0, Status='live', Symbol='Su(fu)', Aliases='Dmel_CG6054, CG6054, Dmel\\CG6054, SU(FU), SUFU, Su(Fu), SuFu, Sufu, dSufu, su(fu), sufu', description='Suppressor of fused', other_designations='CG6054-PA|Su(fu)-PA|suppressor of fused', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12689873, end_position_on_the_genomic_accession=12691468, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41566, CurrentID=0, Status='live', Symbol='Arp1', Aliases='Dmel_CG6174, ARP1, ARP87C, Actr87C, Arp, Arp-1, Arp87C, Arp87c, CG6174, Dmel\\CG6174, arp1, gdl, grid', description='Actin-related protein 1', other_designations='Actin-related protein 87C|Arp1-PA|CG6174-PA|gridlock', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12691743, end_position_on_the_genomic_accession=12694232, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41567, CurrentID=0, Status='live', Symbol='CG14391', Aliases='Dmel_ BcDNA:AT22529, Dmel\\CG14391', description='CG14391 gene product from transcript CG14391-RA', other_designations='CG14391-PA|uncharacterized protein', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12694229, end_position_on_the_genomic_accession=12694842, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41568, CurrentID=0, Status='live', Symbol='NijC', Aliases='Dmel_CG14394, CG14394, Dmel\\CG14394', description='Ninjurin C', other_designations='CG14394-PB|CG14394-PC|CG14394-PD|CG14394-PE|CG14394-PF|CG14394-PG|NijC-PB|NijC-PC|NijC-PD|NijC-PE|NijC-PF|NijC-PG|ninjurin C', map_location='87C5-87C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12695129, end_position_on_the_genomic_accession=12697379, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41569, CurrentID=0, Status='live', Symbol='Past1', Aliases='Dmel_CG6148, CG6148, CG66148, Dmel\\CG6148, Ehd, PAST, PAST-1, Past-1, dEHD', description='Putative Achaete Scute Target 1', other_designations='CG6148-PA|CG6148-PB|EH domain containing protein|Past1-PA|Past1-PB|putative achaete scute target 1', map_location='87C6-87C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12697322, end_position_on_the_genomic_accession=12704875, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41570, CurrentID=0, Status='live', Symbol='CG12279', Aliases='Dmel_ BcDNA:RH54517, Dmel\\CG12279', description='CG12279 gene product from transcript CG12279-RA', other_designations='CG12279-PA|uncharacterized protein', map_location='87C6-87C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12705094, end_position_on_the_genomic_accession=12705818, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41571, CurrentID=0, Status='live', Symbol='mus308', Aliases='Dmel_CG6019, 0443/09, AAF54858, CG6019, Dmel\\CG6019, Mus308, Nuc3, POLQ, Pol theta, PolQ, mus-308', description='mutagen-sensitive 308', other_designations='CG6019-PA|DNA polymerase theta|Nuclease-3|mus308-PA|pol theta', map_location='87C6-87C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12705523, end_position_on_the_genomic_accession=12712299, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41573, CurrentID=0, Status='live', Symbol='CG5724', Aliases='Dmel_ Dmel\\CG5724', description='CG5724 gene product from transcript CG5724-RA', other_designations='CG5724-PA', map_location='87C8-87C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12739642, end_position_on_the_genomic_accession=12741417, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41574, CurrentID=0, Status='live', Symbol='CG5999', Aliases='Dmel_ Dmel\\CG5999', description='CG5999 gene product from transcript CG5999-RB', other_designations='CG5999-PB', map_location='87C8-87C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12741958, end_position_on_the_genomic_accession=12743680, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41575, CurrentID=0, Status='live', Symbol='beat-Vc', Aliases='Dmel_CG14390, Beat-Vc, CG14390, Dmel\\CG14390, beat Vc', description='CG14390 gene product from transcript CG14390-RA', other_designations='CG14390-PA|beat-Vc-PA', map_location='87C8-87D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12746036, end_position_on_the_genomic_accession=12783192, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41576, CurrentID=0, Status='live', Symbol='CG31345', Aliases='Dmel_ CG10135, CG14388, Dmel\\CG31345', description='CG31345 gene product from transcript CG31345-RA', other_designations='CG31345-PA|uncharacterized protein', map_location='87D1-87D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12811145, end_position_on_the_genomic_accession=12817029, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41578, CurrentID=0, Status='live', Symbol='beat-Va', Aliases='Dmel_CG10134, Beat-Va, CG10134, Dmel\\CG10134, beat Va', description='CG10134 gene product from transcript CG10134-RC', other_designations='CG10134-PA|CG10134-PB|CG10134-PC|beat-Va-PA|beat-Va-PB|beat-Va-PC', map_location='87D2-87D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12840560, end_position_on_the_genomic_accession=12848419, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41579, CurrentID=0, Status='live', Symbol='CG10126', Aliases='Dmel_ CT28495, Dmel\\CG10126', description='CG10126 gene product from transcript CG10126-RB', other_designations='CG10126-PA|CG10126-PB|uncharacterized protein', map_location='87D2-87D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12855544, end_position_on_the_genomic_accession=12859246, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41580, CurrentID=0, Status='live', Symbol='d-cup', Aliases='Dmel_CG14387, CG14387, Dmel\\CG14387', description='davis-cup', other_designations='CG14387-PA|d-cup-PA', map_location='87D3-87D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12891956, end_position_on_the_genomic_accession=12893158, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318226, CurrentID=0, Status='live', Symbol='CG32817', Aliases='Dmel_ Dmel\\CG32817', description='CG32817 gene product from transcript CG32817-RC', other_designations='CG32817-PA|CG32817-PB|CG32817-PC|uncharacterized protein', map_location='1B4-1B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=479066, end_position_on_the_genomic_accession=480166, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41583, CurrentID=0, Status='live', Symbol='beat-Vb', Aliases='Dmel_CG31298, Beat-Vb, CG14385, CG14386, CG31298, Dmel\\CG31298, beat Vb', description='CG31298 gene product from transcript CG31298-RB', other_designations='CG31298-PA|CG31298-PB|beat-Vb-PA|beat-Vb-PB', map_location='87D4-87D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12916056, end_position_on_the_genomic_accession=12934813, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41585, CurrentID=0, Status='live', Symbol='Spc25', Aliases='Dmel_CG7242, BcDNA:LD37196, CG7242, Cal2, Dmel\\CG7242, Mitch, S3, cal2, dmR, l(3)87Da, l(3)A34-1, l(3)A46-1, l(3)S3, l(3)c00064', description='CG7242 gene product from transcript CG7242-RA', other_designations='CG7242-PA|Spc25-PA|chromosome alignment defect 2|lethal (3) 87Da|lethal (3) c00064|mitch|mitotic chaos', map_location='3-51.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12961745, end_position_on_the_genomic_accession=12963249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41586, CurrentID=0, Status='live', Symbol='Cyp304a1', Aliases='Dmel_CG7241, 304a1, BcDNA:LP06355, CG7241, CYP304A1-Dm, Cyp304A1, Dmel\\CG7241', description='CG7241 gene product from transcript CG7241-RA', other_designations='CG7241-PA|Cyp304a1-PA|cytochrome P450', map_location='87D5-87D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12963791, end_position_on_the_genomic_accession=12965923, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41587, CurrentID=0, Status='live', Symbol='grsm', Aliases='Dmel_CG7340, BcDNA:LD41548, CG7340, Dmel\\CG7340, granny-smith, l(3)c00064', description='granny smith', other_designations='CG7340-PB|CG7340-PF|Leucyl-aminopeptidase|aminopeptidase|granny-smith|grsm-PB|grsm-PF|lethal (3) c00064', map_location='87D5-87D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12950042, end_position_on_the_genomic_accession=12967712, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41588, CurrentID=0, Status='live', Symbol='CG14384', Aliases='Dmel_ Dmel\\CG14384', description='CG14384 gene product from transcript CG14384-RB', other_designations='CG14384-PA|CG14384-PB|uncharacterized protein', map_location='87D5-87D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12965938, end_position_on_the_genomic_accession=12967712, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41589, CurrentID=0, Status='live', Symbol='CG7381', Aliases='Dmel_ CT22687, Dmel\\CG7381', description='CG7381 gene product from transcript CG7381-RE', other_designations='CG7381-PA|CG7381-PB|CG7381-PD|CG7381-PE|CG7381-PF', map_location='87D5-87D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12967775, end_position_on_the_genomic_accession=12977471, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  41590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41590, CurrentID=0, Status='live', Symbol='CG7091', Aliases='Dmel_ Dmel\\CG7091', description='CG7091 gene product from transcript CG7091-RB', other_designations='CG7091-PB', map_location='87D6-87D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12978000, end_position_on_the_genomic_accession=12980535, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41591, CurrentID=0, Status='live', Symbol='Paip2', Aliases='Dmel_CG12358, CG12358, Dmel\\CG12358, dPaip2, paip2', description='polyA-binding protein interacting protein 2', other_designations='CG12358-PA|CG12358-PB|Paip2-PA|Paip2-PB|PolyA-binding protein interacting protein 2', map_location='87D6-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12980792, end_position_on_the_genomic_accession=12985488, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41592, CurrentID=0, Status='live', Symbol='CG31342', Aliases='Dmel_ CG18551, CG7472, Dmel\\CG31342', description='CG31342 gene product from transcript CG31342-RC', other_designations='CG31342-PB|CG31342-PC|uncharacterized protein', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12986041, end_position_on_the_genomic_accession=12991543, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  8674068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674068, CurrentID=0, Status='live', Symbol='CG42516', Aliases='Dmel_ Dmel\\CG42516, Dromel_CG11508_FBtr0088817_uORF', description='CG42516 gene product from transcript CG42516-RB', other_designations='CG42516-PA|CG42516-PB|CG42516-PC|uncharacterized protein', map_location='44B3-44B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8129354, end_position_on_the_genomic_accession=8131323, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41595, CurrentID=0, Status='live', Symbol='yellow-f', Aliases='Dmel_CG18550, CG18550, Dmel\\CG18550, yell-f', description='CG18550 gene product from transcript CG18550-RC', other_designations='CG18550-PA|CG18550-PB|CG18550-PC|yellow f|yellow-f-PA|yellow-f-PB|yellow-f-PC', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12991449, end_position_on_the_genomic_accession=12993201, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41596, CurrentID=0, Status='live', Symbol='yellow-f2', Aliases='Dmel_CG8063, CG8063, Dmel\\CG8063, yel-f2', description='CG8063 gene product from transcript CG8063-RA', other_designations='CG8063-PA|yellow-f2-PA', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12993395, end_position_on_the_genomic_accession=12995147, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41597, CurrentID=0, Status='live', Symbol='CG7488', Aliases='Dmel_ Dmel\\CG7488', description='CG7488 gene product from transcript CG7488-RA', other_designations='CG7488-PA', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12995620, end_position_on_the_genomic_accession=12996925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41598, CurrentID=0, Status='live', Symbol='CG17327', Aliases='Dmel_ Dmel\\CG17327', description='CG17327 gene product from transcript CG17327-RC', other_designations='CG17327-PA|CG17327-PC|uncharacterized protein', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12996876, end_position_on_the_genomic_accession=12997726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674069, CurrentID=0, Status='live', Symbol='CG42585', Aliases='Dmel_ Dmel\\CG42585', description='CG42585 gene product from transcript CG42585-RA', other_designations='CG42585-PA|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14739914, end_position_on_the_genomic_accession=14741138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41600, CurrentID=0, Status='live', Symbol='CG11656', Aliases='Dmel_ Dmel\\CG11656, anon-WO0140519.199', description='CG11656 gene product from transcript CG11656-RA', other_designations='CG11656-PA|uncharacterized protein', map_location='87D8-87D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13009895, end_position_on_the_genomic_accession=13011086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41601, CurrentID=0, Status='live', Symbol='CG8031', Aliases='Dmel_ Dmel\\CG8031, dMEMO', description='CG8031 gene product from transcript CG8031-RA', other_designations='CG8031-PA', map_location='87D8-87D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13008384, end_position_on_the_genomic_accession=13011310, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41602, CurrentID=0, Status='live', Symbol='CtBP', Aliases='Dmel_CG7583, 0256/16, 256/16, BcDNA:GH07572, CG7583, CTBP, Dm-CtBP, Dmel\\CG7583, EP(3)3528, G9, RE01030, RE61183, ctbp, dCtBP, l(3)03463, l(3)03464, l(3)87DE, l(3)87De, l(3)G9', description='C-terminal Binding Protein', other_designations='C-terminal binding protein|C-terminal-Binding-Protein|C-terminal-binding protein|CG7583-PA|CG7583-PB|CG7583-PC|CG7583-PD|CG7583-PE|CG7583-PF|CG7583-PG|CG7583-PH|CG7583-PI|CG7583-PJ|CtBP-PA|CtBP-PB|CtBP-PC|CtBP-PD|CtBP-PE|CtBP-PF|CtBP-PG|CtBP-PH|CtBP-PI|CtBP-PJ|drosophila C-terminal binding protein|lethal(3)87De', map_location='87D8-87D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13011649, end_position_on_the_genomic_accession=13026275, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  41605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41605, CurrentID=0, Status='live', Symbol='ry', Aliases='Dmel_CG7642, CG7642, Dmel\\CG7642, XDH, XOR, Xdh, Xdh/ry, xdh', description='rosy', other_designations='CG7642-PA|ry-PA|xanthine DH|xanthine dehydrogenase', map_location='3-52.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13032528, end_position_on_the_genomic_accession=13038020, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41606, CurrentID=0, Status='live', Symbol='CG11668', Aliases='Dmel_ Dmel\\CG11668, SP115', description='CG11668 gene product from transcript CG11668-RA', other_designations='CG11668-PA|uncharacterized protein', map_location='87D9-87D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13038375, end_position_on_the_genomic_accession=13040184, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41607, CurrentID=0, Status='live', Symbol='snk', Aliases='Dmel_CG7996, CG7996, Dmel\\CG7996, SNK, SP26, Snk, c-SP26, l(3)87Dg, me(3)4', description='snake', other_designations='CG7996-PA|CG7996-PB|CG7996-PC|snk-PA|snk-PB|snk-PC', map_location='3-52.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13040165, end_position_on_the_genomic_accession=13042770, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41608, CurrentID=0, Status='live', Symbol='CG11670', Aliases='Dmel_ Dmel\\CG11670, SP27', description='CG11670 gene product from transcript CG11670-RD', other_designations='CG11670-PD|uncharacterized protein', map_location='87D9-87D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13042050, end_position_on_the_genomic_accession=13044086, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41609, CurrentID=0, Status='live', Symbol='Hsc70-2', Aliases='Dmel_CG7756, CG7756, Dmel\\CG7756, HSC-70, HSC2, HSC70, HSC70-2, HSP7B_DROME, Hsc-2, Hsc2, Hsc70, Hsc70.2, dhsc70, hsc2, hsc70', description='Heat shock protein cognate 2', other_designations='CG7756-PA|Hs protein cognate 2|Hsc70-2-PA|Hsp cognate 2|heat shock protein cognate 2|heat shock protein cognate-2', map_location='87D10-87D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13044759, end_position_on_the_genomic_accession=13047388, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41610, CurrentID=0, Status='live', Symbol='CG7966', Aliases='Dmel_ Dmel\\CG7966', description='CG7966 gene product from transcript CG7966-RA', other_designations='CG7966-PA', map_location='87D10-87D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13049398, end_position_on_the_genomic_accession=13051453, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41611, CurrentID=0, Status='live', Symbol='pic', Aliases='Dmel_CG7769, 0263/16, 1258/16, 1366/03, 153636_at, 263/16, CG7769, D-DDB1, DDB-1, DDB1, Dbd1, Ddb1, Dmel\\CG7769, d-ddb1, dDdb1, l(3)87Dh, l(3)S136603', description='piccolo', other_designations='CG7769-PA|DDB1|Damage-specific DNA binding protein 1|damaged DNA binding protein 1|drosophila damage-specific DNA-binding protein 1|pic-PA', map_location='3-52.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13052245, end_position_on_the_genomic_accession=13056725, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41612, CurrentID=0, Status='live', Symbol='sim', Aliases='Dmel_CG7771, 0483/09, 0716/08, 0899/14, 0953/08, 1002/10, 1034/02, 1050/13, 1110/04, 1111/10, 1330/08, 1479/04, CG7771, Dmel\\CG7771, S8, SIM, Sim, bHLHe16, l(3)87Ea, l(3)E320, l(3)RD, l(3)S8, l(3)s8, schm, shm', description='single-minded', other_designations='CG7771-PA|CG7771-PB|CG7771-PC|CG7771-PD|Single-minded|Singled-minded|schmal|sim-PA|sim-PB|sim-PC|sim-PD|single minded|singleminded', map_location='3-52.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13057755, end_position_on_the_genomic_accession=13078124, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  41615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41615, CurrentID=0, Status='live', Symbol='timeout', Aliases='Dmel_CG7855, CG14381, CG14382, CG7855, CG8148, Dmel\\CG7855, TIM2, dtim2, tim2', description='CG7855 gene product from transcript CG7855-RA', other_designations='CG7855-PA|timeless 2|timeless2|timeout-PA', map_location='87D12-87D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13088651, end_position_on_the_genomic_accession=13163876, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  41616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41616, CurrentID=0, Status='live', Symbol='2mit', Aliases='Dmel_CG42573, CG10148, CG17319, CG42573, CT28541, CT38314, Dmel\\CG42573, Dmel_CG10148, Dmel_CG17319', description='CG42573 gene product from transcript CG42573-RA', other_designations='2mit-PA|2mit-PB|2mit-PC|2mit-PD|CG42573-PA|CG42573-PB|CG42573-PC|CG42573-PD', map_location='87D13-87D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13121951, end_position_on_the_genomic_accession=13155687, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41619, CurrentID=0, Status='live', Symbol='CG8138', Aliases='Dmel_ Dmel\\CG8138', description='CG8138 gene product from transcript CG8138-RA', other_designations='CG8138-PA', map_location='87E1-87E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13181304, end_position_on_the_genomic_accession=13182154, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41620, CurrentID=0, Status='live', Symbol='CG8508', Aliases='Dmel_ Dmel\\CG8508, anon-WO0140519.216', description='CG8508 gene product from transcript CG8508-RA', other_designations='CG8508-PA', map_location='87E1-87E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13185526, end_position_on_the_genomic_accession=13186496, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41621, CurrentID=0, Status='live', Symbol='CG14380', Aliases='Dmel_ Dmel\\CG14380', description='CG14380 gene product from transcript CG14380-RB', other_designations='CG14380-PB|uncharacterized protein', map_location='87E1-87E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13188519, end_position_on_the_genomic_accession=13189568, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41622, CurrentID=0, Status='live', Symbol='CG8141', Aliases='Dmel_ BcDNA:AT26376, Dmel\\CG8141', description='CG8141 gene product from transcript CG8141-RA', other_designations='CG8141-PA', map_location='87E1-87E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13191599, end_position_on_the_genomic_accession=13192509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41623, CurrentID=0, Status='live', Symbol='CG8483', Aliases='Dmel_ Dmel\\CG8483', description='CG8483 gene product from transcript CG8483-RA', other_designations='CG8483-PA', map_location='87E2-87E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13208754, end_position_on_the_genomic_accession=13214465, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41624, CurrentID=0, Status='live', Symbol='CG8476', Aliases='Dmel_ Dmel\\CG8476', description='CG8476 gene product from transcript CG8476-RA', other_designations='CG8476-PA', map_location='87E2-87E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13219644, end_position_on_the_genomic_accession=13220578, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41625, CurrentID=0, Status='live', Symbol='Ace', Aliases='Dmel_CG17907, ACE, AChE, AcChE, AchE, CG17907, CHE, DmAChE, Dmel\\CG17907, FBgn0000024, ace, ace-2, ache, dAChE, dmAChE, l(3)26, l(3)87Ed', description='Acetylcholine esterase', other_designations='Ace-PA|Ace-PB|Ace-PC|CG17907-PA|CG17907-PB|CG17907-PC|acetyl choline esterase|acetylcholine esterase|acetylcholinesterase', map_location='3-52.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13222951, end_position_on_the_genomic_accession=13259517, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  41626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41626, CurrentID=0, Status='live', Symbol='Ravus', Aliases='Dmel_CG15889, CG15889, Dmel\\CG15889', description='CG15889 gene product from transcript CG15889-RA', other_designations='CG15889-PA|Ravus-PA|ravus', map_location='87E3-87E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13262733, end_position_on_the_genomic_accession=13264061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41627, CurrentID=0, Status='live', Symbol='Su(var)3-7', Aliases='Dmel_CG8599, CG8599, DmSu(var)3-7, Dmel\\CG8599, SU(VAR)3-7, Su(Var)3-7, Su(var)(3)3, Su(var)(3)7, Su-var(3)7, Suvar(3)7, l(3)87El', description='Suppressor of variegation 3-7', other_designations='CG8599-PA|CG8599-PB|Su(var)3-7-PA|Su(var)3-7-PB|suppressor of variegation 3-7', map_location='3-50.9 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13264364, end_position_on_the_genomic_accession=13270038, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41628, CurrentID=0, Status='live', Symbol='CG8449', Aliases='Dmel_ Dmel\\CG8449', description='CG8449 gene product from transcript CG8449-RA', other_designations='CG8449-PA', map_location='87E3-87E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13269137, end_position_on_the_genomic_accession=13272011, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41629, CurrentID=0, Status='live', Symbol='CG8630', Aliases='Dmel_ Dmel\\CG8630', description='CG8630 gene product from transcript CG8630-RA', other_designations='CG8630-PA', map_location='87E3-87E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13279721, end_position_on_the_genomic_accession=13284527, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41630, CurrentID=0, Status='live', Symbol='CG15888', Aliases='Dmel_ Dmel\\CG15888', description='CG15888 gene product from transcript CG15888-RB', other_designations='CG15888-PB|uncharacterized protein', map_location='87E4-87E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13284946, end_position_on_the_genomic_accession=13286529, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41631, CurrentID=0, Status='live', Symbol='CG15887', Aliases='Dmel_ BcDNA:RE53127, Dmel\\CG15887', description='CG15887 gene product from transcript CG15887-RA', other_designations='CG15887-PA|uncharacterized protein', map_location='87E4-87E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13287655, end_position_on_the_genomic_accession=13288813, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41632, CurrentID=0, Status='live', Symbol='Osi22', Aliases='Dmel_CG8644, CG8644, Dmel\\CG8644', description='Osiris 22', other_designations='CG8644-PA|Osi22-PA|osiris|osiris 22', map_location='87E4-87E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13290629, end_position_on_the_genomic_accession=13291846, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41633, CurrentID=0, Status='live', Symbol='wntD', Aliases='Dmel_CG8458, CG8458, DWnt-8, DWnt8, Dm DWnt8, Dmel\\CG8458, Wnt8, WntD, wnt8', description='wnt inhibitor of Dorsal', other_designations='CG8458-PA|wnt inhibitor of dorsal|wntD-PA', map_location='87E4-87E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13292052, end_position_on_the_genomic_accession=13293198, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41634, CurrentID=0, Status='live', Symbol='CG8773', Aliases='Dmel_ Dmel\\CG8773', description='CG8773 gene product from transcript CG8773-RB', other_designations='CG8773-PB', map_location='87E4-87E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13294748, end_position_on_the_genomic_accession=13298288, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41635, CurrentID=0, Status='live', Symbol='CG8774', Aliases='Dmel_ Dmel\\CG8774', description='CG8774 gene product from transcript CG8774-RB', other_designations='CG8774-PA|CG8774-PB', map_location='87E4-87E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13298768, end_position_on_the_genomic_accession=13302334, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41636, CurrentID=0, Status='live', Symbol='CG32473', Aliases='Dmel_ CG11688, CG8775, Dmel\\CG32473, GM08240p', description='CG32473 gene product from transcript CG32473-RA', other_designations='CG32473-PA|CG32473-PB|CG32473-PC|uncharacterized protein', map_location='87E5-87E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13302648, end_position_on_the_genomic_accession=13312785, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  41638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41638, CurrentID=0, Status='live', Symbol='PK2-R2', Aliases='Dmel_CG8795, CG8795, Dm CG8795, Dmel\\CG8795, anon-WO0170980.193, anon-WO0170980.194', description='Pyrokinin 2 receptor 2', other_designations='CG8795-PA|CG8795-PB|PK2-R2-PA|PK2-R2-PB|pyrokinin 2 receptor|pyrokinin 2 receptor 2', map_location='87E6-87E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13329070, end_position_on_the_genomic_accession=13334166, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41639, CurrentID=0, Status='live', Symbol='PK2-R1', Aliases='Dmel_CG8784, CG8784, Dm CG8784, Dmel\\CG8784, anon-WO0170980.190, anon-WO0170980.191', description='Pyrokinin 2 receptor 1', other_designations='CG8784-PA|PK2-R1-PA|pyrokinin 2 receptor|pyrokinin 2 receptor 1', map_location='87E6-87E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13338330, end_position_on_the_genomic_accession=13345106, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41640, CurrentID=0, Status='live', Symbol='Dic1', Aliases='Dmel_CG8790, BcDNA:GH02431, CG8790, DmDic1p, Dmel\\CG8790', description='Dicarboxylate carrier 1', other_designations='CG8790-PA|CG8790-PB|CG8790-PC|Dic1-PA|Dic1-PB|Dic1-PC|dicarboxylate carrier 1', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13364938, end_position_on_the_genomic_accession=13369278, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41641, CurrentID=0, Status='live', Symbol='poly', Aliases='Dmel_CG9829, 0081/31, CG9829, Dmel\\CG9829, did, did/alt1, l(3)05137, l(3)S008131', description='CG9829 gene product from transcript CG9829-RB', other_designations='CG9829-PA|CG9829-PB|lethal (3) 05137|poly-PA|poly-PB', map_location='87E7-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13361816, end_position_on_the_genomic_accession=13366910, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41642, CurrentID=0, Status='live', Symbol='CheA87a', Aliases='Dmel_CG14379, CG14379, Dmel\\CG14379', description='Chemosensory protein A 87a', other_designations='CG14379-PA|CheA87a-PA|chemosensory protein A 87a', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13369648, end_position_on_the_genomic_accession=13370346, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41643, CurrentID=0, Status='live', Symbol='Lip3', Aliases='Dmel_CG8823, CG8823, Dmel\\CG8823, Dmlip3, LIP3, LIP3Dm, d dilp3, lip3', description='CG8823 gene product from transcript CG8823-RA', other_designations='CG8823-PA|Lip3-PA|Lipase-3|dLipase-3|lipase 3|lipase3', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13370240, end_position_on_the_genomic_accession=13371904, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41644, CurrentID=0, Status='live', Symbol='CG9813', Aliases='Dmel_ Dmel\\CG9813, ORE-7', description='CG9813 gene product from transcript CG9813-RC', other_designations='CG9813-PA|CG9813-PB|CG9813-PC|CG9813-PD|CG9813-PE|CG9813-PF|CG9813-PG', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13373104, end_position_on_the_genomic_accession=13379775, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41645, CurrentID=0, Status='live', Symbol='CG8870', Aliases='Dmel_ Dmel\\CG8870, SP38', description='CG8870 gene product from transcript CG8870-RA', other_designations='CG8870-PA', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13379842, end_position_on_the_genomic_accession=13381509, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41646, CurrentID=0, Status='live', Symbol='Droj2', Aliases='Dmel_CG8863, CG8863, DROJ2, Dmel\\CG8863, anon-WO0118547.404', description='DnaJ-like-2', other_designations='CG8863-PA|CG8863-PB|CG8863-PC|CG8863-PD|CG8863-PE|DnaJ-like-2 chaperone|Droj2-PA|Droj2-PB|Droj2-PC|Droj2-PD|Droj2-PE', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13382106, end_position_on_the_genomic_accession=13385283, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41647, CurrentID=0, Status='live', Symbol='CG9799', Aliases='Dmel_ Dmel\\CG9799', description='CG9799 gene product from transcript CG9799-RA', other_designations='CG9799-PA', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13385648, end_position_on_the_genomic_accession=13388709, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41648, CurrentID=0, Status='live', Symbol='CCHa2', Aliases='Dmel_CG14375, BcDNA:RE10132, CCH2, CCM, CG14375, Dm-CCHa2, Dmel\\CG14375', description='CCHamide-2', other_designations='CCHa2-PA|CCHa2-PB|CCHa2-PC|CCHamide|CCHamide 2|CG14375-PA|CG14375-PB|CG14375-PC|Drm-CCHamide-2', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13389059, end_position_on_the_genomic_accession=13395197, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41649, CurrentID=0, Status='live', Symbol='CG14377', Aliases='Dmel_ Dmel\\CG14377', description='CG14377 gene product from transcript CG14377-RA', other_designations='CG14377-PA|CG14377-PB|uncharacterized protein', map_location='87E9-87E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13397835, end_position_on_the_genomic_accession=13398583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41650, CurrentID=0, Status='live', Symbol='GILT1', Aliases='Dmel_CG9796, CG9796, Dmel\\CG9796, GILT', description='Gamma-interferon-inducible lysosomal thiol reductase 1', other_designations='CG9796-PA|GILT1-PA|Gamma-interferon-inducible lysosomal thiol reductase', map_location='87E9-87E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13398526, end_position_on_the_genomic_accession=13401862, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41651, CurrentID=0, Status='live', Symbol='yellow-e3', Aliases='Dmel_CG17045, CG17045, Dmel\\CG17045, e3, yel-e3', description='CG17045 gene product from transcript CG17045-RA', other_designations='CG17045-PA|Yellow-d4|yellow-e3-PA', map_location='87E10-87E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13403423, end_position_on_the_genomic_accession=13404870, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41652, CurrentID=0, Status='live', Symbol='yellow-e2', Aliases='Dmel_CG17044, CG17044, Dmel\\CG17044, e2, yel-e2', description='CG17044 gene product from transcript CG17044-RB', other_designations='CG17044-PB|Yellow-d3|yellow-e2-PB', map_location='87E10-87E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13405109, end_position_on_the_genomic_accession=13406998, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41653, CurrentID=0, Status='live', Symbol='yellow-e', Aliases='Dmel_CG9792, CG9792, Dmel\\CG9792, e, yell-e', description='CG9792 gene product from transcript CG9792-RA', other_designations='CG9792-PA|yellow-e-PA', map_location='87E10-87E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13409543, end_position_on_the_genomic_accession=13415286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41654, CurrentID=0, Status='live', Symbol='Ir87a', Aliases='Dmel_CG14376, CG14376, DmelIR87a, Dmel\\CG14376, IR87a', description='Ionotropic receptor 87a', other_designations='CG14376-PB|Ir87a-PB|ionotropic receptor 87a', map_location='87E10-87E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13420279, end_position_on_the_genomic_accession=13423140, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41656, CurrentID=0, Status='live', Symbol='yrt', Aliases='Dmel_CG9764, CG9764, Dmel\\CG9764, Nbl4, Yrt, dlhB, l(3)87Ek, l(3)E19, l(3)m112, m112', description='yurt', other_designations='CG9764-PA|CG9764-PB|CG9764-PC|CG9764-PD|dorsal holes B|lethal(3)87Ek|yrt-PA|yrt-PB|yrt-PC|yrt-PD', map_location='3-52 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13428867, end_position_on_the_genomic_accession=13434984, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  41657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41657, CurrentID=0, Status='live', Symbol='CG14372', Aliases='Dmel_ 14372, BP1073, CT34011, Dmel\\CG14372', description='CG14372 gene product from transcript CG14372-RD', other_designations='CG14372-PC|CG14372-PD|uncharacterized protein', map_location='87E12-87E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13450590, end_position_on_the_genomic_accession=13464184, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  41658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41658, CurrentID=0, Status='live', Symbol='CG12538', Aliases='Dmel_ Dmel\\CG12538', description='CG12538 gene product from transcript CG12538-RA', other_designations='CG12538-PA|uncharacterized protein', map_location='87F1-87F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13522951, end_position_on_the_genomic_accession=13523535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41659, CurrentID=0, Status='live', Symbol='CG14370', Aliases='Dmel_ Dmel\\CG14370', description='CG14370 gene product from transcript CG14370-RA', other_designations='CG14370-PA|uncharacterized protein', map_location='87F3-87F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13587997, end_position_on_the_genomic_accession=13588725, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41660, CurrentID=0, Status='live', Symbol='CG14369', Aliases='Dmel_ Dmel\\CG14369', description='CG14369 gene product from transcript CG14369-RA', other_designations='CG14369-PA|uncharacterized protein', map_location='87F3-87F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13591632, end_position_on_the_genomic_accession=13592657, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41661, CurrentID=0, Status='live', Symbol='CG9759', Aliases='Dmel_ BcDNA:GH02603, Dmel\\CG9759', description='CG9759 gene product from transcript CG9759-RB', other_designations='CG9759-PA|CG9759-PB|CG9759-PC', map_location='87F4-87F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13607078, end_position_on_the_genomic_accession=13609246, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41662, CurrentID=0, Status='live', Symbol='CG9757', Aliases='Dmel_ Dmel\\CG9757, GCR 20, GCR-20, GCR20, Pen50DE, anon-Pen50DE, pen50DE', description='CG9757 gene product from transcript CG9757-RB', other_designations='CG9757-PA|CG9757-PB', map_location='87F4-87F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13611583, end_position_on_the_genomic_accession=13612433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41663, CurrentID=0, Status='live', Symbol='CG9269', Aliases='Dmel_ Dmel\\CG9269', description='CG9269 gene product from transcript CG9269-RA', other_designations='CG9269-PA', map_location='87F5-87F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13615917, end_position_on_the_genomic_accession=13616492, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41665, CurrentID=0, Status='live', Symbol='CG10841', Aliases='Dmel_ Dmel\\CG10841, anon-WO0140519.42', description='CG10841 gene product from transcript CG10841-RA', other_designations='CG10841-PA|CG10841-PB|uncharacterized protein', map_location='87F5-87F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13631942, end_position_on_the_genomic_accession=13634861, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41666, CurrentID=0, Status='live', Symbol='sqd', Aliases='Dmel_CG16901, BcDNA:GM03761, CG16901, CG17791, CR33328, Dmel\\CG16901, Dmel_CR33328, HRP40, Hrb87, Hrb87Fb, Hrp40, RRM3, Rbp3, SQD, Sqd, Squid, fs(3)00048, hnRNP 1, hnRNP A, hrp40, l(3)EP3631, l(3)j4B4, l(3)j6E3, l(3)j8E7, rrm3', description='squid', other_designations='CG16901-PA|CG16901-PB|CG16901-PC|CG16901-PD|CG16901-PE|RNA-binding protein 3|heterogeneous-nuclear-ribonucleoprotein-87Fb|lethal (3) j4B4|sqd-PA|sqd-PB|sqd-PC|sqd-PD|sqd-PE', map_location='87F5-87F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13634862, end_position_on_the_genomic_accession=13646304, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41668, CurrentID=0, Status='live', Symbol='Rbp4', Aliases='Dmel_CG9654, CG9654, Dmel\\CG9654, RRM4, TSR, Tsr, rrm4, tsr', description='RNA-binding protein 4', other_designations='CG9654-PA|Rbp4-PA|testis specific RRM protein', map_location='87F7-87F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13654556, end_position_on_the_genomic_accession=13656394, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41670, CurrentID=0, Status='live', Symbol='B52', Aliases='Dmel_CG10851/SRp55, B52/dSRp55, CG10851, Dmel\\CG10851, E(Dfd), E726, MabB52, RRM8, RSp55, Rbp8, SR55, SRp55, dSRp55, l(3)s2249, rrm8', description='CG10851 gene product from transcript CG10851-RN', other_designations='B52-PA|B52-PB|B52-PC|B52-PD|B52-PF|B52-PI|B52-PK|B52-PM|B52-PN|B52-PO|CG10851-PA|CG10851-PB|CG10851-PC|CG10851-PD|CG10851-PF|CG10851-PI|CG10851-PK|CG10851-PM|CG10851-PN|CG10851-PO|Monoclonal-antibody-B52|RNA-binding protein 8|Serine/arginine rich protein 55|enhancer of deformed|serine-arginine protein 55|splicing factor B52', map_location='87F7-87F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13661311, end_position_on_the_genomic_accession=13666891, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  41671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41671, CurrentID=0, Status='live', Symbol='Task6', Aliases='Dmel_CG9637, CG9637, Dmel\\CG9637, K2P-TASK6, anon-WO0109301.11, anon-WO0109301.53, anon-WO0109301.54, anon-WO0109301.55, anon-WO0109301.56, anon-WO0109301.57, anon-WO0109301.58, anon-WO0109301.59, anon-WO0109301.60, dTASK-6', description='TWIK-related acid-sensitive K[+] channel 6', other_designations='CG9637-PC|CG9637-PD|CG9637-PE|CG9637-PF|Task6-PC|Task6-PD|Task6-PE|Task6-PF', map_location='87F7-87F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13667046, end_position_on_the_genomic_accession=13674542, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41672, CurrentID=0, Status='live', Symbol='CG9588', Aliases='Dmel_ 9588, Dmel\\CG9588, dp27, p27', description='CG9588 gene product from transcript CG9588-RA', other_designations='CG9588-PA', map_location='87F7-87F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13675092, end_position_on_the_genomic_accession=13676681, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41673, CurrentID=0, Status='live', Symbol='Lkb1', Aliases='Dmel_CG9374, CG9374, DmLKB1, Dmel\\CG9374, LKB1, LKB1/PEUTZ JEGHERS KINASE, PAR-4, PAR4, STK11, anon-EST:Posey135, dLKB1, dlkb1, lkb1, par-4', description='Lkb1 kinase', other_designations='CG9374-PA|CG9374-PB|CG9374-PC|CG9374-PD|CG9374-PE|CG9374-PF|CG9374-PH|CG9374-PI|CG9374-PJ|Lkb1-PA|Lkb1-PB|Lkb1-PC|Lkb1-PD|Lkb1-PE|Lkb1-PF|Lkb1-PH|Lkb1-PI|Lkb1-PJ|lkb1', map_location='87F7-87F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13676549, end_position_on_the_genomic_accession=13679430, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41674, CurrentID=0, Status='live', Symbol='omd', Aliases='Dmel_CG9591, CG9591, Dmel\\CG9591, IntS5, Omd', description='oocyte maintenance defects', other_designations='CG9591-PA|CG9591-PB|CG9591-PC|integrator 5|omd-PA|omd-PB|omd-PC', map_location='87F9-87F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13679440, end_position_on_the_genomic_accession=13687210, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41675, CurrentID=0, Status='live', Symbol='flfl', Aliases='Dmel_CG9351, CG9351, Dmel\\CG9351, Flfl, PP4R3, Ppp4R3, l(3)L4179', description='falafel', other_designations='CG9351-PB|CG9351-PD|CG9351-PE|flfl-PB|flfl-PD|flfl-PE', map_location='87F10-87F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13683921, end_position_on_the_genomic_accession=13693905, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41676, CurrentID=0, Status='live', Symbol='CG14367', Aliases='Dmel_ Dmel\\CG14367', description='CG14367 gene product from transcript CG14367-RA', other_designations='CG14367-PA|uncharacterized protein', map_location='87F11-87F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13694053, end_position_on_the_genomic_accession=13695513, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41677, CurrentID=0, Status='live', Symbol='f-cup', Aliases='Dmel_CG9611, BcDNA:GH09045, CG9611, Dmel\\CG9611', description='flyers-cup', other_designations='CG9611-PA|CG9611-PB|CG9611-PC|CG9611-PD|CG9611-PE|CG9611-PF|CG9611-PG|CG9611-PH|CG9611-PI|f-cup-PA|f-cup-PB|f-cup-PC|f-cup-PD|f-cup-PE|f-cup-PF|f-cup-PG|f-cup-PH|f-cup-PI|gp150-like', map_location='87F11-87F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13695263, end_position_on_the_genomic_accession=13700838, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  41678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41678, CurrentID=0, Status='live', Symbol='ninaB', Aliases='Dmel_CG9347, CG9347, DRPE65, Dmel\\CG9347, RPE65, anon-WO0148162.1, anon-WO0148163.1, beta-diox', description='neither inactivation nor afterpotential B', other_designations='CG9347-PA|beta-carotene dioxygenase|ninaB-PA', map_location='3-53.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13700565, end_position_on_the_genomic_accession=13703164, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41679, CurrentID=0, Status='live', Symbol='Adgf-D', Aliases='Dmel_CG9621, ADGF-D, ADGFD, AdgfD, CG9621, Dmel\\CG9621, adgfd', description='Adenosine deaminase-related growth factor D', other_designations='Adgf-D-PA|CG9621-PA|adenosine deaminase-related growth factor D', map_location='87F11-87F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13703084, end_position_on_the_genomic_accession=13705402, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41680, CurrentID=0, Status='live', Symbol='Adgf-C', Aliases='Dmel_CG9345, ADGF-C, AdgfC, CG9345, Dmel\\CG9345, adgfc', description='Adenosine deaminase-related growth factor C', other_designations='Adgf-C-PA|CG9345-PA|adenosine deaminase-related growth factor C', map_location='87F12-87F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13706757, end_position_on_the_genomic_accession=13709129, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41682, CurrentID=0, Status='live', Symbol='Ubc87F', Aliases='Dmel_CG9602, 83A, AT76, BcDNA:GH25305, CG9602, Dmel\\CG9602, UBCb, Ubc87F', description='Ubiquitin conjugating enzyme 87F', other_designations='CG9602-PA|Ubc87F-PA|ubiquitin conjugating enzyme 87F|ubiquitin conjugating enzyme 87Fb', map_location='87F12-87F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13717454, end_position_on_the_genomic_accession=13718176, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41686, CurrentID=0, Status='live', Symbol='CG9312', Aliases='Dmel_ CT26517, Dmel\\CG9312', description='CG9312 gene product from transcript CG9312-RA', other_designations='CG9312-PA', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13765756, end_position_on_the_genomic_accession=13767103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41687, CurrentID=0, Status='live', Symbol='Cht5', Aliases='Dmel_CG9307, CG9307, DmCHT5, Dmel\\CG9307', description='CG9307 gene product from transcript CG9307-RA', other_designations='CG9307-PA|Cht5-PA|chitin-binding peritrophin-A', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13767549, end_position_on_the_genomic_accession=13770732, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41688, CurrentID=0, Status='live', Symbol='CG9297', Aliases='Dmel_ Dmel\\CG9297', description='CG9297 gene product from transcript CG9297-RC', other_designations='CG9297-PA|CG9297-PB|CG9297-PC|CG9297-PD|CG9297-PE', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13771495, end_position_on_the_genomic_accession=13779317, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  41689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41689, CurrentID=0, Status='live', Symbol='CG9288', Aliases='Dmel_ Dmel\\CG9288, Dromel_CG9288_FBtr0082829_pyruvatekinasedomain_mORF', description='CG9288 gene product from transcript CG9288-RA', other_designations='CG9288-PA', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13779624, end_position_on_the_genomic_accession=13780795, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41690, CurrentID=0, Status='live', Symbol='CG9286', Aliases='Dmel_ Dmel\\CG9286', description='CG9286 gene product from transcript CG9286-RA', other_designations='CG9286-PA', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13780964, end_position_on_the_genomic_accession=13782082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41693, CurrentID=0, Status='live', Symbol='Mst87F', Aliases='Dmel_CG17956, BEST:GH23081, CG17956, Dmel\\CG17956, Mst, bs35f12.y1, mst(3)87F, mst(3)gl-9, mst(3)gl9, mst-9, mst87F', description='Male-specific RNA 87F', other_designations='CG17956-PA|CG17956-PB|Mst87F-PA|Mst87F-PB', map_location='87F14-87F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13827151, end_position_on_the_genomic_accession=13828008, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41694, CurrentID=0, Status='live', Symbol='Nsf2', Aliases='Dmel_CG33101, CG33101, CG9931, DmNSF2, Dmel\\CG33101, EC3-11, NSF-2, NSF2, Nsf-2, dNSF, dNSF-2, dNSF2, dNsf2, dnsf2, lotus', description='N-ethylmaleimide-sensitive factor 2', other_designations='CG33101-PA|CG33101-PB|CG33101-PC|N-ethylmaleimide sensitive factor 2|NEM-sensitive fusion protein 2|Nsf2-PA|Nsf2-PB|Nsf2-PC|lotus', map_location='87F15-87F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13833497, end_position_on_the_genomic_accession=13837991, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318245, CurrentID=0, Status='live', Symbol='CG32847', Aliases='Dmel_ Dmel\\CG32847', description='CG32847 gene product from transcript CG32847-RB', other_designations='CG32847-PB|uncharacterized protein', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15121572, end_position_on_the_genomic_accession=15122301, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41697, CurrentID=0, Status='live', Symbol='ems', Aliases='Dmel_CG2988, CG2988, Dmel\\CG2988, E4, EMS, Ems, W13, ant', description='empty spiracles', other_designations='CG2988-PA|antenna|empty-spiracle|empty-spiracles|ems-PA', map_location='3-54.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13901853, end_position_on_the_genomic_accession=13904620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41698, CurrentID=0, Status='live', Symbol='Art9', Aliases='Dmel_CG9929, CG9929, DART9, Dmel\\CG9929', description='Arginine methyltransferase 9', other_designations='Art9-PA|CG9929-PA|arginine methyltransferase 9', map_location='88A2-88A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13942238, end_position_on_the_genomic_accession=13943460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41699, CurrentID=0, Status='live', Symbol='Art6', Aliases='Dmel_CG9927, CG9927, DART6, Dmel\\CG9927', description='Arginine methyltransferase 6', other_designations='Art6-PA|CG9927-PA|arginine methyltransferase 6', map_location='88A2-88A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13944654, end_position_on_the_genomic_accession=13945812, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41700, CurrentID=0, Status='live', Symbol='CG9926', Aliases='Dmel_ Dmel\\CG9926', description='CG9926 gene product from transcript CG9926-RB', other_designations='CG9926-PA|CG9926-PB', map_location='88A3-88A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13949454, end_position_on_the_genomic_accession=13951138, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41701, CurrentID=0, Status='live', Symbol='Ipp', Aliases='Dmel_CG3028, CG3028, Dmel\\CG3028, anon-WO02059370.48, ipp', description='Inositol polyphosphate 1-phosphatase', other_designations='CG3028-PA|Ipp-PA|inositol polyphosphate 1-phosphatase', map_location='88A3-88A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13958789, end_position_on_the_genomic_accession=13960474, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41702, CurrentID=0, Status='live', Symbol='CG9925', Aliases='Dmel_ Dmel\\CG9925', description='CG9925 gene product from transcript CG9925-RA', other_designations='CG9925-PA', map_location='88A3-88A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13960712, end_position_on_the_genomic_accession=13964174, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41703, CurrentID=0, Status='live', Symbol='Orc2', Aliases='Dmel_CG3041, CG3041, DmORC 2, DmORC2, DmOrc2, DmOrc2p, Dmel\\CG3041, ORC, ORC2, Orc-2p, dOrc2, fs(3)293, fs(3)293-19, k43, l(3)88Ab, l(3)A5, l(3)K43, l(3)SG44, l(3)dsl5, l(3)k43, orc2, rDmORC, Orc2', description='Origin recognition complex subunit 2', other_designations='CG3041-PA|Orc2-PA|lethal(3)88Ab|origin recognition complex|origin recognition complex 2|origin recognition complex subunit 2', map_location='3-51.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13964885, end_position_on_the_genomic_accession=13967075, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41704, CurrentID=0, Status='live', Symbol='rdx', Aliases='Dmel_CG12537, 0869/09, BEST:GM07940, CG10235, CG12537, CG9924, D-SPOP, Dmel\\CG12537, Dmel_CG9924, Dromel_CG9924_FBtr0082871_rdx_mORF, HIB, Rdx, SPOP, Spop, anon-WO0118547.577, hib, l(3)03477, l(3)2004, l(3)88Aa, l(3)A6, l(3)S086909, l(3)SG40, l(3)dsl-6, l(3)dsl6, l(3)j1D1, l(3)j4C7, mei-P19', description='roadkill', other_designations='CG12537-PA|CG12537-PB|CG12537-PC|CG12537-PE|CG12537-PF|Cul3-Hedghog-induced MATH and BTB domain-containing protein|Hh-induced MATH and BTB domain containing protein|hedegehog-induced BTB protein|hedgehog- induced BTB protein|lethal (3) S086909|meiotic P19|rdx-PA|rdx-PB|rdx-PC|rdx-PE|rdx-PF', map_location='3-50.3 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13967061, end_position_on_the_genomic_accession=14032057, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  41706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41706, CurrentID=0, Status='live', Symbol='Cyp6d5', Aliases='Dmel_CG3050, 6d5, CG3050, Dmel\\CG3050, cyp6d5', description='CG3050 gene product from transcript CG3050-RA', other_designations='CG3050-PA|Cyp6d5-PA|cytochrome P450', map_location='88A4-88A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14029228, end_position_on_the_genomic_accession=14031483, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41707, CurrentID=0, Status='live', Symbol='CG3061', Aliases='Dmel_ Dmel\\CG3061', description='CG3061 gene product from transcript CG3061-RA', other_designations='CG3061-PA|dnaJ like2', map_location='88A4-88A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14032371, end_position_on_the_genomic_accession=14035041, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41708, CurrentID=0, Status='live', Symbol='CG9922', Aliases='Dmel_ BcDNA:RE55111, Dmel\\CG9922, NEST:bs28g09', description='CG9922 gene product from transcript CG9922-RB', other_designations='CG9922-PA|CG9922-PB', map_location='88A4-88A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14035018, end_position_on_the_genomic_accession=14036398, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41709, CurrentID=0, Status='live', Symbol='foxo', Aliases='Dmel_CG3143, 3143, Afx, CG3143, Dfoxo, DmQ95V5, Dmel\\CG3143, FKHR, FOXO, FOXO1, FoxO, Foxo, Q95V55, anon-WO0118547.374, cg3143, dFOXO, dFoxO, dFoxo, dfoxo, fd88A, foxO|dFOXO, foxo', description='forkhead box, sub-group O', other_designations='CG3143-PB|CG3143-PC|CG3143-PF|CG3143-PG|CG3143-PH|forkhead Box O|forkhead bOX-containing protein, subfamily O|forkhead box|forkhead box sub-group O|forkhead box type O|forkhead box, subgroup O|forkhead box-containing protein, O subfamily|forkhead transcription factor|foxo-PB|foxo-PC|foxo-PF|foxo-PG|foxo-PH', map_location='88A5-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14056945, end_position_on_the_genomic_accession=14087978, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41710, CurrentID=0, Status='live', Symbol='Npc2b', Aliases='Dmel_CG3153, CG3153, CT10556, DmML2, Dmel\\CG3153, NPC2, npc2b', description='Niemann-Pick type C-2b', other_designations='CG3153-PA|CG3153-PB|Npc2b-PA|Npc2b-PB', map_location='88A8-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14089517, end_position_on_the_genomic_accession=14092425, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41711, CurrentID=0, Status='live', Symbol='CCHa1', Aliases='Dmel_CG14358, CCH1, CCM, CG14358, Dm-CCHa1, Dmel\\CG14358', description='CCHamide-1', other_designations='CCHa1-PB|CCHamide|CCHamide 1|CCHamide1|CG14358-PB|Drm-CCHamide-1', map_location='88A8-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14092504, end_position_on_the_genomic_accession=14096835, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41712, CurrentID=0, Status='live', Symbol='CG9920', Aliases='Dmel_ BEST:GH20473, Dmel\\CG9920, anon-WO0140519.48', description='CG9920 gene product from transcript CG9920-RA', other_designations='CG9920-PA', map_location='88A8-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14097293, end_position_on_the_genomic_accession=14098080, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41713, CurrentID=0, Status='live', Symbol='PK1-R', Aliases='Dmel_CG9918, CG9918, Dmel\\CG9918, Pk1R, Pk1r, anon-WO0170980.112, anon-WO0170980.113', description='Pyrokinin 1 receptor', other_designations='CG9918-PD|CG9918-PE|Drm-PK-1 receptor|Lst receptor|PK1-R-PD|PK1-R-PE|pyrokinin 1 receptor|pyrokinin-1 receptor', map_location='88A8-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14099832, end_position_on_the_genomic_accession=14103122, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41714, CurrentID=0, Status='live', Symbol='CG12402', Aliases='Dmel_ Dmel\\CG12402, FBXL11', description='CG12402 gene product from transcript CG12402-RA', other_designations='CG12402-PA|CG12402-PB|CG12402-PC|uncharacterized protein', map_location='88A8-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14104488, end_position_on_the_genomic_accession=14108321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41715, CurrentID=0, Status='live', Symbol='Or88a', Aliases='Dmel_CG14360, 88A.1, 88a, CG14360, DOR99, DmOr88a, Dmel\\CG14360, or88a', description='Odorant receptor 88a', other_designations='CG14360-PA|Or88a-PA|odorant receptor 88a', map_location='88A8-88A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14109445, end_position_on_the_genomic_accession=14110823, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41716, CurrentID=0, Status='live', Symbol='CG14357', Aliases='Dmel_ Dmel\\CG14357', description='CG14357 gene product from transcript CG14357-RB', other_designations='CG14357-PB|uncharacterized protein', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14111151, end_position_on_the_genomic_accession=14112166, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41717, CurrentID=0, Status='live', Symbol='Kif19A', Aliases='Dmel_CG9913, CG9913, DmKIF19A, DmKlp19A, Dmel\\CG9913, KIF 19A, KIF19A', description='Kinesin family member 19A ortholog', other_designations='CG9913-PB|Kif19A-PB|kinesin family member 19A ortholog|kinesin superfamily protein 19A', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14112451, end_position_on_the_genomic_accession=14118450, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41718, CurrentID=0, Status='live', Symbol='Abi', Aliases='Dmel_CG9749-1, Ablphilin, CG9749, Dmel\\CG9749, abi, dAbi, Abi', description='Abelson interacting protein', other_designations='Abi-PA|Abi-PB|Abi-PC|Abi-PD|Abi-PE|Abi-PF|CG9749-PA|CG9749-PB|CG9749-PC|CG9749-PD|CG9749-PE|CG9749-PF|abelson interacting protein|abelson interactor', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14119262, end_position_on_the_genomic_accession=14122420, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41719, CurrentID=0, Status='live', Symbol='twf', Aliases='Dmel_CG3172, A6, CG3172, Dmel\\CG3172, Twf', description='twinfilin', other_designations='CG3172-PA|twf-PA|twinfillin', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14122654, end_position_on_the_genomic_accession=14124474, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41720, CurrentID=0, Status='live', Symbol='140up', Aliases='Dmel_CG9852, CG9852, DmRP140-upstream, DmRP140up, Dmel\\CG9852, l(3)88Ba, l(3)Z6', description='upstream of RpII140', other_designations='140up-PA|140up-PB|CG9852-PA|CG9852-PB', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14124163, end_position_on_the_genomic_accession=14125548, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41721, CurrentID=0, Status='live', Symbol='RpII140', Aliases='Dmel_CG3180, CG3180, DmRP140, Dmel\\CG3180, H5, II, Pol II, Pol IIc, PolII, PolIIc, RNA Pol II, RNA Pol II 140, RNA Pol IIc, RNA pol II, RNA polII, RNAP, RNAP II, RNApolII, RP140, RPB2_DROME, RPII140[wimp], Rpb2, Rpll140, l(3)RplII140, pol II, polII, rpII140/wimp, wimp, RpII140', description='RNA polymerase II 140kD subunit', other_designations='150kDa polypeptide|CG3180-PA|CG3180-PB|Pol II|RNA pol II second largest subunit|RNA polymerase II|RNA polymerase II 140 kD subunit|RNA polymerase II second largest subunit|RNA polymerase IIc|RNA polymerase subunit IIc|RNA polymeraseII-140 kd subunit|RpII140-PA|RpII140-PB|Rpb2|subunit IIc of RNA Pol II|wimp', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14125791, end_position_on_the_genomic_accession=14130111, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41722, CurrentID=0, Status='live', Symbol='CG14356', Aliases='Dmel_ Dmel\\CG14356', description='CG14356 gene product from transcript CG14356-RA', other_designations='CG14356-PA|uncharacterized protein', map_location='88A9-88A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14130680, end_position_on_the_genomic_accession=14132072, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41723, CurrentID=0, Status='live', Symbol='CG14355', Aliases='Dmel_ Dmel\\CG14355', description='CG14355 gene product from transcript CG14355-RA', other_designations='CG14355-PA|uncharacterized protein', map_location='88A9-88A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14148508, end_position_on_the_genomic_accession=14151940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41724, CurrentID=0, Status='live', Symbol='CG9722', Aliases='Dmel_ Dmel\\CG9722', description='CG9722 gene product from transcript CG9722-RA', other_designations='CG9722-PA', map_location='88A10-88A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14152885, end_position_on_the_genomic_accession=14153829, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41725, CurrentID=0, Status='live', Symbol='CG3199', Aliases='Dmel_ Dmel\\CG3199', description='CG3199 gene product from transcript CG3199-RA', other_designations='CG3199-PA', map_location='88A10-88A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14162709, end_position_on_the_genomic_accession=14163625, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41726, CurrentID=0, Status='live', Symbol='Dop1R1', Aliases='Dmel_CG9652, CG9652, D-Dop1, DA1, DR1, DmDOP1, DmDop1, Dmdop1, Dmel\\CG9652, DopR, DopR1, DopR35F, dDA1, dDA1/DmDOP1, dumb', description='Dopamine 1-like receptor 1', other_designations='CG9652-PB|CG9652-PC|CG9652-PD|CG9652-PE|CG9652-PF|D1 dopamine receptor|D1-Like dopamine receptor|D1-like DA receptor|D1-like receptor|DA D[[1]]-like receptor|D[[1]] dopamine receptor|Dop1R1-PB|Dop1R1-PC|Dop1R1-PD|Dop1R1-PE|Dop1R1-PF|dopamine 1 receptor|dopamine 1-like receptor 1|dopamine D1 receptor|dopamine D1-like receptor|dopamine receptor|dumb|dumb DA receptor', map_location='88A10-88A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14165379, end_position_on_the_genomic_accession=14214810, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41727, CurrentID=0, Status='live', Symbol='CG9649', Aliases='Dmel_ Dmel\\CG9649, SP55', description='CG9649 gene product from transcript CG9649-RA', other_designations='CG9649-PA', map_location='88A12-88A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14215459, end_position_on_the_genomic_accession=14217882, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41728, CurrentID=0, Status='live', Symbol='CG31326', Aliases='Dmel_ CG9645, Dmel\\CG31326, SPH144, cg9645', description='CG31326 gene product from transcript CG31326-RA', other_designations='CG31326-PA|uncharacterized protein', map_location='88A12-88A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14218136, end_position_on_the_genomic_accession=14220205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41729, CurrentID=0, Status='live', Symbol='CG9631', Aliases='Dmel_ BcDNA:GH08420, Dmel\\CG9631, SP60', description='CG9631 gene product from transcript CG9631-RA', other_designations='CG9631-PA', map_location='88A12-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14225098, end_position_on_the_genomic_accession=14226762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41730, CurrentID=0, Status='live', Symbol='CG9624', Aliases='Dmel_ Dmel\\CG9624', description='CG9624 gene product from transcript CG9624-RA', other_designations='CG9624-PA', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14228018, end_position_on_the_genomic_accession=14229145, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41731, CurrentID=0, Status='live', Symbol='CG9616', Aliases='Dmel_ Dmel\\CG9616', description='CG9616 gene product from transcript CG9616-RB', other_designations='CG9616-PB', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14244804, end_position_on_the_genomic_accession=14245471, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41733, CurrentID=0, Status='live', Symbol='Aats-met-m', Aliases='Dmel_CG31322, Aats-met, CG31322, CG8684, CG9612, Dmel\\CG31322, MRS, MetRS', description='mitochondrial methionyl-tRNA synthetase', other_designations='Aats-met-m-PA|Aats-met-m-PB|Aats-met-m-PC|CG31322-PA|CG31322-PB|CG31322-PC|Methionyl-tRNA synthetase', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14249043, end_position_on_the_genomic_accession=14259991, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41734, CurrentID=0, Status='live', Symbol='CG14840', Aliases='Dmel_ Dmel\\CG14840', description='CG14840 gene product from transcript CG14840-RA', other_designations='CG14840-PA|uncharacterized protein', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14260208, end_position_on_the_genomic_accession=14261213, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41735, CurrentID=0, Status='live', Symbol='CG14841', Aliases='Dmel_ BcDNA:AT04690, Dmel\\CG14841', description='CG14841 gene product from transcript CG14841-RA', other_designations='CG14841-PA|uncharacterized protein', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14261391, end_position_on_the_genomic_accession=14262418, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41736, CurrentID=0, Status='live', Symbol='CG14839', Aliases='Dmel_ Dmel\\CG14839', description='CG14839 gene product from transcript CG14839-RA', other_designations='CG14839-PA|uncharacterized protein', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14262387, end_position_on_the_genomic_accession=14263367, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41737, CurrentID=0, Status='live', Symbol='trx', Aliases='Dmel_CG8651, 4720, 4733, CG8651, DMTRXIII, Dmel\\CG8651, NR0A5, R-bx, Rg-bx, TRX, Trx, Trx-G, l(3)bx[v], l(3)j14A6, l(3)j3A2, l(3)s5452, tqrx-g, trxI and trxII, trx', description='trithorax', other_designations='CG8651-PA|CG8651-PB|CG8651-PC|CG8651-PD|CG8651-PE|regulator of bithorax|trx-PA|trx-PB|trx-PC|trx-PD|trx-PE', map_location='3-54.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14263358, end_position_on_the_genomic_accession=14286903, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41738, CurrentID=0, Status='live', Symbol='CG12207', Aliases='Dmel_ Dmel\\CG12207', description='CG12207 gene product from transcript CG12207-RG', other_designations='CG12207-PA|CG12207-PB|CG12207-PC|CG12207-PD|CG12207-PE|CG12207-PF|CG12207-PG|uncharacterized protein', map_location='88B1-88B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14289071, end_position_on_the_genomic_accession=14302276, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41739, CurrentID=0, Status='live', Symbol='CG3259', Aliases='Dmel_ Dmel\\CG3259, IFT54, MIP-T3', description='CG3259 gene product from transcript CG3259-RA', other_designations='CG3259-PA', map_location='88B3-88B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14302428, end_position_on_the_genomic_accession=14304458, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41740, CurrentID=0, Status='live', Symbol='su(Hw)', Aliases='Dmel_CG8573, CG8573, Dmel\\CG8573, SU(HW), SUHW, Su(HW), Su(Hw), Su(hw), Su[Hw], su(HW), su(hw), suHw', description='suppressor of Hairy wing', other_designations='CG8573-PA|CG8573-PB|CG8573-PC|su(Hw)-PA|su(Hw)-PB|su(Hw)-PC|suppressor of Hair-Wing|suppressor of hairy wing|suppressor of hairy-wing', map_location='3-54.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14304455, end_position_on_the_genomic_accession=14308589, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41741, CurrentID=0, Status='live', Symbol='RpII15', Aliases='Dmel_CG3284, CG3284, Dmel\\CG3284, H5, Pol II, PolII, RNA Pol II, RNA pol II, RNA polII, RNAP, RNAP II, RNApolII, RPII15, RpII215, Rpb9, dRPB9, l(3)88Be, l(3)Z23, pol II, polII', description='RNA polymerase II 15kD subunit', other_designations='CG3284-PA|CG3284-PB|CG3284-PC|RNA polymerase II|RNA polymerase II 15kd subunit|RpII15-PA|RpII15-PB|RpII15-PC', map_location='88B3-88B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14309089, end_position_on_the_genomic_accession=14310178, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41742, CurrentID=0, Status='live', Symbol='Crz', Aliases='Dmel_CG3302, CG3302, COR, CORZ, CRZ, Cora, Dm-Crz, Dmel\\CG3302, Drm-COR, cor, crz', description='Corazonin', other_designations='CG3302-PA|Crz-PA|corazonin|preprocorazonin', map_location='88B3-88B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14314854, end_position_on_the_genomic_accession=14315716, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41743, CurrentID=0, Status='live', Symbol='Set8', Aliases='Dmel_CG3307, CG3307, Dmel\\CG3307, PR-SET7, PR-Set7, PR/SET07, PRSET7, Pr-set-7, SET8, Set7, dPR-Set7, dSET8, dSet8, l(3)neo41, pr-Set7, pr-set7', description='SET domain containing 8', other_designations='CG3307-PA|CG3307-PB|CG3307-PD|Set8-PA|Set8-PB|Set8-PD|pr-set7', map_location='88B3-88B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14316038, end_position_on_the_genomic_accession=14320494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41744, CurrentID=0, Status='live', Symbol='Afti', Aliases='Dmel_CG8538, CG8538, Dmel\\CG8538, dAfti', description='Aftiphilin', other_designations='Afti-PB|Afti-PC|Afti-PD|CG8538-PB|CG8538-PC|CG8538-PD|aftiphilin|dAftiphilin', map_location='88B3-88B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14320191, end_position_on_the_genomic_accession=14326689, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41745, CurrentID=0, Status='live', Symbol='ATPsynE', Aliases='Dmel_CG3321, ATPJ, CG3321, Dmel\\CG3321, anon-88Bd, c550', description='ATP synthase, subunit E', other_designations='ATP synthase e chain|ATPsynE-PA|ATPsynE-PB|ATPsynE-PC|CG3321-PA|CG3321-PB|CG3321-PC|F0/F1 ATP synthase E', map_location='88B4-88B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14326947, end_position_on_the_genomic_accession=14327735, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41746, CurrentID=0, Status='live', Symbol='spn-B', Aliases='Dmel_CG3325, CG3325, Dmel\\CG3325, Rad51, Rad51-like, Spn-B, SpnB, XRCC3, rad51-like, spnB', description='spindle B', other_designations='CG3325-PA|Spindle-B|SpindleB|spn-B-PA', map_location='88B4-88B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14328051, end_position_on_the_genomic_accession=14329608, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41747, CurrentID=0, Status='live', Symbol='NK7.1', Aliases='Dmel_CG8524, CG8524, Dmel\\CG8524, NK-7.1', description='CG8524 gene product from transcript CG8524-RC', other_designations='CG8524-PA|CG8524-PB|CG8524-PC|CG8524-PD|Dm.HboxNK-7.1|NK7.1-PA|NK7.1-PB|NK7.1-PC|NK7.1-PD', map_location='88B4-88B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14329413, end_position_on_the_genomic_accession=14372951, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41748, CurrentID=0, Status='live', Symbol='soti', Aliases='Dmel_CG8489, CG8489, Dmel\\CG8489', description='scotti', other_designations='CG8489-PA|soti-PA', map_location='88B7-88B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14380663, end_position_on_the_genomic_accession=14381699, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41749, CurrentID=0, Status='live', Symbol='cv-c', Aliases='Dmel_CG34389, CG14845, CG14846, CG14847, CG14848, CG14849, CG31319, CG31330, CG34389, CG8480, Cv-c, Dmel\\CG34389, Dmel_CG14846, Dmel_CG14847, Dmel_CG31319, Dmel_CG31330, RhoGAP88C, l(3)06951, rhoGAP88C', description='crossveinless c', other_designations='CG34389-PA|CG34389-PC|CG34389-PD|crossveinless-c|cv-c-PA|cv-c-PC|cv-c-PD|lethal (3) 06951', map_location='3-54.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14391683, end_position_on_the_genomic_accession=14481713, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  41756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41756, CurrentID=0, Status='live', Symbol='HtrA2', Aliases='Dmel_CG8464, CG8464, DmHtrA2, Dmel\\CG8464/Omi, Omi, dOmi, domi, htrA2, htrA2/omi, omi, HtrA2', description='HTRA2-related serine protease', other_designations='CG8464-PA|CG8464-PB|HtrA2-PA|HtrA2-PB|serine protease HTRA2', map_location='88C3-88C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14509049, end_position_on_the_genomic_accession=14511712, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41757, CurrentID=0, Status='live', Symbol='mRpL11', Aliases='Dmel_CG3351, BcDNA:RE67638, CG3351, Dmel\\CG3351, L11, rpl-11', description='mitochondrial ribosomal protein L11', other_designations='CG3351-PA|mRpL11-PA', map_location='88C3-88C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14511838, end_position_on_the_genomic_accession=14512691, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41758, CurrentID=0, Status='live', Symbol='CG8461', Aliases='Dmel_ Dmel\\CG8461', description='CG8461 gene product from transcript CG8461-RA', other_designations='CG8461-PA', map_location='88C3-88C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14512639, end_position_on_the_genomic_accession=14513637, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41759, CurrentID=0, Status='live', Symbol='Cyp313a1', Aliases='Dmel_CG3360, 313a1, 313a1p, CG3360, Cyp313A1, Cyp313ap, Dmel\\CG3360', description='CG3360 gene product from transcript CG3360-RA', other_designations='CG3360-PA|Cyp313a1-PA', map_location='88C4-88C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14521049, end_position_on_the_genomic_accession=14523064, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41760, CurrentID=0, Status='live', Symbol='Pde6', Aliases='Dmel_CG8279, CG8279, DmPDE5/6, DmPDE6, Dmel\\CG8279, PDE6', description='Phosphodiesterase 6', other_designations='CG8279-PB|CG8279-PC|CG8279-PD|Pde6-PB|Pde6-PC|Pde6-PD|phosphodiesterase 6', map_location='88C3-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14513901, end_position_on_the_genomic_accession=14558304, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  41761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41761, CurrentID=0, Status='live', Symbol='CG14854', Aliases='Dmel_ Dmel\\CG14854', description='CG14854 gene product from transcript CG14854-RB', other_designations='CG14854-PB|uncharacterized protein', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14559166, end_position_on_the_genomic_accession=14560054, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41762, CurrentID=0, Status='live', Symbol='CG14850', Aliases='Dmel_ Dmel\\CG14850', description='CG14850 gene product from transcript CG14850-RA', other_designations='CG14850-PA|uncharacterized protein', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14560935, end_position_on_the_genomic_accession=14561521, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41763, CurrentID=0, Status='live', Symbol='CG14851', Aliases='Dmel_ Dmel\\CG14851', description='CG14851 gene product from transcript CG14851-RA', other_designations='CG14851-PA|uncharacterized protein', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14562340, end_position_on_the_genomic_accession=14562845, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41764, CurrentID=0, Status='live', Symbol='CG8087', Aliases='Dmel_ Dmel\\CG8087', description='CG8087 gene product from transcript CG8087-RA', other_designations='CG8087-PA', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14564323, end_position_on_the_genomic_accession=14564805, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41765, CurrentID=0, Status='live', Symbol='CG14852', Aliases='Dmel_ Dmel\\CG14852', description='CG14852 gene product from transcript CG14852-RA', other_designations='CG14852-PA|uncharacterized protein', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14565927, end_position_on_the_genomic_accession=14566513, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41766, CurrentID=0, Status='live', Symbol='CG8066', Aliases='Dmel_ Dmel\\CG8066', description='CG8066 gene product from transcript CG8066-RA', other_designations='CG8066-PA|CG8066-PB', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14567611, end_position_on_the_genomic_accession=14568847, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41767, CurrentID=0, Status='live', Symbol='Cys', Aliases='Dmel_CG8050, BcDNA:RH72992, CG8050, Dmel\\CG8050, anon-EST:Liang-2.34, clone 2.34', description='Cystatin-like', other_designations='CG8050-PA|Cys-PA|cystatin', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14569103, end_position_on_the_genomic_accession=14569644, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41768, CurrentID=0, Status='live', Symbol='CG7987', Aliases='Dmel_ Dmel\\CG7987, anon-WO0118547.245', description='CG7987 gene product from transcript CG7987-RA', other_designations='CG7987-PA|CG7987-PB', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14570472, end_position_on_the_genomic_accession=14575887, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41770, CurrentID=0, Status='live', Symbol='stumps', Aliases='Dmel_CG31317, 0837/10, C, CG18485, CG31317, CG3375, DOF, Dmel\\CG31317, Dof, Dof/stumps/heartbroken, Hbr, Hbr/Dof, P1740, anon-EST:CL47, anon-estC, dof, dof/hbr, dof/hbr/sms, dof1, gene C, hbr, heartbroken/dof, i21, i28, l(3)09904, l(3)S083710, sms/dof', description='CG31317 gene product from transcript CG31317-RC', other_designations='CG31317-PB|CG31317-PC|CG31317-PD|CG31317-PE|CG31317-PF|Downstream-of-FGFR|downstream of FGF|downstream of FGF receptor|downstream of FGFR|downstream of FGFR/Heartbroken/Stumps|downstream-of-FGF receptor|heartbroken|stumps-PB|stumps-PC|stumps-PD|stumps-PE|stumps-PF', map_location='88C6-88C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14577217, end_position_on_the_genomic_accession=14608079, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41771, CurrentID=0, Status='live', Symbol='CG14853', Aliases='Dmel_ Dmel\\CG14853', description='CG14853 gene product from transcript CG14853-RB', other_designations='CG14853-PA|CG14853-PB|CG14853-PC|uncharacterized protein', map_location='88C9-88C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14609642, end_position_on_the_genomic_accession=14618987, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41772, CurrentID=0, Status='live', Symbol='put', Aliases='Dmel_CG7904, Act-r, Atr, Atr-II, Atr88CD, AtrII, CG7904, Dmel\\CG7904, PUNT, Punt, Put, STK-C, TGF-B, TGF-beta, Tgf-r, atrII, dlhC, l(3)10460, l(3)j5A5, pun', description='punt', other_designations='Activin-A-receptor-88CD|CG7904-PA|CG7904-PB|CG7904-PC|DPP receptor|TGF-beta-type-receptor-like|activin receptor|activin receptor type II|dorsal holes C|put-PA|put-PB|put-PC', map_location='3-58 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14619030, end_position_on_the_genomic_accession=14625703, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41773, CurrentID=0, Status='live', Symbol='His4r', Aliases='Dmel_CG3379, BcDNA:RH52884, CG3379, Dmel\\CG3379, FBtr0082962, H4r, His4-88CD, His4R', description='Histone H4 replacement', other_designations='CG3379-PA|CG3379-PB|CG3379-PC|CG3379-PD|His4r-PA|His4r-PB|His4r-PC|His4r-PD|histone H4 replacement', map_location='88C9-88C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14625755, end_position_on_the_genomic_accession=14626614, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41774, CurrentID=0, Status='live', Symbol='Cad88C', Aliases='Dmel_CG3389, CG3389, CT11389, DCad88C, Dm Cad88C, Dmel\\CG3389, cad88C', description='Cadherin 88C', other_designations='CG3389-PB|Cad88C-PB|cadherin 88C', map_location='88C10-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14630478, end_position_on_the_genomic_accession=14641776, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  41775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41775, CurrentID=0, Status='live', Symbol='CG7886', Aliases='Dmel_ Dmel\\CG7886, anon-WO0140519.200', description='CG7886 gene product from transcript CG7886-RA', other_designations='CG7886-PA', map_location='88C9-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14626676, end_position_on_the_genomic_accession=14642563, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41776, CurrentID=0, Status='live', Symbol='l(3)L1231', Aliases='Dmel_CG7832, BEST:LD14744, CG7832, Dmel\\CG7832, LD14744', description='lethal (3) L1231', other_designations='CG7832-PA|CG7832-PC|CG7832-PD|l(3)L1231-PA|l(3)L1231-PC|l(3)L1231-PD', map_location='88C10-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14642672, end_position_on_the_genomic_accession=14655716, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41777, CurrentID=0, Status='live', Symbol='CG3505', Aliases='Dmel_ Dmel\\CG3505, SPH9, proPO-AE', description='CG3505 gene product from transcript CG3505-RA', other_designations='CG3505-PA', map_location='88C10-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14657447, end_position_on_the_genomic_accession=14659228, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41778, CurrentID=0, Status='live', Symbol='Hexim', Aliases='Dmel_CG3508, CG3508, Dmel\\CG3508, dHEXIM', description='HEXIM ortholog', other_designations='CG3508-PA|CG3508-PB|Hexim-PA|Hexim-PB', map_location='88C10-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14659438, end_position_on_the_genomic_accession=14661177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41779, CurrentID=0, Status='live', Symbol='Rad17', Aliases='Dmel_CG7825, CG7825, Dm/24, Dmel\\CG7825', description='CG7825 gene product from transcript CG7825-RA', other_designations='CG7825-PA|Rad17-PA', map_location='88C10-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14661176, end_position_on_the_genomic_accession=14663197, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41780, CurrentID=0, Status='live', Symbol='BigH1', Aliases='Dmel_CG3509, CG3509, Dmel\\CG3509, dBigH1', description='Histone H1 variant BigH1', other_designations='BigH1-PA|CG3509-PA|histone H1 variant BigH1', map_location='88C10-88C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14663450, end_position_on_the_genomic_accession=14664822, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41783, CurrentID=0, Status='live', Symbol='kibra', Aliases='Dmel_CG33967, CG12600, CG33967, CG7552, Dmel\\CG33967, Kib, Kibra, ORE-2, RE26350, kbr, kib, l(3)02404, l(3)j5C3, l(3)neo42', description='kibra ortholog', other_designations='CG33967-PA|CG33967-PB|kibra-PA|kibra-PB|lethal (3) neo42', map_location='88D1-88D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14697689, end_position_on_the_genomic_accession=14724261, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41784, CurrentID=0, Status='live', Symbol='CG7530', Aliases='Dmel_ Dmel\\CG7530', description='CG7530 gene product from transcript CG7530-RB', other_designations='CG7530-PA|CG7530-PB|CG7530-PC|CG7530-PD', map_location='88D1-88D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14724760, end_position_on_the_genomic_accession=14731509, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41785, CurrentID=0, Status='live', Symbol='eff', Aliases='Dmel_CG7425, 3527, 6535, CG7425, Dmel\\CG7425, E(Pc)88D, EFF, Eff, SS3-3, UBC4, UBCD-1, UBCD1, Ubc-D1, UbcD, UbcD1, anon-WO0118547.343, fs(3)01295, l(3)01462, l(3)j2C8, l(3)s1782, mei68, ms(3)01295, ms(3)1, ms(3)3, ms(3)88D, ms(3)neo2, ms(3)ry1, ms(3)ry3, ubcD1', description='effete', other_designations='CG7425-PA|CG7425-PB|CG7425-PC|eff-PA|eff-PB|eff-PC|effette|meiotic 68|suppressor of GMR-sina 3-3|ubiquitin conjugating enzyme 1', map_location='88D2-88D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14732414, end_position_on_the_genomic_accession=14741319, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41786, CurrentID=0, Status='live', Symbol='smp-30', Aliases='Dmel_CG7390, CG7390, CT22723, Dca, Dmel\\CG7390, Smp-30', description='Senescence marker protein-30', other_designations='CG7390-PA|CG7390-PB|CG7390-PC|CG7390-PD|cold acclimation|drosophila cold acclimation gene|senescence marker protein-30|smp-30-PA|smp-30-PB|smp-30-PC|smp-30-PD', map_location='88D2-88D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14745901, end_position_on_the_genomic_accession=14751246, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41787, CurrentID=0, Status='live', Symbol='CG7362', Aliases='Dmel_ Dmel\\CG7362', description='CG7362 gene product from transcript CG7362-RB', other_designations='CG7362-PB', map_location='88D2-88D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14751748, end_position_on_the_genomic_accession=14754460, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41788, CurrentID=0, Status='live', Symbol='jvl', Aliases='Dmel_CG43162, CG33330, CG3563, CG43162, Dmel\\CG43162, Dmel_CG33330, Dmel_CG3563, anon-WO0153538.4, anon-WO0153538.5', description='javelin-like', other_designations='CG43162-PA|CG43162-PB|CG43162-PC|CG43162-PD|CG43162-PE|CG43162-PF|CG43162-PG|Javelin-Like|javelinlike|jvl-PA|jvl-PB|jvl-PC|jvl-PD|jvl-PE|jvl-PF|jvl-PG', map_location='3-56.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14741582, end_position_on_the_genomic_accession=14795868, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  41789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41789, CurrentID=0, Status='live', Symbol='CG14855', Aliases='Dmel_ Dmel\\CG14855', description='CG14855 gene product from transcript CG14855-RA', other_designations='CG14855-PA|uncharacterized protein', map_location='88D3-88D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14796305, end_position_on_the_genomic_accession=14798474, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41790, CurrentID=0, Status='live', Symbol='CG14856', Aliases='Dmel_ Dmel\\CG14856', description='CG14856 gene product from transcript CG14856-RA', other_designations='CG14856-PA|uncharacterized protein', map_location='88D4-88D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14798986, end_position_on_the_genomic_accession=14801163, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41791, CurrentID=0, Status='live', Symbol='CG14857', Aliases='Dmel_ Dmel\\CG14857', description='CG14857 gene product from transcript CG14857-RA', other_designations='CG14857-PA|uncharacterized protein', map_location='88D4-88D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14801773, end_position_on_the_genomic_accession=14804342, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41795, CurrentID=0, Status='live', Symbol='CG3610', Aliases='Dmel_ Dmel\\CG3610', description='CG3610 gene product from transcript CG3610-RA', other_designations='CG3610-PA', map_location='88D5-88D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14845043, end_position_on_the_genomic_accession=14846737, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41797, CurrentID=0, Status='live', Symbol='CG3631', Aliases='Dmel_ Dmel\\CG3631', description='CG3631 gene product from transcript CG3631-RC', other_designations='CG3631-PC|CG3631-PD|CG3631-PE', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14879457, end_position_on_the_genomic_accession=14882631, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41798, CurrentID=0, Status='live', Symbol='Rrp6', Aliases='Dmel_CG7292, CG7292, Dm Dmel\\CG7292, dRrp6, rrp6', description='CG7292 gene product from transcript CG7292-RC', other_designations='CG7292-PB|CG7292-PC|Rrp6-PB|Rrp6-PC', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14882498, end_position_on_the_genomic_accession=14886246, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41800, CurrentID=0, Status='live', Symbol='UQCR-C1', Aliases='Dmel_CG3731, CG 3731, CG3731, Dmel\\CG3731, MPPB, UCR1', description='Ubiquinol-cytochrome c reductase core protein 1', other_designations='CG3731-PA|CG3731-PB|UQCR-C1-PA|UQCR-C1-PB|Ubiquinol-cytochrome C reductase complex core protein I|Ubiquinol-cytochrome c oxidoreductase core protein 1|mitochondrial processing peptidase beta subunit', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14890231, end_position_on_the_genomic_accession=14892387, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41801, CurrentID=0, Status='live', Symbol='CycC', Aliases='Dmel_CG7281, CG1C_DROME, CG7281, CYCC, Cln3, Cyc-C, DmcycC, Dmel\\CG7281, SRB11, cycC, cyclinC, dCycC, dTRAP33', description='Cyclin C', other_designations='CG7281-PA|CycC-PA|cyclin C', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14889004, end_position_on_the_genomic_accession=14890205, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41802, CurrentID=0, Status='live', Symbol='CG7265', Aliases='Dmel_ Dmel\\CG7265', description='CG7265 gene product from transcript CG7265-RA', other_designations='CG7265-PA', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14892392, end_position_on_the_genomic_accession=14893954, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41803, CurrentID=0, Status='live', Symbol='CG14860', Aliases='Dmel_ Dmel\\CG14860', description='CG14860 gene product from transcript CG14860-RB', other_designations='CG14860-PB|uncharacterized protein', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14894784, end_position_on_the_genomic_accession=14896446, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41804, CurrentID=0, Status='live', Symbol='Nup93-2', Aliases='Dmel_CG7262, CG7262, Cg7262, Dmel\\CG7262, Nup93, nup93', description='Nucleoporin 93kD-2', other_designations='CG7262-PA|Nup93-2-PA|nucleoporin 93|nucleoporin 93kD-2', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14895703, end_position_on_the_genomic_accession=14898372, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41805, CurrentID=0, Status='live', Symbol='CG3817', Aliases='Dmel_ Dmel\\CG3817', description='CG3817 gene product from transcript CG3817-RA', other_designations='CG3817-PA', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14898739, end_position_on_the_genomic_accession=14899617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41806, CurrentID=0, Status='live', Symbol='VhaPPA1-2', Aliases='Dmel_CG7026, CG7026, Dmel\\CG7026, vhaPPA1-2', description='Vacuolar H[+] ATPase PPA1 subunit 2', other_designations='CG7026-PB|VhaPPA1-2-PB|vacuolar H[+] ATPase PPA1 subunit 2|vacuolar H[+] ATPase subunit PPA1-2', map_location='88D6-88E', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14899606, end_position_on_the_genomic_accession=14900277, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41807, CurrentID=0, Status='live', Symbol='RpS5b', Aliases='Dmel_CG7014, 142157_at, CG7014, Dmel\\CG7014, RS5B_DROME, RbS5b, S5b', description='Ribosomal protein S5b', other_designations='CG7014-PA|CG7014-PB|RpS5b-PA|RpS5b-PB|ribosomal protein S5b', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14900432, end_position_on_the_genomic_accession=14901563, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41809, CurrentID=0, Status='live', Symbol='Sdr', Aliases='Dmel_CG3837, CG3837, Dmel\\CG3837', description='Secreted decoy of InR', other_designations='CG3837-PA|CG3837-PB|Sdr-PA|Sdr-PB|secreted decoy of InR', map_location='88D7-88D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14918628, end_position_on_the_genomic_accession=14923083, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41810, CurrentID=0, Status='live', Symbol='CG14861', Aliases='Dmel_ Dmel\\CG14861', description='CG14861 gene product from transcript CG14861-RA', other_designations='CG14861-PA|uncharacterized protein', map_location='88D9-88D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15014770, end_position_on_the_genomic_accession=15016034, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41811, CurrentID=0, Status='live', Symbol='RpL10Aa', Aliases='Dmel_CG3843, BcDNA:AT11516, CG3843, Dmel\\CG3843, Rpl10Aa', description='Ribosomal protein L10Aa', other_designations='CG3843-PA|RpL10Aa-PA|ribosomal protein L10Aa', map_location='88D10-88D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15037366, end_position_on_the_genomic_accession=15038149, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41815, CurrentID=0, Status='live', Symbol='CG6974', Aliases='Dmel_ Dmel\\CG6974', description='CG6974 gene product from transcript CG6974-RA', other_designations='CG6974-PA', map_location='88E1-88E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15078530, end_position_on_the_genomic_accession=15079285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41816, CurrentID=0, Status='live', Symbol='CG6966', Aliases='Dmel_ BcDNA:AT05246, Dmel\\CG6966', description='CG6966 gene product from transcript CG6966-RC', other_designations='CG6966-PB|CG6966-PC', map_location='88E1-88E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15079320, end_position_on_the_genomic_accession=15095631, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41817, CurrentID=0, Status='live', Symbol='CG42788', Aliases='Dmel_ CG18442, CG31304, CG6934, CG6960, Dmel\\CG42788, Dmel_CG31304, Dmel_CG6934, anon-WO0118547.79', description='CG42788 gene product from transcript CG42788-RC', other_designations='CG42788-PB|CG42788-PC|uncharacterized protein', map_location='88E1-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15095957, end_position_on_the_genomic_accession=15131926, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  41820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41820, CurrentID=0, Status='live', Symbol='CG6912', Aliases='Dmel_ Dmel\\CG6912, anon-WO0118547.571, anon-WO0140519.208', description='CG6912 gene product from transcript CG6912-RA', other_designations='CG6912-PA', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15135138, end_position_on_the_genomic_accession=15137004, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41821, CurrentID=0, Status='live', Symbol='CG3984', Aliases='Dmel_ Dmel\\CG3984', description='CG3984 gene product from transcript CG3984-RA', other_designations='CG3984-PA', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15137148, end_position_on_the_genomic_accession=15139007, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41822, CurrentID=0, Status='live', Symbol='CG3987', Aliases='Dmel_ Dmel\\CG3987, anon-WO0140519.215', description='CG3987 gene product from transcript CG3987-RA', other_designations='CG3987-PA', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15139958, end_position_on_the_genomic_accession=15141715, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41823, CurrentID=0, Status='live', Symbol='GlyS', Aliases='Dmel_CG6904, CG6904, CG6940, Dmel\\CG6904, GS, anon-WO0118547.108, dGS', description='Glycogen synthase', other_designations='CG6904-PA|CG6904-PB|CG6904-PC|CG6904-PD|GlyS-PA|GlyS-PB|GlyS-PC|GlyS-PD|glycogen synthase', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15141843, end_position_on_the_genomic_accession=15147531, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41824, CurrentID=0, Status='live', Symbol='Mf', Aliases='Dmel_CG6803, BcDNA:GH23305, CG6803, DmMf, Dmel\\CG6803, Zeelin1, z1', description='Myofilin', other_designations='CG6803-PA|CG6803-PB|CG6803-PC|CG6803-PD|CG6803-PE|CG6803-PF|CG6803-PG|CG6803-PH|CG6803-PI|CG6803-PJ|CG6803-PK|CG6803-PL|CG6803-PM|CG6803-PN|CG6803-PO|CG6803-PP|CG6803-PQ|Mf-PA|Mf-PB|Mf-PC|Mf-PD|Mf-PE|Mf-PF|Mf-PG|Mf-PH|Mf-PI|Mf-PJ|Mf-PK|Mf-PL|Mf-PM|Mf-PN|Mf-PO|Mf-PP|Mf-PQ|myofilin|zeelin|zeelin1', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15170625, end_position_on_the_genomic_accession=15181510, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  41825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41825, CurrentID=0, Status='live', Symbol='Gyc88E', Aliases='Dmel_CG4154, CG4154, Dmel\\CG4154, GYC, Gyc-88E', description='Guanylyl cyclase at 88E', other_designations='CG4154-PD|CG4154-PE|CG4154-PF|Gyc88E-PD|Gyc88E-PE|Gyc88E-PF|guanylyl cyclase at 88E', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15150375, end_position_on_the_genomic_accession=15191242, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  41826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41826, CurrentID=0, Status='live', Symbol='CG6752', Aliases='Dmel_ Dmel\\CG6752', description='CG6752 gene product from transcript CG6752-RA', other_designations='CG6752-PA|CG6752-PB', map_location='88E2-88E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15191058, end_position_on_the_genomic_accession=15196210, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41829, CurrentID=0, Status='live', Symbol='Spn88Eb', Aliases='Dmel_CG6687, CG6687, Dmel\\CG6687', description='Serpin 88Eb', other_designations='CG6687-PA|Dm-serpin-88Eb|Spn88Eb-PA|serpin 88Eb', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15206703, end_position_on_the_genomic_accession=15208762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41830, CurrentID=0, Status='live', Symbol='CG4203', Aliases='Dmel_ Dmel\\CG4203, FBgn0038300, MAU-2, Mau-2', description='CG4203 gene product from transcript CG4203-RA', other_designations='CG4203-PA', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15209204, end_position_on_the_genomic_accession=15211347, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41831, CurrentID=0, Status='live', Symbol='CG6654', Aliases='Dmel_ Dmel\\CG6654', description='CG6654 gene product from transcript CG6654-RA', other_designations='CG6654-PA', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15211277, end_position_on_the_genomic_accession=15213658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41832, CurrentID=0, Status='live', Symbol='CG4210', Aliases='Dmel_ BcDNA:SD13446, BcDNA:SD26038, Dmel\\CG4210, FBgn 38302', description='CG4210 gene product from transcript CG4210-RA', other_designations='CG4210-PA', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15213974, end_position_on_the_genomic_accession=15214845, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41833, CurrentID=0, Status='live', Symbol='SIDL', Aliases='Dmel_CG6623, CG6623, Dmel\\CG6623', description='Shal Interactor of Di-Leucine Motif', other_designations='CG6623-PA|SIDL-PA|shal interactor of Di-Leucine motif', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15214762, end_position_on_the_genomic_accession=15218874, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41834, CurrentID=0, Status='live', Symbol='CG12241', Aliases='Dmel_ Dmel\\CG12241', description='CG12241 gene product from transcript CG12241-RA', other_designations='CG12241-PA|uncharacterized protein', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15219256, end_position_on_the_genomic_accession=15222371, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41835, CurrentID=0, Status='live', Symbol='CG31344', Aliases='Dmel_ CG6572, Dmel\\CG31344, anon-WO0118547.636', description='CG31344 gene product from transcript CG31344-RA', other_designations='CG31344-PA|uncharacterized protein', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15223378, end_position_on_the_genomic_accession=15225329, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41836, CurrentID=0, Status='live', Symbol='Caf1-55', Aliases='Dmel_CG4236, 154659_at, 55, CAF-1, CAF1, CAF1-55, CAF1-p55, CAF1p55, CG4236, Caf-1, Caf1, Caf1/p55, Caf1p55, Dmel\\CG4236, MSI1/RbAp48/CAC3/LIN-53, NURF, NURF 55, NURF-55, NURF55, Nurf, Nurf 55, Nurf-55, Nurf55, P55, RBAP48, RbAp46/48, RbAp48, S(ls)3, caf-1, caf1, caf1 p55, d-CAF1, dCAF-1, dCAF-1 p55, dCAF-1-p55, dCAF1, dCAF1-p55, dNURF, p55, p55 CAF1, p55/NURF-55, p55CAF1', description='Chromatin assembly factor 1, p55 subunit', other_designations='CG4236-PA|CG4236-PB|Caf1-55-PA|Caf1-55-PB|Nurf55/CAF1|chromatin assembly Factor-1|chromatin assembly factor 1|chromatin assembly factor 1 subunit|chromatin assembly factor 1, p55 subunit|chromatin associated factor-1 subunit|nucleosome remodeling factor|nucleosome remodeling factor - 55kD|p55', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15225576, end_position_on_the_genomic_accession=15227564, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41837, CurrentID=0, Status='live', Symbol='Art3', Aliases='Dmel_CG6563, CG6563, DART3, Dmel\\CG6563', description='Arginine methyltransferase 3', other_designations='Art3-PA|Art3-PB|CG6563-PA|CG6563-PB|arginine methyltransferase 3', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15227605, end_position_on_the_genomic_accession=15229408, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41838, CurrentID=0, Status='live', Symbol='mRpS10', Aliases='Dmel_CG4247, BcDNA:GM25447, CG4247, Dmel\\CG4247, MRP-S10, MrpS10', description='mitochondrial ribosomal protein S10', other_designations='CG4247-PA|CG4247-PB|mRpS10-PA|mRpS10-PB', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15229665, end_position_on_the_genomic_accession=15230516, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41839, CurrentID=0, Status='live', Symbol='tefu', Aliases='Dmel_CG6535, ATM, Atm, CG6535, Dmel\\CG6535, Tefu, atm, atm/tefu, dATM, tef', description='telomere fusion', other_designations='A-T mutated|CG6535-PB|ataxia telangiectasia mutated|ataxia telengiesctasia mutated|ataxia-telangiectasia mutated|drosophila ATM|tefu-PB', map_location='88E3-88E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15232212, end_position_on_the_genomic_accession=15242236, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  41840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41840, CurrentID=0, Status='live', Symbol='Hsc70-4', Aliases='Dmel_CG4264, BAP74, Bap74, CG4264, Dmel\\CG4264, E(csp)1545, E(nd)195, HSC-70, HSC4, HSC70, HSC70-4, Hsc-4, Hsc-70-4, Hsc4, Hsc4p, Hsc70, Hsp-c4, Hsp70, Q3KN45_DROME, anon-WO0118547.237, bs17d06.y1, chr3R:11070155..11070315, dhsc70, hsc4, hsc70, hsc70-4, hsp70, i190, l(3)03550, l(3)L3929, l(3)j7A4, mtHsp 70, scd', description='Heat shock protein cognate 4', other_designations='Bap74|CG4264-PA|CG4264-PB|CG4264-PC|CG4264-PD|CG4264-PE|CG4264-PF|CG4264-PG|Hsc70-4-PA|Hsc70-4-PB|Hsc70-4-PC|Hsc70-4-PD|Hsc70-4-PE|Hsc70-4-PF|Hsc70-4-PG|heat shock 70-kDa protein cognate 4|heat shock cognate 4|heat shock cognate 70|heat shock cognate protein 4|heat shock cognate protein 70-4|heat shock protein 70|heat shock protein cognate 4|heat-shock protein 70|hsp cognate 4|scattered', map_location='3-56.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15242687, end_position_on_the_genomic_accession=15246614, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41841, CurrentID=0, Status='live', Symbol='Amt', Aliases='Dmel_CG6499, CG6499, Dmel\\CG6499', description='Ammonium transporter', other_designations='Amt-PB|CG6499-PB|ammonium transporter', map_location='88E4-88E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15247163, end_position_on_the_genomic_accession=15249137, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41842, CurrentID=0, Status='live', Symbol='CG42404', Aliases='Dmel_ BcDNA:SD12744, CG31154, CG34128, CG4285, Dmel\\CG42404, Dmel_CG34128, Dmel_CG4285, lincRNA.S6902', description='CG42404 gene product from transcript CG42404-RF', other_designations='CG42404-PD|CG42404-PE|CG42404-PF|uncharacterized protein', map_location='88E5-88E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15250094, end_position_on_the_genomic_accession=15257765, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41843, CurrentID=0, Status='live', Symbol='eIF-2gamma', Aliases='Dmel_CG43665, 51kDa protein, CG43665, CG6476, Dmel\\CG43665, eIF-2G, eIF2G, eIF2gamma', description='Eukaryotic initiation factor 2gamma', other_designations='CG43665-PB|CG43665-PC|CG43665-PD|eIF-2gamma-PB|eIF-2gamma-PC|eIF-2gamma-PD|eukaryotic initiation factor 2gamma|gamma subunit of the translation initiation factor eIF2', map_location='88E6-88E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15257844, end_position_on_the_genomic_accession=15263458, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41844, CurrentID=0, Status='live', Symbol='Set', Aliases='Dmel_CG4299, CG4299, Dmel\\CG4299, SET, TAF-Ibeta, dSET, d', description='CG4299 gene product from transcript CG4299-RA', other_designations='CG4299-PA|Set-PA', map_location='88E8-88E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15263552, end_position_on_the_genomic_accession=15265253, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41845, CurrentID=0, Status='live', Symbol='ATPsynO', Aliases='Dmel_CG4307, ATPO, ATPO_DROME, CG4307, Dmel\\CG4307, OSCP, Oscp', description='ATP synthase, oligomycin sensitivity conferring protein', other_designations='ATP synthase OSCP|ATP synthase oligomycin sensitivity conferral protein|ATPsynO-PA|CG4307-PA|F0/F1 ATP synthase OSCP|OLIGOMYCIN SENSITIVITY CONFERRING PROTEIN|oligimycin sensitive-conferring protein|oligomycin sensitivity conferring protein precursor|oligomycin sensitivity-conferring protein', map_location='88E8-88E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15265591, end_position_on_the_genomic_accession=15266690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41846, CurrentID=0, Status='live', Symbol='CG14864', Aliases='Dmel_ BcDNA:AT13896, Dmel\\CG14864', description='CG14864 gene product from transcript CG14864-RB', other_designations='CG14864-PA|CG14864-PB|uncharacterized protein', map_location='88E9-88E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15266684, end_position_on_the_genomic_accession=15268538, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41847, CurrentID=0, Status='live', Symbol='Zip88E', Aliases='Dmel_CG4334, CG4334, Dmel\\CG4334, cg4334, dZip88E', description='Zinc/iron regulated transporter-related protein 88E', other_designations='CG4334-PB|Zip88E-PB', map_location='88E9-88E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15268662, end_position_on_the_genomic_accession=15270532, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41848, CurrentID=0, Status='live', Symbol='Cp190', Aliases='Dmel_CG6384, AbS1-4, Bx-63, Bx63, CEN190, CG6384, CP-190, CP-60, CP190, CP190_DROME, Cen185, Cen190, DMAP 190, DMAP-190, DMAP190, Dmap190, Dmel\\CG6384, E(mod)4-1, Map190, Rb188, cp190, dmap190', description='Centrosomal protein 190kD', other_designations='CG6384-PA|CG6384-PB|Cp190-PA|Cp190-PB|E(mod)4-1|centrosomal protein 190|centrosomal protein 190 kD|centrosomal protein 190 kDa|centrosomal protein 190kD|centrosomal protein 90|centrosomal protein of 190 kDa|enhancer of mod(mdg4)4-1', map_location='88E9-88E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15270432, end_position_on_the_genomic_accession=15275241, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41849, CurrentID=0, Status='live', Symbol='CG4338', Aliases='Dmel_ Dmel\\CG4338', description='CG4338 gene product from transcript CG4338-RA', other_designations='CG4338-PA', map_location='88E11-88E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15275522, end_position_on_the_genomic_accession=15276474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41850, CurrentID=0, Status='live', Symbol='MRG15', Aliases='Dmel_CG6363, CG6363, DmMRG15, Dmel\\CG6363, Dmrg15, Mrg15, dMRG15, dMrg15, l(3)j6A3', description='MORF-related gene 15', other_designations='CG6363-PA|CG6363-PB|MRG15-PA|MRG15-PB', map_location='88E11-88E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15276440, end_position_on_the_genomic_accession=15278109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41851, CurrentID=0, Status='live', Symbol='l(3)neo43', Aliases='Dmel_CG14865, CG14865, Dmel\\CG14865', description='lethal (3) neo43', other_designations='CG14865-PB|CG14865-PC|l(3)neo43-PB|l(3)neo43-PC', map_location='88E11-88E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15277981, end_position_on_the_genomic_accession=15278668, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41852, CurrentID=0, Status='live', Symbol='Tm1', Aliases='Dmel_CG4898, 1305/10, 2299, BcDNA:GH09289, BcDNA:LD37158, BcDNA:SD21996, CG4898, Dm Tm1, Dm TmH33, Dm TmH34, DmTm1, Dmel\\CG4898, PmI, TM, TM1, TMII, Tm, TmH, TmH-33, TmH-34, TmH33, TmH34, TmII, Tmr33, Tmr34, TnH, TnH-33, TnH-34, cTM, cTm, cTmII, chr3R:11122272..11122408, l(3)02299, l(3)S130510, l(3)s2958, mTmII, region 3, tm1, tmII, tropomyosin', description='Tropomyosin 1', other_designations='CG4898-PA|CG4898-PC|CG4898-PE|CG4898-PF|CG4898-PG|CG4898-PH|CG4898-PI|CG4898-PJ|CG4898-PK|CG4898-PL|CG4898-PM|CG4898-PN|CG4898-PO|CG4898-PP|CG4898-PQ|CG4898-PR|CG4898-PS|Tm1-PA|Tm1-PC|Tm1-PE|Tm1-PF|Tm1-PG|Tm1-PH|Tm1-PI|Tm1-PJ|Tm1-PK|Tm1-PL|Tm1-PM|Tm1-PN|Tm1-PO|Tm1-PP|Tm1-PQ|Tm1-PR|Tm1-PS|Tropomyosin-1|cytoplasmic tropomyosin|cytoskeletal tropomyosin|tropomyosin|tropomyosin 1|tropomyosin I|tropomyosin II|tropomyosinII|tropomysin II|troponin H|troponin-H|troponin-H isoform 34', map_location='88E12-88E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15281371, end_position_on_the_genomic_accession=15308779, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  41853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41853, CurrentID=0, Status='live', Symbol='Tm2', Aliases='Dmel_CG4843, CG4843, DROTROPI1, Dmel\\CG4843, IP16005p, Ifm(3)3, Ifm(3)5, Ifm-TmI, TM2, Tm, Tm1, Tm127, TmI, TmII, Tn-H, dro Tm, l(3)nc99Eb, mTmI', description='Tropomyosin 2', other_designations='CG4843-PA|CG4843-PB|CG4843-PC|CG4843-PE|CG4843-PF|CG4843-PG|Tm2-PA|Tm2-PB|Tm2-PC|Tm2-PE|Tm2-PF|Tm2-PG|Tropomyosin-2|lethal(3)nc99Eb|tropomyosin|tropomyosin 127|tropomyosin 2|tropomyosin I|tropomyosin2', map_location='3-54.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15310193, end_position_on_the_genomic_accession=15315197, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41854, CurrentID=0, Status='live', Symbol='CG14866', Aliases='Dmel_ Dmel\\CG14866', description='CG14866 gene product from transcript CG14866-RA', other_designations='CG14866-PA|CG14866-PB|uncharacterized protein', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15316471, end_position_on_the_genomic_accession=15318285, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41855, CurrentID=0, Status='live', Symbol='CG6276', Aliases='Dmel_ Dmel\\CG6276', description='CG6276 gene product from transcript CG6276-RA', other_designations='CG6276-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15318203, end_position_on_the_genomic_accession=15320069, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41857, CurrentID=0, Status='live', Symbol='CG6236', Aliases='Dmel_ Dmel\\CG6236, gi21357449, gi23171351, gi7299997', description='CG6236 gene product from transcript CG6236-RB', other_designations='CG6236-PA|CG6236-PB', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15325325, end_position_on_the_genomic_accession=15328269, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41858, CurrentID=0, Status='live', Symbol='ea', Aliases='Dmel_CG4920, CG4920, Dmel\\CG4920, EA, SP24, c-SP24', description='easter', other_designations='CG4920-PA|CG4920-PB|ea-PA|ea-PB', map_location='3-57 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15328729, end_position_on_the_genomic_accession=15330408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41859, CurrentID=0, Status='live', Symbol='mRpL9', Aliases='Dmel_CG31478, CG31478, CG4923, Dmel\\CG31478, L9', description='mitochondrial ribosomal protein L9', other_designations='CG31478-PA|mRpL9-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15330738, end_position_on_the_genomic_accession=15331588, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41860, CurrentID=0, Status='live', Symbol='FK506-bp1', Aliases='Dmel_CG6226, CG6226, Dmel\\CG6226, FK506bp1, FKBP39, FKbp39, IFK506-bp1', description='FK506-binding protein 1', other_designations='CG6226-PA|FK506-binding protein of 39 kDa|FK506-bp1-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15331584, end_position_on_the_genomic_accession=15333133, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41861, CurrentID=0, Status='live', Symbol='Sra-1', Aliases='Dmel_CG4931, CG4931, CYFIP, CYFIP1/2, DSra-1, Dmel\\CG4931, PIR121, PIR121/Sra-1, Sra/CYFIP, Sra1, Sra1/PIR121, cyfip, dSra, sra-1, sra1, Sra-1', description='specifically Rac1-associated protein 1', other_designations='CG4931-PA|Sra-1-PA|specifically, Rac-1 associated protein 1', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15333266, end_position_on_the_genomic_accession=15338365, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41862, CurrentID=0, Status='live', Symbol='Aats-ser', Aliases='Dmel_CG4938, Aat-Ser, CG4938, DmSRS2, Dmel\\CG4938, SRS, SerRS, seryl-tRNA synthetase', description='Seryl-tRNA synthetase', other_designations='Aats-ser-PA|Aats-ser-PB|CG4938-PA|CG4938-PB', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15340667, end_position_on_the_genomic_accession=15343218, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41863, CurrentID=0, Status='live', Symbol='CG6218', Aliases='Dmel_ Dmel\\CG6218', description='CG6218 gene product from transcript CG6218-RA', other_designations='CG6218-PA|CG6218-PB', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15337908, end_position_on_the_genomic_accession=15343256, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41864, CurrentID=0, Status='live', Symbol='Surf4', Aliases='Dmel_CG6202, CG6202, Dmel\\CG6202, SURF-4, Surf-4', description='Surfeit 4', other_designations='CG6202-PA|Surf4-PA|surfeit 4', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15343653, end_position_on_the_genomic_accession=15346176, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41865, CurrentID=0, Status='live', Symbol='CG31301', Aliases='Dmel_ CG12258, Dmel\\CG31301', description='CG31301 gene product from transcript CG31301-RA', other_designations='CG31301-PA|uncharacterized protein', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15346205, end_position_on_the_genomic_accession=15348012, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41866, CurrentID=0, Status='live', Symbol='Trs33', Aliases='Dmel_CG6196, BcDNA:RH37427, CG6196, Dmel\\CG6196p, Trs33', description='TRAPP subunit 33 ortholog (S. cerevisiae)', other_designations='CG6196-PA|TRAPP subunit 33 ortholog|Trs33-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15350345, end_position_on_the_genomic_accession=15350995, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41867, CurrentID=0, Status='live', Symbol='CG5038', Aliases='Dmel_ Dmel\\CG5038', description='CG5038 gene product from transcript CG5038-RA', other_designations='CG5038-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15351225, end_position_on_the_genomic_accession=15353926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41868, CurrentID=0, Status='live', Symbol='Atg4b', Aliases='Dmel_CG6194, APG4, AUT2, Atg4, CG6194, DmAtg4-like, Dmel\\CG6194', description='Autophagy-related 4b', other_designations='Atg4b-PA|Autophagy-specific gene 4b|CG6194-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15353720, end_position_on_the_genomic_accession=15356924, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41869, CurrentID=0, Status='live', Symbol='CG5044', Aliases='Dmel_ CT16187, Dmel\\CG5044, HIBCH', description='CG5044 gene product from transcript CG5044-RB', other_designations='CG5044-PA|CG5044-PB', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15357229, end_position_on_the_genomic_accession=15359263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41870, CurrentID=0, Status='live', Symbol='ldlCp', Aliases='Dmel_CG6177, CG6177, Dmel\\CG6177, ldl-Cp', description='ldlCp-related protein', other_designations='CG6177-PA|ldlCp-PA', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15358481, end_position_on_the_genomic_accession=15361657, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41871, CurrentID=0, Status='live', Symbol='Trax', Aliases='Dmel_CG5063, CG5063, Dmel\\CG5063, trax', description='Translin associated factor X', other_designations='CG5063-PB|CG5063-PC|Translin-associated factor X|Trax-PB|Trax-PC|translin associated factor X', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15361692, end_position_on_the_genomic_accession=15362850, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41872, CurrentID=0, Status='live', Symbol='CG6171', Aliases='Dmel_ Dmel\\CG6171, anon-Becker2', description='CG6171 gene product from transcript CG6171-RB', other_designations='Anon-becker2|CG6171-PA|CG6171-PB', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15362796, end_position_on_the_genomic_accession=15363857, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41873, CurrentID=0, Status='live', Symbol='CG34404', Aliases='Dmel_ CG18496, CG6156, Dmel\\CG34404, Dmel_CG18496, Dmel_CG6156, anon-FF', description='CG34404 gene product from transcript CG34404-RD', other_designations='CG34404-PC|CG34404-PD|CG34404-PE|uncharacterized protein', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15363786, end_position_on_the_genomic_accession=15367665, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41875, CurrentID=0, Status='live', Symbol='CG14868', Aliases='Dmel_ Dmel\\CG14868', description='CG14868 gene product from transcript CG14868-RA', other_designations='CG14868-PA|uncharacterized protein', map_location='88F1-88F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15367816, end_position_on_the_genomic_accession=15369945, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41876, CurrentID=0, Status='live', Symbol='Ccm3', Aliases='Dmel_CG5073, CG5073, Dmel\\CG5073', description='Cerebral cavernous malformation 3 ortholog', other_designations='CG5073-PA|Ccm3-PA|cerebral cavernous malformation 3|cerebral cavernous malformation 3 ortholog', map_location='88F2-88F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15371957, end_position_on_the_genomic_accession=15374237, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41877, CurrentID=0, Status='live', Symbol='CG6136', Aliases='Dmel_ Dmel\\CG6136', description='CG6136 gene product from transcript CG6136-RD', other_designations='CG6136-PB|CG6136-PC|CG6136-PD', map_location='88F2-88F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15370036, end_position_on_the_genomic_accession=15372212, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41879, CurrentID=0, Status='live', Symbol='h-cup', Aliases='Dmel_CG6130, CG6130, Dmel\\CG6130, anon-WO0140519.172', description='heineken-cup', other_designations='CG6130-PA|h-cup-PA', map_location='88F3-88F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15381407, end_position_on_the_genomic_accession=15382951, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41880, CurrentID=0, Status='live', Symbol='Rbp', Aliases='Dmel_CG43073, CG14867, CG31302, CG43073, CG5143, CG5152, DRBP, Dmel\\CG43073, Dmel_CG14867, Dmel_CG31302, drbp', description='RIM-binding protein', other_designations='CG43073-PF|CG43073-PG|CG43073-PI|CG43073-PJ|CG43073-PK|CG43073-PL|CG43073-PM|CG43073-PN|CG43073-PO|Rab3-interacting molecule (RIM)-binding protein|Rbp-PF|Rbp-PG|Rbp-PI|Rbp-PJ|Rbp-PK|Rbp-PL|Rbp-PM|Rbp-PN|Rbp-PO', map_location='88F3-88F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15374635, end_position_on_the_genomic_accession=15402543, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  8674116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674116, CurrentID=0, Status='live', Symbol='CG42460', Aliases='Dmel_ Dmel\\CG42460', description='CG42460 gene product from transcript CG42460-RB', other_designations='CG42460-PA|CG42460-PB|uncharacterized protein', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3387882, end_position_on_the_genomic_accession=3388496, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41882, CurrentID=0, Status='live', Symbol='CG6125', Aliases='Dmel_ Dmel\\CG6125', description='CG6125 gene product from transcript CG6125-RC', other_designations='CG6125-PA|CG6125-PB|CG6125-PC', map_location='88F4-88F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15402598, end_position_on_the_genomic_accession=15407070, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41883, CurrentID=0, Status='live', Symbol='Atx2', Aliases='Dmel_CG5166, ATX2, Atx-2, Atxain-2, BEST:GH27029, CG5166, CG5166a, Datx2, Dmel\\CG5166, GH27029, SCA2, Sca2, anon-EST:fe2E1, anon-WO0118547.351, anon-WO0140519.1, atx-2, atx2, dAtx2, dSCA2, l(3)06490', description='Ataxin-2', other_designations='ATAXIN-2|Atx2-PA|Atx2-PB|Atx2-PC|Atx2-PD|CG5166-PA|CG5166-PB|CG5166-PC|CG5166-PD|anon-fast-evolving-2E1|ataxin 2|dAtaxin-2', map_location='88F4-88F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15408423, end_position_on_the_genomic_accession=15417277, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41884, CurrentID=0, Status='live', Symbol='CG6118', Aliases='Dmel_ Dmel\\CG6118', description='CG6118 gene product from transcript CG6118-RD', other_designations='CG6118-PB|CG6118-PC|CG6118-PD', map_location='88F4-88F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15421108, end_position_on_the_genomic_accession=15434738, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41885, CurrentID=0, Status='live', Symbol='Act88F', Aliases='Dmel_CG5178, Act(88F), Act88-F, Act88f, Actin, CG5178, Dmel\\CG5178, E, Ifm(3)7, act 88F, act88F, actin, rsd', description='Actin 88F', other_designations='88F actin|Act88F-PA|CG5178-PA|F-actin|actin|actin 88F|actin88F|arthrin', map_location='3-57.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15439969, end_position_on_the_genomic_accession=15442177, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41887, CurrentID=0, Status='live', Symbol='AdamTS-A', Aliases='Dmel_CG14869, CG14869, CG6107, CT19033, Dmel\\CG14869', description='ADAM metallopeptidase with thrombospondin type 1 motif A', other_designations='AdamTS-A-PB|AdamTS-A-PC|AdamTS-A-PE|CG14869-PB|CG14869-PC|CG14869-PE', map_location='88F5-88F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15442246, end_position_on_the_genomic_accession=15476193, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  41888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41888, CurrentID=0, Status='live', Symbol='B9d1', Aliases='Dmel_CG14870, B9D1, CG14870, Dmel\\CG14870', description='CG14870 gene product from transcript CG14870-RA', other_designations='B9d1-PA|CG14870-PA', map_location='88F6-88F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15481906, end_position_on_the_genomic_accession=15482904, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41889, CurrentID=0, Status='live', Symbol='Rh6', Aliases='Dmel_CG5192, CG5192, DMELRH6, Dm Rh6, Dmel\\CG5192, R8, RH6, Rh, rh6', description='Rhodopsin 6', other_designations='CG5192-PC|R8 rhodopsin|Rh6-PC|Rhodopsin6|rhodopsin|rhodopsin 6|rhodopsin-6', map_location='88F6-88F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15483158, end_position_on_the_genomic_accession=15484699, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41890, CurrentID=0, Status='live', Symbol='Trissin', Aliases='Dmel_CG14871, CG14871, Dmel\\CG14871', description='CG14871 gene product from transcript CG14871-RB', other_designations='CG14871-PB|Trissin-PB|trissin', map_location='88F6-88F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15484730, end_position_on_the_genomic_accession=15485396, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41891, CurrentID=0, Status='live', Symbol='CG5205', Aliases='Dmel_ Dmel\\CG5205, cg5205', description='CG5205 gene product from transcript CG5205-RA', other_designations='CG5205-PA', map_location='88F6-88F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15486050, end_position_on_the_genomic_accession=15493310, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41892, CurrentID=0, Status='live', Symbol='CG5213', Aliases='Dmel_ Dmel\\CG5213, cg5213', description='CG5213 gene product from transcript CG5213-RA', other_designations='CG5213-PA', map_location='88F6-88F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15494602, end_position_on_the_genomic_accession=15495620, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41893, CurrentID=0, Status='live', Symbol='CG44014', Aliases='Dmel_ CG14872, Dmel\\CG44014', description='CG44014 gene product from transcript CG44014-RA', other_designations='CG44014-PA|uncharacterized protein', map_location='88F7-88F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15507940, end_position_on_the_genomic_accession=15508886, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41894, CurrentID=0, Status='live', Symbol='AOX1', Aliases='Dmel_CG18522, CG18522, CT42272, DmAO1, Dmel\\CG18522, LPO, PO, Po, lpo', description='Aldehyde oxidase 1', other_designations='AOX1-PA|CG18522-PA|aldehyde oxidase 1|low-pyridoxal-oxidase|pyridoxal oxidase', map_location='3-57.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15532137, end_position_on_the_genomic_accession=15537292, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41895, CurrentID=0, Status='live', Symbol='AOX2', Aliases='Dmel_CG18519, CG18519, DmAO2, Dmel\\CG18519', description='Aldehyde oxidase 2', other_designations='AOX2-PC|CG18519-PC|aldehyde oxidase 2', map_location='88F7-88F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15537713, end_position_on_the_genomic_accession=15542414, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41896, CurrentID=0, Status='live', Symbol='AOX3', Aliases='Dmel_CG6045, CG6045, DmAO3, Dmel\\CG6045', description='Aldehyde oxidase 3', other_designations='AOX3-PA|AOX3-PB|CG6045-PA|CG6045-PB|aldehyde oxidase 3', map_location='88F7-88F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15542748, end_position_on_the_genomic_accession=15548188, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41897, CurrentID=0, Status='live', Symbol='AOX4', Aliases='Dmel_CG18516, CG18516, DmAO4, Dmel\\CG18516', description='Aldehyde oxidase 4', other_designations='AOX4-PA|CG18516-PA|aldehyde oxidase 4', map_location='88F7-88F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15548628, end_position_on_the_genomic_accession=15552833, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41899, CurrentID=0, Status='live', Symbol='pxb', Aliases='Dmel_CG33207, BcDNA:RE16319, CG14873, CG14874, CG33207, CG33207/Pxb, Dmel\\CG33207, Pxb', description='CG33207 gene product from transcript CG33207-RC', other_designations='CG33207-PB|CG33207-PC|baikal|paxillinB|pxb-PB|pxb-PC', map_location='89A1-89A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15666183, end_position_on_the_genomic_accession=15690513, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41900, CurrentID=0, Status='live', Symbol='CG5399', Aliases='Dmel_ Dmel\\CG5399', description='CG5399 gene product from transcript CG5399-RB', other_designations='CG5399-PA|CG5399-PB', map_location='89A2-89A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15695241, end_position_on_the_genomic_accession=15697142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41901, CurrentID=0, Status='live', Symbol='CG5404', Aliases='Dmel_ Dmel\\CG5404', description='CG5404 gene product from transcript CG5404-RA', other_designations='CG5404-PA', map_location='89A2-89A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15704676, end_position_on_the_genomic_accession=15708434, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41902, CurrentID=0, Status='live', Symbol='CG4520', Aliases='Dmel_ Dmel\\CG4520', description='CG4520 gene product from transcript CG4520-RB', other_designations='CG4520-PA|CG4520-PB', map_location='89A3-89A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15744654, end_position_on_the_genomic_accession=15746411, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41903, CurrentID=0, Status='live', Symbol='CG12784', Aliases='Dmel_ Dmel\\CG12784', description='CG12784 gene product from transcript CG12784-RA', other_designations='CG12784-PA|uncharacterized protein', map_location='89A4-89A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15763115, end_position_on_the_genomic_accession=15763551, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41904, CurrentID=0, Status='live', Symbol='CG5623', Aliases='Dmel_ BcDNA:AT10995, Dmel\\CG5623', description='CG5623 gene product from transcript CG5623-RA', other_designations='CG5623-PA', map_location='89A4-89A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15766246, end_position_on_the_genomic_accession=15767438, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41905, CurrentID=0, Status='live', Symbol='CG4525', Aliases='Dmel_ Dmel\\CG4525, TTC26', description='CG4525 gene product from transcript CG4525-RA', other_designations='CG4525-PA', map_location='89A4-89A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15774416, end_position_on_the_genomic_accession=15777640, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41906, CurrentID=0, Status='live', Symbol='CG5614', Aliases='Dmel_ Dmel\\CG5614', description='CG5614 gene product from transcript CG5614-RA', other_designations='CG5614-PA', map_location='89A4-89A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15779805, end_position_on_the_genomic_accession=15781092, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41907, CurrentID=0, Status='live', Symbol='CG9590', Aliases='Dmel_ Dmel\\CG9590', description='CG9590 gene product from transcript CG9590-RA', other_designations='CG9590-PA', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15786703, end_position_on_the_genomic_accession=15788339, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41908, CurrentID=0, Status='live', Symbol='CG9589', Aliases='Dmel_ BcDNA:AT11468, Dmel\\CG9589', description='CG9589 gene product from transcript CG9589-RA', other_designations='CG9589-PA', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15788521, end_position_on_the_genomic_accession=15789494, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41910, CurrentID=0, Status='live', Symbol='Acyp2', Aliases='Dmel_CG18505, AcPDro2, CG18505, Dmel\\CG18505', description='Acylphosphatase 2', other_designations='Acyp2-PA|Acyp2-PB|CG18505-PA|CG18505-PB|acylphosphatase|acylphosphatase 2', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15792688, end_position_on_the_genomic_accession=15793487, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41911, CurrentID=0, Status='live', Symbol='nsl1', Aliases='Dmel_CG4699, CG4699, Dmel\\CG4699, E(nos), NSL1, Nsl1, l(3)06536, l(3)S009413, wah', description='non-specific lethal 1', other_designations='CG4699-PD|CG4699-PE|CG4699-PF|CG4699-PH|CG4699-PI|CG4699-PJ|CG4699-PK|CG4699-PL|CG4699-PM|CG4699-PN|CG4699-PO|CG4699-PP|lethal (3) 06536|nonspecific lethal 1|nsl1-PD|nsl1-PE|nsl1-PF|nsl1-PH|nsl1-PI|nsl1-PJ|nsl1-PK|nsl1-PL|nsl1-PM|nsl1-PN|nsl1-PO|nsl1-PP|waharan', map_location='89A4-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15793782, end_position_on_the_genomic_accession=15802102, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41912, CurrentID=0, Status='live', Symbol='CG9593', Aliases='Dmel_ AAF55227, Dmel\\CG9593, FBpp0082632', description='CG9593 gene product from transcript CG9593-RB', other_designations='CG9593-PB|Fibrinogen-related protein', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15802275, end_position_on_the_genomic_accession=15803653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41913, CurrentID=0, Status='live', Symbol='alpha-Man-IIb', Aliases="Dmel_CG4606, CG4606, DM-GII.2, Dmel\\CG4606, GM, GTD#24/1, Man'ase, ManIIb, alpha-M-II, alpha-Man-II b", description='alpha-Mannosidase class II b', other_designations='CG4606-PA|CG4606-PB|CG4606-PC|alpha mannosidase IIb|alpha-D-mannosidase|alpha-Man-IIb-PA|alpha-Man-IIb-PB|alpha-Man-IIb-PC|alpha-mannosidase|alpha-mannosidase 2|alpha-mannosidase-IIb|golgi alpha-mannosidase IIb|mannosidase II', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15804248, end_position_on_the_genomic_accession=15824118, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41914, CurrentID=0, Status='live', Symbol='CG4576', Aliases='Dmel_ Dmel\\CG4576', description='CG4576 gene product from transcript CG4576-RA', other_designations='CG4576-PA', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15825747, end_position_on_the_genomic_accession=15830321, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41916, CurrentID=0, Status='live', Symbol='CG43317', Aliases='Dmel_ CG14876, Dmel\\CG43317, Dmel_CG14876', description='CG43317 gene product from transcript CG43317-RA', other_designations='CG43317-PA|uncharacterized protein', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15833462, end_position_on_the_genomic_accession=15834425, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41917, CurrentID=0, Status='live', Symbol='Arpc3A', Aliases='Dmel_CG4560, ARC-P21, ARPC, Arpc3a, CG4560, Dmel\\CG4560', description='Actin-related protein 2/3 complex, subunit 3A', other_designations='Arpc3A-PC|Arpc3A-PE|CG4560-PC|CG4560-PE', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15834648, end_position_on_the_genomic_accession=15836021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41919, CurrentID=0, Status='live', Symbol='spn-E', Aliases='Dmel_CG3158, CG3158, Dmel\\CG3158, SPN-E, Spn-E, SpnE, fs(1)hls, fs(3)hls, hes, hls, homeless, spin-E, spn E/hls, spnE, spn-E', description='spindle E', other_designations='CG3158-PA|Spindle-E (homeless)|SpindleE|female sterile(3)homeless|homeless|homeless/spindle-E|spindle-E|spn-E-PA', map_location='3-62 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15837419, end_position_on_the_genomic_accession=15844415, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41920, CurrentID=0, Status='live', Symbol='Pbp45', Aliases='Dmel_CG9597, CG9597, DmPBP, DmPBP45, DmSNAP43, DmSNAPc, Dmel\\CG9597, SNAP43', description='PSEA-binding protein 45kD', other_designations='CG9597-PA|PSEA-binding protein|Pbp45-PA|drosophila melanogaster PSE-binding protein 45-kDa subunit', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15838146, end_position_on_the_genomic_accession=15839515, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41921, CurrentID=0, Status='live', Symbol='cv-d', Aliases='Dmel_CG31150, CG31150, CG9622, CG9625, CG9627, Cv-d, Dmel\\CG31150', description='crossveinless d', other_designations='CG31150-PA|cv-d-PA', map_location='3-65 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15844466, end_position_on_the_genomic_accession=15853581, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  41922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41922, CurrentID=0, Status='live', Symbol='CG4546', Aliases='Dmel_ Dmel\\CG4546', description='CG4546 gene product from transcript CG4546-RA', other_designations='CG4546-PA', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15846534, end_position_on_the_genomic_accession=15848436, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41925, CurrentID=0, Status='live', Symbol='Hmt-1', Aliases='Dmel_CG4225, CG4225, Dmel\\CG4225, EP(3)3387, hmt-1', description='Heavy metal tolerance factor 1', other_designations='CG4225-PA|Hmt-1-PA|heavy metal tolerance factor 1', map_location='89A6-89A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15853868, end_position_on_the_genomic_accession=15857470, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41926, CurrentID=0, Status='live', Symbol='CG9632', Aliases='Dmel_ BcDNA:AT06908, Dmel\\CG9632, NEST:bs19g07, bs19g07.y1', description='CG9632 gene product from transcript CG9632-RA', other_designations='CG9632-PA|CG9632-PB', map_location='89A6-89A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15860196, end_position_on_the_genomic_accession=15861478, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41927, CurrentID=0, Status='live', Symbol='CG31183', Aliases='Dmel_ CG11846, CG4224, Dmel\\CG31183, FBgn0038379, GYC', description='CG31183 gene product from transcript CG31183-RB', other_designations='CG31183-PA|CG31183-PB|uncharacterized protein', map_location='89A6-89A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15857871, end_position_on_the_genomic_accession=15875939, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  41929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41929, CurrentID=0, Status='live', Symbol='CG14877', Aliases='Dmel_ Dmel\\CG14877', description='CG14877 gene product from transcript CG14877-RA', other_designations='CG14877-PA|uncharacterized protein', map_location='89A7-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15889194, end_position_on_the_genomic_accession=15896078, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41930, CurrentID=0, Status='live', Symbol='glob1', Aliases='Dmel_CG9734, CG9734, DmHb, DmHb1, DmeGb1, DmeGlob1, Dmel\\CG9734, Glob1, Hb, dmeglob1, globin1', description='globin 1', other_designations='CG9734-PA|CG9734-PB|CG9734-PC|CG9734-PD|CG9734-PE|CG9734-PF|glob1-PA|glob1-PB|glob1-PC|glob1-PD|glob1-PE|glob1-PF|globin|hemoglobin|hemoglobin gene 1', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15896926, end_position_on_the_genomic_accession=15904521, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  41931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41931, CurrentID=0, Status='live', Symbol='CG3303', Aliases='Dmel_ Dmel\\CG3303', description='CG3303 gene product from transcript CG3303-RA', other_designations='CG3303-PA', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15908398, end_position_on_the_genomic_accession=15910367, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41932, CurrentID=0, Status='live', Symbol='CG45105', Aliases='Dmel_ CG14878, CG31291, CG34275, CG5552, Dmel\\CG45105, Dmel_CG31291, Dmel_CG34275', description='CG45105 gene product from transcript CG45105-RC', other_designations='CG45105-PB|CG45105-PC|CG45105-PD|CG45105-PE|CG45105-PF|CG45105-PG|CG45105-PH|CG45105-PI|uncharacterized protein', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15910372, end_position_on_the_genomic_accession=15918007, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  41934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41934, CurrentID=0, Status='live', Symbol='CG34276', Aliases='Dmel_ CG5470, Dmel\\CG34276', description='CG34276 gene product from transcript CG34276-RB', other_designations='CG34276-PA|CG34276-PB|uncharacterized protein', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15925629, end_position_on_the_genomic_accession=15927577, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41935, CurrentID=0, Status='live', Symbol='Fbxl7', Aliases='Dmel_CG4221, CG4221, Dmel\\CG4221, FBXL7, FbxL7', description='F-box and leucine-rich repeat protein 7', other_designations='CG4221-PA|Fbxl7-PA', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15920898, end_position_on_the_genomic_accession=15932887, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41936, CurrentID=0, Status='live', Symbol='CG5478', Aliases='Dmel_ Dmel\\CG5478', description='CG5478 gene product from transcript CG5478-RA', other_designations='CG5478-PA', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15932700, end_position_on_the_genomic_accession=15933734, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41937, CurrentID=0, Status='live', Symbol='blp', Aliases='Dmel_CG5268, BcDNA:RE11908, Blp, CG5268, Dmel\\CG5268, l(3)01618', description='black pearl', other_designations='CG5268-PA|CG5268-PB|black-pearl|blp-PA|blp-PB', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15933847, end_position_on_the_genomic_accession=15934940, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41938, CurrentID=0, Status='live', Symbol='Sap47', Aliases='Dmel_CG8884, CG8884, Dmel\\CG8884, SAP 47, SAP-47, SAP47, dsap-47, sap47', description='Synapse-associated protein 47kD', other_designations='CG8884-PA|CG8884-PB|CG8884-PC|CG8884-PD|CG8884-PE|CG8884-PF|CG8884-PG|CG8884-PI|CG8884-PJ|CG8884-PK|Sap47-PA|Sap47-PB|Sap47-PC|Sap47-PD|Sap47-PE|Sap47-PF|Sap47-PG|Sap47-PI|Sap47-PJ|Sap47-PK', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15935100, end_position_on_the_genomic_accession=15962124, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41939, CurrentID=0, Status='live', Symbol='CG4287', Aliases='Dmel_ Dmel\\CG4287', description='CG4287 gene product from transcript CG4287-RB', other_designations='CG4287-PA|CG4287-PB', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15962167, end_position_on_the_genomic_accession=15963867, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41940, CurrentID=0, Status='live', Symbol='CG5516', Aliases='Dmel_ Dmel\\CG5516', description='CG5516 gene product from transcript CG5516-RA', other_designations='CG5516-PA|CG5516-PB', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15963884, end_position_on_the_genomic_accession=15964761, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41941, CurrentID=0, Status='live', Symbol='Rbf2', Aliases='Dmel_CG5083, CG5083, Dmel\\CG5083, RB, RBF2, rbf2', description='Retinoblastoma-family protein 2', other_designations='CG5083-PA|Rbf2-PA', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15965446, end_position_on_the_genomic_accession=15968138, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  41942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41942, CurrentID=0, Status='live', Symbol='mor', Aliases='Dmel_CG18740, BAP155, CG18740, CG4275, Dmel\\CG18740, E(E2F)3B, MOIRA, MOR, MOR/BAP155, Moi, Moira, Mor, Swi3D, dmMOIRA, l(3)89B1', description='moira', other_designations='Bap155/Moira|CG18740-PA|CG18740-PB|brahma associated protein 155 kDa|brahma-associated protein 155|brahma-like|mor-PA|mor-PB', map_location='3-58.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15968134, end_position_on_the_genomic_accession=15972913, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41943, CurrentID=0, Status='live', Symbol='Hel89B', Aliases='Dmel_CG4261, 89B, BTAF1, CG4261, Dmel\\CG4261, Helicase89B, Mot1', description='Helicase 89B', other_designations='89B helicase|CG4261-PB|CG4261-PC|Hel89B-PB|Hel89B-PC|Helicase89B|helicase 89B', map_location='89A11-89A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15973298, end_position_on_the_genomic_accession=15983451, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41944, CurrentID=0, Status='live', Symbol='srp', Aliases='Dmel_CG3992, A7.1, ABF, Abf, CG3992, DmGATAb, Dmel\\CG3992, GATA, GATAb, SRP, Srp, abf, dGATAb, l(3)01549, l(3)89B2, l(3)neo45, sptD, srp', description='serpent', other_designations='A-box-binding factor|CG3992-PA|CG3992-PB|CG3992-PD|CG3992-PE|CG3992-PF|CG3992-PG|serpentD|srp-PA|srp-PB|srp-PD|srp-PE|srp-PF|srp-PG', map_location='3-58 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15986152, end_position_on_the_genomic_accession=16004081, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  41945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41945, CurrentID=0, Status='live', Symbol='GATAe', Aliases='Dmel_CG10278, CG10278, CG10278-PA, Dmel\\CG10278, dGATA-E, d', description='CG10278 gene product from transcript CG10278-RB', other_designations='CG10278-PA|CG10278-PB|CG10278-PC|GATAe-PA|GATAe-PB|GATAe-PC', map_location='89A12-89A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16009105, end_position_on_the_genomic_accession=16019133, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41948, CurrentID=0, Status='live', Symbol='CG10264', Aliases='Dmel_ Dmel\\CG10264', description='CG10264 gene product from transcript CG10264-RA', other_designations='CG10264-PA|uncharacterized protein', map_location='89B1-89B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16043498, end_position_on_the_genomic_accession=16044832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41949, CurrentID=0, Status='live', Symbol='CG10407', Aliases='Dmel_ Dmel\\CG10407, cg10407', description='CG10407 gene product from transcript CG10407-RB', other_designations='CG10407-PA|CG10407-PB|CG10407-PC|uncharacterized protein', map_location='89B1-89B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16044747, end_position_on_the_genomic_accession=16046130, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41951, CurrentID=0, Status='live', Symbol='CG5013', Aliases='Dmel_ Dmel\\CG5013', description='CG5013 gene product from transcript CG5013-RA', other_designations='CG5013-PA', map_location='89B1-89B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16049670, end_position_on_the_genomic_accession=16050693, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41952, CurrentID=0, Status='live', Symbol='msps', Aliases='Dmel_CG5000, CG5000, Dis1, Dmel\\CG5000, MSPS, Msps, TOG, XMAP215, anon-WO0140519.136', description='mini spindles', other_designations='CG5000-PB|CG5000-PC|CG5000-PD|CG5000-PE|CG5000-PF|mini-spindles|minispindles|msps-PB|msps-PC|msps-PD|msps-PE|msps-PF', map_location='89B1-89B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16050801, end_position_on_the_genomic_accession=16060675, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  41953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41953, CurrentID=0, Status='live', Symbol='CG10185', Aliases='Dmel_ Dmel\\CG10185', description='CG10185 gene product from transcript CG10185-RA', other_designations='CG10185-PA|uncharacterized protein', map_location='89B2-89B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16060589, end_position_on_the_genomic_accession=16067594, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  41954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41954, CurrentID=0, Status='live', Symbol='sxe2', Aliases='Dmel_CG4979, CG4979, Dmel\\CG4979', description='sex-specific enzyme 2', other_designations='CG4979-PA|CG4979-PB|male sex-specific enzyme 2|sxe2-PA|sxe2-PB', map_location='89B2-89B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16083878, end_position_on_the_genomic_accession=16087222, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41955, CurrentID=0, Status='live', Symbol='Mhcl', Aliases='Dmel_CG31045, CG10218, CG10224, CG31045, DmPDZ, Dmel\\CG31045, MHCL(PDZ)', description='Myosin heavy chain-like', other_designations='CG31045-PA|CG31045-PB|CG31045-PC|CG31045-PD|CG31045-PF|CG31045-PG|CG31045-PH|CG31045-PI|Mhcl-PA|Mhcl-PB|Mhcl-PC|Mhcl-PD|Mhcl-PF|Mhcl-PG|Mhcl-PH|Mhcl-PI|Myosin-heavy-chain-like|PDZ-myosin|myosin 18|myosin XVIII|myosin XVIII (PDZ)|myosin heavy chain|myosin heavy chain-like', map_location='89B2-89B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16068523, end_position_on_the_genomic_accession=16098761, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  41957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41957, CurrentID=0, Status='live', Symbol='Akt1', Aliases='Dmel_CG4006, AKT, AKT/PKB, AKT1, Akt, Akt/PKB|PKB, CG4006, D-Akt, DAKT1, DAKT1/PKB, DAkt, DAkt1, DPKB, DRAC-PK, DRAC-PK66, DRAC-PK85, Dakt, Dakt1, Dmel\\CG4006, Dpkb, PKB, PKB/AKT, PKB/Akt, PKB/dAKT, RAC, RacPK, akt, akt1, dAKT, dAKT/dPKB, dAKT1, dAkt, dAkt/PKB, dAkt1, dPKB, dakt, dakt1, l(3)04226, l(3)89Bq, p-Akt, pAkt', description='CG4006 gene product from transcript CG4006-RE', other_designations='Akt kinase|Akt1-PA|Akt1-PB|Akt1-PC|Akt1-PD|Akt1-PE|CG4006-PA|CG4006-PB|CG4006-PC|CG4006-PD|CG4006-PE|dAkt kinase|protein kinase B|related to PKA to PKC protein kinases|related to the A and C kinases', map_location='89B3-89B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16099228, end_position_on_the_genomic_accession=16105670, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  41958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41958, CurrentID=0, Status='live', Symbol='Sb', Aliases='Dmel_CG4316, CG4316, Dmel\\CG4316, SP56-sbd, c-SP56, sbd, st-sb, Sb', description='Stubble', other_designations='CG4316-PA|CG4316-PB|Sb-PA|Sb-PB|Stubble-stubbloid|stubble|stubbloid', map_location='3-58.22 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16116161, end_position_on_the_genomic_accession=16145284, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  41959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41959, CurrentID=0, Status='live', Symbol='CG5903', Aliases='Dmel_ Dmel\\CG5903, anon-EST:Posey87, anon-WO0118547.99, bs05g07.y1', description='CG5903 gene product from transcript CG5903-RC', other_designations='CG5903-PA|CG5903-PB|CG5903-PC', map_location='89B6-89B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16144401, end_position_on_the_genomic_accession=16146855, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41960, CurrentID=0, Status='live', Symbol='CG5916', Aliases='Dmel_ Dmel\\CG5916', description='CG5916 gene product from transcript CG5916-RB', other_designations='CG5916-PA|CG5916-PB', map_location='89B6-89B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16147029, end_position_on_the_genomic_accession=16149636, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41961, CurrentID=0, Status='live', Symbol='Fer2', Aliases='Dmel_CG5952, CG5952, Dmel\\CG5952, bHLHa33, fer2', description='48 related 2', other_designations='48-related 2|48-related-2|CG5952-PB|CG5952-PC|CG5952-PD|CG5952-PE|Fer2-PB|Fer2-PC|Fer2-PD|Fer2-PE', map_location='89B6-89B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16149369, end_position_on_the_genomic_accession=16151312, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41962, CurrentID=0, Status='live', Symbol='CG6006', Aliases='Dmel_ BcDNA:SD04753, Dmel\\CG6006', description='CG6006 gene product from transcript CG6006-RC', other_designations='CG6006-PB|CG6006-PC|CG6006-PD|CG6006-PE|CG6006-PF', map_location='89B6-89B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16154982, end_position_on_the_genomic_accession=16171766, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  41963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41963, CurrentID=0, Status='live', Symbol='CG8925', Aliases='Dmel_ Dmel\\CG8925', description='CG8925 gene product from transcript CG8925-RA', other_designations='CG8925-PA|CG8925-PB', map_location='89B6-89B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16171991, end_position_on_the_genomic_accession=16180475, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  41964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41964, CurrentID=0, Status='live', Symbol='CG8927', Aliases='Dmel_ Dmel\\CG8927', description='CG8927 gene product from transcript CG8927-RB', other_designations='CG8927-PB|CG8927-PC', map_location='89B6-89B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16181020, end_position_on_the_genomic_accession=16186235, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41965, CurrentID=0, Status='live', Symbol='Bin1', Aliases='Dmel_CG6046, 150001_at, BcDNA:RE02417, Bin 1, CG6046, Dmel\\CG6046, SAP18, bin1, bin1/dSAP18, dSAP18, dSap18, dsap18, sap18', description='Bicoid interacting protein 1', other_designations='Bin1-PA|Bin1-PB|Bin1-PC|CG6046-PA|CG6046-PB|CG6046-PC|SIN3-associated polypeptide 18|Sin3A-associated polypeptide p18|bicoid interacting protein 1', map_location='89B7-89B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16186156, end_position_on_the_genomic_accession=16189564, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41967, CurrentID=0, Status='live', Symbol='CG6126', Aliases='Dmel_ CT19169, Dmel\\CG6126', description='CG6126 gene product from transcript CG6126-RA', other_designations='CG6126-PA', map_location='89B7-89B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16198000, end_position_on_the_genomic_accession=16203083, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41968, CurrentID=0, Status='live', Symbol='mtSSB', Aliases='Dmel_CG4337, 0046/13, CG4337, Dmel\\CG4337, DmmtSSB, SSB, SSB_DROME, c3a, lopo', description='mitochondrial single stranded DNA-binding protein', other_designations='CG4337-PA|low power|mitochondrial single strand binding protein|mitochondrial single-stranded DNA-binding|mitochondrial single-stranded DNA-binding protein|mtSSB-PA', map_location='89B7-89B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16204194, end_position_on_the_genomic_accession=16205250, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41971, CurrentID=0, Status='live', Symbol='asun', Aliases='Dmel_CG6814, ASUN, Asu, CG6814, Dmel\\CG6814, IntS13, Mat89Bb', description='asunder', other_designations='CG6814-PA|asun-PA|integrator 13|maternal transcript 89Bb', map_location='89B9-89B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16264271, end_position_on_the_genomic_accession=16266891, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41972, CurrentID=0, Status='live', Symbol='Mst89B', Aliases='Dmel_CG6864, CG6864, Dmel\\CG6864', description='CG6864 gene product from transcript CG6864-RA', other_designations='CG6864-PA|Mst89B-PA', map_location='89B9-89B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16267012, end_position_on_the_genomic_accession=16268433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41973, CurrentID=0, Status='live', Symbol='Mat89Ba', Aliases='Dmel_CG12785, CG12785, Dmel\\CG12785', description='Maternal transcript 89Ba', other_designations='CG12785-PA|Mat89Ba-PA|maternal transcript 89Ba', map_location='89B9-89B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16268329, end_position_on_the_genomic_accession=16272415, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  41975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41975, CurrentID=0, Status='live', Symbol='Zip89B', Aliases='Dmel_CG6898, CG6898, Dmel\\CG6898, ZIP3, Zip3, cg6898, dZip3, dZip89B', description='Zinc/iron regulated transporter-related protein 89B', other_designations='CG6898-PA|Zinc-iron transporter|Zinc/iron regulated transporter-related protein 3|Zinc/iron transporter-related protein 3|Zip89B-PA', map_location='89B12-89B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16305758, end_position_on_the_genomic_accession=16308700, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41977, CurrentID=0, Status='live', Symbol='CG6901', Aliases='Dmel_ Dmel\\CG6901', description='CG6901 gene product from transcript CG6901-RA', other_designations='CG6901-PA', map_location='89B12-89B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16323415, end_position_on_the_genomic_accession=16325128, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41978, CurrentID=0, Status='live', Symbol='CG17929', Aliases='Dmel_ Dmel\\CG17929', description='CG17929 gene product from transcript CG17929-RB', other_designations='CG17929-PB|uncharacterized protein', map_location='89B12-89B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16325492, end_position_on_the_genomic_accession=16327331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41979, CurrentID=0, Status='live', Symbol='CG17930', Aliases='Dmel_ Dmel\\CG17930', description='CG17930 gene product from transcript CG17930-RA', other_designations='CG17930-PA|uncharacterized protein', map_location='89B12-89B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16327849, end_position_on_the_genomic_accession=16329622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41981, CurrentID=0, Status='live', Symbol='pad', Aliases='Dmel_CG10309, CG10309, Dmel\\CG10309', description='poils au dos', other_designations='CG10309-PB|pad-PB', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16342318, end_position_on_the_genomic_accession=16345829, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  41982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41982, CurrentID=0, Status='live', Symbol='CG14879', Aliases='Dmel_ BcDNA:RE20991, Dmel\\CG14879', description='CG14879 gene product from transcript CG14879-RC', other_designations='CG14879-PA|CG14879-PB|CG14879-PC|CG14879-PD|uncharacterized protein', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16345825, end_position_on_the_genomic_accession=16349226, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41983, CurrentID=0, Status='live', Symbol='Manf', Aliases='Dmel_CG7013, ARP-like, CG7013, DmMANF, DmManf, Dmel\\CG7013, anon-EST:ParkEST188', description='Mesencephalic astrocyte-derived neurotrophic factor ortholog', other_designations='ARP-like|CG7013-PA|Manf-PA|mesencephalic astrocyte-derived neurotrophic factor|mesencephalic astrocyte-derived neurotrophic factor ortholog', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16349404, end_position_on_the_genomic_accession=16350801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41984, CurrentID=0, Status='live', Symbol='CG10311', Aliases='Dmel_ BEST:GH22794, Dmel\\CG10311, dmCG10311, gh22794', description='CG10311 gene product from transcript CG10311-RB', other_designations='CG10311-PA|CG10311-PB|uncharacterized protein', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16350786, end_position_on_the_genomic_accession=16357015, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  41985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41985, CurrentID=0, Status='live', Symbol='CG42446', Aliases='Dmel_ Dmel\\CG42446', description='CG42446 gene product from transcript CG42446-RA', other_designations='CG42446-PA|uncharacterized protein', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16358144, end_position_on_the_genomic_accession=16360108, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41986, CurrentID=0, Status='live', Symbol='CG14880', Aliases='Dmel_ Dmel\\CG14880', description='CG14880 gene product from transcript CG14880-RA', other_designations='CG14880-PA|CG14880-PB|uncharacterized protein', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16360329, end_position_on_the_genomic_accession=16368306, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  41987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41987, CurrentID=0, Status='live', Symbol='CG10317', Aliases='Dmel_ Dmel\\CG10317', description='CG10317 gene product from transcript CG10317-RD', other_designations='CG10317-PC|CG10317-PD|uncharacterized protein', map_location='89B13-89B14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16368855, end_position_on_the_genomic_accession=16370533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41988, CurrentID=0, Status='live', Symbol='ss', Aliases='Dmel_CG6993, AHR, AhR, CG6993, Dmel\\CG6993, SS, Ssa, ss', description='spineless', other_designations='CG6993-PA|CG6993-PC|CG6993-PD|aristapedia|spineless-aristapedia|ss-PA|ss-PC|ss-PD', map_location='3-58.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16374434, end_position_on_the_genomic_accession=16403690, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  41989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41989, CurrentID=0, Status='live', Symbol='CG17565', Aliases='Dmel_ Dmel\\CG17565, FNTB', description='CG17565 gene product from transcript CG17565-RA', other_designations='CG17565-PA|farnesyl transferase beta|uncharacterized protein', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16432998, end_position_on_the_genomic_accession=16434740, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41990, CurrentID=0, Status='live', Symbol='CG14881', Aliases='Dmel_ Dmel\\CG14881', description='CG14881 gene product from transcript CG14881-RB', other_designations='CG14881-PA|CG14881-PB|uncharacterized protein', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16434899, end_position_on_the_genomic_accession=16436066, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41991, CurrentID=0, Status='live', Symbol='mRpS33', Aliases='Dmel_CG10406, CG10406, Dmel\\CG10406, MRP-S33', description='mitochondrial ribosomal protein S33', other_designations='CG10406-PA|mRpS33-PA', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16436045, end_position_on_the_genomic_accession=16436597, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41992, CurrentID=0, Status='live', Symbol='ema', Aliases='Dmel_CG12753, CG12753, Dmel\\CG12753', description='endosomal maturation defective', other_designations='CG12753-PA|CG12753-PB|CG12753-PC|ema-PA|ema-PB|ema-PC', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16436880, end_position_on_the_genomic_accession=16441661, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  41993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41993, CurrentID=0, Status='live', Symbol='CG14894', Aliases='Dmel_ Dmel\\CG14894', description='CG14894 gene product from transcript CG14894-RB', other_designations='CG14894-PA|CG14894-PB|uncharacterized protein', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16441843, end_position_on_the_genomic_accession=16443007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41994, CurrentID=0, Status='live', Symbol='CG14882', Aliases='Dmel_ BcDNA:SD10961, Dmel\\CG14882', description='CG14882 gene product from transcript CG14882-RB', other_designations='CG14882-PA|CG14882-PB|uncharacterized protein', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16443207, end_position_on_the_genomic_accession=16445010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41995, CurrentID=0, Status='live', Symbol='Pak3', Aliases='Dmel_CG14895, CG14895, D- DPAK3, DPak3, DmPAK3, Dmel\\CG14895, Dpak3, PAK3, PaK3, anon-WO0118547.285, dPAK3, dpak3, pak3', description='CG14895 gene product from transcript CG14895-RD', other_designations='CG14895-PA|CG14895-PC|CG14895-PD|Pak3-PA|Pak3-PC|Pak3-PD|p21-activated kinase 3', map_location='89B16-89B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16445007, end_position_on_the_genomic_accession=16453821, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  41996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41996, CurrentID=0, Status='live', Symbol='CG10405', Aliases='Dmel_ Dmel\\CG10405, SP155', description='CG10405 gene product from transcript CG10405-RB', other_designations='CG10405-PB|uncharacterized protein', map_location='89B17-89B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16454492, end_position_on_the_genomic_accession=16455797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41997, CurrentID=0, Status='live', Symbol='CG14883', Aliases='Dmel_ Dmel\\CG14883', description='CG14883 gene product from transcript CG14883-RA', other_designations='CG14883-PA|uncharacterized protein', map_location='89B17-89B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16456052, end_position_on_the_genomic_accession=16458233, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42000, CurrentID=0, Status='live', Symbol='CSN5', Aliases='Dmel_CG14884, BcDNA.LD14392, BcDNA:LD14392, CG14884, CH5, CSN-5, Csn5, DCH5, Dch5, Dmel\\CG14884, JAB1, JAB1/CSN5, Jab1, JadBp, Q9XZ58, csn5, dch5, jab1/csn5, l(3)89Cf, l(3)L4032, l(3)M203, quo', description='COP9 signalosome subunit 5', other_designations='CG14884-PA|CG14884-PB|COP9 complex homolog subunit 5|COP9 complex subunit 5|CSN5-PA|CSN5-PB|JUN activation domain-binding protein-1|Jun activation domain binding protein|drosophila COP9 signalosome homolog 5', map_location='89B17-89B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16470099, end_position_on_the_genomic_accession=16471406, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42001, CurrentID=0, Status='live', Symbol='Gyc-89Da', Aliases='Dmel_CG14885, CG14885, Dmel\\CG14885, GYCb', description='Guanylyl cyclase at 89Da', other_designations='CG14885-PB|CG14885-PC|Gyc-89Da-PB|Gyc-89Da-PC|guanylyl cyclase at 89Da', map_location='89B17-89B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16472601, end_position_on_the_genomic_accession=16475477, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42002, CurrentID=0, Status='live', Symbol='Gyc-89Db', Aliases='Dmel_CG14886, CG14886, Dmel\\CG14886', description='Guanylyl cyclase at 89Db', other_designations='CG14886-PA|Gyc-89Db-PA|guanylyl cyclase at 89Db', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16477110, end_position_on_the_genomic_accession=16481070, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42003, CurrentID=0, Status='live', Symbol='Dhfr', Aliases='Dmel_CG14887, CG14887, DHFR, Dmel\\CG14887, dhfr', description='Dihydrofolate reductase', other_designations='CG14887-PA|CG14887-PB|Dhfr-PA|Dhfr-PB|dihydrofolate reductase', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16481167, end_position_on_the_genomic_accession=16481878, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42004, CurrentID=0, Status='live', Symbol='Sdhaf3', Aliases='Dmel_CG14898, BcDNA:RE23580, CG14898, Dmel\\CG14898, Sdh7, dSdhaf3, sdh7', description='Succinate dehydrogenase assembly factor 3', other_designations='CG14898-PA|Sdhaf3-PA|succinate dehydrogenase assembly factor 3', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16481861, end_position_on_the_genomic_accession=16482452, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42005, CurrentID=0, Status='live', Symbol='Der-2', Aliases='Dmel_CG14899, CG14899, Cg14899, DER1, Dmel\\CG14899', description='Derlin-2', other_designations='CG14899-PA|Der-2-PA', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16482694, end_position_on_the_genomic_accession=16484223, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42006, CurrentID=0, Status='live', Symbol='Cad89D', Aliases='Dmel_CG14900, CG14900, CT34724, DCad89D, Dmel\\CG14900', description='Cadherin 89D', other_designations='CG14900-PB|Cad89D-PB|cadherin 89D', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16484566, end_position_on_the_genomic_accession=16495422, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  42007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42007, CurrentID=0, Status='live', Symbol='Gr89a', Aliases='Dmel_CG14901, CG14901, Dmel\\CG14901, GR89a', description='Gustatory receptor 89a', other_designations='CG14901-PA|Gr89a-PA|gustatory receptor 89a', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16497817, end_position_on_the_genomic_accession=16498959, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42008, CurrentID=0, Status='live', Symbol='Decay', Aliases='Dmel_CG14902, CAS3, CC3, CG14902, Cas3, Casp3, Caspase-3, DECAY, Dmel\\CG14902, Drosophila executioner caspase related to Apopain/Yama, cas3, caspase 3, caspase-3, decay', description='Death executioner caspase related to Apopain/Yama', other_designations='CG14902-PA|Caspase-3|Decay-PA|caspase 3|caspase-3 protein|cleaved caspase 3|death executioner caspase related to Apopain/Yama|decay', map_location='89B18-89B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16499058, end_position_on_the_genomic_accession=16501050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42013, CurrentID=0, Status='live', Symbol='CG14891', Aliases='Dmel_ Dmel\\CG14891, FBXL13', description='CG14891 gene product from transcript CG14891-RA', other_designations='CG14891-PA|uncharacterized protein', map_location='89B20-89B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16571515, end_position_on_the_genomic_accession=16573158, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42014, CurrentID=0, Status='live', Symbol='CG14903', Aliases='Dmel_ Dmel\\CG14903', description='CG14903 gene product from transcript CG14903-RA', other_designations='CG14903-PA|uncharacterized protein', map_location='89B20-89B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16573179, end_position_on_the_genomic_accession=16573752, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42015, CurrentID=0, Status='live', Symbol='Scp2', Aliases='Dmel_CG14904, CE, CG14904, Cex, Dcabp-A.2, Dmel\\CG14904, JHDK, SCP2, cex, dSCP2', description='Sarcoplasmic calcium-binding protein 2', other_designations='CG14904-PA|Ca2+-sensing low molecular weight GTPase|Cex Ca2+-sensing low molecular weight|Scp2-PA|calexcitin|drosophila Ca[2+]-binding protein A.2|sarcoplasmic calcium-binding protein 2', map_location='89B20-89D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16574543, end_position_on_the_genomic_accession=16583663, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42016, CurrentID=0, Status='live', Symbol='CG14892', Aliases='Dmel_ Dmel\\CG14892, SP175', description='CG14892 gene product from transcript CG14892-RA', other_designations='CG14892-PA|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16594414, end_position_on_the_genomic_accession=16597241, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42017, CurrentID=0, Status='live', Symbol='CG10345', Aliases='Dmel_ BcDNA, BcDNA:GH08773, Dmel\\CG10345', description='CG10345 gene product from transcript CG10345-RA', other_designations='CG10345-PA|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16597073, end_position_on_the_genomic_accession=16604861, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42018, CurrentID=0, Status='live', Symbol='NPF', Aliases='Dmel_CG10342, CG10342, Dm-NPF, DmNPF, Dmel\\CG10342, Drm-NPF, NP-PP-A1, NPF-A2, NPF89D3, Npf, dNPF, dnpf, npf, NPF', description='neuropeptide F', other_designations='CG10342-PA|CG10342-PB|CG10342-PC|NPF-PA|NPF-PB|NPF-PC|NPY-like neuropeptide F|long neuropeptide F|neuropeptide F 1|neuropeptide-F', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16609220, end_position_on_the_genomic_accession=16611325, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42019, CurrentID=0, Status='live', Symbol='CG12783', Aliases='Dmel_ Dmel\\CG12783', description='CG12783 gene product from transcript CG12783-RC', other_designations='CG12783-PB|CG12783-PC|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16611419, end_position_on_the_genomic_accession=16613400, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42020, CurrentID=0, Status='live', Symbol='CG10340', Aliases='Dmel_ Dmel\\CG10340, anon-EST:fe2A5, anon2A5', description='CG10340 gene product from transcript CG10340-RA', other_designations='CG10340-PA|anon-fast-evolving-2A5|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16613398, end_position_on_the_genomic_accession=16614391, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42021, CurrentID=0, Status='live', Symbol='CG17562', Aliases='Dmel_ Dmel\\CG17562', description='CG17562 gene product from transcript CG17562-RA', other_designations='CG17562-PA|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16615309, end_position_on_the_genomic_accession=16617154, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42022, CurrentID=0, Status='live', Symbol='CG17560', Aliases='Dmel_ CT38785, Dmel\\CG17560', description='CG17560 gene product from transcript CG17560-RA', other_designations='CG17560-PA|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16617950, end_position_on_the_genomic_accession=16619878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42023, CurrentID=0, Status='live', Symbol='CG14893', Aliases='Dmel_ CT34717, Dmel\\CG14893', description='CG14893 gene product from transcript CG14893-RB', other_designations='CG14893-PB|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16620778, end_position_on_the_genomic_accession=16622702, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42024, CurrentID=0, Status='live', Symbol='CG14905', Aliases='Dmel_ Dmel\\CG14905', description='CG14905 gene product from transcript CG14905-RA', other_designations='CG14905-PA|uncharacterized protein', map_location='89D5-89D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16622735, end_position_on_the_genomic_accession=16624717, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42025, CurrentID=0, Status='live', Symbol='Fas1', Aliases='Dmel_CG6588, CG6588, DROFAS, DROFAS1, DROFASI, Dmel\\CG6588, Fas I, FasI, MRE24, fas I, fas-I, fas1, fasI, gasI', description='Fasciclin 1', other_designations='CG6588-PA|CG6588-PB|CG6588-PC|CG6588-PD|CG6588-PE|CG6588-PF|CG6588-PG|Fas1-PA|Fas1-PB|Fas1-PC|Fas1-PD|Fas1-PE|Fas1-PF|Fas1-PG|Fasciclin-1|fasciclin 1|fasciclin I|mRNA-like ncRNA in embryogenesis 24', map_location='89D5-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16625238, end_position_on_the_genomic_accession=16639771, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  42026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42026, CurrentID=0, Status='live', Symbol='nonA-l', Aliases='Dmel_CG10328, BEST:LD09360, CG10328, Dmel\\CG10328, LD09360, PSF, Z, nonA 1, nonA-1', description='nonA-like', other_designations='CG10328-PA|CG10328-PB|no-on transient A gene|no-on transient A-like|nonA-l-PA|nonA-l-PB|predicted gene Z', map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16640205, end_position_on_the_genomic_accession=16643112, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42027, CurrentID=0, Status='live', Symbol='CG10326', Aliases='Dmel_ Dmel\\CG10326', description='CG10326 gene product from transcript CG10326-RA', other_designations='CG10326-PA|CG10326-PB|uncharacterized protein', map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16643605, end_position_on_the_genomic_accession=16645171, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42028, CurrentID=0, Status='live', Symbol='CG10324', Aliases='Dmel_ Dmel\\CG10324', description='CG10324 gene product from transcript CG10324-RA', other_designations='CG10324-PA|uncharacterized protein', map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16645072, end_position_on_the_genomic_accession=16647004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42029, CurrentID=0, Status='live', Symbol='Cctgamma', Aliases='Dmel_CG8977, CG8977, Cct-gamma, Cct3, Cctg, Dmel\\CG8977, TCPG_DROME, Tcp-1-gamma, Tcp-1gamma, Tcp1-gamma, Y, cct-gamma, cctg, cctgamma, geneY, vine', description='CG8977 gene product from transcript CG8977-RB', other_designations='CCT-gamma protein|CCTgamma-like|CG8977-PA|CG8977-PB|CG8977-PC|CG8977-PD|Cctgamma-PA|Cctgamma-PB|Cctgamma-PC|Cctgamma-PD|TCP-1 subunit 3gamma|cctgamma|chaperonin-containing t-complex protein-1 gamma subunit like|predicted gene Y|vine', map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16647605, end_position_on_the_genomic_accession=16649978, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42030, CurrentID=0, Status='live', Symbol='CG14906', Aliases='Dmel_ Dmel\\CG14906, ORF2, anon-89Eb', description='CG14906 gene product from transcript CG14906-RC', other_designations='CG14906-PC|open reading frame 2|uncharacterized protein', map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16649981, end_position_on_the_genomic_accession=16651157, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42031, CurrentID=0, Status='live', Symbol='CG14907', Aliases='Dmel_ Dmel\\CG14907', description='CG14907 gene product from transcript CG14907-RA', other_designations='CG14907-PA|uncharacterized protein', map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16651217, end_position_on_the_genomic_accession=16652319, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42032, CurrentID=0, Status='live', Symbol='modSP', Aliases='Dmel_CG31217, CG10100, CG17557, CG31217, CT28389, Dm-SP, Dmel\\CG31217, LDL, Ldlr, ModSP, SP172, SP53, Ser89E, W, X', description='modular serine protease', other_designations="CG31217-PA|low density lipoprotein receptor-like 'a' repeats|modSP-PA|predicted gene W|predicted gene X|serine protease like 89E|serine protease-like", map_location='89D6-89D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16652446, end_position_on_the_genomic_accession=16655662, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42034, CurrentID=0, Status='live', Symbol='Ubx', Aliases='Dmel_CG10388, BX-C, Bxl, CG10388, Cbx, DUbx, Dm Ubx, DmUbx, Dmel\\CG10388, Hm, UBX, UBx, Uab1b, abx, bx, bx[D], bxd, bxd[D], bxl, l(3)89Eb, pbx, ubx, Ubx', description='Ultrabithorax', other_designations='CG10388-PA|CG10388-PB|CG10388-PC|CG10388-PD|CG10388-PE|CG10388-PF|Ubx-PA|Ubx-PB|Ubx-PC|Ubx-PD|Ubx-PE|Ubx-PF|anterobithorax|bithorax|bithorax complex|bithoraxoid|contrabithorax|haltere mimic|postbithorax|ultabithorax|ultarbithorax|ultraabdominal|ultrabithorax', map_location='3-58.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16656623, end_position_on_the_genomic_accession=16734426, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42036, CurrentID=0, Status='live', Symbol='Glut3', Aliases='Dmel_CG3853, CG3853, Dmel\\CG3853, GLU, glut-l', description='Glucose transporter type 3', other_designations='CG3853-PA|Glut3-PA|glucose transporter type 3|glucose transporter-like', map_location='89E1-89E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16788311, end_position_on_the_genomic_accession=16790021, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42037, CurrentID=0, Status='live', Symbol='abd-A', Aliases='Dmel_CG10325, ABD-A, Abd-A, AbdA, Abda, BX-C, CG10325, Cbxd, DmabdA, Dmabda, Dmel\\CG10325, Fab-3, Fab-4, Hab, abd A, abd-a, abdA, abd_a, abda, iab, iab-2, iab-3, iab-4, iab-5, iab2, iab3, iab4, l(3)89Ec', description='abdominal A', other_designations='Abdominal-A|CG10325-PA|CG10325-PB|CG10325-PC|CG10325-PD|abd-A-PA|abd-A-PB|abd-A-PC|abd-A-PD|abdominalA|contrabithoraxoid|hyperabdominal|infra-abdominal 2|infra-abdominal 3|infra-abdominal 4|infraabdominal|ultraabdominal', map_location='3-58.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16807214, end_position_on_the_genomic_accession=16830049, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42041, CurrentID=0, Status='live', Symbol='VhaM9.7-d', Aliases='Dmel_CG14909, CG14909, Dmel\\CG14909, vhaM9.7-4', description='Vacuolar H[+] ATPase M9.7 subunit d', other_designations='CG14909-PA|VhaM9.7-d-PA|vacuolar H[+] ATPase M9.7 subunit d|vacuolar H[+] ATPase subunit M9.7-d', map_location='89E5-89E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16975204, end_position_on_the_genomic_accession=16975697, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42043, CurrentID=0, Status='live', Symbol='Ahcy89E', Aliases='Dmel_CG8956, AHCY, CG8956, Dmel\\CG8956, S-AHCY, pH200', description='Adenosylhomocysteinase 89E', other_designations='AdoHcyase-like|Ahcy89E-PC|Ahcy89E-PD|CG8956-PC|CG8956-PD|S-adenosyl homocysteine hydrolase-like|adenosylhomocysteinase 89E', map_location='89E5-89E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16978439, end_position_on_the_genomic_accession=16982158, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42044, CurrentID=0, Status='live', Symbol='CG18622', Aliases='Dmel_ BcDNA:RH18149, Dmel\\CG18622', description='CG18622 gene product from transcript CG18622-RA', other_designations='CG18622-PA|CG18622-PB|CG18622-PC|uncharacterized protein', map_location='89E5-89E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16983510, end_position_on_the_genomic_accession=16987732, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42045, CurrentID=0, Status='live', Symbol='CG3678', Aliases='Dmel_ Dmel\\CG3678, RE62710p', description='CG3678 gene product from transcript CG3678-RB', other_designations='CG3678-PA|CG3678-PB', map_location='89E6-89E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16990334, end_position_on_the_genomic_accession=16991878, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42046, CurrentID=0, Status='live', Symbol='CG17556', Aliases='Dmel_ Dmel\\CG17556', description='CG17556 gene product from transcript CG17556-RA', other_designations='CG17556-PA|CG17556-PB|uncharacterized protein', map_location='89E6-89E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16994542, end_position_on_the_genomic_accession=16996071, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42047, CurrentID=0, Status='live', Symbol='CG3534', Aliases='Dmel_ Dmel\\CG3534', description='CG3534 gene product from transcript CG3534-RA', other_designations='CG3534-PA', map_location='89E6-89E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16996446, end_position_on_the_genomic_accession=16999449, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42048, CurrentID=0, Status='live', Symbol='CG16941', Aliases='Dmel_ Dmel\\CG16941, SF3a1, SF3a120, Sf3a1, l(2)SH3 146, l(3)SH146', description='CG16941 gene product from transcript CG16941-RA', other_designations='CG16941-PA|lethal (3) SH146|splicing factor 3a, subunit 1|uncharacterized protein', map_location='89E6-89E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16999384, end_position_on_the_genomic_accession=17002355, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42049, CurrentID=0, Status='live', Symbol='Irc', Aliases='Dmel_CG8913, CG8913, Dmel\\CG8913, IRC', description='Immune-regulated catalase', other_designations='CG8913-PA|CG8913-PB|Irc-PA|Irc-PB|immune-regulated catalase|immune-related catalase|peroxidase', map_location='89E7-89E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17002875, end_position_on_the_genomic_accession=17007208, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42050, CurrentID=0, Status='live', Symbol='CG8907', Aliases='Dmel_ Dmel\\CG8907, LP01469p', description='CG8907 gene product from transcript CG8907-RA', other_designations='CG8907-PA|CG8907-PB', map_location='89E8-89E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17008269, end_position_on_the_genomic_accession=17012533, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42051, CurrentID=0, Status='live', Symbol='AdSL', Aliases='Dmel_CG3590, CG3590, Dmel\\CG3590', description='Adenylosuccinate Lyase', other_designations='AdSL-PA|CG3590-PA|adenylosuccinate lyase', map_location='89E8-89E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17018780, end_position_on_the_genomic_accession=17021120, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42053, CurrentID=0, Status='live', Symbol='CG5225', Aliases='Dmel_ Dmel\\CG5225', description='CG5225 gene product from transcript CG5225-RA', other_designations='CG5225-PA', map_location='89E9-89E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17028693, end_position_on_the_genomic_accession=17030701, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42054, CurrentID=0, Status='live', Symbol='CG4009', Aliases='Dmel_ Dmel\\CG4009', description='CG4009 gene product from transcript CG4009-RC', other_designations='CG4009-PB|CG4009-PC', map_location='89E10-89E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17033588, end_position_on_the_genomic_accession=17036045, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42055, CurrentID=0, Status='live', Symbol='CG18213', Aliases='Dmel_ Dmel\\CG18213', description='CG18213 gene product from transcript CG18213-RD', other_designations='CG18213-PB|CG18213-PC|CG18213-PD|uncharacterized protein', map_location='89E10-89E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17036394, end_position_on_the_genomic_accession=17040203, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  42056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42056, CurrentID=0, Status='live', Symbol='CG5220', Aliases='Dmel_ Dm TRM7, Dmel\\CG5220', description='CG5220 gene product from transcript CG5220-RA', other_designations='CG5220-PA', map_location='89E10-89E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17040682, end_position_on_the_genomic_accession=17041895, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42057, CurrentID=0, Status='live', Symbol='CG3995', Aliases='Dmel_ Dmel\\CG3995', description='CG3995 gene product from transcript CG3995-RA', other_designations='CG3995-PA', map_location='89E10-89E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17041896, end_position_on_the_genomic_accession=17043136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42058, CurrentID=0, Status='live', Symbol='Patr-1', Aliases='Dmel_CG5208, BcDNA:LD27979, CG5208, Dmel\\CG5208, HPat, LD27979, anon-WO0118547.282, hpat, l(3)06442, patr-1', description='Protein associated with topo II related - 1', other_designations='CG5208-PA|CG5208-PB|Patr-1-PA|Patr-1-PB|lethal (3) 06442|protein associated with topo II related - 1', map_location='89E10-89E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17043564, end_position_on_the_genomic_accession=17048421, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42059, CurrentID=0, Status='live', Symbol='Dad', Aliases='Dmel_CG5201, 1883, CG5201, DAD1, Dmel\\CG5201, EP(3)3196, EP3196, anon-EST:fe1F5, dad, l(3)j1E4, Dad', description='Daughters against dpp', other_designations='CG5201-PA|CG5201-PB|CG5201-PC|Dad-PA|Dad-PB|Dad-PC|anon-fast-evolving-1F5|daughter against dpp|daughters against Dpp|daughters against dpp', map_location='89E11-89E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17054007, end_position_on_the_genomic_accession=17070408, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42060, CurrentID=0, Status='live', Symbol='Ns1', Aliases='Dmel_CG3983, CG3983, Dmel\\CG3983, NS1, dmNS1, ns1', description='Nucleostemin ortholog (H. sapiens) 1', other_designations='CG3983-PA|Ns1-PA|nucleostemin 1|nucleostemin ortholog 1', map_location='89E11-89E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17070523, end_position_on_the_genomic_accession=17072675, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42061, CurrentID=0, Status='live', Symbol='mRpS11', Aliases='Dmel_CG5184, BcDNA:RH70248, CG5184, Dmel\\CG5184, GH1, MRP-S11', description='mitochondrial ribosomal protein S11', other_designations='CG5184-PA|mRpS11-PA', map_location='89E11-89E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17072732, end_position_on_the_genomic_accession=17073795, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42062, CurrentID=0, Status='live', Symbol='Keap1', Aliases='Dmel_CG3962, CG3962, Dmel\\CG3962, KEAP1, Nrf-2, dKEAP1, d dkeap-1, dkeap1, keap1', description='CG3962 gene product from transcript CG3962-RC', other_designations='CG3962-PA|CG3962-PB|CG3962-PC|Keap1-PA|Keap1-PB|Keap1-PC', map_location='89E11-89E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17073942, end_position_on_the_genomic_accession=17079945, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42063, CurrentID=0, Status='live', Symbol='kuk', Aliases='Dmel_CG5175, CG5175, Dmel\\CG5175, Kuk, char/char, kuk', description='kugelkern', other_designations='CG5175-PA|CG5175-PB|CG5175-PC|Kugelkern/Charleston|charleston|kugelkem|kuk-PA|kuk-PB|kuk-PC', map_location='89E12-89E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17080244, end_position_on_the_genomic_accession=17084781, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42064, CurrentID=0, Status='live', Symbol='GckIII', Aliases='Dmel_CG5169, CG5169, Dmel\\CG5169, GCKIII, STLK3, dGCKIII, wheezy', description='Germinal centre kinase III', other_designations='CG5169-PA|GckIII-PA|germinal center kinase III|germinal centre kinase III|wheezy', map_location='89E12-89E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17085513, end_position_on_the_genomic_accession=17087988, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42065, CurrentID=0, Status='live', Symbol='cal1', Aliases='Dmel_CG5148, CAL1, CG5148, CLD2, Cal1, Dmel\\CG5148', description='chromosome alignment defect 1', other_designations='CG5148-PA|cal1-PA', map_location='89E12-89E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17088280, end_position_on_the_genomic_accession=17091938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42066, CurrentID=0, Status='live', Symbol='cher', Aliases='Dmel_CG3937, CG3937, Cher, Dmel\\CG3937, FLN, Fil, joy, sko', description='cheerio', other_designations='CG3937-PA|CG3937-PB|CG3937-PF|CG3937-PG|CG3937-PH|CG3937-PI|CG3937-PK|CG3937-PL|CG3937-PM|CG3937-PN|Filamin-1|Filamin-A|cher-PA|cher-PB|cher-PF|cher-PG|cher-PH|cher-PI|cher-PK|cher-PL|cher-PM|cher-PN|filamin|filamin B|filamin isoform B|filamin-240|shi kong', map_location='3-60 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17091533, end_position_on_the_genomic_accession=17125389, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  42067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42067, CurrentID=0, Status='live', Symbol='CG17477', Aliases='Dmel_ Dmel\\CG17477, SPH174', description='CG17477 gene product from transcript CG17477-RA', other_designations='CG17477-PA|chymotrypsin-type peptidase|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17126088, end_position_on_the_genomic_accession=17126938, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42068, CurrentID=0, Status='live', Symbol='CG31265', Aliases='Dmel_ CG5233, Dmel\\CG31265, SP126', description='CG31265 gene product from transcript CG31265-RA', other_designations='CG31265-PA|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17128212, end_position_on_the_genomic_accession=17129120, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42069, CurrentID=0, Status='live', Symbol='CG17475', Aliases='Dmel_ Dmel\\CG17475, SP131', description='CG17475 gene product from transcript CG17475-RB', other_designations='CG17475-PA|CG17475-PB|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17129339, end_position_on_the_genomic_accession=17130426, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42070, CurrentID=0, Status='live', Symbol='CG4053', Aliases='Dmel_ Dmel\\CG4053, SP185', description='CG4053 gene product from transcript CG4053-RB', other_designations='CG4053-PB', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17131211, end_position_on_the_genomic_accession=17132198, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42071, CurrentID=0, Status='live', Symbol='CG31266', Aliases='Dmel_ CG5240, Dmel\\CG31266, SPH146', description='CG31266 gene product from transcript CG31266-RB', other_designations='CG31266-PB|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17132662, end_position_on_the_genomic_accession=17133719, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42072, CurrentID=0, Status='live', Symbol='CG5246', Aliases='Dmel_ BcDNA:LP06628, Dmel\\CG5246, SP96', description='CG5246 gene product from transcript CG5246-RA', other_designations='CG5246-PA', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17136111, end_position_on_the_genomic_accession=17137195, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42073, CurrentID=0, Status='live', Symbol='CG5255', Aliases='Dmel_ Dmel\\CG5255, SP141', description='CG5255 gene product from transcript CG5255-RA', other_designations='CG5255-PA', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17137893, end_position_on_the_genomic_accession=17138865, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42074, CurrentID=0, Status='live', Symbol='CG5265', Aliases='Dmel_ Dmel\\CG5265', description='CG5265 gene product from transcript CG5265-RA', other_designations='CG5265-PA', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17138999, end_position_on_the_genomic_accession=17141168, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42075, CurrentID=0, Status='live', Symbol='TwdlW', Aliases='Dmel_CG4060, CG4060, Dmel\\CG4060', description='TweedleW', other_designations='CG4060-PA|TwdlW-PA|tweedle W', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17141208, end_position_on_the_genomic_accession=17142404, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42076, CurrentID=0, Status='live', Symbol='m-cup', Aliases='Dmel_CG11896, CG11896, Dmel\\CG11896', description='mann-cup', other_designations='CG11896-PA|CG11896-PB|m-cup-PA|m-cup-PB', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17144475, end_position_on_the_genomic_accession=17149412, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42077, CurrentID=0, Status='live', Symbol='Det', Aliases='Dmel_CG12265, CG12265, Dmel\\CG12265, scpo, svn', description='Deterin', other_designations='CG12265-PA|Det-PA|dSurvivin|deterin|scapolo|survivin', map_location='3-59.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17149667, end_position_on_the_genomic_accession=17150455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42078, CurrentID=0, Status='live', Symbol='CG5285', Aliases='Dmel_ Dmel\\CG5285', description='CG5285 gene product from transcript CG5285-RA', other_designations='CG5285-PA', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17150587, end_position_on_the_genomic_accession=17151630, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42079, CurrentID=0, Status='live', Symbol='CG5292', Aliases='Dmel_ BcDNA:RE41712, Dmel\\CG5292', description='CG5292 gene product from transcript CG5292-RB', other_designations='CG5292-PA|CG5292-PB', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17151861, end_position_on_the_genomic_accession=17153771, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42080, CurrentID=0, Status='live', Symbol='Mur89F', Aliases='Dmel_CG4090, CG4090, CT13556, Dmel\\CG4090', description='Mucin related 89F', other_designations='CG4090-PB|CG4090-PC|Mur89F-PB|Mur89F-PC|mucin related 89F', map_location='89F1-89F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17152995, end_position_on_the_genomic_accession=17202130, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42082, CurrentID=0, Status='live', Symbol='beat-IIb', Aliases='Dmel_CG4135, Beat-IIb, CG14337, CG4135, CT13620, Dmel\\CG4135, beat IIb', description='CG4135 gene product from transcript CG4135-RC', other_designations='CG4135-PA|CG4135-PC|beat-IIb-PA|beat-IIb-PC', map_location='89F3-90A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17232253, end_position_on_the_genomic_accession=17265049, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42084, CurrentID=0, Status='live', Symbol='CG31418', Aliases='Dmel_ CG5329, Dmel\\CG31418', description='CG31418 gene product from transcript CG31418-RA', other_designations='CG31418-PA|uncharacterized protein', map_location='90A1-90A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17283795, end_position_on_the_genomic_accession=17284511, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42086, CurrentID=0, Status='live', Symbol='beat-IIa', Aliases='Dmel_CG14334, BcDNA:RE17794, Beat-IIa, CG14334, CG14335, CT33965, Dmel\\CG14334, beat IIa', description='beaten path IIa', other_designations='CG14334-PA|CG14334-PB|beat-IIa-PA|beat-IIa-PB', map_location='90A2-90A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17320004, end_position_on_the_genomic_accession=17366119, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42087, CurrentID=0, Status='live', Symbol='Brf', Aliases='Dmel_CG31256, BRF, BRF11, CG31256, CG4155, CG5419, Dmel\\CG31256, TFIIIB, brf', description='CG31256 gene product from transcript CG31256-RA', other_designations='Brf-PA|Brf-PB|CG31256-PA|CG31256-PB|RNA Pol III transcription factor', map_location='90A3-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17374833, end_position_on_the_genomic_accession=17391337, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42089, CurrentID=0, Status='live', Symbol='lute', Aliases='Dmel_CG43226, CG14333, CG43226, CG5319, Dmel\\CG43226, Dmel_CG14333, Dmel_CG5319, anon-WO0118547.289', description='CG43226 gene product from transcript CG43226-RC', other_designations='CG43226-PA|CG43226-PB|CG43226-PC|CG43226-PD|lute-PA|lute-PB|lute-PC|lute-PD', map_location='90A4-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17379944, end_position_on_the_genomic_accession=17390265, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42091, CurrentID=0, Status='live', Symbol='sds22', Aliases='Dmel_CG5851, CG5851, CT18355, Dmel\\CG5851, Sds22', description='CG5851 gene product from transcript CG5851-RA', other_designations='CG5851-PA|sds22-PA', map_location='90A5-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17391594, end_position_on_the_genomic_accession=17393020, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42092, CurrentID=0, Status='live', Symbol='CREG', Aliases='Dmel_CG5413, BcDNA:LP11958, CG5413, Dmel\\CG5413, creg, dCREG', description='Cellular Repressor of E1A-stimulated Genes', other_designations='CG5413-PA|CG5413-PB|CG5413-PC|CG5413-PD|CREG-PA|CREG-PB|CREG-PC|CREG-PD|cellular repressor of E1A-stimulated genes', map_location='90A5-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17393272, end_position_on_the_genomic_accession=17394640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42093, CurrentID=0, Status='live', Symbol='Sur-8', Aliases='Dmel_CG5407, CG5407, CT17146, Dmel\\CG5407, SUR-8, Sur8', description='CG5407 gene product from transcript CG5407-RA', other_designations='CG5407-PA|CG5407-PB|CG5407-PE|CG5407-PF|Sur-8-PA|Sur-8-PB|Sur-8-PE|Sur-8-PF', map_location='90A5-90A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17395199, end_position_on_the_genomic_accession=17402515, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42094, CurrentID=0, Status='live', Symbol='CG17283', Aliases='Dmel_ Dmel\\CG17283', description='CG17283 gene product from transcript CG17283-RA', other_designations='CG17283-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17407876, end_position_on_the_genomic_accession=17409460, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42095, CurrentID=0, Status='live', Symbol='CG5860', Aliases='Dmel_ Dmel\\CG5860, FBgn0038506', description='CG5860 gene product from transcript CG5860-RA', other_designations='CG5860-PA', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17409619, end_position_on_the_genomic_accession=17410840, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42096, CurrentID=0, Status='live', Symbol='CG5863', Aliases='Dmel_ Dmel\\CG5863', description='CG5863 gene product from transcript CG5863-RA', other_designations='CG5863-PA', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17411108, end_position_on_the_genomic_accession=17412680, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42097, CurrentID=0, Status='live', Symbol='CG5866', Aliases='Dmel_ Dmel\\CG5866', description='CG5866 gene product from transcript CG5866-RB', other_designations='CG5866-PB', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17415739, end_position_on_the_genomic_accession=17416505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42098, CurrentID=0, Status='live', Symbol='CG14332', Aliases='Dmel_ Dmel\\CG14332', description='CG14332 gene product from transcript CG14332-RA', other_designations='CG14332-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17421690, end_position_on_the_genomic_accession=17422202, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42099, CurrentID=0, Status='live', Symbol='CG14331', Aliases='Dmel_ Dmel\\CG14331', description='CG14331 gene product from transcript CG14331-RC', other_designations='CG14331-PC|uncharacterized protein', map_location='90B1-90B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17438968, end_position_on_the_genomic_accession=17440728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42100, CurrentID=0, Status='live', Symbol='CG5873', Aliases='Dmel_ Dmel\\CG5873', description='CG5873 gene product from transcript CG5873-RA', other_designations='CG5873-PA', map_location='90A6-90B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17425768, end_position_on_the_genomic_accession=17446499, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42101, CurrentID=0, Status='live', Symbol='Eh', Aliases='Dmel_CG5400, CG5400, DMEHAB, Dmel\\CG5400, Drm-EH, EH, EHAB, eh', description='Eclosion hormone', other_designations='CG5400-PA|CG5400-PB|Eh-PA|Eh-PB|eclosion hormone', map_location='90B1-90B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17446642, end_position_on_the_genomic_accession=17448858, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42102, CurrentID=0, Status='live', Symbol='CG14330', Aliases='Dmel_ Dmel\\CG14330', description='CG14330 gene product from transcript CG14330-RA', other_designations='CG14330-PA|uncharacterized protein', map_location='90B1-90B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17455890, end_position_on_the_genomic_accession=17456776, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42103, CurrentID=0, Status='live', Symbol='Dscam3', Aliases='Dmel_CG31190, CG12536, CG31190, CG8106, CT33960, Dmel\\CG31190, Dscam-like', description='Down syndrome cell adhesion molecule 3', other_designations='CG31190-PC|CG31190-PD|CG31190-PE|CG31190-PF|Dscam3-PC|Dscam3-PD|Dscam3-PE|Dscam3-PF|down syndrome cell adhesion molecule 3', map_location='90B1-90B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17473935, end_position_on_the_genomic_accession=17505667, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  42105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42105, CurrentID=0, Status='live', Symbol='CG5823', Aliases='Dmel_ Dmel\\CG5823', description='CG5823 gene product from transcript CG5823-RB', other_designations='CG5823-PA|CG5823-PB', map_location='90B3-90B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17532325, end_position_on_the_genomic_accession=17534587, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42106, CurrentID=0, Status='live', Symbol='P5cr-2', Aliases='Dmel_CG5840, CG5840, CT18319, Dmel\\CG5840', description='Pyrroline-5-carboxylate reductase-like 2', other_designations='CG5840-PA|CG5840-PB|CG5840-PC|CG5840-PD|P5cr-2-PA|P5cr-2-PB|P5cr-2-PC|P5cr-2-PD|Pyrroline-5-carboxylate reductase', map_location='90B3-90B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17534815, end_position_on_the_genomic_accession=17537136, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42109, CurrentID=0, Status='live', Symbol='Prx3', Aliases='Dmel_CG5826, CG5826, DPx-5037, DPx5037, Dmel\\CG5826, Prx5037, dPrx3, mTPx, mtTPx, peroxiredoxin 5037', description='Peroxiredoxin 3', other_designations='CG5826-PA|Prx3-PA|mitochondrial peroxiredoxin|peroxiredoxin 3|peroxiredoxin 5037|thioredoxin peroxidase 3|thioredoxin-peroxidase-3', map_location='90B4-90B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17545222, end_position_on_the_genomic_accession=17546531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42110, CurrentID=0, Status='live', Symbol='Hmx', Aliases='Dmel_CG43748, CG18139, CG34419, CG43748, CG5829, CG5832, DHmx, Dmel\\CG43748, Dmel_CG18139, Dmel_CG34419, Dmel_CG5829, Dmel_CG5832', description='H6-like-homeobox', other_designations='CG43748-PC|CG43748-PD|CG43748-PE|Hmx-PC|Hmx-PD|Hmx-PE', map_location='90B5-90B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17558930, end_position_on_the_genomic_accession=17569097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42113, CurrentID=0, Status='live', Symbol='CG7587', Aliases='Dmel_ Dmel\\CG7587', description='CG7587 gene product from transcript CG7587-RA', other_designations='CG7587-PA', map_location='90B5-90B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17594386, end_position_on_the_genomic_accession=17595057, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42114, CurrentID=0, Status='live', Symbol='Sgs5', Aliases='Dmel_CG7596, BcDNA:LP02306, CG7596, Dmel\\CG7596, LP02306, SGS-5, Sgs, Sgs-5, sgs5', description='Salivary gland secretion 5', other_designations='CG7596-PA|Sgs5-PA|salivary gland secretion 5', map_location='90B5-90B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17595362, end_position_on_the_genomic_accession=17596112, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42115, CurrentID=0, Status='live', Symbol='sll', Aliases='Dmel_CG7623, 1422/14, CG7623, Dmel\\CG7623, Sll, anon-WO0118547.555, anon-WO0172774.85, l(3)S142214', description='slalom', other_designations='CG7623-PA|CG7623-PB|sll-PA|sll-PB', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17606106, end_position_on_the_genomic_accession=17608541, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42116, CurrentID=0, Status='live', Symbol='CG14329', Aliases='Dmel_ Dmel\\CG14329', description='CG14329 gene product from transcript CG14329-RA', other_designations='CG14329-PA|uncharacterized protein', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17609359, end_position_on_the_genomic_accession=17610533, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42117, CurrentID=0, Status='live', Symbol='Edg91', Aliases='Dmel_CG7539, CG7539, Dmel\\CG7539, EDG-91, EDG91', description='Ecdysone-dependent gene 91', other_designations='CG7539-PA|CG7539-PB|Edg91-PA|Edg91-PB', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17610856, end_position_on_the_genomic_accession=17611511, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42118, CurrentID=0, Status='live', Symbol='CG14327', Aliases='Dmel_ BcDNA:RE25364, Dmel\\CG14327', description='CG14327 gene product from transcript CG14327-RB', other_designations='CG14327-PA|CG14327-PB|uncharacterized protein', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17613709, end_position_on_the_genomic_accession=17614933, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42119, CurrentID=0, Status='live', Symbol='CG14324', Aliases='Dmel_ BcDNA:RE34075, Dmel\\CG14324', description='CG14324 gene product from transcript CG14324-RA', other_designations='CG14324-PA|uncharacterized protein', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17618281, end_position_on_the_genomic_accession=17618799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42120, CurrentID=0, Status='live', Symbol='CG14326', Aliases='Dmel_ BcDNA:RE19820, Dmel\\CG14326', description='CG14326 gene product from transcript CG14326-RA', other_designations='CG14326-PA|uncharacterized protein', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17619100, end_position_on_the_genomic_accession=17619688, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42121, CurrentID=0, Status='live', Symbol='CG14323', Aliases='Dmel_ Dmel\\CG14323', description='CG14323 gene product from transcript CG14323-RA', other_designations='CG14323-PA|uncharacterized protein', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17621199, end_position_on_the_genomic_accession=17622112, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42122, CurrentID=0, Status='live', Symbol='AttD', Aliases='Dmel_CG7629, Att, Attacin D, CG7629, Dmel\\CG7629, att, att D, attD', description='Attacin-D', other_designations='AttD-PA|AttacinD|CG7629-PA|attacin|attacin D|attackin', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17625268, end_position_on_the_genomic_accession=17626086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42123, CurrentID=0, Status='live', Symbol='CG14325', Aliases='Dmel_ Dmel\\CG14325', description='CG14325 gene product from transcript CG14325-RB', other_designations='CG14325-PB|uncharacterized protein', map_location='90B6-90B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17627127, end_position_on_the_genomic_accession=17629012, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42124, CurrentID=0, Status='live', Symbol='CG14322', Aliases='Dmel_ Dmel\\CG14322', description='CG14322 gene product from transcript CG14322-RE', other_designations='CG14322-PA|CG14322-PB|CG14322-PD|CG14322-PE|uncharacterized protein', map_location='90B7-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17647679, end_position_on_the_genomic_accession=17669985, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42125, CurrentID=0, Status='live', Symbol='CG7523', Aliases='Dmel_ Dmel\\CG7523', description='CG7523 gene product from transcript CG7523-RA', other_designations='CG7523-PA', map_location='90C1-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17669144, end_position_on_the_genomic_accession=17670895, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42126, CurrentID=0, Status='live', Symbol='Mps1', Aliases='Dmel_CG7643, Ald/Mps1, CG7643, Dmel\\CG7643, Msp1, TKK, ald, dMps1, mps1', description='Monopolar spindle 1', other_designations='CG7643-PA|CG7643-PB|Mps1-PA|Mps1-PB|altered disjunction|monopolar spindle 1|monopolar spindles 1', map_location='3-61 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17671271, end_position_on_the_genomic_accession=17673828, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42127, CurrentID=0, Status='live', Symbol='alt', Aliases='Dmel_CG18212, CG18212, Dmel\\CG18212, Q9VEH0_DROME', description='aluminum tubes', other_designations='CG18212-PA|CG18212-PB|CG18212-PD|CG18212-PE|CG18212-PF|CG18212-PG|CG18212-PH|CG18212-PI|alt-PA|alt-PB|alt-PD|alt-PE|alt-PF|alt-PG|alt-PH|alt-PI', map_location='90C1-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17673732, end_position_on_the_genomic_accession=17682515, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42128, CurrentID=0, Status='live', Symbol='CG7655', Aliases='Dmel_ Dmel\\CG7655', description='CG7655 gene product from transcript CG7655-RA', other_designations='CG7655-PA', map_location='90C1-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17682669, end_position_on_the_genomic_accession=17683773, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42129, CurrentID=0, Status='live', Symbol='CG31249', Aliases='Dmel_ CG7477, Dmel\\CG31249, Q9VEG8', description='CG31249 gene product from transcript CG31249-RA', other_designations='CG31249-PA|uncharacterized protein', map_location='90C1-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17685689, end_position_on_the_genomic_accession=17687071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42130, CurrentID=0, Status='live', Symbol='osa', Aliases='Dmel_CG7467, C819, CG7467, Dmel\\CG7467, E(E2F)3C, OSA, Osa, anon-WO0118547.314, anon-WO0172774.126, eld, en(lz)4F/4H, eyelid, l(3)00090, l(3)04539, l(3)j9C3, p300', description='CG7467 gene product from transcript CG7467-RE', other_designations='CG7467-PA|CG7467-PB|CG7467-PC|CG7467-PD|CG7467-PE|CG7467-PF|eye lid|eyelid|osa-PA|osa-PB|osa-PC|osa-PD|osa-PE|osa-PF', map_location='3-60.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17687818, end_position_on_the_genomic_accession=17718350, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  42131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42131, CurrentID=0, Status='live', Symbol='Pxt', Aliases='Dmel_CG7660, CG7660, Dmel\\CG7660, Dpxt, anon-WO0172774.128, pxd, pxt', description='Peroxinectin-like', other_designations='CG7660-PB|Pxt-PB|pxd|pxt', map_location='90C2-90C', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17719454, end_position_on_the_genomic_accession=17722955, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42132, CurrentID=0, Status='live', Symbol='Atg8b', Aliases='Dmel_CG12334, ATG8, ATG8b, Atg8, Atg8B, CG12334, Dmel\\CG12334, atg8B, atg8b', description='Autophagy-related 8b', other_designations='Atg8b-PA|Atg8b-PB|CG12334-PA|CG12334-PB|autophagy related gene 8|autophagy-specific gene 8b', map_location='90C2-90C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17724410, end_position_on_the_genomic_accession=17725063, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42133, CurrentID=0, Status='live', Symbol='Lgr1', Aliases='Dmel_CG7665, CG7665, DLGR-1, DLGR1, Dmel\\CG7665, FSH, FSH-TSH, Fsh, Fsh/CG7665, LGR1, LH/CG receptor, RH44949p, TSH, dLGR1', description='Leucine-rich repeat-containing G protein-coupled receptor 1', other_designations='CG7665-PA|CG7665-PB|Fsh-Tsh-like receptor|Leu-rich repeats-containing G protein-coupled receptor|Lgr1-PA|Lgr1-PB|glycoprotein A2/Glycoprotein B5 receptor', map_location='90C2-90C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17725777, end_position_on_the_genomic_accession=17733207, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  42134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42134, CurrentID=0, Status='live', Symbol='CG14321', Aliases='Dmel_ Dmel\\CG14321', description='CG14321 gene product from transcript CG14321-RA', other_designations='CG14321-PA|uncharacterized protein', map_location='90C2-90C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17733598, end_position_on_the_genomic_accession=17734381, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  140439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=140439, CurrentID=0, Status='live', Symbol='Tre1', Aliases='Dmel_CG3171, BEST:LD12308, CG3171, Dmel\\CG3171, LD12308, Tre, Tre-1, anon-WO0170980.16, anon-WO0170980.17, anon-WO0170981.1, sctt, tre-1, tre1', description='Trapped in endoderm 1', other_designations='CG3171-PA|CG3171-PB|CG3171-PC|G protein coupled receptor|Tre1-PA|Tre1-PB|Tre1-PC|scattershot|trapped in endoderm 1|trapped in endoderm-1', map_location='5A11-5A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5673830, end_position_on_the_genomic_accession=5679977, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42136, CurrentID=0, Status='live', Symbol='TyrR', Aliases='Dmel_CG7431, CG7431, DmCG7431, Dmel\\CG7431, DrmTR, TR, anon-WO0170980.82, anon-WO0170980.83', description='Tyramine receptor', other_designations='CG7431-PA|CG7431-PB|CG7431-PC|TyrR-PA|TyrR-PB|TyrR-PC|tyramine receptor', map_location='90C2-90C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17750129, end_position_on_the_genomic_accession=17763188, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42137, CurrentID=0, Status='live', Symbol='CG43102', Aliases='Dmel_ CG14318, CG7397, Dmel\\CG43102, Dmel_CG14318, Dmel_CG7397, FBgn0038543', description='CG43102 gene product from transcript CG43102-RA', other_designations='CG43102-PA|CG43102-PB|CG43102-PC|CG43102-PD|CG43102-PE|CG43102-PF|CG43102-PG|uncharacterized protein', map_location='90C4-90C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17768272, end_position_on_the_genomic_accession=17794490, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  42139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42139, CurrentID=0, Status='live', Symbol='CG7713', Aliases='Dmel_ Dmel\\CG7713, dmCG7713', description='CG7713 gene product from transcript CG7713-RB', other_designations='CG7713-PA|CG7713-PB', map_location='90C5-90C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17795075, end_position_on_the_genomic_accession=17796317, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42140, CurrentID=0, Status='live', Symbol='CG7379', Aliases='Dmel_ Dmel\\CG7379, ING1', description='CG7379 gene product from transcript CG7379-RA', other_designations='CG7379-PA', map_location='90C5-90C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17796376, end_position_on_the_genomic_accession=17798304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42141, CurrentID=0, Status='live', Symbol='CG17803', Aliases='Dmel_ Dmel\\CG17803', description='CG17803 gene product from transcript CG17803-RB', other_designations='CG17803-PA|CG17803-PB|uncharacterized protein', map_location='90C5-90C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17798877, end_position_on_the_genomic_accession=17801512, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42142, CurrentID=0, Status='live', Symbol='CG17806', Aliases='Dmel_ Dmel\\CG17806', description='CG17806 gene product from transcript CG17806-RA', other_designations='CG17806-PA|uncharacterized protein', map_location='90C5-90C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17801542, end_position_on_the_genomic_accession=17803260, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42143, CurrentID=0, Status='live', Symbol='CG17802', Aliases='Dmel_ Dmel\\CG17802', description='CG17802 gene product from transcript CG17802-RA', other_designations='CG17802-PA|uncharacterized protein', map_location='90C5-90C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17803225, end_position_on_the_genomic_accession=17805249, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42144, CurrentID=0, Status='live', Symbol='CG17801', Aliases='Dmel_ Dmel\\CG17801', description='CG17801 gene product from transcript CG17801-RB', other_designations='CG17801-PB|uncharacterized protein', map_location='90C5-90C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17805679, end_position_on_the_genomic_accession=17807085, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42145, CurrentID=0, Status='live', Symbol='CG7357', Aliases='Dmel_ Dmel\\CG7357', description='CG7357 gene product from transcript CG7357-RA', other_designations='CG7357-PA', map_location='90C6-90C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17807407, end_position_on_the_genomic_accession=17809222, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42146, CurrentID=0, Status='live', Symbol='CG18012', Aliases='Dmel_ Alg1, Dmel\\CG18012, dAlg1', description='CG18012 gene product from transcript CG18012-RA', other_designations='CG18012-PA|uncharacterized protein', map_location='90C6-90C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17809401, end_position_on_the_genomic_accession=17811134, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42148, CurrentID=0, Status='live', Symbol='tinc', Aliases='Dmel_CG31247, CG14319, CG14320, CG31246, CG31247, CG7723, Dmel\\CG31247, Dmel_CG31246', description='tincar', other_designations='CG31247-PA|CG31247-PB|CG31247-PE|CG31247-PF|CG31247-PG|tinc-PA|tinc-PB|tinc-PE|tinc-PF|tinc-PG', map_location='90C7-90C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17811541, end_position_on_the_genomic_accession=17872507, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42150, CurrentID=0, Status='live', Symbol='Rim', Aliases='Dmel_CG33547, CG33547, CG7301, CG7305, CG7321, DmRIM, Dmel\\CG33547, RIM, UNC-10/RIM, dm-Rim', description='Rab3 interacting molecule', other_designations='CG33547-PAA|CG33547-PF|CG33547-PK|CG33547-PP|CG33547-PQ|CG33547-PR|CG33547-PS|CG33547-PT|CG33547-PU|CG33547-PV|CG33547-PW|CG33547-PX|CG33547-PY|CG33547-PZ|Rim-PAA|Rim-PF|Rim-PK|Rim-PP|Rim-PQ|Rim-PR|Rim-PS|Rim-PT|Rim-PU|Rim-PV|Rim-PW|Rim-PX|Rim-PY|Rim-PZ', map_location='90C9-90C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17872748, end_position_on_the_genomic_accession=17916578, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  42152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42152, CurrentID=0, Status='live', Symbol='CG12347', Aliases='Dmel_ Dmel\\CG12347', description='CG12347 gene product from transcript CG12347-RA', other_designations='CG12347-PA|CG12347-PB|uncharacterized protein', map_location='90C10-90C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17906445, end_position_on_the_genomic_accession=17907451, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42158, CurrentID=0, Status='live', Symbol='CG7785', Aliases='Dmel_ CLLD6, Dmel\\CG7785', description='CG7785 gene product from transcript CG7785-RA', other_designations='CG7785-PA', map_location='90E1-90E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18021600, end_position_on_the_genomic_accession=18023689, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42159, CurrentID=0, Status='live', Symbol='CG7794', Aliases='Dmel_ Dmel\\CG7794, anon-WO0140519.191', description='CG7794 gene product from transcript CG7794-RA', other_designations='CG7794-PA', map_location='90E1-90E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18023707, end_position_on_the_genomic_accession=18025378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42160, CurrentID=0, Status='live', Symbol='htl', Aliases='Dmel_CG7223, CG7223, CT22273, CT39172, DFGF-R1, DFGF-R2, DFR-1, DFR1, DFR1/DFGF-R2, DPR3, DTRK(FR1), Dfr-1, Dfr1, DmHD-38, Dmel\\CG7223, Dtk1, EMS2, FGF-R2, FGFR, FGFR2, FR1, Fr1, HD-38, HTL/FGFR1, Htl, Tk1, dtk1, i100, i150, i79, j372', description='heartless', other_designations='CG7223-PA|CG7223-PB|CG7223-PC|FGF receptor|fibroblast growth factor receptor|fibroblast growth factor receptor 1|htl-PA|htl-PB|htl-PC', map_location='90E2-90E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18045216, end_position_on_the_genomic_accession=18052984, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42161, CurrentID=0, Status='live', Symbol='CG14317', Aliases='Dmel_ Dmel\\CG14317, anon-WO0172774.95', description='CG14317 gene product from transcript CG14317-RA', other_designations='CG14317-PA|uncharacterized protein', map_location='90E4-90E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18081048, end_position_on_the_genomic_accession=18082583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42162, CurrentID=0, Status='live', Symbol='sr', Aliases='Dmel_CG7847, CG7847, Dmel\\CG7847, Sr, SrB, l(3)03999, l(3)06948, l(3)06948b', description='stripe', other_designations='CG7847-PA|CG7847-PB|CG7847-PC|CG7847-PD|CG7847-PE|STRIPE|StripeB|sr-PA|sr-PB|sr-PC|sr-PD|sr-PE|striped', map_location='3-62.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18090803, end_position_on_the_genomic_accession=18135583, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42163, CurrentID=0, Status='live', Symbol='CG14316', Aliases='Dmel_ Dmel\\CG14316', description='CG14316 gene product from transcript CG14316-RA', other_designations='CG14316-PA|uncharacterized protein', map_location='90F3-90F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18152120, end_position_on_the_genomic_accession=18153640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42164, CurrentID=0, Status='live', Symbol='CG14315', Aliases='Dmel_ Dmel\\CG14315', description='CG14315 gene product from transcript CG14315-RA', other_designations='CG14315-PA|uncharacterized protein', map_location='90F4-90F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18158701, end_position_on_the_genomic_accession=18159471, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42165, CurrentID=0, Status='live', Symbol='CG7218', Aliases='Dmel_ Dmel\\CG7218', description='CG7218 gene product from transcript CG7218-RB', other_designations='CG7218-PA|CG7218-PB', map_location='90F4-90F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18161392, end_position_on_the_genomic_accession=18164397, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42166, CurrentID=0, Status='live', Symbol='Cbp20', Aliases='Dmel_CG12357, CBP20, CG12357, Dmel\\CG12357, cbp20', description='cap binding protein 20', other_designations='CG12357-PA|Cbp20-PA', map_location='90F4-90F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18164661, end_position_on_the_genomic_accession=18165439, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42169, CurrentID=0, Status='live', Symbol="PP2A-B'", Aliases="Dmel_CG7913, B', B'/PR61, B56-1, BcDNA:GM05554, CG 7913, CG7901, CG7913, Dmel\\CG7913, PP2A, PP2A B', PP2A-B, PP2A[B'-1], Wrd, anon-WO0118547.420, dB56-1, dPP2A, dPP2A-B56-1, i234, pp2A-B', wrd", description='CG7913 gene product from transcript CG7913-RK', other_designations="CG7913-PA|CG7913-PB|CG7913-PC|CG7913-PD|CG7913-PE|CG7913-PI|CG7913-PJ|CG7913-PK|CG7913-PL|CG7913-PM|CG7913-PN|CG7913-PO|CG7913-PP|CG7913-PQ|PP2A B' regulatory subunit|PP2A B' subunit|PP2A-B'-PA|PP2A-B'-PB|PP2A-B'-PC|PP2A-B'-PD|PP2A-B'-PE|PP2A-B'-PI|PP2A-B'-PJ|PP2A-B'-PK|PP2A-B'-PL|PP2A-B'-PM|PP2A-B'-PN|PP2A-B'-PO|PP2A-B'-PP|PP2A-B'-PQ|protein phosphatase 2A|protein phosphatase 2A B'|protein phosphatase 2A regulatory B subunit|protein phosphatase 2A-B'|regulatory B' subunit of PP2A|well rounded|well-rounded", map_location='90F4-90F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18167757, end_position_on_the_genomic_accession=18180049, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  42171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42171, CurrentID=0, Status='live', Symbol='cdm', Aliases='Dmel_CG7212, 3R23, CG 7212, CG7212, Dmel\\CG7212, Imp13, imp13, sd-5', description='cadmus', other_designations='CG7212-PA|CG7212-PB|cdm-PA|cdm-PB|complementation group 5|importin 13|sd-5', map_location='90F5-90F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18179275, end_position_on_the_genomic_accession=18183479, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42172, CurrentID=0, Status='live', Symbol='CG7208', Aliases='Dmel_ CG 7208, Dmel\\CG7208', description='CG7208 gene product from transcript CG7208-RA', other_designations='CG7208-PA|CG7208-PB', map_location='90F5-90F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18183656, end_position_on_the_genomic_accession=18185715, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42173, CurrentID=0, Status='live', Symbol='Arp5', Aliases='Dmel_CG7940, CG7940, Dmel\\CG7940, dArp5', description='Actin-related protein 5', other_designations='Arp5-PA|Arp5-PB|CG7940-PA|CG7940-PB', map_location='90F6-90F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18185498, end_position_on_the_genomic_accession=18187875, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42174, CurrentID=0, Status='live', Symbol='CG12321', Aliases='Dmel_ Dmel\\CG12321', description='CG12321 gene product from transcript CG12321-RA', other_designations='CG12321-PA|uncharacterized protein', map_location='90F6-90F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18187805, end_position_on_the_genomic_accession=18188707, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42175, CurrentID=0, Status='live', Symbol='MED17', Aliases='Dmel_CG7957, CG7957, Dmel\\CG7957, Med17, TRAP80, Trap80, dTRAP80, dTRAP80/SRB4, dTrap80, l(3)s2956, p70', description='Mediator complex subunit 17', other_designations='CG7957-PA|MED17-PA|mediator|mediator complex subunit 17', map_location='90F6-90F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18189072, end_position_on_the_genomic_accession=18191265, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42176, CurrentID=0, Status='live', Symbol='CG14313', Aliases='Dmel_ Dmel\\CG14313', description='CG14313 gene product from transcript CG14313-RA', other_designations='CG14313-PA|uncharacterized protein', map_location='90F6-90F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18191244, end_position_on_the_genomic_accession=18191967, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42177, CurrentID=0, Status='live', Symbol='Ssdp', Aliases='Dmel_CG7187, CG7187, Dmel\\CG7187, l(3)neo48, ssdp', description='Sequence-specific single-stranded DNA-binding protein', other_designations='CG7187-PA|CG7187-PB|CG7187-PC|CG7187-PD|Sequence-specific single-strand DNA-binding protein|Ssdp-PA|Ssdp-PB|Ssdp-PC|Ssdp-PD', map_location='90F6-90F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18192494, end_position_on_the_genomic_accession=18199595, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42178, CurrentID=0, Status='live', Symbol='CG7985', Aliases='Dmel_ BcDNA:GH04120, Dmel\\CG7985', description='CG7985 gene product from transcript CG7985-RB', other_designations='CG7985-PA|CG7985-PB', map_location='90F7-90F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18199857, end_position_on_the_genomic_accession=18215001, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42179, CurrentID=0, Status='live', Symbol='CG14314', Aliases='Dmel_ Dmel\\CG14314', description='CG14314 gene product from transcript CG14314-RA', other_designations='CG14314-PA|uncharacterized protein', map_location='90F7-90F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18216142, end_position_on_the_genomic_accession=18218014, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42180, CurrentID=0, Status='live', Symbol='CG7988', Aliases='Dmel_ Dmel\\CG7988', description='CG7988 gene product from transcript CG7988-RA', other_designations='CG7988-PA', map_location='90F7-90F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18220526, end_position_on_the_genomic_accession=18221473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42181, CurrentID=0, Status='live', Symbol='CG7183', Aliases='Dmel_ Dmel\\CG7183', description='CG7183 gene product from transcript CG7183-RA', other_designations='CG7183-PA|CG7183-PB', map_location='90F7-90F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18218147, end_position_on_the_genomic_accession=18220699, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42182, CurrentID=0, Status='live', Symbol='mTerf5', Aliases='Dmel_CG7175, CG7175, D-MTERF5, Dmel\\CG7175', description='CG7175 gene product from transcript CG7175-RA', other_designations='CG7175-PA|mTerf5-PA', map_location='90F7-90F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18221522, end_position_on_the_genomic_accession=18223682, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42183, CurrentID=0, Status='live', Symbol='Non3', Aliases='Dmel_CG7993, CG7993, Dmel\\CG7993', description='Novel nucleolar protein 3', other_designations='CG7993-PA|Non3-PA|novel nucleolar protein 3', map_location='90F8-90F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18223722, end_position_on_the_genomic_accession=18224942, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42184, CurrentID=0, Status='live', Symbol='CG7168', Aliases='Dmel_ Dmel\\CG7168', description='CG7168 gene product from transcript CG7168-RA', other_designations='CG7168-PA', map_location='90F8-90F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18224902, end_position_on_the_genomic_accession=18226082, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42185, CurrentID=0, Status='live', Symbol='Mdh2', Aliases='Dmel_CG7998, CG 7998, CG7998, Dmel\\CG7998, MDH-2, Mdh, l(3)psg7, mMdh, mdh2, psg7', description='Malate dehydrogenase 2', other_designations='CG7998-PA|L-Malate dehydrogenase|Mdh2-PA|lethal (3) persistent salivary gland 7|malate dehydrogenase|malate dehydrogenase 2|malic DH|mitochondrial malate dehydrogenase|mitochondrial-Mdh', map_location='3-62.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18226396, end_position_on_the_genomic_accession=18228718, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42186, CurrentID=0, Status='live', Symbol='14-3-3epsilon', Aliases='Dmel_CG31196, 14-3-3, 14-3-3 epsilon, 14-3-3-e, 14-3-3-epsilon, 14-3-30epsilon, 14-3-3EPSILON, 14-3-3e, 14-3-3omicron, 14.3.3 epsilon, 1433epsilon, 143E_DROME, CG31196, CT24092, D14-3-3e, D Dmel\\CG31196, EK3-5, PAR-5, PAR5, Par-5, SR3-9, Su(Raf)3B, anon-WO0172774.141, anon-WO02059370.52, d14-3-3epsilon, e-14-3-3, eps, l(3)j2B10, par-5', description='CG31196 gene product from transcript CG31196-RD', other_designations='14-3-3&epsilon|14-3-3epsilon-PA|14-3-3epsilon-PB|14-3-3epsilon-PC|14-3-3epsilon-PD|CG31196-PA|CG31196-PB|CG31196-PC|CG31196-PD|Tyrosine-3-monooxygenase|epsilon|suppressor of Ras85D 3-9', map_location='90F10-90F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18242531, end_position_on_the_genomic_accession=18250550, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42187, CurrentID=0, Status='live', Symbol='CG7156', Aliases='Dmel_ Dmel\\CG7156', description='CG7156 gene product from transcript CG7156-RA', other_designations='CG7156-PA', map_location='90F10-90F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18239420, end_position_on_the_genomic_accession=18242159, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42188, CurrentID=0, Status='live', Symbol='CG18598', Aliases='Dmel_ CT42539, Dmel\\CG18598', description='CG18598 gene product from transcript CG18598-RA', other_designations='CG18598-PA|uncharacterized protein', map_location='90F10-90F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18251115, end_position_on_the_genomic_accession=18251539, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42189, CurrentID=0, Status='live', Symbol='CG12320', Aliases='Dmel_ Aar2, Dmel\\CG12320', description='CG12320 gene product from transcript CG12320-RB', other_designations='CG12320-PA|CG12320-PB|uncharacterized protein', map_location='90F10-90F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18251069, end_position_on_the_genomic_accession=18252734, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42191, CurrentID=0, Status='live', Symbol='CG18599', Aliases='Dmel_ Dmel\\CG18599', description='CG18599 gene product from transcript CG18599-RA', other_designations='CG18599-PA|uncharacterized protein', map_location='90F11-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18267789, end_position_on_the_genomic_accession=18271172, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42192, CurrentID=0, Status='live', Symbol='Vps39', Aliases='Dmel_CG7146, CG7146, Dmel\\CG7146, dVps39, vps39', description='Vacuolar protein sorting 39', other_designations='CG7146-PA|Vps39-PA|vacuolar protein sorting 39', map_location='91A1-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18271189, end_position_on_the_genomic_accession=18274446, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42194, CurrentID=0, Status='live', Symbol='CG7142', Aliases='Dmel_ Dmel\\CG7142, SP132', description='CG7142 gene product from transcript CG7142-RA', other_designations='CG7142-PA', map_location='91A1-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18278181, end_position_on_the_genomic_accession=18279522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42195, CurrentID=0, Status='live', Symbol='CG14312', Aliases='Dmel_ Dmel\\CG14312', description='CG14312 gene product from transcript CG14312-RA', other_designations='CG14312-PA|uncharacterized protein', map_location='91A1-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18282978, end_position_on_the_genomic_accession=18283528, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42196, CurrentID=0, Status='live', Symbol='CG8064', Aliases='Dmel_ Dmel\\CG8064', description='CG8064 gene product from transcript CG8064-RA', other_designations='CG8064-PA', map_location='91A1-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18283660, end_position_on_the_genomic_accession=18286624, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42197, CurrentID=0, Status='live', Symbol='CG7131', Aliases='Dmel_ Dmel\\CG7131, anon-WO0140519.212', description='CG7131 gene product from transcript CG7131-RA', other_designations='CG7131-PA', map_location='91A1-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18286620, end_position_on_the_genomic_accession=18288777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42199, CurrentID=0, Status='live', Symbol='Dlc90F', Aliases='Dmel_CG12363, CG12363, Dmel\\CG12363, TcTex-1, Tctex, Tctex-1, Tctex1, bs35g12.y1, dlc90F, dtctex-1, l(3)04091, tctex, tctex-1', description='Dynein light chain 90F', other_designations='CG12363-PA|Dlc90F-PA|dynein light chain 90F|dynein light chain; tctex-1', map_location='91A2-91A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18299229, end_position_on_the_genomic_accession=18300044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42201, CurrentID=0, Status='live', Symbol='CG18600', Aliases='Dmel_ BEST:CK01205, CK01205, Dmel\\CG18600', description='CG18600 gene product from transcript CG18600-RA', other_designations='CG18600-PA|uncharacterized protein', map_location='91A2-91A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18300195, end_position_on_the_genomic_accession=18301728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42202, CurrentID=0, Status='live', Symbol='CG7126', Aliases='Dmel_ Dmel\\CG7126', description='CG7126 gene product from transcript CG7126-RA', other_designations='CG7126-PA', map_location='91A2-91A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18302170, end_position_on_the_genomic_accession=18302849, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42203, CurrentID=0, Status='live', Symbol='PKD', Aliases='Dmel_CG7125, 150131_at, CG7125, Dmel\\CG7125, PKCm, dPKD', description='Protein Kinase D', other_designations='CG7125-PA|CG7125-PB|CG7125-PC|CG7125-PD|CG7125-PE|CG7125-PF|CG7125-PG|CG7125-PH|CG7125-PI|PKD-PA|PKD-PB|PKD-PC|PKD-PD|PKD-PE|PKD-PF|PKD-PG|PKD-PH|PKD-PI|protein kinase D', map_location='91A2-91A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18303279, end_position_on_the_genomic_accession=18314407, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42204, CurrentID=0, Status='live', Symbol='CG31235', Aliases='Dmel_ CG14311, CG7117, Dmel\\CG31235', description='CG31235 gene product from transcript CG31235-RB', other_designations='CG31235-PA|CG31235-PB|uncharacterized protein', map_location='91A2-91A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18316075, end_position_on_the_genomic_accession=18328942, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42206, CurrentID=0, Status='live', Symbol='CG15803', Aliases='Dmel_ Dmel\\CG15803, lincRNA.804', description='CG15803 gene product from transcript CG15803-RB', other_designations='CG15803-PB|CG15803-PC|uncharacterized protein', map_location='91A3-91A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18330932, end_position_on_the_genomic_accession=18360728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42207, CurrentID=0, Status='live', Symbol='hmw', Aliases='Dmel_CG7669, CG7669, Dmel\\CG7669, anon-WO0140519.198', description='hemingway', other_designations='CG7669-PA|Hmw|hmw-PA', map_location='91A3-91A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18361994, end_position_on_the_genomic_accession=18364277, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42208, CurrentID=0, Status='live', Symbol='WRNexo', Aliases='Dmel_CG7670, CG7670, DmWRNexo, Dmel\\CG7670', description='WRN exonuclease', other_designations='CG7670-PA|CG7670-PB|CG7670-PC|WRNexo-PA|WRNexo-PB|WRNexo-PC', map_location='91A3-91A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18364244, end_position_on_the_genomic_accession=18366162, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42209, CurrentID=0, Status='live', Symbol='Nup43', Aliases='Dmel_CG7671, CG7671, Dmel\\CG7671, nup43', description='Nucleoporin 43kD', other_designations='CG7671-PA|Nup43-PA|nucleoporin 43|nucleoporin 43kD', map_location='91A3-91A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18366638, end_position_on_the_genomic_accession=18367913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42210, CurrentID=0, Status='live', Symbol='gl', Aliases='Dmel_CG7672, CG7672, Dmel\\CG7672, E(sina)5, Gla, Glass, SS3-2, SY3-3a, none, rh, gl', description='glass', other_designations='CG7672-PA|CG7672-PB|CG7672-PC|GLASS|enhancer of seven in absentia 5|gl-PA|gl-PB|gl-PC|no-ocelli--narrow-eyes|rauhig|suppressor of GMR-sina 3-2', map_location='3-63.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18367927, end_position_on_the_genomic_accession=18372738, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42211, CurrentID=0, Status='live', Symbol='CG7675', Aliases='Dmel_ Dmel\\CG7675', description='CG7675 gene product from transcript CG7675-RC', other_designations='CG7675-PA|CG7675-PB|CG7675-PC|CG7675-PD|CG7675-PE', map_location='91A4-91A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18378124, end_position_on_the_genomic_accession=18381178, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42212, CurrentID=0, Status='live', Symbol='CG14309', Aliases='Dmel_ Dmel\\CG14309', description='CG14309 gene product from transcript CG14309-RA', other_designations='CG14309-PA|uncharacterized protein', map_location='91A5-91A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18381580, end_position_on_the_genomic_accession=18385072, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42213, CurrentID=0, Status='live', Symbol='cona', Aliases='Dmel_CG7676, CG7676, Dmel\\CG7676', description='corona', other_designations='CG7676-PA|cona-PA', map_location='91A5-91A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18385828, end_position_on_the_genomic_accession=18386887, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42214, CurrentID=0, Status='live', Symbol='Vha100-4', Aliases='Dmel_CG7678, CG7678, Dmel\\CG7678, vha100-4', description='Vacuolar H[+] ATPase 100kD subunit 4', other_designations='CG7678-PA|Vha100-4-PA|vacuolar H[+] ATPase 100kD subunit 4|vacuolar H[+] ATPase subunit 100-4', map_location='91A5-91A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18386841, end_position_on_the_genomic_accession=18389805, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42215, CurrentID=0, Status='live', Symbol='Vti1b', Aliases='Dmel_CG44009, CG31232, CG44009, CG7682, Dmel\\CG44009, EP(3)1037', description='Vesicle transport through interaction with t-SNAREs 1b', other_designations='CG44009-PA|CG44009-PB|Vti1b-PA|Vti1b-PB|vesicle transport through interaction with t-SNAREs 1b', map_location='91A5-91A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18396995, end_position_on_the_genomic_accession=18400339, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42216, CurrentID=0, Status='live', Symbol='Vha100-2', Aliases='Dmel_CG18617, BcDNA:LD21735, CG18617, CG7679, Dmel\\CG18617, anon-WO0118547.296, vha100-2', description='Vacuolar H[+] ATPase 100kD subunit 2', other_designations='CG18617-PA|CG18617-PB|Vha100-2-PA|Vha100-2-PB|vacuolar H + ATPase subunit 100-2|vacuolar H[+] ATPase 100kD subunit 2|vacuolar H[+] ATPase subunit 100-2', map_location='91A5-91A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18390323, end_position_on_the_genomic_accession=18399169, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  39804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39804, CurrentID=0, Status='live', Symbol='CG13040', Aliases='Dmel_ Dmel\\CG13040', description='CG13040 gene product from transcript CG13040-RB', other_designations='CG13040-PA|CG13040-PB|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16325083, end_position_on_the_genomic_accession=16326757, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42218, CurrentID=0, Status='live', Symbol='CG12333', Aliases='Dmel_ Dmel\\CG12333', description='CG12333 gene product from transcript CG12333-RA', other_designations='CG12333-PA|uncharacterized protein', map_location='91A5-91A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18400373, end_position_on_the_genomic_accession=18402716, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42220, CurrentID=0, Status='live', Symbol='CG7685', Aliases='Dmel_ Dmel\\CG7685', description='CG7685 gene product from transcript CG7685-RA', other_designations='CG7685-PA|CG7685-PB', map_location='91A6-91A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18404940, end_position_on_the_genomic_accession=18406002, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42221, CurrentID=0, Status='live', Symbol='CG31122', Aliases='Dmel_ CG14308, CG7687, Dmel\\CG31122', description='CG31122 gene product from transcript CG31122-RC', other_designations='CG31122-PA|CG31122-PB|CG31122-PC|uncharacterized protein', map_location='91A6-91A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18406064, end_position_on_the_genomic_accession=18412874, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42222, CurrentID=0, Status='live', Symbol='CG10864', Aliases='Dmel_ Dmel\\CG10864', description='CG10864 gene product from transcript CG10864-RA', other_designations='CG10864-PA|uncharacterized protein', map_location='91A6-91A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18410080, end_position_on_the_genomic_accession=18411631, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42226, CurrentID=0, Status='live', Symbol='fru', Aliases='Dmel_CG14307, BTB-VI, BtbVI, CG14307, CG14307/CG33541, CG7688, CG7689, CG7690, CT22773, Dmel\\CG14307, Dmfru, FRU, Fru, cg7688, cg7689C, fty, ms(3)06411, sat, fru', description='fruitless', other_designations='BTB-protein-VI|CG14307-PA|CG14307-PB|CG14307-PC|CG14307-PD|CG14307-PE|CG14307-PF|CG14307-PG|CG14307-PH|CG14307-PI|CG14307-PJ|CG14307-PK|CG14307-PL|CG14307-PM|CG14307-PN|CG14307-PO|fru-PA|fru-PB|fru-PC|fru-PD|fru-PE|fru-PF|fru-PG|fru-PH|fru-PI|fru-PJ|fru-PK|fru-PL|fru-PM|fru-PN|fru-PO|fru-satori|fruity|satori', map_location='91A6-91B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18414273, end_position_on_the_genomic_accession=18545586, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  42228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42228, CurrentID=0, Status='live', Symbol='CG7691', Aliases='Dmel_ Dmel\\CG7691', description='CG7691 gene product from transcript CG7691-RA', other_designations='CG7691-PA', map_location='91B4-91B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18559869, end_position_on_the_genomic_accession=18561353, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39806, CurrentID=0, Status='live', Symbol='CG13058', Aliases='Dmel_ Dmel\\CG13058', description='CG13058 gene product from transcript CG13058-RA', other_designations='CG13058-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16328841, end_position_on_the_genomic_accession=16329534, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42230, CurrentID=0, Status='live', Symbol='CG7694', Aliases='Dmel_ BcDNA:LP11469, Dmel\\CG7694', description='CG7694 gene product from transcript CG7694-RA', other_designations='CG7694-PA|CG7694-PB|CG7694-PC', map_location='91B4-91B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18573091, end_position_on_the_genomic_accession=18590726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42232, CurrentID=0, Status='live', Symbol='CG14304', Aliases='Dmel_ Dmel\\CG14304', description='CG14304 gene product from transcript CG14304-RA', other_designations='CG14304-PA|CG14304-PB|CG14304-PC|uncharacterized protein', map_location='91B5-91B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18592801, end_position_on_the_genomic_accession=18621327, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42233, CurrentID=0, Status='live', Symbol='CG14305', Aliases='Dmel_ Dmel\\CG14305', description='CG14305 gene product from transcript CG14305-RC', other_designations='CG14305-PA|CG14305-PC|CG14305-PD|uncharacterized protein', map_location='91B5-91B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18598691, end_position_on_the_genomic_accession=18600097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42234, CurrentID=0, Status='live', Symbol='CG7695', Aliases='Dmel_ Dmel\\CG7695', description='CG7695 gene product from transcript CG7695-RA', other_designations='CG7695-PA', map_location='91B6-91B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18616754, end_position_on_the_genomic_accession=18617511, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42235, CurrentID=0, Status='live', Symbol='CG14301', Aliases='Dmel_ Dmel\\CG14301', description='CG14301 gene product from transcript CG14301-RB', other_designations='CG14301-PB|uncharacterized protein', map_location='91B7-91B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18629151, end_position_on_the_genomic_accession=18641457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42236, CurrentID=0, Status='live', Symbol='qin', Aliases='Dmel_CG43726, CG14303, CG14306, CG43726, Dmel\\CG43726, Dmel_CG14303, Dmel_CG14306, kumo', description='CG43726 gene product from transcript CG43726-RB', other_designations='CG43726-PB|qin-PB', map_location='91B7-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18591027, end_position_on_the_genomic_accession=18650137, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42237, CurrentID=0, Status='live', Symbol='CG31224', Aliases='Dmel_ CG17796, CG7696, Dmel\\CG31224, l(3)c00083', description='CG31224 gene product from transcript CG31224-RB', other_designations='CG31224-PA|CG31224-PB|CG31224-PC|lethal (3) c00083|uncharacterized protein', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18650121, end_position_on_the_genomic_accession=18659309, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42239, CurrentID=0, Status='live', Symbol='CstF-64', Aliases='Dmel_CG7697, CG7697, CStF 64, CstF 64, CstF64, Dmel\\CG7697, dCstF64', description='Cleavage stimulation factor 64 kilodalton subunit', other_designations='CG7697-PA|CstF-64-PA|cleavage stimulation factor|cleavage stimulation factor 64 kilodalton subunit', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18659618, end_position_on_the_genomic_accession=18661126, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42240, CurrentID=0, Status='live', Symbol='Cpsf73', Aliases='Dmel_CG7698, CG7698, CPSF-73, CPSF73, Dmel\\CG7698, cpsf-73, dCPSF-73', description='Cleavage and polyadenylation specificity factor 73', other_designations='CG7698-PA|Cpsf73-PA|cleavage and polyadenylation specificity factor 73|cleavage and polyadenylation specificity factor, 73kD-subunit', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18661121, end_position_on_the_genomic_accession=18663552, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42242, CurrentID=0, Status='live', Symbol='CG7702', Aliases='Dmel_ 7702, CT23351, Dmel\\CG7702', description='CG7702 gene product from transcript CG7702-RA', other_designations='CG7702-PA|CG7702-PB|connectin-like', map_location='91B8-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18669390, end_position_on_the_genomic_accession=18674679, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42243, CurrentID=0, Status='live', Symbol='CG7705', Aliases='Dmel_ Dmel\\CG7705', description='CG7705 gene product from transcript CG7705-RA', other_designations='CG7705-PA|CG7705-PB', map_location='91C1-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18674626, end_position_on_the_genomic_accession=18677539, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42244, CurrentID=0, Status='live', Symbol='CG7706', Aliases='Dmel_ Dmel\\CG7706', description='CG7706 gene product from transcript CG7706-RA', other_designations='CG7706-PA', map_location='91C1-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18677818, end_position_on_the_genomic_accession=18680321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42245, CurrentID=0, Status='live', Symbol='CG7708', Aliases='Dmel_ Dmel\\CG7708, dSLC5A15', description='CG7708 gene product from transcript CG7708-RB', other_designations='CG7708-PA|CG7708-PB', map_location='91C1-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18680604, end_position_on_the_genomic_accession=18686739, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42246, CurrentID=0, Status='live', Symbol='Muc91C', Aliases='Dmel_CG7709, CG7709, Dmel\\CG7709', description='Mucin 91C', other_designations='CG7709-PA|CG7709-PB|Muc91C-PA|Muc91C-PB|mucin 91C', map_location='91C1-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18687556, end_position_on_the_genomic_accession=18692719, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42248, CurrentID=0, Status='live', Symbol='CG34282', Aliases='Dmel_ CG7710, Dmel\\CG34282', description='CG34282 gene product from transcript CG34282-RB', other_designations='CG34282-PA|CG34282-PB|uncharacterized protein', map_location='91C1-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18695252, end_position_on_the_genomic_accession=18695995, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42249, CurrentID=0, Status='live', Symbol='ChAT', Aliases='Dmel_CG12345, CG12345, CHAT, CLAT_DROME, CT23399, Cat, ChAt, Cha, ChaT, DChAT, Dmel\\CG12345, cha, dChAT, l(3)91Cc', description='Choline acetyltransferase', other_designations='CG12345-PA|CG12345-PB|ChAT-PA|ChAT-PB|Choline-acetyltransferase|cholinacetyltransferase|choline acetyl transferase|choline acetyltransferase|choline-O-acetyltransferase', map_location='3-64.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18705501, end_position_on_the_genomic_accession=18732344, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42250, CurrentID=0, Status='live', Symbol='CG7714', Aliases='Dmel_ Dmel\\CG7714, anon-EST:ParkEST066', description='CG7714 gene product from transcript CG7714-RA', other_designations='CG7714-PA', map_location='91C4-91C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18725399, end_position_on_the_genomic_accession=18726434, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42251, CurrentID=0, Status='live', Symbol='CG7715', Aliases='Dmel_ Dmel\\CG7715', description='CG7715 gene product from transcript CG7715-RA', other_designations='CG7715-PA', map_location='91C4-91C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18727573, end_position_on_the_genomic_accession=18728468, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42252, CurrentID=0, Status='live', Symbol='CG14302', Aliases='Dmel_ BcDNA:LP02994, Dmel\\CG14302', description='CG14302 gene product from transcript CG14302-RA', other_designations='CG14302-PA|uncharacterized protein', map_location='91C5-91C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18732681, end_position_on_the_genomic_accession=18733034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42253, CurrentID=0, Status='live', Symbol='Mekk1', Aliases='Dmel_CG7717, CG7717, D-MEKK1, D-Mekk1, DmMEKK1, Dmel\\CG7717, MEKK, MEKK1, MEKK4/4, dMEKK, dMEKK1, dMEKK4, dMekk1, mekk1, mekk1a, mekk1b', description='CG7717 gene product from transcript CG7717-RB', other_designations='CG7717-PB|CG7717-PD|Mekk1-PB|Mekk1-PD', map_location='91C5-91C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18733106, end_position_on_the_genomic_accession=18743594, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42254, CurrentID=0, Status='live', Symbol='CG7718', Aliases='Dmel_ Dmel\\CG7718, PL D', description='CG7718 gene product from transcript CG7718-RA', other_designations='CG7718-PA', map_location='91C5-91C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18744063, end_position_on_the_genomic_accession=18746041, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42256, CurrentID=0, Status='live', Symbol='CG14299', Aliases='Dmel_ Dmel\\CG14299', description='CG14299 gene product from transcript CG14299-RD', other_designations='CG14299-PC|CG14299-PD|uncharacterized protein', map_location='91C6-91C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18750436, end_position_on_the_genomic_accession=18758873, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  42257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42257, CurrentID=0, Status='live', Symbol='CG7720', Aliases='Dmel_ BcDNA:RH05973, Dmel\\CG7720, ORE-8, dSLC5A1', description='CG7720 gene product from transcript CG7720-RF', other_designations='CG7720-PA|CG7720-PB|CG7720-PC|CG7720-PD|CG7720-PE|CG7720-PF', map_location='91C6-91D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18759236, end_position_on_the_genomic_accession=18795169, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  42258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42258, CurrentID=0, Status='live', Symbol='CG18208', Aliases='Dmel_ Dmel\\CG18208', description='CG18208 gene product from transcript CG18208-RC', other_designations='CG18208-PA|CG18208-PB|CG18208-PC|uncharacterized protein', map_location='91D1-91D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18799280, end_position_on_the_genomic_accession=18849224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42259, CurrentID=0, Status='live', Symbol='CG14298', Aliases='Dmel_ Dmel\\CG14298', description='CG14298 gene product from transcript CG14298-RB', other_designations='CG14298-PA|CG14298-PB|uncharacterized protein', map_location='91D3-91D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18875528, end_position_on_the_genomic_accession=18877204, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42260, CurrentID=0, Status='live', Symbol='CG14297', Aliases='Dmel_ Dmel\\CG14297, LMW-PTP', description='CG14297 gene product from transcript CG14297-RA', other_designations='CG14297-PA|uncharacterized protein', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18898121, end_position_on_the_genomic_accession=18899342, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42261, CurrentID=0, Status='live', Symbol='Rh2', Aliases='Dmel_CG16740, CG16740, DMELRH2, Dm Rh2, Dmel\\CG16740, RH2, Rh, rh2', description='Rhodopsin 2', other_designations='CG16740-PA|Rh2-PA|rhodopsin|rhodopsin 2', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18899423, end_position_on_the_genomic_accession=18901019, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42262, CurrentID=0, Status='live', Symbol='CG14294', Aliases='Dmel_ BcDNA:AT18358, Dmel\\CG14294', description='CG14294 gene product from transcript CG14294-RA', other_designations='CG14294-PA|uncharacterized protein', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18901663, end_position_on_the_genomic_accession=18902264, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42264, CurrentID=0, Status='live', Symbol='CG14292', Aliases='Dmel_ Dmel\\CG14292', description='CG14292 gene product from transcript CG14292-RA', other_designations='CG14292-PA|uncharacterized protein', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18904933, end_position_on_the_genomic_accession=18905655, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42265, CurrentID=0, Status='live', Symbol='EndoA', Aliases='Dmel_CG14296, CG14296, D-endo, D-endoA, Dendo-A, Dmel\\CG14296, Endo, endo, endoA', description='Endophilin A', other_designations='CG14296-PA|CG14296-PB|CG14296-PC|EndoA-PA|EndoA-PB|EndoA-PC|Endophilin-A|endophilin|endophilin A|endophilinA', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18906609, end_position_on_the_genomic_accession=18910755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42266, CurrentID=0, Status='live', Symbol='CG14291', Aliases='Dmel_ Dmel\\CG14291', description='CG14291 gene product from transcript CG14291-RA', other_designations='CG14291-PA|uncharacterized protein', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18911053, end_position_on_the_genomic_accession=18912922, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42267, CurrentID=0, Status='live', Symbol='Xrp1', Aliases='Dmel_CG17836, 2515, CG17836, DM28, Dmel\\CG17836, XRP1, anon-EST:Liang-2.44, clone 2.44, l(3)02515, xrp1', description='CG17836 gene product from transcript CG17836-RB', other_designations='CG17836-PA|CG17836-PB|CG17836-PC|CG17836-PD|CG17836-PE|CG17836-PF|CG17836-PG|Xrp1-PA|Xrp1-PB|Xrp1-PC|Xrp1-PD|Xrp1-PE|Xrp1-PF|Xrp1-PG|lethal (3) 02515', map_location='91D3-91D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18914968, end_position_on_the_genomic_accession=18926989, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42268, CurrentID=0, Status='live', Symbol='Mpc1', Aliases='Dmel_CG14290, CG14290, Dmel\\CG14290, dMPC1', description='Mitochondrial pyruvate carrier', other_designations='CG14290-PB|CG14290-PC|Mpc1-PB|Mpc1-PC|mitochondrial pyruvate carrier', map_location='91D5-91D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18926996, end_position_on_the_genomic_accession=18928659, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42269, CurrentID=0, Status='live', Symbol='CG42613', Aliases='Dmel_ CG14289, CG31149, CG31222, CG5449, CT17110, CT33918, Dmel\\CG42613, Dmel_CG31149, Dmel_CG31222, tld-like', description='CG42613 gene product from transcript CG42613-RE', other_designations='CG42613-PE|CG42613-PF|CG42613-PG|CG42613-PH|uncharacterized protein', map_location='91E1-91E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18929196, end_position_on_the_genomic_accession=18978088, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  42271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42271, CurrentID=0, Status='live', Symbol='CG12269', Aliases='Dmel_ Dmel\\CG12269', description='CG12269 gene product from transcript CG12269-RA', other_designations='CG12269-PA|CG12269-PB|uncharacterized protein', map_location='91E1-91E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18978273, end_position_on_the_genomic_accession=18978809, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42272, CurrentID=0, Status='live', Symbol='Smu1', Aliases='Dmel_CG5451, CG5451, Dmel\\CG5451, dsmu1', description='SMU1 ortholog', other_designations='CG5451-PA|CG5451-PB|CG5451-PC|Smu1-PA|Smu1-PB|Smu1-PC', map_location='91E1-91E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18978980, end_position_on_the_genomic_accession=18981324, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42273, CurrentID=0, Status='live', Symbol='dnk', Aliases='Dmel_CG5452, CG5452, DM-dNK, DNK, Dm-dNK, DmdNK, Dmel\\CG5452, dNK', description='deoxyribonucleoside kinase', other_designations="2'-deoxynucleoside kinase|CG5452-PA|CG5452-PB|deoxynucleoside kinase|deoxyribonucleoside|dnk-PA|dnk-PB|kinase", map_location='91E2-91E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18980995, end_position_on_the_genomic_accession=18982749, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42274, CurrentID=0, Status='live', Symbol='snRNP-U1-C', Aliases='Dmel_CG5454, BcDNA:RH29440, C, CG5454, Dmel\\CG5454, U1C, dU1-C, snRNP-U1, snRNPC', description='small ribonucleoprotein particle U1 subunit C', other_designations='CG5454-PA|snRNP-U1|snRNP-U1-C-PA|snRNPC', map_location='91E2-91E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18982964, end_position_on_the_genomic_accession=18983650, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42279, CurrentID=0, Status='live', Symbol='CG6005', Aliases='Dmel_ Dmel\\CG6005', description='CG6005 gene product from transcript CG6005-RB', other_designations='CG6005-PA|CG6005-PB', map_location='91E3-91E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19022458, end_position_on_the_genomic_accession=19024030, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42280, CurrentID=0, Status='live', Symbol='CG14286', Aliases='Dmel_ 14286, Dmel\\CG14286', description='CG14286 gene product from transcript CG14286-RA', other_designations='CG14286-PA|uncharacterized protein', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19024092, end_position_on_the_genomic_accession=19024853, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42281, CurrentID=0, Status='live', Symbol='CG14285', Aliases='Dmel_ Dmel\\CG14285', description='CG14285 gene product from transcript CG14285-RA', other_designations='CG14285-PA|uncharacterized protein', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19025099, end_position_on_the_genomic_accession=19026203, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42282, CurrentID=0, Status='live', Symbol='NP15.6', Aliases='Dmel_CG6008, CG6008, Dmel\\CG6008, anon-WO0118547.281', description='CG6008 gene product from transcript CG6008-RA', other_designations='CG6008-PA|NP15.6-PA', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19027368, end_position_on_the_genomic_accession=19028015, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42283, CurrentID=0, Status='live', Symbol='GatA', Aliases='Dmel_CG6007, 0929/02, CG6007, Dmel\\CG6007, bene, benedict, gatA, l(3)S092902', description='Glutamyl-tRNA(Gln) amidotransferase, subunit A', other_designations='CG6007-PA|GatA-PA|benedict|gatA|glutamyl-tRNA amidotransferase A|glutamyl-tRNA amidotransferase subunit A|lethal (3) S092902', map_location='91E4-91F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19026174, end_position_on_the_genomic_accession=19028857, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42284, CurrentID=0, Status='live', Symbol='P5cr', Aliases='Dmel_CG6009, CG6009, Dmel\\CG6009, P5CR', description='Pyrroline 5-carboyxlate reductase', other_designations='CG6009-PA|P5cr-PA|pyrroline 5-carboxylate|pyrroline 5-carboyxlate reductase|reductase', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19028931, end_position_on_the_genomic_accession=19030159, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42285, CurrentID=0, Status='live', Symbol='Prp18', Aliases='Dmel_CG6011, CG6011, Dmel\\CG6011, prp18', description='CG6011 gene product from transcript CG6011-RA', other_designations='CG6011-PA|Prp18-PA', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19029968, end_position_on_the_genomic_accession=19031271, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42286, CurrentID=0, Status='live', Symbol='CG6013', Aliases='Dmel_ Dmel\\CG6013', description='CG6013 gene product from transcript CG6013-RA', other_designations='CG6013-PA', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19031503, end_position_on_the_genomic_accession=19032409, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42287, CurrentID=0, Status='live', Symbol='CG6026', Aliases='Dmel_ Dmel\\CG6026', description='CG6026 gene product from transcript CG6026-RB', other_designations='CG6026-PB', map_location='91E4-91E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19033538, end_position_on_the_genomic_accession=19044693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42289, CurrentID=0, Status='live', Symbol='cdi', Aliases='Dmel_CG6027, 91F, B9.3, CDI/TSKI, CG6027, Cdi, Cdi/TESK1, Dmel\\CG6027/TESK1, l(3)07013, cdi', description='center divider', other_designations='91F gene|CG6027-PA|CG6027-PB|cdi-PA|cdi-PB|central divider|centre divider|centrer divider', map_location='91E4-91F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19044882, end_position_on_the_genomic_accession=19094270, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42290, CurrentID=0, Status='live', Symbol='mRpL55', Aliases='Dmel_CG14283, BcDNA:RH10246, CG14283, Dmel\\CG14283, MRP-L55, tmb', description='mitochondrial ribosomal protein L55', other_designations='CG14283-PA|mRpL55-PA|tom thumb', map_location='91F1-91F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19094394, end_position_on_the_genomic_accession=19094905, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42291, CurrentID=0, Status='live', Symbol='ATPsynD', Aliases='Dmel_CG6030, ATP syn-d, ATPQ, ATPsyn-D, ATPsyn-d, ATPsyn-delta, CG6030, Dmel\\CG6030', description='ATP synthase, subunit D', other_designations='ATP synthase D chain|ATP synthase chain D|ATP synthase subunits d|ATP synthase, subunit d|ATP synthase,d|ATPase synthase, subunit d|ATPsynD-PA|ATPsynD-PB|ATPsynD-PC|CG6030-PA|CG6030-PB|CG6030-PC|F0/F1 ATP synthase D', map_location='91F1-91F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19095000, end_position_on_the_genomic_accession=19096271, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42292, CurrentID=0, Status='live', Symbol='CG6040', Aliases='Dmel_ Dmel\\CG6040', description='CG6040 gene product from transcript CG6040-RA', other_designations='CG6040-PA', map_location='91F2-91F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19096961, end_position_on_the_genomic_accession=19131929, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42293, CurrentID=0, Status='live', Symbol='Cyp12a5', Aliases='Dmel_CG11821, 12a5, CG11821, Dmel\\CG11821', description='CG11821 gene product from transcript CG11821-RA', other_designations='CG11821-PA|Cyp12a5-PA|cytochrome P450', map_location='91F2-91F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19132648, end_position_on_the_genomic_accession=19134704, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42294, CurrentID=0, Status='live', Symbol='Cyp12a4', Aliases='Dmel_CG6042, 12a4, CG6042, CYP12A4, Dmel\\CG6042, P450, cyp12a4', description='CG6042 gene product from transcript CG6042-RA', other_designations='CG6042-PA|Cyp12a4-PA|cytochrome P450', map_location='91F3-91F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19135158, end_position_on_the_genomic_accession=19137136, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42295, CurrentID=0, Status='live', Symbol='Ppcs', Aliases='Dmel_CG5629, BcDNA:GH03502, CG5629, Dmel\\CG5629, GH03502, PPCS, anon-WO0118547.401, dPPCS', description='Phosphopantothenoylcysteine synthetase', other_designations="4'-phosphopantothenoylcysteine synthetase|CG5629-PA|CG5629-PC|CG5629-PD|Ppcs-PA|Ppcs-PC|Ppcs-PD|phosphopantothenoylcysteine synthetase", map_location='91F3-91F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19139054, end_position_on_the_genomic_accession=19147910, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42296, CurrentID=0, Status='live', Symbol='CG5835', Aliases='Dmel_ Dmel\\CG5835', description='CG5835 gene product from transcript CG5835-RA', other_designations='CG5835-PA', map_location='91F4-91F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19148632, end_position_on_the_genomic_accession=19153686, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42297, CurrentID=0, Status='live', Symbol='nos', Aliases='Dmel_CG5637, CG5637, DRONANOS, Dmel\\CG5637, Dronanos, NANOS, NOS, Nos, l(3)07117, l(3)j3B6, l(3)j4B6, nNOS', description='nanos', other_designations='CG5637-PA|CG5637-PB|nos-PA|nos-PB', map_location='3-66.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19157238, end_position_on_the_genomic_accession=19160202, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42298, CurrentID=0, Status='live', Symbol='CG11779', Aliases='Dmel_CG11779, BEST:GM03811, BcDNA:GM038119, Dmel\\CG11779, GH-18370', description='CG11779 gene product from transcript CG11779-RB', other_designations='CG11779-PA|CG11779-PB|CG11779-PE|uncharacterized protein', map_location='91F4-91F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19154097, end_position_on_the_genomic_accession=19159878, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42299, CurrentID=0, Status='live', Symbol='CG42359', Aliases='Dmel_ CG5558, Dmel\\CG42359, Dmel_CG5558', description='CG42359 gene product from transcript CG42359-RA', other_designations='CG42359-PA|uncharacterized protein', map_location='91F4-91F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19161259, end_position_on_the_genomic_accession=19162300, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42300, CurrentID=0, Status='live', Symbol='sqz', Aliases='Dmel_CG5557, CG5557, Dmel\\CG5557, SQZ, Sqz, anon-WO0118547.697, l(3)02102', description='squeeze', other_designations='CG5557-PA|sqz-PA', map_location='91F4-91F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19164211, end_position_on_the_genomic_accession=19170576, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42301, CurrentID=0, Status='live', Symbol='CG14282', Aliases='Dmel_ Dmel\\CG14282', description='CG14282 gene product from transcript CG14282-RA', other_designations='CG14282-PA|uncharacterized protein', map_location='91F4-91F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19177000, end_position_on_the_genomic_accession=19177986, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42302, CurrentID=0, Status='live', Symbol='CG5555', Aliases='Dmel_ Dmel\\CG5555', description='CG5555 gene product from transcript CG5555-RB', other_designations='CG5555-PA|CG5555-PB', map_location='91F10-91F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19178085, end_position_on_the_genomic_accession=19180538, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42303, CurrentID=0, Status='live', Symbol='CG31475', Aliases='Dmel_ CG14281, CG3768, Dmel\\CG31475', description='CG31475 gene product from transcript CG31475-RA', other_designations='CG31475-PA|CG31475-PB|uncharacterized protein', map_location='91F11-91F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19180659, end_position_on_the_genomic_accession=19201096, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42305, CurrentID=0, Status='live', Symbol='cry', Aliases='Dmel_CG3772, CG3772, CRY, Cry, DCry, Dm-CRY1, DmCRY, DmCRY1, Dmcry, Dmel\\CG3772, anon-WO0140519.17, anon-WO0140519.19, anon-WO0140519.20, anon-WO0172774.15b, dCRY, dCry, dcry, cry', description='cryptochrome', other_designations='Blue-light receptor|CG3772-PA|cry-PA|crybaby', map_location='91F11-91F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19212154, end_position_on_the_genomic_accession=19215442, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42306, CurrentID=0, Status='live', Symbol='vib', Aliases='Dmel_CG5269, 0450/02, 1104/16, CG5269, Dmel\\CG5269, PITP, PITP-DM, Vib, anon-WO0142479.3, anon-WO0172774.12, anon-WO0172774.16, anon-WO0172774.28, gio, l(3)S045002, l(3)S045002c, l(3)S110416, l(3)j4A3, l(3)j5A6, l(3)j7A3, l(3)j7C8', description='vibrator', other_designations='CG5269-PA|CG5269-PB|CG5269-PC|PITP|giotto|lethal (3) S045002c|lethal (3) S110416|lethal (3) j5A6|vib-PA|vib-PB|vib-PC', map_location='91F11-91F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19215434, end_position_on_the_genomic_accession=19226651, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42307, CurrentID=0, Status='live', Symbol='CG11703', Aliases='Dmel_ Dmel\\CG11703, JYbeta2, anon-WO0140519.49', description='CG11703 gene product from transcript CG11703-RA', other_designations='CG11703-PA|uncharacterized protein', map_location='91F12-91F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19226903, end_position_on_the_genomic_accession=19228180, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42308, CurrentID=0, Status='live', Symbol='CG5250', Aliases='Dmel_ Dmel\\CG5250', description='CG5250 gene product from transcript CG5250-RA', other_designations='CG5250-PA', map_location='91F12-91F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19228217, end_position_on_the_genomic_accession=19229276, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42309, CurrentID=0, Status='live', Symbol='CG3773', Aliases='Dmel_ Dmel\\CG3773, anon-EST:fe2D10, anon-EST:fe2D12', description='CG3773 gene product from transcript CG3773-RA', other_designations='CG3773-PA|anon-fast-evolving-2D10|anon-fast-evolving-2D12', map_location='91F12-91F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19229463, end_position_on_the_genomic_accession=19230832, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42310, CurrentID=0, Status='live', Symbol='unc79', Aliases='Dmel_CG5237, CG5237, DUNC79, Dmel\\CG5237, UNC79, anon-WO0153538.74, anon-WO0172774.15, d', description='CG5237 gene product from transcript CG5237-RE', other_designations='CG5237-PB|CG5237-PC|CG5237-PD|CG5237-PE|CG5237-PF|unc79-PB|unc79-PC|unc79-PD|unc79-PE|unc79-PF', map_location='91F12-91F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19230764, end_position_on_the_genomic_accession=19246130, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  42311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42311, CurrentID=0, Status='live', Symbol='CG5217', Aliases='Dmel_ Dmel\\CG5217, anon-WO0140519.188', description='CG5217 gene product from transcript CG5217-RA', other_designations='CG5217-PA', map_location='91F12-91F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19247052, end_position_on_the_genomic_accession=19248136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42312, CurrentID=0, Status='live', Symbol='CG14280', Aliases='Dmel_ Dmel\\CG14280', description='CG14280 gene product from transcript CG14280-RB', other_designations='CG14280-PA|CG14280-PB|uncharacterized protein', map_location='91F13-91F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19256777, end_position_on_the_genomic_accession=19260132, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42313, CurrentID=0, Status='live', Symbol='Dl', Aliases='Dmel_CG3619, 0495/20, 0926/11, 1053/14, 1119/09, 1304/03, 1423/11, 1440/11, 1485/04, C1, CG3619, CT12133, D, DL, Dmel\\CG3619, E(ls)2, EC3-5, anon-WO0118547.269, delta, delta D1, dmDelta, l(3)05151, l(3)92Ab, l(3)j8C3', description='Delta', other_designations='CG3619-PA|CG3619-PB|CG3619-PC|Dl-PA|Dl-PB|Dl-PC|complementation group 1|delta|overflow', map_location='3-66.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19302731, end_position_on_the_genomic_accession=19326217, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42314, CurrentID=0, Status='live', Symbol='Ino80', Aliases='Dmel_CG31212, CG14279, CG18491, CG31212, CG3602, Dmel\\CG31212, INO80, dINO80, dIno80, psg25', description='INO80 complex subunit', other_designations='CG31212-PA|Ino80-PA', map_location='3-69.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19368995, end_position_on_the_genomic_accession=19403153, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  42315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42315, CurrentID=0, Status='live', Symbol='CG3581', Aliases='Dmel_ Dmel\\CG3581', description='CG3581 gene product from transcript CG3581-RA', other_designations='CG3581-PA', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19373757, end_position_on_the_genomic_accession=19374789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42316, CurrentID=0, Status='live', Symbol='CG31245', Aliases='Dmel_ BcDNA:AT17586, CG3575, Dmel\\CG31245', description='CG31245 gene product from transcript CG31245-RA', other_designations='CG31245-PA|uncharacterized protein', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19376195, end_position_on_the_genomic_accession=19377225, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42318, CurrentID=0, Status='live', Symbol='CG3734', Aliases='Dmel_ Dmel\\CG3734', description='CG3734 gene product from transcript CG3734-RA', other_designations='CG3734-PA|CG3734-PB', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19387861, end_position_on_the_genomic_accession=19389866, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42319, CurrentID=0, Status='live', Symbol='CG18493', Aliases='Dmel_ Dmel\\CG18493', description='CG18493 gene product from transcript CG18493-RA', other_designations='CG18493-PA|uncharacterized protein', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19390270, end_position_on_the_genomic_accession=19392132, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42320, CurrentID=0, Status='live', Symbol='CG3739', Aliases='Dmel_ Dmel\\CG3739', description='CG3739 gene product from transcript CG3739-RB', other_designations='CG3739-PB', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19393360, end_position_on_the_genomic_accession=19395188, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42322, CurrentID=0, Status='live', Symbol='CG5316', Aliases='Dmel_ Dmel\\CG5316', description='CG5316 gene product from transcript CG5316-RC', other_designations='CG5316-PA|CG5316-PB|CG5316-PC|CG5316-PD', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19403455, end_position_on_the_genomic_accession=19406294, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42323, CurrentID=0, Status='live', Symbol='CG11626', Aliases='Dmel_ Dmel\\CG11626', description='CG11626 gene product from transcript CG11626-RB', other_designations='CG11626-PB|uncharacterized protein', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19409175, end_position_on_the_genomic_accession=19411265, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42324, CurrentID=0, Status='live', Symbol='CG3517', Aliases='Dmel_ Dmel\\CG3517', description='CG3517 gene product from transcript CG3517-RA', other_designations='CG3517-PA', map_location='92A4-92A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19423779, end_position_on_the_genomic_accession=19425199, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42325, CurrentID=0, Status='live', Symbol='CG31221', Aliases='Dmel_ BcDNA:GH25289, CG11625, CG7237, Dmel\\CG31221, anon-EST:Posey251', description='CG31221 gene product from transcript CG31221-RA', other_designations='CG31221-PA|CG31221-PB|CG31221-PC|uncharacterized protein', map_location='92A3-92A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19412689, end_position_on_the_genomic_accession=19460311, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42326, CurrentID=0, Status='live', Symbol='CG6255', Aliases='Dmel_ AAF55672, Dmel\\CG6255, anon-WO0140519.197', description='CG6255 gene product from transcript CG6255-RA', other_designations='CG6255-PA', map_location='92A5-92A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19461879, end_position_on_the_genomic_accession=19463066, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42327, CurrentID=0, Status='live', Symbol='Dys', Aliases='Dmel_CG34157, CG17750, CG31175, CG34157, CG7240, CG7243, CG7344, DLP, DLP1, DLP186, DLP2, DLP3, DmDLP, DmDYS, Dmel\\CG34157, Dp117, Dp186, Dp205, GI3046716, IDLP, det, dmDLP, dmDp186, dmDys, dys', description='Dystrophin', other_designations='CG34157-PA|CG34157-PB|CG34157-PC|CG34157-PD|CG34157-PE|CG34157-PF|CG34157-PG|CG34157-PH|CG34157-PI|CG34157-PJ|CG34157-PK|CG34157-PL|Dys-PA|Dys-PB|Dys-PC|Dys-PD|Dys-PE|Dys-PF|Dys-PG|Dys-PH|Dys-PI|Dys-PJ|Dys-PK|Dys-PL|Dystrophin-like protein 1|Dystrophin-like protein 186|Dystrophin-like protein 2|Dystrophin-like protein 3|detached|dystrophin|dystrophin Dp186', map_location='3-72.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19461085, end_position_on_the_genomic_accession=19597288, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  42328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42328, CurrentID=0, Status='live', Symbol='CG15025', Aliases='Dmel_ Dmel\\CG15025', description='CG15025 gene product from transcript CG15025-RA', other_designations='CG15025-PA|uncharacterized protein', map_location='92A6-92A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19489570, end_position_on_the_genomic_accession=19491115, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42333, CurrentID=0, Status='live', Symbol='Cpr92A', Aliases='Dmel_CG6240, CG6240, DmelCpr92A, Dmel\\CG6240', description='Cuticular protein 92A', other_designations='CG6240-PA|Cpr92A-PA|cuticular protein 92A', map_location='92A10-92A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19597502, end_position_on_the_genomic_accession=19598955, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42334, CurrentID=0, Status='live', Symbol='CG7333', Aliases='Dmel_ Dmel\\CG7333', description='CG7333 gene product from transcript CG7333-RB', other_designations='CG7333-PB', map_location='92A10-92A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19600635, end_position_on_the_genomic_accession=19602629, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42335, CurrentID=0, Status='live', Symbol='CG7342', Aliases='Dmel_ Dmel\\CG7342', description='CG7342 gene product from transcript CG7342-RB', other_designations='CG7342-PA|CG7342-PB', map_location='92A10-92A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19603375, end_position_on_the_genomic_accession=19606475, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42336, CurrentID=0, Status='live', Symbol='CG17751', Aliases='Dmel_ Dmel\\CG17751', description='CG17751 gene product from transcript CG17751-RB', other_designations='CG17751-PB|uncharacterized protein', map_location='92A10-92A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19607254, end_position_on_the_genomic_accession=19609322, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42337, CurrentID=0, Status='live', Symbol='CG17752', Aliases='Dmel_ Dmel\\CG17752', description='CG17752 gene product from transcript CG17752-RA', other_designations='CG17752-PA|uncharacterized protein', map_location='92A10-92A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19610225, end_position_on_the_genomic_accession=19612437, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42338, CurrentID=0, Status='live', Symbol='CG16727', Aliases='Dmel_ Dmel\\CG16727', description='CG16727 gene product from transcript CG16727-RA', other_designations='CG16727-PA|uncharacterized protein', map_location='92A10-92A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19613381, end_position_on_the_genomic_accession=19615784, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42339, CurrentID=0, Status='live', Symbol='CG6231', Aliases='Dmel_ Dmel\\CG6231', description='CG6231 gene product from transcript CG6231-RC', other_designations='CG6231-PA|CG6231-PB|CG6231-PC|CG6231-PD|CG6231-PE|CG6231-PF|organic cation transporter', map_location='92A10-92A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19616024, end_position_on_the_genomic_accession=19632718, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  42340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42340, CurrentID=0, Status='live', Symbol='subdued', Aliases='Dmel_CG16718, CG16718, Dmel\\CG16718', description='CG16718 gene product from transcript CG16718-RA', other_designations='CG16718-PA|CG16718-PB|CG16718-PD|CG16718-PE|CG16718-PF|subdued-PA|subdued-PB|subdued-PD|subdued-PE|subdued-PF', map_location='92A11-92A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19633418, end_position_on_the_genomic_accession=19642722, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  42341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42341, CurrentID=0, Status='live', Symbol='Vha13', Aliases='Dmel_CG6213, ATP6V1G, BcDNA:AT14009, CG6213, Dmel\\CG6213, VhaG, chr3R:15469897..15470039, l(3)05113, vha13', description='Vacuolar H[+] ATPase 13kD subunit', other_designations='CG6213-PA|CG6213-PB|Vha13-PA|Vha13-PB|vacuolar H[+] ATPase 13kD subunit|vacuolar H[+] ATPase G-subunit', map_location='92A11-92A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19643018, end_position_on_the_genomic_accession=19644386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42342, CurrentID=0, Status='live', Symbol='Nup58', Aliases='Dmel_CG7360, CG7360, Dm NUP58, Dmel\\CG7360, RE52572p', description='Nucleoporin 58kD', other_designations='CG7360-PA|Nup58-PA|nucleoporin 58|nucleoporin 58kD', map_location='92A11-92A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19645026, end_position_on_the_genomic_accession=19646906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42343, CurrentID=0, Status='live', Symbol='CG6195', Aliases='Dmel_ BEST:GH15838, Dmel\\CG6195', description='CG6195 gene product from transcript CG6195-RA', other_designations='CG6195-PA', map_location='92A11-92A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19646930, end_position_on_the_genomic_accession=19648410, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42344, CurrentID=0, Status='live', Symbol='CG31219', Aliases='Dmel_ CG16731, Dmel\\CG31219', description='CG31219 gene product from transcript CG31219-RB', other_designations='CG31219-PB|uncharacterized protein', map_location='92A12-92A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19651018, end_position_on_the_genomic_accession=19652430, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42345, CurrentID=0, Status='live', Symbol='CG6184', Aliases='Dmel_ Dmel\\CG6184', description='CG6184 gene product from transcript CG6184-RE', other_designations='CG6184-PA|CG6184-PE|CG6184-PF', map_location='92A13-92A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19652369, end_position_on_the_genomic_accession=19657866, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42347, CurrentID=0, Status='live', Symbol='CG7432', Aliases='Dmel_ Dmel\\CG7432, SP19, c-SP19', description='CG7432 gene product from transcript CG7432-RB', other_designations='CG7432-PB', map_location='92A13-92A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19665007, end_position_on_the_genomic_accession=19670827, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42348, CurrentID=0, Status='live', Symbol='Naam', Aliases='Dmel_CG31216, AAM52663, CG31216, CG7444, CG7500, D-NAAM, Dmel\\CG31216, anon-WO0172774.59, anon-WO0172774.60', description='Nicotinamide amidase', other_designations='CG31216-PA|CG31216-PB|Naam-PA|Naam-PB|nicotinamidase|nicotinamide amidase', map_location='92A13-92B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19671450, end_position_on_the_genomic_accession=19698742, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42350, CurrentID=0, Status='live', Symbol='GluClalpha', Aliases='Dmel_CG7535, BcDNA:HL07853, CG7535, CT23049, Dm-GluCl, DmGlu, Dm Dmel\\CG7535, DrosGlu-Cl-alpha, DrosGluCl, DrosGluCl-alpha, DrosGluCl-alpha1, GLUCL, GluCl, GluCla, dGluCl-alpha, glc, glu', description='CG7535 gene product from transcript CG7535-RK', other_designations='CG7535-PE|CG7535-PF|CG7535-PH|CG7535-PI|CG7535-PJ|CG7535-PK|CG7535-PL|CG7535-PM|CG7535-PN|CG7535-PO|CG7535-PP|GluClalpha-PE|GluClalpha-PF|GluClalpha-PH|GluClalpha-PI|GluClalpha-PJ|GluClalpha-PK|GluClalpha-PL|GluClalpha-PM|GluClalpha-PN|GluClalpha-PO|GluClalpha-PP|glutamate-gated chloride channel|indefinite', map_location='92B1-92B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19730715, end_position_on_the_genomic_accession=19775231, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  42351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42351, CurrentID=0, Status='live', Symbol='CG6300', Aliases='Dmel_ Dmel\\CG6300', description='CG6300 gene product from transcript CG6300-RA', other_designations='CG6300-PA', map_location='92B2-92B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19771382, end_position_on_the_genomic_accession=19773225, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42352, CurrentID=0, Status='live', Symbol='CG11659', Aliases='Dmel_ Dmel\\CG11659', description='CG11659 gene product from transcript CG11659-RA', other_designations='CG11659-PA|uncharacterized protein', map_location='92B2-92B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19775459, end_position_on_the_genomic_accession=19777280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42353, CurrentID=0, Status='live', Symbol='CG11391', Aliases='Dmel_ Dmel\\CG11391', description='CG11391 gene product from transcript CG11391-RA', other_designations='CG11391-PA|uncharacterized protein', map_location='92B2-92B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19778750, end_position_on_the_genomic_accession=19780727, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42354, CurrentID=0, Status='live', Symbol='CG11407', Aliases='Dmel_ Dmel\\CG11407', description='CG11407 gene product from transcript CG11407-RA', other_designations='CG11407-PA|uncharacterized protein', map_location='92B2-92B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19781246, end_position_on_the_genomic_accession=19783145, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42355, CurrentID=0, Status='live', Symbol='CG11453', Aliases='Dmel_ Dmel\\CG11453', description='CG11453 gene product from transcript CG11453-RA', other_designations='CG11453-PA|uncharacterized protein', map_location='92B2-92B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19783522, end_position_on_the_genomic_accession=19785519, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42356, CurrentID=0, Status='live', Symbol='bnl', Aliases='Dmel_CG4608, BNL, Bnl, CG4608, DFGF, DmBnl, Dmel\\CG4608, FGF, l(3)06916, l(3)neo52', description='branchless', other_designations='CG4608-PA|CG4608-PC|CG4608-PD|bnl-PA|bnl-PC|bnl-PD|fibroblast growth factor|fibroblast growth factor protein|fibroblast growth factor)', map_location='92B2-92B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19794073, end_position_on_the_genomic_accession=19836817, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42357, CurrentID=0, Status='live', Symbol='CG4662', Aliases='Dmel_ Dmel\\CG4662', description='CG4662 gene product from transcript CG4662-RB', other_designations='CG4662-PA|CG4662-PB|CG4662-PC|CG4662-PD|CG4662-PE|CG4662-PF', map_location='92B3-92B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19847311, end_position_on_the_genomic_accession=19854403, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42358, CurrentID=0, Status='live', Symbol='Ire1', Aliases='Dmel_CG4583, CG4583, Dmel\\CG4583, IRE-like, IRE1, Ire, Ire-1, dIRE1, dire-1, ire-1, ire1', description='Inositol-requiring enzyme-1', other_designations='CG4583-PB|Ire|Ire1-PB|inositol required 1|ire-1', map_location='92B3-92B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19853908, end_position_on_the_genomic_accession=19861291, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42359, CurrentID=0, Status='live', Symbol='CG11447', Aliases='Dmel_ Dmel\\CG11447', description='CG11447 gene product from transcript CG11447-RA', other_designations='CG11447-PA|uncharacterized protein', map_location='92B3-92B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19861610, end_position_on_the_genomic_accession=19862600, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42360, CurrentID=0, Status='live', Symbol='CG4572', Aliases='Dmel_ CT14806, Dmel\\CG4572, anon-WO02059370.13', description='CG4572 gene product from transcript CG4572-RD', other_designations='CG4572-PA|CG4572-PB|CG4572-PD', map_location='92B3-92B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19862517, end_position_on_the_genomic_accession=19864393, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42361, CurrentID=0, Status='live', Symbol='CG4686', Aliases='Dmel_ BcDNA:AT09463, Dmel\\CG4686', description='CG4686 gene product from transcript CG4686-RA', other_designations='CG4686-PA|CG4686-PB', map_location='92B4-92B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19864496, end_position_on_the_genomic_accession=19866869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42362, CurrentID=0, Status='live', Symbol='CG4562', Aliases='Dmel_ CT14766, Dme_CG4562, Dmel\\CG4562', description='CG4562 gene product from transcript CG4562-RC', other_designations='CG4562-PA|CG4562-PB|CG4562-PC|CG4562-PD|CG4562-PE', map_location='92B4-92B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19865417, end_position_on_the_genomic_accession=19875678, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42363, CurrentID=0, Status='live', Symbol='CG17186', Aliases='Dmel_ Dmel\\CG17186', description='CG17186 gene product from transcript CG17186-RA', other_designations='CG17186-PA|uncharacterized protein', map_location='92B4-92B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19875974, end_position_on_the_genomic_accession=19877514, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42364, CurrentID=0, Status='live', Symbol='Arc42', Aliases='Dmel_CG4703, CG4703, Dmel\\CG4703, dARC42', description='CG4703 gene product from transcript CG4703-RA', other_designations='Arc42-PA|CG4703-PA', map_location='92B4-92B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19877737, end_position_on_the_genomic_accession=19879309, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42366, CurrentID=0, Status='live', Symbol='Ask1', Aliases='Dmel_CG4720, ASK1, CG4720, D-ASK, DASK, DASK1, DPk92B/ASK1, Dmel\\CG4720, MEKK5/ASK1, Pk92B, dASK, dAsk1, pk92B', description='Apoptotic signal-regulating kinase 1', other_designations='Ask1-PB|Ask1-PC|CG4720-PB|CG4720-PC|apoptotic signal-regulating kinase 1|protein kinase at 92B', map_location='92B4-92B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19879719, end_position_on_the_genomic_accession=19886030, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42367, CurrentID=0, Status='live', Symbol='ninaE', Aliases='Dmel_CG4550, 143283_at, 1F9, BEST:GH11778, CG4550, DMELRH1, DRh1, Dm Rh1, DmRh1, Dmel\\CG4550, FBgn0002940, Nina E, NinaE, R, RH1, Rh, Rh-1, Rh1, Rh1(ninaE), Rh1/ninaE, Rhodopsin-1, opsin, ora, rh1, rh1/ninaE, rhodopsin', description='neither inactivation nor afterpotential E', other_designations='CG4550-PA|ninaE-PA|rhodopsin|rhodopsin 1|rhodopsin-1|rhodopsin1', map_location='3-66.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19886255, end_position_on_the_genomic_accession=19888206, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42368, CurrentID=0, Status='live', Symbol='CG4733', Aliases="Dmel_ B'', Dmel\\CG4733, PP2A like, PP2A-B'', PR-72, PR72, dPP2A-PR72, dPR72", description='CG4733 gene product from transcript CG4733-RD', other_designations='CG4733-PB|CG4733-PC|CG4733-PD|CG4733-PE', map_location='92B4-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19890689, end_position_on_the_genomic_accession=19898738, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42369, CurrentID=0, Status='live', Symbol='CG4538', Aliases='Dmel_ ClpX, Dmel\\CG4538', description='CG4538 gene product from transcript CG4538-RC', other_designations='CG4538-PA|CG4538-PC|CG4538-PD', map_location='92B6-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19897372, end_position_on_the_genomic_accession=19903689, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42370, CurrentID=0, Status='live', Symbol='Surf6', Aliases='Dmel_CG4510, CG4510, Dmel\\CG4510', description='Surfeit 6', other_designations='CG4510-PA|Surf6-PA|surfeit 6', map_location='92B6-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19903665, end_position_on_the_genomic_accession=19904868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42371, CurrentID=0, Status='live', Symbol='RhoGAP92B', Aliases='Dmel_CG4755, CG4755, Dmel\\CG4755, dRich, drich', description='Rho GTPase activating protein at 92B', other_designations='CG4755-PA|CG4755-PB|CG4755-PC|CG4755-PD|CG4755-PE|RhoGAP92B-PA|RhoGAP92B-PB|RhoGAP92B-PC|RhoGAP92B-PD|RhoGAP92B-PE', map_location='92B6-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19905164, end_position_on_the_genomic_accession=19914272, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42373, CurrentID=0, Status='live', Symbol='Xport', Aliases='Dmel_CG4468, CG4468, Dmel\\CG4468, xport', description='exit protein of rhodopsin and TRP', other_designations='CG4468-PA|Xport-PA', map_location='92B6-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19914396, end_position_on_the_genomic_accession=19915412, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42374, CurrentID=0, Status='live', Symbol='CG4465', Aliases='Dmel_ Dmel\\CG4465', description='CG4465 gene product from transcript CG4465-RA', other_designations='CG4465-PA', map_location='92B6-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19915868, end_position_on_the_genomic_accession=19917839, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42375, CurrentID=0, Status='live', Symbol='CG4770', Aliases='Dmel_ Dmel\\CG4770', description='CG4770 gene product from transcript CG4770-RA', other_designations='CG4770-PA', map_location='92B6-92B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19918507, end_position_on_the_genomic_accession=19920584, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42376, CurrentID=0, Status='live', Symbol='CG4462', Aliases='Dmel_ Dmel\\CG4462', description='CG4462 gene product from transcript CG4462-RA', other_designations='CG4462-PA', map_location='92B7-92B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19921744, end_position_on_the_genomic_accession=19925371, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42377, CurrentID=0, Status='live', Symbol='CG4459', Aliases='Dmel_ Dmel\\CG4459', description='CG4459 gene product from transcript CG4459-RA', other_designations='CG4459-PA', map_location='92B8-92B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19926322, end_position_on_the_genomic_accession=19928644, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42379, CurrentID=0, Status='live', Symbol='mira', Aliases='Dmel_CG12249, CG12249, Dmel\\CG12249, MIR, MIRA, Mir, Mira, Miranda, mir', description='miranda', other_designations='CG12249-PA|CG12249-PB|CG12249-PC|mira-PA|mira-PB|mira-PC', map_location='3-67.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19931799, end_position_on_the_genomic_accession=19935120, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42380, CurrentID=0, Status='live', Symbol='Hs6st', Aliases='Dmel_CG4451, CG11683, CG17188, CG18533, CG4451, DmHs6st, Dmel\\CG4451, dHS6ST, hs6st', description='Heparan sulfate 6-O-sulfotransferase', other_designations='6-O sulfotransferase|6-O-sulfotransferase|CG4451-PA|CG4451-PB|Hs6st-PA|Hs6st-PB|heparan sulfate 6-O-sulfotransferase', map_location='92B9-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19929277, end_position_on_the_genomic_accession=20008512, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42381, CurrentID=0, Status='live', Symbol='CG4783', Aliases='Dmel_ Dmel\\CG4783, GH26202', description='CG4783 gene product from transcript CG4783-RA', other_designations='CG4783-PA|CG4783-PB', map_location='92B9-92B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19948443, end_position_on_the_genomic_accession=19948907, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42382, CurrentID=0, Status='live', Symbol='CG31213', Aliases='Dmel_ CG11460, CG4794, Dmel\\CG31213', description='CG31213 gene product from transcript CG31213-RB', other_designations='CG31213-PB|CG31213-PC|uncharacterized protein', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19972898, end_position_on_the_genomic_accession=19979569, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42385, CurrentID=0, Status='live', Symbol='MED25', Aliases='Dmel_CG12254, Arc92, CG12254, Dmel\\CG12254, LD07688, Med25, dARC92, dDrip97, med25', description='Mediator complex subunit 25', other_designations='CG12254-PA|MED25-PA|mediator complex subunit 25', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20008947, end_position_on_the_genomic_accession=20011712, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42386, CurrentID=0, Status='live', Symbol='CG17190', Aliases='Dmel_ Dmel\\CG17190', description='CG17190 gene product from transcript CG17190-RB', other_designations='CG17190-PB|uncharacterized protein', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20011776, end_position_on_the_genomic_accession=20013504, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42387, CurrentID=0, Status='live', Symbol='CG4836', Aliases='Dmel_ DM_7300579, Dmel\\CG4836, SODH-1, Srd, anon-WO0140519.105, anon-WO0140519.33, bs15h04.y1, bs36e10.y1', description='CG4836 gene product from transcript CG4836-RC', other_designations='CG4836-PA|CG4836-PB|CG4836-PC|CG4836-PD|CG4836-PE|L-iditol 2-dehydrogenase|sorbitol dehydrogenase', map_location='3-65.3 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20013872, end_position_on_the_genomic_accession=20018248, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42388, CurrentID=0, Status='live', Symbol='CG4433', Aliases='Dmel_ Dmel\\CG4433', description='CG4433 gene product from transcript CG4433-RD', other_designations='CG4433-PB|CG4433-PC|CG4433-PD', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20018269, end_position_on_the_genomic_accession=20021338, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42389, CurrentID=0, Status='live', Symbol='psidin', Aliases='Dmel_CG4845, CG4845, Dmel\\CG4845, ird16, psid', description='phagocyte signaling impaired', other_designations='CG4845-PA|immune response deficient 16|psidin-PA', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20021704, end_position_on_the_genomic_accession=20026482, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42390, CurrentID=0, Status='live', Symbol='CG4424', Aliases='Dmel_ Dmel\\CG4424', description='CG4424 gene product from transcript CG4424-RA', other_designations='CG4424-PA', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20026487, end_position_on_the_genomic_accession=20028040, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42391, CurrentID=0, Status='live', Symbol='CG4854', Aliases='Dmel_ Dmel\\CG4854', description='CG4854 gene product from transcript CG4854-RA', other_designations='CG4854-PA', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20028058, end_position_on_the_genomic_accession=20029561, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42392, CurrentID=0, Status='live', Symbol='trem', Aliases='Dmel_CG4413, CG4413, Dmel\\CG4413', description='trade embargo', other_designations='CG4413-PA|trem-PA', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20029560, end_position_on_the_genomic_accession=20031309, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42393, CurrentID=0, Status='live', Symbol='CG4936', Aliases='Dmel_ Dmel\\CG4936', description='CG4936 gene product from transcript CG4936-RA', other_designations='CG4936-PA', map_location='92C1-92C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20031823, end_position_on_the_genomic_accession=20034330, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42394, CurrentID=0, Status='live', Symbol='CG10889', Aliases='Dmel_ Dmel\\CG10889', description='CG10889 gene product from transcript CG10889-RA', other_designations='CG10889-PA|uncharacterized protein', map_location='92C1-92C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20033881, end_position_on_the_genomic_accession=20036056, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42396, CurrentID=0, Status='live', Symbol='CG4390', Aliases='Dmel_ Dmel\\CG4390', description='CG4390 gene product from transcript CG4390-RB', other_designations='CG4390-PA|CG4390-PB', map_location='92C4-92C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20056418, end_position_on_the_genomic_accession=20059750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42397, CurrentID=0, Status='live', Symbol='mdlc', Aliases='Dmel_CG4973, CG4973, Dmel\\CG4973', description='midlife crisis', other_designations='CG4973-PA|mdlc-PA', map_location='92C4-92C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20057748, end_position_on_the_genomic_accession=20059291, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42398, CurrentID=0, Status='live', Symbol='Rh3', Aliases='Dmel_CG10888, CG10888, DMELRH3, Dm Rh3, Dmel\\CG10888, FBgn0003249, RH3, Rh, rh3', description='Rhodopsin 3', other_designations='CG10888-PA|Rh3-PA|Rhodopsin3|rhodopsin|rhodopsin 3', map_location='92C5-92C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20081060, end_position_on_the_genomic_accession=20082450, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42399, CurrentID=0, Status='live', Symbol='CG10887', Aliases='Dmel_ Dmel\\CG10887', description='CG10887 gene product from transcript CG10887-RA', other_designations='CG10887-PA|uncharacterized protein', map_location='92D1-92D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20114838, end_position_on_the_genomic_accession=20117139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42400, CurrentID=0, Status='live', Symbol='CG5023', Aliases='Dmel_ Dmel\\CG5023', description='CG5023 gene product from transcript CG5023-RA', other_designations='CG5023-PA', map_location='92D2-92D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20169435, end_position_on_the_genomic_accession=20174872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42401, CurrentID=0, Status='live', Symbol='CG17199', Aliases='Dmel_ Dmel\\CG17199', description='CG17199 gene product from transcript CG17199-RA', other_designations='CG17199-PA|uncharacterized protein', map_location='92D5-92D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20208201, end_position_on_the_genomic_accession=20209791, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42402, CurrentID=0, Status='live', Symbol='Nlg4', Aliases='Dmel_CG34139, BcDNA:AT13703, CG15682, CG15683, CG31209, CG31499, CG34139, CG5030, CG5035, Dmel\\CG34139, dnlg4', description='Neuroligin 4', other_designations='CG34139-PB|CG34139-PC|CG34139-PD|CG34139-PE|Nlg4-PB|Nlg4-PC|Nlg4-PD|Nlg4-PE|neuroligin 4', map_location='92D5-92D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20204809, end_position_on_the_genomic_accession=20245473, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  13084068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=13084068, CurrentID=0, Status='live', Symbol='CG43355', Aliases='Dmel_ Dmel\\CG43355, chr2L:11486634..11486740', description='CG43355 gene product from transcript CG43355-RC', other_designations='CG43355-PA|CG43355-PC|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11484707, end_position_on_the_genomic_accession=11486949, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13084069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=13084069, CurrentID=0, Status='live', Symbol='CG43346', Aliases='Dmel_ CG9243, Dmel\\CG43346, Dmel_CG9243', description='CG43346 gene product from transcript CG43346-RC', other_designations='CG43346-PA|CG43346-PB|CG43346-PC|uncharacterized protein', map_location='39A7-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21166046, end_position_on_the_genomic_accession=21168581, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42409, CurrentID=0, Status='live', Symbol='CG4367', Aliases='Dmel_ Dmel\\CG4367', description='CG4367 gene product from transcript CG4367-RB', other_designations='CG4367-PA|CG4367-PB', map_location='92E2-92E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20303955, end_position_on_the_genomic_accession=20304954, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42410, CurrentID=0, Status='live', Symbol='CG4362', Aliases='Dmel_ Dmel\\CG4362', description='CG4362 gene product from transcript CG4362-RA', other_designations='CG4362-PA', map_location='92E2-92E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20306216, end_position_on_the_genomic_accession=20307537, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42412, CurrentID=0, Status='live', Symbol='CG42668', Aliases='Dmel_ CG15684, CG5077, Dmel\\CG42668, Dmel_CG15684, Dmel_CG5077', description='CG42668 gene product from transcript CG42668-RK', other_designations='CG42668-PD|CG42668-PE|CG42668-PF|CG42668-PG|CG42668-PH|CG42668-PK|CG42668-PL|CG42668-PM|CG42668-PN|uncharacterized protein', map_location='92E2-92E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20309344, end_position_on_the_genomic_accession=20324196, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  42413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42413, CurrentID=0, Status='live', Symbol='CG4360', Aliases='Dmel_ Dmel\\CG4360', description='CG4360 gene product from transcript CG4360-RA', other_designations='CG4360-PA|CG4360-PB', map_location='92E2-92E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20324274, end_position_on_the_genomic_accession=20328003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42414, CurrentID=0, Status='live', Symbol='Sirt2', Aliases='Dmel_CG5085, CG5085, D.mel2, Dmel\\CG5085, SIRT2, dSIRT2', description='Sirtuin 2', other_designations='CG5085-PA|CG5085-PB|Sirt2-PA|Sirt2-PB|sirtuin 2', map_location='92E3-92E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20328465, end_position_on_the_genomic_accession=20330915, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42415, CurrentID=0, Status='live', Symbol='Ir92a', Aliases='Dmel_CG15685, CG15685, DmelIR92a, Dmel\\CG15685, IR92a', description='Ionotropic receptor 92a', other_designations='CG15685-PC|Ir92a-PC|ionotropic receptor 92a', map_location='92E4-92E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20338973, end_position_on_the_genomic_accession=20343902, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42416, CurrentID=0, Status='live', Symbol='MtnC', Aliases='Dmel_CG5097, CG5097, Dmel\\CG5097', description='Metallothionein C', other_designations='CG5097-PA|MtnC-PA|metallothionein|metallothionein C', map_location='92E4-92E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20360308, end_position_on_the_genomic_accession=20360634, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42421, CurrentID=0, Status='live', Symbol='CG4335', Aliases='Dmel_ Dmel\\CG4335', description='CG4335 gene product from transcript CG4335-RA', other_designations='CG4335-PA', map_location='92E7-92E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20406615, end_position_on_the_genomic_accession=20407871, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42422, CurrentID=0, Status='live', Symbol='eIF3gb', Aliases='Dmel_CG10881, CG10881, Dmel\\CG10881, cg10881, eIF3-S4', description='eIF3g paralogue b', other_designations='CG10881-PA|eIF3gb-PA', map_location='92E8-92E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20408265, end_position_on_the_genomic_accession=20409243, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42423, CurrentID=0, Status='live', Symbol='Dic2', Aliases='Dmel_CG4323, CG4323, DmDic2p, Dmel\\CG4323', description='Dicarboxylate carrier 2', other_designations='CG4323-PA|CG4323-PC|Dic2-PA|Dic2-PC|dicarboxylate carrier 2', map_location='92E12-92E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20499992, end_position_on_the_genomic_accession=20502015, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42424, CurrentID=0, Status='live', Symbol='MtnB', Aliases='Dmel_CG4312, CG4312, Dmel\\CG4312, DrosMto, MTN, MTO, MtnA, Mto, mtnB', description='Metallothionein B', other_designations='CG4312-PA|CG4312-PB|CG4312-PC|MtnB-PA|MtnB-PB|MtnB-PC|metallothionein|metallothionein B', map_location='92E12-92E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20503102, end_position_on_the_genomic_accession=20503746, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42425, CurrentID=0, Status='live', Symbol='Or92a', Aliases='Dmel_CG17916, 92E.1, 92a, CG17916, DOR111, Dmel\\CG17916, OR 92a, Or92A', description='Odorant receptor 92a', other_designations='CG17916-PA|Or92a-PA|odorant receptor 92a', map_location='92E13-92E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20504788, end_position_on_the_genomic_accession=20506186, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42426, CurrentID=0, Status='live', Symbol='MFS9', Aliases='Dmel_CG4288, CG4288, Dmel\\CG4288', description='Major Facilitator Superfamily Transporter 9', other_designations='CG4288-PA|CG4288-PB|MFS9-PA|MFS9-PB|major facilitator superfamily transporter 9', map_location='92E14-92E14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20507392, end_position_on_the_genomic_accession=20512468, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42428, CurrentID=0, Status='live', Symbol='Stat92E', Aliases='Dmel_CG4257, CG4257, D-STAT, D-Stat, D-stat, D-stat/stat92E, DRODSRC, DSRC, DSTAT, DmSTAT, Dmel\\CG4257, Dstat, Dstat 92E, Dstat92E, SD-stat, STAT, STAT 92E, STAT92E, STAT92e, Stat, Stat1alpha-like, Stat92, Stat92e, d-STAT, dSTAT, dSTAT92E, dSTAT92E/marelle, dstat, l(3)06346, l(3)j6C8, mrL, mrl, mrl stat92E, pSTAT, stat, stat92, stat92E, stat92e, stat[92E]', description='Signal-transducer and activator of transcription protein at 92E', other_designations='CG4257-PB|CG4257-PC|CG4257-PE|CG4257-PF|CG4257-PG|CG4257-PH|CG4257-PI|CG4257-PJ|CG4257-PK|CG4257-PL|CG4257-PM|Signal-transducer and activator of transcription protein|Stat92E-PB|Stat92E-PC|Stat92E-PE|Stat92E-PF|Stat92E-PG|Stat92E-PH|Stat92E-PI|Stat92E-PJ|Stat92E-PK|Stat92E-PL|Stat92E-PM|marelle|signal transducer and activator of transcription|signal transducers and activator of transcription', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20535323, end_position_on_the_genomic_accession=20552311, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42429, CurrentID=0, Status='live', Symbol='att-ORFA', Aliases='Dmel_CG4241, CG4241, Dmel\\CG4241, att', description='alternative testis transcripts open reading frame A', other_designations='CG4241-PA|CG4241-PB|CG4241-PC|CG4241-PD|CG4241-PE|CG4241-PF|alternative testis transcripts|att-ORFA-PA|att-ORFA-PB|att-ORFA-PC|att-ORFA-PD|att-ORFA-PE|att-ORFA-PF', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20552385, end_position_on_the_genomic_accession=20557667, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42431, CurrentID=0, Status='live', Symbol='CG5191', Aliases='Dmel_ Dmel\\CG5191', description='CG5191 gene product from transcript CG5191-RE', other_designations='CG5191-PA|CG5191-PB|CG5191-PC|CG5191-PD|CG5191-PE|CG5191-PF', map_location='92F1-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20560865, end_position_on_the_genomic_accession=20573000, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42432, CurrentID=0, Status='live', Symbol='CG10877', Aliases='Dmel_ Dmel\\CG10877', description='CG10877 gene product from transcript CG10877-RA', other_designations='CG10877-PA|uncharacterized protein', map_location='92F1-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20562486, end_position_on_the_genomic_accession=20564067, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42433, CurrentID=0, Status='live', Symbol='TFAM', Aliases='Dmel_CG4217, CG4217, D-mtTFA, Dmel\\CG4217, MTTFA, Tfam, d-TFAM, d-mttfa, d-tfam, mtTFA, mttfa, tfam', description='mitochondrial transcription factor A', other_designations='CG4217-PA|CG4217-PB|TFAM-PA|TFAM-PB|mitochondiral transcription fctor a|mitochondrial transcription factor', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20575245, end_position_on_the_genomic_accession=20577010, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42434, CurrentID=0, Status='live', Symbol='CG5412', Aliases='Dmel_ Dmel\\CG5412', description='CG5412 gene product from transcript CG5412-RA', other_designations='CG5412-PA', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20577053, end_position_on_the_genomic_accession=20578131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42435, CurrentID=0, Status='live', Symbol='EloB', Aliases='Dmel_CG4204, CG4204, Dmel\\CG4204, Elo-B, Elongin-B', description='Elongin B', other_designations='CG4204-PA|EloB-PA|Elongin-B|ElonginB|d-elongin B|elongin B', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20578222, end_position_on_the_genomic_accession=20579850, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42436, CurrentID=0, Status='live', Symbol='Srp14', Aliases='Dmel_CG5417, BcDNA:RE48595, CG5417, Dmel\\CG5417', description='Signal recognition particle protein 14', other_designations='CG5417-PB|Srp14-PB|signal recognition particle protein 14', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20580115, end_position_on_the_genomic_accession=20580764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42437, CurrentID=0, Status='live', Symbol='CG16953', Aliases='Dmel_ Dmel\\CG16953', description='CG16953 gene product from transcript CG16953-RB', other_designations='CG16953-PB|uncharacterized protein', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20581770, end_position_on_the_genomic_accession=20585833, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42438, CurrentID=0, Status='live', Symbol='Sep2', Aliases='Dmel_CG4173, 38961, CG4173, Dmel\\CG4173, sep2', description='Septin 2', other_designations='CG4173-PA|CG4173-PB|Sep2-PA|Sep2-PB|Septin2|septin|septin 2|septin-2', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20586129, end_position_on_the_genomic_accession=20588356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42439, CurrentID=0, Status='live', Symbol='Srp72', Aliases='Dmel_CG5434, CG5434, Dmel\\CG5434', description='Signal recognition particle protein 72', other_designations='CG5434-PA|Srp72-PA|signal recognition particle protein 72', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20588711, end_position_on_the_genomic_accession=20591040, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42440, CurrentID=0, Status='live', Symbol='CG4159', Aliases='Dmel_ Dmel\\CG4159', description='CG4159 gene product from transcript CG4159-RA', other_designations='CG4159-PA|CG4159-PB', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20591024, end_position_on_the_genomic_accession=20592671, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42443, CurrentID=0, Status='live', Symbol='CG15923', Aliases='Dmel_ Dmel\\CG15923', description='CG15923 gene product from transcript CG15923-RB', other_designations='CG15923-PB|CG15923-PC|uncharacterized protein', map_location='92F3-92F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20611251, end_position_on_the_genomic_accession=20616901, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42444, CurrentID=0, Status='live', Symbol='CG5466', Aliases='Dmel_ Dmel\\CG5466', description='CG5466 gene product from transcript CG5466-RA', other_designations='CG5466-PA', map_location='92F3-92F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20617883, end_position_on_the_genomic_accession=20621269, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42445, CurrentID=0, Status='live', Symbol='H', Aliases='Dmel_CG5460, CG5460, Dmel\\CG5460', description='Hairless', other_designations='CG5460-PA|CG5460-PB|CG5460-PC|CG5460-PD|H-PA|H-PB|H-PC|H-PD|hairless|hairy', map_location='3-69.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20621938, end_position_on_the_genomic_accession=20628915, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42446, CurrentID=0, Status='live', Symbol='Pi3K92E', Aliases="Dmel_CG4141, CG4141, DP110, Dmel\\CG4141, Dmp110, Dp110, Dp110/PI3K, Dpp110, PI(3)K, PI-3 kinase, PI-3-K, PI3'K, PI3K, PI3K-92D, PI3K-92E/Dp110, PI3K-Dp110, PI3K-dp110, PI3K92E, PI3k, PI[[3]]K, Pi3K, Pi3K92D|p110, Pi3Kp110, Pi3k, PiK92E, anon-92Ed, dP110, dP13K, dPI3K, dPI3K[92E], dp110, dp110alpha, p110, p120, pi3k92e, rea, type-1 PI3K", description='CG4141 gene product from transcript CG4141-RB', other_designations='CG4141-PA|CG4141-PB|CG4141-PC|PI 3-kinase|PI3 kinase|PI3-kinase|PI3Kinase92E|Phosphatidylinositol-3-kinase|Phosphoinositide-3 kinase|Pi3K92E-PA|Pi3K92E-PB|Pi3K92E-PC|class I PI(3)K|dPI 3-kinase|insulin-regulated PI3 kinase|p110 PI3 kinase|phosphatidylinositol 3-kinase|phosphatidylinositol-3 kinase|phospho-inositol-3 kinase at 92E|phosphoinositide 3 kinase Dp110|phosphoinositide 3-kinase|phosphotidylinositol 3 kinase', map_location='92F3-92F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20628972, end_position_on_the_genomic_accession=20634732, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42447, CurrentID=0, Status='live', Symbol='Lrrk', Aliases='Dmel_CG5483, CG5483, CT17358, Dmel\\CG5483, LRRK, LRRK2, dLRRK, lrrk', description='Leucine-rich repeat kinase', other_designations='CG5483-PB|CG5483-PC|CG5483-PD|LRRK|Lrrk-PB|Lrrk-PC|Lrrk-PD', map_location='92F4-92F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20635480, end_position_on_the_genomic_accession=20644279, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  42449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42449, CurrentID=0, Status='live', Symbol='Nep4', Aliases='Dmel_CG4058, CG4058, DmeNEP4, Dmel\\CG4058, NEP4, dNEP4, nep4', description='Neprilysin 4', other_designations='CG4058-PA|CG4058-PB|CG4058-PC|Nep4-PA|Nep4-PB|Nep4-PC|neprilysin 4|neprilysin4', map_location='92F4-92F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20652567, end_position_on_the_genomic_accession=20659904, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42450, CurrentID=0, Status='live', Symbol='Cpr92F', Aliases='Dmel_CG5494, CG5494, Dmel\\CG5494', description='Cuticular protein 92F', other_designations='CG5494-PA|Cpr92F-PA|cuticular protein 92F', map_location='92F5-92F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20669571, end_position_on_the_genomic_accession=20672821, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42451, CurrentID=0, Status='live', Symbol='CG4000', Aliases='Dmel_ Dmel\\CG4000', description='CG4000 gene product from transcript CG4000-RA', other_designations='CG4000-PA', map_location='92F6-92F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20676224, end_position_on_the_genomic_accession=20677351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42452, CurrentID=0, Status='live', Symbol='Oct-TyrR', Aliases='Dmel_CG7485, CG7485, DmTAR1, DmTyrR, Dmel\\CG7485, Dmoct/tyr, DrmOTR, OA-R, OAR_DROME, OTR, OcR, Ocr, OctoR1, OctyR99AB, TAR1, TYR, TyR, Tyr, Tyr-dro, TyrR, hono, l(3)neo30', description='Octopamine-Tyramine receptor', other_designations='CG7485-PA|CG7485-PB|Oct-TyrR-PA|Oct-TyrR-PB|Oct/Tyr receptor|Tyramine-receptor|honoka|octopamine receptor|octopamine receptor 1|octopamine/tyramine receptor|tyramine receptor', map_location='79B1-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22035003, end_position_on_the_genomic_accession=22065279, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42453, CurrentID=0, Status='live', Symbol='Cdk2', Aliases='Dmel_CG10498, CDC2C, CDC2c, CDK2, CDK2/CDC2c, CG10498, Cdc2c, Dcdc2c, DmCdc2, DmCdk2, Dmcdc2c, Dmel\\CG10498, S(Sev-CycE)3A, cdc2c, cdk2, dCdk2', description='Cyclin-dependent kinase 2', other_designations='CG10498-PA|CG10498-PB|CG10498-PC|Cdk2-PA|Cdk2-PB|Cdk2-PC|cdc2c|cyclin E|cyclin E-CDK2|cyclin-dependent kinase 2', map_location='92F10-92F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20735623, end_position_on_the_genomic_accession=20738414, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42454, CurrentID=0, Status='live', Symbol='CG17267', Aliases='Dmel_ Dmel\\CG17267', description='CG17267 gene product from transcript CG17267-RC', other_designations='CG17267-PC|uncharacterized protein', map_location='92F10-92F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20737325, end_position_on_the_genomic_accession=20739100, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42455, CurrentID=0, Status='live', Symbol='CG42322', Aliases='Dmel_ CG15688, CG15689, CG15690, CG31203, Dmel\\CG42322, Dmel_CG15690, Dmel_CG31203', description='CG42322 gene product from transcript CG42322-RN', other_designations='CG42322-PD|CG42322-PI|CG42322-PJ|CG42322-PL|CG42322-PM|CG42322-PN|uncharacterized protein', map_location='92F12-92F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20740108, end_position_on_the_genomic_accession=20756639, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  42456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42456, CurrentID=0, Status='live', Symbol='CG31199', Aliases='Dmel_ CG17837, Dmel\\CG31199', description='CG31199 gene product from transcript CG31199-RA', other_designations='CG31199-PA|uncharacterized protein', map_location='92F10-92F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20742800, end_position_on_the_genomic_accession=20743869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42457, CurrentID=0, Status='live', Symbol='Prosalpha4T1', Aliases='Dmel_CG17268, CG17268, Dmel\\CG17268, Pros28.1A, Pros28.2, a4-t1, alpha4T1, alpha4t1, alpha4t1_dm, dalpha7, pros28.1A', description='Proteasome alpha4 subunit, Testis-specific 1', other_designations='20s proteasome|CG17268-PA|Pros28.1A|Prosalpha4T1-PA|proteasome 28kD subunit 1A|proteasome 28kD subunit 2|proteasome alpha4 subunit, Testis-specific 1|testes-specific proteasome subunit', map_location='92F-92F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20748184, end_position_on_the_genomic_accession=20749074, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42460, CurrentID=0, Status='live', Symbol='Syp', Aliases='Dmel_CG17838, AI945337, CG17838, Dmel\\CG17838, anon-WO0118547.613, cg17838, l(3)03806, syp', description='Syncrip', other_designations='CG17838-PA|CG17838-PB|CG17838-PC|CG17838-PD|CG17838-PE|CG17838-PF|CG17838-PH|CG17838-PI|CG17838-PJ|CG17838-PK|CG17838-PL|CG17838-PM|CG17838-PN|CG17838-PO|CG17838-PP|CG17838-PQ|CG17838-PR|Syp-PA|Syp-PB|Syp-PC|Syp-PD|Syp-PE|Syp-PF|Syp-PH|Syp-PI|Syp-PJ|Syp-PK|Syp-PL|Syp-PM|Syp-PN|Syp-PO|Syp-PP|Syp-PQ|Syp-PR|syncrip', map_location='92F12-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20756729, end_position_on_the_genomic_accession=20810726, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  42462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42462, CurrentID=0, Status='live', Symbol='CG17270', Aliases='Dmel_ Dmel\\CG17270, anon-Liu', description='CG17270 gene product from transcript CG17270-RA', other_designations='CG17270-PA|CG17270-PB|uncharacterized protein', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20816445, end_position_on_the_genomic_accession=20818929, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42463, CurrentID=0, Status='live', Symbol='CG17271', Aliases='Dmel_ Dm_3R:88623, Dmel\\CG17271', description='CG17271 gene product from transcript CG17271-RB', other_designations='CG17271-PA|CG17271-PB|uncharacterized protein', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20819129, end_position_on_the_genomic_accession=20821848, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42464, CurrentID=0, Status='live', Symbol='RpS20', Aliases='Dmel_CG15693, CG15693, Dmel\\CG15693, Rps20, anon-EST:Posey111, anon-EST:Posey122, anon-EST:Posey147', description='Ribosomal protein S20', other_designations='CG15693-PA|RpS20-PA|ribosomal protein S20', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20822181, end_position_on_the_genomic_accession=20823356, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42465, CurrentID=0, Status='live', Symbol='CG17272', Aliases='Dmel_ BEST:GH27755, Dmel\\CG17272', description='CG17272 gene product from transcript CG17272-RA', other_designations='CG17272-PA|uncharacterized protein', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20823313, end_position_on_the_genomic_accession=20824239, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42466, CurrentID=0, Status='live', Symbol='AdSS', Aliases='Dmel_CG17273, BcDNA:LD32788, CG17273, Dmel\\CG17273, PURA_DROME', description='Adenylosuccinate Synthetase', other_designations='AdSS-PA|CG17273-PA|adenylosuccinate synthetase', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20824667, end_position_on_the_genomic_accession=20830580, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42467, CurrentID=0, Status='live', Symbol='Synd', Aliases='Dmel_CG33094, CG15694, CG33094, D-Synd, Dmel\\CG33094, synd', description='Syndapin', other_designations='CG33094-PB|CG33094-PC|CG33094-PD|Synd-PB|Synd-PC|Synd-PD|syndapin', map_location='93A1-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20832799, end_position_on_the_genomic_accession=20840118, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42468, CurrentID=0, Status='live', Symbol='CG15695', Aliases='Dmel_ Dmel\\CG15695, gi21356013, gi7300665', description='CG15695 gene product from transcript CG15695-RA', other_designations='CG15695-PA|uncharacterized protein', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20842140, end_position_on_the_genomic_accession=20844981, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42469, CurrentID=0, Status='live', Symbol='CG15696', Aliases='Dmel_ Dmel\\CG15696', description='CG15696 gene product from transcript CG15696-RA', other_designations='CG15696-PA|uncharacterized protein', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20847068, end_position_on_the_genomic_accession=20847748, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42470, CurrentID=0, Status='live', Symbol='RpS30', Aliases='Dmel_CG15697, CG15697, Dmel\\CG15697, Q9VDH8_DROME, S30, anon-EST:Posey127', description='Ribosomal protein S30', other_designations='CG15697-PA|CG15697-PB|CG15697-PC|RpS30-PA|RpS30-PB|RpS30-PC|ribosomal protein S30', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20850516, end_position_on_the_genomic_accession=20851427, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42471, CurrentID=0, Status='live', Symbol='Ir93a', Aliases='Dmel_CG42315, CG17274, CG17275, CG42315, CT35907, DmelIR93a, Dmel\\CG42315, Dmel_CG17274, Dmel_CG17275, IR93a, ir93a', description='Ionotropic receptor 93a', other_designations='CG42315-PD|Ir93a-PD|ionotropic receptor 93a', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20851709, end_position_on_the_genomic_accession=20859945, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42473, CurrentID=0, Status='live', Symbol='CG3822', Aliases='Dmel_ CT12759, DKaiRID, DmelCG3822, Dmel\\CG3822', description='CG3822 gene product from transcript CG3822-RA', other_designations='CG3822-PA', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20860187, end_position_on_the_genomic_accession=20866231, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  42474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42474, CurrentID=0, Status='live', Symbol='TotB', Aliases='Dmel_CG5609, BcDNA:LP06769, CG5609, Dmel\\CG5609, Tot B, tot', description='Turandot B', other_designations='CG5609-PA|TotB-PA|turandot|turandot B', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20873938, end_position_on_the_genomic_accession=20874522, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42475, CurrentID=0, Status='live', Symbol='Ktl', Aliases='Dmel_CG10830, CG10830, Dmel\\CG10830', description='Kctd12-like', other_designations='CG10830-PA|CG10830-PB|Ktl-PA|Ktl-PB', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20894452, end_position_on_the_genomic_accession=20897320, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42476, CurrentID=0, Status='live', Symbol='CG5621', Aliases='Dmel_ CT17756, DKaiRIC, DmelCG5621, Dmel\\CG5621, anon-WO0138359.5', description='CG5621 gene product from transcript CG5621-RA', other_designations='CG5621-PA|CG5621-PC|CG5621-PD', map_location='93A2-93A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20901956, end_position_on_the_genomic_accession=20908263, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42477, CurrentID=0, Status='live', Symbol='CG31191', Aliases='Dmel_ CG17276, CG17277, Dmel\\CG31191', description='CG31191 gene product from transcript CG31191-RA', other_designations='CG31191-PA|CG31191-PD|uncharacterized protein', map_location='93A3-93A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20909213, end_position_on_the_genomic_accession=20945530, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42478, CurrentID=0, Status='live', Symbol='hdly', Aliases='Dmel_CG5630, CG5630, Dmel\\CG5630', description='hadley', other_designations='CG5630-PA|CG5630-PB|CG5630-PC|hdly-PA|hdly-PB|hdly-PC', map_location='93A3-93A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20914296, end_position_on_the_genomic_accession=20929213, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42481, CurrentID=0, Status='live', Symbol='Calx', Aliases='Dmel_CG5685, 9C5, CALX, CALX1.1, CG5685, CalX, Dmel/Ncx, Dmel\\CG5685, NCX, Ncx, anon-93ABa, calX, calx', description='Na/Ca-exchange protein', other_designations='CG5685-PA|CG5685-PB|CG5685-PC|CG5685-PD|CG5685-PE|CG5685-PF|CG5685-PG|CG5685-PH|CG5685-PI|Calx-PA|Calx-PB|Calx-PC|Calx-PD|Calx-PE|Calx-PF|Calx-PG|Calx-PH|Calx-PI', map_location='93A4-93B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20978183, end_position_on_the_genomic_accession=21015613, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  42482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42482, CurrentID=0, Status='live', Symbol='CG10827', Aliases='Dmel_ Dmel\\CG10827', description='CG10827 gene product from transcript CG10827-RA', other_designations='CG10827-PA|uncharacterized protein', map_location='93B3-93B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21005160, end_position_on_the_genomic_accession=21006990, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42483, CurrentID=0, Status='live', Symbol='CG5697', Aliases='Dmel_ BcDNA:RH32180, Dmel\\CG5697', description='CG5697 gene product from transcript CG5697-RA', other_designations='CG5697-PA|CG5697-PB|CG5697-PC|CG5697-PD', map_location='93B4-93B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21015949, end_position_on_the_genomic_accession=21017120, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42484, CurrentID=0, Status='live', Symbol='Rlip', Aliases='Dmel_CG11622, CG11622, D-RLIP, DRLIP, Dmel\\CG11622, RLIP, RLIP1, RhoGAP-93B3, dRLIP, dRalBP, dRlip', description='Ral interacting protein', other_designations='CG11622-PA|CG11622-PB|Rlip-PA|Rlip-PB', map_location='93B4-93B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21017684, end_position_on_the_genomic_accession=21020753, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42485, CurrentID=0, Status='live', Symbol='Dhc93AB', Aliases='Dmel_CG3723, CG3723, DHC, Dhc, Dhc 93AB, Dmel\\CG3723, dhc, dhc93AB', description='Dynein heavy chain at 93AB', other_designations='CG3723-PB|CG3723-PC|Dhc93AB-PB|Dhc93AB-PC|dynein heavy chain at 93AB|dynein-related heavy chain polypeptide', map_location='93B4-93B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21024580, end_position_on_the_genomic_accession=21042499, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  42486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42486, CurrentID=0, Status='live', Symbol='CG12278', Aliases='Dmel_ Dmel\\CG12278', description='CG12278 gene product from transcript CG12278-RA', other_designations='CG12278-PA|uncharacterized protein', map_location='93B7-93B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21044068, end_position_on_the_genomic_accession=21044905, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42487, CurrentID=0, Status='live', Symbol='CG31189', Aliases='Dmel_ CG7096, CG7096-C, Dmel\\CG31189', description='CG31189 gene product from transcript CG31189-RB', other_designations='CG31189-PB|uncharacterized protein', map_location='93B7-93B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21044898, end_position_on_the_genomic_accession=21045842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42488, CurrentID=0, Status='live', Symbol='CG7079', Aliases='Dmel_ Dmel\\CG7079, cg7079', description='CG7079 gene product from transcript CG7079-RB', other_designations='CG7079-PA|CG7079-PB', map_location='93B7-93B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21047462, end_position_on_the_genomic_accession=21048829, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42489, CurrentID=0, Status='live', Symbol='CG17279', Aliases='Dmel_ Dmel\\CG17279', description='CG17279 gene product from transcript CG17279-RC', other_designations='CG17279-PB|CG17279-PC|uncharacterized protein', map_location='93B7-93B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21050014, end_position_on_the_genomic_accession=21050943, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42490, CurrentID=0, Status='live', Symbol='Mvl', Aliases='Dmel_CG3671, CG3671, DMT1, Dmel\\CG3671, NRM2, dNRAMP, mvl', description='Malvolio', other_designations='/Divalent metal transporter|CG3671-PB|CG3671-PD|CG3671-PE|CG3671-PF|CG3671-PG|CG3671-PH|Mvl-PB|Mvl-PD|Mvl-PE|Mvl-PF|Mvl-PG|Mvl-PH|malvolio', map_location='93B7-93B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21051383, end_position_on_the_genomic_accession=21060800, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  42491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42491, CurrentID=0, Status='live', Symbol='Cortactin', Aliases='Dmel_CG3637, CG3637, Dmel\\CG3637, cortactin, cttn', description='CG3637 gene product from transcript CG3637-RD', other_designations='CG3637-PA|CG3637-PB|CG3637-PC|CG3637-PD|Cortactin-PA|Cortactin-PB|Cortactin-PC|Cortactin-PD|DCortactin|cortactin', map_location='93B8-93B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21061680, end_position_on_the_genomic_accession=21064471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42492, CurrentID=0, Status='live', Symbol='AnxB9', Aliases='Dmel_CG5730, ANXB9, Ann, AnnIX, Anx B9, CG5730, Dmel\\CG5730, annexin', description='Annexin B9', other_designations='AnxB9-PA|AnxB9-PB|AnxB9-PC|AnxB9-PD|AnxB9-PE|AnxB9-PF|CG5730-PA|CG5730-PB|CG5730-PC|CG5730-PD|CG5730-PE|CG5730-PF|annexin|annexin B9|annexin B9a|annexin IX', map_location='93B9-93B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21065241, end_position_on_the_genomic_accession=21070917, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42493, CurrentID=0, Status='live', Symbol='r-l', Aliases='Dmel_CG3593, CG3593, Dmel\\CG3593, OMPDC, OPRT, UMPS, ral', description='rudimentary-like', other_designations='CG3593-PA|UMP synthase|orotidine phosphate decarboxylase|r-l-PA', map_location='3-70.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21070557, end_position_on_the_genomic_accession=21073319, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42494, CurrentID=0, Status='live', Symbol='dmrt93B', Aliases='Dmel_CG5737, CG5737, Dmel\\CG5737, dmrt93b', description='doublesex-Mab related 93B', other_designations='CG5737-PA|dmrt93B-PA', map_location='93B10-93B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21074007, end_position_on_the_genomic_accession=21076798, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42495, CurrentID=0, Status='live', Symbol='HHEX', Aliases='Dmel_CG7056, CG7056, Dmel\\CG7056, dHHEX', description='HHEX ortholog (H. sapiens)', other_designations='CG7056-PA|HHEX ortholog|HHEX-PA', map_location='93B10-93B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21076862, end_position_on_the_genomic_accession=21080926, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42496, CurrentID=0, Status='live', Symbol='RhoGAP93B', Aliases='Dmel_CG3421, CG3421, CrGAP, Dmel\\CG3421, RhoGAP, Vilse, vilse', description='Rho GTPase activating protein at 93B', other_designations='CG3421-PA|RhoGAP93B-PA', map_location='93B10-93B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21083043, end_position_on_the_genomic_accession=21097827, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  42497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42497, CurrentID=0, Status='live', Symbol='CG7044', Aliases='Dmel_ Dmel\\CG7044', description='CG7044 gene product from transcript CG7044-RA', other_designations='CG7044-PA', map_location='93B11-93B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21098229, end_position_on_the_genomic_accession=21101718, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42498, CurrentID=0, Status='live', Symbol='CG5745', Aliases='Dmel_ Dmel\\CG5745', description='CG5745 gene product from transcript CG5745-RA', other_designations='CG5745-PA', map_location='93B12-93B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21102199, end_position_on_the_genomic_accession=21105113, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42499, CurrentID=0, Status='live', Symbol='Sec15', Aliases='Dmel_CG7034, 3R41, CG7034, Dmel\\CG7034p, dsec15, sec15, Sec15', description='Sec15 ortholog (S. cerevisiae)', other_designations='CG7034-PA|Sec15 ortholog|Sec15-PA|sec15', map_location='93B12-93B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21105046, end_position_on_the_genomic_accession=21107972, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42500, CurrentID=0, Status='live', Symbol='rtet', Aliases='Dmel_CG5760, BcDNA:LD28419, CG5760, Dmel\\CG5760, Drtet, anon-WO0119857.1, l(3)07086', description='tetracycline resistance', other_designations='CG5760-PA|rtet-PA', map_location='93B12-93B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21108446, end_position_on_the_genomic_accession=21110834, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42501, CurrentID=0, Status='live', Symbol='Rab11', Aliases='Dmel_CG5771, AAF55850, CG5771, DRAB11, D Dm Rab11, DmRab11, Dmel\\CG5771, Drab11, O18335, RAB11, Rab 11, Rab-11, Rab-r11, l(3)93Bi, l(3)j2D1, rab-11, rab11, rlr7', description='CG5771 gene product from transcript CG5771-RA', other_designations='CG5771-PA|CG5771-PB|Rab-protein 11|Rab11-PA|Rab11-PB|lethal(3)93Bi', map_location='93B12-93B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21111408, end_position_on_the_genomic_accession=21116146, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42502, CurrentID=0, Status='live', Symbol='ppan', Aliases='Dmel_CG5786, CG5786, Dmel\\CG5786, PPAN, Ppan, l(3)02231, l(3)j6B6', description='peter pan', other_designations='CG5786-PA|peterpan|ppan-PA', map_location='93B13-93B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21116254, end_position_on_the_genomic_accession=21117899, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42503, CurrentID=0, Status='live', Symbol='Bdbt', Aliases='Dmel_CG17282, CG17282, Dmel\\CG17282, bdbt', description='Bride of doubletime', other_designations='Bdbt-PA|CG17282-PA|bride of dbt|bride of doubletime', map_location='93B13-93B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21117857, end_position_on_the_genomic_accession=21118856, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42504, CurrentID=0, Status='live', Symbol='slmb', Aliases='Dmel_CG3412, BcDNA:GM02031, CG3412, Dmel\\CG3412, FBXW1, MENE (3R)-B, MENE(3R)-B, SLMB, Slimb, Slimb/Beta-TRCP1, Slimb/beta-TRCP, Slimb/betaTrCP, Slmb, crd, ica, l(3)00295, shv, slimb, wel', description='supernumerary limbs', other_designations='CG3412-PA|CG3412-PB|CG3412-PC|SLIMB|centrosome replication defective|icare|pingiel|shiva|slmb-PA|slmb-PB|slmb-PC|wellman', map_location='93B13-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21119252, end_position_on_the_genomic_accession=21127485, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42505, CurrentID=0, Status='live', Symbol='CG5793', Aliases='Dmel_ Dmel\\CG5793, FBgn0038858', description='CG5793 gene product from transcript CG5793-RE', other_designations='CG5793-PB|CG5793-PD|CG5793-PE', map_location='93C1-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21127573, end_position_on_the_genomic_accession=21129601, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42506, CurrentID=0, Status='live', Symbol='Obp93a', Aliases='Dmel_CG17284, CG17284, DmelObp93a, Dmel\\CG17284', description='Odorant-binding protein 93a', other_designations='CG17284-PA|Obp93a-PA', map_location='93C1-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21128950, end_position_on_the_genomic_accession=21129717, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42507, CurrentID=0, Status='live', Symbol='Ice2', Aliases='Dmel_CG10825, CG10825, Dmel\\CG10825', description='Interacts with the C terminus of ELL 1', other_designations='CG10825-PA|CG10825-PC|Ice2-PA|Ice2-PC|interacts with the C terminus of ELL 1', map_location='93C1-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21130216, end_position_on_the_genomic_accession=21132720, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42508, CurrentID=0, Status='live', Symbol='CG7009', Aliases='Dmel_ Dmel\\CG7009', description='CG7009 gene product from transcript CG7009-RA', other_designations='CG7009-PA', map_location='93C1-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21132734, end_position_on_the_genomic_accession=21133881, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42509, CurrentID=0, Status='live', Symbol='Usp8', Aliases='Dmel_CG5798, CG5798, CT18196, Dmel\\CG5798, Q9VDD8, UBPY, UBPY/USP8, USP8, UbpY, Ubpy, dUBPY, usp8', description='Ubiquitin specific protease 8', other_designations='CG5798-PA|CG5798-PB|UBPY ortholog|Usp8-PA|Usp8-PB|ubiquitin specific protease 8', map_location='93C1-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21134262, end_position_on_the_genomic_accession=21138214, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42510, CurrentID=0, Status='live', Symbol='meigo', Aliases='Dmel_CG5802, CG5802, Dmel\\CG5802, Meigo, NEST:bs03h03, anon-EST:fe1H6', description='medial glomeruli', other_designations='CG5802-PA|CG5802-PB|NEST:bs03h03|anon-fast-evolving-1H6|meigo-PA|meigo-PB', map_location='93C1-93C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21138270, end_position_on_the_genomic_accession=21140365, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42512, CurrentID=0, Status='live', Symbol='cDIP', Aliases='Dmel_CG10824, 10824, CG10824, CT11359, DIPc, Dmel\\CG10824', description='Common Dpr-interacting protein', other_designations='CG10824-PA|cDIP-PA|common DIP|common Dpr-interacting protein', map_location='93C2-93C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21159238, end_position_on_the_genomic_accession=21161092, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42513, CurrentID=0, Status='live', Symbol='CG5810', Aliases='Dmel_ CT18214, Dmel\\CG5810', description='CG5810 gene product from transcript CG5810-RA', other_designations='CG5810-PA', map_location='93C2-93C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21163778, end_position_on_the_genomic_accession=21165161, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42514, CurrentID=0, Status='live', Symbol='Snmp1', Aliases='Dmel_CG7000, CG7000, Dmel\\CG7000, SNMP, SNMP1, Snmp, snmp, snmp1', description='Sensory neuron membrane protein 1', other_designations='CG7000-PA|CG7000-PB|Snmp1-PA|Snmp1-PB|sensory neuron membrane protein|sensory neuron membrane protein 1', map_location='93C2-93C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21165167, end_position_on_the_genomic_accession=21168205, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42515, CurrentID=0, Status='live', Symbol='SNF4Agamma', Aliases='Dmel_CG17299, 0050/42, 0127/19, AMPK&ggr, AMPKgamma, AMPgamma, CG17299, CG5806, Dmel\\CG17299, EP3015b, FBgn0025803, SNF4, SNF4A, SNF4Ag, SNF4a, SNF4gamma, Snfg, Snfgamma, anon-WO0118547.338, anon-WO0118547.575, anon-WO0257455.21, l(3)S005042, l(3)S012719, loe, loeI, snf4', description='SNF4/AMP-activated protein kinase gamma subunit', other_designations='AMP-activated protein kinase gamma subunit|CG17299-PA|CG17299-PB|CG17299-PC|CG17299-PE|CG17299-PF|CG17299-PG|CG17299-PI|CG17299-PJ|CG17299-PK|CG17299-PM|CG17299-PN|CG17299-PO|CG17299-PP|CG17299-PQ|CG17299-PR|CG17299-PS|SNF4/AMP-activated protein kinase gamma|SNF4Agamma-PA|SNF4Agamma-PB|SNF4Agamma-PC|SNF4Agamma-PE|SNF4Agamma-PF|SNF4Agamma-PG|SNF4Agamma-PI|SNF4Agamma-PJ|SNF4Agamma-PK|SNF4Agamma-PM|SNF4Agamma-PN|SNF4Agamma-PO|SNF4Agamma-PP|SNF4Agamma-PQ|SNF4Agamma-PR|SNF4Agamma-PS|lethal (3) S005042|lethal (3) S012719|lochrig|loechrig', map_location='93C1-93C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21140739, end_position_on_the_genomic_accession=21214269, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  42516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42516, CurrentID=0, Status='live', Symbol='CG5862', Aliases='Dmel_ Dmel\\CG5862', description='CG5862 gene product from transcript CG5862-RA', other_designations='CG5862-PA', map_location='93C6-93C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21215079, end_position_on_the_genomic_accession=21216226, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42517, CurrentID=0, Status='live', Symbol='CG3353', Aliases='Dmel_ Dmel\\CG3353', description='CG3353 gene product from transcript CG3353-RA', other_designations='CG3353-PA', map_location='93C6-93C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21217268, end_position_on_the_genomic_accession=21218791, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42518, CurrentID=0, Status='live', Symbol='Oga', Aliases='Dmel_CG5871, CG5871, Dmel\\CG5871, GlcNAcase, MGEA5, OGA', description='O-GlcNAcase', other_designations='CG5871-PA|GlcNAcase|Oga-PA', map_location='93C6-93C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21219077, end_position_on_the_genomic_accession=21222843, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42519, CurrentID=0, Status='live', Symbol='CG3337', Aliases='Dmel_ Dmel\\CG3337, mat1', description='CG3337 gene product from transcript CG3337-RB', other_designations='CG3337-PA|CG3337-PB', map_location='93C7-93C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21220704, end_position_on_the_genomic_accession=21223584, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42520, CurrentID=0, Status='live', Symbol='Nelf-A', Aliases='Dmel_CG5874, BcDNA:SD07139, CG5874, Dmel\\CG5874, NELF, NELF-A', description='Negative elongation factor A', other_designations='CG5874-PA|CG5874-PB|CG5874-PC|NELF-A|Nelf-A-PA|Nelf-A-PB|Nelf-A-PC|negative elongation factor A', map_location='93C7-93C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21223704, end_position_on_the_genomic_accession=21229512, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42521, CurrentID=0, Status='live', Symbol='e', Aliases='Dmel_CG3331, CG3331, Dmel\\CG3331', description='ebony', other_designations='CG3331-PA|NBAD-synthase|e-PA', map_location='3-70.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21229839, end_position_on_the_genomic_accession=21237177, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42522, CurrentID=0, Status='live', Symbol='CG5892', Aliases='Dmel_ Dmel\\CG5892', description='CG5892 gene product from transcript CG5892-RB', other_designations='CG5892-PB', map_location='93D1-93D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21242618, end_position_on_the_genomic_accession=21245418, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42523, CurrentID=0, Status='live', Symbol='ETHR', Aliases='Dmel_CG5911, CG5911, CG5911a, CG5911b, Dm a, Dmel\\CG5911, anon-WO0170980.37, anon-WO0170980.38', description='CG5911 gene product from transcript CG5911-RB', other_designations='CG5911-PA|CG5911-PB|CG5911-PC|ETH receptor|ETHR-PA|ETHR-PB|ETHR-PC|ecdysis triggering hormone receptor', map_location='93D1-93D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21247011, end_position_on_the_genomic_accession=21265134, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42524, CurrentID=0, Status='live', Symbol='Rab1', Aliases='Dmel_CG3320, AAF55873, CG3320, DRAB1, DRab1, Dm Rab1, DmRab1, Dmel\\CG3320, O18332, RAB1aa, dRab1, dar6, drab1, rab1', description='CG3320 gene product from transcript CG3320-RA', other_designations='CG3320-PA|Rab-protein 1|Rab1-PA|dendritic arbor reduction 6', map_location='93D2-93D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21266029, end_position_on_the_genomic_accession=21269060, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42525, CurrentID=0, Status='live', Symbol='AP-2sigma', Aliases='Dmel_CG6056, AP-2 sigma, AP2, CG6056, Dmel\\CG6056, aps2, cg6056, sigma2', description='Adaptor Protein complex 2, sigma subunit', other_designations='AP-2sigma-PA|CG6056-PA|adaptor protein complex 2, sigma subunit', map_location='93D2-93D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21269209, end_position_on_the_genomic_accession=21270075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42526, CurrentID=0, Status='live', Symbol='CG5919', Aliases='Dmel_ Dmel\\CG5919, IPPI', description='CG5919 gene product from transcript CG5919-RA', other_designations='CG5919-PA|CG5919-PB', map_location='93D2-93D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21270126, end_position_on_the_genomic_accession=21271408, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42527, CurrentID=0, Status='live', Symbol='CG3308', Aliases='Dmel_ Dmel\\CG3308', description='CG3308 gene product from transcript CG3308-RA', other_designations='CG3308-PA', map_location='93D2-93D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21271078, end_position_on_the_genomic_accession=21273817, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42528, CurrentID=0, Status='live', Symbol='CG3301', Aliases='Dmel_ Dmel\\CG3301, FOHSDR', description='CG3301 gene product from transcript CG3301-RB', other_designations='CG3301-PA|CG3301-PB|CG3301-PC', map_location='93D2-93D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21274243, end_position_on_the_genomic_accession=21276249, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42529, CurrentID=0, Status='live', Symbol='CG17298', Aliases='Dmel_ BcDNA:RH37844, CT35916, Dmel\\CG17298', description='CG17298 gene product from transcript CG17298-RA', other_designations='CG17298-PA|uncharacterized protein', map_location='93D3-93D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21283197, end_position_on_the_genomic_accession=21284050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42530, CurrentID=0, Status='live', Symbol='SIFaR', Aliases='Dmel_CG10823, CG10823, Dmel\\CG10823, DrmSIFR, SIF-R, SIFR, SIFa-R, anon-WO0170980.43, anon-WO0170980.44', description='SIFamide receptor', other_designations='CG10823-PA|CG10823-PC|SIFaR-PA|SIFaR-PC|SIFamide-receptor', map_location='93D2-93D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21276664, end_position_on_the_genomic_accession=21293998, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42531, CurrentID=0, Status='live', Symbol='CG16791', Aliases='Dmel_ Dmel\\CG16791, anon-WO0172774.78', description='CG16791 gene product from transcript CG16791-RD', other_designations='CG16791-PA|CG16791-PB|CG16791-PC|CG16791-PD|uncharacterized protein', map_location='93D5-93D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21318300, end_position_on_the_genomic_accession=21349601, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  42536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42536, CurrentID=0, Status='live', Symbol='tin', Aliases='Dmel_CG7895, CG7895, DROHOXHK4, DROHOXNK4, DmNK-4, Dmel\\CG7895, HOX, NK-4, NK-4/msh-2, NK4, NK4/msh-2, NKX2.5, TIN, Tin, msh-2, msh2/NK4, tin', description='tinman', other_designations='CG7895-PA|muscle-specific-homeodomain-2|tin-PA', map_location='93D9-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21378977, end_position_on_the_genomic_accession=21381970, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42537, CurrentID=0, Status='live', Symbol='bap', Aliases='Dmel_CG7902, BAP, Bap, CG7902, Dmel\\CG7902, NK-3, NK3, bgp, dNK-3, nk-3', description='bagpipe', other_designations='CG7902-PA|bap-PA', map_location='93D10-93D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21388880, end_position_on_the_genomic_accession=21390449, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42538, CurrentID=0, Status='live', Symbol='CG6475', Aliases='Dmel_ Dmel\\CG6475', description='CG6475 gene product from transcript CG6475-RD', other_designations='CG6475-PD', map_location='93D10-93E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21397781, end_position_on_the_genomic_accession=21399742, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42539, CurrentID=0, Status='live', Symbol='CG7907', Aliases='Dmel_ Dmel\\CG7907', description='CG7907 gene product from transcript CG7907-RB', other_designations='CG7907-PB|CG7907-PC', map_location='93E1-93E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21403735, end_position_on_the_genomic_accession=21406690, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42540, CurrentID=0, Status='live', Symbol='CheB93a', Aliases='Dmel_CG15503, CG15503, Dmel\\CG15503', description='Chemosensory protein B 93a', other_designations='CG15503-PA|CheB93a-PA|chemosensory protein B 93a', map_location='93E1-93E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21407445, end_position_on_the_genomic_accession=21408324, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42541, CurrentID=0, Status='live', Symbol='lbl', Aliases='Dmel_CG6570, CG6570, Dmel\\CG6570, Lb, Lbl, NK?, NKch4, lb, nkch4', description='ladybird late', other_designations='CG6570-PA|CG6570-PB|CG6570-PC|NK-cluster homeobox 4|lady-bird-late|ladybird|lbl-PA|lbl-PB|lbl-PC', map_location='93E1-93E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21409979, end_position_on_the_genomic_accession=21434133, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42542, CurrentID=0, Status='live', Symbol='lbe', Aliases='Dmel_CG6545, CG6545, Dmel\\CG6545, Hox11-D125, Lb, Lbe, NK?, lb', description='ladybird early', other_designations='CG6545-PA|lady-bird-early|ladybird|lbe-PA', map_location='93E1-93E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21441577, end_position_on_the_genomic_accession=21446537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42543, CurrentID=0, Status='live', Symbol='Fancm', Aliases='Dmel_CG7922, CG7922, Dmel\\CG7922, cg7922', description='Fanconi anemia group M helicase', other_designations='CG7922-PA|CG7922-PB|Fancm-PA|Fancm-PB|fanconi anemia group M (FANCM) helicase|fanconi anemia group M helicase', map_location='93E2-93E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21482325, end_position_on_the_genomic_accession=21487444, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42544, CurrentID=0, Status='live', Symbol='C15', Aliases='Dmel_CG7937, 311, 93Bal, CG7937, Cll, Dmel\\CG7937, Ect5, Hox11-311, c15, cll', description='CG7937 gene product from transcript CG7937-RA', other_designations='C15-PA|CG7937-PA|clawless', map_location='93E3-93E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21499986, end_position_on_the_genomic_accession=21512814, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3354888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354888, CurrentID=0, Status='live', Symbol='rl', Aliases='Dmel_CG12559, 12559, BcDNA:RE08694, CG12559, CG18732, CT34260, CT39192, D-ERK, DERK, DERK-A, DmERK-A, DmERKA, DmErk, DmMAPK, Dmel\\CG12559, DpERK, DpErk, Dsor2, E(sina)7, EC2-1, EK2-1, ERK, ERK-A, ERKA, ERKa, EY2-2, Erk, Erk/Map kinase, Erk1, ErkA, GroupII, MAP-k, MAPK, MapK, Mapk, RL, Rl, SEM, SR2-1, Sem, Su(Raf)2B, dERK, dm-dpERK, dp-ERK, dpERK, dpERK1, dpERk, dpErk, dpMAPK, erk, l(2)41Ac, l(2R)EMS45-39, mapk, pERK, pMAPK, pMapK/ERK, rl/MAPK, rll, sem, rl', description='rolled', other_designations='CG12559-PA|CG12559-PC|CG12559-PD|CG12559-PE|CG12559-PG|CG12559-PH|CG12559-PI|Diphospho-ERK|ERK|ERK/rolled|Erk MAP kinase|Extracellular-regulated kinase A|Extracellular-signal-related kinase A|MAP kinase|MAP-kinase|drosophila ERK|enhancer of seven in absentia 7|extracellular signal-regulated kinase|extracellular signal-regulated kinase A|mitogen activated protein kinase|mitogen-activated protein kinase|phospho-ERK|rl-PA|rl-PC|rl-PD|rl-PE|rl-PG|rl-PH|rl-PI|roll|sevenmaker', map_location='2-55.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1071462, end_position_on_the_genomic_accession=1125927, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42547, CurrentID=0, Status='live', Symbol='slou', Aliases='Dmel_CG6534, CG6534, Dmel\\CG6534, NK-1, NK1, NK1/S59, Nk1, S-59, S59, S59/NK-1, Slou, prd9, s59/NK1, slou', description='slouch', other_designations='CG6534-PA|CG6534-PB|PRD gene 9|paired-like 9|slou-PA|slou-PB', map_location='93E4-93E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21548661, end_position_on_the_genomic_accession=21557182, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42548, CurrentID=0, Status='live', Symbol='CG15498', Aliases='Dmel_ Dmel\\CG15498', description='CG15498 gene product from transcript CG15498-RA', other_designations='CG15498-PA|uncharacterized protein', map_location='93E4-93E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21567519, end_position_on_the_genomic_accession=21569149, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42549, CurrentID=0, Status='live', Symbol='InR', Aliases='Dmel_CG18402, 18402, CG18402, DIHR, DILR, DIR, DIRH, DIRbeta, DInR, DInr, Dir-a, Dir-b, Dmel\\CG18402, INR, INS, IR, Inr, Inr-alpha, Inr-beta, InsR, dINR, dIR, dIRH, dInR, dInr, dInsR, dinr, dir, er10, inR, inr, insulin/insulin-like growth factor receptor, l(3)05545, l(3)93Dj, l(3)er10, lnR', description='Insulin-like receptor', other_designations='CG18402-PA|CG18402-PB|CG18402-PC|CG18402-PD|InR-PA|InR-PB|InR-PC|InR-PD|Insulin-receptor|drosophila insulin receptor|insulin|insulin receptor|insulin receptor homolog|insulin receptor homologue|insulin-like receptor|lethal(3)93Dj', map_location='93E4-93E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21570248, end_position_on_the_genomic_accession=21619321, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  42550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42550, CurrentID=0, Status='live', Symbol='E2f1', Aliases='Dmel_CG6376, CG6376, DRTF1/E2F, DmE2F-1, Dmel\\CG6376, Dp, E(Sev-CycE)3A, E(var)3-93E, E(var)3-95E, E(var)93E, E2F, E2F-1, E2F1, E2f, Evar(3)164, dE2F, dE2F1, dE2f, dE2f1, de2f1, def21, drosE2F1, e2f, e2f1, l(3)07172, l(3)j3B1, l(3)j3C2, l(3)rM729', description='E2F transcription factor 1', other_designations='CG6376-PA|CG6376-PB|CG6376-PC|CG6376-PD|CG6376-PE|CG6376-PF|E2-promoter binding factor|E2F transcription factor|E2f1-PA|E2f1-PB|E2f1-PC|E2f1-PD|E2f1-PE|E2f1-PF', map_location='93E9-93F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21620304, end_position_on_the_genomic_accession=21660871, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  42551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42551, CurrentID=0, Status='live', Symbol='CG6353', Aliases='Dmel_ BcDNA:SD28072, Dmel\\CG6353', description='CG6353 gene product from transcript CG6353-RA', other_designations='CG6353-PA', map_location='93F1-93F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21661186, end_position_on_the_genomic_accession=21662972, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42552, CurrentID=0, Status='live', Symbol='CG15497', Aliases='Dmel_ Dmel\\CG15497', description='CG15497 gene product from transcript CG15497-RA', other_designations='CG15497-PA|uncharacterized protein', map_location='93F2-93F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21664389, end_position_on_the_genomic_accession=21666676, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42553, CurrentID=0, Status='live', Symbol='DNApol-alpha180', Aliases='Dmel_CG6349, 183 kDa pol alpha, AAB24152, CG6349, DNA Pol-alpha180, DNA polalpha, DNA polymerase alpha 180 kD, DNA-Polalpha, DNApol, DNApol-alpha, DNApola, DNApolalpha, DNApolalpha180, DmPolalpha, Dmel\\CG6349, Dmpol alpha, E(mus304), POLA, POlalpha 180kDa, Pol, Pol alpha, Polalpha, pol alpha, polalpha, polalphap180', description='DNA polymerase alpha 180kD', other_designations='180 kDa subunit|180 kDa subunit of polymerase-primase|CG6349-PA|DNA pol 180|DNA polymerase 180K|DNA polymerase alpha|DNA polymerase alpha 148,000 alpha subunit|DNA polymerase alpha 180 kD|DNA polymerase alpha 182,000 alpha subunit|DNA polymerase-primase 180 kda subunit|DNA-polymerase alpha|DNApol-alpha180-PA|enhancer of mus304', map_location='93F2-93F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21666562, end_position_on_the_genomic_accession=21671750, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42554, CurrentID=0, Status='live', Symbol='CG31176', Aliases='Dmel_ CG15495, CG6328, Dmel\\CG31176, EP(3)3662, EP3662, RE56061', description='CG31176 gene product from transcript CG31176-RA', other_designations='CG31176-PA|uncharacterized protein', map_location='93F2-93F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21672303, end_position_on_the_genomic_accession=21719660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42556, CurrentID=0, Status='live', Symbol='CG5849', Aliases='Dmel_ Dmel\\CG5849', description='CG5849 gene product from transcript CG5849-RB', other_designations='CG5849-PA|CG5849-PB', map_location='93F7-93F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21740105, end_position_on_the_genomic_accession=21744431, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42557, CurrentID=0, Status='live', Symbol='CG31233', Aliases='Dmel_ CG5839, Dmel\\CG31233, anon-WO0140519.125, anon-WO0140519.65', description='CG31233 gene product from transcript CG31233-RB', other_designations='CG31233-PA|CG31233-PB|uncharacterized protein', map_location='93F8-93F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21749853, end_position_on_the_genomic_accession=21753237, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42558, CurrentID=0, Status='live', Symbol='CG42335', Aliases='Dmel_ CG13420, CG31177, CG5845, Dmel\\CG42335, Dmel_CG13420, Dmel_CG31177', description='CG42335 gene product from transcript CG42335-RE', other_designations='CG42335-PE|uncharacterized protein', map_location='93F9-93F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21758764, end_position_on_the_genomic_accession=21769301, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42560, CurrentID=0, Status='live', Symbol='Burs', Aliases='Dmel_CG13419, BURS, Bur alpha alpha, Bursalpha, CG13419, Dmel\\CG13419, burs, burs alpha, burs-alpha, bursalpha, frl, Burs', description='Bursicon', other_designations='Burs-PA|CG13419-PA|bursicon|bursicon alpha|bursicon alpha subunit|bursicon alpha-subunit|furled wing', map_location='93F9-93F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21769518, end_position_on_the_genomic_accession=21770279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42561, CurrentID=0, Status='live', Symbol='GABA-B-R2', Aliases='Dmel_CG6706, BcDNA:GH07312, CG6706, D-GABA[[B]]R2, D-Gaba2, Dmel\\CG6706, GABA B R2, GABA-[[B]]R2, GABABR2, GABA[[B]]-R2, GABA[[B]]R2, GB2, GBR, GH07312, anon-WO0170980.58, anon-WO0170980.59', description='metabotropic GABA-B receptor subtype 2', other_designations='CG6706-PA|CG6706-PB|CG6706-PC|GABA-B-R2-PA|GABA-B-R2-PB|GABA-B-R2-PC|GABA[[B]]- 2 receptor|GABA[[B]]receptor2|metabotropic GABA[[B]] receptor 2', map_location='93F9-93F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21770333, end_position_on_the_genomic_accession=21775840, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  42562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42562, CurrentID=0, Status='live', Symbol='CG6800', Aliases='Dmel_ AC014407, Dmel\\CG6800', description='CG6800 gene product from transcript CG6800-RA', other_designations='CDK7-like|CG6800-PA', map_location='93F9-93F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21776138, end_position_on_the_genomic_accession=21777213, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42563, CurrentID=0, Status='live', Symbol='RpI12', Aliases='Dmel_CG13418, CG13418, Dmel\\CG13418', description='CG13418 gene product from transcript CG13418-RA', other_designations='CG13418-PA|RpI12-PA', map_location='93F9-93F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21777203, end_position_on_the_genomic_accession=21777847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42564, CurrentID=0, Status='live', Symbol='tsl', Aliases='Dmel_CG6705, CG6705, Dm-tsl, Dmel\\CG6705, Tsl, fs(3)00617', description='torso-like', other_designations='CG6705-PA|CG6705-PB|CG6705-PC|Torso-like|fs(3)torso-like|torso like|torsolike|tsl-PA|tsl-PB|tsl-PC', map_location='93F9-93F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21778114, end_position_on_the_genomic_accession=21782536, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42567, CurrentID=0, Status='live', Symbol='CASK', Aliases='Dmel_CG6703, CAKI, CAMGUK, CG13412, CG13413, CG6703, CMG, CMG/CASK, Caki, CamGUK, CamguK, Camguk, Cask, Cmg, DLin-2, Dmel\\CG6703, RE09582p, caki, camguk, cask, cmg, dCASK', description='CASK ortholog', other_designations='CASK-PA|CASK-PB|CASK-PD|CASK-PE|CASK-PF|CASK-PG|CASK-PH|CG6703-PA|CG6703-PB|CG6703-PD|CG6703-PE|CG6703-PF|CG6703-PG|CG6703-PH|Calcium/calmodulin-dependent protein kinase|Camguk/CASK adaptor protein|calmodulin-dependent kinase|camguk', map_location='93F10-93F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21783400, end_position_on_the_genomic_accession=21822674, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  42568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42568, CurrentID=0, Status='live', Symbol='Gr93a', Aliases='Dmel_CG13417, 93F.1, CG13417, Dmel\\CG13417, GR93, GR93F.1, GR93a, Gr93F1', description='Gustatory receptor 93a', other_designations='CG13417-PA|Gr93a-PA|gustatory receptor 93F.1|gustatory receptor 93a', map_location='93F13-93F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21832503, end_position_on_the_genomic_accession=21834013, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42571, CurrentID=0, Status='live', Symbol='glec', Aliases='Dmel_CG6575, CG6575, Dmel\\CG6575', description='gliolectin', other_designations='CG6575-PA|glec-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21850694, end_position_on_the_genomic_accession=21856266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42572, CurrentID=0, Status='live', Symbol='lsn', Aliases='Dmel_CG6637, CG6637, Dmel\\CG6637, Lsn, Vps22, vps22', description='larsen', other_designations='CG6637-PA|lsn-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21865168, end_position_on_the_genomic_accession=21866238, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42573, CurrentID=0, Status='live', Symbol='CG6569', Aliases='Dmel_ Dmel\\CG6569', description='CG6569 gene product from transcript CG6569-RA', other_designations='CG6569-PA|CG6569-PB|CG6569-PC|CG6569-PD', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21866172, end_position_on_the_genomic_accession=21867919, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42574, CurrentID=0, Status='live', Symbol='Cby', Aliases='Dmel_CG13415, CG13415, Dmel\\CG13415, cby', description='Chibby', other_designations='CG13415-PC|CG13415-PE|Cby-PC|Cby-PE|chibby', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21868003, end_position_on_the_genomic_accession=21868770, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42575, CurrentID=0, Status='live', Symbol='CG31178', Aliases='Dmel_ BEST:LP09246, BcDNA:LP09246, CG13414, CT32770, Dmel\\CG31178, Dromel_CG31178_FBtr0084138_mORF, Tes87', description='CG31178 gene product from transcript CG31178-RB', other_designations='CG31178-PB|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21873739, end_position_on_the_genomic_accession=21874922, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42576, CurrentID=0, Status='live', Symbol='CG6656', Aliases='Dmel_ Dmel\\CG6656', description='CG6656 gene product from transcript CG6656-RA', other_designations='CG6656-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21875293, end_position_on_the_genomic_accession=21877788, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42577, CurrentID=0, Status='live', Symbol='CG31465', Aliases='Dmel_ CG6670, Dmel\\CG31465', description='CG31465 gene product from transcript CG31465-RB', other_designations='CG31465-PA|CG31465-PB|CG31465-PC|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21880810, end_position_on_the_genomic_accession=21883990, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42578, CurrentID=0, Status='live', Symbol='fit', Aliases='Dmel_CG17820, CG17820, Dmel\\CG17820', description='female-specific independent of transformer', other_designations='CG17820-PA|female sterile independent of transformer|fit-PA|independent of transformer', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21883924, end_position_on_the_genomic_accession=21884454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42579, CurrentID=0, Status='live', Symbol='CG17819', Aliases='Dmel_ BcDNA:LP07781, Dmel\\CG17819', description='CG17819 gene product from transcript CG17819-RA', other_designations='CG17819-PA|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21884728, end_position_on_the_genomic_accession=21885550, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42580, CurrentID=0, Status='live', Symbol='dnd', Aliases='Dmel_CG6560, ARL3, Arl3, CG6560, Dmel\\CG6560, Dnd, Q9VD64', description='dead end', other_designations='ADP-ribosylation factor-like 3|Arf-like 3|Arf-like-3|Arf-like3|CG6560-PA|dnd-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21885676, end_position_on_the_genomic_accession=21887355, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42581, CurrentID=0, Status='live', Symbol='CG6678', Aliases='Dmel_ Dmel\\CG6678', description='CG6678 gene product from transcript CG6678-RA', other_designations='CG6678-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21887029, end_position_on_the_genomic_accession=21907182, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42582, CurrentID=0, Status='live', Symbol='CG6690', Aliases='Dmel_ Dmel\\CG6690', description='CG6690 gene product from transcript CG6690-RA', other_designations='CG6690-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21911816, end_position_on_the_genomic_accession=21913591, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42583, CurrentID=0, Status='live', Symbol='CG17843', Aliases='Dmel_ DmQSOX2, Dmel\\CG17843, Erv1, anon-WO0140519.66', description='CG17843 gene product from transcript CG17843-RA', other_designations='CG17843-PA|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21913747, end_position_on_the_genomic_accession=21915546, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42585, CurrentID=0, Status='live', Symbol='CG6332', Aliases='Dmel_ Dmel\\CG6332, anon-WO0140519.231, cg6332', description='CG6332 gene product from transcript CG6332-RA', other_designations='CG6332-PA', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22014697, end_position_on_the_genomic_accession=22016212, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42586, CurrentID=0, Status='live', Symbol='CG6439', Aliases='Dmel_ Dmel\\CG6439, FBgn0038922, Idh', description='CG6439 gene product from transcript CG6439-RA', other_designations='CG6439-PA|CG6439-PB|isocitrate dehydrogenase', map_location='93F14-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22016391, end_position_on_the_genomic_accession=22019382, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42587, CurrentID=0, Status='live', Symbol='Mitofilin', Aliases='Dmel_CG6455, CG 6455, CG6455, Dmel\\CG6455', description='CG6455 gene product from transcript CG6455-RC', other_designations='CG6455-PC|CG6455-PD|CG6455-PE|Mitofilin-PC|Mitofilin-PD|Mitofilin-PE|Motor-protein|dmitofilin|mitofilin|motor protein', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22019505, end_position_on_the_genomic_accession=22023335, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42588, CurrentID=0, Status='live', Symbol='mRpL35', Aliases='Dmel_CG13410, BcDNA:RE35766, CG13410, Dmel\\CG13410, MRP-L35', description='mitochondrial ribosomal protein L35', other_designations='CG13410-PA|mRpL35-PA', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22023527, end_position_on_the_genomic_accession=22024341, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42589, CurrentID=0, Status='live', Symbol='CG6028', Aliases='Dmel_ Dmel\\CG6028', description='CG6028 gene product from transcript CG6028-RA', other_designations='CG6028-PA', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22024259, end_position_on_the_genomic_accession=22025506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42590, CurrentID=0, Status='live', Symbol='Cchl', Aliases='Dmel_CG6022, CG6022, Dmel\\CG6022, cchL, cchl', description='Cytochrome c heme lyase', other_designations='CG6022-PA|CG6022-PB|Cchl-PA|Cchl-PB|cytochrome c haem lyase|cytochrome c heme lyase', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22025975, end_position_on_the_genomic_accession=22027480, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42591, CurrentID=0, Status='live', Symbol='ND-42', Aliases='Dmel_CG6343, A10, CG6343, Dmel\\CG6343, ND42, NUDM, NdufA10', description='NADH dehydrogenase (ubiquinone) 42 kDa subunit', other_designations='CG6343-PA|CG6343-PB|CG6343-PC|NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 10|NADH ubiquinone oxidoreductase 42 kDa|NADH-ubiquinone oxidoreductase 42 kDa subunit|NADH-ubiquinone oxidoreductase subunit|NADH: UBIQUINONE REDUCTASE 42 KD SUBUNIT PRECURSOR|NADH:ubiquinone oxidoreductase 42 kDa subunit|NADH:ubiquinone reductase 42 kD|NADH:ubiquinone reductase 42kD subunit precursor|ND-42-PA|ND-42-PB|ND-42-PC', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22027709, end_position_on_the_genomic_accession=22029393, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42592, CurrentID=0, Status='live', Symbol='CG13409', Aliases='Dmel_ Dmel\\CG13409', description='CG13409 gene product from transcript CG13409-RA', other_designations='CG13409-PA|uncharacterized protein', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22029440, end_position_on_the_genomic_accession=22031900, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  45328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45328, CurrentID=0, Status='live', Symbol='sens', Aliases='Dmel_CG32120, 1228/04, CG10714, CG32120, Dmel\\CG32120, Gfi1, Ly, Ly-1, SENS, Sen, Sens, Ss, l(3)70Ad, sene, sens', description='senseless', other_designations='CG32120-PA|lyra|sens-PA|sensesless|sensless', map_location='3-40.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13396228, end_position_on_the_genomic_accession=13401125, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42594, CurrentID=0, Status='live', Symbol='Fadd', Aliases='Dmel_CG12297, BG4, BG4_DROME, CG12297, Dmel\\CG12297, FADD, dFADD, dFadd, dfadd, fadd', description='Fas-associated death domain ortholog', other_designations='BG4|CG12297-PA|Fadd-PA|Fas-associated death domain', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22034455, end_position_on_the_genomic_accession=22035617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42595, CurrentID=0, Status='live', Symbol='pit', Aliases='Dmel_CG6375, CG6375, DmRH15, Dmel\\CG6375, Hel93F, PIT, Pit, hel93F', description='pitchoune', other_designations='CG6375-PA|CG6375-PB|CG6375-PC|helicase at 93F|pit-PA|pit-PB|pit-PC', map_location='93F13-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22034298, end_position_on_the_genomic_accession=22038925, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42596, CurrentID=0, Status='live', Symbol='how', Aliases='Dmel_CG10293, 0904/17, 24B, CG10293, Dmel\\CG10293, HOW, How, KH93F, P62, SZ1, Who/How, anon-EST:Liang-2.39, clone 2.39, l(3)S053606, l(3)S090417, l(3)j5B5, l(3)j5D5, l(3)s2612, qkr, qkr[93F], sci, stru, who', description='held out wings', other_designations='CG10293-PA|CG10293-PB|CG10293-PC|CG10293-PD|CG10293-PE|CG10293-PF|Held-out wings|KH motif at 93F|held Out wing|how-PA|how-PB|how-PC|how-PD|how-PE|how-PF|lethal (3) S053606|lethal (3) S090417|quaking related, 93F|scorpion|struthio|wings held out|wings held-out', map_location='94A1-87F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22042191, end_position_on_the_genomic_accession=22079321, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42597, CurrentID=0, Status='live', Symbol='CG13408', Aliases='Dmel_ Dmel\\CG13408', description='CG13408 gene product from transcript CG13408-RA', other_designations='CG13408-PA|uncharacterized protein', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22080382, end_position_on_the_genomic_accession=22083473, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42598, CurrentID=0, Status='live', Symbol='CG5778', Aliases='Dmel_ BcDNA:GH13168, CT18134, Dmel\\CG5778', description='CG5778 gene product from transcript CG5778-RA', other_designations='CG5778-PA|CG5778-PC', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22085374, end_position_on_the_genomic_accession=22086397, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42599, CurrentID=0, Status='live', Symbol='CG13407', Aliases='Dmel_ Dmel\\CG13407', description='CG13407 gene product from transcript CG13407-RA', other_designations='CG13407-PA|uncharacterized protein', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22103414, end_position_on_the_genomic_accession=22105590, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42600, CurrentID=0, Status='live', Symbol='CG42390', Aliases='Dmel_ CG33334, CG5740, Dmel\\CG42390, Dmel_CG33334, Dmel_CG5740', description='CG42390 gene product from transcript CG42390-RE', other_designations='CG42390-PB|CG42390-PD|CG42390-PE|uncharacterized protein', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22120061, end_position_on_the_genomic_accession=22133367, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  42601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42601, CurrentID=0, Status='live', Symbol='SKIP', Aliases='Dmel_CG31163, CG13863, CG13864, CG31163, CG5736, CG7086, CG7087, Dmel\\CG31163', description='Shal K[+] channel interacting protein', other_designations='CG31163-PB|CG31163-PE|CG31163-PF|CG31163-PG|CG31163-PH|CG31163-PI|CG31163-PJ|SKIP-PB|SKIP-PE|SKIP-PF|SKIP-PG|SKIP-PH|SKIP-PI|SKIP-PJ|shal K[+] channel interacting protein|skal K[+] channel interacting protein', map_location='94A2-94A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22133830, end_position_on_the_genomic_accession=22296126, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  42602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42602, CurrentID=0, Status='live', Symbol='Gld2', Aliases='Dmel_CG5732, CG5732, Dm1, DmCG5732, DmGLD2, Dmel\\CG5732', description='CG5732 gene product from transcript CG5732-RC', other_designations='CG5732-PB|CG5732-PC|Gld2-PB|Gld2-PC', map_location='94A3-94A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22211824, end_position_on_the_genomic_accession=22217415, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42606, CurrentID=0, Status='live', Symbol='CG7084', Aliases='Dmel_ Dmel\\CG7084', description='CG7084 gene product from transcript CG7084-RF', other_designations='CG7084-PB|CG7084-PC|CG7084-PE|CG7084-PF', map_location='94A4-94A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22298805, end_position_on_the_genomic_accession=22304420, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42608, CurrentID=0, Status='live', Symbol='CG34377', Aliases='Dmel_ CG13858, CG13859, CG13860, CG13861, Dmel\\CG34377, Dmel_CG13858, Dmel_CG13859, Dmel_CG13860, Dmel_CG13861', description='CG34377 gene product from transcript CG34377-RB', other_designations='CG34377-PA|CG34377-PB|CG34377-PC|CG34377-PD|CG34377-PE|uncharacterized protein', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22308663, end_position_on_the_genomic_accession=22324094, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42609, CurrentID=0, Status='live', Symbol='CG7080', Aliases='Dmel_ Dmel\\CG7080', description='CG7080 gene product from transcript CG7080-RA', other_designations='CG7080-PA', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22329300, end_position_on_the_genomic_accession=22330546, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42610, CurrentID=0, Status='live', Symbol='CG13862', Aliases='Dmel_ Dmel\\CG13862', description='CG13862 gene product from transcript CG13862-RA', other_designations='CG13862-PA|uncharacterized protein', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22331991, end_position_on_the_genomic_accession=22332559, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42611, CurrentID=0, Status='live', Symbol='CG5391', Aliases='Dmel_ Dmel\\CG5391', description='CG5391 gene product from transcript CG5391-RA', other_designations='CG5391-PA', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22333088, end_position_on_the_genomic_accession=22334150, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42612, CurrentID=0, Status='live', Symbol='CG5388', Aliases='Dmel_ Dmel\\CG5388', description='CG5388 gene product from transcript CG5388-RB', other_designations='CG5388-PA|CG5388-PB', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22334257, end_position_on_the_genomic_accession=22340307, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42613, CurrentID=0, Status='live', Symbol='CG5386', Aliases='Dmel_ Dmel\\CG5386', description='CG5386 gene product from transcript CG5386-RA', other_designations='CG5386-PA', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22341236, end_position_on_the_genomic_accession=22342251, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42614, CurrentID=0, Status='live', Symbol='rdhB', Aliases='Dmel_CG7077, CG7077, Dmel\\CG7077', description='retinol dehydrogenase B', other_designations='CG7077-PA|CG7077-PB|rdhB-PA|rdhB-PB', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22351696, end_position_on_the_genomic_accession=22356264, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42615, CurrentID=0, Status='live', Symbol='Sar1', Aliases='Dmel_CG7073, 5712, CG7073, Dm Sar1, Dmel\\CG7073, NP_732719, Q95SY7, SAR1, anon-EST:Liang-1.48, bro5, clone 1.48, dar3, dsar1, l(3)05712, sar1', description='Sar1 ortholog (S. cerevisiae)', other_designations='CG7073-PA|CG7073-PC|CG7073-PD|CG7073-PE|Sar1 ortholog|Sar1-PA|Sar1-PC|Sar1-PD|Sar1-PE|dendritic arbor reduction 3|sar1|secretion-associated RAS-related 1', map_location='94A5-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22356685, end_position_on_the_genomic_accession=22360711, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42616, CurrentID=0, Status='live', Symbol='PSR', Aliases='Dmel_CG5383, BEST:LD22859, CG5383, Dmel\\CG5383, L0022859, PS, Psr, dPSR, psr', description='phosphatidylserine receptor', other_designations='CG5383-PA|CG5383-PB|PSR-PA|PSR-PB|phosphatidyl-serine receptor', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22360782, end_position_on_the_genomic_accession=22362371, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42617, CurrentID=0, Status='live', Symbol='CG7071', Aliases='Dmel_ BcDNA:RE40914, Dmel\\CG7071, Dromel_CG7071_FBtr0084211_mORF', description='CG7071 gene product from transcript CG7071-RC', other_designations='CG7071-PA|CG7071-PB|CG7071-PC', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22362620, end_position_on_the_genomic_accession=22364580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42618, CurrentID=0, Status='live', Symbol='CG5382', Aliases='Dmel_ Dmel\\CG5382', description='CG5382 gene product from transcript CG5382-RB', other_designations='CG5382-PB', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22364573, end_position_on_the_genomic_accession=22365983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42619, CurrentID=0, Status='live', Symbol='CG5380', Aliases='Dmel_ Dmel\\CG5380', description='CG5380 gene product from transcript CG5380-RB', other_designations='CG5380-PA|CG5380-PB', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22366080, end_position_on_the_genomic_accession=22367373, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42620, CurrentID=0, Status='live', Symbol='PyK', Aliases='Dmel_CG7070, CG7070, Dmel\\CG7070, PK, PYK, Pyk, l(3)rQ178, l(3)r[Q178]', description='Pyruvate kinase', other_designations='CG7070-PA|CG7070-PB|PyK-PA|PyK-PB|lethal (3) rQ178|pyruvate kinase', map_location='94A6-94A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22367536, end_position_on_the_genomic_accession=22372746, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42621, CurrentID=0, Status='live', Symbol='CG7069', Aliases='Dmel_ Dmel\\CG7069', description='CG7069 gene product from transcript CG7069-RA', other_designations='CG7069-PA', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22372835, end_position_on_the_genomic_accession=22375579, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42622, CurrentID=0, Status='live', Symbol='CG18596', Aliases='Dmel_ Dmel\\CG18596', description='CG18596 gene product from transcript CG18596-RA', other_designations='CG18596-PA|uncharacterized protein', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22375923, end_position_on_the_genomic_accession=22379955, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42624, CurrentID=0, Status='live', Symbol='CG43343', Aliases='Dmel_ CG18427, CG31169, Dmel\\CG43343, Dmel_CG31169, NEST:bs15d11, NEST:bs16c09, gpdh-3', description='CG43343 gene product from transcript CG43343-RB', other_designations='CG43343-PA|CG43343-PB|CG43343-PC|uncharacterized protein', map_location='94A6-94A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22382035, end_position_on_the_genomic_accession=22386548, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42625, CurrentID=0, Status='live', Symbol='CG5379', Aliases='Dmel_ Dmel\\CG5379', description='CG5379 gene product from transcript CG5379-RB', other_designations='CG5379-PA|CG5379-PB', map_location='94A11-94A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22386489, end_position_on_the_genomic_accession=22389714, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42626, CurrentID=0, Status='live', Symbol='CG7059', Aliases='Dmel_ Dmel\\CG7059', description='CG7059 gene product from transcript CG7059-RA', other_designations='CG7059-PA|CG7059-PC|CG7059-PD', map_location='94A11-94A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22389780, end_position_on_the_genomic_accession=22392349, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42627, CurrentID=0, Status='live', Symbol='CG13857', Aliases='Dmel_ Dmel\\CG13857', description='CG13857 gene product from transcript CG13857-RA', other_designations='CG13857-PA|uncharacterized protein', map_location='94A11-94A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22392189, end_position_on_the_genomic_accession=22394104, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42628, CurrentID=0, Status='live', Symbol='CG13856', Aliases='Dmel_ Dmel\\CG13856', description='CG13856 gene product from transcript CG13856-RA', other_designations='CG13856-PA|uncharacterized protein', map_location='94A12-94A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22402655, end_position_on_the_genomic_accession=22405059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42629, CurrentID=0, Status='live', Symbol='CG13855', Aliases='Dmel_ Dmel\\CG13855', description='CG13855 gene product from transcript CG13855-RA', other_designations='CG13855-PA|uncharacterized protein', map_location='94A12-94A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22405939, end_position_on_the_genomic_accession=22408640, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42630, CurrentID=0, Status='live', Symbol='CG13850', Aliases='Dmel_ Dmel\\CG13850, GH07286p', description='CG13850 gene product from transcript CG13850-RA', other_designations='CG13850-PA|CG13850-PB|uncharacterized protein', map_location='94A12-94A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22408566, end_position_on_the_genomic_accession=22410915, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42632, CurrentID=0, Status='live', Symbol='lqfR', Aliases='Dmel_CG42250, CG13853, CG13854, CG31170, CG31285, CG42250, Dmel\\CG42250, Dmel_CG31170, Dmel_CG31285, Epsin, LqfR, Lqfr, Tel2, XII-10, epsin, epsin-like, epsin02, epsinR, l(3)03685, l(3)A9, l(3)SG62, l(3)XII-10, l(3)dsl-9, l(3)dsl9', description='liquid facets-Related', other_designations='CG42250-PC|CG42250-PD|EpsinR|epsin-2|epsin-like|lethal (3) SG62|liquid facet related|liquid facets related|liquid-facets related|lqfR-PC|lqfR-PD', map_location='3-76.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22411301, end_position_on_the_genomic_accession=22419051, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42633, CurrentID=0, Status='live', Symbol='Nop56', Aliases='Dmel_CG13849, CG13849, Dm Dmel\\CG13849, Dnop56, FBgn0038964, nop56', description='CG13849 gene product from transcript CG13849-RA', other_designations='CG13849-PA|Nop56-PA', map_location='94A12-94A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22419095, end_position_on_the_genomic_accession=22421068, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42634, CurrentID=0, Status='live', Symbol='mats', Aliases='Dmel_CG13852, CG13852, DMob1, Dmel\\CG13852, Dmob1, Mats, Mob1, anon-WO0118547.303, dmob1', description='mob as tumor suppressor', other_designations='CG13852-PA|CG13852-PB|CG13852-PC|Mob as tumor suppresso|Mob as tumor suppressor|mats-PA|mats-PB|mats-PC', map_location='94A12-94A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22421161, end_position_on_the_genomic_accession=22423302, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42635, CurrentID=0, Status='live', Symbol='pinta', Aliases='Dmel_CG13848, CG13848, CT33362, Dmel\\CG13848, Pinta', description='prolonged depolarization afterpotential (PDA) is not apparent', other_designations='CG13848-PA|CG13848-PB|Lyso-PA acyltransferase|pinta-PA|pinta-PB', map_location='94A12-94A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22423316, end_position_on_the_genomic_accession=22424929, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42636, CurrentID=0, Status='live', Symbol='CG13847', Aliases='Dmel_ Dmel\\CG13847', description='CG13847 gene product from transcript CG13847-RC', other_designations='CG13847-PB|CG13847-PC|uncharacterized protein', map_location='94A13-94A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22425214, end_position_on_the_genomic_accession=22436150, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42637, CurrentID=0, Status='live', Symbol='CG12499', Aliases='Dmel_ Dmel\\CG12499', description='CG12499 gene product from transcript CG12499-RA', other_designations='CG12499-PA|uncharacterized protein', map_location='94A14-94A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22440152, end_position_on_the_genomic_accession=22446587, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42638, CurrentID=0, Status='live', Symbol='CG34376', Aliases='Dmel_ CG13845, CG13846, CG13851, CG31160, Dmel\\CG34376, Dmel_CG13845, Dmel_CG31160', description='CG34376 gene product from transcript CG34376-RB', other_designations='CG34376-PA|CG34376-PB|uncharacterized protein', map_location='94A15-94A15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22448099, end_position_on_the_genomic_accession=22460749, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42641, CurrentID=0, Status='live', Symbol='Rpn7', Aliases='Dmel_CG5378, 141231_at, CG5378, Dmel\\CG5378, S10a, anon-WO0172774.103, anon-WO0172774.106, p42A', description='Regulatory particle non-ATPase 7', other_designations='CG5378-PA|Rpn7-PA|regulatory particle non-ATPase 7', map_location='94A15-94A15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22460662, end_position_on_the_genomic_accession=22462054, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42642, CurrentID=0, Status='live', Symbol='AP-2mu', Aliases='Dmel_CG7057, AP-50, AP2mu2, AP50, CG7057, Dmel\\CG7057, Dmu2, MENE (3R)-D, MENE(3R)-D, anon-EST:Posey50, anon-WO0172774.104, apm4, cg7057, mu2', description='Adaptor Protein complex 2, mu subunit', other_designations='AP-2mu-PA|AP-2mu-PB|AP-2mu-PC|AP-50|CG7057-PA|CG7057-PB|CG7057-PC|adaptor protein complex 2|adaptor protein complex 2, mu subunit|mu2 medium chain of adaptin AP2', map_location='94A15-94A16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22462455, end_position_on_the_genomic_accession=22464603, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42643, CurrentID=0, Status='live', Symbol='CG7054', Aliases='Dmel_ Dmel\\CG7054', description='CG7054 gene product from transcript CG7054-RA', other_designations='CG7054-PA', map_location='94A16-94A16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22464985, end_position_on_the_genomic_accession=22465938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42644, CurrentID=0, Status='live', Symbol='Pebp1', Aliases='Dmel_CG18594, BcDNA:LP12095, CG18594, Dmel\\CG18594, PEBP1, T2', description='Phosphatidylethanolamine-binding protein 1', other_designations='CG18594-PA|Pebp1-PA', map_location='94A16-94A16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22466034, end_position_on_the_genomic_accession=22466679, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42645, CurrentID=0, Status='live', Symbol='CG5377', Aliases='Dmel_ Dmel\\CG5377', description='CG5377 gene product from transcript CG5377-RA', other_designations='CG5377-PA', map_location='94A16-94A16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22466905, end_position_on_the_genomic_accession=22468157, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42646, CurrentID=0, Status='live', Symbol='Nrx-1', Aliases='Dmel_CG7050, CG7050, CT21808, DNrx, Dmel\\CG7050, Nrx, Nrx1, dnrx, dnrx-1, dnrx1, nrx', description='Neurexin 1', other_designations='CG7050-PA|CG7050-PB|CG7050-PC|CG7050-PD|CG7050-PE|CG7050-PF|CG7050-PG|CG7050-PH|CG7050-PI|Neurexin-1|Neurexin-like|Nrx-1-PA|Nrx-1-PB|Nrx-1-PC|Nrx-1-PD|Nrx-1-PE|Nrx-1-PF|Nrx-1-PG|Nrx-1-PH|Nrx-1-PI|drosophila neurexin|neurexin|neurexin 1', map_location='94A16-94B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22468512, end_position_on_the_genomic_accession=22486969, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  42647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42647, CurrentID=0, Status='live', Symbol='CG7048', Aliases='Dmel_ Dmel\\CG7048', description='CG7048 gene product from transcript CG7048-RB', other_designations='CG7048-PA|CG7048-PB', map_location='94B1-94B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22487039, end_position_on_the_genomic_accession=22488110, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42648, CurrentID=0, Status='live', Symbol='CG5376', Aliases='Dmel_ Dmel\\CG5376', description='CG5376 gene product from transcript CG5376-RA', other_designations='CG5376-PA', map_location='94B1-94B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22488033, end_position_on_the_genomic_accession=22488955, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42649, CurrentID=0, Status='live', Symbol='T-cp1', Aliases='Dmel_CG5374, BEST:GH05123, CCT1, CG5374, Dmel\\CG5374, T-CP1, T-cpl, TCP-1alpha, TCP1alpha, TCPA_DROME, Tcp-1, Tcp1, Tcp1-alpha, gh05123, tcp1', description='Tcp1-like', other_designations='CG5374-PA|CG5374-PB|T-complex 1|T-cp1-PA|T-cp1-PB', map_location='94B1-94B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22489105, end_position_on_the_genomic_accession=22491439, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42650, CurrentID=0, Status='live', Symbol='tHMG1', Aliases='Dmel_CG7045, CG7045, Dmel\\CG7045, tHMG-1, thmg-1', description='testis-enriched HMG-box-containing protein 1', other_designations='CG7045-PA|tHMG1-PA|testis-enriched HMG-box-containing protein-1', map_location='94B1-94B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22491829, end_position_on_the_genomic_accession=22492381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42651, CurrentID=0, Status='live', Symbol='tHMG2', Aliases='Dmel_CG7046, BcDNA:AT28425, CG7046, Dmel\\CG7046, tHMG-2, thmg-2', description='testis-enriched HMG-box-containing protein 2', other_designations='CG7046-PA|CG7046-PB|tHMG2-PA|tHMG2-PB|testis-enriched HMG-box-containing protein-2', map_location='94B1-94B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22492953, end_position_on_the_genomic_accession=22493746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42652, CurrentID=0, Status='live', Symbol='Octbeta1R', Aliases='Dmel_CG6919, CG6919, DmCG6919, DmOctBeta1R, DmOctbeta1R, Dmel\\CG6919, Dmoa2, OA2, Oa2, OctBeta1, anon-WO0170980.67, anon-WO0170980.68, oa2, octbeta1R', description='Octopamine beta1 receptor', other_designations='CG6919-PA|CG6919-PB|CG6919-PC|CG6919-PE|Octbeta1R-PA|Octbeta1R-PB|Octbeta1R-PC|Octbeta1R-PE|octopamine beta1 receptor|octopamine receptor 2', map_location='94B1-94B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22497215, end_position_on_the_genomic_accession=22520078, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42653, CurrentID=0, Status='live', Symbol='CG5346', Aliases='Dmel_ Dmel\\CG5346', description='CG5346 gene product from transcript CG5346-RA', other_designations='CG5346-PA', map_location='94B2-94B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22520087, end_position_on_the_genomic_accession=22528663, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42654, CurrentID=0, Status='live', Symbol='CG33093', Aliases='Dmel_ CG5340, Dmel\\CG33093', description='CG33093 gene product from transcript CG33093-RA', other_designations='CG33093-PA|uncharacterized protein', map_location='94B2-94B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22530694, end_position_on_the_genomic_accession=22531922, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42655, CurrentID=0, Status='live', Symbol='CG5326', Aliases='Dmel_ Dmel\\CG5326', description='CG5326 gene product from transcript CG5326-RA', other_designations='CG5326-PA', map_location='94B2-94B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22532174, end_position_on_the_genomic_accession=22539919, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42656, CurrentID=0, Status='live', Symbol='AdipoR', Aliases='Dmel_CG5315, CG5315, Dmel\\CG5315, dAdipoR', description='Adiponectin receptor', other_designations='AdipoR-PA|AdipoR-PB|AdipoR-PC|AdipoR-PD|AdipoR-PE|CG5315-PA|CG5315-PB|CG5315-PC|CG5315-PD|CG5315-PE|adiponectin receptor', map_location='94B3-94B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22543693, end_position_on_the_genomic_accession=22547846, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42657, CurrentID=0, Status='live', Symbol='bond', Aliases='Dmel_CG6921, CG6921, Dmel\\CG6921, James bond', description='james bond', other_designations='CG6921-PA|CG6921-PB|CG6921-PC|bond-PA|bond-PB|bond-PC', map_location='3-75.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22550680, end_position_on_the_genomic_accession=22556494, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42658, CurrentID=0, Status='live', Symbol='CG5278', Aliases='Dmel_ Dmel\\CG5278', description='CG5278 gene product from transcript CG5278-RA', other_designations='CG5278-PA', map_location='94B4-94B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22558342, end_position_on_the_genomic_accession=22562572, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42659, CurrentID=0, Status='live', Symbol='CG33110', Aliases='Dmel_ CG13844, CG6926, Dmel\\CG33110', description='CG33110 gene product from transcript CG33110-RA', other_designations='CG33110-PA|uncharacterized protein', map_location='94B4-94B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22570188, end_position_on_the_genomic_accession=22580392, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42661, CurrentID=0, Status='live', Symbol='CSN6', Aliases='Dmel_CG6932, CG6932, CH6, Csn6, Dch6, Dmel\\CG6932, Q9VCY3', description='COP9 signalosome subunit 6', other_designations='CG6932-PA|COP9 complex homolog subunit 6|CSN6-PA|drosophila COP9 signalosome homolog 6', map_location='94B5-94B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22580773, end_position_on_the_genomic_accession=22582033, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42662, CurrentID=0, Status='live', Symbol='CG6937', Aliases='Dmel_ Dmel\\CG6937, cg6937', description='CG6937 gene product from transcript CG6937-RA', other_designations='CG6937-PA', map_location='94B5-94B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22584773, end_position_on_the_genomic_accession=22586209, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42663, CurrentID=0, Status='live', Symbol='Dph5', Aliases='Dmel_CG31289, BcDNA.LD12153, BcDNA:LD12153, CG31289, CG5275, Dmel\\CG31289, l(3)L4910', description='Diphthamide methyltransferase', other_designations='CG31289-PA|Dph5-PA|diphthamide methyltransferase', map_location='94B5-94B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22582028, end_position_on_the_genomic_accession=22583225, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42664, CurrentID=0, Status='live', Symbol='btn', Aliases='Dmel_CG5264, 1261/13, BTN, CG5264, Dmel\\CG5264/MEOX, btn', description='buttonless', other_designations='CG5264-PA|btn-PA', map_location='94B5-94B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22586149, end_position_on_the_genomic_accession=22586625, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42665, CurrentID=0, Status='live', Symbol='Efa6', Aliases='Dmel_CG31158, CG18185, CG31158, CG6941, Dmel\\CG31158, EFA6, dEfa6, dPsd', description='Exchange factor for Arf 6 ortholog (H. sapiens)', other_designations='CG31158-PC|CG31158-PD|CG31158-PG|CG31158-PH|CG31158-PI|Efa6 pleckstrin and Sec7 domain containing|Efa6-PC|Efa6-PD|Efa6-PG|Efa6-PH|Efa6-PI|exchange factor for Arf 6 ortholog', map_location='94B5-94B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22587755, end_position_on_the_genomic_accession=22608842, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  42668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42668, CurrentID=0, Status='live', Symbol='CG13843', Aliases='Dmel_ Dmel\\CG13843', description='CG13843 gene product from transcript CG13843-RA', other_designations='CG13843-PA|uncharacterized protein', map_location='94B6-94B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22609485, end_position_on_the_genomic_accession=22610715, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42669, CurrentID=0, Status='live', Symbol='CG31156', Aliases='Dmel_ CG5253, CG5257, Dmel\\CG31156', description='CG31156 gene product from transcript CG31156-RA', other_designations='CG31156-PA|uncharacterized protein', map_location='94B6-94B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22608670, end_position_on_the_genomic_accession=22613656, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42670, CurrentID=0, Status='live', Symbol='EF-G2', Aliases='Dmel_CG31159, CG31159, CG6943, Dmel\\CG31159, dEF-G2', description='Elongation Factor G2', other_designations='CG31159-PA|CG31159-PB|EF-G2-PA|EF-G2-PB|elongation factor G2', map_location='94B6-94B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22613686, end_position_on_the_genomic_accession=22616622, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42671, CurrentID=0, Status='live', Symbol='mRpL45', Aliases='Dmel_CG6949, CG6949, Dmel\\CG6949, MRP-L45, XI-2, l(3)A8, l(3)SG61, l(3)dsl-8, l(3)dsl8', description='mitochondrial ribosomal protein L45', other_designations='CG6949-PA|lethal (3) SG61|mRpL45-PA|mitochondrial ribosomal protein large subunit 45', map_location='3-76.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22618492, end_position_on_the_genomic_accession=22619826, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42672, CurrentID=0, Status='live', Symbol='loco', Aliases='Dmel_CG5248, CG17229, CG5248, Dmel\\CG5248, Loco, dLoco', description='locomotion defects', other_designations='CG5248-PB|CG5248-PC|CG5248-PD|CG5248-PF|CG5248-PH|loco-PB|loco-PC|loco-PD|loco-PF|loco-PH|locomotion defect', map_location='94B6-94B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22619968, end_position_on_the_genomic_accession=22639387, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42676, CurrentID=0, Status='live', Symbol='wake', Aliases='Dmel_CG45058, CG17622, CG17623, CG42686, CG45058, CG6954, Dmel\\CG45058, Dmel_CG17622, Dmel_CG17623, Dmel_CG42686, Dmel_CG6954, ban', description='wide awake', other_designations='CG45058-PB|CG45058-PC|CG45058-PD|CG45058-PE|CG45058-PF|CG45058-PG|CG45058-PH|banderuola|wake-PB|wake-PC|wake-PD|wake-PE|wake-PF|wake-PG|wake-PH', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22639808, end_position_on_the_genomic_accession=22683827, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  42678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42678, CurrentID=0, Status='live', Symbol='CG17625', Aliases='Dmel_ Dmel\\CG17625', description='CG17625 gene product from transcript CG17625-RA', other_designations='CG17625-PA|uncharacterized protein', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22686849, end_position_on_the_genomic_accession=22687407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42679, CurrentID=0, Status='live', Symbol='wfs1', Aliases='Dmel_CG4917, CG4917, Dmel\\CG4917', description='wolfram syndrome 1', other_designations='CG4917-PA|CG4917-PC|wfs1-PA|wfs1-PC', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22687542, end_position_on_the_genomic_accession=22690380, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42680, CurrentID=0, Status='live', Symbol='Nup133', Aliases='Dmel_CG6958, CG6958, Dmel\\CG6958, Nup 133, nup133', description='Nucleoporin 133kD', other_designations='CG6958-PA|Nup133-PA|nucleoporin 133|nucleoporin 133kD', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22690607, end_position_on_the_genomic_accession=22694772, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42681, CurrentID=0, Status='live', Symbol='cd', Aliases='Dmel_CG6969, CG6969, Dmel\\CG6969', description='cardinal', other_designations='CG6969-PA|cd-PA', map_location='3-75.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22694959, end_position_on_the_genomic_accession=22697990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42682, CurrentID=0, Status='live', Symbol='Cyp6d4', Aliases='Dmel_CG12800, 6d4, CG12800, Dmel\\CG12800', description='CG12800 gene product from transcript CG12800-RA', other_designations='CG12800-PA|Cyp6d4-PA|cytochrome P450', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22698319, end_position_on_the_genomic_accession=22700307, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42683, CurrentID=0, Status='live', Symbol='CCAP', Aliases='Dmel_CG4910, CG4910, Ccap, DmCCAP, Dmel\\CG4910, ccap', description='Crustacean cardioactive peptide', other_designations='CCAP prepropeptide|CCAP-PA|CCAP-PB|CG4910-PA|CG4910-PB|cardioacceleratory peptide|crustacean cardioactive peptide', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22700582, end_position_on_the_genomic_accession=22702084, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42684, CurrentID=0, Status='live', Symbol='CG6972', Aliases='Dmel_ Dmel\\CG6972', description='CG6972 gene product from transcript CG6972-RB', other_designations='CG6972-PA|CG6972-PB', map_location='94C1-94C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22702007, end_position_on_the_genomic_accession=22704338, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42685, CurrentID=0, Status='live', Symbol='CG13842', Aliases='Dmel_ Dmel\\CG13842', description='CG13842 gene product from transcript CG13842-RB', other_designations='CG13842-PB|uncharacterized protein', map_location='94C2-94C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22704254, end_position_on_the_genomic_accession=22707278, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42686, CurrentID=0, Status='live', Symbol='CG4907', Aliases='Dmel_ Dmel\\CG4907', description='CG4907 gene product from transcript CG4907-RA', other_designations='CG4907-PA', map_location='94C2-94C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22707786, end_position_on_the_genomic_accession=22710762, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42687, CurrentID=0, Status='live', Symbol='wge', Aliases='Dmel_CG31151, CG17621, CG31151, CG4902, Dmel\\CG31151', description='winged eye', other_designations='CG31151-PC|CG31151-PD|CG31151-PE|CG31151-PF|wge-PC|wge-PD|wge-PE|wge-PF', map_location='94C2-94C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22710883, end_position_on_the_genomic_accession=22721150, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42689, CurrentID=0, Status='live', Symbol='Irp-1A', Aliases='Dmel_CG4900, C-Acon, CG4900, Dmel\\CG4900, IRF, IRP, IRP-1, IRP-1A, IRP1, c-Acon, cACON, cAcon, dIRP, irp1A', description='Iron regulatory protein 1A', other_designations='CG4900-PA|Irp-1A-PA|c-aconitase|cytoplasmic aconitase|cytosolic aconitase|iron regulatory factor|iron regulatory protein|iron regulatory protein 1A', map_location='94C3-94C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22721320, end_position_on_the_genomic_accession=22724587, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42690, CurrentID=0, Status='live', Symbol='CG4813', Aliases='Dmel_ Dmel\\CG4813', description='CG4813 gene product from transcript CG4813-RA', other_designations='CG4813-PA', map_location='94C3-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22724884, end_position_on_the_genomic_accession=22726414, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42691, CurrentID=0, Status='live', Symbol='CG45049', Aliases='Dmel_ CG17618, CG6982, Dmel\\CG45049, Dmel_CG17618, Dmel_CG6982', description='CG45049 gene product from transcript CG45049-RC', other_designations='CG45049-PB|CG45049-PC|CG45049-PD|CG45049-PE|CG45049-PF|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22726967, end_position_on_the_genomic_accession=22732362, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42692, CurrentID=0, Status='live', Symbol='Takl2', Aliases='Dmel_CG4803, CG4803, Dmel\\CG4803, TAK', description='Tak1-like 2', other_designations='CG4803-PB|Takl2-PB', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22732356, end_position_on_the_genomic_accession=22733851, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42693, CurrentID=0, Status='live', Symbol='Dcr-1', Aliases='Dmel_CG4792, CG4792, DCR-1, DCR1, DICER, Dcr, Dcr1, Dicer, Dicer1, Dmel\\CG4792, cg4792, dcr 1, dcr-1, dcr1, dicer, dicer-1, dicer1', description='Dicer-1', other_designations='CG4792-PA|Dcr-1-PA|dicer|dicer 1|dicer1', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22734007, end_position_on_the_genomic_accession=22741091, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42694, CurrentID=0, Status='live', Symbol='CG6985', Aliases='Dmel_ Dmel\\CG6985', description='CG6985 gene product from transcript CG6985-RA', other_designations='CG6985-PA', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22741150, end_position_on_the_genomic_accession=22743081, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42695, CurrentID=0, Status='live', Symbol='vret', Aliases='Dmel_CG4771, CG4771, Dmel\\CG4771', description='vreteno', other_designations='CG4771-PA|avocado|vret-PA|vretano', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22742954, end_position_on_the_genomic_accession=22745581, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42696, CurrentID=0, Status='live', Symbol='HP1c', Aliases='Dmel_CG6990, CG6990, Dmel\\CG6990, HP1, HP1C, Hp1c, hp1c', description='Heterochromatin Protein 1c', other_designations='CG6990-PA|HP1c-PA|heterochromatin protein 1c', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22745620, end_position_on_the_genomic_accession=22746605, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42697, CurrentID=0, Status='live', Symbol='CG17141', Aliases='Dmel_ Dmel\\CG17141', description='CG17141 gene product from transcript CG17141-RA', other_designations='CG17141-PA|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22746446, end_position_on_the_genomic_accession=22747599, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42698, CurrentID=0, Status='live', Symbol='CG31139', Aliases='Dmel_ BcDNA:AT07872, CG17138, Dmel\\CG31139', description='CG31139 gene product from transcript CG31139-RC', other_designations='CG31139-PA|CG31139-PC|partner of rumi|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22747690, end_position_on_the_genomic_accession=22749056, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42699, CurrentID=0, Status='live', Symbol='CG4725', Aliases='Dmel_ Dmel\\CG4725', description='CG4725 gene product from transcript CG4725-RA', other_designations='CG4725-PA', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22758810, end_position_on_the_genomic_accession=22761076, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42700, CurrentID=0, Status='live', Symbol='CG4723', Aliases='Dmel_ Dmel\\CG4723', description='CG4723 gene product from transcript CG4723-RA', other_designations='CG4723-PA', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22761440, end_position_on_the_genomic_accession=22763711, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42701, CurrentID=0, Status='live', Symbol='CG4721', Aliases='Dmel_ Dmel\\CG4721, anon-EST:Posey132', description='CG4721 gene product from transcript CG4721-RA', other_designations='CG4721-PA', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22763960, end_position_on_the_genomic_accession=22766172, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42702, CurrentID=0, Status='live', Symbol='Usp12-46', Aliases='Dmel_CG7023, CG7023, Dmel\\CG7023, Q9VCT9', description='Ubiquitin-specific protease 12/46 ortholog', other_designations='CG7023-PB|Usp12-46-PB', map_location='94C4-94C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22750901, end_position_on_the_genomic_accession=22776374, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42703, CurrentID=0, Status='live', Symbol='CG7029', Aliases='Dmel_ Dmel\\CG7029', description='CG7029 gene product from transcript CG7029-RC', other_designations='CG7029-PA|CG7029-PC|CG7029-PD|CG7029-PF|CG7029-PG|CG7029-PI', map_location='94C4-94C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22750901, end_position_on_the_genomic_accession=22794030, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42704, CurrentID=0, Status='live', Symbol='CG7031', Aliases='Dmel_ Dmel\\CG7031', description='CG7031 gene product from transcript CG7031-RB', other_designations='CG7031-PA|CG7031-PB|CG7031-PC', map_location='94C8-94C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22804291, end_position_on_the_genomic_accession=22807290, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42705, CurrentID=0, Status='live', Symbol='CG13840', Aliases='Dmel_ Dmel\\CG13840', description='CG13840 gene product from transcript CG13840-RC', other_designations='CG13840-PC|uncharacterized protein', map_location='94C8-94C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22814483, end_position_on_the_genomic_accession=22816178, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42706, CurrentID=0, Status='live', Symbol='CG4704', Aliases='Dmel_ Dmel\\CG4704', description='CG4704 gene product from transcript CG4704-RB', other_designations='CG4704-PA|CG4704-PB', map_location='94D1-94D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22844093, end_position_on_the_genomic_accession=22846302, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42707, CurrentID=0, Status='live', Symbol='klg', Aliases='Dmel_CG6669, CG6669, CT20712, Dmel\\CG6669, Klg, h214, l(3)rN712', description='klingon', other_designations='CG6669-PA|klg-PA|ruslan', map_location='3-78.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22899284, end_position_on_the_genomic_accession=22961731, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42708, CurrentID=0, Status='live', Symbol='CG6660', Aliases='Dmel_ Dmel\\CG6660', description='CG6660 gene product from transcript CG6660-RA', other_designations='CG6660-PA', map_location='94D3-94D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22923352, end_position_on_the_genomic_accession=22924525, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42709, CurrentID=0, Status='live', Symbol='CG17244', Aliases='Dmel_ BcDNA:RH17470, Dmel\\CG17244', description='CG17244 gene product from transcript CG17244-RB', other_designations='CG17244-PA|CG17244-PB|uncharacterized protein', map_location='94D6-94D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22977270, end_position_on_the_genomic_accession=22978164, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42710, CurrentID=0, Status='live', Symbol='Nha2', Aliases='Dmel_CG43442, CG31052, CG43205, CG43442, CG4682, CG4693, DMG8-chr3R.34.001, DmNHA2, Dmel\\CG43442, Dmel_CG31052, TANGO12, Tango12, nha2, tango12', description='Na[+]/H[+] hydrogen antiporter 2', other_designations='CG43442-PA|CG43442-PB|Nha2-PA|Nha2-PB|transport and golgi organization 12', map_location='94D7-94D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22981459, end_position_on_the_genomic_accession=22998528, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  42711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42711, CurrentID=0, Status='live', Symbol='Or94a', Aliases='Dmel_CG17241, 94D.1, 94a, CG17241, DOR108, Dmel\\CG17241', description='Odorant receptor 94a', other_designations='CG17241-PA|Or94a-PA|odorant receptor 94a', map_location='94D7-94D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22987281, end_position_on_the_genomic_accession=22988503, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42712, CurrentID=0, Status='live', Symbol='Or94b', Aliases='Dmel_CG6679, 94D.2, 94b, CG6679, DOR109, Dmel\\CG6679, anon-JP2002534971-A.31, anon-JP2002534971-A.32', description='Odorant receptor 94b', other_designations='CG6679-PA|Or94b-PA|odorant receptor 94b', map_location='94D7-94D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22989295, end_position_on_the_genomic_accession=22990568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42715, CurrentID=0, Status='live', Symbol='CG34375', Aliases='Dmel_ CG13834, CG13835, Dmel\\CG34375, Dmel_CG13834, Dmel_CG13835', description='CG34375 gene product from transcript CG34375-RA', other_designations='CG34375-PA|CG34375-PB|CG34375-PC|uncharacterized protein', map_location='94D8-94D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23000139, end_position_on_the_genomic_accession=23010562, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42716, CurrentID=0, Status='live', Symbol='CG6688', Aliases='Dmel_ Dmel\\CG6688', description='CG6688 gene product from transcript CG6688-RA', other_designations='CG6688-PA', map_location='94D9-94D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23012412, end_position_on_the_genomic_accession=23015475, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42717, CurrentID=0, Status='live', Symbol='lmd', Aliases='Dmel_CG4677, CG4677, Dmel\\CG4677, K, LMD, Lmd, anon-EST:CL2d4, gfl, minc', description='lame duck', other_designations='CG4677-PB|CG4677-PC|gleeful|lameduck|lmd-PB|lmd-PC|myoblasts incompetent', map_location='94D9-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23021222, end_position_on_the_genomic_accession=23026713, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42718, CurrentID=0, Status='live', Symbol='CG13833', Aliases='Dmel_ Dmel\\CG13833', description='CG13833 gene product from transcript CG13833-RA', other_designations='CG13833-PA|uncharacterized protein', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23033284, end_position_on_the_genomic_accession=23035563, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42719, CurrentID=0, Status='live', Symbol='CG13838', Aliases='Dmel_ Dmel\\CG13838', description='CG13838 gene product from transcript CG13838-RA', other_designations='CG13838-PA|uncharacterized protein', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23042241, end_position_on_the_genomic_accession=23042669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42720, CurrentID=0, Status='live', Symbol='CG13837', Aliases='Dmel_ Dmel\\CG13837', description='CG13837 gene product from transcript CG13837-RA', other_designations='CG13837-PA|uncharacterized protein', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23043990, end_position_on_the_genomic_accession=23044661, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42721, CurrentID=0, Status='live', Symbol='CG17121', Aliases='Dmel_ Dmel\\CG17121', description='CG17121 gene product from transcript CG17121-RA', other_designations='CG17121-PA|uncharacterized protein', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23044664, end_position_on_the_genomic_accession=23048214, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42723, CurrentID=0, Status='live', Symbol='CG17119', Aliases='Dmel_ CTNS, Dmel\\CG17119', description='CG17119 gene product from transcript CG17119-RC', other_designations='CG17119-PA|CG17119-PC|uncharacterized protein', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23054410, end_position_on_the_genomic_accession=23058958, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42726, CurrentID=0, Status='live', Symbol='Ublcp1', Aliases='Dmel_CG6697, 4972728, BcDNA:LD21504, CG6697, DmUBLCP1, Dmel\\CG6697, UBLCP1', description='Ubiquitin-like domain-containing C-terminal domain phosphatase 1', other_designations='CG6697-PA|Ublcp1-PA', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23061493, end_position_on_the_genomic_accession=23062751, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42727, CurrentID=0, Status='live', Symbol='CG17111', Aliases='Dmel_ Dmel\\CG17111', description='CG17111 gene product from transcript CG17111-RB', other_designations='CG17111-PB|uncharacterized protein', map_location='94D10-94D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23062713, end_position_on_the_genomic_accession=23067259, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42728, CurrentID=0, Status='live', Symbol='CG6726', Aliases='Dmel_ Dmel\\CG6726', description='CG6726 gene product from transcript CG6726-RC', other_designations='CG6726-PA|CG6726-PB|CG6726-PC', map_location='94D13-94D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23091724, end_position_on_the_genomic_accession=23095184, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42729, CurrentID=0, Status='live', Symbol='CG17110', Aliases='Dmel_ Dmel\\CG17110', description='CG17110 gene product from transcript CG17110-RC', other_designations='CG17110-PA|CG17110-PB|CG17110-PC|uncharacterized protein', map_location='94D13-94D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23091724, end_position_on_the_genomic_accession=23095441, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42730, CurrentID=0, Status='live', Symbol='CG17109', Aliases='Dmel_ Dmel\\CG17109', description='CG17109 gene product from transcript CG17109-RB', other_designations='CG17109-PA|CG17109-PB|uncharacterized protein', map_location='94D13-94D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23095690, end_position_on_the_genomic_accession=23097240, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42731, CurrentID=0, Status='live', Symbol='CG6733', Aliases='Dmel_ Dmel\\CG6733', description='CG6733 gene product from transcript CG6733-RA', other_designations='CG6733-PA', map_location='94D13-94D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23097728, end_position_on_the_genomic_accession=23099191, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42732, CurrentID=0, Status='live', Symbol='CG6738', Aliases='Dmel_ Dmel\\CG6738', description='CG6738 gene product from transcript CG6738-RA', other_designations='CG6738-PA', map_location='94D13-94D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23099533, end_position_on_the_genomic_accession=23100983, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42733, CurrentID=0, Status='live', Symbol='Cow', Aliases='Dmel_CG13830, CG13830, Dmel\\CG13830, cow', description='Carrier of Wingless', other_designations='CG13830-PC|CG13830-PD|CG13830-PE|Cow-PC|Cow-PD|Cow-PE|carrier of wingless', map_location='94D12-94D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23070516, end_position_on_the_genomic_accession=23112730, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42734, CurrentID=0, Status='live', Symbol='Rassf', Aliases='Dmel_CG4656, CG4656, Dmel\\CG4656, RASSF, dRASSF, dRassf, dmRASSF', description='Ras association family member', other_designations='CG4656-PA|Ras-associated domain family protein|Rassf-PA|drosophila Ras association family', map_location='94D13-94E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23114014, end_position_on_the_genomic_accession=23117931, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42735, CurrentID=0, Status='live', Symbol='CenB1A', Aliases='Dmel_CG6742, CG6742, Cen1ba, Dmel\\CG6742, cenB1A', description='Centaurin beta 1A', other_designations='CG6742-PA|CG6742-PB|CenB1A-PA|CenB1A-PB|centaurin beta 1A', map_location='94E1-94E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23118400, end_position_on_the_genomic_accession=23124217, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42736, CurrentID=0, Status='live', Symbol='CG31365', Aliases='Dmel_ CG4639, Dmel\\CG31365', description='CG31365 gene product from transcript CG31365-RA', other_designations='CG31365-PA|uncharacterized protein', map_location='94E1-94E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23123962, end_position_on_the_genomic_accession=23126422, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42737, CurrentID=0, Status='live', Symbol='hh', Aliases='Dmel_CG4637, CG4637, Dmel\\CG4637, HH, Hh, Mir, Mrt, anon-WO0134654.19, anon-WO0182946.19, bar, bar-3, bar3, hg, l(3)hh, l(3)neo56, l(3)neo57', description='hedgehog', other_designations='CG4637-PB|bar-on-3|hedghog|hh-PB|mirabile|moonrat', map_location='3-81.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23128169, end_position_on_the_genomic_accession=23141906, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42738, CurrentID=0, Status='live', Symbol='unk', Aliases='Dmel_CG4620, CG4620, Dmel\\CG4620', description='unkempt', other_designations='CG4620-PA|CG4620-PB|CG4620-PC|CG4620-PD|CG4620-PE|CG4620-PF|CG4620-PG|unk-PA|unk-PB|unk-PC|unk-PD|unk-PE|unk-PF|unk-PG', map_location='3-79.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23149653, end_position_on_the_genomic_accession=23157862, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42739, CurrentID=0, Status='live', Symbol='VhaAC39-2', Aliases='Dmel_CG4624, CG4624, Dmel\\CG4624, anon-WO0118547.227, vhaAC39-2', description='Vacuolar H[+] ATPase AC39 subunit 2', other_designations='CG4624-PA|VhaAC39-2-PA|vacuolar H[+] ATPase AC39 subunit 2|vacuolar H[+] ATPase subunit AC39-2', map_location='94E2-94E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23158017, end_position_on_the_genomic_accession=23159316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42740, CurrentID=0, Status='live', Symbol='CG13829', Aliases='Dmel_ Dmel\\CG13829', description='CG13829 gene product from transcript CG13829-RA', other_designations='CG13829-PA|uncharacterized protein', map_location='94E3-94E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23160477, end_position_on_the_genomic_accession=23161338, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42742, CurrentID=0, Status='live', Symbol='Irk1', Aliases='Dmel_CG44159, CG13836, CG44159, CG6747, DIR, Dir, Dm_3R:100326, Dm_3R:100349, Dmel\\CG44159, Dmel_CG13836, Dmel_CG6747, Ir, dKirI, ir, irk1', description='Inwardly rectifying potassium channel 1', other_designations='CG44159-PB|CG44159-PC|CG44159-PD|CG44159-PE|CG44159-PF|CG44159-PG|CG44159-PH|Ir|Irk1-PB|Irk1-PC|Irk1-PD|Irk1-PE|Irk1-PF|Irk1-PG|Irk1-PH|drosophila inward rectifier|inwardly rectifying potassium channel|inwardly rectifying potassium channel 1', map_location='94E4-94E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23168317, end_position_on_the_genomic_accession=23185642, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42743, CurrentID=0, Status='live', Symbol='cnc', Aliases='Dmel_CG43286, 5134, BcDNA:RE05559, CG13826, CG17894, CG43286, CG4566, CG4578, CNC, CNC_DROME, Cnc, DM12, Dmel\\CG43286, Dmel_CG13826, Dmel_CG17894, Nrf2, anon-WO0153538.6, anon-WO0153538.7-C, l(3)03921, l(3)j5E7, cnc', description='cap-n-collar', other_designations="CG43286-PA|CG43286-PB|CG43286-PC|CG43286-PD|CG43286-PE|CG43286-PF|CG43286-PG|CG43286-PH|CG43286-PI|CG43286-PJ|CG43286-PK|CG43286-PL|CG43286-PM|CG43286-PN|CG43286-PO|CG43286-PP|Cap 'n' collar|Cap'n'collar|Cap'n'collar B|cap and collar|cap n collar|cnc-PA|cnc-PB|cnc-PC|cnc-PD|cnc-PE|cnc-PF|cnc-PG|cnc-PH|cnc-PI|cnc-PJ|cnc-PK|cnc-PL|cnc-PM|cnc-PN|cnc-PO|cnc-PP", map_location='3-81.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23185580, end_position_on_the_genomic_accession=23226711, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  42745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42745, CurrentID=0, Status='live', Symbol='fzo', Aliases='Dmel_CG4568, CG4568, Dmel\\CG4568, anon-WO0125274.1, anon-WO0125274.25, anon-WO0125274.26', description='fuzzy onions', other_designations='CG4568-PA|fzo-PA', map_location='94E6-94E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23201644, end_position_on_the_genomic_accession=23204034, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42747, CurrentID=0, Status='live', Symbol='CG4467', Aliases='Dmel_ CG13825, Dmel\\CG4467', description='CG4467 gene product from transcript CG4467-RB', other_designations='CG4467-PA|CG4467-PB', map_location='94E7-94E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23227727, end_position_on_the_genomic_accession=23248903, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  42748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42748, CurrentID=0, Status='live', Symbol='CG4449', Aliases='Dmel_ Dmel\\CG4449', description='CG4449 gene product from transcript CG4449-RB', other_designations='CG4449-PA|CG4449-PB|CG4449-PC|CG4449-PD|CG4449-PE|CG4449-PF', map_location='94E8-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23249443, end_position_on_the_genomic_accession=23251477, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42749, CurrentID=0, Status='live', Symbol='EloA', Aliases='Dmel_CG6755, CG6755, Dmel\\CG6755, Elo-A, dEloA', description='Elongin A', other_designations='CG6755-PA|CG6755-PB|EloA-PA|EloA-PB|Elongin-A|ElonginA|elongin A', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23251563, end_position_on_the_genomic_accession=23255071, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42750, CurrentID=0, Status='live', Symbol='wda', Aliases='Dmel_CG4448, CG4448, Dmel\\CG4448, WDA, WFA, Wda', description='will decrease acetylation', other_designations='CG4448-PA|wda-PA', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23255070, end_position_on_the_genomic_accession=23258087, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42751, CurrentID=0, Status='live', Symbol='CG13827', Aliases='Dmel_ BcDNA:RE30473, Dmel\\CG13827, PEX11C', description='CG13827 gene product from transcript CG13827-RA', other_designations='CG13827-PA|CG13827-PB|uncharacterized protein', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23258365, end_position_on_the_genomic_accession=23260602, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42752, CurrentID=0, Status='live', Symbol='orb', Aliases='Dmel_CG10868, 151666_s_at, CG10868, Dmel\\CG10868, ORB, Oo18, Orb, fs(3)00107, oo18', description='oo18 RNA-binding protein', other_designations='CG10868-PA|CG10868-PB|CG10868-PC|CG10868-PD|CG10868-PE|CG10868-PF|CG10868-PG|ORB|Oo18 RNA-binding protein|oo18 RNA binding|orb-PA|orb-PB|orb-PC|orb-PD|orb-PE|orb-PF|orb-PG', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23260610, end_position_on_the_genomic_accession=23280856, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  42753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42753, CurrentID=0, Status='live', Symbol='Cdc16', Aliases='Dmel_CG6759, APC6, Apc6, CDC16, CG6759, Dmcdc16, Dmel\\CG6759, cdc16', description='Cell division cycle 16 ortholog', other_designations='CG6759-PA|Cdc16-PA|cdc16|cell division cycle 16 ortholog', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23281201, end_position_on_the_genomic_accession=23284123, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42754, CurrentID=0, Status='live', Symbol='CG6763', Aliases='Dmel_ Dmel\\CG6763', description='CG6763 gene product from transcript CG6763-RA', other_designations='CG6763-PA|CG6763-PB', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23284002, end_position_on_the_genomic_accession=23286041, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42755, CurrentID=0, Status='live', Symbol='CG17083', Aliases='Dmel_ Dmel\\CG17083', description='CG17083 gene product from transcript CG17083-RA', other_designations='CG17083-PA|uncharacterized protein', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23286093, end_position_on_the_genomic_accession=23288051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42756, CurrentID=0, Status='live', Symbol='CG4434', Aliases='Dmel_ Dmel\\CG4434', description='CG4434 gene product from transcript CG4434-RA', other_designations='CG4434-PA', map_location='94E9-94E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23288217, end_position_on_the_genomic_accession=23290189, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42757, CurrentID=0, Status='live', Symbol='pnt', Aliases='Dmel_CG17077, 0123/09, 0608/07, 0998/12, 3520, CG17077, D-Ets-2, D-ets-2, DMPOINT1A, Dmel\\CG17077, E(E2F)3D, EC3-2, EK3-2, ETS2, EY3-1, Ets, Ets2, Ets58AB, Ets94F, PNT-P1, PNTP1, PNTP2, POINT, Pnt, Pnt-P1, Pnt2, PntP1, PntP2, Pointed-P1, ets94F, l(3)07825, l(3)j1B7, l(3)s118306, png-P1, pnt-P2, pntP2, pntegfr, pointed-RC, ptd, pnt', description='pointed', other_designations='CG17077-PB|CG17077-PC|CG17077-PD|CG17077-PE|Ets at 58AB|Ets protein pointed P2|Point2|Pointed2|no terminal cell clones-R|pnt-PB|pnt-PC|pnt-PD|pnt-PE|pointed P2|pointed-P1|pointed-P2|pointedP2', map_location='3-79.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23290231, end_position_on_the_genomic_accession=23346167, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42758, CurrentID=0, Status='live', Symbol='DNApol-epsilon255', Aliases='Dmel_CG6768, 6768, BAB17608, CG6768, DNA pol epsilon, DNApol-epsilon, Dmel\\CG6768, dpolepsilon, dpolepsilonp255, l(3)pl10, l(3)pl10R, pl[10], pol-epsilon', description='DNA polymerase epsilon 255kD subunit', other_designations='CG6768-PA|DNA polymerase epsilon|DNApol-epsilon|DNApol-epsilon255-PA|DNApolymarase E|lethal (3) pl10', map_location='94E13-94E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23348128, end_position_on_the_genomic_accession=23355163, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42759, CurrentID=0, Status='live', Symbol='ATPsynCF6', Aliases='Dmel_CG4412, ATPR, ATPsyn-Cf6, BcDNA:RH08870, CG4412, Dmel\\CG4412', description='ATP synthase, coupling factor 6', other_designations='ATP synthase coupling factor 6|ATPase coupling factor 6|ATPase coupling-factor 6|ATPsynCF6-PA|ATPsynCF6-PB|ATPsynCF6-PC|CG4412-PA|CG4412-PB|CG4412-PC|F0/F1 ATP synthase coupling factor 6', map_location='94E13-94E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23355220, end_position_on_the_genomic_accession=23356064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42761, CurrentID=0, Status='live', Symbol='RpS3', Aliases='Dmel_CG6779, 6779, CG6779, DmRpS3, Dmel\\CG6779, M, M(3)124, M(3)95A, M(3)B, M(3)B2, M(3)Fla, M(3)w, M(3R)w, M95A, M[[w]], Mw, RPS3, RS3, Rp S3, Rps3, S3, X72921, chr3R:19185518..19185701, dS3, rpS3, rps3', description='Ribosomal protein S3', other_designations='CG6779-PA|CG6779-PB|Minute(3)95A|RpS3-PA|RpS3-PB|minute|ribosomal protein S3|rps3 gene', map_location='3-79.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23358812, end_position_on_the_genomic_accession=23360207, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42762, CurrentID=0, Status='live', Symbol='CG4408', Aliases='Dmel_ BEST:GH09055, Dmel\\CG4408', description='CG4408 gene product from transcript CG4408-RA', other_designations='CG4408-PA|CG4408-PB|Dmel_CG4408', map_location='94E13-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23361889, end_position_on_the_genomic_accession=23363574, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42763, CurrentID=0, Status='live', Symbol='CG42827', Aliases='Dmel_ CG6784, Dmel\\CG42827, Dmel_CG6784', description='CG42827 gene product from transcript CG42827-RC', other_designations='CG42827-PC|CG42827-PD|uncharacterized protein', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23372924, end_position_on_the_genomic_accession=23373600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42764, CurrentID=0, Status='live', Symbol='CG4393', Aliases='Dmel_ Dmel\\CG4393', description='CG4393 gene product from transcript CG4393-RD', other_designations='CG4393-PB|CG4393-PC|CG4393-PD', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23375928, end_position_on_the_genomic_accession=23384777, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  42765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42765, CurrentID=0, Status='live', Symbol='Ir94e', Aliases='Dmel_CG42298, CG17379, CG42298, DmelIR94e, Dmel\\CG42298, Dmel_CG17379, IR94e', description='Ionotropic receptor 94e', other_designations='CG42298-PA|Ir94e-PA|ionotropic receptor 94e', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23405509, end_position_on_the_genomic_accession=23407402, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42766, CurrentID=0, Status='live', Symbol='CG17380', Aliases='Dmel_ BcDNA:AT18288, Dmel\\CG17380', description='CG17380 gene product from transcript CG17380-RB', other_designations='CG17380-PB|uncharacterized protein', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23402608, end_position_on_the_genomic_accession=23403728, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42767, CurrentID=0, Status='live', Symbol='CG4374', Aliases='Dmel_ Dmel\\CG4374', description='CG4374 gene product from transcript CG4374-RB', other_designations='CG4374-PB|CG4374-PC', map_location='94F1-94F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23408684, end_position_on_the_genomic_accession=23472278, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42768, CurrentID=0, Status='live', Symbol='Ir94g', Aliases='Dmel_CG17381, CG17381, DmelIR94g, Dmel\\CG17381, IR94g', description='Ionotropic receptor 94g', other_designations='CG17381-PB|Ir94g-PB|ionotropic receptor 94g', map_location='94F6-94F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23483863, end_position_on_the_genomic_accession=23485500, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42769, CurrentID=0, Status='live', Symbol='Ir94h', Aliases='Dmel_CG17382, CG17382, DmelIR94h, Dmel\\CG17382, IR94h', description='Ionotropic receptor 94h', other_designations='CG17382-PB|Ir94h-PB|ionotropic receptor 94h', map_location='94F6-94F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23487501, end_position_on_the_genomic_accession=23489180, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42770, CurrentID=0, Status='live', Symbol='Irk2', Aliases='Dmel_CG4370, CG10180, CG4370, Dmel\\CG4370, dKir, dKirII, irk2', description='Inwardly rectifying potassium channel 2', other_designations='CG4370-PA|CG4370-PB|CG4370-PC|CG4370-PD|Irk2-PA|Irk2-PB|Irk2-PC|Irk2-PD|inwardly rectifying potassium channel 2', map_location='95A1-95A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23513617, end_position_on_the_genomic_accession=23519445, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42772, CurrentID=0, Status='live', Symbol='CG10177', Aliases='Dmel_ Dmel\\CG10177', description='CG10177 gene product from transcript CG10177-RA', other_designations='CG10177-PA|MLCK-like|uncharacterized protein', map_location='95A1-95A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23520146, end_position_on_the_genomic_accession=23521583, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42773, CurrentID=0, Status='live', Symbol='CG10175', Aliases='Dmel_ Dmel\\CG10175', description='CG10175 gene product from transcript CG10175-RE', other_designations='CG10175-PA|CG10175-PB|CG10175-PC|CG10175-PD|CG10175-PE|uncharacterized protein', map_location='95A1-95A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23522264, end_position_on_the_genomic_accession=23530986, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42774, CurrentID=0, Status='live', Symbol='CG10170', Aliases='Dmel_ Dmel\\CG10170', description='CG10170 gene product from transcript CG10170-RA', other_designations='CG10170-PA|uncharacterized protein', map_location='95A1-95A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23532945, end_position_on_the_genomic_accession=23534767, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42775, CurrentID=0, Status='live', Symbol='CG16732', Aliases='Dmel_ Dmel\\CG16732', description='CG16732 gene product from transcript CG16732-RB', other_designations='CG16732-PA|CG16732-PB|uncharacterized protein', map_location='95A1-95A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23534900, end_position_on_the_genomic_accession=23536710, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42776, CurrentID=0, Status='live', Symbol='CG10168', Aliases='Dmel_ Dmel\\CG10168', description='CG10168 gene product from transcript CG10168-RA', other_designations='CG10168-PA|uncharacterized protein', map_location='95A1-95A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23536993, end_position_on_the_genomic_accession=23538908, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42777, CurrentID=0, Status='live', Symbol='CG10164', Aliases='Dmel_ Dmel\\CG10164, NEST:bs02f09', description='CG10164 gene product from transcript CG10164-RB', other_designations='CG10164-PA|CG10164-PB|uncharacterized protein', map_location='95A2-95A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23554323, end_position_on_the_genomic_accession=23556943, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42778, CurrentID=0, Status='live', Symbol='beat-IV', Aliases='Dmel_CG10152, Beat-IV, CG10152, CG13823, Dmel\\CG10152, beat IV, beatIV', description='CG10152 gene product from transcript CG10152-RB', other_designations='CG10152-PA|CG10152-PB|beat-IV-PA|beat-IV-PB', map_location='95A2-95A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23557412, end_position_on_the_genomic_accession=23580077, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42780, CurrentID=0, Status='live', Symbol='CG10182', Aliases='Dmel_ Dmel\\CG10182', description='CG10182 gene product from transcript CG10182-RA', other_designations='CG10182-PA|uncharacterized protein', map_location='95A2-95A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23587120, end_position_on_the_genomic_accession=23589819, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42781, CurrentID=0, Status='live', Symbol='CG16723', Aliases='Dmel_ Dmel\\CG16723', description='CG16723 gene product from transcript CG16723-RB', other_designations='CG16723-PB|uncharacterized protein', map_location='95A3-95A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23593625, end_position_on_the_genomic_accession=23596273, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42782, CurrentID=0, Status='live', Symbol='CG10183', Aliases='Dmel_ Dmel\\CG10183', description='CG10183 gene product from transcript CG10183-RB', other_designations='CG10183-PB|uncharacterized protein', map_location='95A3-95A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23596561, end_position_on_the_genomic_accession=23599034, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42783, CurrentID=0, Status='live', Symbol='CG10184', Aliases='Dmel_ Dmel\\CG10184', description='CG10184 gene product from transcript CG10184-RA', other_designations='CG10184-PA|uncharacterized protein', map_location='95A4-95A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23602507, end_position_on_the_genomic_accession=23605495, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42784, CurrentID=0, Status='live', Symbol='CG31145', Aliases='Dmel_ CG10187, CG13820, CG13821, Dmel\\CG31145', description='CG31145 gene product from transcript CG31145-RC', other_designations='CG31145-PA|CG31145-PB|CG31145-PC|CG31145-PD|CG31145-PF|uncharacterized protein', map_location='95A4-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23605751, end_position_on_the_genomic_accession=23669659, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  42787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42787, CurrentID=0, Status='live', Symbol='GILT3', Aliases='Dmel_CG13822, CG13822, Dmel\\CG13822, GILT', description='Gamma-interferon-inducible lysosomal thiol reductase 3', other_designations='CG13822-PA|GILT3-PA|Gamma-interferon-inducible lysosomal thiol reductase', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23677358, end_position_on_the_genomic_accession=23678421, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42788, CurrentID=0, Status='live', Symbol='GILT2', Aliases='Dmel_CG10157, CG10157, Dmel\\CG10157, GILT, anon-EST:Posey120', description='Gamma-interferon-inducible lysosomal thiol reductase 2', other_designations='CG10157-PA|GILT2-PA|Gamma-interferon-inducible lysosomal thiol reductase', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23678895, end_position_on_the_genomic_accession=23679855, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42789, CurrentID=0, Status='live', Symbol='eIF-3p66', Aliases='Dmel_CG10161, 153314_at, BEST:GM10514, CG10161, Dmel\\CG10161, GM04979, GM10514, eIF3-S7, eIF3d, eIF3p66, eiF3p66', description='Eukaryotic initiation factor 3 p66 subunit', other_designations='CG10161-PB|eIF-3p66-PB|eukaryotic initiation factor 3 p66 subunit', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23680039, end_position_on_the_genomic_accession=23682317, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42790, CurrentID=0, Status='live', Symbol='CG16710', Aliases='Dmel_ Dmel\\CG16710, SP16', description='CG16710 gene product from transcript CG16710-RB', other_designations='CG16710-PB|uncharacterized protein', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23682503, end_position_on_the_genomic_accession=23683980, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42791, CurrentID=0, Status='live', Symbol='SPE', Aliases='Dmel_CG16705, CG16705, CT37173, Dmel\\CG16705, SP4, Sp4, c-SP4, proPO-AE, spe', description='Spatzle-Processing Enzyme', other_designations='CG16705-PA|SPE-PA|Spatzle-Processing enzyme|Spz-processing enzyme|spz processing enzyme', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23686255, end_position_on_the_genomic_accession=23688147, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42793, CurrentID=0, Status='live', Symbol='CG10254', Aliases='Dmel_ BcDNA:LD22087, Dmel\\CG10254', description='CG10254 gene product from transcript CG10254-RB', other_designations='CG10254-PA|CG10254-PB|CG10254-PC|uncharacterized protein', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23689007, end_position_on_the_genomic_accession=23694983, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42794, CurrentID=0, Status='live', Symbol='CG10252', Aliases='Dmel_ BEST:GH13953, Dmel\\CG10252, anon-WO0140519.133, bs30h03.y1', description='CG10252 gene product from transcript CG10252-RA', other_designations='CG10252-PA|uncharacterized protein', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23695221, end_position_on_the_genomic_accession=23696245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42795, CurrentID=0, Status='live', Symbol='VAChT', Aliases='Dmel_CG32848, CG12345, CG32848, CT41182, Chn, Dmel\\CG32848, VACHT, VACht, VaChT, Vacht, chn, vAChT, vacht', description='Vesicular acetylcholine transporter', other_designations="CG32848-PA|Cha's neighbour|VAChT-PA|vesicular acetylcholine transporter", map_location='91C7-97D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18705501, end_position_on_the_genomic_accession=18718916, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42796, CurrentID=0, Status='live', Symbol='CG31148', Aliases='Dmel_ CG10299, CT28913, Dmel\\CG31148, anon-WO0140519.234', description='CG31148 gene product from transcript CG31148-RA', other_designations='CG31148-PA|glucosylceramidase|uncharacterized protein', map_location='95A8-95A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23700621, end_position_on_the_genomic_accession=23702605, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42797, CurrentID=0, Status='live', Symbol='CG10301', Aliases='Dmel_ Dmel\\CG10301', description='CG10301 gene product from transcript CG10301-RB', other_designations='CG10301-PB|uncharacterized protein', map_location='95A10-95A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23710158, end_position_on_the_genomic_accession=23711338, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42798, CurrentID=0, Status='live', Symbol='CG10300', Aliases='Dmel_ Dmel\\CG10300', description='CG10300 gene product from transcript CG10300-RA', other_designations='CG10300-PA|uncharacterized protein', map_location='95A10-95A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23711527, end_position_on_the_genomic_accession=23713313, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42799, CurrentID=0, Status='live', Symbol='nau', Aliases='Dmel_CG10250, CG10250, Dmel\\CG10250, Dmyd, MYOD, Mdrf, MyoD, NAU, Nau', description='nautilus', other_designations='CG10250-PA|CG10250-PB|myogenic determination|nau-PA|nau-PB', map_location='95A10-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23712814, end_position_on_the_genomic_accession=23718357, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42800, CurrentID=0, Status='live', Symbol='CG10232', Aliases='Dmel_ Dmel\\CG10232, SP14', description='CG10232 gene product from transcript CG10232-RD', other_designations='CG10232-PD|uncharacterized protein', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23719343, end_position_on_the_genomic_accession=23721625, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42801, CurrentID=0, Status='live', Symbol='CG10365', Aliases='Dmel_ Dmel\\CG10365, anon-WO0118547.386', description='CG10365 gene product from transcript CG10365-RC', other_designations='CG10365-PA|CG10365-PB|CG10365-PC|CG10365-PD|uncharacterized protein', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23723331, end_position_on_the_genomic_accession=23731499, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42802, CurrentID=0, Status='live', Symbol='Rpn9', Aliases='Dmel_CG10230, 0718/06, 718/06, CG10230, CT28749, Dmel\\CG10230, PSMD13, Rpn44, S11, l(3)S071806, l(3)S071806b, p39A, rpn9, Rpn9', description='Regulatory particle non-ATPase 9', other_designations='CG10230-PA|CG10230-PB|Rpn9-PA|Rpn9-PB|nobody|p39A|regulatory particle non-ATPase 9', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23731375, end_position_on_the_genomic_accession=23733462, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42803, CurrentID=0, Status='live', Symbol='Hmgcr', Aliases='Dmel_CG10367, BcDNA:LD15354, CG10367, Clb, DmHMG-CoAr, Dmel\\CG10367, HMGCR, HMGCR_Dm, HMGCoAR, HMGCoARr, HMGCoAr, HMGR, HmG-CoAR, Hmg, bro1, bs32f05.y1, clb, hmgc, hmgcr, hmgrc, l(3)01152, l(3)04684', description='HMG Coenzyme A reductase', other_designations='3-Hydroxy-3-Methylglutaryl CoA reductase|3-hydroxy-3-methylglutaryl coenzyme A reductase|CG10367-PA|HMG coenzyme A reductase|HMG-CoA reductase|HMG-CoA reductase class I|HMGCo-A reductase|HMGCoA reductase|Hmg-CoA-reductase|Hmgcr-PA|columbus|lethal (3) 04684', map_location='3-80.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23733324, end_position_on_the_genomic_accession=23750992, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42804, CurrentID=0, Status='live', Symbol='RanBP3', Aliases='Dmel_CG10225, CG10225, Dmel\\CG10225, GC10225', description='Ran binding protein 3', other_designations='CG10225-PA|CG10225-PB|RanBP3-PA|RanBP3-PB', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23750937, end_position_on_the_genomic_accession=23752834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42805, CurrentID=0, Status='live', Symbol='Rpt5', Aliases="Dmel_CG10370, BcDNA.GH12068, BcDNA:GH12068, CG10370, DmTBP-1, Dm_Rpt5a, Dm_Rpt5b, Dmel\\CG10370, Pros26S, Pros26S-RS6A, S6', TBP-1, Tbp-1, Tbp1, anon-WO0153538.10, anon-WO0153538.8, anon-WO0153538.9, dRpt5, l(3)04210, p50, tbp-1", description='Regulatory particle triple-A ATPase 5', other_designations='CG10370-PA|Rpt5-PA|Tat-binding protein-1|proteasome 26S regulatory subunit 6A|regulatory particle triple-A ATPase 5', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23752876, end_position_on_the_genomic_accession=23754623, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42806, CurrentID=0, Status='live', Symbol='Hrd3', Aliases='Dmel_CG10221, BcDNA:LD23587, CG10221, Dmel\\CG10221, HRD3', description='CG10221 gene product from transcript CG10221-RB', other_designations='CG10221-PA|CG10221-PB|Hrd3-PA|Hrd3-PB', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23754623, end_position_on_the_genomic_accession=23758557, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42807, CurrentID=0, Status='live', Symbol='Plip', Aliases='Dmel_CG10371, 150466_at, CG10371, Dmel\\CG10371, PLIP, dPTEN, plip', description='PTEN-like phosphatase', other_designations='CG10371-PA|CG10371-PB|MKP-like|Plip-PA|Plip-PB', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23758339, end_position_on_the_genomic_accession=23759606, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42808, CurrentID=0, Status='live', Symbol='SdhD', Aliases='Dmel_CG10219, BEST:GH27735, CG10219, DHSD, Dmel\\CG10219', description='Succinate dehydrogenase, subunit D', other_designations='CG10219-PA|SUCCINATE DEHYDROGENASE (UBIQUINONE) CYTOCHROME B SMALL SUBUNIT|SdhD-PA|succinate dehydrogenase [ubiquinone] cytochrome B small subunit|succinate dehydrogenase cytochrome b small subunit|succinate dehydrogenase, subunit D', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23759483, end_position_on_the_genomic_accession=23760600, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42809, CurrentID=0, Status='live', Symbol='CG10217', Aliases='Dmel_ Dmel\\CG10217', description='CG10217 gene product from transcript CG10217-RB', other_designations='CG10217-PA|CG10217-PB|uncharacterized protein', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23760737, end_position_on_the_genomic_accession=23763263, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42810, CurrentID=0, Status='live', Symbol='Lsd-1', Aliases='Dmel_CG10374, CG10374, DmPLIN1, Dmel\\CG10374, LSD-1, LSDP1, Lsd1, Lsdp1, PLIN1, lsd1, plin1', description='Lipid storage droplet-1', other_designations='CG10374-PA|CG10374-PB|CG10374-PC|CG10374-PD|Lsd-1-PA|Lsd-1-PB|Lsd-1-PC|Lsd-1-PD|lipid storage droplet protein 1|lipid storage droplet-1|lipid storage droplets surface binding protein|perilipin', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23763857, end_position_on_the_genomic_accession=23766662, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42811, CurrentID=0, Status='live', Symbol='CG10214', Aliases='Dmel_ Dmel\\CG10214', description='CG10214 gene product from transcript CG10214-RA', other_designations='CG10214-PA|uncharacterized protein', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23766519, end_position_on_the_genomic_accession=23767488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42812, CurrentID=0, Status='live', Symbol='CG10375', Aliases='Dmel_ Dmel\\CG10375', description='CG10375 gene product from transcript CG10375-RA', other_designations='CG10375-PA|CG10375-PB|uncharacterized protein', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23767712, end_position_on_the_genomic_accession=23768984, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42813, CurrentID=0, Status='live', Symbol='tst', Aliases='Dmel_CG10210, CG10210, Dmel\\CG10210, SKI2, Ski2, Ski2/tst, cg10210', description='twister', other_designations='CG10210-PA|CG10210-PB|tst-PA|tst-PB', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23768399, end_position_on_the_genomic_accession=23772859, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42814, CurrentID=0, Status='live', Symbol='CG10208', Aliases='Dmel_ CT28701, Dmel\\CG10208', description='CG10208 gene product from transcript CG10208-RA', other_designations='CG10208-PA|uncharacterized protein', map_location='95B1-95B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23772998, end_position_on_the_genomic_accession=23774127, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42816, CurrentID=0, Status='live', Symbol='Nup98-96', Aliases='Dmel_CG10198, CG10198, CG10201, Dmel\\CG10198, NUP98, Nup, Nup 96, Nup 98, Nup-98, Nup96, Nup96-98, Nup98, Nup98-Nup96, Nup98/96, l(3)95BCd, nup145, nup98-96', description='Nucleoporin 98-96kD', other_designations='CG10198-PA|CG10198-PC|CG10198-PD|Nup98|Nup98-96-PA|Nup98-96-PC|Nup98-96-PD|nucleoporin 98|nucleoporin 98-96|nucleoporin 98-96kD|nucleoporin98-96', map_location='95B1-95B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23774225, end_position_on_the_genomic_accession=23781621, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42817, CurrentID=0, Status='live', Symbol='mbc', Aliases='Dmel_CG10379, CG10379, DOCK180, Dm MBC, Dmel\\CG10379, Dock180, MBC, MBC/DOCK180, Mbc, Su(rac)1, dock180', description='myoblast city', other_designations='CG10379-PA|CG10379-PB|mbc-PA|mbc-PB', map_location='95B5-95B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23781765, end_position_on_the_genomic_accession=23801634, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  42819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42819, CurrentID=0, Status='live', Symbol='eIF4G2', Aliases='Dmel_CG10192, CG10192, Dmel\\CG10192, eIF4G-2, ms(3)nc32, nc32, ofs', description='eukaryotic translation initiation factor 4G2', other_designations='CG10192-PA|CG10192-PC|eIF-4G-like protein|eIF4G2-PA|eIF4G2-PC|male sterile (3) nc32|off-schedule', map_location='3-82.9 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23808285, end_position_on_the_genomic_accession=23819237, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42820, CurrentID=0, Status='live', Symbol='CG34355', Aliases='Dmel_ CG12492, CG13818, Dmel\\CG34355, Dmel_CG12492, lincRNA.840', description='CG34355 gene product from transcript CG34355-RF', other_designations='CG34355-PB|CG34355-PE|CG34355-PF|CG34355-PG|uncharacterized protein', map_location='95C1-95C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23803859, end_position_on_the_genomic_accession=23859167, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  42821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42821, CurrentID=0, Status='live', Symbol='Pli', Aliases='Dmel_CG5212, CG5212, Dmel\\CG5212, PAP, pli', description='Pellino', other_designations='CG5212-PA|CG5212-PB|Pli-PA|Pli-PB|pelle associated protein|pellino', map_location='95C5-95C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23860162, end_position_on_the_genomic_accession=23890052, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42822, CurrentID=0, Status='live', Symbol='TfIIA-S', Aliases='Dmel_CG5163, BcDNA:RE44302, CG5163, Dmel\\CG5163, F121, TF-IIAS, TFIIA, TFIIA-S, dTFIIA, dTFIIA-S, tfiia-s', description='Transcription-factor-IIA-S', other_designations='CG5163-PA|TfIIA-S-PA', map_location='95C8-95C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23890388, end_position_on_the_genomic_accession=23891060, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42823, CurrentID=0, Status='live', Symbol='tbrd-1', Aliases='Dmel_CG13597, CG13597, Dmel\\CG13597, tBRD-1', description='testis-specifically expressed bromodomain containing protein-1', other_designations='CG13597-PA|tbrd-1-PA', map_location='95C8-95C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23891597, end_position_on_the_genomic_accession=23893383, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42824, CurrentID=0, Status='live', Symbol='sba', Aliases='Dmel_CG13598, AAF56201, CG13598, DMTF125, Dm Sba, Dmel\\CG13598, F125, TF125, Tf125', description='six-banded', other_designations='CG13598-PC|CG13598-PD|CG13598-PE|CG13598-PF|CG13598-PG|CG13598-PH|CG13598-PI|TF125 type II|sba-PC|sba-PD|sba-PE|sba-PF|sba-PG|sba-PH|sba-PI', map_location='95C9-95C12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23894174, end_position_on_the_genomic_accession=23922973, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42825, CurrentID=0, Status='live', Symbol='Ndc1', Aliases='Dmel_CG5857, CG5857, Dmel\\CG5857, NDC1', description='Ndc1 ortholog (S. cerevisiae)', other_designations='CG5857-PA|Ndc1 ortholog|Ndc1-PA', map_location='95C12-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23924399, end_position_on_the_genomic_accession=23926436, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42826, CurrentID=0, Status='live', Symbol='CG13601', Aliases='Dmel_ Dmel\\CG13601', description='CG13601 gene product from transcript CG13601-RA', other_designations='CG13601-PA|CG13601-PB|uncharacterized protein', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23926372, end_position_on_the_genomic_accession=23927634, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42827, CurrentID=0, Status='live', Symbol='CG43999', Aliases='Dmel_ BcDNA:AT03227, CG13602, CR43999, Dmel\\CG43999', description='CG43999 gene product from transcript CG43999-RC', other_designations='CG43999-PB|CG43999-PC|uncharacterized protein', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23927728, end_position_on_the_genomic_accession=23929000, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42828, CurrentID=0, Status='live', Symbol='Rpt2', Aliases='Dmel_CG5289, 5289, CG5289, Dm-S4-1, DmP26S4, Dm_Rpt2, Dmel\\CG5289, P26s4, PSMC1, Pros26.4, S4, dRpt2, p56, pros26.4', description='Regulatory particle triple-A ATPase 2', other_designations='26S proteasome subunit 4 ATPase|CG5289-PA|CG5289-PB|Rpt2-PA|Rpt2-PB|proteasome 26S|proteasome 26S subunit 4 ATPase|proteasome 26S subunit subunit 4 ATPase|regulatory particle triple-A ATPase 2', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23929873, end_position_on_the_genomic_accession=23931558, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42829, CurrentID=0, Status='live', Symbol='CG13599', Aliases='Dmel_ Dmel\\CG13599', description='CG13599 gene product from transcript CG13599-RA', other_designations='CG13599-PA|uncharacterized protein', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23931639, end_position_on_the_genomic_accession=23932728, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42830, CurrentID=0, Status='live', Symbol='RpS19b', Aliases='Dmel_CG5338, BcDNA:SD22440, CG5338, Dmel\\CG5338, RpS19', description='Ribosomal protein S19b', other_designations='CG5338-PB|CG5338-PC|RpS19b-PB|RpS19b-PC|ribosomal protein S19|ribosomal protein S19b', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23932729, end_position_on_the_genomic_accession=23933530, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42831, CurrentID=0, Status='live', Symbol='CG5854', Aliases='Dmel_ Dmel\\CG5854, anon-WO0118547.306', description='CG5854 gene product from transcript CG5854-RA', other_designations='CG5854-PA', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23933912, end_position_on_the_genomic_accession=23935964, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42832, CurrentID=0, Status='live', Symbol='Gdh', Aliases='Dmel_CG5320, CG5320, DHE3, Dmel\\CG5320, GDH, GLU-D, GLUD, GLUD pre-mRNA, Glud, gdh', description='Glutamate dehydrogenase', other_designations='CG5320-PA|CG5320-PB|CG5320-PD|CG5320-PF|Gdh-PA|Gdh-PB|Gdh-PD|Gdh-PF|NAD-dependent glutamate dehydrogenase|glutamate dehydrogenase', map_location='3-81.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23935958, end_position_on_the_genomic_accession=23943101, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42833, CurrentID=0, Status='live', Symbol='CG12268', Aliases='Dmel_ Dmel\\CG12268', description='CG12268 gene product from transcript CG12268-RB', other_designations='CG12268-PA|CG12268-PB|CG12268-PC|uncharacterized protein', map_location='95D1-95D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23943628, end_position_on_the_genomic_accession=23948424, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42834, CurrentID=0, Status='live', Symbol='Aats-glupro', Aliases='Dmel_CG5394, AATS, AATs-GluPro, Aa-tRNA-syn-glupro, Aats-glu, CG5394, Dmel\\CG5394, EPRS, ERS, FBpp0083898, Glu-ProRS, GluPro-RS, GluProRS, GluRS, glutamyl-prolyl-tRNA synthetase, p200', description='Glutamyl-prolyl-tRNA synthetase', other_designations='Aats-glupro-PA|Aats-glupro-PB|CG5394-PA|CG5394-PB|Glutamyl-tRNA synthetase|aminoacyl-tRNA synthetase|bifunctional aminoacyl-tRNA synthetase|glutamyl-prolyl-tRA synthetase|transfer RNA-Glu-Pro aminoacyl synthetase', map_location='95D1-95D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23952136, end_position_on_the_genomic_accession=23958391, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42835, CurrentID=0, Status='live', Symbol='AP-1sigma', Aliases='Dmel_CG5864, AP-1[sigma]1, AP1sigma1, CG5864, Dmel\\CG5864, aps1, cg5864, sigma1, AP-1sigma', description='Adaptor Protein complex 1, sigma subunit', other_designations='AP-1sigma-PA|AP-1sigma-PB|AP-1sigma-PC|AP-1sigma-PD|CG5864-PA|CG5864-PB|CG5864-PC|CG5864-PD|adaptor protein complex 1, sigma subunit', map_location='95D1-95D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23958651, end_position_on_the_genomic_accession=23962555, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42836, CurrentID=0, Status='live', Symbol='CG31140', Aliases='Dmel_ CG13600, CG31452, CG5875, DGKt, Dmel\\CG31140', description='CG31140 gene product from transcript CG31140-RB', other_designations='CG31140-PA|CG31140-PB|CG31140-PD|CG31140-PE|CG31140-PF|CG31140-PG|CG31140-PH|uncharacterized protein', map_location='95D1-95D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23963180, end_position_on_the_genomic_accession=23983232, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  42838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42838, CurrentID=0, Status='live', Symbol='CG13603', Aliases='Dmel_ Dmel\\CG13603', description='CG13603 gene product from transcript CG13603-RA', other_designations='CG13603-PA|uncharacterized protein', map_location='95D4-95D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23985213, end_position_on_the_genomic_accession=23986569, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42839, CurrentID=0, Status='live', Symbol='CG5902', Aliases='Dmel_ Dmel\\CG5902, anon-WO0118547.372', description='CG5902 gene product from transcript CG5902-RA', other_designations='CG5902-PA|CG5902-PB|CG5902-PC', map_location='95D4-95D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23986642, end_position_on_the_genomic_accession=23988738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42840, CurrentID=0, Status='live', Symbol='CG13604', Aliases='Dmel_ Dmel\\CG13604, UBASH3A', description='CG13604 gene product from transcript CG13604-RD', other_designations='CG13604-PA|CG13604-PB|CG13604-PC|CG13604-PD|CG13604-PE|uncharacterized protein', map_location='95D5-95D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23988806, end_position_on_the_genomic_accession=23992916, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42841, CurrentID=0, Status='live', Symbol='Rab7', Aliases='Dmel_CG5915, AAF56218, CG5915, DRAB7, D Dm Rab7, DmRab7, Dmel\\CG5915, Drab7, O76742, RAB7, Rab-7, Rab-r7, dRab7, drab7, rab7', description='CG5915 gene product from transcript CG5915-RA', other_designations='CG5915-PA|CG5915-PB|CG5915-PC|Rab-protein 7|Rab7-PA|Rab7-PB|Rab7-PC', map_location='95D5-95D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23993008, end_position_on_the_genomic_accession=23995452, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42842, CurrentID=0, Status='live', Symbol='LSm3', Aliases='Dmel_CG31184, BcDNA:RE11655, CG31184, CG5926, Dmel\\CG31184, Lsm3', description='CG31184 gene product from transcript CG31184-RA', other_designations='CG31184-PA|CG31184-PB|LSm3-PA|LSm3-PB', map_location='95D5-95D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23996262, end_position_on_the_genomic_accession=23997017, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42843, CurrentID=0, Status='live', Symbol='KrT95D', Aliases='Dmel_CG5405, CG5405, Dmel\\CG5405', description='Krueppel target at 95D', other_designations='CG5405-PA|CG5405-PB|CG5405-PC|CG5405-PD|CG5405-PF|KrT95D-PA|KrT95D-PB|KrT95D-PC|KrT95D-PD|KrT95D-PF|krueppel target at 95D', map_location='95D5-95D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23997538, end_position_on_the_genomic_accession=24031827, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  42844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42844, CurrentID=0, Status='live', Symbol='Ime4', Aliases='Dmel_CG5933, CG5933, Dmel\\CG5933, IME4, MTA70, Mta70, ime4', description='Inducer of meiosis 4', other_designations='CG5933-PA|Ime4-PA|Mta70 homologue|inducer of MEiosis 4 homolog|inducer of meiosis 4', map_location='95D8-95D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24032157, end_position_on_the_genomic_accession=24034257, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42845, CurrentID=0, Status='live', Symbol='Miro', Aliases='Dmel_CG5410, B682, CG5410, DmMiro, Dmel\\CG5410, Q8IMX7-1, Q8IMX7-2, dMiro, dmiro, miro', description='Mitochondrial Rho', other_designations='CG5410-PC|CG5410-PD|CG5410-PE|CG5410-PF|Miro-PC|Miro-PD|Miro-PE|Miro-PF|mitochondrial Rho|mitochondrial Rho-GTPase', map_location='95D8-95D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24034362, end_position_on_the_genomic_accession=24038847, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42846, CurrentID=0, Status='live', Symbol='spas', Aliases='Dmel_CG5977, CG5977, D-spastin, Dmel\\CG5977, Dspastin, Spas', description='spastin', other_designations='CG5977-PA|CG5977-PB|CG5977-PC|CG5977-PD|CG5977-PE|CG5977-PF|D-spastin|Dm-Spastin|dspastin|spas-PA|spas-PB|spas-PC|spas-PD|spas-PE|spas-PF|spasin', map_location='95D9-95D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24038215, end_position_on_the_genomic_accession=24042982, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42848, CurrentID=0, Status='live', Symbol='Rox8', Aliases='Dmel_CG5422, CG5422, DmTIAR, Dmel\\CG5422, ROX8, RRM12, Rbp12, Rox-8, TIAR, dTIAR, rox8, rrm12', description='CG5422 gene product from transcript CG5422-RD', other_designations='CG5422-PB|CG5422-PC|CG5422-PD|CG5422-PE|CG5422-PG|CG5422-PH|RNA-binding protein 12|Rox8-PB|Rox8-PC|Rox8-PD|Rox8-PE|Rox8-PG|Rox8-PH', map_location='95D9-95D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24043056, end_position_on_the_genomic_accession=24048265, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42849, CurrentID=0, Status='live', Symbol='CG5991', Aliases='Dmel_ BEST:LD21713, Dmel\\CG5991, LD21713, Psd', description='CG5991 gene product from transcript CG5991-RA', other_designations='CG5991-PA|CG5991-PB|CG5991-PC|phosphatidylserine decarboxylase', map_location='95D10-95D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24050783, end_position_on_the_genomic_accession=24053592, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42850, CurrentID=0, Status='live', Symbol='Atg6', Aliases='Dmel_CG5429, ATG6, Beclin-1, Beclin1, CG5429, DmAtg6, Dmel\\CG5429, atg6, beclin, l(3)00096', description='Autophagy-related 6', other_designations='Atg6-PA|Autophagy-specific gene 6|Beclin1|CG5429-PA|beclin|beclin 1', map_location='95D10-95D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24049180, end_position_on_the_genomic_accession=24050782, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42851, CurrentID=0, Status='live', Symbol='CG6000', Aliases='Dmel_ BcDNA:RE30346, Dmel\\CG6000', description='CG6000 gene product from transcript CG6000-RD', other_designations='CG6000-PC|CG6000-PD', map_location='95D11-95D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24054391, end_position_on_the_genomic_accession=24055092, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42852, CurrentID=0, Status='live', Symbol='Hsp68', Aliases='Dmel_CG5436, 68, CG5436, DmHsp68, Dmel\\CG5436, HSP68, HSP68_DROME, Hsp-68, hsp68', description='Heat shock protein 68', other_designations='CG5436-PA|Hsp 68|Hsp68-PA|heat shock protein 68', map_location='95D11-95D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24055080, end_position_on_the_genomic_accession=24057310, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42853, CurrentID=0, Status='live', Symbol='SMC1', Aliases='Dmel_CG6057, AAF56231, CG6057, DmSMC1, Dmel\\CG6057, Smc1, dSMC1, smc1', description='Structural maintenance of chromosomes 1', other_designations='CG6057-PA|SMC1-PA|Smc1 cohesin subunit|cohesin|cohesin complex subunit SMC1|cohesin subunit SMC1|structural maintenance of chromosomes 1', map_location='95D11-95D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24058287, end_position_on_the_genomic_accession=24062815, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42854, CurrentID=0, Status='live', Symbol='Syx1A', Aliases='Dmel_CG31136, CG10716, CG18615, CG31136, CG31136/CG33100, CG5448, CT30033, Dm Syx1, DmSyx1A, Dmel\\CG31136, SYX, SYX 1A, SYX1A, Syt1A, Syx, Syx-1A, Syx1, Syx1a, SyxA, anon-EST:Gibbs4, anon-WO02059370.54, dSyx1, dsyn-1A, l(3)06737, synt, syt-1A, syx, syx-1, syx-1A, syx1, syx1A', description='Syntaxin 1A', other_designations='CG31136-PA|CG31136-PB|CG31136-PC|STX1A|Syx1A-PA|Syx1A-PB|Syx1A-PC|dsyntaxin-1A|syntaxin|syntaxin 1|syntaxin 1A|syntaxin-1|syntaxin-1A|syntaxin1|syntaxin1A|syntx 1', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24101600, end_position_on_the_genomic_accession=24105763, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42855, CurrentID=0, Status='live', Symbol='CG10694', Aliases='Dmel_ Dmel\\CG10694', description='CG10694 gene product from transcript CG10694-RA', other_designations='CG10694-PA|uncharacterized protein', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24107391, end_position_on_the_genomic_accession=24108358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42857, CurrentID=0, Status='live', Symbol='CG18428', Aliases='Dmel_ Dmel\\CG18428', description='CG18428 gene product from transcript CG18428-RB', other_designations='CG18428-PA|CG18428-PB|CG18428-PC|uncharacterized protein', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24109367, end_position_on_the_genomic_accession=24110010, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42858, CurrentID=0, Status='live', Symbol='CG13605', Aliases='Dmel_ Dmel\\CG13605', description='CG13605 gene product from transcript CG13605-RB', other_designations='CG13605-PB|uncharacterized protein', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24109944, end_position_on_the_genomic_accession=24112389, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42859, CurrentID=0, Status='live', Symbol='CG13607', Aliases='Dmel_ 150482_at, Dmel\\CG13607', description='CG13607 gene product from transcript CG13607-RC', other_designations='CG13607-PA|CG13607-PC|uncharacterized protein', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24114043, end_position_on_the_genomic_accession=24116322, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42860, CurrentID=0, Status='live', Symbol='Rootletin', Aliases='Dmel_CG6129, CG6129, Dmel\\CG6129, Q8IMX3_DROME', description='CG6129 gene product from transcript CG6129-RE', other_designations='CG6129-PD|CG6129-PE|CG6129-PF|Rootletin-PD|Rootletin-PE|Rootletin-PF|rootletin', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24113286, end_position_on_the_genomic_accession=24128193, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  42861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42861, CurrentID=0, Status='live', Symbol='CG5463', Aliases='Dmel_ Dmel\\CG5463', description='CG5463 gene product from transcript CG5463-RA', other_designations='CG5463-PA', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24128049, end_position_on_the_genomic_accession=24130255, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42862, CurrentID=0, Status='live', Symbol='Tsc1', Aliases='Dmel_CG6147, CG6147, Dmel\\CG6147, TSC, TSC1, dTSC, dTSC-1, dTSC1, d dtsc1, rocky, tsc1', description='CG6147 gene product from transcript CG6147-RA', other_designations='CG6147-PA|TTuberous sclerosis complex 1|Tsc1-PA|hamartin|jolly green giant|rocky|tuberous sclerosis 1|tuberous sclerosis complex 1', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24130307, end_position_on_the_genomic_accession=24134654, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42863, CurrentID=0, Status='live', Symbol='Sec10', Aliases='Dmel_CG6159, CG6159, Dmel\\CG6159, P71p, dSec10, dsec10, sec10, Sec10', description='Sec10 ortholog (S. cerevisiae)', other_designations='CG6159-PA|P71|Sec10 ortholog|Sec10-PA|sec10', map_location='95E-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24134831, end_position_on_the_genomic_accession=24137538, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42864, CurrentID=0, Status='live', Symbol='Npc2f', Aliases='Dmel_CG6164, BcDNA:LP08842, CG6164, Dmel\\CG6164, npc2f', description='Niemann-Pick type C-2f', other_designations='CG6164-PA|Npc2f-PA', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24137895, end_position_on_the_genomic_accession=24140936, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42865, CurrentID=0, Status='live', Symbol='Kal1', Aliases='Dmel_CG6173, CG6173, CT19368, DmKal-1, Dmel\\CG6173, Dmkal, Kal-1, kal-1', description='Kallmann syndrome 1 ortholog', other_designations='CG6173-PA|Kal1-PA|kal-1|kallmann|kallmann syndrome 1 ortholog', map_location='95E1-95E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24141395, end_position_on_the_genomic_accession=24147490, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42866, CurrentID=0, Status='live', Symbol='p38a', Aliases='Dmel_CG5475, CG5475, D-MPK2, D-P38a, D-p38, D-p38 MAPK, D-p38a, DMPK2, DmMPK2, Dmel\\CG5475, Dmp38a, Dp38, Dp38a, ERK2, Erk2, MAPK2, Mpk2, P38, dMPK2, dp38, dp38a, group 2, mpk2, p38, p38 alpha, p38A, p38Kalpha, p38a', description='p38a MAP kinase', other_designations='CG5475-PA|CG5475-PB|CG5475-PC|Erk-kinase-like|Mpk2|p38 MAP kinase|p38 mitogen-activated protein kinase|p38a-PA|p38a-PB|p38a-PC|stress-activated kinase P38', map_location='95E5-95E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24150416, end_position_on_the_genomic_accession=24152321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42867, CurrentID=0, Status='live', Symbol='CG6178', Aliases='Dmel_ Dmel\\CG6178', description='CG6178 gene product from transcript CG6178-RA', other_designations='CG6178-PA|fatty Acyl-CoA synthetase', map_location='95E5-95E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24152347, end_position_on_the_genomic_accession=24154896, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42869, CurrentID=0, Status='live', Symbol='CG6182', Aliases='Dmel_ BcDNA:LD17019, Dmel\\CG6182', description='CG6182 gene product from transcript CG6182-RA', other_designations='CG6182-PA', map_location='95E6-95E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24161727, end_position_on_the_genomic_accession=24163215, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42870, CurrentID=0, Status='live', Symbol='mRpS24', Aliases='Dmel_CG13608, CG13608, Dmel\\CG13608, MRP-S24, MrpS24', description='mitochondrial ribosomal protein S24', other_designations='CG13608-PA|mRpS24-PA', map_location='95E6-95E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24162829, end_position_on_the_genomic_accession=24163545, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42871, CurrentID=0, Status='live', Symbol='Apc2', Aliases='Dmel_CG6193, APC, APC2, CG6193, D-APC2, Dm APC2, Dmel\\CG6193, E-APC, E-APC dAPC2, apc2, d-APC2, dAPC2, dAPC2/E-APC, dApc2, dapc2, e-apc', description='Adenomatous polyposis coli tumor suppressor homolog 2', other_designations='Apc2-PA|CG6193-PA|adenomatous polyposis coli 2|adenomatous polyposis coli tumor suppressor homolog 2|epithelial adenomatous polyposis coli', map_location='95E6-95E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24163966, end_position_on_the_genomic_accession=24167763, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42872, CurrentID=0, Status='live', Symbol='CG5510', Aliases='Dmel_ Dmel\\CG5510', description='CG5510 gene product from transcript CG5510-RA', other_designations='CG5510-PA', map_location='95E6-95E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24167675, end_position_on_the_genomic_accession=24169562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42873, CurrentID=0, Status='live', Symbol='CG13606', Aliases='Dmel_ Dmel\\CG13606', description='CG13606 gene product from transcript CG13606-RD', other_designations='CG13606-PA|CG13606-PB|CG13606-PC|CG13606-PD|uncharacterized protein', map_location='95E7-95E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24172628, end_position_on_the_genomic_accession=24184144, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42874, CurrentID=0, Status='live', Symbol='CHORD', Aliases='Dmel_CG6198, CG6198, Dmel\\CG6198, Q9U4A3, chp, mora', description='CG6198 gene product from transcript CG6198-RA', other_designations='CG6198-PA|CHORD containing protein|CHORD-PA|morgana|morgana/chp-1', map_location='95E8-95E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24186359, end_position_on_the_genomic_accession=24187785, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42875, CurrentID=0, Status='live', Symbol='CG5515', Aliases='Dmel_ Dmel\\CG5515', description='CG5515 gene product from transcript CG5515-RC', other_designations='CG5515-PA|CG5515-PB|CG5515-PC|CG5515-PD', map_location='95E8-95E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24187724, end_position_on_the_genomic_accession=24188896, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42876, CurrentID=0, Status='live', Symbol='jnj', Aliases='Dmel_CG5524, AAF56254, CG5524, Dmel\\CG5524, Huc-95DE, SMC6, Smc6, huc95E, huc[95D-E], huc[95DE], huc[95E]', description='java no jive', other_designations='CG5524-PA|HU- and caffeine-dependent 95D-E|HU- or caffeine-sensitive 95E|hydroxyurea- and (or) caffeine-sensitive 95E|jnj-PA', map_location='95E8-95F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24189050, end_position_on_the_genomic_accession=24193423, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  42877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42877, CurrentID=0, Status='live', Symbol='CG6204', Aliases='Dmel_ Dmel\\CG6204, GH20028p', description='CG6204 gene product from transcript CG6204-RA', other_designations='CG6204-PA', map_location='95F1-95F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24193832, end_position_on_the_genomic_accession=24196880, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42879, CurrentID=0, Status='live', Symbol='CG17786', Aliases='Dmel_ Dmel\\CG17786', description='CG17786 gene product from transcript CG17786-RA', other_designations='CG17786-PA|uncharacterized protein', map_location='95F2-95F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24214631, end_position_on_the_genomic_accession=24218718, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42880, CurrentID=0, Status='live', Symbol='twin', Aliases='Dmel_CG31137, BEST:GH20274, CCR4, CG17741, CG31137, CG5534, CT39331, Ccr4, Dmel\\CG31137, Twin, ccr4, twn', description='CG31137 gene product from transcript CG31137-RE', other_designations='CG31137-PA|CG31137-PB|CG31137-PC|CG31137-PD|CG31137-PE|CG31137-PF|ccr4|twin-PA|twin-PB|twin-PC|twin-PD|twin-PE|twin-PF', map_location='95F1-95F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24196893, end_position_on_the_genomic_accession=24221965, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42881, CurrentID=0, Status='live', Symbol='cav', Aliases='Dmel_CG6219, 1G5, Anon1G5, CG6219, Caravaggio, Dmel\\CG6219, HOAP, Hoap, anon fe 1G5, anon-EST:fe1G5, anon1G5, hoap, p55', description='caravaggio', other_designations='CG6219-PA|CG6219-PB|CG6219-PD|CG6219-PE|HP1 and Orc-associated protein|HP1/ORC-associated protein|anon-fast-evolving-1G5|anonymous fast evolving gene IG5|cav-PA|cav-PB|cav-PD|cav-PE|drosophila anonymous fast evolving 1G5', map_location='95F2-95F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24221103, end_position_on_the_genomic_accession=24223528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42882, CurrentID=0, Status='live', Symbol='Spps', Aliases='Dmel_CG5669, CG5669, Dmel\\CG5669', description='Sp1-like factor for pairing sensitive-silencing', other_designations='CG5669-PA|CG5669-PB|Sp1-like factor for pairing Sensitive-silencing|Spps-PA|Spps-PB', map_location='95F2-95F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24223445, end_position_on_the_genomic_accession=24226934, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42883, CurrentID=0, Status='live', Symbol='CG13609', Aliases='Dmel_ Dmel\\CG13609', description='CG13609 gene product from transcript CG13609-RA', other_designations='CG13609-PA|uncharacterized protein', map_location='95F3-95F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24227023, end_position_on_the_genomic_accession=24228538, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42885, CurrentID=0, Status='live', Symbol='Spase22-23', Aliases='Dmel_CG5677, CG5677, Dmel\\CG5677, Spase22/23', description='Spase 22/23-subunit', other_designations='CG5677-PA|CG5677-PB|Spase22-23-PA|Spase22-23-PB|spase 22/23-subunit', map_location='95F3-95F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24228342, end_position_on_the_genomic_accession=24230035, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19834759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834759, CurrentID=0, Status='live', Symbol='CG45487', Aliases='Dmel_ Dmel\\CG45487', description='CG45487 gene product from transcript CG45487-RA', other_designations='CG45487-PA|uncharacterized protein', map_location='19A2-19A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19914482, end_position_on_the_genomic_accession=19915281, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42888, CurrentID=0, Status='live', Symbol='beta-PheRS', Aliases='Dmel_CG5706, CG5706, Dmel\\CG5706', description='Phenylalanyl-tRNA synthetase beta subunit', other_designations='CG5706-PA|beta-PheRS-PA', map_location='95F5-95F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24250460, end_position_on_the_genomic_accession=24252679, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42889, CurrentID=0, Status='live', Symbol='jar', Aliases='Dmel_CG5695, 95F, 95F MHC, CG5695, Dm 95F, Dm95F, Dmel\\CG5695, Dro95F, JAG, Jaguar, Jar, M6, MHC95F, MYOVI, Mhc95F, MyoVI, Myosin VI, dMyoVI, jag, ms(3)06746, myo VI, myosins VI', description='jaguar', other_designations='95F myosin|95F unconventional myosin|CG5695-PB|CG5695-PG|CG5695-PH|CG5695-PI|CG5695-PJ|CG5695-PK|CG5695-PL|CG5695-PM|CG5695-PN|jar-PB|jar-PG|jar-PH|jar-PI|jar-PJ|jar-PK|jar-PL|jar-PM|jar-PN|myosin 95F|myosin VI|myosin heavy chain|myosin heavy chain at 95F|myosin-VI|unconventional myosin VI', map_location='95F6-95F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24252826, end_position_on_the_genomic_accession=24271233, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  42890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42890, CurrentID=0, Status='live', Symbol='Orct2', Aliases='Dmel_CG13610, CG13610, Dmel\\CG13610, EP(3)1072, EP1072, cald, orct2', description='Organic cation transporter 2', other_designations='CG13610-PA|CG13610-PB|EP1072|Orct2-PA|Orct2-PB|calderon|organic cation transporter 2', map_location='95F8-95F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24273029, end_position_on_the_genomic_accession=24275728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42891, CurrentID=0, Status='live', Symbol='Orct', Aliases='Dmel_CG6331, CG6331, Dmel\\CG6331, orct', description='Organic cation transporter', other_designations='CG6331-PA|CG6331-PB|Orct-PA|Orct-PB|organic cation transporter', map_location='95F8-95F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24276260, end_position_on_the_genomic_accession=24278792, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42892, CurrentID=0, Status='live', Symbol='CG13611', Aliases='Dmel_ Dmel\\CG13611', description='CG13611 gene product from transcript CG13611-RA', other_designations='CG13611-PA|uncharacterized protein', map_location='95F8-95F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24279572, end_position_on_the_genomic_accession=24281576, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42893, CurrentID=0, Status='live', Symbol='CG6356', Aliases='Dmel_ Dmel\\CG6356', description='CG6356 gene product from transcript CG6356-RA', other_designations='CG6356-PA', map_location='95F9-95F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24283955, end_position_on_the_genomic_accession=24288617, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42894, CurrentID=0, Status='live', Symbol='CG6364', Aliases='Dmel_ CR42873, Dmel\\CG6364, Dmel_CR42873, Uck2', description='CG6364 gene product from transcript CG6364-RD', other_designations='CG6364-PA|CG6364-PB|CG6364-PC|CG6364-PD|CG6364-PE', map_location='95F10-95F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24290172, end_position_on_the_genomic_accession=24293371, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42895, CurrentID=0, Status='live', Symbol='CG5715', Aliases='Dmel_ Dmel\\CG5715', description='CG5715 gene product from transcript CG5715-RA', other_designations='CG5715-PA', map_location='95F10-95F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24293366, end_position_on_the_genomic_accession=24294399, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42896, CurrentID=0, Status='live', Symbol='crb', Aliases='Dmel_CG6383, 0509/20, 1384/04, CG6383, CRB, CT19912, Crb, Crbs, Crumbs, DmCrb, Dmel\\CG6383, crumb, far, l(3)07207, l(3)S050920, l(3)S058104, l(3)j1B5', description='crumbs', other_designations='CG6383-PA|CG6383-PB|CG6383-PC|CG6383-PD|crb-PA|crb-PB|crb-PC|crb-PD|fat rhabdomere|fat rhabdomeres|lethal (3) 07207', map_location='3-82 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24295078, end_position_on_the_genomic_accession=24314541, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  42897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42897, CurrentID=0, Status='live', Symbol='Nab2', Aliases='Dmel_CG5720, BEST:GH27517, BcDNA:LD27873, CG5720, Dmel\\CG5720, DnAB2, RE28534, anon-EST:Liang-1.43, clone 1.43, d', description='CG5720 gene product from transcript CG5720-RA', other_designations='CG5720-PA|Nab2-PA', map_location='95F11-95F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24314660, end_position_on_the_genomic_accession=24318419, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42898, CurrentID=0, Status='live', Symbol='BRWD3', Aliases='Dmel_CG31132, Brwd3, CG18670, CG31132, CG6400, Dmel\\CG31132, anon-WO0118547.520, anon-WO0118547.620, brwd3, d l(3)05842, l(3)s5349, psg13, ram', description='CG31132 gene product from transcript CG31132-RA', other_designations='BRWD3-PA|CG31132-PA|lethal (3) 05842|ramshackle', map_location='3-79.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24319000, end_position_on_the_genomic_accession=24328619, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  42899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42899, CurrentID=0, Status='live', Symbol='CG5728', Aliases='Dmel_ Dmel\\CG5728', description='CG5728 gene product from transcript CG5728-RA', other_designations='CG5728-PA', map_location='95F12-95F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24328455, end_position_on_the_genomic_accession=24333316, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42900, CurrentID=0, Status='live', Symbol='Dis3', Aliases='Dmel_CG6413, CG6413, Dmel\\CG6413, Rrp44, d dRrp44, dis3, taz', description='CG6413 gene product from transcript CG6413-RA', other_designations='CG6413-PA|CG6413-PB|Dis3 ribonuclease|Dis3-PA|Dis3-PB|tazman', map_location='95F13-95F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24333433, end_position_on_the_genomic_accession=24337248, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42901, CurrentID=0, Status='live', Symbol='CG6432', Aliases='Dmel_ Dmel\\CG6432', description='CG6432 gene product from transcript CG6432-RB', other_designations='CG6432-PA|CG6432-PB', map_location='95F14-95F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24336922, end_position_on_the_genomic_accession=24340645, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42903, CurrentID=0, Status='live', Symbol='CG5746', Aliases='Dmel_ Dmel\\CG5746', description='CG5746 gene product from transcript CG5746-RG', other_designations='CG5746-PA|CG5746-PB|CG5746-PC|CG5746-PD|CG5746-PE|CG5746-PF|CG5746-PG', map_location='95F14-95F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24343895, end_position_on_the_genomic_accession=24348791, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  42904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42904, CurrentID=0, Status='live', Symbol='CG6454', Aliases='Dmel_ Dmel\\CG6454', description='CG6454 gene product from transcript CG6454-RC', other_designations='CG6454-PC|CG6454-PD|CG6454-PE', map_location='95F14-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24349182, end_position_on_the_genomic_accession=24356241, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  42905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42905, CurrentID=0, Status='live', Symbol='Golgin84', Aliases='Dmel_CG17785, CG17785, Dmel\\CG17785', description='Golgin-84 ortholog (H. sapiens)', other_designations='CG17785-PA|Golgin-84 ortholog|Golgin84-PA', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24356070, end_position_on_the_genomic_accession=24357938, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42906, CurrentID=0, Status='live', Symbol='CG18528', Aliases='Dmel_ Dmel\\CG18528', description='CG18528 gene product from transcript CG18528-RA', other_designations='CG18528-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24358095, end_position_on_the_genomic_accession=24360144, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42907, CurrentID=0, Status='live', Symbol='CG5762', Aliases='Dmel_ BEST:GH05452, Dmel\\CG5762, anon-WO0140519.76', description='CG5762 gene product from transcript CG5762-RA', other_designations='CG5762-PA', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24360705, end_position_on_the_genomic_accession=24361683, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42909, CurrentID=0, Status='live', Symbol='CG17784', Aliases='Dmel_ Dmel\\CG17784', description='CG17784 gene product from transcript CG17784-RA', other_designations='CG17784-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24364081, end_position_on_the_genomic_accession=24364980, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42910, CurrentID=0, Status='live', Symbol='CG13613', Aliases='Dmel_ Dmel\\CG13613', description='CG13613 gene product from transcript CG13613-RC', other_designations='CG13613-PC|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24365554, end_position_on_the_genomic_accession=24366349, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42911, CurrentID=0, Status='live', Symbol='CG13614', Aliases='Dmel_ Dmel\\CG13614', description='CG13614 gene product from transcript CG13614-RA', other_designations='CG13614-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24369568, end_position_on_the_genomic_accession=24370731, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42912, CurrentID=0, Status='live', Symbol='CG17782', Aliases='Dmel_ Dmel\\CG17782', description='CG17782 gene product from transcript CG17782-RA', other_designations='CG17782-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24370751, end_position_on_the_genomic_accession=24372762, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42913, CurrentID=0, Status='live', Symbol='CG17781', Aliases='Dmel_ Dmel\\CG17781', description='CG17781 gene product from transcript CG17781-RA', other_designations='CG17781-PA|uncharacterized protein', map_location='95F15-96A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24373456, end_position_on_the_genomic_accession=24375661, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42914, CurrentID=0, Status='live', Symbol='CG17780', Aliases='Dmel_ BcDNA:RE41041, Dmel\\CG17780', description='CG17780 gene product from transcript CG17780-RA', other_designations='CG17780-PA|CG17780-PB|uncharacterized protein', map_location='96A1-96A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24382920, end_position_on_the_genomic_accession=24385052, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42915, CurrentID=0, Status='live', Symbol='CG5768', Aliases='Dmel_ Dmel\\CG5768', description='CG5768 gene product from transcript CG5768-RD', other_designations='CG5768-PC|CG5768-PD', map_location='96A1-96A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24386104, end_position_on_the_genomic_accession=24390339, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42916, CurrentID=0, Status='live', Symbol='CG13615', Aliases='Dmel_ Dmel\\CG13615', description='CG13615 gene product from transcript CG13615-RA', other_designations='CG13615-PA|uncharacterized protein', map_location='96A1-96A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24390592, end_position_on_the_genomic_accession=24391870, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42917, CurrentID=0, Status='live', Symbol='CG13616', Aliases='Dmel_ Dmel\\CG13616', description='CG13616 gene product from transcript CG13616-RB', other_designations='CG13616-PA|CG13616-PB|uncharacterized protein', map_location='96A1-96A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24392889, end_position_on_the_genomic_accession=24393762, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42918, CurrentID=0, Status='live', Symbol='nAChRalpha1', Aliases='Dmel_CG5610, ALS, ALs, Acr96A, Acr96Aa, AcrB, CG5610, CG5610-PA, Dalpha1, Dmel\\CG5610, NAcRa96Aa, alpha-like, alpha1, als, dALS, nAChR, nAChRalpha-96Aa, nAcRalpha-96Aa, nAcRs, nAcr96Aa', description='nicotinic Acetylcholine Receptor alpha1', other_designations='ALS|CG5610-PA|CG5610-PB|CG5610-PC|Dalpha1|alpha-like subunit|alpha-like-subunit|nAChRalpha1-PA|nAChRalpha1-PB|nAChRalpha1-PC|nicotinic acetylcholine receptor alpha 96Aa|nicotinic acetylcholine receptor alpha-96Aa|nicotinic acetylcholine receptor alpha-like|nicotinic acetylcholine receptor alpha1', map_location='96A1-96A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24393896, end_position_on_the_genomic_accession=24457157, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42919, CurrentID=0, Status='live', Symbol='nAChRalpha2', Aliases='Dmel_CG6844, Acr96Ab, AcrE, CG6844, CG6844-PA, D[a2], Da2, Dalpha, Dalpha2, Dalpha2/SAD, Dmalpha2, Dmel\\CG6844, SAD, alpha2, alpha[[2]], da2, nAChR, nAChR Dalpha2, nAChRa2, nAChRalpha-96A, nAcRalpha-96Ab, nAcRs, nAchRalpha-96Ab, nAcr96Ab, sad', description='nicotinic Acetylcholine Receptor alpha2', other_designations='CG6844-PA|CG6844-PB|Dalpha2|Dalpha2 nicotimic acetylcholine receptor|SAD|drosophila alpha-like subunit 2|nAChRalpha2-PA|nAChRalpha2-PB|nicotinic acetylcholine receptor alpha 96Ab|nicotinic acetylcholine receptor alpha subunit gene at 96AB|nicotinic acetylcholine receptor alpha-2|nicotinic acetylcholine receptor alpha2|second alpha-like subunit of drosophila', map_location='96A3-96A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24485428, end_position_on_the_genomic_accession=24491147, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42920, CurrentID=0, Status='live', Symbol='nAChRbeta2', Aliases='Dmel_CG6798, Acr96Ac, AcrF, CG6798, Dbeta2, Dmel\\CG6798, SBD, beta2, beta[[2]], nACR beta 96A, nACRbeta, nAChR, nAcRbeta, nAcRbeta-96A, nAcRs, nAcr96Ac, sbd', description='nicotinic Acetylcholine Receptor beta2', other_designations='CG6798-PA|CG6798-PB|Dbeta2|SBD|nAChRbeta2-PA|nAChRbeta2-PB|nicotinic acetylcholine receptor beta 96A|nicotinic acetylcholine receptor beta2|second beta-like subunit of drosophila', map_location='96A5-96A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24509228, end_position_on_the_genomic_accession=24515106, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  42921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42921, CurrentID=0, Status='live', Symbol='CG13617', Aliases='Dmel_ Dmel\\CG13617, Dzip1', description='CG13617 gene product from transcript CG13617-RA', other_designations='CG13617-PA|CG13617-PB|iguana|uncharacterized protein', map_location='96A5-96A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24515278, end_position_on_the_genomic_accession=24518128, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42922, CurrentID=0, Status='live', Symbol='Pp1alpha-96A', Aliases='Dmel_CG6593, 96A, CG6593, DmPp1-96A, Dmel\\CG6593, PP1, PP1 96A, PP1 alpha96a, PP1alpha96A, PP1alpha[[1]], PP1c, PP1c 96A, Pp1-96A, pp1alpha96A', description='Protein phosphatase 1alpha at 96A', other_designations='CG6593-PA|CG6593-PB|Pp1alpha-96A-PA|Pp1alpha-96A-PB|protein phosphatase 1|protein phosphatase 1alpha at 96A', map_location='96A5-96A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24518270, end_position_on_the_genomic_accession=24522331, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42923, CurrentID=0, Status='live', Symbol='CG13622', Aliases='Dmel_ Dmel\\CG13622', description='CG13622 gene product from transcript CG13622-RB', other_designations='CG13622-PB|uncharacterized protein', map_location='96A5-96A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24522478, end_position_on_the_genomic_accession=24523678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42924, CurrentID=0, Status='live', Symbol='CG13618', Aliases='Dmel_ Dmel\\CG13618', description='CG13618 gene product from transcript CG13618-RA', other_designations='CG13618-PA|uncharacterized protein', map_location='96A5-96A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24524794, end_position_on_the_genomic_accession=24527533, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42925, CurrentID=0, Status='live', Symbol='CG6607', Aliases='Dmel_ Dmel\\CG6607', description='CG6607 gene product from transcript CG6607-RA', other_designations='CG6607-PA', map_location='96A6-96A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24528047, end_position_on_the_genomic_accession=24529815, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42926, CurrentID=0, Status='live', Symbol='CG13623', Aliases='Dmel_ BcDNA:GH27470, Dmel\\CG13623', description='CG13623 gene product from transcript CG13623-RA', other_designations='CG13623-PA|CG13623-PB|IscA2 homologue|uncharacterized protein', map_location='96A6-96A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24529773, end_position_on_the_genomic_accession=24530558, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42927, CurrentID=0, Status='live', Symbol='ana1', Aliases='Dmel_CG6631, Ana 1, Ana-1, Ana1, CG6631, Dmel\\CG6631, mecB', description='anastral spindle 1', other_designations='CG6631-PA|CG6631-PC|ana1-PA|ana1-PC|anastral spindle phenotype 1|mechanosensory B|mechanosensory mutant B', map_location='96A6-96A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24530576, end_position_on_the_genomic_accession=24536395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42928, CurrentID=0, Status='live', Symbol='CG5789', Aliases='Dmel_ Dme_CG5789, Dmel\\CG5789', description='CG5789 gene product from transcript CG5789-RA', other_designations='CG5789-PA|CG5789-PB', map_location='96A6-96A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24536204, end_position_on_the_genomic_accession=24543914, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42929, CurrentID=0, Status='live', Symbol='Esyt2', Aliases='Dmel_CG6643, CG6643, Dmel\\CG6643, Tricalbin', description='Extended synaptotagmin-like protein 2 ortholog (H. sapiens)', other_designations='CG6643-PA|CG6643-PB|CG6643-PC|CG6643-PD|Esyt2-PA|Esyt2-PB|Esyt2-PC|Esyt2-PD|extended synaptotagmin-like protein 2 ortholog|tricalbin', map_location='96A7-96A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24544719, end_position_on_the_genomic_accession=24556549, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  42930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42930, CurrentID=0, Status='live', Symbol='REPTOR', Aliases='Dmel_CG13624, CG13624, DM23, DM24, Dmel\\CG13624, anon-WO0153538.65', description='Repressed by TOR', other_designations='CG13624-PA|CG13624-PB|CG13624-PC|CG13624-PD|CG13624-PE|CG13624-PF|CG13624-PG|REPTOR-PA|REPTOR-PB|REPTOR-PC|REPTOR-PD|REPTOR-PE|REPTOR-PF|REPTOR-PG|repressed by TOR', map_location='96A7-96A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24557368, end_position_on_the_genomic_accession=24582864, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42931, CurrentID=0, Status='live', Symbol='CG13625', Aliases='Dmel_ Dmel\\CG13625, anon-WO0118547.376', description='CG13625 gene product from transcript CG13625-RA', other_designations='CG13625-PA|uncharacterized protein', map_location='96A9-96A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24583067, end_position_on_the_genomic_accession=24585160, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42933, CurrentID=0, Status='live', Symbol='Syx18', Aliases='Dmel_CG13626, CG13626, Dm Ufe1, DmSyx18, Dmel\\CG13626, Syx, Ufe1/Syntaxin 18, gtx', description='Syntaxin 18', other_designations='CG13626-PA|Syx18-PA|gtaxin|syntaxin|syntaxin 18', map_location='96A12-96A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24623802, end_position_on_the_genomic_accession=24625562, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42934, CurrentID=0, Status='live', Symbol='atl', Aliases='Dmel_CG6668, ATL, CG6668, Dmel\\CG6668', description='atlastin', other_designations='CG6668-PA|CG6668-PB|CG6668-PC|D-atlastin|atl-PA|atl-PB|atl-PC', map_location='96A13-96A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24625671, end_position_on_the_genomic_accession=24634714, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42935, CurrentID=0, Status='live', Symbol='puf', Aliases='Dmel_CG5794, BcDNA:AT28817, CG5794, Dmel\\CG5794, NEST:bs19d10, Q9VC56, USP34, Usp34', description='puffyeye', other_designations='CG5794-PD|CG5794-PE|CG5794-PF|CG5794-PG|CG5794-PH|CG5794-PI|puf-PD|puf-PE|puf-PF|puf-PG|puf-PH|puf-PI', map_location='96A13-96A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24635270, end_position_on_the_genomic_accession=24650043, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  42936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42936, CurrentID=0, Status='live', Symbol='ash2', Aliases='Dmel_CG6677, 1124/11, 291.8, 703, ASH2, Ash2, CG6677, Dmel\\CG6677, ash-2, l(3)112411, l(3)S112411, l(3)SG65, mad', description='absent, small, or homeotic discs 2', other_designations='CG6677-PB|CG6677-PC|CG6677-PE|absent small or homeotic discs 2|absent, small or homeotic discs 2|ash2-PB|ash2-PC|ash2-PE|many abnormal discs', map_location='3-78.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24650125, end_position_on_the_genomic_accession=24653376, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  42937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42937, CurrentID=0, Status='live', Symbol='CG6695', Aliases='Dmel_ Dmel\\CG6695', description='CG6695 gene product from transcript CG6695-RA', other_designations='CG6695-PA|CG6695-PB', map_location='96A14-96A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24655160, end_position_on_the_genomic_accession=24659379, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42939, CurrentID=0, Status='live', Symbol='Ppox', Aliases='Dmel_CG5796, BcDNA:LD41392, CG5796, Dmel\\CG5796, ESTS:9F2T, PPO', description='Protoporphyrinogen oxidase', other_designations='CG5796-PA|Ppox-PA|protoporphyrinogen oxidase|protoporphyrinogen-oxidase', map_location='96A14-96A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24660197, end_position_on_the_genomic_accession=24661851, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42940, CurrentID=0, Status='live', Symbol='slo', Aliases='Dmel_CG10693, BcDNA:GH10751, CG10693, Dmel\\CG10693, Dslo, Slo, Slo1, dSlo, dSlo1, dslo, fSlo', description='slowpoke', other_designations='CG10693-PA|CG10693-PB|CG10693-PC|CG10693-PD|CG10693-PE|CG10693-PF|CG10693-PH|CG10693-PI|CG10693-PJ|CG10693-PK|CG10693-PL|CG10693-PM|CG10693-PN|CG10693-PO|CG10693-PP|CG10693-PQ|CG10693-PR|CG10693-PS|CG10693-PT|CG10693-PU|CG10693-PV|CG10693-PW|CG10693-PX|slo-PA|slo-PB|slo-PC|slo-PD|slo-PE|slo-PF|slo-PH|slo-PI|slo-PJ|slo-PK|slo-PL|slo-PM|slo-PN|slo-PO|slo-PP|slo-PQ|slo-PR|slo-PS|slo-PT|slo-PU|slo-PV|slo-PW|slo-PX', map_location='3-86 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24662491, end_position_on_the_genomic_accession=24711595, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  42941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42941, CurrentID=0, Status='live', Symbol='CG13627', Aliases='Dmel_ Dmel\\CG13627', description='CG13627 gene product from transcript CG13627-RA', other_designations='CG13627-PA|CG13627-PB|CG13627-PC|uncharacterized protein', map_location='96A17-96A18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24712780, end_position_on_the_genomic_accession=24717430, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42942, CurrentID=0, Status='live', Symbol='Rpb10', Aliases='Dmel_CG13628, BcDNA:SD08670, CG13628, Dmel\\CG13628, RPB10', description='CG13628 gene product from transcript CG13628-RA', other_designations='CG13628-PA|Rpb10-PA', map_location='96A18-96A18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24717511, end_position_on_the_genomic_accession=24717998, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42943, CurrentID=0, Status='live', Symbol='CG13630', Aliases='Dmel_ Dmel\\CG13630, MAP1A', description='CG13630 gene product from transcript CG13630-RA', other_designations='CG13630-PA|uncharacterized protein', map_location='96A18-96A18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24717937, end_position_on_the_genomic_accession=24719364, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42944, CurrentID=0, Status='live', Symbol='tok', Aliases='Dmel_CG6863, CG6863, CT21251, Dmel\\CG6863, TOK, Tlr, Tlr-1, Tok, tldr-1, tldr1, tlr, tlr-1, tlr1', description='tolkin', other_designations='CG6863-PA|CG6863-PB|CG6863-PC|Tolloid-related|piranha|tld-related-1|tok-PA|tok-PB|tok-PC|tolloid-related 1', map_location='96A18-96A19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24719697, end_position_on_the_genomic_accession=24749125, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  42945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42945, CurrentID=0, Status='live', Symbol='tld', Aliases='Dmel_CG6868, CG6868, CT21278, Dmel\\CG6868, TLD, Tld, Tolloid, dTld', description='tolloid', other_designations='CG6868-PA|tld-PA', map_location='3-85 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24749731, end_position_on_the_genomic_accession=24753510, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42946, CurrentID=0, Status='live', Symbol='asp', Aliases='Dmel_CG6875, ASP, Asp, CG6875, Dm Asp, Dmel\\CG6875, anon-96Aa, anon-WO0118547.279', description='abnormal spindle', other_designations='CG6875-PA|abnormal spindle protein|abnormal spindles|asp-PA', map_location='3-85.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24753849, end_position_on_the_genomic_accession=24760513, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42947, CurrentID=0, Status='live', Symbol='AstA', Aliases='Dmel_CG13633, ALLS, AS, ASA, AST, AST-1, AST-3, AST-4, AST-A, Ast, Ast-A-1, AstA-2, AstA-3, AstA-4, BcDNA:RE16553, CG13633, DAP, DAP-A, DST-1A, DST-2A, DST-3A, DST-4A, Dmel\\CG13633, Drm-AST-1, Drm-AST-2, Drm-AST-3, Drm-AST-4, Drm-AST-A, ast, AstA', description='Allatostatin A', other_designations='AstA-PA|AstA-PB|CG13633-PA|CG13633-PB|Drostatin-A1|Drostatin-A3|Drostatin-A4|FGLa-type allatostatin|allatostatin|allatostatin A|allatostatin A-1|allatostatin Diploptera-type|allatostatin-A-like|allatostatin-A4|drostatin-A2', map_location='96A20-96A20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24760530, end_position_on_the_genomic_accession=24765071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42948, CurrentID=0, Status='live', Symbol='CG42331', Aliases='Dmel_ CG13632, CG6879, Dmel\\CG42331, Dmel_CG13632, Dmel_CG6879', description='CG42331 gene product from transcript CG42331-RB', other_designations='CG42331-PB|CG42331-PC|CG42331-PD|uncharacterized protein', map_location='96A21-96A22', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24772181, end_position_on_the_genomic_accession=24784562, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42950, CurrentID=0, Status='live', Symbol='CG5805', Aliases='Dmel_ Dmel\\CG5805', description='CG5805 gene product from transcript CG5805-RB', other_designations='CG5805-PA|CG5805-PB', map_location='96A22-96A22', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24784552, end_position_on_the_genomic_accession=24787421, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42951, CurrentID=0, Status='live', Symbol='CG13634', Aliases='Dmel_ Dmel\\CG13634', description='CG13634 gene product from transcript CG13634-RA', other_designations='CG13634-PA|uncharacterized protein', map_location='96A22-96A22', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24788248, end_position_on_the_genomic_accession=24790136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42952, CurrentID=0, Status='live', Symbol='Ets96B', Aliases='Dmel_CG6892, CG6892, Dmel\\CG6892, Ets, ets96B', description='CG6892 gene product from transcript CG6892-RC', other_designations='CG6892-PB|CG6892-PC|Ets96B-PB|Ets96B-PC', map_location='96A22-96A22', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24788054, end_position_on_the_genomic_accession=24797955, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42953, CurrentID=0, Status='live', Symbol='Ude', Aliases='Dmel_CG18410, CG18410, Dmel\\CG18410, UDE', description='Uracil-DNA degrading factor', other_designations='CG18410-PB|CG18410-PC|CG18410-PD|Ude-PB|Ude-PC|Ude-PD|uracil-DNA degrading factor|uracil-DNA-degrading factor', map_location='96A22-96A22', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24797460, end_position_on_the_genomic_accession=24799668, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42954, CurrentID=0, Status='live', Symbol='polybromo', Aliases='Dmel_CG11375, BAF180, Bap180, CG11375, Dmel\\CG11375, PB, POLYBROMO, Pb, Polybromo, anon-WO0140519.158, baf180, bap180', description='CG11375 gene product from transcript CG11375-RA', other_designations='CG11375-PA|polybromo-PA', map_location='96A22-96A23', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24799967, end_position_on_the_genomic_accession=24805259, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42955, CurrentID=0, Status='live', Symbol='CG6980', Aliases='Dmel_ Dmel\\CG6980, anon-WO0140519.171', description='CG6980 gene product from transcript CG6980-RA', other_designations='CG6980-PA', map_location='96A23-96A23', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24807621, end_position_on_the_genomic_accession=24808677, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42956, CurrentID=0, Status='live', Symbol='CG5807', Aliases='Dmel_ BcDNA:GH12663, Dmel\\CG5807, anon-WO0118547.355', description='CG5807 gene product from transcript CG5807-RB', other_designations='CG5807-PA|CG5807-PB', map_location='96A23-96A23', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24805445, end_position_on_the_genomic_accession=24814012, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42957, CurrentID=0, Status='live', Symbol='CG5808', Aliases='Dmel_ BcDNA:GH01073, Dmel\\CG5808, UKL, cg5808', description='CG5808 gene product from transcript CG5808-RA', other_designations='CG5808-PA', map_location='96A23-96A23', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24814903, end_position_on_the_genomic_accession=24817007, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42958, CurrentID=0, Status='live', Symbol='Saf-B', Aliases='Dmel_CG6995, CG6995, Dmel\\CG6995, LD22822, SAF-B, cg6995, saf-b', description='Scaffold attachment factor B', other_designations='CG6995-PB|CG6995-PD|CG6995-PE|CG6995-PF|Saf-B-PB|Saf-B-PD|Saf-B-PE|Saf-B-PF|scaffold attachment factor B', map_location='96A23-96A23', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24817464, end_position_on_the_genomic_accession=24822350, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42959, CurrentID=0, Status='live', Symbol='niki', Aliases='Dmel_CG10951, CG10951, Dmel\\CG10951, NEK1, Niki', description='nimA-like kinase', other_designations='CG10951-PA|niki-PA', map_location='96A23-96A24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24822401, end_position_on_the_genomic_accession=24825648, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42960, CurrentID=0, Status='live', Symbol='RabX4', Aliases='Dmel_CG31118, AAF56345, CG13638, CG31118, Dm Dmel\\CG31118, rabX4', description='CG31118 gene product from transcript CG31118-RA', other_designations='CG31118-PA|RabX4-PA', map_location='96A24-96A24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24826665, end_position_on_the_genomic_accession=24828409, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  42961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42961, CurrentID=0, Status='live', Symbol='sosie', Aliases='Dmel_CG13636, BcDNA:RE05274, BcDNA:RE20019, CG13636, Dmel\\CG13636, sie', description='CG13636 gene product from transcript CG13636-RB', other_designations='CG13636-PA|CG13636-PB|CG13636-PC|sosie-PA|sosie-PB|sosie-PC', map_location='96A24-96A25', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24833952, end_position_on_the_genomic_accession=24840488, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42962, CurrentID=0, Status='live', Symbol='Esp', Aliases='Dmel_CG7005, CG7005, Dmel\\CG7005', description='Epidermal stripes and patches', other_designations='CG7005-PA|CG7005-PB|Esp-PA|Esp-PB|epidermal stripes and patches', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24841935, end_position_on_the_genomic_accession=24850412, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42963, CurrentID=0, Status='live', Symbol='CG7006', Aliases='Dmel_ BcDNA:GM12126, Dmel\\CG7006, anon-WO0118547.410, dNIP7', description='CG7006 gene product from transcript CG7006-RA', other_designations='CG7006-PA', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24850627, end_position_on_the_genomic_accession=24851385, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42964, CurrentID=0, Status='live', Symbol='nct', Aliases='Dmel_CG7012, CG7012, Dmel\\CG7012, Nct, Nic, S5A, ST96B, agro, dNct, nic', description='nicastrin', other_designations='CG7012-PA|CG7012-PC|CG7012-PD|CG7012-PE|agoraphobic|dNicastrin|nct-PA|nct-PC|nct-PD|nct-PE', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24851601, end_position_on_the_genomic_accession=24854869, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42965, CurrentID=0, Status='live', Symbol='CG10899', Aliases='Dmel_ BcDNA:AT02719, Dmel\\CG10899', description='CG10899 gene product from transcript CG10899-RA', other_designations='CG10899-PA|CG10899-PC|uncharacterized protein', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24856383, end_position_on_the_genomic_accession=24857559, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  42967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42967, CurrentID=0, Status='live', Symbol='CG13640', Aliases='Dmel_ BcDNA:RE70528, Dmel\\CG13640', description='CG13640 gene product from transcript CG13640-RA', other_designations='CG13640-PA|uncharacterized protein', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24861659, end_position_on_the_genomic_accession=24862165, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42968, CurrentID=0, Status='live', Symbol='CG7016', Aliases='Dmel_ Dmel\\CG7016', description='CG7016 gene product from transcript CG7016-RA', other_designations='CG7016-PA', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24862931, end_position_on_the_genomic_accession=24864216, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42969, CurrentID=0, Status='live', Symbol='CG13641', Aliases='Dmel_ Dmel\\CG13641', description='CG13641 gene product from transcript CG13641-RA', other_designations='CG13641-PA|uncharacterized protein', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24864191, end_position_on_the_genomic_accession=24864732, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42970, CurrentID=0, Status='live', Symbol='Elal', Aliases='Dmel_CG7021, CG7021, Dmel\\CG7021, Ela', description='Elastin-like', other_designations='CG7021-PA|CG7021-PB|CG7021-PC|Elal-PA|Elal-PB|Elal-PC|elastin-like', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24865350, end_position_on_the_genomic_accession=24867651, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42971, CurrentID=0, Status='live', Symbol='CycB3', Aliases='Dmel_CG5814, CG5814, Cyc B3, Dmel\\CG5814, anon-WO0118547.414, cycB3, l(3)L6540', description='Cyclin B3', other_designations='CG5814-PA|CG5814-PB|CycB3-PA|CycB3-PB|CyclinB3|cyclin B3', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24867934, end_position_on_the_genomic_accession=24870816, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42972, CurrentID=0, Status='live', Symbol='CG3744', Aliases='Dmel_ Dmel\\CG3744', description='CG3744 gene product from transcript CG3744-RG', other_designations='CG3744-PD|CG3744-PE|CG3744-PF|CG3744-PG', map_location='96B2-96B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24871202, end_position_on_the_genomic_accession=24876654, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42973, CurrentID=0, Status='live', Symbol='CG11089', Aliases='Dmel_ 153761_at, BEST:GH11240, CT30991, Dmel\\CG11089, atic', description='CG11089 gene product from transcript CG11089-RA', other_designations='CG11089-PA|CG11089-PB|uncharacterized protein', map_location='96B2-96B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24877045, end_position_on_the_genomic_accession=24881478, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19834846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834846, CurrentID=0, Status='live', Symbol='CG45307', Aliases='Dmel_ Dmel\\CG45307, lincRNA.77', description='CG45307 gene product from transcript CG45307-RA', other_designations='CG45307-PA|uncharacterized protein', map_location='29E6-29E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8751715, end_position_on_the_genomic_accession=8752122, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  42975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42975, CurrentID=0, Status='live', Symbol='CG31121', Aliases='Dmel_ BcDNA:GH19411, CG11764, CG18633, CG18655, Dmel\\CG31121', description='CG31121 gene product from transcript CG31121-RA', other_designations='CG31121-PA|CG31121-PB|uncharacterized protein', map_location='96B2-96B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24881466, end_position_on_the_genomic_accession=24894605, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  42976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42976, CurrentID=0, Status='live', Symbol='CG11069', Aliases='Dmel_ Dmel\\CG11069', description='CG11069 gene product from transcript CG11069-RA', other_designations='CG11069-PA|uncharacterized protein', map_location='96B4-96B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24894742, end_position_on_the_genomic_accession=24897613, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  42978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42978, CurrentID=0, Status='live', Symbol='CG10845', Aliases='Dmel_ Dmel\\CG10845, KIF26A/B, Klp96B', description='CG10845 gene product from transcript CG10845-RA', other_designations='CG10845-PA|uncharacterized protein', map_location='96B4-96B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24901116, end_position_on_the_genomic_accession=24903048, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42979, CurrentID=0, Status='live', Symbol='PIG-S', Aliases='Dmel_CG44255, CG13644, CG31120, CG44255, CG6031, Dmel\\CG44255, sud2', description='Phosphatidylinositol glycan anchor biosynthesis, class S ortholog (H. sapiens)', other_designations='CG44255-PC|PIG-S-PC|phosphatidylinositol glycan anchor biosynthesis, class S ortholog', map_location='96B6-96B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24913216, end_position_on_the_genomic_accession=24917864, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  42981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42981, CurrentID=0, Status='live', Symbol='CG11168', Aliases='Dmel_ Dmel\\CG11168', description='CG11168 gene product from transcript CG11168-RC', other_designations='CG11168-PA|CG11168-PC|CG11168-PD|CG11168-PE|uncharacterized protein', map_location='96B6-96B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24918106, end_position_on_the_genomic_accession=24922385, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42982, CurrentID=0, Status='live', Symbol='Mink', Aliases='Dmel_CG11120, CG11120, Dmel\\CG11120', description='Mitotic spindle and nuclear protein', other_designations='CG11120-PA|Mink-PA|mitotic spindle and nuclear protein', map_location='96B8-96B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24922473, end_position_on_the_genomic_accession=24925160, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  42983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42983, CurrentID=0, Status='live', Symbol='Trf4-2', Aliases='Dmel_CG17462, AAF56369, CG17462, DmTRF4-2, Dmel\\CG17462', description='CG17462 gene product from transcript CG17462-RA', other_designations='CG17462-PA|CG17462-PB|Trf4-2-PA|Trf4-2-PB', map_location='96B9-96B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24925216, end_position_on_the_genomic_accession=24927085, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42984, CurrentID=0, Status='live', Symbol='CG11771', Aliases='Dmel_ Dmel\\CG11771, anon-EST:Posey100', description='CG11771 gene product from transcript CG11771-RA', other_designations='CG11771-PA|uncharacterized protein', map_location='96B10-96B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24932010, end_position_on_the_genomic_accession=24934919, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  42985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42985, CurrentID=0, Status='live', Symbol='Osbp', Aliases='Dmel_CG6708, CG6708, Dmel\\CG6708, OSBP-Dm, osbp', description='Oxysterol binding protein', other_designations='CG6708-PA|Osbp-PA|oxysterol binding protein|oxysterol-binding protein', map_location='96B10-96B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24934869, end_position_on_the_genomic_accession=24938222, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42986, CurrentID=0, Status='live', Symbol='ssh', Aliases='Dmel_CG6238, CG6238, Dmel\\CG6238, MKP, Mkph, SSH, Ssh, l(3)01207', description='slingshot', other_designations='CG6238-PA|CG6238-PB|CG6238-PC|CG6238-PD|MAP-kinase-phosphatase|MKP-like|ssh-PA|ssh-PB|ssh-PC|ssh-PD', map_location='96B10-96B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24938710, end_position_on_the_genomic_accession=24944956, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  42987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42987, CurrentID=0, Status='live', Symbol='Nmnat', Aliases='Dmel_CG13645, 3R4, AAL90149, CG13645, DmNMNAT, Dmel\\CG13645, NMNAT1, NmnatB, dNmnat, nmnat, Nmnat', description='Nicotinamide mononucleotide adenylyltransferase', other_designations='CG13645-PA|CG13645-PB|CG13645-PC|Nmnat-PA|Nmnat-PB|Nmnat-PC|nicotinamide mononucleotide adenyltransferase|nicotinamide mononucleotide adenylyl transferase 1|nicotinamide mononucleotide adenylyltransferase|nicotinamide mononucleotide adenylyltransferase 1|nicotinamide mononucleotide adenylytransferase', map_location='96B11-96B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24945237, end_position_on_the_genomic_accession=24947400, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  42988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42988, CurrentID=0, Status='live', Symbol='CG13646', Aliases='Dmel_ Dmel\\CG13646', description='CG13646 gene product from transcript CG13646-RA', other_designations='CG13646-PA|uncharacterized protein', map_location='96B11-96B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24947632, end_position_on_the_genomic_accession=24949539, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42990, CurrentID=0, Status='live', Symbol='tnc', Aliases='Dmel_CG13648, CG13648, Dmel\\CG13648, Mur96B, Tnc', description='tenectin', other_designations='CG13648-PC|CG13648-PD|mucin related 96B|tnc-PC|tnc-PD', map_location='96B15-96B15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24984384, end_position_on_the_genomic_accession=25022867, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  42991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42991, CurrentID=0, Status='live', Symbol='beta4GalT7', Aliases='Dmel_CG11780, CG11780, Dbeta4GalT7, Dmel\\CG11780, GalT, GalTI, beta4Gal-T7, beta4GalTI, betaGalT7, dGAGbeta4GalTI, dbeta4GalT7, dbeta4GalTI', description='beta-4-galactosyltransferase 7', other_designations='CG11780-PA|beta-1,-4-galactosyltransferase|beta1,4-Galactosyltransferase-7|beta4GalT7-PA|galactosyltransferase I|glycosaminoglycan galactosyltransferase I', map_location='96B16-96B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25023256, end_position_on_the_genomic_accession=25024482, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42992, CurrentID=0, Status='live', Symbol='CG11781', Aliases='Dmel_ BcDNA:RE35789, Dmel\\CG11781', description='CG11781 gene product from transcript CG11781-RA', other_designations='CG11781-PA|uncharacterized protein', map_location='96B16-96B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25024473, end_position_on_the_genomic_accession=25025017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  42993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42993, CurrentID=0, Status='live', Symbol='Hr96', Aliases='Dmel_CG11783, CG11783, DHR96, Dhr96, Dmel\\CG11783, HR96, NR1J1, dHR96', description='Hormone receptor-like in 96', other_designations='CG11783-PA|Hr96-PA|hormone receptor-like in 96', map_location='96B16-96B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25025235, end_position_on_the_genomic_accession=25029351, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  42994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42994, CurrentID=0, Status='live', Symbol='Smg6', Aliases='Dmel_CG6369, CG6369, Dmel\\CG6369, SMG6, dSMG6', description='CG6369 gene product from transcript CG6369-RA', other_designations='CG6369-PA|Smg6-PA', map_location='96B17-96B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25029429, end_position_on_the_genomic_accession=25033465, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42995, CurrentID=0, Status='live', Symbol='CG6422', Aliases='Dmel_ Dmel\\CG6422', description='CG6422 gene product from transcript CG6422-RA', other_designations='CG6422-PA|CG6422-PB|CG6422-PC|CG6422-PD', map_location='96B17-96B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25033876, end_position_on_the_genomic_accession=25039249, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  42996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42996, CurrentID=0, Status='live', Symbol='bai', Aliases='Dmel_CG11785, 150547_at, BAI, BcDNA:RE52411, CG11785, Dmel\\CG11785, p23, p24', description='baiser', other_designations='CG11785-PA|bai-PA', map_location='96B17-96B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25039374, end_position_on_the_genomic_accession=25040443, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  42998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42998, CurrentID=0, Status='live', Symbol='CG11786', Aliases='Dmel_ Dmel\\CG11786', description='CG11786 gene product from transcript CG11786-RB', other_designations='CG11786-PB|uncharacterized protein', map_location='96B18-96B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25046323, end_position_on_the_genomic_accession=25047409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42999, CurrentID=0, Status='live', Symbol='CG11790', Aliases='Dmel_ Dmel\\CG11790', description='CG11790 gene product from transcript CG11790-RB', other_designations='CG11790-PA|CG11790-PB|CG11790-PC|uncharacterized protein', map_location='96B18-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25047706, end_position_on_the_genomic_accession=25049455, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43000, CurrentID=0, Status='live', Symbol='CG11791', Aliases='Dmel_ Dmel\\CG11791', description='CG11791 gene product from transcript CG11791-RD', other_designations='CG11791-PA|CG11791-PB|CG11791-PC|CG11791-PD|uncharacterized protein', map_location='96B19-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25049616, end_position_on_the_genomic_accession=25053074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43001, CurrentID=0, Status='live', Symbol='CG31109', Aliases='Dmel_ BcDNA:RE17222, CG13649, Dmel\\CG31109', description='CG31109 gene product from transcript CG31109-RB', other_designations='CG31109-PA|CG31109-PB|uncharacterized protein', map_location='96B19-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25053369, end_position_on_the_genomic_accession=25054653, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43002, CurrentID=0, Status='live', Symbol='Unc-13-4B', Aliases='Dmel_CG34349, BP1013, CG11819, CG34037, CG34349, Dmel\\CG34349, Dmel_CG11819, Dmel_CG34037, dunc-13-4B, unc-13-4B', description='Unc-13 ortholog 4B (C. elegans)', other_designations='CG34349-PC|CG34349-PD|CG34349-PE|Unc-13 ortholog 4B|Unc-13-4B-PC|Unc-13-4B-PD|Unc-13-4B-PE|unc-13-4B', map_location='96B19-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25056286, end_position_on_the_genomic_accession=25064348, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  43003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43003, CurrentID=0, Status='live', Symbol='veli', Aliases='Dmel_CG7662, CG7662, D- DLin-7, DLin7, Dmel\\CG7662, Dveli, Flin-7, Lin-7, Veli, d-Veli, dLin-7, dLin7, dVeli, lin-7, lin7', description='CG7662 gene product from transcript CG7662-RA', other_designations='CG7662-PA|CG7662-PC|veli-PA|veli-PC', map_location='96B19-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25064418, end_position_on_the_genomic_accession=25065591, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43004, CurrentID=0, Status='live', Symbol='PQBP1', Aliases='Dmel_CG11820, CG11820, Dmel\\CG11820, PQBP-1, dPQBP-1, dPQBP1', description='Poly-glutamine tract binding protein 1', other_designations='CG11820-PA|PQBP1-PA', map_location='96B19-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25065791, end_position_on_the_genomic_accession=25066718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43005, CurrentID=0, Status='live', Symbol='OstStt3', Aliases='Dmel_CG7748, BcDNA.GM01838, BcDNA:GM01838, CG7748, Dmel\\CG7748, STT, anon-WO03054008.7, l(3)j2D9', description='Oligosaccharyl transferase 3', other_designations='CG7748-PA|OstStt3-PA|oligosaccharyl transferase 3', map_location='96B19-96B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25066712, end_position_on_the_genomic_accession=25069637, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43006, CurrentID=0, Status='live', Symbol='CG11839', Aliases='Dmel_ CT36969, Dmel\\CG11839, Znf830', description='CG11839 gene product from transcript CG11839-RB', other_designations='CG11839-PB|uncharacterized protein', map_location='96B20-96B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25069927, end_position_on_the_genomic_accession=25070914, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43007, CurrentID=0, Status='live', Symbol='CG11836', Aliases='Dmel_ Dmel\\CG11836, SP23', description='CG11836 gene product from transcript CG11836-RB', other_designations='CG11836-PA|CG11836-PB|CG11836-PC|CG11836-PE|CG11836-PF|CG11836-PG|CG11836-PI|uncharacterized protein', map_location='96B20-96B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25071126, end_position_on_the_genomic_accession=25072712, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43008, CurrentID=0, Status='live', Symbol='CG9996', Aliases='Dmel_ Dmel\\CG9996, dibeta3GnT5', description='CG9996 gene product from transcript CG9996-RB', other_designations='CG9996-PB', map_location='96B20-96B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25072729, end_position_on_the_genomic_accession=25074607, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43009, CurrentID=0, Status='live', Symbol='CG11920', Aliases='Dmel_ Dmel\\CG11920', description='CG11920 gene product from transcript CG11920-RA', other_designations='CG11920-PA|uncharacterized protein', map_location='96B20-96B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25074877, end_position_on_the_genomic_accession=25076382, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43010, CurrentID=0, Status='live', Symbol='fd96Ca', Aliases='Dmel_CG11921, CG11921, Dmel\\CG11921, Dmfd4, FD4, fd4', description='forkhead domain 96Ca', other_designations='CG11921-PA|CG11921-PB|fd96Ca-PA|fd96Ca-PB', map_location='96B20-96B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25082684, end_position_on_the_genomic_accession=25084878, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43011, CurrentID=0, Status='live', Symbol='fd96Cb', Aliases='Dmel_CG11922, CG11922, Dmel\\CG11922, Dmfd5, FD5, fd5', description='forkhead domain 96Cb', other_designations='CG11922-PA|fd96Cb-PA', map_location='96B20-96B21', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25095012, end_position_on_the_genomic_accession=25096099, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43012, CurrentID=0, Status='live', Symbol='CG33095', Aliases='Dmel_ CG11935, CG11938, Dmel\\CG33095, FBgn0039275', description='CG33095 gene product from transcript CG33095-RB', other_designations='CG33095-PA|CG33095-PB|uncharacterized protein', map_location='96B21-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25097672, end_position_on_the_genomic_accession=25102146, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43014, CurrentID=0, Status='live', Symbol='CG13650', Aliases='Dmel_ Dmel\\CG13650', description='CG13650 gene product from transcript CG13650-RB', other_designations='CG13650-PB|uncharacterized protein', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25109226, end_position_on_the_genomic_accession=25112516, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43015, CurrentID=0, Status='live', Symbol='CG31108', Aliases='Dmel_ CG10057, Dmel\\CG31108', description='CG31108 gene product from transcript CG31108-RA', other_designations='CG31108-PA|CG31108-PB|uncharacterized protein', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25112867, end_position_on_the_genomic_accession=25122920, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43016, CurrentID=0, Status='live', Symbol='vig2', Aliases='Dmel_CG11844, BcDNA:SD06613, CG11844, Dmel\\CG11844, anon-WO0118547.290, vig-2', description='CG11844 gene product from transcript CG11844-RB', other_designations='CG11844-PA|CG11844-PB|CG11844-PC|CG11844-PD|vig2-PA|vig2-PB|vig2-PC|vig2-PD', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25123373, end_position_on_the_genomic_accession=25125619, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43017, CurrentID=0, Status='live', Symbol='Mocs2', Aliases='Dmel_CG10238, CG10238, Dmel\\CG10238, Dromel_CG10238_FBtr0084782_mocs_mORF, MBIP/dMoaE, MOCS2, Mbip, dMoaE', description='Molybdenum cofactor synthesis 2', other_designations='CG10238-PA|MAPK upstream kinase-binding inhibitory protein|MUK-binding inhibitory protein|Mocs2-PA|molybdenum cofactor synthesis 2|molybdopterin synthase 2', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25125632, end_position_on_the_genomic_accession=25127359, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43018, CurrentID=0, Status='live', Symbol='Clbn', Aliases='Dmel_CG11847, CG11847, Dmel\\CG11847', description='Caliban', other_designations='CG11847-PB|CG11847-PC|Clbn-PB|Clbn-PC|caliban', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25127502, end_position_on_the_genomic_accession=25131155, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43019, CurrentID=0, Status='live', Symbol='Bili', Aliases='Dmel_CG11848, CG11848, Dmel\\CG11848, dBili', description='Band4.1 inhibitor LRP interactor', other_designations='Bili-PB|Bili-PC|Bili-PD|Bili-PE|Bili-PF|Bili-PG|Bili-PH|CG11848-PB|CG11848-PC|CG11848-PD|CG11848-PE|CG11848-PF|CG11848-PG|CG11848-PH|band 4.1 FERM-like', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25131306, end_position_on_the_genomic_accession=25136709, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  43020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43020, CurrentID=0, Status='live', Symbol='danr', Aliases='Dmel_CG13651, CG13651, Danr, Dmel\\CG13651, hern', description='distal antenna-related', other_designations='CG13651-PA|danr-PA|distal antenna related|hernandez|tintin', map_location='96C2-96C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25138072, end_position_on_the_genomic_accession=25139903, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43023, CurrentID=0, Status='live', Symbol='dan', Aliases='Dmel_CG11849, CG11849, Dan, Dmel\\CG11849, fer', description='distal antenna', other_designations='CG11849-PA|CG11849-PB|CG11849-PC|dan-PA|dan-PB|dan-PC|fernandez|tintin', map_location='96C3-96C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25184383, end_position_on_the_genomic_accession=25192806, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43025, CurrentID=0, Status='live', Symbol='CG13653', Aliases='Dmel_ Dmel\\CG13653', description='CG13653 gene product from transcript CG13653-RA', other_designations='CG13653-PA|uncharacterized protein', map_location='96C4-96C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25195527, end_position_on_the_genomic_accession=25196531, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43026, CurrentID=0, Status='live', Symbol='Cad96Ca', Aliases='Dmel_CG10244, CG10239, CG10244, CT28779, Cad96C, DCad96Ca, Dm Cad96Ca, DmHD-14, Dmel\\CG10244, FGFR1-like, HD-14, cad96C, cad96Ca, stit', description='Cadherin 96Ca', other_designations='CG10244-PA|Cad96Ca-PA|cadherin 96Ca|stitcher', map_location='96C4-96C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25197571, end_position_on_the_genomic_accession=25208283, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43028, CurrentID=0, Status='live', Symbol='CG13654', Aliases='Dmel_ Dmel\\CG13654', description='CG13654 gene product from transcript CG13654-RA', other_designations='CG13654-PA|CG13654-PB|CG13654-PC|uncharacterized protein', map_location='96C4-96C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25199818, end_position_on_the_genomic_accession=25214255, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43029, CurrentID=0, Status='live', Symbol='CG13663', Aliases='Dmel_ Dmel\\CG13663', description='CG13663 gene product from transcript CG13663-RA', other_designations='CG13663-PA|uncharacterized protein', map_location='96C5-96C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25214325, end_position_on_the_genomic_accession=25215188, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43030, CurrentID=0, Status='live', Symbol='CG31103', Aliases='Dmel_ CG13655, Dmel\\CG31103', description='CG31103 gene product from transcript CG31103-RB', other_designations='CG31103-PB|uncharacterized protein', map_location='96C6-96C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25219318, end_position_on_the_genomic_accession=25221444, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43031, CurrentID=0, Status='live', Symbol='CG11851', Aliases='Dmel_ Dmel\\CG11851, dAlg9', description='CG11851 gene product from transcript CG11851-RA', other_designations='CG11851-PA|uncharacterized protein', map_location='96C6-96C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25221589, end_position_on_the_genomic_accession=25224118, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43033, CurrentID=0, Status='live', Symbol='Cad96Cb', Aliases='Dmel_CG13664, CG10421, CG13664, CT29262, Cad96C, DCad96Cb, Dmel\\CG13664, cad96C, cad96Cb, cg13664', description='Cadherin 96Cb', other_designations='CG13664-PB|CG13664-PC|CG13664-PD|Cad96Cb-PB|Cad96Cb-PC|Cad96Cb-PD|cadherin 96Cb', map_location='96C7-96C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25224156, end_position_on_the_genomic_accession=25232321, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43034, CurrentID=0, Status='live', Symbol='CG10420', Aliases='Dmel_ Dmel\\CG10420', description='CG10420 gene product from transcript CG10420-RA', other_designations='CG10420-PA|uncharacterized protein', map_location='96C7-96C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25233636, end_position_on_the_genomic_accession=25235529, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43035, CurrentID=0, Status='live', Symbol='CG11852', Aliases='Dmel_ CT33073, Dmel\\CG11852', description='CG11852 gene product from transcript CG11852-RA', other_designations='CG11852-PA|uncharacterized protein', map_location='96C7-96C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25237614, end_position_on_the_genomic_accession=25238895, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43036, CurrentID=0, Status='live', Symbol='to', Aliases='Dmel_CG11853, CG11853, CT33074, Dmel\\CG11853, To', description='takeout', other_designations='CG11853-PA|CG11853-PB|take-out|to-PA|to-PB', map_location='96C7-96C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25240180, end_position_on_the_genomic_accession=25241668, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43037, CurrentID=0, Status='live', Symbol='CG11854', Aliases='Dmel_ Dmel\\CG11854', description='CG11854 gene product from transcript CG11854-RB', other_designations='CG11854-PB|uncharacterized protein', map_location='96C7-96C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25242020, end_position_on_the_genomic_accession=25243262, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43038, CurrentID=0, Status='live', Symbol='bam', Aliases='Dmel_CG10422, BAM, Bam, Bam-C, BamC, BamF, CG10422, Dmel\\CG10422, alpha, fs(3)neo61, ham', description='bag of marbles', other_designations='CG10422-PA|bag of marble|bag-of-marbles|bam-PA|transcript alpha', map_location='96C7-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25243039, end_position_on_the_genomic_accession=25245057, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43039, CurrentID=0, Status='live', Symbol='RpS27', Aliases='Dmel_CG10423, CG10423, Dmel\\CG10423, Q9VBU9_DROME, Rp S27, anon-EST:Posey295', description='Ribosomal protein S27', other_designations='CG10423-PA|CG10423-PB|RpS27-PA|RpS27-PB|ribosomal protein S27', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25245955, end_position_on_the_genomic_accession=25247514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43040, CurrentID=0, Status='live', Symbol='Nup37', Aliases='Dmel_CG11875, CG11875, Dmel\\CG11875', description='Nucleoporin 37kD', other_designations='CG11875-PA|CG11875-PB|Nup37-PA|Nup37-PB|nucleoporin 37kD', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25247658, end_position_on_the_genomic_accession=25249100, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43041, CurrentID=0, Status='live', Symbol='Nup358', Aliases='Dmel_CG11856, CG11856, Dmel\\CG11856, NUP358, Q9VBU7_DROME, RanBP2, nup358', description='Nucleoporin 358kD', other_designations='CG11856-PA|CG11856-PB|Nup358-PA|Nup358-PB|nucleoporin 358|nucleoporin 358kD', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25249348, end_position_on_the_genomic_accession=25258672, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43042, CurrentID=0, Status='live', Symbol='CG11857', Aliases='Dmel_ Dmel\\CG11857', description='CG11857 gene product from transcript CG11857-RA', other_designations='CG11857-PA|CG11857-PC|uncharacterized protein', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25258898, end_position_on_the_genomic_accession=25260759, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43043, CurrentID=0, Status='live', Symbol='CG10425', Aliases='Dmel_ Dmel\\CG10425', description='CG10425 gene product from transcript CG10425-RA', other_designations='CG10425-PA|uncharacterized protein', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25260963, end_position_on_the_genomic_accession=25262498, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43044, CurrentID=0, Status='live', Symbol='CG11858', Aliases='Dmel_ BcDNA:RH66629, Dmel\\CG11858', description='CG11858 gene product from transcript CG11858-RA', other_designations='CG11858-PA|uncharacterized protein', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25262741, end_position_on_the_genomic_accession=25263292, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43046, CurrentID=0, Status='live', Symbol='RIOK2', Aliases='Dmel_CG11859, CG11859, Dmel\\CG11859, Rio2, dRIOK2', description='RIO kinase 2', other_designations='CG11859-PA|RIOK2-PA', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25265844, end_position_on_the_genomic_accession=25267757, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43050, CurrentID=0, Status='live', Symbol='CG11878', Aliases='Dmel_ CT33083, Dmel\\CG11878', description='CG11878 gene product from transcript CG11878-RA', other_designations='CG11878-PA|uncharacterized protein', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25273837, end_position_on_the_genomic_accession=25275237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43051, CurrentID=0, Status='live', Symbol='CG10513', Aliases='Dmel_ CT29500, Dmel\\CG10513', description='CG10513 gene product from transcript CG10513-RB', other_designations='CG10513-PB|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25275790, end_position_on_the_genomic_accession=25277212, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43052, CurrentID=0, Status='live', Symbol='CG10514', Aliases='Dmel_ CT29508, Dmel\\CG10514', description='CG10514 gene product from transcript CG10514-RA', other_designations='CG10514-PA|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25278427, end_position_on_the_genomic_accession=25279846, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43053, CurrentID=0, Status='live', Symbol='CG11892', Aliases='Dmel_ Dmel\\CG11892', description='CG11892 gene product from transcript CG11892-RA', other_designations='CG11892-PA|CG11892-PB|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25281548, end_position_on_the_genomic_accession=25283237, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43054, CurrentID=0, Status='live', Symbol='CG31104', Aliases='Dmel_ CG13657, Dmel\\CG31104', description='CG31104 gene product from transcript CG31104-RA', other_designations='CG31104-PA|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25285473, end_position_on_the_genomic_accession=25286974, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43055, CurrentID=0, Status='live', Symbol='CG13658', Aliases='Dmel_ Dmel\\CG13658', description='CG13658 gene product from transcript CG13658-RB', other_designations='CG13658-PB|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25287474, end_position_on_the_genomic_accession=25288930, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43056, CurrentID=0, Status='live', Symbol='CG11893', Aliases='Dmel_ Dmel\\CG11893', description='CG11893 gene product from transcript CG11893-RA', other_designations='CG11893-PA|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25289889, end_position_on_the_genomic_accession=25291366, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43057, CurrentID=0, Status='live', Symbol='CG31098', Aliases='Dmel_ BcDNA:GH27183, BcDNA:RE13429, CG10634, Dmel\\CG31098', description='CG31098 gene product from transcript CG31098-RA', other_designations='CG31098-PA|uncharacterized protein', map_location='96C9-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25291370, end_position_on_the_genomic_accession=25292977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43058, CurrentID=0, Status='live', Symbol='CG31097', Aliases='Dmel_ CG13665, Dmel\\CG31097', description='CG31097 gene product from transcript CG31097-RB', other_designations='CG31097-PB|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25296128, end_position_on_the_genomic_accession=25297778, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43059, CurrentID=0, Status='live', Symbol='CG13659', Aliases='Dmel_ Dmel\\CG13659', description='CG13659 gene product from transcript CG13659-RA', other_designations='CG13659-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25300891, end_position_on_the_genomic_accession=25302317, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43060, CurrentID=0, Status='live', Symbol='CG31370', Aliases='Dmel_ CG13660, Dmel\\CG31370', description='CG31370 gene product from transcript CG31370-RB', other_designations='CG31370-PB|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25302464, end_position_on_the_genomic_accession=25303986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43061, CurrentID=0, Status='live', Symbol='CG10550', Aliases='Dmel_ Dmel\\CG10550', description='CG10550 gene product from transcript CG10550-RD', other_designations='CG10550-PB|CG10550-PC|CG10550-PD|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25306407, end_position_on_the_genomic_accession=25308391, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43062, CurrentID=0, Status='live', Symbol='CG31087', Aliases='Dmel_ CG10551, Dmel\\CG31087', description='CG31087 gene product from transcript CG31087-RA', other_designations='CG31087-PA|CG31087-PB|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25311027, end_position_on_the_genomic_accession=25313833, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43063, CurrentID=0, Status='live', Symbol='CG10559', Aliases='Dmel_ Dmel\\CG10559', description='CG10559 gene product from transcript CG10559-RB', other_designations='CG10559-PB|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25314494, end_position_on_the_genomic_accession=25316116, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43064, CurrentID=0, Status='live', Symbol='CG10553', Aliases='Dmel_ CT29612, Dmel\\CG10553', description='CG10553 gene product from transcript CG10553-RA', other_designations='CG10553-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25316461, end_position_on_the_genomic_accession=25318111, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43065, CurrentID=0, Status='live', Symbol='CG10560', Aliases='Dmel_ Dmel\\CG10560', description='CG10560 gene product from transcript CG10560-RA', other_designations='CG10560-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25318550, end_position_on_the_genomic_accession=25320101, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43066, CurrentID=0, Status='live', Symbol='CG10562', Aliases='Dmel_ Dmel\\CG10562', description='CG10562 gene product from transcript CG10562-RA', other_designations='CG10562-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25321053, end_position_on_the_genomic_accession=25322757, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43067, CurrentID=0, Status='live', Symbol='CHKov1', Aliases='Dmel_CG10618, BcDNA:GH03753, CG10618, Dmel\\CG10618, ref(3)D', description='CG10618 gene product from transcript CG10618-RE', other_designations='CG10618-PE|CHKov1-PE|refractory to sigma D', map_location='3-90 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25323156, end_position_on_the_genomic_accession=25329316, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43068, CurrentID=0, Status='live', Symbol='CHKov2', Aliases='Dmel_CG10675, CG10675, Dmel\\CG10675, ref(3)D', description='CG10675 gene product from transcript CG10675-RA', other_designations='CG10675-PA|CHKov2-PA|refractory to sigma D', map_location='3-90 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25329530, end_position_on_the_genomic_accession=25331203, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43069, CurrentID=0, Status='live', Symbol='CG10669', Aliases='Dmel_ Dmel\\CG10669', description='CG10669 gene product from transcript CG10669-RB', other_designations='CG10669-PA|CG10669-PB|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25331363, end_position_on_the_genomic_accession=25335243, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43070, CurrentID=0, Status='live', Symbol='CG11902', Aliases='Dmel_ Dmel\\CG11902, anon-96CDa', description='CG11902 gene product from transcript CG11902-RB', other_designations='CG11902-PB|uncharacterized protein', map_location='96D1-96D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25335257, end_position_on_the_genomic_accession=25340088, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43071, CurrentID=0, Status='live', Symbol='rha', Aliases='Dmel_CG11908, CG11908, Dmel\\CG11908, RADHA, anon-WO0140519.50', description='CG11908 gene product from transcript CG11908-RA', other_designations='CG11908-PA|rha-PA', map_location='96D2-96D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25340252, end_position_on_the_genomic_accession=25344838, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43072, CurrentID=0, Status='live', Symbol='tobi', Aliases='Dmel_CG11909, CG11909, CT33098, Dmel\\CG11909', description='target of brain insulin', other_designations='CG11909-PA|alpha-Glucosidase|tobi-PA', map_location='96D2-96D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25344761, end_position_on_the_genomic_accession=25346921, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43073, CurrentID=0, Status='live', Symbol='ND-49L', Aliases='Dmel_CG11913, CG111913, CG11913, Dmel\\CG11913, NUCM, Q9VBR4_DROME', description='NADH dehydrogenase (ubiquinone) 49 kDa subunit-like', other_designations='CG11913-PB|NADH-ubiquinone oxidoreductase 49 kDa subunit|NADH:ubiquinone oxidoreductase 49 kDa subunit|ND-49L-PB', map_location='96D2-96D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25358285, end_position_on_the_genomic_accession=25360034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43074, CurrentID=0, Status='live', Symbol='alrm', Aliases='Dmel_CG11910, CG11910, CT37074, Dmel\\CG11910', description='astrocytic leucine-rich repeat molecule', other_designations='CG11910-PA|alrm-PA', map_location='96D2-96D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25381065, end_position_on_the_genomic_accession=25382781, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19834948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834948, CurrentID=0, Status='live', Symbol='SclB', Aliases='Dmel_CG45091, BEST:LP06756, BcDNA:RE28911, CG45091, CR31696, Dmel\\CG45091, ORF-B, mel2, pncr003, pncr003:2L, pncr003:2L ORFB, pncr:003:2L, scl', description='Sarcolamban B', other_designations='CG45091-PA|CG45091-PB|CG45091-PC|CG45091-PD|SclB-PA|SclB-PB|SclB-PC|SclB-PD|sarcolamban|sarcolamban B', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16837972, end_position_on_the_genomic_accession=16841884, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43077, CurrentID=0, Status='live', Symbol='Vps33B', Aliases='Dmel_CG5127, CG5127, Dmel\\CG5127, VPS33b, Vps 33B, Vps33b, dVps33B', description='Vacuolar protein sorting 33B', other_designations='CG5127-PA|Vps33B-PA|vacuolar protein sorting 33B', map_location='96E1-96E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25473490, end_position_on_the_genomic_accession=25475756, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43078, CurrentID=0, Status='live', Symbol='CG4553', Aliases='Dmel_ Dmel\\CG4553', description='CG4553 gene product from transcript CG4553-RA', other_designations='CG4553-PA', map_location='96E1-96E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25475636, end_position_on_the_genomic_accession=25476915, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43079, CurrentID=0, Status='live', Symbol='MED28', Aliases='Dmel_CG5121, BcDNA:GM31575, CG5121, Dmel\\CG5121, Med23, Med28, dMED23', description='Mediator complex subunit 28', other_designations='CG5121-PA|MED28-PA|mediator complex subunit 28', map_location='96E1-96E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25477003, end_position_on_the_genomic_accession=25477745, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43080, CurrentID=0, Status='live', Symbol='XNP', Aliases='Dmel_CG4548, ATR-X/ CG4548, DATR-X, DATRX, Dmel\\CG4548, EP(3)0635, EP635, dATRX, dXNP, xnp, xnp/atr-x', description='CG4548 gene product from transcript CG4548-RA', other_designations='CG4548-PA|CG4548-PB|CG4548-PC|CG4548-PD|XNP-PA|XNP-PB|XNP-PC|XNP-PD', map_location='96E1-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25477868, end_position_on_the_genomic_accession=25482834, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43081, CurrentID=0, Status='live', Symbol='CG5116', Aliases='Dmel_ Dmel\\CG5116, Dromel_CG5116_FBtr0084877_mORF', description='CG5116 gene product from transcript CG5116-RB', other_designations='CG5116-PA|CG5116-PB', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25483098, end_position_on_the_genomic_accession=25485855, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43083, CurrentID=0, Status='live', Symbol='CG5112', Aliases='Dmel_ Dmel\\CG5112', description='CG5112 gene product from transcript CG5112-RA', other_designations='CG5112-PA', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25491460, end_position_on_the_genomic_accession=25493262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43084, CurrentID=0, Status='live', Symbol='CG5107', Aliases='Dmel_ Dmel\\CG5107, anon-WO0140519.57', description='CG5107 gene product from transcript CG5107-RA', other_designations='CG5107-PA', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25505191, end_position_on_the_genomic_accession=25505931, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43085, CurrentID=0, Status='live', Symbol='CG5111', Aliases='Dmel_ Dmel\\CG5111', description='CG5111 gene product from transcript CG5111-RA', other_designations='CG5111-PA', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25506275, end_position_on_the_genomic_accession=25508279, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43086, CurrentID=0, Status='live', Symbol='CG4582', Aliases='Dmel_ Dmel\\CG4582', description='CG4582 gene product from transcript CG4582-RA', other_designations='CG4582-PA', map_location='96E4-96E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25581223, end_position_on_the_genomic_accession=25583179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43087, CurrentID=0, Status='live', Symbol='msi', Aliases='Dmel_CG5099, CG5099, DMSIDNA, Dmel\\CG5099, MSI, Msi, SIDNA, anon-EST:Liang-2.35, clone 2.35, dMsi', description='musashi', other_designations='CG5099-PA|CG5099-PB|CG5099-PC|CG5099-PD|CG5099-PE|CG5099-PF|CG5099-PH|msi-PA|msi-PB|msi-PC|msi-PD|msi-PE|msi-PF|msi-PH|mushashi', map_location='96E2-96E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25516118, end_position_on_the_genomic_accession=25608423, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43088, CurrentID=0, Status='live', Symbol='ymp', Aliases='Dmel_CG12250, BcDNA:AT03213, CG12250, Dmel\\CG12250', description='yellow-emperor', other_designations='CG12250-PA|CG12250-PB|CG12250-PC|CG12250-PD|ymp-PA|ymp-PB|ymp-PC|ymp-PD', map_location='96E4-96E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25591502, end_position_on_the_genomic_accession=25609127, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  43089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43089, CurrentID=0, Status='live', Symbol='CG5079', Aliases='Dmel_ Dmel\\CG5079', description='CG5079 gene product from transcript CG5079-RA', other_designations='CG5079-PA', map_location='96E4-96E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25613666, end_position_on_the_genomic_accession=25614451, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43090, CurrentID=0, Status='live', Symbol='CG5071', Aliases='Dmel_ BcDNA:AT31683, Dmel\\CG5071', description='CG5071 gene product from transcript CG5071-RD', other_designations='CG5071-PA|CG5071-PB|CG5071-PC|CG5071-PD', map_location='96E5-96E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25618414, end_position_on_the_genomic_accession=25621838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43091, CurrentID=0, Status='live', Symbol='Npl4', Aliases='Dmel_CG4673, CG4673, Dmel\\CG4673, NPL4, dNPLOC4, dNp14, dNpl4, npl4', description='Npl4 ortholog', other_designations='CG4673-PA|CG4673-PB|CG4673-PD|Npl4-PA|Npl4-PB|Npl4-PD', map_location='96E4-96E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25611013, end_position_on_the_genomic_accession=25624662, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43092, CurrentID=0, Status='live', Symbol='Ssadh', Aliases='Dmel_CG4685, CG4685, DmSSADH, Dmel\\CG4685, SSADH, anon-WO0118547.428', description='Succinic semialdehyde dehydrogenase', other_designations='CG4685-PA|CG4685-PB|CG4685-PC|CG4685-PD|CG4685-PE|Ssadh-PA|Ssadh-PB|Ssadh-PC|Ssadh-PD|Ssadh-PE|succinic semialdehyde dehydrogenase', map_location='96E6-96E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25624805, end_position_on_the_genomic_accession=25626924, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43093, CurrentID=0, Status='live', Symbol='jigr1', Aliases='Dmel_CG17383, CG17383, Dmel\\CG17383, JIGR-1, JIGR1, bs29f05.y1', description='jing interacting gene regulatory 1', other_designations='CG17383-PA|CG17383-PB|CG17383-PC|CG17383-PD|CG17383-PE|CG17383-PF|CG17383-PG|jigr1-PA|jigr1-PB|jigr1-PC|jigr1-PD|jigr1-PE|jigr1-PF|jigr1-PG', map_location='96E6-96E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25637843, end_position_on_the_genomic_accession=25650397, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43095, CurrentID=0, Status='live', Symbol='Tnks', Aliases='Dmel_CG4719, BcDNA:LD22548, CG17487, CG4719, DTNKS, Dm.pARTb, Dmel\\CG4719, dTNKS', description='tankyrase', other_designations='CG4719-PA|CG4719-PB|Tnks-PA|Tnks-PB', map_location='96E7-96E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25653091, end_position_on_the_genomic_accession=25661286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43096, CurrentID=0, Status='live', Symbol='RASSF8', Aliases='Dmel_CG5053, Boa, CG5053, Dmel\\CG5053, d dmRASSF7/8', description='CG5053 gene product from transcript CG5053-RA', other_designations='CG5053-PA|RASSF8-PA', map_location='96E8-96E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25661795, end_position_on_the_genomic_accession=25665640, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43098, CurrentID=0, Status='live', Symbol='Lgr3', Aliases='Dmel_CG31096, CG31096, CG5042, CG5046, CT16193, DLGR3, Dmel\\CG31096, anon-WO0170980.25, anon-WO0170980.26, dLGR3', description='Leucine-rich repeat-containing G protein-coupled receptor 3', other_designations='CG31096-PA|Lgr3-PA', map_location='96E8-96E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25665809, end_position_on_the_genomic_accession=25677214, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  43099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43099, CurrentID=0, Status='live', Symbol='CG4730', Aliases='Dmel_ Dmel\\CG4730', description='CG4730 gene product from transcript CG4730-RA', other_designations='CG4730-PA|CG4730-PB|CG4730-PC', map_location='96E9-96E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25677072, end_position_on_the_genomic_accession=25678682, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43100, CurrentID=0, Status='live', Symbol='CG5039', Aliases='Dmel_ BcDNA:LD35121, Dmel\\CG5039', description='CG5039 gene product from transcript CG5039-RA', other_designations='CG5039-PA', map_location='96E9-96E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25678683, end_position_on_the_genomic_accession=25679529, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43101, CurrentID=0, Status='live', Symbol='CG4743', Aliases='Dmel_ Dmel\\CG4743', description='CG4743 gene product from transcript CG4743-RA', other_designations='CG4743-PA', map_location='96E9-96E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25679242, end_position_on_the_genomic_accession=25680550, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43102, CurrentID=0, Status='live', Symbol='CG5028', Aliases='Dmel_ Dmel\\CG5028', description='CG5028 gene product from transcript CG5028-RB', other_designations='CG5028-PA|CG5028-PB|CG5028-PC|CG5028-PD', map_location='96E9-96E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25680845, end_position_on_the_genomic_accession=25683780, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43103, CurrentID=0, Status='live', Symbol='RpL27', Aliases='Dmel_CG4759, CG4759, Dmel\\CG4759, RPL27e, chr3R:21510055..21510241, RpL27', description='ribosomal protein L27', other_designations='CG4759-PA|CG4759-PB|CG4759-PC|CG4759-PD|RpL27-PA|RpL27-PB|RpL27-PC|RpL27-PD', map_location='96E9-96E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25683623, end_position_on_the_genomic_accession=25685051, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43104, CurrentID=0, Status='live', Symbol='CLS', Aliases='Dmel_CG4774, CG4774, Dmel\\CG4774', description='Cardiolipin synthase', other_designations='CG4774-PA|CG4774-PB|CG4774-PC|CG4774-PD|CLS-PA|CLS-PB|CLS-PC|CLS-PD|cardiolipin synthase', map_location='96E10-96E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25685236, end_position_on_the_genomic_accession=25687370, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43105, CurrentID=0, Status='live', Symbol='LpR2', Aliases='Dmel_CG31092, CG31092, CG4823, CG4834, Dmel\\CG31092, LDLR, Lpr2, lpr2', description='Lipophorin receptor 2', other_designations='CG31092-PA|CG31092-PB|CG31092-PC|CG31092-PD|CG31092-PE|CG31092-PF|CG31092-PG|CG31092-PH|CG31092-PI|LDL-receptor-related protein|LDLR-like|LpR2-PA|LpR2-PB|LpR2-PC|LpR2-PD|LpR2-PE|LpR2-PF|LpR2-PG|LpR2-PH|LpR2-PI|a2M-receptor-like|lipophorin receptor 2|low-density lipoprotein receptor', map_location='96E10-96F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25698727, end_position_on_the_genomic_accession=25740465, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  43110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43110, CurrentID=0, Status='live', Symbol='CG17198', Aliases='Dmel_ Dmel\\CG17198', description='CG17198 gene product from transcript CG17198-RB', other_designations='CG17198-PA|CG17198-PB|uncharacterized protein', map_location='96F2-96F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25773334, end_position_on_the_genomic_accession=25775369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43111, CurrentID=0, Status='live', Symbol='CG17197', Aliases='Dmel_CG17197-PA, Dmel\\CG17197', description='CG17197 gene product from transcript CG17197-RB', other_designations='CG17197-PB|uncharacterized protein', map_location='96F2-96F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25773334, end_position_on_the_genomic_accession=25775369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43112, CurrentID=0, Status='live', Symbol='CG17196', Aliases='Dmel_CG17196-PA, Dmel\\CG17196', description='CG17196 gene product from transcript CG17196-RC', other_designations='CG17196-PA|CG17196-PB|CG17196-PC|uncharacterized protein', map_location='96F2-96F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25775288, end_position_on_the_genomic_accession=25776237, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43113, CurrentID=0, Status='live', Symbol='CG17195', Aliases='Dmel_CG17195-PA, Dmel\\CG17195', description='CG17195 gene product from transcript CG17195-RA', other_designations='CG17195-PA|uncharacterized protein', map_location='96F2-96F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25776276, end_position_on_the_genomic_accession=25777198, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43114, CurrentID=0, Status='live', Symbol='CG4956', Aliases='Dmel_ Dmel\\CG4956', description='CG4956 gene product from transcript CG4956-RA', other_designations='CG4956-PA', map_location='96F2-96F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25777268, end_position_on_the_genomic_accession=25778305, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43115, CurrentID=0, Status='live', Symbol='CG4960', Aliases='Dmel_ BcDNA:AT31258, Dmel\\CG4960', description='CG4960 gene product from transcript CG4960-RB', other_designations='CG4960-PA|CG4960-PB', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25786329, end_position_on_the_genomic_accession=25787298, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43116, CurrentID=0, Status='live', Symbol='CG31093', Aliases='Dmel_ BcDNA:AT30660, CG4964, Dmel\\CG31093', description='CG31093 gene product from transcript CG31093-RA', other_designations='CG31093-PA|uncharacterized protein', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25789819, end_position_on_the_genomic_accession=25790471, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43117, CurrentID=0, Status='live', Symbol='CG5024', Aliases='Dmel_ BcDNA:AT16150, Dmel\\CG5024, gene1', description='CG5024 gene product from transcript CG5024-RA', other_designations='CG5024-PA|CG5024-PB', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25811096, end_position_on_the_genomic_accession=25811775, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43118, CurrentID=0, Status='live', Symbol='CG17770', Aliases='Dmel_ Dmel\\CG17770, gene2', description='CG17770 gene product from transcript CG17770-RA', other_designations='CG17770-PA|uncharacterized protein', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25811911, end_position_on_the_genomic_accession=25813134, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43120, CurrentID=0, Status='live', Symbol='CG14354', Aliases='Dmel_ Dmel\\CG14354', description='CG14354 gene product from transcript CG14354-RA', other_designations='CG14354-PA|uncharacterized protein', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25815479, end_position_on_the_genomic_accession=25816574, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43121, CurrentID=0, Status='live', Symbol='Mst57Dc', Aliases='Dmel_CG4986, Acp57Dc, CG4986, Dmel\\CG4986, Mst75Dc, mst(2)ag-1, mst-1', description='Male-specific transcript 57Dc', other_designations='CG4986-PB|Mst57Dc-PB|Mst57dc for male accessory gland protein|male specific transcript 57D|male-specific RNA 57dc', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25829321, end_position_on_the_genomic_accession=25829698, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43122, CurrentID=0, Status='live', Symbol='Mst57Db', Aliases='Dmel_CG5016, CG5016, Dmel\\CG5016, mst(2)ag-1, mst-1', description='Male-specific RNA 57Db', other_designations='CG5016-PA|Mst57Db-PA|male specific transcript 57D', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25830720, end_position_on_the_genomic_accession=25830914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43124, CurrentID=0, Status='live', Symbol='alpha4GT2', Aliases='Dmel_CG5878, CG5878, Dmel\\CG5878', description='CG5878 gene product from transcript CG5878-RB', other_designations='CG5878-PB|alpha4GT2-PB', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25832230, end_position_on_the_genomic_accession=25842878, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43125, CurrentID=0, Status='live', Symbol='CG5886', Aliases='Dmel_ Dmel\\CG5886', description='CG5886 gene product from transcript CG5886-RA', other_designations='CG5886-PA', map_location='96F3-96F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25851632, end_position_on_the_genomic_accession=25853989, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43126, CurrentID=0, Status='live', Symbol='CG5890', Aliases='Dmel_ Dmel\\CG5890', description='CG5890 gene product from transcript CG5890-RD', other_designations='CG5890-PA|CG5890-PC|CG5890-PD', map_location='96F5-96F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25856154, end_position_on_the_genomic_accession=25872670, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43127, CurrentID=0, Status='live', Symbol='Tsp96F', Aliases='Dmel_CG6120, BcDNA:LD19727, CG6120, Dm.Tsp96F, Dmel\\CG6120, LD19727', description='Tetraspanin 96F', other_designations='CG6120-PA|CG6120-PB|CG6120-PC|Tsp96F-PA|Tsp96F-PB|Tsp96F-PC|tetraspanin 96F', map_location='96F5-96F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25872900, end_position_on_the_genomic_accession=25881427, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43128, CurrentID=0, Status='live', Symbol='SppL', Aliases='Dmel_CG17370, CG17370, CG5901, Dmel\\CG17370, SPPL3, Sppl, sppL', description='Signal peptide peptidase-like', other_designations='CG17370-PA|CG17370-PB|CG17370-PC|CG17370-PD|CG17370-PE|CG17370-PF|SPP-like protein 3|SppL-PA|SppL-PB|SppL-PC|SppL-PD|SppL-PE|SppL-PF|signal peptide peptidase-like', map_location='96F5-96F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25881984, end_position_on_the_genomic_accession=25888230, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43130, CurrentID=0, Status='live', Symbol='Lnk', Aliases='Dmel_CG17367, CG 17367, CG17367, CG5908, Dmel\\CG17367, LNK, dSH2B, lnk', description='CG17367 gene product from transcript CG17367-RC', other_designations='CG17367-PA|CG17367-PB|CG17367-PC|CG17367-PD|CG17367-PE|CG17367-PF|Lnk-PA|Lnk-PB|Lnk-PC|Lnk-PD|Lnk-PE|Lnk-PF', map_location='96F6-96F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25888542, end_position_on_the_genomic_accession=25893270, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43132, CurrentID=0, Status='live', Symbol='CG5913', Aliases='Dmel_ Dmel\\CG5913', description='CG5913 gene product from transcript CG5913-RA', other_designations='CG5913-PA', map_location='96F6-96F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25893495, end_position_on_the_genomic_accession=25895026, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43133, CurrentID=0, Status='live', Symbol='CG5948', Aliases='Dmel_ Dmel\\CG5948', description='CG5948 gene product from transcript CG5948-RA', other_designations='CG5948-PA|CG5948-PB', map_location='96F6-96F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25895822, end_position_on_the_genomic_accession=25897174, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43134, CurrentID=0, Status='live', Symbol='MCO3', Aliases='Dmel_CG5959, CG5959, Dmel\\CG5959', description='multicopper oxidase 3', other_designations='CG5959-PA|MCO3-PA|multicopper oxidase', map_location='96F7-96F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25897476, end_position_on_the_genomic_accession=25899744, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43135, CurrentID=0, Status='live', Symbol='CG31089', Aliases='Dmel_ CG5990, Dmel\\CG31089', description='CG31089 gene product from transcript CG31089-RA', other_designations='CG31089-PA|uncharacterized protein', map_location='96F7-96F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25902643, end_position_on_the_genomic_accession=25904178, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43136, CurrentID=0, Status='live', Symbol='CG42235', Aliases='Dmel_ CG14539, CG31090, CG8932, CG8934, CG8951, CG8957, CG8966, Dmel\\CG42235, Dmel_CG31090, Dmel_CG8932, Dmel_CG8934, Dmel_CG8957, Dmel_CG8966, dSLC5A10', description='CG42235 gene product from transcript CG42235-RE', other_designations='CG42235-PA|CG42235-PB|CG42235-PC|CG42235-PD|CG42235-PE|Na/iodide symporter-like|uncharacterized protein', map_location='96F8-96F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25907022, end_position_on_the_genomic_accession=25921046, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  19835010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835010, CurrentID=0, Status='live', Symbol='CG44956', Aliases='Dmel_ Dmel\\CG44956', description='CG44956 gene product from transcript CG44956-RA', other_designations='CG44956-PA|uncharacterized protein', map_location='87B7-87B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12341674, end_position_on_the_genomic_accession=12342373, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43142, CurrentID=0, Status='live', Symbol='CG14546', Aliases='Dmel_ Dmel\\CG14546', description='CG14546 gene product from transcript CG14546-RA', other_designations='CG14546-PA|uncharacterized protein', map_location='96F8-96F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25921042, end_position_on_the_genomic_accession=25922634, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835016, CurrentID=0, Status='live', Symbol='CG45100', Aliases='Dmel_ Dmel\\CG45100, Dmel_Hdac3_uORF', description='CG45100 gene product from transcript CG45100-RA', other_designations='CG45100-PA|uncharacterized protein', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5472810, end_position_on_the_genomic_accession=5474600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43145, CurrentID=0, Status='live', Symbol='CG14540', Aliases='Dmel_ Dmel\\CG14540', description='CG14540 gene product from transcript CG14540-RA', other_designations='CG14540-PA|uncharacterized protein', map_location='96F8-96F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25959279, end_position_on_the_genomic_accession=25961686, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43146, CurrentID=0, Status='live', Symbol='boss', Aliases='Dmel_CG8285, BOSS, Bos, Boss, CG8285, Dmel\\CG8285, F, X55887', description='bride of sevenless', other_designations='CG8285-PA|F transcript|boss-PA', map_location='3-90.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25961994, end_position_on_the_genomic_accession=25966948, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43148, CurrentID=0, Status='live', Symbol='CG42261', Aliases='Dmel_ CG14541, CG8968, Dmel\\CG42261, Dmel_CG14541, Dmel_CG8968', description='CG42261 gene product from transcript CG42261-RB', other_designations='CG42261-PB|uncharacterized protein', map_location='96F9-96F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25970080, end_position_on_the_genomic_accession=25982736, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43149, CurrentID=0, Status='live', Symbol='Nf1', Aliases='Dmel_CG8318, CG8318, Dmel\\CG8318, NF-1, NF1, Nf-1, dNF1, dNf1, nf1', description='Neurofibromin 1', other_designations='CG8318-PB|CG8318-PC|CG8318-PD|CG8318-PE|CG8318-PF|Neurofibromatosis-1|Nf1-PB|Nf1-PC|Nf1-PD|Nf1-PE|Nf1-PF|neurofibramatosis-1|neurofibromatosis 1|neurofibromatosis factor 1|neurofibromatosis type 1|neurofibromin|neurofibromin 1', map_location='96F9-96F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25983013, end_position_on_the_genomic_accession=25995737, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  43150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43150, CurrentID=0, Status='live', Symbol='E(spl)mdelta-HLH', Aliases='Dmel_CG8328, CG8328, Dmel\\CG8328, E(Spl) HLH-m delta, E(Spl)-HLH-mdelta, E(spl) mdelta, E(spl)-mdelta, E(spl)Mdelta, E(spl)delta, E(spl)mC, E(spl)mdelta, HLH-mdelta, HLHMD, HLHMdelta, HLHmC, HLHmdelta, bHLHc14, mC, md, mdelta, mdelta0.5', description='Enhancer of split mdelta, helix-loop-helix', other_designations='CG8328-PA|E(spl) region transcript mdelta|E(spl)mdelta-HLH-PA|enhancer of split mdelta, helix-loop-helix|enhancer of split region transcript HLHmC|split locus enhancer protein mC', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25997621, end_position_on_the_genomic_accession=25998633, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43151, CurrentID=0, Status='live', Symbol='E(spl)mgamma-HLH', Aliases='Dmel_CG8333, CG8333, Dm-mB, Dmel\\CG8333, E(Spl) HLH-m gamma, E(spl) m gamma, E(spl) mg, E(spl) mgamma, E(spl)-mgamma, E(spl)Mgamma, E(spl)gamma, E(spl)mB, E(spl)mgamma, HLH-mgamma, HLHMgamma, HLHmB, HLHmgamma, HlHmg, Mgamma, anon-EST:Liang-1.49, bHLHb34, clone 1.49, e(spl)mgamma, gamma, mB, mg, mgamma', description='Enhancer of split mgamma, helix-loop-helix', other_designations='CG8333-PA|E(spl) region transcript g|E(spl) region transcript mgamma|E(spl)mgamma-HLH-PA|enhancer of split mgamma|enhancer of split mgamma, helix-loop-helix|enhancer of split region transcript HLHmB|split locus enhancer protein mB', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25999851, end_position_on_the_genomic_accession=26000761, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43152, CurrentID=0, Status='live', Symbol='E(spl)mbeta-HLH', Aliases='Dmel_CG14548, CG14548, Dm-mA, Dmel\\CG14548, E(Spl) HLH-Mbeta, E(Spl)mbeta, E(spl), E(spl) mbeta, E(spl)-Mbeta, E(spl)-mbeta, E(spl)Mbeta, E(spl)beta, E(spl)mA, E(spl)mb, E(spl)mbeta, E(spl)mbeta}, HLH-mbeta, HLHMbeta, HLHmA, HLHmbeta, Mbeta, bHLHb34, m-beta, mA, mb, mbeta', description='Enhancer of split mbeta, helix-loop-helix', other_designations='CG14548-PA|E(spl) region transcript mbeta|E(spl)mbeta-HLH-PA|enhancer of split m-beta|enhancer of split mbeta|enhancer of split mbeta, helix-loop-helix|enhancer of split region transcript HLHmA|split locus enhancer protein mA', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26005046, end_position_on_the_genomic_accession=26006023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43153, CurrentID=0, Status='live', Symbol='E(spl)malpha-BFM', Aliases='Dmel_CG8337, CG8337, DmelCG8337, Dmel\\CG8337, E(spl) malpha, E(spl)ma, E(spl)malpha, E(spl)mmalpha, HLHmalpha, malpha', description='Enhancer of split malpha, Bearded family member', other_designations='CG8337-PA|E(spl) region transcript malpha|E(spl)malpha|E(spl)malpha-BFM-PA|enhancer of split malpha, bearded family member', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26010893, end_position_on_the_genomic_accession=26011568, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43154, CurrentID=0, Status='live', Symbol='Kaz-m1', Aliases='Dmel_CG8342, BcDNA:RE44854, CG8342, Dmel\\CG8342, E(spl)-m1, E(spl)m1, HLHm1, lincRNA.856, m1', description='Kazal-type protease inhibitor m1', other_designations='CG8342-PA|CG8342-PB|E(spl) region transcript m1|Kaz-m1-PA|Kaz-m1-PB', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26014721, end_position_on_the_genomic_accession=26016487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43155, CurrentID=0, Status='live', Symbol='E(spl)m2-BFM', Aliases='Dmel_CG6104, CG6104, DmelCG6104, Dmel\\CG6104, E(Spl)m2, E(spl), E(spl)m2, HLHm2, m2', description='Enhancer of split m2, Bearded family member', other_designations='CG6104-PA|E(spl) region transcript m2|E(spl)m2-BFM-PA|enhancer of split m2, bearded family member', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26016514, end_position_on_the_genomic_accession=26017470, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43156, CurrentID=0, Status='live', Symbol='E(spl)m3-HLH', Aliases='Dmel_CG8346, CG8346, Dmel\\CG8346, E(Spl)-HLH-m3, E(spl)-m3, E(spl)C, E(spl)C m3, E(spl)M3, E(spl)m3, HLH-m3, HLHM3, HLHm3, bHLHc14, m3', description='Enhancer of split m3, helix-loop-helix', other_designations='CG8346-PA|E(spl) region transcript m3|E(spl)m3-HLH-PA|enhancer of split m3|enhancer of split m3, helix-loop-helix|split locus enhancer protein m3', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26021823, end_position_on_the_genomic_accession=26023127, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43157, CurrentID=0, Status='live', Symbol='E(spl)m4-BFM', Aliases='Dmel_CG6099, CG6099, DmelCG6099, Dmel\\CG6099, E(spl) m4, E(spl)m4, HLHm4, m4', description='Enhancer of split m4, Bearded family member', other_designations='CG6099-PA|E(spl) region transcript 4|E(spl) region transcript m4|E(spl)m4-BFM-PA|enhancer of split m4|enhancer of split m4, bearded family member', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26023821, end_position_on_the_genomic_accession=26024579, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43158, CurrentID=0, Status='live', Symbol='E(spl)m5-HLH', Aliases='Dmel_CG6096, CG6096, Dmel\\CG6096, E(spl) m5, E(spl) transcription unit m5, E(spl)-m5, E(spl)M5, E(spl)m5, HLH-m5, HLHm5, bHLHb36, m5', description='Enhancer of split m5, helix-loop-helix', other_designations='CG6096-PA|E(spl) region transcript m5|E(spl)m5-HLH-PA|enhancer of split m5, helix-loop-helix', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26028854, end_position_on_the_genomic_accession=26029735, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43159, CurrentID=0, Status='live', Symbol='E(spl)m6-BFM', Aliases='Dmel_CG8354, CG8354, DmelCG8354, Dmel\\CG8354, E(spl)M6, E(spl)m6, HLHm6, M6, m6', description='Enhancer of split m6, Bearded family member', other_designations='CG8354-PA|E(spl) region transcript m6|E(spl)m6-BFM-PA|enhancer of split m6, bearded family member|enhancer of split region transcript m6', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26032911, end_position_on_the_genomic_accession=26034047, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43160, CurrentID=0, Status='live', Symbol='E(spl)m7-HLH', Aliases='Dmel_CG8361, CG8361, Dmel\\CG8361, E(SPL)m7, E(spl) m7, E(spl)-M7, E(spl)-m7, E(spl)M7, E(spl)m7, ESPL, En(spl)m7, HLH-m7, HLHm7, HlHm7, M7, bHLHb35, m7', description='Enhancer of split m7, helix-loop-helix', other_designations='CG8361-PA|E(spl) region transcript 7|E(spl) region transcript m7|E(spl)m7|E(spl)m7-HLH-PA|enhancer of split m7|enhancer of split m7, helix-loop-helix', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26037038, end_position_on_the_genomic_accession=26038019, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43161, CurrentID=0, Status='live', Symbol='E(spl)m8-HLH', Aliases='Dmel_CG8365, CG8365, Dmel\\CG8365, E(Spl), E(spl), E(spl) m8, E(spl)-C, E(spl)-m8, E(spl)M8, E(spl)bHLH, E(spl)m8, En(spl), En(spl)-C, Espl, HLHm8, Hes, M8, Su(H)[[m8]], anon-EST:Liang-1.71, bHLHb36, clone 1.71, e(spl), l(3)96Fd, m8', description='Enhancer of split m8, helix-loop-helix', other_designations='CG8365-PA|E(spl)m8|E(spl)m8-HLH-PA|Enhancer-of-split|enhancer of split|enhancer of split m8|enhancer of split m8, helix-loop-helix|enhancer-of-split-m8', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26040240, end_position_on_the_genomic_accession=26041008, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43162, CurrentID=0, Status='live', Symbol='gro', Aliases='Dmel_CG8384, BEST:GM01575, BEST:LD15161, BcDNA.LD33829, BcDNA:LD33829, CG8384, DGro, Dmel\\CG8384, E(spl)-WD, E(spl)[2], E(spl)gro, E(spl)m9/m10, En-spl, GRO, Grg/TLE, Gro, Groucho, anon-WO0118547.385, dAES1, dAES2, dGro, l(3)gro, m10, m9, m9-m10, m9/10, m9/m10', description='groucho', other_designations='CG8384-PA|CG8384-PB|CG8384-PC|CG8384-PD|CG8384-PE|CG8384-PF|CG8384-PG|CG8384-PH|CG8384-PI|CG8384-PJ|dgroucho|gro-PA|gro-PB|gro-PC|gro-PD|gro-PE|gro-PF|gro-PG|gro-PH|gro-PI|gro-PJ', map_location='3-89.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26041736, end_position_on_the_genomic_accession=26049911, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  43163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43163, CurrentID=0, Status='live', Symbol='Exo84', Aliases='Dmel_CG6095, CG6095, DExo84, Dmel\\CG6095p, anon-WO0153538.55, exo84, onr, Exo84', description='Exo84 ortholog (S. cerevisiae)', other_designations='CG6095-PA|CG6095-PB|Exo84 ortholog|Exo84-PA|Exo84-PB|exo84|onion ring|onion rings', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26049731, end_position_on_the_genomic_accession=26052257, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43164, CurrentID=0, Status='live', Symbol='Vps2', Aliases='Dmel_CG14542, CG14542, Did4, DmVPS2, Dmel\\CG14542, Dvps2, NEST:bs02f06, dvps2, vps2', description='Vacuolar protein sorting 2', other_designations='CG14542-PA|Vps2-PA|vacuolar protein sorting 2|vps2', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26052821, end_position_on_the_genomic_accession=26054128, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43165, CurrentID=0, Status='live', Symbol='Dak1', Aliases='Dmel_CG6092, CG6092L, Dm.UMP-CMP kinase, Dmel\\CG6092, anon-WO0118547.320, dak1', description='CG6092 gene product from transcript CG6092-RA', other_designations='CG6092-PA|Dak1-PA', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26054237, end_position_on_the_genomic_accession=26055438, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43166, CurrentID=0, Status='live', Symbol='Sld5', Aliases='Dmel_CG14549, CG14549, Dmel\\CG14549, anon-WO0172774.61', description='CG14549 gene product from transcript CG14549-RA', other_designations='CG14549-PA|Sld5-PA', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26055657, end_position_on_the_genomic_accession=26056547, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43167, CurrentID=0, Status='live', Symbol='CG14543', Aliases='Dmel_ Dmel\\CG14543', description='CG14543 gene product from transcript CG14543-RA', other_designations='CG14543-PA|uncharacterized protein', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26056630, end_position_on_the_genomic_accession=26057350, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43168, CurrentID=0, Status='live', Symbol='CG14550', Aliases='Dmel_ BcDNA:GM30016, Dmel\\CG14550, Dromel_CG14550_FBtr0084974_mORF', description='CG14550 gene product from transcript CG14550-RA', other_designations='CG14550-PA|uncharacterized protein', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26057347, end_position_on_the_genomic_accession=26058223, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43169, CurrentID=0, Status='live', Symbol='RpL34a', Aliases='Dmel_CG6090, CG6090, Dmel\\CG6090, L34, RpL34', description='Ribosomal protein L34a', other_designations='CG6090-PA|CG6090-PB|CG6090-PC|RpL34a-PA|RpL34a-PB|RpL34a-PC|ribosomal protein L34a', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26058420, end_position_on_the_genomic_accession=26059267, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43170, CurrentID=0, Status='live', Symbol='CG14544', Aliases='Dmel_ Dmel\\CG14544', description='CG14544 gene product from transcript CG14544-RA', other_designations='CG14544-PA|CG14544-PB|uncharacterized protein', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26059496, end_position_on_the_genomic_accession=26061293, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43171, CurrentID=0, Status='live', Symbol='CG14551', Aliases='Dmel_ Dmel\\CG14551', description='CG14551 gene product from transcript CG14551-RA', other_designations='CG14551-PA|CG14551-PB|uncharacterized protein', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26061268, end_position_on_the_genomic_accession=26063011, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43174, CurrentID=0, Status='live', Symbol='dysf', Aliases='Dmel_CG32474, CG12561, CG14552, CG14553, CG14554, CG32474, Dmel\\CG32474, Dys, NXFL, dNXFL, dys', description='dysfusion', other_designations='CG32474-PB|CG32474-PC|CG32474-PD|NXF-like-factor|cranky|dysf-PB|dysf-PC|dysf-PD', map_location='96F12-96F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26076592, end_position_on_the_genomic_accession=26098711, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  43176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43176, CurrentID=0, Status='live', Symbol='CG14556', Aliases='Dmel_ Dmel\\CG14556', description='CG14556 gene product from transcript CG14556-RB', other_designations='CG14556-PB|uncharacterized protein', map_location='97A1-97A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26126659, end_position_on_the_genomic_accession=26127868, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43177, CurrentID=0, Status='live', Symbol='CG31324', Aliases='Dmel_ BcDNA:RH68619, CG14555, CG14557, Dmel\\CG31324', description='CG31324 gene product from transcript CG31324-RB', other_designations='CG31324-PA|CG31324-PB|CG31324-PC|uncharacterized protein', map_location='96F14-97A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26106868, end_position_on_the_genomic_accession=26130674, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43178, CurrentID=0, Status='live', Symbol='CG6142', Aliases='Dmel_ Dmel\\CG6142', description='CG6142 gene product from transcript CG6142-RA', other_designations='CG6142-PA', map_location='97A1-97A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26133154, end_position_on_the_genomic_accession=26137148, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43179, CurrentID=0, Status='live', Symbol='CG31086', Aliases='Dmel_ CG14558, Dmel\\CG31086', description='CG31086 gene product from transcript CG31086-RB', other_designations='CG31086-PA|CG31086-PB|uncharacterized protein', map_location='97A4-97A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26185251, end_position_on_the_genomic_accession=26186227, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43180, CurrentID=0, Status='live', Symbol='CG6073', Aliases='Dmel_ Dmel\\CG6073', description='CG6073 gene product from transcript CG6073-RA', other_designations='CG6073-PA|CG6073-PB', map_location='97A4-97A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26187018, end_position_on_the_genomic_accession=26188807, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43181, CurrentID=0, Status='live', Symbol='CG34129', Aliases='Dmel_ CG6069, Dmel\\CG34129, SPH192', description='CG34129 gene product from transcript CG34129-RA', other_designations='CG34129-PA|uncharacterized protein', map_location='97A4-97A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26190234, end_position_on_the_genomic_accession=26191205, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43182, CurrentID=0, Status='live', Symbol='CG12290', Aliases='Dmel_ Dmel\\CG12290, anon-WO0170980.49, anon-WO0170980.50, lincRNA.S7704', description='CG12290 gene product from transcript CG12290-RB', other_designations='CG12290-PA|CG12290-PB|uncharacterized protein', map_location='97A6-97A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26230248, end_position_on_the_genomic_accession=26234700, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43183, CurrentID=0, Status='live', Symbol='Ald', Aliases='Dmel_CG6058, ALD, BcDNA:LP07735, CG6058, Dmel\\CG6058, FBA, T12, T4, T6, T7, ald', description='Aldolase', other_designations='Ald-PB|Ald-PC|Ald-PD|Ald-PE|Ald-PH|Ald-PI|Ald-PJ|Ald-PK|Ald-PL|Ald-PM|CG6058-PB|CG6058-PC|CG6058-PD|CG6058-PE|CG6058-PH|CG6058-PI|CG6058-PJ|CG6058-PK|CG6058-PL|CG6058-PM|aldolase|fructose 1,6 biphosphate-aldolase A', map_location='3-91.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26254069, end_position_on_the_genomic_accession=26261591, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43184, CurrentID=0, Status='live', Symbol='CG6154', Aliases='Dmel_ Dmel\\CG6154', description='CG6154 gene product from transcript CG6154-RD', other_designations='CG6154-PC|CG6154-PD', map_location='97A7-97A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26262662, end_position_on_the_genomic_accession=26277050, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43185, CurrentID=0, Status='live', Symbol='CG6036', Aliases='Dmel_ Dmel\\CG6036, PP2Cb', description='CG6036 gene product from transcript CG6036-RA', other_designations='CG6036-PA', map_location='97A9-97A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26344290, end_position_on_the_genomic_accession=26345473, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43186, CurrentID=0, Status='live', Symbol='CG33970', Aliases='Dmel_ CG14559, CG31085, CG6162, CG6166, Dmel\\CG33970', description='CG33970 gene product from transcript CG33970-RC', other_designations='CG33970-PA|CG33970-PB|CG33970-PC|uncharacterized protein', map_location='97A8-97A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26318355, end_position_on_the_genomic_accession=26362588, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  3354995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354995, CurrentID=0, Status='live', Symbol='CG17169', Aliases='Dmel_ Dmel\\CG17169', description='CG17169 gene product from transcript CG17169-RB', other_designations='CG17169-PB|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23084672, end_position_on_the_genomic_accession=23085536, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43188, CurrentID=0, Status='live', Symbol='ppk15', Aliases='Dmel_CG14239, CG14239, Dmel\\CG14239', description='pickpocket 15', other_designations='CG14239-PB|ppk15-PB', map_location='97B1-97B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26401909, end_position_on_the_genomic_accession=26403656, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43189, CurrentID=0, Status='live', Symbol='CG5432', Aliases='Dmel_ Dmel\\CG5432', description='CG5432 gene product from transcript CG5432-RA', other_designations='CG5432-PA|CG5432-PB', map_location='97B2-97B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26414088, end_position_on_the_genomic_accession=26416192, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43190, CurrentID=0, Status='live', Symbol='tx', Aliases='Dmel_CG5441, CG5441, Dei, Delilah, Dmel\\CG5441, bHLHc10, dei, del, delilah', description='taxi', other_designations='CG5441-PA|CG5441-PB|delilah|taxi wings|tx-PA|tx-PB', map_location='3-91 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26440778, end_position_on_the_genomic_accession=26448991, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43191, CurrentID=0, Status='live', Symbol='Hex-t2', Aliases='Dmel_CG32849, CG32849, CG5443, Dmel\\CG32849, HK, Hex, anon-WO0140519.203', description='CG32849 gene product from transcript CG32849-RA', other_designations='CG32849-PA|Hex-t2-PA|hexokinase', map_location='97B3-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26450671, end_position_on_the_genomic_accession=26452327, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43192, CurrentID=0, Status='live', Symbol='CG5447', Aliases='Dmel_ Dmel\\CG5447', description='CG5447 gene product from transcript CG5447-RA', other_designations='CG5447-PA|CG5447-PB', map_location='97B3-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26452534, end_position_on_the_genomic_accession=26454585, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43193, CurrentID=0, Status='live', Symbol='Pdf', Aliases='Dmel_CG6496, BcDNA:RH08487, CG6496, Dmel\\CG6496, Drm-PDF, Drm-PDH, Drm-pdf, PAP, PDF, PDH, anon-WO0140519.261, anon-WO0140519.3, cPDH, pdf', description='Pigment-dispersing factor', other_designations='CG6496-PA|PDF-associated peptide|PIGMENT DISPERSING FACTOR|Pdf-PA|Pigment-dispersing hormone-like peptide|pigment dispersing hormone|pigment dispersion factor|pigment-dispersing hormone', map_location='97B3-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26455462, end_position_on_the_genomic_accession=26456141, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43194, CurrentID=0, Status='live', Symbol='CG14237', Aliases='Dmel_ Dmel\\CG14237', description='CG14237 gene product from transcript CG14237-RB', other_designations='CG14237-PB|uncharacterized protein', map_location='97B3-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26456877, end_position_on_the_genomic_accession=26457242, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43195, CurrentID=0, Status='live', Symbol='CG14238', Aliases='Dmel_ Dmel\\CG14238', description='CG14238 gene product from transcript CG14238-RA', other_designations='CG14238-PA|uncharacterized protein', map_location='97B3-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26458249, end_position_on_the_genomic_accession=26460981, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43196, CurrentID=0, Status='live', Symbol='CG5455', Aliases='Dmel_ CT17296, Dmel\\CG5455', description='CG5455 gene product from transcript CG5455-RA', other_designations='CG5455-PA|CG5455-PB', map_location='97B4-97B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26461527, end_position_on_the_genomic_accession=26470008, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43197, CurrentID=0, Status='live', Symbol='plum', Aliases='Dmel_CG6490, CG6490, CT20213, Dmel\\CG6490, Plum', description='CG6490 gene product from transcript CG6490-RD', other_designations='CG6490-PB|CG6490-PC|CG6490-PD|CG6490-PE|plum-PB|plum-PC|plum-PD|plum-PE', map_location='97B4-97B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26470078, end_position_on_the_genomic_accession=26535217, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43200, CurrentID=0, Status='live', Symbol='TwdlM', Aliases='Dmel_CG5468, CG5468, Dmel\\CG5468', description='TweedleM', other_designations='CG5468-PA|TwdlM-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26619118, end_position_on_the_genomic_accession=26620224, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43201, CurrentID=0, Status='live', Symbol='TwdlP', Aliases='Dmel_CG14240, CG14240, Dmel\\CG14240', description='TweedleP', other_designations='CG14240-PA|TwdlP-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26620608, end_position_on_the_genomic_accession=26621364, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43202, CurrentID=0, Status='live', Symbol='TwdlB', Aliases='Dmel_CG6478, CG6478, Dmel\\CG6478', description='TweedleB', other_designations='CG6478-PA|TwdlB-PA|tweedle B', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26622451, end_position_on_the_genomic_accession=26623543, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43203, CurrentID=0, Status='live', Symbol='TwdlL', Aliases='Dmel_CG6447, CG6447, Dmel\\CG6447, TwdIL', description='TweedleL', other_designations='CG6447-PA|CG6447-PB|TwdlL-PA|TwdlL-PB', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26624587, end_position_on_the_genomic_accession=26625652, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43204, CurrentID=0, Status='live', Symbol='TwdlO', Aliases='Dmel_CG6452, CG6452, Dmel\\CG6452', description='TweedleO', other_designations='CG6452-PA|TwdlO-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26625932, end_position_on_the_genomic_accession=26626738, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43205, CurrentID=0, Status='live', Symbol='TwdlK', Aliases='Dmel_CG6460, CG6460, Dmel\\CG6460', description='TweedleK', other_designations='CG6460-PA|TwdlK-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26627885, end_position_on_the_genomic_accession=26628724, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43206, CurrentID=0, Status='live', Symbol='TwdlJ', Aliases='Dmel_CG5471, CG5471, Dmel\\CG5471', description='TweedleJ', other_designations='CG5471-PB|TwdlJ-PB', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26629918, end_position_on_the_genomic_accession=26630825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43207, CurrentID=0, Status='live', Symbol='TwdlN', Aliases='Dmel_CG5476, BcDNA:RE20092, BcDNA:RE59605, CG5476, Dmel\\CG5476, chr3R:22457526..22457631', description='TweedleN', other_designations='CG5476-PA|TwdlN-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26631869, end_position_on_the_genomic_accession=26632956, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43208, CurrentID=0, Status='live', Symbol='TwdlH', Aliases='Dmel_CG31080, CG14241, CG31080, Dmel\\CG31080', description='TweedleH', other_designations='CG31080-PA|TwdlH-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26633732, end_position_on_the_genomic_accession=26634714, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43209, CurrentID=0, Status='live', Symbol='TwdlS', Aliases='Dmel_CG14242, CG14242, Dmel\\CG14242', description='TweedleS', other_designations='CG14242-PA|TwdlS-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26636688, end_position_on_the_genomic_accession=26637557, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43210, CurrentID=0, Status='live', Symbol='TwdlD', Aliases='Dmel_CG14243, CG14243, Dmel\\CG14243', description='TweedleD', other_designations='CG14243-PA|TwdlD-PA', map_location='97C3-97C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26640025, end_position_on_the_genomic_accession=26641054, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43212, CurrentID=0, Status='live', Symbol='Tb', Aliases='Dmel_CG5480, CG5480, Dmel\\CG5480, Tu, TwdlA', description='Tubby', other_designations='CG5480-PA|Tb-PA|TweedleA|tubby', map_location='3-90.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26656440, end_position_on_the_genomic_accession=26657531, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43213, CurrentID=0, Status='live', Symbol='beat-VII', Aliases='Dmel_CG42242, Beat-VII, CG14248, CG14249, CG42242, Dmel\\CG42242, Dmel_CG14248, Dmel_CG14249, beat VII', description='CG42242 gene product from transcript CG42242-RD', other_designations='CG42242-PD|beat-VII-PD', map_location='97C3-97C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26645282, end_position_on_the_genomic_accession=26681966, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43214, CurrentID=0, Status='live', Symbol='TwdlQ', Aliases='Dmel_CG14250, CG14250, Dmel\\CG14250', description='TweedleQ', other_designations='CG14250-PA|TwdlQ-PA', map_location='97C4-97C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26692546, end_position_on_the_genomic_accession=26693590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43215, CurrentID=0, Status='live', Symbol='amon', Aliases='Dmel_CG6438, Amon, CG6438, DMH#5, Dmel\\CG6438, PC2, PPP, Ppp, dPC2', description='amontillado', other_designations='CG6438-PA|amon-PA|prohormone processing protease', map_location='97C4-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26697697, end_position_on_the_genomic_accession=26714159, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  43216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43216, CurrentID=0, Status='live', Symbol='CG6425', Aliases='Dmel_ Dmel\\CG6425', description='CG6425 gene product from transcript CG6425-RA', other_designations='CG6425-PA', map_location='97C5-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26715165, end_position_on_the_genomic_accession=26717389, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43217, CurrentID=0, Status='live', Symbol='CG5484', Aliases='Dmel_ COP, Dmel\\CG5484, anon-WO0118547.366', description='CG5484 gene product from transcript CG5484-RC', other_designations='CG5484-PA|CG5484-PB|CG5484-PC', map_location='97C5-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26717871, end_position_on_the_genomic_accession=26719806, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43218, CurrentID=0, Status='live', Symbol='CG6420', Aliases='Dmel_ Dmel\\CG6420', description='CG6420 gene product from transcript CG6420-RB', other_designations='CG6420-PA|CG6420-PB', map_location='97C5-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26719847, end_position_on_the_genomic_accession=26725216, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43220, CurrentID=0, Status='live', Symbol='CG6403', Aliases='Dmel_ Dmel\\CG6403', description='CG6403 gene product from transcript CG6403-RB', other_designations='CG6403-PA|CG6403-PB', map_location='97D1-97D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26735417, end_position_on_the_genomic_accession=26736307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43221, CurrentID=0, Status='live', Symbol='CG14247', Aliases='Dmel_ Dmel\\CG14247', description='CG14247 gene product from transcript CG14247-RA', other_designations='CG14247-PA|uncharacterized protein', map_location='97D1-97D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26737610, end_position_on_the_genomic_accession=26738593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43222, CurrentID=0, Status='live', Symbol='Tl', Aliases='Dmel_CG5490, CG5490, CT17414, Dmel\\CG5490, EP(3)1051, EP1051, Fs(1)Tl, Fs(3)Tl, T1, TL, TOL, Toll-1, Toll1, dToll, dToll1, mat(3)9, mel(3)10, mel(3)9, tl, toll, toll-1', description='Toll', other_designations='CG5490-PB|CG5490-PC|CG5490-PD|EP1051|Tl-PB|Tl-PC|Tl-PD|Toll-1|protein toll precursor|toll|toll receptor', map_location='3-91 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26799041, end_position_on_the_genomic_accession=26842403, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43223, CurrentID=0, Status='live', Symbol='Lerp', Aliases='Dmel_CG31072, CG14255, CG14256, CG31072, CG6390, CG6396, Dmel\\CG31072, LERP, anon-WO0118547.383', description='lysosomal enzyme receptor protein', other_designations='CG31072-PA|CG31072-PB|CG31072-PD|CG31072-PE|Lerp-PA|Lerp-PB|Lerp-PD|Lerp-PE', map_location='97D2-97D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26842462, end_position_on_the_genomic_accession=26863130, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  43227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43227, CurrentID=0, Status='live', Symbol='IntS12', Aliases='Dmel_CG5491, CG5491, Dmel\\CG5491, Int12', description='Integrator 12', other_designations='CG5491-PA|IntS12-PA|integrator 12', map_location='97D3-97D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26863376, end_position_on_the_genomic_accession=26864559, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43228, CurrentID=0, Status='live', Symbol='ball', Aliases='Dmel_CG6386, BALL, Ball, BcDNA:LD09009, CG6386, Dmel\\CG6386, NHK-1, NHK1, VRK, nhk-1, trip', description='ballchen', other_designations='CG6386-PA|CG6386-PB|ball-PA|ball-PB|nuclear histone kinase-1|nucleosomal histone kinase-1|triplet', map_location='97D3-97D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26864605, end_position_on_the_genomic_accession=26866807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43229, CurrentID=0, Status='live', Symbol='His2Av', Aliases='Dmel_CG5499, 5499, CG5499, Dmel\\CG5499, H2A, H2A.F/Z, H2A.V, H2A.X, H2A.Z, H2AV, H2AV_DROM, H2AX, H2AZ, H2Av, H2AvD, H2a.V, H2av, H2avD, HIS, HIS2AV, HIS2AVD, His, His2, His2A, His2AVD, HisH2Av, Hist, Hist2av, gamma-H2Av, gamma-HIS2AV, gamma-His2Av, gammaH2Av, h2AvD, his, his2Av, l(3)05146, l(3)810, l(3)97Dd, l(3)L1602, His2Av', description='Histone H2A variant', other_designations='CG5499-PA|CG5499-PB|H2A variant|His2Av-PA|His2Av-PB|Histone-2A|Histone-2av|histone|histone 2A variant|histone H2A variant|histone H2Av|histone variant H2Av|histone2A|histone2Av|histone2AvD', map_location='97D3-97D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26866929, end_position_on_the_genomic_accession=26869220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43230, CurrentID=0, Status='live', Symbol='SPARC', Aliases='Dmel_CG6378, BM-40 Sparc, BM-40-SPARC, BM-40/SPARC, BM40, CG6378, CT19876, Dmel\\CG6378, Sparc, anon-EST:Posey291, dSPARC, dsparc, sparc', description='Secreted protein, acidic, cysteine-rich', other_designations='BM-40-SPARC|CG6378-PA|CG6378-PB|SPARC-PA|SPARC-PB|osteonectin|secreted protein, acidic, cysteine-rich', map_location='97D3-97D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26869238, end_position_on_the_genomic_accession=26871995, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43231, CurrentID=0, Status='live', Symbol='Rb97D', Aliases='Dmel_CG6354, CG6354, Dmel\\CG6354, Hrb97D, R16, RBP97D1, ms(3)03445, rb97d, rbf, Rb97D', description='Ribonuclear protein at 97D', other_designations='CG6354-PA|CG6354-PB|CG6354-PC|CG6354-PD|CG6354-PE|CG6354-PF|CG6354-PG|CG6354-PH|CG6354-PI|Rb97D-PA|Rb97D-PB|Rb97D-PC|Rb97D-PD|Rb97D-PE|Rb97D-PF|Rb97D-PG|Rb97D-PH|Rb97D-PI|ribonuclear protein at 97D', map_location='97D3-97D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26872129, end_position_on_the_genomic_accession=26875742, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43232, CurrentID=0, Status='live', Symbol='ms(3)K81', Aliases='Dmel_CG14251, CG14251, Dmel\\CG14251, K81, ms(3)97Dg, ms(3)k81, ph', description='male sterile (3) K81', other_designations='CG14251-PA|ms(3)K81-PA|paternal haploid|paternal-haploid', map_location='3-91.3 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26875852, end_position_on_the_genomic_accession=26876562, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43233, CurrentID=0, Status='live', Symbol='CG5500', Aliases='Dmel_ Dmel\\CG5500', description='CG5500 gene product from transcript CG5500-RA', other_designations='CG5500-PA', map_location='97D4-97D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26876630, end_position_on_the_genomic_accession=26880459, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43234, CurrentID=0, Status='live', Symbol='ro', Aliases='Dmel_CG6348, CG6348, DROROUGH, Dmel\\CG6348, Ro', description='rough', other_designations='CG6348-PA|CG6348-PB|ro-PA|ro-PB', map_location='3-91.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26876771, end_position_on_the_genomic_accession=26881171, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43235, CurrentID=0, Status='live', Symbol='T48', Aliases='Dmel_CG5507, 48, CG5507, Dmel\\CG5507, Ubx-t48, anon-97D, t48', description='Transcript 48', other_designations="'48' transcript|CG5507-PA|CG5507-PB|T48-PA|T48-PB|transcript 48|transcript-48", map_location='97D5-97D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26881734, end_position_on_the_genomic_accession=26910997, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43236, CurrentID=0, Status='live', Symbol='Ets97D', Aliases='Dmel_CG6338, CG6338, D-Elg, D-elg, DELG, Delg, Dmel\\CG6338, Elg, Ets, d-elg, delg, elg, tne, tny', description='Ets at 97D', other_designations='CG6338-PA|CG6338-PB|Ets like gene|Ets97D-PA|Ets97D-PB|tiny|tiny eggs', map_location='3-91 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26910775, end_position_on_the_genomic_accession=26912886, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43238, CurrentID=0, Status='live', Symbol='CG14252', Aliases='Dmel_ Dmel\\CG14252', description='CG14252 gene product from transcript CG14252-RA', other_designations='CG14252-PA|uncharacterized protein', map_location='97D9-97D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26945534, end_position_on_the_genomic_accession=26947248, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43239, CurrentID=0, Status='live', Symbol='CG18472', Aliases='Dmel_ Dmel\\CG18472', description='CG18472 gene product from transcript CG18472-RA', other_designations='CG18472-PA|uncharacterized protein', map_location='97D9-97D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26947868, end_position_on_the_genomic_accession=26949682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43240, CurrentID=0, Status='live', Symbol='CG6330', Aliases='Dmel_ Dmel\\CG6330', description='CG6330 gene product from transcript CG6330-RB', other_designations='CG6330-PA|CG6330-PB', map_location='97D10-97D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26949692, end_position_on_the_genomic_accession=26958491, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43241, CurrentID=0, Status='live', Symbol='Tsp97E', Aliases='Dmel_CG6323, CG6323, Dm.Tsp97E, Dmel\\CG6323, Tspan31', description='Tetraspanin 97E', other_designations='CG6323-PA|CG6323-PD|CG6323-PE|CG6323-PG|CG6323-PH|Tsp97E-PA|Tsp97E-PD|Tsp97E-PE|Tsp97E-PG|Tsp97E-PH|tetraspanin 97E', map_location='97D11-97D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26961064, end_position_on_the_genomic_accession=26964817, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43242, CurrentID=0, Status='live', Symbol='CG5521', Aliases='Dmel_ Dmel\\CG5521, EP3527', description='CG5521 gene product from transcript CG5521-RB', other_designations='CG5521-PA|CG5521-PB', map_location='97D11-97D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26966930, end_position_on_the_genomic_accession=26977878, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43243, CurrentID=0, Status='live', Symbol='CG14253', Aliases='Dmel_ Dmel\\CG14253', description='CG14253 gene product from transcript CG14253-RB', other_designations='CG14253-PA|CG14253-PB|CG14253-PC|CG14253-PD|CG14253-PE|CG14253-PF|uncharacterized protein', map_location='97D12-97D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26978541, end_position_on_the_genomic_accession=26985706, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43244, CurrentID=0, Status='live', Symbol='CG31075', Aliases='Dmel_ 150697_at, CG6309, CT19744, Dmel\\CG31075', description='CG31075 gene product from transcript CG31075-RA', other_designations='CG31075-PA|CG31075-PB|uncharacterized protein', map_location='97D12-97D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26985624, end_position_on_the_genomic_accession=26988517, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43245, CurrentID=0, Status='live', Symbol='TwdlC', Aliases='Dmel_CG14254, CG14254, Dmel\\CG14254, TwdlB', description='TweedleC', other_designations='CG14254-PA|TwdlC-PA', map_location='97D12-97D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26994832, end_position_on_the_genomic_accession=26996442, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43246, CurrentID=0, Status='live', Symbol='CG6296', Aliases='Dmel_ DmelL7, Dmel\\CG6296', description='CG6296 gene product from transcript CG6296-RA', other_designations='CG6296-PA', map_location='97D12-97D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27003236, end_position_on_the_genomic_accession=27005469, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43247, CurrentID=0, Status='live', Symbol='CG6295', Aliases='Dmel_ DmelL6, Dmel\\CG6295', description='CG6295 gene product from transcript CG6295-RA', other_designations='CG6295-PA|triglyceride lipase', map_location='97D13-97D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27005899, end_position_on_the_genomic_accession=27007055, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43248, CurrentID=0, Status='live', Symbol='CG17192', Aliases='Dmel_ DmelL5, Dmel\\CG17192', description='CG17192 gene product from transcript CG17192-RA', other_designations='CG17192-PA|triglyceride lipase|uncharacterized protein', map_location='97D14-97D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27017180, end_position_on_the_genomic_accession=27018287, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43249, CurrentID=0, Status='live', Symbol='CG17191', Aliases='Dmel_ DmelL4, Dmel\\CG17191', description='CG17191 gene product from transcript CG17191-RA', other_designations='CG17191-PA|triglyceride lipase|uncharacterized protein', map_location='97D14-97D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27018608, end_position_on_the_genomic_accession=27019751, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43250, CurrentID=0, Status='live', Symbol='CG6283', Aliases='Dmel_ DmelL3, Dmel\\CG6283', description='CG6283 gene product from transcript CG6283-RA', other_designations='CG6283-PA|triglyceride lipase', map_location='97D14-97D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27020238, end_position_on_the_genomic_accession=27021378, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43251, CurrentID=0, Status='live', Symbol='CG6277', Aliases='Dmel_ DmelL2, Dmel\\CG6277', description='CG6277 gene product from transcript CG6277-RA', other_designations='CG6277-PA|triglyceride lipase', map_location='97D14-97D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27022071, end_position_on_the_genomic_accession=27023217, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43252, CurrentID=0, Status='live', Symbol='CG6271', Aliases='Dmel_ DmelL1, Dmel\\CG6271', description='CG6271 gene product from transcript CG6271-RA', other_designations='CG6271-PA|triglyceride lipase', map_location='97D14-97D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27024968, end_position_on_the_genomic_accession=27026109, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43253, CurrentID=0, Status='live', Symbol='RYa-R', Aliases='Dmel_CG5811, ATR, CG5811, CG5811/107368, Dm RYa-R, Dmel\\CG5811, NepYr, PR4, anon-WO0131005.13, anon-WO0131005.15', description='RYamide receptor', other_designations='CG5811-PA|CG5811-PB|CG5811-PC|RYa-R-PA|RYa-R-PB|RYa-R-PC|neuropeptide Y receptor-like', map_location='97D14-97E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26996980, end_position_on_the_genomic_accession=27057935, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43255, CurrentID=0, Status='live', Symbol='Nep5', Aliases='Dmel_CG6265, CG6265, DmeNEP5, Dmel\\CG6265, NEP5, nep5', description='Neprilysin 5', other_designations='CG6265-PC|CG6265-PD|CG6265-PE|CG6265-PF|Nep5-PC|Nep5-PD|Nep5-PE|Nep5-PF|neprilysin 5', map_location='97E1-97E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27057908, end_position_on_the_genomic_accession=27064199, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43256, CurrentID=0, Status='live', Symbol='spz', Aliases='Dmel_CG6134, CG6134, CT19282, Dmel\\CG6134, SPZ, Spz, Spz1, mel(3)7', description='spatzle', other_designations='CG6134-PA|CG6134-PB|CG6134-PC|CG6134-PD|CG6134-PI|CG6134-PJ|CG6134-PK|CG6134-PL|protein spaetzle precursor|spaetzle|spazle|spaztle|spz-PA|spz-PB|spz-PC|spz-PD|spz-PI|spz-PJ|spz-PK|spz-PL', map_location='3-92 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27064990, end_position_on_the_genomic_accession=27070070, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43257, CurrentID=0, Status='live', Symbol='CG14257', Aliases='Dmel_ Dmel\\CG14257', description='CG14257 gene product from transcript CG14257-RB', other_designations='CG14257-PB|CG14257-PC|uncharacterized protein', map_location='97E1-97E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27075870, end_position_on_the_genomic_accession=27082818, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43258, CurrentID=0, Status='live', Symbol='Cpr97Ea', Aliases='Dmel_CG6131, CG6131, Dmel\\CG6131', description='Cuticular protein 97Ea', other_designations='CG6131-PA|CG6131-PB|Cpr97Ea-PA|Cpr97Ea-PB|cuticular protein 97Ea', map_location='97E1-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27083130, end_position_on_the_genomic_accession=27087778, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43259, CurrentID=0, Status='live', Symbol='Cpr97Eb', Aliases='Dmel_CG15884, CG15884, CT33880, Dmel\\CG15884', description='Cuticular protein 97Eb', other_designations='CG15884-PA|Cpr97Eb-PA|cuticle protein-like|cuticular protein 97Eb', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27089704, end_position_on_the_genomic_accession=27091130, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43260, CurrentID=0, Status='live', Symbol='CG14258', Aliases='Dmel_ Dmel\\CG14258', description='CG14258 gene product from transcript CG14258-RA', other_designations='CG14258-PA|uncharacterized protein', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27093143, end_position_on_the_genomic_accession=27094224, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43261, CurrentID=0, Status='live', Symbol='CG14259', Aliases='Dmel_ Dmel\\CG14259', description='CG14259 gene product from transcript CG14259-RA', other_designations='CG14259-PA|uncharacterized protein', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27094533, end_position_on_the_genomic_accession=27095856, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43262, CurrentID=0, Status='live', Symbol='eater', Aliases='Dmel_CG6124, C901-like, CG6124, CT19227, Dmel\\CG6124, Eater', description='CG6124 gene product from transcript CG6124-RC', other_designations='CG6124-PC|eater-PC', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27095867, end_position_on_the_genomic_accession=27099679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43263, CurrentID=0, Status='live', Symbol='CG17189', Aliases='Dmel_ Dmel\\CG17189, cg17189', description='CG17189 gene product from transcript CG17189-RA', other_designations='CG17189-PA|uncharacterized protein', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27101406, end_position_on_the_genomic_accession=27102917, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43264, CurrentID=0, Status='live', Symbol='caix', Aliases='Dmel_CG6074, CAIX, CG6074, CT19053, Dmel\\CG6074', description='CG6074 gene product from transcript CG6074-RA', other_designations='CG6074-PA|caix-PA', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27102988, end_position_on_the_genomic_accession=27106184, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43265, CurrentID=0, Status='live', Symbol='gb', Aliases='Dmel_CG6070, CG6070, Dmel\\CG6070, Gb, anon-WO0118547.295', description='genderblind', other_designations='CG6070-PA|CG6070-PB|CG6070-PC|gb-PA|gb-PB|gb-PC', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27108467, end_position_on_the_genomic_accession=27112270, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43266, CurrentID=0, Status='live', Symbol='CG5815', Aliases='Dmel_ BcDNA:GH07089, BcDNA:GH07346, Dmel\\CG5815, Spt2', description='CG5815 gene product from transcript CG5815-RA', other_designations='CG5815-PA|CG5815-PC', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27112566, end_position_on_the_genomic_accession=27115126, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43267, CurrentID=0, Status='live', Symbol='CG6066', Aliases='Dmel_ Dmel\\CG6066', description='CG6066 gene product from transcript CG6066-RA', other_designations='CG6066-PA', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27115152, end_position_on_the_genomic_accession=27116707, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43268, CurrentID=0, Status='live', Symbol='CG5880', Aliases='Dmel_ Dmel\\CG5880', description='CG5880 gene product from transcript CG5880-RA', other_designations='CG5880-PA', map_location='97E2-97E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27116745, end_position_on_the_genomic_accession=27118532, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43269, CurrentID=0, Status='live', Symbol='CG5882', Aliases='Dmel_ Dmel\\CG5882', description='CG5882 gene product from transcript CG5882-RA', other_designations='CG5882-PA', map_location='97E2-97E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27118646, end_position_on_the_genomic_accession=27121590, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43270, CurrentID=0, Status='live', Symbol='CG6059', Aliases='Dmel_ Dmel\\CG6059', description='CG6059 gene product from transcript CG6059-RA', other_designations='CG6059-PA', map_location='97E3-97E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27121618, end_position_on_the_genomic_accession=27124688, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43271, CurrentID=0, Status='live', Symbol='CG6051', Aliases='Dmel_ Dmel\\CG6051, blr2', description='CG6051 gene product from transcript CG6051-RD', other_designations='CG6051-PB|CG6051-PC|CG6051-PD|blade runner 2', map_location='97E3-97E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27124759, end_position_on_the_genomic_accession=27151069, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43273, CurrentID=0, Status='live', Symbol='grass', Aliases='Dmel_CG5896, CG5896, Dmel\\CG5896, Grass, SP1, SP2, c-SP1', description='Gram-positive Specific Serine protease', other_designations='CG5896-PA|CG5896-PB|Gram-positive specific serine protease|Gram-positive-specific serine protease|gram-positive specific serine protease|grass-PA|grass-PB', map_location='97E5-97E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27157948, end_position_on_the_genomic_accession=27159781, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43274, CurrentID=0, Status='live', Symbol='CG5909', Aliases='Dmel_ Dmel\\CG5909, SP8, c-SP8', description='CG5909 gene product from transcript CG5909-RA', other_designations='CG5909-PA', map_location='97E5-97E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27161158, end_position_on_the_genomic_accession=27162911, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43275, CurrentID=0, Status='live', Symbol='Ser', Aliases='Dmel_CG6127, Bd, CG6127, CT18858, Dmel\\CG6127, Rpw, SER, mel(3)8, ser, std', description='Serrate', other_designations='CG6127-PA|CG6127-PB|Ser-PA|Ser-PB|beaded|dmSerrate|ripped wing|serrate|serratoid', map_location='97E6-97E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27172090, end_position_on_the_genomic_accession=27194267, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  43276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43276, CurrentID=0, Status='live', Symbol='CG42813', Aliases='Dmel_ CG31063, CG6001, Dmel\\CG42813, Dmel_CG31063, Nudix', description='CG42813 gene product from transcript CG42813-RB', other_designations='CG42813-PA|CG42813-PB|uncharacterized protein', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27227070, end_position_on_the_genomic_accession=27228561, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43277, CurrentID=0, Status='live', Symbol='CG31064', Aliases='Dmel_ CG14261, CG6001, Dmel\\CG31064', description='CG31064 gene product from transcript CG31064-RB', other_designations='CG31064-PA|CG31064-PB|CG31064-PE|CG31064-PF|CG31064-PG|CG31064-PH|uncharacterized protein', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27229505, end_position_on_the_genomic_accession=27239085, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43278, CurrentID=0, Status='live', Symbol='DNApol-alpha73', Aliases='Dmel_CG5923, CG5923, DNA pol-73, DNA polalpha, DNA polymerase alpha 73 kD, DNApol-alpha, DNApol-alpha72, DNApolalpha, Dmel\\CG5923, Dmpol alpha, POLA, Pol, Polalpha, Polalpha 73kDa, pol alpha, polalpha', description='DNA polymerase alpha 73kD', other_designations='B subunit|CG5923-PC|DNA polymerase 73K|DNA polymerase alpha|DNA polymerase alpha primase 73 kda subunit|DNA-polymerase alpha|DNApol-alpha73-PC', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27239163, end_position_on_the_genomic_accession=27241228, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43279, CurrentID=0, Status='live', Symbol='CG17991', Aliases='Dmel_ Dmel\\CG17991', description='CG17991 gene product from transcript CG17991-RA', other_designations='CG17991-PA|uncharacterized protein', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27241267, end_position_on_the_genomic_accession=27242330, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43281, CurrentID=0, Status='live', Symbol='CG5984', Aliases='Dmel_ Dmel\\CG5984', description='CG5984 gene product from transcript CG5984-RA', other_designations='CG5984-PA|Filamin-like', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27246403, end_position_on_the_genomic_accession=27247992, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43282, CurrentID=0, Status='live', Symbol='CG5987', Aliases='Dmel_ Dmel\\CG5987', description='CG5987 gene product from transcript CG5987-RA', other_designations='CG5987-PA', map_location='97E11-97E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27248234, end_position_on_the_genomic_accession=27251007, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43283, CurrentID=0, Status='live', Symbol='pll', Aliases='Dmel_CG5974, CG5974, Dmel\\CG5974, PELLE/IL-1, p11/IRAK1, pll', description='pelle', other_designations='CG5974-PA|CG5974-PB|Serine/threonine-protein kinase pelle|pll-PA|pll-PB', map_location='3-92 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27251131, end_position_on_the_genomic_accession=27253182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43284, CurrentID=0, Status='live', Symbol='TfIIA-L', Aliases='Dmel_CG5930, CG5930, CT18631, Dmel\\CG5930, TFIIA, TFIIA-L, TFIIAL, dTFIIA, tfiia-l', description='Transcription factor IIA L', other_designations='CG5930-PA|CG5930-PB|CG5930-PC|TfIIA-L-PA|TfIIA-L-PB|TfIIA-L-PC|transcription factor IIA L', map_location='97E11-97E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27251112, end_position_on_the_genomic_accession=27256463, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43286, CurrentID=0, Status='live', Symbol='CG14262', Aliases='Dmel_ Dmel\\CG14262', description='CG14262 gene product from transcript CG14262-RA', other_designations='CG14262-PA|uncharacterized protein', map_location='97E11-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27264374, end_position_on_the_genomic_accession=27265430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43287, CurrentID=0, Status='live', Symbol='CG14260', Aliases='Dmel_ Dmel\\CG14260', description='CG14260 gene product from transcript CG14260-RA', other_designations='CG14260-PA|CG14260-PB|uncharacterized protein', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27265457, end_position_on_the_genomic_accession=27266423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43288, CurrentID=0, Status='live', Symbol='l(3)mbt', Aliases='Dmel_CG5954, CG5954, Dmel\\CG5954, L(3)MBT, L(3)MBT/LIN-61, L(3)mbt, L3mbt, MBT, dL(3)mbt, dl(3)MBT, l(3)malignant braintumor, mbt', description='lethal (3) malignant brain tumor', other_designations='CG5954-PB|l(3)mbt-PB|tumor-supressor', map_location='3-93.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27266326, end_position_on_the_genomic_accession=27273273, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43289, CurrentID=0, Status='live', Symbol='CG5934', Aliases='Dmel_ Dmel\\CG5934', description='CG5934 gene product from transcript CG5934-RA', other_designations='CG5934-PA', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27273705, end_position_on_the_genomic_accession=27274565, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43290, CurrentID=0, Status='live', Symbol='CG5938', Aliases='Dmel_ BEST:LD40095, Dmel\\CG5938', description='CG5938 gene product from transcript CG5938-RD', other_designations='CG5938-PA|CG5938-PB|CG5938-PC|CG5938-PD|CG5938-PF', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27274794, end_position_on_the_genomic_accession=27277418, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43291, CurrentID=0, Status='live', Symbol='mrt', Aliases='Dmel_CG3361, CG3361, Dmel\\CG3361', description='martik', other_designations='CG3361-PA|CG3361-PB|CG3361-PC|CG3361-PD|CG3361-PE|CG3361-PF|mrt-PA|mrt-PB|mrt-PC|mrt-PD|mrt-PE|mrt-PF', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27277249, end_position_on_the_genomic_accession=27282661, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43292, CurrentID=0, Status='live', Symbol='CG3368', Aliases='Dmel_ Dmel\\CG3368, SD09067p', description='CG3368 gene product from transcript CG3368-RA', other_designations='CG3368-PA', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27283780, end_position_on_the_genomic_accession=27287448, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43293, CurrentID=0, Status='live', Symbol='bigmax', Aliases='Dmel_CG3350, BIGMAX, Bigmax, CG3350, DmMlx, Dmel\\CG3350, Mlx, bHLHd13, cg3350, mlx', description='CG3350 gene product from transcript CG3350-RA', other_designations='CG3350-PA|big max|bigmax-PA', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27287366, end_position_on_the_genomic_accession=27288526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43294, CurrentID=0, Status='live', Symbol='Hmu', Aliases='Dmel_CG3373, CG3373, Dmel\\CG3373, RRM5, Rbp5, SPH210, rrm5', description='Hemomucin', other_designations='CG3373-PA|Hmu-PA|RNA-binding protein 5|hemomucin', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27288769, end_position_on_the_genomic_accession=27291592, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43295, CurrentID=0, Status='live', Symbol='CG3339', Aliases='Dmel_ Dmel\\CG3339', description='CG3339 gene product from transcript CG3339-RC', other_designations='CG3339-PC', map_location='97F1-97F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27294286, end_position_on_the_genomic_accession=27322325, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  43296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43296, CurrentID=0, Status='live', Symbol='CG3330', Aliases='Dmel_ Dmel\\CG3330, anon-WO0140519.147', description='CG3330 gene product from transcript CG3330-RA', other_designations='CG3330-PA', map_location='97F3-97F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27329693, end_position_on_the_genomic_accession=27330882, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19836004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836004, CurrentID=0, Status='live', Symbol='CG44574', Aliases='Dmel_ Dmel\\CG44574', description='CG44574 gene product from transcript CG44574-RA', other_designations='CG44574-PA|uncharacterized protein', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5883741, end_position_on_the_genomic_accession=5884345, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43298, CurrentID=0, Status='live', Symbol='CG42534', Aliases='Dmel_ CG13980, CG14264, CG6599, Dmel\\CG42534, Dmel_CG13980, Dmel_CG14264, Dmel_CG6599', description='CG42534 gene product from transcript CG42534-RC', other_designations='CG42534-PC|CG42534-PD|uncharacterized protein', map_location='97F6-97F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27340622, end_position_on_the_genomic_accession=27372574, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43300, CurrentID=0, Status='live', Symbol='side', Aliases='Dmel_CG31062, CG12510, CG13973, CG13974, CG13975, CG13979, CG31062, CT33527, CT33528, Dmel\\CG31062, Doe, Side, doe', description='sidestep', other_designations='CG31062-PA|CG31062-PB|CG31062-PC|CG31062-PD|downstream of eyeless|side-PA|side-PB|side-PC|side-PD', map_location='97F6-98A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27375088, end_position_on_the_genomic_accession=27437482, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  43303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43303, CurrentID=0, Status='live', Symbol='CG13978', Aliases='Dmel_ Dmel\\CG13978', description='CG13978 gene product from transcript CG13978-RA', other_designations='CG13978-PA|uncharacterized protein', map_location='98A1-98A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27464189, end_position_on_the_genomic_accession=27467080, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43304, CurrentID=0, Status='live', Symbol='Cyp6a18', Aliases='Dmel_CG13977, 6a18, CG13977, Dmel\\CG13977', description='CG13977 gene product from transcript CG13977-RB', other_designations='CG13977-PA|CG13977-PB|Cyp6a18-PA|Cyp6a18-PB|cytochrome P450', map_location='98A2-98A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27478675, end_position_on_the_genomic_accession=27482904, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43305, CurrentID=0, Status='live', Symbol='Gr98a', Aliases='Dmel_CG13976, 98B.4, CG13976, Dmel\\CG13976, GR98B.4, GR98a, Gr98B4', description='Gustatory receptor 98a', other_designations='CG13976-PA|Gr98a-PA|gustatory receptor 98B.4|gustatory receptor 98a', map_location='98A2-98A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27487927, end_position_on_the_genomic_accession=27489968, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43306 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43306, CurrentID=0, Status='live', Symbol='CG5402', Aliases='Dmel_ Dmel\\CG5402, anon-WO0140519.87', description='CG5402 gene product from transcript CG5402-RA', other_designations='CG5402-PA', map_location='98A3-98A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27502090, end_position_on_the_genomic_accession=27502956, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43307, CurrentID=0, Status='live', Symbol='CG13972', Aliases='Dmel_ Dmel\\CG13972', description='CG13972 gene product from transcript CG13972-RA', other_designations='CG13972-PA|uncharacterized protein', map_location='98A3-98A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27515911, end_position_on_the_genomic_accession=27517257, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43308, CurrentID=0, Status='live', Symbol='CG12885', Aliases='Dmel_ Dmel\\CG12885', description='CG12885 gene product from transcript CG12885-RB', other_designations='CG12885-PA|CG12885-PB|uncharacterized protein', map_location='98A4-98A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27520866, end_position_on_the_genomic_accession=27533934, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43309, CurrentID=0, Status='live', Symbol='Gr98d', Aliases='Dmel_CG31061, 98B.1, CG12884, CG31061, Dmel\\CG31061, GR98B.1, GR98d, Gr98A1, Gr98B1', description='Gustatory receptor 98d', other_designations='CG31061-PA|Gr98d-PA|gustatory receptor 98B.1|gustatory receptor 98d', map_location='98A4-98A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27546052, end_position_on_the_genomic_accession=27547462, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43310, CurrentID=0, Status='live', Symbol='Klp98A', Aliases='Dmel_CG5658, CG5658, DmKlp98A, DmSNX23, Dmel\\CG5658, KIF 16A, KIF16A/B, KLP6, KLP98A, Klp6, Klp98a', description='Kinesin-like protein at 98A', other_designations='CG5658-PA|CG5658-PB|Klp98A-PA|Klp98A-PB', map_location='98A4-98A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27547775, end_position_on_the_genomic_accession=27559145, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  43311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43311, CurrentID=0, Status='live', Symbol='CG5646', Aliases='Dmel_ Dmel\\CG5646', description='CG5646 gene product from transcript CG5646-RA', other_designations='CG5646-PA', map_location='98A6-98A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27559379, end_position_on_the_genomic_accession=27561319, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43312, CurrentID=0, Status='live', Symbol='wdb', Aliases="Dmel_CG5643, 0318/07, B56-2, BEST:LD02456, BcDNA:LD34343, CG5643, Dmel\\CG5643, EP3559, LD02456, PP2-AB, PP2A, PP2A[B'-2], PP2a, WDB, Wdb, dB56-2, dPP2A, dPP2A-B56-2, l(3)S031807", description='widerborst', other_designations='CG5643-PA|CG5643-PB|CG5643-PC|CG5643-PD|CG5643-PE|CG5643-PF|PP2A B subunit|PP2A regulatory subunit|protein phosphatase 2a|protein phosphatase PP2A|wdb-PA|wdb-PB|wdb-PC|wdb-PD|wdb-PE|wdb-PF|wiederborst', map_location='98A6-98A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27561675, end_position_on_the_genomic_accession=27582241, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43314, CurrentID=0, Status='live', Symbol='CG5639', Aliases='Dmel_ CT17816, Dmel\\CG5639', description='CG5639 gene product from transcript CG5639-RA', other_designations='CG5639-PA|N-like', map_location='98A8-98A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27586474, end_position_on_the_genomic_accession=27596130, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43315, CurrentID=0, Status='live', Symbol='dsd', Aliases='Dmel_CG5634, CG5634, CT17798, Dmel\\CG5634', description='distracted', other_designations='CG5634-PB|CG5634-PC|attractin|dsd-PB|dsd-PC', map_location='98A10-98A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27597563, end_position_on_the_genomic_accession=27607396, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43316, CurrentID=0, Status='live', Symbol='CG5612', Aliases='Dmel_ Dmel\\CG5612', description='CG5612 gene product from transcript CG5612-RA', other_designations='CG5612-PA', map_location='98A11-98A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27623359, end_position_on_the_genomic_accession=27626011, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43317, CurrentID=0, Status='live', Symbol='Tusp', Aliases='Dmel_CG5586, CG5586, DMG2-MB6.chr3R.24.043, Dmel\\CG5586, TUSP', description='CG5586 gene product from transcript CG5586-RE', other_designations='CG5586-PD|CG5586-PE|CG5586-PF|CG5586-PG|Tusp-PD|Tusp-PE|Tusp-PF|Tusp-PG|dmWSBa Tubby-like|tusp', map_location='98A12-98A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27607879, end_position_on_the_genomic_accession=27634932, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  43318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43318, CurrentID=0, Status='live', Symbol='CG5611', Aliases='Dmel_ Dmel\\CG5611', description='CG5611 gene product from transcript CG5611-RC', other_designations='CG5611-PA|CG5611-PC', map_location='98A12-98A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27634625, end_position_on_the_genomic_accession=27636060, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43319, CurrentID=0, Status='live', Symbol='Mtl', Aliases='Dmel_CG5588, 5588, CG5588, Dmel\\CG5588, MTL, Mt1/Mig2-like, Rac, Rac GTPase, Rac-like, dMig-2, mtl', description='Mig-2-like', other_designations='CG5588-PA|CG5588-PB|CG5588-PC|Mtl(Rac-like)|Mtl-PA|Mtl-PB|Mtl-PC|Rac|Rac GTPase|mig-2-like', map_location='98A13-98A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27636412, end_position_on_the_genomic_accession=27639871, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43321, CurrentID=0, Status='live', Symbol='RpS10a', Aliases='Dmel_CG12275, BcDNA:LD32148, CG12275, Dmel\\CG12275, RpS10, S10', description='Ribosomal protein S10a', other_designations='CG12275-PA|RpS10a-PA|ribosomal protein S10a', map_location='98A14-98A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27652760, end_position_on_the_genomic_accession=27653363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43323, CurrentID=0, Status='live', Symbol='Mlc1', Aliases='Dmel_CG5596, CG5596, DmMLC1, Dmel\\CG5596, ELC, FBgn0002772, LC1, MLC, MLC-ALK, MLC1, Mlc-1, MlcI, Mlci, chr3R:23484557..23484663, mlc, mlc-alk, mlc1, mlcl', description='Myosin alkali light chain 1', other_designations='CG5596-PA|CG5596-PB|CG5596-PD|Mlc1-PA|Mlc1-PB|Mlc1-PD|myosin|myosin 3R alkali light chain l|myosin alkali light chain 1|myosin alkali light chain l|myosin light Chain-1|myosin light chain|myosin light chain alkali', map_location='98A14-98A15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27657078, end_position_on_the_genomic_accession=27660177, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43324, CurrentID=0, Status='live', Symbol='unc80', Aliases='Dmel_CG18437, CG18437, Dmel\\CG18437, UNC80', description='CG18437 gene product from transcript CG18437-RF', other_designations='CG18437-PC|CG18437-PD|CG18437-PE|CG18437-PF|CG18437-PG|unc80-PC|unc80-PD|unc80-PE|unc80-PF|unc80-PG', map_location='98A15-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27662439, end_position_on_the_genomic_accession=27677099, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  43325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43325, CurrentID=0, Status='live', Symbol='CG5590', Aliases='Dmel_ Dmel\\CG5590', description='CG5590 gene product from transcript CG5590-RA', other_designations='CG5590-PA', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27677522, end_position_on_the_genomic_accession=27679013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43326, CurrentID=0, Status='live', Symbol='CG12883', Aliases='Dmel_ Dmel\\CG12883', description='CG12883 gene product from transcript CG12883-RA', other_designations='CG12883-PA|uncharacterized protein', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27679291, end_position_on_the_genomic_accession=27680336, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43327, CurrentID=0, Status='live', Symbol='Sce', Aliases='Dmel_CG5595, CG5595, DRING, Ding, Dmel\\CG5595, Dring, RING, RING1, Ring, Ring/Sce, Rnf2, SCE/dRing, dRING, dRING1, dRing, dRing1, dring, sce, Sce', description='Sex combs extra', other_designations='CG5595-PA|Sce-PA|Sex comb extra|dRING1|ring', map_location='3-92 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27680431, end_position_on_the_genomic_accession=27682029, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43328, CurrentID=0, Status='live', Symbol='CG12880', Aliases='Dmel_ BEST:LD29228, Dmel\\CG12880', description='CG12880 gene product from transcript CG12880-RA', other_designations='CG12880-PA|uncharacterized protein', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27685962, end_position_on_the_genomic_accession=27692790, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43329, CurrentID=0, Status='live', Symbol='CG43320', Aliases='Dmel_ CG12879, Dmel\\CG43320', description='CG43320 gene product from transcript CG43320-RA', other_designations='CG43320-PA|uncharacterized protein', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27695235, end_position_on_the_genomic_accession=27696531, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43330, CurrentID=0, Status='live', Symbol='ALiX', Aliases='Dmel_CG12876, AIP1, ALIX, Aip1, Alix, CG12876, DAIP1, Dmel\\CG12876, Snf7, alix', description='ALG-2 interacting protein X', other_designations='ALG-2 interacting protein 1|ALiX-PA|CG12876-PA|apoptosis linked gene 2 interacting X', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27696466, end_position_on_the_genomic_accession=27700110, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43331, CurrentID=0, Status='live', Symbol='btz', Aliases='Dmel_CG12878, Btz, CG12878, Dmel\\CG12878, wkl', description='barentsz', other_designations='CG12878-PA|CG12878-PB|CG12878-PC|CG12878-PD|Dm SELOR|barentz|btz-PA|btz-PB|btz-PC|btz-PD|weak localizer', map_location='98B1-98B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27700361, end_position_on_the_genomic_accession=27703445, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43332, CurrentID=0, Status='live', Symbol='CG12428', Aliases='Dmel_ Dmel\\CG12428', description='CG12428 gene product from transcript CG12428-RB', other_designations='CG12428-PA|CG12428-PB|CG12428-PD|CG12428-PE|carnitine palmitoyltransferase|uncharacterized protein', map_location='98B2-98B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27703379, end_position_on_the_genomic_accession=27706751, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43333 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43333, CurrentID=0, Status='live', Symbol='CG12877', Aliases='Dmel_ BcDNA:LD30051, Dmel\\CG12877', description='CG12877 gene product from transcript CG12877-RC', other_designations='CG12877-PC|uncharacterized protein', map_location='98B2-98B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27707100, end_position_on_the_genomic_accession=27710857, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43334, CurrentID=0, Status='live', Symbol='Ets98B', Aliases='Dmel_CG5583, CG5583, D-Ets-4, D-ets-4, D-ets4, Dmel\\CG5583, ETS4, Ets, Ets4, ets98B', description='Ets at 98B', other_designations='CG5583-PA|CG5583-PB|Ets98B-PA|Ets98B-PB', map_location='98B2-98B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27736538, end_position_on_the_genomic_accession=27748366, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19835207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835207, CurrentID=0, Status='live', Symbol='CG45095', Aliases='Dmel_ Dmel\\CG45095', description='CG45095 gene product from transcript CG45095-RA', other_designations='CG45095-PA|uncharacterized protein', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25102553, end_position_on_the_genomic_accession=25103617, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43341, CurrentID=0, Status='live', Symbol='Or98a', Aliases='Dmel_CG5540, 98B.1, 98a, CG5540, DOR110, DmOr98a, Dmel\\CG5540, Or98A', description='Odorant receptor 98a', other_designations='CG5540-PA|Or98a-PA|odorant receptor 98a', map_location='98B4-98B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27845950, end_position_on_the_genomic_accession=27847260, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43342, CurrentID=0, Status='live', Symbol='CG12426', Aliases='Dmel_ Dmel\\CG12426', description='CG12426 gene product from transcript CG12426-RB', other_designations='CG12426-PB|uncharacterized protein', map_location='98B5-98B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27871501, end_position_on_the_genomic_accession=27872390, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43343, CurrentID=0, Status='live', Symbol='mil', Aliases='Dmel_CG5017, CG5017, Dmel\\CG5017, hanabi', description='milkah', other_designations='CG5017-PA|hanabi|mil-PA|milkah-1', map_location='98B5-98B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27912447, end_position_on_the_genomic_accession=27913785, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43344, CurrentID=0, Status='live', Symbol='CG5003', Aliases='Dmel_ Dmel\\CG5003, FBXL8', description='CG5003 gene product from transcript CG5003-RA', other_designations='CG5003-PA', map_location='98B5-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27913777, end_position_on_the_genomic_accession=27916516, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43345, CurrentID=0, Status='live', Symbol='mRpS22', Aliases='Dmel_CG12261, AAF56757, CG12261, Dmel\\CG12261, MRP-S22, MrpS22', description='mitochondrial ribosomal protein S22', other_designations='CG12261-PA|mRpS22-PA', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27916908, end_position_on_the_genomic_accession=27918248, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43347, CurrentID=0, Status='live', Symbol='CG12259', Aliases='Dmel_ Dmel\\CG12259', description='CG12259 gene product from transcript CG12259-RA', other_designations='CG12259-PA|uncharacterized protein', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27919844, end_position_on_the_genomic_accession=27921089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43348, CurrentID=0, Status='live', Symbol='BCAS2', Aliases='Dmel_CG4980, CG4980, Dmel\\CG4980, dBCAS2', description='BCAS2 ortholog', other_designations='BCAS2-PA|CG4980-PA', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27921283, end_position_on_the_genomic_accession=27922356, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43349, CurrentID=0, Status='live', Symbol='RpL4', Aliases='Dmel_CG5502, CG5502, Dmel\\CG5502, L4, Rp L4, RpL1, anon-WO0172774.76, chr3R:23749174..23749334, rpL1', description='Ribosomal protein L4', other_designations='CG5502-PA|RpL4-PA|ribosomal protein L1|ribosomal protein L4', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27922821, end_position_on_the_genomic_accession=27924888, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43350, CurrentID=0, Status='live', Symbol='mino', Aliases='Dmel_CG5508, 152088_at, BcDNA:GH07066, CG5508, Dmel\\CG5508, GPAT1', description='minotaur', other_designations='CG5508-PA|CG5508-PB|CG5508-PC|CG5508-PE|mino-PA|mino-PB|mino-PC|mino-PE', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27925191, end_position_on_the_genomic_accession=27933077, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43351, CurrentID=0, Status='live', Symbol='Mes-4', Aliases='Dmel_CG4976, CG4976, Dmel\\CG4976, MES-4, dMES-4, d', description='CG4976 gene product from transcript CG4976-RB', other_designations='CG4976-PA|CG4976-PB|Mes-4-PA|Mes-4-PB', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27932982, end_position_on_the_genomic_accession=27937853, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43352, CurrentID=0, Status='live', Symbol='CG5514', Aliases='Dmel_ Dmel\\CG5514', description='CG5514 gene product from transcript CG5514-RB', other_designations='CG5514-PA|CG5514-PB|CG5514-PC|CG5514-PD|CG5514-PE', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27938245, end_position_on_the_genomic_accession=27942778, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43353, CurrentID=0, Status='live', Symbol='mfrn', Aliases='Dmel_CG4963, CG4963, Dmel\\CG4963, HD35059, Mfrn, anon-EST:fe2H5, dmfrn', description='mitoferrin', other_designations='CG4963-PA|anon-fast-evolving-2H5|mfrn-PA', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27942796, end_position_on_the_genomic_accession=27948021, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43354 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43354, CurrentID=0, Status='live', Symbol='Gp93', Aliases='Dmel_CG5520, CG5520, CT17486, Dmel\\CG5520, GP93, Hsp90, dGRP94, gp93, p93', description='Glycoprotein 93', other_designations='CG5520-PA|Gp93-PA|glycoprotein 93|glycoprotein93', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27948711, end_position_on_the_genomic_accession=27952151, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43355, CurrentID=0, Status='live', Symbol='CG4951', Aliases='Dmel_ Dmel\\CG4951', description='CG4951 gene product from transcript CG4951-RB', other_designations='CG4951-PA|CG4951-PB', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27952051, end_position_on_the_genomic_accession=27954112, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43356, CurrentID=0, Status='live', Symbol='CG5527', Aliases='Dmel_ Dmel\\CG5527', description='CG5527 gene product from transcript CG5527-RA', other_designations='CG5527-PA', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27954917, end_position_on_the_genomic_accession=27957326, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43357, CurrentID=0, Status='live', Symbol='CG4884', Aliases='Dmel_ Dmel\\CG4884', description='CG4884 gene product from transcript CG4884-RB', other_designations='CG4884-PA|CG4884-PB', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27957356, end_position_on_the_genomic_accession=27959245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43358, CurrentID=0, Status='live', Symbol='CG4849', Aliases='Dmel_ Dmel\\CG4849, EFTUD2, Snu114, eEF2, eftud2', description='CG4849 gene product from transcript CG4849-RA', other_designations='CG4849-PA', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27963415, end_position_on_the_genomic_accession=27966958, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43359, CurrentID=0, Status='live', Symbol='betaTub97EF', Aliases='Dmel_CG4869, 4t, B4t, BETA 98B, CG4869, DTB1, Dmbeta4, Dmel\\CG4869, T, Tub, beta-Tub, beta-Tub97EF, beta-tub, beta4, beta4t, beta97F, betaTub4, betaTub98BC, betaTub98C', description='beta-Tubulin at 97EF', other_designations='Beta4-tubulin|CG4869-PB|CG4869-PC|beta tubulin|beta-Tubulin|beta-Tubulin 97EF|beta4 tubulin|betaTub97EF-PB|betaTub97EF-PC|betatubulin|tubulin', map_location='98B6-98B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27967008, end_position_on_the_genomic_accession=27988384, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43360, CurrentID=0, Status='live', Symbol='CG4815', Aliases='Dmel_ Dmel\\CG4815, SP189', description='CG4815 gene product from transcript CG4815-RA', other_designations='CG4815-PA', map_location='98B7-98B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27989699, end_position_on_the_genomic_accession=27990751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43362, CurrentID=0, Status='live', Symbol='CG34362', Aliases='Dmel_ BcDNA:RE10833, CG12870, CG12872, Dmel\\CG34362, Dmel_CG12870, Dmel_CG12872, Tia1-like', description='CG34362 gene product from transcript CG34362-RI', other_designations='CG34362-PA|CG34362-PE|CG34362-PF|CG34362-PI|CG34362-PJ|uncharacterized protein', map_location='98B8-98C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28026615, end_position_on_the_genomic_accession=28090740, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19835236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835236, CurrentID=0, Status='live', Symbol='CG45154', Aliases='Dmel_ CR45154, Dmel\\CG45154', description='CG45154 gene product from transcript CG45154-RB', other_designations='CG45154-PB|uncharacterized protein', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21146281, end_position_on_the_genomic_accession=21147041, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43367, CurrentID=0, Status='live', Symbol='Mst98Cb', Aliases='Dmel_CG18396, CG18396, Dmel\\CG18396, MST98Cb, Mst, Mst98C, anon-WO0140519.46, mst(3)gl-13', description='Male-specific RNA 98Cb', other_designations='CG18396-PA|Mst98Cb-PA', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28292798, end_position_on_the_genomic_accession=28293863, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43369, CurrentID=0, Status='live', Symbol='UQCR-14L', Aliases='Dmel_CG17856, CG17856, Dmel\\CG17856, UCR6', description='Ubiquinol-cytochrome c reductase 14 kDa subunit-like', other_designations='CG17856-PA|UQCR-14L-PA|Ubiquinol-cytochrome C reductase complex 14 kDa protein|Ubiquinol-cytochrome c reductase 14 kDa protein|complex III', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28294048, end_position_on_the_genomic_accession=28294550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43370, CurrentID=0, Status='live', Symbol='CG12516', Aliases='Dmel_ Dmel\\CG12516', description='CG12516 gene product from transcript CG12516-RB', other_designations='CG12516-PB|uncharacterized protein', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28303631, end_position_on_the_genomic_accession=28304605, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43373, CurrentID=0, Status='live', Symbol='Gfat2', Aliases='Dmel_CG1345, CG1345, Dmel/GFAT2, Dmel\\CG1345', description='Glutamine:fructose-6-phosphate aminotransferase 2', other_designations='CG1345-PA|Gfat2-PA', map_location='98C4-98C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28337019, end_position_on_the_genomic_accession=28339765, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43374, CurrentID=0, Status='live', Symbol='Moca-cyp', Aliases='Dmel_CG1866, BEST:LD38313, CG1866, Dmel\\CG1866, anon-WO0147981.13, anon-WO0147981.4, anon-WO0147981.5, moca', description='CG1866 gene product from transcript CG1866-RC', other_designations='CG1866-PA|CG1866-PB|CG1866-PC|Moca-cyp-PA|Moca-cyp-PB|Moca-cyp-PC', map_location='98C4-98C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28340334, end_position_on_the_genomic_accession=28345169, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43375, CurrentID=0, Status='live', Symbol='Or98b', Aliases='Dmel_CG1867, 98C.1, 98b, CG1867, DOR121, Dmel\\CG1867P, or98b, Or98b', description='Odorant receptor 98b', other_designations='CG1867-PB|Or98b-PB|odorant receptor 98b', map_location='98C4-98C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28352242, end_position_on_the_genomic_accession=28353620, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43377, CurrentID=0, Status='live', Symbol='beat-VI', Aliases='Dmel_CG14064, Beat-VI, CG12089, CG14064, Dmel\\CG14064, beat VI', description='CG14064 gene product from transcript CG14064-RB', other_designations='CG14064-PA|CG14064-PB|beat-VI-PA|beat-VI-PB', map_location='98D1-98D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28418208, end_position_on_the_genomic_accession=28474814, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43378, CurrentID=0, Status='live', Symbol='CG1894', Aliases='Dmel_ Dmel\\CG1894, dmHAG407', description='CG1894 gene product from transcript CG1894-RA', other_designations='CG1894-PA', map_location='98D2-98D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28484557, end_position_on_the_genomic_accession=28486062, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43379, CurrentID=0, Status='live', Symbol='Dhc98D', Aliases='Dmel_CG1842, CG1842, DHC, Dhc, Dmel\\CG1842, dhc', description='Dynein heavy chain at 89D', other_designations='CG1842-PB|CG1842-PC|CG1842-PD|Dhc98D-PB|Dhc98D-PC|Dhc98D-PD|dynein heavy chain at 89D|dynein-related heavy chain polypeptide', map_location='98D2-98D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28486953, end_position_on_the_genomic_accession=28519398, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  43382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43382, CurrentID=0, Status='live', Symbol='CG12413', Aliases='Dmel_ Dmel\\CG12413', description='CG12413 gene product from transcript CG12413-RA', other_designations='CG12413-PA|uncharacterized protein', map_location='98D4-98D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28542274, end_position_on_the_genomic_accession=28544817, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43383, CurrentID=0, Status='live', Symbol='fkh', Aliases='Dmel_CG10002, CG10002, Dmel\\CG10002, Dmfkh, FKH, Fkh, Sebp2, dFoxA', description='fork head', other_designations='CG10002-PA|CG10002-PB|CG10002-PC|CG10002-PD|CG10002-PE|fkh-PA|fkh-PB|fkh-PC|fkh-PD|fkh-PE|forkhead|secretion enhancer-binding protein 2', map_location='3-95 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28580976, end_position_on_the_genomic_accession=28585264, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43384, CurrentID=0, Status='live', Symbol='Noa36', Aliases='Dmel_CG10009, CG10009, Dmel\\CG10009, NOA 36, NOA36, noa36', description='CG10009 gene product from transcript CG10009-RA', other_designations='CG10009-PA|Noa36-PA', map_location='98D6-98D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28598059, end_position_on_the_genomic_accession=28599288, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43385, CurrentID=0, Status='live', Symbol='Hrb98DE', Aliases='Dmel_CG9983, CG9983, Dmel\\CG9983, E7Delta6, HDP(P9), HRB98DE, HRB98DE/hrp38, Hrb1, Hrb98 de/Hrp38, Hrp38, Pen9, anon-E7Delta6, hnRNP A2/B1, hrb98 de, hrb98DE, hrp38, p9, Hrb98DE', description='Heterogeneous nuclear ribonucleoprotein at 98DE', other_designations='CG9983-PA|CG9983-PB|CG9983-PC|CG9983-PD|CG9983-PE|CG9983-PF|Hrb98DE-PA|Hrb98DE-PB|Hrb98DE-PC|Hrb98DE-PD|Hrb98DE-PE|Hrb98DE-PF|heterogeneous nuclear ribonucleoprotein at 98DE|hnRNA-binding protein 1|s', map_location='98D6-98D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28599834, end_position_on_the_genomic_accession=28605777, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43386, CurrentID=0, Status='live', Symbol='CG9986', Aliases='Dmel_ Dmel\\CG9986', description='CG9986 gene product from transcript CG9986-RA', other_designations='CG9986-PA', map_location='98D6-98D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28606079, end_position_on_the_genomic_accession=28608372, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43387, CurrentID=0, Status='live', Symbol='CG10011', Aliases='Dmel_ Dmel\\CG10011', description='CG10011 gene product from transcript CG10011-RA', other_designations='CG10011-PA|uncharacterized protein', map_location='98D6-98D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28607939, end_position_on_the_genomic_accession=28635099, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43388, CurrentID=0, Status='live', Symbol='CG9988', Aliases='Dmel_ Dmel\\CG9988, gi23172514', description='CG9988 gene product from transcript CG9988-RA', other_designations='CG9988-PA', map_location='98D7-98D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28636411, end_position_on_the_genomic_accession=28638799, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43389, CurrentID=0, Status='live', Symbol='CG14062', Aliases='Dmel_ Dmel\\CG14062, LP23408p', description='CG14062 gene product from transcript CG14062-RB', other_designations='CG14062-PB|uncharacterized protein', map_location='98D7-98D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28641585, end_position_on_the_genomic_accession=28642891, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43390, CurrentID=0, Status='live', Symbol='Sid', Aliases='Dmel_CG9989, CG9989, Dmel\\CG9989, sid', description='Stress induced DNase', other_designations='CG9989-PA|Sid-PA|stress induced DNase', map_location='98E1-98E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28670704, end_position_on_the_genomic_accession=28672021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43391, CurrentID=0, Status='live', Symbol='CG9990', Aliases='Dmel_ Dmel\\CG9990', description='CG9990 gene product from transcript CG9990-RA', other_designations='CG9990-PA|CG9990-PB|CG9990-PD|CG9990-PE|CG9990-PF|CG9990-PG', map_location='98E1-98E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28675041, end_position_on_the_genomic_accession=28695939, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  43392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43392, CurrentID=0, Status='live', Symbol='htt', Aliases='Dmel_CG9995, CG9995, DmHtt, Dmel\\CG9995, HD, HTT, Hsap\\HD, Htt, dHtt, dhtt', description='huntingtin', other_designations='CG9995-PA|CG9995-PB|htt-PA|htt-PB', map_location='98E2-98E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28696242, end_position_on_the_genomic_accession=28734816, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  43393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43393, CurrentID=0, Status='live', Symbol='AstA-R2', Aliases='Dmel_CG10001, AR-2, ASTR, CG10001, D.AlstR2, DAR, DAR-2, DAR2, Dmel\\CG10001, anon-WO0131005.7, anon-WO0170980.115, anon-WO0170980.116', description='Allatostatin A receptor 2', other_designations='ASA-receptor|AstA-R2-PA|AstA-R2-PB|AstA-R2-PC|CG10001-PA|CG10001-PB|CG10001-PC|allatostatin A receptor 2|allatostatin receptor 2', map_location='98E2-98E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28737109, end_position_on_the_genomic_accession=28744782, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43394, CurrentID=0, Status='live', Symbol='CG10000', Aliases='Dmel_ Dmel\\CG10000, dppGalNAcT10', description='CG10000 gene product from transcript CG10000-RA', other_designations='CG10000-PA|uncharacterized protein', map_location='98E3-98E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28748384, end_position_on_the_genomic_accession=28751661, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43395, CurrentID=0, Status='live', Symbol='CG9997', Aliases='Dmel_ Dmel\\CG9997', description='CG9997 gene product from transcript CG9997-RB', other_designations='CG9997-PA|CG9997-PB', map_location='98E3-98E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28755484, end_position_on_the_genomic_accession=28756979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43396, CurrentID=0, Status='live', Symbol='aqrs', Aliases='Dmel_CG14061, CG14061, Dmel\\CG14061', description='aquarius', other_designations='CG14061-PA|aqrs-PA', map_location='98E3-98E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28756999, end_position_on_the_genomic_accession=28758234, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43397, CurrentID=0, Status='live', Symbol='intr', Aliases='Dmel_CG12558, CG12558, Dmel\\CG12558, SPH202', description='intrepid', other_designations='CG12558-PA|intr-PA', map_location='98E4-98E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28790522, end_position_on_the_genomic_accession=28791461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43398, CurrentID=0, Status='live', Symbol='wdn', Aliases='Dmel_CG1454, CG1454, Dmel\\CG1454, Sryc, pqp, sry h-l, sry-c, sry-h1', description='wings down', other_designations='CG1454-PA|pourquoi-pas|serendipity cognate|serendipity-cognate|wdn-PA', map_location='98E5-98E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28807983, end_position_on_the_genomic_accession=28811337, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43399, CurrentID=0, Status='live', Symbol='CG1646', Aliases='Dmel_ Dmel\\CG1646, Prp39', description='CG1646 gene product from transcript CG1646-RC', other_designations='CG1646-PA|CG1646-PB|CG1646-PC|CG1646-PE|CG1646-PH', map_location='98E5-98E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28812102, end_position_on_the_genomic_accession=28817354, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43400, CurrentID=0, Status='live', Symbol='CG1523', Aliases='Dmel_ Dmel\\CG1523', description='CG1523 gene product from transcript CG1523-RA', other_designations='CG1523-PA', map_location='98E5-98E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28817342, end_position_on_the_genomic_accession=28819668, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43401, CurrentID=0, Status='live', Symbol='CG1647', Aliases='Dmel_ Dmel\\CG1647', description='CG1647 gene product from transcript CG1647-RA', other_designations='CG1647-PA|CG1647-PB', map_location='98E5-98E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28820221, end_position_on_the_genomic_accession=28824508, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43402, CurrentID=0, Status='live', Symbol='WASp', Aliases='Dmel_CG1520, CG1520, D-WASP, Dm WASP, DmWASP, Dmel\\CG1520, WASP, WSP, Wasp, Wsp, dWASP, wasp, wsp', description='CG1520 gene product from transcript CG1520-RA', other_designations='CG1520-PA|CG1520-PB|CG1520-PC|CG1520-PD|WASP|WASp-PA|WASp-PB|WASp-PC|WASp-PD|Wiscott-Aldrich syndrome protein|Wiskott-Aldrich syndrome protein', map_location='98E5-98E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28824506, end_position_on_the_genomic_accession=28832024, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43404, CurrentID=0, Status='live', Symbol='spg', Aliases='Dmel_CG31048, CG11754, CG14530, CG31048, Dm Dock4, Dmel\\CG31048, mat(3)6, mat3(6)', description='sponge', other_designations='CG31048-PD|CG31048-PE|CG31048-PF|CG31048-PG|CG31048-PH|Dock4|early-D|spg-PD|spg-PE|spg-PF|spg-PG|spg-PH', map_location='3-95 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28845270, end_position_on_the_genomic_accession=28870439, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43408, CurrentID=0, Status='live', Symbol='CG14529', Aliases='Dmel_ Dmel\\CG14529', description='CG14529 gene product from transcript CG14529-RA', other_designations='CG14529-PA|uncharacterized protein', map_location='98F1-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28872905, end_position_on_the_genomic_accession=28875031, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43410, CurrentID=0, Status='live', Symbol='CG14528', Aliases='Dmel_ Dmel\\CG14528', description='CG14528 gene product from transcript CG14528-RA', other_designations='CG14528-PA|uncharacterized protein', map_location='98F1-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28875621, end_position_on_the_genomic_accession=28877805, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43411, CurrentID=0, Status='live', Symbol='CG14523', Aliases='Dmel_ Dmel\\CG14523', description='CG14523 gene product from transcript CG14523-RA', other_designations='CG14523-PA|uncharacterized protein', map_location='98F1-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28878024, end_position_on_the_genomic_accession=28880122, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43412, CurrentID=0, Status='live', Symbol='CG14527', Aliases='Dmel_ Dmel\\CG14527', description='CG14527 gene product from transcript CG14527-RA', other_designations='CG14527-PA|uncharacterized protein', map_location='98F1-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28880792, end_position_on_the_genomic_accession=28882945, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43413, CurrentID=0, Status='live', Symbol='CG14526', Aliases='Dmel_ BcDNA:GH07188, Dmel\\CG14526', description='CG14526 gene product from transcript CG14526-RC', other_designations='CG14526-PC|uncharacterized protein', map_location='98F1-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28883693, end_position_on_the_genomic_accession=28886015, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43414, CurrentID=0, Status='live', Symbol='CG33203', Aliases='Dmel_ CG14522, CG14524, CG14525, CG31046, Dmel\\CG33203', description='CG33203 gene product from transcript CG33203-RE', other_designations='CG33203-PD|CG33203-PE|CG33203-PF|uncharacterized protein', map_location='98F1-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28870606, end_position_on_the_genomic_accession=28887709, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43415, CurrentID=0, Status='live', Symbol='Doa', Aliases='Dmel_CG42320, 0844/02, CG1658, CG31049, CG33204, CG33553, CG42320, DOA, DOA/CLK2, Dmel\\CG42320, Dmel_CG33204, Dmel_CG33553, Msu, doa, l(3)01705, l(3)S084402, l(3)s2784', description='Darkener of apricot', other_designations='CG42320-PAA|CG42320-PJ|CG42320-PK|CG42320-PL|CG42320-PM|CG42320-PN|CG42320-PP|CG42320-PR|CG42320-PS|CG42320-PT|CG42320-PU|CG42320-PV|CG42320-PX|CG42320-PY|CG42320-PZ|Doa-PAA|Doa-PJ|Doa-PK|Doa-PL|Doa-PM|Doa-PN|Doa-PP|Doa-PR|Doa-PS|Doa-PT|Doa-PU|Doa-PV|Doa-PX|Doa-PY|Doa-PZ|darkener of apricot|lethal (3) S084402|lethal (3) s2784|mosaic suppressor', map_location='3-99 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28888073, end_position_on_the_genomic_accession=28922772, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  43416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43416, CurrentID=0, Status='live', Symbol='CG11828', Aliases='Dmel_ Dmel\\CG11828', description='CG11828 gene product from transcript CG11828-RD', other_designations='CG11828-PD|uncharacterized protein', map_location='98F2-98F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28923044, end_position_on_the_genomic_accession=28925086, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43417, CurrentID=0, Status='live', Symbol='DIP-gamma', Aliases='Dmel_CG14521, 14521, CG14521, CT34248, Dmel\\CG14521, anon-WO0140519.196', description='Dpr-interacting protein gamma', other_designations='CG14521-PA|DIP-gamma-PA', map_location='98F2-98F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28925004, end_position_on_the_genomic_accession=28970372, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43420, CurrentID=0, Status='live', Symbol='wat', Aliases='Dmel_CG1443, CG1443, Dmel\\CG1443, Wat', description='waterproof', other_designations='CG1443-PA|wat-PA', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28992977, end_position_on_the_genomic_accession=29020406, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43421, CurrentID=0, Status='live', Symbol='CG14518', Aliases='Dmel_ Dmel\\CG14518', description='CG14518 gene product from transcript CG14518-RA', other_designations='CG14518-PA|uncharacterized protein', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29024971, end_position_on_the_genomic_accession=29026159, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43422, CurrentID=0, Status='live', Symbol='eIF4E-6', Aliases='Dmel_CG1442, BcDNA:RH55324, CG1442, Dmel\\CG1442, eIF4E', description='CG1442 gene product from transcript CG1442-RA', other_designations='CG1442-PA|CG1442-PB|CG1442-PC|eIF4E-6-PA|eIF4E-6-PB|eIF4E-6-PC', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29026439, end_position_on_the_genomic_accession=29027174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43423, CurrentID=0, Status='live', Symbol='CG1951', Aliases='Dmel_ Dmel\\CG1951', description='CG1951 gene product from transcript CG1951-RA', other_designations='CG1951-PA', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29027236, end_position_on_the_genomic_accession=29030938, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43424, CurrentID=0, Status='live', Symbol='Ssl2', Aliases='Dmel_CG11833, CG11833, DmSSl2, Dmel\\CG11833, SSl2', description='Strictosidine synthase-like 2', other_designations='CG11833-PB|Ssl2-PB|strictosidine synthase-like 2', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29030951, end_position_on_the_genomic_accession=29032642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43425, CurrentID=0, Status='live', Symbol='beta4GalNAcTB', Aliases='Dmel_CG14517, CG14517, Dmel\\CG14517, beta4GalTB, dbeta4GalTB', description='CG14517 gene product from transcript CG14517-RA', other_designations='CG14517-PA|beta4GalNAcTB-PA', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29032660, end_position_on_the_genomic_accession=29034008, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43426, CurrentID=0, Status='live', Symbol='Cpsf100', Aliases='Dmel_CG1957, BcDNA:LD14168, CG1957, CPSF-100, CPSF100, Dmel\\CG1957', description='Cleavage and polyadenylation specificity factor 100', other_designations='CG1957-PA|Cpsf100-PA|cleavage and polyadenylation specificity factor 100', map_location='98F5-98F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29034573, end_position_on_the_genomic_accession=29037062, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43427, CurrentID=0, Status='live', Symbol='Slu7', Aliases='Dmel_CG1420, CG1420, Dmel\\CG1420, SLU7', description='CG1420 gene product from transcript CG1420-RA', other_designations='CG1420-PA|Slu7-PA', map_location='98F5-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29037058, end_position_on_the_genomic_accession=29039071, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43428, CurrentID=0, Status='live', Symbol='Pkc98E', Aliases='Dmel_CG1954, 98F, CG1954, Dmel\\CG1954, Dpkc3, Nc98F, PK-C, PKC, PKC 98F, PKC d98F, PKC-98F, PKC98C, PKC98E, PKCdelta, Pkc3, dPKC98F, nPKC', description='Protein C kinase 98E', other_designations='CG1954-PA|CG1954-PB|CG1954-PC|CG1954-PD|Neural-conserved-at-98F|PKC-brain|Pkc98E-PA|Pkc98E-PB|Pkc98E-PC|Pkc98E-PD|protein C kinase 98E|protein kinase C|protein kinase-c3', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29039156, end_position_on_the_genomic_accession=29049148, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  43429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43429, CurrentID=0, Status='live', Symbol='CG11837', Aliases='Dmel_ Dim1, Dmel\\CG11837, FBgn0039627', description='CG11837 gene product from transcript CG11837-RA', other_designations='CG11837-PA|uncharacterized protein', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29049493, end_position_on_the_genomic_accession=29050721, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43430, CurrentID=0, Status='live', Symbol='CG11841', Aliases='Dmel_ BcDNA:GH26583, Dmel\\CG11841, SP45, anon-WO0140519.18, anon-WO0140519.6', description='CG11841 gene product from transcript CG11841-RA', other_designations='CG11841-PA|uncharacterized protein', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29051056, end_position_on_the_genomic_accession=29052267, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43431, CurrentID=0, Status='live', Symbol='CG11842', Aliases='Dmel_ Dmel\\CG11842, SP68', description='CG11842 gene product from transcript CG11842-RA', other_designations='CG11842-PA|uncharacterized protein', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29052680, end_position_on_the_genomic_accession=29053913, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43432, CurrentID=0, Status='live', Symbol='CG11843', Aliases='Dmel_ Dmel\\CG11843, SP43', description='CG11843 gene product from transcript CG11843-RB', other_designations='CG11843-PA|CG11843-PB|uncharacterized protein', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29054347, end_position_on_the_genomic_accession=29055837, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43433, CurrentID=0, Status='live', Symbol='Sirt7', Aliases='Dmel_CG11305, CG11305, D.mel5, Dmel\\CG11305, SIRT7, dSIRT7, dmSRT409', description='Sirtuin 7', other_designations='CG11305-PA|Sirt7-PA|sirtuin 7', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29055927, end_position_on_the_genomic_accession=29059204, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43434, CurrentID=0, Status='live', Symbol='Cul5', Aliases='Dmel_CG1401, CG1401, CUL5, Cul-5, Cullin5, Dmel\\CG1401, cul-5, cul5', description='Cullin 5', other_designations='CG1401-PA|Cul5-PA|Cullin-5|cullin|cullin 5', map_location='98F6-98F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29059272, end_position_on_the_genomic_accession=29062975, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43435, CurrentID=0, Status='live', Symbol='CG11873', Aliases='Dmel_ Dmel\\CG11873, anon-WO0140519.162', description='CG11873 gene product from transcript CG11873-RB', other_designations='CG11873-PA|CG11873-PB|CG11873-PC|CG11873-PD|uncharacterized protein', map_location='98F6-98F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29063040, end_position_on_the_genomic_accession=29108948, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43436, CurrentID=0, Status='live', Symbol='alpha-Man-Ib', Aliases='Dmel_CG11874, CG11874, Dmel\\CG11874, ORE-11, dMas-3, mas-3', description='alpha-Mannosidase class I b', other_designations='CG11874-PA|alpha mannosidase Ib|alpha-Man-Ib-PA', map_location='98F10-98F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29109053, end_position_on_the_genomic_accession=29112822, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43437, CurrentID=0, Status='live', Symbol='CG11876', Aliases='Dmel_ CG 11876, Dmel\\CG11876, E1beta, PDHB', description='CG11876 gene product from transcript CG11876-RA', other_designations='CG11876-PA|CG11876-PB|CG11876-PC|CG11876-PD|pyruvate dehydrogenase E1beta subunit|uncharacterized protein', map_location='98F10-98F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29113071, end_position_on_the_genomic_accession=29116197, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43438, CurrentID=0, Status='live', Symbol='Atg14', Aliases='Dmel_CG11877, CG11877, Dmel\\CG11877', description='Autophagy-related 14', other_designations='Atg14-PA|Autophagy-specific gene 14|CG11877-PA', map_location='98F10-98F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29116391, end_position_on_the_genomic_accession=29118390, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43439, CurrentID=0, Status='live', Symbol='yem', Aliases='Dmel_CG14513, CG11879, CG14513, Dmel\\CG14513, Yemalpha, yG4.5, yT4.5-alpha, yema, yemalpha, yem', description='yemanuclein', other_designations='CG14513-PA|CG14513-PB|yem-PA|yem-PB|yem-alpha|yema nuclein alpha|yema-gene-4.5|yemalpha|yemanuclein alpha|yemanuclein-alpha', map_location='98F10-98F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29118330, end_position_on_the_genomic_accession=29122879, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43440, CurrentID=0, Status='live', Symbol='CG11880', Aliases='Dmel_ Dmel\\CG11880', description='CG11880 gene product from transcript CG11880-RA', other_designations='CG11880-PA|CG11880-PB|CG11880-PC|uncharacterized protein', map_location='98F10-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29123370, end_position_on_the_genomic_accession=29128247, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43441, CurrentID=0, Status='live', Symbol='dgt6', Aliases='Dmel_CG11881, CG11881, Dgt6, Dmel\\CG11881, msd4', description='dim gamma-tubulin 6', other_designations='CG11881-PA|dgt6-PA|mitotic spindle density 4', map_location='98F12-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29128584, end_position_on_the_genomic_accession=29130972, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43442, CurrentID=0, Status='live', Symbol='Vha100-1', Aliases='Dmel_CG1709, BcDNA:LD21248, CG1709, Dmel\\CG1709, V0, V100, v0a1, v100, vha, vha100-1', description='Vacuolar H[+] ATPase 100kD subunit 1', other_designations='CG1709-PA|CG1709-PB|CG1709-PC|CG1709-PD|CG1709-PE|CG1709-PF|CG1709-PG|CG1709-PI|CG1709-PJ|CG1709-PK|V0-ATPase a1|Vha100-1-PA|Vha100-1-PB|Vha100-1-PC|Vha100-1-PD|Vha100-1-PE|Vha100-1-PF|Vha100-1-PG|Vha100-1-PI|Vha100-1-PJ|Vha100-1-PK|neuron-specific V[[0]] subunit a1|vacuolar ATPase V0, subunit a1|vacuolar H[+] ATPase 100kD subunit 1|vacuolar H[+] ATPase subunit 100-1', map_location='98F12-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29131193, end_position_on_the_genomic_accession=29137980, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43443, CurrentID=0, Status='live', Symbol='CG14512', Aliases='Dmel_ Alg13, Dmel\\CG14512, dAlg13', description='CG14512 gene product from transcript CG14512-RA', other_designations='CG14512-PA|uncharacterized protein', map_location='98F12-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29137970, end_position_on_the_genomic_accession=29138573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43444, CurrentID=0, Status='live', Symbol='CG14516', Aliases='Dmel_ Dmel\\CG14516', description='CG14516 gene product from transcript CG14516-RB', other_designations='CG14516-PA|CG14516-PB|CG14516-PC|uncharacterized protein', map_location='98F12-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29138662, end_position_on_the_genomic_accession=29145787, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43445, CurrentID=0, Status='live', Symbol='CG14511', Aliases='Dmel_ Dmel\\CG14511', description='CG14511 gene product from transcript CG14511-RA', other_designations='CG14511-PA|uncharacterized protein', map_location='98F12-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29145678, end_position_on_the_genomic_accession=29146893, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43446, CurrentID=0, Status='live', Symbol='CG11882', Aliases='Dmel_ Dmel\\CG11882', description='CG11882 gene product from transcript CG11882-RA', other_designations='CG11882-PA|CG11882-PB|uncharacterized protein', map_location='98F12-98F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29147016, end_position_on_the_genomic_accession=29148172, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43447, CurrentID=0, Status='live', Symbol='Pglym78', Aliases='Dmel_CG1721, CG1721, Dmel\\CG1721, PGAM, PGLYM, Pglym, pglym78', description='Phosphoglyceromutase', other_designations='CG1721-PA|CG1721-PB|CG1721-PC|Pglym78-PA|Pglym78-PB|Pglym78-PC|phosphoglycerate mutase|phosphoglycerol mutase|phosphoglyceromutase', map_location='98F12-98F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29148248, end_position_on_the_genomic_accession=29149901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43448, CurrentID=0, Status='live', Symbol='Slbp', Aliases='Dmel_CG11886, CG11886, Dmel\\CG11886, SLBP, dSLBP, dSLPB, slbp', description='Stem-loop binding protein', other_designations='CG11886-PA|Slbp-PA|stem loop binding protein', map_location='98F13-98F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29152074, end_position_on_the_genomic_accession=29153266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43449, CurrentID=0, Status='live', Symbol='Rpn2', Aliases='Dmel_CG11888, 11888, CG11888, Dmel\\CG11888, PSMC1, RPN2, S1, dRpn2, p110, rpn2', description='Regulatory particle non-ATPase 2', other_designations='CG11888-PA|CG11888-PB|CG11888-PC|Rpn2-PA|Rpn2-PB|Rpn2-PC|regulatory particle non-ATPase 2', map_location='98F13-98F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29153524, end_position_on_the_genomic_accession=29157255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43450, CurrentID=0, Status='live', Symbol='CG11897', Aliases='Dmel_ Dmel\\CG11897', description='CG11897 gene product from transcript CG11897-RB', other_designations='CG11897-PA|CG11897-PB|uncharacterized protein', map_location='98F13-98F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29158122, end_position_on_the_genomic_accession=29164966, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43451, CurrentID=0, Status='live', Symbol='CG11898', Aliases='Dmel_ Dme_CG11898, Dmel\\CG11898', description='CG11898 gene product from transcript CG11898-RB', other_designations='CG11898-PB|uncharacterized protein', map_location='98F13-98F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29165643, end_position_on_the_genomic_accession=29171508, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43453, CurrentID=0, Status='live', Symbol='CG14509', Aliases='Dmel_ CG14510, Dmel\\CG14509', description='CG14509 gene product from transcript CG14509-RD', other_designations='CG14509-PB|CG14509-PC|CG14509-PD|CG14509-PE|uncharacterized protein', map_location='98F13-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29171191, end_position_on_the_genomic_accession=29194994, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43454, CurrentID=0, Status='live', Symbol='CG14515', Aliases='Dmel_ Dmel\\CG14515', description='CG14515 gene product from transcript CG14515-RA', other_designations='CG14515-PA|uncharacterized protein', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29197556, end_position_on_the_genomic_accession=29199211, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43456, CurrentID=0, Status='live', Symbol='Mesh1', Aliases='Dmel_CG11900, CG11900, Dmel\\CG11900, Q9VAM9', description='Metazoan SpoT homolog-1', other_designations='CG11900-PA|Mesh1-PA|metazoan SpoT homolog-1', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29201756, end_position_on_the_genomic_accession=29202525, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43457, CurrentID=0, Status='live', Symbol='Cyt-c1L', Aliases='Dmel_CG14508, CG14508, CY1, Dmel\\CG14508', description='Cytochrome c1-like', other_designations='CG14508-PA|Cyt-c1L-PA|Ubiquinol-cytochrome c reductase cytochrome C1, heme protein|cytochrome c1, heme protein, mitochondrial|cytochrome c1-like', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29202660, end_position_on_the_genomic_accession=29204133, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43459, CurrentID=0, Status='live', Symbol='Cisd2', Aliases='Dmel_CG1458, BcDNA:RE49709, CG1458, Dmel\\CG1458, MitoNEET, cisd2, dWFS2', description='CDGSH iron sulfur domain 2', other_designations='CG1458-PA|Cisd2-PA', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29222217, end_position_on_the_genomic_accession=29223428, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43460, CurrentID=0, Status='live', Symbol='Brd8', Aliases='Dmel_CG14514, CG14514, Dmel\\CG14514, dBrd8', description='Bromodomain containing 8', other_designations='Brd8-PA|CG14514-PA|bromodomain containing 8', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29223485, end_position_on_the_genomic_accession=29226276, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43461, CurrentID=0, Status='live', Symbol='CG14507', Aliases='Dmel_ Dmel\\CG14507', description='CG14507 gene product from transcript CG14507-RC', other_designations='CG14507-PC|uncharacterized protein', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29226204, end_position_on_the_genomic_accession=29227725, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43462, CurrentID=0, Status='live', Symbol='Usp1', Aliases='Dmel_CG15817, BcDNA:LD22910, CG15817, Dmel\\CG15817, LD22910, Q8I077', description='Ubiquitin specific protease 1', other_designations='CG15817-PC|CG15817-PD|CG15817-PE|Usp1-PC|Usp1-PD|Usp1-PE|ubiquitin specific protease 1', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29229161, end_position_on_the_genomic_accession=29235302, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43463, CurrentID=0, Status='live', Symbol='CG11951', Aliases='Dmel_ Dmel\\CG11951', description='CG11951 gene product from transcript CG11951-RA', other_designations='CG11951-PA|CG11951-PB|uncharacterized protein', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29235322, end_position_on_the_genomic_accession=29238704, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43464, CurrentID=0, Status='live', Symbol='CG31427', Aliases='Dmel_ CG11955, Dmel\\CG31427', description='CG31427 gene product from transcript CG31427-RD', other_designations='CG31427-PC|CG31427-PD|CG31427-PE|CG31427-PF|uncharacterized protein', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29238830, end_position_on_the_genomic_accession=29240363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43466, CurrentID=0, Status='live', Symbol='stg', Aliases='Dmel_CG1395, 0224/06, 0245/03, 0439/22, 0730/13, 0896/05, 0967/05, 0980/06, 1083/13, 1089/08, 1143/02, 5473, CDC25, CDC25[string], CG1395, Cdc25, Cdc25[Stg], Cdc25[String], Cdc25[stg], Cdc25[string], Cdc25c, Dmel\\CG1395, EP1213, S(rux)3A, STG, SY3-4, Stg, String/Cdc25, anon-EST:Liang-2.21, cdc25, cdc25[string], clone 2.21, l(3)01235, l(3)j10B9, l(3)j1D3, l(3)j1E3, l(3)j3D1, l(3)s2213, st[cdc25], str/cdc25, stg', description='string', other_designations='CG1395-PA|CG1395-PB|Cdc25|String/Cdc25 phosphatase|cdc25/string|stg-PA|stg-PB|string/cdc25|suppressor of rux 3A', map_location='99A5-99A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29251826, end_position_on_the_genomic_accession=29255800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43467, CurrentID=0, Status='live', Symbol='CG14506', Aliases='Dmel_ Dmel\\CG14506, cg14506', description='CG14506 gene product from transcript CG14506-RB', other_designations='CG14506-PB|CG14506-PC|uncharacterized protein', map_location='99A5-99A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29293743, end_position_on_the_genomic_accession=29297180, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43469, CurrentID=0, Status='live', Symbol='Ptp99A', Aliases='Dmel_CG11516, CG11515, CG11516, CG11517, CG2005, CT6383, DPTP99A, DPTP[[99A]], Dmel\\CG11516, PTP99A, R-PTP 99A, dptp99A, ptp99A', description='Protein tyrosine phosphatase 99A', other_designations='CG11516-PA|CG11516-PB|CG11516-PC|CG11516-PF|CG11516-PG|Ptp99A-PA|Ptp99A-PB|Ptp99A-PC|Ptp99A-PF|Ptp99A-PG|protein tyrosine phosphatase 99A', map_location='99A7-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29377646, end_position_on_the_genomic_accession=29487131, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  43471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43471, CurrentID=0, Status='live', Symbol='CG2321', Aliases='Dmel_ Dmel\\CG2321', description='CG2321 gene product from transcript CG2321-RB', other_designations='CG2321-PA|CG2321-PB', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29484783, end_position_on_the_genomic_accession=29489265, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43472, CurrentID=0, Status='live', Symbol='CG2006', Aliases='Dmel_ Dmel\\CG2006', description='CG2006 gene product from transcript CG2006-RA', other_designations='CG2006-PA|CG2006-PB', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29489468, end_position_on_the_genomic_accession=29491383, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43473, CurrentID=0, Status='live', Symbol='CG2310', Aliases='Dmel_ Dmel\\CG2310', description='CG2310 gene product from transcript CG2310-RA', other_designations='CG2310-PA|CG2310-PB', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29491524, end_position_on_the_genomic_accession=29493518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43474, CurrentID=0, Status='live', Symbol='Diedel', Aliases='Dmel_CG11501, BcDNA:RH38110, CG11501, Dmel\\CG11501', description='CG11501 gene product from transcript CG11501-RA', other_designations='CG11501-PA|Diedel-PA|diedel', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29493852, end_position_on_the_genomic_accession=29494267, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43475, CurrentID=0, Status='live', Symbol='CG2010', Aliases='Dmel_ Dmel\\CG2010, FBXL3', description='CG2010 gene product from transcript CG2010-RB', other_designations='CG2010-PA|CG2010-PB|CG2010-PC', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29494782, end_position_on_the_genomic_accession=29498923, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43476, CurrentID=0, Status='live', Symbol='Trc8', Aliases='Dmel_CG2304, CG2304, DTrc8, Dmel\\CG2304, Dtrc8, TRC8', description='TRC8 ortholog', other_designations='CG2304-PA|CG2304-PB|CG2304-PD|CG2304-PE|Trc8-PA|Trc8-PB|Trc8-PD|Trc8-PE', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29499025, end_position_on_the_genomic_accession=29503797, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43477, CurrentID=0, Status='live', Symbol='ND-20L', Aliases='Dmel_CG2014, CG2014, Dmel\\CG2014, ND20-like, NUKM', description='NADH dehydrogenase (ubiquinone) 20 kDa subunit-like', other_designations='CG2014-PA|CG2014-PB|NADH-ubiquinone oxidoreductase 20 kDa subunit|NADH: ubiquinone oxidoreductase-like, 20kDa subunit|NADH:ubiquinone oxidoreductase 20 kDa subunit|ND-20L-PA|ND-20L-PB', map_location='99B2-99B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29515717, end_position_on_the_genomic_accession=29516527, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43479, CurrentID=0, Status='live', Symbol='CG7567', Aliases='Dmel_ BcDNA:RE47447, Dmel\\CG7567', description='CG7567 gene product from transcript CG7567-RA', other_designations='CG7567-PA', map_location='99B4-99B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29587237, end_position_on_the_genomic_accession=29588104, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  338392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=338392, CurrentID=0, Status='live', Symbol='CG33932', Aliases='Dmel_ BcDNA:RE70209, CG12533, CG32530, CG32530-ORFB, Dmel\\CG33932', description='CG33932 gene product from transcript CG33932-RB', other_designations='CG33932-PA|CG33932-PB|uncharacterized protein', map_location='18E5-18E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19744431, end_position_on_the_genomic_accession=19745898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43481, CurrentID=0, Status='live', Symbol='alph', Aliases='Dmel_CG1906, CG1906, Dmel\\CG1906, PP2Cb, SK3-1, pp2c99B', description='alphabet', other_designations='CG1906-PA|CG1906-PB|CG1906-PC|CG1906-PD|CG1906-PE|CG1906-PF|alph-PA|alph-PB|alph-PC|alph-PD|alph-PE|alph-PF|pp2c99B', map_location='99B4-99B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29591831, end_position_on_the_genomic_accession=29617157, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  43482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43482, CurrentID=0, Status='live', Symbol='CG7568', Aliases='Dmel_ Dmel\\CG7568', description='CG7568 gene product from transcript CG7568-RE', other_designations='CG7568-PD|CG7568-PE', map_location='99B4-99B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29598670, end_position_on_the_genomic_accession=29600862, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43483, CurrentID=0, Status='live', Symbol='CG1907', Aliases='Dmel_ CG 1907, Dmel\\CG1907', description='CG1907 gene product from transcript CG1907-RA', other_designations='CG1907-PA', map_location='99B5-99B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29619065, end_position_on_the_genomic_accession=29620758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43484, CurrentID=0, Status='live', Symbol='Dop1R2', Aliases='Dmel_CG18741, CG18741, CG7569, DAMB, DDR2, DOPR2, DOPR99B, DR2, Damb, Dar2, Dmel\\CG18741, DopD1, DopR2, DopR99B, DopRII, K13sc', description='Dopamine 1-like receptor 2', other_designations='CG18741-PA|CG18741-PB|CG18741-PC|DAMB DA receptor|D[[2]]-like receptor|Dop1R2-PA|Dop1R2-PB|Dop1R2-PC|K13sc|dopamine 1-like receptor 2|dopamine receptor 2|dopamine receptor in mushroom bodies|dopamine-2 receptor', map_location='99B5-99B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29630304, end_position_on_the_genomic_accession=29659984, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43485, CurrentID=0, Status='live', Symbol='ppk21', Aliases='Dmel_CG12048, CG12048, Dmel\\CG12048, PPK21', description='pickpocket 21', other_designations='CG12048-PA|CG12048-PB|ppk21-PA|ppk21-PB', map_location='99B6-99B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29664779, end_position_on_the_genomic_accession=29666783, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43486, CurrentID=0, Status='live', Symbol='ppk20', Aliases='Dmel_CG7577, CG7577, Dmel\\CG7577, PPK20', description='pickpocket 20', other_designations='CG7577-PA|CG7577-PB|CG7577-PC|ppk20-PA|ppk20-PB|ppk20-PC', map_location='99B6-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29666814, end_position_on_the_genomic_accession=29668877, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43487, CurrentID=0, Status='live', Symbol='ppk30', Aliases='Dmel_CG18110, CG18110, Dmel\\CG18110', description='pickpocket 30', other_designations='CG18110-PA|CG18110-PB|ppk30-PA|ppk30-PB', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29669149, end_position_on_the_genomic_accession=29671478, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43488, CurrentID=0, Status='live', Symbol='Obp99a', Aliases='Dmel_CG18111, 99a, BcDNA:RH70762, CG18111, DmelObp99a, Dmel\\CG18111, OBP99a', description='Odorant-binding protein 99a', other_designations='CG18111-PA|CG18111-PB|Obp99a-PA|Obp99a-PB|odorant binding protein 99a', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29671657, end_position_on_the_genomic_accession=29672334, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43489, CurrentID=0, Status='live', Symbol='ppk19', Aliases='Dmel_CG18287, CG18287, Dmel\\CG18287, PPK19', description='pickpocket 19', other_designations='CG18287-PA|Pickpocket19|ppk19-PA', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29672456, end_position_on_the_genomic_accession=29674223, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43490, CurrentID=0, Status='live', Symbol='Bub3', Aliases='Dmel_CG7581, BUB3, CG7581, Dmbub3, Dmel\\CG7581, bub3, dBUB3', description='CG7581 gene product from transcript CG7581-RB', other_designations='Bub3-PA|Bub3-PB|CG7581-PA|CG7581-PB|bub3', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29674007, end_position_on_the_genomic_accession=29675421, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43491, CurrentID=0, Status='live', Symbol='Cap-D2', Aliases='Dmel_CG1911, CAP-D2, CG1911, CapD2, DmCAP-D2, Dmel\\CG1911, dCap-D2', description='CAP-D2 condensin subunit', other_designations='CG1911-PA|Cap-D2-PA', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29675651, end_position_on_the_genomic_accession=29680257, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43492, CurrentID=0, Status='live', Symbol='CG7582', Aliases='Dmel_ Dmel\\CG7582', description='CG7582 gene product from transcript CG7582-RB', other_designations='CG7582-PA|CG7582-PB', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29680241, end_position_on_the_genomic_accession=29682278, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43493, CurrentID=0, Status='live', Symbol='Gycalpha99B', Aliases='Dmel_CG1912, CG1912, Dgc1, Dgcalpha1, Dmel\\CG1912, GCalpha1, GYCa 99B, Gcalpha1, Gycalpha-99B, Gycalpha63A, SGCalpha, dgc1, dgca1, dgcalpha1, gycalpha, sGC, sGCalpha1, sGCalpha99B', description='Guanylyl cyclase alpha-subunit at 99B', other_designations='CG1912-PA|Gycalpha99B-PA|guanyl cyclase alpha-subunit at 63A|guanylate cyclase 99B|guanylyl cyclase alpha-subunit at 99B|soluble guanylate cyclase|soluble guanylyl cyclase alpha', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29682503, end_position_on_the_genomic_accession=29685758, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43494, CurrentID=0, Status='live', Symbol='Obp99c', Aliases='Dmel_CG7584, 99c, CG7584, DmelObp99c, Dmel\\CG7584, Obp 99c, Obp99b', description='Odorant-binding protein 99c', other_designations='CG7584-PA|CG7584-PB|Obp99c-PA|Obp99c-PB', map_location='99B7-99B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29685736, end_position_on_the_genomic_accession=29689311, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43495, CurrentID=0, Status='live', Symbol='dmrt99B', Aliases='Dmel_CG15504, CG15504, Dmel\\CG15504', description='doublesex-Mab related 99B', other_designations='CG15504-PA|dmrt99B-PA', map_location='99B8-99B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29692591, end_position_on_the_genomic_accession=29697025, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43496, CurrentID=0, Status='live', Symbol='Obp99d', Aliases='Dmel_CG15505, 99d, CG15505, CG155505, DmelObp99d, Dmel\\CG15505, Obp99c', description='Odorant-binding protein 99d', other_designations='CG15505-PA|Obp99d-PA', map_location='99B8-99B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29714505, end_position_on_the_genomic_accession=29715004, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43497, CurrentID=0, Status='live', Symbol='Obp99b', Aliases='Dmel_CG7592, 99b, BcDNA:LP06572, CG7592, DmelObp99b, Dmel\\CG7592, OBP99b, Obp99d, obp99b, tsx', description='Odorant-binding protein 99b', other_designations='CG7592-PA|CG7592-PB|Obp99b-PA|Obp99b-PB|Odordant-binding protein 99b|general odorant-binding protein 99b precursor|odorant binding protein 99b|turn on sex-specificity', map_location='99B8-99B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29716254, end_position_on_the_genomic_accession=29716896, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43498, CurrentID=0, Status='live', Symbol='CG15506', Aliases='Dmel_ Dmel\\CG15506', description='CG15506 gene product from transcript CG15506-RA', other_designations='CG15506-PA|uncharacterized protein', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29718663, end_position_on_the_genomic_accession=29722388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43499, CurrentID=0, Status='live', Symbol='Pcd', Aliases='Dmel_CG1963, BcDNA:LP07248, CG1963, Dmel\\CG1963, PCD', description='pterin-4a-carbinolamine dehydratase', other_designations='CG1963-PA|Pcd-PA', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29723163, end_position_on_the_genomic_accession=29724226, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43500, CurrentID=0, Status='live', Symbol='CG7593', Aliases='Dmel_ Dmel\\CG7593', description='CG7593 gene product from transcript CG7593-RA', other_designations='CG7593-PA|Ubiquinol-cytochrome C reductase complex 11 kDa protein', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29724159, end_position_on_the_genomic_accession=29724950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43501, CurrentID=0, Status='live', Symbol='Kul', Aliases='Dmel_CG1964, ADAM10, CG1964, CT3146, Dmel\\CG1964', description='Kuzbanian-like', other_designations='CG1964-PA|CG1964-PB|CG1964-PC|Kul-PA|Kul-PB|Kul-PC', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29725830, end_position_on_the_genomic_accession=29735360, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43502, CurrentID=0, Status='live', Symbol='CG7601', Aliases='Dmel_ BcDNA:GH06026, Dmel\\CG7601', description='CG7601 gene product from transcript CG7601-RA', other_designations='CG7601-PA', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29735269, end_position_on_the_genomic_accession=29736745, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43503, CurrentID=0, Status='live', Symbol='CIA30', Aliases='Dmel_CG7598, CG7598, CI30, Dmel\\CG7598, dCIA30', description='Complex I intermediate-associated protein, 30 kDa', other_designations='CG7598-PA|CIA30-PA|complex I intermediate- associate protein 30|complex I intermediate-associated protein 30|complex I intermediate-associated protein, 30 kDa', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29736965, end_position_on_the_genomic_accession=29738070, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43504, CurrentID=0, Status='live', Symbol='CG1969', Aliases='Dmel_ Dmel\\CG1969', description='CG1969 gene product from transcript CG1969-RG', other_designations='CG1969-PA|CG1969-PB|CG1969-PD|CG1969-PE|CG1969-PF|CG1969-PG|CG1969-PH', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29738275, end_position_on_the_genomic_accession=29741952, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43505, CurrentID=0, Status='live', Symbol='CG7609', Aliases='Dmel_ BcDNA:LD21720, Dmel\\CG7609, FBtr0085543, SEA2', description='CG7609 gene product from transcript CG7609-RB', other_designations='CG7609-PB|CG7609-PC', map_location='99B9-99B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29741972, end_position_on_the_genomic_accession=29745092, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43506, CurrentID=0, Status='live', Symbol='IntS11', Aliases='Dmel_CG1972, CG1972, Dmel\\CG1972, Int11, dRC-68', description='Integrator 11', other_designations='CG1972-PA|IntS11-PA|integrator 11', map_location='99B10-99B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29745337, end_position_on_the_genomic_accession=29747513, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43507, CurrentID=0, Status='live', Symbol='ATPsyngamma', Aliases='Dmel_CG7610, ATPG, ATPG_DROME, ATPsyn-gam, ATPsyn-gamma, CG7610, Dmel\\CG7610', description='ATP synthase, gamma subunit', other_designations='ATP synthase gamma chain|ATP synthase subunit gamma|ATP synthase-gamma chain|ATPsyn-gamma chain|ATPsyngamma-PA|ATPsyngamma-PB|ATPsyngamma-PC|CG7610-PA|CG7610-PB|CG7610-PC|F0/F1 ATP synthase gamma|F1ATP synthase|catalytic core of F1 ATP synthase of complex V', map_location='99B10-99B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29747704, end_position_on_the_genomic_accession=29749163, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43508, CurrentID=0, Status='live', Symbol='yata', Aliases='Dmel_CG1973, CG1973, Dmel\\CG1973, Yata', description='CG1973 gene product from transcript CG1973-RB', other_designations='CG1973-PA|CG1973-PB|yata-PA|yata-PB', map_location='99B10-99B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29749219, end_position_on_the_genomic_accession=29754528, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43511, CurrentID=0, Status='live', Symbol='fig', Aliases='Dmel_CG7615, CG7615, Dmel\\CG7615', description='fos intronic gene', other_designations='CG7615-PA|fig-PA|fos intron gene|fos-intronic gene', map_location='99C1-99C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29779273, end_position_on_the_genomic_accession=29780452, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43512, CurrentID=0, Status='live', Symbol='sro', Aliases='Dmel_CG12068, CG12068, Dmel\\CG12068', description='shroud', other_designations='CG12068-PA|sro-PA', map_location='3-100 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29794597, end_position_on_the_genomic_accession=29795805, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43513, CurrentID=0, Status='live', Symbol='CG7837', Aliases='Dmel_ Dmel\\CG7837', description='CG7837 gene product from transcript CG7837-RA', other_designations='CG7837-PA', map_location='99C1-99C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29795697, end_position_on_the_genomic_accession=29799354, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43514, CurrentID=0, Status='live', Symbol='Drice', Aliases='Dmel_CG7788, CG7788, DRICE, Dmel\\CG7788, DrICE, DrIce, Drive, ICE, Ice, caspase 3, crice, drICE, drIce, drice, ice', description='Death related ICE-like caspase', other_designations='CG7788-PA|Drice-PA|Ice|caspase-3|death related ICE-like caspase|interleukin-1beta-converting enzyme', map_location='99C1-99C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29799658, end_position_on_the_genomic_accession=29801584, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43515, CurrentID=0, Status='live', Symbol='CG7834', Aliases='Dmel_ CG 7834, Dmel\\CG7834, ETFB', description='CG7834 gene product from transcript CG7834-RB', other_designations='CG7834-PA|CG7834-PB|electron transfer flavoprotein beta subunit|electron transfer flavoprotein beta-subunit', map_location='99C1-99C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29801599, end_position_on_the_genomic_accession=29802945, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43516, CurrentID=0, Status='live', Symbol='CG7789', Aliases='Dmel_ Dmel\\CG7789', description='CG7789 gene product from transcript CG7789-RA', other_designations='CG7789-PA', map_location='99C1-99C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29803196, end_position_on_the_genomic_accession=29804389, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43517, CurrentID=0, Status='live', Symbol='ncd', Aliases='Dmel_CG7831, CG7831, DmNCD, DmNcd, Dmel\\CG7831, KIF C1, KIFC1, NCD, Ncd, ca, ca[nd], cd, l(3)05884, ndc', description='non-claret disjunctional', other_designations='CG7831-PA|CG7831-PB|Non-Claret disjunctional|claret nondisjunctional|kinesin-14|ncd-PA|ncd-PB|non-claret disjunction|non-claret-disjunctional|nonclaret disjunction|nonclaret disjunctional', map_location='3-100.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29804290, end_position_on_the_genomic_accession=29807127, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43518, CurrentID=0, Status='live', Symbol='ca', Aliases='Dmel_CG31037, CG18133, CG18134, CG31037, CG7799, Dmel\\CG31037, bw-p, bw[G]', description='claret', other_designations='CG31037-PA|CG31037-PC|Rab32 GEF|ca-PA|ca-PC', map_location='3-100.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29806811, end_position_on_the_genomic_accession=29815047, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19835392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835392, CurrentID=0, Status='live', Symbol='CG44475', Aliases='Dmel_ Dmel\\CG44475', description='CG44475 gene product from transcript CG44475-RA', other_designations='CG44475-PA|uncharacterized protein', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15882783, end_position_on_the_genomic_accession=15883190, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43521, CurrentID=0, Status='live', Symbol='Vps16B', Aliases='Dmel_CG18112, CG18112, Dmel\\CG18112, Vps16b, dVps16B, fob', description='Vacuolar protein sorting 16B', other_designations='CG18112-PA|Vps16B-PA|full-of-bacteria|vacuolar protein sorting 16B', map_location='99C2-99C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29814921, end_position_on_the_genomic_accession=29816451, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43522, CurrentID=0, Status='live', Symbol='CG7829', Aliases='Dmel_ Dmel\\CG7829, SP136', description='CG7829 gene product from transcript CG7829-RA', other_designations='CG7829-PA', map_location='99C2-99C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29816738, end_position_on_the_genomic_accession=29817764, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43523, CurrentID=0, Status='live', Symbol='neo', Aliases='Dmel_CG7802, CG7802, Dmel\\CG7802', description='neyo', other_designations='CG7802-PA|CG7802-PB|neo-PA|neo-PB', map_location='99C2-99C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29819131, end_position_on_the_genomic_accession=29825800, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43526, CurrentID=0, Status='live', Symbol='CG15510', Aliases='Dmel_ BcDNA:AT14346, Dmel\\CG15510', description='CG15510 gene product from transcript CG15510-RB', other_designations='CG15510-PB|CG15510-PC|CG15510-PD|uncharacterized protein', map_location='99C2-99C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29835950, end_position_on_the_genomic_accession=29837447, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43528, CurrentID=0, Status='live', Symbol='Cad99C', Aliases='Dmel_CG31009, CA-10, CG15511, CG310034, CG31009, CG7805, CT35624, CT4930, Cad99c, DCad99C, Dm Cad99C, Dmel\\CG31009, cad99C', description='Cadherin 99C', other_designations='CG31009-PA|CG31009-PB|Cad99C-PA|Cad99C-PB|Cadherin99C|cadherin 99C', map_location='99C4-99C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29841657, end_position_on_the_genomic_accession=29854699, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  43529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43529, CurrentID=0, Status='live', Symbol='dgt1', Aliases='Dmel_CG18041, CG18041, Dgt1, Dmel\\CG18041, NSL2', description='dim gamma-tubulin 1', other_designations='CG18041-PA|CG18041-PB|CG18041-PC|dgt1-PA|dgt1-PB|dgt1-PC|non-specific lethal 2', map_location='99C4-99C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29854695, end_position_on_the_genomic_accession=29857491, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43530, CurrentID=0, Status='live', Symbol='CG7824', Aliases='Dmel_ Dmel\\CG7824', description='CG7824 gene product from transcript CG7824-RC', other_designations='CG7824-PA|CG7824-PC', map_location='99C4-99C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29857652, end_position_on_the_genomic_accession=29858936, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43531, CurrentID=0, Status='live', Symbol='CG15514', Aliases='Dmel_ Dmel\\CG15514', description='CG15514 gene product from transcript CG15514-RB', other_designations='CG15514-PA|CG15514-PB|uncharacterized protein', map_location='99C4-99C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29860055, end_position_on_the_genomic_accession=29861494, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43532, CurrentID=0, Status='live', Symbol='RpS8', Aliases='Dmel_CG7808, BcDNA:RH06886, BcDNA:SD17528, CG7808, Dmel\\CG7808, Rp S8, Rps8, S8, anon-EST:Liang-1.56, anon-EST:fe1A2, chr3R:25688477..25688608, clone 1.56', description='Ribosomal protein S8', other_designations='CG7808-PC|CG7808-PE|CG7808-PF|CG7808-PG|RpS8-PC|RpS8-PE|RpS8-PF|RpS8-PG|anon-fast-evolving-1A2|ribosomal protein S8', map_location='99C4-99C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29861582, end_position_on_the_genomic_accession=29863902, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43533, CurrentID=0, Status='live', Symbol='Zip99C', Aliases='Dmel_CG7816, CG7816, Dmel\\CG7816, cg7816, dZIP13, dZip99C', description='Zinc/iron regulated transporter-related protein 99C', other_designations='CG7816-PA|CG7816-PC|CG7816-PD|CG7816-PE|CG7816-PF|CG7816-PG|CG7816-PH|CG7816-PI|CG7816-PJ|Zip99C-PA|Zip99C-PC|Zip99C-PD|Zip99C-PE|Zip99C-PF|Zip99C-PG|Zip99C-PH|Zip99C-PI|Zip99C-PJ', map_location='99C4-99C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29863701, end_position_on_the_genomic_accession=29868895, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43535, CurrentID=0, Status='live', Symbol='CG31038', Aliases='Dmel_ CG15512, CG7854, Dmel\\CG31038', description='CG31038 gene product from transcript CG31038-RA', other_designations='CG31038-PA|CG31038-PB|CG31038-PC|CG31038-PD|CG31038-PE|CG31038-PG|CG31038-PH|uncharacterized protein', map_location='99C5-99C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29877059, end_position_on_the_genomic_accession=29903231, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43537, CurrentID=0, Status='live', Symbol='CG15517', Aliases='Dmel_ Dmel\\CG15517', description='CG15517 gene product from transcript CG15517-RA', other_designations='CG15517-PA|uncharacterized protein', map_location='99C6-99C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29903240, end_position_on_the_genomic_accession=29903825, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43538, CurrentID=0, Status='live', Symbol='CG15515', Aliases='Dmel_ BcDNA:RE21592, Dmel\\CG15515', description='CG15515 gene product from transcript CG15515-RB', other_designations='CG15515-PB|uncharacterized protein', map_location='99C6-99C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29905464, end_position_on_the_genomic_accession=29906448, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43539, CurrentID=0, Status='live', Symbol='CG31036', Aliases='Dmel_ CG15516, CG15518, Dmel\\CG31036', description='CG31036 gene product from transcript CG31036-RA', other_designations='CG31036-PA|CG31036-PD|uncharacterized protein', map_location='99C6-99C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29908107, end_position_on_the_genomic_accession=29912892, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43541, CurrentID=0, Status='live', Symbol='Capa', Aliases='Dmel_CG15520, CAP, CAP-2b, CAP2B-1, CAP2B-2, CAP2B-3, CAP2B/pyrokinin, CAP2b, CAP2b-1, CAP2b-2, CAP2b-3, CAPA, CAPA-PK, CAPA-PVK-1, CAPA-PVK-2, CAPA-PVK/PK, CAP[[2b]], CG15520, CPPB, Cap-1, Cap-2, Cap-2b, Cap-3-1, Capa-2, Dmel\\CG15520, Drm-CAP-1, Drm-CAP-2, Drm-CAP2b/MT, Drm-MT, Drm-PK, Drm-PK-1, Drm-PVK, Drm-PVK-1, Drm-PVK-2, Drm-capa-1, PVK-2, capa, capa 2, capa-1, capa-2, capa-3, pyrokinin/PBAN-like, Capa', description='Capability', other_designations='CAP|CAP2B-like|CG15520-PA|Capa-PA|Drm-myotropin|TGPSASSGLWFGPRLamide|capability|cardio acceleratory peptide|cardioacceleratory peptide-2b|myotropin|myotropin-CAP2b-like|periviscerokinin|pyrokinin|pyrokinin-1|pyrokinin-like', map_location='99C6-99C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29912991, end_position_on_the_genomic_accession=29914346, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43542, CurrentID=0, Status='live', Symbol='trp', Aliases='Dmel_CG7875, 559, CG7875, DmTRP, Dmel\\CG7875, TRP, Trp, dTRP, dTrp, gene B, idl-B', description='transient receptor potential', other_designations='CG7875-PA|CG7875-PB|Cosens-Manning mutant|INAD localization-B|transient photreceptor potential|transient receptor potential protein|trp-PA|trp-PB', map_location='99C6-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29914521, end_position_on_the_genomic_accession=29920289, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  43543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43543, CurrentID=0, Status='live', Symbol='Jon99Ciii', Aliases='Dmel_CG31362, 99Calpha, CG15519, CG31362, Dmel\\CG31362, Jon99B, Jon99C, Jon99C1, Jon99C2, Jon99Calpha, Jon99Calphaii, Jon99Cbeta, Jon99Cbetaiii, JonC, SER2, SP143, Ser2, Ser5, Ser99Db', description='Jonah 99Ciii', other_designations='CG31362-PA|CG31362-PB|Jon99Ciii-PA|Jon99Ciii-PB|jonah 99C alpha|jonah 99C beta|jonah 99Ciii|serine protease 2|serine protease 5', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29922216, end_position_on_the_genomic_accession=29923455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43544, CurrentID=0, Status='live', Symbol='Jon99Cii', Aliases='Dmel_CG31034, 99Calpha, CG31034, CG7877, Dmel\\CG31034, Jon99B, Jon99C, Jon99C1, Jon99C2, Jon99Calpha, Jon99Calphai, Jon99Cbeta, Jon99Cbetaii, JonC, SER1, SP122, Ser1, Ser5, Ser99Da, Serl', description='Jonah 99Cii', other_designations='CG31034-PA|Jon99Cii-PA|jonah 99C alpha|jonah 99C beta|jonah 99Cii|serine protease 1|serine protease 5', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29923754, end_position_on_the_genomic_accession=29924615, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43545, CurrentID=0, Status='live', Symbol='Jon99Ci', Aliases='Dmel_CG31039, 143341_at, 99Calpha, CG17951, CG31039, Dmel\\CG31039, Jon99B, Jon99C, Jon99C1, Jon99C2, Jon99Calpha, Jon99Calphai, Jon99Cbeta, Jon99Cbetai, JonC, SER3, SP150, Ser3, Ser99Dc, ser99Dc', description='Jonah 99Ci', other_designations='CG31039-PA|Jon99Ci-PA|jonah 99C alpha|jonah 99C beta|jonah 99Ci|serine protease|serine protease 3', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29925226, end_position_on_the_genomic_accession=29926189, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43546, CurrentID=0, Status='live', Symbol='CG15522', Aliases='Dmel_ Dmel\\CG15522', description='CG15522 gene product from transcript CG15522-RB', other_designations='CG15522-PA|CG15522-PB|uncharacterized protein', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29927255, end_position_on_the_genomic_accession=29928766, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43547, CurrentID=0, Status='live', Symbol='Cog7', Aliases='Dmel_CG31040, CG31040, CG7880, Dmel\\CG31040', description='CG31040 gene product from transcript CG31040-RA', other_designations='CG31040-PA|Cog7-PA', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29930551, end_position_on_the_genomic_accession=29933057, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43548, CurrentID=0, Status='live', Symbol='CG42557', Aliases='Dmel_ CG7866, Dmel\\CG42557, Dmel_CG7866', description='CG42557 gene product from transcript CG42557-RA', other_designations='CG42557-PA|uncharacterized protein', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29933287, end_position_on_the_genomic_accession=29934587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43549, CurrentID=0, Status='live', Symbol='eIF2B-alpha', Aliases='Dmel_CG7883, CG7883, DeIF2Balpha, Dmel\\CG7883', description='CG7883 gene product from transcript CG7883-RA', other_designations='CG7883-PA|eIF2B-alpha-PA', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29934582, end_position_on_the_genomic_accession=29935888, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43550, CurrentID=0, Status='live', Symbol='CG15523', Aliases='Dmel_ Dmel\\CG15523, Eif2b', description='CG15523 gene product from transcript CG15523-RA', other_designations='CG15523-PA|CG15523-PB|uncharacterized protein', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29935978, end_position_on_the_genomic_accession=29948207, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43551, CurrentID=0, Status='live', Symbol='TkR99D', Aliases='Dmel_CG7887, CG7887, DTKR, DmTRPR, Dmel\\CG7887, Dtkr, TAKR99D, TKR99D, TLR2_DROME, Tak-r, Takr99D, anon-WO0131005.11, anon-WO0131005.9, dtkr, l(3)s2222', description='Tachykinin-like receptor at 99D', other_designations='CG7887-PA|CG7887-PB|CG7887-PC|DTK receptor|TkR99D-PA|TkR99D-PB|TkR99D-PC|drosophila tachykinin receptor|tachykinin receptor', map_location='99C8-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29950796, end_position_on_the_genomic_accession=29969447, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43552, CurrentID=0, Status='live', Symbol='CG7896', Aliases='Dmel_ CT23878, Dmel\\CG7896, tartan/capricious-like', description='CG7896 gene product from transcript CG7896-RA', other_designations='CG7896-PA', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29983426, end_position_on_the_genomic_accession=29990785, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43554, CurrentID=0, Status='live', Symbol='CG7903', Aliases='Dmel_ Dmel\\CG7903, cg7903', description='CG7903 gene product from transcript CG7903-RB', other_designations='CG7903-PA|CG7903-PB', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29993475, end_position_on_the_genomic_accession=29995231, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43555, CurrentID=0, Status='live', Symbol='Sas-6', Aliases='Dmel_CG15524, CG15524, DSAS-6, DSas-6, DSas6, Dmel\\CG15524, Dsas6, SAS-6, SAS6, Sas6, dsas-6, sas-6', description='Spindle assembly abnormal 6 ortholog (C. elegans)', other_designations='CG15524-PA|Sas-6-PA|spindle assembly abnormal 6|spindle assembly abnormal 6 ortholog', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29995357, end_position_on_the_genomic_accession=29997034, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43556, CurrentID=0, Status='live', Symbol='CG15525', Aliases='Dmel_ BcDNA:SD23244, Dmel\\CG15525', description='CG15525 gene product from transcript CG15525-RA', other_designations='CG15525-PA|uncharacterized protein', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29996961, end_position_on_the_genomic_accession=29997676, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43557, CurrentID=0, Status='live', Symbol='CG11504', Aliases='Dmel_ Dmel\\CG11504', description='CG11504 gene product from transcript CG11504-RB', other_designations='CG11504-PA|CG11504-PB|CG11504-PC|uncharacterized protein', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29997900, end_position_on_the_genomic_accession=29999755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43558, CurrentID=0, Status='live', Symbol='Tace', Aliases='Dmel_CG7908, CG7908, CT23908, DTACE, Dmel\\CG7908, TACE', description='CG7908 gene product from transcript CG7908-RB', other_designations='CG7908-PA|CG7908-PB|Tace-PA|Tace-PB|tace|tumor necrosis factor-alpha-converting enzyme', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30000530, end_position_on_the_genomic_accession=30003678, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43559, CurrentID=0, Status='live', Symbol='Nph', Aliases='Dmel_CG7911, CG7911, Dmel\\CG7911', description='Nucleophosmin', other_designations='CG7911-PA|CG7911-PB|Nph-PA|Nph-PB|nucleophosmin|nucleoplasmin', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30003681, end_position_on_the_genomic_accession=30004861, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43560, CurrentID=0, Status='live', Symbol='Nlp', Aliases='Dmel_CG7917, CG7917, CRP1, Crp1, Dmel\\CG7917, NLP, dNLP, dNLP-S, nlp, p22', description='Nucleoplasmin', other_designations='CG7917-PA|CG7917-PB|Nlp-PA|Nlp-PB|Nucleoplasmin-like|chromatin decondensation protein 1|nucleophosmin|nucleoplasmin|nucleoplasmin-1', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30004785, end_position_on_the_genomic_accession=30010352, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43561, CurrentID=0, Status='live', Symbol='CG7912', Aliases='Dmel_ Dmel\\CG7912', description='CG7912 gene product from transcript CG7912-RA', other_designations='CG7912-PA', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30006034, end_position_on_the_genomic_accession=30009189, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43562, CurrentID=0, Status='live', Symbol='CG7920', Aliases='Dmel_ BEST:GH02390, Dmel\\CG7920, GH02390.5prime, anon-WO0153538.36, anon-WO0153538.37, i130', description='CG7920 gene product from transcript CG7920-RA', other_designations='4-hydroxybutyrate Co-A transferase|CG7920-PA|CG7920-PC', map_location='99D1-99D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30010476, end_position_on_the_genomic_accession=30015284, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43563, CurrentID=0, Status='live', Symbol='Mgat2', Aliases='Dmel_CG7921, CG7921, Dmel\\CG7921, GlcNAc-TII, GlcNAcT2, dMGAT2', description='CG7921 gene product from transcript CG7921-RB', other_designations='CG7921-PA|CG7921-PB|CG7921-PC|Mgat2-PA|Mgat2-PB|Mgat2-PC', map_location='99D1-99D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30015513, end_position_on_the_genomic_accession=30022221, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43564, CurrentID=0, Status='live', Symbol='RpS28a', Aliases='Dmel_CG15527, CG15527, Dmel\\CG15527', description='Ribosomal protein S28a', other_designations='CG15527-PA|RpS28a-PA|ribosomal protein S28a', map_location='99D2-99D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30022345, end_position_on_the_genomic_accession=30022648, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43565, CurrentID=0, Status='live', Symbol='Axn', Aliases='Dmel_CG7926, 0442/30, APC, CG7926, D-Axin, DAxin, Daxin, Dmel\\CG7926, GSK3beta, axin, axn, d-axin, dAXIN, dAxin, din, l(3)S044230', description='Axin', other_designations='Axn-PA|Axn-PB|Axn-PC|Axn-PD|Axn-PE|Axn-PF|Axn-PG|CG7926-PA|CG7926-PB|CG7926-PC|CG7926-PD|CG7926-PE|CG7926-PF|CG7926-PG|D-axin|Dm axin|axin|axn|dAxin', map_location='99D2-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30022842, end_position_on_the_genomic_accession=30035311, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43566, CurrentID=0, Status='live', Symbol='ZIPIC', Aliases='Dmel_CG7928, CG7928, Dmel\\CG7928', description='Zinc-finger protein interacting with CP190', other_designations='CG7928-PA|ZIPIC-PA', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30035240, end_position_on_the_genomic_accession=30036870, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43567, CurrentID=0, Status='live', Symbol='ocn', Aliases='Dmel_CG7929, BcDNA:GH02250, CG7929, Dmel\\CG7929, Ocn, bs29e10.y1', description='ocnus', other_designations='CG7929-PA|ocn-PA', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30037003, end_position_on_the_genomic_accession=30037730, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43568, CurrentID=0, Status='live', Symbol='janB', Aliases='Dmel_CG7931, CG31024, CG7931, Dmel\\CG7931, Jan, Jan-B, bs30f02.y1, jan, jan B, janus', description='janus B', other_designations='CG7931-PA|janB-PA|janus|janusB', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30037984, end_position_on_the_genomic_accession=30038728, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43569, CurrentID=0, Status='live', Symbol='janA', Aliases='Dmel_CG7933, BcDNA:AT12574, CG7933, Dmel\\CG7933, Jan, Y, jan, jan A, janus', description='janus A', other_designations='CG7933-PA|CG7933-PB|CG7933-PC|janA-PA|janA-PB|janA-PC|janus', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30038622, end_position_on_the_genomic_accession=30039533, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43570, CurrentID=0, Status='live', Symbol='Sry-beta', Aliases='Dmel_CG7938, CG7938, Dmel\\CG7938, Srybeta, ser, sry, sry beta, sry-b, sry-beta, sryb, srybeta', description='Serendipity beta', other_designations='CG7938-PA|CG7938-PB|Sry-beta-PA|Sry-beta-PB|serendipity|serendipity beta', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30039719, end_position_on_the_genomic_accession=30043071, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43571, CurrentID=0, Status='live', Symbol='Sry-alpha', Aliases='Dmel_CG17957, CG17957, Dmel\\CG17957, Sry alpha, Sryalpha, ser, sry, sry alpha, sry-a, sry-alpha, srya, sryalpha', description='Serendipity alpha', other_designations='CG17957-PA|CG17957-PB|Sry-alpha-PA|Sry-alpha-PB|serendipity|serendipity alpha|serendipity-alpha', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30039719, end_position_on_the_genomic_accession=30043071, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43572, CurrentID=0, Status='live', Symbol='Sry-delta', Aliases='Dmel_CG17958, CG17958, Dmel\\CG17958, EH8, Sry delta, Srydelta, anon-EH8, sry, sry delta, sry-delta, srydelta', description='Serendipity delta', other_designations='CG17958-PA|Sry-delta-PA|serendipity delta|serendipity-delta', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30043341, end_position_on_the_genomic_accession=30044886, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43573, CurrentID=0, Status='live', Symbol='RpL32', Aliases='Dmel_CG7939, 143250_at, BcDNA:RH03940, CG7939, Dmel\\CG7939, L32, M(3)99D, RP-49, RP49, RPL32, Rp-49, Rp49, Rp49/RpL32, Rpl32, bs30a02.y1, rp49, rp49/RpL32', description='Ribosomal protein L32', other_designations='CG7939-PA|CG7939-PB|CG7939-PC|CG7939-PD|Minute(3)99D|RpL32-PA|RpL32-PB|RpL32-PC|RpL32-PD|r-protein 49|ribosomal protein 49|ribosomal protein 49/L32|ribosomal protein L32', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30045229, end_position_on_the_genomic_accession=30046161, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43574, CurrentID=0, Status='live', Symbol='CG7943', Aliases='Dmel_ Dmel\\CG7943', description='CG7943 gene product from transcript CG7943-RB', other_designations='CG7943-PA|CG7943-PB', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30046470, end_position_on_the_genomic_accession=30048665, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43575, CurrentID=0, Status='live', Symbol='CG15528', Aliases='Dmel_ Dmel\\CG15528', description='CG15528 gene product from transcript CG15528-RB', other_designations='CG15528-PB|MKP-like|uncharacterized protein', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30048635, end_position_on_the_genomic_accession=30052112, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43576, CurrentID=0, Status='live', Symbol='CG7946', Aliases='Dmel_ Dmel\\CG7946', description='CG7946 gene product from transcript CG7946-RA', other_designations='CG7946-PA|CG7946-PB', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30052514, end_position_on_the_genomic_accession=30054697, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43577, CurrentID=0, Status='live', Symbol='spn-A', Aliases='Dmel_CG7948, CG7948, D37788, DMR, DMR/DroRAD51, DMR1, Dm Rad51, DmRAD51, DmRad51, Dmel\\CG7948, Dmrad51, RA51, RA51_DROME, RAD51, Rad51, Rad51dm, Spn-A, SpnA, dmRAD51, dmrad51, rad51, spnA', description='spindle A', other_designations='CG7948-PA|CG7948-PB|Rad51-like|Spindle-A|spn-A-PA|spn-A-PB', map_location='3-96 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30054696, end_position_on_the_genomic_accession=30056037, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43578, CurrentID=0, Status='live', Symbol='CG15526', Aliases='Dmel_ Dmel\\CG15526', description='CG15526 gene product from transcript CG15526-RA', other_designations='CG15526-PA|uncharacterized protein', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30055957, end_position_on_the_genomic_accession=30056603, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43579, CurrentID=0, Status='live', Symbol='CG7950', Aliases='Dmel_ Dmel\\CG7950, Dromel_CG7950_FBtr0085563_CHCHD4inmorf_mORF', description='CG7950 gene product from transcript CG7950-RB', other_designations='CG7950-PA|CG7950-PB', map_location='99D3-99D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30056859, end_position_on_the_genomic_accession=30058131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43580, CurrentID=0, Status='live', Symbol='sima', Aliases='Dmel_CG45051, 7951, CG31031, CG45051, CG7951, DMU43090, Dmel\\CG45051, Dmel_CG31031, Dmel_CG7951, HIF alpha, HIF-1, HIF-1alpha, HIF-alpha/sima, Hif-1A, Hif1alpha, SIMA, Sima, bHLHe56, dHIF-1alpha, dHIF-alpha, l(3)j11B7', description='similar', other_designations='CG45051-PA|CG45051-PB|CG45051-PC|CG45051-PD|HIF alpha|HIF-alpha|hypoxia inducible factor|hypoxia inducible factor ortholog|hypoxia-inducible factor-1|sima-PA|sima-PB|sima-PC|sima-PD', map_location='99D4-99D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30058311, end_position_on_the_genomic_accession=30121798, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43581, CurrentID=0, Status='live', Symbol='Ppi1', Aliases='Dmel_CG31025, CG2164, CG31025, Dmel\\CG31025, anon-WO0140519.94, ppi1', description='Protein phosphatase 1c interacting protein 1', other_designations='CG31025-PA|CG31025-PB|Ppi1-PA|Ppi1-PB|protein phosphatase 1c interacting protein 1', map_location='99D7-99D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30125633, end_position_on_the_genomic_accession=30129119, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43582, CurrentID=0, Status='live', Symbol='Tpi', Aliases='Dmel_CG2171, CG2171, Dmel\\CG2171, ND14, TPI, TPIS, sgk, tpi, wstd', description='Triose phosphate isomerase', other_designations='CG2171-PA|CG2171-PB|CG2171-PC|Tpi-PA|Tpi-PB|Tpi-PC|neurodegeneration 14|sugarkill|triose phosphate isomerase|triosephosphate isomerase|wasted away', map_location='3-101.3 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30132981, end_position_on_the_genomic_accession=30134599, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43583, CurrentID=0, Status='live', Symbol='AdoR', Aliases='Dmel_CG9753, CG9753, DmAdoR, Dmel\\CG9753, adoR, anon-WO0170980.109, anon-WO0170980.110, dADOR, dAdoR', description='Adenosine receptor', other_designations='AdoR-PA|AdoR-PB|CG9753-PA|CG9753-PB|adenosine receptor', map_location='99D8-99D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30134856, end_position_on_the_genomic_accession=30149606, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43584, CurrentID=0, Status='live', Symbol='CG15529', Aliases='Dmel_ Dmel\\CG15529', description='CG15529 gene product from transcript CG15529-RB', other_designations='CG15529-PA|CG15529-PB|uncharacterized protein', map_location='99D9-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30152059, end_position_on_the_genomic_accession=30153699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43585, CurrentID=0, Status='live', Symbol='CG11498', Aliases='Dmel_ Dmel\\CG11498', description='CG11498 gene product from transcript CG11498-RC', other_designations='CG11498-PB|CG11498-PC|uncharacterized protein', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30153022, end_position_on_the_genomic_accession=30160473, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43586, CurrentID=0, Status='live', Symbol='CG31028', Aliases='Dmel_ CG2176, Dmel\\CG31028', description='CG31028 gene product from transcript CG31028-RC', other_designations='CG31028-PC|uncharacterized protein', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30166631, end_position_on_the_genomic_accession=30171207, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43587, CurrentID=0, Status='live', Symbol='Mlc2', Aliases='Dmel_CG2184, CG2184, DmMLC2, Dmel\\CG2184, Dmlc, Dmlc2, Ifm(3)99Eb, LC2, MLC, MLC 2, MLC-2, MLC2, MRLC, MYL2, RLC, RLC2, dro MLC-2, l(3)99Eb, mlc-2, mlc2', description='Myosin light chain 2', other_designations='CG2184-PA|Mlc2-PA|muscle-specific myosin regulatory light chain|myosin LC2|myosin alkali light chain 2|myosin light Chain-2|myosin light chain 2|myosin light-chain 2|myosin regulatory light chain', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30171438, end_position_on_the_genomic_accession=30174551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43588, CurrentID=0, Status='live', Symbol='CG1983', Aliases='Dmel_ Dmel\\CG1983', description='CG1983 gene product from transcript CG1983-RB', other_designations='CG1983-PA|CG1983-PB', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30175786, end_position_on_the_genomic_accession=30177700, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43589, CurrentID=0, Status='live', Symbol='CG15530', Aliases='Dmel_ Dmel\\CG15530', description='CG15530 gene product from transcript CG15530-RB', other_designations='CG15530-PA|CG15530-PB|CG15530-PC|uncharacterized protein', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30177877, end_position_on_the_genomic_accession=30180178, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43590, CurrentID=0, Status='live', Symbol='Bet5', Aliases='Dmel_CG1359, BET5p, CG1359, Dmel\\CG1359, Bet5', description='BET5 ortholog (S. cerevisiae)', other_designations='BET5 ortholog|Bet5-PA|CG1359-PA', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30180232, end_position_on_the_genomic_accession=30181486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43591, CurrentID=0, Status='live', Symbol='CG9747', Aliases='Dmel_ Dmel\\CG9747', description='CG9747 gene product from transcript CG9747-RB', other_designations='CG9747-PA|CG9747-PB', map_location='99E1-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30185423, end_position_on_the_genomic_accession=30191252, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43592, CurrentID=0, Status='live', Symbol='CG15531', Aliases='Dmel_ Dmel\\CG15531', description='CG15531 gene product from transcript CG15531-RA', other_designations='CG15531-PA|uncharacterized protein', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30194790, end_position_on_the_genomic_accession=30196228, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43593, CurrentID=0, Status='live', Symbol='CG9743', Aliases='Dmel_ Dmel\\CG9743, anon-WO0118547.264', description='CG9743 gene product from transcript CG9743-RA', other_designations='CG9743-PA', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30196360, end_position_on_the_genomic_accession=30202998, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43594, CurrentID=0, Status='live', Symbol='RpS7', Aliases='Dmel_CG1883, CG1883, Dmel\\CG1883', description='Ribosomal protein S7', other_designations='CG1883-PA|CG1883-PC|CG1883-PD|CG1883-PE|RpS7-PA|RpS7-PC|RpS7-PD|RpS7-PE|ribosomal protein S7', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30206447, end_position_on_the_genomic_accession=30209020, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43595, CurrentID=0, Status='live', Symbol='Anp', Aliases='Dmel_CG1361, And, CEC, CG1361, Cec, Dmel\\CG1361, k-9', description='Andropin', other_designations='Anp-PA|Anp-PB|CG1361-PA|CG1361-PB|andropin', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30209948, end_position_on_the_genomic_accession=30210382, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43596, CurrentID=0, Status='live', Symbol='CecA1', Aliases='Dmel_CG1365, CEC, CG1365, Cec, Cec A1, Cec-A1, CecA, DIM 25, Dmel\\CG1365, cec, cecA, cecA1, crpA1', description='Cecropin A1', other_designations='CG1365-PA|CecA1-PA|Cecropin-A1/A2 precursor|CecropinA|cecropin|cecropin A|cecropin A1|cecropin-A|cecropin-A1|cecropinA1', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30210874, end_position_on_the_genomic_accession=30211273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43597, CurrentID=0, Status='live', Symbol='CecA2', Aliases='Dmel_CG1367, CG1367, Cec, Cec A2, CecA, Dmel\\CG1367, cec, cecA, cecA2, crpA2', description='Cecropin A2', other_designations='CG1367-PA|CecA2-PA|Cecropin-A|Cecropin-A1/A2 precursor|Cecropin-A2|cecropin|cecropin A|cecropin A2|cecropinA|cecropinA2', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30212156, end_position_on_the_genomic_accession=30212568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43598, CurrentID=0, Status='live', Symbol='CecB', Aliases='Dmel_CG1878, CG1878, Cec, Cec B, Dmel\\CG1878, cec, cecB, cedcB, crpB', description='Cecropin B', other_designations='CG1878-PA|CecB-PA|Cecropin-B|Cecropin-B precursor|cecropin|cecropin B|cercropin B', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30213315, end_position_on_the_genomic_accession=30213797, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43599, CurrentID=0, Status='live', Symbol='CecC', Aliases='Dmel_CG1373, BcDNA:RH33561, CG1373, Cec, Cec-C, Dmel\\CG1373, cec, cecC', description='Cecropin C', other_designations='CG1373-PA|CecC-PA|Cecropin-C|Cecropin-C precursor|cecropin|cecropin C', map_location='99E2-99E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30216484, end_position_on_the_genomic_accession=30216938, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43600, CurrentID=0, Status='live', Symbol='CG9737', Aliases='Dmel_ Dmel\\CG9737, SP5, c-SP5', description='CG9737 gene product from transcript CG9737-RA', other_designations='CG9737-PA', map_location='99E3-99E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30241268, end_position_on_the_genomic_accession=30243312, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43601, CurrentID=0, Status='live', Symbol='CG9733', Aliases='Dmel_ Dmel\\CG9733, SP10, c-SP10, proAE, proPO-AE', description='CG9733 gene product from transcript CG9733-RB', other_designations='CG9733-PA|CG9733-PB|POAE-like serine protease|pro-phenoloxidase AE', map_location='99E3-99E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30243978, end_position_on_the_genomic_accession=30246703, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43602, CurrentID=0, Status='live', Symbol='CG9682', Aliases='Dmel_ Dmel\\CG9682, anon-WO0140519.240', description='CG9682 gene product from transcript CG9682-RC', other_designations='CG9682-PB|CG9682-PC', map_location='99E3-99E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30250928, end_position_on_the_genomic_accession=30252830, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43603, CurrentID=0, Status='live', Symbol='CG18404', Aliases='Dmel_ Dmel\\CG18404, anon-WO0140519.106', description='CG18404 gene product from transcript CG18404-RA', other_designations='CG18404-PA|uncharacterized protein', map_location='99E3-99E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30253182, end_position_on_the_genomic_accession=30254097, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43604, CurrentID=0, Status='live', Symbol='hdc', Aliases='Dmel_CG15532, CG15532, Dmel\\CG15532, Hdc, l(3)B4-3-20, psg12', description='headcase', other_designations='CG15532-PA|CG15532-PB|CG15532-PC|CG15532-PD|Fusion-6|hdc-PA|hdc-PB|hdc-PC|hdc-PD', map_location='99E4-99F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30277932, end_position_on_the_genomic_accession=30372382, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43607, CurrentID=0, Status='live', Symbol='CG2217', Aliases='Dmel_ BcDNA:GH11688, Dmel\\CG2217', description='CG2217 gene product from transcript CG2217-RD', other_designations='CG2217-PA|CG2217-PB|CG2217-PC|CG2217-PD', map_location='99F3-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30418767, end_position_on_the_genomic_accession=30422971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43608, CurrentID=0, Status='live', Symbol='CG15535', Aliases='Dmel_ Dmel\\CG15535', description='CG15535 gene product from transcript CG15535-RB', other_designations='CG15535-PA|CG15535-PB|uncharacterized protein', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30423032, end_position_on_the_genomic_accession=30423929, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43609, CurrentID=0, Status='live', Symbol='mRpS18C', Aliases='Dmel_CG9688, CG9688, Dmel\\CG9688, MRP-S18-(1), mRpS18a', description='mitochondrial ribosomal protein S18C', other_designations='CG9688-PA|mRpS18C-PA|mitochondrial ribosomal protein S18a', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30423974, end_position_on_the_genomic_accession=30424556, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43610, CurrentID=0, Status='live', Symbol='CG15536', Aliases='Dmel_ Dmel\\CG15536', description='CG15536 gene product from transcript CG15536-RA', other_designations='CG15536-PA|uncharacterized protein', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30424499, end_position_on_the_genomic_accession=30425162, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43611, CurrentID=0, Status='live', Symbol='CG2218', Aliases='Dmel_ Dmel\\CG2218, UIP5', description='CG2218 gene product from transcript CG2218-RA', other_designations='BT044264.1|CG2218-PA|CG2218-PB', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30425427, end_position_on_the_genomic_accession=30428276, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43612, CurrentID=0, Status='live', Symbol='CG15533', Aliases='Dmel_ Dmel\\CG15533, aSMase', description='CG15533 gene product from transcript CG15533-RA', other_designations='CG15533-PA|acid sphingomyelinase|uncharacterized protein', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30428820, end_position_on_the_genomic_accession=30431362, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43613, CurrentID=0, Status='live', Symbol='CG15534', Aliases='Dmel_ Dmel\\CG15534, aSMase', description='CG15534 gene product from transcript CG15534-RA', other_designations='CG15534-PA|uncharacterized protein', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30432064, end_position_on_the_genomic_accession=30435139, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43614, CurrentID=0, Status='live', Symbol='CG15537', Aliases='Dmel_ Dmel\\CG15537', description='CG15537 gene product from transcript CG15537-RC', other_designations='CG15537-PB|CG15537-PC|uncharacterized protein', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30435177, end_position_on_the_genomic_accession=30443599, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43615, CurrentID=0, Status='live', Symbol='Osi23', Aliases='Dmel_CG15538, CG15538, Dmel\\CG15538', description='Osiris 23', other_designations='CG15538-PB|Osi23-PB|osiris|osiris 23', map_location='99F4-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30444076, end_position_on_the_genomic_accession=30445046, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43616, CurrentID=0, Status='live', Symbol='aralar1', Aliases='Dmel_CG2139, Aralar, BcDNA:GH07418, CG2139, DmAralar, Dmel\\CG2139, aralar', description='CG2139 gene product from transcript CG2139-RC', other_designations='Aralar1|CG2139-PA|CG2139-PB|CG2139-PC|CG2139-PD|CG2139-PE|CG2139-PF|aralar1-PA|aralar1-PB|aralar1-PC|aralar1-PD|aralar1-PE|aralar1-PF', map_location='99F4-99F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30445635, end_position_on_the_genomic_accession=30455867, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  43617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43617, CurrentID=0, Status='live', Symbol='CG2224', Aliases='Dmel_ Amsh, Dmel\\CG2224, Q9VA71', description='CG2224 gene product from transcript CG2224-RB', other_designations='CG2224-PA|CG2224-PB', map_location='99F5-99F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30456441, end_position_on_the_genomic_accession=30459300, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43618, CurrentID=0, Status='live', Symbol='CDase', Aliases='Dmel_CG1471, CG1471, Dmel\\CG1471, anon-WO0118547.407, cdase, fcd, slab', description='Ceramidase', other_designations='CDase-PA|CDase-PB|CDase-PC|CDase-PD|CDase-PE|CDase-PF|CG1471-PA|CG1471-PB|CG1471-PC|CG1471-PD|CG1471-PE|CG1471-PF|ceramidase|slug-a-bed', map_location='99F5-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30459521, end_position_on_the_genomic_accession=30465736, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43620, CurrentID=0, Status='live', Symbol='PH4alphaEFB', Aliases='Dmel_CG31022, BcDNA:SD05564, CG12088, CG1473, CG31022, Dmel\\CG31022, PH4-aEFB, PH4alpha-EFB', description='prolyl-4-hydroxylase-alpha EFB', other_designations='CG31022-PA|PH4alphaEFB-PA|propyl-4-hydroxylase-alpha EFB', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30466050, end_position_on_the_genomic_accession=30485799, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43621, CurrentID=0, Status='live', Symbol='Jon99Fii', Aliases='Dmel_CG2229, CG2229, Dmel\\CG2229, Jon99F, SP135', description='Jonah 99Fii', other_designations='CG2229-PA|Jon99Fii-PA|jonah 99F|jonah 99Fii', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30487016, end_position_on_the_genomic_accession=30487928, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43622, CurrentID=0, Status='live', Symbol='Jon99Fi', Aliases='Dmel_CG18030, CG18030, Dmel\\CG18030, Jon99F, SPH181', description='Jonah 99Fi', other_designations='CG18030-PA|Jon99Fi-PA|jonah 99F|jonah 99Fi', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30488963, end_position_on_the_genomic_accession=30489895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43623, CurrentID=0, Status='live', Symbol='PH4alphaSG2', Aliases='Dmel_CG1546, CG1546, Dmel\\CG1546, SG2', description='prolyl-4-hydroxylase-alpha SG2', other_designations='CG1546-PA|PH4alphaSG2-PA', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30490074, end_position_on_the_genomic_accession=30491987, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43624, CurrentID=0, Status='live', Symbol='PH4alphaMP', Aliases='Dmel_CG9726, 59D7T, CG9726, DM59D7T, Dmel\\CG9726, ESTS:59D7T', description='prolyl-4-hydroxylase-alpha MP', other_designations='CG9726-PA|PH4alphaMP-PA|alpha subunit|prolyl-4-hydroxylase-alpha', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30491813, end_position_on_the_genomic_accession=30496153, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43625, CurrentID=0, Status='live', Symbol='PH4alphaNE1', Aliases='Dmel_CG9728, CG9728, Dmel\\CG9728', description='prolyl-4-hydroxylase-alpha NE1', other_designations='CG9728-PA|PH4alphaNE1-PA', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30496751, end_position_on_the_genomic_accession=30500326, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43626, CurrentID=0, Status='live', Symbol='CG31371', Aliases='Dmel_ CG15542, Dmel\\CG31371', description='CG31371 gene product from transcript CG31371-RC', other_designations='CG31371-PB|CG31371-PC|uncharacterized protein', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30501124, end_position_on_the_genomic_accession=30503506, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43627, CurrentID=0, Status='live', Symbol='CG15539', Aliases='Dmel_CG15539-NE4, Dmel\\CG15539', description='CG15539 gene product from transcript CG15539-RA', other_designations='CG15539-PA|CG15539-PB|CG15539-PC|uncharacterized protein', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30505987, end_position_on_the_genomic_accession=30507977, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43628, CurrentID=0, Status='live', Symbol='PH4alphaNE2', Aliases='Dmel_CG9720, CG9720, Dmel\\CG9720', description='prolyl-4-hydroxylase-alpha NE2', other_designations='CG9720-PA|PH4alphaNE2-PA', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30510083, end_position_on_the_genomic_accession=30512214, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43629, CurrentID=0, Status='live', Symbol='CG9698', Aliases='Dmel_CG9698-E, Dmel\\CG9698', description='CG9698 gene product from transcript CG9698-RA', other_designations='CG9698-PA', map_location='99F7-99F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30514777, end_position_on_the_genomic_accession=30516980, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43630, CurrentID=0, Status='live', Symbol='jdp', Aliases='Dmel_CG2239, CG2239, Dmel\\CG2239, JDP, dJDP, p25a', description='CG2239 gene product from transcript CG2239-RB', other_designations='CG2239-PA|CG2239-PB|CG2239-PC|jdp-PA|jdp-PB|jdp-PC', map_location='99F7-99F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30517059, end_position_on_the_genomic_accession=30527037, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19835504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835504, CurrentID=0, Status='live', Symbol='CG45065', Aliases='Dmel_ CG9517, Dm GMCepsilon1, Dmel\\CG45065', description='CG45065 gene product from transcript CG45065-RA', other_designations='CG45065-PA|uncharacterized protein', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14875370, end_position_on_the_genomic_accession=14882586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43633, CurrentID=0, Status='live', Symbol='tmod', Aliases='Dmel_CG1539, 1309/10, CG11493, CG1539, CG15540, CT3919, CT41421, Dmel\\CG1539, E42, Tmod, l(3)00848, l(3)S130910, spdo', description='tropomodulin', other_designations='CG1539-PA|CG1539-PB|CG1539-PC|CG1539-PD|CG1539-PE|CG1539-PF|CG1539-PG|CG1539-PH|CG1539-PI|CG1539-PJ|CG1539-PK|CG1539-PL|CG1539-PM|CG1539-PN|CG1539-PO|CG1539-PP|CG1539-PQ|sanpodo|tmod-PA|tmod-PB|tmod-PC|tmod-PD|tmod-PE|tmod-PF|tmod-PG|tmod-PH|tmod-PI|tmod-PJ|tmod-PK|tmod-PL|tmod-PM|tmod-PN|tmod-PO|tmod-PP|tmod-PQ|tropmodulin', map_location='99F7-99F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30532607, end_position_on_the_genomic_accession=30578905, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  43634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43634, CurrentID=0, Status='live', Symbol='CG9702', Aliases='Dmel_ Dmel\\CG9702', description='CG9702 gene product from transcript CG9702-RA', other_designations='CG9702-PA|CG9702-PB|CG9702-PC', map_location='99F9-99F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30579429, end_position_on_the_genomic_accession=30582117, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43635, CurrentID=0, Status='live', Symbol='Rpt6R', Aliases='Dmel_CG2241, CG2241, DmCG2241, Dm_Rpt6b, Dmel\\CG2241, Rpt6, Rpt6b, S8', description='Regulatory particle triple-A ATPase 6-related', other_designations='CG2241-PA|Rpt6R-PA|regulatory particle triple-A ATPase 6-related', map_location='99F9-99F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30582198, end_position_on_the_genomic_accession=30583697, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43636, CurrentID=0, Status='live', Symbol='CG9717', Aliases='Dmel_ Dmel\\CG9717', description='CG9717 gene product from transcript CG9717-RA', other_designations='CG9717-PA', map_location='99F10-99F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30584020, end_position_on_the_genomic_accession=30587302, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43637, CurrentID=0, Status='live', Symbol='CG2246', Aliases='Dmel_ Dmel\\CG2246', description='CG2246 gene product from transcript CG2246-RK', other_designations='CG2246-PA|CG2246-PB|CG2246-PC|CG2246-PD|CG2246-PE|CG2246-PF|CG2246-PG|CG2246-PH|CG2246-PI|CG2246-PJ|CG2246-PK|CG2246-PL|CG2246-PM|CG2246-PN|CG2246-PO|CG2246-PP', map_location='99F10-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30592971, end_position_on_the_genomic_accession=30601489, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43638, CurrentID=0, Status='live', Symbol='CG31021', Aliases='Dmel_ CG9708, Dmel\\CG31021', description='CG31021 gene product from transcript CG31021-RB', other_designations='CG31021-PB|uncharacterized protein', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30602206, end_position_on_the_genomic_accession=30604041, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43639, CurrentID=0, Status='live', Symbol='CG2267', Aliases='Dmel_ Dmel\\CG2267, anon-WO0140519.167', description='CG2267 gene product from transcript CG2267-RB', other_designations='CG2267-PA|CG2267-PB', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30627241, end_position_on_the_genomic_accession=30629028, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43640, CurrentID=0, Status='live', Symbol='PH4alphaPV', Aliases='Dmel_CG31015, CG31015, CG9713, Dmel\\CG31015', description='prolyl-4-hydroxylase-alpha PV', other_designations='CG31015-PA|PH4alphaPV-PA', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30646106, end_position_on_the_genomic_accession=30648551, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43642, CurrentID=0, Status='live', Symbol='Spn100A', Aliases='Dmel_CG1342, CG1342, Dmel\\CG1342, Spn100', description='Serpin 100A', other_designations='CG1342-PA|Spn100A-PA|serpin 100A', map_location='100A2-100A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30690642, end_position_on_the_genomic_accession=30692768, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43643, CurrentID=0, Status='live', Symbol='CG12069', Aliases='Dmel_ Dmel\\CG12069, Pka-like', description='CG12069 gene product from transcript CG12069-RA', other_designations='CG12069-PA|CG12069-PC|PKA-like|uncharacterized protein', map_location='100A2-100A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30693572, end_position_on_the_genomic_accession=30695498, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43644, CurrentID=0, Status='live', Symbol='Pka-C2', Aliases='Dmel_CG12066, CG12066, CdkB, DC1, DC1a, Dmel\\CG12066, PKA, PKA C2', description='Protein kinase, cAMP-dependent, catalytic subunit 2', other_designations='CG12066-PB|CG12066-PC|Pka-C2-PB|Pka-C2-PC|cAMP-dependent protein kinase 2|cAMP-dependent protein kinase type II|protein kinase, cAMP-dependent, catalytic subunit 2', map_location='100A2-100A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30695537, end_position_on_the_genomic_accession=30697138, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43645, CurrentID=0, Status='live', Symbol='CG1340', Aliases='Dmel_ Dmel\\CG1340, cg1340, eIF4H', description='CG1340 gene product from transcript CG1340-RC', other_designations='CG1340-PA|CG1340-PC', map_location='100A3-100A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30726087, end_position_on_the_genomic_accession=30728052, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43646, CurrentID=0, Status='live', Symbol='CG11313', Aliases='Dmel_ Dmel\\CG11313, SP11, c-SP11', description='CG11313 gene product from transcript CG11313-RC', other_designations='CG11313-PC|CG11313-PD|uncharacterized protein', map_location='100A3-100A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30729837, end_position_on_the_genomic_accession=30731165, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43647, CurrentID=0, Status='live', Symbol='CG15543', Aliases='Dmel_ Dmel\\CG15543', description='CG15543 gene product from transcript CG15543-RB', other_designations='CG15543-PA|CG15543-PB|uncharacterized protein', map_location='100A3-100A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30742880, end_position_on_the_genomic_accession=30744474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43648, CurrentID=0, Status='live', Symbol='Npc2g', Aliases='Dmel_CG11314, BcDNA:RE56164, CG11314, Dmel\\CG11314, npc2g', description='Niemann-Pick type C-2g', other_designations='CG11314-PA|CG11314-PB|Npc2g-PA|Npc2g-PB', map_location='100A3-100A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30744605, end_position_on_the_genomic_accession=30745263, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43649, CurrentID=0, Status='live', Symbol='Npc2h', Aliases='Dmel_CG11315, BcDNA:GH26608, BcDNA:RH04252, BcDNA:RH68460, CG11315, DmML1A, Dmel\\CG11315, npc2h', description='Niemann-Pick type C-2h', other_designations='CG11315-PA|CG11315-PB|CG11315-PC|Npc2h-PA|Npc2h-PB|Npc2h-PC', map_location='100A3-100A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30745479, end_position_on_the_genomic_accession=30746290, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43650, CurrentID=0, Status='live', Symbol='zfh1', Aliases='Dmel_CG1322, CG1322, Dmel\\CG1322, ZFH-1, ZFH1, Zfh-1, Zfh1, Zfh1a, l(3)00865, zfh-1, zfl-1, zhf1', description='Zn finger homeodomain 1', other_designations='CG1322-PA|CG1322-PB|CG1322-PE|Zinc-finger homeodomain protein 1|Zn homeodomain 1|zfh1-PA|zfh1-PB|zfh1-PE|zinc finger homeodomain 1|zinc finger homeodomain protein-1|zinc finger homeodomain-1|zinc-finger homeobox gene|zinc-finger homeodomain1', map_location='100A4-100A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30765926, end_position_on_the_genomic_accession=30789202, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43651, CurrentID=0, Status='live', Symbol='wts', Aliases='Dmel_CG12072, CG12072, Dlats, Dmel\\CG12072, LATS, LATS1, Lats, Lats/Warts, MENE (3R)-G, WTS/LATS, Warts, Warts/Lats, Wts, Wts/Lats, dmLATS, l(3)100Aa, lats, lts, wart, warts/lats/lats, wts', description='warts', other_designations='CG12072-PA|Warts/Lats|large tumor suppressor|large tumor supressor|miracle-gro|warts kinase|warts/D-lats|warts/large tumour suppressor|wts-PA', map_location='100A5-100A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30789625, end_position_on_the_genomic_accession=30806619, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43652, CurrentID=0, Status='live', Symbol='dj-1beta', Aliases='Dmel_CG1349, CG1349, DJ-1, DJ-1B, DJ-1b, DJ-1beta, Dmel\\CG1349', description='CG1349 gene product from transcript CG1349-RB', other_designations='CG1349-PB|DJ-1B|dj-1beta-PB', map_location='100A5-100A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30807117, end_position_on_the_genomic_accession=30807978, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19835525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835525, CurrentID=0, Status='live', Symbol='CG44624', Aliases='Dmel_ Dmel\\CG44624', description='CG44624 gene product from transcript CG44624-RA', other_designations='CG44624-PA|uncharacterized protein', map_location='56F10-56F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20234995, end_position_on_the_genomic_accession=20235345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43654, CurrentID=0, Status='live', Symbol='cindr', Aliases='Dmel_CG31012, BcDNA:GH03163, CD2AP, CG11316, CG1408, CG31012, Cindr, Dmel\\CG31012', description='CIN85 and CD2AP orthologue', other_designations='CD2AP and CIN85 orthologue|CG31012-PA|CG31012-PB|CG31012-PC|CG31012-PD|CG31012-PE|CG31012-PF|CG31012-PG|CG31012-PH|CG31012-PI|CG31012-PJ|cindr-PA|cindr-PB|cindr-PC|cindr-PD|cindr-PE|cindr-PF|cindr-PG|cindr-PH|cindr-PI|cindr-PJ', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30808617, end_position_on_the_genomic_accession=30827005, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43655, CurrentID=0, Status='live', Symbol='CG15544', Aliases='Dmel_ Dmel\\CG15544', description='CG15544 gene product from transcript CG15544-RB', other_designations='CG15544-PA|CG15544-PB|uncharacterized protein', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30827819, end_position_on_the_genomic_accession=30842034, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43656, CurrentID=0, Status='live', Symbol='tll', Aliases='Dmel_CG1378, CG1378, Dmel\\CG1378, NR2E2, TLL, Tll, dm-tll', description='tailless', other_designations='CG1378-PA|tailles|tll-PA', map_location='3-102 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30852315, end_position_on_the_genomic_accession=30854400, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43657, CurrentID=0, Status='live', Symbol='Cpr100A', Aliases='Dmel_CG12045, CG12045, DmelCpr100A, Dmel\\CG12045', description='Cuticular protein 100A', other_designations='CG12045-PA|Cpr100A-PA|cuticular protein 100A', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30867521, end_position_on_the_genomic_accession=30869184, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43658, CurrentID=0, Status='live', Symbol='CG15545', Aliases='Dmel_ Dmel\\CG15545', description='CG15545 gene product from transcript CG15545-RA', other_designations='CG15545-PA|uncharacterized protein', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30870939, end_position_on_the_genomic_accession=30872351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43659, CurrentID=0, Status='live', Symbol='CG15546', Aliases='Dmel_ Dmel\\CG15546', description='CG15546 gene product from transcript CG15546-RA', other_designations='CG15546-PA|uncharacterized protein', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30874365, end_position_on_the_genomic_accession=30876186, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43660, CurrentID=0, Status='live', Symbol='CG12071', Aliases='Dmel_ Dmel\\CG12071', description='CG12071 gene product from transcript CG12071-RE', other_designations='CG12071-PD|CG12071-PE|CG12071-PF|CG12071-PG|uncharacterized protein', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30877084, end_position_on_the_genomic_accession=30881027, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43661, CurrentID=0, Status='live', Symbol='CG15547', Aliases='Dmel_ Dmel\\CG15547', description='CG15547 gene product from transcript CG15547-RA', other_designations='CG15547-PA|uncharacterized protein', map_location='100A6-100A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30881792, end_position_on_the_genomic_accession=30883503, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43662, CurrentID=0, Status='live', Symbol='Sap-r', Aliases='Dmel_CG12070, Apa, BcDNA:GH08312, CG12070, DAPA, Dmel\\CG12070, Ov9, P110, Sap-R, Sapr, anon-EST:ParkEST270, sap-r', description='Saposin-related', other_designations='110kD protein|CG12070-PA|CG12070-PB|Sap-r-PA|Sap-r-PB|Saposin-r|aggregation promoting activity|drosophila aggregation promoting activity', map_location='100A6-100A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30883464, end_position_on_the_genomic_accession=30889062, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43663, CurrentID=0, Status='live', Symbol='Cyp4c3', Aliases='Dmel_CG1438, 4C3, CG1438, Dmel\\CG1438, P-450, P450, cyp4c3', description='Cytochrome P450-4c3', other_designations='CG1438-PA|Cyp4c3-PA|cytochrome P450|cytochrome P450-4c3', map_location='100A7-100A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30894038, end_position_on_the_genomic_accession=30900688, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43664, CurrentID=0, Status='live', Symbol='Ptx1', Aliases='Dmel_CG1447, CG1447, D- Dmel\\CG1447, Ptx, ptx1', description='CG1447 gene product from transcript CG1447-RC', other_designations='CG1447-PA|CG1447-PC|CG1447-PE|Ptx1-PA|Ptx1-PC|Ptx1-PE', map_location='100A7-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30912754, end_position_on_the_genomic_accession=30930682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43665, CurrentID=0, Status='live', Symbol='CG15549', Aliases='Dmel_ BcDNA:AT07554, Dmel\\CG15549', description='CG15549 gene product from transcript CG15549-RA', other_designations='CG15549-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30935396, end_position_on_the_genomic_accession=30936507, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43666, CurrentID=0, Status='live', Symbol='CG15550', Aliases='Dmel_ Dmel\\CG15550', description='CG15550 gene product from transcript CG15550-RA', other_designations='CG15550-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30936589, end_position_on_the_genomic_accession=30937102, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43667, CurrentID=0, Status='live', Symbol='CG15548', Aliases='Dmel_ Dmel\\CG15548', description='CG15548 gene product from transcript CG15548-RA', other_designations='CG15548-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30959700, end_position_on_the_genomic_accession=30962307, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43668, CurrentID=0, Status='live', Symbol='Ctr1C', Aliases='Dmel_CG15551, BcDNA:AT31946, CG15551, Dmel\\CG15551, FBgn0039813', description='Copper transporter 1C', other_designations='CG15551-PA|Ctr1C-PA|copper transporter 1C', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30963290, end_position_on_the_genomic_accession=30964320, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43669, CurrentID=0, Status='live', Symbol='5-HT7', Aliases='Dmel_CG12073, 5-HT-dro, 5-HT-dro1Dro, 5-HT[7]Dro, 5-HT[[7Dro]], 5-HT[[7]], 5-HT[[7]]Dro, 5-HT[[dro1]], 5-HTdro1, 5HT-7, 5HT-R1, 5HT-dro, 5HT-dro1, 5HT7, 5HT[[7Dro]], 5htdror, CG12073, DRO1, Dm5HTdro1, Dmel\\CG12073, 5-HT7', description='5-hydroxytryptamine (serotonin) receptor 7', other_designations='5-HT7-PA|5-HT7-PB|CG12073-PA|CG12073-PB|serotonin receptor|serotonin receptor 1|serotonin receptor 7', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30968236, end_position_on_the_genomic_accession=31016875, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43670, CurrentID=0, Status='live', Symbol='CanA1', Aliases='Dmel_CG1455, CAN A1/PP2B, CG1455, CNA1, Can, Cana1, CnA, CnnA21EF, D14, DmCanA1-100B, Dmel\\CG1455, PP2B, PP2B 21EF, PpD14, canA, canA1, pMY4', description='Calcineurin A1', other_designations='CG1455-PA|CG1455-PB|CG1455-PD|CanA1-PA|CanA1-PB|CanA1-PD|calcineurin|calcineurin A|calcineurin A1|protein phosphatase D14|protein phosphatase from PCR fragment D14', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31039374, end_position_on_the_genomic_accession=31046821, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  43671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43671, CurrentID=0, Status='live', Symbol='Alp4', Aliases='Dmel_CG1462, Aph-4, CG1462, Dmel\\CG1462, aph-4, l(3)07028, pMY51', description='Alkaline phosphatase 4', other_designations='Alp4-PA|Alp4-PB|CG1462-PA|CG1462-PB|alkaline phosphatase|alkaline phosphatase 4|alkaline phosphatase-4', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31048929, end_position_on_the_genomic_accession=31052540, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43672, CurrentID=0, Status='live', Symbol='l(3)03670', Aliases='Dmel_CG1715, 0205/14, 0353/13, 0812/02, 0966/01, 1274/08, CG1715, Dmel\\CG1715, anon-l, l(3)3670, l(3)96601, l(3)S127408, pMY8, x94917', description='lethal (3) 03670', other_designations='CG1715-PA|l(3)03670-PA', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31052616, end_position_on_the_genomic_accession=31053915, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43673, CurrentID=0, Status='live', Symbol='dco', Aliases='Dmel_CG2048, 0538/13, 0915/10, 1396/02, 1447/01, 1460/09, CG2048, CK1delta/epsilon, CK1epsilon, CKIepsilon, DBT, DBT/CK1epsilon, DCO, DCO/CKIe, Dbt, Dco, Dco/CK1, Dmel\\CG2048, ck1epsilon, dCKIepsilon, dbt-1, ddbt, l(3)S053813, l(3)S144701, l(3)dco, l(3)dco-1, l(3)discs overgrown, l(3)discs overgrown-1, l(3)j3B9, l(3)rK215, dco', description='discs overgrown', other_designations='CG2048-PA|CG2048-PB|CG2048-PC|CG2048-PD|casein kinase I epsilon|casein kinase iepsilon|dco-PA|dco-PB|dco-PC|dco-PD|disc overgrown|double time|double-time|doubletime', map_location='100B1-100B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31054085, end_position_on_the_genomic_accession=31061210, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19835548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835548, CurrentID=0, Status='live', Symbol='CG45068', Aliases='Dmel_ CG3419, Dmel\\CG45068', description='CG45068 gene product from transcript CG45068-RA', other_designations='CG45068-PA|uncharacterized protein', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24411870, end_position_on_the_genomic_accession=24415295, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43677, CurrentID=0, Status='live', Symbol='CG15553', Aliases='Dmel_ Dmel\\CG15553', description='CG15553 gene product from transcript CG15553-RC', other_designations='CG15553-PC|uncharacterized protein', map_location='100B2-100B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31098868, end_position_on_the_genomic_accession=31101169, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43678, CurrentID=0, Status='live', Symbol='CG11318', Aliases='Dmel_ Dmel\\CG11318, RE14222p, anon-WO0170980.121, anon-WO0170980.122', description='CG11318 gene product from transcript CG11318-RA', other_designations='CG11318-PA|uncharacterized protein', map_location='100B2-100B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31106362, end_position_on_the_genomic_accession=31109191, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43679, CurrentID=0, Status='live', Symbol='Prosalpha3T', Aliases='Dmel_CG1736, :Prosalpha3T, CG1736, Dmel\\CG1736, Pros20S-alpha3t, alpha3T, prosalpha3T', description='Proteasome alpha3 subunit, Testis-specific', other_designations='20s proteasome|CG1736-PA|Prosalpha3T-PA|proteasome 20S alpha3t subunit|proteasome alpha3 subunit, Testis-specific|proteasome alpha3T subunit', map_location='100B3-100B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31122527, end_position_on_the_genomic_accession=31123415, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43680, CurrentID=0, Status='live', Symbol='CG15554', Aliases='Dmel_ BcDNA:RE10288, Dmel\\CG15554', description='CG15554 gene product from transcript CG15554-RA', other_designations='CG15554-PA|uncharacterized protein', map_location='100B3-100B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31126149, end_position_on_the_genomic_accession=31127335, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43681, CurrentID=0, Status='live', Symbol='CG15556', Aliases='Dmel_ AT07595p, Dmel\\CG15556, anon-WO0170980.178, anon-WO0170980.179', description='CG15556 gene product from transcript CG15556-RA', other_designations='CG15556-PA|CG15556-PB|CG15556-PC|uncharacterized protein', map_location='100B3-100B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31127965, end_position_on_the_genomic_accession=31130666, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43682, CurrentID=0, Status='live', Symbol='Gycbeta100B', Aliases='Dmel_CG1470, CG1470, Dgc beta 1, Dgcbeta1, Dmel\\CG1470, Gycbeta-100B, dgcb1, dgcbeta1, gycbeta100B, sGCbeta1, sGCbeta100B', description='Guanylyl cyclase beta-subunit at 100B', other_designations='CG1470-PA|Gycbeta100B-PA|guanylyl cyclase beta-subunit at 100B', map_location='100B4-100B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31134392, end_position_on_the_genomic_accession=31178278, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  43683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43683, CurrentID=0, Status='live', Symbol='stops', Aliases='Dmel_CG31006, CG15557, CG15558, CG15559, CG31006, Dmel\\CG31006, su(rdgB)69', description='slow termination of phototransduction', other_designations='CG31006-PA|CG31006-PB|CG31006-PC|stops-PA|stops-PB|stops-PC|suppressor (rdgB) 69', map_location='3-101 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31158156, end_position_on_the_genomic_accession=31169106, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43687, CurrentID=0, Status='live', Symbol='bnk', Aliases='Dmel_CG1480, Bnk, CG1480, Dmel\\CG1480', description='bottleneck', other_designations='CG1480-PA|bnk-PA', map_location='100B5-100B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31193362, end_position_on_the_genomic_accession=31194389, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43688, CurrentID=0, Status='live', Symbol='mesh', Aliases='Dmel_CG31004, CG12074, CG15560, CG31004, CT5150, Dmel\\CG31004, anon-WO0140519.233', description='CG31004 gene product from transcript CG31004-RC', other_designations='CG31004-PA|CG31004-PB|CG31004-PC|CG31004-PD|CG31004-PE|mesh-PA|mesh-PB|mesh-PC|mesh-PD|mesh-PE', map_location='100B5-100B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31178389, end_position_on_the_genomic_accession=31195762, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  43689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43689, CurrentID=0, Status='live', Symbol='CG1544', Aliases='Dmel_ Dmel\\CG1544', description='CG1544 gene product from transcript CG1544-RA', other_designations='CG1544-PA', map_location='100B5-100B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31196916, end_position_on_the_genomic_accession=31203722, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  43690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43690, CurrentID=0, Status='live', Symbol='chp', Aliases='Dmel_CG1744, 24B10, 516, BSH, CG1744, CHP, CT5040, Chaoptin, Chp, Dmel\\CG1744, MAb 24B10, MAb24B10, Mab24B10, Nsp, idl-A, mAB24B10, mAb 24B10, mAb24B10', description='chaoptin', other_designations='CG1744-PA|CG1744-PC|CG1744-PD|CG1744-PE|CG1744-PF|INAD localization-A|Neural-specific-protein|chaoptic|chp-PA|chp-PC|chp-PD|chp-PE|chp-PF', map_location='100B5-100B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31204180, end_position_on_the_genomic_accession=31210730, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  43691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43691, CurrentID=0, Status='live', Symbol='CG1542', Aliases='Dmel_ Dmel\\CG1542', description='CG1542 gene product from transcript CG1542-RA', other_designations='CG1542-PA', map_location='100B6-100B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31213458, end_position_on_the_genomic_accession=31214784, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43692, CurrentID=0, Status='live', Symbol='CG15561', Aliases='Dmel_ Dmel\\CG15561', description='CG15561 gene product from transcript CG15561-RA', other_designations='CG15561-PA|uncharacterized protein', map_location='100B7-100B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31214776, end_position_on_the_genomic_accession=31215942, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43693, CurrentID=0, Status='live', Symbol='ATPsynC', Aliases='Dmel_CG1746, AT91, CG1746, Dmel\\CG1746, anon-EST:Posey224', description='ATP synthase, subunit C', other_designations='ATP synthase Lipid-binding protein P1|ATP synthase lipid-binding protein|ATPsynC-PA|ATPsynC-PB|ATPsynC-PC|ATPsynC-PE|ATPsynC-PF|ATPsynC-PG|CG1746-PA|CG1746-PB|CG1746-PC|CG1746-PE|CG1746-PF|CG1746-PG|F0/F1 ATP synthase Lipid-binding protein', map_location='100B7-100B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31216145, end_position_on_the_genomic_accession=31219630, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43694, CurrentID=0, Status='live', Symbol='CG12054', Aliases='Dmel_ Dmel\\CG12054', description='CG12054 gene product from transcript CG12054-RE', other_designations='CG12054-PA|CG12054-PB|CG12054-PE|CG12054-PF|uncharacterized protein', map_location='100B8-100B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31220702, end_position_on_the_genomic_accession=31231259, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43695, CurrentID=0, Status='live', Symbol='CG15563', Aliases='Dmel_ Dmel\\CG15563', description='CG15563 gene product from transcript CG15563-RA', other_designations='CG15563-PA|uncharacterized protein', map_location='100B8-100B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31231685, end_position_on_the_genomic_accession=31232336, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43696, CurrentID=0, Status='live', Symbol='CG15564', Aliases='Dmel_ Dmel\\CG15564', description='CG15564 gene product from transcript CG15564-RA', other_designations='CG15564-PA|uncharacterized protein', map_location='100B8-100B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31232415, end_position_on_the_genomic_accession=31235362, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43697, CurrentID=0, Status='live', Symbol='qless', Aliases='Dmel_CG31005, CG15565, CG31005, CG3684, Dmel\\CG31005, anon-WO0118547.422', description='CG31005 gene product from transcript CG31005-RA', other_designations='CG31005-PA|CG31005-PB|qless-PA|qless-PB', map_location='100B8-100B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31236035, end_position_on_the_genomic_accession=31252608, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43698, CurrentID=0, Status='live', Symbol='mRpL32', Aliases='Dmel_CG12220, BcDNA:LD41522, CG12220, Dmel\\CG12220, L32, RpL32', description='mitochondrial ribosomal protein L32', other_designations='CG12220-PA|mRpL32-PA', map_location='100B8-100B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31252872, end_position_on_the_genomic_accession=31253686, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43699, CurrentID=0, Status='live', Symbol='CG1750', Aliases='Dmel_ Dmel\\CG1750', description='CG1750 gene product from transcript CG1750-RA', other_designations='CG1750-PA', map_location='100B8-100B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31253659, end_position_on_the_genomic_accession=31254900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43700, CurrentID=0, Status='live', Symbol='spn-F', Aliases='Dmel_CG12114, 1130/03, CG12114, Dmel\\CG12114, Spn-F, l(3)S113003b', description='CG12114 gene product from transcript CG12114-RA', other_designations='CG12114-PA|CG12114-PB|Spindle-F|lethal (3) S113003b|spindle F|spn-F-PA|spn-F-PB', map_location='3-100 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31255127, end_position_on_the_genomic_accession=31256659, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43701, CurrentID=0, Status='live', Symbol='CG3669', Aliases='Dmel_ Dmel\\CG3669', description='CG3669 gene product from transcript CG3669-RB', other_designations='CG3669-PB', map_location='100B9-100B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31257599, end_position_on_the_genomic_accession=31260049, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43702, CurrentID=0, Status='live', Symbol='ppk24', Aliases='Dmel_CG15555, CG15555, Dmel\\CG15555', description='pickpocket 24', other_designations='CG15555-PA|ppk24-PA', map_location='100B9-100B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31265730, end_position_on_the_genomic_accession=31268511, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43703, CurrentID=0, Status='live', Symbol='CG11340', Aliases='Dmel_ CT5896, Dmel\\CG11340, lincRNA.914', description='CG11340 gene product from transcript CG11340-RB', other_designations='CG11340-PA|CG11340-PB|glycine-gated chloride channel-like|uncharacterized protein', map_location='100C1-100C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31300918, end_position_on_the_genomic_accession=31305384, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43707, CurrentID=0, Status='live', Symbol='CG1607', Aliases='Dmel_ Cg1607, Dmel\\CG1607', description='CG1607 gene product from transcript CG1607-RA', other_designations='CG1607-PA|CG1607-PB|CG1607-PC|CG1607-PD', map_location='100C2-100C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31382109, end_position_on_the_genomic_accession=31391893, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  43708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43708, CurrentID=0, Status='live', Symbol='CG31002', Aliases='Dmel_CG31002, CG15569-PA, Dmel\\CG31002', description='CG31002 gene product from transcript CG31002-RA', other_designations='CG31002-PA|uncharacterized protein', map_location='100C3-100C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31393582, end_position_on_the_genomic_accession=31395304, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43709, CurrentID=0, Status='live', Symbol='Gcn2', Aliases='Dmel_CG1609, CG1609, DGCN2, Dmel\\CG1609, EIF2, GCN2, Gcn-2, dGCN2', description='CG1609 gene product from transcript CG1609-RA', other_designations='CG1609-PA|CG1609-PB|CG1609-PC|GC nonderepressing 2 kinase|Gcn2-PA|Gcn2-PB|Gcn2-PC|general control nonderepressing 2 kinase', map_location='100C3-100C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31395660, end_position_on_the_genomic_accession=31402013, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43710, CurrentID=0, Status='live', Symbol='CG11337', Aliases='Dmel_ Dmel\\CG11337', description='CG11337 gene product from transcript CG11337-RD', other_designations='CG11337-PA|CG11337-PB|CG11337-PD|CG11337-PE|uncharacterized protein', map_location='100C3-100C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31401560, end_position_on_the_genomic_accession=31405452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43712, CurrentID=0, Status='live', Symbol='lox', Aliases='Dmel_CG11335, CG11335, DmLOXL-1, Dmel\\CG11335, Dmloxl-1, LOX, Lox, dmlox-1', description='lysyl oxidase-like', other_designations='CG11335-PA|lox-PA', map_location='100C4-100C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31458380, end_position_on_the_genomic_accession=31459869, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43713, CurrentID=0, Status='live', Symbol='CG11334', Aliases='Dmel_ Dmel\\CG11334', description='CG11334 gene product from transcript CG11334-RA', other_designations='CG11334-PA|CG11334-PB|CG11334-PC|uncharacterized protein', map_location='100C4-100C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31459822, end_position_on_the_genomic_accession=31463037, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43714, CurrentID=0, Status='live', Symbol='CG11333', Aliases='Dmel_ BcDNA:AT02774, Dmel\\CG11333', description='CG11333 gene product from transcript CG11333-RA', other_designations='CG11333-PA|uncharacterized protein', map_location='100C4-100C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31463301, end_position_on_the_genomic_accession=31464134, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43715, CurrentID=0, Status='live', Symbol='mey', Aliases='Dmel_CG12063, CG12063, Dmel\\CG12063', description='morpheyus', other_designations='CG12063-PA|CG12063-PB|mey-PA|mey-PB', map_location='100C4-100C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31499235, end_position_on_the_genomic_accession=31506992, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43716, CurrentID=0, Status='live', Symbol='nyo', Aliases='Dmel_CG1499, CG1499, Dmel\\CG1499, anon-EST:Liang-2.47, clone 2.47', description='nyobe', other_designations='CG1499-PC|CG1499-PD|nyo-PC|nyo-PD', map_location='100C4-100C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31543032, end_position_on_the_genomic_accession=31566165, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43717, CurrentID=0, Status='live', Symbol='gammaCOP', Aliases='Dmel_CG1528, 0573/02, 573.02, CG1528, COPIgamma, Dmel\\CG1528, anon-WO0118547.334, copg, gamma-COP, gammaCop, l(3)S057302, l(3)S057302b', description='Coat Protein (coatomer) gamma', other_designations='CG1528-PA|CG1528-PB|CG1528-PC|CG1528-PD|COPI gamma subunit|coat protein (coatomer) gamma|gamma-coatomer protein|gammaCOP-PA|gammaCOP-PB|gammaCOP-PC|gammaCOP-PD|gammaCop|lethal (3) S057302b', map_location='100C6-100C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31572143, end_position_on_the_genomic_accession=31575707, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43718, CurrentID=0, Status='live', Symbol='pygo', Aliases='Dmel_CG11518, BEST:LD21971, CG11518, Dm Pygo, Dmel\\CG11518, Pygo, anon-WO02077023.1, dPygo, gam, pNB5.2, pyg, pypo', description='pygopus', other_designations='CG11518-PA|gammy legs|pygo-PA|pygopous', map_location='3-102 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31575835, end_position_on_the_genomic_accession=31580699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43719, CurrentID=0, Status='live', Symbol='rod', Aliases='Dmel_CG1569, CG1569, Dmel\\CG1569, EP3408, ROD, Rod, pNB20A', description='rough deal', other_designations='CG1569-PA|rod-PA|roughdeal', map_location='3-105.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31580951, end_position_on_the_genomic_accession=31588244, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43720, CurrentID=0, Status='live', Symbol='CG1635', Aliases='Dmel_ Dmel\\CG1635', description='CG1635 gene product from transcript CG1635-RA', other_designations='CG1635-PA', map_location='100C7-100C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31588673, end_position_on_the_genomic_accession=31590392, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43721, CurrentID=0, Status='live', Symbol='CG1638', Aliases='Dmel_ Dmel\\CG1638, bs31f06.y1', description='CG1638 gene product from transcript CG1638-RB', other_designations='CG1638-PB', map_location='100C7-100C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31590581, end_position_on_the_genomic_accession=31592316, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43722, CurrentID=0, Status='live', Symbol='CG1774', Aliases='Dmel_ Dmel\\CG1774', description='CG1774 gene product from transcript CG1774-RA', other_designations='CG1774-PA|CG1774-PB', map_location='100C7-100C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31592803, end_position_on_the_genomic_accession=31594719, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43723, CurrentID=0, Status='live', Symbol='RpL6', Aliases='Dmel_CG11522, CG11522, Dmel\\CG11522, RPL6, Rp L6, rpl6', description='Ribosomal protein L6', other_designations='CG11522-PA|CG11522-PB|RpL6-PA|RpL6-PB|ribosomal protein L6', map_location='100C7-100C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31594792, end_position_on_the_genomic_accession=31596204, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43724, CurrentID=0, Status='live', Symbol='CycG', Aliases='Dmel_CG11525, CG 11525, CG11525, Dmel\\CG11525, anon-EST:Liang-1.25, anon-EST:Liang-2.24, chr3R:27424276..27424475, clone 1.25, clone 2.24, cycG, dCycG, i205, i261', description='Cyclin G', other_designations='CG11525-PA|CG11525-PB|CG11525-PC|CG11525-PD|CG11525-PF|CycG-PA|CycG-PB|CycG-PC|CycG-PD|CycG-PF|cyclin G|cyclin G1', map_location='100C7-100C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31596642, end_position_on_the_genomic_accession=31612664, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43725, CurrentID=0, Status='live', Symbol='Med', Aliases='Dmel_CG1775, CG1775, Dmel\\CG1775, E(zen)3, MED, SMAD4, Smad4, anon-EST:Posey121, dSmad4, l(3)11m-254, l(3)12m-137, l(3)SG36, l(3)SG70, l(3)XIIm137, med, medea', description='Medea', other_designations='CG1775-PA|CG1775-PB|CG1775-PC|Med-PA|Med-PB|Med-PC|lethal(3)SG36|lethal(3)SG70|medea|media', map_location='3-106 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31611046, end_position_on_the_genomic_accession=31615375, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43726, CurrentID=0, Status='live', Symbol='CG11539', Aliases='Dmel_ Dmel\\CG11539, NAT9', description='CG11539 gene product from transcript CG11539-RB', other_designations='CG11539-PA|CG11539-PB|uncharacterized protein', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31615284, end_position_on_the_genomic_accession=31616160, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43727, CurrentID=0, Status='live', Symbol='CG1792', Aliases='Dmel_ Dmel\\CG1792', description='CG1792 gene product from transcript CG1792-RA', other_designations='CG1792-PA', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31616303, end_position_on_the_genomic_accession=31617837, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43728, CurrentID=0, Status='live', Symbol='pasha', Aliases='Dmel_CG1800, CG1800, DGCR8, Dmel\\CG1800, PASHA, Pasha, cg1800, dmDGCR8', description='partner of drosha', other_designations='CG1800-PA|CG1800-PB|CG1800-PC|DiGeorge critical region 8|pasha-PA|pasha-PB|pasha-PC', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31617977, end_position_on_the_genomic_accession=31622217, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43729, CurrentID=0, Status='live', Symbol='kek6', Aliases='Dmel_CG1804, CG1804, Dmel\\CG1804', description='CG1804 gene product from transcript CG1804-RA', other_designations='CG1804-PA|CG1804-PB|kek6-PA|kek6-PB|kekkon-like', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31622136, end_position_on_the_genomic_accession=31664646, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43730, CurrentID=0, Status='live', Symbol='CG1815', Aliases='Dmel_ Dmel\\CG1815', description='CG1815 gene product from transcript CG1815-RC', other_designations='CG1815-PB|CG1815-PC', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31669465, end_position_on_the_genomic_accession=31677971, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43731, CurrentID=0, Status='live', Symbol='CG11550', Aliases='Dmel_ Dmel\\CG11550', description='CG11550 gene product from transcript CG11550-RA', other_designations='CG11550-PA|uncharacterized protein', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31678147, end_position_on_the_genomic_accession=31679880, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43734, CurrentID=0, Status='live', Symbol='CstF-50', Aliases='Dmel_CG2261, CG2261, CstF50, Dmel\\CG2261', description='CG2261 gene product from transcript CG2261-RA', other_designations='CG2261-PA|CstF-50-PA', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31735503, end_position_on_the_genomic_accession=31736963, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43735, CurrentID=0, Status='live', Symbol='CG11563', Aliases='Dmel_ BcDNA:RE07994, Dmel\\CG11563', description='CG11563 gene product from transcript CG11563-RA', other_designations='CG11563-PA|uncharacterized protein', map_location='100D1-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31737035, end_position_on_the_genomic_accession=31737999, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43736, CurrentID=0, Status='live', Symbol='Ef1alpha100E', Aliases='Dmel_CG1873, CG1873, Dmel\\CG1873, EF-1, EF-1alpha, EF-1alpha F2, EF1-a 100E, EF1-alpha, EF1a, EF1alpha, EF1alpha100, Ef1 alpha100E, Ef1a-100E, Ef1a48, Ef1alpha2, Efla2, F2, eEF1alpha, ef-1alpha', description='Elongation factor 1alpha100E', other_designations='CG1873-PA|CG1873-PB|CG1873-PC|CG1873-PD|Ef1alpha100E-PA|Ef1alpha100E-PB|Ef1alpha100E-PC|Ef1alpha100E-PD|S|elongation factor 1-alpha|elongation factor 1-alpha F2|elongation factor 1alpha 100E|elongation factor 1alpha100E', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31738468, end_position_on_the_genomic_accession=31743306, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43737, CurrentID=0, Status='live', Symbol='CG1890', Aliases='Dmel_ Dmel\\CG1890', description='CG1890 gene product from transcript CG1890-RA', other_designations='CG1890-PA', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31743490, end_position_on_the_genomic_accession=31744248, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43738, CurrentID=0, Status='live', Symbol='CG1896', Aliases='Dmel_ Dmel\\CG1896', description='CG1896 gene product from transcript CG1896-RA', other_designations='CG1896-PA', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31744282, end_position_on_the_genomic_accession=31745178, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43739, CurrentID=0, Status='live', Symbol='awd', Aliases='Dmel_CG2210, 1084/08, Awd, BcDNA:GM19775, BcDNA:RH27794, CG2210, Dmel\\CG2210, K-pn, Kpn, NDKB, NDPK, Nm23/awd, anon-EST:Liang-2.27, anon-EST:Liang-2.28, anon-WO0172774.80, anon-WO0172774.82, clone 2.27, clone 2.28, e(shi)A, eshiA, l(3)L8700, l(3)j2A4', description='abnormal wing discs', other_designations='CG2210-PC|CG2210-PD|CG2210-PE|Killer-of-prune|NDP kinase|abnormal wing disc|abnormal wings disc|awd-PC|awd-PD|awd-PE|killer of prune|nucleoside diphosphate kinase', map_location='3-104.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31745172, end_position_on_the_genomic_accession=31746077, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43740, CurrentID=0, Status='live', Symbol='nero', Aliases='Dmel_CG2245, CG2245, DOHH, Dmel\\CG2245, l(3)s1921', description='CG2245 gene product from transcript CG2245-RA', other_designations='CG2245-PA|deoxyhypusine hydroxylase|lethal (3) s1921|nero-PA', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31746445, end_position_on_the_genomic_accession=31747558, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43741, CurrentID=0, Status='live', Symbol='CG1910', Aliases='Dmel_ 1A3, Anon1A3, Dmel\\CG1910, anon-EST:fe1A3, anon-WO0118547.398, anon-WO0118547.745, anon1A3', description='CG1910 gene product from transcript CG1910-RA', other_designations='CG1910-PA|CG1910-PC|CG1910-PD|anon-fast-evolving-1A3', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31747643, end_position_on_the_genomic_accession=31751894, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43742, CurrentID=0, Status='live', Symbol='salt', Aliases='Dmel_CG2196, CG2196, Dmel\\CG2196, dSLC5A7', description='salty dog', other_designations='CG2196-PB|CG2196-PC|salt-PB|salt-PC|salty', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31751825, end_position_on_the_genomic_accession=31754705, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43743, CurrentID=0, Status='live', Symbol='CG2187', Aliases='Dmel_ D3-100EF, Dmel\\CG2187, anon-100EF-D3, dSLC5A6', description='CG2187 gene product from transcript CG2187-RA', other_designations='CG2187-PA|CG2187-PB', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31755356, end_position_on_the_genomic_accession=31758637, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43744, CurrentID=0, Status='live', Symbol='Smvt', Aliases='Dmel_CG2191, CG2191, Dmel\\CG2191, SMVT, dSLC5A5, smvt', description='Sodium-dependent multivitamin transporter', other_designations='CG2191-PA|Smvt-PA', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31759404, end_position_on_the_genomic_accession=31761579, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43745, CurrentID=0, Status='live', Symbol='CG2150', Aliases='Dmel_ Dmel\\CG2150, GCR 1, GCR-1, GCR1, Pen100DE, anon-Pen100DE, pen100DE', description='CG2150 gene product from transcript CG2150-RA', other_designations='CG2150-PA', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31763304, end_position_on_the_genomic_accession=31764772, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43746, CurrentID=0, Status='live', Symbol='betaGlu', Aliases='Dmel_CG2135, CG2135, Dmel\\CG2135, Glc, Glu, beta-GUS, gus', description='beta glucuronidase', other_designations='CG2135-PA|beta-Glucuronidase|betaGlu-PA', map_location='100D2-100F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31769293, end_position_on_the_genomic_accession=31771848, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43747, CurrentID=0, Status='live', Symbol='sip3', Aliases='Dmel_CG1937, CG1937, Dmel\\CG1937, HRD1, Hrd1, Sip3, dSyno, hrd1', description='septin interacting protein 3', other_designations='CG1937-PA|CG1937-PB|septin-interacting protein 3|sip3-PA|sip3-PB|synoviolin', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31772219, end_position_on_the_genomic_accession=31775163, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43748, CurrentID=0, Status='live', Symbol='CG2126', Aliases='Dmel_ Dmel\\CG2126, anon-WO0118547.412', description='CG2126 gene product from transcript CG2126-RA', other_designations='CG2126-PA|CG2126-PB', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31775049, end_position_on_the_genomic_accession=31776652, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43749, CurrentID=0, Status='live', Symbol='faf', Aliases='Dmel_CG1945, BcDNA.LD22582, BcDNA:LD22582, CG1945, Dmel\\CG1945, Faf, Fat, P55824', description='fat facets', other_designations='CG1945-PA|CG1945-PC|CG1945-PD|CG1945-PE|CG1945-PG|CG1945-PH|CG1945-PI|Fat facets|faf-PA|faf-PC|faf-PD|faf-PE|faf-PG|faf-PH|faf-PI|fat-facets', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31777020, end_position_on_the_genomic_accession=31790681, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  43750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43750, CurrentID=0, Status='live', Symbol='CG2118', Aliases='Dmel_ Dmel\\CG2118, MCC, Q9V9T5_DROME, mcc', description='CG2118 gene product from transcript CG2118-RC', other_designations='3-methyl-crotonyl-CoA carboxylase|CG2118-PA|CG2118-PB|CG2118-PC|CG2118-PD', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31790592, end_position_on_the_genomic_accession=31793907, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43751, CurrentID=0, Status='live', Symbol='Acf', Aliases='Dmel_CG1966, 175K protein, ACF, ACF/CHRAC, ACF1-1, Acf1, CAF, CG1966, CHRAC, CHRAC-175, Dmel\\CG1966, acf1, dACF, dCHRAC, p170, p170/p185, p175, Acf', description='ATP-dependent chromatin assembly factor large subunit', other_designations='ATP-utilizing chromatin assembly and remodeling factor|Acf-PA|CG1966-PA|chromatin accessibility complex', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31793981, end_position_on_the_genomic_accession=31799799, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43753, CurrentID=0, Status='live', Symbol='Ir100a', Aliases='Dmel_CG11575, CG11575, DmelIR100a, Dmel\\CG11575, IR100a', description='Ionotropic receptor 100a', other_designations='CG11575-PA|Ir100a-PA|ionotropic receptor 100a', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31800963, end_position_on_the_genomic_accession=31802774, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43754, CurrentID=0, Status='live', Symbol='CG42233', Aliases='Dmel_ CG11574, CG2112, Dmel\\CG42233, Dmel_CG11574, Dmel_CG2112', description='CG42233 gene product from transcript CG42233-RB', other_designations='CG42233-PB|uncharacterized protein', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31799634, end_position_on_the_genomic_accession=31805278, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43755 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43755, CurrentID=0, Status='live', Symbol='CG1971', Aliases='Dmel_ Dmel\\CG1971', description='CG1971 gene product from transcript CG1971-RA', other_designations='CG1971-PA|CG1971-PB|CG1971-PC|CG1971-PD', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31805468, end_position_on_the_genomic_accession=31806769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43756 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43756, CurrentID=0, Status='live', Symbol='CG11576', Aliases='Dmel_ Dmel\\CG11576', description='CG11576 gene product from transcript CG11576-RA', other_designations='CG11576-PA|uncharacterized protein', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31806867, end_position_on_the_genomic_accession=31810186, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43757, CurrentID=0, Status='live', Symbol='FeCH', Aliases='Dmel_CG2098, CG2098, Dmel\\CG2098, ferrochelatase', description='Ferrochelatase', other_designations='CG2098-PA|CG2098-PB|CG2098-PC|FeCH-PA|FeCH-PB|FeCH-PC|ferrochelatase', map_location='100D2-100D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31810460, end_position_on_the_genomic_accession=31812296, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43758, CurrentID=0, Status='live', Symbol='RhoGAP100F', Aliases='Dmel_CG1976, 1976, CG1976, DSyd-1, DSyd1, Dmel\\CG1976, RHOGAP100F, Syd-1, dsyd-1, syd-1', description='Rho GTPase activating protein at 100F', other_designations='CG1976-PB|CG1976-PC|CG1976-PD|CG1976-PE|CG1976-PF|CG1976-PH|CG1976-PI|CG1976-PJ|RhoGAP100F-PB|RhoGAP100F-PC|RhoGAP100F-PD|RhoGAP100F-PE|RhoGAP100F-PF|RhoGAP100F-PH|RhoGAP100F-PI|RhoGAP100F-PJ|synpase defective-1', map_location='100D2-100D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31812609, end_position_on_the_genomic_accession=31846570, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  43761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43761, CurrentID=0, Status='live', Symbol='CG2003', Aliases='Dmel_ Dmel\\CG2003', description='CG2003 gene product from transcript CG2003-RA', other_designations='CG2003-PA|CG2003-PB|CG2003-PC', map_location='100E1-100E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31978121, end_position_on_the_genomic_accession=31990909, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43762, CurrentID=0, Status='live', Symbol='CG2053', Aliases='Dmel_ Dmel\\CG2053', description='CG2053 gene product from transcript CG2053-RB', other_designations='CG2053-PA|CG2053-PB', map_location='100E3-100E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=32038330, end_position_on_the_genomic_accession=32039753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835635, CurrentID=0, Status='live', Symbol='CG44622', Aliases='Dmel_ Dmel\\CG44622, lincRNA.354', description='CG44622 gene product from transcript CG44622-RA', other_designations='CG44622-PA|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19953795, end_position_on_the_genomic_accession=19954645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43764, CurrentID=0, Status='live', Symbol='mod', Aliases='Dmel_CG2050, 0952/14, CG2050, Dmel\\CG2050, Mod, ms(3)100EF', description='modulo', other_designations='CG2050-PA|CG2050-PC|male sterile(3)100EF|mod-PA|mod-PC', map_location='100E3-100E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=32052060, end_position_on_the_genomic_accession=32054974, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43765, CurrentID=0, Status='live', Symbol='Map205', Aliases='Dmel_CG1483, 205-kDa MAP, 205K MAP, 205kD MAP, 205kDa MAP, CG1483, Dmel\\CG1483, MAP205, MAP4, map205', description='Microtubule-associated protein 205', other_designations='CG1483-PA|CG1483-PB|CG1483-PC|Map205-PA|Map205-PB|Map205-PC', map_location='100E3-100E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=32055279, end_position_on_the_genomic_accession=32068441, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43766, CurrentID=0, Status='live', Symbol='PlexB', Aliases='Dmel_CG17245, CG17245, Dmel\\CG17245, PLEXB, plex, plex B, plexB', description='Plexin B', other_designations='CG17245-PA|CG17245-PB|PlexB-PA|PlexB-PB|Plexin-B|plexin|plexin B|plexin and met-hepatocyte growth factor receptor-like|plexinB', map_location='102A1-102A1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=32478, end_position_on_the_genomic_accession=43778, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43767, CurrentID=0, Status='live', Symbol='ci', Aliases='Dmel_CG2125, CG2125, CI, CID, Ce, Ci, Ci/GLI, Ci/Gli, Ci155, CiD, Ci[D], Dmel\\CG2125, Gli, Siah-D, ci155, ciD, ci[D], l(4)102ABc, l(4)13, l(4)17, ci', description='cubitus interruptus', other_designations='CG2125-PA|CG2125-PB|CG2125-PC|cell|ci-PA|ci-PB|ci-PC|cubitis interruptus|cubitus interuptus|cubitus-interruptus|cubitus-interruptus-Dominant', map_location='4-0.0 cM', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=47710, end_position_on_the_genomic_accession=57041, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43768 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43768, CurrentID=0, Status='live', Symbol='RpS3A', Aliases='Dmel_CG2168, C3, CG2168, Dmel\\CG2168, M(4)101, M(4)57g, M(4)[57g], M57g, M[57g], RPS3A, RPS3a, RpS3a, Rps3A, Rsp3A, anon-EST:Liang-2.29, chr4:86830..87021, chr4:87101..87234, chr4:87503..87595, clone 2.29, l(4)102ABa', description='Ribosomal protein S3A', other_designations='CG2168-PA|CG2168-PB|CG2168-PF|Minute(4)101|RpS3A-PA|RpS3A-PB|RpS3A-PF|clone 3|ribosomal protein S3A|ribosomal protein S3a|ribosomal protein S3alpha', map_location='101F1-101F1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=65225, end_position_on_the_genomic_accession=67207, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43769, CurrentID=0, Status='live', Symbol='pan', Aliases='Dmel_CG34403, CG17964, CG32005, CG34403, DTCF, DTcf, Dm Pan, Dmel\\CG34403, Dmel_CG17964, Dmel_CG32005, IA5, LEF-1, LEF/TCF, LEF/TCF-1, LEF1, LEF1/TCF, Lef, Lef1, PAN, Pan, TCF, TCF/LEF, TCF/LEF1, Tcf, Tcf-1, Tcf/LEF, cTCF, d-TCF, dTCF, dTcf, l(4)102ABb, l(4)13, lef1.dTCF, tcf, pan', description='pangolin', other_designations='CG34403-PA|CG34403-PB|CG34403-PD|CG34403-PH|CG34403-PI|CG34403-PJ|CG34403-PQ|CG34403-PR|CG34403-PS|T cell factor|T-cell factor|TCF|lethal 13|lymphocyte enhancer binding factor 1/T cell factor|pan-PA|pan-PB|pan-PD|pan-PH|pan-PI|pan-PJ|pan-PQ|pan-PR|pan-PS', map_location='102A3-102A4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=69326, end_position_on_the_genomic_accession=114270, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  43770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43770, CurrentID=0, Status='live', Symbol='Ank', Aliases='Dmel_CG1651, ANK, CG1651, CG32008, DAnk1, Dank1, Dmel\\CG1651, ank, dank1', description='Ankyrin', other_designations='Ank-PB|Ank-PC|Ank-PD|Ank-PE|CG1651-PB|CG1651-PC|CG1651-PD|CG1651-PE|ankyrin', map_location='102A3-102A3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=115074, end_position_on_the_genomic_accession=129849, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  43771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43771, CurrentID=0, Status='live', Symbol='Arl4', Aliases='Dmel_CG2219, ARL4B, BcDNA:AT26185, CG 2219, CG2219, Dmel\\CG2219, Q9V4B2', description='ADP ribosylation factor-like 4', other_designations='Arl4-PA|Arl4-PB|Arl4-PC|CG2219-PA|CG2219-PB|CG2219-PC', map_location='102A6-102A6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=177348, end_position_on_the_genomic_accession=179389, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43772, CurrentID=0, Status='live', Symbol='CG2316', Aliases='Dmel_ Dmel\\CG2316', description='CG2316 gene product from transcript CG2316-RB', other_designations='CG2316-PA|CG2316-PB|CG2316-PC|CG2316-PD|CG2316-PG|CG2316-PH', map_location='102A6-102A6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=184076, end_position_on_the_genomic_accession=193829, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  43774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43774, CurrentID=0, Status='live', Symbol='CG31998', Aliases='Dmel_ CG11572, CG11578, Dmel\\CG31998', description='CG31998 gene product from transcript CG31998-RB', other_designations='CG31998-PA|CG31998-PB|uncharacterized protein', map_location='102A7-102A7', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=196675, end_position_on_the_genomic_accession=205945, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43775, CurrentID=0, Status='live', Symbol='Crk', Aliases='Dmel_CG1587, CG1587, CRK, D-CRK, D-Crk, DCrk, Dcrk, Dmel\\CG1587, crk, dCRK, dCrk', description='Crk oncogene', other_designations='CG1587-PA|CG1587-PB|CG1587-PC|CG1587-PD|CG1587-PE|CG1587-PG|CG1587-PH|Crk-PA|Crk-PB|Crk-PC|Crk-PD|Crk-PE|Crk-PG|Crk-PH', map_location='102A8-102A8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=210224, end_position_on_the_genomic_accession=213165, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43777, CurrentID=0, Status='live', Symbol='CG31999', Aliases='Dmel_ CG10323, CG11288, CT31491, CT7856, Dmel\\CG31999', description='CG31999 gene product from transcript CG31999-RC', other_designations='CG31999-PA|CG31999-PC|fibrillin-like|uncharacterized protein', map_location='102A8-102B1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=215009, end_position_on_the_genomic_accession=227748, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  43779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43779, CurrentID=0, Status='live', Symbol='yellow-h', Aliases='Dmel_CG1629, CG1629, Dmel\\CG1629, h, yell-h', description='CG1629 gene product from transcript CG1629-RA', other_designations='CG1629-PA|yellow-h-PA', map_location='102B1-102B1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=227922, end_position_on_the_genomic_accession=230428, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43780, CurrentID=0, Status='live', Symbol='CG1674', Aliases='Dmel_ Dmel\\CG1674', description='CG1674 gene product from transcript CG1674-RB', other_designations='CG1674-PB|CG1674-PD|CG1674-PE|CG1674-PF|CG1674-PH|CG1674-PI|CG1674-PJ|CG1674-PK|CG1674-PL|CG1674-PM', map_location='102B1-102B1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=230730, end_position_on_the_genomic_accession=245903, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  43781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43781, CurrentID=0, Status='live', Symbol='RhoGAP102A', Aliases='Dmel_CG42316, BP1050, CG1748, CG34013, CG42316, Dm_4:1183, Dmel\\CG42316, Dmel_CG1748, Dmel_CG34013, mlncRNA102B1, rhoGAP102A', description='Rho GTPase activating protein at 102A', other_designations='CG42316-PC|CG42316-PD|CG42316-PE|CG42316-PF|RhoGAP102A-PC|RhoGAP102A-PD|RhoGAP102A-PE|RhoGAP102A-PF', map_location='102B1-102B1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=254997, end_position_on_the_genomic_accession=265574, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  43782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43782, CurrentID=0, Status='live', Symbol='NfI', Aliases='Dmel_CG2380, CG2380, Dmel\\CG2380, NFI', description='Nuclear factor I', other_designations='CG2380-PB|CG2380-PC|NfI-PB|NfI-PC|nuclear factor I', map_location='102B1-102B2', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=268175, end_position_on_the_genomic_accession=276369, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43783, CurrentID=0, Status='live', Symbol='Syt7', Aliases='Dmel_CG2381, CG10322, CG2381, Dmel\\CG2381, SYT7, Syt 7, Syt VII, dsyt7, syt7', description='Synaptotagmin 7', other_designations='CG2381-PA|CG2381-PB|CG2381-PF|CG2381-PI|CG2381-PM|Syt7-PA|Syt7-PB|Syt7-PF|Syt7-PI|Syt7-PM|synaptotagmin 7|synaptotagmin VII', map_location='102B3-102B3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=278685, end_position_on_the_genomic_accession=306481, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  43785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43785, CurrentID=0, Status='live', Symbol='Rad23', Aliases='Dmel_CG1836, CG1836, DHR23, Dmel\\CG1836, RAD23, rad23', description='CG1836 gene product from transcript CG1836-RC', other_designations='CG1836-PA|CG1836-PB|CG1836-PC|Rad23-PA|Rad23-PB|Rad23-PC', map_location='102B3-102B3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=308093, end_position_on_the_genomic_accession=310822, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43786, CurrentID=0, Status='live', Symbol='Zip102B', Aliases='Dmel_CG2177, CG2177, Dmel\\CG2177, cg2177, dZip102B', description='Zinc/iron regulated transporter-related protein 102B', other_designations='CG2177-PB|CG2177-PD|CG2177-PE|CG2177-PF|Zip102B-PB|Zip102B-PD|Zip102B-PE|Zip102B-PF', map_location='102B3-102B3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=310930, end_position_on_the_genomic_accession=314087, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43787, CurrentID=0, Status='live', Symbol='PMCA', Aliases='Dmel_CG42314, BEST:CK, BEST:CK01140, BEST:CK02288, BP1021, CG2165, CG34036, CG42314, CK02288, Dmel\\CG42314, Dmel_CG2165, Dmel_CG34036', description='plasma membrane calcium ATPase', other_designations='CG42314-PI|CG42314-PJ|CG42314-PK|CG42314-PL|CG42314-PM|CG42314-PN|CG42314-PO|CG42314-PP|CG42314-PQ|CG42314-PR|CG42314-PS|PMCA-PI|PMCA-PJ|PMCA-PK|PMCA-PL|PMCA-PM|PMCA-PN|PMCA-PO|PMCA-PP|PMCA-PQ|PMCA-PR|PMCA-PS|plasma membrane Ca2+ ATPase|plasma membrane Ca[2+] ATPase', map_location='102B5-102B5', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=328193, end_position_on_the_genomic_accession=358917, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  43788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43788, CurrentID=0, Status='live', Symbol='Hcf', Aliases='Dmel_CG1710, CG1710, Dmel\\CG1710, HCF, HCF1, HCF_DROME1, anon-WO0172774.48, dHCF, dHcf, dHcf1, Hcf', description='Host cell factor', other_designations='CG1710-PA|CG1710-PB|CG1710-PE|CG1710-PF|CG1710-PG|FBgn0039904|Hcf-PA|Hcf-PB|Hcf-PE|Hcf-PF|Hcf-PG|heat cell factor|host cell factor|host cell factor 1', map_location='102B5-102B6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=359938, end_position_on_the_genomic_accession=375554, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  43789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43789, CurrentID=0, Status='live', Symbol='dati', Aliases='Dmel_CG2052, CG10204, CG2052, DmLin29, Dmel\\CG2052, Lin29', description='datilografo', other_designations='CG2052-PA|CG2052-PB|CG2052-PC|CG2052-PD|dati-PA|dati-PB|dati-PC|dati-PD', map_location='102B6-102B7', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=375640, end_position_on_the_genomic_accession=395605, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  43791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43791, CurrentID=0, Status='live', Symbol='lgs', Aliases='Dmel_CG2041, BCL9, CG2041, Dm Lgs, Dmel\\CG2041, Lgl, Lgs', description='legless', other_designations='CG2041-PA|lgs-PA', map_location='102B8-102C1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=436957, end_position_on_the_genomic_accession=443911, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43792, CurrentID=0, Status='live', Symbol='CaMKI', Aliases='Dmel_CG1495, CG1495, Calcium/calmodulin-dependent protein kinase, CamKI, Dmel\\CG1495, caMKI, camKI, dCKI', description='Calcium/calmodulin-dependent protein kinase I', other_designations='CG1495-PA|CG1495-PB|CG1495-PG|CG1495-PH|CG1495-PI|CaMKI-PA|CaMKI-PB|CaMKI-PG|CaMKI-PH|CaMKI-PI|calmodulin-dependent kinase', map_location='102C1-102C1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=445504, end_position_on_the_genomic_accession=453947, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43793, CurrentID=0, Status='live', Symbol='bip2', Aliases='Dmel_CG2009, BIP2, Bip2, CG2009, DmTAF3, Dmel\\CG2009, TAF155, TAF3, TAFII155, TAF[[II]]155, TFIID, bip-II, dBIP2, dTAFII3, dTAF[[II]]155, dmTAF3, i163, i31', description='CG2009 gene product from transcript CG2009-RA', other_designations='Bric-a-brac interacting protein 2|CG2009-PA|TATA binding protein associated factor 3|bip2-PA|bric a brac interacting protein 2', map_location='102C1-102C1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=459887, end_position_on_the_genomic_accession=466828, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43794, CurrentID=0, Status='live', Symbol='Asator', Aliases='Dmel_CG11533, CG11533, CK I, CKI-like, Dmel\\CG11533', description='CG11533 gene product from transcript CG11533-RI', other_designations='Asator-PD|Asator-PE|Asator-PF|Asator-PG|Asator-PH|Asator-PI|Asator-PJ|Asator-PK|Asator-PL|CG11533-PD|CG11533-PE|CG11533-PF|CG11533-PG|CG11533-PH|CG11533-PI|CG11533-PJ|CG11533-PK|CG11533-PL|asator', map_location='102C1-102C1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=469409, end_position_on_the_genomic_accession=488994, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  43795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43795, CurrentID=0, Status='live', Symbol='zfh2', Aliases='Dmel_CG1449, CG1449, Dmel\\CG1449, ZFH-2, ZFH2, Zfh-2, Zfh2, zfh-2', description='Zn finger homeodomain 2', other_designations='CG1449-PA|CG1449-PB|CG1449-PC|zfh2-PA|zfh2-PB|zfh2-PC|zinc finger homeodomain 2|zinc finger homeodomain-2', map_location='102C3-102C4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=501810, end_position_on_the_genomic_accession=540874, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  43796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43796, CurrentID=0, Status='live', Symbol='Thd1', Aliases='Dmel_CG1981, 1981, CG1981, Dmel\\CG1981, G/T-mismatch-specific-thymine-DNA-glycosylase', description='CG1981 gene product from transcript CG1981-RC', other_designations='CG1981-PA|CG1981-PB|CG1981-PC|Thd1-PA|Thd1-PB|Thd1-PC', map_location='102C5-102C5', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=538332, end_position_on_the_genomic_accession=557854, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43797, CurrentID=0, Status='live', Symbol='Pur-alpha', Aliases='Dmel_CG1507, CG1507, Dmel\\CG1507, PURalpha, Pur alpha, Pur-a, dPur alpha, pur-alpha', description='Purine-rich binding protein-alpha', other_designations='CG1507-PA|CG1507-PB|CG1507-PC|CG1507-PD|CG1507-PE|CG1507-PF|CG1507-PG|Pur-alpha-PA|Pur-alpha-PB|Pur-alpha-PC|Pur-alpha-PD|Pur-alpha-PE|Pur-alpha-PF|Pur-alpha-PG', map_location='102C5-102C5', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=558016, end_position_on_the_genomic_accession=567257, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  43798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43798, CurrentID=0, Status='live', Symbol='ND-49', Aliases='Dmel_CG1970, CG 1970, CG1970, Dmel\\CG1970, NUCM', description='NADH dehydrogenase (ubiquinone) 49 kDa subunit', other_designations='CG1970-PA|CG1970-PB|FBpp0088174|NADH ubiquinone oxidoreductase 49 kDa|NADH-ubiquinone oxidoreductase 49 kDa subunit|NADH:ubiquinone oxidoreductase 49 kDa subunit|ND-49-PA|ND-49-PB', map_location='102C5-102C5', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=568296, end_position_on_the_genomic_accession=571554, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43799, CurrentID=0, Status='live', Symbol='Ephrin', Aliases='Dmel_CG1862, CG1862, Dmel\\CG1862, ephrin, lincRNA.922', description='CG1862 gene product from transcript CG1862-RC', other_designations='CG1862-PA|CG1862-PB|CG1862-PC|Ephrin-PA|Ephrin-PB|Ephrin-PC|dephrin|ephrin', map_location='102C5-102C6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=571755, end_position_on_the_genomic_accession=580122, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43800, CurrentID=0, Status='live', Symbol='CG1909', Aliases='Dmel_ Dmel\\CG1909', description='CG1909 gene product from transcript CG1909-RA', other_designations='CG1909-PA|CG1909-PB', map_location='102C6-102C6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=583108, end_position_on_the_genomic_accession=589348, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43803, CurrentID=0, Status='live', Symbol='Eph', Aliases='Dmel_CG1511, CG1511, CT3831, DEK, Dek, Dek7, Deph, Dmel\\CG1511, Ek7, RPTK Dek7, dek, eph, lincRNA.923', description='Eph receptor tyrosine kinase', other_designations='CG1511-PA|CG1511-PB|CG1511-PC|CG1511-PD|CG1511-PE|CG1511-PF|Eph kinase|Eph receptor|Eph-PA|Eph-PB|Eph-PC|Eph-PD|Eph-PE|Eph-PF|drosophila eph|ephrin receptors', map_location='102D1-102D1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=610684, end_position_on_the_genomic_accession=622470, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  43804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43804, CurrentID=0, Status='live', Symbol='mav', Aliases='Dmel_CG1901, CG1901, Dmel\\CG1901, MAV, Mav, TGF-b', description='maverick', other_designations='CG1901-PA|mav-PA', map_location='102D1-102D1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=623219, end_position_on_the_genomic_accession=626031, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43805, CurrentID=0, Status='live', Symbol='Gat', Aliases='Dmel_CG1732, CG1732, CG1732a, CG6218, DmGAT, Dmel\\CG1732, GAT, anon-WO0149848.11, dGAT, gat', description='GABA transporter', other_designations='CG1732-PA|CG1732-PB|Gat-PA|Gat-PB', map_location='102D1-102D1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=629691, end_position_on_the_genomic_accession=635350, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43806, CurrentID=0, Status='live', Symbol='Ekar', Aliases='Dmel_CG9935, CG9935, CT36399, DmelCG9935, Dmel\\CG9935, ekar', description='Eye-enriched kainate receptor', other_designations='CG9935-PA|CG9935-PB|CG9935-PD|Ekar-PA|Ekar-PB|Ekar-PD|eye-enriched kainate receptor', map_location='102D1-102D1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=636131, end_position_on_the_genomic_accession=646556, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  318597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318597, CurrentID=0, Status='live', Symbol='CG31115', Aliases='Dmel_ Dmel\\CG31115', description='CG31115 gene product from transcript CG31115-RB', other_designations='CG31115-PB|uncharacterized protein', map_location='96B17-96B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25036944, end_position_on_the_genomic_accession=25037982, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43808, CurrentID=0, Status='live', Symbol='gw', Aliases='Dmel_CG31992, CG11484, CG31992, CG9905, Dmel\\CG31992, GQ182, GW, GW182, Gw, anon-EST:Posey205, dGW182182, gw', description='gawky', other_designations='CG31992-PA|CG31992-PB|CG31992-PE|CG31992-PF|CG31992-PI|CG31992-PJ|gawki|gw-PA|gw-PB|gw-PE|gw-PF|gw-PI|gw-PJ', map_location='102D2-102D3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=649041, end_position_on_the_genomic_accession=663103, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  43809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43809, CurrentID=0, Status='live', Symbol='Slip1', Aliases='Dmel_CG1783, CG1783, Dmel\\CG1783, dSLIP1', description='CG1783 gene product from transcript CG1783-RA', other_designations='CG1783-PA|CG1783-PB|CG1783-PC|Slip1-PA|Slip1-PB|Slip1-PC', map_location='102D4-102D4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=663621, end_position_on_the_genomic_accession=669884, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43810, CurrentID=0, Status='live', Symbol='CG11360', Aliases='Dmel_ Dmel\\CG11360, cg11360', description='CG11360 gene product from transcript CG11360-RC', other_designations='CG11360-PA|CG11360-PC|CG11360-PD|CG11360-PE|uncharacterized protein', map_location='102D4-102D4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=671142, end_position_on_the_genomic_accession=690650, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43811, CurrentID=0, Status='live', Symbol='myo', Aliases='Dmel_CG1838, BcDNA:LD02307, CG1838, Dmel\\CG1838, MYO, Myo, Myoglianin, myg', description='myoglianin', other_designations='CG1838-PA|CG1838-PC|CG1838-PD|myo-PA|myo-PC|myo-PD', map_location='102D4-102D4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=690686, end_position_on_the_genomic_accession=696055, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43812, CurrentID=0, Status='live', Symbol='ey', Aliases='Dmel_CG1464, CG1464, DPax-6, Dmel\\CG1464, EY, EYEL, Ey, Ey/Pax6, OK107, Pax-6, Pax6e, l(4)102CDh, l(4)102CDr, l(4)33, pax6, ey', description='eyeless', other_designations='CG1464-PA|CG1464-PB|CG1464-PC|CG1464-PD|ey-PA|ey-PB|ey-PC|ey-PD|tumor-head-63', map_location='4-2.0 cM', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=697689, end_position_on_the_genomic_accession=721173, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43814, CurrentID=0, Status='live', Symbol='bt', Aliases='Dmel_CG32019, 39c-18, Bt, CG10285, CG1479, CG32019, CT3598, CT8086, Dmel\\CG32019, Prj, l(2)2, l(2)23, l(4)102CDa, l(4)2, l(4)21, l(4)23, l(4)37, l(4)38, l(4)PT-2, projectin, titin', description='bent', other_designations='C-protein|CG32019-PC|CG32019-PF|CG32019-PH|CG32019-PI|CG32019-PJ|bt-PC|bt-PF|bt-PH|bt-PI|bt-PJ|lethal(4) Powell, Tennessee-2|myosin LCK|projectin|projectin myosin light chain kinase|twitchin', map_location='4-1.4 cM', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=724400, end_position_on_the_genomic_accession=776474, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  43816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43816, CurrentID=0, Status='live', Symbol='MED26', Aliases='Dmel_CG1793, ARC 70, Arc70, CG1793, CG1823, Dmel\\CG1793, Med26, dARC70', description='Mediator complex subunit 26', other_designations='CG1793-PA|CG1793-PB|MED26-PA|MED26-PB|mediator complex subunit 26|mediator subunit 26', map_location='102E1-102E1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=778516, end_position_on_the_genomic_accession=788978, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43819, CurrentID=0, Status='live', Symbol='pho', Aliases='Dmel_CG17743, CG17743, Dmel\\CG17743, PHO, PHO/YY1, Pho, YY1, dYY1, l(4)102EFc, l(4)29, l(4)BU-2, l(4)OC-1, l(4)pho', description='pleiohomeotic', other_designations='CG17743-PA|CG17743-PB|PHO|peiohomeotic|pho-PA|pho-PB|pleihomeotic|polyhomeotic', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1172469, end_position_on_the_genomic_accession=1181628, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43820, CurrentID=0, Status='live', Symbol='Kif3C', Aliases='Dmel_CG17461, CG17459, CG17461, CP38609, DmKIF3C, DmKif3C, Dmel\\CG17461, KIF 24, KIF 3C, KIF17, KIF3C, KIF3Cdm, Kif3c', description='Kinesin family member 3C ortholog', other_designations='CG17461-PA|Kif3C-PA|kinesin family member 3C ortholog|kinesin superfamily protein 3C', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1147466, end_position_on_the_genomic_accession=1169239, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  43822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43822, CurrentID=0, Status='live', Symbol='CG11155', Aliases='Dmel_ CT30863, DmelCG11155, Dmel\\CG11155', description='CG11155 gene product from transcript CG11155-RD', other_designations='CG11155-PA|CG11155-PB|CG11155-PD|uncharacterized protein', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1126602, end_position_on_the_genomic_accession=1137919, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  43823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43823, CurrentID=0, Status='live', Symbol='Arf102F', Aliases='Dmel_CG11027, ARF2, ARFII, Arf102E, Arf4, Arf5, CG11027, Dmel\\CG11027, P40945, arf102F, dARFII', description='ADP ribosylation factor at 102F', other_designations='ADP ribosylation factor 102F|ADP ribosylation factor II|Arf102F-PA|Arf102F-PB|CG11027-PA|CG11027-PB', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1124167, end_position_on_the_genomic_accession=1125325, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43824, CurrentID=0, Status='live', Symbol='Cals', Aliases='Dmel_CG11059, CALS, CG11059, CT30877, Cad102F, DCad102F, Dmel\\CG11059, cals', description='Calsyntenin-1', other_designations='CG11059-PA|CG11059-PB|CG11059-PC|CG11059-PD|Cals-PA|Cals-PB|Cals-PC|Cals-PD|calsyntenin-1|calsyntenin-1-like', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1112874, end_position_on_the_genomic_accession=1123357, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  43825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43825, CurrentID=0, Status='live', Symbol='sv', Aliases='Dmel_CG11049, CG11049, Cat, D-Pax2, D-pax-2, D-pax2, DPax-2, DPax2, Dmel\\CG11049, Dpax-2, Dpax2, PAX2, Pax2, Pax2/5/8, Pax258, Sv, d-Pax2, dPax2, dPax258, en(lz)4G/I, l(4)40, pax2/sparkling, pol, shaven/sparkling, spa, spa-sv', description='shaven', other_designations='CG11049-PA|CG11049-PC|CG11049-PD|CG11049-PF|CG11049-PG|CG11049-PH|CG11049-PI|cataract|poliert|sparklin poliert|sparkling|sparkling poliert|sv-PA|sv-PC|sv-PD|sv-PF|sv-PG|sv-PH|sv-PI', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1088798, end_position_on_the_genomic_accession=1113317, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  43826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43826, CurrentID=0, Status='live', Symbol='Actbeta', Aliases='Dmel_CG11062, Act, Act-beta, Activin, CG11062, Dmel\\CG11062, act, actbeta, activin-beta, anon-WO02059370.79, beta-Activin, dACT, dAct, dActivin, dact', description='Activin-beta', other_designations='Actbeta-PA|CG11062-PA|activan|activin|activin 102F|activin beta|activin-b|dActivin', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1077331, end_position_on_the_genomic_accession=1084796, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  43827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43827, CurrentID=0, Status='live', Symbol='Rfabg', Aliases='Dmel_CG11064, ApoL1, ApoL2, ApoLI, ApoLII, CG11064, DRBP, Dmel\\CG11064, Lipophorin, Lpp, RFABG, RFABP, RFBP, RfaBP, RfaBp, T3, apoLI, apoLII, apoLPP, apoLpp, apolpp, chr4:1088013..1088173, lipophorin, rfabg', description='Retinoid- and fatty acid-binding glycoprotein', other_designations='CG11064-PA|CG11064-PB|CG11064-PC|CG11064-PD|Retinoid- and fatty acid-binding protein|Retinoid- and fatty-acid binding protein|RfaBp|Rfabg-PA|Rfabg-PB|Rfabg-PC|Rfabg-PD|apolipophorin|apolipophorin I|apolipophorin II|apolipophorin-2|lipophorin|retinoid and fatty acid binding glycoprotein|retinoid and fatty acid binding protein|retinoid and fatty-acid binding protein|retinoid- and fatty acid-binding gene|retinoid- and fatty acid-binding glycoprotein|retinoid-fatty acid binding protein', map_location='102F7-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1060969, end_position_on_the_genomic_accession=1077273, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  43828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43828, CurrentID=0, Status='live', Symbol='CaMKII', Aliases='Dmel_CG18069, CAMKII, CAMKIIalpha, CAMKIId, CDPK1, CG18069, Ca2+/calmodulin-dependent protein kinase II, CaM, CaM II, CaM KII, CaM kinase, CaMK II, CaMK-II, Calcium/calmodulin-dependent protein kinase, Cam, CamKII, CamKIIalpha, Camkii, DCK, Dmel\\CG18069, calcium/calmodulin-dependent protein kinase II, camKII, dCAMKII, dCaMKII, l(4)102EFb, l(4)16, l(4)ar', description='Calcium/calmodulin-dependent protein kinase II', other_designations='CG18069-PA|CG18069-PB|CG18069-PC|CG18069-PD|CG18069-PE|CG18069-PG|CG18069-PH|CG18069-PI|CG18069-PJ|CG18069-PK|CG18069-PL|CG18069-PM|Ca2+ calmodulin-dependent kinase II|Ca2+/calmodulin-dependent protein kinase II|CaM kinase|CaM kinase II|CaM-kinase-II|CaMKII-PA|CaMKII-PB|CaMKII-PC|CaMKII-PD|CaMKII-PE|CaMKII-PG|CaMKII-PH|CaMKII-PI|CaMKII-PJ|CaMKII-PK|CaMKII-PL|CaMKII-PM|Ca[2+]/calmodulin-dependent protein kinase II|Calcium/calmodulin-dependent protein kinase|calcium/ calmodulin-dependent protein kinase II|calcium/calmodulin-dependent protein kinase II|calmodulin kinase II|calmodulin-dependent kinase|calmodulin-dependent protein kinase II|lethal (4) 102EFb', map_location='4-1 cM', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1035288, end_position_on_the_genomic_accession=1053703, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  43829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43829, CurrentID=0, Status='live', Symbol='ATPsynbeta', Aliases='Dmel_CG11154, ATP, ATP syn-beta, ATP-syn-B, ATP-synbeta, ATPB, ATPase beta, ATPasebeta, ATPsyn b, ATPsyn-&bgr, ATPsyn-b, ATPsyn-beta, CG11154, Comp V beta, Dmel\\CG11154, beta-ATPase', description='ATP synthase, beta subunit', other_designations='ATP synthase|ATP synthase beta|ATP synthase beta chain|ATP synthase beta subunit|ATP synthase subunit beta|ATP synthase subunit beta, mitochondrial|ATP synthase subunits beta|ATP synthase-beta|ATPsynbeta-PA|ATPsynbeta-PC|ATPsynbeta-PD|CG11154-PA|CG11154-PC|CG11154-PD|F0/F1 ATP synthase beta|F1ATP synthase|H+-ATPase beta subunit|beta-F1-ATPase|complex V beta', map_location='102F6-102F6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1031484, end_position_on_the_genomic_accession=1034917, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43830, CurrentID=0, Status='live', Symbol='CG11076', Aliases='Dmel_ Dmel\\CG11076', description='CG11076 gene product from transcript CG11076-RE', other_designations='CG11076-PA|CG11076-PD|CG11076-PE|uncharacterized protein', map_location='102F6-102F6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1029711, end_position_on_the_genomic_accession=1031757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43831, CurrentID=0, Status='live', Symbol='CG11077', Aliases='Dmel_ Dmel\\CG11077', description='CG11077 gene product from transcript CG11077-RA', other_designations='CG11077-PA|uncharacterized protein', map_location='102F6-102F6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1028675, end_position_on_the_genomic_accession=1029361, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43832, CurrentID=0, Status='live', Symbol='PlexA', Aliases='Dmel_CG11081, BcDNA:GM05237, CG11081, D-Plex A, DPlexA, Dmel\\CG11081, Plex11, lincRNA.927, plex, plex A, plexA, PlexA', description='Plexin A', other_designations='CG11081-PA|CG11081-PB|CG11081-PC|CG11081-PF|PlexA-PA|PlexA-PB|PlexA-PC|PlexA-PF|PlexinA|plexin|plexin A|plexin-A', map_location='102F5-102F6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1009895, end_position_on_the_genomic_accession=1027101, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43833, CurrentID=0, Status='live', Symbol='toy', Aliases='Dmel_CG11186, CG11186, Dmel\\CG11186, Pax-6, Toy, l(4)102CDg, l(4)8', description='twin of eyeless', other_designations='CG11186-PA|CG11186-PC|lethal(4)102CDg|toy-PA|toy-PC|twin-of-eyeless', map_location='4-0 cM', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=989725, end_position_on_the_genomic_accession=1007922, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43835, CurrentID=0, Status='live', Symbol='fuss', Aliases='Dmel_CG11093, CG11093, CORL, Dmel\\CG11093', description='fussel', other_designations='CG11093-PB|CG11093-PC|CG11093-PD|CORL|fuss-PB|fuss-PC|fuss-PD', map_location='102F4-102F4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=962341, end_position_on_the_genomic_accession=978489, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  43836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43836, CurrentID=0, Status='live', Symbol='4E-T', Aliases='Dmel_CG32016, CG11097, CG11232, CG32016, Dm4E-T, Dmel\\CG32016', description='eIF4E-Transporter', other_designations='4E-T-PB|4E-T-PC|4E-T-PG|4E-T-PH|CG32016-PB|CG32016-PC|CG32016-PG|CG32016-PH', map_location='102F3-102F3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=951542, end_position_on_the_genomic_accession=958392, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43838, CurrentID=0, Status='live', Symbol='mGluR', Aliases='Dmel_CG11144, CAA67993.1_Dr, CG11144, DmGlu-A, DmGlu-RA, DmGluA, DmGluAR, DmGluR, DmGluRA, DmGluR[[A]], DmGlu[[A]]R, Dmel\\CG11144, Glu-RA, GluR, dmGluRA, mGLuRA, mGluR', description='metabotropic Glutamate Receptor', other_designations='CG11144-PA|CG11144-PB|CG11144-PC|CG11144-PD|mGluR-PA|mGluR-PB|mGluR-PC|mGluR-PD|metabotropic glutamate receptor', map_location='102F1-102F3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=936819, end_position_on_the_genomic_accession=949441, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43839, CurrentID=0, Status='live', Symbol='eIF4G', Aliases='Dmel_CG10811, CG10811, Dmel\\CG10811, EIF4G, Eif4G, deIF4G, eIF-4G, eIF4g, p200', description='eukaryotic translation initiation factor 4G', other_designations='CG10811-PA|CG10811-PB|CG10811-PC|Eukaryotic-initiation-factor-4G|eIF-4G|eIF4G-PA|eIF4G-PB|eIF4G-PC|eukaryotic initiation factor 4G|translation initiation factor|translation initiation factor eIF4G', map_location='102F1-102F1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=915297, end_position_on_the_genomic_accession=930715, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  43841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43841, CurrentID=0, Status='live', Symbol='unc-13', Aliases='Dmel_CG2999, CG2999, DUNC-13, Dmel\\CG2999, D Dunc13, Munc-13, Munc-13/UNC-13, UNC-13, UNC-13-B, Unc-13, dUNC-13, dUnc-13, drunc13, dunc-13, l(4)ry16, unc', description='CG2999 gene product from transcript CG2999-RD', other_designations='CG2999-PA|CG2999-PB|CG2999-PC|CG2999-PD|CG2999-PE|CG2999-PF|CG2999-PG|lethal (4) ry16|unc-13-PA|unc-13-PB|unc-13-PC|unc-13-PD|unc-13-PE|unc-13-PF|unc-13-PG', map_location='102E5-102F1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=867824, end_position_on_the_genomic_accession=912331, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  43842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43842, CurrentID=0, Status='live', Symbol='CG11148', Aliases='Dmel_ Dmel\\CG11148', description='CG11148 gene product from transcript CG11148-RH', other_designations='CG11148-PD|CG11148-PF|CG11148-PG|CG11148-PH|uncharacterized protein', map_location='102E5-102E5', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=853557, end_position_on_the_genomic_accession=867025, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43843, CurrentID=0, Status='live', Symbol='fd102C', Aliases='Dmel_CG11152, CG11152, Dmcg11152, Dmel\\CG11152', description='forkhead domain 102C', other_designations='CG11152-PA|fd102C-PA', map_location='102E3-102E3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=836986, end_position_on_the_genomic_accession=844529, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43844, CurrentID=0, Status='live', Symbol='Sox102F', Aliases='Dmel_CG11153, CG11153, Dmel\\CG11153, SOX102D, SOXD, Sox D, Sox102D, SoxD', description='CG11153 gene product from transcript CG11153-RB', other_designations='CG11153-PA|CG11153-PB|CG11153-PC|CG11153-PD|Sox102F-PA|Sox102F-PB|Sox102F-PC|Sox102F-PD', map_location='102E1-102E3', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=800436, end_position_on_the_genomic_accession=834393, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43845, CurrentID=0, Status='live', Symbol='CG11231', Aliases='Dmel_ Dmel\\CG11231', description='CG11231 gene product from transcript CG11231-RA', other_designations='CG11231-PA|uncharacterized protein', map_location='102E1-102E1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=798474, end_position_on_the_genomic_accession=800222, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835723, CurrentID=0, Status='live', Symbol='CG45011', Aliases='Dmel_ Dmel\\CG45011', description='CG45011 gene product from transcript CG45011-RB', other_designations='CG45011-PA|CG45011-PB|uncharacterized protein', map_location='33A2-33A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11834086, end_position_on_the_genomic_accession=11834497, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43852, CurrentID=0, Status='live', Symbol='a', Aliases='Dmel_CG6741, Arc, CG13505, CG6741, Dmel\\CG6741, bran', description='arc', other_designations='CG6741-PA|CG6741-PB|CG6741-PC|CG6741-PD|a-PA|a-PB|a-PC|a-PD|broad angular', map_location='2-99.2 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22136968, end_position_on_the_genomic_accession=22172834, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19835727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835727, CurrentID=0, Status='live', Symbol='CG44247', Aliases='Dmel_ CG30420, CG3749, Dmel\\CG44247', description='CG44247 gene product from transcript CG44247-RB', other_designations='CG44247-PA|CG44247-PB|uncharacterized protein', map_location='60E4-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24876691, end_position_on_the_genomic_accession=24881221, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43856, CurrentID=0, Status='live', Symbol='nej', Aliases='Dmel_CG15319, CBP, CBP/p300, CBP_, CG15319, CG15321, Cbp, Crbp, Dmel\\CG15319, Dmel_CG15321, Nej, Nejire, anon-WO0147981.11, anon-WO03040301.89, cbp, dCBP, dKAT3, dmCBP, l(1)G0112, l(1)G0350, l(1)G0470 CBP, p300, p300/CBP, nej', description='nejire', other_designations='CBP/Nejire|CG15319-PB|CG15319-PC|CG15319-PD|CG15319-PE|CREB binding protein|CREB-binding protein|cyclic AMP response element binding protein|drosophila CREB-binding|drosophila CREB-binding protein|lethal (1) G0112|lethal (1) G0350|lethal (1) G0470|nej-PB|nej-PC|nej-PD|nej-PE|protein', map_location='8F7-8F9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9665442, end_position_on_the_genomic_accession=9687066, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  43862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43862, CurrentID=0, Status='live', Symbol='HnRNP-K', Aliases='Dmel_CG13425, BL, CG13425, Dmel\\CG13425, Hrb57A, Q18, bl, dhnRNP K, hnRNP K, hnRNP-A1L, l(2)k08305', description='Heterogeneous nuclear ribonucleoprotein K', other_designations='CG13425-PA|CG13425-PB|CG13425-PC|CG13425-PD|CG13425-PE|CG13425-PF|HnRNP-K-PA|HnRNP-K-PB|HnRNP-K-PC|HnRNP-K-PD|HnRNP-K-PE|HnRNP-K-PF|bancal|bancal hnRNP|heterogeneous nuclear ribonucleoprotein K|heterogeneous nuclear ribonucleoprotein at 57A', map_location='57A6-57A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20587551, end_position_on_the_genomic_accession=20606854, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  43864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43864, CurrentID=0, Status='live', Symbol='RpL10', Aliases='Dmel_CG17521, BcDNA:LD24589, CG17521, DQM, Dmel\\CG17521, L10, Qm, Rp L10, anon-EST:Posey179, qm', description='Ribosomal protein L10', other_designations='60S ribosomal protein L10|CG17521-PC|CG17521-PD|CG17521-PE|Qm|RpL10-PC|RpL10-PD|RpL10-PE|ribosomal protein L10|ribosome RL10', map_location='80D1-80D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23093190, end_position_on_the_genomic_accession=23095471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43867, CurrentID=0, Status='live', Symbol='Atf3', Aliases='Dmel_CG11405, A3-3, ATF3, CG11405, DM11, Dmel\\CG11405, EG:33C11.1, atf3, dATF3', description='Activating transcription factor 3', other_designations='A3-3|Atf3-PA|Atf3-PB|Atf3-PC|CG11405-PA|CG11405-PB|CG11405-PC|activating transcription factor 3', map_location='1E5-1F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1242740, end_position_on_the_genomic_accession=1275284, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43869, CurrentID=0, Status='live', Symbol='amx', Aliases='Dmel_CG12127, BcDNA:GH02974, CG12127, Dmel\\CG12127, fs(1)amx', description='almondex', other_designations='CG12127-PA|CG12127-PB|amx-PA|amx-PB', map_location='8D2-8D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9245915, end_position_on_the_genomic_accession=9247336, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43872, CurrentID=0, Status='live', Symbol='Ppn', Aliases='Dmel_CG33103, CG1540, CG18436, CG33103, CT3186, CT41988, Dmel\\CG33103, ppl, ppn', description='Papilin', other_designations='CG33103-PC|CG33103-PE|CG33103-PF|CG33103-PG|Ppn-PC|Ppn-PE|Ppn-PF|Ppn-PG|lacunin|lacunin-like|papilin|papilin precursor', map_location='98D3-98D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28520665, end_position_on_the_genomic_accession=28540324, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  43873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43873, CurrentID=0, Status='live', Symbol='ras', Aliases='Dmel_CG1799, CG11485, CG1799, Dmel\\CG1799, EP(X)1093, IMPDH, IMPdH, LD02673, LD06825, Ras, l(1)9Eb, l(1)G0002, l(1)G0056, l(1)G0098, l(1)G0127, l(1)G0238, l(1)G0351, l(1)G0380, l(1)G0388, l(1)G0391, l(1)G0436, l(1)G0482-l, raspberry/impd, ras', description='raspberry', other_designations='CG1799-PA|CG1799-PC|CG1799-PD|CG1799-PE|IMP dehydrogenase|inosine monophosphate dehydrogenase|lethal (1) G0380|ras-PA|ras-PC|ras-PD|ras-PE|raspberry-lethal', map_location='9E1-9E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10744502, end_position_on_the_genomic_accession=10749097, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43875, CurrentID=0, Status='live', Symbol='Ret', Aliases='Dmel_CG14396, CG1061, CG14396, D-ret, DRET, DRet, DmHD-59, DmRet, Dmel\\CG14396, Dret, HD-59, MEN2, RETo, dRET, dRet, dret, ret, Ret', description='Ret oncogene', other_designations='CG14396-PA|CG14396-PB|CG14396-PC|Ret-PA|Ret-PB|Ret-PC|dret', map_location='39B1-39B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21182446, end_position_on_the_genomic_accession=21198762, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19835749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835749, CurrentID=0, Status='live', Symbol='CG44245', Aliases='Dmel_ CG9480, Dmel\\CG44245', description='CG44245 gene product from transcript CG44245-RA', other_designations='CG44245-PA|CG44245-PB|uncharacterized protein', map_location='57D1-57D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21200483, end_position_on_the_genomic_accession=21212969, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19835750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835750, CurrentID=0, Status='live', Symbol='CG44296', Aliases='Dmel_ Dmel\\CG44296', description='CG44296 gene product from transcript CG44296-RA', other_designations='CG44296-PA|uncharacterized protein', map_location='46D9-46D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10050348, end_position_on_the_genomic_accession=10050874, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43892, CurrentID=0, Status='live', Symbol='sif', Aliases='Dmel_CG34418, CG32414, CG34418, CG5256, CG5406, Dm_3L:26493, Dmel\\CG34418, Dmel_CG32414, Dmel_CG5406, RhoGEF, SIF, Sif, lincRNA.447, slf', description='still life', other_designations='CG34418-PA|CG34418-PB|CG34418-PC|CG34418-PF|CG34418-PI|CG34418-PJ|CG34418-PK|CG34418-PL|CG34418-PM|CG34418-PN|CG34418-PO|CG34418-PP|CG34418-PQ|sif-PA|sif-PB|sif-PC|sif-PF|sif-PI|sif-PJ|sif-PK|sif-PL|sif-PM|sif-PN|sif-PO|sif-PP|sif-PQ|still life type 1|still life type 2|still-life', map_location='64E1-64E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5657913, end_position_on_the_genomic_accession=5749599, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  43895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43895, CurrentID=0, Status='live', Symbol='E23', Aliases='Dmel_CG3327, CG15410, CG3327, Dmel\\CG3327', description='Early gene at 23', other_designations='CG3327-PC|CG3327-PD|CG3327-PE|E23-PC|E23-PD|E23-PE|early gene at 23', map_location='23E4-23E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3334875, end_position_on_the_genomic_accession=3354829, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43900, CurrentID=0, Status='live', Symbol='Crg-1', Aliases='Dmel_CG32788, CG32788, Crg1, Dmel\\CG32788, crg1', description='Circadianly Regulated Gene', other_designations='CG32788-PB|Crg-1-PB|circadianly regulated gene', map_location='3F1-3F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3793223, end_position_on_the_genomic_accession=3797339, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43901, CurrentID=0, Status='live', Symbol='chb', Aliases='Dmel_CG32435, 0686/07, BcDNA:LD31673, CG32435, CLASP, Chb, DMAP160, Dmel\\CG32435, MAST, MAST/Orbit, MESR7, Mast, Mast/Orbit, NEST:bs13a06, NEST:bs17d05, NEST:bs28g05, Orbit, Orbit/MAST, Orbit/Mast, anon-WO0104295(v40), chb/mast/orbit, clasp, dCLASP, mast, mast/orbit, orb, orbit, orbit/MAST, orbit/mast, v40, chb', description='chromosome bows', other_designations='CG32435-PA|CG32435-PB|CG32435-PC|MAST/Orbit|Orbit/Mast/CLASP|chb-PA|chb-PB|chb-PC|clasp|mast|misexpression suppressor of ras 7|multiple asters|orbit|orbit/clasp', map_location='3-46.6 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21172853, end_position_on_the_genomic_accession=21183278, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43904, CurrentID=0, Status='live', Symbol='AMPKalpha', Aliases='Dmel_CG3051, AK, AMPK, AmpKalpha, CG3051, DmAMPK alpha, Dmel\\CG3051, EG:132E8.2, FBgn0023169, Gprk-4, Gprk4, SNF1A, ampk, ampkalpha, dAMPK, dAMPKa, dAMPKalpha, snf1A, snf1a', description='AMP-activated protein kinase alpha subunit', other_designations="5'monophosphate-activated kinase|AMP activated kinase|AMP-activated kinase|AMP-activated protein kinase|AMP-dependent kinase|AMPK catalytic subunit|AMPKalpha-PA|AMPKalpha-PB|AMPKalpha-PC|CG3051-PA|CG3051-PB|CG3051-PC|G protein-coupled receptor kinase 4|SNF1A/AMP-activated protein kinase|adenosine monophosphate-activated protein kinase", map_location='2A1-2A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1375536, end_position_on_the_genomic_accession=1378896, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43905, CurrentID=0, Status='live', Symbol='syd', Aliases='Dmel_CG8110, CG8110, Dmel\\CG8110, JIP-3, Syd, jip3', description='sunday driver', other_designations='CG8110-PA|CG8110-PB|CG8110-PC|CG8110-PD|CG8110-PE|CG8110-PF|CG8110-PG|CG8110-PH|CG8110-PI|CG8110-PJ|CG8110-PK|syd-PA|syd-PB|syd-PC|syd-PD|syd-PE|syd-PF|syd-PG|syd-PH|syd-PI|syd-PJ|syd-PK', map_location='66A20-66A20', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7916650, end_position_on_the_genomic_accession=7936120, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  43906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43906, CurrentID=0, Status='live', Symbol='skd', Aliases='Dmel_CG9936, CG9936, Dmel\\CG9936, LD28662, MED13, Med13, Pap, Pap/Trap, SPH203, Scad78, Skd, Skd/Med13, Skd/Pap/Bli, Srb9/AMIB/PAP/TRAP240, TRAP240, Trap240, bli, bls, dMED13, dTRAP240, dTrap240, dmMED13, l(3)L7062, l(3)rK760, pap, pap/dTRAP240, siren9', description='skuld', other_designations='CG9936-PC|CG9936-PD|CG9936-PE|CG9936-PF|blind spot|flytrap|mediator complex subunit 13|poils aux pattes|skd-PC|skd-PD|skd-PE|skd-PF|suppressor of constitutively activated Dpp signaling 78', map_location='3-51 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20992814, end_position_on_the_genomic_accession=21027434, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  43912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43912, CurrentID=0, Status='live', Symbol='SF1', Aliases='Dmel_CG5836, CG5836, Dmel\\CG5836, cg5836, dSF1, p70', description='Splicing factor 1', other_designations='CG5836-PA|CG5836-PB|CG5836-PC|SF1-PA|SF1-PB|SF1-PC|splicing factor 1', map_location='90B4-90B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17540421, end_position_on_the_genomic_accession=17545148, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  43913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43913, CurrentID=0, Status='live', Symbol='Gyk', Aliases='Dmel_CG18374, CG18374, Dmel\\CG18374, dGyk', description='Glycerol kinase', other_designations='CG18374-PA|CG18374-PB|Gyk-PA|Gyk-PB|glycerol kinase', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=202098, end_position_on_the_genomic_accession=205844, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19835787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835787, CurrentID=0, Status='live', Symbol='CG43784', Aliases='Dmel_ CR43784, Dmel\\CG43784', description='CG43784 gene product from transcript CG43784-RB', other_designations='CG43784-PB|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19957548, end_position_on_the_genomic_accession=19958342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  43916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43916, CurrentID=0, Status='live', Symbol='Rok', Aliases='Dmel_CG9774, CG9774, DROK, DRhk, DRok, Dmel\\CG9774, Drok, ROCK, ROCK-1, ROCK1, ROCK2, ROK, ROKalpha/beta, Rhk, RhoK, Rock/ROCK, dROK, dRok, drok, rok, Rok', description='Rho kinase', other_designations='CG9774-PA|CG9774-PB|CG9774-PC|CG9774-PD|DRho-kinase|ROCK|Rho-associated kinase|Rho-associated protein kinase|Rho-kinase|RhoA kinase|Rok-PA|Rok-PB|Rok-PC|Rok-PD|drosophila rho-associated kinase|rho kinase', map_location='14F2-14F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16624012, end_position_on_the_genomic_accession=16637414, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  43918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43918, CurrentID=0, Status='live', Symbol='noe', Aliases='Dmel_CG32172, CG32172, Dmel\\CG32172, Noe, anon-EST:Posey101, anon-EST:Posey151, anon-EST:Posey243, anon-EST:Posey292, anon-EST:Posey35, anon-WO02059370.51', description='CG32172 gene product from transcript CG32172-RC', other_designations='CG32172-PB|CG32172-PC|noe-PB|noe-PC', map_location='74B1-74B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17400347, end_position_on_the_genomic_accession=17401348, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43923, CurrentID=0, Status='live', Symbol='axo', Aliases='Dmel_CG43225, CG13715, CG13717, CG18296, CG32234, CG42664, CG43225, CT1277, Dmel\\CG43225, Dmel_CG13715, Dmel_CG13717, Dmel_CG18296, Dmel_CG32234, Dmel_CG42664', description='axotactin', other_designations='CG43225-PB|CG43225-PC|CG43225-PD|CG43225-PF|CG43225-PG|CG43225-PH|CG43225-PI|axo-PB|axo-PC|axo-PD|axo-PF|axo-PG|axo-PH|axo-PI', map_location='64B12-64B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4629604, end_position_on_the_genomic_accession=4687317, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  43924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43924, CurrentID=0, Status='live', Symbol='jim', Aliases='Dmel_CG11352, BcDNA:SD14927, CG11352, Dmel\\CG11352, Jim, OVK, ovfc.K, ovk', description='CG11352 gene product from transcript CG11352-RC', other_designations='CG11352-PC|CG11352-PD|CG11352-PE|CG11352-PF|CG11352-PG|CG11352-PH|CG11352-PI|jim-PC|jim-PD|jim-PE|jim-PF|jim-PG|jim-PH|jim-PI', map_location='80A2-80A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22789233, end_position_on_the_genomic_accession=22835194, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43926, CurrentID=0, Status='live', Symbol='fan', Aliases='Dmel_CG7919, CG7919, Dmel\\CG7919', description='farinelli', other_designations='CG7919-PA|fan-PA', map_location='66B6-66B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8109132, end_position_on_the_genomic_accession=8110043, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  43927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43927, CurrentID=0, Status='live', Symbol='fal', Aliases='Dmel_CG9670, CG9670, Dmel\\CG9670, Q9VVY4_DROME', description='falten', other_designations='CG9670-PB|CG9670-PC|CG9670-PD|fal-PB|fal-PC|fal-PD', map_location='76A5-76A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19294564, end_position_on_the_genomic_accession=19298830, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43928, CurrentID=0, Status='live', Symbol='chm', Aliases='Dmel_CG5229, CG5229, Chm, Dmel\\CG5229, HAT, HAT1, HBO1, dmHAG404', description='chameau', other_designations='CG5229-PA|CG5229-PB|CG5229-PC|chm-PA|chm-PB|chm-PC', map_location='27F3-27F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7412417, end_position_on_the_genomic_accession=7422736, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  43930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43930, CurrentID=0, Status='live', Symbol='prc', Aliases='Dmel_CG5700, BEST:CK01593, CG5700, CK01593, Dmel\\CG5700, EC11, PC, PRC, Pc, Prc, pc', description='pericardin', other_designations='CG5700-PB|prc-PB', map_location='68E3-68E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11958872, end_position_on_the_genomic_accession=11966131, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  43934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43934, CurrentID=0, Status='live', Symbol='gom', Aliases='Dmel_CG6727, CG6727, Dmel\\CG6727', description='gomdanji', other_designations='CG6727-PA|gom-PA', map_location='58C1-58C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22139935, end_position_on_the_genomic_accession=22141364, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  43935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43935, CurrentID=0, Status='live', Symbol='NtR', Aliases='Dmel_CG6698, CG6698, D Dmel\\CG6698, anon-WO0138359.3', description='CG6698 gene product from transcript CG6698-RA', other_designations='CG6698-PA|CG6698-PB|NtR-PA|NtR-PB', map_location='58B10-58C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22127808, end_position_on_the_genomic_accession=22130184, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43936, CurrentID=0, Status='live', Symbol='Men-b', Aliases='Dmel_CG5889, CG5889, Dmel\\CG5889, MDH, Mdh, anon-EST:Posey79', description='Malic enzyme b', other_designations='CG5889-PA|CG5889-PB|CG5889-PC|Mdh|Men-b-PA|Men-b-PB|Men-b-PC|malic DH|malic enzyme b', map_location='97E5-97E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27151662, end_position_on_the_genomic_accession=27157603, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  43938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43938, CurrentID=0, Status='live', Symbol='cl', Aliases='Dmel_CG11024, BcDNA:RE06767, CG11024, Dmel\\CG11024', description='clot', other_designations='CG11024-PA|CG11024-PB|CG11024-PC|cl-PA|cl-PB|cl-PC', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5520135, end_position_on_the_genomic_accession=5521382, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43943, CurrentID=0, Status='live', Symbol='or', Aliases='Dmel_CG3029, AP-3sigma, CG3029, Dmel\\CG3029, Or, aps3, cg3029, sigma3', description='orange', other_designations='CG3029-PA|adaptor protein complex 3, sigma subunit|or-PA', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23967939, end_position_on_the_genomic_accession=23968943, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  43944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43944, CurrentID=0, Status='live', Symbol='sbr', Aliases='Dmel_CG1664, CG1664, CG17335, CT4634, DNXF-1, Dm nxf1, DmNXF1, Dmel\\CG1664, NFX1, NXF1, NXF1-p15, Nxf1, PABPr, dTap, l(1)9Fd, l(1)G1, l(1)ts, l(1)ts403, nxf1, sbt', description='small bristles', other_designations='CG1664-PA|locus 1|nuclear export factor 1|sbr-PA|shorter-bristles', map_location='9F4-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10832752, end_position_on_the_genomic_accession=10847092, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  43946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43946, CurrentID=0, Status='live', Symbol='wb', Aliases='Dmel_CG42677, &agr1, 2, 2Lan, 2]], A1, BEST:CK02229, BG:DS03792.1, BG:DS03792.2, CG15288, CG32970, CG42677, CK02229, CT35236, DLAM, Dmel\\CG42677, Dmel_CG15288, Dmel_CG32970, LM-A/alpha2, LanA1, alpha[[1, fus5, l(2)09437, l(2)34Fb, l(2)br1, l34Fb', description='wing blister', other_designations='CG42677-PB|CG42677-PC|CG42677-PD|CG42677-PE|CG42677-PF|CG42677-PG|D-laminin alpha2|SGP cluster fusion defects 5|alpha laminin|laminin|laminin A2|laminin alpha1,2|laminin alpha[[1,2]]|lamininW|lethal 1 in the black-reduced region|lethal group A1|wb-PB|wb-PC|wb-PD|wb-PE|wb-PF|wb-PG|wing blistered', map_location='35A3-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14263027, end_position_on_the_genomic_accession=14328262, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  43950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43950, CurrentID=0, Status='live', Symbol='CdsA', Aliases='Dmel_CG7962, CDS, CG7962, Dmel\\CG7962, cds, cdsA, eye-CDS, eye-cds', description='CDP diglyceride synthetase', other_designations='CDP-DAG synthase|CDP-diacylglycerol synthase|CDP-diacylglycerol synthetase|CDS-diacylglycerol synthase|CG7962-PA|CG7962-PB|CdsA-PA|CdsA-PB|cytidine diphosphate (CDP)-DAG synthase|diacyl-glycerol-synthetase|winded', map_location='66B7-66B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8127762, end_position_on_the_genomic_accession=8131675, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  43951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43951, CurrentID=0, Status='live', Symbol='e(r)', Aliases='Dmel_CG1871, CG1871, Dmel\\CG1871, ER, ER_DROME', description='enhancer of rudimentary', other_designations='CG1871-PA|CG1871-PB|CG1871-PC|CG1871-PD|CG1871-PE|e(r)-PA|e(r)-PB|e(r)-PC|e(r)-PD|e(r)-PE', map_location='8B7-8B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8904144, end_position_on_the_genomic_accession=8906248, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43952, CurrentID=0, Status='live', Symbol='MED20', Aliases='Dmel_CG18780, CG18267, CG18780, Dmel\\CG18780, Med20, Srb2/TRFP, Tmr, TrfP, Trfp, dTrap26, dTrfp, l(2)28DE, med20, p29', description='Mediator complex subunit 20', other_designations='CG18780-PA|CG18780-PB|MED20-PA|MED20-PB|Trf-proximal|lethal(2)28DE|mediator|mediator complex subunit 20|transcription mediator-related', map_location='28E1-28E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8071812, end_position_on_the_genomic_accession=8072848, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43958, CurrentID=0, Status='live', Symbol='spag', Aliases='Dmel_CG13570, CG13570, Dmel\\CG13570, Spag, l(2)k12101', description='spaghetti', other_designations='CG13570-PA|spag-PA', map_location='60C1-60C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24159384, end_position_on_the_genomic_accession=24161503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3355124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355124, CurrentID=0, Status='live', Symbol='RpL5', Aliases='Dmel_CG17489, CG17489, Dmel\\CG17489, E-2d, L5, M(2)40B, Rp L5, Rpl5, dRPL5, yip6', description='Ribosomal protein L5', other_designations='CG17489-PA|CG17489-PB|CG17489-PG|CG17489-PH|RpL5-PA|RpL5-PB|RpL5-PG|RpL5-PH|enhancer on chromosome 2, complementation group d|ribosomal protein L5|yippee interacting protein 6', map_location='40F7-40F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22532279, end_position_on_the_genomic_accession=22536884, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43973, CurrentID=0, Status='live', Symbol='Lip1', Aliases='Dmel_CG7279, CG7279, Dmel\\CG7279, Dmlip1, LIP1Dm, Lip, dLip1, dilp1, lip1', description='Lipase 1', other_designations='CG7279-PA|CG7279-PB|Lip1-PA|Lip1-PB|Lipase1|lipase 1', map_location='32A4-32A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10699408, end_position_on_the_genomic_accession=10701544, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43974, CurrentID=0, Status='live', Symbol='trio', Aliases='Dmel_CG18214, 0368/10, 0959/14, 1372/03, 1386/06, BEST:LD36950, CG18214, CG9208, DTrio, Dmel\\CG18214, M89, TRIO, Trio, UNC-73/Trio, dTrio, l(3)036810, l(3)6D104, l(3)S036810, l(3)S095914, l(3)S137203, l(3)S138606, l(3)S[1372/3], l(3)S[1386/06], l(3)', description='CG18214 gene product from transcript CG18214-RA', other_designations='CG18214-PA|CG18214-PB|CG18214-PC|CG18214-PD|CG18214-PE|CG18214-PF|trio-PA|trio-PB|trio-PC|trio-PD|trio-PE|trio-PF', map_location='61E1-61E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=995982, end_position_on_the_genomic_accession=1034875, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  43978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43978, CurrentID=0, Status='live', Symbol='Plap', Aliases='Dmel_CG5105, CG5105, Dmel\\CG5105, UFD3, dPLAP, dmPLAP, plap', description='Phospholipase A2 activator protein', other_designations='CG5105-PA|CG5105-PB|CG5105-PC|Plap-PA|Plap-PB|Plap-PC|phospholipase A2 activator protein', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1167172, end_position_on_the_genomic_accession=1170365, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43980, CurrentID=0, Status='live', Symbol='Lip2', Aliases='Dmel_CG17116, CG17116, Dmel\\CG17116, Dmlip2, dilp2, lip2', description='Lipase 2', other_designations='CG17116-PA|CG17116-PB|Lip2-PA|Lip2-PB|lipase 2', map_location='32A4-32A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10701940, end_position_on_the_genomic_accession=10709453, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  43981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43981, CurrentID=0, Status='live', Symbol='DIP1', Aliases='Dmel_CG17686, CG17686, Dmel\\CG17686, TO34, TO67, UbxBP1, clot 10495, clot 2020, dip1, dip1a, klt', description='DISCO Interacting Protein 1', other_designations='CG17686-PB|CG17686-PC|CG17686-PD|CG17686-PE|CG17686-PF|CG17686-PG|CG17686-PH|CG17686-PI|DIP1-PB|DIP1-PC|DIP1-PD|DIP1-PE|DIP1-PF|DIP1-PG|DIP1-PH|DIP1-PI|DISCO interacting protein 1|disconnected interacting protein 1|klett|ultrabithorax binding protein 1', map_location='20A4-20A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21624796, end_position_on_the_genomic_accession=21630148, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  43982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43982, CurrentID=0, Status='live', Symbol='Oamb', Aliases='Dmel_CG3856, CG15698, CG3856, Dm-OAMB, DmOAMB, DmOCT1B, DmOctalpha1Rb, Dmel\\CG3856, Dmoa1, GPCR, OAMB, OAMB-1a, OAMB-AS, OAMB-K3 a, Oamb b, OctR, OctR1, oamb, Oamb', description='Octopamine receptor in mushroom bodies', other_designations='CG3856-PB|CG3856-PC|CG3856-PD|CG3856-PE|CG3856-PF|CG3856-PG|Oamb-PB|Oamb-PC|Oamb-PD|Oamb-PE|Oamb-PF|Oamb-PG|octopamine mushroom body receptor|octopamine receptor|octopamine receptor OAMB|octopamine receptor in mushroom bodies|octopamine-activated G-protein-coupled receptor', map_location='92F6-92F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20689018, end_position_on_the_genomic_accession=20714577, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  43984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43984, CurrentID=0, Status='live', Symbol='Sdic1', Aliases='Dmel_CG9580, CG9580, Dmel\\CG9580, Sdic:CG9580', description='Sperm-specific dynein intermediate chain 1', other_designations='CG9580-PA|CG9580-PC|Sdic1-PA|Sdic1-PC|Sdic:CG9580', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20174217, end_position_on_the_genomic_accession=20178638, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  318627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318627, CurrentID=0, Status='live', Symbol='CG31206', Aliases='Dmel_ Dmel\\CG31206, NEST:bs16c12', description='CG31206 gene product from transcript CG31206-RA', other_designations='CG31206-PA|CG31206-PB|uncharacterized protein', map_location='92D1-92D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20106936, end_position_on_the_genomic_accession=20107963, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19835861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835861, CurrentID=0, Status='live', Symbol='CG44532', Aliases='Dmel_ Dmel\\CG44532, lincRNA.970', description='CG44532 gene product from transcript CG44532-RA', other_designations='CG44532-PA|uncharacterized protein', map_location='8C14-8C15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9162092, end_position_on_the_genomic_accession=9163153, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43990, CurrentID=0, Status='live', Symbol='Sep5', Aliases='Dmel_CG2916, CG2916, Dmel\\CG2916, Sep, sep5', description='Septin 5', other_designations='CG2916-PA|CG2916-PB|Sep5-PA|Sep5-PB|Septin-5|septin|septin 5', map_location='43F8-43F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7960133, end_position_on_the_genomic_accession=7961921, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  43991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43991, CurrentID=0, Status='live', Symbol='Pten', Aliases='Dmel_CG5671, CG5671, DPTEN, Dmel\\CG5671, PTEN, PTEN3, dPTEN, dPten, dpten, pten', description='Phosphatase and tensin homolog', other_designations='CG5671-PA|CG5671-PB|CG5671-PC|CG5671-PD|CG5671-PE|CG5671-PG|CG5671-PH|PI 3-phosphatase|PTEN|PtdIns(3,4,5)P3 phosphatase|Pten-PA|Pten-PB|Pten-PC|Pten-PD|Pten-PE|Pten-PG|Pten-PH|drosophila phosphatase and tensin homologue deleted on chromosome 10|phosphatase and tensin homolog|phosphatase and tensin homologue|phosphatidylinositol (3,4,5)-trisphosphate phosphatase', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10256319, end_position_on_the_genomic_accession=10261049, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19835865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835865, CurrentID=0, Status='live', Symbol='CG44253', Aliases='Dmel_ CG30194, Dmel\\CG44253', description='CG44253 gene product from transcript CG44253-RA', other_designations='CG44253-PA|uncharacterized protein', map_location='59C2-59C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23058248, end_position_on_the_genomic_accession=23058793, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835868, CurrentID=0, Status='live', Symbol='CG44271', Aliases='Dmel_ Dmel\\CG44271', description='CG44271 gene product from transcript CG44271-RA', other_designations='CG44271-PA|uncharacterized protein', map_location='27B1-27B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6747421, end_position_on_the_genomic_accession=6747914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  43997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43997, CurrentID=0, Status='live', Symbol='jbug', Aliases='Dmel_CG30092, AAF46896, CG11605, CG13525, CG30092, CG3550, Dmel\\CG30092, Jbug', description='jitterbug', other_designations='CG30092-PC|CG30092-PF|CG30092-PG|CG30092-PH|CG30092-PI|CG30092-PJ|CG30092-PK|CG30092-PL|CG30092-PM|CG30092-PN|filamin|jbug-PC|jbug-PF|jbug-PG|jbug-PH|jbug-PI|jbug-PJ|jbug-PK|jbug-PL|jbug-PM|jbug-PN', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22717321, end_position_on_the_genomic_accession=22740459, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  43999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43999, CurrentID=0, Status='live', Symbol='trbl', Aliases='Dmel_CG5408, CG5408, Dmel\\CG5408, Trbl, trb', description='tribbles', other_designations='CG5408-PA|trbl-PA', map_location='77C1-77C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20396288, end_position_on_the_genomic_accession=20401609, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44000, CurrentID=0, Status='live', Symbol='Spt6', Aliases='Dmel_CG12225, CG12225, Dmel\\CG12225, Dspt6, SPT6, d dspt6, l(1)G0063, spt 6, spt6', description='CG12225 gene product from transcript CG12225-RA', other_designations='CG12225-PA|CG12225-PB|Spt6-PA|Spt6-PB', map_location='5E4-5E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6267797, end_position_on_the_genomic_accession=6274984, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  44001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44001, CurrentID=0, Status='live', Symbol='Sucb', Aliases='Dmel_CG10622, 142911_at, CG10622, Dmel\\CG10622, sucB, sucb', description='CG10622 gene product from transcript CG10622-RA', other_designations='CG10622-PA|CG10622-PC|Sucb-PA|Sucb-PC|sucb', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5561701, end_position_on_the_genomic_accession=5563691, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44002, CurrentID=0, Status='live', Symbol='Sras', Aliases='Dmel_CG4852, CG4852, Dmel\\CG4852, FAC2_DROME', description='severas', other_designations='CG4852-PA|Sras-PA|prenyl protease type II', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5563774, end_position_on_the_genomic_accession=5565103, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835876, CurrentID=0, Status='live', Symbol='CG44261', Aliases='Dmel_ Dmel\\CG44261', description='CG44261 gene product from transcript CG44261-RA', other_designations='CG44261-PA|uncharacterized protein', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9407694, end_position_on_the_genomic_accession=9408169, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44005, CurrentID=0, Status='live', Symbol='NHP2', Aliases='Dmel_CG5258, CG5258, Dmel\\CG5258, FBgn0029148', description='CG5258 gene product from transcript CG5258-RB', other_designations='CG5258-PA|CG5258-PB|NHP2-PA|NHP2-PB', map_location='70F6-70F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14801328, end_position_on_the_genomic_accession=14802168, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44006, CurrentID=0, Status='live', Symbol='Fad2', Aliases='Dmel_CG7923, CG7923, Dmel\\CG7923, desatF, desatF-alpha, fad, fad2', description='CG7923 gene product from transcript CG7923-RA', other_designations='CG7923-PA|Fad2-PA', map_location='68A1-68A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11023535, end_position_on_the_genomic_accession=11024846, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44007, CurrentID=0, Status='live', Symbol='COQ7', Aliases='Dmel_CG14437, CG14437, Dclk1, Dmel\\CG14437', description='CG14437 gene product from transcript CG14437-RA', other_designations='CG14437-PA|COQ7-PA', map_location='6C13-6C13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6771225, end_position_on_the_genomic_accession=6773039, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44008, CurrentID=0, Status='live', Symbol='Hex-C', Aliases='Dmel_CG8094, CG8094, CT22069, DM2, Dmel\\CG8094, Fk, HEX, HEX-3, HEX-C, HK, HK-C, Hex, Hex-3, Hex-c, HexC, Hk', description='Hexokinase C', other_designations='CG8094-PA|Hex-C-PA|Hexokinase-3|HexokinaseC|fructokinase|hexokinase|hexokinase C|hexokinase-C|hexose kinase', map_location='51E7-51E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15218837, end_position_on_the_genomic_accession=15220329, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44010, CurrentID=0, Status='live', Symbol='Pgm', Aliases='Dmel_CG5165, CG5165, Dmel\\CG5165, PGM, PGM-1-1, pgm, Pgm', description='phosphoglucose mutase', other_designations='CG5165-PA|Pgm-PA|phosphoglucomutase|phosphogluconate mutase', map_location='72D8-72D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16126018, end_position_on_the_genomic_accession=16128531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44011, CurrentID=0, Status='live', Symbol='pwn', Aliases='Dmel_CG11101, CG11101, CT31067, Dmel\\CG11101, l(2)pwn', description='pawn', other_designations='CG11101-PB|CG11101-PC|N-like|pwn-PB|pwn-PC', map_location='2-55.4 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7282572, end_position_on_the_genomic_accession=7300806, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  44013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44013, CurrentID=0, Status='live', Symbol='sls', Aliases='Dmel_CG1915, 0020/01, CG18242, CG18245, CG18857, CG1915, CT41299, D-Titin, D-titin, Dmel\\CG1915, KZ, Ket, MCP, SLS, Sls, TITIN, Titin, anon-CREST, d-titin, ket, kettin, l(3)62Ca, l(3)Ca, l(3)S002001, l(3)dre8, l(3)j1D7, l(3)rL182, sal, sam, titin', description='sallimus', other_designations='CG1915-PA|CG1915-PD|CG1915-PP|CG1915-PQ|CG1915-PR|CG1915-PS|CG1915-PT|CG1915-PU|CG1915-PV|CG1915-PW|CG1915-PX|CG1915-PY|CG1915-PZ|D-Titin/Kettin|D-titan|D-titin|D-titin-KZ|dtitin|kettin|lethal (3) S002001|lethal(3)62Ca|mitotic chromosomal protein|salimus|sls-PA|sls-PD|sls-PP|sls-PQ|sls-PR|sls-PS|sls-PT|sls-PU|sls-PV|sls-PW|sls-PX|sls-PY|sls-PZ|titin', map_location='62C2-62C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2039681, end_position_on_the_genomic_accession=2115611, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  44014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44014, CurrentID=0, Status='live', Symbol='sol', Aliases='Dmel_CG1391, CG1391, CalpD, DmSol, Dmel\\CG1391, PC79, SOL, Sol, nofE, w5000', description='small optic lobes', other_designations='CG1391-PA|CG1391-PB|CG1391-PE|CG1391-PF|small optic lobe|sol-PA|sol-PB|sol-PE|sol-PF', map_location='19F5-19F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21350280, end_position_on_the_genomic_accession=21357912, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19835887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835887, CurrentID=0, Status='live', Symbol='CG44362', Aliases='Dmel_ Dmel\\CG44362', description='CG44362 gene product from transcript CG44362-RA', other_designations='CG44362-PA|uncharacterized protein', map_location='68C6-68C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11397542, end_position_on_the_genomic_accession=11397865, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835889, CurrentID=0, Status='live', Symbol='CG45546', Aliases='Dmel_ Dmel\\CG45546', description='CG45546 gene product from transcript CG45546-RA', other_designations='CG45546-PA|uncharacterized protein', map_location='99A7-99A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29369338, end_position_on_the_genomic_accession=29369777, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44018, CurrentID=0, Status='live', Symbol='cas', Aliases='Dmel_CG2102, CG2102, Cas, Dmel\\CG2102, l(3)j1C2, l(3)neo33, ming', description='castor', other_designations='CG2102-PA|CG2102-PC|cas-PA|cas-PC', map_location='83C1-83C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5712975, end_position_on_the_genomic_accession=5716805, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44021, CurrentID=0, Status='live', Symbol='nmd', Aliases='Dmel_CG5395, CG5395, DmCG5395, Dmel\\CG5395, Msp, Msp1, Nmd, anon-EST:Liang-1.79, clone 1.79, l(2)08774, ms(2)4, ms(2)ry4', description='no mitochondrial derivative', other_designations='CG5395-PA|CG5395-PB|CG5395-PC|mitochondrial sorting protein 1|nmd-PA|nmd-PB|nmd-PC', map_location='31D1-31D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10302173, end_position_on_the_genomic_accession=10303947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318633, CurrentID=0, Status='live', Symbol='CG31226', Aliases='Dmel_ BcDNA:AT26083, Dmel\\CG31226, NEST:bs14h12', description='CG31226 gene product from transcript CG31226-RB', other_designations='CG31226-PA|CG31226-PB|uncharacterized protein', map_location='91D2-91D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18851305, end_position_on_the_genomic_accession=18851798, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44027, CurrentID=0, Status='live', Symbol='sbb', Aliases='Dmel_CG5580, Bks, Brakeless, CG5580, Dmel\\CG5580, GM04742, Group IIa, Sbb, bks, disc6, l(2)04440, mtv/mtv, sbb', description='scribbler', other_designations='CG5580-PC|CG5580-PD|CG5580-PE|CG5580-PF|CG5580-PG|CG5580-PH|CG5580-PJ|brakeless|master of thickveins|mother of thick-vein|sbb-PC|sbb-PD|sbb-PE|sbb-PF|sbb-PG|sbb-PH|sbb-PJ', map_location='55C2-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18278198, end_position_on_the_genomic_accession=18357296, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19835900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835900, CurrentID=0, Status='live', Symbol='CG45490', Aliases='Dmel_ Dmel\\CG45490', description='CG45490 gene product from transcript CG45490-RA', other_designations='CG45490-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19933541, end_position_on_the_genomic_accession=19934294, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44029, CurrentID=0, Status='live', Symbol='Ref1', Aliases='Dmel_CG1101, ALY, Aly, BcDNA:LD24793, CG1101, DmREF1, Dmel\\CG1101, REF1, aly, cg1101, dmREF1, l(3)02267, ref1', description='RNA and export factor binding protein 1', other_designations='Aly|CG1101-PA|Ref1-PA', map_location='84B2-84B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7077904, end_position_on_the_genomic_accession=7080387, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44030, CurrentID=0, Status='live', Symbol='msn', Aliases='Dmel_CG16973, 6286, CG16973, DMSN, Dmel\\CG16973, HGK, JNKKKK, MESR5, MSN, Msn, NIK, Nik, l(3)03349, l(3)06946, l(3)6286, l(3)j1E2, msm', description='misshapen', other_designations='CG16973-PA|CG16973-PB|CG16973-PC|CG16973-PD|CG16973-PF|CG16973-PG|CG16973-PH|CG16973-PI|NIK|mishappen|msn-PA|msn-PB|msn-PC|msn-PD|msn-PF|msn-PG|msn-PH|msn-PI', map_location='62E6-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2554847, end_position_on_the_genomic_accession=2586540, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19835907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835907, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-L', Aliases='Dmel_CG44880, CG44880, Dmel\\CG44880', description='CG44880 gene product from transcript CG44880-RA', other_designations='CG44880-PA|CG44880-PB|pre-mod(mdg4)-L-PA|pre-mod(mdg4)-L-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21355692, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44038, CurrentID=0, Status='live', Symbol='Hf', Aliases='Dmel_CG10658, CG10658, DHF, Dmel\\CG10658, GH21, Os9, dhf', description='Helical Factor', other_designations='CG10658-PA|Hf-PA|Olfactory-specific 9|drosophila helical factor|helical factor', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19966727, end_position_on_the_genomic_accession=19967685, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44039, CurrentID=0, Status='live', Symbol='Pak', Aliases='Dmel_CG10295, CG10295, D-Pak, DPAK, DPak, DPak1, DmPAK1, Dmel\\CG10295, Dpak, Dpak1, PAK, PAK1, PAK2, PaK1, dPAK, dPAK1, dPak, dpak, dpak1, l(3)psg4, p65[PAK], pak, pak1, psg4, Pak', description='p21-activated kinase', other_designations='CG10295-PA|CG10295-PB|CG10295-PC|CG10295-PE|CG10295-PF|CG10295-PG|CG10295-PH|CG10295-PI|CG10295-PJ|P21-activated kinase|P21-activating kinase|PAK kinase|PAK-kinase|Pak-PA|Pak-PB|Pak-PC|Pak-PE|Pak-PF|Pak-PG|Pak-PH|Pak-PI|Pak-PJ|Pak1 kinase|lethal (3) persistent salivary gland 4|p-21 activated kinase|p21 activated kinase|p21- activated kinase|p21-activated serine/threonine kinase|p21-associated kinase', map_location='83E4-83E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6350896, end_position_on_the_genomic_accession=6361614, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  44046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44046, CurrentID=0, Status='live', Symbol='sun', Aliases='Dmel_CG9032, ATP5E, ATPE, ATPsynepsilon, BcDNA:RE19513, BcDNA:RH48911, CG9032, Dmel\\CG9032, Sun, l(1)13Fe, l(1)XV', description='stunted', other_designations='ATP synthase epsilon chain|ATP synthase, epsilon subunit|CG9032-PA|CG9032-PB|F0/F1 ATP synthase epsilon|sun-PA|sun-PB', map_location='13F12-13F12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15847859, end_position_on_the_genomic_accession=15849258, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44052, CurrentID=0, Status='live', Symbol='Shc', Aliases='Dmel_CG3715, CG3715, D-Shc, DSHC, Dmel\\CG3715, dShc, dshc, shc', description='SHC-adaptor protein', other_designations='CG3715-PA|Shc-PA', map_location='67B5-67B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9427010, end_position_on_the_genomic_accession=9428671, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44054, CurrentID=0, Status='live', Symbol='Ubc7', Aliases='Dmel_CG4443, CG4443, Crl, Dmel\\CG4443, UBC7, Ubc47D, col, crl', description='Ubiquitin conjugating enzyme 7', other_designations='CG4443-PA|CG4443-PB|Ubc7-PA|Ubc7-PB|Ubiquitin-conjugating-enzyme-47D|courtless|ubiquitin conjugating enzyme 7', map_location='47D-47D', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16593677, end_position_on_the_genomic_accession=16595867, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44058, CurrentID=0, Status='live', Symbol='Dot', Aliases='Dmel_CG2788, CG2788, Dmel\\CG2788', description='Dorothy', other_designations='CG2788-PA|CG2788-PB|Dot-PA|Dot-PB|dorothy', map_location='24A1-24A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3619097, end_position_on_the_genomic_accession=3621573, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44059, CurrentID=0, Status='live', Symbol='RpL30', Aliases='Dmel_CG10652, BcDNA:RE25263, BcDNA:RH09938, CG10652, Dmel\\CG10652, L30, RPL30, Rp L30, Rpl30, anon-EST:fe1D4, l(2)01265, l(2)SH0229, l(2)SH2 0229, l(2)k09918, plume', description='Ribosomal protein L30', other_designations='CG10652-PB|CG10652-PC|CG10652-PE|L30 ribosomal gene|RpL30-PB|RpL30-PC|RpL30-PE|anon-fast-evolving-1D4|lethal (2) SH0229|plume|ribosomal protein L30', map_location='37B9-37B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19008031, end_position_on_the_genomic_accession=19009306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44071, CurrentID=0, Status='live', Symbol='Mcr', Aliases='Dmel_CG7586, CG7586, DmTEP6, Dmel\\CG7586, E(br)155, TEP6, TEP6mel, Tep6, TepVI, alpha-2-Mrp, anon-28DE, dTEPVI, luf, mcr', description='Macroglobulin complement-related', other_designations='CG7586-PA|E(br)155|Mcr-PA|TEP|Tep6|alpha-2-macroglobulin related|enhancer of broad|enhancer of broad 155|luftschlange|macroglobulin complement related|macroglobulin complement-related|thiolester containing protein VI', map_location='28E1-28E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8074117, end_position_on_the_genomic_accession=8082982, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  44072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44072, CurrentID=0, Status='live', Symbol='Ran', Aliases='Dmel_CG1404, AAF30287, CG1404, Dmel\\CG1404, Q9VZ23GTP, dran, l(1)G0075, ran, ran10A', description='CG1404 gene product from transcript CG1404-RC', other_designations='CG1404-PA|CG1404-PB|CG1404-PC|CG1404-PD|Ran GTPase|Ran-PA|Ran-PB|Ran-PC|Ran-PD|ran', map_location='10A8-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11144852, end_position_on_the_genomic_accession=11148716, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44079, CurrentID=0, Status='live', Symbol='Ag5r2', Aliases='Dmel_CG9540, Agr2, CG9540, Dmel\\CG9540', description='Antigen 5-related 2', other_designations='Ag5r2-PA|CG9540-PA|antigen 5-related 2', map_location='12F4-12F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14782962, end_position_on_the_genomic_accession=14784052, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835957, CurrentID=0, Status='live', Symbol='CG43920', Aliases='Dmel_ CR43920, Dmel\\CG43920', description='CG43920 gene product from transcript CG43920-RB', other_designations='CG43920-PB|CG43920-PC|uncharacterized protein', map_location='54C12-54C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17598845, end_position_on_the_genomic_accession=17599253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44086, CurrentID=0, Status='live', Symbol='DCTN2-p50', Aliases='Dmel_CG8269, CG8269, DMN, Dmel\\CG8269, Dmn, Dynamitin, P50, dDyn, dmn, dyna, l(2)k16109, l(2)k16218, p50, p50/Dmn, p50/dynamitin, p50[Dmn]', description='Dynactin 2, p50 subunit', other_designations='CG8269-PA|DCTN2-p50-PA|dynactin 2, p50 subunit|dynamitin|p50 dynamitin|p50-dynamitin|p50/Dynamitin', map_location='44F3-44F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8891832, end_position_on_the_genomic_accession=8893339, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44087, CurrentID=0, Status='live', Symbol='Rs1', Aliases='Dmel_CG2173, 153844_at, BcDNA:GM05306, CG2173, DmRH21, Dmel\\CG2173, Drs1, cg2173, l(2)k09514', description='CG2173 gene product from transcript CG2173-RB', other_designations='CG2173-PA|CG2173-PB|Rs1-PA|Rs1-PB', map_location='44B7-44B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8163564, end_position_on_the_genomic_accession=8167769, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  44095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44095, CurrentID=0, Status='live', Symbol='fus', Aliases='Dmel_CG8205, CG8205, Dmel\\CG8205, E(cact[E10])7, cg8205', description='fusilli', other_designations='CG8205-PA|CG8205-PB|CG8205-PC|CG8205-PD|CG8205-PE|CG8205-PF|CG8205-PG|CG8205-PH|enhancer of cactus[E10], number 7|fus-PA|fus-PB|fus-PC|fus-PD|fus-PE|fus-PF|fus-PG|fus-PH', map_location='52B3-52B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15657318, end_position_on_the_genomic_accession=15676669, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  44096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44096, CurrentID=0, Status='live', Symbol='wit', Aliases='Dmel_CG10776, 1262/15, ALK3/BMPRII, BMP, CG10776, Dmel\\CG10776, SE20, STK-D, Stk-D, WIT, Wit, Wit-C, l(3)64Aa, l(3)S126215, l(3)SH12', description='wishful thinking', other_designations='CG10776-PA|CG10776-PB|Serine/threonine kinase-D|bone morphogenetic protein|wit-PA|wit-PB', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4064152, end_position_on_the_genomic_accession=4071717, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  44097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44097, CurrentID=0, Status='live', Symbol='sns', Aliases='Dmel_CG33141, 43-49, CG12495, CG13752, CG13753, CG13754, CG13755, CG18464, CG2385, CG33141, CG8278, CT33230, CT33232, Dmel\\CG33141, SNS, Sns, rost20, sns', description='sticks and stones', other_designations='CG33141-PA|CG33141-PB|CG33141-PC|sns-PA|sns-PB|sns-PC|stick-and-stones|sticks and stone|sticks-and-stones', map_location='44F1-44F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8798489, end_position_on_the_genomic_accession=8856091, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  44098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44098, CurrentID=0, Status='live', Symbol='rasp', Aliases='Dmel_CG11495, 7F21, CG11495, Dmel\\CG11495, RASP, Rasp, cmn, cnm, l(3)63Bg, l(3)M47, l(3)SG3, l(3)m47, rsp, sig/ski, sit, ski, ski/sit', description='CG11495 gene product from transcript CG11495-RA', other_designations='CG11495-PA|central missing|lethal(3)63Bg|lethal(3)SG3|rasp-PA|sightless|skinny hedgehog', map_location='3-4.3 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3162180, end_position_on_the_genomic_accession=3163852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44100, CurrentID=0, Status='live', Symbol='Patj', Aliases='Dmel_CG12021, BcDNA:LD22238, CG12021, D-Patj, DLT, DPATJ, DPatj, Dlt, Dmel\\CG12021, Dpatj, PATJ, PaTJ, PatJ/Dlt, dPATJ, dPatj, dlt, dlt:discs lost, dpatj, patj', description='CG12021 gene product from transcript CG12021-RC', other_designations='CG12021-PA|CG12021-PB|CG12021-PC|Patj-PA|Patj-PB|Patj-PC|discs lost|discslost|patj', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1798803, end_position_on_the_genomic_accession=1802184, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34657, CurrentID=0, Status='live', Symbol='CG5435', Aliases='Dmel_ BcDNA:AT11334, Dmel\\CG5435', description='CG5435 gene product from transcript CG5435-RB', other_designations='CG5435-PA|CG5435-PB', map_location='33E4-33E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12429827, end_position_on_the_genomic_accession=12430892, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44108, CurrentID=0, Status='live', Symbol='Optix', Aliases='Dmel_CG18455, CG18455, D-Six3, Dmel\\CG18455, Dsix3, OPTIX, Six3, anon-WO0153538.79, opt, optix, opx, six3', description='CG18455 gene product from transcript CG18455-RC', other_designations='CG18455-PA|CG18455-PB|CG18455-PC|Optix-PA|Optix-PB|Optix-PC|optix', map_location='44A3-44A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8031283, end_position_on_the_genomic_accession=8042529, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44110, CurrentID=0, Status='live', Symbol='Mgstl', Aliases='Dmel_CG1742, BcDNA:RE10696, BcDNA:RH59553, CG1742, Dmel\\CG1742, Mgst1', description='Microsomal glutathione S-transferase-like', other_designations='CG1742-PA|CG1742-PB|CG1742-PC|Mgstl-PA|Mgstl-PB|Mgstl-PC|microsomal glutathione|microsomal glutathione S-transferase-like|microsomal glutathione transferase 1|transferase', map_location='19E7-19E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21042125, end_position_on_the_genomic_accession=21044107, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44111, CurrentID=0, Status='live', Symbol='slpr', Aliases='Dmel_CG2272, CG2272, D-MLK, Dmel\\CG2272, JNKKK, MLK, MLK2, Mlk2, Slpr, dMLK, dMLK2', description='slipper', other_designations='CG2272-PA|CG2272-PB|CG2272-PC|CG2272-PD|mixed lineage protein kinase 2|slpr-PA|slpr-PB|slpr-PC|slpr-PD', map_location='7D17-7D17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8170229, end_position_on_the_genomic_accession=8177185, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  44115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44115, CurrentID=0, Status='live', Symbol='Rgl', Aliases='Dmel_CG8865, CG8865, DRgl, Dmel\\CG8865, RalGDS, RasGDS, dRGL, rgl', description='Ral guanine nucleotide dissociation stimulator-like ortholog (M. musculus)', other_designations='CG8865-PA|CG8865-PB|CG8865-PC|CG8865-PD|CG8865-PE|Ral guanine nucleotide dissociation stimulator-like ortholog|Ral guanine nucleotide exchange factor|Ral guanine nucleotide exchange factor 2|Rgl-PA|Rgl-PB|Rgl-PC|Rgl-PD|Rgl-PE', map_location='70C5-70C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13906897, end_position_on_the_genomic_accession=13925335, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  44117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44117, CurrentID=0, Status='live', Symbol='bgm', Aliases='Dmel_CG4501, 2a9hom, BG:DS01514.2, BG:DS0514.2, Bgm, CG4501, Dmel\\CG4501', description='bubblegum', other_designations='CG4501-PA|CG4501-PC|bgm-PA|bgm-PC', map_location='34F1-34F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14013004, end_position_on_the_genomic_accession=14019328, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  44118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44118, CurrentID=0, Status='live', Symbol='vih', Aliases='Dmel_CG10682, 1105/01, CG10682, Dmel\\CG10682, anon-WO0172774.54, anon-WO0172774.55, anon-WO0172774.56, l(3)S110501', description='vihar', other_designations='CG10682-PA|lethal (3) S110501|vih-PA', map_location='69C4-69C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12504292, end_position_on_the_genomic_accession=12505821, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44121, CurrentID=0, Status='live', Symbol='TotA', Aliases='Dmel_CG31509, CG31509, DIM 31, Dmel\\CG31509, Tot, Tot A, Tot-A, Turt, tot, totA', description='Turandot A', other_designations='CG31509-PA|Stress-inducible humoral factor turandot A|TotA-PA|turandot|turandot A', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20871036, end_position_on_the_genomic_accession=20871700, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44125, CurrentID=0, Status='live', Symbol='ETH', Aliases='Dmel_CG18105, CG18105, DmETH, Dmel\\CG18105, Drm-ETH-1, Drm-ETH-2-1, ETH-2, ETH1, ETH2, eth, ETH', description='Ecdysis triggering hormone', other_designations='CG18105-PA|ETH-PA|Ecdysis-triggering hormone|ecdysis triggering hormone|ecdysis triggering hormone 1|ecdysis triggering hormone-1|ecdysis triggering hormone-2|ecdysis-triggering hormone-1|ecdysis-triggering hormone-2|ecdysis-triggering-hormone-1', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24794670, end_position_on_the_genomic_accession=24795482, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44126, CurrentID=0, Status='live', Symbol='AstA-R1', Aliases='Dmel_CG2872, ASR, AST-R, AlstR, CG2872, D.AlstR1, DAR-1, DAR1, DGR, Dmel\\CG2872, EG:121E7.2, GR', description='Allatostatin A receptor 1', other_designations='ASA-receptor|AlstR|AstA-R1-PB|AstA-R1-PD|CG2872-PB|CG2872-PD|allatostatin A receptor 1|allatostatin receptor|galanin receptor', map_location='3D6-3E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3574536, end_position_on_the_genomic_accession=3666724, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  44127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44127, CurrentID=0, Status='live', Symbol='mab-21', Aliases='Dmel_CG4746, CG4746, Dmel\\CG4746, mab-2', description='CG4746 gene product from transcript CG4746-RA', other_designations='CG4746-PA|mab-2|mab-21-PA', map_location='5D2-5D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6008183, end_position_on_the_genomic_accession=6013181, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44128, CurrentID=0, Status='live', Symbol='lmgB', Aliases='Dmel_CG34441, CG18042, CG34441, Dmel\\CG34441, Dmel_CG18042, Dromel_CG18042_FBtr0079695_lmg_mORF, ORF2', description='lemming B', other_designations='CG34441-PA|CG34441-PB|CG34441-PC|CG34441-PD|lmgB-PA|lmgB-PB|lmgB-PC|lmgB-PD', map_location='29D4-29D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8524999, end_position_on_the_genomic_accession=8528618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44129, CurrentID=0, Status='live', Symbol='hbs', Aliases='Dmel_CG7449, CG7449, CT22841, Dmel\\CG7449, Hbs, icon', description='hibris', other_designations='CG7449-PA|CG7449-PB|CG7449-PC|hbs-PA|hbs-PB|hbs-PC|inappropriate contact|inappropriate contacts', map_location='51D8-51D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15011396, end_position_on_the_genomic_accession=15042134, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  44131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44131, CurrentID=0, Status='live', Symbol='Mkrn1', Aliases='Dmel_CG7184, CG7184, Dmel\\CG7184, MKRN1', description='Makorin 1', other_designations='CG7184-PA|CG7184-PB|CG7184-PC|Mkrn1-PA|Mkrn1-PB|Mkrn1-PC|makorin 1', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21529959, end_position_on_the_genomic_accession=21533713, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44132, CurrentID=0, Status='live', Symbol='slmo', Aliases='Dmel_CG9131, CG9131, CT26166, Dmel\\CG9131, kisir', description='slowmo', other_designations='CG9131-PA|CG9131-PB|CG9131-PC|CG9131-PD|CG9131-PE|kisir|slmo-PA|slmo-PB|slmo-PC|slmo-PD|slmo-PE', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6046516, end_position_on_the_genomic_accession=6048706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44135, CurrentID=0, Status='live', Symbol='ect', Aliases='Dmel_CG6611, CG11965, CG6611, DROECT, Difl, Dmel\\CG6611, Ifl', description='ectodermal', other_designations='CG6611-PA|CG6611-PB|CG6611-PD|CG6611-PE|CG6611-PF|Interferon-like protein|ect-PA|ect-PB|ect-PD|ect-PE|ect-PF', map_location='67D10-67D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10190276, end_position_on_the_genomic_accession=10201837, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19836008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836008, CurrentID=0, Status='live', Symbol='CG45084', Aliases='Dmel_ Dmel\\CG45084, Dromel_CG4824_FBtr0080804_bicc_uORF', description='CG45084 gene product from transcript CG45084-RB', other_designations='CG45084-PA|CG45084-PB|uncharacterized protein', map_location='35E2-35E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16042047, end_position_on_the_genomic_accession=16048757, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19836032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836032, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-J', Aliases='Dmel_CG44882, CG44882, Dmel\\CG44882', description='CG44882 gene product from transcript CG44882-RA', other_designations='CG44882-PA|CG44882-PB|pre-mod(mdg4)-J-PA|pre-mod(mdg4)-J-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21355692, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3355155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355155, CurrentID=0, Status='live', Symbol='ND-AGGG', Aliases='Dmel_CG40002, CG40002, Dmel\\CG40002', description='NADH dehydrogenase (ubiquinone) AGGG subunit', other_designations='CG40002-PA|CG40002-PB|NADH ubiquinone oxidoreductase AGGG|NADH:ubiquinone oxidoreductase AGGG subunit|ND-AGGG-PA|ND-AGGG-PB', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24976254, end_position_on_the_genomic_accession=24977118, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44148, CurrentID=0, Status='live', Symbol='PpD5', Aliases='Dmel_CG10138, CG10138, D5, Dmel\\CG10138, anon-WO0140519.107', description='Protein phosphatase D5', other_designations='CG10138-PA|PP1-like|PpD5-PA|protein phosphatase D5|protein phosphatase from PCR fragment D5', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22041667, end_position_on_the_genomic_accession=22042945, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44149, CurrentID=0, Status='live', Symbol='Acon', Aliases='Dmel_CG9244, ACON-1M, BEST:GH10550, CG9244, Dmel\\CG9244, FBgn0010100, M-Acon, acon, hsp60, i204, i59, l(2)07054, m-Acon, m-acon, mACON, mAc, mAcon, Acon', description='Aconitase', other_designations='ACONITASE|Acon-PB|Acon-PD|Acon-PE|Acon-PF|CG9244-PB|CG9244-PD|CG9244-PE|CG9244-PF|aconitase|m-aconitase|mitochondrial aconitase', map_location='39A7-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21168570, end_position_on_the_genomic_accession=21173185, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44150, CurrentID=0, Status='live', Symbol='RpS15Aa', Aliases='Dmel_CG2033, CG2033, Dmel\\CG2033, Rp S15 Aa, RpS15A, RpS15a, S15, S15Aa, anon-EST:Posey44', description='Ribosomal protein S15Aa', other_designations='CG2033-PA|CG2033-PC|CG2033-PD|CG2033-PE|CG2033-PF|RpS15Aa-PA|RpS15Aa-PC|RpS15Aa-PD|RpS15Aa-PE|RpS15Aa-PF|ribosomal protein S15A|ribosomal protein S15Aa', map_location='11E11-11E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13335637, end_position_on_the_genomic_accession=13337705, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44154, CurrentID=0, Status='live', Symbol='Hmgs', Aliases='Dmel_CG4311, CG16796, CG4311, Dmel\\CG4311, HMGS, Ov25, l(2)06214', description='HMG Coenzyme A synthase', other_designations='CG4311-PA|CG4311-PB|CG4311-PC|CG4311-PD|CG4311-PE|CG4311-PF|HMG coenzyme A synthase|HMG-CoA synthase|Hmgs-PA|Hmgs-PB|Hmgs-PC|Hmgs-PD|Hmgs-PE|Hmgs-PF|hydroxy-3-methylglutaryl-CoA synthase', map_location='53C1-53C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16336090, end_position_on_the_genomic_accession=16341242, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44155, CurrentID=0, Status='live', Symbol='CG10939', Aliases='Dmel_ Dmel\\CG10939, Sip1, l(2)06373', description='CG10939 gene product from transcript CG10939-RA', other_designations='CG10939-PA|SRY interacting protein 1|uncharacterized protein', map_location='54B6-54B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17394593, end_position_on_the_genomic_accession=17405171, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19836030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836030, CurrentID=0, Status='live', Symbol='CG45489', Aliases='Dmel_ Dmel\\CG45489', description='CG45489 gene product from transcript CG45489-RA', other_designations='CG45489-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19928416, end_position_on_the_genomic_accession=19929178, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44160, CurrentID=0, Status='live', Symbol='sw', Aliases='Dmel_CG18000, 17-234, CG18000, Cdic, Cdic2b, DIC, Dic, Dic19B, Dic19C, Dmel\\CG18000, IC, IC74, cDIC, dic, l(1)19Bb', description='short wing', other_designations='CG18000-PA|CG18000-PB|CG18000-PC|CG18000-PD|CG18000-PE|CG18000-PF|CG18000-PG|CG18000-PH|CG18000-PI|CG18000-PJ|CG18000-PK|CG18000-PL|CG18000-PM|CG18000-PN|CG18000-PO|cytoplasmic dynein|cytoplasmic dynein intermediate chain|cytoplasmic dynein intermediate polypeptide chain|dynein|dynein IC|dynein intermediate chain|dynein intermediate-chain|sw-PA|sw-PB|sw-PC|sw-PD|sw-PE|sw-PF|sw-PG|sw-PH|sw-PI|sw-PJ|sw-PK|sw-PL|sw-PM|sw-PN|sw-PO|swallow', map_location='1-64.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20227961, end_position_on_the_genomic_accession=20236382, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19836034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836034, CurrentID=0, Status='live', Symbol='CG44250', Aliases='Dmel_ CG3884, Dmel\\CG44250', description='CG44250 gene product from transcript CG44250-RB', other_designations='CG44250-PB|CG44250-PC|uncharacterized protein', map_location='49E4-49E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12950404, end_position_on_the_genomic_accession=12952293, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44172, CurrentID=0, Status='live', Symbol='RabX1', Aliases='Dmel_CG3870, AAF47018, CG3870, Chrw, Dm Dmel\\CG3870, RAB5c, chr, chrw, l(2)k06908, rabX1', description='CG3870 gene product from transcript CG3870-RA', other_designations='CG3870-PA|RabX1-PA|chrowded', map_location='59E2-59E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23519839, end_position_on_the_genomic_accession=23523613, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44173, CurrentID=0, Status='live', Symbol='hoip', Aliases='Dmel_CG3949, CG3949, Dmel\\CG3949, HOIP, Hoip', description='hoi-polloi', other_designations='CG3949-PA|CG3949-PB|Hoi polloi|hoip-PA|hoip-PB', map_location='30C5-30C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9699791, end_position_on_the_genomic_accession=9700459, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19836050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836050, CurrentID=0, Status='live', Symbol='CG44623', Aliases='Dmel_ Dmel\\CG44623, lincRNA.354', description='CG44623 gene product from transcript CG44623-RA', other_designations='CG44623-PA|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19956195, end_position_on_the_genomic_accession=19957004, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44183, CurrentID=0, Status='live', Symbol='ScpX', Aliases='Dmel_CG17320, CG17320, Dmel\\CG17320', description='Sterol carrier protein X-related thiolase', other_designations='CG17320-PA|ScpX-PA|sterol carrier protein X-related thiolase|sterol carrier protein x', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18831229, end_position_on_the_genomic_accession=18833630, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19836061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836061, CurrentID=0, Status='live', Symbol='CG45078', Aliases='Dmel_ CG6544, Dmel\\CG45078', description='CG45078 gene product from transcript CG45078-RA', other_designations='CG45078-PA|uncharacterized protein', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10773339, end_position_on_the_genomic_accession=10777374, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318663, CurrentID=0, Status='live', Symbol='CG31287', Aliases='Dmel_ Dmel\\CG31287, NEST:bs24f08', description='CG31287 gene product from transcript CG31287-RB', other_designations='CG31287-PA|CG31287-PB|uncharacterized protein', map_location='89B7-89B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16203371, end_position_on_the_genomic_accession=16204961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44205, CurrentID=0, Status='live', Symbol='spen', Aliases='Dmel_CG18497, 136/24, BcDNA:GM01870, CG18497, Dmel\\CG18497, E(E2F)2A, E(Raf)2A, E(Sev-CycE)2A, EK2-9, EY2-7, En(yan[ACT])2-7, SPEN, Spen, Splitends, l(2)03350, l(2)k06805, l(2)k07612, l(2)k08102, poc, rno, spe, yip1', description='split ends', other_designations='CG18497-PA|CG18497-PB|CG18497-PC|CG18497-PD|CG18497-PE|CG18497-PF|CG18497-PG|CG18497-PH|polycephalon|rhinoceros|spen-PA|spen-PB|spen-PC|spen-PD|spen-PE|spen-PF|spen-PG|spen-PH|split-ends|splitends|yippee interacting protein 1', map_location='21B3-21B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=159032, end_position_on_the_genomic_accession=203397, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  44207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44207, CurrentID=0, Status='live', Symbol='ND-23', Aliases='Dmel_CG3944, CG3944, Dmel\\CG3944, ND 23, ND23, NUIM', description='NADH dehydrogenase (ubiquinone) 23 kDa subunit', other_designations='CG3944-PA|COMPLEX I: 23 kDa subunit|NADH ubiquinone oxidoreductase 23 kDa|NADH-UBIQUINONE OXIDOREDUCTASE 23 KDA SUBUNIT|NADH: UBIQUINONE REDUCTASE 23 KD SUBUNIT PRECURSOR|NADH:ubiquinone oxidoreductase 23 kDa subunit|NADH:ubiquinone reductase 23kD subunit precursor|ND-23-PA', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15836048, end_position_on_the_genomic_accession=15837117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740474, CurrentID=0, Status='live', Symbol='CG34171', Aliases='Dmel_ Dmel\\CG34171', description='CG34171 gene product from transcript CG34171-RC', other_designations='CG34171-PB|CG34171-PC|uncharacterized protein', map_location='36F3-36F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18378591, end_position_on_the_genomic_accession=18379909, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19836082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836082, CurrentID=0, Status='live', Symbol='CG44569', Aliases='Dmel_ Dmel\\CG44569', description='CG44569 gene product from transcript CG44569-RA', other_designations='CG44569-PA|uncharacterized protein', map_location='56E4-56E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19842495, end_position_on_the_genomic_accession=19843668, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19836089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836089, CurrentID=0, Status='live', Symbol='CG44329', Aliases='Dmel_ Dmel\\CG44329, lincRNA.951', description='CG44329 gene product from transcript CG44329-RA', other_designations='CG44329-PA|uncharacterized protein', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5812273, end_position_on_the_genomic_accession=5812977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44219, CurrentID=0, Status='live', Symbol='Sr-CII', Aliases='Dmel_CG8856, CG8856, CT25430, Dmel\\CG8856, SR-CII, dSR-CII', description='Scavenger receptor class C, type II', other_designations='CG8856-PA|CG8856-PB|Sr-CII-PA|Sr-CII-PB|scavenger receptor|scavenger receptor class C, type II', map_location='48E10-48E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12208893, end_position_on_the_genomic_accession=12212155, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44226, CurrentID=0, Status='live', Symbol='Xbp1', Aliases='Dmel_CG9415, 138/3, CG 9415, CG9415, DM18, Dmel\\CG9415, Dxbp-1, XBP-1, XBP1, Xbp-1, XbpI, anon-EST:Liang-1.41, clone 1.41, dXbp-1, l(2)k13803, xbp-1, xbp1', description='X box binding protein-1', other_designations='CG9415-PD|CG9415-PE|X box binding protein 1|X-box binding protein|X-box-binding protein 1|Xbp1-PD|Xbp1-PE', map_location='57C3-57C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21143545, end_position_on_the_genomic_accession=21145750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44228, CurrentID=0, Status='live', Symbol='ND-PDSW', Aliases='Dmel_CG8844, BcDNA:RE65754, CG8844, Dmel\\CG8844, NIDM, PDSW, Pdsw, anon-EST:Posey91, l(2)k10101, pdsw', description='NADH dehydrogenase (ubiquinone) PDSW subunit', other_designations='CG8844-PDSW-PA|CG8844-PDSW-PB|NADH ubiquinone oxidoreductase PDSW|NADH-ubiquinone oxidoreductase PDSW subunit|NADH:ubiquinone oxidoreductase PDSW subunit|ND-PDSW-PA|ND-PDSW-PB|pdsw', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3466157, end_position_on_the_genomic_accession=3467334, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19836101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836101, CurrentID=0, Status='live', Symbol='CG45690', Aliases='Dmel_ Dmel\\CG45690', description='CG45690 gene product from transcript CG45690-RA', other_designations='CG45690-PA|uncharacterized protein', map_location='32A2-32A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10658281, end_position_on_the_genomic_accession=10658559, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44233, CurrentID=0, Status='live', Symbol='fray', Aliases='Dmel_CG7693, 4624, CG7693, Dmel\\CG7693, Dosr1, Fray, STE20-like/SPAK, dOSR1, dSTRAD, l(3)07551, lag', description='frayed', other_designations='CG7693-PA|CG7693-PB|CG7693-PC|CG7693-PE|CG7693-PF|CG7693-PG|CG7693-PH|CG7693-PI|Ste20-like|Ste20-like protein kinase|fray-PA|fray-PB|fray-PC|fray-PE|fray-PF|fray-PG|fray-PH|fray-PI|laggard|oxidative-stress-responsive kinase 1', map_location='91B4-91B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18573091, end_position_on_the_genomic_accession=18589459, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  44235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44235, CurrentID=0, Status='live', Symbol='bon', Aliases='Dmel_CG5206, 0241/08, 0249/12, 0434/20, 0487/06, 150268_at, 150269_at, Bon, Bonus, CG15687, CG5206, Dmel\\CG5206, anon-WO0118547.418, l(3)S043420', description='bonus', other_designations='CG5206-PA|CG5206-PB|CG5206-PC|CG5206-PD|CG5206-PE|CG5206-PF|bon-PA|bon-PB|bon-PC|bon-PD|bon-PE|bon-PF', map_location='92F2-92F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20593139, end_position_on_the_genomic_accession=20612540, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  44238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44238, CurrentID=0, Status='live', Symbol='sip1', Aliases='Dmel_CG7238, BcDNA:GH11690, CG7238, Dmel\\CG7238, GH11690', description='septin interacting protein 1', other_designations='CG7238-PA|CG7238-PB|sip1-PA|sip1-PB', map_location='25F3-25F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5721051, end_position_on_the_genomic_accession=5723944, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19836116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836116, CurrentID=0, Status='live', Symbol='CG45064', Aliases='Dmel_ CG9517, Dmel\\CG45064', description='CG45064 gene product from transcript CG45064-RA', other_designations='CG45064-PA|uncharacterized protein', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14875370, end_position_on_the_genomic_accession=14882586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19836117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836117, CurrentID=0, Status='live', Symbol='CG44428', Aliases='Dmel_ Dmel\\CG44428', description='CG44428 gene product from transcript CG44428-RA', other_designations='CG44428-PA|uncharacterized protein', map_location='12A2-12A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13529656, end_position_on_the_genomic_accession=13530249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44250, CurrentID=0, Status='live', Symbol='Cpsf160', Aliases='Dmel_CG10110, CG10110, CPSF, CPSF-160, CPSF160, Cpsf, Dmel\\CG10110, anon-WO0118547.217, cpsf, dCPSF-160', description='Cleavage and polyadenylation specificity factor 160', other_designations='CG10110-PA|CG10110-PB|Cpsf160-PA|Cpsf160-PB|cleavage and polyadenylation specificity factor|cleavage and polyadenylation specificity factor 160|drosophila cleavage and polyadenylation stimulation factor-160', map_location='51A4-51A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14497347, end_position_on_the_genomic_accession=14503488, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44251, CurrentID=0, Status='live', Symbol='RpL8', Aliases='Dmel_CG1263, CG1263, Dmel\\CG1263, L8, M(3)62F, M(3)LS2, RPL8, Rp L8, anon-EST:Posey5, rpL8', description='Ribosomal protein L8', other_designations='CG1263-PA|CG1263-PB|Minute(3)62F|RpL8-PA|RpL8-PB|ribosomal protein L8', map_location='62E7-62E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2587456, end_position_on_the_genomic_accession=2588883, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44254, CurrentID=0, Status='live', Symbol='bab2', Aliases='Dmel_CG9102, BAB, BAB2, BTB-II, Bab2, BtbII, CG13911, CG9102, Dmel\\CG9102, bab, bab-II', description='bric a brac 2', other_designations='BTB-protein-II|Bric-a-brac 2|CG9102-PA|CG9102-PB|bab2-PA|bab2-PB|bric a brac|bric a brac II|bric-a-brac|bric-a-brac-2|bric-a-brac2', map_location='61F2-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1140429, end_position_on_the_genomic_accession=1177276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44258, CurrentID=0, Status='live', Symbol='ps', Aliases='Dmel_CG42670, BcDNA:RE47781, CG16765, CG16776, CG16777, CG18509, CG42670, CG8144, Dmel\\CG42670, Dmel_CG16765, Dmel_CG16777, EP(3)3406, NOVA1, PS, cg8144, l(3)10615', description='pasilla', other_designations='CG42670-PB|CG42670-PC|CG42670-PD|CG42670-PF|CG42670-PG|CG42670-PH|CG42670-PI|CG42670-PJ|CG42670-PK|CG42670-PL|CG42670-PM|CG42670-PN|CG42670-PO|CG42670-PP|CG42670-PQ|CG42670-PR|CG42670-PS|CG42670-PT|lethal (3) 10615|ps-PB|ps-PC|ps-PD|ps-PF|ps-PG|ps-PH|ps-PI|ps-PJ|ps-PK|ps-PL|ps-PM|ps-PN|ps-PO|ps-PP|ps-PQ|ps-PR|ps-PS|ps-PT', map_location='85D16-85D17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9417940, end_position_on_the_genomic_accession=9455500, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  44259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44259, CurrentID=0, Status='live', Symbol='nbs', Aliases='Dmel_CG6754, 6754, CG6754, Dmel\\CG6754, D1, NBS, NBS1, Nbs, Nbs1, l(3)67BDp, l(3)67BDr, l(3)e77A1', description='CG6754 gene product from transcript CG6754-RD', other_designations='CG6754-PD|CG6754-PE|lethal (3) 67BDp|lethal (3) 67BDr|nbs-PD|nbs-PE|nijmegen breakage syndrome', map_location='67B1-67C5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9720414, end_position_on_the_genomic_accession=9723194, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44260, CurrentID=0, Status='live', Symbol='eIF-1A', Aliases='Dmel_CG8053, CG8053, D-eIF1A, D-eIF4c, Dmel\\CG8053, EP(3)0935, EP935, Group I, anon-EST:Liang-2.3, anon-EST:Posey4, clone 2.3, eIF-4c, eIF1A, group I, lincRNA.S7087', description='Eukaryotic initiation factor 1A', other_designations='CG8053-PA|CG8053-PB|CG8053-PC|eIF-1A-PA|eIF-1A-PB|eIF-1A-PC|eukaryotic initiation factor 1A|eukaryotic initiation factor 4c', map_location='91A1-91A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18274665, end_position_on_the_genomic_accession=18283005, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44261, CurrentID=0, Status='live', Symbol='eIF5B', Aliases='Dmel_CG10840, A1, CG10840, CG14970, Dmel\\CG10840, Eif5b, IF2, anon-WO0118547.362, cIF2, dIF2, yIF2', description='CG10840 gene product from transcript CG10840-RE', other_designations='CG10840-PB|CG10840-PC|CG10840-PD|CG10840-PE|CG10840-PF|cytoplasmic initiation factor 2|eIF5B-PB|eIF5B-PC|eIF5B-PD|eIF5B-PE|eIF5B-PF|eukaryotic initiation factor 5B', map_location='63E1-63E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3428658, end_position_on_the_genomic_accession=3460977, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  44263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44263, CurrentID=0, Status='live', Symbol='Sara', Aliases='Dmel_CG15667, CG15667, Dmel\\CG15667, SARA, dSARA, sara', description='Smad anchor for receptor activation', other_designations='CG15667-PA|Sara-PA|smad anchor for receptor activation', map_location='57E6-57E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21481927, end_position_on_the_genomic_accession=21487107, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  44266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44266, CurrentID=0, Status='live', Symbol='Inx3', Aliases='Dmel_CG1448, CG1448, Dm-inx3, Dmel\\CG1448, FBpp0084719, INX-3, Ix3, anon-EST:fe2B11, inx-3, inx3', description='Innexin 3', other_designations='CG1448-PA|CG1448-PB|CG1448-PC|Innexin-3|Inx3-PA|Inx3-PB|Inx3-PC|anon-fast-evolving-2B11|innexin|innexin 3|innexin3', map_location='98E6-98F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28853606, end_position_on_the_genomic_accession=28859497, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19836139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836139, CurrentID=0, Status='live', Symbol='CG45494', Aliases='Dmel_ Dmel\\CG45494', description='CG45494 gene product from transcript CG45494-RA', other_designations='CG45494-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19952814, end_position_on_the_genomic_accession=19953608, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44268, CurrentID=0, Status='live', Symbol='sut3', Aliases='Dmel_CG17976, CG17976, Dmel\\CG17976', description='sugar transporter 3', other_designations='CG17976-PA|CG17976-PB|sut3-PA|sut3-PB', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8089482, end_position_on_the_genomic_accession=8091297, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44269, CurrentID=0, Status='live', Symbol='sut2', Aliases='Dmel_CG17975, CG17975, Dmel\\CG17975', description='sugar transporter 2', other_designations='CG17975-PA|sut2-PA', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8087396, end_position_on_the_genomic_accession=8089297, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44271, CurrentID=0, Status='live', Symbol='xmas-2', Aliases='Dmel_CG32562, CG32562, CG8919, Dmel\\CG32562, Xmas-2', description='CG32562 gene product from transcript CG32562-RA', other_designations='CG32562-PA|X-linked male sterile 2|xmas-2-PA', map_location='15E7-15F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17154110, end_position_on_the_genomic_accession=17159306, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44272, CurrentID=0, Status='live', Symbol='tio', Aliases='Dmel_CG12630, CG12630, Dmel\\CG12630, Tio', description='tiptop', other_designations='CG12630-PA|CG12630-PB|tio-PA|tio-PB', map_location='40D3-40D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22019208, end_position_on_the_genomic_accession=22025734, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44274, CurrentID=0, Status='live', Symbol='disp', Aliases='Dmel_CG2019, CG2019, Disp, Dmel\\CG2019, anon-WO0146227.1', description='dispatched', other_designations='CG2019-PA|disp-PA', map_location='83C5-83C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5849697, end_position_on_the_genomic_accession=5856311, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  44275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44275, CurrentID=0, Status='live', Symbol='SoxN', Aliases='Dmel_CG18024, CG18024, DM64, Dmel\\CG18024, SOX29F, SOXB1, Sox B1, Sry-rDM64, soxN', description='SoxNeuro', other_designations='CG18024-PA|CG18024-PB|Sox-neuro|SoxN-PA|SoxN-PB|Sry-related DM64|soxneuro', map_location='29F2-29F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8825645, end_position_on_the_genomic_accession=8829670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44276, CurrentID=0, Status='live', Symbol='Crtp', Aliases='Dmel_CG13580, CG13580, Dmel\\CG13580', description='Caldesmon-related protein', other_designations='CG13580-PA|Crtp-PA|caldesmon-related testes protein', map_location='60D1-60D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24494518, end_position_on_the_genomic_accession=24497024, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44279, CurrentID=0, Status='live', Symbol='arr', Aliases='Dmel_CG5912, Arr, BEST:CK00539, CG5912, CK00539, CT15575, Dm Arr, Dmel\\CG5912, Irp6, LRP, LRP/Arr, LRP5/6, LRP6, l(2)k08131', description='arrow', other_designations='CG5912-PA|Dm arrow|LDL-related protein 6|LDLR-like|LRP5/6/Arrow|arr-PA', map_location='2-66 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13453640, end_position_on_the_genomic_accession=13483229, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  317902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317902, CurrentID=0, Status='live', Symbol='CG32189', Aliases='Dmel_ BcDNA:RE39957, Dmel\\CG32189', description='CG32189 gene product from transcript CG32189-RA', other_designations='CG32189-PA|uncharacterized protein', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17834745, end_position_on_the_genomic_accession=17835174, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44289, CurrentID=0, Status='live', Symbol='flw', Aliases='Dmel_CG2096, CG15305, CG2096, DmPp1-9C, Dmel\\CG2096, FLW, FLW/PP1B, PP-1, PP1, PP1 9C, PP1 b9C, PP1 beta9c, PP19C, PP1B_DROME, PP1beta, PP1beta-9C, PP1beta9C, PP1c, PP1cbeta, Pp1-9C, Pp1beta-9C, Pp1beta9C, XE55, anon-WO03040301.120, l(1)G0172', description='flapwing', other_designations='CG2096-PA|CG2096-PB|CG2096-PC|PP1beta9C|flap wing|flw-PA|flw-PB|flw-PC|protein phosphatase 1|protein phosphatase 1beta at 9C', map_location='1-31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10385960, end_position_on_the_genomic_accession=10409817, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  44291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44291, CurrentID=0, Status='live', Symbol='Idh', Aliases='Dmel_CG7176, CG7176, CT22171, Dmel\\CG7176, ICDH, IDH, IDH-NADP-NADP, cIdh, idh, isocitrate dehydrogenase, l(3)L3852, Idh', description='Isocitrate dehydrogenase', other_designations='CG7176-PA|CG7176-PB|CG7176-PC|CG7176-PD|CG7176-PF|CG7176-PG|CG7176-PI|CG7176-PJ|CG7176-PK|ISOCITRATE DEHYDROGENASE|Idh-PA|Idh-PB|Idh-PC|Idh-PD|Idh-PF|Idh-PG|Idh-PI|Idh-PJ|Idh-PK|NADP-IDH|isocitrate dehydrogenase|isocitrate dehydrogenase (NADP[+])|isocitrate dehydrolase', map_location='66C8-66C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8356404, end_position_on_the_genomic_accession=8360906, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  44297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44297, CurrentID=0, Status='live', Symbol='ome', Aliases='Dmel_CG42280, BEST:LD21715, CG17705, CG32143, CG32145, CG42280, CG9370, Dmel\\CG42280, Dmel_CG32143, Dmel_CG32145, LD21715, Ome', description='omega', other_designations='CG42280-PE|CG42280-PF|CG42280-PG|CG42280-PH|ome-PE|ome-PF|ome-PG|ome-PH', map_location='3-36 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14672839, end_position_on_the_genomic_accession=14747868, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19836171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836171, CurrentID=0, Status='live', Symbol='CG44434', Aliases='Dmel_ CG15086, Dmel\\CG44434', description='CG44434 gene product from transcript CG44434-RA', other_designations='CG44434-PA|uncharacterized protein', map_location='55E6-55E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18678680, end_position_on_the_genomic_accession=18684198, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19836175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836175, CurrentID=0, Status='live', Symbol='CG44242', Aliases='Dmel_ CG8446, Dmel\\CG44242', description='CG44242 gene product from transcript CG44242-RB', other_designations='CG44242-PA|CG44242-PB|uncharacterized protein', map_location='52F7-52F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16196384, end_position_on_the_genomic_accession=16196928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19836177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836177, CurrentID=0, Status='live', Symbol='CG45218', Aliases='Dmel_ Dmel\\CG45218', description='CG45218 gene product from transcript CG45218-RA', other_designations='CG45218-PA|uncharacterized protein', map_location='88E12-88E12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15290309, end_position_on_the_genomic_accession=15290799, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44307, CurrentID=0, Status='live', Symbol='Vha16-1', Aliases='Dmel_CG3161, ATP6V0C, BcDNA:SD02875, CG3161, Dmel\\CG3161, DucC, EP2372, MRE15, V-ATPase, Vha1, Vha16, Vhac, vha16, vha16-1', description='Vacuolar H[+] ATPase 16kD subunit 1', other_designations='CG3161-PA|CG3161-PB|CG3161-PC|CG3161-PD|Vha16-1-PA|Vha16-1-PB|Vha16-1-PC|Vha16-1-PD|c subunit|ductin|ductin subunit C|ductin, vacuolar H(+)-ATPase subunit C proteolipid|mRNA-like ncRNA in embryogenesis 15|vacuolar H[+] ATPase 16kD subunit|vacuolar H[+] ATPase 16kD subunit 1|vacuolar H[+] ATPase subunit 16-1|vacuolar H[+]-ATPase c subunit', map_location='42B2-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6625242, end_position_on_the_genomic_accession=6633090, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19836181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836181, CurrentID=0, Status='live', Symbol='CG43998', Aliases='Dmel_ BcDNA:AT03227, CG13602, CR43998, Dmel\\CG43998', description='CG43998 gene product from transcript CG43998-RB', other_designations='CG43998-PB|CG43998-PC|CG43998-PD|uncharacterized protein', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23927728, end_position_on_the_genomic_accession=23929000, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44315, CurrentID=0, Status='live', Symbol='pie', Aliases='Dmel_CG5354, CG5354, Dmel\\CG5354, l(2)31Ek, l(2)54, mat(2)QM47, mat(2)ea-E, mat(2)earlyQM47', description='pineapple eye', other_designations='CG5354-PA|CG5354-PB|lethal(2)54|maternal effect early arrest E|pie-PA|pie-PB|rumpled', map_location='2-43 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10420625, end_position_on_the_genomic_accession=10422780, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19836188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836188, CurrentID=0, Status='live', Symbol='CG44426', Aliases='Dmel_ Dmel\\CG44426, lincRNA.1032', description='CG44426 gene product from transcript CG44426-RA', other_designations='CG44426-PA|uncharacterized protein', map_location='19E5-19E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20978286, end_position_on_the_genomic_accession=20978868, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44317, CurrentID=0, Status='live', Symbol='CanB', Aliases='Dmel_CG4209, BcDNA:RH11383, CG4209, CNB, Can, CnB, Dmel\\CG4209, canB-4F, dCnB1', description='Calcineurin B', other_designations='CG4209-PA|CG4209-PB|CanB-PA|CanB-PB|calcineurin B', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5332499, end_position_on_the_genomic_accession=5334449, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44324, CurrentID=0, Status='live', Symbol='Ms', Aliases='Dmel_CG6440, CG6440, DMS, Dm-MS, Dmel\\CG6440, Dms, Drm-MS, Drome-MS, FLRFa, MEMS, MS, Nems, cg6440, dms', description='Myosuppressin', other_designations='CG6440-PA|CG6440-PB|CG6440-PC|Drm-myosuppressin|Ms-PA|Ms-PB|Ms-PC|TDVDHVFLRFamide|dromyosuppressin|myosuppressin|neomyosuppressin', map_location='95F14-95F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24341551, end_position_on_the_genomic_accession=24343825, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19836197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836197, CurrentID=0, Status='live', Symbol='CG45073', Aliases='Dmel_ CG18682, Dmel\\CG45073', description='CG45073 gene product from transcript CG45073-RA', other_designations='CG45073-PA|uncharacterized protein', map_location='99D7-99D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30123801, end_position_on_the_genomic_accession=30124735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44326, CurrentID=0, Status='live', Symbol='mRpL12', Aliases='Dmel_CG5012, BcDNA:RE53508, CG5012, Dmel\\CG5012, L7, L7/12A, L7/L12, Mrp17, l(3)10534, l(3)j4B2, l(3)rL424, mRpL7-L12', description='mitochondrial ribosomal protein L12', other_designations='CG5012-PA|mRpL12-PA|mitochondrial ribosomal protein L5/L17|mitochondrial ribosomal protein L7/L12', map_location='66E5-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8956791, end_position_on_the_genomic_accession=8957584, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44339, CurrentID=0, Status='live', Symbol='imd', Aliases='Dmel_CG5576, BG5, CG5576, Dmel\\CG5576, IMD, Imd, anon-WO0172774.166, shadok', description='immune deficiency', other_designations='CG5576-PA|CG5576-PB|imd-PA|imd-PB|immune deficient|immunodeficiency', map_location='55C8-55C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18409796, end_position_on_the_genomic_accession=18413599, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19836213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836213, CurrentID=0, Status='live', Symbol='CG45486', Aliases='Dmel_ Dmel\\CG45486', description='CG45486 gene product from transcript CG45486-RA', other_designations='CG45486-PA|uncharacterized protein', map_location='19A2-19A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19909318, end_position_on_the_genomic_accession=19910117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19836220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836220, CurrentID=0, Status='live', Symbol='CG45092', Aliases='Dmel_ Dmel\\CG45092', description='CG45092 gene product from transcript CG45092-RF', other_designations='CG45092-PD|CG45092-PE|CG45092-PF|uncharacterized protein', map_location='42C5-42C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6699695, end_position_on_the_genomic_accession=6715892, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  5740413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740413, CurrentID=0, Status='live', Symbol='CG34295', Aliases='Dmel_ Dmel\\CG34295', description='CG34295 gene product from transcript CG34295-RA', other_designations='CG34295-PA|uncharacterized protein', map_location='98E4-98E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28789617, end_position_on_the_genomic_accession=28790483, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44353, CurrentID=0, Status='live', Symbol='Shark', Aliases='Dmel_CG18247, CG18247, Dmel\\CG18247, Dtk7, SHARK, SYK/SHARK, Syk, Tk7, dtk7, l(2)W4, l(2R)W4, shark', description='SH2 ankyrin repeat kinase', other_designations='CG18247-PA|Shark-PA|Src homology 2 ankyrin repeat tyrosine kinase|Src homology 2, ankyrin repeat, tyrosine kinase|Src homology2, ankyrin repeat, tyrosine kinase|tyrosine kinase 7', map_location='52F8-52F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16199426, end_position_on_the_genomic_accession=16202884, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44355, CurrentID=0, Status='live', Symbol='drl', Aliases='Dmel_CG17348, CG10758, CG17348, DRL, Dmel\\CG17348, Drl, lin, lio, lio/drl', description='derailed', other_designations='CG17348-PA|CG17348-PB|drl-PA|drl-PB|linotte|lionette', map_location='37C7-37C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19190342, end_position_on_the_genomic_accession=19208621, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19836228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836228, CurrentID=0, Status='live', Symbol='CG45413', Aliases='Dmel_ Dmel\\CG45413', description='CG45413 gene product from transcript CG45413-RA', other_designations='CG45413-PA|CG45413-PB|uncharacterized protein', map_location='65D3-65D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6931023, end_position_on_the_genomic_accession=6931950, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44359, CurrentID=0, Status='live', Symbol='sda', Aliases='Dmel_CG5518, BcDNA:GH07466, CG5518, Dmel\\CG5518, Sda', description='slamdance', other_designations='CG5518-PA|CG5518-PB|CG5518-PC|CG5518-PD|sda-PA|sda-PB|sda-PC|sda-PD', map_location='97D7-97D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26915761, end_position_on_the_genomic_accession=26945309, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  44360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44360, CurrentID=0, Status='live', Symbol='Rab18', Aliases='Dmel_CG3129, 24639915, AAF46057, CG3129, DRABR4, DRabRP4, DmRab18, Dmel\\CG3129, O18337, Rab-RP4, Rab-r4/RP4, rab18', description='CG3129 gene product from transcript CG3129-RA', other_designations='CG3129-PA|Rab18-PA|rab-related protein 4', map_location='5A11-5A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5670827, end_position_on_the_genomic_accession=5671812, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19836240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836240, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-AB', Aliases='Dmel_CG44877, CG44877, Dmel\\CG44877', description='CG44877 gene product from transcript CG44877-RA', other_designations='CG44877-PA|CG44877-PB|pre-mod(mdg4)-AB-PA|pre-mod(mdg4)-AB-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21362968, end_position_on_the_genomic_accession=21365290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19836250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836250, CurrentID=0, Status='live', Symbol='Cyp6a16', Aliases='Dmel_CG7249, 6a16, CG7249, CR7249Psi, Cyp6a16p, Dmel\\CG7249', description='CG7249 gene product from transcript CG7249-RB', other_designations='CG7249-PB|CG7249-PC|Cyp6a16-PB|Cyp6a16-PC|Cyp6a16Psi', map_location='25F1-25F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5620615, end_position_on_the_genomic_accession=5623762, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44380, CurrentID=0, Status='live', Symbol='fok', Aliases='Dmel_CG43690, 38B.11, CG10746, CG42849, CG43690, Dmel\\CG43690, Dmel_CG10746, Dmel_CG42849, Fok, l(2)k07614, noncoding_3938', description='fledgling of Klp38B', other_designations='CG43690-PA|CG43690-PC|CG43690-PD|CG43690-PE|fok-PA|fok-PC|fok-PD|fok-PE', map_location='38B4-38B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20079270, end_position_on_the_genomic_accession=20085070, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44381, CurrentID=0, Status='live', Symbol='Tim17b2', Aliases='Dmel_CG15257, BG:DS01845.2, BcDNA:AT05822, BcDNA:AT13511, CG15257, Dmel\\CG15257, Tim17, dmTim17', description='Translocase inner membrane 17', other_designations='CG15257-PA|CG15257-PC|Tim17b2-PA|Tim17b2-PC|translocase inner membrane 17', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15484729, end_position_on_the_genomic_accession=15485545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44382, CurrentID=0, Status='live', Symbol='Sr-CIII', Aliases='Dmel_CG31962, CG31962, Dmel\\CG31962, SR-CIII, dSR-CIII', description='Scavenger receptor class C, type III', other_designations='CG31962-PA|Sr-CIII-PA|scavenger receptor|scavenger receptor class C, type III', map_location='24D6-24D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4120343, end_position_on_the_genomic_accession=4121627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318693, CurrentID=0, Status='live', Symbol='CG31347', Aliases='Dmel_ Dmel\\CG31347', description='CG31347 gene product from transcript CG31347-RA', other_designations='CG31347-PA|uncharacterized protein', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12691733, end_position_on_the_genomic_accession=12692597, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44390, CurrentID=0, Status='live', Symbol='RFeSP', Aliases='Dmel_CG7361, CG7361, Dmel\\CG7361, H, Noble, RISP, RfeSP, UCRI, UQCR-RFeSP, anon-EST:Posey219, l(2)k11704', description='Rieske iron-sulfur protein', other_designations='CG7361-PB|CG7361-PD|CG7361-PE|Not out of the blue|RFeSP-PB|RFeSP-PD|RFeSP-PE|RIESKE IRON-SULFUR PROTEIN|Ubiquinol-cytochrome C reductase iron-sulfur subunit|Ubiquinol-cytochrome c oxidoreductase Iron-sulfur|Ubiquinol-cytochrome c reductase rieske iron sulfur protein|rieske iron sulfur protein|rieske iron sulphur protein|rieske iron- sulfur protein of complex III|rieske iron-sulfur protein', map_location='22A3-22A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1612429, end_position_on_the_genomic_accession=1614160, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44391, CurrentID=0, Status='live', Symbol='Nnp-1', Aliases='Dmel_CG12396, CG12396, Dmel\\CG12396, NNP-1, Nnp1, l(2)k07826, nnp-1', description='CG12396 gene product from transcript CG12396-RA', other_designations='CG12396-PA|Nnp-1-PA|Nnp1', map_location='34B7-34B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13363308, end_position_on_the_genomic_accession=13366189, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44392, CurrentID=0, Status='live', Symbol='Rpp30', Aliases='Dmel_CG11606, #19/1, CG11606, Dmel\\CG11606, RNase P, l(2)k01901', description='RNaseP protein p30', other_designations='CG11606-PA|CG11606-PB|Rpp30-PA|Rpp30-PB', map_location='21B5-21B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=251039, end_position_on_the_genomic_accession=252373, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44409, CurrentID=0, Status='live', Symbol='Cbp80', Aliases='Dmel_CG7035, CBP80, CG7035, Dmel\\CG7035, EG:84H4.3, cbp80', description='cap binding protein 80', other_designations='CG7035-PA|CG7035-PB|Cap binding protein 80|Cap-binding protein 80|Cbp80-PA|Cbp80-PB', map_location='4C11-4C11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4650015, end_position_on_the_genomic_accession=4658259, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  44432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44432, CurrentID=0, Status='live', Symbol='IKKbeta', Aliases='Dmel_CG4201, CG4201, DIK, DLAK, DmIKK, DmIKKBeta, DmIKKb, DmIKKbeta, DmIkkbeta, Dmel\\CG4201, Dmikkb, Dmikkbeta, IK, IKK, IKK-beta, IKKB, IKKb, IRD-5, IRD5, Ik, Ikkb, Ikkbeta, Ird-5, Ird5, LRK, Lak, anon-89Bd, dIKK, dIKK-beta, dIKKbeta, dLak, dmIKKbeta, ikk, ird, ird-5, ird5, ird5/DmIkk[beta]', description='I-kappaB kinase beta', other_designations='CG4201-PB|IKK-like protein|IKKbeta-PB|LPS-responsive kinase|LPS-responsive-kinase|cactus kinase|immune response deficient 5|immune response-deficient 5|lipopolysaccharide activated kinase|mmune response deficient 5', map_location='89B1-89B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16046347, end_position_on_the_genomic_accession=16049421, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44433, CurrentID=0, Status='live', Symbol='Hs2st', Aliases='Dmel_CG10234, CG10234, DmHs2st, Dmel\\CG10234, HS2ST-2, dHS2ST, dmHS-2OST, dmHs2st, Hs2st', description='Heparan sulfate 2-O-sulfotransferase', other_designations='2-O sulfotransferase|CG10234-PA|HS 2-O|Hs2st-PA|heparan sulfate 2-O-sulfotransferase|heparan-sulfate-2-sulfotransferase', map_location='37E1-37E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19442833, end_position_on_the_genomic_accession=19444750, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44437, CurrentID=0, Status='live', Symbol='Sec13', Aliases='Dmel_CG6773, BEST:CK00043, BcDNA:LD03471, CG6773, CK00043, Dmel\\CG6773, Dsec13, SEC13, l(3)01031, sec13', description='Sec13 ortholog (S. cerevisiae)', other_designations='CG6773-PA|CG6773-PB|Sec13 ortholog|Sec13-PA|Sec13-PB|sec13', map_location='94E13-94E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23356278, end_position_on_the_genomic_accession=23358531, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44438, CurrentID=0, Status='live', Symbol='EfTuM', Aliases='Dmel_CG6050, CG6050, Dmel\\CG6050, efTuM, eftu-M, l(3)L4569, mtEF-Tu', description='Elongation factor Tu mitochondrial', other_designations='CG6050-PA|CG6050-PB|ELONGATION FACTOR Tu MITOCHONDRIAL|EfTuM-PA|EfTuM-PB|elongation factor Tu mitochondrial', map_location='50A14-50A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13507557, end_position_on_the_genomic_accession=13509332, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44441, CurrentID=0, Status='live', Symbol='Bem46', Aliases='Dmel_CG18642, CG18642, Dmel\\CG18642, c1.6', description='CG18642 gene product from transcript CG18642-RA', other_designations='Bem46-PA|CG18642-PA', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2988116, end_position_on_the_genomic_accession=2990075, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44446, CurrentID=0, Status='live', Symbol='HDAC3', Aliases='Dmel_CG2128, CG2128, DHDAC3, DmHDAC3, Dmel\\CG2128, Dromel_CG2128_FBtr0078767_hdac3_mORF, HDAC, Hdac 3, Hdac3, dHDAC3, dmHDA402, hdac3', description='Histone deacetylase 3', other_designations='CG2128-PA|HDAC3-PA|Hdac3|histone Deacetylase-3|histone deacetylase 3', map_location='83A4-83A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5472810, end_position_on_the_genomic_accession=5474600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44448, CurrentID=0, Status='live', Symbol='scrib', Aliases='Dmel_CG43398, 0424/05, CG31082, CG42614, CG43398, CG5462, CG5467, CT17324, Dmel\\CG43398, Dmel_CG31082, Dmel_CG42614, Dmel_CG5462, Dmel_CG5467, MENE (3R)-F, SCR, Scb, Scrb, Scrib, ird, ird15, l(3)673, l(3)S042405, l(3)c00119, l(3)j7B3, scrb, smi97B, vart', description='scribbled', other_designations='CG43398-PA|CG43398-PB|CG43398-PC|CG43398-PD|CG43398-PH|CG43398-PI|CG43398-PJ|CG43398-PK|CG43398-PL|CG43398-PM|CG43398-PN|CG43398-PO|CG43398-PP|CG43398-PQ|CG43398-PR|CG43398-PS|CG43398-PT|CG43398-PU|gp150-like|immune response deficient 15|lethal (3) c00119|scrib-PA|scrib-PB|scrib-PC|scrib-PD|scrib-PH|scrib-PI|scrib-PJ|scrib-PK|scrib-PL|scrib-PM|scrib-PN|scrib-PO|scrib-PP|scrib-PQ|scrib-PR|scrib-PS|scrib-PT|scrib-PU|scribble|scribbles|smell impaired 97B|vartul', map_location='97B9-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26536350, end_position_on_the_genomic_accession=26604051, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  44451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44451, CurrentID=0, Status='live', Symbol='ear', Aliases='Dmel_CG4913, CG4913, Dmel\\CG4913, ENL/AF-9, ENL/AF9, Ear, dENL/AF9', description='ENL/AF9-related', other_designations='CG4913-PA|CG4913-PB|CG4913-PC|ear-PA|ear-PB|ear-PC', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15320756, end_position_on_the_genomic_accession=15325279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  34717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=34717, CurrentID=0, Status='live', Symbol='CG15482', Aliases='Dmel_ Dmel\\CG15482', description='CG15482 gene product from transcript CG15482-RA', other_designations='CG15482-PA|uncharacterized protein', map_location='34A9-34A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13210773, end_position_on_the_genomic_accession=13211814, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44481, CurrentID=0, Status='live', Symbol='Met75Cb', Aliases='Dmel_CG18064, CG18064, Dmel\\CG18064, bs14h01.y1', description='CG18064 gene product from transcript CG18064-RA', other_designations='CG18064-PA|Met75Cb-PA', map_location='75C7-75C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18469378, end_position_on_the_genomic_accession=18469618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44482, CurrentID=0, Status='live', Symbol='Met75Ca', Aliases='Dmel_CG32197, CG32197, Dmel\\CG32197', description='CG32197 gene product from transcript CG32197-RA', other_designations='CG32197-PA|Met75Ca-PA', map_location='75C7-75C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18472143, end_position_on_the_genomic_accession=18472384, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44483, CurrentID=0, Status='live', Symbol='GluRIID', Aliases='Dmel_CG18039, CG18039, CT13538, DGluRIID, DKaiR-IA, DmelGluRIID, Dmel\\CG18039, GluR-IID, GluRIA, Kai-RIA, KaiRIA, brec, gluRIID', description='Glutamate receptor IID', other_designations='CG18039-PA|GluRIID-PA|KaiRIA|bad reception|glutamate receptor IID', map_location='92F4-92F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20648595, end_position_on_the_genomic_accession=20652450, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  44484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44484, CurrentID=0, Status='live', Symbol='GluRIB', Aliases='Dmel_CG43743, CG42509, CG43743, CG4481, DGluR-IB, Dm_3L:43677, DmelGlu-R1B, Dmel\\CG43743, Dmel_CG42509, Dmel_CG4481, Glu-R1B, Glu-RIB, GluRIb, dglur-IB', description='Glutamate receptor IB', other_designations='CG43743-PA|CG43743-PB|CG43743-PC|CG43743-PD|GluRIB-PA|GluRIB-PB|GluRIB-PC|GluRIB-PD|glutamate receptor|glutamate receptor IB', map_location='67A4-67A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9251978, end_position_on_the_genomic_accession=9288522, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14462405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462405, CurrentID=0, Status='live', Symbol='CG43997', Aliases='Dmel_ CG12433, Dmel\\CG43997', description='CG43997 gene product from transcript CG43997-RA', other_designations='CG43997-PA|uncharacterized protein', map_location='15D3-15D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17023499, end_position_on_the_genomic_accession=17024042, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462410, CurrentID=0, Status='live', Symbol='CG43841', Aliases='Dmel_ CG32551, Dmel\\CG43841', description='CG43841 gene product from transcript CG43841-RA', other_designations='CG43841-PA|uncharacterized protein', map_location='17B1-17B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18439667, end_position_on_the_genomic_accession=18440076, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462411, CurrentID=0, Status='live', Symbol='CG43843', Aliases='Dmel_ CG15049, Dmel\\CG43843', description='CG43843 gene product from transcript CG43843-RA', other_designations='CG43843-PA|uncharacterized protein', map_location='17B3-17B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18459138, end_position_on_the_genomic_accession=18460418, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44496, CurrentID=0, Status='live', Symbol='Uba2', Aliases='Dmel_CG7528, CG7528, DmSAE2, DmSae2, DmUba2, Dmel\\CG7528, Dmuba2, L36, SAE2, Sae2, UBA2, Ubc2, dUba2, sip2, uba2', description='Ubiquitin activating enzyme 2', other_designations='CG7528-PA|Smt3 activating enzyme 2|Smt3-activating enzyme 2|Uba2-PA|ubiquitin activating enzyme 2|ubiquitin-activating enzyme (E1)', map_location='66B7-66B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8125327, end_position_on_the_genomic_accession=8127818, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  44497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44497, CurrentID=0, Status='live', Symbol='Tollo', Aliases='Dmel_CG6890, CG6890, CT21344, Dmel\\CG6890, TOLL 8, Tl-8, Toll-8, dTLR8, dToll8, toll, tollo, tollo/toll-8', description='CG6890 gene product from transcript CG6890-RA', other_designations='CG6890-PA|Toll-8|Tollo-PA|toll receptor|tollo', map_location='71C1-71C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15235619, end_position_on_the_genomic_accession=15242832, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44498, CurrentID=0, Status='live', Symbol='Surf1', Aliases='Dmel_CG9943, BcDNA:GH23574, CG9943, Dmel\\CG9943, SUR1', description='Surfeit 1', other_designations='CG9943-PA|Surf1-PA|surfeit 1|surfeit locus protein 1', map_location='65D4-65D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6958346, end_position_on_the_genomic_accession=6959586, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44505, CurrentID=0, Status='live', Symbol='br', Aliases='Dmel_CG11491, 2B5, 2Bc, A18, BR, BR-C, BR-C Z1, BR-c, BRC, BRC-Z1, BRC-Z2, BRC-Z3, BRC-Z4, BR_C, Br, Br-C, Br-C Z2, Br-C-Z3, Br-Z4, BrC, BrC-Z1, BrZ3, CG11491, CG11509, CG11511, CG11514, Dmel\\CG11491, Dmel_CG11509, EG:123F11.1, EG:17A9.1, EG:25D2.1, PP1, PP2, Z1, Z2, Z3, Z4-C, br-Z1, br-Z2, br-Z3, br-Z4, de12, ecs, l(1)2Ba, l(1)2Bab, l(1)2Bad, l(1)2Bb, l(1)2Bc, l(1)2Bd, l(1)ESHS5, l(1)G0018, l(1)G0042, l(1)G0284, l(1)G0284a, l(1)G0318, l(1)G0401, l(1)PP1, l(1)d norm-12, l(1)d.norm.1, l(1)dn1, l(1)n34, l(1)npr-1, l(1)npr1, l(1)pp-1, l(1)pp-2, l(1)pp1, l(1)pp2, l(1)ts132, l(1)ts144, l(1)ts358, l(1)ts376, npr, npr-1, npr1, nprl, nrp, nrp1, o.c.c., rbp, rdp, rds, uq, br', description='broad', other_designations='Br-C|CG11491-PA|CG11491-PB|CG11491-PC|CG11491-PE|CG11491-PG|CG11491-PH|CG11491-PI|CG11491-PJ|CG11491-PL|CG11491-PM|CG11491-PN|CG11491-PO|CG11491-PP|CG11491-PQ|br-PA|br-PB|br-PC|br-PE|br-PG|br-PH|br-PI|br-PJ|br-PL|br-PM|br-PN|br-PO|br-PP|br-PQ|broad complex|broad-complex|ecdysone-sensitivity|lethal(1)2Bab|lethal(1)2Bc|lethal(1)2Bd|lethal(1)discs normal|lethal(1)non-pupariating-1|non-pupariating|non-pupariating-1|overlapping-complementation-complex|reduced bristles on palps|reduced palps|unequal wings', map_location='1-0.28 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1575132, end_position_on_the_genomic_accession=1664869, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  5740826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740826, CurrentID=0, Status='live', Symbol='CG34189', Aliases='Dmel_ Dmel\\CG34189', description='CG34189 gene product from transcript CG34189-RA', other_designations='CG34189-PA|uncharacterized protein', map_location='53D6-53D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16754444, end_position_on_the_genomic_accession=16754925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44510, CurrentID=0, Status='live', Symbol='cv', Aliases='Dmel_CG12410, CG12410, Cv, Dmel\\CG12410, Tsg2-1, tsg2, cv', description='crossveinless', other_designations='CG12410-PA|CG12410-PB|CG12410-PC|crossvein|cv-PA|cv-PB|cv-PC|twisted gastrulation 2', map_location='1-13.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5690002, end_position_on_the_genomic_accession=5693084, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44513, CurrentID=0, Status='live', Symbol='PPO3', Aliases='Dmel_CG42640, CG2952, CG42640, DmePPO3, Dmel\\CG42640, Dmel_CG2952, Dox-A3, Pro-PO3, ProPO59, proPO59', description='Prophenoloxidase 3', other_designations='CG42640-PA|PPO3-PA|diphenol OxidaseA3|prophenoloxidase 3|prophenoloxidase 59', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23064769, end_position_on_the_genomic_accession=23067183, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14462434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462434, CurrentID=0, Status='live', Symbol='CG44088', Aliases='Dmel_ Dmel\\CG44088', description='CG44088 gene product from transcript CG44088-RA', other_designations='CG44088-PA|uncharacterized protein', map_location='82D5-82D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4774516, end_position_on_the_genomic_accession=4775053, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44522, CurrentID=0, Status='live', Symbol='robo2', Aliases='Dmel_CG5481, CG14347, CG14348, CG5481, CG5574, CT17326, D-Robo2, D-robo2, Dmel\\CG5481, ROBO2, Robo 2, Robo-2, Robo2, Robo2 (Lea), anon-EST:Liang-1.75, clone 1.75, dRobo-2, fus4, lea, robo-2', description='roundabout 2', other_designations='CG5481-PA|CG5481-PB|Robo2|Roundabout2|SGP cluster fusion defects 4|leak|robo2-PA|robo2-PB', map_location='2-3 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1380086, end_position_on_the_genomic_accession=1420453, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14462445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462445, CurrentID=0, Status='live', Symbol='CG43675', Aliases='Dmel_ BcDNA:RE70963, CG16750, CG16751, CG33187, CG33188, CG3393, CG33936, CG33937, Dmel\\CG43675, anon-85Da, drn', description='CG43675 gene product from transcript CG43675-RC', other_designations='CG43675-PB|CG43675-PC|CG43675-PD|uncharacterized protein', map_location='85D11-85D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9355495, end_position_on_the_genomic_accession=9359189, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44529, CurrentID=0, Status='live', Symbol='CG2841', Aliases='Dmel_ Dmel\\CG2841, EG:BACH48C10.5, ptr', description='CG2841 gene product from transcript CG2841-RD', other_designations='CG2841-PA|CG2841-PB|CG2841-PC|CG2841-PD|CG2841-PE|CG2841-PF|proximal to raf', map_location='3A1-3A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2344949, end_position_on_the_genomic_accession=2366356, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  44531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44531, CurrentID=0, Status='live', Symbol='rg', Aliases='Dmel_CG44835, AKAP, Akap550, CG15464, CG15465/CG6775, CG15466, CG44835, CG6775, DAKAP550, Dmel\\CG44835, Dmel_CG15464, Dmel_CG6775, fas', description='rugose', other_designations='A kinase anchor protein 550|A-kinase anchoring protein|CG44835-PL|CG44835-PM|CG44835-PN|CG44835-PO|CG44835-PP|CG44835-PQ|CG44835-PR|drosophila A kinase anchor protein|facetious|pebbled|rg-PL|rg-PM|rg-PN|rg-PO|rg-PP|rg-PQ|rg-PR', map_location='1-11.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5085759, end_position_on_the_genomic_accession=5254864, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  14462452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462452, CurrentID=0, Status='live', Symbol='CG12256', Aliases='Dmel_ Dmel\\CG12256, SP133', description='CG12256 gene product from transcript CG12256-RA', other_designations='CG12256-PA|uncharacterized protein', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12225282, end_position_on_the_genomic_accession=12226339, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14462454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462454, CurrentID=0, Status='live', Symbol='CG44038', Aliases='Dmel_ Dmel\\CG44038', description='CG44038 gene product from transcript CG44038-RA', other_designations='CG44038-PA|uncharacterized protein', map_location='87B4-87B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12260969, end_position_on_the_genomic_accession=12261230, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462455, CurrentID=0, Status='live', Symbol='CG44037', Aliases='Dmel_ Dmel\\CG44037', description='CG44037 gene product from transcript CG44037-RA', other_designations='CG44037-PA|uncharacterized protein', map_location='87B4-87B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12258737, end_position_on_the_genomic_accession=12262879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44540, CurrentID=0, Status='live', Symbol='ct', Aliases='Dmel_CG11387, BcDNA:GH10590, CG11387, CT, CUT, Ct, Cut, Dmel\\CG11387, kf, l(1)7Ba, l(1)7Bb, l(1)VE614', description='cut', other_designations='CG11387-PA|CG11387-PC|CG11387-PD|CG11387-PE|Cut|ct-PA|ct-PC|ct-PD|ct-PE|kinked-femur', map_location='1-20.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7609148, end_position_on_the_genomic_accession=7678859, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  44543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44543, CurrentID=0, Status='live', Symbol='enc', Aliases='Dmel_CG10847, BB129, CG10847, Dmel\\CG10847, Enc, fs(3)63F-64A', description='encore', other_designations='CG10847-PA|CG10847-PB|CG10847-PD|CG10847-PE|CG10847-PF|CG10847-PG|CG10847-PH|enc-PA|enc-PB|enc-PD|enc-PE|enc-PF|enc-PG|enc-PH', map_location='63F1-63F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3818679, end_position_on_the_genomic_accession=3842326, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  44544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44544, CurrentID=0, Status='live', Symbol='Hnf4', Aliases='Dmel_CG9310, CG9310, DmHNF4, Dmel\\CG9310, HNF-4, HNF-4(D), HNF4, HNF4alpha, Hnf-4, Hnf-4h, NR2A4, dHNF-4, dHNF4', description='Hepatocyte nuclear factor 4', other_designations='CG9310-PA|CG9310-PB|CG9310-PD|CG9310-PE|CG9310-PF|Hnf4-PA|Hnf4-PB|Hnf4-PD|Hnf4-PE|Hnf4-PF|hepatocyte nuclear factor 4|hepatocyte nuclear factor 4 homologue', map_location='29E4-29E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8687226, end_position_on_the_genomic_accession=8709611, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  44548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44548, CurrentID=0, Status='live', Symbol='lola', Aliases='Dmel_CG12052, BEST:LD03274, BTB-IV, BcDNA:LD17006, BcDNA:RH31485, BtbIV, CG12052, CG18376, CG18378, CG18379, CG18380, CG18381, CG30012, CG30013, CG30014, Dmel\\CG12052, LD03274, Lola, NEST:bs06a08, bs06a08.y1, eyeful, fus7, l(2)00642, l(2)s3697, sw59', description='longitudinals lacking', other_designations='BTB-protein-IV|CG12052-PA|CG12052-PAA|CG12052-PB|CG12052-PC|CG12052-PD|CG12052-PE|CG12052-PF|CG12052-PG|CG12052-PH|CG12052-PI|CG12052-PJ|CG12052-PK|CG12052-PL|CG12052-PM|CG12052-PO|CG12052-PP|CG12052-PQ|CG12052-PR|CG12052-PT|CG12052-PU|CG12052-PV|CG12052-PW|CG12052-PX|CG12052-PY|CG12052-PZ|SGP cluster fusion defects 7|e|eyeful|lola-PA|lola-PAA|lola-PB|lola-PC|lola-PD|lola-PE|lola-PF|lola-PG|lola-PH|lola-PI|lola-PJ|lola-PK|lola-PL|lola-PM|lola-PO|lola-PP|lola-PQ|lola-PR|lola-PT|lola-PU|lola-PV|lola-PW|lola-PX|lola-PY|lola-PZ|longitudinal lacking|longitudinals absent|longitudinals-lacking|misguided', map_location='47A11-47A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10481894, end_position_on_the_genomic_accession=10543291, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  44554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44554, CurrentID=0, Status='live', Symbol='Bzd', Aliases='Dmel_CG13761, CG13761, Dmel\\CG13761, EG:BACR7C10.4', description='Buzidau', other_designations='Bzd-PB|CG13761-PB|buzidau', map_location='1-1.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2606680, end_position_on_the_genomic_accession=2608876, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14462475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462475, CurrentID=0, Status='live', Symbol='CG44040', Aliases='Dmel_ Dmel\\CG44040', description='CG44040 gene product from transcript CG44040-RA', other_designations='CG44040-PA|uncharacterized protein', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15320378, end_position_on_the_genomic_accession=15320691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462476, CurrentID=0, Status='live', Symbol='CG44013', Aliases='Dmel_ CG14872, Dmel\\CG44013', description='CG44013 gene product from transcript CG44013-RC', other_designations='CG44013-PB|CG44013-PC|uncharacterized protein', map_location='88F7-88F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15508896, end_position_on_the_genomic_accession=15509917, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462479, CurrentID=0, Status='live', Symbol='CG44044', Aliases='Dmel_ Dmel\\CG44044', description='CG44044 gene product from transcript CG44044-RA', other_designations='CG44044-PA|uncharacterized protein', map_location='89B12-89B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16321838, end_position_on_the_genomic_accession=16322189, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462482, CurrentID=0, Status='live', Symbol='CG44090', Aliases='Dmel_ Dmel\\CG44090', description='CG44090 gene product from transcript CG44090-RA', other_designations='CG44090-PA|uncharacterized protein', map_location='90A7-90A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17433991, end_position_on_the_genomic_accession=17434430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462483, CurrentID=0, Status='live', Symbol='CG44158', Aliases='Dmel_ Dmel\\CG44158', description='CG44158 gene product from transcript CG44158-RA', other_designations='CG44158-PA|uncharacterized protein', map_location='90F7-90F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18208779, end_position_on_the_genomic_accession=18209282, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462484, CurrentID=0, Status='live', Symbol='Rpb4', Aliases='Dmel_CG43662, Ada2a, CG31237, CG33520, CG43662, CG7150, DmRPB4, Dm Dmel\\CG43662, Dmel_CG33520, RPB4, dAda2a, dRPB4, dRpb4', description='CG43662 gene product from transcript CG43662-RF', other_designations='CG43662-PF|Rpb4-PF', map_location='90F10-90F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18252843, end_position_on_the_genomic_accession=18255487, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14462485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462485, CurrentID=0, Status='live', Symbol='koko', Aliases='Dmel_CG44010, CG31232, CG44010, CG7683, Dmel\\CG44010, Dmel_CG31232', description='kokopelli', other_designations='CG44010-PB|CG44010-PC|CG44010-PD|CG44010-PE|koko-PB|koko-PC|koko-PD|koko-PE', map_location='91A6-91A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18403026, end_position_on_the_genomic_accession=18404726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14462486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462486, CurrentID=0, Status='live', Symbol='CG43732', Aliases='Dmel_ Dm_3R:78579, Dmel\\CG43732', description='CG43732 gene product from transcript CG43732-RA', other_designations='CG43732-PA|CG43732-PB|uncharacterized protein', map_location='91E1-91E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18950196, end_position_on_the_genomic_accession=18976557, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  14462489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462489, CurrentID=0, Status='live', Symbol='CG44174', Aliases='Dmel_ Dmel\\CG44174, lincRNA.811', description='CG44174 gene product from transcript CG44174-RA', other_designations='CG44174-PA|uncharacterized protein', map_location='92A1-92A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19297246, end_position_on_the_genomic_accession=19297695, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44573, CurrentID=0, Status='live', Symbol='MAPk-Ak2', Aliases='Dmel_CG3086, CG3086, DmMAPKAPK-2, Dmel\\CG3086, MAPKAP, MAPKAPK-2, MK2, dMK2', description='MAP kinase activated protein-kinase-2', other_designations='CG3086-PB|CG3086-PD|CG3086-PE|CG3086-PF|MAP kinase activated protein kinase-2|MAPk-Ak2-PB|MAPk-Ak2-PD|MAPk-Ak2-PE|MAPk-Ak2-PF', map_location='5B8-5B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5759908, end_position_on_the_genomic_accession=5764203, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14462512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462512, CurrentID=0, Status='live', Symbol='CG43694', Aliases='Dmel_ Dmel\\CG43694', description='CG43694 gene product from transcript CG43694-RA', other_designations='CG43694-PA|uncharacterized protein', map_location='94F3-94F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23425602, end_position_on_the_genomic_accession=23426029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462523, CurrentID=0, Status='live', Symbol='CG43935', Aliases='Dmel_ Dmel\\CG43935', description='CG43935 gene product from transcript CG43935-RA', other_designations='CG43935-PA|uncharacterized protein', map_location='97E1-97E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27065436, end_position_on_the_genomic_accession=27065807, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462537, CurrentID=0, Status='live', Symbol='CG44102', Aliases='Dmel_ Dmel\\CG44102', description='CG44102 gene product from transcript CG44102-RC', other_designations='CG44102-PB|CG44102-PC|uncharacterized protein', map_location='41C5-41C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4895730, end_position_on_the_genomic_accession=4907685, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44620, CurrentID=0, Status='live', Symbol='CG14630', Aliases='Dmel_ B, Dmel\\CG14630, EG:BACR7A4.9, anon-1Eb', description='CG14630 gene product from transcript CG14630-RA', other_designations='CG14630-PA|transcription unit B|uncharacterized protein', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1031496, end_position_on_the_genomic_accession=1033165, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44631, CurrentID=0, Status='live', Symbol='Ag5r', Aliases='Dmel_CG9538, Agr, Agr5, CG9538, Dmel\\CG9538, agr, anon-WO0140519.74', description='Antigen 5-related', other_designations='Ag5r-PA|Ag5r-PB|Ag5r-PC|CG9538-PA|CG9538-PB|CG9538-PC|antigen 5-related', map_location='12F4-12F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14784664, end_position_on_the_genomic_accession=14786551, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44632, CurrentID=0, Status='live', Symbol='Cyp4e1', Aliases='Dmel_CG2062, 4e1, CG2062, CYP4E1, Dmel\\CG2062, P-450, P450', description='Cytochrome P450-4e1', other_designations='CG2062-PA|Cyp4e1-PA|cytochrome P450-4e1', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8447451, end_position_on_the_genomic_accession=8449745, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318735, CurrentID=0, Status='live', Symbol='CG31437', Aliases='Dmel_ Dmel\\CG31437', description='CG31437 gene product from transcript CG31437-RA', other_designations='CG31437-PA|uncharacterized protein', map_location='96D3-96D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25383700, end_position_on_the_genomic_accession=25384623, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462556, CurrentID=0, Status='live', Symbol='CG43646', Aliases='Dmel_ Dmel\\CG43646, lincRNA.258, mlncRNA42E5-2', description='CG43646 gene product from transcript CG43646-RA', other_designations='CG43646-PA|uncharacterized protein', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7018760, end_position_on_the_genomic_accession=7019390, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44637, CurrentID=0, Status='live', Symbol='cutlet', Aliases='Dmel_CG33122, CG10030, CG33122, Dmel\\CG33122, ctl', description='CG33122 gene product from transcript CG33122-RA', other_designations='CG33122-PA|cutlet-PA|gilead', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3707230, end_position_on_the_genomic_accession=3711748, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  44642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44642, CurrentID=0, Status='live', Symbol='Apc', Aliases='Dmel_CG1451, APC, APC11, CG1451, D-APC, D-APC1, DAPC, Dm APC1, Dmel\\CG1451, apc, apc 1, apc1, d-APC, dAPC, dAPC1, dApc, dapc1, Apc', description='APC-like', other_designations='Apc-PA|Apc-PB|Apc-PC|CG1451-PA|CG1451-PB|CG1451-PC|adenomatous polyposis coli|adenomatous polyposis coli 1|adenomatous polyposis coli protein', map_location='98E5-98E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28832301, end_position_on_the_genomic_accession=28844748, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44643, CurrentID=0, Status='live', Symbol='Cnx99A', Aliases='Dmel_CG11958, CG11958, Cnx, Cnx99a, Dmel\\CG11958, cnx', description='Calnexin 99A', other_designations='CG11958-PA|CG11958-PB|CG11958-PD|CG11958-PE|Cnx99A-PA|Cnx99A-PB|Cnx99A-PD|Cnx99A-PE|calnexin|calnexin 99A|calnexin99A', map_location='99A7-99A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29315254, end_position_on_the_genomic_accession=29319644, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  44653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44653, CurrentID=0, Status='live', Symbol='wech', Aliases='Dmel_CG42396, CG1624, CG42396, Dmel\\CG42396, Dpld, Wech, dpld, lin-41', description='CG42396 gene product from transcript CG42396-RC', other_designations='CG42396-PA|CG42396-PB|CG42396-PC|CG42396-PD|CG42396-PE|dappled|wech-PA|wech-PB|wech-PC|wech-PD|wech-PE', map_location='43C5-43C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7478853, end_position_on_the_genomic_accession=7489823, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14462577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462577, CurrentID=0, Status='live', Symbol='pre-lola-G', Aliases='Dmel_CG44025, CG44025, Dmel\\CG44025', description='CG44025 gene product from transcript CG44025-RC', other_designations='CG44025-PC|CG44025-PD|pre-lola-G-PC|pre-lola-G-PD', map_location='47A11-47A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10481894, end_position_on_the_genomic_accession=10485252, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44660, CurrentID=0, Status='live', Symbol='xmas-1', Aliases='Dmel_CG32561, CG32561, CG8919, Dmel\\CG32561, l(1)15Ea, m(1)3, ms(1)3, ms3, xmas', description='CG32561 gene product from transcript CG32561-RA', other_designations='CG32561-PA|transcript g|xmas-1-PA', map_location='15E7-15E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17151231, end_position_on_the_genomic_accession=17154592, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14462585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462585, CurrentID=0, Status='live', Symbol='CG43691', Aliases='Dmel_ CG17049, CG17050, Dmel\\CG43691, Dmel_CG17049, Dmel_CG17050', description='CG43691 gene product from transcript CG43691-RD', other_designations='CG43691-PC|CG43691-PD|uncharacterized protein', map_location='50A3-50A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13389495, end_position_on_the_genomic_accession=13393361, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44668, CurrentID=0, Status='live', Symbol='SmD1', Aliases='Dmel_CG10753, BcDNA:RE39488, CG10753, D1-Sm, Dmel\\CG10753, dD1, l(2)SH3 003, l(3)SH003, smd1, snRNP, snRNP-69D, snRNP69D', description='Small ribonucleoprotein particle protein SmD1', other_designations='CG10753-PA|SmD1-PA|lethal (3) SH003|small nuclear ribonucleoprotein at 69D|small ribonucleoprotein particle protein SmD1', map_location='69D3-69D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12733364, end_position_on_the_genomic_accession=12734611, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44669, CurrentID=0, Status='live', Symbol='cwo', Aliases='Dmel_CG17100, AAF24476.1, CG17100, CWO, Cwo, Dmel\\CG17100, GM05287, bHLHe49, mel_cwo', description='clockwork orange', other_designations='CG17100-PA|CG17100-PB|CG17100-PC|CLOCKWORK ORANGE|clockworkorange|cwo-PA|cwo-PB|cwo-PC|stich', map_location='86B3-86B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10388258, end_position_on_the_genomic_accession=10400752, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14462591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462591, CurrentID=0, Status='live', Symbol='CG43829', Aliases='Dmel_ Dmel\\CG43829', description='CG43829 gene product from transcript CG43829-RA', other_designations='CG43829-PA|uncharacterized protein', map_location='51E5-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15193996, end_position_on_the_genomic_accession=15194505, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44672, CurrentID=0, Status='live', Symbol='Lk6', Aliases='Dmel_CG17342, CG17342, CG6929, Dmel\\CG17342, LK6, MESR8, MNK/LK6, anon-WO0172774.84, lk6', description='Lk6 kinase', other_designations='CG17342-PA|CG17342-PB|Lk6-PA|Lk6-PB', map_location='86E18-86E18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11750931, end_position_on_the_genomic_accession=11764457, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14462596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462596, CurrentID=0, Status='live', Symbol='CG43788', Aliases='Dmel_ Dmel\\CG43788', description='CG43788 gene product from transcript CG43788-RA', other_designations='CG43788-PA|uncharacterized protein', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16937597, end_position_on_the_genomic_accession=16938011, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462597, CurrentID=0, Status='live', Symbol='CG43789', Aliases='Dmel_ Dmel\\CG43789', description='CG43789 gene product from transcript CG43789-RA', other_designations='CG43789-PA|uncharacterized protein', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16938696, end_position_on_the_genomic_accession=16938995, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44679, CurrentID=0, Status='live', Symbol='ergic53', Aliases='Dmel_CG6822, CG6822, Dmel\\CG6822, ERGIC53, dERGIC53, d l(3)S1760, l(3)s1760, rhea', description='CG6822 gene product from transcript CG6822-RB', other_designations='CG6822-PA|CG6822-PB|ergic-53|ergic53-PA|ergic53-PB', map_location='66D6-66D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8534977, end_position_on_the_genomic_accession=8538013, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14462605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462605, CurrentID=0, Status='live', Symbol='Rgk1', Aliases='Dmel_CG44011, CG30127, CG44011, CG9811, CT27512, Dmel\\CG44011, Dmel_CG30127, Dmel_CG9811, RGK1', description='Rad, Gem/Kir family member 1', other_designations='CG44011-PA|CG44011-PB|CG44011-PC|Rgk1-PA|Rgk1-PB|Rgk1-PC', map_location='56C11-56D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19311148, end_position_on_the_genomic_accession=19356525, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14462611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462611, CurrentID=0, Status='live', Symbol='CG43709', Aliases='Dmel_ Dmel\\CG43709', description='CG43709 gene product from transcript CG43709-RA', other_designations='CG43709-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20775795, end_position_on_the_genomic_accession=20776409, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462612, CurrentID=0, Status='live', Symbol='CG43710', Aliases='Dmel_ Dmel\\CG43710', description='CG43710 gene product from transcript CG43710-RA', other_designations='CG43710-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20776727, end_position_on_the_genomic_accession=20777749, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462613, CurrentID=0, Status='live', Symbol='CG43711', Aliases='Dmel_ Dmel\\CG43711', description='CG43711 gene product from transcript CG43711-RA', other_designations='CG43711-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20778249, end_position_on_the_genomic_accession=20779487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462614, CurrentID=0, Status='live', Symbol='CG43666', Aliases='Dmel_ CG15226, Dmel\\CG43666', description='CG43666 gene product from transcript CG43666-RA', other_designations='CG43666-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20788195, end_position_on_the_genomic_accession=20789217, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462615, CurrentID=0, Status='live', Symbol='CG43667', Aliases='Dmel_ CG15226, Dmel\\CG43667', description='CG43667 gene product from transcript CG43667-RA', other_designations='CG43667-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20789295, end_position_on_the_genomic_accession=20790201, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462616, CurrentID=0, Status='live', Symbol='CG43677', Aliases='Dmel_ Dmel\\CG43677', description='CG43677 gene product from transcript CG43677-RA', other_designations='CG43677-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20790245, end_position_on_the_genomic_accession=20791147, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462617, CurrentID=0, Status='live', Symbol='CG43668', Aliases='Dmel_ CG15226, Dmel\\CG43668', description='CG43668 gene product from transcript CG43668-RA', other_designations='CG43668-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20791462, end_position_on_the_genomic_accession=20792192, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462620, CurrentID=0, Status='live', Symbol='CG43982', Aliases='Dmel_ Dmel\\CG43982', description='CG43982 gene product from transcript CG43982-RA', other_designations='CG43982-PA|uncharacterized protein', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21763995, end_position_on_the_genomic_accession=21764695, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44701, CurrentID=0, Status='live', Symbol='Coprox', Aliases='Dmel_CG3433, BcDNA:GM14838, CG3433, Dmel\\CG3433, l(2)k10617, l(2)k11018', description='Coproporphyrinogen oxidase', other_designations='CG3433-PA|CG3433-PB|Coprox-PA|Coprox-PB|coproporphyrinogen oxidase|corpoporphyrinogen III oxidase', map_location='27C7-27C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6964421, end_position_on_the_genomic_accession=6965887, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44702, CurrentID=0, Status='live', Symbol='Vha36-1', Aliases='Dmel_CG8186, ATP6V1D, CG8186, Dmel\\CG8186, Vha36, VhaD, anon-WO0153538.20, anon-WO0153538.21, anon-WO0153538.28, anon-WO02059370.27, l(2)k07207, vha36, vha36-1', description='Vacuolar H[+] ATPase 36kD subunit 1', other_designations='CG8186-PA|Vha36|Vha36-1-PA|vacuolar H-ATPase subunit D|vacuolar H[+] ATPase 36kD subunit 1|vacuolar H[+] ATPase subunit 36-1', map_location='52A10-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15541435, end_position_on_the_genomic_accession=15542498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44703, CurrentID=0, Status='live', Symbol='lolal', Aliases='Dmel_CG5738, 25/12, Ban, CG5738, Dmel\\CG5738, EP0647, EP647, LOLAL, Lolal, anon-EST:fe3D3, anon-WO0172774.186, ban, bat, l(2)k02512, l(2)kO2512', description='lola like', other_designations='CG5738-PA|CG5738-PB|CG5738-PC|CG5738-PD|CG5738-PE|CG5738-PF|CG5738-PG|CG5738-PH|anon-fast-evolving-3D3|batman|lola-like|lolal-PA|lolal-PB|lolal-PC|lolal-PD|lolal-PE|lolal-PF|lolal-PG|lolal-PH', map_location='55B9-55B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18154095, end_position_on_the_genomic_accession=18158192, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14462625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462625, CurrentID=0, Status='live', Symbol='CG43659', Aliases='Dmel_ Dm_2R:85606, Dmel\\CG43659', description='CG43659 gene product from transcript CG43659-RA', other_designations='CG43659-PA|uncharacterized protein', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23079380, end_position_on_the_genomic_accession=23079715, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462630, CurrentID=0, Status='live', Symbol='CG43776', Aliases='Dmel_ CG13569, CG13569-2, Dmel\\CG43776', description='CG43776 gene product from transcript CG43776-RA', other_designations='CG43776-PA|CG43776-PB|uncharacterized protein', map_location='60B11-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24123167, end_position_on_the_genomic_accession=24126630, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14462631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462631, CurrentID=0, Status='live', Symbol='CG43777', Aliases='Dmel_ CG13569, CG13569-3, Dmel\\CG43777', description='CG43777 gene product from transcript CG43777-RA', other_designations='CG43777-PA|CG43777-PB|CG43777-PC|uncharacterized protein', map_location='60B11-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24123167, end_position_on_the_genomic_accession=24126630, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  44724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44724, CurrentID=0, Status='live', Symbol='san', Aliases='Dmel_CG12352, CG12352, DmAAF34715, Dmel\\CG12352, NAA50, atado, span', description='separation anxiety', other_designations='CG12352-PA|atado|complementation group 2|san-PA|sanpodo', map_location='47F5-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11360706, end_position_on_the_genomic_accession=11361871, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44728, CurrentID=0, Status='live', Symbol='MED6', Aliases='Dmel_CG9473, BcDNA:LD15729, CG9473, Dmel\\CG9473, Med6, Trap32, anon-EST:Posey279, dMED6, dMed6, dTRAP32, med6, p35', description='Mediator complex subunit 6', other_designations='CG9473-PA|CG9473-PB|CG9473-PC|CG9473-PD|CG9473-PE|MED6-PA|MED6-PB|MED6-PC|MED6-PD|MED6-PE|mediator|mediator complex subunit 6', map_location='85E8-85E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9766031, end_position_on_the_genomic_accession=9768984, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  317917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317917, CurrentID=0, Status='live', Symbol='CG32212', Aliases='Dmel_ Dmel\\CG32212', description='CG32212 gene product from transcript CG32212-RA', other_designations='CG32212-PA|uncharacterized protein', map_location='76B4-76B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19487560, end_position_on_the_genomic_accession=19488348, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44736, CurrentID=0, Status='live', Symbol='CG3719', Aliases='Dmel_ BcDNA:LD03613, Dmel\\CG3719, EG:132E8.3', description='CG3719 gene product from transcript CG3719-RA', other_designations='CG3719-PA|thioredoxin', map_location='2A1-2A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1378885, end_position_on_the_genomic_accession=1380480, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462660, CurrentID=0, Status='live', Symbol='CG44072', Aliases='Dmel_ Dmel\\CG44072', description='CG44072 gene product from transcript CG44072-RA', other_designations='CG44072-PA|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1520034, end_position_on_the_genomic_accession=1520387, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462661, CurrentID=0, Status='live', Symbol='Gr22d', Aliases='Dmel_CG31930, CG31930, CR31930, Dmel\\CG31930, GR22dP, Gr2940.2, Gr22d', description='Gustatory receptor 22d', other_designations='CG31930-PB|Gr22d-PB|gustatory receptor 22d|gustatory receptor 2940.2', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1789141, end_position_on_the_genomic_accession=1790361, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318753, CurrentID=0, Status='live', Symbol='CG31473', Aliases='Dmel_ CG2649, Dmel\\CG31473', description='CG31473 gene product from transcript CG31473-RA', other_designations='CG31473-PA|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7811814, end_position_on_the_genomic_accession=7813009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44747, CurrentID=0, Status='live', Symbol='msk', Aliases='Dmel_CG7935, 61, CG7935, CIP-61, DIM-7, Dim-7, Dim7, DmRanBP7, Dmel\\CG7935, Imp7, Imp7/8, Msk, S(ls)2, dim-7, dim7', description='moleskin', other_designations='CG7935-PA|Importin-7|corkscrew interacting Protein-61|diminished-7|drosophila Importin-7|msk-PA', map_location='66B6-66B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8116179, end_position_on_the_genomic_accession=8121579, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  44748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44748, CurrentID=0, Status='live', Symbol='NiPp1', Aliases='Dmel_CG8980, CG8980, Dmel\\CG8980, NIPP-1Dm, NIPP1, NIPP1Dm, NiPp-1, nipp1', description='Nuclear inhibitor of Protein phosphatase 1', other_designations='CG8980-PA|NiPp1-PA|nuclear inhibitor of PP1|nuclear inhibitor of protein phosphatase 1', map_location='53E4-53E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17003563, end_position_on_the_genomic_accession=17005149, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44750, CurrentID=0, Status='live', Symbol='Ady43A', Aliases='Dmel_CG1851, Adk43A, Ady, CG1851, DmAdk3, Dmel\\CG1851, ady43A', description='CG1851 gene product from transcript CG1851-RA', other_designations='Ady43A-PA|CG1851-PA', map_location='43A1-43A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7243024, end_position_on_the_genomic_accession=7245776, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14462673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462673, CurrentID=0, Status='live', Symbol='CG43750', Aliases='Dmel_ Dmel\\CG43750', description='CG43750 gene product from transcript CG43750-RB', other_designations='CG43750-PA|CG43750-PB|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2389578, end_position_on_the_genomic_accession=2390689, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462674, CurrentID=0, Status='live', Symbol='CG44139', Aliases='Dmel_ Dmel\\CG44139', description='CG44139 gene product from transcript CG44139-RA', other_designations='CG44139-PA|uncharacterized protein', map_location='22E1-22E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2424797, end_position_on_the_genomic_accession=2425041, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462678, CurrentID=0, Status='live', Symbol='CG43851', Aliases='Dmel_ Dmel\\CG43851', description='CG43851 gene product from transcript CG43851-RA', other_designations='CG43851-PA|uncharacterized protein', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2647886, end_position_on_the_genomic_accession=2648361, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462685, CurrentID=0, Status='live', Symbol='CG43815', Aliases='Dmel_ Dmel\\CG43815', description='CG43815 gene product from transcript CG43815-RA', other_designations='CG43815-PA|uncharacterized protein', map_location='24A1-24A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3576210, end_position_on_the_genomic_accession=3576623, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462687, CurrentID=0, Status='live', Symbol='CG43703', Aliases='Dmel_ Dmel\\CG43703', description='CG43703 gene product from transcript CG43703-RA', other_designations='CG43703-PA|uncharacterized protein', map_location='24C3-24C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3765494, end_position_on_the_genomic_accession=3765863, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462692, CurrentID=0, Status='live', Symbol='CG43797', Aliases='Dmel_ Dmel\\CG43797', description='CG43797 gene product from transcript CG43797-RA', other_designations='CG43797-PA|uncharacterized protein', map_location='24D2-24D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4000018, end_position_on_the_genomic_accession=4000806, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462693, CurrentID=0, Status='live', Symbol='CG43774', Aliases='Dmel_ CG10039, Dmel\\CG43774', description='CG43774 gene product from transcript CG43774-RA', other_designations='CG43774-PA|uncharacterized protein', map_location='24D3-24D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4008653, end_position_on_the_genomic_accession=4009090, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462699, CurrentID=0, Status='live', Symbol='CG43798', Aliases='Dmel_ Dmel\\CG43798, lincRNA.43', description='CG43798 gene product from transcript CG43798-RA', other_designations='CG43798-PA|uncharacterized protein', map_location='25A2-25A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4647051, end_position_on_the_genomic_accession=4647920, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14462701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462701, CurrentID=0, Status='live', Symbol='CG44000', Aliases='Dmel_ CG31919, CG5822, Dmel\\CG44000, Dromel_CG31919_FBtr0079006_mORF', description='CG44000 gene product from transcript CG44000-RC', other_designations='CG44000-PB|CG44000-PC|CG44000-PD|CG44000-PE|CG44000-PF|uncharacterized protein', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5043270, end_position_on_the_genomic_accession=5048365, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  44783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44783, CurrentID=0, Status='live', Symbol='RpL37A', Aliases='Dmel_CG5827, BcDNA:RE23595, BcDNA:RH41593, CG5827, DmL37a, Dmel\\CG5827, L37A, M(2)25C, M(2)S1, RpL37a, chr2L:5071724..5071820, l(2)25Cb', description='Ribosomal protein L37A', other_designations='CG5827-PA|CG5827-PB|CG5827-PC|RpL37A-PA|RpL37A-PB|RpL37A-PC|minute (2) 25C|ribosomal protein L37A|ribosomal protein L37a', map_location='2-15.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5070957, end_position_on_the_genomic_accession=5072055, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  44785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44785, CurrentID=0, Status='live', Symbol='tan', Aliases='Dmel_CG6586, CG6586, Dmel\\CG6586', description='tantalus', other_designations='CG6586-PA|CG6586-PB|tan-PA|tan-PB', map_location='65A4-65A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6055136, end_position_on_the_genomic_accession=6057002, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44786, CurrentID=0, Status='live', Symbol='nerfin-1', Aliases='Dmel_CG13906, CG13906, Dmel\\CG13906, Nerfin, Nerfin-1, nerfin', description='nervous fingers 1', other_designations='CG13906-PA|Nerfin-1|nerfin|nerfin 1|nerfin-1-PA|nervous finger-1', map_location='61D4-61D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=909112, end_position_on_the_genomic_accession=912319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44787, CurrentID=0, Status='live', Symbol='miple2', Aliases='Dmel_CG18321, CG18321, Dmel\\CG18321', description='midkine and pleiotrophin 2', other_designations='CG18321-PA|CG18321-PD|CG18321-PE|CG18321-PF|CG18321-PG|Miple2|miple2-PA|miple2-PD|miple2-PE|miple2-PF|miple2-PG', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=271200, end_position_on_the_genomic_accession=275018, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  44790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44790, CurrentID=0, Status='live', Symbol='TwdlT', Aliases='Dmel_CG5812, CG5812, Dmel\\CG5812, GCR(ich)', description='TweedleT', other_designations='CG5812-PA|GCR(ich)|TwdlT-PA', map_location='97D14-97D14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27049220, end_position_on_the_genomic_accession=27051004, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44791, CurrentID=0, Status='live', Symbol='Ef1gamma', Aliases='Dmel_CG11901, CG11901, Dmel\\CG11901, EF1gamma, Ef1Bggr, Ef1y, eEF1gamma', description='CG11901 gene product from transcript CG11901-RC', other_designations='CG11901-PA|CG11901-PB|CG11901-PC|Ef1gamma-PA|Ef1gamma-PB|Ef1gamma-PC', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29220172, end_position_on_the_genomic_accession=29222109, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14462712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462712, CurrentID=0, Status='live', Symbol='CG43922', Aliases='Dmel_ Dmel\\CG43922', description='CG43922 gene product from transcript CG43922-RA', other_designations='CG43922-PA|uncharacterized protein', map_location='26C2-26C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6308377, end_position_on_the_genomic_accession=6308751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44793, CurrentID=0, Status='live', Symbol='abo', Aliases='Dmel_CG6093, Abo, CG6093, Dmel\\CG6093', description='abnormal oocyte', other_designations='CG6093-PA|abo-PA', map_location='2-44.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10973443, end_position_on_the_genomic_accession=10975293, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462714, CurrentID=0, Status='live', Symbol='CG43799', Aliases='Dmel_ Dmel\\CG43799', description='CG43799 gene product from transcript CG43799-RA', other_designations='CG43799-PA|uncharacterized protein', map_location='27E6-27E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7227734, end_position_on_the_genomic_accession=7228641, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462715, CurrentID=0, Status='live', Symbol='CG43800', Aliases='Dmel_ Dmel\\CG43800', description='CG43800 gene product from transcript CG43800-RA', other_designations='CG43800-PA|CG43800-PB|uncharacterized protein', map_location='27E6-27E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7232165, end_position_on_the_genomic_accession=7232813, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44801, CurrentID=0, Status='live', Symbol='bsk', Aliases='Dmel_CG5680, BSK, Bsk, CG5680, D-JNK, D-junk, DBSK/JNK, DJNK, DJNK/bsk, Dmel\\CG5680, JNK, JNK/SAPK, Jnk, Junk, SAPKa, dJNK, jnk, pJNK', description='basket', other_designations='CG5680-PB|CG5680-PE|CG5680-PF|JNK|JUN kinase|Jun N-terminal kinase|Jun NH2-terminal kinase|Jun-N-terminal kinase|Jun-kinase|bsk-PB|bsk-PE|bsk-PF|c-Jun N-terminal kinase|c-Jun aminoterminal kinase|c-Jun-N-terminal kinase|pJNK', map_location='2-33 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10247502, end_position_on_the_genomic_accession=10250501, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14462726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462726, CurrentID=0, Status='live', Symbol='CG43796', Aliases='Dmel_ Dmel\\CG43796', description='CG43796 gene product from transcript CG43796-RA', other_designations='CG43796-PA|uncharacterized protein', map_location='29D1-29D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8482374, end_position_on_the_genomic_accession=8483005, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44811, CurrentID=0, Status='live', Symbol='E(bx)', Aliases='Dmel_CG32346, CG10894, CG17135, CG32346, CG32478, CG7022, Dmel\\CG32346, E(BX), E(Bx), Ebx, En-bx, NURF, NURF 215, NURF-215, NURF-301, NURF215, NURF301, NuRF301, Nurf 301, Nurf-215, Nurf301, dNURF, dNURF301, e(bx), l(3)122, l(3)ry122, nurf301, p215, p301', description='Enhancer of bithorax', other_designations='CG32346-PA|CG32346-PB|CG32346-PC|CG32346-PE|CG32346-PF|CG32346-PG|CG32346-PH|CG32346-PI|CG32346-PJ|E(bx)-PA|E(bx)-PB|E(bx)-PC|E(bx)-PE|E(bx)-PF|E(bx)-PG|E(bx)-PH|E(bx)-PI|E(bx)-PJ|enhancer of bithorax|nucleosome remodeling factor|nucleosome remodeling factor - 215kD', map_location='61B2-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=233926, end_position_on_the_genomic_accession=246912, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  318765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318765, CurrentID=0, Status='live', Symbol='CG31493', Aliases='Dmel_ CG2640, Dmel\\CG31493', description='CG31493 gene product from transcript CG31493-RA', other_designations='CG31493-PA|CG31493-PB|uncharacterized protein', map_location='84C4-84C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7131560, end_position_on_the_genomic_accession=7132695, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44817, CurrentID=0, Status='live', Symbol='for', Aliases='Dmel_CG10033, 142251_at, BcDNA:GM08338, CG10033, DG2, Dg2, Dmel\\CG10033, FOR/PKG, For, PKG, Pkg2, Pkg24A, anon-WO0140519.260, anon-WO02059370.47, dg2, l(2)06860', description='foraging', other_designations='CG10033-PA|CG10033-PC|CG10033-PD|CG10033-PE|CG10033-PG|CG10033-PH|CG10033-PI|CG10033-PJ|CG10033-PK|for-PA|for-PC|for-PD|for-PE|for-PG|for-PH|for-PI|for-PJ|for-PK', map_location='2-10 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3622074, end_position_on_the_genomic_accession=3656953, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  44819 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44819, CurrentID=0, Status='live', Symbol='g', Aliases='Dmel_CG10986, AP-3delta, AP3delta, CG10986, CG11197, Dmel\\CG10986, apl5, cg11197, delta, l(1)Mc19', description='garnet', other_designations='AP-3, delta subunit|CG10986-PB|CG10986-PD|CG10986-PF|adaptor protein complex 3, delta subunit|g-PB|g-PD|g-PF|lethal(1)Mc19|salmon', map_location='1-44.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13727204, end_position_on_the_genomic_accession=13736278, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14462740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462740, CurrentID=0, Status='live', Symbol='CG43725', Aliases='Dmel_ Dmel\\CG43725', description='CG43725 gene product from transcript CG43725-RA', other_designations='CG43725-PA|uncharacterized protein', map_location='32D2-32D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11075513, end_position_on_the_genomic_accession=11075962, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44830, CurrentID=0, Status='live', Symbol='kto', Aliases='Dmel_CG8491, Brm @ Kto, CG8491, Dmel\\CG8491, Kto, Kto/Med12, MED12, Med12, SRB8, Srb8/KTO/TRAP230, Trap230, dTRAP230, dTrap230, l(3)76BDc, med12, med12/kto, psg14, siren1', description='kohtalo', other_designations='CG8491-PA|CG8491-PB|khotalo|kto-PA|kto-PB|mediator complex subunit 12', map_location='3-46 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19836614, end_position_on_the_genomic_accession=19844944, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14462751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462751, CurrentID=0, Status='live', Symbol='CG43813', Aliases='Dmel_ Dmel\\CG43813', description='CG43813 gene product from transcript CG43813-RA', other_designations='CG43813-PA|uncharacterized protein', map_location='33E1-33E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12343643, end_position_on_the_genomic_accession=12344266, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44835, CurrentID=0, Status='live', Symbol='m', Aliases='Dmel_CG9369, CG9369, Dmel\\CG9369, M, Minin, m', description='miniature', other_designations='CG9369-PA|m-PA', map_location='1-36.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11754273, end_position_on_the_genomic_accession=11769508, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  5740837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740837, CurrentID=0, Status='live', Symbol='CG34425', Aliases='Dmel_ CG12584, CG41467, Dmel\\CG34425, Dmel_CG12584, Dmel_CG41467', description='CG34425 gene product from transcript CG34425-RA', other_designations='CG34425-PA|uncharacterized protein', map_location='82C1-82C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4504612, end_position_on_the_genomic_accession=4506365, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44839, CurrentID=0, Status='live', Symbol='mud', Aliases='Dmel_CG12047, CG12047, Dmel\\CG12047, KS63, Mud, NuMA', description='mushroom body defect', other_designations='CG12047-PF|CG12047-PG|CG12047-PH|CG12047-PI|CG12047-PJ|CG12047-PK|CG12047-PL|mud-PF|mud-PG|mud-PH|mud-PI|mud-PJ|mud-PK|mud-PL|mushroom bodies deranged|mushroom body defective', map_location='1-50 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14247732, end_position_on_the_genomic_accession=14258595, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14462761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462761, CurrentID=0, Status='live', Symbol='CG43779', Aliases='Dmel_ CG15639, Dmel\\CG43779', description='CG43779 gene product from transcript CG43779-RA', other_designations='CG43779-PA|uncharacterized protein', map_location='34B5-34B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13305920, end_position_on_the_genomic_accession=13306755, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462762, CurrentID=0, Status='live', Symbol='CG43850', Aliases='Dmel_ Dmel\\CG43850', description='CG43850 gene product from transcript CG43850-RA', other_designations='CG43850-PA|CG43850-PB|uncharacterized protein', map_location='34B11-34B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13415431, end_position_on_the_genomic_accession=13420645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44843, CurrentID=0, Status='live', Symbol='nkd', Aliases='Dmel_CG11614, CG11614, CG18224, Dmel\\CG11614, Nkd, naked', description='naked cuticle', other_designations='CG11614-PA|CG11614-PB|CG11614-PC|naked|nkd-PA|nkd-PB|nkd-PC', map_location='3-47 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19003741, end_position_on_the_genomic_accession=19045449, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  44848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44848, CurrentID=0, Status='live', Symbol='plu', Aliases='Dmel_CG9183, CG9183, Dmel\\CG9183, PLU, Plu', description='plutonium', other_designations='CG9183-PA|plu-PA', map_location='2-88 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20262958, end_position_on_the_genomic_accession=20263684, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44849, CurrentID=0, Status='live', Symbol='pnr', Aliases='Dmel_CG3978, CG3978, CG3978-PA, Dmel\\CG3978, GATA, GATA4, GATAa, GATAa/Pannier, PNR, Pnr, dGATAa, dGATAa/pannier, l(3)89B3, prn', description='pannier', other_designations='CG3978-PA|CG3978-PB|CG3978-PC|CG3978-PD|pnr-PA|pnr-PB|pnr-PC|pnr-PD', map_location='3-58 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16026079, end_position_on_the_genomic_accession=16041796, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  44851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44851, CurrentID=0, Status='live', Symbol='raw', Aliases='Dmel_CG12437, CG12437, CG9321, Dmel\\CG12437, cyr, ensh, l(2)06825, l(2)k03514', description='CG12437 gene product from transcript CG12437-RB', other_designations='CG12437-PA|CG12437-PB|CG12437-PC|CG12437-PD|CG12437-PE|cyrano|ensheathment defects|raw-PA|raw-PB|raw-PC|raw-PD|raw-PE', map_location='2-19 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8709880, end_position_on_the_genomic_accession=8740660, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14462774 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462774, CurrentID=0, Status='live', Symbol='CG44141', Aliases='Dmel_ Dmel\\CG44141', description='CG44141 gene product from transcript CG44141-RA', other_designations='CG44141-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15659041, end_position_on_the_genomic_accession=15659474, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44855, CurrentID=0, Status='live', Symbol='rib', Aliases='Dmel_CG7230, CG7230, Dmel\\CG7230, RIB, fus1, gene 5', description='ribbon', other_designations='CG7230-PA|CG7230-PB|CG7230-PC|SGP cluster fusion defects 1|rib-PA|rib-PB|rib-PC', map_location='2-88 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19270078, end_position_on_the_genomic_accession=19281396, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44856, CurrentID=0, Status='live', Symbol='ru', Aliases='Dmel_CG1214, CG1214, Dmel\\CG1214, Rho-3, Rho-related [62A1], rho-3, rho3', description='roughoid', other_designations='CG1214-PA|Rhomboid-3|rhomboid|rhomboid 3|roughoid/rhomboid-3|ru-PA', map_location='3-0.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1370628, end_position_on_the_genomic_accession=1387437, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14462777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462777, CurrentID=0, Status='live', Symbol='CG43924', Aliases='Dmel_ Dmel\\CG43924', description='CG43924 gene product from transcript CG43924-RA', other_designations='CG43924-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15710579, end_position_on_the_genomic_accession=15711099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  44858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44858, CurrentID=0, Status='live', Symbol='sad', Aliases='Dmel_CG14728, 315a1, CG14728, CYP315A1, Cyp315a1, Dmel\\CG14728, Sad', description='shadow', other_designations='CG14728-PA|cytochrome P450|ecdysteroid hydroxylase|karmoisin ghost|sad-PA', map_location='3-51 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11877235, end_position_on_the_genomic_accession=11879754, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  38592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38592, CurrentID=0, Status='live', Symbol='ATPsynCF6L', Aliases='Dmel_CG12027, ATPR, CG12027, Dmel\\CG12027', description='ATP synthase, coupling factor 6-like', other_designations='ATP synthase coupling factor 6|ATPsynCF6L-PA|CG12027-PA', map_location='64C4-64C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5011444, end_position_on_the_genomic_accession=5012133, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44861, CurrentID=0, Status='live', Symbol='sdt', Aliases='Dmel_CG32717, BP1063, CG12657, CG12658, CG15339, CG15340, CG15341, CG15342, CG1617, CG32717, Dmel\\CG32717, PALS1, Pals1, Sdt, Std, anon-EST:fe2E6, cGUK1, cMAGUK1, l(1)7Ef, pal1, std', description='stardust', other_designations='BP1063|CG32717-PB|CG32717-PD|CG32717-PE|CG32717-PF|CG32717-PG|CG32717-PH|CG32717-PI|CG32717-PJ|CG32717-PK|CG32717-PL|CG32717-PM|CG32717-PN|CG32717-PO|CG32717-PP|anon-fast-evolving-2E6|sdt-PB|sdt-PD|sdt-PE|sdt-PF|sdt-PG|sdt-PH|sdt-PI|sdt-PJ|sdt-PK|sdt-PL|sdt-PM|sdt-PN|sdt-PO|sdt-PP', map_location='1-23 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8178549, end_position_on_the_genomic_accession=8240474, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  14462783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462783, CurrentID=0, Status='live', Symbol='CG43816', Aliases='Dmel_ Dmel\\CG43816', description='CG43816 gene product from transcript CG43816-RB', other_designations='CG43816-PA|CG43816-PB|uncharacterized protein', map_location='35E1-35E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16004325, end_position_on_the_genomic_accession=16005461, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44870, CurrentID=0, Status='live', Symbol='sub', Aliases='Dmel_CG12298, CG12298, DmKlp54E, DmSub, Dmel\\CG12298, Dub, KIF 20A, KIF20A/B, Mklp2, SUB, Sub, mei-1794', description='subito', other_designations='CG12298-PA|CG12298-PB|double or nothing|double-or-nothing|sub-PA|sub-PB', map_location='2-86 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17745500, end_position_on_the_genomic_accession=17748078, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44873, CurrentID=0, Status='live', Symbol='Gr5a', Aliases='Dmel_CG15779, CG15779, Dmel5a, Dmel\\CG15779, GR5A, GR5a, GRLU.7, Gr5, Gr5A1, GrLU7, LU.7, Tre, han, tre', description='Gustatory receptor 5a', other_designations='CG15779-PA|Gr5a-PA|Trehalose-sensitivity|gustatory receptor 5a|gustatory receptor LU.7|trehalose receptor|trehalose sensitivity', map_location='1-13.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5680617, end_position_on_the_genomic_accession=5682488, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14462795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462795, CurrentID=0, Status='live', Symbol='CG43814', Aliases='Dmel_ Dmel\\CG43814', description='CG43814 gene product from transcript CG43814-RA', other_designations='CG43814-PA|uncharacterized protein', map_location='37A4-37A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18804491, end_position_on_the_genomic_accession=18805347, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44879, CurrentID=0, Status='live', Symbol='rhi', Aliases='Dmel_CG10683, CG10683, Dmel\\CG10683, HP1D, HP1d, RHI, Rhi, rno', description='rhino', other_designations='CG10683-PA|rhi-PA', map_location='54D2-54D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17628491, end_position_on_the_genomic_accession=17633516, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44885, CurrentID=0, Status='live', Symbol='mys', Aliases='Dmel_CG1560, BetaPS, BetaPS-Int, CG1560, CT40473, Dmel\\CG1560, EM28, ItgbPS, MAb6G11, Mys, PS, PS[[beta]], PSbeta, bPS, b[[PS]], beta integrin, beta-Int, beta-PS, beta-int, beta-integrin, beta1, betaInt, betaPS, betaPS Int, betaPS integrin, betaPS-integrin, betaPS1, beta[[PS]], beta[[PS]]-integrin, beta[[P]]S, beta[[mys]], betamys, integrin beta[[PS]], l(1)7Db, l(1)7Dn, l(1)93p, l(1)968, l(1)DA551, l(1)EM28, l(1)G0233, l(1)G0281, l(1)mys, nj42, olfC', description='myospheroid', other_designations='CG1560-PA|CG1560-PB|CG1560-PC|CG1560-PD|Integrin-betaPS|PS integrin beta subunit|PS-beta-integrin|PSbeta-integrin|beta PS integrin|beta integrin|beta-PS integrin|beta-PS-Integrin|beta-PS1 integrin|beta-integrin|beta1 integrin|betaIntegrin|betaPS|betaPS integrin|betaPS-Integrin|betaPS.*integrin|beta[[PS]] integrin|beta[[PS]] integrin subunit|beta[[PS]]-integrin|beta[[mys]]-integrin|integrin PSbeta|integrin betaPS|integrin betaPS subunit|integrin beta[[PS]] subunit|integrinbeta|lethal myospheroid|lethal(1)93 pupal|lethal(1)myospheroid|myo-spheroid|myospheriod|mys-PA|mys-PB|mys-PC|mys-PD|non-jumper-42|olfactory C|olfactory-C', map_location='7D5-7D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8061645, end_position_on_the_genomic_accession=8070237, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  44887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44887, CurrentID=0, Status='live', Symbol='mdy', Aliases='Dmel_CG31991, CG13273, CG17937, CG17938, CG31991, CT39962, DGAT, DGAT1, Dgat1, Dmel\\CG31991, Mdy, mdcds_3562', description='midway', other_designations='CG31991-PA|CG31991-PB|CG31991-PC|CG31991-PE|diacylglycerol acyltransferase|mdy-PA|mdy-PB|mdy-PC|mdy-PE', map_location='36B1-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16816660, end_position_on_the_genomic_accession=16826964, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  44889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44889, CurrentID=0, Status='live', Symbol='ph-d', Aliases='Dmel_CG3895, CG3895, DROZFP, Dmel\\CG3895, EG:BACN25G24.3, PH, PH-d, PHD, Ph, Ph-d, PhD, ph, ph-D, ph[D], ph[d], phd, phm', description='polyhomeotic distal', other_designations='CG3895-PA|CG3895-PB|ph-d-PA|ph-d-PB|polyhomeotic|polyhomeotic proximal|polyhomeotic-distal', map_location='1-0.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2115727, end_position_on_the_genomic_accession=2123286, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14462810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462810, CurrentID=0, Status='live', Symbol='CG43739', Aliases='Dmel_ Dm_2L:110584, Dmel\\CG43739', description='CG43739 gene product from transcript CG43739-RA', other_designations='CG43739-PA|uncharacterized protein', map_location='38F3-38F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20871335, end_position_on_the_genomic_accession=20918931, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  44900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44900, CurrentID=0, Status='live', Symbol='Larp7', Aliases='Dmel_CG42569, CG12058, CG42569, Dmel\\CG42569, Dmel_CG12058, dLARP7', description='La related protein 7', other_designations='CG42569-PA|Larp7-PA', map_location='8D2-8D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9243817, end_position_on_the_genomic_accession=9245925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  44910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44910, CurrentID=0, Status='live', Symbol='unc-45', Aliases='Dmel_CG2708, CG2708, DmUNC-45, DmUnc-45, Dmel\\CG2708, Tom34, dUNC-45, d l(3)03692, lincRNA.721', description='CG2708 gene product from transcript CG2708-RB', other_designations='CG2708-PA|CG2708-PB|translocase of outer membrane 34|unc-45-PA|unc-45-PB', map_location='84E9-84E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8033301, end_position_on_the_genomic_accession=8037403, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  44913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44913, CurrentID=0, Status='live', Symbol='Acam', Aliases='Dmel_CG17769, ACaM, And, CG17769, Camr97A, Dmel\\CG17769', description='Androcam', other_designations='Acam-PA|Acam-PB|CG17769-PA|CG17769-PB|Calmodulin-related 97A|androcalmodulin|androcam', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25811911, end_position_on_the_genomic_accession=25813134, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44915, CurrentID=0, Status='live', Symbol='DNApol-alpha60', Aliases='Dmel_CG5553, BcDNA:GM13640, CG5553, DNA polalpha, DNAprim, DNAprimase, Dmel\\CG5553, Dmpol alpha, GM13640, Pol, Polalpha 60 kDa, S(faf)240, l(3)77ABe, l(3)j10B2, pol alpha, polalpha', description='DNA polymerase alpha 60kD', other_designations='CG5553-PA|DNA polymerase alpha 58,000 beta subunit|DNA polymerase alpha 60,000 beta subunit|DNA primase|DNA replication primer|DNApol-alpha60-PA|DNAprim', map_location='77B6-77B', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20359833, end_position_on_the_genomic_accession=20361585, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462836, CurrentID=0, Status='live', Symbol='CG43781', Aliases='Dmel_ CG8600, Dmel\\CG43781', description='CG43781 gene product from transcript CG43781-RB', other_designations='CG43781-PA|CG43781-PB|uncharacterized protein', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7333544, end_position_on_the_genomic_accession=7335226, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14462837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462837, CurrentID=0, Status='live', Symbol='CG44195', Aliases='Dmel_ Dmel\\CG44195', description='CG44195 gene product from transcript CG44195-RA', other_designations='CG44195-PA|uncharacterized protein', map_location='66A8-66A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7658409, end_position_on_the_genomic_accession=7695654, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14462838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462838, CurrentID=0, Status='live', Symbol='CG43880', Aliases='Dmel_ Dmel\\CG43880, lincRNA.466', description='CG43880 gene product from transcript CG43880-RA', other_designations='CG43880-PA|uncharacterized protein', map_location='66A21-66A21', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7945713, end_position_on_the_genomic_accession=7946322, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44920, CurrentID=0, Status='live', Symbol='Hsp70Ab', Aliases='Dmel_CG18743, CG18743, DMHSP7A2, Dm-hsp70, Dmel\\CG18743, GRP78, HSP70, HSP70Ab, Hsp 70, Hsp-70, Hsp70, Hsp70-Ab, Hsp70A, Hsp70aB, Hsp70ab, Hsp70b, dHsp70, dhsp70, hs70, hsp 70, hsp-70, hsp70, hsp70 87A, hsp70(87A), hsp70A, hsp70Ab, hsp70a', description='Heat-shock-protein-70Ab', other_designations='87A locus|CG18743-PA|Hsp70Ab-PA|heat shock 70|heat shock protein 70|heat shock protein 70Ab|heat-shock protein 70|protein 70', map_location='87A3-87A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11958535, end_position_on_the_genomic_accession=11960909, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44921, CurrentID=0, Status='live', Symbol='Hsp70Ba', Aliases='Dmel_CG31449, CG31449, Dm-hsp70, Dmel\\CG31449, HSP70, Hsp 70, Hsp-70, Hsp70, Hsp70(87C), Hsp70B, dHsp70, dhsp70, hsp-70, hsp70, hsp70 87C, hsp70 Ba, hsp70B, hsp70Ba, hsp70b', description='Heat-shock-protein-70Ba', other_designations='87C locus|CG31449-PA|Hsp70Ba-PA|heat shock 70|heat shock protein 70BA|heat-shock protein 70|hsp70 87C|protein 70', map_location='87B12-87B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12465453, end_position_on_the_genomic_accession=12467776, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462843, CurrentID=0, Status='live', Symbol='CG43965', Aliases='Dmel_ Dmel\\CG43965', description='CG43965 gene product from transcript CG43965-RA', other_designations='CG43965-PA|uncharacterized protein', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8725520, end_position_on_the_genomic_accession=8725774, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462845, CurrentID=0, Status='live', Symbol='CG43783', Aliases='Dmel_ CG43113, CG5741, Dmel\\CG43783', description='CG43783 gene product from transcript CG43783-RG', other_designations='CG43783-PA|CG43783-PB|CG43783-PE|CG43783-PF|CG43783-PG|uncharacterized protein', map_location='66E4-66E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8932235, end_position_on_the_genomic_accession=8954272, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14462850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462850, CurrentID=0, Status='live', Symbol='CG43993', Aliases='Dmel_ Dmel\\CG43993', description='CG43993 gene product from transcript CG43993-RA', other_designations='CG43993-PA|uncharacterized protein', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12032557, end_position_on_the_genomic_accession=12032999, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462851, CurrentID=0, Status='live', Symbol='CG43638', Aliases='Dmel_ Dmel\\CG43638', description='CG43638 gene product from transcript CG43638-RA', other_designations='CG43638-PA|uncharacterized protein', map_location='68F1-68F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12044579, end_position_on_the_genomic_accession=12045356, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44936, CurrentID=0, Status='live', Symbol='Eip93F', Aliases='Dmel_CG18389, CG18389, Dmel\\CG18389, E93, Eip93, Eip93f, anon-WO02059370.58, eip93f, l(3)93, l(3)ry93, lincRNA.833, psg11', description='Ecdysone-induced protein 93F', other_designations='CG18389-PA|CG18389-PB|CG18389-PC|Eip93F-PA|Eip93F-PB|Eip93F-PC|lethal (3) ry93', map_location='93F5-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21949114, end_position_on_the_genomic_accession=22014539, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14462860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462860, CurrentID=0, Status='live', Symbol='CG43678', Aliases='Dmel_ Dmel\\CG43678', description='CG43678 gene product from transcript CG43678-RA', other_designations='CG43678-PA|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15049179, end_position_on_the_genomic_accession=15049495, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462861, CurrentID=0, Status='live', Symbol='CG43679', Aliases='Dmel_ Dmel\\CG43679', description='CG43679 gene product from transcript CG43679-RA', other_designations='CG43679-PA|CG43679-PB|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15050856, end_position_on_the_genomic_accession=15051425, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462862, CurrentID=0, Status='live', Symbol='CG43680', Aliases='Dmel_ Dmel\\CG43680', description='CG43680 gene product from transcript CG43680-RA', other_designations='CG43680-PA|uncharacterized protein', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15052998, end_position_on_the_genomic_accession=15053351, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462875, CurrentID=0, Status='live', Symbol='CG43954', Aliases='Dmel_ Dmel\\CG43954', description='CG43954 gene product from transcript CG43954-RA', other_designations='CG43954-PA|uncharacterized protein', map_location='73B6-73C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16676406, end_position_on_the_genomic_accession=16711677, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14462878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462878, CurrentID=0, Status='live', Symbol='CG44006', Aliases='Dmel_ CG34252, Dmel\\CG44006', description='CG44006 gene product from transcript CG44006-RA', other_designations='CG44006-PA|uncharacterized protein', map_location='75B6-75B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18068728, end_position_on_the_genomic_accession=18070150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14462879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462879, CurrentID=0, Status='live', Symbol='CG44004', Aliases='Dmel_ CG34252, Dmel\\CG44004', description='CG44004 gene product from transcript CG44004-RB', other_designations='CG44004-PB|CG44004-PC|uncharacterized protein', map_location='75B6-75B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18072362, end_position_on_the_genomic_accession=18073863, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  44964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44964, CurrentID=0, Status='live', Symbol='Actn3', Aliases='Dmel_CG8953, CG8953, Dmel\\CG8953, ORF1, acta-3, dabp', description='alpha actinin 3', other_designations='Actn3-PA|CG8953-PA|alpha-actinin-like|alpha-actinin3|open reading frame 1', map_location='89E5-89E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16975885, end_position_on_the_genomic_accession=16977795, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  44965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44965, CurrentID=0, Status='live', Symbol='Fer2LCH', Aliases='Dmel_CG1469, 2LCH, CG1469, Dmel\\CG1469, Fe2LCH, Fer-H, Fer2LC, FerLCH, LCH, anon-EST:ParkEST264, fer2, fer2lch, l(3)00035, l(3)07016, l(3)j2A3, l(3)neo60, l(3)neo63, l(3)s2083', description='Ferritin 2 light chain homologue', other_designations='CG1469-PA|CG1469-PB|CG1469-PC|CG1469-PD|CG1469-PE|Fer2LCH-PA|Fer2LCH-PB|Fer2LCH-PC|Fer2LCH-PD|Fer2LCH-PE|Ferritin-Heavy|Ferritin-Light|ferritin|ferritin 2 light chain|ferritin 2 light chain homologue|ferritin L|ferritin LCH|ferritin light chain', map_location='99F2-99F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30387804, end_position_on_the_genomic_accession=30390584, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14462886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462886, CurrentID=0, Status='live', Symbol='CG43885', Aliases='Dmel_ Dmel\\CG43885', description='CG43885 gene product from transcript CG43885-RA', other_designations='CG43885-PA|uncharacterized protein', map_location='76B6-76B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19543100, end_position_on_the_genomic_accession=19543622, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462897, CurrentID=0, Status='live', Symbol='CG43931', Aliases='Dmel_ Dmel\\CG43931', description='CG43931 gene product from transcript CG43931-RA', other_designations='CG43931-PA|uncharacterized protein', map_location='77F4-77F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20897557, end_position_on_the_genomic_accession=20898501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  44978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=44978, CurrentID=0, Status='live', Symbol='Tom40', Aliases='Dmel_CG12157, CG12157, Dmel\\CG12157, dtom40, l(1)G0216, mit, ms(1)15, ms15, tom40', description='Translocase of outer membrane 40', other_designations='CG12157-PA|CG12157-PB|CG12157-PC|Tom40-PA|Tom40-PB|Tom40-PC|membrane import protein|translocase of outer membrane 40', map_location='7B7-7B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7736893, end_position_on_the_genomic_accession=7739978, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14462900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462900, CurrentID=0, Status='live', Symbol='CG43938', Aliases='Dmel_ Dmel\\CG43938', description='CG43938 gene product from transcript CG43938-RA', other_designations='CG43938-PA|uncharacterized protein', map_location='78B2-78B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21121653, end_position_on_the_genomic_accession=21123415, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462901, CurrentID=0, Status='live', Symbol='CG43702', Aliases='Dmel_ Dmel\\CG43702', description='CG43702 gene product from transcript CG43702-RA', other_designations='CG43702-PA|CG43702-PB|uncharacterized protein', map_location='78C6-78C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21298675, end_position_on_the_genomic_accession=21299063, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462910, CurrentID=0, Status='live', Symbol='CG43895', Aliases='Dmel_ Dmel\\CG43895', description='CG43895 gene product from transcript CG43895-RA', other_designations='CG43895-PA|uncharacterized protein', map_location='78F2-78F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21813005, end_position_on_the_genomic_accession=21813698, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462915, CurrentID=0, Status='live', Symbol='CG43968', Aliases='Dmel_ Dmel\\CG43968, lincRNA.665', description='CG43968 gene product from transcript CG43968-RA', other_designations='CG43968-PA|uncharacterized protein', map_location='80F9-80F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23784231, end_position_on_the_genomic_accession=23787405, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45012, CurrentID=0, Status='live', Symbol='Vha68-2', Aliases='Dmel_CG3762, A subunit, ATP6V1A, CG3762, Dmel\\CG3762, MENE-R6, RE30552p, Vha68, Vha68.2, l(2)01510, l(2)s4214, vha67-2, vha68-2', description='Vacuolar H[+] ATPase 68 kDa subunit 2', other_designations='CG3762-PA|CG3762-PB|CG3762-PC|CG3762-PD|CG3762-PE|CG3762-PF|V-ATPase 69 kDa subunit 2|Vha68-2-PA|Vha68-2-PB|Vha68-2-PC|Vha68-2-PD|Vha68-2-PE|Vha68-2-PF|vacuolar ATPase|vacuolar H[+] ATPase 68 kDa subunit 2|vacuolar H[+] ATPase subunit 68-2', map_location='34A3-34A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12974322, end_position_on_the_genomic_accession=12978607, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  45017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45017, CurrentID=0, Status='live', Symbol='Lcp65Af', Aliases='Dmel_CG10533, CG10533, DCP7, DMLCP65f, Dmel Dmel\\CG10533, Lcp-f', description='CG10533 gene product from transcript CG10533-RA', other_designations='CG10533-PA|Lcp65Af-PA', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6136215, end_position_on_the_genomic_accession=6136791, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  45018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45018, CurrentID=0, Status='live', Symbol='Lcp65Ae', Aliases='Dmel_CG10529, CG10529, DCP6, DMLCP65e, Dmel Dmel\\CG10529, Lcp-e', description='CG10529 gene product from transcript CG10529-RA', other_designations='CG10529-PA|Lcp65Ae-PA', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6137583, end_position_on_the_genomic_accession=6138118, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  45021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45021, CurrentID=0, Status='live', Symbol='Zfrp8', Aliases='Dmel_CG3260, 137/5, CG3260, Dmel\\CG3260, PDCD2, l(2)k13705, zfrp8', description='Zinc finger protein RP-8', other_designations='CG3260-PA|CG3260-PB|Zfrp8-PA|Zfrp8-PB|programmed cell death 2|zinc finger RP8|zinc finger protein RP-8', map_location='60B11-60B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24133989, end_position_on_the_genomic_accession=24135405, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  45031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45031, CurrentID=0, Status='live', Symbol='Pp4-19C', Aliases='Dmel_CG32505, CG1459, CG1596, CG18339, CG32505, DmPpp4-19C, Dmel\\CG32505, PP1, PP4, PP4 19C, PP4-19C, PP4-19c, PP4C, PP4c, PPP4c, PPX, PpX, Ppp4c, dPP4, pp4', description='Protein phosphatase 19C', other_designations='CG32505-PA|CG32505-PE|CG32505-PG|CG32505-PH|CG32505-PI|Pp4-19C-PA|Pp4-19C-PE|Pp4-19C-PG|Pp4-19C-PH|Pp4-19C-PI|phosphoprotein phosphatase 4|protein phosphatase 1|protein phosphatase 19C|protein phosphatase 4|protein phosphatase X', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20421087, end_position_on_the_genomic_accession=20424002, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45039, CurrentID=0, Status='live', Symbol='Sox100B', Aliases='Dmel_CG15552, CG12098, CG15552, CG18674, CT6009, Dmel\\CG15552, SOX100B, SOXE, Sox100, Sox9, SoxB', description='CG15552 gene product from transcript CG15552-RA', other_designations='CG15552-PA|Sox100B-PA', map_location='100B2-100B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31065253, end_position_on_the_genomic_accession=31083407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  45040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45040, CurrentID=0, Status='live', Symbol='Spn43Ab', Aliases='Dmel_CG1865, CG1865, Dmel\\CG1865, SER2, Ser2, anon-WO0118057.2, spn43Ab', description='Serpin 43Ab', other_designations='CG1865-PB|Spn43Ab-PB|serine protease inhibitor 2|serine protease inhibitor 43Ab|serpin 43Ab', map_location='42F3-42F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7151662, end_position_on_the_genomic_accession=7153642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  45041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45041, CurrentID=0, Status='live', Symbol='Spn43Aa', Aliases='Dmel_CG12172, CG12172, Dmel\\CG12172, SER1, Ser1, Spn43 Aa, Spn43A, anon-WO0118057.1', description='Serpin 43Aa', other_designations='CG12172-PA|Spn43Aa-PA|serine protease inhibitor 1|serine protease inhibitor 43Aa|serpin 43Aa', map_location='42F2-42F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7147692, end_position_on_the_genomic_accession=7149289, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45064, CurrentID=0, Status='live', Symbol='Nsun2', Aliases='Dmel_CG6133, CG6133, Dmel\\CG6133, EG:EG0007.9, dNsun2', description='NOP2-Sun domain family, member 2 ortholog', other_designations='CG6133-PA|Nsun2-PA', map_location='4A5-4A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4111156, end_position_on_the_genomic_accession=4113734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246610, CurrentID=0, Status='live', Symbol='CG30430', Aliases='Dmel_ Dmel\\CG30430, NEST:bs01g11, NEST:bs02g04, NEST:bs04a11, NEST:bs12g08, NEST:bs13e09, NEST:bs14e02, NEST:bs15a10, NEST:bs15d07, NEST:bs17g03, NEST:bs27g03, NEST:bs29b02', description='CG30430 gene product from transcript CG30430-RA', other_designations='CG30430-PA|CG30430-PB|uncharacterized protein', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25129432, end_position_on_the_genomic_accession=25129985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740480, CurrentID=0, Status='live', Symbol='CG34247', Aliases='Dmel_ Dmel\\CG34247', description='CG34247 gene product from transcript CG34247-RA', other_designations='CG34247-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16261185, end_position_on_the_genomic_accession=16261605, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740847, CurrentID=0, Status='live', Symbol='CG17636', Aliases='Dmel_ Dmel\\CG17636, EG:23E12.1', description='CG17636 gene product from transcript CG17636-RA', other_designations='CG17636-PA|CG17636-PB|CG17636-PC|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=124370, end_position_on_the_genomic_accession=126714, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740699 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740699, CurrentID=0, Status='live', Symbol='CG34174', Aliases='Dmel_ Dmel\\CG34174', description='CG34174 gene product from transcript CG34174-RA', other_designations='CG34174-PA|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2214266, end_position_on_the_genomic_accession=2215167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  45233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45233, CurrentID=0, Status='live', Symbol='capt', Aliases='Dmel_CG33979, Acap, BcDNA:LD24380, CAP, CG33979, CG5061, CG5074, Cap, Capt, Dcap, Dmel\\CG33979, acu, cap, capping protein, cyclase-associated protein, l(2)06955, punc', description='capulet', other_designations='Act up|CG33979-PA|CG33979-PB|CG33979-PC|Cyclase-associated protein|adenylyl cyclase-associated protein|capt-PA|capt-PB|capt-PC|cyclase associated protein capulet|punctate', map_location='21F1-21F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1151678, end_position_on_the_genomic_accession=1158262, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  45234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45234, CurrentID=0, Status='live', Symbol='Ggamma30A', Aliases='Dmel_CG3694, CG18511, CG3694, CG43733, Dmel\\CG3694, Dmel_CG43733, G-gamma30A, G[[gamma30A]], G[[gamma]], Galpha30A, Ggamma, Ggamma[[e]], Ggammae, lincRNA.S701', description='CG3694 gene product from transcript CG3694-RC', other_designations='CG3694-PA|CG3694-PB|CG3694-PC|CG3694-PF|CG3694-PG|CG3694-PH|CG3694-PI|CG3694-PJ|G protein gamma30A|Ggamma30A-PA|Ggamma30A-PB|Ggamma30A-PC|Ggamma30A-PF|Ggamma30A-PG|Ggamma30A-PH|Ggamma30A-PI|Ggamma30A-PJ', map_location='30A6-30A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9256563, end_position_on_the_genomic_accession=9302494, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  45247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45247, CurrentID=0, Status='live', Symbol='VhaPPA1-1', Aliases='Dmel_CG7007, CG7007, Dmel\\CG7007, VhaPPA1, vhaPPA1-1', description='Vacuolar H[+] ATPase PPA1 subunit 1', other_designations='CG7007-PA|CG7007-PB|V-ATPase proton pump|VhaPPA1-1-PA|VhaPPA1-1-PB|vacuolar H[+] ATPase PPA1 subunit 1|vacuolar H[+] ATPase subunit PPA1-1', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14902330, end_position_on_the_genomic_accession=14904061, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45248, CurrentID=0, Status='live', Symbol='Nckx30C', Aliases='Dmel_CG18660, CG18660, CG4106, Dmel\\CG18660, NcKX30C, NckX30C, potassium-dependent sodium/calcium', description='CG18660 gene product from transcript CG18660-RF', other_designations='CG18660-PA|CG18660-PB|CG18660-PD|CG18660-PE|CG18660-PF|Nckx30C-PA|Nckx30C-PB|Nckx30C-PD|Nckx30C-PE|Nckx30C-PF', map_location='30C6-30C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9711512, end_position_on_the_genomic_accession=9746495, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  45249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45249, CurrentID=0, Status='live', Symbol='Ef1beta', Aliases='Dmel_CG6341, 154264_at, BcDNA.LD24492, BcDNA:LD24492, CG6341, Dmel\\CG6341, EG:EG0003.7, Elf 1 beta', description='Elongation factor 1 beta', other_designations='CG6341-PB|CG6341-PC|Ef1beta-PB|Ef1beta-PC|elongation facto 1 beta|elongation factor 1 beta|elongation factor 1beta', map_location='53D14-53D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16868222, end_position_on_the_genomic_accession=16869280, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45250, CurrentID=0, Status='live', Symbol='deltaCOP', Aliases='Dmel_CG14813, 63B12S, ARCN1, Arch, CG14813, DM63B12S, Delta COP, DeltaCop, Dmel\\CG14813, EG:63B12.10, ESTS:63B12S, anon-WO03040301.246, delta-Cop, deltaCop, l(1)G0051, l(1)G0301, l(1)G0450', description='Coat Protein (coatomer) delta', other_designations='CG14813-PA|CG14813-PB|COPI delta subunit|archain|coat protein (coatomer) delta|coatomer complex COPI delta-COP subunit|delta-coatomer protein|deltaCOP-PA|deltaCOP-PB', map_location='2B12-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1859141, end_position_on_the_genomic_accession=1861790, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  5740851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740851, CurrentID=0, Status='live', Symbol='CG41284', Aliases='Dmel_ Dmel\\CG41284', description='CG41284 gene product from transcript CG41284-RC', other_designations='CG41284-PC|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=26298455, end_position_on_the_genomic_accession=26305572, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45268, CurrentID=0, Status='live', Symbol='cathD', Aliases='Dmel_CG1548, CATH, CG1548, CT3996, CTSD, CathD, Dmel\\CG1548, T17', description='CG1548 gene product from transcript CG1548-RA', other_designations='CG1548-PA|Cathepsin-D|cathD-PA|cathepsin D', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7822111, end_position_on_the_genomic_accession=7823569, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  45269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45269, CurrentID=0, Status='live', Symbol='Prosbeta5', Aliases='Dmel_CG12323, Beta-5, CG12323, Dmel\\CG12323, LD08717, PSMB5, Pros b5, anon-WO0118547.116, beta-5, beta5, dbeta5', description='Proteasome beta5 subunit', other_designations='20S proteasome|CG12323-PA|CG12323-PB|Prosbeta5-PA|Prosbeta5-PB|beta5|proteasome beta5 subunit', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10819978, end_position_on_the_genomic_accession=10821178, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45278, CurrentID=0, Status='live', Symbol='csw', Aliases='Dmel_CG3954, 19-106, CG3954, CSW, Csw, Csw/Shp2, Dmel\\CG3954, E(sev)1A, EG:BACN25G24.2, SHP-2, Shp-2, Shp2/csw, anon-WO03040301.207, anon-WO03040301.209, anon-WO03040301.219/SHP-2, l(1)2Db, l(1)2Dd, l(1)G0170, l(1)GA114, l(1)csw, csw', description='corkscrew', other_designations='CG3954-PA|CG3954-PB|CG3954-PC|CG3954-PD|csw-PA|csw-PB|csw-PC|csw-PD|enhancer of sevenless 1A', map_location='1-1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2094234, end_position_on_the_genomic_accession=2115701, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  45280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45280, CurrentID=0, Status='live', Symbol='cv-2', Aliases='Dmel_CG15671, BMPER, CG15671, CT35855, Cv 2, Cv-2, Cv2, Dmel\\CG15671, cv 2', description='crossveinless 2', other_designations='CG15671-PA|crossveinless-2|cv-2-PA', map_location='2-96.2 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21355015, end_position_on_the_genomic_accession=21379464, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  45282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45282, CurrentID=0, Status='live', Symbol='dib', Aliases='Dmel_CG12028, 302a1, CG12028, CYP302A1, Cyp, Cyp302a1, Dib, Dmel\\CG12028, l(3)64Ak, l(3)SH14, l(3)dib', description='disembodied', other_designations='CG12028-PA|cytochrome P450|dib-PA', map_location='3-12 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4062292, end_position_on_the_genomic_accession=4064551, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318843, CurrentID=0, Status='live', Symbol='CG43394', Aliases='Dmel_ BcDNA:RE27850, CG31607, CG31900, Dmel\\CG43394, Dmel_CG31607, Dmel_CG31900', description='CG43394 gene product from transcript CG43394-RC', other_designations='CG43394-PA|CG43394-PB|CG43394-PC|CG43394-PD|CG43394-PE|uncharacterized protein', map_location='28E8-28E9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8145758, end_position_on_the_genomic_accession=8158786, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  45285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45285, CurrentID=0, Status='live', Symbol='Dr', Aliases='Dmel_CG1897, 153515_at, 99B, CG1897, Dmel\\CG1897/msh, Msh, Msx, l(3)FA30, ltt, msh, msh-1, msh1, Dr', description='Drop', other_designations='CG1897-PA|Dr-PA|Dr-distal|drop|lottchen|muscle segment homeobox|muscle specific homeobox|muscle-specific homeodomain|muscle-specific homeodomain 1', map_location='3-99.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29556386, end_position_on_the_genomic_accession=29565244, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3354861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354861, CurrentID=0, Status='live', Symbol='MFS17', Aliases='Dmel_CG40263, BEST:LD07241, CG40263, Dmel\\CG40263, bs24e02.y1', description='Major Facilitator Superfamily Transporter 17', other_designations='CG40263-PF|CG40263-PG|CG40263-PH|CG40263-PI|CG40263-PJ|MFS17-PF|MFS17-PG|MFS17-PH|MFS17-PI|MFS17-PJ|major facilitator superfamily transporter 17', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=2494742, end_position_on_the_genomic_accession=2565385, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  3354863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354863, CurrentID=0, Status='live', Symbol='klhl10', Aliases='Dmel_CG12423, CG12423, Dmel\\CG12423, anon-WO0140519.152', description='CG12423 gene product from transcript CG12423-RB', other_designations='CG12423-PA|CG12423-PB|klhl10-PA|klhl10-PB', map_location='', chromosome='Unknown', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  45307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45307, CurrentID=0, Status='live', Symbol='fz', Aliases='Dmel_CG17697, CG17697, CG3646, DFZ1, DFz, DFz1, Dfz, Dfz1, Dm Fz1, Dmel\\CG17697, FZ, Frz, Fz, Fz1, dFz, frz1, fz[[1]], fz', description='frizzled', other_designations='CG17697-PA|CG17697-PC|Frizzled1|frizzled 1|frizzled[[1]]|fz-PA|fz-PC', map_location='3-41.7 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14274343, end_position_on_the_genomic_accession=14368639, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  45318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45318, CurrentID=0, Status='live', Symbol='kn', Aliases='Dmel_CG10197, CG10197, Col, Col/Kn, DmCOE, Dmel\\CG10197, Kn, coe, col, collier', description='knot', other_designations='CG10197-PA|CG10197-PB|CG10197-PC|CG10197-PD|CG10197-PE|Collier-Knot|KNOT|collier|collier/knot|kn-PA|kn-PB|kn-PC|kn-PD|kn-PE|knotted', map_location='2-72.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14772650, end_position_on_the_genomic_accession=14807061, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  45320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45320, CurrentID=0, Status='live', Symbol='trol', Aliases='Dmel_CG33950, BcDNA:GM02481, CG12497, CG33675, CG33950, CG7981, CT23996, Dmel\\CG33950, EG:BACR25B3.1, EG:BACR25B3.10, EG:BACR25B3.11, EG:BACR25B3.2, GC7891, MRE7, Pcan, Trol, Trol-A, Trol-B, Troll, ZW-1, anon-WO0153538.72, l(1)3Ac, l(1)9-96, l(1)G0023, l(1)G0271, l(1)G0374, l(1)VA51, l(1)trol, l(1)zw1, l(1)zwl, pcanl, zw-1, zw1, trol', description='terribly reduced optic lobes', other_designations='CG33950-PAE|CG33950-PAF|CG33950-PAG|CG33950-PAH|CG33950-PAI|CG33950-PAK|CG33950-PAL|CG33950-PAM|CG33950-PAN|CG33950-PAO|CG33950-PAP|CG33950-PAQ|CG33950-PAR|CG33950-PAS|CG33950-PAT|CG33950-PAU|CG33950-PAV|CG33950-PAW|CG33950-PAX|CG33950-PAY|CG33950-PAZ|CG33950-PBA|Trol/perlecan|dPerlecan|lethal (1) G0023|lethal (1) G0271|lethal (1) G0374|mRNA-like ncRNA in embryogenesis 7|perlecan|trol-PAE|trol-PAF|trol-PAG|trol-PAH|trol-PAI|trol-PAK|trol-PAL|trol-PAM|trol-PAN|trol-PAO|trol-PAP|trol-PAQ|trol-PAR|trol-PAS|trol-PAT|trol-PAU|trol-PAV|trol-PAW|trol-PAX|trol-PAY|trol-PAZ|trol-PBA', map_location='1-1.04 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2470460, end_position_on_the_genomic_accession=2545410, orientation='minus', exon_count=55, OMIM='', no_hdr0=''),
  3354890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354890, CurrentID=0, Status='live', Symbol='CG40498', Aliases='Dmel_ Dmel\\CG40498', description='CG40498 gene product from transcript CG40498-RG', other_designations='CG40498-PF|CG40498-PG|CG40498-PI|CG40498-PJ|CG40498-PK|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=924636, end_position_on_the_genomic_accession=951252, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  45325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45325, CurrentID=0, Status='live', Symbol='lin', Aliases='Dmel_CG11770, CG11770, Dmel\\CG11770, Lin', description='lines', other_designations='CG11770-PA|lin-PA', map_location='2-59 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8914524, end_position_on_the_genomic_accession=8918234, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45326, CurrentID=0, Status='live', Symbol='Lsp2', Aliases='Dmel_CG6806, C23, CG11538, CG6806, DmeLSP2, Dmel\\CG6806, LHP, LSP 2, LSP-2, LSP2, Lsp-2, Pt-1', description='Larval serum protein 2', other_designations='CG6806-PA|CG6806-PB|Lsp2-PA|Lsp2-PB|Protein-1|larval serum protein 2|larval serum protein-2|larval-hemolymph-protein', map_location='3-37 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12129393, end_position_on_the_genomic_accession=12131869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3354896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354896, CurrentID=0, Status='live', Symbol='vtd', Aliases='Dmel_CG17436, 80Fh, CG17436, CG40222, DRAD21, DRad21, DmRAD21, Dmel\\CG17436, Drad21, RAD21, Rad21, Rad21/Scc1, Scc1, VTD, dRad21, drad21, l(3)80Fh, l(3)Lh3, l3, lethal 3, rad21, scc1', description='verthandi', other_designations='CG17436-PA|CG17436-PB|Rad21|alpha kleisin|alpha-kleisin|cohesin|lethal 3|lethal(3)80Fh|vtd-PA|vtd-PB', map_location='3-46 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=27136525, end_position_on_the_genomic_accession=27157999, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  45329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45329, CurrentID=0, Status='live', Symbol='RpL12', Aliases='Dmel_CG3195, CG3195, Dmel\\CG3195, L12, Q9W1B9_DROME, RPL12, Rp L12, Rpl12, anon-EST:Posey236', description='Ribosomal protein L12', other_designations='CG3195-PA|CG3195-PB|CG3195-PC|RpL12-PA|RpL12-PB|RpL12-PC|ribosomal protein L12', map_location='60B7-60B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24060378, end_position_on_the_genomic_accession=24061860, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  45338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45338, CurrentID=0, Status='live', Symbol='na', Aliases='Dmel_CG1517, CG1517, Dma1U, Dmalpha1U, Dmel\\CG1517, NA, NALCN, Na, alpha1U, har, har 38, har 85, har38, har85, harA', description='narrow abdomen', other_designations='CG1517-PF|CG1517-PG|NARROW ABDOMEN|halothane-resistant|na-PF|na-PG', map_location='1-46.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14266367, end_position_on_the_genomic_accession=14277416, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  45339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45339, CurrentID=0, Status='live', Symbol='net', Aliases='Dmel_CG11450, CG11450, Dmel\\CG11450, Group IId, Shout, bHLHc42, shout', description='CG11450 gene product from transcript CG11450-RB', other_designations='CG11450-PB|CG11450-PC|net-PB|net-PC', map_location='2-0.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=82421, end_position_on_the_genomic_accession=87387, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  45343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45343, CurrentID=0, Status='live', Symbol='pk', Aliases='Dmel_CG11084, CG11084, CG12830, Dmel\\CG11084, Pk, Pk-Sple, Sple-sple, pkM, sple, pk', description='prickle', other_designations='CG11084-PA|CG11084-PB|CG11084-PC|pk-PA|pk-PB|pk-PC|prickle-spiny legs|prickle-spiny-legs|spiny legs', map_location='2-55.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7150703, end_position_on_the_genomic_accession=7223378, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  5740854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740854, CurrentID=0, Status='live', Symbol='CG34409', Aliases='Dmel_ CG12949, CG31411, CG31412, Dmel\\CG34409, Dmel_CG31411, Dmel_CG31412, SP12, c-SP12', description='CG34409 gene product from transcript CG34409-RB', other_designations='CG34409-PB|uncharacterized protein', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9742877, end_position_on_the_genomic_accession=9746367, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3354918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354918, CurrentID=0, Status='live', Symbol='RpL15', Aliases='Dmel_CG17420, 80Fi, CG17420, CG40199, DmRpL15, Dmel\\CG17420, L15, RPL15, l(3)80Fi, l(3)Lh2, l2, lethal 2', description='Ribosomal protein L15', other_designations='CG17420-PA|CG17420-PB|CG17420-PC|CG17420-PE|CG17420-PF|CG17420-PG|CG17420-PH|CG17420-PI|RpL15-PA|RpL15-PB|RpL15-PC|RpL15-PE|RpL15-PF|RpL15-PG|RpL15-PH|RpL15-PI|lethal (3) 80Fi|lethal 2|ribosomal L15|ribosomal protein L15', map_location='80F-80F', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=27267255, end_position_on_the_genomic_accession=27268445, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3354919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354919, CurrentID=0, Status='live', Symbol='Dbp80', Aliases='Dmel_CG17023, BcDNA:RE44177, CG17023, DBP80, DmDbp80, Dmel\\CG17023, HEL80, Hel40, Hel80, hel-40, mdcds_36124', description='Dead box protein 80', other_designations='CG17023-PB|CG17023-PD|CG17023-PE|CG17023-PF|CG17023-PG|Dbp80-PB|Dbp80-PD|Dbp80-PE|Dbp80-PF|Dbp80-PG|dead box protein 80|helicase 40|helicase 80', map_location='80F-80F', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=27279316, end_position_on_the_genomic_accession=27422159, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  3354921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354921, CurrentID=0, Status='live', Symbol='Gfat1', Aliases='Dmel_CG12449, CG12449, CG40197, Dmel/Gfat1, Dmel\\CG12449, GFAT, Gfat, gfat, gfat1', description='Glutamine:fructose-6-phosphate aminotransferase 1', other_designations='CG12449-PA|CG12449-PD|CG12449-PF|CG12449-PH|CG12449-PJ|CG12449-PK|CG12449-PL|CG12449-PM|CG12449-PN|Gfat1-PA|Gfat1-PD|Gfat1-PF|Gfat1-PH|Gfat1-PJ|Gfat1-PK|Gfat1-PL|Gfat1-PM|Gfat1-PN', map_location='81F-81F', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4126442, end_position_on_the_genomic_accession=4137882, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3354922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354922, CurrentID=0, Status='live', Symbol='CG40198', Aliases='Dmel_ BcDNA:RE21703, Dmel\\CG40198', description='CG40198 gene product from transcript CG40198-RC', other_designations='CG40198-PB|CG40198-PC|uncharacterized protein', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4160610, end_position_on_the_genomic_accession=4162235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3354925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354925, CurrentID=0, Status='live', Symbol='Maf1', Aliases='Dmel_CG40196, CG40196, CR41604, Dmel\\CG40196, d', description='CG40196 gene product from transcript CG40196-RC', other_designations='CG40196-PA|CG40196-PB|CG40196-PC|CG40196-PD|CG40196-PE|Maf1-PA|Maf1-PB|Maf1-PC|Maf1-PD|Maf1-PE', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1512810, end_position_on_the_genomic_accession=1527095, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  45360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45360, CurrentID=0, Status='live', Symbol='shu', Aliases='Dmel_CG4735, CG4735, Dmel\\CG4735, Shu', description='shutdown', other_designations='CG4735-PA|shu-PA|shut down', map_location='2-105 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23841461, end_position_on_the_genomic_accession=23843286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3354930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354930, CurrentID=0, Status='live', Symbol='CG40191', Aliases='Dmel_ BEST:HL02313, BcDNA:LD12604, Dmel\\CG40191', description='CG40191 gene product from transcript CG40191-RC', other_designations='CG40191-PA|CG40191-PB|CG40191-PC|CG40191-PD|CG40191-PE|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1264140, end_position_on_the_genomic_accession=1267747, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  45368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45368, CurrentID=0, Status='live', Symbol='Treh', Aliases='Dmel_CG9364, 87D3T, CG9364, Dmel\\CG9364, ESTS:87D3T, TRE, TREH, Tre, anon-EST:Posey159, anon-sts38', description='Trehalase', other_designations='CG9364-PA|CG9364-PB|CG9364-PC|CG9364-PD|CG9364-PE|CG9364-PF|CG9364-PG|Treh-PA|Treh-PB|Treh-PC|Treh-PD|Treh-PE|Treh-PF|Treh-PG|trehalase', map_location='57B16-57B19', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21074018, end_position_on_the_genomic_accession=21088945, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  3354941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354941, CurrentID=0, Status='live', Symbol='CG40178', Aliases='Dmel_CG40178/CG41283, Dmel\\CG40178', description='CG40178 gene product from transcript CG40178-RA', other_designations='CG40178-PA|CG40178-PB|CG40178-PC|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=25798441, end_position_on_the_genomic_accession=26014112, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  45374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45374, CurrentID=0, Status='live', Symbol='fwd', Aliases='Dmel_CG7004, CG7004, Dmel\\CG7004, Fwd, PI4K, anon-WO0118547.259', description='four wheel drive', other_designations='CG7004-PA|CG7004-PB|CG7004-PC|CG7004-PD|Four-wheel drive|fwd-PA|fwd-PB|fwd-PC|fwd-PD', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=305393, end_position_on_the_genomic_accession=319884, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  318859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318859, CurrentID=0, Status='live', Symbol='CG31626', Aliases='Dmel_ BcDNA:RE20756, Dmel\\CG31626', description='CG31626 gene product from transcript CG31626-RA', other_designations='CG31626-PA|CG31626-PB|uncharacterized protein', map_location='39B3-39B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21232495, end_position_on_the_genomic_accession=21236733, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45380, CurrentID=0, Status='live', Symbol='spin', Aliases='Dmel_CG8428, 25/11, CG8428, Dmel\\CG8428, Spin, bnch, l(2)10403, l(2)W5, l(2)k02511, l(2)k09905, l(2R)W5', description='spinster', other_designations='CG8428-PA|CG8428-PB|CG8428-PC|CG8428-PD|CG8428-PE|Spinster/Benchwarmer|benchwarmer|diphthong|lethal (2) W5|spin-PA|spin-PB|spin-PC|spin-PD|spin-PE', map_location='52E6-52E', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16124981, end_position_on_the_genomic_accession=16138238, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  45382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45382, CurrentID=0, Status='live', Symbol='ogre', Aliases='Dmel_CG3039, CB8, CG3039, Dm-inx1, Dmel\\CG3039, FBpp0070995, Inx1, Ogre, cb-8, inx-1, inx1, jnL3, l(1)6Ec, l(1)jnL3, l(1)ogre', description='optic ganglion reduced', other_designations='CG3039-PA|CG3039-PB|CG3039-PC|CG3039-PD|CG3039-PE|CG3039-PF|contrast-blind-8|innexin 1|innexin-1|ogre-PA|ogre-PB|ogre-PC|ogre-PD|ogre-PE|ogre-PF|optic ganglia-reduced', map_location='1-18.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6973620, end_position_on_the_genomic_accession=6991318, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  45396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45396, CurrentID=0, Status='live', Symbol='E5', Aliases='Dmel_CG9930, CG9930, Dmel\\CG9930', description='CG9930 gene product from transcript CG9930-RA', other_designations='CG9930-PA|E5-PA', map_location='88A1-88A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13867345, end_position_on_the_genomic_accession=13874916, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3354965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354965, CurrentID=0, Status='live', Symbol='CG40160', Aliases='Dmel_ BcDNA:LD13269, Dmel\\CG40160', description='CG40160 gene product from transcript CG40160-RH', other_designations='CG40160-PA|CG40160-PD|CG40160-PE|CG40160-PF|CG40160-PG|CG40160-PH|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=26215815, end_position_on_the_genomic_accession=26229433, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45398, CurrentID=0, Status='live', Symbol='Aldh-III', Aliases='Dmel_CG11140, ALDH, ALDH-III, CG11140, CT41571, Dhap, Dmel\\CG11140, FALDH, l(2)03610', description='Aldehyde dehydrogenase type III', other_designations='Aldh-III-PE|Aldh-III-PI|Aldh-III-PK|Aldh-III-PL|Aldh-III-PM|Aldh-III-PN|Aldh-III-PO|Aldh-III-PP|Aldh-III-PQ|Aldh-III-PR|Aldh-III-PS|Aldh-III-PU|CG11140-PE|CG11140-PI|CG11140-PK|CG11140-PL|CG11140-PM|CG11140-PN|CG11140-PO|CG11140-PP|CG11140-PQ|CG11140-PR|CG11140-PS|CG11140-PU|aldehyde dehydrogenase type III|fatty aldehyde dehydrogenase', map_location='43C4-43C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7466879, end_position_on_the_genomic_accession=7477162, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  45399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45399, CurrentID=0, Status='live', Symbol='Aats-trp', Aliases='Dmel_CG9735, 85D-WRS, CG9735, Dmel\\CG9735, TrpRS, WRS, WRS-85D, l(3)03559, l(3)03560, l(3)3559, wars', description='Tryptophanyl-tRNA synthetase', other_designations='Aats-trp-PB|CG9735-PB', map_location='85D7-85D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9265027, end_position_on_the_genomic_accession=9266904, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  45401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45401, CurrentID=0, Status='live', Symbol='ox', Aliases='Dmel_CG8764, BcDNA:RH44662, CG8764, Dmel\\CG8764, QCR9, Qcr10, Qcr9, UCRX, UQCR-7.2, ms(2)00815', description='oxen', other_designations='CG8764-PA|Ubiquinol-cytochrome C reductase|Ubiquinol-cytochrome C reductase complex 7.2 kDa protein|Ubiquinol-cytochrome c oxidoreductase 7.2 kDa|Ubiquinol-cytochrome c reductase 7.2 kDa protein|ox-PA', map_location='49C2-49C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12757133, end_position_on_the_genomic_accession=12757529, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3354971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354971, CurrentID=0, Status='live', Symbol='Set1', Aliases='Dmel_CG40351, CG17395, CG17396, CG40351, Dmel\\CG40351, SET1, aaf45425 Dm, dSet1, dset1, l(3)h5', description='SET domain containing 1', other_designations='CG40351-PA|CG40351-PB|CG40351-PC|CG40351-PD|CG40351-PE|CG40351-PF|CG40351-PG|CG40351-PH|CG40351-PI|Set1-PA|Set1-PB|Set1-PC|Set1-PD|Set1-PE|Set1-PF|Set1-PG|Set1-PH|Set1-PI|lethal 5', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=25107594, end_position_on_the_genomic_accession=25115180, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3354977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354977, CurrentID=0, Status='live', Symbol='MED21', Aliases='Dmel_CG17397, CG17397, CG40364, Dmel\\CG17397, Med21, Trap19, dSRB7, dTRAP19, p18', description='Mediator complex subunit 21', other_designations='CG17397-PA|CG17397-PB|MED21-PA|MED21-PB|mediator|mediator complex subunit 21', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=25114924, end_position_on_the_genomic_accession=25115946, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3354987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354987, CurrentID=0, Status='live', Symbol='tyn', Aliases='Dmel_CG17131, CG17131, CG17131-ORFA, CG17131-ORFB, Dmel\\CG17131, SP71', description='trynity', other_designations='CG17131-PA|CG17131-PB|CG17131-PC|CG17131-PD|CG17131-PE|CG17131-PF|SP71|tyn-PA|tyn-PB|tyn-PC|tyn-PD|tyn-PE|tyn-PF', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=142208, end_position_on_the_genomic_accession=200663, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  3354988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354988, CurrentID=0, Status='live', Symbol='CG17163', Aliases='Dmel_ Dmel\\CG17163', description='CG17163 gene product from transcript CG17163-RB', other_designations='CG17163-PA|CG17163-PB|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23080759, end_position_on_the_genomic_accession=23082709, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3354990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354990, CurrentID=0, Status='live', Symbol='CG12061', Aliases='Dmel_ CG40148, Dmel\\CG12061', description='CG12061 gene product from transcript CG12061-RD', other_designations='CG12061-PD|CG12061-PF|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23036066, end_position_on_the_genomic_accession=23047678, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3354991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354991, CurrentID=0, Status='live', Symbol='CG17167', Aliases='Dmel_ CG40146, Dmel\\CG17167', description='CG17167 gene product from transcript CG17167-RE', other_designations='CG17167-PC|CG17167-PD|CG17167-PE|CG17167-PF|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23068301, end_position_on_the_genomic_accession=23081144, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3354992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354992, CurrentID=0, Status='live', Symbol='zyd', Aliases='Dmel_CG2893, BEST:CK02262, CG2893, CG40147, CK02262, Dmel\\CG2893', description='zydeco', other_designations='CG2893-PA|CG2893-PD|CG2893-PE|CG2893-PF|CG2893-PG|CG2893-PH|CG2893-PI|zyd-PA|zyd-PD|zyd-PE|zyd-PF|zyd-PG|zyd-PH|zyd-PI', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23022898, end_position_on_the_genomic_accession=23035550, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3354994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354994, CurrentID=0, Status='live', Symbol='su(f)', Aliases='Dmel_CG17170, 19, 28, 29, 3-DES, 3DES, CG17170, CstF-77, CstF77, Cstf77, Dmel\\CG17170, EC235, R-9-18, R9-18, SU-f, Su(f), l(1)20Eb, l(1)3DES, l(1)EC235, l(1)G0393, l(1)R-9-18, l(1)X1, l(1)X3, l(1)su(f), l3DES, su-f, su[+]-f', description='suppressor of forked', other_designations='CG17170-PE|CG17170-PF|CG17170-PG|lethal3Des|section 19|section 28|section 29|su(f)-PE|su(f)-PF|su(f)-PG', map_location='1-65.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23096771, end_position_on_the_genomic_accession=23101431, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  318867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318867, CurrentID=0, Status='live', Symbol='CG31643', Aliases='Dmel_CG31643, BcDNA:LD32258PA, Dmel\\CG31643', description='CG31643 gene product from transcript CG31643-RA', other_designations='CG31643-PA|CG31643-PB|uncharacterized protein', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6038117, end_position_on_the_genomic_accession=6041283, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3354996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3354996, CurrentID=0, Status='live', Symbol='CG17168', Aliases='Dmel_ Dmel\\CG17168', description='CG17168 gene product from transcript CG17168-RA', other_designations='CG17168-PA|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23082738, end_position_on_the_genomic_accession=23084576, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45432, CurrentID=0, Status='live', Symbol='can', Aliases='Dmel_CG6577, CG6577, Can, Dmel\\CG6577, TAF5L, Taf5L, dTAF5L, dTAF[[II]]80, dmTAF5b', description='cannonball', other_designations='CG6577-PA|can-PA', map_location='3-31 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10263909, end_position_on_the_genomic_accession=10267111, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3355010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355010, CurrentID=0, Status='live', Symbol='IntS3', Aliases='Dmel_CG17665, CG17665, Dmel\\CG17665, IR23, Int3, NC19, NC28, l(2)IR23, l(2)NC19, l(2)NC28, l(2Rh)IR23', description='Integrator 3', other_designations='CG17665-PA|IntS3-PA|integrator 3|lethal (2) IR23|lethal (2) NC19|lethal (2) NC28', map_location='h44-h46', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4235779, end_position_on_the_genomic_accession=4242934, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  3355011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355011, CurrentID=0, Status='live', Symbol='CG17683', Aliases='Dmel_ Dmel\\CG17683, NAR1p, NC37, NC38, NC70, l(2)NC37, l(2)NC38, l(2)NC70', description='CG17683 gene product from transcript CG17683-RD', other_designations='CG17683-PA|CG17683-PC|CG17683-PD|CG17683-PE|lethal (2) NC37|lethal (2) NC38|lethal (2) NC70|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4130974, end_position_on_the_genomic_accession=4133484, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3355013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355013, CurrentID=0, Status='live', Symbol='Gprk1', Aliases='Dmel_CG40129, CG40129, CT4320, Dmel\\CG40129, GPRK1, Gprk-1, RK, Rok, gprk1', description='G protein-coupled receptor kinase 1', other_designations='CG40129-PA|CG40129-PB|CG40129-PC|Gprk1-PA|Gprk1-PB|Gprk1-PC|rhodopsin kinase', map_location='41B3-41C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4322295, end_position_on_the_genomic_accession=4471060, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  3355016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355016, CurrentID=0, Status='live', Symbol='RNASEK', Aliases='Dmel_CG40127, BcDNA:GM16138, CG40127, Dmel\\CG40127, GM16138, NC110, anon-EST:Posey156, dRNASEK, l(2)NC110', description='Ribonuclease kappa', other_designations='CG40127-PA|CG40127-PB|RNASEK-PA|RNASEK-PB|lethal (2) NC110|ribonuclease kappa', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4186835, end_position_on_the_genomic_accession=4188127, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45460, CurrentID=0, Status='live', Symbol='Pex1', Aliases='Dmel_CG6760, 0848/07, CG6760, DmCAB51031.1, DmelPex1, Dmel\\CG6760, PEX1, anon-WO0172774.100, anon-WO0172774.101, anon-WO0172774.102, anon-WO0172774.96, anon-WO0172774.98, anon-WO0172774.99, l(3)02402, l(3)084807, l(3)70Da, l(3)S084807, l(3)s4868, pex1', description='Peroxin 1', other_designations='CG6760-PA|Pex1-PA|lethal (3) 70Da|peroxin 1', map_location='70D1-70D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14071892, end_position_on_the_genomic_accession=14076123, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  45467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45467, CurrentID=0, Status='live', Symbol='mtd', Aliases='Dmel_CG32464, CG10196, CG10199, CG32464, Dmel\\CG32464, L(3)82Fd, L82, anon-EST:Posey263, l(3)82FD, l(3)82Fd', description='mustard', other_designations='CG32464-PA|CG32464-PAA|CG32464-PB|CG32464-PC|CG32464-PF|CG32464-PG|CG32464-PH|CG32464-PI|CG32464-PJ|CG32464-PK|CG32464-PM|CG32464-PN|CG32464-PO|CG32464-PP|CG32464-PQ|CG32464-PR|CG32464-PS|CG32464-PT|CG32464-PU|CG32464-PV|CG32464-PW|CG32464-PX|CG32464-PY|CG32464-PZ|l(3)82Fd|late puff gene at 82F|mtd-PA|mtd-PAA|mtd-PB|mtd-PC|mtd-PF|mtd-PG|mtd-PH|mtd-PI|mtd-PJ|mtd-PK|mtd-PM|mtd-PN|mtd-PO|mtd-PP|mtd-PQ|mtd-PR|mtd-PS|mtd-PT|mtd-PU|mtd-PV|mtd-PW|mtd-PX|mtd-PY|mtd-PZ', map_location='82F8-82F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5270116, end_position_on_the_genomic_accession=5350443, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  3355040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355040, CurrentID=0, Status='live', Symbol='CG17514', Aliases='Dmel_ Dmel\\CG17514', description='CG17514 gene product from transcript CG17514-RA', other_designations='CG17514-PA|CG17514-PG|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=27967244, end_position_on_the_genomic_accession=28012599, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3355041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355041, CurrentID=0, Status='live', Symbol='eIF-4B', Aliases='Dmel_CG10837, CG10837, Dm-eIF4B, Dmel\\CG10837, EIF-4B, Eif4B-L, eIF-4B-S, eIF-4b, eIF4B, eIF4B-L, eIF4B-S, eIf4b, eIF-4B', description='Eukaryotic initiation factor 4B', other_designations='CG10837-PA|CG10837-PB|CG10837-PE|CG10837-PF|eIF-4B-PA|eIF-4B-PB|eIF-4B-PE|eIF-4B-PF|eukaryotic initiation factor 4B', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=27842669, end_position_on_the_genomic_accession=27858271, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3355064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355064, CurrentID=0, Status='live', Symbol='Haspin', Aliases='Dmel_CG40080, BcDNA:LD07633, CG40080, Dmel\\CG40080, haspin', description='CG40080 gene product from transcript CG40080-RA', other_designations='CG40080-PA|Haspin-PA|haspin', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4014111, end_position_on_the_genomic_accession=4049342, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3355066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355066, CurrentID=0, Status='live', Symbol='CG17684', Aliases='Dmel_ CG40072, CG40260, Dmel\\CG17684', description='CG17684 gene product from transcript CG17684-RC', other_designations='CG17684-PC|CG17684-PD|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1866080, end_position_on_the_genomic_accession=2262115, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  3355069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355069, CurrentID=0, Status='live', Symbol='CG17691', Aliases='Dmel_ Dmel\\CG17691', description='CG17691 gene product from transcript CG17691-RC', other_designations='CG17691-PC|CG17691-PE|CG17691-PF|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1590484, end_position_on_the_genomic_accession=1596457, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3355072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355072, CurrentID=0, Status='live', Symbol='CG41099', Aliases='Dmel_ BEST:GM10035, CG17416, CG17419, Dmel\\CG41099, GM10035', description='CG41099 gene product from transcript CG41099-RD', other_designations='CG41099-PB|CG41099-PC|CG41099-PD|CG41099-PE|D-Rabankyrin|Rabankyrin-5|uncharacterized protein', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=3846729, end_position_on_the_genomic_accession=3868401, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3355079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355079, CurrentID=0, Status='live', Symbol='CG40045', Aliases='Dmel_ BcDNA:RE63412, Dmel\\CG40045', description='CG40045 gene product from transcript CG40045-RB', other_designations='CG40045-PA|CG40045-PB|CG40045-PC|CG40045-PD|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24028515, end_position_on_the_genomic_accession=24035859, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3355084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355084, CurrentID=0, Status='live', Symbol='Snap25', Aliases='Dmel_CG40452, CG17676, CG17884, CG40452, Dmel\\CG40452, SNAP-25, SNAP25, snap-25, snp25', description='Synaptosomal-associated protein 25kDa', other_designations='CG40452-PA|CG40452-PC|Snap25-PA|Snap25-PC|synapse protein 25|synaptosomal-associated protein 25', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24074180, end_position_on_the_genomic_accession=24299205, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3355089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355089, CurrentID=0, Status='live', Symbol='mRpS5', Aliases='Dmel_CG40049, BcDNA:RE47909, CG40049, Dmel\\CG40049, MRP-S5, MrpS5', description='mitochondrial ribosomal protein S5', other_designations='CG40049-PA|CG40049-PC|CG40049-PD|mRpS5-PA|mRpS5-PC|mRpS5-PD', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24537236, end_position_on_the_genomic_accession=24543534, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45524, CurrentID=0, Status='live', Symbol='Cyp4p1', Aliases='Dmel_CG10842, 4p1, CG10842, Dmel\\CG10842, P-450, P450, anon-WO0140519.35', description='Cytochrome P450-4p1', other_designations='CG10842-PA|Cyp4p1-PA|cytochrome P450|cytochrome P450-4p1', map_location='45B7-45B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9240041, end_position_on_the_genomic_accession=9242139, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  45525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45525, CurrentID=0, Status='live', Symbol='eIF-4E', Aliases='Dmel_CG4035, 0260/09, 0587/11, 0589/11, 0919/12, 1004/13, 2, 7238, CBP, CG4035, CT13384, CT39424, CT39426, D-eIF4E, Dmel\\CG4035, EIF4E, Eif4E, Eif4e, IF4E, dEif4e, deIF-4E, deIF4E, eIF 4E1, eIF-4E2, eIF-4EII, eIF-4e, eIF4E, eIF4E-1, eIF4E-1/2, eIF4E-2, eIF4E1, eIF4EI, eIF4e, eif-4E, eif-4e, eif4e, elF4E, l(3)07238, l(3)67Af, l(3)S025007, l(3)S026009, l(3)S058711, l(3)S091912, eIF-4E', description='Eukaryotic initiation factor 4E', other_designations='CG4035-PA|CG4035-PB|CG4035-PC|CG4035-PD|CG4035-PE|CG4035-PF|CG4035-PG|CG4035-PH|CG4035-PI|cap binding protein|eIF-4E-PA|eIF-4E-PB|eIF-4E-PC|eIF-4E-PD|eIF-4E-PE|eIF-4E-PF|eIF-4E-PG|eIF-4E-PH|eIF-4E-PI|eukaryotic initiation factor 4E|eukaryotic translation initiation factor 4E|lethal(3)67Af', map_location='67B4-67B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9399618, end_position_on_the_genomic_accession=9402460, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3355094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355094, CurrentID=0, Status='live', Symbol='CG12567', Aliases='Dmel_ Dmel\\CG12567', description='CG12567 gene product from transcript CG12567-RF', other_designations='CG12567-PA|CG12567-PC|CG12567-PD|CG12567-PE|CG12567-PF|CG12567-PG|uncharacterized protein', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=23161122, end_position_on_the_genomic_accession=23174484, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  5740860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740860, CurrentID=0, Status='live', Symbol='CG34251', Aliases='Dmel_ Dmel\\CG34251', description='CG34251 gene product from transcript CG34251-RA', other_designations='CG34251-PA|CG34251-PB|uncharacterized protein', map_location='74E5-74F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17671622, end_position_on_the_genomic_accession=17683581, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3355096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355096, CurrentID=0, Status='live', Symbol='Tim23', Aliases='Dmel_CG40042, BEST:GH02609, BcDNA:SD19278, CG40042, Dmel\\CG40042, LP07554', description='Translocase of inner mitochondrial membrane 23', other_designations='CG40042-PA|Tim23-PA|translocase of inner mitochondrial membrane 23', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=23312465, end_position_on_the_genomic_accession=23313578, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3355097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355097, CurrentID=0, Status='live', Symbol='Gpb5', Aliases='Dmel_CG40041, CG40041, Dmbeta, Dmel\\CG40041, GPB5, beta-glycoprotein-hormone-related, gpb5', description='Glycoprotein hormone beta 5 ortholog (H. sapiens)', other_designations='CG40041-PA|CG40041-PC|CG40041-PD|Gpb5-PA|Gpb5-PC|Gpb5-PD|beta-glycoprotein-hormone-related|beta-glycoprotein-hormone-related-polypeptide|glycoprotein hormone beta 5|glycoprotein hormone beta 5 ortholog', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=23311700, end_position_on_the_genomic_accession=23312464, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355102, CurrentID=0, Status='live', Symbol='Scp1', Aliases='Dmel_CG15848, BcDNA:GH15296, CG15848, Dcabp-A.1, Dmel\\CG15848, SCP, SCP1', description='Sarcoplasmic calcium-binding protein 1', other_designations='CG15848-PA|CG15848-PB|Scp1-PA|Scp1-PB|calcium biding protein|drosophila Ca[2+]-binding protein A.1|sarcoplasmic calcium binding protein 1|sarcoplasmic calcium-binding protein 1', map_location='80A-80B', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=2932595, end_position_on_the_genomic_accession=2968052, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  3355106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355106, CurrentID=0, Status='live', Symbol='Alg-2', Aliases='Dmel_CG40410, ALG-2, CG17661, CG40410, DALG-2, Dmel\\CG40410', description='Apoptosis-linked gene-2', other_designations='Alg-2-PC|CG40410-PC', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=3623143, end_position_on_the_genomic_accession=3627690, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3355107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355107, CurrentID=0, Status='live', Symbol='Tim17b', Aliases='Dmel_CG40451, CG40451, Dmel\\CG40451', description='CG40451 gene product from transcript CG40451-RA', other_designations='CG40451-PA|CG40451-PB|Tim17b-PA|Tim17b-PB', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=3734855, end_position_on_the_genomic_accession=3745306, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3355108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355108, CurrentID=0, Status='live', Symbol='CG41128', Aliases='Dmel_ BcDNA:RH61753, Dmel\\CG41128, Minos1b, RH61753p', description='CG41128 gene product from transcript CG41128-RB', other_designations='CG41128-PB|mitochondrial INner membrane organizing system 1|uncharacterized protein', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=3750363, end_position_on_the_genomic_accession=3752353, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355109, CurrentID=0, Status='live', Symbol='Parp', Aliases='Dmel_CG40411, BEST:LD21673, CG17685, CG17696, CG17718, CG40411, D.PARP, Dm.pARTa, Dmel\\CG40411, LD21673.3prime, PARP, PARP-1, PARP11, dPARP, dmparp, parp, Parp', description='Poly-(ADP-ribose) polymerase', other_designations='(Poly (ADP-ribose) polymeras|CG40411-PB|Parp-PB|Poly(ADP ribose) polymerase 1|Poly(ADP)-Ribose polymerase|Poly(ADP-)Ribose polymerase|Poly(ADP-ribose) polymerase|Poly(ADP-ribose) polymerase 1|poly (ADP-ribose) polymerase|poly ADP ribose polymerase', map_location='81F-81F', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=3461351, end_position_on_the_genomic_accession=3587054, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3355111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355111, CurrentID=0, Status='live', Symbol='FASN3', Aliases='Dmel_CG17374, CG17374, DM_7289423, Dmel\\CG17374, FAS, FASN[CG17374], FAS[CG17374]', description='Fatty acid synthase 3', other_designations='CG17374-PC|FASN3-PC|fatty acid synthase 3', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=25752859, end_position_on_the_genomic_accession=25762779, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  3355114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355114, CurrentID=0, Status='live', Symbol='Spf45', Aliases='Dmel_CG17540, CG17540, Dmel\\CG17540, RBM17, SPF45, dRBM17, dSPF45, spf-45, spf45', description='CG17540 gene product from transcript CG17540-RB', other_designations='CG17540-PA|CG17540-PB|CG17540-PC|RNA binding motif protein 1|Spf45-PA|Spf45-PB|Spf45-PC', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=23066874, end_position_on_the_genomic_accession=23068386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3355116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355116, CurrentID=0, Status='live', Symbol='Cht3', Aliases='Dmel_CG18140, BcDNA:LP04696, CG17682, CG18140, CG18140-PA3, CHT1, CHT3, Cht1, Cht10, DmCHT1, DmCHT10, DmCHT3, Dmel\\CG18140', description='Chitinase 3', other_designations='CG18140-PA|CG18140-PB|CG18140-PC|Cht3-PA|Cht3-PB|Cht3-PC|chitinase|chitinase 1|chitinase 3|chitinase-1|chitinase-3', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22999574, end_position_on_the_genomic_accession=23025915, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  3355120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355120, CurrentID=0, Status='live', Symbol='CG40439', Aliases='Dmel_ Dmel\\CG40439', description='CG40439 gene product from transcript CG40439-RB', other_designations='CG40439-PA|CG40439-PB|uncharacterized protein', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22797519, end_position_on_the_genomic_accession=22798276, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3355123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355123, CurrentID=0, Status='live', Symbol='CG40006', Aliases='Dmel_ BcDNA:RE68569, Dmel\\CG40006', description='CG40006 gene product from transcript CG40006-RA', other_designations='CG40006-PA|uncharacterized protein', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22622417, end_position_on_the_genomic_accession=22756349, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  45556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45556, CurrentID=0, Status='live', Symbol='Cyp6a17', Aliases='Dmel_CG10241, 61D6S, 6a17, CG10241, Cyp6A17, Dmel\\CG10241, ESTS:61D6S, cyp6a17', description='Cytochrome P450-6a17', other_designations='CG10241-PA|CG10241-PB|Cyp6a17-PA|Cyp6a17-PB|cytochrome P450|cytochrome P450-6a17', map_location='51D1-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14873953, end_position_on_the_genomic_accession=14875739, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355125, CurrentID=0, Status='live', Symbol='CG17490', Aliases='Dmel_ Dmel\\CG17490', description='CG17490 gene product from transcript CG17490-RG', other_designations='CG17490-PE|CG17490-PG|uncharacterized protein', map_location='40F7-40F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22537668, end_position_on_the_genomic_accession=22540597, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355126, CurrentID=0, Status='live', Symbol='CG17493', Aliases='Dmel_ CG40016, Dmel\\CG17493', description='CG17493 gene product from transcript CG17493-RB', other_designations='CG17493-PB|CG17493-PC|CG17493-PD|uncharacterized protein', map_location='40F7-40F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22550064, end_position_on_the_genomic_accession=22556295, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355127, CurrentID=0, Status='live', Symbol='Slmap', Aliases='Dmel_CG17494, CG17494, Dmel\\CG17494', description='Sarcolemma associated protein', other_designations='CG17494-PA|CG17494-PD|Slmap-PA|Slmap-PD|sarcolemma associated protein|sarcolemma associated protein ortholog', map_location='40F7-40F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22579072, end_position_on_the_genomic_accession=22585843, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3355129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355129, CurrentID=0, Status='live', Symbol='CG17715', Aliases='Dmel_ BcDNA:RE07178, BcDNA:RE63764, CG18603, Dmel\\CG17715', description='CG17715 gene product from transcript CG17715-RG', other_designations='CG17715-PA|CG17715-PB|CG17715-PC|CG17715-PD|CG17715-PE|CG17715-PF|CG17715-PG|CG17715-PH|uncharacterized protein', map_location='', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22844220, end_position_on_the_genomic_accession=22854541, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3355130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355130, CurrentID=0, Status='live', Symbol='lt', Aliases='Dmel_CG18028, CG18028, Dm-lt(het), Dmel\\CG18028, Lt, Vps41, l(2)40Fb, vps41', description='light', other_designations='CG18028-PA|CG18028-PC|CG18028-PE|CG18028-PF|lt-PA|lt-PC|lt-PE|lt-PF|pink-wing|pinkoid', map_location='2-55.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22924795, end_position_on_the_genomic_accession=22943135, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  3355131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355131, CurrentID=0, Status='live', Symbol='cta', Aliases='Dmel_CG17678, CG17678, CG40010, Cta, Dmel\\CG17678, conc, mat(2)cta, mat(2)fog', description='concertina', other_designations='CG17678-PA|CG17678-PB|Galpha[[12/13]] subunit|cta-PA|cta-PB', map_location='2-54.8 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22981816, end_position_on_the_genomic_accession=22992951, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3355132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355132, CurrentID=0, Status='live', Symbol='CG14464', Aliases='Dmel_ BcDNA:LD29015, CG40285, Dmel\\CG14464', description='CG14464 gene product from transcript CG14464-RD', other_designations='CG14464-PC|CG14464-PD|uncharacterized protein', map_location='41C3-41C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4829246, end_position_on_the_genomic_accession=4830587, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355133, CurrentID=0, Status='live', Symbol='conu', Aliases='Dmel_CG17082, BcDNA:LD04957, CG17082, CG40286, Dmel\\CG17082', description='conundrum', other_designations='CG17082-PA|CG17082-PB|CG17082-PD|CG17082-PG|CG17082-PI|conu-PA|conu-PB|conu-PD|conu-PG|conu-PI', map_location='41C1-41C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4744348, end_position_on_the_genomic_accession=4758048, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  3355134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355134, CurrentID=0, Status='live', Symbol='CG17528', Aliases='Dmel_ CAMK-like, CG40283, Dmel\\CG17528, PK1, Pk1', description='CG17528 gene product from transcript CG17528-RD', other_designations='CG17528-PC|CG17528-PD|protein kinase 1|uncharacterized protein', map_location='41C2-41C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4815426, end_position_on_the_genomic_accession=4824319, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3355135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355135, CurrentID=0, Status='live', Symbol='Stlk', Aliases='Dmel_CG40293, BcDNA:LD05623, CG40293, Dmel\\CG40293, LD05623, stlk', description='Ste20-like kinase', other_designations='CG40293-PA|CG40293-PC|Stlk-PA|Stlk-PC', map_location='41A3-41A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4546691, end_position_on_the_genomic_accession=4565281, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3355136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355136, CurrentID=0, Status='live', Symbol='Nipped-B', Aliases='Dmel_CG17704, CG17704, CG40281, Dmel\\CG17704, NIPPED-B, NipB, l(2)02047, nipped-B', description='CG17704 gene product from transcript CG17704-RG', other_designations='CG17704-PE|CG17704-PF|CG17704-PG|CG17704-PH|CG17704-PI|CG17704-PJ|Nipped-B-PE|Nipped-B-PF|Nipped-B-PG|Nipped-B-PH|Nipped-B-PI|Nipped-B-PJ|nipped|nipped B', map_location='41B3-41C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4691672, end_position_on_the_genomic_accession=4729017, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  3355138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355138, CurrentID=0, Status='live', Symbol='CG17486', Aliases='Dmel_ CG40291, Dmel\\CG17486', description='CG17486 gene product from transcript CG17486-RB', other_designations='CG17486-PB|uncharacterized protein', map_location='41B1-41B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4619030, end_position_on_the_genomic_accession=4620833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3355139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355139, CurrentID=0, Status='live', Symbol='CG17883', Aliases='Dmel_ CG40289, Dmel\\CG17883', description='CG17883 gene product from transcript CG17883-RB', other_designations='CG17883-PA|CG17883-PB|CG17883-PC|CG17883-PD|uncharacterized protein', map_location='41B3-41B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4671853, end_position_on_the_genomic_accession=4673639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  45574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45574, CurrentID=0, Status='live', Symbol='ida', Aliases='Dmel_CG10850, APC5, Apc5, CG10850, Dmel\\CG10850, IDA, l(3)63Fb, l(3)SH7', description='imaginal discs arrested', other_designations='CG10850-PA|ida-PA', map_location='63F6-63F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3905344, end_position_on_the_genomic_accession=3907813, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355143, CurrentID=0, Status='live', Symbol='p120ctn', Aliases='Dmel_CG17484, CG17484, Dmel\\CG17484, Dp120[ctn], Dp120ctn, P120, P120-catenin, P120ctn, dP120ctn, dp120ctn, p120, p120[ctn]', description='Adherens junction protein p120', other_designations='Adherens-junction-protein-p120|CG17484-PA|P120-Catenin|adherens junction protein p120|p120 catenin|p120catenin|p120ctn-PA', map_location='41B1-41B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4595288, end_position_on_the_genomic_accession=4609492, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3355144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355144, CurrentID=0, Status='live', Symbol='RpL38', Aliases="Dmel_CG18001, 41Af, 41Ag, BcDNA:RE42506, CG18001, CG40278, Dmel\\CG18001, E-2b, GroupIV, M(2)41A, M(2)S2, M(2)S3, Rp L38, l(2)41Af, l(2)41Ag, l(2)41Ai, l(2)EMS4-72, l(2)EMS45-72, l(2R)A'", description='Ribosomal protein L38', other_designations='CG18001-PA|CG18001-PB|Minute(2)41A|RpL38-PA|RpL38-PB|enhancer on chromosome 2, complementation group b|lethal(2)41Af|nipped|ribosomal protein 38|ribosomal protein L38', map_location='2-55.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4515388, end_position_on_the_genomic_accession=4516458, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  45577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45577, CurrentID=0, Status='live', Symbol='Faa', Aliases='Dmel_CG14993, CG14993, Dmel\\CG14993, dfaa, l(3)64Al, l(3)SH11', description='Fumarylacetoacetase', other_designations='CG14993-PA|Faa-PA|fumarylacetoacetase', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4072032, end_position_on_the_genomic_accession=4073954, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3355149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355149, CurrentID=0, Status='live', Symbol='CG40298', Aliases='Dmel_ Dmel\\CG40298', description='CG40298 gene product from transcript CG40298-RA', other_designations='CG40298-PA|uncharacterized protein', map_location='80F9-80F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23726829, end_position_on_the_genomic_accession=23727989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3355150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355150, CurrentID=0, Status='live', Symbol='AGO3', Aliases='Dmel_CG40300, Ago-3, Ago3, BEST:LD15785, CG40300, Dmel\\CG40300, ago3, dAgo3', description='Argonaute 3', other_designations='AGO3-PD|AGO3-PE|AGO3-PF|AGO3-PG|Argonaute-3|Argonaute3|BEST:LD15785|CG40300-PD|CG40300-PE|CG40300-PF|CG40300-PG|argonaute 3', map_location='80F9-80F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23554593, end_position_on_the_genomic_accession=23689640, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3355151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355151, CurrentID=0, Status='live', Symbol='scro', Aliases='Dmel_CG17594, CG17594, CG17595, CG18452, Dmel\\CG17594, mdcds_36132u, scro', description='scarecrow', other_designations='CG17594-PA|CG17594-PB|CG17594-PC|CG17594-PF|scro-PA|scro-PB|scro-PC|scro-PF', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24724162, end_position_on_the_genomic_accession=24758036, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  45587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45587, CurrentID=0, Status='live', Symbol='nompA', Aliases='Dmel_CG13207, CG13207, Dmel\\CG13207, NOMPA, Nomp-A, NompA, nomp', description='no mechanoreceptor potential A', other_designations='CG13207-PA|CG13207-PB|CG13207-PC|CG13207-PD|CG13207-PF|nompA-PA|nompA-PB|nompA-PC|nompA-PD|nompA-PF', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11310335, end_position_on_the_genomic_accession=11318676, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  45588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45588, CurrentID=0, Status='live', Symbol='Pdp1', Aliases='Dmel_CG17888, CG17888, DM1, DM32, Dmel\\CG17888, PDP, PDP1, PDP1epsilon, Pdp-1 epsilon, Pdp1e, Pdp1epsilon, anon-WO0140519.206, mel_pdp1, pdp, pdp1, pdp1epsilon, Pdp1', description='PAR-domain protein 1', other_designations='CG17888-PB|CG17888-PC|CG17888-PD|CG17888-PF|CG17888-PG|CG17888-PI|CG17888-PJ|CG17888-PK|CG17888-PL|CG17888-PM|CG17888-PN|CG17888-PO|CG17888-PP|PAR domain protein 1epsilon|PAR-DOMAIN-PROTEIN 1|PAR-Domain protein 1|PAR-protein domain1|Par domain protein 1|Pdp1-PB|Pdp1-PC|Pdp1-PD|Pdp1-PF|Pdp1-PG|Pdp1-PI|Pdp1-PJ|Pdp1-PK|Pdp1-PL|Pdp1-PM|Pdp1-PN|Pdp1-PO|Pdp1-PP', map_location='66A14-66A17', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7811458, end_position_on_the_genomic_accession=7867369, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  3355158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355158, CurrentID=0, Status='live', Symbol='CG13865', Aliases='Dmel_ CG40301, Dmel\\CG13865, FBgn0039965', description='CG13865 gene product from transcript CG13865-RC', other_designations='CG13865-PA|CG13865-PB|CG13865-PC|uncharacterized protein', map_location='20F3-20F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23013495, end_position_on_the_genomic_accession=23016586, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3355160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355160, CurrentID=0, Status='live', Symbol='CG42346', Aliases='Dmel_ CG40500, CG40500-PA, CG41105, Dmel\\CG42346, Dmel_CG40500, Dmel_CG41105', description='CG42346 gene product from transcript CG42346-RD', other_designations='CG42346-PC|CG42346-PD|uncharacterized protein', map_location='20F1-20F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22832399, end_position_on_the_genomic_accession=22885541, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3355162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355162, CurrentID=0, Status='live', Symbol='CG40486', Aliases='Dmel_ CG12466, Dmel\\CG40486, FOHSDR', description='CG40486 gene product from transcript CG40486-RC', other_designations='CG40486-PB|CG40486-PC|uncharacterized protein', map_location='20F3-20F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22958856, end_position_on_the_genomic_accession=22960091, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3355163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355163, CurrentID=0, Status='live', Symbol='Rab21', Aliases='Dmel_CG17515, AAF45341, CG17515, CG40304, Dm Dmel\\CG17515, Q7PLE8, anon-EST:Posey237, rab21', description='CG17515 gene product from transcript CG17515-RD', other_designations='CG17515-PA|CG17515-PC|CG17515-PD|Rab21-PA|Rab21-PC|Rab21-PD', map_location='20F3-20F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23012140, end_position_on_the_genomic_accession=23013409, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3355164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355164, CurrentID=0, Status='live', Symbol='stnA', Aliases='Dmel_CG12500, 20, 31, 32, 8P1, CG12473/CG12500, CG12500, CG40306, Dmel\\CG12500, ORF1, STNA, X-3, X-43, X3, l(1)13-20, l(1)20Bc, l(1)X-3, l8P1, lR9-10, sesC, stn, stn-A, x-3', description='stoned A', other_designations='CG12500-PC|CG12500-PD|CG12500-PE|CG12500-PF|section 20|section 31|section 32|stnA-PC|stnA-PD|stnA-PE|stnA-PF|stoned|stonedA|stress-sensitive-C', map_location='20F3-20F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22980722, end_position_on_the_genomic_accession=23004648, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3355165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355165, CurrentID=0, Status='live', Symbol='CG40485', Aliases='Dmel_ Dmel\\CG40485, FOHSDR', description='CG40485 gene product from transcript CG40485-RB', other_designations='CG40485-PA|CG40485-PB|CG40485-PC|uncharacterized protein', map_location='20F3-20F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22960609, end_position_on_the_genomic_accession=22962360, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3355167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355167, CurrentID=0, Status='live', Symbol='CG41106', Aliases='Dmel_ Dmel\\CG41106', description='CG41106 gene product from transcript CG41106-RA', other_designations='CG41106-PA|uncharacterized protein', map_location='20F2-20F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22945321, end_position_on_the_genomic_accession=22946339, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3355171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355171, CurrentID=0, Status='live', Symbol='FucTC', Aliases='Dmel_CG40305, CG40305, Dmel\\CG40305', description='CG40305 gene product from transcript CG40305-RB', other_designations='CG40305-PB|FucTC-PB|alpha1,3-fucosyltransferase', map_location='20F3-20F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23008766, end_position_on_the_genomic_accession=23011303, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3355173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355173, CurrentID=0, Status='live', Symbol='Pp1-Y2', Aliases='Dmel_CG40448, CG40448, DmPp1-Y2, Dmel\\CG40448, Pp1Y2', description='Protein phosphatase 1, Y-linked 2', other_designations='CG40448-PC|Pp1-Y2-PC|protein phosphatase 1, Y-linked 2', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3539437, end_position_on_the_genomic_accession=3540722, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3355174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355174, CurrentID=0, Status='live', Symbol='ORY', Aliases='Dmel_CG40446, CG40446, Dmel\\CG40446, Ory, ory', description='Occludin-Related Y', other_designations='CG40446-PD|ORY-PD', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3561745, end_position_on_the_genomic_accession=3618339, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3355180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355180, CurrentID=0, Status='live', Symbol='Ppr-Y', Aliases='Dmel_CG40440, CG40440, Dmel\\CG40440, PPr-Y, PprY', description='CG40440 gene product from transcript CG40440-RB', other_designations='CG40440-PB|Ppr-Y-PB', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=1636574, end_position_on_the_genomic_accession=1884846, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3355181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355181, CurrentID=0, Status='live', Symbol='kl-2', Aliases='Dmel_CG17866, CG17866, CG40438, Dmel\\CG17866, Lms2, kl2', description='male fertility factor kl2', other_designations='1beta-dynein heavy chain|CG17866-PD|kl-2-PD|male fertility factor kl-2', map_location='h10-h11', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=573088, end_position_on_the_genomic_accession=664173, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  45655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45655, CurrentID=0, Status='live', Symbol='dom', Aliases='Dmel_CG9696, 27/4, CG9696, DOM, DOM-A, Dmel\\CG9696, Dom, DomA, Domino, anon-WO0172774.190, l(2)81/8, l(2)k02704, l(2)k08108, p400', description='domino', other_designations='CG9696-PA|CG9696-PD|CG9696-PE|CG9696-PG|Domino-A|dom-PA|dom-PD|dom-PE|dom-PG|domino DNA dependent ATPase', map_location='57D11-57D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21323430, end_position_on_the_genomic_accession=21341847, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  318907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318907, CurrentID=0, Status='live', Symbol='CG31710', Aliases='Dmel_ Dmel\\CG31710, EP2160', description='CG31710 gene product from transcript CG31710-RD', other_designations='CG31710-PA|CG31710-PC|CG31710-PD|uncharacterized protein', map_location='30C5-30C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9702442, end_position_on_the_genomic_accession=9705954, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  45668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45668, CurrentID=0, Status='live', Symbol='Vha68-1', Aliases='Dmel_CG12403, A subunit, ATP6V1A, CG12403, Dmel\\CG12403, Vha, Vha-68-1, Vha68, VhaA, vha67-1, vha68-1', description='Vacuolar H[+] ATPase 68kD subunit 1', other_designations='CG12403-PA|CG12403-PB|Vha68-1-PA|Vha68-1-PB|vacuolar ATPase subunit A|vacuolar H[+] ATPase 68kD subunit 1|vacuolar H[+] ATPase subunit 68-1|vacuolar H[+]-ATPase 68|vacuolar H[+]-ATPase 68kD A subunit|vacuolar H[+]-ATPase A subunit', map_location='34B-34A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13000770, end_position_on_the_genomic_accession=13004107, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  45680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45680, CurrentID=0, Status='live', Symbol='Tapdelta', Aliases='Dmel_CG9035, BcDNA:GM12291, CG9035, Dmel\\CG9035, TRAPdelta, l(2)k17005', description='Translocon-associated protein delta', other_designations='CG9035-PA|CG9035-PB|Tapdelta-PA|Tapdelta-PB', map_location='47F7-47F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11370300, end_position_on_the_genomic_accession=11371335, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  45682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45682, CurrentID=0, Status='live', Symbol='eIF-3p40', Aliases='Dmel_CG9124, BcDNA:GM14618, CG9124, Dmel\\CG9124, Q9U9Q4, eIF-3, eIF3-S3, eIF3h, eIf3-p44, l(2)k09003', description='Eukaryotic initiation factor 3 p40 subunit', other_designations='CG9124-PA|CG9124-PB|eIF-3p40-PA|eIF-3p40-PB|eukaryotic initiation factor 3 p40 subunit', map_location='25C3-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5051193, end_position_on_the_genomic_accession=5053094, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45683, CurrentID=0, Status='live', Symbol='milt', Aliases='Dmel_CG43227, CG13777, CG31630, CG43227, Dmel\\CG43227, Dmel_CG13777, Dmel_CG31630, anon-WO0153538.19, l(2)k04704', description='milton', other_designations='CG43227-PA|CG43227-PB|CG43227-PC|CG43227-PD|CG43227-PE|CG43227-PF|milt-PA|milt-PB|milt-PC|milt-PD|milt-PE|milt-PF', map_location='27D5-27D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7037823, end_position_on_the_genomic_accession=7056010, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  45706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45706, CurrentID=0, Status='live', Symbol='Cyp4d14', Aliases='Dmel_CG3540, 4d14, CG3540, CYP4D14, Dmel\\CG3540, EG:152A3.2', description='CG3540 gene product from transcript CG3540-RA', other_designations='CG3540-PA|CG3540-PB|Cyp4d14-PA|Cyp4d14-PB|cytochrome P450', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2173415, end_position_on_the_genomic_accession=2175315, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45707, CurrentID=0, Status='live', Symbol='CG18031', Aliases='Dmel_ Dmel\\CG18031, EG:103B4.2', description='CG18031 gene product from transcript CG18031-RA', other_designations='CG18031-PA|uncharacterized protein', map_location='2E2-2E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2201820, end_position_on_the_genomic_accession=2203866, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  5740866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740866, CurrentID=0, Status='live', Symbol='CG34301', Aliases='Dmel_ Dmel\\CG34301', description='CG34301 gene product from transcript CG34301-RA', other_designations='CG34301-PA|uncharacterized protein', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8773435, end_position_on_the_genomic_accession=8773810, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45739, CurrentID=0, Status='live', Symbol='SuUR', Aliases='Dmel_CG7869, CG7869, Dmel\\CG7869, Dromel_CG7869_FBtr0076203_suur_mORF, SU(UR), SUUR, Su(UR), Su(UR)ES, su(UR)ES', description='Suppressor of Under-Replication', other_designations='CG7869-PA|SuUR-PA|suppressor of Under- replication|suppressor of Under-Replication|suppressor of underreplication|suppressor of underreplication protein', map_location='68A4-68A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11064991, end_position_on_the_genomic_accession=11069025, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45757, CurrentID=0, Status='live', Symbol="beta'COP", Aliases="Dmel_CG6699, CG6699, COPIbeta', Dmel\\CG6699, beta'-COP, beta'Cop, dbeta'COP", description="Coat Protein (coatomer) beta'", other_designations="CG6699-PA|COPI beta' subunit|beta'-coatomer protein|beta'COP-PA|coat protein (coatomer) beta'", map_location='34B9-34B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13378902, end_position_on_the_genomic_accession=13382382, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45775, CurrentID=0, Status='live', Symbol='mei-P26', Aliases='Dmel_CG12218, AF145661, BcDNA:GH10646, CG12218, Dmel\\CG12218, MEI-P26, Mei-P26, mei-p26', description='meiotic P26', other_designations='CG12218-PA|CG12218-PB|CG12218-PC|CG12218-PD|CG12218-PE|mei-P26-PA|mei-P26-PB|mei-P26-PC|mei-P26-PD|mei-P26-PE', map_location='8C15-8C17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9163255, end_position_on_the_genomic_accession=9189309, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  45780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45780, CurrentID=0, Status='live', Symbol='Prosalpha1', Aliases='Dmel_CG18495, 18495, Alpha-1, CG18495, Dmel\\CG18495, PSA6_DROME, Prosalpha6, alpha-1, alpha1, l(2)SH2 2342, l(2)SH2342, prosalpha1, prosalpha6', description='Proteasome alpha1 subunit', other_designations='20S proteasome subunit alpha1|20s proteasome|CG18495-PA|CG18495-PB|Prosalpha1-PA|Prosalpha1-PB|alpha-subunit of 20S proteasome|alpha1|lethal (2) SH2342|proteasome alpha1 subunit|proteasome alpha6 subunit', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7827303, end_position_on_the_genomic_accession=7828678, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  45783 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45783, CurrentID=0, Status='live', Symbol='Aats-val', Aliases='Dmel_CG4062, BEST:LD45671, CG4062, Dmel\\CG4062, VRS, ValRS, l(2)03531, l(2)rI255, ld45671', description='Valyl-tRNA synthetase', other_designations='Aats-val-PA|Aats-val-PB|Aats-val-PC|CG4062-PA|CG4062-PB|CG4062-PC', map_location='49F7-49F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13198850, end_position_on_the_genomic_accession=13202709, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  45784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45784, CurrentID=0, Status='live', Symbol='Aats-thr', Aliases='Dmel_CG5353, 42/3, CG5353, Dmel\\CG5353, P539, TRS, ThrRS, l(2)k04203, threonyl-tRNA synthetase', description='Threonyl-tRNA synthetase', other_designations='Aats-thr-PA|Aats-thr-PB|Aats-thr-PC|Aats-thr-PD|CG5353-PA|CG5353-PB|CG5353-PC|CG5353-PD|threonyl-tRNA synthetase', map_location='33C4-33C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12103751, end_position_on_the_genomic_accession=12107170, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  45785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45785, CurrentID=0, Status='live', Symbol='Aats-ile', Aliases='Dmel_CG11471, CG11471, Dmel\\CG11471, IRS, IleRS, l(3)00827', description='Isoleucyl-tRNA synthetase', other_designations='Aats-ile-PA|Aats-ile-PC|Aats-ile-PD|CG11471-PA|CG11471-PC|CG11471-PD', map_location='79D4-79D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22288665, end_position_on_the_genomic_accession=22292810, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  45786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45786, CurrentID=0, Status='live', Symbol='Aats-gln', Aliases='Dmel_CG10506, Aast-g1n, BcDNA:GH11673, CG10506, Dmel\\CG10506, GlnRS, QRS, l(3)05461, qars', description='Glutaminyl-tRNA synthetase', other_designations='Aats-gln-PA|CG10506-PA', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25263267, end_position_on_the_genomic_accession=25265721, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  45815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45815, CurrentID=0, Status='live', Symbol='Spn27A', Aliases='Dmel_CG11331, CG11331, Dm-Spn27A, Dmel\\CG11331, EST1, Spn27, Spn27a, sp27A, spn27A', description='Serpin 27A', other_designations='CG11331-PA|CG11331-PB|Dm-serpin-27A|Serpin-27A|Serpin27A|Spn27A-PA|Spn27A-PB|serpin 27A', map_location='26F6-26F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6671065, end_position_on_the_genomic_accession=6673347, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  45816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45816, CurrentID=0, Status='live', Symbol='onecut', Aliases='Dmel_CG1922, A1, CG1922, Dmel\\CG1922', description='CG1922 gene product from transcript CG1922-RB', other_designations='CG1922-PA|CG1922-PB|D-onecut|onecut-PA|onecut-PB', map_location='102D1-102D1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=598555, end_position_on_the_genomic_accession=607649, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  45821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45821, CurrentID=0, Status='live', Symbol='Abl', Aliases='Dmel_CG4032, 4674, ABL1, AblK, Ableson, Am ABL, C-abl, CG4032, D-Abl, D-abl, D-ash, DAbl, DROTKABL3, Dabl, Dash, Ddash/abl, Dmel\\CG4032, Dsrc7, abl, abl1, c-abl, cAbl, dAbl, l(3)04674, l(3)73Ba, l(3)c-abl, Abl', description='Abl tyrosine kinase', other_designations='Abl oncogene|Abl-PA|Abl-PB|Abl-PC|Abl-PD|Abl-PE|Abl-PF|Abl-PG|Abl-PH|Abl-PI|Abl-PJ|CG4032-PA|CG4032-PB|CG4032-PC|CG4032-PD|CG4032-PE|CG4032-PF|CG4032-PG|CG4032-PH|CG4032-PI|CG4032-PJ|D-abelson|abelson|abelson kinase|abelson tyrosine kinase|abl kinase', map_location='73B1-73B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16615866, end_position_on_the_genomic_accession=16647882, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  45826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45826, CurrentID=0, Status='live', Symbol='bel', Aliases='Dmel_CG9748, BEL, Bel, CG9748, DmRH6, Dmel\\CG9748, L3, anon-85Ab, cap, cg9748, l(3)85Ac, l(3)L3, l(3)L4740, l(3)S047110, l(3)S097074, ms(3)Pneo85A, ms(3)neo30, psg9', description='belle', other_designations='CG9748-PA|CG9748-PB|bel-PA|bel-PB|capon|lethal (3) L4740|lethal (3) S047110', map_location='3-44.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8654975, end_position_on_the_genomic_accession=8660761, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  45827 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45827, CurrentID=0, Status='live', Symbol='bic', Aliases='Dmel_CG3644, 107/12, 5A, BcDNA.GM05329, BcDNA:GM05329, Bic, Btf, CG3644, Dmel\\CG3644, E(2)Bic, E(Bic), l(2)49Da, l(2)k10712, l(2)vr22, vr22', description='bicaudal', other_designations='CG3644-PA|CG3644-PB|bic-PA|bic-PB|enhancer of bicaudal', map_location='2-64.0 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12870249, end_position_on_the_genomic_accession=12871794, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45830, CurrentID=0, Status='live', Symbol='bur', Aliases='Dmel_CG9242, CG9242, Dmel\\CG9242, ES2-2, GMPS, Gmps, P174, anon-WO0118547.193, anon-WO0118547.204, gua2, l(2)10523', description='burgundy', other_designations='CG9242-PB|CG9242-PC|GMP synthetase|bur-PB|bur-PC|enhancer of GMR-sina 2-2', map_location='2-55.7 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21174182, end_position_on_the_genomic_accession=21179683, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  246611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246611, CurrentID=0, Status='live', Symbol='CG30432', Aliases='Dmel_ Dmel\\CG30432', description='CG30432 gene product from transcript CG30432-RB', other_designations='CG30432-PB|uncharacterized protein', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6197387, end_position_on_the_genomic_accession=6198266, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  45840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45840, CurrentID=0, Status='live', Symbol='cpo', Aliases='Dmel_CG43738, 1824, 3603, 5339, 6015, BcDNA:AT31405, CG12349, CG18434, CG18435, CG31243, CG42457, CG43738, CPO, Cpo, Dmel\\CG43738, Dmel_CG31243, Dmel_CG42457, E(sda)N, Line N, ORE-17, cg18435, l(3)01432, l(3)6015, l(3)j4A1, l(3)j4D4, l(3)j7E2, l(3)j9B3, l(3)rJ553, l(3)s2336', description='couch potato', other_designations='CG43738-PN|CG43738-PO|CG43738-PP|CG43738-PQ|CG43738-PR|CG43738-PS|CG43738-PT|CG43738-PU|CG43738-PV|CG43738-PW|CG43738-PX|CG43738-PY|cpo-PN|cpo-PO|cpo-PP|cpo-PQ|cpo-PR|cpo-PS|cpo-PT|cpo-PU|cpo-PV|cpo-PW|cpo-PX|cpo-PY', map_location='90D1-90E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17919832, end_position_on_the_genomic_accession=18018892, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  45845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45845, CurrentID=0, Status='live', Symbol='Dsk', Aliases='Dmel_CG18090, BcDNA:RH46857, CCK, CG18090, DSK, DSK I, DSK II, DSK-0, DSK-1, DSK-2, DSK-I, Dmel\\CG18090, Drm-SK, Drm-SK-0, Drm-SK-1, Drm-SK-2, DrmSK-I, DrmSK-II, SK-0, dsk', description='Drosulfakinin', other_designations='CG18090-PA|Dsk-PA|cholecystokinin-like|drosophila sulfakinin|drososulfakinin|drosulfakinin|sulfakinin|sulfakinin-0', map_location='81F6-81F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4189692, end_position_on_the_genomic_accession=4190260, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  45848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45848, CurrentID=0, Status='live', Symbol='e(y)2', Aliases='Dmel_CG15191, CG15191, Dmel\\CG15191, E(y)2, ENY2, Sus1, d, e(y)[2]', description='enhancer of yellow 2', other_designations='CG15191-PA|e(y)2-PA', map_location='10C7-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11576118, end_position_on_the_genomic_accession=11576620, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  40411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40411, CurrentID=0, Status='live', Symbol='CG11248', Aliases='Dmel_ Dmel\\CG11248', description='CG11248 gene product from transcript CG11248-RC', other_designations='CG11248-PA|CG11248-PB|CG11248-PC|uncharacterized protein', map_location='78E3-78E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21680050, end_position_on_the_genomic_accession=21683231, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45871, CurrentID=0, Status='live', Symbol='grau', Aliases='Dmel_CG33133, CG3282, CG33133, Dmel\\CG33133', description='grauzone', other_designations='CG33133-PA|grau-PA', map_location='2-97 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21009764, end_position_on_the_genomic_accession=21011921, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  45875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45875, CurrentID=0, Status='live', Symbol='Hex-A', Aliases='Dmel_CG3001, CG3001, DM1, Dmel\\CG3001, HEX, HEX-A, HK, HK-A, Hex:B, HexA, hexokinase, Hex-A', description='Hexokinase A', other_designations='CG3001-PA|CG3001-PB|CG3001-PC|Hex-A-PA|Hex-A-PB|Hex-A-PC|hexokinase|hexokinase A', map_location='1-29.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9585675, end_position_on_the_genomic_accession=9589813, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  45879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45879, CurrentID=0, Status='live', Symbol='ImpE1', Aliases='Dmel_CG32356, BcDNA:RE39082, CG13668, CG32356, CG7116, CT22003, CT33103, Dmel\\CG32356, IMP-E1, Imp-E1', description='Ecdysone-inducible gene E1', other_designations='CG32356-PA|CG32356-PB|CG32356-PC|ImpE1-PA|ImpE1-PB|ImpE1-PC|Y1-like', map_location='66C8-66C10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8372405, end_position_on_the_genomic_accession=8391171, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  45880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45880, CurrentID=0, Status='live', Symbol='ImpL3', Aliases='Dmel_CG10160, CG10160, Dmel\\CG10160, IMP-L3, IMPL3, Imp-L3, Impl3, LDH, Ld, Ldh, dLDH, dLdha, dmLDH, impl3, ldh, lincRNA.S4223', description='Ecdysone-inducible gene L3', other_designations='CG10160-PA|CG10160-PB|CG10160-PC|CG10160-PD|Ecdysone-induced gene L3|ImpL3-PA|ImpL3-PB|ImpL3-PC|ImpL3-PD|lactate dehydrogenase|lactate dehydrogenase B|lactate/malate dehydrogenease|lactic DH', map_location='65A11-65A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6259105, end_position_on_the_genomic_accession=6262694, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  45881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45881, CurrentID=0, Status='live', Symbol='ix', Aliases='Dmel_CG13201, CG13201, Dmel\\CG13201, MED29', description='intersex', other_designations='CG13201-PA|ix-PA|mediator complex subunit 29', map_location='2-60.5 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11359437, end_position_on_the_genomic_accession=11360163, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  45883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45883, CurrentID=0, Status='live', Symbol='kar', Aliases='Dmel_CG12286, CG12286, Dmel\\CG12286', description='karmoisin', other_designations='CG12286-PA|CG12286-PB|kar-PA|kar-PB', map_location='3-51.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12682186, end_position_on_the_genomic_accession=12689367, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  45884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45884, CurrentID=0, Status='live', Symbol='kkv', Aliases='Dmel_CG2666, Blimp, CG2666, CS-1, CS-1/kkv, ChSB, DmCS-1, DmeChSB, Dmel\\CG2666, Kkv, blimp, l(3)82Fh', description='krotzkopf verkehrt', other_designations='CG2666-PA|CG2666-PC|CG2666-PD|chitin Syn1|chitin synthase 1|chitin synthase B|chitin synthase-1|kkv-PA|kkv-PC|kkv-PD|krotzkopf-verkehrt|short of breath', map_location='3-47.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5378093, end_position_on_the_genomic_accession=5392866, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  45894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45894, CurrentID=0, Status='live', Symbol='lds', Aliases='Dmel_CG2684, CG2684, DmF2, Dmel\\CG2684, Fs(3)Hor, Fs(3)Horka, Fs(3)Sz11, Horka, Lds, NTef2, early, lod', description='lodestar', other_designations='CG2684-PA|early-3|early-5|early-B|factor 2|female sterile (3) horka|lds-PA|negative transcription elongation factor 2|transcript A', map_location='3-47.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7920431, end_position_on_the_genomic_accession=7924491, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318946, CurrentID=0, Status='live', Symbol='CG31798', Aliases='Dmel_ Dmel\\CG31798, anon-WO0140519.187', description='CG31798 gene product from transcript CG31798-RA', other_designations='CG31798-PA|CG31798-PB|uncharacterized protein', map_location='37D4-37D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19394160, end_position_on_the_genomic_accession=19396190, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  45906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45906, CurrentID=0, Status='live', Symbol='mr', Aliases='Dmel_CG3060, APC2, APC2/mr, Apc2, CG3060, Dmel\\CG3060, MR, dAPC2', description='morula', other_designations='CG3060-PA|mr-PA', map_location='2-106.7 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23972431, end_position_on_the_genomic_accession=23975408, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  45908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45908, CurrentID=0, Status='live', Symbol='nec', Aliases='Dmel_CG1857, CG1857, Dmel\\CG1857, NEC, Nec, SER3, Ser3, Spn43Ac, anon-WO0118057.3', description='necrotic', other_designations='CG1857-PA|nec-PA|necrotic serpin|serine protease inhibitor 3|serpin43Ac', map_location='42F3-42F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7156477, end_position_on_the_genomic_accession=7158241, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45910, CurrentID=0, Status='live', Symbol='ort', Aliases='Dmel_CG7411, CG7411, CT22815, DMHisClalpha1, Dm-hclA, DmHisCl2, Dmel\\CG7411, GluCI, HA-CI I, HclA, HisC12, HisCI2, HisCl, HisCl-alpha1, HisCl2, Hist1, ORT, Ort, hclA, hisCl2, ora', description='ora transientless', other_designations='CG7411-PA|Histamine-gated chloride channel subunit 2|glycine receptor alpha2-like|histamine-gated chloride channel subunit 1|indefinite|ort-PA|outer rhabdomere absent|outer rhabdomeres absent|outer rhabdomeres transientless|outer-rhabdomeres-absent', map_location='3-66.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19659748, end_position_on_the_genomic_accession=19663672, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  45922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45922, CurrentID=0, Status='live', Symbol='rap', Aliases='Dmel_CG3000, APC/C[Fzr/Cdh1], CDH1, CG3000, Cdh1, Cdh1/Fzr, Cdh1[Fzr], Cdh[Fzr], Dmel\\CG3000, Fxr, Fzr, Fzr/rap, Hec/Cdh[Fzr], Rap/Fzr, anon-WO03040301.193, cdh1, dFZR, fzr, fzr/Cdh1, l(1)G0326, l(1)G0418/Fzr, rap/fizzy-related, rap/fzr, redax, wex, rap', description='retina aberrant in pattern', other_designations='CG3000-PA|CG3000-PB|Cdh1|Fizzy-related|Fizzy-related/Cdh1|fizzy related|fzr|rap-PA|rap-PB|redax|reduced axons|westfalenexpress', map_location='4C11-4C12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4667290, end_position_on_the_genomic_accession=4679537, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  45928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45928, CurrentID=0, Status='live', Symbol='shi', Aliases='Dmel_CG18102, CG18102, DNM3, DYN, Ddyn, Ddyn3, Ddyn4, Dmel\\CG18102, Dyn, Shi, anon-WO0153538.12, anon-WO0153538.13, anon-WO0153538.14, dDyn, dyn, dynamin, l(1)VII, l(1)shibere, shi', description='shibire', other_designations='CG18102-PA|CG18102-PB|CG18102-PC|CG18102-PE|CG18102-PF|CG18102-PG|CG18102-PH|CG18102-PI|CG18102-PJ|CG18102-PK|CG18102-PL|CG18102-PM|CG18102-PN|CG18102-PO|CG18102-PP|dynamin|shi-PA|shi-PB|shi-PC|shi-PE|shi-PF|shi-PG|shi-PH|shi-PI|shi-PJ|shi-PK|shi-PL|shi-PM|shi-PN|shi-PO|shi-PP|shibere|shibiri', map_location='1-51.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15892116, end_position_on_the_genomic_accession=15906716, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  45931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45931, CurrentID=0, Status='live', Symbol='spir', Aliases='Dmel_CG10076, 38C.34, 38C.37, CG10076, CG18621, Dmel\\CG10076, NP2788, Spir, Spire, l(2)08327, p150-Spir, p150Spir', description='spire', other_designations='CG10076-PA|CG10076-PB|CG10076-PC|CG10076-PD|CG10076-PE|CG10076-PF|CG10076-PG|CG10076-PH|CG10076-PI|CG10076-PJ|spir-PA|spir-PB|spir-PC|spir-PD|spir-PE|spir-PF|spir-PG|spir-PH|spir-PI|spir-PJ', map_location='2-54 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20311219, end_position_on_the_genomic_accession=20348386, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  45959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45959, CurrentID=0, Status='live', Symbol='mts', Aliases='Dmel_CG7109, 5559, CG7109, DmPp2A-28D, Dmel\\CG7109, ER2-6, MTS/PP2A, Mts, PP2, PP2A, PP2A 28D, PP2A C, PP2A-C, PP2A/MTS, PP2A[C], PP2A[[C]], PP2Ac, PP2a, PP2a 28D, Pp2A, Pp2A-28D, dPP2A, dPP2A-C, l(2)02496, l(2)s5286', description='microtubule star', other_designations='CG7109-PA|CG7109-PB|CG7109-PC|Microtubule-star|PP2A catalytic subunit|catalytic subunit C of type 2A protein phosphatase|mts-PA|mts-PB|mts-PC|phosphatase 2A catalytic subunit|protein phosphatase 2A|protein phosphatase 2A at 28D', map_location='28D1-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7827650, end_position_on_the_genomic_accession=7833576, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  45960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45960, CurrentID=0, Status='live', Symbol='aly', Aliases='Dmel_CG2075, 143450_at, Aly, CG2075, Dmel\\CG2075, ms(3)2, ms(3)ry2', description='always early', other_designations='CG2075-PA|CG2075-PB|aly-PA|aly-PB', map_location='3-4.4 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3024129, end_position_on_the_genomic_accession=3026521, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  45961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45961, CurrentID=0, Status='live', Symbol='Pep', Aliases='Dmel_CG6143, CG6143, Dmel\\CG6143, PEP, PEP/X4, X4/PEP, p100, pep', description='Protein on ecdysone puffs', other_designations='74F gene|CG6143-PA|CG6143-PB|CG6143-PC|CG6143-PD|Pep-PA|Pep-PB|Pep-PC|Pep-PD|protein on ecdysone puffs', map_location='74E4-74E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17638103, end_position_on_the_genomic_accession=17645036, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  45962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45962, CurrentID=0, Status='live', Symbol='msopa', Aliases='Dmel_CG14560, CG14560, Dmel\\CG14560', description='male-specific opa containing gene', other_designations='CG14560-PA|CG14560-PB|dromsopa|drosophila male specific opa containing gene|msopa-PA|msopa-PB', map_location='79B3-79B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22082720, end_position_on_the_genomic_accession=22083367, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  45969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45969, CurrentID=0, Status='live', Symbol='gwl', Aliases='Dmel_CG7719, 1J, CG7719, CG7719-related, CT23435, Dmel\\CG7719, EP0515, Gwl, Pk91C, Pk?2, anon-WO2004063362.73, scant', description='greatwall', other_designations='CG7719-PA|CG7719-PB|gwl-PA|gwl-PB|protein kinase-like 91C|scott of the antarctic', map_location='91C5-91C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18746370, end_position_on_the_genomic_accession=18750360, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  45970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45970, CurrentID=0, Status='live', Symbol='S6KL', Aliases='Dmel_CG7001, 3-10, BIN4, BcdK, Bin4, CG7001, Dmel\\CG7001, Pk17E, Pk?4, bin4, pk17e', description='S6 Kinase Like', other_designations='Bcd interacting protein K|CG7001-PA|CG7001-PB|CG7001-PC|S6 kinase like|S6KL-PA|S6KL-PB|S6KL-PC|bicoid interacting protein 4|protein kinase-like 17E', map_location='17D6-17E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18788464, end_position_on_the_genomic_accession=18806057, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  45973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45973, CurrentID=0, Status='live', Symbol='fs(1)N', Aliases='Dmel_CG11411, 371, 8D8.1, CG11411, Dmel\\CG11411, EG:8D8.1, Fs(1)N, Nas, fs(1)A371, fs(1)K1540, fs(1)M6as, fs(1)Nasrat, fs(1)[nas], nas, fs(1)N', description='female sterile (1) Nasrat', other_designations='CG11411-PA|female sterile (1) nasrat|fs(1)N-PA|fs(1)Nasrat|nasrat', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1322513, end_position_on_the_genomic_accession=1329438, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  45976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45976, CurrentID=0, Status='live', Symbol='retn', Aliases='Dmel_CG5403, 2535, BcDNA:LD35748, CG5403, DRI, Dead Ringer/Retain, Dmel\\CG5403, Dri, EHMT, EY2-10, Retn, dri, l(2)02535/dri, retn', description='retained', other_designations='CG5403-PA|CG5403-PB|CG5403-PC|CG5403-PD|CG5403-PE|dead Ringer/Retain|dead ringer|dead-ringer|retn-PA|retn-PB|retn-PC|retn-PD|retn-PE', map_location='59F5-59F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23632610, end_position_on_the_genomic_accession=23654820, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  5740407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740407, CurrentID=0, Status='live', Symbol='CG34163', Aliases='Dmel_ Dmel\\CG34163', description='CG34163 gene product from transcript CG34163-RA', other_designations='CG34163-PA|CG34163-PB|uncharacterized protein', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11988270, end_position_on_the_genomic_accession=11988801, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  45994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=45994, CurrentID=0, Status='live', Symbol='fs(1)M3', Aliases='Dmel_CG4790, CG4790, Dmel\\CG4790, Fs(1)K646, fs(1)K646, fs(1)ph, fs(1)phl, fs(1)pole hole, fs(1)polehole', description='female sterile (1) M3', other_designations='CG4790-PA|fs(1)M3-PA|fs(1)polehole|polehole', map_location='1-15 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5958207, end_position_on_the_genomic_accession=5964247, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  46000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46000, CurrentID=0, Status='live', Symbol='scw', Aliases='Dmel_CG31695, BcDNA:LD30945, CG31695, Dmel\\CG31695, SCW, Scw, l(2)37Ff, l(2)IG76, mor2', description='screw', other_designations='CG31695-PA|CG31695-PB|morphological defects 2|scw-PA|scw-PB', map_location='2-53 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19699379, end_position_on_the_genomic_accession=19700968, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  46006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46006, CurrentID=0, Status='live', Symbol='east', Aliases='Dmel_CG4399, CG4399, Dmel\\CG4399, EAST, EG:133E12.4, East, anon-WO03040301.224, l(1)G0014, l(1)G0500, su(w[sp])', description='enhanced adult sensory threshold', other_designations='CG4399-PB|CG4399-PC|CG4399-PD|CG4399-PE|CG4399-PF|CG4399-PG|east-PB|east-PC|east-PD|east-PE|east-PF|east-PG|enhanced adult salt tolerance|suppressor of white-spotted', map_location='1-0.16 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2010860, end_position_on_the_genomic_accession=2024647, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  46008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46008, CurrentID=0, Status='live', Symbol='Gel', Aliases='Dmel_CG1106, CG1106, DGS, Dmel\\CG1106, gel', description='Gelsolin', other_designations='CG1106-PA|CG1106-PB|CG1106-PD|CG1106-PF|CG1106-PJ|CG1106-PK|CG1106-PL|Gel-PA|Gel-PB|Gel-PD|Gel-PF|Gel-PJ|Gel-PK|Gel-PL|drosophila gelsolin|gelsolin', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4235217, end_position_on_the_genomic_accession=4241057, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  46015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46015, CurrentID=0, Status='live', Symbol='tutl', Aliases='Dmel_CG15427, CG15426, CG15427, CG15428, CT35489, CT35490, Dmel\\CG15427, Tutl, jf4, l(2)01085, l(2)24Df, l(2)jf4, mdcds_7441', description='turtle', other_designations='CG15427-PA|CG15427-PC|CG15427-PD|CG15427-PE|CG15427-PF|CG15427-PG|CG15427-PH|CG15427-PI|CG15427-PJ|CG15427-PK|lethal(2)24Df|tutl-PA|tutl-PC|tutl-PD|tutl-PE|tutl-PF|tutl-PG|tutl-PH|tutl-PI|tutl-PJ|tutl-PK', map_location='24E1-24E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4283148, end_position_on_the_genomic_accession=4323781, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  46017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46017, CurrentID=0, Status='live', Symbol='l(2)01289', Aliases='Dmel_CG9432, CG30444, CG9432, CG9435, Dmel\\CG9432', description='lethal (2) 01289', other_designations='CG9432-PA|CG9432-PB|CG9432-PF|CG9432-PG|CG9432-PH|CG9432-PJ|CG9432-PK|CG9432-PL|CG9432-PM|Fusion-2|Fusion2|l(2)01289-PA|l(2)01289-PB|l(2)01289-PF|l(2)01289-PG|l(2)01289-PH|l(2)01289-PJ|l(2)01289-PK|l(2)01289-PL|l(2)01289-PM', map_location='42C6-42C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6721101, end_position_on_the_genomic_accession=6740672, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  46020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46020, CurrentID=0, Status='live', Symbol='NAT1', Aliases='Dmel_CG3845, CG3845, Dmel\\CG3845, Nat1, d l(2)01424', description='CG3845 gene product from transcript CG3845-RA', other_designations='CG3845-PA|CG3845-PB|CG3845-PD|NAT1-PA|NAT1-PB|NAT1-PD|lethal (2) 01424', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12915994, end_position_on_the_genomic_accession=12923734, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  46027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46027, CurrentID=0, Status='live', Symbol='kermit', Aliases='Dmel_CG11546, BEST:LD04728, CG11546, Dmel\\CG11546, LD04728, dGIPC, dKermit, l(2)02045', description='CG11546 gene product from transcript CG11546-RB', other_designations='CG11546-PA|CG11546-PB|CG11546-PC|CG11546-PD|CG11546-PE|kermit-PA|kermit-PB|kermit-PC|kermit-PD|kermit-PE|lethal (2) 02045', map_location='44B5-44B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8146918, end_position_on_the_genomic_accession=8156654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  46032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46032, CurrentID=0, Status='live', Symbol='Bka', Aliases='Dmel_CG4539, BcDNA:RE14949, CG4539, Dmel\\CG4539, l(2)02695, mip', description='Bekka', other_designations='Bka-PA|CG4539-PA|bekka|missing imaginal precursors', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9895287, end_position_on_the_genomic_accession=9896104, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  46035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46035, CurrentID=0, Status='live', Symbol='Ccs', Aliases='Dmel_CG17753, CCD, CCS, CG-17753, CG17753, DmCCS, Dmel\\CG17753, anon-EST:Posey231, dCCS, l(2)03221', description='Copper chaperone for superoxide dismutase', other_designations='CCS|CG17753-PB|CG17753-PC|Ccs-PB|Ccs-PC|Cu-chaperone for superoxide dismutase|copper chaperone for superoxide dismutase', map_location='46F1-46F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10121486, end_position_on_the_genomic_accession=10122574, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  46038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46038, CurrentID=0, Status='live', Symbol='Phb2', Aliases='Dmel_CG15081, CG15081, Dmel\\CG15081, anon-WO0118547.157, anon-WO02079478.4, l (2) 03709, l(2)03709, phb2', description='Prohibitin 2', other_designations='CG15081-PA|CG15081-PB|CG15081-PC|CG15081-PD|CG15081-PE|CG15081-PF|Phb2-PA|Phb2-PB|Phb2-PC|Phb2-PD|Phb2-PE|Phb2-PF|lethal (2) 03709|prohibitin 2', map_location='55F5-55F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18819467, end_position_on_the_genomic_accession=18822250, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  46040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46040, CurrentID=0, Status='live', Symbol='cype', Aliases='Dmel_CG14028, BcDNA:GH04604, CG14028, CG18323, COX6C, COXH, COXVIc, Cyc, Cype, Dmel\\CG14028, VIc, cyc, l(2)03771', description='cyclope', other_designations='CG14028-PA|CG14028-PB|CG14028-PC|cype-PA|cype-PB|cype-PC|cytochrome c oxidase VIc|cytochrome c oxidase polypeptide VIc|cytochrome c oxidase subunit 6C|subunit VIc', map_location='25D6-25D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5327227, end_position_on_the_genomic_accession=5327969, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  46058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46058, CurrentID=0, Status='live', Symbol='Prosbeta1', Aliases='Dmel_CG8392, 1(2)05070, 8392, Beta-1, CG8392, Dmel\\CG8392, beta-1, beta1, l(2)05070, prosbeta1', description='Proteasome beta1 subunit', other_designations='CG8392-PA|Prosbeta1-PA|beta1|lethal (2) 05070|proteasome beta1 subunit', map_location='52D13-52D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16013805, end_position_on_the_genomic_accession=16014787, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  46059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46059, CurrentID=0, Status='live', Symbol='Sply', Aliases='Dmel_CG8946, CG8946, Dmel\\CG8946, SPL, Spl, anon-WO0142479.5, l(2)05091, spl, sply', description='Sphingosine-1-phosphate lyase', other_designations='CG8946-PA|CG8946-PB|Sply-PA|Sply-PB|sphinganine-1-phosphate lyase|sphingosine-1-phosphate lyase', map_location='53F6-53F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17088441, end_position_on_the_genomic_accession=17091958, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  46066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46066, CurrentID=0, Status='live', Symbol='l(2)05714', Aliases='Dmel_CG8886, CG8886, Dmel\\CG8886, l(2)05715', description='lethal (2) 05714', other_designations='CG8886-PA|CG8886-PB|CG8886-PC|l(2)05714-PA|l(2)05714-PB|l(2)05714-PC', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4970135, end_position_on_the_genomic_accession=4973921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  46068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46068, CurrentID=0, Status='live', Symbol='Acsl', Aliases='Dmel_CG8732, 7447, CG8732, Dmel\\CG8732, FACS, FCS, anon-WO0172774.173, dAcsl, dFAC, l(2)05847, l(2)44DEa, l(2)7447, l(2)SH0465, l(2)SH2 0465, l(2)k05304', description='Acyl-CoA synthetase long-chain', other_designations='Acsl-PA|Acsl-PB|Acsl-PC|Acsl-PD|Acsl-PE|Acsl-PF|Acsl-PG|Acsl-PH|Acsl-PI|Acsl-PJ|CG8732-PA|CG8732-PB|CG8732-PC|CG8732-PD|CG8732-PE|CG8732-PF|CG8732-PG|CG8732-PH|CG8732-PI|CG8732-PJ|Long-chain acyl-CoA synthetase|fatty acyl CoA synthetase|fatty acyl coenzyme A (CoA) synthetase|fatty acyl-CoA synthetase|lethal (2) 44DEa|lethal (2) SH0465', map_location='44D3-44D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8665813, end_position_on_the_genomic_accession=8683022, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  46069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46069, CurrentID=0, Status='live', Symbol='ATPsynG', Aliases='Dmel_CG6105, ATPN, BcDNA:RE20862, BcDNA:RH28372, CG6105, Dmel\\CG6105, I(2)06225, anon-EST:Posey10, l(2)06225', description='ATP synthase', other_designations='ATP synthase g chain|ATP synthase, subunit G|ATPsynG-PA|ATPsynG-PB|ATPsynG-PD|CG6105-PA|CG6105-PB|CG6105-PD|F0/F1 ATP synthase G|lethal (2) 06225', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10975522, end_position_on_the_genomic_accession=10976399, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  46074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46074, CurrentID=0, Status='live', Symbol='DCTN3-p24', Aliases='Dmel_CG9893, CG9893, Dmel\\CG9893, l(2)06 496, l(2)06496, p22/p24, p24', description='Dynactin 3, p24 subunit', other_designations='CG9893-PA|CG9893-PB|DCTN3-p24-PA|DCTN3-p24-PB|dynactin 3, p24 subunit|dynactin, p24 subunit|lethal (2) 06496|lethal-(2)06 496', map_location='59C2-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23061420, end_position_on_the_genomic_accession=23062593, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  46078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46078, CurrentID=0, Status='live', Symbol='vlc', Aliases='Dmel_CG8390, CG8390, Dmel\\CG8390, X, anon-41Fa, anon-WO0118547.98, l(2)07022', description='vulcan', other_designations='CG8390-PA|CG8390-PB|CG8390-PC|CG8390-PE|CG8390-PF|CG8390-PG|CG8390-PH|CG8390-PI|anon-41Fa|transcription unit X|vlc-PA|vlc-PB|vlc-PC|vlc-PE|vlc-PF|vlc-PG|vlc-PH|vlc-PI', map_location='41F9-41F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5702112, end_position_on_the_genomic_accession=5705439, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  46080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46080, CurrentID=0, Status='live', Symbol='Sec61beta', Aliases='Dmel_CG10130, CG10130, DSec61beta, Dmel\\CG10130, Sec1, Sec61, Secbeta61, l(2)07214, sec61b, sec61beta', description='Sec61 beta subunit', other_designations='CG10130-PA|Sec61beta-PA', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14618771, end_position_on_the_genomic_accession=14619681, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  46085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46085, CurrentID=0, Status='live', Symbol='MFS14', Aliases='Dmel_CG15095, 8717, CG15095, Dmel\\CG15095, l(2)08717, l208717, lethal (2) 08717, lethal(2)08717', description='Major Facilitator Superfamily Transporter 14', other_designations='CG15095-PA|CG15095-PB|CG15095-PC|MFS14-PA|MFS14-PB|MFS14-PC|lethal (2) 08717|major facilitator superfamily transporter 14', map_location='55F3-55F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18796975, end_position_on_the_genomic_accession=18803727, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  46091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46091, CurrentID=0, Status='live', Symbol='Nup214', Aliases='Dmel_CG3820, CAN, CAN/Nup214, CG3820, DCAN, DNup214, Dmel\\CG3820, NUP214, Nup, Nup 214, dCAN, l(2)10444, nup214', description='Nucleoporin 214kD', other_designations='CG3820-PA|Nup214-PA|nucleoporin 214|nucleoporin 214kD', map_location='59B4-59B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22918937, end_position_on_the_genomic_accession=22924742, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  318981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318981, CurrentID=0, Status='live', Symbol='CG31849', Aliases='Dmel_ BcDNA:LD34806, Dmel\\CG31849, dMGAT3', description='CG31849 gene product from transcript CG31849-RA', other_designations='CG31849-PA|uncharacterized protein', map_location='34A8-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13180398, end_position_on_the_genomic_accession=13182596, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  46121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46121, CurrentID=0, Status='live', Symbol='l(3)01239', Aliases='Dmel_CG6302, CG6302, Dmel\\CG6302, P01239, l(3)j9B4', description='lethal (3) 01239', other_designations='CG6302-PA|CG6302-PB|l(3)01239-PA|l(3)01239-PB', map_location='68A4-68A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11073292, end_position_on_the_genomic_accession=11074388, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  46140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46140, CurrentID=0, Status='live', Symbol='l(3)02640', Aliases='Dmel_CG9165, 2640, 3544, CG9165, Dmel\\CG9165, PBGD, anon-WO0153538.70, l(3)3544, l(3)7E130, l(3)j4E3', description='lethal (3) 02640', other_designations='CG9165-PA|hydroxymethylbilane synthase|l(3)02640-PA|porphobilinogen deaminase', map_location='61F6-61F7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1334889, end_position_on_the_genomic_accession=1339415, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  46155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46155, CurrentID=0, Status='live', Symbol='CG31195', Aliases='Dmel_ Dmel\\CG31195, l(3)03806', description='CG31195 gene product from transcript CG31195-RB', other_designations='CG31195-PB|uncharacterized protein', map_location='92F12-92F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20757269, end_position_on_the_genomic_accession=20762369, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  46156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46156, CurrentID=0, Status='live', Symbol='Gug', Aliases='Dmel_CG6964, Atro, CG6964, Dmel\\CG6964, GUG, atro, gug, l(3)01323, l(3)03928, l(3)S146907, l(3)j5A3, l(3)rO116', description='Grunge', other_designations='CG6964-PA|CG6964-PC|CG6964-PD|CG6964-PE|CG6964-PF|CG6964-PG|CG6964-PH|CG6964-PI|CG6964-PJ|Gug-PA|Gug-PC|Gug-PD|Gug-PE|Gug-PF|Gug-PG|Gug-PH|Gug-PI|Gug-PJ|atrophin|grunge', map_location='3-26 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8449922, end_position_on_the_genomic_accession=8472669, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  46158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46158, CurrentID=0, Status='live', Symbol='l(3)04053', Aliases='Dmel_CG11238, CG11238, Dmel\\CG11238, l(3)j8B2', description='lethal (3) 04053', other_designations='CG11238-PA|l(3)04053-PA', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22738602, end_position_on_the_genomic_accession=22741388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318993, CurrentID=0, Status='live', Symbol='CG31870', Aliases='Dmel_ BcDNA:AT11120, Dmel\\CG31870, NEST:bs19e04, NEST:bs28a02, bs19e04.y1, bs28a02.y1', description='CG31870 gene product from transcript CG31870-RD', other_designations='CG31870-PA|CG31870-PC|CG31870-PD|CG31870-PE|uncharacterized protein', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10790222, end_position_on_the_genomic_accession=10791450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  46187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46187, CurrentID=0, Status='live', Symbol='sinu', Aliases='Dmel_CG10624, 6524, CG10624, Dmel\\CG10624, Sinu, i8, l(3)06524', description='sinuous', other_designations='CG10624-PA|sinu-PA', map_location='64D3-64D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5559470, end_position_on_the_genomic_accession=5561598, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  46194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46194, CurrentID=0, Status='live', Symbol='Spn', Aliases='Dmel_CG16757, BcDNA:GM16129, CG16757, Dmel\\CG16757, E62, EL62, c62E-5, l(3)06911, spn', description='Spinophilin', other_designations='CG16757-PB|CG16757-PC|CG16757-PD|CG16757-PE|CG16757-PF|CG16757-PG|CG16757-PH|CG16757-PI|CG16757-PJ|CG16757-PS|CG16757-PT|Spn-PB|Spn-PC|Spn-PD|Spn-PE|Spn-PF|Spn-PG|Spn-PH|Spn-PI|Spn-PJ|Spn-PS|Spn-PT|d-spinophilin|spinophilin', map_location='62E4-62E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2505245, end_position_on_the_genomic_accession=2554292, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  46201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46201, CurrentID=0, Status='live', Symbol='l(3)07882', Aliases='Dmel_CG5824, CG5824, Dmel\\CG5824', description='lethal (3) 07882', other_designations='CG5824-PA|CG5824-PB|l(3)07882-PA|l(3)07882-PB', map_location='90B3-90B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17535533, end_position_on_the_genomic_accession=17539963, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  46234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46234, CurrentID=0, Status='live', Symbol='fbl', Aliases='Dmel_CG5725, CG5725, Dmel\\CG5725, Fbl, PANK, dPANK, dPANK/Fbl, dPanK', description='fumble', other_designations='CG5725-PA|CG5725-PB|CG5725-PC|CG5725-PD|CG5725-PE|CG5725-PF|CG5725-PG|fbl-PA|fbl-PB|fbl-PC|fbl-PD|fbl-PE|fbl-PF|fbl-PG|pantothenate kinase', map_location='77B9-77C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20384429, end_position_on_the_genomic_accession=20391913, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  46243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46243, CurrentID=0, Status='live', Symbol='poe', Aliases='Dmel_CG14472, 3420, CG14472, Calo, Dmel\\CG14472, POE, POE_DROME, Poe, calo, ms(2)01659, ms(2)27C, ms(2)3420, ms(2)5970, push', description='purity of essence', other_designations='CG14472-PA|CG14472-PB|calossin|male sterile(2)27C|male sterile(2)3420|poe-PA|poe-PB|pushover', map_location='28D11-28E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8054322, end_position_on_the_genomic_accession=8071624, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  46246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46246, CurrentID=0, Status='live', Symbol='Pglym87', Aliases='Dmel_CG17645, CG17645, Dmel\\CG17645, Pglym, pglym87', description='CG17645 gene product from transcript CG17645-RA', other_designations='CG17645-PA|Pglym87-PA', map_location='87B3-87B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12231645, end_position_on_the_genomic_accession=12232729, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  46260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46260, CurrentID=0, Status='live', Symbol='ND-SGDH', Aliases='Dmel_CG9762, CG9762, Dmel\\CG9762, NISM, l(3)neo18', description='NADH dehydrogenase (ubiquinone) SGDH subunit', other_designations='CG9762-PA|CG9762-PB|NADH ubiquinone oxidoreductase SGDH|NADH-ubiquinone oxidoreductase SGDH subunit|NADH:ubiquinone oxidoreductase SGDH subunit|ND-SGDH-PA|ND-SGDH-PB|lethal (3) neo18', map_location='68F5-68F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12137496, end_position_on_the_genomic_accession=12138584, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8674081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674081, CurrentID=0, Status='live', Symbol='CG33655', Aliases='Dmel_ BcDNA:AT03788, CR33655, Dmel\\CG33655, Dmel_CR33655, NEST:bs30e09', description='CG33655 gene product from transcript CG33655-RB', other_designations='CG33655-PB|CG33655-PC|CR33655|NEST:bs30e09|uncharacterized protein', map_location='57D11-57D11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21317793, end_position_on_the_genomic_accession=21320017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798158, CurrentID=0, Status='live', Symbol='CG43438', Aliases='Dmel_ Dmel\\CG43438, lincRNA.111', description='CG43438 gene product from transcript CG43438-RA', other_designations='CG43438-PA|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11536145, end_position_on_the_genomic_accession=11536844, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  46326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46326, CurrentID=0, Status='live', Symbol='l(2)dtl', Aliases='Dmel_CG11295, CG11295, Cdt2, DTL83, Dmel\\CG11295, L2DTL, cdt2', description='lethal-(2)-denticleless', other_designations='CG11295-PA|l(2)dtl-PA', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23702477, end_position_on_the_genomic_accession=23705403, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  46386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46386, CurrentID=0, Status='live', Symbol='CG13397', Aliases='Dmel_ 172F5T, Dmel\\CG13397, ESTS:172F5T, anon-sts19', description='CG13397 gene product from transcript CG13397-RB', other_designations='CG13397-PA|CG13397-PB|uncharacterized protein', map_location='29C3-29C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8408888, end_position_on_the_genomic_accession=8411868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319027, CurrentID=0, Status='live', Symbol='CG31921', Aliases='Dmel_ Dmel\\CG31921, NEST:bs08h12', description='CG31921 gene product from transcript CG31921-RA', other_designations='CG31921-PA|uncharacterized protein', map_location='21C1-21C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=329910, end_position_on_the_genomic_accession=332335, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  46391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46391, CurrentID=0, Status='live', Symbol='CG11899', Aliases='Dmel_ 39C10S, Dmel\\CG11899, ESTS:39C10S, Ov8, anon-sts29', description='CG11899 gene product from transcript CG11899-RA', other_designations='CG11899-PA|uncharacterized protein', map_location='99A1-99A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29200454, end_position_on_the_genomic_accession=29201770, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  46415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46415, CurrentID=0, Status='live', Symbol='Fer1HCH', Aliases='Dmel_CG2216, 1HCH, BEST:LD36673, CG2216, Dmel\\CG2216, FER1, FER1HCH, Fe1HCH, Fer-L, Fer1, Fer1HC, FerHCH, HCH, T21, T9, fer1, fer1hch, l(3)00451, l(3)j10B4', description='Ferritin 1 heavy chain homologue', other_designations='CG2216-PA|CG2216-PB|CG2216-PC|CG2216-PD|CG2216-PE|CG2216-PF|CG2216-PG|CG2216-PI|CG2216-PJ|Fer1HCH-PA|Fer1HCH-PB|Fer1HCH-PC|Fer1HCH-PD|Fer1HCH-PE|Fer1HCH-PF|Fer1HCH-PG|Fer1HCH-PI|Fer1HCH-PJ|Ferritin-Heavy|Ferritin-Light|ferritin|ferritin 1 heavy chain|ferritin 1 heavy chain homolog|ferritin 1 heavy chain homologue|ferritin H|ferritin HCH|ferritin heavy chain|ferritin subunit 1', map_location='99F2-99F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30385573, end_position_on_the_genomic_accession=30388178, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  46456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46456, CurrentID=0, Status='live', Symbol='CanB2', Aliases='Dmel_CG11217, CG11217, CS2-1, Dmel\\CG11217, canB2, dCNB2, dCnB2', description='Calcineurin B2', other_designations='CG11217-PA|CG11217-PB|CanB2-PA|CanB2-PB|calcineurin B2|calcineurin suppressor, chromosome 2, group 1', map_location='43E16-43E16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7786010, end_position_on_the_genomic_accession=7789053, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  7354405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354405, CurrentID=0, Status='live', Symbol='CG42391', Aliases='Dmel_ Dmel\\CG42391', description='CG42391 gene product from transcript CG42391-RA', other_designations='CG42391-PA|uncharacterized protein', map_location='51E5-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15202193, end_position_on_the_genomic_accession=15202987, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740490, CurrentID=0, Status='live', Symbol='CG34210', Aliases='Dmel_ Dmel\\CG34210', description='CG34210 gene product from transcript CG34210-RA', other_designations='CG34210-PA|uncharacterized protein', map_location='59D7-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23351518, end_position_on_the_genomic_accession=23352144, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  46662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46662, CurrentID=0, Status='live', Symbol='l(2)09851', Aliases='Dmel_CG12792, CG12792, Dmel\\CG12792, GRWD1, I(2)09851, l(2)k08138', description='lethal (2) 09851', other_designations='CG12792-PA|CG12792-PB|l(2)09851-PA|l(2)09851-PB', map_location='41F9-41F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5754517, end_position_on_the_genomic_accession=5756327, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  35089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35089, CurrentID=0, Status='live', Symbol='CG7200', Aliases='Dmel_ Dmel\\CG7200', description='CG7200 gene product from transcript CG7200-RB', other_designations='CG7200-PA|CG7200-PB', map_location='36E6-36E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18157110, end_position_on_the_genomic_accession=18159272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  46717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46717, CurrentID=0, Status='live', Symbol='arg', Aliases='Dmel_CG18104, CG18104, Dmel\\CG18104, EG:171D11.4, EG:65F1.3', description='arginase', other_designations='CG18104-PA|arg-PA', map_location='1B7-1B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=509435, end_position_on_the_genomic_accession=528028, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  46719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46719, CurrentID=0, Status='live', Symbol='dlt', Aliases='Dmel_CG32315, CG1977, CG32315, DLT, Dlt, Dmel\\CG32315, Van, dre1, l(3)04276, l(3)62Ba, l(3)dre1', description='discs lost', other_designations='CG32315-PA|dlt-PA|lethal(3)62Ba|vanaso', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1789631, end_position_on_the_genomic_accession=1795328, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  46813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=46813, CurrentID=0, Status='live', Symbol='CG6421', Aliases='Dmel_ Dmel\\CG6421, EG:EG0003.8', description='CG6421 gene product from transcript CG6421-RA', other_designations='CG6421-PA', map_location='53D15-53D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16872250, end_position_on_the_genomic_accession=16872906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35113, CurrentID=0, Status='live', Symbol='Atac2', Aliases='Dmel_CG10414, CG10414, Dmel\\CG10414, KAT14, dATAC2', description='Ada2a-containing complex component 2', other_designations='ATAC complex component 2|Atac2-PA|CG10414-PA|lysine acetyltransferase 14', map_location='36F7-36F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18609576, end_position_on_the_genomic_accession=18612249, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36724, CurrentID=0, Status='live', Symbol='CG12963', Aliases='Dmel_ BcDNA:AT28470, Dmel\\CG12963', description='CG12963 gene product from transcript CG12963-RL', other_designations='CG12963-PA|CG12963-PC|CG12963-PD|CG12963-PE|CG12963-PF|CG12963-PG|CG12963-PH|CG12963-PI|CG12963-PJ|CG12963-PK|CG12963-PL|CG12963-PM|CG12963-PN|uncharacterized protein', map_location='52A11-52A12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15555446, end_position_on_the_genomic_accession=15562436, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  38662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38662, CurrentID=0, Status='live', Symbol='CG5592', Aliases='Dmel_ Dmel\\CG5592', description='CG5592 gene product from transcript CG5592-RA', other_designations='CG5592-PA', map_location='64F4-64F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5897638, end_position_on_the_genomic_accession=5899395, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  47065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47065, CurrentID=0, Status='live', Symbol='qsm', Aliases='Dmel_CG13432, CG13432, CG13433, Dmel\\CG13432, EP2356, EP2587, L(2)05510, QSM, l(2)05510, l(2)k009920, lethal (2) 05510', description='quasimodo', other_designations='CG13432-PA|CG13432-PB|lethal (2) 05510|qsm-PA|qsm-PB', map_location='57A5-57A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20563034, end_position_on_the_genomic_accession=20583591, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  47080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47080, CurrentID=0, Status='live', Symbol='acj6', Aliases='Dmel_CG9151, ACJ6, Acj6, CG9151, Dmel\\CG9151, I-POU, Ipou, tI-POU', description='abnormal chemosensory jump 6', other_designations='CG9151-PA|CG9151-PC|CG9151-PD|CG9151-PE|CG9151-PF|CG9151-PG|CG9151-PH|CG9151-PI|CG9151-PJ|CG9151-PK|CG9151-PL|CG9151-PM|CG9151-PN|abnormal chemosensory jumb|acj6-PA|acj6-PC|acj6-PD|acj6-PE|acj6-PF|acj6-PG|acj6-PH|acj6-PI|acj6-PJ|acj6-PK|acj6-PL|acj6-PM|acj6-PN|altered chemosensory jump 6|twin of I-POU', map_location='1-49.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15350239, end_position_on_the_genomic_accession=15379064, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  47082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47082, CurrentID=0, Status='live', Symbol='Adf1', Aliases='Dmel_CG15845, Adf 1, Adf-1, CG15845, Dmel\\CG15845, adf-1, adf1, l(2)01349, l(2)04065, nal', description='Adh transcription factor 1', other_designations='Adf1-PA|Adf1-PB|Adf1-PC|Adf1-PE|CG15845-PA|CG15845-PB|CG15845-PC|CG15845-PE|nalyot|naylot', map_location='42C3-42C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6662265, end_position_on_the_genomic_accession=6668085, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  47091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47091, CurrentID=0, Status='live', Symbol='comt', Aliases='Dmel_CG1618, CG 1618, CG1618, Comt, DmNSF, Dmel\\CG1618, NSF, NSF-1, NSF1, Nsf, com, dNSF, dNSF-1, dNSF1, dNsf1, dnsf1', description='comatose', other_designations='CG1618-PA|CG1618-PB|N-ethylmaleimide-sensitive fusion protein|NEM-sensitive fusion protein|comt-PA|comt-PB', map_location='1-40.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13259021, end_position_on_the_genomic_accession=13262488, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  7354371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354371, CurrentID=0, Status='live', Symbol='CG42374', Aliases='Dmel_ Dmel\\CG42374, Dromel_CG9666_FBtr0074991_uORF', description='CG42374 gene product from transcript CG42374-RB', other_designations='CG42374-PA|CG42374-PB|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19263766, end_position_on_the_genomic_accession=19266838, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  7354373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354373, CurrentID=0, Status='live', Symbol='CG42357', Aliases='Dmel_ CG32308, Dmel\\CG42357, Dmel_CG32308', description='CG42357 gene product from transcript CG42357-RA', other_designations='CG42357-PA|uncharacterized protein', map_location='62C1-62C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2020554, end_position_on_the_genomic_accession=2021958, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354374, CurrentID=0, Status='live', Symbol='CG42356', Aliases='Dmel_ CG32308, Dmel\\CG42356', description='CG42356 gene product from transcript CG42356-RA', other_designations='CG42356-PA|uncharacterized protein', map_location='62C2-62C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2022070, end_position_on_the_genomic_accession=2023391, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354375, CurrentID=0, Status='live', Symbol='lectin-22C', Aliases='Dmel_CG42295, AC 004716A, AC004716a, CG15378, CG42295, Dmel\\CG42295, Dmel_CG15378', description='CG42295 gene product from transcript CG42295-RB', other_designations='CG42295-PB|Lectin22C|lectin|lectin 22C|lectin-22C-PB', map_location='22C1-22C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2036409, end_position_on_the_genomic_accession=2037265, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7354376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354376, CurrentID=0, Status='live', Symbol='tal-2A', Aliases='Dmel_CG42385, CG42385, Dm2, Dmel\\CG42385, ORF 2A, ORF2, Pep2, pri, tal', description='CG42385 gene product from transcript CG42385-RA', other_designations='CG42385-PA|tal-2A-PA', map_location='87F14-87F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13813109, end_position_on_the_genomic_accession=13814648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354377, CurrentID=0, Status='live', Symbol='tal-AA', Aliases='Dmel_CG42387, CG42387, Dm4, Dm_3R:52381, Dmel\\CG42387, ORF AA, ORF4, Pep4, pri, tal', description='CG42387 gene product from transcript CG42387-RA', other_designations='CG42387-PA|tal-AA-PA', map_location='87F14-87F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13813109, end_position_on_the_genomic_accession=13814648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354378, CurrentID=0, Status='live', Symbol='tal-3A', Aliases='Dmel_CG42386, CG42386, Dm3, Dmel\\CG42386, ORF 3A, ORF3, Pep3, pri, tal', description='CG42386 gene product from transcript CG42386-RA', other_designations='CG42386-PA|tal-3A-PA', map_location='87F14-87F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13813109, end_position_on_the_genomic_accession=13814648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354379, CurrentID=0, Status='live', Symbol='CG42372', Aliases='Dmel_ Dmel\\CG42372, Dromel_CG30100_FBtr0087140_uORF', description='CG42372 gene product from transcript CG42372-RA', other_designations='CG42372-PA|CG42372-PC|uncharacterized protein', map_location='53B1-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16291453, end_position_on_the_genomic_accession=16292269, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354380, CurrentID=0, Status='live', Symbol='CG42375', Aliases='Dmel_ CG9288, Dmel\\CG42375, Dromel_CG9288_FBtr0082829_pyruvatekinasedomain_uORF', description='CG42375 gene product from transcript CG42375-RA', other_designations='CG42375-PA|uncharacterized protein', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13779624, end_position_on_the_genomic_accession=13780795, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354381, CurrentID=0, Status='live', Symbol='tal-1A', Aliases='Dmel_CG42384, CG42384, Dm1, Dmel\\CG42384, ORF 1A, ORF1, Pep1, Tal1A, pri, tal', description='CG42384 gene product from transcript CG42384-RB', other_designations='CG42384-PB|tal-1A-PB', map_location='87F14-87F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13813109, end_position_on_the_genomic_accession=13814648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354382, CurrentID=0, Status='live', Symbol='CG42305', Aliases='Dmel_ Dmel\\CG42305, Dromel_CG17325_FBtr0081102_uORF', description='CG42305 gene product from transcript CG42305-RC', other_designations='CG42305-PA|CG42305-PB|CG42305-PC|uncharacterized protein', map_location='37A4-37A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18812648, end_position_on_the_genomic_accession=18813941, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47121, CurrentID=0, Status='live', Symbol='dup', Aliases='Dmel_CG8171, 19k, CDT1, CDT1/DUP, CG8171, Cdt1, DUP, DUP/CDT1, Dmel\\CG8171, Dup, anon-WO0257455.31, cdt1, cdt1/dup, fs(2)51Fa, fs(2)PA77, fs(2)ltoPA77, l(2)51Ec, l(2)k03308', description='double parked', other_designations='CG8171-PA|CG8171-PB|double park|double-parked|doubleparked|dup-PA|dup-PB|female sterile(2)51Fa|lethal(2)51Ec|speckled', map_location='51F11-51F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15376619, end_position_on_the_genomic_accession=15380988, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7354398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354398, CurrentID=0, Status='live', Symbol='CG42238', Aliases='Dmel_ CG9266, Dmel\\CG42238', description='CG42238 gene product from transcript CG42238-RC', other_designations='CG42238-PA|CG42238-PB|CG42238-PC|CG42238-PD|CG42238-PE|uncharacterized protein', map_location='38F6-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20965520, end_position_on_the_genomic_accession=21052164, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  7354399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354399, CurrentID=0, Status='live', Symbol='CG17454', Aliases='Dmel_ Dmel\\CG17454, SPF30', description='CG17454 gene product from transcript CG17454-RA', other_designations='CG17454-PA|CG17454-PB|uncharacterized protein', map_location='80F9-80F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23715023, end_position_on_the_genomic_accession=23716689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  7354400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354400, CurrentID=0, Status='live', Symbol='Ir60b', Aliases='Dmel_CG42289, CG13582, CG42289, DmelIR60b, Dmel\\CG42289, Dmel_CG13582, IR60b', description='Ionotropic receptor 60b', other_designations='CG42289-PA|Ir60b-PA|ionotropic receptor 60b', map_location='60D3-60D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24515685, end_position_on_the_genomic_accession=24517458, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  47137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47137, CurrentID=0, Status='live', Symbol='knk', Aliases='Dmel_CG6217, CG6217, Dmel\\CG6217, Gnarled, Knk, dyspneic, gnar, l(3)85Eg', description='knickkopf', other_designations='CG6217-PA|dyspneic|gnarled|knk-PA', map_location='3-49.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10122365, end_position_on_the_genomic_accession=10125664, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  7354402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354402, CurrentID=0, Status='live', Symbol='Ir60d', Aliases='Dmel_CG42291, CG13582, CG42291, DmelIR60d, Dmel\\CG42291, Dmel_CG13582, IR60d', description='Ionotropic receptor 60d', other_designations='CG42291-PB|Ir60d-PB|ionotropic receptor 60d', map_location='60D3-60D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24519423, end_position_on_the_genomic_accession=24521147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354403, CurrentID=0, Status='live', Symbol='Tfb5', Aliases='Dmel_CG42373, CG14037, CG42373, Dmel\\CG42373, Dmp8, Dromel_CG31917_FBtr0079037_uORF, TFB5, TFB5 (uORF), TTDA, l(2)SH1279, l(2)SH2 1279, lincRNA.843, p8, uORF', description='TFB5 ortholog (S. cerevisiae)', other_designations='CG42373-PB|TFB5 ortholog|Tfb5-PB|lethal (2) SH1279', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5009931, end_position_on_the_genomic_accession=5010834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354404, CurrentID=0, Status='live', Symbol='Yeti', Aliases='Dmel_CG40218, 41Aa, BEST:GH01620, BcDNA:RE36623, CG40218, Dmel\\CG40218, GroupI, l(2)41Aa, l(2)EMS31, l(2R)B, l(2R)EMS-31', description='CG40218 gene product from transcript CG40218-RB', other_designations='CG40218-PA|CG40218-PB|Yeti-PA|Yeti-PB|lethal (2) 41Aa|yeti', map_location='2-55.1 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=1343403, end_position_on_the_genomic_accession=1345119, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  47141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47141, CurrentID=0, Status='live', Symbol='ebo', Aliases='Dmel_CG3923, CG3923, Dmel\\CG3923, EBO, EG:165H7.3, Exp1, Exp6', description='ellipsoid body open', other_designations='CG3923-PA|CG3923-PB|CG3923-PC|Exportin-6|ebo-PA|ebo-PB|ebo-PC|exportin 6', map_location='1-3.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=471019, end_position_on_the_genomic_accession=476579, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  7354406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354406, CurrentID=0, Status='live', Symbol='CG42363', Aliases='Dmel_ CG15675, Dmel\\CG42363, Dmel_CG15675', description='CG42363 gene product from transcript CG42363-RB', other_designations='CG42363-PA|CG42363-PB|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21664126, end_position_on_the_genomic_accession=21666108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354407, CurrentID=0, Status='live', Symbol='CG42362', Aliases='Dmel_ CG15675, Dmel\\CG42362, Dmel_CG15675', description='CG42362 gene product from transcript CG42362-RB', other_designations='CG42362-PA|CG42362-PB|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21664126, end_position_on_the_genomic_accession=21666108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354408, CurrentID=0, Status='live', Symbol='CG42364', Aliases='Dmel_ CG15675, Dmel\\CG42364, Dmel_CG15675', description='CG42364 gene product from transcript CG42364-RB', other_designations='CG42364-PB|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21665940, end_position_on_the_genomic_accession=21666579, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354409, CurrentID=0, Status='live', Symbol='CG42381', Aliases='Dmel_ CG9865(uORF1), Dmel\\CG42381, Dromel_CG9865_FBtr0071680_cx9c_uorf1', description='CG42381 gene product from transcript CG42381-RA', other_designations='CG42381-PA|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21667541, end_position_on_the_genomic_accession=21669916, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7354410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354410, CurrentID=0, Status='live', Symbol='CG42379', Aliases='Dmel_ CG9865(uORF2), Dmel\\CG42379, Dromel_CG9865_FBtr0071680_uorf2', description='CG42379 gene product from transcript CG42379-RA', other_designations='CG42379-PA|CG42379-PB|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21667541, end_position_on_the_genomic_accession=21669916, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7354411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354411, CurrentID=0, Status='live', Symbol='CG42380', Aliases='Dmel_ CG9865(uORF3), Dmel\\CG42380, Dromel_CG9865_FBtr0071680_cx9c_uorf3', description='CG42380 gene product from transcript CG42380-RA', other_designations='CG42380-PA|CG42380-PB|uncharacterized protein', map_location='57F7-57F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21667541, end_position_on_the_genomic_accession=21669916, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7354412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354412, CurrentID=0, Status='live', Symbol='Ir94d', Aliases='Dmel_CG42297, CG17379, CG42297, DmelIR94d, Dmel\\CG42297, Dmel_CG17379, IR94d', description='Ionotropic receptor 94d', other_designations='CG42297-PA|Ir94d-PA|ionotropic receptor 94d', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23402808, end_position_on_the_genomic_accession=23405139, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354413, CurrentID=0, Status='live', Symbol='CG42263', Aliases='Dmel_ Dmel\\CG42263', description='CG42263 gene product from transcript CG42263-RB', other_designations='CG42263-PB|uncharacterized protein', map_location='76D6-76D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19954547, end_position_on_the_genomic_accession=19955799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  7354414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354414, CurrentID=0, Status='live', Symbol='CG42308', Aliases='Dmel_ CG14428, Dmel\\CG42308', description='CG42308 gene product from transcript CG42308-RA', other_designations='CG42308-PA|CG42308-PB|uncharacterized protein', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7011357, end_position_on_the_genomic_accession=7012137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7354415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354415, CurrentID=0, Status='live', Symbol='CG42246', Aliases='Dmel_ Dm_X:38130, Dmel\\CG42246', description='CG42246 gene product from transcript CG42246-RA', other_designations='CG42246-PA|CG42246-PB|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7411267, end_position_on_the_genomic_accession=7411906, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354416, CurrentID=0, Status='live', Symbol='Ir7d', Aliases='Dmel_CG42294, CG15327, CG42294, DmelIR7d, Dmel\\CG42294, Dmel_CG15327, IR7d', description='Ionotropic receptor 7d', other_designations='CG42294-PA|Ir7d-PA|ionotropic receptor 7d', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7876963, end_position_on_the_genomic_accession=7878836, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354417, CurrentID=0, Status='live', Symbol='Ir7e', Aliases='Dmel_CG42293, CG15327, CG42293, DmelIR7e, Dmel\\CG42293, Dmel_CG15327, IR7e', description='Ionotropic receptor 7e', other_designations='CG42293-PA|Ir7e-PA|ionotropic receptor 7e', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7879800, end_position_on_the_genomic_accession=7881679, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354418, CurrentID=0, Status='live', Symbol='Ir7f', Aliases='Dmel_CG42292, CG15327, CG42292, DmelIR7f, Dmel\\CG42292, Dmel_CG15327, IR7f', description='Ionotropic receptor 7f', other_designations='CG42292-PA|Ir7f-PA|ionotropic receptor 7f', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7881939, end_position_on_the_genomic_accession=7884026, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7354420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354420, CurrentID=0, Status='live', Symbol='CG42376', Aliases='Dmel_ Dmel\\CG42376', description='CG42376 gene product from transcript CG42376-RA', other_designations='CG42376-PA|CG42376-PB|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=254876, end_position_on_the_genomic_accession=255524, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  7354421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354421, CurrentID=0, Status='live', Symbol='Nsun5', Aliases='Dmel_CG42358, CG42358, CG5558, Dmel\\CG42358, Dmel_CG5558, dNsun5', description='Nop2/Sun-like domain containing protein 5', other_designations='CG42358-PA|Nsun5-PA', map_location='91F4-91F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19160099, end_position_on_the_genomic_accession=19162300, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  7354422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354422, CurrentID=0, Status='live', Symbol='CG42239', Aliases='Dmel_ Dmel\\CG42239', description='CG42239 gene product from transcript CG42239-RA', other_designations='CG42239-PA|uncharacterized protein', map_location='54C9-54C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17572829, end_position_on_the_genomic_accession=17573330, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354423, CurrentID=0, Status='live', Symbol='CG42307', Aliases='Dmel_ Dmel\\CG42307', description='CG42307 gene product from transcript CG42307-RD', other_designations='CG42307-PA|CG42307-PB|CG42307-PC|CG42307-PD|uncharacterized protein', map_location='65F4-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7329132, end_position_on_the_genomic_accession=7333471, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  7354424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354424, CurrentID=0, Status='live', Symbol='CG42394', Aliases='Dmel_ Dmel\\CG42394', description='CG42394 gene product from transcript CG42394-RC', other_designations='CG42394-PA|CG42394-PB|CG42394-PC|CG42394-PD|uncharacterized protein', map_location='86C7-86C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10849825, end_position_on_the_genomic_accession=10851788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7354425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354425, CurrentID=0, Status='live', Symbol='CG42306', Aliases='Dmel_ Dmel\\CG42306, Dromel_CG5469_FBtr0086654_gint3_uORF', description='CG42306 gene product from transcript CG42306-RA', other_designations='CG42306-PA|CG42306-PB|CG42306-PC|uncharacterized protein', map_location='55E5-55E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18658418, end_position_on_the_genomic_accession=18661548, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7354426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354426, CurrentID=0, Status='live', Symbol='CG42300', Aliases='Dmel_ CG18620, CG32584, Dmel\\CG42300, Dmel_CG32584', description='CG42300 gene product from transcript CG42300-RA', other_designations='CG42300-PA|uncharacterized protein', map_location='13D2-13D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15555253, end_position_on_the_genomic_accession=15556045, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354427, CurrentID=0, Status='live', Symbol='CG42287', Aliases='Dmel_ CG13353, Dmel\\CG42287, Dmel_CG13353', description='CG42287 gene product from transcript CG42287-RA', other_designations='CG42287-PA|uncharacterized protein', map_location='50D5-50D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14090325, end_position_on_the_genomic_accession=14090972, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7354428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354428, CurrentID=0, Status='live', Symbol='CG42245', Aliases='Dmel_ Dmel\\CG42245', description='CG42245 gene product from transcript CG42245-RA', other_designations='CG42245-PA|uncharacterized protein', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2631715, end_position_on_the_genomic_accession=2641984, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  7354429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354429, CurrentID=0, Status='live', Symbol='Snup', Aliases='Dmel_CG42303, CG1247, CG32297, CG42303, Dmel\\CG42303, Dmel_CG32297, dSNUP', description='Snurportin', other_designations='CG42303-PA|Snup-PA|snurportin', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2652088, end_position_on_the_genomic_accession=2655455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  7354430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354430, CurrentID=0, Status='live', Symbol='CG42392', Aliases='Dmel_ Dmel\\CG42392', description='CG42392 gene product from transcript CG42392-RA', other_designations='CG42392-PA|uncharacterized protein', map_location='53D7-53D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16765361, end_position_on_the_genomic_accession=16766000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7354431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354431, CurrentID=0, Status='live', Symbol='CG42266', Aliases='Dmel_ BG:DS02740.19, CG18115, CG31737, CG31818, DS02740.19, Dmel\\CG42266, Dmel_CG31737, Dmel_CG31818', description='CG42266 gene product from transcript CG42266-RA', other_designations='CG42266-PA|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16337989, end_position_on_the_genomic_accession=16340368, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7354432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354432, CurrentID=0, Status='live', Symbol='CG42231', Aliases='Dmel_ Dmel\\CG42231', description='CG42231 gene product from transcript CG42231-RA', other_designations='CG42231-PA|uncharacterized protein', map_location='35F1-35F12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16342430, end_position_on_the_genomic_accession=16352628, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  7354433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354433, CurrentID=0, Status='live', Symbol='DIP-epsilon', Aliases='Dmel_CG42368, 42368, CG13981, CG33295, CG42368, CG9508, CG9511, Dmel\\CG42368, Dmel_CG9508, Dmel_CG9511', description='Dpr-interacting protein epsilon', other_designations='CG42368-PA|DIP-epsilon-PA', map_location='26C4-26D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6377771, end_position_on_the_genomic_accession=6411326, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  7354434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354434, CurrentID=0, Status='live', Symbol='CG42265', Aliases='Dmel_ CG15776, CG15777, CG17758, Dmel\\CG42265, Dmel_CG15776, Dmel_CG15777, Dmel_CG17758', description='CG42265 gene product from transcript CG42265-RE', other_designations='CG42265-PD|CG42265-PE|CG42265-PF|uncharacterized protein', map_location='5A10-5A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5650567, end_position_on_the_genomic_accession=5662861, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7354435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354435, CurrentID=0, Status='live', Symbol='CG42302', Aliases='Dmel_ CG6857, Dmel\\CG42302, Dmel_CG6857', description='CG42302 gene product from transcript CG42302-RA', other_designations='CG42302-PA|uncharacterized protein', map_location='17D3-17D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18724914, end_position_on_the_genomic_accession=18725672, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7354436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354436, CurrentID=0, Status='live', Symbol='CG42395', Aliases='Dmel_ Dmel\\CG42395', description='CG42395 gene product from transcript CG42395-RD', other_designations='CG42395-PD|uncharacterized protein', map_location='8E10-8E10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9575126, end_position_on_the_genomic_accession=9576247, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  47173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47173, CurrentID=0, Status='live', Symbol='Men', Aliases='Dmel_CG10120, CG10120, Dmel\\CG10120, MDH, ME, MEN, Mdh-NADP, Me, anon-WO0118547.278, mem, men', description='Malic enzyme', other_designations='CG10120-PA|CG10120-PB|CG10120-PC|Men-PA|Men-PB|Men-PC|NADP-dependent malate dehydrogenase|NADP-malic enzyme|malic enzyme', map_location='3-51.73 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12713103, end_position_on_the_genomic_accession=12722442, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  7354447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354447, CurrentID=0, Status='live', Symbol='SteXh:CG42398', Aliases='Dmel_CG42398, CG42398, Dmel\\CG42398, RR41752', description='CG42398 gene product from transcript CG42398-RA', other_designations='CG42398-PA|SteXh:CG42398-PA', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  47186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47186, CurrentID=0, Status='live', Symbol='mus205', Aliases='Dmel_CG1925, 154926_at, AAF59191, CG1925, DhRev3, DmREV3, Dmel\\CG1925, Dmpolzeta, Mus205, mus 205, mus(2)205, mus-205, rev3', description='mutagen-sensitive 205', other_designations='CG1925-PA|mus205-PA', map_location='2-54.9 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7793625, end_position_on_the_genomic_accession=7801044, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  47198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47198, CurrentID=0, Status='live', Symbol='Sgs7', Aliases='Dmel_CG18087, 68C GIII, CG18087, Dmel\\CG18087, Sgs-7, sgs-7, sgs7', description='Salivary gland secretion 7', other_designations='CG18087-PA|Sgs7-PA|gene III|group III|salivary gland secretion 7', map_location='3-35.0 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11509861, end_position_on_the_genomic_accession=11510247, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354465, CurrentID=0, Status='live', Symbol='CG42232', Aliases='Dmel_ CG14896, CG14897, Dmel\\CG42232, Dmel_CG14896, Dmel_CG14897, anon-EST:Liang-1.14, clone 1.14', description='CG42232 gene product from transcript CG42232-RB', other_designations='CG42232-PB|uncharacterized protein', map_location='89B17-89B17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16458280, end_position_on_the_genomic_accession=16469663, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  7354466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354466, CurrentID=0, Status='live', Symbol='CG42342', Aliases='Dmel_ CG11769, CG14888, CG14889, CG14890, CG42243, Dmel\\CG42342, Dmel_CG11769, Dmel_CG14888, Dmel_CG14889, Dmel_CG14890', description='CG42342 gene product from transcript CG42342-RH', other_designations='CG42342-PD|CG42342-PH|CG42342-PK|CG42342-PL|CG42342-PM|CG42342-PN|CG42342-PO|CG42342-PP|CG42342-PQ|CG42342-PR|CG42342-PS|CG42342-PT|uncharacterized protein', map_location='89B20-89B20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16502770, end_position_on_the_genomic_accession=16571363, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  7354470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354470, CurrentID=0, Status='live', Symbol='CG42402', Aliases='Dmel_ CG12453, CG41358, Dmel\\CG42402, Dmel_CG12453, Dmel_CG41358', description='CG42402 gene product from transcript CG42402-RD', other_designations='CG42402-PC|CG42402-PD|CG42402-PE|uncharacterized protein', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4077939, end_position_on_the_genomic_accession=4104326, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  7354471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354471, CurrentID=0, Status='live', Symbol='CG42240', Aliases='Dmel_ CG15893, Dmel\\CG42240', description='CG42240 gene product from transcript CG42240-RC', other_designations='CG42240-PA|CG42240-PB|CG42240-PC|uncharacterized protein', map_location='5F1-5F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6330842, end_position_on_the_genomic_accession=6333028, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  47208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47208, CurrentID=0, Status='live', Symbol='term', Aliases='Dmel_CG4216, Bsg75C, CG4216, Dmel\\CG4216, Term, ter', description='terminus', other_designations='CG4216-PA|term-PA', map_location='75D2-75D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18592303, end_position_on_the_genomic_accession=18593771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  7354473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354473, CurrentID=0, Status='live', Symbol='moi', Aliases='Dmel_CG42350, 0967/13, CG31241, CG42350, DTL, DTLu, Dmel\\CG42350, Dromel_CG31241_FBtr0083570_dtl_uORF, Dtl uORF, l(3)CB02140, l(3)S096713, tlu, uORF', description='modigliani', other_designations='CG42350-PA|lethal (3) S096713|moi-PA|upstream ORF of tat-like', map_location='90F11-90F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18255530, end_position_on_the_genomic_accession=18257677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354474, CurrentID=0, Status='live', Symbol='CG42393', Aliases='Dmel_ Dmel\\CG42393', description='CG42393 gene product from transcript CG42393-RA', other_designations='CG42393-PA|uncharacterized protein', map_location='75B5-75B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18042733, end_position_on_the_genomic_accession=18044214, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  7354475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354475, CurrentID=0, Status='live', Symbol='CG42259', Aliases='Dmel_ C, Dmel\\CG42259, anon-1Ea', description='CG42259 gene product from transcript CG42259-RA', other_designations='CG42259-PA|CG42259-PB|anon-1Ea|transcription unit C|uncharacterized protein', map_location='1E1-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1046224, end_position_on_the_genomic_accession=1047009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354476, CurrentID=0, Status='live', Symbol='CG42371', Aliases='Dmel_ Dmel\\CG42371', description='CG42371 gene product from transcript CG42371-RA', other_designations='CG42371-PA|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2216503, end_position_on_the_genomic_accession=2217246, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  47216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47216, CurrentID=0, Status='live', Symbol='Klp54D', Aliases='Dmel_CG43323, CG15844, CG43323, DmKlp54D, Dmel\\CG43323, Dmel_CG15844, KIF 12, KIF12, KLP1, KLP54D, Klp1', description='Kinesin-like protein at 54D', other_designations='CG43323-PA|CG43323-PB|CG43323-PC|CG43323-PD|CG43323-PE|Kinesin-like protein 54D|Klp54D-PA|Klp54D-PB|Klp54D-PC|Klp54D-PD|Klp54D-PE', map_location='54C5-54C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17557031, end_position_on_the_genomic_accession=17563406, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  47218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47218, CurrentID=0, Status='live', Symbol='RpS14a', Aliases='Dmel_CG1524, CG1524, Dmel\\CG1524, RPS14A, Rp S14a, RpS14, RpS14A, S14, S14a, anon-EST:Posey115, anon-EST:Posey131, anon-EST:Posey154, anon-EST:Posey77, anon-WO03040301.136, anon-WO03040301.138, rpS14A', description='Ribosomal protein S14a', other_designations='40S ribosomal protein S14|CG1524-PA|CG1524-PB|RpS14a-PA|RpS14a-PB|ribosomal protein S14a', map_location='7C6-7C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7933742, end_position_on_the_genomic_accession=7934677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47219, CurrentID=0, Status='live', Symbol='RpS14b', Aliases='Dmel_CG1527, CG1527, Dmel\\CG1527, RPS14B, RpS14, RpS14B, S14, S14b, anon-EST:Posey115, anon-EST:Posey131, anon-EST:Posey154, anon-EST:Posey77, rpS14B', description='Ribosomal protein S14b', other_designations='CG1527-PA|CG1527-PB|RpS14b-PA|RpS14b-PB|ribosomal protein S14b', map_location='7C8-7C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7934898, end_position_on_the_genomic_accession=7935792, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  47220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47220, CurrentID=0, Status='live', Symbol='Fur1', Aliases='Dmel_CG10772, CG10772, Dfur1, Dfurin1, Dmel\\CG10772, Furin, Furl, Klip1, dFur1, dFurin 1, dKLIP-1, fur-1, fur1, fur[1], furin, l(3)rL205', description='Furin 1', other_designations='CG10772-PA|CG10772-PB|CG10772-PC|CG10772-PD|CG10772-PF|CG10772-PG|CG10772-PH|CG10772-PI|CG10772-PJ|Dfurin1|Fur1-PA|Fur1-PB|Fur1-PC|Fur1-PD|Fur1-PF|Fur1-PG|Fur1-PH|Fur1-PI|Fur1-PJ|Furin-like protease|Furin1|furin 1', map_location='96D2-96E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25347431, end_position_on_the_genomic_accession=25473058, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  5740766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740766, CurrentID=0, Status='live', Symbol='CG34313', Aliases='Dmel_ CG4826, Dmel\\CG34313', description='CG34313 gene product from transcript CG34313-RA', other_designations='CG34313-PA|uncharacterized protein', map_location='36B3-36B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16879517, end_position_on_the_genomic_accession=16886319, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  47249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47249, CurrentID=0, Status='live', Symbol='woc', Aliases='Dmel_CG5965, CG5965, Dmel\\CG5965, WOC, Woc', description='without children', other_designations='CG5965-PA|CG5965-PB|CG5965-PC|CG5965-PD|CG5965-PE|with out children|woc-PA|woc-PB|woc-PC|woc-PD|woc-PE', map_location='97E11-97E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27256595, end_position_on_the_genomic_accession=27264055, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  47251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47251, CurrentID=0, Status='live', Symbol='Gtp-bp', Aliases='Dmel_CG2522, BEST:CK00309, CG2522, CK00309, Dmel\\CG2522, GTP-bp, SrpRalpha, mAb 13D5', description='GTP-binding protein', other_designations='CG2522-PA|GTP binding protein|Gtp-bp-PA', map_location='10A6-10A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11126877, end_position_on_the_genomic_accession=11129087, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  47253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47253, CurrentID=0, Status='live', Symbol='dmGlut', Aliases='Dmel_CG5304, CG5304, DMGLUT, Dmel\\CG5304, I(2)01810, MFS13, anon-WO0118547.88, l(2)01810, lethal(2)01810', description='Dietary and metabolic glutamate transporter', other_designations='CG5304-PA|dietary and metabolic glutamate transporter|dmGlut-PA|lethal (2) 01810', map_location='33B9-33B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12027438, end_position_on_the_genomic_accession=12034122, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  47260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47260, CurrentID=0, Status='live', Symbol='l(3)05822', Aliases='Dmel_CG7129, 5822, CG7129, Dmel\\CG7129, L(3)05822, LD12970, clot 3142, l(3)5822', description='lethal (3) 05822', other_designations='CG7129-PA|CG7129-PB|CG7129-PC|CG7129-PD|l(3)05822-PA|l(3)05822-PB|l(3)05822-PC|l(3)05822-PD', map_location='91A1-91A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18288866, end_position_on_the_genomic_accession=18298879, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  35185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35185, CurrentID=0, Status='live', Symbol='CG17344', Aliases='Dmel_ BcDNA:AT10305, Dmel\\CG17344', description='CG17344 gene product from transcript CG17344-RA', other_designations='CG17344-PA|uncharacterized protein', map_location='37C1-37C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19102044, end_position_on_the_genomic_accession=19102909, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47272, CurrentID=0, Status='live', Symbol='cbx', Aliases='Dmel_CG10536, CG10536, Cbx, Dmel\\CG10536, Ft1, ms(2)05704, ms(2)46C', description='crossbronx', other_designations='CG10536-PA|CG10536-PB|CG10536-PC|cbx-PA|cbx-PB|cbx-PC|male sterile(2)46C', map_location='46B4-46C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9864510, end_position_on_the_genomic_accession=9875072, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  47283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47283, CurrentID=0, Status='live', Symbol='SrpRbeta', Aliases='Dmel_CG33162, BcDNA:GM04779, CG33162, CG5950, Dm SRbeta, Dmel\\CG33162, Q9VSN9, SRbeta, l(3)rK561, l(3)rL537, srpRbeta', description='Signal recognition particle receptor beta', other_designations='CG33162-PA|CG33162-PB|SrpRbeta-PA|SrpRbeta-PB|signal recognition particle receptor beta', map_location='66D11-66D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8718949, end_position_on_the_genomic_accession=8720225, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  47285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47285, CurrentID=0, Status='live', Symbol='repo', Aliases='Dmel_CG31240, 3702, AbRK2, CG31240, CG8045(CT24072), CT24072, Dmel\\CG31240, REPO, RK2, Repo, l(3)03702, rk2', description='reversed polarity', other_designations='CG31240-PA|antibody RK2|repo-PA|reverse polarity', map_location='90F9-90F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18236192, end_position_on_the_genomic_accession=18239602, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798326, CurrentID=0, Status='live', Symbol='CG43107', Aliases='Dmel_ Dmel\\CG43107, MIP12466, lincRNA.331', description='CG43107 gene product from transcript CG43107-RA', other_designations='CG43107-PA|uncharacterized protein', map_location='54A1-54A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17140495, end_position_on_the_genomic_accession=17140865, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19836138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836138, CurrentID=0, Status='live', Symbol='CG44406', Aliases='Dmel_ Dmel\\CG44406', description='CG44406 gene product from transcript CG44406-RA', other_designations='CG44406-PA|uncharacterized protein', map_location='36B6-36B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16981867, end_position_on_the_genomic_accession=16982301, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740713, CurrentID=0, Status='live', Symbol='CG34269', Aliases='Dmel_ Dmel\\CG34269', description='CG34269 gene product from transcript CG34269-RA', other_designations='CG34269-PA|uncharacterized protein', map_location='61C3-61C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=528337, end_position_on_the_genomic_accession=528996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798338, CurrentID=0, Status='live', Symbol='CG43069', Aliases='Dmel_ Dmel\\CG43069, chr2R_14750100_14750772.0', description='CG43069 gene product from transcript CG43069-RA', other_designations='CG43069-PA|uncharacterized protein', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18862862, end_position_on_the_genomic_accession=18863277, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740417, CurrentID=0, Status='live', Symbol='CG34242', Aliases='Dmel_ Dmel\\CG34242', description='CG34242 gene product from transcript CG34242-RB', other_designations='CG34242-PA|CG34242-PB|uncharacterized protein', map_location='69C6-69C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12533996, end_position_on_the_genomic_accession=12534403, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47384, CurrentID=0, Status='live', Symbol='sra', Aliases='Dmel_CG6072, 89B, 89B11, CG6072, Dmel\\CG6072, Sra, anon-WO0118547.92, anon-WO0172774.10, l(3)05203, nebula, nla', description='sarah', other_designations='CG6072-PA|CG6072-PB|lethal (3) 05203|nebula|sara|sra-PA|sra-PB', map_location='3-58.2 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16186999, end_position_on_the_genomic_accession=16197584, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  47396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47396, CurrentID=0, Status='live', Symbol='Tor', Aliases='Dmel_CG5092, 5092, CG5092, CT16317, CT24745, CT24817, DmTOR, Dmel\\CG5092, FRAP/TOR, TOR, TORC1|TOR, dTOR, dTor, dtor, l(2)k17004, mTOR, tor, Tor', description='Target of rapamycin', other_designations='CG5092-PA|CG5092-PB|Target-of-rapamycin|Tor-PA|Tor-PB|target of rapamycin|target of rapamycin kinase', map_location='34A4-34A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13004480, end_position_on_the_genomic_accession=13013197, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  47399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47399, CurrentID=0, Status='live', Symbol='Reph', Aliases='Dmel_CG3920, CG3920, Dmel\\CG3920, l(2)k16918, reph', description='Regulator of eph expression', other_designations='CG3920-PA|CG3920-PB|Reph-PA|Reph-PB|lethal (2) k16918|regulator of eph expression', map_location='24C5-24C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3803545, end_position_on_the_genomic_accession=3810671, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  47408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47408, CurrentID=0, Status='live', Symbol='Vps28', Aliases='Dmel_CG12770, CG12770, DmVps28, Dmel\\CG12770, Dvps28, ESCRT-I, ESCRT-II, dVps28, dvps28, l(2)k16503, vps28', description='Vacuolar protein sorting 28', other_designations='CG12770-PA|CG12770-PB|Vps28-PA|Vps28-PB|vacuolar protein sorting 28', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8078869, end_position_on_the_genomic_accession=8080041, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19835622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835622, CurrentID=0, Status='live', Symbol='CG44435', Aliases='Dmel_ CG15086, Dmel\\CG44435', description='CG44435 gene product from transcript CG44435-RA', other_designations='CG44435-PA|CG44435-PB|uncharacterized protein', map_location='55E6-55E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18678680, end_position_on_the_genomic_accession=18684198, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  47457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47457, CurrentID=0, Status='live', Symbol='ytr', Aliases='Dmel_CG18426, CG18426, Dmel\\CG18426, YTR, b(2)gcn, bgcn, l(2)60A-F, l(2)k13108, l(2)k13214', description='yantar', other_designations='CG18426-PA|CG18426-PB|ytr-PA|ytr-PB', map_location='60A3-60A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23847161, end_position_on_the_genomic_accession=23851480, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  5740714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740714, CurrentID=0, Status='live', Symbol='CG34327', Aliases='Dmel_ Dmel\\CG34327', description='CG34327 gene product from transcript CG34327-RA', other_designations='CG34327-PA|uncharacterized protein', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17222968, end_position_on_the_genomic_accession=17223815, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  47594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47594, CurrentID=0, Status='live', Symbol='aPKC', Aliases='Dmel_CG42783, CG10261, CG30475, CG42783, DaPKC, Dmel\\CG42783, Dmel_CG10261, Dmel_CG30475, PKCl, PKCzeta, a-PKC-zeta, aPKCzeta, aPkc, alphaPKCzeta, apkc, daPKC, dapkc, i, l(2)k06403, pkc-3, psu, aPKC', description='atypical protein kinase C', other_designations='CG42783-PA|CG42783-PB|CG42783-PC|CG42783-PD|CG42783-PE|CG42783-PF|CG42783-PG|CG42783-PI|CG42783-PJ|CG42783-PK|CG42783-PL|CG42783-PM|aPKC-PA|aPKC-PB|aPKC-PC|aPKC-PD|aPKC-PE|aPKC-PF|aPKC-PG|aPKC-PI|aPKC-PJ|aPKC-PK|aPKC-PL|aPKC-PM|atypical PKC|atypical protein kinase|atypical-Protein kinase C|pschur', map_location='51D5-51D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14944458, end_position_on_the_genomic_accession=14962969, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  47611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47611, CurrentID=0, Status='live', Symbol='Gpo-1', Aliases='Dmel_CG8256, 1(2) K05713, CG 8256, CG18786, CG8256, Dmel\\CG8256, GPO, GPO-1, Gpo, alpha-GPO, alphaGPO, anon-WO0118547.114, l(2)k05713', description='Glycerophosphate oxidase-1', other_designations='CG8256-PA|CG8256-PB|CG8256-PC|CG8256-PD|Gpo-1-PA|Gpo-1-PB|Gpo-1-PC|Gpo-1-PD|alpha glycerophosphate oxidase|alpha-Glycerophosphate oxidase|glycerophosphate oxidase|glycerophosphate oxidase-1|lethal (2) k05713|mitochondrial GPD|mitochondrial glycerol-3-phosphate dehydrogenase', map_location='52C8-52C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15857752, end_position_on_the_genomic_accession=15862755, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  47718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47718, CurrentID=0, Status='live', Symbol='CG17829', Aliases='Dmel_ Dmel\\CG17829, EG:115C2.6', description='CG17829 gene product from transcript CG17829-RA', other_designations='CG17829-PA|CG17829-PB|CG17829-PC|uncharacterized protein', map_location='1B12-1B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=625542, end_position_on_the_genomic_accession=628016, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  40722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=40722, CurrentID=0, Status='live', Symbol='CG42675', Aliases='Dmel_ CG10999, CG31554, Dmel\\CG42675, Dmel_CG10999, Dmel_CG31554, anon-WO0140519.247', description='CG42675 gene product from transcript CG42675-RF', other_designations='CG42675-PB|CG42675-PC|CG42675-PD|CG42675-PE|CG42675-PF|CG42675-PG|uncharacterized protein', map_location='83C4-83C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5813609, end_position_on_the_genomic_accession=5818820, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  47729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47729, CurrentID=0, Status='live', Symbol='CG16833', Aliases='Dmel_ BEST:LD13681, Dmel\\CG16833, LD13681', description='CG16833 gene product from transcript CG16833-RM', other_designations='CG16833-PD|CG16833-PE|CG16833-PF|CG16833-PG|CG16833-PH|CG16833-PI|CG16833-PK|CG16833-PL|CG16833-PM|CG16833-PN|uncharacterized protein', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10989401, end_position_on_the_genomic_accession=10997897, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  47730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47730, CurrentID=0, Status='live', Symbol='CG12253', Aliases='Dmel_ BEST:LD08487, DPLG1, Dmel\\CG12253, LD08487', description='CG12253 gene product from transcript CG12253-RA', other_designations='CG12253-PA|PIF-like gene 1|uncharacterized protein', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10987806, end_position_on_the_genomic_accession=10989241, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  47746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47746, CurrentID=0, Status='live', Symbol='Lk', Aliases='Dmel_CG13480, CG13480, DLK, Dmel\\CG13480, Drm-KIN, LCK, LK, Leuk, leucokinin, leukokinin, lk, pp', description='Leucokinin', other_designations='CG13480-PA|Leucokinin-1|Leucokinin-like|Lk-PA|drosokinin|kinin|leucokinin|leucokinin precursor|leukokinin|leukokinin neuropeptide|myokinin-like', map_location='70E2-70E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14560139, end_position_on_the_genomic_accession=14560764, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  47757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47757, CurrentID=0, Status='live', Symbol='betaggt-II', Aliases='Dmel_CG18627, CG18627, Dmel\\CG18627, GGT-II beta, RabGGTbeta, betaggt, betaggt II', description='beta subunit of type II geranylgeranyl transferase', other_designations='CG18627-PA|betaggt-II-PA', map_location='23C1-23C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2879112, end_position_on_the_genomic_accession=2880708, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47762, CurrentID=0, Status='live', Symbol='FKBP59', Aliases='Dmel_CG4535, CG4535, Dmel\\CG4535-bp1, anon-WO0118547.161, dFKBP59, l(2)k00424, l(2)k09010, FKBP59', description='FK506-binding protein FKBP59', other_designations='CG4535-PA|CG4535-PB|FKBP59-PA|FKBP59-PB', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9888005, end_position_on_the_genomic_accession=9890119, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  47763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47763, CurrentID=0, Status='live', Symbol='Abd-B', Aliases='Dmel_CG11648, 9, ABD-B, ABDB, Abd B, Abd-b, Abd0B, AbdB, AbdB(CA)[[26]], AbdB-I, AbdB-II, AbdBm, Abdb, AdbB, BX-C, CG10291, CG11648, DROABDB, DmAbdB, Dmel\\CG11648, Droabdb, FAB, Fab-6, Fab-7, Fab-8, Fab7, IAB 5, IAB5, IAB7, IAB8, Iab-7, MCP, Mc, Mcp, Sab, Tab, Uab-5[Sab], abd-B, abdB, bex, iab, iab-5, iab-6, iab-7, iab-8, iab5, iab6, iab7, iab8, iab8.9, iab9, l(3)89Ed, pH189, tuh-3, twig', description='Abdominal B', other_designations='Abd-B-PA|Abd-B-PB|Abd-B-PC|Abd-B-PD|Abd-B-PE|Abd-B-PF|Abd-B-PG|Abd-B-PH|Abd-B-PI|Abd-B-PJ|Abdominal-B|AbdominalB|CG11648-PA|CG11648-PB|CG11648-PC|CG11648-PD|CG11648-PE|CG11648-PF|CG11648-PG|CG11648-PH|CG11648-PI|CG11648-PJ|abdominal B|abdominal-A|corset|infra-abdominal 5|infra-abdominal 6|infra-abdominal 7|infra-abdominal 8|infra-abdominal 9|infraabdominal|infraabdominal-8.9|male chauvinist pigmentation|microcephalus|miscadastral pigmentation|superabdominal|transabdominal', map_location='3-58.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16927212, end_position_on_the_genomic_accession=16972236, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  47764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47764, CurrentID=0, Status='live', Symbol='Amy-p', Aliases='Dmel_CG18730, AMY, Amy, Amy(p)roximal, Amy1, Amy2, Amy4, Amy5, AmyA, CG18640, CG18730, Dmel\\CG18730, alpha-AMY, alpha-Amy, amy, amy-p, Amy-p', description='Amylase proximal', other_designations='Amy-p-PA|Amy-p-PB|Amylase-A|CG18730-PA|CG18730-PB|alpha-amylase|alpha-amylase proximal|amylase|amylase proximal', map_location='2-77.9 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17118636, end_position_on_the_genomic_accession=17120303, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  47765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47765, CurrentID=0, Status='live', Symbol='c(3)G', Aliases='Dmel_CG17604, C(3)G, C3G, CG17604, Dmel\\CG17604, c(3)g, c3G, cx', description='crossover suppressor on 3 of Gowen', other_designations='CG17604-PA|CG17604-PB|CG17604-PC|c(3)G-PA|c(3)G-PB|c(3)G-PC|crossover suppressor in chromosome 3 of gowen|crossover suppressor on 3 of gowan|crossover suppressor on 3 of gowen', map_location='3-57.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15789473, end_position_on_the_genomic_accession=15792579, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  47767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47767, CurrentID=0, Status='live', Symbol='crc', Aliases='Dmel_CG8669, 929, ATF-4, ATF4, ATF4/crc, Atf4, CG8669, DM6, Dmel\\CG8669, Ire1, dATF-4, dATF4, l(2)crc', description='cryptocephal', other_designations='CG8669-PA|CG8669-PB|CG8669-PE|CG8669-PF|crc-PA|crc-PB|crc-PE|crc-PF', map_location='39C2-39C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21310866, end_position_on_the_genomic_accession=21319527, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  47769 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47769, CurrentID=0, Status='live', Symbol='Dip-C', Aliases='Dmel_CG5663, CG5663, Dmel\\CG5663, dipC', description='Dipeptidase C', other_designations='CG5663-PA|CG5663-PB|Dip-C-PA|Dip-C-PB|dipeptidase C', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12411809, end_position_on_the_genomic_accession=12413759, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  47781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47781, CurrentID=0, Status='live', Symbol='fs(1)K10', Aliases='Dmel_CG3218, CG3218, Dmel\\CG3218, EG:30B8.5, K-10, K10, f(s)K10, fs(1)M9, fs(1)k10, k10', description='female sterile (1) K10', other_designations='CG3218-PA|fs(1)K10-PA', map_location='1-0.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2242138, end_position_on_the_genomic_accession=2246041, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740498, CurrentID=0, Status='live', Symbol='CG34202', Aliases='Dmel_ Dmel\\CG34202', description='CG34202 gene product from transcript CG34202-RA', other_designations='CG34202-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20752586, end_position_on_the_genomic_accession=20753739, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  38691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38691, CurrentID=0, Status='live', Symbol='CG6602', Aliases='Dmel_ Dmel\\CG6602', description='CG6602 gene product from transcript CG6602-RA', other_designations='CG6602-PA', map_location='65A5-65A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6072139, end_position_on_the_genomic_accession=6073539, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  47852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47852, CurrentID=0, Status='live', Symbol='Acp95EF', Aliases='Dmel_CG17924, Acp, CG17924, DTA-D, DTA-E, Dmel\\CG17924, Mst95E, Mst95F, mst 316, mst(3)ag-3, mst-3, mst316, mst316F', description='Accessory gland protein 95EF', other_designations='Acp95EF-PA|CG17924-PA|accessory gland protein 95EF|accessory gland-specific peptide 95EF', map_location='95F3-95F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24228975, end_position_on_the_genomic_accession=24229279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47854, CurrentID=0, Status='live', Symbol='Mst98Ca', Aliases='Dmel_CG11719, CG11719, Dmel\\CG11719, MST98Ca, Mst, Mst98C, Mst98CE, mst(3)gl-12, mst336', description='Male-specific RNA 98Ca', other_designations='CG11719-PA|Mst98Ca-PA', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28291338, end_position_on_the_genomic_accession=28292656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  47869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47869, CurrentID=0, Status='live', Symbol='vir', Aliases='Dmel_CG3496, CG3496, Dmel\\CG3496, Vir', description='virilizer', other_designations='CG3496-PA|vir-PA', map_location='2-103.3 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23360192, end_position_on_the_genomic_accession=23366355, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  47873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47873, CurrentID=0, Status='live', Symbol='bgcn', Aliases='Dmel_CG30170, BGCN, Bgcn, CG10330, CG10331, CG30170, Dmel\\CG30170, PNBP, b(2)cgn, b(2)gcn, bcgn, ben(2)gcn, cg10331, fs(2)eo-C, fs(2)eo6, fs(2)eoQS2, gcn, pep', description='benign gonial cell neoplasm', other_designations='CG30170-PA|CG30170-PB|benign gonial|benign gonial cell neoplasia|benign-(2)-gonial-cell-neoplasm|bgcn-PA|bgcn-PB|cell neoplasm|gonial-cell-neoplasm|peppercorn|putative nuclear binding protein', map_location='60A4-60A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23853270, end_position_on_the_genomic_accession=23858921, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  47874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47874, CurrentID=0, Status='live', Symbol='Zw10', Aliases='Dmel_CG9900, CG9900, Dmel\\CG9900, Dsl1, EG:BACR7C10.3, Mit(1)15, ZW-10, ZW10, abe, dZW10, l(1)3Af, l(1)6-99, l(1)zw10, mit(1)15, mit(l)15, zw-10, zw10', description='Zeste-white 10', other_designations='CG9900-PB|Zw10-PB|abnormal-eye|mitotic 15', map_location='1-1.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2604197, end_position_on_the_genomic_accession=2606824, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47877, CurrentID=0, Status='live', Symbol='tws', Aliases='Dmel_CG6235, 1466/06, 2414, B/PR55, CG6235, DPR55, Dmel\\CG6235, PP2A, PP2A-B55, PP2A[B], PP2a, PR55, Pp2A-85F, Tws, aar, dPP2A, dPP2A-B55, dPP2A-Balpha, dPR55, l(3)01436, l(3)02414, l(3)S022205b, l(3)S022361, l(3)S022440, l(3)S023013, l(3)S023141a, l(3)S023206, l(3)S023309, l(3)S023938, l(3)S024455, l(3)S024834a, l(3)S024838, l(3)S025806, l(3)S025832, l(3)S025913, l(3)S026226, l(3)S026326, l(3)S027127, l(3)S027313, l(3)S028707, l(3)S029110, l(3)S029403, l(3)S029701a, l(3)S031006, l(3)S032303, l(3)S032708c, l(3)S033903, l(3)S035505b, l(3)S042629, l(3)S042630, l(3)S043008b, l(3)S043029, l(3)S045519, l(3)S046918, l(3)S048006, l(3)S048013, l(3)S048507, l(3)S049902, l(3)S052810, l(3)S053011, l(3)S060203, l(3)S060804, l(3)S061805, l(3)S061915, l(3)S063110, l(3)S066017, l(3)S066813, l(3)S067109b, l(3)S067915, l(3)S069206a, l(3)S075110, l(3)S075515b, l(3)S075902b, l(3)S076415, l(3)S080409, l(3)S088513, l(3)S091905, l(3)S101413b, l(3)S105605, l(3)S110008a, l(3)S110815, l(3)S111515, l(3)S119908, l(3)S132907, l(3)S134601a, l(3)S141309, l(3)S146606, l(3)j11C8, l(3)s1801, tw/aar, v158, tws', description='twins', other_designations='CG6235-PA|CG6235-PB|CG6235-PC|CG6235-PD|CG6235-PE|CG6235-PF|CG6235-PG|CG6235-PH|CG6235-PI|PP2A regulatory subunit|PP2A regulatory subunit twins|abnormal anaphase resolution|phosphoprotein phosphatase 2A 55 kDa|protein phosphatase 2A|protein phosphatase 2A at 85F|protein phosphatase type 2A|regulatory B subunit of type 2A protein phosphatase|tws-PA|tws-PB|tws-PC|tws-PD|tws-PE|tws-PF|tws-PG|tws-PH|tws-PI', map_location='85F13-85F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10126129, end_position_on_the_genomic_accession=10139772, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  47878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47878, CurrentID=0, Status='live', Symbol='Eip63F-1', Aliases='Dmel_CG15855, 63F-1, CG15855, Ca[2+] binding protein, Dmel\\CG15855, E63-1, E63F, Eip63-1, Eip63F1, Eip63Fa, eip63F-1', description='Ecdysone-induced protein 63F 1', other_designations='CG15855-PA|CG15855-PB|CG15855-PC|CG15855-PD|Eip63F-1-PA|Eip63F-1-PB|Eip63F-1-PC|Eip63F-1-PD', map_location='63F6-64A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3909973, end_position_on_the_genomic_accession=3935617, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  47894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47894, CurrentID=0, Status='live', Symbol='hbn', Aliases='Dmel_CG33152, 57San, CG10614, CG15647, CG15648, CG33152, Dmel\\CG33152, Hbn', description='homeobrain', other_designations='CG33152-PA|hbn-PA', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20954668, end_position_on_the_genomic_accession=20960909, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  47895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47895, CurrentID=0, Status='live', Symbol='Nmdmc', Aliases='Dmel_CG18466, 151767_at, CG18466, DNMDMC, Dmel\\CG18466, Dnmdmc, Mthfd2, dNMDMC, nmdmc', description='NAD-dependent methylenetetrahydrofolate dehydrogenase', other_designations='CG18466-PA|CG18466-PB|CG18466-PC|NAD dependent methylenetetrahydrofolate dehydrogenase|Nmdmc-PA|Nmdmc-PB|Nmdmc-PC', map_location='85C3-85C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9042508, end_position_on_the_genomic_accession=9050555, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  47900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47900, CurrentID=0, Status='live', Symbol='Taf5', Aliases='Dmel_CG7704, CG7704, Dmel\\CG7704, TAF, TAF5, TAF80, TAF80/85, TAFII80, TAFII85, TAF[II]80, TAF[[II]], TAF[[II]]80, TAF[[II]]80/85, TAF[[II]]85, TFIID, TFIID 85, Taf80, Taf85, Taf[[II]]80, d80, dTAF5, dTAFII85, dTAF[[II]]80, dTAFii80, dmTAF5, dmTaf5, p85', description='TBP-associated factor 5', other_designations='CG7704-PA|TBP associated factor 80|TBP-associated factor|TBP-associated factor 80|TBP-associated factor 80kD|Taf5-PA', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10877312, end_position_on_the_genomic_accession=10879673, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  47901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47901, CurrentID=0, Status='live', Symbol='betaTry', Aliases='Dmel_CG18211, CG18211, DmBetaTry, Dmel\\CG18211, SP119, Try, beta', description='betaTrypsin', other_designations='CG18211-PA|CG18211-PB|betaTry-PA|betaTry-PB', map_location='47F4-47F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11346325, end_position_on_the_genomic_accession=11347395, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  47905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47905, CurrentID=0, Status='live', Symbol='l(2)03659', Aliases='Dmel_CG8799, A, CG11803, CG8799, Dme_CG8799, Dmel\\CG8799, hspr', description='lethal (2) 03659', other_designations='CG8799-PB|CG8799-PC|Hsp related|l(2)03659-PB|l(2)03659-PC', map_location='45D1-45D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9390165, end_position_on_the_genomic_accession=9395281, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  47906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47906, CurrentID=0, Status='live', Symbol='Dek', Aliases='Dmel_CG5935, CG593, CG5935, DEK, Dmel\\CG5935, EG:EG0003.6, anon-WO0172774.191, anon-WO0172774.192, anon-WO0172774.194, dDEK, dek, l(2)04154, l(2)k09907, lincRNA.S2954', description='CG5935 gene product from transcript CG5935-RE', other_designations='CG5935-PA|CG5935-PB|CG5935-PC|CG5935-PD|CG5935-PE|Dek-PA|Dek-PB|Dek-PC|Dek-PD|Dek-PE', map_location='53D14-53D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16856220, end_position_on_the_genomic_accession=16861738, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  5740636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740636, CurrentID=0, Status='live', Symbol='CG34309', Aliases='Dmel_ CG14378, Dm_3R:49679, Dm_3R:49681, Dmel\\CG34309', description='CG34309 gene product from transcript CG34309-RA', other_designations='CG34309-PA|uncharacterized protein', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13372630, end_position_on_the_genomic_accession=13373326, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47942, CurrentID=0, Status='live', Symbol='scat', Aliases='Dmel_CG3766, CG3766, Dmel\\CG3766, Scat, Vps54, ms(2)05289, ms(2)30B', description='scattered', other_designations='CG3766-PA|scat-PA', map_location='30B3-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9431405, end_position_on_the_genomic_accession=9435341, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  47954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47954, CurrentID=0, Status='live', Symbol='Mst57Da', Aliases='Dmel_CG9074, BcDNA:GH19893, CG9074, Dmel\\CG9074, Mst57D, bs30d07.y1, mst(2)ag-1, mst-1, mst57D', description='Male-specific RNA 57Da', other_designations='CG9074-PA|Mst57Da-PA', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25831575, end_position_on_the_genomic_accession=25831899, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  47998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=47998, CurrentID=0, Status='live', Symbol='rin', Aliases='Dmel_CG9412, CG9412, Dmel\\CG9412, Nts, RIN, Rin, cg9412, rasp', description='rasputin', other_designations='CG9412-PB|CG9412-PE|CG9412-PF|CG9412-PG|next to squid|rin-PB|rin-PE|rin-PF|rin-PG', map_location='87F6-87F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13647004, end_position_on_the_genomic_accession=13654527, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  48071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48071, CurrentID=0, Status='live', Symbol='Su(z)12', Aliases='Dmel_CG8013, CG8013, Dmel\\CG8013, SU(Z)12, SUZ12, Su(Z)12, Su(z)[12], l(3)76BDo, su(z)12', description='CG8013 gene product from transcript CG8013-RB', other_designations='CG8013-PA|CG8013-PB|Su(z)12-PA|Su(z)12-PB|Suppressor-of-zeste-12|suppressor of zeste 12|suppressor of zeste[12]', map_location='3-49.5 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19918242, end_position_on_the_genomic_accession=19923462, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12798461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798461, CurrentID=0, Status='live', Symbol='CG43329', Aliases='Dmel_ CG12377, Dmel\\CG43329, Dmel_CG12377', description='CG43329 gene product from transcript CG43329-RA', other_designations='CG43329-PA|uncharacterized protein', map_location='79E6-79E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22504487, end_position_on_the_genomic_accession=22505218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19988923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19988923, CurrentID=0, Status='live', Symbol='ARY', Aliases='Dmel_CG40064, CG40064, Dmel\\CG40064', description='Aldehyde reductase Y', other_designations='ARY-PC|ARY-PD|CG40064-PC|CG40064-PD|aldehyde reductase Y', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=702430, end_position_on_the_genomic_accession=704048, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12798465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798465, CurrentID=0, Status='live', Symbol='CG43194', Aliases='Dmel_ Dmel\\CG43194', description='CG43194 gene product from transcript CG43194-RB', other_designations='CG43194-PA|CG43194-PB|uncharacterized protein', map_location='91B4-91B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18564264, end_position_on_the_genomic_accession=18565173, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798466, CurrentID=0, Status='live', Symbol='CG43236', Aliases='Dmel_ Dmel\\CG43236', description='CG43236 gene product from transcript CG43236-RB', other_designations='CG43236-PA|CG43236-PB|uncharacterized protein', map_location='51B7-51B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14638538, end_position_on_the_genomic_accession=14638935, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38510, CurrentID=0, Status='live', Symbol='Cpr64Ac', Aliases='Dmel_CG15008, CG15008, DmelCpr64Ac, Dmel\\CG15008', description='Cuticular protein 64Ac', other_designations='CG15008-PA|Cpr64Ac-PA|cuticular protein 64Ac', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4211964, end_position_on_the_genomic_accession=4212915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798477, CurrentID=0, Status='live', Symbol='CG43068', Aliases='Dmel_ Dmel\\CG43068', description='CG43068 gene product from transcript CG43068-RA', other_designations='CG43068-PA|uncharacterized protein', map_location='51E5-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15185158, end_position_on_the_genomic_accession=15185952, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48224, CurrentID=0, Status='live', Symbol='Aprt', Aliases='Dmel_CG18315, BcDNA:LP07986, CG18315, Dmel\\CG18315, aprt', description='Adenine phosphoribosyltransferase', other_designations='Aprt-PA|Aprt-PC|Aprt-PD|CG18315-PA|CG18315-PC|CG18315-PD|adenine phosphoribosyltransferase', map_location='3-1.49 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1883795, end_position_on_the_genomic_accession=1885019, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  48228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48228, CurrentID=0, Status='live', Symbol='DNaseII', Aliases='Dmel_CG7780, CG7780, DNase, DNase 1, DNase-1, DNase1, Dmel\\CG7780, dDNase II, dnaseII, dnaseII[lo]', description='Deoxyribonuclease II', other_designations='CG7780-PA|DNaseII-PA|acid DNase|deoxyribonuclease 1|deoxyribonuclease II', map_location='3-61.8 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18019375, end_position_on_the_genomic_accession=18021161, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740662, CurrentID=0, Status='live', Symbol='CG34296', Aliases='Dmel_ Dmel\\CG34296', description='CG34296 gene product from transcript CG34296-RA', other_designations='CG34296-PA|uncharacterized protein', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29722538, end_position_on_the_genomic_accession=29723013, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  48239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48239, CurrentID=0, Status='live', Symbol='ImpE3', Aliases='Dmel_CG2723, CG11046, CG2723, Dmel\\CG2723, IMP-E3, Imp-E3', description='Ecdysone-inducible gene E3', other_designations='CG2723-PB|ImpE3-PB', map_location='84E8-84E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8028453, end_position_on_the_genomic_accession=8029546, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740681, CurrentID=0, Status='live', Symbol='CG34432', Aliases='Dmel_ CG15541, Dmel\\CG34432, Dmel_CG15541', description='CG34432 gene product from transcript CG34432-RA', other_designations='CG34432-PA|uncharacterized protein', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30649151, end_position_on_the_genomic_accession=30649873, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48309, CurrentID=0, Status='live', Symbol='mus101', Aliases='Dmel_CG11156, CG11156, DmMus101, Dmel\\CG11156, Mus101, TOPBP1, anon-WO0148202.1, fs(1)K451, mus 101, mus(1)101, mus-101', description='mutagen-sensitive 101', other_designations='CG11156-PA|mus101-PA', map_location='1-44.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13722246, end_position_on_the_genomic_accession=13727113, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48311, CurrentID=0, Status='live', Symbol='Pkc53E', Aliases='Dmel_CG6622, CG6622, DPKC, Dmel\\CG6622, Dpkc1, PK-C, PKC, PKC 53B, PKC 53E, PKC-53B, PKC-53E, PKC1, PKC53E, Pkc1, dPKC, dPKC53E(br), pkc53e', description='Protein C kinase 53E', other_designations='CG6622-PA|CG6622-PB|CG6622-PC|CG6622-PD|CG6622-PE|CG6622-PF|Pkc53E-PA|Pkc53E-PB|Pkc53E-PC|Pkc53E-PD|Pkc53E-PE|Pkc53E-PF|protein C kinase 53E|protein kinase C|protein kinase-C 53E|protein kinase-c1', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16924655, end_position_on_the_genomic_accession=16945126, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  48312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48312, CurrentID=0, Status='live', Symbol='RpII18', Aliases='Dmel_CG1163, BcDNA:RH21608, CG1163, Dm6, Dmel\\CG1163, H5, Pol II, PolII, RNA Pol II, RNA pol II, RNA polII, RNAP, RNAP II, RNApolII, RPB6_DROME, RPII18, RpABC14, Rpll18, pol II, polII', description='RNA polymerase II 18kD subunit', other_designations='CG1163-PA|RNA polymerase II|RNA polymerase II (18-kDa subunit)|RNA-polymerase ABC 14kD subunit|RpII18-PA', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5371795, end_position_on_the_genomic_accession=5372397, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  48316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48316, CurrentID=0, Status='live', Symbol='alphaTry', Aliases='Dmel_CG18444, CG18444, DmAlphaTry, Dmel\\CG18444, SP102, Try, alpha', description='alphaTrypsin', other_designations='CG18444-PA|alphaTry-PA', map_location='47F4-47F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11344260, end_position_on_the_genomic_accession=11345119, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  48317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48317, CurrentID=0, Status='live', Symbol='ttk', Aliases='Dmel_CG1856, 0037/17, 0250/25, 0438/31, 0702/07, 1049/07, 1119/04, 1184/16, 1209/05, 1209/10, 1260/10, 1281/04, 1325/15, 1372/08, 1396/14, 1418/06, 3540, 5125, 5311, CG11558, CG1856, Dmel\\CG1856, E(yan)100D, FTZ-F2, TTK, TTK69, TTK88, TTKA, Ttk, Ttk69, Ttk88, anon-EST:Liang-2.9, clone 2.9, ftz-f2, ftzf2/ttk, l(3)02667, l(3)j2A1, l(3)j7B8, osn, oss, ovs, tramtrak/FTZ-F2, ttk69, ttkp69, twk, ttk', description='tramtrack', other_designations='CG1856-PA|CG1856-PB|CG1856-PC|CG1856-PD|CG1856-PE|CG1856-PF|CG1856-PG|CG1856-PH|Tramtrack69|Tramtrack88|Ttk88 repressor|enhancer of yan at 100D|oversensitive|schmal|tramtrack-69|tramtrak|ttk-PA|ttk-PB|ttk-PC|ttk-PD|ttk-PE|ttk-PF|ttk-PG|ttk-PH', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31713867, end_position_on_the_genomic_accession=31735392, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  2768715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768715, CurrentID=0, Status='live', Symbol='CheB42c', Aliases='Dmel_CG33350, CG15241, CG33350, Dmel\\CG33350', description='Chemosensory protein B 42c', other_designations='CG33350-PB|CheB42c-PB|chemosensory protein B 42c', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6944614, end_position_on_the_genomic_accession=6971637, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12798498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798498, CurrentID=0, Status='live', Symbol='CG43446', Aliases='Dmel_ Dmel\\CG43446', description='CG43446 gene product from transcript CG43446-RA', other_designations='CG43446-PA|CG43446-PB|uncharacterized protein', map_location='93B3-93B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20986504, end_position_on_the_genomic_accession=20987936, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48335, CurrentID=0, Status='live', Symbol='GstD2', Aliases='Dmel_CG4181, CG4181, DmGST21, DmGSTD2, Dmel\\CG4181, GST, GST D2, GST-D2, GSTD2, Gst1, Gstd21, gst D21, gstD21, gstd2, gstd21, GstD2', description='Glutathione S transferase D2', other_designations='CG4181-PA|GstD2-PA|glutathione S transferase D2|glutathione S transferase D21|glutathione-S-transferase D2', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12371971, end_position_on_the_genomic_accession=12372704, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48336, CurrentID=0, Status='live', Symbol='GstD3', Aliases='Dmel_CG4381, BcDNA:LP11313, CG4381, D3, DmGST22, DmGSTD3, Dmel\\CG4381, GST-D3, GSTD3, Gst, GstD22, gstD22, gstD3, gstd3', description='Glutathione S transferase D3', other_designations='CG4381-PA|GstD3-PA|glutathione S transferase D22|glutathione S transferase D3', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12373068, end_position_on_the_genomic_accession=12373813, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48337, CurrentID=0, Status='live', Symbol='GstD4', Aliases='Dmel_CG11512, CG11512, DmGST23, DmGSTD4, DmGstd4, Dmel\\CG11512, GST, GSTD4, Gst, GstD23, gstD23, gstd4', description='Glutathione S transferase D4', other_designations='CG11512-PA|Glutathione-S-transferase|GstD4-PA|glutathione S transferase D23|glutathione S transferase D4', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12374100, end_position_on_the_genomic_accession=12374824, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48338, CurrentID=0, Status='live', Symbol='GstD5', Aliases='Dmel_CG12242, CG12242, D5, DmGST24, DmGSTD5, Dmel\\CG12242, GST, GSTD5, Gst, GstD24, gstD24, gstd5', description='Glutathione S transferase D5', other_designations='CG12242-PA|GstD5-PA|glutathione S transferase D24|glutathione S transferase D5|glutathione-S-transferase D5', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12375733, end_position_on_the_genomic_accession=12376572, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48339, CurrentID=0, Status='live', Symbol='GstD6', Aliases='Dmel_CG4423, CG4423, DmGST25, DmGSTD6, DmGstd6, Dmel\\CG4423, GST, GSTD6, Gst, GstD25, gstD25, gstd6', description='Glutathione S transferase D6', other_designations='CG4423-PA|GstD6-PA|glutathione S transferase D25|glutathione S transferase D6', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12377128, end_position_on_the_genomic_accession=12377907, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48340, CurrentID=0, Status='live', Symbol='GstD7', Aliases='Dmel_CG4371, CG4371, DmGST26, DmGSTD7, DmGstd7, Dmel\\CG4371, GSTD7, Gst, GstD26, gstD26, gstd7', description='Glutathione S transferase D7', other_designations='CG4371-PA|GstD7-PA|glutathione S transferase D26|glutathione S transferase D7', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12378500, end_position_on_the_genomic_accession=12379255, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48341, CurrentID=0, Status='live', Symbol='GstD8', Aliases='Dmel_CG4421, CG4421, DmGST27, DmGSTD8, Dmel\\CG4421, GST, GST D27, GSTD8, Gst, GstD27, gstD27, gstD8, gstd8', description='Glutathione S transferase D8', other_designations='CG4421-PA|Glutathione-S-transferase|GstD8-PA|glutathione S transferase D27|glutathione S transferase D8', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12380008, end_position_on_the_genomic_accession=12380815, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48343, CurrentID=0, Status='live', Symbol='deltaTry', Aliases='Dmel_CG12351, CG12351, DmDeltaTry, Dmel\\CG12351, Try, delta, endo-I', description='deltaTrypsin', other_designations='CG12351-PA|deltaTry-PA|endoproteinase I', map_location='47F5-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11355319, end_position_on_the_genomic_accession=11356128, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772334, CurrentID=0, Status='live', Symbol='CG33928', Aliases='Dmel_ Dmel\\CG33928', description='CG33928 gene product from transcript CG33928-RA', other_designations='CG33928-PA|uncharacterized protein', map_location='36C8-36C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17435891, end_position_on_the_genomic_accession=17436600, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  48373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48373, CurrentID=0, Status='live', Symbol='l(2)k10201', Aliases='Dmel_CG13951, B, CG13951, CG8803, Dmel\\CG13951', description='lethal (2) k10201', other_designations='CG13951-PA|l(2)k10201-PA', map_location='45D3-45D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9399080, end_position_on_the_genomic_accession=9399899, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48381, CurrentID=0, Status='live', Symbol='Lcp65Ab2', Aliases='Dmel_CG18773, BcDNA:RE39879, CG18773, CP5, DCP-beta, Dmel\\CG18773, LCP-5, LCP5, LCP65Ab2, L[[3]]CP5, Lcp-b, Lcp-b2, Lcp5, Lcp65A-b, Lcp65Ab', description='CG18773 gene product from transcript CG18773-RA', other_designations='CG18773-PA|Lcp65Ab2-PA', map_location='3-11 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6146674, end_position_on_the_genomic_accession=6147089, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48382, CurrentID=0, Status='live', Symbol='Lcp65Ab1', Aliases='Dmel_CG32400, CG18776, CG32400, CP5, DCP-alpha, Dmel\\CG32400, LCP-5, LCP10, LCP5, LCP65Ab1, L[[3]]CP5, Lcp-b, Lcp-b1, Lcp10, Lcp5, Lcp65A-b, Lcp65Ab, Lcp65Abl, Rho', description='CG32400 gene product from transcript CG32400-RA', other_designations='CG32400-PA|Lcp65Ab1-PA|larval cuticle protein 10', map_location='3-11 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6149548, end_position_on_the_genomic_accession=6149960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48387, CurrentID=0, Status='live', Symbol='l(2)k14710', Aliases='Dmel_CG8325, BcDNA:LD21719, CG8325, Dmel\\CG8325, l(2)k15603', description='lethal (2) k14710', other_designations='CG8325-PA|CG8325-PB|l(2)k14710-PA|l(2)k14710-PB', map_location='42A13-42A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6209782, end_position_on_the_genomic_accession=6214122, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768717, CurrentID=0, Status='live', Symbol='CheB42b', Aliases='Dmel_CG33351, CG15241, CG33351, Dmel\\CG33351', description='Chemosensory protein B 42b', other_designations='CG33351-PA|CheB42b-PA|chemosensory protein B 42b', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6943096, end_position_on_the_genomic_accession=6943995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798510, CurrentID=0, Status='live', Symbol='CG43342', Aliases='Dmel_ CG12757, CG31169, Dmel\\CG43342, Dmel_CG31169, gpdh-3', description='CG43342 gene product from transcript CG43342-RA', other_designations='CG43342-PA|uncharacterized protein', map_location='94A6-94A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22381340, end_position_on_the_genomic_accession=22382147, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36528, CurrentID=0, Status='live', Symbol='CG13339', Aliases='Dmel_ BcDNA:RE69336, Dmel\\CG13339', description='CG13339 gene product from transcript CG13339-RA', other_designations='CG13339-PA|uncharacterized protein', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13824858, end_position_on_the_genomic_accession=13825572, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  48421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48421, CurrentID=0, Status='live', Symbol='CG3558', Aliases='Dmel_ BEST:LD23852, Dmel\\CG3558, FBgn0025681, LD23852, c4.3', description='CG3558 gene product from transcript CG3558-RD', other_designations='CG3558-PA|CG3558-PB|CG3558-PC|CG3558-PD', map_location='23C4-23C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2992870, end_position_on_the_genomic_accession=2998726, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  48422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48422, CurrentID=0, Status='live', Symbol='CG5707', Aliases='Dmel_ BEST:HL03644, Dmel\\CG5707, HL03644', description='CG5707 gene product from transcript CG5707-RA', other_designations='CG5707-PA|CG5707-PB|CG5707-PC', map_location='62D2-62D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2189160, end_position_on_the_genomic_accession=2192221, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12798515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798515, CurrentID=0, Status='live', Symbol='CG43092', Aliases='Dmel_ Dmel\\CG43092', description='CG43092 gene product from transcript CG43092-RA', other_designations='CG43092-PA|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22768883, end_position_on_the_genomic_accession=22769667, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48440, CurrentID=0, Status='live', Symbol='Kebab', Aliases='Dmel_CG31672, BEST:LD15963, CG31672, CG4248, Dmel\\CG31672, keb, kebab, n(2)k09932', description='Kinetochore and EB1 associated basic protein', other_designations='CG31672-PA|CG31672-PB|Kebab-PA|Kebab-PB|kinetochore and EB1 associated basic protein', map_location='22C1-22C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2045494, end_position_on_the_genomic_accession=2048054, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3885620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885620, CurrentID=0, Status='live', Symbol='CG34049', Aliases='Dmel_ BP1065, Dmel\\CG34049', description='CG34049 gene product from transcript CG34049-RB', other_designations='CG34049-PB|uncharacterized protein', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2264231, end_position_on_the_genomic_accession=2265321, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  48445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48445, CurrentID=0, Status='live', Symbol='Acph-1', Aliases='Dmel_CG7899, ACP, ACPH, APH, Acph, CG7899, Dmel\\CG7899', description='Acid phosphatase 1', other_designations='Acph-1-PA|Acph-1-PB|CG7899-PA|CG7899-PB|acid phosphatase|acid phosphatase 1', map_location='3-101.1 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29991144, end_position_on_the_genomic_accession=29993411, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  48448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48448, CurrentID=0, Status='live', Symbol='CkIIalpha', Aliases='Dmel_CG17520, 0958/08, CG17520, CJIIalpha, CK II, CK-2, CK-II, CK-II alpha, CK2, CK2 alpha, CK2-alpha, CK2a, CK2alpha, CKII, CKII alpha, CKII-alpha, CKIIa, CKIIalfa, CKIIalpha, Cask-II-a, Ck2alpha, CkII, DmCK2alpha, DmCKIIalpha, Dmel\\CG17520, Tik, alphaCK2, anon-WO02059370.53, ck2alpha, ckIIalpha, dCK2alpha, dCKII, dCKII alpha, l(3)S095808', description='casein kinase IIalpha', other_designations='CG17520-PA|CG17520-PB|CG17520-PC|CG17520-PD|CG17520-PE|CG17520-PF|CG17520-PG|CkIIalpha-PA|CkIIalpha-PB|CkIIalpha-PC|CkIIalpha-PD|CkIIalpha-PE|CkIIalpha-PF|CkIIalpha-PG|NAP1-kinase|alpha subunit of drosophila casein kinase II|casein kinase 2|casein kinase II|casein kinase II alpha|casein kinase II alpha subunit|casein kinase II-alpha subunit|lethal (3) S095808|protein kinase CK2|timekeeper', map_location='80D1-80D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23097072, end_position_on_the_genomic_accession=23103460, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  48450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48450, CurrentID=0, Status='live', Symbol='Dip-B', Aliases='Dmel_CG9285, CG9285, DipB, Dmel\\CG9285', description='Dipeptidase B', other_designations='CG9285-PA|CG9285-PB|CG9285-PC|CG9285-PD|Dip-B-PA|Dip-B-PB|Dip-B-PC|Dip-B-PD|dipeptidase B', map_location='3-53.6 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13782288, end_position_on_the_genomic_accession=13785676, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  48481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48481, CurrentID=0, Status='live', Symbol='Grip91', Aliases='Dmel_CG10988, CG10988, DGrip91, Dd4, Dgrip9, Dgrip91, Dgrips91, Dmel\\CG10988, GCP3, GRIP91, anon-WO03040301.197, dGRIP91, dd4, l(1)G0122, l(1)d deg-4, l(1)d.deg.-4, l(1)d.deg.4, l(1)dd4', description='Gamma-tubulin ring protein 91', other_designations='CG10988-PA|Dgrip91|Dgrips91|Grip91-PA|lethal (1) discs degenerate 4', map_location='1-44.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13736570, end_position_on_the_genomic_accession=13744683, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12798528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798528, CurrentID=0, Status='live', Symbol='CG43222', Aliases='Dmel_ DMG4-MB6.chr3R.35.021, Dmel\\CG43222', description='CG43222 gene product from transcript CG43222-RA', other_designations='CG43222-PA|CG43222-PB|uncharacterized protein', map_location='96A13-96A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24634625, end_position_on_the_genomic_accession=24635165, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48531, CurrentID=0, Status='live', Symbol='Pp1-13C', Aliases='Dmel_CG9156, 13C, CG9156, DmPp1-13C, Dmel\\CG9156, PP1, PP1 13C, PP1 beta13c, PP1-13C, PP13C, PP1alpha-13C, PP1alpha13C, PP1c, Pp1, pp1-13C', description='Protein phosphatase 1 at 13C', other_designations='CG9156-PA|Pp1-13C-PA|protein phosphatase 1|protein phosphatase 1 13C|protein phosphatase 1 at 13C', map_location='13B9-13B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15359101, end_position_on_the_genomic_accession=15361061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48532, CurrentID=0, Status='live', Symbol='PpY-55A', Aliases='Dmel_CG10930, CG10930, D19, DmPpY-55A, Dmel\\CG10930, PPY, PPY 55A, PpD19', description='Protein phosphatase Y at 55A', other_designations='CG10930-PA|CG10930-PB|PpY-55A-PA|PpY-55A-PB|protein phosphatase D19|protein phosphatase Y at 55A|protein phosphatase from PCR fragment D19|protein phosphatase-Y', map_location='55A2-55A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17955396, end_position_on_the_genomic_accession=17956570, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  48535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48535, CurrentID=0, Status='live', Symbol='Hrb87F', Aliases='Dmel_CG12749, CG12749, Dmel\\CG12749, HRB87F, HRB87F/hrp36, HRP36, Hrb2, Hrb85CDa, Hrb87f, Hrp36, P11, P11/Hrb87F, Q14, Q16, hnRNP A2/B1, hnRNP36, hnRNPA1, hnRNPA1/A2, hnRNPA2, hrp36, hrp40, p38, Hrb87F', description='Heterogeneous nuclear ribonucleoprotein at 87F', other_designations='CG12749-PA|CG12749-PC|CG12749-PD|CG12749-PE|Hrb87F-PA|Hrb87F-PC|Hrb87F-PD|Hrb87F-PE|heterogeneous nuclear ribonucleoprotein at 85CD|heterogeneous nuclear ribonucleoprotein at 87F|hnRNA-binding protein 2', map_location='87F7-87F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13656846, end_position_on_the_genomic_accession=13660531, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  48552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48552, CurrentID=0, Status='live', Symbol='Sam-S', Aliases='Dmel_CG2674, CG2674, Dmel\\CG2674, M(2)21A-B, M(2)21AB, METK, SamS, Su(z)5, m(2)21ab', description='S-adenosylmethionine Synthetase', other_designations='CG2674-PA|CG2674-PB|CG2674-PC|CG2674-PD|CG2674-PE|CG2674-PF|CG2674-PG|CG2674-PH|CG2674-PI|CG2674-PJ|CG2674-PK|CG2674-PL|S-Adenosylmethionine synthetase|S-adenosylmethionine synthetase|Sam-S-PA|Sam-S-PB|Sam-S-PC|Sam-S-PD|Sam-S-PE|Sam-S-PF|Sam-S-PG|Sam-S-PH|Sam-S-PI|Sam-S-PJ|Sam-S-PK|Sam-S-PL|methionine adenosyl transferase|minute (2) 21AB', map_location='2-0.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=106903, end_position_on_the_genomic_accession=114432, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  48571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48571, CurrentID=0, Status='live', Symbol='heph', Aliases='Dmel_CG31000, CG2094, CG2290, CG31000, DmPTB, Dmel\\CG31000, HEPH, Heph, PTB, Ptb, dPTB, dmPTB, ema, hnRNP I, l(3)03429, l(3)j11B9, ptb', description='hephaestus', other_designations='CG31000-PA|CG31000-PB|CG31000-PC|CG31000-PD|CG31000-PE|CG31000-PF|CG31000-PG|CG31000-PH|CG31000-PI|CG31000-PJ|CG31000-PK|CG31000-PL|CG31000-PM|CG31000-PN|CG31000-PO|CG31000-PP|CG31000-PQ|CG31000-PR|CG31000-PS|CG31000-PT|CG31000-PU|CG31000-PV|CG31000-PW|CG31000-PY|CG31000-PZ|ectopic margin|heph-PA|heph-PB|heph-PC|heph-PD|heph-PE|heph-PF|heph-PG|heph-PH|heph-PI|heph-PJ|heph-PK|heph-PL|heph-PM|heph-PN|heph-PO|heph-PP|heph-PQ|heph-PR|heph-PS|heph-PT|heph-PU|heph-PV|heph-PW|heph-PY|heph-PZ|heterogeneous nuclear RibonucleoProtein I|polypyrimidine tract binding protein|polypyrimidine tract-binding protein', map_location='100D3-100E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31846987, end_position_on_the_genomic_accession=32015520, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  48572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48572, CurrentID=0, Status='live', Symbol='Hsp60B', Aliases='Dmel_CG2830, 152031_at, CG2830, Dmel\\CG2830, Hsp60b, anon-WO0140519.61, hsp60B, ms(2)06619, ms(2)21D', description='Heat shock protein 60 related', other_designations='CG2830-PA|Hsp 60 related|Hsp60B-PA|heat shock protein 60 related|male sterile(2)21D', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=728362, end_position_on_the_genomic_accession=730564, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798539, CurrentID=0, Status='live', Symbol='CG43149', Aliases='Dmel_ Dmel\\CG43149', description='CG43149 gene product from transcript CG43149-RA', other_designations='CG43149-PA|uncharacterized protein', map_location='61A6-61A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=53873, end_position_on_the_genomic_accession=54751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  48581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48581, CurrentID=0, Status='live', Symbol='Hsp70Aa', Aliases='Dmel_CG31366, 87A, CG31366, Dm-hsp70, DmHSP70AA, Dmel\\CG31366, HSP70, Hsp 70, Hsp-70, Hsp-70Aa, Hsp70, Hsp70A, Hsp70a, dHsp70, dhsp70, dhsp70Aa, hs70, hsp 70, hsp-70, hsp70, hsp70 87A, hsp70 Aa, hsp70(87A), hsp70A, hsp70Aa, hsp70a', description='Heat-shock-protein-70Aa', other_designations='87A locus|CG31366-PA|Hsp70Aa-PA|heat shock 70|heat shock protein-70|heat-shock protein 70|hsp70 87A|protein 70', map_location='87A2-87A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11954163, end_position_on_the_genomic_accession=11956544, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  48582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48582, CurrentID=0, Status='live', Symbol='Hsp70Bb', Aliases='Dmel_CG31359, Bb, CG31359, DMHSP7D1, Dm-hsp70, Dmel\\CG31359, HSP70, Hsp 70, Hsp-70, Hsp70, Hsp70(87C), Hsp70A7d, Hsp70Bc, Hsp70Bc, dHsp70, dhsp70, hsp-70, hsp70, hsp70 87C, hsp70B, hsp70Bb, Hsp70Bb', description='Heat-shock-protein-70Bb', other_designations='87C locus|CG31359-PA|Hsp70Bb-PA|heat shock protein 70|heat-shock protein 70|protein 70', map_location='87B14-87B14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12505793, end_position_on_the_genomic_accession=12508139, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48583, CurrentID=0, Status='live', Symbol='Hsp70Bc', Aliases='Dmel_CG6489, Bc, CG6489, Dm-hsp70, Dmel\\CG6489, HSP70, HSP70Bc, Hsp 70, Hsp-70, Hsp-70Bc, Hsp70, Hsp70(87C), Hsp70-BC, Hsp70A7d, Hsp70B, Hsp70Bbc, Hsp70Bd, dHsp70, dhsp70, hsp-70, hsp70, hsp70 87C, hsp70B, hsp70Bc', description='Heat-shock-protein-70Bc', other_designations='87C locus|CG6489-PA|Hsp70Bc-PA|heat shock 70|heat shock protein 70|heat-shock protein 70|heat-shock-protein 70bc|protein 70', map_location='87B14-87B15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12509076, end_position_on_the_genomic_accession=12511461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  117327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117327, CurrentID=0, Status='live', Symbol='Gr98b', Aliases='Dmel_CG31059, 98B.2, CG12884, CG31059, Dmel\\CG31059, GR98B.2, GR98b, Gr98A1, Gr98B2', description='Gustatory receptor 98b', other_designations='CG31059-PA|Gr98b-PA|gustatory receptor 98B.2|gustatory receptor 98b', map_location='98A4-98A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27542497, end_position_on_the_genomic_accession=27543888, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798544, CurrentID=0, Status='live', Symbol='CG43337', Aliases='Dmel_ BcDNA:SD03580, CG32345, Dmel\\CG43337', description='CG43337 gene product from transcript CG43337-RA', other_designations='CG43337-PA|CG43337-PB|uncharacterized protein', map_location='61C7-61C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=620695, end_position_on_the_genomic_accession=622316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48613, CurrentID=0, Status='live', Symbol='CG5704', Aliases='Dmel_ BEST:LD30049, Dmel\\CG5704, LD30049', description='CG5704 gene product from transcript CG5704-RB', other_designations='CG5704-PA|CG5704-PB', map_location='62D1-62D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2169553, end_position_on_the_genomic_accession=2170971, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48632, CurrentID=0, Status='live', Symbol='Act87E', Aliases='Dmel_CG18290, Act87e, Actin, Actin87E, CG18290, Dmel\\CG18290, E, act 87E, act87E, actin, mus87E', description='Actin 87E', other_designations='Act87E-PA|Act87E-PB|Act87E-PC|CG18290-PA|CG18290-PB|CG18290-PC|actin|actin 87E|actin87E', map_location='87E11-87E11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13425983, end_position_on_the_genomic_accession=13428375, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798550, CurrentID=0, Status='live', Symbol='CG43165', Aliases='Dmel_ Dmel\\CG43165', description='CG43165 gene product from transcript CG43165-RA', other_designations='CG43165-PA|uncharacterized protein', map_location='24B2-24B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3694431, end_position_on_the_genomic_accession=3695020, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798551, CurrentID=0, Status='live', Symbol='CG43145', Aliases='Dmel_ Dmel\\CG43145, lincRNA.26', description='CG43145 gene product from transcript CG43145-RA', other_designations='CG43145-PA|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3701826, end_position_on_the_genomic_accession=3702248, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885656, CurrentID=0, Status='live', Symbol='CG34045', Aliases='Dmel_ BP1009, Dmel\\CG34045', description='CG34045 gene product from transcript CG34045-RB', other_designations='CG34045-PB|uncharacterized protein', map_location='56D1-56D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19330889, end_position_on_the_genomic_accession=19335188, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  39810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39810, CurrentID=0, Status='live', Symbol='CG4925', Aliases='Dmel_ Dmel\\CG4925', description='CG4925 gene product from transcript CG4925-RA', other_designations='CG4925-PA', map_location='72E2-72E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16348747, end_position_on_the_genomic_accession=16351951, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38719, CurrentID=0, Status='live', Symbol='CG7386', Aliases='Dmel_ Dmel\\CG7386', description='CG7386 gene product from transcript CG7386-RA', other_designations='CG7386-PA', map_location='65A7-65A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6189745, end_position_on_the_genomic_accession=6192049, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798556, CurrentID=0, Status='live', Symbol='CG43055', Aliases='Dmel_ Dmel\\CG43055, MB8.chr2L.pasa.5', description='CG43055 gene product from transcript CG43055-RA', other_designations='CG43055-PA|uncharacterized protein', map_location='24F1-24F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4411155, end_position_on_the_genomic_accession=4411824, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885664, CurrentID=0, Status='live', Symbol='CG33258', Aliases='Dmel_ BP1001, CR33258, Dmel\\CG33258', description='CG33258 gene product from transcript CG33258-RB', other_designations='CG33258-PB|CG33258-PC|uncharacterized protein', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16082551, end_position_on_the_genomic_accession=16083638, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885665, CurrentID=0, Status='live', Symbol='CG34031', Aliases='Dmel_ BP1068, CR34031, Dmel\\CG34031, Dmel_CR34031', description='CG34031 gene product from transcript CG34031-RC', other_designations='CG34031-PC|uncharacterized protein', map_location='80D4-80D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23174446, end_position_on_the_genomic_accession=23185750, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885667, CurrentID=0, Status='live', Symbol='CG42729', Aliases='Dmel_ CG17014, CG33984, CG7973, Dmel\\CG42729, Dmel_CG33984', description='CG42729 gene product from transcript CG42729-RA', other_designations='CG42729-PA|uncharacterized protein', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15526910, end_position_on_the_genomic_accession=15527516, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3885668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885668, CurrentID=0, Status='live', Symbol='CG33965', Aliases='Dmel_ CG1211, Dmel\\CG33965', description='CG33965 gene product from transcript CG33965-RA', other_designations='CG33965-PA|uncharacterized protein', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1232882, end_position_on_the_genomic_accession=1234015, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798569, CurrentID=0, Status='live', Symbol='CG43307', Aliases='Dmel_ Dmel\\CG43307, lincRNA.52', description='CG43307 gene product from transcript CG43307-RA', other_designations='CG43307-PA|uncharacterized protein', map_location='25F5-25F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5811574, end_position_on_the_genomic_accession=5811925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  48772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48772, CurrentID=0, Status='live', Symbol='NimC3', Aliases='Dmel_CG16880, BG:DS05899.2, CG16880, Dmel\\CG16880, I(2)34Fa, br30, l(2)34Fa, l(2)br30, l(2)k00811, l34Fa, nimC3', description='Nimrod C3', other_designations='CG16880-PB|NimC3-PB|lethal (2) 34Fa|nimrod C3', map_location='34F1-34F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14019357, end_position_on_the_genomic_accession=14020208, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  48775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48775, CurrentID=0, Status='live', Symbol='Pgant35A', Aliases='Dmel_CG7480, A3, BG:DS01068.8, CG7480, Dmel\\CG7480, GalNAc-T, br12, dGalNAc-T1, l(2)35Aa, l(2)br12, l35Aa, pgant35, pgant35A', description='Polypeptide N-acetylgalactosaminyltransferase 35A', other_designations='CG7480-PA|Pgant35A-PA|lethal 12 in the black-reduced region|lethal group A3|polypeptide N-acetylgalactosaminyltransferase 35A', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14359252, end_position_on_the_genomic_accession=14361411, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768730, CurrentID=0, Status='live', Symbol='CG12158', Aliases='Dmel_ BcDNA:RH45738, Dmel\\CG12158, anon-WO0153538.61', description='CG12158 gene product from transcript CG12158-RA', other_designations='CG12158-PA|uncharacterized protein', map_location='45D1-45D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9362574, end_position_on_the_genomic_accession=9362906, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  48782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48782, CurrentID=0, Status='live', Symbol='l(2)35Df', Aliases='Dmel_CG4152, BG:DS09217.2, BcDNA:GH07290, CG4152, Dmel\\CG4152, Mtr4, br44, cg4152, l(2)br44, l(2)k14422, l(2)k14423, l.2.35Df, lethal(2)35Df', description='lethal (2) 35Df', other_designations='CG4152-PA|l(2)35Df-PA', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15752263, end_position_on_the_genomic_accession=15755900, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  48785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48785, CurrentID=0, Status='live', Symbol='wek', Aliases='Dmel_CG4148, BG:DS09217.5, CG4148, Dmel\\CG4148, l(2)35Ea, l(2)br47, l35Ea', description='weckle', other_designations='CG4148-PA|CG4148-PB|lethal(2)35Ea|wek-PA|wek-PB', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15766133, end_position_on_the_genomic_accession=15768349, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  48805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48805, CurrentID=0, Status='live', Symbol='Catsup', Aliases='Dmel_CG10449, Bc, CG10449, Dmel\\CG10449, catsup, cg10449, dCATSUP, fs(2)ltoQT50, l(2)37Bc, l(2)E13, l(2)k05402, l(2)k05424', description='Catecholamines up', other_designations='CG10449-PA|Catsup-PA|catecholamines up|catsup', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19041862, end_position_on_the_genomic_accession=19043614, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740723, CurrentID=0, Status='live', Symbol='CG34303', Aliases='Dmel_ Dmel\\CG34303', description='CG34303 gene product from transcript CG34303-RA', other_designations='CG34303-PA|uncharacterized protein', map_location='85F15-85F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10151153, end_position_on_the_genomic_accession=10151697, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  48844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48844, CurrentID=0, Status='live', Symbol='l(2)tid', Aliases='Dmel_CG5504, CG5504, Dmel\\CG5504, DnaJ, TID56, cDNA 10D, l(2)701, l(2)td, l(2)tud, l(2)tumorous disc, l(2)tumorous imaginal discs, tid', description='lethal (2) tumorous imaginal discs', other_designations='CG5504-PA|CG5504-PB|CG5504-PC|CG5504-PD|CG5504-PE|l(2)tid-PA|l(2)tid-PB|l(2)tid-PC|l(2)tid-PD|l(2)tid-PE', map_location='59F6-59F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23706913, end_position_on_the_genomic_accession=23709026, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798585, CurrentID=0, Status='live', Symbol='CG43098', Aliases='Dmel_ Dmel\\CG43098, chr3L_3658997_3659727.0', description='CG43098 gene product from transcript CG43098-RA', other_designations='CG43098-PA|uncharacterized protein', map_location='63E6-63E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3659068, end_position_on_the_genomic_accession=3659724, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  48903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48903, CurrentID=0, Status='live', Symbol='l(3)73Ah', Aliases='Dmel_CG4195, CG4195, Dmel\\CG4195, I(3)73Ah, L(3)73Ah, l(3)02540, l(3)133.18', description='lethal (3) 73Ah', other_designations='CG4195-PA|CG4195-PB|l(3)73Ah-PA|l(3)73Ah-PB|lethal(3)73Ah', map_location='73A10-73A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16588475, end_position_on_the_genomic_accession=16590132, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768659, CurrentID=0, Status='live', Symbol='CG33293', Aliases='Dmel_ BcDNA:GH07964, CG12163, Dmel\\CG33293, NEST:bs06b05, NEST:bs07d08, NEST:bs34a01, NEST:bs35d06, bs06b05.y1, bs07d08.y1, bs34a01.y1, bs35d06.y1', description='CG33293 gene product from transcript CG33293-RA', other_designations='CG33293-PA|uncharacterized protein', map_location='82F7-82F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5240014, end_position_on_the_genomic_accession=5240735, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768661, CurrentID=0, Status='live', Symbol='Fer1', Aliases='Dmel_CG33323, BcDNA:RH30329, CG10066, CG14611, CG33323, Dmel\\CG33323, bHLHa30, fer1', description='48 related 1', other_designations='CG33323-PA|CG33323-PB|Fer1-PA|Fer1-PB', map_location='84C5-84C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7136206, end_position_on_the_genomic_accession=7141972, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768664, CurrentID=0, Status='live', Symbol='CG33325', Aliases='Dmel_ Dmel\\CG33325', description='CG33325 gene product from transcript CG33325-RB', other_designations='CG33325-PB|uncharacterized protein', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8678954, end_position_on_the_genomic_accession=8680034, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768665, CurrentID=0, Status='live', Symbol='CG17207', Aliases='Dmel_ BcDNA:AT11646, Dmel\\CG17207', description='CG17207 gene product from transcript CG17207-RA', other_designations='CG17207-PA|uncharacterized protein', map_location='87B10-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12437817, end_position_on_the_genomic_accession=12439139, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768666, CurrentID=0, Status='live', Symbol='Sp212', Aliases='Dmel_CG33329, CG33329, Dmel\\CG33329, ORE-5, SP212', description='Serine-peptidase 212', other_designations='CG33329-PC|Sp212-PC', map_location='88A12-88A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14222074, end_position_on_the_genomic_accession=14224155, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768668, CurrentID=0, Status='live', Symbol='CG33331', Aliases='Dmel_ BcDNA:SD17447, CG3641, Dmel\\CG33331', description='CG33331 gene product from transcript CG33331-RA', other_designations='CG33331-PA|uncharacterized protein', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14886456, end_position_on_the_genomic_accession=14887739, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768669, CurrentID=0, Status='live', Symbol='CG33332', Aliases='Dmel_ BcDNA:RE07395, CG3641, Dmel\\CG33332', description='CG33332 gene product from transcript CG33332-RA', other_designations='CG33332-PA|uncharacterized protein', map_location='88D6-88D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14886456, end_position_on_the_genomic_accession=14889146, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768670, CurrentID=0, Status='live', Symbol='dpr9', Aliases='Dmel_CG33485, BcDNA:GH01517, CG12601, CG14862, CG14863, CG31114, CG33485, CT34680, Dmel\\CG33485, Dpr-9, Dpr9', description='defective proboscis extension response 9', other_designations='CG33485-PA|CG33485-PB|dpr9-PA|dpr9-PB', map_location='88E1-88E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15054930, end_position_on_the_genomic_accession=15077206, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  2768671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768671, CurrentID=0, Status='live', Symbol='Pxd', Aliases='Dmel_CG3477, CG3477, DmPO, Dmel\\CG3477, Dpxd, PO', description='Peroxidase', other_designations='CG3477-PA|CG3477-PB|CG3477-PC|Pxd-PA|Pxd-PB|Pxd-PC|peroxidase', map_location='89E8-89E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17014576, end_position_on_the_genomic_accession=17027987, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  2768672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768672, CurrentID=0, Status='live', Symbol='CG33333', Aliases='Dmel_ Dmel\\CG33333', description='CG33333 gene product from transcript CG33333-RA', other_designations='CG33333-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17417467, end_position_on_the_genomic_accession=17418042, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768673, CurrentID=0, Status='live', Symbol='att-ORFB', Aliases='Dmel_CG33488, CG33488, Dmel\\CG33488, att', description='alternative testis transcripts open reading frame B', other_designations='CG33488-PA|alternative testis transcripts|att-ORFB-PA', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20552635, end_position_on_the_genomic_accession=20554409, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2768674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768674, CurrentID=0, Status='live', Symbol='Fancd2', Aliases='Dmel_CG17269, CG17269, CG31192, CG31194, Dmel\\CG17269, FAD, FANCD2, dmFANCD2', description='CG17269 gene product from transcript CG17269-RA', other_designations='CG17269-PA|Fancd2-PA', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20810804, end_position_on_the_genomic_accession=20816190, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  2768677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768677, CurrentID=0, Status='live', Symbol='p53', Aliases='Dmel_CG33336, CG10873, CG31325, CG33336, D- DMP53, Dm-P53, DmP53, Dmel\\CG33336, Dmp53, Dp53, dmp53, dp53, prac', description='CG33336 gene product from transcript CG33336-RB', other_designations='CG33336-PA|CG33336-PB|CG33336-PC|CG33336-PE|Dmp53|drosophila p53|p53-PA|p53-PB|p53-PC|p53-PE|p53-like regulator of apoptosis and cell cycle|protein 53', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23049657, end_position_on_the_genomic_accession=23054082, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  2768678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768678, CurrentID=0, Status='live', Symbol='CG33337', Aliases='Dmel_ Dmel\\CG33337', description='CG33337 gene product from transcript CG33337-RB', other_designations='CG33337-PB|uncharacterized protein', map_location='95A3-95A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23590141, end_position_on_the_genomic_accession=23592769, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768679, CurrentID=0, Status='live', Symbol='p38c', Aliases='Dmel_CG33338, CG33338, Dmel\\CG33338, Dp38c, Mpk2B, P38c, dMPK2B, dp38c', description='p38c MAP kinase', other_designations='CG33338-PA|Mpk2B|p38c-PA', map_location='95E5-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24149020, end_position_on_the_genomic_accession=24150203, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768680, CurrentID=0, Status='live', Symbol='Myo95E', Aliases='Dmel_CG5501, 95E, CG31134, CG31135, CG5501, Dm95E, DmAE003746, Dmel\\CG5501, myo 95E', description='Myosin 95E', other_designations='CG5501-PB|CG5501-PD|CG5501-PE|CG5501-PF|CG5501-PG|CG5501-PJ|Myo95E-PB|Myo95E-PD|Myo95E-PE|Myo95E-PF|Myo95E-PG|Myo95E-PJ|myosin 95E', map_location='95E6-95E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24154826, end_position_on_the_genomic_accession=24161498, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  2768681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768681, CurrentID=0, Status='live', Symbol='CG33339', Aliases='Dmel_ CG13612, Dmel\\CG33339', description='CG33339 gene product from transcript CG33339-RB', other_designations='CG33339-PB|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24361847, end_position_on_the_genomic_accession=24363142, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  2768682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768682, CurrentID=0, Status='live', Symbol='CG33340', Aliases='Dmel_ BEST:GH24664, BcDNA:AT31672, CG13612, Dmel\\CG33340', description='CG33340 gene product from transcript CG33340-RA', other_designations='CG33340-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24363138, end_position_on_the_genomic_accession=24364289, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768683, CurrentID=0, Status='live', Symbol='CG33341', Aliases='Dmel_ Dmel\\CG33341', description='CG33341 gene product from transcript CG33341-RA', other_designations='CG33341-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24368358, end_position_on_the_genomic_accession=24369419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768684, CurrentID=0, Status='live', Symbol='CG33342', Aliases='Dmel_CG33342, BP1089, Dmel\\CG33342, mdcds_42691', description='HDC15381', other_designations='CG33342-PB', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24376865, end_position_on_the_genomic_accession=24377705, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768685, CurrentID=0, Status='live', Symbol='mld', Aliases='Dmel_CG34100, CG13620, CG31312, CG33343, CG34100, CG9469, DTS-3, Dmel\\CG34100, L(3)3[DTS], Mld, l(3)3[DTS], l(3)DTS3', description='molting defective', other_designations='CG34100-PA|CG34100-PC|CG34100-PD|CG34100-PE|CG34100-PF|defective|dominant temperature-sensitive 3|lethal (3) DTS3|mld-PA|mld-PC|mld-PD|mld-PE|mld-PF|molting', map_location='3-84.5 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24585653, end_position_on_the_genomic_accession=24623899, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768686, CurrentID=0, Status='live', Symbol='CG33494', Aliases='Dmel_ BcDNA:LP10657, Dmel\\CG33494', description='CG33494 gene product from transcript CG33494-RB', other_designations='CG33494-PA|CG33494-PB|uncharacterized protein', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25493481, end_position_on_the_genomic_accession=25495161, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768687, CurrentID=0, Status='live', Symbol='LpR1', Aliases='Dmel_CG31094, CG31094, CG31095, CG4861, CG4870, CG4874, CT15569, Dmel\\CG31094, lpr1', description='Lipophorin receptor 1', other_designations='CG31094-PD|CG31094-PG|CG31094-PH|CG31094-PJ|CG31094-PK|CG31094-PL|CG31094-PP|CG31094-PQ|CG31094-PR|CG31094-PS|LDL-receptor-related protein|LpR1-PD|LpR1-PG|LpR1-PH|LpR1-PJ|LpR1-PK|LpR1-PL|LpR1-PP|LpR1-PQ|LpR1-PR|LpR1-PS|lipophorin receptor 1|yolkless-like', map_location='96F1-96F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25741348, end_position_on_the_genomic_accession=25763105, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  2768688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768688, CurrentID=0, Status='live', Symbol='CCAP-R', Aliases='Dmel_CG33344, CCAP receptor, CG14547, CG33344, CG6111, CcapR, DCCAPR, Dmel\\CG33344, anon-WO0170980.46, anon-WO0170980.47', description='Crustacean cardioactive peptide receptor', other_designations='CCAP receptor|CCAP-R-PC|CG33344-PC|cardioacceleratory peptide receptor|crustacean cardioactive peptide receptor', map_location='96F8-96F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25923499, end_position_on_the_genomic_accession=25956270, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  2768690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768690, CurrentID=0, Status='live', Symbol='CG33346', Aliases='Dmel_ Dmel\\CG33346', description='CG33346 gene product from transcript CG33346-RB', other_designations='CG33346-PB|uncharacterized protein', map_location='98E1-98E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28669233, end_position_on_the_genomic_accession=28670452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768691, CurrentID=0, Status='live', Symbol='Dup99B', Aliases='Dmel_CG33495, CG33495, DUP, DUP99B, Dmel\\CG33495, anon-SAGE:Wang-013', description='Ductus ejaculatorius peptide 99B', other_designations='CG33495-PA|CG33495-PB|Dup99B-PA|Dup99B-PB|anon-SAGE:Wang-013|ductus ejaculatorius peptide 99B', map_location='99B8-99B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29715068, end_position_on_the_genomic_accession=29715377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768693, CurrentID=0, Status='live', Symbol='CG33483', Aliases='Dmel_ Dmel\\CG33483', description='CG33483 gene product from transcript CG33483-RB', other_designations='CG33483-PA|CG33483-PB|uncharacterized protein', map_location='100A7-100A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30900718, end_position_on_the_genomic_accession=30902161, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  2768714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768714, CurrentID=0, Status='live', Symbol='Ir41a', Aliases='Dmel_CG33492, CG17587, CG33492, DmelIR41a, Dmel\\CG33492, IR41a', description='Ionotropic receptor 41a', other_designations='CG33492-PC|Ir41a-PC|ionotropic receptor 41a', map_location='41C6-41D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4917519, end_position_on_the_genomic_accession=5021955, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  48971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48971, CurrentID=0, Status='live', Symbol='Atpalpha', Aliases='Dmel_CG5670, (Na[+]/k[+])-ATPase, ATP-alpha, ATPBagr, ATPa, ATPalpha, Atp a, Atpa, CG5670, Da-47, Dmel\\CG5670, JY, K ATPase, K-ATPase, K[+] -ATPase, K[+]-ATPase, K[+]-ATPase alpha subunit, K[+]-ATPase alpha-subunit, Na, Na+-K+-ATPase, Na+/K+ ATPase alpha subunit, Na-p, Na/K ATPase, Na/K-ATPase, Na/K-ATPase alpha-subunit, NaK, NaK ATPase, Na[+], Na[+]-K[+]-ATPase alpha[[5]]-subunit, Na[+]/K[+] ATPase, Na[+]/K[+]-ATPase, Na[+]/K[+]ATPase, Na[+]K[+] ATPase, Na[+]K[+]-ATPase, Na[+]K[+]ATPase, atpalpha, dnATPase, l(3)01164, l(3)01453, l(3)04492, l(3)j3B2, l(3)j3B7, l(3)j6A4, l(3)rM454, l(3)ry136, sodium/potassium transporting ATPase alpha', description='Na pump alpha subunit', other_designations='(Na[+] + K[+])-ATPase alpha-subunit|(Na[+],K[+])ATPase alpha subunit|ATPalpha|ATPase alpha|ATPase alpha subunit|Atpalpha-PA|Atpalpha-PB|Atpalpha-PC|Atpalpha-PD|Atpalpha-PE|Atpalpha-PF|Atpalpha-PG|Atpalpha-PH|Atpalpha-PI|Atpalpha-PJ|Atpalpha-PK|CG5670-PA|CG5670-PB|CG5670-PC|CG5670-PD|CG5670-PE|CG5670-PF|CG5670-PG|CG5670-PH|CG5670-PI|CG5670-PJ|CG5670-PK|Na+,K+ ATPase alpha-subunit|Na, K ATPase a subunit|Na,K ATPase|Na,K ATPase alpha subunit|Na,K-ATPase|Na,K-ATPase alpha|Na,K-ATPase alpha subunit|Na,K-ATPase alpha-subunit|Na[ +], K[ +] -ATPase|Na[+] pump alpha subunit|Na[+], K[+] ATPase alpha|Na[+],K[+] ATPase alpha subunit|Na[+],K[+]-ATPase alpha subunit|Na[+]-K[+] ATPase|Na[+]/K[+] ATPase (alpha subunit)|Na[+]/K[+] ATPase alpha subunit|Na[+]K[+] ATPase alpha subunit|Na[+]K[+]ATPase|alpha-subunit of ATPase|sodium pump alpha subunit|sodium pump alpha-subunit', map_location='93A4-93A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20948714, end_position_on_the_genomic_accession=20976239, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  2768716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768716, CurrentID=0, Status='live', Symbol='mim', Aliases='Dmel_CG33558, CG33352, CG33558, CG9469, Dm_2R:10851, Dmel\\CG33558, MIM, Mim, anon-EST:fe1E4, dmim', description='missing-in-metastasis', other_designations='CG33558-PB|CG33558-PC|CG33558-PD|CG33558-PE|CG33558-PF|CG33558-PG|CG33558-PH|CG33558-PI|CG33558-PJ|CG33558-PK|CG33558-PL|CG33558-PM|CG33558-PN|anon-fast-evolving-1E4|mim-PB|mim-PC|mim-PD|mim-PE|mim-PF|mim-PG|mim-PH|mim-PI|mim-PJ|mim-PK|mim-PL|mim-PM|mim-PN', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6924687, end_position_on_the_genomic_accession=6971637, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  48973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48973, CurrentID=0, Status='live', Symbol='Src64B', Aliases='Dmel_CG7524, C-src1, CG7524, D-Src64B, D-src, DSRC64, DSrc, DSrc64, DSrc64B, Dm SRC1, DmHD-358, Dmel\\CG7524, Dsrc, Dsrc64, Dsrc64B, HD-358, MRE5, SRC 64B, SRC64B, Src, Src1, Src64, Src64b, c-src, dSrc, dsrc, src, src1, src64, src64B, src64b', description='Src oncogene at 64B', other_designations='CG7524-PB|CG7524-PC|CG7524-PD|CG7524-PE|CG7524-PF|CG7524-PH|CG7524-PI|CG7524-PJ|CG7524-PK|Src proto oncogene sequence|Src64B-PB|Src64B-PC|Src64B-PD|Src64B-PE|Src64B-PF|Src64B-PH|Src64B-PI|Src64B-PJ|Src64B-PK|mRNA-like ncRNA in embryogenesis 5', map_location='64B11-64B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4592447, end_position_on_the_genomic_accession=4626196, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768718, CurrentID=0, Status='live', Symbol='qvr', Aliases='Dmel_CG33472, BcDNA:GH06755, CG30032, CG33472, Dmel\\CG33472, anon-WO02059370.1, mdcds_22455, sss', description='quiver', other_designations='CG33472-PB|CG33472-PC|qvr-PB|qvr-PC|sleepless', map_location='2-62 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11442505, end_position_on_the_genomic_accession=11451604, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768719, CurrentID=0, Status='live', Symbol='luna', Aliases='Dmel_CG33473, CG17326, CG33473, CG9087, DKLF, Dmel\\CG33473, KLF, Klf, LUNA, lincRNA.284', description='CG33473 gene product from transcript CG33473-RC', other_designations='CG33473-PB|CG33473-PC|CG33473-PD|luna-PB|luna-PC|luna-PD', map_location='47D1-47D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10977956, end_position_on_the_genomic_accession=11116378, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  2768720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768720, CurrentID=0, Status='live', Symbol='Cyp12d1-d', Aliases='Dmel_CG33503, 12D1, 12d1, CG18239, CG18240, CG33503, CT26082, CYP12D1, Cyp 12d1, Cyp12d1, Cyp12d2, Dmel\\CG33503', description='CG33503 gene product from transcript CG33503-RA', other_designations='CG33503-PA|Cyp12d1-d-PA|cytochrome P450|cytochrome P450-12d1', map_location='47D4-47D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11123658, end_position_on_the_genomic_accession=11125647, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768721, CurrentID=0, Status='live', Symbol='CheB42a', Aliases='Dmel_CG33348, CG15242, CG33348, CheB42A, Dmel\\CG33348', description='Chemosensory protein B 42a', other_designations='CG33348-PA|CheB42a-PA|chemosensory protein B 42a', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6949482, end_position_on_the_genomic_accession=6950284, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768722, CurrentID=0, Status='live', Symbol='ppk25', Aliases='Dmel_CG33349, CG15242, CG33349, Dmel\\CG33349, llz', description='pickpocket 25', other_designations='CG33349-PA|lounge lizard|ppk25-PA', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6950385, end_position_on_the_genomic_accession=6952158, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  2768724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768724, CurrentID=0, Status='live', Symbol='CG33475', Aliases='Dmel_ Dmel\\CG33475', description='CG33475 gene product from transcript CG33475-RA', other_designations='CG33475-PA|uncharacterized protein', map_location='47A6-47A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10410770, end_position_on_the_genomic_accession=10411213, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768725, CurrentID=0, Status='live', Symbol='CG33474', Aliases='Dmel_ Dmel\\CG33474', description='CG33474 gene product from transcript CG33474-RA', other_designations='CG33474-PA|CG33474-PB|uncharacterized protein', map_location='47A7-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10421706, end_position_on_the_genomic_accession=10422989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768726, CurrentID=0, Status='live', Symbol='CG33477', Aliases='Dmel_ Dmel\\CG33477', description='CG33477 gene product from transcript CG33477-RA', other_designations='CG33477-PA|uncharacterized protein', map_location='47A6-47A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10409155, end_position_on_the_genomic_accession=10409751, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768727, CurrentID=0, Status='live', Symbol='CG33476', Aliases='Dmel_ Dmel\\CG33476', description='CG33476 gene product from transcript CG33476-RA', other_designations='CG33476-PA|uncharacterized protein', map_location='47A6-47A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10409985, end_position_on_the_genomic_accession=10410570, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768728, CurrentID=0, Status='live', Symbol='Or46a', Aliases='Dmel_CG33478, 46F.1, 46a, 46b, AN8, AN9, CG17848, CG17849, CG33478, DOR19, DOR20, DOR46F.1, DOR46F.2, Dmel\\CG33478, Or19, Or46F.1, Or46F.2A, Or46aB, Or46aa, Or46b, dor19, Or46a', description='Odorant receptor 46a', other_designations='CG33478-PA|CG33478-PB|Or46a-PA|Or46a-PB|odorant receptor 46a|odorant receptor 46b|olfactory receptor 46F.1|olfactory receptor 46F.2', map_location='46F2-46F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10139264, end_position_on_the_genomic_accession=10142074, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  48986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=48986, CurrentID=0, Status='live', Symbol='boca', Aliases='Dmel_CG30498, Boca, CG1364, CG30498, Dmel\\CG30498, dmBoca, l(2)43Ea', description='CG30498 gene product from transcript CG30498-RA', other_designations='CG30498-PA|boca-PA|lethal(2)43Ea', map_location='43D7-43D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7565873, end_position_on_the_genomic_accession=7566738, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768732, CurrentID=0, Status='live', Symbol='CG33228', Aliases='Dmel_ CG4806, Dmel\\CG33228, l(2)SH0542, l(2)SH2 0542', description='CG33228 gene product from transcript CG33228-RB', other_designations='CG33228-PB|CG33228-PC|lethal (2) SH0542|uncharacterized protein', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24668411, end_position_on_the_genomic_accession=24669330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38730 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38730, CurrentID=0, Status='live', Symbol='CG13300', Aliases='Dmel_ Dmel\\CG13300', description='CG13300 gene product from transcript CG13300-RA', other_designations='CG13300-PA|uncharacterized protein', map_location='65A14-65A14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6329324, end_position_on_the_genomic_accession=6352390, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49045, CurrentID=0, Status='live', Symbol='Gprk2', Aliases='Dmel_CG17998, 154978_at, CG17998, Dmel\\CG17998, Gprk-2, gprk2, nosh', description='G protein-coupled receptor kinase 2', other_designations='CG17998-PA|CG17998-PB|Gprk2-PA|Gprk2-PB|nosh', map_location='100C3-100C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31405245, end_position_on_the_genomic_accession=31457873, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  49070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49070, CurrentID=0, Status='live', Symbol='Mbs', Aliases='Dmel_CG32156, 0573/06, 0953/04, CG32156, CG5600, CG5891, DMBS, DMYPT, DMbs, Dmel\\CG32156, MBS, Mypt, l(3)03802, l(3)72Dd, l(3)S005331b, l(3)S057306, l(3)S057306b, l(3)S095304, l(3)j3B4, l(3)j7A1, mbs', description='Myosin binding subunit', other_designations='CG32156-PE|CG32156-PF|CG32156-PG|CG32156-PH|CG32156-PI|CG32156-PJ|CG32156-PK|CG32156-PL|CG32156-PM|CG32156-PN|CG32156-PO|CG32156-PP|CG32156-PQ|MLC-phosphatase|Mbs-PE|Mbs-PF|Mbs-PG|Mbs-PH|Mbs-PI|Mbs-PJ|Mbs-PK|Mbs-PL|Mbs-PM|Mbs-PN|Mbs-PO|Mbs-PP|Mbs-PQ|lethal (3) S057306b|lethal (3) S095304|lethal(3)72Dd|myosin binding subunit|myosin phosphatase|myosin phosphatase target subunit|myosin phosphatase-binding subunit', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16051959, end_position_on_the_genomic_accession=16082026, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  49077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49077, CurrentID=0, Status='live', Symbol='CSN8', Aliases='Dmel_CG42522, CG13383, CG33297, CG42522, CH8, Csn8, Dch8, Dmel\\CG42522, Dmel_CG13383, csn8, l(2)SH1829, l(2)SH2 1829', description='COP9 signalosome subunit 8', other_designations='CG42522-PB|COP9 complex homolog subunit 8|CSN8-PB|drosophila COP9 signalosome homolog 8|lethal (2) SH1829', map_location='29B2-29B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8364855, end_position_on_the_genomic_accession=8365663, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  49080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49080, CurrentID=0, Status='live', Symbol='epsilonTry', Aliases='Dmel_CG18681, CG18681, DmEpsilonTry, Dmel\\CG18681, SP88, Try, epsilon', description='epsilonTrypsin', other_designations='CG18681-PA|epsilonTry-PA', map_location='47F4-47F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11345237, end_position_on_the_genomic_accession=11346066, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768831, CurrentID=0, Status='live', Symbol='CG33470', Aliases='Dmel_ Dmel\\CG33470', description='CG33470 gene product from transcript CG33470-RA', other_designations='CG33470-PA|CG33470-PC|uncharacterized protein', map_location='50A5-50A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13411739, end_position_on_the_genomic_accession=13413085, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768832, CurrentID=0, Status='live', Symbol='CG33469', Aliases='Dmel_ Dmel\\CG33469', description='CG33469 gene product from transcript CG33469-RB', other_designations='CG33469-PA|CG33469-PB|uncharacterized protein', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14605652, end_position_on_the_genomic_accession=14606302, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768833, CurrentID=0, Status='live', Symbol='CG33468', Aliases='Dmel_ Dmel\\CG33468', description='CG33468 gene product from transcript CG33468-RA', other_designations='CG33468-PA|uncharacterized protein', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14606307, end_position_on_the_genomic_accession=14606980, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  49090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49090, CurrentID=0, Status='live', Symbol='RyR', Aliases='Dmel_CG10844, CG10844, D-RyR, DRR, DRY, Dmel\\CG10844, RYRs, Rya-44F, Rya-R44F, Rya-r44F, Rya-r76CD, Ryr, dRyR, dRyr, dry, dya, l(2)k00424, l(2)k04913, rya-r44F, ryr, RyR', description='Ryanodine receptor', other_designations='CG10844-PA|CG10844-PB|CG10844-PC|CG10844-PD|CG10844-PE|CG10844-PF|CG10844-PG|CG10844-PH|CG10844-PI|CG10844-PJ|RyR-PA|RyR-PB|RyR-PC|RyR-PD|RyR-PE|RyR-PF|RyR-PG|RyR-PH|RyR-PI|RyR-PJ|ryanodine receptor|ryanodine receptor 44F', map_location='44F1-44F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8860364, end_position_on_the_genomic_accession=8888097, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  2768836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768836, CurrentID=0, Status='live', Symbol='Fs', Aliases='Dmel_CG33466, CG12955, CG12956, CG33466, Dmel\\CG33466, Fol1, dFS, dFol1, dfs, fs', description='Follistatin', other_designations='CG33466-PA|CG33466-PB|Fs-PA|Fs-PB|dFollistatin|follistatin|follistatin 1', map_location='51E7-51E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15223560, end_position_on_the_genomic_accession=15241589, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768840, CurrentID=0, Status='live', Symbol='CG33461', Aliases='Dmel_ CG30086, CG8215, Dmel\\CG33461, SP41', description='CG33461 gene product from transcript CG33461-RB', other_designations='CG33461-PB|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15695386, end_position_on_the_genomic_accession=15696478, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768841, CurrentID=0, Status='live', Symbol='CG33462', Aliases='Dmel_ CG30086, CG8215, Dmel\\CG33462, SP41', description='CG33462 gene product from transcript CG33462-RA', other_designations='CG33462-PA|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15696794, end_position_on_the_genomic_accession=15698010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768844, CurrentID=0, Status='live', Symbol='CG33463', Aliases='Dmel_ Dmel\\CG33463', description='CG33463 gene product from transcript CG33463-RA', other_designations='CG33463-PA|uncharacterized protein', map_location='52D1-52D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15895829, end_position_on_the_genomic_accession=15896427, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768846, CurrentID=0, Status='live', Symbol='CG33458', Aliases='Dmel_ Dmel\\CG33458', description='CG33458 gene product from transcript CG33458-RA', other_designations='CG33458-PA|uncharacterized protein', map_location='53C6-53C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16486674, end_position_on_the_genomic_accession=16487801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768847, CurrentID=0, Status='live', Symbol='CG33459', Aliases='Dmel_ Dmel\\CG33459, HDC06756', description='CG33459 gene product from transcript CG33459-RA', other_designations='CG33459-PA|uncharacterized protein', map_location='53C6-53C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16488053, end_position_on_the_genomic_accession=16489182, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768848, CurrentID=0, Status='live', Symbol='mute', Aliases='Dmel_CG34415, CG15610, CG33455, CG33456, CG34415, Dmel\\CG34415, Dmel_CG33455, Dmel_CG33456, Mute', description='muscle wasted', other_designations='CG34415-PC|CG34415-PD|mute-PC|mute-PD', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16948296, end_position_on_the_genomic_accession=16955266, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  2768850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768850, CurrentID=0, Status='live', Symbol='CG33454', Aliases='Dmel_ Dmel\\CG33454', description='CG33454 gene product from transcript CG33454-RA', other_designations='CG33454-PA|uncharacterized protein', map_location='56C4-56C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19251027, end_position_on_the_genomic_accession=19251666, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768851, CurrentID=0, Status='live', Symbol='CG33453', Aliases='Dmel_ Dmel\\CG33453', description='CG33453 gene product from transcript CG33453-RB', other_designations='CG33453-PA|CG33453-PB|uncharacterized protein', map_location='56C5-56C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19252501, end_position_on_the_genomic_accession=19253228, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768852, CurrentID=0, Status='live', Symbol='par-1', Aliases='Dmel_CG8201, 27C1, BcDNA:RH48823, CG11960, CG16701, CG30131, CG30132, CG8201, DPAR-1, DPar1, Dmel\\CG8201, EMK, MARK, PAR-1, PAR1, PAR1/MARK2, Par-1, Par1, anon-WO0210402.19, dMARK, dPAR-1, dPAR1, dPar-1, dPar1, l(2)27C1, l(2)k06323, par1', description='CG8201 gene product from transcript CG8201-RV', other_designations='C-TAK1-like|CG8201-PA|CG8201-PH|CG8201-PL|CG8201-PP|CG8201-PQ|CG8201-PR|CG8201-PS|CG8201-PT|CG8201-PV|CG8201-PW|CG8201-PX|ELKL motif kinase|Par-1 kinase|Par-1 protein kinase|lethal (2) 27C1|par-1-PA|par-1-PH|par-1-PL|par-1-PP|par-1-PQ|par-1-PR|par-1-PS|par-1-PT|par-1-PV|par-1-PW|par-1-PX|partitioning defective-1', map_location='56D9-57D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19456170, end_position_on_the_genomic_accession=19485686, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  2768853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768853, CurrentID=0, Status='live', Symbol='mei-W68', Aliases='Dmel_CG7753, CG30132, CG7753, Dmel\\CG7753, Mei-W68, MeiW68, W68, l(2)k05603, mei, mei W-68, meiW68, spo11', description='meiotic W68', other_designations='CG7753-PA|mei-W68-PA', map_location='2-94 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19456170, end_position_on_the_genomic_accession=19459523, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768858, CurrentID=0, Status='live', Symbol='dpr1', Aliases='Dmel_CG13439, CG13439, DPR1, Dmel\\CG13439, Dpr-1, Dpr1, dpr', description='defective proboscis extension response 1', other_designations='CG13439-PA|CG13439-PB|defective proboscis extension response|dpr1-PA|dpr1-PB', map_location='57B1-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20698483, end_position_on_the_genomic_accession=20750442, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768859, CurrentID=0, Status='live', Symbol='CG33226', Aliases='Dmel_ CG10450, Dmel\\CG33226, SP37, SPH37', description='CG33226 gene product from transcript CG33226-RC', other_designations='CG33226-PC|uncharacterized protein', map_location='57E11-57E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21538250, end_position_on_the_genomic_accession=21539293, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768860, CurrentID=0, Status='live', Symbol='CG33225', Aliases='Dmel_ Dmel\\CG33225', description='CG33225 gene product from transcript CG33225-RC', other_designations='CG33225-PB|CG33225-PC|uncharacterized protein', map_location='57F2-57F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21590275, end_position_on_the_genomic_accession=21591584, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768865, CurrentID=0, Status='live', Symbol='dpr7', Aliases='Dmel_CG33481, CG33481, Dmel\\CG33481, Dpr-7, Dpr7', description='defective proboscis extension response 7', other_designations='CG33481-PF|CG33481-PG|dpr7-PF|dpr7-PG', map_location='102B1-102B1', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=247124, end_position_on_the_genomic_accession=253329, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  2768867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768867, CurrentID=0, Status='live', Symbol='CG33223', Aliases='Dmel_ Dmel\\CG33223', description='CG33223 gene product from transcript CG33223-RA', other_designations='CG33223-PA|CG33223-PB|uncharacterized protein', map_location='7F1-7F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8465690, end_position_on_the_genomic_accession=8483885, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768869, CurrentID=0, Status='live', Symbol='CG34330', Aliases='Dmel_ CG33254, Dmel\\CG34330', description='CG34330 gene product from transcript CG34330-RB', other_designations='CG34330-PA|CG34330-PB|uncharacterized protein', map_location='18A2-18A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19068275, end_position_on_the_genomic_accession=19068973, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768870, CurrentID=0, Status='live', Symbol='AlkB', Aliases='Dmel_CG33250, BEST:LD02396, CG33250, Dmel\\CG33250, alkB', description='CG33250 gene product from transcript CG33250-RA', other_designations='AlkB-PA|CG33250-PA', map_location='13F14-13F16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15858134, end_position_on_the_genomic_accession=15859379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768871 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768871, CurrentID=0, Status='live', Symbol='CG33235', Aliases='Dmel_ CG15192, Dmel\\CG33235', description='CG33235 gene product from transcript CG33235-RA', other_designations='CG33235-PA|uncharacterized protein', map_location='10B17-10C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11486546, end_position_on_the_genomic_accession=11491573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768872, CurrentID=0, Status='live', Symbol='Ste:CG33237', Aliases='Dmel_CG33237, CG32605, CG33237, Dmel\\CG33237', description='CG33237 gene product from transcript CG33237-RB', other_designations='CG33237-PB|Ste:CG33237-PB', map_location='12D4-12D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14059434, end_position_on_the_genomic_accession=14060157, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768873, CurrentID=0, Status='live', Symbol='Ste:CG33236', Aliases='Dmel_CG33236, CG32605, CG33236, Dmel\\CG33236', description='CG33236 gene product from transcript CG33236-RB', other_designations='CG33236-PB|Ste:CG33236-PB', map_location='12D4-12D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14060702, end_position_on_the_genomic_accession=14061817, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768875, CurrentID=0, Status='live', Symbol='CG33502', Aliases='Dmel_ BcDNA:GM32035, CG18856, Dmel\\CG33502', description='CG33502 gene product from transcript CG33502-RA', other_designations='CG33502-PA|uncharacterized protein', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21609357, end_position_on_the_genomic_accession=21611526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768877, CurrentID=0, Status='live', Symbol='CG5541', Aliases='Dmel_ 5541, Dmel\\CG5541', description='CG5541 gene product from transcript CG5541-RA', other_designations='CG5541-PA|CG5541-PB', map_location='13A5-13A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15044123, end_position_on_the_genomic_accession=15051842, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768879, CurrentID=0, Status='live', Symbol='CG32500', Aliases='Dmel_ BcDNA:GM32035, CG18856, Dmel\\CG32500', description='CG32500 gene product from transcript CG32500-RA', other_designations='CG32500-PA|uncharacterized protein', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21621541, end_position_on_the_genomic_accession=21623750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768881, CurrentID=0, Status='live', Symbol='CG9132', Aliases='Dmel_ CG32566, Dmel\\CG9132', description='CG9132 gene product from transcript CG9132-RA', other_designations='CG9132-PA|CG9132-PB|CG9132-PC', map_location='15A8-15A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16796902, end_position_on_the_genomic_accession=16800710, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2768882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768882, CurrentID=0, Status='live', Symbol='CG9125', Aliases='Dmel_ CG32566, Dmel\\CG9125', description='CG9125 gene product from transcript CG9125-RA', other_designations='CG9125-PA', map_location='15A9-15A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16801929, end_position_on_the_genomic_accession=16803344, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768883, CurrentID=0, Status='live', Symbol='Sdic2', Aliases='Dmel_CG33497, CG33497, Dmel\\CG33497, Sdic:CG33497', description='Sperm-specific dynein intermediate chain 2', other_designations='CG33497-PA|CG33497-PB|CG33497-PC|Sdic2-PA|Sdic2-PB|Sdic2-PC|Sdic:CG33497', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20181590, end_position_on_the_genomic_accession=20186487, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2768886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768886, CurrentID=0, Status='live', Symbol='Sdic4', Aliases='Dmel_CG33499, CG33499, Dmel\\CG33499, Sdic:CG33499', description='Sperm-specific dynein intermediate chain 4', other_designations='CG33499-PE|Sdic4-PE|Sdic:CG33499', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20220666, end_position_on_the_genomic_accession=20225020, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  2768889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768889, CurrentID=0, Status='live', Symbol='CG33253', Aliases='Dmel_ BEST:CK01423, CG14192, CG7633, CK01423, Dmel\\CG33253', description='CG33253 gene product from transcript CG33253-RA', other_designations='CG33253-PA|CG33253-PB|CG33253-PC|uncharacterized protein', map_location='18A5-18A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19183421, end_position_on_the_genomic_accession=19187278, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768892, CurrentID=0, Status='live', Symbol='CG42271', Aliases='Dmel_ CG1846, CG33248, CG33249, Dmel\\CG42271, Dmel_CG33248, Dmel_CG33249, FBgn0259166, INPP4A', description='CG42271 gene product from transcript CG42271-RC', other_designations='CG42271-PB|CG42271-PC|uncharacterized protein', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14228060, end_position_on_the_genomic_accession=14234066, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768895, CurrentID=0, Status='live', Symbol='CG34332', Aliases='Dmel_ Dmel\\CG34332', description='CG34332 gene product from transcript CG34332-RA', other_designations='CG34332-PA|uncharacterized protein', map_location='19B3-19B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20141010, end_position_on_the_genomic_accession=20141476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768896, CurrentID=0, Status='live', Symbol='Ste:CG33244', Aliases='Dmel_CG33244, CG32605, CG33244, Dmel\\CG33244', description='CG33244 gene product from transcript CG33244-RB', other_designations='CG33244-PB|Ste:CG33244-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14050549, end_position_on_the_genomic_accession=14051272, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768897, CurrentID=0, Status='live', Symbol='Ste:CG33243', Aliases='Dmel_CG33243, CG32605, CG33243, Dmel\\CG33243', description='CG33243 gene product from transcript CG33243-RB', other_designations='CG33243-PB|Ste:CG33243-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14051818, end_position_on_the_genomic_accession=14052541, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768898, CurrentID=0, Status='live', Symbol='Ste:CG33242', Aliases='Dmel_CG33242, CG32605, CG33242, Dmel\\CG33242', description='CG33242 gene product from transcript CG33242-RB', other_designations='CG33242-PB|Ste:CG33242-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14053087, end_position_on_the_genomic_accession=14053811, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768899, CurrentID=0, Status='live', Symbol='Ste:CG33241', Aliases='Dmel_CG33241, CG32605, CG33241, Dmel\\CG33241', description='CG33241 gene product from transcript CG33241-RB', other_designations='CG33241-PB|Ste:CG33241-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14054357, end_position_on_the_genomic_accession=14055080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768900, CurrentID=0, Status='live', Symbol='Ste:CG33240', Aliases='Dmel_CG33240, CG32605, CG33240, Dmel\\CG33240', description='CG33240 gene product from transcript CG33240-RB', other_designations='CG33240-PB|Ste:CG33240-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14055626, end_position_on_the_genomic_accession=14056349, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768901, CurrentID=0, Status='live', Symbol='Ste:CG33239', Aliases='Dmel_CG33239, CG32605, CG33239, Dmel\\CG33239', description='CG33239 gene product from transcript CG33239-RB', other_designations='CG33239-PB|Ste:CG33239-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14056895, end_position_on_the_genomic_accession=14057618, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768902, CurrentID=0, Status='live', Symbol='Ste:CG33238', Aliases='Dmel_CG33238, CG32605, CG33238, Dmel\\CG33238', description='CG33238 gene product from transcript CG33238-RB', other_designations='CG33238-PB|Ste:CG33238-PB', map_location='12D3-12D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14058164, end_position_on_the_genomic_accession=14058889, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768903, CurrentID=0, Status='live', Symbol='CG33252', Aliases='Dmel_ Dmel\\CG33252', description='CG33252 gene product from transcript CG33252-RA', other_designations='CG33252-PA|uncharacterized protein', map_location='15B5-15B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16896215, end_position_on_the_genomic_accession=16897187, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768904, CurrentID=0, Status='live', Symbol='Ste:CG33247', Aliases='Dmel_CG33247, CG32605, CG33247, Dmel\\CG33247', description='CG33247 gene product from transcript CG33247-RB', other_designations='CG33247-PB|Ste:CG33247-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14046772, end_position_on_the_genomic_accession=14047495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768905, CurrentID=0, Status='live', Symbol='Ste:CG33246', Aliases='Dmel_CG33246, CG32605, CG33246, Dmel\\CG33246', description='CG33246 gene product from transcript CG33246-RB', other_designations='CG33246-PB|Ste:CG33246-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14048011, end_position_on_the_genomic_accession=14048734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768906, CurrentID=0, Status='live', Symbol='Ste:CG33245', Aliases='Dmel_CG33245, CG32605, CG33245, Dmel\\CG33245', description='CG33245 gene product from transcript CG33245-RB', other_designations='CG33245-PB|Ste:CG33245-PB', map_location='12D3-12D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14049280, end_position_on_the_genomic_accession=14050003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768908, CurrentID=0, Status='live', Symbol='CG3857', Aliases='Dmel_ Dmel\\CG3857, EG:39E1.3', description='CG3857 gene product from transcript CG3857-RA', other_designations='CG3857-PA', map_location='2B16-2B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1937700, end_position_on_the_genomic_accession=1940040, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  49165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49165, CurrentID=0, Status='live', Symbol='Cyp303a1', Aliases='Dmel_CG4163, 303a1, BG:DS02740.6, CG4163, CYP303A1-Dm, Cyp303A1, Dmel\\CG4163, I(2)35Fb, l(2)35Fb, nompH', description='CG4163 gene product from transcript CG4163-RB', other_designations='CG4163-PA|CG4163-PB|Cyp303a1-PA|Cyp303a1-PB|lethal(2)35Fb', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16294183, end_position_on_the_genomic_accession=16296969, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768910, CurrentID=0, Status='live', Symbol='CG33310', Aliases='Dmel_ BG:DS00929.6, CG15274, Dmel\\CG33310', description='CG33310 gene product from transcript CG33310-RB', other_designations='CG33310-PB|uncharacterized protein', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15026882, end_position_on_the_genomic_accession=15029513, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768911, CurrentID=0, Status='live', Symbol='CG13970', Aliases='Dmel_ 38B.19, Dmel\\CG13970', description='CG13970 gene product from transcript CG13970-RA', other_designations='CG13970-PA|uncharacterized protein', map_location='38C1-38C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20103386, end_position_on_the_genomic_accession=20104150, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  49168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49168, CurrentID=0, Status='live', Symbol='l(2)37Cc', Aliases='Dmel_CG10691, CG10691, Cc, Dmel\\CG10691, fs(2)HH32, l (2) 37Cc, l(2)Cc', description='lethal (2) 37Cc', other_designations='CG10691-PA|CG10691-PB|CG10691-PD|l(2)37Cc-PA|l(2)37Cc-PB|l(2)37Cc-PD|transcription unit No. 4', map_location='37C1-37C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19121345, end_position_on_the_genomic_accession=19123869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768914, CurrentID=0, Status='live', Symbol='CG33307', Aliases='Dmel_ BG:DS00941.11, BcDNA:GH07914, CG16887, DS00941.11, Dmel\\CG33307', description='CG33307 gene product from transcript CG33307-RB', other_designations='CG33307-PA|CG33307-PB|uncharacterized protein', map_location='34D7-34D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13839262, end_position_on_the_genomic_accession=13841520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768915, CurrentID=0, Status='live', Symbol='CG33306', Aliases='Dmel_ BG:DS00941.11, BcDNA:GM14937, CG16887, DS00941.11, Dmel\\CG33306', description='CG33306 gene product from transcript CG33306-RA', other_designations='CG33306-PA|uncharacterized protein', map_location='34D7-34D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13842888, end_position_on_the_genomic_accession=13843742, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768916, CurrentID=0, Status='live', Symbol='CG33303', Aliases='Dmel_ CG5364, Dmel\\CG33303, OST', description='CG33303 gene product from transcript CG33303-RB', other_designations='CG33303-PA|CG33303-PB|uncharacterized protein', map_location='31D11-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10373369, end_position_on_the_genomic_accession=10375757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768917, CurrentID=0, Status='live', Symbol='CG33301', Aliases='Dmel_ CG4791, Dmel\\CG33301', description='CG33301 gene product from transcript CG33301-RB', other_designations='CG33301-PA|CG33301-PB|uncharacterized protein', map_location='31A1-31A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10049580, end_position_on_the_genomic_accession=10051201, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768918, CurrentID=0, Status='live', Symbol='Cpr31A', Aliases='Dmel_CG33302, CG33302, CG4791, CG4791-PB, Dmel\\CG33302', description='Cuticular protein 31A', other_designations='CG33302-PA|Cpr31A-PA|cuticular protein 31A', map_location='31A1-31A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10054213, end_position_on_the_genomic_accession=10055935, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768920, CurrentID=0, Status='live', Symbol='squ', Aliases='Dmel_CG4711, CG4711, Dmel\\CG4711, SQU, atg, mat(2)cell-A, mat(2)cellHK35', description='squash', other_designations='CG4711-PA|CG4711-PB|CG4711-PC|adjacent-to-grapes|squ-PA|squ-PB|squ-PC', map_location='2-54 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16680054, end_position_on_the_genomic_accession=16681653, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768923, CurrentID=0, Status='live', Symbol='CG18636', Aliases='Dmel_ Dmel\\CG18636, SPH130', description='CG18636 gene product from transcript CG18636-RA', other_designations='CG18636-PA|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14783301, end_position_on_the_genomic_accession=14784598, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768924, CurrentID=0, Status='live', Symbol='CG18420', Aliases='Dmel_ BG:DS06874.6, DS06874.6, Dmel\\CG18420, SPH111', description='CG18420 gene product from transcript CG18420-RA', other_designations='CG18420-PA|uncharacterized protein', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14794516, end_position_on_the_genomic_accession=14796642, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768925, CurrentID=0, Status='live', Symbol='CG33308', Aliases='Dmel_ BP1076, Dmel\\CG33308, HDC02577, anon-SAGE:Wang-042', description='CG33308 gene product from transcript CG33308-RA', other_designations='CG33308-PA|HDC02577|anon-SAGE:Wang-042|uncharacterized protein', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14802334, end_position_on_the_genomic_accession=14803293, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768926, CurrentID=0, Status='live', Symbol='CG33309', Aliases='Dmel_ BG:DS06874.7, DS06874.7, Dmel\\CG33309, NEST:bs22a11', description='CG33309 gene product from transcript CG33309-RA', other_designations='CG33309-PA|uncharacterized protein', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14810761, end_position_on_the_genomic_accession=14812599, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768929, CurrentID=0, Status='live', Symbol='CG33298', Aliases='Dmel_ CG13112, CG18419, DmCG33298, Dmel\\CG33298', description='CG33298 gene product from transcript CG33298-RB', other_designations='CG33298-PA|CG33298-PB|CG33298-PC|CG33298-PD|uncharacterized protein', map_location='30B5-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9495474, end_position_on_the_genomic_accession=9520696, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  2768931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768931, CurrentID=0, Status='live', Symbol='Muc30E', Aliases='Dmel_CG33300, CG33300, Dmel\\CG33300', description='Mucin 30E', other_designations='CG33300-PC|Muc30E-PC|mucin 30E', map_location='30E1-30E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9897820, end_position_on_the_genomic_accession=9903128, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768932, CurrentID=0, Status='live', Symbol='CG33322', Aliases='Dmel_ 38E.23, BcDNA:AT04844, CG9329, Dmel\\CG33322', description='CG33322 gene product from transcript CG33322-RA', other_designations='CG33322-PA|uncharacterized protein', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20814337, end_position_on_the_genomic_accession=20816240, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2768933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768933, CurrentID=0, Status='live', Symbol='CheB38b', Aliases='Dmel_CG33321, 38E.23, CG33321, CG9329, Dmel\\CG33321', description='Chemosensory protein B 38b', other_designations='CG33321-PA|CheB38b-PA|chemosensory protein B 38b', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20816456, end_position_on_the_genomic_accession=20817270, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768934, CurrentID=0, Status='live', Symbol='CheB38a', Aliases='Dmel_CG33320, 38E.23, CG33320, CG9329, Dmel\\CG33320', description='Chemosensory protein B 38a', other_designations='CG33320-PA|CheB38a-PA|chemosensory protein B 38a', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20817815, end_position_on_the_genomic_accession=20818532, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768938, CurrentID=0, Status='live', Symbol='CG33281', Aliases='Dmel_ CG15407, Dmel\\CG33281', description='CG33281 gene product from transcript CG33281-RA', other_designations='CG33281-PA|uncharacterized protein', map_location='23E4-23E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3324237, end_position_on_the_genomic_accession=3326637, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  2768939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768939, CurrentID=0, Status='live', Symbol='CG33282', Aliases='Dmel_ CG15407, Dmel\\CG33282', description='CG33282 gene product from transcript CG33282-RB', other_designations='CG33282-PB|uncharacterized protein', map_location='23E4-23E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3327903, end_position_on_the_genomic_accession=3329792, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  2768940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768940, CurrentID=0, Status='live', Symbol='Pp2A-29B', Aliases='Dmel_CG17291, 29B, A/PR65, CG13383, CG17291, CG33297, DPR65, Dmel\\CG17291, PP2A, PP2A 29B, PP2A A, PP2A-29B, PP2A-A, PP2A[A], PR65, Pp2A, Pp2A29B, dPP2A A, pp2A-29B', description='Protein phosphatase 2A at 29B', other_designations='65 kDa regulatory subunit of protein phosphatase|CG17291-PA|CG17291-PB|CG17291-PC|CG17291-PD|CG17291-PE|PP2A 65kD subunit A|Pp2A-29B-PA|Pp2A-29B-PB|Pp2A-29B-PC|Pp2A-29B-PD|Pp2A-29B-PE|protein phosphatase 2A|protein phosphatase 2A at 29B|regulatory subunit A of type 2A protein phosphatase', map_location='29B2-29B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8366038, end_position_on_the_genomic_accession=8370090, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768941, CurrentID=0, Status='live', Symbol='CG33296', Aliases='Dmel_ Dmel\\CG33296', description='CG33296 gene product from transcript CG33296-RC', other_designations='CG33296-PB|CG33296-PC|uncharacterized protein', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7713998, end_position_on_the_genomic_accession=7716645, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768944, CurrentID=0, Status='live', Symbol='rho-5', Aliases='Dmel_CG33304, CG33304, CG5364, CT17046, Dmel\\CG33304, Rho-related[31D10], iRhom', description='rhomboid-5', other_designations='CG33304-PA|rho-5-PA', map_location='31D10-31D11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10368025, end_position_on_the_genomic_accession=10373900, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  2768945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768945, CurrentID=0, Status='live', Symbol='CG42458', Aliases='Dmel_ CG18656, CG33277, CG33279, Dmel\\CG42458, Dmel_CG33277, Dmel_CG33279, cg18656', description='CG42458 gene product from transcript CG42458-RA', other_designations='CG42458-PA|uncharacterized protein', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7423827, end_position_on_the_genomic_accession=7454717, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768946, CurrentID=0, Status='live', Symbol='CG33278', Aliases='Dmel_ CG18656, Dmel\\CG33278, cg18656', description='CG33278 gene product from transcript CG33278-RA', other_designations='CG33278-PA|uncharacterized protein', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7450853, end_position_on_the_genomic_accession=7452159, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768948, CurrentID=0, Status='live', Symbol='CG33275', Aliases='Dmel_ CG33167, CG7511, CG7513, CG7515, CG7520, Dmel\\CG33275', description='CG33275 gene product from transcript CG33275-RA', other_designations='CG33275-PA|CG33275-PB|CG33275-PC|CG33275-PD|uncharacterized protein', map_location='66A6-66A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7593091, end_position_on_the_genomic_accession=7639573, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  2768949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768949, CurrentID=0, Status='live', Symbol='CG33276', Aliases='Dmel_ Dmel\\CG33276, anon-AE003558.1', description='CG33276 gene product from transcript CG33276-RB', other_designations='CG33276-PB|uncharacterized protein', map_location='66A8-66A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7639811, end_position_on_the_genomic_accession=7640446, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2768950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768950, CurrentID=0, Status='live', Symbol='CG33259', Aliases='Dmel_ CG7489, Dmel\\CG33259', description='CG33259 gene product from transcript CG33259-RA', other_designations='CG33259-PA|uncharacterized protein', map_location='71C4-71C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15407832, end_position_on_the_genomic_accession=15408267, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768951, CurrentID=0, Status='live', Symbol='CG33217', Aliases='Dmel_ BcDNA:RE15216, Dmel\\CG33217', description='CG33217 gene product from transcript CG33217-RA', other_designations='CG33217-PA|uncharacterized protein', map_location='80D2-80D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23126025, end_position_on_the_genomic_accession=23132209, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  2768952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768952, CurrentID=0, Status='live', Symbol='CG33263', Aliases='Dmel_ CG14106, Dmel\\CG33263', description='CG33263 gene product from transcript CG33263-RA', other_designations='CG33263-PA|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13423320, end_position_on_the_genomic_accession=13424168, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768954, CurrentID=0, Status='live', Symbol='CG12355', Aliases='Dmel_ Dmel\\CG12355', description='CG12355 gene product from transcript CG12355-RD', other_designations='CG12355-PA|CG12355-PB|CG12355-PD|CG12355-PE|uncharacterized protein', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15492964, end_position_on_the_genomic_accession=15495549, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  2768955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768955, CurrentID=0, Status='live', Symbol='Obp73a', Aliases='Dmel_CG13027, CG13027, DmelObp73a, Dmel\\CG13027', description='Odorant-binding protein 73a', other_designations='CG13027-PB|Obp73a-PB', map_location='73E1-73E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16944423, end_position_on_the_genomic_accession=16945878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  2768957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768957, CurrentID=0, Status='live', Symbol='CG33500', Aliases='Dmel_ Dmel\\CG33500', description='CG33500 gene product from transcript CG33500-RA', other_designations='CG33500-PA|inverted repeat 1|uncharacterized protein', map_location='68C11-68C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11508637, end_position_on_the_genomic_accession=11508903, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768958, CurrentID=0, Status='live', Symbol='CG33272', Aliases='Dmel_ CG17981, Dmel\\CG33272, HDC09626', description='CG33272 gene product from transcript CG33272-RA', other_designations="CG33272-PA|inverted repeat 1'|uncharacterized protein", map_location='68C11-68C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11510323, end_position_on_the_genomic_accession=11510589, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768960, CurrentID=0, Status='live', Symbol='CG33257', Aliases='Dmel_ CG13037, Dmel\\CG33257', description='CG33257 gene product from transcript CG33257-RA', other_designations='CG33257-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16346323, end_position_on_the_genomic_accession=16347620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768961, CurrentID=0, Status='live', Symbol='CG7369', Aliases='Dmel_ Dmel\\CG7369', description='CG7369 gene product from transcript CG7369-RA', other_designations='CG7369-PA', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22741793, end_position_on_the_genomic_accession=22748614, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768962, CurrentID=0, Status='live', Symbol='CG6244', Aliases='Dmel_ Dmel\\CG6244', description='CG6244 gene product from transcript CG6244-RA', other_designations='CG6244-PA', map_location='72A1-72A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15808322, end_position_on_the_genomic_accession=15808883, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768963, CurrentID=0, Status='live', Symbol='CG13064', Aliases='Dmel_ Dmel\\CG13064', description='CG13064 gene product from transcript CG13064-RA', other_designations='CG13064-PA|uncharacterized protein', map_location='72D12-72D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16280404, end_position_on_the_genomic_accession=16280860, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768964, CurrentID=0, Status='live', Symbol='Or69a', Aliases='Dmel_CG33264, 69F.1, 69a, 69b, CG17902, CG32116, CG33264, DOR82, DOR83, Dmel\\CG33264, Or69AA, Or69aB, Or69ab, Or69b, or69a, Or69a', description='Odorant receptor 69a', other_designations='CG33264-PA|CG33264-PB|Or69a-PA|Or69a-PB|odorant receptor 69a|odorant receptor 69b', map_location='69F2-69F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12964360, end_position_on_the_genomic_accession=12967733, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2768965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768965, CurrentID=0, Status='live', Symbol='AsnS', Aliases='Dmel_CG33486, AS, CG33486, Dmel\\CG33486, asparagine-synthetase', description='Asparagine synthetase', other_designations='AsnS-PA|CG33486-PA|asparagine synthetase|asparagine-synthetase', map_location='78C1-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21173118, end_position_on_the_genomic_accession=21174973, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768967, CurrentID=0, Status='live', Symbol='CG13028', Aliases='Dmel_ Dmel\\CG13028', description='CG13028 gene product from transcript CG13028-RA', other_designations='CG13028-PA|uncharacterized protein', map_location='73D6-73D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16923252, end_position_on_the_genomic_accession=16923994, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768968, CurrentID=0, Status='live', Symbol='CG33288', Aliases='Dmel_ Dmel\\CG33288', description='CG33288 gene product from transcript CG33288-RC', other_designations='CG33288-PC|uncharacterized protein', map_location='78D2-78D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21427790, end_position_on_the_genomic_accession=21433066, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768969, CurrentID=0, Status='live', Symbol='CG33267', Aliases='Dmel_ Dmel\\CG33267', description='CG33267 gene product from transcript CG33267-RB', other_designations='CG33267-PB|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11564483, end_position_on_the_genomic_accession=11572796, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768970, CurrentID=0, Status='live', Symbol='CG33490', Aliases='Dmel_ CG6110, Dmel\\CG33490', description='CG33490 gene product from transcript CG33490-RA', other_designations='CG33490-PA|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11572797, end_position_on_the_genomic_accession=11574663, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  2768971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768971, CurrentID=0, Status='live', Symbol='CG33489', Aliases='Dmel_ CG6110, Dmel\\CG33489', description='CG33489 gene product from transcript CG33489-RA', other_designations='CG33489-PA|CG33489-PB|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11574764, end_position_on_the_genomic_accession=11577044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  49228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49228, CurrentID=0, Status='live', Symbol='mod(mdg4)', Aliases='Dmel_CG32491, BcDNA:GH07769, BcDNA:SD03001, CG15500, CG15501, CG15802, CG18151, CG32491, CG7836, CG7859, CG8076, Dmel\\CG32491, Doom, E(var)129, E(var)3-93D, E(var)93D, E-(var)3-93D, E-var(3)1, E-var(3)3, MOD(MDG4), MOD(MDG4)56.3, MOD2.2, Mod(mdg 4), Mod(mdg 4)67.2, Mod(mdg)4, Mod(mdg4), Mod(mdg4)-67.2, Mod(mdg4)2.2, Mod(mdg4)67.2, Mod2.2, Mod67.2, Mod[mdg4], bpd, doom, l(3)03852, l(3)L3101, l(3)j2B7, mdg4, mnm, mnn, mod, mod (mdg4), mod(gypsy), mod(mdg-4), mod(mdg4-67.2, mod(mdg4)67.2, mod2.2, modMDG4, mod[mdg4], pf-2, mod(mdg4)', description='modifier of mdg4', other_designations='CG32491-PA|CG32491-PAA|CG32491-PAB|CG32491-PAC|CG32491-PAD|CG32491-PAE|CG32491-PAF|CG32491-PB|CG32491-PC|CG32491-PD|CG32491-PE|CG32491-PF|CG32491-PG|CG32491-PH|CG32491-PI|CG32491-PJ|CG32491-PK|CG32491-PL|CG32491-PM|CG32491-PN|CG32491-PO|CG32491-PP|CG32491-PR|CG32491-PS|CG32491-PT|CG32491-PU|CG32491-PV|CG32491-PW|CG32491-PX|CG32491-PY|CG32491-PZ|Modifier67.2|branch point disrupted|enhancer of variegation 93D|mod(mdg4 isoform N)|mod(mdg4)-PA|mod(mdg4)-PAA|mod(mdg4)-PAB|mod(mdg4)-PAC|mod(mdg4)-PAD|mod(mdg4)-PAE|mod(mdg4)-PAF|mod(mdg4)-PB|mod(mdg4)-PC|mod(mdg4)-PD|mod(mdg4)-PE|mod(mdg4)-PF|mod(mdg4)-PG|mod(mdg4)-PH|mod(mdg4)-PI|mod(mdg4)-PJ|mod(mdg4)-PK|mod(mdg4)-PL|mod(mdg4)-PM|mod(mdg4)-PN|mod(mdg4)-PO|mod(mdg4)-PP|mod(mdg4)-PR|mod(mdg4)-PS|mod(mdg4)-PT|mod(mdg4)-PU|mod(mdg4)-PV|mod(mdg4)-PW|mod(mdg4)-PX|mod(mdg4)-PY|mod(mdg4)-PZ|modifier mdg4|modifier of (mdg4) 67.2|modifier of mdg 4 67.2|modifier of mdg4 67.2|modifier of mdg4 in meiosis|modifier of midget4', map_location='3-70.7 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21350846, end_position_on_the_genomic_accession=21377401, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  2768973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768973, CurrentID=0, Status='live', Symbol='Rpn12R', Aliases='Dmel_CG11552, CG11552, Dmel\\CG11552', description='Regulatory particle non-ATPase 12-related', other_designations='CG11552-PA|Rpn12R-PA|regulatory particle non-ATPase 12-related', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15041631, end_position_on_the_genomic_accession=15042570, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768974, CurrentID=0, Status='live', Symbol='ind', Aliases='Dmel_CG11551, CG11551, Dmel\\CG11551, Ind/Gsx, ind', description='intermediate neuroblasts defective', other_designations='CG11551-PB|ind-PB|intermediate nerve cord defective', map_location='71B2-71B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15042799, end_position_on_the_genomic_accession=15043924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768975, CurrentID=0, Status='live', Symbol='CG33262', Aliases='Dmel_ Dmel\\CG33262', description='CG33262 gene product from transcript CG33262-RB', other_designations='CG33262-PB|uncharacterized protein', map_location='70A1-70A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13136269, end_position_on_the_genomic_accession=13137066, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  2768976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768976, CurrentID=0, Status='live', Symbol='CG32318', Aliases='Dmel_CG32318-PA, CR32318, Dmel\\CG32318', description='CG32318 gene product from transcript CG32318-RA', other_designations='CG32318-PA|uncharacterized protein', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1246073, end_position_on_the_genomic_accession=1246923, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768977, CurrentID=0, Status='live', Symbol='CG33260', Aliases='Dmel_ CG13470, Dmel\\CG33260', description='CG33260 gene product from transcript CG33260-RC', other_designations='CG33260-PC|uncharacterized protein', map_location='70F4-70F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14760520, end_position_on_the_genomic_accession=14761844, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  2768978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768978, CurrentID=0, Status='live', Symbol='CG7377', Aliases='Dmel_ Dmel\\CG7377, HDC09644', description='CG7377 gene product from transcript CG7377-RA', other_designations='CG7377-PA', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11589695, end_position_on_the_genomic_accession=11589979, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768979, CurrentID=0, Status='live', Symbol='CG33268', Aliases='Dmel_ CG14138, Dmel\\CG33268', description='CG33268 gene product from transcript CG33268-RA', other_designations='CG33268-PA|uncharacterized protein', map_location='68C14-68C14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11590872, end_position_on_the_genomic_accession=11591231, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768980, CurrentID=0, Status='live', Symbol='Muc68Ca', Aliases='Dmel_CG18331, CG18331, Dmel\\CG18331, cg18331', description='Mucin 68Ca', other_designations='CG18331-PA|Muc68Ca-PA|mucin 68Ca', map_location='68C15-68C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11610578, end_position_on_the_genomic_accession=11620101, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768981, CurrentID=0, Status='live', Symbol='Trl', Aliases='Dmel_CG33261, Adf-2, Adf-2-519, Adf2, CG33261, CG9343, Dmel\\CG33261, E(var)3-trl, E(var)62, GAF, GAGA, GAGA factor, Gaf, Gaga, NC70F, Nc70F, TRL, TfGAGA/Adf-2-GAGA, anon-EST:fe2E12, gaga, l(3)s2325, trl, Trl', description='Trithorax-like', other_designations='Adh transcription factor 2|CG33261-PA|CG33261-PB|CG33261-PC|CG33261-PD|CG33261-PE|CG33261-PF|CG33261-PH|CG33261-PI|CG33261-PJ|CG33261-PK|GAGA element|GAGA factor|GAGA transcription factor|GAGA-factor|Trithorax-related|Trl-PA|Trl-PB|Trl-PC|Trl-PD|Trl-PE|Trl-PF|Trl-PH|Trl-PI|Trl-PJ|Trl-PK|anon-fast-evolving-2E12|neural conserved at 70F|transcription-factor-Adh2|trithorax like|trithorax-like', map_location='70F4-70F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14747929, end_position_on_the_genomic_accession=14761049, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  2768982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768982, CurrentID=0, Status='live', Symbol='CG33270', Aliases='Dmel_ CG14139, CG32084, CG6100, Dmel\\CG33270', description='CG33270 gene product from transcript CG33270-RA', other_designations='CG33270-PA|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11578236, end_position_on_the_genomic_accession=11578623, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768983, CurrentID=0, Status='live', Symbol='CG33269', Aliases='Dmel_ CG14139, CG32084, CG6100, Dmel\\CG33269', description='CG33269 gene product from transcript CG33269-RB', other_designations='CG33269-PA|CG33269-PB|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11581415, end_position_on_the_genomic_accession=11582197, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  2768984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768984, CurrentID=0, Status='live', Symbol='CG33286', Aliases='Dmel_ CG12982, Dmel\\CG33286', description='CG33286 gene product from transcript CG33286-RB', other_designations='CG33286-PB|uncharacterized protein', map_location='78A1-78A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20970812, end_position_on_the_genomic_accession=20974700, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  49241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49241, CurrentID=0, Status='live', Symbol='rec', Aliases='Dmel_CG31293, CG14875, CG31293, CG4828, Dmel\\CG31293, MCM8-1, rec1, rec', description='recombination-defective', other_designations='CG31293-PA|CG31293-PB|mini-chromosome maintenance 8|rec-PA|rec-PB|recombination defective', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15823019, end_position_on_the_genomic_accession=15832632, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768986, CurrentID=0, Status='live', Symbol='CG33287', Aliases='Dmel_ CG12982, Dmel\\CG33287', description='CG33287 gene product from transcript CG33287-RB', other_designations='CG33287-PB|uncharacterized protein', map_location='78A1-78A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20966766, end_position_on_the_genomic_accession=20970663, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  2768987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768987, CurrentID=0, Status='live', Symbol='ppk5', Aliases='Dmel_CG33289, CG33289, Dmel\\CG33289', description='pickpocket 5', other_designations='CG33289-PB|CG33289-PC|ppk5-PB|ppk5-PC', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21534660, end_position_on_the_genomic_accession=21536861, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  2768988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768988, CurrentID=0, Status='live', Symbol='CG33285', Aliases='Dmel_ Dmel\\CG33285, NEST:bs33a04, NEST:bs33a07', description='CG33285 gene product from transcript CG33285-RA', other_designations='CG33285-PA|uncharacterized protein', map_location='78B1-78B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21088223, end_position_on_the_genomic_accession=21088989, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26067037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067037, CurrentID=0, Status='live', Symbol='CG33939', Aliases='Dmel_CG33939, BcDNA:RE54004, CR33939, Dmel\\CG33939, pncr004:X', description='putative noncoding RNA 004:X', other_designations='CG33939-PB|pncr004:X', map_location='18A4-18A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19171077, end_position_on_the_genomic_accession=19174900, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768990, CurrentID=0, Status='live', Symbol='Muc68E', Aliases='Dmel_CG33265, CG33265, Dmel\\CG33265', description='Mucin 68E', other_designations='CG33265-PA|Muc68E-PA|mucin 68E', map_location='68E4-68E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11967343, end_position_on_the_genomic_accession=11972979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26067039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067039, CurrentID=0, Status='live', Symbol='CG33947', Aliases='Dmel_CG33947, BEST:GH03576, BcDNA:GH03576, CR33947, Dmel\\CG33947, GH03576, pncr011:3L', description='putative noncoding RNA 011:3L', other_designations='CG33947-PD', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11096496, end_position_on_the_genomic_accession=11098556, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768992, CurrentID=0, Status='live', Symbol='Ilp5', Aliases='Dmel_CG33273, CG33273, DILP, DILP-5, DILP5, Dilp, Dilp 5, Dilp-5, Dilp5, Dmel\\CG33273, ILP5, Ilp, Ilp-5, dILP-5, dILP5, dIlp5, dilp, dilp-5, dilp5, ilp5', description='Insulin-like peptide 5', other_designations='CG33273-PA|Ilp5-PA|drosophila insulin-like peptide 5|insulin like peptide 5|insulin-like peptide|insulin-like peptide-5', map_location='67C9-67C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9823349, end_position_on_the_genomic_accession=9823897, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26067041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067041, CurrentID=0, Status='live', Symbol='Mst77Y-4', Aliases='Dmel_CG45763, CG45763, Dmel\\CG45763', description='Male-specific transcript 77Y 4', other_designations='CG45763-PA|Mst77Y-4-PA', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3284939, end_position_on_the_genomic_accession=3285868, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768994, CurrentID=0, Status='live', Symbol='CG33493', Aliases='Dmel_ CG6463, Dmel\\CG33493', description='CG33493 gene product from transcript CG33493-RB', other_designations='CG33493-PA|CG33493-PB|uncharacterized protein', map_location='67E7-67E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10692530, end_position_on_the_genomic_accession=10693167, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  2768995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768995, CurrentID=0, Status='live', Symbol='CG43980', Aliases='Dmel_ CG11250, CG33291, CG33292, Dmel\\CG43980, Dmel_CG33291, Dmel_CG33292', description='CG43980 gene product from transcript CG43980-RB', other_designations='CG43980-PB|CG43980-PC|uncharacterized protein', map_location='78E2-78E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21644958, end_position_on_the_genomic_accession=21666306, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  2768996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768996, CurrentID=0, Status='live', Symbol='CG33229', Aliases='Dmel_ CG16971, CG32492, Dmel\\CG33229', description='CG33229 gene product from transcript CG33229-RB', other_designations='CG33229-PA|CG33229-PB|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=223852, end_position_on_the_genomic_accession=225605, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  2768997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768997, CurrentID=0, Status='live', Symbol='CG12038', Aliases='Dmel_ Dmel\\CG12038', description='CG12038 gene product from transcript CG12038-RA', other_designations='CG12038-PA|uncharacterized protein', map_location='61E1-61E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=993983, end_position_on_the_genomic_accession=995703, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2768998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768998, CurrentID=0, Status='live', Symbol='CG33230', Aliases='Dmel_ Dmel\\CG33230', description='CG33230 gene product from transcript CG33230-RA', other_designations='CG33230-PA|uncharacterized protein', map_location='62A8-62A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1618414, end_position_on_the_genomic_accession=1620397, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768999, CurrentID=0, Status='live', Symbol='CG42676', Aliases='Dmel_ CG12022, CG13923, CG32317, Dmel\\CG42676, Dmel_CG12022, Dmel_CG13923', description='CG42676 gene product from transcript CG42676-RF', other_designations='CG42676-PC|CG42676-PD|CG42676-PE|CG42676-PF|uncharacterized protein', map_location='62B4-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1803107, end_position_on_the_genomic_accession=1806920, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  26067048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067048, CurrentID=0, Status='live', Symbol='CG45770', Aliases='Dmel_ Dmel\\CG45770', description='CG45770 gene product from transcript CG45770-RA', other_designations='CG45770-PA|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3349651, end_position_on_the_genomic_accession=3350374, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2769001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2769001, CurrentID=0, Status='live', Symbol='zormin', Aliases='Dmel_CG33484, BcDNA:GH09541, CG11850, CG11952, CG11953, CG1282, CG32307, CG32309, CG32310, CG32311, CG33484, CG5699, CT36990, Dmel\\CG33484', description='CG33484 gene product from transcript CG33484-RK', other_designations='CG33484-PE|CG33484-PF|CG33484-PG|CG33484-PI|CG33484-PJ|CG33484-PK|D-titin|titin|zormin-PE|zormin-PF|zormin-PG|zormin-PI|zormin-PJ|zormin-PK', map_location='62C4-62D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2117466, end_position_on_the_genomic_accession=2151700, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  26067050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067050, CurrentID=0, Status='live', Symbol='CG45782', Aliases='Dmel_ CG40467, CG41087, CG41343, Dmel\\CG45782, Dmel_CG40467, Dmel_CG41087, Dmel_CG41343, GH10292, Slc45-2', description='CG45782 gene product from transcript CG45782-RA', other_designations='CG45782-PA|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=25236940, end_position_on_the_genomic_accession=25670980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  26067051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067051, CurrentID=0, Status='live', Symbol='Pzl', Aliases='Dmel_CG45783, CG40188, CG40232, CG40249, CG40328, CG41056, CG41238, CG41369, CG42617, CG42618, CG45783, CR41056, Dmel\\CG45783, Dmel_CG40188, Dmel_CG40249, Dmel_CG41056, Dmel_CG41238, Dmel_CG41369, Dmel_CG42617, Dmel_CG42618, Dmel_CR41056', description='Piezo-like', other_designations='CG45783-PE|Pzl-PE', map_location='', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=2554162, end_position_on_the_genomic_accession=3263582, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  49260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49260, CurrentID=0, Status='live', Symbol='Pp1-87B', Aliases='Dmel_CG5650, 87B, CG5650, DmPp1-87B, Dmel\\CG5650, PP-1alpha, PP1, PP1 87B, PP1 beta87b, PP1-87B, PP187B, PP1A87B, PP1B, PP1alpha-87B, PP1alpha87B, PP1alpha[[2]], PP1c, PP1c 87B, PPI 87B, Pp-1alpha, Pp1b, Pp1alpha-87B, Ppl-87B, Su(Var)3-6, Su(var)(3)6, Su(var)3-6, Su-var(3)6, Suvar(3)6, ck19, dPP1beta, l(3)87Bg, l(3)ck19, l(3)j6E7, pp1-87B, su-var(3)6, Pp1-87B', description='Protein phosphatase 1 at 87B', other_designations='CG5650-PA|Pp1-87B-PA|protein phosphatase 1|protein phosphatase 1 at 87B|suppressor of variegation', map_location='87B9-87B10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12423621, end_position_on_the_genomic_accession=12425917, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26067053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067053, CurrentID=0, Status='live', Symbol='kl-3', Aliases='Dmel_CG45785, CG17629, CG40443, CG40936, CG40992, CG45785, Dmel\\CG45785, Dmel_CG17629, Dmel_CG40936, Dmel_CG40992, Lms3, Y104b, kl-4, kl3, ms(Y)104b', description='male fertility factor kl3', other_designations='CG45785-PC|gamma-dynein heavy chain|kl-3-PC|male fertility factor kl-3', map_location='h8-h8', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=336381, end_position_on_the_genomic_accession=563165, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  26067054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067054, CurrentID=0, Status='live', Symbol='kl-5', Aliases='Dmel_CG45786, CG12573, CG17616, CG40444, CG41497, CG45786, Dhc, Dhc-Yh3, DhcYh3, Dmel\\CG45786, Dmel_CG40444, Dmel_CG41497, Kl-5, Lms4, kl5', description='male fertility factor kl5', other_designations='CG45786-PC|Y-linked flagellar dynein|dynein heavy chain in Yh3|dynein-related heavy chain|kl-5-PC|male fertility factor kl-5', map_location='h2-h3', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=22250, end_position_on_the_genomic_accession=137486, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  2769007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2769007, CurrentID=0, Status='live', Symbol='Rdh', Aliases='Dmel_CG14975, CG14975, Dmel\\CG14975', description='CG14975 gene product from transcript CG14975-RA', other_designations='CG14975-PA|Rdh-PA|Red-herring', map_location='63F2-63F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3825449, end_position_on_the_genomic_accession=3828525, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2769008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2769008, CurrentID=0, Status='live', Symbol='Dscam4', Aliases='Dmel_CG42330, CG13672, CG18630, CG33274, CG42330, CG7060, CT21829, Dmel\\CG42330, Dmel_CG13672, Dmel_CG33274', description='Down syndrome cell adhesion molecule 4', other_designations='CG42330-PD|CG42330-PE|CG42330-PH|CG42330-PI|CG42330-PJ|Dscam4-PD|Dscam4-PE|Dscam4-PH|Dscam4-PI|Dscam4-PJ|down syndrome cell adhesion molecule 4', map_location='66C3-66C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8238647, end_position_on_the_genomic_accession=8282398, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  26067057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067057, CurrentID=0, Status='live', Symbol='CCY', Aliases='Dmel_CG45930, CG40948, CG40968, CG42620, CG45930, Ccy, Dmel\\CG45930, Dmel_CG40948, Dmel_CG40968, Dmel_CG42620, NEST:bs20e03, bs20e03.y1, ccy', description='Coiled-Coils Y', other_designations='CCY-PB|CG45930-PB', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3649996, end_position_on_the_genomic_accession=3666928, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26067058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067058, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-G', Aliases='Dmel_CG45986, CG45986, Dmel\\CG45986', description='CG45986 gene product from transcript CG45986-RA', other_designations='CG45986-PA|CG45986-PB|pre-mod(mdg4)-G-PA|pre-mod(mdg4)-G-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21354445, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  26067059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067059, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-H', Aliases='Dmel_CG45987, CG45987, Dmel\\CG45987', description='CG45987 gene product from transcript CG45987-RA', other_designations='CG45987-PA|CG45987-PB|pre-mod(mdg4)-H-PA|pre-mod(mdg4)-H-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21354445, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  26067060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067060, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-AA', Aliases='Dmel_CG45988, CG45988, Dmel\\CG45988', description='CG45988 gene product from transcript CG45988-RA', other_designations='CG45988-PA|CG45988-PB|pre-mod(mdg4)-AA-PA|pre-mod(mdg4)-AA-PB', map_location='93D9-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21370918, end_position_on_the_genomic_accession=21373575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26067061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067061, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-O', Aliases='Dmel_CG45989, CG45989, Dmel\\CG45989', description='CG45989 gene product from transcript CG45989-RA', other_designations='CG45989-PA|CG45989-PB|pre-mod(mdg4)-O-PA|pre-mod(mdg4)-O-PB', map_location='93D9-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21370918, end_position_on_the_genomic_accession=21373575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26067062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067062, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-N', Aliases='Dmel_CG45990, CG45990, Dmel\\CG45990', description='CG45990 gene product from transcript CG45990-RA', other_designations='CG45990-PA|CG45990-PB|pre-mod(mdg4)-N-PA|pre-mod(mdg4)-N-PB', map_location='93D9-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21370918, end_position_on_the_genomic_accession=21373575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26067063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067063, CurrentID=0, Status='live', Symbol='CG46025', Aliases='Dmel_ CG14538, CG7196, Dmel\\CG46025, Dmel_CG14538, Dmel_CG7196', description='CG46025 gene product from transcript CG46025-RC', other_designations='CG46025-PA|CG46025-PB|CG46025-PC|CG46025-PD|uncharacterized protein', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7754174, end_position_on_the_genomic_accession=7765915, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26067064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067064, CurrentID=0, Status='live', Symbol='CG46026', Aliases='Dmel_ CG31556, Dmel\\CG46026, mdcds_43655', description='CG46026 gene product from transcript CG46026-RA', other_designations='CG46026-PA|uncharacterized protein', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6306884, end_position_on_the_genomic_accession=6309030, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26067065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067065, CurrentID=0, Status='live', Symbol='CG46059', Aliases='Dmel_ CG17567, CR17567, Dmel\\CG46059', description='CG46059 gene product from transcript CG46059-RC', other_designations='CG46059-PA|CG46059-PB|CG46059-PC|uncharacterized protein', map_location='37C7-37C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19183093, end_position_on_the_genomic_accession=19183787, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26067066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067066, CurrentID=0, Status='live', Symbol='CG46121', Aliases='Dmel_ CG14177, CG3280, Dmel\\CG46121, Dmel_CG14177, Dmel_CG3280', description='CG46121 gene product from transcript CG46121-RC', other_designations='CG46121-PC|CG46121-PD|uncharacterized protein', map_location='67C2-67C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9597764, end_position_on_the_genomic_accession=9614001, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  26067067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067067, CurrentID=0, Status='live', Symbol='PRAS40', Aliases='Dmel_CG46146, CG10109, CG46146, Dmel\\CG46146, dPRAS40', description='PRAS40 ortholog', other_designations='CG46146-PA|PRAS40-PA', map_location='51A4-51A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14481187, end_position_on_the_genomic_accession=14497312, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26067068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067068, CurrentID=0, Status='live', Symbol='Fatp', Aliases='Dmel_CG46149, CG45083, CG46149, CG7400, Dmel\\CG46149, Dmel_CG45083, Dmel_CG7400, Dromel_CG7400_FBtr0080133_fatp_mORF, Dromel_CG7400_FBtr0080133_fatp_uORF, anon-WO0121795.74, dFATP, dFatp, fatp, l(2)k10307', description='Fatty acid (long chain) transport protein', other_designations='CG46149-PA|CG46149-PB|CG46149-PE|CG46149-PF|Fatp-PA|Fatp-PB|Fatp-PE|Fatp-PF|fatty acid (long chain) transport protein|fatty acid transport protein', map_location='31F4-31F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10510672, end_position_on_the_genomic_accession=10517218, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26067069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067069, CurrentID=0, Status='live', Symbol='CG46191', Aliases='Dmel_ Dmel\\CG46191', description='CG46191 gene product from transcript CG46191-RA', other_designations='CG46191-PA|Mst35Ba-related|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3215547, end_position_on_the_genomic_accession=3215855, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26067070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067070, CurrentID=0, Status='live', Symbol='CG46192', Aliases='Dmel_ Dmel\\CG46192', description='CG46192 gene product from transcript CG46192-RA', other_designations='CG46192-PA|Mst35Ba-related|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3252956, end_position_on_the_genomic_accession=3253357, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26067071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067071, CurrentID=0, Status='live', Symbol='CG46193', Aliases='Dmel_ Dmel\\CG46193', description='CG46193 gene product from transcript CG46193-RA', other_designations='CG46193-PA|Mst35Ba-related|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3263180, end_position_on_the_genomic_accession=3263488, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26067072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067072, CurrentID=0, Status='live', Symbol='CG46244', Aliases='Dmel_ Dm_2L:103252, Dmel\\CG46244', description='CG46244 gene product from transcript CG46244-RA', other_designations='CG46244-PA|uncharacterized protein', map_location='37F2-38A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19606569, end_position_on_the_genomic_accession=19731561, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  26067073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067073, CurrentID=0, Status='live', Symbol='CG46270', Aliases='Dmel_ Dmel\\CG46270', description='CG46270 gene product from transcript CG46270-RA', other_designations='CG46270-PA|uncharacterized protein', map_location='86B4-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10428137, end_position_on_the_genomic_accession=10428424, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26067074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067074, CurrentID=0, Status='live', Symbol='CG46275', Aliases='Dmel_ Dmel\\CG46275', description='CG46275 gene product from transcript CG46275-RA', other_designations='CG46275-PA|uncharacterized protein', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20197319, end_position_on_the_genomic_accession=20201933, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26067075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067075, CurrentID=0, Status='live', Symbol='CG46276', Aliases='Dmel_ Dmel\\CG46276', description='CG46276 gene product from transcript CG46276-RA', other_designations='CG46276-PA|uncharacterized protein', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20205043, end_position_on_the_genomic_accession=20209603, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26067076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067076, CurrentID=0, Status='live', Symbol='CG46277', Aliases='Dmel_ Dmel\\CG46277', description='CG46277 gene product from transcript CG46277-RA', other_designations='CG46277-PA|uncharacterized protein', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20212738, end_position_on_the_genomic_accession=20217259, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26067077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067077, CurrentID=0, Status='live', Symbol='CG46280', Aliases='Dmel_ CG12360, Dmel\\CG46280', description='CG46280 gene product from transcript CG46280-RB', other_designations='CG46280-PA|CG46280-PB|uncharacterized protein', map_location='87D9-87D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13026283, end_position_on_the_genomic_accession=13030659, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  26067078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067078, CurrentID=0, Status='live', Symbol='CG46281', Aliases='Dmel_ CG12360, Dmel\\CG46281', description='CG46281 gene product from transcript CG46281-RA', other_designations='CG46281-PA|uncharacterized protein', map_location='87D9-87D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13028920, end_position_on_the_genomic_accession=13030307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  35521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35521, CurrentID=0, Status='live', Symbol='CG7881', Aliases='Dmel_ Dmel\\CG7881', description='CG7881 gene product from transcript CG7881-RB', other_designations='CG7881-PA|CG7881-PB', map_location='42A4-42A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5960275, end_position_on_the_genomic_accession=5966856, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26067080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067080, CurrentID=0, Status='live', Symbol='vas', Aliases='Dmel_CG46283, BG:DS00929.14, CG3506, CG43081, CG46283, DmRH25, Dmel\\CG46283, Dmel_CG33678, Dmel_CG3506, Dmel_CG43081, EP(2)0812, VAS, VASA, Vas, Vasa, cgt, fs(2)ltoRJ36', description='vasa', other_designations='CG46283-PA|CG46283-PB|CG46283-PC|VASA|Vas|courgette|female sterile(2)ltoRJ36|no-relish|vas-PA|vas-PB|vas-PC', map_location='35B10-35C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15061656, end_position_on_the_genomic_accession=15074311, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  49297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49297, CurrentID=0, Status='live', Symbol='SERCA', Aliases='Dmel_CG3725, 17136664, A4UZU0_DROME, ATC1_DROME, CA-P60A, CG3725, Ca-ATPase, Ca-P60, Ca-P60A, Ca-p, CaP60A, Ca[2+]ATPase, Dmel\\CG3725, Kum2, Serca, Serca2a, dSERCA, dserca, kum, l(2)SH0211, l(2)k08308ab, l(2)k09025, SERCA', description='Sarco/endoplasmic reticulum Ca(2+)-ATPase', other_designations='CG3725-PA|CG3725-PB|CG3725-PC|CG3725-PD|CG3725-PE|CG3725-PF|CG3725-PG|CG3725-PH|CG3725-PI|Ca[2+] ATPase|SERCA-PA|SERCA-PB|SERCA-PC|SERCA-PD|SERCA-PE|SERCA-PF|SERCA-PG|SERCA-PH|SERCA-PI|calcium ATPase|calcium ATPase at 60A|kumbhakarna|lethal (2) SH0211|organellar-type Ca-ATPase|sarco endoplasmic reticulum calcium-ATPase|sarco/endoplasmic reticulum-type Ca-2+-ATPase|sarcoplasmic endoplasmic reticulum calcium ATPase type 2', map_location='2-106 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23925077, end_position_on_the_genomic_accession=23935884, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  26067038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067038, CurrentID=0, Status='live', Symbol='CG33946', Aliases='Dmel_CG33946, BcDNA:RH62830, CR33946, Dmel\\CG33946, pncr008:3L', description='putative noncoding RNA 008:3L', other_designations='CG33946-PB', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6220578, end_position_on_the_genomic_accession=6220949, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  49424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49424, CurrentID=0, Status='live', Symbol='CIAPIN1', Aliases='Dmel_CG4180, B, BG:DS00929.9, CG4180, D-CIAPIN1, D1, Dmel\\CG4180, anon-35Ba, br26, l(2)35Bg, l(2)br26', description='CIAPIN1 ortholog', other_designations='CG4180-PA|CIAPIN1-PA|lethal (2) 35Bg|lethal 26 in the black-reduced region|lethal group D1|transcription unit B', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15036293, end_position_on_the_genomic_accession=15038861, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49425, CurrentID=0, Status='live', Symbol='mRpL4', Aliases='Dmel_CG5818, BG:DS02740.17, CG5818, Dmel\\CG5818, L4, Rpl4, l(2)35Fe, l(2)k14608, l35Fe, mrpL4', description='mitochondrial ribosomal protein L4', other_designations='CG5818-PA|lethal(2)35Fe|mRpL4-PA|mitochondrial ribosomal protein large subunit 4', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16328059, end_position_on_the_genomic_accession=16329132, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  49427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49427, CurrentID=0, Status='live', Symbol='l(2)37Bb', Aliases='Dmel_CG10655, Bb, CG10655, Dmel\\CG10655, l(2)35B2, l(2)E58', description='lethal (2) 37Bb', other_designations='CG10655-PA|l(2)37Bb-PA', map_location='2-53.9 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19049843, end_position_on_the_genomic_accession=19051612, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  49505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49505, CurrentID=0, Status='live', Symbol='fs(2)ltoPP43', Aliases='Dmel_CG10528, BEST:LD12957, CG10528, Dmel\\CG10528, LD12957, l(2)38Ac', description='female sterile (2) ltoPP43', other_designations='CG10528-PA|CG10528-PB|CG10528-PC|fs(2)ltoPP43-PA|fs(2)ltoPP43-PB|fs(2)ltoPP43-PC', map_location='2-51 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19792773, end_position_on_the_genomic_accession=19795851, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  49529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49529, CurrentID=0, Status='live', Symbol='ms(2)35Ci', Aliases='Dmel_CG31733, BcDNA:AT28918, CG31733, Dmel\\CG31733, NEST:bs29a09, ms(2)02316, ms(2)46AB', description='male sterile (2) 35Ci', other_designations='CG31733-PA|CG31733-PC|male sterile(2)46AB|ms(2)35Ci-PA|ms(2)35Ci-PC', map_location='35D1-35D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15292468, end_position_on_the_genomic_accession=15294136, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41023, CurrentID=0, Status='live', Symbol='CG11753', Aliases='Dmel_ Dmel\\CG11753', description='CG11753 gene product from transcript CG11753-RC', other_designations='CG11753-PA|CG11753-PB|CG11753-PC|uncharacterized protein', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8652821, end_position_on_the_genomic_accession=8653847, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  49580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49580, CurrentID=0, Status='live', Symbol='Invadolysin', Aliases='Dmel_CG3953, CG3953, DmIX-14, Dmel\\CG3953, IX-14, inv, invadolysin, l(3)IX-14, l(3)SG34', description='CG3953 gene product from transcript CG3953-RA', other_designations='CG3953-PA|Invadolysin-PA|invadolysin|lethal (3) IX-14', map_location='3-49.0 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10144668, end_position_on_the_genomic_accession=10159134, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  318079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318079, CurrentID=0, Status='live', Symbol='CG32537', Aliases='Dmel_ CG8020, Dmel\\CG32537', description='CG32537 gene product from transcript CG32537-RA', other_designations='CG32537-PA|CG32537-PB|uncharacterized protein', map_location='18B7-18B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19273349, end_position_on_the_genomic_accession=19276539, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835677, CurrentID=0, Status='live', Symbol='CG44259', Aliases='Dmel_ Dmel\\CG44259', description='CG44259 gene product from transcript CG44259-RA', other_designations='CG44259-PA|uncharacterized protein', map_location='89F4-89F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17253581, end_position_on_the_genomic_accession=17254129, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  49636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49636, CurrentID=0, Status='live', Symbol='hrg', Aliases='Dmel_CG9854, CG9854, Dmel\\CG9854, PAP, Pap, l(2)07619, l(2)k07619', description='hiiragi', other_designations='CG9854-PA|CG9854-PB|CG9854-PC|CG9854-PD|Poly(A) polymerase|hrg-PA|hrg-PB|hrg-PC|hrg-PD|poly (A) polymerase', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19663919, end_position_on_the_genomic_accession=19669832, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  49638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49638, CurrentID=0, Status='live', Symbol='drm', Aliases='Dmel_CG10016, CG10016, Dmel\\CG10016, drum', description='drumstick', other_designations='CG10016-PA|CG10016-PB|CG10016-PC|drm-PA|drm-PB|drm-PC', map_location='24A1-24A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3539247, end_position_on_the_genomic_accession=3548423, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  8673955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673955, CurrentID=0, Status='live', Symbol='CG42592', Aliases='Dmel_ Dmel\\CG42592', description='CG42592 gene product from transcript CG42592-RA', other_designations='CG42592-PA|uncharacterized protein', map_location='99C2-99C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29835610, end_position_on_the_genomic_accession=29836288, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462401, CurrentID=0, Status='live', Symbol='CG43673', Aliases='Dmel_ CG33173, CG8473, Dmel\\CG43673, Dmel_CG33173', description='CG43673 gene product from transcript CG43673-RA', other_designations='CG43673-PA|uncharacterized protein', map_location='13E17-13E17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15779146, end_position_on_the_genomic_accession=15780214, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  49701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49701, CurrentID=0, Status='live', Symbol='gish', Aliases='Dmel_CG6963, CG6963, CK1, CK1-gamma, CK1[[gamma]], CK1gamma, CKI-related, CKIgamma, Dmel\\CG6963, Gish, Hrr25, NEST:bs27c08, Spider, anon-WO0118547.425, ms(3)89B, spider', description='gilgamesh', other_designations='CG6963-PA|CG6963-PB|CG6963-PC|CG6963-PD|CG6963-PE|CG6963-PF|CG6963-PG|CG6963-PH|CG6963-PI|CG6963-PJ|CG6963-PK|CG6963-PL|CG6963-PM|CKIg-like|NEST:bs27c08|casein kinase 1gamma|gish-PA|gish-PB|gish-PC|gish-PD|gish-PE|gish-PF|gish-PG|gish-PH|gish-PI|gish-PJ|gish-PK|gish-PL|gish-PM|spider', map_location='89B9-89B12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16272452, end_position_on_the_genomic_accession=16304768, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  49713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49713, CurrentID=0, Status='live', Symbol='alpha-catenin-related', Aliases='Dmel_CG2987, CG2987, Dmel\\CG2987, alpha-catenin-Related', description='alpha-catenin related', other_designations='CG2987-PA|CG2987-PC|alpha-catenin-related-PA|alpha-catenin-related-PC', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23957842, end_position_on_the_genomic_accession=23961631, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318939, CurrentID=0, Status='live', Symbol='CG31784', Aliases='Dmel_ Dmel\\CG31784, NEST:bs05c10', description='CG31784 gene product from transcript CG31784-RC', other_designations='CG31784-PC|uncharacterized protein', map_location='36C1-36C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17123307, end_position_on_the_genomic_accession=17125961, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  49762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49762, CurrentID=0, Status='live', Symbol='l(3)87Df', Aliases='Dmel_CG7620, BcDNA:SD27354, CG7620, Dmel\\CG7620, S12, l(3)S12, l(3)s12', description='lethal (3) 87Df', other_designations='CG7620-PA|l(3)87Df-PA', map_location='87D9-87D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13030913, end_position_on_the_genomic_accession=13031598, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246383, CurrentID=0, Status='live', Symbol='CG30001', Aliases='Dmel_ Dmel\\CG30001', description='CG30001 gene product from transcript CG30001-RA', other_designations='CG30001-PA|uncharacterized protein', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9821168, end_position_on_the_genomic_accession=9822955, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  246384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246384, CurrentID=0, Status='live', Symbol='CG30002', Aliases='Dmel_ Dmel\\CG30002', description='CG30002 gene product from transcript CG30002-RB', other_designations='CG30002-PB|uncharacterized protein', map_location='46A1-46A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9704384, end_position_on_the_genomic_accession=9706147, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246386, CurrentID=0, Status='live', Symbol='GstT2', Aliases='Dmel_CG30005, CG12930, CG12930A, CG30005, Dmel\\CG30005, GSTT2, gstt2', description='Glutathione S transferase T2', other_designations='CG30005-PA|GstT2-PA|glutathione S transferase T2', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9607292, end_position_on_the_genomic_accession=9608350, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  49779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49779, CurrentID=0, Status='live', Symbol='PpD3', Aliases='Dmel_CG8402, CG8402, D3, DmPP5, DmPpp5-85E, Dmel\\CG8402, PP5, PP5 85E, Pp5-85E', description='Protein phosphatase D3', other_designations='CG8402-PA|CG8402-PB|CG8402-PC|PpD3-PA|PpD3-PB|PpD3-PC|protein phosphatase D3|protein phosphatase from PCR fragment D3', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9747875, end_position_on_the_genomic_accession=9751094, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  49780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49780, CurrentID=0, Status='live', Symbol='PpD6', Aliases='Dmel_CG8822, CG8822, D6, Dmel\\CG8822, PP1', description='Protein phosphatase D6', other_designations='CG8822-PA|PP1-like|PpD6-PA|protein phosphatase D6|protein phosphatase from PCR fragment D6', map_location='23D1-23D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3109672, end_position_on_the_genomic_accession=3110998, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246389, CurrentID=0, Status='live', Symbol='CG30010', Aliases='Dmel_ BcDNA:SD18306, Dmel\\CG30010', description='CG30010 gene product from transcript CG30010-RB', other_designations='CG30010-PA|CG30010-PB|uncharacterized protein', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9863651, end_position_on_the_genomic_accession=9864576, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246393, CurrentID=0, Status='live', Symbol='CG30016', Aliases='Dmel_ Dmel\\CG30016, anon-AE003828.1', description='CG30016 gene product from transcript CG30016-RA', other_designations='CG30016-PA|uncharacterized protein', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10874892, end_position_on_the_genomic_accession=10875392, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246395, CurrentID=0, Status='live', Symbol='mthl13', Aliases='Dmel_CG30018, CG30018, Dmel\\CG30018', description='methuselah-like 13', other_designations='CG30018-PD|mthl13-PD', map_location='47B13-47B13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10791459, end_position_on_the_genomic_accession=10793109, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  246397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246397, CurrentID=0, Status='live', Symbol='sprt', Aliases='Dmel_CG30023, CG30023, CG9026, Dmel\\CG30023', description='sprite', other_designations='CG30023-PA|sprt-PA', map_location='47F8-47F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11384743, end_position_on_the_genomic_accession=11397432, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246398, CurrentID=0, Status='live', Symbol='CG30025', Aliases='Dmel_ Dmel\\CG30025', description='CG30025 gene product from transcript CG30025-RA', other_designations='CG30025-PA|uncharacterized protein', map_location='47F4-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11353564, end_position_on_the_genomic_accession=11354373, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246399, CurrentID=0, Status='live', Symbol='CG30026', Aliases='Dmel_ Dmel\\CG30026', description='CG30026 gene product from transcript CG30026-RB', other_designations='CG30026-PB|uncharacterized protein', map_location='47F10-47F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11408576, end_position_on_the_genomic_accession=11409449, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246404, CurrentID=0, Status='live', Symbol='CG30031', Aliases='Dmel_ Dmel\\CG30031, endo-I', description='CG30031 gene product from transcript CG30031-RA', other_designations='CG30031-PA|endoproteinase I|uncharacterized protein', map_location='47F4-47F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11350811, end_position_on_the_genomic_accession=11351621, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246406, CurrentID=0, Status='live', Symbol='CG30033', Aliases='Dmel_ BcDNA:RH07028, BcDNA:RH63918, Dmel\\CG30033', description='CG30033 gene product from transcript CG30033-RB', other_designations='CG30033-PB|CG30033-PC|uncharacterized protein', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11292531, end_position_on_the_genomic_accession=11293591, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246408, CurrentID=0, Status='live', Symbol='CG30036', Aliases='Dmel_ CG8976, Dmel\\CG30036', description='CG30036 gene product from transcript CG30036-RA', other_designations='CG30036-PA|uncharacterized protein', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11977536, end_position_on_the_genomic_accession=11979018, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246409, CurrentID=0, Status='live', Symbol='CG30037', Aliases='Dmel_ CG8976, Dmel\\CG30037', description='CG30037 gene product from transcript CG30037-RA', other_designations='CG30037-PA|uncharacterized protein', map_location='48D5-48D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11979094, end_position_on_the_genomic_accession=11980538, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  49802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49802, CurrentID=0, Status='live', Symbol='IM2', Aliases='Dmel_CG18106, BcDNA:RH08291, CG18106, DIM 2, DIM-2, DIM2, Dmel\\CG18106, Im2, anon-WO0140519.2', description='Immune induced molecule 2', other_designations='CG18106-PA|IM2-PA|Immune-induced molecule 2|Immune-induced peptide 2 precursor|immune induced molecule 2|immune induced peptide 2|immune molecule 2', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18386602, end_position_on_the_genomic_accession=18387030, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  49803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49803, CurrentID=0, Status='live', Symbol='Spn55B', Aliases='Dmel_CG10913, CG10913, Dmel\\CG10913, Sp6, Spn6, sp-6, sp6', description='Serpin 55B', other_designations='CG10913-PA|Spn55B-PA|serine protease inhibitor 6|serpin 55B|serpin 6', map_location='55B7-55B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18126347, end_position_on_the_genomic_accession=18128210, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  49804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49804, CurrentID=0, Status='live', Symbol='Spn88Ea', Aliases='Dmel_CG18525, CG18525, Dmel\\CG18525, Sp5, Spn5, sp5, spn5', description='Serpin 88Ea', other_designations='CG18525-PA|CG18525-PB|Dm-serpin-88Ea|Spn88Ea-PA|Spn88Ea-PB|serine protease inhibitor 5|serpin|serpin 5|serpin 88Ea|serpin5', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15203069, end_position_on_the_genomic_accession=15206447, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  49805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49805, CurrentID=0, Status='live', Symbol='Spn42Da', Aliases='Dmel_CG9453, CG9453, Dmel\\CG9453, Serp2, Sp4, Spn4, Spn4A, dSerp2, sp4, spn4', description='Serpin 42Da', other_designations='CG9453-PA|CG9453-PB|CG9453-PD|CG9453-PE|CG9453-PF|CG9453-PG|CG9453-PH|CG9453-PI|CG9453-PJ|CG9453-PK|CG9453-PL|Serp2|Spn42Da-PA|Spn42Da-PB|Spn42Da-PD|Spn42Da-PE|Spn42Da-PF|Spn42Da-PG|Spn42Da-PH|Spn42Da-PI|Spn42Da-PJ|Spn42Da-PK|Spn42Da-PL|neuroserpin|serine protease inhibitor 4|serpin 4|serpin 42Da', map_location='42D-42E', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6875906, end_position_on_the_genomic_accession=6878927, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  49806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49806, CurrentID=0, Status='live', Symbol='Spn38F', Aliases='Dmel_CG9334, CG9334, Dmel\\CG9334, Sp3, Spn3, sp3', description='Serpin 38F', other_designations='CG9334-PA|CG9334-PB|Spn38F-PA|Spn38F-PB|serine protease inhibitor 3|serpin 3|serpin 38F', map_location='38F1-38F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20836621, end_position_on_the_genomic_accession=20837985, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49807, CurrentID=0, Status='live', Symbol='Spn28F', Aliases='Dmel_CG8137, CG8137, Dmel\\CG8137, Sp2, Spn2, sp2, spn2', description='Serpin 28F', other_designations='CG8137-PA|Spn28F-PA|serine protease inhibitor 2|serpin 2|serpin 28F', map_location='28F5-28F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8240414, end_position_on_the_genomic_accession=8241800, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49808, CurrentID=0, Status='live', Symbol='Spn42Dd', Aliases='Dmel_CG9456, CG9456, Dmel\\CG9456, Spn1, sp1', description='Serpin 42Dd', other_designations='CG9456-PB|Spn42Dd-PB|serine protease inhibitor 1|serpin 1|serpin 42Dd', map_location='42D6-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6883280, end_position_on_the_genomic_accession=6884873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  49809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49809, CurrentID=0, Status='live', Symbol='Rbp2', Aliases='Dmel_CG4429, CG4429, Dmel\\CG4429, MRE9, RRM2, anon-EST:fe1B9, cg4429, eIF4H, rrm2', description='RNA-binding protein 2', other_designations='CG4429-PA|CG4429-PB|CG4429-PC|CG4429-PD|CG4429-PE|Rbp2-PA|Rbp2-PB|Rbp2-PC|Rbp2-PD|Rbp2-PE|anon-fast-evolving-1B9|mRNA-like ncRNA in embryogenesis 9', map_location='14C6-14C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16448665, end_position_on_the_genomic_accession=16454972, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246418, CurrentID=0, Status='live', Symbol='CG30051', Aliases='Dmel_ Dmel\\CG30051, anon-AE003822.1', description='CG30051 gene product from transcript CG30051-RC', other_designations='CG30051-PC|uncharacterized protein', map_location='49A6-49A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12436120, end_position_on_the_genomic_accession=12436990, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246419, CurrentID=0, Status='live', Symbol='CG30053', Aliases='Dmel_ CG8769, Dmel\\CG30053', description='CG30053 gene product from transcript CG30053-RA', other_designations='CG30053-PA|uncharacterized protein', map_location='49B12-49B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12688733, end_position_on_the_genomic_accession=12689876, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246420, CurrentID=0, Status='live', Symbol='CG30054', Aliases='Dmel_ Dmel\\CG30054', description='CG30054 gene product from transcript CG30054-RE', other_designations='CG30054-PE|CG30054-PF|uncharacterized protein', map_location='49B9-49B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12623726, end_position_on_the_genomic_accession=12625352, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  246422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246422, CurrentID=0, Status='live', Symbol='CG30056', Aliases='Dmel_ Dmel\\CG30056, NEST:bs07b11, NEST:bs07h07, NEST:bs33g02', description='CG30056 gene product from transcript CG30056-RA', other_designations='CG30056-PA|uncharacterized protein', map_location='49B3-49B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12562060, end_position_on_the_genomic_accession=12562655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  49815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49815, CurrentID=0, Status='live', Symbol='Cht4', Aliases='Dmel_CG3986, BcDNA:RE62779, CG3986, CHT4, DmCHT4, Dmel\\CG3986', description='Chitinase 4', other_designations='CG3986-PA|Cht4-PA|chitinase 4', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21065379, end_position_on_the_genomic_accession=21067087, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  49816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49816, CurrentID=0, Status='live', Symbol='CycK', Aliases='Dmel_CG15218, CG15218, Dmel\\CG15218', description='Cyclin K', other_designations='CG15218-PA|CG15218-PB|CG15218-PC|CycK-PA|CycK-PB|CycK-PC|CyclinK|cyclin K', map_location='40E4-40E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22124285, end_position_on_the_genomic_accession=22130142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246425, CurrentID=0, Status='live', Symbol='CG30059', Aliases='Dmel_ Dmel\\CG30059, anon-WO0140519.22', description='CG30059 gene product from transcript CG30059-RA', other_designations='CG30059-PA|uncharacterized protein', map_location='50A5-50A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13412769, end_position_on_the_genomic_accession=13414620, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246426, CurrentID=0, Status='live', Symbol='CG30060', Aliases='Dmel_ Dmel\\CG30060, anon-SAGE:Wang-129', description='CG30060 gene product from transcript CG30060-RA', other_designations='CG30060-PA|anon-SAGE:Wang-129|uncharacterized protein', map_location='50A7-50A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13422978, end_position_on_the_genomic_accession=13423647, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246428, CurrentID=0, Status='live', Symbol='CG30062', Aliases='Dmel_ Dmel\\CG30062', description='CG30062 gene product from transcript CG30062-RB', other_designations='CG30062-PB|uncharacterized protein', map_location='50A9-50A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13445700, end_position_on_the_genomic_accession=13446309, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  49821 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49821, CurrentID=0, Status='live', Symbol='EbpIII', Aliases='Dmel_CG11390, BcDNA:RH74005, CG11390, DmelCSP2, Dmel\\CG11390, PEB, PEBmeIII, PEBmelIII, PebIII, Phk-2, phk-2', description='Ejaculatory bulb protein III', other_designations='CG11390-PA|CG11390-PB|CG11390-PC|EbpIII-PA|EbpIII-PB|EbpIII-PC|ejaculatory bulb protein III|pherokine 2', map_location='60B4-60B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24026172, end_position_on_the_genomic_accession=24027518, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246431, CurrentID=0, Status='live', Symbol='CG30065', Aliases='Dmel_ BcDNA:AT19876, Dmel\\CG30065', description='CG30065 gene product from transcript CG30065-RA', other_designations='CG30065-PA|uncharacterized protein', map_location='50A8-50A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13433028, end_position_on_the_genomic_accession=13433760, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246432, CurrentID=0, Status='live', Symbol='Obp50a', Aliases='Dmel_CG30067, CG30067, DmelObp50a, Dmel\\CG30067, anon-AE003814.2', description='Odorant-binding protein 50a', other_designations='CG30067-PA|CG30067-PB|Obp50a-PA|Obp50a-PB', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14370331, end_position_on_the_genomic_accession=14371116, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246434, CurrentID=0, Status='live', Symbol='Obp50c', Aliases='Dmel_CG30072, CG30072, DmelObp50c, Dmel\\CG30072', description='Odorant-binding protein 50c', other_designations='CG30072-PC|Obp50c-PC', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14371186, end_position_on_the_genomic_accession=14373006, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246435, CurrentID=0, Status='live', Symbol='Obp50b', Aliases='Dmel_CG30073, CG13940, CG30073, DmelObp50b, Dmel\\CG30073', description='Odorant-binding protein 50b', other_designations='CG30073-PB|Obp50b-PB', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14371186, end_position_on_the_genomic_accession=14373006, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246436, CurrentID=0, Status='live', Symbol='Obp50d', Aliases='Dmel_CG30074, CG30074, DmelObp50d, Dmel\\CG30074', description='Odorant-binding protein 50d', other_designations='CG30074-PA|Obp50d-PA', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14373108, end_position_on_the_genomic_accession=14373808, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246437, CurrentID=0, Status='live', Symbol='CG30075', Aliases='Dmel_ Dmel\\CG30075', description='CG30075 gene product from transcript CG30075-RA', other_designations='CG30075-PA|uncharacterized protein', map_location='50F7-50F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14375921, end_position_on_the_genomic_accession=14376677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246438, CurrentID=0, Status='live', Symbol='CG30076', Aliases='Dmel_ Dmel\\CG30076', description='CG30076 gene product from transcript CG30076-RA', other_designations='CG30076-PA|uncharacterized protein', map_location='51A8-51A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14544150, end_position_on_the_genomic_accession=14546372, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246439, CurrentID=0, Status='live', Symbol='Blos1', Aliases='Dmel_CG30077, BLOS1, CG10127, CG30077, Dmel\\CG30077, NP_725401, blos1', description='Biogenesis of lysosome-related organelles complex 1, subunit 1', other_designations='BLOC-1 subunit 1|Blos1-PB|CG30077-PB|biogenesis of lysosome-related organelles complex 1, subunit 1', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14601278, end_position_on_the_genomic_accession=14602010, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246441, CurrentID=0, Status='live', Symbol='CG30080', Aliases='Dmel_ BcDNA:RE42467, BcDNA:RE70859, CG12965, Dmel\\CG30080', description='CG30080 gene product from transcript CG30080-RC', other_designations='CG30080-PA|CG30080-PC|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15698726, end_position_on_the_genomic_accession=15701696, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246442, CurrentID=0, Status='live', Symbol='Ir51b', Aliases='Dmel_CG30081, CG30081, DmelIR51b, Dmel\\CG30081, IR51b', description='Ionotropic receptor 51b', other_designations='CG30081-PA|Ir51b-PA|ionotropic receptor 51b', map_location='51E6-51E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15207993, end_position_on_the_genomic_accession=15209786, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246443, CurrentID=0, Status='live', Symbol='CG30082', Aliases='Dmel_ CR30082, Dmel\\CG30082, Dmel_CR30082', description='CG30082 gene product from transcript CG30082-RB', other_designations='CG30082-PB|uncharacterized protein', map_location='52B5-52B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15685004, end_position_on_the_genomic_accession=15686082, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246444, CurrentID=0, Status='live', Symbol='CG30083', Aliases='Dmel_ Dmel\\CG30083', description='CG30083 gene product from transcript CG30083-RB', other_designations='CG30083-PB|uncharacterized protein', map_location='52C2-52C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15711597, end_position_on_the_genomic_accession=15712644, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246445, CurrentID=0, Status='live', Symbol='Rif1', Aliases='Dmel_CG30085, BcDNA:LD22218, CG30085, Dmel\\CG30085, dRif1', description='Rap1 interacting factor 1', other_designations='CG30085-PA|Rap1 interacting factor 1 homolog|Rif1-PA', map_location='52C8-52C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15864309, end_position_on_the_genomic_accession=15869147, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246446, CurrentID=0, Status='live', Symbol='CG30087', Aliases='Dmel_ CG8215, CR30087, Dmel\\CG30087, SP41', description='CG30087 gene product from transcript CG30087-RA', other_designations='CG30087-PA|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15692448, end_position_on_the_genomic_accession=15693751, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246447, CurrentID=0, Status='live', Symbol='CG30088', Aliases='Dmel_ CG8215, Dmel\\CG30088, SP41', description='CG30088 gene product from transcript CG30088-RB', other_designations='CG30088-PB|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15691113, end_position_on_the_genomic_accession=15692216, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246448, CurrentID=0, Status='live', Symbol='CG30090', Aliases='Dmel_ BcDNA:GH21666, CG8215, Dmel\\CG30090, SP41', description='CG30090 gene product from transcript CG30090-RA', other_designations='CG30090-PA|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15689826, end_position_on_the_genomic_accession=15691014, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246449, CurrentID=0, Status='live', Symbol='CG30091', Aliases='Dmel_ CG8212, Dmel\\CG30091', description='CG30091 gene product from transcript CG30091-RA', other_designations='CG30091-PA|uncharacterized protein', map_location='52B5-52B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15680545, end_position_on_the_genomic_accession=15682477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246451, CurrentID=0, Status='live', Symbol='COX6AL', Aliases='Dmel_CG30093, CG30093, COXD, COXE, Dmel\\CG30093, NEST:bs02a12', description='Cytochrome c oxidase subunit 6A-like', other_designations='CG30093-PA|COX6AL-PA|cytochrome c oxidase polypeptide VIA|cytochrome c oxidase subunit 6A-like', map_location='52D3-52D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15944028, end_position_on_the_genomic_accession=15944567, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246452, CurrentID=0, Status='live', Symbol='CG30094', Aliases='Dmel_ BcDNA:SD13256, CG15700, Dmel\\CG30094', description='CG30094 gene product from transcript CG30094-RA', other_designations='CG30094-PA|uncharacterized protein', map_location='52E10-52E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16142202, end_position_on_the_genomic_accession=16142935, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246453, CurrentID=0, Status='live', Symbol='CG30095', Aliases='Dmel_ CG8425, Dmel\\CG30095', description='CG30095 gene product from transcript CG30095-RA', other_designations='CG30095-PA|uncharacterized protein', map_location='52E5-52E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16123484, end_position_on_the_genomic_accession=16124577, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246454, CurrentID=0, Status='live', Symbol='CG30096', Aliases='Dmel_ Dmel\\CG30096', description='CG30096 gene product from transcript CG30096-RB', other_designations='CG30096-PB|uncharacterized protein', map_location='53B4-53B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16320815, end_position_on_the_genomic_accession=16322124, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49847, CurrentID=0, Status='live', Symbol='Cyp6a22', Aliases='Dmel_CG10240, 6a22, BF6-2, CG10240, Cyt-P450-rBF6-2, Dmel\\CG10240, P-450, P450, cyp6a22', description='CG10240 gene product from transcript CG10240-RB', other_designations='CG10240-PA|CG10240-PB|CG10240-PC|Cyp6a22-PA|Cyp6a22-PB|Cyp6a22-PC|cytochrome P450|cytochrome P450 related BF6-2', map_location='51C7-51D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14871366, end_position_on_the_genomic_accession=14873705, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246456, CurrentID=0, Status='live', Symbol='CG30098', Aliases='Dmel_ Dmel\\CG30098', description='CG30098 gene product from transcript CG30098-RA', other_designations='CG30098-PA|uncharacterized protein', map_location='53C2-53C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16354224, end_position_on_the_genomic_accession=16355215, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246457, CurrentID=0, Status='live', Symbol='CG30100', Aliases='Dmel_ Dmel\\CG30100, Dromel_CG30100_FBtr0087140_mORF', description='CG30100 gene product from transcript CG30100-RA', other_designations='CG30100-PA|CG30100-PB|uncharacterized protein', map_location='53B1-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16291453, end_position_on_the_genomic_accession=16292269, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246458, CurrentID=0, Status='live', Symbol='NT5E-2', Aliases='Dmel_CG30104, CG30104, CG4837, Dmel\\CG30104', description="Ecto-5'-nucleotidase 2", other_designations='CG30104-PA|CG30104-PB|NT5E-2-PA|NT5E-2-PB', map_location='54B17-54B17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17485002, end_position_on_the_genomic_accession=17493169, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246459, CurrentID=0, Status='live', Symbol='CG30105', Aliases='Dmel_ BcDNA:RE11282, Dmel\\CG30105', description='CG30105 gene product from transcript CG30105-RA', other_designations='CG30105-PA|uncharacterized protein', map_location='54C12-54C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17591804, end_position_on_the_genomic_accession=17592509, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246460, CurrentID=0, Status='live', Symbol='CG30108', Aliases='Dmel_ Dmel\\CG30108', description='CG30108 gene product from transcript CG30108-RB', other_designations='CG30108-PB|uncharacterized protein', map_location='54D4-54D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17669395, end_position_on_the_genomic_accession=17670000, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246461, CurrentID=0, Status='live', Symbol='CG30109', Aliases='Dmel_ Dmel\\CG30109', description='CG30109 gene product from transcript CG30109-RB', other_designations='CG30109-PB|uncharacterized protein', map_location='54D4-54D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17670082, end_position_on_the_genomic_accession=17670748, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246462, CurrentID=0, Status='live', Symbol='CG30110', Aliases='Dmel_ Dmel\\CG30110, NEST:bs09e07', description='CG30110 gene product from transcript CG30110-RA', other_designations='CG30110-PA|uncharacterized protein', map_location='54F5-54F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17880194, end_position_on_the_genomic_accession=17882245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246465, CurrentID=0, Status='live', Symbol='CG30114', Aliases='Dmel_ Dmel\\CG30114', description='CG30114 gene product from transcript CG30114-RA', other_designations='CG30114-PA|uncharacterized protein', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18207755, end_position_on_the_genomic_accession=18208978, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246467, CurrentID=0, Status='live', Symbol='CG30120', Aliases='Dmel_ Dmel\\CG30120', description='CG30120 gene product from transcript CG30120-RB', other_designations='CG30120-PB|uncharacterized protein', map_location='55C9-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18446548, end_position_on_the_genomic_accession=18449420, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246471, CurrentID=0, Status='live', Symbol='Ir56a', Aliases='Dmel_CG30125, CG30125, DmelIR56a, Dmel\\CG30125, IR56a', description='Ionotropic receptor 56a', other_designations='CG30125-PA|Ir56a-PA|ionotropic receptor 56a', map_location='56B3-56B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19094267, end_position_on_the_genomic_accession=19096174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246484, CurrentID=0, Status='live', Symbol='CG30151', Aliases='Dmel_ BcDNA:RE70748, Dmel\\CG30151', description='CG30151 gene product from transcript CG30151-RA', other_designations='CG30151-PA|uncharacterized protein', map_location='57A4-57A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20528503, end_position_on_the_genomic_accession=20529464, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246485, CurrentID=0, Status='live', Symbol='CG30152', Aliases='Dmel_ CG10404, Cg30152, Dmel\\CG30152, FAM96a', description='CG30152 gene product from transcript CG30152-RC', other_designations='CG30152-PA|CG30152-PB|CG30152-PC|CG30152-PD|uncharacterized protein', map_location='57A10-57A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20670636, end_position_on_the_genomic_accession=20673495, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246487, CurrentID=0, Status='live', Symbol='CG30154', Aliases='Dmel_ Dmel\\CG30154', description='CG30154 gene product from transcript CG30154-RA', other_designations='CG30154-PA|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20533359, end_position_on_the_genomic_accession=20533934, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246488, CurrentID=0, Status='live', Symbol='CG30156', Aliases='Dmel_ BEST:GH05253, BcDNA:AT02529, Dmel\\CG30156', description='CG30156 gene product from transcript CG30156-RB', other_designations='CG30156-PA|CG30156-PB|uncharacterized protein', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6981446, end_position_on_the_genomic_accession=6985712, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246489, CurrentID=0, Status='live', Symbol='CG30157', Aliases='Dmel_ Dmel\\CG30157', description='CG30157 gene product from transcript CG30157-RA', other_designations='CG30157-PA|CG30157-PB|uncharacterized protein', map_location='42E1-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6974038, end_position_on_the_genomic_accession=6975133, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246490, CurrentID=0, Status='live', Symbol='CG30160', Aliases='Dmel_ Dmel\\CG30160', description='CG30160 gene product from transcript CG30160-RA', other_designations='CG30160-PA|uncharacterized protein', map_location='42E4-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7005157, end_position_on_the_genomic_accession=7006136, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246492, CurrentID=0, Status='live', Symbol='Cpr60D', Aliases='Dmel_CG30163, BcDNA:RH24125, CG30163, CG4856, Dmel\\CG30163, cpr60d', description='Cuticular protein 60D', other_designations='CG30163-PA|CG30163-PB|Cpr60D-PA|Cpr60D-PB|cuticular protein 60D', map_location='60D13-60D13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24662682, end_position_on_the_genomic_accession=24663406, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246493, CurrentID=0, Status='live', Symbol='prom', Aliases='Dmel_CG42310, CG30164, CG30165, CG30166, CG42310, CG4556, Dmel\\CG42310, Dmel_CG30164, Dmel_CG30165, Dmel_CG30166, eyc', description='prominin', other_designations='CG42310-PB|CG42310-PC|CG42310-PD|CG42310-PE|eyes closed|prom-PB|prom-PC|prom-PD|prom-PE', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24415410, end_position_on_the_genomic_accession=24428387, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  246496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246496, CurrentID=0, Status='live', Symbol='CG30172', Aliases='Dmel_ BcDNA:RH70879, DmelCSP1, Dmel\\CG30172, anon-AE003462.1', description='CG30172 gene product from transcript CG30172-RB', other_designations='CG30172-PA|CG30172-PB|uncharacterized protein', map_location='60B5-60B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24030546, end_position_on_the_genomic_accession=24031427, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246497, CurrentID=0, Status='live', Symbol='HSPC300', Aliases='Dmel_CG30173, Brk1, CG30173, Dmel\\CG30173, SIP1, dHSPC', description='CG30173 gene product from transcript CG30173-RB', other_designations='CG30173-PA|CG30173-PB|HSPC300-PA|HSPC300-PB|syntaxin interacting protein 1', map_location='60B4-60B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24027717, end_position_on_the_genomic_accession=24028789, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  49892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49892, CurrentID=0, Status='live', Symbol='Fis1', Aliases='Dmel_CG17510, BcDNA:RE29957, CG17510, Dmel\\CG17510, NEST:bs15a08, bs15a08.y1, d', description='CG17510 gene product from transcript CG17510-RA', other_designations='CG17510-PA|CG17510-PC|CG17510-PD|CG17510-PE|CG17510-PF|CG17510-PG|Fis1-PA|Fis1-PC|Fis1-PD|Fis1-PE|Fis1-PF|Fis1-PG', map_location='41F5-41F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5603587, end_position_on_the_genomic_accession=5604834, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  49893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49893, CurrentID=0, Status='live', Symbol='CG17508', Aliases='Dmel_ Dmel\\CG17508', description='CG17508 gene product from transcript CG17508-RA', other_designations='CG17508-PA|CG17508-PB|uncharacterized protein', map_location='41F5-41F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5608635, end_position_on_the_genomic_accession=5611586, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  49895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49895, CurrentID=0, Status='live', Symbol='CG17018', Aliases='Dmel_ BcDNA:RE40762, CG17410, Dmel\\CG17018', description='CG17018 gene product from transcript CG17018-RE', other_designations='CG17018-PA|CG17018-PC|CG17018-PD|CG17018-PE|CG17018-PF|uncharacterized protein', map_location='40F7-40F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22308755, end_position_on_the_genomic_accession=22368796, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  246504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246504, CurrentID=0, Status='live', Symbol='ppk3', Aliases='Dmel_CG30181, CG15872, CG30181, Dmel\\CG30181, cg30181', description='pickpocket 3', other_designations='CG30181-PA|ppk3-PA', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23538631, end_position_on_the_genomic_accession=23540773, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246505, CurrentID=0, Status='live', Symbol='CG42559', Aliases='Dmel_ CG30182, Dmel\\CG42559, Dmel_CG30182', description='CG42559 gene product from transcript CG42559-RA', other_designations='CG42559-PA|CG42559-PB|CG42559-PC|uncharacterized protein', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23599526, end_position_on_the_genomic_accession=23600787, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  246506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246506, CurrentID=0, Status='live', Symbol='CG30184', Aliases='Dmel_ Dmel\\CG30184', description='CG30184 gene product from transcript CG30184-RA', other_designations='CG30184-PA|uncharacterized protein', map_location='59F1-59F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23563973, end_position_on_the_genomic_accession=23564780, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246507, CurrentID=0, Status='live', Symbol='CG30185', Aliases='Dmel_ CG5365, Dmel\\CG30185, anon-EST:Posey25', description='CG30185 gene product from transcript CG30185-RA', other_designations='CG30185-PA|uncharacterized protein', map_location='59E3-59E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23544024, end_position_on_the_genomic_accession=23544901, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246509, CurrentID=0, Status='live', Symbol='CG30187', Aliases='Dmel_ BcDNA:SD12357, Dmel\\CG30187', description='CG30187 gene product from transcript CG30187-RC', other_designations='CG30187-PC|CG30187-PD|uncharacterized protein', map_location='59B4-59B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22916843, end_position_on_the_genomic_accession=22918790, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  246510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246510, CurrentID=0, Status='live', Symbol='CG30192', Aliases='Dmel_ BcDNA:AT30929, Dmel\\CG30192', description='CG30192 gene product from transcript CG30192-RA', other_designations='CG30192-PA|uncharacterized protein', map_location='59C5-59D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23138589, end_position_on_the_genomic_accession=23139085, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246511, CurrentID=0, Status='live', Symbol='CG18853', Aliases='Dmel_CG18853, 143303_at-PA, Dmel\\CG18853', description='CG18853 gene product from transcript CG18853-RA', other_designations='CG18853-PA|uncharacterized protein', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7829062, end_position_on_the_genomic_accession=7830429, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246513, CurrentID=0, Status='live', Symbol='CG30197', Aliases='Dmel_ BcDNA:RE05231, Dmel\\CG30197', description='CG30197 gene product from transcript CG30197-RA', other_designations='CG30197-PA|uncharacterized protein', map_location='51A6-51A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14514030, end_position_on_the_genomic_accession=14514992, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246514, CurrentID=0, Status='live', Symbol='CG30203', Aliases='Dmel_ CG17740, CT39323, Dmel\\CG30203', description='CG30203 gene product from transcript CG30203-RB', other_designations='CG30203-PB|uncharacterized protein', map_location='48F7-48F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12310689, end_position_on_the_genomic_accession=12313834, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  246515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246515, CurrentID=0, Status='live', Symbol='CG43325', Aliases='Dmel_ CG13514, CG30217, Dmel\\CG43325, Dmel_CG30217', description='CG43325 gene product from transcript CG43325-RA', other_designations='CG43325-PA|CG43325-PB|CG43325-PC|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22655549, end_position_on_the_genomic_accession=22657135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246518, CurrentID=0, Status='live', Symbol='CG30271', Aliases='Dmel_ BcDNA:AT17081, CG13528, Dmel\\CG30271', description='CG30271 gene product from transcript CG30271-RI', other_designations='CG30271-PF|CG30271-PG|CG30271-PH|CG30271-PI|uncharacterized protein', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22816474, end_position_on_the_genomic_accession=22822111, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  246519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246519, CurrentID=0, Status='live', Symbol='MFS1', Aliases='Dmel_CG30272, CG11682, CG30272, Dmel\\CG30272', description='Major Facilitator Superfamily Transporter 1', other_designations='CG30272-PA|MFS1-PA|major facilitator superfamily transporter 1', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22829644, end_position_on_the_genomic_accession=22831877, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246520, CurrentID=0, Status='live', Symbol='CG30273', Aliases='Dmel_ CG13515, Dmel\\CG30273', description='CG30273 gene product from transcript CG30273-RB', other_designations='CG30273-PB|uncharacterized protein', map_location='58F4-58F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22667155, end_position_on_the_genomic_accession=22668848, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246521, CurrentID=0, Status='live', Symbol='CG30274', Aliases='Dmel_ CG13529, Dmel\\CG30274', description='CG30274 gene product from transcript CG30274-RA', other_designations='CG30274-PA|uncharacterized protein', map_location='59B2-59B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22825093, end_position_on_the_genomic_accession=22828903, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246522, CurrentID=0, Status='live', Symbol='CG30275', Aliases='Dmel_ CG13520, Dmel\\CG30275', description='CG30275 gene product from transcript CG30275-RM', other_designations='CG30275-PL|CG30275-PM|CG30275-PN|uncharacterized protein', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22707794, end_position_on_the_genomic_accession=22712724, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  246523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246523, CurrentID=0, Status='live', Symbol='CG30278', Aliases='Dmel_ CG3387, Dmel\\CG30278', description='CG30278 gene product from transcript CG30278-RC', other_designations='CG30278-PA|CG30278-PB|CG30278-PC|uncharacterized protein', map_location='58D2-58D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22217285, end_position_on_the_genomic_accession=22218500, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246525, CurrentID=0, Status='live', Symbol='CG30281', Aliases='Dmel_ AAM68209, CG6676, Dmel\\CG30281, FBpp0071674', description='CG30281 gene product from transcript CG30281-RA', other_designations='CG30281-PA|Fibrinogen-related protein|uncharacterized protein', map_location='58B9-58B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22114997, end_position_on_the_genomic_accession=22116189, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246527, CurrentID=0, Status='live', Symbol='CG30284', Aliases='Dmel_ BcDNA:AT12437, BcDNA:AT22714, Dmel\\CG30284, NEST:bs16b04', description='CG30284 gene product from transcript CG30284-RA', other_designations='CG30284-PA|CG30284-PB|uncharacterized protein', map_location='57F6-57F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21650820, end_position_on_the_genomic_accession=21651999, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246528, CurrentID=0, Status='live', Symbol='CG30285', Aliases='Dmel_ Dmel\\CG30285', description='CG30285 gene product from transcript CG30285-RA', other_designations='CG30285-PA|uncharacterized protein', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21687903, end_position_on_the_genomic_accession=21688545, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246529, CurrentID=0, Status='live', Symbol='CG30286', Aliases='Dmel_ CG10450, Dmel\\CG30286, SP37, SPH37', description='CG30286 gene product from transcript CG30286-RB', other_designations='CG30286-PB|CG30286-PC|uncharacterized protein', map_location='57E9-57E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21535716, end_position_on_the_genomic_accession=21536795, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246530, CurrentID=0, Status='live', Symbol='CG30287', Aliases='Dmel_ CG10450, Dmel\\CG30287, SP37, SPH37', description='CG30287 gene product from transcript CG30287-RA', other_designations='CG30287-PA|Dmel_CG30287|uncharacterized protein', map_location='57E10-57E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21536894, end_position_on_the_genomic_accession=21538015, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246531, CurrentID=0, Status='live', Symbol='CG30288', Aliases='Dmel_ Dmel\\CG30288', description='CG30288 gene product from transcript CG30288-RC', other_designations='CG30288-PC|uncharacterized protein', map_location='57E9-57E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21519243, end_position_on_the_genomic_accession=21520393, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246532, CurrentID=0, Status='live', Symbol='CG30289', Aliases='Dmel_ Dmel\\CG30289', description='CG30289 gene product from transcript CG30289-RA', other_designations='CG30289-PA|uncharacterized protein', map_location='57E9-57E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21520413, end_position_on_the_genomic_accession=21521848, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246533, CurrentID=0, Status='live', Symbol='Ppcdc', Aliases='Dmel_CG30290, BcDNA:LD37882, CG30290, Dmel\\CG30290, Dromel_CG30290_FBtr0071672_cx9c_mORF, PPCDC, dPPCDC', description='Phosphopantothenoylcysteine decarboxylase', other_designations="(R)-4'-phospho-N-pantothenoylcysteine decarboxylase|CG30290-PA|CG30290-PB|Ppcdc-PA|Ppcdc-PB|phosphopantothenoylcysteine decarboxylase", map_location='57F8-57F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21691295, end_position_on_the_genomic_accession=21692764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246534, CurrentID=0, Status='live', Symbol='CG30291', Aliases='Dmel_ CG3282, Dmel\\CG30291', description='CG30291 gene product from transcript CG30291-RA', other_designations='CG30291-PA|uncharacterized protein', map_location='57B5-57B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21007755, end_position_on_the_genomic_accession=21009576, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246535, CurrentID=0, Status='live', Symbol='Cht12', Aliases='Dmel_CG30293, BcDNA:AT18578, CG30293, Dmel\\CG30293, NEST:bs09g08, NEST:bs28h08, NEST:bs33a06', description='CG30293 gene product from transcript CG30293-RA', other_designations='CG30293-PA|Cht12-PA', map_location='57B16-57B16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21063178, end_position_on_the_genomic_accession=21064986, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246538, CurrentID=0, Status='live', Symbol='Sema-2b', Aliases='Dmel_CG33960, CG30322, CG33960, CG4383, Dmel\\CG33960, Sema2b, SemaIIB, sema-2b', description='Semaphorin-2b', other_designations='CG33960-PA|CG33960-PB|CG33960-PC|CG33960-PD|Sema-2b-PA|Sema-2b-PB|Sema-2b-PC|Sema-2b-PD|semaphorin|semaphorin2b|semiphorin', map_location='53C4-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16386524, end_position_on_the_genomic_accession=16431758, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  246539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246539, CurrentID=0, Status='live', Symbol='CG30323', Aliases='Dmel_ Dmel\\CG30323', description='CG30323 gene product from transcript CG30323-RB', other_designations='CG30323-PB|CG30323-PC|uncharacterized protein', map_location='54F1-54F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17853295, end_position_on_the_genomic_accession=17854296, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246540, CurrentID=0, Status='live', Symbol='CG30324', Aliases='Dmel_ CG7750, Dmel\\CG30324', description='CG30324 gene product from transcript CG30324-RA', other_designations='CG30324-PA|uncharacterized protein', map_location='53B1-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16287586, end_position_on_the_genomic_accession=16288233, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246541, CurrentID=0, Status='live', Symbol='CG30325', Aliases='Dmel_ Dmel\\CG30325', description='CG30325 gene product from transcript CG30325-RA', other_designations='CG30325-PA|uncharacterized protein', map_location='54F5-54F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17877543, end_position_on_the_genomic_accession=17880012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246543, CurrentID=0, Status='live', Symbol='Vha100-3', Aliases='Dmel_CG30329, BcDNA:AT03238, CG30329, Dmel\\CG30329, vha100-3', description='Vacuolar H[+] ATPase 100kD subunit 3', other_designations='CG30329-PA|Vha100-3-PA|vacuolar H[+] ATPase 100kD subunit 3|vacuolar H[+] ATPase subunit 100-3', map_location='55F11-55F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18908961, end_position_on_the_genomic_accession=18911895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246545, CurrentID=0, Status='live', Symbol='CG30334', Aliases='Dmel_ Dmel\\CG30334, NEST:bs09a11', description='CG30334 gene product from transcript CG30334-RB', other_designations='CG30334-PA|CG30334-PB|uncharacterized protein', map_location='49A6-49A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12440633, end_position_on_the_genomic_accession=12441320, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246548, CurrentID=0, Status='live', Symbol='CG30338', Aliases='Dmel_ Dmel\\CG30338', description='CG30338 gene product from transcript CG30338-RA', other_designations='CG30338-PA|uncharacterized protein', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9594045, end_position_on_the_genomic_accession=9599505, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246549, CurrentID=0, Status='live', Symbol='CG30339', Aliases='Dmel_ BcDNA:RE54230, Dmel\\CG30339', description='CG30339 gene product from transcript CG30339-RA', other_designations='CG30339-PA|uncharacterized protein', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9601431, end_position_on_the_genomic_accession=9602901, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246550, CurrentID=0, Status='live', Symbol='CG30340', Aliases='Dmel_ BACR48G21.1, BG:BACR48G21.1, Dmel\\CG30340', description='CG30340 gene product from transcript CG30340-RA', other_designations='CG30340-PA|uncharacterized protein', map_location='45F6-45F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9599909, end_position_on_the_genomic_accession=9601252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246551, CurrentID=0, Status='live', Symbol='CG30343', Aliases='Dmel_ BcDNA:GM04716, Dmel\\CG30343', description='CG30343 gene product from transcript CG30343-RA', other_designations='CG30343-PA|CG30343-PB|uncharacterized protein', map_location='45C1-45C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9252021, end_position_on_the_genomic_accession=9253255, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246552, CurrentID=0, Status='live', Symbol='CG30344', Aliases='Dmel_ CG8054, Dmel\\CG30344, MET', description='CG30344 gene product from transcript CG30344-RA', other_designations='CG30344-PA|multidrug efflux transporter|uncharacterized protein', map_location='45A12-45A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9145260, end_position_on_the_genomic_accession=9148512, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  246553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246553, CurrentID=0, Status='live', Symbol='CG30345', Aliases='Dmel_ Dmel\\CG30345', description='CG30345 gene product from transcript CG30345-RB', other_designations='CG30345-PB|uncharacterized protein', map_location='45A13-45A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9148689, end_position_on_the_genomic_accession=9151892, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  246556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246556, CurrentID=0, Status='live', Symbol='CG30350', Aliases='Dmel_ CG8278, Dmel\\CG30350', description='CG30350 gene product from transcript CG30350-RA', other_designations='CG30350-PA|uncharacterized protein', map_location='44F1-44F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8826391, end_position_on_the_genomic_accession=8827755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  49952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49952, CurrentID=0, Status='live', Symbol='CG17450', Aliases='Dmel_ Dmel\\CG17450', description='CG17450 gene product from transcript CG17450-RC', other_designations='CG17450-PA|CG17450-PB|CG17450-PC|tektin|uncharacterized protein', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21605884, end_position_on_the_genomic_accession=21608920, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49953, CurrentID=0, Status='live', Symbol='GCS2alpha', Aliases='Dmel_CG14476, BcDNA:GH04962, CG14476, Dmel\\CG14476, clot#312', description='Glucosidase 2 alpha subunit', other_designations='CG14476-PA|CG14476-PB|CG14476-PC|CG14476-PD|CG14476-PE|GCS2alpha-PA|GCS2alpha-PB|GCS2alpha-PC|GCS2alpha-PD|GCS2alpha-PE|alpha-glucosidase|glucosidase 2 alpha subunit|glucosidase II', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21600502, end_position_on_the_genomic_accession=21605264, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246562, CurrentID=0, Status='live', Symbol='CG30356', Aliases='Dmel_ Dmel\\CG30356', description='CG30356 gene product from transcript CG30356-RA', other_designations='CG30356-PA|uncharacterized protein', map_location='44E3-44E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8728964, end_position_on_the_genomic_accession=8729501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246563, CurrentID=0, Status='live', Symbol='CG30357', Aliases='Dmel_ BcDNA:AT27059, Dmel\\CG30357', description='CG30357 gene product from transcript CG30357-RA', other_designations='CG30357-PA|uncharacterized protein', map_location='44E3-44E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8720100, end_position_on_the_genomic_accession=8721624, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246565, CurrentID=0, Status='live', Symbol='Mal-A6', Aliases='Dmel_CG30360, CG30360, CG8689, Dmel\\CG30360, Drome_A6, mal_A6', description='Maltase A6', other_designations='CG30360-PC|CG30360-PD|Mal-A6-PC|Mal-A6-PD|maltase A6', map_location='44D1-44D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8465158, end_position_on_the_genomic_accession=8467595, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246566, CurrentID=0, Status='live', Symbol='cola', Aliases='Dmel_CG30363, CG2149, CG30363, Dmel\\CG30363', description='comas sola', other_designations='CG30363-PA|cola-PA', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8363272, end_position_on_the_genomic_accession=8365345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246567, CurrentID=0, Status='live', Symbol='hubl', Aliases='Dmel_CG30364, CG2149, CG30364, Dmel\\CG30364, anon-WO0140519.168', description='hug-bell', other_designations='CG30364-PA|CG30364-PB|hubl-PA|hubl-PB', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8359466, end_position_on_the_genomic_accession=8362156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246568, CurrentID=0, Status='live', Symbol='spaw', Aliases='Dmel_CG30365, BcDNA:AT04457, CG2149, CG30365, Dmel\\CG30365', description='spacewatch', other_designations='CG30365-PA|CG30365-PB|spaw-PA|spaw-PB', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8359466, end_position_on_the_genomic_accession=8362156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246569, CurrentID=0, Status='live', Symbol='swif', Aliases='Dmel_CG30366, CG2149, CG30366, Dmel\\CG30366', description='swift-tuttle', other_designations='CG30366-PA|CG30366-PB|swif-PA|swif-PB', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8363272, end_position_on_the_genomic_accession=8365345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246572, CurrentID=0, Status='live', Symbol='CG30369', Aliases='Dmel_ Dmel\\CG30369, NEST:bs06h11', description='CG30369 gene product from transcript CG30369-RA', other_designations='CG30369-PA|uncharacterized protein', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8259286, end_position_on_the_genomic_accession=8259784, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246573, CurrentID=0, Status='live', Symbol='CG30373', Aliases='Dmel_ BcDNA:RH67738, Dmel\\CG30373', description='CG30373 gene product from transcript CG30373-RB', other_designations='CG30373-PA|CG30373-PB|uncharacterized protein', map_location='44B7-44B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8164000, end_position_on_the_genomic_accession=8164869, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246575, CurrentID=0, Status='live', Symbol='CG30375', Aliases='Dmel_ CG2280, Dmel\\CG30375, SP80', description='CG30375 gene product from transcript CG30375-RA', other_designations='CG30375-PA|uncharacterized protein', map_location='44C3-44C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8265693, end_position_on_the_genomic_accession=8267295, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  49968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49968, CurrentID=0, Status='live', Symbol='Cadps', Aliases='Dmel_CG33653, BcDNA:GH07283, CAPS, CG17881, CG18026, CG33653, CG40477, Caps, Dmel\\CG33653, SDCAP, cap, dCAPS, dcaps, l(4)102Fb', description='Calcium-dependent secretion activator', other_designations='CG33653-PA|CG33653-PB|CG33653-PD|CG33653-PE|CG33653-PF|Cadps-PA|Cadps-PB|Cadps-PD|Cadps-PE|Cadps-PF|Calcium-activated protein for secretion|calcium activated protein for secretion', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1230837, end_position_on_the_genomic_accession=1271759, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  246577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246577, CurrentID=0, Status='live', Symbol='CG30378', Aliases='Dmel_ Dmel\\CG30378, NEST:bs08d04', description='CG30378 gene product from transcript CG30378-RA', other_designations='CG30378-PA|uncharacterized protein', map_location='43F6-43F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7950339, end_position_on_the_genomic_accession=7951492, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246578, CurrentID=0, Status='live', Symbol='CG30379', Aliases='Dmel_ Dmel\\CG30379', description='CG30379 gene product from transcript CG30379-RB', other_designations='CG30379-PB|uncharacterized protein', map_location='43F4-43F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7939956, end_position_on_the_genomic_accession=7941870, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246579, CurrentID=0, Status='live', Symbol='CG30380', Aliases='Dmel_ Dmel\\CG30380', description='CG30380 gene product from transcript CG30380-RA', other_designations='CG30380-PA|uncharacterized protein', map_location='43F4-43F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7939201, end_position_on_the_genomic_accession=7939659, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246580, CurrentID=0, Status='live', Symbol='CG30381', Aliases='Dmel_ Dmel\\CG30381', description='CG30381 gene product from transcript CG30381-RB', other_designations='CG30381-PB|CG30381-PC|uncharacterized protein', map_location='43F2-43F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7929658, end_position_on_the_genomic_accession=7931016, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246581, CurrentID=0, Status='live', Symbol='GstE9', Aliases='Dmel_CG17534, CG17534, CT38753, DmGSTE9, Dmel\\CG17534, E9, GST3-1, GSTE9, Gst3-1, gste9', description='Glutathione S transferase E9', other_designations='CG17534-PA|GstE9-PA|glutathion S-transferase|glutathione S transferase E9', map_location='55C8-55C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18409048, end_position_on_the_genomic_accession=18409835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246582, CurrentID=0, Status='live', Symbol='CG30382', Aliases='Dmel_ 30382, Dmel\\CG30382, Prosalpha1R, Prosalpha6', description='CG30382 gene product from transcript CG30382-RB', other_designations='CG30382-PA|CG30382-PB|alpha-subunit of 20S proteasome|proteasome alpha1 subunit-related|uncharacterized protein', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7830413, end_position_on_the_genomic_accession=7831771, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246583, CurrentID=0, Status='live', Symbol='CG30383', Aliases='Dmel_ Dmel\\CG30383', description='CG30383 gene product from transcript CG30383-RA', other_designations='CG30383-PA|uncharacterized protein', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7823943, end_position_on_the_genomic_accession=7825141, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246584, CurrentID=0, Status='live', Symbol='CG30384', Aliases='Dmel_ CG11074, Dmel\\CG30384', description='CG30384 gene product from transcript CG30384-RA', other_designations='CG30384-PA|uncharacterized protein', map_location='43A1-43A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7182997, end_position_on_the_genomic_accession=7185298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246585, CurrentID=0, Status='live', Symbol='CG30385', Aliases='Dmel_ BcDNA:AT12446, CG11074, Dmel\\CG30385', description='CG30385 gene product from transcript CG30385-RA', other_designations='CG30385-PA|uncharacterized protein', map_location='43A1-43A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7181326, end_position_on_the_genomic_accession=7183020, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246587, CurrentID=0, Status='live', Symbol='CG30392', Aliases='Dmel_ CG10509, Dmel\\CG30392, anon-WO0153538.80', description='CG30392 gene product from transcript CG30392-RB', other_designations='CG30392-PA|CG30392-PB|uncharacterized protein', map_location='57D8-57D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21290914, end_position_on_the_genomic_accession=21292181, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246588, CurrentID=0, Status='live', Symbol='CG30393', Aliases='Dmel_ CG15662, Dmel\\CG30393', description='CG30393 gene product from transcript CG30393-RA', other_designations='CG30393-PA|uncharacterized protein', map_location='57D4-57D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21266364, end_position_on_the_genomic_accession=21267579, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246589, CurrentID=0, Status='live', Symbol='CG30395', Aliases='Dmel_ Dmel\\CG30395', description='CG30395 gene product from transcript CG30395-RB', other_designations='CG30395-PB|uncharacterized protein', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21765489, end_position_on_the_genomic_accession=21770196, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  246591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246591, CurrentID=0, Status='live', Symbol='CG30398', Aliases='Dmel_ Dmel\\CG30398', description='CG30398 gene product from transcript CG30398-RA', other_designations='CG30398-PA|uncharacterized protein', map_location='57F10-57F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21717804, end_position_on_the_genomic_accession=21718832, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246593, CurrentID=0, Status='live', Symbol='CG30401', Aliases='Dmel_ CG13496, Dmel\\CG30401', description='CG30401 gene product from transcript CG30401-RC', other_designations='CG30401-PA|CG30401-PC|uncharacterized protein', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21985626, end_position_on_the_genomic_accession=21986938, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246595, CurrentID=0, Status='live', Symbol='CG30403', Aliases='Dmel_ BcDNA:RE26725, Dmel\\CG30403', description='CG30403 gene product from transcript CG30403-RB', other_designations='CG30403-PB|uncharacterized protein', map_location='57F10-57F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21709895, end_position_on_the_genomic_accession=21711613, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246596, CurrentID=0, Status='live', Symbol='Tango11', Aliases='Dmel_CG30404, BEST:LD33633, BcDNA:LD22567, CG30404, Dmel\\CG30404, TANGO11, anon-EST:fe1F11, tango11', description='Transport and Golgi organization 11', other_designations='CG30404-PA|CG30404-PB|CG30404-PC|Tango11-PA|Tango11-PB|Tango11-PC|anon-fast-evolving-1F11|tango11|transport and golgi organization|transport and golgi organization 11', map_location='57F10-57F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21716780, end_position_on_the_genomic_accession=21719963, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246599, CurrentID=0, Status='live', Symbol='Rpi', Aliases='Dmel_CG30410, CG30410, CG9872, Dmel\\CG30410, rpi', description='Ribose-5-phosphate isomerase', other_designations='CG30410-PB|CG30410-PC|Rpi-PB|Rpi-PC|ribose-5-phosphate isomerase', map_location='59D8-59D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23355183, end_position_on_the_genomic_accession=23356441, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  49992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49992, CurrentID=0, Status='live', Symbol='CG15219', Aliases='Dmel_ BcDNA:LP09525, Dmel\\CG15219, cg15219', description='CG15219 gene product from transcript CG15219-RC', other_designations='CG15219-PA|CG15219-PC|uncharacterized protein', map_location='40E3-40E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22104891, end_position_on_the_genomic_accession=22105512, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246601, CurrentID=0, Status='live', Symbol='CG30413', Aliases='Dmel_ Dmel\\CG30413', description='CG30413 gene product from transcript CG30413-RA', other_designations='CG30413-PA|uncharacterized protein', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23329543, end_position_on_the_genomic_accession=23330068, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246602, CurrentID=0, Status='live', Symbol='CG30415', Aliases='Dmel_ CG3487, Dmel\\CG30415, anon-SAGE:Wang-079', description='CG30415 gene product from transcript CG30415-RB', other_designations='CG30415-PA|CG30415-PB|anon-SAGE:Wang-079|uncharacterized protein', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23385464, end_position_on_the_genomic_accession=23386463, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246603, CurrentID=0, Status='live', Symbol='CG30416', Aliases='Dmel_ BcDNA:AT11052, CG9803, Dmel\\CG30416', description='CG30416 gene product from transcript CG30416-RB', other_designations='CG30416-PB|CG30416-PC|uncharacterized protein', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23321481, end_position_on_the_genomic_accession=23323399, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246604, CurrentID=0, Status='live', Symbol='tbrd-3', Aliases='Dmel_CG30417, CG17692, CG30417, Dmel\\CG30417, tBRD-3', description='testis-specifically expressed bromodomain containing protein-3', other_designations='CG30417-PA|tbrd-3-PA', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23326739, end_position_on_the_genomic_accession=23327581, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  49997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=49997, CurrentID=0, Status='live', Symbol='CG11052', Aliases='Dmel_ Dmel\\CG11052', description='CG11052 gene product from transcript CG11052-RB', other_designations='CG11052-PA|CG11052-PB|CG11052-PC|CG11052-PD|uncharacterized protein', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8008249, end_position_on_the_genomic_accession=8009297, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246606, CurrentID=0, Status='live', Symbol='CG30424', Aliases='Dmel_ Dmel\\CG30424, mdcds_18556', description='CG30424 gene product from transcript CG30424-RB', other_designations='CG30424-PB|CG30424-PC|CG30424-PD|CG30424-PE|uncharacterized protein', map_location='60E11-60E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24965275, end_position_on_the_genomic_accession=24967298, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246608, CurrentID=0, Status='live', Symbol='CG30428', Aliases='Dmel_ BcDNA:SD21514, Dmel\\CG30428', description='CG30428 gene product from transcript CG30428-RB', other_designations='CG30428-PA|CG30428-PB|uncharacterized protein', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25251927, end_position_on_the_genomic_accession=25254535, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246609, CurrentID=0, Status='live', Symbol='CG30429', Aliases='Dmel_ Dmel\\CG30429, Q8MLN3_DROME', description='CG30429 gene product from transcript CG30429-RA', other_designations='CG30429-PA|uncharacterized protein', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25246467, end_position_on_the_genomic_accession=25247584, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50002, CurrentID=0, Status='live', Symbol='COX7A', Aliases='Dmel_CG9603, CG9603, COXJ, COXK, Dmel\\CG9603, VIIa', description='Cytochrome c oxidase subunit 7A', other_designations='CG9603-PA|CG9603-PB|COX7A-PA|COX7A-PB|cytochrome c oxidase VIIa|cytochrome c oxidase polypeptide VIIa|cytochrome c oxidase subunit 7A', map_location='84F13-84F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8339944, end_position_on_the_genomic_accession=8341043, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  50003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50003, CurrentID=0, Status='live', Symbol='CG45263', Aliases='Dmel_ 43462, BcDNA:AT05815, BcDNA:GH22170, CG11718, CG11729, CG11730, CG11731, CG11733, CG31264, CG31369, CG31369/CG31462, CG31462, CG42285, CG42286, CG43462, CT32186, Dmel\\CG45263, Dmel_CG11718, Dmel_CG31369, Dmel_CG31462, Dmel_CG42285, Dmel_CG42286, Dmel_CG43462, GH22170, GS3056, MRE1, anon-EST:Posey178', description='CG45263 gene product from transcript CG45263-RC', other_designations='CG45263-PB|CG45263-PC|CG45263-PD|CG45263-PE|CG45263-PF|CG45263-PG|mRNA-like ncRNA in embryogenesis 1|uncharacterized protein', map_location='85A1-85A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8364328, end_position_on_the_genomic_accession=8477617, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  50005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50005, CurrentID=0, Status='live', Symbol='CG8369', Aliases='Dmel_ BcDNA:GH11984, Dmel\\CG8369, anon-EST:Posey161', description='CG8369 gene product from transcript CG8369-RC', other_designations='CG8369-PA|CG8369-PB|CG8369-PC', map_location='85B2-85B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8822008, end_position_on_the_genomic_accession=8822838, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246614, CurrentID=0, Status='live', Symbol='CG30438', Aliases='Dmel_ BcDNA:LD09936, BcDNA:RE54684, Dmel\\CG30438, GlcAT-P, dGlcAT-BSII', description='CG30438 gene product from transcript CG30438-RB', other_designations='CG30438-PA|CG30438-PB|CG30438-PC|CG30438-PD|CG30438-PE|uncharacterized protein', map_location='41F2-41F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5496674, end_position_on_the_genomic_accession=5549543, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  50007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50007, CurrentID=0, Status='live', Symbol='CG11985', Aliases='Dmel_ Dmel\\CG11985', description='CG11985 gene product from transcript CG11985-RA', other_designations='CG11985-PA|uncharacterized protein', map_location='85C2-85C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9014972, end_position_on_the_genomic_accession=9015341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246616, CurrentID=0, Status='live', Symbol='CG30441', Aliases='Dmel_ Dmel\\CG30441, IFT20', description='CG30441 gene product from transcript CG30441-RB', other_designations='CG30441-PB|uncharacterized protein', map_location='41E5-41E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5351717, end_position_on_the_genomic_accession=5352122, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674004, CurrentID=0, Status='live', Symbol='CG42495', Aliases='Dmel_ Dmel\\CG42495', description='CG42495 gene product from transcript CG42495-RA', other_designations='CG42495-PA|uncharacterized protein', map_location='75D5-75D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18667361, end_position_on_the_genomic_accession=18668135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246618, CurrentID=0, Status='live', Symbol='Opbp', Aliases='Dmel_CG30443, BcDNA:RE37461, CG12397, CG30443, Dmel\\CG30443, OBP, Obp', description='Optix-binding protein', other_designations='CG30443-PA|Opbp-PA|optix binding protein', map_location='42C1-42C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6653217, end_position_on_the_genomic_accession=6655267, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246620, CurrentID=0, Status='live', Symbol='Tdc2', Aliases='Dmel_CG30446, CG30446, CG3686, Dmel\\CG30446, TDC2, dTdc2, tdc, tdc2', description='Tyrosine decarboxylase 2', other_designations='CG30446-PA|Tdc2-PA|tyrosine decarboxylase|tyrosine decarboxylase 2', map_location='42C4-42C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6684745, end_position_on_the_genomic_accession=6689974, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  246621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246621, CurrentID=0, Status='live', Symbol='CG30447', Aliases='Dmel_ Dmel\\CG30447', description='CG30447 gene product from transcript CG30447-RB', other_designations='CG30447-PA|CG30447-PB|CG30447-PC|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19947196, end_position_on_the_genomic_accession=19948080, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246624, CurrentID=0, Status='live', Symbol='CG30458', Aliases='Dmel_ CG17288, Dmel\\CG30458', description='CG30458 gene product from transcript CG30458-RA', other_designations='CG30458-PA|uncharacterized protein', map_location='54A2-54A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17158325, end_position_on_the_genomic_accession=17159038, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246625, CurrentID=0, Status='live', Symbol='CG43328', Aliases='Dmel_ CG30459, Dmel\\CG43328, Dmel_CG30459', description='CG43328 gene product from transcript CG43328-RA', other_designations='CG43328-PA|CG43328-PB|uncharacterized protein', map_location='53E10-53E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17032487, end_position_on_the_genomic_accession=17034012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246627, CurrentID=0, Status='live', Symbol='CG30463', Aliases='Dmel_ CG15713, CG8456, CG8463, Dmel\\CG30463, dppGalNAcT9', description='CG30463 gene product from transcript CG30463-RA', other_designations='CG30463-PA|CG30463-PB|CG30463-PC|CG30463-PD|CG30463-PE|uncharacterized protein', map_location='53D1-53D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16673235, end_position_on_the_genomic_accession=16732916, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  246629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246629, CurrentID=0, Status='live', Symbol='CG30466', Aliases='Dmel_ Dmel\\CG30466, FBXO19', description='CG30466 gene product from transcript CG30466-RB', other_designations='CG30466-PA|CG30466-PB|uncharacterized protein', map_location='52A13-52A13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15568960, end_position_on_the_genomic_accession=15573170, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50022, CurrentID=0, Status='live', Symbol='Hsp70Bbb', Aliases='Dmel_CG5834, CG31354, CG5834, Dm-hsp70, Dmel\\CG5834, HSP70, HSP70Bbb, Hsp-70, Hsp70, Hsp70B, Hsp70Bb, dHsp70, hsp70, hsp70B, hsp70Bb-prime', description='CG5834 gene product from transcript CG5834-RA', other_designations='CG5834-PA|Hsp70Bbb-PA|heat shock protein 70|heat-shock protein 70', map_location='87B14-87B14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12502510, end_position_on_the_genomic_accession=12504856, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246631, CurrentID=0, Status='live', Symbol='Ir52b', Aliases='Dmel_CG30469, CG12961, CG30469, DmelIR52b, Dmel\\CG30469, IF52b, IR52b', description='Ionotropic receptor 52b', other_designations='CG30469-PB|Ir52b-PB|ionotropic receptor 52b', map_location='52A6-52A6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15501820, end_position_on_the_genomic_accession=15503610, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50024, CurrentID=0, Status='live', Symbol='CG11686', Aliases='Dmel_ BcDNA:RH19248, Dmel\\CG11686', description='CG11686 gene product from transcript CG11686-RB', other_designations='CG11686-PA|CG11686-PB|uncharacterized protein', map_location='87E3-87E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13260356, end_position_on_the_genomic_accession=13262227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246633, CurrentID=0, Status='live', Symbol='CG30471', Aliases='Dmel_ CG8185, Dmel\\CG30471', description='CG30471 gene product from transcript CG30471-RB', other_designations='CG30471-PB|uncharacterized protein', map_location='52A10-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15533520, end_position_on_the_genomic_accession=15537782, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  50026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50026, CurrentID=0, Status='live', Symbol='CG14374', Aliases='Dmel_ Dmel\\CG14374', description='CG14374 gene product from transcript CG14374-RA', other_designations='CG14374-PA|uncharacterized protein', map_location='87E9-87E9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13396694, end_position_on_the_genomic_accession=13397140, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246638, CurrentID=0, Status='live', Symbol='ave', Aliases='Dmel_CG30476, BcDNA:RE67675, CG30476, CG30476/Ave, Dmel\\CG30476, HYP, hyp', description='aveugle', other_designations='CG30476-PA|ave-PA|hyphen', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14752521, end_position_on_the_genomic_accession=14753269, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50032, CurrentID=0, Status='live', Symbol='CG42500', Aliases='Dmel_ CG14359, Dmel\\CG42500', description='CG42500 gene product from transcript CG42500-RB', other_designations='CG42500-PA|CG42500-PB|uncharacterized protein', map_location='88A10-88A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14152211, end_position_on_the_genomic_accession=14152650, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246644, CurrentID=0, Status='live', Symbol='CG30485', Aliases='Dmel_ CG13342, Dmel\\CG30485', description='CG30485 gene product from transcript CG30485-RA', other_designations='CG30485-PA|uncharacterized protein', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13849882, end_position_on_the_genomic_accession=13852983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246645, CurrentID=0, Status='live', Symbol='CG30486', Aliases='Dmel_ Dmel\\CG30486', description='CG30486 gene product from transcript CG30486-RA', other_designations='CG30486-PA|uncharacterized protein', map_location='49D4-49D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12848034, end_position_on_the_genomic_accession=12848993, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246646, CurrentID=0, Status='live', Symbol='CG30487', Aliases='Dmel_ Dmel\\CG30487, NEST:bs16h04, bs33e02.y1', description='CG30487 gene product from transcript CG30487-RA', other_designations='CG30487-PA|uncharacterized protein', map_location='49E1-49E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12937616, end_position_on_the_genomic_accession=12938343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246647, CurrentID=0, Status='live', Symbol='antr', Aliases='Dmel_CG30488, CG30488, Dmel\\CG30488', description='antares', other_designations='CG30488-PA|CG30488-PB|antr-PA|antr-PB', map_location='49D4-49D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12846848, end_position_on_the_genomic_accession=12848004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246648, CurrentID=0, Status='live', Symbol='Cyp12d1-p', Aliases='Dmel_CG30489, 12D1, 12d1, CG18240, CG30489, CYP12D1, Cyp 12d1, Cyp12d1, Dmel\\CG30489', description='CG30489 gene product from transcript CG30489-RB', other_designations='CG30489-PA|CG30489-PB|Cyp12d1-p-PA|Cyp12d1-p-PB|cytochrome P450|cytochrome P450-12d1', map_location='47D4-47D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11117225, end_position_on_the_genomic_accession=11121884, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246650, CurrentID=0, Status='live', Symbol='CG30493', Aliases='Dmel_ CG1952, Dmel\\CG30493', description='CG30493 gene product from transcript CG30493-RB', other_designations='CG30493-PB|uncharacterized protein', map_location='43E11-43E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7699253, end_position_on_the_genomic_accession=7703501, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246651, CurrentID=0, Status='live', Symbol='CG30495', Aliases='Dmel_ CG17986, Dmel\\CG30495', description='CG30495 gene product from transcript CG30495-RB', other_designations='CG30495-PA|CG30495-PB|uncharacterized protein', map_location='43E7-43E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7656895, end_position_on_the_genomic_accession=7659061, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246652, CurrentID=0, Status='live', Symbol='CG30496', Aliases='Dmel_ BcDNA:RE73310, CG1952, Dmel\\CG30496, anon-WO0118547.218', description='CG30496 gene product from transcript CG30496-RA', other_designations='CG30496-PA|uncharacterized protein', map_location='43E11-43E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7699500, end_position_on_the_genomic_accession=7701878, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246653, CurrentID=0, Status='live', Symbol='CG30497', Aliases='Dmel_ BEST:LD33067, CG1383, CG1557, CG30947, Dmel\\CG30497, EP(2)2618, NEST:bs36c04, anon-WO0118547.176', description='CG30497 gene product from transcript CG30497-RD', other_designations='CG30497-PA|CG30497-PB|CG30497-PC|CG30497-PD|uncharacterized protein', map_location='43E13-43E16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7737019, end_position_on_the_genomic_accession=7782637, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246656, CurrentID=0, Status='live', Symbol='CG30503', Aliases='Dmel_ BcDNA:RH50933, CG1672, Dmel\\CG30503', description='CG30503 gene product from transcript CG30503-RB', other_designations='CG30503-PA|CG30503-PB|uncharacterized protein', map_location='43C1-43C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7443824, end_position_on_the_genomic_accession=7445792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50050, CurrentID=0, Status='live', Symbol='CG17278', Aliases='Dmel_ BcDNA:SD04019, Dmel\\CG17278', description='CG17278 gene product from transcript CG17278-RA', other_designations='CG17278-PA|uncharacterized protein', map_location='93B4-93B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21021044, end_position_on_the_genomic_accession=21024163, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246659, CurrentID=0, Status='live', Symbol='CheA29a', Aliases='Dmel_CG33194, CG33194, Dmel\\CG33194', description='Chemosensory protein A 29a', other_designations='CG33194-PA|CheA29a-PA|chemosensory protein A 29a', map_location='29D1-29D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8477349, end_position_on_the_genomic_accession=8478088, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246663, CurrentID=0, Status='live', Symbol='Prx2540-1', Aliases='Dmel_CG12405, 1-cys peroxiredoxin DPx- 2540-1, BEST:LD11278, CG12405, DPx-2540-1, DPx2540, Dmel\\CG12405, Prx2540', description='Peroxiredoxin 2540-1', other_designations='AOP2-related|CG12405-PA|Prx2540-1-PA|peroxiredoxin 2540|peroxiredoxin 2540-1', map_location='47A7-47A7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10424436, end_position_on_the_genomic_accession=10425471, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  246666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246666, CurrentID=0, Status='live', Symbol='Obp51a', Aliases='Dmel_CG30473, 51a, CG30473, DmelObp51a, Dmel\\CG30473, Obp-c, anon-AE003812.1', description='Odorant-binding protein 51a', other_designations='CG30473-PA|Obp51a-PA', map_location='51D8-51D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15024375, end_position_on_the_genomic_accession=15024839, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246667, CurrentID=0, Status='live', Symbol='Obp56i', Aliases='Dmel_CG30448, 56I, 56i, CG30448, DmelObp56i, Dmel\\CG30448, Obp-a, anon-AE003794.2', description='Odorant-binding protein 56i', other_designations='CG30448-PA|CG30448-PB|Obp56i-PA|Obp56i-PB', map_location='56E4-56E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19864717, end_position_on_the_genomic_accession=19868714, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246668, CurrentID=0, Status='live', Symbol='Obp56f', Aliases='Dmel_CG30450, 56f, CG30450, DmelObp56f, Dmel\\CG30450, Obp-d, anon-AE003794.1', description='Odorant-binding protein 56f', other_designations='CG30450-PA|Obp56f-PA', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19713395, end_position_on_the_genomic_accession=19713976, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246669, CurrentID=0, Status='live', Symbol='Obp57b', Aliases='Dmel_CG30142, 57b, CG30142, DmelObp57b, Dmel\\CG30142, Obp-e, anon-AE003792.2, obp57b', description='Odorant-binding protein 57b', other_designations='CG30142-PA|CG30142-PB|Obp57b-PA|Obp57b-PB', map_location='57A4-57A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20504251, end_position_on_the_genomic_accession=20505136, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246670, CurrentID=0, Status='live', Symbol='Obp57a', Aliases='Dmel_CG30141, 57a, CG30141, DmelObp57a, Dmel\\CG30141, OBP57a, Obp-f, Obp57c, anon-AE003792.1, obp57a', description='Odorant-binding protein 57a', other_designations='CG30141-PA|Obp57a-PA|general odorant-binding protein 57a precursor', map_location='57A4-57A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20505150, end_position_on_the_genomic_accession=20505758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246671, CurrentID=0, Status='live', Symbol='Obp57d', Aliases='Dmel_CG30150, 57d, CG30150, DmelObp57d, Dmel\\CG30150, Dmel\\Obp57d, OBP57d, Obp-g, Obp57e, anon-AE003791.1', description='Odorant-binding protein 57d', other_designations='CG30150-PB|Obp57d-PB', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20548368, end_position_on_the_genomic_accession=20550117, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246672, CurrentID=0, Status='live', Symbol='Obp22a', Aliases='Dmel_CG31941, 22a, CG31941, DmelObp22a, Dmel\\CG31941, Obp-b, anon-AE003584.1, obp22', description='Odorant-binding protein 22a', other_designations='CG31941-PB|CG31941-PC|Obp22a-PB|Obp22a-PC', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1991381, end_position_on_the_genomic_accession=1992160, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50068, CurrentID=0, Status='live', Symbol='CG33111', Aliases='Dmel_ CG10193, CG11945, CG11957, Dmel\\CG33111', description='CG33111 gene product from transcript CG33111-RC', other_designations='CG33111-PA|CG33111-PB|CG33111-PC|CG33111-PD|CG33111-PE|uncharacterized protein', map_location='95B7-95C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23801714, end_position_on_the_genomic_accession=23819237, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  50070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50070, CurrentID=0, Status='live', Symbol='mask', Aliases='Dmel_CG33106, CG18671, CG31138, CG33106, CG6268, CG6313, Dmel\\CG33106, EC3-6, MASK, Mask, cg6313', description='multiple ankyrin repeats single KH domain', other_designations='CG33106-PA|CG33106-PB|CG33106-PC|CG33106-PD|CG33106-PE|CG33106-PF|mask-PA|mask-PB|mask-PC|mask-PD|mask-PE|mask-PF', map_location='95F3-95F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24230562, end_position_on_the_genomic_accession=24250248, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  50072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50072, CurrentID=0, Status='live', Symbol='CG42812', Aliases='Dmel_ CG18669, Dmel\\CG42812, Dmel_CG18669', description='CG42812 gene product from transcript CG42812-RA', other_designations='CG42812-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24367360, end_position_on_the_genomic_accession=24368164, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50073, CurrentID=0, Status='live', Symbol='CG13631', Aliases='Dmel_ BcDNA:GH19356, Dmel\\CG13631', description='CG13631 gene product from transcript CG13631-RB', other_designations='CG13631-PA|CG13631-PB|uncharacterized protein', map_location='96A20-96A20', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24769351, end_position_on_the_genomic_accession=24770933, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246682, CurrentID=0, Status='live', Symbol='CG17724', Aliases='Dmel_CG17724a, CG4037, Dmel\\CG17724', description='CG17724 gene product from transcript CG17724-RB', other_designations='CG17724-PA|CG17724-PB|CG17724-PD|CG17724-PE|CG17724-PF|uncharacterized protein', map_location='49F4-49F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13176094, end_position_on_the_genomic_accession=13190086, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  50075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50075, CurrentID=0, Status='live', Symbol='CG14545', Aliases='Dmel_ BcDNA:RE55691, Dmel\\CG14545', description='CG14545 gene product from transcript CG14545-RA', other_designations='CG14545-PA|uncharacterized protein', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25850410, end_position_on_the_genomic_accession=25851291, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50079, CurrentID=0, Status='live', Symbol='CG6503', Aliases='Dmel_ BcDNA:RH63361, CT20261, Dmel\\CG6503, pncr007:3R', description='CG6503 gene product from transcript CG6503-RA', other_designations='CG6503-PA|CG6503-PB|CG6503-PC|putative noncoding RNA 007:3R', map_location='97B2-97B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26424698, end_position_on_the_genomic_accession=26425105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26067043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067043, CurrentID=0, Status='live', Symbol='CG45765', Aliases='Dmel_ Dmel\\CG45765', description='CG45765 gene product from transcript CG45765-RA', other_designations='CG45765-PA|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3293606, end_position_on_the_genomic_accession=3294480, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50082, CurrentID=0, Status='live', Symbol='CG3348', Aliases='Dmel_ Dmel\\CG3348', description='CG3348 gene product from transcript CG3348-RA', other_designations='CG3348-PA|CG3348-PB', map_location='97F1-97F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27291535, end_position_on_the_genomic_accession=27292209, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50092, CurrentID=0, Status='live', Symbol='CheB98a', Aliases='Dmel_CG14531, CG14531p, Dmel\\CG14531, CheB98a', description='Chemosensory protein A 98a', other_designations='CG14531-PB|CheB98a-PB|CheB98ap|chemosensory protein A 98a', map_location='98E5-98E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28806702, end_position_on_the_genomic_accession=28807501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50096, CurrentID=0, Status='live', Symbol='Spase12', Aliases='Dmel_CG11500, CG11500, CG31042, Dmel\\CG11500, spase12', description='Spase 12-subunit', other_designations='CG11500-PA|Spase12-PA|spase 12-subunit', map_location='99B1-99B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29490544, end_position_on_the_genomic_accession=29491150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50101, CurrentID=0, Status='live', Symbol='CG18672', Aliases='Dmel_ Dmel\\CG18672', description='CG18672 gene product from transcript CG18672-RC', other_designations='CG18672-PB|CG18672-PC|CG18672-PD|uncharacterized protein', map_location='100B9-100B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31260247, end_position_on_the_genomic_accession=31261667, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50102, CurrentID=0, Status='live', Symbol='CG18673', Aliases='Dmel_ Dmel\\CG18673', description='CG18673 gene product from transcript CG18673-RC', other_designations='CG18673-PC|uncharacterized protein', map_location='100B9-100B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31262598, end_position_on_the_genomic_accession=31263878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50105, CurrentID=0, Status='live', Symbol='CG32428', Aliases='Dmel_ CG13249, CG5145, Dmel\\CG32428', description='CG32428 gene product from transcript CG32428-RC', other_designations='CG32428-PA|CG32428-PB|CG32428-PC|uncharacterized protein', map_location='77C4-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20469674, end_position_on_the_genomic_accession=20473723, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  50107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50107, CurrentID=0, Status='live', Symbol='CG4186', Aliases='Dmel_ Dmel\\CG4186', description='CG4186 gene product from transcript CG4186-RB', other_designations='CG4186-PA|CG4186-PB', map_location='77E6-77E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20769586, end_position_on_the_genomic_accession=20770138, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50109, CurrentID=0, Status='live', Symbol='CG13255', Aliases='Dmel_ Dmel\\CG13255', description='CG13255 gene product from transcript CG13255-RB', other_designations='CG13255-PA|CG13255-PB|uncharacterized protein', map_location='77F1-77F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20837488, end_position_on_the_genomic_accession=20838839, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50110, CurrentID=0, Status='live', Symbol='CG11458', Aliases='Dmel_ Dmel\\CG11458', description='CG11458 gene product from transcript CG11458-RA', other_designations='CG11458-PA|uncharacterized protein', map_location='77F4-77F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20905795, end_position_on_the_genomic_accession=20906331, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19836231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836231, CurrentID=0, Status='live', Symbol='CG44625', Aliases='Dmel_ Dmel\\CG44625', description='CG44625 gene product from transcript CG44625-RA', other_designations='CG44625-PA|uncharacterized protein', map_location='56F10-56F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20235545, end_position_on_the_genomic_accession=20235945, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50121, CurrentID=0, Status='live', Symbol='CG11666', Aliases='Dmel_ Dmel\\CG11666', description='CG11666 gene product from transcript CG11666-RA', other_designations='CG11666-PA|uncharacterized protein', map_location='19E6-19E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20984686, end_position_on_the_genomic_accession=20985246, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50122, CurrentID=0, Status='live', Symbol='CG15461', Aliases='Dmel_ Dmel\\CG15461', description='CG15461 gene product from transcript CG15461-RA', other_designations='CG15461-PA|uncharacterized protein', map_location='19D2-19D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20486180, end_position_on_the_genomic_accession=20486588, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50123, CurrentID=0, Status='live', Symbol='CG15456', Aliases='Dmel_ Dmel\\CG15456', description='CG15456 gene product from transcript CG15456-RB', other_designations='CG15456-PA|CG15456-PB|uncharacterized protein', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20426431, end_position_on_the_genomic_accession=20427042, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50124, CurrentID=0, Status='live', Symbol='CG15458', Aliases='Dmel_ Dmel\\CG15458', description='CG15458 gene product from transcript CG15458-RA', other_designations='CG15458-PA|uncharacterized protein', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20413652, end_position_on_the_genomic_accession=20414199, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50126, CurrentID=0, Status='live', Symbol='IM4', Aliases='Dmel_CG15231, BcDNA:RH25931, CG15231, DIM 4, DIM-4, DIM4, Dmel\\CG15231', description='Immune induced molecule 4', other_designations='CG15231-PA|CG15231-PB|IM4-PA|IM4-PB|drosophila immune-induced molecule 4|immune induced molecule 4|immune induced peptide 4', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20868496, end_position_on_the_genomic_accession=20869321, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50129, CurrentID=0, Status='live', Symbol='CG30222', Aliases='Dmel_ CG12709, CG15672, Dmel\\CG30222', description='CG30222 gene product from transcript CG30222-RF', other_designations='CG30222-PE|CG30222-PF|CG30222-PG|uncharacterized protein', map_location='57F2-57F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21583492, end_position_on_the_genomic_accession=21585368, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50133, CurrentID=0, Status='live', Symbol='CG13551', Aliases='Dmel_ BcDNA:RE33866, Dmel\\CG13551, SK[[66]]', description='CG13551 gene product from transcript CG13551-RA', other_designations='CG13551-PA|CG13551-PB|uncharacterized protein', map_location='59D9-59D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23377866, end_position_on_the_genomic_accession=23383424, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  50139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50139, CurrentID=0, Status='live', Symbol='CG12848', Aliases='Dmel_ Dmel\\CG12848', description='CG12848 gene product from transcript CG12848-RA', other_designations='CG12848-PA|CG12848-PB|uncharacterized protein', map_location='60E1-60E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24762566, end_position_on_the_genomic_accession=24763094, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50143, CurrentID=0, Status='live', Symbol='e(y)2b', Aliases='Dmel_CG14612, CG14612, Dmel\\CG14612', description='enhancer of yellow 2b', other_designations='CG14612-PA|CG14612-PB|e(y)2b-PA|e(y)2b-PB', map_location='84B6-84B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7093935, end_position_on_the_genomic_accession=7094727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50147, CurrentID=0, Status='live', Symbol='CG9445', Aliases='Dmel_ Dmel\\CG9445', description='CG9445 gene product from transcript CG9445-RA', other_designations='CG9445-PA|CG9445-PB', map_location='42D6-42D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6838074, end_position_on_the_genomic_accession=6839055, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50149, CurrentID=0, Status='live', Symbol='CG30158', Aliases='Dmel_ BcDNA:RE28276, CG15243, CG15244, CG9464, Dmel\\CG30158', description='CG30158 gene product from transcript CG30158-RA', other_designations='CG30158-PA|uncharacterized protein', map_location='42D6-42E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6891760, end_position_on_the_genomic_accession=6922613, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  50152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50152, CurrentID=0, Status='live', Symbol='CG11373', Aliases='Dmel_ Dmel\\CG11373', description='CG11373 gene product from transcript CG11373-RA', other_designations='CG11373-PA|uncharacterized protein', map_location='83D1-83D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5960294, end_position_on_the_genomic_accession=5960987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50157, CurrentID=0, Status='live', Symbol='CG12589', Aliases='Dmel_ Dmel\\CG12589', description='CG12589 gene product from transcript CG12589-RC', other_designations='CG12589-PA|CG12589-PB|CG12589-PC|uncharacterized protein', map_location='82F1-82F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5137057, end_position_on_the_genomic_accession=5138989, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50160, CurrentID=0, Status='live', Symbol='CG14645', Aliases='Dmel_ Dmel\\CG14645', description='CG14645 gene product from transcript CG14645-RA', other_designations='CG14645-PA|uncharacterized protein', map_location='82A4-82A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4335098, end_position_on_the_genomic_accession=4335506, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50161, CurrentID=0, Status='live', Symbol='CG12483', Aliases='Dmel_ BcDNA:RH38463, Dmel\\CG12483', description='CG12483 gene product from transcript CG12483-RA', other_designations='CG12483-PA|uncharacterized protein', map_location='61A6-61A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=97628, end_position_on_the_genomic_accession=98375, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50167, CurrentID=0, Status='live', Symbol='CG14974', Aliases='Dmel_ Dmel\\CG14974', description='CG14974 gene product from transcript CG14974-RC', other_designations='CG14974-PA|CG14974-PC|uncharacterized protein', map_location='63E5-63E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3620433, end_position_on_the_genomic_accession=3621032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50169, CurrentID=0, Status='live', Symbol='CG18675', Aliases='Dmel_ Dmel\\CG18675', description='CG18675 gene product from transcript CG18675-RA', other_designations='CG18675-PA|uncharacterized protein', map_location='64A9-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4173884, end_position_on_the_genomic_accession=4192968, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50170, CurrentID=0, Status='live', Symbol='Teh3', Aliases='Dmel_CG18676, CG18676, Dmel\\CG18676, TEH3', description='tipE homolog 3', other_designations='CG18676-PA|Teh3-PA', map_location='64A10-64A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4182645, end_position_on_the_genomic_accession=4184663, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50172, CurrentID=0, Status='live', Symbol='CG15024', Aliases='Dmel_ Dmel\\CG15024', description='CG15024 gene product from transcript CG15024-RB', other_designations='CG15024-PB|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4462974, end_position_on_the_genomic_accession=4463823, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50178, CurrentID=0, Status='live', Symbol='ND-B8', Aliases='Dmel_CG15434, CG15434, Dmel\\CG15434, NI8M', description='NADH dehydrogenase (ubiquinone) B8 subunit', other_designations='CG15434-PA|NADH ubiquinone oxidoreductase B8|NADH-ubiquinone oxidoreductase B8 subunit|NADH:ubiquinone oxidoreductase B8 subunit|ND-B8-PA', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4454626, end_position_on_the_genomic_accession=4455207, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50188, CurrentID=0, Status='live', Symbol='CG15386', Aliases='Dmel_ Dmel\\CG15386', description='CG15386 gene product from transcript CG15386-RA', other_designations='CG15386-PA|uncharacterized protein', map_location='22D4-22D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2216503, end_position_on_the_genomic_accession=2217246, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50190, CurrentID=0, Status='live', Symbol='Nplp4', Aliases='Dmel_CG15361, CG15361, Dmel\\CG15361, NPLP4, YSY', description='Neuropeptide-like precursor 4', other_designations='CG15361-PA|CG15361-PB|Nplp4-PA|Nplp4-PB|YSY peptide|neuropeptide-like precursor 4', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2008414, end_position_on_the_genomic_accession=2009072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50191, CurrentID=0, Status='live', Symbol='CG15353', Aliases='Dmel_ BcDNA:RH63135, Dmel\\CG15353', description='CG15353 gene product from transcript CG15353-RA', other_designations='CG15353-PA|uncharacterized protein', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2006763, end_position_on_the_genomic_accession=2007193, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50192, CurrentID=0, Status='live', Symbol='CG15357', Aliases='Dmel_ Dmel\\CG15357', description='CG15357 gene product from transcript CG15357-RB', other_designations='CG15357-PA|CG15357-PB|uncharacterized protein', map_location='22B6-22B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1945150, end_position_on_the_genomic_accession=1945709, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50196, CurrentID=0, Status='live', Symbol='CG5011', Aliases='Dmel_ Dmel\\CG5011', description='CG5011 gene product from transcript CG5011-RB', other_designations='CG5011-PA|CG5011-PB', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1200671, end_position_on_the_genomic_accession=1201220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50198, CurrentID=0, Status='live', Symbol='CG13946', Aliases='Dmel_ Dmel\\CG13946', description='CG13946 gene product from transcript CG13946-RA', other_designations='CG13946-PA|uncharacterized protein', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=776536, end_position_on_the_genomic_accession=776941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50202, CurrentID=0, Status='live', Symbol='CG15126', Aliases='Dmel_ Dmel\\CG15126', description='CG15126 gene product from transcript CG15126-RA', other_designations='CG15126-PA|uncharacterized protein', map_location='56E2-56E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19690798, end_position_on_the_genomic_accession=19691078, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50203, CurrentID=0, Status='live', Symbol='CG15127', Aliases='Dmel_ CG18365, Dmel\\CG15127', description='CG15127 gene product from transcript CG15127-RA', other_designations='CG15127-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19680768, end_position_on_the_genomic_accession=19681391, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50205, CurrentID=0, Status='live', Symbol='CG16926', Aliases='Dmel_ 141315_at, Dmel\\CG16926', description='CG16926 gene product from transcript CG16926-RB', other_designations='CG16926-PA|CG16926-PB|uncharacterized protein', map_location='56D14-56D14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19500295, end_position_on_the_genomic_accession=19501418, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50206, CurrentID=0, Status='live', Symbol='CG15068', Aliases='Dmel_ CT34939, Dmel\\CG15068', description='CG15068 gene product from transcript CG15068-RA', other_designations='CG15068-PA|uncharacterized protein', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18390353, end_position_on_the_genomic_accession=18390629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50207, CurrentID=0, Status='live', Symbol='CG15065', Aliases='Dmel_ CT34936, Dmel\\CG15065, IM2-like', description='CG15065 gene product from transcript CG15065-RA', other_designations='CG15065-PA|uncharacterized protein', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18389959, end_position_on_the_genomic_accession=18390242, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50208, CurrentID=0, Status='live', Symbol='CG16836', Aliases='Dmel_ Dmel\\CG16836', description='CG16836 gene product from transcript CG16836-RA', other_designations='CG16836-PA|uncharacterized protein', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18388995, end_position_on_the_genomic_accession=18389443, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50209, CurrentID=0, Status='live', Symbol='IM3', Aliases='Dmel_CG16844, BcDNA:RH58911, CG16844, CT37419, DIM 3, Dmel\\CG16844, im3', description='Immune induced molecule 3', other_designations='CG16844-PA|CG16844-PB|IM3-PA|IM3-PB|immune induced molecule 3', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18387895, end_position_on_the_genomic_accession=18388734, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50210, CurrentID=0, Status='live', Symbol='Tango8', Aliases='Dmel_CG14503, CG14503, Dmel\\CG14503, TANGO8', description='Transport and Golgi organization 8', other_designations='CG14503-PA|Tango8-PA|transport and golgi organization 8', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18311822, end_position_on_the_genomic_accession=18311998, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50216, CurrentID=0, Status='live', Symbol='CG15919', Aliases='Dmel_ Dmel\\CG15919', description='CG15919 gene product from transcript CG15919-RA', other_designations='CG15919-PA|uncharacterized protein', map_location='53D9-53D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16792642, end_position_on_the_genomic_accession=16793179, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50220, CurrentID=0, Status='live', Symbol='CG12853', Aliases='Dmel_ Dmel\\CG12853, NEST:bs26d12, bs26d12.y1', description='CG12853 gene product from transcript CG12853-RA', other_designations='CG12853-PA|uncharacterized protein', map_location='51D3-51D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14933515, end_position_on_the_genomic_accession=14933913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50224, CurrentID=0, Status='live', Symbol='CG13018', Aliases='Dmel_ Dmel\\CG13018', description='CG13018 gene product from transcript CG13018-RA', other_designations='CG13018-PA|uncharacterized protein', map_location='50E1-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14157050, end_position_on_the_genomic_accession=14157488, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50225, CurrentID=0, Status='live', Symbol='Prosap', Aliases='Dmel_CG30483, BcDNA:GH23107, CG13354, CG30483, CG8122, D-ProSAP, Dmel\\CG30483, ProSAP, Shank, anon-WO0118547.149', description='CG30483 gene product from transcript CG30483-RD', other_designations='CG30483-PA|CG30483-PC|CG30483-PD|CG30483-PE|Prosap-PA|Prosap-PC|Prosap-PD|Prosap-PE|prosap|shank', map_location='50D3-50E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14060456, end_position_on_the_genomic_accession=14140902, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  50227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50227, CurrentID=0, Status='live', Symbol='CG17059', Aliases='Dmel_ BcDNA:GH04581, Dmel\\CG17059', description='CG17059 gene product from transcript CG17059-RA', other_designations='CG17059-PA|uncharacterized protein', map_location='49F10-49F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13220283, end_position_on_the_genomic_accession=13221015, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50228, CurrentID=0, Status='live', Symbol='CG17580', Aliases='Dmel_ BcDNA:AT18054, Dmel\\CG17580', description='CG17580 gene product from transcript CG17580-RA', other_designations='CG17580-PA|uncharacterized protein', map_location='49D3-49D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12806614, end_position_on_the_genomic_accession=12807811, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26067044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067044, CurrentID=0, Status='live', Symbol='CG45766', Aliases='Dmel_ Dmel\\CG45766', description='CG45766 gene product from transcript CG45766-RA', other_designations='CG45766-PA|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3301827, end_position_on_the_genomic_accession=3302701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50236, CurrentID=0, Status='live', Symbol='CG18336', Aliases='Dmel_ CG13212, Dmel\\CG18336', description='CG18336 gene product from transcript CG18336-RA', other_designations='CG18336-PA|CG18336-PB|uncharacterized protein', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11293776, end_position_on_the_genomic_accession=11294586, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50237, CurrentID=0, Status='live', Symbol='CG13230', Aliases='Dmel_ Dmel\\CG13230', description='CG13230 gene product from transcript CG13230-RA', other_designations='CG13230-PA|uncharacterized protein', map_location='47D5-47D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11154896, end_position_on_the_genomic_accession=11155293, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50246, CurrentID=0, Status='live', Symbol='COX7C', Aliases='Dmel_CG2249, BcDNA:LD14731, CG2249, COXO, CoAVIIc, CoVIIc, Dmel\\CG2249, VIIc, anon-EST:fe3A1', description='Cytochrome c oxidase subunit 7C', other_designations='CG2249-PA|CG2249-PB|CG2249-PC|CG2249-PD|COX7C-PA|COX7C-PB|COX7C-PC|COX7C-PD|anon-fast-evolving-3A1|cytochrome c oxidase VIIc|cytochrome c oxidase polypeptide VIIc|cytochrome c oxidase subunit 7C|cytochrome c oxidase subunit VIIc|subunit VIIc', map_location='46D8-46D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10048353, end_position_on_the_genomic_accession=10050066, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50250, CurrentID=0, Status='live', Symbol='CG14767', Aliases='Dmel_ CG8575, Dmel\\CG14767, dmCG14767', description='CG14767 gene product from transcript CG14767-RA', other_designations='CG14767-PA|CG14767-PB|CG14767-PC|uncharacterized protein', map_location='44E2-44E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8718017, end_position_on_the_genomic_accession=8723839, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  50251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50251, CurrentID=0, Status='live', Symbol='CG17977', Aliases='Dmel_ Dmel\\CG17977', description='CG17977 gene product from transcript CG17977-RA', other_designations='CG17977-PA|uncharacterized protein', map_location='44A4-44A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8066274, end_position_on_the_genomic_accession=8067655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50252, CurrentID=0, Status='live', Symbol='CG30377', Aliases='Dmel_ CG14595, CG1918, Dmel\\CG30377', description='CG30377 gene product from transcript CG30377-RB', other_designations='CG30377-PA|CG30377-PB|uncharacterized protein', map_location='43F1-43F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7890025, end_position_on_the_genomic_accession=7915032, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50253, CurrentID=0, Status='live', Symbol='Atg10', Aliases='Dmel_CG12821, ATG10, CG12821, Dmel\\CG12821', description='Autophagy-related 10', other_designations='Atg10-PB|Atg10-PD|Autophagy-specific gene 10|CG12821-PB|CG12821-PD', map_location='43E18-43E18', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7832171, end_position_on_the_genomic_accession=7835518, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50257, CurrentID=0, Status='live', Symbol='karr', Aliases='Dmel_CG15323, CG15323, Dmel\\CG15323', description='CG15323 gene product from transcript CG15323-RA', other_designations='CG15323-PA|karr-PA', map_location='19B1-19B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20110110, end_position_on_the_genomic_accession=20110575, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50259, CurrentID=0, Status='live', Symbol='CG14104', Aliases='Dmel_ Dmel\\CG14104', description='CG14104 gene product from transcript CG14104-RB', other_designations='CG14104-PB|uncharacterized protein', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19806811, end_position_on_the_genomic_accession=19807199, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50263, CurrentID=0, Status='live', Symbol='geko', Aliases='Dmel_CG13695, BcDNA:RE30284, CG13695, CG13696, CT33153, Dmel\\CG13695, Geko, gk', description='CG13695 gene product from transcript CG13695-RB', other_designations='CG13695-PB|geko-PB', map_location='75B11-75B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18139402, end_position_on_the_genomic_accession=18145418, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50266, CurrentID=0, Status='live', Symbol='CG7630', Aliases='Dmel_ BcDNA:RH19679, Dmel\\CG7630, anon-EST:Posey153, anon-EST:Posey228', description='CG7630 gene product from transcript CG7630-RA', other_designations='CG7630-PA', map_location='74B3-74B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17425477, end_position_on_the_genomic_accession=17426282, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50267, CurrentID=0, Status='live', Symbol='CG13056', Aliases='Dmel_ Dmel\\CG13056', description='CG13056 gene product from transcript CG13056-RA', other_designations='CG13056-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16333313, end_position_on_the_genomic_accession=16333782, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50268, CurrentID=0, Status='live', Symbol='CG13038', Aliases='Dmel_ Dmel\\CG13038', description='CG13038 gene product from transcript CG13038-RA', other_designations='CG13038-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16327258, end_position_on_the_genomic_accession=16327893, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50270, CurrentID=0, Status='live', Symbol='CG13066', Aliases='Dmel_ Dmel\\CG13066', description='CG13066 gene product from transcript CG13066-RB', other_designations='CG13066-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16272909, end_position_on_the_genomic_accession=16273608, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50271, CurrentID=0, Status='live', Symbol='CG13069', Aliases='Dmel_ Dmel\\CG13069', description='CG13069 gene product from transcript CG13069-RA', other_designations='CG13069-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16265385, end_position_on_the_genomic_accession=16265804, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50274, CurrentID=0, Status='live', Symbol='CG13053', Aliases='Dmel_ Dmel\\CG13053', description='CG13053 gene product from transcript CG13053-RB', other_designations='CG13053-PA|CG13053-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16259037, end_position_on_the_genomic_accession=16259722, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50281, CurrentID=0, Status='live', Symbol='BobA', Aliases='Dmel_CG12487, Bob, Bob-A, CG12487, Dmel\\CG12487, bobA', description='Brother of Bearded A', other_designations='BobA-PA|CG12487-PA|brother of bearded|brother of bearded A', map_location='71A2-71A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14945499, end_position_on_the_genomic_accession=14946125, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50282, CurrentID=0, Status='live', Symbol='CG13465', Aliases='Dmel_ Dmel\\CG13465', description='CG13465 gene product from transcript CG13465-RA', other_designations='CG13465-PA|uncharacterized protein', map_location='71A2-71A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14943730, end_position_on_the_genomic_accession=14944331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8673976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673976, CurrentID=0, Status='live', Symbol='CG42397', Aliases='Dmel_ Dmel\\CG42397', description='CG42397 gene product from transcript CG42397-RA', other_designations='CG42397-PA|uncharacterized protein', map_location='68E4-68E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11973085, end_position_on_the_genomic_accession=11973790, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50285, CurrentID=0, Status='live', Symbol='CG8750', Aliases='Dmel_ Dmel\\CG8750', description='CG8750 gene product from transcript CG8750-RB', other_designations='CG8750-PA|CG8750-PB', map_location='70C4-70C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13842377, end_position_on_the_genomic_accession=13843017, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50286, CurrentID=0, Status='live', Symbol='Nplp2', Aliases='Dmel_CG11051, BcDNA:RH08410, CG11051, Dmel\\CG11051, NEF, NPLP 2, NPLP2, nplp2', description='Neuropeptide-like precursor 2', other_designations='CG11051-PA|CG11051-PB|CG11051-PC|NEF peptide|Nplp2-PA|Nplp2-PB|Nplp2-PC|neuropeptide-like precursor 2', map_location='70A6-70A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13356376, end_position_on_the_genomic_accession=13357223, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50287, CurrentID=0, Status='live', Symbol='CG14113', Aliases='Dmel_ Dmel\\CG14113', description='CG14113 gene product from transcript CG14113-RA', other_designations='CG14113-PA|uncharacterized protein', map_location='70A5-70A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13329135, end_position_on_the_genomic_accession=13329699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50289, CurrentID=0, Status='live', Symbol='CG12521', Aliases='Dmel_ Dmel\\CG12521', description='CG12521 gene product from transcript CG12521-RA', other_designations='CG12521-PA|uncharacterized protein', map_location='68E2-68E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11847383, end_position_on_the_genomic_accession=11847803, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50290, CurrentID=0, Status='live', Symbol='CG14132', Aliases='Dmel_ DIGF-7, Dmel\\CG14132', description='CG14132 gene product from transcript CG14132-RB', other_designations='CG14132-PB|uncharacterized protein', map_location='68D3-68D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11712229, end_position_on_the_genomic_accession=11715285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50296, CurrentID=0, Status='live', Symbol='dpr6', Aliases='Dmel_CG14162, BcDNA:LD13525, CG14162, Dmel\\CG14162, Dpr-6, Dpr6', description='defective proboscis extension response 6', other_designations='CG14162-PA|CG14162-PC|CG14162-PD|CG14162-PE|CG14162-PF|defective proboscis response 6|dpr6-PA|dpr6-PC|dpr6-PD|dpr6-PE|dpr6-PF', map_location='67D2-67D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9974651, end_position_on_the_genomic_accession=10139360, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  50300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50300, CurrentID=0, Status='live', Symbol='CG13315', Aliases='Dmel_ BcDNA:RH39779, Dmel\\CG13315', description='CG13315 gene product from transcript CG13315-RA', other_designations='CG13315-PA|CG13315-PB|uncharacterized protein', map_location='66F5-66F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9096877, end_position_on_the_genomic_accession=9097494, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50301, CurrentID=0, Status='live', Symbol='CG13306', Aliases='Dmel_ Dmel\\CG13306', description='CG13306 gene product from transcript CG13306-RA', other_designations='CG13306-PA|uncharacterized protein', map_location='66D12-66D12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8749139, end_position_on_the_genomic_accession=8749726, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50302, CurrentID=0, Status='live', Symbol='CG32354', Aliases='Dmel_ BEST:CK01670, CG17355, CG7027, CG7159, CK01670, Dmel\\CG32354', description='CG32354 gene product from transcript CG32354-RA', other_designations='CG32354-PA|uncharacterized protein', map_location='66C11-66C12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8415857, end_position_on_the_genomic_accession=8422205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  50305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50305, CurrentID=0, Status='live', Symbol='CG8012', Aliases='Dmel_ BcDNA:GH26995, Dmel\\CG8012', description='CG8012 gene product from transcript CG8012-RB', other_designations='CG8012-PA|CG8012-PB', map_location='66C1-66C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8215280, end_position_on_the_genomic_accession=8216198, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50310, CurrentID=0, Status='live', Symbol='CG8620', Aliases='Dmel_ Dmel\\CG8620', description='CG8620 gene product from transcript CG8620-RA', other_designations='CG8620-PA|CG8620-PB', map_location='65E6-65E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7158099, end_position_on_the_genomic_accession=7158874, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673977, CurrentID=0, Status='live', Symbol='CG42586', Aliases='Dmel_ Dmel\\CG42586', description='CG42586 gene product from transcript CG42586-RB', other_designations='CG42586-PA|CG42586-PB|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14741161, end_position_on_the_genomic_accession=14741840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50315, CurrentID=0, Status='live', Symbol='CG15212', Aliases='Dmel_ BcDNA:RE28658, Dmel\\CG15212', description='CG15212 gene product from transcript CG15212-RA', other_designations='CG15212-PA|uncharacterized protein', map_location='64C6-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5131545, end_position_on_the_genomic_accession=5132127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50316, CurrentID=0, Status='live', Symbol='CG15213', Aliases='Dmel_ Dmel\\CG15213', description='CG15213 gene product from transcript CG15213-RA', other_designations='CG15213-PA|uncharacterized protein', map_location='64C7-64C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5132796, end_position_on_the_genomic_accession=5133220, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50320, CurrentID=0, Status='live', Symbol='dpr12', Aliases='Dmel_CG34385, CG14469, CG34385, CG7796, CT34165, Dmel\\CG34385, Dmel_CG14469, Dmel_CG7796, Dpr-12, Dpr12', description='defective proboscis extension response 12', other_designations='CG34385-PC|CG34385-PD|CG34385-PE|dpr12-PC|dpr12-PD|dpr12-PE', map_location='42A1-42A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5804712, end_position_on_the_genomic_accession=5860831, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  50323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50323, CurrentID=0, Status='live', Symbol='CG15210', Aliases='Dmel_ Dmel\\CG15210', description='CG15210 gene product from transcript CG15210-RA', other_designations='CG15210-PA|uncharacterized protein', map_location='9F4-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10839230, end_position_on_the_genomic_accession=10839895, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  35695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=35695, CurrentID=0, Status='live', Symbol='CG1360', Aliases='Dmel_ Dmel\\CG1360', description='CG1360 gene product from transcript CG1360-RA', other_designations='CG1360-PA', map_location='43E3-43E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7587773, end_position_on_the_genomic_accession=7590109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318943, CurrentID=0, Status='live', Symbol='CG31789', Aliases='Dmel_ Dmel\\CG31789', description='CG31789 gene product from transcript CG31789-RC', other_designations='CG31789-PA|CG31789-PB|CG31789-PC|uncharacterized protein', map_location='36F7-36F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18605928, end_position_on_the_genomic_accession=18606684, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50344, CurrentID=0, Status='live', Symbol='CG12479', Aliases='Dmel_ Dmel\\CG12479, Minos1a', description='CG12479 gene product from transcript CG12479-RA', other_designations='CG12479-PA|mitochondrial INner membrane organizing system 1|uncharacterized protein', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14151579, end_position_on_the_genomic_accession=14152006, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31323, CurrentID=0, Status='live', Symbol='CG13021', Aliases='Dmel_ BcDNA:AT30881, Dmel\\CG13021', description='CG13021 gene product from transcript CG13021-RA', other_designations='CG13021-PA|CG13021-PB|uncharacterized protein', map_location='3D5-3D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3514316, end_position_on_the_genomic_accession=3515757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50350, CurrentID=0, Status='live', Symbol='CG12994', Aliases='Dmel_ Dmel\\CG12994', description='CG12994 gene product from transcript CG12994-RA', other_designations='CG12994-PA|uncharacterized protein', map_location='16A5-16A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17422320, end_position_on_the_genomic_accession=17443994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50356, CurrentID=0, Status='live', Symbol='CG34401', Aliases='Dmel_ CG15037, CG17757, CG32542, CG7307, CG7317, Dmel\\CG34401, Dmel_CG17757, Dmel_CG32542', description='CG34401 gene product from transcript CG34401-RB', other_designations='CG34401-PA|CG34401-PB|uncharacterized protein', map_location='17F2-17F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18863950, end_position_on_the_genomic_accession=18884075, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  50358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50358, CurrentID=0, Status='live', Symbol='Diedel3', Aliases='Dmel_CG34329, CG14189, CG34329, Diedel-3, Dmel\\CG34329', description='Diedel 3', other_designations='CG34329-PA|CG34329-PB|Diedel-3|Diedel3-PA|Diedel3-PB|diedel 3', map_location='18A2-18A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19050442, end_position_on_the_genomic_accession=19053160, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50363, CurrentID=0, Status='live', Symbol='ksh', Aliases='Dmel_CG14199, CG14199, Dmel\\CG14199', description='kish', other_designations='CG14199-PB|ksh-PB', map_location='18C8-18C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19494420, end_position_on_the_genomic_accession=19494992, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50364, CurrentID=0, Status='live', Symbol='CG13364', Aliases='Dmel_ CG18451, Dmel\\CG13364, EG:115C2.12, anon-EST:Posey36', description='CG13364 gene product from transcript CG13364-RA', other_designations='CG13364-PA|uncharacterized protein', map_location='1B12-1B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=624757, end_position_on_the_genomic_accession=625203, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50365, CurrentID=0, Status='live', Symbol='CG3713', Aliases='Dmel_ BcDNA:GH13231, Dmel\\CG3713, EG:BACR7A4.2', description='CG3713 gene product from transcript CG3713-RA', other_designations='CG3713-PA|CG3713-PB', map_location='1D1-1D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=846171, end_position_on_the_genomic_accession=847157, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50366, CurrentID=0, Status='live', Symbol='CG14634', Aliases='Dmel_ Dmel\\CG14634', description='CG14634 gene product from transcript CG14634-RB', other_designations='CG14634-PB|uncharacterized protein', map_location='1D1-1D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=862608, end_position_on_the_genomic_accession=863200, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50373, CurrentID=0, Status='live', Symbol='CG17776', Aliases='Dmel_ Dmel\\CG17776', description='CG17776 gene product from transcript CG17776-RB', other_designations='CG17776-PA|CG17776-PB|uncharacterized protein', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2184725, end_position_on_the_genomic_accession=2185258, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50374, CurrentID=0, Status='live', Symbol='CG17777', Aliases='Dmel_ Dmel\\CG17777', description='CG17777 gene product from transcript CG17777-RC', other_designations='CG17777-PB|CG17777-PC|uncharacterized protein', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2186855, end_position_on_the_genomic_accession=2187741, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50378, CurrentID=0, Status='live', Symbol='CG15577', Aliases='Dmel_ Dmel\\CG15577', description='CG15577 gene product from transcript CG15577-RA', other_designations='CG15577-PA|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4276493, end_position_on_the_genomic_accession=4277000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50379, CurrentID=0, Status='live', Symbol='CG15578', Aliases='Dmel_ Dmel\\CG15578', description='CG15578 gene product from transcript CG15578-RA', other_designations='CG15578-PA|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4277547, end_position_on_the_genomic_accession=4277961, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50380, CurrentID=0, Status='live', Symbol='CG15579', Aliases='Dmel_ Dmel\\CG15579', description='CG15579 gene product from transcript CG15579-RB', other_designations='CG15579-PA|CG15579-PB|uncharacterized protein', map_location='4B6-4B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4311279, end_position_on_the_genomic_accession=4311799, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50381, CurrentID=0, Status='live', Symbol='mRpL33', Aliases='Dmel_CG3712, BcDNA:GH15938, CG3712, Dmel\\CG3712, L33', description='mitochondrial ribosomal protein L33', other_designations='CG3712-PA|mRpL33-PA', map_location='4B6-4B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4321764, end_position_on_the_genomic_accession=4322231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50391, CurrentID=0, Status='live', Symbol='Grip', Aliases='Dmel_CG14447, CG14447, CG5980, DGrip, Dgrip, Dmel\\CG14447, GRIP, dGRIP, dgrip, grip', description='Glutamate receptor binding protein', other_designations='CG14447-PA|CG14447-PB|CG14447-PC|Glutamate-receptor-interacting-protein|Grip-PA|Grip-PB|Grip-PC|glutamate receptor binding protein|glutamate receptor interacting protein', map_location='5D1-5D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5966279, end_position_on_the_genomic_accession=5986699, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  50392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50392, CurrentID=0, Status='live', Symbol='schlank', Aliases='Dmel_CG3576, CG15898, CG3576, DLAG1, DLag1, Dlag1, Dmel\\CG3576, Lag1, l(1)G0061, l(1)G0365, l(1)G0489, lag1, lon', description='CG3576 gene product from transcript CG3576-RA', other_designations='CG3576-PA|CG3576-PB|Psc|longevity|longevity assurance gene 1|schlank-PA|schlank-PB', map_location='5E3-5E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6257668, end_position_on_the_genomic_accession=6267228, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  50395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50395, CurrentID=0, Status='live', Symbol='CG32736', Aliases='Dmel_ BcDNA:RE60462, CG14429, Dmel\\CG32736', description='CG32736 gene product from transcript CG32736-RB', other_designations='CG32736-PA|CG32736-PB|uncharacterized protein', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7011357, end_position_on_the_genomic_accession=7012137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50396, CurrentID=0, Status='live', Symbol='CG15036', Aliases='Dmel_ Dmel\\CG15036', description='CG15036 gene product from transcript CG15036-RA', other_designations='CG15036-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7424527, end_position_on_the_genomic_accession=7424842, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50397, CurrentID=0, Status='live', Symbol='CG11368', Aliases='Dmel_ BcDNA:RH67809, Dmel\\CG11368', description='CG11368 gene product from transcript CG11368-RA', other_designations='CG11368-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7452209, end_position_on_the_genomic_accession=7452643, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50403, CurrentID=0, Status='live', Symbol='CG12659', Aliases='Dmel_ Dmel\\CG12659', description='CG12659 gene product from transcript CG12659-RC', other_designations='CG12659-PC|uncharacterized protein', map_location='7F7-7F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8590775, end_position_on_the_genomic_accession=8591729, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50405, CurrentID=0, Status='live', Symbol='CG9034', Aliases='Dmel_ BcDNA:RH45008, Dmel\\CG9034, GH20078', description='CG9034 gene product from transcript CG9034-RB', other_designations='CG9034-PA|CG9034-PB', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9159514, end_position_on_the_genomic_accession=9160052, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50415, CurrentID=0, Status='live', Symbol='CG15308', Aliases='Dmel_ Dmel\\CG15308', description='CG15308 gene product from transcript CG15308-RB', other_designations='CG15308-PB|CG15308-PC|uncharacterized protein', map_location='9B4-9B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10217913, end_position_on_the_genomic_accession=10219139, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50416, CurrentID=0, Status='live', Symbol='CG12643', Aliases='Dmel_ Dmel\\CG12643', description='CG12643 gene product from transcript CG12643-RA', other_designations='CG12643-PA|uncharacterized protein', map_location='9B5-9B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10264233, end_position_on_the_genomic_accession=10265263, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50417 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50417, CurrentID=0, Status='live', Symbol='Neb-cGP', Aliases='Dmel_CG15304, BcDNA:RE31692, BcDNA:RH45340, CG15304, Dmel\\CG15304, pncr006:X', description='CG15304 gene product from transcript CG15304-RB', other_designations='CG15304-PA|CG15304-PB|Neb-cGP-PA|Neb-cGP-PB|putative noncoding RNA 006:X', map_location='9C4-9C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10460736, end_position_on_the_genomic_accession=10461878, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50423, CurrentID=0, Status='live', Symbol='CG13998', Aliases='Dmel_ Dmel\\CG13998', description='CG13998 gene product from transcript CG13998-RA', other_designations='CG13998-PA|uncharacterized protein', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5956489, end_position_on_the_genomic_accession=5956819, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50424, CurrentID=0, Status='live', Symbol='Muc26B', Aliases='Dmel_CG13990, BcDNA:RE30702, CG13990, Dmel\\CG13990', description='Mucin 26B', other_designations='CG13990-PA|Muc26B-PA|mucin 26B', map_location='26B8-26B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6151969, end_position_on_the_genomic_accession=6153701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50428, CurrentID=0, Status='live', Symbol='Sem1', Aliases='Dmel_CG13779, BcDNA:RH34416, CG13779, Dmel\\CG13779p, dmdss1, dss1, Sem1', description='Sem1 ortholog (S. cerevisiae)', other_designations='CG13779-PA|Sem1 ortholog|Sem1-PA', map_location='27D7-27D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7063986, end_position_on_the_genomic_accession=7064467, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50432, CurrentID=0, Status='live', Symbol='Peritrophin-15b', Aliases='Dmel_CG31893, CG13395, CG31893, Dm15b, Dmel\\CG31893, Peritrophi-15b', description='CG31893 gene product from transcript CG31893-RC', other_designations='CG31893-PA|CG31893-PB|CG31893-PC|Dm-peritrophin-15b|Peritrophin-15b-PA|Peritrophin-15b-PB|Peritrophin-15b-PC', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8394446, end_position_on_the_genomic_accession=8394940, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50433, CurrentID=0, Status='live', Symbol='Peritrophin-15a', Aliases='Dmel_CG17814, CG17814, CT39472, Dm15a, Dmel\\CG17814', description='CG17814 gene product from transcript CG17814-RA', other_designations='CG17814-PA|Dm-peritrophin-15a|Peritrophin-15a-PA', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8395665, end_position_on_the_genomic_accession=8396096, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50434, CurrentID=0, Status='live', Symbol='CG32983', Aliases='Dmel_ CG18663, Dmel\\CG32983', description='CG32983 gene product from transcript CG32983-RA', other_designations='CG32983-PA|uncharacterized protein', map_location='29F3-29F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8865850, end_position_on_the_genomic_accession=8866670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50437, CurrentID=0, Status='live', Symbol='CG18662', Aliases='Dmel_ Dmel\\CG18662', description='CG18662 gene product from transcript CG18662-RA', other_designations='CG18662-PA|uncharacterized protein', map_location='29F6-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8980659, end_position_on_the_genomic_accession=8981432, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  50438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50438, CurrentID=0, Status='live', Symbol='CG18661', Aliases='Dmel_ BcDNA:SD17344, Dmel\\CG18661', description='CG18661 gene product from transcript CG18661-RA', other_designations='CG18661-PA|CG18661-PB|uncharacterized protein', map_location='29F7-29F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8997605, end_position_on_the_genomic_accession=8998544, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50441, CurrentID=0, Status='live', Symbol='CG18666', Aliases='Dmel_ Dmel\\CG18666', description='CG18666 gene product from transcript CG18666-RA', other_designations='CG18666-PA|uncharacterized protein', map_location='32D1-32D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11034795, end_position_on_the_genomic_accession=11035308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50442, CurrentID=0, Status='live', Symbol='CG14933', Aliases='Dmel_ Dmel\\CG14933', description='CG14933 gene product from transcript CG14933-RC', other_designations='CG14933-PA|CG14933-PC|uncharacterized protein', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11842828, end_position_on_the_genomic_accession=11844012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50446, CurrentID=0, Status='live', Symbol='CG16978', Aliases='Dmel_ CT35582, Dmel\\CG16978', description='CG16978 gene product from transcript CG16978-RB', other_designations='CG16978-PB|uncharacterized protein', map_location='34A6-34A6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13143214, end_position_on_the_genomic_accession=13143622, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50454, CurrentID=0, Status='live', Symbol='CG15283', Aliases='Dmel_ BG:DS08340.1, DS08340.1, Dmel\\CG15283', description='CG15283 gene product from transcript CG15283-RB', other_designations='CG15283-PB|uncharacterized protein', map_location='35B1-35B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14450282, end_position_on_the_genomic_accession=14450922, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50456, CurrentID=0, Status='live', Symbol='CG13245', Aliases='Dmel_ BEST:GH06096, BEST:GH09435, BEST:GH21951, Dmel\\CG13245, bs34a02.y1', description='CG13245 gene product from transcript CG13245-RB', other_designations='CG13245-PA|CG13245-PB|CG13245-PC|uncharacterized protein', map_location='35E1-35E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16004325, end_position_on_the_genomic_accession=16005461, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50457, CurrentID=0, Status='live', Symbol='CG34168', Aliases='Dmel_ CG4959, Dmel\\CG34168', description='CG34168 gene product from transcript CG34168-RA', other_designations='CG34168-PA|uncharacterized protein', map_location='35E3-35E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16134714, end_position_on_the_genomic_accession=16135272, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50458, CurrentID=0, Status='live', Symbol='CG4440', Aliases='Dmel_ BcDNA:RE16123, Dmel\\CG4440', description='CG4440 gene product from transcript CG4440-RB', other_designations='CG4440-PA|CG4440-PB', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16329925, end_position_on_the_genomic_accession=16330521, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50459, CurrentID=0, Status='live', Symbol='CG6115', Aliases='Dmel_ BcDNA:GH16758, Dmel\\CG6115', description='CG6115 gene product from transcript CG6115-RB', other_designations='CG6115-PA|CG6115-PB', map_location='36B3-36B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16887109, end_position_on_the_genomic_accession=16887966, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50463, CurrentID=0, Status='live', Symbol='CG5693', Aliases='Dmel_ Dmel\\CG5693', description='CG5693 gene product from transcript CG5693-RA', other_designations='CG5693-PA', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18006627, end_position_on_the_genomic_accession=18007216, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50466, CurrentID=0, Status='live', Symbol='CG10570', Aliases='Dmel_ Dmel\\CG10570, Dromel_CG10570_FBtr0081122_mORF', description='CG10570 gene product from transcript CG10570-RC', other_designations='CG10570-PA|CG10570-PB|CG10570-PC|uncharacterized protein', map_location='37A4-37A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18808694, end_position_on_the_genomic_accession=18811252, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50467, CurrentID=0, Status='live', Symbol='CG17325', Aliases='Dmel_ BcDNA:RH02271, Dmel\\CG17325, Dromel_CG17325_FBtr0081102_mORF', description='CG17325 gene product from transcript CG17325-RC', other_designations='CG17325-PA|CG17325-PB|CG17325-PC|uncharacterized protein', map_location='37A4-37A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18812648, end_position_on_the_genomic_accession=18813941, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50470, CurrentID=0, Status='live', Symbol='CG12617', Aliases='Dmel_ 38C.31, Dmel\\CG12617', description='CG12617 gene product from transcript CG12617-RC', other_designations='CG12617-PA|CG12617-PB|CG12617-PC|uncharacterized protein', map_location='38C4-38C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20294693, end_position_on_the_genomic_accession=20295804, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673983, CurrentID=0, Status='live', Symbol='Sfp53D', Aliases='Dmel_CG42477, CG42477, Dmel\\CG42477, HDC06796', description='Seminal fluid protein 53D', other_designations='CG42477-PA|CG42477-PB|Sfp53D-PA|Sfp53D-PB|seminal fluid protein 53D', map_location='53D5-53D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16744295, end_position_on_the_genomic_accession=16745114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318109, CurrentID=0, Status='live', Symbol='CG32603', Aliases='Dmel_ Dmel\\CG32603', description='CG32603 gene product from transcript CG32603-RA', other_designations='CG32603-PA|uncharacterized protein', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14184561, end_position_on_the_genomic_accession=14185807, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26067046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067046, CurrentID=0, Status='live', Symbol='Mst77Y-7', Aliases='Dmel_CG45768, CG45768, Dmel\\CG45768', description='Male-specific transcript 77Y 7', other_designations='CG45768-PA|Mst77Y-7-PA', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3327556, end_position_on_the_genomic_accession=3328279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673985, CurrentID=0, Status='live', Symbol='CG42538', Aliases='Dmel_ Dmel\\CG42538', description='CG42538 gene product from transcript CG42538-RA', other_designations='CG42538-PA|uncharacterized protein', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16034524, end_position_on_the_genomic_accession=16034853, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673987, CurrentID=0, Status='live', Symbol='Sfp96F', Aliases='Dmel_CG42609, CG42609, Dmel\\CG42609', description='Seminal fluid protein 96F', other_designations='CG42609-PA|Sfp96F-PA|seminal fluid protein 96F', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25830061, end_position_on_the_genomic_accession=25830228, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740465, CurrentID=0, Status='live', Symbol='CG34266', Aliases='Dmel_ Dmel\\CG34266', description='CG34266 gene product from transcript CG34266-RA', other_designations='CG34266-PA|uncharacterized protein', map_location='64B13-64B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4677832, end_position_on_the_genomic_accession=4678233, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8673988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673988, CurrentID=0, Status='live', Symbol='CG42455', Aliases='Dmel_ Dmel\\CG42455, Dromel_CG11989_FBtr0076348_ard1_uORF', description='CG42455 gene product from transcript CG42455-RC', other_designations='CG42455-PA|CG42455-PB|CG42455-PC|uncharacterized protein', map_location='67D2-67D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9972312, end_position_on_the_genomic_accession=9974023, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  41213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41213, CurrentID=0, Status='live', Symbol='CG9459', Aliases='Dmel_ CT26804, Dmel\\CG9459', description='CG9459 gene product from transcript CG9459-RA', other_designations='CG9459-PA', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9817428, end_position_on_the_genomic_accession=9818395, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673990, CurrentID=0, Status='live', Symbol='Sfp24Bc', Aliases='Dmel_CG42602, CG42602, Dmel\\CG42602, SFP24BC', description='Seminal fluid protein 24Bc', other_designations='CG42602-PA|Sfp24Bc-PA|seminal fluid protein 24Bc', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3670131, end_position_on_the_genomic_accession=3670581, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19836253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19836253, CurrentID=0, Status='live', Symbol='CG44341', Aliases='Dmel_ Dmel\\CG44341', description='CG44341 gene product from transcript CG44341-RA', other_designations='CG44341-PA|uncharacterized protein', map_location='49F2-49F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13103895, end_position_on_the_genomic_accession=13104895, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462442, CurrentID=0, Status='live', Symbol='CG44227', Aliases='Dmel_ Dmel\\CG44227', description='CG44227 gene product from transcript CG44227-RA', other_designations='CG44227-PA|uncharacterized protein', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8684793, end_position_on_the_genomic_accession=8685176, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740640, CurrentID=0, Status='live', Symbol='CG34289', Aliases='Dmel_ Dmel\\CG34289', description='CG34289 gene product from transcript CG34289-RC', other_designations='CG34289-PB|CG34289-PC|uncharacterized protein', map_location='94F6-94F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23506908, end_position_on_the_genomic_accession=23507476, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26067049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067049, CurrentID=0, Status='live', Symbol='CG45781', Aliases='Dmel_ 40378, CG40378, CG42644, CG43676, Dmel\\CG45781, Dmel_CG40378, Dmel_CG42644, Dmel_CG43676', description='CG45781 gene product from transcript CG45781-RA', other_designations='CG45781-PA|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=432219, end_position_on_the_genomic_accession=705848, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  8674000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674000, CurrentID=0, Status='live', Symbol='CG42542', Aliases='Dmel_ CG31303, CG4196, Dmel\\CG42542, Dmel_CG31303, Dmel_CG4196', description='CG42542 gene product from transcript CG42542-RE', other_designations='CG42542-PB|CG42542-PC|CG42542-PD|CG42542-PE|CG42542-PF|CG42542-PG|uncharacterized protein', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15196598, end_position_on_the_genomic_accession=15203074, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  8674001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674001, CurrentID=0, Status='live', Symbol='CG42496', Aliases='Dmel_ Dmel\\CG42496, Dromel_CG30290_FBtr0071672_cx9c_uORF', description='CG42496 gene product from transcript CG42496-RB', other_designations='CG42496-PA|CG42496-PB|uncharacterized protein', map_location='57F8-57F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21691295, end_position_on_the_genomic_accession=21692764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12797860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797860, CurrentID=0, Status='live', Symbol='CG43209', Aliases='Dmel_ DMG4-MB6.chr2R.33.028, Dmel\\CG43209', description='CG43209 gene product from transcript CG43209-RA', other_designations='CG43209-PA|uncharacterized protein', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23601981, end_position_on_the_genomic_accession=23602456, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797866, CurrentID=0, Status='live', Symbol='CG43106', Aliases='Dmel_ Dmel\\CG43106', description='CG43106 gene product from transcript CG43106-RA', other_designations='CG43106-PA|uncharacterized protein', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25138627, end_position_on_the_genomic_accession=25139627, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797872, CurrentID=0, Status='live', Symbol='CG43348', Aliases='Dmel_ CG43180, Dmel\\CG43348, chr2L_1199311_1199719.0', description='CG43348 gene product from transcript CG43348-RB', other_designations='CG43348-PA|CG43348-PB|uncharacterized protein', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1199412, end_position_on_the_genomic_accession=1199924, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797873, CurrentID=0, Status='live', Symbol='CG43349', Aliases='Dmel_ Dmel\\CG43349', description='CG43349 gene product from transcript CG43349-RB', other_designations='CG43349-PA|CG43349-PB|uncharacterized protein', map_location='21F2-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1200205, end_position_on_the_genomic_accession=1200663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797875, CurrentID=0, Status='live', Symbol='CG43401', Aliases='Dmel_ Dmel\\CG43401', description='CG43401 gene product from transcript CG43401-RB', other_designations='CG43401-PA|CG43401-PB|uncharacterized protein', map_location='22A1-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1428474, end_position_on_the_genomic_accession=1429159, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797876, CurrentID=0, Status='live', Symbol='CG43402', Aliases='Dmel_ Dmel\\CG43402', description='CG43402 gene product from transcript CG43402-RB', other_designations='CG43402-PA|CG43402-PB|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1442627, end_position_on_the_genomic_accession=1443832, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797883, CurrentID=0, Status='live', Symbol='CG43449', Aliases='Dmel_ Dmel\\CG43449, lincRNA.748', description='CG43449 gene product from transcript CG43449-RB', other_designations='CG43449-PA|CG43449-PB|uncharacterized protein', map_location='86D2-86D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11088420, end_position_on_the_genomic_accession=11089356, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797888, CurrentID=0, Status='live', Symbol='CG43062', Aliases='Dmel_ Dmel\\CG43062', description='CG43062 gene product from transcript CG43062-RA', other_designations='CG43062-PA|uncharacterized protein', map_location='86E17-86E17', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11750333, end_position_on_the_genomic_accession=11750916, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797889, CurrentID=0, Status='live', Symbol='CG43450', Aliases='Dmel_ Dmel\\CG43450', description='CG43450 gene product from transcript CG43450-RA', other_designations='CG43450-PA|uncharacterized protein', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12195294, end_position_on_the_genomic_accession=12195574, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797890, CurrentID=0, Status='live', Symbol='CG14742', Aliases='Dmel_ DMG8-chr3R.14.009, Dmel\\CG14742', description='CG14742 gene product from transcript CG14742-RA', other_designations='CG14742-PA|uncharacterized protein', map_location='87B2-87B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12214501, end_position_on_the_genomic_accession=12215659, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797893, CurrentID=0, Status='live', Symbol='CG43630', Aliases='Dmel_ Dmel\\CG43630', description='CG43630 gene product from transcript CG43630-RA', other_designations='CG43630-PA|uncharacterized protein', map_location='87C5-87C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12676010, end_position_on_the_genomic_accession=12676260, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12797894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797894, CurrentID=0, Status='live', Symbol='CG43063', Aliases='Dmel_ Dmel\\CG43063, MIP20204', description='CG43063 gene product from transcript CG43063-RA', other_designations='CG43063-PA|uncharacterized protein', map_location='87D12-87D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13086224, end_position_on_the_genomic_accession=13086761, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797895 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797895, CurrentID=0, Status='live', Symbol='CG43208', Aliases='Dmel_ Dmel\\CG43208, MB6.chr3R.pasa.10072', description='CG43208 gene product from transcript CG43208-RA', other_designations='CG43208-PA|uncharacterized protein', map_location='87E5-87E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13303426, end_position_on_the_genomic_accession=13303856, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797901, CurrentID=0, Status='live', Symbol='CG43291', Aliases='Dmel_ Dmel\\CG43291', description='CG43291 gene product from transcript CG43291-RC', other_designations='CG43291-PA|CG43291-PB|CG43291-PC|uncharacterized protein', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14234984, end_position_on_the_genomic_accession=14244677, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12797902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797902, CurrentID=0, Status='live', Symbol='CG43179', Aliases='Dmel_ Dmel\\CG43179, chr3R_10081851_10082659.0', description='CG43179 gene product from transcript CG43179-RA', other_designations='CG43179-PA|uncharacterized protein', map_location='88B1-88B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14256108, end_position_on_the_genomic_accession=14256930, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8674005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674005, CurrentID=0, Status='live', Symbol='Sfp78E', Aliases='Dmel_CG42607, CG42607, Dmel\\CG42607', description='Seminal fluid protein 78E', other_designations='CG42607-PA|CG42607-PB|Sfp78E-PA|Sfp78E-PB|seminal fluid protein 78E', map_location='78E4-78E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21683098, end_position_on_the_genomic_accession=21684042, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12797910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797910, CurrentID=0, Status='live', Symbol='CG43248', Aliases='Dmel_ Dmel\\CG43248', description='CG43248 gene product from transcript CG43248-RD', other_designations='CG43248-PA|CG43248-PD|uncharacterized protein', map_location='71E1-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15530085, end_position_on_the_genomic_accession=15531473, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797911, CurrentID=0, Status='live', Symbol='CG43083', Aliases='Dmel_ Dmel\\CG43083', description='CG43083 gene product from transcript CG43083-RA', other_designations='CG43083-PA|uncharacterized protein', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15648571, end_position_on_the_genomic_accession=15649125, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12797912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797912, CurrentID=0, Status='live', Symbol='CG43084', Aliases='Dmel_ Dmel\\CG43084', description='CG43084 gene product from transcript CG43084-RA', other_designations='CG43084-PA|uncharacterized protein', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15650852, end_position_on_the_genomic_accession=15651392, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797913, CurrentID=0, Status='live', Symbol='CG43082', Aliases='Dmel_ Dmel\\CG43082', description='CG43082 gene product from transcript CG43082-RA', other_designations='CG43082-PA|uncharacterized protein', map_location='71E5-71E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15661867, end_position_on_the_genomic_accession=15662279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12797916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797916, CurrentID=0, Status='live', Symbol='CG43295', Aliases='Dmel_ Dmel\\CG43295', description='CG43295 gene product from transcript CG43295-RA', other_designations='CG43295-PA|uncharacterized protein', map_location='72D6-72D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16105852, end_position_on_the_genomic_accession=16106428, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797917, CurrentID=0, Status='live', Symbol='CG43249', Aliases='Dmel_ Dmel\\CG43249, lincRNA.583', description='CG43249 gene product from transcript CG43249-RA', other_designations='CG43249-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16304282, end_position_on_the_genomic_accession=16304662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797919, CurrentID=0, Status='live', Symbol='CG43206', Aliases='Dmel_ Dmel\\CG43206, chr3L_16559870_16560394.0', description='CG43206 gene product from transcript CG43206-RA', other_designations='CG43206-PA|uncharacterized protein', map_location='73A8-73A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16566750, end_position_on_the_genomic_accession=16567300, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797926, CurrentID=0, Status='live', Symbol='CG43085', Aliases='Dmel_ Dmel\\CG43085', description='CG43085 gene product from transcript CG43085-RB', other_designations='CG43085-PA|CG43085-PB|uncharacterized protein', map_location='74E4-74E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17645486, end_position_on_the_genomic_accession=17646737, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740553, CurrentID=0, Status='live', Symbol='CG34277', Aliases='Dmel_ Dmel\\CG34277', description='CG34277 gene product from transcript CG34277-RD', other_designations='CG34277-PD|CG34277-PE|uncharacterized protein', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5369540, end_position_on_the_genomic_accession=5371744, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  8674006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674006, CurrentID=0, Status='live', Symbol='CG42591', Aliases='Dmel_ Dmel\\CG42591', description='CG42591 gene product from transcript CG42591-RA', other_designations='CG42591-PA|uncharacterized protein', map_location='66B4-66B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8069453, end_position_on_the_genomic_accession=8069960, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12797937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797937, CurrentID=0, Status='live', Symbol='CG43049', Aliases='Dmel_ Dmel\\CG43049', description='CG43049 gene product from transcript CG43049-RA', other_designations='CG43049-PA|CG43049-PB|uncharacterized protein', map_location='75E6-75E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18925090, end_position_on_the_genomic_accession=18926108, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12797938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797938, CurrentID=0, Status='live', Symbol='CG43407', Aliases='Dmel_ Dmel\\CG43407', description='CG43407 gene product from transcript CG43407-RB', other_designations='CG43407-PA|CG43407-PB|uncharacterized protein', map_location='75F5-75F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19052902, end_position_on_the_genomic_accession=19053389, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12797946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797946, CurrentID=0, Status='live', Symbol='CG43198', Aliases='Dmel_ DMG6-MB6.chrX.13.005, Dmel\\CG43198', description='CG43198 gene product from transcript CG43198-RA', other_designations='CG43198-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7452694, end_position_on_the_genomic_accession=7453336, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797949, CurrentID=0, Status='live', Symbol='CG43138', Aliases='Dmel_ Dmel\\CG43138', description='CG43138 gene product from transcript CG43138-RA', other_designations='CG43138-PA|uncharacterized protein', map_location='7B3-7B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7528304, end_position_on_the_genomic_accession=7528796, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797953, CurrentID=0, Status='live', Symbol='CG43255', Aliases='Dmel_ Dmel\\CG43255', description='CG43255 gene product from transcript CG43255-RA', other_designations='CG43255-PA|uncharacterized protein', map_location='8B8-8B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8923960, end_position_on_the_genomic_accession=8924449, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797970, CurrentID=0, Status='live', Symbol='CG43156', Aliases='Dmel_ Dmel\\CG43156, lincRNA.980', description='CG43156 gene product from transcript CG43156-RA', other_designations='CG43156-PA|uncharacterized protein', map_location='10F4-10F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11906452, end_position_on_the_genomic_accession=11906868, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12797977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797977, CurrentID=0, Status='live', Symbol='CG43088', Aliases='Dmel_ Dmel\\CG43088, MB8.chrX.pasa.104, MB8.chrX.pasa.104-MIP02663', description='CG43088 gene product from transcript CG43088-RA', other_designations='CG43088-PA|uncharacterized protein', map_location='12F2-12F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14702358, end_position_on_the_genomic_accession=14703654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12797978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797978, CurrentID=0, Status='live', Symbol='CG43075', Aliases='Dmel_ Dmel\\CG43075', description='CG43075 gene product from transcript CG43075-RA', other_designations='CG43075-PA|uncharacterized protein', map_location='13B3-13B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15220271, end_position_on_the_genomic_accession=15221176, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797979, CurrentID=0, Status='live', Symbol='CG43076', Aliases='Dmel_ Dmel\\CG43076, lincRNA.997', description='CG43076 gene product from transcript CG43076-RA', other_designations='CG43076-PA|uncharacterized protein', map_location='13B3-13B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15222657, end_position_on_the_genomic_accession=15223582, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797990, CurrentID=0, Status='live', Symbol='CG43077', Aliases='Dmel_ Dmel\\CG43077', description='CG43077 gene product from transcript CG43077-RA', other_designations='CG43077-PA|uncharacterized protein', map_location='15E1-15E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17067304, end_position_on_the_genomic_accession=17068523, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12797992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797992, CurrentID=0, Status='live', Symbol='CG43221', Aliases='Dmel_ DMG8-chr2L.31.003, Dmel\\CG43221', description='CG43221 gene product from transcript CG43221-RA', other_designations='CG43221-PA|uncharacterized protein', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17485780, end_position_on_the_genomic_accession=17486005, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797993, CurrentID=0, Status='live', Symbol='CG43220', Aliases='Dmel_ DMG8-chr2L.31.003, Dmel\\CG43220', description='CG43220 gene product from transcript CG43220-RB', other_designations='CG43220-PB|uncharacterized protein', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17486597, end_position_on_the_genomic_accession=17486779, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12797997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797997, CurrentID=0, Status='live', Symbol='CG43054', Aliases='Dmel_ Dmel\\CG43054, lincRNA.173', description='CG43054 gene product from transcript CG43054-RA', other_designations='CG43054-PA|uncharacterized protein', map_location='36D1-36D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17627859, end_position_on_the_genomic_accession=17628389, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12797998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12797998, CurrentID=0, Status='live', Symbol='CG43231', Aliases='Dmel_ Dmel\\CG43231', description='CG43231 gene product from transcript CG43231-RA', other_designations='CG43231-PA|uncharacterized protein', map_location='36D2-36D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17704681, end_position_on_the_genomic_accession=17706142, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798001, CurrentID=0, Status='live', Symbol='CG43271', Aliases='Dmel_ Dmel\\CG43271, lincRNA.174', description='CG43271 gene product from transcript CG43271-RD', other_designations='CG43271-PA|CG43271-PB|CG43271-PC|CG43271-PD|uncharacterized protein', map_location='36E2-36E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17914365, end_position_on_the_genomic_accession=17915023, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798006, CurrentID=0, Status='live', Symbol='CG43362', Aliases='Dmel_ Dmel\\CG43362', description='CG43362 gene product from transcript CG43362-RA', other_designations='CG43362-PA|uncharacterized protein', map_location='36F2-36F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18291253, end_position_on_the_genomic_accession=18291462, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798009, CurrentID=0, Status='live', Symbol='CG43354', Aliases='Dmel_ Dmel\\CG43354, lincRNA.182', description='CG43354 gene product from transcript CG43354-RA', other_designations='CG43354-PA|uncharacterized protein', map_location='36F2-36F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18309187, end_position_on_the_genomic_accession=18309446, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798024, CurrentID=0, Status='live', Symbol='CG43233', Aliases='Dmel_ Dmel\\CG43233', description='CG43233 gene product from transcript CG43233-RB', other_designations='CG43233-PA|CG43233-PB|uncharacterized protein', map_location='38C8-38C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20472793, end_position_on_the_genomic_accession=20473385, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798044, CurrentID=0, Status='live', Symbol='CG43448', Aliases='Dmel_ Dmel\\CG43448, lincRNA.898', description='CG43448 gene product from transcript CG43448-RA', other_designations='CG43448-PA|uncharacterized protein', map_location='99E3-99E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30237094, end_position_on_the_genomic_accession=30237734, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798047, CurrentID=0, Status='live', Symbol='CG43290', Aliases='Dmel_ Dmel\\CG43290', description='CG43290 gene product from transcript CG43290-RA', other_designations='CG43290-PA|uncharacterized protein', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7348849, end_position_on_the_genomic_accession=7349552, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798054, CurrentID=0, Status='live', Symbol='CG43254', Aliases='Dmel_ Dmel\\CG43254, lincRNA.715', description='CG43254 gene product from transcript CG43254-RA', other_designations='CG43254-PA|CG43254-PB|uncharacterized protein', map_location='84D13-84D13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7738965, end_position_on_the_genomic_accession=7739563, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798056, CurrentID=0, Status='live', Symbol='CG43060', Aliases='Dmel_ Dmel\\CG43060', description='CG43060 gene product from transcript CG43060-RA', other_designations='CG43060-PA|CG43060-PB|uncharacterized protein', map_location='84E10-84E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8041224, end_position_on_the_genomic_accession=8055970, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798057, CurrentID=0, Status='live', Symbol='CG43618', Aliases='Dmel_ Dmel\\CG43618', description='CG43618 gene product from transcript CG43618-RA', other_designations='CG43618-PA|uncharacterized protein', map_location='84E10-84E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8049167, end_position_on_the_genomic_accession=8049456, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798058, CurrentID=0, Status='live', Symbol='CG43061', Aliases='Dmel_ Dmel\\CG43061', description='CG43061 gene product from transcript CG43061-RA', other_designations='CG43061-PA|uncharacterized protein', map_location='84E10-84E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8049673, end_position_on_the_genomic_accession=8050059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798078, CurrentID=0, Status='live', Symbol='CG43441', Aliases='Dmel_ Dmel\\CG43441', description='CG43441 gene product from transcript CG43441-RA', other_designations='CG43441-PA|uncharacterized protein', map_location='85E2-85E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9652659, end_position_on_the_genomic_accession=9653228, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798080, CurrentID=0, Status='live', Symbol='CG43123', Aliases='Dmel_ Dmel\\CG43123, chr2R_3290447_3291368.0', description='CG43123 gene product from transcript CG43123-RA', other_designations='CG43123-PA|uncharacterized protein', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7402757, end_position_on_the_genomic_accession=7403870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798081, CurrentID=0, Status='live', Symbol='CG43267', Aliases='Dmel_ Dmel\\CG43267', description='CG43267 gene product from transcript CG43267-RA', other_designations='CG43267-PA|uncharacterized protein', map_location='43B2-43B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7404035, end_position_on_the_genomic_accession=7404441, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798088, CurrentID=0, Status='live', Symbol='CG43341', Aliases='Dmel_ CG30492, Dmel\\CG43341, Dmel_CG30492', description='CG43341 gene product from transcript CG43341-RA', other_designations='CG43341-PA|uncharacterized protein', map_location='43E6-43E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7632841, end_position_on_the_genomic_accession=7633880, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798106, CurrentID=0, Status='live', Symbol='CG43288', Aliases='Dmel_ Dmel\\CG43288, lincRNA.945', description='CG43288 gene product from transcript CG43288-RA', other_designations='CG43288-PA|uncharacterized protein', map_location='4B3-4B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4213025, end_position_on_the_genomic_accession=4213421, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798107, CurrentID=0, Status='live', Symbol='CG43134', Aliases='Dmel_ Dmel\\CG43134', description='CG43134 gene product from transcript CG43134-RA', other_designations='CG43134-PA|CG43134-PB|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4247313, end_position_on_the_genomic_accession=4247785, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798108, CurrentID=0, Status='live', Symbol='CG43135', Aliases='Dmel_ Dmel\\CG43135, lincRNA.946', description='CG43135 gene product from transcript CG43135-RA', other_designations='CG43135-PA|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4248377, end_position_on_the_genomic_accession=4248657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798119, CurrentID=0, Status='live', Symbol='CG43136', Aliases='Dmel_ Dmel\\CG43136, chrX_5747778_5748511.0', description='CG43136 gene product from transcript CG43136-RA', other_designations='CG43136-PA|uncharacterized protein', map_location='5C4-5C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5853761, end_position_on_the_genomic_accession=5854464, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798120, CurrentID=0, Status='live', Symbol='CG43137', Aliases='Dmel_ Dmel\\CG43137, lincRNA.952', description='CG43137 gene product from transcript CG43137-RA', other_designations='CG43137-PA|uncharacterized protein', map_location='5C5-5C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5864550, end_position_on_the_genomic_accession=5865151, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798121, CurrentID=0, Status='live', Symbol='CG43115', Aliases='Dmel_ Dmel\\CG43115', description='CG43115 gene product from transcript CG43115-RA', other_designations='CG43115-PA|uncharacterized protein', map_location='5D3-5D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6065729, end_position_on_the_genomic_accession=6066416, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798122, CurrentID=0, Status='live', Symbol='CG43090', Aliases='Dmel_ Dmel\\CG43090', description='CG43090 gene product from transcript CG43090-RA', other_designations='CG43090-PA|uncharacterized protein', map_location='5E1-5E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6199216, end_position_on_the_genomic_accession=6199891, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798124, CurrentID=0, Status='live', Symbol='CG43195', Aliases='Dmel_ Dmel\\CG43195', description='CG43195 gene product from transcript CG43195-RA', other_designations='CG43195-PA|CG43195-PB|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19944263, end_position_on_the_genomic_accession=19945877, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12798125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798125, CurrentID=0, Status='live', Symbol='CG43308', Aliases='Dmel_ Dmel\\CG43308', description='CG43308 gene product from transcript CG43308-RB', other_designations='CG43308-PA|CG43308-PB|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20562463, end_position_on_the_genomic_accession=20563001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798138, CurrentID=0, Status='live', Symbol='CG43393', Aliases='Dmel_ Dmel\\CG43393', description='CG43393 gene product from transcript CG43393-RA', other_designations='CG43393-PA|uncharacterized protein', map_location='58B5-58B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22089423, end_position_on_the_genomic_accession=22090691, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798140, CurrentID=0, Status='live', Symbol='CG43326', Aliases='Dmel_ CG13514, CG30217, Dmel\\CG43326, Dmel_CG30217', description='CG43326 gene product from transcript CG43326-RB', other_designations='CG43326-PA|CG43326-PB|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22655549, end_position_on_the_genomic_accession=22656998, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798144, CurrentID=0, Status='live', Symbol='CG43207', Aliases='Dmel_ DMG8-chr2R.33.031, Dmel\\CG43207, lincRNA.382', description='CG43207 gene product from transcript CG43207-RA', other_designations='CG43207-PA|uncharacterized protein', map_location='59D1-59D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23152748, end_position_on_the_genomic_accession=23153318, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798145, CurrentID=0, Status='live', Symbol='HP1Lcsd', Aliases='Dmel_CG43352, CG43352, Dmel\\CG43352, lincRNA.384', description='Heterochromatin Protein 1L chromoshadow domain', other_designations='CG43352-PA|HP1Lcsd-PA|heterochromatin protein 1L chromoshadow domain', map_location='59D6-59D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23313017, end_position_on_the_genomic_accession=23313854, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798151, CurrentID=0, Status='live', Symbol='CG43133', Aliases='Dmel_ Dmel\\CG43133', description='CG43133 gene product from transcript CG43133-RA', other_designations='CG43133-PA|uncharacterized protein', map_location='16F7-16F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18129585, end_position_on_the_genomic_accession=18130178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798152, CurrentID=0, Status='live', Symbol='CG43229', Aliases='Dmel_ Dmel\\CG43229, chrX_18032634_18033088.0, lincRNA.1011', description='CG43229 gene product from transcript CG43229-RA', other_designations='CG43229-PA|CG43229-PB|uncharacterized protein', map_location='16F7-16F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18138542, end_position_on_the_genomic_accession=18139000, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798157, CurrentID=0, Status='live', Symbol='CG43153', Aliases='Dmel_ Dmel\\CG43153, lincRNA.110', description='CG43153 gene product from transcript CG43153-RA', other_designations='CG43153-PA|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11521765, end_position_on_the_genomic_accession=11522776, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38810, CurrentID=0, Status='live', Symbol='CG43780', Aliases='Dmel_ CG8600, Dmel\\CG43780', description='CG43780 gene product from transcript CG43780-RB', other_designations='CG43780-PA|CG43780-PB|uncharacterized protein', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7333544, end_position_on_the_genomic_accession=7335226, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798159, CurrentID=0, Status='live', Symbol='CG43167', Aliases='Dmel_ Dmel\\CG43167', description='CG43167 gene product from transcript CG43167-RA', other_designations='CG43167-PA|uncharacterized protein', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11657410, end_position_on_the_genomic_accession=11657992, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798172, CurrentID=0, Status='live', Symbol='CG43245', Aliases='Dmel_ Dmel\\CG43245', description='CG43245 gene product from transcript CG43245-RA', other_designations='CG43245-PA|CG43245-PB|uncharacterized protein', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10771925, end_position_on_the_genomic_accession=10773399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798184, CurrentID=0, Status='live', Symbol='CG43050', Aliases='Dmel_ DMG8-chr2L.24.001, Dmel\\CG43050, lincRNA.126', description='CG43050 gene product from transcript CG43050-RA', other_designations='CG43050-PA|uncharacterized protein', map_location='34A5-34A5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13119294, end_position_on_the_genomic_accession=13120767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12798198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798198, CurrentID=0, Status='live', Symbol='CG43376', Aliases='Dmel_ Dmel\\CG43376', description='CG43376 gene product from transcript CG43376-RA', other_designations='CG43376-PA|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13773997, end_position_on_the_genomic_accession=13774355, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798201, CurrentID=0, Status='live', Symbol='CG43390', Aliases='Dmel_ Dmel\\CG43390', description='CG43390 gene product from transcript CG43390-RA', other_designations='CG43390-PA|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11579743, end_position_on_the_genomic_accession=11579950, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798202, CurrentID=0, Status='live', Symbol='CG43391', Aliases='Dmel_ Dmel\\CG43391', description='CG43391 gene product from transcript CG43391-RA', other_designations='CG43391-PA|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11580829, end_position_on_the_genomic_accession=11581019, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674015, CurrentID=0, Status='live', Symbol='CG42465', Aliases='Dmel_ Dmel\\CG42465', description='CG42465 gene product from transcript CG42465-RA', other_designations='CG42465-PA|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3700195, end_position_on_the_genomic_accession=3700658, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798207, CurrentID=0, Status='live', Symbol='CG43064', Aliases='Dmel_ Dmel\\CG43064, MIP22993', description='CG43064 gene product from transcript CG43064-RB', other_designations='CG43064-PA|CG43064-PB|uncharacterized protein', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11834721, end_position_on_the_genomic_accession=11836721, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798213, CurrentID=0, Status='live', Symbol='CG43332', Aliases='Dmel_ BG:DS01523.2, CG15291, CG32972, Dmel\\CG43332, Dmel_CG32972', description='CG43332 gene product from transcript CG43332-RA', other_designations='CG43332-PA|uncharacterized protein', map_location='35A1-35A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14161144, end_position_on_the_genomic_accession=14163222, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798214, CurrentID=0, Status='live', Symbol='CG43052', Aliases='Dmel_ Dmel\\CG43052, MB6.chr2L.pasa.1468', description='CG43052 gene product from transcript CG43052-RB', other_designations='CG43052-PB|uncharacterized protein', map_location='35A4-35A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14352077, end_position_on_the_genomic_accession=14352619, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798218, CurrentID=0, Status='live', Symbol='CG43294', Aliases='Dmel_ Dmel\\CG43294', description='CG43294 gene product from transcript CG43294-RB', other_designations='CG43294-PA|CG43294-PB|uncharacterized protein', map_location='68E3-68E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11958291, end_position_on_the_genomic_accession=11958909, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798223, CurrentID=0, Status='live', Symbol='CG43230', Aliases='Dmel_ Dmel\\CG43230', description='CG43230 gene product from transcript CG43230-RA', other_designations='CG43230-PA|uncharacterized protein', map_location='35C4-35C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15215222, end_position_on_the_genomic_accession=15216160, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798230, CurrentID=0, Status='live', Symbol='CG43131', Aliases='Dmel_ Dmel\\CG43131', description='CG43131 gene product from transcript CG43131-RA', other_designations='CG43131-PA|uncharacterized protein', map_location='82F1-82F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5108678, end_position_on_the_genomic_accession=5109178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798232, CurrentID=0, Status='live', Symbol='CG14664', Aliases='Dmel_ Dmel\\CG14664, FBgn0040682', description='CG14664 gene product from transcript CG14664-RC', other_designations='CG14664-PA|CG14664-PB|CG14664-PC|uncharacterized protein', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5224139, end_position_on_the_genomic_accession=5225542, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798240, CurrentID=0, Status='live', Symbol='CG43147', Aliases='Dmel_ Dmel\\CG43147, lincRNA.541', description='CG43147 gene product from transcript CG43147-RB', other_designations='CG43147-PB|CG43147-PC|uncharacterized protein', map_location='70A4-70A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13299841, end_position_on_the_genomic_accession=13300152, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798242, CurrentID=0, Status='live', Symbol='CG43184', Aliases='Dmel_ Dmel\\CG43184, chr3L_13830834_13831226.0', description='CG43184 gene product from transcript CG43184-RA', other_designations='CG43184-PA|uncharacterized protein', map_location='70C4-70C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13837756, end_position_on_the_genomic_accession=13838080, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  248102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=248102, CurrentID=0, Status='live', Symbol='Gos28', Aliases='Dmel_CG7700, CG7700, Dmel\\CG7700, FSG28, GOS28, dGos28, dgos28, gos28', description='Golgi SNARE, 28 kDa', other_designations='CG7700-PA|Gos28-PA|golgi SNARE, 28 kDa', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18663738, end_position_on_the_genomic_accession=18665307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798252, CurrentID=0, Status='live', Symbol='CG43120', Aliases='Dmel_ Dmel\\CG43120', description='CG43120 gene product from transcript CG43120-RA', other_designations='CG43120-PA|uncharacterized protein', map_location='70D6-70D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14381913, end_position_on_the_genomic_accession=14383299, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798253, CurrentID=0, Status='live', Symbol='CG43246', Aliases='Dmel_ Dmel\\CG43246', description='CG43246 gene product from transcript CG43246-RA', other_designations='CG43246-PA|uncharacterized protein', map_location='70E1-70E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14536634, end_position_on_the_genomic_accession=14537272, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798254, CurrentID=0, Status='live', Symbol='CG43121', Aliases='Dmel_ Dmel\\CG43121', description='CG43121 gene product from transcript CG43121-RA', other_designations='CG43121-PA|uncharacterized protein', map_location='70E7-70E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14671935, end_position_on_the_genomic_accession=14672437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798256, CurrentID=0, Status='live', Symbol='CG43273', Aliases='Dmel_ Dmel\\CG43273', description='CG43273 gene product from transcript CG43273-RA', other_designations='CG43273-PA|CG43273-PC|uncharacterized protein', map_location='96A24-96A24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24830301, end_position_on_the_genomic_accession=24830973, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798257, CurrentID=0, Status='live', Symbol='CG13639', Aliases='Dmel_ Dmel\\CG13639, anon-SAGE:Wang-045', description='CG13639 gene product from transcript CG13639-RA', other_designations='CG13639-PA|anon-SAGE:Wang-045|uncharacterized protein', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24859900, end_position_on_the_genomic_accession=24860366, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798270, CurrentID=0, Status='live', Symbol='CG43116', Aliases='Dmel_ Dmel\\CG43116', description='CG43116 gene product from transcript CG43116-RA', other_designations='CG43116-PA|uncharacterized protein', map_location='96F9-96F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25995879, end_position_on_the_genomic_accession=25996996, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798272, CurrentID=0, Status='live', Symbol='CG43117', Aliases='Dmel_ Dmel\\CG43117', description='CG43117 gene product from transcript CG43117-RB', other_designations='CG43117-PB|uncharacterized protein', map_location='97B3-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26454794, end_position_on_the_genomic_accession=26455112, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798275, CurrentID=0, Status='live', Symbol='CG43447', Aliases='Dmel_ Dmel\\CG43447, lincRNA.869', description='CG43447 gene product from transcript CG43447-RA', other_designations='CG43447-PA|uncharacterized protein', map_location='98A1-98A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27440312, end_position_on_the_genomic_accession=27440845, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798276, CurrentID=0, Status='live', Symbol='CG43139', Aliases='Dmel_ DMG4-MB6.chr3R.40.017, Dmel\\CG43139', description='CG43139 gene product from transcript CG43139-RB', other_designations='CG43139-PA|CG43139-PB|uncharacterized protein', map_location='98A3-98A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27501366, end_position_on_the_genomic_accession=27502025, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798278, CurrentID=0, Status='live', Symbol='CG43319', Aliases='Dmel_ CG12879, Dmel\\CG43319', description='CG43319 gene product from transcript CG43319-RA', other_designations='CG43319-PA|uncharacterized protein', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27694318, end_position_on_the_genomic_accession=27694541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798279, CurrentID=0, Status='live', Symbol='Acp98AB', Aliases='Dmel_CG43619, Acp98, Acp98Ab, Acp98B, CG31056, CG43619, Dmel\\CG43619, Dmel_CG31056', description='Accessory gland protein 98AB', other_designations='Acp98AB-PB|CG43619-PB|accessory gland protein 98AB|accessory gland-specific peptide 98AB|accessory gland-specific peptide 98B', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27694883, end_position_on_the_genomic_accession=27695057, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798281, CurrentID=0, Status='live', Symbol='CG43125', Aliases='Dmel_ Dmel\\CG43125', description='CG43125 gene product from transcript CG43125-RA', other_designations='CG43125-PA|uncharacterized protein', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28262735, end_position_on_the_genomic_accession=28263622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798282, CurrentID=0, Status='live', Symbol='CG43124', Aliases='Dmel_ Dmel\\CG43124', description='CG43124 gene product from transcript CG43124-RA', other_designations='CG43124-PA|uncharacterized protein', map_location='98C3-98C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28263871, end_position_on_the_genomic_accession=28264773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798292, CurrentID=0, Status='live', Symbol='CG43181', Aliases='Dmel_ Dmel\\CG43181, chrX_244575_245164.0', description='CG43181 gene product from transcript CG43181-RA', other_designations='CG43181-PA|uncharacterized protein', map_location='1A5-1A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=350566, end_position_on_the_genomic_accession=351133, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798293, CurrentID=0, Status='live', Symbol='CG43182', Aliases='Dmel_ Dmel\\CG43182', description='CG43182 gene product from transcript CG43182-RA', other_designations='CG43182-PA|uncharacterized protein', map_location='1A5-1A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=353067, end_position_on_the_genomic_accession=353317, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462463, CurrentID=0, Status='live', Symbol='CG44194', Aliases='Dmel_ Dmel\\CG44194', description='CG44194 gene product from transcript CG44194-RB', other_designations='CG44194-PA|CG44194-PB|uncharacterized protein', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12996876, end_position_on_the_genomic_accession=12997726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798297, CurrentID=0, Status='live', Symbol='Diedel2', Aliases='Dmel_CG43228, CG43228, Chr3R_25434193_25434663.0, Diedel-2, Dmel\\CG43228', description='Diedel 2', other_designations='CG43228-PA|Diedel-2|Diedel2-PA|diedel 2', map_location='99B4-99B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29608477, end_position_on_the_genomic_accession=29608854, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798298, CurrentID=0, Status='live', Symbol='CG43321', Aliases='Dmel_ CG13776, Dmel\\CG43321', description='CG43321 gene product from transcript CG43321-RA', other_designations='CG43321-PA|uncharacterized protein', map_location='27D4-27D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7012262, end_position_on_the_genomic_accession=7012803, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798301, CurrentID=0, Status='live', Symbol='CG43235', Aliases='Dmel_ Dmel\\CG43235', description='CG43235 gene product from transcript CG43235-RA', other_designations='CG43235-PA|uncharacterized protein', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7693030, end_position_on_the_genomic_accession=7693821, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798305, CurrentID=0, Status='live', Symbol='CG43057', Aliases='Dmel_ Dmel\\CG43057', description='CG43057 gene product from transcript CG43057-RB', other_designations='CG43057-PA|CG43057-PB|uncharacterized protein', map_location='28F5-28F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8243323, end_position_on_the_genomic_accession=8244003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798308, CurrentID=0, Status='live', Symbol='CG43400', Aliases='Dmel_ Dmel\\CG43400, lincRNA.75', description='CG43400 gene product from transcript CG43400-RA', other_designations='CG43400-PA|uncharacterized protein', map_location='29D1-29D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8490739, end_position_on_the_genomic_accession=8491360, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798309, CurrentID=0, Status='live', Symbol='CG43350', Aliases='Dmel_ CG42845, Chr2L_9349524_9350014.0, Dmel\\CG43350, lincRNA.88', description='CG43350 gene product from transcript CG43350-RB', other_designations='CG43350-PA|CG43350-PB|uncharacterized protein', map_location='30A9-30A9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9349546, end_position_on_the_genomic_accession=9349963, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798317, CurrentID=0, Status='live', Symbol='CG43187', Aliases='Dmel_ DMG8-chr2R.23.004, Dmel\\CG43187, lincRNA.322', description='CG43187 gene product from transcript CG43187-RB', other_designations='CG43187-PA|CG43187-PB|uncharacterized protein', map_location='53C3-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16380516, end_position_on_the_genomic_accession=16385765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798323, CurrentID=0, Status='live', Symbol='CG43327', Aliases='Dmel_ CG30459, Dmel\\CG43327, Dmel_CG30459', description='CG43327 gene product from transcript CG43327-RA', other_designations='CG43327-PA|uncharacterized protein', map_location='53E10-53E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17032487, end_position_on_the_genomic_accession=17034012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798324, CurrentID=0, Status='live', Symbol='CG43371', Aliases='Dmel_ Dmel\\CG43371', description='CG43371 gene product from transcript CG43371-RA', other_designations='CG43371-PA|uncharacterized protein', map_location='53E11-53E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17034045, end_position_on_the_genomic_accession=17034609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798325, CurrentID=0, Status='live', Symbol='CG43103', Aliases='Dmel_ Dm_2R:55417, Dmel\\CG43103', description='CG43103 gene product from transcript CG43103-RA', other_designations='CG43103-PA|uncharacterized protein', map_location='54A1-54A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17139450, end_position_on_the_genomic_accession=17139931, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462464, CurrentID=0, Status='live', Symbol='CG44142', Aliases='Dmel_ Dmel\\CG44142', description='CG44142 gene product from transcript CG44142-RA', other_designations='CG44142-PA|uncharacterized protein', map_location='87E5-87E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13312855, end_position_on_the_genomic_accession=13313188, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798327, CurrentID=0, Status='live', Symbol='CG43237', Aliases='Dmel_ Dmel\\CG43237, lincRNA.333', description='CG43237 gene product from transcript CG43237-RA', other_designations='CG43237-PA|uncharacterized protein', map_location='54A3-54A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17205208, end_position_on_the_genomic_accession=17205743, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798328, CurrentID=0, Status='live', Symbol='CG43272', Aliases='Dmel_ Dmel\\CG43272, lincRNA.337', description='CG43272 gene product from transcript CG43272-RA', other_designations='CG43272-PA|CG43272-PB|uncharacterized protein', map_location='54B1-54B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17255241, end_position_on_the_genomic_accession=17255737, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798329, CurrentID=0, Status='live', Symbol='CG43108', Aliases='Dmel_ Dmel\\CG43108, lincRNA.338', description='CG43108 gene product from transcript CG43108-RA', other_designations='CG43108-PA|uncharacterized protein', map_location='54B1-54B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17257156, end_position_on_the_genomic_accession=17257886, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798330, CurrentID=0, Status='live', Symbol='CG43110', Aliases='Dmel_ Dmel\\CG43110, MB7.chr2R.pasa.4777', description='CG43110 gene product from transcript CG43110-RB', other_designations='CG43110-PA|CG43110-PB|uncharacterized protein', map_location='54B17-54B17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17477565, end_position_on_the_genomic_accession=17486220, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12798331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798331, CurrentID=0, Status='live', Symbol='CG43324', Aliases='Dmel_ CG15844, Dmel\\CG43324, Dmel_CG15844', description='CG43324 gene product from transcript CG43324-RA', other_designations='CG43324-PA|uncharacterized protein', map_location='54C8-54C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17563606, end_position_on_the_genomic_accession=17564351, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798334, CurrentID=0, Status='live', Symbol='CG43202', Aliases='Dmel_ CG42737, DMG8-chr2R.26.014, Dmel\\CG43202, chr2R.1731', description='CG43202 gene product from transcript CG43202-RA', other_designations='CG43202-PA|uncharacterized protein', map_location='55C4-55C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18381393, end_position_on_the_genomic_accession=18381844, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  248194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=248194, CurrentID=0, Status='live', Symbol='Gclm', Aliases='Dmel_CG4919, 4919, CG4919, DmGCLM, DmGCSl, Dmel\\CG4919, GCLM, GCLm, GCSl, Gcsl, box, gclm', description='Glutamate-cysteine ligase modifier subunit', other_designations='CG4919-PA|CG4919-PB|Gclm-PA|Gclm-PB|gamma-Glutamylcysteine synthetase regulatory subunit', map_location='94C1-94C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22685006, end_position_on_the_genomic_accession=22686510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798339, CurrentID=0, Status='live', Symbol='CG43070', Aliases='Dmel_ Dmel\\CG43070, lincRNA.347', description='CG43070 gene product from transcript CG43070-RA', other_designations='CG43070-PA|CG43070-PB|uncharacterized protein', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18863896, end_position_on_the_genomic_accession=18864558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798340, CurrentID=0, Status='live', Symbol='CG43071', Aliases='Dmel_ Dmel\\CG43071, lincRNA.347', description='CG43071 gene product from transcript CG43071-RC', other_designations='CG43071-PB|CG43071-PC|uncharacterized protein', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18864621, end_position_on_the_genomic_accession=18865218, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798341, CurrentID=0, Status='live', Symbol='CG43109', Aliases='Dmel_ Dmel\\CG43109', description='CG43109 gene product from transcript CG43109-RA', other_designations='CG43109-PA|uncharacterized protein', map_location='55F8-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18867848, end_position_on_the_genomic_accession=18868167, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798343, CurrentID=0, Status='live', Symbol='CG43351', Aliases='Dmel_ CG42725, Dmel\\CG43351', description='CG43351 gene product from transcript CG43351-RA', other_designations='CG43351-PA|uncharacterized protein', map_location='56B1-56B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19030009, end_position_on_the_genomic_accession=19030250, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798347, CurrentID=0, Status='live', Symbol='CG43111', Aliases='Dmel_ Dmel\\CG43111', description='CG43111 gene product from transcript CG43111-RA', other_designations='CG43111-PA|uncharacterized protein', map_location='56D15-56D15', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19552071, end_position_on_the_genomic_accession=19552672, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798348, CurrentID=0, Status='live', Symbol='CG42878', Aliases='Dmel_ Dmel\\CG42878, chr2R_15474834_15475129.0', description='CG42878 gene product from transcript CG42878-RA', other_designations='CG42878-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19587372, end_position_on_the_genomic_accession=19587763, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798350, CurrentID=0, Status='live', Symbol='CG43277', Aliases='Dmel_ Dmel\\CG43277', description='CG43277 gene product from transcript CG43277-RA', other_designations='CG43277-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19637559, end_position_on_the_genomic_accession=19638041, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798363, CurrentID=0, Status='live', Symbol='CG43293', Aliases='Dmel_ Dmel\\CG43293', description='CG43293 gene product from transcript CG43293-RA', other_designations='CG43293-PA|CG43293-PB|uncharacterized protein', map_location='65A7-65A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6185189, end_position_on_the_genomic_accession=6185853, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798364, CurrentID=0, Status='live', Symbol='CG43168', Aliases='Dmel_ Dmel\\CG43168, chr3L_6268776_6270114.0, lincRNA.451', description='CG43168 gene product from transcript CG43168-RA', other_designations='CG43168-PA|uncharacterized protein', map_location='65A11-65A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6275787, end_position_on_the_genomic_accession=6277035, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798365, CurrentID=0, Status='live', Symbol='CG43439', Aliases='Dmel_ Dmel\\CG43439', description='CG43439 gene product from transcript CG43439-RA', other_designations='CG43439-PA|uncharacterized protein', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6983473, end_position_on_the_genomic_accession=6984586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798367, CurrentID=0, Status='live', Symbol='CG43292', Aliases='Dmel_ Dmel\\CG43292, lincRNA.463, mlncRNA66A2', description='CG43292 gene product from transcript CG43292-RA', other_designations='CG43292-PA|uncharacterized protein', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7475570, end_position_on_the_genomic_accession=7476247, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798369, CurrentID=0, Status='live', Symbol='CG43059', Aliases='Dmel_ Dmel\\CG43059', description='CG43059 gene product from transcript CG43059-RA', other_designations='CG43059-PA|uncharacterized protein', map_location='66C6-66C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8318542, end_position_on_the_genomic_accession=8318966, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798372, CurrentID=0, Status='live', Symbol='CG43169', Aliases='Dmel_ Dmel\\CG43169, MB8.chr3L.pasa.50', description='CG43169 gene product from transcript CG43169-RA', other_designations='CG43169-PA|uncharacterized protein', map_location='66E3-66E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8888145, end_position_on_the_genomic_accession=8888633, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798379, CurrentID=0, Status='live', Symbol='CG43296', Aliases='Dmel_ Dmel\\CG43296', description='CG43296 gene product from transcript CG43296-RA', other_designations='CG43296-PA|uncharacterized protein', map_location='44C6-44C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8417176, end_position_on_the_genomic_accession=8417706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798381, CurrentID=0, Status='live', Symbol='CG43183', Aliases='Dmel_ DMG4-MB6.chr2R.8.009, DMG4-MB6.chr2R.8.009-MIP07212, Dmel\\CG43183, lincRNA.270', description='CG43183 gene product from transcript CG43183-RA', other_designations='CG43183-PA|uncharacterized protein', map_location='44D4-44D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8592195, end_position_on_the_genomic_accession=8592969, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798390, CurrentID=0, Status='live', Symbol='CG43200', Aliases='Dmel_ Dmel\\CG43200, nc_5744.a', description='CG43200 gene product from transcript CG43200-RA', other_designations='CG43200-PA|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10351533, end_position_on_the_genomic_accession=10352161, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798391, CurrentID=0, Status='live', Symbol='CG43172', Aliases='Dmel_ Dmel\\CG43172, nc_5744.a, noncoding_5744', description='CG43172 gene product from transcript CG43172-RA', other_designations='CG43172-PA|CG43172-PB|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10351533, end_position_on_the_genomic_accession=10352161, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798392, CurrentID=0, Status='live', Symbol='CG43201', Aliases='Dmel_ Dmel\\CG43201, MIP03715', description='CG43201 gene product from transcript CG43201-RB', other_designations='CG43201-PA|CG43201-PB|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10352456, end_position_on_the_genomic_accession=10353114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798393, CurrentID=0, Status='live', Symbol='CG43171', Aliases='Dmel_ Dmel\\CG43171, MIP03715, MIP03715_ORF_1', description='CG43171 gene product from transcript CG43171-RB', other_designations='CG43171-PA|CG43171-PB|CG43171-PC|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10352456, end_position_on_the_genomic_accession=10353114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798394, CurrentID=0, Status='live', Symbol='CG43397', Aliases='Dmel_ Dmel\\CG43397', description='CG43397 gene product from transcript CG43397-RB', other_designations='CG43397-PA|CG43397-PB|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10354724, end_position_on_the_genomic_accession=10355831, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798395, CurrentID=0, Status='live', Symbol='CG43396', Aliases='Dmel_ Dmel\\CG43396', description='CG43396 gene product from transcript CG43396-RA', other_designations='CG43396-PA|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10360213, end_position_on_the_genomic_accession=10360840, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798403, CurrentID=0, Status='live', Symbol='CG43188', Aliases='Dmel_ Dmel\\CG43188, MIP03057, noncoding_5956', description='CG43188 gene product from transcript CG43188-RA', other_designations='CG43188-PA|uncharacterized protein', map_location='47D3-47D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11085586, end_position_on_the_genomic_accession=11088300, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798404, CurrentID=0, Status='live', Symbol='CG43189', Aliases='Dmel_ Dmel\\CG43189', description='CG43189 gene product from transcript CG43189-RA', other_designations='CG43189-PA|uncharacterized protein', map_location='47D3-47D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11086245, end_position_on_the_genomic_accession=11086845, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798405, CurrentID=0, Status='live', Symbol='CG43112', Aliases='Dmel_ Dmel\\CG43112, MIP12787', description='CG43112 gene product from transcript CG43112-RA', other_designations='CG43112-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11249594, end_position_on_the_genomic_accession=11249933, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798406, CurrentID=0, Status='live', Symbol='CG43114', Aliases='Dmel_ Dmel\\CG43114, MIP09856', description='CG43114 gene product from transcript CG43114-RA', other_designations='CG43114-PA|CG43114-PB|uncharacterized protein', map_location='47F1-47F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11310831, end_position_on_the_genomic_accession=11311165, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798407, CurrentID=0, Status='live', Symbol='CG43190', Aliases='Dmel_ DMG8-chr2R.14.008, Dmel\\CG43190', description='CG43190 gene product from transcript CG43190-RA', other_designations='CG43190-PA|uncharacterized protein', map_location='48C4-48C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11844378, end_position_on_the_genomic_accession=11845099, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798408, CurrentID=0, Status='live', Symbol='CG43191', Aliases='Dmel_ Dmel\\CG43191, noncoding_6153', description='CG43191 gene product from transcript CG43191-RA', other_designations='CG43191-PA|uncharacterized protein', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12035921, end_position_on_the_genomic_accession=12037165, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12798413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798413, CurrentID=0, Status='live', Symbol='CG43335', Aliases='Dmel_ CG5302, Dmel\\CG43335, SPH73', description='CG43335 gene product from transcript CG43335-RA', other_designations='CG43335-PA|uncharacterized protein', map_location='89A1-89A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15635468, end_position_on_the_genomic_accession=15636489, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798414, CurrentID=0, Status='live', Symbol='CG43336', Aliases='Dmel_ CG5302, Dmel\\CG43336, SPH73', description='CG43336 gene product from transcript CG43336-RA', other_designations='CG43336-PA|uncharacterized protein', map_location='89A1-89A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15637168, end_position_on_the_genomic_accession=15638181, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798419, CurrentID=0, Status='live', Symbol='CG43315', Aliases='Dmel_ CG13169, Dmel\\CG43315, Dmel_CG13169', description='CG43315 gene product from transcript CG43315-RB', other_designations='CG43315-PA|CG43315-PB|uncharacterized protein', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12241218, end_position_on_the_genomic_accession=12241872, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798420, CurrentID=0, Status='live', Symbol='CG43244', Aliases='Dmel_ Dmel\\CG43244', description='CG43244 gene product from transcript CG43244-RA', other_designations='CG43244-PA|uncharacterized protein', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12243999, end_position_on_the_genomic_accession=12244701, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798422, CurrentID=0, Status='live', Symbol='CG43318', Aliases='Dmel_ CG14876, Dmel\\CG43318, Dmel_CG14876', description='CG43318 gene product from transcript CG43318-RA', other_designations='CG43318-PA|uncharacterized protein', map_location='89A5-89A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15832708, end_position_on_the_genomic_accession=15833432, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798425, CurrentID=0, Status='live', Symbol='CG43204', Aliases='Dmel_ Dmel\\CG43204, chr2R_8318110_8318385.0', description='CG43204 gene product from transcript CG43204-RA', other_designations='CG43204-PA|uncharacterized protein', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12430595, end_position_on_the_genomic_accession=12431296, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798428, CurrentID=0, Status='live', Symbol='CG43251', Aliases='Dmel_ Dmel\\CG43251, lincRNA.639', description='CG43251 gene product from transcript CG43251-RA', other_designations='CG43251-PA|uncharacterized protein', map_location='77E4-77E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20758883, end_position_on_the_genomic_accession=20759192, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798436, CurrentID=0, Status='live', Symbol='CG43072', Aliases='Dmel_ Dmel\\CG43072, chr3L_21058978_21059336.0', description='CG43072 gene product from transcript CG43072-RA', other_designations='CG43072-PA|uncharacterized protein', map_location='78B1-78B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21065852, end_position_on_the_genomic_accession=21066238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798437, CurrentID=0, Status='live', Symbol='CG43219', Aliases='Dmel_ DMG4-MB6.chr3L.36.014, Dmel\\CG43219', description='CG43219 gene product from transcript CG43219-RA', other_designations='CG43219-PA|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21220400, end_position_on_the_genomic_accession=21223118, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798438, CurrentID=0, Status='live', Symbol='CG43218', Aliases='Dmel_ Dmel\\CG43218, MB8.chr3L.pasa.58', description='CG43218 gene product from transcript CG43218-RA', other_designations='CG43218-PA|uncharacterized protein', map_location='78C3-78C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21258020, end_position_on_the_genomic_accession=21258758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798442, CurrentID=0, Status='live', Symbol='CG43192', Aliases='Dmel_ Dmel\\CG43192, noncoding_6564, noncoding_6907', description='CG43192 gene product from transcript CG43192-RA', other_designations='CG43192-PA|uncharacterized protein', map_location='50A11-50A11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13477342, end_position_on_the_genomic_accession=13478079, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  41383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41383, CurrentID=0, Status='live', Symbol='CG14708', Aliases='Dmel_ Dmel\\CG14708', description='CG14708 gene product from transcript CG14708-RA', other_designations='CG14708-PA|uncharacterized protein', map_location='86E10-86E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11554145, end_position_on_the_genomic_accession=11554869, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798445, CurrentID=0, Status='live', Symbol='CG43058', Aliases='Dmel_ Dmel\\CG43058', description='CG43058 gene product from transcript CG43058-RA', other_designations='CG43058-PA|uncharacterized protein', map_location='50B5-50B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13600645, end_position_on_the_genomic_accession=13601099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798449, CurrentID=0, Status='live', Symbol='CG43175', Aliases='Dmel_ Dmel\\CG43175', description='CG43175 gene product from transcript CG43175-RB', other_designations='CG43175-PA|CG43175-PB|uncharacterized protein', map_location='90B7-90B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17630717, end_position_on_the_genomic_accession=17645262, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798450, CurrentID=0, Status='live', Symbol='CG43445', Aliases='Dmel_ Dmel\\CG43445', description='CG43445 gene product from transcript CG43445-RA', other_designations='CG43445-PA|uncharacterized protein', map_location='90C10-90C10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17911376, end_position_on_the_genomic_accession=17912045, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12798457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798457, CurrentID=0, Status='live', Symbol='CG43392', Aliases='Dmel_ Dmel\\CG43392', description='CG43392 gene product from transcript CG43392-RA', other_designations='CG43392-PA|uncharacterized protein', map_location='79B1-79B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22029453, end_position_on_the_genomic_accession=22029770, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798458, CurrentID=0, Status='live', Symbol='CG43312', Aliases='Dmel_ Dmel\\CG43312', description='CG43312 gene product from transcript CG43312-RA', other_designations='CG43312-PA|uncharacterized protein', map_location='79E5-79E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22501600, end_position_on_the_genomic_accession=22502417, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12798459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798459, CurrentID=0, Status='live', Symbol='CG43331', Aliases='Dmel_ CG12377, Dmel\\CG43331, Dmel_CG12377', description='CG43331 gene product from transcript CG43331-RA', other_designations='CG43331-PA|uncharacterized protein', map_location='79E5-79E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22502978, end_position_on_the_genomic_accession=22503830, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12798460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798460, CurrentID=0, Status='live', Symbol='CG43330', Aliases='Dmel_ CG12377, Dmel\\CG43330, Dmel_CG12377', description='CG43330 gene product from transcript CG43330-RA', other_designations='CG43330-PA|uncharacterized protein', map_location='79E5-79E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22503869, end_position_on_the_genomic_accession=22504423, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3885565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885565, CurrentID=0, Status='live', Symbol='PIP4K', Aliases='Dmel_CG17471, CG17471, Dmel\\CG17471, PIP5K 102E, dPIP4K', description='Phosphatidylinositol 5-phosphate 4-kinase', other_designations='CG17471-PA|CG17471-PB|PIP4K-PA|PIP4K-PB|phosphatidylinositol 5-phosphate 4-kinase', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1192419, end_position_on_the_genomic_accession=1196848, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  3885567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885567, CurrentID=0, Status='live', Symbol='Dyrk3', Aliases='Dmel_CG40478, BcDNA:RE60792, CG40478, DDYRK3, DYRK3, Dm3, DmDyrk3, Dmel\\CG40478, FBtr0100406', description='Dual-specificity tyrosine phosphorylation-regulated kinase 3', other_designations='CG40478-PA|CG40478-PC|CG40478-PE|Dyrk3-PA|Dyrk3-PC|Dyrk3-PE|unknown-telomeric-protein-gene', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1225089, end_position_on_the_genomic_accession=1230713, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3885569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885569, CurrentID=0, Status='live', Symbol='CG34026', Aliases='Dmel_ BP1064, Dmel\\CG34026', description='CG34026 gene product from transcript CG34026-RB', other_designations='CG34026-PA|CG34026-PB|uncharacterized protein', map_location='8E7-8E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9533818, end_position_on_the_genomic_accession=9534625, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885570, CurrentID=0, Status='live', Symbol='CG34015', Aliases='Dmel_ BP1006, Dmel\\CG34015', description='CG34015 gene product from transcript CG34015-RA', other_designations='CG34015-PA|uncharacterized protein', map_location='14C6-14C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16447568, end_position_on_the_genomic_accession=16448302, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798468, CurrentID=0, Status='live', Symbol='CG43101', Aliases='Dmel_ Dmel\\CG43101', description='CG43101 gene product from transcript CG43101-RA', other_designations='CG43101-PA|uncharacterized protein', map_location='51D8-51D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15013047, end_position_on_the_genomic_accession=15013522, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885573, CurrentID=0, Status='live', Symbol='Octbeta3R', Aliases='Dmel_CG42244, CG18554, CG31348, CG31350, CG31351, CG33959, CG34055, CG42244, CG7078, DmCG7078, DmOctbeta3R, Dmel\\CG42244, Dmel_CG18554, Dmel_CG33959, Dmel_CG34055, OctBeta3, Octb3R, anon-WO0170980.187, anon-WO0170980.188, octRbeta3, octbeta3R', description='Octopamine beta3 receptor', other_designations='CG42244-PF|CG42244-PG|CG42244-PJ|CG42244-PK|Octbeta3R-PF|Octbeta3R-PG|Octbeta3R-PJ|Octbeta3R-PK|octopamine beta3 receptor|octopamine receptor beta3', map_location='87B15-87C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12511570, end_position_on_the_genomic_accession=12548229, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  3885575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885575, CurrentID=0, Status='live', Symbol='CG33977', Aliases='Dmel_ Dmel\\CG33977', description='CG33977 gene product from transcript CG33977-RA', other_designations='CG33977-PA|uncharacterized protein', map_location='87E9-87E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13402559, end_position_on_the_genomic_accession=13403051, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885576, CurrentID=0, Status='live', Symbol='CG33721', Aliases='Dmel_ CR33721, Dmel\\CG33721', description='CG33721 gene product from transcript CG33721-RA', other_designations='CG33721-PA|uncharacterized protein', map_location='94A5-94A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22331069, end_position_on_the_genomic_accession=22331801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3885577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885577, CurrentID=0, Status='live', Symbol='CG34008', Aliases='Dmel_ BP1002, Dmel\\CG34008', description='CG34008 gene product from transcript CG34008-RA', other_designations='CG34008-PA|uncharacterized protein', map_location='92F2-92F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20580683, end_position_on_the_genomic_accession=20581522, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3885578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885578, CurrentID=0, Status='live', Symbol='CG42823', Aliases='Dmel_ BP1087, CG34053, Dmel\\CG42823, Dmel_CG34053, HDC14119', description='CG42823 gene product from transcript CG42823-RA', other_designations='CG42823-PA|uncharacterized protein', map_location='90A5-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17404487, end_position_on_the_genomic_accession=17405109, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3885579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885579, CurrentID=0, Status='live', Symbol='CG42821', Aliases='Dmel_ BP1011, CG34035, Dmel\\CG42821, Dmel_CG34035', description='CG42821 gene product from transcript CG42821-RA', other_designations='CG42821-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17413077, end_position_on_the_genomic_accession=17413639, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798476, CurrentID=0, Status='live', Symbol='CG43203', Aliases='Dmel_ CG42793, Dmel\\CG43203, noncoding_17732', description='CG43203 gene product from transcript CG43203-RA', other_designations='CG43203-PA|uncharacterized protein', map_location='92A2-92A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19354051, end_position_on_the_genomic_accession=19354884, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885581, CurrentID=0, Status='live', Symbol='CG34023', Aliases='Dmel_ BP1072, Dmel\\CG34023', description='CG34023 gene product from transcript CG34023-RA', other_designations='CG34023-PA|uncharacterized protein', map_location='84D6-84D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7420268, end_position_on_the_genomic_accession=7420698, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885583, CurrentID=0, Status='live', Symbol='CG34041', Aliases='Dmel_ BP1033, CG31018, Dmel\\CG34041', description='CG34041 gene product from transcript CG34041-RF', other_designations='CG34041-PD|CG34041-PE|CG34041-PF|uncharacterized protein', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30508023, end_position_on_the_genomic_accession=30510146, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  3885584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885584, CurrentID=0, Status='live', Symbol='CG34034', Aliases='Dmel_ BP1088, Dmel\\CG34034, HDC14863', description='CG34034 gene product from transcript CG34034-RB', other_designations='CG34034-PB|uncharacterized protein', map_location='93F7-93F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21738345, end_position_on_the_genomic_accession=21739110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3885585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885585, CurrentID=0, Status='live', Symbol='CG34027', Aliases='Dmel_ BP1028, Dmel\\CG34027', description='CG34027 gene product from transcript CG34027-RA', other_designations='CG34027-PA|uncharacterized protein', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25103670, end_position_on_the_genomic_accession=25109032, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  3885587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885587, CurrentID=0, Status='live', Symbol='CG33465', Aliases='Dmel_ BP1080, CR33465, Dmel\\CG33465, HDC06631', description='CG33465 gene product from transcript CG33465-RC', other_designations='CG33465-PC|uncharacterized protein', map_location='52C7-52C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15819114, end_position_on_the_genomic_accession=15821114, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3885588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885588, CurrentID=0, Status='live', Symbol='CG33460', Aliases='Dmel_ CG30086, CG8215, CR33460, Dmel\\CG33460, SP41', description='CG33460 gene product from transcript CG33460-RA', other_designations='CG33460-PA|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15694092, end_position_on_the_genomic_accession=15695232, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3885589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885589, CurrentID=0, Status='live', Symbol='CG34021', Aliases='Dmel_ BP1044, Dmel\\CG34021', description='CG34021 gene product from transcript CG34021-RA', other_designations='CG34021-PA|uncharacterized protein', map_location='48F10-48F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12336890, end_position_on_the_genomic_accession=12337718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885590, CurrentID=0, Status='live', Symbol='CG33632', Aliases='Dmel_ CR33632, Dmel\\CG33632', description='CG33632 gene product from transcript CG33632-RA', other_designations='CG33632-PA|uncharacterized protein', map_location='49A10-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12522023, end_position_on_the_genomic_accession=12522731, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3885592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885592, CurrentID=0, Status='live', Symbol='CG34029', Aliases='Dmel_ BP1066, Dmel\\CG34029', description='CG34029 gene product from transcript CG34029-RA', other_designations='CG34029-PA|uncharacterized protein', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21970767, end_position_on_the_genomic_accession=21971285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885594, CurrentID=0, Status='live', Symbol='CG33998', Aliases='Dmel_ BP1058, Dmel\\CG33998', description='CG33998 gene product from transcript CG33998-RA', other_designations='CG33998-PA|uncharacterized protein', map_location='55E11-55E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18738492, end_position_on_the_genomic_accession=18739143, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885596, CurrentID=0, Status='live', Symbol='CG34038', Aliases='Dmel_ BP1023, Dm_2R:93795, Dm_2R:93799, Dmel\\CG34038', description='CG34038 gene product from transcript CG34038-RB', other_designations='CG34038-PB|CG34038-PC|uncharacterized protein', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25167659, end_position_on_the_genomic_accession=25169855, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3885597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885597, CurrentID=0, Status='live', Symbol='CG34005', Aliases='Dmel_ BP1019, Dmel\\CG34005', description='CG34005 gene product from transcript CG34005-RB', other_designations='CG34005-PB|CG34005-PC|uncharacterized protein', map_location='55B2-55B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18048108, end_position_on_the_genomic_accession=18048909, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885598, CurrentID=0, Status='live', Symbol='dpr13', Aliases='Dmel_CG33996, CG12557, CG14493, CG33996, CT34204, Dm_2R:58906, Dmel\\CG33996, Dpr-13, Dpr13', description='defective proboscis extension response 13', other_designations='CG33996-PB|CG33996-PC|dpr13-PB|dpr13-PC', map_location='55A3-55A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17975122, end_position_on_the_genomic_accession=18007902, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3885599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885599, CurrentID=0, Status='live', Symbol='CG13516', Aliases='Dmel_ BP1036, Dmel\\CG13516', description='CG13516 gene product from transcript CG13516-RB', other_designations='CG13516-PA|CG13516-PB|uncharacterized protein', map_location='58F7-58F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22669145, end_position_on_the_genomic_accession=22670622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36639, CurrentID=0, Status='live', Symbol='CG10151', Aliases='Dmel_ Dmel\\CG10151', description='CG10151 gene product from transcript CG10151-RA', other_designations='CG10151-PA|CG10151-PB|CG10151-PC|CG10151-PD|uncharacterized protein', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14753337, end_position_on_the_genomic_accession=14756288, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3885601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885601, CurrentID=0, Status='live', Symbol='CG34054', Aliases='Dmel_ BP1079, Dmel\\CG34054, HDC05792', description='CG34054 gene product from transcript CG34054-RA', other_designations='CG34054-PA|uncharacterized protein', map_location='47F11-47F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11409810, end_position_on_the_genomic_accession=11410438, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885602, CurrentID=0, Status='live', Symbol='CG34033', Aliases='Dmel_ BP1078, Dmel\\CG34033, HDC05479', description='CG34033 gene product from transcript CG34033-RA', other_designations='CG34033-PA|uncharacterized protein', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9820563, end_position_on_the_genomic_accession=9821186, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885603, CurrentID=0, Status='live', Symbol='IM14', Aliases='Dmel_CG33990, CG33990, DIM 14, Dmel\\CG33990, im14', description='Immune induced molecule 14', other_designations='CG33990-PA|IM14-PA|immune induced molecule 14', map_location='57B3-57B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20870392, end_position_on_the_genomic_accession=20870678, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885606, CurrentID=0, Status='live', Symbol='CG34043', Aliases='Dmel_ BP1017, Dmel\\CG34043', description='CG34043 gene product from transcript CG34043-RB', other_designations='CG34043-PB|uncharacterized protein', map_location='31B1-31C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10278951, end_position_on_the_genomic_accession=10280872, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3885607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885607, CurrentID=0, Status='live', Symbol='CG33995', Aliases='Dmel_ Dmel\\CG33995, Dromel_CG31919_FBtr0079006_uORF', description='CG33995 gene product from transcript CG33995-RB', other_designations='CG33995-PA|CG33995-PB|CG33995-PC|uncharacterized protein', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5043270, end_position_on_the_genomic_accession=5048365, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3885610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885610, CurrentID=0, Status='live', Symbol='CG34007', Aliases='Dmel_ BP1020, Dmel\\CG34007', description='CG34007 gene product from transcript CG34007-RB', other_designations='CG34007-PA|CG34007-PB|uncharacterized protein', map_location='38C8-38C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20495030, end_position_on_the_genomic_accession=20495758, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885611, CurrentID=0, Status='live', Symbol='NimB3', Aliases='Dmel_CG34003, BP1035, CG16882, CG34003, Dmel\\CG34003, nimB3', description='Nimrod B3', other_designations='CG34003-PA|NimB3-PA|nimrod B3', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13967464, end_position_on_the_genomic_accession=13967985, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885612, CurrentID=0, Status='live', Symbol='ppk16', Aliases='Dmel_CG34059, CG34059, CG4110, Dmel\\CG34059, PPK16', description='pickpocket 16', other_designations='CG34059-PA|pickpocket16|ppk16-PA', map_location='30C8-30C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9771155, end_position_on_the_genomic_accession=9773044, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  117294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117294, CurrentID=0, Status='live', Symbol='Dsp1', Aliases='Dmel_CG12223, CG12223, DSP1, Dmel\\CG12223, Hmg14B, Hmg14D, Ssrp2, dsp1, hmg, ssrp2', description='Dorsal switch protein 1', other_designations='CG12223-PA|CG12223-PB|CG12223-PC|CG12223-PD|CG12223-PE|CG12223-PF|CG12223-PG|Dsp1-PA|Dsp1-PB|Dsp1-PC|Dsp1-PD|Dsp1-PE|Dsp1-PF|Dsp1-PG|Hmg14B|dorsal switch protein 1|dorsal switch protein-1|structure specific recognition protein 2', map_location='14C1-14B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16333612, end_position_on_the_genomic_accession=16340624, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  3885616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885616, CurrentID=0, Status='live', Symbol='CG17140', Aliases='Dmel_ CG31722, CG31722-B, Dmel\\CG17140', description='CG17140 gene product from transcript CG17140-RA', other_designations='CG17140-PA|CG17140-PB|uncharacterized protein', map_location='32B1-32B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10843014, end_position_on_the_genomic_accession=10845664, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3885617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885617, CurrentID=0, Status='live', Symbol='ppk10', Aliases='Dmel_CG34042, BP1007, CG34042, Dmel\\CG34042, PPK10', description='pickpocket 10', other_designations='CG34042-PB|ppk10-PB', map_location='31E3-31E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10450134, end_position_on_the_genomic_accession=10452396, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12798514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798514, CurrentID=0, Status='live', Symbol='CG43091', Aliases='Dmel_ Dmel\\CG43091', description='CG43091 gene product from transcript CG43091-RA', other_designations='CG43091-PA|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22768121, end_position_on_the_genomic_accession=22768733, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885619, CurrentID=0, Status='live', Symbol='CG34011', Aliases='Dmel_ BP1004, Dmel\\CG34011', description='CG34011 gene product from transcript CG34011-RC', other_designations='CG34011-PC|uncharacterized protein', map_location='25F3-25F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5716701, end_position_on_the_genomic_accession=5718067, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  117300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117300, CurrentID=0, Status='live', Symbol='clt', Aliases='Dmel_CG9858, CG9858, Dmel\\CG9858, Est-R, anon-57Fb', description='cricklet', other_designations='CG9858-PA|JHE-related|clt-PA', map_location='2-96 cM', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21617061, end_position_on_the_genomic_accession=21619079, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12798517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798517, CurrentID=0, Status='live', Symbol='CG43094', Aliases='Dmel_ Dmel\\CG43094', description='CG43094 gene product from transcript CG43094-RA', other_designations='CG43094-PA|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22770613, end_position_on_the_genomic_accession=22770987, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12798518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798518, CurrentID=0, Status='live', Symbol='CG43095', Aliases='Dmel_ Dmel\\CG43095', description='CG43095 gene product from transcript CG43095-RA', other_designations='CG43095-PA|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22771156, end_position_on_the_genomic_accession=22771770, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885627, CurrentID=0, Status='live', Symbol='CG34010', Aliases='Dmel_ BP1022, Dmel\\CG34010', description='CG34010 gene product from transcript CG34010-RA', other_designations='CG34010-PA|uncharacterized protein', map_location='28D3-28D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8009603, end_position_on_the_genomic_accession=8010652, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3885628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885628, CurrentID=0, Status='live', Symbol='Ir62a', Aliases='Dmel_CG33971, CG12090, CG33971, DmelIR62a, Dmel\\CG33971, IR62a', description='Ionotropic receptor 62a', other_designations='CG33971-PA|Ir62a-PA|ionotropic receptor 62a', map_location='62A3-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1529439, end_position_on_the_genomic_accession=1531259, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3885632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885632, CurrentID=0, Status='live', Symbol='CG34012', Aliases='Dmel_ BP1014, Dmel\\CG34012', description='CG34012 gene product from transcript CG34012-RA', other_designations='CG34012-PA|uncharacterized protein', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11108820, end_position_on_the_genomic_accession=11109505, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3885633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885633, CurrentID=0, Status='live', Symbol='CG34025', Aliases='Dmel_ BP1082, Dmel\\CG34025, HDC08169', description='CG34025 gene product from transcript CG34025-RB', other_designations='CG34025-PB|uncharacterized protein', map_location='63A2-63A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2959722, end_position_on_the_genomic_accession=2961049, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12798530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798530, CurrentID=0, Status='live', Symbol='CG43403', Aliases='Dmel_ Dmel\\CG43403', description='CG43403 gene product from transcript CG43403-RA', other_designations='CG43403-PA|uncharacterized protein', map_location='23A1-23A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2620187, end_position_on_the_genomic_accession=2620782, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3885636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885636, CurrentID=0, Status='live', Symbol='CG34030', Aliases='Dmel_ BP1084, Dmel\\CG34030, HDC08855', description='CG34030 gene product from transcript CG34030-RA', other_designations='CG34030-PA|uncharacterized protein', map_location='65E3-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7073476, end_position_on_the_genomic_accession=7075247, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3885637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885637, CurrentID=0, Status='live', Symbol='CG33993', Aliases='Dmel_ CG13289, CG13290, Dmel\\CG33993', description='CG33993 gene product from transcript CG33993-RA', other_designations='CG33993-PA|CG33993-PB|uncharacterized protein', map_location='65A1-65A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5933036, end_position_on_the_genomic_accession=5950658, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3885638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885638, CurrentID=0, Status='live', Symbol='CG33986', Aliases='Dmel_ CG17014, CG7973, Dmel\\CG33986, FBgn0036507', description='CG33986 gene product from transcript CG33986-RB', other_designations='CG33986-PB|uncharacterized protein', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15524325, end_position_on_the_genomic_accession=15525253, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885639, CurrentID=0, Status='live', Symbol='CG33985', Aliases='Dmel_ CG17014, CG7973, Dmel\\CG33985, FBgn0036507', description='CG33985 gene product from transcript CG33985-RA', other_designations='CG33985-PA|uncharacterized protein', map_location='71D4-71D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15525891, end_position_on_the_genomic_accession=15526812, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3885640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885640, CurrentID=0, Status='live', Symbol='CG34001', Aliases='Dmel_ BP1041, Dmel\\CG34001', description='CG34001 gene product from transcript CG34001-RB', other_designations='CG34001-PB|CG34001-PC|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10631079, end_position_on_the_genomic_accession=10634612, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885642, CurrentID=0, Status='live', Symbol='CG33966', Aliases='Dmel_ CG1211, Dmel\\CG33966', description='CG33966 gene product from transcript CG33966-RA', other_designations='CG33966-PA|uncharacterized protein', map_location='61F4-61F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1234130, end_position_on_the_genomic_accession=1235420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885643, CurrentID=0, Status='live', Symbol='CG34039', Aliases='Dmel_ BP1005, Dmel\\CG34039', description='CG34039 gene product from transcript CG34039-RB', other_designations='CG34039-PB|uncharacterized protein', map_location='70E2-70E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14560000, end_position_on_the_genomic_accession=14562613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3885644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885644, CurrentID=0, Status='live', Symbol='CG34002', Aliases='Dmel_ BP1025, Dmel\\CG34002, MIP08535', description='CG34002 gene product from transcript CG34002-RB', other_designations='CG34002-PB|uncharacterized protein', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17858474, end_position_on_the_genomic_accession=17860482, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885645, CurrentID=0, Status='live', Symbol='CG34057', Aliases='Dmel_ CG13904, Dmel\\CG34057, dC1GalT6', description='CG34057 gene product from transcript CG34057-RA', other_designations='CG34057-PA|CG34057-PB|uncharacterized protein', map_location='61D4-61D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=897380, end_position_on_the_genomic_accession=898772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3885646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885646, CurrentID=0, Status='live', Symbol='CG33978', Aliases='Dmel_ Dmel\\CG33978, FBgn0053978', description='CG33978 gene product from transcript CG33978-RD', other_designations='CG33978-PC|CG33978-PD|CG33978-PE|CG33978-PF|CG33978-PG|uncharacterized protein', map_location='102A4-102A6', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=157575, end_position_on_the_genomic_accession=175639, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  3885647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885647, CurrentID=0, Status='live', Symbol='Mitf', Aliases='Dmel_CG43369, CG17469, CG40476, CG41130, CG43369, Dmel\\CG43369, Dmel_CG17469, Dmel_CG41130, MITF, d', description='CG43369 gene product from transcript CG43369-RA', other_designations='CG43369-PA|CG43369-PB|CG43369-PC|CG43369-PD|Mitf-PA|Mitf-PB|Mitf-PC|Mitf-PD|mitf', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1198852, end_position_on_the_genomic_accession=1224467, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  3885648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885648, CurrentID=0, Status='live', Symbol='CG34028', Aliases='Dmel_ BP1074, Dmel\\CG34028', description='CG34028 gene product from transcript CG34028-RA', other_designations='CG34028-PA|uncharacterized protein', map_location='8E4-8E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9511524, end_position_on_the_genomic_accession=9512162, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3885650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885650, CurrentID=0, Status='live', Symbol='Cp7Fa', Aliases='Dmel_CG33962, A, CG15348, CG15349, CG33962, Dmel\\CG33962', description='Chorion protein a at 7F', other_designations='CG33962-PA|Cp7Fa-PA|chorion protein a at 7F', map_location='7F1-7F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8470888, end_position_on_the_genomic_accession=8473087, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  117331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117331, CurrentID=0, Status='live', Symbol='VepD', Aliases='Dmel_CG33200, CG33200, D, Dmel\\CG33200, anon-EST:CL57, ventrally-expressed-protein-D, vepD', description='Ventrally-expressed protein D', other_designations='CG33200-PA|CG33200-PB|CG33200-PC|VepD-PA|VepD-PB|VepD-PC|ventrally expressed protein-D|ventrally-expressed-protein-D', map_location='58D7-58D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22408525, end_position_on_the_genomic_accession=22409342, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117332, CurrentID=0, Status='live', Symbol='Rheb', Aliases='Dmel_CG1081, CG1081, DReb, Dm Rheb, DmRheb, Dmel\\CG1081, Q9VND8, RHEB, RheB, dRheb, rheb', description='Ras homolog enriched in brain ortholog (H. sapiens)', other_designations='CG1081-PA|CG1081-PB|RAS homolog enriched in brain|Ras homolog enriched in brain ortholog|Ras homologue enriched in brain|Ras homology enriched in brain|Rheb-PA|Rheb-PB', map_location='83B2-83B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5568921, end_position_on_the_genomic_accession=5570491, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3885654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885654, CurrentID=0, Status='live', Symbol='CG33981', Aliases='Dmel_ APC13, BEST:GH01256, Dmel\\CG33981', description='CG33981 gene product from transcript CG33981-RB', other_designations='BEST:GH01256|CG33981-PA|CG33981-PB|CG33981-PC|uncharacterized protein', map_location='54E9-54E9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17773964, end_position_on_the_genomic_accession=17781930, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3885655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885655, CurrentID=0, Status='live', Symbol='CG34040', Aliases='Dmel_ BP1015, Dmel\\CG34040', description='CG34040 gene product from transcript CG34040-RA', other_designations='CG34040-PA|uncharacterized protein', map_location='58A2-58A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21832501, end_position_on_the_genomic_accession=21833551, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117336, CurrentID=0, Status='live', Symbol='Gr98c', Aliases='Dmel_CG31060, 98B.3, CG12884, CG31060, Dmel\\CG31060, GR98B.3, GR98c, Gr98A1, Gr98A2, Gr98B3', description='Gustatory receptor 98c', other_designations='CG31060-PA|Gr98c-PA|gustatory receptor 98A2|gustatory receptor 98B.3|gustatory receptor 98c', map_location='98A4-98A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27544443, end_position_on_the_genomic_accession=27545836, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  117338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117338, CurrentID=0, Status='live', Symbol='Gr97a', Aliases='Dmel_CG33083, 97D.1, CG31073, CG31074, CG33083, Dmel\\CG33083, GR97D.1, GR97a, Gr97D1', description='Gustatory receptor 97a', other_designations='CG33083-PA|CG33083-PB|Gr97a-PA|Gr97a-PB|gustatory receptor 97D.1|gustatory receptor 97a', map_location='97D11-97D11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26965124, end_position_on_the_genomic_accession=26966485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117339, CurrentID=0, Status='live', Symbol='Gr94a', Aliases='Dmel_CG31280, 94E.1, CG31280, Dmel\\CG31280, GR94E.1, GR94a, Gr94E1', description='Gustatory receptor 94a', other_designations='CG31280-PA|Gr94a-PA|gustatory receptor 94E.1|gustatory receptor 94a', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23051689, end_position_on_the_genomic_accession=23052903, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3885660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3885660, CurrentID=0, Status='live', Symbol='CG34051', Aliases='Dmel_ BP1077, Dmel\\CG34051, HDC03337', description='CG34051 gene product from transcript CG34051-RB', other_designations='CG34051-PA|CG34051-PB|uncharacterized protein', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19960069, end_position_on_the_genomic_accession=19961030, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12798557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798557, CurrentID=0, Status='live', Symbol='CG43056', Aliases='Dmel_ Dmel\\CG43056', description='CG43056 gene product from transcript CG43056-RA', other_designations='CG43056-PA|uncharacterized protein', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4437787, end_position_on_the_genomic_accession=4438201, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117342, CurrentID=0, Status='live', Symbol='CG32395', Aliases='Dmel_ 65C.1, Dmel\\CG32395, GR65C.1, Gr65C1, Gr65a', description='CG32395 gene product from transcript CG32395-RB', other_designations='CG32395-PB|Gr65a|gustatory receptor 65C.1|uncharacterized protein', map_location='65C1-65C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6666638, end_position_on_the_genomic_accession=6667816, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117343, CurrentID=0, Status='live', Symbol='Gr59d', Aliases='Dmel_CG30330, 59D.1, 59D1, CG13543, CG30330, Dmel\\CG30330, GR59D.1, GR59d, Gr59D1', description='Gustatory receptor 59d', other_designations='CG30330-PA|Gr59d-PA|gustatory receptor 59D.1|gustatory receptor 59d', map_location='59C4-59C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23103941, end_position_on_the_genomic_accession=23105169, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117344, CurrentID=0, Status='live', Symbol='Gr58a', Aliases='Dmel_CG30396, 58A.1, 58A1, CG30396, Dmel\\CG30396, GR58A.1, Gr58A1', description='Gustatory receptor 58a', other_designations='CG30396-PA|Gr58a-PA|gustatory receptor 58A.1|gustatory receptor 58a', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21979860, end_position_on_the_genomic_accession=21981099, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117345, CurrentID=0, Status='live', Symbol='Gr47b', Aliases='Dmel_CG30030, 47F.1, CG30030, Dmel\\CG30030, GR47F.1, GR47b, Gr47F1', description='Gustatory receptor 47b', other_designations='CG30030-PA|Gr47b-PA|gustatory receptor 47F.1|gustatory receptor 47b', map_location='47F5-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11361900, end_position_on_the_genomic_accession=11363405, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117346, CurrentID=0, Status='live', Symbol='Gr39a', Aliases='Dmel_CG31622, 39D.2, 39D.2a, 39D.2b, 39D.2c, 39D.2d, 39D2a, 39D2b, 39D2c, 39D2d, CG31622, Dmel\\CG31622, GR39D.2, GR39a, Gr39D2A, Gr39D2a, Gr39D2b, Gr39D2c, Gr39D2d.a, Gr39a.b, Gr39a.c, Gr39a.d, Gr39aA, Gr39aB, Gr39aC, Gr39aD, gene Y, l(2)k05106, Gr39a', description='Gustatory receptor 39a', other_designations='CG31622-PA|CG31622-PB|CG31622-PC|CG31622-PD|Gr39a-PA|Gr39a-PB|Gr39a-PC|Gr39a-PD|gustatory receptor 39D.2|gustatory receptor 39a|lethal (2) k05106', map_location='39C1-39C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21290734, end_position_on_the_genomic_accession=21297686, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  117347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117347, CurrentID=0, Status='live', Symbol='Gr39b', Aliases='Dmel_CG31620, 39D.1, 39D1, CG31620, Dmel\\CG31620, GR39D.1, GR39b, Gr39D1', description='Gustatory receptor 39b', other_designations='CG31620-PA|Gr39b-PA|gustatory receptor 39D.1|gustatory receptor 39b', map_location='39D2-39D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21367908, end_position_on_the_genomic_accession=21369706, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  117348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117348, CurrentID=0, Status='live', Symbol='Gr28a', Aliases='Dmel_CG13787, 28A.1, CG13787, Dmel\\CG13787, GR28A.1, GR28a, Gr28A1, Gr28A2', description='Gustatory receptor 28a', other_designations='CG13787-PA|Gr28a-PA|gustatory receptor 28A.1|gustatory receptor 28A2|gustatory receptor 28a', map_location='28A1-28A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7452520, end_position_on_the_genomic_accession=7454212, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  117349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117349, CurrentID=0, Status='live', Symbol='Gr22f', Aliases='Dmel_CG31932, 22B.1, CG31932, Dmel\\CG31932, GR22B.1, GR22b1, GR22f, Gr22B1', description='Gustatory receptor 22f', other_designations='CG31932-PA|Gr22f-PA|gustatory receptor 22B.1|gustatory receptor 22f', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1755253, end_position_on_the_genomic_accession=1756451, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117353, CurrentID=0, Status='live', Symbol='Cct1', Aliases='Dmel_CG1049, CCT1, CG1049, CT41283, Cct, Dmel\\CG1049, cct1', description='CTP:phosphocholine cytidylyltransferase 1', other_designations='CG1049-PA|CG1049-PB|CG1049-PC|CG1049-PD|CG1049-PE|CG1049-PF|CTP:phosphocholine cytidylyltransferase|Cct1-PA|Cct1-PB|Cct1-PC|Cct1-PD|Cct1-PE|Cct1-PF|cytidylyltransferase 1|phosphocholine cytidylyltransferase', map_location='62A3-62A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1546092, end_position_on_the_genomic_accession=1553504, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12798570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798570, CurrentID=0, Status='live', Symbol='CG43185', Aliases='Dmel_ Dmel\\CG43185', description='CG43185 gene product from transcript CG43185-RD', other_designations='CG43185-PC|CG43185-PD|uncharacterized protein', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5884888, end_position_on_the_genomic_accession=5885491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12798574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798574, CurrentID=0, Status='live', Symbol='CG43389', Aliases='Dmel_ Dmel\\CG43389', description='CG43389 gene product from transcript CG43389-RA', other_designations='CG43389-PA|uncharacterized protein', map_location='63C1-63C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3248015, end_position_on_the_genomic_accession=3249044, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117361, CurrentID=0, Status='live', Symbol='FASN2', Aliases='Dmel_CG3524, CG3524, DM_7295849, Dmel\\CG3524, FAS, FAS-mb, FASN[CG3524], FAS[CG3524], mFAS, v(2)k05816', description='Fatty acid synthase 2', other_designations='CG3524-PA|CG3524-PB|FASN2-PA|FASN2-PB|fatty acid synthase 2|fatty acid synthase, methyl-branched specific|mbCHC-specific FAS|methyl-branched specific FAS|v(2)k05816', map_location='23C5-23C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3043947, end_position_on_the_genomic_accession=3056143, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12798579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798579, CurrentID=0, Status='live', Symbol='CG43074', Aliases='Dmel_ Dmel\\CG43074', description='CG43074 gene product from transcript CG43074-RA', other_designations='CG43074-PA|uncharacterized protein', map_location='63E2-63E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3483832, end_position_on_the_genomic_accession=3484653, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  117364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117364, CurrentID=0, Status='live', Symbol='Hex-t1', Aliases='Dmel_CG33102, CG33102, CG5443, CG54430, DHK-453, Dmel\\CG33102, HEX-t1, HK, Hex', description='CG33102 gene product from transcript CG33102-RA', other_designations='CG33102-PA|Hex-t1-PA|Hexokinase-t1|hexokinase', map_location='97B2-97B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26449079, end_position_on_the_genomic_accession=26450585, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  117365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117365, CurrentID=0, Status='live', Symbol='HBS1', Aliases='Dmel_CG1898, CG1898, Dmel\\CG1898, Hbs1, hbs1', description='CG1898 gene product from transcript CG1898-RB', other_designations='CG1898-PA|CG1898-PB|HBS1-PA|HBS1-PB', map_location='62B7-62B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1862881, end_position_on_the_genomic_accession=1865240, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  117366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117366, CurrentID=0, Status='live', Symbol='Gpi1', Aliases='Dmel_CG32578, CG32578, Dm_CAC10056/216-441, Dmel\\CG32578, d i19e, l(1)14Ca, l(1)14Cd, l(1)19-134, l(1)i19, l(1)i19e', description='CG32578 gene product from transcript CG32578-RC', other_designations='CG32578-PB|CG32578-PC|Gpi1-PB|Gpi1-PC|lethal (1) 14Ca', map_location='14C1-14C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16359143, end_position_on_the_genomic_accession=16361275, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12798583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798583, CurrentID=0, Status='live', Symbol='CG43232', Aliases='Dmel_ Dmel\\CG43232', description='CG43232 gene product from transcript CG43232-RB', other_designations='CG43232-PA|CG43232-PB|uncharacterized protein', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6932121, end_position_on_the_genomic_accession=6932954, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117368, CurrentID=0, Status='live', Symbol='prt', Aliases='Dmel_CG10251, CG10251, Dmel\\CG10251', description='portabella', other_designations='CG10251-PB|prt-PB', map_location='95A7-95A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23696372, end_position_on_the_genomic_accession=23699027, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  117369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117369, CurrentID=0, Status='live', Symbol='Desat1', Aliases='Dmel_CG5887, BEST:SD05462, CG5887, Dmel\\CG5887, Fad, anon-WO0118547.726, anon-WO0118547.730, desat, desat1, ds1, fad, l(3)S028813', description='Desaturase 1', other_designations='CG5887-PA|CG5887-PB|CG5887-PC|CG5887-PD|CG5887-PE|Desat1-PA|Desat1-PB|Desat1-PC|Desat1-PD|Desat1-PE|desat1|desaturase|desaturase 1|desaturase1|fatty acid desaturase|lethal (3) S028813|stearoyl CoA desaturase|stearoyl-CoA 9-desaturase', map_location='87B10-87B11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12439600, end_position_on_the_genomic_accession=12447616, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12798588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798588, CurrentID=0, Status='live', Symbol='CG43089', Aliases='Dmel_ Dmel\\CG43089, lincRNA.427', description='CG43089 gene product from transcript CG43089-RC', other_designations='CG43089-PB|CG43089-PC|uncharacterized protein', map_location='63E8-63E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3714027, end_position_on_the_genomic_accession=3714717, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117406, CurrentID=0, Status='live', Symbol='CG32396', Aliases='Dmel_ CG8339, CT42364, Dmel\\CG32396', description='CG32396 gene product from transcript CG32396-RA', other_designations='CG32396-PA|uncharacterized protein', map_location='65B4-65B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6530729, end_position_on_the_genomic_accession=6532429, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117407, CurrentID=0, Status='live', Symbol='Tsp68C', Aliases='Dmel_CG32136, CG32136, Dm.Tsp68C, Dmel\\CG32136', description='Tetraspanin 68C', other_designations='CG32136-PA|CG32136-PC|Tsp68C-PA|Tsp68C-PC|tetraspanin 68C', map_location='70C4-70C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13843656, end_position_on_the_genomic_accession=13845316, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462474, CurrentID=0, Status='live', Symbol='CG44094', Aliases='Dmel_ Dmel\\CG44094', description='CG44094 gene product from transcript CG44094-RA', other_designations='CG44094-PA|uncharacterized protein', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14576331, end_position_on_the_genomic_accession=14576611, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674029, CurrentID=0, Status='live', Symbol='CG42561', Aliases='Dmel_ CG14481, Dmel\\CG42561, Dmel_CG14481', description='CG42561 gene product from transcript CG42561-RA', other_designations='CG42561-PA|uncharacterized protein', map_location='54D5-54D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17684855, end_position_on_the_genomic_accession=17685548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117414, CurrentID=0, Status='live', Symbol='mRpS14', Aliases='Dmel_CG32531, CG12211, CG32531, Dmel\\CG32531, MRP-S14, MrpS14, S14, cg12211', description='mitochondrial ribosomal protein S14', other_designations='CG32531-PA|CG32531-PB|mRpS14-PA|mRpS14-PB', map_location='18C7-18C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19479382, end_position_on_the_genomic_accession=19480238, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26067055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067055, CurrentID=0, Status='live', Symbol='PRY', Aliases='Dmel_CG45787, CG40930, CG41367, CG42619, CG45787, Dmel\\CG45787, Dmel_CG40930, Dmel_CG41367, Dmel_CG42619', description='polycystine-related-Y', other_designations='CG45787-PA|PRY-PA', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=146092, end_position_on_the_genomic_accession=192020, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  117419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117419, CurrentID=0, Status='live', Symbol='Irbp', Aliases='Dmel_CG5247, CG5247, DmKu70, Dmel\\CG5247, IRBP, IRBP/Ku70, Ku, Ku70, YPF1, YPF1beta, Ypf1b, dp70, irbp, ku70, p70', description='Inverted repeat-binding protein', other_designations='CG5247-PA|CG5247-PB|Irbp-PA|Irbp-PB|inverted repeat-binding protein|yolk protein factor 1b', map_location='86E4-86E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11410047, end_position_on_the_genomic_accession=11412440, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768834, CurrentID=0, Status='live', Symbol='CG33467', Aliases='Dmel_ Dmel\\CG33467', description='CG33467 gene product from transcript CG33467-RA', other_designations='CG33467-PA|CG33467-PB|uncharacterized protein', map_location='51D9-51D9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15055503, end_position_on_the_genomic_accession=15056241, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  117459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117459, CurrentID=0, Status='live', Symbol='TotZ', Aliases='Dmel_CG31507, CG31507, Dmel\\CG31507, Totz, tot', description='Turandot Z', other_designations='CG31507-PA|TotZ-PA|turandot Z', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20877736, end_position_on_the_genomic_accession=20878363, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117460, CurrentID=0, Status='live', Symbol='TotX', Aliases='Dmel_CG31193, CG31193, Dmel\\CG31193, TOTX, Tot X, tot', description='Turandot X', other_designations='CG31193-PA|TotX-PA|turandot|turandot X', map_location='93A3-93A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20904917, end_position_on_the_genomic_accession=20905518, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  117461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117461, CurrentID=0, Status='live', Symbol='TotF', Aliases='Dmel_CG31691, CG13963, CG31691, Dmel\\CG31691, Tot F, tot', description='Turandot F', other_designations='CG31691-PA|TotF-PA|turandot F', map_location='38A7-38A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19911815, end_position_on_the_genomic_accession=19912377, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117462, CurrentID=0, Status='live', Symbol='TotC', Aliases='Dmel_CG31508, CG31508, Dmel\\CG31508, TOTc, tot, totC', description='Turandot C', other_designations='CG31508-PA|Stress-inducible humoral factor turandot C|TotC-PA|turandot|turandot C', map_location='93A2-93A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20872988, end_position_on_the_genomic_accession=20873580, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117463, CurrentID=0, Status='live', Symbol='Ste12DOR', Aliases='Dmel_CG32616, CG32616, Dmel\\CG32616, SC-1, Ste, SteDOR', description='Stellate 12D orphon', other_designations='CG32616-PB|Ste12DOR-PB|stellate 12D orphon|stellate orphon', map_location='12D2-12D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14000210, end_position_on_the_genomic_accession=14000889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117464, CurrentID=0, Status='live', Symbol='He', Aliases='Dmel_CG31770, CG31770, Dmel\\CG31770, H2, he, hem', description='Hemese', other_designations='CG31770-PB|CG31770-PC|He-PB|He-PC|gustav|hemes|hemese', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13972096, end_position_on_the_genomic_accession=13972947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117467, CurrentID=0, Status='live', Symbol='mthl11', Aliases='Dmel_CG31147, CG31147, Dmel\\CG31147', description='methuselah-like 11', other_designations='CG31147-PC|CG31147-PD|Mth-like 11|Mth-like-11|mthl11-PC|mthl11-PD', map_location='86B3-86B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10401032, end_position_on_the_genomic_accession=10402966, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  117470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117470, CurrentID=0, Status='live', Symbol='Gr93d', Aliases='Dmel_CG31335, 93F.4, CG31335, DmGr93d, Dmel\\CG31335, GR93F.4, GR93d, Gr93F4', description='Gustatory receptor 93d', other_designations='CG31335-PA|Gr93d-PA|gustatory receptor 93F.4|gustatory receptor 93d', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21845454, end_position_on_the_genomic_accession=21846659, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117471, CurrentID=0, Status='live', Symbol='Gr93c', Aliases='Dmel_CG31173, 93F.3, CG13411, CG31173, Dmel\\CG31173, GR93F.3, GR93c, Gr93F3', description='Gustatory receptor 93c', other_designations='CG31173-PA|Gr93c-PA|gustatory receptor 93F.3|gustatory receptor 93c', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21844014, end_position_on_the_genomic_accession=21845265, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117472, CurrentID=0, Status='live', Symbol='Gr93b', Aliases='Dmel_CG31336, 93F.2, CG13411, CG31336, Dmel\\CG31336, GR93F.2, GR93b, Gr93F2', description='Gustatory receptor 93b', other_designations='CG31336-PB|Gr93b-PB|gustatory receptor 93F.2|gustatory receptor 93b', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21842521, end_position_on_the_genomic_accession=21843765, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117473, CurrentID=0, Status='live', Symbol='Gr92a', Aliases='Dmel_CG31208, CG31208, DmGr92a, Dmel\\CG31208, GR92a, Gr92D1', description='Gustatory receptor 92a', other_designations='CG31208-PA|Gr92a-PA|gustatory receptor 92D1|gustatory receptor 92a', map_location='92D1-92D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20121060, end_position_on_the_genomic_accession=20122277, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117475, CurrentID=0, Status='live', Symbol='Gr85a', Aliases='Dmel_CG31405, CG31405, Dmel\\CG31405', description='Gustatory receptor 85a', other_designations='CG31405-PA|Gr85a-PA|gustatory receptor 85a', map_location='85F5-85F5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9995544, end_position_on_the_genomic_accession=9996789, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117476, CurrentID=0, Status='live', Symbol='Gr77a', Aliases='Dmel_CG32433, 77E.1, CG32433, Dmel\\CG32433, GR77E.1, GR77a, Gr77E1', description='Gustatory receptor 77a', other_designations='CG32433-PA|Gr77a-PA|gustatory receptor 77E.1|gustatory receptor 77a', map_location='77E4-77E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20737569, end_position_on_the_genomic_accession=20739430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117477, CurrentID=0, Status='live', Symbol='Gr64f', Aliases='Dmel_CG32255, CG14988, CG32255, DmGr64f, Dmel64f, Dmel\\CG32255, GR64f, Gr64, Gr64A3', description='Gustatory receptor 64f', other_designations='CG32255-PA|Gr64f-PA|gustatory receptor 64A3|gustatory receptor 64f', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4035799, end_position_on_the_genomic_accession=4037575, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  117479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117479, CurrentID=0, Status='live', Symbol='Gr64c', Aliases='Dmel_CG32256, CG14986, CG32256, Dmel64c, Dmel\\CG32256, GR64c, Gr64, gr64c', description='Gustatory receptor 64c', other_designations='CG32256-PA|CG32256-PB|Gr64c-PA|Gr64c-PB|gustatory receptor 64c', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4028605, end_position_on_the_genomic_accession=4032032, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  117480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117480, CurrentID=0, Status='live', Symbol='Gr64b', Aliases='Dmel_CG32257, CG14986, CG32257, Dmel64b, Dmel\\CG32257, GR64b, Gr64, anon-WO0118547.604', description='Gustatory receptor 64b', other_designations='CG32257-PA|Gr64b-PA|gustatory receptor 64b', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4028605, end_position_on_the_genomic_accession=4032032, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  117481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117481, CurrentID=0, Status='live', Symbol='Gr64a', Aliases='Dmel_CG32261, CG14986, CG32261, Dmel64a, Dmel\\CG32261, GR64a, GRLU.3, Gr64, Gr64A1, GrLU3, LU.3', description='Gustatory receptor 64a', other_designations='CG32261-PA|Gr64a-PA|gustatory receptor 64a|gustatory receptor LU.3', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4026769, end_position_on_the_genomic_accession=4028600, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  117484 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117484, CurrentID=0, Status='live', Symbol='Gr59b', Aliases='Dmel_CG30191, CG13537, CG30191, Dmel\\CG30191, GR59C.2, GR59b, Gr59C1, Gr59D3', description='Gustatory receptor 59b', other_designations='CG30191-PA|Gr59b-PA|gustatory receptor 59C1|gustatory receptor 59D3|gustatory receptor 59b', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23082203, end_position_on_the_genomic_accession=23083366, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117485, CurrentID=0, Status='live', Symbol='Gr59a', Aliases='Dmel_CG30189, CG13537, CG30189, Dmel\\CG30189, GR59C.1, GR59a', description='Gustatory receptor 59a', other_designations='CG30189-PA|Gr59a-PA|gustatory receptor 59a', map_location='59C3-59C3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23080561, end_position_on_the_genomic_accession=23081725, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117486, CurrentID=0, Status='live', Symbol='Gr36c', Aliases='Dmel_CG31748, CG15137, CG31748, Dmel\\CG31748, GR36c, Gr36B.3, Gr36B3', description='Gustatory receptor 36c', other_designations='CG31748-PA|Gr36c-PA|gustatory receptor 36B3|gustatory receptor 36c', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17180338, end_position_on_the_genomic_accession=17181632, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117487, CurrentID=0, Status='live', Symbol='Gr36b', Aliases='Dmel_CG31744, CG15137, CG31744, Dmel\\CG31744, GR36b, Gr36B.2, Gr36B2', description='Gustatory receptor 36b', other_designations='CG31744-PA|Gr36b-PA|gustatory receptor 36B2|gustatory receptor 36b', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17178772, end_position_on_the_genomic_accession=17180064, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117488, CurrentID=0, Status='live', Symbol='Gr36a', Aliases='Dmel_CG31747, CG31747, Dmel\\CG31747, GR36a, GRLU.1, Gr36B1, GrLU1, LU.1', description='Gustatory receptor 36a', other_designations='CG31747-PA|Gr36a-PA|gustatory receptor 36a|gustatory receptor LU.1', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17176618, end_position_on_the_genomic_accession=17177921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  117492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117492, CurrentID=0, Status='live', Symbol='Gr22b', Aliases='Dmel_CG31931, CG31931, CR31931, Dmel\\CG31931, GR22bP, Gr2940.4, Gr22b', description='Gustatory receptor 22b', other_designations='CG31931-PA|Gr22b-PA|gustatory receptor 22b|gustatory receptor 2940.4', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1792057, end_position_on_the_genomic_accession=1793268, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117496, CurrentID=0, Status='live', Symbol='Gr28b', Aliases='Dmel_CG13788, 28A3, CG13788, Dmel\\CG13788, GR28B, GR28b, GRLU.2, Gr28A3, Gr28A4.a, Gr28b.b, Gr28b.c, Gr28b.d, Gr28b.e, Gr28bA, Gr28bB, Gr28bC, Gr28bD, Gr28bE, Gr28be, GrLU2, LU.2, gr28b, Gr28b', description='Gustatory receptor 28b', other_designations='CG13788-PA|CG13788-PB|CG13788-PC|CG13788-PD|CG13788-PE|Gr28b-PA|Gr28b-PB|Gr28b-PC|Gr28b-PD|Gr28b-PE|gustatory receptor 28A4|gustatory receptor 28b|gustatory receptor LU.2', map_location='28A1-28A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7454826, end_position_on_the_genomic_accession=7462247, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  117498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117498, CurrentID=0, Status='live', Symbol='Gr22e', Aliases='Dmel_CG31936, CG31936, Dmel\\CG31936, GR22e, Gr2940.1', description='Gustatory receptor 22e', other_designations='CG31936-PA|Gr22e-PA|gustatory receptor 22e|gustatory receptor 2940.1', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1784062, end_position_on_the_genomic_accession=1785286, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117499, CurrentID=0, Status='live', Symbol='Gr22c', Aliases='Dmel_CG31929, CG31929, Dmel\\CG31929, GR22c, GRLU.6, Gr2940.3, GrLU6, LU.6', description='Gustatory receptor 22c', other_designations='CG31929-PA|Gr22c-PA|gustatory receptor 22c|gustatory receptor 2940.3|gustatory receptor LU.6', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1790617, end_position_on_the_genomic_accession=1791823, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117500, CurrentID=0, Status='live', Symbol='Gr22a', Aliases='Dmel_CG31662, CG31662, Dmel\\CG31662, GR22a, Gr2940.5', description='Gustatory receptor 22a', other_designations='CG31662-PA|Gr22a-PA|gustatory receptor 22a|gustatory receptor 2940.5', map_location='22B2-22B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1793668, end_position_on_the_genomic_accession=1794917, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117501, CurrentID=0, Status='live', Symbol='Gr10a', Aliases='Dmel_CG32664, CG32664, Dmel\\CG32664, GR10a, Gr10B1, gr10a', description='Gustatory receptor 10a', other_designations='CG32664-PB|Gr10a-PB|gustatory receptor 10B1|gustatory receptor 10a', map_location='10B10-10B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11413467, end_position_on_the_genomic_accession=11417422, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  117503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=117503, CurrentID=0, Status='live', Symbol='RacGAP84C', Aliases='Dmel_CG2595, CG2595, Dmel\\CG2595, GAP84C, P1.7, RacGAP(84C), RhoGAP-84C8, racGAP, rnGAP, rnRac-GAP, rnRacGAP, rnRacGap, rnprot, rnprot1.7, rnracGAP, rnracGap', description='Rac GTPase activating protein at 84C', other_designations='CG2595-PA|CG2595-PB|RacGAP84C-PA|RacGAP84C-PB|RnRacGAP|RotundRACGAP|rotund RacGAP', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7285962, end_position_on_the_genomic_accession=7287629, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772140, CurrentID=0, Status='live', Symbol='CG14246', Aliases='Dmel_ CG32950, Dmel\\CG14246', description='CG14246 gene product from transcript CG14246-RC', other_designations='CG14246-PA|CG14246-PB|CG14246-PC|uncharacterized protein', map_location='97C5-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26727314, end_position_on_the_genomic_accession=26730117, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26067056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067056, CurrentID=0, Status='live', Symbol='WDY', Aliases='Dmel_CG45799, CG40245, CG40449, CG40551, CG40583, CG41020, CG45799, Dmel\\CG45799, Dmel_CG40245, Dmel_CG40551, Dmel_CG40583, Dmel_CG41020', description='WD40 Y', other_designations='CG45799-PC|WDY-PC', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=2273422, end_position_on_the_genomic_accession=2487777, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3772680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772680, CurrentID=0, Status='live', Symbol='CG33640', Aliases='Dmel_ Dmel\\CG33640', description='CG33640 gene product from transcript CG33640-RB', other_designations='CG33640-PB|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13527841, end_position_on_the_genomic_accession=13528537, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31196, CurrentID=0, Status='live', Symbol='CG3457', Aliases='Dmel_ Dmel\\CG3457, EG:17E2.1', description='CG3457 gene product from transcript CG3457-RB', other_designations='CG3457-PA|CG3457-PB', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2187820, end_position_on_the_genomic_accession=2188919, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740467, CurrentID=0, Status='live', Symbol='CG34220', Aliases='Dmel_ Dmel\\CG34220', description='CG34220 gene product from transcript CG34220-RA', other_designations='CG34220-PA|uncharacterized protein', map_location='46D7-46D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10041065, end_position_on_the_genomic_accession=10043507, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31197, CurrentID=0, Status='live', Symbol='CG14053', Aliases='Dmel_ BcDNA:AT10359, Dmel\\CG14053', description='CG14053 gene product from transcript CG14053-RB', other_designations='CG14053-PA|CG14053-PB|uncharacterized protein', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2191639, end_position_on_the_genomic_accession=2193909, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38836 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38836, CurrentID=0, Status='live', Symbol='Tsp66A', Aliases='Dmel_CG16991, CG1699, CG16991, Dm.Tsp66A, Dmel\\CG16991', description='Tetraspanin 66A', other_designations='CG16991-PC|CG16991-PD|CG16991-PE|CG16991-PF|Tsp66A-PC|Tsp66A-PD|Tsp66A-PE|Tsp66A-PF|tetraspanin 66A', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7465509, end_position_on_the_genomic_accession=7468808, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  318168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318168, CurrentID=0, Status='live', Symbol='CG32718', Aliases='Dmel_ Dmel\\CG32718', description='CG32718 gene product from transcript CG32718-RC', other_designations='CG32718-PB|CG32718-PC|CG32718-PD|uncharacterized protein', map_location='7B8-7B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7778205, end_position_on_the_genomic_accession=7779370, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318169, CurrentID=0, Status='live', Symbol='CG32719', Aliases='Dmel_ Dmel\\CG32719', description='CG32719 gene product from transcript CG32719-RA', other_designations='CG32719-PA|uncharacterized protein', map_location='7B3-7B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7508651, end_position_on_the_genomic_accession=7547992, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  38840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38840, CurrentID=0, Status='live', Symbol='Cyp316a1', Aliases='Dmel_CG8540, 316a1, CG8540, Dmel\\CG8540, cyp316a1', description='CG8540 gene product from transcript CG8540-RB', other_designations='CG8540-PB|Cyp316a1-PB', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7491864, end_position_on_the_genomic_accession=7493484, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12798516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=12798516, CurrentID=0, Status='live', Symbol='CG43093', Aliases='Dmel_ Dmel\\CG43093', description='CG43093 gene product from transcript CG43093-RA', other_designations='CG43093-PA|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22769775, end_position_on_the_genomic_accession=22770307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835655, CurrentID=0, Status='live', Symbol='CG44403', Aliases='Dmel_ Dmel\\CG44403', description='CG44403 gene product from transcript CG44403-RA', other_designations='CG44403-PA|uncharacterized protein', map_location='54D1-54D1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17616499, end_position_on_the_genomic_accession=17617895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674046, CurrentID=0, Status='live', Symbol='CG42498', Aliases='Dmel_ Dmel\\CG42498, Dromel_CG14550_FBtr0084974_uORF', description='CG42498 gene product from transcript CG42498-RA', other_designations='CG42498-PA|uncharacterized protein', map_location='96F10-96F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26057347, end_position_on_the_genomic_accession=26058223, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36038, CurrentID=0, Status='live', Symbol='CG12134', Aliases='Dmel_ Dmel\\CG12134', description='CG12134 gene product from transcript CG12134-RB', other_designations='CG12134-PA|CG12134-PB|uncharacterized protein', map_location='46C9-46C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9971143, end_position_on_the_genomic_accession=9972780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  249072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=249072, CurrentID=0, Status='live', Symbol='Liprin-alpha', Aliases='Dmel_CG11199, CG11199, DLiprin, D Dliprin-alpha, Dmel\\CG11199, LIP, LIP.1, LIP1, Liprin, lip-alpha, lpr, oos', description='CG11199 gene product from transcript CG11199-RC', other_designations='CG11199-PA|CG11199-PB|CG11199-PC|CG11199-PD|CG11199-PE|LAR-interacting protein 1|Liprin-a|Liprin-alpha-PA|Liprin-alpha-PB|Liprin-alpha-PC|Liprin-alpha-PD|Liprin-alpha-PE|dliprin|dliprin-alpha|liprin|liprin-alpha|out of step', map_location='27A1-27A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6723903, end_position_on_the_genomic_accession=6732766, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  318176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318176, CurrentID=0, Status='live', Symbol='CG32727', Aliases='Dmel_ Dmel\\CG32727, anon-SAGE:Wang-041', description='CG32727 gene product from transcript CG32727-RA', other_designations='CG32727-PA|anon-SAGE:Wang-041|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7407208, end_position_on_the_genomic_accession=7407854, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674051, CurrentID=0, Status='live', Symbol='Sfp33A2', Aliases='Dmel_CG42473, CG42473, Dmel\\CG42473', description='Seminal fluid protein 33A2', other_designations='CG42473-PA|Sfp33A2-PA|seminal fluid protein 33A2', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11837713, end_position_on_the_genomic_accession=11838049, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  249170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=249170, CurrentID=0, Status='live', Symbol='CG32236', Aliases='Dmel_ BcDNA:GH16763, Dmel\\CG32236, NEST:bs27g09', description='CG32236 gene product from transcript CG32236-RA', other_designations='CG32236-PA|uncharacterized protein', map_location='64B13-64B13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4665927, end_position_on_the_genomic_accession=4667389, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38849, CurrentID=0, Status='live', Symbol='CG14837', Aliases='Dmel_ Dmel\\CG14837', description='CG14837 gene product from transcript CG14837-RC', other_designations='CG14837-PC|uncharacterized protein', map_location='66A5-66A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7555780, end_position_on_the_genomic_accession=7560037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36669, CurrentID=0, Status='live', Symbol='CG10253', Aliases='Dmel_ ADPS, Dmel\\CG10253, dAGPS', description='CG10253 gene product from transcript CG10253-RB', other_designations='CG10253-PA|CG10253-PB|CG10253-PC|CG10253-PD|alkyldihydroxyacetonephosphate synthase|uncharacterized protein', map_location='51D3-51D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14929703, end_position_on_the_genomic_accession=14934861, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19835657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835657, CurrentID=0, Status='live', Symbol='SclA', Aliases='Dmel_CG45090, BEST:LP06756, BcDNA:RE28911, CG45090, CR31696, Dmel\\CG45090, ORF-A, mel1, pncr003, pncr003:2L, pncr003:2L ORFA, pncr:003:2L, scl', description='Sarcolamban A', other_designations='CG45090-PA|CG45090-PB|CG45090-PC|SclA-PA|SclA-PB|SclA-PC|sarcolamban|sarcolamban A', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16837972, end_position_on_the_genomic_accession=16841884, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  41581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41581, CurrentID=0, Status='live', Symbol='CG10909', Aliases='Dmel_ Dmel\\CG10909', description='CG10909 gene product from transcript CG10909-RA', other_designations='CG10909-PA|uncharacterized protein', map_location='87D4-87D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12910120, end_position_on_the_genomic_accession=12911292, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36121, CurrentID=0, Status='live', Symbol='CG11883', Aliases='Dmel_ Dmel\\CG11883', description='CG11883 gene product from transcript CG11883-RD', other_designations='CG11883-PA|CG11883-PB|CG11883-PC|CG11883-PD|uncharacterized protein', map_location='47B1-47B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10617662, end_position_on_the_genomic_accession=10641807, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  8674063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674063, CurrentID=0, Status='live', Symbol='CG42529', Aliases='Dmel_ Dmel\\CG42529', description='CG42529 gene product from transcript CG42529-RA', other_designations='CG42529-PA|uncharacterized protein', map_location='76C3-76C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19728994, end_position_on_the_genomic_accession=19730402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  118436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=118436, CurrentID=0, Status='live', Symbol='Ranbp16', Aliases='Dmel_CG33180, CG33180, CG9126, CG9136, Dmel\\CG33180, RanBP16', description='CG33180 gene product from transcript CG33180-RE', other_designations='CG33180-PB|CG33180-PC|CG33180-PE|CG33180-PF|CG33180-PG|CG33180-PH|Ranbp16-PB|Ranbp16-PC|Ranbp16-PE|Ranbp16-PF|Ranbp16-PG|Ranbp16-PH', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15927185, end_position_on_the_genomic_accession=15933867, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19893533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893533, CurrentID=0, Status='live', Symbol='COX1', Aliases='', description='cytochrome c oxidase subunit I', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=1474, end_position_on_the_genomic_accession=3009, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  170887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=170887, CurrentID=0, Status='live', Symbol='CG31750', Aliases='Dmel_ Dmel\\CG31750, Gr36d', description='CG31750 gene product from transcript CG31750-RA', other_designations='CG31750-PA|Gr36d|uncharacterized protein', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17181938, end_position_on_the_genomic_accession=17183654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36127, CurrentID=0, Status='live', Symbol='Rab3', Aliases='Dmel_CG7576, AAF58762, CG7576, DRAB3, DRab3, DmRab3, Dmel\\CG7576, Drab3, P25228A, dm-Rab3, drab3, rab3', description='CG7576 gene product from transcript CG7576-RA', other_designations='CG7576-PA|Rab-protein 3|Rab3-PA', map_location='47B7-47B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10721335, end_position_on_the_genomic_accession=10729203, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19893538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893538, CurrentID=0, Status='live', Symbol='ATP8', Aliases='', description='ATP synthase F0 subunit 8', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=3907, end_position_on_the_genomic_accession=4068, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  36679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36679, CurrentID=0, Status='live', Symbol='CG8089', Aliases='Dmel_ Dmel\\CG8089', description='CG8089 gene product from transcript CG8089-RA', other_designations='CG8089-PA', map_location='51E2-51E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15145825, end_position_on_the_genomic_accession=15147824, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36131, CurrentID=0, Status='live', Symbol='CG33144', Aliases='Dmel_ AAF58756.2, CG12339, CG7621, Dmel\\CG33144, lincRNA.281', description='CG33144 gene product from transcript CG33144-RB', other_designations='CG33144-PA|CG33144-PB|uncharacterized protein', map_location='47B9-47B14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10740054, end_position_on_the_genomic_accession=10798709, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  10178784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178784, CurrentID=0, Status='live', Symbol='CG42716', Aliases='Dmel_ Dmel\\CG42716', description='CG42716 gene product from transcript CG42716-RA', other_designations='CG42716-PA|uncharacterized protein', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16033778, end_position_on_the_genomic_accession=16034170, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  41593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41593, CurrentID=0, Status='live', Symbol='CG14383', Aliases='Dmel_ Dmel\\CG14383', description='CG14383 gene product from transcript CG14383-RA', other_designations='CG14383-PA|uncharacterized protein', map_location='87D7-87D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12987376, end_position_on_the_genomic_accession=12988511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36134, CurrentID=0, Status='live', Symbol='CG12935', Aliases='Dmel_ Dmel\\CG12935', description='CG12935 gene product from transcript CG12935-RA', other_designations='CG12935-PA|CG12935-PB|uncharacterized protein', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10809397, end_position_on_the_genomic_accession=10810655, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835659, CurrentID=0, Status='live', Symbol='CG45492', Aliases='Dmel_ Dmel\\CG45492', description='CG45492 gene product from transcript CG45492-RA', other_designations='CG45492-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19943357, end_position_on_the_genomic_accession=19944147, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  41599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41599, CurrentID=0, Status='live', Symbol='CG7518', Aliases='Dmel_ Dmel\\CG7518', description='CG7518 gene product from transcript CG7518-RJ', other_designations='CG7518-PF|CG7518-PG|CG7518-PH|CG7518-PI|CG7518-PJ', map_location='87D7-87D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12998089, end_position_on_the_genomic_accession=13008404, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  36141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36141, CurrentID=0, Status='live', Symbol='Lsm10', Aliases='Dmel_CG12938, CG12938, Dmel\\CG12938, d', description='CG12938 gene product from transcript CG12938-RA', other_designations='CG12938-PA|Lsm10-PA', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10821098, end_position_on_the_genomic_accession=10821640, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674067, CurrentID=0, Status='live', Symbol='CG42560', Aliases='Dmel_ CG30182, Dmel\\CG42560, Dmel_CG30182', description='CG42560 gene product from transcript CG42560-RA', other_designations='CG42560-PA|CG42560-PB|uncharacterized protein', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23599526, end_position_on_the_genomic_accession=23600787, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36143, CurrentID=0, Status='live', Symbol='CG12343', Aliases='Dmel_ Dmel\\CG12343, Syf2', description='CG12343 gene product from transcript CG12343-RA', other_designations='CG12343-PA|uncharacterized protein', map_location='47C1-47C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10822400, end_position_on_the_genomic_accession=10823372, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19893552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893552, CurrentID=0, Status='live', Symbol='ND4L', Aliases='', description='NADH dehydrogenase subunit 4L', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=9545, end_position_on_the_genomic_accession=9835, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  19893555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19893555, CurrentID=0, Status='live', Symbol='ND6', Aliases='', description='NADH dehydrogenase subunit 6', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_024511.2', start_position_on_the_genomic_accession=9971, end_position_on_the_genomic_accession=10495, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  249663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=249663, CurrentID=0, Status='live', Symbol='hoe1', Aliases='Dmel_CG12787, BcDNA:GH02636, CG11918, CG12787, CG2831, Dmel\\CG12787, Hoe1, hoe', description='hoepel1', other_designations='CG12787-PA|CG12787-PB|CG12787-PC|CG12787-PD|CG12787-PE|CG12787-PF|CG12787-PH|CG12787-PI|CG12787-PJ|hoe1-PA|hoe1-PB|hoe1-PC|hoe1-PD|hoe1-PE|hoe1-PF|hoe1-PH|hoe1-PI|hoe1-PJ|hoepel', map_location='25B3-25B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4908581, end_position_on_the_genomic_accession=4932384, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  38865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38865, CurrentID=0, Status='live', Symbol='CG7457', Aliases='Dmel_ CT13960, Dmel\\CG7457', description='CG7457 gene product from transcript CG7457-RB', other_designations='CG7457-PA|CG7457-PB', map_location='66A10-66A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7719899, end_position_on_the_genomic_accession=7726473, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36150, CurrentID=0, Status='live', Symbol='nclb', Aliases='Dmel_CG6751, CG6751, Dmel\\CG6751', description='no child left behind', other_designations='CG6751-PA|nclb-PA', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10875748, end_position_on_the_genomic_accession=10877296, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674071, CurrentID=0, Status='live', Symbol='CG42497', Aliases='Dmel_ Dmel\\CG42497, Dromel_CG9878_FBtr0071676_tim10_uORF', description='CG42497 gene product from transcript CG42497-RA', other_designations='CG42497-PA|uncharacterized protein', map_location='57F8-57F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21690215, end_position_on_the_genomic_accession=21691139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38869, CurrentID=0, Status='live', Symbol='CG13685', Aliases='Dmel_ Dmel\\CG13685', description='CG13685 gene product from transcript CG13685-RC', other_designations='CG13685-PA|CG13685-PB|CG13685-PC|CG13685-PD|uncharacterized protein', map_location='66A11-66A11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7747344, end_position_on_the_genomic_accession=7750552, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771723, CurrentID=0, Status='live', Symbol='His3:CG33836', Aliases='Dmel_CG33836, CG33836, Dmel\\CG33836', description='CG33836 gene product from transcript CG33836-RA', other_designations='CG33836-PA|His3:CG33836-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21489830, end_position_on_the_genomic_accession=21490358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36172, CurrentID=0, Status='live', Symbol='CG9084', Aliases='Dmel_ Dmel\\CG9084', description='CG9084 gene product from transcript CG9084-RD', other_designations='CG9084-PB|CG9084-PC|CG9084-PD|CG9084-PE', map_location='47D7-47D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11217416, end_position_on_the_genomic_accession=11219771, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3771728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771728, CurrentID=0, Status='live', Symbol='CG33713', Aliases='Dmel_ CG1704, CG33071, CG33071-ORFB, CG33071ORFB, Dmel\\CG33713, Dromel_CG33713_FBtr0091704_mORF, cg1704', description='CG33713 gene product from transcript CG33713-RA', other_designations='CG33713-PA|CG33713-PB|uncharacterized protein', map_location='19F4-19F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21317936, end_position_on_the_genomic_accession=21319296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36177, CurrentID=0, Status='live', Symbol='CG13218', Aliases='Dmel_ Dmel\\CG13218', description='CG13218 gene product from transcript CG13218-RA', other_designations='CG13218-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11233192, end_position_on_the_genomic_accession=11233714, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36178, CurrentID=0, Status='live', Symbol='CG13228', Aliases='Dmel_ Dmel\\CG13228', description='CG13228 gene product from transcript CG13228-RA', other_designations='CG13228-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11233995, end_position_on_the_genomic_accession=11234465, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36180, CurrentID=0, Status='live', Symbol='CG13217', Aliases='Dmel_ BcDNA:RE32413, Dmel\\CG13217', description='CG13217 gene product from transcript CG13217-RA', other_designations='CG13217-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11236203, end_position_on_the_genomic_accession=11236686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771733, CurrentID=0, Status='live', Symbol='CG33642', Aliases='Dmel_ Dmel\\CG33642', description='CG33642 gene product from transcript CG33642-RA', other_designations='CG33642-PA|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13529975, end_position_on_the_genomic_accession=13530736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36182, CurrentID=0, Status='live', Symbol='CG13215', Aliases='Dmel_ Dmel\\CG13215', description='CG13215 gene product from transcript CG13215-RA', other_designations='CG13215-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11239494, end_position_on_the_genomic_accession=11240114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674075, CurrentID=0, Status='live', Symbol='CG42526', Aliases='Dmel_ Dmel\\CG42526', description='CG42526 gene product from transcript CG42526-RA', other_designations='CG42526-PA|uncharacterized protein', map_location='67B6-67B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9433686, end_position_on_the_genomic_accession=9435399, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36184, CurrentID=0, Status='live', Symbol='CG13226', Aliases='Dmel_ Dmel\\CG13226', description='CG13226 gene product from transcript CG13226-RA', other_designations='CG13226-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11244247, end_position_on_the_genomic_accession=11245199, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36186, CurrentID=0, Status='live', Symbol='CG34227', Aliases='Dmel_ CG7738, Dmel\\CG34227', description='CG34227 gene product from transcript CG34227-RA', other_designations='CG34227-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11248126, end_position_on_the_genomic_accession=11248610, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  38873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38873, CurrentID=0, Status='live', Symbol='CG32369', Aliases='Dmel_ CG13684, CG7388, Dmel\\CG32369', description='CG32369 gene product from transcript CG32369-RC', other_designations='CG32369-PA|CG32369-PC|uncharacterized protein', map_location='66A13-66A14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7777105, end_position_on_the_genomic_accession=7809825, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3771749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771749, CurrentID=0, Status='live', Symbol='CG14929', Aliases='Dmel_ CG32953, Dmel\\CG14929, anon-WO0140519.184', description='CG14929 gene product from transcript CG14929-RA', other_designations='CG14929-PA|CG14929-PB|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11519832, end_position_on_the_genomic_accession=11521444, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772149, CurrentID=0, Status='live', Symbol='His3:CG33803', Aliases='Dmel_CG33803, CG33803, Dmel\\CG33803', description='CG33803 gene product from transcript CG33803-RA', other_designations='CG33803-PA|His3:CG33803-PA|histone H3', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21420129, end_position_on_the_genomic_accession=21420657, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740630, CurrentID=0, Status='live', Symbol='CG34204', Aliases='Dmel_ Dmel\\CG34204', description='CG34204 gene product from transcript CG34204-RA', other_designations='CG34204-PA|CG34204-PB|uncharacterized protein', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21796395, end_position_on_the_genomic_accession=21796941, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771760, CurrentID=0, Status='live', Symbol='CG33703', Aliases='Dmel_ Dmel\\CG33703', description='CG33703 gene product from transcript CG33703-RA', other_designations='CG33703-PA|uncharacterized protein', map_location='67B11-67B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9512008, end_position_on_the_genomic_accession=9512660, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  53425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53425, CurrentID=0, Status='live', Symbol='Alk', Aliases='Dmel_CG8250, ALK, CG8250, DAlk, DAlk53, Dmel\\CG8250, alk, dALK, mili', description='Anaplastic lymphoma kinase', other_designations='Alk-PA|Alk-PB|CG8250-PA|CG8250-PB|anaplastic lymphoma kinase', map_location='53C12-53C13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16625293, end_position_on_the_genomic_accession=16640011, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  53426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53426, CurrentID=0, Status='live', Symbol='ACXE', Aliases='Dmel_CG17178, AC, CG17178, D Dmel\\CG17178', description='CG17178 gene product from transcript CG17178-RB', other_designations='ACXE-PB|CG17178-PB', map_location='34A2-34A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12925611, end_position_on_the_genomic_accession=12929823, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53427, CurrentID=0, Status='live', Symbol='ACXB', Aliases='Dmel_CG17174, AC, AC34A, Ac34A, CG17174, D Dmel\\CG17174', description='CG17174 gene product from transcript CG17174-RA', other_designations='ACXB-PA|CG17174-PA|adenylyl cyclase|adenylyl cyclase 34A', map_location='34A2-34A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12916814, end_position_on_the_genomic_accession=12921008, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53428, CurrentID=0, Status='live', Symbol='wds', Aliases='Dmel_CG17437, CG17437, Dmel\\CG17437, EG:BACR25B3.7, WD5, WDR5, WDS, Wds, l(1)16-94, l(1)3Ad, l(1)G0251, l(1)zw8, zw-8, zw8', description='will die slowly', other_designations='CG17437-PA|CG17437-PB|lethal(1)3Ad|wds-PA|wds-PB', map_location='1-1.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2585980, end_position_on_the_genomic_accession=2588217, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53431, CurrentID=0, Status='live', Symbol='Best1', Aliases='Dmel_CG6264, CG6264, Dbest, Dmel\\CG6264, anon-WO0118547.380, best, dBest1, dbest1, dmBest1', description='Bestrophin 1', other_designations='Best1-PA|Best1-PB|Best1-PC|Best1-PD|CG6264-PA|CG6264-PB|CG6264-PC|CG6264-PD|bestrophin 1', map_location='85F15-85F16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10161464, end_position_on_the_genomic_accession=10170487, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  53432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53432, CurrentID=0, Status='live', Symbol='ACXA', Aliases='Dmel_CG17176, AC, AC62D, Ac62D, CG17176, D Dmel\\CG17176', description='CG17176 gene product from transcript CG17176-RG', other_designations='ACXA-PA|ACXA-PC|ACXA-PD|ACXA-PG|ACXA-PH|AdenylateCyclase A|CG17176-PA|CG17176-PC|CG17176-PD|CG17176-PG|CG17176-PH|adenylyl cyclase|adenylyl cyclase 62D', map_location='34A2-34A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12921091, end_position_on_the_genomic_accession=12925453, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53433, CurrentID=0, Status='live', Symbol='Vhl', Aliases='Dmel_CG13221, BcDNA:RH61560, CG13221, DVhl, Dmel\\CG13221, Dvhl, VHL, d-VHL, d-vhl, dVHL, dmVHL, vhl', description='von Hippel-Lindau', other_designations='CG13221-PA|CG13221-PB|Vhl-PA|Vhl-PB|von hippel lindau|von hippel lindau protein', map_location='47E5-47E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11279725, end_position_on_the_genomic_accession=11281033, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53434, CurrentID=0, Status='live', Symbol='c12.2', Aliases='Dmel_CG12149, CG12149, Dmel\\CG12149', description='CG12149 gene product from transcript CG12149-RB', other_designations='CG12149-PA|CG12149-PB|c12.2-PA|c12.2-PB', map_location='8D7-8D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9319359, end_position_on_the_genomic_accession=9324458, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  53435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53435, CurrentID=0, Status='live', Symbol='c12.1', Aliases='Dmel_CG12135, C12.1, CG12135, Dmel\\CG12135, cwc15', description='CG12135 gene product from transcript CG12135-RA', other_designations='CG12135-PA|c12.1-PA', map_location='8D7-8D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9317915, end_position_on_the_genomic_accession=9319223, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53436, CurrentID=0, Status='live', Symbol='c11.1', Aliases='Dmel_CG12132, C11.1, CG12132, Dmel\\CG12132', description='CG12132 gene product from transcript CG12132-RB', other_designations='CG12132-PA|CG12132-PB|c11.1-PA|c11.1-PB', map_location='8D4-8D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9275318, end_position_on_the_genomic_accession=9283117, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  53437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53437, CurrentID=0, Status='live', Symbol='Sulf1', Aliases='Dmel_CG6725, CG6725, DSulf, DSulf1, Dmel\\CG6725, Dsulf1, Neu5, sulf1', description='Sulfated', other_designations='CG6725-PA|CG6725-PB|CG6725-PC|Sulf1-PA|Sulf1-PB|Sulf1-PC|sulfated', map_location='89B12-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16310542, end_position_on_the_genomic_accession=16339575, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  53438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53438, CurrentID=0, Status='live', Symbol='RecQ4', Aliases='Dmel_CG7487, CG7487, DmRECQ4, Dmel\\CG7487, RECQL4, cg7487, dRecQ4', description='RecQ4 helicase', other_designations='CG7487-PA|RecQ4-PA', map_location='66B10-66B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8170640, end_position_on_the_genomic_accession=8175978, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53439, CurrentID=0, Status='live', Symbol='pont', Aliases='Dmel_CG4003, 0229/05, CG4003, Dmel\\CG4003, PON, PONT, Pont, Pontin, RUVBL1, Tip49, Tip49/Pontin, anon-WO0118547.698, dPon, dPontin, dpon, l(3)S022905, pon', description='pontin', other_designations='CG4003-PA|dpontin|pont-PA', map_location='86A3-86A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10262123, end_position_on_the_genomic_accession=10263899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53442, CurrentID=0, Status='live', Symbol='Spt5', Aliases='Dmel_CG7626, CG7626, DSIF, Dmel\\CG7626, Dspt5, SPT5, dDSIF, d spt5', description='CG7626 gene product from transcript CG7626-RA', other_designations='CG7626-PA|Spt5-PA|complementation group C', map_location='56D5-56D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19437483, end_position_on_the_genomic_accession=19441898, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53443, CurrentID=0, Status='live', Symbol='SF2', Aliases='Dmel_CG6987, 154911_at, ASF, ASF/SF2, CG6987, Dmel\\CG6987, SRp30, cg6987, dASF, dSF2/ASF, dSRp30, dmSF2/p28', description='Splicing factor 2', other_designations='CG6987-PA|CG6987-PB|SF2-PA|SF2-PB|SRp30|splicing factor 2', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16338943, end_position_on_the_genomic_accession=16341988, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  53444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53444, CurrentID=0, Status='live', Symbol='SC35', Aliases='Dmel_CG5442, CG15494, CG5442, CG5542, Dmel\\CG5442, SC-35, Sc35, anon-EST:Posey284, cg5442, d sc35', description='CG5442 gene product from transcript CG5442-RC', other_designations='CG5442-PA|CG5442-PB|CG5442-PC|CG5442-PD|SC35-PA|SC35-PB|SC35-PC|SC35-PD', map_location='33E4-33E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12431075, end_position_on_the_genomic_accession=12434665, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53445, CurrentID=0, Status='live', Symbol='Roc1b', Aliases='Dmel_CG16988, BcDNA:AT21612, CG16988, Dmel\\CG16988, ROC1B, ROC1b, ROC2, Roc, dRoc1b, roc1b', description='Regulator of cullins 1b', other_designations='CG16988-PA|Roc1b-PA|regulator of cullins 1b', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=357981, end_position_on_the_genomic_accession=358486, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53446, CurrentID=0, Status='live', Symbol='HGTX', Aliases='Dmel_CG13475, CG13475, CG4745, Dmel\\CG13475, Nk6, Nkx6, dNkx6, nkx6', description='CG13475 gene product from transcript CG13475-RA', other_designations='CG13475-PA|HGTX-PA|Nk6 homeobox', map_location='70E3-70E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14583222, end_position_on_the_genomic_accession=14599540, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  53448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53448, CurrentID=0, Status='live', Symbol='meso18E', Aliases='Dmel_CG14233, CG14233, Dmel\\CG14233, D Meso18E, Mmeso18E, anon-18DEd, eso18E, ld26007', description='CG14233 gene product from transcript CG14233-RA', other_designations='CG14233-PA|CG14233-PB|CG14233-PC|CG14233-PD|meso18E-PA|meso18E-PB|meso18E-PC|meso18E-PD', map_location='18E2-18E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19706418, end_position_on_the_genomic_accession=19713045, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  10178777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178777, CurrentID=0, Status='live', Symbol='CG42835', Aliases='Dmel_ Dmel\\CG42835', description='CG42835 gene product from transcript CG42835-RA', other_designations='CG42835-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17420638, end_position_on_the_genomic_accession=17420989, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178781, CurrentID=0, Status='live', Symbol='CG42867', Aliases='Dmel_ Dmel\\CG42867', description='CG42867 gene product from transcript CG42867-RB', other_designations='CG42867-PB|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22665646, end_position_on_the_genomic_accession=22666145, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178782, CurrentID=0, Status='live', Symbol='CG31808', Aliases='Dmel_ BcDNA:RE70695, CR31808, Dmel\\CG31808, Dmel_CR31808, MRE33, RE70695', description='CG31808 gene product from transcript CG31808-RC', other_designations='CG31808-PC|mRNA-like ncRNA in embryogenesis 33|uncharacterized protein', map_location='36A11-36A11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16704286, end_position_on_the_genomic_accession=16716195, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53471, CurrentID=0, Status='live', Symbol='SP555', Aliases='Dmel_CG14041, BcDNA:AT09933, CG14041, Dmel\\CG14041, NEST:bs13e02, dmSSBa', description='CG14041 gene product from transcript CG14041-RB', other_designations='CG14041-PA|CG14041-PB|CG14041-PC|CG14041-PD|SP555-PA|SP555-PB|SP555-PC|SP555-PD', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5037709, end_position_on_the_genomic_accession=5040818, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53472, CurrentID=0, Status='live', Symbol='Aplip1', Aliases='Dmel_CG1200, APLIP1, CG1200, Dmel\\CG1200, E(Khc)ek4, EK4, Ek4, JIP1, Jip1/2, SP512, aplip1, dp1, ek4, jip1', description='APP-like protein interacting protein 1', other_designations='Aplip1-PA|Aplip1-PB|CG1200-PA|CG1200-PB|JNK interacting protein-1|c-Jun N-terminal kinase (JNK)-interacting protein', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1196317, end_position_on_the_genomic_accession=1199339, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  53473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53473, CurrentID=0, Status='live', Symbol='SP1029', Aliases='Dmel_CG11956, CG11956, Dmel\\CG11956, ptl, ptls', description='CG11956 gene product from transcript CG11956-RA', other_designations='CG11956-PA|CG11956-PB|CG11956-PC|CG11956-PD|CG11956-PE|SP1029-PA|SP1029-PB|SP1029-PC|SP1029-PD|SP1029-PE|pathless', map_location='99A4-99A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29244961, end_position_on_the_genomic_accession=29250451, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  10178787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178787, CurrentID=0, Status='live', Symbol='CG42852', Aliases='Dmel_ Dmel\\CG42852, lincRNA.585', description='CG42852 gene product from transcript CG42852-RA', other_designations='CG42852-PA|uncharacterized protein', map_location='73C2-73C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16755489, end_position_on_the_genomic_accession=16755720, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178792, CurrentID=0, Status='live', Symbol='CG42694', Aliases='Dmel_ Dmel\\CG42694', description='CG42694 gene product from transcript CG42694-RB', other_designations='CG42694-PA|CG42694-PB|uncharacterized protein', map_location='59B7-59B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22941877, end_position_on_the_genomic_accession=22945147, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  10178793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178793, CurrentID=0, Status='live', Symbol='CG42789', Aliases='Dmel_ DMG8-chr3R.40.015, DMG8-chr3R.40.015-MIP08885, Dmel\\CG42789, lincRNA.863', description='CG42789 gene product from transcript CG42789-RC', other_designations='CG42789-PB|CG42789-PC|uncharacterized protein', map_location='97D2-97D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26783609, end_position_on_the_genomic_accession=26784208, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10178795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178795, CurrentID=0, Status='live', Symbol='CG42762', Aliases='Dmel_ Dmel\\CG42762', description='CG42762 gene product from transcript CG42762-RA', other_designations='CG42762-PA|uncharacterized protein', map_location='28E7-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8143401, end_position_on_the_genomic_accession=8144047, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178796, CurrentID=0, Status='live', Symbol='CG42780', Aliases='Dmel_ Dmel\\CG42780', description='CG42780 gene product from transcript CG42780-RA', other_designations='CG42780-PA|CG42780-PB|uncharacterized protein', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7331265, end_position_on_the_genomic_accession=7332191, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10178797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178797, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-X', Aliases='Dmel_CG42772, CG42772, Dmel\\CG42772', description='CG42772 gene product from transcript CG42772-RA', other_designations='CG42772-PA|CG42772-PB|pre-mod(mdg4)-X-PA|pre-mod(mdg4)-X-PB', map_location='93D8-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21369330, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  10178805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178805, CurrentID=0, Status='live', Symbol='CG42654', Aliases='Dmel_ Dmel\\CG42654', description='CG42654 gene product from transcript CG42654-RB', other_designations='CG42654-PB|CG42654-PC|uncharacterized protein', map_location='76D6-76D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19953794, end_position_on_the_genomic_accession=19954526, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10178806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178806, CurrentID=0, Status='live', Symbol='CG42656', Aliases='Dmel_ DMG8-chr3R.6.013, Dmel\\CG42656', description='CG42656 gene product from transcript CG42656-RA', other_designations='CG42656-PA|uncharacterized protein', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7492961, end_position_on_the_genomic_accession=7493271, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178809, CurrentID=0, Status='live', Symbol='CG42854', Aliases='Dmel_ Dmel\\CG42854', description='CG42854 gene product from transcript CG42854-RA', other_designations='CG42854-PA|uncharacterized protein', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17245353, end_position_on_the_genomic_accession=17247016, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  10178812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178812, CurrentID=0, Status='live', Symbol='CG42815', Aliases='Dmel_ CG32186, CG5587, CG5944, Dmel\\CG42815, Dmel_CG32186', description='CG42815 gene product from transcript CG42815-RA', other_designations='CG42815-PA|uncharacterized protein', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17791079, end_position_on_the_genomic_accession=17791486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  53501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53501, CurrentID=0, Status='live', Symbol='Ugt86Dj', Aliases='Dmel_CG15902, AC 006491J, CG15902, Dmel\\CG15902', description='CG15902 gene product from transcript CG15902-RA', other_designations='CG15902-PA|Ugt86Dj-PA', map_location='86D5-86D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11173441, end_position_on_the_genomic_accession=11175408, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53502, CurrentID=0, Status='live', Symbol='Ugt86Di', Aliases='Dmel_CG6658, AC 006491I, CG6658, Dmel\\CG6658', description='CG6658 gene product from transcript CG6658-RB', other_designations='CG6658-PA|CG6658-PB|Ugt86Di-PA|Ugt86Di-PB', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11151026, end_position_on_the_genomic_accession=11153849, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  53503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53503, CurrentID=0, Status='live', Symbol='Ugt86Dh', Aliases='Dmel_CG4772, AC 006491H, CG4772, Dmel\\CG4772', description='CG4772 gene product from transcript CG4772-RB', other_designations='CG4772-PA|CG4772-PB|Ugt86Dh-PA|Ugt86Dh-PB', map_location='86D6-86D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11175529, end_position_on_the_genomic_accession=11177910, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53504, CurrentID=0, Status='live', Symbol='Ugt86Dg', Aliases='Dmel_CG17200, AC 006491G, CG17200, CG17200-PA, Dmel\\CG17200', description='CG17200 gene product from transcript CG17200-RB', other_designations='CG17200-PB|Ugt86Dg-PB', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11164423, end_position_on_the_genomic_accession=11166074, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178817, CurrentID=0, Status='live', Symbol='CG42869', Aliases='Dmel_ Dmel\\CG42869', description='CG42869 gene product from transcript CG42869-RB', other_designations='CG42869-PB|uncharacterized protein', map_location='93F7-93F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21736513, end_position_on_the_genomic_accession=21736849, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53506, CurrentID=0, Status='live', Symbol='Ugt86De', Aliases='Dmel_CG6653, AC 006491E, CG6653, CG6653-PA, Dmel\\CG6653, Ugt86DE', description='CG6653 gene product from transcript CG6653-RA', other_designations='CG6653-PA|Ugt86De-PA', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11166177, end_position_on_the_genomic_accession=11167981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53507, CurrentID=0, Status='live', Symbol='Ugt86Dd', Aliases='Dmel_CG6633, AC 006491D, CG6633, Dmel\\CG6633', description='CG6633 gene product from transcript CG6633-RB', other_designations='CG6633-PB|Ugt86Dd-PB', map_location='86D4-86D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11126597, end_position_on_the_genomic_accession=11128328, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53508, CurrentID=0, Status='live', Symbol='Ugt86Dc', Aliases='Dmel_CG4739, AC 006491C, CG4739, Dmel\\CG4739', description='CG4739 gene product from transcript CG4739-RA', other_designations='CG4739-PA|Ugt86Dc-PA', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11154626, end_position_on_the_genomic_accession=11156513, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53510, CurrentID=0, Status='live', Symbol='Ugt86Da', Aliases='Dmel_CG18578, AC 006491A, BEST:GM04645, CG18578, Dmel\\CG18578, GM04645, ugt86Da', description='CG18578 gene product from transcript CG18578-RA', other_designations='CG18578-PA|Ugt86Da-PA', map_location='86D5-86D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11157098, end_position_on_the_genomic_accession=11159744, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53511, CurrentID=0, Status='live', Symbol='Ugt36Bc', Aliases='Dmel_CG17932, AC 005121C, CG17932, Dmel\\CG17932', description='CG17932 gene product from transcript CG17932-RB', other_designations='CG17932-PA|CG17932-PB|CG17932-PC|Ugt36Bc-PA|Ugt36Bc-PB|Ugt36Bc-PC', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16799025, end_position_on_the_genomic_accession=16801584, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53512, CurrentID=0, Status='live', Symbol='Ugt36Bb', Aliases='Dmel_CG13271, AC 005121B, CG13271, Dmel\\CG13271', description='CG13271 gene product from transcript CG13271-RA', other_designations='CG13271-PA|Ugt36Bb-PA', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16796595, end_position_on_the_genomic_accession=16798273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53513, CurrentID=0, Status='live', Symbol='Ugt36Ba', Aliases='Dmel_CG13270, AC 005121A, CG13270, Dmel\\CG13270', description='CG13270 gene product from transcript CG13270-RB', other_designations='CG13270-PA|CG13270-PB|Ugt36Ba-PA|Ugt36Ba-PB', map_location='36B1-36B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16794211, end_position_on_the_genomic_accession=16796009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53515, CurrentID=0, Status='live', Symbol='Myo28B1', Aliases='Dmel_CG6976, 28B, CG6976, DmVIIB, Dmel\\CG6976, Myo28B, VIIb 28B, myo 28B1, myo28B', description='Myosin 28B1', other_designations='CG6976-PA|CG6976-PB|CG6976-PC|CG6976-PD|Myo28B1-PA|Myo28B1-PB|Myo28B1-PC|Myo28B1-PD|myosin 28B1|myosin VII|myosin VII (28B)|myosin VII 28B1-B4|myosin VIIB', map_location='28C1-28C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7666047, end_position_on_the_genomic_accession=7689459, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  53516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53516, CurrentID=0, Status='live', Symbol='lectin-33A', Aliases='Dmel_CG16834, AC 007083A, AC007083a, CG16834, Dmel\\CG16834, lecti-33A, lectin-33a', description='CG16834 gene product from transcript CG16834-RB', other_designations='CG16834-PB|lectin-33A-PB', map_location='32C1-32C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11002072, end_position_on_the_genomic_accession=11002738, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178831, CurrentID=0, Status='live', Symbol='CG42846', Aliases='Dmel_ Dmel\\CG42846', description='CG42846 gene product from transcript CG42846-RA', other_designations='CG42846-PA|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=230971, end_position_on_the_genomic_accession=231604, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178832, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-W', Aliases='Dmel_CG42775, CG42775, Dmel\\CG42775', description='CG42775 gene product from transcript CG42775-RB', other_designations='CG42775-PA|CG42775-PB|pre-mod(mdg4)-W-PA|pre-mod(mdg4)-W-PB', map_location='93D8-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21370249, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  53522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53522, CurrentID=0, Status='live', Symbol='lectin-46Cb', Aliases='Dmel_CG1652, AC 007330B, AC007330b, CG1652, Dmel\\CG1652, Lectin-46Cb, lectin46Cb', description='CG1652 gene product from transcript CG1652-RB', other_designations='CG1652-PB|lectin 46cb|lectin-46Cb-PB', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9815510, end_position_on_the_genomic_accession=9816600, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53523, CurrentID=0, Status='live', Symbol='lectin-46Ca', Aliases='Dmel_CG1656, AC 007330A, AC007330a, CG1656, Dmel\\CG1656, Lectin-46CA, Lectin-46Ca', description='CG1656 gene product from transcript CG1656-RA', other_designations='CG1656-PA|lectin-44Ca|lectin-46Ca-PA', map_location='46B4-46B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9817300, end_position_on_the_genomic_accession=9818535, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178837, CurrentID=0, Status='live', Symbol='CG42740', Aliases='Dmel_ Dmel\\CG42740', description='CG42740 gene product from transcript CG42740-RC', other_designations='CG42740-PA|CG42740-PB|CG42740-PC|uncharacterized protein', map_location='99F3-99F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30418767, end_position_on_the_genomic_accession=30422971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  10178839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178839, CurrentID=0, Status='live', Symbol='CG42690', Aliases='Dmel_ Dmel\\CG42690', description='CG42690 gene product from transcript CG42690-RB', other_designations='CG42690-PA|CG42690-PB|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19944263, end_position_on_the_genomic_accession=19945877, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  10178840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178840, CurrentID=0, Status='live', Symbol='CG42837', Aliases='Dmel_ CG8446, Dmel\\CG42837', description='CG42837 gene product from transcript CG42837-RA', other_designations='CG42837-PA|uncharacterized protein', map_location='52F7-52F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16197042, end_position_on_the_genomic_accession=16199378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10178844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178844, CurrentID=0, Status='live', Symbol='CG42831', Aliases='Dmel_ Dmel\\CG42831', description='CG42831 gene product from transcript CG42831-RA', other_designations='CG42831-PA|uncharacterized protein', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10814819, end_position_on_the_genomic_accession=10816084, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178846, CurrentID=0, Status='live', Symbol='CG42661', Aliases='Dmel_ Dmel\\CG42661', description='CG42661 gene product from transcript CG42661-RB', other_designations='CG42661-PA|CG42661-PB|uncharacterized protein', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7465509, end_position_on_the_genomic_accession=7468808, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  10178847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178847, CurrentID=0, Status='live', Symbol='CG42662', Aliases='Dmel_ Dmel\\CG42662', description='CG42662 gene product from transcript CG42662-RC', other_designations='CG42662-PA|CG42662-PB|CG42662-PC|CG42662-PD|uncharacterized protein', map_location='52C1-52C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15698726, end_position_on_the_genomic_accession=15701696, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10178851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178851, CurrentID=0, Status='live', Symbol='CG42857', Aliases='Dmel_ DMG6-MB6.chr3R.10.010, Dmel\\CG42857', description='CG42857 gene product from transcript CG42857-RA', other_designations='CG42857-PA|uncharacterized protein', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9820514, end_position_on_the_genomic_accession=9821191, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53540, CurrentID=0, Status='live', Symbol='CG3706', Aliases='Dmel_ Dmel\\CG3706, EG:BAC7A4.20, EG:BACR7A4.20, ESTS:99F6T', description='CG3706 gene product from transcript CG3706-RA', other_designations='CG3706-PA', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=930931, end_position_on_the_genomic_accession=932475, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  53541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53541, CurrentID=0, Status='live', Symbol='lectin-30A', Aliases='Dmel_CG17011, AC 005889A, AC005889a, CG17011, Dmel\\CG17011, Lectin-30A', description='CG17011 gene product from transcript CG17011-RA', other_designations='CG17011-PA|Lectin30A|lectin-30A-PA', map_location='30A6-30A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9254076, end_position_on_the_genomic_accession=9254854, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  53542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53542, CurrentID=0, Status='live', Symbol='lectin-28C', Aliases='Dmel_CG7106, AC 004723A, AC004723a, CG7106, Dmel\\CG7106', description='CG7106 gene product from transcript CG7106-RC', other_designations='CG7106-PB|CG7106-PC|Lectin28C|lectin-28C-PB|lectin-28C-PC', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7857077, end_position_on_the_genomic_accession=7860120, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53543, CurrentID=0, Status='live', Symbol='lectin-24A', Aliases='Dmel_CG3410, AC 005149A, AC005149a, CG3410, Dmel\\CG3410', description='CG3410 gene product from transcript CG3410-RA', other_designations='CG3410-PA|Lectin24A|lectin-24A-PA', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3716800, end_position_on_the_genomic_accession=3717773, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  53544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53544, CurrentID=0, Status='live', Symbol='CG42296', Aliases='Dmel_ CG15378, Dmel\\CG42296, Dmel_CG15378', description='CG42296 gene product from transcript CG42296-RA', other_designations='CG42296-PA|uncharacterized protein', map_location='22C1-22C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2032519, end_position_on_the_genomic_accession=2036041, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178857, CurrentID=0, Status='live', Symbol='CG42692', Aliases='Dmel_ Dmel\\CG42692', description='CG42692 gene product from transcript CG42692-RA', other_designations='CG42692-PA|CG42692-PB|uncharacterized protein', map_location='34F1-34F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14008101, end_position_on_the_genomic_accession=14008820, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53546, CurrentID=0, Status='live', Symbol='CG3021', Aliases='Dmel_ Dmel\\CG3021, E, EG:BACR7A4.8, anon-1Ed', description='CG3021 gene product from transcript CG3021-RA', other_designations='CG3021-PA|transcription unit E', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1026791, end_position_on_the_genomic_accession=1028290, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53547, CurrentID=0, Status='live', Symbol='lectin-29Ca', Aliases='Dmel_CG17799, AC 004423A, AC004423a, Acp29AB-I, CG17799, Dmel\\CG17799, Lectin-29Ca', description='CG17799 gene product from transcript CG17799-RA', other_designations='CG17799-PA|Lectin29Ca|lectin-29Ca-PA', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8393489, end_position_on_the_genomic_accession=8394280, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178860, CurrentID=0, Status='live', Symbol='CG42790', Aliases='Dmel_ Dmel\\CG42790, lincRNA.863', description='CG42790 gene product from transcript CG42790-RB', other_designations='CG42790-PB|CG42790-PC|uncharacterized protein', map_location='97D2-97D2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26784338, end_position_on_the_genomic_accession=26785753, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178861, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-T', Aliases='Dmel_CG42769, CG42769, Dmel\\CG42769, Pre-MOD(MDG4)-T', description='CG42769 gene product from transcript CG42769-RB', other_designations='CG42769-PB|pre-mod(mdg4)-T-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21360176, end_position_on_the_genomic_accession=21362131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53550, CurrentID=0, Status='live', Symbol='lectin-24Db', Aliases='Dmel_CG2958, AC 004371A, AC004371a, CG2958, Dmel\\CG2958', description='CG2958 gene product from transcript CG2958-RA', other_designations='CG2958-PA|Lectin24DB|lectin|lectin-24Db-PA', map_location='24D8-24D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4188123, end_position_on_the_genomic_accession=4189284, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53552, CurrentID=0, Status='live', Symbol='lectin-21Ca', Aliases='Dmel_CG2826, AC 004573A, AC004573a, CG2826, Dmel\\CG2826, Lectin-21Ca', description='CG2826 gene product from transcript CG2826-RA', other_designations='CG2826-PA|Lectin21Ca|lectin-21Ca-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=624617, end_position_on_the_genomic_accession=625556, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771784 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771784, CurrentID=0, Status='live', Symbol='CG33631', Aliases='Dmel_ CG12808, Dmel\\CG33631', description='CG33631 gene product from transcript CG33631-RA', other_designations='CG33631-PA|uncharacterized protein', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10064007, end_position_on_the_genomic_accession=10064645, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53554, CurrentID=0, Status='live', Symbol='krz', Aliases='Dmel_CG1487, 0952/14, CG1487, Dmel\\CG1487, Krz, Kurz, l(3)S095214', description='kurtz', other_designations='CG1487-PA|CG1487-PB|beta-arrestin|krz-PA|krz-PB|kurtz-arrestin', map_location='100E3-100E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=32045314, end_position_on_the_genomic_accession=32051993, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53556, CurrentID=0, Status='live', Symbol='dbo', Aliases='Dmel_CG6224, CG6224, DIABLO, Dbo, Dm.DBO, Dmel\\CG6224, Smac', description='diablo', other_designations='CG6224-PA|CG6224-PB|DIABLO|dbo-PA|dbo-PB', map_location='72A2-72A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15826888, end_position_on_the_genomic_accession=15832004, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53557, CurrentID=0, Status='live', Symbol='esn', Aliases='Dmel_CG43722, CG12833, CG12835, CG43722, Dm_2R:11728, Dmel\\CG43722, Dmel_CG12833, Dmel_CG12835', description='espinas', other_designations='CG43722-PA|CG43722-PB|CG43722-PC|CG43722-PD|esn-PA|esn-PB|esn-PC|esn-PD', map_location='42F1-42F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7086316, end_position_on_the_genomic_accession=7127868, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53558, CurrentID=0, Status='live', Symbol='wun2', Aliases='Dmel_CG8805, BEST:CK02248, CG8805, CK02248, Dmel\\CG8805, DrPAP2[G], LPP-like, N14, Pap2G, Tunen, WUN-like, Wun2, wun-2', description='wunen-2', other_designations='CG8805-PA|CG8805-PB|Wunen2|wun2-PA|wun2-PB|wunen 2', map_location='45D4-45D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9414342, end_position_on_the_genomic_accession=9417846, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3771785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771785, CurrentID=0, Status='live', Symbol='His2A:CG33826', Aliases='Dmel_CG33826, CG33826, Dmel\\CG33826', description='CG33826 gene product from transcript CG33826-RA', other_designations='CG33826-PA|His2A:CG33826-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21473141, end_position_on_the_genomic_accession=21473633, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53560, CurrentID=0, Status='live', Symbol='cic', Aliases='Dmel_CG43122, CG43122, CG5060, CG5067, CIC, Cic, Dmel\\CG43122, Dmel_CG5060, Dmel_CG5067, E(Dl)D49s, fet, cic', description='capicua', other_designations='CG43122-PA|CG43122-PB|CG43122-PC|CG43122-PD|CG43122-PG|CG43122-PH|cic-PA|cic-PB|cic-PC|cic-PD|cic-PG|cic-PH|enhancer of delta D49|fettucine', map_location='3-72.4 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20252770, end_position_on_the_genomic_accession=20303942, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  53561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53561, CurrentID=0, Status='live', Symbol='cmet', Aliases='Dmel_CG6392, CENP-E, CENP-E meta, CENP-META, CENP-Meta, CENPE, CENPmeta, CG6392, Cenp-E, Cenp-M, Cenp-meta, Cenp_E, Cmet, DmCmet, DmCmeta, Dmel\\CG6392, KIF 10, KIF10, l(2)04431', description='CENP-meta', other_designations='CG6392-PC|CG6392-PD|CG6392-PE|CG6392-PF|Cenp-meta|CenpMeta|cmet-PC|cmet-PD|cmet-PE|cmet-PF', map_location='32E2-32E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11261026, end_position_on_the_genomic_accession=11269002, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  53562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53562, CurrentID=0, Status='live', Symbol='tara', Aliases='Dmel_CG6889, 5472, CG6889, Dmel\\CG6889, EP(3)3463, Tara, anon-WO0257455.25, l(3)03881, l(3)j1C5, l(3)jlC5', description='taranis', other_designations='CG6889-PA|CG6889-PB|tara-PA|tara-PB', map_location='89B8-89B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16225651, end_position_on_the_genomic_accession=16260329, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53563, CurrentID=0, Status='live', Symbol='gukh', Aliases='Dmel_CG31043, CG14287, CG14288, CG31043, CG5456, CG6003, Dmel\\CG31043, GUKH, GUKh, GukH, Gukh, anon-WO0118547.205', description='GUK-holder', other_designations='CG31043-PA|CG31043-PB|CG31043-PC|CG31043-PE|GUK holder|gukh-PA|gukh-PB|gukh-PC|gukh-PE|gukholder', map_location='91E2-91E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18984025, end_position_on_the_genomic_accession=19022436, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53564, CurrentID=0, Status='live', Symbol='retinin', Aliases='Dmel_CG13057, 4A11, CG13057, Dmel\\CG13057', description='CG13057 gene product from transcript CG13057-RA', other_designations='CG13057-PA|retinin-PA', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16331933, end_position_on_the_genomic_accession=16332789, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53565, CurrentID=0, Status='live', Symbol='bor', Aliases='Dmel_CG6815, ATAD3A, Bor, CG6815, CP89BC, Dmel\\CG6815, dATAD3A', description='belphegor', other_designations='CG6815-PA|CG6815-PB|bor-PA|bor-PB|cytoplasmic protein 89BC', map_location='89B9-89B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16261355, end_position_on_the_genomic_accession=16264115, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53566, CurrentID=0, Status='live', Symbol='Kat60', Aliases='Dmel_CG10229, CG10229, Dm-Kat60, DmCG10229, Dmel\\CG10229, Katanin-60, kat60, katanin-60, katanin60-dm', description='Katanin 60', other_designations='CG10229-PA|CG10229-PB|Kat60-PA|Kat60-PB|Katanin-60|katanin|katanin 60', map_location='82F6-82F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5227289, end_position_on_the_genomic_accession=5232209, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53567, CurrentID=0, Status='live', Symbol='larp', Aliases='Dmel_CG42551, 0031/08, 0225/29, BcDNA:LD02611, CG14065, CG14066, CG42551, Dmel\\CG42551, Dmel_CG14065, Dmel_CG14066, Larp, dlarp, l(3)S003108, l(3)S0225291, larp', description='La related protein', other_designations='CG42551-PA|CG42551-PB|CG42551-PD|CG42551-PE|CG42551-PF|CG42551-PG|CG42551-PH|La-related protein|drosophila La related protein|larp-PA|larp-PB|larp-PD|larp-PE|larp-PF|larp-PG|larp-PH|lethal (3) S003108|lethal (3) S022529|meteor', map_location='98C3-98C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28318162, end_position_on_the_genomic_accession=28336440, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  10178880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178880, CurrentID=0, Status='live', Symbol='CG42853', Aliases='Dmel_ Dmel\\CG42853', description='CG42853 gene product from transcript CG42853-RA', other_designations='CG42853-PA|uncharacterized protein', map_location='75B10-75B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18120932, end_position_on_the_genomic_accession=18121989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53569, CurrentID=0, Status='live', Symbol='pins', Aliases='Dmel_CG5692, CG5692, Dmel\\CG5692, PINS, Pins, Rad, Raps, rad, raps, rapsyn', description='partner of inscuteable', other_designations='CG5692-PA|partner of inscutable|pins-PA|rapsinoid|rapsynoid', map_location='98A8-98A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27582840, end_position_on_the_genomic_accession=27586376, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53573, CurrentID=0, Status='live', Symbol='Ilk', Aliases='Dmel_CG10504, CG10504, CT29478, DmILK, Dmel\\CG10504, ILK, ilk, l(3)78Ca', description='Integrin linked kinase', other_designations='CG10504-PA|CG10504-PB|Ilk-PA|Ilk-PB|Integrin-linked kinase|integrin linked kinase', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21216901, end_position_on_the_genomic_accession=21219240, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10178888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178888, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-AD', Aliases='Dmel_CG42773, CG42773, Dmel\\CG42773', description='CG42773 gene product from transcript CG42773-RB', other_designations='CG42773-PA|CG42773-PB|pre-mod(mdg4)-AD-PA|pre-mod(mdg4)-AD-PB', map_location='93D8-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21368279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53577, CurrentID=0, Status='live', Symbol='Jafrac2', Aliases='Dmel_CG1274, 1274, CG1274, DPx-4156, DPx4156, Dmel\\CG1274, JafRac2, Jafrac, Jafrac-2, Prx4, Prx4156, dPrx4, jafrac2, thioredoxin peroxidase 2 isoform A', description='Thioredoxin peroxidase 2', other_designations='CG1274-PA|CG1274-PB|CG1274-PC|Jafrac2-PA|Jafrac2-PB|Jafrac2-PC|jafrac2|peroxiredoxin 4|peroxiredoxin 4156|thioredoxin peroxidase|thioredoxin peroxidase 2', map_location='63A3-63A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3042359, end_position_on_the_genomic_accession=3044174, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53578, CurrentID=0, Status='live', Symbol='Jafrac1', Aliases='Dmel_CG1633, 1633, CG1633, DPx-4783, DPx4783, DmTPx-1, Dmel\\CG1633, Dpx-4783, Jafrac, PRDX1_DROME, Prx, Prx4783, SSP5111, T20, cTPx, dPrx I, jafrac1, jf1', description='thioredoxin peroxidase 1', other_designations='CG1633-PA|CG1633-PB|CG1633-PC|CG1633-PD|CG1633-PE|Jafrac1-PA|Jafrac1-PB|Jafrac1-PC|Jafrac1-PD|Jafrac1-PE|peroxidase|peroxiredoxin|peroxiredoxin 1|peroxiredoxin 4783|peroxiredoxin I|thioredoxin peroxidase', map_location='11E11-11E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13329824, end_position_on_the_genomic_accession=13331994, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53579, CurrentID=0, Status='live', Symbol='Dlip3', Aliases='Dmel_CG12767, CG12767, Dip3, Dmel\\CG12767, anon-WO0118547.169, dip3', description='Dorsal interacting protein 3', other_designations='CG12767-PA|Dlip3-PA|dorsal interacting protein 3', map_location='55B7-55B7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18128494, end_position_on_the_genomic_accession=18130246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53580, CurrentID=0, Status='live', Symbol='Dlip1', Aliases='Dmel_CG15367, CG15367, DIP1, Dip1, Dmel\\CG15367, dip1', description='Dorsal interacting protein 1', other_designations='CG15367-PA|CG15367-PB|Dlip1-PA|Dlip1-PB|dorsal interacting protein 1', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9158622, end_position_on_the_genomic_accession=9159375, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53581, CurrentID=0, Status='live', Symbol='Gclc', Aliases='Dmel_CG2259, 2259, CG2259, DmGCLC, DmGCS, DmGCSh, Dmel\\CG2259, GCL, GCLC, GCLc, GCSh, Gcs, Gcsh, gclc', description='Glutamate-cysteine ligase catalytic subunit', other_designations='CG2259-PA|CG2259-PB|CG2259-PC|CG2259-PD|Gclc-PA|Gclc-PB|Gclc-PC|Gclc-PD|catalytic subunit of glutamate- cysteine ligase|gamma-Glutamylcysteine synthetase|gamma-Glutamylcysteine synthetase catalytic subunit|gamma-glutamylcysteine ligase|glutamate cysteine ligase|glutamate-cysteine ligase', map_location='7D6-7D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8102181, end_position_on_the_genomic_accession=8116147, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  10178894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178894, CurrentID=0, Status='live', Symbol='CG42809', Aliases='Dmel_ CG16849, Dmel\\CG42809, Dmel_CG16849', description='CG42809 gene product from transcript CG42809-RA', other_designations='CG42809-PA|uncharacterized protein', map_location='34C1-34C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13457497, end_position_on_the_genomic_accession=13458073, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53583, CurrentID=0, Status='live', Symbol='Ugt37c1', Aliases='Dmel_CG8652, AC 004641A, CG8652, DmeUgt37c1, Dmel\\CG8652, EG:EG0003.4, LD15335', description='UDP-glycosyltransferase 37c1', other_designations='CG8652-PA|Ugt37c1-PA', map_location='53D12-53D12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16843296, end_position_on_the_genomic_accession=16845038, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  53584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53584, CurrentID=0, Status='live', Symbol='Ugt37b1', Aliases='Dmel_CG9481, AC 005285A, CG9481, DmeUgt37b1, Dmel\\CG9481', description='UDP-glycosyltransferase 37b1', other_designations='CG9481-PA|Ugt37b1-PA', map_location='26B11-26B11', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6225048, end_position_on_the_genomic_accession=6226842, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53585, CurrentID=0, Status='live', Symbol='Debcl', Aliases='Dmel_CG33134, BG1, BOK, Bcl-2, Bok, Bok(Debcl), CG12397, CG33134, DBok, DEBCL, Dbok/Drob-1/dBorg-1/dBok, Debcl/dBorg-1/dBok/Drob-1, Debcl/dBorg-1/dRob-1, Dmel\\CG33134, Drob-1, Drob-1/Debcl/dBORG-1, Drob-1/Debcl/dBorg-1/DBok, Drob-1/Debcl/dBorg-1/dBok, bcl-2, borg1/debcl, dBOK, dBORG-1/DROB-1/DEBCL/dBOK, dBok, dBorg-1, dBorg01, dBorg1, dBrog-1, dbok, dbok/debcl, dborg1, debc1, debcl, debcl/Borg1, debok, drob-1, Debcl', description='Death executioner Bcl-2 homologue', other_designations='Bcl-2|CG33134-PA|Debcl-PA|death executioner Bcl-2 homologue|death inducing Bcl-2-like protein|deborg', map_location='42B2-42C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6648654, end_position_on_the_genomic_accession=6652842, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=53586, CurrentID=0, Status='live', Symbol='Dhap-at', Aliases='Dmel_CG4625, 8-1, CG4625, DHAP-AT, Dmel\\CG4625, GPAT, Gpat', description='Dihydroxyacetone phosphate acyltransferase', other_designations='CG4625-PA|CG4625-PB|CG4625-PC|Dhap-at-PA|Dhap-at-PB|Dhap-at-PC|dihydroxyacetone phosphate acyltansferase|dihydroxyacetone phosphate acyltransferase|glycerol 3-phosphate acyltransferase|glycerol-3-phosphate acyltransferase|glycerol-3-phosphate-acyltransferase', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25485618, end_position_on_the_genomic_accession=25490936, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10178899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178899, CurrentID=0, Status='live', Symbol='CG42680', Aliases='Dmel_ Dmel\\CG42680', description='CG42680 gene product from transcript CG42680-RA', other_designations='CG42680-PA|uncharacterized protein', map_location='35B3-35B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14587920, end_position_on_the_genomic_accession=14588234, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178900, CurrentID=0, Status='live', Symbol='CG42758', Aliases='Dmel_ CG13479, Dmel\\CG42758', description='CG42758 gene product from transcript CG42758-RA', other_designations='CG42758-PA|uncharacterized protein', map_location='70E2-70E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14567440, end_position_on_the_genomic_accession=14568133, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178901, CurrentID=0, Status='live', Symbol='CG42659', Aliases='Dmel_ Dmel\\CG42659, noncoding_3557', description='CG42659 gene product from transcript CG42659-RB', other_designations='CG42659-PB|uncharacterized protein', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18007360, end_position_on_the_genomic_accession=18007842, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178905, CurrentID=0, Status='live', Symbol='CG42826', Aliases='Dmel_ CG6640, Dmel\\CG42826, Dmel_CG6640', description='CG42826 gene product from transcript CG42826-RA', other_designations='CG42826-PA|uncharacterized protein', map_location='67D8-67D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10099357, end_position_on_the_genomic_accession=10100881, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178906, CurrentID=0, Status='live', Symbol='CG42693', Aliases='Dmel_ Dmel\\CG42693', description='CG42693 gene product from transcript CG42693-RA', other_designations='CG42693-PA|uncharacterized protein', map_location='100D1-100D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31707870, end_position_on_the_genomic_accession=31708340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178908, CurrentID=0, Status='live', Symbol='CG42782', Aliases='Dmel_ Dmel\\CG42782', description='CG42782 gene product from transcript CG42782-RB', other_designations='CG42782-PA|CG42782-PB|uncharacterized protein', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12418797, end_position_on_the_genomic_accession=12419395, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178909, CurrentID=0, Status='live', Symbol='CG42830', Aliases='Dmel_ CG7527, Dmel\\CG42830, anon-AE003656.1', description='CG42830 gene product from transcript CG42830-RA', other_designations='CG42830-PA|anon-AE003656.1|uncharacterized protein', map_location='36D3-36D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17811953, end_position_on_the_genomic_accession=17812672, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178910, CurrentID=0, Status='live', Symbol='CG42834', Aliases='Dmel_ Dmel\\CG42834', description='CG42834 gene product from transcript CG42834-RA', other_designations='CG42834-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17418759, end_position_on_the_genomic_accession=17419294, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178911, CurrentID=0, Status='live', Symbol='CG42870', Aliases='Dmel_ Dmel\\CG42870, lincRNA.828', description='CG42870 gene product from transcript CG42870-RB', other_designations='CG42870-PB|uncharacterized protein', map_location='93F7-93F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21736933, end_position_on_the_genomic_accession=21737364, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178912, CurrentID=0, Status='live', Symbol='CG42779', Aliases='Dmel_ Dmel\\CG42779', description='CG42779 gene product from transcript CG42779-RB', other_designations='CG42779-PA|CG42779-PB|uncharacterized protein', map_location='89E13-89E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17121577, end_position_on_the_genomic_accession=17122087, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771792, CurrentID=0, Status='live', Symbol='His3:CG33830', Aliases='Dmel_CG33830, CG33830, Dmel\\CG33830', description='CG33830 gene product from transcript CG33830-RA', other_designations='CG33830-PA|His3:CG33830-PA|histone H3', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21479636, end_position_on_the_genomic_accession=21480164, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178914, CurrentID=0, Status='live', Symbol='CG42876', Aliases='Dmel_ Dmel\\CG42876', description='CG42876 gene product from transcript CG42876-RB', other_designations='CG42876-PA|CG42876-PB|uncharacterized protein', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15810571, end_position_on_the_genomic_accession=15812141, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178916, CurrentID=0, Status='live', Symbol='CG42822', Aliases='Dmel_ BP1011, CG34035, Dmel\\CG42822, Dmel_CG34035', description='CG42822 gene product from transcript CG42822-RA', other_designations='CG42822-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17413901, end_position_on_the_genomic_accession=17414607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10178923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178923, CurrentID=0, Status='live', Symbol='CG42753', Aliases='Dmel_ Dmel\\CG42753', description='CG42753 gene product from transcript CG42753-RA', other_designations='CG42753-PA|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19635279, end_position_on_the_genomic_accession=19635941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178926, CurrentID=0, Status='live', Symbol='CG42682', Aliases='Dmel_ Dmel\\CG42682', description='CG42682 gene product from transcript CG42682-RB', other_designations='CG42682-PB|uncharacterized protein', map_location='35B6-35B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14842358, end_position_on_the_genomic_accession=14843108, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178927, CurrentID=0, Status='live', Symbol='CG42851', Aliases='Dmel_ Dmel\\CG42851, chr2R_20696695_20697629.0', description='CG42851 gene product from transcript CG42851-RB', other_designations='CG42851-PB|uncharacterized protein', map_location='60E2-60E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24809325, end_position_on_the_genomic_accession=24810107, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178928, CurrentID=0, Status='live', Symbol='CG42655', Aliases='Dmel_ Dmel\\CG42655', description='CG42655 gene product from transcript CG42655-RB', other_designations='CG42655-PB|uncharacterized protein', map_location='76D6-76D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19952971, end_position_on_the_genomic_accession=19953712, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10178929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178929, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-Z', Aliases='Dmel_CG42770, CG42770, Dmel\\CG42770', description='CG42770 gene product from transcript CG42770-RA', other_designations='CG42770-PA|pre-mod(mdg4)-Z-PA', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21360176, end_position_on_the_genomic_accession=21362131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10178930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178930, CurrentID=0, Status='live', Symbol='CG42759', Aliases='Dmel_ Dmel\\CG42759', description='CG42759 gene product from transcript CG42759-RA', other_designations='CG42759-PA|uncharacterized protein', map_location='85F14-85F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10143070, end_position_on_the_genomic_accession=10143483, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178932, CurrentID=0, Status='live', Symbol='MtnE', Aliases='Dmel_CG42872, CG42872, DMG8-chr3R.29.007, Dmel\\CG42872', description='Metallothionein E', other_designations='CG42872-PA|CG42872-PB|MtnE-PA|MtnE-PB|metallothionein E', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20534021, end_position_on_the_genomic_accession=20534704, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178933, CurrentID=0, Status='live', Symbol='CG42841', Aliases='Dmel_ Dmel\\CG42841', description='CG42841 gene product from transcript CG42841-RB', other_designations='CG42841-PA|CG42841-PB|uncharacterized protein', map_location='62C1-62C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2003109, end_position_on_the_genomic_accession=2005651, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178934, CurrentID=0, Status='live', Symbol='CG42828', Aliases='Dmel_ CG6784, Dmel\\CG42828, Dmel_CG6784', description='CG42828 gene product from transcript CG42828-RB', other_designations='CG42828-PB|uncharacterized protein', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23373703, end_position_on_the_genomic_accession=23374245, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178935, CurrentID=0, Status='live', Symbol='CG42649', Aliases='Dmel_ Dmel\\CG42649', description='CG42649 gene product from transcript CG42649-RA', other_designations='CG42649-PA|uncharacterized protein', map_location='54C12-54C12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17599389, end_position_on_the_genomic_accession=17599607, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178939, CurrentID=0, Status='live', Symbol='CG42660', Aliases='Dmel_ Dmel\\CG42660', description='CG42660 gene product from transcript CG42660-RB', other_designations='CG42660-PA|CG42660-PB|uncharacterized protein', map_location='66A2-66A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7465509, end_position_on_the_genomic_accession=7468808, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  10178940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178940, CurrentID=0, Status='live', Symbol='CG42718', Aliases='Dmel_ Dmel\\CG42718', description='CG42718 gene product from transcript CG42718-RA', other_designations='CG42718-PA|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16320830, end_position_on_the_genomic_accession=16321272, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178941, CurrentID=0, Status='live', Symbol='CG42811', Aliases='Dmel_ CG18669, Dmel\\CG42811, Dmel_CG18669', description='CG42811 gene product from transcript CG42811-RA', other_designations='CG42811-PA|uncharacterized protein', map_location='95F15-95F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24366527, end_position_on_the_genomic_accession=24367220, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178943, CurrentID=0, Status='live', Symbol='CG42691', Aliases='Dmel_ Dmel\\CG42691', description='CG42691 gene product from transcript CG42691-RA', other_designations='CG42691-PA|CG42691-PB|uncharacterized protein', map_location='56E5-56E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19945731, end_position_on_the_genomic_accession=19946814, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  10178945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178945, CurrentID=0, Status='live', Symbol='CG42710', Aliases='Dmel_ Dmel\\CG42710, HDC01517', description='CG42710 gene product from transcript CG42710-RA', other_designations='CG42710-PA|uncharacterized protein', map_location='29F1-29F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8781799, end_position_on_the_genomic_accession=8782500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178946, CurrentID=0, Status='live', Symbol='CG42792', Aliases='Dmel_ Dmel\\CG42792, HDC02744', description='CG42792 gene product from transcript CG42792-RB', other_designations='CG42792-PA|CG42792-PB|uncharacterized protein', map_location='35D5-35D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15864929, end_position_on_the_genomic_accession=15865461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178948, CurrentID=0, Status='live', Symbol='CG42787', Aliases='Dmel_ Dmel\\CG42787, Mzt1', description='CG42787 gene product from transcript CG42787-RA', other_designations='CG42787-PA|MOZART1|mitotic spindle OrganiZing protein associated with a ring of g-Tubulin|uncharacterized protein', map_location='62E1-62E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2389265, end_position_on_the_genomic_accession=2389650, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771798, CurrentID=0, Status='live', Symbol='CG33689', Aliases='Dmel_ Dmel\\CG33689', description='CG33689 gene product from transcript CG33689-RB', other_designations='CG33689-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16200229, end_position_on_the_genomic_accession=16200926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10178950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178950, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-V', Aliases='Dmel_CG42776, CG42776, Dmel\\CG42776', description='CG42776 gene product from transcript CG42776-RB', other_designations='CG42776-PA|CG42776-PB|pre-mod(mdg4)-V-PA|pre-mod(mdg4)-V-PB', map_location='93D9-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21370249, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10178951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178951, CurrentID=0, Status='live', Symbol='CG42658', Aliases='Dmel_ Dmel\\CG42658', description='CG42658 gene product from transcript CG42658-RA', other_designations='CG42658-PA|uncharacterized protein', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2287025, end_position_on_the_genomic_accession=2287933, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10178954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178954, CurrentID=0, Status='live', Symbol='CG42798', Aliases='Dmel_ Dmel\\CG42798', description='CG42798 gene product from transcript CG42798-RB', other_designations='CG42798-PA|CG42798-PB|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17406678, end_position_on_the_genomic_accession=17407534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10178957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178957, CurrentID=0, Status='live', Symbol='CG42751', Aliases='Dmel_ Dmel\\CG42751', description='CG42751 gene product from transcript CG42751-RA', other_designations='CG42751-PA|uncharacterized protein', map_location='33A4-33A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11853430, end_position_on_the_genomic_accession=11853925, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178958, CurrentID=0, Status='live', Symbol='CG42814', Aliases='Dmel_ CG31063, CG6001, Dmel\\CG42814, Dmel_CG31063', description='CG42814 gene product from transcript CG42814-RA', other_designations='CG42814-PA|uncharacterized protein', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27228673, end_position_on_the_genomic_accession=27229468, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178960, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-AE', Aliases='Dmel_CG42774, CG42774, Dmel\\CG42774', description='CG42774 gene product from transcript CG42774-RB', other_designations='CG42774-PA|CG42774-PB|pre-mod(mdg4)-AE-PA|pre-mod(mdg4)-AE-PB', map_location='93D8-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21368279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178961, CurrentID=0, Status='live', Symbol='CG42808', Aliases='Dmel_ CG13335, CG6183, Dmel\\CG42808, Dmel_CG13335', description='CG42808 gene product from transcript CG42808-RA', other_designations='CG42808-PA|uncharacterized protein', map_location='50B2-50B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13573618, end_position_on_the_genomic_accession=13574257, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178963, CurrentID=0, Status='live', Symbol='CG42764', Aliases='Dmel_ Dmel\\CG42764', description='CG42764 gene product from transcript CG42764-RB', other_designations='CG42764-PA|CG42764-PB|CG42764-PC|uncharacterized protein', map_location='77C4-77C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20464714, end_position_on_the_genomic_accession=20466041, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10178965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178965, CurrentID=0, Status='live', Symbol='CG42833', Aliases='Dmel_ Dmel\\CG42833', description='CG42833 gene product from transcript CG42833-RA', other_designations='CG42833-PA|uncharacterized protein', map_location='77A2-77A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20200093, end_position_on_the_genomic_accession=20200487, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740660, CurrentID=0, Status='live', Symbol='CG12828', Aliases='Dmel_ Dmel\\CG12828', description='CG12828 gene product from transcript CG12828-RA', other_designations='CG12828-PA|uncharacterized protein', map_location='42F3-42F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7149174, end_position_on_the_genomic_accession=7149677, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  7354437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354437, CurrentID=0, Status='live', Symbol='CG42354', Aliases='Dmel_ CG9216, Dmel\\CG42354, Dmel_CG9216', description='CG42354 gene product from transcript CG42354-RB', other_designations='CG42354-PA|CG42354-PB|CG42354-PC|CG42354-PD|CG42354-PE|CG42354-PF|uncharacterized protein', map_location='14A9-14A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16093059, end_position_on_the_genomic_accession=16104946, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  3771803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771803, CurrentID=0, Status='live', Symbol='His1:CG33852', Aliases='Dmel_CG33852, CG33852, Dmel\\CG33852', description='CG33852 gene product from transcript CG33852-RA', other_designations='CG33852-PA|His1:CG33852-PA|histone H1', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21516773, end_position_on_the_genomic_accession=21517795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740537, CurrentID=0, Status='live', Symbol='CG34302', Aliases='Dmel_ Dmel\\CG34302', description='CG34302 gene product from transcript CG34302-RB', other_designations='CG34302-PB|uncharacterized protein', map_location='85E10-85E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9822999, end_position_on_the_genomic_accession=9823823, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36254, CurrentID=0, Status='live', Symbol='Ir48b', Aliases='Dmel_CG13195, CG13195, DmelIR48b, Dmel\\CG13195, IR48b', description='Ionotropic receptor 48b', other_designations='CG13195-PA|Ir48b-PA|ionotropic receptor 48b', map_location='48C2-48C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11700647, end_position_on_the_genomic_accession=11702449, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771809, CurrentID=0, Status='live', Symbol='His2B:CG33872', Aliases='Dmel_CG33872, CG33872, Dmel\\CG33872', description='CG33872 gene product from transcript CG33872-RA', other_designations='CG33872-PA|His2B:CG33872-PA|histone H2B', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21527618, end_position_on_the_genomic_accession=21528212, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674090, CurrentID=0, Status='live', Symbol='CG42597', Aliases='Dmel_ Dmel\\CG42597', description='CG42597 gene product from transcript CG42597-RA', other_designations='CG42597-PA|uncharacterized protein', map_location='40E1-40E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22064604, end_position_on_the_genomic_accession=22065343, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  38547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38547, CurrentID=0, Status='live', Symbol='CG12607', Aliases='Dmel_ BcDNA:RE67729, Dmel\\CG12607', description='CG12607 gene product from transcript CG12607-RC', other_designations='CG12607-PB|CG12607-PC|CG12607-PD|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4446573, end_position_on_the_genomic_accession=4448072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771816, CurrentID=0, Status='live', Symbol='His1:CG33816', Aliases='Dmel_CG33816, CG33816, Dmel\\CG33816', description='CG33816 gene product from transcript CG33816-RA', other_designations='CG33816-PA|His1:CG33816-PA|histone H1', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21456113, end_position_on_the_genomic_accession=21457135, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771818, CurrentID=0, Status='live', Symbol='His1:CG33804', Aliases='Dmel_CG33804, CG33804, Dmel\\CG33804', description='CG33804 gene product from transcript CG33804-RA', other_designations='CG33804-PA|His1:CG33804-PA|histone H1', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21426916, end_position_on_the_genomic_accession=21427938, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674092, CurrentID=0, Status='live', Symbol='CG42507', Aliases='Dmel_ Dmel\\CG42507', description='CG42507 gene product from transcript CG42507-RB', other_designations='CG42507-PA|CG42507-PB|uncharacterized protein', map_location='70F4-70F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14747929, end_position_on_the_genomic_accession=14757962, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  38329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38329, CurrentID=0, Status='live', Symbol='CG1143', Aliases='Dmel_ Dmel\\CG1143', description='CG1143 gene product from transcript CG1143-RB', other_designations='CG1143-PA|CG1143-PB', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2645368, end_position_on_the_genomic_accession=2647192, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674093, CurrentID=0, Status='live', Symbol='Sfp84E', Aliases='Dmel_CG42484, CG42484, Dmel\\CG42484', description='Seminal fluid protein 84E', other_designations='CG42484-PA|Sfp84E-PA|seminal fluid protein 84E', map_location='84E6-84E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7979588, end_position_on_the_genomic_accession=7980322, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3771825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771825, CurrentID=0, Status='live', Symbol='CG33654', Aliases='Dmel_ Dmel\\CG33654', description='CG33654 gene product from transcript CG33654-RA', other_designations='CG33654-PA|uncharacterized protein', map_location='85D26-85D26', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9582515, end_position_on_the_genomic_accession=9583222, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19835263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835263, CurrentID=0, Status='live', Symbol='CG44954', Aliases='Dmel_ Dmel\\CG44954, lincRNA.900', description='CG44954 gene product from transcript CG44954-RA', other_designations='CG44954-PA|uncharacterized protein', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30618626, end_position_on_the_genomic_accession=30621906, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26067042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067042, CurrentID=0, Status='live', Symbol='CG45764', Aliases='Dmel_ Dmel\\CG45764', description='CG45764 gene product from transcript CG45764-RA', other_designations='CG45764-PA|uncharacterized protein', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3289209, end_position_on_the_genomic_accession=3290215, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  7354472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=7354472, CurrentID=0, Status='live', Symbol='CG42340', Aliases='Dmel_ CG12543, CG3367, Dmel\\CG42340, Dmel_CG12543, Dmel_CG3367', description='CG42340 gene product from transcript CG42340-RD', other_designations='CG42340-PB|CG42340-PC|CG42340-PD|uncharacterized protein', map_location='6A3-6B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6423123, end_position_on_the_genomic_accession=6524323, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3771838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771838, CurrentID=0, Status='live', Symbol='His1:CG33813', Aliases='Dmel_CG33813, CG33813, Dmel\\CG33813', description='CG33813 gene product from transcript CG33813-RA', other_designations='CG33813-PA|His1:CG33813-PA|histone H1', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21451052, end_position_on_the_genomic_accession=21452074, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771840, CurrentID=0, Status='live', Symbol='CG33768', Aliases='Dmel_ CG14458, Dmel\\CG33768', description='CG33768 gene product from transcript CG33768-RA', other_designations='CG33768-PA|uncharacterized protein', map_location='79F3-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22647187, end_position_on_the_genomic_accession=22647748, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318222, CurrentID=0, Status='live', Symbol='CG32811', Aliases='Dmel_ Dmel\\CG32811', description='CG32811 gene product from transcript CG32811-RB', other_designations='CG32811-PB|uncharacterized protein', map_location='1F2-1F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1297984, end_position_on_the_genomic_accession=1298387, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318223, CurrentID=0, Status='live', Symbol='CG32814', Aliases='Dmel_ Dmel\\CG32814', description='CG32814 gene product from transcript CG32814-RB', other_designations='CG32814-PB|uncharacterized protein', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1019795, end_position_on_the_genomic_accession=1026700, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318224, CurrentID=0, Status='live', Symbol='CG32815', Aliases='Dmel_ CG16994, CG18830, Dmel\\CG32815', description='CG32815 gene product from transcript CG32815-RD', other_designations='CG32815-PC|CG32815-PD|uncharacterized protein', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=932466, end_position_on_the_genomic_accession=939424, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36299, CurrentID=0, Status='live', Symbol='CG8378', Aliases='Dmel_ BcDNA:LD29892, Dmel\\CG8378', description='CG8378 gene product from transcript CG8378-RA', other_designations='CG8378-PA', map_location='48E2-48E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12140631, end_position_on_the_genomic_accession=12143272, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740644, CurrentID=0, Status='live', Symbol='CG34288', Aliases='Dmel_ Dmel\\CG34288', description='CG34288 gene product from transcript CG34288-RB', other_designations='CG34288-PB|uncharacterized protein', map_location='94A14-94A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22446763, end_position_on_the_genomic_accession=22447263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3771854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771854, CurrentID=0, Status='live', Symbol='His4:CG33869', Aliases='Dmel_CG33869, CG33869, Dmel\\CG33869', description='CG33869 gene product from transcript CG33869-RA', other_designations='CG33869-PA|His4:CG33869-PA|histone H4', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21534085, end_position_on_the_genomic_accession=21534610, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771856, CurrentID=0, Status='live', Symbol='CheA84a', Aliases='Dmel_CG33779, CG33779, Dmel\\CG33779', description='Chemosensory protein A 84a', other_designations='CG33779-PA|CheA84a-PA|chemosensory protein A 84a', map_location='84F3-84F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8204195, end_position_on_the_genomic_accession=8204830, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674100, CurrentID=0, Status='live', Symbol='CG42511', Aliases='Dmel_ BBIP1, Dmel\\CG42511', description='CG42511 gene product from transcript CG42511-RA', other_designations='CG42511-PA|uncharacterized protein', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9404385, end_position_on_the_genomic_accession=9407857, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3771860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771860, CurrentID=0, Status='live', Symbol='CG33752', Aliases='Dmel_ Dmel\\CG33752', description='CG33752 gene product from transcript CG33752-RA', other_designations='CG33752-PA|uncharacterized protein', map_location='49B3-49B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12545689, end_position_on_the_genomic_accession=12546301, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36323, CurrentID=0, Status='live', Symbol='CG13171', Aliases='Dmel_ Dmel\\CG13171', description='CG13171 gene product from transcript CG13171-RA', other_designations='CG13171-PA|uncharacterized protein', map_location='48E10-48E10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12214010, end_position_on_the_genomic_accession=12214395, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  250710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=250710, CurrentID=0, Status='live', Symbol='Proc', Aliases='Dmel_CG7105, CG7105, Dmel\\CG7105, PROC, PROCT, Prctt, proct, Proc', description='Proctolin', other_designations='CG7105-PA|CG7105-PB|Proc-PA|Proc-PB|proctolin', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7864642, end_position_on_the_genomic_accession=7871432, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36325, CurrentID=0, Status='live', Symbol='CG13170', Aliases='Dmel_ Dmel\\CG13170', description='CG13170 gene product from transcript CG13170-RA', other_designations='CG13170-PA|uncharacterized protein', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12240399, end_position_on_the_genomic_accession=12240813, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36327, CurrentID=0, Status='live', Symbol='CG13168', Aliases='Dmel_ Dmel\\CG13168', description='CG13168 gene product from transcript CG13168-RC', other_designations='CG13168-PB|CG13168-PC|CG13168-PD|uncharacterized protein', map_location='48F1-48F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12253992, end_position_on_the_genomic_accession=12255701, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  250735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=250735, CurrentID=0, Status='live', Symbol='fdl', Aliases='Dmel_CG8824, CG8824, Dmel\\CG8824, FDL, Fdl, l(2)k17040', description='fused lobes', other_designations='CG8824-PB|CG8824-PC|CG8824-PE|CG8824-PF|fdl-PB|fdl-PC|fdl-PE|fdl-PF', map_location='49A9-49A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12487451, end_position_on_the_genomic_accession=12507333, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  250736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=250736, CurrentID=0, Status='live', Symbol='bwa', Aliases='Dmel_CG13969, 38B.6, CG13969, Dmel\\CG13969, dacer', description='brain washing', other_designations='CG13969-PA|CG13969-PB|brainwashing|bwa-PA|bwa-PB', map_location='38B2-38B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20066391, end_position_on_the_genomic_accession=20073719, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31264, CurrentID=0, Status='live', Symbol='CG2709', Aliases='Dmel_ Dmel\\CG2709, EG:95B7.5', description='CG2709 gene product from transcript CG2709-RA', other_designations='CG2709-PA', map_location='3B3-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2726678, end_position_on_the_genomic_accession=2727568, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36338, CurrentID=0, Status='live', Symbol='CG8490', Aliases='Dmel_ Dmel\\CG8490', description='CG8490 gene product from transcript CG8490-RB', other_designations='CG8490-PA|CG8490-PB', map_location='48F10-48F10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12335316, end_position_on_the_genomic_accession=12336205, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771891, CurrentID=0, Status='live', Symbol='His2B:CG33878', Aliases='Dmel_CG33878, CG33878, Dmel\\CG33878', description='CG33878 gene product from transcript CG33878-RA', other_designations='CG33878-PA|His2B:CG33878-PA|histone H2B', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21512887, end_position_on_the_genomic_accession=21513481, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771892, CurrentID=0, Status='live', Symbol='gdl-ORF39', Aliases='Dmel_CG33755, CG33088, CG33755, CG7268, Dmel\\CG33755, ORF39', description='CG33755 gene product from transcript CG33755-RA', other_designations='CG33755-PA|gdl-ORF39-PA', map_location='71D3-71D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15509445, end_position_on_the_genomic_accession=15510837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  250811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=250811, CurrentID=0, Status='live', Symbol='bru-2', Aliases='Dmel_CG43065, Bru-2, CG17008, CG31761, CG31763, CG43065, CG6319, Dm Bru-2, Dmel\\CG43065, Dmel_CG31761, Dmel_CG31763', description='bruno-2', other_designations='CG43065-PA|CG43065-PD|CG43065-PE|CG43065-PF|CG43065-PG|CG43065-PH|CG43065-PI|CG43065-PJ|CG43065-PK|CG43065-PL|CG43065-PM|bru-2-PA|bru-2-PD|bru-2-PE|bru-2-PF|bru-2-PG|bru-2-PH|bru-2-PI|bru-2-PJ|bru-2-PK|bru-2-PL|bru-2-PM|bruno2', map_location='33E1-33E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12344219, end_position_on_the_genomic_accession=12397528, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  3771893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771893, CurrentID=0, Status='live', Symbol='His4:CG33909', Aliases='Dmel_CG33909, CG33909, Dmel\\CG33909', description='CG33909 gene product from transcript CG33909-RA', other_designations='CG33909-PA|His4:CG33909-PA|histone H4', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21419531, end_position_on_the_genomic_accession=21419943, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740688, CurrentID=0, Status='live', Symbol='RpS28-like', Aliases='Dmel_CG34182, CG34182, Dmel\\CG34182', description='Ribosomal protein S28-like', other_designations='CG34182-PA|RpS28-like-PA|ribosomal protein S28-like', map_location='30B3-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9436804, end_position_on_the_genomic_accession=9437227, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768912, CurrentID=0, Status='live', Symbol='CG33299', Aliases='Dmel_ Dmel\\CG33299', description='CG33299 gene product from transcript CG33299-RB', other_designations='CG33299-PB|CG33299-PC|uncharacterized protein', map_location='30C2-30C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9656600, end_position_on_the_genomic_accession=9657424, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771905, CurrentID=0, Status='live', Symbol='CG43736', Aliases='Dmel_ CG14433, CG32737, CG32737 ORF A, CG32737-ORFA, CG32737-ORFB, CG33691, CG33692, CG4564, CG4571, Dmel\\CG43736, Dmel_CG33691, Dmel_CG33692', description='CG43736 gene product from transcript CG43736-RK', other_designations='CG43736-PA|CG43736-PB|CG43736-PC|CG43736-PD|CG43736-PE|CG43736-PF|CG43736-PG|CG43736-PI|CG43736-PJ|CG43736-PK|CG43736-PL|uncharacterized protein', map_location='6E2-6E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6863888, end_position_on_the_genomic_accession=6881580, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  318235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318235, CurrentID=0, Status='live', Symbol='CG32829', Aliases='Dmel_ Dmel\\CG32829', description='CG32829 gene product from transcript CG32829-RA', other_designations='CG32829-PA|uncharacterized protein', map_location='27C4-27C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6915581, end_position_on_the_genomic_accession=6916445, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771908, CurrentID=0, Status='live', Symbol='His4:CG33905', Aliases='Dmel_CG33905, CG33905, Dmel\\CG33905', description='CG33905 gene product from transcript CG33905-RA', other_designations='CG33905-PA|His4:CG33905-PA|histone H4', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21434703, end_position_on_the_genomic_accession=21435115, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771910, CurrentID=0, Status='live', Symbol='His1:CG33828', Aliases='Dmel_CG33828, CG33828, Dmel\\CG33828', description='CG33828 gene product from transcript CG33828-RA', other_designations='CG33828-PA|His1:CG33828-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21476302, end_position_on_the_genomic_accession=21477324, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36359, CurrentID=0, Status='live', Symbol='CG13154', Aliases='Dmel_ Dmel\\CG13154', description='CG13154 gene product from transcript CG13154-RA', other_designations='CG13154-PA|uncharacterized protein', map_location='49A4-49A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12426883, end_position_on_the_genomic_accession=12427861, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3771912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771912, CurrentID=0, Status='live', Symbol='His1:CG33849', Aliases='Dmel_CG33849, CG33849, Dmel\\CG33849', description='CG33849 gene product from transcript CG33849-RA', other_designations='CG33849-PA|His1:CG33849-PA|histone H1', map_location='39D5-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21511728, end_position_on_the_genomic_accession=21512750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771913, CurrentID=0, Status='live', Symbol='CG33645', Aliases='Dmel_ Dmel\\CG33645', description='CG33645 gene product from transcript CG33645-RB', other_designations='CG33645-PB|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13532872, end_position_on_the_genomic_accession=13533454, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768993, CurrentID=0, Status='live', Symbol='CG33290', Aliases='Dmel_ Dmel\\CG33290', description='CG33290 gene product from transcript CG33290-RA', other_designations='CG33290-PA|uncharacterized protein', map_location='78D5-78D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21542359, end_position_on_the_genomic_accession=21543012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771914, CurrentID=0, Status='live', Symbol='CG33796', Aliases='Dmel_ Dmel\\CG33796', description='CG33796 gene product from transcript CG33796-RA', other_designations='CG33796-PA|uncharacterized protein', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16169455, end_position_on_the_genomic_accession=16170106, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674111, CurrentID=0, Status='live', Symbol='CG42445', Aliases='Dmel_ CG7498, Dmel\\CG42445', description='CG42445 gene product from transcript CG42445-RA', other_designations='CG42445-PA|CG42445-PB|uncharacterized protein', map_location='66B10-66B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8154681, end_position_on_the_genomic_accession=8158900, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771915, CurrentID=0, Status='live', Symbol='CG33672', Aliases='Dmel_ CG16804, CG16804a, CG33009, CG33009-ORFA, CG33009ORFA, CG8810, Dmel\\CG33672, Dromel_CG33672_FBtr0091651_uORF', description='CG33672 gene product from transcript CG33672-RA', other_designations='CG33672-PA|uncharacterized protein', map_location='49B10-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12634166, end_position_on_the_genomic_accession=12636141, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318237, CurrentID=0, Status='live', Symbol='CG32832', Aliases='Dmel_ Dmel\\CG32832, NEST:bs10d12, NEST:bs21e09, bs21e09.y1', description='CG32832 gene product from transcript CG32832-RB', other_designations='CG32832-PA|CG32832-PB|uncharacterized protein', map_location='36B3-36B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16878326, end_position_on_the_genomic_accession=16879290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771916, CurrentID=0, Status='live', Symbol='Gr64e', Aliases='Dmel_CG32258, CG14988, CG32258, CG33157, DmGr64e, Dmel64e, Dmel\\CG32258, GR64e, Gr64, Gr64A2, gr64e', description='Gustatory receptor 64e', other_designations='CG32258-PA|CG32258-PB|Gr64e-PA|Gr64e-PB|gustatory receptor 64A2|gustatory receptor 64e', map_location='64A4-64A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4032141, end_position_on_the_genomic_accession=4035703, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14462557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462557, CurrentID=0, Status='live', Symbol='CG43647', Aliases='Dmel_ Dmel\\CG43647, lincRNA.258, mlncRNA42E5-1', description='CG43647 gene product from transcript CG43647-RA', other_designations='CG43647-PA|uncharacterized protein', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7020133, end_position_on_the_genomic_accession=7020746, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3771920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771920, CurrentID=0, Status='live', Symbol='PGRP-LD', Aliases='Dmel_CG33717, CG32912, CG33717, CG5523, Dmel\\CG33717', description='Peptidoglycan recognition protein LD', other_designations='CG33717-PA|CG33717-PB|CG33717-PC|CG33717-PD|PGRP-LD-PA|PGRP-LD-PB|PGRP-LD-PC|PGRP-LD-PD|peptidoglycan recognition protein LD|peptidoglycan recognition protein-LD', map_location='64E7-64E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5780051, end_position_on_the_genomic_accession=5784685, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  8674113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674113, CurrentID=0, Status='live', Symbol='CG42571', Aliases='Dmel_ Dmel\\CG42571, chr3L.16.093a', description='CG42571 gene product from transcript CG42571-RB', other_designations='CG42571-PB|uncharacterized protein', map_location='71F2-71F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15716453, end_position_on_the_genomic_accession=15717749, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  5740542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740542, CurrentID=0, Status='live', Symbol='CG41561', Aliases='Dmel_ Dmel\\CG41561', description='CG41561 gene product from transcript CG41561-RB', other_designations='CG41561-PB|uncharacterized protein', map_location='', chromosome='3CEN', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  4379817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379817, CurrentID=0, Status='live', Symbol='RunxA', Aliases='Dmel_CG34145, CG1379, CG15453, CG34145, Dmel\\CG34145', description='CG34145 gene product from transcript CG34145-RB', other_designations='CG34145-PB|CG34145-PC|CG34145-PD|RunxA-PB|RunxA-PC|RunxA-PD', map_location='19D3-19E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20590252, end_position_on_the_genomic_accession=20623773, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3771934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771934, CurrentID=0, Status='live', Symbol='cana', Aliases='Dmel_CG33694, CENP-E ana, CENPana, CG32955, CG33694, CG4831, CP15516, Cana, Cenp-E, Cenp_E, DmCana, Dmel\\CG33694, KIF 10, KIF10, l(2)k07716', description='CENP-ana', other_designations='CG33694-PA|CG33694-PB|CG33694-PC|cana-PA|cana-PB|cana-PC', map_location='32E2-32E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11269780, end_position_on_the_genomic_accession=11277846, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  4379834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379834, CurrentID=0, Status='live', Symbol='stnB', Aliases='Dmel_CG12473, CG12473, CG12473/CG12500, CG40306, Dmel\\CG12473, ORF2, STNB, StnB, stn, stn-B', description='stoned B', other_designations='CG12473-PE|CG12473-PF|CG12473-PG|CG12473-PH|StonedB|stnB-PE|stnB-PF|stnB-PG|stnB-PH|stoned|stoned-B', map_location='20F3-20F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22980722, end_position_on_the_genomic_accession=23004648, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3771935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771935, CurrentID=0, Status='live', Symbol='His4:CG33901', Aliases='Dmel_CG33901, CG33901, Dmel\\CG33901', description='CG33901 gene product from transcript CG33901-RA', other_designations='CG33901-PA|His4:CG33901-PA|histone H4', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21453790, end_position_on_the_genomic_accession=21454202, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  4379842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379842, CurrentID=0, Status='live', Symbol='CG34112', Aliases='Dmel_ CG14649, CG41471, Dmel\\CG34112, Dmel_CG41471', description='CG34112 gene product from transcript CG34112-RB', other_designations='CG34112-PA|CG34112-PB|CG34112-PC|CG34112-PD|uncharacterized protein', map_location='82A5-82A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4391711, end_position_on_the_genomic_accession=4394259, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  4379844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379844, CurrentID=0, Status='live', Symbol='CG34113', Aliases='Dmel_ 34113, CG10989, CG14677, CG14678, CT34460, Dm_3R:9314, Dmel\\CG34113', description='CG34113 gene product from transcript CG34113-RP', other_designations='CG34113-PO|CG34113-PP|uncharacterized protein', map_location='83C6-83D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5862865, end_position_on_the_genomic_accession=5959555, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  4379850 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379850, CurrentID=0, Status='live', Symbol='CG34135', Aliases='Dmel_ Dmel\\CG34135', description='CG34135 gene product from transcript CG34135-RC', other_designations='CG34135-PB|CG34135-PC|CG34135-PD|uncharacterized protein', map_location='85A10-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8749206, end_position_on_the_genomic_accession=8768191, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  4379851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379851, CurrentID=0, Status='live', Symbol='CG34117', Aliases='Dmel_ Dmel\\CG34117', description='CG34117 gene product from transcript CG34117-RA', other_designations='CG34117-PA|uncharacterized protein', map_location='85D15-85D15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9399570, end_position_on_the_genomic_accession=9400053, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  4379852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379852, CurrentID=0, Status='live', Symbol='CG34107', Aliases='Dmel_ CG12816, Dmel\\CG34107', description='CG34107 gene product from transcript CG34107-RA', other_designations='CG34107-PA|uncharacterized protein', map_location='86A1-86A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10224909, end_position_on_the_genomic_accession=10225587, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771938, CurrentID=0, Status='live', Symbol='His4:CG33899', Aliases='Dmel_CG33899, CG33899, Dmel\\CG33899', description='CG33899 gene product from transcript CG33899-RA', other_designations='CG33899-PA|His4:CG33899-PA|histone H4', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21458853, end_position_on_the_genomic_accession=21459265, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  4379854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379854, CurrentID=0, Status='live', Symbol='CG34114', Aliases='Dmel_ 34114, CG14698, CG4695, Dmel\\CG34114', description='CG34114 gene product from transcript CG34114-RB', other_designations='CG34114-PB|CG34114-PC|uncharacterized protein', map_location='86C15-86D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10972381, end_position_on_the_genomic_accession=11050175, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  4379856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379856, CurrentID=0, Status='live', Symbol='CG34138', Aliases='Dmel_ Dmel\\CG34138', description='CG34138 gene product from transcript CG34138-RA', other_designations='CG34138-PA|uncharacterized protein', map_location='92A11-92A11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19634949, end_position_on_the_genomic_accession=19635879, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  4379860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379860, CurrentID=0, Status='live', Symbol='Muted', Aliases='Dmel_CG34131, CG34131, CG7071, Dmel\\CG34131, Dmel_CG7071, Dromel_CG7071_FBtr0084211_uORF, NP_001036744, muted', description='Muted ortholog (H. sapiens)', other_designations='CG34131-PA|CG34131-PB|Muted-PA|Muted-PB|muted|muted ortholog', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22362620, end_position_on_the_genomic_accession=22364580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  4379861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379861, CurrentID=0, Status='live', Symbol='CG34149', Aliases='Dmel_ CG12757, Dmel\\CG34149', description='CG34149 gene product from transcript CG34149-RA', other_designations='CG34149-PA|CG34149-PB|uncharacterized protein', map_location='94A6-94A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22380128, end_position_on_the_genomic_accession=22381253, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379863, CurrentID=0, Status='live', Symbol='CG34150', Aliases='Dmel_ Dmel\\CG34150', description='CG34150 gene product from transcript CG34150-RB', other_designations='CG34150-PB|uncharacterized protein', map_location='96A23-96A23', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24814262, end_position_on_the_genomic_accession=24814930, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  4379864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379864, CurrentID=0, Status='live', Symbol='lobo', Aliases='Dmel_CG34110, CG13661, CG13662, CG34110, Dmel\\CG34110', description='lost boys', other_designations='CG34110-PC|lobo-PC', map_location='96C2-96C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25169565, end_position_on_the_genomic_accession=25193776, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  4379865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379865, CurrentID=0, Status='live', Symbol='BG642167', Aliases='Dmel_CG34103, 1a8(1a8), CG34103, Dmel\\CG34103, anon-SAGE:Wang-011', description='CG34103 gene product from transcript CG34103-RA', other_designations='BG642167-PA|CG34103-PA|anon-SAGE:Wang-011', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25806445, end_position_on_the_genomic_accession=25806961, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379866, CurrentID=0, Status='live', Symbol='CG34130', Aliases='Dmel_ CG6069, Dmel\\CG34130, SPH192', description='CG34130 gene product from transcript CG34130-RA', other_designations='CG34130-PA|uncharacterized protein', map_location='97A4-97A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26189004, end_position_on_the_genomic_accession=26189977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  4379867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379867, CurrentID=0, Status='live', Symbol='CG34155', Aliases='Dmel_ Dmel\\CG34155', description='CG34155 gene product from transcript CG34155-RA', other_designations='CG34155-PA|uncharacterized protein', map_location='99F8-99F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30554554, end_position_on_the_genomic_accession=30560199, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  4379870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379870, CurrentID=0, Status='live', Symbol='CG34141', Aliases='Dmel_ Dmel\\CG34141', description='CG34141 gene product from transcript CG34141-RB', other_designations='CG34141-PB|CG34141-PC|CG34141-PD|CG34141-PE|uncharacterized protein', map_location='45C1-45C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9260278, end_position_on_the_genomic_accession=9263971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  4379878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379878, CurrentID=0, Status='live', Symbol='mRpL34', Aliases='Dmel_CG34147, CG34147, Dmel\\CG34147', description='mitochondrial ribosomal protein L34', other_designations='CG34147-PA|mRpL34-PA', map_location='52E4-52E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16098948, end_position_on_the_genomic_accession=16099337, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379879, CurrentID=0, Status='live', Symbol='Acp54A1', Aliases='Dmel_CG34098, CG34098, Dmel\\CG34098', description='Accessory gland protein 54A1', other_designations='Acp54A1-PA|CG34098-PA|accessory gland protein 54A1', map_location='54A1-54A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17133445, end_position_on_the_genomic_accession=17133750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4379880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379880, CurrentID=0, Status='live', Symbol='CG34115', Aliases='Dmel_ Dmel\\CG34115', description='CG34115 gene product from transcript CG34115-RA', other_designations='CG34115-PA|uncharacterized protein', map_location='57B5-57B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20975488, end_position_on_the_genomic_accession=20976934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  4379881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379881, CurrentID=0, Status='live', Symbol='CG34105', Aliases='Dmel_ Dmel\\CG34105', description='CG34105 gene product from transcript CG34105-RA', other_designations='CG34105-PA|uncharacterized protein', map_location='59F7-59F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23749318, end_position_on_the_genomic_accession=23749891, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178785, CurrentID=0, Status='live', Symbol='CG42832', Aliases='Dmel_ Dmel\\CG42832', description='CG42832 gene product from transcript CG42832-RA', other_designations='CG42832-PA|uncharacterized protein', map_location='68C15-68C15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11610135, end_position_on_the_genomic_accession=11610386, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  4379884 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379884, CurrentID=0, Status='live', Symbol='CG34132', Aliases='Dmel_ Dmel\\CG34132', description='CG34132 gene product from transcript CG34132-RA', other_designations='CG34132-PA|uncharacterized protein', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7885503, end_position_on_the_genomic_accession=7886109, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379885, CurrentID=0, Status='live', Symbol='CG14042', Aliases='Dmel_ CG14041, Dmel\\CG14042, Dmel_CG14041, NEST:bs13e02', description='CG14042 gene product from transcript CG14042-RB', other_designations='CG14042-PA|CG14042-PB|CG14042-PC|uncharacterized protein', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5037709, end_position_on_the_genomic_accession=5040818, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379886, CurrentID=0, Status='live', Symbol='CG34125', Aliases='Dmel_ CG31652, Dmel\\CG34125', description='CG34125 gene product from transcript CG34125-RA', other_designations='CG34125-PA|uncharacterized protein', map_location='25B5-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4965673, end_position_on_the_genomic_accession=4966399, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379887, CurrentID=0, Status='live', Symbol='CG34124', Aliases='Dmel_ CG31652, Dmel\\CG34124', description='CG34124 gene product from transcript CG34124-RA', other_designations='CG34124-PA|uncharacterized protein', map_location='25B4-25B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4958411, end_position_on_the_genomic_accession=4965430, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  4379888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379888, CurrentID=0, Status='live', Symbol='BG642163', Aliases='Dmel_CG34102(1a3), CG34102, Dmel\\CG34102', description='CG34102 gene product from transcript CG34102-RB', other_designations='BG642163-PA|BG642163-PB|CG34102-PA|CG34102-PB', map_location='25A2-25A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4641296, end_position_on_the_genomic_accession=4642872, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379889, CurrentID=0, Status='live', Symbol='CG34134', Aliases='Dmel_ CG8475, Dmel\\CG34134', description='CG34134 gene product from transcript CG34134-RA', other_designations='CG34134-PA|uncharacterized protein', map_location='28F1-28F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8190297, end_position_on_the_genomic_accession=8190781, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  4379890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379890, CurrentID=0, Status='live', Symbol='CG34109', Aliases='Dmel_ CG12441, CG13128, CG13129, Dmel\\CG34109', description='CG34109 gene product from transcript CG34109-RC', other_designations='CG34109-PC|uncharacterized protein', map_location='31A2-31A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10094088, end_position_on_the_genomic_accession=10125736, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  4379891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379891, CurrentID=0, Status='live', Symbol='CG34106', Aliases='Dmel_ Dmel\\CG34106', description='CG34106 gene product from transcript CG34106-RB', other_designations='CG34106-PA|CG34106-PB|uncharacterized protein', map_location='36C2-36C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17131024, end_position_on_the_genomic_accession=17173118, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  4379892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379892, CurrentID=0, Status='live', Symbol='CG34136', Aliases='Dmel_ Dmel\\CG34136', description='CG34136 gene product from transcript CG34136-RA', other_designations='CG34136-PA|uncharacterized protein', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21096922, end_position_on_the_genomic_accession=21098932, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771945, CurrentID=0, Status='live', Symbol='CG33922', Aliases='Dmel_ Dmel\\CG33922', description='CG33922 gene product from transcript CG33922-RA', other_designations='CG33922-PA|uncharacterized protein', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1343443, end_position_on_the_genomic_accession=1344171, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  4379897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379897, CurrentID=0, Status='live', Symbol='CG41434', Aliases='Dmel_ Dmel\\CG41434, lincRNA.226', description='CG41434 gene product from transcript CG41434-RC', other_designations='CG41434-PB|CG41434-PC|uncharacterized protein', map_location='40D4-40D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22037860, end_position_on_the_genomic_accession=22039772, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  4379900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379900, CurrentID=0, Status='live', Symbol='CG34137', Aliases='Dmel_ Dmel\\CG34137', description='CG34137 gene product from transcript CG34137-RA', other_designations='CG34137-PA|CG34137-PB|uncharacterized protein', map_location='40F1-40F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22151623, end_position_on_the_genomic_accession=22152640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771946, CurrentID=0, Status='live', Symbol='CG13175', Aliases='Dmel_ CG33154, Dmel\\CG13175', description='CG13175 gene product from transcript CG13175-RA', other_designations='CG13175-PA|uncharacterized protein', map_location='48E6-48E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12182639, end_position_on_the_genomic_accession=12185506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  4379907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379907, CurrentID=0, Status='live', Symbol='Mkp', Aliases='Dmel_CG34099, CG34099, CG7042, DMKP, Dmel\\CG34099, MKP, mkp', description='MAP kinase-specific phosphatase', other_designations='CG34099-PA|CG34099-PB|CG34099-PC|CG34099-PD|MAP kinase phosphatase|Mkp-PA|Mkp-PB|Mkp-PC|Mkp-PD', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=182522, end_position_on_the_genomic_accession=184540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  4379908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379908, CurrentID=0, Status='live', Symbol='CG34140', Aliases='Dmel_ CG7042, Dmel\\CG34140', description='CG34140 gene product from transcript CG34140-RA', other_designations='CG34140-PA|uncharacterized protein', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=182522, end_position_on_the_genomic_accession=184424, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  4379909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379909, CurrentID=0, Status='live', Symbol='BG642312', Aliases='Dmel_CG33943, 4h1(4h1), BcDNA:AT24650, CG33943, CR33943, Dmel\\CG33943, pncr010:3L', description='CG33943 gene product from transcript CG33943-RB', other_designations='BG642312-PA|BG642312-PB|CG33943-PA|CG33943-PB|putative noncoding RNA 010:3L', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6232469, end_position_on_the_genomic_accession=6233381, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  4379911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379911, CurrentID=0, Status='live', Symbol='CG17672', Aliases='Dmel_ BcDNA:RH29995, CG11271, Dmel\\CG17672, Dmel_CG11271', description='CG17672 gene product from transcript CG17672-RA', other_designations='CG17672-PA|uncharacterized protein', map_location='69F6-69F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13028962, end_position_on_the_genomic_accession=13036504, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  4379912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379912, CurrentID=0, Status='live', Symbol='CG34116', Aliases='Dmel_ CG14103, Dmel\\CG34116', description='CG34116 gene product from transcript CG34116-RA', other_designations='CG34116-PA|uncharacterized protein', map_location='76C6-76C6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19805241, end_position_on_the_genomic_accession=19806036, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740761, CurrentID=0, Status='live', Symbol='CG34445', Aliases='Dmel_ CG13513, CG30219, Dmel\\CG34445, Dmel_CG30219', description='CG34445 gene product from transcript CG34445-RA', other_designations='CG34445-PA|uncharacterized protein', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22647892, end_position_on_the_genomic_accession=22648478, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  4379919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=4379919, CurrentID=0, Status='live', Symbol='CG17698', Aliases='Dmel_ CAMKIIB, CAMKIIb, CG40297, Dmel\\CG17698', description='CG17698 gene product from transcript CG17698-RF', other_designations='CG17698-PA|CG17698-PD|CG17698-PE|CG17698-PF|uncharacterized protein', map_location='80F9-80F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23737960, end_position_on_the_genomic_accession=23749058, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  318244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318244, CurrentID=0, Status='live', Symbol='CG32846', Aliases='Dmel_ Dmel\\CG32846', description='CG32846 gene product from transcript CG32846-RB', other_designations='CG32846-PA|CG32846-PB|uncharacterized protein', map_location='63A2-63A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2950174, end_position_on_the_genomic_accession=2951618, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771951, CurrentID=0, Status='live', Symbol='Prx5', Aliases='Dmel_CG7217, CG 32920, CG 7217, CG32920, CG7217, Dmel\\CG7217, dPrx5, dPrxV, dprx5, prx5', description='Peroxiredoxin 5', other_designations='CG7217-PA|CG7217-PB|Prx5-PA|Prx5-PB|peroxiredoxin 5|peroxiredoxin V', map_location='90F4-90F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18165442, end_position_on_the_genomic_accession=18167273, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  38915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38915, CurrentID=0, Status='live', Symbol='CG7387', Aliases='Dmel_ Dmel\\CG7387, anon-WO0140519.72', description='CG7387 gene product from transcript CG7387-RA', other_designations='CG7387-PA', map_location='66B11-66B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8191522, end_position_on_the_genomic_accession=8193116, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3771952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771952, CurrentID=0, Status='live', Symbol='CG33665', Aliases='Dmel_ Dmel\\CG33665', description='CG33665 gene product from transcript CG33665-RB', other_designations='CG33665-PB|uncharacterized protein', map_location='6A2-6A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6478138, end_position_on_the_genomic_accession=6478849, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674119, CurrentID=0, Status='live', Symbol='Kmn2', Aliases='Dmel_CG42550, CG42550, Dmel\\CG42550, HDC12388', description='kinetochore Mis12-Ndc80 network component 2', other_designations='CG42550-PA|CG42550-PD|Kmn2-PA|Kmn2-PD', map_location='84E8-84E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8000548, end_position_on_the_genomic_accession=8001309, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3771957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771957, CurrentID=0, Status='live', Symbol='His2B:CG33896', Aliases='Dmel_CG33896, CG33896, Dmel\\CG33896', description='CG33896 gene product from transcript CG33896-RA', other_designations='CG33896-PA|His2B:CG33896-PA|histone H2B', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21467375, end_position_on_the_genomic_accession=21467969, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771958, CurrentID=0, Status='live', Symbol='CG33783', Aliases='Dmel_ Dmel\\CG33783', description='CG33783 gene product from transcript CG33783-RA', other_designations='CG33783-PA|uncharacterized protein', map_location='85F9-85F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10066770, end_position_on_the_genomic_accession=10067567, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3771960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771960, CurrentID=0, Status='live', Symbol='waw', Aliases='Dmel_CG1410, CG1410, CG32527, Dmel\\CG1410, anon-19Fb, tu2', description='waclaw', other_designations='CG1410-PA|waw-PA', map_location='19F5-19F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21363028, end_position_on_the_genomic_accession=21373145, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38553, CurrentID=0, Status='live', Symbol='CG7465', Aliases='Dmel_ Dmel\\CG7465', description='CG7465 gene product from transcript CG7465-RA', other_designations='CG7465-PA', map_location='64B8-64B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4480283, end_position_on_the_genomic_accession=4481487, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771965, CurrentID=0, Status='live', Symbol='CG9515', Aliases='Dmel_ BEST:GH06087, CG33085, Dmel\\CG9515, gh06087', description='CG9515 gene product from transcript CG9515-RB', other_designations='CG9515-PA|CG9515-PB', map_location='29F5-29F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8941945, end_position_on_the_genomic_accession=8943257, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3771966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771966, CurrentID=0, Status='live', Symbol='CG9960', Aliases='Dmel_ CG32951, DAMT, Dmel\\CG9960, Dromel_CG9960_CG9958_FBtr007773_snapin_mORF, damt', description='CG9960 gene product from transcript CG9960-RA', other_designations='CG9960-PA|DNA adenine methyl-transferase|DNA adenine methyltransferase', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2758427, end_position_on_the_genomic_accession=2759677, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  251270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251270, CurrentID=0, Status='live', Symbol='CG33191', Aliases='Dmel_ BEST:GH01761, BcDNA:GH01761, Dmel\\CG33191, anon-WO0140519.60', description='CG33191 gene product from transcript CG33191-RA', other_designations='CG33191-PA|uncharacterized protein', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8778379, end_position_on_the_genomic_accession=8779923, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3771970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771970, CurrentID=0, Status='live', Symbol='CG33641', Aliases='Dmel_ Dmel\\CG33641', description='CG33641 gene product from transcript CG33641-RA', other_designations='CG33641-PA|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13529042, end_position_on_the_genomic_accession=13529821, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835902, CurrentID=0, Status='live', Symbol='CG45086', Aliases='Dmel_ Dmel\\CG45086, Dromel_CG17759_galpha49b_uORF', description='CG45086 gene product from transcript CG45086-RB', other_designations='CG45086-PA|CG45086-PB|uncharacterized protein', map_location='49B8-49B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12612740, end_position_on_the_genomic_accession=12619820, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  251324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251324, CurrentID=0, Status='live', Symbol='CG30270', Aliases='Dmel_ BcDNA:AT30770, Dmel\\CG30270', description='CG30270 gene product from transcript CG30270-RC', other_designations='CG30270-PC|CG30270-PD|uncharacterized protein', map_location='59B1-59B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22784304, end_position_on_the_genomic_accession=22784932, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3771981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771981, CurrentID=0, Status='live', Symbol='His1:CG33819', Aliases='Dmel_CG33819, CG33819, Dmel\\CG33819', description='CG33819 gene product from transcript CG33819-RA', other_designations='CG33819-PA|His1:CG33819-PA|histone H1', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21461173, end_position_on_the_genomic_accession=21462195, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3771984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771984, CurrentID=0, Status='live', Symbol='CG18749', Aliases='Dmel_ CG18238, CG31188, Dmel\\CG18749', description='CG18749 gene product from transcript CG18749-RC', other_designations='CG18749-PB|CG18749-PC|uncharacterized protein', map_location='84F5-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8244055, end_position_on_the_genomic_accession=8248054, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3771985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771985, CurrentID=0, Status='live', Symbol='Pif1B', Aliases='Dmel_CG33720, BcDNA:GH12504, CG32940, CG33719/CG33720, CG33720, CG9808, Dmel\\CG33720, PIF-1B, Pif1', description='PFTAIRE-interacting factor 1B', other_designations='CG33720-PA|CG33720-PB|CG33720-PC|CG33720-PD|CG33720-PE|Pif1B-PA|Pif1B-PB|Pif1B-PC|Pif1B-PD|Pif1B-PE', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8774286, end_position_on_the_genomic_accession=8801176, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  318251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318251, CurrentID=0, Status='live', Symbol='CG32856', Aliases='Dmel_ BcDNA:GM32296, Dmel\\CG32856', description='CG32856 gene product from transcript CG32856-RB', other_designations='CG32856-PA|CG32856-PB|uncharacterized protein', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15964678, end_position_on_the_genomic_accession=15965351, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36439, CurrentID=0, Status='live', Symbol='CG13326', Aliases='Dmel_ Dmel\\CG13326', description='CG13326 gene product from transcript CG13326-RA', other_designations='CG13326-PA|uncharacterized protein', map_location='49F2-49F2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13129511, end_position_on_the_genomic_accession=13132455, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  251414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251414, CurrentID=0, Status='live', Symbol='zpg', Aliases='Dmel_CG10125, CG10125, Dm-inx4, Dmel\\CG10125, FBpp0076692, Inx4, Zpg, inx-4, inx4', description='zero population growth', other_designations='CG10125-PA|CG10125-PB|Innexin-4|innexin 4|zpg-PA|zpg-PB', map_location='65B5-65B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6604366, end_position_on_the_genomic_accession=6606032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  251430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251430, CurrentID=0, Status='live', Symbol='pen-2', Aliases='Dmel_CG33198, CG33198, Dmel\\CG33198, Pen-2, anon-AE003800.1, dPen-2', description='presenilin enhancer', other_designations='CG33198-PA|CG33198-PB|pen-2-PA|pen-2-PB|presenilin enhancer 2', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18277439, end_position_on_the_genomic_accession=18278762, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  251466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251466, CurrentID=0, Status='live', Symbol='RpL41', Aliases='Dmel_CG30425, BEST:LP07070, BcDNA:LP07070, BcDNA:RE63504, CG30425, CR30425, Dmel\\CG30425, L41, pncr005:2R', description='Ribosomal protein L41', other_designations='CG30425-PA|RpL41-PA|putative noncoding RNA 005:2R|ribosomal protein L41', map_location='60E5-60E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24903416, end_position_on_the_genomic_accession=24903935, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772007, CurrentID=0, Status='live', Symbol='Rpp20', Aliases='Dmel_CG33931, BcDNA:RE70209, CG12533, CG32530, CG32530-ORFA, CG33931, Dmel\\CG33931, d', description='CG33931 gene product from transcript CG33931-RB', other_designations='CG33931-PA|CG33931-PB|Rpp20-PA|Rpp20-PB', map_location='18E5-18E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19744431, end_position_on_the_genomic_accession=19745898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772011, CurrentID=0, Status='live', Symbol='CG33643', Aliases='Dmel_ Dmel\\CG33643', description='CG33643 gene product from transcript CG33643-RA', other_designations='CG33643-PA|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13531074, end_position_on_the_genomic_accession=13531662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772012, CurrentID=0, Status='live', Symbol='CG17768', Aliases='Dmel_ BcDNA:RE35747, CG31990, CG5356, Dmel\\CG17768', description='CG17768 gene product from transcript CG17768-RA', other_designations='CG17768-PA|CG17768-PB|uncharacterized protein', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10403197, end_position_on_the_genomic_accession=10405203, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3772013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772013, CurrentID=0, Status='live', Symbol='His2B:CG33906', Aliases='Dmel_CG33906, CG33906, Dmel\\CG33906', description='CG33906 gene product from transcript CG33906-RA', other_designations='CG33906-PA|His2B:CG33906-PA|histone H2B', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21433122, end_position_on_the_genomic_accession=21433716, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36463, CurrentID=0, Status='live', Symbol='fsd', Aliases='Dmel_CG12765, CG12765, Dmel\\CG12765, FBXO14', description='fates-shifted', other_designations='CG12765-PA|fsd-PA', map_location='49F11-49F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13230583, end_position_on_the_genomic_accession=13231663, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  251610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251610, CurrentID=0, Status='live', Symbol='CG31142', Aliases='Dmel_ BcDNA:RE39515, Dmel\\CG31142', description='CG31142 gene product from transcript CG31142-RA', other_designations='CG31142-PA|uncharacterized protein', map_location='95C13-95C13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23929243, end_position_on_the_genomic_accession=23929988, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3346145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346145, CurrentID=0, Status='live', Symbol='CG33557', Aliases='Dmel_ CG12648, Dmel\\CG33557', description='CG33557 gene product from transcript CG33557-RA', other_designations='CG33557-PA|uncharacterized protein', map_location='9A4-9A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10009210, end_position_on_the_genomic_accession=10010049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38992, CurrentID=0, Status='live', Symbol='CG13305', Aliases='Dmel_ BcDNA:RE33542, CT32600, Dmel\\CG13305', description='CG13305 gene product from transcript CG13305-RC', other_designations='CG13305-PB|CG13305-PC|uncharacterized protein', map_location='66D9-66D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8644916, end_position_on_the_genomic_accession=8646822, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740546, CurrentID=0, Status='live', Symbol='CG15140', Aliases='Dmel_ Dmel\\CG15140', description='CG15140 gene product from transcript CG15140-RB', other_designations='CG15140-PB|uncharacterized protein', map_location='36C5-36C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17286240, end_position_on_the_genomic_accession=17287663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  251645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251645, CurrentID=0, Status='live', Symbol='CG32276', Aliases='Dmel_ BcDNA:RE27904, Dmel\\CG32276', description='CG32276 gene product from transcript CG32276-RB', other_designations='CG32276-PA|CG32276-PB|uncharacterized protein', map_location='63B11-63B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3197863, end_position_on_the_genomic_accession=3199047, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318261, CurrentID=0, Status='live', Symbol='CG32945', Aliases='Dmel_ Dmel\\CG32945', description='CG32945 gene product from transcript CG32945-RA', other_designations='CG32945-PA|uncharacterized protein', map_location='82C2-82C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4577939, end_position_on_the_genomic_accession=4578644, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  251690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251690, CurrentID=0, Status='live', Symbol='CG32147', Aliases='Dmel_ BcDNA:RE07960, Dmel\\CG32147', description='CG32147 gene product from transcript CG32147-RA', other_designations='CG32147-PA|uncharacterized protein', map_location='71B5-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15143063, end_position_on_the_genomic_accession=15144059, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  251711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251711, CurrentID=0, Status='live', Symbol='CG31517', Aliases='Dmel_ BcDNA:LD34214, Dmel\\CG31517', description='CG31517 gene product from transcript CG31517-RB', other_designations='CG31517-PA|CG31517-PB|uncharacterized protein', map_location='88A3-88A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13948420, end_position_on_the_genomic_accession=13949742, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19835597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835597, CurrentID=0, Status='live', Symbol='CG45072', Aliases='Dmel_ CG18682, Dmel\\CG45072', description='CG45072 gene product from transcript CG45072-RB', other_designations='CG45072-PB|uncharacterized protein', map_location='99D7-99D7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30122023, end_position_on_the_genomic_accession=30123263, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772047, CurrentID=0, Status='live', Symbol='CG33767', Aliases='Dmel_ CG14458, Dmel\\CG33767', description='CG33767 gene product from transcript CG33767-RA', other_designations='CG33767-PA|uncharacterized protein', map_location='79F3-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22646307, end_position_on_the_genomic_accession=22646959, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835307, CurrentID=0, Status='live', Symbol='CG44476', Aliases='Dmel_ Dmel\\CG44476, lincRNA.179', description='CG44476 gene product from transcript CG44476-RA', other_designations='CG44476-PA|uncharacterized protein', map_location='36E5-36E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18112830, end_position_on_the_genomic_accession=18113179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772049, CurrentID=0, Status='live', Symbol='CG33647', Aliases='Dmel_ Dmel\\CG33647', description='CG33647 gene product from transcript CG33647-RA', other_designations='CG33647-PA|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19236360, end_position_on_the_genomic_accession=19237057, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31297, CurrentID=0, Status='live', Symbol='CG14265', Aliases='Dmel_ BcDNA:RE43371, Dmel\\CG14265, EG:96G10.8', description='CG14265 gene product from transcript CG14265-RB', other_designations='CG14265-PB|uncharacterized protein', map_location='3C10-3C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3236130, end_position_on_the_genomic_accession=3236686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36504, CurrentID=0, Status='live', Symbol='CG13332', Aliases='Dmel_ Dmel\\CG13332', description='CG13332 gene product from transcript CG13332-RA', other_designations='CG13332-PA|uncharacterized protein', map_location='50B1-50B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13525377, end_position_on_the_genomic_accession=13526448, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772058, CurrentID=0, Status='live', Symbol='His2B:CG33902', Aliases='Dmel_CG33902, CG33902, Dmel\\CG33902', description='CG33902 gene product from transcript CG33902-RA', other_designations='CG33902-PA|His2B:CG33902-PA|histone H2B', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21452211, end_position_on_the_genomic_accession=21452805, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772065, CurrentID=0, Status='live', Symbol='CG33688', Aliases='Dmel_ Dmel\\CG33688', description='CG33688 gene product from transcript CG33688-RA', other_designations='CG33688-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16199220, end_position_on_the_genomic_accession=16199926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36517, CurrentID=0, Status='live', Symbol='CG6209', Aliases='Dmel_ Dmel\\CG6209, mst101(2)', description='CG6209 gene product from transcript CG6209-RA', other_designations='CG6209-PA', map_location='50B9-50B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13661495, end_position_on_the_genomic_accession=13663598, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772073, CurrentID=0, Status='live', Symbol='CheB74a', Aliases='Dmel_CG33924, CG33924, Dmel\\CG33924', description='Chemosensory protein B 74a', other_designations='CG33924-PA|CheB74a-PA|chemosensory protein B 74a', map_location='74F3-74F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17719193, end_position_on_the_genomic_accession=17719854, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  251900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251900, CurrentID=0, Status='live', Symbol='GlcAT-I', Aliases='Dmel_CG32775, BcDNA:GH05057, CAA21824, CG32775, D DmGlcAT-I, Dmel\\CG32775, EG:EG0007.5, GH05057p, GlcAT1, RE38067, dGlcAT-I', description='CG32775 gene product from transcript CG32775-RA', other_designations='CG32775-PA|GlcAT-I-PA', map_location='4B1-4B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4131211, end_position_on_the_genomic_accession=4132833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462588, CurrentID=0, Status='live', Symbol='CG43919', Aliases='Dmel_ Dmel\\CG43919', description='CG43919 gene product from transcript CG43919-RA', other_designations='CG43919-PA|uncharacterized protein', map_location='51B1-51B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14553765, end_position_on_the_genomic_accession=14554309, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36523, CurrentID=0, Status='live', Symbol='Cpr50Ca', Aliases='Dmel_CG13338, CG13338, Dmel\\CG13338', description='Cuticular protein 50Ca', other_designations='CG13338-PA|Cpr50Ca-PA|cuticular protein 50Ca', map_location='50C3-50C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13769915, end_position_on_the_genomic_accession=13787802, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  251916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251916, CurrentID=0, Status='live', Symbol='CG32235', Aliases='Dmel_ BcDNA:AT25643, Dmel\\CG32235', description='CG32235 gene product from transcript CG32235-RA', other_designations='CG32235-PA|uncharacterized protein', map_location='64C1-64C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4859810, end_position_on_the_genomic_accession=4860470, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772077, CurrentID=0, Status='live', Symbol='His1:CG33837', Aliases='Dmel_CG33837, CG33837, Dmel\\CG33837', description='CG33837 gene product from transcript CG33837-RA', other_designations='CG33837-PA|His1:CG33837-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21491541, end_position_on_the_genomic_accession=21492563, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  251924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251924, CurrentID=0, Status='live', Symbol='CG32148', Aliases='Dmel_ BcDNA:AT21341, Dmel\\CG32148, NEST:bs21b09', description='CG32148 gene product from transcript CG32148-RA', other_designations='CG32148-PA|uncharacterized protein', map_location='71B1-71B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15023492, end_position_on_the_genomic_accession=15024009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  251937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251937, CurrentID=0, Status='live', Symbol='CG31468', Aliases='Dmel_ BcDNA:AT08725, Dmel\\CG31468, NEST:bs12h04, NEST:bs18e02, NEST:bs21c04, NEST:bs34b06, Tes37', description='CG31468 gene product from transcript CG31468-RB', other_designations='CG31468-PA|CG31468-PB|uncharacterized protein', map_location='95A8-95A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23699328, end_position_on_the_genomic_accession=23700354, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36529, CurrentID=0, Status='live', Symbol='CG6329', Aliases='Dmel_ Dmel\\CG6329', description='CG6329 gene product from transcript CG6329-RE', other_designations='CG6329-PA|CG6329-PB|CG6329-PC|CG6329-PD|CG6329-PE|CG6329-PF', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13826371, end_position_on_the_genomic_accession=13833503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36530, CurrentID=0, Status='live', Symbol='CG6337', Aliases='Dmel_ Dmel\\CG6337', description='CG6337 gene product from transcript CG6337-RB', other_designations='CG6337-PA|CG6337-PB', map_location='50C6-50C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13833575, end_position_on_the_genomic_accession=13835096, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772083, CurrentID=0, Status='live', Symbol='His2B:CG33876', Aliases='Dmel_CG33876, CG33876, Dmel\\CG33876', description='CG33876 gene product from transcript CG33876-RA', other_designations='CG33876-PA|His2B:CG33876-PA|histone H2B', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21517932, end_position_on_the_genomic_accession=21518526, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772084, CurrentID=0, Status='live', Symbol='CG33668', Aliases='Dmel_ Dmel\\CG33668', description='CG33668 gene product from transcript CG33668-RB', other_designations='CG33668-PB|uncharacterized protein', map_location='6A1-6A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6418760, end_position_on_the_genomic_accession=6419434, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318271, CurrentID=0, Status='live', Symbol='CG32971', Aliases='Dmel_ Dmel\\CG32971', description='CG32971 gene product from transcript CG32971-RB', other_designations='CG32971-PA|CG32971-PB|CG32971-PC|uncharacterized protein', map_location='35A3-35A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14274821, end_position_on_the_genomic_accession=14278002, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  251984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=251984, CurrentID=0, Status='live', Symbol='Jheh1', Aliases='Dmel_CG15101, 142196_at, BEST:GH06241, CG15101, Dmel\\CG15101, JH, JHEH, Jheh', description='Juvenile hormone epoxide hydrolase 1', other_designations='CG15101-PA|JH-epoxide hydrolase|Jheh1-PA|juvenile hormone epoxide hydrolase|juvenile hormone epoxide hydrolase 1|microsomal apoxide hydrolase', map_location='55F7-55F8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18853434, end_position_on_the_genomic_accession=18855087, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  10178814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178814, CurrentID=0, Status='live', Symbol='CG42806', Aliases='Dmel_ Dmel\\CG42806', description='CG42806 gene product from transcript CG42806-RB', other_designations='CG42806-PB|uncharacterized protein', map_location='50C19-50C20', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13964653, end_position_on_the_genomic_accession=13965455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772094, CurrentID=0, Status='live', Symbol='His2B:CG33900', Aliases='Dmel_CG33900, CG33900, Dmel\\CG33900', description='CG33900 gene product from transcript CG33900-RA', other_designations='CG33900-PA|His2B:CG33900-PA|histone H2B', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21457272, end_position_on_the_genomic_accession=21457866, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318273, CurrentID=0, Status='live', Symbol='CG32987', Aliases='Dmel_ CG15866, Dmel\\CG32987', description='CG32987 gene product from transcript CG32987-RA', other_designations='CG32987-PA|uncharacterized protein', map_location='29F3-29F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8863811, end_position_on_the_genomic_accession=8864847, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772096, CurrentID=0, Status='live', Symbol='His4:CG33897', Aliases='Dmel_CG33897, CG33897, Dmel\\CG33897', description='CG33897 gene product from transcript CG33897-RA', other_designations='CG33897-PA|His4:CG33897-PA|histone H4', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21463913, end_position_on_the_genomic_accession=21464325, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772099, CurrentID=0, Status='live', Symbol='His2B:CG33908', Aliases='Dmel_CG33908, CG33908, Dmel\\CG33908', description='CG33908 gene product from transcript CG33908-RA', other_designations='CG33908-PA|His2B:CG33908-PA|histone H2B', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21428075, end_position_on_the_genomic_accession=21428669, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36548, CurrentID=0, Status='live', Symbol='St4', Aliases='Dmel_CG6704, CG6704, Dmel\\CG6704, dmST4, st4', description='Sulfotransferase 4', other_designations='CG6704-PA|CG6704-PB|St4-PA|St4-PB|sulfotransferase 4', map_location='50C20-50C20', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13969910, end_position_on_the_genomic_accession=13971263, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318274, CurrentID=0, Status='live', Symbol='CG32988', Aliases='Dmel_ CG18663, Dmel\\CG32988', description='CG32988 gene product from transcript CG32988-RA', other_designations='CG32988-PA|uncharacterized protein', map_location='29F3-29F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8864864, end_position_on_the_genomic_accession=8865692, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36549, CurrentID=0, Status='live', Symbol='CG18568', Aliases='Dmel_ Dmel\\CG18568, anon-WO0140519.113', description='CG18568 gene product from transcript CG18568-RB', other_designations='CG18568-PA|CG18568-PB|uncharacterized protein', map_location='50C21-50C22', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13971325, end_position_on_the_genomic_accession=13972834, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772104, CurrentID=0, Status='live', Symbol='His2B:CG33888', Aliases='Dmel_CG33888, CG33888, Dmel\\CG33888', description='CG33888 gene product from transcript CG33888-RA', other_designations='CG33888-PA|His2B:CG33888-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21487655, end_position_on_the_genomic_accession=21488249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740549, CurrentID=0, Status='live', Symbol='CG34446', Aliases='Dmel_ CG13513, CG30219, Dmel\\CG34446, Dmel_CG30219', description='CG34446 gene product from transcript CG34446-RA', other_designations='CG34446-PA|uncharacterized protein', map_location='58F3-58F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22647458, end_position_on_the_genomic_accession=22647807, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36556, CurrentID=0, Status='live', Symbol='CG18371', Aliases='Dmel_ Dmel\\CG18371', description='CG18371 gene product from transcript CG18371-RA', other_designations='CG18371-PA|uncharacterized protein', map_location='50D2-50D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14028620, end_position_on_the_genomic_accession=14029096, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772111, CurrentID=0, Status='live', Symbol='CG33792', Aliases='Dmel_ Dmel\\CG33792', description='CG33792 gene product from transcript CG33792-RB', other_designations='CG33792-PB|uncharacterized protein', map_location='49B10-49B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12632674, end_position_on_the_genomic_accession=12633454, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772113, CurrentID=0, Status='live', Symbol='His4:CG33873', Aliases='Dmel_CG33873, CG33873, Dmel\\CG33873', description='CG33873 gene product from transcript CG33873-RA', other_designations='CG33873-PA|His4:CG33873-PA|histone H4', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21524399, end_position_on_the_genomic_accession=21524926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178820 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178820, CurrentID=0, Status='live', Symbol='CG42704', Aliases='Dmel_ Dmel\\CG42704', description='CG42704 gene product from transcript CG42704-RA', other_designations='CG42704-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7412668, end_position_on_the_genomic_accession=7413296, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740103, CurrentID=0, Status='live', Symbol='CG34216', Aliases='Dmel_ Dmel\\CG34216', description='CG34216 gene product from transcript CG34216-RB', other_designations='CG34216-PB|uncharacterized protein', map_location='43E11-43E12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7703506, end_position_on_the_genomic_accession=7705095, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3346141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346141, CurrentID=0, Status='live', Symbol='CG33521', Aliases='Dmel_ 2D9, BcDNA:RE73965, CG17467, CG32015, Dmel\\CG33521, anon-EST:fe2A9, anon-EST:fe2D9, anon-EST:fe2E7, anon2D9', description='CG33521 gene product from transcript CG33521-RC', other_designations='BcDNA:RE73965|CG33521-PA|CG33521-PC|CG33521-PD|CG33521-PG|anon-fast-evolving-2A9|anon-fast-evolving-2D9|anon-fast-evolving-2E7|uncharacterized protein', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1185768, end_position_on_the_genomic_accession=1192985, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  3346142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346142, CurrentID=0, Status='live', Symbol='CG13376', Aliases='Dmel_ Dmel\\CG13376', description='CG13376 gene product from transcript CG13376-RC', other_designations='CG13376-PB|CG13376-PC|CG13376-PD|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=265724, end_position_on_the_genomic_accession=267028, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772129, CurrentID=0, Status='live', Symbol='His4:CG33881', Aliases='Dmel_CG33881, CG33881, Dmel\\CG33881', description='CG33881 gene product from transcript CG33881-RA', other_designations='CG33881-PA|His4:CG33881-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21504377, end_position_on_the_genomic_accession=21504789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31313 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31313, CurrentID=0, Status='live', Symbol='CG16782', Aliases='Dmel_ Dmel\\CG16782', description='CG16782 gene product from transcript CG16782-RA', other_designations='CG16782-PA|uncharacterized protein', map_location='3D3-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3459016, end_position_on_the_genomic_accession=3460240, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36581, CurrentID=0, Status='live', Symbol='CG8503', Aliases='Dmel_ Dmel\\CG8503', description='CG8503 gene product from transcript CG8503-RA', other_designations='CG8503-PA', map_location='50E5-50E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14243778, end_position_on_the_genomic_accession=14246066, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3346150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346150, CurrentID=0, Status='live', Symbol='CG32547', Aliases='Dmel_ CG12610, CG15049, CG15050, Dmel\\CG32547', description='CG32547 gene product from transcript CG32547-RD', other_designations='CG32547-PC|CG32547-PD|uncharacterized protein', map_location='17B1-17B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18443027, end_position_on_the_genomic_accession=18479487, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31314, CurrentID=0, Status='live', Symbol='CG10801', Aliases='Dmel_ Dmel\\CG10801', description='CG10801 gene product from transcript CG10801-RA', other_designations='CG10801-PA|CG10801-PB|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3460592, end_position_on_the_genomic_accession=3462500, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317813, CurrentID=0, Status='live', Symbol='CG18789', Aliases='Dmel_ Dmel\\CG18789', description='CG18789 gene product from transcript CG18789-RA', other_designations='CG18789-PA|uncharacterized protein', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11993989, end_position_on_the_genomic_accession=11995432, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317815, CurrentID=0, Status='live', Symbol='CG32000', Aliases='Dmel_ BcDNA:LD05303, BcDNA:LD19039, BcDNA:SD17880, Dmel\\CG32000', description='CG32000 gene product from transcript CG32000-RH', other_designations='CG32000-PA|CG32000-PB|CG32000-PC|CG32000-PD|CG32000-PE|CG32000-PF|CG32000-PG|CG32000-PH|CG32000-PI|uncharacterized protein', map_location='102A3-102A4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=132220, end_position_on_the_genomic_accession=143123, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  317817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317817, CurrentID=0, Status='live', Symbol='CG32006', Aliases='Dmel_ BcDNA:RE51913, DmCg32006, Dmel\\CG32006', description='CG32006 gene product from transcript CG32006-RA', other_designations='CG32006-PA|uncharacterized protein', map_location='102A4-102A4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=150764, end_position_on_the_genomic_accession=153538, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3772138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772138, CurrentID=0, Status='live', Symbol='IM18', Aliases='Dmel_CG33706, BcDNA:AT19986, CG10332, CG33706, DIM 18, Dmel\\CG33706, Dmel_CG10332, anon-WO0153538.56, anon-WO0153538.57, anon-WO0153538.58', description='Immune induced molecule 18', other_designations='CG33706-PA|CG33706-PB|IM18-PA|IM18-PB|immune induced molecule 18', map_location='59F4-59F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23600914, end_position_on_the_genomic_accession=23601773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317823, CurrentID=0, Status='live', Symbol='CG32017', Aliases='Dmel_ Dmel\\CG32017', description='CG32017 gene product from transcript CG32017-RD', other_designations='CG32017-PB|CG32017-PC|CG32017-PD|uncharacterized protein', map_location='102F8-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1137830, end_position_on_the_genomic_accession=1162158, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  317824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317824, CurrentID=0, Status='live', Symbol='Zyx', Aliases='Dmel_CG32018, BcDNA:LD06023, CG32018, DLim-1, DmLIM-1, Dmel\\CG32018, Lim102EF, ZYX102102, Zyx102EF, lim, zyx, zyx102, zyxin, Zyx', description='Zyxin', other_designations='CG32018-PA|CG32018-PB|CG32018-PG|CG32018-PH|CG32018-PL|Zyx-PA|Zyx-PB|Zyx-PG|Zyx-PH|Zyx-PL|Zyx102EF|zyxin', map_location='102F7-102F8', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=1057365, end_position_on_the_genomic_accession=1065001, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  317826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317826, CurrentID=0, Status='live', Symbol='CG32022', Aliases='Dmel_ BcDNA:AT11343, CG5950, Chd5, Dmel\\CG32022, Wrb', description='CG32022 gene product from transcript CG32022-RB', other_designations='CG32022-PA|CG32022-PB|uncharacterized protein', map_location='66D11-66D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8720280, end_position_on_the_genomic_accession=8721272, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772139, CurrentID=0, Status='live', Symbol='His2A:CG33859', Aliases='Dmel_CG33859, CG33859, Dmel\\CG33859', description='CG33859 gene product from transcript CG33859-RA', other_designations='CG33859-PA|His2A:CG33859-PA|histone H2A', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21528341, end_position_on_the_genomic_accession=21528833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317828, CurrentID=0, Status='live', Symbol='CG32024', Aliases='Dmel_ Dmel\\CG32024', description='CG32024 gene product from transcript CG32024-RA', other_designations='CG32024-PA|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8807601, end_position_on_the_genomic_accession=8808407, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317829, CurrentID=0, Status='live', Symbol='CG32026', Aliases='Dmel_ BcDNA:GH12815, Dmel\\CG32026, anon-WO0140519.202', description='CG32026 gene product from transcript CG32026-RA', other_designations='CG32026-PA|uncharacterized protein', map_location='66E1-66E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8836462, end_position_on_the_genomic_accession=8838843, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317833, CurrentID=0, Status='live', Symbol='ghi', Aliases='Dmel_CG32038, CG32038, Dmel\\CG32038, fro', description='ghiberti', other_designations='CG32038-PA|frodo|ghi-PA', map_location='67B7-67B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9451876, end_position_on_the_genomic_accession=9452790, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317834, CurrentID=0, Status='live', Symbol='CG32039', Aliases='Dmel_ BEST:GH02734, BcDNA:RE03722, Dmel\\CG32039', description='CG32039 gene product from transcript CG32039-RB', other_designations='CG32039-PA|CG32039-PB|uncharacterized protein', map_location='67B2-67B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9358999, end_position_on_the_genomic_accession=9359996, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317835, CurrentID=0, Status='live', Symbol='CG32040', Aliases='Dmel_ Dmel\\CG32040', description='CG32040 gene product from transcript CG32040-RA', other_designations='CG32040-PA|uncharacterized protein', map_location='67B1-67B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9354033, end_position_on_the_genomic_accession=9355457, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317838, CurrentID=0, Status='live', Symbol='Ir67c', Aliases='Dmel_CG32058, CG32058, DmelIR67c, Dmel\\CG32058, IR67c', description='Ionotropic receptor 67c', other_designations='CG32058-PA|Ir67c-PA|ionotropic receptor 67c', map_location='67E3-67E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10432384, end_position_on_the_genomic_accession=10434105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317841, CurrentID=0, Status='live', Symbol='Elo68alpha', Aliases='Dmel_CG32072, CG32072, CG6261, Dmel\\CG32072, elo68alpha', description='Elongase 68alpha', other_designations='CG32072-PA|Elo68alpha-PA|elongase 68alpha', map_location='68A6-68A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11101875, end_position_on_the_genomic_accession=11103030, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317842, CurrentID=0, Status='live', Symbol='CG32074', Aliases='Dmel_ Dmel\\CG32074', description='CG32074 gene product from transcript CG32074-RA', other_designations='CG32074-PA|uncharacterized protein', map_location='68A8-68A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11131883, end_position_on_the_genomic_accession=11132322, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317846, CurrentID=0, Status='live', Symbol='Vha16-3', Aliases='Dmel_CG32090, CG32090, CG7547, Dmel\\CG32090, vha16-3', description='Vacuolar H[+] ATPase 16kD subunit 3', other_designations='CG32090-PA|CG32090-PB|Vha16-3-PA|Vha16-3-PB|vacuolar H[+] ATPase 16kD subunit 3|vacuolar H[+] ATPase subunit 16-3', map_location='68C9-68C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11473291, end_position_on_the_genomic_accession=11473972, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317849, CurrentID=0, Status='live', Symbol='CG32095', Aliases='Dmel_ BEST:GH03217, BcDNA:GH03217, DPLG3, Dmel\\CG32095', description='CG32095 gene product from transcript CG32095-RA', other_designations='CG32095-PA|PIF-like gene 3|uncharacterized protein', map_location='68E1-68E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11824207, end_position_on_the_genomic_accession=11825732, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317852, CurrentID=0, Status='live', Symbol='CG32099', Aliases='Dmel_ BcDNA:GH12334, Dmel\\CG32099, EAP, eap', description='CG32099 gene product from transcript CG32099-RA', other_designations='CG32099-PA|eap|ebony activating protein|uncharacterized protein', map_location='68F6-68F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12149273, end_position_on_the_genomic_accession=12150329, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317854, CurrentID=0, Status='live', Symbol='CG32102', Aliases='Dmel_ BcDNA:RE42781, Dmel\\CG32102', description='CG32102 gene product from transcript CG32102-RA', other_designations='CG32102-PA|uncharacterized protein', map_location='69C2-69C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12469072, end_position_on_the_genomic_accession=12469784, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317855, CurrentID=0, Status='live', Symbol='CG32106', Aliases='Dmel_ BcDNA:GH24302, Dmel\\CG32106, mlncRNA69E2', description='CG32106 gene product from transcript CG32106-RD', other_designations='CG32106-PD|CG32106-PE|uncharacterized protein', map_location='69E2-69E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12773951, end_position_on_the_genomic_accession=12778716, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  317856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317856, CurrentID=0, Status='live', Symbol='CG32107', Aliases='Dmel_ Dmel\\CG32107', description='CG32107 gene product from transcript CG32107-RD', other_designations='CG32107-PD|CG32107-PE|uncharacterized protein', map_location='69E6-69E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12857282, end_position_on_the_genomic_accession=12861305, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3346160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346160, CurrentID=0, Status='live', Symbol='dpr4', Aliases='Dmel_CG33512, CG12593, CG14707, CG33512, CT34497, Dmel\\CG33512, Dpr-4, Dpr4', description='defective proboscis extension response 4', other_designations='CG33512-PB|dpr4-PB', map_location='86E8-86E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11533483, end_position_on_the_genomic_accession=11542949, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  317858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317858, CurrentID=0, Status='live', Symbol='CG32110', Aliases='Dmel_ Dmel\\CG32110, anon-AE003540.1', description='CG32110 gene product from transcript CG32110-RA', other_designations='CG32110-PA|uncharacterized protein', map_location='69E4-69E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12808721, end_position_on_the_genomic_accession=12810096, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317860, CurrentID=0, Status='live', Symbol='CG32112', Aliases='Dmel_ CG10982, Dmel\\CG32112', description='CG32112 gene product from transcript CG32112-RE', other_designations='CG32112-PB|CG32112-PC|CG32112-PD|CG32112-PE|uncharacterized protein', map_location='69E1-69E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12741718, end_position_on_the_genomic_accession=12745154, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  317862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317862, CurrentID=0, Status='live', Symbol='CG32115', Aliases='Dmel_ BcDNA:RE40923, Dmel\\CG32115', description='CG32115 gene product from transcript CG32115-RA', other_designations='CG32115-PA|uncharacterized protein', map_location='69F1-69F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12912027, end_position_on_the_genomic_accession=12913646, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36593, CurrentID=0, Status='live', Symbol='CG8613', Aliases='Dmel_ Dmel\\CG8613', description='CG8613 gene product from transcript CG8613-RA', other_designations='CG8613-PA', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14352231, end_position_on_the_genomic_accession=14355674, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  317864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317864, CurrentID=0, Status='live', Symbol='CG32117', Aliases='Dmel_ Dmel\\CG32117', description='CG32117 gene product from transcript CG32117-RA', other_designations='CG32117-PA|uncharacterized protein', map_location='69F1-69F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12902925, end_position_on_the_genomic_accession=12903711, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317866, CurrentID=0, Status='live', Symbol='CG32119', Aliases='Dmel_ BcDNA:AT05691, Dmel\\CG32119, NEST:bs34d09', description='CG32119 gene product from transcript CG32119-RA', other_designations='CG32119-PA|uncharacterized protein', map_location='70A3-70A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13246864, end_position_on_the_genomic_accession=13248702, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317867, CurrentID=0, Status='live', Symbol='CG32121', Aliases='Dmel_ Dmel\\CG32121', description='CG32121 gene product from transcript CG32121-RB', other_designations='CG32121-PB|uncharacterized protein', map_location='70B1-70B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13416576, end_position_on_the_genomic_accession=13422336, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  36594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36594, CurrentID=0, Status='live', Symbol='CG8617', Aliases='Dmel_ Dmel\\CG8617', description='CG8617 gene product from transcript CG8617-RA', other_designations='CG8617-PA', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14355769, end_position_on_the_genomic_accession=14357281, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317878, CurrentID=0, Status='live', Symbol='Nxf3', Aliases='Dmel_CG32135, CG32135, Dmel\\CG32135, NXF3, nxf3', description='Nuclear export factor 3', other_designations='CG32135-PB|CG32135-PC|Nxf3-PB|Nxf3-PC|nuclear export factor 3', map_location='70C8-70C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13965203, end_position_on_the_genomic_accession=13973067, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  317880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317880, CurrentID=0, Status='live', Symbol='CG32141', Aliases='Dmel_ Dmel\\CG32141', description='CG32141 gene product from transcript CG32141-RA', other_designations='CG32141-PA|uncharacterized protein', map_location='70D3-70D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14203323, end_position_on_the_genomic_accession=14204294, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317885, CurrentID=0, Status='live', Symbol='CG32152', Aliases='Dmel_ Dmel\\CG32152', description='CG32152 gene product from transcript CG32152-RC', other_designations='CG32152-PB|CG32152-PC|uncharacterized protein', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16139501, end_position_on_the_genomic_accession=16144036, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317887, CurrentID=0, Status='live', Symbol='l(3)72Dp', Aliases='Dmel_CG32155, CG32155, CG5082, Dmel\\CG32155', description='lethal (3) 72Dp', other_designations='CG32155-PA|l(3)72Dp-PA', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16234140, end_position_on_the_genomic_accession=16236999, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  317890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317890, CurrentID=0, Status='live', Symbol='CG32161', Aliases='Dmel_ Dmel\\CG32161', description='CG32161 gene product from transcript CG32161-RA', other_designations='CG32161-PA|uncharacterized protein', map_location='73D1-73D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16846227, end_position_on_the_genomic_accession=16847000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317891, CurrentID=0, Status='live', Symbol='CG32163', Aliases='Dmel_ Dmel\\CG32163', description='CG32163 gene product from transcript CG32163-RB', other_designations='CG32163-PA|CG32163-PB|uncharacterized protein', map_location='73A10-73A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16587157, end_position_on_the_genomic_accession=16588143, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3346166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346166, CurrentID=0, Status='live', Symbol='CG14300', Aliases='Dmel_ Dmel\\CG14300', description='CG14300 gene product from transcript CG14300-RA', other_designations='CG14300-PA|uncharacterized protein', map_location='91C1-91C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18693293, end_position_on_the_genomic_accession=18693711, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317894, CurrentID=0, Status='live', Symbol='CG32170', Aliases='Dmel_ BcDNA:RE28470, CT12913, Dmel\\CG32170', description='CG32170 gene product from transcript CG32170-RA', other_designations='CG32170-PA|uncharacterized protein', map_location='73D4-73D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16877309, end_position_on_the_genomic_accession=16879372, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317897, CurrentID=0, Status='live', Symbol='CG32181', Aliases='Dmel_ BcDNA:AT19257, Dmel\\CG32181', description='CG32181 gene product from transcript CG32181-RB', other_designations='CG32181-PB|uncharacterized protein', map_location='75A1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17759697, end_position_on_the_genomic_accession=17762904, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  317898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317898, CurrentID=0, Status='live', Symbol='CG32182', Aliases='Dmel_ Dmel\\CG32182', description='CG32182 gene product from transcript CG32182-RA', other_designations='CG32182-PA|uncharacterized protein', map_location='75A1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17758112, end_position_on_the_genomic_accession=17759553, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317900 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317900, CurrentID=0, Status='live', Symbol='edin', Aliases='Dmel_CG32185, CG32185, Dmel\\CG32185, Edin', description='elevated during infection', other_designations='CG32185-PA|edin-PA', map_location='74D1-74D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17494880, end_position_on_the_genomic_accession=17495300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317901, CurrentID=0, Status='live', Symbol='CG32188', Aliases='Dmel_ Dmel\\CG32188', description='CG32188 gene product from transcript CG32188-RA', other_designations='CG32188-PA|uncharacterized protein', map_location='75A4-75A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17833277, end_position_on_the_genomic_accession=17833625, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31317, CurrentID=0, Status='live', Symbol='CG10804', Aliases='Dmel_ Dmel\\CG10804', description='CG10804 gene product from transcript CG10804-RD', other_designations='CG10804-PA|CG10804-PB|CG10804-PC|CG10804-PD|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3476463, end_position_on_the_genomic_accession=3482322, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  317903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317903, CurrentID=0, Status='live', Symbol='CG32191', Aliases='Dmel_ CG5584, Dmel\\CG32191', description='CG32191 gene product from transcript CG32191-RC', other_designations='CG32191-PB|CG32191-PC|uncharacterized protein', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17806451, end_position_on_the_genomic_accession=17809146, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  317904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317904, CurrentID=0, Status='live', Symbol='CG32192', Aliases='Dmel_ Dmel\\CG32192, NEST:bs33g12, NEST:bs36a12, bs33g12.y1', description='CG32192 gene product from transcript CG32192-RB', other_designations='CG32192-PB|CG32192-PC|uncharacterized protein', map_location='75B4-75B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18040317, end_position_on_the_genomic_accession=18041738, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  317907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317907, CurrentID=0, Status='live', Symbol='CG32195', Aliases='Dmel_ Dmel\\CG32195', description='CG32195 gene product from transcript CG32195-RA', other_designations='CG32195-PA|CG32195-PB|CG32195-PC|uncharacterized protein', map_location='75B10-75B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18112959, end_position_on_the_genomic_accession=18119150, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  317908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317908, CurrentID=0, Status='live', Symbol='CG32196', Aliases='Dmel_ Dmel\\CG32196', description='CG32196 gene product from transcript CG32196-RC', other_designations='CG32196-PB|CG32196-PC|uncharacterized protein', map_location='75C7-75C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18459464, end_position_on_the_genomic_accession=18461024, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317909, CurrentID=0, Status='live', Symbol='CG32198', Aliases='Dmel_ Dmel\\CG32198', description='CG32198 gene product from transcript CG32198-RB', other_designations='CG32198-PB|uncharacterized protein', map_location='75D1-75D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18526976, end_position_on_the_genomic_accession=18527445, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317911, CurrentID=0, Status='live', Symbol='CG32201', Aliases='Dmel_ CG13382, CG18230, Dmel\\CG32201', description='CG32201 gene product from transcript CG32201-RB', other_designations='CG32201-PB|uncharacterized protein', map_location='75D4-75D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18609620, end_position_on_the_genomic_accession=18611862, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  317912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317912, CurrentID=0, Status='live', Symbol='Spn75F', Aliases='Dmel_CG32203, 6h1, BG642378, BG642378(6h1), CG32203, Dmel\\CG32203', description='Serpin 75F', other_designations='BG642378|CG32203-PB|Spn75F-PB|serpin 75F', map_location='75F9-75F9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19128450, end_position_on_the_genomic_accession=19129665, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317913, CurrentID=0, Status='live', Symbol='CG32204', Aliases='Dmel_ Dmel\\CG32204', description='CG32204 gene product from transcript CG32204-RB', other_designations='CG32204-PB|uncharacterized protein', map_location='75E6-75E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18922419, end_position_on_the_genomic_accession=18946038, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  317916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317916, CurrentID=0, Status='live', Symbol='825-Oak', Aliases='Dmel_CG32208, CG32208, Dmel\\CG32208', description='CG32208 gene product from transcript CG32208-RB', other_designations='825-Oak-PB|CG32208-PB', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19471656, end_position_on_the_genomic_accession=19472228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36602, CurrentID=0, Status='live', Symbol='CG10104', Aliases='Dmel_ Dmel\\CG10104', description='CG10104 gene product from transcript CG10104-RA', other_designations='CG10104-PA|uncharacterized protein', map_location='51A1-51A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14390696, end_position_on_the_genomic_accession=14392209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317918, CurrentID=0, Status='live', Symbol='CG32213', Aliases='Dmel_ Dmel\\CG32213', description='CG32213 gene product from transcript CG32213-RB', other_designations='CG32213-PB|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19475839, end_position_on_the_genomic_accession=19476423, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740551, CurrentID=0, Status='live', Symbol='CG34260', Aliases='Dmel_ Dmel\\CG34260', description='CG34260 gene product from transcript CG34260-RC', other_designations='CG34260-PC|uncharacterized protein', map_location='77E3-77E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20714100, end_position_on_the_genomic_accession=20715158, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317922, CurrentID=0, Status='live', Symbol='CG32219', Aliases='Dmel_ BcDNA:AT25132, Dmel\\CG32219', description='CG32219 gene product from transcript CG32219-RA', other_designations='CG32219-PA|uncharacterized protein', map_location='76D6-76D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19950023, end_position_on_the_genomic_accession=19951135, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317923, CurrentID=0, Status='live', Symbol='Csas', Aliases='Dmel_CG32220, CG32220, CSAS, DmCSAS, Dmel\\CG32220', description='CMP-sialic acid synthase', other_designations='CG32220-PA|Csas-PA|cytidine monophosphate-sialic acid synthetase', map_location='76D8-76D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19988027, end_position_on_the_genomic_accession=19988947, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317924, CurrentID=0, Status='live', Symbol='CG32221', Aliases='Dmel_ Dmel\\CG32221, FBXL18, anon-WO0118547.270', description='CG32221 gene product from transcript CG32221-RB', other_designations='CG32221-PB|uncharacterized protein', map_location='76D1-76D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19834670, end_position_on_the_genomic_accession=19836413, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317926, CurrentID=0, Status='live', Symbol='CG32225', Aliases='Dmel_ Dmel\\CG32225', description='CG32225 gene product from transcript CG32225-RA', other_designations='CG32225-PA|uncharacterized protein', map_location='77B2-77B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20304310, end_position_on_the_genomic_accession=20306283, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  317928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317928, CurrentID=0, Status='live', Symbol='ND-MLRQ', Aliases='Dmel_CG32230, BEST:GH04411, BcDNA:RE56733, CG32230, Dmel\\CG32230, NUML', description='NADH dehydrogenase (ubiquinone) MLRQ subunit', other_designations='CG32230-PA|CG32230-PB|CG32230-PC|NADH ubiquinone oxidoreductase MLRQ|NADH-ubiquinone oxidoreductase MLRQ subunit|NADH:ubiquinone oxidoreductase MLRQ subunit|ND-MLRQ-PA|ND-MLRQ-PB|ND-MLRQ-PC', map_location='80F2-80F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23339326, end_position_on_the_genomic_accession=23345936, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  317930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317930, CurrentID=0, Status='live', Symbol='CG32232', Aliases='Dmel_ Dmel\\CG32232', description='CG32232 gene product from transcript CG32232-RA', other_designations='CG32232-PA|uncharacterized protein', map_location='64C4-64C4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4993950, end_position_on_the_genomic_accession=4996877, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317933, CurrentID=0, Status='live', Symbol='CG32240', Aliases='Dmel_ BEST:GH04494, BcDNA:GH04494, Dmel\\CG32240', description='CG32240 gene product from transcript CG32240-RA', other_designations='CG32240-PA|uncharacterized protein', map_location='64B3-64B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4380864, end_position_on_the_genomic_accession=4381629, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317934, CurrentID=0, Status='live', Symbol='CG32241', Aliases='Dmel_ Dmel\\CG32241', description='CG32241 gene product from transcript CG32241-RB', other_designations='CG32241-PB|uncharacterized protein', map_location='64B6-64B6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4465105, end_position_on_the_genomic_accession=4466630, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317935, CurrentID=0, Status='live', Symbol='CG32243', Aliases='Dmel_ BcDNA:RE12890, Dmel\\CG32243', description='CG32243 gene product from transcript CG32243-RA', other_designations='CG32243-PA|uncharacterized protein', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4539302, end_position_on_the_genomic_accession=4541244, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  317936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317936, CurrentID=0, Status='live', Symbol='CG32246', Aliases='Dmel_ Dmel\\CG32246', description='CG32246 gene product from transcript CG32246-RB', other_designations='CG32246-PB|uncharacterized protein', map_location='64B12-64B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4621219, end_position_on_the_genomic_accession=4622985, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317938, CurrentID=0, Status='live', Symbol='CG32248', Aliases='Dmel_ Dmel\\CG32248', description='CG32248 gene product from transcript CG32248-RB', other_designations='CG32248-PB|uncharacterized protein', map_location='64B7-64B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4471749, end_position_on_the_genomic_accession=4472438, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317939, CurrentID=0, Status='live', Symbol='CG32249', Aliases='Dmel_ Dmel\\CG32249', description='CG32249 gene product from transcript CG32249-RB', other_designations='CG32249-PB|uncharacterized protein', map_location='64B7-64B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4470001, end_position_on_the_genomic_accession=4471012, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317942, CurrentID=0, Status='live', Symbol='CG32259', Aliases='Dmel_ Dmel\\CG32259, NEST:bs27e11', description='CG32259 gene product from transcript CG32259-RD', other_designations='CG32259-PC|CG32259-PD|uncharacterized protein', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4069369, end_position_on_the_genomic_accession=4070281, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317943, CurrentID=0, Status='live', Symbol='CG32260', Aliases='Dmel_ BcDNA:GH24648, Dmel\\CG32260', description='CG32260 gene product from transcript CG32260-RC', other_designations='CG32260-PA|CG32260-PC|uncharacterized protein', map_location='64A8-64A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4142364, end_position_on_the_genomic_accession=4145966, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  317944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317944, CurrentID=0, Status='live', Symbol='CG32263', Aliases='Dmel_ CG10854, Dmel\\CG32263', description='CG32263 gene product from transcript CG32263-RA', other_designations='CG32263-PA|uncharacterized protein', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3810312, end_position_on_the_genomic_accession=3811114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317945, CurrentID=0, Status='live', Symbol='CG32266', Aliases='Dmel_ BcDNA:GM24338, Dmel\\CG32266', description='CG32266 gene product from transcript CG32266-RA', other_designations='CG32266-PA|uncharacterized protein', map_location='63F1-63F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3790804, end_position_on_the_genomic_accession=3791560, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  317946 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317946, CurrentID=0, Status='live', Symbol='CG32267', Aliases='Dmel_ Dmel\\CG32267', description='CG32267 gene product from transcript CG32267-RA', other_designations='CG32267-PA|uncharacterized protein', map_location='63E2-63E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3471555, end_position_on_the_genomic_accession=3472113, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  317952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317952, CurrentID=0, Status='live', Symbol='CG32280', Aliases='Dmel_ BcDNA:GM08588, Dmel\\CG32280', description='CG32280 gene product from transcript CG32280-RD', other_designations='CG32280-PA|CG32280-PB|CG32280-PC|CG32280-PD|uncharacterized protein', map_location='63B8-63B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3165181, end_position_on_the_genomic_accession=3169080, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  317953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317953, CurrentID=0, Status='live', Symbol='CG32281', Aliases='Dmel_ CG12032, Dmel\\CG32281', description='CG32281 gene product from transcript CG32281-RA', other_designations='CG32281-PA|uncharacterized protein', map_location='63B8-63B8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3169375, end_position_on_the_genomic_accession=3170913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317954, CurrentID=0, Status='live', Symbol='Drsl4', Aliases='Dmel_CG32282, CG32282, Dmel\\CG32282, Dro-4, Dro4, Drs-lF, dro-4, dro4, dros4', description='Drosomycin-like 4', other_designations='CG32282-PA|Drosomycin-4|Drosomycin-like F|Drsl4-PA|drosmycin-4|drosomycin|drosomycin 4|drosomycin-F', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3315620, end_position_on_the_genomic_accession=3315942, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317955, CurrentID=0, Status='live', Symbol='Drsl3', Aliases='Dmel_CG32283, CG10813, CG32283, CT30310, Dmel\\CG32283, Dro-3, Dro3, Drs-lE, dro3', description='Drosomycin-like 3', other_designations='CG32283-PA|Drosomycin-like E|Drosomycin3|Drsl3-PA|drosomycin|drosomycin 3|drosomycin-3|drosomycin-E', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3314996, end_position_on_the_genomic_accession=3315355, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317956, CurrentID=0, Status='live', Symbol='CG32284', Aliases='Dmel_ BcDNA:LP03531, Dmel\\CG32284', description='CG32284 gene product from transcript CG32284-RA', other_designations='CG32284-PA|uncharacterized protein', map_location='63B10-63B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3186664, end_position_on_the_genomic_accession=3187245, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3346177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346177, CurrentID=0, Status='live', Symbol='CG33506', Aliases='Dmel_ CG10220, Dmel\\CG33506', description='CG33506 gene product from transcript CG33506-RA', other_designations='CG33506-PA|uncharacterized protein', map_location='51C5-51C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14857447, end_position_on_the_genomic_accession=14858640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317962, CurrentID=0, Status='live', Symbol='CG32295', Aliases='Dmel_ BEST:GH20547, CG16757, Dmel\\CG32295, NEST:bs09a12', description='CG32295 gene product from transcript CG32295-RA', other_designations='CG32295-PA|uncharacterized protein', map_location='62E5-62E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2525799, end_position_on_the_genomic_accession=2526973, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317963, CurrentID=0, Status='live', Symbol='CG32298', Aliases='Dmel_ Dmel\\CG32298, NEST:bs10a03', description='CG32298 gene product from transcript CG32298-RB', other_designations='CG32298-PB|uncharacterized protein', map_location='62E1-62E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2389583, end_position_on_the_genomic_accession=2390989, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317964, CurrentID=0, Status='live', Symbol='CG32299', Aliases='Dmel_ BcDNA:AT26229, Dmel\\CG32299, NEST:bs19a11', description='CG32299 gene product from transcript CG32299-RB', other_designations='CG32299-PA|CG32299-PB|uncharacterized protein', map_location='62E1-62E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2389583, end_position_on_the_genomic_accession=2390989, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317966, CurrentID=0, Status='live', Symbol='obst-I', Aliases='Dmel_CG32304, CG13805, CG32304, Dmel\\CG32304', description='obstructor-I', other_designations='CG32304-PA|obst-I-PA', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2267858, end_position_on_the_genomic_accession=2268590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317967, CurrentID=0, Status='live', Symbol='CG32305', Aliases='Dmel_ AC, CG8970, Dmel\\CG32305', description='CG32305 gene product from transcript CG32305-RB', other_designations='CG32305-PB|uncharacterized protein', map_location='62D4-62D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2248635, end_position_on_the_genomic_accession=2252729, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  317969 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317969, CurrentID=0, Status='live', Symbol='CG42355', Aliases='Dmel_ CG32308, Dmel\\CG42355, Dmel_CG32308', description='CG42355 gene product from transcript CG42355-RD', other_designations='CG42355-PC|CG42355-PD|uncharacterized protein', map_location='62C2-62C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2037371, end_position_on_the_genomic_accession=2038224, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317973, CurrentID=0, Status='live', Symbol='CG32313', Aliases='Dmel_ Dmel\\CG32313', description='CG32313 gene product from transcript CG32313-RA', other_designations='CG32313-PA|CG32313-PB|CG32313-PC|uncharacterized protein', map_location='62A3-62A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1544828, end_position_on_the_genomic_accession=1545743, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317974 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317974, CurrentID=0, Status='live', Symbol='CG33791', Aliases='Dmel_ CG13918, CG32316, CG32316-ORFB, CG7934, Dmel\\CG33791, OGDH', description='CG33791 gene product from transcript CG33791-RB', other_designations='CG33791-PA|CG33791-PB|CG33791-PC|CG33791-PD|CG33791-PE|CG33791-PF|uncharacterized protein', map_location='62A7-62A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1602954, end_position_on_the_genomic_accession=1617355, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  317976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317976, CurrentID=0, Status='live', Symbol='CG32319', Aliases='Dmel_ CG12083, Dmel\\CG32319', description='CG32319 gene product from transcript CG32319-RA', other_designations='CG32319-PA|uncharacterized protein', map_location='62A2-62A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1470850, end_position_on_the_genomic_accession=1472060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  317977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317977, CurrentID=0, Status='live', Symbol='CG32320', Aliases='Dmel_ Dmel\\CG32320', description='CG32320 gene product from transcript CG32320-RD', other_designations='CG32320-PD|uncharacterized protein', map_location='62A1-62A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1398586, end_position_on_the_genomic_accession=1401612, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3772165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772165, CurrentID=0, Status='live', Symbol='Pex6', Aliases='Dmel_CG11919, CG11919, CG30019, DmCG11919, DmelPex6, Dmel\\CG11919, PEX6', description='Peroxin 6', other_designations='CG11919-PC|CG11919-PD|Pex6-PC|Pex6-PD|peroxin 6', map_location='47C5-47C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10879974, end_position_on_the_genomic_accession=10883202, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  317987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317987, CurrentID=0, Status='live', Symbol='Vps11', Aliases='Dmel_CG32350, BcDNA:LD20292, CG32350, Dmel\\CG32350', description='Vacuolar protein sorting 11', other_designations='CG32350-PA|Vps11-PA|vacuolar protein sorting 11', map_location='80D3-80D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23147656, end_position_on_the_genomic_accession=23150499, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772166, CurrentID=0, Status='live', Symbol='His2B:CG33892', Aliases='Dmel_CG33892, CG33892, Dmel\\CG33892, His2b:CG33892', description='CG33892 gene product from transcript CG33892-RA', other_designations='CG33892-PA|His2B:CG33892-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21477461, end_position_on_the_genomic_accession=21478055, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  38958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38958, CurrentID=0, Status='live', Symbol='mtrm', Aliases='Dmel_CG18543, CG18543, D52, Dmel\\CG18543, anon-D52, anonD52', description='matrimony', other_designations='CG18543-PA|mtrm-PA', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8413012, end_position_on_the_genomic_accession=8414103, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317996, CurrentID=0, Status='live', Symbol='tut', Aliases='Dmel_CG32364, CG32364, Dmel\\CG32364, anon-WO0118547.705', description='tumorous testis', other_designations='CG32364-PA|tut-PA', map_location='66B13-66C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8209577, end_position_on_the_genomic_accession=8210743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317997, CurrentID=0, Status='live', Symbol='CG32368', Aliases='Dmel_ BcDNA:RH61042, Dmel\\CG32368', description='CG32368 gene product from transcript CG32368-RA', other_designations='CG32368-PA|uncharacterized protein', map_location='66A19-66A19', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7915437, end_position_on_the_genomic_accession=7915846, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  317999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317999, CurrentID=0, Status='live', Symbol='CG32371', Aliases='Dmel_ BEST:GH23475, Dmel\\CG32371, EB1', description='CG32371 gene product from transcript CG32371-RA', other_designations='CG32371-PA|uncharacterized protein', map_location='66A12-66A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7767269, end_position_on_the_genomic_accession=7768352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318000, CurrentID=0, Status='live', Symbol='CG32373', Aliases='Dmel_ BcDNA:RE64043, CG3237, Dmel\\CG32373', description='CG32373 gene product from transcript CG32373-RB', other_designations='CG32373-PA|CG32373-PB|CG32373-PC|uncharacterized protein', map_location='66A8-66A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7650548, end_position_on_the_genomic_accession=7759123, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  318002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318002, CurrentID=0, Status='live', Symbol='CG32376', Aliases='Dmel_ CG8528, Dmel\\CG32376, SP148', description='CG32376 gene product from transcript CG32376-RA', other_designations='CG32376-PA|uncharacterized protein', map_location='66A5-66A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7541582, end_position_on_the_genomic_accession=7542457, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318005, CurrentID=0, Status='live', Symbol='sphinx1', Aliases='Dmel_CG32383, CG32383, CG8555, Dmel\\CG32383, SPH194, Sphinx-1, Sphinx1', description='CG32383 gene product from transcript CG32383-RB', other_designations='CG32383-PB|sphinx1-PB', map_location='66A1-66A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7438421, end_position_on_the_genomic_accession=7439405, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318007, CurrentID=0, Status='live', Symbol='CG32391', Aliases='Dmel_ Dmel\\CG32391', description='CG32391 gene product from transcript CG32391-RC', other_designations='CG32391-PC|uncharacterized protein', map_location='65E3-65E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7068211, end_position_on_the_genomic_accession=7069199, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318008, CurrentID=0, Status='live', Symbol='jv', Aliases='Dmel_CG32397, CG32397, Dmel\\CG32397, anon-WO0138581.6', description='javelin', other_designations='CG32397-PB|CG32397-PC|CG32397-PD|jv-PB|jv-PC|jv-PD', map_location='3-19.2 cM', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6563197, end_position_on_the_genomic_accession=6589479, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  318011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318011, CurrentID=0, Status='live', Symbol='Or65a', Aliases='Dmel_CG32401, 65a, CG32401, DOR66, Dmel\\CG32401, LU27.2', description='Odorant receptor 65a', other_designations='CG32401-PA|Or65a-PA|odorant receptor 65a', map_location='65A14-65A14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6320611, end_position_on_the_genomic_accession=6322098, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318012, CurrentID=0, Status='live', Symbol='Or65b', Aliases='Dmel_CG32402, 65b, CG32402, DOR63, Dmel\\CG32402, LU27.1', description='Odorant receptor 65b', other_designations='CG32402-PA|Or65b-PA|odorant receptor 65b', map_location='65A14-65A14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6325612, end_position_on_the_genomic_accession=6327049, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318013, CurrentID=0, Status='live', Symbol='Or65c', Aliases='Dmel_CG32403, 65c, CG32403, DOR65, Dmel\\CG32403, LU27.3', description='Odorant receptor 65c', other_designations='CG32403-PA|Or65c-PA|odorant receptor 65c', map_location='65A14-65A14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6327608, end_position_on_the_genomic_accession=6329069, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318014, CurrentID=0, Status='live', Symbol='Cpr65Av', Aliases='Dmel_CG32405, BcDNA:RE31218, CG10461, CG32405, DmelCpr65Av, Dmel\\CG32405', description='Cuticular protein 65Av', other_designations='CG32405-PA|Cpr65Av-PA|cuticular protein 65Av', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6139223, end_position_on_the_genomic_accession=6139974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  252479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=252479, CurrentID=0, Status='live', Symbol='DIP2', Aliases='Dmel_CG7020, CG7020, Ddip2, Dmel\\CG7020', description='DISCO Interacting Protein 2', other_designations='CG7020-PA|CG7020-PB|CG7020-PC|CG7020-PD|CG7020-PE|DIP2-PA|DIP2-PB|DIP2-PC|DIP2-PD|DIP2-PE|DISCO interacting protein 2', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=250229, end_position_on_the_genomic_accession=257954, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  318016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318016, CurrentID=0, Status='live', Symbol='CG32409', Aliases='Dmel_ BcDNA:LD24586, Dmel\\CG32409', description='CG32409 gene product from transcript CG32409-RA', other_designations='CG32409-PA|uncharacterized protein', map_location='64F4-64F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5899528, end_position_on_the_genomic_accession=5900936, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318018, CurrentID=0, Status='live', Symbol='CG32413', Aliases='Dmel_ CG10487, Dmel\\CG32413', description='CG32413 gene product from transcript CG32413-RC', other_designations='CG32413-PC|uncharacterized protein', map_location='64F5-64F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5902887, end_position_on_the_genomic_accession=5904052, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318020, CurrentID=0, Status='live', Symbol='CG42272', Aliases='Dmel_ CG32415, CG32416, Dmel\\CG42272, Dmel_CG32415, Dmel_CG32416', description='CG42272 gene product from transcript CG42272-RC', other_designations='CG42272-PC|CG42272-PE|uncharacterized protein', map_location='64E8-64E11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5803033, end_position_on_the_genomic_accession=5805749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36620, CurrentID=0, Status='live', Symbol='CG12868', Aliases='Dmel_ CT32008, Dmel\\CG12868', description='CG12868 gene product from transcript CG12868-RB', other_designations='CG12868-PB|uncharacterized protein', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14604549, end_position_on_the_genomic_accession=14605297, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318030, CurrentID=0, Status='live', Symbol='CG32440', Aliases='Dmel_ CG5051, Dmel\\CG32440, bs36c07.y1', description='CG32440 gene product from transcript CG32440-RA', other_designations='CG32440-PA|uncharacterized protein', map_location='78C8-78C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21360179, end_position_on_the_genomic_accession=21361229, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772173, CurrentID=0, Status='live', Symbol='His3:CG33863', Aliases='Dmel_CG33863, CG33863, Dmel\\CG33863', description='CG33863 gene product from transcript CG33863-RA', other_designations='CG33863-PA|His3:CG33863-PA|histone H3', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21534636, end_position_on_the_genomic_accession=21535164, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318032, CurrentID=0, Status='live', Symbol='CG32448', Aliases='Dmel_ Dmel\\CG32448', description='CG32448 gene product from transcript CG32448-RC', other_designations='CG32448-PB|CG32448-PC|uncharacterized protein', map_location='78F4-78F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21827774, end_position_on_the_genomic_accession=21828480, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318034, CurrentID=0, Status='live', Symbol='CG32450', Aliases='Dmel_ BEST:GH25462, BcDNA:GH25462, Dmel\\CG32450, anon-WO0140519.221', description='CG32450 gene product from transcript CG32450-RA', other_designations='CG32450-PA|uncharacterized protein', map_location='79E1-79E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22329778, end_position_on_the_genomic_accession=22330657, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318035, CurrentID=0, Status='live', Symbol='CG32452', Aliases='Dmel_ BcDNA:SD04526, Dmel\\CG32452', description='CG32452 gene product from transcript CG32452-RC', other_designations='CG32452-PA|CG32452-PB|CG32452-PC|uncharacterized protein', map_location='79F6-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22717728, end_position_on_the_genomic_accession=22719943, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318036, CurrentID=0, Status='live', Symbol='CG32453', Aliases='Dmel_ Dmel\\CG32453', description='CG32453 gene product from transcript CG32453-RB', other_designations='CG32453-PA|CG32453-PB|uncharacterized protein', map_location='79F6-79F6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22697230, end_position_on_the_genomic_accession=22697970, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318037, CurrentID=0, Status='live', Symbol='CG32454', Aliases='Dmel_ CG11367, Dmel\\CG32454', description='CG32454 gene product from transcript CG32454-RA', other_designations='CG32454-PA|uncharacterized protein', map_location='80A1-80A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22729766, end_position_on_the_genomic_accession=22731600, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318040, CurrentID=0, Status='live', Symbol='CG32457', Aliases='Dmel_ Dmel\\CG32457', description='CG32457 gene product from transcript CG32457-RB', other_designations='CG32457-PB|CG32457-PC|uncharacterized protein', map_location='80C2-80C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22987642, end_position_on_the_genomic_accession=22988262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318042, CurrentID=0, Status='live', Symbol='CG32461', Aliases='Dmel_ Dmel\\CG32461', description='CG32461 gene product from transcript CG32461-RA', other_designations='CG32461-PA|uncharacterized protein', map_location='80C1-80C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22954421, end_position_on_the_genomic_accession=22955302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318043, CurrentID=0, Status='live', Symbol='CG32462', Aliases='Dmel_ CG8220, Dmel\\CG32462', description='CG32462 gene product from transcript CG32462-RA', other_designations='CG32462-PA|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22852270, end_position_on_the_genomic_accession=22853512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318044, CurrentID=0, Status='live', Symbol='Tengl2', Aliases='Dmel_CG32463, CG17006, CG32463, Dmel\\CG32463', description='Testis EndoG-Like 2', other_designations='CG32463-PA|Tengl2-PA|testis EndoG-Like 2', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2238304, end_position_on_the_genomic_accession=2239384, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318046, CurrentID=0, Status='live', Symbol='mthl14', Aliases='Dmel_CG32476, BcDNA:RE31310, CG32476, Dmel\\CG32476', description='methuselah-like 14', other_designations='CG32476-PA|mthl14-PA', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=247491, end_position_on_the_genomic_accession=250305, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  36624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36624, CurrentID=0, Status='live', Symbol='CG10131', Aliases='Dmel_ Dmel\\CG10131', description='CG10131 gene product from transcript CG10131-RA', other_designations='CG10131-PA|uncharacterized protein', map_location='51B6-51B6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14629910, end_position_on_the_genomic_accession=14631023, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318050, CurrentID=0, Status='live', Symbol='CG32483', Aliases='Dmel_ Dmel\\CG32483', description='CG32483 gene product from transcript CG32483-RA', other_designations='CG32483-PA|uncharacterized protein', map_location='61C8-61C8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=688523, end_position_on_the_genomic_accession=690081, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318053, CurrentID=0, Status='live', Symbol='CG32495', Aliases='Dmel_ 32495, Dmel\\CG32495, GS', description='CG32495 gene product from transcript CG32495-RD', other_designations='CG32495-PB|CG32495-PD|CG32495-PE|CG32495-PF|CG32495-PG|CG32495-PH|CG32495-PI|glutathione synthetase|uncharacterized protein', map_location='16F1-16F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17899702, end_position_on_the_genomic_accession=17906025, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  318059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318059, CurrentID=0, Status='live', Symbol='CG32506', Aliases='Dmel_ Dmel\\CG32506', description='CG32506 gene product from transcript CG32506-RD', other_designations='CG32506-PC|CG32506-PD|uncharacterized protein', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20449331, end_position_on_the_genomic_accession=20467065, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  318060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318060, CurrentID=0, Status='live', Symbol='CG32507', Aliases='Dmel_ Dmel\\CG32507', description='CG32507 gene product from transcript CG32507-RA', other_designations='CG32507-PA|uncharacterized protein', map_location='19E3-19E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20780150, end_position_on_the_genomic_accession=20781220, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318064, CurrentID=0, Status='live', Symbol='stg1', Aliases='Dmel_CG33670, CG10985, CG11568, CG11569, CG32514, CG32514-ORFA, CG33670, Dmel\\CG33670, Dro STG1, STG1, Stg1', description='stargazin-like protein', other_designations='CG33670-PA|stg1-PA', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21093844, end_position_on_the_genomic_accession=21111417, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318066, CurrentID=0, Status='live', Symbol='CG34120', Aliases='Dmel_ BcDNA:AT04622, CG1819, CG32516, Dmel\\CG34120', description='CG34120 gene product from transcript CG34120-RE', other_designations='CG34120-PB|CG34120-PC|CG34120-PD|CG34120-PE|uncharacterized protein', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21123634, end_position_on_the_genomic_accession=21140951, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  3772179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772179, CurrentID=0, Status='live', Symbol='primo-1', Aliases='Dmel_CG33748, BEO2, CG31311, CG33748, CG9599, Dmel\\CG33748, LMW-PTP, Primo-1, primo', description='CG33748 gene product from transcript CG33748-RA', other_designations='CG33748-PA|CG33748-PB|primo-1-PA|primo-1-PB', map_location='87F12-87F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13710017, end_position_on_the_genomic_accession=13711835, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318071, CurrentID=0, Status='live', Symbol='CG32523', Aliases='Dmel_ CG1497, Dmel\\CG32523, SP50', description='CG32523 gene product from transcript CG32523-RA', other_designations='CG32523-PA|uncharacterized protein', map_location='19F2-19F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21217878, end_position_on_the_genomic_accession=21218835, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36628, CurrentID=0, Status='live', Symbol='CG12861', Aliases='Dmel_ Dmel\\CG12861', description='CG12861 gene product from transcript CG12861-RA', other_designations='CG12861-PA|uncharacterized protein', map_location='51B9-51B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14664246, end_position_on_the_genomic_accession=14665356, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36629, CurrentID=0, Status='live', Symbol='CG12860', Aliases='Dmel_ Dmel\\CG12860, anon-WO0140519.75, bs33f06.y1', description='CG12860 gene product from transcript CG12860-RA', other_designations='CG12860-PA|uncharacterized protein', map_location='51B9-51B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14665641, end_position_on_the_genomic_accession=14666817, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318082, CurrentID=0, Status='live', Symbol='CG32544', Aliases='Dmel_ Dmel\\CG32544', description='CG32544 gene product from transcript CG32544-RA', other_designations='CG32544-PA|CG32544-PB|uncharacterized protein', map_location='17F1-17F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18860284, end_position_on_the_genomic_accession=18861964, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318085, CurrentID=0, Status='live', Symbol='CG32551', Aliases='Dmel_ Dmel\\CG32551', description='CG32551 gene product from transcript CG32551-RB', other_designations='CG32551-PB|uncharacterized protein', map_location='17B1-17B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18438868, end_position_on_the_genomic_accession=18439260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318087, CurrentID=0, Status='live', Symbol='CG32553', Aliases='Dmel_ Dmel\\CG32553', description='CG32553 gene product from transcript CG32553-RA', other_designations='CG32553-PA|uncharacterized protein', map_location='16F7-16F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=18124575, end_position_on_the_genomic_accession=18127027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  252554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=252554, CurrentID=0, Status='live', Symbol='sav', Aliases='Dmel_CG33193, CG13831, CG13832, CG33193, Dmel\\CG33193, Salvador/Shar-pei, Sav, anon-WO0172774.152, shar-pei', description='salvador', other_designations='CG33193-PA|CG33193-PB|salvador/shar-pei|sav-PA|sav-PB', map_location='94D10-94D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23058609, end_position_on_the_genomic_accession=23061304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36631, CurrentID=0, Status='live', Symbol='CG30479', Aliases='Dmel_ CG18373, Dmel\\CG30479', description='CG30479 gene product from transcript CG30479-RA', other_designations='CG30479-PA|CG30479-PB|uncharacterized protein', map_location='51B10-51B10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14674465, end_position_on_the_genomic_accession=14678136, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318092, CurrentID=0, Status='live', Symbol='TwdlZ', Aliases='Dmel_CG32569, CG32569, Dmel\\CG32569', description='TweedleZ', other_designations='CG32569-PB|TwdlZ-PB', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16738178, end_position_on_the_genomic_accession=16738868, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318093, CurrentID=0, Status='live', Symbol='TwdlY', Aliases='Dmel_CG32570, CG32570, Dmel\\CG32570', description='TweedleY', other_designations='CG32570-PA|TwdlY-PA|tweedle Y', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16734947, end_position_on_the_genomic_accession=16735980, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318094, CurrentID=0, Status='live', Symbol='TwdlX', Aliases='Dmel_CG32571, CG32571, Dmel\\CG32571', description='TweedleX', other_designations='CG32571-PA|TwdlX-PA', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16731091, end_position_on_the_genomic_accession=16732630, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318095, CurrentID=0, Status='live', Symbol='CG32576', Aliases='Dmel_ BcDNA:RE29988, Dmel\\CG32576, NEST:bs01h05, anon-AE003501.1', description='CG32576 gene product from transcript CG32576-RB', other_designations='CG32576-PA|CG32576-PB|uncharacterized protein', map_location='14B7-14B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16285953, end_position_on_the_genomic_accession=16288533, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318096, CurrentID=0, Status='live', Symbol='CG32579', Aliases='Dmel_ Dmel\\CG32579', description='CG32579 gene product from transcript CG32579-RC', other_designations='CG32579-PA|CG32579-PB|CG32579-PC|uncharacterized protein', map_location='14B12-14B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16323286, end_position_on_the_genomic_accession=16325768, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318097, CurrentID=0, Status='live', Symbol='Muc14A', Aliases='Dmel_CG32580, CG32580, Dmel\\CG32580', description='Mucin 14A', other_designations='CG32580-PA|Muc14A-PA|mucin 14A', map_location='14A6-14A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16009114, end_position_on_the_genomic_accession=16061722, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  318098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318098, CurrentID=0, Status='live', Symbol='CG32581', Aliases='Dmel_ Dmel\\CG32581', description='CG32581 gene product from transcript CG32581-RB', other_designations='CG32581-PA|CG32581-PB|uncharacterized protein', map_location='13F1-13F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15796908, end_position_on_the_genomic_accession=15799162, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318101, CurrentID=0, Status='live', Symbol='CG32588', Aliases='Dmel_ BcDNA:AT31812, Dmel\\CG32588, NEST:bs07c08, bs07c08.y1', description='CG32588 gene product from transcript CG32588-RA', other_designations='CG32588-PA|uncharacterized protein', map_location='13B3-13B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15205807, end_position_on_the_genomic_accession=15206636, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318102, CurrentID=0, Status='live', Symbol='drd', Aliases='Dmel_CG33968, BcDNA:GM06434, CG32589, CG33968, CG5652, Dmel\\CG33968, beltless, blt, lwf', description='drop dead', other_designations="CG33968-PA|beltless|drd-PA|drop-dead|lot's wife|lots wife", map_location='13A12-13B', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15125753, end_position_on_the_genomic_accession=15137311, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  318103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318103, CurrentID=0, Status='live', Symbol='CG32590', Aliases='Dmel_ Dmel\\CG32590', description='CG32590 gene product from transcript CG32590-RA', other_designations='CG32590-PA|uncharacterized protein', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14838352, end_position_on_the_genomic_accession=14839106, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318104, CurrentID=0, Status='live', Symbol='CG32591', Aliases='Dmel_ Dmel\\CG32591, NEST:bs09h01, NEST:bs18d12', description='CG32591 gene product from transcript CG32591-RB', other_designations='CG32591-PA|CG32591-PB|uncharacterized protein', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14927202, end_position_on_the_genomic_accession=14928944, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318105, CurrentID=0, Status='live', Symbol='hog', Aliases='Dmel_CG32595, CG32595, Dmel\\CG32595', description='CG32595 gene product from transcript CG32595-RA', other_designations='CG32595-PA|hog-PA', map_location='12F3-12F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14718273, end_position_on_the_genomic_accession=14718942, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318107, CurrentID=0, Status='live', Symbol='betaNACtes6', Aliases='Dmel_CG32598, CG32598, Dmel\\CG32598', description='CG32598 gene product from transcript CG32598-RA', other_designations='CG32598-PA|betaNACtes6-PA', map_location='12E8-12E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14316130, end_position_on_the_genomic_accession=14317150, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318108, CurrentID=0, Status='live', Symbol='CG32599', Aliases='Dmel_ CG12047, Dmel\\CG32599, GM07807', description='CG32599 gene product from transcript CG32599-RA', other_designations='CG32599-PA|uncharacterized protein', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14243686, end_position_on_the_genomic_accession=14246514, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  36634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36634, CurrentID=0, Status='live', Symbol='CG12858', Aliases='Dmel_ Dmel\\CG12858', description='CG12858 gene product from transcript CG12858-RB', other_designations='CG12858-PA|CG12858-PB|uncharacterized protein', map_location='51C1-51C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14739359, end_position_on_the_genomic_accession=14745335, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3346203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346203, CurrentID=0, Status='live', Symbol='Acp53C14c', Aliases='Dmel_CG33530, CG33530, Dmel\\CG33530, HDC06801', description='Accessory gland protein 53C14c', other_designations='Acp53C14c-PA|Acp53C14c-PB|CG33530-PA|CG33530-PB|accessory gland protein 53C14c', map_location='53D6-53D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16750776, end_position_on_the_genomic_accession=16751495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8673956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673956, CurrentID=0, Status='live', Symbol='CG7606', Aliases='Dmel_ Dmel\\CG7606', description='CG7606 gene product from transcript CG7606-RC', other_designations='CG7606-PB|CG7606-PC', map_location='90B5-90B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17596536, end_position_on_the_genomic_accession=17598453, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673957, CurrentID=0, Status='live', Symbol='Mst36Fb', Aliases='Dmel_CG42626, CG15165, CG31791, CG42626, Dmel\\CG42626, Dmel_CG31791, mst36Fb', description='Male-specific transcript 36Fb', other_designations='CG42626-PB|Mst36Fb-PB', map_location='36F10-36F10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18673363, end_position_on_the_genomic_accession=18675094, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8673960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673960, CurrentID=0, Status='live', Symbol='CG42467', Aliases='Dmel_ Dmel\\CG42467', description='CG42467 gene product from transcript CG42467-RA', other_designations='CG42467-PA|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3701290, end_position_on_the_genomic_accession=3701699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3346204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346204, CurrentID=0, Status='live', Symbol='CheB53b', Aliases='Dmel_CG33524, CG33457, CG33524, Dmel\\CG33524', description='Chemosensory protein B 53b', other_designations='CG33524-PA|CheB53b-PA|chemosensory protein B 53b', map_location='53C5-53C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16468209, end_position_on_the_genomic_accession=16468934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8673962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673962, CurrentID=0, Status='live', Symbol='CG42486', Aliases='Dmel_ CG42501, Dmel\\CG42486', description='CG42486 gene product from transcript CG42486-RA', other_designations='CG42486-PA|CG42486-PB|uncharacterized protein', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11843137, end_position_on_the_genomic_accession=11844012, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8673963 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673963, CurrentID=0, Status='live', Symbol='Sfp93F', Aliases='Dmel_CG42608, CG42608, Dmel\\CG42608', description='Seminal fluid protein 93F', other_designations='CG42608-PA|Sfp93F-PA|seminal fluid protein 93F', map_location='93F7-93F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21735909, end_position_on_the_genomic_accession=21736296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673965, CurrentID=0, Status='live', Symbol='CG42638', Aliases='Dmel_ Dmel\\CG42638, noncoding_13567', description='CG42638 gene product from transcript CG42638-RA', other_designations='CG42638-PA|uncharacterized protein', map_location='76C3-76C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19745834, end_position_on_the_genomic_accession=19749744, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8673966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673966, CurrentID=0, Status='live', Symbol='Skadu', Aliases='Dmel_CG42448, CG42448, Dmel\\CG42448, SKADU', description='CG42448 gene product from transcript CG42448-RA', other_designations='CG42448-PA|Skadu-PA|skadu', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15547519, end_position_on_the_genomic_accession=15548243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772189, CurrentID=0, Status='live', Symbol='His3:CG33866', Aliases='Dmel_CG33866, CG33866, Dmel\\CG33866', description='CG33866 gene product from transcript CG33866-RA', other_designations='CG33866-PA|His3:CG33866-PA|histone H3', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21541153, end_position_on_the_genomic_accession=21541681, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318128, CurrentID=0, Status='live', Symbol='CG32625', Aliases='Dmel_ Dmel\\CG32625', description='CG32625 gene product from transcript CG32625-RA', other_designations='CG32625-PA|uncharacterized protein', map_location='12B10-12B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13756164, end_position_on_the_genomic_accession=13756826, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673970, CurrentID=0, Status='live', Symbol='CG40472', Aliases='Dmel_ CR40472, Dmel\\CG40472', description='CG40472 gene product from transcript CG40472-RC', other_designations='CG40472-PC|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=24927130, end_position_on_the_genomic_accession=24927710, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8673971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673971, CurrentID=0, Status='live', Symbol='Sfp38D', Aliases='Dmel_CG42606, CG42606, Dmel\\CG42606', description='Seminal fluid protein 38D', other_designations='CG42606-PA|CG42606-PB|Sfp38D-PA|Sfp38D-PB|seminal fluid protein 38D', map_location='38D1-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20640917, end_position_on_the_genomic_accession=20642216, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673972, CurrentID=0, Status='live', Symbol='CG42531', Aliases='Dmel_ Dmel\\CG42531', description='CG42531 gene product from transcript CG42531-RA', other_designations='CG42531-PA|CG42531-PB|uncharacterized protein', map_location='48C5-48C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11864161, end_position_on_the_genomic_accession=11864941, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3346206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346206, CurrentID=0, Status='live', Symbol='CG42313', Aliases='Dmel_ 42313, BG:DS01219.3, BG:DS03144.1, BG:SD01219.3, CG15275, CG15276, CG15277, CG33515, CT35222, Dmel\\CG42313, Dmel_CG15277, Dmel_CG33515, cg15277', description='CG42313 gene product from transcript CG42313-RB', other_designations='CG42313-PA|CG42313-PB|uncharacterized protein', map_location='35B7-35B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14907489, end_position_on_the_genomic_accession=14972300, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  318134 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318134, CurrentID=0, Status='live', Symbol='CG32639', Aliases='Dmel_ Dmel\\CG32639', description='CG32639 gene product from transcript CG32639-RA', other_designations='CG32639-PA|uncharacterized protein', map_location='11E6-11E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13235925, end_position_on_the_genomic_accession=13236566, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318135, CurrentID=0, Status='live', Symbol='CG32640', Aliases='Dmel_ Dmel\\CG32640', description='CG32640 gene product from transcript CG32640-RA', other_designations='CG32640-PA|uncharacterized protein', map_location='11E3-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13178005, end_position_on_the_genomic_accession=13178523, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318136, CurrentID=0, Status='live', Symbol='CG32641', Aliases='Dmel_ Dmel\\CG32641', description='CG32641 gene product from transcript CG32641-RA', other_designations='CG32641-PA|uncharacterized protein', map_location='11E3-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13184478, end_position_on_the_genomic_accession=13184996, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318137, CurrentID=0, Status='live', Symbol='CG32642', Aliases='Dmel_ Dmel\\CG32642', description='CG32642 gene product from transcript CG32642-RE', other_designations='CG32642-PB|CG32642-PE|uncharacterized protein', map_location='11D5-11D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12959220, end_position_on_the_genomic_accession=12960790, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318138, CurrentID=0, Status='live', Symbol='CG32643', Aliases='Dmel_ Dmel\\CG32643, NEST:bs08h10', description='CG32643 gene product from transcript CG32643-RA', other_designations='CG32643-PA|uncharacterized protein', map_location='11D6-11D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12984524, end_position_on_the_genomic_accession=12985071, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318139, CurrentID=0, Status='live', Symbol='Mur11Da', Aliases='Dmel_CG32644, CG32644, Dmel\\CG32644', description='Mucin related 11Da', other_designations='CG32644-PD|Mur11Da-PD|mucin related 11Da', map_location='11D5-11D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12967888, end_position_on_the_genomic_accession=12970056, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8673980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673980, CurrentID=0, Status='live', Symbol='Sfp60F', Aliases='Dmel_CG42478, CG42478, Dmel\\CG42478', description='Seminal fluid protein 60F', other_designations='CG42478-PA|CG42478-PB|Sfp60F-PA|Sfp60F-PB|seminal fluid protein 60F', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25172245, end_position_on_the_genomic_accession=25172695, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673981, CurrentID=0, Status='live', Symbol='Sfp26Ad', Aliases='Dmel_CG42603, CG42603, Dmel\\CG42603', description='Seminal fluid protein 26Ad', other_designations='CG42603-PA|CG42603-PB|Sfp26Ad-PA|Sfp26Ad-PB|seminal fluid protein 26Ad', map_location='26A3-26A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5942623, end_position_on_the_genomic_accession=5943304, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673982, CurrentID=0, Status='live', Symbol='CG43127', Aliases='Dmel_ CG34237, CG42519, CG42520, Dmel\\CG43127, Dmel_CG34237, Dmel_CG42519, Dmel_CG42520', description='CG43127 gene product from transcript CG43127-RD', other_designations='CG43127-PC|CG43127-PD|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10624728, end_position_on_the_genomic_accession=10625599, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318143, CurrentID=0, Status='live', Symbol='CG32655', Aliases='Dmel_ Dmel\\CG32655', description='CG32655 gene product from transcript CG32655-RA', other_designations='CG32655-PA|uncharacterized protein', map_location='11A7-11A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12234562, end_position_on_the_genomic_accession=12235877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673984, CurrentID=0, Status='live', Symbol='Sfp24Ba', Aliases='Dmel_CG42461, CG42461, Dmel\\CG42461, SFP24BA', description='Seminal fluid protein 24Ba', other_designations='CG42461-PA|Sfp24Ba-PA|seminal fluid protein 24Ba', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3667977, end_position_on_the_genomic_accession=3668551, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3346208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346208, CurrentID=0, Status='live', Symbol='dpr3', Aliases='Dmel_CG33516, CG15379, CG15380, CG15381, CG33516, CRACM2, CT35425, Dmel\\CG33516, Dpr-3, Dpr3', description='defective proboscis extension response 3', other_designations='CG33516-PB|CG33516-PC|CG33516-PD|CRAC modulator 2|dpr3-PB|dpr3-PC|dpr3-PD', map_location='22C1-22C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2058790, end_position_on_the_genomic_accession=2109878, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  318146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318146, CurrentID=0, Status='live', Symbol='ssp7', Aliases='Dmel_CG32667, CG32667, CG32677, Dmel\\CG32667, Ssp7', description='short spindle 7', other_designations='CG32667-PA|ssp7-PA', map_location='10A7-10A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11139762, end_position_on_the_genomic_accession=11140263, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318147, CurrentID=0, Status='live', Symbol='CG32668', Aliases='Dmel_ Dmel\\CG32668', description='CG32668 gene product from transcript CG32668-RA', other_designations='CG32668-PA|uncharacterized protein', map_location='10A8-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11156676, end_position_on_the_genomic_accession=11159760, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  318148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318148, CurrentID=0, Status='live', Symbol='Rab9E', Aliases='Dmel_CG32673, CG32673, Dmel\\CG32673, anon-AE003451.2', description='Rab at 9E', other_designations='CG32673-PA|Rab GTPase 9E|Rab9E-PA', map_location='9E1-9E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10660955, end_position_on_the_genomic_accession=10661843, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318149, CurrentID=0, Status='live', Symbol='Rab9D', Aliases='Dmel_CG32678, CG32678, Dmel\\CG32678, anon-AE003451.1', description='Rab at 9D', other_designations='CG32678-PA|Rab GTPase 9D|Rab9D-PA', map_location='9D3-9D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10589089, end_position_on_the_genomic_accession=10589985, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318150, CurrentID=0, Status='live', Symbol='CG32679', Aliases='Dmel_ Dmel\\CG32679', description='CG32679 gene product from transcript CG32679-RA', other_designations='CG32679-PA|uncharacterized protein', map_location='9B12-9B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10366402, end_position_on_the_genomic_accession=10367472, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36641, CurrentID=0, Status='live', Symbol='Trs31', Aliases='Dmel_CG10153, CG10153, Dmel\\CG10153p, Trs31', description='TRAPP subunit 31 ortholog (S. cerevisiae)', other_designations='CG10153-PA|TRAPP subunit 31 ortholog|Trs31-PA', map_location='51C2-51C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14757064, end_position_on_the_genomic_accession=14757846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8673992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673992, CurrentID=0, Status='live', Symbol='Nost', Aliases='Dmel_CG42388, CG15364, CG33224, CG4040, CG42388, Dmel\\CG42388, Dmel_CG15364, Dmel_CG33224, Dmel_CG4040, NOSTRIN, nost', description='Nostrin', other_designations='CG42388-PC|CG42388-PD|CG42388-PE|CG42388-PF|CG42388-PG|CG42388-PH|Nost-PC|Nost-PD|Nost-PE|Nost-PF|Nost-PG|Nost-PH|nostrin', map_location='8C4-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9030205, end_position_on_the_genomic_accession=9054575, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  8673993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673993, CurrentID=0, Status='live', Symbol='CG42508', Aliases='Dmel_ Dmel\\CG42508', description='CG42508 gene product from transcript CG42508-RB', other_designations='CG42508-PA|CG42508-PB|uncharacterized protein', map_location='92B6-92B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19914320, end_position_on_the_genomic_accession=19915412, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8673994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673994, CurrentID=0, Status='live', Symbol='Ir75b', Aliases='Dmel_CG42643, CG14586, CG14586-B, CG42643, DmelIR75b, Dmel\\CG42643, IR75abc, IR75b', description='Ionotropic receptor 75b', other_designations='CG42643-PA|Ir75b-PA|ionotropic receptor 75b', map_location='75A2-75A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17818073, end_position_on_the_genomic_accession=17820193, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  8673995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673995, CurrentID=0, Status='live', Symbol='Sfp70A4', Aliases='Dmel_CG42480, CG42480, Dmel\\CG42480', description='Seminal fluid protein 70A4', other_designations='CG42480-PA|CG42480-PB|Sfp70A4-PA|Sfp70A4-PB|seminal fluid protein 70A4', map_location='70A4-70A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13296530, end_position_on_the_genomic_accession=13296860, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318158, CurrentID=0, Status='live', Symbol='Gr9a', Aliases='Dmel_CG32693, CG32693, Dmel\\CG32693, GR9a', description='Gustatory receptor 9a', other_designations='CG32693-PA|Gr9a-PA|gustatory receptor 9a', map_location='9B1-9B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10061627, end_position_on_the_genomic_accession=10062785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8673999 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673999, CurrentID=0, Status='live', Symbol='CG42577', Aliases='Dmel_ Dmel\\CG42577', description='CG42577 gene product from transcript CG42577-RA', other_designations='CG42577-PA|uncharacterized protein', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20259239, end_position_on_the_genomic_accession=20259656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318160, CurrentID=0, Status='live', Symbol='CG32706', Aliases='Dmel_ Dmel\\CG32706', description='CG32706 gene product from transcript CG32706-RA', other_designations='CG32706-PA|uncharacterized protein', map_location='8C7-8C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9087129, end_position_on_the_genomic_accession=9088073, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318161, CurrentID=0, Status='live', Symbol='CG32708', Aliases='Dmel_ CG6997, Dmel\\CG32708', description='CG32708 gene product from transcript CG32708-RA', other_designations='CG32708-PA|uncharacterized protein', map_location='8C7-8C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9085932, end_position_on_the_genomic_accession=9086864, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674002, CurrentID=0, Status='live', Symbol='CG42544', Aliases='Dmel_ Dmel\\CG42544, HCD12266', description='CG42544 gene product from transcript CG42544-RA', other_designations='CG42544-PA|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7304592, end_position_on_the_genomic_accession=7304954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318164, CurrentID=0, Status='live', Symbol='lawc', Aliases='Dmel_CG32711, BcDNA:RH09582, CG32711, Dmel\\CG32711, l(1)7Ea, lawC', description='leg arista wing complex', other_designations='CG32711-PA|CG32711-PB|CG32711-PC|CG32711-PD|lawc-PA|lawc-PB|lawc-PC|lawc-PD|leg-arista-wing complex|lethal(1)7Ea', map_location='1-23.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8408008, end_position_on_the_genomic_accession=8411658, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318165, CurrentID=0, Status='live', Symbol='CG32712', Aliases='Dmel_ Dmel\\CG32712', description='CG32712 gene product from transcript CG32712-RB', other_designations='CG32712-PB|uncharacterized protein', map_location='7E9-7E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8433087, end_position_on_the_genomic_accession=8433809, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318166, CurrentID=0, Status='live', Symbol='CG32713', Aliases='Dmel_ Dmel\\CG32713', description='CG32713 gene product from transcript CG32713-RB', other_designations='CG32713-PA|CG32713-PB|uncharacterized protein', map_location='7E11-7E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8456837, end_position_on_the_genomic_accession=8457632, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674007, CurrentID=0, Status='live', Symbol='Sfp33A4', Aliases='Dmel_CG42604, CG42604, Dmel\\CG42604, anon-SAGE:Wang-036', description='Seminal fluid protein 33A4', other_designations='CG42604-PA|Sfp33A4-PA|anon-SAGE:Wang-036|seminal fluid protein 33A4', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11837243, end_position_on_the_genomic_accession=11837486, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674008, CurrentID=0, Status='live', Symbol='Sfp51E', Aliases='Dmel_CG42476, CG42476, Dmel\\CG42476', description='Seminal fluid protein 51E', other_designations='CG42476-PA|Sfp51E-PA|seminal fluid protein 51E', map_location='51E5-51E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15192628, end_position_on_the_genomic_accession=15193228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674009, CurrentID=0, Status='live', Symbol='Sfp65A', Aliases='Dmel_CG42479, CG42479, Dmel\\CG42479', description='Seminal fluid protein 65A', other_designations='CG42479-PA|Sfp65A-PA|seminal fluid protein 65A', map_location='65A15-65A15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6350276, end_position_on_the_genomic_accession=6350829, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318170, CurrentID=0, Status='live', Symbol='CG32720', Aliases='Dmel_ Dmel\\CG32720, NEST:bs01f06', description='CG32720 gene product from transcript CG32720-RB', other_designations='CG32720-PB|uncharacterized protein', map_location='7B3-7B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7535625, end_position_on_the_genomic_accession=7536959, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674011, CurrentID=0, Status='live', Symbol='CG42570', Aliases='Dmel_ Dmel\\CG42570, new ORF', description='CG42570 gene product from transcript CG42570-RA', other_designations='CG42570-PA|CG42570-PB|uncharacterized protein', map_location='71F2-71F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15717819, end_position_on_the_genomic_accession=15720073, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  8674012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674012, CurrentID=0, Status='live', Symbol='CG42521', Aliases='Dmel_ CG34237, Dmel\\CG42521, Dmel_CG34237', description='CG42521 gene product from transcript CG42521-RA', other_designations='CG42521-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10623066, end_position_on_the_genomic_accession=10623576, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674013, CurrentID=0, Status='live', Symbol='CG42493', Aliases='Dmel_ Dmel\\CG42493', description='CG42493 gene product from transcript CG42493-RA', other_designations='CG42493-PA|uncharacterized protein', map_location='48C2-48C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11778526, end_position_on_the_genomic_accession=11779235, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8674014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674014, CurrentID=0, Status='live', Symbol='Sfp79B', Aliases='Dmel_CG42483, CG42483, Dmel\\CG42483', description='Seminal fluid protein 79', other_designations='CG42483-PA|Sfp79B-PA|seminal fluid protein 79|seminal fluid protein 79B', map_location='79B2-79B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22080322, end_position_on_the_genomic_accession=22080554, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318175, CurrentID=0, Status='live', Symbol='CG32726', Aliases='Dmel_ BcDNA:RH59280, Dmel\\CG32726', description='CG32726 gene product from transcript CG32726-RA', other_designations='CG32726-PA|CG32726-PB|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7425220, end_position_on_the_genomic_accession=7425743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674016, CurrentID=0, Status='live', Symbol='Sfp24F', Aliases='Dmel_CG42468, CG42468, Dmel\\CG42468, SFP24F', description='Seminal fluid protein 24F', other_designations='CG42468-PA|CG42468-PB|Sfp24F-PA|Sfp24F-PB|seminal fluid protein 24F', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4441429, end_position_on_the_genomic_accession=4442060, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674018, CurrentID=0, Status='live', Symbol='CG42505', Aliases='Dmel_ Dmel\\CG42505', description='CG42505 gene product from transcript CG42505-RA', other_designations='CG42505-PA|uncharacterized protein', map_location='87A5-87A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12028070, end_position_on_the_genomic_accession=12031757, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8674019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674019, CurrentID=0, Status='live', Symbol='mtTFB1', Aliases='Dmel_CG42631, CG42631, CG7319, Dmel\\CG42631, Dmel_CG7319, TFB1, d-TFB1M, d-mtTFB1', description='Mitochondrial Transcription Factor B1', other_designations='CG42631-PA|mitochondrial transcription factor B1|mitochondrial translation factor B1|mtTFB1-PA', map_location='68D2-68D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11708979, end_position_on_the_genomic_accession=11710570, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674020, CurrentID=0, Status='live', Symbol='CG42471', Aliases='Dmel_ Dmel\\CG42471', description='CG42471 gene product from transcript CG42471-RA', other_designations='CG42471-PA|uncharacterized protein', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11753640, end_position_on_the_genomic_accession=11754022, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674021, CurrentID=0, Status='live', Symbol='CG42503', Aliases='Dmel_ Dmel\\CG42503, Dromel_CG10238_FBtr0084782_mocs2_uORF', description='CG42503 gene product from transcript CG42503-RA', other_designations='CG42503-PA|uncharacterized protein', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25125632, end_position_on_the_genomic_accession=25127359, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8674022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674022, CurrentID=0, Status='live', Symbol='CG14397', Aliases='Dmel_ Dmel\\CG14397', description='CG14397 gene product from transcript CG14397-RA', other_designations='CG14397-PA|uncharacterized protein', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21091227, end_position_on_the_genomic_accession=21094864, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8674023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674023, CurrentID=0, Status='live', Symbol='CG42535', Aliases='Dmel_ Dmel\\CG42535', description='CG42535 gene product from transcript CG42535-RA', other_designations='CG42535-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10630034, end_position_on_the_genomic_accession=10630935, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8674024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674024, CurrentID=0, Status='live', Symbol='CG42558', Aliases='Dmel_ CG7866, Dmel\\CG42558, Dmel_CG7866', description='CG42558 gene product from transcript CG42558-RA', other_designations='CG42558-PA|uncharacterized protein', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29933287, end_position_on_the_genomic_accession=29934587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  8674025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674025, CurrentID=0, Status='live', Symbol='CG42579', Aliases='Dmel_ Dmel\\CG42579, HDC19846', description='CG42579 gene product from transcript CG42579-RA', other_designations='CG42579-PA|uncharacterized protein', map_location='19D2-19D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20527631, end_position_on_the_genomic_accession=20528043, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674026, CurrentID=0, Status='live', Symbol='Gyc76C', Aliases='Dmel_CG42636, CG32215, CG42636, CG8742, DGC1, Dmel\\CG42636, Dmel_CG8742, DrGC-1, GYC 76C, drgc, gim, gyc76C, l(3)76BDl, l(3)L0090', description='Guanylyl cyclase at 76C', other_designations='CG42636-PA|CG42636-PB|CG42636-PC|CG42636-PD|Gyc76C-PA|Gyc76C-PB|Gyc76C-PC|Gyc76C-PD|gimli|guanylyl cyclase at 76C|l(3)76BDl|lethal (3) L0090|receptor-type guanylate cyclase', map_location='76C2-100', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19750943, end_position_on_the_genomic_accession=19793569, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  8674028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674028, CurrentID=0, Status='live', Symbol='CG42463', Aliases='Dmel_ Dmel\\CG42463', description='CG42463 gene product from transcript CG42463-RB', other_designations='CG42463-PB|uncharacterized protein', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3668691, end_position_on_the_genomic_accession=3669134, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318189, CurrentID=0, Status='live', Symbol='CG32751', Aliases='Dmel_ CG3648, Dmel\\CG32751', description='CG32751 gene product from transcript CG32751-RB', other_designations='CG32751-PA|CG32751-PB|pantetheine hydrolase|uncharacterized protein', map_location='5D8-5D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6196580, end_position_on_the_genomic_accession=6198654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8674030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674030, CurrentID=0, Status='live', Symbol='CG17931', Aliases='Dmel_ BcDNA:AT19233, Dmel\\CG17931', description='CG17931 gene product from transcript CG17931-RD', other_designations='CG17931-PA|CG17931-PC|CG17931-PD|CG17931-PE|uncharacterized protein', map_location='89B13-89B13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16357535, end_position_on_the_genomic_accession=16359466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674031, CurrentID=0, Status='live', Symbol='CG42556', Aliases='Dmel_ CG15133, Dmel\\CG42556, Dmel_CG15133', description='CG42556 gene product from transcript CG42556-RA', other_designations='CG42556-PA|uncharacterized protein', map_location='36B3-36B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16889908, end_position_on_the_genomic_accession=16891239, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318192, CurrentID=0, Status='live', Symbol='CG32755', Aliases='Dmel_ Dmel\\CG32755', description='CG32755 gene product from transcript CG32755-RA', other_designations='CG32755-PA|uncharacterized protein', map_location='5C10-5C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5950251, end_position_on_the_genomic_accession=5951711, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318193, CurrentID=0, Status='live', Symbol='CG32756', Aliases='Dmel_ BcDNA:LD29716, Dmel\\CG32756', description='CG32756 gene product from transcript CG32756-RB', other_designations='CG32756-PA|CG32756-PB|uncharacterized protein', map_location='5C10-5C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5945682, end_position_on_the_genomic_accession=5948353, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  8674035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674035, CurrentID=0, Status='live', Symbol='CG42489', Aliases='Dmel_ Dmel\\CG42489', description='CG42489 gene product from transcript CG42489-RB', other_designations='CG42489-PA|CG42489-PB|CG42489-PC|uncharacterized protein', map_location='85A5-85A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8653802, end_position_on_the_genomic_accession=8654766, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674036, CurrentID=0, Status='live', Symbol='CG42615', Aliases='Dmel_ Dmel\\CG42615, Dmel_CG11861, Dmel_CG31829, gft', description='CG42615 gene product from transcript CG42615-RB', other_designations='CG42615-PB|uncharacterized protein', map_location='44E3-44E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8737370, end_position_on_the_genomic_accession=8737790, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674037, CurrentID=0, Status='live', Symbol='CG42488', Aliases='Dmel_ Dmel\\CG42488, Dromel_CG5116_FBtr0084877_uORF', description='CG42488 gene product from transcript CG42488-RB', other_designations='CG42488-PB|uncharacterized protein', map_location='96E2-96E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25483098, end_position_on_the_genomic_accession=25485481, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318198, CurrentID=0, Status='live', Symbol='CG32762', Aliases='Dmel_ Dmel\\CG32762', description='CG32762 gene product from transcript CG32762-RB', other_designations='CG32762-PA|CG32762-PB|uncharacterized protein', map_location='5A1-5A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5500208, end_position_on_the_genomic_accession=5501229, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674039, CurrentID=0, Status='live', Symbol='CG42518', Aliases='Dmel_ CG5134, Dmel\\CG42518, Dmel_CG5134', description='CG42518 gene product from transcript CG42518-RB', other_designations='CG42518-PA|CG42518-PB|mediator complex subunit 9|uncharacterized protein', map_location='55B11-55B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18167227, end_position_on_the_genomic_accession=18168380, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  8674042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674042, CurrentID=0, Status='live', Symbol='CG42598', Aliases='Dmel_ Dmel\\CG42598', description='CG42598 gene product from transcript CG42598-RB', other_designations='CG42598-PB|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=26354140, end_position_on_the_genomic_accession=26371356, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8674043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674043, CurrentID=0, Status='live', Symbol='Sfp77F', Aliases='Dmel_CG42482, CG42482, Dmel\\CG42482', description='Seminal fluid protein 77F', other_designations='CG42482-PA|CG42482-PB|Sfp77F-PA|Sfp77F-PB|seminal fluid protein 77F', map_location='77F4-77F4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20896268, end_position_on_the_genomic_accession=20897417, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674044, CurrentID=0, Status='live', Symbol='dpr21', Aliases='Dmel_CG42596, CG42596, Dmel\\CG42596, Dpr21', description='defective proboscis extension response 21', other_designations='CG42596-PC|dpr21-PC', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=3066499, end_position_on_the_genomic_accession=3191011, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318206, CurrentID=0, Status='live', Symbol='Tlk', Aliases='Dmel_CG34412, BcDNA:GH07910, CG12462, CG2829, CG32781, CG32782, CG33219, CG34412, Dmel\\CG34412, Dmel_CG32781, Dmel_CG32782, EP(X)1413, EP1413, TLK, anon-WO03040301.114, anon-WO03040301.116, dtlk, l(1)G0054, tlk', description='Tousled-like kinase', other_designations='CG34412-PB|CG34412-PC|CG34412-PE|CG34412-PF|CG34412-PG|CG34412-PH|CG34412-PI|CG34412-PJ|CG34412-PK|CG34412-PL|CG34412-PM|CG34412-PN|Tlk-PB|Tlk-PC|Tlk-PE|Tlk-PF|Tlk-PG|Tlk-PH|Tlk-PI|Tlk-PJ|Tlk-PK|Tlk-PL|Tlk-PM|Tlk-PN|tousled like kinase|tousled-like kinase', map_location='3F1-3E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3720007, end_position_on_the_genomic_accession=3789770, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  252671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=252671, CurrentID=0, Status='live', Symbol='X11L', Aliases='Dmel_CG5675, CG5675, CG5678, Dmel\\CG5675, Dmint, Dmint1, MINT, Mint, X11, X11-like, X11/Mintalpha, dX11, dX11L, x11/Mint, x11Lalpha', description='CG5675 gene product from transcript CG5675-RF', other_designations='CG5675-PA|CG5675-PB|CG5675-PC|CG5675-PE|CG5675-PF|X11-like|X11/Mint|X11L-PA|X11L-PB|X11L-PC|X11L-PE|X11L-PF|dX11/Mint/Lin-10', map_location='16B7-16B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17599608, end_position_on_the_genomic_accession=17606879, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  8674048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674048, CurrentID=0, Status='live', Symbol='Sfp36F', Aliases='Dmel_CG42605, CG42605, Dmel\\CG42605', description='Seminal fluid protein 36F', other_designations='CG42605-PA|CG42605-PB|Sfp36F-PA|Sfp36F-PB|seminal fluid protein 36F', map_location='36F2-36F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18292522, end_position_on_the_genomic_accession=18292833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674049, CurrentID=0, Status='live', Symbol='Pif1A', Aliases='Dmel_CG42599, BcDNA:GH12504, CG11773, CG11774, CG11776, CG32940, CG33719, CG33719/CG33720, CG42599, CG9808, DM22, Dmel\\CG42599, Dmel_CG11776, Dmel_CG33719, PIF-1A, Pif1', description='PFTAIRE-interacting factor 1A', other_designations='CG42599-PA|CG42599-PG|CG42599-PH|CG42599-PI|CG42599-PJ|Pif1A-PA|Pif1A-PG|Pif1A-PH|Pif1A-PI|Pif1A-PJ', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8774286, end_position_on_the_genomic_accession=8801176, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  318210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318210, CurrentID=0, Status='live', Symbol='CG32786', Aliases='Dmel_ Dmel\\CG32786', description='CG32786 gene product from transcript CG32786-RB', other_designations='CG32786-PB|uncharacterized protein', map_location='3F8-3F9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3940847, end_position_on_the_genomic_accession=3944265, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318211, CurrentID=0, Status='live', Symbol='HIP', Aliases='Dmel_CG32789, CG32789, Dmel\\CG32789, anon-EST:fe1A9', description='Hsc/Hsp70-interacting protein', other_designations='CG32789-PA|HIP-PA|anon-fast-evolving-1A9', map_location='3F2-3F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3799150, end_position_on_the_genomic_accession=3801570, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  8674052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674052, CurrentID=0, Status='live', Symbol='CG42633', Aliases='Dmel_ CG6547, Dmel\\CG42633', description='CG42633 gene product from transcript CG42633-RA', other_designations='CG42633-PA|uncharacterized protein', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10780965, end_position_on_the_genomic_accession=10782267, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318213, CurrentID=0, Status='live', Symbol='ppk8', Aliases='Dmel_CG32792, CG32792, Dmel\\CG32792, ppk29', description='pickpocket 8', other_designations='CG32792-PG|CG32792-PH|ppk8-PG|ppk8-PH', map_location='3D6-3D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3557812, end_position_on_the_genomic_accession=3572000, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  8674054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674054, CurrentID=0, Status='live', Symbol='Sfp23F', Aliases='Dmel_CG42459, CG42459, Dmel\\CG42459', description='Seminal fluid protein 23F', other_designations='CG42459-PA|Sfp23F-PA|seminal fluid protein 23F', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3387354, end_position_on_the_genomic_accession=3387743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674055, CurrentID=0, Status='live', Symbol='mgl', Aliases='Dmel_CG42611, CG12139, CG12654, CG15316, CG34339, CG34352, CG42611, CT7830, Dm CG34352, Dmel\\CG42611, Dmel_CG12139, Dmel_CG12654, Dmel_CG34339, Dmel_CG34352, LDLR, SOSIE, anon-WO0140519.108', description='Megalin', other_designations='CG42611-PA|CG42611-PC|CG42611-PD|CG42611-PE|LDL-receptor-related protein|LDLR-like|megalin|mgl-PA|mgl-PC|mgl-PD|mgl-PE', map_location='8E1-8E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9358496, end_position_on_the_genomic_accession=9500129, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  8674056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674056, CurrentID=0, Status='live', Symbol='CG42487', Aliases='Dmel_ Dmel\\CG42487', description='CG42487 gene product from transcript CG42487-RA', other_designations='CG42487-PA|uncharacterized protein', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27957356, end_position_on_the_genomic_accession=27959245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3346220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346220, CurrentID=0, Status='live', Symbol='CG33552', Aliases='Dmel_ Dmel\\CG33552', description='CG33552 gene product from transcript CG33552-RA', other_designations='CG33552-PA|uncharacterized protein', map_location='36A10-36A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16682503, end_position_on_the_genomic_accession=16683290, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674058, CurrentID=0, Status='live', Symbol='CG42481', Aliases='Dmel_ Dmel\\CG42481', description='CG42481 gene product from transcript CG42481-RB', other_designations='CG42481-PA|CG42481-PB|uncharacterized protein', map_location='70A4-70A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=13298797, end_position_on_the_genomic_accession=13299023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674059, CurrentID=0, Status='live', Symbol='Sfp24C1', Aliases='Dmel_CG42466, CG42466, Dmel\\CG42466', description='Seminal fluid protein 24C1', other_designations='CG42466-PA|Sfp24C1-PA|seminal fluid protein 24C1', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3700674, end_position_on_the_genomic_accession=3701092, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318220, CurrentID=0, Status='live', Symbol='CG32806', Aliases='Dmel_ Dmel\\CG32806, NEST:bs06b11', description='CG32806 gene product from transcript CG32806-RA', other_designations='CG32806-PA|uncharacterized protein', map_location='2B9-2B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1789566, end_position_on_the_genomic_accession=1790418, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318221, CurrentID=0, Status='live', Symbol='CG32808', Aliases='Dmel_ Dmel\\CG32808, EG:80H7.1', description='CG32808 gene product from transcript CG32808-RA', other_designations='CG32808-PA|uncharacterized protein', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1468693, end_position_on_the_genomic_accession=1469817, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674062, CurrentID=0, Status='live', Symbol='CG43339', Aliases='Dmel_ CG31791, CG42627, Dmel\\CG43339, Dmel_CG31791, Dmel_CG42627', description='CG43339 gene product from transcript CG43339-RC', other_designations='CG43339-PB|CG43339-PC|uncharacterized protein', map_location='36F10-36F10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18671604, end_position_on_the_genomic_accession=18673285, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3346221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346221, CurrentID=0, Status='live', Symbol='lectin-37Db', Aliases='Dmel_CG33533, AC 007082C, AC007082c, CG33533, CG9978, DL3, Dmel\\CG33533', description='CG33533 gene product from transcript CG33533-RA', other_designations='CG33533-PA|CG33533-PB|drosophila lectin 3|lectin|lectin-37Db-PA|lectin-37Db-PB', map_location='37D6-37D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19419075, end_position_on_the_genomic_accession=19419767, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674064, CurrentID=0, Status='live', Symbol='Sfp35C', Aliases='Dmel_CG42475, CG42475, Dmel\\CG42475', description='Seminal fluid protein 35C', other_designations='CG42475-PA|CG42475-PB|Sfp35C-PA|Sfp35C-PB|seminal fluid protein 35C', map_location='35C3-35C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15201613, end_position_on_the_genomic_accession=15202092, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318225, CurrentID=0, Status='live', Symbol='CG32816', Aliases='Dmel_ Dmel\\CG32816', description='CG32816 gene product from transcript CG32816-RC', other_designations='CG32816-PA|CG32816-PB|CG32816-PC|CG32816-PD|CG32816-PE|CG32816-PF|uncharacterized protein', map_location='1A2-1B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=316576, end_position_on_the_genomic_accession=470726, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  8674066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674066, CurrentID=0, Status='live', Symbol='CG42492', Aliases='Dmel_ CG11473, CG17761, Dmel\\CG42492, Dmel_CG11473, Dmel_CG17761, anon-WO0140519.181', description='CG42492 gene product from transcript CG42492-RE', other_designations='CG42492-PA|CG42492-PC|CG42492-PD|CG42492-PE|CG42492-PF|CG42492-PG|uncharacterized protein', map_location='5A9-5A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5625676, end_position_on_the_genomic_accession=5648487, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  318227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318227, CurrentID=0, Status='live', Symbol='CG32819', Aliases='Dmel_ Dmel\\CG32819', description='CG32819 gene product from transcript CG32819-RA', other_designations='CG32819-PA|CG32819-PB|tektin|uncharacterized protein', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21617889, end_position_on_the_genomic_accession=21621104, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318228, CurrentID=0, Status='live', Symbol='CG32820', Aliases='Dmel_ Dmel\\CG32820', description='CG32820 gene product from transcript CG32820-RA', other_designations='CG32820-PA|CG32820-PB|tektin|uncharacterized protein', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21611797, end_position_on_the_genomic_accession=21615239, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3346222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346222, CurrentID=0, Status='live', Symbol='lectin-37Da', Aliases='Dmel_CG33532, AC 007082B, AC007082b, CG33532, CG9978, DL2, Dmel\\CG33532', description='CG33532 gene product from transcript CG33532-RA', other_designations='CG33532-PA|CG33532-PB|drosophila lectin 2|lectin|lectin-37Da-PA|lectin-37Da-PB', map_location='37D6-37D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19418317, end_position_on_the_genomic_accession=19419029, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318231, CurrentID=0, Status='live', Symbol='Sdic3', Aliases='Dmel_CG32823, CG32823, Dmel\\CG32823, Sdic:CG32823', description='Sperm-specific dynein intermediate chain 3', other_designations='CG32823-PE|CG32823-PF|CG32823-PG|Sdic3-PE|Sdic3-PF|Sdic3-PG|Sdic:CG32823', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20189440, end_position_on_the_genomic_accession=20194227, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8674072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674072, CurrentID=0, Status='live', Symbol='CG42502', Aliases='Dmel_ Dmel\\CG42502, Dromel_CG10570_FBtr0081122_uORF', description='CG42502 gene product from transcript CG42502-RC', other_designations='CG42502-PA|CG42502-PB|CG42502-PC|uncharacterized protein', map_location='37A4-37A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18808694, end_position_on_the_genomic_accession=18811252, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674073, CurrentID=0, Status='live', Symbol='CG42554', Aliases='Dmel_ CG13886, CT33419, Dmel\\CG42554, Dmel_CG13886', description='CG42554 gene product from transcript CG42554-RA', other_designations='CG42554-PA|uncharacterized protein', map_location='61D2-61D2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=835683, end_position_on_the_genomic_accession=836770, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8674074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674074, CurrentID=0, Status='live', Symbol='CG42537', Aliases='Dmel_ BRG-A3, CG10284, Dmel\\CG42537', description='CG42537 gene product from transcript CG42537-RA', other_designations='CG42537-PA|uncharacterized protein', map_location='83E5-83E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6377797, end_position_on_the_genomic_accession=6378233, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3346223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346223, CurrentID=0, Status='live', Symbol='CG33511', Aliases='Dmel_ CG14398, Dmel\\CG33511', description='CG33511 gene product from transcript CG33511-RB', other_designations='CG33511-PB|uncharacterized protein', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21081127, end_position_on_the_genomic_accession=21082494, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674076, CurrentID=0, Status='live', Symbol='Sfp87B', Aliases='Dmel_CG42485, CG42485, Dmel\\CG42485, GH21617p', description='Seminal fluid protein 87B', other_designations='CG42485-PA|Sfp87B-PA|seminal fluid protein 87B', map_location='87B4-87B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12273126, end_position_on_the_genomic_accession=12273521, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674077, CurrentID=0, Status='live', Symbol='FDY', Aliases='Dmel_CG41503, CG41503, CR41503, Dmel\\CG41503, fdy', description='flagrante delicto Y', other_designations='CG41503-PB|CG41503-PC|FDY-PB|FDY-PC', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_024512.1', start_position_on_the_genomic_accession=3022289, end_position_on_the_genomic_accession=3024026, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318238, CurrentID=0, Status='live', Symbol='CG32834', Aliases='Dmel_ CG30331, CG9898, Dmel\\CG32834, SPH196', description='CG32834 gene product from transcript CG32834-RB', other_designations='CG32834-PB|uncharacterized protein', map_location='59C1-59C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22982542, end_position_on_the_genomic_accession=22984293, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674080, CurrentID=0, Status='live', Symbol='CG42512', Aliases='Dmel_ CG42528, Dmel\\CG42512, Dromel_CG32573_FBtr0074315_uORF', description='CG42512 gene product from transcript CG42512-RA', other_designations='CG42512-PA|CG42512-PB|uncharacterized protein', map_location='14F5-14F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16651376, end_position_on_the_genomic_accession=16653275, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3346224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346224, CurrentID=0, Status='live', Symbol='CG33510', Aliases='Dmel_ CG14398, Dmel\\CG33510', description='CG33510 gene product from transcript CG33510-RB', other_designations='CG33510-PB|uncharacterized protein', map_location='39A1-39A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21082643, end_position_on_the_genomic_accession=21084326, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674082, CurrentID=0, Status='live', Symbol='CG42523', Aliases='Dmel_ Dmel\\CG42523', description='CG42523 gene product from transcript CG42523-RB', other_designations='CG42523-PA|CG42523-PB|uncharacterized protein', map_location='25A4-25A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4762895, end_position_on_the_genomic_accession=4763951, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318243, CurrentID=0, Status='live', Symbol='CG32845', Aliases='Dmel_ Dmel\\CG32845', description='CG32845 gene product from transcript CG32845-RA', other_designations='CG32845-PA|uncharacterized protein', map_location='61B3-61B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=271420, end_position_on_the_genomic_accession=273391, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674084, CurrentID=0, Status='live', Symbol='CG42524', Aliases='Dmel_ Dmel\\CG42524', description='CG42524 gene product from transcript CG42524-RC', other_designations='CG42524-PA|CG42524-PB|CG42524-PC|uncharacterized protein', map_location='52B2-52B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15613854, end_position_on_the_genomic_accession=15657214, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  8674085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674085, CurrentID=0, Status='live', Symbol='Sfp33A1', Aliases='Dmel_CG42472, CG42472, Dmel\\CG42472', description='Seminal fluid protein 33A1', other_designations='CG42472-PA|Sfp33A1-PA|seminal fluid protein 33A1', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11754085, end_position_on_the_genomic_accession=11754502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318246, CurrentID=0, Status='live', Symbol='CG32850', Aliases='Dmel_ BEST:GH26713, BcDNA:RE41137, Dmel\\CG32850, RNF-11', description='CG32850 gene product from transcript CG32850-RB', other_designations='CG32850-PA|CG32850-PB|uncharacterized protein', map_location='102B3-102B5', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=314923, end_position_on_the_genomic_accession=331387, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  8674087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674087, CurrentID=0, Status='live', Symbol='CG42494', Aliases='Dmel_ Dmel\\CG42494', description='CG42494 gene product from transcript CG42494-RB', other_designations='CG42494-PA|CG42494-PB|CG42494-PC|uncharacterized protein', map_location='63B10-63B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3184404, end_position_on_the_genomic_accession=3186549, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  8674088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674088, CurrentID=0, Status='live', Symbol='CG42634', Aliases='Dmel_ CG15149, Dmel\\CG42634, Dmel_CG15149', description='CG42634 gene product from transcript CG42634-RA', other_designations='CG42634-PA|uncharacterized protein', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18007935, end_position_on_the_genomic_accession=18008962, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318249, CurrentID=0, Status='live', Symbol='mRpS21', Aliases='Dmel_CG32854, BcDNA:RE23482, CG32854, Dmel\\CG32854, MRP-S21, MrpS21', description='mitochondrial ribosomal protein S21', other_designations='CG32854-PA|CG32854-PB|mRpS21-PA|mRpS21-PB', map_location='87E8-87E8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13381276, end_position_on_the_genomic_accession=13381809, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318250, CurrentID=0, Status='live', Symbol='CG32855', Aliases='Dmel_ Dmel\\CG32855', description='CG32855 gene product from transcript CG32855-RA', other_designations='CG32855-PA|uncharacterized protein', map_location='89B2-89B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16078980, end_position_on_the_genomic_accession=16080992, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674091 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674091, CurrentID=0, Status='live', Symbol='CG42594', Aliases='Dmel_ CG15470, CG32770, CG32771, CG6952, Dmel\\CG42594, Dmel_CG15470, Dmel_CG32770, Dmel_CG32771', description='CG42594 gene product from transcript CG42594-RD', other_designations='CG42594-PA|CG42594-PC|CG42594-PD|uncharacterized protein', map_location='4D5-4D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4840833, end_position_on_the_genomic_accession=4906575, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  318252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318252, CurrentID=0, Status='live', Symbol='CG32857', Aliases='Dmel_ Dmel\\CG32857', description='CG32857 gene product from transcript CG32857-RA', other_designations='CG32857-PA|uncharacterized protein', map_location='20A4-20A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21615449, end_position_on_the_genomic_accession=21617617, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36658, CurrentID=0, Status='live', Symbol='Pcf11', Aliases='Dmel_CG10228, CG10228, Dmel\\CG10228, Inr-a, anon-WO0118547.146, dPcf11, l(2)k08015, pcf11', description='Protein 1 of cleavage and polyadenylation factor 1', other_designations='CG10228-PB|CG10228-PC|CG10228-PD|CG10228-PE|CG10228-PF|CG10228-PG|CG10228-PH|Pcf11-PB|Pcf11-PC|Pcf11-PD|Pcf11-PE|Pcf11-PF|Pcf11-PG|Pcf11-PH|inverse regulator a|lethal (2) k08015|protein 1 of cleavage and polyadenylation factor 1', map_location='44C1-50B9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14860936, end_position_on_the_genomic_accession=14871071, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  8674094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674094, CurrentID=0, Status='live', Symbol='Sfp24Bb', Aliases='Dmel_CG42462, CG42462, Dmel\\CG42462, SFP24BB', description='Seminal fluid protein 24Bb', other_designations='CG42462-PA|Sfp24Bb-PA|seminal fluid protein 24Bb', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3669261, end_position_on_the_genomic_accession=3669775, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674095, CurrentID=0, Status='live', Symbol='Sfp26Ac', Aliases='Dmel_CG42469, CG42469, Dmel\\CG42469', description='Seminal fluid protein 26Ac', other_designations='CG42469-PA|Sfp26Ac-PA|seminal fluid protein 26Ac', map_location='26A1-26A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5884350, end_position_on_the_genomic_accession=5884774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  8674097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674097, CurrentID=0, Status='live', Symbol='CG42525', Aliases='Dmel_ Dmel\\CG42525', description='CG42525 gene product from transcript CG42525-RB', other_designations='CG42525-PA|CG42525-PB|uncharacterized protein', map_location='63B10-63B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3188025, end_position_on_the_genomic_accession=3189747, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  8674098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674098, CurrentID=0, Status='live', Symbol='CanA-14F', Aliases='Dmel_CG9819, CG9819, CnnA14D, CnnA14D(I), D27, DmCanA-14F, Dmel\\CG9819, MRE12, PP2B, PpD27, canA-14F', description='Calcineurin A at 14F', other_designations='CG9819-PA|CG9819-PB|CG9819-PC|CG9819-PD|CanA-14F-PA|CanA-14F-PB|CanA-14F-PC|CanA-14F-PD|calcineurin|calcineurin A at 14F|mRNA-like ncRNA in embryogenesis 12|protein phosphatase D27|protein phosphatase from PCR fragment D27', map_location='14E3-14F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16560684, end_position_on_the_genomic_accession=16587172, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3772211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772211, CurrentID=0, Status='live', Symbol='His4:CG33877', Aliases='Dmel_CG33877, CG33877, Dmel\\CG33877', description='CG33877 gene product from transcript CG33877-RA', other_designations='CG33877-PA|His4:CG33877-PA|histone H4', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21514466, end_position_on_the_genomic_accession=21514878, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318260, CurrentID=0, Status='live', Symbol='CG32939', Aliases='Dmel_ Dmel\\CG32939', description='CG32939 gene product from transcript CG32939-RA', other_designations='CG32939-PA|uncharacterized protein', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9693079, end_position_on_the_genomic_accession=9694461, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31198, CurrentID=0, Status='live', Symbol='Mct1', Aliases='Dmel_CG3456, CG3456, Dmel\\CG3456, EG:103B4.3', description='Monocarboxylate transporter 1', other_designations='CG3456-PA|CG3456-PB|Mct1-PA|Mct1-PB|monocarboxylate transporter 1', map_location='2E1-2E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2189188, end_position_on_the_genomic_accession=2205373, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19835642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835642, CurrentID=0, Status='live', Symbol='CG45603', Aliases='Dmel_ Dmel\\CG45603', description='CG45603 gene product from transcript CG45603-RA', other_designations='CG45603-PA|uncharacterized protein', map_location='11A7-11A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12274635, end_position_on_the_genomic_accession=12275155, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674103, CurrentID=0, Status='live', Symbol='CG42583', Aliases='Dmel_ Dmel\\CG42583', description='CG42583 gene product from transcript CG42583-RA', other_designations='CG42583-PA|uncharacterized protein', map_location='19B3-19B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20139793, end_position_on_the_genomic_accession=20140170, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674104, CurrentID=0, Status='live', Symbol='CG42568', Aliases='Dmel_ CG12240, CG13571, Dmel\\CG42568', description='CG42568 gene product from transcript CG42568-RA', other_designations='CG42568-PA|CG42568-PB|uncharacterized protein', map_location='60C1-60C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24157668, end_position_on_the_genomic_accession=24159312, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8674106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674106, CurrentID=0, Status='live', Symbol='CG42578', Aliases='Dmel_ Dmel\\CG42578', description='CG42578 gene product from transcript CG42578-RA', other_designations='CG42578-PA|uncharacterized protein', map_location='19D1-19D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20446371, end_position_on_the_genomic_accession=20446703, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674107, CurrentID=0, Status='live', Symbol='CG42590', Aliases='Dmel_ Dmel\\CG42590', description='CG42590 gene product from transcript CG42590-RA', other_designations='CG42590-PA|uncharacterized protein', map_location='66B4-66B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8068413, end_position_on_the_genomic_accession=8069267, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  8674108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674108, CurrentID=0, Status='live', Symbol='CG42514', Aliases='Dmel_ AC, CG32158, CG4526, Dmel\\CG42514, Dmel_CG32158', description='CG42514 gene product from transcript CG42514-RB', other_designations='CG42514-PA|CG42514-PB|CG42514-PC|uncharacterized protein', map_location='73A3-73A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16491838, end_position_on_the_genomic_accession=16496236, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3772213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772213, CurrentID=0, Status='live', Symbol='Tim9b', Aliases='Dmel_CG17767, CG17767, CG33066, Dmel\\CG17767, Dromel_CG12788_CG17767_FBtr0074_tim9b_uORF, Tim9, tim9', description='Translocase of inner membrane 9b', other_designations='CG17767-PC|CG17767-PD|Tim9b-PC|Tim9b-PD|translocase of inner membrane 9b', map_location='18D3-18D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19616663, end_position_on_the_genomic_accession=19618537, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31200, CurrentID=0, Status='live', Symbol='msta', Aliases='Dmel_CG33548, CG18033, CG32799, CG32800, CG33548, Dmel\\CG33548, EG:103B4.4, EG:103B4.5, MSTA, Msta', description='CG33548 gene product from transcript CG33548-RD', other_designations='CG33548-PA|CG33548-PB|CG33548-PC|CG33548-PD|msta-PA|msta-PB|msta-PC|msta-PD', map_location='2E2-2E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2205515, end_position_on_the_genomic_accession=2213809, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  8674114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674114, CurrentID=0, Status='live', Symbol='inaF-B', Aliases='Dmel_CG42447, CG42447, Dmel\\CG42447, INAF, INAF-B, InaF, inaF', description='CG42447 gene product from transcript CG42447-RA', other_designations='CG42447-PA|inaF-B-PA', map_location='10D8-10E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11724668, end_position_on_the_genomic_accession=11732770, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  8674115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674115, CurrentID=0, Status='live', Symbol='Sfp33A3', Aliases='Dmel_CG42474, CG42474, Dmel\\CG42474', description='Seminal fluid protein 33A3', other_designations='CG42474-PA|Sfp33A3-PA|seminal fluid protein 33A3', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11839912, end_position_on_the_genomic_accession=11840329, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740122, CurrentID=0, Status='live', Symbol='CG34334', Aliases='Dmel_ CG9561, Dmel\\CG34334, cg9561', description='CG34334 gene product from transcript CG34334-RA', other_designations='CG34334-PA|uncharacterized protein', map_location='20F1-20F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22872812, end_position_on_the_genomic_accession=22875367, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8674117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674117, CurrentID=0, Status='live', Symbol='CG42464', Aliases='Dmel_ Dmel\\CG42464', description='CG42464 gene product from transcript CG42464-RA', other_designations='CG42464-PA|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3699464, end_position_on_the_genomic_accession=3699826, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326113, CurrentID=0, Status='live', Symbol='CG31016', Aliases='Dmel_ CG9708, Dmel\\CG31016', description='CG31016 gene product from transcript CG31016-RB', other_designations='CG31016-PB|CG31016-PC|uncharacterized protein', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30606189, end_position_on_the_genomic_accession=30610456, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  8674120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8674120, CurrentID=0, Status='live', Symbol='CG42566', Aliases='Dmel_ CG13512, Dmel\\CG42566, Dmel_CG13512', description='CG42566 gene product from transcript CG42566-RB', other_designations='CG42566-PA|CG42566-PB|uncharacterized protein', map_location='58F4-58F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22661934, end_position_on_the_genomic_accession=22663241, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  5740117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740117, CurrentID=0, Status='live', Symbol='CG41242', Aliases='Dmel_ Dmel\\CG41242', description='CG41242 gene product from transcript CG41242-RB', other_designations='CG41242-PB|uncharacterized protein', map_location='', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=3363857, end_position_on_the_genomic_accession=3364402, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31203, CurrentID=0, Status='live', Symbol='CG14052', Aliases='Dmel_ Dmel\\CG14052', description='CG14052 gene product from transcript CG14052-RB', other_designations='CG14052-PB|uncharacterized protein', map_location='2F1-2F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2223754, end_position_on_the_genomic_accession=2224969, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31204 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31204, CurrentID=0, Status='live', Symbol='pcx', Aliases='Dmel_CG3443, CG3443, DMPCX, Dmel\\CG3443, EG:30B8.4, fs(1)12-1012, fs(1)12-1743, fs(1)12-2135, fs(1)12-3014, fs(1)12-3102, fs(1)12-4169, fs(1)14-1153, fs(1)14-567, fs(1)A387, fs(1)M10, fs(1)pcx, mel(1)R, mel(1)R1', description='pecanex', other_designations='CG3443-PC|CG3443-PE|CG3443-PF|pcx-PC|pcx-PE|pcx-PF', map_location='1-0.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2225296, end_position_on_the_genomic_accession=2238045, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  3772218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772218, CurrentID=0, Status='live', Symbol='CG33649', Aliases='Dmel_ CG33084, CG33084-ORFA, CG8969, Dmel\\CG33649', description='CG33649 gene product from transcript CG33649-RA', other_designations='CG33649-PA|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13831482, end_position_on_the_genomic_accession=13833105, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326117 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326117, CurrentID=0, Status='live', Symbol='CG31099', Aliases='Dmel_ CG10551, Dmel\\CG31099', description='CG31099 gene product from transcript CG31099-RA', other_designations='CG31099-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25309429, end_position_on_the_genomic_accession=25310943, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36667, CurrentID=0, Status='live', Symbol='Cyp317a1', Aliases='Dmel_CG17453, 317a1, CG17453, Dmel\\CG17453', description='CG17453 gene product from transcript CG17453-RB', other_designations='CG17453-PA|CG17453-PB|CG17453-PC|Cyp317a1-PA|Cyp317a1-PB|Cyp317a1-PC', map_location='51D2-51D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14898312, end_position_on_the_genomic_accession=14903095, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326119, CurrentID=0, Status='live', Symbol='5PtaseI', Aliases='Dmel_CG42283, BcDNA:RE64196, CG31107, CG31110, CG42283, CG7613, Dmel\\CG42283, Dmel_CG31107, Dmel_CG31110, IPP, dm', description='CG42283 gene product from transcript CG42283-RB', other_designations='5PtaseI-PB|5PtaseI-PC|5PtaseI-PF|CG42283-PB|CG42283-PC|CG42283-PF', map_location='96B17-96B18', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25040427, end_position_on_the_genomic_accession=25045151, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3346237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346237, CurrentID=0, Status='live', Symbol='nab', Aliases='Dmel_CG33545, CG15000, CG15001, CG33545, Dmel\\CG33545, NAB, Nab, dNAB, d l(2)SH3 143, l(3)64Aj, l(3)SH143, l(3)SH21', description='CG33545 gene product from transcript CG33545-RC', other_designations='CG33545-PB|CG33545-PC|CG33545-PD|lethal (3) SH143|nab-PB|nab-PC|nab-PD', map_location='64A8-64A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4153618, end_position_on_the_genomic_accession=4160932, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31208, CurrentID=0, Status='live', Symbol='crn', Aliases='Dmel_CG3193, 62D9.a, CG18842, CG3193, Crn, Dmel\\CG3193, EG:30B8.1, N10, N5, RC63, l(1)2, l(1)2Fa, l(1)CRN, l(1)N7, l(1)crn, l30, yok', description='crooked neck', other_designations='CG3193-PA|crn-PA|yolky', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2247594, end_position_on_the_genomic_accession=2249983, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3346238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346238, CurrentID=0, Status='live', Symbol='form3', Aliases='Dmel_CG33556, CG14824, CG18138, CG32384, CG33556, CG8633, Dmel\\CG33556, Form3, INF2, ah1644, formin/DIA-like', description='formin 3', other_designations='CG33556-PA|CG33556-PB|form3-PA|form3-PB|formin3', map_location='65E3-65E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7094818, end_position_on_the_genomic_accession=7123297, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  326121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326121, CurrentID=0, Status='live', Symbol='CG31126', Aliases='Dmel_ CG13621, Dmel\\CG31126', description='CG31126 gene product from transcript CG31126-RA', other_designations='CG31126-PA|uncharacterized protein', map_location='96A14-96A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24659621, end_position_on_the_genomic_accession=24660265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36671, CurrentID=0, Status='live', Symbol='CG10257', Aliases='Dmel_ BcDNA:RH51312, Dmel\\CG10257', description='CG10257 gene product from transcript CG10257-RA', other_designations='CG10257-PA|uncharacterized protein', map_location='51D5-51D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14942398, end_position_on_the_genomic_accession=14944469, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31210 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31210, CurrentID=0, Status='live', Symbol='CG3091', Aliases='Dmel_ Dmel\\CG3091', description='CG3091 gene product from transcript CG3091-RC', other_designations='CG3091-PA|CG3091-PC', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2252332, end_position_on_the_genomic_accession=2254197, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326123, CurrentID=0, Status='live', Symbol='CG31161', Aliases='Dmel_ CG6943, Dmel\\CG31161', description='CG31161 gene product from transcript CG31161-RB', other_designations='CG31161-PB|uncharacterized protein', map_location='94B6-94B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22616511, end_position_on_the_genomic_accession=22618300, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772225, CurrentID=0, Status='live', Symbol='His1:CG33840', Aliases='Dmel_CG33840, CG33840, Dmel\\CG33840', description='CG33840 gene product from transcript CG33840-RA', other_designations='CG33840-PA|His1:CG33840-PA|histone H1', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21496586, end_position_on_the_genomic_accession=21497608, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326124, CurrentID=0, Status='live', Symbol='CG31200', Aliases='Dmel_ CG17837, Dmel\\CG31200, SPH169', description='CG31200 gene product from transcript CG31200-RA', other_designations='CG31200-PA|uncharacterized protein', map_location='92F10-92F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20744234, end_position_on_the_genomic_accession=20745338, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  42135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42135, CurrentID=0, Status='live', Symbol='TyrRII', Aliases='Dmel_CG16766, CG16766, DmCG16766, Dmel\\CG16766, anon-WO0170980.124, anon-WO0170980.125', description='Tyramine receptor II', other_designations='CG16766-PA|CG16766-PB|CG16766-PC|CG16766-PD|TyrRII-PA|TyrRII-PB|TyrRII-PC|TyrRII-PD|tyramine receptor II', map_location='90C2-90C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17736611, end_position_on_the_genomic_accession=17747470, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  326125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326125, CurrentID=0, Status='live', Symbol='CG31207', Aliases='Dmel_ CG7096, CG7096-N, Dmel\\CG31207', description='CG31207 gene product from transcript CG31207-RA', other_designations='CG31207-PA|uncharacterized protein', map_location='93B7-93B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21046032, end_position_on_the_genomic_accession=21047027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326126, CurrentID=0, Status='live', Symbol='CG31220', Aliases='Dmel_ CG16731, Dmel\\CG31220, SP6, c-SP6', description='CG31220 gene product from transcript CG31220-RB', other_designations='CG31220-PB|uncharacterized protein', map_location='92A12-92A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19648487, end_position_on_the_genomic_accession=19650586, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326127, CurrentID=0, Status='live', Symbol='CG31223', Aliases='Dmel_ CG15694, Dmel\\CG31223', description='CG31223 gene product from transcript CG31223-RA', other_designations='CG31223-PA|uncharacterized protein', map_location='93A1-93A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20831127, end_position_on_the_genomic_accession=20832455, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326129, CurrentID=0, Status='live', Symbol='CG31244', Aliases='Dmel_ CG18491, Dmel\\CG31244', description='CG31244 gene product from transcript CG31244-RA', other_designations='CG31244-PA|uncharacterized protein', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19397430, end_position_on_the_genomic_accession=19398567, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772231 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772231, CurrentID=0, Status='live', Symbol='His3:CG33857', Aliases='Dmel_CG33857, CG33857, Dmel\\CG33857', description='CG33857 gene product from transcript CG33857-RA', other_designations='CG33857-PA|His3:CG33857-PA|histone H3', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21524952, end_position_on_the_genomic_accession=21525480, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31218, CurrentID=0, Status='live', Symbol='CG14050', Aliases='Dmel_ Dmel\\CG14050, EG:BACH48C10.1, H48C10.1', description='CG14050 gene product from transcript CG14050-RB', other_designations='CG14050-PA|CG14050-PB|uncharacterized protein', map_location='2F6-2F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2324748, end_position_on_the_genomic_accession=2326217, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36680, CurrentID=0, Status='live', Symbol='CG7544', Aliases='Dmel_ Dmel\\CG7544', description='CG7544 gene product from transcript CG7544-RB', other_designations='CG7544-PA|CG7544-PB', map_location='51E2-51E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15147860, end_position_on_the_genomic_accession=15149709, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31219, CurrentID=0, Status='live', Symbol='CG2854', Aliases='Dmel_ Dmel\\CG2854, EG:BACH48C10.2', description='CG2854 gene product from transcript CG2854-RA', other_designations='CG2854-PA', map_location='2F6-3A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2327128, end_position_on_the_genomic_accession=2328941, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326132, CurrentID=0, Status='live', Symbol='CG31300', Aliases='Dmel_ CG13657, Dmel\\CG31300', description='CG31300 gene product from transcript CG31300-RA', other_designations='CG31300-PA|uncharacterized protein', map_location='96C9-96C9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25283620, end_position_on_the_genomic_accession=25285224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740126, CurrentID=0, Status='live', Symbol='CG34287', Aliases='Dmel_ Dmel\\CG34287', description='CG34287 gene product from transcript CG34287-RA', other_designations='CG34287-PA|uncharacterized protein', map_location='83C5-83C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5853332, end_position_on_the_genomic_accession=5854282, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326133, CurrentID=0, Status='live', Symbol='CG31343', Aliases='Dmel_ BEST:GH20248, CG5839, Dmel\\CG31343, anon-WO0140519.86', description='CG31343 gene product from transcript CG31343-RA', other_designations='CG31343-PA|uncharacterized protein', map_location='93F8-93F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21745266, end_position_on_the_genomic_accession=21749003, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  326136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326136, CurrentID=0, Status='live', Symbol='Npc2e', Aliases='Dmel_CG31410, CG31410, CG3928, Dmel\\CG31410, npc2e', description='Niemann-Pick type C-2e', other_designations='CG31410-PB|Npc2e-PB', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10053332, end_position_on_the_genomic_accession=10054348, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  36687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36687, CurrentID=0, Status='live', Symbol='CG11808', Aliases='Dmel_ Dmel\\CG11808', description='CG11808 gene product from transcript CG11808-RA', other_designations='CG11808-PA|uncharacterized protein', map_location='51E7-51E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15216207, end_position_on_the_genomic_accession=15217070, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31226, CurrentID=0, Status='live', Symbol='CG12496', Aliases='Dmel_ Dmel\\CG12496, EG:BACH7M4.4', description='CG12496 gene product from transcript CG12496-RA', other_designations='CG12496-PA|uncharacterized protein', map_location='3A2-3A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2409365, end_position_on_the_genomic_accession=2410901, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326139, CurrentID=0, Status='live', Symbol='CG31441', Aliases='Dmel_ CG5135, Dmel\\CG31441', description='CG31441 gene product from transcript CG31441-RB', other_designations='CG31441-PA|CG31441-PB|uncharacterized protein', map_location='86D8-86D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11236286, end_position_on_the_genomic_accession=11237624, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326140, CurrentID=0, Status='live', Symbol='CG31445', Aliases='Dmel_ CG11955, Dmel\\CG31445', description='CG31445 gene product from transcript CG31445-RB', other_designations='CG31445-PA|CG31445-PB|uncharacterized protein', map_location='99A1-99A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29240580, end_position_on_the_genomic_accession=29244827, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  326141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326141, CurrentID=0, Status='live', Symbol='mRpS18A', Aliases='Dmel_CG31450, BcDNA:SD13983, CG11744, CG31450, Dmel\\CG31450, MRP-S18-(3), mRpS18c', description='mitochondrial ribosomal protein S18A', other_designations='CG31450-PA|mRpS18A-PA|mitochondrial ribosomal protein S18c', map_location='85A3-85A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8561917, end_position_on_the_genomic_accession=8562593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326142, CurrentID=0, Status='live', Symbol='CG31457', Aliases='Dmel_ CG4639, Dmel\\CG31457', description='CG31457 gene product from transcript CG31457-RB', other_designations='CG31457-PB|uncharacterized protein', map_location='94E1-94E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23126581, end_position_on_the_genomic_accession=23127981, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14462622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462622, CurrentID=0, Status='live', Symbol='CG43742', Aliases='Dmel_ Dmel\\CG43742', description='CG43742 gene product from transcript CG43742-RB', other_designations='CG43742-PB|uncharacterized protein', map_location='58A4-58A4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21935002, end_position_on_the_genomic_accession=21936863, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36695, CurrentID=0, Status='live', Symbol='ND-51L2', Aliases='Dmel_CG8102, CG8102, Dmel\\CG8102, NUBM', description='NADH dehydrogenase (ubiquinone) 51 kDa subunit-like 2', other_designations='CG8102-PA|CG8102-PB|NADH-UBIQUINONE OXIDOREDUCTASE 51 KDA SUBUNIT|NADH:ubiquinone oxidoreductase 51 kDa subunit|ND-51L2-PA|ND-51L2-PB', map_location='51F3-51F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15296764, end_position_on_the_genomic_accession=15298329, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326146, CurrentID=0, Status='live', Symbol='CG31510', Aliases='Dmel_ CG10057, Dmel\\CG31510', description='CG31510 gene product from transcript CG31510-RA', other_designations='CG31510-PA|uncharacterized protein', map_location='96C1-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25117218, end_position_on_the_genomic_accession=25122920, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772248, CurrentID=0, Status='live', Symbol='His2B:CG33898', Aliases='Dmel_CG33898, CG33898, Dmel\\CG33898', description='CG33898 gene product from transcript CG33898-RA', other_designations='CG33898-PA|His2B:CG33898-PA|histone H2B', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21462332, end_position_on_the_genomic_accession=21462926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326147, CurrentID=0, Status='live', Symbol='CG31515', Aliases='Dmel_ Dmel\\CG31515', description='CG31515 gene product from transcript CG31515-RB', other_designations='CG31515-PB|uncharacterized protein', map_location='94F1-94F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23388373, end_position_on_the_genomic_accession=23388898, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326148, CurrentID=0, Status='live', Symbol='CG31523', Aliases='Dmel_ CG9798, Dmel\\CG31523', description='CG31523 gene product from transcript CG31523-RB', other_designations='CG31523-PA|CG31523-PB|CG31523-PC|CG31523-PD|CG31523-PE|CG31523-PF|CG31523-PG|uncharacterized protein', map_location='82B2-82B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4465421, end_position_on_the_genomic_accession=4479207, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14462623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462623, CurrentID=0, Status='live', Symbol='CG44163', Aliases='Dmel_ Dmel\\CG44163', description='CG44163 gene product from transcript CG44163-RA', other_designations='CG44163-PA|uncharacterized protein', map_location='58B5-58B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22090795, end_position_on_the_genomic_accession=22091395, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326149, CurrentID=0, Status='live', Symbol='CG31642', Aliases='Dmel_ CG5981, Dmel\\CG31642', description='CG31642 gene product from transcript CG31642-RA', other_designations='CG31642-PA|uncharacterized protein', map_location='26B7-26B7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6118512, end_position_on_the_genomic_accession=6120522, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31239, CurrentID=0, Status='live', Symbol='egh', Aliases='Dmel_CG9659, 4ManT, CG9659, Dmel\\CG9659, EG:BACR25B3.8, Egh, ZW-4, beta1, l(1)3Ae, l(1)zw4, zw-4, zw4', description='egghead', other_designations='CG9659-PB|CG9659-PC|CG9659-PD|egh-PB|egh-PC|egh-PD|zeste-white 4', map_location='1-1.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2588503, end_position_on_the_genomic_accession=2599111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19835348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835348, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-P', Aliases='Dmel_CG44881, CG44881, Dmel\\CG44881', description='CG44881 gene product from transcript CG44881-RA', other_designations='CG44881-PA|CG44881-PB|pre-mod(mdg4)-P-PA|pre-mod(mdg4)-P-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21355692, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31240, CurrentID=0, Status='live', Symbol='Klp3A', Aliases='Dmel_CG8590, 3A7Kin, BcDNA:LD21815, CG8590, DmKLP3A, DmKlp3A, Dmel\\CG8590, EG:BACR25B3.9, KIF 4A, KIF4, KIF4A, KLP-3A, KLP3A, Klp 3A, Klp3a, fs(1)M4, klp3A, klp3a, mei-352', description='Kinesin-like protein at 3A', other_designations='CG8590-PA|CG8590-PB|Kinesin-Like-Protein-at-3A|Klp3A-PA|Klp3A-PB|female sterile(1)M4|meiotic 352', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2599158, end_position_on_the_genomic_accession=2604052, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3772254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772254, CurrentID=0, Status='live', Symbol='His4:CG33903', Aliases='Dmel_CG33903, CG33903, Dmel\\CG33903', description='CG33903 gene product from transcript CG33903-RA', other_designations='CG33903-PA|His4:CG33903-PA|histone H4', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21439589, end_position_on_the_genomic_accession=21440001, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740555, CurrentID=0, Status='live', Symbol='CG34293', Aliases='Dmel_ Dmel\\CG34293', description='CG34293 gene product from transcript CG34293-RB', other_designations='CG34293-PA|CG34293-PB|uncharacterized protein', map_location='97E6-97E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27164708, end_position_on_the_genomic_accession=27165575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326155, CurrentID=0, Status='live', Symbol='CG31739', Aliases='Dmel_ CG17938, Dmel\\CG31739', description='CG31739 gene product from transcript CG31739-RA', other_designations='CG31739-PA|CG31739-PB|CG31739-PC|uncharacterized protein', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16827078, end_position_on_the_genomic_accession=16832093, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3772257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772257, CurrentID=0, Status='live', Symbol='CG33626', Aliases='Dmel_ Dmel\\CG33626', description='CG33626 gene product from transcript CG33626-RB', other_designations='CG33626-PB|uncharacterized protein', map_location='49A3-49A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12410079, end_position_on_the_genomic_accession=12410696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326156, CurrentID=0, Status='live', Symbol='CG31743', Aliases='Dmel_ 152851_at, CG4826, Dmel\\CG31743, dC4ST', description='CG31743 gene product from transcript CG31743-RA', other_designations='CG31743-PA|CG31743-PB|uncharacterized protein', map_location='36B3-36B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16879517, end_position_on_the_genomic_accession=16886319, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  326157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326157, CurrentID=0, Status='live', Symbol='mon2', Aliases='Dmel_CG8683, CG14276, CG31756, CG8683, Dmel\\CG8683, Mon2, noan', description='CG8683 gene product from transcript CG8683-RB', other_designations='CG8683-PB|CG8683-PC|mon2-PB|mon2-PC|no anchor', map_location='28E6-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8116965, end_position_on_the_genomic_accession=8124010, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  5740131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740131, CurrentID=0, Status='live', Symbol='CG34383', Aliases='Dmel_ CG14365, CG14366, CG9322, Dm_3R:52023, Dmel\\CG34383, Dmel_CG14365, Dmel_CG14366, Dmel_CG9322', description='CG34383 gene product from transcript CG34383-RJ', other_designations='CG34383-PB|CG34383-PD|CG34383-PE|CG34383-PF|CG34383-PG|CG34383-PH|CG34383-PI|CG34383-PJ|CG34383-PK|CG34383-PL|uncharacterized protein', map_location='87F13-87F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13715207, end_position_on_the_genomic_accession=13764667, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  326158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326158, CurrentID=0, Status='live', Symbol='CG31759', Aliases='Dmel_ 3635, BcDNA:AT13596, BcDNA:LD05405, Dmel\\CG31759, d3635, ms(2)01284', description='CG31759 gene product from transcript CG31759-RD', other_designations='CG31759-PC|CG31759-PD|uncharacterized protein', map_location='33D3-33D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12207709, end_position_on_the_genomic_accession=12210110, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318552, CurrentID=0, Status='live', Symbol='gskt', Aliases='Dmel_CG31003, BEST:GH16447, BcDNA:AT21229, CG11338, CG31003, CT4237, Dm Gskt, Dmel\\CG31003, GSK3b, Gsk3b, Gskt, NEST:bs22e06, NGSK, mjl', description='gasket', other_designations='CG31003-PA|gskt-PA|mojoless', map_location='100C2-100C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31372427, end_position_on_the_genomic_accession=31374198, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318553, CurrentID=0, Status='live', Symbol='CG31007', Aliases='Dmel_ Dmel\\CG31007', description='CG31007 gene product from transcript CG31007-RA', other_designations='CG31007-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30991700, end_position_on_the_genomic_accession=30992457, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318554, CurrentID=0, Status='live', Symbol='CG31008', Aliases='Dmel_ BEST:GH26101, Dmel\\CG31008, bs08f11.y1', description='CG31008 gene product from transcript CG31008-RA', other_designations='CG31008-PA|uncharacterized protein', map_location='100B1-100B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30989478, end_position_on_the_genomic_accession=30990006, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318555, CurrentID=0, Status='live', Symbol='CG31010', Aliases='Dmel_ BcDNA:AT12704, Dmel\\CG31010, NEST:bs21b05', description='CG31010 gene product from transcript CG31010-RA', other_designations='CG31010-PA|CG31010-PB|uncharacterized protein', map_location='100A2-100A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30704762, end_position_on_the_genomic_accession=30706121, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318558, CurrentID=0, Status='live', Symbol='CG31019', Aliases='Dmel_ Dmel\\CG31019', description='CG31019 gene product from transcript CG31019-RA', other_designations='CG31019-PA|CG31019-PB|uncharacterized protein', map_location='99F11-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30597004, end_position_on_the_genomic_accession=30599296, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318559 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318559, CurrentID=0, Status='live', Symbol='spdo', Aliases='Dmel_CG31020, 0970/02, BcDNA:RE23355, CG31020, Dmel\\CG31020, Spdo', description='sanpodo', other_designations='CG31020-PA|spdo-PA', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30476684, end_position_on_the_genomic_accession=30478849, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31248, CurrentID=0, Status='live', Symbol='sgg', Aliases='Dmel_CG2621, CG2621, DMSGG3, DMZ3K25Z, Dm Zw3, Dmel\\CG2621, Dmsgg3, EG:155E2.3, EG:BACR7C10.8, GSK, GSK-3, GSK-3B, GSK-3beta, GSK-3betaSgg, GSK3, GSK3-beta, GSK3A, GSK3Beta, GSK3beta, GSK3beta/SGG, GSK3betagr, GSKbeta, Gsk-3, Gsk-3beta, Gsk3, SGG, SGG/GSK3b, Sgg, Sgg/GSK-3b, Sgg/Zw3/GSK3, ZW-3, ZW3, Zw-3, Zw3, Zw3-sh, Zw3/GSK3, Zw3[sgg], anon-WO03040301.254, gsk3, l(1)3Ba, l(1)G0055, l(1)G0183, l(1)G0263, l(1)G0335, l(1)zw3-zw3, sgg/zw3, sgg10, sgg39, sgg46, zw-3, zw3, zw3(sgg), zw3/GSK3beta, zw3/sgg, zw3[sgg], zw[3], sgg', description='shaggy', other_designations='CG2621-PA|CG2621-PB|CG2621-PC|CG2621-PD|CG2621-PE|CG2621-PF|CG2621-PG|CG2621-PH|CG2621-PI|CG2621-PJ|CG2621-PK|CG2621-PM|CG2621-PN|CG2621-PO|CG2621-PP|CG2621-PQ|CG2621-PR|GSK-3 kinase|GSK3betagr|Shaggy/GSK-3beta|Shaggy/Glycogen synthase kinase 3|Shaggy/Zeste-white-3/Glycogen synthase kinase 3beta|Zeste-white3|Zw3 kinase|glycogen synthase Kinase-3|glycogen synthase kinase 3|glycogen synthase kinase 3Beta|glycogen synthase kinase-3beta|saggy|sgg-PA|sgg-PB|sgg-PC|sgg-PD|sgg-PE|sgg-PF|sgg-PG|sgg-PH|sgg-PI|sgg-PJ|sgg-PK|sgg-PM|sgg-PN|sgg-PO|sgg-PP|sgg-PQ|sgg-PR|shaggy-zeste white 3|shaggy/zeste-white 3|zest-white 3|zeste white 3|zeste white-3|zeste white3|zeste-white 3|zestewhite-3', map_location='1-1.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2633952, end_position_on_the_genomic_accession=2679553, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  318562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318562, CurrentID=0, Status='live', Symbol='CG31029', Aliases='Dmel_ BcDNA:AT12234, CG2164, Dmel\\CG31029', description='CG31029 gene product from transcript CG31029-RA', other_designations='CG31029-PA|uncharacterized protein', map_location='99D8-99D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30129279, end_position_on_the_genomic_accession=30132395, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318563, CurrentID=0, Status='live', Symbol='CG31030', Aliases='Dmel_ BcDNA:RH09162, CG2176, Dmel\\CG31030, anon-EST:Posey81', description='CG31030 gene product from transcript CG31030-RB', other_designations='CG31030-PA|CG31030-PB|uncharacterized protein', map_location='99E1-99E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30164692, end_position_on_the_genomic_accession=30166419, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318566, CurrentID=0, Status='live', Symbol='CG34133', Aliases='Dmel_ CG31035, CG7814, Dmel\\CG34133, anon-WO0118547.417', description='CG34133 gene product from transcript CG34133-RC', other_designations='CG34133-PB|CG34133-PC|CG34133-PE|uncharacterized protein', map_location='99C5-99C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29869365, end_position_on_the_genomic_accession=29875159, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  326161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326161, CurrentID=0, Status='live', Symbol='CG31793', Aliases='Dmel_ CG10441, CG17338, Dmel\\CG31793', description='CG31793 gene product from transcript CG31793-RC', other_designations='CG31793-PA|CG31793-PC|uncharacterized protein', map_location='37B9-37B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19003367, end_position_on_the_genomic_accession=19008234, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318571, CurrentID=0, Status='live', Symbol='CG31050', Aliases='Dmel_ Dmel\\CG31050, NEST:bs32b08, bs32b08.y1', description='CG31050 gene product from transcript CG31050-RB', other_designations='CG31050-PA|CG31050-PB|uncharacterized protein', map_location='98E1-98E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28649909, end_position_on_the_genomic_accession=28650480, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318572, CurrentID=0, Status='live', Symbol='CG31053', Aliases='Dmel_ Dmel\\CG31053', description='CG31053 gene product from transcript CG31053-RA', other_designations='CG31053-PA|uncharacterized protein', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27927357, end_position_on_the_genomic_accession=27928174, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318574, CurrentID=0, Status='live', Symbol='CG31055', Aliases='Dmel_ BcDNA:AT11054, BcDNA:AT22836, Dmel\\CG31055, NEST:bs01f03, NEST:bs23c05, bs23c05.y1', description='CG31055 gene product from transcript CG31055-RB', other_designations='CG31055-PA|CG31055-PB|CG31055-PC|uncharacterized protein', map_location='98B1-98B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27685632, end_position_on_the_genomic_accession=27688134, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318575, CurrentID=0, Status='live', Symbol='ppk31', Aliases='Dmel_CG31065, CG31065, Dmel\\CG31065', description='pickpocket 31', other_designations='CG31065-PB|ppk31-PB', map_location='97E5-97E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27163124, end_position_on_the_genomic_accession=27165131, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318578, CurrentID=0, Status='live', Symbol='CG31068', Aliases='Dmel_ BcDNA:AT10318, Dmel\\CG31068', description='CG31068 gene product from transcript CG31068-RC', other_designations='CG31068-PB|CG31068-PC|uncharacterized protein', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27233698, end_position_on_the_genomic_accession=27236360, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318579, CurrentID=0, Status='live', Symbol='spn-D', Aliases='Dmel_CG31069, BcDNA:RE19845, CG31069, Dmel\\CG31069, Spn-D, rad51C, spnD, spndD', description='spindle D', other_designations='CG31069-PA|spn-D-PA', map_location='3-91 cM', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27165716, end_position_on_the_genomic_accession=27167122, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318582, CurrentID=0, Status='live', Symbol='CG31076', Aliases='Dmel_ CG6309, CT19744, Dmel\\CG31076', description='CG31076 gene product from transcript CG31076-RB', other_designations='CG31076-PB|uncharacterized protein', map_location='97D12-97D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26989035, end_position_on_the_genomic_accession=26989852, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318583 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318583, CurrentID=0, Status='live', Symbol='CG31077', Aliases='Dmel_ Dmel\\CG31077', description='CG31077 gene product from transcript CG31077-RB', other_designations='CG31077-PB|uncharacterized protein', map_location='97D12-97D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26999752, end_position_on_the_genomic_accession=27002961, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326164, CurrentID=0, Status='live', Symbol='CG31826', Aliases='Dmel_ BG:DS09217.6, CG13241, Dmel\\CG31826', description='CG31826 gene product from transcript CG31826-RC', other_designations='CG31826-PB|CG31826-PC|uncharacterized protein', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15771263, end_position_on_the_genomic_accession=15773113, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318586, CurrentID=0, Status='live', Symbol='TwdlR', Aliases='Dmel_CG31081, CG14240, CG14241, CG31081, Dmel\\CG31081', description='TweedleR', other_designations='CG31081-PA|TwdlR-PA', map_location='97C2-97C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26634885, end_position_on_the_genomic_accession=26635998, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318590, CurrentID=0, Status='live', Symbol='CG31088', Aliases='Dmel_ Dmel\\CG31088', description='CG31088 gene product from transcript CG31088-RA', other_designations='CG31088-PA|uncharacterized protein', map_location='89A8-89A8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15950201, end_position_on_the_genomic_accession=15951088, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318591, CurrentID=0, Status='live', Symbol='CG31091', Aliases='Dmel_ CG5990, Dmel\\CG31091', description='CG31091 gene product from transcript CG31091-RC', other_designations='CG31091-PB|CG31091-PC|uncharacterized protein', map_location='96F7-96F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25899903, end_position_on_the_genomic_accession=25902172, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318593, CurrentID=0, Status='live', Symbol='ppk22', Aliases='Dmel_CG31105, CG31105, Dmel\\CG31105, FBgn0051105', description='pickpocket 22', other_designations='CG31105-PB|CG31105-PC|ppk22-PB|ppk22-PC', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24857585, end_position_on_the_genomic_accession=24859801, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  318594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318594, CurrentID=0, Status='live', Symbol='CG31106', Aliases='Dmel_ CG13655, Dmel\\CG31106', description='CG31106 gene product from transcript CG31106-RB', other_designations='CG31106-PB|uncharacterized protein', map_location='96C5-96C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25216297, end_position_on_the_genomic_accession=25218708, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  318595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318595, CurrentID=0, Status='live', Symbol='CG31111', Aliases='Dmel_ CG13649, Dmel\\CG31111', description='CG31111 gene product from transcript CG31111-RA', other_designations='CG31111-PA|uncharacterized protein', map_location='96B19-96B19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25055076, end_position_on_the_genomic_accession=25056177, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31254, CurrentID=0, Status='live', Symbol='CG2662', Aliases='Dmel_ Dmel\\CG2662, EG:100G10.6, MAD, snuts', description='CG2662 gene product from transcript CG2662-RB', other_designations='CG2662-PA|CG2662-PB|shrunken nuts', map_location='3B2-3B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2698701, end_position_on_the_genomic_accession=2701033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318599, CurrentID=0, Status='live', Symbol='CG31125', Aliases='Dmel_ BcDNA:SD02419, Dmel\\CG31125', description='CG31125 gene product from transcript CG31125-RA', other_designations='CG31125-PA|uncharacterized protein', map_location='96A13-96A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24653371, end_position_on_the_genomic_accession=24654820, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318600, CurrentID=0, Status='live', Symbol='Wsck', Aliases='Dmel_CG31127, BcDNA:LD25626, BcDNA:SD05152, CG31127, Dmel\\CG31127, WSCK', description='CG31127 gene product from transcript CG31127-RA', other_designations='CG31127-PA|Wsck-PA|wsck', map_location='96A13-96A13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24627066, end_position_on_the_genomic_accession=24630075, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318601, CurrentID=0, Status='live', Symbol='CG31128', Aliases='Dmel_ BcDNA:AT04449, Dmel\\CG31128', description='CG31128 gene product from transcript CG31128-RA', other_designations='CG31128-PA|uncharacterized protein', map_location='96A2-96A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24469397, end_position_on_the_genomic_accession=24470310, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326167, CurrentID=0, Status='live', Symbol='ZnT33D', Aliases='Dmel_CG31860, CG17215, CG31860, Dmel\\CG31860, cg31860, dZnT33D', description='Zinc transporter 33D', other_designations='CG31860-PB|CG31860-PC|ZnT33D-PB|ZnT33D-PC|zinc transporter 33D', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12164241, end_position_on_the_genomic_accession=12166704, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318604, CurrentID=0, Status='live', Symbol='Slimp', Aliases='Dmel_CG31133, BcDNA:LD24627, CG31133, Dmel\\CG31133, SLIMP', description='Seryl-tRNA synthetase-like insect mitochondrial protein', other_designations='CG31133-PA|Slimp-PA', map_location='95E5-95E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24147436, end_position_on_the_genomic_accession=24148967, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318605, CurrentID=0, Status='live', Symbol='CG31141', Aliases='Dmel_ Dmel\\CG31141', description='CG31141 gene product from transcript CG31141-RA', other_designations='CG31141-PA|uncharacterized protein', map_location='95C12-95C12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23923300, end_position_on_the_genomic_accession=23924319, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318608, CurrentID=0, Status='live', Symbol='CG31157', Aliases='Dmel_ Dmel\\CG31157', description='CG31157 gene product from transcript CG31157-RD', other_designations='CG31157-PC|CG31157-PD|uncharacterized protein', map_location='87D10-87D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13047731, end_position_on_the_genomic_accession=13049294, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31256, CurrentID=0, Status='live', Symbol='eIF2B-beta', Aliases='Dmel_CG2677, CG2677, DMC100G10.3, Dmel\\CG2677, EG:100G10.3, EIF2b, eIF2Bbeta', description='CG2677 gene product from transcript CG2677-RA', other_designations='CG2677-PA|eIF2B-beta-PA', map_location='3B2-3B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2704571, end_position_on_the_genomic_accession=2706217, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318610, CurrentID=0, Status='live', Symbol='Ir94a', Aliases='Dmel_CG31164, CG31164, DmelIR94a, Dmel\\CG31164, IR94a', description='Ionotropic receptor 94a', other_designations='CG31164-PA|Ir94a-PA|ionotropic receptor 94a', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22113427, end_position_on_the_genomic_accession=22115214, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326169, CurrentID=0, Status='live', Symbol='CG31883', Aliases='Dmel_ CG17022, Dmel\\CG31883, anon-WO0140519.218', description='CG31883 gene product from transcript CG31883-RA', other_designations='CG31883-PA|uncharacterized protein', map_location='30B10-30B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9528515, end_position_on_the_genomic_accession=9529648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318616, CurrentID=0, Status='live', Symbol='CG31174', Aliases='Dmel_ BcDNA:SD11171, Dmel\\CG31174', description='CG31174 gene product from transcript CG31174-RB', other_designations='CG31174-PA|CG31174-PB|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21868681, end_position_on_the_genomic_accession=21869514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36719, CurrentID=0, Status='live', Symbol='CG30467', Aliases='Dmel_ CG8185, Dmel\\CG30467', description='CG30467 gene product from transcript CG30467-RA', other_designations='CG30467-PA|uncharacterized protein', map_location='52A10-52A10', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15538175, end_position_on_the_genomic_accession=15540014, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31258, CurrentID=0, Status='live', Symbol='CG2681', Aliases='Dmel_ 100G10.2, Dmel\\CG2681, EG:100G10.2', description='CG2681 gene product from transcript CG2681-RA', other_designations='CG2681-PA', map_location='3B2-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2708684, end_position_on_the_genomic_accession=2711668, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318622, CurrentID=0, Status='live', Symbol='CG31198', Aliases='Dmel_ CG5845, Dmel\\CG31198, anon-WO0140519.12, anon-WO0140519.24', description='CG31198 gene product from transcript CG31198-RA', other_designations='CG31198-PA|uncharacterized protein', map_location='93F8-93F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21754324, end_position_on_the_genomic_accession=21757758, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318623, CurrentID=0, Status='live', Symbol='GluRIIE', Aliases='Dmel_CG31201, CG31201, DmelGluRIIE, Dmel\\CG31201, GluR-IIE', description='Glutamate receptor IIE', other_designations='CG31201-PB|GluRIIE-PB|glutamate receptor IIE', map_location='92F4-92F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20644154, end_position_on_the_genomic_accession=20648716, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  318624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318624, CurrentID=0, Status='live', Symbol='alpha-Man-Ic', Aliases='Dmel_CG31202, CG31202, CG7880, Dmel\\CG31202, dMas-2, mas-2', description='alpha-Mannosidase class I c', other_designations='CG31202-PA|alpha-Man-Ic-PA', map_location='99C7-99C7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29928701, end_position_on_the_genomic_accession=29930386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318625, CurrentID=0, Status='live', Symbol='CG31204', Aliases='Dmel_ BcDNA:AT25127, CG15569, Dmel\\CG31204', description='CG31204 gene product from transcript CG31204-RA', other_designations='CG31204-PA|uncharacterized protein', map_location='100C3-100C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31391894, end_position_on_the_genomic_accession=31393527, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318626, CurrentID=0, Status='live', Symbol='CG31205', Aliases='Dmel_ Dmel\\CG31205', description='CG31205 gene product from transcript CG31205-RC', other_designations='CG31205-PB|CG31205-PC|uncharacterized protein', map_location='92F10-92F10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20731631, end_position_on_the_genomic_accession=20732707, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31259, CurrentID=0, Status='live', Symbol='CG2685', Aliases='Dmel_ Dmel\\CG2685, EG:100G10.1', description='CG2685 gene product from transcript CG2685-RA', other_designations='CG2685-PA', map_location='3B3-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2711756, end_position_on_the_genomic_accession=2713960, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318632, CurrentID=0, Status='live', Symbol='Ir94f', Aliases='Dmel_CG31225, CG31225, DmelIR94f, Dmel\\CG31225, IR94f', description='Ionotropic receptor 94f', other_designations='CG31225-PB|Ir94f-PB|ionotropic receptor 94f', map_location='94F5-94F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23481865, end_position_on_the_genomic_accession=23483541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31260, CurrentID=0, Status='live', Symbol='CG2694', Aliases='Dmel_ Dmel\\CG2694, EG:100G10.8, EG:95B7.10', description='CG2694 gene product from transcript CG2694-RB', other_designations='CG2694-PA|CG2694-PB|CG2694-PD', map_location='3B3-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2714079, end_position_on_the_genomic_accession=2717654, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318636, CurrentID=0, Status='live', Symbol='CG31229', Aliases='Dmel_ Dmel\\CG31229', description='CG31229 gene product from transcript CG31229-RA', other_designations='CG31229-PA|uncharacterized protein', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18667852, end_position_on_the_genomic_accession=18669083, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318637, CurrentID=0, Status='live', Symbol='CG31230', Aliases='Dmel_ Dmel\\CG31230', description='CG31230 gene product from transcript CG31230-RA', other_designations='CG31230-PA|uncharacterized protein', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18671679, end_position_on_the_genomic_accession=18672094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318638, CurrentID=0, Status='live', Symbol='CG31231', Aliases='Dmel_ BcDNA:AT20662, BcDNA:GH08511, Dmel\\CG31231, anon-WO0140519.92', description='CG31231 gene product from transcript CG31231-RA', other_designations='CG31231-PA|uncharacterized protein', map_location='91B8-91B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18665361, end_position_on_the_genomic_accession=18667733, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326173, CurrentID=0, Status='live', Symbol='CG44001', Aliases='Dmel_ CG31919, CG5822, Dmel\\CG44001, Dromel_CG31919_FBtr0079006_mORF', description='CG44001 gene product from transcript CG44001-RA', other_designations='CG44001-PA|CG44001-PB|CG44001-PC|CG44001-PD|CG44001-PE|uncharacterized protein', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5043270, end_position_on_the_genomic_accession=5048365, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318645, CurrentID=0, Status='live', Symbol='CG31251', Aliases='Dmel_ CG7477, Dmel\\CG31251, NudC-L', description='CG31251 gene product from transcript CG31251-RA', other_designations='CG31251-PA|NudC-like|uncharacterized protein', map_location='90C1-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17683740, end_position_on_the_genomic_accession=17684915, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318646, CurrentID=0, Status='live', Symbol='CG31253', Aliases='Dmel_ BcDNA:RE50074, Dmel\\CG31253', description='CG31253 gene product from transcript CG31253-RA', other_designations='CG31253-PA|uncharacterized protein', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27959346, end_position_on_the_genomic_accession=27960573, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318649, CurrentID=0, Status='live', Symbol='CG31262', Aliases='Dmel_ Dmel\\CG31262, dSLC5A14', description='CG31262 gene product from transcript CG31262-RB', other_designations='CG31262-PB|CG31262-PD|CG31262-PE|CG31262-PF|uncharacterized protein', map_location='89F3-89F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17217116, end_position_on_the_genomic_accession=17223074, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  318651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318651, CurrentID=0, Status='live', Symbol='CG31267', Aliases='Dmel_ CG5240, Dmel\\CG31267, SPH146', description='CG31267 gene product from transcript CG31267-RA', other_designations='CG31267-PA|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17134197, end_position_on_the_genomic_accession=17135283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318652, CurrentID=0, Status='live', Symbol='CG31268', Aliases='Dmel_ Dmel\\CG31268', description='CG31268 gene product from transcript CG31268-RB', other_designations='CG31268-PA|CG31268-PB|uncharacterized protein', map_location='89E9-89E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17031802, end_position_on_the_genomic_accession=17033515, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318653, CurrentID=0, Status='live', Symbol='CG31269', Aliases='Dmel_ CG5233, Dmel\\CG31269, SP126', description='CG31269 gene product from transcript CG31269-RB', other_designations='CG31269-PB|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17127126, end_position_on_the_genomic_accession=17128201, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318655 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318655, CurrentID=0, Status='live', Symbol='CG31274', Aliases='Dmel_ BcDNA:LP05795, Dmel\\CG31274', description='CG31274 gene product from transcript CG31274-RB', other_designations='CG31274-PB|uncharacterized protein', map_location='89E6-89E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16993016, end_position_on_the_genomic_accession=16994532, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318657, CurrentID=0, Status='live', Symbol='CG31279', Aliases='Dmel_ Dmel\\CG31279', description='CG31279 gene product from transcript CG31279-RD', other_designations='CG31279-PB|CG31279-PC|CG31279-PD|CG31279-PE|CG31279-PF|CG31279-PG|uncharacterized protein', map_location='89B16-89B16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16430681, end_position_on_the_genomic_accession=16432897, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318658, CurrentID=0, Status='live', Symbol='Tpl94D', Aliases='Dmel_CG31281, CG31281, Dmel\\CG31281, NEST:bs11b10, NEST:bs30g09, Tpl[94D], tpl, tpl94D, tpl[94D]', description='Transition protein-like 94D', other_designations='CG31281-PA|CG31281-PB|Tpl94D-PA|Tpl94D-PB|transition protein-like 94D|transition protein-like[94D]', map_location='94D5-94D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22966284, end_position_on_the_genomic_accession=22967842, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772278, CurrentID=0, Status='live', Symbol='His2A:CG33862', Aliases='Dmel_CG33862, CG33862, Dmel\\CG33862', description='CG33862 gene product from transcript CG33862-RA', other_designations='CG33862-PA|His2A:CG33862-PA|histone H2A', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21533184, end_position_on_the_genomic_accession=21533676, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318662, CurrentID=0, Status='live', Symbol='CG31286', Aliases='Dmel_ BcDNA:AT25568, Dmel\\CG31286, NEST:bs36b12', description='CG31286 gene product from transcript CG31286-RB', other_designations='CG31286-PB|uncharacterized protein', map_location='84A1-84A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6641232, end_position_on_the_genomic_accession=6643346, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  326177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326177, CurrentID=0, Status='live', Symbol='CG31937', Aliases='Dmel_ CG17711, CT39259, Dmel\\CG31937', description='CG31937 gene product from transcript CG31937-RA', other_designations='CG31937-PA|uncharacterized protein', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1729887, end_position_on_the_genomic_accession=1731070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318664, CurrentID=0, Status='live', Symbol='CG31288', Aliases='Dmel_ CG13665, Dmel\\CG31288', description='CG31288 gene product from transcript CG31288-RA', other_designations='CG31288-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25297960, end_position_on_the_genomic_accession=25299557, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318666, CurrentID=0, Status='live', Symbol='CG31294', Aliases='Dmel_ BcDNA:AT21457, Dmel\\CG31294, NEST:bs15e12, NEST:bs25d10', description='CG31294 gene product from transcript CG31294-RA', other_designations='CG31294-PA|uncharacterized protein', map_location='89A2-89A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15715505, end_position_on_the_genomic_accession=15716585, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318668, CurrentID=0, Status='live', Symbol='CG31296', Aliases='Dmel_ Dmel\\CG31296, NEST:bs01d08', description='CG31296 gene product from transcript CG31296-RB', other_designations='CG31296-PA|CG31296-PB|uncharacterized protein', map_location='88F7-88F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15530072, end_position_on_the_genomic_accession=15531473, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326178, CurrentID=0, Status='live', Symbol='CG31948', Aliases='Dmel_ CG9966, Dmel\\CG31948', description='CG31948 gene product from transcript CG31948-RB', other_designations='CG31948-PA|CG31948-PB|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2415359, end_position_on_the_genomic_accession=2416720, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772280, CurrentID=0, Status='live', Symbol='LSm-4', Aliases='Dmel_CG33677, BcDNA:SD20015, CG31990, CG31990-ORFA, CG33677, Dmel\\CG33677, LSM4, LSm4', description='Like Sm protein 4', other_designations='CG33677-PB|CG33677-PC|LSm-4-PB|LSm-4-PC|like Sm protein 4', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10403197, end_position_on_the_genomic_accession=10405203, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326179, CurrentID=0, Status='live', Symbol='CG31952', Aliases='Dmel_ BcDNA:AT14709, CG15413, Dmel\\CG31952', description='CG31952 gene product from transcript CG31952-RB', other_designations='CG31952-PA|CG31952-PB|uncharacterized protein', map_location='23F1-23F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3416693, end_position_on_the_genomic_accession=3418068, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318677, CurrentID=0, Status='live', Symbol='CG31313', Aliases='Dmel_ BcDNA:RH02566, Dmel\\CG31313', description='CG31313 gene product from transcript CG31313-RA', other_designations='CG31313-PA|CG31313-PB|CG31313-PC|uncharacterized protein', map_location='88C6-88C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14566432, end_position_on_the_genomic_accession=14567429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318680, CurrentID=0, Status='live', Symbol='HEATR2', Aliases='Dmel_CG31320, CG31320, Dmel\\CG31320, NEST:bs27f03', description='HEAT repeat containing 2', other_designations='CG31320-PA|HEATR2-PA', map_location='88B6-88B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14358356, end_position_on_the_genomic_accession=14361786, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318681, CurrentID=0, Status='live', Symbol='CG31321', Aliases='Dmel_ BcDNA:RH31685, Dmel\\CG31321', description='CG31321 gene product from transcript CG31321-RB', other_designations='CG31321-PB|uncharacterized protein', map_location='88B3-88B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14310362, end_position_on_the_genomic_accession=14314412, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  318682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318682, CurrentID=0, Status='live', Symbol='CG31323', Aliases='Dmel_ CG14558, Dmel\\CG31323', description='CG31323 gene product from transcript CG31323-RA', other_designations='CG31323-PA|CG31323-PB|uncharacterized protein', map_location='97A2-97A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26170847, end_position_on_the_genomic_accession=26187307, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318684, CurrentID=0, Status='live', Symbol='CG31327', Aliases='Dmel_ BcDNA:AT17065, Dmel\\CG31327', description='CG31327 gene product from transcript CG31327-RA', other_designations='CG31327-PA|uncharacterized protein', map_location='88A10-88A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14157560, end_position_on_the_genomic_accession=14160278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772282, CurrentID=0, Status='live', Symbol='His2A:CG33808', Aliases='Dmel_CG33808, CG33808, Dmel\\CG33808', description='CG33808 gene product from transcript CG33808-RA', other_designations='CG33808-PA|His2A:CG33808-PA|histone H2A', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21433862, end_position_on_the_genomic_accession=21434337, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  260645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=260645, CurrentID=0, Status='live', Symbol='NimB2', Aliases='Dmel_CG31839, BG:DS00180.7, CG16882, CG31839, CT37470, DS00180.7, Dmel\\CG31839, nimB2', description='Nimrod B2', other_designations='CG31839-PA|NimB2-PA|nimrod B2|nimrodB2', map_location='34E5-34E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13965012, end_position_on_the_genomic_accession=13967018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318688, CurrentID=0, Status='live', Symbol='CG31337', Aliases='Dmel_ BcDNA:RE40816, Dmel\\CG31337', description='CG31337 gene product from transcript CG31337-RA', other_designations='CG31337-PA|uncharacterized protein', map_location='87F2-87F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13543966, end_position_on_the_genomic_accession=13545400, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326182 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326182, CurrentID=0, Status='live', Symbol='CG31988', Aliases='Dmel_ CG17401, Dmel\\CG31988, anon-WO0140519.26, bs21c01.y1', description='CG31988 gene product from transcript CG31988-RB', other_designations='CG31988-PA|CG31988-PB|uncharacterized protein', map_location='39B1-39B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21201493, end_position_on_the_genomic_accession=21202802, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318698, CurrentID=0, Status='live', Symbol='CG31360', Aliases='Dmel_ BcDNA:RE06652, Dmel\\CG31360', description='CG31360 gene product from transcript CG31360-RA', other_designations='CG31360-PA|uncharacterized protein', map_location='90C1-90C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17685097, end_position_on_the_genomic_accession=17685807, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326183, CurrentID=0, Status='live', Symbol='CG32037', Aliases='Dmel_ CG3426, Dmel\\CG32037', description='CG32037 gene product from transcript CG32037-RB', other_designations='CG32037-PA|CG32037-PB|uncharacterized protein', map_location='67B10-67B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9489502, end_position_on_the_genomic_accession=9492215, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318700, CurrentID=0, Status='live', Symbol='CG31373', Aliases='Dmel_ CG6482, Dmel\\CG31373, PDF', description='CG31373 gene product from transcript CG31373-RA', other_designations='CG31373-PA|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10708487, end_position_on_the_genomic_accession=10709364, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318701, CurrentID=0, Status='live', Symbol='CG31380', Aliases='Dmel_ CG4964, Dmel\\CG31380', description='CG31380 gene product from transcript CG31380-RA', other_designations='CG31380-PA|uncharacterized protein', map_location='96F3-96F3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25787997, end_position_on_the_genomic_accession=25789455, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31272, CurrentID=0, Status='live', Symbol='CG12498', Aliases='Dmel_ Dmel\\CG12498, EG:BACR43E12.1', description='CG12498 gene product from transcript CG12498-RA', other_designations='CG12498-PA|uncharacterized protein', map_location='3C1-3C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2809881, end_position_on_the_genomic_accession=2810846, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318709, CurrentID=0, Status='live', Symbol='CG42726', Aliases='Dmel_ CG12258, CG31392, Dmel\\CG42726, Dmel_CG31392', description='CG42726 gene product from transcript CG42726-RA', other_designations='CG42726-PA|uncharacterized protein', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15348048, end_position_on_the_genomic_accession=15350391, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31273, CurrentID=0, Status='live', Symbol='CG14416', Aliases='Dmel_ Dmel\\CG14416, EG:BACR43E12.7', description='CG14416 gene product from transcript CG14416-RA', other_designations='CG14416-PA|uncharacterized protein', map_location='3C2-3C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2871884, end_position_on_the_genomic_accession=2872853, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318714, CurrentID=0, Status='live', Symbol='CG31404', Aliases='Dmel_ CG3575, Dmel\\CG31404', description='CG31404 gene product from transcript CG31404-RA', other_designations='CG31404-PA|uncharacterized protein', map_location='92A3-92A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19375043, end_position_on_the_genomic_accession=19376023, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318715, CurrentID=0, Status='live', Symbol='CG31406', Aliases='Dmel_ BcDNA:AT13367, Dmel\\CG31406', description='CG31406 gene product from transcript CG31406-RB', other_designations='CG31406-PA|CG31406-PB|uncharacterized protein', map_location='86B1-86B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10348761, end_position_on_the_genomic_accession=10350010, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318716, CurrentID=0, Status='live', Symbol='CG31407', Aliases='Dmel_ Dmel\\CG31407, NEST:bs33d08, NEST:bs35f01', description='CG31407 gene product from transcript CG31407-RB', other_designations='CG31407-PA|CG31407-PB|uncharacterized protein', map_location='85F7-85F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10022845, end_position_on_the_genomic_accession=10025150, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31274, CurrentID=0, Status='live', Symbol='CG14417', Aliases='Dmel_ Dmel\\CG14417, EG:BACR43E12.6', description='CG14417 gene product from transcript CG14417-RA', other_designations='CG14417-PA|uncharacterized protein', map_location='3C2-3C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2875428, end_position_on_the_genomic_accession=2876533, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318720, CurrentID=0, Status='live', Symbol='CG31413', Aliases='Dmel_ CG10299, Dmel\\CG31413', description='CG31413 gene product from transcript CG31413-RA', other_designations='CG31413-PA|uncharacterized protein', map_location='95A8-95A9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23702761, end_position_on_the_genomic_accession=23704542, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318721, CurrentID=0, Status='live', Symbol='CG31414', Aliases='Dmel_ CG10299, Dmel\\CG31414', description='CG31414 gene product from transcript CG31414-RC', other_designations='CG31414-PB|CG31414-PC|glucosylceramidase|uncharacterized protein', map_location='95A9-95A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23704804, end_position_on_the_genomic_accession=23708512, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  318722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318722, CurrentID=0, Status='live', Symbol='CG31415', Aliases='Dmel_ Dmel\\CG31415, NEST:bs07c11', description='CG31415 gene product from transcript CG31415-RA', other_designations='CG31415-PA|uncharacterized protein', map_location='85E6-85E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9743517, end_position_on_the_genomic_accession=9744662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31275, CurrentID=0, Status='live', Symbol='CG14418', Aliases='Dmel_ Dmel\\CG14418, EG:BACR43E12.5', description='CG14418 gene product from transcript CG14418-RB', other_designations='CG14418-PA|CG14418-PB|uncharacterized protein', map_location='3C2-3C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2882226, end_position_on_the_genomic_accession=2883661, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318725, CurrentID=0, Status='live', Symbol='Takl1', Aliases='Dmel_CG31421, CG31421, Dmel\\CG31421, FBgn0046689, TAKL1, anon-AE003732.1', description='Tak1-like 1', other_designations='CG31421-PA|Takl1-PA|extra2', map_location='92F13-92F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20770828, end_position_on_the_genomic_accession=20772601, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318727, CurrentID=0, Status='live', Symbol='Ir94c', Aliases='Dmel_CG31423, CG31423, DmelIR94c, Dmel\\CG31423, IR94c', description='Ionotropic receptor 94c', other_designations='CG31423-PB|Ir94c-PB|ionotropic receptor 94c', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22117640, end_position_on_the_genomic_accession=22119454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318728, CurrentID=0, Status='live', Symbol='Ir94b', Aliases='Dmel_CG31424, CG31424, DmelIR94b, Dmel\\CG31424, IR94b', description='Ionotropic receptor 94b', other_designations='CG31424-PB|Ir94b-PB|ionotropic receptor 94b', map_location='94A2-94A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22115471, end_position_on_the_genomic_accession=22117249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31276, CurrentID=0, Status='live', Symbol='CG14419', Aliases='Dmel_ Dmel\\CG14419', description='CG14419 gene product from transcript CG14419-RA', other_designations='CG14419-PA|uncharacterized protein', map_location='3C2-3C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2885186, end_position_on_the_genomic_accession=2885914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318734, CurrentID=0, Status='live', Symbol='CG31436', Aliases='Dmel_ CG13660, Dmel\\CG31436', description='CG31436 gene product from transcript CG31436-RA', other_designations='CG31436-PA|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25304137, end_position_on_the_genomic_accession=25305651, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31277 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31277, CurrentID=0, Status='live', Symbol='CG3526', Aliases='Dmel_ Dmel\\CG3526, EG:BACR43E12.4', description='CG3526 gene product from transcript CG3526-RC', other_designations='CG3526-PA|CG3526-PC', map_location='3C2-3C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2890768, end_position_on_the_genomic_accession=2894232, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318736, CurrentID=0, Status='live', Symbol='CheB93b', Aliases='Dmel_CG31438, CG31438, Dmel\\CG31438', description='Chemosensory protein B 93b', other_designations='CG31438-PA|CheB93b-PA|chemosensory protein B 93b', map_location='93E1-93E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21409123, end_position_on_the_genomic_accession=21409924, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318737, CurrentID=0, Status='live', Symbol='Muc96D', Aliases='Dmel_CG31439, CG31439, Dmel\\CG31439', description='Mucin 96D', other_designations='CG31439-PA|Muc96D-PA|mucin 96D', map_location='96D4-96D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25397788, end_position_on_the_genomic_accession=25400645, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318741, CurrentID=0, Status='live', Symbol='CG31446', Aliases='Dmel_ BcDNA:RE40372, Dmel\\CG31446', description='CG31446 gene product from transcript CG31446-RB', other_designations='CG31446-PA|CG31446-PB|uncharacterized protein', map_location='89A2-89A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15693086, end_position_on_the_genomic_accession=15694194, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318744, CurrentID=0, Status='live', Symbol='CG31454', Aliases='Dmel_ CG11740, Dmel\\CG31454', description='CG31454 gene product from transcript CG31454-RA', other_designations='CG31454-PA|uncharacterized protein', map_location='85A2-85A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8521658, end_position_on_the_genomic_accession=8523520, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318746, CurrentID=0, Status='live', Symbol='CG31459', Aliases='Dmel_ BcDNA:AT26603, Dmel\\CG31459, NEST:bs03f05, NEST:bs29d02', description='CG31459 gene product from transcript CG31459-RA', other_designations='CG31459-PA|uncharacterized protein', map_location='92B3-92B3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19840845, end_position_on_the_genomic_accession=19842040, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318747, CurrentID=0, Status='live', Symbol='CG31460', Aliases='Dmel_ BcDNA:RE67710, CG11745, Dmel\\CG31460', description='CG31460 gene product from transcript CG31460-RB', other_designations='CG31460-PA|CG31460-PB|uncharacterized protein', map_location='85A3-85A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8562673, end_position_on_the_genomic_accession=8563445, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318749, CurrentID=0, Status='live', Symbol='CG31464', Aliases='Dmel_ CG11673, Dmel\\CG31464', description='CG31464 gene product from transcript CG31464-RB', other_designations='CG31464-PB|uncharacterized protein', map_location='84F9-84F9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8291240, end_position_on_the_genomic_accession=8292034, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318751, CurrentID=0, Status='live', Symbol='CG31467', Aliases='Dmel_ CG6471, Dmel\\CG31467', description='CG31467 gene product from transcript CG31467-RA', other_designations='CG31467-PA|uncharacterized protein', map_location='86C4-86C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10705170, end_position_on_the_genomic_accession=10706863, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31280 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31280, CurrentID=0, Status='live', Symbol='CG32793', Aliases='Dmel_ CG18089, CG18852, CG3591, Dmel\\CG32793, EG:100G7.1, EG:100G7.2, anon-3Ca, anon-3Cb', description='CG32793 gene product from transcript CG32793-RA', other_designations='CG32793-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2929796, end_position_on_the_genomic_accession=2932626, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31282, CurrentID=0, Status='live', Symbol='CG3592', Aliases='Dmel_ Dmel\\CG3592, frag7', description='CG3592 gene product from transcript CG3592-RA', other_designations='CG3592-PA', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2935127, end_position_on_the_genomic_accession=2935966, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31283 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31283, CurrentID=0, Status='live', Symbol='CG3598', Aliases='Dmel_ Dmel\\CG3598, EG:100G7.3', description='CG3598 gene product from transcript CG3598-RA', other_designations='CG3598-PA', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2937027, end_position_on_the_genomic_accession=2938073, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318776, CurrentID=0, Status='live', Symbol='CG31516', Aliases='Dmel_ BcDNA:RE68446, Dmel\\CG31516', description='CG31516 gene product from transcript CG31516-RA', other_designations='CG31516-PA|uncharacterized protein', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4218463, end_position_on_the_genomic_accession=4220130, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31284 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31284, CurrentID=0, Status='live', Symbol='CG14420', Aliases='Dmel_ Dmel\\CG14420', description='CG14420 gene product from transcript CG14420-RA', other_designations='CG14420-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2938856, end_position_on_the_genomic_accession=2940527, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318778, CurrentID=0, Status='live', Symbol='Or82a', Aliases='Dmel_CG31519, 82a, CG31519, DOR61, DmOR82A, Dmel\\CG31519', description='Odorant receptor 82a', other_designations='CG31519-PA|Or82a-PA|odorant receptor 82a', map_location='82A1-82A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4256972, end_position_on_the_genomic_accession=4258444, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  318781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318781, CurrentID=0, Status='live', Symbol='CG31524', Aliases='Dmel_ Dmel\\CG31524', description='CG31524 gene product from transcript CG31524-RB', other_designations='CG31524-PA|CG31524-PB|uncharacterized protein', map_location='99F6-99F7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30512325, end_position_on_the_genomic_accession=30514355, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  318782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318782, CurrentID=0, Status='live', Symbol='CG31525', Aliases='Dmel_ Dmel\\CG31525', description='CG31525 gene product from transcript CG31525-RA', other_designations='CG31525-PA|uncharacterized protein', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4427030, end_position_on_the_genomic_accession=4428200, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31285, CurrentID=0, Status='live', Symbol='CG14421', Aliases='Dmel_ Dmel\\CG14421', description='CG14421 gene product from transcript CG14421-RA', other_designations='CG14421-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2941408, end_position_on_the_genomic_accession=2942329, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318785, CurrentID=0, Status='live', Symbol='CG31528', Aliases='Dmel_ BEST:GH26112, BcDNA:GH26112, Dmel\\CG31528', description='CG31528 gene product from transcript CG31528-RA', other_designations='CG31528-PA|uncharacterized protein', map_location='82B1-82B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4416438, end_position_on_the_genomic_accession=4417665, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740775, CurrentID=0, Status='live', Symbol='CG34450', Aliases='Dmel_ CG12136, CG32705, Dmel\\CG34450, Dmel_CG32705', description='CG34450 gene product from transcript CG34450-RA', other_designations='CG34450-PA|uncharacterized protein', map_location='8D8-8D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9329016, end_position_on_the_genomic_accession=9330487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772299, CurrentID=0, Status='live', Symbol='His2B:CG33910', Aliases='Dmel_CG33910, CG33910, Dmel\\CG33910', description='CG33910 gene product from transcript CG33910-RA', other_designations='CG33910-PA|His2B:CG33910-PA|histone H2B', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21418112, end_position_on_the_genomic_accession=21418706, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318788, CurrentID=0, Status='live', Symbol='CG31533', Aliases='Dmel_ BEST:LP07046, BcDNA:AT30147, Dmel\\CG31533, NEST:bs16a07', description='CG31533 gene product from transcript CG31533-RA', other_designations='CG31533-PA|uncharacterized protein', map_location='88A10-88A10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14154728, end_position_on_the_genomic_accession=14157431, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31286 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31286, CurrentID=0, Status='live', Symbol='CG14422', Aliases='Dmel_ Dmel\\CG14422', description='CG14422 gene product from transcript CG14422-RA', other_designations='CG14422-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2943386, end_position_on_the_genomic_accession=2944569, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318790, CurrentID=0, Status='live', Symbol='CG31538', Aliases='Dmel_ BEST:GH11373, BEST:GH20014, BEST:GH27336, BEST:LP11873, BcDNA:AT27831, Dmel\\CG31538, NEST:bs02h07, NEST:bs09e09, NEST:bs09g01, NEST:bs10d08, NEST:bs26g11', description='CG31538 gene product from transcript CG31538-RA', other_designations='CG31538-PA|uncharacterized protein', map_location='82F8-82F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5275678, end_position_on_the_genomic_accession=5277835, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318791, CurrentID=0, Status='live', Symbol='CG31542', Aliases='Dmel_ BcDNA:AT15479, Dmel\\CG31542, NEST:bs01d10, NEST:bs08f12, NEST:bs29h01', description='CG31542 gene product from transcript CG31542-RC', other_designations='CG31542-PA|CG31542-PC|uncharacterized protein', map_location='83A1-83A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5369540, end_position_on_the_genomic_accession=5371744, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318792, CurrentID=0, Status='live', Symbol='CG31544', Aliases='Dmel_ Dmel\\CG31544', description='CG31544 gene product from transcript CG31544-RA', other_designations='CG31544-PA|uncharacterized protein', map_location='84D6-84D6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7418998, end_position_on_the_genomic_accession=7420132, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318794, CurrentID=0, Status='live', Symbol='CG31548', Aliases='Dmel_ BcDNA:RH24570, CG2907, Dmel\\CG31548', description='CG31548 gene product from transcript CG31548-RA', other_designations='CG31548-PA|uncharacterized protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5635901, end_position_on_the_genomic_accession=5636852, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318795, CurrentID=0, Status='live', Symbol='CG31551', Aliases='Dmel_ CG11005, Dmel\\CG31551', description='CG31551 gene product from transcript CG31551-RA', other_designations='CG31551-PA|uncharacterized protein', map_location='83B7-83B7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=5627852, end_position_on_the_genomic_accession=5632033, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318801, CurrentID=0, Status='live', Symbol='Osi16', Aliases='Dmel_CG31561, CG10314, CG31561, Dmel\\CG31561', description='Osiris 16', other_designations='CG31561-PA|Osi16-PA|osiris|osiris 16', map_location='83E2-83E2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=6305247, end_position_on_the_genomic_accession=6306366, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326202, CurrentID=0, Status='live', Symbol='CG32228', Aliases='Dmel_ Dmel\\CG32228', description='CG32228 gene product from transcript CG32228-RA', other_designations='CG32228-PA|uncharacterized protein', map_location='77B1-77B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20265814, end_position_on_the_genomic_accession=20268800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31291, CurrentID=0, Status='live', Symbol='CG4116', Aliases='Dmel_ Dmel\\CG4116', description='CG4116 gene product from transcript CG4116-RA', other_designations='CG4116-PA|CG4116-PB', map_location='3C6-3C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3066615, end_position_on_the_genomic_accession=3067562, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772305 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772305, CurrentID=0, Status='live', Symbol='CG10731', Aliases='Dmel_ ATPW, CG33038, Dmel\\CG10731', description='CG10731 gene product from transcript CG10731-RC', other_designations='ATP synthase coupling factor B|CG10731-PB|CG10731-PC|uncharacterized protein', map_location='52E11-52E11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16148812, end_position_on_the_genomic_accession=16149795, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318824, CurrentID=0, Status='live', Symbol='dpy', Aliases='Dmel_CG33196, BcDNA:SD02173, CG15637, CG31724, CG31725, CG31726, CG31727, CG31964, CG31965, CG31966, CG31967, CG31968, CG33000, CG33086, CG33196, CR33195, CT35799, DP, DPY, Dmel\\CG33196, Dp, Mur24F, Q8IQ18_DROME, SP460, dp, l(2)24Fd, olv[D]', description='dumpy', other_designations='CG33196-PAA|CG33196-PI|CG33196-PJ|CG33196-PM|CG33196-PO|CG33196-PP|CG33196-PQ|CG33196-PR|CG33196-PS|CG33196-PT|CG33196-PU|CG33196-PV|CG33196-PW|CG33196-PX|CG33196-PY|CG33196-PZ|dpy-PAA|dpy-PI|dpy-PJ|dpy-PM|dpy-PO|dpy-PP|dpy-PQ|dpy-PR|dpy-PS|dpy-PT|dpy-PU|dpy-PV|dpy-PW|dpy-PX|dpy-PY|dpy-PZ|fibrillin-like|mucin related 24F', map_location='2-13.0 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4477462, end_position_on_the_genomic_accession=4614300, orientation='minus', exon_count=85, OMIM='', no_hdr0=''),
  318825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318825, CurrentID=0, Status='live', Symbol='mRpL27', Aliases='Dmel_CG33002, CG15442, CG33002, Dmel\\CG33002', description='mitochondrial ribosomal protein L27', other_designations='CG33002-PA|CG33002-PB|CG33002-PC|mRpL27-PA|mRpL27-PB|mRpL27-PC|ribosomal protein L27A', map_location='24F3-24F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4458595, end_position_on_the_genomic_accession=4459638, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318826, CurrentID=0, Status='live', Symbol='CG33003', Aliases='Dmel_ Dmel\\CG33003', description='CG33003 gene product from transcript CG33003-RB', other_designations='CG33003-PA|CG33003-PB|uncharacterized protein', map_location='24F1-24F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4400953, end_position_on_the_genomic_accession=4403142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326205, CurrentID=0, Status='live', Symbol='Claspin', Aliases='Dmel_CG32251, CG1326, CG32251, Dmel\\CG32251', description='CG32251 gene product from transcript CG32251-RB', other_designations='CG32251-PA|CG32251-PB|Claspin-PA|Claspin-PB|claspin', map_location='64B1-64B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4313483, end_position_on_the_genomic_accession=4318608, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318833, CurrentID=0, Status='live', Symbol='CG33057', Aliases='Dmel_ Dmel\\CG33057', description='CG33057 gene product from transcript CG33057-RB', other_designations='CG33057-PA|CG33057-PB|uncharacterized protein', map_location='66C11-66C11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8402168, end_position_on_the_genomic_accession=8405277, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318834, CurrentID=0, Status='live', Symbol='CG31600', Aliases='Dmel_ Dmel\\CG31600', description='CG31600 gene product from transcript CG31600-RB', other_designations='CG31600-PA|CG31600-PB|uncharacterized protein', map_location='40B4-40B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21914269, end_position_on_the_genomic_accession=21915358, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318835, CurrentID=0, Status='live', Symbol='CG31601', Aliases='Dmel_ Dmel\\CG31601', description='CG31601 gene product from transcript CG31601-RC', other_designations='CG31601-PB|CG31601-PC|uncharacterized protein', map_location='40D4-40D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22040727, end_position_on_the_genomic_accession=22043877, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  326206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326206, CurrentID=0, Status='live', Symbol='CG11583', Aliases='Dmel_ CG32253, Dmel\\CG11583, dBRX1', description='CG11583 gene product from transcript CG11583-RA', other_designations='CG11583-PA|uncharacterized protein', map_location='64A12-64A12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4281504, end_position_on_the_genomic_accession=4282817, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318838, CurrentID=0, Status='live', Symbol='CG33061', Aliases='Dmel_ CG13057, Dmel\\CG33061', description='CG33061 gene product from transcript CG33061-RC', other_designations='CG33061-PB|CG33061-PC|uncharacterized protein', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16329795, end_position_on_the_genomic_accession=16331227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318840, CurrentID=0, Status='live', Symbol='CG33062', Aliases='Dmel_ BcDNA:SD02058, Dmel\\CG33062', description='CG33062 gene product from transcript CG33062-RB', other_designations='CG33062-PA|CG33062-PB|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19441719, end_position_on_the_genomic_accession=19446444, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318841, CurrentID=0, Status='live', Symbol='Bsg', Aliases='Dmel_CG31605, 135/10, BcDNA:GH21853, CG31605, Dmel\\CG31605, NP6293, bsg, gel, l(2)06243, l(2)SH1217, l(2)SH2 1217, l(2)k06338, l(2)k09030, l(2)k13638, l(2)k14308, ms(2)08318', description='Basigin', other_designations='Bsg-PA|Bsg-PB|Bsg-PC|Bsg-PD|Bsg-PE|Bsg-PF|Bsg-PG|Bsg-PH|Bsg-PI|Bsg-PJ|CG31605-PA|CG31605-PB|CG31605-PC|CG31605-PD|CG31605-PE|CG31605-PF|CG31605-PG|CG31605-PH|CG31605-PI|CG31605-PJ|D-basigin|basigin|gelded|lethal (2) SH1217', map_location='28E3-28E6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8083420, end_position_on_the_genomic_accession=8116531, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  318842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318842, CurrentID=0, Status='live', Symbol='CG31606', Aliases='Dmel_ Dmel\\CG31606', description='CG31606 gene product from transcript CG31606-RB', other_designations='CG31606-PB|uncharacterized protein', map_location='28E7-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8147611, end_position_on_the_genomic_accession=8148356, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31295, CurrentID=0, Status='live', Symbol='CG3939', Aliases='Dmel_ BEST:SD07039, Dmel\\CG3939, EG:140G11.5', description='CG3939 gene product from transcript CG3939-RA', other_designations='CG3939-PA|CG3939-PB', map_location='3C9-3C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3175018, end_position_on_the_genomic_accession=3176223, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318845, CurrentID=0, Status='live', Symbol='CG31609', Aliases='Dmel_ Dmel\\CG31609, anon-AE003623.1', description='CG31609 gene product from transcript CG31609-RB', other_designations='CG31609-PB|uncharacterized protein', map_location='30A1-30A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9097563, end_position_on_the_genomic_accession=9098175, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318846, CurrentID=0, Status='live', Symbol='His4:CG31611', Aliases='Dmel_CG31611, CG31611, Dmel\\CG31611', description='CG31611 gene product from transcript CG31611-RA', other_designations='CG31611-PA|His4:CG31611-PA|histone H4', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21424593, end_position_on_the_genomic_accession=21425005, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318847 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318847, CurrentID=0, Status='live', Symbol='His3:CG31613', Aliases='Dmel_CG31613, CG31613, Dmel\\CG31613, His3', description='CG31613 gene product from transcript CG31613-RA', other_designations='CG31613-PA|His3:CG31613-PA|histone H3', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21425193, end_position_on_the_genomic_accession=21425721, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318854, CurrentID=0, Status='live', Symbol='His1:CG31617', Aliases='Dmel_CG31617, CG31617, Dmel\\CG31617', description='CG31617 gene product from transcript CG31617-RA', other_designations='CG31617-PA|His1:CG31617-PA|histone H1', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21421855, end_position_on_the_genomic_accession=21422877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318855, CurrentID=0, Status='live', Symbol='His2A:CG31618', Aliases='Dmel_CG31618, BEST:CK01438, CG31618, Dmel\\CG31618', description='CG31618 gene product from transcript CG31618-RA', other_designations='CG31618-PA|His2A:CG31618-PA|histone H2A', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21423737, end_position_on_the_genomic_accession=21424229, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318856, CurrentID=0, Status='live', Symbol='dtr', Aliases='Dmel_CG31623, CG31623, Dmel\\CG31623, Dtr', description='defective transmitter release', other_designations='CG31623-PA|dtr-PA', map_location='39D1-39D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21361216, end_position_on_the_genomic_accession=21367288, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318857, CurrentID=0, Status='live', Symbol='CG33082', Aliases='Dmel_ Dmel\\CG33082', description='CG33082 gene product from transcript CG33082-RB', other_designations='CG33082-PA|CG33082-PB|uncharacterized protein', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1030758, end_position_on_the_genomic_accession=1033119, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772311, CurrentID=0, Status='live', Symbol='CG33777', Aliases='Dmel_ Dmel\\CG33777', description='CG33777 gene product from transcript CG33777-RA', other_designations='CG33777-PA|uncharacterized protein', map_location='64B9-64B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4554537, end_position_on_the_genomic_accession=4555236, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318860, CurrentID=0, Status='live', Symbol='CG31627', Aliases='Dmel_ Dmel\\CG31627, NEST:bs05g10', description='CG31627 gene product from transcript CG31627-RB', other_designations='CG31627-PB|uncharacterized protein', map_location='39A7-39A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21173222, end_position_on_the_genomic_accession=21174179, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318865, CurrentID=0, Status='live', Symbol='CG31636', Aliases='Dmel_ CG13770, Dmel\\CG31636', description='CG31636 gene product from transcript CG31636-RB', other_designations='CG31636-PA|CG31636-PB|uncharacterized protein', map_location='26F5-26F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6656459, end_position_on_the_genomic_accession=6658152, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318866, CurrentID=0, Status='live', Symbol='CG31639', Aliases='Dmel_ Dmel\\CG31639, NEST:bs21c05, NEST:bs29d12', description='CG31639 gene product from transcript CG31639-RB', other_designations='CG31639-PA|CG31639-PB|uncharacterized protein', map_location='26B8-26B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6153840, end_position_on_the_genomic_accession=6155141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326211, CurrentID=0, Status='live', Symbol='CG32379', Aliases='Dmel_ CG18415, Dmel\\CG32379', description='CG32379 gene product from transcript CG32379-RB', other_designations='CG32379-PB|uncharacterized protein', map_location='65F11-65F11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7396355, end_position_on_the_genomic_accession=7398208, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318868 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318868, CurrentID=0, Status='live', Symbol='CG31644', Aliases='Dmel_ BcDNA:AT20031, Dmel\\CG31644', description='CG31644 gene product from transcript CG31644-RB', other_designations='CG31644-PA|CG31644-PB|CG31644-PC|uncharacterized protein', map_location='25F5-25F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5817890, end_position_on_the_genomic_accession=5818443, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318872, CurrentID=0, Status='live', Symbol='CG34126', Aliases='Dmel_ CG31653, CG3805, Dm_2L:23517, Dm_2L:23546, Dmel\\CG34126', description='CG34126 gene product from transcript CG34126-RE', other_designations='CG34126-PB|CG34126-PD|CG34126-PE|CG34126-PF|CG34126-PG|uncharacterized protein', map_location='25B5-25B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4981614, end_position_on_the_genomic_accession=4992702, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  10178859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178859, CurrentID=0, Status='live', Symbol='CG42728', Aliases='Dmel_ CG17014, CG33984, CG7973, Dmel\\CG42728, Dmel_CG33984', description='CG42728 gene product from transcript CG42728-RA', other_designations='CG42728-PA|uncharacterized protein', map_location='71D4-71E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15527620, end_position_on_the_genomic_accession=15528196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318874, CurrentID=0, Status='live', Symbol='Nnf1b', Aliases='Dmel_CG31658, CG31658, Dmel\\CG31658, dmNnf1R-1', description='CG31658 gene product from transcript CG31658-RA', other_designations='CG31658-PA|Nnf1-related protein 1|Nnf1b-PA', map_location='21E2-21E2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=810577, end_position_on_the_genomic_accession=811531, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318875, CurrentID=0, Status='live', Symbol='CG31659', Aliases='Dmel_ CG16933, Dmel\\CG31659', description='CG31659 gene product from transcript CG31659-RA', other_designations='CG31659-PA|CG31659-PB|uncharacterized protein', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1358925, end_position_on_the_genomic_accession=1359790, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  3772314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772314, CurrentID=0, Status='live', Symbol='His4:CG33893', Aliases='Dmel_CG33893, CG33893, Dmel\\CG33893', description='CG33893 gene product from transcript CG33893-RA', other_designations='CG33893-PA|His4:CG33893-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21473999, end_position_on_the_genomic_accession=21474411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318878 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318878, CurrentID=0, Status='live', Symbol='CG31674', Aliases='Dmel_ CG9332, Dmel\\CG31674', description='CG31674 gene product from transcript CG31674-RA', other_designations='CG31674-PA|CG31674-PB|uncharacterized protein', map_location='38F1-38F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20831262, end_position_on_the_genomic_accession=20832859, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318879, CurrentID=0, Status='live', Symbol='CG31675', Aliases='Dmel_ BcDNA:RE19849, Dmel\\CG31675', description='CG31675 gene product from transcript CG31675-RA', other_designations='CG31675-PA|uncharacterized protein', map_location='38F3-38F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20867755, end_position_on_the_genomic_accession=20869650, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318880, CurrentID=0, Status='live', Symbol='CG31677', Aliases='Dmel_ Dmel\\CG31677', description='CG31677 gene product from transcript CG31677-RA', other_designations='CG31677-PA|uncharacterized protein', map_location='38D1-38D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20603535, end_position_on_the_genomic_accession=20604196, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318881, CurrentID=0, Status='live', Symbol='CG31680', Aliases='Dmel_ BcDNA:AT19802, Dmel\\CG31680', description='CG31680 gene product from transcript CG31680-RA', other_designations='CG31680-PA|uncharacterized protein', map_location='38D1-38D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20640121, end_position_on_the_genomic_accession=20640718, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318882 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318882, CurrentID=0, Status='live', Symbol='CG31681', Aliases='Dmel_ Dmel\\CG31681', description='CG31681 gene product from transcript CG31681-RA', other_designations='CG31681-PA|uncharacterized protein', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2285925, end_position_on_the_genomic_accession=2286873, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318883, CurrentID=0, Status='live', Symbol='Tengl1', Aliases='Dmel_CG31682, BcDNA:AT02809, CG17006, CG31682, Dmel\\CG31682', description='Testis EndoG-Like 1', other_designations='CG31682-PA|Tengl1-PA|testis EndoG-Like 1', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2239414, end_position_on_the_genomic_accession=2240593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318886 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318886, CurrentID=0, Status='live', Symbol='CG31687', Aliases='Dmel_CG31687-PA, Dmel\\CG31687', description='CG31687 gene product from transcript CG31687-RC', other_designations='CG31687-PB|CG31687-PC|uncharacterized protein', map_location='38C7-38C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20450225, end_position_on_the_genomic_accession=20456978, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318887, CurrentID=0, Status='live', Symbol='CG31693', Aliases='Dmel_ BcDNA:GH25177, Dmel\\CG31693', description='CG31693 gene product from transcript CG31693-RA', other_designations='CG31693-PA|CG31693-PB|uncharacterized protein', map_location='40D3-40D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22015190, end_position_on_the_genomic_accession=22017334, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318888 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318888, CurrentID=0, Status='live', Symbol='CG31698', Aliases='Dmel_ BcDNA:LP02725, Dmel\\CG31698', description='CG31698 gene product from transcript CG31698-RB', other_designations='CG31698-PA|CG31698-PB|uncharacterized protein', map_location='23D4-23D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3173205, end_position_on_the_genomic_accession=3173840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318890, CurrentID=0, Status='live', Symbol='p24-2', Aliases='Dmel_CG33105, CG33105, CG9443, Dmel\\CG33105, p24', description='p24-related-2', other_designations='CG33105-PA|CG33105-PB|p24-2-PA|p24-2-PB', map_location='85E4-85E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9687402, end_position_on_the_genomic_accession=9692794, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  326215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326215, CurrentID=0, Status='live', Symbol='SMC5', Aliases='Dmel_CG32438, AAF51749, CG32438, CG7783, DmSMC5, Dmel\\CG32438, Smc5', description='Structural maintenance of chromosomes 5', other_designations='CG32438-PE|CG32438-PF|CG32438-PG|CG32438-PH|SMC5-PE|SMC5-PF|SMC5-PG|SMC5-PH|Smc5|structural maintenance of chromosomes 5', map_location='78D6-78D7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21569176, end_position_on_the_genomic_accession=21573523, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  318892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318892, CurrentID=0, Status='live', Symbol='CG33125', Aliases='Dmel_ Dmel\\CG33125', description='CG33125 gene product from transcript CG33125-RC', other_designations='CG33125-PB|CG33125-PC|uncharacterized protein', map_location='23A2-23A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2648360, end_position_on_the_genomic_accession=2652414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318893 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318893, CurrentID=0, Status='live', Symbol='CG33127', Aliases='Dmel_ Dmel\\CG33127', description='CG33127 gene product from transcript CG33127-RA', other_designations='CG33127-PA|uncharacterized protein', map_location='21B8-21B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=322101, end_position_on_the_genomic_accession=323091, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318894 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318894, CurrentID=0, Status='live', Symbol='CG33136', Aliases='Dmel_ Dmel\\CG33136, NEST:bs04e10', description='CG33136 gene product from transcript CG33136-RC', other_designations='CG33136-PA|CG33136-PB|CG33136-PC|uncharacterized protein', map_location='55E6-55E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18664449, end_position_on_the_genomic_accession=18665792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326216, CurrentID=0, Status='live', Symbol='Atox1', Aliases='Dmel_CG32446, CG12562, CG32446, Dm Dmel\\CG32446', description='CG32446 gene product from transcript CG32446-RA', other_designations='Atox1-PA|Atox1-PB|CG32446-PA|CG32446-PB', map_location='78E2-78E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21640995, end_position_on_the_genomic_accession=21642599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318901 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318901, CurrentID=0, Status='live', Symbol='CG33159', Aliases='Dmel_ CG11532, Dmel\\CG33159, SP176', description='CG33159 gene product from transcript CG33159-RA', other_designations='CG33159-PA|uncharacterized protein', map_location='63B12-63B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3203675, end_position_on_the_genomic_accession=3204448, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318904, CurrentID=0, Status='live', Symbol='CG31702', Aliases='Dmel_ Dmel\\CG31702', description='CG31702 gene product from transcript CG31702-RA', other_designations='CG31702-PA|uncharacterized protein', map_location='40E1-40E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22071144, end_position_on_the_genomic_accession=22072902, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318905, CurrentID=0, Status='live', Symbol='CG31703', Aliases='Dmel_ Dmel\\CG31703', description='CG31703 gene product from transcript CG31703-RA', other_designations='CG31703-PA|uncharacterized protein', map_location='40E3-40E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22102626, end_position_on_the_genomic_accession=22104401, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318906 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318906, CurrentID=0, Status='live', Symbol='CG31704', Aliases='Dmel_ Dmel\\CG31704', description='CG31704 gene product from transcript CG31704-RB', other_designations='CG31704-PA|CG31704-PB|uncharacterized protein', map_location='33A3-33A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11840691, end_position_on_the_genomic_accession=11841204, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772319, CurrentID=0, Status='live', Symbol='His4:CG33879', Aliases='Dmel_CG33879, CG33879, Dmel\\CG33879', description='CG33879 gene product from transcript CG33879-RA', other_designations='CG33879-PA|His4:CG33879-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21509422, end_position_on_the_genomic_accession=21509834, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318908, CurrentID=0, Status='live', Symbol='Apf', Aliases='Dmel_CG31713, CG31713, Dmel\\CG31713, IP04585', description='diadenosine tetraphosphate hydrolase', other_designations='Apf-PA|CG31713-PA', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9938178, end_position_on_the_genomic_accession=9938790, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318909, CurrentID=0, Status='live', Symbol='CG31715', Aliases='Dmel_ BcDNA:RE44624, Dmel\\CG31715', description='CG31715 gene product from transcript CG31715-RA', other_designations='CG31715-PA|CG31715-PB|uncharacterized protein', map_location='31D9-31D9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10341616, end_position_on_the_genomic_accession=10342523, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318910 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318910, CurrentID=0, Status='live', Symbol='CG43672', Aliases='Dmel_ CG33173, CG8473, Dmel\\CG43672, Dmel_CG33173', description='CG43672 gene product from transcript CG43672-RA', other_designations='CG43672-PA|uncharacterized protein', map_location='13E16-13E17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15773171, end_position_on_the_genomic_accession=15778703, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  318911 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318911, CurrentID=0, Status='live', Symbol='CG31717', Aliases='Dmel_ BcDNA:AT30094, Css2, Dmel\\CG31717', description='CG31717 gene product from transcript CG31717-RA', other_designations='CG31717-PA|CG31717-PB|uncharacterized protein', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10250737, end_position_on_the_genomic_accession=10251714, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318912, CurrentID=0, Status='live', Symbol='Ir31a', Aliases='Dmel_CG31718, CG31718, DmelIR31a, Dmel\\CG31718, IR31a', description='Ionotropic receptor 31a', other_designations='CG31718-PC|CG31718-PD|Ir31a-PC|Ir31a-PD|ionotropic receptor 31a', map_location='31E1-31E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10405565, end_position_on_the_genomic_accession=10408197, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  318913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318913, CurrentID=0, Status='live', Symbol='CG33178', Aliases='Dmel_ BcDNA:RH14671, CG5579, Dmel\\CG33178', description='CG33178 gene product from transcript CG33178-RC', other_designations='CG33178-PA|CG33178-PB|CG33178-PC|uncharacterized protein', map_location='13A9-13A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15082827, end_position_on_the_genomic_accession=15084017, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318914, CurrentID=0, Status='live', Symbol='mthl15', Aliases='Dmel_CG31720, CG31720, Dmel\\CG31720', description='methuselah-like 15', other_designations='CG31720-PC|CG31720-PD|mthl15-PC|mthl15-PD', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10234585, end_position_on_the_genomic_accession=10238317, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  318916 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318916, CurrentID=0, Status='live', Symbol='CG33181', Aliases='Dmel_ CG1521, CG32714, Dmel\\CG33181', description='CG33181 gene product from transcript CG33181-RF', other_designations='CG33181-PC|CG33181-PD|CG33181-PE|CG33181-PF|CG33181-PG|CG33181-PH|CG33181-PI|CG33181-PJ|uncharacterized protein', map_location='7F1-7F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8489348, end_position_on_the_genomic_accession=8517029, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  318918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318918, CurrentID=0, Status='live', Symbol='Kdm4B', Aliases='Dmel_CG33182, CG33182, CG4037, Dmel\\CG33182, KDM4B, dJMJD2(2), dKDM4B, dKdm4B', description='Lysine (K)-specific demethylase 4B', other_designations='CG33182-PA|CG33182-PB|CG33182-PC|CG33182-PD|CG33182-PE|Kdm4B-PA|Kdm4B-PB|Kdm4B-PC|Kdm4B-PD|Kdm4B-PE|histone demethylase 4B|lysine (K)-specific demethylase 4B', map_location='49F7-49F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13176553, end_position_on_the_genomic_accession=13198337, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  318920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318920, CurrentID=0, Status='live', Symbol='CG31730', Aliases='Dmel_ Dmel\\CG31730', description='CG31730 gene product from transcript CG31730-RA', other_designations='CG31730-PA|uncharacterized protein', map_location='34B3-34B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13259064, end_position_on_the_genomic_accession=13259743, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326220, CurrentID=0, Status='live', Symbol='CG32488', Aliases='Dmel_ 20151345, CG2077, Dmel\\CG32488, Phos32488', description='CG32488 gene product from transcript CG32488-RB', other_designations='CG32488-PA|CG32488-PB|uncharacterized protein', map_location='63A3-63A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3014761, end_position_on_the_genomic_accession=3016069, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318922, CurrentID=0, Status='live', Symbol='CG31735', Aliases='Dmel_ BcDNA:AT20140, Dmel\\CG31735', description='CG31735 gene product from transcript CG31735-RA', other_designations='CG31735-PA|CG31735-PC|uncharacterized protein', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15895454, end_position_on_the_genomic_accession=15897071, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  318924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318924, CurrentID=0, Status='live', Symbol='Prosbeta5R2', Aliases='Dmel_CG31742, 31742, CG31742, Dmel\\CG31742, anon-AE003657.1, beta5R2', description='Proteasome beta5 subunit-related 2', other_designations='AE003657 - proteasome beta subunit|CG31742-PA|CG31742-PB|Prosbeta5R2-PA|Prosbeta5R2-PB|proteasome beta5 subunit-related 2', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17990418, end_position_on_the_genomic_accession=17991951, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772322, CurrentID=0, Status='live', Symbol='CG33687', Aliases='Dmel_ Dmel\\CG33687', description='CG33687 gene product from transcript CG33687-RB', other_designations='CG33687-PB|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16197868, end_position_on_the_genomic_accession=16198852, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326221, CurrentID=0, Status='live', Symbol='CG32557', Aliases='Dmel_ CG7871, Dmel\\CG32557, MCO4', description='CG32557 gene product from transcript CG32557-RA', other_designations='CG32557-PA|uncharacterized protein', map_location='16E1-16E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17852958, end_position_on_the_genomic_accession=17856129, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318930, CurrentID=0, Status='live', Symbol='NimA', Aliases='Dmel_CG42282, BG:DS00180.5, CG31765, CG31841, CG42282, CG8855, CT25428, DS00180.5, Dmel\\CG42282, Dmel_CG31765, Dmel_CG31841, j, nimA', description='Nimrod A', other_designations='CG42282-PD|CG42282-PE|CG42282-PF|NimA-PD|NimA-PE|NimA-PF|jaunty|nimrod A', map_location='2-48.7 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13957539, end_position_on_the_genomic_accession=13963381, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  326222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326222, CurrentID=0, Status='live', Symbol='CG32564', Aliases='Dmel_ CG5095, Dmel\\CG32564', description='CG32564 gene product from transcript CG32564-RA', other_designations='CG32564-PA|uncharacterized protein', map_location='15F3-15F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17206407, end_position_on_the_genomic_accession=17207854, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318934, CurrentID=0, Status='live', Symbol='CG31769', Aliases='Dmel_ BcDNA:RH26533, Dmel\\CG31769', description='CG31769 gene product from transcript CG31769-RA', other_designations='CG31769-PA|uncharacterized protein', map_location='34F4-34F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14109597, end_position_on_the_genomic_accession=14110739, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318935, CurrentID=0, Status='live', Symbol='CG31775', Aliases='Dmel_ BG:DS00810.2, BcDNA:LP08165, BcDNA:RE38252, BcDNA:RH25462, Dmel\\CG31775', description='CG31775 gene product from transcript CG31775-RA', other_designations='CG31775-PA|CG31775-PB|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14737090, end_position_on_the_genomic_accession=14737769, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318936 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318936, CurrentID=0, Status='live', Symbol='Acp24A4', Aliases='Dmel_CG31779, CG31779, Dmel\\CG31779', description='Accessory gland protein 24A4', other_designations='Acp24A4-PB|Acp24A4-PC|CG31779-PB|CG31779-PC|accessory gland protein 24A4', map_location='24B1-24B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3693184, end_position_on_the_genomic_accession=3693905, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318937, CurrentID=0, Status='live', Symbol='CG31780', Aliases='Dmel_ Dmel\\CG31780', description='CG31780 gene product from transcript CG31780-RB', other_designations='CG31780-PB|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15696038, end_position_on_the_genomic_accession=15697750, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31311, CurrentID=0, Status='live', Symbol='CG12535', Aliases='Dmel_ Dmel\\CG12535', description='CG12535 gene product from transcript CG12535-RF', other_designations='CG12535-PB|CG12535-PC|CG12535-PD|CG12535-PE|CG12535-PF|uncharacterized protein', map_location='3D2-3D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3387944, end_position_on_the_genomic_accession=3448943, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  318940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318940, CurrentID=0, Status='live', Symbol='CG31785', Aliases='Dmel_ Dmel\\CG31785', description='CG31785 gene product from transcript CG31785-RB', other_designations='CG31785-PB|uncharacterized protein', map_location='36E5-36E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18124098, end_position_on_the_genomic_accession=18125749, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318941 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318941, CurrentID=0, Status='live', Symbol='CG31787', Aliases='Dmel_ Dmel\\CG31787, p24', description='CG31787 gene product from transcript CG31787-RA', other_designations='CG31787-PA|uncharacterized protein', map_location='36F2-36F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18284845, end_position_on_the_genomic_accession=18285720, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318942, CurrentID=0, Status='live', Symbol='CG31788', Aliases='Dmel_ Dmel\\CG31788', description='CG31788 gene product from transcript CG31788-RB', other_designations='CG31788-PA|CG31788-PB|uncharacterized protein', map_location='36F1-36F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18257494, end_position_on_the_genomic_accession=18257978, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772325, CurrentID=0, Status='live', Symbol='His4:CG33895', Aliases='Dmel_CG33895, CG33895, Dmel\\CG33895', description='CG33895 gene product from transcript CG33895-RA', other_designations='CG33895-PA|His4:CG33895-PA|histone H4', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21468956, end_position_on_the_genomic_accession=21469368, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31312, CurrentID=0, Status='live', Symbol='CG14269', Aliases='Dmel_ Dmel\\CG14269', description='CG14269 gene product from transcript CG14269-RB', other_designations='CG14269-PA|CG14269-PB|uncharacterized protein', map_location='3D3-3D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3434943, end_position_on_the_genomic_accession=3435807, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772172, CurrentID=0, Status='live', Symbol='His4:CG33885', Aliases='Dmel_CG33885, CG33885, Dmel\\CG33885', description='CG33885 gene product from transcript CG33885-RA', other_designations='CG33885-PA|His4:CG33885-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21494279, end_position_on_the_genomic_accession=21494691, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318947, CurrentID=0, Status='live', Symbol='CG31800', Aliases='Dmel_ Dmel\\CG31800', description='CG31800 gene product from transcript CG31800-RA', other_designations='CG31800-PA|uncharacterized protein', map_location='37B10-37B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19033657, end_position_on_the_genomic_accession=19034576, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318948, CurrentID=0, Status='live', Symbol='Mst36Fa', Aliases='Dmel_CG31801, CG15165, CG31801, Dmel\\CG31801, mst36Fa', description='Male-specific transcript 36Fa', other_designations='CG31801-PC|Mst36Fa-PC', map_location='36F10-36F10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18675183, end_position_on_the_genomic_accession=18676998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  318949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318949, CurrentID=0, Status='live', Symbol='CG31802', Aliases='Dmel_ BcDNA:AT22559, CG15158, Dmel\\CG31802', description='CG31802 gene product from transcript CG31802-RA', other_designations='CG31802-PA|uncharacterized protein', map_location='36F1-36F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18208415, end_position_on_the_genomic_accession=18209237, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318950, CurrentID=0, Status='live', Symbol='CG31803', Aliases='Dmel_ BcDNA:AT29119, Dmel\\CG31803', description='CG31803 gene product from transcript CG31803-RA', other_designations='CG31803-PA|uncharacterized protein', map_location='36C7-36C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17411649, end_position_on_the_genomic_accession=17412633, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318951 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318951, CurrentID=0, Status='live', Symbol='CG31804', Aliases='Dmel_ Dmel\\CG31804, NEST:bs12c10, NEST:bs13h04, NEST:bs27h11, NEST:bs34f07', description='CG31804 gene product from transcript CG31804-RA', other_designations='CG31804-PA|uncharacterized protein', map_location='36C6-36C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17363496, end_position_on_the_genomic_accession=17364351, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318952, CurrentID=0, Status='live', Symbol='CG31805', Aliases='Dmel_ Dmel\\CG31805', description='CG31805 gene product from transcript CG31805-RB', other_designations='CG31805-PB|CG31805-PC|CG31805-PD|uncharacterized protein', map_location='36B4-36B4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16898175, end_position_on_the_genomic_accession=16900473, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318953, CurrentID=0, Status='live', Symbol='CG31807', Aliases='Dmel_ Dmel\\CG31807', description='CG31807 gene product from transcript CG31807-RB', other_designations='CG31807-PA|CG31807-PB|uncharacterized protein', map_location='36A10-36A10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16681819, end_position_on_the_genomic_accession=16684008, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318954, CurrentID=0, Status='live', Symbol='CG31809', Aliases='Dmel_ Dmel\\CG31809', description='CG31809 gene product from transcript CG31809-RB', other_designations='CG31809-PB|CG31809-PC|uncharacterized protein', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16853620, end_position_on_the_genomic_accession=16858714, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318955, CurrentID=0, Status='live', Symbol='CG31810', Aliases='Dmel_ BcDNA:RE06583, Dmel\\CG31810', description='CG31810 gene product from transcript CG31810-RA', other_designations='CG31810-PA|uncharacterized protein', map_location='36B2-36B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16849018, end_position_on_the_genomic_accession=16852368, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318956 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318956, CurrentID=0, Status='live', Symbol='CG31812', Aliases='Dmel_ Dmel\\CG31812', description='CG31812 gene product from transcript CG31812-RB', other_designations='CG31812-PB|uncharacterized protein', map_location='36A12-36A12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16729777, end_position_on_the_genomic_accession=16731199, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318957, CurrentID=0, Status='live', Symbol='CG31813', Aliases='Dmel_ BG:DS08249.4, BcDNA:RE62259, Dmel\\CG31813', description='CG31813 gene product from transcript CG31813-RB', other_designations='CG31813-PA|CG31813-PB|uncharacterized protein', map_location='34D1-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13736138, end_position_on_the_genomic_accession=13737030, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318958, CurrentID=0, Status='live', Symbol='CG31814', Aliases='Dmel_ 31814, BG:DS08249.1, Dmel\\CG31814', description='CG31814 gene product from transcript CG31814-RA', other_designations='CG31814-PA|CG31814-PB|uncharacterized protein', map_location='34C6-34D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13676860, end_position_on_the_genomic_accession=13717320, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  318959 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318959, CurrentID=0, Status='live', Symbol='CG31815', Aliases='Dmel_ Dmel\\CG31815', description='CG31815 gene product from transcript CG31815-RA', other_designations='CG31815-PA|uncharacterized protein', map_location='36A6-36A7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16605822, end_position_on_the_genomic_accession=16607321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318960, CurrentID=0, Status='live', Symbol='CG31816', Aliases='Dmel_ Dmel\\CG31816, NEST:bs11b11, NEST:bs19f07, NEST:bs34h07', description='CG31816 gene product from transcript CG31816-RA', other_designations='CG31816-PA|uncharacterized protein', map_location='36A4-36A4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16542546, end_position_on_the_genomic_accession=16543263, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318961, CurrentID=0, Status='live', Symbol='CG31820', Aliases='Dmel_ Dmel\\CG31820, NEST:bs10d03, NEST:bs36g10', description='CG31820 gene product from transcript CG31820-RB', other_designations='CG31820-PA|CG31820-PB|uncharacterized protein', map_location='35E1-35E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16005789, end_position_on_the_genomic_accession=16006279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326227, CurrentID=0, Status='live', Symbol='CG32650', Aliases='Dmel_ CG4410, Dmel\\CG32650', description='CG32650 gene product from transcript CG32650-RA', other_designations='CG32650-PA|uncharacterized protein', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12897503, end_position_on_the_genomic_accession=12898669, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318965, CurrentID=0, Status='live', Symbol='CG31827', Aliases='Dmel_ Dmel\\CG31827', description='CG31827 gene product from transcript CG31827-RA', other_designations='CG31827-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15705171, end_position_on_the_genomic_accession=15706220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318970, CurrentID=0, Status='live', Symbol='CG31832', Aliases='Dmel_ AAF44911, BG:DS00929.7, DS00929.7, Dmel\\CG31832, FBpp0080285', description='CG31832 gene product from transcript CG31832-RB', other_designations='CG31832-PB|Fibrinogen-related protein|uncharacterized protein', map_location='35B8-35B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15030928, end_position_on_the_genomic_accession=15031750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318972, CurrentID=0, Status='live', Symbol='CG31835', Aliases='Dmel_ BG:DS00810.1, Dmel\\CG31835', description='CG31835 gene product from transcript CG31835-RA', other_designations='CG31835-PA|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14735843, end_position_on_the_genomic_accession=14737067, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326229, CurrentID=0, Status='live', Symbol='CG32669', Aliases='Dmel_ Dmel\\CG32669, dSLC5A4', description='CG32669 gene product from transcript CG32669-RA', other_designations='CG32669-PA|uncharacterized protein', map_location='9F5-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10842697, end_position_on_the_genomic_accession=10845081, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318977, CurrentID=0, Status='live', Symbol='mRpS23', Aliases='Dmel_CG31842, CG31842, Dmel\\CG31842, l(2)SH0272, l(2)SH2 0272', description='mitochondrial ribosomal protein S23', other_designations='CG31842-PA|lethal (2) SH0272|mRpS23-PA', map_location='34D6-34D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13834250, end_position_on_the_genomic_accession=13835072, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318980, CurrentID=0, Status='live', Symbol='CG31848', Aliases='Dmel_ Dmel\\CG31848', description='CG31848 gene product from transcript CG31848-RB', other_designations='CG31848-PB|uncharacterized protein', map_location='34B3-34B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13264996, end_position_on_the_genomic_accession=13265876, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31318, CurrentID=0, Status='live', Symbol='CG10802', Aliases='Dmel_ Dmel\\CG10802, anon-EST:Posey113', description='CG10802 gene product from transcript CG10802-RA', other_designations='CG10802-PA|CG10802-PB|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3482965, end_position_on_the_genomic_accession=3484555, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318982, CurrentID=0, Status='live', Symbol='CG31851', Aliases='Dmel_ CG16821, Dmel\\CG31851', description='CG31851 gene product from transcript CG31851-RB', other_designations='CG31851-PA|CG31851-PB|uncharacterized protein', map_location='34B2-34B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13257477, end_position_on_the_genomic_accession=13258644, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318983, CurrentID=0, Status='live', Symbol='CG31855', Aliases='Dmel_ CG6537, Dmel\\CG31855', description='CG31855 gene product from transcript CG31855-RB', other_designations='CG31855-PA|CG31855-PB|uncharacterized protein', map_location='34B8-34B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13369204, end_position_on_the_genomic_accession=13370329, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318986, CurrentID=0, Status='live', Symbol='t-cup', Aliases='Dmel_CG31858, CG31858, Dmel\\CG31858, cup', description='tetleys-cup', other_designations='CG31858-PA|t-cup-PA', map_location='33E7-33E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12496616, end_position_on_the_genomic_accession=12497333, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  318988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318988, CurrentID=0, Status='live', Symbol='CG31861', Aliases='Dmel_ CG17215, Dmel\\CG31861', description='CG31861 gene product from transcript CG31861-RA', other_designations='CG31861-PA|uncharacterized protein', map_location='33D2-33D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12161767, end_position_on_the_genomic_accession=12162605, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318989, CurrentID=0, Status='live', Symbol='CG31862', Aliases='Dmel_ Dmel\\CG31862', description='CG31862 gene product from transcript CG31862-RA', other_designations='CG31862-PA|uncharacterized protein', map_location='33E1-33E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12338848, end_position_on_the_genomic_accession=12339477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318990, CurrentID=0, Status='live', Symbol='Qtzl', Aliases='Dmel_CG31864, CG17749, CG31864, CG31864-PA, Dmel\\CG31864', description='Quetzalcoatl', other_designations='CG31864-PA|CG31864-PC|Qtzl-PA|Qtzl-PC|quetzalcoatl', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11995740, end_position_on_the_genomic_accession=11996723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318991, CurrentID=0, Status='live', Symbol='Ada1-1', Aliases='Dmel_CG31865, Ada1, BcDNA:RE30472, CG31865, Dmel\\CG31865, ada1-1', description='transcriptional Adaptor 1-1', other_designations='Ada1|Ada1-1-PA|Ada1-1-PB|CG31865-PA|CG31865-PB|transcriptional adaptor 1-1', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11996809, end_position_on_the_genomic_accession=11997965, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318992 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318992, CurrentID=0, Status='live', Symbol='Ada1-2', Aliases='Dmel_CG31866, Ada1, CG31866, Dmel\\CG31866, ada1-2', description='transcriptional Adaptor 1-2', other_designations='Ada1|Ada1-2-PA|CG31866-PA|transcriptional adaptor 1-2', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11992902, end_position_on_the_genomic_accession=11994054, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31320, CurrentID=0, Status='live', Symbol='CG10803', Aliases='Dmel_ Dmel\\CG10803', description='CG10803 gene product from transcript CG10803-RE', other_designations='CG10803-PA|CG10803-PB|CG10803-PC|CG10803-PD|CG10803-PE|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3485650, end_position_on_the_genomic_accession=3490915, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  318995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318995, CurrentID=0, Status='live', Symbol='CG31874', Aliases='Dmel_ Dmel\\CG31874, NEST:bs10e11', description='CG31874 gene product from transcript CG31874-RA', other_designations='CG31874-PA|uncharacterized protein', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10189820, end_position_on_the_genomic_accession=10191496, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318996, CurrentID=0, Status='live', Symbol='CG31875', Aliases='Dmel_ BcDNA:AT29184, CG5831, Desr, Dmel\\CG31875', description='CG31875 gene product from transcript CG31875-RB', other_designations='CG31875-PA|CG31875-PB|CG31875-PC|CG31875-PD|drosophila Elm2-Sant retrogene|uncharacterized protein', map_location='30F5-30F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9995561, end_position_on_the_genomic_accession=10002736, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  318997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318997, CurrentID=0, Status='live', Symbol='Cpr30F', Aliases='Dmel_CG31876, CG31876, CG5879, DmelCpr30F, Dmel\\CG31876', description='Cuticular protein 30F', other_designations='CG31876-PA|Cpr30F-PA|cuticular protein 30F', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9932272, end_position_on_the_genomic_accession=9933163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  318998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318998, CurrentID=0, Status='live', Symbol='CG42367', Aliases='Dmel_ CG31878, CG4588, Dmel\\CG42367, Dmel_CG31711, Dmel_CG31878', description='CG42367 gene product from transcript CG42367-RC', other_designations='CG42367-PC|uncharacterized protein', map_location='30F1-30F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9930580, end_position_on_the_genomic_accession=9931131, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  319000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319000, CurrentID=0, Status='live', Symbol='CG31882', Aliases='Dmel_ Dmel\\CG31882, NEST:bs35a07, bs35a07.y1', description='CG31882 gene product from transcript CG31882-RA', other_designations='CG31882-PA|uncharacterized protein', map_location='30C1-30C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9624452, end_position_on_the_genomic_accession=9625285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31322, CurrentID=0, Status='live', Symbol='DIP-alpha', Aliases='Dmel_CG32791, 32791, CG13019, CG13020, CG32791, Dmel\\CG32791', description='Dpr-interacting protein alpha', other_designations='CG32791-PA|CG32791-PB|CG32791-PC|DIP-alpha-PA|DIP-alpha-PB|DIP-alpha-PC', map_location='3D4-3D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3507245, end_position_on_the_genomic_accession=3571649, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  3346207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346207, CurrentID=0, Status='live', Symbol='CG33543', Aliases='Dmel_ 33543, CG15354, CG15355, CT35363, CT35364, Dmel\\CG33543', description='CG33543 gene product from transcript CG33543-RD', other_designations='CG33543-PD|uncharacterized protein', map_location='22B8-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1997102, end_position_on_the_genomic_accession=2006313, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  3772336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772336, CurrentID=0, Status='live', Symbol='His2B:CG33890', Aliases='Dmel_CG33890, CG33890, Dmel\\CG33890', description='CG33890 gene product from transcript CG33890-RA', other_designations='CG33890-PA|His2B:CG33890-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21482504, end_position_on_the_genomic_accession=21483098, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319011, CurrentID=0, Status='live', Symbol='CG31897', Aliases='Dmel_ Dmel\\CG31897, NEST:bs30d08', description='CG31897 gene product from transcript CG31897-RA', other_designations='CG31897-PA|uncharacterized protein', map_location='29C5-29C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8442447, end_position_on_the_genomic_accession=8445815, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319012, CurrentID=0, Status='live', Symbol='CG31898', Aliases='Dmel_ BcDNA:LD28680, Dmel\\CG31898', description='CG31898 gene product from transcript CG31898-RA', other_designations='CG31898-PA|uncharacterized protein', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8400421, end_position_on_the_genomic_accession=8401592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319014, CurrentID=0, Status='live', Symbol='Mur29B', Aliases='Dmel_CG31901, BcDNA:LP08232, CG31901, Dmel\\CG31901', description='Mucin related 29B', other_designations='CG31901-PA|Mur29B-PA|mucin related 29B', map_location='29B1-29B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8352187, end_position_on_the_genomic_accession=8355172, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  319016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319016, CurrentID=0, Status='live', Symbol='CG31904', Aliases='Dmel_CG31904-PD, Dmel\\CG31904, anon-WO0149848.5', description='CG31904 gene product from transcript CG31904-RD', other_designations='CG31904-PB|CG31904-PD|uncharacterized protein', map_location='28C2-28C2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7734415, end_position_on_the_genomic_accession=7738809, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  319019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319019, CurrentID=0, Status='live', Symbol='CG31907', Aliases='Dmel_ Dmel\\CG31907, NEST:bs06f02, NEST:bs25f07, NEST:bs29b10', description='CG31907 gene product from transcript CG31907-RB', other_designations='CG31907-PA|CG31907-PB|uncharacterized protein', map_location='27D7-27D7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7059214, end_position_on_the_genomic_accession=7060851, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319020, CurrentID=0, Status='live', Symbol='CG31909', Aliases='Dmel_ Dmel\\CG31909, NEST:bs19e08', description='CG31909 gene product from transcript CG31909-RC', other_designations='CG31909-PB|CG31909-PC|uncharacterized protein', map_location='27E7-27E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7257512, end_position_on_the_genomic_accession=7258198, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  319021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319021, CurrentID=0, Status='live', Symbol='CG31910', Aliases='Dmel_ BcDNA:AT21561, Dmel\\CG31910, NEST:bs09e03', description='CG31910 gene product from transcript CG31910-RA', other_designations='CG31910-PA|uncharacterized protein', map_location='27A2-27A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6744663, end_position_on_the_genomic_accession=6745622, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319023, CurrentID=0, Status='live', Symbol='CG31913', Aliases='Dmel_ Dmel\\CG31913, NEST:bs09g05, NEST:bs14h07, NEST:bs23g06, NEST:bs28e01, NEST:bs35c08', description='CG31913 gene product from transcript CG31913-RA', other_designations='CG31913-PA|uncharacterized protein', map_location='25F4-25F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5784031, end_position_on_the_genomic_accession=5784893, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319025, CurrentID=0, Status='live', Symbol='CG31915', Aliases='Dmel_ CG6922, Dmel\\CG31915', description='CG31915 gene product from transcript CG31915-RA', other_designations='CG31915-PA|uncharacterized protein', map_location='25E5-25E5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5528123, end_position_on_the_genomic_accession=5530682, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  319026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319026, CurrentID=0, Status='live', Symbol='CG43755', Aliases='Dmel_ BcDNA:GH10778, CG31920, CG33992, Dmel\\CG43755, Dmel_CG31920, Dmel_CG33992, NEST:bs21f09, NEST:bs23f04, anon-WO0140519.62', description='CG43755 gene product from transcript CG43755-RD', other_designations='CG43755-PB|CG43755-PD|uncharacterized protein', map_location='21C1-21C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=325522, end_position_on_the_genomic_accession=329761, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740147, CurrentID=0, Status='live', Symbol='CG17159', Aliases='Dmel_ CG40460, Dmel\\CG17159, FBgn0058460, anon-20Ea', description='CG17159 gene product from transcript CG17159-RF', other_designations='CG17159-PB|CG17159-PF|uncharacterized protein', map_location='20E-20E', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23084960, end_position_on_the_genomic_accession=23092631, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  319028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319028, CurrentID=0, Status='live', Symbol='CG31922', Aliases='Dmel_ 20K, Dmel\\CG31922, anon-AE003587.1', description='CG31922 gene product from transcript CG31922-RA', other_designations='CG31922-PA|uncharacterized protein', map_location='21F1-21F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1170181, end_position_on_the_genomic_accession=1170789, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326238, CurrentID=0, Status='live', Symbol='CG32750', Aliases='Dmel_ CG3648, Dmel\\CG32750', description='CG32750 gene product from transcript CG32750-RA', other_designations='CG32750-PA|pantetheine hydrolase|uncharacterized protein', map_location='5D8-5D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6194737, end_position_on_the_genomic_accession=6196439, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319030, CurrentID=0, Status='live', Symbol='CG31924', Aliases='Dmel_ 21064207, BcDNA:AT29272, CG5564, Dmel\\CG31924', description='CG31924 gene product from transcript CG31924-RB', other_designations='CG31924-PB|uncharacterized protein', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1354101, end_position_on_the_genomic_accession=1354919, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319033, CurrentID=0, Status='live', Symbol='Rrp40', Aliases='Dmel_CG31938, CG31938, Dmel\\CG31938, d', description='CG31938 gene product from transcript CG31938-RB', other_designations='CG31938-PA|CG31938-PB|Rrp40-PA|Rrp40-PB', map_location='22B1-22B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1728732, end_position_on_the_genomic_accession=1729762, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326240, CurrentID=0, Status='live', Symbol='CG32803', Aliases='Dmel_ CG14819, Dmel\\CG32803', description='CG32803 gene product from transcript CG32803-RD', other_designations='CG32803-PA|CG32803-PC|CG32803-PD|uncharacterized protein', map_location='2B14-2B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1900699, end_position_on_the_genomic_accession=1901911, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  319042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319042, CurrentID=0, Status='live', Symbol='CG31949', Aliases='Dmel_ Dmel\\CG31949', description='CG31949 gene product from transcript CG31949-RB', other_designations='CG31949-PA|CG31949-PB|uncharacterized protein', map_location='22E1-22E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2300103, end_position_on_the_genomic_accession=2300888, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  319043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319043, CurrentID=0, Status='live', Symbol='CG31950', Aliases='Dmel_ Dmel\\CG31950', description='CG31950 gene product from transcript CG31950-RB', other_designations='CG31950-PA|CG31950-PB|uncharacterized protein', map_location='23C1-23C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2878344, end_position_on_the_genomic_accession=2879001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319045, CurrentID=0, Status='live', Symbol='CG31955', Aliases='Dmel_ Dmel\\CG31955', description='CG31955 gene product from transcript CG31955-RC', other_designations='CG31955-PA|CG31955-PB|CG31955-PC|CG31955-PD|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3711881, end_position_on_the_genomic_accession=3713105, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  319046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319046, CurrentID=0, Status='live', Symbol='CG31957', Aliases='Dmel_ Dmel\\CG31957', description='CG31957 gene product from transcript CG31957-RB', other_designations='CG31957-PA|CG31957-PB|uncharacterized protein', map_location='24C8-24C8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3871271, end_position_on_the_genomic_accession=3872458, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319047, CurrentID=0, Status='live', Symbol='CG31960', Aliases='Dmel_ CG10022, Dmel\\CG31960', description='CG31960 gene product from transcript CG31960-RA', other_designations='CG31960-PA|uncharacterized protein', map_location='24C4-24C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3784947, end_position_on_the_genomic_accession=3785622, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319048, CurrentID=0, Status='live', Symbol='CG31961', Aliases='Dmel_ BcDNA:LD34582, CG3381, Dmel\\CG31961', description='CG31961 gene product from transcript CG31961-RA', other_designations='CG31961-PA|CG31961-PB|uncharacterized protein', map_location='24D8-24D8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4186250, end_position_on_the_genomic_accession=4187793, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835366, CurrentID=0, Status='live', Symbol='CG44355', Aliases='Dmel_ Dmel\\CG44355', description='CG44355 gene product from transcript CG44355-RA', other_designations='CG44355-PA|uncharacterized protein', map_location='69B1-69B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12310805, end_position_on_the_genomic_accession=12311961, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319052, CurrentID=0, Status='live', Symbol='CG31975', Aliases='Dmel_ CG18501, Dmel\\CG31975', description='CG31975 gene product from transcript CG31975-RC', other_designations='CG31975-PB|CG31975-PC|uncharacterized protein', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=138384, end_position_on_the_genomic_accession=140992, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  36792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36792, CurrentID=0, Status='live', Symbol='CG8441', Aliases='Dmel_ Dmel\\CG8441', description='CG8441 gene product from transcript CG8441-RB', other_designations='CG8441-PA|CG8441-PB', map_location='52F5-52F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16185122, end_position_on_the_genomic_accession=16186506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772345, CurrentID=0, Status='live', Symbol='His2A:CG33814', Aliases='Dmel_CG33814, CG33814, Dmel\\CG33814', description='CG33814 gene product from transcript CG33814-RA', other_designations='CG33814-PA|His2A:CG33814-PA|histone H2A', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21452934, end_position_on_the_genomic_accession=21453426, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=319064, CurrentID=0, Status='live', Symbol='CG31997', Aliases='Dmel_ BEST:GH25016, BcDNA:RE42475, Dmel\\CG31997', description='CG31997 gene product from transcript CG31997-RB', other_designations='CG31997-PA|CG31997-PB|uncharacterized protein', map_location='102A4-102A4', chromosome='4', genomic_nucleotide_accession_version='NC_004353.4', start_position_on_the_genomic_accession=154654, end_position_on_the_genomic_accession=156060, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326244, CurrentID=0, Status='live', Symbol='CG32985', Aliases='Dmel_ CG17789, Dmel\\CG32985, gi21430910, gi22945998', description='CG32985 gene product from transcript CG32985-RA', other_designations='CG32985-PA|uncharacterized protein', map_location='29F6-29F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8962533, end_position_on_the_genomic_accession=8964620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772346 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772346, CurrentID=0, Status='live', Symbol='His2A:CG33853', Aliases='Dmel_CG33853, CG33853, Dmel\\CG33853', description='CG33853 gene product from transcript CG33853-RA', other_designations='CG33853-PA|His2A:CG33853-PA|histone H2A', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21518655, end_position_on_the_genomic_accession=21519147, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772349, CurrentID=0, Status='live', Symbol='Tyler', Aliases='Dmel_CG14208, CG14208, CG33075, Dmel\\CG14208', description='CG14208 gene product from transcript CG14208-RF', other_designations='CG14208-PB|CG14208-PF|CG14208-PG|Tyler-PB|Tyler-PF|Tyler-PG|tyler', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19611526, end_position_on_the_genomic_accession=19615617, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  36799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36799, CurrentID=0, Status='live', Symbol='CG15706', Aliases='Dmel_ Dmel\\CG15706', description='CG15706 gene product from transcript CG15706-RA', other_designations='CG15706-PA|CG15706-PB|CG15706-PD|CG15706-PE|CG15706-PF|MFSD6|uncharacterized protein', map_location='52F11-52F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16223218, end_position_on_the_genomic_accession=16230680, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326250, CurrentID=0, Status='live', Symbol='CG34148', Aliases='Dmel_ Dmel\\CG34148', description='CG34148 gene product from transcript CG34148-RB', other_designations='CG34148-PB|CG34148-PD|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21877870, end_position_on_the_genomic_accession=21878686, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326251, CurrentID=0, Status='live', Symbol='CG33096', Aliases='Dmel_ CG11935, Dmel\\CG33096', description='CG33096 gene product from transcript CG33096-RG', other_designations='CG33096-PC|CG33096-PD|CG33096-PE|CG33096-PF|CG33096-PG|uncharacterized protein', map_location='96B21-96C1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25097672, end_position_on_the_genomic_accession=25102146, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  326253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326253, CurrentID=0, Status='live', Symbol='CG33098', Aliases='Dmel_ BcDNA:AT09395, BcDNA:RH08915, CG17756, CT39356, Dmel\\CG33098', description='CG33098 gene product from transcript CG33098-RF', other_designations='CG33098-PB|CG33098-PC|CG33098-PD|CG33098-PE|CG33098-PF|uncharacterized protein', map_location='87B9-87B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12387188, end_position_on_the_genomic_accession=12388794, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  326254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326254, CurrentID=0, Status='live', Symbol='CG33099', Aliases='Dmel_ CG5340, Dmel\\CG33099', description='CG33099 gene product from transcript CG33099-RA', other_designations='CG33099-PA|uncharacterized protein', map_location='94B2-94B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22528920, end_position_on_the_genomic_accession=22530572, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31343, CurrentID=0, Status='live', Symbol='CG15239', Aliases='Dmel_ Dmel\\CG15239', description='CG15239 gene product from transcript CG15239-RA', other_designations='CG15239-PA|uncharacterized protein', map_location='3F4-3F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3884168, end_position_on_the_genomic_accession=3890486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326256, CurrentID=0, Status='live', Symbol='CG33107', Aliases='Dmel_ CG5275, Dmel\\CG33107', description='CG33107 gene product from transcript CG33107-RA', other_designations='CG33107-PA|uncharacterized protein', map_location='94B5-94B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22583352, end_position_on_the_genomic_accession=22584718, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3772360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772360, CurrentID=0, Status='live', Symbol='His2A:CG33841', Aliases='Dmel_CG33841, CG33841, Dmel\\CG33841', description='CG33841 gene product from transcript CG33841-RA', other_designations='CG33841-PA|His2A:CG33841-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21498468, end_position_on_the_genomic_accession=21498960, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31347, CurrentID=0, Status='live', Symbol='CG2938', Aliases='Dmel_ Dmel\\CG2938', description='CG2938 gene product from transcript CG2938-RD', other_designations='CG2938-PB|CG2938-PC|CG2938-PD', map_location='3F7-3F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3932000, end_position_on_the_genomic_accession=3936469, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  326260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326260, CurrentID=0, Status='live', Symbol='CG33120', Aliases='Dmel_ CG10599, Dmel\\CG33120', description='CG33120 gene product from transcript CG33120-RA', other_designations='CG33120-PA|uncharacterized protein', map_location='37B1-37B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18841400, end_position_on_the_genomic_accession=18844579, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  326261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326261, CurrentID=0, Status='live', Symbol='Spn28Db', Aliases='Dmel_CG33121, CG12318, CG33121, Dmel\\CG33121', description='Serpin 28Db', other_designations='CG33121-PA|Spn28Db-PA|serpin 28Db', map_location='28D2-28D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7889288, end_position_on_the_genomic_accession=7890594, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31350, CurrentID=0, Status='live', Symbol='lva', Aliases='Dmel_CG6450, ABP2, CG#6450, CG6450, Dmel\\CG6450, LVA, LVA_DROME, Lava, Lva, Lvl, lava, lvl', description='lava lamp', other_designations='CG6450-PC|CG6450-PD|LAVA LAMP|Lava-Lamp|actin-binding protein 2|lavalamp|lva-PC|lva-PD', map_location='4A1-4A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3955543, end_position_on_the_genomic_accession=3965327, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31351, CurrentID=0, Status='live', Symbol='CG6428', Aliases='Dmel_ CG-6428, Dmel\\CG6428', description='CG6428 gene product from transcript CG6428-RA', other_designations='CG6428-PA', map_location='4A1-4A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3965435, end_position_on_the_genomic_accession=3968655, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  2769005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2769005, CurrentID=0, Status='live', Symbol='CG33233', Aliases='Dmel_ CG12036, Dmel\\CG33233', description='CG33233 gene product from transcript CG33233-RC', other_designations='CG33233-PC|CG33233-PE|uncharacterized protein', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2659639, end_position_on_the_genomic_accession=2661501, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31352, CurrentID=0, Status='live', Symbol='CG6414', Aliases='Dmel_ Dmel\\CG6414', description='CG6414 gene product from transcript CG6414-RB', other_designations='CG6414-PA|CG6414-PB', map_location='4A1-4A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3968788, end_position_on_the_genomic_accession=3971350, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31353, CurrentID=0, Status='live', Symbol='CG43689', Aliases='Dmel_ CG15376, CG32778, CG32790, CG43284, CG4985, Dmel\\CG43689, Dmel_CG15376, Dmel_CG32778, Dmel_CG32790, Dmel_CG43284', description='CG43689 gene product from transcript CG43689-RH', other_designations='CG43689-PC|CG43689-PD|CG43689-PE|CG43689-PF|CG43689-PG|CG43689-PH|uncharacterized protein', map_location='4A5-4A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4075978, end_position_on_the_genomic_accession=4092237, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  36815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36815, CurrentID=0, Status='live', Symbol='CG15705', Aliases='Dmel_ Dmel\\CG15705', description='CG15705 gene product from transcript CG15705-RA', other_designations='CG15705-PA|uncharacterized protein', map_location='53A5-53A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16280950, end_position_on_the_genomic_accession=16281629, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326267, CurrentID=0, Status='live', Symbol='CG33170', Aliases='Dmel_ BcDNA:AT30788, CG8244, Dmel\\CG33170', description='CG33170 gene product from transcript CG33170-RB', other_designations='CG33170-PA|CG33170-PB|CG33170-PC|uncharacterized protein', map_location='80B1-80B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22864600, end_position_on_the_genomic_accession=22867535, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  2769006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2769006, CurrentID=0, Status='live', Symbol='CG33234', Aliases='Dmel_ CG12036, Dmel\\CG33234', description='CG33234 gene product from transcript CG33234-RD', other_designations='CG33234-PD|CG33234-PE|uncharacterized protein', map_location='62E8-62E8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=2661622, end_position_on_the_genomic_accession=2663622, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  326270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326270, CurrentID=0, Status='live', Symbol='MtnD', Aliases='Dmel_CG33192, CG33192, Dmel\\CG33192, mtnD', description='Metallothionein D', other_designations='CG33192-PB|MtnD-PB|metallothionein|metallothionein D', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20534962, end_position_on_the_genomic_accession=20535301, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31361, CurrentID=0, Status='live', Symbol='dgt4', Aliases='Dmel_CG4865, BcDNA:SD26403, CG4865, Dgt4, Dmel\\CG4865, EG0007.10, EG:EG0007.10, msd3', description='dim gamma-tubulin 4', other_designations='CG4865-PA|CG4865-PB|CG4865-PC|dgt4-PA|dgt4-PB|dgt4-PC|mitotic spindle density 3', map_location='4A6-4A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4113701, end_position_on_the_genomic_accession=4114929, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36818, CurrentID=0, Status='live', Symbol='CG30099', Aliases='Dmel_ CG7750, Dmel\\CG30099', description='CG30099 gene product from transcript CG30099-RA', other_designations='CG30099-PA|uncharacterized protein', map_location='53B1-53B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16288208, end_position_on_the_genomic_accession=16288969, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772381, CurrentID=0, Status='live', Symbol='CG33758', Aliases='Dmel_ Dmel\\CG33758', description='CG33758 gene product from transcript CG33758-RA', other_designations='CG33758-PA|uncharacterized protein', map_location='45D8-45D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9453409, end_position_on_the_genomic_accession=9454055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  36830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36830, CurrentID=0, Status='live', Symbol='Nup62', Aliases='Dmel_CG6251, CG6251, Dmel\\CG6251, p62', description='Nucleoporin 62kD', other_designations='CG6251-PA|CG6251-PB|Nup62-PA|Nup62-PB|nucleoporin|nucleoporin 62|nucleoporin 62kD', map_location='53C1-53C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16344377, end_position_on_the_genomic_accession=16345828, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772383, CurrentID=0, Status='live', Symbol='CG15891', Aliases='Dmel_ CG15892, CG32918, Dmel\\CG15891', description='CG15891 gene product from transcript CG15891-RD', other_designations='CG15891-PA|CG15891-PD|uncharacterized protein', map_location='5E6-5E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6294517, end_position_on_the_genomic_accession=6295345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178873, CurrentID=0, Status='live', Symbol='CG42824', Aliases='Dmel_ CG34053, Dmel\\CG42824, Dmel_CG34053, HDC14119', description='CG42824 gene product from transcript CG42824-RA', other_designations='CG42824-PA|uncharacterized protein', map_location='90A6-90A6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17406048, end_position_on_the_genomic_accession=17406511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772390, CurrentID=0, Status='live', Symbol='CG33700', Aliases='Dmel_ Dmel\\CG33700', description='CG33700 gene product from transcript CG33700-RC', other_designations='CG33700-PC|uncharacterized protein', map_location='67B11-67B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9514119, end_position_on_the_genomic_accession=9525544, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3772393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772393, CurrentID=0, Status='live', Symbol='CG11566', Aliases='Dmel_ CG11569, CG32514, Dmel\\CG11566', description='CG11566 gene product from transcript CG11566-RA', other_designations='CG11566-PA|uncharacterized protein', map_location='19F1-19F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=21093844, end_position_on_the_genomic_accession=21111417, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  10178875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178875, CurrentID=0, Status='live', Symbol='CG42752', Aliases='Dmel_ Dmel\\CG42752', description='CG42752 gene product from transcript CG42752-RA', other_designations='CG42752-PA|uncharacterized protein', map_location='37A2-37A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18739320, end_position_on_the_genomic_accession=18740110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36844, CurrentID=0, Status='live', Symbol='S-Lap8', Aliases='Dmel_CG4439, CG4439, Dmel\\CG4439, S-LAP 8, anon-WO0140519.41', description='Sperm-Leucylaminopeptidase 8', other_designations='CG4439-PA|S-Lap8-PA', map_location='53C6-53C6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16491100, end_position_on_the_genomic_accession=16493233, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31383, CurrentID=0, Status='live', Symbol='CG3568', Aliases='Dmel_ Dmel\\CG3568', description='CG3568 gene product from transcript CG3568-RA', other_designations='CG3568-PA', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4542630, end_position_on_the_genomic_accession=4544497, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772219, CurrentID=0, Status='live', Symbol='CG33627', Aliases='Dmel_ Dmel\\CG33627', description='CG33627 gene product from transcript CG33627-RA', other_designations='CG33627-PA|uncharacterized protein', map_location='49A3-49A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12411011, end_position_on_the_genomic_accession=12411678, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10178877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178877, CurrentID=0, Status='live', Symbol='CG42818', Aliases='Dmel_ BG:DS07473.1, CG12455, DS07473.1A, Dmel\\CG42818, Dmel_CG12455', description='CG42818 gene product from transcript CG42818-RB', other_designations='CG42818-PB|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16241939, end_position_on_the_genomic_accession=16245559, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772409, CurrentID=0, Status='live', Symbol='His1:CG33864', Aliases='Dmel_CG33864, CG33864, Dmel\\CG33864', description='CG33864 gene product from transcript CG33864-RA', other_designations='CG33864-PA|His1:CG33864-PA|histone H1', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21536346, end_position_on_the_genomic_accession=21537368, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772412, CurrentID=0, Status='live', Symbol='CG4970', Aliases='Dmel_ CG32953, Dmel\\CG4970, anon-WO0140519.184', description='CG4970 gene product from transcript CG4970-RB', other_designations='CG4970-PB', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11517391, end_position_on_the_genomic_accession=11519783, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326191, CurrentID=0, Status='live', Symbol='CG32073', Aliases='Dmel_ Dmel\\CG32073', description='CG32073 gene product from transcript CG32073-RB', other_designations='CG32073-PA|CG32073-PB|uncharacterized protein', map_location='68A7-68A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11107008, end_position_on_the_genomic_accession=11107582, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772423, CurrentID=0, Status='live', Symbol='CG33673', Aliases='Dmel_ Dmel\\CG33673', description='CG33673 gene product from transcript CG33673-RB', other_designations='CG33673-PB|uncharacterized protein', map_location='22B6-22B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1945815, end_position_on_the_genomic_accession=1946728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772424, CurrentID=0, Status='live', Symbol='CG33771', Aliases='Dmel_ CG14458, Dmel\\CG33771', description='CG33771 gene product from transcript CG33771-RC', other_designations='CG33771-PC|uncharacterized protein', map_location='79F3-79F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22649396, end_position_on_the_genomic_accession=22649988, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  36875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36875, CurrentID=0, Status='live', Symbol='CG5267', Aliases='Dmel_ Dmel\\CG5267', description='CG5267 gene product from transcript CG5267-RA', other_designations='CG5267-PA', map_location='53D6-53D6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16755101, end_position_on_the_genomic_accession=16755823, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835383 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835383, CurrentID=0, Status='live', Symbol='PIG-V', Aliases='Dmel_CG44239, CG44239, CG6657, Dmel\\CG44239, gpi-mt2', description='Phosphatidylinositol glycan anchor biosynthesis, class V ortholog (H. sapiens)', other_designations='CG44239-PA|CG44239-PB|CG44239-PC|GPI mannosyltransferase II|PIG-V-PA|PIG-V-PB|PIG-V-PC|phosphatidylinositol glycan anchor biosynthesis, class V ortholog', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16955458, end_position_on_the_genomic_accession=16957308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740165, CurrentID=0, Status='live', Symbol='CG34427', Aliases='Dmel_ CG13307, Dmel\\CG34427, Dmel_CG13307', description='CG34427 gene product from transcript CG34427-RB', other_designations='CG34427-PB|uncharacterized protein', map_location='66D15-66D15', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8804764, end_position_on_the_genomic_accession=8805812, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326328, CurrentID=0, Status='live', Symbol='CG33189', Aliases='Dmel_ Dmel\\CG33189', description='CG33189 gene product from transcript CG33189-RA', other_designations='CG33189-PA|uncharacterized protein', map_location='85B1-85B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8780113, end_position_on_the_genomic_accession=8780789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772432 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772432, CurrentID=0, Status='live', Symbol='Adhr', Aliases="Dmel_CG3484, 3'ORF, ADHR, ADH[r], Adh-dup, Adh[r], BG:DS01486.9, CG32954, CG3484, Dmel\\CG3484", description='Adh-related', other_designations='Adh related|Adhr-PA|Adhr-PB|Adhr-PC|CG3484-PA|CG3484-PB|CG3484-PC|alcohol dehydrogenase related', map_location='2-50.1 cM', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14615552, end_position_on_the_genomic_accession=14618902, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  3772438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772438, CurrentID=0, Status='live', Symbol='CG33912', Aliases='Dmel_ Dmel\\CG33912', description='CG33912 gene product from transcript CG33912-RA', other_designations='CG33912-PA|uncharacterized protein', map_location='77E4-77E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=20754976, end_position_on_the_genomic_accession=20755685, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3772439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772439, CurrentID=0, Status='live', Symbol='CG11891', Aliases='Dmel_ CG33091, CT37044, Dmel\\CG11891', description='CG11891 gene product from transcript CG11891-RD', other_designations='CG11891-PD|CG11891-PE|uncharacterized protein', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25270231, end_position_on_the_genomic_accession=25273267, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772447, CurrentID=0, Status='live', Symbol='His2A:CG33829', Aliases='Dmel_CG33829, CG33829, Dmel\\CG33829', description='CG33829 gene product from transcript CG33829-RA', other_designations='CG33829-PA|His2A:CG33829-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21478184, end_position_on_the_genomic_accession=21478676, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36896, CurrentID=0, Status='live', Symbol='CG8910', Aliases='Dmel_ Dmel\\CG8910, EG:EG0003.1, anon-WO0140519.220', description='CG8910 gene product from transcript CG8910-RA', other_designations='CG8910-PA', map_location='53D15-53E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16874724, end_position_on_the_genomic_accession=16895652, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  3772449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772449, CurrentID=0, Status='live', Symbol='CG14245', Aliases='Dmel_ CG32950, Dmel\\CG14245', description='CG14245 gene product from transcript CG14245-RB', other_designations='CG14245-PA|CG14245-PB|uncharacterized protein', map_location='97C5-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26728727, end_position_on_the_genomic_accession=26729341, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  36899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36899, CurrentID=0, Status='live', Symbol='CG15614', Aliases='Dmel_ Dmel\\CG15614', description='CG15614 gene product from transcript CG15614-RA', other_designations='CG15614-PA|uncharacterized protein', map_location='53E2-53E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16945555, end_position_on_the_genomic_accession=16947231, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740170, CurrentID=0, Status='live', Symbol='CG34215', Aliases='Dmel_ Dmel\\CG34215', description='CG34215 gene product from transcript CG34215-RA', other_designations='CG34215-PA|uncharacterized protein', map_location='42D4-42D4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6831296, end_position_on_the_genomic_accession=6831793, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36904, CurrentID=0, Status='live', Symbol='CG9010', Aliases='Dmel_ Dmel\\CG9010, anon-WO0140519.132', description='CG9010 gene product from transcript CG9010-RA', other_designations='CG9010-PA', map_location='53E3-53E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16958943, end_position_on_the_genomic_accession=16960195, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772464, CurrentID=0, Status='live', Symbol='CG33914', Aliases='Dmel_ Dmel\\CG33914', description='CG33914 gene product from transcript CG33914-RC', other_designations='CG33914-PB|CG33914-PC|uncharacterized protein', map_location='42E5-42E5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7036768, end_position_on_the_genomic_accession=7039050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32681, CurrentID=0, Status='live', Symbol='CG13002', Aliases='Dmel_ Dmel\\CG13002', description='CG13002 gene product from transcript CG13002-RA', other_designations='CG13002-PA|uncharacterized protein', map_location='15B5-15B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16889810, end_position_on_the_genomic_accession=16890637, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740782, CurrentID=0, Status='live', Symbol='whip', Aliases='Dmel_CG34218, CG2149, CG34218, Dmel\\CG34218, NEST:bs34h06, NEST:bs36g01', description='whipple', other_designations='CG34218-PA|NEST:bs34h06|NEST:bs36g01|whip-PA', map_location='44C4-44C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8362359, end_position_on_the_genomic_accession=8363218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  13084051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=13084051, CurrentID=0, Status='live', Symbol='CG43178', Aliases='Dmel_ Dmel\\CG43178, MIP03715_ORF_2', description='CG43178 gene product from transcript CG43178-RB', other_designations='CG43178-PA|CG43178-PB|CG43178-PC|uncharacterized protein', map_location='47A2-47A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=10352456, end_position_on_the_genomic_accession=10353114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835393, CurrentID=0, Status='live', Symbol='CG44437', Aliases='Dmel_ CG1903, Dmel\\CG44437', description='CG44437 gene product from transcript CG44437-RA', other_designations='CG44437-PA|uncharacterized protein', map_location='11E3-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13201170, end_position_on_the_genomic_accession=13202262, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740175, CurrentID=0, Status='live', Symbol='CG34160', Aliases='Dmel_ Dmel\\CG34160, NEST:bs34b09', description='CG34160 gene product from transcript CG34160-RA', other_designations='CG34160-PA|NEST:bs34b09|uncharacterized protein', map_location='32A1-32A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10579753, end_position_on_the_genomic_accession=10580324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13084062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=13084062, CurrentID=0, Status='live', Symbol='CG43210', Aliases='Dmel_ Dmel\\CG43210', description='CG43210 gene product from transcript CG43210-RD', other_designations='CG43210-PB|CG43210-PC|CG43210-PD|uncharacterized protein', map_location='91B4-91B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18564264, end_position_on_the_genomic_accession=18565173, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3772489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772489, CurrentID=0, Status='live', Symbol='His3:CG33812', Aliases='Dmel_CG33812, CG33812, Dmel\\CG33812', description='CG33812 gene product from transcript CG33812-RA', other_designations='CG33812-PA|His3:CG33812-PA|histone H3', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21440187, end_position_on_the_genomic_accession=21440715, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772494, CurrentID=0, Status='live', Symbol='CG33679', Aliases='Dmel_ BP1003, Dmel\\CG33679', description='CG33679 gene product from transcript CG33679-RA', other_designations='CG33679-PA|uncharacterized protein', map_location='35B9-35B10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15058303, end_position_on_the_genomic_accession=15059120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740178, CurrentID=0, Status='live', Symbol='CG34286', Aliases='Dmel_ Dmel\\CG34286', description='CG34286 gene product from transcript CG34286-RB', other_designations='CG34286-PA|CG34286-PB|uncharacterized protein', map_location='92B1-92B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=19715186, end_position_on_the_genomic_accession=19715927, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  36944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36944, CurrentID=0, Status='live', Symbol='CG12699', Aliases='Dmel_ BEST:GH14656, Dmel\\CG12699, GH14656, NEST:bs24f01, anon-WO0140519.144, bs24f01.y1', description='CG12699 gene product from transcript CG12699-RB', other_designations='CG12699-PB|uncharacterized protein', map_location='54B1-54B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17246530, end_position_on_the_genomic_accession=17247383, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772502, CurrentID=0, Status='live', Symbol='His2B:CG33894', Aliases='Dmel_CG33894, CG33894, Dmel\\CG33894', description='CG33894 gene product from transcript CG33894-RA', other_designations='CG33894-PA|His2B:CG33894-PA|histone H2B', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21472418, end_position_on_the_genomic_accession=21473012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19835398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835398, CurrentID=0, Status='live', Symbol='CG44521', Aliases='Dmel_ Dmel\\CG44521', description='CG44521 gene product from transcript CG44521-RA', other_designations='CG44521-PA|uncharacterized protein', map_location='64E2-64E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5713331, end_position_on_the_genomic_accession=5714029, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772504, CurrentID=0, Status='live', Symbol='CG33630', Aliases='Dmel_ CG12808, Dmel\\CG33630', description='CG33630 gene product from transcript CG33630-RA', other_designations='CG33630-PA|uncharacterized protein', map_location='85F8-85F8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10062624, end_position_on_the_genomic_accession=10063327, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740431, CurrentID=0, Status='live', Symbol='CG34223', Aliases='Dmel_ Dmel\\CG34223', description='CG34223 gene product from transcript CG34223-RA', other_designations='CG34223-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11235860, end_position_on_the_genomic_accession=11236204, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772505, CurrentID=0, Status='live', Symbol='His2A:CG33865', Aliases='Dmel_CG33865, CG33865, Dmel\\CG33865', description='CG33865 gene product from transcript CG33865-RA', other_designations='CG33865-PA|His2A:CG33865-PA|histone H2A', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21538228, end_position_on_the_genomic_accession=21538720, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772506, CurrentID=0, Status='live', Symbol='CG11889', Aliases='Dmel_ CG33091, Dmel\\CG11889', description='CG11889 gene product from transcript CG11889-RA', other_designations='CG11889-PA|CG11889-PB|uncharacterized protein', map_location='96C8-96C8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25270209, end_position_on_the_genomic_accession=25273267, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772509, CurrentID=0, Status='live', Symbol='His4:CG33887', Aliases='Dmel_CG33887, CG33887, Dmel\\CG33887', description='CG33887 gene product from transcript CG33887-RA', other_designations='CG33887-PA|His4:CG33887-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21489234, end_position_on_the_genomic_accession=21489646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  3346205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3346205, CurrentID=0, Status='live', Symbol='CheB53a', Aliases='Dmel_CG33550, CG16797, CG33457, CG33550, Dmel\\CG33550', description='Chemosensory protein B 53a', other_designations='CG33550-PA|CheB53a-PA|chemosensory protein B 53a', map_location='53C5-53C5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16466709, end_position_on_the_genomic_accession=16467522, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772511, CurrentID=0, Status='live', Symbol='CG33667', Aliases='Dmel_ Dmel\\CG33667', description='CG33667 gene product from transcript CG33667-RB', other_designations='CG33667-PB|uncharacterized protein', map_location='6A2-6A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6471673, end_position_on_the_genomic_accession=6472378, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772513, CurrentID=0, Status='live', Symbol='CG33690', Aliases='Dmel_ Dmel\\CG33690', description='CG33690 gene product from transcript CG33690-RA', other_designations='CG33690-PA|uncharacterized protein', map_location='72D10-72D10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16201328, end_position_on_the_genomic_accession=16202032, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772517, CurrentID=0, Status='live', Symbol='His3:CG33860', Aliases='Dmel_CG33860, CG33860, Dmel\\CG33860', description='CG33860 gene product from transcript CG33860-RA', other_designations='CG33860-PA|His3:CG33860-PA|histone H3', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21529795, end_position_on_the_genomic_accession=21530323, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772518, CurrentID=0, Status='live', Symbol='His3:CG33851', Aliases='Dmel_CG33851, CG33851, Dmel\\CG33851', description='CG33851 gene product from transcript CG33851-RA', other_designations='CG33851-PA|His3:CG33851-PA|histone H3', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21515062, end_position_on_the_genomic_accession=21515590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772519, CurrentID=0, Status='live', Symbol='His4:CG33891', Aliases='Dmel_CG33891, CG33891, Dmel\\CG33891', description='CG33891 gene product from transcript CG33891-RA', other_designations='CG33891-PA|His4:CG33891-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21479042, end_position_on_the_genomic_accession=21479454, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740183, CurrentID=0, Status='live', Symbol='CG34236', Aliases='Dmel_ Dmel\\CG34236', description='CG34236 gene product from transcript CG34236-RC', other_designations='CG34236-PB|CG34236-PC|uncharacterized protein', map_location='50B5-50B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13604495, end_position_on_the_genomic_accession=13605655, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740186, CurrentID=0, Status='live', Symbol='CG34265', Aliases='Dmel_ Dmel\\CG34265', description='CG34265 gene product from transcript CG34265-RB', other_designations='CG34265-PB|uncharacterized protein', map_location='63C3-63C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3268461, end_position_on_the_genomic_accession=3268994, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31523, CurrentID=0, Status='live', Symbol='CG12236', Aliases='Dmel_ Dmel\\CG12236', description='CG12236 gene product from transcript CG12236-RA', other_designations='CG12236-PA|CG12236-PB|uncharacterized protein', map_location='5C7-5C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5908135, end_position_on_the_genomic_accession=5913079, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  39032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39032, CurrentID=0, Status='live', Symbol='CG5068', Aliases='Dmel_ Dmel\\CG5068', description='CG5068 gene product from transcript CG5068-RB', other_designations='CG5068-PA|CG5068-PB', map_location='66F1-66F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8986710, end_position_on_the_genomic_accession=8989135, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  36988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36988, CurrentID=0, Status='live', Symbol='CG14480', Aliases='Dmel_ Dmel\\CG14480', description='CG14480 gene product from transcript CG14480-RA', other_designations='CG14480-PA|CG14480-PB|uncharacterized protein', map_location='54C8-54C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17564340, end_position_on_the_genomic_accession=17565852, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  3772541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772541, CurrentID=0, Status='live', Symbol='His2A:CG33820', Aliases='Dmel_CG33820, CG33820, Dmel\\CG33820', description='CG33820 gene product from transcript CG33820-RA', other_designations='CG33820-PA|His2A:CG33820-PA|histone H2A', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21463055, end_position_on_the_genomic_accession=21463547, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772542, CurrentID=0, Status='live', Symbol='Menl-1', Aliases='Dmel_CG7964, CG30097, CG7964, Dmel\\CG7964', description='Malic enzyme like-1', other_designations='CG7964-PD|Menl-1-PD|malic enzyme like-1', map_location='53C4-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16414815, end_position_on_the_genomic_accession=16419398, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  36991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36991, CurrentID=0, Status='live', Symbol='UQCR-6.4', Aliases='Dmel_CG14482, CG14482, Dmel\\CG14482, UCRY', description='Ubiquinol-cytochrome c reductase 6.4 kDa subunit', other_designations='CG14482-PA|CG14482-PB|UQCR-6.4-PA|UQCR-6.4-PB|Ubiquinol-cytochrome C reductase complex 6.4 kDa protein|Ubiquinol-cytochrome c oxidoreductase 6.4 kDa|Ubiquinol-cytochrome c reductase 6.4 kDa protein', map_location='54C9-54C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17572096, end_position_on_the_genomic_accession=17572784, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37000, CurrentID=0, Status='live', Symbol='CG10936', Aliases='Dmel_ Dmel\\CG10936', description='CG10936 gene product from transcript CG10936-RA', other_designations='CG10936-PA|uncharacterized protein', map_location='54D1-54D2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17604951, end_position_on_the_genomic_accession=17639138, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  3772556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772556, CurrentID=0, Status='live', Symbol='His2A:CG33856', Aliases='Dmel_CG33856, CG33856, Dmel\\CG33856', description='CG33856 gene product from transcript CG33856-RA', other_designations='CG33856-PA|His2A:CG33856-PA|histone H2A', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21523498, end_position_on_the_genomic_accession=21523990, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37007, CurrentID=0, Status='live', Symbol='Sema-1b', Aliases='Dmel_CG6446, BcDNA:GH03186, CG6446, D-semaIII, Dmel\\CG6446, Sema 1b, Sema III, Sema1b, sema-III, sema1b', description='CG6446 gene product from transcript CG6446-RA', other_designations='CG6446-PA|CG6446-PB|CG6446-PC|Sema-1b-PA|Sema-1b-PB|Sema-1b-PC|semaphorin|semaphorin-I|semaphorin-III|semaphorin-like|semiphorin', map_location='54D4-54D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17672839, end_position_on_the_genomic_accession=17683499, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  37008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37008, CurrentID=0, Status='live', Symbol='CG42562', Aliases='Dmel_ CG14481, Dmel\\CG42562, Dmel_CG14481', description='CG42562 gene product from transcript CG42562-RA', other_designations='CG42562-PA|uncharacterized protein', map_location='54D5-54D5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17685814, end_position_on_the_genomic_accession=17686410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772563, CurrentID=0, Status='live', Symbol='CG33644', Aliases='Dmel_ Dmel\\CG33644', description='CG33644 gene product from transcript CG33644-RA', other_designations='CG33644-PA|uncharacterized protein', map_location='34C4-34C4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13532085, end_position_on_the_genomic_accession=13532636, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37013, CurrentID=0, Status='live', Symbol='CG10934', Aliases='Dmel_ Dmel\\CG10934, anon-WO0140519.130', description='CG10934 gene product from transcript CG10934-RA', other_designations='CG10934-PA|uncharacterized protein', map_location='54E1-54E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17722339, end_position_on_the_genomic_accession=17723381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772566, CurrentID=0, Status='live', Symbol='CG11714', Aliases='Dmel_ CG33047, Dmel\\CG11714', description='CG11714 gene product from transcript CG11714-RD', other_designations='CG11714-PA|CG11714-PC|CG11714-PD|uncharacterized protein', map_location='68C12-68C12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11518798, end_position_on_the_genomic_accession=11522713, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37017, CurrentID=0, Status='live', Symbol='CG4984', Aliases='Dmel_ Dmel\\CG4984', description='CG4984 gene product from transcript CG4984-RD', other_designations='CG4984-PA|CG4984-PB|CG4984-PC|CG4984-PD', map_location='54E2-54E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17726914, end_position_on_the_genomic_accession=17732987, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  37022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37022, CurrentID=0, Status='live', Symbol='Ir54a', Aliases='Dmel_CG14487, CG14487, DmelIR54a, Dmel\\CG14487, IR54a', description='Ionotropic receptor 54a', other_designations='CG14487-PA|Ir54a-PA|ionotropic receptor 54a', map_location='54E7-54E7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17744101, end_position_on_the_genomic_accession=17746005, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37024, CurrentID=0, Status='live', Symbol='CG10931', Aliases='Dmel_ Dmel\\CG10931', description='CG10931 gene product from transcript CG10931-RA', other_designations='CG10931-PA|uncharacterized protein', map_location='54E7-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17748935, end_position_on_the_genomic_accession=17750136, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  39040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39040, CurrentID=0, Status='live', Symbol='CG5144', Aliases='Dmel_ Dmel\\CG5144', description='CG5144 gene product from transcript CG5144-RA', other_designations='CG5144-PA', map_location='66F3-66F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9047439, end_position_on_the_genomic_accession=9048724, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740195, CurrentID=0, Status='live', Symbol='CG34232', Aliases='Dmel_ Dmel\\CG34232', description='CG34232 gene product from transcript CG34232-RC', other_designations='CG34232-PA|CG34232-PB|CG34232-PC|uncharacterized protein', map_location='48E4-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12164909, end_position_on_the_genomic_accession=12165812, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772580, CurrentID=0, Status='live', Symbol='CG13430', Aliases='Dmel_ CG33008, Dmel\\CG13430, SP70', description='CG13430 gene product from transcript CG13430-RA', other_designations='CG13430-PA|CG13430-PB|uncharacterized protein', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20550334, end_position_on_the_genomic_accession=20552090, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772581, CurrentID=0, Status='live', Symbol='His1:CG33858', Aliases='Dmel_CG33858, CG33858, Dmel\\CG33858', description='CG33858 gene product from transcript CG33858-RA', other_designations='CG33858-PA|His1:CG33858-PA|histone H1', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21526459, end_position_on_the_genomic_accession=21527481, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37031, CurrentID=0, Status='live', Symbol='CG18635', Aliases='Dmel_ Dmel\\CG18635', description='CG18635 gene product from transcript CG18635-RA', other_designations='CG18635-PA|uncharacterized protein', map_location='54E8-54E8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17767808, end_position_on_the_genomic_accession=17769989, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37045, CurrentID=0, Status='live', Symbol='CG14492', Aliases='Dmel_ Dmel\\CG14492', description='CG14492 gene product from transcript CG14492-RA', other_designations='CG14492-PA|uncharacterized protein', map_location='55A1-55A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17893682, end_position_on_the_genomic_accession=17894342, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37051, CurrentID=0, Status='live', Symbol='CG5084', Aliases='Dmel_ Dmel\\CG5084', description='CG5084 gene product from transcript CG5084-RB', other_designations='CG5084-PA|CG5084-PB', map_location='55B1-55B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18018665, end_position_on_the_genomic_accession=18020211, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740569, CurrentID=0, Status='live', Symbol='CG34337', Aliases='Dmel_ Dmel\\CG34337', description='CG34337 gene product from transcript CG34337-RA', other_designations='CG34337-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7417253, end_position_on_the_genomic_accession=7417693, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37055, CurrentID=0, Status='live', Symbol='CG5767', Aliases='Dmel_ CT18118, Dmel\\CG5767', description='CG5767 gene product from transcript CG5767-RA', other_designations='CG5767-PA', map_location='55B1-55B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18046386, end_position_on_the_genomic_accession=18047484, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772611, CurrentID=0, Status='live', Symbol='CG33710', Aliases='Dmel_ Dmel\\CG33710, NEST:bs03f11, bs03f11.y1', description='CG33710 gene product from transcript CG33710-RA', other_designations='CG33710-PA|uncharacterized protein', map_location='76D6-76D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19951954, end_position_on_the_genomic_accession=19952790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772618, CurrentID=0, Status='live', Symbol='His2A:CG33838', Aliases='Dmel_CG33838, CG33838, Dmel\\CG33838', description='CG33838 gene product from transcript CG33838-RA', other_designations='CG33838-PA|His2A:CG33838-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21493423, end_position_on_the_genomic_accession=21493915, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740203, CurrentID=0, Status='live', Symbol='CG34179', Aliases='Dmel_ Dmel\\CG34179', description='CG34179 gene product from transcript CG34179-RC', other_designations='CG34179-PA|CG34179-PB|CG34179-PC|uncharacterized protein', map_location='26B3-26B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6046651, end_position_on_the_genomic_accession=6048706, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772625, CurrentID=0, Status='live', Symbol='CG33795', Aliases='Dmel_ Dmel\\CG33795', description='CG33795 gene product from transcript CG33795-RB', other_designations='CG33795-PB|uncharacterized protein', map_location='72D9-72D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16170729, end_position_on_the_genomic_accession=16171380, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740205, CurrentID=0, Status='live', Symbol='CG34261', Aliases='Dmel_ Dmel\\CG34261', description='CG34261 gene product from transcript CG34261-RB', other_designations='CG34261-PB|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21196831, end_position_on_the_genomic_accession=21197738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772632, CurrentID=0, Status='live', Symbol='His2A:CG33832', Aliases='Dmel_CG33832, CG33832, Dmel\\CG33832', description='CG33832 gene product from transcript CG33832-RA', other_designations='CG33832-PA|His2A:CG33832-PA|histone H2A', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21483227, end_position_on_the_genomic_accession=21483719, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740483, CurrentID=0, Status='live', Symbol='CG34308', Aliases='Dmel_ Dmel\\CG34308', description='CG34308 gene product from transcript CG34308-RB', other_designations='CG34308-PB|uncharacterized protein', map_location='87D12-87D12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13089242, end_position_on_the_genomic_accession=13090081, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  42545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42545, CurrentID=0, Status='live', Symbol='CG7956', Aliases='Dmel_ Dmel\\CG7956', description='CG7956 gene product from transcript CG7956-RD', other_designations='CG7956-PA|CG7956-PD|CG7956-PE|CG7956-PF|CG7956-PG', map_location='93E4-93E4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21520928, end_position_on_the_genomic_accession=21528598, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  37086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37086, CurrentID=0, Status='live', Symbol='CG14499', Aliases='Dmel_ CR14499, Dmel\\CG14499, Dmel_CR14499', description='CG14499 gene product from transcript CG14499-RC', other_designations='CG14499-PC|uncharacterized protein', map_location='55B12-55B12', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18187707, end_position_on_the_genomic_accession=18188573, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740207, CurrentID=0, Status='live', Symbol='CG34323', Aliases='Dmel_ Dmel\\CG34323', description='CG34323 gene product from transcript CG34323-RA', other_designations='CG34323-PA|uncharacterized protein', map_location='11A3-11A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12070437, end_position_on_the_genomic_accession=12071263, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772640, CurrentID=0, Status='live', Symbol='CG33786', Aliases='Dmel_ CG10438, CG10438b, CG33005, CG33005-ORFB, CG33005ORFB, Dmel\\CG33786, Dromel_CG33786_FBtr0091786_mORF', description='CG33786 gene product from transcript CG33786-RA', other_designations='CG33786-PA|uncharacterized protein', map_location='57A8-57A9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20642466, end_position_on_the_genomic_accession=20643785, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772641, CurrentID=0, Status='live', Symbol='Shawn', Aliases='Dmel_CG14209, CG14209, CG33075, Dmel\\CG14209', description='CG14209 gene product from transcript CG14209-RE', other_designations='CG14209-PB|CG14209-PC|CG14209-PD|CG14209-PE|Shawn-PB|Shawn-PC|Shawn-PD|Shawn-PE|shawn', map_location='18D3-18D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19611526, end_position_on_the_genomic_accession=19615617, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3772644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772644, CurrentID=0, Status='live', Symbol='CG33664', Aliases='Dmel_ Dmel\\CG33664', description='CG33664 gene product from transcript CG33664-RB', other_designations='CG33664-PB|uncharacterized protein', map_location='5F5-5F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6384220, end_position_on_the_genomic_accession=6384899, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37094, CurrentID=0, Status='live', Symbol='CG18537', Aliases='Dmel_ Dmel\\CG18537', description='CG18537 gene product from transcript CG18537-RA', other_designations='CG18537-PA|uncharacterized protein', map_location='55C2-55C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18295928, end_position_on_the_genomic_accession=18296656, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772652, CurrentID=0, Status='live', Symbol='CG33923', Aliases='Dmel_ Dmel\\CG33923', description='CG33923 gene product from transcript CG33923-RB', other_designations='CG33923-PB|uncharacterized protein', map_location='22A1-22A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1341365, end_position_on_the_genomic_accession=1342095, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  3772658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772658, CurrentID=0, Status='live', Symbol='CG33722', Aliases='Dmel_ CG18238, CG31188, CG31188-ORFA, Dmel\\CG33722', description='CG33722 gene product from transcript CG33722-RE', other_designations='CG33722-PB|CG33722-PC|CG33722-PD|CG33722-PE|uncharacterized protein', map_location='84F5-84F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=8244055, end_position_on_the_genomic_accession=8248054, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  3772662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772662, CurrentID=0, Status='live', Symbol='CG7215', Aliases='Dmel_ CG32920, Dmel\\CG7215', description='CG7215 gene product from transcript CG7215-RA', other_designations='CG7215-PA|CG7215-PB', map_location='90F4-90F4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18165442, end_position_on_the_genomic_accession=18167273, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740212, CurrentID=0, Status='live', Symbol='CG34274', Aliases='Dmel_ Dmel\\CG34274', description='CG34274 gene product from transcript CG34274-RA', other_designations='CG34274-PA|uncharacterized protein', map_location='89A2-89A2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15731746, end_position_on_the_genomic_accession=15732694, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  3772665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772665, CurrentID=0, Status='live', Symbol='His1:CG33822', Aliases='Dmel_CG33822, CG33822, Dmel\\CG33822', description='CG33822 gene product from transcript CG33822-RA', other_designations='CG33822-PA|His1:CG33822-PA|histone H1', map_location='39D4-39D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21466216, end_position_on_the_genomic_accession=21467238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740687, CurrentID=0, Status='live', Symbol='CG34338', Aliases='Dmel_ Dmel\\CG34338', description='CG34338 gene product from transcript CG34338-RB', other_designations='CG34338-PB|uncharacterized protein', map_location='7B8-7B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7768737, end_position_on_the_genomic_accession=7769161, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835431 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835431, CurrentID=0, Status='live', Symbol='CG45099', Aliases='Dmel_ Dmel\\CG45099', description='CG45099 gene product from transcript CG45099-RA', other_designations='CG45099-PA|uncharacterized protein', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22032150, end_position_on_the_genomic_accession=22034286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  37118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37118, CurrentID=0, Status='live', Symbol='CG12263', Aliases='Dmel_ Dmel\\CG12263, anon-WO0140519.205', description='CG12263 gene product from transcript CG12263-RA', other_designations='CG12263-PA|uncharacterized protein', map_location='55C9-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18426421, end_position_on_the_genomic_accession=18430333, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  10178816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178816, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-U', Aliases='Dmel_CG42777, CG42777, Dmel\\CG42777', description='CG42777 gene product from transcript CG42777-RA', other_designations='CG42777-PA|CG42777-PB|pre-mod(mdg4)-U-PA|pre-mod(mdg4)-U-PB', map_location='93D9-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21370249, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5740790 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740790, CurrentID=0, Status='live', Symbol='Rpb12', Aliases='Dmel_CG34186, CG34186, Dmel\\CG34186, RPB12', description='CG34186 gene product from transcript CG34186-RA', other_designations='CG34186-PA|Rpb12-PA', map_location='51E2-51E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=15144968, end_position_on_the_genomic_accession=15145519, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740214, CurrentID=0, Status='live', Symbol='CG34170', Aliases='Dmel_ Dmel\\CG34170', description='CG34170 gene product from transcript CG34170-RB', other_designations='CG34170-PA|CG34170-PB|uncharacterized protein', map_location='36C5-36C5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17311468, end_position_on_the_genomic_accession=17311820, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37124, CurrentID=0, Status='live', Symbol='CG5190', Aliases='Dmel_ Dmel\\CG5190', description='CG5190 gene product from transcript CG5190-RA', other_designations='CG5190-PA', map_location='55C9-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18451284, end_position_on_the_genomic_accession=18452736, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37125, CurrentID=0, Status='live', Symbol='CG17669', Aliases='Dmel_ Dmel\\CG17669', description='CG17669 gene product from transcript CG17669-RA', other_designations='CG17669-PA|uncharacterized protein', map_location='55C9-55C9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18452737, end_position_on_the_genomic_accession=18454462, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36181, CurrentID=0, Status='live', Symbol='CG13216', Aliases='Dmel_ Dmel\\CG13216', description='CG13216 gene product from transcript CG13216-RA', other_designations='CG13216-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11238108, end_position_on_the_genomic_accession=11238995, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178822, CurrentID=0, Status='live', Symbol='CG42697', Aliases='Dmel_ Dmel\\CG42697', description='CG42697 gene product from transcript CG42697-RA', other_designations='CG42697-PA|uncharacterized protein', map_location='55E1-55E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18611645, end_position_on_the_genomic_accession=18612730, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10178823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178823, CurrentID=0, Status='live', Symbol='CG42778', Aliases='Dmel_ Dmel\\CG42778, lincRNA.767', description='CG42778 gene product from transcript CG42778-RC', other_designations='CG42778-PC|uncharacterized protein', map_location='87F2-87F2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13528597, end_position_on_the_genomic_accession=13529356, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  5740572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740572, CurrentID=0, Status='live', Symbol='CG34193', Aliases='Dmel_ Dmel\\CG34193', description='CG34193 gene product from transcript CG34193-RB', other_designations='CG34193-PA|CG34193-PB|uncharacterized protein', map_location='54C1-54C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=17508995, end_position_on_the_genomic_accession=17511394, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178824, CurrentID=0, Status='live', Symbol='CG42688', Aliases='Dmel_ DMG8-chr2L.34.001, Dmel\\CG42688, anon-SAGE:Wang-006', description='CG42688 gene product from transcript CG42688-RB', other_designations='CG42688-PA|CG42688-PB|anon-SAGE:Wang-006|uncharacterized protein', map_location='37C6-37C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19181965, end_position_on_the_genomic_accession=19182810, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178825, CurrentID=0, Status='live', Symbol='CG42650', Aliases='Dmel_ DMG4-MB6.chr3R.6.014, Dmel\\CG42650', description='CG42650 gene product from transcript CG42650-RA', other_designations='CG42650-PA|uncharacterized protein', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7391413, end_position_on_the_genomic_accession=7392280, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  42593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=42593, CurrentID=0, Status='live', Symbol='CG6015', Aliases='Dmel_ CDC40, Dmel\\CG6015, prp17', description='CG6015 gene product from transcript CG6015-RA', other_designations='CG6015-PA', map_location='94A1-94A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22032150, end_position_on_the_genomic_accession=22034286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10178828 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178828, CurrentID=0, Status='live', Symbol='CG42703', Aliases='Dmel_ Dmel\\CG42703', description='CG42703 gene product from transcript CG42703-RA', other_designations='CG42703-PA|uncharacterized protein', map_location='59C4-59C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23101232, end_position_on_the_genomic_accession=23101777, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740216, CurrentID=0, Status='live', Symbol='CG34244', Aliases='Dmel_ Dmel\\CG34244', description='CG34244 gene product from transcript CG34244-RB', other_designations='CG34244-PA|CG34244-PB|uncharacterized protein', map_location='70D4-70D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=14235053, end_position_on_the_genomic_accession=14235941, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772691, CurrentID=0, Status='live', Symbol='Kaz1-ORFB', Aliases='Dmel_CG1220, CG1220, CG33045, Dmel\\CG1220, Kaz1, ORF B, ORF C, ORF D, ORF E', description='CG1220 gene product from transcript CG1220-RE', other_designations='CG1220-PA|CG1220-PE|CG1220-PF|CG1220-PH|KAZ1 type C1|KAZ1 type C2|KAZ1 type C3|KAZ1 type D|KAZ1 type E|KAZ1 type beta|Kaz1-ORFB-PA|Kaz1-ORFB-PE|Kaz1-ORFB-PF|Kaz1-ORFB-PH', map_location='61B2-61B2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=216625, end_position_on_the_genomic_accession=218058, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3772692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772692, CurrentID=0, Status='live', Symbol='Menl-2', Aliases='Dmel_CG7969, CG30097, CG7969, Dmel\\CG7969', description='Malic enzyme like-2', other_designations='CG7969-PA|CG7969-PD|CG7969-PE|Menl-2-PA|Menl-2-PD|Menl-2-PE|malic enzyme like-2', map_location='53C4-53C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=16414815, end_position_on_the_genomic_accession=16419398, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  5740219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740219, CurrentID=0, Status='live', Symbol='CG34211', Aliases='Dmel_ Dmel\\CG34211', description='CG34211 gene product from transcript CG34211-RA', other_designations='CG34211-PA|uncharacterized protein', map_location='42A8-42A8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6047195, end_position_on_the_genomic_accession=6047392, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37150, CurrentID=0, Status='live', Symbol='CG44433', Aliases='Dmel_ CG15086, Dmel\\CG44433', description='CG44433 gene product from transcript CG44433-RA', other_designations='CG44433-PA|uncharacterized protein', map_location='55E6-55E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18684230, end_position_on_the_genomic_accession=18686775, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772708, CurrentID=0, Status='live', Symbol='His4:CG33871', Aliases='Dmel_CG33871, CG33871, Dmel\\CG33871', description='CG33871 gene product from transcript CG33871-RA', other_designations='CG33871-PA|His4:CG33871-PA|histone H4', map_location='39E1-39E1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21529242, end_position_on_the_genomic_accession=21529769, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  10178853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178853, CurrentID=0, Status='live', Symbol='CG42866', Aliases='Dmel_ Dmel\\CG42866', description='CG42866 gene product from transcript CG42866-RA', other_designations='CG42866-PA|uncharacterized protein', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19961224, end_position_on_the_genomic_accession=19961685, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59128, CurrentID=0, Status='live', Symbol='CG15107', Aliases='Dmel_ Dmel\\CG15107', description='CG15107 gene product from transcript CG15107-RA', other_designations='CG15107-PA|CG15107-PB|uncharacterized protein', map_location='56A2-56A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18971309, end_position_on_the_genomic_accession=18972537, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  59137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59137, CurrentID=0, Status='live', Symbol='CG18735', Aliases='Dmel_ Dmel\\CG18735, SP22', description='CG18735 gene product from transcript CG18735-RA', other_designations='CG18735-PA|uncharacterized protein', map_location='58A1-58A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21798017, end_position_on_the_genomic_accession=21799195, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  37163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37163, CurrentID=0, Status='live', Symbol='Cyp12b2', Aliases='Dmel_CG15077, 12b2, CG15077, Dmel\\CG15077', description='CG15077 gene product from transcript CG15077-RA', other_designations='CG15077-PA|Cyp12b2-PA', map_location='55F1-55F1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18754995, end_position_on_the_genomic_accession=18757083, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  59140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59140, CurrentID=0, Status='live', Symbol='CG18744', Aliases='Dmel_ Dmel\\CG18744', description='CG18744 gene product from transcript CG18744-RB', other_designations='CG18744-PB|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7797078, end_position_on_the_genomic_accession=7798771, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  59141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59141, CurrentID=0, Status='live', Symbol='CG18745', Aliases='Dmel_ Dmel\\CG18745', description='CG18745 gene product from transcript CG18745-RA', other_designations='CG18745-PA|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7783843, end_position_on_the_genomic_accession=7786119, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  59142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59142, CurrentID=0, Status='live', Symbol='CG18746', Aliases='Dmel_ Dmel\\CG18746', description='CG18746 gene product from transcript CG18746-RA', other_designations='CG18746-PA|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7788282, end_position_on_the_genomic_accession=7790187, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  59143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59143, CurrentID=0, Status='live', Symbol='CG18747', Aliases='Dmel_ Dmel\\CG18747, Q9I7L1_DROME', description='CG18747 gene product from transcript CG18747-RB', other_designations='CG18747-PB|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7780884, end_position_on_the_genomic_accession=7783446, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  59144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59144, CurrentID=0, Status='live', Symbol='CG18748', Aliases='Dmel_ CR18748, Dmel\\CG18748', description='CG18748 gene product from transcript CG18748-RB', other_designations='CG18748-PB|CG18748-PC|uncharacterized protein', map_location='84E1-84E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7786170, end_position_on_the_genomic_accession=7787923, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  59145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59145, CurrentID=0, Status='live', Symbol='CG18754', Aliases='Dmel_ Dmel\\CG18754, SP61', description='CG18754 gene product from transcript CG18754-RB', other_designations='CG18754-PB|CG18754-PC|uncharacterized protein', map_location='95A7-95A7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23684057, end_position_on_the_genomic_accession=23686537, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59149, CurrentID=0, Status='live', Symbol='CG18765', Aliases='Dmel_ Dmel\\CG18765', description='CG18765 gene product from transcript CG18765-RB', other_designations='CG18765-PB|uncharacterized protein', map_location='86E16-86E16', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11690386, end_position_on_the_genomic_accession=11691799, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59150, CurrentID=0, Status='live', Symbol='CG18766', Aliases='Dmel_ CG5995, Dmel\\CG18766', description='CG18766 gene product from transcript CG18766-RB', other_designations='CG18766-PA|CG18766-PB|CG18766-PC|uncharacterized protein', map_location='97E10-97E10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27242335, end_position_on_the_genomic_accession=27245249, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  59151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59151, CurrentID=0, Status='live', Symbol='mRpL36', Aliases='Dmel_CG18767, BcDNA:GM01757, CG18767, Dmel\\CG18767', description='mitochondrial ribosomal protein L36', other_designations='CG18767-PA|mRpL36-PA', map_location='66B7-66B7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8124584, end_position_on_the_genomic_accession=8125132, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59157, CurrentID=0, Status='live', Symbol='Cpr65Ax2', Aliases='Dmel_CG18777, BcDNA:RE62456, CG18777, DmelCpr65Ax2, Dmel\\CG18777', description='Cuticular protein 65Ax2', other_designations='CG18777-PA|CG18777-PB|Cpr65Ax2-PA|Cpr65Ax2-PB|cuticular protein 65Ax2', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6150058, end_position_on_the_genomic_accession=6150857, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59158, CurrentID=0, Status='live', Symbol='Cpr65Au', Aliases='Dmel_CG18778, BcDNA:GH21145, CG18778, DmelCpr65Au, Dmel\\CG18778', description='Cuticular protein 65Au', other_designations='CG18778-PA|CG18778-PB|Cpr65Au-PA|Cpr65Au-PB|cuticular protein 65Au', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6131242, end_position_on_the_genomic_accession=6132109, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59159, CurrentID=0, Status='live', Symbol='Lcp65Ag3', Aliases='Dmel_CG18779, CG18779, DmelLcp65Ag3, Dmel\\CG18779, Lcp-g3, g3', description='Larval cuticle protein', other_designations='CG18779-PA|Lcp65Ag3-PA|larval cuticle protein', map_location='65A5-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6130386, end_position_on_the_genomic_accession=6130989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59163, CurrentID=0, Status='live', Symbol='CG18787', Aliases='Dmel_ Dmel\\CG18787', description='CG18787 gene product from transcript CG18787-RB', other_designations='CG18787-PB|uncharacterized protein', map_location='33B5-33B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11997900, end_position_on_the_genomic_accession=11999340, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59164, CurrentID=0, Status='live', Symbol='CG18788', Aliases='Dmel_ Dmel\\CG18788', description='CG18788 gene product from transcript CG18788-RB', other_designations='CG18788-PB|uncharacterized protein', map_location='33B9-33B9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=12034190, end_position_on_the_genomic_accession=12036131, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  59167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59167, CurrentID=0, Status='live', Symbol='OS9', Aliases='Dmel_CG18806, CG18806, Dmel\\CG18806', description='Olfactory-specific 9', other_designations='CG18806-PB|OS9-PB', map_location='38A8-38A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19965664, end_position_on_the_genomic_accession=19966510, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  3772720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772720, CurrentID=0, Status='live', Symbol='CG33919', Aliases='Dmel_ Dmel\\CG33919', description='CG33919 gene product from transcript CG33919-RA', other_designations='CG33919-PA|uncharacterized protein', map_location='42B2-42B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6623832, end_position_on_the_genomic_accession=6624577, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59170, CurrentID=0, Status='live', Symbol='CG18809', Aliases='Dmel_ CG1809, Dmel\\CG18809', description='CG18809 gene product from transcript CG18809-RB', other_designations='CG18809-PA|CG18809-PB|CG18809-PC|CG18809-PD|uncharacterized protein', map_location='18E5-18E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19741224, end_position_on_the_genomic_accession=19744335, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59171, CurrentID=0, Status='live', Symbol='CG18810', Aliases='Dmel_ Dmel\\CG18810', description='CG18810 gene product from transcript CG18810-RA', other_designations='CG18810-PA|uncharacterized protein', map_location='38D2-38D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20648912, end_position_on_the_genomic_accession=20650079, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59172, CurrentID=0, Status='live', Symbol='Capr', Aliases='Dmel_CG18811, CG18811, Dmel\\CG18811, NEST:bs19b07, anon-WO0118547.253, dFAP', description='Caprin', other_designations='CG18811-PA|CG18811-PB|Capr-PA|Capr-PB|caprin|dFMRP-associated protein', map_location='75D5-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18668153, end_position_on_the_genomic_accession=18673324, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59173, CurrentID=0, Status='live', Symbol='CG18812', Aliases='Dmel_ CG12040, CG18813, Dmel\\CG18812', description='CG18812 gene product from transcript CG18812-RB', other_designations='CG18812-PA|CG18812-PB|CG18812-PC|CG18812-PD|CG18812-PE|CG18812-PF|uncharacterized protein', map_location='43E12-43E13', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=7716807, end_position_on_the_genomic_accession=7735622, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  59175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59175, CurrentID=0, Status='live', Symbol='CG18814', Aliases='Dmel_ Dmel\\CG18814', description='CG18814 gene product from transcript CG18814-RB', other_designations='CG18814-PB|uncharacterized protein', map_location='72E3-72E3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16369826, end_position_on_the_genomic_accession=16370945, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59176, CurrentID=0, Status='live', Symbol='CG18815', Aliases='Dmel_ APT1, Dmel\\CG18815, anon-EST:Posey232', description='CG18815 gene product from transcript CG18815-RD', other_designations='Acyl-protein thioesterase 1|CG18815-PA|CG18815-PB|CG18815-PC|CG18815-PD|CG18815-PE|uncharacterized protein', map_location='68D1-68D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11628037, end_position_on_the_genomic_accession=11630749, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  59177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59177, CurrentID=0, Status='live', Symbol='Tsp42Eb', Aliases='Dmel_CG18816, BEST:LP06288, CG18816, Dm.Tsp42Eb, Dmel\\CG18816, LP06288, Tetrasp42Eb', description='Tetraspanin 42Eb', other_designations='CG18816-PB|Tsp42Eb-PB|tetraspanin 42E|tetraspanin 42Eb', map_location='42E4-42E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6998848, end_position_on_the_genomic_accession=6999848, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  59178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59178, CurrentID=0, Status='live', Symbol='Tsp42Ea', Aliases='Dmel_CG18817, BcDNA:GH05668, BcDNA:GM15292, CG18742, CG18817, CG3364, Dm.Tsp42Ea, Dmel\\CG18817, GH05668, Q9U3V4', description='Tetraspanin 42Ea', other_designations='CG18817-PA|CG18817-PB|CG18817-PC|Tsp42Ea-PA|Tsp42Ea-PB|Tsp42Ea-PC|tetraspanin 42E|tetraspanin 42Ea', map_location='42E4-42E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6992676, end_position_on_the_genomic_accession=7004259, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  10178866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178866, CurrentID=0, Status='live', Symbol='CG42855', Aliases='Dmel_ Dmel\\CG42855', description='CG42855 gene product from transcript CG42855-RB', other_designations='CG42855-PB|uncharacterized protein', map_location='55E6-55E6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18687372, end_position_on_the_genomic_accession=18688641, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  59184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59184, CurrentID=0, Status='live', Symbol='CG18823', Aliases='Dmel_ Dmel\\CG18823', description='CG18823 gene product from transcript CG18823-RA', other_designations='CG18823-PA|uncharacterized protein', map_location='1D4-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=990518, end_position_on_the_genomic_accession=991184, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3772188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772188, CurrentID=0, Status='live', Symbol='CG33648', Aliases='Dmel_ Dmel\\CG33648', description='CG33648 gene product from transcript CG33648-RB', other_designations='CG33648-PB|CG33648-PC|uncharacterized protein', map_location='35B2-35B2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14524312, end_position_on_the_genomic_accession=14525656, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  10178870 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178870, CurrentID=0, Status='live', Symbol='CG42727', Aliases='Dmel_ CG12258, CG31392, Dmel\\CG42727, Dmel_CG31392', description='CG42727 gene product from transcript CG42727-RA', other_designations='CG42727-PA|uncharacterized protein', map_location='88F1-88F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15348048, end_position_on_the_genomic_accession=15350391, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  59214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59214, CurrentID=0, Status='live', Symbol='Or19a', Aliases='Dmel_CG18859, 19A.1, 19a, CG18859, DOR37, Dmel\\CG18859-1, Or19a', description='Odorant receptor 19a', other_designations='CG18859-PB|Or19a-PB|odorant receptor 19a', map_location='19B3-19B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20141819, end_position_on_the_genomic_accession=20143188, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  59215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59215, CurrentID=0, Status='live', Symbol='CG32164', Aliases='Dmel_ CG18219, CG18860, Dmel\\CG32164', description='CG32164 gene product from transcript CG32164-RB', other_designations='CG32164-PA|CG32164-PB|uncharacterized protein', map_location='73B1-73B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16603053, end_position_on_the_genomic_accession=16608146, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  59216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59216, CurrentID=0, Status='live', Symbol='CG32165', Aliases='Dmel_ CG18218, CG18861, Dmel\\CG32165, Imp4/5, anon-73Bb', description='CG32165 gene product from transcript CG32165-RC', other_designations='CG32165-PA|CG32165-PB|CG32165-PC|uncharacterized protein', map_location='73A11-73B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16595181, end_position_on_the_genomic_accession=16600459, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  59217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59217, CurrentID=0, Status='live', Symbol='CG18869', Aliases='Dmel_ Dmel\\CG18869, Tes36, anon-WO0140519.104, anon-WO0153538.77', description='CG18869 gene product from transcript CG18869-RA', other_designations='CG18869-PA|CG18869-PB|uncharacterized protein', map_location='64A5-64A5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4059770, end_position_on_the_genomic_accession=4061923, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59218 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59218, CurrentID=0, Status='live', Symbol='CG18870', Aliases='Dmel_ Dmel\\CG18870, anon-57Fa', description='CG18870 gene product from transcript CG18870-RB', other_designations='CG18870-PA|CG18870-PB|uncharacterized protein', map_location='57F4-57F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21613845, end_position_on_the_genomic_accession=21616907, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  59223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59223, CurrentID=0, Status='live', Symbol='CG18769', Aliases='Dmel_ CG8352, Dmel\\CG18769', description='CG18769 gene product from transcript CG18769-RH', other_designations='CG18769-PA|CG18769-PB|CG18769-PC|CG18769-PD|CG18769-PE|CG18769-PG|CG18769-PH|CG18769-PI|uncharacterized protein', map_location='65B4-65B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6550738, end_position_on_the_genomic_accession=6593940, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59224, CurrentID=0, Status='live', Symbol='CG17239', Aliases='Dmel_ Dmel\\CG17239, SP138', description='CG17239 gene product from transcript CG17239-RB', other_designations='CG17239-PA|CG17239-PB|uncharacterized protein', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2251510, end_position_on_the_genomic_accession=2252341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  59225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59225, CurrentID=0, Status='live', Symbol='CG17234', Aliases='Dmel_ Dmel\\CG17234, SP113', description='CG17234 gene product from transcript CG17234-RA', other_designations='CG17234-PA|uncharacterized protein', map_location='22D5-22D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2253045, end_position_on_the_genomic_accession=2253852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59226, CurrentID=0, Status='live', Symbol='CG17242', Aliases='Dmel_ Dmel\\CG17242, SPH187, anon-SAGE:Wang-110', description='CG17242 gene product from transcript CG17242-RB', other_designations='CG17242-PA|CG17242-PB|anon-SAGE:Wang-110|uncharacterized protein', map_location='22D6-22D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2268106, end_position_on_the_genomic_accession=2269140, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  59227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59227, CurrentID=0, Status='live', Symbol='CG17376', Aliases='Dmel_ CG18461, Dmel\\CG17376, bs29c06.y1', description='CG17376 gene product from transcript CG17376-RB', other_designations='CG17376-PA|CG17376-PB|CG17376-PC|uncharacterized protein', map_location='27B2-27B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6799946, end_position_on_the_genomic_accession=6800894, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59232, CurrentID=0, Status='live', Symbol='CG16741', Aliases='Dmel_ CG18463, Dmel\\CG16741', description='CG16741 gene product from transcript CG16741-RA', other_designations='CG16741-PA|uncharacterized protein', map_location='56F17-56F17', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20347041, end_position_on_the_genomic_accession=20347589, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59233, CurrentID=0, Status='live', Symbol='CG13713', Aliases='Dmel_ CG18465, Dmel\\CG13713, HL2', description='CG13713 gene product from transcript CG13713-RA', other_designations='CG13713-PA|uncharacterized protein', map_location='64B14-64B14', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4697601, end_position_on_the_genomic_accession=4698058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59235, CurrentID=0, Status='live', Symbol='Nplp3', Aliases='Dmel_CG13061, BcDNA:RH16169, CG13061, CG18502, Dmel\\CG13061, NPLP3, SHA, VVIamide, nplp3', description='Neuropeptide-like precursor 3', other_designations='CG13061-PA|Nplp3-PA|SHA peptide|VVIamide peptide', map_location='72E1-72E1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16318024, end_position_on_the_genomic_accession=16318690, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  59239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59239, CurrentID=0, Status='live', Symbol='CG18764', Aliases='Dmel_ Dmel\\CG18764', description='CG18764 gene product from transcript CG18764-RA', other_designations='CG18764-PA|uncharacterized protein', map_location='86E13-86E13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11626889, end_position_on_the_genomic_accession=11628452, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59240, CurrentID=0, Status='live', Symbol='GstD10', Aliases='Dmel_CG18548, CG17392, CG18548, DmGSTD10, Dmel\\CG18548, GSTD10, gstd10', description='Glutathione S transferase D10', other_designations='CG18548-PA|CG18548-PB|GstD10-PA|GstD10-PB|glutathione S transferase D10', map_location='87B8-87B8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12364776, end_position_on_the_genomic_accession=12365735, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59241, CurrentID=0, Status='live', Symbol='CG18530', Aliases='Dmel_ CG12294, Dmel\\CG18530', description='CG18530 gene product from transcript CG18530-RA', other_designations='CG18530-PA|uncharacterized protein', map_location='87C3-87C3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=12633359, end_position_on_the_genomic_accession=12634670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  59246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59246, CurrentID=0, Status='live', Symbol='CG18731', Aliases='Dmel_ Dmel\\CG18731', description='CG18731 gene product from transcript CG18731-RB', other_designations='CG18731-PA|CG18731-PB|uncharacterized protein', map_location='99B9-99B9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29725027, end_position_on_the_genomic_accession=29725690, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  59247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59247, CurrentID=0, Status='live', Symbol='Zwilch', Aliases='Dmel_CG18729, CG18639, CG18729, Dmel\\CG18729, zwilch', description='CG18729 gene product from transcript CG18729-RA', other_designations='CG18729-PA|Zwilch-PA|zwilch', map_location='100B6-100B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31210966, end_position_on_the_genomic_accession=31213376, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59253, CurrentID=0, Status='live', Symbol='CG18508', Aliases='Dmel_ CG4015, Dmel\\CG18508, EG:140G11.3', description='CG18508 gene product from transcript CG18508-RB', other_designations='CG18508-PA|CG18508-PB|uncharacterized protein', map_location='3C9-3C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3172268, end_position_on_the_genomic_accession=3173140, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740226, CurrentID=0, Status='live', Symbol='CG34264', Aliases='Dmel_ Dmel\\CG34264, HDC07730', description='CG34264 gene product from transcript CG34264-RC', other_designations='CG34264-PB|CG34264-PC|CG34264-PD|uncharacterized protein', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=317084, end_position_on_the_genomic_accession=317789, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  59260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59260, CurrentID=0, Status='live', Symbol='Tgs1', Aliases='Dmel_CG44890, 0967/13, CG31241, CG44890, CG8045, CT24102, DTL, DTLd, Dmel\\CG44890, Dmel_CG31241, Dromel_CG31241_FBtr0083570_dtl_mORF, Dtl, TGS1, dORF, dtl, l(3)S096713, tatl', description='Trimethylguanosine synthase 1 ortholog (H. sapiens)', other_designations='CG44890-PB|Tat like|Tgs1-PB|drosophila Tat-like|lethal (3) S096713|tat-like|trimethylguanosine synthase 1 ortholog', map_location='90F11-90F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18255530, end_position_on_the_genomic_accession=18257677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  59261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59261, CurrentID=0, Status='live', Symbol='CG3267', Aliases='Dmel_ Dmel\\CG3267, FBgn0042083, MCCC2_DROME, l(2)4524', description='CG3267 gene product from transcript CG3267-RA', other_designations='CG3267-PA', map_location='42C8-42C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=6746381, end_position_on_the_genomic_accession=6748954, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  10178885 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178885, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-Y', Aliases='Dmel_CG42771, CG42771, Dmel\\CG42771', description='CG42771 gene product from transcript CG42771-RB', other_designations='CG42771-PA|CG42771-PB|pre-mod(mdg4)-Y-PA|pre-mod(mdg4)-Y-PB', map_location='93D8-93D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21365609, end_position_on_the_genomic_accession=21368279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246388, CurrentID=0, Status='live', Symbol='CG30008', Aliases='Dmel_ CG12138, Dmel\\CG30008, anon-WO0140519.246', description='CG30008 gene product from transcript CG30008-RA', other_designations='CG30008-PA|uncharacterized protein', map_location='46C2-46C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9889878, end_position_on_the_genomic_accession=9891141, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  10178889 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178889, CurrentID=0, Status='live', Symbol='CG42763', Aliases='Dmel_ Dmel\\CG42763', description='CG42763 gene product from transcript CG42763-RA', other_designations='CG42763-PA|uncharacterized protein', map_location='28E7-28E7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8144391, end_position_on_the_genomic_accession=8145107, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740457, CurrentID=0, Status='live', Symbol='CG17162', Aliases='Dmel_ CG40460, Dmel\\CG17162, FBgn0058460, anon-20Ea', description='CG17162 gene product from transcript CG17162-RH', other_designations='CG17162-PB|CG17162-PC|CG17162-PD|CG17162-PE|CG17162-PF|CG17162-PG|CG17162-PH|uncharacterized protein', map_location='20E-20E', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=23084960, end_position_on_the_genomic_accession=23096735, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  10178890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178890, CurrentID=0, Status='live', Symbol='CG42713', Aliases='Dmel_ DMG4-MB6.chr2L.16.012, Dmel\\CG42713', description='CG42713 gene product from transcript CG42713-RA', other_designations='CG42713-PA|CG42713-PB|uncharacterized protein', map_location='29F3-29F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8891440, end_position_on_the_genomic_accession=8892180, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178891, CurrentID=0, Status='live', Symbol='B9d2', Aliases='Dmel_CG42730, CG42730, CG9227, Dmel\\CG42730, Dmel_CG9227', description='CG42730 gene product from transcript CG42730-RA', other_designations='B9d2-PA|CG42730-PA', map_location='26B7-26B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6127872, end_position_on_the_genomic_accession=6130651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  10178897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178897, CurrentID=0, Status='live', Symbol='CG42717', Aliases='Dmel_ DMG4-MB6.chr3L.27.007, DMG4-MB6.chr3L.27.007-MIP19202, Dmel\\CG42717, HDC10387', description='CG42717 gene product from transcript CG42717-RA', other_designations='CG42717-PA|uncharacterized protein', map_location='72D1-72D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=16032762, end_position_on_the_genomic_accession=16033190, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178898, CurrentID=0, Status='live', Symbol='CG42705', Aliases='Dmel_ CG47205, Dmel\\CG42705', description='CG42705 gene product from transcript CG42705-RB', other_designations='CG42705-PA|CG42705-PB|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7413869, end_position_on_the_genomic_accession=7414662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740575, CurrentID=0, Status='live', Symbol='CG34241', Aliases='Dmel_ Dmel\\CG34241, NEST:bs11f11, NEST:bs29g03, NEST:bs31h06', description='CG34241 gene product from transcript CG34241-RA', other_designations='CG34241-PA|NEST:bs11f11|NEST:bs29g03|NEST:bs31h06|uncharacterized protein', map_location='68F4-68F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=12127979, end_position_on_the_genomic_accession=12128461, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  318567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318567, CurrentID=0, Status='live', Symbol='CG31041', Aliases='Dmel_ Dmel\\CG31041', description='CG31041 gene product from transcript CG31041-RA', other_designations='CG31041-PA|uncharacterized protein', map_location='99B4-99B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29590070, end_position_on_the_genomic_accession=29590906, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835450, CurrentID=0, Status='live', Symbol='CG45545', Aliases='Dmel_ Dmel\\CG45545', description='CG45545 gene product from transcript CG45545-RA', other_designations='CG45545-PA|uncharacterized protein', map_location='84D9-84D9', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7552008, end_position_on_the_genomic_accession=7552298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  10178913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178913, CurrentID=0, Status='live', Symbol='CG42847', Aliases='Dmel_ Dmel\\CG42847', description='CG42847 gene product from transcript CG42847-RA', other_designations='CG42847-PA|uncharacterized protein', map_location='30B3-30B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9427870, end_position_on_the_genomic_accession=9428740, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740234, CurrentID=0, Status='live', Symbol='CG34284', Aliases='Dmel_ Dmel\\CG34284', description='CG34284 gene product from transcript CG34284-RA', other_designations='CG34284-PA|uncharacterized protein', map_location='91D4-91D4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18902338, end_position_on_the_genomic_accession=18902933, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740215, CurrentID=0, Status='live', Symbol='CG13177', Aliases='Dmel_ Dmel\\CG13177', description='CG13177 gene product from transcript CG13177-RB', other_designations='CG13177-PA|CG13177-PB|uncharacterized protein', map_location='48E4-48E4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12163675, end_position_on_the_genomic_accession=12164896, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  318207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318207, CurrentID=0, Status='live', Symbol='CG32783', Aliases='Dmel_ Dmel\\CG32783', description='CG32783 gene product from transcript CG32783-RB', other_designations='CG32783-PB|uncharacterized protein', map_location='3F9-3F9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3945043, end_position_on_the_genomic_accession=3948460, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  10178949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178949, CurrentID=0, Status='live', Symbol='CG42681', Aliases='Dmel_ Dmel\\CG42681', description='CG42681 gene product from transcript CG42681-RA', other_designations='CG42681-PA|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14744421, end_position_on_the_genomic_accession=14744817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37258, CurrentID=0, Status='live', Symbol='CG15125', Aliases='Dmel_ Dmel\\CG15125', description='CG15125 gene product from transcript CG15125-RA', other_designations='CG15125-PA|CG15125-PB|uncharacterized protein', map_location='56E1-56E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=19657573, end_position_on_the_genomic_accession=19659599, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5740242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740242, CurrentID=0, Status='live', Symbol='CG34185', Aliases='Dmel_ Dmel\\CG34185', description='CG34185 gene product from transcript CG34185-RA', other_designations='CG34185-PA|uncharacterized protein', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14373829, end_position_on_the_genomic_accession=14374395, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740141, CurrentID=0, Status='live', Symbol='CG34268', Aliases='Dmel_ Dmel\\CG34268', description='CG34268 gene product from transcript CG34268-RA', other_designations='CG34268-PA|uncharacterized protein', map_location='61C3-61C3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=500965, end_position_on_the_genomic_accession=501400, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740243, CurrentID=0, Status='live', Symbol='CG34203', Aliases='Dmel_ Dmel\\CG34203', description='CG34203 gene product from transcript CG34203-RA', other_designations='CG34203-PA|uncharacterized protein', map_location='57D6-57D7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21281063, end_position_on_the_genomic_accession=21281656, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  10178968 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178968, CurrentID=0, Status='live', Symbol='CG42781', Aliases='Dmel_ Dmel\\CG42781, MIP22495', description='CG42781 gene product from transcript CG42781-RA', other_designations='CG42781-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7382733, end_position_on_the_genomic_accession=7382990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32914, CurrentID=0, Status='live', Symbol='CG7884', Aliases='Dmel_ Dmel\\CG7884', description='CG7884 gene product from transcript CG7884-RA', other_designations='CG7884-PA|CG7884-PB', map_location='18B2-18B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19220622, end_position_on_the_genomic_accession=19235910, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246411, CurrentID=0, Status='live', Symbol='CG30039', Aliases='Dmel_ BcDNA:AT09538, Dmel\\CG30039, NEST:bs15e10, NEST:bs25c07', description='CG30039 gene product from transcript CG30039-RA', other_designations='CG30039-PA|uncharacterized protein', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12050977, end_position_on_the_genomic_accession=12051755, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246412, CurrentID=0, Status='live', Symbol='CG30043', Aliases='Dmel_ CG13161, Dmel\\CG30043', description='CG30043 gene product from transcript CG30043-RA', other_designations='CG30043-PA|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12359436, end_position_on_the_genomic_accession=12363165, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  39099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39099, CurrentID=0, Status='live', Symbol='CG44838', Aliases='Dmel_ CG14180, CG18632, CG32043, CG3628, Dmel\\CG44838, Dmel_CG14180, Dmel_CG32043, anon-WO0118547.368, anon-WO0118547.427', description='CG44838 gene product from transcript CG44838-RE', other_designations='CG44838-PD|CG44838-PE|CG44838-PF|CG44838-PG|CG44838-PH|CG44838-PI|rugose|uncharacterized protein', map_location='67B9-67B9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9461050, end_position_on_the_genomic_accession=9475475, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  246414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246414, CurrentID=0, Status='live', Symbol='CG30047', Aliases='Dmel_ Dmel\\CG30047', description='CG30047 gene product from transcript CG30047-RA', other_designations='CG30047-PA|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12351793, end_position_on_the_genomic_accession=12355413, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  5740254 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740254, CurrentID=0, Status='live', Symbol='CG34217', Aliases='Dmel_ Dmel\\CG34217', description='CG34217 gene product from transcript CG34217-RA', other_designations='CG34217-PA|uncharacterized protein', map_location='44C2-44C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8227990, end_position_on_the_genomic_accession=8228727, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  59169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=59169, CurrentID=0, Status='live', Symbol='CG18808', Aliases='Dmel_ Dmel\\CG18808', description='CG18808 gene product from transcript CG18808-RA', other_designations='CG18808-PA|uncharacterized protein', map_location='64B17-64B17', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4806053, end_position_on_the_genomic_accession=4807406, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246415, CurrentID=0, Status='live', Symbol='CG30048', Aliases='Dmel_ Dmel\\CG30048', description='CG30048 gene product from transcript CG30048-RD', other_designations='CG30048-PC|CG30048-PD|uncharacterized protein', map_location='49A2-49A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12394907, end_position_on_the_genomic_accession=12399142, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246416, CurrentID=0, Status='live', Symbol='CG30049', Aliases='Dmel_CG30049-PA, Dmel\\CG30049', description='CG30049 gene product from transcript CG30049-RA', other_designations='CG30049-PA|uncharacterized protein', map_location='49A1-49A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12355671, end_position_on_the_genomic_accession=12359087, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19835476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835476, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-C', Aliases='Dmel_CG44876, CG44876, Dmel\\CG44876', description='CG44876 gene product from transcript CG44876-RB', other_designations='CG44876-PB|CG44876-PC|pre-mod(mdg4)-C-PB|pre-mod(mdg4)-C-PC', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21362968, end_position_on_the_genomic_accession=21365290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  246424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246424, CurrentID=0, Status='live', Symbol='CG30058', Aliases='Dmel_ Dmel\\CG30058', description='CG30058 gene product from transcript CG30058-RA', other_designations='CG30058-PA|CG30058-PB|uncharacterized protein', map_location='49F9-49F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13209809, end_position_on_the_genomic_accession=13212237, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  317919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=317919, CurrentID=0, Status='live', Symbol='CG32214', Aliases='Dmel_ Dmel\\CG32214', description='CG32214 gene product from transcript CG32214-RC', other_designations='CG32214-PB|CG32214-PC|uncharacterized protein', map_location='76B3-76B3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19467346, end_position_on_the_genomic_accession=19468160, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462759, CurrentID=0, Status='live', Symbol='CG43925', Aliases='Dmel_ Dmel\\CG43925', description='CG43925 gene product from transcript CG43925-RA', other_designations='CG43925-PA|uncharacterized protein', map_location='34A8-34A8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13176423, end_position_on_the_genomic_accession=13176888, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39111, CurrentID=0, Status='live', Symbol='CG42673', Aliases='Dmel_ BcDNA:RE71517, BcDNA:RE71585, CG14178, CG17357, CG3179, CG32048, CG34272, Dm_3L:44995, Dmel\\CG42673, Dmel_CG32048, Dmel_CG34272', description='CG42673 gene product from transcript CG42673-RD', other_designations='CG42673-PB|CG42673-PC|CG42673-PD|uncharacterized protein', map_location='67C1-67C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9547220, end_position_on_the_genomic_accession=9596713, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19834718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834718, CurrentID=0, Status='live', Symbol='sud1', Aliases='Dmel_CG44254, CG18761, CG31120, CG44254, Dmel\\CG44254', description='sudestada1', other_designations='CG44254-PA|sud1-PA', map_location='96B6-96B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24915840, end_position_on_the_genomic_accession=24917864, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19834720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834720, CurrentID=0, Status='live', Symbol='CG44385', Aliases='Dmel_ Dmel\\CG44385', description='CG44385 gene product from transcript CG44385-RA', other_designations='CG44385-PA|uncharacterized protein', map_location='10A6-10A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11131340, end_position_on_the_genomic_accession=11132412, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19834721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834721, CurrentID=0, Status='live', Symbol='CG44405', Aliases='Dmel_ Dmel\\CG44405', description='CG44405 gene product from transcript CG44405-RA', other_designations='CG44405-PA|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16340392, end_position_on_the_genomic_accession=16342324, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  39114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39114, CurrentID=0, Status='live', Symbol='CG3222', Aliases='Dmel_ Dmel\\CG3222, NEST:bs17c12, bs17c12.y1', description='CG3222 gene product from transcript CG3222-RA', other_designations='CG3222-PA', map_location='67C2-67C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9598988, end_position_on_the_genomic_accession=9600341, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37399, CurrentID=0, Status='live', Symbol='CG9394', Aliases='Dmel_ Dmel\\CG9394, anon-EST:Posey227', description='CG9394 gene product from transcript CG9394-RA', other_designations='CG9394-PA|CG9394-PB', map_location='57C1-57C1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=21118394, end_position_on_the_genomic_accession=21120705, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835488, CurrentID=0, Status='live', Symbol='CG45122', Aliases='Dmel_ Dmel\\CG45122', description='CG45122 gene product from transcript CG45122-RA', other_designations='CG45122-PA|uncharacterized protein', map_location='87D9-87D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13044157, end_position_on_the_genomic_accession=13044529, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178835, CurrentID=0, Status='live', Symbol='CG42766', Aliases='Dmel_ Dmel\\CG42766, FBgn0261834', description='CG42766 gene product from transcript CG42766-RB', other_designations='CG42766-PA|CG42766-PB|uncharacterized protein', map_location='14A4-14A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15976868, end_position_on_the_genomic_accession=15977660, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740271, CurrentID=0, Status='live', Symbol='Vml', Aliases='Dmel_CG34333, CG34333, CRM-80, Dmel\\CG34333, vml', description='Vitelline membrane-like', other_designations='CG34333-PA|CG34333-PB|Vml-PA|Vml-PB|vitelline membrane-like', map_location='2F4-2F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2275822, end_position_on_the_genomic_accession=2277781, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246561, CurrentID=0, Status='live', Symbol='CG30355', Aliases='Dmel_ BcDNA:AT24812, Dmel\\CG30355, NEST:bs04f07, NEST:bs06e10, NEST:bs10b08, NEST:bs23a07, NEST:bs23e07', description='CG30355 gene product from transcript CG30355-RA', other_designations='CG30355-PA|uncharacterized protein', map_location='44E2-44E2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8679146, end_position_on_the_genomic_accession=8679953, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31953 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31953, CurrentID=0, Status='live', Symbol='CG34408', Aliases='Dmel_ BcDNA:GH09358, CG2967, CG32689, Dmel\\CG34408, Dmel_CG2967, Dmel_CG32689', description='CG34408 gene product from transcript CG34408-RF', other_designations='CG34408-PA|CG34408-PB|CG34408-PC|CG34408-PD|CG34408-PE|CG34408-PF|CG34408-PG|CG34408-PH|CG34408-PI|CG34408-PJ|uncharacterized protein', map_location='9B4-9B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10220110, end_position_on_the_genomic_accession=10230088, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  19834757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834757, CurrentID=0, Status='live', Symbol='CG44335', Aliases='Dmel_ Dmel\\CG44335, lincRNA.706, lincRNA.707', description='CG44335 gene product from transcript CG44335-RA', other_designations='CG44335-PA|uncharacterized protein', map_location='84D3-84D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7325597, end_position_on_the_genomic_accession=7326459, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19834760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834760, CurrentID=0, Status='live', Symbol='CG45076', Aliases='Dmel_ CG6544, Dmel\\CG45076', description='CG45076 gene product from transcript CG45076-RH', other_designations='CG45076-PA|CG45076-PE|CG45076-PH|CG45076-PI|uncharacterized protein', map_location='86C6-86C6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10768001, end_position_on_the_genomic_accession=10777374, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19834761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834761, CurrentID=0, Status='live', Symbol='CG44249', Aliases='Dmel_ 65K, CG11478, CG30327, CG42257, Dmel\\CG44249, cg11478', description='CG44249 gene product from transcript CG44249-RA', other_designations='CG44249-PA|uncharacterized protein', map_location='58B1-58B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22066883, end_position_on_the_genomic_accession=22068372, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740585, CurrentID=0, Status='live', Symbol='CG34224', Aliases='Dmel_ Dmel\\CG34224', description='CG34224 gene product from transcript CG34224-RA', other_designations='CG34224-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11237172, end_position_on_the_genomic_accession=11237644, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740281, CurrentID=0, Status='live', Symbol='CG34219', Aliases='Dmel_ Dmel\\CG34219', description='CG34219 gene product from transcript CG34219-RA', other_designations='CG34219-PA|CG34219-PB|uncharacterized protein', map_location='44F7-44F7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8917708, end_position_on_the_genomic_accession=8918877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740536, CurrentID=0, Status='live', Symbol='CG34310', Aliases='Dmel_ Dmel\\CG34310', description='CG34310 gene product from transcript CG34310-RD', other_designations='CG34310-PA|CG34310-PC|CG34310-PD|CG34310-PE|uncharacterized protein', map_location='26F6-26F7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6663664, end_position_on_the_genomic_accession=6674780, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19834795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834795, CurrentID=0, Status='live', Symbol='CG45012', Aliases='Dmel_ Dmel\\CG45012', description='CG45012 gene product from transcript CG45012-RA', other_designations='CG45012-PA|uncharacterized protein', map_location='33A2-33A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11834608, end_position_on_the_genomic_accession=11834874, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462776, CurrentID=0, Status='live', Symbol='CG44140', Aliases='Dmel_ Dmel\\CG44140', description='CG44140 gene product from transcript CG44140-RA', other_designations='CG44140-PA|uncharacterized protein', map_location='35D3-35D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15694038, end_position_on_the_genomic_accession=15694471, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50039, CurrentID=0, Status='live', Symbol='CG34281', Aliases='Dmel_ CG14328, Dmel\\CG34281', description='CG34281 gene product from transcript CG34281-RA', other_designations='CG34281-PA|uncharacterized protein', map_location='90B6-90B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17612887, end_position_on_the_genomic_accession=17613404, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462778, CurrentID=0, Status='live', Symbol='CG43760', Aliases='Dmel_ Dmel\\CG43760', description='CG43760 gene product from transcript CG43760-RA', other_designations='CG43760-PA|uncharacterized protein', map_location='35D4-35D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15776442, end_position_on_the_genomic_accession=15776852, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740285, CurrentID=0, Status='live', Symbol='CG34173', Aliases='Dmel_ Dmel\\CG34173', description='CG34173 gene product from transcript CG34173-RA', other_designations='CG34173-PA|uncharacterized protein', map_location='40D5-40D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=22053834, end_position_on_the_genomic_accession=22054302, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50042, CurrentID=0, Status='live', Symbol='CG14310', Aliases='Dmel_ Dmel\\CG14310', description='CG14310 gene product from transcript CG14310-RC', other_designations='CG14310-PC|uncharacterized protein', map_location='91A3-91A4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18372928, end_position_on_the_genomic_accession=18373329, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19834814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834814, CurrentID=0, Status='live', Symbol='CG44314', Aliases='Dmel_ Dmel\\CG44314', description='CG44314 gene product from transcript CG44314-RB', other_designations='CG44314-PA|CG44314-PB|CG44314-PC|uncharacterized protein', map_location='48D8-48D8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=12052944, end_position_on_the_genomic_accession=12056547, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  38814 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38814, CurrentID=0, Status='live', Symbol='CG14834', Aliases='Dmel_ CT34650, Dmel\\CG14834', description='CG14834 gene product from transcript CG14834-RA', other_designations='CG14834-PA|uncharacterized protein', map_location='65F5-65F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=7345698, end_position_on_the_genomic_accession=7347123, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19835713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835713, CurrentID=0, Status='live', Symbol='CG45081', Aliases='Dmel_ Dmel\\CG45081', description='CG45081 gene product from transcript CG45081-RB', other_designations='CG45081-PA|CG45081-PB|uncharacterized protein', map_location='76A3-76A3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=19266061, end_position_on_the_genomic_accession=19266838, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50044, CurrentID=0, Status='live', Symbol='CG17193', Aliases='Dmel_ BcDNA:RE41733, Dmel\\CG17193', description='CG17193 gene product from transcript CG17193-RB', other_designations='CG17193-PA|CG17193-PB|CG17193-PC|uncharacterized protein', map_location='92C2-92C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20036421, end_position_on_the_genomic_accession=20056097, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19834823 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834823, CurrentID=0, Status='live', Symbol='CG45544', Aliases='Dmel_ Dmel\\CG45544, lincRNA.886', description='CG45544 gene product from transcript CG45544-RA', other_designations='CG45544-PA|CG45544-PB|uncharacterized protein', map_location='99A5-99A5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29256485, end_position_on_the_genomic_accession=29257425, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19834824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834824, CurrentID=0, Status='live', Symbol='CG45101', Aliases='Dmel_ Dmel\\CG45101, Dmel_SuUR_uORF', description='CG45101 gene product from transcript CG45101-RA', other_designations='CG45101-PA|uncharacterized protein', map_location='68A4-68A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11064991, end_position_on_the_genomic_accession=11069025, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32035 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32035, CurrentID=0, Status='live', Symbol='CG1552', Aliases='Dmel_ Dmel\\CG1552, anon-EST:Posey230', description='CG1552 gene product from transcript CG1552-RC', other_designations='CG1552-PA|CG1552-PB|CG1552-PC', map_location='10A3-10A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11052171, end_position_on_the_genomic_accession=11053556, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19834832 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834832, CurrentID=0, Status='live', Symbol='CG44869', Aliases='Dmel_ Dmel\\CG44869', description='CG44869 gene product from transcript CG44869-RA', other_designations='CG44869-PA|uncharacterized protein', map_location='35D2-35D2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15417929, end_position_on_the_genomic_accession=15418361, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19834833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834833, CurrentID=0, Status='live', Symbol='CG45488', Aliases='Dmel_ Dmel\\CG45488', description='CG45488 gene product from transcript CG45488-RA', other_designations='CG45488-PA|uncharacterized protein', map_location='19A2-19A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19919641, end_position_on_the_genomic_accession=19920435, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  5740610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740610, CurrentID=0, Status='live', Symbol='CG34254', Aliases='Dmel_ Dmel\\CG34254', description='CG34254 gene product from transcript CG34254-RA', other_designations='CG34254-PA|uncharacterized protein', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18680409, end_position_on_the_genomic_accession=18680779, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  37509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37509, CurrentID=0, Status='live', Symbol='CG13500', Aliases='Dmel_ Dmel\\CG13500', description='CG13500 gene product from transcript CG13500-RA', other_designations='CG13500-PA|uncharacterized protein', map_location='58B1-58B1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22056467, end_position_on_the_genomic_accession=22057681, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50048, CurrentID=0, Status='live', Symbol='CG15922', Aliases='Dmel_ BcDNA:RH34302, Dmel\\CG15922', description='CG15922 gene product from transcript CG15922-RB', other_designations='CG15922-PA|CG15922-PB|uncharacterized protein', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20559708, end_position_on_the_genomic_accession=20560280, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19834845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834845, CurrentID=0, Status='live', Symbol='CG45691', Aliases='Dmel_ Dmel\\CG45691, lincRNA.169', description='CG45691 gene product from transcript CG45691-RA', other_designations='CG45691-PA|uncharacterized protein', map_location='36C6-36C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17320250, end_position_on_the_genomic_accession=17320712, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19835998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835998, CurrentID=0, Status='live', Symbol='CG45061', Aliases='Dmel_ CG15245, CG15246, CG32696, CG42262, Dmel\\CG45061', description='CG45061 gene product from transcript CG45061-RA', other_designations='CG45061-PA|uncharacterized protein', map_location='9A2-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9871098, end_position_on_the_genomic_accession=9871702, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740392, CurrentID=0, Status='live', Symbol='CG34162', Aliases='Dmel_ Dmel\\CG34162', description='CG34162 gene product from transcript CG34162-RB', other_designations='CG34162-PB|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11531820, end_position_on_the_genomic_accession=11533756, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  5740297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740297, CurrentID=0, Status='live', Symbol='CG34256', Aliases='Dmel_ Dmel\\CG34256', description='CG34256 gene product from transcript CG34256-RB', other_designations='CG34256-PB|uncharacterized protein', map_location='75F2-75F2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18988736, end_position_on_the_genomic_accession=18989815, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  192481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192481, CurrentID=0, Status='live', Symbol='Gr59c', Aliases='Dmel_CG30186, 59D.2, 59D2, CG13543, CG30186, Dmel\\CG30186, GR59D.2, GR59c, Gr59D2', description='Gustatory receptor 59c', other_designations='CG30186-PA|Gr59c-PA|gustatory receptor 59D.2|gustatory receptor 59c', map_location='59C4-59C4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23102477, end_position_on_the_genomic_accession=23103733, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  192506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192506, CurrentID=0, Status='live', Symbol='CG5986', Aliases='Dmel_ BEST:LP03871, Dmel\\CG5986', description='CG5986 gene product from transcript CG5986-RA', other_designations='CG5986-PA|CG5986-PB', map_location='95D10-95D10', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24047790, end_position_on_the_genomic_accession=24049168, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  192507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192507, CurrentID=0, Status='live', Symbol='CG4747', Aliases='Dmel_ BEST:GH19401, BEST:LD22483, BEST:LD29743, Dm_AI388468, Dmel\\CG4747, LD29743', description='CG4747 gene product from transcript CG4747-RB', other_designations='CG4747-PA|CG4747-PB|CG4747-PC', map_location='30F5-30F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10003832, end_position_on_the_genomic_accession=10008889, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  192508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192508, CurrentID=0, Status='live', Symbol='CG5180', Aliases='Dmel_ BEST:LD05530, Dmel\\CG5180, LD05530', description='CG5180 gene product from transcript CG5180-RC', other_designations='CG5180-PB|CG5180-PC', map_location='92F1-92F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=20557930, end_position_on_the_genomic_accession=20559760, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192509, CurrentID=0, Status='live', Symbol='CG12084', Aliases='Dmel_ BEST:LD02105, Dmel\\CG12084, LD02105', description='CG12084 gene product from transcript CG12084-RB', other_designations='CG12084-PA|CG12084-PB|uncharacterized protein', map_location='62A2-62A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1490951, end_position_on_the_genomic_accession=1496643, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  5740807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740807, CurrentID=0, Status='live', Symbol='CG34225', Aliases='Dmel_ Dmel\\CG34225', description='CG34225 gene product from transcript CG34225-RA', other_designations='CG34225-PA|uncharacterized protein', map_location='47E1-47E1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11245406, end_position_on_the_genomic_accession=11245995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14462793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462793, CurrentID=0, Status='live', Symbol='CG43645', Aliases='Dmel_ Dmel\\CG43645, mlncRNA36C10', description='CG43645 gene product from transcript CG43645-RA', other_designations='CG43645-PA|uncharacterized protein', map_location='36C10-36C10', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=17486121, end_position_on_the_genomic_accession=17486365, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  192528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192528, CurrentID=0, Status='live', Symbol='CG31381', Aliases='Dmel_ BEST:LD10347, Dmel\\CG31381', description='CG31381 gene product from transcript CG31381-RA', other_designations='CG31381-PA|CG31381-PB|uncharacterized protein', map_location='96B2-96B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24877045, end_position_on_the_genomic_accession=24878827, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  3772203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772203, CurrentID=0, Status='live', Symbol='His2B:CG33904', Aliases='Dmel_CG33904, CG33904, Dmel\\CG33904', description='CG33904 gene product from transcript CG33904-RA', other_designations='CG33904-PA|His2B:CG33904-PA|histone H2B', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21438170, end_position_on_the_genomic_accession=21438764, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  192535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192535, CurrentID=0, Status='live', Symbol='CG32032', Aliases='Dmel_ BEST:GH25183, BcDNA:RH42907, Dmel\\CG32032', description='CG32032 gene product from transcript CG32032-RA', other_designations='CG32032-PA|uncharacterized protein', map_location='66F5-66F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9093053, end_position_on_the_genomic_accession=9094856, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19834892 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834892, CurrentID=0, Status='live', Symbol='CG45080', Aliases='Dmel_ Dmel\\CG45080', description='CG45080 gene product from transcript CG45080-RA', other_designations='CG45080-PA|uncharacterized protein', map_location='87E5-87E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13318991, end_position_on_the_genomic_accession=13319270, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  64867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64867, CurrentID=0, Status='live', Symbol='nol', Aliases='Dmel_CG32077, CG32077, Dmel\\CG32077', description='no optic lobe', other_designations='CG32077-PA|nol-PA', map_location='68A8-68A8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11130835, end_position_on_the_genomic_accession=11131402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  192611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=192611, CurrentID=0, Status='live', Symbol='CG32388', Aliases='Dmel_ BEST:GH26265, BcDNA:AT03632, BcDNA:GH26265, Dmel\\CG32388, anon-SAGE:Wang-131, anon-WO0140519.36', description='CG32388 gene product from transcript CG32388-RA', other_designations='CG32388-PA|anon-SAGE:Wang-131|uncharacterized protein', map_location='65D5-65D5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6983473, end_position_on_the_genomic_accession=6984586, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64872, CurrentID=0, Status='live', Symbol='MESK4', Aliases='Dmel_CG31447, BcDNA:LD15753, CG31447, Dmel\\CG31447', description='Misexpression suppressor of KSR 4', other_designations='CG31447-PA|CG31447-PB|MESK4-PA|MESK4-PB|misexpression suppressor of KSR 4', map_location='89E6-89E6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16988814, end_position_on_the_genomic_accession=16990419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37567, CurrentID=0, Status='live', Symbol='CG10384', Aliases='Dmel_ Dmel\\CG10384, cg10384', description='CG10384 gene product from transcript CG10384-RE', other_designations='CG10384-PD|CG10384-PE|uncharacterized protein', map_location='58E2-58E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22454741, end_position_on_the_genomic_accession=22472995, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  64874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64874, CurrentID=0, Status='live', Symbol='eIF2D', Aliases='Dmel_CG31426, BcDNA:LD24540, CG31426, Dmel\\CG31426, ligatin', description='eukaryotic translation initiation factor 2D', other_designations='CG31426-PA|eIF2D-PA|ligatin', map_location='98F13-98F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29149896, end_position_on_the_genomic_accession=29151794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32107, CurrentID=0, Status='live', Symbol='nod', Aliases='Dmel_CG1763, CG1763, DmNod, Dmel\\CG1763, DnMod, KIF 22, KIF22, NOD, Nod, l(1)TW6', description='no distributive disjunction', other_designations='CG1763-PA|CG1763-PB|No distributive disjunction|kinesin-10|kineskin-10|nod-PA|nod-PB', map_location='1-36 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11580772, end_position_on_the_genomic_accession=11585750, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  37569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37569, CurrentID=0, Status='live', Symbol='CG4752', Aliases='Dmel_ Dmel\\CG4752', description='CG4752 gene product from transcript CG4752-RA', other_designations='CG4752-PA', map_location='58E3-58E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22478379, end_position_on_the_genomic_accession=22483054, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  3355093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355093, CurrentID=0, Status='live', Symbol='CG40470', Aliases='Dmel_ Dmel\\CG40470', description='CG40470 gene product from transcript CG40470-RA', other_designations='CG40470-PA|uncharacterized protein', map_location='', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=23839860, end_position_on_the_genomic_accession=23986410, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  64877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64877, CurrentID=0, Status='live', Symbol='cpx', Aliases='Dmel_CG32490, BcDNA:GH27718, BcDNA:HL04936, BcDNA:RH26459, BcDNA:RH31085, CG32490, Cpx, DmCpx, Dmel\\CG32490, dmCplx', description='complexin', other_designations='CG32490-PA|CG32490-PAA|CG32490-PAB|CG32490-PC|CG32490-PE|CG32490-PG|CG32490-PH|CG32490-PI|CG32490-PJ|CG32490-PK|CG32490-PL|CG32490-PM|CG32490-PO|CG32490-PP|CG32490-PS|CG32490-PU|CG32490-PV|CG32490-PW|CG32490-PX|CG32490-PY|CG32490-PZ|cpx-PA|cpx-PAA|cpx-PAB|cpx-PC|cpx-PE|cpx-PG|cpx-PH|cpx-PI|cpx-PJ|cpx-PK|cpx-PL|cpx-PM|cpx-PO|cpx-PP|cpx-PS|cpx-PU|cpx-PV|cpx-PW|cpx-PX|cpx-PY|cpx-PZ', map_location='82A1-82A3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=4280221, end_position_on_the_genomic_accession=4304604, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  32111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32111, CurrentID=0, Status='live', Symbol='CG2247', Aliases='Dmel_ Dmel\\CG2247, FBXL5', description='CG2247 gene product from transcript CG2247-RA', other_designations='CG2247-PA|CG2247-PB|CG2247-PC|CG2247-PD', map_location='10C10-10D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11606290, end_position_on_the_genomic_accession=11611652, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  64880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64880, CurrentID=0, Status='live', Symbol='chico', Aliases='Dmel_CG5686, BcDNA.GH11263, BcDNA:GH11263, CG5686, CHICO, Chico, Dmel\\CG5686, IRS, IRS/ anon-WO0078940.1, anon-WO0078940.4, chic, dIRS, fs(2)4, fs(2)ry4', description='CG5686 gene product from transcript CG5686-RB', other_designations='CG5686-PA|CG5686-PB|CG5686-PC|chico-PA|chico-PB|chico-PC|flipper|insulin receptor substrate|insulin-like receptor substrate', map_location='31B1-31B1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10242512, end_position_on_the_genomic_accession=10247078, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19834907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834907, CurrentID=0, Status='live', Symbol='CG44477', Aliases='Dmel_ Dmel\\CG44477', description='CG44477 gene product from transcript CG44477-RA', other_designations='CG44477-PA|uncharacterized protein', map_location='32F4-32F4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11663259, end_position_on_the_genomic_accession=11663872, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50061, CurrentID=0, Status='live', Symbol='CG13841', Aliases='Dmel_ Dmel\\CG13841', description='CG13841 gene product from transcript CG13841-RA', other_designations='CG13841-PA|CG13841-PB|uncharacterized protein', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22766732, end_position_on_the_genomic_accession=22767699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19834908 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834908, CurrentID=0, Status='live', Symbol='CG45071', Aliases='Dmel_ CG6854, Dmel\\CG45071, Polypeptide A', description='CG45071 gene product from transcript CG45071-RA', other_designations='CG45071-PA|CG45071-PB|uncharacterized protein', map_location='71B4-71B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15105882, end_position_on_the_genomic_accession=15110730, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14462799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462799, CurrentID=0, Status='live', Symbol='CG43731', Aliases='Dmel_ Dmel\\CG43731', description='CG43731 gene product from transcript CG43731-RA', other_designations='CG43731-PA|uncharacterized protein', map_location='37C1-37C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19102044, end_position_on_the_genomic_accession=19102909, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  2768972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=2768972, CurrentID=0, Status='live', Symbol='CG33271', Aliases='Dmel_ CG14139, CG32084, CG6100, Dmel\\CG33271', description='CG33271 gene product from transcript CG33271-RA', other_designations='CG33271-PA|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11577006, end_position_on_the_genomic_accession=11577495, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  41695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=41695, CurrentID=0, Status='live', Symbol='CG14362', Aliases='Dmel_ Dmel\\CG14362', description='CG14362 gene product from transcript CG14362-RA', other_designations='CG14362-PA|uncharacterized protein', map_location='88A1-88A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13854599, end_position_on_the_genomic_accession=13855454, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19834929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834929, CurrentID=0, Status='live', Symbol='CG45428', Aliases='Dmel_ Dmel\\CG45428', description='CG45428 gene product from transcript CG45428-RA', other_designations='CG45428-PA|uncharacterized protein', map_location='80A4-80A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=22836555, end_position_on_the_genomic_accession=22837070, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19834930 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834930, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-I', Aliases='Dmel_CG44883, CG44883, Dmel\\CG44883', description='CG44883 gene product from transcript CG44883-RA', other_designations='CG44883-PA|CG44883-PB|pre-mod(mdg4)-I-PA|pre-mod(mdg4)-I-PB', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21355692, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19834931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834931, CurrentID=0, Status='live', Symbol='CG45085', Aliases='Dmel_ Dmel\\CG45085, Dromel_CG8026_FBtr0088611_uORF', description='CG45085 gene product from transcript CG45085-RB', other_designations='CG45085-PA|CG45085-PB|uncharacterized protein', map_location='45B3-45B3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=9186670, end_position_on_the_genomic_accession=9192096, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19834937 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834937, CurrentID=0, Status='live', Symbol='CG45075', Aliases='Dmel_ CG18783, CG9167, Dmel\\CG45075', description='CG45075 gene product from transcript CG45075-RA', other_designations='CG45075-PA|uncharacterized protein', map_location='26B5-26B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=6091412, end_position_on_the_genomic_accession=6096498, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19834939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834939, CurrentID=0, Status='live', Symbol='CG44438', Aliases='Dmel_ Dmel\\CG44438, lincRNA.249', description='CG44438 gene product from transcript CG44438-RA', other_designations='CG44438-PA|uncharacterized protein', map_location='42A2-42A2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5879196, end_position_on_the_genomic_accession=5880395, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  37609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37609, CurrentID=0, Status='live', Symbol='Obp58d', Aliases='Dmel_CG13519, CG13519, DmelObp58d, Dmel\\CG13519', description='Odorant-binding protein 58d', other_designations='CG13519-PA|Obp58d-PA', map_location='59A3-59A3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22706934, end_position_on_the_genomic_accession=22707772, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  260658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=260658, CurrentID=0, Status='live', Symbol='CG31819', Aliases='Dmel_ BG:DS02795.3, BcDNA:AT12868, Dmel\\CG31819', description='CG31819 gene product from transcript CG31819-RA', other_designations='CG31819-PA|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16211459, end_position_on_the_genomic_accession=16213304, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  260659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=260659, CurrentID=0, Status='live', Symbol='CG31817', Aliases='Dmel_ BG:DS02740.18, DS02740.18, Dmel\\CG31817, anon-WO03042407.98, anon-WO03042407.99', description='CG31817 gene product from transcript CG31817-RB', other_designations='CG31817-PB|uncharacterized protein', map_location='35F1-35F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16330560, end_position_on_the_genomic_accession=16337885, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3772206 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3772206, CurrentID=0, Status='live', Symbol='CheA56a', Aliases='Dmel_CG33724, CG33724, Dmel\\CG33724', description='Chemosensory protein A 56a', other_designations='CG33724-PA|CheA56a-PA|chemosensory protein A 56a', map_location='55E3-55E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18640194, end_position_on_the_genomic_accession=18640802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  8673967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673967, CurrentID=0, Status='live', Symbol='CG42536', Aliases='Dmel_ Dmel\\CG42536', description='CG42536 gene product from transcript CG42536-RA', other_designations='CG42536-PA|uncharacterized protein', map_location='67E5-67E5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10629083, end_position_on_the_genomic_accession=10630069, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19834964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19834964, CurrentID=0, Status='live', Symbol='CG45087', Aliases='Dmel_ Dmel\\CG45087, Dromel_CG17725_FBtr0086701_pepck_uORF', description='CG45087 gene product from transcript CG45087-RA', other_designations='CG45087-PA|uncharacterized protein', map_location='55D3-55D3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18536767, end_position_on_the_genomic_accession=18539416, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50074, CurrentID=0, Status='live', Symbol='CG13643', Aliases='Dmel_ CG13642, Dmel\\CG13643', description='CG13643 gene product from transcript CG13643-RE', other_designations='CG13643-PB|CG13643-PC|CG13643-PD|CG13643-PE|uncharacterized protein', map_location='96B4-96B5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24897685, end_position_on_the_genomic_accession=24906004, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  318603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318603, CurrentID=0, Status='live', Symbol='CG31131', Aliases='Dmel_ Dmel\\CG31131', description='CG31131 gene product from transcript CG31131-RB', other_designations='CG31131-PB|uncharacterized protein', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27961992, end_position_on_the_genomic_accession=27963219, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740319, CurrentID=0, Status='live', Symbol='CG34462', Aliases='Dmel_ CG7072, Cpr66Ca, DmelCpr66Ca, Dmel\\CG34462', description='CG34462 gene product from transcript CG34462-RA', other_designations='CG34462-PA|uncharacterized protein', map_location='66C7-66C7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8328583, end_position_on_the_genomic_accession=8330237, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  37657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37657, CurrentID=0, Status='live', Symbol='CG44252', Aliases='Dmel_ CG30194, Dmel\\CG44252', description='CG44252 gene product from transcript CG44252-RE', other_designations='CG44252-PB|CG44252-PD|CG44252-PE|uncharacterized protein', map_location='59C2-59C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23055474, end_position_on_the_genomic_accession=23061549, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  50080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=50080, CurrentID=0, Status='live', Symbol='CG14244', Aliases='Dmel_ Dmel\\CG14244', description='CG14244 gene product from transcript CG14244-RB', other_designations='CG14244-PB|uncharacterized protein', map_location='97C5-97C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=26726133, end_position_on_the_genomic_accession=26726681, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19835022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835022, CurrentID=0, Status='live', Symbol='CG44270', Aliases='Dmel_ Dmel\\CG44270, lincRNA.207', description='CG44270 gene product from transcript CG44270-RA', other_designations='CG44270-PA|uncharacterized protein', map_location='38C9-38C9', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20531066, end_position_on_the_genomic_accession=20531833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32234, CurrentID=0, Status='live', Symbol='CG3775', Aliases='Dmel_ BEST:GH06882, Dmel\\CG3775', description='CG3775 gene product from transcript CG3775-RA', other_designations='CG3775-PA|CG3775-PB|zinc metalloprotease', map_location='11C2-11C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12801661, end_position_on_the_genomic_accession=12804232, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  261614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261614, CurrentID=0, Status='live', Symbol='CG31867', Aliases='Dmel_ BcDNA:AT02643, CG6275, Dmel\\CG31867', description='CG31867 gene product from transcript CG31867-RA', other_designations='CG31867-PA|CG31867-PB|uncharacterized protein', map_location='32D5-32D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11144579, end_position_on_the_genomic_accession=11145417, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  261615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261615, CurrentID=0, Status='live', Symbol='CG31806', Aliases='Dmel_ CG4855, Dmel\\CG31806', description='CG31806 gene product from transcript CG31806-RB', other_designations='CG31806-PA|CG31806-PB|uncharacterized protein', map_location='36B6-36B6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=16984978, end_position_on_the_genomic_accession=16985634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  261616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261616, CurrentID=0, Status='live', Symbol='CG31778', Aliases='Dmel_ BcDNA:HL03084, Dmel\\CG31778', description='CG31778 gene product from transcript CG31778-RA', other_designations='CG31778-PA|CG31778-PB|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3704106, end_position_on_the_genomic_accession=3705488, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740437, CurrentID=0, Status='live', Symbol='CG34196', Aliases='Dmel_ Dmel\\CG34196', description='CG34196 gene product from transcript CG34196-RA', other_designations='CG34196-PA|uncharacterized protein', map_location='55B5-55B5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18109985, end_position_on_the_genomic_accession=18110494, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  261617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261617, CurrentID=0, Status='live', Symbol='CG31777', Aliases='Dmel_ BEST:RH19272, Dmel\\CG31777', description='CG31777 gene product from transcript CG31777-RA', other_designations='CG31777-PA|uncharacterized protein', map_location='24C1-24C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3706551, end_position_on_the_genomic_accession=3707063, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  261618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261618, CurrentID=0, Status='live', Symbol='CG31740', Aliases='Dmel_ BEST:GH22513, BcDNA:AT23551, Dmel\\CG31740', description='CG31740 gene product from transcript CG31740-RA', other_designations='CG31740-PA|uncharacterized protein', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18008953, end_position_on_the_genomic_accession=18009694, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  261620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261620, CurrentID=0, Status='live', Symbol='CG31709', Aliases='Dmel_ BEST:LP04712, Dmel\\CG31709', description='CG31709 gene product from transcript CG31709-RB', other_designations='CG31709-PA|CG31709-PB|uncharacterized protein', map_location='30B12-30B12', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=9578120, end_position_on_the_genomic_accession=9579638, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  3771932 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771932, CurrentID=0, Status='live', Symbol='CG33702', Aliases='Dmel_ Dmel\\CG33702', description='CG33702 gene product from transcript CG33702-RA', other_designations='CG33702-PA|uncharacterized protein', map_location='67B11-67B11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9512797, end_position_on_the_genomic_accession=9513409, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  261621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261621, CurrentID=0, Status='live', Symbol='CG31706', Aliases='Dmel_ CG6528, Dmel\\CG31706', description='CG31706 gene product from transcript CG31706-RB', other_designations='CG31706-PA|CG31706-PB|uncharacterized protein', map_location='32F2-32F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11529038, end_position_on_the_genomic_accession=11531052, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32247, CurrentID=0, Status='live', Symbol='CG15721', Aliases='Dmel_ Dmel\\CG15721', description='CG15721 gene product from transcript CG15721-RB', other_designations='CG15721-PB|uncharacterized protein', map_location='11D5-11D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12970711, end_position_on_the_genomic_accession=12973967, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  261626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261626, CurrentID=0, Status='live', Symbol='CG31495', Aliases='Dmel_ CG9931, Dmel\\CG31495', description='CG31495 gene product from transcript CG31495-RA', other_designations='CG31495-PA|CG31495-PB|uncharacterized protein', map_location='87F15-87F15', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13838240, end_position_on_the_genomic_accession=13839934, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32252, CurrentID=0, Status='live', Symbol='CG12715', Aliases='Dmel_ Dmel\\CG12715', description='CG12715 gene product from transcript CG12715-RB', other_designations='CG12715-PA|CG12715-PB|uncharacterized protein', map_location='11D8-11D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13052847, end_position_on_the_genomic_accession=13055328, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32253, CurrentID=0, Status='live', Symbol='rad', Aliases='Dmel_CG44424, CG15720, CG18646, CG42628, CG42629, CG4346, CG4368, CG44424, Dmel\\CG44424, Dmel_CG15720, Dmel_CG18646, Dmel_CG42628, Dmel_CG42629, Dmel_CG4346, Rad, rsh', description='radish', other_designations='CG44424-PI|CG44424-PJ|CG44424-PK|CG44424-PL|CG44424-PM|CG44424-PN|rad-PI|rad-PJ|rad-PK|rad-PL|rad-PM|rad-PN', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12988615, end_position_on_the_genomic_accession=13079070, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  37167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37167, CurrentID=0, Status='live', Symbol='CG15080', Aliases='Dmel_ Dmel\\CG15080', description='CG15080 gene product from transcript CG15080-RA', other_designations='CG15080-PA|uncharacterized protein', map_location='55F3-55F3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=18783299, end_position_on_the_genomic_accession=18794594, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19835051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835051, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-B', Aliases='Dmel_CG44884, CG44884, Dmel\\CG44884', description='CG44884 gene product from transcript CG44884-RB', other_designations='CG44884-PA|CG44884-PB|CG44884-PC|pre-mod(mdg4)-B-PA|pre-mod(mdg4)-B-PB|pre-mod(mdg4)-B-PC', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21354445, end_position_on_the_genomic_accession=21359768, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  261634 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261634, CurrentID=0, Status='live', Symbol='CG31051', Aliases='Dmel_ CG1540, CT3186, Dmel\\CG31051', description='CG31051 gene product from transcript CG31051-RA', other_designations='CG31051-PA|CG31051-PB|uncharacterized protein', map_location='98D3-98D3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=28519548, end_position_on_the_genomic_accession=28520515, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  318151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318151, CurrentID=0, Status='live', Symbol='CG32681', Aliases='Dmel_ Dmel\\CG32681', description='CG32681 gene product from transcript CG32681-RA', other_designations='CG32681-PA|uncharacterized protein', map_location='9B12-9B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10376651, end_position_on_the_genomic_accession=10381352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740596, CurrentID=0, Status='live', Symbol='CG34166', Aliases='Dmel_ CG15281, Dmel\\CG34166', description='CG34166 gene product from transcript CG34166-RD', other_designations='CG34166-PA|CG34166-PB|CG34166-PC|CG34166-PD|uncharacterized protein', map_location='35B5-35B5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=14743193, end_position_on_the_genomic_accession=14743879, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  39181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39181, CurrentID=0, Status='live', Symbol='CG8072', Aliases='Dmel_ Dmel\\CG8072', description='CG8072 gene product from transcript CG8072-RA', other_designations='CG8072-PA', map_location='67D9-67D9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10151827, end_position_on_the_genomic_accession=10152984, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14462830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462830, CurrentID=0, Status='live', Symbol='CG43915', Aliases='Dmel_ Dmel\\CG43915', description='CG43915 gene product from transcript CG43915-RA', other_designations='CG43915-PA|uncharacterized protein', map_location='64A2-64A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3972899, end_position_on_the_genomic_accession=3973784, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835072, CurrentID=0, Status='live', Symbol='CG44303', Aliases='Dmel_ CG3184, Dmel\\CG44303', description='CG44303 gene product from transcript CG44303-RA', other_designations='CG44303-PA|uncharacterized protein', map_location='6C6-6C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6686963, end_position_on_the_genomic_accession=6687420, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  5740338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740338, CurrentID=0, Status='live', Symbol='CG34177', Aliases='Dmel_ CG12677, Dmel\\CG34177, Dmel_CG12677', description='CG34177 gene product from transcript CG34177-RA', other_designations='CG34177-PA|uncharacterized protein', map_location='24F2-24F2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=4433908, end_position_on_the_genomic_accession=4435057, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246500, CurrentID=0, Status='live', Symbol='CG30177', Aliases='Dmel_ Dmel\\CG30177', description='CG30177 gene product from transcript CG30177-RA', other_designations='CG30177-PA|uncharacterized protein', map_location='60A1-60A1', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23801542, end_position_on_the_genomic_accession=23803070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  246501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=246501, CurrentID=0, Status='live', Symbol='CG30178', Aliases='Dmel_ CG17263, Dmel\\CG30178', description='CG30178 gene product from transcript CG30178-RA', other_designations='CG30178-PA|uncharacterized protein', map_location='60A16-60A16', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23986615, end_position_on_the_genomic_accession=23987308, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  318153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=318153, CurrentID=0, Status='live', Symbol='CG32686', Aliases='Dmel_ BcDNA:AT25933, Dmel\\CG32686, anon-WO02059370.11', description='CG32686 gene product from transcript CG32686-RA', other_designations='CG32686-PA|CG32686-PB|uncharacterized protein', map_location='9B12-9B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10371717, end_position_on_the_genomic_accession=10375316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835563, CurrentID=0, Status='live', Symbol='CG45088', Aliases='Dmel_ Dmel\\CG45088, Dromel_CG6191_FBtr0087678_uORF', description='CG45088 gene product from transcript CG45088-RA', other_designations='CG45088-PA|uncharacterized protein', map_location='50B3-50B4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13580586, end_position_on_the_genomic_accession=13593339, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  3771947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3771947, CurrentID=0, Status='live', Symbol='His4:CG33883', Aliases='Dmel_CG33883, CG33883, Dmel\\CG33883', description='CG33883 gene product from transcript CG33883-RA', other_designations='CG33883-PA|His4:CG33883-PA|histone H4', map_location='39D5-39D5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21499324, end_position_on_the_genomic_accession=21499736, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14462840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=14462840, CurrentID=0, Status='live', Symbol='CG43886', Aliases='Dmel_ Dmel\\CG43886, lincRNA.469', description='CG43886 gene product from transcript CG43886-RA', other_designations='CG43886-PA|uncharacterized protein', map_location='66B10-66B10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8159361, end_position_on_the_genomic_accession=8159774, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326159, CurrentID=0, Status='live', Symbol='CG31773', Aliases='Dmel_ CG3390, Dmel\\CG31773', description='CG31773 gene product from transcript CG31773-RA', other_designations='CG31773-PA|uncharacterized protein', map_location='24C9-24D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3911796, end_position_on_the_genomic_accession=3914288, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740347 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740347, CurrentID=0, Status='live', Symbol='Blos3', Aliases='Dmel_CG34255, BLOS3, CG34255, Dm Blos3, Dmel\\CG34255, NP_001097640, blos3', description='Biogenesis of lysosome-related organelles complex 1, subunit 3', other_designations='BLOC-1 subunit 3|Blos3-PA|CG34255-PA|biogenesis of lysosome-related organelles complex 1, subunit 3', map_location='75E2-75E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18870224, end_position_on_the_genomic_accession=18870897, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740348, CurrentID=0, Status='live', Symbol='Cpr65Ay', Aliases='Dmel_CG34271, CG34271, Dmel\\CG34271', description='Cuticular protein 65Ay', other_designations='CG34271-PA|Cpr65Ay-PA|cuticular protein 65Ay', map_location='65A6-65A6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6158075, end_position_on_the_genomic_accession=6158630, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835127, CurrentID=0, Status='live', Symbol='pre-mod(mdg4)-E', Aliases='Dmel_CG44878, CG44878, Dmel\\CG44878', description='CG44878 gene product from transcript CG44878-RB', other_designations='CG44878-PA|CG44878-PB|CG44878-PC|pre-mod(mdg4)-E-PA|pre-mod(mdg4)-E-PB|pre-mod(mdg4)-E-PC', map_location='93D8-93D8', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21362968, end_position_on_the_genomic_accession=21365488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835131, CurrentID=0, Status='live', Symbol='CG45493', Aliases='Dmel_ Dmel\\CG45493', description='CG45493 gene product from transcript CG45493-RA', other_designations='CG45493-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19948079, end_position_on_the_genomic_accession=19948873, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835136, CurrentID=0, Status='live', Symbol='CG45066', Aliases='Dmel_ CG14955, Dmel\\CG45066', description='CG45066 gene product from transcript CG45066-RA', other_designations='CG45066-PA|uncharacterized protein', map_location='63B1-63B1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3110128, end_position_on_the_genomic_accession=3111227, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835138, CurrentID=0, Status='live', Symbol='CG44439', Aliases='Dmel_ Dmel\\CG44439, lincRNA.265', description='CG44439 gene product from transcript CG44439-RA', other_designations='CG44439-PA|uncharacterized protein', map_location='44C2-44C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8249096, end_position_on_the_genomic_accession=8249595, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326163, CurrentID=0, Status='live', Symbol='CG31823', Aliases='Dmel_ BG:DS00365.3, CG7647, Dmel\\CG31823', description='CG31823 gene product from transcript CG31823-RA', other_designations='CG31823-PA|uncharacterized protein', map_location='35D6-35D6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15898879, end_position_on_the_genomic_accession=15900524, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835139, CurrentID=0, Status='live', Symbol='CG45062', Aliases='Dmel_ CG3701, Dmel\\CG45062, alphalike', description='CG45062 gene product from transcript CG45062-RB', other_designations='CG45062-PA|CG45062-PB|uncharacterized protein', map_location='58E3-58E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22472995, end_position_on_the_genomic_accession=22476314, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  37813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37813, CurrentID=0, Status='live', Symbol='CG13566', Aliases='Dmel_ Dmel\\CG13566', description='CG13566 gene product from transcript CG13566-RB', other_designations='CG13566-PB|uncharacterized protein', map_location='60A14-60A14', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=23961477, end_position_on_the_genomic_accession=23962577, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  326165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326165, CurrentID=0, Status='live', Symbol='CG42784', Aliases='Dmel_ BcDNA:GH04046, BcDNA:HL01958, CG16850, CG16850/CG31846, CG16851, CG31846, CG31847, CR31846, Dm_2L:67149, Dmel\\CG42784, Dmel_CG16850, Dmel_CG31846, Dmel_CG31847, Dmel_CR31846, mdcds_2231', description='CG42784 gene product from transcript CG42784-RF', other_designations='CG42784-PC|CG42784-PD|CG42784-PF|CG42784-PJ|uncharacterized protein', map_location='34C3-34C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13415431, end_position_on_the_genomic_accession=13500427, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19835571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835571, CurrentID=0, Status='live', Symbol='CG45079', Aliases='Dmel_ Dmel\\CG45079', description='CG45079 gene product from transcript CG45079-RA', other_designations='CG45079-PA|uncharacterized protein', map_location='89F1-89F1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17124477, end_position_on_the_genomic_accession=17124786, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  3355145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=3355145, CurrentID=0, Status='live', Symbol='CG12547', Aliases='Dmel_ CG40277, Dmel\\CG12547, anon-EST:Posey268', description='CG12547 gene product from transcript CG12547-RB', other_designations='CG12547-PA|CG12547-PB|CG12547-PC|anon-EST:Posey268|uncharacterized protein', map_location='41C2-41C2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=4785855, end_position_on_the_genomic_accession=4788965, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  10178852 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178852, CurrentID=0, Status='live', Symbol='CG42819', Aliases='Dmel_ CG13394, Dmel\\CG42819, Dmel_CG13394', description='CG42819 gene product from transcript CG42819-RA', other_designations='CG42819-PA|uncharacterized protein', map_location='29C1-29C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8388728, end_position_on_the_genomic_accession=8389331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  36222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=36222, CurrentID=0, Status='live', Symbol='CG13202', Aliases='Dmel_ Dmel\\CG13202', description='CG13202 gene product from transcript CG13202-RA', other_designations='CG13202-PA|uncharacterized protein', map_location='47F5-47F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=11356114, end_position_on_the_genomic_accession=11356600, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835181, CurrentID=0, Status='live', Symbol='CG44812', Aliases='Dmel_ Dmel\\CG44812', description='CG44812 gene product from transcript CG44812-RA', other_designations='CG44812-PA|uncharacterized protein', map_location='60C7-60C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24350969, end_position_on_the_genomic_accession=24351389, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326174, CurrentID=0, Status='live', Symbol='CG31926', Aliases='Dmel_ CG10872, Dmel\\CG31926', description='CG31926 gene product from transcript CG31926-RB', other_designations='CG31926-PA|CG31926-PB|aspartic-type endopeptidase|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1495460, end_position_on_the_genomic_accession=1497071, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37863, CurrentID=0, Status='live', Symbol='CG43775', Aliases='Dmel_ CG13569, CG13569-1, Dmel\\CG43775', description='CG43775 gene product from transcript CG43775-RA', other_designations='CG43775-PA|uncharacterized protein', map_location='60B11-60B11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24123167, end_position_on_the_genomic_accession=24126630, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  326175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326175, CurrentID=0, Status='live', Symbol='CG31928', Aliases='Dmel_ CG10872, Dmel\\CG31928', description='CG31928 gene product from transcript CG31928-RB', other_designations='CG31928-PA|CG31928-PB|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1492295, end_position_on_the_genomic_accession=1493859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  37891 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37891, CurrentID=0, Status='live', Symbol='CG13578', Aliases='Dmel_ Dmel\\CG13578', description='CG13578 gene product from transcript CG13578-RA', other_designations='CG13578-PA|uncharacterized protein', map_location='60C7-60C7', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24377831, end_position_on_the_genomic_accession=24378740, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  5740368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740368, CurrentID=0, Status='live', Symbol='CG34201', Aliases='Dmel_ Dmel\\CG34201', description='CG34201 gene product from transcript CG34201-RA', other_designations='CG34201-PA|uncharacterized protein', map_location='57B2-57B2', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20750914, end_position_on_the_genomic_accession=20752135, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326181, CurrentID=0, Status='live', Symbol='CG31976', Aliases='Dmel_ CG11623, Dmel\\CG31976', description='CG31976 gene product from transcript CG31976-RB', other_designations='CG31976-PB|CG31976-PC|uncharacterized protein', map_location='21B3-21B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=138384, end_position_on_the_genomic_accession=140992, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  37899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=37899, CurrentID=0, Status='live', Symbol='Pgam5-2', Aliases='Dmel_CG15874, CG15874, Dmel\\CG15874, dPGAM5-2', description='Phosphoglycerate mutase 5-2', other_designations='CG15874-PA|Pgam5-2-PA|phosphoglycerate mutase 5-2', map_location='60C8-60C8', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=24431344, end_position_on_the_genomic_accession=24432733, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  5740370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740370, CurrentID=0, Status='live', Symbol='CG34307', Aliases='Dmel_ Dmel\\CG34307', description='CG34307 gene product from transcript CG34307-RB', other_designations='CG34307-PB|uncharacterized protein', map_location='87A1-87A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11897582, end_position_on_the_genomic_accession=11898419, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  39216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=39216, CurrentID=0, Status='live', Symbol='CG12523', Aliases='Dmel_ BcDNA:RE74758, Dmel\\CG12523', description='CG12523 gene product from transcript CG12523-RB', other_designations='CG12523-PA|CG12523-PB|uncharacterized protein', map_location='67F1-67F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10801190, end_position_on_the_genomic_accession=10802350, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19835590 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835590, CurrentID=0, Status='live', Symbol='CG45491', Aliases='Dmel_ Dmel\\CG45491', description='CG45491 gene product from transcript CG45491-RA', other_designations='CG45491-PA|uncharacterized protein', map_location='19A3-19A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=19938232, end_position_on_the_genomic_accession=19938994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835591, CurrentID=0, Status='live', Symbol='CG44290', Aliases='Dmel_ Dmel\\CG44290, lincRNA.869', description='CG44290 gene product from transcript CG44290-RA', other_designations='CG44290-PA|uncharacterized protein', map_location='98A1-98A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27438871, end_position_on_the_genomic_accession=27440944, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835258, CurrentID=0, Status='live', Symbol='CG45676', Aliases='Dmel_ Dmel\\CG45676', description='CG45676 gene product from transcript CG45676-RA', other_designations='CG45676-PA|uncharacterized protein', map_location='86E5-86E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=11418885, end_position_on_the_genomic_accession=11419295, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  8673979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=8673979, CurrentID=0, Status='live', Symbol='CG42470', Aliases='Dmel_ Dmel\\CG42470', description='CG42470 gene product from transcript CG42470-RA', other_designations='CG42470-PA|uncharacterized protein', map_location='33A1-33A1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=11752973, end_position_on_the_genomic_accession=11753555, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19835262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835262, CurrentID=0, Status='live', Symbol='CG45089', Aliases='Dmel_ Dmel\\CG45089, Dromel_CG18624_FBtr0071140_uORF', description='CG45089 gene product from transcript CG45089-RA', other_designations='CG45089-PA|CG45089-PB|CG45089-PC|uncharacterized protein', map_location='7C2-7C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7893053, end_position_on_the_genomic_accession=7893798, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  30970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30970, CurrentID=0, Status='live', Symbol='CG3038', Aliases='Dmel_ Dmel\\CG3038, EG:BACR37P7.1, GalT3, dbeta3GnT', description='CG3038 gene product from transcript CG3038-RC', other_designations='CG3038-PA|CG3038-PB|CG3038-PC', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=243954, end_position_on_the_genomic_accession=245856, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30971, CurrentID=0, Status='live', Symbol='G9a', Aliases='Dmel_CG2995, CG2995, Dmel\\CG2995, EG:BACR37P7.2, EHMT, cab65850  Dm, dG9a', description='CG2995 gene product from transcript CG2995-RB', other_designations='CG2995-PA|CG2995-PB|G9a-PA|G9a-PB|dG9a|euchromatin histone methyltransferase|histone methyltransferase', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=245978, end_position_on_the_genomic_accession=254650, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  30972 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30972, CurrentID=0, Status='live', Symbol='CG13377', Aliases='Dmel_ BG01596, Dmel\\CG13377, EG:BACR37P7.9', description='CG13377 gene product from transcript CG13377-RB', other_designations='BG01596|CG13377-PA|CG13377-PB|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=258335, end_position_on_the_genomic_accession=263999, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30973, CurrentID=0, Status='live', Symbol='cin', Aliases='Dmel_CG2945, CG2945, CIN, Dmel\\CG2945, EG:BACR37P7.3, fs(1)M50, l(1)1Ae, l(1)G0142', description='cinnamon', other_designations='CG2945-PA|CG2945-PB|cin-PA|cin-PB', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=255674, end_position_on_the_genomic_accession=258469, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  30975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30975, CurrentID=0, Status='live', Symbol='ewg', Aliases='Dmel_CG3114, CG3114, Dmel\\CG3114, EC3, EG:BACR37P7.7, EWG, Ewg, NRF-1, l(1)1Ag, l(1)DA659, l(1)EC3, l(1)EC3f, l(1)VA720, l[EC3]', description='erect wing', other_designations='CG3114-PA|CG3114-PB|CG3114-PC|CG3114-PD|CG3114-PE|CG3114-PF|CG3114-PG|CG3114-PH|CG3114-PJ|CG3114-PK|CG3114-PL|CG3114-PM|CG3114-PN|erect wings|ewg-PA|ewg-PB|ewg-PC|ewg-PD|ewg-PE|ewg-PF|ewg-PG|ewg-PH|ewg-PJ|ewg-PK|ewg-PL|ewg-PM|ewg-PN', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=268509, end_position_on_the_genomic_accession=279718, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  30976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30976, CurrentID=0, Status='live', Symbol='CG13375', Aliases='Dmel_ Dmel\\CG13375, EG:BACR37P7.8', description='CG13375 gene product from transcript CG13375-RB', other_designations='CG13375-PA|CG13375-PB|CG13375-PC|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=283186, end_position_on_the_genomic_accession=294962, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  30977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30977, CurrentID=0, Status='live', Symbol='CG12470', Aliases='Dmel_ Dmel\\CG12470, EG:BACR37P7.5, NEST:bs07e10, NEST:bs17h08', description='CG12470 gene product from transcript CG12470-RA', other_designations='CG12470-PA|uncharacterized protein', map_location='1A1-1A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=307684, end_position_on_the_genomic_accession=308935, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  30978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30978, CurrentID=0, Status='live', Symbol='Or1a', Aliases='Dmel_CG17885, 1A.1, 1a, CG17885, DOR68, Dmel\\CG17885, OR1a', description='Odorant receptor 1a', other_designations='CG17885-PA|Or1a-PA|odorant receptor 1a', map_location='1A2-1A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=315123, end_position_on_the_genomic_accession=316566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  30979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30979, CurrentID=0, Status='live', Symbol='CG3777', Aliases='Dmel_ BEST:CK01670, CK01670, Dmel\\CG3777, EG:125H10.1', description='CG3777 gene product from transcript CG3777-RB', other_designations='CG3777-PA|CG3777-PB|CG3777-PC|CG3777-PD|CG3777-PE', map_location='1A1-1A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=279703, end_position_on_the_genomic_accession=350504, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  30980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30980, CurrentID=0, Status='live', Symbol='y', Aliases='Dmel_CG3757, CG3757, Dmel\\CG3757, EG:125H10.2, T6, Y', description='yellow', other_designations='CG3757-PA|y-PA', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=356509, end_position_on_the_genomic_accession=361245, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  30981 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30981, CurrentID=0, Status='live', Symbol='ac', Aliases='Dmel_CG3796, 990 E5 F1, AC, AS-C, AS-C T5, AS-C T5ac, ASC, Ac, CG3796, Dmel\\CG3796, EG:125H10.3, Hw, T5, ascT5, sc/T5', description='achaete', other_designations='Achaete/Scute|CG3796-PA|Hairy-wing|ac-PA|achaete-scute|acheate|achete|asc|hairy wing', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=370031, end_position_on_the_genomic_accession=370947, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  30982 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30982, CurrentID=0, Status='live', Symbol='sc', Aliases='Dmel_CG3827, AS-C T4, AS-C T4sc, CG3827, DROACS2, Dmel\\CG3827, EG:198A6.1, Hw, Sc, T4, T4 AS-C, ascT4, bHLHc28, l(1)1Ba/T4, scalpha, scute/sisterlessB, sis b, sis-b, sisB, sisb, sc', description='scute', other_designations='Achaete/Scute|CG3827-PA|Hairy-wing|achaete-scute|asc|sc-PA|scute alpha|scute-T4-transcript|scute/sisterless-b|sisterless-b', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=396060, end_position_on_the_genomic_accession=397497, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  30983 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30983, CurrentID=0, Status='live', Symbol='l(1)sc', Aliases="Dmel_CG3839, AS-C T3, AS-T3, ASC-T3, CG3839, Dmel\\CG3839, EG:198A6.2, L'Sc, L'sc, L(1)sc, T3, ascT3, bHLHc28, l'sc, l(1)1Ba, l-sc, l[1]sc, lsc, sc/T3", description='lethal of scute', other_designations='CG3839-PA|Lethal(1)scute|Lethal-scute|l(1)sc-PA|lethal at scute', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=409721, end_position_on_the_genomic_accession=410817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  30984 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30984, CurrentID=0, Status='live', Symbol='pcl', Aliases='Dmel_CG13374, CG13374, Dmel\\CG13374, EG:EG0001.1, PCL, T2, scbeta', description='pepsinogen-like', other_designations='CG13374-PA|pcl-PA|scute beta', map_location='1B2-1B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=436139, end_position_on_the_genomic_accession=437448, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  30985 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30985, CurrentID=0, Status='live', Symbol='ase', Aliases='Dmel_CG3258, AS-C T8, AS-C T8ase, AS-T8, ASC, Ase, CG3258, Dmel\\CG3258, EG:165H7.2, T1, T1A, T1a, T8, as-T8, ascT8, bHLHc28, sc/T8', description='asense', other_designations='CG3258-PA|ase-PA', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=460747, end_position_on_the_genomic_accession=463176, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  30986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30986, CurrentID=0, Status='live', Symbol='Cyp4g1', Aliases='Dmel_CG3972, 4g1, ASC-T1, CG3972, CYP4G1, Cyt-P450-A1, Dmel\\CG3972, EG:165H7.1, P-450, P-450-A, P450, P450-A, P450A, T1, anon-WO0140519.102, cyp4g1, scgamma', description='Cytochrome P450-4g1', other_designations='Achaete-scute complex transcript 1|CG3972-PA|Cyp4g1-PA|cytochrome P450|cytochrome P450 4g1|cytochrome P450 A1|cytochrome P450-4g1', map_location='1B4-1B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=467057, end_position_on_the_genomic_accession=469335, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  30988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30988, CurrentID=0, Status='live', Symbol='CG13373', Aliases='Dmel_ Dmel\\CG13373', description='CG13373 gene product from transcript CG13373-RB', other_designations='CG13373-PA|CG13373-PB|uncharacterized protein', map_location='1B4-1B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=476984, end_position_on_the_genomic_accession=478191, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  30990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30990, CurrentID=0, Status='live', Symbol='CG3176', Aliases='Dmel_ BcDNA:RH59219, Dmel\\CG3176', description='CG3176 gene product from transcript CG3176-RB', other_designations='CG3176-PA|CG3176-PB|CG3176-PC|CG3176-PD', map_location='1B4-1B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=481862, end_position_on_the_genomic_accession=482768, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  30991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30991, CurrentID=0, Status='live', Symbol='CG18273', Aliases='Dmel_ CG13372, Dmel\\CG18273, EG:171D11.6', description='CG18273 gene product from transcript CG18273-RA', other_designations='CG18273-PA|uncharacterized protein', map_location='1B4-1B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=482777, end_position_on_the_genomic_accession=487581, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  30994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30994, CurrentID=0, Status='live', Symbol='CG3156', Aliases='Dmel_ Dmel\\CG3156, EG:171D11.2', description='CG3156 gene product from transcript CG3156-RA', other_designations='CG3156-PA', map_location='1B5-1B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=487878, end_position_on_the_genomic_accession=490725, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  30995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30995, CurrentID=0, Status='live', Symbol='CG17896', Aliases='Dmel_ 153323_at, Dmel\\CG17896, EG:171D11.1', description='CG17896 gene product from transcript CG17896-RB', other_designations='CG17896-PA|CG17896-PB|uncharacterized protein', map_location='1B5-1B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=490787, end_position_on_the_genomic_accession=493495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30996, CurrentID=0, Status='live', Symbol='CG17778', Aliases='Dmel_ BcDNA:RH02374, Dmel\\CG17778, EG:171D11.5', description='CG17778 gene product from transcript CG17778-RB', other_designations='CG17778-PA|CG17778-PB|uncharacterized protein', map_location='1B5-1B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=494614, end_position_on_the_genomic_accession=496369, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  30998 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=30998, CurrentID=0, Status='live', Symbol='svr', Aliases='Dmel_CG4122, CG18503, CG32818, CG4122, CP, CPD, CpepE, CypD, Dmel\\CG4122, EG:171D11.3, EP(X)0356, cc, dwp, l(1)1Bd, poi, slv', description='silver', other_designations='CG4122-PB|CG4122-PD|CG4122-PE|CG4122-PF|CG4122-PG|CG4122-PH|CG4122-PI|CG4122-PJ|CG4122-PL|CG4122-PM|CG4122-PN|DmSilver|carboxypeptidase|carboxypeptidase D|carboxypeptidase E|chlorotic|dwarp|pointed|svr-PB|svr-PD|svr-PE|svr-PF|svr-PG|svr-PH|svr-PI|svr-PJ|svr-PL|svr-PM|svr-PN', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=496446, end_position_on_the_genomic_accession=509072, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  5740386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740386, CurrentID=0, Status='live', Symbol='CG40813', Aliases='Dmel_ Dmel\\CG40813', description='CG40813 gene product from transcript CG40813-RB', other_designations='CG40813-PB|uncharacterized protein', map_location='', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=22143102, end_position_on_the_genomic_accession=22143816, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31000, CurrentID=0, Status='live', Symbol='elav', Aliases='Dmel_CG4262, 44C11, 9F8A9, CG4262, Dmel\\CG4262, EC7, EG:65F1.2, ELAV, END1-2, ElaV, Elav, Elav-9F8A9-1, elav-2, elav-3, fliJ, l(1)1Be, l(1)EC7, l(1)G0031, l(1)G0319, weg, elav', description='embryonic lethal abnormal vision', other_designations='CG4262-PA|CG4262-PB|CG4262-PC|CG4262-PD|elav-PA|elav-PB|elav-PC|elav-PD|embryonic lethal abnormal visual system|embryonic lethal, abnormal vision|weniger', map_location='1B7-1B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=509235, end_position_on_the_genomic_accession=523226, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31001, CurrentID=0, Status='live', Symbol='CG4293', Aliases='Dmel_ 154711_at, Dmel\\CG4293, EG:65F1.1', description='CG4293 gene product from transcript CG4293-RD', other_designations='CG4293-PA|CG4293-PB|CG4293-PC|CG4293-PD|CG4293-PE', map_location='1B8-1B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=527941, end_position_on_the_genomic_accession=530291, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31002, CurrentID=0, Status='live', Symbol='Appl', Aliases='Dmel_CG7727, APP, APP-like, APPL, Abeta, BcDNA:GH04413, CG7727, Dmel\\CG7727, EG:65F1.5, appl', description='beta amyloid protein precursor-like', other_designations='APP-like|Appl-PA|Appl-PB|Appl-PC|Appl-PD|Appl-PE|Appl-PF|CG7727-PA|CG7727-PB|CG7727-PC|CG7727-PD|CG7727-PE|CG7727-PF|amyloid Precursor-Like protein|amyloid precursor protein|amyloid precursor protein-like|beta-amyloid-protein-precursor-like', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=530501, end_position_on_the_genomic_accession=579044, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  31003 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31003, CurrentID=0, Status='live', Symbol='vnd', Aliases='Dmel_CG6172, CG6172, Dm vnd, Dm-NK2, Dmel\\CG6172, EC6, EG:118B3.1, HDC16822, NK-2, NK2, Nkx2, VND, Vnd, dNK-2, l(1)1Bf, l(1)EA142, l(1)EC6, l(1)GA100, l(1)GA122, l(1)RC24, l(1)VA208, l(1)VE769, lab, nk-2/NK-2, vnd', description='ventral nervous system defective', other_designations='CG6172-PA|CG6172-PB|HDC16822|ventral nerve cord defective|ventral nervous system condensation defective|ventral neuroblasts defective|ventral neurons defective|vnd-PA|vnd-PB', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=582121, end_position_on_the_genomic_accession=598151, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31004 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31004, CurrentID=0, Status='live', Symbol='CG13366', Aliases='Dmel_ AAD55419, Dmel\\CG13366, EG:118B3.2', description='CG13366 gene product from transcript CG13366-RG', other_designations='CG13366-PA|CG13366-PB|CG13366-PC|CG13366-PD|CG13366-PE|CG13366-PF|CG13366-PG|SPECC1L ortholog|SPECC1L-ortholog|split discs|uncharacterized protein', map_location='1B11-1B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=585081, end_position_on_the_genomic_accession=620195, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  31005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31005, CurrentID=0, Status='live', Symbol='CG13365', Aliases='Dmel_ BcDNA:LP07417, Dmel\\CG13365', description='CG13365 gene product from transcript CG13365-RA', other_designations='CG13365-PA|uncharacterized protein', map_location='1B12-1B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=620717, end_position_on_the_genomic_accession=621215, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31006, CurrentID=0, Status='live', Symbol='mod(r)', Aliases='Dmel_CG17828, CG17828, Dmel\\CG17828, EG:115C2.5, Vps37, anon-WO0153538.78, vps37A', description='modifier of rudimentary', other_designations='CG17828-PA|CG17828-PB|CG17828-PC|CG17828-PD|CG17828-PE|CG17828-PF|mod(r)-PA|mod(r)-PB|mod(r)-PC|mod(r)-PD|mod(r)-PE|mod(r)-PF', map_location='1B12-1B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=621530, end_position_on_the_genomic_accession=623051, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31007, CurrentID=0, Status='live', Symbol='CG13369', Aliases='Dmel_ Dmel\\CG13369, EG:115C2.1', description='CG13369 gene product from transcript CG13369-RA', other_designations='CG13369-PA|uncharacterized protein', map_location='1B12-1B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=623468, end_position_on_the_genomic_accession=624765, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31009, CurrentID=0, Status='live', Symbol='RpL36', Aliases='Dmel_CG7622, BcDNA:AT29875, CG7622, Dmel\\CG7622, EG:115C2.7, L36a, M(1)1B, M(1)Bld, Rp L36, Vi, anon-EST:Posey274, l(1)1Bj, l(1)G0471', description='Ribosomal protein L36', other_designations='CG7622-PA|CG7622-PB|CG7622-PC|CG7622-PD|CG7622-PE|Minute(1)1B|Minute(1)Blond|RpL36-PA|RpL36-PB|RpL36-PC|RpL36-PD|RpL36-PE|ribosomal protein L36|viability', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=628338, end_position_on_the_genomic_accession=629483, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31010 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31010, CurrentID=0, Status='live', Symbol='l(1)1Bi', Aliases='Dmel_CG6189, CG6189, Dmel\\CG6189, EG:115C2.2, HM-419, HM-52, l(1)Bi, l(1)lBi', description='lethal (1) 1Bi', other_designations='CG6189-PA|l(1)1Bi-PA', map_location='1-0.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=629773, end_position_on_the_genomic_accession=633633, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31011, CurrentID=0, Status='live', Symbol='Dredd', Aliases='Dmel_CG7486, CASP8_DROME, CG7486, DCP-2, DCP-2/DREDD, DCP2, DREDD, Dcp-2, Dcp-2/Dredd, Dcp2, Dedd, Dmel\\CG7486/DCP-2, EG:115C2.9, Redd, anon-1BCa, dDredd, dcp-2/dredd, dredd, Dredd', description='Death related ced-3/Nedd2-like caspase', other_designations='CG7486-PD|CG7486-PE|CG7486-PF|CG7486-PG|Ced-3-like/Nedd2-like protein|Dredd-PD|Dredd-PE|Dredd-PF|Dredd-PG|caspase|death caspase-2|death related ced-3/Nedd2-like caspase|death related ced-3/Nedd2-like protein|death-related ced-3/neural precursor cell expressed developmentally downregulated 2 (NEDD2)-like protein|drosophila caspase', map_location='1B12-1B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=633622, end_position_on_the_genomic_accession=635545, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31012 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31012, CurrentID=0, Status='live', Symbol='su(sable)', Aliases='Dmel_CG6222, CG6222, Dmel\\CG6222, EG:115C2.3, su(s), su-s, su-v, su[2]-s', description='suppressor of sable', other_designations='CG6222-PA|CG6222-PB|CG6222-PC|su(sable)-PA|su(sable)-PB|su(sable)-PC|suppressor-of-sable', map_location='1B13-1B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=635890, end_position_on_the_genomic_accession=643760, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31013, CurrentID=0, Status='live', Symbol='CG13367', Aliases='Dmel_ Dmel\\CG13367, EG:115C2.8, anon-1BCb', description='CG13367 gene product from transcript CG13367-RA', other_designations='CG13367-PA|uncharacterized protein', map_location='1B13-1B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=644761, end_position_on_the_genomic_accession=646386, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31014, CurrentID=0, Status='live', Symbol='Roc1a', Aliases='Dmel_CG16982, BEST:CK01110, CG16982, CK01110, Dmel\\CG16982, EG:115C2.11, ROC1, Rbx1, Rbx1/Roclpha, anon-EST:Posey65, dRbx1, dRoc1, dRoc1a, Roc1a', description='Regulator of cullins 1a', other_designations='CG16982-PA|CG16982-PC|CG16982-PD|Roc1a-PA|Roc1a-PC|Roc1a-PD|regulator of cullins 1a', map_location='1B13-1B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=647304, end_position_on_the_genomic_accession=648597, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31015, CurrentID=0, Status='live', Symbol='Hmt4-20', Aliases='Dmel_CG13363, BcDNA:LD36415, CG13363, Dmel\\CG13363, EG:115C2.10, SUV4-20, Su(var)4-20, Suv4-20, Suv4.20, dSuv4-20, hmt4-20', description='Histone methyltransferase 4-20', other_designations='CG13363-PA|CG13363-PB|Hmt4-20-PA|Hmt4-20-PB|Suv4-20|histone methyltransferase 4-20', map_location='1B13-1B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=648483, end_position_on_the_genomic_accession=656033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31016, CurrentID=0, Status='live', Symbol='SkpA', Aliases='Dmel_CG16983, CG16983, Dmel\\CG16983, EG:115C2.4, SKP1, SKPA, Skp, Skp1, anon-WO03040301.234, anon-WO03040301.236, anon-WO03040301.238, dSKP1, dSkp-1, dSkpA, dskpA, l(1)G0037, l(1)G0058, l(1)G0109, l(1)G0389, skpA, spkA', description='SKP1-related A', other_designations='CG16983-PA|CG16983-PB|CG16983-PC|CG16983-PD|CG16983-PE|CG16983-PF|CG16983-PG|CG16983-PH|CG16983-PI|Skp1-related gene A|SkpA-PA|SkpA-PB|SkpA-PC|SkpA-PD|SkpA-PE|SkpA-PF|SkpA-PG|SkpA-PH|SkpA-PI|skpA', map_location='1B14-1B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=656114, end_position_on_the_genomic_accession=657899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31017, CurrentID=0, Status='live', Symbol='sdk', Aliases='Dmel_CG5227, BcDNA:LD22322, BcDNA:RE63453, CG5227, CT16627, Dmel\\CG5227, EG:BACR19J1.1', description='sidekick', other_designations='CG5227-PA|CG5227-PB|CG5227-PC|CG5227-PD|CG5227-PE|CG5227-PF|CG5227-PG|CG5227-PH|CG5227-PI|sdk-PA|sdk-PB|sdk-PC|sdk-PD|sdk-PE|sdk-PF|sdk-PG|sdk-PH|sdk-PI', map_location='1C2-1C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=686834, end_position_on_the_genomic_accession=749496, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  31018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31018, CurrentID=0, Status='live', Symbol='CG13362', Aliases='Dmel_ Dmel\\CG13362', description='CG13362 gene product from transcript CG13362-RA', other_designations='CG13362-PA|uncharacterized protein', map_location='1C3-1C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=749241, end_position_on_the_genomic_accession=751301, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31019, CurrentID=0, Status='live', Symbol='CG13361', Aliases='Dmel_ Dmel\\CG13361', description='CG13361 gene product from transcript CG13361-RA', other_designations='CG13361-PA|uncharacterized protein', map_location='1C3-1C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=752011, end_position_on_the_genomic_accession=753653, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31020, CurrentID=0, Status='live', Symbol='CG5254', Aliases='Dmel_ CT16777, Dmel\\CG5254, EG:BACR19J1.2', description='CG5254 gene product from transcript CG5254-RB', other_designations='CG5254-PA|CG5254-PB', map_location='1C4-1C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=750347, end_position_on_the_genomic_accession=758719, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31021, CurrentID=0, Status='live', Symbol='CG5273', Aliases='Dmel_ Dmel\\CG5273, EG:BACR19J1.3', description='CG5273 gene product from transcript CG5273-RB', other_designations='CG5273-PA|CG5273-PB|CG5273-PC', map_location='1C4-1C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=759594, end_position_on_the_genomic_accession=761939, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31022, CurrentID=0, Status='live', Symbol='RpL22', Aliases='Dmel_CG7434, CG7434, Dmel\\CG7434, EG:BACR19J1.4, L22, PBP-12, RL22, RL22_DROME, RPL22, Rp L22, Rpl22, anon-EST:Posey53, rpL22, rpl22', description='Ribosomal protein L22', other_designations='CG7434-PA|PARP-binding protein-12|RpL22-PA|ribosomal protein L22', map_location='1C4-1C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=762560, end_position_on_the_genomic_accession=765130, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31023 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31023, CurrentID=0, Status='live', Symbol='fz3', Aliases='Dmel_CG16785, CG16785, DFz3, Dfz3, Dm Fz3, Dmel\\CG16785, EG:34F3.6, Fz3, dFz3, dfz3', description='frizzled 3', other_designations='CG16785-PA|CG16785-PB|CG16785-PC|Dfrizzled-3|dFrizzled3|frizzled3|fz3-PA|fz3-PB|fz3-PC', map_location='1C4-1C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=769945, end_position_on_the_genomic_accession=784222, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31024, CurrentID=0, Status='live', Symbol='tw', Aliases='Dmel_CG12311, CG12311, DPOMT2, DmPOMT2, Dmel\\CG12311, EG:34F3.7, POMT2, Pomt2, Tw, dPOMT2, dPomt2', description='Protein O-mannosyltransferase 2', other_designations='CG12311-PA|CG12311-PB|protein O-mannosyltransferase 2|tricky-dickey|tw-PA|tw-PB|twisted', map_location='1-0.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=784746, end_position_on_the_genomic_accession=788215, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31025 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31025, CurrentID=0, Status='live', Symbol='CG13360', Aliases='Dmel_ Dmel\\CG13360, EG:34F3.5', description='CG13360 gene product from transcript CG13360-RA', other_designations='CG13360-PA|uncharacterized protein', map_location='1C4-1C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=787855, end_position_on_the_genomic_accession=790089, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31026, CurrentID=0, Status='live', Symbol='CG16989', Aliases='Dmel_ Dmel\\CG16989, EG:34F3.4', description='CG16989 gene product from transcript CG16989-RB', other_designations='CG16989-PA|CG16989-PB|uncharacterized protein', map_location='1C5-1C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=793147, end_position_on_the_genomic_accession=797061, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31027, CurrentID=0, Status='live', Symbol='Rbf', Aliases='Dmel_CG7413, CG7413, Dmel\\CG7413, EG:34F3.3, FBF, RB, RBF, RBF1, Rb, Rb1, RbF1, dRBF, pRb, rb, rbf, rbf1, Rbf', description='Retinoblastoma-family protein', other_designations='CG7413-PA|Rb|Rbf-PA|retinoblastoma|retinoblastoma factor|retinoblastoma factor-1|retinoblastoma family protein 1|retinoblastoma-family homolog', map_location='1C5-1C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=797090, end_position_on_the_genomic_accession=801479, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31028, CurrentID=0, Status='live', Symbol='CG13359', Aliases='Dmel_ Dmel\\CG13359, EG:34F3.9', description='CG13359 gene product from transcript CG13359-RA', other_designations='CG13359-PA|uncharacterized protein', map_location='1C5-1C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=801669, end_position_on_the_genomic_accession=803470, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31029, CurrentID=0, Status='live', Symbol='CG13358', Aliases='Dmel_ Dmel\\CG13358, EG:34F3.10', description='CG13358 gene product from transcript CG13358-RD', other_designations='CG13358-PD|uncharacterized protein', map_location='1C5-1C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=803424, end_position_on_the_genomic_accession=804921, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31030 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31030, CurrentID=0, Status='live', Symbol='Sec22', Aliases='Dmel_CG7359, 34F3.8, CG7359, Dmel\\CG7359, EG:34F3.8, SEC22p, Sec22', description='Sec22 ortholog (S. cerevisiae)', other_designations='CG7359-PA|CG7359-PB|Sec22 ortholog|Sec22-PA|Sec22-PB', map_location='1C5-1C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=806076, end_position_on_the_genomic_accession=808200, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31031 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31031, CurrentID=0, Status='live', Symbol='CG43867', Aliases='Dmel_ CG12467, CG14632, CG14633, CG42248, Dmel\\CG43867, Dmel_CG12467, Dmel_CG14632, Dmel_CG14633, Dmel_CG42248, EG:34F3.1, EG:34F3.2', description='CG43867 gene product from transcript CG43867-RA', other_designations='CG43867-PA|CG43867-PB|CG43867-PC|CG43867-PD|CG43867-PF|CG43867-PH|CG43867-PI|CG43867-PJ|CG43867-PK|CG43867-PL|uncharacterized protein', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=808217, end_position_on_the_genomic_accession=927975, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  31032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31032, CurrentID=0, Status='live', Symbol='CG14635', Aliases='Dmel_ Dmel\\CG14635', description='CG14635 gene product from transcript CG14635-RA', other_designations='CG14635-PA|uncharacterized protein', map_location='1D1-1D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=844823, end_position_on_the_genomic_accession=845297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31033, CurrentID=0, Status='live', Symbol='CG11664', Aliases='Dmel_ Dmel\\CG11664, EG:BACR7A4.3, SP193', description='CG11664 gene product from transcript CG11664-RA', other_designations='CG11664-PA|uncharacterized protein', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=875732, end_position_on_the_genomic_accession=876612, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31036, CurrentID=0, Status='live', Symbol='CG11663', Aliases='Dmel_ BcDNA:AT05678, Dmel\\CG11663', description='CG11663 gene product from transcript CG11663-RA', other_designations='CG11663-PA|CG11663-PB|uncharacterized protein', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=894482, end_position_on_the_genomic_accession=895200, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31037, CurrentID=0, Status='live', Symbol='CG3711', Aliases='Dmel_ Dmel\\CG3711, EG:BACR7A4.19', description='CG3711 gene product from transcript CG3711-RE', other_designations='CG3711-PD|CG3711-PE', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=922889, end_position_on_the_genomic_accession=927582, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31038, CurrentID=0, Status='live', Symbol='MED22', Aliases='Dmel_CG3034, 144358_at, CG18830, CG3034, Dmel\\CG3034, EG:BACR7A4.6, Med22, Med24, Srb6/Surf5, anon-EST:Liang-1.82, clone 1.82, dMED24, med22', description='Mediator complex subunit 22', other_designations='CG3034-PA|MED22-PA|mediator complex subunit 22', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=928139, end_position_on_the_genomic_accession=928860, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31039, CurrentID=0, Status='live', Symbol='CG3708', Aliases='Dmel_ 152685_at, Dmel\\CG3708, EG:BACR7A4.18', description='CG3708 gene product from transcript CG3708-RB', other_designations='CG3708-PB', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=928798, end_position_on_the_genomic_accession=930098, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31042, CurrentID=0, Status='live', Symbol='TRAM', Aliases='Dmel_CG11642, CG11642, CG18830, Dmel\\CG11642, EG:BACR7A4.5', description='CG11642 gene product from transcript CG11642-RB', other_designations='CG11642-PA|CG11642-PB|CG11642-PC|CG11642-PD|CG11642-PE|TRAM-PA|TRAM-PB|TRAM-PC|TRAM-PD|TRAM-PE', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=935232, end_position_on_the_genomic_accession=937822, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31043, CurrentID=0, Status='live', Symbol='CG3704', Aliases='Dmel_ Dmel\\CG3704, EG:BACR7A4.17, XAB1', description='CG3704 gene product from transcript CG3704-RA', other_designations='CG3704-PA', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=939170, end_position_on_the_genomic_accession=940805, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31044 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31044, CurrentID=0, Status='live', Symbol='mus81', Aliases='Dmel_CG3026, CG3026, Dmel\\CG3026, EG:BACR7A4.16, Mus81', description='CG3026 gene product from transcript CG3026-RA', other_designations='CG3026-PA|mus81-PA', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=941075, end_position_on_the_genomic_accession=943229, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31045, CurrentID=0, Status='live', Symbol='CG3703', Aliases='Dmel_ BACR7A4.15, Dmel\\CG3703, EG:BACR7A4.15', description='CG3703 gene product from transcript CG3703-RA', other_designations='CG3703-PA', map_location='1D2-1D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=942474, end_position_on_the_genomic_accession=945773, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31046, CurrentID=0, Status='live', Symbol='CG3699', Aliases='Dmel_ CT12411, Dmel\\CG3699, EG:BACR7A4.14', description='CG3699 gene product from transcript CG3699-RA', other_designations='CG3699-PA', map_location='1D2-1D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=946662, end_position_on_the_genomic_accession=947500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31047, CurrentID=0, Status='live', Symbol='CG3690', Aliases='Dmel_ Dmel\\CG3690, EG:BACR7A4.13, anon-WO0140519.250', description='CG3690 gene product from transcript CG3690-RB', other_designations='CG3690-PA|CG3690-PB', map_location='1D3-1D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=948585, end_position_on_the_genomic_accession=951900, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31048 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31048, CurrentID=0, Status='live', Symbol='TfIIA-S-2', Aliases='Dmel_CG11639, CG11639, Dmel\\CG11639, EG:BACR7A4.7, TFIIA, tfiia-s-2', description='CG11639 gene product from transcript CG11639-RA', other_designations='CG11639-PA|TfIIA-S-2-PA', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1011101, end_position_on_the_genomic_accession=1011643, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31049, CurrentID=0, Status='live', Symbol='CG14631', Aliases='Dmel_ Dmel\\CG14631', description='CG14631 gene product from transcript CG14631-RA', other_designations='CG14631-PA|uncharacterized protein', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1012035, end_position_on_the_genomic_accession=1012624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31050, CurrentID=0, Status='live', Symbol='CG11638', Aliases='Dmel_ Dmel\\CG11638, EG:BACR7A4.12', description='CG11638 gene product from transcript CG11638-RA', other_designations='CG11638-PA|uncharacterized protein', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1013984, end_position_on_the_genomic_accession=1026703, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31052, CurrentID=0, Status='live', Symbol='CDC45L', Aliases='Dmel_CG3658, CDC45, CG3658, Cdc45, D, DmCdc45, Dmel\\CG3658, EG:BACR7A4.11, anon-1Ec, cdc45, dCDC45, d', description='CG3658 gene product from transcript CG3658-RA', other_designations='CDC45L-PA|CG3658-PA|transcription unit D', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1028291, end_position_on_the_genomic_accession=1030556, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31054 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31054, CurrentID=0, Status='live', Symbol='su(w[a])', Aliases='Dmel_CG3019, CG3019, DmSWAP, Dmel\\CG3019, EG:BACR7A4.10, SWAP, su(w)-1, su(w[a])-2, su(wa), su-w, su[wa], su(w[a])', description='suppressor of white-apricot', other_designations='CG3019-PA|CG3019-PB|CG3019-PC|CG3019-PG|CG3019-PI|CG3019-PJ|CG3019-PM|su(w[a])-PA|su(w[a])-PB|su(w[a])-PC|su(w[a])-PG|su(w[a])-PI|su(w[a])-PJ|su(w[a])-PM|suppressor of apricot', map_location='1-0.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1012779, end_position_on_the_genomic_accession=1046040, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31055, CurrentID=0, Status='live', Symbol='CG14629', Aliases='Dmel_ Dmel\\CG14629, EG:103E12.2', description='CG14629 gene product from transcript CG14629-RA', other_designations='CG14629-PA|uncharacterized protein', map_location='1E1-1E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1051525, end_position_on_the_genomic_accession=1052631, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31056, CurrentID=0, Status='live', Symbol='CG3655', Aliases='Dmel_ Dmel\\CG3655, EG:103E12.3, RE126B2', description='CG3655 gene product from transcript CG3655-RB', other_designations='CG3655-PA|CG3655-PB|CG3655-PC|CG3655-PD', map_location='1E1-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1073041, end_position_on_the_genomic_accession=1134896, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31057, CurrentID=0, Status='live', Symbol='CG14628', Aliases='Dmel_ Dmel\\CG14628, EG:BACR42I17.12, cg14628', description='CG14628 gene product from transcript CG14628-RA', other_designations='CG14628-PA|uncharacterized protein', map_location='1E3-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1145644, end_position_on_the_genomic_accession=1146132, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31059, CurrentID=0, Status='live', Symbol='eIF4E-7', Aliases='Dmel_CG32859, CG11392, CG11393, CG32859, CT31794, Dmel\\CG32859, EG:BACR42I17.1, eIF4E', description='CG32859 gene product from transcript CG32859-RA', other_designations='CG32859-PA|eIF4E-7-PA', map_location='1E3-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1158599, end_position_on_the_genomic_accession=1162389, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31060, CurrentID=0, Status='live', Symbol='CG11378', Aliases='Dmel_ Dmel\\CG11378, EG:BACR42I17.2', description='CG11378 gene product from transcript CG11378-RB', other_designations='CG11378-PA|CG11378-PB|CG11378-PC|uncharacterized protein', map_location='1E3-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1170828, end_position_on_the_genomic_accession=1173797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31061 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31061, CurrentID=0, Status='live', Symbol='CG11384', Aliases='Dmel_ Dmel\\CG11384, EG:BACR42I17.3', description='CG11384 gene product from transcript CG11384-RA', other_designations='CG11384-PA|uncharacterized protein', map_location='1E3-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1173877, end_position_on_the_genomic_accession=1176035, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31062, CurrentID=0, Status='live', Symbol='CG11379', Aliases='Dmel_ Dmel\\CG11379, EG:BACR42I17.4, bs31h12.y1', description='CG11379 gene product from transcript CG11379-RB', other_designations='CG11379-PA|CG11379-PB|uncharacterized protein', map_location='1E3-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1180165, end_position_on_the_genomic_accession=1181551, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31063, CurrentID=0, Status='live', Symbol='CG14627', Aliases='Dmel_ Dmel\\CG14627, EG:BACR42I17.5', description='CG14627 gene product from transcript CG14627-RA', other_designations='CG14627-PA|CG14627-PB|uncharacterized protein', map_location='1E3-1E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1181620, end_position_on_the_genomic_accession=1183502, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31064 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31064, CurrentID=0, Status='live', Symbol='CG14626', Aliases='Dmel_ Dmel\\CG14626, EG:BACR42I17.6', description='CG14626 gene product from transcript CG14626-RA', other_designations='CG14626-PA|uncharacterized protein', map_location='1E3-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1184778, end_position_on_the_genomic_accession=1188585, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31065, CurrentID=0, Status='live', Symbol='CG11380', Aliases='Dmel_ Dmel\\CG11380, EG:BACR42I17.7', description='CG11380 gene product from transcript CG11380-RB', other_designations='CG11380-PA|CG11380-PB|uncharacterized protein', map_location='1E4-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1191308, end_position_on_the_genomic_accession=1196660, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31066 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31066, CurrentID=0, Status='live', Symbol='CG14625', Aliases='Dmel_ Dmel\\CG14625, EG:BACR42I17.8', description='CG14625 gene product from transcript CG14625-RB', other_designations='CG14625-PB|uncharacterized protein', map_location='1E4-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1199546, end_position_on_the_genomic_accession=1203157, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31067, CurrentID=0, Status='live', Symbol='CG11381', Aliases='Dmel_ Dmel\\CG11381', description='CG11381 gene product from transcript CG11381-RA', other_designations='CG11381-PA|uncharacterized protein', map_location='1E4-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1203777, end_position_on_the_genomic_accession=1205433, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31068, CurrentID=0, Status='live', Symbol='CG14624', Aliases='Dmel_ BACR42I17.9, Dmel\\CG14624, EG:BACR42I17.9', description='CG14624 gene product from transcript CG14624-RA', other_designations='CG14624-PA|uncharacterized protein', map_location='1E4-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1206737, end_position_on_the_genomic_accession=1207555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31069, CurrentID=0, Status='live', Symbol='CG11382', Aliases='Dmel_ Dmel\\CG11382, EG:BACR42I17.10', description='CG11382 gene product from transcript CG11382-RB', other_designations='CG11382-PB|CG11382-PC|uncharacterized protein', map_location='1E4-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1209970, end_position_on_the_genomic_accession=1211853, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31070, CurrentID=0, Status='live', Symbol='CG11398', Aliases='Dmel_ Dmel\\CG11398, EG:BACR42I17.11', description='CG11398 gene product from transcript CG11398-RD', other_designations='CG11398-PA|CG11398-PD|uncharacterized protein', map_location='1E4-1E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1211989, end_position_on_the_genomic_accession=1215156, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31071, CurrentID=0, Status='live', Symbol='CG3638', Aliases='Dmel_ BG00376, Dmel\\CG3638, EG:33C11.3, TTYH2', description='CG3638 gene product from transcript CG3638-RJ', other_designations='BG00376|CG3638-PB|CG3638-PF|CG3638-PG|CG3638-PH|CG3638-PI|CG3638-PJ|CG3638-PK|CG3638-PL|tweety-2', map_location='1E4-1E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1215289, end_position_on_the_genomic_accession=1236684, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  31072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31072, CurrentID=0, Status='live', Symbol='CG11403', Aliases='Dmel_ Dmel\\CG11403, EG:33C11.2, cg11403', description='CG11403 gene product from transcript CG11403-RA', other_designations='CG11403-PA|uncharacterized protein', map_location='1E5-1E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1236873, end_position_on_the_genomic_accession=1239884, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31074, CurrentID=0, Status='live', Symbol='CG32812', Aliases='Dmel_ CG11408, Dmel\\CG32812, EG:114D9.1', description='CG32812 gene product from transcript CG32812-RA', other_designations='CG32812-PA|CG32812-PB|uncharacterized protein', map_location='1F1-1F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1287290, end_position_on_the_genomic_accession=1288945, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31075, CurrentID=0, Status='live', Symbol='DAAM', Aliases='Dmel_CG14622, CG14622, Daam, Dmel\\CG14622, EG:114D9.2, anon-EST:Posey148, dDAAM, dDaam, daam', description='Dishevelled Associated Activator of Morphogenesis', other_designations='CG14622-PA|CG14622-PB|CG14622-PD|CG14622-PE|CG14622-PF|CG14622-PG|DAAM-PA|DAAM-PB|DAAM-PD|DAAM-PE|DAAM-PF|DAAM-PG|DIA-like|dishevelled associated activator of morphogenesis|formin', map_location='1F2-1F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1289713, end_position_on_the_genomic_accession=1322362, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  31078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31078, CurrentID=0, Status='live', Symbol='CG11409', Aliases='Dmel_ Dmel\\CG11409, EG:8D8.2', description='CG11409 gene product from transcript CG11409-RC', other_designations='CG11409-PB|CG11409-PC|uncharacterized protein', map_location='1F3-1F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1329518, end_position_on_the_genomic_accession=1341587, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31079, CurrentID=0, Status='live', Symbol='CG11412', Aliases='Dmel_ DmAAF45606, Dmel\\CG11412, EG:8D8.6, NAA30', description='CG11412 gene product from transcript CG11412-RA', other_designations='CG11412-PA|CG11412-PD|uncharacterized protein', map_location='1F3-1F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1342106, end_position_on_the_genomic_accession=1344770, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31080, CurrentID=0, Status='live', Symbol='Tsp2A', Aliases='Dmel_CG11415, CG11415, D8.7, Dm.Tsp2A, Dmel\\CG11415, EG:8D8.7', description='Tetraspanin 2A', other_designations='CG11415-PA|CG11415-PB|Tsp2A-PA|Tsp2A-PB|tetraspanin 2A', map_location='1F4-1F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1346553, end_position_on_the_genomic_accession=1348143, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31081, CurrentID=0, Status='live', Symbol='CG11418', Aliases='Dmel_ DmNP_569904, Dmel\\CG11418, EG:8D8.8', description='CG11418 gene product from transcript CG11418-RC', other_designations='CG11418-PA|CG11418-PB|CG11418-PC|uncharacterized protein', map_location='1F4-1F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1344872, end_position_on_the_genomic_accession=1349357, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31082 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31082, CurrentID=0, Status='live', Symbol='CG12773', Aliases='Dmel_ Dmel\\CG12773, EG:8D8.3', description='CG12773 gene product from transcript CG12773-RC', other_designations='CG12773-PA|CG12773-PB|CG12773-PC|uncharacterized protein', map_location='1F4-1F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1349623, end_position_on_the_genomic_accession=1353981, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31083, CurrentID=0, Status='live', Symbol='CG11417', Aliases='Dmel_ Dmel\\CG11417, EG:8D8.4', description='CG11417 gene product from transcript CG11417-RA', other_designations='CG11417-PA|uncharacterized protein', map_location='1F4-1F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1354014, end_position_on_the_genomic_accession=1356473, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31084, CurrentID=0, Status='live', Symbol='png', Aliases='Dmel_CG11420, CG11420, Dmel\\CG11420, EG:8D8.5, PNG, Png, fs(1)M2, pan-gnu', description='pan gu', other_designations='CG11420-PA|png-PA', map_location='1F4-1F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1356488, end_position_on_the_genomic_accession=1357601, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31085, CurrentID=0, Status='live', Symbol='CG14770', Aliases='Dmel_ Dmel\\CG14770', description='CG14770 gene product from transcript CG14770-RA', other_designations='CG14770-PA|uncharacterized protein', map_location='2A1-2A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1369249, end_position_on_the_genomic_accession=1370176, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31086, CurrentID=0, Status='live', Symbol='ssx', Aliases='Dmel_CG3056, CG3056, Dmel\\CG3056, EG:132E8.1, anon-EST:Posey102, cg3056', description='sister-of-Sex-lethal', other_designations='CG3056-PB|CG3056-PC|CG3056-PD|CG3056-PE|ssx-PB|ssx-PC|ssx-PD|ssx-PE', map_location='2A1-2A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1370580, end_position_on_the_genomic_accession=1375257, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31089, CurrentID=0, Status='live', Symbol='CG32813', Aliases='Dmel_ CG11446, CG14771, Dmel\\CG32813, anon-WO0153538.75', description='CG32813 gene product from transcript CG32813-RD', other_designations='CG32813-PB|CG32813-PD|CG32813-PH|CG32813-PI|CG32813-PJ|CG32813-PK|CG32813-PL|CG32813-PM|CG32813-PN|uncharacterized protein', map_location='2A1-2A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1380914, end_position_on_the_genomic_accession=1392843, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31090, CurrentID=0, Status='live', Symbol='CG11448', Aliases='Dmel_ Dmel\\CG11448, EG:132E8.4', description='CG11448 gene product from transcript CG11448-RB', other_designations='CG11448-PA|CG11448-PB|CG11448-PC|uncharacterized protein', map_location='2A1-2A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1393054, end_position_on_the_genomic_accession=1397337, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31093, CurrentID=0, Status='live', Symbol='Gr2a', Aliases='Dmel_CG18531, CG18531, Dmel\\CG18531, EG:BACN32G11.1, GR1F.1, GR2B1, GR2a, Gr1F1, Gr2B1', description='Gustatory receptor 2a', other_designations='CG18531-PB|CG18531-PC|Gr2a-PB|Gr2a-PC|gustatory receptor 1F.1|gustatory receptor 2B1|gustatory receptor 2a', map_location='2A3-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1452430, end_position_on_the_genomic_accession=1454177, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31094, CurrentID=0, Status='live', Symbol='CG14785', Aliases='Dmel_ Dmel\\CG14785, EG:BACN32G11.2', description='CG14785 gene product from transcript CG14785-RA', other_designations='CG14785-PA|uncharacterized protein', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1455215, end_position_on_the_genomic_accession=1457054, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31095, CurrentID=0, Status='live', Symbol='Lrpprc2', Aliases='Dmel_CG14786, CG14786, DmLrpprc2, Dmel\\CG14786, EG:BACN32G11.3', description='Leucine-rich pentatricopeptide repeat containing 2', other_designations='CG14786-PA|Lrpprc2-PA', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1457175, end_position_on_the_genomic_accession=1460617, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31096 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31096, CurrentID=0, Status='live', Symbol='CG14787', Aliases='Dmel_ Dmel\\CG14787, EG:BACN32G11.4', description='CG14787 gene product from transcript CG14787-RA', other_designations='CG14787-PA|uncharacterized protein', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1460807, end_position_on_the_genomic_accession=1461676, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31097, CurrentID=0, Status='live', Symbol='Ns3', Aliases='Dmel_CG14788, A60, BACN32G11.5, CG14788, Dmel\\CG14788, EG:BACN32G11.5, dNS3, dmNuc3, l(1)2Ad, l(1)BA11, l(1)G0431, ns3', description='Nucleostemin ortholog (H. sapiens) 3', other_designations='CG14788-PA|CG14788-PB|Ns3-PA|Ns3-PB|lethal (1) 2Ad|lethal (1) G0431|nucleostemin 3|nucleostemin ortholog 3', map_location='2B1-2B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1461844, end_position_on_the_genomic_accession=1464220, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31098, CurrentID=0, Status='live', Symbol='O-fut2', Aliases='Dmel_CG14789, CG14789, Dmel\\CG14789, EG:BACN32G11.6, O-FUT2, O-fut, OFut2, Ofut2', description='O-fucosyltransferase 2', other_designations='CG14789-PA|O-fut2-PA', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1464363, end_position_on_the_genomic_accession=1465996, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31099, CurrentID=0, Status='live', Symbol='CG14777', Aliases='Dmel_ BcDNA:RH63812, Dmel\\CG14777, EG:80H7.10', description='CG14777 gene product from transcript CG14777-RF', other_designations='CG14777-PB|CG14777-PC|CG14777-PD|CG14777-PE|CG14777-PF|uncharacterized protein', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1466134, end_position_on_the_genomic_accession=1468692, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31100, CurrentID=0, Status='live', Symbol='CG14778', Aliases='Dmel_ Dmel\\CG14778', description='CG14778 gene product from transcript CG14778-RA', other_designations='CG14778-PA|uncharacterized protein', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1467493, end_position_on_the_genomic_accession=1470614, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31101, CurrentID=0, Status='live', Symbol='pck', Aliases='Dmel_CG14779, CG14779, Dmel\\CG14779, EG:80H7.2, Mega, PCK, Pck, l(1)G0044, mega', description='pickel', other_designations='CG14779-PA|CG14779-PC|CG14779-PD|megatrachea|pck-PA|pck-PC|pck-PD', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1470666, end_position_on_the_genomic_accession=1472325, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31102, CurrentID=0, Status='live', Symbol='CG14780', Aliases='Dmel_ Dmel\\CG14780, EG:80H7.3, SP95', description='CG14780 gene product from transcript CG14780-RA', other_designations='CG14780-PA|uncharacterized protein', map_location='2A4-2A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1472522, end_position_on_the_genomic_accession=1473727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31103, CurrentID=0, Status='live', Symbol='Rab27', Aliases='Dmel_CG14791, 18543235, AAF45634, AAF45635, CG14791, Dm Dmel\\CG14791, EG:80H7.4, O76901, dm-Rab27, rab27', description='CG14791 gene product from transcript CG14791-RB', other_designations='CG14791-PB|CG14791-PC|Rab27-PB|Rab27-PC', map_location='2A4-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1473884, end_position_on_the_genomic_accession=1476974, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31104, CurrentID=0, Status='live', Symbol='mei-38', Aliases='Dmel_CG14781, CG14781, D-TPX2, Dmel\\CG14781, EG:80H7.11, Mei-38, Ssp1, ssp1', description='meiotic 38', other_designations='CG14781-PA|CG14781-PB|mei-38-PA|mei-38-PB|short spindle 1', map_location='2B1-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1477063, end_position_on_the_genomic_accession=1478482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31105 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31105, CurrentID=0, Status='live', Symbol='rush', Aliases='Dmel_CG14782, CG14782, Dmel\\CG14782, EG:80H7.5', description='rush hour', other_designations='CG14782-PA|CG14782-PB|rush-PA|rush-PB', map_location='2B1-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1478552, end_position_on_the_genomic_accession=1481420, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31106, CurrentID=0, Status='live', Symbol='sta', Aliases='Dmel_CG14792, 37LRP/p40, CG14792, Dmel\\CG14792, EG:80H7.6, Lam-r, P40, Rp 40, Rp40, RpSA, STA, Sa, Sta, anon-EST:Liang-2.56, chrX:1339078..1339216, clone 2.56, k14, l(1)2Af, l(1)G0130, l(1)G0448', description='stubarista', other_designations='CG14792-PA|CG14792-PB|CG14792-PE|CG14792-PF|Laminin-receptor|laminin receptor|ribosomal protein SA|sta-PA|sta-PB|sta-PE|sta-PF', map_location='2B1-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1481797, end_position_on_the_genomic_accession=1483413, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31107, CurrentID=0, Status='live', Symbol='Nmdar2', Aliases='Dmel_CG33513, 79G8T, CG14793, CG14794, CG33513, CT34603, DmelNmdar2, Dmel\\CG33513, DrNR2, EG:80H7.7, ESTS:79G8T, NMDAR2, NR2, anon-sts34, dNR2, dnmdar-II, dsNR2', description='NMDA receptor 2', other_designations='CG33513-PA|CG33513-PB|CG33513-PC|CG33513-PD|CG33513-PE|CG33513-PF|CG33513-PG|NMDA receptor|Nmdar2-PA|Nmdar2-PB|Nmdar2-PC|Nmdar2-PD|Nmdar2-PE|Nmdar2-PF|Nmdar2-PG|dNR2', map_location='2B1-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1483625, end_position_on_the_genomic_accession=1507809, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  31109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31109, CurrentID=0, Status='live', Symbol='CG14795', Aliases='Dmel_ Dmel\\CG14795, EG:196F3.3', description='CG14795 gene product from transcript CG14795-RB', other_designations='CG14795-PB|uncharacterized protein', map_location='2B1-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1512227, end_position_on_the_genomic_accession=1512803, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31110, CurrentID=0, Status='live', Symbol='inc', Aliases='Dmel_CG32810, CG14783, CG14784, CG32810, Dmel\\CG32810, EG:196F3.2, ins', description='insomniac', other_designations='CG32810-PB|CG32810-PC|CG32810-PD|inc-PB|inc-PC|inc-PD', map_location='2B1-2B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1483479, end_position_on_the_genomic_accession=1517615, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31111, CurrentID=0, Status='live', Symbol='Mur2B', Aliases='Dmel_CG14796, CG14796, Dmel\\CG14796, EG:56G7.1', description='Mucin related 2B', other_designations='CG14796-PA|CG14796-PB|CG14796-PC|Mur2B-PA|Mur2B-PB|Mur2B-PC|mucin related 2B', map_location='2B2-2B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1523554, end_position_on_the_genomic_accession=1668788, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  38667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38667, CurrentID=0, Status='live', Symbol='CG32407', Aliases='Dmel_ BcDNA:AT29938, CG10727, Dmel\\CG32407', description='CG32407 gene product from transcript CG32407-RA', other_designations='CG32407-PA|CG32407-PB|uncharacterized protein', map_location='64F5-64F5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5922055, end_position_on_the_genomic_accession=5923567, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31118, CurrentID=0, Status='live', Symbol='dor', Aliases='Dmel_CG3093, CG3093, Deep-orange, Dmel\\CG3093, Dof, Dor, EG:171E4.1, Vps181, l(1)2Be, l(1)2Bw, l(1)7, l(1)76, dor', description='deep orange', other_designations='CG3093-PA|Deep-orange|dor-PA', map_location='1-0.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1665052, end_position_on_the_genomic_accession=1668451, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31119, CurrentID=0, Status='live', Symbol='CG3740', Aliases='Dmel_ Dmel\\CG3740, EG:171E4.4', description='CG3740 gene product from transcript CG3740-RA', other_designations='CG3740-PA', map_location='2B5-2B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1668406, end_position_on_the_genomic_accession=1669297, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31120, CurrentID=0, Status='live', Symbol='hfw', Aliases='Dmel_CG3095, CG3095, CT10406, Dmel\\CG3095, EG:171E4.2, SWI, l(1)2Bf, l(1)nfw, swi', description='halfway', other_designations='CG3095-PA|CG3095-PB|hfw-PA|hfw-PB|singed wings|singed-wings', map_location='1-0.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1669570, end_position_on_the_genomic_accession=1673608, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31121, CurrentID=0, Status='live', Symbol='CG32809', Aliases='Dmel_ CG3737, CG3783, CG3791, Dmel\\CG32809, EG:171E4.3, EG:73D1.1, EG:9D2.2', description='CG32809 gene product from transcript CG32809-RJ', other_designations='CG32809-PB|CG32809-PD|CG32809-PE|CG32809-PF|CG32809-PG|CG32809-PH|CG32809-PI|CG32809-PJ|CG32809-PK|CG32809-PL|CG32809-PM|uncharacterized protein', map_location='2B6-2B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1673110, end_position_on_the_genomic_accession=1692768, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  31123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31123, CurrentID=0, Status='live', Symbol='CG14797', Aliases='Dmel_ Dmel\\CG14797', description='CG14797 gene product from transcript CG14797-RC', other_designations='CG14797-PB|CG14797-PC|uncharacterized protein', map_location='2B7-2B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1681566, end_position_on_the_genomic_accession=1684891, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31124, CurrentID=0, Status='live', Symbol='b6', Aliases='Dmel_CG3100, B6, CG3100, Dmel\\CG3100, EG:9D2.1', description='CG3100 gene product from transcript CG3100-RA', other_designations='CG3100-PA|b6-PA', map_location='2B7-2B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1686624, end_position_on_the_genomic_accession=1690187, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31126, CurrentID=0, Status='live', Symbol='CG14798', Aliases='Dmel_ BcDNA:RE57135, Dmel\\CG14798', description='CG14798 gene product from transcript CG14798-RA', other_designations='CG14798-PA|uncharacterized protein', map_location='2B7-2B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1692824, end_position_on_the_genomic_accession=1693843, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31127, CurrentID=0, Status='live', Symbol='a6', Aliases='Dmel_CG3771, CG3771, Dmel\\CG3771, EG:9D2.3', description='CG3771 gene product from transcript CG3771-RC', other_designations='CG3771-PA|CG3771-PB|CG3771-PC|a6-PA|a6-PB|a6-PC', map_location='2B7-2B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1693615, end_position_on_the_genomic_accession=1695889, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31128, CurrentID=0, Status='live', Symbol='CG3795', Aliases='Dmel_ BcDNA:GH24985, Dmel\\CG3795, EG:9D2.4, SP160', description='CG3795 gene product from transcript CG3795-RB', other_designations='CG3795-PA|CG3795-PB', map_location='2B7-2B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1708806, end_position_on_the_genomic_accession=1710182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31129 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31129, CurrentID=0, Status='live', Symbol='Scgdelta', Aliases='Dmel_CG14808, BEST:GH21860, CG14808, CT34621, DmSCG-gammadelta, Dmel\\CG14808, EG:4F1.1, GH21860, Sgcd', description='Sarcoglycan delta', other_designations='CG14808-PA|CG14808-PC|CG14808-PD|CG14808-PE|Scgdelta-PA|Scgdelta-PC|Scgdelta-PD|Scgdelta-PE|delta-sarcoglycan|sarcoglycan|sarcoglycan delta', map_location='2B8-2B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1735945, end_position_on_the_genomic_accession=1754065, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31130, CurrentID=0, Status='live', Symbol='Adar', Aliases='Dmel_CG12598, ADAR, ADAR1, CG12598, Dmel\\CG12598, EG:BACN35H14.1, adar, adr, cg12598, dADAR, dAdar, hypnos-2', description='Adenosine deaminase acting on RNA', other_designations='Adar-PA|Adar-PB|Adar-PC|Adar-PD|Adar-PE|Adar-PF|Adar-PH|Adar-PI|Adar-PJ|Adar-PK|Adar-PM|Adar-PN|Adar-PO|Adar-PP|CG12598-PA|CG12598-PB|CG12598-PC|CG12598-PD|CG12598-PE|CG12598-PF|CG12598-PH|CG12598-PI|CG12598-PJ|CG12598-PK|CG12598-PM|CG12598-PN|CG12598-PO|CG12598-PP|adenosine deaminase acting on RNA|adenosine deaminases acting on RNA|hypnos-2|hypoxia, anoxia, sensitive 2', map_location='2B9-2B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1773725, end_position_on_the_genomic_accession=1788065, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  31131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31131, CurrentID=0, Status='live', Symbol='CG14810', Aliases='Dmel_ Dmel\\CG14810', description='CG14810 gene product from transcript CG14810-RB', other_designations='CG14810-PA|CG14810-PB|uncharacterized protein', map_location='2B10-2B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1819495, end_position_on_the_genomic_accession=1820341, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31132, CurrentID=0, Status='live', Symbol='CG14811', Aliases='Dmel_ Dmel\\CG14811', description='CG14811 gene product from transcript CG14811-RB', other_designations='CG14811-PB|uncharacterized protein', map_location='2B10-2B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1821469, end_position_on_the_genomic_accession=1822381, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31136, CurrentID=0, Status='live', Symbol='CG42666', Aliases='Dmel_ CG14799, CG14800, CG14801, CG17968, Dmel\\CG42666, Dmel_CG14799, Dmel_CG14801, EG:131F2.2, EG:137E7.1', description='CG42666 gene product from transcript CG42666-RG', other_designations='CG42666-PA|CG42666-PB|CG42666-PC|CG42666-PD|CG42666-PE|CG42666-PF|CG42666-PG|CG42666-PH|CG42666-PI|CG42666-PJ|CG42666-PK|uncharacterized protein', map_location='2B10-2B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1773719, end_position_on_the_genomic_accession=1853667, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31137, CurrentID=0, Status='live', Symbol='CG14812', Aliases='Dmel_ BcDNA:RE59545, Dmel\\CG14812, EG:131F2.3, dHBXIP', description='CG14812 gene product from transcript CG14812-RB', other_designations='CG14812-PA|CG14812-PB|uncharacterized protein', map_location='2B12-2B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1857889, end_position_on_the_genomic_accession=1858971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31139, CurrentID=0, Status='live', Symbol='CG14814', Aliases='Dmel_ Dmel\\CG14814, EG:63B12.6', description='CG14814 gene product from transcript CG14814-RD', other_designations='CG14814-PA|CG14814-PB|CG14814-PD|uncharacterized protein', map_location='2B13-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1862009, end_position_on_the_genomic_accession=1865835, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31140, CurrentID=0, Status='live', Symbol='MED18', Aliases='Dmel_CG14802, CG14802, Dmel\\CG14802, EG:63B12.13, Med18, Srb5/Med28b, dp28b, p28, p28/CG14802', description='Mediator complex subunit 18', other_designations='CG14802-PA|CG14802-PC|MED18-PA|MED18-PC|mediator complex subunit 18', map_location='2B13-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1865909, end_position_on_the_genomic_accession=1866887, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31141, CurrentID=0, Status='live', Symbol='Pex5', Aliases='Dmel_CG14815, CG14815, DmelPex5, Dmel\\CG14815, EG:63B12.5, PEX5', description='Peroxin 5', other_designations='CG14815-PA|CG14815-PB|Pex5-PA|Pex5-PB|peroxin 5', map_location='2B13-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1866832, end_position_on_the_genomic_accession=1869061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31142, CurrentID=0, Status='live', Symbol='CG14803', Aliases='Dmel_ Dmel\\CG14803, EG:63B12.9', description='CG14803 gene product from transcript CG14803-RA', other_designations='CG14803-PA|CG14803-PB|uncharacterized protein', map_location='2B13-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1868870, end_position_on_the_genomic_accession=1872408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31143, CurrentID=0, Status='live', Symbol='Pgam5', Aliases='Dmel_CG14816, CG14816, DPGAM5, Dmel\\CG14816, EG:63B12.4, PGAM5, dPGAM5', description='Phosphoglycerate mutase 5', other_designations='CG14816-PA|CG14816-PB|Pgam5-PA|Pgam5-PB|phosphoglycerate mutase 5', map_location='2B13-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1872400, end_position_on_the_genomic_accession=1873565, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31144, CurrentID=0, Status='live', Symbol='Vps26', Aliases='Dmel_CG14804, 131F2S, CG14804, DmVps26, Dmel\\CG14804, Dvps26, EG:63B12.8, ESTS:131F2S, anon-sts13, dVps26, dvps26, vps26', description='Vacuolar protein sorting 26', other_designations='CG14804-PA|Vps26-PA|vacuolar protein sorting 26', map_location='2B13-2B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1873853, end_position_on_the_genomic_accession=1876474, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31145, CurrentID=0, Status='live', Symbol='CG14817', Aliases='Dmel_ Dmel\\CG14817, EG:63B12.11', description='CG14817 gene product from transcript CG14817-RB', other_designations='CG14817-PA|CG14817-PB|uncharacterized protein', map_location='2B14-2B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1876294, end_position_on_the_genomic_accession=1877270, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31146, CurrentID=0, Status='live', Symbol='CG14805', Aliases='Dmel_ Dmel\\CG14805, EG:63B12.7, i207', description='CG14805 gene product from transcript CG14805-RA', other_designations='CG14805-PA|uncharacterized protein', map_location='2B14-2B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1877317, end_position_on_the_genomic_accession=1879077, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31147, CurrentID=0, Status='live', Symbol='CG14818', Aliases='Dmel_ BcDNA:RE09176, BcDNA:RH52516, CT34631, Dmel\\CG14818, EG:63B12.12', description='CG14818 gene product from transcript CG14818-RB', other_designations='CG14818-PA|CG14818-PB|uncharacterized protein', map_location='2B14-2B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1879019, end_position_on_the_genomic_accession=1879774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31148, CurrentID=0, Status='live', Symbol='CG14806', Aliases='Dmel_ Apop1, Dmel\\CG14806', description='CG14806 gene product from transcript CG14806-RA', other_designations='CG14806-PA|CG14806-PB|uncharacterized protein', map_location='2B14-2B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1880282, end_position_on_the_genomic_accession=1881432, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31149, CurrentID=0, Status='live', Symbol='trr', Aliases='Dmel_CG3848, CG3848, Dmel\\CG3848, EG:63B12.3, TRR, Trr, anon-2Bb', description='trithorax-related', other_designations='CG3848-PC|CG3848-PD|Trithorax-related|trithorax related|trithorax-related protein|trr-PC|trr-PD', map_location='2B14-2B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1881313, end_position_on_the_genomic_accession=1889950, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31150, CurrentID=0, Status='live', Symbol='mRpL16', Aliases='Dmel_CG3109, CG3109, D-15, D15, Dmel\\CG3109, EG:63B12.2, L16, anon-2Bd', description='mitochondrial ribosomal protein L16', other_designations='CG3109-PA|CG3109-PB|mRpL16-PA|mRpL16-PB', map_location='2B14-2B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1890256, end_position_on_the_genomic_accession=1891482, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31151, CurrentID=0, Status='live', Symbol='arm', Aliases='Dmel_CG11579, ARM, Arm, CG11579, Dm Arm, Dmel\\CG11579, EG:86E4.6, beta-Cat, beta-cat, beta-cat-arm, beta-catenin, beta-catenin/Arm, betacat, bgr;-Cat, l(1)2Bv, l(1)G0192, l(1)G0234, l(1)G0410, l(1)arm, t12687 ALR Dm', description='armadillo', other_designations='Armadillo(Arm)/beta-catenin|Armadillo/beta-catenin|Armadillo/bgr;-catenin|Beta-catenin/Arm|CG11579-PA|CG11579-PB|CG11579-PC|CG11579-PD|CG11579-PE|CG11579-PF|alpha-Catenin|arm-PA|arm-PB|arm-PC|arm-PD|arm-PE|arm-PF|b-catenin|beta-Catenin|beta-catenin/Armadillo|betacatenin|bgr;-catenin|catenin', map_location='1-1.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1891401, end_position_on_the_genomic_accession=1900646, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31152, CurrentID=0, Status='live', Symbol='CG32801', Aliases='Dmel_ CG14819, Dmel\\CG32801', description='CG32801 gene product from transcript CG32801-RA', other_designations='CG32801-PA|uncharacterized protein', map_location='2B15-2B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1902082, end_position_on_the_genomic_accession=1903177, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31153, CurrentID=0, Status='live', Symbol='mip130', Aliases='Dmel_CG3480, 86E4.4, CG3480, Dmel\\CG3480, EG:86E4.4, Mip130, Mip130/TWIT, TWIT, Twit, twit', description='Myb-interacting protein 130', other_designations='CG3480-PA|mip130-PA|twilight', map_location='2B15-2B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1908239, end_position_on_the_genomic_accession=1920318, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31154, CurrentID=0, Status='live', Symbol='Edem1', Aliases='Dmel_CG3810, CG3810, DM-pI, Dmel\\CG3810, EDEM1, EG.86E4.2, EG:86E4.2, edem1', description='ER degradation enhancer, mannosidase alpha-like 1', other_designations='CG3810-PA|CG3810-PB|CG3810-PC|CG3810-PD|CG3810-PE|Edem1-PA|Edem1-PB|Edem1-PC|Edem1-PD|Edem1-PE', map_location='2B15-2B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1903704, end_position_on_the_genomic_accession=1908827, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31155 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31155, CurrentID=0, Status='live', Symbol='Rbcn-3B', Aliases='Dmel_CG17766, CG17766, Dmel\\CG17766, EG:86E4.3, FK39', description='Rabconnectin-3B', other_designations='CG17766-PA|CG17766-PB|Rabconnectin-3beta|Rbcn-3B-PA|Rbcn-3B-PB', map_location='2B15-2B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1908920, end_position_on_the_genomic_accession=1916825, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31156, CurrentID=0, Status='live', Symbol='eIF2B-epsilon', Aliases='Dmel_CG3806, CG3806, Dmel\\CG3806, EG:86E4.1, eIF-2bepsilon, eIF2Bepsilon, elf2B-epsilon', description='CG3806 gene product from transcript CG3806-RB', other_designations='CG3806-PA|CG3806-PB|eIF2B-epsilon-PA|eIF2B-epsilon-PB', map_location='2B16-2B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1920166, end_position_on_the_genomic_accession=1922874, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31157, CurrentID=0, Status='live', Symbol='Ocrl', Aliases='Dmel_CG3573, CG3573, Dmel\\CG3573, EG:86E4.5, dOCRL, ocrl', description='Oculocerebrorenal syndrome of Lowe', other_designations='CG3573-PA|CG3573-PB|Ocrl-PA|Ocrl-PB|inositol 5-phosphatase|inositol polyphosphate 5-phosphatase|oculocerebrorenal syndrome of lowe|oculocerebrorenal syndrome of lowe 1 ortholog|oculocerebrorenal syndrome of lowe ortholog', map_location='2B16-2B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1922981, end_position_on_the_genomic_accession=1927524, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31158, CurrentID=0, Status='live', Symbol='CG11596', Aliases='Dmel_ Dmel\\CG11596, EG:39E1.1', description='CG11596 gene product from transcript CG11596-RA', other_designations='CG11596-PA|CG11596-PB|CG11596-PC|CG11596-PD|CG11596-PE|uncharacterized protein', map_location='2B16-2B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1927683, end_position_on_the_genomic_accession=1930517, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31160, CurrentID=0, Status='live', Symbol='CG3587', Aliases='Dmel_ Dmel\\CG3587, EG:39E1.2', description='CG3587 gene product from transcript CG3587-RA', other_designations='CG3587-PA', map_location='2B16-2B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1940307, end_position_on_the_genomic_accession=1941536, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31162, CurrentID=0, Status='live', Symbol='Hr4', Aliases='Dmel_CG43934, CG16902, CG3600, CG42527, CG43692, CG43934, DHR4, Dmel\\CG43934, Dmel_CG16902, Dmel_CG3600, Dmel_CG42527, EG:133E12.2, EG:BACH61I5.1, EP(X)1232, EP1232, GRF, NR6A2, null', description='CG43934 gene product from transcript CG43934-RK', other_designations='CG43934-PC|CG43934-PD|CG43934-PE|CG43934-PF|CG43934-PG|CG43934-PH|CG43934-PI|CG43934-PJ|CG43934-PK|Hr4-PC|Hr4-PD|Hr4-PE|Hr4-PF|Hr4-PG|Hr4-PH|Hr4-PI|Hr4-PJ|Hr4-PK', map_location='2C1-2C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=1941940, end_position_on_the_genomic_accession=2007956, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  31163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31163, CurrentID=0, Status='live', Symbol='CG4406', Aliases='Dmel_ Dmel\\CG4406, EG:133E12.3, cg4406', description='CG4406 gene product from transcript CG4406-RA', other_designations='CG4406-PA', map_location='2C1-2C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2007911, end_position_on_the_genomic_accession=2009364, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31165, CurrentID=0, Status='live', Symbol='usp', Aliases='Dmel_CG4380, 2C, CF1, CF1/USP, CF1/Usp, CG4380, Cf1, DmUSP, DmUsp, Dmel\\CG4380, EG:22E5.1, NR2B4, USP, Usp, XR2C, cf1, dUSP, dmUSP, l(1)2Cf, l(1)usp', description='ultraspiracle', other_designations='CG4380-PA|CG4380-PB|X receptor at 2C|chorion factor 1|chorion factor CF1|ultraspiricle|usp-PA|usp-PB', map_location='2C7-2C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2038798, end_position_on_the_genomic_accession=2042311, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31166, CurrentID=0, Status='live', Symbol='Actn', Aliases='Dmel_CG4376, ACTN, CG4376, CT14163, CT14232, Dmel\\CG4376, EG:133E12.1, GA17, HM-29, actn, alpha-actinin, alphaActn, fliA, l(1)2Cb, l(1)EA82, y', description='alpha actinin', other_designations='Actn-PA|Actn-PB|Actn-PC|Actn-PD|Actn-PF|Actn-PG|Actn-PH|Actn-PI|Actn-PJ|Alpha-actinin|CG4376-PA|CG4376-PB|CG4376-PC|CG4376-PD|CG4376-PF|CG4376-PG|CG4376-PH|CG4376-PI|CG4376-PJ|a-actinin|flightless A|flightless-A|lethal(1)2cb', map_location='2C4-2C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2023822, end_position_on_the_genomic_accession=2042311, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  31167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31167, CurrentID=0, Status='live', Symbol='CG4325', Aliases='Dmel_ Dmel\\CG4325, EG:22E5.12', description='CG4325 gene product from transcript CG4325-RB', other_designations='CG4325-PA|CG4325-PB', map_location='2C7-2C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2042730, end_position_on_the_genomic_accession=2043818, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31168, CurrentID=0, Status='live', Symbol='moody', Aliases='Dmel_CG4322, 1631, CG4322, Dmel\\CG4322, EG:22E5.11, EP1631, Moody', description='CG4322 gene product from transcript CG4322-RD', other_designations='CG4322-PA|CG4322-PB|CG4322-PC|CG4322-PD|CG4322-PE|CG4322-PF|moody receptor|moody-PA|moody-PB|moody-PC|moody-PD|moody-PE|moody-PF', map_location='2C7-2C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2044292, end_position_on_the_genomic_accession=2050120, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31169, CurrentID=0, Status='live', Symbol='CG4313', Aliases='Dmel_ Dmel\\CG4313, EG:22E5.10', description='CG4313 gene product from transcript CG4313-RE', other_designations='CG4313-PC|CG4313-PD|CG4313-PE', map_location='2C8-2C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2041868, end_position_on_the_genomic_accession=2055414, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31170, CurrentID=0, Status='live', Symbol='Sik2', Aliases='Dmel_CG4290, CG4290, Dmel\\CG4290, EG:22E5.8, SIK, SIK2', description='Salt-inducible kinase 2', other_designations='CG4290-PA|Salt-induced kinase 2|Sik2-PA', map_location='2C8-2C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2055310, end_position_on_the_genomic_accession=2061915, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31171, CurrentID=0, Status='live', Symbol='CG4281', Aliases='Dmel_ Dmel\\CG4281, EG:22E5.7', description='CG4281 gene product from transcript CG4281-RA', other_designations='CG4281-PA', map_location='2C8-2C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2062313, end_position_on_the_genomic_accession=2065142, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31172, CurrentID=0, Status='live', Symbol='CG14054', Aliases='Dmel_ BcDNA:AT13882, Dmel\\CG14054', description='CG14054 gene product from transcript CG14054-RA', other_designations='CG14054-PA|uncharacterized protein', map_location='2C8-2C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2065273, end_position_on_the_genomic_accession=2066359, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31173, CurrentID=0, Status='live', Symbol='CG4194', Aliases='Dmel_ Dmel\\CG4194, EG:22E5.6', description='CG4194 gene product from transcript CG4194-RA', other_designations='CG4194-PA', map_location='2C9-2C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2068592, end_position_on_the_genomic_accession=2069688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31174, CurrentID=0, Status='live', Symbol='CG4199', Aliases='Dmel_ 143005_at, Dmel\\CG4199, EG:22E5.5', description='CG4199 gene product from transcript CG4199-RG', other_designations='CG4199-PA|CG4199-PB|CG4199-PC|CG4199-PD|CG4199-PG|CG4199-PH|CG4199-PI', map_location='2C9-2C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2066435, end_position_on_the_genomic_accession=2073483, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31175 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31175, CurrentID=0, Status='live', Symbol='CG4061', Aliases='Dmel_ Dmel\\CG4061, EG:22E5.3, RTC1_DROME', description='CG4061 gene product from transcript CG4061-RA', other_designations='CG4061-PA|CG4061-PB', map_location='2C10-2C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2072413, end_position_on_the_genomic_accession=2074890, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31176, CurrentID=0, Status='live', Symbol='CG4045', Aliases='Dmel_ Dmel\\CG4045, EG:22E5.4', description='CG4045 gene product from transcript CG4045-RA', other_designations='CG4045-PA', map_location='2C10-2C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2074795, end_position_on_the_genomic_accession=2075900, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31177, CurrentID=0, Status='live', Symbol='CG4025', Aliases='Dmel_ DRAM, Dmel\\CG4025, EG:22E5.9, anon-WO0140519.169', description='CG4025 gene product from transcript CG4025-RA', other_designations='CG4025-PA', map_location='2C10-2C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2076966, end_position_on_the_genomic_accession=2078455, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31178, CurrentID=0, Status='live', Symbol='CG16903', Aliases='Dmel_ Dmel\\CG16903, EG:67A9.2, anon-WO03040301.211', description='CG16903 gene product from transcript CG16903-RA', other_designations='CG16903-PA|uncharacterized protein', map_location='2C10-2C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2078343, end_position_on_the_genomic_accession=2080405, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31179, CurrentID=0, Status='live', Symbol='Unc-76', Aliases='Dmel_CG3981, BcDNA:GH10260, CG18851, CG3981, Dmel\\CG3981, Dunc-76, EG:67A9.1, UNC-76, l(1)G0066, l(1)G0158, l(1)G0310, l(1)G0333, l(1)G0360, unc-76', description='CG3981 gene product from transcript CG3981-RC', other_designations='CG3981-PA|CG3981-PB|CG3981-PC|CG3981-PD|Unc-76-PA|Unc-76-PB|Unc-76-PC|Unc-76-PD', map_location='2C10-2D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2080610, end_position_on_the_genomic_accession=2093948, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31181 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31181, CurrentID=0, Status='live', Symbol='ph-p', Aliases='Dmel_CG18412, CG18412, CG18414, Dmel\\CG18412, EG:87B1.5, PH, PH-p, PHP, Ph, Ph-p, PhP, dPh, ph, ph-P, ph[P], ph[p], phm, php', description='polyhomeotic proximal', other_designations='CG18412-PA|CG18412-PB|CG18412-PC|CG18412-PD|Polyhomeotic-proximal|ph-p-PA|ph-p-PB|ph-p-PC|ph-p-PD|ph-proximal|polyhomeotic|polyhomeotic proxima', map_location='1-0.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2126210, end_position_on_the_genomic_accession=2138371, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31184, CurrentID=0, Status='live', Symbol='CG3835', Aliases='Dmel_ Dmel\\CG3835, EG87B1, EG:87B1.3', description='CG3835 gene product from transcript CG3835-RC', other_designations='CG3835-PA|CG3835-PB|CG3835-PC', map_location='2D4-2D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2141489, end_position_on_the_genomic_accession=2144937, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31185, CurrentID=0, Status='live', Symbol='Pgd', Aliases='Dmel_CG3724, 6-PGD, 6-Pgd, 6-pgd, 6PGD, 6PGDH, 6Pgd, 6Pgdh, 6pgdh, A7, CG3724, Dmel\\CG3724, EG:87B1.4, N1, PGD, Pdg, l(1)2Dc, l(1)2De, l(1)A7, l(1)G0385, l(1)N1, l(1)N3[90], l(1)Pgd, l(1)Pgd-A, l35, pgd', description='Phosphogluconate dehydrogenase', other_designations='6-phosphogluconate|6-phosphogluconate dehydrogenase|CG3724-PA|CG3724-PB|NADP-6 phosphogluconate dehydrogenase|Pgd-PA|Pgd-PB|dehydrogenase|phosphogluconate dehydrogenase', map_location='2D4-2D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2145136, end_position_on_the_genomic_accession=2148566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31186, CurrentID=0, Status='live', Symbol='bcn92', Aliases='Dmel_CG3717, BcDNA:RE57896, CG3717, Dmel\\CG3717, EG:87B1.6, NEST:bs15a04', description='CG3717 gene product from transcript CG3717-RA', other_designations='CG3717-PA|bcn92-PA', map_location='2D4-2D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2148463, end_position_on_the_genomic_accession=2149647, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31187, CurrentID=0, Status='live', Symbol='wapl', Aliases='Dmel_CG3707, C204, CG3707, Dmel\\CG3707, EG:87B1.2, N2, RNA D, RNAD, Wapl, l(1)2Dd, l(1)2Df, l(1)C204, l(1)N2, l(1)N5[72], l74, pasc, wap', description='wings apart-like', other_designations='CG3707-PA|CG3707-PB|CG3707-PC|parallel sister chromatids protein|wapl-PA|wapl-PB|wapl-PC|wings-apart-like', map_location='1-0.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2149936, end_position_on_the_genomic_accession=2158983, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31188, CurrentID=0, Status='live', Symbol='Cyp4d1', Aliases='Dmel_CG3656, 4d1, CG3656, CYP4D1, CYP6D1, Cyp4D1, Cyt-P450-4D1, Dmel\\CG3656, EG:87B1.1, P-450, P450, cyp4d1', description='Cytochrome P450-4d1', other_designations='CG3656-PA|CG3656-PB|Cyp4d1-PA|Cyp4d1-PB|P450 CYPD1|cytochrome P450|cytochrome P450 4D1|cytochrome P450-4d1', map_location='2D5-2D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2159145, end_position_on_the_genomic_accession=2162180, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31189, CurrentID=0, Status='live', Symbol='CG3630', Aliases='Dmel_ 152A3.3, Dmel\\CG3630, EG:152A3.3', description='CG3630 gene product from transcript CG3630-RD', other_designations='CG3630-PA|CG3630-PB|CG3630-PC|CG3630-PD', map_location='2D6-2D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2162196, end_position_on_the_genomic_accession=2172281, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31190, CurrentID=0, Status='live', Symbol='ND-B14.5A', Aliases='Dmel_CG3621, 152A3.7, BcDNA:RE75088, CG3621, Dmel\\CG3621, EG:152A3.7, N4AM, RNAA, anon-2D, anon-2Db', description='NADH dehydrogenase (ubiquinone) B14.5 A subunit', other_designations='CG3621-PA|CG3621-PB|FBpp0070370|NADH ubiquinone oxidoreductase B14.5A|NADH-ubiquinone oxidoreductase subunit B14.5a|NADH:ubiquinone oxidoreductase B14.5A subunit|ND-B14.5A-PA|ND-B14.5A-PB', map_location='2D6-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2172526, end_position_on_the_genomic_accession=2173432, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31192, CurrentID=0, Status='live', Symbol='Cyp4d2', Aliases='Dmel_CG3466, 4d2, CG3466, CYP4D2, Cyp4D2, Cyt-P450-4D2, Dmel\\CG3466, EG:152A3.4, P-450, P450', description='Cytochrome P450-4d2', other_designations='CG3466-PA|CG3466-PB|Cyp4d2-PA|Cyp4d2-PB|cytochrome P-450|cytochrome P450|cytochrome P450 4D2|cytochrome P450-4d2', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2176006, end_position_on_the_genomic_accession=2178728, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31193, CurrentID=0, Status='live', Symbol='Cyp4ae1', Aliases='Dmel_CG10755, 4ae1, CG10755, CYP4E4, Cyp4e4, Cypae1, Dmel\\CG10755, EG:152A3.6, I, P-450, P450, TcD20, anon-2Ea', description='Cytochrome P450-4ae1', other_designations='CG10755-PA|Cyp4ae1-PA|cytochrome P450-4ae1|transcription unit I', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2179007, end_position_on_the_genomic_accession=2180998, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31194, CurrentID=0, Status='live', Symbol='pn', Aliases='Dmel_CG3461, CG3461, Dmel\\CG3461, EG:152A3.5, N3, N4, TcD37', description='prune', other_designations='CG3461-PA|pn-PA|se-like-62|transcription unit II', map_location='1-0.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2181211, end_position_on_the_genomic_accession=2182916, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31195, CurrentID=0, Status='live', Symbol='Nmd3', Aliases='Dmel_CG3460, CG3460, Dmel\\CG3460, EG:152A3.1, II, TcD12, TcD13, anon-2Eb, dNmd3', description='Nonsense-mediated mRNA 3', other_designations='CG3460-PA|Nmd3-PA|transcription unit III', map_location='2E1-2E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2182818, end_position_on_the_genomic_accession=2184668, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326108, CurrentID=0, Status='live', Symbol='EbpII', Aliases='Dmel_CG2665, CG2665, Dmel\\CG2665, PEB, PEB2, PEBII, PEBmeII, PebII', description='Ejaculatory bulb protein II', other_designations='CG2665-PA|CG2665-PB|EbpII-PA|EbpII-PB|ejaculatory bulb protein II|protein ejaculatory bulb II', map_location='60F5-60F5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=25173568, end_position_on_the_genomic_accession=25174230, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326109, CurrentID=0, Status='live', Symbol='CG18858', Aliases='Dmel_ 38C.53, Dmel\\CG18858', description='CG18858 gene product from transcript CG18858-RB', other_designations='CG18858-PA|CG18858-PB|uncharacterized protein', map_location='38C7-38C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20456847, end_position_on_the_genomic_accession=20459075, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  326110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326110, CurrentID=0, Status='live', Symbol='Obp57e', Aliases='Dmel_CG30145, 57e, CG13429, CG30145, DmelObp57e, Dmel\\CG30145, Dmel\\Obp57e, OBP57e, Obp57d', description='Odorant-binding protein 57e', other_designations='CG30145-PA|Obp57e-PA', map_location='57A5-57A5', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=20547816, end_position_on_the_genomic_accession=20548291, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326111 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326111, CurrentID=0, Status='live', Symbol='CG31013', Aliases='Dmel_ CG9713, Dmel\\CG31013', description='CG31013 gene product from transcript CG31013-RA', other_designations='CG31013-PA|uncharacterized protein', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30643767, end_position_on_the_genomic_accession=30645880, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  326112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326112, CurrentID=0, Status='live', Symbol='PH4alphaSG1', Aliases='Dmel_CG31014, CG15542, CG31014, Dmel\\CG31014, PHalphaSG1, Ph4alphaSG1, SG1, alphaSG1', description='prolyl-4-hydroxylase-alpha SG1', other_designations='CG31014-PA|PH4alphaSG1-PA|SG specific ER protein prolyl-4-hydroxylase|prolyl 4-hydroxylase alpha salivary gland 1', map_location='99F6-99F6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30503665, end_position_on_the_genomic_accession=30505661, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31201, CurrentID=0, Status='live', Symbol='Vinc', Aliases='Dmel_CG3299, CG3299, CT11081, Dmel\\CG3299, Dmvincp, EG:103B4.1, Vin, Vin2EF, vinc', description='Vinculin', other_designations='CG3299-PA|CG3299-PB|CG3299-PC|Vinc-PA|Vinc-PB|Vinc-PC|vinculin|vinculin at 2EF', map_location='2E2-2E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2214347, end_position_on_the_genomic_accession=2222567, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  326114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326114, CurrentID=0, Status='live', Symbol='PH4alphaNE3', Aliases='Dmel_CG31017, CG31017, CG9708, Dmel\\CG31017', description='prolyl-4-hydroxylase-alpha NE3', other_designations='CG31017-PA|PH4alphaNE3-PA', map_location='100A1-100A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=30604181, end_position_on_the_genomic_accession=30606144, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  326115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326115, CurrentID=0, Status='live', Symbol='Atg16', Aliases='Dmel_CG31033, CG15513, CG18040, CG18683, CG31033, Dmel\\CG31033', description='Autophagy-related 16', other_designations='Atg16-PA|Atg16-PB|Atg16-PC|Atg16-PF|Autophagy-specific gene 16|CG31033-PA|CG31033-PB|CG31033-PC|CG31033-PF', map_location='99C2-99C2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29827281, end_position_on_the_genomic_accession=29835101, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  326116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326116, CurrentID=0, Status='live', Symbol='tau', Aliases='Dmel_CG45110, BcDNA:RE16764, CG12881, CG31057, CG31058, CG34294, CG45110, CG5606, Dmel\\CG45110, Dmel_CG31057, Dmel_CG31058, Dmel_CG34294, Tau, d- dTau, dtau', description='CG45110 gene product from transcript CG45110-RH', other_designations='CG45110-PA|CG45110-PB|CG45110-PC|CG45110-PE|CG45110-PF|CG45110-PG|CG45110-PH|CG45110-PI|CG45110-PJ|CG45110-PK|CG45110-PL|CG45110-PM|CG45110-PN|tau-PA|tau-PB|tau-PC|tau-PE|tau-PF|tau-PG|tau-PH|tau-PI|tau-PJ|tau-PK|tau-PL|tau-PM|tau-PN', map_location='98A14-98A14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27639886, end_position_on_the_genomic_accession=27656996, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  31205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31205, CurrentID=0, Status='live', Symbol='kz', Aliases='Dmel_CG3228, CG3228, Dmel\\CG3228, EG:30B8.2, N1, N8, l(1)2Eb, l(1)N4, l67', description='kurz', other_designations='CG3228-PA|CG3228-PB|kz-PA|kz-PB', map_location='1-0.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2237858, end_position_on_the_genomic_accession=2242035, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  326118 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326118, CurrentID=0, Status='live', Symbol='CG31102', Aliases='Dmel_ CG10634, Dmel\\CG31102', description='CG31102 gene product from transcript CG31102-RA', other_designations='CG31102-PA|CG31102-PB|uncharacterized protein', map_location='96D1-96D1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=25293458, end_position_on_the_genomic_accession=25295594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31207, CurrentID=0, Status='live', Symbol='Or2a', Aliases='Dmel_CG3206, 2F.1, 2a, AN4, CG3206, DOR2F.1, DOR62, Dmel\\CG3206, Dor62, EG:30B8.7, Or2F.1, Or62, dor62, or2a', description='Odorant receptor 2a', other_designations='CG3206-PA|Or2a-PA|odorant receptor 2a|odorant receptor 62|olfactory receptor 2F.1', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2246231, end_position_on_the_genomic_accession=2247567, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326120, CurrentID=0, Status='live', Symbol='HDAC11', Aliases='Dmel_CG31119, CG31119, DHDACX, Dmel\\CG31119, HDACX, Hdac11, HdacX, dHDAC11, dmHDA403', description='Histone deacetylase 11', other_designations='CG31119-PA|CG31119-PB|HDAC11-PA|HDAC11-PB|histone deacetylase 11|histone deacetylase X', map_location='96B1-96B1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24854496, end_position_on_the_genomic_accession=24856243, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31209, CurrentID=0, Status='live', Symbol='CG3191', Aliases='Dmel_ 62D9.a, BEST:GM01072, CG18842, Dmel\\CG3191, EG:30B8.3', description='CG3191 gene product from transcript CG3191-RA', other_designations='CG3191-PA|CG3191-PC', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2250145, end_position_on_the_genomic_accession=2251993, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326122, CurrentID=0, Status='live', Symbol='rumi', Aliases='Dmel_CG31152, CG17138, CG31152, Dmel\\CG31152, Rumi', description='CG31152 gene product from transcript CG31152-RB', other_designations='CG31152-PA|CG31152-PB|rumi-PA|rumi-PB', map_location='94C4-94C4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=22749014, end_position_on_the_genomic_accession=22750597, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31211, CurrentID=0, Status='live', Symbol='CG3078', Aliases='Dmel_ Dmel\\CG3078, EG:30B8.6', description='CG3078 gene product from transcript CG3078-RA', other_designations='CG3078-PA|CG3078-PB', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2254403, end_position_on_the_genomic_accession=2259863, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31212, CurrentID=0, Status='live', Symbol='temp', Aliases='Dmel_CG3073, CG3073, Dmel\\CG3073, l(1)G0144', description='tempura', other_designations='CG3073-PA|CG3073-PB|lethal (1) G0144|temp-PA|temp-PB', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2260207, end_position_on_the_genomic_accession=2262515, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31213, CurrentID=0, Status='live', Symbol='CG3071', Aliases='Dmel_ Dmel\\CG3071, EG:25E8.3', description='CG3071 gene product from transcript CG3071-RA', other_designations='CG3071-PA', map_location='2F2-2F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2262837, end_position_on_the_genomic_accession=2264825, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31214, CurrentID=0, Status='live', Symbol='CG2924', Aliases='Dmel_ Dmel\\CG2924, EG:25E8.2, NICE5', description='CG2924 gene product from transcript CG2924-RE', other_designations='CG2924-PA|CG2924-PB|CG2924-PD|CG2924-PE', map_location='2F2-2F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2264657, end_position_on_the_genomic_accession=2270761, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31215, CurrentID=0, Status='live', Symbol='CG2918', Aliases='Dmel_ CT9894, Dmel\\CG2918, EG:25E8.1, GRP170', description='CG2918 gene product from transcript CG2918-RB', other_designations='CG2918-PA|CG2918-PB', map_location='2F3-2F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2271008, end_position_on_the_genomic_accession=2275068, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  326128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326128, CurrentID=0, Status='live', Symbol='Ada2a', Aliases='Dmel_CG43663, ADA2a, Ada2A, CG31318, CG33520, CG43663, CG7150, DmADA2a, DmAda2a, Dmel\\CG43663, ada2a, dADA2a, dAda2A, dAda2a', description='transcriptional Adaptor 2a', other_designations='Ada2a-PA|Ada2a-PB|Ada2a-PC|Ada2a-PE|Ada2a-PG|CG43663-PA|CG43663-PB|CG43663-PC|CG43663-PE|CG43663-PG|transcriptional adaptor 2a', map_location='90F10-90F11', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=18252843, end_position_on_the_genomic_accession=18255487, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31217, CurrentID=0, Status='live', Symbol='CG2865', Aliases='Dmel_ Dmel\\CG2865, EG:25E8.4, anon-WO03040301.165', description='CG2865 gene product from transcript CG2865-RB', other_designations='CG2865-PA|CG2865-PB', map_location='2F4-2F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2278031, end_position_on_the_genomic_accession=2293411, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326130, CurrentID=0, Status='live', Symbol='CG31272', Aliases='Dmel_ CG6529, Dmel\\CG31272', description='CG31272 gene product from transcript CG31272-RA', other_designations='CG31272-PA|uncharacterized protein', map_location='86C5-86C5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10756296, end_position_on_the_genomic_accession=10759398, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  326131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326131, CurrentID=0, Status='live', Symbol='CG31275', Aliases='Dmel_ BcDNA:GH08635, CG14908, Dmel\\CG31275', description='CG31275 gene product from transcript CG31275-RA', other_designations='CG31275-PA|late transcript|uncharacterized protein', map_location='89E1-89E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=16750741, end_position_on_the_genomic_accession=16751334, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31220 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31220, CurrentID=0, Status='live', Symbol='Ilp6', Aliases='Dmel_CG14049, CG14049, DILP, DILP-6, DILP6, Dilp, Dilp 6, Dilp6, Dmel\\CG14049, Ilp, Ilp-6, dILP6, dIlp6, dilp, dilp-6, dilp6, ilp6', description='Insulin-like peptide 6', other_designations='CG14049-PA|CG14049-PB|CG14049-PC|CG14049-PD|Ilp6-PA|Ilp6-PB|Ilp6-PC|Ilp6-PD|drosophila insulin-like peptide 6|insulin-like peptide|insulin-like peptide 6', map_location='3A1-3A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2331493, end_position_on_the_genomic_accession=2334658, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31221, CurrentID=0, Status='live', Symbol='Raf', Aliases='Dmel_CG2845, 11-29, C110, CG2845, D-RAF, D-Raf, D-raf, D-raf1, DRaf, DRaf1, Dmel\\CG2845, Draf, Draf-1, Draf1, EG:BACH48C10.3, Phl, RAF/phl, Raf1, dRaf, draf, l(1)2Fe, l(1)G0475, l(1)ph, l(1)phl, l(1)pole hole, l(1)polehole, l(1)polehole/draf, l(1)raf, lincRNA.937, ph, ph1, phl, raf, raf-1, raf1, rafl, Raf', description='Raf oncogene', other_designations='CG2845-PA|CG2845-PE|Raf-PA|Raf-PE|l(1)pole hole|pole hole|pole-hole|polehole|raf-like-oncogene-1', map_location='1-0.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2295466, end_position_on_the_genomic_accession=2343870, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31222, CurrentID=0, Status='live', Symbol='mRpL14', Aliases='Dmel_CG14048, BcDNA:RH18819, CG14048, Dmel\\CG14048, EG:BACH48C10.6, L14, mRPL14', description='mitochondrial ribosomal protein L14', other_designations='CG14048-PA|CG14048-PB|mRpL14-PA|mRpL14-PB', map_location='3A1-3A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2343997, end_position_on_the_genomic_accession=2344649, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326135, CurrentID=0, Status='live', Symbol='CG31357', Aliases='Dmel_ CG13638, Dmel\\CG31357', description='CG31357 gene product from transcript CG31357-RA', other_designations='CG31357-PA|uncharacterized protein', map_location='96A24-96A24', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24828279, end_position_on_the_genomic_accession=24829959, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31224, CurrentID=0, Status='live', Symbol='PsGEF', Aliases='Dmel_CG43947, CG14045, CG14047, CG43947, DmPsGEF, Dmel\\CG43947, Dmel_CG14045, Dmel_CG14047, EG:BACH48C10.4, EG:BACH7M4.1, EG:BACH7M4.2, FBgn0264598, PsGef, null', description='Protostome-specific GEF', other_designations='CG43947-PA|CG43947-PB|CG43947-PC|CG43947-PE|CG43947-PF|CG43947-PG|CG43947-PH|CG43947-PI|PsGEF-PA|PsGEF-PB|PsGEF-PC|PsGEF-PE|PsGEF-PF|PsGEF-PG|PsGEF-PH|PsGEF-PI|protostome-specific GEF', map_location='3A2-3A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2367121, end_position_on_the_genomic_accession=2387664, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  326137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326137, CurrentID=0, Status='live', Symbol='CG31419', Aliases='Dmel_ CG5329, Dmel\\CG31419', description='CG31419 gene product from transcript CG31419-RB', other_designations='CG31419-PB|uncharacterized protein', map_location='90A1-90A1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=17284648, end_position_on_the_genomic_accession=17285421, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326138, CurrentID=0, Status='live', Symbol='CG31431', Aliases='Dmel_ CG13414, CT32770, Dmel\\CG31431', description='CG31431 gene product from transcript CG31431-RC', other_designations='CG31431-PA|CG31431-PB|CG31431-PC|uncharacterized protein', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21870681, end_position_on_the_genomic_accession=21873640, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31227, CurrentID=0, Status='live', Symbol='gt', Aliases='Dmel_CG7952, CG7952, DM17, Dmel\\CG7952, EG:BACH7M4.5, GIAN_DROME, GT, Giant, Gt, l(1)3Aa, l(1)giant', description='giant', other_designations='CG7952-PB|gt-PB', map_location='1-0.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2427113, end_position_on_the_genomic_accession=2428967, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31228, CurrentID=0, Status='live', Symbol='tko', Aliases='Dmel_CG7925, CG7925, Dmel\\CG7925, EG:BACH59J11.1, MRP-S12, MRPS12, S12, Tko, l(1)3Ab, l(1)tko, mRpS12, mt-rps12, rps12', description='technical knockout', other_designations='CG7925-PB|CG7925-PC|mitochondrial ribosomal protein S12|mitoribosomal protein S12|tko-PB|tko-PC', map_location='1-1.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2442313, end_position_on_the_genomic_accession=2443976, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31229, CurrentID=0, Status='live', Symbol='boi', Aliases='Dmel_CG32796, Boi, CG13756, CG32796, CG7894, CT23737, CT33235, Dmel\\CG32796, EG:BACH59J11.2', description='brother of ihog', other_designations='CG32796-PA|CG32796-PB|CG32796-PE|CG32796-PF|CG32796-PG|boi-PA|boi-PB|boi-PE|boi-PF|boi-PG|brother of iHog', map_location='3A3-3A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2444167, end_position_on_the_genomic_accession=2470043, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  31230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31230, CurrentID=0, Status='live', Symbol='z', Aliases='Dmel_CG7803, CG7803, Dmel\\CG7803, EG:BACH59J11.3, Z, e(bx), en-bx', description='zeste', other_designations='CG7803-PA|CG7803-PB|CG7803-PC|enhancer-of-bithorax|z-PA|z-PB|z-PC', map_location='1-1.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2447769, end_position_on_the_genomic_accession=2450550, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326144, CurrentID=0, Status='live', Symbol='ATPsynepsilonL', Aliases='Dmel_CG31477, ATPE, CG12810, CG31477, Dmel\\CG31477', description='ATP synthase, epsilon subunit-like', other_designations='ATP synthase epsilon chain|ATPsynepsilonL-PA|ATPsynepsilonL-PB|ATPsynepsilonL-PC|CG31477-PA|CG31477-PB|CG31477-PC', map_location='85F13-85F13', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=10121773, end_position_on_the_genomic_accession=10122320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326145, CurrentID=0, Status='live', Symbol='CG31496', Aliases='Dmel_ CG14604, Dmel\\CG31496', description='CG31496 gene product from transcript CG31496-RA', other_designations='CG31496-PA|uncharacterized protein', map_location='84D5-84D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=7345069, end_position_on_the_genomic_accession=7346354, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31234, CurrentID=0, Status='live', Symbol='Pdfr', Aliases='Dmel_CG13758, BACR25B3.3, CG13758, DmeCG13758, Dmel\\CG13758, EG:BACR25B3.3, Han, PDF-R1, PDFR, Pdf-R1, PdfR, anon-WO0170980.157, anon-WO0170980.158, gop, han, pdfr', description='Pigment-dispersing factor receptor', other_designations='CG13758-PA|CG13758-PB|CG13758-PC|CG13758-PD|PDF receptor|PDF-receptor|Pdf recep- tor|Pdfr-PA|Pdfr-PB|Pdfr-PC|Pdfr-PD|groom-of-pdf|pigment dispersing factor receptor|pigment dispersing factor receptor 1|pigment-dispersing factor receptor', map_location='3A4-3A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2552206, end_position_on_the_genomic_accession=2578640, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  31235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31235, CurrentID=0, Status='live', Symbol='Vha36-3', Aliases='Dmel_CG8310, CG8310, Dmel\\CG8310, EG:BACR25B3.4, vha36-3', description='Vacuolar H[+] ATPase 36kD subunit 3', other_designations='CG8310-PB|CG8310-PC|Vha36-3-PB|Vha36-3-PC|vacuolar H[+] ATPase 36kD subunit 3|vacuolar H[+] ATPase subunit 36-3', map_location='3A6-3A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2579091, end_position_on_the_genomic_accession=2580560, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31236 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31236, CurrentID=0, Status='live', Symbol='CG13759', Aliases='Dmel_ BcDNA:AT19803, CT33239, Dmel\\CG13759, EG:BACR25B3.5', description='CG13759 gene product from transcript CG13759-RB', other_designations='CG13759-PA|CG13759-PB|CG13759-PC|CG13759-PD|uncharacterized protein', map_location='3A6-3A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2580958, end_position_on_the_genomic_accession=2583823, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31237, CurrentID=0, Status='live', Symbol='CG13760', Aliases='Dmel_ Dmel\\CG13760, EG:BACR25B3.6', description='CG13760 gene product from transcript CG13760-RC', other_designations='CG13760-PC|CG13760-PD|uncharacterized protein', map_location='3A6-3A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2583745, end_position_on_the_genomic_accession=2585699, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  326150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326150, CurrentID=0, Status='live', Symbol='CG31650', Aliases='Dmel_ CG9112, Dmel\\CG31650, RCN2', description='CG31650 gene product from transcript CG31650-RF', other_designations='CG31650-PA|CG31650-PB|CG31650-PC|CG31650-PD|CG31650-PE|CG31650-PF|CG31650-PG|uncharacterized protein', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5040648, end_position_on_the_genomic_accession=5044712, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326151, CurrentID=0, Status='live', Symbol='pgant5', Aliases='Dmel_CG31651, CG31651, CG9152, Dmel\\CG31651, PGANT5, l(2)e02157', description='polypeptide GalNAc transferase 5', other_designations='CG31651-PA|CG31651-PB|lethal (2) e02157|pgant5-PA|pgant5-PB', map_location='25C4-25C6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5072170, end_position_on_the_genomic_accession=5093592, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  326152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326152, CurrentID=0, Status='live', Symbol='erm', Aliases='Dmel_CG31670, CG31670, CG7312, Dmel\\CG31670, dFezf, dfezl', description='earmuff', other_designations='CG31670-PA|CG31670-PB|CG31670-PC|CG31670-PD|erm-PA|erm-PB|erm-PC|erm-PD', map_location='22B6-22B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1950239, end_position_on_the_genomic_accession=1970362, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326153, CurrentID=0, Status='live', Symbol='tho2', Aliases='Dmel_CG31671, CG31671, CG4263, Dmel\\CG31671, THO2, Tho2, macadamia', description='CG31671 gene product from transcript CG31671-RA', other_designations='CG31671-PA|CG31671-PB|CG31671-PC|tho2-PA|tho2-PB|tho2-PC', map_location='22D1-22D1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2137350, end_position_on_the_genomic_accession=2145187, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  326154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326154, CurrentID=0, Status='live', Symbol='CG31689', Aliases='Dmel_ CG9892, Dmel\\CG31689', description='CG31689 gene product from transcript CG31689-RH', other_designations='CG31689-PA|CG31689-PB|CG31689-PC|CG31689-PD|CG31689-PE|CG31689-PF|CG31689-PG|CG31689-PH|uncharacterized protein', map_location='23A3-23A3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=2725674, end_position_on_the_genomic_accession=2739830, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  31243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31243, CurrentID=0, Status='live', Symbol='eIF3ga', Aliases='Dmel_CG8636, CG8636, Dmel\\CG8636, cg8636, eIF s4, eIF-3, eIF3-S4', description='eIF3g paralogue a', other_designations='CG8636-PA|eIF3ga-PA', map_location='3A7-3A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2609247, end_position_on_the_genomic_accession=2610336, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31244, CurrentID=0, Status='live', Symbol='Tsp3A', Aliases='Dmel_CG10742, BACR7C10.6, CG10742, Dm.Tsp3A, Dmel\\CG10742, EG:BACR7C10.6, tsp3A', description='Tetraspanin 3A', other_designations='CG10742-PA|CG10742-PB|Tsp3A-PA|Tsp3A-PB|tetraspanin 3A', map_location='3A7-3A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2617758, end_position_on_the_genomic_accession=2619738, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31245, CurrentID=0, Status='live', Symbol='Seipin', Aliases='Dmel_CG9904, CG9904, Dmel\\CG9904, EG:BACR7C10.1, d', description='CG9904 gene product from transcript CG9904-RA', other_designations='CG9904-PA|Seipin-PA|dSeipin|seipin', map_location='3A7-3A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2619799, end_position_on_the_genomic_accession=2622140, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31246, CurrentID=0, Status='live', Symbol='brv3', Aliases='Dmel_CG13762, CG13762, CT33244, Dmel\\CG13762, EG:BACR7C10.7', description='brivido-3', other_designations='CG13762-PB|CG13762-PC|brv3-PB|brv3-PC', map_location='3A8-3A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2622069, end_position_on_the_genomic_accession=2624575, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31247 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31247, CurrentID=0, Status='live', Symbol='PI4KIIIalpha', Aliases='Dmel_CG10260, CG10260, Dmel\\CG10260, EG:BACR7C10.2, PI4K, PI4KIII[[alpha]], PI4KinaseIIIalpha, Pi4KIIIalpha, ZW-2, l(1)17-59, l(1)3Ah, l(1)VA722, l(1)zw2, stt4, zw-2, zw2', description='Phosphatidylinositol 4-kinase III alpha', other_designations='CG10260-PC|CG10260-PD|PI4KIIIalpha-PC|PI4KIIIalpha-PD|lethal (1) 3Ah|phosphatidylinositol 4-kinase III alpha|phosphatidylinositol 4-kinase IIIalpha|zeste-white 2', map_location='1-1.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2624804, end_position_on_the_genomic_accession=2633672, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  326160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326160, CurrentID=0, Status='live', Symbol='ninaD', Aliases='Dmel_CG31783, CG31783, CG5750, Dmel\\CG31783, NINAD', description='neither inactivation nor afterpotential D', other_designations='CG31783-PA|CG31783-PB|ninaD-PA|ninaD-PB', map_location='36E3-36E3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=18081628, end_position_on_the_genomic_accession=18083608, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31249, CurrentID=0, Status='live', Symbol='HLH3B', Aliases='Dmel_CG2655, CG2655, Dmel\\CG2655, DroSCL, EG:155E2.2, HLH3b, bHLHa20', description='Helix loop helix protein 3B', other_designations='CG2655-PA|HLH3B-PA|helix loop helix protein 3B', map_location='3B1-3B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2680199, end_position_on_the_genomic_accession=2681723, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31250, CurrentID=0, Status='live', Symbol='CG2652', Aliases='Dmel_ Dmel\\CG2652, EG:155E2.5', description='CG2652 gene product from transcript CG2652-RA', other_designations='CG2652-PA', map_location='3B1-3B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2681987, end_position_on_the_genomic_accession=2682932, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31251 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31251, CurrentID=0, Status='live', Symbol='per', Aliases='Dmel_CG2647, CG2647, Clk, Dmel\\CG2647, EG:155E2.4, PER, Per, clk-6, dPER, dmper, dper, mel_per', description='period', other_designations='CG2647-PA|CG2647-PB|PERIOD|clock|clock-6|dperiod|per-PA|per-PB|period clock protein', map_location='3B1-3B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2685580, end_position_on_the_genomic_accession=2692780, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31252 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31252, CurrentID=0, Status='live', Symbol='CG2650', Aliases='Dmel_ 0.9, 0.9 gene, Dmel\\CG2650, EG:155E2.1, anon-3B1.2', description='CG2650 gene product from transcript CG2650-RA', other_designations='0.9kb transcript|CG2650-PA|anonymous 0.9 kb RNA', map_location='1-1.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2692732, end_position_on_the_genomic_accession=2693886, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31253 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31253, CurrentID=0, Status='live', Symbol='CG2658', Aliases='Dmel_ Dm100G10.aj, Dmel\\CG2658, EG:100G10.7, anon-3Ba', description='CG2658 gene product from transcript CG2658-RC', other_designations='CG2658-PA|CG2658-PC|paraplegin_DROME', map_location='3B2-3B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2694392, end_position_on_the_genomic_accession=2698211, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  326166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326166, CurrentID=0, Status='live', Symbol='Tap42', Aliases='Dmel_CG31852, CG31852, CG6537, Dmel\\CG31852, Tap-42/CG31852, dTap42', description='Two A-associated protein of 42kDa', other_designations='CG31852-PA|Tap42-PA', map_location='34B7-34B8', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13367617, end_position_on_the_genomic_accession=13369021, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31255, CurrentID=0, Status='live', Symbol='Csat', Aliases='Dmel_CG2675, CG2675, CSAT, DmNST, DmUGT, Dmel\\CG2675, EG:100G10.5, UDP-Gal/UDP-GalNAc, anon-3Bb, dNST-1, dNST-2, ugt', description='CG2675 gene product from transcript CG2675-RC', other_designations='CG2675-PA|CG2675-PB|CG2675-PC|Csat-PA|Csat-PB|Csat-PC|D.melanogaster nucleotide sugar transporter|csat', map_location='3B2-3B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2701124, end_position_on_the_genomic_accession=2703852, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326168, CurrentID=0, Status='live', Symbol='Mulk', Aliases='Dmel_CG31873, CG18162, CG31873, Dmel\\CG31873, Dmulk, anon-31BCb', description='Multi-substrate lipid kinase', other_designations='CG31873-PA|Mulk-PA|anon-31BCb', map_location='31D-31D', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=10270051, end_position_on_the_genomic_accession=10271465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31257, CurrentID=0, Status='live', Symbol='CG2680', Aliases='Dmel_ 3218519, Dmel\\CG2680, EG:100G10.4, Phos2680', description='CG2680 gene product from transcript CG2680-RB', other_designations='CG2680-PB', map_location='3B2-3B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2706151, end_position_on_the_genomic_accession=2707427, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326170, CurrentID=0, Status='live', Symbol='CG31886', Aliases='Dmel_ CG9524, Dmel\\CG31886, anon-WO0172774.159, anon-WO0172774.161', description='CG31886 gene product from transcript CG31886-RB', other_designations='CG31886-PA|CG31886-PB|CG31886-PC|uncharacterized protein', map_location='29F5-29F5', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=8953481, end_position_on_the_genomic_accession=8959051, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  326171 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326171, CurrentID=0, Status='live', Symbol='CG31917', Aliases='Dmel_ BcDNA:RE52596, CG14038, Dmel\\CG31917, Dmp18, Dromel_CG31917_FBtr0079037_mORF', description='CG31917 gene product from transcript CG31917-RA', other_designations='CG31917-PA|uncharacterized protein', map_location='25C1-25C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5009931, end_position_on_the_genomic_accession=5010834, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326172, CurrentID=0, Status='live', Symbol='CG31918', Aliases='Dmel_ CG18592, Dmel\\CG31918, anon-WO0118547.477', description='CG31918 gene product from transcript CG31918-RA', other_designations='CG31918-PA|uncharacterized protein', map_location='25C3-25C3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=5055982, end_position_on_the_genomic_accession=5058248, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31261, CurrentID=0, Status='live', Symbol='CG2701', Aliases='Dmel_ Dmel\\CG2701, EG:95B7.9, II, anon-3Bd', description='CG2701 gene product from transcript CG2701-RA', other_designations='CG2701-PA|coding region II|transcript II', map_location='3B3-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2717638, end_position_on_the_genomic_accession=2720327, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31262, CurrentID=0, Status='live', Symbol='fs(1)Yb', Aliases='Dmel_CG2706, CG2706, Dmel\\CG2706, EG:95B7.8, YB, Yb, fs(1)M104, fs(1)Y[b], nov', description='female sterile (1) Yb', other_designations='CG2706-PA|female sterile(1)M104|female sterile(1)Yb|fs(1)Yb-PA|no ovaries', map_location='1-1.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2720468, end_position_on_the_genomic_accession=2724158, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31263, CurrentID=0, Status='live', Symbol='fs(1)Ya', Aliases='Dmel_CG2707, CG2707, Dmel\\CG2707, EG:95B7.4, YA, Ya, fs(1)73, fs(1)M12, fs(1)Y2, fs(1)Y[a]', description='female sterile (1) Young arrest', other_designations='CG2707-PA|female sterile (1) young arrest|fs(1)Ya-PA|young arrest', map_location='3B3-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2724286, end_position_on_the_genomic_accession=2726577, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326176, CurrentID=0, Status='live', Symbol='CG31933', Aliases='Dmel_ CG15622, Dmel\\CG31933, gi21064045, gi22945479', description='CG31933 gene product from transcript CG31933-RA', other_designations='CG31933-PA|uncharacterized protein', map_location='22B3-22B3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1821256, end_position_on_the_genomic_accession=1823331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31265, CurrentID=0, Status='live', Symbol='dwg', Aliases='Dmel_CG2711, CG2711, CG2714, DWG, Dmel\\CG2711, EG:95B7.6, SBP, Sbp, ZW-5, ZW5, Zw-5, Zw5, l(1)3Be, l(1)zw5, ves, zw-5, zw5', description='deformed wings', other_designations='CG2711-PA|deformed wing|dwg-PA|scs binding protein|scs-binding protein|vestigium|zeste-white 5|zeste-white5', map_location='1-1.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2727698, end_position_on_the_genomic_accession=2730042, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31266, CurrentID=0, Status='live', Symbol='ttm50', Aliases='Dmel_CG2713, 17944223, CG2713, Dmel\\CG2713, EG:95B7.3, Gp99, i238, i90', description='tiny tim 50', other_designations='CG2713-PA|CG2713-PB|Gp99|ttm50-PA|ttm50-PB', map_location='3B3-3B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2730077, end_position_on_the_genomic_accession=2732190, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31267, CurrentID=0, Status='live', Symbol='CG2712', Aliases='Dmel_ Dmel\\CG2712, EG:95B7.7, III, anon-3Be', description='CG2712 gene product from transcript CG2712-RA', other_designations='CG2712-PA|coding region III', map_location='3B3-3B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2732469, end_position_on_the_genomic_accession=2734402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31268, CurrentID=0, Status='live', Symbol='crm', Aliases='Dmel_CG2714, CG2714, Dmel\\CG2714, EG:95B7.2, ZW-9, l(1)3C1, l(1)3Ca, l(1)zw9, sa, sta[P], swa, zw-9, zw9', description='cramped', other_designations='CG2714-PA|CG2714-PB|crm-PA|crm-PB|sparse-arista|stubarista from P32|swollen-antenna', map_location='1-1.48 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2734248, end_position_on_the_genomic_accession=2738773, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31269, CurrentID=0, Status='live', Symbol='Syx4', Aliases='Dmel_CG2715, CG2715, Dm Syx4, DmSyx4, Dmel\\CG2715, EG:95B7.1, EG:95B7.11, Syx, Syx-4, syx4', description='Syntaxin 4', other_designations='CG2715-PA|CG2715-PB|Syx4-PA|Syx4-PB|syntaxin|syntaxin 4', map_location='3B4-3B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2739089, end_position_on_the_genomic_accession=2744076, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31270 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31270, CurrentID=0, Status='live', Symbol='CG32795', Aliases='Dmel_ CG2716, CG2766, Dmel\\CG32795, EG:BACN33B1.2', description='CG32795 gene product from transcript CG32795-RD', other_designations='CG32795-PA|CG32795-PB|CG32795-PC|CG32795-PD|CG32795-PE|uncharacterized protein', map_location='3B6-3B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2782906, end_position_on_the_genomic_accession=2790567, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31271, CurrentID=0, Status='live', Symbol='w', Aliases='Dmel_CG2759, BACN33B1.1, CG2759, DMWHITE, Dmel\\CG2759, EG:BACN33B1.1, Hid, W, c23, e(g), m(g), mw(AT)[[13]], w', description='white', other_designations='CG2759-PA|enhancer of garnet|mini-white|modifier of garnet|w-PA|white[+]', map_location='1-1.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2790599, end_position_on_the_genomic_accession=2796466, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326184 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326184, CurrentID=0, Status='live', Symbol='CG32052', Aliases='Dmel_ BEST:HL01355, CG8163, Dmel\\CG32052, Smase, aSMase', description='CG32052 gene product from transcript CG32052-RE', other_designations='CG32052-PC|CG32052-PD|CG32052-PE|uncharacterized protein', map_location='67C8-67C9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=9787046, end_position_on_the_genomic_accession=9808713, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  326185 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326185, CurrentID=0, Status='live', Symbol='CG32054', Aliases='Dmel_ BcDNA:GM13904, CG6645, Dmel\\CG32054', description='CG32054 gene product from transcript CG32054-RA', other_designations='CG32054-PA|uncharacterized protein', map_location='67D8-67D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10092768, end_position_on_the_genomic_accession=10094738, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326186 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326186, CurrentID=0, Status='live', Symbol='scramb1', Aliases='Dmel_CG32056, CG32056, CG6590, CT20492, Dmel\\CG32056, Scramb1', description='scramblase 1', other_designations='CG32056-PA|CG32056-PB|CG32056-PC|scramb1-PA|scramb1-PB|scramb1-PC|scramblase', map_location='67D11-67D11', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10220483, end_position_on_the_genomic_accession=10225240, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326187, CurrentID=0, Status='live', Symbol='S-Lap3', Aliases='Dmel_CG32063, BEST:LP10167, CG32063, CG8040, Dmel\\CG32063, S-LAP 3, anon-WO0140519.83, lincRNA.S4635', description='Sperm-Leucylaminopeptidase 3', other_designations='CG32063-PB|CG32063-PC|CG32063-PD|S-Lap3-PB|S-Lap3-PC|S-Lap3-PD|Sperm-LeucylAminoPeptidase 3', map_location='67E3-67E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10465187, end_position_on_the_genomic_accession=10468974, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326188, CurrentID=0, Status='live', Symbol='S-Lap4', Aliases='Dmel_CG32064, CG32064, CG8040, Dmel\\CG32064, GH8, S-LAP 4, anon-WO0140519.141, anon-WO0140519.223, bs04d09.y1', description='Sperm-Leucylaminopeptidase 4', other_designations='CG32064-PA|S-Lap4-PA', map_location='67E4-67E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10490861, end_position_on_the_genomic_accession=10492934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326189, CurrentID=0, Status='live', Symbol='Adi1', Aliases='Dmel_CG32068, CG32068, CG8000, Dadi1, Dmel\\CG32068, adi1', description='Acireductone dioxygenase 1', other_designations='Adi1-PB|CG32068-PB|aci-reductone dioxygenase 1|acireductone dioxygenase 1', map_location='67E6-67E6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=10663376, end_position_on_the_genomic_accession=10664435, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31278, CurrentID=0, Status='live', Symbol='CG3588', Aliases='Dmel_ Dmel\\CG3588, EG:100G7.6', description='CG3588 gene product from transcript CG3588-RH', other_designations='CG3588-PE|CG3588-PF|CG3588-PG|CG3588-PH|CG3588-PI', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2919770, end_position_on_the_genomic_accession=2924426, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31279, CurrentID=0, Status='live', Symbol='CG14424', Aliases='Dmel_ Dmel\\CG14424, EG:100G7.5', description='CG14424 gene product from transcript CG14424-RA', other_designations='CG14424-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2925884, end_position_on_the_genomic_accession=2927258, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326192 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326192, CurrentID=0, Status='live', Symbol='CG32075', Aliases='Dmel_ CG6316, Dmel\\CG32075', description='CG32075 gene product from transcript CG32075-RA', other_designations='CG32075-PA|CG32075-PB|uncharacterized protein', map_location='68A3-68A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11062726, end_position_on_the_genomic_accession=11064941, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  326193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326193, CurrentID=0, Status='live', Symbol='Alg10', Aliases='Dmel_CG32076, CG32076, CG7624, Dmel\\CG32076, alg10, dAlg10', description='Alpha 3 glucosyltransferase', other_designations='Alg10-PA|CG32076-PA|alpha 3 glucosyltransferase', map_location='68A9-68A9', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11164971, end_position_on_the_genomic_accession=11166620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326194, CurrentID=0, Status='live', Symbol='CG32081', Aliases='Dmel_ BcDNA:AT27573, CG11806, Dmel\\CG32081, FBgn 52081', description='CG32081 gene product from transcript CG32081-RA', other_designations='CG32081-PA|uncharacterized protein', map_location='68A1-68A2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11027579, end_position_on_the_genomic_accession=11029444, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  326195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326195, CurrentID=0, Status='live', Symbol='CG32086', Aliases='Dmel_ CG6100, Dmel\\CG32086', description='CG32086 gene product from transcript CG32086-RB', other_designations='CG32086-PA|CG32086-PB|uncharacterized protein', map_location='68C13-68C13', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11582401, end_position_on_the_genomic_accession=11584064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326196, CurrentID=0, Status='live', Symbol='sunn', Aliases='Dmel_CG32088, CG32088, CG7308, Dmel\\CG32088', description='sisters unbound', other_designations='CG32088-PB|sunn-PB', map_location='68D3-68D3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=11728348, end_position_on_the_genomic_accession=11732016, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  326197 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326197, CurrentID=0, Status='live', Symbol='Ndfip', Aliases='Dmel_CG32177, CG32177, CG8056, Dmel\\CG32177, anon-EST:Liang-2.19, clone 2.19, dNdfip', description='Nedd4 family interacting protein', other_designations='CG32177-PA|Ndfip-PA', map_location='74E4-74E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17647442, end_position_on_the_genomic_accession=17654280, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326198, CurrentID=0, Status='live', Symbol='Krn', Aliases='Dmel_CG32179, CG32179, CG8056, CT19073, Dmel\\CG32179, Ker, gritz, keren, keren/spitz2, krn, s-krn, skrn', description='Keren', other_designations='CG32179-PA|Krn-PA|keren|spitz-2|spitz2', map_location='74E4-74E4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17649562, end_position_on_the_genomic_accession=17654280, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31287 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31287, CurrentID=0, Status='live', Symbol='CG14423', Aliases='Dmel_ Dmel\\CG14423', description='CG14423 gene product from transcript CG14423-RA', other_designations='CG14423-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2945403, end_position_on_the_genomic_accession=2946589, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31288, CurrentID=0, Status='live', Symbol='CG17959', Aliases='Dmel_ Dmel\\CG17959', description='CG17959 gene product from transcript CG17959-RA', other_designations='CG17959-PA|uncharacterized protein', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2947094, end_position_on_the_genomic_accession=2947580, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31289 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31289, CurrentID=0, Status='live', Symbol='CG3603', Aliases='Dmel_ BcDNA:RH09070, Dmel\\CG3603', description='CG3603 gene product from transcript CG3603-RB', other_designations='CG3603-PA|CG3603-PB', map_location='3C3-3C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2948034, end_position_on_the_genomic_accession=2949120, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31290, CurrentID=0, Status='live', Symbol='rst', Aliases='Dmel_CG4125, CG4125, CT13684, Dmel\\CG4125, IREE-C, IRREC/RST, Irre-C, IrreC, IrreC-Rst, IrreC-rst, IrreC/Rst, IrrecC, Rst, Rst/IrreC, UB883, irreC, irreC-rst, irrecC-rst-irreC, rst-irrecC, rst/irrecC, rst', description='roughest', other_designations='C-Roughest|CG4125-PA|CG4125-PB|ChiasmC-roughest|Irregular-chiasm-C|irreC-roughest|irregular chiasm C-roughest|irregular chiasm C-roughest protein|irregular chiasm-C-roughest|irregular chiasmC-roughest|irregular optic chiasma C|irregular optic chiasma C/Roughest|rst-PA|rst-PB', map_location='1-2.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2950399, end_position_on_the_genomic_accession=2973834, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  326203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326203, CurrentID=0, Status='live', Symbol='CG32238', Aliases='Dmel_ CG7474, Dmel\\CG32238', description='CG32238 gene product from transcript CG32238-RA', other_designations='CG32238-PA|uncharacterized protein', map_location='64B12-64B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=4648276, end_position_on_the_genomic_accession=4650305, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31292, CurrentID=0, Status='live', Symbol='kirre', Aliases='Dmel_CG3653, CG3653, CT12279, DUF, DUF/KIRRE, Dmel\\CG3653, Duf, Duf/Kirre, EG:163A10.1, Kirre, RH09239, duf, duf/kirre', description='kin of irre', other_designations='CG3653-PA|CG3653-PB|CG3653-PC|CG3653-PD|CG3653-PE|CG3653-PF|CG3653-PG|Dumbfounded/Kirre|Kin of irre|Kin of irregular-chiasm-C|Kin-of-irre|dumbfounded|kin of irreC|kin-of-irreC|kirre-PA|kirre-PB|kirre-PC|kirre-PD|kirre-PE|kirre-PF|kirre-PG', map_location='3C3-3C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=2740384, end_position_on_the_genomic_accession=3134532, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  31293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31293, CurrentID=0, Status='live', Symbol='N', Aliases='Dmel_CG3936, 1.1, 16-178, 16-55, Ax, CG3936, CT13012, Chp, Co, Dmel\\CG3936, EG:140G11.1, EG:163A10.2ECD, NICD, RICN, anon-EST:Liang-1.12, clone 1.12, co, dNotch, fa, l(1)3Cb, l(1)Ax, l(1)N, n[fah], nd, notch, shd, spl, swb, N', description='Notch', other_designations='CG3936-PA|CG3936-PB|N-PA|N-PB|NICD|abruptex|chopped|confluens|dNotch|facet|notch|notchoid|split|strawberry', map_location='1-3.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3134870, end_position_on_the_genomic_accession=3172221, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31294, CurrentID=0, Status='live', Symbol='Fcp3C', Aliases='Dmel_CG4015, CG4015, Dmel\\CG4015, EG:140G11.2, Ecp3C, VMP3C', description='Follicle cell protein 3C', other_designations='CG4015-PA|Fcp3C-PA|follicle cell protein 3C', map_location='3C9-3C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3173376, end_position_on_the_genomic_accession=3174474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326207, CurrentID=0, Status='live', Symbol='Drsl1', Aliases='Dmel_CG32274, CG11520, CG32274, Dmel\\CG32274, Dro-1, Dro-H, Dro1, Drs-I, Drs-l, Drs-lC, dro1b', description='Drosomycin-like 1', other_designations='CG32274-PA|Drosomycin-like|Drosomycin-like C|Drs-like|Drsl1-PA|drosomycin|drosomycin-1|drosomycin-1b|drosomycin-H', map_location='63D1-63D1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3335579, end_position_on_the_genomic_accession=3335788, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326208, CurrentID=0, Status='live', Symbol='CG32344', Aliases='Dmel_ CG6994, DmRH20, Dmel\\CG32344, cg6994', description='CG32344 gene product from transcript CG32344-RA', other_designations='CG32344-PA|uncharacterized protein', map_location='61C1-61C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=321233, end_position_on_the_genomic_accession=324720, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326209, CurrentID=0, Status='live', Symbol='S-Lap2', Aliases='Dmel_CG32351, BcDNA:LP03301, CG32351, Dmel\\CG32351, Lap CG32351, S-LAP 2, anon-WO0140519.85', description='Sperm-Leucylaminopeptidase 2', other_designations='CG32351-PA|S-Lap2-PA', map_location='66D8-66D8', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=8619888, end_position_on_the_genomic_accession=8622091, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31298, CurrentID=0, Status='live', Symbol='ng2', Aliases='Dmel_CG14266, BcDNA:RE05239, CG14266, Dmel\\CG14266, EG:96G10.4, ng, ng-2', description='new glue 2', other_designations='CG14266-PA|ng2-PA', map_location='3C10-3C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3240026, end_position_on_the_genomic_accession=3240571, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31299, CurrentID=0, Status='live', Symbol='ng3', Aliases='Dmel_CG10788, BcDNA:LP03924, CG10788, Dmel\\CG10788, EG:96G10.3, ng, ng-3', description='new glue 3', other_designations='CG10788-PB|ng3-PB', map_location='3C10-3C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3241491, end_position_on_the_genomic_accession=3242084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31300, CurrentID=0, Status='live', Symbol='ng1', Aliases='Dmel_CG10781, CG10781, Dmel\\CG10781, EG:96G10.5, SC1, ng, ng-1, ngp', description='new glue 1', other_designations='CG10781-PA|nested|new glue|new glue protein|ng1-PA', map_location='3C10-3C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3242412, end_position_on_the_genomic_accession=3242934, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31301, CurrentID=0, Status='live', Symbol='ng4', Aliases='Dmel_CG10789, CG10789, Dmel\\CG10789, EG:96G10.2, ng, ng-4', description='new glue 4', other_designations='CG10789-PA|ng4-PA', map_location='3C10-3C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3243954, end_position_on_the_genomic_accession=3244220, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326214, CurrentID=0, Status='live', Symbol='vito', Aliases='Dmel_CG32418, CG10569, CG32418, Dmel\\CG32418', description='viriato', other_designations='CG32418-PA|vito-PA', map_location='64E7-64E7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=5776324, end_position_on_the_genomic_accession=5777333, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31303, CurrentID=0, Status='live', Symbol='Pig1', Aliases='Dmel_CG10790, CG10790, Dmel\\CG10790, EG:96G10.1, Pig-1, gsg, pig1', description='Pre-intermoult gene 1', other_designations='CG10790-PB|Pig1-PB|Pre-intermolt gene 1|gland specific gene', map_location='3C10-3C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3248600, end_position_on_the_genomic_accession=3249289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31304 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31304, CurrentID=0, Status='live', Symbol='Sgs4', Aliases='Dmel_CG12181, CG12181, Dmel\\CG12181, EG:96G10.6, SGS-4, SGS4, Sgs, Sgs-4, sgs-4, sgs4', description='Salivary gland secretion 4', other_designations='CG12181-PC|Sgs4-PC|salivary gland secretion 4|salivary gland secretion protein 4', map_location='1-3.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3250089, end_position_on_the_genomic_accession=3251040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326199, CurrentID=0, Status='live', Symbol='CG32202', Aliases='Dmel_ CG13381, Dmel\\CG32202', description='CG32202 gene product from transcript CG32202-RA', other_designations='CG32202-PA|uncharacterized protein', map_location='75D6-75D6', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=18682184, end_position_on_the_genomic_accession=18682860, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326219 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326219, CurrentID=0, Status='live', Symbol='Sk2', Aliases='Dmel_CG32484, CG2159, CG32484, Dmel\\CG32484, SK2, Sphk2, sk2', description='Sphingosine kinase 2', other_designations='CG32484-PA|Sk2-PA|sphingosine kinase|sphingosine kinase 2|sphingosine kinase II', map_location='63A3-63A4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3047087, end_position_on_the_genomic_accession=3050159, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31308, CurrentID=0, Status='live', Symbol='CG10793', Aliases='Dmel_ Dmel\\CG10793', description='CG10793 gene product from transcript CG10793-RA', other_designations='CG10793-PA|uncharacterized protein', map_location='3D1-3D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3314744, end_position_on_the_genomic_accession=3316402, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31309, CurrentID=0, Status='live', Symbol='dnc', Aliases='Dmel_CG32498, CG10791, CG10792, CG10797, CG14267, CG14268, CG32498, Dmel\\CG32498, EG:140G11.4, EG:96G10.7, EG:BACN05I09.3, EG:BACN5I9.2, PDE4, PDEII, fs(1)M42', description='dunce', other_designations='CG32498-PA|CG32498-PB|CG32498-PC|CG32498-PD|CG32498-PE|CG32498-PF|CG32498-PG|CG32498-PI|CG32498-PJ|CG32498-PL|CG32498-PN|CG32498-PP|CG32498-PQ|CG32498-PR|CG32498-PS|CG32498-PT|CG32498-PU|dnc-PA|dnc-PB|dnc-PC|dnc-PD|dnc-PE|dnc-PF|dnc-PG|dnc-PI|dnc-PJ|dnc-PL|dnc-PN|dnc-PP|dnc-PQ|dnc-PR|dnc-PS|dnc-PT|dnc-PU', map_location='1-3.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3176440, end_position_on_the_genomic_accession=3343767, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  31310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31310, CurrentID=0, Status='live', Symbol='Myc', Aliases='Dmel_CG10798, CG10798, D- DM, DMYc, Dm, Dmel\\CG10798, Dmyc, EG:BACN5I9.1, MYC, anon-WO03040301.171, bHLHe57, c-MYC, c-Myc, c-myc, d-myc, dMYC, dMyc, dMyc1, dm, dm/dMyc, dm/myc, dmyc, dmyc1, l(1)G0139, l(1)G0354, l(1)G0359, myc', description='CG10798 gene product from transcript CG10798-RB', other_designations='CG10798-PA|CG10798-PB|Myc-PA|Myc-PB|diminuitive|diminutive|lethal (1) G0354|lethal (1) G0359', map_location='1-4.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3373159, end_position_on_the_genomic_accession=3391697, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326223 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326223, CurrentID=0, Status='live', Symbol='Twdlalpha', Aliases='Dmel_CG32574, CG32574, CG4724, Dmel\\CG32574', description='Tweedlealpha', other_designations='CG32574-PA|Twdlalpha-PA|tweedlealpha', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16742847, end_position_on_the_genomic_accession=16744365, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326224, CurrentID=0, Status='live', Symbol='betaNACtes3', Aliases='Dmel_CG32601, CG32601, CG9404, Dmel\\CG32601', description='CG32601 gene product from transcript CG32601-RA', other_designations='CG32601-PA|betaNACtes3-PA', map_location='12E9-12E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14378806, end_position_on_the_genomic_accession=14379826, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326225, CurrentID=0, Status='live', Symbol='CG32633', Aliases='Dmel_ CG1733, Dmel\\CG32633, anon-WO0140519.164', description='CG32633 gene product from transcript CG32633-RA', other_designations='CG32633-PA|CG32633-PB|uncharacterized protein', map_location='12A4-12A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13561671, end_position_on_the_genomic_accession=13563359, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  326226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326226, CurrentID=0, Status='live', Symbol='CG32647', Aliases='Dmel_ BcDNA:RH40632, CG15723, CG18453, CG3794, CT35953, Dmel\\CG32647', description='CG32647 gene product from transcript CG32647-RD', other_designations='CG32647-PA|CG32647-PB|CG32647-PC|CG32647-PD|uncharacterized protein', map_location='11C1-11C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12771844, end_position_on_the_genomic_accession=12784958, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31315, CurrentID=0, Status='live', Symbol='CG16781', Aliases='Dmel_ Dmel\\CG16781', description='CG16781 gene product from transcript CG16781-RA', other_designations='CG16781-PA|CG16781-PB|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3462474, end_position_on_the_genomic_accession=3464044, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31316, CurrentID=0, Status='live', Symbol='CG12206', Aliases='Dmel_ Dmel\\CG12206', description='CG12206 gene product from transcript CG12206-RA', other_designations='CG12206-PA|CG12206-PB|CG12206-PC|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3465737, end_position_on_the_genomic_accession=3474524, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  260693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=260693, CurrentID=0, Status='live', Symbol='Or19b', Aliases='Dmel_CG32825, CG32825, Dmel\\CG32825', description='Odorant receptor 19b', other_designations='CG32825-PA|Or19b-PA|odorant receptor 19b', map_location='19B1-19B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20094398, end_position_on_the_genomic_accession=20095767, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326230, CurrentID=0, Status='live', Symbol='Rab9Fa', Aliases='Dmel_CG32671, CG2532, CG32671, DmRabX3, Dmel\\CG32671, RabX3', description='Rab at 9Fa', other_designations='CG32671-PA|Rab GTPase 9Fa|Rab9Fa-PA', map_location='9F13-9F13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10950160, end_position_on_the_genomic_accession=10951030, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31319, CurrentID=0, Status='live', Symbol='CG14270', Aliases='Dmel_ Dmel\\CG14270', description='CG14270 gene product from transcript CG14270-RA', other_designations='CG14270-PA|uncharacterized protein', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3484605, end_position_on_the_genomic_accession=3485378, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326232 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326232, CurrentID=0, Status='live', Symbol='Tango5', Aliases='Dmel_CG32675, CG1534, CG32675, Dmel\\CG32675, TANGO5', description='Transport and Golgi organization 5', other_designations='CG32675-PA|CG32675-PB|CG32675-PC|CG32675-PD|CG32675-PE|CG32675-PF|Tango5-PA|Tango5-PB|Tango5-PC|Tango5-PD|Tango5-PE|Tango5-PF|transport and golgi organization 5', map_location='9E4-9E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10760048, end_position_on_the_genomic_accession=10764413, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31321, CurrentID=0, Status='live', Symbol='Gas8', Aliases='Dmel_CG14271, CG14271, Dmel\\CG14271', description='Growth arrest specific protein 8', other_designations='CG14271-PB|CG14271-PC|CG14271-PD|Gas8-PB|Gas8-PC|Gas8-PD|growth arrest specific protein 8', map_location='3D4-3D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3491011, end_position_on_the_genomic_accession=3504851, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326234 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326234, CurrentID=0, Status='live', Symbol='l(1)G0320', Aliases='Dmel_CG32701, CG2994, CG32701, Dmel\\CG32701', description='lethal (1) G0320', other_designations='CG32701-PA|CG32701-PB|l(1)G0320-PA|l(1)G0320-PB', map_location='8E7-8E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9551140, end_position_on_the_genomic_accession=9553750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  326235 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326235, CurrentID=0, Status='live', Symbol='Cubn', Aliases='Dmel_CG32702, CG2996, CG32702, CT7882, Dmel\\CG32702, dCubn, tok-like', description='Cubilin ortholog', other_designations='CG32702-PB|Cubn-PB|TMPRSS6 iron sensing|cubilin|cubilin ortholog|dCubilin', map_location='8E7-8E10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9558665, end_position_on_the_genomic_accession=9571308, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  326237 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326237, CurrentID=0, Status='live', Symbol='Ubi-p5E', Aliases='Dmel_CG32744, CG18282, CG32744, CR32744, DmUbi-p5E, Dmel\\CG32744, UB, Ub, Ubi, Ubiq', description='Ubiquitin-5E', other_designations='CG32744-PA|CG32744-PB|Ubi-p5E-PA|Ubi-p5E-PB|ubiquitin', map_location='5E4-5E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6276876, end_position_on_the_genomic_accession=6279647, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31326, CurrentID=0, Status='live', Symbol='CG2875', Aliases='Dmel_ Dmel\\CG2875', description='CG2875 gene product from transcript CG2875-RC', other_designations='CG2875-PA|CG2875-PB|CG2875-PC', map_location='3D6-3D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3571456, end_position_on_the_genomic_accession=3573622, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31328, CurrentID=0, Status='live', Symbol='Ilp7', Aliases='Dmel_CG13317, CG13317, DILP, DILP-7, DILP7, Dilp, Dilp 7, Dilp7, Dmel\\CG13317, ILP7, Ilp, Ilp-7, dILP7, dILP8, dilp, dilp-7, dilp7, ilp7', description='Insulin-like peptide 7', other_designations='CG13317-PA|Ilp7-PA|drosophila insulin-like peptide 7|insulin-like peptide', map_location='3E2-3E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3667621, end_position_on_the_genomic_accession=3668845, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31329, CurrentID=0, Status='live', Symbol='Parg', Aliases='Dmel_CG2864, CG2864, Dmel\\CG2864, EG:114E2.1, HD10914, PARG, parg', description='Poly(ADP-ribose) glycohydrolase', other_designations='CG2864-PB|CG2864-PC|Parg-PB|Parg-PC|Poly(ADP-ribose)glycohydrolase|Poly(ADPribose) glyco-hydrolase', map_location='3E2-3E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3668837, end_position_on_the_genomic_accession=3671502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31331 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31331, CurrentID=0, Status='live', Symbol='Mnt', Aliases='Dmel_CG13316, CG13316, CG2856, Dm Dmel\\CG13316, EG:114E2.2, MNT, Mad, Mad-3E, bHLHe43, dMad, dMnt, dmnt, mnt', description='CG13316 gene product from transcript CG13316-RD', other_designations='CG13316-PA|CG13316-PC|CG13316-PD|CG13316-PE|CG13316-PF|CG13316-PG|CG13316-PH|CG13316-PI|CG13316-PJ|CG13316-PK|Mnt-PA|Mnt-PC|Mnt-PD|Mnt-PE|Mnt-PF|Mnt-PG|Mnt-PH|Mnt-PI|Mnt-PJ|Mnt-PK', map_location='3E4-3E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3671801, end_position_on_the_genomic_accession=3701672, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31332 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31332, CurrentID=0, Status='live', Symbol='Rala', Aliases='Dmel_CG2849, 24639550, CG2849, D-RalA, DRal, DRala, Dm RalA, Dmel\\CG2849, Dral, DralA, Drala, P48555, RALA, Ral, Ral-a, RalA, anon-EST:fe1H4, dRal, l(1)G0174, l(1)G0373, l(1)G0501, ral, ralA', description='Ras-like protein A', other_designations='CG2849-PA|CG2849-PB|CG2849-PC|Rala-PA|Rala-PB|Rala-PC|Ras-related protein|anon-fast-evolving-1H4', map_location='3E5-3E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3704610, end_position_on_the_genomic_accession=3719635, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  326245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326245, CurrentID=0, Status='live', Symbol='CG33052', Aliases='Dmel_ CG3882, Dmel\\CG33052', description='CG33052 gene product from transcript CG33052-RA', other_designations='CG33052-PA|uncharacterized protein', map_location='74B4-74B5', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17427305, end_position_on_the_genomic_accession=17428836, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  326246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326246, CurrentID=0, Status='live', Symbol='CG33056', Aliases='Dmel_ CG10517, Dmel\\CG33056', description='CG33056 gene product from transcript CG33056-RA', other_designations='CG33056-PA|CG33056-PB|CG33056-PC|CG33056-PD|CG33056-PE|uncharacterized protein', map_location='78C2-78C2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21198261, end_position_on_the_genomic_accession=21203232, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31335, CurrentID=0, Status='live', Symbol='HIP-R', Aliases='Dmel_CG2947, CG2947, Dmel\\CG2947, Hip, anon-EST:fe1A9', description='Hsc/Hsp70-interacting protein related', other_designations='CG2947-PA|CG2947-PB|CG2947-PD|HIP-R-PA|HIP-R-PB|HIP-R-PD|anon-fast-evolving-1A9', map_location='3F1-3F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3789969, end_position_on_the_genomic_accession=3792477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31336 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31336, CurrentID=0, Status='live', Symbol='fd3F', Aliases='Dmel_CG44123, CG12632, CG15238, CG18771, CG32779, CG44123, Dmel\\CG44123, Dmel_CG12632, Dmel_CG32779, Fd3F, fd3f', description='forkhead domain 3F', other_designations='CG44123-PC|fd3F-PC', map_location='3F2-3F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3802316, end_position_on_the_genomic_accession=3811040, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31338, CurrentID=0, Status='live', Symbol='CG2901', Aliases='Dmel_ Dmel\\CG2901', description='CG2901 gene product from transcript CG2901-RA', other_designations='CG2901-PA', map_location='3F2-3F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3828197, end_position_on_the_genomic_accession=3831546, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31339, CurrentID=0, Status='live', Symbol='ec', Aliases='Dmel_CG2904, CG2904, Dmel\\CG2904, EG:EG0002.3, Ec, Q1W9Q0', description='echinus', other_designations='CG2904-PA|CG2904-PB|CG2904-PC|ec-PA|ec-PB|ec-PC', map_location='1-5.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3811698, end_position_on_the_genomic_accession=3857621, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31340, CurrentID=0, Status='live', Symbol='yin', Aliases='Dmel_CG44402, CG2913, CG34336, CG44402, Dmel\\CG44402, Dmel_CG2913, Dmel_CG34336, EG:EG0002.1, OPT1, OpT1, Yin, opt, opt1', description='CG44402 gene product from transcript CG44402-RD', other_designations='CG44402-PA|CG44402-PB|CG44402-PC|CG44402-PD|oligopeptide transporter 1|yin-PA|yin-PB|yin-PC|yin-PD', map_location='3F3-3F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3862966, end_position_on_the_genomic_accession=3871939, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31341, CurrentID=0, Status='live', Symbol='CG2930', Aliases='Dmel_ BcDNA:GH06717, CG32780, CG32787, Dmel\\CG2930', description='CG2930 gene product from transcript CG2930-RF', other_designations='CG2930-PA|CG2930-PB|CG2930-PD|CG2930-PE|CG2930-PF', map_location='3F4-3F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3868973, end_position_on_the_genomic_accession=3880379, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  31342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31342, CurrentID=0, Status='live', Symbol='VhaAC39-1', Aliases='Dmel_CG2934, ATP6V0D1, CG2934, Dmel\\CG2934, VhaAC39, Vhaac39, anon-EST:GressD5, vhaAC39, vhaAC39-1', description='Vacuolar H[+] ATPase AC39 subunit 1', other_designations='CG2934-PA|V-ATPase V0 d subunit|VhaAC39|VhaAC39-1-PA|vacuolar ATPsintase subunit 1|vacuolar H[+] ATPase AC39 subunit 1|vacuolar H[+] ATPase subunit AC39-1|vacuolar H[+]-ATPase C39 subunit', map_location='3F4-3F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3880748, end_position_on_the_genomic_accession=3882967, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  326255 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326255, CurrentID=0, Status='live', Symbol='4EHP', Aliases='Dmel_CG33100, 4E-HP, CG10716, CG33100, CT30033, Dmel\\CG33100, d4EHP, eIF4E-8', description='eIF4E-Homologous Protein', other_designations='4EHP-PA|4EHP-PB|4EHP-PC|4EHP-PD|CG33100-PA|CG33100-PB|CG33100-PC|CG33100-PD|eIF4E-Homologous protein', map_location='95E1-95E1', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=24063077, end_position_on_the_genomic_accession=24108960, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31344, CurrentID=0, Status='live', Symbol='CG42541', Aliases='Dmel_ BcDNA:RH43809, CG15240, CG16828, CG16829, CG32776, Dmel\\CG42541, Dmel_CG15240, Dmel_CG32776', description='CG42541 gene product from transcript CG42541-RE', other_designations='CG42541-PC|CG42541-PD|CG42541-PE|CG42541-PF|uncharacterized protein', map_location='3F6-3F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3891497, end_position_on_the_genomic_accession=3930578, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  326257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326257, CurrentID=0, Status='live', Symbol='CG33108', Aliases='Dmel_ BcDNA:RE26840, C19orf54Nt, CG5926, Dmel\\CG33108', description='CG33108 gene product from transcript CG33108-RA', other_designations='CG33108-PA|CG33108-PB|uncharacterized protein', map_location='95D5-95D5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=23995659, end_position_on_the_genomic_accession=23997544, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326258, CurrentID=0, Status='live', Symbol='CG33109', Aliases='Dmel_ CG9645, Dmel\\CG33109, cg9645', description='CG33109 gene product from transcript CG33109-RA', other_designations='CG33109-PA|uncharacterized protein', map_location='88A12-88A12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=14220560, end_position_on_the_genomic_accession=14221584, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326259, CurrentID=0, Status='live', Symbol='CG33116', Aliases='Dmel_ CG10355, Dmel\\CG33116, dCCS1', description='CG33116 gene product from transcript CG33116-RA', other_designations='CG33116-PA|uncharacterized protein', map_location='37F1-37F1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=19533479, end_position_on_the_genomic_accession=19537363, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31348, CurrentID=0, Status='live', Symbol='CG2941', Aliases='Dmel_ Dmel\\CG2941', description='CG2941 gene product from transcript CG2941-RB', other_designations='CG2941-PA|CG2941-PB', map_location='3F7-3F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3936517, end_position_on_the_genomic_accession=3940582, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31349, CurrentID=0, Status='live', Symbol='Vap-33A', Aliases='Dmel_CG5014, BcDNA:GM03307, CG5014, DVAP, DVAP-33A, DVAP33A, Dmel\\CG5014, VAP, VAP-33, VAP-33-1, VAP33-A, VAPB, Vap-33-1, Vap33-1, anon-EST:Liang-2.42, anon-WO03040301.124, clone 2.42, dVAP, dVAP33A, dvap33a, l(1)G0231, vap-33A, vap33-1', description='VAMP-associated protein of 33kDa ortholog A', other_designations='CG5014-PA|CG5014-PB|CG5014-PC|CG5014-PE|VAMP-associated protein of 33kDa A|Vap-33-1|Vap-33A-PA|Vap-33A-PB|Vap-33A-PC|Vap-33A-PE|Vap33-1', map_location='3F9-3F9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=3948975, end_position_on_the_genomic_accession=3955442, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  326262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326262, CurrentID=0, Status='live', Symbol='CG33123', Aliases='Dmel_ CG3229, Dmel\\CG33123, LeuRS, anon-EST:Posey266', description='CG33123 gene product from transcript CG33123-RB', other_designations='CG33123-PA|CG33123-PB|uncharacterized protein', map_location='23F6-23F6', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3510328, end_position_on_the_genomic_accession=3514656, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  326263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326263, CurrentID=0, Status='live', Symbol='CG33128', Aliases='Dmel_ CG10872, Dmel\\CG33128', description='CG33128 gene product from transcript CG33128-RA', other_designations='CG33128-PA|aspartic-type endopeptidase|uncharacterized protein', map_location='22A2-22A2', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=1494070, end_position_on_the_genomic_accession=1495456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326264, CurrentID=0, Status='live', Symbol='AGBE', Aliases='Dmel_CG33138, CG33138, CG4023, Dmel\\CG33138', description='1,4-Alpha-Glucan Branching Enzyme', other_designations='1,4-Alpha-Glucan branching enzyme|AGBE-PA|CG33138-PA', map_location='49F4-49F4', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=13171401, end_position_on_the_genomic_accession=13174796, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326265 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326265, CurrentID=0, Status='live', Symbol='CG33160', Aliases='Dmel_ CG11531, Dmel\\CG33160, SP183', description='CG33160 gene product from transcript CG33160-RA', other_designations='CG33160-PA|uncharacterized protein', map_location='63B12-63B12', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=3205539, end_position_on_the_genomic_accession=3206363, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31355 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31355, CurrentID=0, Status='live', Symbol='CG6379', Aliases='Dmel_ Dmel\\CG6379', description='CG6379 gene product from transcript CG6379-RB', other_designations='CG6379-PA|CG6379-PB', map_location='4A5-4A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4097592, end_position_on_the_genomic_accession=4101031, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31356 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31356, CurrentID=0, Status='live', Symbol='CG15375', Aliases='Dmel_ Dmel\\CG15375, mdcds_59461', description='CG15375 gene product from transcript CG15375-RB', other_designations='CG15375-PB|uncharacterized protein', map_location='4A5-4A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4101318, end_position_on_the_genomic_accession=4102995, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31357, CurrentID=0, Status='live', Symbol='Xpac', Aliases='Dmel_CG6358, CG6358, DhR14, DhXPA, DmXPA, Dmel\\CG6358, Dxpa, EG:EG0007.8, XPAC, XPA_DROME, Xpa, dmXPA, dxpa', description='Xeroderma pigmentosum group A-like', other_designations='CG6358-PA|Xpac-PA|xeroderma pigmentosum group A-like', map_location='4A5-4A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4102955, end_position_on_the_genomic_accession=4104197, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31358, CurrentID=0, Status='live', Symbol='brn', Aliases='Dmel_CG4934, Brn, CG4934, Dmel\\CG4934, EG:EG0007.6, brc, fs(1)107, fs(1)A107, l(1)6P6', description='brainiac', other_designations='CG4934-PA|brn-PA|gene B', map_location='1-5.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4104342, end_position_on_the_genomic_accession=4105791, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31359, CurrentID=0, Status='live', Symbol='cib', Aliases='Dmel_CG4944, CG4944, Dmel\\CG4944, EG0007-11, EG:EG0007.11', description='ciboulot', other_designations='CG4944-PA|CG4944-PB|CG4944-PC|CG4944-PD|CG4944-PE|beta-thymosin|cib-PA|cib-PB|cib-PC|cib-PD|cib-PE|ciboulet|ciboulout', map_location='4A5-4A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4106113, end_position_on_the_genomic_accession=4110981, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  326273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326273, CurrentID=0, Status='live', Symbol='His2B:CG17949', Aliases='Dmel_CG17949, CG17949, Dmel\\CG17949', description='CG17949 gene product from transcript CG17949-RA', other_designations='CG17949-PA|His2B:CG17949-PA|histone H2B', map_location='39D3-39D3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=21423014, end_position_on_the_genomic_accession=21423608, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31362, CurrentID=0, Status='live', Symbol='Tip60', Aliases='Dmel_CG6121, 6121, CG6121, DmEG0007.7, Dmel/TIP60, Dmel\\CG6121, Dmel\\TIP60, EG0007.7, EG:EG0007.7, TIP-60, TIP60, dTIP60, dTip60, dmHAG405', description='Tat interactive protein 60kDa', other_designations='CG6121-PA|CG6121-PB|CG6121-PC|Tat-interactive protein|Tat-interactive protein-60 KDa|Tip60-PA|Tip60-PB|Tip60-PC', map_location='4A6-4B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4114800, end_position_on_the_genomic_accession=4117487, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31363, CurrentID=0, Status='live', Symbol='tyf', Aliases='Dmel_CG4857, CG485, CG4857, Dmel\\CG4857, EG:EG0007.12, EG:EG0007.4, TYF', description='twenty-four', other_designations='CG4857-PB|CG4857-PC|CG4857-PF|CG4857-PG|CG4857-PH|CG4857-PI|CG4857-PJ|CG4857-PK|CG4857-PL|CG4857-PM|CG4857-PN|TWENTY-FOUR|twenty four|tyf-PB|tyf-PC|tyf-PF|tyf-PG|tyf-PH|tyf-PI|tyf-PJ|tyf-PK|tyf-PL|tyf-PM|tyf-PN', map_location='4B1-4B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4117020, end_position_on_the_genomic_accession=4130016, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31364, CurrentID=0, Status='live', Symbol='Fas2', Aliases='Dmel_CG3665, 1D4, Ab 1D4, CG3665, CT12301, Dmel\\CG3665, EG:EG0007.3, FAS 11, FAS II, FAS2, FASII, Fas, Fas 2, Fas II, Fas-II, FasII, FascII, Fasii, MAB1D4, anon-EST:Liang-1.60, clone 1.60, fas II, fas-II, fas2, fasII, fasciclin II, l(1)G0032, l(1)G0048, l(1)G0081, l(1)G0293, l(1)G0336, mAB1D4, mAb 1D4, mAb1D4', description='Fasciclin 2', other_designations='CG3665-PA|CG3665-PB|CG3665-PC|CG3665-PD|CG3665-PF|CG3665-PG|CG3665-PH|FASCICLIN II|Fas|Fas2-PA|Fas2-PB|Fas2-PC|Fas2-PD|Fas2-PF|Fas2-PG|Fas2-PH|FasciclinII|fasII|fascicilin II|fasciclin|fasciclin 2|fasciclin-2|fasciclin-II|fasciclin2|fasciculin II|fascilin II|fasiclin II', map_location='4B1-4B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4132887, end_position_on_the_genomic_accession=4206093, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  31365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31365, CurrentID=0, Status='live', Symbol='CG15571', Aliases='Dmel_ Dmel\\CG15571', description='CG15571 gene product from transcript CG15571-RA', other_designations='CG15571-PA|uncharacterized protein', map_location='4B3-4B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4216873, end_position_on_the_genomic_accession=4217394, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31366, CurrentID=0, Status='live', Symbol='CG15570', Aliases='Dmel_ Dmel\\CG15570', description='CG15570 gene product from transcript CG15570-RA', other_designations='CG15570-PA|uncharacterized protein', map_location='4B3-4B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4218170, end_position_on_the_genomic_accession=4223331, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  31367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31367, CurrentID=0, Status='live', Symbol='Femcoat', Aliases='Dmel_CG15573, CG15573, Dmel\\CG15573, femcoat', description='CG15573 gene product from transcript CG15573-RB', other_designations='CG15573-PB|Femcoat-PB|femcoat', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4241887, end_position_on_the_genomic_accession=4242826, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31368, CurrentID=0, Status='live', Symbol='Muc4B', Aliases='Dmel_CG32774, CG15574, CG15575, CG32774, Dmel\\CG32774, muc4B', description='Mucin 4B', other_designations='CG32774-PB|CG32774-PC|Muc4B-PB|Muc4B-PC|mucin 4B', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4244095, end_position_on_the_genomic_accession=4246139, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31369 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31369, CurrentID=0, Status='live', Symbol='CG15576', Aliases='Dmel_ Dmel\\CG15576', description='CG15576 gene product from transcript CG15576-RB', other_designations='CG15576-PB|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4251870, end_position_on_the_genomic_accession=4252790, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31370 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31370, CurrentID=0, Status='live', Symbol='CG12691', Aliases='Dmel_ Dmel\\CG12691', description='CG12691 gene product from transcript CG12691-RA', other_designations='CG12691-PA|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4266097, end_position_on_the_genomic_accession=4266669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31371, CurrentID=0, Status='live', Symbol='CG15572', Aliases='Dmel_ Dmel\\CG15572', description='CG15572 gene product from transcript CG15572-RA', other_designations='CG15572-PA|uncharacterized protein', map_location='4B4-4B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4268701, end_position_on_the_genomic_accession=4271083, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31372 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31372, CurrentID=0, Status='live', Symbol='CG12692', Aliases='Dmel_ Dmel\\CG12692', description='CG12692 gene product from transcript CG12692-RC', other_designations='CG12692-PB|CG12692-PC|uncharacterized protein', map_location='4B5-4B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4300117, end_position_on_the_genomic_accession=4302346, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31373, CurrentID=0, Status='live', Symbol='mei-9', Aliases='Dmel_CG3697, CG3697, DhR1, DmXPF, Dmel\\CG3697, MEI-9, Mei-9, Rad1p, mei9, mus(1)110, mus110, xpf/mei-9', description='meiotic 9', other_designations='CG3697-PA|CG3697-PB|mei-9-PA|mei-9-PB|meiotic-9', map_location='1-6.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4314230, end_position_on_the_genomic_accession=4318801, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31374, CurrentID=0, Status='live', Symbol='CG2982', Aliases='Dmel_ Dmel\\CG2982, anon-4Bd', description='CG2982 gene product from transcript CG2982-RB', other_designations='CG2982-PB|CG2982-PC|anon-4Bd', map_location='4B6-4B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4319438, end_position_on_the_genomic_accession=4321745, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31375, CurrentID=0, Status='live', Symbol='CG12693', Aliases='Dmel_ Dmel\\CG12693', description='CG12693 gene product from transcript CG12693-RB', other_designations='CG12693-PB|uncharacterized protein', map_location='4B6-4B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4327080, end_position_on_the_genomic_accession=4329457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31376 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31376, CurrentID=0, Status='live', Symbol='norpA', Aliases='Dmel_CG3620, CG3620, CdkA, DIP2, Dmel\\CG3620, MRE18, NORPA, NorpA, PI-PLC, PLC, PLC-beta, PLCbeta, anon-EST:Posey221, nofC, rof, x12, x13, x16, x24', description='no receptor potential A', other_designations='CG3620-PA|CG3620-PB|CG3620-PC|CG3620-PD|CG3620-PE|CG3620-PF|No receptor potential A|eye-specific PLC|mRNA-like ncRNA in embryogenesis 18|no receptor potential|no receptor potentials A|no-receptor potential A|no-receptor-potential A|no-receptor-potential-A|norp A|norpA-PA|norpA-PB|norpA-PC|norpA-PD|norpA-PE|norpA-PF|phospholipase C|phospholipase C-beta|phospholipase C[[beta]]|phospholipase cbeta|phospholipase-C', map_location='1-6.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4322626, end_position_on_the_genomic_accession=4365408, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  31377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31377, CurrentID=0, Status='live', Symbol='CG3626', Aliases='Dmel_ Dmel\\CG3626', description='CG3626 gene product from transcript CG3626-RA', other_designations='CG3626-PA', map_location='4C2-4C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4366417, end_position_on_the_genomic_accession=4371136, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31378, CurrentID=0, Status='live', Symbol='CG12688', Aliases='Dmel_ Dmel\\CG12688', description='CG12688 gene product from transcript CG12688-RB', other_designations='CG12688-PB|uncharacterized protein', map_location='4C2-4C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4380758, end_position_on_the_genomic_accession=4382466, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31379, CurrentID=0, Status='live', Symbol='bi', Aliases='Dmel_CG3578, BI, CG3578, Dm-OMB, Dm-omb, Dmel\\CG3578, OMB, Omb, Qd, T3, TBX2, dm-omb, l(1)bi, l(1)omb, omb', description='bifid', other_designations='CG3578-PA|CG3578-PB|CG3578-PC|CG3578-PD|CG3578-PE|CG3578-PF|CG3578-PG|Optimotor-blind|bi-PA|bi-PB|bi-PC|bi-PD|bi-PE|bi-PF|bi-PG|lethal(l)optomotor-blind|optimotor blind|optomer-blind|optomoter blind|optomotor blind|optomotor-blind|optomotor-blind gene|optomotorblind|quadroon', map_location='4C3-4C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4412856, end_position_on_the_genomic_accession=4485647, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31380 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31380, CurrentID=0, Status='live', Symbol='CG3556', Aliases='Dmel_ Dmel\\CG3556', description='CG3556 gene product from transcript CG3556-RB', other_designations='CG3556-PA|CG3556-PB', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4530231, end_position_on_the_genomic_accession=4535464, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31381, CurrentID=0, Status='live', Symbol='rb', Aliases='Dmel_CG11427, AP-3beta, CG11427, Dmel\\CG11427, Rb, apl6, beta3, cg11427', description='ruby', other_designations='CG11427-PA|CG11427-PC|CG11427-PD|adaptor protein complex 3, beta subunit|rb-PA|rb-PC|rb-PD', map_location='1-7.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4535635, end_position_on_the_genomic_accession=4540626, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31382, CurrentID=0, Status='live', Symbol='CHOp24', Aliases='Dmel_CG3564, CG3564, Dmel\\CG3564, Emp24, emp24, p24, p24/p24beta1', description='CG3564 gene product from transcript CG3564-RB', other_designations='CG3564-PA|CG3564-PB|CHOp24-PA|CHOp24-PB', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4540908, end_position_on_the_genomic_accession=4542506, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740389, CurrentID=0, Status='live', Symbol='CG34200', Aliases='Dmel_ Dm_2R:4171, Dmel\\CG34200', description='CG34200 gene product from transcript CG34200-RA', other_designations='CG34200-PA|uncharacterized protein', map_location='41F9-41F9', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=5757662, end_position_on_the_genomic_accession=5758067, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31384, CurrentID=0, Status='live', Symbol='Usf', Aliases='Dmel_CG17592, CG17592, Dm  DmUsf, Dmel\\CG17592, bHLHb13, dUSF, dUSF2', description='CG17592 gene product from transcript CG17592-RC', other_designations='CG17592-PB|CG17592-PC|Usf-PB|Usf-PC', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4544617, end_position_on_the_genomic_accession=4548794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31385, CurrentID=0, Status='live', Symbol='CG15912', Aliases='Dmel_ 24639695, Dmel\\CG15912', description='CG15912 gene product from transcript CG15912-RA', other_designations='CG15912-PA|uncharacterized protein', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4549070, end_position_on_the_genomic_accession=4549957, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31386, CurrentID=0, Status='live', Symbol='CG11436', Aliases='Dmel_ Dmel\\CG11436', description='CG11436 gene product from transcript CG11436-RA', other_designations='CG11436-PA|uncharacterized protein', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4549883, end_position_on_the_genomic_accession=4551878, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31387, CurrentID=0, Status='live', Symbol='CG3527', Aliases='Dmel_ Dmel\\CG3527, dNEP1', description='CG3527 gene product from transcript CG3527-RA', other_designations='CG3527-PA', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4551973, end_position_on_the_genomic_accession=4553006, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31388 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31388, CurrentID=0, Status='live', Symbol='CG11444', Aliases='Dmel_ Dmel\\CG11444', description='CG11444 gene product from transcript CG11444-RB', other_designations='CG11444-PA|CG11444-PB|uncharacterized protein', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4553117, end_position_on_the_genomic_accession=4554398, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31389, CurrentID=0, Status='live', Symbol='CG3546', Aliases='Dmel_ Dmel\\CG3546', description='CG3546 gene product from transcript CG3546-RA', other_designations='CG3546-PA', map_location='4C7-4C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4558072, end_position_on_the_genomic_accession=4560317, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31390 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31390, CurrentID=0, Status='live', Symbol='CG12684', Aliases='Dmel_ Dmel\\CG12684', description='CG12684 gene product from transcript CG12684-RA', other_designations='CG12684-PA|uncharacterized protein', map_location='4C8-4C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4570991, end_position_on_the_genomic_accession=4571929, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31391 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31391, CurrentID=0, Status='live', Symbol='peb', Aliases='Dmel_CG12212, CG12212, Dmel\\CG12212, EG:66A1.1, EP55, HNT, Hin, Hind, Hnt, PEB, Peb, anon-4Cg, hnd, hnt', description='pebbled', other_designations='CG12212-PA|CG12212-PB|hindsight|peb-PA|peb-PB', map_location='1-7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4617564, end_position_on_the_genomic_accession=4627206, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31392, CurrentID=0, Status='live', Symbol='CG12184', Aliases='Dmel_ Dmel\\CG12184, EG:66A1.3, GH15731P', description='CG12184 gene product from transcript CG12184-RA', other_designations='CG12184-PA|CG12184-PB|CG12184-PC|uncharacterized protein', map_location='4C10-4C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4627380, end_position_on_the_genomic_accession=4630936, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31393, CurrentID=0, Status='live', Symbol='CG12179', Aliases='Dmel_ Dmel\\CG12179, EG:66A1.2', description='CG12179 gene product from transcript CG12179-RD', other_designations='CG12179-PC|CG12179-PD|uncharacterized protein', map_location='4C10-4C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4631052, end_position_on_the_genomic_accession=4636004, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31394, CurrentID=0, Status='live', Symbol='pon', Aliases='Dmel_CG3346, CG3346, Dmel\\CG3346, PON, Pon', description='partner of numb', other_designations='CG3346-PA|pon-PA', map_location='4C10-4C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4636318, end_position_on_the_genomic_accession=4639112, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31395, CurrentID=0, Status='live', Symbol='CG3081', Aliases='Dmel_ Dmel\\CG3081, EG:84H4.4', description='CG3081 gene product from transcript CG3081-RB', other_designations='CG3081-PA|CG3081-PB', map_location='4C10-4C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4639259, end_position_on_the_genomic_accession=4643987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31396, CurrentID=0, Status='live', Symbol='CG3062', Aliases='Dmel_ Dmel\\CG3062, EG:84H4.5', description='CG3062 gene product from transcript CG3062-RA', other_designations='CG3062-PA', map_location='4C10-4C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4641269, end_position_on_the_genomic_accession=4644868, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31397, CurrentID=0, Status='live', Symbol='HLH4C', Aliases='Dmel_CG3052, CG3052, Dmel\\CG3052, DroNHLH, EG:84H4.2, bHLH36', description='Helix loop helix protein 4C', other_designations='CG3052-PA|CG3052-PB|HLH4C-PA|HLH4C-PB|helix loop helix protein 4C', map_location='4C10-4C11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4645127, end_position_on_the_genomic_accession=4647414, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31398, CurrentID=0, Status='live', Symbol='mRpL30', Aliases='Dmel_CG7038, CG7038, Dmel\\CG7038, L30, mRPL30', description='mitochondrial ribosomal protein L30', other_designations='CG7038-PA|mRpL30-PA', map_location='4C11-4C11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4647871, end_position_on_the_genomic_accession=4648646, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31399, CurrentID=0, Status='live', Symbol='Torsin', Aliases='Dmel_CG3024, CG3024, DmTorp4a, Dmel\\CG3024, EG:84H4.1, TORP4A, Torp4a, dtor, torp4a, torsin', description='CG3024 gene product from transcript CG3024-RB', other_designations='CG3024-PA|CG3024-PB|Torsin-PA|Torsin-PB|dtorsin|torp4a|torsin', map_location='4C11-4C11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4648592, end_position_on_the_genomic_accession=4650032, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31401, CurrentID=0, Status='live', Symbol='CG15473', Aliases='Dmel_ Dmel\\CG15473', description='CG15473 gene product from transcript CG15473-RA', other_designations='CG15473-PA|uncharacterized protein', map_location='4C11-4C11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4651955, end_position_on_the_genomic_accession=4654497, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31402, CurrentID=0, Status='live', Symbol='CG3009', Aliases='Dmel_ Dmel\\CG3009', description='CG3009 gene product from transcript CG3009-RD', other_designations='CG3009-PA|CG3009-PB|CG3009-PD', map_location='4C11-4C11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4658158, end_position_on_the_genomic_accession=4666700, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31404 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31404, CurrentID=0, Status='live', Symbol='Pp2C1', Aliases='Dmel_CG2984, CG2984, Dmel\\CG2984, PP2C, PP2C1, anon-WO03040301.132, dpp2c1', description='Protein phosphatase 2C', other_designations='CG2984-PA|CG2984-PB|Pp2C1-PA|Pp2C1-PB|protein phosphatase 2C', map_location='4C12-4C13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4679988, end_position_on_the_genomic_accession=4686903, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31405 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31405, CurrentID=0, Status='live', Symbol='ctp', Aliases='Dmel_CG6998, CG6998, Cdlc1, Ctp, DLC8, Ddlc-1, Ddlc1, Dlc, Dmel\\CG6998, LC8, anon-WO03040301.128, ddlc1, dlc, l(1)G0153, l(1)G0204, l(1)G0207, l(1)G0244, l(1)G0248, l(1)G0338, l(1)G0371, l(1)G0387, l(1)G0432, l(1)G0445, lincRNA.S8635', description='cut up', other_designations='8kd dynein light chain|CG6998-PA|CG6998-PB|CG6998-PC|CG6998-PE|ctp-PA|ctp-PB|ctp-PC|ctp-PE|cytoplasmic dynein|cytoplasmic dynein 8kD light chain|cytoplasmic dynein light chain 1|dynein|dynein light chain|dynein light chain 1|dynein light chain LC8|dynein light-chain|dynein light-chain 1', map_location='4C13-4C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4687800, end_position_on_the_genomic_accession=4702034, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31406, CurrentID=0, Status='live', Symbol='l(1)G0334', Aliases='Dmel_CG7010, CG 7010, CG7010, Dmel\\CG7010, E1-PDH, L(1)g0334, PDH, Pdh, l (1) G0334', description='lethal (1) G0334', other_designations='CG7010-PA|CG7010-PB|CG7010-PC|CG7010-PD|PDH complex|l(1)G0334-PA|l(1)G0334-PB|l(1)G0334-PC|l(1)G0334-PD|pyruvate dehydrogenase', map_location='4C14-4C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4702284, end_position_on_the_genomic_accession=4705265, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31407, CurrentID=0, Status='live', Symbol='CG7024', Aliases='Dmel_ Dmel\\CG7024', description='CG7024 gene product from transcript CG7024-RA', other_designations='CG7024-PA', map_location='4C14-4C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4705309, end_position_on_the_genomic_accession=4707504, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31408, CurrentID=0, Status='live', Symbol='Proc-R', Aliases='Dmel_CG6986, CG6986, Dmel\\CG6986, Proc R, ProcR, anon-WO0170980.73, anon-WO0170980.74', description='Proctolin receptor', other_designations='CG6986-PA|CG6986-PB|CG6986-PC|CG6986-PD|CG6986-PE|Proc-R-PA|Proc-R-PB|Proc-R-PC|Proc-R-PD|Proc-R-PE|proctolin receptor', map_location='4C14-4C15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4707916, end_position_on_the_genomic_accession=4729487, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31409, CurrentID=0, Status='live', Symbol='CG15472', Aliases='Dmel_ Dmel\\CG15472', description='CG15472 gene product from transcript CG15472-RA', other_designations='CG15472-PA|uncharacterized protein', map_location='4D1-4D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4768727, end_position_on_the_genomic_accession=4770137, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31410, CurrentID=0, Status='live', Symbol='Bteb2', Aliases='Dmel_CG2932, BTEB2, BteB2, CG2932, Dmel\\CG2932', description='CG2932 gene product from transcript CG2932-RA', other_designations='Bteb2-PA|CG2932-PA', map_location='4D1-4D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4770557, end_position_on_the_genomic_accession=4771588, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31411 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31411, CurrentID=0, Status='live', Symbol='CG2871', Aliases='Dmel_ Dmel\\CG2871', description='CG2871 gene product from transcript CG2871-RA', other_designations='CG2871-PA', map_location='4D1-4D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4775984, end_position_on_the_genomic_accession=4777375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31412, CurrentID=0, Status='live', Symbol='CG15471', Aliases='Dmel_ Dmel\\CG15471', description='CG15471 gene product from transcript CG15471-RB', other_designations='CG15471-PB|CG15471-PC|uncharacterized protein', map_location='4D1-4D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4782201, end_position_on_the_genomic_accession=4782789, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31413, CurrentID=0, Status='live', Symbol='CG6978', Aliases='Dmel_ Dmel\\CG6978', description='CG6978 gene product from transcript CG6978-RA', other_designations='CG6978-PA', map_location='4D1-4D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4784075, end_position_on_the_genomic_accession=4786059, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31414, CurrentID=0, Status='live', Symbol='CG2861', Aliases='Dmel_ Dmel\\CG2861', description='CG2861 gene product from transcript CG2861-RE', other_designations='CG2861-PA|CG2861-PC|CG2861-PD|CG2861-PE', map_location='4D1-4D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4797231, end_position_on_the_genomic_accession=4804862, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31415, CurrentID=0, Status='live', Symbol='CG12682', Aliases='Dmel_ Dmel\\CG12682', description='CG12682 gene product from transcript CG12682-RA', other_designations='CG12682-PA|uncharacterized protein', map_location='4D2-4D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4825974, end_position_on_the_genomic_accession=4826833, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31416, CurrentID=0, Status='live', Symbol='CG12681', Aliases='Dmel_ Dmel\\CG12681', description='CG12681 gene product from transcript CG12681-RA', other_designations='CG12681-PA|uncharacterized protein', map_location='4D3-4D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4875670, end_position_on_the_genomic_accession=4877263, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=326329, CurrentID=0, Status='live', Symbol='CG33199', Aliases='Dmel_ Dmel\\CG33199', description='CG33199 gene product from transcript CG33199-RB', other_designations='CG33199-PA|CG33199-PB|uncharacterized protein', map_location='44F11-44F11', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=8947681, end_position_on_the_genomic_accession=8951086, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31419, CurrentID=0, Status='live', Symbol='CG6927', Aliases='Dmel_ CT21430, Dmel\\CG6927', description='CG6927 gene product from transcript CG6927-RA', other_designations='CG6927-PA|glycine-gated chloride channel-like', map_location='4D5-4D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4906900, end_position_on_the_genomic_accession=4908821, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31420, CurrentID=0, Status='live', Symbol='CG32772', Aliases='Dmel_ CG15469, CG6911, Dmel\\CG32772', description='CG32772 gene product from transcript CG32772-RE', other_designations='CG32772-PC|CG32772-PD|CG32772-PE|CG32772-PF|uncharacterized protein', map_location='4D5-4D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4909102, end_position_on_the_genomic_accession=4915161, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31422, CurrentID=0, Status='live', Symbol='CG4041', Aliases='Dmel_ Dmel\\CG4041', description='CG4041 gene product from transcript CG4041-RB', other_designations='CG4041-PB', map_location='4D5-4D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4915280, end_position_on_the_genomic_accession=4918447, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31423, CurrentID=0, Status='live', Symbol='CG6903', Aliases='Dmel_ Dmel\\CG6903', description='CG6903 gene product from transcript CG6903-RA', other_designations='CG6903-PA|CG6903-PB', map_location='4D6-4D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4918730, end_position_on_the_genomic_accession=4921648, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31424, CurrentID=0, Status='live', Symbol='CG44774', Aliases='Dmel_ CG4068, CR43698, CR43699, Dmel\\CG44774, Dmel_CG4068, Dmel_CR43698, Dmel_CR43699', description='CG44774 gene product from transcript CG44774-RH', other_designations='CG44774-PA|CG44774-PB|CG44774-PC|CG44774-PD|CG44774-PE|CG44774-PF|CG44774-PG|CG44774-PH|CG44774-PI|uncharacterized protein', map_location='4D6-4D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4921678, end_position_on_the_genomic_accession=4938105, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31425, CurrentID=0, Status='live', Symbol='Ptp4E', Aliases='Dmel_CG6899, 4E, CG6899, CT21187, DPTP4E, Dmel\\CG6899, PTP4E, PtP4E, R-PTP 4E', description='Protein tyrosine phosphatase 4E', other_designations='CG6899-PA|CG6899-PB|CG6899-PC|CG6899-PD|CG6899-PF|CG6899-PG|CG6899-PH|Ptp4E-PA|Ptp4E-PB|Ptp4E-PC|Ptp4E-PD|Ptp4E-PF|Ptp4E-PG|Ptp4E-PH|protein tyrosine phosphatase 4E', map_location='4D7-4D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4939128, end_position_on_the_genomic_accession=4955916, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  31426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31426, CurrentID=0, Status='live', Symbol='CG15468', Aliases='Dmel_ Dmel\\CG15468, SIP3', description='CG15468 gene product from transcript CG15468-RA', other_designations='CG15468-PA|syntaxin interacting protein 3|uncharacterized protein', map_location='4D7-4D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4959929, end_position_on_the_genomic_accession=4961128, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31427, CurrentID=0, Status='live', Symbol='CG12680', Aliases='Dmel_ Dmel\\CG12680', description='CG12680 gene product from transcript CG12680-RC', other_designations='CG12680-PB|CG12680-PC|uncharacterized protein', map_location='4E1-4E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=4997441, end_position_on_the_genomic_accession=4998388, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31429, CurrentID=0, Status='live', Symbol='ovo', Aliases='Dmel_CG6824, CG15467, CG6824, Dmel\\CG6824, Fs(1)K1103, Fs(1)K1237, Fs(1)K155, OVO, Ovo, Ovo-D, Shv, Svb, Svb/Ovo, fs(1)K1237, fs(1)M1, fs(1)M38/shavenbaby, ovo/svb, ovoD, svb', description='CG6824 gene product from transcript CG6824-RE', other_designations='CG6824-PA|CG6824-PB|CG6824-PC|CG6824-PD|CG6824-PE|ovo-PA|ovo-PB|ovo-PC|ovo-PD|ovo-PE|ovo/shaven-baby|shaven baby|shavenbaby', map_location='1-10.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5048273, end_position_on_the_genomic_accession=5070929, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31430, CurrentID=0, Status='live', Symbol='CG32767', Aliases='Dmel_ BcDNA:GM08530, CG5086, CG5113, Dmel\\CG32767, LD34931, i7', description='CG32767 gene product from transcript CG32767-RH', other_designations='CG32767-PC|CG32767-PD|CG32767-PE|CG32767-PF|CG32767-PG|CG32767-PH|uncharacterized protein', map_location='4E2-4E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5071027, end_position_on_the_genomic_accession=5084768, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31433 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31433, CurrentID=0, Status='live', Symbol='Rpn13R', Aliases='Dmel_CG6789, CG6789, Dmel\\CG6789', description='Regulatory particle non-ATPase 13-related', other_designations='CG6789-PA|Rpn13R-PA|regulatory particle non-ATPase 13-related', map_location='4F1-4F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5131898, end_position_on_the_genomic_accession=5132985, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31434, CurrentID=0, Status='live', Symbol='CG15465', Aliases='Dmel_CG15465/CG6775, Dmel\\CG15465', description='CG15465 gene product from transcript CG15465-RB', other_designations='CG15465-PA|CG15465-PB|uncharacterized protein', map_location='4F2-4F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5085759, end_position_on_the_genomic_accession=5254864, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  31435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31435, CurrentID=0, Status='live', Symbol='CG5062', Aliases='Dmel_ BcDNA:AT13663, Dmel\\CG5062, FBgn0029747', description='CG5062 gene product from transcript CG5062-RA', other_designations='CG5062-PA', map_location='4F2-4F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5180182, end_position_on_the_genomic_accession=5182369, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31438, CurrentID=0, Status='live', Symbol='CG42749', Aliases='Dmel_ CG15785, CG15786, Dmel\\CG42749, Dmel_CG15785, Dmel_CG15786', description='CG42749 gene product from transcript CG42749-RD', other_designations='CG42749-PB|CG42749-PC|CG42749-PD|uncharacterized protein', map_location='4F4-4F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5296066, end_position_on_the_genomic_accession=5306288, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31439, CurrentID=0, Status='live', Symbol='CG3323', Aliases='Dmel_ Dmel\\CG3323', description='CG3323 gene product from transcript CG3323-RA', other_designations='CG3323-PA', map_location='4F4-4F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5302988, end_position_on_the_genomic_accession=5305691, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31440, CurrentID=0, Status='live', Symbol='CG17764', Aliases='Dmel_ Dmel\\CG17764', description='CG17764 gene product from transcript CG17764-RA', other_designations='CG17764-PA|uncharacterized protein', map_location='4F4-4F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5306261, end_position_on_the_genomic_accession=5307417, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31441, CurrentID=0, Status='live', Symbol='Cdk7', Aliases='Dmel_CG3319, CD 7, CDK7, CG3319, DmCDK7, DmCdk7, DmMO15, DmMo15, Dmcdk7, Dmel\\CG3319, Mo15, cdk-7, cdk7, dCdk7, p40[MO15]', description='Cyclin-dependent kinase 7', other_designations='CG3319-PA|Cdk7-PA|Mo15', map_location='4F4-4F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5307465, end_position_on_the_genomic_accession=5308983, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31442, CurrentID=0, Status='live', Symbol='snf', Aliases="Dmel_CG4528, CG4528, D25, Dmel\\CG4528, SNF, SNF/D25, Snf, U1-A, U2 snRNP, U2 snRNP B'' protein, fs(1)1621, fs(1)A16121, fs(1)A1621, liz, snRNP B''1621, snf", description='sans fille', other_designations='CG4528-PA|Sans-fille|simply not fertile|snf-PA|splicing necessary factor', map_location='1-11.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5309242, end_position_on_the_genomic_accession=5310501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31443, CurrentID=0, Status='live', Symbol='TrxT', Aliases='Dmel_CG3315, 3315, CG3315, DmTrxT, DmelTrxT, Dmel\\CG3315', description='Thioredoxin T', other_designations='CG3315-PA|CG3315-PB|ThioredoxinT|TrxT-PA|TrxT-PB|thioredoxin|thioredoxin T', map_location='4F4-4F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5310581, end_position_on_the_genomic_accession=5311888, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31444, CurrentID=0, Status='live', Symbol='dhd', Aliases='Dmel_CG4193, 4193, CG4193, Dhd, DmTrx-1, Dmdhd, Dmel\\CG4193, Dmeldhd, Trx-1', description='deadhead', other_designations='CG4193-PA|CG4193-PB|dhd-PA|dhd-PB|thioredoxin|thioredoxin-like', map_location='4F4-4F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5312171, end_position_on_the_genomic_accession=5313125, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31445 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31445, CurrentID=0, Status='live', Symbol='CG4198', Aliases='Dmel_ Dmel\\CG4198', description='CG4198 gene product from transcript CG4198-RA', other_designations='CG4198-PA|CG4198-PB', map_location='4F4-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5313258, end_position_on_the_genomic_accession=5315190, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31446, CurrentID=0, Status='live', Symbol='CG15930', Aliases='Dmel_ Dmel\\CG15930', description='CG15930 gene product from transcript CG15930-RB', other_designations='CG15930-PB|uncharacterized protein', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5315130, end_position_on_the_genomic_accession=5317295, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31447, CurrentID=0, Status='live', Symbol='Sas10', Aliases='Dmel_CG4202, CG4202, Dmel\\CG4202, dSAS10, l(1)G0106, sas-10, sas10', description='Something about silencing 10 ortholog (S. cerevisiae)', other_designations='CG4202-PA|Sas10-PA|something about silencing 10 ortholog', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5317409, end_position_on_the_genomic_accession=5319071, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31448, CurrentID=0, Status='live', Symbol='Rnp4F', Aliases='Dmel_CG3312, 4F-rnp, 4f-rnp, CG3312, Dmel\\CG3312, RNP-4F, Rnp-4F, rnp-4f', description='RNA-binding protein 4F', other_designations='CG3312-PA|CG3312-PB|CG3312-PD|CG3312-PF|Rnp4F-PA|Rnp4F-PB|Rnp4F-PD|Rnp4F-PF|splicing assembly factor rnp-4f', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5318928, end_position_on_the_genomic_accession=5322696, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31449 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31449, CurrentID=0, Status='live', Symbol='Mcm3', Aliases='Dmel_CG4206, CG4206, DmMCM3, DmMcm3, DmeMCM3, Dmel\\CG4206, MCM3, McM3, Mcp-PCR2, PCR2, dMCM3, dMcm3', description='Minichromosome maintenance 3', other_designations='CG4206-PA|CG4206-PB|Mcm3-PA|Mcm3-PB|minichromosome maintenance 3|minichromosome maintenance-PCR2', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5323855, end_position_on_the_genomic_accession=5326738, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31450, CurrentID=0, Status='live', Symbol='CG3309', Aliases='Dmel_ BEST:LD39762, Dmel\\CG3309', description='CG3309 gene product from transcript CG3309-RB', other_designations='CG3309-PA|CG3309-PB', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5326675, end_position_on_the_genomic_accession=5329347, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31451 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31451, CurrentID=0, Status='live', Symbol='XRCC1', Aliases='Dmel_CG4208, 4208, CG4208, Dmel\\CG4208, Xrcc1', description='CG4208 gene product from transcript CG4208-RA', other_designations='CG4208-PA|XRCC1-PA', map_location='4F5-4F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5329876, end_position_on_the_genomic_accession=5332435, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31456 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31456, CurrentID=0, Status='live', Symbol='SK', Aliases='Dmel_CG10706, BcDNA:GM01744, CG10706, CG12733, CG4179, Dmel\\CG10706, Sk, dSK', description='small conductance calcium-activated potassium channel', other_designations='CG10706-PH|CG10706-PI|CG10706-PJ|CG10706-PK|CG10706-PL|CG10706-PM|CG10706-PN|CG10706-PO|CG10706-PP|CG10706-PQ|CG10706-PR|CG10706-PS|CG10706-PT|CG10706-PU|SK-PH|SK-PI|SK-PJ|SK-PK|SK-PL|SK-PM|SK-PN|SK-PO|SK-PP|SK-PQ|SK-PR|SK-PS|SK-PT|SK-PU', map_location='4F5-4F9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5340236, end_position_on_the_genomic_accession=5405138, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  31457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31457, CurrentID=0, Status='live', Symbol='NAAT1', Aliases='Dmel_CG3252, CG3252, DmNAT1, Dmel\\CG3252, anon-WO0149848.3, dmNAT1', description='Nutrient Amino Acid Transporter 1', other_designations='CG3252-PA|NAAT1-PA|nutrient amino acid transporter 1', map_location='4F9-4F9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5405238, end_position_on_the_genomic_accession=5410849, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31458, CurrentID=0, Status='live', Symbol='Usp16-45', Aliases='Dmel_CG4165, CG4165, Dmel\\CG4165, Q9W4C3, USP45, Usp16, Usp45', description='Ubiquitin specific protease 16/45', other_designations='CG4165-PA|CG4165-PB|CG4165-PD|CG4165-PE|CG4165-PF|Usp16-45-PA|Usp16-45-PB|Usp16-45-PD|Usp16-45-PE|Usp16-45-PF|ubiquitin specific protease 16/45', map_location='4F10-4F10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5421533, end_position_on_the_genomic_accession=5427219, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31459, CurrentID=0, Status='live', Symbol='spoon', Aliases='Dmel_CG3249, BcDNA:GM04319, CG3249, DAKAP550, Dmel\\CG3249, PKAAP, Yu, yu', description='spoonbill', other_designations='AKAP Yu|CG3249-PA|CG3249-PC|CG3249-PD|CG3249-PE|PKA anchor protein|spoon-PA|spoon-PC|spoon-PD|spoon-PE|yu', map_location='1-14 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5412386, end_position_on_the_genomic_accession=5420867, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31460, CurrentID=0, Status='live', Symbol='CG16756', Aliases='Dmel_ BcDNA:RH17440, Dmel\\CG16756', description='CG16756 gene product from transcript CG16756-RA', other_designations='CG16756-PA|uncharacterized protein', map_location='4F10-4F10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5427517, end_position_on_the_genomic_accession=5428217, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31461, CurrentID=0, Status='live', Symbol='CG15784', Aliases='Dmel_ Dmel\\CG15784', description='CG15784 gene product from transcript CG15784-RB', other_designations='CG15784-PA|CG15784-PB|uncharacterized protein', map_location='4F10-4F10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5429571, end_position_on_the_genomic_accession=5432526, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31463, CurrentID=0, Status='live', Symbol='SPR', Aliases='Dmel_CG16752, CG12731, CG16752, Dmel\\CG16752', description='Sex peptide receptor', other_designations='CG16752-PA|CG16752-PB|CG16752-PC|SPR-PA|SPR-PB|SPR-PC|sex peptide receptor', map_location='4F11-5A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5446656, end_position_on_the_genomic_accession=5494012, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31464, CurrentID=0, Status='live', Symbol='frma', Aliases='Dmel_CG3239, BEST:GH09377, CG3239, Dmel\\CG3239', description='fra mauro', other_designations='CG3239-PF|CG3239-PG|Dmel_CG3239|frma-PF|frma-PG', map_location='5A1-5A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5493014, end_position_on_the_genomic_accession=5496052, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31465 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31465, CurrentID=0, Status='live', Symbol='CG4151', Aliases='Dmel_ Dmel\\CG4151', description='CG4151 gene product from transcript CG4151-RB', other_designations='CG4151-PA|CG4151-PB', map_location='5A2-5A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5501476, end_position_on_the_genomic_accession=5502782, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31466 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31466, CurrentID=0, Status='live', Symbol='CG12730', Aliases='Dmel_ BcDNA:RE01432, Dmel\\CG12730', description='CG12730 gene product from transcript CG12730-RB', other_designations='CG12730-PA|CG12730-PB|uncharacterized protein', map_location='5A2-5A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5516608, end_position_on_the_genomic_accession=5518771, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31469 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31469, CurrentID=0, Status='live', Symbol='Vsx2', Aliases='Dmel_CG33980, CG15781, CG15782, CG15783, CG33980, Chx2, DChx2, Dmel\\CG33980, dChx2, dVsx2', description='Visual system homeobox 2 ortholog', other_designations='CG33980-PA|CG33980-PB|Chx2|Vsx2-PA|Vsx2-PB|visual system homeobox 2 ortholog', map_location='5A3-5A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5528927, end_position_on_the_genomic_accession=5560183, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31470, CurrentID=0, Status='live', Symbol='Vsx1', Aliases='Dmel_CG4136, CG4136, Chx1, DChx1, DVSX1, Dchx1, Dmel\\CG4136, Vsx-1, dChx1, dVsx1, lincRNA.S8688', description='Visual system homeobox 1 ortholog', other_designations='CG4136-PA|CG4136-PB|Chx1|Vsx1-PA|Vsx1-PB|visual system homeobox 1 ortholog', map_location='5A6-5A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5594255, end_position_on_the_genomic_accession=5612865, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31471, CurrentID=0, Status='live', Symbol='CG34435', Aliases='Dmel_ CG15780, CG17763, Dmel\\CG34435, Dmel_CG15780, Dmel_CG17763', description='CG34435 gene product from transcript CG34435-RA', other_designations='CG34435-PA|uncharacterized protein', map_location='5A7-5A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5613325, end_position_on_the_genomic_accession=5614467, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31472, CurrentID=0, Status='live', Symbol='CG34434', Aliases='Dmel_ CG15780, CG17763, Dmel\\CG34434, Dmel_CG15780, Dmel_CG17763', description='CG34434 gene product from transcript CG34434-RB', other_designations='CG34434-PB|uncharacterized protein', map_location='5A8-5A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5614554, end_position_on_the_genomic_accession=5616370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31473, CurrentID=0, Status='live', Symbol='RhoGAP5A', Aliases='Dmel_CG3208, CG3208, Dmel\\CG3208, RhoGAP', description='Rho GTPase activating protein at 5A', other_designations='CG3208-PA|CG3208-PB|RhoGAP5A-PA|RhoGAP5A-PB', map_location='5A8-5A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5616572, end_position_on_the_genomic_accession=5618806, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31474, CurrentID=0, Status='live', Symbol='Mlc-c', Aliases='Dmel_CG3201, CG3201, DmMLC_C, Dmel\\CG3201, MyoII, mlc-c, nmELC', description='Myosin light chain cytoplasmic', other_designations='CG3201-PA|CG3201-PB|Mlc-c-PA|Mlc-c-PB|Myosin-II|alkali light chain of non-muscle myosin-II|essential light chain-cytoplasmic|essential myosin light chain-cytoplasmic|myosin|myosin II|myosin light chain cytoplasmic|myosin light chain-cytoplasmic|non-muscle myosin II', map_location='5A8-5A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5619136, end_position_on_the_genomic_accession=5622539, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31475, CurrentID=0, Status='live', Symbol='l(1)G0045', Aliases='Dmel_CG32763, BcDNA:LD21623, CG32763, CG4129, Dmel\\CG32763, l(1)G0312', description='lethal (1) G0045', other_designations='CG32763-PA|l(1)G0045-PA', map_location='5A8-5A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5622648, end_position_on_the_genomic_accession=5625493, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31480, CurrentID=0, Status='live', Symbol='Sirt4', Aliases='Dmel_CG3187, CG3187, D.mel3, Dmel\\CG3187, dSIRT4, dmSRT407', description='Sirtuin 4', other_designations='CG3187-PA|CG3187-PB|CG3187-PC|Sirt4-PA|Sirt4-PB|Sirt4-PC|sirtuin 4', map_location='5A10-5A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5664628, end_position_on_the_genomic_accession=5665940, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31481, CurrentID=0, Status='live', Symbol='CG4119', Aliases='Dmel_ BcDNA:LD23634, Dmel\\CG4119, cg4119', description='CG4119 gene product from transcript CG4119-RB', other_designations='CG4119-PA|CG4119-PB', map_location='5A10-5A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5666052, end_position_on_the_genomic_accession=5670652, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31483, CurrentID=0, Status='live', Symbol='RpL35', Aliases='Dmel_CG4111, BcDNA:RE09547, CG4111, Dmel\\CG4111, Rp L35e, anon-EST:Posey157, anon-EST:fe1C12, RpL35', description='Ribosomal protein L35', other_designations='CG4111-PA|CG4111-PB|CG4111-PC|RpL35-PA|RpL35-PB|RpL35-PC|anon-fast-evolving-1C12|ribosomal protein L35', map_location='5A11-5A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5672401, end_position_on_the_genomic_accession=5673547, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31486 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31486, CurrentID=0, Status='live', Symbol='CG42449', Aliases='Dmel_ CG15778, Dmel\\CG42449, Dmel_CG15778', description='CG42449 gene product from transcript CG42449-RB', other_designations='CG42449-PA|CG42449-PB|uncharacterized protein', map_location='5A12-5A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5683240, end_position_on_the_genomic_accession=5684456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31487, CurrentID=0, Status='live', Symbol='CG3160', Aliases='Dmel_ Dmel\\CG3160', description='CG3160 gene product from transcript CG3160-RA', other_designations='CG3160-PA', map_location='5A12-5A13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5685253, end_position_on_the_genomic_accession=5689473, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31489, CurrentID=0, Status='live', Symbol='CG3149', Aliases='Dmel_ BcDNA:GH08388, Dmel\\CG3149', description='CG3149 gene product from transcript CG3149-RA', other_designations='CG3149-PA', map_location='5A13-5A14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5692945, end_position_on_the_genomic_accession=5695205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31490, CurrentID=0, Status='live', Symbol='CG4096', Aliases='Dmel_ CT13580, Dmel\\CG4096', description='CG4096 gene product from transcript CG4096-RB', other_designations='CG4096-PB|CG4096-PC', map_location='5B1-5B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5698557, end_position_on_the_genomic_accession=5708633, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  31491 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31491, CurrentID=0, Status='live', Symbol='CG33080', Aliases='Dmel_ CG15774, CG15775, CG3141, Dmel\\CG33080', description='CG33080 gene product from transcript CG33080-RA', other_designations='CG33080-PA|CG33080-PB|CG33080-PC|CG33080-PD|uncharacterized protein', map_location='5B1-5B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5708559, end_position_on_the_genomic_accession=5717778, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31493, CurrentID=0, Status='live', Symbol='CG32758', Aliases='Dmel_ CG11465, CG3130, CG3138, DmSNX27, Dmel\\CG32758', description='CG32758 gene product from transcript CG32758-RD', other_designations='CG32758-PA|CG32758-PC|CG32758-PD|uncharacterized protein', map_location='5B2-5B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5719374, end_position_on_the_genomic_accession=5738031, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31494, CurrentID=0, Status='live', Symbol='CG15773', Aliases='Dmel_ Dmel\\CG15773', description='CG15773 gene product from transcript CG15773-RA', other_designations='CG15773-PA|CG15773-PB|uncharacterized protein', map_location='5B3-5B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5732740, end_position_on_the_genomic_accession=5734614, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31496, CurrentID=0, Status='live', Symbol='IntS6', Aliases='Dmel_CG3125, CG3125, Dmel\\CG3125, Int6, l(1)G0060', description='Integrator 6', other_designations='CG3125-PA|CG3125-PB|CG3125-PC|CG3125-PD|CG3125-PE|IntS6-PA|IntS6-PB|IntS6-PC|IntS6-PD|IntS6-PE|integrator 6', map_location='5B4-5B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5738663, end_position_on_the_genomic_accession=5747254, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31497, CurrentID=0, Status='live', Symbol='CG4078', Aliases='Dmel_ Dmel\\CG4078', description='CG4078 gene product from transcript CG4078-RB', other_designations='CG4078-PA|CG4078-PB', map_location='5B6-5B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5747747, end_position_on_the_genomic_accession=5752077, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31498, CurrentID=0, Status='live', Symbol='CG15772', Aliases='Dmel_ Dmel\\CG15772', description='CG15772 gene product from transcript CG15772-RB', other_designations='CG15772-PA|CG15772-PB|uncharacterized protein', map_location='5B6-5B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5752249, end_position_on_the_genomic_accession=5754041, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31499, CurrentID=0, Status='live', Symbol='lin-52', Aliases='Dmel_CG15929, CG15929, Dmel\\CG15929, Lin-52, Lin52, dLIN-52, dLin-52, dLin52, d dlin52', description='CG15929 gene product from transcript CG15929-RB', other_designations='CG15929-PA|CG15929-PB|lin-52-PA|lin-52-PB', map_location='5B7-5B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5753390, end_position_on_the_genomic_accession=5754958, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31500, CurrentID=0, Status='live', Symbol='CG15771', Aliases='Dmel_ 28571108, CR33074, Dmel\\CG15771', description='CG15771 gene product from transcript CG15771-RD', other_designations='CG15771-PA|CG15771-PC|CG15771-PD|uncharacterized protein', map_location='5B8-5B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5755574, end_position_on_the_genomic_accession=5759751, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31503, CurrentID=0, Status='live', Symbol='CG3097', Aliases='Dmel_ Dmel\\CG3097', description='CG3097 gene product from transcript CG3097-RA', other_designations='CG3097-PA', map_location='5B8-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5765855, end_position_on_the_genomic_accession=5770780, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31504, CurrentID=0, Status='live', Symbol='CG42264', Aliases='Dmel_ CG15768, CG15769, Dmel\\CG42264, Dmel_CG15768, Dmel_CG15769', description='CG42264 gene product from transcript CG42264-RB', other_designations='CG42264-PA|CG42264-PB|uncharacterized protein', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5773023, end_position_on_the_genomic_accession=5776647, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  43480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43480, CurrentID=0, Status='live', Symbol='CG11470', Aliases='Dmel_ Dmel\\CG11470', description='CG11470 gene product from transcript CG11470-RB', other_designations='CG11470-PA|CG11470-PB|uncharacterized protein', map_location='99B4-99B4', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=29588993, end_position_on_the_genomic_accession=29589762, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31506, CurrentID=0, Status='live', Symbol='CG3108', Aliases='Dmel_ Dmel\\CG3108', description='CG3108 gene product from transcript CG3108-RA', other_designations='CG3108-PA', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5777195, end_position_on_the_genomic_accession=5781620, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31507, CurrentID=0, Status='live', Symbol='CG42699', Aliases='Dmel_ CG15770, CG4064, Dmel\\CG42699, Dmel_CG15770, Dmel_CG4064', description='CG42699 gene product from transcript CG42699-RC', other_designations='CG42699-PC|uncharacterized protein', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5764553, end_position_on_the_genomic_accession=5789576, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31508, CurrentID=0, Status='live', Symbol='CG15767', Aliases='Dmel_ Dmel\\CG15767', description='CG15767 gene product from transcript CG15767-RA', other_designations='CG15767-PA|uncharacterized protein', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5789487, end_position_on_the_genomic_accession=5790919, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31509, CurrentID=0, Status='live', Symbol='CG12239', Aliases='Dmel_ Dmel\\CG12239', description='CG12239 gene product from transcript CG12239-RA', other_designations='CG12239-PA|uncharacterized protein', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5792140, end_position_on_the_genomic_accession=5794543, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31510, CurrentID=0, Status='live', Symbol='Cpr5C', Aliases='Dmel_CG4052, CG4052, Dmel\\CG4052', description='Cuticular protein 5C', other_designations='CG4052-PA|Cpr5C-PA|cuticular protein 5C', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5804677, end_position_on_the_genomic_accession=5805344, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31511, CurrentID=0, Status='live', Symbol='Prosbeta2R1', Aliases='Dmel_CG18341, CG18341, Dmel\\CG18341, beta2R1', description='Proteasome beta2 subunit-related 1', other_designations='CG18341-PA|Prosbeta2R1-PA|proteasome beta2 subunit-related 1|proteasome beta2R1 subunit', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5808954, end_position_on_the_genomic_accession=5810158, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31512, CurrentID=0, Status='live', Symbol='CG15766', Aliases='Dmel_ Dmel\\CG15766', description='CG15766 gene product from transcript CG15766-RA', other_designations='CG15766-PA|uncharacterized protein', map_location='5C2-5C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5814864, end_position_on_the_genomic_accession=5815820, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31513, CurrentID=0, Status='live', Symbol='CG15765', Aliases='Dmel_ Dmel\\CG15765', description='CG15765 gene product from transcript CG15765-RA', other_designations='CG15765-PA|CG15765-PB|uncharacterized protein', map_location='5C2-5C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5818559, end_position_on_the_genomic_accession=5840823, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31515, CurrentID=0, Status='live', Symbol='CG12729', Aliases='Dmel_ Dmel\\CG12729', description='CG12729 gene product from transcript CG12729-RA', other_designations='CG12729-PA|uncharacterized protein', map_location='5C3-5C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5845280, end_position_on_the_genomic_accession=5846106, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31516, CurrentID=0, Status='live', Symbol='CG15764', Aliases='Dmel_ BcDNA:AT16792, Dmel\\CG15764', description='CG15764 gene product from transcript CG15764-RB', other_designations='CG15764-PA|CG15764-PB|uncharacterized protein', map_location='5C5-5C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5869095, end_position_on_the_genomic_accession=5870266, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31517, CurrentID=0, Status='live', Symbol='GAA1', Aliases='Dmel_CG3033, CG3033, Dmel\\CG3033', description='Glycosylphosphatidylinositol anchor attachment 1 ortholog (H. sapiens)', other_designations='CG3033-PA|GAA1-PA|glycosylphosphatidylinositol anchor attachment 1 ortholog', map_location='5C5-5C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5870841, end_position_on_the_genomic_accession=5873357, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31518, CurrentID=0, Status='live', Symbol='mof', Aliases='Dmel_CG3025, CG3025, Dmel\\CG3025, MOF, Mof, dMOF, dmMOF, max', description='males absent on the first', other_designations='CG3025-PA|MOF histone acetylase|males absent on first|males-absent on the X|males-absent-on-the-first|mof-PA', map_location='5C5-5C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5873840, end_position_on_the_genomic_accession=5876853, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31519 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31519, CurrentID=0, Status='live', Symbol='Usp30', Aliases='Dmel_CG3016, CG3016, Dmel\\CG3016, Q9W462, USP30, dUSP30', description='Ubiquitin specific protease 30', other_designations='CG3016-PA|CG3016-PB|Usp30-PA|Usp30-PB|ubiquitin specific peptidase 30|ubiquitin specific protease 30', map_location='5C5-5C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5877000, end_position_on_the_genomic_accession=5879787, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31520, CurrentID=0, Status='live', Symbol='CG16721', Aliases='Dmel_ Dmel\\CG16721', description='CG16721 gene product from transcript CG16721-RA', other_designations='CG16721-PA|CG16721-PB|CG16721-PC|uncharacterized protein', map_location='5C6-5C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5879869, end_position_on_the_genomic_accession=5883391, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31521, CurrentID=0, Status='live', Symbol='Act5C', Aliases='Dmel_CG4027, A, A4V404_DROME, ACT, ACT1_DROME, ACT5C, Ac5C, Act, Act-5C, Act5, Act5c, Actin, Actin/BAP47, Actin5C, BAP47, Bap47, CG4027, Dmel\\CG4027, M32055, T11, act, act 5C, act5C, actin, actin5C, anon-EST:fe2D2, beta-actin, beta-actin/Bap47, chrX:5748184..5748304, cyt5C, l(1)G0009, l(1)G0010, l(1)G0025, l(1)G0079, l(1)G0117, l(1)G0177, l(1)G0245, l(1)G0330, l(1)G0420, l(1)G0486', description='Actin 5C', other_designations='5C actin|Act5C-PA|Act5C-PB|Act5C-PC|Act5C-PD|Act5C-PE|CG4027-PA|CG4027-PB|CG4027-PC|CG4027-PD|CG4027-PE|Dm-actin5C|F-actin|G-actin|actin|actin 5 C|actin 5C|actin 5c|actin A1|actin5C|anon-fast-evolving-2D2|beta-actin|cellular cytoskeletal beta-actin', map_location='5C7-5C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5900861, end_position_on_the_genomic_accession=5905399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31522, CurrentID=0, Status='live', Symbol='CG4020', Aliases='Dmel_ Dmel\\CG4020', description='CG4020 gene product from transcript CG4020-RB', other_designations='CG4020-PA|CG4020-PB', map_location='5C7-5C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5905746, end_position_on_the_genomic_accession=5908001, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  338395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=338395, CurrentID=0, Status='live', Symbol='CG33213', Aliases='Dmel_ CG12260, Dmel\\CG33213', description='CG33213 gene product from transcript CG33213-RA', other_designations='CG33213-PA|uncharacterized protein', map_location='98B6-98B6', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=27918182, end_position_on_the_genomic_accession=27919460, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31524, CurrentID=0, Status='live', Symbol='CG3011', Aliases='Dmel_ 144561_at, Dmel\\CG3011, Shmt2, anon-WO03040301.100', description='CG3011 gene product from transcript CG3011-RA', other_designations='CG3011-PA|CG3011-PB|CG3011-PC|glycine hydroxylmethyltransferase', map_location='5C7-5C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5913067, end_position_on_the_genomic_accession=5916797, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31525, CurrentID=0, Status='live', Symbol='CG3726', Aliases='Dmel_ Dmel\\CG3726', description='CG3726 gene product from transcript CG3726-RB', other_designations='CG3726-PA|CG3726-PB', map_location='5C7-5C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5917097, end_position_on_the_genomic_accession=5941852, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31526, CurrentID=0, Status='live', Symbol='CG12728', Aliases='Dmel_ Dmel\\CG12728, XRCC4', description='CG12728 gene product from transcript CG12728-RB', other_designations='CG12728-PA|CG12728-PB|X-ray repair cross-complementing protein 4|uncharacterized protein', map_location='5C10-5C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5942606, end_position_on_the_genomic_accession=5945250, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31527 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31527, CurrentID=0, Status='live', Symbol='CG6041', Aliases='Dmel_ Dmel\\CG6041, SP86', description='CG6041 gene product from transcript CG6041-RA', other_designations='CG6041-PA', map_location='5C10-5C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5948160, end_position_on_the_genomic_accession=5950208, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31528, CurrentID=0, Status='live', Symbol='CG6048', Aliases='Dmel_ Dmel\\CG6048, SP108', description='CG6048 gene product from transcript CG6048-RA', other_designations='CG6048-PA', map_location='5C10-5C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5952565, end_position_on_the_genomic_accession=5954046, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31529, CurrentID=0, Status='live', Symbol='CG6067', Aliases='Dmel_ Dmel\\CG6067', description='CG6067 gene product from transcript CG6067-RA', other_designations='CG6067-PA', map_location='5D1-5D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5955629, end_position_on_the_genomic_accession=5958283, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  5740876 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740876, CurrentID=0, Status='live', Symbol='CG34452', Aliases='Dmel_ BP1049, CG33999, Dmel\\CG34452, Dmel_CG33999', description='CG34452 gene product from transcript CG34452-RC', other_designations='CG34452-PB|CG34452-PC|uncharacterized protein', map_location='71F1-71F1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=15686352, end_position_on_the_genomic_accession=15687902, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31532, CurrentID=0, Status='live', Symbol='CG5966', Aliases='Dmel_ 141233_at, Dmel\\CG5966', description='CG5966 gene product from transcript CG5966-RB', other_designations='CG5966-PA|CG5966-PB|triglyceride lipase', map_location='5D1-5D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5988791, end_position_on_the_genomic_accession=5992640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31533, CurrentID=0, Status='live', Symbol='CG4766', Aliases='Dmel_ BcDNA:GH11415, Dmel\\CG4766, gh11415', description='CG4766 gene product from transcript CG4766-RA', other_designations='CG4766-PA', map_location='5D1-5D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=5997508, end_position_on_the_genomic_accession=6000797, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31535, CurrentID=0, Status='live', Symbol='rux', Aliases='Dmel_CG4336, CG4336, Dmel\\CG4336, Rux', description='roughex', other_designations='CG4336-PA|CG4336-PB|Rough-ex|rux-PA|rux-PB', map_location='1-15.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6035753, end_position_on_the_genomic_accession=6037800, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31536, CurrentID=0, Status='live', Symbol='MCTS1', Aliases='Dmel_CG5941, CG5941, Dmel\\CG5941, MCT-1, anon-WO03040301.150, anon-WO03040301.151', description='Malignant T cell amplified sequence 1', other_designations='CG5941-PA|CG5941-PB|MCTS1-PA|MCTS1-PB|malignant T cell amplified sequence 1|multiple copies in T-cell lymphoma-1', map_location='5D2-5D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6037892, end_position_on_the_genomic_accession=6039576, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31537, CurrentID=0, Status='live', Symbol='CG5937', Aliases='Dmel_ Dmel\\CG5937', description='CG5937 gene product from transcript CG5937-RD', other_designations='CG5937-PC|CG5937-PD', map_location='5D2-5D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6040005, end_position_on_the_genomic_accession=6050381, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31538, CurrentID=0, Status='live', Symbol='CG5921', Aliases='Dmel_ Dmel\\CG5921', description='CG5921 gene product from transcript CG5921-RC', other_designations='CG5921-PC|CG5921-PD|CG5921-PE', map_location='5D3-5D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6050828, end_position_on_the_genomic_accession=6065646, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31539, CurrentID=0, Status='live', Symbol='CG5928', Aliases='Dmel_ Dmel\\CG5928', description='CG5928 gene product from transcript CG5928-RC', other_designations='CG5928-PA|CG5928-PB|CG5928-PC', map_location='5D3-5D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6067958, end_position_on_the_genomic_accession=6069264, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31540, CurrentID=0, Status='live', Symbol='Tsp5D', Aliases='Dmel_CG4690, CG4690, Dm.Tsp5D, Dmel\\CG4690, tsp5D', description='Tetraspanin 5D', other_designations='CG4690-PA|CG4690-PB|CG4690-PC|Tsp5D-PA|Tsp5D-PB|Tsp5D-PC|tetraspanin 5D', map_location='5D3-5D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6069131, end_position_on_the_genomic_accession=6074857, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31541, CurrentID=0, Status='live', Symbol='CG4666', Aliases='Dmel_ Dmel\\CG4666', description='CG4666 gene product from transcript CG4666-RB', other_designations='CG4666-PA|CG4666-PB', map_location='5D3-5D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6076072, end_position_on_the_genomic_accession=6077786, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31542, CurrentID=0, Status='live', Symbol='CG4660', Aliases='Dmel_ BcDNA:RH58440, Dmel\\CG4660', description='CG4660 gene product from transcript CG4660-RA', other_designations='CG4660-PA', map_location='5D4-5D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6078708, end_position_on_the_genomic_accession=6081372, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31543, CurrentID=0, Status='live', Symbol='raptor', Aliases='Dmel_CG4320, 4320, CG4320, Dmel\\CG4320, Raptor, dRap, dRaptor/CG4320', description='CG4320 gene product from transcript CG4320-RC', other_designations='CG4320-PB|CG4320-PC|RAPTOR|dRaptor|raptor-PB|raptor-PC|target of rapamycin (TOR)-associated protein', map_location='5D4-5D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6082634, end_position_on_the_genomic_accession=6090201, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31544, CurrentID=0, Status='live', Symbol='Mipp2', Aliases='Dmel_CG4317, CG4317, Dmel\\CG4317, l(1)G0050, l(1)G0303', description='Multiple inositol polyphosphate phosphatase 2', other_designations='CG4317-PA|Mipp2-PA|multiple inositol polyphosphate phosphatase 2', map_location='5D4-5D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6082709, end_position_on_the_genomic_accession=6091471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31547, CurrentID=0, Status='live', Symbol='Nep1', Aliases='Dmel_CG5905, CG5894, CG5905, CG5914, DmeNEP1, Dmel\\CG5905, NEP, NEP1, dNEP1, nep1', description='Neprilysin 1', other_designations='CG5905-PA|CG5905-PB|CG5905-PC|CG5905-PD|Nep1-PA|Nep1-PB|Nep1-PC|Nep1-PD|neprilysin 1', map_location='5D4-5D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6091205, end_position_on_the_genomic_accession=6103971, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  31550 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31550, CurrentID=0, Status='live', Symbol='Ca-alpha1T', Aliases='Dmel_CG15899, CG12544, CG15899, CG4222, Ca-alphaT-RC, Ca1G, Caalpha1T, DmalphaG, Dmca1G, Dmel\\CG15899, alpha1T, cat', description='Ca[2+]-channel protein alpha[[1]] subunit T', other_designations='CG15899-PB|CG15899-PC|CG15899-PD|CG15899-PE|CG15899-PF|Ca-alpha1T-PB|Ca-alpha1T-PC|Ca-alpha1T-PD|Ca-alpha1T-PE|Ca-alpha1T-PF|Ca1G', map_location='5D5-5D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6104085, end_position_on_the_genomic_accession=6192483, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  31551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31551, CurrentID=0, Status='live', Symbol='vanin-like', Aliases='Dmel_CG32754, CG32754, CG3648, Dmel\\CG32754, Vanin-like', description='CG32754 gene product from transcript CG32754-RA', other_designations='CG32754-PA|pantetheine hydrolase|vanin-like-PA', map_location='5E1-5E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6200188, end_position_on_the_genomic_accession=6202559, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31552, CurrentID=0, Status='live', Symbol='Btnd', Aliases='Dmel_CG3599, BTD, Btd, CG3599, CT12113, Dmel\\CG3599', description='Biotinidase', other_designations='Btnd-PA|CG3599-PA|biotinidase', map_location='5E1-5E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6217268, end_position_on_the_genomic_accession=6219327, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31553, CurrentID=0, Status='live', Symbol='Efr', Aliases='Dmel_CG3774, CG3774, Dmel\\CG3774', description='ER GDP-fucose transporter', other_designations='CG3774-PA|CG3774-PB|ER GFP-fucose transporter|Efr-PA|Efr-PB', map_location='5E1-5E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6219957, end_position_on_the_genomic_accession=6222719, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31554, CurrentID=0, Status='live', Symbol='sqh', Aliases='Dmel_CG3595, CG3595, DmMRLC_C, Dmel\\CG3595, MLC, MLRC, MRLC, Myo, Myo II, Myo-II, Myo-II RLC, MyoII, NMM, RLC, RMLC, Sqh, myoII, myosin II, sgh, spg, squ, ssq', description='spaghetti squash', other_designations='CG3595-PA|CG3595-PB|CG3595-PC|CG3595-PD|CG3595-PE|Myo-II regulatory light chain|MyoII regulatory light chain|Myosin-II|Myosin-II light chain|Spaghetti-Squash|light chain of nonmuscle myosin II|myosin|myosin 2|myosin II|myosin II regulatory light chain|myosin LC|myosin light chain|myosin regulatory light chain|myosinII|non-muscle myosin II|non-muscle myosin II regulatory light chain|nonmuscle myosin II|nonmuscle myosin II regulatory light chain|phospho-Myosin light chain 2|regulatory light chain|regulatory light chain of myosin|regulatory light chain of myosin II|regulatory light chain of nonmuscle myosin II|regulatory light chain of nonmuscle myosin-II|regulatory myosin light chain|sqh-PA|sqh-PB|sqh-PC|sqh-PD|sqh-PE|squash|type II myosin regulatory light chain', map_location='1-13 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6223061, end_position_on_the_genomic_accession=6225140, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31555 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31555, CurrentID=0, Status='live', Symbol='CG14446', Aliases='Dmel_ Dmel\\CG14446, MRE2, lincRNA.S8753', description='CG14446 gene product from transcript CG14446-RC', other_designations='CG14446-PB|CG14446-PC|CG14446-PD|CG14446-PE|mRNA-like ncRNA in embryogenesis 2|uncharacterized protein', map_location='5E1-5E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6225647, end_position_on_the_genomic_accession=6236218, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  31556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31556, CurrentID=0, Status='live', Symbol='CG14445', Aliases='Dmel_ Dmel\\CG14445', description='CG14445 gene product from transcript CG14445-RC', other_designations='CG14445-PA|CG14445-PB|CG14445-PC|CG14445-PD|uncharacterized protein', map_location='5E1-5E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6237315, end_position_on_the_genomic_accession=6240558, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31557, CurrentID=0, Status='live', Symbol='Rbcn-3A', Aliases='Dmel_CG3585, CG3585, DmX, Dmel\\CG3585, Dmx, FV10, X, anon-X', description='Rabconnectin-3A', other_designations='CG3585-PB|Rabconnectin-3alpha|Rbcn-3A-PB', map_location='5E1-5E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6240745, end_position_on_the_genomic_accession=6255487, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  31558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31558, CurrentID=0, Status='live', Symbol='Spx', Aliases='Dmel_CG3780, CG3780, Dmel\\CG3780, SPX, spx', description='Spliceosomal protein on the X', other_designations='CG3780-PA|Spx-PA|spliceosomal protein on the X', map_location='5E3-5E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6255822, end_position_on_the_genomic_accession=6257596, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31560 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31560, CurrentID=0, Status='live', Symbol='CG3781', Aliases='Dmel_ Dmel\\CG3781, Q9W422', description='CG3781 gene product from transcript CG3781-RA', other_designations='CG3781-PA', map_location='5E3-5E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6261169, end_position_on_the_genomic_accession=6262119, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31562, CurrentID=0, Status='live', Symbol='CG3566', Aliases='Dmel_ BcDNA:RH45308, Dmel\\CG3566', description='CG3566 gene product from transcript CG3566-RB', other_designations='CG3566-PB|CG3566-PC|CG3566-PD', map_location='5E4-5E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6274972, end_position_on_the_genomic_accession=6276724, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31564, CurrentID=0, Status='live', Symbol='CG11700', Aliases='Dmel_ CR11700, Dmel\\CG11700, ESTS:143G11T, ESTS:199A6T, Ubi-p5E5, Ubiquitin, anon-sts41, poly-ub', description='CG11700 gene product from transcript CG11700-RB', other_designations='CG11700-PB|ubiquitin|uncharacterized protein', map_location='5E5-5E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6282037, end_position_on_the_genomic_accession=6283575, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31565, CurrentID=0, Status='live', Symbol='Top3beta', Aliases='Dmel_CG3458, CG3458, Dmel\\CG3458, TOP3, Top3, Topo IIIbeta, dTop3beta, top3, top3[beta], top3beta, topo III[beta]', description='Topoisomerase 3beta', other_designations='CG3458-PA|CG3458-PB|Top 3beta|Top3beta-PA|Top3beta-PB|topoisomerase 3|topoisomerase 3beta', map_location='5E5-5E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6284138, end_position_on_the_genomic_accession=6287990, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31566, CurrentID=0, Status='live', Symbol='wuho', Aliases='Dmel_CG15897, CG15897, Dmel\\CG15897, wh', description='CG15897 gene product from transcript CG15897-RA', other_designations='CG15897-PA|wuho-PA', map_location='5E5-5E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6288100, end_position_on_the_genomic_accession=6289562, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31567, CurrentID=0, Status='live', Symbol='Rpt4', Aliases='Dmel_CG3455, CG3455, Dm_Rpt4a, Dm_Rpt4b, Dmel\\CG3455, Dmp42D, S10b, l(1)G0114, l(1)G0227, l(1)G0345, p42D', description='Regulatory particle triple-A ATPase 4', other_designations='CG3455-PB|Rpt4-PB|regulatory particle triple-A ATPase 4', map_location='5E5-5E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6289470, end_position_on_the_genomic_accession=6291245, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31568, CurrentID=0, Status='live', Symbol='CG15896', Aliases='Dmel_ Dmel\\CG15896', description='CG15896 gene product from transcript CG15896-RB', other_designations='CG15896-PA|CG15896-PB|uncharacterized protein', map_location='5E5-5E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6291303, end_position_on_the_genomic_accession=6294177, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31571, CurrentID=0, Status='live', Symbol='CG3815', Aliases='Dmel_ Dmel\\CG3815, PF1, Pf1', description='CG3815 gene product from transcript CG3815-RA', other_designations='CG3815-PA', map_location='5E6-5E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6295389, end_position_on_the_genomic_accession=6298970, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31572, CurrentID=0, Status='live', Symbol='CG12219', Aliases='Dmel_ BEST:LD26135, Dmel\\CG12219', description='CG12219 gene product from transcript CG12219-RA', other_designations='CG12219-PA|uncharacterized protein', map_location='5E7-5E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6299054, end_position_on_the_genomic_accession=6301285, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31573, CurrentID=0, Status='live', Symbol='CG3823', Aliases='Dmel_ Dmel\\CG3823', description='CG3823 gene product from transcript CG3823-RA', other_designations='CG3823-PA', map_location='5E7-5E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6303975, end_position_on_the_genomic_accession=6306745, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31574, CurrentID=0, Status='live', Symbol='CG15894', Aliases='Dmel_ Dmel\\CG15894', description='CG15894 gene product from transcript CG15894-RB', other_designations='CG15894-PA|CG15894-PB|CG15894-PC|CG15894-PD|uncharacterized protein', map_location='5F1-5F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6309212, end_position_on_the_genomic_accession=6329828, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31576, CurrentID=0, Status='live', Symbol='CG3842', Aliases='Dmel_ Dmel\\CG3842, lincRNA.959', description='CG3842 gene product from transcript CG3842-RD', other_designations='CG3842-PA|CG3842-PB|CG3842-PC|CG3842-PD', map_location='5F2-5F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6337762, end_position_on_the_genomic_accession=6343731, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31577 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31577, CurrentID=0, Status='live', Symbol='CG3847', Aliases='Dmel_ Dmel\\CG3847', description='CG3847 gene product from transcript CG3847-RA', other_designations='CG3847-PA', map_location='5F2-5F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6347103, end_position_on_the_genomic_accession=6349236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31578 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31578, CurrentID=0, Status='live', Symbol='ND-B16.6', Aliases='Dmel_CG3446, CG3446, Dmel\\CG3446, NB6M', description='NADH dehydrogenase (ubiquinone) B16.6 subunit', other_designations='CG3446-PA|CG3446-PB|CG3446-PC|NADH ubiquinone oxidoreductase B16.6|NADH-ubiquinone oxidoreductase B16.6 subunit|NADH:ubiquinone oxidoreductase B16.6 subunit|ND-B16.6-PA|ND-B16.6-PB|ND-B16.6-PC', map_location='5F2-5F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6349016, end_position_on_the_genomic_accession=6350401, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31579, CurrentID=0, Status='live', Symbol='kdn', Aliases='Dmel_CG3861, BEST:GM01832, CG 3861, CG3861, CS, Csyn, Dmel\\CG3861, l (1) G0030, l(1)G0030, l(1)G0033, l(1)G0159', description='knockdown', other_designations='CG3861-PA|CG3861-PB|citrate synthase|kdn-PA|kdn-PB|lethal (1) G0030', map_location='5F2-5F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6350629, end_position_on_the_genomic_accession=6361323, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31580, CurrentID=0, Status='live', Symbol='swa', Aliases='Dmel_CG3429, CG3429, Dmel\\CG3429, Swa, anon-WO0140519.70, fs(1)1497, fs(1)1502, fs(1)M11-999, fs(1)M13C82, fs(1)M44, fs(1)m44, sww', description='swallow', other_designations='CG3429-PA|swa-PA', map_location='1-15.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6363418, end_position_on_the_genomic_accession=6365708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31581, CurrentID=0, Status='live', Symbol='Marf', Aliases='Dmel_CG3869, CG3869, CG38869, Dmel\\CG3869, MARF-1, Mfn, Mfn-2, anon-WO0125274.3, dMFN, dMfn, dmfn, marf, mfn, mfn2, Marf', description='Mitochondrial assembly regulatory factor', other_designations='CG3869-PB|CG3869-PC|CG3869-PD|Marf-PB|Marf-PC|Marf-PD|drosophila mitofusin|mitochondria assembly regulatory factor|mitochondrial assembly regulatory factor|mitofusin|mitofusin/MARF', map_location='5F4-5F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6365949, end_position_on_the_genomic_accession=6370924, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31582, CurrentID=0, Status='live', Symbol='PpV', Aliases='Dmel_CG12217, CG12217, DmPpV-6A, Dmel\\CG12217, PP6, PPPV6A, PPV, PPV 6A[[C]], Ppv, dPP6, PpV', description='Protein phosphatase V', other_designations='CG12217-PA|CG12217-PB|CG12217-PC|CG12217-PD|PpV-PA|PpV-PB|PpV-PC|PpV-PD|protein phosphatase V', map_location='5F4-5F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6371065, end_position_on_the_genomic_accession=6372915, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31585, CurrentID=0, Status='live', Symbol='CG3918', Aliases='Dmel_ Dmel\\CG3918', description='CG3918 gene product from transcript CG3918-RA', other_designations='CG3918-PA', map_location='6B1-6B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6524464, end_position_on_the_genomic_accession=6526342, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31586, CurrentID=0, Status='live', Symbol='CG3342', Aliases='Dmel_ Dmel\\CG3342', description='CG3342 gene product from transcript CG3342-RA', other_designations='CG3342-PA', map_location='6B1-6B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6526106, end_position_on_the_genomic_accession=6527888, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31587, CurrentID=0, Status='live', Symbol='Spat', Aliases='Dmel_CG3926, CG3926, Dm-Spat, Dmel\\CG3926, spa, spat', description='Serine pyruvate aminotransferase', other_designations='CG3926-PA|Spat-PA|serine pyruvate aminotransferase', map_location='6B1-6B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6528948, end_position_on_the_genomic_accession=6530726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31588, CurrentID=0, Status='live', Symbol='RpL7A', Aliases='Dmel_CG3314, CG3314, Dmel\\CG3314, L7A, L7a, Rp L7A, RpL7a, Surf-3/rpL7a, Surf3/rpL7a', description='Ribosomal protein L7A', other_designations='CG3314-PA|CG3314-PD|CG3314-PE|CG3314-PF|CG3314-PG|RpL7A-PA|RpL7A-PD|RpL7A-PE|RpL7A-PF|RpL7A-PG|ribosomal protein L7A', map_location='6B1-6B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6530897, end_position_on_the_genomic_accession=6533348, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31589, CurrentID=0, Status='live', Symbol='dx', Aliases='Dmel_CG3929, CG3929, DMDELTEX, Dmel\\CG3929, Dx', description='deltex', other_designations='CG3929-PA|dx-PA', map_location='1-17.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6533999, end_position_on_the_genomic_accession=6538575, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31591, CurrentID=0, Status='live', Symbol='CG34417', Aliases='Dmel_ BcDNA:GH23906, CG3950, CG3960, Dmel\\CG34417, Dmel_CG3950, Dmel_CG3960', description='CG34417 gene product from transcript CG34417-RI', other_designations='CG34417-PC|CG34417-PD|CG34417-PF|CG34417-PH|CG34417-PI|CG34417-PJ|CG34417-PK|CG34417-PL|CG34417-PM|CG34417-PN|CG34417-PO|CG34417-PP|CG34417-PQ|CG34417-PR|CG34417-PS|CG34417-PT|CG34417-PU|CG34417-PV|CG34417-PW|CG34417-PX|CG34417-PY|uncharacterized protein', map_location='6B3-6C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6540771, end_position_on_the_genomic_accession=6589173, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  31592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31592, CurrentID=0, Status='live', Symbol='CG17717', Aliases='Dmel_ BcDNA:AT19932, Dmel\\CG17717', description='CG17717 gene product from transcript CG17717-RA', other_designations='CG17717-PA|CG17717-PB|uncharacterized protein', map_location='6C1-6C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6589290, end_position_on_the_genomic_accession=6590646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31593, CurrentID=0, Status='live', Symbol='Pat1', Aliases='Dmel_CG10695, CG10695, Dmel\\CG10695, PAT1, pat1', description='Protein interacting with APP tail-1', other_designations='CG10695-PA|CG10695-PB|CG10695-PC|Pat1-PA|Pat1-PB|Pat1-PC|protein interacting with APP tail-1', map_location='6C1-6C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6590776, end_position_on_the_genomic_accession=6596954, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31594, CurrentID=0, Status='live', Symbol='APC7', Aliases='Dmel_CG14444, Apc7, CG14444, Dmel\\CG14444', description='Anaphase Promoting Complex subunit 7', other_designations='APC7-PA|APC7-PB|CG14444-PA|CG14444-PB|anaphase promoting complex 7|anaphase promoting complex subunit 7', map_location='6C1-6C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6597088, end_position_on_the_genomic_accession=6599699, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31595, CurrentID=0, Status='live', Symbol='CG14443', Aliases='Dmel_ Dmel\\CG14443, cg14443', description='CG14443 gene product from transcript CG14443-RB', other_designations='CG14443-PB|uncharacterized protein', map_location='6C2-6C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6633959, end_position_on_the_genomic_accession=6636169, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31596, CurrentID=0, Status='live', Symbol='pigs', Aliases='Dmel_CG3973, CG3973, Dmel\\CG3973', description='pickled eggs', other_designations='CG3973-PA|CG3973-PB|CG3973-PC|pigs-PA|pigs-PB|pigs-PC', map_location='6C1-6C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6600062, end_position_on_the_genomic_accession=6653015, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31597, CurrentID=0, Status='live', Symbol='CG3226', Aliases='Dmel_ Dmel\\CG3226, Q9W3Y3', description='CG3226 gene product from transcript CG3226-RA', other_designations='CG3226-PA', map_location='6C3-6C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6653113, end_position_on_the_genomic_accession=6654248, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31598, CurrentID=0, Status='live', Symbol='Cdc7', Aliases='Dmel_CG32742, CG32742, CG8655, CG8656, CT25076, Dmel\\CG32742, L(1)G0148, anon-WO03040301.250, cdc7, cdc7-related, l(1)G0148, l(1)G0149, l(1)G0461', description='Cdc7 kinase', other_designations='CG32742-PA|CG32742-PB|Cdc7-PA|Cdc7-PB|lethal (1) G0148', map_location='6C3-6C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6654541, end_position_on_the_genomic_accession=6659709, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31600 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31600, CurrentID=0, Status='live', Symbol='CG3224', Aliases='Dmel_ Dmel\\CG3224', description='CG3224 gene product from transcript CG3224-RA', other_designations='CG3224-PA', map_location='6C3-6C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6660552, end_position_on_the_genomic_accession=6661462, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31601, CurrentID=0, Status='live', Symbol='Ctr1A', Aliases='Dmel_CG3977, BcDNA:RH42635, CG3977, DmCtr1A, Dmel\\CG3977, ctr1A', description='Copper transporter 1A', other_designations='CG3977-PA|CG3977-PB|CG3977-PC|CG3977-PD|Ctr1A-PA|Ctr1A-PB|Ctr1A-PC|Ctr1A-PD|copper transporter 1A', map_location='6C4-6C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6661647, end_position_on_the_genomic_accession=6668258, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31602, CurrentID=0, Status='live', Symbol='CG3198', Aliases='Dmel_ Dmel\\CG3198', description='CG3198 gene product from transcript CG3198-RA', other_designations='CG3198-PA', map_location='6C4-6C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6668539, end_position_on_the_genomic_accession=6673973, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31603, CurrentID=0, Status='live', Symbol='Mcm6', Aliases='Dmel_CG4039, CG4039, DmMCM6, DmeMCM6, Dmel\\CG4039, MCM6, McM6, dMcm6, fs(1)K1214, mcm6', description='Minichromosome maintenance 6', other_designations='CG4039-PA|Mcm6-PA|female sterile(1)K1214|minichromosome maintenance 6', map_location='1-18.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6674387, end_position_on_the_genomic_accession=6677176, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31604, CurrentID=0, Status='live', Symbol='ND-ASHI', Aliases='Dmel_CG3192, CG3192, Dmel\\CG3192, NIAM', description='NADH dehydrogenase (ubiquinone) ASHI subunit', other_designations='CG3192-PA|CG3192-PB|CG3192-PC|NADH ubiquinone oxidoreductase ASHI|NADH-ubiquinone oxidoreductase ASHI subunit|NADH:ubiquinone oxidoreductase ASHI subunit|ND-ASHI-PA|ND-ASHI-PB|ND-ASHI-PC', map_location='6C5-6C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6677262, end_position_on_the_genomic_accession=6679118, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31605, CurrentID=0, Status='live', Symbol='l(1)G0255', Aliases='Dmel_CG4094, CG 4094, CG4094, Dmel\\CG4094, FBgn0029889, l(1)G0290', description='lethal (1) G0255', other_designations='CG4094-PA|CG4094-PB|l(1)G0255-PA|l(1)G0255-PB', map_location='6C5-6C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6678424, end_position_on_the_genomic_accession=6681845, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31606, CurrentID=0, Status='live', Symbol='CG4095', Aliases='Dmel_ Dmel\\CG4095', description='CG4095 gene product from transcript CG4095-RA', other_designations='CG4095-PA', map_location='6C5-6C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6681767, end_position_on_the_genomic_accession=6683696, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31607, CurrentID=0, Status='live', Symbol='Pink1', Aliases='Dmel_CG4523, BEST:GH23468, CG4523, Dmel\\CG4523, PINK, PINK1, dPINK1, dPink1, pink1', description='PTEN-induced putative kinase 1', other_designations='CG4523-PA|CG4523-PB|CG4523-PC|CG4523-PD|CG4523-PE|CG4523-PF|CG4523-PG|CG4523-PH|CG4523-PI|PTEN induced putative kinase 1|PTEN-induced kinase 1|Pink1-PA|Pink1-PB|Pink1-PC|Pink1-PD|Pink1-PE|Pink1-PF|Pink1-PG|Pink1-PH|Pink1-PI|phosphatase and tensin homolog-induced putative kinase 1', map_location='6C6-6C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6683798, end_position_on_the_genomic_accession=6687302, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31608, CurrentID=0, Status='live', Symbol='CG3184', Aliases='Dmel_ Dmel\\CG3184', description='CG3184 gene product from transcript CG3184-RA', other_designations='CG3184-PA', map_location='6C6-6C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6687596, end_position_on_the_genomic_accession=6688774, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31609, CurrentID=0, Status='live', Symbol='CG14442', Aliases='Dmel_ Dmel\\CG14442', description='CG14442 gene product from transcript CG14442-RA', other_designations='CG14442-PA|CG14442-PB|CG14442-PC|CG14442-PD|CG14442-PE|uncharacterized protein', map_location='6C6-6C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6689296, end_position_on_the_genomic_accession=6695007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31610, CurrentID=0, Status='live', Symbol='CG14440', Aliases='Dmel_ Dmel\\CG14440', description='CG14440 gene product from transcript CG14440-RA', other_designations='CG14440-PA|CG14440-PB|uncharacterized protein', map_location='6C7-6C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6695866, end_position_on_the_genomic_accession=6698567, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31611, CurrentID=0, Status='live', Symbol='CG14441', Aliases='Dmel_ Dmel\\CG14441', description='CG14441 gene product from transcript CG14441-RA', other_designations='CG14441-PA|CG14441-PB|CG14441-PC|CG14441-PD|uncharacterized protein', map_location='6C7-6C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6699508, end_position_on_the_genomic_accession=6719389, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31612, CurrentID=0, Status='live', Symbol='CG3168', Aliases='Dmel_ BEST:LP06294, Dmel\\CG3168, LP06294.5prime', description='CG3168 gene product from transcript CG3168-RG', other_designations='CG3168-PA|CG3168-PB|CG3168-PC|CG3168-PD|CG3168-PE|CG3168-PF|CG3168-PG', map_location='6C9-6C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6720004, end_position_on_the_genomic_accession=6739986, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31613, CurrentID=0, Status='live', Symbol='RpL17', Aliases='Dmel_CG3203, CG3203, Dmel\\CG3203, Rpl17, anon-EST:Posey276', description='Ribosomal protein L17', other_designations='CG3203-PA|CG3203-PB|CG3203-PC|CG3203-PD|CG3203-PE|RpL17-PA|RpL17-PB|RpL17-PC|RpL17-PD|RpL17-PE|ribosomal protein L17', map_location='6C10-6C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6741537, end_position_on_the_genomic_accession=6744238, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31614 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31614, CurrentID=0, Status='live', Symbol='CG14439', Aliases='Dmel_ Dmel\\CG14439', description='CG14439 gene product from transcript CG14439-RB', other_designations='CG14439-PA|CG14439-PB|CG14439-PC|uncharacterized protein', map_location='6C11-6C12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6747544, end_position_on_the_genomic_accession=6755752, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31615, CurrentID=0, Status='live', Symbol='CG14438', Aliases='Dmel_ Dmel\\CG14438', description='CG14438 gene product from transcript CG14438-RA', other_designations='CG14438-PA|CG14438-PB|CG14438-PC|uncharacterized protein', map_location='6C12-6C13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6756322, end_position_on_the_genomic_accession=6770712, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31617, CurrentID=0, Status='live', Symbol='shf', Aliases='Dmel_CG3135, CG3134, CG3135, CT10514, D-Wif, DmWIF, DmWif, Dmel\\CG3135, Shf-ov, shf', description='shifted', other_designations='CG3135-PA|CG3135-PB|CG3135-PC|CG3135-PD|WNT inhibitory factor|shf-PA|shf-PB|shf-PC|shf-PD', map_location='1-17.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6771463, end_position_on_the_genomic_accession=6779659, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31618 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31618, CurrentID=0, Status='live', Symbol='C3G', Aliases='Dmel_CG42328, C(3)G, CG14436, CG3126, CG42328, DC3-G, DC3G, Dmel\\CG42328, Dmel_CG14436, Dmel_CG3126, c(3)G', description='C3G ortholog (H. sapiens)', other_designations='C3G ortholog|C3G-PC|C3G-PD|C3G-PE|C3G-PF|C3G-PG|C3G-PH|C3G-PI|C3G-PJ|C3G-PK|C3G-PL|CG42328-PC|CG42328-PD|CG42328-PE|CG42328-PF|CG42328-PG|CG42328-PH|CG42328-PI|CG42328-PJ|CG42328-PK|CG42328-PL', map_location='6D2-6D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6780790, end_position_on_the_genomic_accession=6803593, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  31620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31620, CurrentID=0, Status='live', Symbol='pod1', Aliases='Dmel_CG4532, BEST:LD37601, BcDNA:LD08951, CG4532, Dmel\\CG4532, D POD-1, Pod1, coronin, dmPod, pod-1', description='CG4532 gene product from transcript CG4532-RG', other_designations='CG4532-PA|CG4532-PB|CG4532-PC|CG4532-PD|CG4532-PE|CG4532-PF|CG4532-PG|POD-1/coronin|coronin|pod1-PA|pod1-PB|pod1-PC|pod1-PD|pod1-PE|pod1-PF|pod1-PG', map_location='6D2-6D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6803684, end_position_on_the_genomic_accession=6813830, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  31621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31621, CurrentID=0, Status='live', Symbol='iav', Aliases='Dmel_CG4536, CG4536, CT14708, DmIav, Dmel\\CG4536, IAV, Iav, TRPV, hypoB, osm-9', description='inactive', other_designations='CG4536-PA|iav-PA', map_location='1-18.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6813978, end_position_on_the_genomic_accession=6818156, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31622, CurrentID=0, Status='live', Symbol='Nf-YC', Aliases='Dmel_CG3075, CG3075, Dmel\\CG3075, NF-YC, nf-yc', description='Nuclear factor Y-box C', other_designations='CG3075-PA|NF-YC-like|Nf-YC-PA|nuclear factor Y-box C', map_location='6D3-6D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6818178, end_position_on_the_genomic_accession=6820919, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31623, CurrentID=0, Status='live', Symbol='xit', Aliases='Dmel_CG4542, CG4542, Dalg8, Dmel\\CG4542, dAlg8', description='xiantuan', other_designations='CG4542-PA|xit-PA', map_location='6D3-6D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6821016, end_position_on_the_genomic_accession=6822840, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31624, CurrentID=0, Status='live', Symbol='Atx-1', Aliases='Dmel_CG4547, CG4547, Dmel\\CG4547, dAtx-1, dAtx1', description='Ataxin 1', other_designations='Ataxin1|Atx-1-PB|CG4547-PB|ataxin 1|ataxin-1', map_location='6D3-6D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6823204, end_position_on_the_genomic_accession=6824203, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31625, CurrentID=0, Status='live', Symbol='nonC', Aliases='Dmel_CG32743, CG32743, CG4549, CG4555, Dmel\\CG32743, SMG1, Smg, Smg1, anon-EST:Liang-1.24, anon-WO0170980.133, clone 1.24, dSmg1, rst(1a)cyr, smg1', description='no-on-and-no-off transient C', other_designations='CG32743-PA|CG32743-PB|FRAP-related|Smg1|no on or off transient C|nonC-PA|nonC-PB|resistance 1a cyromazine', map_location='6D3-6D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6824496, end_position_on_the_genomic_accession=6838122, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31626, CurrentID=0, Status='live', Symbol='CG12796', Aliases='Dmel_ Dmel\\CG12796', description='CG12796 gene product from transcript CG12796-RA', other_designations='CG12796-PA|uncharacterized protein', map_location='6D4-6D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6831397, end_position_on_the_genomic_accession=6832684, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31628, CurrentID=0, Status='live', Symbol='CG14435', Aliases='Dmel_ Dmel\\CG14435', description='CG14435 gene product from transcript CG14435-RC', other_designations='CG14435-PA|CG14435-PC|uncharacterized protein', map_location='6D4-6D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6838174, end_position_on_the_genomic_accession=6843921, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31629, CurrentID=0, Status='live', Symbol='CG4557', Aliases='Dmel_ AAF46211, Dmel\\CG4557', description='CG4557 gene product from transcript CG4557-RA', other_designations='CG4557-PA', map_location='6D5-6D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6844341, end_position_on_the_genomic_accession=6847847, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31630, CurrentID=0, Status='live', Symbol='Cht11', Aliases='Dmel_CG3044, CG3044, Dmel\\CG3044', description='CG3044 gene product from transcript CG3044-RA', other_designations='CG3044-PA|Cht11-PA', map_location='6D6-6D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6847877, end_position_on_the_genomic_accession=6849517, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31631, CurrentID=0, Status='live', Symbol='CG4558', Aliases='Dmel_ Dmel\\CG4558', description='CG4558 gene product from transcript CG4558-RA', other_designations='CG4558-PA|CG4558-PB', map_location='6D6-6D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6849770, end_position_on_the_genomic_accession=6851884, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31632, CurrentID=0, Status='live', Symbol='CG14434', Aliases='Dmel_ Dmel\\CG14434', description='CG14434 gene product from transcript CG14434-RC', other_designations='CG14434-PA|CG14434-PB|CG14434-PC|uncharacterized protein', map_location='6D7-6E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6851909, end_position_on_the_genomic_accession=6867194, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31638, CurrentID=0, Status='live', Symbol='CG32732', Aliases='Dmel_ CG12542, CG3042, Dmel\\CG32732', description='CG32732 gene product from transcript CG32732-RA', other_designations='CG32732-PA|uncharacterized protein', map_location='6E3-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6958032, end_position_on_the_genomic_accession=6968750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31639, CurrentID=0, Status='live', Symbol='CG14431', Aliases='Dmel_ Dmel\\CG14431', description='CG14431 gene product from transcript CG14431-RC', other_designations='CG14431-PB|CG14431-PC|uncharacterized protein', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6908277, end_position_on_the_genomic_accession=6968829, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31641, CurrentID=0, Status='live', Symbol='CG4586', Aliases='Dmel_ Dmel\\CG4586', description='CG4586 gene product from transcript CG4586-RA', other_designations='CG4586-PA|acyl-coenzyme A oxidase', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6969918, end_position_on_the_genomic_accession=6972265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31642, CurrentID=0, Status='live', Symbol='CG3040', Aliases='Dmel_ CT9706, Dmel\\CG3040', description='CG3040 gene product from transcript CG3040-RA', other_designations='CG3040-PA', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6972324, end_position_on_the_genomic_accession=6973393, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31644, CurrentID=0, Status='live', Symbol='bou', Aliases='Dmel_CG14430, BcDNA:RE28342, CG14430, Dmel\\CG14430, JNXL4, jnL4, l(1)6Ea, l(1)jnL4', description='boudin', other_designations='CG14430-PA|bou-PA|lethal (1) 6Ea', map_location='1-18.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6986486, end_position_on_the_genomic_accession=6987188, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31645, CurrentID=0, Status='live', Symbol='Inx6', Aliases='Dmel_CG17063, CG17063, Dm-inx6, Dmel\\CG17063, FBpp0077031, inx-6, inx6, prp6', description='Innexin 6', other_designations='CG17063-PA|Inx6-PA|innexin 6|inx6|pas related protein 6', map_location='19C1-19C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=20267821, end_position_on_the_genomic_accession=20269769, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31646, CurrentID=0, Status='live', Symbol='Inx2', Aliases='Dmel_CG4590, CG4590, D-inx-2, Dm-inx, Dm-inx2, Dmel\\CG4590, FBpp0291074, INX-2, Ix2, inx-2, inx2, kropf, l(1)G0035, l(1)G0036, l(1)G0043, l(1)G0059, l(1)G0118, l(1)G0157, l(1)G0317, l(1)G0364, prp33', description='Innexin 2', other_designations='CG4590-PA|CG4590-PB|CG4590-PC|CG4590-PD|Innexin-2|Inx2-PA|Inx2-PB|Inx2-PC|Inx2-PD|innexin|innexin 2|innexin2|pas related protein 33', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=6998672, end_position_on_the_genomic_accession=7004151, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31647, CurrentID=0, Status='live', Symbol='cm', Aliases='Dmel_CG3035, AP-3, AP-3mu, AP3mu3, CG3035, CG32731, Dmel\\CG3035, Dmu3, Mu3, ap-3, apm3, cg3035, l(1)VEM119, mu3', description='carmine', other_designations='CG3035-PA|CG3035-PB|adaptor protein complex 3, mu subunit|cm-PA|cm-PB', map_location='1-18.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7009006, end_position_on_the_genomic_accession=7011145, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31648, CurrentID=0, Status='live', Symbol='CG3032', Aliases='Dmel_ Dmel\\CG3032', description='CG3032 gene product from transcript CG3032-RB', other_designations='CG3032-PA|CG3032-PB', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7012239, end_position_on_the_genomic_accession=7013881, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31649, CurrentID=0, Status='live', Symbol='CG4593', Aliases='Dmel_ Dmel\\CG4593', description='CG4593 gene product from transcript CG4593-RA', other_designations='CG4593-PA|CG4593-PB', map_location='6E4-6E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7013887, end_position_on_the_genomic_accession=7015009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31650, CurrentID=0, Status='live', Symbol='CG12541', Aliases='Dmel_ Dmel\\CG12541', description='CG12541 gene product from transcript CG12541-RE', other_designations='CG12541-PD|CG12541-PE|uncharacterized protein', map_location='6E4-6F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7021608, end_position_on_the_genomic_accession=7066052, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31651, CurrentID=0, Status='live', Symbol='CG14427', Aliases='Dmel_ Dmel\\CG14427', description='CG14427 gene product from transcript CG14427-RA', other_designations='CG14427-PA|uncharacterized protein', map_location='6F1-6F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7040603, end_position_on_the_genomic_accession=7041769, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31652, CurrentID=0, Status='live', Symbol='nullo', Aliases='Dmel_CG14426, BcDNA:RE47733, CG14426, Dmel\\CG14426, Nullo', description='CG14426 gene product from transcript CG14426-RB', other_designations='CG14426-PB|nullo-PB', map_location='6F1-6F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7049476, end_position_on_the_genomic_accession=7050321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31653 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31653, CurrentID=0, Status='live', Symbol='CG4607', Aliases='Dmel_ Dmel\\CG4607', description='CG4607 gene product from transcript CG4607-RA', other_designations='CG4607-PA|CG4607-PB', map_location='6F1-6F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7069413, end_position_on_the_genomic_accession=7072689, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31656, CurrentID=0, Status='live', Symbol='CG4615', Aliases='Dmel_ Dmel\\CG4615', description='CG4615 gene product from transcript CG4615-RA', other_designations='CG4615-PA', map_location='6F5-6F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7098578, end_position_on_the_genomic_accession=7102300, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31657, CurrentID=0, Status='live', Symbol='CG4617', Aliases='Dmel_ Dmel\\CG4617', description='CG4617 gene product from transcript CG4617-RA', other_designations='CG4617-PA|CG4617-PC', map_location='6F5-6F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7102555, end_position_on_the_genomic_accession=7104491, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31658 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31658, CurrentID=0, Status='live', Symbol='CG8300', Aliases='Dmel_ Dmel\\CG8300', description='CG8300 gene product from transcript CG8300-RA', other_designations='CG8300-PA', map_location='6F5-6F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7104885, end_position_on_the_genomic_accession=7106345, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31659, CurrentID=0, Status='live', Symbol='fz4', Aliases='Dmel_CG4626, CG4626, DFz4, Dfz4, Dm Fz4, Dmel\\CG4626, Fz4, anon-WO0170980.10, anon-WO0170980.11', description='frizzled 4', other_designations='CG4626-PA|CG4626-PB|fz4-PA|fz4-PB', map_location='6F5-6F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7106272, end_position_on_the_genomic_accession=7110321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31660, CurrentID=0, Status='live', Symbol='CG9650', Aliases='Dmel_ Dmel\\CG9650', description='CG9650 gene product from transcript CG9650-RG', other_designations='CG9650-PG|CG9650-PH|CG9650-PI|CG9650-PJ', map_location='7A2-7A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7137723, end_position_on_the_genomic_accession=7241932, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31661, CurrentID=0, Status='live', Symbol='CG1958', Aliases='Dmel_ Dmel\\CG1958', description='CG1958 gene product from transcript CG1958-RA', other_designations='CG1958-PA', map_location='7A7-7A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7266904, end_position_on_the_genomic_accession=7267883, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31662, CurrentID=0, Status='live', Symbol='CG1677', Aliases='Dmel_ Dmel\\CG1677', description='CG1677 gene product from transcript CG1677-RB', other_designations='CG1677-PA|CG1677-PB|CG1677-PC|CG1677-PD', map_location='7A8-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7281089, end_position_on_the_genomic_accession=7405797, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31663, CurrentID=0, Status='live', Symbol='CG2059', Aliases='Dmel_ Dmel\\CG2059', description='CG2059 gene product from transcript CG2059-RB', other_designations='CG2059-PA|CG2059-PB', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7286358, end_position_on_the_genomic_accession=7288117, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31664, CurrentID=0, Status='live', Symbol='unc-119', Aliases='Dmel_CG1659, CG1659, DmUNC-119, Dmel\\CG1659, Unc-119, Unc119, anon-EST:Posey123, dm ', description='CG1659 gene product from transcript CG1659-RC', other_designations='CG1659-PA|CG1659-PC|unc-119-PA|unc-119-PC', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7288367, end_position_on_the_genomic_accession=7291298, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31665, CurrentID=0, Status='live', Symbol='brk', Aliases='Dmel_CG9653, Brk, CG9653, Dm brk, Dmel\\CG9653, ssg-1', description='brinker', other_designations='CG9653-PA|brk-PA|small salivary glands-1', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7307939, end_position_on_the_genomic_accession=7311132, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31666, CurrentID=0, Status='live', Symbol='Atg5', Aliases='Dmel_CG1643, ATG5, CG1643, DmAtg5, Dmel\\CG1643, atg5', description='Autophagy-related 5', other_designations='Atg5-PA|CG1643-PA|autophagy-specific gene 5', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7314622, end_position_on_the_genomic_accession=7323770, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31667 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31667, CurrentID=0, Status='live', Symbol='Dok', Aliases='Dmel_CG2079, CG2079, Ddok, Dmel\\CG2079, l(1)G0331', description='Downstream of kinase', other_designations='CG2079-PA|Dok-PA|downstream of kinase|lethal (1) G0331', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7324214, end_position_on_the_genomic_accession=7328806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31668, CurrentID=0, Status='live', Symbol='CG18155', Aliases='Dmel_ Dmel\\CG18155, FBgn0029945', description='CG18155 gene product from transcript CG18155-RB', other_designations='CG18155-PB|uncharacterized protein', map_location='7B1-7B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7329046, end_position_on_the_genomic_accession=7331255, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31669 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31669, CurrentID=0, Status='live', Symbol='CBP', Aliases='Dmel_CG1435, BcDNA.GH12350, BcDNA:GH12350, CG1435, Dmel\\CG1435, cbp, dCBP', description='sarcoplasmic calcium-binding protein', other_designations='CBP-PA|CBP-PB|CBP-PC|CG1435-PA|CG1435-PB|CG1435-PC|Ca2+ - binding protein|calcium binding protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7337352, end_position_on_the_genomic_accession=7342731, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31670, CurrentID=0, Status='live', Symbol='CG15034', Aliases='Dmel_ Dmel\\CG15034', description='CG15034 gene product from transcript CG15034-RB', other_designations='CG15034-PA|CG15034-PB|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7346259, end_position_on_the_genomic_accession=7347506, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31671, CurrentID=0, Status='live', Symbol='CG1999', Aliases='Dmel_ Dmel\\CG1999, anon-WO0140519.160', description='CG1999 gene product from transcript CG1999-RA', other_designations='CG1999-PA', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7358796, end_position_on_the_genomic_accession=7360058, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31672, CurrentID=0, Status='live', Symbol='CheA7a', Aliases='Dmel_CG15033, CG15033, Dmel\\CG15033', description='Chemosensory protein A 7a', other_designations='CG15033-PA|CheA7a-PA|chemosensory protein A 7a', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7378140, end_position_on_the_genomic_accession=7378953, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31673, CurrentID=0, Status='live', Symbol='CG15035', Aliases='Dmel_ Dmel\\CG15035', description='CG15035 gene product from transcript CG15035-RA', other_designations='CG15035-PA|uncharacterized protein', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7401698, end_position_on_the_genomic_accession=7403174, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31674, CurrentID=0, Status='live', Symbol='CG9657', Aliases='Dmel_ Dmel\\CG9657, cg9657, dSLC5A9', description='CG9657 gene product from transcript CG9657-RB', other_designations='CG9657-PA|CG9657-PB', map_location='7B2-7B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7418594, end_position_on_the_genomic_accession=7422142, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31675, CurrentID=0, Status='live', Symbol='CG11369', Aliases='Dmel_ Dmel\\CG11369', description='CG11369 gene product from transcript CG11369-RA', other_designations='CG11369-PA|uncharacterized protein', map_location='7B4-7B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7574290, end_position_on_the_genomic_accession=7576767, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31676, CurrentID=0, Status='live', Symbol='CG12689', Aliases='Dmel_ Dmel\\CG12689', description='CG12689 gene product from transcript CG12689-RA', other_designations='CG12689-PA|uncharacterized protein', map_location='7B4-7B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7585303, end_position_on_the_genomic_accession=7586131, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31678, CurrentID=0, Status='live', Symbol='CHES-1-like', Aliases='Dmel_CG12690, CG12690, Ches-1-like, Ches1, Dmches-1, Dmel\\CG12690, FoxN, anon-EST:Posey7, ches1', description='Checkpoint suppressor homologue', other_designations='CG12690-PA|CG12690-PB|CG12690-PC|CG12690-PD|CHES-1-like-PA|CHES-1-like-PB|CHES-1-like-PC|CHES-1-like-PD|checkpoint suppressor homologue|ches-1-like', map_location='7B6-7B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7682338, end_position_on_the_genomic_accession=7709385, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31679, CurrentID=0, Status='live', Symbol='CG15478', Aliases='Dmel_ Dmel\\CG15478', description='CG15478 gene product from transcript CG15478-RA', other_designations='CG15478-PA|CG15478-PB|uncharacterized protein', map_location='7B6-7B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7712783, end_position_on_the_genomic_accession=7717932, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31680, CurrentID=0, Status='live', Symbol='Hira', Aliases='Dmel_CG12153, CG12153, Dhh, Dmel\\CG12153, HIRA, d dhira, hira, ssm', description='CG12153 gene product from transcript CG12153-RA', other_designations='CG12153-PA|Hira-PA|drosophila hira homolog|drosophila homolog of HIRA|hira|sesame', map_location='7B7-7B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7722119, end_position_on_the_genomic_accession=7725917, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31681 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31681, CurrentID=0, Status='live', Symbol='NELF-B', Aliases='Dmel_CG32721, BcDNA:GH10333, CG12152, CG32721, Dmel\\CG32721, NELF', description='CG32721 gene product from transcript CG32721-RA', other_designations='CG32721-PA|NELF-B-PA', map_location='7B7-7B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7725996, end_position_on_the_genomic_accession=7728361, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31682, CurrentID=0, Status='live', Symbol='CG12155', Aliases='Dmel_ Dmel\\CG12155', description='CG12155 gene product from transcript CG12155-RA', other_designations='CG12155-PA|uncharacterized protein', map_location='7B7-7B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7728755, end_position_on_the_genomic_accession=7731729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31683, CurrentID=0, Status='live', Symbol='Pdp', Aliases='Dmel_CG12151, 153332_at, CG12151, Dmel\\CG12151, PDP, pdp', description='Pyruvate dehydrogenase phosphatase', other_designations='CG12151-PA|CG12151-PB|Pdp-PA|Pdp-PB|pyruvate dehydrogenase phosphatase', map_location='7B7-7B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7731779, end_position_on_the_genomic_accession=7734651, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31684, CurrentID=0, Status='live', Symbol='Rab39', Aliases='Dmel_CG12156, 24640395, AAF46271, CG12156, Dm Dmel\\CG12156, rab39', description='CG12156 gene product from transcript CG12156-RB', other_designations='CG12156-PA|CG12156-PB|Rab39-PA|Rab39-PB', map_location='7B7-7B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7734923, end_position_on_the_genomic_accession=7736756, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31686, CurrentID=0, Status='live', Symbol='Ir7a', Aliases='Dmel_CG12663, CG12663, DmelIR7a, Dmel\\CG12663, IR7a', description='Ionotropic receptor 7a', other_designations='CG12663-PA|Ir7a-PA|ionotropic receptor 7a', map_location='7B8-7B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7764982, end_position_on_the_genomic_accession=7766826, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31687, CurrentID=0, Status='live', Symbol='CG1402', Aliases='Dmel_ Dmel\\CG1402', description='CG1402 gene product from transcript CG1402-RC', other_designations='CG1402-PB|CG1402-PC', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7826197, end_position_on_the_genomic_accession=7842679, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31688, CurrentID=0, Status='live', Symbol='CG10920', Aliases='Dmel_ Dmel\\CG10920', description='CG10920 gene product from transcript CG10920-RA', other_designations='CG10920-PA|uncharacterized protein', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7852481, end_position_on_the_genomic_accession=7854244, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31689, CurrentID=0, Status='live', Symbol='CG1409', Aliases='Dmel_ BcDNA:AT29287, D.M.BLPp, Dmel\\CG1409', description='CG1409 gene product from transcript CG1409-RB', other_designations='CG1409-PB', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7858173, end_position_on_the_genomic_accession=7858794, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31690, CurrentID=0, Status='live', Symbol='Ir7b', Aliases='Dmel_CG15326, CG15326, DmelIR7b, Dmel\\CG15326, IR7b', description='Ionotropic receptor 7b', other_designations='CG15326-PB|Ir7b-PB|ionotropic receptor 7b', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7864948, end_position_on_the_genomic_accession=7866968, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31691, CurrentID=0, Status='live', Symbol='dec-1', Aliases='Dmel_CG2175, 39051, CG2175, Dmel\\CG2175, Fc, Fc1, Fcp7C, dec, dec1, dec[[1]], fs(1)11-549, fs(1)12-1873, fs(1)12-2514, fs(1)12-3512, fs(1)12-365, fs(1)12-3907, fs(1)12-403, fs(1)12-4860, fs(1)13-453, fs(1)13C-57, fs(1)13C-73, fs(1)14-963, fs(1)14A-114, fs(1)14E19, fs(1)1501, fs(1)384, fs(1)5, fs(1)A257, fs(1)A384, fs(1)C1, fs(1)M102, fs(1)M1021', description='defective chorion 1', other_designations='CG2175-PA|CG2175-PB|CG2175-PC|CG2175-PE|CG2175-PF|CG2175-PG|dec-1-PA|dec-1-PB|dec-1-PC|dec-1-PE|dec-1-PF|dec-1-PG|defective chorion|follicle cell protein 7C', map_location='1-20.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7867161, end_position_on_the_genomic_accession=7873222, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31692, CurrentID=0, Status='live', Symbol='Ir7c', Aliases='Dmel_CG15324, CG15324, CT35324, DmelIR7c, Dmel\\CG15324, IR7c', description='Ionotropic receptor 7c', other_designations='CG15324-PA|CG15324-PB|Ir7c-PA|Ir7c-PB|ionotropic receptor 7c', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7874394, end_position_on_the_genomic_accession=7876847, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31694, CurrentID=0, Status='live', Symbol='Ir7g', Aliases='Dmel_CG15325, CG15325, DmelIR7g, Dmel\\CG15325, IR7g', description='Ionotropic receptor 7g', other_designations='CG15325-PA|Ir7g-PA|ionotropic receptor 7g', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7884660, end_position_on_the_genomic_accession=7886976, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31695, CurrentID=0, Status='live', Symbol='CG10932', Aliases='Dmel_ 154229_at, Dmel\\CG10932', description='CG10932 gene product from transcript CG10932-RB', other_designations='CG10932-PA|CG10932-PB|uncharacterized protein', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7887130, end_position_on_the_genomic_accession=7888785, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31696, CurrentID=0, Status='live', Symbol='Nek2', Aliases='Dmel_CG17256, CG17256, Dm  DmNek2, Dmel\\CG17256, NEK2, dNek2', description='CG17256 gene product from transcript CG17256-RA', other_designations='CG17256-PA|Nek2-PA', map_location='7C1-7C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7889066, end_position_on_the_genomic_accession=7891646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31697 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31697, CurrentID=0, Status='live', Symbol='ND-MNLL', Aliases='Dmel_CG18624, CG15328, CG18624, Dmel\\CG18624, Dromel_CG18624_FBtr0071140_mORF, NINM', description='NADH dehydrogenase (ubiquinone) MNLL subunit', other_designations='CG18624-PA|CG18624-PB|CG18624-PC|CG18624-PD|CG18624-PE|NADH ubiquinone oxidoreductase MNLL|NADH-ubiquinone oxidoreductase MNLL subunit|NADH:ubiquinone oxidoreductase MNLL subunit|ND-MNLL-PA|ND-MNLL-PB|ND-MNLL-PC|ND-MNLL-PD|ND-MNLL-PE', map_location='7C2-7C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7893053, end_position_on_the_genomic_accession=7893798, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31698, CurrentID=0, Status='live', Symbol='p115', Aliases='Dmel_CG1422, CG1422, Dmel\\CG1422, d l(1)G0306', description='CG1422 gene product from transcript CG1422-RB', other_designations='CG1422-PA|CG1422-PB|p115-PA|p115-PB', map_location='7C2-7C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7894031, end_position_on_the_genomic_accession=7896991, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31700, CurrentID=0, Status='live', Symbol='RpS6', Aliases='Dmel_CG10944, (Rp)S6, CG10944, DS6, Dmel\\CG10944, M(1)7BC, M(1)7C, RPS6, Rp S6, Rps6, S6, air8, air[8], anon-WO02059370.61, hen, l(1)air8, l(1)air[8], l(1)hen, pp30, rpS6, rps6', description='Ribosomal protein S6', other_designations='CG10944-PA|CG10944-PB|RpS6-PA|RpS6-PB|S6 ribosomal protein|hemocytes enlarged|lethal(1)aberrant immune response[8]|lethal-(1)-haemocytes enlarged|lethal-aberrant-immune-response-8|minute|minute (1) 7BC|ribosomal protein S6', map_location='7C1-7C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7899624, end_position_on_the_genomic_accession=7901331, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31701, CurrentID=0, Status='live', Symbol='bys', Aliases='Dmel_CG1430, CG1430, Dmel\\CG1430', description='by S6', other_designations='CG1430-PA|bys-PA|bystin', map_location='1-20.45 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7901880, end_position_on_the_genomic_accession=7903643, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31702 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31702, CurrentID=0, Status='live', Symbol='dpr14', Aliases='Dmel_CG10946, CG10946, CT30653, Dmel\\CG10946, Dpr-14, Dpr14', description='defective proboscis extension response 14', other_designations='CG10946-PA|CG10946-PB|dpr14-PA|dpr14-PB', map_location='7C2-7C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7903704, end_position_on_the_genomic_accession=7908286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31703, CurrentID=0, Status='live', Symbol='CG1444', Aliases='Dmel_ Dmel\\CG1444, KAR', description='CG1444 gene product from transcript CG1444-RA', other_designations='CG1444-PA', map_location='7C2-7C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7908931, end_position_on_the_genomic_accession=7911885, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31704, CurrentID=0, Status='live', Symbol='snz', Aliases='Dmel_CG1514, CG1514, DmSNX25, Dmel\\CG1514, Snz', description='snazarus', other_designations='CG1514-PA|CG1514-PB|snz-PA|snz-PB', map_location='7C2-7C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7912089, end_position_on_the_genomic_accession=7918444, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31705, CurrentID=0, Status='live', Symbol='hdm', Aliases='Dmel_CG15329, CG15329, Dmel\\CG15329', description="hold'em", other_designations='CG15329-PA|hdm-PA', map_location='7C3-7C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7918517, end_position_on_the_genomic_accession=7920131, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31706, CurrentID=0, Status='live', Symbol='Ykt6', Aliases='Dmel_CG1515, CG1515, Dmel\\CG1515, YKT6p, cg1515, l(1)G0155, Ykt6', description='YKT6 ortholog (S. cerevisiae)', other_designations='CG1515-PA|YKT6 ortholog|Ykt6-PA|lethal (1) G0155', map_location='7C3-7C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7920409, end_position_on_the_genomic_accession=7921966, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31707, CurrentID=0, Status='live', Symbol='CG10777', Aliases='Dmel_ DmRH1, Dmel\\CG10777, cg10777', description='CG10777 gene product from transcript CG10777-RB', other_designations='CG10777-PB|CG10777-PC|uncharacterized protein', map_location='7C3-7C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7922158, end_position_on_the_genomic_accession=7931382, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31708, CurrentID=0, Status='live', Symbol='CG10778', Aliases='Dmel_ Dmel\\CG10778', description='CG10778 gene product from transcript CG10778-RB', other_designations='CG10778-PA|CG10778-PB|uncharacterized protein', map_location='7C4-7C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7931745, end_position_on_the_genomic_accession=7933351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31712 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31712, CurrentID=0, Status='live', Symbol='l(1)G0193', Aliases='Dmel_CG2206, CG2206, Dmel\\CG2206, l(1)G0327', description='lethal (1) G0193', other_designations='CG2206-PA|CG2206-PB|l(1)G0193-PA|l(1)G0193-PB', map_location='7C9-7C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7941478, end_position_on_the_genomic_accession=7946908, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31713, CurrentID=0, Status='live', Symbol='Ubr3', Aliases='Dmel_CG42593, CG1530, CG1531, CG42593, Dmel\\CG42593, Dmel_CG1530, Dmel_CG1531, ubr3', description='Ubr3 ubiquitin ligase', other_designations='CG42593-PA|CG42593-PB|CG42593-PC|Ubr3-PA|Ubr3-PB|Ubr3-PC', map_location='7C9-7D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7935896, end_position_on_the_genomic_accession=7953883, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31714, CurrentID=0, Status='live', Symbol='CG15332', Aliases='Dmel_ Dmel\\CG15332', description='CG15332 gene product from transcript CG15332-RA', other_designations='CG15332-PA|uncharacterized protein', map_location='7D1-7D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7953852, end_position_on_the_genomic_accession=7955990, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31715, CurrentID=0, Status='live', Symbol='CG15330', Aliases='Dmel_ Dmel\\CG15330', description='CG15330 gene product from transcript CG15330-RA', other_designations='CG15330-PA|uncharacterized protein', map_location='7D1-7D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7956064, end_position_on_the_genomic_accession=7956702, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31716, CurrentID=0, Status='live', Symbol='sws', Aliases='Dmel_CG2212, CG2212, Dmel\\CG2212, PNPLA6, Sws, olfE', description='swiss cheese', other_designations='CG2212-PA|CG2212-PB|CG2212-PC|olfactory E|olfactory-E|swiss-cheese|swisscheese|sws-PA|sws-PB|sws-PC', map_location='1-22 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7956820, end_position_on_the_genomic_accession=7968236, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  31717 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31717, CurrentID=0, Status='live', Symbol='sn', Aliases='Dmel_CG32858, CG15331, CG1536, CG32858, Dmel\\CG32858, SN, Sn, fs(1)A1057, fs(1)K1421, fs(1)K418, fs(1)K473, fs(1)K743, fs(1)M45', description='singed', other_designations='CG32858-PA|CG32858-PB|CG32858-PE|CG32858-PG|fascin|sn-PA|sn-PB|sn-PE|sn-PG', map_location='1-21.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7964024, end_position_on_the_genomic_accession=7986601, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31718, CurrentID=0, Status='live', Symbol='Tbh', Aliases='Dmel_CG1543, CG1543, DBH, Dbh, Dmel\\CG1543, T beta h, TBH, TBetaH, TBetah, TBh, TbH, TbetaH, TbetaM, Tbetah, tbetah, tbh', description='Tyramine beta hydroxylase', other_designations='CG1543-PB|CG1543-PC|Tbh-PB|Tbh-PC|dopamine beta hydroxylase|tyramine beta hydroxylase|tyramine beta-hydroxylase|tyramine beta-monooxygenase|tyramine-beta-hydroxylase|tyroside beta-hydroxylase|tyrosine beta hydroxylase', map_location='7D2-7D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=7995697, end_position_on_the_genomic_accession=8027394, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31720, CurrentID=0, Status='live', Symbol='CG15333', Aliases='Dmel_ CG18293, Dmel\\CG15333', description='CG15333 gene product from transcript CG15333-RA', other_designations='CG15333-PA|uncharacterized protein', map_location='7D2-7D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8019523, end_position_on_the_genomic_accession=8021345, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31721, CurrentID=0, Status='live', Symbol='CG2233', Aliases='Dmel_ Dmel\\CG2233', description='CG2233 gene product from transcript CG2233-RA', other_designations='CG2233-PA', map_location='7D3-7D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8035741, end_position_on_the_genomic_accession=8037517, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31722, CurrentID=0, Status='live', Symbol='fs(1)h', Aliases='Dmel_CG2252, CG2252, Dmel\\CG2252, Fs(1)h, Fs(1)hL, Fs(1)hS, Fsh, N72, ORE-1, anon-EST:Liang-1.81, anon-EST:fe1G2, clone 1.81, fs(1)1456, fs(1)26/26A, fs(1)M16, fs(1)R10.4, fsh, fst(1)h, l(1)7Da, l(1)G0093, l(1)G0495, rnc', description='female sterile (1) homeotic', other_designations='CG2252-PA|CG2252-PB|CG2252-PC|CG2252-PD|CG2252-PE|CG2252-PF|CG2252-PG|CG2252-PH|CG2252-PI|anon-fast-evolving-1G2|female sterile (1) homoeotic|female sterile homeotic|fs(1)h-PA|fs(1)h-PB|fs(1)h-PC|fs(1)h-PD|fs(1)h-PE|fs(1)h-PF|fs(1)h-PG|fs(1)h-PH|fs(1)h-PI|rancor', map_location='1-21.65 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8037808, end_position_on_the_genomic_accession=8061436, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  31724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31724, CurrentID=0, Status='live', Symbol='Upf2', Aliases='Dmel_CG2253, CG2253, Dmel\\CG2253, UPF2, smg-3, upf2', description='CG2253 gene product from transcript CG2253-RA', other_designations='CG2253-PA|Upf2-PA|up-frameshift suppressor 2', map_location='7D5-7D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8070244, end_position_on_the_genomic_accession=8075466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31725, CurrentID=0, Status='live', Symbol='CG1571', Aliases='Dmel_ Dmel\\CG1571', description='CG1571 gene product from transcript CG1571-RA', other_designations='CG1571-PA', map_location='7D5-7D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8075877, end_position_on_the_genomic_accession=8078240, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31726, CurrentID=0, Status='live', Symbol='CG2254', Aliases='Dmel_ Dmel\\CG2254', description='CG2254 gene product from transcript CG2254-RA', other_designations='CG2254-PA', map_location='7D5-7D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8078350, end_position_on_the_genomic_accession=8081605, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31727, CurrentID=0, Status='live', Symbol='CG2256', Aliases='Dmel_ Dmel\\CG2256', description='CG2256 gene product from transcript CG2256-RC', other_designations='CG2256-PC', map_location='7D5-7D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8084018, end_position_on_the_genomic_accession=8087213, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31728, CurrentID=0, Status='live', Symbol='UbcE2H', Aliases='Dmel_CG2257, CG2257, Dmel\\CG2257, Ubc-E2H, Ubc-E2Hs, UbcDE2H, dUbc-E2H, ubcE2h', description='Ubiquitin conjugating enzyme E2H', other_designations='CG2257-PA|CG2257-PB|CG2257-PC|Ubc-E2H|UbcE2H-PA|UbcE2H-PB|UbcE2H-PC|ubiquitin conjugating enzyme E2H', map_location='7D6-7D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8087544, end_position_on_the_genomic_accession=8094064, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31729, CurrentID=0, Status='live', Symbol='CG2258', Aliases='Dmel_ Dmel\\CG2258', description='CG2258 gene product from transcript CG2258-RD', other_designations='CG2258-PA|CG2258-PB|CG2258-PC|CG2258-PD', map_location='7D6-7D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8094250, end_position_on_the_genomic_accession=8102095, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  31731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31731, CurrentID=0, Status='live', Symbol='CG1575', Aliases='Dmel_ Dmel\\CG1575', description='CG1575 gene product from transcript CG1575-RA', other_designations='CG1575-PA|CG1575-PB', map_location='7D6-7D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8117012, end_position_on_the_genomic_accession=8119130, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31732, CurrentID=0, Status='live', Symbol='CG2260', Aliases='Dmel_ Dmel\\CG2260', description='CG2260 gene product from transcript CG2260-RA', other_designations='CG2260-PA', map_location='7D6-7D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8119082, end_position_on_the_genomic_accession=8121147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31733, CurrentID=0, Status='live', Symbol='cyr', Aliases='Dmel_CG15335, CG15335, CG15338, Dmel\\CG15335', description='cypher', other_designations='CG15335-PB|CG15335-PC|CG15335-PD|cyr-PB|cyr-PC|cyr-PD', map_location='7D6-7D14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8121280, end_position_on_the_genomic_accession=8135713, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31735 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31735, CurrentID=0, Status='live', Symbol='CG2116', Aliases='Dmel_ Dmel\\CG2116', description='CG2116 gene product from transcript CG2116-RA', other_designations='CG2116-PA', map_location='7D12-7D12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8129812, end_position_on_the_genomic_accession=8131783, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31736 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31736, CurrentID=0, Status='live', Symbol='CG10958', Aliases='Dmel_ Dmel\\CG10958', description='CG10958 gene product from transcript CG10958-RA', other_designations='CG10958-PA|uncharacterized protein', map_location='7D12-7D13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8132285, end_position_on_the_genomic_accession=8134967, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31737, CurrentID=0, Status='live', Symbol='CG2120', Aliases='Dmel_ Dmel\\CG2120', description='CG2120 gene product from transcript CG2120-RB', other_designations='CG2120-PB', map_location='7D14-7D14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8135824, end_position_on_the_genomic_accession=8137046, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31738, CurrentID=0, Status='live', Symbol='Smox', Aliases='Dmel_CG2262, CG2262, DSMAD2, DSmad2, Dmel\\CG2262, SMAD2, SMOX, Sad, Smad, Smad2, SmoX, dSMAD2, dSmad2, dsmad2, l(1)G0348, sad, smad2, smox, ted, tmp', description='Smad on X', other_designations='CG2262-PA|CG2262-PB|Smox-PA|Smox-PB|Smox/dSmad2|sisters against dpp|smad on X|temporarily deranged', map_location='7D14-7D15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8137041, end_position_on_the_genomic_accession=8142279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31739, CurrentID=0, Status='live', Symbol='CG17982', Aliases='Dmel_ Dmel\\CG17982', description='CG17982 gene product from transcript CG17982-RB', other_designations='CG17982-PA|CG17982-PB|uncharacterized protein', map_location='7D15-7D15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8142550, end_position_on_the_genomic_accession=8143835, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31740, CurrentID=0, Status='live', Symbol='alpha-PheRS', Aliases='Dmel_CG2263, CG2263, Dmel\\CG2263', description='Phenylalanyl-tRNA synthetase alpha subunit', other_designations='CG2263-PA|alpha-PheRS-PA', map_location='7D15-7D15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8143904, end_position_on_the_genomic_accession=8145602, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31741, CurrentID=0, Status='live', Symbol='CG2129', Aliases='Dmel_ Dmel\\CG2129', description='CG2129 gene product from transcript CG2129-RA', other_designations='CG2129-PA', map_location='7D15-7D16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8145839, end_position_on_the_genomic_accession=8147855, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31742, CurrentID=0, Status='live', Symbol='CG15336', Aliases='Dmel_ Dmel\\CG15336', description='CG15336 gene product from transcript CG15336-RD', other_designations='CG15336-PD|uncharacterized protein', map_location='7D16-7D16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8147949, end_position_on_the_genomic_accession=8148705, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31743, CurrentID=0, Status='live', Symbol='CG10959', Aliases='Dmel_ Dmel\\CG10959', description='CG10959 gene product from transcript CG10959-RA', other_designations='CG10959-PA|CG10959-PB|uncharacterized protein', map_location='7D16-7D16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8148783, end_position_on_the_genomic_accession=8150686, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31744 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31744, CurrentID=0, Status='live', Symbol='Gbeta5', Aliases='Dmel_CG10763, CG10763, Dmel\\CG10763, G[[beta5]], G[[beta]], Gbeta, Gbeta 5, Gbetagamma', description='CG10763 gene product from transcript CG10763-RA', other_designations='CG10763-PA|G-protein beta 5|Gbeta5-PA', map_location='7D16-7D16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8150731, end_position_on_the_genomic_accession=8152275, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31745, CurrentID=0, Status='live', Symbol='CG18262', Aliases='Dmel_ Dmel\\CG18262', description='CG18262 gene product from transcript CG18262-RA', other_designations='CG18262-PA|uncharacterized protein', map_location='7D16-7D16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8152948, end_position_on_the_genomic_accession=8154533, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31746, CurrentID=0, Status='live', Symbol='Traf6', Aliases='Dmel_CG10961, CG10961, DTRAF2, Dmel\\CG10961, EP325, TRAF, TRAF2, TRAF6, Traf2, dTRAF2, dTRAF6, dTraf2, dTraf6, dtraf2, traf2', description='TNF-receptor-associated factor 6', other_designations='CG10961-PA|EP325|TNF-receptor-associated factor 2|Traf6-PA|tumor necrosis factor receptor 2', map_location='7D16-7D16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8154599, end_position_on_the_genomic_accession=8157386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31747, CurrentID=0, Status='live', Symbol='GIIIspla2', Aliases='Dmel_CG1583, BEST:HL07869, CG1583, Dmel\\CG1583, GIIIsPLA2, Giiispla2', description='CG1583 gene product from transcript CG1583-RC', other_designations='CG1583-PB|CG1583-PC|GIIIspla2-PB|GIIIspla2-PC|phospholipase A[[2]]', map_location='7D17-7D17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8157628, end_position_on_the_genomic_accession=8159348, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31748, CurrentID=0, Status='live', Symbol='CG15337', Aliases='Dmel_ Dmel\\CG15337', description='CG15337 gene product from transcript CG15337-RA', other_designations='CG15337-PA|uncharacterized protein', map_location='7D17-7D17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8160054, end_position_on_the_genomic_accession=8161552, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31749 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31749, CurrentID=0, Status='live', Symbol='CG10761', Aliases='Dmel_ Dmel\\CG10761', description='CG10761 gene product from transcript CG10761-RA', other_designations='CG10761-PA|uncharacterized protein', map_location='7D17-7D17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8161540, end_position_on_the_genomic_accession=8162792, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31750 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31750, CurrentID=0, Status='live', Symbol='Or7a', Aliases='Dmel_CG10759, 7D.1, 7a, CG10759, DOR30, DmOR7a, Dmel\\CG10759', description='Odorant receptor 7a', other_designations='CG10759-PA|Or7a-PA|odorant receptor 7a', map_location='7D17-7D17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8164266, end_position_on_the_genomic_accession=8165708, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31751, CurrentID=0, Status='live', Symbol='CG2278', Aliases='Dmel_ Dmel\\CG2278', description='CG2278 gene product from transcript CG2278-RC', other_designations='CG2278-PA|CG2278-PB|CG2278-PC', map_location='7D17-7D17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8165833, end_position_on_the_genomic_accession=8170266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835394, CurrentID=0, Status='live', Symbol='CG45060', Aliases='Dmel_ CG15245, CG15246, CG32696, CG42262, Dmel\\CG45060', description='CG45060 gene product from transcript CG45060-RA', other_designations='CG45060-PA|uncharacterized protein', map_location='9A2-9A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9868068, end_position_on_the_genomic_accession=9868568, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31759, CurrentID=0, Status='live', Symbol='CG2147', Aliases='Dmel_ BcDNA:LP02728, Dmel\\CG2147', description='CG2147 gene product from transcript CG2147-RA', other_designations='CG2147-PA', map_location='7D22-7D22', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8240684, end_position_on_the_genomic_accession=8241483, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31760, CurrentID=0, Status='live', Symbol='Trxr-1', Aliases='Dmel_CG2151, 2151, CG2151, DTR, Dm-TrxR, DmTR, DmTrx, DmTrxR, DmTrxR-1, Dmel\\CG2151, GR, Gr, TH, TRR, Trx, TrxR, TrxR-1, Trxr1, anon-WO03040301.185, anon-WO03040301.187, dTrxR, dmtrxr-1, gr, l(1)G0154, l(1)G0379, l(1)G0477, l(1)G0481, trxr-1', description='Thioredoxin reductase-1', other_designations='CG2151-PA|CG2151-PB|CG2151-PC|Trxr-1-PA|Trxr-1-PB|Trxr-1-PC|disulfide reductase|glutathione reductase|thioredoxin reductase|thioredoxin reductase-1', map_location='7D22-7E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8242528, end_position_on_the_genomic_accession=8248033, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31761, CurrentID=0, Status='live', Symbol='sni', Aliases='Dmel_CG10964, CG10964, Dmel\\CG10964, SNI, Sni, anon-WO03040301.183', description='sniffer', other_designations='CG10964-PA|CG10964-PB|sni-PA|sni-PB', map_location='7D22-7D22', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8241501, end_position_on_the_genomic_accession=8244016, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31762, CurrentID=0, Status='live', Symbol='ND-75', Aliases='Dmel_CG2286, CG2286, Dmel\\CG2286, ND75, NDUFS1, NUAM', description='NADH dehydrogenase (ubiquinone) 75 kDa subunit', other_designations='CG2286-PA|CG2286-PB|FBpp0071128|NADH ubiquinone oxidoreductase 75 kDa|NADH-ubiquinone oxidoreductase 75 kDa subunit|NADH:ubiquinone oxidoreductase 75 kDa subunit|NADH:ubiquinone reductase 75 kDa subunit|NADH:ubiquinone reductase 75-kD subunit precursor|NADH:ubiquinone reductase 75kD subunit precursor|ND-75-PA|ND-75-PB', map_location='7E1-7E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8248651, end_position_on_the_genomic_accession=8251687, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31763, CurrentID=0, Status='live', Symbol='CG1632', Aliases='Dmel_ Dmel\\CG1632, SPH198', description='CG1632 gene product from transcript CG1632-RA', other_designations='CG1632-PA|CG1632-PB', map_location='7E1-7E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8252670, end_position_on_the_genomic_accession=8272462, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  31764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31764, CurrentID=0, Status='live', Symbol='Corp', Aliases='Dmel_CG10965, CG10965, Dmel\\CG10965, corp', description='Companion of reaper', other_designations='CG10965-PA|CG10965-PB|Corp-PA|Corp-PB|companion of reaper', map_location='7E1-7E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8269898, end_position_on_the_genomic_accession=8271409, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31765, CurrentID=0, Status='live', Symbol='CG15343', Aliases='Dmel_ Dmel\\CG15343', description='CG15343 gene product from transcript CG15343-RA', other_designations='CG15343-PA|uncharacterized protein', map_location='7E1-7E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8272324, end_position_on_the_genomic_accession=8273806, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31766, CurrentID=0, Status='live', Symbol='CG1636', Aliases='Dmel_ Dmel\\CG1636', description='CG1636 gene product from transcript CG1636-RB', other_designations='CG1636-PA|CG1636-PB|CG1636-PC', map_location='7E1-7E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8273412, end_position_on_the_genomic_accession=8276450, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31767, CurrentID=0, Status='live', Symbol='nAChRalpha3', Aliases='Dmel_CG2302, Acr7E, CG2302, D[a3], Da3, Dalpha3, Dmel\\CG2302, alpha3, lincRNA.965, nACRalpha-7E, nAChR, nAcRalpha-7E, nAcRe', description='nicotinic Acetylcholine Receptor alpha3', other_designations='CG2302-PB|CG2302-PC|Dalpha3|drosophila alpha-like subunit 3|nAChRalpha3-PB|nAChRalpha3-PC|nAcRalpha-7E|nicotinic acetylcholine receptor alpha 7E|nicotinic acetylcholine receptor alpha3', map_location='7E1-7E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8276200, end_position_on_the_genomic_accession=8367169, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  31770 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31770, CurrentID=0, Status='live', Symbol='CG1387', Aliases='Dmel_ Dmel\\CG1387', description='CG1387 gene product from transcript CG1387-RA', other_designations='CG1387-PA', map_location='7E5-7E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8385781, end_position_on_the_genomic_accession=8389504, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31771 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31771, CurrentID=0, Status='live', Symbol='CG10555', Aliases='Dmel_ Dmel\\CG10555', description='CG10555 gene product from transcript CG10555-RB', other_designations='CG10555-PA|CG10555-PB|uncharacterized protein', map_location='7E6-7E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8398940, end_position_on_the_genomic_accession=8404516, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31772 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31772, CurrentID=0, Status='live', Symbol='CG11190', Aliases='Dmel_ Dm*AAF46367/27-522, Dmel\\CG11190', description='CG11190 gene product from transcript CG11190-RA', other_designations='CG11190-PA|uncharacterized protein', map_location='7E6-7E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8404535, end_position_on_the_genomic_accession=8407263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31773 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31773, CurrentID=0, Status='live', Symbol='Trf2', Aliases='Dmel_CG18009, CG11195, CG18009, Dmel\\CG18009, TLF, TRF2, dTRF2, dTRF2L, dTRF2S, dTrf2, l(1)G0039, l(1)G0071, l(1)G0295, l(1)G0425, p79/TRF2, trf2', description='TATA box binding protein-related factor 2', other_designations='CG18009-PE|CG18009-PF|CG18009-PG|CG18009-PH|CG18009-PI|CG18009-PJ|TATA box-binding protein related factor 2|TBP related factor 2|TBP-like factor|TBP-related factor 2|Trf2-PE|Trf2-PF|Trf2-PG|Trf2-PH|Trf2-PI|Trf2-PJ|lethal (1) G0039|lethal (1) G0071|lethal (1) G0295|lethal (1) G0425', map_location='7E7-7E', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8409272, end_position_on_the_genomic_accession=8433011, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31775 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31775, CurrentID=0, Status='live', Symbol='CG12125', Aliases='Dmel_ Dmel\\CG12125', description='CG12125 gene product from transcript CG12125-RC', other_designations='CG12125-PA|CG12125-PB|CG12125-PC|uncharacterized protein', map_location='7E9-7E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8433923, end_position_on_the_genomic_accession=8437820, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31776 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31776, CurrentID=0, Status='live', Symbol='CG1440', Aliases='Dmel_ Dmel\\CG1440', description='CG1440 gene product from transcript CG1440-RD', other_designations='CG1440-PA|CG1440-PB|CG1440-PC|CG1440-PD', map_location='7E9-7E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8437057, end_position_on_the_genomic_accession=8441264, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31777 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31777, CurrentID=0, Status='live', Symbol='CG12123', Aliases='Dmel_ Dmel\\CG12123', description='CG12123 gene product from transcript CG12123-RA', other_designations='CG12123-PA|uncharacterized protein', map_location='7E9-7E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8441247, end_position_on_the_genomic_accession=8442348, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31778 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31778, CurrentID=0, Status='live', Symbol='org-1', Aliases='Dmel_CG11202, CG11202, Dm-ORG-1, Dmel\\CG11202, Org-1', description='optomotor-blind-related-gene-1', other_designations='CG11202-PA|CG11202-PB|Optomotor-blind-related-1|Optomotor-blind-related-gene 1|omb-related gene-1|optomotor-blind-related gene-1|optomotor-blind-related-protein-1|org-1-PA|org-1-PB', map_location='7E9-7E10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8447691, end_position_on_the_genomic_accession=8455069, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31779 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31779, CurrentID=0, Status='live', Symbol='CG15347', Aliases='Dmel_ Dmel\\CG15347', description='CG15347 gene product from transcript CG15347-RA', other_designations='CG15347-PA|CG15347-PB|uncharacterized protein', map_location='7E11-7E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8458494, end_position_on_the_genomic_accession=8460898, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31780 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31780, CurrentID=0, Status='live', Symbol='Es2', Aliases='Dmel_CG1474, CG1474, Dmel\\CG1474, ES2, es', description='CG1474 gene product from transcript CG1474-RA', other_designations='CG1474-PA|Es2-PA', map_location='7E11-7E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8461479, end_position_on_the_genomic_accession=8463349, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31781 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31781, CurrentID=0, Status='live', Symbol='Sptr', Aliases='Dmel_CG12117, CG12117, Dmel\\CG12117, SR, sptr', description='Sepiapterin reductase', other_designations='CG12117-PA|Sptr-PA|sepiapterin reductase', map_location='7E11-7E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8463296, end_position_on_the_genomic_accession=8464438, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31782 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31782, CurrentID=0, Status='live', Symbol='CG12116', Aliases='Dmel_ 144701_at, Dmel\\CG12116', description='CG12116 gene product from transcript CG12116-RA', other_designations='CG12116-PA|uncharacterized protein', map_location='7F1-7F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8467443, end_position_on_the_genomic_accession=8469012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31785 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31785, CurrentID=0, Status='live', Symbol='Cp7Fb', Aliases='Dmel_CG15350, B, C, CG15350, Dmel\\CG15350, cp7Fb', description='Chorion protein b at 7F', other_designations='CG15350-PA|CG15350-PB|Cp7Fb-PA|Cp7Fb-PB|chorion protein b at 7F', map_location='7F1-7F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8473837, end_position_on_the_genomic_accession=8476071, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31786 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31786, CurrentID=0, Status='live', Symbol='Cp7Fc', Aliases='Dmel_CG15351, CG15351, D, Dmel\\CG15351, E, s37', description='Chorion protein c at 7F', other_designations='CG15351-PA|Cp7Fc-PA|chorion protein c at 7F', map_location='7F1-7F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8477117, end_position_on_the_genomic_accession=8478320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31787 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31787, CurrentID=0, Status='live', Symbol='Cp36', Aliases='Dmel_CG1478, CG1478, CP36, Dmel\\CG1478, Dms36, E4, S36, cp36, dec-2, dec2, dec[[2]], fs(1)C2, s36, s36-1, s365', description='Chorion protein 36', other_designations='CG1478-PA|Cp36-PA|Shell-36|chorion protein 36|defective chorion 2', map_location='1-23.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8479319, end_position_on_the_genomic_accession=8480523, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31788 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31788, CurrentID=0, Status='live', Symbol='Cp38', Aliases='Dmel_CG11213, ACE1, CG11213, CP38, Dmel\\CG11213, Dms38, E3, S38, cp38, s38, s38-1', description='Chorion protein 38', other_designations='CG11213-PA|CG11213-PB|Cp38-PA|Cp38-PB|Shell-38|chorion prot 38|chorion protein 38', map_location='7F1-7F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8481837, end_position_on_the_genomic_accession=8483311, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31789 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31789, CurrentID=0, Status='live', Symbol='otu', Aliases='Dmel_CG12743, CG12743, DROOTUA, Dmel\\CG12743, K, OTU, Otu, fs(1)231, fs(1)23l, fs(1)M101, fs(1)jA265, fs(1)otu', description='ovarian tumor', other_designations='CG12743-PA|CG12743-PB|CG12743-PC|CG12743-PD|CG12743-PE|otu-PA|otu-PB|otu-PC|otu-PD|otu-PE|ovarian tumour', map_location='1-22.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8485008, end_position_on_the_genomic_accession=8489535, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31791 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31791, CurrentID=0, Status='live', Symbol='PIP82', Aliases='Dmel_CG11219, CG11219, Dmel\\CG11219', description='CG11219 gene product from transcript CG11219-RA', other_designations='CG11219-PA|PIP82-PA', map_location='7F3-7F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8525387, end_position_on_the_genomic_accession=8529698, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31792 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31792, CurrentID=0, Status='live', Symbol='Nrg', Aliases='Dmel_CG1634, CG1634, CT4318, Dmel\\CG1634, NFASC, NRG, Ngl167, Nrg180, RA35, ceb, fs(1)M72, ibx, l(1)7Fa, l(1)G0099, l(1)G0413, l(1)G0488, nrg, Nrg', description='Neuroglian', other_designations='CG1634-PA|CG1634-PB|CG1634-PC|CG1634-PD|CG1634-PE|CG1634-PF|CG1634-PG|CG1634-PH|CG1634-PI|Nrg-PA|Nrg-PB|Nrg-PC|Nrg-PD|Nrg-PE|Nrg-PF|Nrg-PG|Nrg-PH|Nrg-PI|central brain deranged|female sterile(1)M72|icebox|lethal (1) G0488|neuroglian', map_location='1-23.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8517373, end_position_on_the_genomic_accession=8555170, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  31793 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31793, CurrentID=0, Status='live', Symbol='IntS4', Aliases='Dmel_CG12113, CG12113, Dmel\\CG12113, Int4, l(1)G0095', description='Integrator 4', other_designations='CG12113-PA|IntS4-PA|integrator 4|lethal (1) G0095', map_location='7F4-7F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8555239, end_position_on_the_genomic_accession=8558781, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31794 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31794, CurrentID=0, Status='live', Symbol='CG12112', Aliases='Dmel_ Dmel\\CG12112', description='CG12112 gene product from transcript CG12112-RA', other_designations='CG12112-PA|CG12112-PB|uncharacterized protein', map_location='7F4-7F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8558879, end_position_on_the_genomic_accession=8560192, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31795 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31795, CurrentID=0, Status='live', Symbol='Trf4-1', Aliases='Dmel_CG11265, AAF46390, CG11265, DmTRF4-1, Dmel\\CG11265, Trf4', description='CG11265 gene product from transcript CG11265-RE', other_designations='CG11265-PC|CG11265-PE|CG11265-PF|CG11265-PG|CG11265-PH|CG11265-PI|CG11265-PJ|Trf4-1-PC|Trf4-1-PE|Trf4-1-PF|Trf4-1-PG|Trf4-1-PH|Trf4-1-PI|Trf4-1-PJ', map_location='7F4-7F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8560322, end_position_on_the_genomic_accession=8570159, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31796 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31796, CurrentID=0, Status='live', Symbol='CG12111', Aliases='Dmel_ Dmel\\CG12111', description='CG12111 gene product from transcript CG12111-RB', other_designations='CG12111-PA|CG12111-PB|uncharacterized protein', map_location='7F4-7F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8570208, end_position_on_the_genomic_accession=8570948, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31797 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31797, CurrentID=0, Status='live', Symbol='spirit', Aliases='Dmel_CG2056, CG2056, Dmel\\CG2056, SP33, Spirit, c-SP33', description='Serine Protease Immune Response Integrator', other_designations='CG2056-PA|CG2056-PB|CG2056-PC|CG2056-PD|serine protease immune response integrator|spirit-PA|spirit-PB|spirit-PC|spirit-PD', map_location='7F5-7F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8571099, end_position_on_the_genomic_accession=8573593, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31798 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31798, CurrentID=0, Status='live', Symbol='CG12065', Aliases='Dmel_ Dmel\\CG12065', description='CG12065 gene product from transcript CG12065-RF', other_designations='CG12065-PA|CG12065-PB|CG12065-PC|CG12065-PD|CG12065-PE|CG12065-PF|uncharacterized protein', map_location='7F5-7F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8574306, end_position_on_the_genomic_accession=8588528, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31799 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31799, CurrentID=0, Status='live', Symbol='CG12081', Aliases='Dmel_ Dmel\\CG12081', description='CG12081 gene product from transcript CG12081-RB', other_designations='CG12081-PA|CG12081-PB|uncharacterized protein', map_location='7F7-7F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8588811, end_position_on_the_genomic_accession=8590774, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31800 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31800, CurrentID=0, Status='live', Symbol='Crag', Aliases='Dmel_CG12737, CG12737, CJ, CRAG, Dmel\\CG12737', description='Calmodulin-binding protein related to a Rab3 GDP/GTP exchange protein', other_designations='CG12737-PA|CG12737-PB|CG12737-PD|Calmodulin-binding protein related to a Rab3 GDP-GTP exchange protein|Crag-PA|Crag-PB|Crag-PD', map_location='7F7-7F8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8592031, end_position_on_the_genomic_accession=8602036, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  31801 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31801, CurrentID=0, Status='live', Symbol='Caf1-180', Aliases='Dmel_CG12109, CAF-1 p180, CAF1-180, CAF1-p180, CAF1p180, CG12109, Dmel\\CG12109, NURF, dCAF-1-180, dCAF-1-p180, p180', description='Chromatin assembly factor 1, p180 subunit', other_designations='CAF-1 large subunit|CG12109-PB|Caf1-180-PB|chromatin assembly factor 1, p180 subunit', map_location='7F8-7F8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8602137, end_position_on_the_genomic_accession=8606304, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31802 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31802, CurrentID=0, Status='live', Symbol='oc', Aliases='Dmel_CG12154, CG12154, Dmel\\CG12154, Oc, Otd, Otx, l(1)7Ff, l(1)8Ac, ort, otd, uvi', description='ocelliless', other_designations='CG12154-PC|CG12154-PD|CG12154-PE|CG12154-PF|CG12154-PG|oc-PC|oc-PD|oc-PE|oc-PF|oc-PG|orthodentical|orthodenticle|uvi-insensitive', map_location='1-23.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8630159, end_position_on_the_genomic_accession=8650681, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31803 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31803, CurrentID=0, Status='live', Symbol='CG12772', Aliases='Dmel_ Dmel\\CG12772, FBgn 30055', description='CG12772 gene product from transcript CG12772-RB', other_designations='CG12772-PA|CG12772-PB|rubicon|uncharacterized protein', map_location='8A1-8A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8677063, end_position_on_the_genomic_accession=8680912, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31804 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31804, CurrentID=0, Status='live', Symbol='CG11284', Aliases='Dmel_ CG11285, Dmel\\CG11284', description='CG11284 gene product from transcript CG11284-RE', other_designations='CG11284-PA|CG11284-PB|CG11284-PC|CG11284-PD|CG11284-PE|uncharacterized protein', map_location='8A1-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8681252, end_position_on_the_genomic_accession=8688760, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31805 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31805, CurrentID=0, Status='live', Symbol='Ppt1', Aliases='Dmel_CG12108, CG12108, DmPpt1, Dmel\\CG12108, PPT1, ppt1', description='Palmitoyl-protein thioesterase 1', other_designations='CG12108-PA|CG12108-PB|Ppt1-PA|Ppt1-PB|palmitoyl protein thioesterase 1|palmitoyl-protein thioesterase 1', map_location='8A2-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8688627, end_position_on_the_genomic_accession=8690227, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31806 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31806, CurrentID=0, Status='live', Symbol='Ogg1', Aliases='Dmel_CG1795, 1795, BcDNA:LD19945, CG1795, Dmel\\CG1795, dOGG1', description='8-oxoguanine DNA glycosylase', other_designations='CG1795-PA|Ogg1-PA', map_location='8A2-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8690327, end_position_on_the_genomic_accession=8691811, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31807 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31807, CurrentID=0, Status='live', Symbol='CG11294', Aliases='Dmel_ CT31519, Dmel\\CG11294', description='CG11294 gene product from transcript CG11294-RB', other_designations='CG11294-PA|CG11294-PB|uncharacterized protein', map_location='8A2-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8691820, end_position_on_the_genomic_accession=8693620, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31808 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31808, CurrentID=0, Status='live', Symbol='Hexo2', Aliases='Dmel_CG1787, CG1787, DmHex2, Dmel\\CG1787, HEX 2, HEX2, HEXO2', description='Hexosaminidase 2', other_designations='CG1787-PA|Hexo2-PA|hexosaminidase 2', map_location='8A2-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8709652, end_position_on_the_genomic_accession=8712478, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31809 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31809, CurrentID=0, Status='live', Symbol='CG2004', Aliases='Dmel_ Dmel\\CG2004', description='CG2004 gene product from transcript CG2004-RG', other_designations='CG2004-PA|CG2004-PC|CG2004-PD|CG2004-PF|CG2004-PG', map_location='8A2-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8712472, end_position_on_the_genomic_accession=8714924, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31810 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31810, CurrentID=0, Status='live', Symbol='CG1785', Aliases='Dmel_ Dmel\\CG1785', description='CG1785 gene product from transcript CG1785-RA', other_designations='CG1785-PA', map_location='8A2-8A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8714936, end_position_on_the_genomic_accession=8717065, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31811 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31811, CurrentID=0, Status='live', Symbol='l(1)G0020', Aliases='Dmel_CG1994, CG1994, Dmel\\CG1994, NAT10, p110', description='lethal (1) G0020', other_designations='CG1994-PA|CG1994-PB|l(1)G0020-PA|l(1)G0020-PB', map_location='8A2-8A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8716713, end_position_on_the_genomic_accession=8720277, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31812 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31812, CurrentID=0, Status='live', Symbol='CG1789', Aliases='Dmel_ Dmel\\CG1789', description='CG1789 gene product from transcript CG1789-RA', other_designations='CG1789-PA', map_location='8A3-8A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8720520, end_position_on_the_genomic_accession=8721503, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31813 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31813, CurrentID=0, Status='live', Symbol='Lim1', Aliases='Dmel_CG11354, BcDNA:GH04929, CG10760, CG11354, Dlim1, Dmel\\CG11354, Lim, d dlim-1, dlim1, lim1', description='CG11354 gene product from transcript CG11354-RB', other_designations='CG11354-PA|CG11354-PB|Lim kinase|Lim1-PA|Lim1-PB', map_location='8A5-8B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8756539, end_position_on_the_genomic_accession=8805804, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31815 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31815, CurrentID=0, Status='live', Symbol='CG12075', Aliases='Dmel_ Dmel\\CG12075', description='CG12075 gene product from transcript CG12075-RC', other_designations='CG12075-PA|CG12075-PB|CG12075-PC|CG12075-PD|CG12075-PE|uncharacterized protein', map_location='8B4-8B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8838323, end_position_on_the_genomic_accession=8872774, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31816 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31816, CurrentID=0, Status='live', Symbol='Moe', Aliases='Dmel_CG10701, CG10701, D17, Dmel\\CG10701, Dmoe, EMR1, ERM, Emr1s, anon-WO03040301.155, anon-WO03040301.157, anon-WO03040301.159, anon-WO03040301.161, dMoe, gma, l(1)G0067, l(1)G0323, l(1)G0404, l(1)G0415, moe, moesin/ezrin/radixin homolog mRNA, Moe', description='Moesin', other_designations='CG10701-PA|CG10701-PB|CG10701-PC|CG10701-PD|CG10701-PE|CG10701-PF|CG10701-PG|CG10701-PH|CG10701-PI|CG10701-PJ|CG10701-PK|CG10701-PL|CG10701-PM|Ezrin-Radixin-Moesin|Ezrin-moesin-radixin-1|Moe-PA|Moe-PB|Moe-PC|Moe-PD|Moe-PE|Moe-PF|Moe-PG|Moe-PH|Moe-PI|Moe-PJ|Moe-PK|Moe-PL|Moe-PM|Moesin-like|dmoesin|moesin', map_location='8B4-8B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8873012, end_position_on_the_genomic_accession=8898332, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  31817 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31817, CurrentID=0, Status='live', Symbol='CG1885', Aliases='Dmel_ Dmel\\CG1885', description='CG1885 gene product from transcript CG1885-RB', other_designations='CG1885-PA|CG1885-PB', map_location='8B6-8B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8898554, end_position_on_the_genomic_accession=8900264, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31818 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31818, CurrentID=0, Status='live', Symbol='Rbm13', Aliases='Dmel_CG10648, CG10648, Dmel\\CG10648, RBM13', description='RNA-binding motif protein 13', other_designations='CG10648-PA|CG10648-PB|Rbm13-PA|Rbm13-PB', map_location='8B7-8B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8901699, end_position_on_the_genomic_accession=8903761, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31822 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31822, CurrentID=0, Status='live', Symbol='CG12661', Aliases='Dmel_ Dmel\\CG12661', description='CG12661 gene product from transcript CG12661-RB', other_designations='CG12661-PA|CG12661-PB|uncharacterized protein', map_location='8B8-8B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8922888, end_position_on_the_genomic_accession=8923687, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31824 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31824, CurrentID=0, Status='live', Symbol='CG10962', Aliases='Dmel_ BcDNA:RE29909, CG7090, Dmel\\CG10962, EP(X)1565, EP1565', description='CG10962 gene product from transcript CG10962-RB', other_designations='CG10962-PB|uncharacterized protein', map_location='8C1-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8972908, end_position_on_the_genomic_accession=8974255, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31825 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31825, CurrentID=0, Status='live', Symbol='CG12662', Aliases='Dmel_ Dmel\\CG12662', description='CG12662 gene product from transcript CG12662-RA', other_designations='CG12662-PA|uncharacterized protein', map_location='8C1-8C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8983634, end_position_on_the_genomic_accession=8984190, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31826 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31826, CurrentID=0, Status='live', Symbol='rdgA', Aliases='Dmel_CG42667, CG10966, CG12660, CG15352, CG34344, CG42667, DGK, DGK2, Dmel\\CG42667, Dmel_CG10966, Dmel_CG12660, Dmel_CG15352, Dmel_CG34344, RDGA, dDGK2, dgk-5, x35', description='retinal degeneration A', other_designations='CG42667-PA|CG42667-PB|CG42667-PC|CG42667-PD|CG42667-PE|CG42667-PG|CG42667-PH|CG42667-PI|CG42667-PJ|CG42667-PK|DAG kinase|DAG-kinase|diacylglycerol kinase|rdgA-PA|rdgA-PB|rdgA-PC|rdgA-PD|rdgA-PE|rdgA-PG|rdgA-PH|rdgA-PI|rdgA-PJ|rdgA-PK|receptor degeneration A|receptor-degeneration-I', map_location='1-26.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=8907068, end_position_on_the_genomic_accession=9029476, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  31829 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31829, CurrentID=0, Status='live', Symbol='CG15365', Aliases='Dmel_ Dmel\\CG15365', description='CG15365 gene product from transcript CG15365-RA', other_designations='CG15365-PA|CG15365-PB|CG15365-PC|uncharacterized protein', map_location='8C4-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9054698, end_position_on_the_genomic_accession=9067983, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31830 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31830, CurrentID=0, Status='live', Symbol='CG10970', Aliases='Dmel_ BcDNA:HL08157, Dmel\\CG10970', description='CG10970 gene product from transcript CG10970-RB', other_designations='CG10970-PA|CG10970-PB|uncharacterized protein', map_location='8C4-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9069599, end_position_on_the_genomic_accession=9071408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31831 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31831, CurrentID=0, Status='live', Symbol='CG44325', Aliases='Dmel_ CG15366, CG7267, Dmel\\CG44325, Dmel_CG15366, Dmel_CG7267', description='CG44325 gene product from transcript CG44325-RD', other_designations='CG44325-PA|CG44325-PB|CG44325-PC|CG44325-PD|CG44325-PE|uncharacterized protein', map_location='8C4-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9070318, end_position_on_the_genomic_accession=9078499, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31833 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31833, CurrentID=0, Status='live', Symbol='CG7246', Aliases='Dmel_ Dmel\\CG7246, dUTP6', description='CG7246 gene product from transcript CG7246-RA', other_designations='CG7246-PA', map_location='8C4-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9078823, end_position_on_the_genomic_accession=9081057, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31834 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31834, CurrentID=0, Status='live', Symbol='HP1b', Aliases='Dmel_CG7041, CG7041, Dmel\\CG7041, HP1, HP1B, Hp1b', description='Heterochromatin Protein 1b', other_designations='CG7041-PA|CG7041-PB|CG7041-PC|HP1b-PA|HP1b-PB|HP1b-PC|heterochromatin protein 1B|heterochromatin protein 1b', map_location='8C4-8C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9081063, end_position_on_the_genomic_accession=9082746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31835 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31835, CurrentID=0, Status='live', Symbol='APC4', Aliases='Dmel_CG32707, CG32707, CG4350, Dmel\\CG32707', description='Anaphase Promoting Complex subunit 4', other_designations='APC4-PA|APC4-PC|CG32707-PA|CG32707-PC|anaphase promoting complex 4|anaphase promoting complex subunit 4', map_location='8C4-8C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9082931, end_position_on_the_genomic_accession=9085867, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31837 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31837, CurrentID=0, Status='live', Symbol='CG6999', Aliases='Dmel_ Dmel\\CG6999', description='CG6999 gene product from transcript CG6999-RA', other_designations='CG6999-PA', map_location='8C8-8C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9088367, end_position_on_the_genomic_accession=9089314, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31838 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31838, CurrentID=0, Status='live', Symbol='CG7033', Aliases='Dmel_ CCT2, Cct2, Dmel\\CG7033, Q8IRM8_DROME, Q9W392_DROME, SD02216p, TCP-1beta, TCP1beta, Tcp1-beta', description='CG7033 gene product from transcript CG7033-RA', other_designations='CG7033-PA|TCP-1 subunit 2beta', map_location='8C8-8C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9089513, end_position_on_the_genomic_accession=9092826, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31839 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31839, CurrentID=0, Status='live', Symbol='CG7766', Aliases='Dmel_ Dmel\\CG7766', description='CG7766 gene product from transcript CG7766-RI', other_designations='CG7766-PA|CG7766-PB|CG7766-PC|CG7766-PD|CG7766-PG|CG7766-PH|CG7766-PI', map_location='8C8-8C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9093612, end_position_on_the_genomic_accession=9103739, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  31840 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31840, CurrentID=0, Status='live', Symbol='Bx42', Aliases='Dmel_CG8264, BX42, CG8264, Dmel\\CG8264, SKIP, Skip, bx42', description='CG8264 gene product from transcript CG8264-RB', other_designations='Bx42-PA|Bx42-PB|CG8264-PA|CG8264-PB|Ski-interacting protein', map_location='8C9-8C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9103870, end_position_on_the_genomic_accession=9106604, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31841 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31841, CurrentID=0, Status='live', Symbol='Arfrp1', Aliases='Dmel_CG7039, ARFRP1, Arp1, CG7039, Dmel\\CG7039, Q9W389', description='ADP-ribosylation factor related protein 1', other_designations='Arfrp1-PA|CG7039-PA', map_location='8C9-8C9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9107069, end_position_on_the_genomic_accession=9108354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31842 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31842, CurrentID=0, Status='live', Symbol='AP-1gamma', Aliases='Dmel_CG9113, AP-1, AP-1 gamma, AP-1[gamma], AP1gamma, AP1gamma1, Ap-1, CG9113, Dmel\\CG9113, apl4, cg9113, dgamma, gamma, gamma-Ada', description='Adaptor Protein complex 1, gamma subunit', other_designations='AP-1gamma-PA|AP-1gamma-PB|AP-1gamma-PC|AP-1gamma-PD|AP-1gamma-PE|AP-1gamma-PF|AP-1gamma-PG|AP-1gamma-PH|AP-1gamma-PI|CG9113-PA|CG9113-PB|CG9113-PC|CG9113-PD|CG9113-PE|CG9113-PF|CG9113-PG|CG9113-PH|CG9113-PI|adaptor protein complex 1, gamma subunit|gamma subunit of AP-1 clathrin|gamma-adaptin', map_location='8C9-8C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9108691, end_position_on_the_genomic_accession=9117744, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  31843 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31843, CurrentID=0, Status='live', Symbol='fend', Aliases='Dmel_CG12664, BcDNA:RH56795, CG12664, Dmel\\CG12664, Dromel_CG12664_FBtr0071349_ld14_mORF, LD14, ld14', description='forked end', other_designations='CG12664-PB|CG12664-PC|fend-PB|fend-PC|ld14', map_location='8C11-8C13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9121973, end_position_on_the_genomic_accession=9135099, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31844 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31844, CurrentID=0, Status='live', Symbol='CG7065', Aliases='Dmel_ Dmel\\CG7065', description='CG7065 gene product from transcript CG7065-RA', other_designations='CG7065-PA', map_location='8C13-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9142493, end_position_on_the_genomic_accession=9147894, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31845 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31845, CurrentID=0, Status='live', Symbol='fh', Aliases='Dmel_CG8971, BcDNA:AT09528, CG8971, DFH, Dfh, Dmel\\CG8971, X25, dfh', description='frataxin homolog', other_designations='CG8971-PA|CG8971-PB|fh-PA|fh-PB|frataxin|frataxin homologue|frataxin-like', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9148080, end_position_on_the_genomic_accession=9149071, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31846 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31846, CurrentID=0, Status='live', Symbol='Bap111', Aliases='Dmel_CG7055, BAF57, BAP 111, BAP111, CG7055, Dalao, Dmel\\CG7055, anon-EST:Liang-1.83, clone 1.83, dalao', description='Brahma associated protein 111kD', other_designations='BRM-associated protein 111|Bap111-PA|CG7055-PA|brahma associated protein 111kD|dalao', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9149309, end_position_on_the_genomic_accession=9151976, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31848 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31848, CurrentID=0, Status='live', Symbol='His3.3B', Aliases='Dmel_CG8989, CG8989, Dmel\\CG8989, H3.3, H3.3B, His 3.3B, His3.3, HisH3-3Q, PH3, anon-EST:fe1B6, dH3.3B', description='Histone H3.3B', other_designations='CG8989-PA|CG8989-PB|CG8989-PC|CG8989-PE|His3.3B-PA|His3.3B-PB|His3.3B-PC|His3.3B-PE|HistoneH3|anon-fast-evolving-1B6|histone H3-3Q|histone H3.3|histone H3.3B|histone variant 3.3|histone variant H3.3', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9152523, end_position_on_the_genomic_accession=9156669, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31849 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31849, CurrentID=0, Status='live', Symbol='Ost48', Aliases='Dmel_CG9022, CG9022, DmOST50, Dmel\\CG9022, DrOST, OST, OST48, OST48/WBP1, Ots48, anon-EST:Liang-1.11, clone 1.11, ost48', description='Oligosaccharyltransferase 48kD subunit', other_designations='CG9022-PA|Ost48-PA|oligosaccaryltransferase 48kDa subunit|oligosaccharyltransferase 48kD subunit', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9156929, end_position_on_the_genomic_accession=9158515, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31851 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31851, CurrentID=0, Status='live', Symbol='Zpr1', Aliases='Dmel_CG9060, CG9060, Dmel\\CG9060', description='CG9060 gene product from transcript CG9060-RA', other_designations='CG9060-PA|Zpr1-PA|zinc finger protein 1', map_location='8C14-8C14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9160124, end_position_on_the_genomic_accession=9162090, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31853 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31853, CurrentID=0, Status='live', Symbol='CG12115', Aliases='Dmel_ BcDNA:LP02922, Dmel\\CG12115', description='CG12115 gene product from transcript CG12115-RA', other_designations='CG12115-PA|uncharacterized protein', map_location='8C17-8C17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9189326, end_position_on_the_genomic_accession=9190605, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31854 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31854, CurrentID=0, Status='live', Symbol='CG12057', Aliases='Dmel_ BcDNA:LP05863, Dmel\\CG12057', description='CG12057 gene product from transcript CG12057-RA', other_designations='CG12057-PA|CG12057-PB|uncharacterized protein', map_location='8C17-8C17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9191657, end_position_on_the_genomic_accession=9192548, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31855 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31855, CurrentID=0, Status='live', Symbol='CG12056', Aliases='Dmel_ Dmel\\CG12056', description='CG12056 gene product from transcript CG12056-RA', other_designations='CG12056-PA|uncharacterized protein', map_location='8C17-8C17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9192669, end_position_on_the_genomic_accession=9193620, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31856 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31856, CurrentID=0, Status='live', Symbol='CG12106', Aliases='Dmel_ BcDNA:GH03263, Dmel\\CG12106', description='CG12106 gene product from transcript CG12106-RA', other_designations='CG12106-PA|uncharacterized protein', map_location='8C17-8D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9193898, end_position_on_the_genomic_accession=9195333, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31857 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31857, CurrentID=0, Status='live', Symbol='CG12118', Aliases='Dmel_ Dmel\\CG12118', description='CG12118 gene product from transcript CG12118-RB', other_designations='CG12118-PA|CG12118-PB|uncharacterized protein', map_location='8D1-8D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9194750, end_position_on_the_genomic_accession=9196414, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31858 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31858, CurrentID=0, Status='live', Symbol='su(r)', Aliases='Dmel_CG2194, CG2194, Dmel\\CG2194, Dreg-3, Pyd1, Reg-3, pyd1', description='suppressor of rudimentary', other_designations='CG2194-PC|drosophila rhythmically expressed gene 3|rhythmically expressed gene 3|su(r)-PC', map_location='1-27.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9196808, end_position_on_the_genomic_accession=9201820, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31859 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31859, CurrentID=0, Status='live', Symbol='CG12119', Aliases='Dmel_ Dmel\\CG12119, lincRNA.971', description='CG12119 gene product from transcript CG12119-RB', other_designations='CG12119-PA|CG12119-PB|uncharacterized protein', map_location='8D1-8D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9206126, end_position_on_the_genomic_accession=9210887, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31860 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31860, CurrentID=0, Status='live', Symbol='Obp8a', Aliases='Dmel_CG12665, 8a, CG12665, DmelObp8a, Dmel\\CG12665', description='Odorant-binding protein 8a', other_designations='CG12665-PA|Obp8a-PA', map_location='8D1-8D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9214876, end_position_on_the_genomic_accession=9215699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31861 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31861, CurrentID=0, Status='live', Symbol='CG15368', Aliases='Dmel_ Dmel\\CG15368', description='CG15368 gene product from transcript CG15368-RA', other_designations='CG15368-PA|uncharacterized protein', map_location='8D1-8D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9215785, end_position_on_the_genomic_accession=9216540, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31862, CurrentID=0, Status='live', Symbol='CG15369', Aliases='Dmel_ Dmel\\CG15369', description='CG15369 gene product from transcript CG15369-RB', other_designations='CG15369-PA|CG15369-PB|uncharacterized protein', map_location='8D1-8D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9216872, end_position_on_the_genomic_accession=9217466, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31863 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31863, CurrentID=0, Status='live', Symbol='t', Aliases='Dmel_CG12120, CG12120, Dmel\\CG12120', description='tan', other_designations='CG12120-PA|CG12120-PB|Tan|on-transient defective|t-PA|t-PB', map_location='1-27.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9217655, end_position_on_the_genomic_accession=9223257, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31864, CurrentID=0, Status='live', Symbol='CG15370', Aliases='Dmel_ Dmel\\CG15370', description='CG15370 gene product from transcript CG15370-RA', other_designations='CG15370-PA|uncharacterized protein', map_location='8D2-8D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9225437, end_position_on_the_genomic_accession=9226688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31865 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31865, CurrentID=0, Status='live', Symbol='Gr8a', Aliases='Dmel_CG15371, CG15371, Dmel\\CG15371, GR8a, Gr8D1', description='Gustatory receptor 8a', other_designations='CG15371-PA|Gr8a-PA|gustatory receptor 8D1|gustatory receptor 8a', map_location='8D2-8D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9227137, end_position_on_the_genomic_accession=9228827, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31866 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31866, CurrentID=0, Status='live', Symbol='CG12121', Aliases='Dmel_ Dmel\\CG12121', description='CG12121 gene product from transcript CG12121-RA', other_designations='CG12121-PA|uncharacterized protein', map_location='8D2-8D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9228796, end_position_on_the_genomic_accession=9232378, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31867 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31867, CurrentID=0, Status='live', Symbol='Ir8a', Aliases='Dmel_CG32704, CG12122, CG12666, CG32704, CT7438, DmelIR8a, Dmel\\CG32704, IR8a', description='Ionotropic receptor 8a', other_designations='CG32704-PA|Ir8a-PA|ionotropic receptor 8a', map_location='8D2-8D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9232726, end_position_on_the_genomic_accession=9236585, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31869 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31869, CurrentID=0, Status='live', Symbol='mxc', Aliases='Dmel_CG12124, CG12058, CG12124, Dmel\\CG12124, Dmel_CG12058, MPM2, MXC, Mxc, NPAT, dNPAT, l(1)mbn', description='multi sex combs', other_designations='CG12124-PA|CG12124-PB|lethal(1)malignant blood neoplasm|mxc-PA|mxc-PB', map_location='1-27.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9236801, end_position_on_the_genomic_accession=9243467, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31872 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31872, CurrentID=0, Status='live', Symbol='Dsor1', Aliases='Dmel_CG15793, CG15793, D-MEK, D-MEK/Dsor, D-Mek, D-SOR, D-Sor, D-mek, D-sor, D-sor-1, DMEK-1, DRODSOR1, DSOR1, DSORT, DSor, DSor1, Dmek, Dmel\\CG15793, Dsor, EK1-1, MAPKK, MEK, MEK/Dsor1, MEK1/2, Mek, SOR, Sor, Su(Raf)34B, dMEK, dSor, dSor1, dsor1, mek, sor, sor/MEK1', description='Downstream of raf1', other_designations='CG15793-PA|CG15793-PB|Dsor1-PA|Dsor1-PB|MAPK kinase|Map kinase kinase|downstream of raf1|downstream suppressor of Raf 1|drosophila MEK|mitogen-activated protein kinase kinase', map_location='1-27.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9247342, end_position_on_the_genomic_accession=9250037, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31873 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31873, CurrentID=0, Status='live', Symbol='CG17754', Aliases='Dmel_ BcDNA:LP02641, Dmel\\CG17754', description='CG17754 gene product from transcript CG17754-RG', other_designations='CG17754-PA|CG17754-PC|CG17754-PD|CG17754-PE|CG17754-PF|CG17754-PG|uncharacterized protein', map_location='8D3-8D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9250420, end_position_on_the_genomic_accession=9255953, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31874 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31874, CurrentID=0, Status='live', Symbol='ric8a', Aliases='Dmel_CG15797, CG15797, DmRic8, Dmel\\CG15797, RIC-8, Ric-8, l(1)G0397, ric-8, ric-8a', description='CG15797 gene product from transcript CG15797-RB', other_designations='CG15797-PA|CG15797-PB|ric8a-PA|ric8a-PB', map_location='8D3-8D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9256168, end_position_on_the_genomic_accession=9259202, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31877 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31877, CurrentID=0, Status='live', Symbol='Erk7', Aliases='Dmel_CG32703, CG17441, CG2309, CG32703, Dmel\\CG32703, ERK7', description='Extracellularly regulated kinase 7', other_designations='CG32703-PA|CG32703-PB|CG32703-PC|Erk7-PA|Erk7-PB|Erk7-PC|extracellularly regulated kinase 7', map_location='8D3-8D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9261376, end_position_on_the_genomic_accession=9267698, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  31879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31879, CurrentID=0, Status='live', Symbol='CG17440', Aliases='Dmel_ Dmel\\CG17440', description='CG17440 gene product from transcript CG17440-RA', other_designations='CG17440-PA|uncharacterized protein', map_location='8D4-8D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9268438, end_position_on_the_genomic_accession=9269925, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31880 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31880, CurrentID=0, Status='live', Symbol='Cfp1', Aliases='Dmel_CG17446, CG17446, CXXC1, Dmel\\CG17446, dCXXC1, dCfp1', description='CXXC finger protein 1', other_designations='CG17446-PA|CG17446-PB|Cfp1-PA|Cfp1-PB', map_location='8D4-8D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9270434, end_position_on_the_genomic_accession=9272960, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31881 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31881, CurrentID=0, Status='live', Symbol='CG16892', Aliases='Dmel_ Aladin, Dmel\\CG16892', description='CG16892 gene product from transcript CG16892-RB', other_designations='CG16892-PA|CG16892-PB|uncharacterized protein', map_location='8D4-8D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9273413, end_position_on_the_genomic_accession=9275653, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31883 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31883, CurrentID=0, Status='live', Symbol='lz', Aliases='Dmel_CG1689, CG1689, Dmel\\CG1689, Lz, amx, fs(1)A1569, fs(1)M69, spe', description='lozenge', other_designations='CG1689-PA|CG1689-PB|lz-PA|lz-PB|spectacled', map_location='1-27.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9284649, end_position_on_the_genomic_accession=9303636, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31887 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31887, CurrentID=0, Status='live', Symbol='CG34449', Aliases='Dmel_ CG12136, CG15372, CG17251, CG32705, Dm_X:47283, Dmel\\CG34449, Dmel_CG32705, Pen16, anon-Pen16, p16, pen16', description='CG34449 gene product from transcript CG34449-RA', other_designations='CG34449-PA|CG34449-PB|CG34449-PC|CG34449-PD|CG34449-PE|CG34449-PF|CG34449-PG|uncharacterized protein', map_location='8D7-8D9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9324508, end_position_on_the_genomic_accession=9356227, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  31890 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31890, CurrentID=0, Status='live', Symbol='BCL7-like', Aliases='Dmel_CG17252, Bcl7-like, CG17252, Dmel\\CG17252', description='CG17252 gene product from transcript CG17252-RA', other_designations='BCL7-like-PA|CG17252-PA', map_location='8D9-8D9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9356634, end_position_on_the_genomic_accession=9358594, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  10178862 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178862, CurrentID=0, Status='live', Symbol='CG42714', Aliases='Dmel_ DMG8-chr3L.11.020, Dmel\\CG42714', description='CG42714 gene product from transcript CG42714-RA', other_designations='CG42714-PA|uncharacterized protein', map_location='65A10-65A10', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=6233491, end_position_on_the_genomic_accession=6234056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31896 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31896, CurrentID=0, Status='live', Symbol='CG32700', Aliases='Dmel_ BcDNA:GH19985, CG12651, CG12652, CG18429, CG2994, CR43497, Dmel\\CG32700, Dmel_CR43497, Tes89', description='CG32700 gene product from transcript CG32700-RB', other_designations='CG32700-PA|CG32700-PB|uncharacterized protein', map_location='8E4-8E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9502826, end_position_on_the_genomic_accession=9550838, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31897 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31897, CurrentID=0, Status='live', Symbol='RpS28b', Aliases='Dmel_CG2998, CG2998, Dmel\\CG2998, RS28_DROME', description='Ribosomal protein S28b', other_designations='CG2998-PA|CG2998-PB|RpS28b-PA|RpS28b-PB|ribosomal protein S28b', map_location='8E7-8E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9554623, end_position_on_the_genomic_accession=9555661, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31898 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31898, CurrentID=0, Status='live', Symbol='CG15317', Aliases='Dmel_ Dmel\\CG15317', description='CG15317 gene product from transcript CG15317-RA', other_designations='CG15317-PA|uncharacterized protein', map_location='8E7-8E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9555607, end_position_on_the_genomic_accession=9556544, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31899 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31899, CurrentID=0, Status='live', Symbol='LPCAT', Aliases='Dmel_CG32699, CG15318, CG18535, CG2996, CG32699, CT7882, Dmel\\CG326991/2, tok-like, LPCAT', description='Lysophosphatidylcholine acyltransferase', other_designations='CG32699-PB|LPCAT-PB|Lyso-PA acyltransferase|lysophosphatidylcholine acyltransferase', map_location='8E10-8E10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9571537, end_position_on_the_genomic_accession=9579987, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31902 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31902, CurrentID=0, Status='live', Symbol='Gga', Aliases='Dmel_CG3002, CG3002, DmGGA1, Dmel\\CG3002, GGA, dGGA, dmGGA', description='Golgi-localized, gamma-adaptin ear containing, ARF binding protein', other_designations='CG3002-PB|Gga-PB|Golgi-localised, gamma-ear containing, ADP ribosylation factor-binding', map_location='8F1-8F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9601030, end_position_on_the_genomic_accession=9604280, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  31903 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31903, CurrentID=0, Status='live', Symbol='Lst8', Aliases='Dmel_CG3004, CG3004, Dmel\\CG3004, Gbl, dGbetaL, dLST8, lst8', description='CG3004 gene product from transcript CG3004-RA', other_designations='CG3004-PA|Lst8-PA', map_location='8F1-8F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9604154, end_position_on_the_genomic_accession=9605969, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31904 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31904, CurrentID=0, Status='live', Symbol='Aats-lys', Aliases='Dmel_CG12141, CG12141, Dmel\\CG12141, KRS, LysRS', description='Lysyl-tRNA synthetase', other_designations='Aats-lys-PA|Aats-lys-PB|Aats-lys-PC|CG12141-PA|CG12141-PB|CG12141-PC', map_location='8F1-8F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9606363, end_position_on_the_genomic_accession=9608884, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31905 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31905, CurrentID=0, Status='live', Symbol='CG42797', Aliases='Dmel_ BT058014.1, CG3003, CG3099, Dmel\\CG42797, Dmel_CG3003, Dmel_CG3099', description='CG42797 gene product from transcript CG42797-RE', other_designations='CG42797-PC|CG42797-PD|CG42797-PE|CG42797-PF|uncharacterized protein', map_location='8F2-8F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9608946, end_position_on_the_genomic_accession=9623243, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  31907 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31907, CurrentID=0, Status='live', Symbol='Ptpmeg2', Aliases='Dmel_CG32697, CG3101, CG3102, CG32697, Dmel\\CG32697, L(1)G0232, MEG2, PTP-Meg2, dMeg2, l(1)G0232, lG0232', description='Protein tyrosine phosphatase Meg2', other_designations='CG32697-PA|CG32697-PB|CG32697-PC|CG32697-PD|CG32697-PE|CG32697-PF|CG32697-PG|CG32697-PH|CG32697-PI|Ptpmeg2-PA|Ptpmeg2-PB|Ptpmeg2-PC|Ptpmeg2-PD|Ptpmeg2-PE|Ptpmeg2-PF|Ptpmeg2-PG|Ptpmeg2-PH|Ptpmeg2-PI|lethal (1) G0232|protein tyrosine phosphatase Meg2', map_location='8F2-8F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9626294, end_position_on_the_genomic_accession=9659145, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  31909 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31909, CurrentID=0, Status='live', Symbol='CG3106', Aliases='Dmel_ Dmel\\CG3106', description='CG3106 gene product from transcript CG3106-RB', other_designations='CG3106-PA|CG3106-PB', map_location='8F6-8F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9659095, end_position_on_the_genomic_accession=9662903, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31912 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31912, CurrentID=0, Status='live', Symbol='btd', Aliases='Dmel_CG12653, BTD, Btd, CG12653, Dmel\\CG12653', description='buttonhead', other_designations='CG12653-PA|btd-PA|button-head', map_location='1-31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9694186, end_position_on_the_genomic_accession=9697558, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31913 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31913, CurrentID=0, Status='live', Symbol='Sp1', Aliases='Dmel_CG1343, CG1343, D- D-Sp8, D.Sp1, Dmel\\CG1343, SP1, dSp1, sp1', description='CG1343 gene product from transcript CG1343-RE', other_designations='CG1343-PB|CG1343-PD|CG1343-PE|CG1343-PF|Sp1-PB|Sp1-PD|Sp1-PE|Sp1-PF', map_location='9A1-9A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9729648, end_position_on_the_genomic_accession=9756369, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31914 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31914, CurrentID=0, Status='live', Symbol='CG1354', Aliases='Dmel_ Dmel\\CG1354', description='CG1354 gene product from transcript CG1354-RD', other_designations='CG1354-PA|CG1354-PB|CG1354-PC|CG1354-PD|CG1354-PE', map_location='9A1-9A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9778742, end_position_on_the_genomic_accession=9782094, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31915 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31915, CurrentID=0, Status='live', Symbol='CG32698', Aliases='Dmel_ CG9678, CG9683, Dmel\\CG32698', description='CG32698 gene product from transcript CG32698-RD', other_designations='CG32698-PA|CG32698-PB|CG32698-PC|CG32698-PD|uncharacterized protein', map_location='9A2-9A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9783223, end_position_on_the_genomic_accession=9855449, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  31917 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31917, CurrentID=0, Status='live', Symbol='Ser7', Aliases='Dmel_CG2045, CG2045, Dmel\\CG2045, SP#2, SP2, SP3, c-SP3', description='CG2045 gene product from transcript CG2045-RA', other_designations='CG2045-PA|Ser7-PA|serine protease 2', map_location='9A2-9A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9865233, end_position_on_the_genomic_accession=9867070, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31918 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31918, CurrentID=0, Status='live', Symbol='CG32695', Aliases='Dmel_ BcDNA:RH68724, CG15246, Dmel\\CG32695', description='CG32695 gene product from transcript CG32695-RC', other_designations='CG32695-PA|CG32695-PB|CG32695-PC|uncharacterized protein', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9872382, end_position_on_the_genomic_accession=9874057, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31919 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31919, CurrentID=0, Status='live', Symbol='CG15247', Aliases='Dmel_ Dmel\\CG15247', description='CG15247 gene product from transcript CG15247-RA', other_designations='CG15247-PA|uncharacterized protein', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9874493, end_position_on_the_genomic_accession=9876672, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31920 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31920, CurrentID=0, Status='live', Symbol='CG1468', Aliases='Dmel_ BcDNA:GH18955, Dmel\\CG1468', description='CG1468 gene product from transcript CG1468-RA', other_designations='CG1468-PA', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9878422, end_position_on_the_genomic_accession=9879279, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31921 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31921, CurrentID=0, Status='live', Symbol='CG9686', Aliases='Dmel_ BcDNA:RH74092, CT6802, Dmel\\CG9686, anon-EST:Posey3', description='CG9686 gene product from transcript CG9686-RA', other_designations='CG9686-PA', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9879870, end_position_on_the_genomic_accession=9880535, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31922 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31922, CurrentID=0, Status='live', Symbol='CG9689', Aliases='Dmel_ BcDNA:GH05731, Dmel\\CG9689, lincRNA.S9020', description='CG9689 gene product from transcript CG9689-RA', other_designations='CG9689-PA|CG9689-PB', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9881953, end_position_on_the_genomic_accession=9883670, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31923 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31923, CurrentID=0, Status='live', Symbol='CG9691', Aliases='Dmel_ BcDNA:RH12290, CT27394, Dmel\\CG9691', description='CG9691 gene product from transcript CG9691-RB', other_designations='CG9691-PA|CG9691-PB', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9884442, end_position_on_the_genomic_accession=9885044, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31924 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31924, CurrentID=0, Status='live', Symbol='CG15249', Aliases='Dmel_ Dmel\\CG15249', description='CG15249 gene product from transcript CG15249-RB', other_designations='CG15249-PB|CG15249-PD|uncharacterized protein', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9887972, end_position_on_the_genomic_accession=9899585, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31925 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31925, CurrentID=0, Status='live', Symbol='CG1986', Aliases='Dmel_ Dmel\\CG1986', description='CG1986 gene product from transcript CG1986-RB', other_designations='CG1986-PA|CG1986-PB', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9908663, end_position_on_the_genomic_accession=9911126, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31926 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31926, CurrentID=0, Status='live', Symbol='Idgf4', Aliases='Dmel_CG1780, CG1780, DmIDGF4, Dmel\\CG1780, IDGF, IDGF-4, IDGF4, idgf4', description='Imaginal disc growth factor 4', other_designations='CG1780-PA|CG1780-PB|CG1780-PC|Idgf4-PA|Idgf4-PB|Idgf4-PC|imaginal disc growth factor 4', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9927973, end_position_on_the_genomic_accession=9930948, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31927 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31927, CurrentID=0, Status='live', Symbol='CG1791', Aliases='Dmel_ AAF46535, AAM52597, Dmel\\CG1791, FBpp0071364', description='CG1791 gene product from transcript CG1791-RC', other_designations='CG1791-PA|CG1791-PB|CG1791-PC|Fibrinogen-related protein', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9963453, end_position_on_the_genomic_accession=9964878, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31928 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31928, CurrentID=0, Status='live', Symbol='CG1889', Aliases='Dmel_ AAF46536, AAN09619, Dmel\\CG1889, FBpp0071343, FBpp0071344', description='CG1889 gene product from transcript CG1889-RA', other_designations='CG1889-PA|CG1889-PC|CG1889-PD|CG1889-PE|Fibrinogen-related protein', map_location='9A3-9A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9965241, end_position_on_the_genomic_accession=9966959, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31929 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31929, CurrentID=0, Status='live', Symbol='CG43386', Aliases='Dmel_ BcDNA:RE65233, CG15250, CG15251, Dmel\\CG43386, Dmel_CG15250, Dmel_CG15251', description='CG43386 gene product from transcript CG43386-RB', other_designations='CG43386-PB|CG43386-PC|CG43386-PD|uncharacterized protein', map_location='9A3-9A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9966615, end_position_on_the_genomic_accession=9977405, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31931 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31931, CurrentID=0, Status='live', Symbol='CG15252', Aliases='Dmel_ Dmel\\CG15252', description='CG15252 gene product from transcript CG15252-RB', other_designations='CG15252-PB|CG15252-PC|uncharacterized protein', map_location='9A4-9A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9978153, end_position_on_the_genomic_accession=9980564, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31933 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31933, CurrentID=0, Status='live', Symbol='CG32694', Aliases='Dmel_ BcDNA:RE02049, CG12634, CG12649, CG18512, Dmel\\CG32694', description='CG32694 gene product from transcript CG32694-RH', other_designations='CG32694-PA|CG32694-PB|CG32694-PC|CG32694-PE|CG32694-PF|CG32694-PG|CG32694-PH|uncharacterized protein', map_location='9A4-9A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=9989690, end_position_on_the_genomic_accession=10006882, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  31934 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31934, CurrentID=0, Status='live', Symbol='CG2990', Aliases='Dmel_ Dmel\\CG2990', description='CG2990 gene product from transcript CG2990-RB', other_designations='CG2990-PB|CG2990-PC', map_location='9A4-9A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10010481, end_position_on_the_genomic_accession=10014279, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31935 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31935, CurrentID=0, Status='live', Symbol='Cht6', Aliases='Dmel_CG43374, CG15313, CG15314, CG2989, CG43374, CT5624, Dmel\\CG43374, Dmel_CG15313, Dmel_CG15314, Dmel_CG2989', description='CG43374 gene product from transcript CG43374-RK', other_designations='CG43374-PC|CG43374-PF|CG43374-PG|CG43374-PH|CG43374-PI|CG43374-PJ|CG43374-PK|Cht6-PC|Cht6-PF|Cht6-PG|Cht6-PH|Cht6-PI|Cht6-PJ|Cht6-PK', map_location='9A5-9A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10016845, end_position_on_the_genomic_accession=10050803, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  31938 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31938, CurrentID=0, Status='live', Symbol='Yp2', Aliases='Dmel_CG2979, CG2979, Dmel\\CG2979, VG2, VIT2Dm, YP, YP-2, YP2, YPII, YP[[2]], fs(1)A59, fs(1)K313, fs(1)M35, fs(1)m35, yp, yp2', description='Yolk protein 2', other_designations='CG2979-PA|CG2979-PB|Yp2-PA|Yp2-PB|vitellogenin|yolk protein|yolk protein 2', map_location='1-30 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10050951, end_position_on_the_genomic_accession=10052636, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31939 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31939, CurrentID=0, Status='live', Symbol='Yp1', Aliases='Dmel_CG2985, CG2985, DmYP1, Dmel\\CG2985, VG1, VIT1Dm, YP, YP-1, YP1, YPI, YP[[1]], fs(1)1163, fs(1)A1163, yp, yp-1, yp1', description='Yolk protein 1', other_designations='CG2985-PA|CG2985-PB|Yok protein 1|Yp1-PA|Yp1-PB|vitellogenin|yolk protein|yolk protein 1|yolk protein I|yolk-protein 1', map_location='1-30 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10053811, end_position_on_the_genomic_accession=10055498, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31940 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31940, CurrentID=0, Status='live', Symbol='CG15312', Aliases='Dmel_ CT35293, Dmel\\CG15312', description='CG15312 gene product from transcript CG15312-RE', other_designations='CG15312-PA|CG15312-PB|CG15312-PC|CG15312-PD|CG15312-PE|CG15312-PF|uncharacterized protein', map_location='9B1-9B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10055629, end_position_on_the_genomic_accession=10070868, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31942 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31942, CurrentID=0, Status='live', Symbol='ZAP3', Aliases='Dmel_CG32685, CG12646, CG12647, CG2971, CG32685, Dmel\\CG32685', description='CG32685 gene product from transcript CG32685-RC', other_designations='CG32685-PC|CG32685-PD|CG32685-PE|CG32685-PF|ZAP3-PC|ZAP3-PD|ZAP3-PE|ZAP3-PF', map_location='9B1-9B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10071858, end_position_on_the_genomic_accession=10082492, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  31943 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31943, CurrentID=0, Status='live', Symbol='CG2972', Aliases='Dmel_ Dmel\\CG2972', description='CG2972 gene product from transcript CG2972-RB', other_designations='CG2972-PA|CG2972-PB', map_location='9B1-9B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10083154, end_position_on_the_genomic_accession=10085050, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31944 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31944, CurrentID=0, Status='live', Symbol='CG2974', Aliases='Dmel_ Dmel\\CG2974', description='CG2974 gene product from transcript CG2974-RB', other_designations='CG2974-PB', map_location='9B1-9B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10084914, end_position_on_the_genomic_accession=10086171, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31945 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31945, CurrentID=0, Status='live', Symbol='RhoU', Aliases='Dmel_CG34104, CG12094, CG12102, CG34104, Dmel\\CG34104', description='CG34104 gene product from transcript CG34104-RB', other_designations='CG34104-PA|CG34104-PB|CG34104-PC|RhoU-PA|RhoU-PB|RhoU-PC', map_location='9B1-9B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10086527, end_position_on_the_genomic_accession=10148876, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  31947 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31947, CurrentID=0, Status='live', Symbol='CG12645', Aliases='Dmel_ Dmel\\CG12645', description='CG12645 gene product from transcript CG12645-RB', other_designations='CG12645-PB|CG12645-PC|uncharacterized protein', map_location='9B2-9B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10153612, end_position_on_the_genomic_accession=10155009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  31948 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31948, CurrentID=0, Status='live', Symbol='CG43902', Aliases='Dmel_ CG12644, CG15311, Dmel\\CG43902, Dmel_CG12644, Dmel_CG15311', description='CG43902 gene product from transcript CG43902-RA', other_designations='CG43902-PA|uncharacterized protein', map_location='9B2-9B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10158434, end_position_on_the_genomic_accession=10195337, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31949 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31949, CurrentID=0, Status='live', Symbol='CG15309', Aliases='Dmel_ BcDNA:GH10478, Dmel\\CG15309', description='CG15309 gene product from transcript CG15309-RC', other_designations='CG15309-PA|CG15309-PB|CG15309-PC|CG15309-PD|CG15309-PE|uncharacterized protein', map_location='9B4-9B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10205292, end_position_on_the_genomic_accession=10212374, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31950 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31950, CurrentID=0, Status='live', Symbol='ATPsyndelta', Aliases='Dmel_CG2968, ATP5D, ATPD, CG2968, Dmel\\CG2968, anon-WO03040301.109, anon-WO03040301.110, l (1) G0230, l(1)G0230', description='ATP synthase, delta subunit', other_designations='ATP synthase delta chain|ATPsyndelta-PA|ATPsyndelta-PB|ATPsyndelta-PC|CG2968-PA|CG2968-PB|CG2968-PC|F0/F1 ATP synthase delta|lethal (1) G0230', map_location='9B4-9B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10212970, end_position_on_the_genomic_accession=10217393, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31952 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31952, CurrentID=0, Status='live', Symbol='CG2962', Aliases='Dmel_ Dmel\\CG2962', description='CG2962 gene product from transcript CG2962-RA', other_designations='CG2962-PA', map_location='9B4-9B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10214858, end_position_on_the_genomic_accession=10216642, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  261329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261329, CurrentID=0, Status='live', Symbol='CG8176', Aliases='Dmel_ CG31123, Dmel\\CG8176, FCHo, FCHo2', description='CG8176 gene product from transcript CG8176-RF', other_designations='CG8176-PA|CG8176-PC|CG8176-PE|CG8176-PF|CG8176-PG', map_location='85D19-85D19', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9516007, end_position_on_the_genomic_accession=9527237, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  31954 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31954, CurrentID=0, Status='live', Symbol='Ipod', Aliases='Dmel_CG2961, CG2961, Dmel\\CG2961, IPOD', description='Interaction partner of Dnmt2', other_designations='CG2961-PA|CG2961-PB|CG2961-PC|Ipod-PA|Ipod-PB|Ipod-PC|interaction partner of Dnmt2', map_location='9B4-9B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10231002, end_position_on_the_genomic_accession=10233080, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31955 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31955, CurrentID=0, Status='live', Symbol='Hk', Aliases='Dmel_CG43388, CG17779, CG2287, CG32688, CG34319, CG43388, Dm_X:51771, Dmel\\CG43388, Dmel_CG32688, Dmel_CG34319, hk, lincRNA.975', description='Hyperkinetic', other_designations='CG43388-PD|CG43388-PE|CG43388-PF|CG43388-PG|CG43388-PH|CG43388-PI|CG43388-PK|CG43388-PL|CG43388-PM|Hk-PD|Hk-PE|Hk-PF|Hk-PG|Hk-PH|Hk-PI|Hk-PK|Hk-PL|Hk-PM|hyperk|hyperkinetic|kvbeta channel subunit', map_location='1-30.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10233347, end_position_on_the_genomic_accession=10263485, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  31957 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31957, CurrentID=0, Status='live', Symbol='alpha-Man-Ia', Aliases="Dmel_CG42275, CG17809, CG18799, CG32682, CG32684, CG42275, DM-GI, Dmel\\CG42275, Dmel_CG32682, Dmel_CG32684, MA12, MAN-1, Man'ase, ManI, alpha-Man-I, alpha-Man-l, dManI, dMas-1, mas-1, mas1", description='alpha-Mannosidase class I a', other_designations='CG42275-PH|CG42275-PI|CG42275-PJ|CG42275-PK|CG42275-PL|CG42275-PM|CG42275-PN|CG42275-PO|CG42275-PP|alpha 1,2 mannosidase|alpha mannosidase I|alpha-1,2-mannosidase I|alpha-D-mannosidase|alpha-Man-Ia-PH|alpha-Man-Ia-PI|alpha-Man-Ia-PJ|alpha-Man-Ia-PK|alpha-Man-Ia-PL|alpha-Man-Ia-PM|alpha-Man-Ia-PN|alpha-Man-Ia-PO|alpha-Man-Ia-PP|alpha-mannosidase|mannosidase 1', map_location='1-30.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10266792, end_position_on_the_genomic_accession=10325958, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  31958 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31958, CurrentID=0, Status='live', Symbol='CG2909', Aliases='Dmel_ Dmel\\CG2909', description='CG2909 gene product from transcript CG2909-RA', other_designations='CG2909-PA', map_location='9B7-9B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10326464, end_position_on_the_genomic_accession=10328946, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31960 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31960, CurrentID=0, Status='live', Symbol='Gip', Aliases='Dmel_CG2227, CG2227, Dmel\\CG2227, T15, gene C', description='GIP-like', other_designations='CG2227-PA|Gip-PA|glyoxylate-induced protein|hydroxypyruvate isomerase', map_location='9B7-9B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10328877, end_position_on_the_genomic_accession=10329872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31961 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31961, CurrentID=0, Status='live', Symbol='CG15306', Aliases='Dmel_ Dmel\\CG15306', description='CG15306 gene product from transcript CG15306-RA', other_designations='CG15306-PA|CG15306-PB|uncharacterized protein', map_location='9B7-9B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10330160, end_position_on_the_genomic_accession=10331818, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31962 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31962, CurrentID=0, Status='live', Symbol='CG43740', Aliases='Dmel_ CG12641, CG15307, Dmel\\CG43740, Dmel_CG12641, Dmel_CG15307', description='CG43740 gene product from transcript CG43740-RA', other_designations='CG43740-PA|CG43740-PB|CG43740-PC|uncharacterized protein', map_location='9B7-9B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10333652, end_position_on_the_genomic_accession=10347794, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31964 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31964, CurrentID=0, Status='live', Symbol='l(1)G0289', Aliases='Dmel_CG2221, CG2221, Dmel\\CG2221, Mgn1457', description='lethal (1) G0289', other_designations='CG2221-PA|CG2221-PB|CG2221-PC|l(1)G0289-PA|l(1)G0289-PB|l(1)G0289-PC', map_location='9B11-9B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10348680, end_position_on_the_genomic_accession=10366273, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31965 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31965, CurrentID=0, Status='live', Symbol='CG17841', Aliases='Dmel_ BcDNA:GH12326, Dmel\\CG17841, anon-EST:Liang-2.12, clone 2.12', description='CG17841 gene product from transcript CG17841-RA', other_designations='CG17841-PA|uncharacterized protein', map_location='9B12-9B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10368052, end_position_on_the_genomic_accession=10383580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31966 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31966, CurrentID=0, Status='live', Symbol='CG2898', Aliases='Dmel_ Dmel\\CG2898', description='CG2898 gene product from transcript CG2898-RA', other_designations='CG2898-PA', map_location='9B12-9B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10375374, end_position_on_the_genomic_accession=10376628, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31967 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31967, CurrentID=0, Status='live', Symbol='Psf3', Aliases='Dmel_CG2222, CG2222, Dmel\\CG2222', description='CG2222 gene product from transcript CG2222-RC', other_designations='CG2222-PA|CG2222-PB|CG2222-PC|Psf3-PA|Psf3-PB|Psf3-PC', map_location='9B14-9B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10383353, end_position_on_the_genomic_accession=10385861, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  31970 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31970, CurrentID=0, Status='live', Symbol='CG32683', Aliases='Dmel_ CG15303, CG2881, CG2883, Dmel\\CG32683', description='CG32683 gene product from transcript CG32683-RA', other_designations='CG32683-PA|CG32683-PB|uncharacterized protein', map_location='9C1-9C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10410391, end_position_on_the_genomic_accession=10454895, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  31971 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31971, CurrentID=0, Status='live', Symbol='RabX2', Aliases='Dmel_CG2885, CG2885, Dm Dmel\\CG2885', description='CG2885 gene product from transcript CG2885-RA', other_designations='CG2885-PA|RabX2-PA', map_location='9C1-9C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10417429, end_position_on_the_genomic_accession=10418219, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  31973 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31973, CurrentID=0, Status='live', Symbol='CG12640', Aliases='Dmel_ Dmel\\CG12640', description='CG12640 gene product from transcript CG12640-RA', other_designations='CG12640-PA|uncharacterized protein', map_location='9C2-9C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10432465, end_position_on_the_genomic_accession=10432812, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31975 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31975, CurrentID=0, Status='live', Symbol='Or9a', Aliases='Dmel_CG15302, 9E.1, 9a, CG15302, DOR95, DmOr9a, Dmel\\CG15302, OR9a, or9a', description='Odorant receptor 9a', other_designations='CG15302-PA|Or9a-PA|odorant receptor 9a', map_location='9C4-9C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10458186, end_position_on_the_genomic_accession=10459676, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31976 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31976, CurrentID=0, Status='live', Symbol='nocte', Aliases='Dmel_CG17255, CG17255, Dmel\\CG17255', description='no circadian temperature entrainment', other_designations='CG17255-PA|CG17255-PB|CG17255-PC|CG17255-PD|nocte-PA|nocte-PB|nocte-PC|nocte-PD', map_location='9C5-9D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10461865, end_position_on_the_genomic_accession=10474675, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  31977 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31977, CurrentID=0, Status='live', Symbol='CG2889', Aliases='Dmel_ Dmel\\CG2889', description='CG2889 gene product from transcript CG2889-RA', other_designations='CG2889-PA|CG2889-PB', map_location='9D1-9D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10473938, end_position_on_the_genomic_accession=10477201, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  31978 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31978, CurrentID=0, Status='live', Symbol='CG2887', Aliases='Dmel_ Dmel\\CG2887', description='CG2887 gene product from transcript CG2887-RA', other_designations='CG2887-PA', map_location='9D1-9D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10477260, end_position_on_the_genomic_accession=10478493, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31979 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31979, CurrentID=0, Status='live', Symbol='PPP4R2r', Aliases='Dmel_CG2890, BcDNA:LD29930, CG2890, Dmel\\CG2890, PP4, PP4R, PP4R2, PP4R2r, PPP4R2, Ppp4R2, R2', description='Protein phosphatase 4 regulatory subunit 2-related protein', other_designations='CG2890-PA|CG2890-PB|CG2890-PC|CG2890-PD|PP4 regulatory subunit|PPP4R2r-PA|PPP4R2r-PB|PPP4R2r-PC|PPP4R2r-PD|protein phosphatase 4 regulatory subunit 2-related protein', map_location='9D1-9D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10478838, end_position_on_the_genomic_accession=10481943, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31980 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31980, CurrentID=0, Status='live', Symbol='CG32687', Aliases='Dmel_ BcDNA:RH29237, CG2143, CG2892, CT6984, CT7080, Cg32687, Dmel\\CG32687', description='CG32687 gene product from transcript CG32687-RB', other_designations='CG32687-PA|CG32687-PB|uncharacterized protein', map_location='9D2-9D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10482028, end_position_on_the_genomic_accession=10487205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31986 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31986, CurrentID=0, Status='live', Symbol='CG15296', Aliases='Dmel_ BcDNA:AT26906, Dmel\\CG15296', description='CG15296 gene product from transcript CG15296-RB', other_designations='CG15296-PA|CG15296-PB|uncharacterized protein', map_location='9D3-9D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10565842, end_position_on_the_genomic_accession=10566591, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  31987 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31987, CurrentID=0, Status='live', Symbol='spri', Aliases='Dmel_CG34414, CG12638, CG15297, CG15298, CG15299, CG15300, CG15301, CG32674, CG32680, CG33175, CG34414, Dmel\\CG34414, Dmel_CG32680, Dmel_CG33175, Rin1, Spri, anon-WO03040301.275', description='sprint', other_designations='CG34414-PA|CG34414-PE|CG34414-PF|CG34414-PG|CG34414-PI|CG34414-PJ|CG34414-PL|CG34414-PM|CG34414-PN|CG34414-PO|CG34414-PP|CG34414-PQ|CG34414-PR|spri-PA|spri-PE|spri-PF|spri-PG|spri-PI|spri-PJ|spri-PL|spri-PM|spri-PN|spri-PO|spri-PP|spri-PQ|spri-PR|sprint-a|sprint-b', map_location='9D2-9D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10490069, end_position_on_the_genomic_accession=10585422, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  31988 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31988, CurrentID=0, Status='live', Symbol='Hmr', Aliases='Dmel_CG1619, BcDNA:LD22117, CG1619, Dmel\\CG1619, HMR-mel, Hmr[[mel]], hmr, Hmr', description='Hybrid male rescue', other_designations='CG1619-PB|Hmr-PB|hybrid male rescue', map_location='1-31.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10590920, end_position_on_the_genomic_accession=10595855, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  31989 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31989, CurrentID=0, Status='live', Symbol='CG2124', Aliases='Dmel_ Dmel\\CG2124', description='CG2124 gene product from transcript CG2124-RA', other_designations='CG2124-PA', map_location='9D3-9D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10596122, end_position_on_the_genomic_accession=10599169, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  31990 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31990, CurrentID=0, Status='live', Symbol='CG1628', Aliases='Dmel_ Dmel\\CG1628', description='CG1628 gene product from transcript CG1628-RC', other_designations='CG1628-PB|CG1628-PC', map_location='9D3-9D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10598146, end_position_on_the_genomic_accession=10602714, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  31991 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31991, CurrentID=0, Status='live', Symbol='CG43347', Aliases='Dmel_ BcDNA:RH11383, CG15295, CG9817, Dmel\\CG43347, Dmel_CG15295, Dmel_CG9817', description='CG43347 gene product from transcript CG43347-RC', other_designations='CG43347-PB|CG43347-PC|CG43347-PD|uncharacterized protein', map_location='9D4-9D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10602641, end_position_on_the_genomic_accession=10620190, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  31993 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31993, CurrentID=0, Status='live', Symbol='Rab9Db', Aliases='Dmel_CG9807, CG9807, Dmel\\CG9807', description='Rab at 9Db', other_designations='CG9807-PA|Rab GTPase 9Db|Rab9Db-PA', map_location='9D4-9D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10622355, end_position_on_the_genomic_accession=10623225, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31994 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31994, CurrentID=0, Status='live', Symbol='CG9806', Aliases='Dmel_ Dmel\\CG9806', description='CG9806 gene product from transcript CG9806-RA', other_designations='CG9806-PA', map_location='9D4-9D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10623725, end_position_on_the_genomic_accession=10627112, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  31995 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31995, CurrentID=0, Status='live', Symbol='CG2111', Aliases='Dmel_ Dmel\\CG2111, pepCG2111', description='CG2111 gene product from transcript CG2111-RA', other_designations='CG2111-PA', map_location='9E1-9E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10657680, end_position_on_the_genomic_accession=10660671, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  31996 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31996, CurrentID=0, Status='live', Symbol='CG12637', Aliases='Dmel_ Dmel\\CG12637', description='CG12637 gene product from transcript CG12637-RA', other_designations='CG12637-PA|uncharacterized protein', map_location='9E1-9E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10672433, end_position_on_the_genomic_accession=10673466, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  31997 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=31997, CurrentID=0, Status='live', Symbol='X11Lbeta', Aliases='Dmel_CG32677, CG1852, CG1861, CG32677, Dmel\\CG32677, Dmint2, X11B, x11Lbeta', description='CG32677 gene product from transcript CG32677-RE', other_designations='CG32677-PA|CG32677-PD|CG32677-PE|X11Lbeta-PA|X11Lbeta-PD|X11Lbeta-PE', map_location='9D4-9E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10627737, end_position_on_the_genomic_accession=10710947, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  32000 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32000, CurrentID=0, Status='live', Symbol='BTBD9', Aliases='Dmel_CG1826, CG1826, Dmel\\CG1826, dBTBD9', description='BTB (POZ) domain containing 9 ortholog', other_designations='BTBD9-PA|BTBD9-PB|CG1826-PA|CG1826-PB', map_location='9E7-9E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10768537, end_position_on_the_genomic_accession=10772154, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32001 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32001, CurrentID=0, Status='live', Symbol='Atg8a', Aliases='Dmel_CG32672, ATG8, ATG8/LC3, ATG8a, Atg8, Atg8/LC3, Atg8A, BcDNA:LD05816, CG1534, CG32672, Dmel\\CG32672, DrAtg8a, LC3, LC3/Atg8, atg8, atg8A, atg8a', description='Autophagy-related 8a', other_designations='Atg8a-PA|Atg8a-PB|Atg8a-PC|Autophagy-specific gene 8a|CG32672-PA|CG32672-PB|CG32672-PC|autophagy-related 8a', map_location='9E6-9E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10764878, end_position_on_the_genomic_accession=10768338, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32002 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32002, CurrentID=0, Status='live', Symbol='Rph', Aliases='Dmel_CG11556, CG11556, Dmel\\CG11556', description='Rabphilin', other_designations='CG11556-PA|CG11556-PB|CG11556-PC|CG11556-PD|Rph-PA|Rph-PB|Rph-PC|Rph-PD|dm-rabphilin|rabphilin|rabphilin-3A', map_location='9E2-9E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10749077, end_position_on_the_genomic_accession=10757666, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32005 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32005, CurrentID=0, Status='live', Symbol='CG32676', Aliases='Dmel_ CG9725, CG9732, Dmel\\CG32676', description='CG32676 gene product from transcript CG32676-RB', other_designations='CG32676-PA|CG32676-PB|uncharacterized protein', map_location='9E1-9E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10711163, end_position_on_the_genomic_accession=10744943, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32006 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32006, CurrentID=0, Status='live', Symbol='CG15211', Aliases='Dmel_ Dmel\\CG15211', description='CG15211 gene product from transcript CG15211-RA', other_designations='CG15211-PA|CG15211-PB|CG15211-PC|CG15211-PD|CG15211-PE|uncharacterized protein', map_location='9E8-9E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10772541, end_position_on_the_genomic_accession=10778981, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32007 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32007, CurrentID=0, Status='live', Symbol='sesB', Aliases='Dmel_CG16944, A/A-T, A/A-T/sesB, ADP/ATP translocase, ADT2, ANT, ANT1, CG16944, Dmel\\CG16944, EM12, Hmr, S12, SesB, anon-WO02059370.55, dANT, jive, l(1)9Ed, l(1)9Fa, l(1)DC701, l(1)EM31, l(1)G0126, l(1)G0247, l(1)G0386, l(1)S12, orangi, ses B', description='stress-sensitive B', other_designations='ADP,ATP carrier protein|CG16944-PA|CG16944-PB|CG16944-PC|CG16944-PD|CG16944-PE|Stress-sensitive B|adenine nuclear translocase|adenine nucleotide translocase|drosophila ANT|lethal(1)9Ed|sesB-PA|sesB-PB|sesB-PC|sesB-PD|sesB-PE|stress sensitive B', map_location='1-32.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10780893, end_position_on_the_genomic_accession=10786958, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32008 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32008, CurrentID=0, Status='live', Symbol='Ant2', Aliases='Dmel_CG1683, ANT2, CG1683, Dmel\\CG1683, l(1)G0126, l(1)G0247, l(1)G0386', description='Adenine nucleotide translocase 2', other_designations='ADP/ATP translocase|ADP/ATP translocase 2|Ant2-PA|Ant2-PB|Ant2-PC|CG1683-PA|CG1683-PB|CG1683-PC|adenine nucleotide translocase 2', map_location='9E9-9F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10778954, end_position_on_the_genomic_accession=10786907, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32009 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32009, CurrentID=0, Status='live', Symbol='Imp', Aliases='Dmel_CG1691, CG1691, Dmel\\CG1691, IGF-II, IGF2BP1, IMP, IMP-1, MRE11, ORE-9, anon-EST:fe2B3, anon-WO0140519.39, cg1691, dIMP, imp, l(1)G0072, mre11', description='IGF-II mRNA-binding protein', other_designations='CG1691-PA|CG1691-PB|CG1691-PC|CG1691-PD|CG1691-PE|CG1691-PF|CG1691-PG|CG1691-PH|CG1691-PJ|CG1691-PK|CG1691-PL|CG1691-PM|IGF-II messenger RNA binding protein|Imp-PA|Imp-PB|Imp-PC|Imp-PD|Imp-PE|Imp-PF|Imp-PG|Imp-PH|Imp-PJ|Imp-PK|Imp-PL|Imp-PM|Insulin-like growth factor 2 mRNA binding protein 1|KH-domain protein|anon-fast-evolving-2B3|insulin-like growth factor II|mRNA-like ncRNA in embryogenesis 11', map_location='9F1-9F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10787869, end_position_on_the_genomic_accession=10824277, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  32011 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32011, CurrentID=0, Status='live', Symbol='CG15209', Aliases='Dmel_ BcDNA:SD22308, Dmel\\CG15209, anon-EST:Posey262', description='CG15209 gene product from transcript CG15209-RA', other_designations='CG15209-PA|uncharacterized protein', map_location='9F5-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10842205, end_position_on_the_genomic_accession=10842995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32013 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32013, CurrentID=0, Status='live', Symbol='CG17333', Aliases='Dmel_ Dmel\\CG17333, Pgls', description='CG17333 gene product from transcript CG17333-RA', other_designations='CG17333-PA|uncharacterized protein', map_location='9F5-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10847438, end_position_on_the_genomic_accession=10848534, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32014 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32014, CurrentID=0, Status='live', Symbol='CG2202', Aliases='Dmel_ Dmel\\CG2202', description='CG2202 gene product from transcript CG2202-RA', other_designations='CG2202-PA', map_location='9F5-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10848757, end_position_on_the_genomic_accession=10852327, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32015 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32015, CurrentID=0, Status='live', Symbol='feo', Aliases='Dmel_CG11207, CG11207, Dmel\\CG11207, EA86, Feo, PRC1, Q54, l(1)9Fe, l(1)BP1', description='fascetto', other_designations='CG11207-PA|feo-PA|lethal(1)9Fe', map_location='1-32.88 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10852446, end_position_on_the_genomic_accession=10854869, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32016 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32016, CurrentID=0, Status='live', Symbol='sofe', Aliases='Dmel_CG1655, CG1655, Dmel\\CG1655', description='sister of feo', other_designations='CG1655-PA|CG1655-PB|sofe-PA|sofe-PB', map_location='9F5-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10855000, end_position_on_the_genomic_accession=10857413, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32017 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32017, CurrentID=0, Status='live', Symbol='CG2186', Aliases='Dmel_ Dmel\\CG2186', description='CG2186 gene product from transcript CG2186-RB', other_designations='CG2186-PA|CG2186-PB', map_location='9F5-9F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10857655, end_position_on_the_genomic_accession=10866246, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32018 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32018, CurrentID=0, Status='live', Symbol='CG2157', Aliases='Dmel_ Dmel\\CG2157', description='CG2157 gene product from transcript CG2157-RB', other_designations='CG2157-PA|CG2157-PB', map_location='9F5-9F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10869284, end_position_on_the_genomic_accession=10870705, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32019 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32019, CurrentID=0, Status='live', Symbol='CG1637', Aliases='Dmel_ Dmel\\CG1637', description='CG1637 gene product from transcript CG1637-RC', other_designations='CG1637-PA|CG1637-PB|CG1637-PC|CG1637-PD|CG1637-PE', map_location='9F5-9F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10870594, end_position_on_the_genomic_accession=10877893, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32020 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32020, CurrentID=0, Status='live', Symbol='Ork1', Aliases='Dmel_CG1615, CG1615, Dmel\\CG1615, DorK, KCNK0, K[[2P]], ORK, ORK-1, ORK1, anon-WO0161006.1, anon-WO0161006.22, anon-WO0161006.23, d-ORK1, dORK, dORK1, dOrk, dOrk1, ork1', description='Open rectifier K[+] channel 1', other_designations='CG1615-PA|CG1615-PB|CG1615-PC|CG1615-PD|Ork1-PA|Ork1-PB|Ork1-PC|Ork1-PD|open rectifier K+ channel 1|open rectifier K[+] channel 1|open rectifier potassium channel', map_location='9F7-9F8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10878054, end_position_on_the_genomic_accession=10891250, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32021 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32021, CurrentID=0, Status='live', Symbol='CG1582', Aliases='Dmel_ Dmel\\CG1582, cg1582', description='CG1582 gene product from transcript CG1582-RB', other_designations='CG1582-PA|CG1582-PB', map_location='9F8-9F8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10894747, end_position_on_the_genomic_accession=10899302, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32022 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32022, CurrentID=0, Status='live', Symbol='CG15208', Aliases='Dmel_ Dmel\\CG15208, anon-WO0140519.64, lincRNA.977', description='CG15208 gene product from transcript CG15208-RB', other_designations='CG15208-PA|CG15208-PB|CG15208-PC|uncharacterized protein', map_location='9F8-9F8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10899965, end_position_on_the_genomic_accession=10901811, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32024 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32024, CurrentID=0, Status='live', Symbol='CG43901', Aliases='Dmel_ CG11203, CG15207, Dmel\\CG43901, Dmel_CG11203, Dmel_CG15207', description='CG43901 gene product from transcript CG43901-RA', other_designations='CG43901-PA|CG43901-PB|CG43901-PC|uncharacterized protein', map_location='9F10-9F11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10905767, end_position_on_the_genomic_accession=10923436, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32026 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32026, CurrentID=0, Status='live', Symbol='v', Aliases='Dmel_CG2155, 152559_at, CG2155, DmTDO, Dmel\\CG2155, NP_511113, TDO, Verer, vm, v', description='vermilion', other_designations='CG2155-PA|tryptophan 2,3-dioxygenase|v-PA|vermillion', map_location='1-33.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10923972, end_position_on_the_genomic_accession=10925631, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32027 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32027, CurrentID=0, Status='live', Symbol='CG2145', Aliases='Dmel_ Dmel\\CG2145', description='CG2145 gene product from transcript CG2145-RA', other_designations='CG2145-PA|CG2145-PB', map_location='9F11-9F12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10927399, end_position_on_the_genomic_accession=10929712, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32028 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32028, CurrentID=0, Status='live', Symbol='Myo10A', Aliases='Dmel_CG43657, 10A, AAF47980, CG16922, CG2136, CG2174, CG43657, DmXV, Dmel\\CG43657, Dmel_CG16922, Dmel_CG2174, Syph, myo10A', description='Myosin 10A', other_designations='CG43657-PC|CG43657-PD|CG43657-PE|CG43657-PF|Myo10A-PC|Myo10A-PD|Myo10A-PE|Myo10A-PF|myosin 10A|myosin XV|sisyphus|unconventional myosin class XV', map_location='9F12-9F13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10930840, end_position_on_the_genomic_accession=10964046, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32029 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32029, CurrentID=0, Status='live', Symbol='Rab9Fb', Aliases='Dmel_CG32670, CG2532, CG32670, DmRabX3, Dmel\\CG32670, RabX3', description='Rab at 9Fb', other_designations='CG32670-PA|Rab GTPase 9Fb|Rab9Fb-PA', map_location='9F13-9F13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10951609, end_position_on_the_genomic_accession=10952486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32032 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32032, CurrentID=0, Status='live', Symbol='C901', Aliases='Dmel_CG1567, 10, CG1567, CT4052, Dmel\\CG1567, anon-10Ab, c901, cDNA 901', description='CG1567 gene product from transcript CG1567-RA', other_designations='C901-PA|CG1567-PA', map_location='9F13-9F13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10964032, end_position_on_the_genomic_accession=10967134, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32033 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32033, CurrentID=0, Status='live', Symbol='CG42339', Aliases='Dmel_ CG12626, CG15204, Dmel\\CG42339, Dmel_CG12626, Dmel_CG15204', description='CG42339 gene product from transcript CG42339-RB', other_designations='CG42339-PA|CG42339-PB|uncharacterized protein', map_location='10A3-10A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=10967722, end_position_on_the_genomic_accession=11039901, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32034 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32034, CurrentID=0, Status='live', Symbol='CG11160', Aliases='Dmel_ Dmel\\CG11160', description='CG11160 gene product from transcript CG11160-RC', other_designations='CG11160-PA|CG11160-PB|CG11160-PC|uncharacterized protein', map_location='10A3-10A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11044756, end_position_on_the_genomic_accession=11049540, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  38107 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38107, CurrentID=0, Status='live', Symbol='CG13905', Aliases='Dmel_ Dmel\\CG13905', description='CG13905 gene product from transcript CG13905-RB', other_designations='CG13905-PB|uncharacterized protein', map_location='61D4-61D4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=902898, end_position_on_the_genomic_accession=903810, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32036 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32036, CurrentID=0, Status='live', Symbol='CG1545', Aliases='Dmel_ BcDNA:GH01560, Dmel\\CG1545, anon-EST:Posey256', description='CG1545 gene product from transcript CG1545-RB', other_designations='CG1545-PB', map_location='10A3-10A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11057545, end_position_on_the_genomic_accession=11059619, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32037 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32037, CurrentID=0, Status='live', Symbol='CG1537', Aliases='Dmel_ BcDNA:GH02938, Dmel\\CG1537', description='CG1537 gene product from transcript CG1537-RA', other_designations='CG1537-PA', map_location='10A3-10A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11062027, end_position_on_the_genomic_accession=11063170, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32038 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32038, CurrentID=0, Status='live', Symbol='CG15203', Aliases='Dmel_ 144845_at, DIGF-1, Dmel\\CG15203', description='CG15203 gene product from transcript CG15203-RB', other_designations='CG15203-PA|CG15203-PB|uncharacterized protein', map_location='10A4-10A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11066552, end_position_on_the_genomic_accession=11068379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32039 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32039, CurrentID=0, Status='live', Symbol='sev', Aliases='Dmel_CG18085, 7LES, 7less, CG18085, CT3980, DROSEV, DROSEV1, DmHD-265, Dmel\\CG18085, HD-265, SEV, Sev(AC)[[14]], sev', description='sevenless', other_designations='CG18085-PA|sev-PA', map_location='1-33.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11071441, end_position_on_the_genomic_accession=11086299, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  32040 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32040, CurrentID=0, Status='live', Symbol='Vago', Aliases='Dmel_CG2081, BcDNA:RH70969, CG2081, DIGF-2, Dmel\\CG2081', description='CG2081 gene product from transcript CG2081-RD', other_designations='CG2081-PB|CG2081-PD|Vago-PB|Vago-PD|vago', map_location='10A4-10A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11089290, end_position_on_the_genomic_accession=11090219, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32041 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32041, CurrentID=0, Status='live', Symbol='CG2076', Aliases='Dmel_ Dmel\\CG2076', description='CG2076 gene product from transcript CG2076-RB', other_designations='CG2076-PA|CG2076-PB', map_location='10A4-10A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11090210, end_position_on_the_genomic_accession=11092098, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32042 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32042, CurrentID=0, Status='live', Symbol='CG2061', Aliases='Dmel_ BcDNA:LD28247, Dmel\\CG2061', description='CG2061 gene product from transcript CG2061-RD', other_designations='CG2061-PA|CG2061-PB|CG2061-PD', map_location='10A4-10A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11092492, end_position_on_the_genomic_accession=11094947, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32043 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32043, CurrentID=0, Status='live', Symbol='CG42249', Aliases='Dmel_ CG11126, CG1961, Dmel\\CG42249, Dmel_CG11126, Dmel_CG1961', description='CG42249 gene product from transcript CG42249-RC', other_designations='CG42249-PB|CG42249-PC|CG42249-PD|uncharacterized protein', map_location='10A4-10A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11096273, end_position_on_the_genomic_accession=11108148, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  32045 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32045, CurrentID=0, Status='live', Symbol='Hsp60', Aliases='Dmel_CG12101, 12, BP5, CG12101, Cpn60, DmHsp60, Dmel\\CG12101, Dmhsp60, G62, HSP60, HSP60A, Hsp 60A, IEF16, Mmp-P1, SSP 7506, cpn60, d-hsp60, hsp60, hsp60A, l(1)10Ac, l(1)BP5, l(1)G8, l(1)HM21, l(1)L12, l(1)dp025, Hsp60', description='Heat shock protein 60', other_designations='60-kDa heat shock protein, mitochondrial|CG12101-PA|CG12101-PB|Hsp 60|Hsp60-PA|Hsp60-PB|heat shock protein 60|heat shock protein 60 kDa|lethal(1)10Ac|mitochondrial matrix protein P1', map_location='1-33.51 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11108482, end_position_on_the_genomic_accession=11112308, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32046 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32046, CurrentID=0, Status='live', Symbol='CG11122', Aliases='Dmel_ Dmel\\CG11122', description='CG11122 gene product from transcript CG11122-RA', other_designations='CG11122-PA|uncharacterized protein', map_location='10A4-10A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11112395, end_position_on_the_genomic_accession=11124726, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32047 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32047, CurrentID=0, Status='live', Symbol='Rpt3', Aliases='Dmel_CG16916, 16916, CG16916, Dm_Rpt3a, Dmel\\CG16916, Dmp48A, S6, S6B, TBP7, l(1)G0052, p48A', description='Regulatory particle triple-A ATPase 3', other_designations='CG16916-PA|CG16916-PB|Rpt3-PA|Rpt3-PB|regulatory particle triple-A ATPase 3', map_location='10A6-10A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11125101, end_position_on_the_genomic_accession=11126785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32049 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32049, CurrentID=0, Status='live', Symbol='Klp10A', Aliases='Dmel_CG1453, CG1453, DmCG1453, DmKLP10A, DmKlp10A, Dmel\\CG1453, KIF 2A, KIF2A/B/C, KLP10A, KLP10a, Kif2A, anon-WO03040301.142, klp10A', description='Kinesin-like protein at 10A', other_designations='CG1453-PA|CG1453-PB|CG1453-PC|CG1453-PD|CG1453-PE|CG1453-PF|Klp10A-PA|Klp10A-PB|Klp10A-PC|Klp10A-PD|Klp10A-PE|Klp10A-PF|kinesin-13|kinesin-13 motor|kinesin-like protein 10A', map_location='10A6-10A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11129362, end_position_on_the_genomic_accession=11137118, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32050 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32050, CurrentID=0, Status='live', Symbol='CDK2AP1', Aliases='Dmel_CG18292, CG18292, DOC1, Dmel\\CG18292, anon-WO03040301.144', description='CDK2-associated protein 1', other_designations='CDK2AP1-PA|CDK2AP1-PB|CG18292-PA|CG18292-PB|deleted in oral cancer 1', map_location='10A6-10A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11137859, end_position_on_the_genomic_accession=11139689, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32051 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32051, CurrentID=0, Status='live', Symbol='CG15199', Aliases='Dmel_ DIGF-3, Dmel\\CG15199', description='CG15199 gene product from transcript CG15199-RA', other_designations='CG15199-PA|CG15199-PB|uncharacterized protein', map_location='10A7-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11140474, end_position_on_the_genomic_accession=11141835, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32052 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32052, CurrentID=0, Status='live', Symbol='CG15202', Aliases='Dmel_ BcDNA:GH13342, DIGF-4, Dmel\\CG15202', description='CG15202 gene product from transcript CG15202-RB', other_designations='CG15202-PA|CG15202-PB|uncharacterized protein', map_location='10A8-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11142541, end_position_on_the_genomic_accession=11143755, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32053 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32053, CurrentID=0, Status='live', Symbol='CG15201', Aliases='Dmel_ BcDNA:RH06304, DIGF-5, Dmel\\CG15201', description='CG15201 gene product from transcript CG15201-RA', other_designations='CG15201-PA|uncharacterized protein', map_location='10A8-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11144120, end_position_on_the_genomic_accession=11144809, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32055 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32055, CurrentID=0, Status='live', Symbol='Lint-1', Aliases='Dmel_CG1908, CG1908, Dmel\\CG1908, dLint-1', description='l(3)mbt interacting protein 1', other_designations='CG1908-PA|CG1908-PC|CG1908-PD|CG1908-PE|Lint-1-PA|Lint-1-PC|Lint-1-PD|Lint-1-PE', map_location='10A8-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11149323, end_position_on_the_genomic_accession=11152592, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32056 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32056, CurrentID=0, Status='live', Symbol='rtv', Aliases='Dmel_CG1397, 1, CG1397, Dmel\\CG1397, L1, l(1)10Ag, l(1)G13, l(1)L1', description='retroactive', other_designations='CG1397-PA|CG1397-PB|CG1397-PC|CG1397-PD|rtv-PA|rtv-PB|rtv-PC|rtv-PD', map_location='10A8-10A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11152484, end_position_on_the_genomic_accession=11153990, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32057 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32057, CurrentID=0, Status='live', Symbol='Dlic', Aliases='Dmel_CG1938, CG1938, D-LIC, DLIC, DLIC22, Dmel\\CG1938, dlic, dlic2, l(1)G0065, l(1)G0190, Dlic', description='Dynein light intermediate chain', other_designations='CG1938-PA|CG1938-PB|CG1938-PD|Dlic-PA|Dlic-PB|Dlic-PD|Dlic2|DyneinLC|dandelion clock|dynein light interme- diate chain|dynein light intermediate chain|dynein light intermediate chain 2', map_location='10A8-10A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11154421, end_position_on_the_genomic_accession=11164670, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32058 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32058, CurrentID=0, Status='live', Symbol='antdh', Aliases='Dmel_CG1386, AntDH, Antdh, CG1386, Dmel\\CG1386, FOHSDR', description='CG1386 gene product from transcript CG1386-RA', other_designations='CG1386-PA|antdh-PA', map_location='10A10-10A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11168030, end_position_on_the_genomic_accession=11169058, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32059 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32059, CurrentID=0, Status='live', Symbol='CG1394', Aliases='Dmel_ BEST:GH24085, Dmel\\CG1394, anon-WO0140519.177, bs30g07.y1', description='CG1394 gene product from transcript CG1394-RB', other_designations='CG1394-PA|CG1394-PB', map_location='10A10-10A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11169811, end_position_on_the_genomic_accession=11170678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32060 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32060, CurrentID=0, Status='live', Symbol='CG15200', Aliases='Dmel_ Dmel\\CG15200', description='CG15200 gene product from transcript CG15200-RA', other_designations='CG15200-PA|uncharacterized protein', map_location='10A10-10A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11182931, end_position_on_the_genomic_accession=11183476, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32062 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32062, CurrentID=0, Status='live', Symbol='CG11106', Aliases='Dmel_ BcDNA:AT24116, Dmel\\CG11106', description='CG11106 gene product from transcript CG11106-RA', other_designations='CG11106-PA|CG11106-PB|uncharacterized protein', map_location='10B1-10B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11202248, end_position_on_the_genomic_accession=11202806, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32063 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32063, CurrentID=0, Status='live', Symbol='CG44422', Aliases='Dmel_ CG11105, CG12625, CG34144, CG42683, Dm_X:57642, Dmel\\CG44422, Dmel_CG11105, Dmel_CG12625, Dmel_CG34144, Dmel_CG42683', description='CG44422 gene product from transcript CG44422-RD', other_designations='CG44422-PA|CG44422-PB|CG44422-PC|CG44422-PD|CG44422-PE|CG44422-PF|uncharacterized protein', map_location='10B2-10B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11219379, end_position_on_the_genomic_accession=11302783, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32065 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32065, CurrentID=0, Status='live', Symbol='CG15198', Aliases='Dmel_ Dmel\\CG15198', description='CG15198 gene product from transcript CG15198-RA', other_designations='CG15198-PA|uncharacterized protein', map_location='10B1-10B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11248236, end_position_on_the_genomic_accession=11248885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32067 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32067, CurrentID=0, Status='live', Symbol='Ir10a', Aliases='Dmel_CG34143, CG12623, CG12624, CG34143, DmelIR10a, Dmel\\CG34143, IR10a', description='Ionotropic receptor 10a', other_designations='CG34143-PC|Ir10a-PC|ionotropic receptor 10a', map_location='10B2-10B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11308533, end_position_on_the_genomic_accession=11311208, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32068 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32068, CurrentID=0, Status='live', Symbol='sisA', Aliases='Dmel_CG1641, CG1641, DM29, Dmel\\CG1641, sis A, sis a, sis-A, sis-a, sisa', description='sisterless A', other_designations='CG1641-PA|sisA-PA', map_location='1-34.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11321579, end_position_on_the_genomic_accession=11322357, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32069 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32069, CurrentID=0, Status='live', Symbol='l(1)10Bb', Aliases='Dmel_CG1639, 10Ba, CG1639, Dmel\\CG1639, L(1)10Bb, l(1)G0169, l(1)G14, l(1)GLM14', description='lethal (1) 10Bb', other_designations='CG1639-PA|l(1)10Bb-PA|lethal(1)10Bb', map_location='1-34.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11324294, end_position_on_the_genomic_accession=11325308, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32070 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32070, CurrentID=0, Status='live', Symbol='CG1657', Aliases='Dmel_ Dmel\\CG1657', description='CG1657 gene product from transcript CG1657-RA', other_designations='CG1657-PA', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11325385, end_position_on_the_genomic_accession=11334071, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32071 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32071, CurrentID=0, Status='live', Symbol='Kap3', Aliases='Dmel_CG11759, CG11759, DmKAP, DmKap, Dmel\\CG11759, KAP, KAP3, Kap, dKAP3, kap3', description='Kinesin associated protein 3', other_designations='CG11759-PF|CG11759-PG|CG11759-PH|CG11759-PM|CG11759-PN|Kap3-PF|Kap3-PG|Kap3-PH|Kap3-PM|Kap3-PN|Kinesin-2|Kinesin-associated protein 3|kinesin II accessory protein|kinesin accessory protein|kinesin associated protein 3', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11334187, end_position_on_the_genomic_accession=11340084, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32072 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32072, CurrentID=0, Status='live', Symbol='CG34348', Aliases='Dmel_ CG11757, CG15197, Dmel\\CG34348, Dmel_CG11757', description='CG34348 gene product from transcript CG34348-RA', other_designations='CG34348-PA|uncharacterized protein', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11339873, end_position_on_the_genomic_accession=11344734, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32073 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32073, CurrentID=0, Status='live', Symbol='GCS1', Aliases='Dmel_CG1597, CG1597, Dmel\\CG1597', description='Glucosidase 1', other_designations='CG1597-PB|CG1597-PC|GCS1-PB|GCS1-PC|glucosidase 1', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11344506, end_position_on_the_genomic_accession=11348001, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32074 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32074, CurrentID=0, Status='live', Symbol='CG11756', Aliases='Dmel_ BcDNA:AT06648, Dmel\\CG11756', description='CG11756 gene product from transcript CG11756-RA', other_designations='CG11756-PA|uncharacterized protein', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11348120, end_position_on_the_genomic_accession=11348865, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32075 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32075, CurrentID=0, Status='live', Symbol='CG1738', Aliases='Dmel_ BcDNA:RE11532, Dmel\\CG1738', description='CG1738 gene product from transcript CG1738-RA', other_designations='CG1738-PA', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11349049, end_position_on_the_genomic_accession=11349847, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32076 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32076, CurrentID=0, Status='live', Symbol='CG11752', Aliases='Dmel_ BcDNA:RH44935, Dmel\\CG11752, anon-EST:Posey214', description='CG11752 gene product from transcript CG11752-RA', other_designations='CG11752-PA|uncharacterized protein', map_location='10B3-10B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11349998, end_position_on_the_genomic_accession=11350474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32077 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32077, CurrentID=0, Status='live', Symbol='CG1737', Aliases='Dmel_ Dmel\\CG1737', description='CG1737 gene product from transcript CG1737-RB', other_designations='CG1737-PA|CG1737-PB|CG1737-PC', map_location='10B3-10B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11350608, end_position_on_the_genomic_accession=11355539, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32078 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32078, CurrentID=0, Status='live', Symbol='dsh', Aliases='Dmel_CG18361, CG1720, CG18361, DSH, Dm Dsh, Dmel\\CG18361, Dsh, Dvl, l(1)10Bd, l(1)G0267, l(1)G17, l(1)dsh', description='dishevelled', other_designations='CG18361-PA|disheveled|disshevelled|dsh-PA', map_location='1-34.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11355881, end_position_on_the_genomic_accession=11358724, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32079, CurrentID=0, Status='live', Symbol='Pa1', Aliases='Dmel_CG11750, BcDNA:RE09126, BcDNA:RH21749, CG11750, Dmel\\CG11750, PA1, dPA1', description='PTIP associated 1', other_designations='CG11750-PA|Pa1-PA', map_location='10B5-10B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11359236, end_position_on_the_genomic_accession=11360929, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32080 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32080, CurrentID=0, Status='live', Symbol='hop', Aliases='Dmel_CG1594, 4, CG1594, Dm JAK, DmHD-160, Dmel\\CG1594, HD-160, HOP, Hop, Hop1, JAK, Jak, L4, Tum, Tum-1, d-jak/JAK, jak, l(1)10Be, l(1)G18, l(1)L4, l(1)hop, msvl, hop', description='hopscotch', other_designations='CG1594-PA|JAK kinase|Janus-family kinase|hop-PA|hopskotch|janus kinase|locus 18|tumorous', map_location='1-34.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11360930, end_position_on_the_genomic_accession=11368098, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32081 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32081, CurrentID=0, Status='live', Symbol='Tim8', Aliases='Dmel_CG1728, BcDNA:SD06593, CG1728, Dmel\\CG1728', description='CG1728 gene product from transcript CG1728-RB', other_designations='CG1728-PA|CG1728-PB|Tim8-PA|Tim8-PB', map_location='10B6-10B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11368445, end_position_on_the_genomic_accession=11369283, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32083 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32083, CurrentID=0, Status='live', Symbol='dlg1', Aliases='Dmel_CG1725, 11, CG1725, CG1730, CPD, DLG, DLG-A, DLG1, Discs-large, Dlg, Dlg-A, Dlg1, DlgA, DlgS97, Dmel\\CG1725, Drodlg, PSD95, SAP97, anon-EST:Posey93, anon-WO03040301.258, anon-WO03040301.260, anon-WO03040301.268, d. lg.-1, dlg, dlg-1, dlg-A, dlgA, dlgS97, l(1)10Bf, l(1)G0276, l(1)G0342, l(1)G0456, l(1)G19, l(1)L11, l(1)bwn, l(1)d.lg-1, l(1)d.lg.-1, l(1)discs large, l(1)dlg, l(1)dlg-1, l(1)dlg1, l(1)l.pr.-2, l(1)lpr-2, misb', description='discs large 1', other_designations='CG1725-PA|CG1725-PB|CG1725-PC|CG1725-PD|CG1725-PE|CG1725-PF|CG1725-PG|CG1725-PH|CG1725-PI|CG1725-PJ|CG1725-PK|CG1725-PL|CG1725-PM|CG1725-PN|CG1725-PO|CG1725-PP|CG1725-PQ|CG1725-PR|CG1725-PS|CG1725-PT|CG1725-PU|Disc-Large|Discs-large|disc large|disclarge|discs large|discslarge|disk large|dlg1-PA|dlg1-PB|dlg1-PC|dlg1-PD|dlg1-PE|dlg1-PF|dlg1-PG|dlg1-PH|dlg1-PI|dlg1-PJ|dlg1-PK|dlg1-PL|dlg1-PM|dlg1-PN|dlg1-PO|dlg1-PP|dlg1-PQ|dlg1-PR|dlg1-PS|dlg1-PT|dlg1-PU|lethal(1)benign wing imaginal disc neoplasm|lethal(1)discs large|lethal(1)discs-large|lethal(2)discs large', map_location='1-34.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11369665, end_position_on_the_genomic_accession=11409776, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  32084 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32084, CurrentID=0, Status='live', Symbol='CG15196', Aliases='Dmel_ Dmel\\CG15196', description='CG15196 gene product from transcript CG15196-RB', other_designations='CG15196-PB|CG15196-PC|uncharacterized protein', map_location='10B10-10B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11409947, end_position_on_the_genomic_accession=11411667, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32085 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32085, CurrentID=0, Status='live', Symbol='Gr10b', Aliases='Dmel_CG12622, CG12622, Dmel\\CG12622, GR10b, Gr10B2', description='Gustatory receptor 10b', other_designations='CG12622-PA|Gr10b-PA|gustatory receptor 10B2|gustatory receptor 10b', map_location='10B10-10B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11411783, end_position_on_the_genomic_accession=11413190, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32086 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32086, CurrentID=0, Status='live', Symbol='Or10a', Aliases='Dmel_CG17867, 10B.1, 10a, CG17867, DOR92, Dmel\\CG17867', description='Odorant receptor 10a', other_designations='CG17867-PB|Or10a-PB|odorant receptor 10a', map_location='10B10-10B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11413467, end_position_on_the_genomic_accession=11417422, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32087 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32087, CurrentID=0, Status='live', Symbol='Gs2', Aliases='Dmel_CG1743, CG1743, Dmel\\CG1743, GS2, GsII, gs-c, gs2', description='Glutamine synthetase 2', other_designations='CG1743-PA|CG1743-PB|CG1743-PC|CG1743-PD|CG1743-PE|CG1743-PF|Glutamate-ammonia ligase|Gs2-PA|Gs2-PB|Gs2-PC|Gs2-PD|Gs2-PE|Gs2-PF|cytoplasmic glutamine synthetase|glutamine synthase 2|glutamine synthetase 2', map_location='10B11-10B11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11422832, end_position_on_the_genomic_accession=11430197, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32088 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32088, CurrentID=0, Status='live', Symbol='Evi5', Aliases='Dmel_CG11727, BcDNA:GH08923, BcDNA:LD11130, CG11727, Dmel\\CG11727, MRE21', description='Ecotropic viral integration site 5 ortholog (H. sapiens)', other_designations='CG11727-PA|CG11727-PB|CG11727-PC|CG11727-PD|CG11727-PE|Evi5-PA|Evi5-PB|Evi5-PC|Evi5-PD|Evi5-PE|ecotropic viral integration site 5 ortholog|mRNA-like ncRNA in embryogenesis 21', map_location='10B11-10B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11430523, end_position_on_the_genomic_accession=11454386, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32089 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32089, CurrentID=0, Status='live', Symbol='Sk1', Aliases='Dmel_CG1747, CG1747, Dmel\\CG1747, SK1, Sphk1', description='Sphingosine kinase 1', other_designations='CG1747-PA|CG1747-PB|Sk1-PA|Sk1-PB|sphingosine kinase 1|sphingosine kinase I', map_location='10B13-10B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11442539, end_position_on_the_genomic_accession=11446012, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32090 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32090, CurrentID=0, Status='live', Symbol='HP5', Aliases='Dmel_CG1745, CG1745, Dmel\\CG1745', description='Heterochromatin protein 5', other_designations='CG1745-PA|CG1745-PB|CG1745-PC|HP5-PA|HP5-PB|HP5-PC|heterochromatin protein 5', map_location='10B14-10B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11454786, end_position_on_the_genomic_accession=11459227, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32092 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32092, CurrentID=0, Status='live', Symbol='CG43155', Aliases='Dmel_ CG15194, CG1756, Dmel\\CG43155, Dmel_CG15194, Dmel_CG1756, anon-WO0109301.51, anon-WO0109301.52, anon-WO0109301.9', description='CG43155 gene product from transcript CG43155-RC', other_designations='CG43155-PC|uncharacterized protein', map_location='10B15-10B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11459786, end_position_on_the_genomic_accession=11476504, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32093 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32093, CurrentID=0, Status='live', Symbol='Cyp4g15', Aliases='Dmel_CG11715, 152658_at, 4g15, CG11715, Dmel\\CG11715', description='CG11715 gene product from transcript CG11715-RC', other_designations='CG11715-PA|CG11715-PC|Cyp4g15-PA|Cyp4g15-PC|cytochrome P450', map_location='10B15-10B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11477058, end_position_on_the_genomic_accession=11483233, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32094 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32094, CurrentID=0, Status='live', Symbol='CG1749', Aliases='Dmel_ Dmel\\CG1749, dMoaD', description='CG1749 gene product from transcript CG1749-RA', other_designations='CG1749-PA|Uba5 homologue', map_location='10B15-10B15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11483518, end_position_on_the_genomic_accession=11485060, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32095 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32095, CurrentID=0, Status='live', Symbol='Spase25', Aliases='Dmel_CG1751, CG1751, Dmel\\CG1751', description='Spase 25-subunit', other_designations='CG1751-PA|CG1751-PB|CG1751-PC|Spase25-PA|Spase25-PB|Spase25-PC|spase 25-subunit', map_location='10B15-10B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11485250, end_position_on_the_genomic_accession=11486566, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32097 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32097, CurrentID=0, Status='live', Symbol='Drak', Aliases='Dmel_CG32666, CG11713, CG1760, CG32666, Dmel\\CG32666, EP(X)1452, EP1452, drak', description='Death-associated protein kinase related', other_designations='CG32666-PC|CG32666-PD|CG32666-PE|Drak-PC|Drak-PD|Drak-PE', map_location='10C1-10C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11493135, end_position_on_the_genomic_accession=11555045, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32098 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32098, CurrentID=0, Status='live', Symbol='CG1572', Aliases='Dmel_ Dmel\\CG1572', description='CG1572 gene product from transcript CG1572-RC', other_designations='CG1572-PA|CG1572-PB|CG1572-PC|CG1572-PD', map_location='10C5-10C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11555050, end_position_on_the_genomic_accession=11558654, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32099 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32099, CurrentID=0, Status='live', Symbol='PGRP-SA', Aliases='Dmel_CG11709, BcDNA:AT30827, CG11709, Dm PGRP-SA, Dmel\\CG11709, SA, SEMMELWEIS, seml', description='Peptidoglycan recognition protein SA', other_designations='CG11709-PA|CG11709-PB|PG recognition Protein-SA|PGRP-SA-PA|PGRP-SA-PB|Peptidoglycan-recognition protein-SA precursor|peptidoglycan recognition protein SA|peptidoglycan recognition protein-SA|protein semmelweis|semmelweis|semmelweiss', map_location='10C6-10C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11561643, end_position_on_the_genomic_accession=11562779, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32100 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32100, CurrentID=0, Status='live', Symbol='RpII215', Aliases='Dmel_CG1554, 5, 8WG16, CG1554, CTD, Dmel\\CG1554, H5, II, IIo, L5, POL, Pol II, Pol II CTD, Pol II Ser5P, Pol II Ser5p, Pol II0[ser2], Pol II0[ser5], Pol II[ser2], Pol IIa, Pol IIo, Pol IIo[Ser2], Pol IIo[ser2], Pol IIo[ser5], Pol-IIa, PolII, PolIIa, PolIIo, PolIIo[ser2], PolIIo[ser5], RNA Pol II, RNA Pol II CTD, RNA PolI 215, RNA PolII, RNA pol II, RNA pol IIo, RNA polII, RNA-PolII, RNAP, RNAP II, RNAP II LS, RNAPII, RNAPII0, RNApol, RNApol2, RNApolII, RPB1, RPII215, RpII, Rpb1, Rpll215, Ser5-P Pol II, Ubl, alphaPol IIo[ser2], dRPB1, dRpb1, l(1)10Ca, l(1)DC912, l(1)DF912, l(1)G0040, l(1)L5, pol II, polII, rpII1', description='RNA polymerase II 215kD subunit', other_designations='CG1554-PA|Pol II CTD|Pol II large subunit|Pol II large subunit CTD|RNA Pol II|RNA polymerase II|RNA polymerase II large subunit|RNA polymerase IIo|RNA polymeraseII-215 kd subunit|RNA-polymerase II|RpII215-PA|Ultrabithorax-like|polymerase II CTD', map_location='1-35.66 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11562800, end_position_on_the_genomic_accession=11570326, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32101 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32101, CurrentID=0, Status='live', Symbol='CG11699', Aliases='Dmel_ Dmel\\CG11699', description='CG11699 gene product from transcript CG11699-RB', other_designations='CG11699-PA|CG11699-PB|uncharacterized protein', map_location='10C6-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11570621, end_position_on_the_genomic_accession=11571718, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32102 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32102, CurrentID=0, Status='live', Symbol='Kmn1', Aliases='Dmel_CG1558, CG1558, CT4038, Dmel\\CG1558, Nsl1, anon-WO03040301.201, dmNsl1R, l(1)G0237, nsl1', description='kinetochore Mis12-Ndc80 network component 1', other_designations='CG1558-PA|CG1558-PB|Kmn1-PA|Kmn1-PB|Nsl1-related protein|l(1)G0237|lethal (1) G0237', map_location='10C6-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11571383, end_position_on_the_genomic_accession=11572442, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32103 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32103, CurrentID=0, Status='live', Symbol='CG11697', Aliases='Dmel_ Dmel\\CG11697, a, anon-WO03040301.203', description='CG11697 gene product from transcript CG11697-RA', other_designations='CG11697-PA|CG11697-PB|uncharacterized protein', map_location='10C7-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11572707, end_position_on_the_genomic_accession=11573640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32104 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32104, CurrentID=0, Status='live', Symbol='CG11696', Aliases='Dmel_ Dmel\\CG11696, b', description='CG11696 gene product from transcript CG11696-RA', other_designations='CG11696-PA|uncharacterized protein', map_location='10C7-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11573731, end_position_on_the_genomic_accession=11576152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32106 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32106, CurrentID=0, Status='live', Symbol='CG11695', Aliases='Dmel_ Dmel\\CG11695, c', description='CG11695 gene product from transcript CG11695-RB', other_designations='CG11695-PA|CG11695-PB|uncharacterized protein', map_location='10C7-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11576883, end_position_on_the_genomic_accession=11578806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64875 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64875, CurrentID=0, Status='live', Symbol='disco-r', Aliases='Dmel_CG32577, CG32577, CG9219, CG9223, Dmel\\CG32577', description='disco-related', other_designations='CG32577-PA|CG32577-PB|CG32577-PC|CG32577-PD|disco-r-PA|disco-r-PB|disco-r-PC|disco-r-PD|disconnected-related', map_location='14B1-14B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16118067, end_position_on_the_genomic_accession=16148977, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32108 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32108, CurrentID=0, Status='live', Symbol='CG1561', Aliases='Dmel_ Dmel\\CG1561, e', description='CG1561 gene product from transcript CG1561-RB', other_designations='CG1561-PB', map_location='10C7-10C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11581957, end_position_on_the_genomic_accession=11584175, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32109 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32109, CurrentID=0, Status='live', Symbol='rho-4', Aliases='Dmel_CG1697, CG1697, Dmel\\CG1697, Rho-4, Rho-related [10C6], rho 4, rho4', description='rhomboid-4', other_designations='CG1697-PA|CG1697-PB|Rhomboid-4|rho-4-PA|rho-4-PB|rhomboid 4', map_location='10C10-10C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11599838, end_position_on_the_genomic_accession=11606288, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32110 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32110, CurrentID=0, Status='live', Symbol='CG2533', Aliases='Dmel_ BcDNA:AT28768, Dmel\\CG2533', description='CG2533 gene product from transcript CG2533-RA', other_designations='CG2533-PA', map_location='10C10-10C10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11603243, end_position_on_the_genomic_accession=11604194, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64879 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=64879, CurrentID=0, Status='live', Symbol='Adgf-A2', Aliases='Dmel_CG32178, ADGF-A2, BcDNA:AT05468, CG32178, Dmel\\CG32178, MSI, Msi', description='Adenosine deaminase-related growth factor A2', other_designations='Adgf-A2-PA|Adgf-A2-PB|Adgf-A2-PC|CG32178-PA|CG32178-PB|CG32178-PC|Male-specific insect derived growth factor|adenosine deaminase-related growth factor A2|male-specific IDGF|male-specific-imaginal-disc-growth-factor', map_location='75A1-75A1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17754936, end_position_on_the_genomic_accession=17764158, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32112 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32112, CurrentID=0, Status='live', Symbol='CG1703', Aliases='Dmel_ Dmel\\CG1703', description='CG1703 gene product from transcript CG1703-RB', other_designations='CG1703-PA|CG1703-PB', map_location='10D1-10D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11611927, end_position_on_the_genomic_accession=11615128, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32113 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32113, CurrentID=0, Status='live', Symbol='RPA3', Aliases='Dmel_CG15220, BcDNA:RH55360, CG15220, Dmel\\CG15220, l(1)G0241', description='Replication protein A3', other_designations='CG15220-PA|CG15220-PB|RPA3-PA|RPA3-PB|lethal (1) G0241|replication protein A|replication protein A3', map_location='10D1-10D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11615097, end_position_on_the_genomic_accession=11615840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32114 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32114, CurrentID=0, Status='live', Symbol='Met', Aliases='Dmel_CG1705, CG1705, DmMet, Dmel\\CG1705, MET/Met1, Mett, Rst(1)JH, bHLHe59, met, Met', description='Methoprene-tolerant', other_designations='CG1705-PA|CG1705-PB|Met-PA|Met-PB|Methoprene-tolerant protein|juvenile hormone resistance|methoprene tolerant|methoprene-tolerant|resistance (1) juvenile H|resistance to juvenile hormone', map_location='1-35.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11616124, end_position_on_the_genomic_accession=11621280, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32115 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32115, CurrentID=0, Status='live', Symbol='Ptp10D', Aliases='Dmel_CG1817, 10D, CG1817, CT4920, DPTP, DPTP10D, DPTP[[10D]], Dmel\\CG1817, PTP10D, Ptp10, R-PTP 10D, dptp, ptp10D', description='Protein tyrosine phosphatase 10D', other_designations='CG1817-PB|CG1817-PE|CG1817-PF|CG1817-PG|CG1817-PH|CG1817-PI|Ptp10D-PB|Ptp10D-PE|Ptp10D-PF|Ptp10D-PG|Ptp10D-PH|Ptp10D-PI|protein tyrosine phosphatase 10B|protein tyrosine phosphatase 10D', map_location='10D1-10D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11622015, end_position_on_the_genomic_accession=11677338, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  32116 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32116, CurrentID=0, Status='live', Symbol='CG2371', Aliases='Dmel_ CG2442, CG2443, Dmel\\CG2371', description='CG2371 gene product from transcript CG2371-RC', other_designations='CG2371-PA|CG2371-PB|CG2371-PC', map_location='10D4-10D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11677344, end_position_on_the_genomic_accession=11678589, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32119 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32119, CurrentID=0, Status='live', Symbol='bif', Aliases='Dmel_CG1822, CG1822, Dmel\\CG1822', description='bifocal', other_designations='CG1822-PB|CG1822-PC|CG1822-PD|CG1822-PE|CG1822-PF|bif-PB|bif-PC|bif-PD|bif-PE|bif-PF', map_location='10D4-10D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11678970, end_position_on_the_genomic_accession=11694740, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32120 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32120, CurrentID=0, Status='live', Symbol='PhKgamma', Aliases='Dmel_CG1830, CG1830, DPhK-gamma, Dmel\\CG1830, EP779, ESTS:194B3T, Phkgamma, anon-sts6', description='Phosphorylase kinase gamma', other_designations='CG1830-PB|CG1830-PD|CG1830-PE|CG1830-PF|CG1830-PH|PhKgamma-PB|PhKgamma-PD|PhKgamma-PE|PhKgamma-PF|PhKgamma-PH|phosphorylase kinase gamma', map_location='10D5-10D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11695137, end_position_on_the_genomic_accession=11704394, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  32121 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32121, CurrentID=0, Status='live', Symbol='CG2444', Aliases='Dmel_ BcDNA:LP01642, DIGF-6, Dmel\\CG2444', description='CG2444 gene product from transcript CG2444-RB', other_designations='CG2444-PA|CG2444-PB', map_location='10D5-10D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11704310, end_position_on_the_genomic_accession=11705300, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32122 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32122, CurrentID=0, Status='live', Symbol='FucT6', Aliases='Dmel_CG2448, 6-FucT, BEST:CK00490, CG2448, CK00490, Dm alpha1, Dmel\\CG2448, Fuc-TVIII, dalpha6Fut', description='alpha1,6-fucosyltransferase', other_designations='CG2448-PA|FucT6-PA|core alpha1,6-fucosyltransferase', map_location='10D5-10D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11706575, end_position_on_the_genomic_accession=11709951, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32123 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32123, CurrentID=0, Status='live', Symbol='Amun', Aliases='Dmel_CG2446, CG2446, Dmel\\CG2446, EP1503, anon-WO03040301.104, anon-WO03040301.105, anon-WO03040301.107', description='CG2446 gene product from transcript CG2446-RB', other_designations='Amun-PA|Amun-PB|Amun-PC|Amun-PD|Amun-PE|Amun-PF|Amun-PG|Amun-PH|Amun-PI|Amun-PJ|Amun-PK|Amun-PL|CG2446-PA|CG2446-PB|CG2446-PC|CG2446-PD|CG2446-PE|CG2446-PF|CG2446-PG|CG2446-PH|CG2446-PI|CG2446-PJ|CG2446-PK|CG2446-PL|amun', map_location='10D6-10D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11710120, end_position_on_the_genomic_accession=11715565, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32124 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32124, CurrentID=0, Status='live', Symbol='prtp', Aliases='Dmel_CG1837, CG1837, Dmel\\CG1837, anon-EST:Posey253, p47', description='pretaporter', other_designations='CG1837-PA|CG1837-PB|CG1837-PC|CG1837-PD|pret-a-porter|prtp-PA|prtp-PB|prtp-PC|prtp-PD', map_location='10D7-10D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11716756, end_position_on_the_genomic_accession=11719894, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32125 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32125, CurrentID=0, Status='live', Symbol='Tango10', Aliases='Dmel_CG1841, CG1841, Dmel\\CG1841, TANGO10, tango10', description='Transport and Golgi organization 10', other_designations='CG1841-PA|CG1841-PB|CG1841-PC|Tango10-PA|Tango10-PB|Tango10-PC|transport and golgi organization|transport and golgi organization 10', map_location='10D7-10D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11720245, end_position_on_the_genomic_accession=11723854, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32126 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32126, CurrentID=0, Status='live', Symbol='inaF-D', Aliases='Dmel_CG42563, CG2457, CG42563, Dmel\\CG42563, Dmel_CG2457, INAF, InaF, anon-WO0172774.148, inaF', description='CG42563 gene product from transcript CG42563-RB', other_designations='CG42563-PB|inaF|inaF-D-PB', map_location='10D8-10E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11724668, end_position_on_the_genomic_accession=11729099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32127 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32127, CurrentID=0, Status='live', Symbol='CG15221', Aliases='Dmel_ Dmel\\CG15221', description='CG15221 gene product from transcript CG15221-RB', other_designations='CG15221-PB|uncharacterized protein', map_location='10D8-10E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11724246, end_position_on_the_genomic_accession=11736617, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32128 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32128, CurrentID=0, Status='live', Symbol='CG9360', Aliases='Dmel_ Dmel\\CG9360, FOHSDR', description='CG9360 gene product from transcript CG9360-RA', other_designations='CG9360-PA', map_location='10E2-10E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11763474, end_position_on_the_genomic_accession=11764590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32130 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32130, CurrentID=0, Status='live', Symbol='dy', Aliases='Dmel_CG9355, CG9355, Dmel\\CG9355, ala1, dy', description='dusky', other_designations='CG9355-PA|alae parvae|dy-PA', map_location='1-36.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11769769, end_position_on_the_genomic_accession=11778784, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32131 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32131, CurrentID=0, Status='live', Symbol='Karl', Aliases='Dmel_CG4139, CG4139, Dmel\\CG4139, karl', description='CG4139 gene product from transcript CG4139-RC', other_designations='CG4139-PA|CG4139-PB|CG4139-PC|CG4139-PD|Karl-PA|Karl-PB|Karl-PC|Karl-PD|karl', map_location='10E3-10E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11787381, end_position_on_the_genomic_accession=11792253, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32132 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32132, CurrentID=0, Status='live', Symbol='CkIIbeta', Aliases='Dmel_CG15224, And, CCK2, CG15224, CK II, CK-2, CK-II beta, CK-IIbeta, CK2, CK2-beta, CK2beta, CKII, CKII beta, CKII-beta1, CKIIb, CKIIbeta, Cask-II-b, Ck-IIbeta, Ck2beta, CkII, CkII-beta1, Ckiibeta, DmCK2beta, DmCKIIbeta, DmCkIIbeta, Dmel\\CG15224, Ds cas kin, KC2B_DROME, betaCK2, ck2beta, dCK2beta, dCKII, dCKII beta, mbu, CkIIbeta', description='Casein kinase II beta subunit', other_designations='CG15224-PB|CG15224-PC|CG15224-PD|CG15224-PE|CG15224-PF|CG15224-PG|CG15224-PH|CG15224-PI|CG15224-PJ|CG15224-PK|CK2 (beta subunit)|CkIIbeta-PB|CkIIbeta-PC|CkIIbeta-PD|CkIIbeta-PE|CkIIbeta-PF|CkIIbeta-PG|CkIIbeta-PH|CkIIbeta-PI|CkIIbeta-PJ|CkIIbeta-PK|andante|casein kinase 2|casein kinase 2 regulatory subunit|casein kinase II|casein kinase II beta|casein kinase II beta subunit|mushroom bodies undersized', map_location='10E3-10E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11792533, end_position_on_the_genomic_accession=11801587, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32133 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32133, CurrentID=0, Status='live', Symbol='Hsc70-3', Aliases='Dmel_CG4147, A4V4C4_DROME, BiP, Bip, CG1579, CG4147, Dmel\\CG4147, GRP78, Grp78, HSC-70, HSC3, HSC70, HSC70-3, HSP7C_DROME, Hsc-3, Hsc-70-3, Hsc3, Hsc3p, Hsc70, Hsp-c3, anon-WO0138581.7, bip, dBiP, dhsc70, dmHsc72, hsc3, hsc70, hsc70-3, hsc72, l(1)G0102, l(1)G0111, l(1)G0292, l(1)G0341, l(1)G0407, l(1)G0466', description='Heat shock 70-kDa protein cognate 3', other_designations='CG4147-PA|CG4147-PB|CG4147-PC|CG4147-PD|CG4147-PE|Hsc70-3-PA|Hsc70-3-PB|Hsc70-3-PC|Hsc70-3-PD|Hsc70-3-PE|heat shock 70-kDa protein cognate 3|heat shock cognate protein 3|heat shock protein cognate 3|heat shock protein cognate 70-3|heat shock protein cognate 72|heat-shock cognate 70-3|immunoglobulin binding protein', map_location='10E3-10E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11801696, end_position_on_the_genomic_accession=11807117, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32135 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32135, CurrentID=0, Status='live', Symbol='CG1578', Aliases='Dmel_ Dmel\\CG1578', description='CG1578 gene product from transcript CG1578-RB', other_designations='CG1578-PA|CG1578-PB|CG1578-PC', map_location='10E5-10E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11808506, end_position_on_the_genomic_accession=11816551, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32136 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32136, CurrentID=0, Status='live', Symbol='rudhira', Aliases='Dmel_CG43154, CG1905, CG32662, CG32663, CG43154, DRudh, Dmel\\CG43154, Dmel_CG32662, Dmel_CG32663, Drudh, Rudh, anon-EST:Liang-2.41, clone 2.41', description='CG43154 gene product from transcript CG43154-RD', other_designations='CG43154-PB|CG43154-PC|CG43154-PD|CG43154-PE|CG43154-PF|rudhira-PB|rudhira-PC|rudhira-PD|rudhira-PE|rudhira-PF', map_location='10E6-10F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11816898, end_position_on_the_genomic_accession=11837460, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32137 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32137, CurrentID=0, Status='live', Symbol='CG15741', Aliases='Dmel_ Dmel\\CG15741', description='CG15741 gene product from transcript CG15741-RA', other_designations='CG15741-PA|uncharacterized protein', map_location='10F1-10F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11837543, end_position_on_the_genomic_accession=11838026, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32138 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32138, CurrentID=0, Status='live', Symbol='Cyp28c1', Aliases='Dmel_CG1895, 28c1, CG1895, Dmel\\CG1895', description='CG1895 gene product from transcript CG1895-RB', other_designations='CG1895-PA|CG1895-PB|Cyp28c1-PA|Cyp28c1-PB|cytochrome P450', map_location='10F1-10F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11839608, end_position_on_the_genomic_accession=11841503, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32139 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32139, CurrentID=0, Status='live', Symbol='CG15740', Aliases='Dmel_ Dmel\\CG15740, i175', description='CG15740 gene product from transcript CG15740-RB', other_designations='CG15740-PB|uncharacterized protein', map_location='10F1-10F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11842662, end_position_on_the_genomic_accession=11846099, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32140 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32140, CurrentID=0, Status='live', Symbol='p24-1', Aliases='Dmel_CG1967, CG1967, Dmel\\CG1967, anon-EST:Liang-1.45, clone 1.45, p24, p24.1, p26/p24gamma4', description='p24-related-1', other_designations='CG1967-PA|CG1967-PB|p24-1-PA|p24-1-PB', map_location='10F1-10F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11850909, end_position_on_the_genomic_accession=11852530, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32141 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32141, CurrentID=0, Status='live', Symbol='CG10347', Aliases='Dmel_ Dmel\\CG10347, HSP20, Q9VYT5', description='CG10347 gene product from transcript CG10347-RB', other_designations='CG10347-PA|CG10347-PB|uncharacterized protein', map_location='10F1-10F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11852692, end_position_on_the_genomic_accession=11854931, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32142 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32142, CurrentID=0, Status='live', Symbol='ATP7', Aliases='Dmel_CG1886, CG1886, Dm Dmel\\CG1886, dmATP7, lincRNA.S9122', description='CG1886 gene product from transcript CG1886-RC', other_designations='ATP7-PB|ATP7-PC|CG1886-PB|CG1886-PC', map_location='10F1-10F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11855022, end_position_on_the_genomic_accession=11864754, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32143 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32143, CurrentID=0, Status='live', Symbol='CG2025', Aliases='Dmel_ Dmel\\CG2025', description='CG2025 gene product from transcript CG2025-RA', other_designations='CG2025-PA', map_location='10F2-10F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11864651, end_position_on_the_genomic_accession=11868893, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32144 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32144, CurrentID=0, Status='live', Symbol='CG1847', Aliases='Dmel_ Dmel\\CG1847', description='CG1847 gene product from transcript CG1847-RA', other_designations='CG1847-PA|CG1847-PC', map_location='10F2-10F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11869170, end_position_on_the_genomic_accession=11871168, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32145 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32145, CurrentID=0, Status='live', Symbol='CG11802', Aliases='Dmel_ Dmel\\CG11802', description='CG11802 gene product from transcript CG11802-RA', other_designations='CG11802-PA|uncharacterized protein', map_location='10F2-10F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11871299, end_position_on_the_genomic_accession=11872779, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32146 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32146, CurrentID=0, Status='live', Symbol='CG15739', Aliases='Dmel_ 17945012, Dmel\\CG15739, FPPase, Phos15739', description='CG15739 gene product from transcript CG15739-RA', other_designations='CG15739-PA|CG15739-PB|uncharacterized protein', map_location='10F2-10F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11873319, end_position_on_the_genomic_accession=11875346, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32147 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32147, CurrentID=0, Status='live', Symbol='CG10352', Aliases='Dmel_ Dmel\\CG10352, Phos10352', description='CG10352 gene product from transcript CG10352-RB', other_designations='CG10352-PB|uncharacterized protein', map_location='10F2-10F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11875306, end_position_on_the_genomic_accession=11877168, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32148 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32148, CurrentID=0, Status='live', Symbol='CG10353', Aliases='Dmel_ Dmel\\CG10353', description='CG10353 gene product from transcript CG10353-RF', other_designations='CG10353-PC|CG10353-PD|CG10353-PE|CG10353-PF|CG10353-PI|CG10353-PK|CG10353-PL|uncharacterized protein', map_location='10F2-10F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11877737, end_position_on_the_genomic_accession=11887162, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  32149 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32149, CurrentID=0, Status='live', Symbol='SelG', Aliases='Dmel_CG1844, BcDNA:GH03581, CG1844, Dmel\\CG1844, G-rich, SelK, dSelK, dselG, dselK', description='Selenoprotein G', other_designations='CG1844-PA|G-rich protein|SelG-PA|Seleno-protein G|selenoprotein G', map_location='10F4-10F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11887284, end_position_on_the_genomic_accession=11888240, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32150 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32150, CurrentID=0, Status='live', Symbol='CG1840', Aliases='Dmel_ AAF48112, BcDNA:LD12750, Dmel\\CG1840', description='CG1840 gene product from transcript CG1840-RA', other_designations='CG1840-PA|CG1840-PB', map_location='10F4-10F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11888352, end_position_on_the_genomic_accession=11889186, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32151 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32151, CurrentID=0, Status='live', Symbol='sicily', Aliases='Dmel_CG15738, CG15738, Dmel\\CG15738, XE07', description='severe impairment of CI with lengthened youth', other_designations='CG15738-PA|CG15738-PB|sicily-PA|sicily-PB', map_location='10F4-10F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11891732, end_position_on_the_genomic_accession=11893745, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32152 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32152, CurrentID=0, Status='live', Symbol='wisp', Aliases='Dmel_CG15737, CG15737, DmCG15737, Dmel\\CG15737, TAMP, Wisp, fs(1)M19', description='wispy', other_designations='CG15737-PA|CG15737-PB|CG15737-PC|WISPY|female sterile(1)M19|temprano|toll pathway activation mediating protein|wisp-PA|wisp-PB|wisp-PC', map_location='1-37 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11893943, end_position_on_the_genomic_accession=11899920, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32153 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32153, CurrentID=0, Status='live', Symbol='Upf1', Aliases='Dmel_CG1559, CG1559, Dm-Upf1, DmUpf1, Dmel\\CG1559, UPF1p, smg-2, upf1', description='CG1559 gene product from transcript CG1559-RA', other_designations='CG1559-PA|CG1559-PB|Upf1-PA|Upf1-PB|up-frameshift suppressor 1', map_location='10F4-10F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11899996, end_position_on_the_genomic_accession=11905371, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32154 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32154, CurrentID=0, Status='live', Symbol='pot', Aliases='Dmel_CG2467, CG2467, Dmel\\CG2467', description='papillote', other_designations='CG2467-PA|CG2467-PB|CG2467-PC|pot-PA|pot-PB|pot-PC', map_location='10F4-10F11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11907409, end_position_on_the_genomic_accession=11917260, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32156 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32156, CurrentID=0, Status='live', Symbol='Sclp', Aliases='Dmel_CG2471, CG2471, CT8171, Dmel\\CG2471, dSCLP', description='CG2471 gene product from transcript CG2471-RC', other_designations='CG2471-PA|CG2471-PB|CG2471-PC|Sclp-PA|Sclp-PB|Sclp-PC|sclp', map_location='10F7-10F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11920004, end_position_on_the_genomic_accession=11924662, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32157 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32157, CurrentID=0, Status='live', Symbol='CG10362', Aliases='Dmel_ Dmel\\CG10362', description='CG10362 gene product from transcript CG10362-RA', other_designations='CG10362-PA|uncharacterized protein', map_location='10F7-10F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11926424, end_position_on_the_genomic_accession=11930850, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32158 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32158, CurrentID=0, Status='live', Symbol='cac', Aliases='Dmel_CG43368, 13, CAC, CG1522, CG15928, CG43368, Ca-alpha1A, Ca[[V]]2.1, Cac, Dmca1A, Dmcalpha1A, Dmel\\CG43368, Dmel_CG1522, Dmel_CG15928, EE171, H18, H47, L13, P18, alpha1A1, l(1)11Aa, l(1)13, l(1)L13, ldg, nbA, nonB, opm18, opm47, cac', description='cacophony', other_designations='CG43368-PA|CG43368-PB|CG43368-PC|CG43368-PD|CG43368-PE|CG43368-PF|CG43368-PG|CG43368-PH|CG43368-PI|CG43368-PJ|CG43368-PL|CG43368-PM|CG43368-PN|CG43368-PO|CG43368-PP|cac-PA|cac-PB|cac-PC|cac-PD|cac-PE|cac-PF|cac-PG|cac-PH|cac-PI|cac-PJ|cac-PL|cac-PM|cac-PN|cac-PO|cac-PP|cacaphony|cacophany|calcium channel alpha1 subunit, type A|lamina-degeneration|lethal(1)11Aa|night blind A|nightblind|nightblind A|no-on-transient B', map_location='1-36.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11931280, end_position_on_the_genomic_accession=11985087, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  32159 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32159, CurrentID=0, Status='live', Symbol='gd', Aliases='Dmel_CG1505, CG1505, Dmel\\CG1505, GD, Gd, SP186, fs(1)18, fs(1)190-5, fs(1)573, fs(1)A573, fs(1)M18, fs(1)gd, gas', description='gastrulation-defective', other_designations='CG1505-PC|CG1505-PD|CG1505-PE|CG1505-PF|gastrulation defective|gd-PC|gd-PD|gd-PE|gd-PF', map_location='1-36.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11985154, end_position_on_the_genomic_accession=11987764, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32160 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32160, CurrentID=0, Status='live', Symbol='tsg', Aliases='Dmel_CG1502, CG1502, Dmel\\CG1502, RF32, Tsg, dTsg, l(1)11Ab, l(1)11Ac, l(1)RF32, l(1)tsg', description='twisted gastrulation', other_designations='CG1502-PA|lethal(1)11Ac|tsg-PA', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11988062, end_position_on_the_genomic_accession=11988948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32161 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32161, CurrentID=0, Status='live', Symbol='CG18130', Aliases='Dmel_ Dmel\\CG18130', description='CG18130 gene product from transcript CG18130-RA', other_designations='CG18130-PA|uncharacterized protein', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11990118, end_position_on_the_genomic_accession=11992768, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32162 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32162, CurrentID=0, Status='live', Symbol='fw', Aliases='Dmel_CG1500, CG1500, CT3745, Dmel\\CG1500-like, wr, fw', description='furrowed', other_designations='CG1500-PA|CG1500-PB|CG1500-PC|fw-PA|fw-PB|fw-PC|wrinkled', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=11992915, end_position_on_the_genomic_accession=12003914, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  32163 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32163, CurrentID=0, Status='live', Symbol='CG1806', Aliases='Dmel_ Dmel\\CG1806', description='CG1806 gene product from transcript CG1806-RA', other_designations='CG1806-PA|CG1806-PB', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12007176, end_position_on_the_genomic_accession=12009904, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32164 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32164, CurrentID=0, Status='live', Symbol='CG1492', Aliases='Dmel_ Dmel\\CG1492', description='CG1492 gene product from transcript CG1492-RB', other_designations='CG1492-PA|CG1492-PB', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12009711, end_position_on_the_genomic_accession=12012211, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32165 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32165, CurrentID=0, Status='live', Symbol='regucalcin', Aliases='Dmel_CG1803, CG1803, Dmel\\CG1803, Regucalcin, T1', description='CG1803 gene product from transcript CG1803-RA', other_designations='CG1803-PA|CG1803-PB|CG1803-PC|CG1803-PD|regucalcin-PA|regucalcin-PB|regucalcin-PC|regucalcin-PD', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12012377, end_position_on_the_genomic_accession=12017375, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32166 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32166, CurrentID=0, Status='live', Symbol='Chrac-16', Aliases='Dmel_CG15736, 15736, 20K protein, BcDNA:RE59557, CG15736, CHRAC, CHRAC-16, CHRAC16, Chrac16, Dmel\\CG15736, joey', description='Chromatin accessibility complex 16kD protein', other_designations='CG15736-PA|CG15736-PB|Chrac-16-PA|Chrac-16-PB|chromatin accessibility complex 16kD protein', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12017225, end_position_on_the_genomic_accession=12017961, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32167 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32167, CurrentID=0, Status='live', Symbol='CG15735', Aliases='Dmel_ BcDNA:LD05893, Dmel\\CG15735', description='CG15735 gene product from transcript CG15735-RA', other_designations='CG15735-PA|uncharacterized protein', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12018640, end_position_on_the_genomic_accession=12020719, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32168 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32168, CurrentID=0, Status='live', Symbol='Tango4', Aliases='Dmel_CG1796, CG1796, Dmel\\CG1796, TANGO4', description='Transport and Golgi organization 4', other_designations='CG1796-PA|Tango4-PA|transport and golgi organization 4', map_location='11A1-11A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12021068, end_position_on_the_genomic_accession=12022746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32169 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32169, CurrentID=0, Status='live', Symbol='Usp7', Aliases='Dmel_CG1490, CG1490, Dmel\\CG1490, Q9VYQ8, USP7, Usp-7', description='Ubiquitin-specific protease 7', other_designations='CG1490-PB|CG1490-PC|Usp7-PB|Usp7-PC|ubiquitin-specific protease 7', map_location='11A1-11A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12022837, end_position_on_the_genomic_accession=12029032, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32170 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32170, CurrentID=0, Status='live', Symbol='Cyp311a1', Aliases='Dmel_CG1488, 311a1, CG1488, Dmel\\CG1488', description='CG1488 gene product from transcript CG1488-RB', other_designations='CG1488-PB|Cyp311a1-PB|cytochrome P450', map_location='11A2-11A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12029754, end_position_on_the_genomic_accession=12031465, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32172 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32172, CurrentID=0, Status='live', Symbol='Cyp318a1', Aliases='Dmel_CG1786, 318a1, CG11364, CG1786, Cyp318A1, Dmel\\CG1786', description='CG1786 gene product from transcript CG1786-RA', other_designations='CG1786-PA|Cyp318a1-PA|cytochrome P450', map_location='11A2-11A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12032632, end_position_on_the_genomic_accession=12035283, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32173 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32173, CurrentID=0, Status='live', Symbol='Uch-L5R', Aliases='Dmel_CG1950, CG1950, Dmel\\CG1950, Uch37R', description='Uch-L5-related', other_designations='CG1950-PA|Uch-L5R-PA', map_location='11A2-11A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12055338, end_position_on_the_genomic_accession=12056635, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32174 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32174, CurrentID=0, Status='live', Symbol='Muc11A', Aliases='Dmel_CG32656, CG2779, CG32656, Dmel\\CG32656', description='Mucin 11A', other_designations='CG32656-PB|Muc11A-PB|mucin 11A', map_location='11A2-11A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12059795, end_position_on_the_genomic_accession=12065144, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32176 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32176, CurrentID=0, Status='live', Symbol='CG12721', Aliases='Dmel_ BcDNA:RE60404, Dmel\\CG12721', description='CG12721 gene product from transcript CG12721-RA', other_designations='CG12721-PA|uncharacterized protein', map_location='11A3-11A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12072008, end_position_on_the_genomic_accession=12072818, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32177 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32177, CurrentID=0, Status='live', Symbol='CG15734', Aliases='Dmel_ Dmel\\CG15734', description='CG15734 gene product from transcript CG15734-RD', other_designations='CG15734-PC|CG15734-PD|uncharacterized protein', map_location='11A3-11A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12086469, end_position_on_the_genomic_accession=12087025, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32178 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32178, CurrentID=0, Status='live', Symbol='CG11356', Aliases='Dmel_ Arl13b, Dmel\\CG11356', description='CG11356 gene product from transcript CG11356-RA', other_designations='CG11356-PA|uncharacterized protein', map_location='11A4-11A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12102944, end_position_on_the_genomic_accession=12103840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32179 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32179, CurrentID=0, Status='live', Symbol='CG2750', Aliases='Dmel_ Dmel\\CG2750, anon-WO0140519.124', description='CG2750 gene product from transcript CG2750-RB', other_designations='CG2750-PB|CG2750-PC', map_location='11A4-11A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12107200, end_position_on_the_genomic_accession=12113357, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32180 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32180, CurrentID=0, Status='live', Symbol='CG1924', Aliases='Dmel_ Dmel\\CG1924', description='CG1924 gene product from transcript CG1924-RA', other_designations='CG1924-PA', map_location='11A6-11A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12176440, end_position_on_the_genomic_accession=12178626, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32183 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32183, CurrentID=0, Status='live', Symbol='Ten-a', Aliases='Dmel_CG42338, 1.2, CG11270, CG12720, CG15733, CG18182, CG2578, CG2590, CG32659, CG42338, Dmel\\CG42338, Dmel_CG12720, Dmel_CG15733, Dmel_CG32659, DmtenA, Dmtena, Ten11A, Ten[a], cbd, ten-a, tenA, tenA(AT)[[14]], ten[a]', description='Tenascin accessory', other_designations='CG42338-PD|CG42338-PE|CG42338-PF|CG42338-PG|CG42338-PI|CG42338-PK|CG42338-PL|CG42338-PM|CG42338-PN|CG42338-PO|CG42338-PP|Ten-a-PD|Ten-a-PE|Ten-a-PF|Ten-a-PG|Ten-a-PI|Ten-a-PK|Ten-a-PL|Ten-a-PM|Ten-a-PN|Ten-a-PO|Ten-a-PP|central body defect|tenascin accessory|tenascin-like molecule accessory|teneurin', map_location='1-41 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12043698, end_position_on_the_genomic_accession=12335214, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  32188 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32188, CurrentID=0, Status='live', Symbol='CG2577', Aliases='Dmel_ Dmel\\CG2577', description='CG2577 gene product from transcript CG2577-RA', other_designations='CG2577-PA|CKIa-like', map_location='11A9-11A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12373118, end_position_on_the_genomic_accession=12374448, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32189 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32189, CurrentID=0, Status='live', Symbol='Ir11a', Aliases='Dmel_CG15732, CG15732, CT35969, DmelIR11a, Dmel\\CG15732, IR11a, Q9VYN4_DROME', description='Ionotropic receptor 11a', other_designations='CG15732-PB|Ir11a-PB|ionotropic receptor 11a', map_location='11A9-11A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12375596, end_position_on_the_genomic_accession=12377755, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32190 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32190, CurrentID=0, Status='live', Symbol='CG2574', Aliases='Dmel_ Dmel\\CG2574, bs21b01.y1', description='CG2574 gene product from transcript CG2574-RA', other_designations='CG2574-PA', map_location='11A9-11A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12408861, end_position_on_the_genomic_accession=12409825, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32191 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32191, CurrentID=0, Status='live', Symbol='Pkcdelta', Aliases='Dmel_CG42349, CG10524, CG11245, CG42349, Dmel\\CG42349, Dmel_CG10524, Dmel_CG11245, PKCd, PKCdelta', description='Protein kinase C delta', other_designations='CG42349-PC|CG42349-PD|CG42349-PE|Pkcdelta-PC|Pkcdelta-PD|Pkcdelta-PE|protein kinase C|protein kinase C delta', map_location='11A10-11A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12432224, end_position_on_the_genomic_accession=12453650, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  10178864 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=10178864, CurrentID=0, Status='live', Symbol='CG42711', Aliases='Dmel_ Dmel\\CG42711', description='CG42711 gene product from transcript CG42711-RB', other_designations='CG42711-PB|uncharacterized protein', map_location='34E4-34E4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=13930322, end_position_on_the_genomic_accession=13931362, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32193 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32193, CurrentID=0, Status='live', Symbol='CG15927', Aliases='Dmel_ Dmel\\CG15927', description='CG15927 gene product from transcript CG15927-RA', other_designations='CG15927-PA|CG15927-PB|uncharacterized protein', map_location='11A11-11A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12454729, end_position_on_the_genomic_accession=12455683, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32194 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32194, CurrentID=0, Status='live', Symbol='CG15731', Aliases='Dmel_ Dmel\\CG15731', description='CG15731 gene product from transcript CG15731-RA', other_designations='CG15731-PA|uncharacterized protein', map_location='11A11-11A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12456409, end_position_on_the_genomic_accession=12457763, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32195 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32195, CurrentID=0, Status='live', Symbol='Rab40', Aliases='Dmel_CG1900, 24641533, CG1900, Dm Dmel\\CG1900, Q8IR80, rab40', description='CG1900 gene product from transcript CG1900-RA', other_designations='CG1900-PA|CG1900-PB|CG1900-PC|Rab40-PA|Rab40-PB|Rab40-PC', map_location='11A11-11A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12459796, end_position_on_the_genomic_accession=12463112, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32196 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32196, CurrentID=0, Status='live', Symbol='CG42258', Aliases='Dmel_ BcDNA:RE29468, CG15926, CG17788, Dmel\\CG42258, Dmel_CG15926, Dmel_CG17788, Pripl4', description='CG42258 gene product from transcript CG42258-RC', other_designations='CG42258-PC|CG42258-PD|CG42258-PE|CG42258-PF|uncharacterized protein', map_location='11A11-11A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12463528, end_position_on_the_genomic_accession=12478431, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32198 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32198, CurrentID=0, Status='live', Symbol='Cpr11A', Aliases='Dmel_CG2560, CG2560, Dmel\\CG2560, L12', description='Cuticular protein 11A', other_designations='CG2560-PA|Cpr11A-PA|cuticular protein 11A', map_location='11A12-11A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12486970, end_position_on_the_genomic_accession=12489448, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32199 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32199, CurrentID=0, Status='live', Symbol='Lsp1alpha', Aliases='Dmel_CG2559, CG2559, DmeLSP1a, Dmel\\CG2559, LSP 1, LSP-1, LSP-1 alpha, LSP-1alpha, LSP1, LSP1-alpha, LSP1alpha, Lsp-1alpha, Lsp1, Lsp1-alpha', description='Larval serum protein 1 alpha', other_designations='CG2559-PA|Lsp1alpha-PA|larval serum protein|larval serum protein 1 alpha|larval serum protein one alpha', map_location='11A12-11A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12490291, end_position_on_the_genomic_accession=12492982, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32200 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32200, CurrentID=0, Status='live', Symbol='Mks1', Aliases='Dmel_CG15730, CG15730, Dmel\\CG15730, dMks1', description='Meckel syndrome, type 1 homologue', other_designations='CG15730-PA|Mks1-PA|meckel syndrome, type 1 homologue', map_location='11A12-11A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12494395, end_position_on_the_genomic_accession=12496597, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32201 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32201, CurrentID=0, Status='live', Symbol='CG2556', Aliases='Dmel_ Dmel\\CG2556', description='CG2556 gene product from transcript CG2556-RA', other_designations='CG2556-PA', map_location='11A12-11A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12497188, end_position_on_the_genomic_accession=12516565, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32202 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32202, CurrentID=0, Status='live', Symbol='CG43921', Aliases='Dmel_ BcDNA:GH02424, CG11146, CG15729, CG32653, CG42251, Dmel\\CG43921, Dmel_CG11146, Dmel_CG15729, Dmel_CG32653, Dmel_CG42251, lincRNA.981', description='CG43921 gene product from transcript CG43921-RA', other_designations='CG43921-PA|CG43921-PB|CG43921-PC|CG43921-PD|uncharacterized protein', map_location='11B1-11B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12516634, end_position_on_the_genomic_accession=12568452, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  32203 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32203, CurrentID=0, Status='live', Symbol='Cpr11B', Aliases='Dmel_CG2555, CG2555, DmelCpr11B, Dmel\\CG2555', description='Cuticular protein 11B', other_designations='CG2555-PA|CG2555-PB|Cpr11B-PA|Cpr11B-PB|cuticular protein 11B', map_location='11B1-11B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12553124, end_position_on_the_genomic_accession=12554235, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32205 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32205, CurrentID=0, Status='live', Symbol='CG11138', Aliases='Dmel_ BEST:RE18684, BcDNA:RE48052, CG1855, Dmel\\CG11138, EP(X)1313, RE18684', description='CG11138 gene product from transcript CG11138-RE', other_designations='CG11138-PC|CG11138-PD|CG11138-PE|uncharacterized protein', map_location='11B1-11B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12569169, end_position_on_the_genomic_accession=12583193, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32207 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32207, CurrentID=0, Status='live', Symbol='LIMK1', Aliases='Dmel_CG1848, CG1848, D-LIMK1, D-Limk, DLIMK, DLIMK1, Dlimk, Dmel\\CG1848, LIM-kinase 1, LIMK, LIMK-1, LimK, LimK1, Limk, dLIMK, dLIMK1, limk', description='LIM-kinase1', other_designations='CG1848-PA|CG1848-PC|CG1848-PD|CG1848-PE|CG1848-PF|LIM Kinase1|LIM domain protein kinase|LIM-DOMAIN KINASE|LIM-kinase|LIM-kinase 1|LIMK1-PA|LIMK1-PC|LIMK1-PD|LIMK1-PE|LIMK1-PF|Lim kinase|im kinase', map_location='11B2-11B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12584385, end_position_on_the_genomic_accession=12592193, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32208 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32208, CurrentID=0, Status='live', Symbol='CG1824', Aliases='Dmel_ Dmel\\CG1824', description='CG1824 gene product from transcript CG1824-RA', other_designations='CG1824-PA', map_location='11B2-11B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12592555, end_position_on_the_genomic_accession=12595624, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32209 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32209, CurrentID=0, Status='live', Symbol='Sec16', Aliases='Dmel_CG32654, BcDNA:RE70141, CG18079, CG2551, CG32654, Dmel\\CG32654, dSec16, dsec16, sec16, sec26', description='Sec16 ortholog (S. cerevisiae)', other_designations='CG32654-PD|CG32654-PE|CG32654-PF|CG32654-PH|CG32654-PI|CG32654-PJ|Sec16 ortholog|Sec16-PD|Sec16-PE|Sec16-PF|Sec16-PH|Sec16-PI|Sec16-PJ', map_location='11B2-11B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12596047, end_position_on_the_genomic_accession=12609910, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  32211 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32211, CurrentID=0, Status='live', Symbol='CG1463', Aliases='Dmel_ Dmel\\CG1463', description='CG1463 gene product from transcript CG1463-RA', other_designations='CG1463-PA|CG1463-PB|CG1463-PC', map_location='11B3-11B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12609977, end_position_on_the_genomic_accession=12611667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32212 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32212, CurrentID=0, Status='live', Symbol='CG2543', Aliases='Dmel_ Dmel\\CG2543', description='CG2543 gene product from transcript CG2543-RA', other_designations='CG2543-PA|CG2543-PB', map_location='11B3-11B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12611867, end_position_on_the_genomic_accession=12615530, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32213 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32213, CurrentID=0, Status='live', Symbol='CG11085', Aliases='Dmel_ Dmel\\CG11085', description='CG11085 gene product from transcript CG11085-RA', other_designations='CG11085-PA|uncharacterized protein', map_location='11B4-11B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12616316, end_position_on_the_genomic_accession=12621216, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32214 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32214, CurrentID=0, Status='live', Symbol='CG15728', Aliases='Dmel_ Dmel\\CG15728', description='CG15728 gene product from transcript CG15728-RB', other_designations='CG15728-PA|CG15728-PB|uncharacterized protein', map_location='11B5-11B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12624292, end_position_on_the_genomic_accession=12628445, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32215 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32215, CurrentID=0, Status='live', Symbol='Aven', Aliases='Dmel_CG15727, CG15727, Dmel\\CG15727, dAven', description='Aven ortholog', other_designations='Aven-PA|Aven-PB|CG15727-PA|CG15727-PB|aven ortholog', map_location='11B5-11B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12628370, end_position_on_the_genomic_accession=12629640, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32216 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32216, CurrentID=0, Status='live', Symbol='CG2540', Aliases='Dmel_ Dmel\\CG2540', description='CG2540 gene product from transcript CG2540-RA', other_designations='CG2540-PA', map_location='11B5-11B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12629878, end_position_on_the_genomic_accession=12631111, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32217 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32217, CurrentID=0, Status='live', Symbol='Tomosyn', Aliases='Dmel_CG17762, BcDNA:RH19475, CG11056, CG17762, CG2205, Dmel\\CG17762, tomosyn', description='Tomosyn ortholog', other_designations='CG17762-PC|CG17762-PF|CG17762-PH|CG17762-PI|CG17762-PJ|CG17762-PK|CG17762-PL|Tomosyn-PC|Tomosyn-PF|Tomosyn-PH|Tomosyn-PI|Tomosyn-PJ|Tomosyn-PK|Tomosyn-PL|tomosyn|tomosyn ortholog', map_location='11B6-11B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12631605, end_position_on_the_genomic_accession=12653442, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  26067079 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=26067079, CurrentID=0, Status='live', Symbol='solo', Aliases='Dmel_CG46282, BG:DS00929.15, CG33678, CG43081, CG46282, Dmel\\CG46282, cgt, fs(2)ltoRJ36, pf7', description='sisters on the loose', other_designations='CG46282-PA|CG46282-PB|courgette|female sterile(2)ltoRJ36|no-relish|solo-PA|solo-PB', map_location='35B10-35C1', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=15061656, end_position_on_the_genomic_accession=15072133, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32221 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32221, CurrentID=0, Status='live', Symbol='CkIalpha', Aliases='Dmel_CG2028, CG2028, CK I, CK1, CK1a, CK1alpha, CKI, CKI alpha, CKIa, CKIalpha, Ck1alpha, CkI alpha, CkIa, Dmel\\CG2028, PKA-C, anon-WO03040301.93, anon-WO03040301.95, ck1a, ck1alpha, dmCK1, dmckI, l(1)G0492', description='Casein kinase Ialpha', other_designations='CG2028-PA|CG2028-PB|CG2028-PD|CG2028-PE|CG2028-PF|CG2028-PG|CkIalpha-PA|CkIalpha-PB|CkIalpha-PD|CkIalpha-PE|CkIalpha-PF|CkIalpha-PG|caesin kinase I alpha|casein kinase 1|casein kinase 1a|casein kinase 1alpha|casein kinase I|casein kinase ialpha|casein kinase-1', map_location='11B7-11B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12653626, end_position_on_the_genomic_accession=12657834, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32222 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32222, CurrentID=0, Status='live', Symbol='Tis11', Aliases='Dmel_CG4070, CG4070, DTIS11, Dmel\\CG4070, TIS11, TIS11cc1, TIScc1, Ts11, anon-EST:Liang-1.28, clone 1.28, dTTP, dTis11, l(1)G0124, tis-11', description='Tis11 homolog', other_designations='CG4070-PA|CG4070-PC|CG4070-PD|Tis11-PA|Tis11-PC|Tis11-PD', map_location='11B9-11B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12659415, end_position_on_the_genomic_accession=12682554, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  261599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261599, CurrentID=0, Status='live', Symbol='mthl12', Aliases='Dmel_CG32853, CG32853, Dmel\\CG32853, Mthl12', description='methuselah-like 12', other_designations='CG32853-PA|Mth-like 12|Mth-like-12|mthl12-PA', map_location='87E7-87E7', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13353673, end_position_on_the_genomic_accession=13355413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32224 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32224, CurrentID=0, Status='live', Symbol='CG15725', Aliases='Dmel_ Dmel\\CG15725', description='CG15725 gene product from transcript CG15725-RA', other_designations='CG15725-PA|uncharacterized protein', map_location='11B12-11B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12705854, end_position_on_the_genomic_accession=12708841, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32225 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32225, CurrentID=0, Status='live', Symbol='Smr', Aliases='Dmel_CG4013, CG12719, CG4013, Dmel\\CG4013, E52, SMR, SMRTER, anon-WO0153538.73, l(1)G0060, l(1)G0361, lincRNA.983, smr', description='Smrter', other_designations='CG4013-PD|CG4013-PE|CG4013-PF|CG4013-PG|SANT domain protein|SMRT-related ecdysone receptor interacting factor|SMRT-related ecdysone receptor-interacting factor|SMRTER|Smr-PD|Smr-PE|Smr-PF|Smr-PG|lethal (1) G0060|silencing mediator of retinoid and thyroid hormone receptors-related ecdysone receptor-interacting factor|smarter|smrter', map_location='11B10-11B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12684203, end_position_on_the_genomic_accession=12742618, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  32226 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32226, CurrentID=0, Status='live', Symbol='CG4004', Aliases='Dmel_ CG18137, Dmel\\CG4004', description='CG4004 gene product from transcript CG4004-RE', other_designations='CG4004-PA|CG4004-PB|CG4004-PC|CG4004-PD|CG4004-PE', map_location='11B14-11B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12744212, end_position_on_the_genomic_accession=12749785, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  261603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261603, CurrentID=0, Status='live', Symbol='CG32445', Aliases='Dmel_ CG12562, Dmel\\CG32445, anon-WO0140519.131', description='CG32445 gene product from transcript CG32445-RA', other_designations='CG32445-PA|uncharacterized protein', map_location='78E2-78E2', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=21639359, end_position_on_the_genomic_accession=21640710, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32228 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32228, CurrentID=0, Status='live', Symbol='ade5', Aliases='Dmel_CG3989, AIRc-SAICARs, Ade5, CG3989, Dmel\\CG3989', description='CG3989 gene product from transcript CG3989-RB', other_designations='AIR-carboxylase-SAICAR synthetase|CG3989-PA|CG3989-PB|ade5-PA|ade5-PB', map_location='11B16-11B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12757036, end_position_on_the_genomic_accession=12760617, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32229 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32229, CurrentID=0, Status='live', Symbol='CG12717', Aliases='Dmel_ Dmel\\CG12717', description='CG12717 gene product from transcript CG12717-RB', other_designations='CG12717-PA|CG12717-PB|uncharacterized protein', map_location='11B16-11B16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12760642, end_position_on_the_genomic_accession=12763611, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32230 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32230, CurrentID=0, Status='live', Symbol='CG3812', Aliases='Dmel_ AGPAT1, AGPAT2, Dmel\\CG3812', description='CG3812 gene product from transcript CG3812-RB', other_designations='CG3812-PA|CG3812-PB|CG3812-PC', map_location='11B16-11C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12765266, end_position_on_the_genomic_accession=12772851, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  261607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261607, CurrentID=0, Status='live', Symbol='CG32174', Aliases='Dmel_ CG3877, Dmel\\CG32174', description='CG32174 gene product from transcript CG32174-RD', other_designations='CG32174-PB|CG32174-PC|CG32174-PD|uncharacterized protein', map_location='74B5-74C1', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=17428936, end_position_on_the_genomic_accession=17433621, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32233 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32233, CurrentID=0, Status='live', Symbol='Pde9', Aliases='Dmel_CG42276, CG1627, CG32648, CG3754, CG3761, CG3765, CG42276, Dmel\\CG42276, Dmel_CG32648, Dmel_CG3754, PDE9', description='Phosphodiesterase 9', other_designations='CG42276-PB|CG42276-PD|CG42276-PE|CG42276-PF|Pde9-PB|Pde9-PD|Pde9-PE|Pde9-PF|phosphodiesterase 9', map_location='11C4-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12785563, end_position_on_the_genomic_accession=12897327, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  261610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261610, CurrentID=0, Status='live', Symbol='pgant4', Aliases='Dmel_CG31956, CG31956, CG8845, CG8845b, Dmel\\CG31956', description='polypeptide GalNAc transferase 4', other_designations='CG31956-PB|CG31956-PC|pgant4-PB|pgant4-PC', map_location='23F3-23F3', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=3472603, end_position_on_the_genomic_accession=3475338, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  261613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261613, CurrentID=0, Status='live', Symbol='CG31908', Aliases='Dmel_ CG8761, Dmel\\CG31908', description='CG31908 gene product from transcript CG31908-RC', other_designations='CG31908-PA|CG31908-PB|CG31908-PC|uncharacterized protein', map_location='27D4-27D4', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=7024829, end_position_on_the_genomic_accession=7027389, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32238 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32238, CurrentID=0, Status='live', Symbol='CG4661', Aliases='Dmel_ Dmel\\CG4661', description='CG4661 gene product from transcript CG4661-RA', other_designations='CG4661-PA', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12855688, end_position_on_the_genomic_accession=12857722, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32239 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32239, CurrentID=0, Status='live', Symbol='CG32649', Aliases='Dmel_ CG4410, Dmel\\CG32649', description='CG32649 gene product from transcript CG32649-RA', other_designations='CG32649-PA|uncharacterized protein', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12898768, end_position_on_the_genomic_accession=12901114, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32240 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32240, CurrentID=0, Status='live', Symbol='CG4407', Aliases='Dmel_ Dmel\\CG4407, Fad1', description='CG4407 gene product from transcript CG4407-RA', other_designations='CG4407-PA|CG4407-PB', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12901353, end_position_on_the_genomic_accession=12903516, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32241 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32241, CurrentID=0, Status='live', Symbol='CG4404', Aliases='Dmel_ Dmel\\CG4404', description='CG4404 gene product from transcript CG4404-RA', other_designations='CG4404-PA', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12903745, end_position_on_the_genomic_accession=12906018, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32242 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32242, CurrentID=0, Status='live', Symbol='mRpL49', Aliases='Dmel_CG4647, BcDNA:RE35556, CG4647, Dmel\\CG4647, MRP-L49, mRpl49', description='mitochondrial ribosomal protein L49', other_designations='CG4647-PA|CG4647-PB|mRpL49-PA|mRpL49-PB', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12907105, end_position_on_the_genomic_accession=12907942, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32243 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32243, CurrentID=0, Status='live', Symbol='Brms1', Aliases='Dmel_CG4400, BRMS1L, CG4400, Dmel\\CG4400, dBrms1', description='BRMS1 ortholog', other_designations='Brms1-PA|CG4400-PA', map_location='11D1-11D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12907986, end_position_on_the_genomic_accession=12909622, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32244 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32244, CurrentID=0, Status='live', Symbol='CG4645', Aliases='Dmel_ Dmel\\CG4645', description='CG4645 gene product from transcript CG4645-RA', other_designations='CG4645-PA|CG4645-PB|CG4645-PC', map_location='11D1-11D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12909953, end_position_on_the_genomic_accession=12912070, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32245 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32245, CurrentID=0, Status='live', Symbol='fne', Aliases='Dmel_CG4396, BEST:HL02010, BcDNA:HL02010, BcDNA:RE25290, CG4396, CR32646, Dmel\\CG4396, Dmel_CR32646, FNE, Fne, MRE28, cg4396, lincRNA.985', description='found in neurons', other_designations='CG4396-PA|CG4396-PB|CG4396-PC|CG4396-PD|CG4396-PE|CG4396-PG|CG4396-PH|CG4396-PI|CG4396-PJ|CR32646|fne-PA|fne-PB|fne-PC|fne-PD|fne-PE|fne-PG|fne-PH|fne-PI|fne-PJ', map_location='11D4-11D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12917477, end_position_on_the_genomic_accession=12950417, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  32246 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32246, CurrentID=0, Status='live', Symbol='hec', Aliases='Dmel_CG4395, CG4395, Dmel\\CG4395, anon-WO0170980.1, anon-WO0170980.2', description='hector', other_designations='CG4395-PA|CG4395-PB|hec-PA|hec-PB', map_location='11D4-11D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12950436, end_position_on_the_genomic_accession=12957462, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  261623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261623, CurrentID=0, Status='live', Symbol='CG31683', Aliases='Dmel_ BcDNA:GH02384, Dmel\\CG31683', description='CG31683 gene product from transcript CG31683-RA', other_designations='CG31683-PA|CG31683-PB|uncharacterized protein', map_location='38C7-38C7', chromosome='2L', genomic_nucleotide_accession_version='NT_033779.5', start_position_on_the_genomic_accession=20447734, end_position_on_the_genomic_accession=20449964, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32248 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32248, CurrentID=0, Status='live', Symbol='CG12716', Aliases='Dmel_ Dmel\\CG12716', description='CG12716 gene product from transcript CG12716-RA', other_designations='CG12716-PA|uncharacterized protein', map_location='11D5-11D5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=12974683, end_position_on_the_genomic_accession=12976231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32249 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32249, CurrentID=0, Status='live', Symbol='CG15719', Aliases='Dmel_ Dmel\\CG15719', description='CG15719 gene product from transcript CG15719-RD', other_designations='CG15719-PC|CG15719-PD|uncharacterized protein', map_location='11D6-11D6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13003955, end_position_on_the_genomic_accession=13004879, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32250 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32250, CurrentID=0, Status='live', Symbol='CG2209', Aliases='Dmel_ Dmel\\CG2209, anon-WO0140519.110', description='CG2209 gene product from transcript CG2209-RA', other_designations='CG2209-PA', map_location='11D7-11D8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13022198, end_position_on_the_genomic_accession=13022949, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  261628 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261628, CurrentID=0, Status='live', Symbol='CG31469', Aliases='Dmel_ CG9599, Dmel\\CG31469', description='CG31469 gene product from transcript CG31469-RA', other_designations='CG31469-PA|uncharacterized protein', map_location='87F12-87F12', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=13709159, end_position_on_the_genomic_accession=13709939, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  261629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261629, CurrentID=0, Status='live', Symbol='Unc-115a', Aliases='Dmel_CG31352, CG31352, CG9489, Dmel\\CG31352, Dunc-115, anon-WO0140519.21, dunc-115, unc-115', description='CG31352 gene product from transcript CG31352-RE', other_designations='CG31352-PA|CG31352-PB|CG31352-PC|CG31352-PE|CG31352-PF|CG31352-PG|CG31352-PH|CG31352-PJ|Unc-115a-PA|Unc-115a-PB|Unc-115a-PC|Unc-115a-PE|Unc-115a-PF|Unc-115a-PG|Unc-115a-PH|Unc-115a-PJ', map_location='85E4-85E5', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=9694386, end_position_on_the_genomic_accession=9705417, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  261631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261631, CurrentID=0, Status='live', Symbol='Rpb7', Aliases='Dmel_CG31155, CG31155, CG6572, Dmel\\CG31155, RPB7, d', description='CG31155 gene product from transcript CG31155-RA', other_designations='CG31155-PA|Rpb7-PA', map_location='88E3-88E3', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=15222251, end_position_on_the_genomic_accession=15223180, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32256 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32256, CurrentID=0, Status='live', Symbol='hep', Aliases='Dmel_CG4353, CG2190, CG4353, D-MKK7, DHEP/MKK7, DJNKK, DMKK7, Dmel\\CG4353, HEP, HEP/MKK7, Hep, Hep/Mkk7, JNKK, MAP2K7, MAPKK, MKK7, Mkk7, dJNKK, dMKK7, hem/JNKK, hp, hrp, l(1)7P1, l(1)G0107, l(1)G0208, hep', description='hemipterous', other_designations='CG4353-PA|CG4353-PC|CG4353-PD|CG4353-PE|CG4353-PF|JNK kinase|JNK-Kinase|Mitogen-activated protein kinase kinase 7|hempiterous|hep-PA|hep-PC|hep-PD|hep-PE|hep-PF', map_location='11D10-11D10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13079205, end_position_on_the_genomic_accession=13090434, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32257 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32257, CurrentID=0, Status='live', Symbol='lic', Aliases='Dmel_CG12244, CG12244, D-MKK3, DMEK3, DMKK3, DMKK3/lic, DmMKK3, Dmel\\CG12244, MEK3, MEK3/MKK3, MKK3, Mek3, Mkk3, Mpk3, dMKK3, l(1)G0252, p38, p38MAPKK', description='licorne', other_designations='CG12244-PA|MAP kinase kinase 3|MEK-kinase 3|Mkk3|drosophila MAP kinase kinase 3|lic-PA', map_location='11D10-11D10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13091145, end_position_on_the_genomic_accession=13093986, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32258 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32258, CurrentID=0, Status='live', Symbol='CG2200', Aliases='Dmel_ Dmel\\CG2200', description='CG2200 gene product from transcript CG2200-RB', other_designations='CG2200-PA|CG2200-PB', map_location='11D10-11D10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13094299, end_position_on_the_genomic_accession=13095732, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32259 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32259, CurrentID=0, Status='live', Symbol='CG43313', Aliases='Dmel_ CG12714, CG4351, Dmel\\CG43313, Dmel_CG12714, Dmel_CG4351, dCHPF', description='CG43313 gene product from transcript CG43313-RB', other_designations='CG43313-PA|CG43313-PB|uncharacterized protein', map_location='11D11-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13095571, end_position_on_the_genomic_accession=13112462, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32260 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32260, CurrentID=0, Status='live', Symbol='Fer3HCH', Aliases='Dmel_CG4349, CG4349, Dmel\\CG4349', description='Ferritin 3 heavy chain homologue', other_designations='CG4349-PA|Fer3HCH-PA|ferritin 3 heavy chain homologue|mitoch. ferritin', map_location='11D11-11D11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13107382, end_position_on_the_genomic_accession=13108194, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32261 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32261, CurrentID=0, Status='live', Symbol='CG42237', Aliases='Dmel_ C133, CG4346, Dmel\\CG42237, Dmel_CG4346, PLA2', description='CG42237 gene product from transcript CG42237-RB', other_designations='CG42237-PA|CG42237-PB|phospholipase A2|uncharacterized protein', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13115131, end_position_on_the_genomic_accession=13123066, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32262 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32262, CurrentID=0, Status='live', Symbol='CG15717', Aliases='Dmel_ Dmel\\CG15717', description='CG15717 gene product from transcript CG15717-RE', other_designations='CG15717-PA|CG15717-PC|CG15717-PE|uncharacterized protein', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13128952, end_position_on_the_genomic_accession=13134565, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32263 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32263, CurrentID=0, Status='live', Symbol='MFS10', Aliases='Dmel_CG4330, AA202479, CG4330, Dmel\\CG4330', description='Major Facilitator Superfamily Transporter 10', other_designations='CG4330-PA|CG4330-PB|CG4330-PC|MFS10-PA|MFS10-PB|MFS10-PC|major facilitator superfamily transporter 10', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13129018, end_position_on_the_genomic_accession=13134839, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32264 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32264, CurrentID=0, Status='live', Symbol='CG32645', Aliases='Dmel_ CG15716, CG4331, Dmel\\CG32645', description='CG32645 gene product from transcript CG32645-RB', other_designations='CG32645-PB|uncharacterized protein', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13139199, end_position_on_the_genomic_accession=13153462, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32266 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32266, CurrentID=0, Status='live', Symbol='CG4318', Aliases='Dmel_ Dmel\\CG4318', description='CG4318 gene product from transcript CG4318-RB', other_designations='CG4318-PB', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13153461, end_position_on_the_genomic_accession=13154924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32267 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32267, CurrentID=0, Status='live', Symbol='CG4332', Aliases='Dmel_ Dmel\\CG4332', description='CG4332 gene product from transcript CG4332-RA', other_designations='CG4332-PA', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13155038, end_position_on_the_genomic_accession=13157372, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32268 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32268, CurrentID=0, Status='live', Symbol='Bap60', Aliases='Dmel_CG4303, BAF60, BAP60, BAPp60, CG4303, Dmel\\CG4303, bap60, i127, i133', description='Brahma associated protein 60kD', other_designations='BRM-associated protein 60|Bap60-PA|CG4303-PA|brahma associated protein 60 kD|brahma associated protein 60 kDa|brahma associated protein 60kD', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13157445, end_position_on_the_genomic_accession=13159508, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32269 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32269, CurrentID=0, Status='live', Symbol='CG12096', Aliases='Dmel_ 12096, Dmel\\CG12096, S5b, S5b/PSMD5, dS5b', description='CG12096 gene product from transcript CG12096-RA', other_designations='CG12096-PA|CG12096-PB|CG12096-PC|uncharacterized protein', map_location='11E1-11E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13159604, end_position_on_the_genomic_accession=13162854, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32271 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32271, CurrentID=0, Status='live', Symbol='CG12723', Aliases='Dmel_ Dmel\\CG12723', description='CG12723 gene product from transcript CG12723-RA', other_designations='CG12723-PA|uncharacterized protein', map_location='11E3-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13189643, end_position_on_the_genomic_accession=13191883, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32272 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32272, CurrentID=0, Status='live', Symbol='CG2453', Aliases='Dmel_ Dmel\\CG2453', description='CG2453 gene product from transcript CG2453-RC', other_designations='CG2453-PA|CG2453-PB|CG2453-PC', map_location='11E3-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13193094, end_position_on_the_genomic_accession=13194551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32273 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32273, CurrentID=0, Status='live', Symbol='sno', Aliases='Dmel_CG44436, CG1903, CG44436, Dmel\\CG44436, Dmel_CG1903, Sno, g-l, l(1)11Ea', description='strawberry notch', other_designations='CG44436-PA|CG44436-PB|glossy-like|sno-PA|sno-PB', map_location='1-41.8 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13195499, end_position_on_the_genomic_accession=13211674, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  32274 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32274, CurrentID=0, Status='live', Symbol='REG', Aliases='Dmel_CG1591, CG1591, DmREGgamma, Dmel\\CG1591gamma, dREG, dREGgamma, dReggamma', description='CG1591 gene product from transcript CG1591-RB', other_designations='CG1591-PA|CG1591-PB|REG-PA|REG-PB', map_location='11E3-11E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13212325, end_position_on_the_genomic_accession=13214408, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32275 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32275, CurrentID=0, Status='live', Symbol='mew', Aliases='Dmel_CG1771, CG1771, CT5254, Dmel\\CG1771, ItgaPS1, PS1, PS1alpha, PSalpha1, aPS1, alpha-PS1, alpha1Int, alphaPS1, alpha[[PS1]], l(1)G0429, l(1)G0443', description='multiple edematous wings', other_designations='CG1771-PA|CG1771-PB|CG1771-PC|CG1771-PD|CG1771-PE|CG1771-PF|PS1 integrin|Position-specific-1 antigen|Position-specific-antigen-1alpha|alpha integrin|alpha-PS1-integrin|alpha1 integrin|alphaPS integrin|alphaPS1|alphaPS1 integrin|alpha[[PS1]] subunit of integrin|integrin alphaPS1|integrin alphaPS1 subunit|mew-PA|mew-PB|mew-PC|mew-PD|mew-PE|mew-PF|multiple edematous wing', map_location='11E3-11E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13214409, end_position_on_the_genomic_accession=13258490, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32276 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32276, CurrentID=0, Status='live', Symbol='CG15742', Aliases='Dmel_ Dmel\\CG15742', description='CG15742 gene product from transcript CG15742-RC', other_designations='CG15742-PA|CG15742-PB|CG15742-PC|uncharacterized protein', map_location='11E8-11E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13257392, end_position_on_the_genomic_accession=13258495, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32278 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32278, CurrentID=0, Status='live', Symbol='HDAC4', Aliases='Dmel_CG1770, CG1770, DHDAC4, Dmel\\CG1770, GC1770, HDACa, dHDAC4, dmHDA405, hdac4, HDAC4', description='Histone deacetylase 4', other_designations='CG1770-PB|CG1770-PD|CG1770-PE|CG1770-PF|CG1770-PG|CG1770-PH|CG1770-PI|CG1770-PJ|HDAC4-PB|HDAC4-PD|HDAC4-PE|HDAC4-PF|HDAC4-PG|HDAC4-PH|HDAC4-PI|HDAC4-PJ|histone deacetylase 4', map_location='11E8-11E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13262687, end_position_on_the_genomic_accession=13285632, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  32279 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32279, CurrentID=0, Status='live', Symbol='CG15743', Aliases='Dmel_ Dmel\\CG15743', description='CG15743 gene product from transcript CG15743-RC', other_designations='CG15743-PA|CG15743-PB|CG15743-PC|uncharacterized protein', map_location='11E9-11E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13285707, end_position_on_the_genomic_accession=13287741, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32281 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32281, CurrentID=0, Status='live', Symbol='CG1764', Aliases='Dmel_ Dmel\\CG1764, cg1764', description='CG1764 gene product from transcript CG1764-RB', other_designations='CG1764-PA|CG1764-PB', map_location='11E9-11E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13294473, end_position_on_the_genomic_accession=13302415, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32282 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32282, CurrentID=0, Status='live', Symbol='CG1622', Aliases='Dmel_ Dmel\\CG1622', description='CG1622 gene product from transcript CG1622-RC', other_designations='CG1622-PA|CG1622-PB|CG1622-PC', map_location='11E9-11E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13297082, end_position_on_the_genomic_accession=13299910, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32285 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32285, CurrentID=0, Status='live', Symbol='IP3K2', Aliases='Dmel_CG45017, CG12724, CG15745, CG1630, CG34359, CG4501, CG45017, D-IP3K2, DIP3K2, Dmel\\CG45017, Dmel_CG12724, Dmel_CG15745, Dmel_CG1630, Dmel_CG34359, IP[[3]]K-2, IP[[3]]K2, NP2758, dmIP3K2, dmIP3Kbeta, dmIP[[3]]beta, ip3k2', description='Inositol 1,4,5-triphosphate kinase 2', other_designations='CG45017-PA|CG45017-PB|CG45017-PC|CG45017-PD|CG45017-PG|CG45017-PH|IP3K2-PA|IP3K2-PB|IP3K2-PC|IP3K2-PD|IP3K2-PG|IP3K2-PH|inositol 1,4,5-triphosphate kinase 2', map_location='11E11-11E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13303633, end_position_on_the_genomic_accession=13329303, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  261663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=261663, CurrentID=0, Status='live', Symbol='CG32335', Aliases='Dmel_ BcDNA:RE72485, Dmel\\CG32335', description='CG32335 gene product from transcript CG32335-RB', other_designations='CG32335-PB|uncharacterized protein', map_location='61F3-61F3', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1201871, end_position_on_the_genomic_accession=1203036, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32288 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32288, CurrentID=0, Status='live', Symbol='CG15747', Aliases='Dmel_ Dmel\\CG15747', description='CG15747 gene product from transcript CG15747-RA', other_designations='CG15747-PA|CG15747-PC|uncharacterized protein', map_location='11E12-11E13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13338318, end_position_on_the_genomic_accession=13341009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32290 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32290, CurrentID=0, Status='live', Symbol='Syt12', Aliases='Dmel_CG10617, CG10617, Dmel\\CG10617, Syt 12, dsyt12, syt12', description='Synaptotagmin 12', other_designations='CG10617-PA|CG10617-PB|Syt12-PA|Syt12-PB|synaptotagmin 12|syt12', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13359248, end_position_on_the_genomic_accession=13368179, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32291 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32291, CurrentID=0, Status='live', Symbol='dmrt11E', Aliases='Dmel_CG15749, CG15749, Dmel\\CG15749, Dmrt11E, dmrt11', description='doublesex-Mab related 11E', other_designations='CG15749-PA|dmrt11E-PA|doublesex-Mab-related 11E', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13368191, end_position_on_the_genomic_accession=13370472, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32292 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32292, CurrentID=0, Status='live', Symbol='CG1640', Aliases='Dmel_ Dmel\\CG1640', description='CG1640 gene product from transcript CG1640-RB', other_designations='CG1640-PA|CG1640-PB|CG1640-PC|CG1640-PD|CG1640-PE|CG1640-PF|CG1640-PG', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13370606, end_position_on_the_genomic_accession=13380160, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32293 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32293, CurrentID=0, Status='live', Symbol='Rbp1-like', Aliases='Dmel_CG1987, BEST:HL02433, BcDNA:RE47308, CG1987, Dmel\\CG1987, RBP1-like, Rbp1L, i159', description='CG1987 gene product from transcript CG1987-RC', other_designations='CG1987-PA|CG1987-PB|CG1987-PC|Rbp1-like-PA|Rbp1-like-PB|Rbp1-like-PC', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13380166, end_position_on_the_genomic_accession=13384877, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32294 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32294, CurrentID=0, Status='live', Symbol='Tim9a', Aliases='Dmel_CG1660, CG1660, Dmel\\CG1660, Tim9', description='CG1660 gene product from transcript CG1660-RA', other_designations='CG1660-PA|CG1660-PB|Tim9a-PA|Tim9a-PB', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13385210, end_position_on_the_genomic_accession=13385893, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32295 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32295, CurrentID=0, Status='live', Symbol='Yippee', Aliases='Dmel_CG1989, CG1989, Dmel\\CG1989', description='CG1989 gene product from transcript CG1989-RA', other_designations='CG1989-PA|CG1989-PB|Yippee-PA|Yippee-PB|yippee', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13385914, end_position_on_the_genomic_accession=13387417, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32296 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32296, CurrentID=0, Status='live', Symbol='CG1662', Aliases='Dmel_ Dmel\\CG1662', description='CG1662 gene product from transcript CG1662-RA', other_designations='CG1662-PA|CG1662-PB', map_location='11F1-11F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13387563, end_position_on_the_genomic_accession=13389332, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32297 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32297, CurrentID=0, Status='live', Symbol='CG1673', Aliases='Dmel_ Dmel\\CG1673', description='CG1673 gene product from transcript CG1673-RA', other_designations='CG1673-PA', map_location='11F1-11F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13396899, end_position_on_the_genomic_accession=13401224, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32298 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32298, CurrentID=0, Status='live', Symbol='CG12725', Aliases='Dmel_ Dmel\\CG12725', description='CG12725 gene product from transcript CG12725-RA', other_designations='CG12725-PA|uncharacterized protein', map_location='11F4-11F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13401839, end_position_on_the_genomic_accession=13402863, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32299 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32299, CurrentID=0, Status='live', Symbol='GstT4', Aliases='Dmel_CG1681, CG1681, Dmel\\CG1681, GST-theta, GSTT4, gstt4', description='Glutathione S transferase T4', other_designations='CG1681-PA|GstT4-PA|glutathione S transferase T4', map_location='11F4-11F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13404749, end_position_on_the_genomic_accession=13409323, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32300 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32300, CurrentID=0, Status='live', Symbol='CG1998', Aliases='Dmel_ Dmel\\CG1998', description='CG1998 gene product from transcript CG1998-RA', other_designations='CG1998-PA|CG1998-PB|CG1998-PC', map_location='11F4-11F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13409369, end_position_on_the_genomic_accession=13413074, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32301 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32301, CurrentID=0, Status='live', Symbol='Set2', Aliases='Dmel_CG1716, CG1716, Dmel\\CG1716, Hypb, SET2, dHypb, dSet2, set2', description='SET domain containing 2', other_designations='CG1716-PA|CG1716-PB|Set2-PA|Set2-PB|drosophila hypb', map_location='11F4-11F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13413242, end_position_on_the_genomic_accession=13421024, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32302 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32302, CurrentID=0, Status='live', Symbol='CG34324', Aliases='Dmel_ CG12726, Dmel\\CG34324', description='CG34324 gene product from transcript CG34324-RA', other_designations='CG34324-PA|uncharacterized protein', map_location='11F8-11F8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13456534, end_position_on_the_genomic_accession=13457663, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32303 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32303, CurrentID=0, Status='live', Symbol='Neto', Aliases='Dmel_CG44328, CG12727, CG15751, CG15752, CG15753, CG32635, CG44328, CT36003, Dmel\\CG44328, Dmel_CG15753, Dmel_CG32635, net, neto, tld-like', description='Neuropilin and tolloid-like', other_designations='CG44328-PA|CG44328-PB|Neto-PA|Neto-PB|neuropilin and tolloid-like', map_location='12A2-12A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13445903, end_position_on_the_genomic_accession=13527282, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  32307 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32307, CurrentID=0, Status='live', Symbol='CG15754', Aliases='Dmel_ Dmel\\CG15754', description='CG15754 gene product from transcript CG15754-RA', other_designations='CG15754-PA|uncharacterized protein', map_location='12A2-12A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13528006, end_position_on_the_genomic_accession=13529280, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32308 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32308, CurrentID=0, Status='live', Symbol='CG15756', Aliases='Dmel_ Dmel\\CG15756', description='CG15756 gene product from transcript CG15756-RA', other_designations='CG15756-PA|uncharacterized protein', map_location='12A3-12A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13534400, end_position_on_the_genomic_accession=13536202, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32309 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32309, CurrentID=0, Status='live', Symbol='Cpr12A', Aliases='Dmel_CG15757, BcDNA:LP12385, CG15757, Dmel\\CG15757', description='Cuticular protein 12A', other_designations='CG15757-PA|Cpr12A-PA|cuticular protein 12A', map_location='12A4-12A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13543643, end_position_on_the_genomic_accession=13544907, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32310 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32310, CurrentID=0, Status='live', Symbol='CG11816', Aliases='Dmel_ Dmel\\CG11816', description='CG11816 gene product from transcript CG11816-RD', other_designations='CG11816-PA|CG11816-PB|CG11816-PC|CG11816-PD|uncharacterized protein', map_location='12A4-12A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13552559, end_position_on_the_genomic_accession=13555800, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32311 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32311, CurrentID=0, Status='live', Symbol='CG32631', Aliases='Dmel_ CG1733, Dmel\\CG32631', description='CG32631 gene product from transcript CG32631-RA', other_designations='CG32631-PA|uncharacterized protein', map_location='12A4-12A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13556472, end_position_on_the_genomic_accession=13557866, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32312 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32312, CurrentID=0, Status='live', Symbol='Tango13', Aliases='Dmel_CG32632, CG1573, CG15758, CG15759, CG32629, CG32629/CG32632, CG32632, Dmel\\CG32632, Dmel_CG15758, Dmel_CG32629, TANGO13', description='Transport and Golgi organization 13', other_designations='CG32632-PB|CG32632-PC|CG32632-PD|CG32632-PE|CG32632-PF|CG32632-PG|Tango13-PB|Tango13-PC|Tango13-PD|Tango13-PE|Tango13-PF|Tango13-PG|transport and golgi organization 13', map_location='12A6-12A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13563823, end_position_on_the_genomic_accession=13593224, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  32314 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32314, CurrentID=0, Status='live', Symbol='up', Aliases='Dmel_CG7107, BcDNA:LD08591, CG7107, DmTpnT, Dmel\\CG7107, FBgn0004169, Tn-T, TnT, TpnT, cTNT, dro TnT, int, tnTheld2, wupB, up', description='upheld', other_designations='CG7107-PA|CG7107-PB|CG7107-PD|CG7107-PE|CG7107-PG|CG7107-PH|CG7107-PI|CG7107-PJ|CG7107-PK|CG7107-PL|CG7107-PM|CG7107-PN|CG7107-PO|CG7107-PP|CG7107-PQ|CG7107-PR|CG7107-PS|CG7107-PT|indented thorax|troponin|troponin T|troponin-T|troponinT|up-PA|up-PB|up-PD|up-PE|up-PG|up-PH|up-PI|up-PJ|up-PK|up-PL|up-PM|up-PN|up-PO|up-PP|up-PQ|up-PR|up-PS|up-PT', map_location='1-43.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13593474, end_position_on_the_genomic_accession=13602578, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32315 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32315, CurrentID=0, Status='live', Symbol='CG11178', Aliases='Dmel_ Dmel\\CG11178', description='CG11178 gene product from transcript CG11178-RC', other_designations='CG11178-PB|CG11178-PC|uncharacterized protein', map_location='12A8-12A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13605681, end_position_on_the_genomic_accession=13609529, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32316 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32316, CurrentID=0, Status='live', Symbol='Ndc80', Aliases='Dmel_CG9938, CG9938, Dmel\\CG9938, Hec1, NDC80, d dmNdc80, hec1', description='CG9938 gene product from transcript CG9938-RA', other_designations='CG9938-PA|Hec1|Ndc80-PA', map_location='12A8-12A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13609671, end_position_on_the_genomic_accession=13612062, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32317 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32317, CurrentID=0, Status='live', Symbol='BthD', Aliases='Dmel_CG11177, CG11177, Dmel\\CG11177, SelH, dSelH, dselH, dselM', description='BthD selenoprotein', other_designations='BthD-PA|BthDselenoprotein|CG11177-PA', map_location='12A8-12A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13612131, end_position_on_the_genomic_accession=13613228, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32318 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32318, CurrentID=0, Status='live', Symbol='tth', Aliases='Dmel_CG12175, CG12175, Dmel\\CG12175, anon-EST:fe2G7', description='toothrin', other_designations='CG12175-PB|CG12175-PC|anon-fast-evolving-2G7|toothreen|tth-PB|tth-PC', map_location='12A8-12A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13613193, end_position_on_the_genomic_accession=13616586, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32319 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32319, CurrentID=0, Status='live', Symbol='Tango2', Aliases='Dmel_CG11176, CG11176, Dmel\\CG11176, TANGO2', description='Transport and Golgi organization 2', other_designations='CG11176-PA|CG11176-PB|CG11176-PC|CG11176-PD|Tango2-PA|Tango2-PB|Tango2-PC|Tango2-PD|transport and golgi organization 2', map_location='12A8-12A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13616989, end_position_on_the_genomic_accession=13619207, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32320 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32320, CurrentID=0, Status='live', Symbol='CG2691', Aliases='Dmel_ Dmel\\CG2691', description='CG2691 gene product from transcript CG2691-RB', other_designations='CG2691-PA|CG2691-PB', map_location='12A8-12A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13619276, end_position_on_the_genomic_accession=13624408, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32321 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32321, CurrentID=0, Status='live', Symbol='NFAT', Aliases='Dmel_CG11172, CG11172, Dmel\\CG11172, EP1335, MESR1, NF-AT55, Nfat, dNFAT, NFAT', description='NFAT homolog', other_designations='CG11172-PA|CG11172-PB|CG11172-PC|CG11172-PD|CG11172-PE|CG11172-PG|CG11172-PH|NFAT-PA|NFAT-PB|NFAT-PC|NFAT-PD|NFAT-PE|NFAT-PG|NFAT-PH|Rel domain protein|drosophila TonEBP|misexpression suppressor of Ras 1', map_location='12A9-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13624745, end_position_on_the_genomic_accession=13664568, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  32322 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32322, CurrentID=0, Status='live', Symbol='Lig4', Aliases='Dmel_CG12176, CG12176, Dmel\\CG12176, RE37186, lig4, ligIV, ligaseIV', description='Ligase4', other_designations='CG12176-PA|DNA ligase 4|DNA ligase IV|Lig4-PA|ligase 4|ligase IV', map_location='12B2-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13664584, end_position_on_the_genomic_accession=13667811, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32323 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32323, CurrentID=0, Status='live', Symbol='CG11164', Aliases='Dmel_ Dmel\\CG11164', description='CG11164 gene product from transcript CG11164-RB', other_designations='CG11164-PA|CG11164-PB|uncharacterized protein', map_location='12B2-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13668014, end_position_on_the_genomic_accession=13669573, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32324 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32324, CurrentID=0, Status='live', Symbol='CG15760', Aliases='Dmel_ Dmel\\CG15760', description='CG15760 gene product from transcript CG15760-RB', other_designations='CG15760-PA|CG15760-PB|CG15760-PC|uncharacterized protein', map_location='12B2-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13669699, end_position_on_the_genomic_accession=13676232, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32325 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32325, CurrentID=0, Status='live', Symbol='CG11162', Aliases='Dmel_ Dmel\\CG11162', description='CG11162 gene product from transcript CG11162-RA', other_designations='CG11162-PA|uncharacterized protein', map_location='12B2-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13676894, end_position_on_the_genomic_accession=13677882, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32326 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32326, CurrentID=0, Status='live', Symbol='CG12177', Aliases='Dmel_ Dmel\\CG12177', description='CG12177 gene product from transcript CG12177-RA', other_designations='CG12177-PA|CG12177-PB|CG12177-PC|uncharacterized protein', map_location='12B2-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13677899, end_position_on_the_genomic_accession=13680500, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32327 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32327, CurrentID=0, Status='live', Symbol='CG11158', Aliases='Dmel_ Dmel\\CG11158', description='CG11158 gene product from transcript CG11158-RA', other_designations='CG11158-PA|uncharacterized protein', map_location='12B2-12B2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13682952, end_position_on_the_genomic_accession=13684616, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32328 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32328, CurrentID=0, Status='live', Symbol='CG9940', Aliases='Dmel_ 145027_at, Dmel\\CG9940, NADS, Q9VYA0', description='CG9940 gene product from transcript CG9940-RB', other_designations='CG9940-PA|CG9940-PB|CG9940-PC', map_location='12B2-12B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13684380, end_position_on_the_genomic_accession=13718285, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32329 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32329, CurrentID=0, Status='live', Symbol='NnaD', Aliases='Dmel_CG44533, CG11428, CG32627, CG32628, CG44533, Dmel\\CG44533, Dmel_CG32627, Dmel_CG32628', description='Drosophila Nna1 ortholog', other_designations='CG44533-PA|CG44533-PB|CG44533-PC|CG44533-PD|CG44533-PE|CG44533-PF|CG44533-PG|CG44533-PH|CG44533-PI|CG44533-PJ|NnaD-PA|NnaD-PB|NnaD-PC|NnaD-PD|NnaD-PE|NnaD-PF|NnaD-PG|NnaD-PH|NnaD-PI|NnaD-PJ|drosophila Nna1 ortholog', map_location='12B4-12B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13692292, end_position_on_the_genomic_accession=13717835, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  32330 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32330, CurrentID=0, Status='live', Symbol='CG9941', Aliases='Dmel_ A, Dmel\\CG9941, x93', description='CG9941 gene product from transcript CG9941-RA', other_designations='CG9941-PA', map_location='12B4-12B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13718734, end_position_on_the_genomic_accession=13721977, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19835490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=19835490, CurrentID=0, Status='live', Symbol='CG45063', Aliases='Dmel_ CG3701, Dmel\\CG45063', description='CG45063 gene product from transcript CG45063-RA', other_designations='CG45063-PA|uncharacterized protein', map_location='58E3-58E3', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=22476385, end_position_on_the_genomic_accession=22477295, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32334 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32334, CurrentID=0, Status='live', Symbol='CG11134', Aliases='Dmel_ 1, Dmel\\CG11134', description='CG11134 gene product from transcript CG11134-RB', other_designations='CG11134-PA|CG11134-PB|transcript 1|uncharacterized protein', map_location='12B4-12B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13737942, end_position_on_the_genomic_accession=13739177, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32335 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32335, CurrentID=0, Status='live', Symbol='CG11151', Aliases='Dmel_ 2, Dmel\\CG11151', description='CG11151 gene product from transcript CG11151-RA', other_designations='CG11151-PA|CG11151-PB|transcript 2|uncharacterized protein', map_location='12B4-12B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13739744, end_position_on_the_genomic_accession=13741952, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32337 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32337, CurrentID=0, Status='live', Symbol='Pdcd4', Aliases='Dmel_CG10990, CG10990, Dmel\\CG10990', description='Programmed cell death 4 ortholog', other_designations='CG10990-PA|CG10990-PB|CG10990-PC|CG10990-PD|Pdcd4-PA|Pdcd4-PB|Pdcd4-PC|Pdcd4-PD|programmed cell death 4|programmed cell death 4 ortholog', map_location='12B4-12B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13745577, end_position_on_the_genomic_accession=13755406, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32338 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32338, CurrentID=0, Status='live', Symbol='Rtc1', Aliases='Dmel_CG11130, CG11130, Dmel\\CG11130', description='CG11130 gene product from transcript CG11130-RA', other_designations='CG11130-PA|Rtc1-PA', map_location='12B10-12C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13757179, end_position_on_the_genomic_accession=13758598, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32339 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32339, CurrentID=0, Status='live', Symbol='Yp3', Aliases='Dmel_CG11129, CG11129, DIM 30, Dmel\\CG11129, VG3, VIT3Dm, YP, YP-3, YP3, YP[[3]], fs(1)1526, fs(1)A1526, yp, yp3', description='Yolk protein 3', other_designations='CG11129-PA|CG11129-PB|Yp3-PA|Yp3-PB|vitellogenin|yolk protein|yolk protein 3', map_location='1-44 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13759546, end_position_on_the_genomic_accession=13761547, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32340 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32340, CurrentID=0, Status='live', Symbol='rdgB', Aliases='Dmel_CG11111, CG11111, Dmel\\CG11111, DrdgB, DrdgBalpha, RdgB, RdgBalpha, hypoF, ota1, x36', description='retinal degeneration B', other_designations='CG11111-PA|CG11111-PB|CG11111-PC|CG11111-PD|CG11111-PE|CG11111-PF|CG11111-PG|CG11111-PH|CG11111-PI|CG11111-PJ|CG11111-PK|rdgB-PA|rdgB-PB|rdgB-PC|rdgB-PD|rdgB-PE|rdgB-PF|rdgB-PG|rdgB-PH|rdgB-PI|rdgB-PJ|rdgB-PK|receptor degeneration B|receptor-degeneration-II', map_location='1-42.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13762752, end_position_on_the_genomic_accession=13779488, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  32341 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32341, CurrentID=0, Status='live', Symbol='CtsB1', Aliases='Dmel_CG10992, CB, CG-10992, CG10992, CTSB1, Dmel\\CG10992, anon-EST:ParkEST132', description='Cathepsin B1', other_designations='CG10992-PA|CG10992-PC|CtsB1-PA|CtsB1-PC|cathepsin B|cathepsin B1', map_location='12C4-12C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13779847, end_position_on_the_genomic_accession=13782170, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32342 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32342, CurrentID=0, Status='live', Symbol='CG11103', Aliases='Dmel_ BcDNA:LP03404, Dmel\\CG11103', description='CG11103 gene product from transcript CG11103-RB', other_designations='CG11103-PB|uncharacterized protein', map_location='12C4-12C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13782459, end_position_on_the_genomic_accession=13783381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32343 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32343, CurrentID=0, Status='live', Symbol='inaE', Aliases='Dmel_CG33174, CG11096, CG11102, CG11104, CG15761, CG32630, CG33174, DAGL, Dmel\\CG33174, EP(X)1101', description='inactivation no afterpotential E', other_designations='CG33174-PA|CG33174-PD|CG33174-PE|CG33174-PF|CG33174-PG|CG33174-PH|diacylglycerol lipase|inaE-PA|inaE-PD|inaE-PE|inaE-PF|inaE-PG|inaE-PH', map_location='12C4-12C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13783714, end_position_on_the_genomic_accession=13811493, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  32344 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32344, CurrentID=0, Status='live', Symbol='CG10993', Aliases='Dmel_ BcDNA:GH23064, Dmel\\CG10993', description='CG10993 gene product from transcript CG10993-RA', other_designations='CG10993-PA|uncharacterized protein', map_location='12C5-12C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13790338, end_position_on_the_genomic_accession=13791111, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32345 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32345, CurrentID=0, Status='live', Symbol='CG10996', Aliases='Dmel_ Dmel\\CG10996', description='CG10996 gene product from transcript CG10996-RA', other_designations='CG10996-PA|uncharacterized protein', map_location='12C5-12C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13796389, end_position_on_the_genomic_accession=13797638, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32348 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32348, CurrentID=0, Status='live', Symbol='CG11095', Aliases='Dmel_ Dmel\\CG11095', description='CG11095 gene product from transcript CG11095-RA', other_designations='CG11095-PA|uncharacterized protein', map_location='12C5-12C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13811684, end_position_on_the_genomic_accession=13812870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32349 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32349, CurrentID=0, Status='live', Symbol='Clic', Aliases='Dmel_CG10997, CG10997, DCLIC, DmCLIC, Dmel\\CG10997, clic, l(1)G0053, l(1)G0472', description='Chloride intracellular channel', other_designations='CG10997-PA|CG10997-PC|Clic-PA|Clic-PC|chloride intracellular channel', map_location='12C5-12C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13812971, end_position_on_the_genomic_accession=13822307, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32350 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32350, CurrentID=0, Status='live', Symbol='Nup93-1', Aliases='Dmel_CG11092, BcDNA:LD21129, CG11092, Dmel\\CG11092, Nup93, nup93', description='Nucleoporin 93kD-1', other_designations='CG11092-PA|Nup93-1-PA|nucleoporin 93kD-1', map_location='12C6-12C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13822826, end_position_on_the_genomic_accession=13825910, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32351 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32351, CurrentID=0, Status='live', Symbol='jub', Aliases='Dmel_CG11063, CG11063, Dmel\\CG11063, Jub, djub', description='Ajuba LIM protein', other_designations='CG11063-PB|CG11063-PC|CG11063-PD|CG11063-PE|ajuba|ajuba LIM protein|drosophila ajuba LIM protein|jub-PB|jub-PC|jub-PD|jub-PE', map_location='12C6-12C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13826041, end_position_on_the_genomic_accession=13830317, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32352 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32352, CurrentID=0, Status='live', Symbol='AMPdeam', Aliases='Dmel_CG32626, AMP deam, CG11058, CG11065, CG15762, CG32626, Dmel\\CG32626', description='AMP deaminase', other_designations='AMPdeam-PC|AMPdeam-PE|AMPdeam-PF|AMPdeam-PG|AMPdeam-PH|AMPdeam-PI|AMPdeam-PJ|AMPdeam-PK|AMPdeam-PL|CG32626-PC|CG32626-PE|CG32626-PF|CG32626-PG|CG32626-PH|CG32626-PI|CG32626-PJ|CG32626-PK|CG32626-PL', map_location='12C6-12C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13830741, end_position_on_the_genomic_accession=13845866, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32353 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32353, CurrentID=0, Status='live', Symbol='mamo', Aliases='Dmel_CG45477, CG11071, CG11072, CG11075, CG11082, CG32606, CG32611, CG34346, CG44132, CG45477, CR44656, Dmel\\CG45477, Dmel_CG11071, Dmel_CG11072, Dmel_CG32606, Dmel_CG32611, Dmel_CG34346, Dmel_CG44132, lincRNA.990', description='maternal gene required for meiosis', other_designations='CG45477-PC|CG45477-PD|CG45477-PE|CG45477-PF|CG45477-PG|CG45477-PH|CG45477-PI|mamo-PC|mamo-PD|mamo-PE|mamo-PF|mamo-PG|mamo-PH|mamo-PI', map_location='12D1-12D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13846116, end_position_on_the_genomic_accession=13990527, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  32357 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32357, CurrentID=0, Status='live', Symbol='CG11068', Aliases='Dmel_ Dmel\\CG11068', description='CG11068 gene product from transcript CG11068-RA', other_designations='CG11068-PA|uncharacterized protein', map_location='12D1-12D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13955236, end_position_on_the_genomic_accession=13956574, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32358 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32358, CurrentID=0, Status='live', Symbol='ben', Aliases='Dmel_CG18319, BEN, Ben, Bend, CG18319, Dmel\\CG18319, UBC13, Ubc13, UbcD3, anon-WO03040301.242, nj-262, ubc13', description='bendless', other_designations='CG18319-PA|CG18319-PB|CG18319-PC|CG18319-PD|CG18319-PE|ben-PA|ben-PB|ben-PC|ben-PD|ben-PE|nonjumper-262', map_location='1-45.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=13996345, end_position_on_the_genomic_accession=13999805, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32359 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32359, CurrentID=0, Status='live', Symbol='CG12480', Aliases='Dmel_ Dmel\\CG12480', description='CG12480 gene product from transcript CG12480-RA', other_designations='CG12480-PA|CG12480-PB|uncharacterized protein', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14169592, end_position_on_the_genomic_accession=14171545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32360 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32360, CurrentID=0, Status='live', Symbol='betaNACtes1', Aliases='Dmel_CG13402, CG13402, Dmel\\CG13402', description='CG13402 gene product from transcript CG13402-RA', other_designations='CG13402-PA|betaNACtes1-PA', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14171487, end_position_on_the_genomic_accession=14172624, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32361 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32361, CurrentID=0, Status='live', Symbol='CG1368', Aliases='Dmel_ Dmel\\CG1368', description='CG1368 gene product from transcript CG1368-RA', other_designations='CG1368-PA', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14174756, end_position_on_the_genomic_accession=14175670, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32362 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32362, CurrentID=0, Status='live', Symbol='CG11581', Aliases='Dmel_ Dmel\\CG11581', description='CG11581 gene product from transcript CG11581-RB', other_designations='CG11581-PA|CG11581-PB|uncharacterized protein', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14176923, end_position_on_the_genomic_accession=14178456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32363 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32363, CurrentID=0, Status='live', Symbol='CG11584', Aliases='Dmel_ Dmel\\CG11584', description='CG11584 gene product from transcript CG11584-RB', other_designations='CG11584-PB|uncharacterized protein', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14181028, end_position_on_the_genomic_accession=14183315, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32364 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32364, CurrentID=0, Status='live', Symbol='CG12481', Aliases='Dmel_ Dmel\\CG12481', description='CG12481 gene product from transcript CG12481-RB', other_designations='CG12481-PB|uncharacterized protein', map_location='12E2-12E2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14187253, end_position_on_the_genomic_accession=14187841, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32365 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32365, CurrentID=0, Status='live', Symbol='CG11585', Aliases='Dmel_ Dmel\\CG11585', description='CG11585 gene product from transcript CG11585-RB', other_designations='CG11585-PB|uncharacterized protein', map_location='12E3-12E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14189516, end_position_on_the_genomic_accession=14190755, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32366 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32366, CurrentID=0, Status='live', Symbol='CG13403', Aliases='Dmel_ Dmel\\CG13403', description='CG13403 gene product from transcript CG13403-RA', other_designations='CG13403-PA|CG13403-PB|uncharacterized protein', map_location='12E3-12E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14191220, end_position_on_the_genomic_accession=14192018, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32367 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32367, CurrentID=0, Status='live', Symbol='yl', Aliases='Dmel_CG1372, CG1372, CT3114, DmYPR, Dmel\\CG1372, LR13, YL, Yl, fs(1)148, fs(1)5, fs(1)A148, fs(1)K184, fs(1)L186, fs(1)L193, fs(1)L196, fs(1)L211, fs(1)M205, fs(1)M25, fs(1)yl, fs(1)yolkless', description='yolkless', other_designations='CG1372-PA|CG1372-PB|Dm yolkless|YP receptor|yl-PA|yl-PB', map_location='1-48 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14193214, end_position_on_the_genomic_accession=14199958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32368 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32368, CurrentID=0, Status='live', Symbol='CG11590', Aliases='Dmel_ BcDNA:GM24986, Dmel\\CG11590', description='CG11590 gene product from transcript CG11590-RA', other_designations='CG11590-PA|uncharacterized protein', map_location='12E3-12E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14200178, end_position_on_the_genomic_accession=14200907, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32371 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32371, CurrentID=0, Status='live', Symbol='l(1)G0469', Aliases='Dmel_CG32597, CG11595, CG12482, CG32597, Dmel\\CG32597, anon-WO03040301.232', description='lethal (1) G0469', other_designations='CG32597-PA|CG32597-PC|CG32597-PD|l(1)G0469-PA|l(1)G0469-PC|l(1)G0469-PD', map_location='12E4-12E4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14203961, end_position_on_the_genomic_accession=14211457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32373 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32373, CurrentID=0, Status='live', Symbol='l(1)G0007', Aliases='Dmel_CG32604, CG1375, CG1405, CG32604, Dmel\\CG32604, EP764, EP784, Prp16, cg1405, l(1)G0028, l(1)G0103, l(1)G0176, l(1)G0308, l(1)G0416, l(1)G0476, l(1)G0491, prp16', description='lethal (1) G0007', other_designations='CG32604-PA|CG32604-PB|l(1)G0007-PA|l(1)G0007-PB', map_location='12E3-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14201362, end_position_on_the_genomic_accession=14224732, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32374 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32374, CurrentID=0, Status='live', Symbol='CG11674', Aliases='Dmel_ Dmel\\CG11674', description='CG11674 gene product from transcript CG11674-RA', other_designations='CG11674-PA|uncharacterized protein', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14224647, end_position_on_the_genomic_accession=14225879, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32375 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32375, CurrentID=0, Status='live', Symbol='mRpL38', Aliases='Dmel_CG15871, CG15871, Dmel\\CG15871, MRP-L3, mRpL3a', description='mitochondrial ribosomal protein L38', other_designations='CG15871-PA|mRpL38-PA|mitochondrial ribosomal protein L3a', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14226289, end_position_on_the_genomic_accession=14227949, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32377 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32377, CurrentID=0, Status='live', Symbol='CG1434', Aliases='Dmel_ Dmel\\CG1434', description='CG1434 gene product from transcript CG1434-RB', other_designations='CG1434-PA|CG1434-PB', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14237988, end_position_on_the_genomic_accession=14240152, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32378 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32378, CurrentID=0, Status='live', Symbol='Fbxl4', Aliases='Dmel_CG1839, CG1839, Dmel\\CG1839, FBXL4, dFbxl4', description='F box and leucine-rich-repeat gene 4', other_designations='CG1839-PA|Fbxl4-PA', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14234413, end_position_on_the_genomic_accession=14237923, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32379 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32379, CurrentID=0, Status='live', Symbol='mRNA-cap', Aliases='Dmel_CG1810, CG1810, Dmel\\CG1810, PIR', description='mRNA-capping-enzyme', other_designations='CG1810-PA|mRNA capping enzyme|mRNA-cap-PA', map_location='12E5-12E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14240266, end_position_on_the_genomic_accession=14242822, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32381 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32381, CurrentID=0, Status='live', Symbol='CG1461', Aliases='Dmel_ Dmel\\CG1461', description='CG1461 gene product from transcript CG1461-RB', other_designations='CG1461-PA|CG1461-PB', map_location='12E6-12E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14261223, end_position_on_the_genomic_accession=14265379, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32382 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32382, CurrentID=0, Status='live', Symbol='CG13404', Aliases='Dmel_ Dmel\\CG13404', description='CG13404 gene product from transcript CG13404-RB', other_designations='CG13404-PB|uncharacterized protein', map_location='12E6-12E6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14264737, end_position_on_the_genomic_accession=14266250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32384 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32384, CurrentID=0, Status='live', Symbol='Muc12Ea', Aliases='Dmel_CG32602, CG32602, CG5228, Dmel\\CG32602', description='Mucin 12Ea', other_designations='CG32602-PA|Muc12Ea-PA|mucin 12Ea', map_location='12E7-12E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14277735, end_position_on_the_genomic_accession=14288161, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32385 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32385, CurrentID=0, Status='live', Symbol='CG9400', Aliases='Dmel_ Dmel\\CG9400, anon-WO0140519.10', description='CG9400 gene product from transcript CG9400-RA', other_designations='CG9400-PA', map_location='12E7-12E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14288084, end_position_on_the_genomic_accession=14289393, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32386 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32386, CurrentID=0, Status='live', Symbol='betaNACtes2', Aliases='Dmel_CG18157, CG18157, Dmel\\CG18157', description='CG18157 gene product from transcript CG18157-RA', other_designations='CG18157-PA|betaNACtes2-PA', map_location='12E8-12E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14314623, end_position_on_the_genomic_accession=14315583, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32387 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32387, CurrentID=0, Status='live', Symbol='dpr8', Aliases='Dmel_CG32600, CG11477, CG14415, CG32600, CG5291, CT16867, Dmel\\CG32600, Dpr-8, Dpr8', description='defective proboscis extension response 8', other_designations='CG32600-PA|CG32600-PB|dpr8-PA|dpr8-PB', map_location='12E8-12E10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14322226, end_position_on_the_genomic_accession=14462351, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  32389 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32389, CurrentID=0, Status='live', Symbol='betaNACtes4', Aliases='Dmel_CG18313, CG18313, Dmel\\CG18313', description='CG18313 gene product from transcript CG18313-RA', other_designations='CG18313-PA|betaNACtes4-PA', map_location='12E9-12E9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14380351, end_position_on_the_genomic_accession=14381416, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32392 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32392, CurrentID=0, Status='live', Symbol='CG9411', Aliases='Dmel_ Dmel\\CG9411', description='CG9411 gene product from transcript CG9411-RA', other_designations='CG9411-PA', map_location='12E10-12E10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14472408, end_position_on_the_genomic_accession=14480086, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32393 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32393, CurrentID=0, Status='live', Symbol='CG12540', Aliases='Dmel_ BcDNA:RE49992, Dmel\\CG12540', description='CG12540 gene product from transcript CG12540-RA', other_designations='CG12540-PA|uncharacterized protein', map_location='12E11-12E11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14496135, end_position_on_the_genomic_accession=14497240, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32394 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32394, CurrentID=0, Status='live', Symbol='CG14414', Aliases='Dmel_ BcDNA:AT01764, Dmel\\CG14414, cg14414', description='CG14414 gene product from transcript CG14414-RC', other_designations='CG14414-PA|CG14414-PC|CG14414-PD|CG14414-PE|uncharacterized protein', map_location='12F1-12F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14580600, end_position_on_the_genomic_accession=14582174, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32395 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32395, CurrentID=0, Status='live', Symbol='mRpS25', Aliases='Dmel_CG14413, 25mRps, CG14413, Dmel\\CG14413, MRP-S25, MrpS25', description='mitochondrial ribosomal protein S25', other_designations='CG14413-PA|mRpS25-PA', map_location='12F1-12F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14582136, end_position_on_the_genomic_accession=14582982, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32396 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32396, CurrentID=0, Status='live', Symbol='nmdyn-D6', Aliases='Dmel_CG5310, CG5310, Dmel\\CG5310', description='CG5310 gene product from transcript CG5310-RA', other_designations='CG5310-PA|nmdyn-D6-PA', map_location='12F1-12F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14583677, end_position_on_the_genomic_accession=14584616, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32397 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32397, CurrentID=0, Status='live', Symbol='CG9413', Aliases='Dmel_ Dmel\\CG9413', description='CG9413 gene product from transcript CG9413-RD', other_designations='CG9413-PA|CG9413-PB|CG9413-PC|CG9413-PD', map_location='12F1-12F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14584733, end_position_on_the_genomic_accession=14602580, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32398 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32398, CurrentID=0, Status='live', Symbol='NetA', Aliases='Dmel_CG18657, CG18657, CT27014, Dmel\\CG18657, NetrinA, net, netA, netrin, netrin A', description='Netrin-A', other_designations='CG18657-PA|CG18657-PB|CG18657-PC|D-netrin|D-netrin-A|NetA-PA|NetA-PB|NetA-PC|NetrinA|netrin|netrin A', map_location='12F1-12F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14604012, end_position_on_the_genomic_accession=14653928, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32399 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32399, CurrentID=0, Status='live', Symbol='CG5321', Aliases='Dmel_ Dmel\\CG5321, anon-EST:Posey66', description='CG5321 gene product from transcript CG5321-RB', other_designations='CG5321-PB|CG5321-PC', map_location='12F2-12F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14679568, end_position_on_the_genomic_accession=14681418, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32400 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32400, CurrentID=0, Status='live', Symbol='NetB', Aliases='Dmel_CG10521, CG10521, CT29512, Dmel\\CG10521, NETB, net, netB, netrin, netrin B', description='Netrin-B', other_designations='CG10521-PA|CG10521-PB|CG10521-PC|CG10521-PD|CG10521-PE|CG10521-PF|CG10521-PG|D-netrin|D-netrin-B|NetB-PA|NetB-PB|NetB-PC|NetB-PD|NetB-PE|NetB-PF|NetB-PG|netrin|netrin B|netrin-B|netrinB', map_location='12F2-12F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14686038, end_position_on_the_genomic_accession=14749376, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32401 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32401, CurrentID=0, Status='live', Symbol='CG15890', Aliases='Dmel_ Dmel\\CG15890', description='CG15890 gene product from transcript CG15890-RB', other_designations='CG15890-PA|CG15890-PB|uncharacterized protein', map_location='12F4-12F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14757531, end_position_on_the_genomic_accession=14765787, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32402 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32402, CurrentID=0, Status='live', Symbol='CG5334', Aliases='Dmel_ Dmel\\CG5334', description='CG5334 gene product from transcript CG5334-RA', other_designations='CG5334-PA', map_location='12F4-12F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14775308, end_position_on_the_genomic_accession=14776299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32403 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32403, CurrentID=0, Status='live', Symbol='CG5347', Aliases='Dmel_ Dmel\\CG5347', description='CG5347 gene product from transcript CG5347-RA', other_designations='CG5347-PA', map_location='12F4-12F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14776774, end_position_on_the_genomic_accession=14777983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  5740423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=5740423, CurrentID=0, Status='live', Symbol='CG34184', Aliases='Dmel_ Dmel\\CG34184', description='CG34184 gene product from transcript CG34184-RA', other_designations='CG34184-PA|uncharacterized protein', map_location='50F6-50F6', chromosome='2R', genomic_nucleotide_accession_version='NT_033778.4', start_position_on_the_genomic_accession=14365656, end_position_on_the_genomic_accession=14366658, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32406 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32406, CurrentID=0, Status='live', Symbol='rut', Aliases='Dmel_CG9533, AC, Ac12F, CG9533, Dmel\\CG9533, EP1603, RUT, Rut, Rut-AC', description='rutabaga', other_designations='CG9533-PA|CG9533-PB|CG9533-PC|CG9533-PD|CG9533-PE|CG9533-PF|CG9533-PG|RUTABAGA|rut-PA|rut-PB|rut-PC|rut-PD|rut-PE|rut-PF|rut-PG|rutabage', map_location='12F4-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14787478, end_position_on_the_genomic_accession=14826073, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  32407 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32407, CurrentID=0, Status='live', Symbol='CG14408', Aliases='Dmel_ Dmel\\CG14408', description='CG14408 gene product from transcript CG14408-RH', other_designations='CG14408-PA|CG14408-PE|CG14408-PF|CG14408-PG|CG14408-PH|uncharacterized protein', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14826867, end_position_on_the_genomic_accession=14832033, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32408 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32408, CurrentID=0, Status='live', Symbol='CG14411', Aliases='Dmel_ Dmel\\CG14411', description='CG14411 gene product from transcript CG14411-RA', other_designations='CG14411-PA|CG14411-PB|CG14411-PC|CG14411-PD|uncharacterized protein', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14832647, end_position_on_the_genomic_accession=14836262, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32409 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32409, CurrentID=0, Status='live', Symbol='CG14410', Aliases='Dmel_ Dmel\\CG14410', description='CG14410 gene product from transcript CG14410-RA', other_designations='CG14410-PA|uncharacterized protein', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14836725, end_position_on_the_genomic_accession=14837330, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32410 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32410, CurrentID=0, Status='live', Symbol='CG14407', Aliases='Dmel_ BcDNA:RH03087, Dmel\\CG14407, Grx5', description='CG14407 gene product from transcript CG14407-RA', other_designations='CG14407-PA|uncharacterized protein', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14837471, end_position_on_the_genomic_accession=14838198, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32412 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32412, CurrentID=0, Status='live', Symbol='CG12539', Aliases='Dmel_ Dm GMCiota2, Dmel\\CG12539', description='CG12539 gene product from transcript CG12539-RA', other_designations='CG12539-PA|uncharacterized protein', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14842617, end_position_on_the_genomic_accession=14844757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32413 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32413, CurrentID=0, Status='live', Symbol='CG9522', Aliases='Dmel_ Dm GMCiota1, Dmel\\CG9522', description='CG9522 gene product from transcript CG9522-RA', other_designations='CG9522-PA', map_location='12F5-12F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14846472, end_position_on_the_genomic_accession=14848808, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32414 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32414, CurrentID=0, Status='live', Symbol='CG9521', Aliases='Dmel_ Dm GMCtheta2, Dmel\\CG9521', description='CG9521 gene product from transcript CG9521-RA', other_designations='CG9521-PA', map_location='12F6-12F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14851005, end_position_on_the_genomic_accession=14853269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32415 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32415, CurrentID=0, Status='live', Symbol='CG9519', Aliases='Dmel_ Dm GMCtheta1, Dmel\\CG9519', description='CG9519 gene product from transcript CG9519-RA', other_designations='CG9519-PA', map_location='12F6-12F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14854328, end_position_on_the_genomic_accession=14856361, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32416 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32416, CurrentID=0, Status='live', Symbol='CG9518', Aliases='Dmel_ Dm GMCzeta1, Dmel\\CG9518', description='CG9518 gene product from transcript CG9518-RA', other_designations='CG9518-PA|CG9518-PB|CG9518-PC', map_location='12F6-12F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14862989, end_position_on_the_genomic_accession=14871181, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32418 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32418, CurrentID=0, Status='live', Symbol='CG9514', Aliases='Dmel_ Dm GMCdelta1, Dmel\\CG9514', description='CG9514 gene product from transcript CG9514-RA', other_designations='CG9514-PA', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14885952, end_position_on_the_genomic_accession=14896891, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32419 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32419, CurrentID=0, Status='live', Symbol='CG9512', Aliases='Dmel_ CT26924, Dmel\\CG9512', description='CG9512 gene product from transcript CG9512-RB', other_designations='CG9512-PA|CG9512-PB', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14901844, end_position_on_the_genomic_accession=14905549, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32420 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32420, CurrentID=0, Status='live', Symbol='CG9509', Aliases='Dmel_ Dm GMCbeta2, Dmel\\CG9509', description='CG9509 gene product from transcript CG9509-RB', other_designations='CG9509-PA|CG9509-PB', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14906334, end_position_on_the_genomic_accession=14909708, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32421 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32421, CurrentID=0, Status='live', Symbol='CG14406', Aliases='Dmel_ Dmel\\CG14406', description='CG14406 gene product from transcript CG14406-RA', other_designations='CG14406-PA|CG14406-PB|uncharacterized protein', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14909873, end_position_on_the_genomic_accession=14910845, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32422 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32422, CurrentID=0, Status='live', Symbol='CG12398', Aliases='Dmel_ Dm GMCgamma1, Dmel\\CG12398', description='CG12398 gene product from transcript CG12398-RA', other_designations='CG12398-PA|uncharacterized protein', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14911781, end_position_on_the_genomic_accession=14913812, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32423 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32423, CurrentID=0, Status='live', Symbol='Eo', Aliases='Dmel_CG9504, CG9504, DmEO, Dmel\\CG9504, EObeta1', description='Ecdysone oxidase', other_designations='CG9504-PA|Eo-PA|ecdysone oxidase', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14916756, end_position_on_the_genomic_accession=14918800, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32424 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32424, CurrentID=0, Status='live', Symbol='CG9503', Aliases='Dmel_ Dm GMCalpha1, Dmel\\CG9503', description='CG9503 gene product from transcript CG9503-RA', other_designations='CG9503-PA', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14919252, end_position_on_the_genomic_accession=14923493, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32425 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32425, CurrentID=0, Status='live', Symbol='Flo2', Aliases='Dmel_CG32593, CG11547, CG14409, CG32593, DmFlo2, Dmel\\CG32593, FLO2, FLO[Dm]-2, Flo, Flo-2, flo2, mdcds_54791, reggie-1', description='Flotillin 2', other_designations='CG32593-PA|CG32593-PB|CG32593-PD|CG32593-PE|CG32593-PF|CG32593-PG|CG32593-PH|CG32593-PI|CG32593-PJ|CG32593-PK|CG32593-PL|CG32593-PM|Flo2-PA|Flo2-PB|Flo2-PD|Flo2-PE|Flo2-PF|Flo2-PG|Flo2-PH|Flo2-PI|Flo2-PJ|Flo2-PK|Flo2-PL|Flo2-PM|Flotillin-2|flotillin|flotillin 2|flotillin2|flottilin 2|reggie-1|reggie1', map_location='13-13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14839376, end_position_on_the_genomic_accession=14933946, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32426 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32426, CurrentID=0, Status='live', Symbol='pdgy', Aliases='Dmel_CG9009, BcDNA:GH02901, CG9009, Dmel\\CG9009', description='pudgy', other_designations='CG9009-PA|CG9009-PB|CG9009-PC|CG9009-PD|long chain fatty acid CoA ligase|pdgy-PA|pdgy-PB|pdgy-PC|pdgy-PD', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14933670, end_position_on_the_genomic_accession=14953532, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32427 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32427, CurrentID=0, Status='live', Symbol='CG9030', Aliases='Dmel_ Dmel\\CG9030', description='CG9030 gene product from transcript CG9030-RB', other_designations='CG9030-PB', map_location='13A1-13A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14954123, end_position_on_the_genomic_accession=14955100, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32428 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32428, CurrentID=0, Status='live', Symbol='eag', Aliases='Dmel_CG10952, CG10952, CIKE, Dmel\\CG10952, EAG, Eag', description='ether a go-go', other_designations='CG10952-PA|CG10952-PB|CG10952-PC|CG10952-PD|CG10952-PE|CG10952-PF|CG10952-PG|eag-PA|eag-PB|eag-PC|eag-PD|eag-PE|eag-PF|eag-PG|ether-a go-go|ether-a-go-go|ether-a-gogo', map_location='1-50.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=14955477, end_position_on_the_genomic_accession=15009989, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  32429 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32429, CurrentID=0, Status='live', Symbol='hiw', Aliases='Dmel_CG32592, CG32592, CG9041, CG9049, Dmel\\CG32592, Hiw', description='highwire', other_designations='CG32592-PA|CG32592-PB|hiw-PA|hiw-PB', map_location='13A5-13A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15010137, end_position_on_the_genomic_accession=15063429, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  32430 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32430, CurrentID=0, Status='live', Symbol='be', Aliases='Dmel_CG32594, BEST:LD10566, CG15026, CG32594, CG5530, Dmel\\CG32594, E, anon-EST:CL1a5', description='ben', other_designations='CG32594-PB|CG32594-PC|CG32594-PF|E protein|be-PB|be-PC|be-PF|gene A', map_location='13A5-13A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15024436, end_position_on_the_genomic_accession=15042060, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32434 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32434, CurrentID=0, Status='live', Symbol='ND-B18', Aliases='Dmel_CG5548, CG5548, Dmel\\CG5548, NB8M', description='NADH dehydrogenase (ubiquinone) B18 subunit', other_designations='CG5548-PA|CG5548-PB|CG5548-PC|NADH ubiquinone oxidoreductase B18|NADH-ubiquinone oxidoreductase B18 subunit|NADH:ubiquinone oxidoreductase B18 subunit|ND-B18-PA|ND-B18-PB|ND-B18-PC', map_location='13A8-13A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15063733, end_position_on_the_genomic_accession=15064926, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32435 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32435, CurrentID=0, Status='live', Symbol='opm', Aliases='Dmel_CG9053, CG9053, Dmel\\CG9053, opo, p24', description='opossum', other_designations='CG9053-PA|CG9053-PC|opm-PA|opm-PC', map_location='13A8-13A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15064836, end_position_on_the_genomic_accession=15066432, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32436 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32436, CurrentID=0, Status='live', Symbol='dob', Aliases='Dmel_CG5560, CG5560, Dmel\\CG5560, Dob', description='doppelganger von brummer', other_designations='CG5560-PA|dob-PA', map_location='13A8-13A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15067007, end_position_on_the_genomic_accession=15069091, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32437 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32437, CurrentID=0, Status='live', Symbol='Lsd-2', Aliases='Dmel_CG9057, 145098_at, CG9057, DmPLIN2, Dmel\\CG9057, LSD-2, LSD2, LSP-2, Lsd2, PLIN2, plin2', description='Lipid storage droplet-2', other_designations='CG9057-PA|CG9057-PB|CG9057-PD|CG9057-PE|CG9057-PF|Lsd-2-PA|Lsd-2-PB|Lsd-2-PD|Lsd-2-PE|Lsd-2-PF|Perilipin2|lipid storage droplet 2|lipid storage droplet 2 gene|lipid storage droplet protein 2|lipid storage droplet-2|perilipin', map_location='13A8-13A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15069127, end_position_on_the_genomic_accession=15076220, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32438 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32438, CurrentID=0, Status='live', Symbol='CG33177', Aliases='Dmel_ CG5579, Dmel\\CG33177', description='CG33177 gene product from transcript CG33177-RA', other_designations='CG33177-PA|uncharacterized protein', map_location='13A9-13A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15081367, end_position_on_the_genomic_accession=15082369, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32439 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32439, CurrentID=0, Status='live', Symbol='CG9065', Aliases='Dmel_ COXS, Dmel\\CG9065', description='CG9065 gene product from transcript CG9065-RA', other_designations='CG9065-PA|complex IV copper chaperone|cytochrome c oxidase copper chaperone', map_location='13A9-13A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15084029, end_position_on_the_genomic_accession=15084699, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32440 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32440, CurrentID=0, Status='live', Symbol='CG15027', Aliases='Dmel_ BcDNA:SD22530, Dmel\\CG15027', description='CG15027 gene product from transcript CG15027-RA', other_designations='CG15027-PA|uncharacterized protein', map_location='13A9-13A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15085159, end_position_on_the_genomic_accession=15085827, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32441 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32441, CurrentID=0, Status='live', Symbol='CG5599', Aliases='Dmel_ Dmel\\CG5599', description='CG5599 gene product from transcript CG5599-RA', other_designations='CG5599-PA', map_location='13A10-13A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15086125, end_position_on_the_genomic_accession=15088629, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32442 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32442, CurrentID=0, Status='live', Symbol='Rab3-GEF', Aliases='Dmel_CG5627, BEST:GH21304, CG5627, Dmel\\CG5627, anon-EST:Posey246, rab3-GEF', description='Rab3 GDP-GTP exchange factor', other_designations='CG5627-PC|CG5627-PD|CG5627-PE|CG5627-PF|CG5627-PG|CG5627-PH|CG5627-PI|Rab3-GEF-PC|Rab3-GEF-PD|Rab3-GEF-PE|Rab3-GEF-PF|Rab3-GEF-PG|Rab3-GEF-PH|Rab3-GEF-PI|rab3-GEF', map_location='13A11-13A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15089682, end_position_on_the_genomic_accession=15113744, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  32443 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32443, CurrentID=0, Status='live', Symbol='CG9072', Aliases='Dmel_ Dmel\\CG9072', description='CG9072 gene product from transcript CG9072-RA', other_designations='CG9072-PA', map_location='13A11-13A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15095458, end_position_on_the_genomic_accession=15096042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32444 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32444, CurrentID=0, Status='live', Symbol='Cyp4s3', Aliases='Dmel_CG9081, 4s3, CG9081, Dmel\\CG9081', description='CG9081 gene product from transcript CG9081-RA', other_designations='CG9081-PA|Cyp4s3-PA|cytochrome P450', map_location='13A12-13A12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15113302, end_position_on_the_genomic_accession=15116542, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32446 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32446, CurrentID=0, Status='live', Symbol='RpL37a', Aliases='Dmel_CG9091, CG9091, Dmel\\CG9091, RpL37, RpL37A', description='Ribosomal protein L37a', other_designations='CG9091-PA|CG9091-PB|RpL37a-PA|RpL37a-PB|ribosomal protein L37a', map_location='13B1-13B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15137870, end_position_on_the_genomic_accession=15139367, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32447 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32447, CurrentID=0, Status='live', Symbol='CG9095', Aliases='Dmel_ CT26106, Dmel\\CG9095', description='CG9095 gene product from transcript CG9095-RA', other_designations='CG9095-PA|CG9095-PB|CG9095-PC|CG9095-PD', map_location='13B1-13B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15139937, end_position_on_the_genomic_accession=15163225, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32448 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32448, CurrentID=0, Status='live', Symbol='CG43737', Aliases='Dmel_ CG15028, CG15029, Dm_X:77844, Dmel\\CG43737, Dmel_CG15028, Dmel_CG15029', description='CG43737 gene product from transcript CG43737-RB', other_designations='CG43737-PB|CG43737-PC|uncharacterized protein', map_location='13B1-13B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15175441, end_position_on_the_genomic_accession=15189931, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32450 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32450, CurrentID=0, Status='live', Symbol='CG5662', Aliases='Dmel_ Dmel\\CG5662', description='CG5662 gene product from transcript CG5662-RB', other_designations='CG5662-PA|CG5662-PB', map_location='13B3-13B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15210386, end_position_on_the_genomic_accession=15211467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32452 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32452, CurrentID=0, Status='live', Symbol='CG9101', Aliases='Dmel_ Dmel\\CG9101', description='CG9101 gene product from transcript CG9101-RA', other_designations='CG9101-PA', map_location='13B3-13B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15228744, end_position_on_the_genomic_accession=15229875, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32453 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32453, CurrentID=0, Status='live', Symbol='PPYR1', Aliases='Dmel_CG15031, BcDNA:AT19571, CG15031, Dmel\\CG15031', description='Protein phosphatase Y regulator 1', other_designations='CG15031-PA|PPYR1-PA|protein phosphatase Y regulator 1', map_location='13B3-13B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15243859, end_position_on_the_genomic_accession=15244913, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32454 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32454, CurrentID=0, Status='live', Symbol='CG9106', Aliases='Dmel_ Dmel\\CG9106', description='CG9106 gene product from transcript CG9106-RA', other_designations='CG9106-PA|CG9106-PB', map_location='13B4-13B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15271036, end_position_on_the_genomic_accession=15272405, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32455 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32455, CurrentID=0, Status='live', Symbol='CG5877', Aliases='Dmel_ Dmel\\CG5877', description='CG5877 gene product from transcript CG5877-RA', other_designations='CG5877-PA|CG5877-PB|CG5877-PD', map_location='13B4-13B5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15293031, end_position_on_the_genomic_accession=15297169, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32457 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32457, CurrentID=0, Status='live', Symbol='gce', Aliases='Dmel_CG42739, CG15032, CG42739, CG6211, Dmel\\CG42739, Dmel_CG15032, Dmel_CG6211, Dmgce, GCE, Gce/Met2', description='germ cell-expressed bHLH-PAS', other_designations='CG42739-PC|CG42739-PD|CG42739-PE|CG42739-PG|Germ-cell expressed bHLH-PAS|gce-PC|gce-PD|gce-PE|gce-PG|germ cell expressed|germ cell-expressed|germ-cell expressed', map_location='13B6-13B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15297766, end_position_on_the_genomic_accession=15319074, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  32458 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32458, CurrentID=0, Status='live', Symbol='Top1', Aliases='Dmel_CG6146, CG6146, Dmel\\CG6146, TOP1, TopI, Topo I, TopoI, dTopoI, l(1)G0134, l(1)G0201, l(1)G0229, l(1)G0278, l(1)top1, top1, topo I, topoI', description='Topoisomerase 1', other_designations='CG6146-PA|CG6146-PB|CG6146-PC|CG6146-PD|CG6146-PE|Top 1|Top1-PA|Top1-PB|Top1-PC|Top1-PD|Top1-PE|topoisomerase 1|topoisomerase I|topoisomerase1|type I topoisomerase', map_location='13B6-13B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15319479, end_position_on_the_genomic_accession=15328029, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  32459 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32459, CurrentID=0, Status='live', Symbol='dah', Aliases='Dmel_CG6157, Apd, CG6157, Dah, Dmel\\CG6157, fs(1)dah', description='discontinuous actin hexagon', other_designations='CG6157-PA|CG6157-PB|apodystrophin|dah-PA|dah-PB', map_location='13B6-13B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15328342, end_position_on_the_genomic_accession=15331002, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32460 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32460, CurrentID=0, Status='live', Symbol='CG9114', Aliases='Dmel_ Dmel\\CG9114', description='CG9114 gene product from transcript CG9114-RA', other_designations='CG9114-PA', map_location='13B6-13B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15330784, end_position_on_the_genomic_accession=15333154, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32461 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32461, CurrentID=0, Status='live', Symbol='HDAC6', Aliases='Dmel_CG6170, CG6170, DHDAC2, DmHDAC2, Dmel\\CG6170, HDAC, HDAC2, dHDAC2, dHDAC6, dmHDA404, hdac6', description='Histone deacetylase 6', other_designations='CG6170-PA|CG6170-PB|CG6170-PC|CG6170-PD|CG6170-PE|CG6170-PF|CG6170-PG|HDAC6-PA|HDAC6-PB|HDAC6-PC|HDAC6-PD|HDAC6-PE|HDAC6-PF|HDAC6-PG|histone deacetylase 6|histone deacetylase HDA2', map_location='13B6-13B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15333310, end_position_on_the_genomic_accession=15340360, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32462 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32462, CurrentID=0, Status='live', Symbol='CG9123', Aliases='Dmel_ Dmel\\CG9123', description='CG9123 gene product from transcript CG9123-RA', other_designations='CG9123-PA', map_location='13B6-13B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15340373, end_position_on_the_genomic_accession=15342117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32463 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32463, CurrentID=0, Status='live', Symbol='CG12608', Aliases='Dmel_ Dmel\\CG12608', description='CG12608 gene product from transcript CG12608-RA', other_designations='CG12608-PA|uncharacterized protein', map_location='13B7-13B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15342857, end_position_on_the_genomic_accession=15344625, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32464 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32464, CurrentID=0, Status='live', Symbol='CG6227', Aliases='Dmel_ DmRH29, Dmel\\CG6227, Prp5, cg6227', description='CG6227 gene product from transcript CG6227-RA', other_designations='CG6227-PA', map_location='13B7-13B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15345169, end_position_on_the_genomic_accession=15350021, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32467 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32467, CurrentID=0, Status='live', Symbol='CG9164', Aliases='Dmel_ Dmel\\CG9164', description='CG9164 gene product from transcript CG9164-RC', other_designations='CG9164-PB|CG9164-PC|CG9164-PD', map_location='13C1-13C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15380371, end_position_on_the_genomic_accession=15385441, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32468 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32468, CurrentID=0, Status='live', Symbol='Cngl', Aliases='Dmel_CG43395, CG18210, CG43395, CG9176, CNGL, Dmel\\CG43395, Dmel_CG18210, Dmel_CG9176, cngl, dmCNG3, mdcds_55060', description='Cyclic nucleotide-gated ion channel-like', other_designations='CG43395-PD|CG43395-PE|CG43395-PF|CG43395-PI|CG43395-PJ|CNG channel-like|Cngl-PD|Cngl-PE|Cngl-PF|Cngl-PI|Cngl-PJ|cng-like|cyclic nucleotide-gated ion channel-like', map_location='13C1-13C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15386038, end_position_on_the_genomic_accession=15441252, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  32470 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32470, CurrentID=0, Status='live', Symbol='Scamp', Aliases='Dmel_CG9195, BcDNA:GM04934, CG9195, Dmel\\CG9195, SCAMP, Sc, anon-EST:fe1F7, dmSC, dmSCAMP', description='CG9195 gene product from transcript CG9195-RC', other_designations='CG9195-PA|CG9195-PC|CG9195-PD|Scamp-PA|Scamp-PC|Scamp-PD|anon-fast-evolving-1F7|scamp', map_location='13C3-13C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15446179, end_position_on_the_genomic_accession=15449883, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32471 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32471, CurrentID=0, Status='live', Symbol='Ahcy13', Aliases='Dmel_CG11654, CG11654, Dmel\\CG11654, SAHH, ahcY', description='Adenosylhomocysteinase at 13', other_designations='AdoHcyase|Ahcy13-PA|Ahcy13-PB|Ahcy13-PC|CG11654-PA|CG11654-PB|CG11654-PC|S-Adenosyl-L-homocysteine hydrolase|adenosylhomocysteinase at 13', map_location='13C3-13C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15450261, end_position_on_the_genomic_accession=15453167, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32472 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32472, CurrentID=0, Status='live', Symbol='CG11655', Aliases='Dmel_ Dmel\\CG11655', description='CG11655 gene product from transcript CG11655-RA', other_designations='CG11655-PA|uncharacterized protein', map_location='13C3-13C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15453471, end_position_on_the_genomic_accession=15455551, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32473 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32473, CurrentID=0, Status='live', Symbol='shtd', Aliases='Dmel_CG9198, APC1, Apc1, CG9198, Dmel\\CG9198, SHTD, Shtd', description='shattered', other_designations='CG9198-PA|shtd-PA', map_location='1-53.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15455544, end_position_on_the_genomic_accession=15462441, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32474 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32474, CurrentID=0, Status='live', Symbol='CG6294', Aliases='Dmel_ Dmel\\CG6294', description='CG6294 gene product from transcript CG6294-RA', other_designations='CG6294-PA|CG6294-PB', map_location='13C4-13C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15462722, end_position_on_the_genomic_accession=15466812, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32475 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32475, CurrentID=0, Status='live', Symbol='CG6299', Aliases='Dmel_ Dmel\\CG6299', description='CG6299 gene product from transcript CG6299-RB', other_designations='CG6299-PA|CG6299-PB|CG6299-PC|CG6299-PD|CG6299-PE', map_location='13C4-13C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15462722, end_position_on_the_genomic_accession=15466812, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32476 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32476, CurrentID=0, Status='live', Symbol='CG15642', Aliases='Dmel_ Dmel\\CG15642', description='CG15642 gene product from transcript CG15642-RB', other_designations='CG15642-PA|CG15642-PB|uncharacterized protein', map_location='13C4-13C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15466184, end_position_on_the_genomic_accession=15467757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32477 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32477, CurrentID=0, Status='live', Symbol='CG15641', Aliases='Dmel_ Dmel\\CG15641', description='CG15641 gene product from transcript CG15641-RA', other_designations='CG15641-PA|uncharacterized protein', map_location='13C4-13C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15467891, end_position_on_the_genomic_accession=15468303, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32478 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32478, CurrentID=0, Status='live', Symbol='Grip128', Aliases='Dmel_CG9201, CG9201, D Dgrip128, Dgrips128, Dmel\\CG9201, GCP5, grip, grip128', description='CG9201 gene product from transcript CG9201-RB', other_designations='CG9201-PB|DGrip128|Grip128-PB', map_location='13C4-13C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15468281, end_position_on_the_genomic_accession=15472119, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32479 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32479, CurrentID=0, Status='live', Symbol='CG6308', Aliases='Dmel_ Alg14, Dmel\\CG6308, dAlg14', description='CG6308 gene product from transcript CG6308-RA', other_designations='CG6308-PA', map_location='13C4-13C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15472164, end_position_on_the_genomic_accession=15472859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32480 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32480, CurrentID=0, Status='live', Symbol='mh', Aliases='Dmel_CG9203, CG9203, Dmel\\CG9203, fs(1)1182, fs(1)A1182', description='maternal haploid', other_designations='CG9203-PA|fs(1)maternal haploid|mh-PA', map_location='13C4-13C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15472804, end_position_on_the_genomic_accession=15475400, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32481 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32481, CurrentID=0, Status='live', Symbol='CG6324', Aliases='Dmel_ Dmel\\CG6324', description='CG6324 gene product from transcript CG6324-RB', other_designations='CG6324-PA|CG6324-PB', map_location='13C5-13C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15476091, end_position_on_the_genomic_accession=15481722, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32482 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32482, CurrentID=0, Status='live', Symbol='CG6340', Aliases='Dmel_ Dmel\\CG6340', description='CG6340 gene product from transcript CG6340-RE', other_designations='CG6340-PA|CG6340-PB|CG6340-PE|CG6340-PF', map_location='13C5-13C5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15482374, end_position_on_the_genomic_accession=15486612, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32483 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32483, CurrentID=0, Status='live', Symbol='Gmap', Aliases='Dmel_CG33206, CG32587, CG33206, CG6350, CG7821, Dmel\\CG33206, GMAP, GMAP-210, GMAP210, dGMAP, l(1)G0168', description='Golgi microtubule-associated protein', other_designations='CG33206-PA|CG33206-PB|CG33206-PC|Gmap-PA|Gmap-PB|Gmap-PC|golgi microtubule-associated protein|lethal (1) G0168', map_location='13C5-13C7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15486922, end_position_on_the_genomic_accession=15511223, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32485 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32485, CurrentID=0, Status='live', Symbol='Ac13E', Aliases='Dmel_CG9210, AC 13E, AC13E, Ac35C, CG9210, DAC9, Dmel\\CG9210', description='Adenylyl cyclase 35C', other_designations='Ac13E-PA|Ac13E-PB|Ac13E-PC|CG9210-PA|CG9210-PB|CG9210-PC|adenyly cyclase 35C|adenylyl cyclase|adenylyl cyclase 35C', map_location='13C7-13D2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15510877, end_position_on_the_genomic_accession=15554278, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32487 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32487, CurrentID=0, Status='live', Symbol='CG42299', Aliases='Dmel_ CG15644, CG18620, CG32584, Dmel\\CG42299, Dmel_CG32584', description='CG42299 gene product from transcript CG42299-RA', other_designations='CG42299-PA|uncharacterized protein', map_location='13D3-13D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15574947, end_position_on_the_genomic_accession=15575683, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32488 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32488, CurrentID=0, Status='live', Symbol='CG7860', Aliases='Dmel_ Dmel\\CG7860', description='CG7860 gene product from transcript CG7860-RA', other_designations='CG7860-PA|CG7860-PB', map_location='13D3-13D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15568993, end_position_on_the_genomic_accession=15570369, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32489 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32489, CurrentID=0, Status='live', Symbol='CG15643', Aliases='Dmel_ CG32583, Dmel\\CG15643', description='CG15643 gene product from transcript CG15643-RA', other_designations='CG15643-PA|CG15643-PB|uncharacterized protein', map_location='13D3-13D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15570380, end_position_on_the_genomic_accession=15572149, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32490 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32490, CurrentID=0, Status='live', Symbol='CG9213', Aliases='Dmel_ CG32585, Dmel\\CG9213', description='CG9213 gene product from transcript CG9213-RA', other_designations='CG9213-PA|CG9213-PB', map_location='13D3-13D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15572076, end_position_on_the_genomic_accession=15574535, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32492 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32492, CurrentID=0, Status='live', Symbol='cerv', Aliases='Dmel_CG15645, CG15645, Dmel\\CG15645', description='cervantes', other_designations='CG15645-PA|CG15645-PC|CG15645-PD|cerv-PA|cerv-PC|cerv-PD', map_location='13D4-13D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15576986, end_position_on_the_genomic_accession=15578171, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32493 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32493, CurrentID=0, Status='live', Symbol='CG7872', Aliases='Dmel_ Dmel\\CG7872', description='CG7872 gene product from transcript CG7872-RA', other_designations='CG7872-PA', map_location='13D4-13D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15578205, end_position_on_the_genomic_accession=15579857, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32494 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32494, CurrentID=0, Status='live', Symbol='CG9215', Aliases='Dmel_ Dmel\\CG9215', description='CG9215 gene product from transcript CG9215-RA', other_designations='CG9215-PA', map_location='13D4-13D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15580930, end_position_on_the_genomic_accession=15582874, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32495 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32495, CurrentID=0, Status='live', Symbol='CG8097', Aliases='Dmel_ Dmel\\CG8097', description='CG8097 gene product from transcript CG8097-RB', other_designations='CG8097-PA|CG8097-PB', map_location='13D4-13D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15583159, end_position_on_the_genomic_accession=15585249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32496 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32496, CurrentID=0, Status='live', Symbol='CG8105', Aliases='Dmel_ Dmel\\CG8105, FBgn0030661', description='CG8105 gene product from transcript CG8105-RA', other_designations='CG8105-PA', map_location='13D4-13D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15587733, end_position_on_the_genomic_accession=15589365, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32497 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32497, CurrentID=0, Status='live', Symbol='CG9220', Aliases='Dmel_ Dmel\\CG9220, dCHSY', description='CG9220 gene product from transcript CG9220-RC', other_designations='CG9220-PC|CG9220-PD', map_location='13D4-13E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15585289, end_position_on_the_genomic_accession=15593534, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32498 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32498, CurrentID=0, Status='live', Symbol='sog', Aliases='Dmel_CG9224, CG9224, Dm sog, Dmel\\CG9224, SOG, Sog, l(1)G0160, l(1)G0395, l(1)G0479', description='short gastrulation', other_designations='CG9224-PA|CG9224-PB|CG9224-PE|chordin|short of gastrulation|short-gastrulation|sog-PA|sog-PB|sog-PE', map_location='1-53 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15603283, end_position_on_the_genomic_accession=15633434, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32499 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32499, CurrentID=0, Status='live', Symbol='CG8117', Aliases='Dmel_ Dmel\\CG8117, dCrsp70', description='CG8117 gene product from transcript CG8117-RA', other_designations='CG8117-PA', map_location='13E3-13E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15644222, end_position_on_the_genomic_accession=15644991, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32500 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32500, CurrentID=0, Status='live', Symbol='CG8119', Aliases='Dmel_ Dmel\\CG8119', description='CG8119 gene product from transcript CG8119-RA', other_designations='CG8119-PA', map_location='13E3-13E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15662585, end_position_on_the_genomic_accession=15663319, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32501 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32501, CurrentID=0, Status='live', Symbol='CG15646', Aliases='Dmel_ Dmel\\CG15646', description='CG15646 gene product from transcript CG15646-RB', other_designations='CG15646-PA|CG15646-PB|uncharacterized protein', map_location='13E3-13E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15667885, end_position_on_the_genomic_accession=15671029, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32502 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32502, CurrentID=0, Status='live', Symbol='CG12708', Aliases='Dmel_ Dmel\\CG12708', description='CG12708 gene product from transcript CG12708-RB', other_designations='CG12708-PA|CG12708-PB|uncharacterized protein', map_location='13E5-13E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15675476, end_position_on_the_genomic_accession=15677737, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32503 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32503, CurrentID=0, Status='live', Symbol='CG15599', Aliases='Dmel_ Dmel\\CG15599', description='CG15599 gene product from transcript CG15599-RD', other_designations='CG15599-PD|uncharacterized protein', map_location='13E6-13E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15684947, end_position_on_the_genomic_accession=15688857, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32504 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32504, CurrentID=0, Status='live', Symbol='CG8128', Aliases='Dmel_ Dmel\\CG8128', description='CG8128 gene product from transcript CG8128-RA', other_designations='CG8128-PA|CG8128-PB', map_location='13E7-13E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15697494, end_position_on_the_genomic_accession=15700034, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32505 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32505, CurrentID=0, Status='live', Symbol='CG9240', Aliases='Dmel_ Dmel\\CG9240, anon-EST:Posey267', description='CG9240 gene product from transcript CG9240-RA', other_designations='CG9240-PA|CG9240-PB', map_location='13E8-13E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15700172, end_position_on_the_genomic_accession=15701259, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32506 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32506, CurrentID=0, Status='live', Symbol='Pis', Aliases='Dmel_CG9245, CG9245, Dmel\\CG9245, anon-WO02059370.38, dPIS, dpis', description='Phosphatidylinositol synthase', other_designations='CG9245-PA|CG9245-PB|PI synthase|Phosphatidyl-inositol synthase|Pis-PA|Pis-PB|drosophila phosphatidylinositol synthase|phosphatidylinositol synthase', map_location='13E8-13E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15701451, end_position_on_the_genomic_accession=15703770, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32507 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32507, CurrentID=0, Status='live', Symbol='CG8134', Aliases='Dmel_ Dmel\\CG8134', description='CG8134 gene product from transcript CG8134-RB', other_designations='CG8134-PA|CG8134-PB', map_location='13E8-13E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15704125, end_position_on_the_genomic_accession=15705806, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32508 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32508, CurrentID=0, Status='live', Symbol='CG9281', Aliases='Dmel_ BEST:LD38280, Dmel\\CG9281', description='CG9281 gene product from transcript CG9281-RB', other_designations='CG9281-PB|CG9281-PC|CG9281-PD', map_location='13E8-13E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15705806, end_position_on_the_genomic_accession=15710539, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32509 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32509, CurrentID=0, Status='live', Symbol='CG15601', Aliases='Dmel_ Dmel\\CG15601', description='CG15601 gene product from transcript CG15601-RB', other_designations='CG15601-PA|CG15601-PB|CG15601-PC|CG15601-PD|uncharacterized protein', map_location='13E8-13E8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15710647, end_position_on_the_genomic_accession=15712309, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32510 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32510, CurrentID=0, Status='live', Symbol='CG8184', Aliases='Dmel_ Dmel\\CG8184, URE-B1/CG8184', description='CG8184 gene product from transcript CG8184-RD', other_designations='CG8184-PB|CG8184-PC|CG8184-PD|CG8184-PE|CG8184-PF|CG8184-PG|CG8184-PH', map_location='13E8-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15712628, end_position_on_the_genomic_accession=15731935, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  32511 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32511, CurrentID=0, Status='live', Symbol='CG8191', Aliases='Dmel_ Dmel\\CG8191', description='CG8191 gene product from transcript CG8191-RA', other_designations='CG8191-PA', map_location='13E12-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15732178, end_position_on_the_genomic_accession=15733663, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32512 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32512, CurrentID=0, Status='live', Symbol='CG12379', Aliases='Dmel_ Dmel\\CG12379', description='CG12379 gene product from transcript CG12379-RB', other_designations='CG12379-PB|uncharacterized protein', map_location='13E12-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15733222, end_position_on_the_genomic_accession=15733996, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32513 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32513, CurrentID=0, Status='live', Symbol='l(1)G0136', Aliases='Dmel_CG8198, BcDNA:AT26381, CG8198, Dmel\\CG8198', description='lethal (1) G0136', other_designations='CG8198-PA|CG8198-PB|IscA1 homologue|l(1)G0136-PA|l(1)G0136-PB', map_location='13E12-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15734270, end_position_on_the_genomic_accession=15735780, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32514 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32514, CurrentID=0, Status='live', Symbol='Arp6', Aliases='Dmel_CG11678, ARP, ARP(13E), ARP13E, Acrp, Actr13E, Arp13E, Arp4, CG11678, Dmel\\CG11678, actr13E, arp4, dArp6', description='Actin-related protein 6', other_designations='Actin-related protein 13E|Arp6-PA|CG11678-PA|actin-related protein', map_location='13E12-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15735774, end_position_on_the_genomic_accession=15737120, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32515 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32515, CurrentID=0, Status='live', Symbol='CG11679', Aliases='Dmel_ Dmel\\CG11679', description='CG11679 gene product from transcript CG11679-RA', other_designations='CG11679-PA|uncharacterized protein', map_location='13E12-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15737493, end_position_on_the_genomic_accession=15738954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32516 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32516, CurrentID=0, Status='live', Symbol='CG8206', Aliases='Dmel_ Dmel\\CG8206', description='CG8206 gene product from transcript CG8206-RB', other_designations='CG8206-PA|CG8206-PB', map_location='13E12-13E12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15739230, end_position_on_the_genomic_accession=15740417, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32517 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32517, CurrentID=0, Status='live', Symbol='CG8944', Aliases='Dmel_ Dmel\\CG8944', description='CG8944 gene product from transcript CG8944-RB', other_designations='CG8944-PB', map_location='13E12-13E13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15741343, end_position_on_the_genomic_accession=15746231, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32518 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32518, CurrentID=0, Status='live', Symbol='Tcp-1zeta', Aliases='Dmel_CG8231, CCT6, CG8231, Dmel\\CG8231, L(1)g0022, Q9VXQ5_DROME, Rcd3, TCP-1-zeta, TCP-1zeta, Tcp-1-zeta, l(1)G0022, l(1)G0027, l(1)G0057', description='T-cp1zeta', other_designations='CG8231-PA|Tcp-1zeta-PA|lethal (1) G0022|reduction in Cnn dots 3', map_location='13E13-13E14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15746648, end_position_on_the_genomic_accession=15748971, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32520 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32520, CurrentID=0, Status='live', Symbol='CG8239', Aliases='Dmel_ Dmel\\CG8239, MVD, MevPPD', description='CG8239 gene product from transcript CG8239-RB', other_designations='CG8239-PA|CG8239-PB', map_location='13E14-13E14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15749166, end_position_on_the_genomic_accession=15751737, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32521 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32521, CurrentID=0, Status='live', Symbol='CG8260', Aliases='Dmel_ Dmel\\CG8260', description='CG8260 gene product from transcript CG8260-RA', other_designations='CG8260-PA|CG8260-PB', map_location='13E14-13E14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15756318, end_position_on_the_genomic_accession=15757767, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32522 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32522, CurrentID=0, Status='live', Symbol='Graf', Aliases='Dmel_CG8948, CG8948, D-Graf, Dmel\\CG8948, Q9VXQ2_DROME', description='GRAF ortholog (H. sapiens)', other_designations='CG8948-PA|CG8948-PD|CG8948-PE|CG8948-PG|CG8948-PH|CG8948-PI|CG8948-PJ|GRAF ortholog|GTPase regulator associated with focal adhesion kinase ortholog|Graf-PA|Graf-PD|Graf-PE|Graf-PG|Graf-PH|Graf-PI|Graf-PJ', map_location='13E14-13E14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15751674, end_position_on_the_genomic_accession=15760733, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32523 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32523, CurrentID=0, Status='live', Symbol='mRpL3', Aliases='Dmel_CG8288, CG8288, Dmel\\CG8288, L3', description='mitochondrial ribosomal protein L3', other_designations='CG8288-PA|mRpL3-PA', map_location='13E14-13E14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15761248, end_position_on_the_genomic_accession=15762663, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32524 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32524, CurrentID=0, Status='live', Symbol='CG17209', Aliases='Dmel_ Dmel\\CG17209', description='CG17209 gene product from transcript CG17209-RC', other_designations='CG17209-PC|uncharacterized protein', map_location='13E14-13E15', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15762956, end_position_on_the_genomic_accession=15769514, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32525 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32525, CurrentID=0, Status='live', Symbol='CG8952', Aliases='Dmel_ Dmel\\CG8952, SP52', description='CG8952 gene product from transcript CG8952-RA', other_designations='CG8952-PA', map_location='13E14-13E14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15765470, end_position_on_the_genomic_accession=15766445, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32526 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32526, CurrentID=0, Status='live', Symbol='CG33172', Aliases='Dmel_ CG8473, Dmel\\CG33172', description='CG33172 gene product from transcript CG33172-RB', other_designations='CG33172-PB|uncharacterized protein', map_location='13E15-13E16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15769664, end_position_on_the_genomic_accession=15772985, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32528 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32528, CurrentID=0, Status='live', Symbol='Efhc1.1', Aliases='Dmel_CG8959, CG8959, Dmel\\CG8959', description='EFHC1 homologue 1', other_designations='CG8959-PA|Defhc1.1|EFHC1 homolog|Efhc1.1-PA', map_location='13E17-13E18', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15780398, end_position_on_the_genomic_accession=15783710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32529 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32529, CurrentID=0, Status='live', Symbol='Paf-AHalpha', Aliases='Dmel_CG8962, AAC83820, CG8962, Dmaalpha[[PAFAH]], Dmel\\CG8962, PAFAHalpha', description='Platelet-activating factor acetylhydrolase alpha', other_designations='CG8962-PA|CG8962-PB|Paf-AHalpha-PA|Paf-AHalpha-PB|platelet-activating factor acetylhydrolase Ib alpha-subunit homolog|platelet-activating factor acetylhydrolase alpha subunit homolog', map_location='13E18-13E18', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15784009, end_position_on_the_genomic_accession=15785583, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32530 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32530, CurrentID=0, Status='live', Symbol='mRpS30', Aliases='Dmel_CG8470, CG8470, Dmel\\CG8470, MRP-S30, mRSp30', description='mitochondrial ribosomal protein S30', other_designations='CG8470-PA|CG8470-PB|mRpS30-PA|mRpS30-PB', map_location='13E18-13E18', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15786047, end_position_on_the_genomic_accession=15788288, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32531 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32531, CurrentID=0, Status='live', Symbol='Rhp', Aliases='Dmel_CG8497, CG8497, Dmel\\CG8497', description='Rhophilin', other_designations='CG8497-PA|CG8497-PB|Rhp-PA|Rhp-PB|rhophilin', map_location='13E18-13E18', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15788897, end_position_on_the_genomic_accession=15793632, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  32532 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32532, CurrentID=0, Status='live', Symbol='CG8974', Aliases='Dmel_ Dmel\\CG8974', description='CG8974 gene product from transcript CG8974-RD', other_designations='CG8974-PA|CG8974-PB|CG8974-PC|CG8974-PD|CG8974-PE', map_location='13E18-13F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15793721, end_position_on_the_genomic_accession=15796028, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32533 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32533, CurrentID=0, Status='live', Symbol='CG15602', Aliases='Dmel_ Dmel\\CG15602', description='CG15602 gene product from transcript CG15602-RB', other_designations='CG15602-PA|CG15602-PB|uncharacterized protein', map_location='13F1-13F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15799426, end_position_on_the_genomic_accession=15800961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32534 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32534, CurrentID=0, Status='live', Symbol='PGRP-LE', Aliases='Dmel_CG8995, CG8995, Dmel\\CG8995, PGRP, anon-WO0153538.66, pgrp-le', description='Peptidoglycan recognition protein LE', other_designations='CG8995-PA|CG8995-PB|PGRP-LE-PA|PGRP-LE-PB|peptidoglycan recognition protein|peptidoglycan recognition protein LE|peptidoglycan recognition protein-LE', map_location='13F1-13F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15801153, end_position_on_the_genomic_accession=15803880, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32535 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32535, CurrentID=0, Status='live', Symbol='CG8509', Aliases='Dmel_ Dmel\\CG8509', description='CG8509 gene product from transcript CG8509-RA', other_designations='CG8509-PA', map_location='13F1-13F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15809555, end_position_on_the_genomic_accession=15810947, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32536 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32536, CurrentID=0, Status='live', Symbol='sd', Aliases='Dmel_CG8544, CG8544, DROEXO, DROEXO2, Dmel\\CG8544, EP(X)1088, EP1088, SD, Sd, TEAD, anon-EST:Liang-2.14, clone 2.14, l(1)G0239, l(1)G0262, l(1)G0309, l(1)G0315, l(1)G0483, l(1)HF394, l(1)III, sp', description='scalloped', other_designations='CG8544-PA|CG8544-PB|CG8544-PC|CG8544-PD|CG8544-PE|CG8544-PF|CG8544-PG|CG8544-PH|CG8544-PI|CG8544-PJ|CG8544-PK|CG8544-PL|CG8544-PM|CG8544-PN|CG8544-PO|CG8544-PP|CG8544-PQ|CG8544-PR|CG8544-PS|CG8544-PT|CG8544-PU|sd-PA|sd-PB|sd-PC|sd-PD|sd-PE|sd-PF|sd-PG|sd-PH|sd-PI|sd-PJ|sd-PK|sd-PL|sd-PM|sd-PN|sd-PO|sd-PP|sd-PQ|sd-PR|sd-PS|sd-PT|sd-PU|spatula', map_location='1-51.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15804370, end_position_on_the_genomic_accession=15827682, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  32537 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32537, CurrentID=0, Status='live', Symbol='Chc', Aliases='Dmel_CG9012, CG9012, CHC, CLH, Cla, Clh, D-Chc, Dmel\\CG9012, chc, chc1, dCHC, l(1)13Fb, l(1)G0438, l(1)VI', description='Clathrin heavy chain', other_designations='CG9012-PA|CG9012-PB|CG9012-PC|CG9012-PD|CG9012-PE|CG9012-PF|CG9012-PG|Chc-PA|Chc-PB|Chc-PC|Chc-PD|Chc-PE|Chc-PF|Chc-PG|clathrin|clathrin hc|clathrin heavy chain|clathrin heavy-chain', map_location='13F5-13F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15827986, end_position_on_the_genomic_accession=15835490, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32538 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32538, CurrentID=0, Status='live', Symbol='CG8565', Aliases='Dmel_ Dmel\\CG8565, SPRK2', description='CG8565 gene product from transcript CG8565-RA', other_designations='CG8565-PA', map_location='13F9-13F10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15837830, end_position_on_the_genomic_accession=15841160, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32539 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32539, CurrentID=0, Status='live', Symbol='Aats-arg', Aliases='Dmel_CG9020, ArgRS, CG9020, Dmel\\CG9020, RRS', description='Arginyl-tRNA synthetase', other_designations='Aats-arg-PA|Arginyl-tRNA ligase|CG9020-PA', map_location='13F10-13F10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15841769, end_position_on_the_genomic_accession=15844145, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32540 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32540, CurrentID=0, Status='live', Symbol='CG8578', Aliases='Dmel_ Dmel\\CG8578', description='CG8578 gene product from transcript CG8578-RC', other_designations='CG8578-PA|CG8578-PB|CG8578-PC', map_location='13F10-13F12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15844476, end_position_on_the_genomic_accession=15847858, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32541 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32541, CurrentID=0, Status='live', Symbol='UBL3', Aliases='Dmel_CG9038, CG9038, Dmel\\CG9038, HCG-1, MUB, Ubl3', description='CG9038 gene product from transcript CG9038-RC', other_designations='CG9038-PA|CG9038-PB|CG9038-PC|UBL3-PA|UBL3-PB|UBL3-PC|Ubiquitin-like 3', map_location='13F13-13F14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15849731, end_position_on_the_genomic_accession=15853175, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32542 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32542, CurrentID=0, Status='live', Symbol='CG15914', Aliases='Dmel_ Dmel\\CG15914, PIG', description='CG15914 gene product from transcript CG15914-RA', other_designations='CG15914-PA|uncharacterized protein', map_location='13F14-13F14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15853421, end_position_on_the_genomic_accession=15854970, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32543 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32543, CurrentID=0, Status='live', Symbol='Myb', Aliases='Dmel_CG9045, CG9045, D-myb, DMyb, Dm Myb, Dm myb, Dm-Myb, Dm-myb, DmMYB, DmMyb, Dmel\\CG9045, Dmyb, MYB, c-Myb, c-myb, dMyb, dmMYB, dmyb, l(1)XVI, myb, p85', description='Myb oncogene-like', other_designations='CG9045-PA|CG9045-PB|CG9045-PD|CG9045-PE|Myb-PA|Myb-PB|Myb-PD|Myb-PE', map_location='13F14-13F14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15854832, end_position_on_the_genomic_accession=15857823, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32544 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32544, CurrentID=0, Status='live', Symbol='Gbeta13F', Aliases='Dmel_CG10545, CG10545, Dmel\\CG10545, G b13F, G-beta, G-betab, G[[beta13F]], G[[beta]], Gb13F, Gbb, Gbeta, anon-EST:Liang-1.22, clone 1.22, dgbeta, gbeta13F', description='G protein beta-subunit 13F', other_designations='CG10545-PA|CG10545-PB|CG10545-PC|CG10545-PE|CG10545-PF|CG10545-PG|CG10545-PH|G protein beta subunit|G-protein beta 13F|G[[beta]] brain|Gbeta13F-PA|Gbeta13F-PB|Gbeta13F-PC|Gbeta13F-PE|Gbeta13F-PF|Gbeta13F-PG|Gbeta13F-PH', map_location='13F16-13F16', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15860115, end_position_on_the_genomic_accession=15866175, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32545 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32545, CurrentID=0, Status='live', Symbol='Gapdh2', Aliases='Dmel_CG8893, CG8893, Dmel\\CG8893, GA3PDH, GADPH, GAP, GAPDH, GAPDH II, GAPDH2, Gapd, Gapdh, Gapdh-2, Gapdh13F', description='Glyceraldehyde 3 phosphate dehydrogenase 2', other_designations='CG8893-PA|CG8893-PC|G3-P dehydrogenase|GAPDH2|Gapdh2-PA|Gapdh2-PC|glyceraldehyde 3 phosphate dehydrogenase 2|glyceraldehyde 3-phosphate dehydrogenase|glyceraldehyde-3-phosphate dehydrogenase', map_location='13F17-13F17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15868145, end_position_on_the_genomic_accession=15869661, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32546 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32546, CurrentID=0, Status='live', Symbol='CG16952', Aliases='Dmel_ Dmel\\CG16952', description='CG16952 gene product from transcript CG16952-RE', other_designations='CG16952-PA|CG16952-PB|CG16952-PC|CG16952-PD|CG16952-PE|uncharacterized protein', map_location='13F17-13F17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15870183, end_position_on_the_genomic_accession=15875607, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32547 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32547, CurrentID=0, Status='live', Symbol='tay', Aliases='Dmel_CG9056, CG9056, Dmel\\CG9056, XI, dAUTS2, l(1)13Ff, l(1)14Ac, l(1)XI', description='tay bridge', other_designations='CG9056-PA|CG9056-PB|CG9056-PC|CG9056-PD|Tay bridge|lethal (1) 14Ac|tay-PA|tay-PB|tay-PC|tay-PD', map_location='14A1-14B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15876368, end_position_on_the_genomic_accession=15887311, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32548 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32548, CurrentID=0, Status='live', Symbol='MSBP', Aliases='Dmel_CG9066, CG9066, Dmel\\CG9066, dm_MSBP', description='membrane steroid binding protein', other_designations='CG9066-PA|CG9066-PB|CG9066-PC|MSBP-PA|MSBP-PB|MSBP-PC|steroid binding protein', map_location='13F17-13F17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15887567, end_position_on_the_genomic_accession=15890109, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32549 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32549, CurrentID=0, Status='live', Symbol='CG15916', Aliases='Dmel_ BcDNA:RH08992, Dmel\\CG15916', description='CG15916 gene product from transcript CG15916-RA', other_designations='CG15916-PA|uncharacterized protein', map_location='13F17-13F17', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15890589, end_position_on_the_genomic_accession=15891547, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32551 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32551, CurrentID=0, Status='live', Symbol='CycD', Aliases='Dmel_CG9096, CDI3, CG9096, CYCD, Cdi3, Cyc D, DmcycD, Dmel\\CG9096, cdi3, cycD, cyclinD, dCycD', description='Cyclin D', other_designations='CG9096-PD|CG9096-PE|CG9096-PF|CycD-PD|CycD-PE|CycD-PF|Cyclin-dependent kinase interactor 3|cyclin D|cyclinD|dCyclinD', map_location='13F18-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15906822, end_position_on_the_genomic_accession=15909676, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32552 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32552, CurrentID=0, Status='live', Symbol='CG8909', Aliases='Dmel_ CT25574, Dm CG8909, Dmel\\CG8909', description='CG8909 gene product from transcript CG8909-RC', other_designations='CG8909-PB|CG8909-PC|CG8909-PD|yolkless-like', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15910870, end_position_on_the_genomic_accession=15921013, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  32553 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32553, CurrentID=0, Status='live', Symbol='CG8916', Aliases='Dmel_ Dmel\\CG8916', description='CG8916 gene product from transcript CG8916-RB', other_designations='CG8916-PB|CG8916-PC', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15921955, end_position_on_the_genomic_accession=15924558, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32554 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32554, CurrentID=0, Status='live', Symbol='Lcch3', Aliases='Dmel_CG17336, CG17336, CT38330, DmGABAbeta, DmGA_b, Dmel\\CG17336, LCCH3', description='Ligand-gated chloride channel homolog 3', other_designations='CG17336-PC|GABA receptor beta subunit|Lcch3-PC|beta subunit|ligand-gated chloride channel|ligand-gated chloride channel homologue 3', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15925105, end_position_on_the_genomic_accession=15928529, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32556 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32556, CurrentID=0, Status='live', Symbol='Stim', Aliases='Dmel_CG9126, C-Stim, CG9126, D-STIM, D-Stim, Dmel\\CG9126, STIM, dSTIM, dStim, dstim, stim1', description='Stromal interaction molecule', other_designations='CG9126-PA|CG9126-PD|CG9126-PF|CG9126-PG|Stim-PA|Stim-PD|Stim-PF|Stim-PG|drosophila stromal interaction molecule|stromal interaction molecule|stromal interaction molecule 1', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15934495, end_position_on_the_genomic_accession=15940451, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32557 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32557, CurrentID=0, Status='live', Symbol='CG8924', Aliases='Dmel_ Dmel\\CG8924', description='CG8924 gene product from transcript CG8924-RB', other_designations='CG8924-PB', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15941077, end_position_on_the_genomic_accession=15944449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32558 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32558, CurrentID=0, Status='live', Symbol='Rrp47', Aliases='Dmel_CG8928, CG8928, Dmel\\CG8928, d lincRNA.1007', description='CG8928 gene product from transcript CG8928-RB', other_designations='CG8928-PA|CG8928-PB|Rrp47-PA|Rrp47-PB', map_location='14A1-14A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15944897, end_position_on_the_genomic_accession=15946249, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32561 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32561, CurrentID=0, Status='live', Symbol='mmd', Aliases='Dmel_CG42252, CG15603, CG15604, CG42252, CG9163, CT26192, Dmel\\CG42252, Dmel_CG15603, Dmel_CG15604, Dmel_CG9163', description='mind-meld', other_designations='CG42252-PB|CG42252-PD|CG42252-PE|CG42252-PF|CG42252-PG|CG42252-PH|CG42252-PI|CG42252-PJ|Meltrin-like|mind meld|mmd-PB|mmd-PD|mmd-PE|mmd-PF|mmd-PG|mmd-PH|mmd-PI|mmd-PJ', map_location='14A1-14A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15946936, end_position_on_the_genomic_accession=15973159, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  32562 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32562, CurrentID=0, Status='live', Symbol='Or13a', Aliases='Dmel_CG12697, 13F.1, 13a, CG12697, DOR38, Dmel\\CG12697, dOr13a, or13a', description='Odorant receptor 13a', other_designations='CG12697-PA|Or13a-PA|odorant receptor 13a', map_location='14A4-14A4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15974425, end_position_on_the_genomic_accession=15976706, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32563 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32563, CurrentID=0, Status='live', Symbol='CG9170', Aliases='Dmel_ Dmel\\CG9170, anon-WO0140519.214', description='CG9170 gene product from transcript CG9170-RC', other_designations='CG9170-PA|CG9170-PB|CG9170-PC', map_location='14A4-14A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15977967, end_position_on_the_genomic_accession=15982944, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32564 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32564, CurrentID=0, Status='live', Symbol='CG8931', Aliases='Dmel_ Dmel\\CG8931', description='CG8931 gene product from transcript CG8931-RA', other_designations='CG8931-PA', map_location='14A5-14A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15983772, end_position_on_the_genomic_accession=15985405, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32565 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32565, CurrentID=0, Status='live', Symbol='ND-20', Aliases='Dmel_CG9172, 20 kDa, CG9172, Dmel\\CG9172, NUKM', description='NADH dehydrogenase (ubiquinone) 20 kDa subunit', other_designations='CG9172-PA|CG9172-PB|CG9172-PC|FBpp0073949|NADH ubiquinone oxidoreductase 20 kDa|NADH-UBIQUINONE OXIDOREDUCTASE 20 KDA SUBUNIT|NADH:ubiquinone oxidoreductase 20 kDa subunit|ND-20-PA|ND-20-PB|ND-20-PC', map_location='14A5-14A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15985473, end_position_on_the_genomic_accession=15986808, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32566 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32566, CurrentID=0, Status='live', Symbol='eIF5', Aliases='Dmel_CG9177, CG9177, Dmel\\CG9177, EIf5, anon-EST:fe3C6', description='CG9177 gene product from transcript CG9177-RD', other_designations='CG9177-PA|CG9177-PB|CG9177-PC|CG9177-PD|CG9177-PE|CG9177-PF|CG9177-PG|anon-fast-evolving-3C6|eIF5-PA|eIF5-PB|eIF5-PC|eIF5-PD|eIF5-PE|eIF5-PF|eIF5-PG', map_location='14A5-14A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15987008, end_position_on_the_genomic_accession=15991974, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32567 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32567, CurrentID=0, Status='live', Symbol='exd', Aliases='Dmel_CG8933, CG8933, DExd, Dm-EXD, Dmel\\CG8933, Dpbx, EXD, Exd, Pbx1, anon-EST:fe1H3, l(1)IV, lincRNA.S9404, td48', description='extradenticle', other_designations='CG8933-PA|CG8933-PB|CG8933-PD|CG8933-PE|Pbx1-oncogene-like|anon-fast-evolving-1H3|exd-PA|exd-PB|exd-PD|exd-PE', map_location='14A5-14A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15992468, end_position_on_the_genomic_accession=15996509, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32568 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32568, CurrentID=0, Status='live', Symbol='CG8939', Aliases='Dmel_ Dmel\\CG8939, dSpb1', description='CG8939 gene product from transcript CG8939-RA', other_designations='CG8939-PA', map_location='14A5-14A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15997016, end_position_on_the_genomic_accession=15999698, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32569 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32569, CurrentID=0, Status='live', Symbol='vap', Aliases='Dmel_CG9209, CG9209, D-RasGAP, Dmel\\CG9209, RAS-GAP, RasGAP, RasGAP/Vap, RasGap, p120RasGAP, rasGap', description='vacuolar peduncle', other_designations='CG9209-PA|CG9209-PB|CG9209-PC|CG9209-PD|CG9209-PE|CG9209-PF|Ras GTPase-activating protein|vacuolar peduncule|vacuolar pedunculi|vap-PA|vap-PB|vap-PC|vap-PD|vap-PE|vap-PF', map_location='1-54.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=15999951, end_position_on_the_genomic_accession=16005812, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32570 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32570, CurrentID=0, Status='live', Symbol='CG12698', Aliases='Dmel_ Dmel\\CG12698', description='CG12698 gene product from transcript CG12698-RA', other_designations='CG12698-PA|uncharacterized protein', map_location='14A6-14A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16006147, end_position_on_the_genomic_accession=16008721, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32571 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32571, CurrentID=0, Status='live', Symbol='CG12395', Aliases='Dmel_ Dmel\\CG12395', description='CG12395 gene product from transcript CG12395-RA', other_designations='CG12395-PA|uncharacterized protein', map_location='14A6-14A6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16061902, end_position_on_the_genomic_accession=16062898, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32572 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32572, CurrentID=0, Status='live', Symbol='dpr18', Aliases='Dmel_CG14948, CG14948, CT34788, Dmel\\CG14948, Dpr-18, Dpr18', description='defective proboscis extension response 18', other_designations='CG14948-PA|dpr18-PA', map_location='14A6-14A7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16063738, end_position_on_the_genomic_accession=16070304, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32573 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32573, CurrentID=0, Status='live', Symbol='Nipsnap', Aliases='Dmel_CG9212, CG9212, Dmel\\CG9212', description='CG9212 gene product from transcript CG9212-RC', other_designations='CG9212-PA|CG9212-PB|CG9212-PC|CG9212-PF|Nipsnap-PA|Nipsnap-PB|Nipsnap-PC|Nipsnap-PF|nipsnap', map_location='14A7-14A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16070365, end_position_on_the_genomic_accession=16073568, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32574 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32574, CurrentID=0, Status='live', Symbol='Tob', Aliases='Dmel_CG9214, CG9214, Dmel\\CG9214', description='CG9214 gene product from transcript CG9214-RC', other_designations='CG9214-PA|CG9214-PC|CG9214-PE|CG9214-PF|Tob-PA|Tob-PC|Tob-PE|Tob-PF|transducer of ERBB2', map_location='14A8-14A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16074717, end_position_on_the_genomic_accession=16091495, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32575 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32575, CurrentID=0, Status='live', Symbol='CG8958', Aliases='Dmel_ Dmel\\CG8958', description='CG8958 gene product from transcript CG8958-RA', other_designations='CG8958-PA', map_location='14A8-14A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16083232, end_position_on_the_genomic_accession=16085751, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32576 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32576, CurrentID=0, Status='live', Symbol='CG42353', Aliases='Dmel_ CG9216, Dmel\\CG42353, Dmel_CG9216', description='CG42353 gene product from transcript CG42353-RA', other_designations='CG42353-PA|CG42353-PB|uncharacterized protein', map_location='14A9-14A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16093059, end_position_on_the_genomic_accession=16104946, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32579 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32579, CurrentID=0, Status='live', Symbol='disco', Aliases='Dmel_CG9908, CG9908, Disco, Dmel\\CG9908, disc', description='disconnected', other_designations='CG9908-PA|CG9908-PB|disco-PA|disco-PB', map_location='1-53.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16210598, end_position_on_the_genomic_accession=16216750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32580 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32580, CurrentID=0, Status='live', Symbol='CG12507', Aliases='Dmel_ Dmel\\CG12507', description='CG12507 gene product from transcript CG12507-RB', other_designations='CG12507-PB|uncharacterized protein', map_location='14B6-14B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16265242, end_position_on_the_genomic_accession=16266537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32581 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32581, CurrentID=0, Status='live', Symbol='kat80', Aliases='Dmel_CG13956, CG13956, CG9910, Dmel\\CG13956, Kat80, Katanin80', description='katanin 80', other_designations='CG13956-PA|CG13956-PB|CG13956-PC|Katanin80|kat80-PA|kat80-PB|kat80-PC|katanin-80', map_location='14B7-14B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16267007, end_position_on_the_genomic_accession=16276180, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32582 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32582, CurrentID=0, Status='live', Symbol='Mfe2', Aliases='Dmel_CG3415, CG3415, DmMFE-2, Dmel\\CG3415, MFE-2', description='peroxisomal Multifunctional enzyme type 2', other_designations='CG3415-PA|CG3415-PB|Mfe2-PA|Mfe2-PB|multifunctional enzyme type 2|peroxisomal multifunctional enzyme type 2', map_location='14B7-14B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16270951, end_position_on_the_genomic_accession=16273815, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32584 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32584, CurrentID=0, Status='live', Symbol='Prosalpha4', Aliases='Dmel_CG3422, 28.1Pros, 3422, Alpha-4, CG3422, Dmel\\CG3422, PROS-28.1, PROS-Dm28.1, PROSA-28.1, PSA4_DROME, Pros-Dm28.1, Pros28, Pros28.1, a4, alpha-4, alpha4, alpha4_dm, pros28.1', description='Proteasome alpha4 subunit', other_designations='20s proteasome|CG3422-PA|Prosalpha4-PA|alpha-subunit of 20S proteasome|alpha4|proteasome 28kD subunit 1|proteasome alpha4 subunit', map_location='14B7-14B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16276270, end_position_on_the_genomic_accession=16277587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32585 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32585, CurrentID=0, Status='live', Symbol='eas', Aliases='Dmel_CG3525, CG3525, Dmel\\CG3525, EP770, EtnK, PC80, RE10142, RH11, ala', description='easily shocked', other_designations='CG3525-PA|CG3525-PB|CG3525-PC|CG3525-PF|CG3525-PG|CG3525-PH|CG3525-PI|CG3525-PJ|alpha lobes absent|alpha-lobe absent|eas-PA|eas-PB|eas-PC|eas-PF|eas-PG|eas-PH|eas-PI|eas-PJ|easily-shocked|ethanolamine kinase', map_location='1-53.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16278328, end_position_on_the_genomic_accession=16285618, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32586 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32586, CurrentID=0, Status='live', Symbol='UQCR-14', Aliases='Dmel_CG3560, BcDNA:RH44664, CG3560, Dmel\\CG3560, UCR6, Ucrh, Ucrh 14KDa, anon-EST:Posey167', description='Ubiquinol-cytochrome c reductase 14 kDa subunit', other_designations='CG3560-PA|UQCR-14-PA|Ubiquinol-cytochrome C reductase complex 14 kDa protein|Ubiquinol-cytochrome c oxidoreductase 14 kDa|Ubiquinol-cytochrome c reductase 14 kDa protein', map_location='14B8-14B8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16286288, end_position_on_the_genomic_accession=16286988, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32587 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32587, CurrentID=0, Status='live', Symbol='caz', Aliases='Dmel_CG3606, BcDNA:GM09207, CG3606, Caz, Dmel\\CG3606, EWSR1, P19, Pen19, SARFH, Sarfh, TFIID, anon-Pen19, cas, p19, pen19', description='cabeza', other_designations='CG3606-PB|CG3606-PC|CG3606-PD|CG3606-PE|Sarcoma-associated RNA-binding fly homolog|caz-PB|caz-PC|caz-PD|caz-PE', map_location='14B8-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16288810, end_position_on_the_genomic_accession=16293520, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32588 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32588, CurrentID=0, Status='live', Symbol='CG9911', Aliases='Dmel_ CG9912, Dmel\\CG9911, LD41494p', description='CG9911 gene product from transcript CG9911-RD', other_designations='CG9911-PA|CG9911-PB|CG9911-PC|CG9911-PD', map_location='14B9-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16293853, end_position_on_the_genomic_accession=16297718, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32589 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32589, CurrentID=0, Status='live', Symbol='CG3632', Aliases='Dmel_ Dmel\\CG3632', description='CG3632 gene product from transcript CG3632-RI', other_designations='CG3632-PE|CG3632-PF|CG3632-PG|CG3632-PH|CG3632-PI|CG3632-PJ|MYOTUBULARIN-like', map_location='14B9-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16296671, end_position_on_the_genomic_accession=16303308, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  32591 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32591, CurrentID=0, Status='live', Symbol='CG3679', Aliases='Dmel_ BcDNA:LD21675, Dmel\\CG3679', description='CG3679 gene product from transcript CG3679-RB', other_designations='CG3679-PA|CG3679-PB', map_location='14B9-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16303870, end_position_on_the_genomic_accession=16307377, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32592 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32592, CurrentID=0, Status='live', Symbol='CG9914', Aliases='Dmel_ Dmel\\CG9914', description='CG9914 gene product from transcript CG9914-RB', other_designations='CG9914-PA|CG9914-PB|Enoyl-CoA hydratase mitochondrial', map_location='14B9-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16306912, end_position_on_the_genomic_accession=16308667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32593 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32593, CurrentID=0, Status='live', Symbol='CG9915', Aliases='Dmel_ Dmel\\CG9915', description='CG9915 gene product from transcript CG9915-RC', other_designations='CG9915-PB|CG9915-PC', map_location='14B9-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16309817, end_position_on_the_genomic_accession=16313138, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32594 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32594, CurrentID=0, Status='live', Symbol='SWIP', Aliases='Dmel_CG13957, CG13957, Dmel\\CG13957, p116', description='CG13957 gene product from transcript CG13957-RB', other_designations='CG13957-PB|SWIP-PB', map_location='14B9-14B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16313709, end_position_on_the_genomic_accession=16317750, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32595 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32595, CurrentID=0, Status='live', Symbol='Cyp1', Aliases='Dmel_CG9916, BcDNA:GM09255, CG9916, Cpy1, CyP-1, Cyp-1, Dmel\\CG9916, cyp-1, cyp1', description='Cyclophilin 1', other_designations='CG9916-PA|Cyp1-PA|Peptidyl-prolyl cis-trans isomerase|cyclophilin 1|cyclophilin1|peptidylprolyl isomerase F', map_location='14B9-14B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16317673, end_position_on_the_genomic_accession=16320370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32596 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32596, CurrentID=0, Status='live', Symbol='CG9917', Aliases='Dmel_ Dmel\\CG9917, anon-WO0140519.229', description='CG9917 gene product from transcript CG9917-RB', other_designations='CG9917-PA|CG9917-PB', map_location='14B12-14B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16321431, end_position_on_the_genomic_accession=16323231, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32597 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32597, CurrentID=0, Status='live', Symbol='CalpC', Aliases='Dmel_CG3692, CG3692, Dmel\\CG3692', description='Calpain C', other_designations='CG3692-PA|CalpC-PA|calpain C', map_location='14B13-14B13', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16325763, end_position_on_the_genomic_accession=16328523, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32598 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32598, CurrentID=0, Status='live', Symbol='CG9919', Aliases='Dmel_ Dmel\\CG9919', description='CG9919 gene product from transcript CG9919-RB', other_designations='CG9919-PA|CG9919-PB', map_location='14B13-14B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16328437, end_position_on_the_genomic_accession=16330759, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32599 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32599, CurrentID=0, Status='live', Symbol='CG9921', Aliases='Dmel_ Dmel\\CG9921', description='CG9921 gene product from transcript CG9921-RC', other_designations='CG9921-PB|CG9921-PC', map_location='14B14-14B14', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16331281, end_position_on_the_genomic_accession=16333600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32601 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32601, CurrentID=0, Status='live', Symbol='sl', Aliases='Dmel_CG4200, CG4200, DPLCgamma, Dmel\\CG4200, PLC, PLC&ygr;, PLC-gamma, PLC-gamma D, PLC-gammaD, PLCgamma, PLCgamma1, PLCgammaDmel, Plcgamma, p145, plc-gamma, plc-gammad', description='small wing', other_designations='CG4200-PA|phospholipase C gamma|phospholipase cgamma|sl-PA', map_location='1-53.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16350315, end_position_on_the_genomic_accession=16356482, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32602 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32602, CurrentID=0, Status='live', Symbol='U2af50', Aliases='Dmel_CG9998, 9-21, CG9998, DU2AF50, DmU2AF[50], Dmel\\CG9998, U2AF, U2AF 50, U2AF-50, U2AF50, U2af, dU2AF50, dU2AF[50], dUTAf[50], l(1)14Ca, l(1)14Cc, l(1)9-21', description='U2 small nuclear riboprotein auxiliary factor 50', other_designations='CG9998-PA|CG9998-PB|CG9998-PC|CG9998-PD|U2af50-PA|U2af50-PB|U2af50-PC|U2af50-PD', map_location='14B17-14B18', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16356696, end_position_on_the_genomic_accession=16359039, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32603 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32603, CurrentID=0, Status='live', Symbol='nonA', Aliases='Dmel_CG4211, Bj6, Bj6-1, CG4211, Dmel\\CG4211, NONA, NonA, diss, non-A/diss, opm2, p54[nrb], x14, nonA', description='no on or off transient A', other_designations='Bj6-monoclonal-antibody|CG4211-PA|CG4211-PB|CG4211-PC|CG4211-PD|dissonance|no on or off transient-A|no-on transient A|no-on-transient A|no-on-transient-A|non-on-transient-A|nonA-PA|nonA-PB|nonA-PC|nonA-PD|positive spike II group', map_location='14B18-14C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16361569, end_position_on_the_genomic_accession=16371547, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32604 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32604, CurrentID=0, Status='live', Symbol='Fur2', Aliases='Dmel_CG18734, 154521_at, CG18734, CG4235, DMH#11, Dfur2, Dfurin2, Dmel\\CG18734, FUR2, Furin, fur2, furin', description='Furin 2', other_designations='CG18734-PA|CG18734-PB|CG18734-PC|CG18734-PD|CG18734-PE|CG18734-PF|CG18734-PG|CG18734-PH|CG18734-PI|Dfurin2|Fur2-PA|Fur2-PB|Fur2-PC|Fur2-PD|Fur2-PE|Fur2-PF|Fur2-PG|Fur2-PH|Fur2-PI|Furin-like protease|furin 2|furin2', map_location='14C1-14C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16373086, end_position_on_the_genomic_accession=16381782, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  32605 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32605, CurrentID=0, Status='live', Symbol='CG9992', Aliases='Dmel_ Dmel\\CG9992', description='CG9992 gene product from transcript CG9992-RB', other_designations='CG9992-PA|CG9992-PB|CG9992-PC|CG9992-PD|CG9992-PE', map_location='14C2-14C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16381954, end_position_on_the_genomic_accession=16384908, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32606 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32606, CurrentID=0, Status='live', Symbol='CG4239', Aliases='Dmel_ Dmel\\CG4239', description='CG4239 gene product from transcript CG4239-RA', other_designations='CG4239-PA|CG4239-PB|CG4239-PC|CG4239-PD', map_location='14C2-14C2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16385278, end_position_on_the_genomic_accession=16388702, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32607 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32607, CurrentID=0, Status='live', Symbol='TH1', Aliases='Dmel_CG9984, CG9984, Dmel\\CG9984, NELF, NELF-D', description='CG9984 gene product from transcript CG9984-RA', other_designations='CG9984-PA|CG9984-PB|TH1-PA|TH1-PB', map_location='14C2-14C3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16388820, end_position_on_the_genomic_accession=16391438, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32608 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32608, CurrentID=0, Status='live', Symbol='mei-41', Aliases='Dmel_CG4252, 4252, ATR, CG4252, DmATR, Dmel\\CG4252, Dmmei41, MEI41, MEI41/FRP1, Mei-41, Mei-41/DATR, atr, fs(1)M37, mei-195, mei41, meiP41, mus(1)103, mus(1)104, mus-103, mus-104, mus103, mus104', description='meiotic 41', other_designations='ATM-RAD3 related|CG4252-PA|drosophila ATR|mei-41-PA|meiotic-41', map_location='1-54.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16391698, end_position_on_the_genomic_accession=16399850, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32609 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32609, CurrentID=0, Status='live', Symbol='CG9981', Aliases='Dmel_ DmCG9981, Dmel\\CG9981', description='CG9981 gene product from transcript CG9981-RC', other_designations='CG9981-PA|CG9981-PB|CG9981-PC', map_location='14C3-14C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16399782, end_position_on_the_genomic_accession=16405122, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32610 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32610, CurrentID=0, Status='live', Symbol='CG4301', Aliases='Dmel_ DmCG4301, Dmel\\CG4301', description='CG4301 gene product from transcript CG4301-RC', other_designations='CG4301-PA|CG4301-PB|CG4301-PC', map_location='14C4-14C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16405509, end_position_on_the_genomic_accession=16414651, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32611 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32611, CurrentID=0, Status='live', Symbol='Traf-like', Aliases='Dmel_CG4394, CG4394, Dmel\\CG4394, TRAF3, Traf-related, Traf3, dTRAF3', description='TNF-receptor-associated factor-like', other_designations='CG4394-PA|CG4394-PB|CG4394-PC|CG4394-PD|CG4394-PE|CG4394-PF|CG4394-PG|Traf-like-PA|Traf-like-PB|Traf-like-PC|Traf-like-PD|Traf-like-PE|Traf-like-PF|Traf-like-PG|Traf3', map_location='14C4-14C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16415046, end_position_on_the_genomic_accession=16417585, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32612 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32612, CurrentID=0, Status='live', Symbol='AnxB11', Aliases='Dmel_CG9968, ANXB11, Anxb11, BcDNA:GH16395, CG9968, Dmel\\CG9968, annexin', description='Annexin B11', other_designations='AnxB11-PA|AnxB11-PB|AnxB11-PC|AnxB11-PD|AnxB11-PE|AnxB11-PF|AnxB11-PG|CG9968-PA|CG9968-PB|CG9968-PC|CG9968-PD|CG9968-PE|CG9968-PF|CG9968-PG|annexin B11', map_location='14C4-14C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16416958, end_position_on_the_genomic_accession=16422265, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32613 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32613, CurrentID=0, Status='live', Symbol='hang', Aliases='Dmel_CG32575, BcDNA:GM07804, BcDNA:LD22118, CG32575, CG4411, CG4416, Dmel\\CG32575, HANG, LD22118', description='hangover', other_designations='CG32575-PA|CG32575-PC|CG32575-PD|hang-PA|hang-PC|hang-PD', map_location='14C4-14C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16422831, end_position_on_the_genomic_accession=16438454, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  32615 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32615, CurrentID=0, Status='live', Symbol='CG9947', Aliases='Dmel_ Dmel\\CG9947', description='CG9947 gene product from transcript CG9947-RB', other_designations='CG9947-PA|CG9947-PB', map_location='14C4-14C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16438680, end_position_on_the_genomic_accession=16441296, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32616 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32616, CurrentID=0, Status='live', Symbol='rngo', Aliases='Dmel_CG4420, CG4420, Dmel\\CG4420', description='rings lost', other_designations='CG4420-PA|CG4420-PB|rngo-PA|rngo-PB', map_location='14C6-14C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16441433, end_position_on_the_genomic_accession=16445367, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32617 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32617, CurrentID=0, Status='live', Symbol='eIF-2alpha', Aliases='Dmel_CG9946, CG9946, Dmel\\CG9946, EIF-2alpha, Eif2alpha, IF2A_DROME, M(1)14C, M(1)19-153, M(1)8e3-10, deIF2alpha, eIF-2, eIF-2 a, eIF-2 alphaeIF-2alpha, eIF2 alpha, eIF2-alpha, eIF2a, eIF2alpha, eiF-2alpha, elF2alpha, l(1)14Cf, l(1)19-153, eIF-2alpha', description='eukaryotic translation Initiation Factor 2alpha', other_designations='CG9946-PA|CG9946-PB|eIF-2alpha-PA|eIF-2alpha-PB|eukaryotic initiation factor 2alpha|eukaryotic translation initiation factor 2|eukaryotic translation initiation factor 2alpha|minute (1) 14C', map_location='14C6-14C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16445176, end_position_on_the_genomic_accession=16446993, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32619 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32619, CurrentID=0, Status='live', Symbol='para', Aliases='Dmel_CG9907, CG9907, DmNa[[V]], DmNa[[v]], DmNa[[v]]1, DmNav1, Dmel\\CG9907, Ocd, bas, bss, l(1)14Da, l(1)ESHS48, lincRNA.S9469, olfD, par, sbl, sbl-1', description='paralytic', other_designations='CG9907-PA|CG9907-PAA|CG9907-PAB|CG9907-PAC|CG9907-PAD|CG9907-PAE|CG9907-PAF|CG9907-PAG|CG9907-PAH|CG9907-PAI|CG9907-PAJ|CG9907-PAK|CG9907-PAL|CG9907-PAM|CG9907-PAN|CG9907-PAO|CG9907-PAP|CG9907-PAQ|CG9907-PAR|CG9907-PAS|CG9907-PAT|CG9907-PAU|CG9907-PAV|CG9907-PAW|CG9907-PAX|CG9907-PAY|CG9907-PAZ|CG9907-PB|CG9907-PBA|CG9907-PBB|CG9907-PBC|CG9907-PBD|CG9907-PBE|CG9907-PBF|CG9907-PBG|CG9907-PBH|CG9907-PC|CG9907-PD|CG9907-PE|CG9907-PF|CG9907-PG|CG9907-PH|CG9907-PI|CG9907-PJ|CG9907-PK|CG9907-PL|CG9907-PM|CG9907-PN|CG9907-PO|CG9907-PP|CG9907-PQ|CG9907-PR|CG9907-PS|CG9907-PT|CG9907-PU|CG9907-PV|CG9907-PW|CG9907-PX|CG9907-PY|CG9907-PZ|Out cold|Out-cold|bang senseless|bang sensitive|bang-senseless|bangsenseless|para-PA|para-PAA|para-PAB|para-PAC|para-PAD|para-PAE|para-PAF|para-PAG|para-PAH|para-PAI|para-PAJ|para-PAK|para-PAL|para-PAM|para-PAN|para-PAO|para-PAP|para-PAQ|para-PAR|para-PAS|para-PAT|para-PAU|para-PAV|para-PAW|para-PAX|para-PAY|para-PAZ|para-PB|para-PBA|para-PBB|para-PBC|para-PBD|para-PBE|para-PBF|para-PBG|para-PBH|para-PC|para-PD|para-PE|para-PF|para-PG|para-PH|para-PI|para-PJ|para-PK|para-PL|para-PM|para-PN|para-PO|para-PP|para-PQ|para-PR|para-PS|para-PT|para-PU|para-PV|para-PW|para-PX|para-PY|para-PZ|paralyzed|smell-blind|smellblind', map_location='1-55 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16455027, end_position_on_the_genomic_accession=16533096, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  32620 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32620, CurrentID=0, Status='live', Symbol='Cnx14D', Aliases='Dmel_CG9906, CG9906, Cnx, Dmel\\CG9906, Xcbp1', description='Calnexin 14D', other_designations='CG9906-PA|Cnx14D-PA|calnexin|calnexin 14D', map_location='14E1-14E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16522967, end_position_on_the_genomic_accession=16525415, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32621 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32621, CurrentID=0, Status='live', Symbol='CG9903', Aliases='Dmel_ Dmel\\CG9903', description='CG9903 gene product from transcript CG9903-RA', other_designations='CG9903-PA', map_location='14E1-14E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16539153, end_position_on_the_genomic_accession=16540852, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32622 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32622, CurrentID=0, Status='live', Symbol='CG9902', Aliases='Dmel_ Dmel\\CG9902', description='CG9902 gene product from transcript CG9902-RB', other_designations='CG9902-PA|CG9902-PB', map_location='14E1-14E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16542124, end_position_on_the_genomic_accession=16547520, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32623 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32623, CurrentID=0, Status='live', Symbol='Arp2', Aliases='Dmel_CG9901, ARP14D, ARP2, Actr14D, Arp14D, CG9901, Dmel\\CG9901, arp2', description='Actin-related protein 2', other_designations='Actin-related protein 14 Da|Actin-related protein 14D|Arp14D|Arp2-PA|Arp2-PB|Arp2-PC|CG9901-PA|CG9901-PB|CG9901-PC', map_location='14E1-14E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16548290, end_position_on_the_genomic_accession=16553968, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32624 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32624, CurrentID=0, Status='live', Symbol='Pp2B-14D', Aliases='Dmel_CG9842, CG9842, CanA14D, CnA, CnnA14D, D33, DmPp2B-14D, Dmel\\CG9842, PP2B, PP2B 14D, PP2B-14D, Pp12-14D/CaN, Pp2B-14DF, PpD33, canA, Pp2B-14D', description='Protein phosphatase 2B at 14D', other_designations='CG9842-PA|CG9842-PB|CG9842-PC|Pp2B-14D-PA|Pp2B-14D-PB|Pp2B-14D-PC|calcineurin|protein phosphatase 2B at 14D|protein phosphatase D33|protein phosphatase from PCR fragment D33', map_location='14E1-14E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16554627, end_position_on_the_genomic_accession=16560016, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32625 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32625, CurrentID=0, Status='live', Symbol='P5CDh2', Aliases='Dmel_CG33092, CG33092, CG6670, DmP5CDh2, Dmel\\CG33092', description='delta-1-Pyrroline-5-carboxylate dehydrogenase 2', other_designations='CG33092-PA|CG33092-PB|CG33092-PC|CG33092-PD|CG33092-PE|CG33092-PG|P5CDh2-PA|P5CDh2-PB|P5CDh2-PC|P5CDh2-PD|P5CDh2-PE|P5CDh2-PG', map_location='93F14-93F14', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=21877936, end_position_on_the_genomic_accession=21880696, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32626 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32626, CurrentID=0, Status='live', Symbol='CG13014', Aliases='Dmel_ Dmel\\CG13014', description='CG13014 gene product from transcript CG13014-RA', other_designations='CG13014-PA|uncharacterized protein', map_location='14F1-14F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16587873, end_position_on_the_genomic_accession=16588930, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32627 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32627, CurrentID=0, Status='live', Symbol='SMC3', Aliases='Dmel_CG9802, CAP, CG9802, Cap, DmSMC3, DmSMC3/Cap, Dmel\\CG9802, Smc3, cap, dCAP, dSMC3', description='Structural maintenance of chromosomes 3', other_designations='CG9802-PA|Chromosome-associated protein|SMC3-PA|cohesin subunit SMC3|structural maintenance of chromosomes 3', map_location='14F1-14F1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16588996, end_position_on_the_genomic_accession=16593439, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32629 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32629, CurrentID=0, Status='live', Symbol='CG9784', Aliases='Dmel_ Dmel\\CG9784, IPP, lincRNA.S9473', description='CG9784 gene product from transcript CG9784-RA', other_designations='CG9784-PA|CG9784-PB|CG9784-PC', map_location='14F2-14F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16595136, end_position_on_the_genomic_accession=16599215, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32630 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32630, CurrentID=0, Status='live', Symbol='Nup153', Aliases='Dmel_CG4453, CG4453, Dmel\\CG4453, NUP153, Nup 153, dNup153, dmNup153, nup153', description='Nucleoporin 153kD', other_designations='CG4453-PA|CG4453-PC|CG4453-PD|Nup153-PA|Nup153-PC|Nup153-PD|nucleoporin 153|nucleoporin 153kD', map_location='14F2-14F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16599241, end_position_on_the_genomic_accession=16610459, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32631 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32631, CurrentID=0, Status='live', Symbol='mbt', Aliases='Dmel_CG18582, CG18582, DmMBT, DmPAK2, Dmel\\CG18582, Mbt, PAK, Pak2, STE20, Ste20, anon-EST:fe1G7, dPAK2', description='mushroom bodies tiny', other_designations='CG18582-PA|anon-fast-evolving-1G7|mbt-PA', map_location='14F2-14F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16610685, end_position_on_the_genomic_accession=16614856, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32632 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32632, CurrentID=0, Status='live', Symbol='CG9782', Aliases='Dmel_ Dmel\\CG9782', description='CG9782 gene product from transcript CG9782-RA', other_designations='CG9782-PA', map_location='14F2-14F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16615028, end_position_on_the_genomic_accession=16621589, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32633 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32633, CurrentID=0, Status='live', Symbol='CG9777', Aliases='Dmel_ Dmel\\CG9777', description='CG9777 gene product from transcript CG9777-RA', other_designations='CG9777-PA', map_location='14F2-14F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16630673, end_position_on_the_genomic_accession=16632235, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32635 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32635, CurrentID=0, Status='live', Symbol='RpS19a', Aliases='Dmel_CG4464, BcDNA:GM26647, CG4464, Dmel\\CG4464, Rp S19a, RpS19, S19, S19a', description='Ribosomal protein S19a', other_designations='CG4464-PA|CG4464-PB|CG4464-PC|CG4464-PD|CG4464-PE|RpS19a-PA|RpS19a-PB|RpS19a-PC|RpS19a-PD|RpS19a-PE|ribosomal protein S19|ribosomal protein S19a', map_location='14F4-14F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16637894, end_position_on_the_genomic_accession=16639249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32636 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32636, CurrentID=0, Status='live', Symbol='SNRPG', Aliases='Dmel_CG9742, BcDNA:RE61341, BcDNA:RH35475, CG9742, Dmel\\CG9742, G, G-Sm, HD10913, SmG, dG, snRNPG', description='Small nuclear ribonucleoprotein G', other_designations='CG9742-PA|CG9742-PB|SNRPG-PA|SNRPG-PB|small nuclear ribonucleoprotein G|small ribonucleoprotein G|small ribonucleoprotein particle protein SmG|snRNPG', map_location='14F4-14F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16639810, end_position_on_the_genomic_accession=16640419, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32637 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32637, CurrentID=0, Status='live', Symbol='mthl1', Aliases='Dmel_CG4521, CG4521, Dmel\\CG4521, Mth-1, anon-WO0170980.22, anon-WO0170980.23, mist', description='methuselah-like 1', other_designations='CG4521-PA|CG4521-PB|mesoderm-invagination signal transducer|mth-like 1|mthl1-PA|mthl1-PB', map_location='14F4-14F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16640968, end_position_on_the_genomic_accession=16650309, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32638 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32638, CurrentID=0, Status='live', Symbol='CG32573', Aliases='Dmel_ CG13013, CG4524, Dmel\\CG32573, Dromel_CG32573_FBtr0074315_mORF', description='CG32573 gene product from transcript CG32573-RA', other_designations='CG32573-PA|CG32573-PC|uncharacterized protein', map_location='14F5-14F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16651376, end_position_on_the_genomic_accession=16653275, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32639 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32639, CurrentID=0, Status='live', Symbol='CG9723', Aliases='Dmel_ Dmel\\CG9723', description='CG9723 gene product from transcript CG9723-RA', other_designations='CG9723-PA', map_location='14F5-14F5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16653220, end_position_on_the_genomic_accession=16655015, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32640 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32640, CurrentID=0, Status='live', Symbol='r', Aliases='Dmel_CG18572, ACT, CAD, CG18572, CG4601, CPS, DHO, DRORUD, Dmel\\CG18572, GAT, PYR1, Su(b), csp2, fs(1)829, fs(1)M34, l(1)Ab', description='rudimentary', other_designations='CG18572-PA|CG18572-PD|CG18572-PE|Suppressor-of-black|carbamoylphosphate synthetase|r-PA|r-PD|r-PE|rudimentary gene|suppressor of black', map_location='14F5-15A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16655359, end_position_on_the_genomic_accession=16668712, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32641 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32641, CurrentID=0, Status='live', Symbol='CG15865', Aliases='Dmel_ Dmel\\CG15865, anon-15Ab', description='CG15865 gene product from transcript CG15865-RA', other_designations='CG15865-PA|uncharacterized protein', map_location='14F6-14F6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16660898, end_position_on_the_genomic_accession=16664455, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32642 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32642, CurrentID=0, Status='live', Symbol='CG13012', Aliases='Dmel_ Dmel\\CG13012', description='CG13012 gene product from transcript CG13012-RE', other_designations='CG13012-PB|CG13012-PC|CG13012-PD|CG13012-PE|uncharacterized protein', map_location='15A1-15A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16668818, end_position_on_the_genomic_accession=16676753, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32643 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32643, CurrentID=0, Status='live', Symbol='CG13010', Aliases='Dmel_ Dmel\\CG13010', description='CG13010 gene product from transcript CG13010-RA', other_designations='CG13010-PA|uncharacterized protein', map_location='15A1-15A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16678265, end_position_on_the_genomic_accession=16678945, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32644 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32644, CurrentID=0, Status='live', Symbol='sing', Aliases='Dmel_CG13011, CG13011, Dmel\\CG13011les-bar, sing', description='singles bar', other_designations='CG13011-PA|sing-PA', map_location='15A-15A', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16679297, end_position_on_the_genomic_accession=16681784, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32645 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32645, CurrentID=0, Status='live', Symbol='Axs', Aliases='Dmel_CG9703, CG9703, Dmel\\CG9703, axs', description='Abnormal X segregation', other_designations='Axs-PA|CG9703-PA|aberrant X segregation|abnormal X segregation', map_location='1-56 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16680089, end_position_on_the_genomic_accession=16683895, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32646 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32646, CurrentID=0, Status='live', Symbol='Sep4', Aliases='Dmel_CG9699, CG9699, Dmel\\CG9699, sep4', description='Septin 4', other_designations='CG9699-PA|CG9699-PB|CG9699-PC|CG9699-PD|CG9699-PE|CG9699-PF|CG9699-PG|Sep4-PA|Sep4-PB|Sep4-PC|Sep4-PD|Sep4-PE|Sep4-PF|Sep4-PG|septin|septin 4', map_location='15A1-15A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16684216, end_position_on_the_genomic_accession=16695173, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32647 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32647, CurrentID=0, Status='live', Symbol='CG9676', Aliases='Dmel_ Dmel\\CG9676, SP99', description='CG9676 gene product from transcript CG9676-RA', other_designations='CG9676-PA', map_location='15A1-15A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16699152, end_position_on_the_genomic_accession=16700170, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32648 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32648, CurrentID=0, Status='live', Symbol='spheroide', Aliases='Dmel_CG9675, CG9675, Dmel\\CG9675, SPH114, Spheroide, anon-EST:ParkEST153, anon-WO0140519.154', description='CG9675 gene product from transcript CG9675-RA', other_designations='CG9675-PA|spheroide-PA', map_location='15A1-15A2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16700446, end_position_on_the_genomic_accession=16701972, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32649 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32649, CurrentID=0, Status='live', Symbol='CG9673', Aliases='Dmel_ Dmel\\CG9673, SPH81', description='CG9673 gene product from transcript CG9673-RB', other_designations='CG9673-PA|CG9673-PB', map_location='15A2-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16702786, end_position_on_the_genomic_accession=16703884, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32650 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32650, CurrentID=0, Status='live', Symbol='CG4653', Aliases='Dmel_ Dmel\\CG4653, SPH179', description='CG4653 gene product from transcript CG4653-RA', other_designations='CG4653-PA', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16704081, end_position_on_the_genomic_accession=16705278, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32651 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32651, CurrentID=0, Status='live', Symbol='CG9672', Aliases='Dmel_ BcDNA:GH16384, Dmel\\CG9672, SPH161', description='CG9672 gene product from transcript CG9672-RA', other_designations='CG9672-PA', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16705971, end_position_on_the_genomic_accession=16706874, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32652 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32652, CurrentID=0, Status='live', Symbol='CG4678', Aliases='Dmel_ Dmel\\CG4678', description='CG4678 gene product from transcript CG4678-RD', other_designations='CG4678-PD|CG4678-PE|CG4678-PF|CG4678-PG|CG4678-PI|CG4678-PJ|CG4678-PK|CG4678-PL', map_location='15A1-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16696131, end_position_on_the_genomic_accession=16711145, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  32654 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32654, CurrentID=0, Status='live', Symbol='CG13008', Aliases='Dmel_ CG13009, Dmel\\CG13008', description='CG13008 gene product from transcript CG13008-RA', other_designations='CG13008-PA|uncharacterized protein', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16711303, end_position_on_the_genomic_accession=16714372, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32656 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32656, CurrentID=0, Status='live', Symbol='CG34325', Aliases='Dmel_ CG18358, Dmel\\CG34325', description='CG34325 gene product from transcript CG34325-RC', other_designations='CG34325-PA|CG34325-PB|CG34325-PC|uncharacterized protein', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16714449, end_position_on_the_genomic_accession=16721411, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32657 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32657, CurrentID=0, Status='live', Symbol='CG32572', Aliases='Dmel_ CG13006, CG13007, Dmel\\CG32572', description='CG32572 gene product from transcript CG32572-RA', other_designations='CG32572-PA|uncharacterized protein', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16721183, end_position_on_the_genomic_accession=16726948, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32659 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32659, CurrentID=0, Status='live', Symbol='CG32568', Aliases='Dmel_ CG4724, Dmel\\CG32568', description='CG32568 gene product from transcript CG32568-RC', other_designations='CG32568-PB|CG32568-PC|uncharacterized protein', map_location='15A3-15A3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16740906, end_position_on_the_genomic_accession=16742918, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32660 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32660, CurrentID=0, Status='live', Symbol='goe', Aliases='Dmel_CG9634, BcDNA:LD21405, CG9634, Dmel\\CG9634', description='gone early', other_designations='CG9634-PA|CG9634-PB|CG9634-PC|goe-PA|goe-PB|goe-PC', map_location='15A3-15A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16727304, end_position_on_the_genomic_accession=16751910, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  32661 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32661, CurrentID=0, Status='live', Symbol='if', Aliases='Dmel_CG9623, CG9623, CT27194, Dmel\\CG9623, If, ItgaPS2, PS 2, PS2, PS2(c), PS2(m8), PS2(m8)alpha, PS2alpha, PSalpha2, aPS2, alpha2Int, alphaPS2, alphaPS2 integrin, alphaPS2C, alphaPS2m8, alpha[[PS2(ms8)]], alpha[[PS2]], in, inf', description='inflated', other_designations='CG9623-PB|CG9623-PC|CG9623-PD|CG9623-PF|alpha integrin|alpha-PS2-integrin|alpha-integrin|alpha2 integrin|alphaPS2|alphaPS2 integrin|alphaPS2-integrin|alpha[[PS2]] integrin|if-PB|if-PC|if-PD|if-PF|integrin alphaPS2 subunit|position-specific antigen 2', map_location='1-55 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16752189, end_position_on_the_genomic_accession=16783434, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  32662 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32662, CurrentID=0, Status='live', Symbol='mRpL22', Aliases='Dmel_CG4742, CG4742, Dmel\\CG4742, L22', description='mitochondrial ribosomal protein L22', other_designations='CG4742-PA|mRpL22-PA|ribosomal protein L22', map_location='15A7-15A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16783933, end_position_on_the_genomic_accession=16785092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32663 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32663, CurrentID=0, Status='live', Symbol='CG9609', Aliases='Dmel_ Dmel\\CG9609', description='CG9609 gene product from transcript CG9609-RA', other_designations='CG9609-PA', map_location='15A8-15A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16785115, end_position_on_the_genomic_accession=16787008, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32664 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32664, CurrentID=0, Status='live', Symbol='Sap30', Aliases='Dmel_CG4756, CG4756, Dmel\\CG4756, SAP30, cg4756', description='SIN3-associated polypeptide 30', other_designations='CG4756-PA|CG4756-PB|Sap30-PA|Sap30-PB', map_location='15A8-15A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16787283, end_position_on_the_genomic_accession=16789644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32665 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32665, CurrentID=0, Status='live', Symbol='Rrp45', Aliases='Dmel_CG9606, CG9606, Dmel\\CG9606, d rrp45', description='CG9606 gene product from transcript CG9606-RB', other_designations='CG9606-PA|CG9606-PB|Rrp45-PA|Rrp45-PB', map_location='15A8-15A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16789405, end_position_on_the_genomic_accession=16791266, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32666 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32666, CurrentID=0, Status='live', Symbol='CG4768', Aliases='Dmel_ Dmel\\CG4768', description='CG4768 gene product from transcript CG4768-RC', other_designations='CG4768-PA|CG4768-PB|CG4768-PC', map_location='15A8-15A8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16792119, end_position_on_the_genomic_accession=16797742, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32668 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32668, CurrentID=0, Status='live', Symbol='CG4789', Aliases='Dmel_ Dmel\\CG4789, Q9VXA9', description='CG4789 gene product from transcript CG4789-RA', other_designations='CG4789-PA', map_location='15A9-15A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16800930, end_position_on_the_genomic_accession=16802000, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32670 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32670, CurrentID=0, Status='live', Symbol='CG13005', Aliases='Dmel_ Dmel\\CG13005', description='CG13005 gene product from transcript CG13005-RA', other_designations='CG13005-PA|uncharacterized protein', map_location='15A9-15A9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16803650, end_position_on_the_genomic_accession=16805754, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32671 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32671, CurrentID=0, Status='live', Symbol='ppk28', Aliases='Dmel_CG4805, CG4805, Dmel\\CG4805, PPK28', description='pickpocket 28', other_designations='CG4805-PA|ppk28-PA', map_location='15A9-15A10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16806068, end_position_on_the_genomic_accession=16808515, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32672 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32672, CurrentID=0, Status='live', Symbol='CG4829', Aliases='Dmel_ Dmel\\CG4829', description='CG4829 gene product from transcript CG4829-RB', other_designations='CG4829-PA|CG4829-PB|CG4829-PC|CG4829-PD|CG4829-PE|CG4829-PF', map_location='15A10-15A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16811699, end_position_on_the_genomic_accession=16826494, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32673 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32673, CurrentID=0, Status='live', Symbol='CG13004', Aliases='Dmel_ Dmel\\CG13004', description='CG13004 gene product from transcript CG13004-RB', other_designations='CG13004-PA|CG13004-PB|CG13004-PC|CG13004-PD|uncharacterized protein', map_location='15A11-15A11', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16826681, end_position_on_the_genomic_accession=16835520, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32674 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32674, CurrentID=0, Status='live', Symbol='RSG7', Aliases='Dmel_CG9108, CG9108, Dmel\\CG9108, RGS7, dRGS7', description='Regulator of G-protein signalling 7', other_designations='CG9108-PA|CG9108-PB|RSG7-PA|RSG7-PB|regulator of G-protein signalling 7|regulator of g protein signaling 7', map_location='15B1-15B1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16838173, end_position_on_the_genomic_accession=16844433, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32675 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32675, CurrentID=0, Status='live', Symbol='CG13003', Aliases='Dmel_ Dmel\\CG13003', description='CG13003 gene product from transcript CG13003-RB', other_designations='CG13003-PA|CG13003-PB|CG13003-PC|uncharacterized protein', map_location='15B1-15B3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16844807, end_position_on_the_genomic_accession=16876544, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32676, CurrentID=0, Status='live', Symbol='CG4872', Aliases='Dmel_ Dmel\\CG4872', description='CG4872 gene product from transcript CG4872-RA', other_designations='CG4872-PA', map_location='15B4-15B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16881347, end_position_on_the_genomic_accession=16882790, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32677 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32677, CurrentID=0, Status='live', Symbol='Nprl2', Aliases='Dmel_CG9104, CG9104, Dmel\\CG9104, NPR2, npr2, nprl2', description='Nitrogen permease regulator-like 2', other_designations='CG9104-PA|Nprl2-PA|nitrogen permease regulator-like 2', map_location='15B4-15B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16882509, end_position_on_the_genomic_accession=16884365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32678 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32678, CurrentID=0, Status='live', Symbol='Rcp', Aliases='Dmel_CG4875, CG4875, Dmel\\CG4875, dRCP', description='Receptor component protein', other_designations='CG4875-PB|CG4875-PC|Rcp-PB|Rcp-PC|receptor component protein', map_location='15B4-15B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16884581, end_position_on_the_genomic_accession=16886612, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32679 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32679, CurrentID=0, Status='live', Symbol='DENR', Aliases='Dmel_CG9099, CG9099, Dmel\\CG9099', description='Density regulated protein', other_designations='CG9099-PA|CG9099-PB|DENR-PA|DENR-PB|density regulated protein', map_location='15B4-15B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16885546, end_position_on_the_genomic_accession=16886560, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32680 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32680, CurrentID=0, Status='live', Symbol='CG4880', Aliases='Dmel_ Dmel\\CG4880', description='CG4880 gene product from transcript CG4880-RA', other_designations='CG4880-PA|CG4880-PB', map_location='15B4-15B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16886818, end_position_on_the_genomic_accession=16888729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  43676 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=43676, CurrentID=0, Status='live', Symbol='CG11317', Aliases='Dmel_ Dmel\\CG11317', description='CG11317 gene product from transcript CG11317-RA', other_designations='CG11317-PA|CG11317-PB|uncharacterized protein', map_location='100B2-100B2', chromosome='3R', genomic_nucleotide_accession_version='NT_033777.3', start_position_on_the_genomic_accession=31090712, end_position_on_the_genomic_accession=31097321, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32682 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32682, CurrentID=0, Status='live', Symbol='CG4928', Aliases='Dmel_ BcDNA:GH10120, Dmel\\CG4928', description='CG4928 gene product from transcript CG4928-RA', other_designations='CG4928-PA|CG4928-PB|CG4928-PC', map_location='15C1-15C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16913180, end_position_on_the_genomic_accession=16938209, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32683 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32683, CurrentID=0, Status='live', Symbol='wus', Aliases='Dmel_CG9089, CG9089, Dmel\\CG9089, l(1)G0162', description='wurst', other_designations='CG9089-PA|CG9089-PB|wus-PA|wus-PB', map_location='15C4-15C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16938244, end_position_on_the_genomic_accession=16940873, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32684 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32684, CurrentID=0, Status='live', Symbol='CG13001', Aliases='Dmel_ Dmel\\CG13001', description='CG13001 gene product from transcript CG13001-RA', other_designations='CG13001-PA|uncharacterized protein', map_location='15C4-15C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16941086, end_position_on_the_genomic_accession=16942871, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32685 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32685, CurrentID=0, Status='live', Symbol='CG13000', Aliases='Dmel_ Dmel\\CG13000', description='CG13000 gene product from transcript CG13000-RA', other_designations='CG13000-PA|uncharacterized protein', map_location='15C4-15C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16947964, end_position_on_the_genomic_accession=16948471, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32686 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32686, CurrentID=0, Status='live', Symbol='RhoGAP15B', Aliases='Dmel_CG4937, CG4937, Dmel\\CG4937, gap20', description='Rho GTPase activating protein at 15B', other_designations='CG4937-PB|CG4937-PC|RhoGAP15B-PB|RhoGAP15B-PC', map_location='15C4-15C4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16943290, end_position_on_the_genomic_accession=16956234, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  32687 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32687, CurrentID=0, Status='live', Symbol='Ubr1', Aliases='Dmel_CG9086, CG9086, Dmel\\CG9086, ubr1', description='Ubr1 ubiquitin ligase', other_designations='CG9086-PA|CG9086-PB|CG9086-PC|Ubr1-PA|Ubr1-PB|Ubr1-PC', map_location='15C5-15C6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16956232, end_position_on_the_genomic_accession=16968679, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  32688 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32688, CurrentID=0, Status='live', Symbol='CG45002', Aliases='Dmel_ BcDNA:GH06235, CG34326, CG9059, Dmel\\CG45002, Dmel_CG34326, Dmel_CG9059', description='CG45002 gene product from transcript CG45002-RI', other_designations='CG45002-PD|CG45002-PE|CG45002-PF|CG45002-PG|CG45002-PH|CG45002-PI|CG45002-PJ|uncharacterized protein', map_location='15D4-15E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=16969069, end_position_on_the_genomic_accession=17066732, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  32689 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32689, CurrentID=0, Status='live', Symbol='CG43996', Aliases='Dmel_ CG12433, Dmel\\CG43996', description='CG43996 gene product from transcript CG43996-RA', other_designations='CG43996-PA|uncharacterized protein', map_location='15D3-15D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17022304, end_position_on_the_genomic_accession=17023370, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32690 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32690, CurrentID=0, Status='live', Symbol='CG8949', Aliases='Dmel_ BP1032, CG34014, Dmel\\CG8949', description='CG8949 gene product from transcript CG8949-RD', other_designations='CG8949-PA|CG8949-PC|CG8949-PD|CG8949-PE|CG8949-PF|CG8949-PG', map_location='15E1-15E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17068717, end_position_on_the_genomic_accession=17077897, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  32691 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32691, CurrentID=0, Status='live', Symbol='CG4949', Aliases='Dmel_ Dmel\\CG4949', description='CG4949 gene product from transcript CG4949-RA', other_designations='CG4949-PA|CG4949-PC', map_location='15E1-15E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17078836, end_position_on_the_genomic_accession=17081237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32692 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32692, CurrentID=0, Status='live', Symbol='CG4955', Aliases='Dmel_ Dmel\\CG4955, anon-WO0140519.97', description='CG4955 gene product from transcript CG4955-RA', other_designations='CG4955-PA', map_location='15E1-15E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17081447, end_position_on_the_genomic_accession=17083356, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32693 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32693, CurrentID=0, Status='live', Symbol='CG8945', Aliases='Dmel_ Cg8945, Dmel\\CG8945', description='CG8945 gene product from transcript CG8945-RC', other_designations='CG8945-PC', map_location='15E1-15E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17080423, end_position_on_the_genomic_accession=17087675, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32694 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32694, CurrentID=0, Status='live', Symbol='CG16700', Aliases='Dmel_ Dmel\\CG16700', description='CG16700 gene product from transcript CG16700-RA', other_designations='CG16700-PA|uncharacterized protein', map_location='15E2-15E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17091854, end_position_on_the_genomic_accession=17098488, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32695 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32695, CurrentID=0, Status='live', Symbol='CG4991', Aliases='Dmel_ CT15971, Dmel\\CG4991', description='CG4991 gene product from transcript CG4991-RF', other_designations='CG4991-PA|CG4991-PB|CG4991-PD|CG4991-PE|CG4991-PF', map_location='15E3-15E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17091854, end_position_on_the_genomic_accession=17104562, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32696 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32696, CurrentID=0, Status='live', Symbol='Arpc3B', Aliases='Dmel_CG8936, Arpc3, Arpc3b, BcDNA:SD24339, CG8936, Dmel\\CG8936, P21-ARC', description='Actin-related protein 2/3 complex, subunit 3B', other_designations='Arpc3B-PA|Arpc3B-PB|Arpc3B-PC|CG8936-PA|CG8936-PB|CG8936-PC', map_location='15E3-15E3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17104571, end_position_on_the_genomic_accession=17105852, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  38187 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38187, CurrentID=0, Status='live', Symbol='CG12104', Aliases='Dmel_ Dmel\\CG12104', description='CG12104 gene product from transcript CG12104-RB', other_designations='CG12104-PA|CG12104-PB|CG12104-PC|uncharacterized protein', map_location='62A7-62A7', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1600726, end_position_on_the_genomic_accession=1602674, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32698 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32698, CurrentID=0, Status='live', Symbol='CG5004', Aliases='Dmel_ Dmel\\CG5004, anon-15Db', description='CG5004 gene product from transcript CG5004-RA', other_designations='CG5004-PA|CG5004-PB|CG5004-PC|CG5004-PD|CG5004-PE|CG5004-PF|anon-15Db|transcript b', map_location='15E3-15E5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17107582, end_position_on_the_genomic_accession=17135988, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  32700 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32700, CurrentID=0, Status='live', Symbol='RpS5a', Aliases='Dmel_CG8922, CG8922, Dmel\\CG8922, M(1)15D, M(1)15d, M(1)34i28, M(1)o, M[[o]], RS5, RS5A_DROME, Rp S5a, RpS5, S5a, rpS5, rps5', description='Ribosomal protein S5a', other_designations='CG8922-PA|CG8922-PB|CG8922-PC|CG8922-PD|Minute(1)15D|RpS5a-PA|RpS5a-PB|RpS5a-PC|RpS5a-PD|minute 1(o)|ribosomal protein S5|ribosomal protein S5 gene|ribosomal protein S5a|transcript e', map_location='1-56.6 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17143588, end_position_on_the_genomic_accession=17148264, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32701 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32701, CurrentID=0, Status='live', Symbol='CG5010', Aliases='Dmel_ BcDNA:RE73921, Dmel\\CG5010, anon-15Dd', description='CG5010 gene product from transcript CG5010-RA', other_designations='CG5010-PA|CG5010-PB|anon-15Dd|transcript f', map_location='15E7-15E7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17149001, end_position_on_the_genomic_accession=17150647, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32703 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32703, CurrentID=0, Status='live', Symbol='baz', Aliases='Dmel_CG5055, BAZ, Baz, Baz/Par-3, Baz/Par3, Bazooka, CG5055, D-Par3, D-par3, Dmel\\CG5055, PAR-3, PAR3, Par-3, Par3, dPar-3, l(1)G0484, par-3', description='bazooka', other_designations='Bazooka-Par3|Bazooka/Par-3|Bazooka/Par3|CG5055-PA|CG5055-PB|CG5055-PC|CG5055-PD|baz-PA|baz-PB|baz-PC|baz-PD|par-3', map_location='1-56.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17160006, end_position_on_the_genomic_accession=17200705, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  32704 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32704, CurrentID=0, Status='live', Symbol='CG8918', Aliases='Dmel_ Dmel\\CG8918', description='CG8918 gene product from transcript CG8918-RA', other_designations='CG8918-PA', map_location='15F2-15F2', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17200726, end_position_on_the_genomic_accession=17202598, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32705 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32705, CurrentID=0, Status='live', Symbol='CG5070', Aliases='Dmel_ Dmel\\CG5070', description='CG5070 gene product from transcript CG5070-RA', other_designations='CG5070-PA', map_location='15F3-15F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17203112, end_position_on_the_genomic_accession=17203925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32706 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32706, CurrentID=0, Status='live', Symbol='CG32563', Aliases='Dmel_ CG5095, Dmel\\CG32563', description='CG32563 gene product from transcript CG32563-RA', other_designations='CG32563-PA|CG32563-PB|uncharacterized protein', map_location='15F3-15F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17208436, end_position_on_the_genomic_accession=17209063, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32707 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32707, CurrentID=0, Status='live', Symbol='CG12995', Aliases='Dmel_ Dmel\\CG12995', description='CG12995 gene product from transcript CG12995-RA', other_designations='CG12995-PA|uncharacterized protein', map_location='15F3-15F3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17209368, end_position_on_the_genomic_accession=17209997, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32708 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32708, CurrentID=0, Status='live', Symbol='CG18258', Aliases='Dmel_ CT41369, Dmel\\CG18258, anon-SAGE:Wang-051', description='CG18258 gene product from transcript CG18258-RA', other_designations='CG18258-PA|anon-SAGE:Wang-051|uncharacterized protein', map_location='15F3-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17212307, end_position_on_the_genomic_accession=17213977, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32709 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32709, CurrentID=0, Status='live', Symbol='CG5162', Aliases='Dmel_ Dmel\\CG5162', description='CG5162 gene product from transcript CG5162-RB', other_designations='CG5162-PA|CG5162-PB', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17215304, end_position_on_the_genomic_accession=17217365, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32710 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32710, CurrentID=0, Status='live', Symbol='CG12998', Aliases='Dmel_ BcDNA:RH61745, CT32204, Dmel\\CG12998', description='CG12998 gene product from transcript CG12998-RB', other_designations='CG12998-PA|CG12998-PB|uncharacterized protein', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17218667, end_position_on_the_genomic_accession=17219787, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  32711 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32711, CurrentID=0, Status='live', Symbol='CG5172', Aliases='Dmel_ BcDNA:RH06401, CT16557, Dmel\\CG5172', description='CG5172 gene product from transcript CG5172-RD', other_designations='CG5172-PC|CG5172-PD', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17221442, end_position_on_the_genomic_accession=17222651, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32713 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32713, CurrentID=0, Status='live', Symbol='CG10598', Aliases='Dmel_ Dmel\\CG10598, GCR 17, GCR-11, GCR11, GCR17, Pen15EF, anon-Pen15EF, pen15EF', description='CG10598 gene product from transcript CG10598-RC', other_designations='CG10598-PA|CG10598-PC|uncharacterized protein', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17225157, end_position_on_the_genomic_accession=17226867, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32714 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32714, CurrentID=0, Status='live', Symbol='CG10597', Aliases='Dmel_ Dmel\\CG10597', description='CG10597 gene product from transcript CG10597-RB', other_designations='CG10597-PB|uncharacterized protein', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17230275, end_position_on_the_genomic_accession=17231044, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32715 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32715, CurrentID=0, Status='live', Symbol='CG8915', Aliases='Dmel_ Dmel\\CG8915, cg8915', description='CG8915 gene product from transcript CG8915-RB', other_designations='CG8915-PA|CG8915-PB|helicasa', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17240468, end_position_on_the_genomic_accession=17245001, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32716 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32716, CurrentID=0, Status='live', Symbol='CG8675', Aliases='Dmel_ Dmel\\CG8675', description='CG8675 gene product from transcript CG8675-RA', other_designations='CG8675-PA', map_location='15F4-15F4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17248372, end_position_on_the_genomic_accession=17249502, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32718 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32718, CurrentID=0, Status='live', Symbol='f', Aliases='Dmel_CG42864, CG12996, CG42864, CG5424, Dmel\\CG42864, Dmel_CG12996, Dmel_CG5424, X1', description='forked', other_designations='CG42864-PB|CG42864-PC|CG42864-PD|CG42864-PF|CG42864-PG|CG42864-PH|CG42864-PI|CG42864-PJ|CG42864-PK|f-PB|f-PC|f-PD|f-PF|f-PG|f-PH|f-PI|f-PJ|f-PK', map_location='1-56.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17232942, end_position_on_the_genomic_accession=17280964, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  32719 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32719, CurrentID=0, Status='live', Symbol='CG8664', Aliases='Dmel_ Dmel\\CG8664', description='CG8664 gene product from transcript CG8664-RB', other_designations='CG8664-PA|CG8664-PB', map_location='15F7-15F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17280918, end_position_on_the_genomic_accession=17282726, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32720 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32720, CurrentID=0, Status='live', Symbol='CG8661', Aliases='Dmel_ BcDNA:LP01183, Dmel\\CG8661', description='CG8661 gene product from transcript CG8661-RA', other_designations='CG8661-PA', map_location='15F7-15F7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17283762, end_position_on_the_genomic_accession=17284500, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32721 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32721, CurrentID=0, Status='live', Symbol='Fim', Aliases='Dmel_CG8649, CG8649, Dmel\\CG8649, FIM, fim', description='Fimbrin', other_designations='CG8649-PA|CG8649-PC|CG8649-PD|Fim-PA|Fim-PC|Fim-PD|fimbrin|plastin', map_location='15F8-16A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17285704, end_position_on_the_genomic_accession=17301929, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32722 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32722, CurrentID=0, Status='live', Symbol='CG5445', Aliases='Dmel_ Dmel\\CG5445', description='CG5445 gene product from transcript CG5445-RG', other_designations='CG5445-PA|CG5445-PB|CG5445-PD|CG5445-PF|CG5445-PG', map_location='16A1-16A1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17282496, end_position_on_the_genomic_accession=17305475, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  32723 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32723, CurrentID=0, Status='live', Symbol='B-H2', Aliases='Dmel_CG5488, B, BH2, Bar, Bar-H2, BarH, Barh2, CG5488, Dmel\\CG5488, bar-h2', description='BarH2', other_designations='B-H2-PA|B-H2-PB|Bar|CG5488-PA|CG5488-PB', map_location='1-56.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17314581, end_position_on_the_genomic_accession=17324175, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32724 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32724, CurrentID=0, Status='live', Symbol='B-H1', Aliases='Dmel_CG5529, BH1, Bar, Bar H1, Bar-H1, BarH, BarHI, Barh1, CG5529, Dmel\\CG5529, bar-h1, barH1, barh1', description='BarH1', other_designations='B-H1-PA|B-H1-PB|Bar|Bar-H1|CG5529-PA|CG5529-PB', map_location='16A4-16A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17397478, end_position_on_the_genomic_accession=17403278, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32725 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32725, CurrentID=0, Status='live', Symbol='CG8611', Aliases='Dmel_ BcDNA:GH02833, DmRH11, Dmel\\CG8611, cg8611', description='CG8611 gene product from transcript CG8611-RB', other_designations='CG8611-PA|CG8611-PB|CG8611-PC', map_location='16A5-16A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17413267, end_position_on_the_genomic_accession=17422145, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  32726 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32726, CurrentID=0, Status='live', Symbol='CG5613', Aliases='Dmel_ Dmel\\CG5613', description='CG5613 gene product from transcript CG5613-RA', other_designations='CG5613-PA|CG5613-PB', map_location='16A5-16A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17423591, end_position_on_the_genomic_accession=17428735, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32727 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32727, CurrentID=0, Status='live', Symbol='p-cup', Aliases='Dmel_CG12993, CG12993, Dmel\\CG12993', description='presidents-cup', other_designations='CG12993-PA|p-cup-PA', map_location='16A5-16A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17446350, end_position_on_the_genomic_accession=17448774, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  32728 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32728, CurrentID=0, Status='live', Symbol='CG8568', Aliases='Dmel_ Dmel\\CG8568', description='CG8568 gene product from transcript CG8568-RA', other_designations='CG8568-PA', map_location='16A5-16A5', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17448990, end_position_on_the_genomic_accession=17452543, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32729 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32729, CurrentID=0, Status='live', Symbol='CG43658', Aliases='Dmel_ CG12432, CG8557, Dmel\\CG43658, Dmel_CG12432, Dmel_CG8557', description='CG43658 gene product from transcript CG43658-RE', other_designations='CG43658-PA|CG43658-PB|CG43658-PC|CG43658-PD|CG43658-PE|CG43658-PF|CG43658-PG|CG43658-PH|uncharacterized protein', map_location='16B3-16B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17456169, end_position_on_the_genomic_accession=17561180, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  32731 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32731, CurrentID=0, Status='live', Symbol='ppk23', Aliases='Dmel_CG8527, CG8527, Dmel\\CG8527, PPK23, Ppk23', description='pickpocket 23', other_designations='CG8527-PA|CG8527-PB|CG8527-PC|CG8527-PD|ppk23-PA|ppk23-PB|ppk23-PC|ppk23-PD', map_location='16B4-16B4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17566552, end_position_on_the_genomic_accession=17569228, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  32732 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32732, CurrentID=0, Status='live', Symbol='Ankle2', Aliases='Dmel_CG8465, CG8465, Dmel\\CG8465, anon-WO03040301.175, anon-WO03040301.177, anon-WO03040301.179, dAnkle2, l(1)G0222, l(1)G0316', description='ANKLE2 ortholog', other_designations='Ankle2-PD|Ankle2-PE|Ankle2-PF|Ankle2-PG|CG8465-PD|CG8465-PE|CG8465-PF|CG8465-PG|l(1)G0222|lethal (1) G0222', map_location='16B5-16B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17571854, end_position_on_the_genomic_accession=17579229, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  32733 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32733, CurrentID=0, Status='live', Symbol='CG12992', Aliases='Dmel_ Dmel\\CG12992', description='CG12992 gene product from transcript CG12992-RE', other_designations='CG12992-PB|CG12992-PC|CG12992-PD|CG12992-PE|uncharacterized protein', map_location='16B6-16B6', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17588865, end_position_on_the_genomic_accession=17591706, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32734 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32734, CurrentID=0, Status='live', Symbol='CG12991', Aliases='Dmel_ Dmel\\CG12991', description='CG12991 gene product from transcript CG12991-RE', other_designations='CG12991-PA|CG12991-PB|CG12991-PC|CG12991-PD|CG12991-PE|CG12991-PF|uncharacterized protein', map_location='16B6-16B7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17579861, end_position_on_the_genomic_accession=17599314, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32737 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32737, CurrentID=0, Status='live', Symbol='stas', Aliases='Dmel_CG8408, CG8408, Dmel\\CG8408, Tmem41b', description='Stasimon', other_designations='CG8408-PA|CG8408-PB|stas-PA|stas-PB|stasimon|stymied in smn', map_location='16B9-16B9', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17623221, end_position_on_the_genomic_accession=17627895, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32738 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32738, CurrentID=0, Status='live', Symbol='CG8326', Aliases='Dmel_ BEST:LD29336, Dmel\\CG8326', description='CG8326 gene product from transcript CG8326-RA', other_designations='CG8326-PA', map_location='16B9-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17628164, end_position_on_the_genomic_accession=17629492, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  32739 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32739, CurrentID=0, Status='live', Symbol='corolla', Aliases='Dmel_CG8316, CG8316, Dmel\\CG8316, mei-39', description='CG8316 gene product from transcript CG8316-RB', other_designations='CG8316-PA|CG8316-PB|corolla-PA|corolla-PB|meiotic 39', map_location='16B10-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17629791, end_position_on_the_genomic_accession=17632047, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32740 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32740, CurrentID=0, Status='live', Symbol='ND-24', Aliases='Dmel_CG5703, BEST:LP01573, CG 5703, CG5703, Dmel\\CG5703, NUHM', description='NADH dehydrogenase (ubiquinone) 24 kDa subunit', other_designations='CG5703-PA|CG5703-PB|COMPLEX I: 24 kDa subunit|NADH ubiquinone oxidoreductase 24 kDa|NADH-ubiquinone oxidoreductase 24 kDa subunit|NADH:ubiquinone oxidoreductase 24 kDa subunit|ND-24-PA|ND-24-PB', map_location='16B10-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17632089, end_position_on_the_genomic_accession=17633509, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32741 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32741, CurrentID=0, Status='live', Symbol='CG8289', Aliases='Dmel_ Dmel\\CG8289', description='CG8289 gene product from transcript CG8289-RB', other_designations='CG8289-PA|CG8289-PB', map_location='16B10-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17633691, end_position_on_the_genomic_accession=17636044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32742 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32742, CurrentID=0, Status='live', Symbol='CG5800', Aliases='Dmel_ DmRH14, Dmel\\CG5800, cg5800', description='CG5800 gene product from transcript CG5800-RA', other_designations='CG5800-PA', map_location='16B10-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17636281, end_position_on_the_genomic_accession=17639003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32743 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32743, CurrentID=0, Status='live', Symbol='RhoGAPp190', Aliases='Dmel_CG32555, 8267, BcDNA:GH17919, BcDNA:GM05133, CG32555, CG8240, CG8267, D-p190, DRhoGAP, Dmel\\CG32555, RhoGAP, RhoGAP P190, RhoGAP-16B12, mdcds_56566, p190, p190 RhoGAP, p190-RhoGAP, p190RhoGAP', description='Rho GTPase activating protein p190', other_designations='CG32555-PA|CG32555-PB|CG32555-PC|CG32555-PD|CG32555-PE|CG32555-PF|CG32555-PG|RhoGAPp190-PA|RhoGAPp190-PB|RhoGAPp190-PC|RhoGAPp190-PD|RhoGAPp190-PE|RhoGAPp190-PF|RhoGAPp190-PG|Rp190RhoGAP', map_location='16B10-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17639147, end_position_on_the_genomic_accession=17652562, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  32745 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32745, CurrentID=0, Status='live', Symbol='IntS2', Aliases='Dmel_CG8211, CG8211, Dmel\\CG8211, Int2', description='Integrator 2', other_designations='CG8211-PA|IntS2-PA|integrator 2', map_location='16B10-16B10', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17652761, end_position_on_the_genomic_accession=17656818, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32746 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32746, CurrentID=0, Status='live', Symbol='beta-Spec', Aliases='Dmel_CG5870, B-spec, CG5870, Dmel\\CG5870, SPEC8, Spec, Spec-beta, Specbeta, anon-EST:fe1B3, b spectrin, b-Spec, beta, beta-spec, betaSpec, betaspec, i168, i6, kus, l(1)G0074, l(1)G0108, l(1)G0198', description='beta Spectrin', other_designations='CG5870-PA|CG5870-PB|CG5870-PC|anon-fast-evolving-1B3|beta spectrin|beta-Spec-PA|beta-Spec-PB|beta-Spec-PC|beta-spectrin|karussell|spectrin', map_location='16B10-16B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17657372, end_position_on_the_genomic_accession=17669371, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  32747 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32747, CurrentID=0, Status='live', Symbol='CG12990', Aliases='Dmel_ Dmel\\CG12990', description='CG12990 gene product from transcript CG12990-RA', other_designations='CG12990-PA|uncharacterized protein', map_location='16B12-16B12', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17669208, end_position_on_the_genomic_accession=17671888, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32748 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32748, CurrentID=0, Status='live', Symbol='chas', Aliases='Dmel_CG32556, CG12987, CG12988, CG12989, CG32556, Dmel\\CG32556, EP-631', description='chascon', other_designations='CG32556-PA|CG32556-PB|CG32556-PC|CG32556-PD|CG32556-PE|chas-PA|chas-PB|chas-PC|chas-PD|chas-PE', map_location='16B12-16C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17672458, end_position_on_the_genomic_accession=17698730, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  32751 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32751, CurrentID=0, Status='live', Symbol='CG8188', Aliases='Dmel_ Dmel\\CG8188, Ubc, Ube2S', description='CG8188 gene product from transcript CG8188-RD', other_designations='CG8188-PA|CG8188-PB|CG8188-PC|CG8188-PD', map_location='16C1-16C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17698917, end_position_on_the_genomic_accession=17703024, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32752 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32752, CurrentID=0, Status='live', Symbol='par-6', Aliases='Dmel_CG5884, CG5884, DPar-6, DPar6, Dm-Par-6, Dm-Par6, DmPAR-6, DmPar-6, DmPar6, Dmel\\CG5884, Dmpar6, PAR-6, PAR6, Par, Par-6, Par6, dPAR-6, dPar-6, dmPar-6, par, par6', description='CG5884 gene product from transcript CG5884-RA', other_designations='CG5884-PA|CG5884-PB|Par-6|par-6-PA|par-6-PB', map_location='16C1-16C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17703352, end_position_on_the_genomic_accession=17707096, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32753 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32753, CurrentID=0, Status='live', Symbol='CG8173', Aliases='Dmel_ Dmel\\CG8173', description='CG8173 gene product from transcript CG8173-RA', other_designations='CG8173-PA', map_location='16C1-16C1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17707401, end_position_on_the_genomic_accession=17709653, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32754 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32754, CurrentID=0, Status='live', Symbol='CG42684', Aliases='Dmel_ CG15816, CG32559, CG32560, CG42270, CG5898, CG5960, Dmel\\CG42684, Dmel_CG15816, Dmel_CG32559, Dmel_CG32560, Dmel_CG42270', description='CG42684 gene product from transcript CG42684-RB', other_designations='CG42684-PA|CG42684-PB|CG42684-PC|CG42684-PD|CG42684-PE|CG42684-PF|CG42684-PG|CG42684-PH|CG42684-PI|uncharacterized protein', map_location='16C5-16C8', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17710144, end_position_on_the_genomic_accession=17752007, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  38227 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=38227, CurrentID=0, Status='live', Symbol='CG13933', Aliases='Dmel_ Dmel\\CG13933, dmASBa', description='CG13933 gene product from transcript CG13933-RB', other_designations='CG13933-PA|CG13933-PB|uncharacterized protein', map_location='62B3-62B4', chromosome='3L', genomic_nucleotide_accession_version='NT_037436.4', start_position_on_the_genomic_accession=1771809, end_position_on_the_genomic_accession=1775147, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  32757 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32757, CurrentID=0, Status='live', Symbol='unc-4', Aliases='Dmel_CG6269, CG6269, DPHD-1, Dmel\\CG6269, D Unc4', description='CG6269 gene product from transcript CG6269-RC', other_designations='CG6269-PB|CG6269-PC|CG6269-PD|paired HomeoDomain 1|unc-4-PB|unc-4-PC|unc-4-PD', map_location='16C10-16D1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17768569, end_position_on_the_genomic_accession=17778708, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32758 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32758, CurrentID=0, Status='live', Symbol='OdsH', Aliases='Dmel_CG6352, CG6352, Dmel\\CG6352[mel], odsH, odsh, OdsH', description='Ods-site homeobox', other_designations='CG6352-PA|OdsH-PA', map_location='16D1-16D3', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17791010, end_position_on_the_genomic_accession=17814011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  32759 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32759, CurrentID=0, Status='live', Symbol='CG12986', Aliases='Dmel_ Dmel\\CG12986', description='CG12986 gene product from transcript CG12986-RA', other_designations='CG12986-PA|uncharacterized protein', map_location='16D4-16D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17819155, end_position_on_the_genomic_accession=17819685, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  32760 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32760, CurrentID=0, Status='live', Symbol='Socs16D', Aliases='Dmel_CG8146, CG8146, Dmel\\CG8146, SOCS, SOCS16D, Socs 16D, Socs16E, dmSocs16D, socs16D', description='Suppressor of Cytokine Signaling at 16D', other_designations='CG8146-PA|CG8146-PB|Socs16D-PA|Socs16D-PB|suppressor of cytokine signaling at 16D|suppressor of cytokine signaling at 16E', map_location='16D4-16D4', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17820764, end_position_on_the_genomic_accession=17827564, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  32761 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32761, CurrentID=0, Status='live', Symbol='CG6398', Aliases='Dmel_ Dmel\\CG6398', description='CG6398 gene product from transcript CG6398-RB', other_designations='CG6398-PA|CG6398-PB', map_location='16D5-16D7', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17830828, end_position_on_the_genomic_accession=17841705, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32762 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32762, CurrentID=0, Status='live', Symbol='e(y)1', Aliases='Dmel_CG6474, CG6474, Dmel\\CG6474, E(y)1, TAF40, TAF40/42, TAF40/42/e(y)1, TAF42, TAF9, TAFII40, TAFII40(42), TAF[II]40, TAF[[II]], TAF[[II]]40, TAF[[II]]40/42, TAF[[II]]42, TFIID, TFIID 42, TFIID-p42, Taf40, Taf42, Taf9, Taf[[II]]40, d40, dTAF42, dTAF9, dTAFII42, dTAF[[40]], dTAF[[II]]40, dTAF[[II]]42, dmTAF9, dmTaf9, dme-TAFII40, e(y)4, e(y)[1], p42, taf40/e(y)1', description='enhancer of yellow 1', other_designations='CG6474-PA|TBP-associated factor|TBP-associated factor 40kD|e(y)1-PA|enhancer of yellow 4', map_location='1-57.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17842161, end_position_on_the_genomic_accession=17843348, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  32763 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32763, CurrentID=0, Status='live', Symbol='CG8142', Aliases='Dmel_ DmRFC2, Dmel\\CG8142, RFC2, RfC', description='CG8142 gene product from transcript CG8142-RA', other_designations='CG8142-PA', map_location='16E1-16E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17843343, end_position_on_the_genomic_accession=17844704, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  32764 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32764, CurrentID=0, Status='live', Symbol='Ucp4A', Aliases='Dmel_CG6492, CG6492, DmUCP4A, Dmel\\CG6492, UCP4A, Ucp4a, dmUCP4a, ucp4a', description='CG6492 gene product from transcript CG6492-RB', other_designations='CG6492-PA|CG6492-PB|CG6492-PC|Ucp4A-PA|Ucp4A-PB|Ucp4A-PC', map_location='16E1-16E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17844949, end_position_on_the_genomic_accession=17848223, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  32765 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32765, CurrentID=0, Status='live', Symbol='CG15814', Aliases='Dmel_ Dmel\\CG15814', description='CG15814 gene product from transcript CG15814-RB', other_designations='CG15814-PA|CG15814-PB|CG15814-PC|CG15814-PD|uncharacterized protein', map_location='16E1-16E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17848110, end_position_on_the_genomic_accession=17850342, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  32766 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32766, CurrentID=0, Status='live', Symbol='Spt7', Aliases='Dmel_CG6506, CG6506, Dmel\\CG6506, cg6506, d', description='CG6506 gene product from transcript CG6506-RA', other_designations='CG6506-PA|Spt7-PA', map_location='16E1-16E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17850460, end_position_on_the_genomic_accession=17851889, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  32767 : NtData(tax_id=7227, Org_name='Drosophila melanogaster', GeneID=32767, CurrentID=0, Status='live', Symbol='CG32554', Aliases='Dmel_ CG7871, Dmel\\CG32554', description='CG32554 gene product from transcript CG32554-RA', other_designations='CG32554-PA|uncharacterized protein', map_location='16E1-16E1', chromosome='X', genomic_nucleotide_accession_version='NC_004354.4', start_position_on_the_genomic_accession=17851869, end_position_on_the_genomic_accession=17852685, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
}

